# Copyright 2005-2016 Canonical Limited.  All rights reserved.

import os
import subprocess
import multiprocessing

import transaction

from storm.uri import URI

from canonical.lib.undef import Undef


class BenchmarkData(object):
    """Manage database data common to one or more benchmarks.

    This helper will programmatically create the needed data from scratch
    if it's the first run of the benchmarks requiring it, and "cache" it
    using pg_dump for follow-up runs (where it will be loaded using pg_restore
    rather than created from scratch).

    @cvar name: A unique name for this benchmark data, that will be
        used as reference by tests needing it. Must be defined by-subclasses.
    """

    name = Undef()

    def __init__(self, zstorm):
        """
        @param zstorm: The ZStorm instance to use to get the URIs of the
            various databases.
        """
        self._zstorm = zstorm

    def identifier(self):
        """Return a string identifying this dataset instance.

        The identifier combines the dataset name and the current patch level
        of the database.
        """
        main_store = self._zstorm.get("main")
        result = main_store.execute("SELECT max(version) FROM patch")
        return "{}-{}".format(self.name, result.get_one()[0])

    def create(self):
        """Create the database data for the benchmark from scratch."""
        self._create_data()
        transaction.commit()

    def ensure(self, build_area):
        """Ensure that the database is initialized with the benchmark data.

        If a dump is available (from previous runs) it will be loaded,
        otherwise the data will be created programmatically from scratch (and
        dumped for follow-up runs).

        @param build_area: The directory to use to "cache" PostgreSQL dumps.
        """
        dump_dir = self._get_dump_dir(build_area)
        if os.path.exists(dump_dir):
            self._restore_all(dump_dir)
        else:
            self.create()
            self._dump_all(dump_dir)

    def _create_data(self):
        """Hook for sub-classes, to implement custom datasets."""
        raise RuntimeError("Must be implemented by sub-classes.")

    def _get_dump_dir(self, build_area):
        """Get the dump directory path.

        The directory name will contain the patch level, so only dumps against
        the current patch level will be used.
        """
        return os.path.join(build_area, self.identifier())

    def _get_storm_uris(self):
        """Return a list of storm.uri.URIs for all available databases."""
        uris = []
        default_uris = self._zstorm.get_default_uris()
        for name, uri_string in default_uris.iteritems():
            if name.endswith("-autocommit"):
                # Skip stores set as autocommit, like session-autocommit, since
                # they are duplicate
                continue
            uris.append(URI(uri_string))
        return uris

    def _dump_all(self, dump_dir):
        """Dump all available databases.

        @param dump_dir: The directory to create the dumps in.
        """
        os.makedirs(dump_dir)

        for uri in self._get_storm_uris():
            self._dump_one(dump_dir, uri)

    def _dump_one(self, dump_dir, uri):
        """Dump a single database.

        @param dump_dir: The directory to create the dump in.
        @param uri: The storm.uri.URI instance of the database to dump.
        """
        dump_path = os.path.join(dump_dir, uri.database)
        pg_dump = [
            "pg_dump",
            "--jobs={}".format(multiprocessing.cpu_count()),
            "--format=directory",
            "--data-only",
            "--no-acl",
            "--disable-triggers",
            "--exclude-table=patch",
            "--file={}".format(dump_path),
            uri.database,
        ]
        # Silence pg_dump redirecting stderr to /dev/null
        with open(os.devnull, "w") as stderr:
            subprocess.check_call(pg_dump, stderr=stderr)

    def _restore_all(self, dump_dir):
        """Restore all available databases.

        @param dump_dir: The directory to restore the dumps from.
        """
        for uri in self._get_storm_uris():
            self._restore_one(dump_dir, uri)
        transaction.commit()

    def _restore_one(self, dump_dir, uri):
        """Dump a single database.

        @param dump_dir: The directory to restore the dump from.
        @param uri: The storm.uri.URI instance of the database to restore.
        """
        dump_path = os.path.join(dump_dir, uri.database)
        pg_restore = [
            "pg_restore",
            "--jobs={}".format(multiprocessing.cpu_count()),
            "--format=directory",
            "--disable-triggers",
            "--no-owner",
            "--data-only",
            "--dbname={}".format(uri.database),
            dump_path,
        ]

        # Run pg_restore redirecting stderr noise to /dev/null
        with open(os.devnull, "w") as stderr:
            subprocess.check_call(pg_restore, stderr=stderr)
