import os
import transaction
import shutil

from testresources import TestResourceManager

from canonical.landscape.testing.resources import (
    ZStormResourceManager,
    DispatcherResourceManager,
    SetupResourceManager,
)

from benchmark.testing.registry import DATA_REGISTRY


class BenchmarkDataResourceManager(TestResourceManager):
    """Manage a BenchmarkData resource.

    This manager will make sure that the data created the BenchmarkData
    subclass matching the given identifier is available in the database.

    It will optimise multiple runs of the same benchmark by loading database
    dumps instead of re-creating the data from scratch programmatically.
    """

    resources = [
        ("zstorm", ZStormResourceManager(group="benchmark")),
        ("dispatcher", DispatcherResourceManager()),
        ("setup", SetupResourceManager()),
    ]

    def __init__(self, name):
        """
        @param name: The name of the BenchmarkData class to use to generate
            the benchmark dataset.
        """
        super(BenchmarkDataResourceManager, self).__init__()
        self._name = name

    def make(self, dependencies):
        zstorm = dependencies["zstorm"]
        _, zstorm_manager = self.resources[0]

        # Wipe any cache if we're told to force creation
        if self.options.force_create:
            shutil.rmtree(self.options.build_area, ignore_errors=True)

        # Ensure that the build directory exists
        if not os.path.exists(self.options.build_area):
            os.makedirs(self.options.build_area)

        # Lookup the benchmark data class matching the given identifier.
        data_class = DATA_REGISTRY[self._name]
        data = data_class(zstorm)
        identifier = data.identifier()

        if self._get_last_identifier() != identifier:
            self._make_data(data, zstorm, zstorm_manager)

        # Remember what was the last benchmark dataset we handled.
        self._set_last_identifier(identifier)

        return data

    def _make_data(self, data, zstorm, zstorm_manager):
        # We need to first get all stores from ZStorm so they get
        # registered as side-effect.
        for name in zstorm.get_default_uris().keys():
            if name.endswith("-autocommit"):
                continue  # Skip autocommit stores, has they are duplicate
            zstorm.get(name)

        # Force wiping all the data from all stores.
        zstorm_manager.force_delete = True
        zstorm_manager.clean(zstorm)
        zstorm_manager.force_delete = False
        transaction.commit()

        if self.options.no_dumps:
            # Force creating the data from scratch.
            data.create()
        else:
            # Ensure that the benchmark data is there, possibly making use
            # of the cached dumps.
            data.ensure(self.options.build_area)

        # This will make the ZStorm manager skip deleting the data we
        # created.
        zstorm_manager._commits.clear()

    def _get_last_identifier_path(self):
        """Get file path storing the identifier of the last data we handled."""
        return os.path.join(self.options.build_area, "last-identifier")

    def _get_last_identifier(self):
        """Get the identifier of the last benchmark data we handled."""
        last_data_path = self._get_last_identifier_path()
        if not os.path.exists(last_data_path):
            return ""
        with open(last_data_path) as fd:
            return fd.read().strip()

    def _set_last_identifier(self, identifier):
        """Update the identifier of the last benchmark data we handled."""
        with open(self._get_last_identifier_path(), "w") as fd:
            fd.write("{}\n".format(identifier))
