# Copyright 2005-2016 Canonical Limited.  All rights reserved.
from inspect import getargspec

from testtools import (
    MultiTestResult,
    ExtendedToStreamDecorator,
    TextTestResult as TestToolsTextTestResult,
)

from subunit import StreamResultToBytes

# We'll just do this work once
_delta_to_float_args = getargspec(TestToolsTextTestResult._delta_to_float).args


class TestResult(MultiTestResult):
    """Override stock MultiTestResult to provide testresources output."""

    def __init__(self, stream, failfast=None, subunit=None):
        results = [TextTestResult(stream)]
        if subunit:
            result = StreamResultToBytes(open(subunit, "w"))
            result = ExtendedToStreamDecorator(result)
            results.append(result)
        super(TestResult, self).__init__(*results)
        # XXX MultiResultError doesn't play well with failfast, as it doesn't
        #     take it as init parameter and ends up resetting it.
        self.failfast = failfast


class TextTestResult(TestToolsTextTestResult):
    """Override stock TextTestResult to provide more detailed output."""

    def startTest(self, test):
        super(TextTestResult, self).startTest(test)
        self.__start_test = self._now()

    def addSuccess(self, test, details=None):
        super(TextTestResult, self).addSuccess(test, details=details)
        self._show_test(test, "OK")

    def addError(self, test, err=None, details=None):
        super(TextTestResult, self).addError(test, err=err, details=details)
        self._show_test(test, "ERROR")

    def addFailure(self, test, err=None, details=None):
        super(TextTestResult, self).addFailure(test, err=err, details=details)
        self._show_test(test, "FAILURE")

    def addSkip(self, test, reason=None, details=None):
        super(TextTestResult, self).addSkip(test, reason=reason, details=details)
        self._show_test(test, "SKIPPED")

    def _show_test(self, test, label):
        timing = self._delta_to_float(self._now() - self.__start_test)
        self.stream.write("{} {} {:.3f}\n".format(test.id(), label, timing))

    def _delta_to_float(self, a_timedelta):
        # Recent version of testtools have added a precision argument
        # to the _delta_to_float method, this works around some issues
        # with Python3.  Becuase we'd like our code to work with older
        # versions of the library still we override the method and
        # pass the correct args list to the superclass implementation.
        args = [a_timedelta]
        if "precision" in _delta_to_float_args:
            # 3 is the value for precision testtools uses internally
            args.append(3)
        return super(TextTestResult, self)._delta_to_float(*args)
