# Copyright 2005-2016 Canonical Limited.  All rights reserved.

import os

import transaction

from fixtures import TempDir

from canonical.landscape.testing.model import ModelTest
from canonical.landscape.model.main.account import get_account_by_name

from benchmark.testing.options import Options
from benchmark.testing.resources import BenchmarkDataResourceManager
from benchmark.testing.registry import DATA_REGISTRY
from benchmark.testing.tests.data import DummyBenchmarkData


class BenchmarkDataResourceManagerTest(ModelTest):
    def setUp(self):
        super(BenchmarkDataResourceManagerTest, self).setUp()
        self.build_area_dir = self.useFixture(TempDir())
        self.options = Options()
        self.options.build_area = self.build_area_dir.path

        DATA_REGISTRY["dummy"] = DummyBenchmarkData
        self.addCleanup(DATA_REGISTRY.pop, "dummy")
        self.manager = BenchmarkDataResourceManager("dummy")
        self.manager.options = self.options

        # Replace the ZStorm manager we depend on with the unit-test one, so we
        # don't point to the benchmark database group. The other resources
        # aren't really needed, for the sake of unit tests, since they get
        # already setup by ModelTest
        self.manager.resources = [("zstorm", self.get_zstorm_manager())]

        self.dependencies = {"zstorm": self.zstorm}

    def tearDown(self):
        # Force clearing the database, since the resource by-passes this
        # behavior
        zstorm_manager = self.get_zstorm_manager()
        zstorm_manager.force_delete = True
        zstorm_manager.clean(self.zstorm)
        zstorm_manager.force_delete = False
        super(BenchmarkDataResourceManagerTest, self).tearDown()

    def test_make(self):
        """
        The resource builds the benchmark dataset identified by the given name.
        """
        self.manager.make(self.dependencies)
        account = get_account_by_name("account")
        self.assertEqual("Account", account.title)

    def test_make_no_delete(self):
        """
        The resource prevents the ZStorm manager from deleting the data
        it creates.
        """
        self.manager.make(self.dependencies)
        zstorm_manager = self.get_zstorm_manager()
        zstorm_manager.clean(self.zstorm)
        account = get_account_by_name("account")
        self.assertEqual("Account", account.title)

    def test_make_no_data_change(self):
        """
        If the last benchmark dataset is the same we're requested to build,
        then nothing will be done.
        """
        self.manager.make(self.dependencies)
        account = get_account_by_name("account")
        account_id = account.id

        class FailingDummyBenchmarkData(DummyBenchmarkData):
            def ensure(self, build_area):
                raise AssertionError("Shouldn't be called")

        DATA_REGISTRY["dummy"] = FailingDummyBenchmarkData

        self.manager.make(self.dependencies)
        transaction.commit()

        account = get_account_by_name("account")
        self.assertEqual(account_id, account.id)

    def test_force_create(self):
        """
        When the --force-create option is passed, data gets re-created.
        """
        self.manager.make(self.dependencies)
        account = get_account_by_name("account")
        account_id = account.id

        self.options.force_create = True

        self.manager.make(self.dependencies)

        account = get_account_by_name("account")
        self.assertNotEqual(account_id, account.id)

    def test_no_dumps(self):
        """
        When the --no-dumps option is passed, no dumps will be created or
        used.
        """
        self.options.no_dumps = True

        self.manager.make(self.dependencies)

        main_store = self.get_main_store()
        result = main_store.execute("SELECT max(version) FROM patch")
        patch_level = result.get_one()[0]
        dump_dir = self.build_area_dir.join("dummy-{}".format(patch_level))
        self.assertFalse(os.path.exists(dump_dir))
