# Copyright 2005-2017 Canonical Limited.  All rights reserved.

"""
Job handler: retrieve tasks from the AMQP queue and push them to job action
handlers.
"""
from builtins import str
from builtins import object
import json

from twisted.internet.defer import inlineCallbacks, maybeDeferred, fail
from twisted.logger import Logger, globalLogPublisher

from txamqp.content import Content
from txamqp.client import Closed
from txamqp.queue import Closed as QueueClosed

from canonical.amqp.message.publisher import Publisher
from canonical.lib.serialize import json_loads


__all__ = ["JobHandler"]


def append_traceback_log_observer(event):
    """
    Log observer to append traceback to failure log events.

    Traceback are only added from FileLogObserver, but not when emitting
    to the globalLogPublisher, thus this observer.
    """
    if "log_failure" in event:
        try:
            traceback = event["log_failure"].getTraceback()
        except Exception:
            traceback = "(unable to obtain traceback from event)\n"
        event["log_traceback"] = traceback
        event["log_format"] += "\n{log_traceback}"
    globalLogPublisher(event)


class JobHandler(object):
    """
    An AMQP consumer which handles messages sent over a "job-handler" queue to
    launch jobs and forward their results. A message on the queue should
    contain an "action" and an "uuid": the first one will be used to determine
    which tasks to run, the second one to know to which queue send the result.
    It can also contain parameters for the action.

    @ivar accumulator_reconnection_delay: The delay in seconds between each try
        to connect to the broker as an accumulator.
    @ivar default_accumulator_delay: The delay in seconds for message
        accumulation when not specified.
    """

    logger = Logger(observer=append_traceback_log_observer)

    def __init__(
        self,
        registry,
        prefix,
        clock,
        accumulator_reconnection_delay,
        default_accumulator_delay,
        publisher=None,
    ):
        self._registry = registry
        self._prefix = prefix
        self._channel = None
        self._client = None
        self._clock = clock
        if publisher is None:
            publisher = Publisher(prefix)
        self._publisher = publisher
        self._accumulator_reconnection_delay = accumulator_reconnection_delay
        self._default_accumulator_delay = default_accumulator_delay

    def disconnected(self):
        """Called when the connection with the AMQP broker is lost.

        This is invoked by C{txlongpoll.client.AMQFactory} when the connection
        with rabbitmq is lost for some reason. Eventually it should be restored
        and the C{connected} method called.
        """
        self._channel = None
        self._client = None
        self._publisher.disconnected()

    def connected(self, connection_data):
        """Start the job handler and the accumulator."""
        (client, channel) = connection_data
        self.logger.debug("Job handler connected to AMQP broker")
        self._client = client
        self._channel = channel

        self._publisher.connected((client, channel))
        self.run_handler(client, channel).addErrback(self._catch_closed)
        self.run_accumulator(client).addErrback(self._catch_closed)

    def _catch_closed(self, failure):
        failure.trap(QueueClosed)

    @inlineCallbacks
    def run_handler(self, client, channel):
        """
        Declare the handler exchange and the queue, and consumes messages from
        it.

        When a message is received, the appropriate job handler is found, and
        run inside a thread.

        @rtype: Deferred.
        @raises txamqp.client.Closed: (Deferred) if the connection to the
            broker is lost.
        @raises txamqp.queue.Closed: (Deferred) if the queue is closed.
        """
        queue_name = "%s.job-handler-queue" % self._prefix
        exchange_name = "%s.job-handler-exchange" % self._prefix

        yield channel.exchange_declare(exchange=exchange_name, type="direct")
        yield channel.queue_declare(queue=queue_name, durable=True)
        yield channel.queue_bind(
            queue=queue_name, exchange=exchange_name, routing_key=queue_name
        )
        yield channel.basic_consume(
            consumer_tag="%s.job-handler-tag" % self._prefix, queue=queue_name
        )

        queue = yield client.queue("%s.job-handler-tag" % self._prefix)

        while True:
            msg = yield queue.get()
            payload = json_loads(msg.content.body)
            action = payload.pop("action")
            uuid = payload.pop("uuid")
            namespace = payload.pop("namespace", "")
            kwargs = dict((str(key), value) for key, value in list(payload.items()))
            self.logger.info(
                "Starting job {uuid}/{action}/{namespace} {args!r}",
                uuid=uuid,
                action=action,
                namespace=namespace,
                args=kwargs,
            )
            runner = self._registry.get(action)
            if runner is None:
                deferred = fail(RuntimeError("No handler for '%s'" % action))
            else:
                deferred = maybeDeferred(runner.run, **kwargs)

            deferred.addCallback(self._handle_result, uuid, action, namespace, kwargs)
            deferred.addErrback(self._handle_error, uuid, action, namespace, kwargs)
            deferred.addBoth(self._ack, channel, delivery_tag=msg.delivery_tag)

    def _handle_result(self, result, uuid, action, namespace, args):
        log_template = "Finished job {uuid}/{action}/{namespace} {args!r}"
        if any(result):
            log_template += " with result {result!r}"
        self.logger.info(
            log_template,
            uuid=uuid,
            action=action,
            namespace=namespace,
            args=args,
            result=result,
        )
        if uuid and result[0] is not None:
            self._publisher.publish_action(result, uuid, action)
        if namespace and result[1] is not None:
            self._publisher.publish_notification(result[1], uuid, action, namespace)

    def _handle_error(self, error, uuid, action, namespace, args):
        self.logger.failure(
            "Error while handling job {uuid}/{action}/{namespace} {args!r}",
            failure=error,
            uuid=uuid,
            action=action,
            namespace=namespace,
            args=args,
        )
        self._publisher.publish_error(error.value, uuid, action)

    def _ack(self, _, channel, delivery_tag=None):
        """Notify the AMQP broker that we processed the given message.

        This is called after job completes (either successfully or not) and
        will make Rabbit actually discard the message from the queue. It's
        important to run this when the job actually completes, so in case this
        handler instance dies or is stopped another one can pick up the job
        and resume it.

        See also the "Message acknowledgement" paragraph in::

          http://www.rabbitmq.com/tutorials/tutorial-two-python.html
        """
        channel.basic_ack(delivery_tag)

    @inlineCallbacks
    def run_accumulator(self, client, channel_id=2):
        """
        Declare the accumulator exchange and the queue, and consumes messages
        from it.

        When a message is received, other messages of the same type will be
        ignored for a given delay, and then send back to the job handler queue.

        @rtype: Deferred.
        @raises txamqp.client.Closed: (Deferred) if the connection to the
            broker is lost.
        @raises txamqp.queue.Closed: (Deferred) if the queue is closed.
        """
        self._accumulation = {}

        channel = yield client.channel(channel_id)
        yield channel.channel_open()

        yield channel.exchange_declare(
            exchange="%s.job-accumulator-exchange" % self._prefix, type="direct"
        )

        yield channel.queue_declare(
            queue="%s.job-accumulator-queue" % self._prefix, durable=True
        )
        yield channel.queue_bind(
            queue="%s.job-accumulator-queue" % self._prefix,
            exchange="%s.job-accumulator-exchange" % self._prefix,
            routing_key="%s.job-accumulator-queue" % self._prefix,
        )
        try:
            yield channel.basic_consume(
                consumer_tag="%s.job-accumulator-tag" % self._prefix,
                queue="%s.job-accumulator-queue" % self._prefix,
                exclusive=True,
            )
        except Closed:
            self._clock.callLater(
                self._accumulator_reconnection_delay, self.run_accumulator, client
            )
            return

        queue = yield client.queue("%s.job-accumulator-tag" % self._prefix)

        while True:
            msg = yield queue.get()
            payload = json.loads(msg.content.body)
            action = payload["action"]
            namespace = payload.get("namespace", "")
            yield channel.basic_ack(msg.delivery_tag)

            key = "%s.%s" % (action, namespace)
            if key in self._accumulation:
                continue

            delay = payload.pop("accumulate_delay", self._default_accumulator_delay)
            self._accumulation[key] = self._clock.callLater(
                delay, self._fire_accumulated, key, payload, channel_id
            )

    @inlineCallbacks
    def _fire_accumulated(self, key, payload, channel_id):
        self._accumulation.pop(key)
        channel = yield self._client.channel(channel_id)

        body = json.dumps(payload)
        content = Content(body)

        yield channel.basic_publish(
            content=content,
            exchange="%s.job-handler-exchange" % self._prefix,
            routing_key="%s.job-handler-queue" % self._prefix,
        )
        return

    @inlineCallbacks
    def fire_pending_accumulated(self):
        """
        Force the publishing of accumulated messages, at shutdown for example.
        """
        for call in list(self._accumulation.values()):
            yield call.func(*call.args, **call.kw)
