# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import uuid
import logging
import threading
import transaction
from transaction.interfaces import IDataManager

from zope.interface import implementer
from zope.component import getUtility

from twisted.internet.threads import blockingCallFromThread
from twisted.internet import reactor

from canonical.amqp.message.interfaces import IDispatcher, IPublisher


__all__ = ["Dispatcher"]


class DispatcherAttributes(threading.local):
    """
    Thread-local object to store thread-sensitive attributes of
    L{Dispatcher}.
    """

    def __init__(self):
        super(DispatcherAttributes, self).__init__()
        self.queues = []
        self.jobs = []
        self.actions = []
        self.errors = []
        self.notifications = {}
        self.added = False


@implementer(IDispatcher, IDataManager)
class Dispatcher(object):
    """
    Send job requests that will be run by the external job handler.

    @ivar _local: L{DispatcherAttributes} to store thread local variables.
    """

    def __init__(self, prefix):
        self._prefix = prefix
        self._local = DispatcherAttributes()
        self.__publisher = None

    @property
    def _publisher(self):
        if self.__publisher is None:
            self.__publisher = getUtility(IPublisher)
        return self.__publisher

    def _registerDataManager(self):
        """
        Register the dispatcher as L{IDataManager}, so that it's called when
        the transaction ends.

        @note: This is called in a thread which is not the main thread.
        """
        if not self._local.added:
            transaction.manager.get().join(self)
            self._local.added = True

    def create_job_queue(self, namespaces):
        """
        Create a new job queue, used to handle page messages.

        @note: This is called in a thread which is not the main thread.
        """
        self._registerDataManager()
        page_key = str(uuid.uuid4())
        self._local.queues.append((page_key, namespaces))
        return page_key

    def create_job(self, page_key, namespace, action, **kwargs):
        """Create a job request.

        @param page_key: A UUID used to identify the requester of the job. Once
            the job is finished it will be used to notify the requester back
            with the job's result, using a queue named against such UUID (see
            also L{canonical.amqp.message.publisher.Publisher.get_queue_name}).

        @param namespace: If not C{None}, a namespace that will be used to
            broadcast the notification part of the job's result. See also the
            docstring for L{canonical.amqp.job.runner.JobRunner.run}.

        @param action: The name of the job to run.

        @note: This is called in a thread which is not the main thread.
        """
        self._registerDataManager()
        kwargs["uuid"] = page_key
        kwargs["action"] = action
        kwargs["namespace"] = namespace
        self._local.jobs.append(kwargs)

    def send_action(self, result, uuid, action):
        """
        Queue an action to be sent on commit.

        @note: This is called in a thread which is not the main thread.
        """
        self._registerDataManager()
        self._local.actions.append((result, uuid, action))

    def send_error(self, value, uuid, action):
        """
        Queue an error to be sent on commit.

        @note: This is called in a thread which is not the main thread.
        """
        self._registerDataManager()
        self._local.errors.append((value, uuid, action))

    def send_notification(self, result, uuid, action, namespace):
        """
        Queue a notification to be sent on commit.

        @note: This is called in a thread which is not the main thread.
        """
        self._registerDataManager()
        self._local.notifications[(uuid, action, namespace)] = result

    def tpc_finish(self, txn):
        """
        Send created jobs and queues after commit, cleaning up local
        attributes.

        @note: This is called in a thread which is not the main thread.
        """
        if not reactor.running:
            logging.warning("Reactor is not running, unable to send messages.")
            return
        try:
            self._local.added = False

            for page_key, namespaces in self._local.queues:
                # Declare the queues in the main thread
                blockingCallFromThread(
                    reactor, self._publisher.declare_queue, page_key, namespaces
                )
            self._local.queues = []

            for kwargs in self._local.jobs:
                # Send the message in the main thread
                blockingCallFromThread(reactor, self._publisher.publish_job, kwargs)

            self._local.jobs = []

            for action in self._local.actions:
                # Send actions from the main thread
                blockingCallFromThread(reactor, self._publisher.publish_action, *action)
            self._local.actions = []

            for error in self._local.errors:
                # Send errors from the main thread
                blockingCallFromThread(reactor, self._publisher.publish_error, *error)
            self._local.errors = []

            for args, result in self._local.notifications.items():
                # Send notifications from the main thread
                blockingCallFromThread(
                    reactor, self._publisher.publish_notification, result, *args
                )
            self._local.notifications = {}
        except Exception as e:
            logging.warning("Exception during message publication: %s" % (e,))
            try:
                # The connection is likely to be in an inconsistent state,
                # let's reconnect.
                blockingCallFromThread(reactor, self._publisher.disconnect)
            except Exception:
                # We really don't want the transaction to break at that point.
                pass

    def abort(self, txn):
        """Clean created jobs and queues when aborting."""
        self._local.added = False
        self._local.queues = []
        self._local.jobs = []
        self._local.actions = []
        self._local.errors = []
        self._local.notifications = {}

    def tpc_begin(self, txn):
        pass

    def commit(self, txn):
        pass

    def tpc_vote(self, txn):
        pass

    def tpc_abort(self, txn):
        pass

    def sortKey(self):
        return "job_dispatcher_%d" % id(self)


global_dispatcher = Dispatcher("landscape")
