# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import object
from zope.interface import implementer

from twisted.internet.defer import inlineCallbacks
from twisted.python import log

from txamqp.content import Content

from canonical.lib.serialize import json_dumps
from canonical.amqp.message.interfaces import IPublisher

__all__ = ["Publisher"]


@implementer(IPublisher)
class Publisher(object):
    """
    An AMQP publisher which pushes messages to either the AJAX frontend or the
    job handler.
    """

    def __init__(self, prefix):
        self._prefix = prefix
        self._channel = None
        self._client = None

    def disconnected(self):
        self._channel = None
        self._client = None

    def disconnect(self):
        self._client.transport.loseConnection()

    def get_queue_name(self, uuid):
        """Build a queue name from a UUID."""
        return "%s.notifications-queue.%s" % (self._prefix, uuid)

    @inlineCallbacks
    def connected(self, connection_data):
        """Declare the notifications exchange.

        @rtype: Deferred.
        @raises txamqp.client.Closed: (Deferred) if the connection to the
            broker is lost.
        """
        (client, channel) = connection_data
        log.msg("Notifications publisher connected to AMQP broker")

        self._client = client
        self._channel = channel

        yield channel.exchange_declare(
            exchange="%s.job-handler-exchange" % self._prefix, type="direct"
        )
        yield channel.exchange_declare(
            exchange="%s.notifications-exchange" % self._prefix, type="direct"
        )
        yield channel.exchange_declare(
            exchange="%s.job-accumulator-exchange" % self._prefix, type="direct"
        )

    def publish_job(self, arguments):
        """Publish a message to the job handler queue.

        @note: This has to be called in the main thread.
        """
        accumulate = arguments.pop("accumulate", False)
        body = self._dumps(arguments)
        if accumulate:
            exchange = "job-accumulator-exchange"
            key = "job-accumulator-queue"
        else:
            exchange = "job-handler-exchange"
            key = "job-handler-queue"
        return self._channel.basic_publish(
            content=self._make_content(body),
            exchange="%s.%s" % (self._prefix, exchange),
            routing_key="%s.%s" % (self._prefix, key),
        )

    def publish_action(self, result, uuid, action):
        """Sends an action result through the notifications queue."""
        if uuid is None:
            return
        dump = self._dumps({"action": action, "result": result})
        return self._channel.basic_publish(
            exchange="%s.notifications-exchange" % self._prefix,
            content=self._make_content(dump),
            routing_key=self.get_queue_name(uuid),
        )

    def publish_error(self, value, uuid, action):
        """Sends an action error result through the notifications queue."""
        if uuid is None:
            return
        result = self._dumps({"error": str(value), "action": action})
        return self._channel.basic_publish(
            exchange="%s.notifications-exchange" % self._prefix,
            content=self._make_content(result),
            routing_key=self.get_queue_name(uuid),
        )

    def publish_notification(self, result, uuid, action, namespace):
        """Sends a notification to all listeners in the namespace."""
        notification = self._dumps(
            {"result": result, "notification": action, "original-uuid": uuid}
        )
        return self._channel.basic_publish(
            exchange="%s.notifications-exchange" % self._prefix,
            content=self._make_content(notification),
            routing_key=namespace,
        )

    @inlineCallbacks
    def declare_queue(self, page_key, namespaces, expires=300 * 1000):
        """
        Create the queue for the given 'page_key;, bind it to the proper
        'namespaces'. The queue will be deleted after 'expires' milliseconds,
        unless None is passed.

        @note: This has to be called in the main thread.
        """
        queue_name = self.get_queue_name(page_key)
        channel = self._channel
        arguments = {}
        if expires:
            arguments.update({"x-expires": expires})
        yield channel.queue_declare(queue=queue_name, durable=True, arguments=arguments)
        yield channel.queue_bind(
            queue=queue_name,
            exchange="%s.notifications-exchange" % self._prefix,
            routing_key=queue_name,
        )
        for namespace in namespaces:
            yield channel.queue_bind(
                queue=queue_name,
                exchange="%s.notifications-exchange" % self._prefix,
                routing_key=namespace,
            )

    def update_queue_subscription(self, page_key, namespace):
        """
        Update the namespaces of an existing queue with the given C{page_key}.
        """
        queue_name = self.get_queue_name(page_key)
        return self._channel.queue_bind(
            queue=queue_name,
            exchange="%s.notifications-exchange" % self._prefix,
            routing_key=namespace,
        )

    def _dumps(self, obj):
        """Return a json-encoded string of the given object."""
        # Use canonical.lib.serialize.json_dumps in order to automatically
        # un-proxy lists and dicts.
        return json_dumps(obj)

    def _make_content(self, body):
        """Return a L{txamqp.content.Content} object with the given body

        The delivery mode will be set to 2, so messages will be durable.
        """
        return Content(body, properties={"delivery mode": 2})


global_publisher = Publisher("landscape")
