# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from twisted.internet.defer import Deferred, inlineCallbacks, DeferredQueue
from twisted.internet import reactor

from txamqp.client import Closed

from canonical.testing.runner import get_worker
from canonical.testing.twist import TwistedTestCase
from canonical.txlongpoll.client import AMQFactory


class QueueWrapper(object):
    """
    Wrap a queue to have notifications when get is called on this particular
    queue.
    """

    def __init__(self, queue, method="get"):
        self._real_queue = getattr(queue, method)
        self.event_queue = DeferredQueue()
        setattr(queue, method, self.wrap)

    def wrap(self, *args, **kwargs):
        self.event_queue.put(None)
        return self._real_queue(*args, **kwargs)


class AMQTest(TwistedTestCase):

    vhost = "landscape-test"
    user = "landscape"
    password = "landscape"

    def setUp(self):
        """
        At each run, we delete the test vhost and recreate it, to be sure to be
        in a clean environment.
        """
        TwistedTestCase.setUp(self)
        self.queues = set()
        self.exchanges = set()
        self.connected_deferred = Deferred()

        # When running trial -j we need to use different vhosts
        if get_worker() is not None:
            self.vhost += "-%d" % get_worker()

        self.factory = AMQFactory(
            self.user,
            self.password,
            self.vhost,
            self.amq_connected,
            self.amq_disconnected,
            self.amq_failed,
        )
        self.factory.initialDelay = 2.0
        self.connector = reactor.connectTCP("localhost", 5672, self.factory)
        return self.connected_deferred

    @inlineCallbacks
    def tearDown(self):
        self.factory.stopTrying()
        self.connector.disconnect()

        self.connected_deferred = Deferred()
        factory = AMQFactory(
            self.user,
            self.password,
            self.vhost,
            self.amq_connected,
            self.amq_disconnected,
            self.amq_failed,
        )
        connector = reactor.connectTCP("localhost", 5672, factory)
        yield self.connected_deferred
        channel_id = 1
        for queue in self.queues:
            try:
                yield self.channel.queue_delete(queue=queue)
            except Closed:
                channel_id += 1
                self.channel = yield self.client.channel(channel_id)
                yield self.channel.channel_open()
        for exchange in self.exchanges:
            try:
                yield self.channel.exchange_delete(exchange=exchange)
            except Closed:
                channel_id += 1
                self.channel = yield self.client.channel(channel_id)
                yield self.channel.channel_open()
        factory.stopTrying()
        connector.disconnect()
        TwistedTestCase.tearDown(self)

    def amq_connected(self, connection_data):
        """
        Save the channel and client, and fire C{self.connected_deferred}.

        This is the connected_callback that's pased to the L{AMQFactory}.
        """
        (client, channel) = connection_data
        self.real_queue_declare = channel.queue_declare
        channel.queue_declare = self.queue_declare
        self.real_exchange_declare = channel.exchange_declare
        channel.exchange_declare = self.exchange_declare
        self.channel = channel
        self.client = client
        self.connected_deferred.callback(None)

    def amq_disconnected(self):
        """
        This is the disconnected_callback that's passed to the L{AMQFactory}.
        """

    def amq_failed(self, failure_data):
        """
        This is the failed_callback that's passed to the L{AMQFactory}.
        """
        (connector, reason) = failure_data
        self.connected_deferred.errback(reason)

    def queue_declare(self, queue, **kwargs):
        """
        Keep track of the queue declaration, and forward to the real
        queue_declare function.
        """
        self.queues.add(queue)
        return self.real_queue_declare(queue=queue, **kwargs)

    def exchange_declare(self, exchange, **kwargs):
        """
        Keep track of the exchange declaration, and forward to the real
        exchange_declare function.
        """
        self.exchanges.add(exchange)
        return self.real_exchange_declare(exchange=exchange, **kwargs)
