from __future__ import division

# Copyright 2005-2014 Canonical Limited. All rights reserved.

from builtins import object
from past.utils import old_div

DEFAULT_BATCH_SIZE = 25


class BatchingInfo(object):
    def __init__(self, total_items, position=1, batch_size=DEFAULT_BATCH_SIZE):
        """Hold information about a single batch in a partioned sequence.

        Let's say we have a sequence of items [a1, a2, a3, ..., aN] that has
        been partioned in batches of equal size (with the exception of
        the last one which might be smaller). Then a L{BatchingInfo} object
        provides both data about one of those partitions (like the index of the
        first and last items), and data about the partition as a whole (like
        the total number of elements in the sequence and the number of batches
        it has been partioned into).

        For example, if the sequence to partition has 5 items and the
        chosen batch size is 2, then there will be 3 batches: the first
        batch from the first item to the second, the second batch from
        the third item to the fourth, and the last batch with only the fifth
        element.

        @param total_items: The total number of elements in the sequence being
            partitioned. It must be greater than zero.
        @param position: An positive integer indicating which batch of the
            partition we're providing information for. The first batch is
            indicated by "1", the second by "2", and so on.
        @param batch_size: An integer to indicate how many items a batch
            should contain. All batches in the partition will have the same
            size, possibly except that last which might be smaller (if not
            enough elements remain).
        """
        assert total_items > 0, "Total items is not greater than zero"
        assert batch_size > 0, "Batch size is not greater than zero"
        self.total_items = total_items
        self.batch_size = batch_size
        self.position = self._coerce_position(position)

    @property
    def total_batches(self):
        """Return the total number of batches."""
        result = old_div(self.total_items, self.batch_size)
        if self.total_items % self.batch_size != 0:
            # Account for the last batch, which will be smaller
            result += 1
        return result

    @property
    def first_item(self):
        """Return the index of the first item of the batch."""
        return self.batch_size * (self.position - 1) + 1

    @property
    def last_item(self):
        """Return the index of the last item of the batch."""
        return min(self.first_item + self.batch_size - 1, self.total_items)

    def at_first_batch(self):
        """Whether the position is set to first batch."""
        return self.position == 1

    def at_last_batch(self):
        """Whether the position is set to the last batch."""
        return self.position == self.total_batches

    def _coerce_position(self, position):
        """Coerce the provided position value to a valid one.

        If C{position} is a string it will be converted to a digit
        (or default to 1).
        If position is an out-of-range digit it will be changed to the closest
        bound.
        """
        try:
            position = int(position)
        except ValueError:
            position = 1
        if position < 1:
            position = 1
        if position > self.total_batches:
            position = self.total_batches
        return position
