# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""L{CollectionBase} is a superclass for Storm object collections."""

from storm.info import get_cls_info

from canonical.lib.undef import Undef


__all__ = ["CollectionBase"]


class CollectionBase(object):
    """A superclass for collections of instances of Storm model.

    @cvar model_class: The AccountEntity-based class that this
        collection is for.
    """

    model_class = Undef()  # Must be defined by subclasses

    def __init__(self, store, **kwargs):
        self._store = store
        self._where = kwargs
        self._using = None
        self._select = None

    @property
    def filters(self):
        """Return a dict with the filters set on the collection."""
        return self._where

    def find(self):
        """Find instances matching filter criteria.

        @return: A L{ResultSet} that yields filtered instances.
        """
        return self._get_result(self._store, **self._where)

    def using(self, *tables):
        """Specify tables to use explicity.

        The L{find} method generally uses the collection class only.  In cases
        where an explicit SQL join is needed though, the using method can be
        used to construct an explicit join.

        @param tables: A list of expressions to use as tables.
        """
        new = self.__class__(self._store, **self._where)
        new._using = tables
        return new

    def select(self, expr):
        """Specify the select expression explicity.

        The L{find} method normally returns L{ResultSet}s configured to
        select a particular model class. This method makes it possible
        to use an alternate select expression.

        @param expr: The alternate select expr to use.
        """
        new = self.__class__(self._store, **self._where)
        new._select = expr
        return new

    def remove(self):
        """Remove objects in the Collection."""
        primary_key = get_cls_info(self.model_class).primary_key[0]
        result = self.find()
        select = result.find().get_select_expr(primary_key)
        result = self._store.find(self.model_class, primary_key.is_in(select))
        result.remove()

    def reset(self):
        """Return a new Collecion of the same kind without filters."""
        return self.__class__(self._store)

    def _filter_by(self, **kwargs):
        """Return a subset of this collection filtered by C{expressions}.

        This method is protected and intended for use by a subclass.
        """
        where = dict(self._where)
        where.update(kwargs)
        new = self.__class__(self._store, **where)
        new._using = self._using
        return new

    def _get_result(self, store, where=None, **kwargs):
        """Get a Storm C{ResultSet} using C{kwargs} to search C{store}.

        This method must be implemented by a subclass.

        It must use the store and provided criteria to create and return a
        C{ResultSet} containing requested items.  Named keyword arguments for
        each filtering criterion should be defined by the subclass that
        overrides this method.

        @param store: The C{Store} to create the C{ResultSet} from.
        @param where: Optionally, a list of where clauses to include in the
            generated query.
        """
        raise NotImplementedError("Must be implemented by subclass.")
