# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.lib.timestamp import utcnow, TIMESTAMP_FORMAT


__all__ = ["set_download", "set_csv_download", "set_yaml_download"]


def set_download(response, filename):
    """Set the Content-Disposition header for a download.

    @param response: The response to which the header will be added.
    @param filename: The file name to add.
    """
    response.setHeader(
        "Content-Disposition",
        'attachment; filename="%s"' % filename.replace('"', '\\"'),
    )


def set_content_type(response, content_type):
    """Set the Content-Type header for a response.

    @param response: The response to which the header will be added.
    @param content_type: The content type to be added.
    """
    response.setHeader("Content-Type", content_type)


def set_csv_download(response, basename):
    """
    Set the Content-Disposition and Content-Type headers for a CSV download.

    @param response: The response to which the header will be added.
    @param basename: The start of the file name. The current time as
        well as an extension will be appended to this.
    """
    set_download(response, _make_timestamped_filename(basename, ".csv"))
    set_content_type(response, "text/csv")


def set_yaml_download(response, filename):
    """
    Set the Content-Disposition and Content-Type headers for a YAML download.

    @param response: The response to which the header will be added.
    @param filename: The file name to add.
    """
    set_download(response, filename)
    set_content_type(response, "application/x-yaml")


def set_gzipped_tar_download(response, basename, timestamped=True):
    """
    Set the Content-Disposition, Content-Type and Content-Encoding headers for
    a tar download.

    @param basename: The start of the file name. The current time as well as
        an extension will be appended to this.
    @param timestampd: Whether the current time should be appended.
    """
    name = basename
    if timestamped:
        name = _make_timestamped_filename(name, ".tar.gz")
    set_download(response, name)
    set_content_type(response, "application/x-gzip")


def _make_timestamp():
    """Make an ISO8601 timestamp with second resolution."""
    return utcnow().replace(microsecond=0).strftime(TIMESTAMP_FORMAT)


def _make_timestamped_filename(basename, extension):
    """Given a basename and an extension, return a file name with timestamp."""
    return "%s-%s%s" % (basename, _make_timestamp(), extension)
