# Copyright 2005-2022 Canonical Limited. All rights reserved.

import sys
import inspect

from html import escape
from uuid import uuid4

import zope.i18n
import zope.i18nmessageid

from zope.browserpage import namedtemplate
from zope.component import getMultiAdapter
from zope.formlib import form
from zope.interface import implementer, directlyProvides
from zope.formlib.interfaces import (
    IWidgetInputError,
    WidgetInputError,
    IWidgetInputErrorView,
)
from zope.session.interfaces import ISession

from canonical.content.provider import ContentProvider, TemplateContentProvider
from canonical.content.interfaces import IPrimaryAction


Actions = form.Actions
Action = form.Action


@implementer(IPrimaryAction)
class PrimaryAction(form.Action):
    """A primary L{Action}.

    Primary actions are presented differently to the end-user, such
    that the obvious choice in a form is clearly visible.
    """


class action(form.action):
    """
    A decorator that registers an L{Action} item with the form
    provider it's used on. This implementation supports marking the
    registered action as a C{primary} action.
    """

    def __init__(self, label, actions=None, primary=False, disabled=False, **options):
        # Note: sys._getframe() mockery copied from zope.formlib.form,
        # because otherwise it would get our subclass' locals instead.
        caller_locals = sys._getframe(1).f_locals
        if actions is None:
            actions = caller_locals.get("actions")
        if actions is None:
            actions = caller_locals["actions"] = Actions()
        form.action.__init__(self, label, actions, **options)
        self.primary = primary
        self.disabled = disabled

    def __call__(self, success):
        action = form.action.__call__(self, success)
        if self.primary:
            directlyProvides(action, IPrimaryAction)
        action.disabled = self.disabled
        return action


@namedtemplate.implementation(IPrimaryAction)
def render_submit_button(self):
    """Specialized template for a L{PrimaryAction}.

    Adds the C{primary} CSS class to the rendered input control, so that
    it can be styled differently.
    """
    if not self.available():
        return ""
    label = self.label
    if isinstance(label, zope.i18nmessageid.Message):
        label = zope.i18n.translate(self.label, context=self.form.request)
    disabled = " disabled" if getattr(self, "disabled", False) else ""
    return (
        '<input type="submit" id="%s" name="%s" value="%s"'
        ' class="button primary"%s />'
        % (self.__name__, self.__name__, escape(label, quote=True), disabled)
    )


class FormToken(object):
    """Manages a token to be used for CSRF form protection.

    The token is bound to the currently active session for a user, and
    expires when the user logs out.
    """

    session_key = "canonical.content.form.FormToken"
    token_key = "form-security-token"

    def generate_token(self):
        """Generate a CSRF form protection token."""
        return str(uuid4())

    def get(self, request):
        """Get existing token from session, otherwise generate a new one."""
        session = ISession(request)[self.session_key]
        try:
            return session[self.token_key]
        except KeyError:
            session[self.token_key] = token = self.generate_token()
            return token

    def validate(self, request, token=None):
        """Validate a provided token against the current session."""
        if token is None:
            token = request.form.get(self.token_key)
            if token is None:
                return False
        # Don't create a session if there is no token
        session = ISession(request)[self.session_key]
        existing = session.get(self.token_key)
        if existing is None:
            return False
        return token == existing

    def expire(self, request):
        """Expire form token from the current session if it exists."""
        session = ISession(request)[self.session_key]
        try:
            del session[self.token_key]
        except KeyError:
            pass


form_token = FormToken()


def render_form_token(request):
    """Render CSRF protection token as a hidden input field."""
    key = form_token.token_key
    value = form_token.get(request)
    return '<input type="hidden" name="%s" value="%s"/>' % (key, value)


def set_form_token(request):
    """Add an exising form token to the current request."""
    request.form[form_token.token_key] = form_token.get(request)


class Protect(Action):
    """Render a security token as part of the form.

    In order to include the security token as part of the containing
    form, we inject this action in ``available_actions``.

    We are really abusing the framework here, but there doesn't seem
    to be a better way to inject a snippet of HTML into the form
    other than changing the page template itself.
    """

    def __init__(self, **options):
        if not "label" in options:
            options["label"] = "Form Protection"
        super(Protect, self).__init__(**options)

    def submitted(self):
        # This action is never considered submitted. Otherwise,
        # form.handleSubmit() would bail out as soon as it found
        # us. At the same time, that means ``validate`` will never be
        # called by form.handleSubmit(), so we call it from
        # ``FormContentProvider._validate`` instead.
        return False

    def render(self):
        """Render the CSRF protection token as a hidden field."""
        return render_form_token(self.form.request)


class Fields(form.Fields):
    def __init__(self, *args, **kwargs):
        super(Fields, self).__init__(*args, **kwargs)
        self._hidden = set()
        self._disabled = set()

        # Percolate up per-field custom widgets.
        for field in self:
            if field.custom_widget is not None:
                continue
            field.custom_widget = getattr(field.field, "custom_widget", None)

    def disable(self, *names):
        """Disable the form widgets associated with the given C{names}.

        If no field name is passed, then all fields get disabled.
        """
        obj = self.__class__(*list(self))
        obj._disabled.update(self._disabled)
        if not names:
            names = [field.__name__ for field in list(self)]
        obj._disabled.update(names)
        return obj

    def hide(self, *names):
        """Hide form widgets associated with the given C{names}."""
        obj = self.__class__(*list(self))
        obj._hidden.update(self._hidden)
        obj._hidden.update(names)
        return obj

    def select(self, *names):
        """Filter widgets associated with the given C{names}."""
        obj = super(Fields, self).select(*names)
        obj._hidden.update(self._hidden)
        return obj

    def omit(self, *names):
        """Omit widgets associated with the given C{names}."""
        obj = super(Fields, self).omit(*names)
        obj._hidden.update(self._hidden)
        return obj


class FormContentProvider(ContentProvider):

    prefix = "form"

    fields = Fields()
    actions = Actions()

    render_request = True
    render_global_errors = True
    allowed_http_methods = ("POST",)
    dropdown_css_class = "dropdown-spacer"

    def __init__(self, context, request, view):
        super(FormContentProvider, self).__init__(context, request, view)
        self.widget_errors = {}
        self.widget_hidden = {}
        self.global_errors = []

    def available_actions(self):
        """
        Get available actions for this form, injecting a fake action
        to generate a CSRF protection token.
        """
        actions = form.availableActions(self, self.actions)

        # If there's a single action and it's not a L{Protect} action,
        # mark it as an L{IPrimaryAction}.
        if len(actions) == 1 and not isinstance(actions[0], Protect):
            directlyProvides(actions[0], IPrimaryAction)

        # Inject a L{Protect} action into the list of available
        # actions if one doesn't exist yet, so that it's C{render}
        # method is called, thus rendering the hidden input field with
        # the security token.
        protect = [action for action in actions if isinstance(action, Protect)]
        if not len(protect):
            actions.append(Protect().__get__(self))
        return actions

    def get_initial_data(self):
        """Get initial data to populate the form fields with.

        Sub-classes can override this method to provide initial data for
        the fields.

        @return: A C{dict} mapping field names to the value they should be
            set to.
        """
        return {}

    def setup_widgets(self, render_request=True):
        """Setup widget instances with initial data."""
        self.adapters = {}
        self.widgets = form.setUpWidgets(
            self.fields,
            self.prefix,
            self.context,
            self.request,
            data=self.get_initial_data(),
            adapters=self.adapters,
            ignore_request=not render_request,
        )

    def apply_changes(self, context, data, fields=None, adapters=None):
        """Apply form changes to context object."""
        if fields is None:
            fields = self.fields
        return form.applyChanges(context, fields, data, adapters=adapters)

    def render(self, *args, **kwargs):
        """
        Render this content provider if it hasn't been rendered yet and no
        redirect is performed.
        """
        if self.request.response.getStatus() in [302, 303]:
            return ""
        if self._result is None:
            return super(FormContentProvider, self).render(*args, **kwargs)
        return self._result

    def update(self):
        """Called as the first step when rendering the form.

        Prepares widgets, performs CSRF validation and finally handles
        the submitted action.
        """
        self.setup_widgets(self.render_request)

        self.widget_hidden.clear()
        for field in self.fields:
            widget = self.widgets[field.__name__]
            self.widget_hidden[widget.name] = field.__name__ in self.fields._hidden
            if field.__name__ in self.fields._disabled:
                widget.extra = 'disabled="disabled"'

        # Check that we are really allowed to be called at
        # all. Typically, this involves checking for a form
        # authentication token and for the right HTTP method.
        if not self._validate_security():
            self._result = None
            return

        data = {}
        errors, action = form.handleSubmit(self.actions, data, self._validate)

        if not action:
            self._result = None
        elif self.has_errors():
            self._result = action.failure(data, ())
        else:
            self._result = action.success(data)

    def _validate(self, action, data):
        """Validate form data and add per-widget and global errors."""
        has_widget_errors = False
        for error in form.getWidgetsData(self.widgets, self.prefix, data):
            if IWidgetInputError.providedBy(error):
                has_widget_errors = True
            else:
                self.add_error(error)

        if has_widget_errors:
            for widget in self.widgets:
                error_method = getattr(widget, "error", None)
                if error_method:
                    error = error_method()
                    if error:
                        self.add_widget_error(error, widget.context.getName())
                else:
                    raise RuntimeError("Non-BrowserWidget used.")

        for error in check_invariants(self.fields, data):
            self.add_error(error)

        self.validate(action, data)

    def _validate_security(self):
        """Validate CSRF protection token and HTTP method."""
        # Only run this check if some action has been submitted. We
        # don't want to perform the check when the initial form is
        # rendered, but only when it has been submitted.
        for action in self.actions:
            if action.submitted():
                break
        else:
            return True
        if not self.request.method in self.allowed_http_methods:
            self.add_error(
                "Invalid HTTP method (got: %s, expected: %s)"
                % (self.request.method, ", ".join(self.allowed_http_methods))
            )
            return False
        if not form_token.validate(self.request):
            self.add_error("Missing authentication token")
            return False
        return True

    def validate(self, action, data):
        """Hook for validating form data. To be overriden by sub-classes.

        Sub-classes can use this either to perform further checks on the fields
        data that can't be expressed with Interface schemas or constraints, or
        do extract further data from the posted form (without implementing
        custom schemas or widgets).
        """

    def add_widget_error(self, error, field_name):
        """Adds a raw error message for a specific widget."""
        widget = self.widgets[field_name]
        self.widget_errors[widget.name] = error

    def add_error(self, error, field_name=None):
        """Add an error message, either globally or field-specific.

        If a C{field_name} is provided, create a L{WidgetInputError}
        and render the error snippet by looking up an
        L{IWidgetInputErrorView}. Otherwise, add raw error message as
        a global error.
        """
        if field_name is not None:
            widget = self.widgets[field_name]
            input_error = WidgetInputError(widget.name, widget.label, error)
            view = getMultiAdapter((input_error, self.request), IWidgetInputErrorView)
            self.add_widget_error(view.snippet(), field_name)
        else:
            self.global_errors.append(error)

    def has_errors(self):
        """Check if this form has any errors, global or widget-specific."""
        return self.widget_errors or self.global_errors


class TemplateFormContentProvider(FormContentProvider, TemplateContentProvider):
    template_filename = "form.pt"

    summary = ""
    description = ""

    separator = ""


_new_formlib = len(inspect.getargspec(form.checkInvariants).args) == 3


def check_invariants(form_fields, form_data):
    """
    Wrap formlib.checkInvariants, which has backward incompatible changes
    between 4.0.x and 4.1.x.
    """
    args = [form_fields, form_data]
    if _new_formlib:
        args.append(None)
    return form.checkInvariants(*args)
