# Copyright 2005-2022 Canonical Limited. All rights reserved.

import sys
import os

from zope.contentprovider.interfaces import IContentProvider
from zope.component import queryMultiAdapter
from zope.interface import implementer

from z3c.pt.pagetemplate import ViewPageTemplateFile


__all__ = [
    "ContentProvider",
    "TemplateContentProvider",
    "ForwardContentProviderAdapter",
]


@implementer(IContentProvider)
class ContentProvider(object):
    """Base implementation of a content provider."""

    def __init__(self, context, request, view):
        self.context = context
        self.request = request
        self.view = view

    @property
    def __parent__(self):
        return self.view

    def update(self):
        """Update state of the content provider."""

    def render(self, *args, **kwargs):
        """Render content provider."""
        return ""


class ProviderPageTemplateFile(ViewPageTemplateFile):
    @property
    def boolean_attributes(self):
        # The 'hidden' boolean attribute was introduced by HTML5
        result = super(ProviderPageTemplateFile, self).boolean_attributes
        return result.union(["hidden"])

    def pt_getContext(self, instance, request, **_kw):
        namespace = ViewPageTemplateFile.pt_getContext(self, instance, request, **_kw)
        namespace["view"] = instance.view
        namespace["provider"] = instance
        return namespace

    def _pt_get_context(self, instance, request, kwargs):
        namespace = ViewPageTemplateFile._pt_get_context(
            self, instance, request, kwargs
        )
        namespace["view"] = instance.view
        namespace["provider"] = instance
        return namespace


class TemplateContentProviderMeta(type):
    def __new__(cls, name, bases, dict):
        cls = type.__new__(cls, name, bases, dict)
        template_filename = dict.get("template_filename")
        if template_filename is not None:
            if not os.path.isabs(template_filename):
                filename_prefix = os.path.dirname(
                    sys._getframe(1).f_globals["__file__"]
                )
            else:
                filename_prefix = None
            cls.template = ProviderPageTemplateFile(
                cls.template_filename, filename_prefix
            )
        return cls


class TemplateContentProvider(ContentProvider, metaclass=TemplateContentProviderMeta):
    """Content provider which renders a template."""

    template_filename = None
    template = None

    def render(self, *args, **kwargs):
        return self.template(*args, **kwargs)


class ForwardContentProviderAdapter(object):
    """Adapter to forward content provider queries to another provider."""

    def __init__(self, name):
        self._name = name

    def __call__(self, context, request, view):
        return queryMultiAdapter((context, request, view), IContentProvider, self._name)
