# Copyright 2005-2022 Canonical Limited. All rights reserved.

from functools import wraps

from zope.interface import implementer
from zope.component.interfaces import IObjectEvent
from zope.component import (
    adapter,
    provideHandler,
    queryUtility,
    provideUtility,
    handle,
    getUtilitiesFor,
    getGlobalSiteManager,
)

from canonical.event.interfaces import IEventPhaseFactory, IEventPhase


def subscriber(*args, **kwargs):
    """Decorator used to subscribe an event handler to event types.

    @args: A list of event types the handler wants to subscribe to.
    @kwargs: Extra parameters, for now only the 'phase' parameter is supported,
        which will make the handler be executed at the given phase in the
        event notification cycle.
    """
    phase_number = kwargs.get("phase", 0)
    phase_name = str(phase_number)
    phase_class = queryUtility(IEventPhaseFactory, name=phase_name)
    if phase_class is None:
        # This the first time the decorator gets called with this phase number,
        # so create a new marker phase class for it.

        @implementer(IEventPhase)
        class EventPhase:
            number = phase_number

        provideUtility(EventPhase, IEventPhaseFactory, name=phase_name)
        phase_class = EventPhase

    def decorator(function):
        # Register a handler that takes the phase maker into account, so we
        # can callback all handlers of a given event registered against a given
        # phase.
        @wraps(function)
        def wrapper(phase, *args):
            if not IEventPhase.providedBy(phase):
                # The handler being called directly
                args = (phase,) + args
            return function(*args)

        handler = adapter(phase_class, *args)(wrapper)
        provideHandler(handler)
        return handler

    return decorator


def dispatch(event):
    """
    Notify subscribers for the given event.

    To hook this into zope.event, run:

      import zope.event
      zope.event.subscribers.append(dispatch)

    and use zope.event.notify as usual.
    """
    args = (event,)
    object_event_args = None
    if IObjectEvent.providedBy(event):
        # The Landscape code base uses objectEvent a bit inconsistently,
        # sometimes registering subscribers against the event only, and
        # sometimes against both the object of the event and the event itself,
        # so we support both behaviors.
        object_event_args = (event.object, event)

    # This loops through *all* registered phases, even if there are no handlers
    # subscribed for a particular event/phase combination. That makes code
    # easier.
    for phase_name, phase_class in sorted(getUtilitiesFor(IEventPhaseFactory)):
        phase = phase_class()
        handle(phase, *args)
        if object_event_args is not None:
            handle(phase, *object_event_args)


def unsubscribe(*handlers):
    """Unsubscribe the given event handlers."""
    for handler in handlers:
        getGlobalSiteManager().unregisterHandler(handler)
