# Copyright 2005-2022 Canonical Limited. All rights reserved.

import errno
import os
import re
import subprocess


from landscape.lib.twisted_util import spawn_process

from canonical.lib.timestamp import date_from_epoch

RETRY_ATTEMPTS = 1


class GnuPGError(Exception):
    """Raised when gpg fails."""

    def __init__(self, out, err, code=None, signal=None):
        self.out = out
        self.err = err
        self.code = code
        self.signal = signal

        if code is not None:
            reason = "exit code %d" % code
        if signal is not None:
            reason = "signal %d" % signal

        super(GnuPGError, self).__init__(
            "gpg ended with %s (out='%s'," " err='%s')" % (reason, out, err)
        )


class GnuPG(object):
    """
    Handle execution of GnuPG processes.

    @param homedir: The base directory for the installation.
    @param binary_path: The path to the gnupg binary.
    """

    def __init__(self, homedir, binary_path=None, options=None):
        self._homedir = homedir
        self._binary_path = binary_path or "/usr/bin/gpg"
        self._options = options or {}

        if binary_path:
            self._version = [0, 0, 0]
        else:
            version_output = subprocess.check_output([self._binary_path, "--version"])
            self._version = [
                int(v)
                for v in re.match(
                    b"^gpg \(GnuPG\) (\d)\.(\d)\.(\d)", version_output
                ).groups()
            ]

    def _run(self, options, stdin=None):
        """
        Run a gpg command with the specified options.

        @param options: A list of command line options to pass to gpg.

        @return: A deferred firing with stdout and stderr.
        """
        args = ["--homedir", self._homedir, "--no-tty"]
        if self._version[0] >= 2:
            args += ["--batch", "--emit-version", "--pinentry-mode", "loopback"]
        args.extend(options)
        args = [str(arg) for arg in args]

        if not os.path.exists(self._homedir):
            try:
                os.makedirs(self._homedir)
            except OSError as error:
                if error.errno != errno.EEXIST:
                    raise
        os.chmod(self._homedir, 0o700)
        self._write_config()

        try:
            # Causes issues with gpg-agent on gpg2 if missing.
            os.mkdir(os.path.join(self._homedir, "private-keys-v1.d"))
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise

        deferred = spawn_process(self._binary_path, args=args, stdin=stdin)
        deferred.addErrback(self._process_aborted)
        deferred.addCallback(self._check_exit_code, args=args, stdin=stdin)
        return deferred

    def _write_config(self):
        """Write current configuration to gpg.conf."""
        with open(os.path.join(self._homedir, "gpg.conf"), "w") as fd:
            for key, value in list(self._options.items()):
                fd.write("{} {}\n".format(key, value or ""))

    def _check_exit_code(self, error_data, args=[], stdin=None, attempts=0):
        """
        Raise a RuntimeError in case of non-zero exit code.
        gpg can be a bit flaky sometimes, so retry the command just once
        more if the error was an agent connection one.
        """
        (out, err, code) = error_data

        if attempts < RETRY_ATTEMPTS and code == 2 and b"IPC connect" in err:
            retry = spawn_process(self._binary_path, args=args, stdin=stdin)
            retry.addErrback(self._process_aborted)
            retry.addCallback(
                self._check_exit_code, args=args, stdin=stdin, attempts=attempts + 1
            )
            return retry

        if code != 0:
            raise GnuPGError(out, err, code=code)
        return out, err

    def _process_aborted(self, failure):
        """Raise a RuntimeError in case of non-zero exit code."""
        out, err, signal = failure.value.args
        raise GnuPGError(out, err, signal=signal)

    def import_keys(self, keys):
        """Import the given keys in the keyring."""
        deferred = self._run(["--import"], "\n".join(keys))
        return deferred.addCallback(self._parse_import_keys)

    def _parse_import_keys(self, pipes):
        """Parse the output of the import command.

        @return: A list of (key_id, has_secret) tuples.
        """
        (out, err) = pipes
        keys = {}
        key_re = re.compile(
            r"^gpg: key ([0-9A-F]+): " r"(public|secret) key.*imported$"
        )
        for line in err.splitlines():
            line = line.decode()

            match = key_re.match(line)
            if match is not None:
                key_id, key_type = match.groups()
                # trim key_id to last 8 chars, to be compatible with gpg1
                key_id = key_id[-8:]
                has_secret = keys.setdefault(key_id, False)
                if not has_secret and key_type == "secret":
                    keys[key_id] = True

        return list(keys.items())

    def sign(self, text, cleartext=True, passphrase=""):
        """Sign the given text using the default key.

        @param cleartext: If True, make a cleartext signature.
        @param passphrase: The passphrase to unlock key.
        """
        command = "--clearsign" if cleartext else "--sign"
        deferred = self._run([command, "--passphrase", passphrase], stdin=text)
        return deferred

    def _format_fingerprint(self, fingerprint):
        """Return fingerprint in colon-separated format."""
        tokens = (fingerprint[start::4] for start in range(0, 4))
        tokens = list(zip(*tokens))
        tokens = ("".join(token) for token in tokens)
        return ":".join(tokens).lower()

    def list_keys(self, keys=None):
        """
        Return info about keys.

        @param keys: A list of key IDs to list, or all keys if None.
        """
        options = [
            "--fixed-list-mode",
            "--with-colons",
            "--list-keys",
            "--with-fingerprint",
            "--with-fingerprint",
        ]
        if keys is not None:
            options.extend(keys)
        deferred = self._run(options)
        return deferred.addCallback(self._parse_list_keys)

    def _parse_list_keys(self, output):
        """
        Parse the output of key listing.  Description of the output format can
        be found in the DETAILS file in the gnupg package.
        """
        (out, err) = output
        KEY_ALGORITHMS = {
            "1": "RSA",
            "16": "Elgamal (encrypt only)",
            "17": "DSA",
            "20": "Elgamal (sign and encrypt)",
        }

        def basic_key_info(key_line):
            return {
                "id": line[4],
                "length": int(line[2]),
                "algorithm": KEY_ALGORITHMS.get(line[3], "Other"),
                "creation_date": date_from_epoch(int(line[5])),
                "expiration_date": date_from_epoch(int(line[6])) if line[6] else None,
            }

        keys = {}
        current_key = None
        main_key = None
        for line in out.splitlines():
            line = line.decode()

            line = line.split(":")
            key_type = line[0]

            if key_type == "pub":
                short_id = line[4][-8:]
                main_key = current_key = keys.setdefault(short_id, {})
                current_key.update(basic_key_info(line))
                current_key.update({"user_ids": [], "sub_keys": []})
            if key_type == "sub" and main_key is not None:
                current_key = basic_key_info(line)
                main_key["sub_keys"].append(current_key)
            if key_type == "fpr" and current_key is not None:
                current_key["fingerprint"] = self._format_fingerprint(line[9])
            if key_type == "uid" and current_key is not None:
                current_key["user_ids"].append(line[9])
        return keys

    def fingerprint(self, key):
        """Return the fingerprint the key specified by id."""
        deferred = self.list_keys(keys=[key])
        return deferred.addCallback(lambda info: info[key]["fingerprint"])

    def export_keys(self, secret=False, keys=None):
        """Export the GPG keys in the keyring.

        @param secret: If C{True}, export private keys too.
        @param keys: Optionally, a list of key ids to export, if C{None} all
            keys will be exported.
        """
        options = ["--armor"]
        options.append("--export-secret-keys" if secret else "--export")
        if keys is not None:
            options.extend(keys)
        deferred = self._run(options)
        return deferred.addCallback(lambda out_err: out_err[0])
