# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Role Based Access Control
=========================

Methods for managing role based access control (RBAC) within Landscape.

Central to RBAC is the concept of a role. Roles can have permissions, access
groups and administrators associated with them.
"""

from canonical.txapi.server.method import method
from canonical.txapi.server.schema import Schema, Unicode, List

from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.routes import (
    ManagementRoute,
    AccountRoute,
    EntityLookupRoute,
    CollectionLookupRoute,
)
from canonical.landscape.api.methods import (
    AccountMethod,
    FUTURE_VERSION,
    LATEST_VERSION,
)
from canonical.landscape.api.error import SelfRevokeGlobalAdminError
from canonical.landscape.model.main.errors import (
    UnknownPersonsError,
    UnknownAccessContextsError,
)
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.landscape.model.access.permissions import (
    MACRO_PERMISSIONS,
    get_macro_permissions,
)
from canonical.landscape.model.access.errors import (
    DuplicateRoleError,
    UnknownRoleError,
    InvalidPermissionsError,
    InvalidRoleNameError,
    ReadOnlyRoleError,
)
from canonical.landscape.service.accesscontext.lookup import (
    AccessContextLookup,
    AccessContextCollectionLookup,
)
from canonical.landscape.service.accesscontext.remove import RemoveAccessContext
from canonical.landscape.service.accesscontext.create import CreateAccessContext
from canonical.landscape.service.accesscontext.get import GetAccessContexts


@method
class GetRoles(AccountMethod):
    """
    Get all roles in the account.

    Arguments:

    - **names**: Optionally, a list of role names to get. Only matching roles
      will be returned.

    For example, the following request fetches all roles in the caller's
    account::

        ?action=GetRoles

    The method returns a JSON serialized representation of the account roles::

        [{"key": 1012,
          "name": "MyRole",
          "description": "",
          "permissions": [],
          "global_permissions": [],
          "persons": [],
          "access_groups": []}
        ]
    """

    requires = "ViewRole"

    schema = Schema(
        name="GetRoles",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc="A list of role names to limit the result.",
            )
        ],
        errors=[],
        doc="""Get all available roles.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Return all account roles."""
        roles = context.get_roles(names=call.args.names or None)
        return [role.get_state() for role in roles]


@method
class CopyRole(AccountMethod):
    """
    Copy an existing access role to an access role with a new name.

    Arguments:

    - **name**: The name of an existing access role to copy.
    - **destination_name**: The name of the copied role. It must start with a
      letter and can contain alphanumeric characters, '-' and '+'.
    - **description**: Optional description of the role.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name is found.
    - **DuplicateRole**: A role with the specified name exists.
    - **InvalidRoleName**: The provided name is not valid for a role.

    For example, the following request creates a role named 'BaseRole1'
    from BaseRole::

        ?action=CopyRole&name=BaseRole&destination_name=BaseRole1

    The method returns a JSON serialized representation of the new role::

        {"key": 1012,
         "name": "BaseRole1",
         "description": "",
         "permissions": [],
         "persons": [],
         "access_groups": []
        }

    """

    requires = "EditRole"

    schema = Schema(
        name="CopyRole",
        parameters=[
            Unicode("name", doc="The name of the existing role."),
            Unicode("destination_name", doc="The name of the new role."),
            Unicode(
                "description", optional=True, doc="The description of the new role."
            ),
        ],
        errors=[UnknownRoleError, DuplicateRoleError, InvalidRoleNameError],
        doc="""Copy an existing access role to a role with a new name.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Copies an existing access role"""
        role = context.copy_role(
            call.args.name, call.args.destination_name, call.args.description
        )
        return role.get_state()


@method
class CreateRole(AccountMethod):
    """
    Create a new access role.

    Arguments:

    - **name**: The name of the role. It must start with a letter and can
      contain alphanumeric characters, '-' and '+'.
    - **description**: Optional description of the role.

    The following errors may be raised:

    - **DuplicateRole**: A role with the specified name exists.
    - **InvalidRoleName**: The provided name is not valid for a role.

    For example, the following request creates a role named 'MyRole'::

        ?action=CreateRole&name=MyRole

    The method returns a JSON serialized representation of the new role::

        {"key": 1012,
         "name": "MyRole",
         "description": "",
         "permissions": [],
         "persons": [],
         "access_groups": []
        }
    """

    requires = "EditRole"

    schema = Schema(
        name="CreateRole",
        parameters=[
            Unicode("name", doc="The name of the role."),
            Unicode("description", optional=True, doc="The description of the role."),
        ],
        errors=[DuplicateRoleError, InvalidRoleNameError],
        doc="""Create a new access role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Creates a new access role"""
        role = context.create_role(call.args.name, call.args.description)
        return role.get_state()


@method
class AddPermissionsToRole(AccountMethod):
    """
    Add permissions to a role.

    Arguments:

    - **name**: The name of the role to modify.
    - **permissions**: A list of permissions to add.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name was found.
    - **InvalidPermissions**: One or more of the given permissions are invalid.
    - **ReadOnlyRole**: The specified role can't be modified.

    For example, the following request adds the 'ExecuteScript' permission
    to the role named 'MyRole'::

        ?action=AddPermissionsToRole&name=MyRole&permissions.1=ExecuteScript

    The method returns a JSON serialized representation of the edited role::

        {"key": 1012,
         "name": "MyRole",
         "description": "",
         "permissions": ["ExecuteScript"],
         "persons": [],
         "access_groups": []
        }
    """

    requires = "EditRole"

    schema = Schema(
        name="AddPermissionsToRole",
        parameters=[
            Unicode("name", doc="The name of the role to modify."),
            List("permissions", Unicode(), doc="A list of permissions to add."),
        ],
        errors=[UnknownRoleError, InvalidPermissionsError, ReadOnlyRoleError],
        doc="""Add permissions to a role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Add permissions to an access role"""
        role_name = call.args.name
        role = context.get_role(role_name)
        permissions = role.get_permissions()
        permissions.extend(call.args.permissions)
        new_permissions = set(permissions)
        context.edit_role(role_name, permissions=new_permissions)

        return role.get_state()


@method
class RemovePermissionsFromRole(AccountMethod):
    """
    Remove permissions from a role.

    Arguments:

    - **name**: The name of the role to modify.
    - **permissions**: A list of permissions to remove.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name was found.
    - **InvalidPermissions**: One or more of the given permissions are invalid.
    - **ReadOnlyRole**: The specified role can't be modified.

    For example, the following request will remove the 'ExecuteScript'
    permission to the role named 'MyRole'::

        ?action=RemovePermissionsFromRole&name=MyRole
        &permissions.1=ExecuteScript

    The method returns a JSON serialized representation of the edited role::

        {"key": 1012,
         "name": "MyRole",
         "description": "",
         "permissions": [],
         "persons": [],
         "access_groups": []
        }
    """

    requires = "EditRole"

    schema = Schema(
        name="RemovePermissionsFromRole",
        parameters=[
            Unicode("name", doc="The name of the role to modify."),
            List("permissions", Unicode(), doc="A list of permissions to remove."),
        ],
        errors=[UnknownRoleError, InvalidPermissionsError, ReadOnlyRoleError],
        doc="""Remove permissions from a role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Remove permissions from an access role"""
        role_name = call.args.name
        role = context.get_role(role_name)
        existing_permissions = role.get_permissions()
        permissions_to_remove = call.args.permissions
        invalid_macro_permissions = set(permissions_to_remove) - set(
            get_macro_permissions(system=False)
        )
        if invalid_macro_permissions:
            raise InvalidPermissionsError(invalid_macro_permissions)
        new_permissions = set(existing_permissions).difference(permissions_to_remove)
        context.edit_role(role_name, permissions=new_permissions)
        return role.get_state()


@method
class RemoveRole(AccountMethod):
    """
    Removes an access role.

    Arguments:

    - **name**: The name of the role.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name was found.
    - **InvalidRoleName**: When trying to remove the default GlobalAdmin role.

    For example, the following request removes a role named 'MyRole'::

        ?action=RemoveRole&name=MyRole

    An empty response is returned is the role is successfuly removed::

        {}
    """

    requires = "EditRole"

    schema = Schema(
        name="RemoveRole",
        doc="Remove an access role.",
        parameters=[Unicode("name", doc="The name of the role.")],
        errors=[UnknownRoleError, ReadOnlyRoleError],
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Remove an access role"""
        context.remove_role(call.args.name)
        return {}


@method
class AddPersonsToRole(AccountMethod):
    """
    Add the given persons to a role. Those persons will be granted the role.

    Arguments:

    - **name**: The name of the role to modify.
    - **persons**: A list of emails of persons to add.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name was found.
    - **UnknownPersons**: One or more of the given emails are unknown.

    For example, the following request grants the role named 'MyRole' to
    the person in the account with email 'john@example.com'::

        ?action=AddPersonsToRole&name=MyRole&persons.1=john@example.com

    The method returns a JSON serialized representation of the edited role::

        {"key": 1012,
          "name": "MyRole",
          "description": "",
          "permissions": [],
          "persons": ["john@example.com"],
          "access_groups": []
        }
    """

    requires = "EditRole"

    schema = Schema(
        name="AddPersonsToRole",
        parameters=[
            Unicode("name", doc="The name of the role to modify."),
            List("persons", Unicode(), doc="A list of emails of persons to add."),
        ],
        errors=[UnknownRoleError, UnknownPersonsError],
        doc="""Add permissions to a role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Add persons to an access role"""
        context.edit_membership(call.args.name, added=call.args.persons)
        role = context.get_role(call.args.name)
        return role.get_state()


@method
class RemovePersonsFromRole(AccountMethod):
    """
    Remove the given people from a role.

    Arguments:

    - **name**: The name of the role to modify.
    - **persons**: A list of the email addresses of people to remove.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name was found.
    - **UnknownPersons**: One or more of the given email addresses are unknown.

    For example, the following request removes the role named 'MyRole' from
    the person in the account with email 'john@example.com'::

        ?action=RemovePersonsFromRole&name=MyRole&persons.1=john@example.com

    The method returns a JSON serialized representation of the edited role::

        {"key": 1012,
          "name": "MyRole",
          "description": "",
          "permissions": [],
          "persons": [],
          "access_groups": []
        }
    """

    requires = "EditRole"

    schema = Schema(
        name="RemovePersonsFromRole",
        parameters=[
            Unicode("name", doc="The name of the role to modify."),
            List(
                "persons",
                Unicode(),
                doc="A list of email addresses of people to remove.",
            ),
        ],
        errors=[UnknownRoleError, UnknownPersonsError, SelfRevokeGlobalAdminError],
        doc="""Remove people from a role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Remove people from a role"""
        role = context.get_role(call.args.name)
        if (
            role.name == "GlobalAdmin"
            and context.person.email in call.args.persons
            and context.person in role.get_persons()
        ):
            raise SelfRevokeGlobalAdminError(
                "Can't revoke yourself the 'GlobalAdmin' role"
            )

        context.edit_membership(call.args.name, removed=call.args.persons)
        return role.get_state()


@method
class AddAccessGroupsToRole(AccountMethod):
    """
    Add the given access groups to a role.

    Arguments:

    - **name**: The name of the role to modify.
    - **access_groups**: A list of names of access groups to add.

    The following errors may be raised:

    - **UnknownRole**: No role with the specified name was found.
    - **UnknownAccessGroups**: One or more of the given access groups are
      unknown.
    - **ReadOnlyRoleError**: The role's access groups are read-only.

    For example, the following request adds the access group 'my-group' to the
    role named 'MyRole'. Any person with this role granted will inherit
    'my-group' permissions::

        ?action=AddAccessGroupsToRole&name=MyRole
        &access_groups.1=my-group

    The method returns a JSON serialized representation of the edited role::

        {"key": 1012,
         "name": "MyRole",
         "description": "",
         "permissions": [],
         "persons": [],
         "access_groups": ["my-group"]
        }
    """

    requires = "EditRole"

    schema = Schema(
        name="AddAccessGroupsToRole",
        parameters=[
            Unicode("name", doc="The name of the role to modify."),
            List(
                "access_groups",
                Unicode(),
                doc=("A list of names of access groups to add " "to the role."),
            ),
        ],
        errors=[UnknownRoleError, UnknownAccessContextsError, ReadOnlyRoleError],
        doc="""Add the given access groups to a role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Add access groups to an access role"""
        role = context.get_role(call.args.name)
        access_contexts = [i.name for i in role.get_access_contexts()]
        access_contexts.extend(call.args.access_groups)
        context.edit_role(call.args.name, access_contexts=access_contexts)
        return role.get_state()


@method
class RemoveAccessGroupsFromRole(AccountMethod):
    """
    Remove the given access groups from a role.

    Arguments:

    - **name**: The name of the role to modify.
    - **access_groups**: A list of names of access groups to remove.

    For example, the following request removes the access group 'my-group' from
    the role named 'MyRole'. Any person with this role granted will lose
    'my-group' permissions::

        ?action=RemoveAccessGroupsFromRole&name=MyRole
        &access_groups.1=my-group

    The method returns a JSON serialized representation of the edited role::

        {"key": 1012,
         "name": "MyRole",
         "description": "",
         "permissions": [],
         "persons": [],
         "access_groups": []
        }

    """

    requires = "EditRole"

    schema = Schema(
        name="RemoveAccessGroupsFromRole",
        parameters=[
            Unicode("name", doc="The name of the role to modify."),
            List(
                "access_groups",
                Unicode(),
                doc=("A list of names of access groups to remove " "from the role."),
            ),
        ],
        errors=[UnknownRoleError, ReadOnlyRoleError, UnknownAccessContextsError],
        doc="""Remove the given access groups to a role.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        """Remove access groups from an access role"""
        role = context.get_role(call.args.name)
        # Pre-check to make sure these are valid access contexts.
        account_access_contexts = [
            i.name for i in context.account.get_access_contexts()
        ]
        invalid_ones = set(call.args.access_groups) - set(account_access_contexts)
        if invalid_ones:
            invalid_ones = sorted(list(invalid_ones))
            raise UnknownAccessContextsError(invalid_ones, [], context.account)
        current_access_contexts = [i.name for i in role.get_access_contexts()]
        to_remove = set(current_access_contexts) - set(call.args.access_groups)
        context.edit_role(call.args.name, access_contexts=to_remove)
        return role.get_state()


@method
class GetAccessGroups(AccountMethod):
    """
    Get all access groups in the account.

    Arguments:

    - **names**: Optionally, a list of access group names to get. Only matching
      access groups will be returned.

    For example, the following request fetches all access groups in the
    caller's account::

        ?action=GetAccessGroups

    The method returns a JSON serialized representation of the account access
    groups::

        [{"name": "MyAccessGroup",
          "title": "",
          "parent": "ParentAccessGroup",
          "children": ""}]
    """

    schema = Schema(
        name="GetAccessGroups",
        parameters=[
            List("names", Unicode(), optional=True, doc="The name of the access group.")
        ],
        errors=GetAccessContexts.errors(),
        doc="""List all access groups in the account.""",
    )

    route = CollectionLookupRoute(AccessContextCollectionLookup, "names")

    def query(self, context, call):
        """List a access groups"""
        operation = GetAccessContexts(context, call.principal)
        access_contexts = operation()
        v2 = call.version == FUTURE_VERSION
        return [access_context.get_state(v2=v2) for access_context in access_contexts]


@method
class CreateAccessGroup(AccountMethod):
    """
    Create a new access group.

    Arguments:

    - **title**: The title of the access group.
    - **parent**: Optionally, the name of the access group that this access
      group should be added as a child of.  If this parameter is omitted the
      child will be added below the root access group of the account.

    The following errors may be raised:

    - **DuplicateAccessGroup**: An access group with the specified title
      already exists.
    - **InvalidAccessGroup**: The provided name is not valid for an access
      group.

    For example, the following request creates an access group named
    'Production' as child of the access group 'Servers'::

        ?action=CreateAccessGroup&name=Production
        &title=Production&parent=Servers

    The method returns a JSON serialized representation of the new access
    group::

        {"title":   "MyAccessGroup",
         "parent":  "ParentAccessGroup",
         "children": ""
        }
    """

    schema = Schema(
        name="CreateAccessGroup",
        parameters=[
            UnicodeTitle("title", doc="The title of the access group."),
            Unicode(
                "parent", optional=True, doc="The title of the parent access group."
            ),
        ],
        errors=CreateAccessContext.errors(),
        doc="""Create a new access group.""",
    )

    schema_v2 = Schema(
        name="CreateAccessGroup",
        parameters=[
            Unicode("name", doc="The name of the access group."),
            Unicode(
                "description", optional=True, doc="Description of the access group."
            ),
            Unicode(
                "parent",
                optional=True,
                doc="The name or ID of the parent access group.",
            ),
        ],
        errors=CreateAccessContext.errors(),
        doc="""Create a new access group.""",
    )

    schemas = {LATEST_VERSION: schema, FUTURE_VERSION: schema_v2}

    route = AccountRoute()

    def query(self, context, call):
        """Creates a new access group"""
        operation = CreateAccessContext(context, call.principal)
        v2 = call.version == FUTURE_VERSION
        if v2:
            name = call.args.name
            description = call.args.description
        else:
            name = None
            description = call.args.title
        access_context = operation(
            name, description=description, parent=call.args.parent
        )
        return access_context.get_state(v2=v2)


@method
class RemoveAccessGroup(AccountMethod):
    """
    Remove an access group.

    Arguments:

    - **name**: The name of the access group to remove.

    For example, the following request removes an access group named
    'MyAccessGroup', a child of the access group 'ParentAccessGroup'::

        ?action=RemoveAccessGroup&name=MyAccessGroup

    The method returns a JSON serialized representation of the state of the
    parent access group after the child is removed::

        {"name": "ParentAccessGroup",
         "title":  "parent",
         "parent": "GrandParentAccessGroup",
         "children": ""
        }
    """

    schema = Schema(
        name="RemoveAccessGroup",
        parameters=[Unicode("name", doc="The name of the access group to remove.")],
        errors=AccessContextLookup.errors() | RemoveAccessContext.errors(),
        doc="""Remove an access group.""",
    )

    route = EntityLookupRoute(factory=AccessContextLookup)

    def query(self, context, call):
        """Removes a new access group"""
        operation = RemoveAccessContext(context, call.principal)
        parent_access_context = operation()
        return parent_access_context.get_state()


@method
class GetPermissions(AccountMethod):
    """
    Get all available permissions.

    Example of valid call::

        ?action=GetPermissions

    The method returns a JSON serialized list of permissions::

        [{"name": "ViewComputer",
          "title": "View Computers"},
         {"name": "ManageComputer",
          "title": "Manage Computers"},
         ...]
    """

    requires = "ViewAccount"

    schema = Schema(name="GetPermissions", doc="""List all available permissions.""")

    route = AccountRoute()

    def query(self, context, call):
        """List macro-permissions"""
        return [
            {"name": permission.name, "title": permission.title}
            for permission in list(MACRO_PERMISSIONS.values())
        ]
