# Copyright 2005-2023 Canonical Limited. All rights reserved.

from datetime import datetime

from zope.component import getUtility

from canonical.txapi.server.schema import Schema, Unicode, Integer, Date, Bool, List
from canonical.txapi.server.method import method

from canonical.lib.timestamp import utctoday

from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.routes import GlobalRoute
from canonical.landscape.api.error import (
    MultipleAccountsError,
    UnknownUserError,
    InvalidDateError,
    MissingParameterError,
    InvalidParameterCombinationError,
    NoLDSLicenseError,
    InvalidAccountNameError,
    DuplicateAccountError,
    DuplicateEmailAddressError,
    PersonHasNoRoleError,
    UnauthorisedError,
    PersonAlreadyHasRoleError,
    DuplicateSalesforceAccountKeyError,
    UnknownAccountError,
    HasComputersError,
    LicenseNotFoundError,
)
from canonical.landscape.api.methods import Method
from canonical.landscape.model.account.invitation import InvitationManagement
from canonical.landscape.model.configuration import (
    LANDSCAPE_SYSTEM_EMAIL,
    is_standalone,
    set_root_url,
    set_default_system_email,
    ISystemConfiguration,
)
from canonical.landscape.model.main.person import (
    get_person_by_identity,
    create_person,
    get_person_by_email,
)
from canonical.landscape.model.main.account import (
    Account,
    bootstrap_account,
    get_account_by_name,
    is_valid_account_name,
    generate_account_name,
    get_trial_accounts,
    get_account_by_salesforce_account_key,
    PersonNotMemberOfAccountError,
    DEFAULT_STANDALONE_ACCOUNT_MAX_PEOPLE_COUNT,
)
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.account.errors import RepeatedInvitationError
from canonical.landscape.model.main.schema import InvalidEmailAddress
from canonical.landscape.model.main.standalone import AccountLDSLicenseAPI
from canonical.landscape.model.main.license import LDS_ROLES
from canonical.lib.mail import is_email_address
from canonical.security.role import UnknownRoleError


@method
class AddLicense(Method):
    """Add a new L{License} to an account.

    @param owner_id: The OpenID identity of the L{Person} creating the license,
        used to find the account.
    @param account_name: The name of the L{Account} this license is to be added
        to.
    @param seats: The number of seats the license covers.
    @param expiration_date: The date the license becomes invalid.
    @param role: Optional parameter with the name of the role to specify for
        the new L{License} default to BasicFeatures.
    @param salesforce_asset: Optional parameter with the asset key this
        license represents.

    Note: owner_id and account_name are mutually exclusive.
    """

    hidden = True
    requires = "CreateLicense"
    schema = Schema(
        name="AddLicense",
        parameters=[
            Unicode(
                "owner_id",
                optional=True,
                doc=(
                    "The OpenID identity URL of the person to get "
                    "account information for."
                ),
            ),
            Unicode(
                "account_name",
                optional=True,
                doc=("The name of the L{Account} to get information about."),
            ),
            Integer("seats", doc="The number of seats the license covers."),
            Unicode("expiration_date", doc="The date the license becomes invalid."),
            Unicode(
                "role",
                default="BasicFeatures",
                optional=True,
                doc=(
                    "The name of the role to specify for the new "
                    "L{License} default to BasicFeatures."
                ),
            ),
            Unicode(
                "salesforce_asset",
                optional=True,
                doc="The salesforce asset for this license.",
            ),
        ],
        errors=[
            UnknownRoleError,
            MissingParameterError,
            UnknownUserError,
            UnknownAccountError,
            InvalidParameterCombinationError,
            MultipleAccountsError,
            InvalidDateError,
            NoLDSLicenseError,
        ],
        doc="""
    Add a new L{License} to an account.

    Note: owner_id and account_name are mutually exclusive.""",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Retrieve the person, get its (hopefully) only account, and add a
        license to it.
        """
        account = self.get_account(call)
        expiration_date = self.get_expiration_date(
            call.args.expiration_date, "expiration_date"
        )
        role = call.args.role
        salesforce_asset = call.args.salesforce_asset

        lds_license = None
        if role in LDS_ROLES:
            api = AccountLDSLicenseAPI(account)
            lds_license = api.get_license()
            if lds_license is None:
                raise NoLDSLicenseError("Account not enabled for LDS")

        license = LicenseManagement(account).create_license(
            role, call.args.seats, expiration_date, salesforce_asset=salesforce_asset
        )

        if lds_license is not None:
            lds_license.add_license(license)
            lds_license.regenerate_contents()

        return license.get_state()

    def get_account(self, call):
        """
        Return the L{Account} to manage licenses for.

        The call arguments must contain either C{owner_id} or C{account_name}
        parameters.

        @raises MissingParameterError: if neither the C{owner_id} nor
            C{account_name} parameters are included in the request.
        @raises InvalidParameterCombinationError: if both the C{owner_id} and
            C{account_name} parameters are included in the request.
        @raises UnknownUserError: if the provided L{owner_id} is unknown.
        @raises MultipleAccountError: if the person is linked to more than one
            account.
        @raises UnknownAccountError: if no account is found.
        """
        if not call.args.owner_id and not call.args.account_name:
            raise MissingParameterError(
                "The request must contain an owner_id or account_name parameter."
            )
        if call.args.owner_id and call.args.account_name:
            raise InvalidParameterCombinationError(
                "The request cannot contain both the "
                "owner_id and account_name parameters."
            )

        if call.args.owner_id:
            person = self.get_person_by_identity(call.args.owner_id)
            account = self.get_account_by_person(person)
        else:
            account = self.get_account_by_name(call.args.account_name)
        return account

    def get_expiration_date(self, expiration, parameter_name):
        """
        Check that the provided expiration date is valid.
        The expected date format is 'YYYY-MM-DD'.

        @param expiration: a string with the expiration date or time
        @param parameter_name: name of expiration parameter in call
        @return: A C{datetime.date} instance representing the expiration date.
        """
        try:
            expiration_date = datetime.strptime(expiration, "%Y-%m-%d").date()
        except ValueError:
            raise InvalidDateError(
                f"Invalid format for {parameter_name}, should be YYYY-MM-DD"
            )

        if expiration_date < utctoday():
            raise InvalidDateError(f"{parameter_name} in the past: '{expiration}'")
        return expiration_date


@method
class EditLicense(Method):
    """Edit an existing L{License} for an account.

    @param owner_id: The OpenID identity of the L{Person} editing the license,
        used to find the account.
    @param account_name: The name of the L{Account} this license is associated
        with.
    @param seats: The number of seats the license covers.
    @param expiration_date: The date the license becomes invalid.
    @param role: The name of the role of the L{License}.
    @param salesforce_asset: The Salesforce asset key for this license.

    Note: owner_id and account_name are mutually exclusive.
    """

    hidden = True
    requires = "CreateLicense"
    schema = Schema(
        name="EditLicense",
        parameters=[
            Unicode(
                "owner_id",
                optional=True,
                doc=(
                    "The OpenID identity URL of the person to get "
                    "account information for."
                ),
            ),
            Unicode(
                "account_name",
                optional=True,
                doc=("The name of the L{Account} this license is associated with."),
            ),
            Integer("seats", doc="The number of seats the license covers."),
            Unicode("expiration_date", doc="The date the license becomes invalid."),
            Unicode(
                "role",
                default="BasicFeatures",
                doc=("The name of the role of the L{License}."),
            ),
            Unicode(
                "salesforce_asset",
                doc="The Salesforce asset key for this license.",
            ),
        ],
        errors=[
            UnknownRoleError,
            MissingParameterError,
            UnknownUserError,
            UnknownAccountError,
            InvalidParameterCombinationError,
            MultipleAccountsError,
            InvalidDateError,
            NoLDSLicenseError,
            LicenseNotFoundError,
        ],
        doc="""
    Edit an existing L{License}.

    Note: owner_id and account_name are mutually exclusive.""",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Retrieve the person, get its (hopefully) only account, and edit
        the matching license.
        """
        account = self.get_account(call)
        expiration_date = self.get_expiration_date(
            call.args.expiration_date, "expiration_date"
        )
        role = call.args.role
        salesforce_asset = call.args.salesforce_asset

        license = LicenseManagement(account).find_license(
            role, call.args.seats, salesforce_asset
        )
        if license is None:
            raise LicenseNotFoundError("No matching license found.")
        license = LicenseManagement(account).edit_license(
            license, expiration_date, salesforce_asset
        )

        return license.get_state()

    def get_account(self, call):
        """
        Return the L{Account} to manage licenses for.

        The call arguments must contain either C{owner_id} or C{account_name}
        parameters.

        @raises MissingParameterError: if neither the C{owner_id} nor
            C{account_name} parameters are included in the request.
        @raises InvalidParameterCombinationError: if both the C{owner_id} and
            C{account_name} parameters are included in the request.
        @raises UnknownUserError: if the provided L{owner_id} is unknown.
        @raises MultipleAccountError: if the person is linked to more than one
            account.
        @raises UnknownAccountError: if no account is found.
        """
        if not call.args.owner_id and not call.args.account_name:
            raise MissingParameterError(
                "The request must contain an owner_id or account_name parameter."
            )
        if call.args.owner_id and call.args.account_name:
            raise InvalidParameterCombinationError(
                "The request cannot contain both the "
                "owner_id and account_name parameters."
            )

        if call.args.owner_id:
            person = self.get_person_by_identity(call.args.owner_id)
            account = self.get_account_by_person(person)
        else:
            account = self.get_account_by_name(call.args.account_name)
        return account

    def get_expiration_date(self, expiration, parameter_name):
        """
        Check that the provided expiration date is valid.
        The expected date format is 'YYYY-MM-DD'.

        @param expiration: a string with the expiration date or time
        @param parameter_name: name of expiration parameter in call
        @return: A C{datetime.date} instance representing the expiration date.
        """
        try:
            expiration_date = datetime.strptime(expiration, "%Y-%m-%d").date()
        except ValueError:
            raise InvalidDateError(
                f"Invalid format for {parameter_name}, should be YYYY-MM-DD"
            )

        if expiration_date < utctoday():
            raise InvalidDateError(f"{parameter_name} in the past: '{expiration}'")
        return expiration_date


@method
class CreateAccount(Method):
    """
    Create a new L{Account}.

    @param name: The "slug" name for this Account, this should be a
        url-friendly version of the title. Optional.
    @param title: The proper name of the Account.
    @param salesforce_key: The Salesforce Key to be associated with the Person
        in the Account. **This is optional for backwards compatibility and
        ignored even if provided.**
    @param creator_email: The email address of the L{Account} creator.
    @param creator_name: The full name of the L{Account} creator.
    @param creator_user_id: The Ubuntu SSO OpenID of the L{Account} creator.
    @param invitation_name: The Name of the Person to be invited to the
        L{Account}.
    @param invitation_email: The email address to be sent the newly created
       invitation.

    NOTE: C{invitation_name} and C{creator_name} are mutually exclusive.
    """

    hidden = True
    requires = "CreateAccount"
    schema = Schema(
        name="CreateAccount",
        parameters=[
            Unicode(
                "name",
                optional=True,
                doc=(
                    'The "slug" name for this Account, this should be '
                    "a url-friendly version of the title."
                ),
            ),
            UnicodeTitle("title", doc="The proper name of the Account."),
            Unicode(
                "salesforce_key",
                optional=True,
                doc=(
                    "The Salesforce Key to be associated with the "
                    "Person in the Account. **Optional and ignored.**"
                ),
            ),
            Bool(
                "support_only",
                optional=True,
                doc=(
                    "If specified, the role will be set to SupportUser "
                    "instead of Owner."
                ),
            ),
            Unicode(
                "creator_email",
                optional=True,
                doc="The email address of the L{Account} creator.",
            ),
            Unicode(
                "creator_name",
                optional=True,
                doc="The full name of the L{Account} creator.",
            ),
            Unicode(
                "creator_user_id",
                optional=True,
                doc="The Ubuntu SSO OpenID of the L{Account} creator.",
            ),
            Unicode(
                "invitation_name",
                optional=True,
                doc=("The Name of the Person to be invited to the L{Account}."),
            ),
            Unicode(
                "invitation_email",
                optional=True,
                doc=("The email address to be sent the newly created invitation."),
            ),
        ],
        errors=[
            UnknownAccountError,
            InvalidAccountNameError,
            UnknownUserError,
            DuplicateAccountError,
            RepeatedInvitationError,
            DuplicateEmailAddressError,
            MissingParameterError,
            InvalidParameterCombinationError,
        ],
        doc="""
    Create a new L{Account}.

    NOTE: invitation_name and creator_name are mutually exclusive.
""",
    )

    def _get_account_name(self, title, name):
        """
        If we have a name, then we check its validity, otherwise, we generate a
        new one based upon the account title.
        """
        if name:
            if not is_valid_account_name(name):
                raise InvalidAccountNameError(f"Name '{name}' is invalid.")

            account = get_account_by_name(name)
            if account:
                raise DuplicateAccountError(f"Account '{name}' already exists.")
        else:
            name = generate_account_name(title)
        return name

    def _verify_parameters(self, arguments):
        """
        If creator_name is supplied as a parameter to the request, check for
        the presence of creator_email and creator_user_id parameters, otherwise
        raise a MissingParameter error.

        If invitation_name is supplied as a parameter, check for the presence
        of an invitation_email parameter, otherwise raise a MissingParameter
        error.

        If both creator_name and invitation_name parameters are supplied, raise
        an InvalidParameterCombination error.

        If neither creator_name nor invitation_name are supplied, raise a
        MissingParameter error.
        """
        if not arguments.creator_name and not arguments.invitation_name:
            raise MissingParameterError(
                "The request must contain the creator_name "
                "or invitation_name parameter"
            )

        if arguments.creator_name and arguments.invitation_name:
            raise InvalidParameterCombinationError(
                "The request cannot contain both the "
                "creator_name and invitation_name parameters"
            )
        if arguments.creator_name:
            if not arguments.creator_email:
                raise MissingParameterError(
                    "The request must contain the creator_email parameter"
                )
            if not arguments.creator_user_id:
                raise MissingParameterError(
                    "The request must contain the creator_user_id parameter"
                )
        elif arguments.invitation_name:
            if not arguments.invitation_email:
                raise MissingParameterError(
                    "The request must contain the invitation_email parameter"
                )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Instantiate an L{Account}, and add the L{Person} referenced by the
        supplied creator_user_id to the account as an admin.
        """
        self._verify_parameters(call.args)
        title = call.args.title

        account_name = self._get_account_name(title, call.args.name)

        account = bootstrap_account(account_name, title)

        if call.args.support_only:
            role_name = "SupportUser"
        else:
            role_name = "Owner"

        if call.args.creator_name:
            identity_url = call.args.creator_user_id
            person = get_person_by_identity(identity_url)
            if not person:
                if get_person_by_email(call.args.creator_email):
                    raise DuplicateEmailAddressError(
                        f"Email address '{call.args.creator_email}' already exists."
                    )
                person = create_person(
                    call.args.creator_name,
                    call.args.creator_email,
                    identity=identity_url,
                )
            account.add_person(person, roles=[role_name])

        elif call.args.invitation_name:
            management = InvitationManagement(account, context.person)
            invitation = management.invite(
                call.args.invitation_name,
                call.args.invitation_email,
                role_names=[role_name],
            )

            invitation.send_invitation()

        return account.get_state()


@method
class RemoveAccount(Method):
    """
    Remove an L{Account}.

    @param account_name: The name of the Account to remove.
    @param allow_reuse: If true, also remove the account from the
        removed_account_name table to permit its reuse. Defaults to false.
    """

    hidden = True
    requires = "RemoveAccount"
    schema = Schema(
        name="RemoveAccount",
        parameters=[
            Unicode("account_name", doc=("The name of the L{Account} to remove")),
            Bool(
                "allow_reuse",
                doc=(
                    "If true, don't remember the account name, permitting "
                    "its re-use later."
                ),
            ),
        ],
        errors=[UnknownAccountError, HasComputersError],
        doc=(
            "Remove the specified account from the system, optionally "
            "permitting it to be re-created in future."
        ),
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Remove the L{Account} identified by C{account_name}.

        @raises UnknownAccountError: Raised if the specified account name
            doesn't match a known L{Account}.
        @raises HasComputersError: Raised if any computers are still associated
            with the L{Account}.
        """
        account = get_account_by_name(call.args.account_name)
        if account is None:
            raise UnknownAccountError("No account matches the given account_name.")
        if account.get_computers_count():
            raise HasComputersError(
                "The account still has computers associated with it."
            )
        account.remove(call.args.allow_reuse)
        return {}


def bootstrap_account_helper(
    admin_email=None,
    admin_name=None,
    admin_password=None,
    system_email=None,
    registration_key=None,
    root_url=None,
    identity=None,
):
    """
    See BootstrapLDS API class for docstring.
    """

    if not is_standalone():
        raise UnauthorisedError("BootstrapLDS only available on LDS installs")

    account = get_account_by_name("standalone")
    if account:
        raise DuplicateAccountError("Account 'standalone' already exists.")

    if not is_email_address(admin_email):
        raise InvalidEmailAddress(admin_email)

    if system_email:
        if not is_email_address(system_email):
            raise InvalidEmailAddress(system_email)

    # Create LDS account and person
    #
    # XXX we need to create the account first and the person next, without
    # leveraging the 'person' parameter of bootstrap_account, because if
    # we create the person first, that will trigger a NewUserEvent which
    # in turn will eventually create a 'standalone' account. See #1397930.
    account = bootstrap_account(
        "standalone",
        "Standalone Account",
        registration_key=registration_key,
        max_people_count=DEFAULT_STANDALONE_ACCOUNT_MAX_PEOPLE_COUNT,
    )
    person = create_person(
        name=admin_name,
        email=admin_email,
        password=admin_password,
        identity=identity,
    )
    account.add_person(person, roles=["Owner"])
    credentials = person.create_api_credentials(account)

    # Final configuration to avoid registration forms upon first UI login
    if system_email:
        configuration = getUtility(ISystemConfiguration)
        configuration.set(LANDSCAPE_SYSTEM_EMAIL, system_email)

    if root_url:
        root_url = root_url.rstrip("/") + "/"
        set_root_url(root_url)
        if not system_email:  # Set a default email based on the root URL
            set_default_system_email(root_url)

    return {
        "LANDSCAPE_API_KEY": credentials.access_key_id.decode(),
        "LANDSCAPE_API_SECRET": credentials.access_secret_key.decode(),
    }


@method
class BootstrapLDS(Method):
    """
    Initialise the standalone L{Account} with an account admin.

    This method will only succeed if no L{Account}s are defined.

    @param admin_email: The email address of the L{Account} administrator.
    @param admin_name: The full name of the L{Account} administrator.
    @param admin_password: The password of the L{Account} administrator.
    @param registration_key: The optional registration key of the L{Account}.
    @param system_email: Optional email address that will appear on emails
        sent by Landscape. If unset and no root url, the admin will be prompted
        for this data upon first login to the UI. If there is a URL then a
        default will be created based on that, similar to the initial UI form
    @param root_url: Optional URL used to access Landscape's UI. In charmed
        deployment, this is of the form https://<apache_unit_ip>/. If unset,
        the admin will be prompted for this data upon first login to the UI.

    Return the API credentials for the account administrator created
    """

    hidden = True
    requires = None  # No permissions required for anonymous methods
    route = staticmethod(lambda _: (None, None))  # No context for anonymous methods

    schema = Schema(
        name="BootstrapLDS",
        parameters=[
            Unicode(
                "admin_email", doc="The email address of the L{Account} administrator."
            ),
            Unicode("admin_name", doc="The full name of the L{Account} administrator."),
            Unicode(
                "admin_password", doc="The password of the L{Account} administrator."
            ),
            Unicode(
                "system_email",
                optional=True,
                doc=("The email address that will appear on emails sent by Landscape."),
            ),
            Unicode(
                "registration_key", optional=True, doc="The account registration key."
            ),
            Unicode(
                "root_url",
                optional=True,
                doc=(
                    "The URL used to access Landscape's UI. In charmed "
                    "deployment, this takes the form "
                    "https://<apache_unit_ip>/."
                ),
            ),
        ],
        errors=[DuplicateAccountError, UnauthorisedError, InvalidEmailAddress],
        doc="""
    Initialise the standalone L{Account} with an account administrator.

    This method performs Landscape On Premises account registration
    and creates the specified account administrator. In response, it provides
    the created administrator's API credentials.
""",
    )

    def query(self, context, call):
        """
        Instantiate the C{standalone} L{Account} and add the L{Person}
        defined by the supplied admin_email and admin_name parameters as the
        C{standalone} account admin.
        """

        return bootstrap_account_helper(
            admin_email=call.args.admin_email,
            admin_name=call.args.admin_name,
            admin_password=call.args.admin_password,
            system_email=call.args.system_email,
            registration_key=call.args.registration_key,
            root_url=call.args.root_url,
        )


@method
class UpdateUser(Method):
    """Update a Person.

    @param user_id: The OpenID identity URL of the person to edit.
    @param salesforce_contact_id: The Salesforce contact ID used by the
        customer to access support cases.
    """

    hidden = True
    requires = "EditSalesforceKey"
    schema = Schema(
        name="UpdateUser",
        parameters=[
            Unicode("user_id", doc="The OpenID identity URL of the person to edit."),
            Unicode(
                "salesforce_contact_id",
                doc=(
                    "The Salesforce contact ID used by the customer to "
                    "access support cases."
                ),
            ),
        ],
        errors=[UnknownAccountError, UnknownUserError, MultipleAccountsError],
        doc="Update a Person with a Salesforce contact ID.",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Update the L{Person} identified by user_id to have the salesforce_key
        in salesforce_contact_id.
        """
        person = self.get_person_by_identity(call.args.user_id)
        account = self.get_account_by_person(person)
        salesforce_key = call.args.salesforce_contact_id
        salesforce_user = account.get_salesforce_user(person)

        if salesforce_user is None:
            account.create_salesforce_user(person, salesforce_key)
        else:
            salesforce_user.salesforce_key = salesforce_key
        return {}


@method
class ChangeRole(Method):
    """Change a Person role.

    @param user_id: The OpenID identity URL of the person to edit.
    @param current_role: The current role to change.
    @param new_role: The new role to give to the Person.
    """

    hidden = True

    requires = "EditPersonRole"
    schema = Schema(
        name="ChangeRole",
        doc="Change a Person role.",
        parameters=[
            Unicode("user_id", doc="The OpenID identity URL of the person to edit."),
            Unicode("current_role", doc="The current role to change."),
            Unicode("new_role", doc="The new role to give to the Person."),
        ],
        errors=[
            PersonHasNoRoleError,
            PersonAlreadyHasRoleError,
            MultipleAccountsError,
            UnknownAccountError,
            UnknownUserError,
            UnknownRoleError,
        ],
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Replace the current role with the new one for the L{Person} identified
        by user_id.
        """
        person = self.get_person_by_identity(call.args.user_id)
        account = self.get_account_by_person(person)

        current_role = self.get_role(call.args.current_role)
        new_role = self.get_role(call.args.new_role)
        access_context = account.access_context

        if not person.has_role(current_role, access_context):
            raise PersonHasNoRoleError(
                f"The specified person does not have role '{call.args.current_role}'"
            )
        if person.has_role(new_role, access_context):
            raise PersonAlreadyHasRoleError(
                f"The specified person already has role '{call.args.new_role}'"
            )

        access_context.remove_role(person, current_role)
        access_context.add_role(person, new_role)
        return sorted(role.name for role in person.get_all_roles()[access_context.id])


@method
class GetUserGlobalRoles(Method):
    """Return global roles for a Person.

    @param user_id: The OpenID identity URL of the person.
    """

    hidden = True

    requires = "ViewPerson"
    schema = Schema(
        name="GetUserGlobalRoles",
        doc="Return global roles for a Person.",
        parameters=[Unicode("user_id", doc="The OpenID identity URL of the person.")],
        errors=[UnknownUserError],
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Return global roles for a L{Person}.
        """
        person = self.get_person_by_identity(call.args.user_id)
        global_roles = person.get_all_roles().get(None, set())
        # Only return stored editable roles (that have a database key).
        return sorted(role.name for role in global_roles if role.key is not None)


@method
class EditUserGlobalRoles(Method):
    """Edit global roles for a Person.

    @param user_id: The OpenID identity URL of the person.
    @param add: List of roles to add.
    @param remove: List of roles to remove.
    """

    hidden = True

    requires = "EditPersonRole"
    schema = Schema(
        name="EditUserGlobalRoles",
        doc="Edit global roles for a Person.",
        parameters=[
            Unicode("user_id", doc="The OpenID identity URL of the person."),
            List("add", Unicode(), optional=True, doc="List of roles to add."),
            List("remove", Unicode(), optional=True, doc="List of roles to remove."),
        ],
        errors=[
            PersonHasNoRoleError,
            PersonAlreadyHasRoleError,
            UnknownUserError,
            UnknownRoleError,
            MissingParameterError,
        ],
    )

    route = GlobalRoute()

    _GLOBAL_ROLES = (
        "SupportProvider",
        "Operator",
        "KnowledgeBaseReviewer",
        "AccountManager",
    )

    def _check_roles(self, person, add_roles, remove_roles):
        """Check that provided roles are valid global roles, and they are
        consistent with current user roles.

        Raise errors if invalid role names are passed, the user already has
        roles to add or doesn't have roles to remove.
        """
        unknown_roles = (add_roles | remove_roles) - set(self._GLOBAL_ROLES)
        if unknown_roles:
            unknown_roles_msg = ", ".join(sorted(unknown_roles))
            raise UnknownRoleError(f"Unknown role(s): {unknown_roles_msg}")

        if not add_roles and not remove_roles:
            raise MissingParameterError(
                "The request must contain at least one 'add' or 'remove' parameter."
            )

        for role in remove_roles:
            if not person.has_global_role(role):
                raise PersonHasNoRoleError(
                    f"The specified person does not have role '{role}'"
                )
        for role in add_roles:
            if person.has_global_role(role):
                raise PersonAlreadyHasRoleError(
                    f"The specified person already has role '{role}'"
                )

    def query(self, context, call):
        """
        Edit global roles for the L{person} specified by C{user_id}.
        """
        person = self.get_person_by_identity(call.args.user_id)
        add_roles = set(call.args.add)
        remove_roles = set(call.args.remove)

        self._check_roles(person, add_roles, remove_roles)

        for role in add_roles:
            person.add_global_role(role)
        for role in remove_roles:
            person.remove_global_role(role)

        global_roles = person.get_all_roles().get(None, set())
        return sorted(role.name for role in global_roles if role.key is not None)


@method
class GetTrialAccounts(Method):
    """Get a list of trial accounts created in a specified interval.

    @param created_after_time: Return information about all trial accounts
        created after this time, in ISO 8601 format.
    @param created_before_time: Filter trial accounts created before this time,
        in ISO 8601 format (Optional).
    """

    hidden = True
    requires = "ViewAllAccounts"
    schema = Schema(
        name="GetTrialAccounts",
        parameters=[
            Date(
                "created_after_time",
                doc=(
                    "Return information about all trial accounts "
                    "created after this time, in ISO 8601 format."
                ),
            ),
            Date(
                "created_before_time",
                doc=(
                    "Filter trial accounts created before this time, "
                    "in ISO 8601 format"
                ),
                optional=True,
            ),
        ],
        errors=[],
        doc="Get a list of trial accounts created in a specified interval.",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """Perform a query to get information about trial L{Account}s.

        @return: A list of trial L{Account}s.
        """
        result = get_trial_accounts(
            call.args.created_after_time, call.args.created_before_time
        )
        return [account.get_state() for (trial, account) in result]


@method
class GetAccounts(Method):
    """Get a list of L{Account}s associated with a given L{Person}.

    This method may only be invoked by an C{AccountManager} and takes the
    following required parameters:

     * account_name: The name of the L{Account} to get information about.
     * owner_id: The OpenID identity URL of the person to get account
       information for.
     * email: The email address of the person to get account information for.

    One of the C{account_name}, C{owner_id}, or C{email} parameters must be
    provided.  C{account_name} and C{owner_id} can not both be provided in the
    same request.  If the specified user is unknown an C{UnknownUser} error
    is returned.
    """

    hidden = True
    requires = "ViewAllAccounts"

    schema = Schema(
        name="GetAccounts",
        parameters=[
            Unicode(
                "account_name",
                optional=True,
                doc=("The name of the L{Account} to get information about."),
            ),
            Unicode(
                "owner_id",
                optional=True,
                doc=(
                    "The OpenID identity URL of the person to get "
                    "account information for."
                ),
            ),
            Unicode(
                "email",
                optional=True,
                doc=(
                    "The email address of the person to get L{Account} "
                    "information for."
                ),
            ),
        ],
        errors=[
            InvalidParameterCombinationError,
            MissingParameterError,
            UnknownUserError,
            UnknownAccountError,
        ],
        doc="""
    Get a list of L{Account}s associated with a given L{Person}.

    This method may only be invoked by an C{AccountManager}.

    One of the C{account_name}, C{owner_id}, or C{email} parameters must be
    provided.  C{account_name} and C{owner_id} can not both be provided in the
    same request.
""",
    )

    def _verify_parameters(self, arguments):
        """Validate the arguments passed to the request.

        @raises MissingParameterError: raised if neither the
            C{owner_id}, C{account_name}, nor C{email} parameters are included
            in the request.
        @raises InvalidParameterCombinationError: raised if
            both the C{owner_id} and C{account_name} parameters are included in
            the request.
        """
        if (
            not arguments.owner_id
            and not arguments.account_name
            and not arguments.email
        ):
            raise MissingParameterError(
                "The request must contain an owner_id, account_name, or email "
                "parameter."
            )
        if arguments.owner_id and arguments.account_name:
            raise InvalidParameterCombinationError(
                "The request cannot contain both the owner_id and "
                "account_name parameters."
            )

    route = GlobalRoute()

    def query(self, context, call):
        """Perform a query to get L{Account}s associated with a L{Person}.

        @raises UnknownUserError: Raised if the specified OpenID identity URL
            doesn't match a known L{Person}.
        @raises UnknownAccountError: Raised if the specified account name
            doesn't match a known L{Account}.
        @return: A list of L{Account}s associated with the matching L{Person}.
        """
        self._verify_parameters(call.args)
        # Try owner_id, then account_name, then email
        result = None
        if call.args.owner_id:
            person = get_person_by_identity(call.args.owner_id)
            if person is not None:
                result = [
                    account.get_state()
                    for account in person.accounts.order_by(Account.name)
                ]
        elif call.args.account_name:
            account_name = str(call.args.account_name)
            account = get_account_by_name(account_name)
            if account is not None:
                result = [account.get_state()]
        if call.args.email:
            person = get_person_by_email(call.args.email)
            if person is not None:
                result = [
                    acc.get_state() for acc in person.accounts.order_by(Account.name)
                ]

        if result is None:
            # No matches found
            if call.args.account_name:
                error_type = UnknownAccountError
                if call.args.email:
                    message = "No account matches the given account_name or email."
                else:
                    message = "No account matches the given account_name."
            elif call.args.owner_id:
                error_type = UnknownUserError
                if call.args.email:
                    message = "No person matches the given owner_id or email."
                else:
                    message = "No person matches the given owner_id."
            else:
                error_type = UnknownUserError
                message = "No person matches the given email."
            raise error_type(message)
        return result


@method
class DisableAdminInAccount(Method):
    """Disable an Administrator identified by identity in a named account.

    @param user_id: The OpenID identity URL of the person to disable.
    @param account_name: The name of the Account to disable the administrator
        in.
    """

    hidden = True
    requires = "DisableAccountMembership"
    schema = Schema(
        name="DisableAdminInAccount",
        parameters=[
            Unicode("user_id", doc="The OpenID identity URL of the person to disable."),
            Unicode(
                "account_name",
                doc=(
                    "The name of the L{Account} to disable the "
                    "administrator in.about."
                ),
            ),
        ],
        errors=[UnknownAccountError, PersonNotMemberOfAccountError, UnknownUserError],
        doc=(
            "Disable an Administrator identified by an OpenID identity in a "
            "named account."
        ),
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Disable the L{Person} identified by C{user_id} in the account
        identified by C{account_name}.
        """
        account = self.get_account_by_name(call.args.account_name)
        person = self.get_person_by_identity(call.args.user_id)
        account.disable_membership(person)
        return {}


@method
class SetSalesforceAccountKey(Method):
    """Update the C{salesforce_account_key} for a given L{Account}.

    @param account_name: The C{Account.name} for the L{Account} to update
    @param salesforce_account_key: The Salesforce Key to be assigned to this
        L{Account}.
    """

    hidden = True
    requires = "EditSalesforceKey"
    schema = Schema(
        name="SetSalesforceAccountKey",
        parameters=[
            Unicode(
                "account_name", doc="The C{Account.name} for the L{Account} to update."
            ),
            Unicode(
                "salesforce_account_key",
                doc=("The Salesforce Key to be assigned to this L{Account}."),
            ),
        ],
        errors=[UnknownAccountError, DuplicateSalesforceAccountKeyError],
        doc="Update the C{salesforce_account_key} for a given L{Account}.",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Update the L{Account} identified by C{account_name} to have the
        supplied C{salesforce_account_key}.
        """
        account = self.get_account_by_name(call.args.account_name)
        salesforce_key = call.args.salesforce_account_key
        other_account = get_account_by_salesforce_account_key(salesforce_key)
        if other_account is not None:
            raise DuplicateSalesforceAccountKeyError(
                f"salesforce_account_key '{salesforce_key}' already in use."
            )

        account.salesforce_account_key = salesforce_key
        return {}


@method
class EnableLDS(Method):
    """Enable an account for LDS.

    @param account_name: The name of the L{Account} to be enabled.

    This method instantiates an LDSLicense which is a prerequisite to adding
    LDSLicenses.
    """

    hidden = True
    requires = "CreateLicense"
    schema = Schema(
        name="EnableLDS",
        parameters=[
            Unicode("account_name", doc="The name of the L{Account} to be enabled.")
        ],
        errors=[UnknownAccountError],
        doc="""
    Enable an account for LDS.

    This method instantiates an LDSLicense which is a prerequisite to adding
    LDSLicenses.
""",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Add an LDSLicense to contain LDS seats to the specified L{Account}.
        """
        account = self.get_account_by_name(call.args.account_name)
        api = AccountLDSLicenseAPI(account)
        if not api.get_license():
            api.create_license()
        return {}


@method
class DisableLDS(Method):
    """Disable an account for LDS.

    @param account_name: The name of the L{Account} to be disabled.
    """

    hidden = True
    requires = "EditLicense"
    schema = Schema(
        name="DisableLDS",
        parameters=[
            Unicode("account_name", doc="The name of the L{Account} to be disabled.")
        ],
        errors=[UnknownAccountError],
        doc="""
    Disable an account for LDS.

    This method removes an existing LDSLicense.
""",
    )

    route = GlobalRoute()

    def query(self, context, call):
        """
        Remove an LDSLicense from the specified L{Account}.
        """
        account = self.get_account_by_name(call.args.account_name)
        api = AccountLDSLicenseAPI(account)
        if api.get_license():
            api.remove_license()
        return {}
