# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Administrators
==============

The methods available here help you invite, list and disable administrators in
your account.
"""

from canonical.txapi.server.schema import Schema, Unicode, UnicodeLine, List
from canonical.txapi.server.method import method

from canonical.landscape.api.error import MissingParameterError, UnknownUserError
from canonical.landscape.model.main.account import PersonNotMemberOfAccountError
from canonical.landscape.model.account.errors import (
    MaximumAccountAdministratorsError,
    RepeatedInvitationError,
)
from canonical.landscape.model.access.errors import UnknownRoleError
from canonical.landscape.model.main.schema import InvalidEmailAddress
from canonical.landscape.model.main.person import get_person_by_email
from canonical.landscape.model.account.invitation import InvitationManagement
from canonical.landscape.model.account.management import AccountManagement
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.routes import ManagementRoute


__all__ = ["InviteAdministrator", "GetAdministrators", "DisableAdministrator"]


@method
class InviteAdministrator(AccountMethod):
    """
    Invite an administrator to your account.

    - **name**: The name of the person to invite.
    - **email**: The email address of the administrator, to which the
      invitation will be send.

    - **roles**: If specified, the roles that the administrator is going to
      have in your account. Default to GlobalAdmin.

    For example::

        ?action=InviteAdministrator&name=John%2ODoe&email=john@example.com
    """

    requires = "InviteAdministrator"

    schema = Schema(
        name="InviteAdministrator",
        parameters=[
            UnicodeLine("name", doc="The name of the person to invite."),
            UnicodeLine(
                "email",
                doc="The email address of the administrator, to "
                "which the invitation will be send.",
            ),
            List(
                "roles",
                Unicode(),
                optional=True,
                default=["GlobalAdmin"],
                doc="If specified, the roles that the administrator is "
                "going to have in your account. "
                "Default to GlobalAdmin",
            ),
        ],
        errors=[
            UnknownRoleError,
            RepeatedInvitationError,
            MissingParameterError,
            MaximumAccountAdministratorsError,
            InvalidEmailAddress,
        ],
        doc="""Invite an administrator to your account.""",
    )

    route = ManagementRoute(factory=InvitationManagement)

    def query(self, context, call):
        invitation = context.invite(
            call.args.name, call.args.email, role_names=call.args.roles
        )
        invitation.send_invitation()
        return {}


@method
class GetAdministrators(AccountMethod):
    """Retrieve the list of administrators in the account.

    Here's an example output::

      [{"email": u"john@example.com",
        "id": 40042,
        "name": u"John Doe",
        "role": u"Owner"}]
    """

    requires = "ViewAccount"
    schema = Schema(
        name="GetAdministrators",
        parameters=[],
        doc="""Retrieve the list of administrators in the account.""",
    )

    route = ManagementRoute(factory=AccessControlManagement)

    def query(self, context, call):
        persons_roles = context.get_persons_roles()

        result = []
        for person, roles in list(persons_roles.items()):
            data = person.get_state()
            data["roles"] = sorted(role.name for role in roles)
            result.append(data)
        return result


@method
class DisableAdministrator(AccountMethod):
    """
    Disable an administrator of your account.

    - **email**: The email address of the administrator to disable.

    For example::

        ?action=DisableAdministrator&email=john@example.com
    """

    requires = "DisableMembership"

    schema = Schema(
        name="DisableAdministrator",
        parameters=[Unicode("email", doc="The name of the person to disable.")],
        errors=[UnknownUserError, PersonNotMemberOfAccountError],
        doc="""Disable an administrator of your account.""",
    )

    route = ManagementRoute(factory=AccountManagement)

    def query(self, context, call):
        management = context
        person = get_person_by_email(call.args.email)
        if person is None:
            raise UnknownUserError("No person matches the given email address.")
        management.disable_membership(person)
        return {}
