# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Alerts
======

Methods for retrieving or editing alerts for an account.
"""

from canonical.txapi.server.method import method
from canonical.txapi.server.schema import Unicode

from canonical.landscape.model.alert.rule import (
    get_rule_info_class_by_name,
    UnknownAlertTypeError,
    InvalidAlertTypeError,
)
from canonical.landscape.model.alert.management import AlertManagement
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.api.methods import (
    AccountMethod,
    Method,
    Schema,
    TagAssociationMethod,
)
from canonical.landscape.api.routes import AccountRoute, SubjectRoute
from canonical.landscape.service.alert.get import GetAlertsOperation
from canonical.landscape.service.alert.subscribe import SubscribeToAlertOperation


@method
class GetAlerts(AccountMethod):
    """Get a list of the alerts on the account and computers.

    Example of a valid request::

        ?action=GetAlerts

    The method returns a JSON serialised list of the alerts::

        [{u'alert_type': u'PackageReporterAlert',
          u'all_computers': True,
          u'description': u'Alert when package reporting fails',
          u'scope': u'computer',
          u'status': u'OK',
          u'subscribed': u'Yes',
          u'tags': [u'file-server',
                    u'web-server',
                    u'desktop',
                    u'laptop']},]

    """

    schema = Schema(
        name="GetAlerts", parameters=[], doc="""Get a list of alerts in an account."""
    )

    route = AccountRoute()

    def query(self, context, call):
        """Get a list of alerts in an account."""
        operation = GetAlertsOperation(context, call.principal)
        return [
            {
                "alert_type": rule["alert_type"],
                "description": rule["description"],
                "subscribed": rule["subscribed"],
                "status": rule["status"],
                "scope": rule["scope"],
                "all_computers": rule["all_computers"],
                "tags": rule["tags"],
            }
            for rule in operation()
        ]


@method
class GetAlertSubscribers(Method):
    """Get a list of the subscribers to a given alert type.

    - **alert_type**: The name of the alert whose subscribers should be
      listed. Valid names can be seen by calling the GetAlerts method.

    Example of a valid request::

        ?action=GetAlertSubscribers&alert_type=ComputerOfflineAlert

    The method returns a JSON serialized list of the alert subscribers::

        [{u'email': u'person1@example.com',
          u'name': u'Person',
          u'id': 1},
         {u'email': u'person2@example.com',
          u'name': u'Person',
          u'id': 2}]

    The following errors may be raised:

    - **UnknownAlertTypeError**: If an unknown alert_type is passed.

    """

    requires = "GetAlertSubscription"
    schema = Schema(
        name="GetAlertSubscribers",
        parameters=[
            Unicode("alert_type", doc="The alert type to check the subscription on.")
        ],
        errors=[UnknownAlertTypeError],
        doc="Get a a list of people who subscribe to a given alert.",
    )

    route = AccountRoute()

    def query(self, context, call):
        account = context
        alert_type = call.args.alert_type
        info_class = get_rule_info_class_by_name(alert_type)
        [rule] = account.get_rules(info_classes=[info_class])
        if rule is None:
            return []
        return [
            subscriber.get_state()
            for subscriber in list(rule.get_subscribers().order_by("id"))
        ]


@method
class SubscribeToAlert(AccountMethod):
    """Subscribe your user account to a given alert types notifications.

    - **alert_type**: The name of the alert type you wish to subscribe to.
      Valid names can be seen by calling the GetAlerts method.

    Example of a valid request::

        ?action=SubscribeToAlert&alert_type=ComputerOfflineAlert

    The following errors may be raised:

    - **UnknownAlertTypeError**: If an unknown alert_type is passed.
    """

    schema = Schema(
        name="SubscribeToAlert",
        parameters=[
            Unicode("alert_type", doc="The alert type to add a subscription to.")
        ],
        errors=SubscribeToAlertOperation.errors(),
        doc="Subscribe to an alert.",
    )

    route = AccountRoute()

    def query(self, context, call):
        """Subscribe to a given alert type on this account."""
        operation = SubscribeToAlertOperation(context, call.principal)
        operation(alert_type=call.args.alert_type)
        return {}


@method
class UnsubscribeFromAlert(Method):
    """Unsubscribe your user account from a given alert types notifications.

    - **alert_type**: The name of the alert type from which you wish to remove
      your subscription. Valid names can be seen by calling the GetAlerts
      method.

    Example of a valid request::

        ?action=UnsubscribeFromAlert&alert_type=ComputerOfflineAlert

    The following errors may be raised:

    - **UnknownAlertTypeError**: If an unknown alert_type is passed.

    """

    requires = "SubscribeAlert"
    schema = Schema(
        name="UnsubscribeFromAlert",
        parameters=[
            Unicode("alert_type", doc="The alert type to remove a subscription from.")
        ],
        errors=[UnknownAlertTypeError],
        doc="Unsubscribe from an alert.",
    )

    route = AccountRoute()

    def query(self, context, call):
        """Unsubscribe from a given alert in a given account."""
        account = context
        person = get_person_by_id(call.principal.id)
        alert_type = call.args.alert_type
        info_class = get_rule_info_class_by_name(alert_type)
        [rule] = account.get_rules(info_classes=[info_class])
        rule.unsubscribe(person)
        return {}


class AlertTagAssociationMethod(TagAssociationMethod):
    """A common superclass for alert tag association/disassociation methods."""

    def _get_state(self, context, call):
        """Generate the return value of the API call.

        We override this method to restrict the results
        returned to the computers the person making the call is allowed to
        see alert status information about.
        """
        person = get_person_by_id(call.principal.id)
        account = get_account_by_id(call.principal.account_id)
        computer_ids = person.get_computer_ids(account)
        return context.get_state_for_computers(computer_ids=computer_ids)


@method
class AssociateAlert(AlertTagAssociationMethod):
    """Associate an alert to computers with specific tags or to all
    computers.

    - **name**: Name of alert type.
    - **tags.#**: List of tag names to associate to the alert.
    - **all_computers**: true if the alert should be associated to all
      computers. This parameter is *optional* and defaults to
      *false*. Individual tags associated to the alert will remain, but they
      will only be effective if the "all_computers" flag is later disabled.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request that associates the alert 'ComputerOfflineAlert'
    with the tags 'desktop' and 'laptop'::

        ?action=AssociateAlert&name=ComputerOfflineAlert
            &tags.1=desktop
            &tags.2=laptop
            &all_computers=false

    The following errors may be raised:

    - **UnknownAlertTypeError**: No alert with the specified name exists.
    - **InvalidAlertTypeError**: Raised when trying to associate tags with an
      alert with account scope instead of computer scope.
    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.

    The state of the alert will be returned.
    """

    requires = "EditAlert"

    schema = TagAssociationMethod.schema.extend(
        name="AssociateAlert",
        errors=[UnknownAlertTypeError, InvalidAlertTypeError],
        doc="""
    Associate an alert to computers with specific tags or to all computers.
    If a tag with a given name doesn't exist, it will be automatically
    created. An 'all_computers' value of 'true' and a list of tags are mutually
    exclusive. Only one or the other may be passed.""",
    )

    route = SubjectRoute(
        factory=AlertManagement,
        getter="get_computer_alert_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownAlertTypeError,
    )


@method
class DisassociateAlert(AlertTagAssociationMethod):
    """
    Disassociate an alert from computers with specified tags or from all
    computers.

    If **all_computers=true**, the alert will be unflagged as applying to all
    computers, but will still be enabled for computers which have specific tags
    associated with it.

    - **name**: Name of the alert type.
    - **tags.#**: List of tag names to disassociate from the alert.
    - **all_computers**: If true, the alert will only remain enabled for
      computers with tags associated to the alert. This parameter is
      *optional* and defaults to *false*.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=DisassociateAlert&name=ComputerOfflineAlert
            &tags.1=desktop&tags.2=laptop

    The following errors may be raised:

    - **UnknownAlertTypeError**: No alert with the specified name exists.
    - **UnknownTag**: No tag with the specified name exists.
    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.
    - **InvalidAlertTypeError**: Raised when trying to discassociate tags with
      an alert with account scope instead of computer scope.

    The state of the alert will be returned.
    """

    requires = "EditAlert"

    associate = False

    schema = TagAssociationMethod.schema.extend(
        name="DisassociateAlert",
        errors=[UnknownAlertTypeError, InvalidAlertTypeError],
        doc="""Disassociate an alert from computers with specific tags or from
        all computers.""",
    )

    route = SubjectRoute(
        factory=AlertManagement,
        getter="get_computer_alert_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownAlertTypeError,
    )
