# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Upgrade profiles
================

The methods defined here let you schedule automatic upgrades.
"""

from zope.security.interfaces import Unauthorized

from canonical.txapi.server.method import method
from canonical.txapi.server.schema import (
    Schema,
    Unicode,
    Bool,
    Integer,
    MissingParameterError,
    List,
)

from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.error import InvalidParameterCombinationError
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.api.methods import AccountMethod, TagAssociationMethod
from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.errors import UnknownAccessContextError
from canonical.landscape.model.package.autoupgrade import (
    UpgradeProfileManagement,
    ScheduleFormatException,
    UPGRADE_TYPE_MAP,
    get_upgrade_type,
)
from canonical.landscape.model.package.errors import UnknownUpgradeProfileError


__all__ = [
    "AssociateUpgradeProfile",
    "CreateUpgradeProfile",
    "DisassociateUpgradeProfile",
    "EditUpgradeProfile",
    "RemoveUpgradeProfile",
    "GetUpgradeProfiles",
]


FREQUENCIES = {"week": "WEEKLY", "hour": "HOURLY"}


def _validate_upgrade_type(upgrade_type):
    """
    Validate that the passed upgrade type is None or one of the valid keys used
    by L{UpgradeProfile.upgrade_type}
    """
    valid = False
    if upgrade_type is None:
        valid = True
    if upgrade_type in list(UPGRADE_TYPE_MAP.keys()):
        valid = True
    return valid


def _validate_frequency(frequency):
    """
    Validate that the passed frquency is one of our whitelisted option.
    """
    return frequency.lower() in list(FREQUENCIES.keys())


def arguments_to_ical(args):
    """
    This function takes the API call's arguments and returns a corresponding
    iCal String. Typically, this method would be passed "call.args" from an
    API method's query().
    """
    on_day = args.on_days

    if args.every is None:
        raise MissingParameterError("every")

    frequency = args.every.lower()
    ical_frequency = FREQUENCIES.get(frequency)
    ical = "FREQ=%s" % ical_frequency  # This is now HOURLY or WEEKLY

    if on_day:
        days_string = ",".join(on_day)
        ical += ";BYDAY=%s" % days_string

    if ical_frequency == "WEEKLY":
        # Days
        if not on_day:
            raise MissingParameterError("on_days")
        # Hour
        if args.at_hour is None:
            raise MissingParameterError("at_hour")
        ical += ";BYHOUR=%s" % args.at_hour
    else:
        # The frequency is "hour".
        if args.at_hour is not None:
            raise InvalidParameterCombinationError(
                "at_hour cannot be used for every='hour'"
            )

    if args.at_minute is None:
        raise MissingParameterError("at_minute")
    ical += ";BYMINUTE=%s" % args.at_minute

    return ical


@method
class CreateUpgradeProfile(AccountMethod):
    """
    Create an upgrade profile.

    Arguments:

    - **title**: The human readable title for this upgrade profile.
    - **every**: The frequency at which you wish this upgrade to be
      executed. Valid choices are "hour" and "week".
    - **on_days**: A list of days of the week on which the upgrade profile
      will be run. The day names must be abbreviated to their first two
      letters, as: "mo", "tu", "we", "th", "fr", "sa", "su". Must be specified
      when the every parameter is "week", but optional if the every parameter
      is "hour".
    - **at_hour**: The hour, in 24h format, at which the upgrade profile
      will be run.
    - **at_minute**: The minute of the hour (0-59) at which the upgrade
      profile will be run.
    - **deliver_within**: An optional number of hours within which the
      upgrade task should be delivered to computers. The window will be
      from the time specified by this API call (on_days, at_hour, at_minute)
      until the provided number of hours later. Defaults to 1 hour.
    - **deliver_delay_window**: Randomise delivery within the given timeframe
      specified in minutes. (optional)
    - **security_upgrade**: Optional, defaults to False. Whether this upgrade
      is a security upgrade or not. Deprecated, use upgrade_type instead.
    - **upgrade_type**: This profile should apply to all package upgrades
      or only security upgrades.  Must be either "security" or "all".
    - **autoremove**: Optional, defaults to False. Whether the upgrade profile
      should also perform an autoremove of old packages.

    - **access_group**: An optional name of the access group to create the
      profile into.

    The following errors may be raised:

    - **InvalidScheduleFormat**: The given schedule string is not valid.
    - **InvalidParameterCombination**: The set of arguments are not
      compatible when specified together.
    - **UnknownAccessGroup**: The given access group is not known.
    - **Unauthorised**: The person is not authorised to copy upgrade profiles
      in the given access group.

    For example, the following request creates an upgrade profile schedule that
    updates all packages every Monday and Sunday at 1:30pm::

        ?action=CreateUpgradeProfile&name=test1&every=week&on_days.1=mo
        &on_days.2=su&at_hour=13&at_minute=30

    The method returns a JSON serialized representation of the newly created
    object::

        [{"id": 1,
          "title": "Test 1",
          "name": "test-1",
          "tags": [],  # This is set via AssociateUpgradeProfile
          "all_computers": False,
          "upgrade_type": "all",       # This can alternatively be "security"
          "autoremove": False,
          "deliver_within": 1,         # hours before activity expiration
          "deliver_delay_window": 0,   # minutes for random delivery timeframe
          "every": "week",
          "at_minute": 30,
          "at_hour": 13,
          "on_days": ["mo", "su"]
         }]
    """

    requires = "ViewAccount"

    schema = Schema(
        name="CreateUpgradeProfile",
        parameters=[
            UnicodeTitle(
                "title", doc="A human readable title for this upgrade profile."
            ),
            Unicode(
                "every",
                validator=_validate_frequency,
                doc=(
                    "The frequency at which you wish this upgrade "
                    'profile to be executed. Valid choices are "hour"'
                    ' and "week".'
                ),
            ),
            List(
                "on_days",
                Unicode(),
                optional=True,
                doc=(
                    "A list of days of the week on which the "
                    "upgrade profile will be run. The day names "
                    "must be abbreviated to their first two letters, "
                    'as: "mo", "tu", "we", "th", "fr", "sa", '
                    '"su". Required when the every parameter is '
                    '"week" but optional when the every parameter is '
                    ' "hour".'
                ),
            ),
            Integer(
                "at_hour",
                optional=True,
                min=0,
                max=23,
                doc=(
                    "The hour, in 24h format, at which the upgrade "
                    "profile will be run."
                ),
            ),
            Integer(
                "at_minute",
                min=0,
                max=59,
                doc=(
                    "The minute of the hour (0-59) at which the "
                    "upgrade profile will be run."
                ),
            ),
            Integer(
                "deliver_within",
                optional=True,
                default=1,
                doc=(
                    "An optional number of hours within which the "
                    "upgrade task should be delivered to computers. "
                    "The window will be from the time specified by this "
                    "API call (on_days, at_hour, at_minute) until the "
                    "provided number of hours later. Defaults to 1 "
                    "hour."
                ),
            ),
            Integer(
                "deliver_delay_window",
                optional=True,
                default=0,
                doc=(
                    "Randomise delivery within the given timeframe "
                    "specified in minutes."
                ),
            ),
            Bool(
                "security_upgrade",
                default=False,
                optional=True,
                doc="(Deprecated) Whether this upgrade is a security "
                "upgrade or not.",
            ),
            Unicode(
                "upgrade_type",
                optional=True,
                validator=_validate_upgrade_type,
                doc=('The type of upgrade profile, either "security" ' 'or "all".'),
            ),
            Bool(
                "autoremove",
                default=False,
                optional=True,
                doc=("Whether this upgrade should also autoremove " "old packages."),
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc="An optional name of the access group to create the "
                "profile into.",
            ),
        ],
        errors=[
            ScheduleFormatException,
            UnknownAccessContextError,
            MissingParameterError,
            Unauthorized,
            InvalidParameterCombinationError,
            InvalidScheduleAfterWindowError,
        ],
        doc="""Create an upgrade profile.""",
    )

    route = ManagementRoute(factory=UpgradeProfileManagement)

    def query(self, context, call):
        """Creates the UpgradeProfile in the database"""
        schedule = arguments_to_ical(call.args)

        if (
            call.args.upgrade_type is not None
            and "security_upgrade" in call.get_raw_params()
        ):
            raise InvalidParameterCombinationError(
                "Only one of security_upgrade (deprecated) or upgrade_type "
                "must be included."
            )

        if call.args.upgrade_type is not None:
            upgrade_type = call.args.upgrade_type
        else:
            upgrade_type = get_upgrade_type(call.args.security_upgrade)

        profile = context.create_upgrade_profile(
            call.args.title,
            upgrade_type,
            schedule,
            deliver_within=call.args.deliver_within,
            access_context_name=call.args.access_group,
            deliver_after_window=call.args.deliver_delay_window,
            autoremove=call.args.autoremove,
        )
        return profile.get_state()


@method
class EditUpgradeProfile(AccountMethod):
    """
    Edit an upgrade profile.

    Arguments:

    - **name**: The name for this upgrade profile.
    - **title**: Optional, the new title for this upgrade profile.
    - **every**: The frequency at which you wish this upgrade profile to be
      executed. Valid choices are "hour" and "week".
    - **on_days**: A list of days of the week on which the upgrade profile
      will be run. The day names must be abbreviated to their first two
      letters, as: "mo", "tu", "we", "th", "fr", "sa", "su". Only needed in
      case the every parameter is "week" and optional when the every parameter
      is "hour".
    - **at_hour**: The hour, in 24h format, at which the upgrade profile
      will be run.
    - **at_minute**: The minute of the hour (0-59) at which the upgrade
      profile will be run.
    - **deliver_within**: An optional number of hours within which the
      upgrade task should be delivered to computers. The window will be
      from the time specified by this API call (on_days, at_hour, at_minute)
      until the provided number of hours later. Defaults to 1 hour.
    - **deliver_delay_window**: Randomise delivery within the given timeframe
      specified in minutes. (optional)
    - **security_upgrade**: Optional, defaults to False. Whether this profile
      will target only security upgrades (instead of all upgrades).
      Deprecated, use upgrade_type instead.
    - **upgrade_type**: This profile should apply to all package upgrades
      or only security upgrades.  Must be either "security" or "all".
    - **autoremove**: Optional, defaults to False. Whether the upgrade profile
      should also perform an autoremove of old packages.

    The following errors may be raised:

    - **UnknownUpgradeProfile**: A profile with the specified name could not be
      found in the database.
    - **InvalidScheduleFormat**: The given schedule string is not valid.
    - **InvalidParameterCombination**: The set of arguments are not
      compatible when specified together.

    For example, the following is a request to update the example created in
    CreateUpgradeProfile with an upgrade schedule to update all packages on an
    hourly basis instead, at 35 past the hour::

        ?action=EditUpgradeProfile&name=test1&every=hour&on_minute=35

    The method returns a JSON serialized representation of the modified
    object::

        [{"id": 1,
          "name": "test1",
          "tags": [],  # This is set via AssociateUpgradeProfile
          "all_computers": False,
          "upgrade_type": "all",
          "autoremove": False,
          "deliver_within": 1,         # hours before activity expiration
          "deliver_delay_window": 0,   # minutes for random delivery timeframe
          "every": "hour",
          "at_minute": 35,
         }]

    """

    requires = "ViewAccount"

    schema = Schema(
        name="EditUpgradeProfile",
        parameters=[
            Unicode("name", doc="The name for this upgrade profile."),
            UnicodeTitle(
                "title", doc="The new title of the upgrade profile.", optional=True
            ),
            Unicode(
                "every",
                optional=True,
                validator=_validate_frequency,
                doc=(
                    "The frequency at which you wish this upgrade "
                    'profile to be executed. Valid choices are "hour" '
                    'and "week".'
                ),
            ),
            List(
                "on_days",
                Unicode(),
                optional=True,
                doc=(
                    "A list of days of the week on which the "
                    "upgrade profile will be run. The day names "
                    "must be abbreviated to their first two letters, "
                    'as: "mo", "tu", "we", "th", "fr", "sa", '
                    '"su". Required when the every parameter is '
                    '"week" but optional when the every parameter is '
                    ' "hour".'
                ),
            ),
            Integer(
                "at_hour",
                optional=True,
                min=0,
                max=23,
                doc=(
                    "The hour, in 24h format, at which the upgrade "
                    "profile will be run."
                ),
            ),
            Integer(
                "at_minute",
                optional=True,
                min=0,
                max=59,
                doc=(
                    "The minute of the hour (0-59) at which the "
                    "upgrade profile will be run."
                ),
            ),
            Integer(
                "deliver_within",
                optional=True,
                default=1,
                doc=(
                    "An optional number of hours within which the "
                    "upgrade task should be delivered to computers. "
                    "The window will be from the time specified by this "
                    "API call (on_days, at_hour, at_minute) until the "
                    "provided number of hours later. Defaults to 1 "
                    "hour."
                ),
            ),
            Integer(
                "deliver_delay_window",
                optional=True,
                doc=(
                    "Randomise delivery within the given timeframe "
                    "specified in minutes."
                ),
            ),
            Bool(
                "security_upgrade",
                default=False,
                optional=True,
                doc="(Deprecated) Whether this upgrade is a security "
                "upgrade or not.",
            ),
            Unicode(
                "upgrade_type",
                optional=True,
                validator=_validate_upgrade_type,
                doc=('The type of upgrade profile, either "security" ' 'or "all".'),
            ),
            Bool(
                "autoremove",
                default=None,
                optional=True,
                doc=("Whether this upgrade should also autoremove " "old packages."),
            ),
        ],
        errors=[
            UnknownUpgradeProfileError,
            ScheduleFormatException,
            InvalidParameterCombinationError,
            MissingParameterError,
            InvalidScheduleAfterWindowError,
        ],
        doc="""Edit an upgrade profile.""",
    )

    route = ManagementRoute(factory=UpgradeProfileManagement)

    def query(self, context, call):
        """Updates the UpgradeProfile in the database."""
        args = call.args
        missing_arguments = (
            args.title is None
            and args.deliver_delay_window is None
            and args.upgrade_type is None
            and args.autoremove is None
            and None in (args.every, args.at_minute)
        )
        if missing_arguments:
            raise InvalidParameterCombinationError(
                "Either title, deliver_delay_window, upgrade_type, autoremove "
                "or every and at_minute must be included."
            )

        if (
            args.upgrade_type is not None
            and "security_upgrade" in call.get_raw_params()
        ):
            raise InvalidParameterCombinationError(
                "Only one of security_upgrade (deprecated) or upgrade_type "
                "must be included."
            )

        security_upgrade = args.security_upgrade
        if "security_upgrade" not in call.get_raw_params():
            # The user did not specify this param, it's just the default
            security_upgrade = None
        if args.upgrade_type is not None:
            # True if "security", False otherwise
            security_upgrade = UPGRADE_TYPE_MAP.get(args.upgrade_type) == 1

        # Don't edit the schedule if not provided
        schedule = arguments_to_ical(args) if args.every else None

        profile = context.edit(
            args.name,
            title=args.title,
            schedule=schedule,
            security_upgrade=security_upgrade,
            deliver_within=args.deliver_within,
            deliver_after_window=args.deliver_delay_window,
            autoremove=args.autoremove,
        )
        return profile.get_state()


@method
class RemoveUpgradeProfile(AccountMethod):
    """
    Remove an existing upgrade profile by name.

    Arguments:

    - **name**: The name of the upgrade profile you wish to cancel.

    The following errors may be raised:

    - **UnknownUpgradeProfile**: A profile with the specified name could not be
      found in the database.

    For example, the following request deletes an upgrade profile schedule with
    the name "test-1"::

        ?action=RemoveUpgradeProfile&name=test-1
    """

    requires = "EditUpgradeProfile"

    schema = Schema(
        name="RemoveUpgradeProfile",
        parameters=[
            Unicode(
                "name", doc=("The name of the upgrade profile you wish to " "cancel.")
            )
        ],
        errors=[UnknownUpgradeProfileError],
        doc="""Remove an existing upgrade profile by name.""",
    )

    route = SubjectRoute(
        factory=UpgradeProfileManagement,
        getter="get_upgrade_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        profile = context
        management = self.get_management(UpgradeProfileManagement, call)
        management.remove_upgrade_profiles([profile])
        return {}


@method
class GetUpgradeProfiles(AccountMethod):
    """
    List all previously created upgrade profiles.

    Arguments:

    - **upgrade_type**: Optional. The type of upgrade you wish to list. This
      can be either "all" or "security", in which case the result will be a
      list of upgrade profiles with an upgrade type of "all" or "security"
      respectively. If omitted, the resulting list will contain all
      upgrade profiles, regardless of their upgrade type.

    The following errors may be raised:

    - **InvalidParameterValue**: If upgrade_type is not "all" or "security".

    For example, the following request lists all upgrade profile schedules::

        ?action=GetUpgradeProfiles

    The method returns a JSON serialized representation of the list::

        [{"id": 1,
          "name": "test1",
          "schedule": "FREQ=HOURLY;BYMINUTE=30",
          "tags": ["my-computers", "lucid", "server"],
          "all_computers": False,
          "upgrade_type": "all"}
         {"id": 2,
          "name": "test2",
          "schedule": "FREQ=WEEKLY",
          "tags": ["my-computers", "lucid", "server"],
          "all_computers": False,
          "upgrade_type": "security"}]
    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetUpgradeProfiles",
        parameters=[
            Unicode(
                "upgrade_type",
                optional=True,
                validator=_validate_upgrade_type,
                doc=(
                    " The type of upgrade you wish to list. This can be "
                    'either "all" or "security", in which case the '
                    "result will be a list of upgrade profiles with "
                    'an upgrade type of "all" or "security" '
                    "respectively. If omitted, the resulting list will "
                    "contain all upgrade profiles, regardless of "
                    "their upgrade type."
                ),
            )
        ],
        errors=[],
        doc="""List all previously created upgrade profiles.""",
    )

    route = ManagementRoute(factory=UpgradeProfileManagement)

    def query(self, context, call):
        profiles = context.get_upgrade_profiles(upgrade_type=call.args.upgrade_type)
        return [profile.get_state() for profile in profiles]


@method
class DisassociateUpgradeProfile(TagAssociationMethod):
    """
    Disassociate an upgrade profile from computers with the specified
    tags, or from all computers.

    Arguments:

    - **name**: Name of the upgrade profile.
    - **tags.#**: A list of tag names to disassociate from the profile.
    - **all_computers**: if true, the profile will only remain enabled for
      computers with tags associated to the profile. This parameter is
      *optional* and defaults to *false*.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=DisassociateUpgradeProfile&name=test-1
            &tags.1=server&tags.2=lucid

    The following errors may be raised:

    - **UnknownUpgradeProfile**: No profile with the specified name exists.
    - **InvalidParameterCombination**: The set of arguments are not
      compatible when specified together.

    The method returns JSON serialized info of the profile status::

        {"name": "test-1",
         "id": 178,
         "upgrade_type": "all",
         "schedule": "FREQ=WEEKLY",
         "tags": ["my-computers"],
         "all_computers": False}
    """

    requires = "EditUpgradeProfile"

    associate = False

    schema = TagAssociationMethod.schema.extend(
        name="DisassociateUpgradeProfile",
        errors=[UnknownUpgradeProfileError],
        doc="""
    Disassociate an upgrade profile from computers with the specified
    tags, or from all computers.

    tags and all_computers=true are mutually exclusive.""",
    )

    route = SubjectRoute(
        factory=UpgradeProfileManagement,
        getter="get_upgrade_profile_by_name",
        identifier_arg="name",
    )


@method
class AssociateUpgradeProfile(TagAssociationMethod):
    """
    Associate an upgrade profile to computers with the specified tags,
    or all computers.

    Arguments:

    - **name**: Name of the upgrade profile.
    - **tags.#**: A list of tag names to associate to the profile.
    - **all_computers**: true if the profile should be associated to all
      computers. This parameter is *optional* and defaults to *false*.
      Individual tags associated to the profile will remain, but they will
      only be effective if the "all_computers" flag is later disabled.

    tags.# and all_computers=true are mutually exclusive.

    Example of a valid request::

        ?action=AssociateUpgradeProfile&name=test-1
            &tags.1=server&tags.2=lucid

    The following errors may be raised:

    - **UnknownUpgradeProfile**: No profile with the specified name exists.
    - **InvalidParameterCombination**: The set of arguments are not
      compatible when specified together.

    The method returns JSON serialized info of the profile status::

        {"name": "test-1",
         "id": 178,
         "upgrade_type": "all",
         "schedule": "FREQ=WEEKLY",
         "tags": ["my-computers", "lucid", "server"],
         "all_computers": False}
    """

    requires = "EditUpgradeProfile"

    schema = TagAssociationMethod.schema.extend(
        name="AssociateUpgradeProfile",
        errors=[UnknownUpgradeProfileError],
        doc="""
    Associate an upgrade profile to computers with the specified tags,
    or all computers.

    tags and all_computers=true are mutually exclusive.""",
    )

    route = SubjectRoute(
        factory=UpgradeProfileManagement,
        getter="get_upgrade_profile_by_name",
        identifier_arg="name",
    )
