# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.txapi.server.method import method
from canonical.txapi.server.schema import Schema, Unicode, Float, Integer, Date

from canonical.landscape.api.methods import Method
from canonical.landscape.api.routes import GlobalRoute
from canonical.landscape.service.billing.cost import (
    GetCurrentUtilityCost,
    UpdateUtilityCost,
)
from canonical.landscape.service.billing.vatchange import ChangeCountryVATRate
from canonical.landscape.service.billing.config import (
    SetInvoiceStorageCredentials,
    SetMarketoCredentials,
)
from canonical.landscape.service.billing.config import SetStripeAPIKeys
from canonical.landscape.model.billing.vat import get_country_vat_rates


@method
class GetCountryVatRates(Method):
    """Return VAT rates for each country."""

    hidden = True

    schema = Schema(
        name="GetCountryVatRates",
        doc="Return a map of country codes and VAT rates for those countries.",
    )

    route = GlobalRoute()

    def query(self, context, call):
        return get_country_vat_rates()


@method
class ChangeCountryVatRate(Method):
    """Change the VAT rate for a country.

    @param country: the 2-letter ISO-3166-1 code for the country.
    @param rate: the new VAT percentage value."""

    hidden = True

    route = GlobalRoute()

    schema = Schema(
        name="ChangeCountryVatRate",
        parameters=[
            Unicode("country", doc="The country two-letter ISO code."),
            Float(
                "rate",
                doc=(
                    "The new VAT rate for the country. "
                    "It must be a percentage expressed as a number "
                    "between 0 and 1"
                ),
            ),
        ],
        doc="Change VAT rate for a country.",
        errors=ChangeCountryVATRate.errors(),
    )

    def query(self, context, call):
        operation = ChangeCountryVATRate(context, call.principal)
        operation(call.args.country, call.args.rate)
        return {}


@method
class GetUtilityCost(Method):
    """Get the current utility cost for a type of machine.

    @param computer_type: the type of computer the rate applies to.
    """

    hidden = True

    route = GlobalRoute()

    schema = Schema(
        name="GetUtilityCost",
        parameters=[
            Unicode("computer_type", doc="The type of computer to return cost for")
        ],
        doc="Return utility cost for a computer type.",
        errors=GetCurrentUtilityCost.errors(),
    )

    def query(self, context, call):
        operation = GetCurrentUtilityCost(context, call.principal)
        return operation(call.args.computer_type)


@method
class ChangeUtilityCost(Method):
    """Change the utility cost for a type of machine.

    @param computer_type: the type of computer the rate applies to.
    @param valid_since: the start validity date for the rate.
    @param hourly_cost: the management cost per hour, in 100th of cents.
    """

    hidden = True

    route = GlobalRoute()

    schema = Schema(
        name="ChangeUtilityCost",
        parameters=[
            Unicode("computer_type", doc="The computer type to change cost for."),
            Date("valid_since", doc="The validity start date for the new rate."),
            Integer("hourly_cost", doc="The hourly cost."),
        ],
        doc="Change utility cost for a computer type",
        errors=UpdateUtilityCost.errors(),
    )

    def query(self, context, call):
        operation = UpdateUtilityCost(context, call.principal)
        operation(call.args.computer_type, call.args.valid_since, call.args.hourly_cost)
        return {}


@method
class SetStripeApiKeys(Method):
    """Set the public and secret API keys for Stripe

    @param public_key: the publishable Stripe API key
    @param secret_key: the secret Stripe API key."""

    hidden = True

    route = GlobalRoute()

    schema = Schema(
        name="SetStripeApiKeys",
        parameters=[
            Unicode("public_key", doc="The Stripe Publishable key."),
            Unicode("secret_key", doc="The Stripe Secret key."),
        ],
        doc="Set Stripe API keys",
        errors=SetStripeAPIKeys.errors(),
    )

    def query(self, context, call):
        operation = SetStripeAPIKeys(context, call.principal)
        operation(call.args.public_key, call.args.secret_key)
        return {}


@method
class SetBillingOpenstackCredentials(Method):
    """Set OpenStack credentials to store billing receipts.

    @param keystone_url: URL for the keystone endpoint
    @param username: The username of the designated cloud user
    @param password: The password of the designated cloud user
    @param tenant: The openstack tenant name where invoices will be stored.
    """

    hidden = True

    route = GlobalRoute()

    schema = Schema(
        name="SetBillingOpenstackCredentials",
        parameters=[
            Unicode("keystone_url", doc="The Keystone endpoint."),
            Unicode("username", doc="The username of the cloud user."),
            Unicode("password", doc="The password of the cloud user."),
            Unicode("tenant", doc="The OpenStack tenant name."),
        ],
        doc="Set OpenStack credentials for receipts upload.",
        errors=SetInvoiceStorageCredentials.errors(),
    )

    def query(self, context, call):
        operation = SetInvoiceStorageCredentials(context, call.principal)
        return operation(
            call.args.keystone_url,
            call.args.username,
            call.args.password,
            call.args.tenant,
        )


@method
class SetMarketingCredentials(Method):
    """Set the credentials for pushing self-service users to marketing.

    @param url: the base url of the marketo account
    @param client_id: the marketo API client ID
    @param client_secret: the marketo API secret key
    """

    hidden = True

    route = GlobalRoute()

    schema = Schema(
        name="SetMarketingCredentials",
        parameters=[
            Unicode("url", doc="the Marketo base url"),
            Unicode("client_id", doc="the Marketo client_id"),
            Unicode("client_secret", doc="the Marketo client_secret"),
        ],
        doc="Set the marketo credentials.",
        errors=SetMarketoCredentials.errors(),
    )

    def query(self, context, call):
        operation = SetMarketoCredentials(context, call.principal)
        return operation(call.args.url, call.args.client_id, call.args.client_secret)
