# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Computers
=========

The methods available here give the ability to retrieve computers and to do
basic operations on them, like tagging.
"""
from storm.expr import And, Join

from canonical.txapi.server.call import Call
from canonical.txapi.server.schema import Schema, Unicode, Integer, Bool, List
from canonical.txapi.server.method import method

from landscape.lib.tag import is_valid_tag

from canonical.lib.text import make_secure_id
from canonical.lib.timestamp import parse_datetime

from canonical.landscape.api.error import (
    InvalidTagError,
    InsufficientLicensesError,
    UnknownAccessGroupError,
    UnknownComputerError,
    InvalidQueryError,
    UnauthorisedError,
    InvalidTimeError,
    InvalidTitleError,
    NotWSLHostError,
)
from canonical.landscape.api.methods import search_schema, AccountMethod
from canonical.landscape.api.routes import AccountRoute, SelectionRoute
from canonical.landscape.api.schema import Mapping
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.computer import (
    Computer,
    ComputerRelationship,
    PendingComputerToAccept,
)
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.distribution import Distribution
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.tag import get_tags
from canonical.landscape.model.main.license import LicenseSeatError
from canonical.landscape.model.main.errors import AcceptanceUnauthorized
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.computer.collection import AccountComputerTwinCollection
from canonical.landscape.model.computer.factory import WSL_HOST_CODE_NAME
from canonical.landscape.model.computer.search import ComputerSearch
from canonical.landscape.model.cloud.instance import CloudRegistrationOTP
from canonical.landscape.model.resource.network import AccountNetworkInfo
from canonical.landscape.model.resource.hardware import HardwareInfoManagement
from canonical.landscape.model.resource.annotation import (
    AnnotationManagement,
    InvalidAnnotationKeyError,
)
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.service.computer.remove import (
    RemoveComputers as RemoveComputersOperation,
)
from canonical.landscape.service.computer.accesscontextchange import (
    ComputerAccessContextChange,
)
from canonical.landscape.service.security import PersonPrincipal
from canonical.landscape.service.collection.errors import UnknownEntityIDsError

from canonical.landscape.model.main.computer import (
    add_tags_to_computers,
    remove_tags_from_computers,
)


__all__ = [
    "GetComputers",
    "AddAnnotationToComputers",
    "RemoveAnnotationFromComputers",
    "AddTagsToComputers",
    "RemoveTagsFromComputers",
    "ChangeComputersAccessGroup",
    "RemoveComputers",
    "GetPendingComputers",
    "AcceptPendingComputers",
    "RejectPendingComputers",
    "CreateCloudOtps",
    "RebootComputers",
    "ShutdownComputers",
    "RenameComputers",
    "GetWSLHosts",
    "StartChildComputers",
    "StopChildComputers",
]


@method
class GetComputers(AccountMethod):
    """
    Get a list of computers associated with the account used for
    authentication.

    .. _computer-query:

    - **query**: A query string with space separated tokens used to filter the
      returned computers. Words provided as search parameters are treated as
      keywords, matching the hostname or the computer title. Selector prefixes
      can be used to further customize the search:

    - *tag*: search for computers with the specified tag.
    - *distribution*: search for computers running a specific Ubuntu
      release (can be code name like 'lucid' or version number like
      '10.04').
    - *hostname*: search for computers with the exact hostname.
    - *title*: search for computers with the exact title.
    - *alert*: search for computers with a named alert being active, alerts
      can be one of "package-upgrades", "security-upgrades",
      "package-profiles", "package-reporter",  "computer-offline",
      "computer-reboot".
    - *access-group*: search for computers that belong to the access group with
      the specified name.
    - *id*: select the specified numeric computer id.
    - *mac*: search for computers with the specified MAC address, which can
      be a substring of the full address.
    - *ip*: search for computers with the specified IP address, which can be
      a substring of the full address. No network classing is done.
    - *search*: select computers based on the result of the named search.
    - *needs:reboot*: search for computers that have the reboot required flag
      set.  Note that with this particular criteria, the only possible value
      for it is the text "reboot".
    - *license-id*: search for computers licensed to the specified license-id.
    - *needs:license* OR *license-id:none*: search for computers that do not
      have a Landscape license, and, as a result, are not managed.
    - *annotation*: search for computers which define the specified annotation
      key. Optionally specify annotation:<key>:<string> which will only return
      computers whose key matches and value also contains the <string>
      specified.
    - *OR*: search for computers matching term A or term B. OR must be in
      all caps.
    - *NOT*: search for computers not matching the next term. NOT must be in
      all caps.

    If values following a prefix contains spaces or non-ASCII characters, it
    must be quoted.

    These prefixes can be mixed and matched with keywords.  For example, the
    following query would match computers with 'appserver' in their title or
    hostname, either with the tag 'server' or running the 'hardy' release of
    Ubuntu::

      appserver tag:server OR distribution:hardy

    It also supports the following optional arguments:

    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.
    - **with_network**: If true, include the details of all network devices
      attached to the computer.
    - **with_hardware**: If true, include the details of all hardware
      information known.
    - **with_annotations**: If true, include the details of all custom
      annotation information known.

    For example, the following request looks for all computers with the tag
    'server', includes their network devices, and with a limit of 20 machines::

        ?action=GetComputers&query=tag:server&limit=20&with_network=true

    The method returns a JSON serialized list of computers::

         [{"access_group": "global",
           "comment": u"",
           "distribution": u"12.04",
           "hostname": u"a_comp.example.com",
           "id": 12345,
           "last_ping_time": None,
           "last_exchange_time": "2011-06-3017:59Z",
           "title": u"A Computer",
           "reboot_required_flag": False,
           "tags": ["server"],
           "total_memory": None,
           "total_swap": None,
           "network_devices": [{"broadcast_address": "192.168.1.255",
                                "interface": "eth0",
                                "ip_address": "192.168.1.2",
                                "mac_address": "00:1e:c9:6c:b8:de",
                                "netmask": "255.255.255.0"}]}]

    """

    requires = "ViewAccount"

    schema = search_schema.extend(
        name="GetComputers",
        parameters=[
            Bool(
                "with_network",
                optional=True,
                default=False,
                doc=(
                    "If true, include the details of all network devices "
                    "attached to the computer."
                ),
            ),
            Bool(
                "with_hardware",
                optional=True,
                default=False,
                doc=(
                    "If true, include the details of all hardware " "information known."
                ),
            ),
            Bool(
                "with_annotations",
                optional=True,
                default=False,
                doc=(
                    "If true, include the details of all custom annotation "
                    "information known."
                ),
            ),
        ],
        doc="""
    Get a list of computers associated with the account.
    """,
    )

    route = SelectionRoute(permission="ViewComputer", error_if_empty=False)

    def query(self, context, call):
        """
        Perform a query to get L{Computer}s associated with the
        L{Account}, using the account attached to the credentials making
        the call.
        """
        computer_data = dict(
            (computer.id, computer.get_state()) for computer in context.get_selected()
        )
        computer_ids = list(computer_data.keys())
        if call.args.with_network:
            info = AccountNetworkInfo(context.account)
            computer_devices = info.get_devices_for_computer_ids(computer_ids)
            for computer_id, devices in list(computer_devices.items()):
                computer_data[computer_id]["network_devices"] = sorted(
                    (device.get_state() for device in devices),
                    key=lambda x: x["interface"],
                )
        if call.args.with_hardware:
            management = HardwareInfoManagement(context.account)
            hardware = management.get_hardware_for_computer_ids(computer_ids)
            for computer_id, data in list(hardware.items()):
                computer_data[computer_id]["hardware"] = data
        if call.args.with_annotations:
            person = get_person_by_id(call.principal.id)
            management = AnnotationManagement(context.account, person)
            annotations = management.get_annotation(computer_ids)
            for computer_id, data in list(annotations.items()):
                computer_data[computer_id]["annotations"] = data
        return list(computer_data.values())


@method
class AddAnnotationToComputers(AccountMethod):
    """
    Add a custom annotation to a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers for which to add
      annotations.
      See :ref:`Computer Queries <computer-query>` for details.

    - **key**: Annotation key to add to the selected computers.
    - **value**: Annotation value associated with the provided key to add
      to the selected computers. (optional)

    For example, the following request will add a custom annotation
    representing physical to all computers running Ubuntu 10.04 (Lucid)::

        ?action=AddAnnotationToComputers&query=distribution:10.04&key=location
            &value=BLDG3:FLR2:CAGE101A
    """

    requires = "ViewAccount"

    route = SelectionRoute(permission="EditComputerDetails", error_if_empty=False)

    schema = Schema(
        name="AddAnnotationToComputers",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "which to add the annotation."
                ),
            ),
            Unicode("key", doc=("Annotation key to add to the selected computers.")),
            Unicode(
                "value",
                optional=True,
                doc=(
                    "Annotation value associated with the provided key to "
                    "add to the selected computers."
                ),
            ),
        ],
        errors=[InvalidQueryError, InvalidAnnotationKeyError],
        doc="""
    Add annotation key and optional value to a selection of computers.""",
    )

    def query(self, context, call):
        """
        Find all computers matching query and add the specified
        annotation for these computers.
        """
        person = get_person_by_id(call.principal.id)
        management = AnnotationManagement(context.account, person)
        annotation = {call.args.key: call.args.value}
        management.add_annotation(
            [computer.id for computer in context.get_selected()], annotation
        )
        return {}


@method
class RemoveAnnotationFromComputers(AccountMethod):
    """
    Remove a custom annotation with the specified key from a selection of
    computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to remove
      annotations from.
      See :ref:`Computer queries <computer-query>` for details.

    - **key**: The annotation key to remove.

    For example, the following request will remove the annotation key
    "location" from all computers with tag *server*::

        ?action=RemoveAnnotationFromComputers&query=tag:server&key=location
    """

    requires = "ViewAccount"

    route = SelectionRoute(permission="EditComputerDetails", error_if_empty=False)

    schema = Schema(
        name="RemoveAnnotationFromComputers",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers from "
                    "which to remove annotation."
                ),
            ),
            Unicode("key", doc="Annotation key to disassociate."),
        ],
        errors=[InvalidQueryError, InvalidAnnotationKeyError],
        doc="""Remove annotation key from a selection of computers.""",
    )

    def query(self, context, call):
        """
        Find all computers matching query and remove the specified
        annotation key for these computers.
        """
        person = get_person_by_id(call.principal.id)
        management = AnnotationManagement(context.account, person)
        management.remove_annotation(
            [computer.id for computer in context.get_selected()], [call.args.key]
        )
        return {}


class ComputerTagHandler(AccountMethod):

    requires = "ViewAccount"

    route = SelectionRoute(permission="EditTags", error_if_empty=False)

    def query(self, context, call):
        """
        Find all computers matching query and add the specified tags to the
        computers.
        """
        person = get_person_by_id(call.principal.id)
        for tag in call.args.tags:
            if not is_valid_tag(tag):
                raise InvalidTagError("Tag '%s' is invalid." % tag)
        self.method.__func__(
            person, get_tags(call.args.tags), list(context.get_selected())
        )
        return {}


@method
class AddTagsToComputers(ComputerTagHandler):
    """
    Add tags to a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to add tags to.
      See :ref:`Computer Queries <computer-query>` for details.

    - **tags**: Tag name to be applied, this can have more than one Tag, by
      numbering the tags with tags.1, tags.2, tags.3 etc.

    For example, the following request will add tags "server" and "lucid" to
    all computers running Ubuntu 10.04 (Lucid)::

        ?action=AddTagsToComputers&query=distribution:10.04&tags.1=server
            &tags.2=lucid
    """

    schema = Schema(
        name="AddTagsToComputers",
        parameters=[
            Unicode(
                "query",
                doc=("A query string used to select the computers to " "add tags to."),
            ),
            List("tags", Unicode(), doc="Tag names to be applied."),
        ],
        errors=[InvalidQueryError, InvalidTagError],
        doc="""Add tags to a selection of computers.""",
    )

    method = add_tags_to_computers


@method
class RemoveTagsFromComputers(ComputerTagHandler):
    """
    Remove tags from a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to remove tags
      from. See :ref:`Computer queries <computer-query>` for details.

    - **tags**: Tag name to be remove, this can have more than one Tag, by
      numbering the tags with tags.1, tags.2, tags.3 etc.

    For example, the following request will remove tags "server" and "lucid"
    from all computers with tag *server*::

        ?action=RemoveTagsFromComputers&query=tag:server&tags.1=server
            &tags.2=lucid
    """

    schema = Schema(
        name="RemoveTagsFromComputers",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "remove tags from."
                ),
            ),
            List("tags", Unicode(), doc="Tag names to be removed."),
        ],
        errors=[InvalidTagError, InvalidQueryError],
        doc="""Remove tags from a selection of computers.""",
    )
    method = remove_tags_from_computers


@method
class ChangeComputersAccessGroup(AccountMethod):
    """
    Change the access group for a selection of computers.

    This method takes two mandatory arguments:

    - **query**: A query string used to select the computers to change access
      group for.  See :ref:`Computer queries <computer-query>` for details.
    - **access_group**: The name of the access group to assign selected
      computers to.

    This is an example of a valid request::

         action=ChangeComputersAccessGroup&query=tag:new-servers
            &access_group=server

    The method returns a JSON serialized list of computers in the selection
    which have successfully changed access group::

         [{"access_group": "server",
           "id": 12345,
           "title": u"A Computer",
           "comment": u"",
           "total_memory": None,
           "total_swap": None,
           "reboot_required_flag": False,
           "hostname": u"a_comp.example.com",
           "last_ping_time": None,
           "last_exchange_time": "2011-06-3017:59Z",
           "tags": ["server"],
           "network_devices": [{"broadcast_address": "192.168.1.255",
                                "interface": "eth0",
                                "ip_address": "192.168.1.2",
                                "mac_address": "00:1e:c9:6c:b8:de",
                                "netmask": "255.255.255.0"}]}]

    The following errors may be raised:

    - **InvalidQueryError**: If query format is invalid.
    """

    schema = Schema(
        name="ChangeComputersAccessGroup",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "change access group for."
                ),
            ),
            Unicode(
                "access_group",
                doc=(
                    "The name of the access group to assign selected " "computers to."
                ),
            ),
        ],
        errors={InvalidQueryError} | ComputerAccessContextChange.errors(),
        doc="""Change the access group for a selection of computers.""",
    )

    # Rely on the security checks in the service operation
    requires = "ViewAccount"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        account = context.account
        collection = AccountComputerTwinCollection(
            store=account.store,
            account=account,
            ids=context.get_selected_computer_ids(),
        )

        principal = PersonPrincipal(call.principal.id)
        operation = ComputerAccessContextChange(collection, principal)
        # Grab get_selected before we update_access_context. Once the access
        # group change occurs, a subsequent get_selected call will filter by
        # RemoveComputerFromAccessGroup permission. If the person doesn't also
        # have RemoveCompFromAG on the new group, our selection will disappear.
        selected_computers = list(context.get_selected())
        computer_ids = operation(call.args.access_group)
        return [
            computer.get_state()
            for computer in selected_computers
            if computer.id in computer_ids
        ]


@method
class RemoveComputers(AccountMethod):
    """
    Remove a list of computers by ID.

    This method takes a mandatory argument:

    - **computer_ids**: A list of computer ids to remove.

    This is an example of a valid request::

        ?action=RemoveComputers&computer_ids.1=30&computer_ids.2=43
    """

    schema = Schema(
        name="RemoveComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to remove.")
        ],
        errors=(list(RemoveComputersOperation.errors()) + [UnknownComputerError]),
        doc="""Remove a list of computers by ID.""",
    )

    route = AccountRoute()

    def query(self, context, call):
        """
        Remove computers with the specified ids.
        """
        collection = AccountComputerTwinCollection(context.store, account=context)
        collection = collection.with_ids(call.args.computer_ids)
        try:
            RemoveComputersOperation(collection, call.principal)()
        except UnknownEntityIDsError as error:
            raise UnknownComputerError("Unknown computer '{}'".format(error.ids.pop(0)))
        return {}


@method
class GetPendingComputers(AccountMethod):
    """
    Get a list of pending computers associated with the account used for
    authentication::

        ?action=GetPendingComputers

    The method returns a JSON serialized list of pending computers::

         [{"id": 12345,
           "title": "My Server",
           "hostname": "server.london.company.com",
           "creation_time": "2011-06-3017:59Z",
           "vm_info": "xen",
           "client_tags": ["london"], ["server"]}]
    """

    schema = Schema(
        name="GetPendingComputers",
        parameters=[],
        doc="""
    Get a list of pending computers associated with the account used for
    authentication.
    """,
    )

    requires = "ViewPendingComputer"

    route = AccountRoute()

    def query(self, context, call):
        """
        Find all L{PendingComputers} associated with an account.
        """
        account = context  # For readability
        person = get_person_by_id(call.principal.id)
        pending_computers = account.get_pending_computers(person)
        return [computer.get_state() for computer in pending_computers]


@method
class AcceptPendingComputers(AccountMethod):
    """
    Accept a list of pending computers associated with the account used for
    authentication.

    This method takes the following arguments:

    - **computer_ids**: A list of pending computer ids to accept.
    - **access_group**: The name of the access group to accept the computers
      to. If not provided, they will be put into the global access
      group. (optional)

    This is an example of a valid request::

        ?action=AcceptPendingComputers?computer_ids.1=1&computer_ids.2=2

    The method returns a JSON serialized list of accepted computers::

         [{"id": 12345,
           "title": u"A Computer",
           "comment": u"",
           "hostname": u"a_comp.example.com",
           "last_exchange_time": "2011-06-3017:59Z"}]

    To replace existing computers, map the pending IDs to the existing
    ones::

        ?action=AcceptPendingComputers?computer_ids.1=1&computer_ids.2=2
            &existing_ids.2=3

    The following errors may be raised:

    - **InsufficientLicenses**: Insufficient licenses available to accept new
      computers.
    - **UnknownAccessGroup**: The access group is not known or the person
      is not authorized to accept pending computers into the access group.
    """

    schema = Schema(
        name="AcceptPendingComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer IDs to accept."),
            Mapping(
                "existing_ids",
                Integer("pending", doc="The ID of the pending computer."),
                Integer("existing", doc="The ID of the computer to replace."),
                optional=True,
                doc=("A mapping from pending computer IDs " "to existing ones."),
            ),
            Unicode(
                "access_group",
                optional=True,
                default=None,
                doc="The access group to put the computers into",
            ),
        ],
        errors=[
            InsufficientLicensesError,
            UnknownComputerError,
            UnknownAccessGroupError,
        ],
        doc="""
    Accept a list of pending computers associated with the account used for
    authentication.""",
    )

    # We rely on the security checks in PendingComputer.accept to keep out the
    # baddies. Oh dear!
    requires = "ViewAccount"

    route = AccountRoute()

    def query(self, context, call):
        """
        Accept L{PendingComputer}s associated with an account.
        """
        account = context  # For readability
        person = get_person_by_id(call.principal.id)
        access_group = call.args.access_group
        pending_ids = call.args.computer_ids
        existing_ids = dict(call.args.existing_ids or ())

        extra = set(existing_ids) - set(pending_ids)
        if extra:
            raise UnknownComputerError(
                ("extra pending computer IDs in existing_ids: {}").format(
                    ",".join(str(cid) for cid in sorted(extra))
                )
            )

        pending = self._get_pending_computers(
            account, person, access_group, pending_ids, existing_ids
        )
        _check_computer_ids([p.computer for p in pending], pending_ids)
        _check_computer_ids(
            [p.existing for p in pending],
            sorted(existing_ids.values()),
            msg="Invalid 'existing' computer IDs: {}",
        )

        try:
            computers = account.accept_computers(person, pending)
        except LicenseSeatError:
            raise InsufficientLicensesError("Insufficient licenses to accept computers")
        except AcceptanceUnauthorized as e:
            raise UnauthorisedError(str(e))

        return sorted(
            [computer.get_state() for computer in computers],
            key=lambda x: list(x.values()),
        )

    def _get_pending_computers(
        self, account, person, access_group, computer_ids, existing_ids
    ):
        get_pending_computer = lambda cid: account.get_pending_computer_by_id(
            person, cid
        )
        get_existing_computer = lambda cid: account.get_computer(cid)

        # Make a closure for get_access_context().
        if access_group:
            access_context = self.get_access_context(account, access_group)
            get_access_context = lambda _: access_context
        else:
            default_access_context = self.get_access_context(
                account, DEFAULT_ACCESS_CONTEXT_NAME
            )

            def get_access_context(computer):
                if computer.access_group:
                    return self.get_access_context(account, computer.access_group)
                else:
                    return default_access_context

        # Return the list of PendingComputerToAccept.
        pending = self._iter_pending_computers(
            computer_ids,
            existing_ids,
            get_pending_computer,
            get_existing_computer,
            get_access_context,
        )
        return list(pending)

    def _iter_pending_computers(
        self, computer_ids, existing_ids, get_pending, get_existing, get_access_context
    ):
        for computer_id in computer_ids:
            existing_id = existing_ids.get(computer_id)
            if existing_id is not None:
                existing = get_existing(existing_id)
            else:
                existing = None

            pending_id = computer_id
            pending = get_pending(pending_id)

            if pending is not None:
                access_context = get_access_context(pending)
            else:
                access_context = None

            yield PendingComputerToAccept(pending, existing, access_context)


@method
class RejectPendingComputers(AccountMethod):
    """
    Reject a list of pending computers associated with the account used for
    authentication::

        ?action=RejectPendingComputers?computer_ids.1=1&computer_ids.2=2
    """

    schema = Schema(
        name="RejectPendingComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer IDs to reject.")
        ],
        errors=[UnknownComputerError],
        doc="""
    Reject a list of pending computers associated with the account used for
    authentication.
    """,
    )

    requires = "RejectPendingComputer"

    route = AccountRoute()

    def query(self, context, call):
        """
        Reject L{PendingComputer}s associated with an account.
        """
        account = context  # For readability
        person = get_person_by_id(call.principal.id)

        computer_ids = call.args.computer_ids
        pending_computers = [
            account.get_pending_computer_by_id(person, computer_id)
            for computer_id in computer_ids
        ]
        _check_computer_ids(pending_computers, computer_ids)

        context.reject_computers(person, pending_computers)
        return {}


def _check_computer_ids(computers, requested, msg=None):
    if msg is None:
        msg = "Invalid computer_ids supplied: {}"
    requested = set(cid for cid in requested if cid is not None)

    computer_ids = set(x.id for x in computers if x)
    missing_ids = requested.difference(computer_ids)
    if missing_ids:
        raise UnknownComputerError(msg.format(",".join(map(str, sorted(missing_ids)))))


@method
class CreateCloudOtps(AccountMethod):
    """
    Create one-time passwords used for registration of cloud instances::

        ?action=CreateCloudOtps?count=3

    The method returns a JSON serialized list of one-time passwords, one for
    each requested::

         ["otp1", "otp2", "otp3"]

    You can then use those OTPs in the client configuration, using cloud-init
    for example.
    """

    schema = Schema(
        name="CreateCloudOtps",
        parameters=[
            Integer(
                "count", doc="The number of OTPs to create", optional=True, default=1
            )
        ],
        doc="""
    Create one-time passwords used for registration of cloud instances.
    """,
    )
    requires = "CreateComputer"

    route = AccountRoute()

    def query(self, context, call):
        result = []
        for i in range(call.args.count):
            otp = make_secure_id()
            result.append(otp)
            context.store.add(CloudRegistrationOTP(otp, account=context))
        return result


@method
class RebootComputers(AccountMethod):
    """
    Reboot a list of computers.

    This method takes a mandatory argument:

    - **computer_ids**: A list of computer ids to reboot.

    It also supports an optional argument:

    - **deliver_after**: Reboot the computer after the specified time.  The
      time format is YYYY-MM-DDTHH:MM:SSZ.

    This is an example of a valid request::

        ?action=RebootComputers&computer_ids.1=30&computer_ids.2=43

    The method returns a JSON serialized activity::

        {u'computer_id': None,
         u'creation_time': u'2012-11-19T18:11:51Z',
         u'creator': {u'email': u'john@example.com', u'id': 3,
                      u'name': u'John Smith'},
         u'id': 141,
         u'parent_id': None,
         u'summary': u'Restart computer',
         u'type': u'ActivityGroup'}
    """

    schema = Schema(
        name="RebootComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to reboot."),
            Unicode(
                "deliver_after",
                optional=True,
                doc="A time in the future to deliver the script.",
            ),
        ],
        errors=[InvalidTimeError, UnknownComputerError],
        doc="""Reboot a list of computers.""",
    )

    requires = "RestartComputer"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """
        Reboot computers with the specified ids.
        """
        person = get_person_by_id(call.principal.id)
        management = ComputerManagement(context, person)

        deliver_after = None
        if call.args.deliver_after:
            try:
                deliver_after = parse_datetime(call.args.deliver_after)
            except ValueError:
                raise InvalidTimeError(
                    "Invalid format for deliver_after: '%s'" % call.args.deliver_after
                )

        activity = management.restart(deliver_after=deliver_after)
        return activity.get_state()


@method
class ShutdownComputers(AccountMethod):
    """
    Shut down a list of computers.

    This method takes a mandatory argument:

    - **computer_ids**: A list of computer ids to shut down.

    It also supports an optional argument:

    - **deliver_after**: Shutdown the computer after the specified time.  The
      time format is YYYY-MM-DDTHH:MM:SSZ.

    This is an example of a valid request::

        ?action=ShutdownComputers&computer_ids.1=30&computer_ids.2=43

    The method returns a JSON serialized activity::

        {u'computer_id': None,
         u'creation_time': u'2012-11-19T18:14:19Z',
         u'creator': {u'email': u'john@example.com', u'id': 3,
                      u'name': u'John Smith'},
         u'id': 147,
         u'parent_id': None,
         u'summary': u'Shutdown computer',
         u'type': u'ActivityGroup'}
    """

    schema = Schema(
        name="ShutdownComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to shutdown."),
            Unicode(
                "deliver_after",
                optional=True,
                doc="A time in the future to deliver the script.",
            ),
        ],
        errors=[InvalidTimeError, UnknownComputerError],
        doc="""Shutdown a list of computers.""",
    )

    requires = "ShutdownComputer"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """
        Shutdown computers with the specified ids.
        """
        person = get_person_by_id(call.principal.id)
        management = ComputerManagement(context, person)

        deliver_after = None
        if call.args.deliver_after:
            try:
                deliver_after = parse_datetime(call.args.deliver_after)
            except ValueError:
                raise InvalidTimeError(
                    "Invalid format for deliver_after: '%s'" % call.args.deliver_after
                )

        activity = management.shutdown(deliver_after=deliver_after)
        return activity.get_state()


@method
class RenameComputers(AccountMethod):
    """
    Rename a set of computers.

    This method takes a single mandatory argument:

    - **computer_titles**: A mapping of computer_ids to titles to set.

    This is an example of a valid request::

        ?action=RenameComputers&computer_titles.30:newname
    """

    schema = Schema(
        name="RenameComputers",
        parameters=[
            Mapping(
                "computer_titles",
                Integer("computer_id", doc="The ID of the computer to rename"),
                Unicode("computer_name", doc="The new name to apply."),
                doc="mapping of computer_ids to computer titles",
            )
        ],
        errors=[UnknownComputerError, InvalidTitleError],
        doc="Rename a set of computers.",
    )

    requires = "EditComputerDetails"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """Rename a set of computers."""
        account = context
        computer_titles = dict(call.args.computer_titles)
        computers = account.get_computers_by_ids(list(computer_titles.keys()))
        for computer in computers:
            title = computer_titles[computer.id]
            # title is required
            if not title:
                raise InvalidTitleError("Invalid title: '{}'".format(title))
            computer.title = title

        return {}


@method
class GetWSLHosts(AccountMethod):
    """
    Gets a list of Windows computers that is hosting at least one WSL instance.
    """

    requires = "ViewAccount"

    schema = search_schema.extend(
        name="GetWSLHosts",
        parameters=[],
        errors=[],
        doc="Gets a list of Windows computers that is hosting at least one WSL instance",
    )

    route = SelectionRoute(permission="ViewComputer", error_if_empty=False)

    def query(self, context: ComputerSelection, call: Call):
        store = context._main_store
        origin = [
            Computer,
            Join(Distribution, Computer.distribution_id == Distribution.id),
            Join(ComputerRelationship, ComputerRelationship.parent_id == Computer.id),
        ]
        search = ComputerSearch(context.account)
        where = search.get_where_clause(call.args.query)

        result = store.using(*origin).find(
            Computer,
            And(*where),
            Distribution.code_name == WSL_HOST_CODE_NAME,
            ComputerRelationship.parent_id != None,
        )
        result = result.config(
            distinct=True,
            offset=call.args.offset,
            limit=call.args.limit,
        )

        return [computer.get_state() for computer in result]


@method
class RemoveWSLHosts(AccountMethod):
    """Removes Windows computers that are hosting at least one WSL instance."""

    schema = Schema(
        name="RemoveWSLHosts",
        parameters=[
            List("computer_ids", Integer(), doc="A list of computer ids to remove."),
            Bool(
                "cascade_to_children",
                optional=True,
                default=False,
                doc="If true, removed computers will also include child computers.",
            ),
        ],
        errors=[
            UnknownComputerError,
            NotWSLHostError,
            *RemoveComputersOperation.errors(),
        ],
        doc="Remove a list of Windows computers that are hosting at least one WSL instance by id",
    )

    route = AccountRoute()

    def query(self, context: Account, call: Call):
        """Remove WSL host computers with the specified ids."""

        # Determine if computers are WSL hosts before removing
        main_store = get_main_store()
        hosts = main_store.using(
            Computer,
            Join(Distribution, Computer.distribution_id == Distribution.id),
            Join(ComputerRelationship, ComputerRelationship.parent_id == Computer.id),
        ).find(
            Computer.id,
            Distribution.code_name == WSL_HOST_CODE_NAME,
            ComputerRelationship.parent_id.is_in(call.args.computer_ids),
        )
        computers = list(hosts)
        not_hosts = set(call.args.computer_ids) - set(computers)

        if not_hosts:
            raise NotWSLHostError(
                f"Computers with id={not_hosts} are not WSL hosts or unknown"
            )

        if call.args.cascade_to_children:
            computers += list(
                context.get_child_computers(computers).values(Computer.id)
            )

        collection = AccountComputerTwinCollection(context.store, account=context)
        collection = collection.with_ids(computers)

        try:
            RemoveComputersOperation(collection, call.principal)()
        except UnknownEntityIDsError as error:
            raise UnknownComputerError(f"Unknown computer '{error.ids.pop(0)}'")

        return {"message": f"Successfully removed computers with id={computers}"}


@method
class DeleteChildComputers(AccountMethod):
    """
    Send a message to a hostagent to delete a child computer instance.
    """

    requires = "RemoveComputer"

    schema = Schema(
        name="DeleteChildComputers",
        parameters=[
            List(
                "computer_ids", Integer(), doc="A list of child computer ids to delete."
            )
        ],
        errors=[UnknownComputerError],
        doc="""Delete a list of child computers by ID.""",
    )

    route = SelectionRoute(permission="RemoveComputer", error_if_empty=False)

    def query(self, context, call):
        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "delete-child-computers", context=context, person=person
        )
        state = handler.handle(computer_ids=context.get_selected_computer_ids())

        return state


@method
class StartChildComputers(AccountMethod):
    """
    Send a message to a hostagent to start a child computer instance.
    """

    requires = "RestartComputer"

    schema = Schema(
        name="StartChildComputers",
        parameters=[
            List(
                "computer_ids", Integer(), doc="A list of child computer ids to start."
            )
        ],
        errors=[UnknownComputerError],
        doc="""Start a list of child computers by ID.""",
    )

    route = SelectionRoute(permission="RestartComputer", error_if_empty=False)

    def query(self, context, call):
        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "start-child-computers", context=context, person=person
        )
        state = handler.handle(computer_ids=context.get_selected_computer_ids())

        return state


@method
class StopChildComputers(AccountMethod):
    """
    Send a message to a hostagent to stop a child computer instance.
    """

    requires = "ShutdownComputer"

    schema = Schema(
        name="StopChildComputers",
        parameters=[
            List("computer_ids", Integer(), doc="A list of child computer ids to stop.")
        ],
        errors=[UnknownComputerError],
        doc="""Stop a list of child computers by ID.""",
    )

    route = SelectionRoute(permission="RestartComputer", error_if_empty=False)

    def query(self, context, call):
        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "stop-child-computers", context=context, person=person
        )
        state = handler.handle(computer_ids=context.get_selected_computer_ids())

        return state
