# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Definitions for the errors that the Landscape API can raise.

This should only contain exceptions that need to be explicitly raised by the
API code. Exceptions from model code can be automatically mapped by using
L{canonical.landscape.api.errormapping}.
"""


from canonical.txapi.server.exception import APIError


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class LandscapeAPIError(APIError):

    status = 400

    def __init__(self, message):
        APIError.__init__(self, self.status, self.code_name, message)

    @classproperty
    def code_name(klass):
        return klass.__name__[: -len("Error")]

    def __str__(self):
        return super(LandscapeAPIError, self).__str__()


class UnknownAccountError(LandscapeAPIError):
    """The specified account couldn't be found."""


class InvalidAccountNameError(LandscapeAPIError):
    """The specified account name is invalid."""


class MultipleAccountsError(LandscapeAPIError):
    """The specified person is in multiple accounts."""


class DuplicateAccountError(LandscapeAPIError):
    """The specified account already exists."""


class UnknownUserError(LandscapeAPIError):
    """A user couldn't be found."""


class InvalidDateError(LandscapeAPIError):
    """Invalid date format. Dates must be in YYYY-MM-DD."""


class InvalidTimeError(LandscapeAPIError):
    """Invalid time format."""


class InvalidTitleError(LandscapeAPIError):
    """Title is invalid."""


class MissingParameterError(LandscapeAPIError):
    """A required parameter was not provided."""


class InvalidParameterCombinationError(LandscapeAPIError):
    """Mutually exclusive parameters must not be provided together."""


class UnknownRoleError(LandscapeAPIError):
    """The specified role does not exist."""


class InvalidRoleError(LandscapeAPIError):
    # Update this documentation if we start using this role for anything else
    """The specified role cannot be be added to an LDS licence."""


class SelfRevokeGlobalAdminError(LandscapeAPIError):
    """A GlobalAdmin can't remove self from the role."""


class UnknownAccessGroupError(LandscapeAPIError):
    """The specified access group couldn't be found."""


class UnknownResourceGroupError(LandscapeAPIError):
    """The specified resource group couldn't be found."""


class NoLDSLicenseError(LandscapeAPIError):
    """The specified account is not enabled for LDS."""


class LicenseNotFoundError(LandscapeAPIError):
    """No license could be found matching the specified criteria."""


class InsufficientLicensesError(LandscapeAPIError):
    """Insufficient licenses are available to accept new computers."""


class DuplicateSalesforceKeyError(LandscapeAPIError):
    """A Salesforce key already exists."""


class DuplicateSalesforceAccountKeyError(LandscapeAPIError):
    """The specified Salesforce account key is already in use."""


class DuplicateEmailAddressError(LandscapeAPIError):
    """Email address already exists."""


class PersonHasNoRoleError(LandscapeAPIError):
    """The person doesn't have the specified role."""


class PersonAlreadyHasRoleError(LandscapeAPIError):
    """The person already has the specified role."""


class InvalidQueryError(LandscapeAPIError):
    """The query is invalid."""


class InvalidActivityStatusChangeError(LandscapeAPIError):
    """
    There was an incompatibility with the status transition of an activity.
    """


class UnknownCloudError(LandscapeAPIError):
    """The specified cloud couldn't be found."""


class InvalidTagError(LandscapeAPIError):
    """The specified tag name is invalid."""


class UnknownTagError(LandscapeAPIError):
    """The specified tag can't be found."""


class UnknownComputerError(LandscapeAPIError):
    """The specified computer couldn't be found."""


class UnknownComputersError(LandscapeAPIError):
    """No computers matching the specified criteria were found."""


class HasComputersError(LandscapeAPIError):
    """The specified account still contains computers."""


class GPGKeyImportError(LandscapeAPIError):
    """A key import failed."""

    code_name = "GPGKeyImportError"


class MultipleGPGKeyError(LandscapeAPIError):
    """Too many GPG keys were found."""


class UnknownGPGKeyError(LandscapeAPIError):
    """The specified GPG key couldn't be found."""


class GPGKeyInUseError(LandscapeAPIError):
    """The GPG key is currently in use and can't be removed."""


class GPGKeyHasNoSecretError(LandscapeAPIError):
    """The GPGkey has no secret key."""


class GPGKeyNotAssociatedError(LandscapeAPIError):
    """
    The GPG key cannot be removed because is not associated with the specified
    pocket.
    """


class GPGKeyAlreadyAssociatedError(LandscapeAPIError):
    """
    The GPG key cannot be reassociated as it is already associated with the
    specified pocket.
    """


class UnauthorisedError(LandscapeAPIError):
    """Permission denied."""

    status = 403


class InvalidCredentialsError(LandscapeAPIError):
    """Credentials supplied are not associated with an account."""

    status = 403


class InvalidPackageConstraintError(LandscapeAPIError):
    """The package constraint couldn't be parsed."""


class UnknownPackagesError(LandscapeAPIError):
    """A package wasn't found."""


class DuplicateMacAddressError(LandscapeAPIError):
    """A provisioned system with that MAC address already exists."""


class UnknownMAASServerError(LandscapeAPIError):
    """A MAAS server could not be found."""


class InvalidSavedSearchNameError(LandscapeAPIError):
    """The name provided for a saved search is invalid."""


class InvalidSavedSearchCriteriaError(LandscapeAPIError):
    """The criteria provided for a saved search is invalid."""


class UnknownSavedSearchError(LandscapeAPIError):
    """The specified saved search could not be found."""


class UnknownScriptError(LandscapeAPIError):
    """The specified script could not be found."""


class RequiresUserError(LandscapeAPIError):
    """The specified script requires a username to be specified."""


class DuplicateSeriesError(LandscapeAPIError):
    """A series with same name already exists in the specified distribution."""


class UnknownDistributionError(LandscapeAPIError):
    """The specified distribution couldn't be found."""


class UnknownSeriesError(LandscapeAPIError):
    """The specified series couldn't be found."""


class InvalidPocketFilterError(LandscapeAPIError):
    """
    The filter_type specified for the pocket is neither blacklist nor
    whitelist.
    """


class DuplicatePocketError(LandscapeAPIError):
    """The pocket specified already exists."""


class NoPocketFilteringError(LandscapeAPIError):
    """The pocket specified does not allow filtering."""


class InvalidPocketModeError(LandscapeAPIError):
    """The mode specified for the pocket is not valid."""


class UnknownPocketError(LandscapeAPIError):
    """The pocket specified couldn't be found."""


class APTSourceInUseError(LandscapeAPIError):
    """
    The APT Source cannot be removed because it is still referenced by
    repository profiles.
    """


class UnknownAPTSourceError(LandscapeAPIError):
    """The specified APT source couldn't be found."""


class SettingsValueError(LandscapeAPIError):
    """The specified settings key/value pair is invalid."""


class MultiError(LandscapeAPIError):
    """
    Multiple errors occurred.
    """

    code_name = "MultiError"

    def __init__(self, action_error):
        self.action_error = action_error
        LandscapeAPIError.__init__(self, "Multiple errors occurred.")

    def get_reasons(self):
        return self.action_error.get_reasons()


class NotWSLHostError(LandscapeAPIError):
    """The specified computer is not a WSL host."""
