# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
A mapping of model exceptions to API error codes.
"""
from canonical.txapi.server.schema import (
    MissingParameterError,
    InvalidParameterValueError,
)

from zope.security.interfaces import Unauthorized

from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError
from canonical.landscape.model.main.schema import InvalidEmailAddress
from canonical.landscape.model.access.errors import (
    DuplicateRoleError,
    InvalidRoleNameError,
    UnknownRoleError,
    InvalidPermissionsError,
    ReadOnlyRoleError,
)
from canonical.security.role import UnknownRoleError as SecurityUnknownRoleError
from canonical.landscape.model.main.errors import (
    UnknownPersonsError,
    UnknownAccessContextsError,
    UnknownAccessContextError,
)
from canonical.landscape.model.main.account import (
    InvalidAccessContext,
    PersonNotMemberOfAccountError,
)
from canonical.landscape.model.main.license import LicenseSeatError
from canonical.landscape.model.account.errors import (
    MaximumAccountAdministratorsError,
    RepeatedInvitationError,
)
from canonical.landscape.model.account.savedsearch import DuplicateSavedSearchError
from canonical.landscape.model.package.profile import (
    InvalidParameterCombinationError as PPInvalidParameterCombinationError,
    NoFoundPackagesError,
    DuplicatePackageProfileError,
    UnknownPackageProfileNameError,
    InvalidPackageProfileNameError,
    InvalidPackageProfileMaterialError,
    EmptyPackageProfileError,
)
from canonical.landscape.model.package.package import (
    InvalidPackageConstraintTypeError,
    InvalidPackageConstraintError,
)
from canonical.landscape.model.package.autoupgrade import ScheduleFormatException
from canonical.landscape.model.package.errors import UnknownUpgradeProfileError
from canonical.landscape.model.repository.distribution import (
    LockError,
    DuplicateDistributionError,
)
from canonical.landscape.model.repository.profile import (
    UnknownRepositoryProfileError,
    DuplicateRepositoryProfileError,
)
from canonical.landscape.model.repository.aptsource import (
    DuplicateAPTSourceError,
    UnknownAPTSourceError,
)
from canonical.landscape.model.repository.management import InvalidAPTLine
from canonical.landscape.model.repository.series import NoPocketsInSeries
from canonical.landscape.model.gpg.key import DuplicateGPGKeyError
from canonical.landscape.model.resource.annotation import InvalidAnnotationKeyError
from canonical.landscape.model.script.errors import (
    UnknownScriptError,
    UnknownScriptAttachmentError,
    DuplicateScriptAttachmentError,
    TooManyScriptAttachmentsError,
    ScriptAttachmentSizeError,
    ScriptEncodingError,
)
from canonical.landscape.model.script.script import (
    DuplicateScriptError,
    EmptyScriptCodeError,
    EmptyScriptInterpreterError,
)
from canonical.landscape.model.computer.removal import (
    UnknownRemovalProfileError,
    DaysWithoutExchangeValueError,
)
from canonical.landscape.service.errors import ServiceError
from canonical.landscape.model.alert.rule import (
    UnknownAlertTypeError,
    InvalidAlertTypeError,
)


def get_api_errors():
    """
    Get the default map of exceptions to error codes.
    """
    return {
        # Generic
        Unauthorized: "Unauthorised",
        UnknownPersonsError: "UnknownPersonEmails",
        MissingParameterError: "MissingParameter",
        InvalidParameterValueError: "InvalidParameterValue",
        # Access Groups
        DuplicateRoleError: "DuplicateRoleName",
        InvalidRoleNameError: "InvalidRoleName",
        UnknownRoleError: "UnknownRole",
        SecurityUnknownRoleError: "UnknownRole",
        InvalidPermissionsError: "InvalidRolePermissions",
        UnknownAccessContextsError: "UnknownAccessGroups",
        InvalidAccessContext: "InvalidAccessGroup",
        UnknownAccessContextError: "UnknownAccessGroup",
        ReadOnlyRoleError: "ReadOnlyRole",
        # Saved Searches
        DuplicateSavedSearchError: "DuplicateSavedSearch",
        # Package Profiles
        NoFoundPackagesError: "NoFoundPackages",
        PPInvalidParameterCombinationError: "InvalidParameterCombination",
        DuplicatePackageProfileError: "DuplicatePackageProfile",
        UnknownPackageProfileNameError: "UnknownPackageProfileName",
        InvalidPackageProfileMaterialError: "InvalidPackageProfileMaterial",
        EmptyPackageProfileError: "EmptyPackageProfile",
        InvalidPackageProfileNameError: "InvalidPackageProfileName",
        InvalidPackageConstraintTypeError: "InvalidConstraintType",
        InvalidPackageConstraintError: "InvalidPackageConstraint",
        # Upgrade Profiles
        ScheduleFormatException: "InvalidScheduleFormat",
        UnknownUpgradeProfileError: "UnknownUpgradeProfile",
        InvalidScheduleAfterWindowError: "InvalidDeliverDelayWindow",
        # Account Management
        RepeatedInvitationError: "RepeatedInvitation",
        PersonNotMemberOfAccountError: "PersonNotMemberOfAccount",
        MaximumAccountAdministratorsError: "MaximumAccountAdministrators",
        InvalidEmailAddress: "InvalidEmailAddress",
        # Repository Management
        InvalidAPTLine: "InvalidAPTLine",
        DuplicateAPTSourceError: "DuplicateAPTSource",
        UnknownAPTSourceError: "UnknownAPTSource",
        LockError: "OperationInProgress",
        DuplicateDistributionError: "DuplicateDistribution",
        NoPocketsInSeries: "EmptySeries",
        # Repository Profiles
        DuplicateRepositoryProfileError: "DuplicateRepositoryProfile",
        UnknownRepositoryProfileError: "UnknownRepositoryProfile",
        # Annotation Keys
        InvalidAnnotationKeyError: "InvalidAnnotationKey",
        # GPG Keys
        DuplicateGPGKeyError: "DuplicateGPGKey",
        # Scripts
        UnknownScriptError: "UnknownScript",
        UnknownScriptAttachmentError: "UnknownScriptAttachment",
        DuplicateScriptAttachmentError: "DuplicateScriptAttachment",
        TooManyScriptAttachmentsError: "TooManyScriptAttachments",
        ScriptAttachmentSizeError: "ScriptAttachmentSize",
        DuplicateScriptError: "DuplicateScript",
        EmptyScriptCodeError: "EmptyScriptCode",
        EmptyScriptInterpreterError: "EmptyScriptInterpreter",
        ScriptEncodingError: "ScriptEncoding",
        LicenseSeatError: "NoLicenseSeat",
        # Removal Profiles
        UnknownRemovalProfileError: "UnknownRemovalProfile",
        DaysWithoutExchangeValueError: "InvalidDaysWithoutExchangeValue",
        # Alert rules
        UnknownAlertTypeError: "UnknownAlertTypeError",
        InvalidAlertTypeError: "InvalidAlertTypeError",
    }


def get_error_code(exception_type):
    """
    Get the API code name for an exception.
    """
    if issubclass(exception_type, ServiceError):
        return exception_type.code()

    # XXX eventually API methods should only fail because of ServiceErrors, but
    # as we complete the transition to build the API on top of the service
    # layer we also consider the legacy errors.
    if isinstance(exception_type, str):
        return exception_type
    if hasattr(exception_type, "code_name"):
        return exception_type.code_name
    codes = get_api_errors()
    return codes.get(exception_type)
