# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Event Log
============

Methods for retrieving or exporting an account's EventLog
"""

from canonical.txapi.server.schema import Schema, Integer
from canonical.txapi.server.method import method

from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.routes import AccountRoute
from canonical.landscape.service.eventlog.get import GetEventLogEntries


@method
class GetEventLog(AccountMethod):
    """Retrieve event log entries for the account, ordered by creation time.

    - **days**: The number of days prior to today from which to fetch log
      entries.
    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.

    For example, the following returns the last 7 days for event log entries
    in the account, limiting the result to 100 entries::

        ?action=GetEventLog&days=7&limit=100


    The method returns a JSON representation of the log entries returned::

        [{"creation_time": u'2014-09-16T18:37:45Z',
          "entity_id": 2,
          "entity_name": u"WebServer1",
          "entity_type": u"Computer",
          "id": 1,
          "message": u"Added tag 'apache' to computer 'WebServer1'",
          "person_id", 3,
          "person_name": u"John Smith"},
          ...
        ]
    """

    schema = Schema(
        name="GetEventLog",
        parameters=[
            Integer(
                "days",
                optional=True,
                doc=(
                    "The number of days prior to today from which to "
                    "fetch log entries. It defaults to 30 days."
                ),
            ),
            Integer(
                "limit",
                default=RESULT_LIMIT,
                optional=True,
                doc=(
                    "The maximum number of results returned by the "
                    "method. It defaults to %d." % RESULT_LIMIT
                ),
            ),
            Integer(
                "offset",
                default=RESULT_OFFSET,
                optional=True,
                doc="The offset inside the list of results.",
            ),
        ],
        errors=GetEventLogEntries.errors(),
        doc="Retrieve event log for the account.",
    )

    route = AccountRoute()

    def query(self, context, call):
        """
        Return the account C{EventLog} over the requested number of C{days}.
        """
        operation = GetEventLogEntries(context, call.principal)
        result = operation(
            days=call.args.days, offset=call.args.offset, limit=call.args.limit
        )
        return [log.get_state() for log in result]
