# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
GPG
=======

The methods available here are for GPG key handling.
"""

import shutil
import tempfile
from datetime import date

from zope.security.interfaces import Unauthorized

from canonical.txapi.server.schema import Schema, Unicode, List
from canonical.txapi.server.method import method

from canonical.gnupg.gnupg import GnuPG
from canonical.lib.text import is_valid_slug

from canonical.landscape.model.gpg.key import DuplicateGPGKeyError
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.error import (
    GPGKeyImportError,
    MultipleGPGKeyError,
    UnknownGPGKeyError,
    GPGKeyInUseError,
)
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.model.gpg.management import GPGKeyManagement


@method
class ImportGPGKey(AccountMethod):
    """
    Import a GPG key.

    - **name**: Name of the GPG key. It must be unique within the account,
      start with an alphanumeric character and only contain lowercase letters,
      numbers and - or + signs.
    - **material**: The text representation of the key.

    Example of a valid request::

        ?action=ImportGPGKey&name=my-key
            &material=-----BEGIN PGP PUBLIC KEY BLOCK-----
              Version: GnuPG v1.4.11 (GNU/Linux)
              mQENBE4xHgQBCADESzhTFEYYCOrvxBCnwtdQZa2DNDf/RbgEEOW/XEh3E5j9kUPj
              .....
              Ci8pOnAXPkVXmT5+um1o8b4bzP4BmGbxemmCdPksJxMt6Qq7n88406M6QPaLj/oz
              -----END PGP PUBLIC KEY BLOCK-----

    The following error may be raised:

    - **DuplicateGPGKey**: A key with same name or fingerprint already exists.
    - **GPGKeyImportFailed**: Key import failed.
    - **MultipleGPGKey**: The material contains multiple keys.
    - **Unauthorised**: The user is not authorised to perform the operation.

    The method returns a JSON representation of the stored key::

        {"fingerprint": "a404:34a3:e40c:1add:94fa:31b4:30a7:5431:a2eb:521a",
         "key_id": "30A75431A2EB521A",
         "id": 11,
         "name": "my-key",
         "has_secret": true}
    """

    requires = "CreateGPGKey"

    schema = Schema(
        name="ImportGPGKey",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    "Name of the GPG key. It must be unique within the "
                    "account, start with an alphanumeric character and "
                    "only contain lowercase letters, numbers and - or + "
                    "signs."
                ),
            ),
            Unicode("material", doc="The text representation of the key."),
        ],
        errors=[
            GPGKeyImportError,
            DuplicateGPGKeyError,
            MultipleGPGKeyError,
            Unauthorized,
        ],
        doc="""Import a GPG key.""",
    )

    def invoke(self, call):
        # XXX This is overridden to load the call object with additional
        # parameters before the transaction is run. It should be kept
        # in sync with the original Method.invoke() implementation.
        call.parse(self.schema)
        deferred = self._import_key(call)
        deferred.addCallback(lambda _: self.transactor.run(self._transaction, call))
        deferred.addErrback(self._handle_errors)
        return deferred

    def _gnupg_failure(self, failure):
        """Handle GPG failures."""
        raise GPGKeyImportError("Key import failed")

    def _check_expired(self, keys_info):
        """Check that the key being imported is not expired."""
        key_info = list(keys_info.values())[0]
        expiration_date = key_info["expiration_date"]
        if expiration_date is not None and expiration_date < date.today():
            raise GPGKeyImportError("GPG key is expired")
        return key_info

    def _check_import(self, keys, gnupg):
        """Validate the material being imported and return info about it.

        @param keys: The keys as returned by GnuPG.list_keys.
        """
        if len(keys) > 1:
            raise MultipleGPGKeyError("Multiple GPG keys in material")

        [(key_id, has_secret)] = keys
        deferred = gnupg.list_keys(keys=[key_id])
        deferred.addCallback(self._check_expired)
        deferred.addCallback(lambda info: (info["fingerprint"], has_secret))
        return deferred

    def _passphrase_failure(self, failure):
        raise GPGKeyImportError("GPG secret key must not require a passphrase")

    def _check_no_passphrase(self, gpg_data, gnupg):
        """
        In case of a secret key, check that it doesn't require a passphrase.
        """
        (fingerprint, has_secret) = gpg_data
        if not has_secret:
            return (fingerprint, has_secret)

        deferred = gnupg.sign("sample text")
        deferred.addCallback(lambda _: (fingerprint, has_secret))
        deferred.addErrback(self._passphrase_failure)
        return deferred

    def _extract_public(self, gpg_data, gnupg, key_material):
        """Extract the public part of a key material."""
        (fingerprint, has_secret) = gpg_data
        if has_secret:
            deferred = gnupg.export_keys()
            deferred.addCallback(
                lambda public: (fingerprint, public.decode("utf8"), key_material)
            )
            return deferred
        else:
            # The given material contains no secret part, nothing to do.
            return (fingerprint, key_material, None)

    def _import_key(self, call):
        """
        Imports the key and returns key id, fingerprint and secret key if
        available.
        """
        key_material = call.args.material
        tempdir = tempfile.mkdtemp()
        gnupg = GnuPG(tempdir)
        deferred = gnupg.import_keys([key_material])
        deferred.addErrback(self._gnupg_failure)
        deferred.addCallback(self._check_import, gnupg)
        deferred.addCallback(self._check_no_passphrase, gnupg)
        deferred.addCallback(self._extract_public, gnupg, key_material)

        def cleanup(result, path):
            # XXX stuff the result in the call object instead of passing
            # explicitly to query(), as the latter is invoked by _transaction()
            call.gpg_data = result
            shutil.rmtree(path, ignore_errors=True)
            return result

        deferred.addBoth(cleanup, tempdir)
        return deferred

    route = ManagementRoute(factory=GPGKeyManagement)

    def query(self, context, call):
        """Import the key."""
        management = context  # for readability
        fingerprint, material, secret_material = call.gpg_data
        key = management.create_key(
            call.args.name, fingerprint, material, secret_material=secret_material
        )
        return key.get_state()


@method
class RemoveGPGKey(AccountMethod):
    """
    Remove a GPG key.

    - **name**: Name of the GPG key.

    Example of a valid request::

        ?action=RemoveGPGKey&name=my-key

    The following error may be raised:

    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **GPGKeyInUse**: The GPG key is currently used by a pocket and can't be
      removed.
    """

    requires = "EditGPGKey"

    schema = Schema(
        name="RemoveGPGKey",
        parameters=[Unicode("name", doc="Name of the GPG key to remove.")],
        errors=[UnknownGPGKeyError, GPGKeyInUseError],
        doc="""Remove a GPG key.""",
    )

    route = SubjectRoute(
        factory=GPGKeyManagement,
        getter="get_key",
        identifier_arg="name",
        unknown_error_factory=UnknownGPGKeyError,
    )

    def query(self, context, call):
        """Get the key and removes it from db."""
        key = context  # for readability

        if key.get_pockets().count() > 0:
            raise GPGKeyInUseError("The key is linked to repository pockets")
        if key.get_apt_sources().count() > 0:
            raise GPGKeyInUseError("The key is linked to APT sources")

        management = self.get_management(GPGKeyManagement, call)
        management.remove_keys((key.id,))
        return {}


@method
class GetGPGKeys(AccountMethod):
    """
    Get info about GPG keys.

    This method takes an optional argument:

    - **names.#**: A list of GPG keys to get info for. If this is not provided,
      the call will return all keys for the account.

      Example call::

       ?action=GetGPGKeys&names.1=mykey

    The method returns JSON serialized info on the keys::

         [{"fingerprint": "aa65:abfd:ffa4:327a:6fa0:6bb1:faa1:83c3:534a:91ee",
           "has_secret": false,
           "id": 10,
           "key_id": "FAA183C3534A91EE",
           "name": "mykey"},
          {"fingerprint": "afd7:2cef:2aa4:b345:0304:0ab4:20ac:3539:5aaf:ffa4",
           "has_secret": true,
           "id": 1,
           "key_id": "20AC35395AAFFFA4",
           "name": "sign-key"}]
    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetGPGKeys",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc=(
                    "A list of GPG keys to get info for. If this is not "
                    "provided, the call will return all keys for the "
                    "account."
                ),
            )
        ],
        errors=[],
        doc="""Get info about GPG keys.""",
    )

    route = ManagementRoute(factory=GPGKeyManagement)

    def query(self, context, call):
        """Get the keys and return info."""
        result = context.get_keys(names=call.args.names or None)
        return [key.get_state() for key in result]
