# Copyright 2005-2022 Canonical Limited. All rights reserved.

import base64

from zope.component import getUtility

from canonical.txapi.server.schema import AbstractParameter as Parameter
from canonical.txapi.server.exception import APIError

from storm.exceptions import NotOneError

from canonical.security.interfaces import IRoleRegistry
from canonical.security.role import UnknownRoleError
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.main.person import (
    get_person_by_identity,
    get_person_by_id,
)
from canonical.landscape.model.main.account import (
    get_account_by_name,
    get_account_by_id,
)
from canonical.landscape.model.main.tag import get_tag
from canonical.landscape.model.main.errors import UnknownAccessContextsError
from canonical.landscape.api.error import (
    UnknownAccountError,
    MultipleAccountsError,
    UnknownUserError,
    UnknownAccessGroupError,
    UnknownRoleError as APIUnknownRoleError,
)


class APIHelper(object):
    """
    Helper class with methods raising appropriate API errors.
    """

    def check_unknown_names(self, names, entities, entity_name):
        """Check if invalid names are found and raise an error if so."""
        if names is None:
            names = set([])

        known_names = [entity.name for entity in entities]
        invalid_names = set(names).difference(known_names)

        if not invalid_names:
            return entities

        def to_capital_case(word):
            if word.islower():
                return word.capitalize()
            else:
                return word

        code = "Unknown" + "".join(
            [to_capital_case(word) for word in entity_name.split()]
        )
        message = "Unknown %s name%s: %s" % (
            entity_name,
            "s" if len(invalid_names) > 1 else "",
            ", ".join(sorted(invalid_names)),
        )
        raise APIError(400, code=code, message=message)

    def get_account_by_name(self, name):
        """
        Find the L{Account} called C{name}.

        @param name: The name of the L{Account}.

        @raises UnknownAccountError: raised if the name doesn't match an
            L{Account} in the system.
        """
        account = get_account_by_name(name)
        if account is None:
            raise UnknownAccountError("Unknown account name '%s'." % name)
        return account

    def get_account_by_person(self, person):
        """
        Find the L{Account} associated with a L{Person}.

        @param person: The person to get the account of.
        @type person: L{canonical.landscape.model.main.person.Person}
        @raises MultipleAccountsError: raised if the person supplied is
            attached to more than one account.
        @raises UnknownAccountError: raised if the person supplied is not
            attached to any accounts.
        """
        try:
            account = person.accounts.one()
        except NotOneError:
            raise MultipleAccountsError("Multiple accounts matching the given person.")
        if account is None:
            raise UnknownAccountError("No account matching the given person.")
        return account

    def get_person_by_identity(self, user_id):
        """
        Find the L{Person} identified by user id.

        @raises UnknownUserError: raised if the user id is not known.
        """
        person = get_person_by_identity(user_id)
        if person is None:
            raise UnknownUserError("Unknown user id: %s" % user_id)
        return person

    def get_role(self, role):
        """Return the role specified by name."""
        registry = getUtility(IRoleRegistry)
        try:
            return registry.get(role)
        except UnknownRoleError:
            raise APIUnknownRoleError("The role '%s' does not exist." % role)

    def get_access_context(self, account, name):
        """Return the C{AccessContext} in the account with specified name.

        @raises UnknownAccessGroupError: if no L{AccessContext} is found.
        """
        try:
            [access_context] = account.get_access_contexts(names=[name])
        except UnknownAccessContextsError:
            raise UnknownAccessGroupError("Unknown access group name '%s'" % name)
        return access_context

    def get_management(self, management_class, call):
        """
        Return a management class for the person and account specified by the
        call L{Principal}.
        """
        account = get_account_by_id(call.principal.account_id)
        person = get_person_by_id(call.principal.id)
        return management_class(account, person)

    def get_tags_by_names(self, names):
        """Return the L{Tags}s with the given C{names}.

        @param names: The names of the tags to return.

        @raises APIError: If some names do not result in existent tags.
        """

        tags = (get_tag(name) for name in names)
        tags = [tag for tag in tags if tag is not None]
        return self.check_unknown_names(names, tags, "tag")

    def event_log_from_call(self, call):
        """Create an C{EventLog} instance from a call."""
        person = get_person_by_id(call.principal.id)
        account = get_account_by_id(call.principal.account_id)
        return EventLog(account, person)


class File(Parameter):
    """
    Parameter representing a file's name and contents. The format is
    C{filename + "$$" + base64content}.
    """

    kind = "file"

    def parse(self, value):
        """
        This is called by the server to unpack the value after receiving it
        from the wire.
        """
        if isinstance(value, bytes):
            value = value.decode()

        (filename, contents) = value.split("$$", 1)
        try:
            binary_contents = base64.b64decode(contents)
        except TypeError:
            # For txapi, any param parse errors need to raise a ValueError.
            # Our invalid Base64 encoding becomes a ValueError
            raise ValueError()
        return (filename, binary_contents.decode())


class Data(Parameter):
    """
    A parameter representing some base64 encoded information.
    """

    kind = "data"

    def parse(self, value):
        """
        This is called by the server to decode the information after receiving
        it from the wire.
        """
        try:
            decoded = base64.b64decode(value)
        except TypeError:
            # For txapi, any param parse errors need to raise a ValueError.
            # Our invalid Base64 encoding becomes a ValueError
            raise ValueError()
        return decoded
