# Copyright 2005-2022 Canonical Limited. All rights reserved.

from copy import copy
from zope.component import getUtility
from zope.security.management import getSecurityPolicy

from canonical.txapi.server.exception import APIError
from canonical.txapi.server.schema import Schema, Unicode, Integer, Bool, List
from canonical.txapi.server.method import Method as Method_

from canonical.landscape.model.main.tag import get_tags
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.action.error import ActionError
from canonical.landscape.api.helpers import APIHelper
from canonical.landscape.api.error import (
    InvalidCredentialsError,
    InvalidQueryError,
    MultiError,
    UnauthorisedError,
    InvalidParameterCombinationError,
    UnknownTagError,
)
from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.participation import Participation
from canonical.landscape.api.errormapping import get_error_code, get_api_errors
from canonical.landscape.model.action.interfaces import IActionRegistry
from canonical.landscape.security.roles import is_support
from canonical.landscape.service.errors import ServiceError
from canonical.landscape.service.security import PersonPrincipal
from canonical.security.status import is_disabled

LATEST_VERSION = "2011-08-01"
FUTURE_VERSION = "2013-11-04"


class UndeclaredException(Exception):
    """
    An exception that was not found in the schema was raised from an API
    method.
    """

    def __init__(self, schema, failure):
        self.schema = schema
        self.failure = failure
        name = type(self.failure.value).__name__
        if name == "APIError":
            name += "(%s)" % (self.failure.value.code,)
        if name == "ActionError":
            name += "(%s)" % (
                [reason.code for reason in self.failure.value.get_reasons()],
            )
        self.exception_name = name

    def __str__(self):
        return (
            "Error type %s was raised from schema %s which only declares "
            "%s. Full traceback:\n %s"
            % (
                self.exception_name,
                self.schema.name,
                [get_error_code(error) for error in self.schema.errors],
                self.failure.getTraceback(),
            )
        )


class Method(Method_, APIHelper):
    """Handle a single HTTP request to the service API.

    @param transactor: The L{Transactor} wrapper to use to run database
       queries.

    @ivar requires: The permission needed to invoke this method, if set
        to C{None} then any principal can invoke this method.
    @ivar errors: A C{dict} mapping model exception classes to API error codes,
        by default the global mapping in errormapping.py is used. When such
        model errors are raised, an APIError with the associated code will
        be returned.
    @cvar cli_name: The command line name for this method. If not given the
        command line generator will derive it for the class name
        (e.g.: FooBar -> foo-bar).
    """

    versions = [LATEST_VERSION, FUTURE_VERSION]
    prefix = "landscape"
    errors = get_api_errors()
    cli_name = None

    def __init__(self, transactor):
        self.transactor = transactor

    def authorize(self, call):
        """
        Return C{True} by default, as permission check is done in the
        C{_transaction} method.
        """
        return True

    def invoke(self, call):
        """Invoke this method for executing the given C{call}.

        This method will call the L{query} method with the
        L{Principal} and the parameters extracted from the request,
        executing it in a separate thread.

        @param call: The API L{Call} to execute.
        @return: A deferred resulting in the response body, or calling
            L{errback} in case of error.
        """
        schemas = getattr(self, "schemas", {})
        # XXX It'd be nicer to pass the schema object around instead of
        # overriding the instance attribute, but this currently breaks a
        # few API methods that override this base class invoke() and
        # duplicate it (e.g ImportGPGKey). We should eventually fix those
        # methods.
        self.schema = schemas.get(call.version, self.schema)
        call.parse(self.schema)
        deferred = self.transactor.run(self._transaction, call)
        deferred.addErrback(self._handle_errors)
        return deferred

    def _transaction(self, call):
        """Run all logic that requires database access."""

        context, account = self.route(call)
        requires = getattr(self, "requires", None)
        if requires is not None:
            # XXX legacy API methods define security via the 'requires' class
            # attributes, which is not needed anymore in new API methods, since
            # security checks are handled completely by the service layer.
            policy = getSecurityPolicy()(Participation(call))
            permission = "landscape." + requires
            if not policy.checkPermission(permission, context):
                raise UnauthorisedError("Permission denied.")
        else:
            # XXX new API methods build on top of the service layer, so we
            # set the principal differently in order to use it for invoking
            # service layer operations (which will take care of enforcing
            # security).
            if not isinstance(call.principal, PersonPrincipal):
                # Avoid side-effect with transaction arguments
                # as routes are called earlier.
                call = copy(call)
                call.principal = PersonPrincipal(call.principal.id)

        if not is_support(call):
            if account is not None:
                if is_disabled(account, default=False):
                    raise UnauthorisedError("Account Disabled.")

        return self.query(context, call)

    def route(self, call):
        """Return a context object based on the parameters of the request.

        The responsibility of this method is to return a context object that
        security will be checked against, ensuring that the accessing person
        has the permission declared by the method on such context.

        In case of collection-like contexts, e.g. a ComputerSelection, this
        method is expected to include in the collection only items that the
        accessing person has the required permission one.
        """
        raise NotImplementedError()

    def query(self, context, call):
        """Interact with the database to perform the requested API call.

        @param context: The model context object returned by the route method.
        @param call: The API L{Call} object, containing the L{Principal}
            and request parameters.
        @return: The body of the response for the request being handled.
        """
        raise NotImplementedError()

    def _handle_errors(self, failure):
        """Handle errors declared in the C{errors} attribute."""
        # We make sure that the error that's being raised here is declared
        # in the schema's errors. This is because we rely on these errors
        # being declared in the client, as well as for potential
        # documentation accuracy purposes.
        if failure.type is APIError:
            # This is weird, but we still have some exceptions being
            # raised as APIErrors with strings for codes, and what's
            # worse, the codes are dynamically constructed with string
            # concatenation (see check_unknown_names). But we still
            # require that an *equivalent* exception (one which maps to
            # the same error code) be listed in the schema's errors.
            for error in self.schema.errors:
                schema_error_code = get_error_code(error)
                if schema_error_code == failure.value.code:
                    # Okay, it's listed, let's pass it through.
                    return failure
        if failure.check(ActionError):
            schema_error_codes = [get_error_code(error) for error in self.schema.errors]
            for reason in failure.value.get_reasons():
                if reason.code not in schema_error_codes:
                    raise UndeclaredException(self.schema, failure)
            raise MultiError(failure.value)
        elif failure.check(ServiceError):
            code = failure.value.code()
            message = str(failure.value)
            status = 400
            if code == "Unauthorised":
                status = 403
            raise APIError(status, code=code, message=message)
        else:
            always_acceptable = set([UnauthorisedError, InvalidCredentialsError])
            acceptable_errors = self.schema.errors.union(always_acceptable)
            if not failure.check(*acceptable_errors):
                raise UndeclaredException(self.schema, failure)

        # The mapping must DIRECTLY contain the error type raised; superclasses
        # don't work. Should we support error inheritance?
        code = self.errors.get(failure.type)
        if code:
            status = 400
            if code == "Unauthorised":
                status = 403
            raise APIError(status, code=code, message=str(failure.value))
        else:
            return failure

    def get_action_handler(self, name, context=None, person=None):
        """Return the L{ActionHandler} for the action with the given name."""
        registry = getUtility(IActionRegistry)
        handler = registry.get(name, context=context, person=person)
        return handler


class AccountMethod(Method, APIHelper):
    """Method subclass for account-related methods.

    This restricts these methods to credentials associated with a specific
    account, rather than "global" credentials used.
    """

    def authorize(self, call):
        """Check if the requested C{call} is authorized.

        If the principal credentials provided are not associated with
        an L{Account} then raise an error.
        """
        if not call.principal.account_id:
            raise InvalidCredentialsError(
                "Credentials supplied not associated with an account"
            )
        return True


class TagAssociationMethod(AccountMethod):
    """
    Base class for API methods for associating/disassociating with tags or all
    computers.
    """

    # The mode the method acts: if true, associates tags, otherwise
    # disassociates.
    associate = True

    schema = Schema(
        name="TagAssociationMethod",
        doc=(
            "Associate profile to computers with specified tags " "or to all computers."
        ),
        errors=[InvalidParameterCombinationError, UnknownTagError],
        parameters=[
            Unicode("name", doc="Name of the entity."),
            List(
                "tags",
                Unicode(),
                optional=True,
                doc=("Tags to change entity association for"),
            ),
            Bool(
                "all_computers",
                optional=True,
                default=False,
                doc=(
                    "If true, change the 'all_computers' flag state "
                    "for the entity. If the flag is enabled, associated "
                    "tags will be kept, but they will not be effective "
                    "until the flag is disabled."
                ),
            ),
        ],
    )

    def query(self, context, call):
        """
        Associate or disassociate tags and/or all computers with the profile
        specified as context object.
        """
        if call.args.tags and call.args.all_computers:
            raise InvalidParameterCombinationError(
                "The request cannot contain both tags.# parameters and "
                "all_computers=true."
            )

        if not call.args.tags and not call.args.all_computers:
            raise InvalidParameterCombinationError(
                "The request must contain either tags.# parameters "
                "or all_computers=true."
            )

        all_computers = None
        if call.args.all_computers:
            all_computers = self.associate

        new_tags = None
        if call.args.tags:
            new_tags = context.get_tags()
            if self.associate:
                new_tags.update(set(get_tags(call.args.tags)))
            else:
                remove_tags = self.get_tags_by_names(call.args.tags)
                new_tags = new_tags.difference(set(remove_tags))

        context.update_association(
            all_computers=all_computers,
            new_tags=new_tags,
            changes={},
            person=get_person_by_id(call.principal.id),
        )

        return self._get_state(context, call)

    def _get_state(self, context, call):
        """Generate the state to be returned after this API call is completed.

        Override this method to alter the data returned. Dhis is
        useful where additional security context is required.
        """
        return context.get_state()


search_schema = Schema(
    name="SearchMethod",
    parameters=[
        Unicode(
            "query",
            default="",
            optional=True,
            doc="A query string with space separated tokens used "
            "to filter the returned result objects.",
        ),
        Integer(
            "limit",
            default=RESULT_LIMIT,
            optional=True,
            doc=(
                "The maximum number of results returned by the "
                "method. It defaults to 1000."
            ),
        ),
        Integer(
            "offset",
            default=RESULT_OFFSET,
            optional=True,
            doc="The offset inside the list of results.",
        ),
    ],
    errors=[InvalidQueryError],
)
