# Copyright 2005-2022 Canonical Limited. All rights reserved.

from builtins import object


class AnonymousCredentials(object):
    """
    Credentials used only for our anonymous API methods such as BootstrapLDS.
    These methods will require anonymous credentials instead of a
    APICredentials based on L{Person}s in our database.
    """

    access_key_id = b"anonymous"
    access_secret_key = b"anonymous"
    account = None
    person = None


class Principal(object):
    """
    Wrap a L{APICredentials} instance, to provide the expected attributes for
    authentication, and add roles.

    @ivar id: The identifier of the associated L{Person}.
    @ivar access_key: The C{access_key_id} of the given L{APICredentials}.
    @ivar secret_key: The C{access_secret_key} of the given L{APICredentials}.
    @ivar roles: All the roles, of the L{Person}, per access group.
    @ivar account_id: The L{Account.id} of the account the L{APICredentials}
        are associated with, or C{None} if the credentials are global.
    @ivar access_context_id: The identifier of the access group the
        L{Account} belongs to, or C{None} if the credentials are global.
    """

    def __init__(self, credentials):
        # Note: be careful to not set any Storm object to instances
        # of the Principal class, because they will be attached to
        # instances txapi.server.call.Call which is passed around
        # crossing transaction (and thread) boundaries
        if credentials.person is not None:
            self.id = credentials.person.id
            self.roles = credentials.person.get_all_roles()
        else:
            self.id = None
            self.roles = None
        self.access_key = credentials.access_key_id
        self.secret_key = credentials.access_secret_key
        if credentials.account is not None:
            self.account_id = credentials.account.id
            self.access_context_id = credentials.account.access_context_id
        else:
            self.account_id = None
            self.access_context_id = None
