# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Removal Profiles
================


Landscape can be configured to automatically remove computers that have not
contacted the server within a set period of time.
"""
from zope.security.interfaces import Unauthorized

from canonical.txapi.server.schema import Bool, Schema, Unicode, Integer
from canonical.txapi.server.method import method

from canonical.landscape.api.methods import AccountMethod, TagAssociationMethod
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.model.computer.removal import (
    RemovalProfileManagement,
    UnknownRemovalProfileError,
    DaysWithoutExchangeValueError,
)
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.errors import UnknownAccessContextError


__all__ = [
    "CreateRemovalProfile",
    "RemoveRemovalProfile",
    "GetRemovalProfiles",
    "EditRemovalProfile",
    "AssociateRemovalProfile",
    "DisassociateRemovalProfile",
]


@method
class CreateRemovalProfile(AccountMethod):
    """
    Create a removal profile.

    - **title**: The human-readable title of the profile.
    - **days_without_exchange**: The length of time after which a computer may
        be removed.
    - **access_group**: If specified, the name of the access group the profile
        will apply to.
    - **cascade_to_children**: If true, removed computers will also include
        child computers (e.g. virtual machines, WSL instances).

    For example, the following command creates a removal profile titled
    "New Profile" that will remove computers after 4 days::

        ?action=CreateRemovalProfile&title=New%20Profile
            &days_without_exchange=4

    The method returns a JSON representation of the created removal profile::

        {"id": 1,
         "name": "new-profile",
         "title": "New Profile",
         "days_without_exchange": 4,
         "access_group": "global"}

    The following errors may be raised:

    - **UnknownAccessGroup**: The given access group is not known.
    - **Unauthorised**: The person is not authorised to create removal
      profiles in the given access group.
    - **InvalidDaysWithoutExchangeValue**: The days_without_exchange value
      is less than 1.
    """

    schema = Schema(
        name="CreateRemovalProfile",
        parameters=[
            Unicode("title", doc="The title of the profile to create."),
            Integer(
                "days_without_exchange",
                min=None,
                doc="The length of time after which a computer may be removed.",
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc="An optional name of an access group the profile will apply to.",
            ),
            Bool(
                "cascade_to_children",
                optional=True,
                default=False,
                doc="If true, removed computers will also include child computers (e.g. virtual machines, WSL instances).",
            ),
        ],
        errors=[UnknownAccessContextError, Unauthorized, DaysWithoutExchangeValueError],
        doc="""Create a removal profile.""",
    )

    # This cannot be EditRemovalProfile because it would interfere with
    # restricted admins.
    requires = "ViewAccount"

    route = ManagementRoute(factory=RemovalProfileManagement)

    def query(self, context, call):
        """Creates the RemovalProfile in the database"""
        profile = context.create_removal_profile(
            call.args.title,
            call.args.days_without_exchange,
            call.args.cascade_to_children,
            access_context_name=call.args.access_group,
        )
        return profile.get_state()


@method
class RemoveRemovalProfile(AccountMethod):
    """
    Remove an existing removal profile by name.

    Arguments:

    - **name**: The name of the removal profile to be removed.

    For example, the following request deletes a removal profile with the name
    "test-1"::

        ?action=RemoveRemovalProfile&name=test-1

    The following errors may be raised:

    - **UnknownRemovalProfile**: A profile with the specified name could not be
      found.
    """

    requires = "EditRemovalProfile"

    schema = Schema(
        name="RemoveRemovalProfile",
        parameters=[
            Unicode(
                "name", doc=("The name of the removal profile you wish to" "remove.")
            )
        ],
        errors=[UnknownRemovalProfileError],
        doc="""Remove an existing removal profile by name.""",
    )

    route = SubjectRoute(
        factory=RemovalProfileManagement,
        getter="get_removal_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """Removes the profile from the database."""
        profile = context
        management = self.get_management(RemovalProfileManagement, call)
        management.remove_removal_profiles([profile])
        return {}


@method
class GetRemovalProfiles(AccountMethod):
    """
    List existing removal profiles.

    For example, the following request lists all removal profiles::

        ?action=GetRemovalProfiles

    The method returns a JSON serialized representation of the list::

        [{"id": 1,
          "name": "servers",
          "title": "Servers",
          "days_without_exchange": 28,
          "access_group", "global"},
         {"id": 2,
          "name": "testing",
          "title": "Testing Systems",
          "days_without_exchange": 14,
          "access_group": "testing"}]
    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetRemovalProfiles",
        parameters=[],
        errors=[],
        doc="""List all existing removal profiles.""",
    )

    route = ManagementRoute(factory=RemovalProfileManagement)

    def query(self, context, call):
        profiles = context.get_removal_profiles()
        return [profile.get_state() for profile in profiles]


@method
class EditRemovalProfile(AccountMethod):
    """
    Edit a removal profile.

    Arguments:

    - **name**: The name of the profile to edit.
    - **title**: The new human-readable title of the profile.
    - **days_without_exchange**: The length of time after which a computer may
        be removed.

    For example, the following modifies the example profile in
    documented in CreateRemovalProfile to have a no communication period of
    31 days::

        ?action=EditRemovalProfile&name=newprofile&days_without_exchange=31

    The method returns a JSON serialized representation of the modified
    object::

        {"id": 1,
         "name": "newprofile",
         "title": "New Profile",
         "days_without_exchange": 31,
         "access_group": "global"}

    The following errors may be raised:

    - **UnknownRemovalProfile**: A profile with the specified name could not be
      found.
    - **Unauthorised**: The person is not authorised to edit removal
      profiles in the associated access group.
    - **InvalidDaysWithoutExchangeValue**: The days_without_exchange value
      is less than 1.
    """

    requires = "ViewAccount"

    schema = Schema(
        name="EditRemovalProfile",
        parameters=[
            Unicode("name", doc="The name of the profile to edit."),
            Unicode("title", optional=True, doc="The new title of the profile."),
            Integer(
                "days_without_exchange",
                optional=True,
                min=None,
                doc=("The length of time after which a computer may be " "removed."),
            ),
        ],
        errors=[
            UnknownRemovalProfileError,
            Unauthorized,
            DaysWithoutExchangeValueError,
        ],
        doc="""Edit an removal profile.""",
    )

    route = ManagementRoute(factory=RemovalProfileManagement)

    def query(self, context, call):
        """Update the RemovalProfile in the database."""
        profile = context.edit_removal_profile(
            name=call.args.name,
            title=call.args.title,
            days_without_exchange=call.args.days_without_exchange,
        )
        return profile.get_state()


@method
class AssociateRemovalProfile(TagAssociationMethod):
    """
    Associate a removal profile to computers with the specified tags,
    or all computers.

    Arguments:

    - **name**: Name of the removal profile.
    - **tags.#**: A list of tag names to associate to the profile.
    - **all_computers**: true if the profile should be associated to all
      computers. This parameter is *optional* and defaults to *false*.
      Individual tags associated to the profile will remain, but they will
      only be effective if the "all_computers" flag is later disabled.

    tags.# and all_computers=true are mutually exclusive.

    Example of a valid request::

        ?action=AssociateRemovalProfile&name=test-1
            &tags.1=server&tags.2=lucid

    The following errors may be raised:

    - **UnknownRemovalProfile**: No profile with the specified name exists.
    - **InvalidParameterCombination**: The set of arguments are not
      compatible when specified together.

    The method returns JSON serialized info of the profile status::

        {"name": "test-1",
         "id": 178,
         "upgrade_type": "all",
         "schedule": "FREQ=WEEKLY",
         "tags": ["my-computers", "lucid", "server"],
         "all_computers": False}
    """

    requires = "EditRemovalProfile"

    schema = TagAssociationMethod.schema.extend(
        name="AssociateRemovalProfile",
        errors=[UnknownRemovalProfileError],
        doc="""
    Associate a removal profile to computers with the specified tags,
    or all computers.

    tags and all_computers=true are mutually exclusive.""",
    )

    route = SubjectRoute(
        factory=RemovalProfileManagement,
        getter="get_removal_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownRemovalProfileError,
    )


@method
class DisassociateRemovalProfile(TagAssociationMethod):
    """
    Disassociate a removal profile from computers with the specified
    tags, or from all computers.

    Arguments:

    - **name**: Name of the removal profile.
    - **tags.#**: A list of tag names to disassociate from the profile.
    - **all_computers**: if true, the profile will only remain enabled for
      computers with tags associated to the profile. This parameter is
      *optional* and defaults to *false*.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=DisassociateRemovalProfile&name=test-1
            &tags.1=server&tags.2=lucid

    The following errors may be raised:

    - **UnknownRemovalProfile**: No profile with the specified name exists.
    - **InvalidParameterCombination**: The set of arguments are not
      compatible when specified together.

    The method returns JSON serialized info of the profile status::

        {"name": "test-1",
         "id": 178,
         "upgrade_type": "all",
         "schedule": "FREQ=WEEKLY",
         "tags": ["my-computers"],
         "all_computers": False}
    """

    requires = "EditRemovalProfile"

    associate = False

    schema = TagAssociationMethod.schema.extend(
        name="DisassociateRemovalProfile",
        errors=[UnknownRemovalProfileError],
        doc="""
    Disassociate a removal profile from computers with the specified
    tags, or from all computers.

    tags and all_computers=true are mutually exclusive.""",
    )

    route = SubjectRoute(
        factory=RemovalProfileManagement,
        getter="get_removal_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownRemovalProfileError,
    )
