# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Reporting
=========

The methods available here give the ability to do reporting on selections of
computers.

"""
from canonical.txapi.server.schema import Bool, Integer, List
from canonical.txapi.server.method import method

from canonical.landscape.api.methods import search_schema, AccountMethod
from canonical.landscape.api.routes import SelectionRoute
from canonical.landscape.model.reporting.management import ReportingManagement


@method
class GetNotPingingComputers(AccountMethod):
    """
    Get the ids of computers that have not pinged the server for a give number
    of minutes.

    This method takes the following arguments:

    - **since_minutes**: An integer number of minutes elapsed in which
      computers should have pinged the server.
    - **query**: A query string used to select the computers you wish to report
      on. See :ref:`Computer Queries <computer-query>` for details. (optional)

    For example, the following request will query for all machines running
    Ubuntu 12.04 (Precise) that have not pinged in more than 10 minutes::

        ?action=GetNotPingingComputers&query=distribution:12.04
            &since_minutes=10

    An example output looks like this::

        [44, 343, 5463]
    """

    schema = search_schema.extend(
        name="GetNotPingingComputers",
        parameters=[
            Integer(
                "since_minutes",
                doc=(
                    "The number of minutes elapsed in which no ping from"
                    " included computers has been seen."
                ),
            )
        ],
        doc=(
            "Report the ids of computers, within a given selection, that "
            "have not pinged the server in a given number of minutes."
        ),
    )

    requires = "ViewComputer"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        reporting_management = ReportingManagement(context)

        return list(
            reporting_management.get_not_pinging_computers(
                since_minutes=call.args.since_minutes or None
            )
        )


@method
class GetComputersNotUpgraded(AccountMethod):
    """
    Get the ids of computers that are not currently covered by an upgrade
    profile..

    This method takes one optional argument:

    - **query**: A query string used to select the computers you wish to report
      on. See :ref:`Computer Queries <computer-query>` for details.

    For example, the following request will query for all machines running
    Ubuntu 12.04 (Precise) that are not covered by an upgrade profile::

        ?action=GetComputersNotUpgraded&query=distribution:12.04

    An example output looks like this::

        [1, 3, 39, 500]
    """

    schema = search_schema.extend(
        name="GetComputersNotUpgraded",
        doc=(
            "Report the ids of computers, within a given selection, that "
            "are not covered by an upgrade profile."
        ),
    )

    requires = "ViewUpgradeProfile"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        reporting_management = ReportingManagement(context)
        return list(reporting_management.get_computers_not_upgraded())


@method
class GetUpgradedComputersByFrequency(AccountMethod):
    """
    Get a summary of the upgrade schedules of computers.

    This method takes one optional argument:

    - **query**: A query string used to select the computers you wish to report
      on. See :ref:`Computer Queries <computer-query>` for details.

    For example, the following request will query for all machines running
    Ubuntu 12.04 (Precise)::

        ?action=GetUpgradedComputersByFrequency&query=distribution:12.04

    Example output looks like::

        {u'Every hour at 30 minutes past the hour': [1, 2, 3, 6, 7, 8]}
    """

    schema = search_schema.extend(
        name="GetUpgradedComputersByFrequency",
        doc=(
            "Return a dictionary of computer IDs broken down by their "
            "upgrade schedule."
        ),
    )

    requires = "ViewUpgradeProfile"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        reporting_management = ReportingManagement(context)
        schedules = reporting_management.get_upgraded_computers_by_frequency()
        result = {}
        for schedule in schedules:
            result[schedule["display_name"]] = schedule["computer_ids"]
        return result


@method
class GetUSNTimeToFix(AccountMethod):
    """
    Get a summary of the lengths of time machines waited to be patched after
    USN releases.

    This method takes four optional arguments:

    - **query**: A query string used to select the computers you wish to report
      on. See :ref:`Computer Queries <computer-query>` for details.

    - **fixed_in_days**: A list of integer periods of days from a USN release
      into which we will group machines.  The default value is [2, 14, 30].

    - **pending_in_days**: The number of days in history to search for USN
      releases that are pending on the machines. Note, this parameter is
      independent of the in_last parameter. The default value is 60.

    - **in_last**: The number of days in history to search for USN releases to
      be considered in patch time statistics. The default value is 30.

    For example, the following request will query for all machines running
    Ubuntu 12.04 (Precise)::

        ?action=GetUSNTimeToFix&query=distribution:12.04

    This example request will return machines fixed in 7,14,21 days::

        ?action=GetUSNTimeToFix&fixed_in_days.1=7&fixed_in_days.2=14
            &fixed_in_days.3=21

    This example request will return statistics for unpatched times on USNs
    released in a window of 50 days prior to today's date::

        ?action=GetUSNTimeToFix&in_last=50

    Example output looks like::

        {u'30': [1, 2, 3, 4, 5, 6, 9, 10, 11],
         u'14': [1, 2, 3, 4, 5, 6, 7, 9, 10, 11],
         u'2': [1, 2, 3, 4, 5, 6, 9, 10, 11],
         u'pending': []}

    """

    schema = search_schema.extend(
        name="GetUSNTimeToFix",
        parameters=[
            List(
                "fixed_in_days",
                Integer(),
                optional=True,
                doc=(
                    "A list of periods of days to report on USN fixes "
                    "being applied in"
                ),
            ),
            Integer(
                "pending_in_days",
                optional=True,
                doc=(
                    "The period of days in the past to search for USNs "
                    "that are pending on a computer.  This is "
                    "independent of the in_last argument."
                ),
            ),
            Integer(
                "in_last",
                optional=True,
                doc=(
                    "The period of days to look into the past to find "
                    "USN releases to be considered in these statistics."
                ),
            ),
        ],
        doc=(
            "Return a break down of machines unpatched periods "
            "following a USN release."
        ),
    )

    requires = "ViewComputer"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        reporting_management = ReportingManagement(context)
        args = {}
        fixed_in_days = call.args.fixed_in_days or None
        pending_in_days = call.args.pending_in_days or None
        in_last = call.args.in_last or None
        if fixed_in_days:
            args["fixed_in_days"] = fixed_in_days
        if pending_in_days:
            args["pending_in_days"] = pending_in_days
        if in_last:
            args["in_last"] = in_last

        usn_times = reporting_management.get_usn_time_to_fix(**args)
        result = {"pending": list(usn_times["pending"])}
        for time, ids in list(usn_times["fixed_in"].items()):
            result[time] = list(ids)
        return result


@method
class GetCSVComplianceData(AccountMethod):
    """
    Get a breakdown of compliance data in CSV format (including a header row).

    The method takes optionals argument:

    - **query**: A query string used to select the computers you wish to report
      on. See :ref:`Computer Queries <computer-query>` for details.

    - **max_days**: Return issues newer than max_days.
      The default value is 30 days.

    - **by_cve**: If by_cve is false (the default), a key will be added for each
      USN released in the last max_days. The key name will be the identifier of
      the USN with timestamp, and the value will indicate whether the issue is
      present and if it is, whether it has been resolved, and when. If by_cve is
      true, CVEs will be used as the column key instead of USNs.

    For example, the following request will query for all machines running
    Ubuntu 12.04 (Precise)::

        ?action=GetCSVComplianceData&query=distribution:12.04

    An example output looks like this::

        u"name,is_pinging,schedule,USN 1234-12 (2012-08-19 14:51)
          Server 1,False,Every hour at 30 minutes past the hour,Not present
          Server 2,False,Every hour at 30 minutes past the hour,Not present
          Web,False,Every hour at 30 minutes past the hour,2012-08-20 14:51
          DB,False,Every hour at 30 minutes past the hour,Not present
         "
    """

    schema = search_schema.extend(
        name="GetCSVComplianceData",
        doc="Return a CSV formatted report of compliance data.",
        parameters=[
            Integer(
                "max_days",
                optional=True,
                doc="Return issues newer than max_days.",
            ),
            Bool(
                "by_cve",
                optional=True,
                doc=(
                    "If by_cve is false (the default), a key will be added for "
                    "each USN released in the last max_days. The key name will "
                    "be the identifier of the USN with timestamp, and the "
                    "value will indicate whether the issue is present and if "
                    "it is, whether it has been resolved, and when. If by_cve "
                    "is true, CVEs will be used as the column key instead."
                ),
            ),
        ],
    )

    requires = "ViewComputer"

    route = SelectionRoute(permission=requires, error_if_empty=False)

    def query(self, context, call):
        kwargs = {}

        if call.args.max_days is not None:
            kwargs["max_days"] = call.args.max_days

        if call.args.by_cve is not None:
            kwargs["by_cve"] = call.args.by_cve

        reporting_management = ReportingManagement(context)
        csv_data = reporting_management.get_csv_data(**kwargs)

        return csv_data
