# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
APT Sources
===================

The methods available here are related to operations on APT sources.
"""

from canonical.txapi.server.schema import Schema, Unicode, List
from canonical.txapi.server.method import method

from zope.security.interfaces import Unauthorized

from canonical.lib.text import is_valid_slug

from canonical.landscape.api.error import APTSourceInUseError
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.api.repository.methods import RepositoryMethod
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.errors import UnknownAccessContextError
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.repository.aptsource import (
    DuplicateAPTSourceError,
    UnknownAPTSourceError,
)
from canonical.landscape.model.repository.management import (
    RepositoryManagement,
    InvalidAPTLine,
)


__all__ = ["GetAPTSources", "CreateAPTSource", "RemoveAPTSource", "RemoveAPTSources"]


@method
class GetAPTSources(RepositoryMethod):
    """
    Get a list of apt sources in the account used for authentication.

    - **names.#**: List of names of the APT source to be returned.
      Multiple names can be supplied. (optional)

    Example of a valid request::

        ?action=GetAPTSources&names.1=lucid-mirror

    The method returns a JSON serialized list of the APT sources::

        [{"id": 100,
          "name": "lucid-mirror",
          "access_group": "global",
          "line": "deb http://archive.ubuntu.com/ubuntu lucid main",
          "gpg_key": null}]
    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetAPTSources",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc=(
                    "List of names of the APT source to be returned. "
                    "Multiple names can be supplied."
                ),
            )
        ],
        errors=[UnknownAccessContextError],
        doc="""
        Get a list of apt sources in the account used for authentication.
        """,
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """
        Return all APT sources in the account that match the call arguments.
        """
        result = context.get_apt_sources(names=call.args.names or None)
        return [source.get_state() for source in result]


@method
class CreateAPTSource(RepositoryMethod):
    """
    Create an APT source in the account used for authentication.

    - **name**: Name of the APT source. It must be unique within the
      account, start with an alphanumeric character and only contain lowercase
      letters, numbers and - or + signs.
    - **apt_line**: the APT line of the source.
    - **gpg_key**: Name of the GPG key used to sign the repository (optional).
    - **access_group**: An optional name of the access group to create the
      APT source into.

    Example of a valid request::

        ?action=CreateAPTSource&name=lucid-mirror
            &apt_line=deb+http://archive.ubuntu.com/ubuntu/+lucid+main

    The method returns a JSON serialized list of the created APT source::

        {"id": 100,
         "name": "lucid-mirror",
         "access_group": "global",
         "uri": "deb http://archive.ubuntu.com/ubuntu lucid main",
         "gpg_key": null}
    """

    requires = "ViewAccount"

    schema = Schema(
        name="CreateAPTSource",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    "Name of the APT source. It must be unique within "
                    "the account, start with an alphanumeric character "
                    "and only contain lowercase letters, numbers and - "
                    "or + signs."
                ),
            ),
            Unicode("apt_line", doc="The APT line of the source."),
            Unicode(
                "gpg_key",
                optional=True,
                doc="Name of the GPG key used to sign the repository",
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc="An optional name of the access group to create the "
                "APT source into.",
            ),
        ],
        errors=[DuplicateAPTSourceError, InvalidAPTLine, Unauthorized],
        doc="""
    Create an APT source in the account used for authentication.
    """,
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """Create a new APT source."""
        person = get_person_by_id(call.principal.id)
        if call.args.gpg_key:
            [gpg_key] = self.get_gpg_keys_by_names(
                context.account, person, names=[call.args.gpg_key]
            )
        else:
            gpg_key = None

        source = context.create_apt_source(
            call.args.name,
            call.args.apt_line,
            gpg_key,
            access_context_name=call.args.access_group,
        )

        return source.get_state()


@method
class RemoveAPTSources(RepositoryMethod):
    """
    DEPRECATED: use RemoveAPTSource instead.

    Remove APT sources in the account used for authentication.

    - **names.#**: List of names of the APT sources be removed. Multiple
      names can be supplied.
    """

    requires = "ViewAccount"

    schema = Schema(
        name="RemoveAPTSources",
        parameters=[
            List(
                "names",
                Unicode(),
                doc=(
                    "List of names of the APT sources be removed. Multiple "
                    "names can be supplied."
                ),
            )
        ],
        errors=[APTSourceInUseError, UnknownAPTSourceError],
        doc="""Deprecated: use RemoveAPTSource instead.""",
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """
        Remove APT sources in the account that matches the call arguments.
        """
        names = call.args.names
        sources = self.get_apt_sources_by_names(context, names=names)
        profiles = list(
            context.get_profiles(apt_sources_ids=[source.id for source in sources])
        )

        if profiles:
            raise APTSourceInUseError(
                "The following profiles are linked to "
                "APT sources: %s" % ", ".join(profile.name for profile in profiles)
            )

        not_removed = context.remove_apt_sources([source.id for source in sources])

        return not_removed


@method
class RemoveAPTSource(RepositoryMethod):
    """
    Remove an apt source.

    This method takes one mandatory argument:

    - **name**: Name of the apt source to be removed.

    Example of a valid request::

        ?action=RemoveAPTSource&name=lucid-mirror
    """

    requires = "EditAPTSource"

    schema = Schema(
        name="RemoveAPTSource",
        doc="Remove apt source from the account.",
        parameters=[Unicode("name", doc="The name of the apt source to be removed.")],
        errors=[UnknownAPTSourceError, APTSourceInUseError],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_apt_source_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """
        Remove a apt source from the account.
        """
        apt_source = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        profiles = list(management.get_profiles(apt_sources_ids=[apt_source.id]))

        if profiles:
            raise APTSourceInUseError(
                "The following profiles are linked to "
                "APT source '%s': %s"
                % (apt_source.name, ", ".join(profile.name for profile in profiles))
            )

        management.remove_apt_sources([apt_source.id])
        return {}
