# Copyright 2005-2022 Canonical Limited. All rights reserved.

import shutil

from canonical.txapi.server.schema import Schema, Unicode, List
from canonical.txapi.server.method import method

from canonical.lib.text import is_valid_slug
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.api.error import UnknownDistributionError
from canonical.landscape.api.repository.methods import RepositoryMethod
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.model.repository.distribution import LockError

from canonical.landscape.service.repository.distributioncreate import (
    CreateDistribution as CreateDistributionOperation,
)


@method
class CreateDistribution(RepositoryMethod):
    """
    Create a repository distribution associated with the account.

    This method takes a mandatory argument:

    - **name**: The name of the distribution. It must be unique within the
      account, start with an alphanumeric character and only contain lowercase
      letters, numbers and - or + signs.
    - **access_group**: An optional name of the access group to create the
      distribution into.

    For example, the following request creates a distribution named 'ubuntu'::

       ?action=CreateDistribution&name=ubuntu

    The following errors may be raised:

    - **DuplicateDistribution**: A distribution with same name already exists.
    - **UnknownAccessGroup**: The access group provided is not defined in the
      account.
    - **Unauthorized**: The user does not have authorized permissions to create
      distributions or on the supplied access_group.

    The method returns JSON serialized info of the created distribution::

        {"name": "ubuntu",
         "access_group": "global",
         "series": [],
         "creation_time": "2011-07-27T09:01:55Z"}
    """

    schema = Schema(
        name="CreateDistribution",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    "The name of the distribution. It must be unique "
                    "within the account, start with an alphanumeric "
                    "character and only contain lowercase letters, "
                    "numbers and - or + signs."
                ),
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc="An optional name of the access group to create the "
                "distribution into.",
            ),
        ],
        errors=(CreateDistributionOperation.errors()),
        doc="""
    Create a repository distribution associated with the account.
    """,
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        operation = CreateDistributionOperation(context, call.principal)
        distribution = operation(call.args.name, access_group=call.args.access_group)
        return distribution.get_state()


@method
class RemoveDistribution(RepositoryMethod):
    """
    Remove the specified repository distribution.

    This method takes a mandatory argument:

    - **name**: The name of the distribution.

    For example, the following request removes a distribution named 'ubuntu'::

       ?action=RemoveDistribution&name=ubuntu

    The following error may be raised:

    - **UnknownDistribution**: Distribution with specified name doesn't exist.
    - **OperationInProgress**: Another operation is in progress on the
      distribution.
    """

    schema = Schema(
        name="RemoveDistribution",
        parameters=[Unicode("name", doc="The name of the distribution to remove.")],
        errors=[UnknownDistributionError, LockError],
        doc="""Remove the specified repository distribution.""",
    )

    requires = "EditDistribution"

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_distribution",
        identifier_arg="name",
        unknown_error_factory=UnknownDistributionError,
    )

    def query(self, context, call):
        management = self.get_management(RepositoryManagement, call)
        distribution = context  # for readability

        distribution.lock()
        shutil.rmtree(distribution.get_basedir(), ignore_errors=True)
        management.remove_distribution(distribution)
        return {}


@method
class GetDistributions(RepositoryMethod):
    """
    Get info about distributions.

    This method takes an optional argument:

    - **names.#**: A list of distribution names to get info for. If this is not
      provided, the call will return all distributions for the account.

    For example, the following request returns info about a distribution named
    'ubuntu'::

       ?action=GetDistributions&names.1=ubuntu

    The method returns JSON serialized info on the distributions::

         [{"creation_time": "2011-07-19T12:51:18Z",
           "name": "ubuntu",
           "series": [
              {"creation_time": "2011-07-19T12:51:18Z",
               "name": "lucid",
               "pockets": [
                  {"architectures": [],
                   "components": ["main", "foo"],
                   "creation_time": "2011-07-19T12:51:18Z",
                   "mirror_suite": "suite",
                   "mirror_uri": "http://example.com",
                   "mode": "mirror",
                   "name": "other"},
                  {"architectures": ["i386", "source"],
                   "components": [],
                   "creation_time": "2011-07-19T12:51:18Z",
                   "mode": "upload",
                   "name": "updates"}]},
              {"creation_time": "2011-07-19T12:51:18Z",
               "name": "natty",
               "pockets": []}]}]
    """

    schema = Schema(
        name="GetDistributions",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc=(
                    "A list of distribution names to get info for. If "
                    "this is not provided, the call will return all "
                    "distributions for the account."
                ),
            )
        ],
        errors=[],
        doc="""Get info about distributions.""",
    )

    requires = "ViewAccount"

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        result = context.get_distributions(names=call.args.names or None)
        return [distribution.get_state() for distribution in result]
