import logging
import os
import re
from collections import OrderedDict
from operator import itemgetter
from time import time

import treq
from bs4 import BeautifulSoup
from twisted.internet.defer import ensureDeferred
from twisted.web.error import SchemeNotSupported
from twisted.internet.error import DNSLookupError

from canonical.landscape.api.error import LandscapeAPIError
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.routes import AccountRoute
from canonical.txapi.server.method import method
from canonical.txapi.server.schema import Schema, Unicode


REPO_CACHE = OrderedDict()
CACHE_MAXSIZE = 128


async def get_repos_cached(url):
    """
    Cache that calls get_repos, expiring after n seconds, and ignores right url slash
    """

    seconds = 60 * 5  # 5 minutes
    cache = REPO_CACHE

    key = "{}-{}".format(time() // seconds, url.strip("/"))

    # Key will expire every n seconds. For example
    # 10 // 5 = 2
    # ...
    # 13 // 5 = 2
    # 14 // 5 = 2
    # 15 // 5 = 3

    if key not in cache:
        cache[key] = await get_repos(url)
        if len(cache) > CACHE_MAXSIZE:
            cache.popitem(last=False)

    return cache[key]


async def get_repos(mirror_uri):
    """Gets info from the repo directories and release file"""

    mirror_uri = mirror_uri.strip("/")
    dists_path = os.path.join(mirror_uri, "dists")

    result = {"flat": False, "ubuntu": False, "repos": []}
    repos = result["repos"]

    # Check if flat repo
    raw_metadata = await download_raw_metadata(mirror_uri)
    if raw_metadata:  # Flat repo
        metadata = get_clean_metadata(raw_metadata, mirror_uri)
        result["flat"] = True
        repos.append(metadata)
        return result

    try:  # Check if an ubuntu or debian repo
        html_text = await get_request(dists_path)
        root_path = dists_path
    except RepoRequestError:  # Not a debian or ubuntu repo
        try:
            html_text = await get_request(mirror_uri)
            root_path = mirror_uri
        except RepoRequestError:  # Not a valid URL
            return result

    dirnames = get_dirnames(html_text)
    max_dirs = 100  # Cut it off if too long. Ubuntu archives are ~ 50 with pockets
    for dirname in dirnames[:max_dirs]:
        dirpath_url = os.path.join(root_path, dirname)
        raw_metadata = await download_raw_metadata(dirpath_url)
        if raw_metadata:
            metadata = get_clean_metadata(raw_metadata, dirpath_url)
            repos.append(metadata)

    if root_path == dists_path:  # means that this is debian or ubuntu
        if repos:
            if repos[0]["origin"] == "Ubuntu":
                result["ubuntu"] = True

    # Sort by version of ubuntu (e.g. 20.04, 22.04)
    repos.sort(key=itemgetter("version", "repo"))
    return result


async def download_raw_metadata(dirpath):
    """Returns beginning of metadata file response given repo url"""

    release_path = get_release_path(dirpath)

    # Only read beginning of file since it's quite large
    headers = {"Range": "bytes=0-500"}
    try:
        data = await get_request(release_path, headers=headers)
    except RepoRequestError:
        return None  # Not a valid mirror repo
    return data


async def get_request(url, headers={}):
    """
    Gets the text response of the url. Throws RepoRequestError when
    there's a bad response
    """
    logging.info(f"Repo Info: {url}")
    try:
        response = await treq.get(url, headers=headers)
    except (SchemeNotSupported, DNSLookupError):
        error_out(f"Invalid URL {url}")
    except Exception as exc:
        logging.error(exc)
        error_out(f"Unknown error! {url}")
    if response.code == 404:
        # We expect a 404 if it's not a flat repo
        error_out()
    if response.code < 200 or response.code > 299:
        error_out(f"Server returned {response.code}")
    text = await response.text()
    return text


def get_clean_metadata(text, repo_url):
    """
    Gets a cleaned up version of the InRelease file adding the repos or
    directory name and the url of the file
    """

    metadata = {}
    for line in text.splitlines():  # Parse metadata from release file
        if re.match(r"\w+\:", line):
            key, value = line.split(":", 1)  # Only split up to first token
            value = value.strip()
            if value:
                metadata[key] = value

    cleaned = {}
    cleaned["repo"] = os.path.basename(repo_url)
    cleaned["url"] = get_release_path(repo_url)

    # Should be architectures but sometimes it's mispelled
    if "Architectures" in metadata:
        cleaned["architectures"] = metadata["Architectures"]
    else:
        cleaned["architectures"] = metadata.get("Architecture", "")

    cleaned["codename"] = metadata.get("Codename", "")
    cleaned["description"] = metadata.get("Description", "")
    cleaned["label"] = metadata.get("Label", "")
    cleaned["origin"] = metadata.get("Origin", "")
    cleaned["version"] = metadata.get("Version", "")
    return cleaned


def get_dirnames(html_text):
    """Gets repo directory names from html, and ignores certain ones like pockets"""
    links = get_links(html_text)
    dirs = []
    for link in links:
        if not link.endswith("/"):  # Directories must end with /
            continue
        if link.startswith("http:") or link.startswith("https:"):
            continue  # This means it's a link to a webpage
        dirname = link.strip("/")
        if not dirname:
            continue
        if dirname == "devel":  # Skip dev/test directory
            continue
        if "-" in dirname:  # Takes too long to go through the pockets
            pockets = ("backports", "proposed", "security", "updates")
            if dirname.split("-")[1] in pockets:
                continue
        dirs.append(dirname)
    return dirs


def get_links(html_text):
    """Get all links from HTML"""
    soup = BeautifulSoup(html_text, "html.parser")
    return [link["href"] for link in soup.find_all("a", href=True)]


def get_release_path(repo_url):
    return os.path.join(repo_url, "InRelease")


def error_out(message=None):
    if message:
        logging.error(message)
    raise RepoRequestError(message)


class RepoInfoError(LandscapeAPIError):
    pass


class RepoRequestError(Exception):
    pass


@method
class GetRepoInfo(AccountMethod):
    """
    Gets info from the repo directories and release file

    This method takes 1 mandatory arguments:

    - **mirror**: The mirror uri of the repo

    For example, the following` uri http://archive.ubuntu.com/ubuntu/dists/

    The following errors may be raised:

    - **RepoInfoError**: When the URL is invalid
    """

    schema = Schema(
        name="GetRepoInfo",
        parameters=[
            Unicode("mirror_uri", doc="The name of mirror uri"),
        ],
        errors=[RepoInfoError],
        doc="""
Given a url to a mirror repo, returns repository information present
in the release file/s and also whether it's flat or an ubuntu repo
""",
    )

    requires = "ViewAccount"

    route = AccountRoute()

    def invoke(self, call):
        # XXX This is overridden to load the call object with additional
        # parameters before the transaction is run. It should be kept
        # in sync with the original Method.invoke() implementation.
        call.parse(self.schema)
        deferred = self.deferred_get_repo(call)
        deferred.addCallback(lambda _: self.transactor.run(self._transaction, call))
        deferred.addErrback(self._handle_errors)
        return deferred

    def deferred_get_repo(self, call):
        """Converts the async get_repo into a deferred and stashes the result"""

        coro = get_repos_cached(call.args.mirror_uri)
        deferred = ensureDeferred(coro)

        def collect(result):
            self.result = result

        deferred.addBoth(collect)
        return deferred

    def query(self, context, call):

        result = self.result
        if not result["repos"]:
            raise RepoInfoError("No repo found at {}".format(call.args.mirror_uri))
        return result
