# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from storm.twisted.transact import transact

from canonical.txapi.server.exception import APIError

from canonical.reprepro.reprepro import RepreproError

from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.model.gpg.management import GPGKeyManagement
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.model.repository.distribution import LockError
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.configuration import (
    is_testing_environment,
    is_standalone,
)


class RepositoryMethod(AccountMethod):
    """Used by repository methods as a base-class."""

    def is_available(self):
        """Repository management is only available for developers and LDS."""
        return is_testing_environment() or is_standalone()

    def get_distributions_by_names(self, context, names=None):
        """Return the L{Distribution}s with the given C{names}.

        @param context: The L{RepositoryManagement} for the L{Account}.
        @param names: The names of the distributions to return, if C{None}
            all distributions in the account will be returned.

        @raises APIError: If names is not C{None} and some of its element
            do not result in actual distributions.
        """
        distributions = list(context.get_distributions(names=names))
        return self.check_unknown_names(names, distributions, "distribution")

    def get_series_by_names(self, distribution, names=None):
        """Return the L{Series}s with the given C{names}.

        @param distribution: The L{Distribution} to search in.
        @param names: The names of the series to return, if C{None} all series
            in the distribution will be returned.

        @raises APIError: If names is not C{None} and some of its element
            do not result in actual series.
        """
        series = list(distribution.get_series(names=names))
        return self.check_unknown_names(names, series, "series")

    def get_pockets_by_names(self, series, names=None):
        """Return the L{Pocket}s with the given C{names}.

        @param series: The L{Series} to search in.
        @param names: The names of the pockets to return, if C{None}
            all pockets in the series will be returned.

        @raises APIError: If names is not C{None} and some of its element
            do not result in actual pockets.
        """
        pockets = list(series.get_pockets(names=names))
        return self.check_unknown_names(names, pockets, "pocket")

    def get_profiles_by_names(self, context, names=None):
        """Return the L{RepositoryProfile}s with the given C{names}.

        @param context: The L{RepositoryManagement} for the L{Account} to
            search in.
        @param names: The names of the profiles to return, if C{None}
            all profiles in the account will be returned.

        @raises APIError: If names is not C{None} and some of its element
            do not result in actual profiles.
        """
        profiles = list(context.get_profiles(names=names))
        return self.check_unknown_names(names, profiles, "repository profile")

    def get_gpg_keys_by_names(self, account, person, names=None):
        """Return the L{GPGKeys}s with the given C{names}.

        @param account: The L{Account} the GPG keys belongs to.
        @param account: The L{Person} makeing the API call.
        @param names: The names of the keys to return, if C{None}
            all distributions in the account will be returned.

        @raises APIError: If names is not C{None} and some of its element
            do not result in actual distributions.
        """
        management = GPGKeyManagement(account, person)
        keys = list(management.get_keys(names=names))
        return self.check_unknown_names(names, keys, "GPGKey")

    def get_apt_sources_by_names(self, context, names=None):
        """Return the L{APTSource}s with the given C{names}.

        @param context: The L{RepositoryManagement} for the L{Account} to
            search in.
        @param names: The names of the sources to return, if C{None} all
            sources in the account will be returned.

        @raises APIError: If names is not C{None} and some of its element
            do not result in actual sources.
        """
        sources = list(context.get_apt_sources(names=names))
        return self.check_unknown_names(names, sources, "APTSource")


class SynchronousRepreproMethod(AccountMethod):
    """
    Base class for local operation on pockets involving calling reprepro and
    wait for it to complete.

    The call fails if another operation is in progress.
    """

    def invoke(self, call):
        deferred = super(SynchronousRepreproMethod, self).invoke(call)
        deferred.addCallback(self._with_lock)
        return deferred

    def _with_lock(self, repo_data):
        """
        Run reprepro outside of the transaction and attach a callback to unlock
        the distribution upon completion.
        """
        (reprepro, account_id, person_id, distribution_name) = repo_data
        deferred = self.run_reprepro(reprepro)
        deferred.addBoth(
            self._unlock_distribution, account_id, person_id, distribution_name
        )
        deferred.addErrback(self._operation_error)
        deferred.addCallback(self.finish)
        return deferred

    def run_reprepro(self, reprepro):
        """
        Override this to run the appropriate reprepro command.
        """
        raise NotImplementedError("run_reprepro() must be implemented")

    def finish(self, result):
        """
        Override this method to process rerepro result and return a different
        result for the API call.
        """
        return result

    def lock_distribution(self, distribution):
        try:
            distribution.lock()
        except LockError:
            raise APIError(
                400,
                code="OperationInProgress",
                message=(
                    "Another operation is already in progress " "on distribution '%s'"
                )
                % distribution.name,
            )

    @transact
    def _unlock_distribution(self, result, account_id, person_id, distribution_name):
        account = get_account_by_id(account_id)
        person = get_person_by_id(person_id)
        management = RepositoryManagement(account, person)
        distribution = management.get_distributions(names=[distribution_name]).one()
        distribution.unlock()
        return result

    def _operation_error(self, failure):
        failure.trap(RepreproError)
        raise APIError(
            400, code="OperationError", message="The requested operation failed"
        )
