# Copyright 2005-2022 Canonical Limited. All rights reserved.

from twisted.internet.defer import inlineCallbacks

from canonical.txapi.server.schema import Schema, Unicode, Bool, List
from canonical.txapi.server.method import method

from canonical.lib.text import is_valid_slug
from canonical.landscape.api.methods import FUTURE_VERSION
from canonical.landscape.api.routes import SubjectRoute
from canonical.landscape.api.error import (
    InvalidPocketModeError,
    InvalidPocketFilterError,
    GPGKeyHasNoSecretError,
    DuplicatePocketError,
    MissingParameterError,
    UnknownPocketError,
    InvalidParameterCombinationError,
    NoPocketFilteringError,
    UnknownSeriesError,
    UnknownDistributionError,
    GPGKeyNotAssociatedError,
    GPGKeyAlreadyAssociatedError,
    UnknownGPGKeyError,
)
from canonical.landscape.api.repository.methods import (
    RepositoryMethod,
    SynchronousRepreproMethod,
)
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.package.package import is_valid_package_name
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.model.repository.distribution import LockError
from canonical.landscape.model.repository.pocket import (
    FILTER_WHITELIST,
    FILTER_BLACKLIST,
)
from canonical.landscape.model.activity.parent import get_parent_activity


class PocketMethod(RepositoryMethod):
    """
    Base class for operations on repository pockets.

    Provide convenience methods to retrieve the pocket from the call
    parameters and perform sanity checks on pocket mode.
    """

    schema = Schema(
        name="PocketMethod",
        parameters=[
            Unicode("name", doc="The name of the pocket."),
            Unicode("series", doc="The name of the series."),
            Unicode("distribution", doc="The name of the distribution."),
        ],
        errors=[UnknownDistributionError, UnknownSeriesError, InvalidPocketModeError],
    )

    pocket_modes = None

    requires = "EditPocket"

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_distribution",
        identifier_arg="distribution",
        unknown_error_factory=UnknownDistributionError,
    )

    def check_pocket_mode(self, pocket):
        """
        Check that the pocket is configured in one of the accepted modes for
        the operation. If C{pocket_modes} is C{None} any mode is accepted.
        """
        if self.pocket_modes is not None:
            if not any(
                getattr(pocket, f"is_in_{mode}_mode")() for mode in self.pocket_modes
            ):
                raise InvalidPocketModeError(
                    "Pocket is not configured in a valid mode for this operation"
                )

        return True

    def get_pocket(self, distribution, call):
        """
        Retrieves pocket specified by call parameters, raising appropriate
        errors if parameters are invalid.

        @return: the pocket
        """
        [series] = self.get_series_by_names(distribution, names=[call.args.series])
        [pocket] = self.get_pockets_by_names(series, names=[call.args.name])

        self.check_pocket_mode(pocket)
        return pocket


class SynchronousPocketMethod(PocketMethod, SynchronousRepreproMethod):
    """
    Base class for local operation on pockets involving calling reprepro and
    wait for it to complete.

    The call fails if another operation is in progress.
    """

    def query(self, context, call):
        distribution = context  # for readability
        pocket = self.get_pocket(distribution, call)

        self.lock_distribution(distribution)
        management = self.get_management(RepositoryManagement, call)
        self.process_pocket(management, pocket, call)
        return (
            pocket.get_reprepro(),
            distribution.account_id,
            call.principal.id,
            distribution.name,
        )

    def process_pocket(self, management, pocket, call):
        """
        This can be overridden to do some work on the pocket before calling
        reprepro.
        """


@method
class CreatePocket(SynchronousPocketMethod):
    """
    Create a pocket associated with a series in the account.

    This method takes the following arguments:

    - **name**: The name of the pocket. It must be unique within series, start
      with an alphanumeric character and only contain lowercase letters,
      numbers and - or + signs.
    - **series**: The name of the series to create the pocket in.
    - **distribution**: The name of the distribution the series belongs to.
    - **components.#**: A list of components the pocket will handle.
    - **architectures.#**: A list of architectures the pocket will handle.
    - **mode**: The pocket mode. Can be 'pull', 'mirror' and 'upload'.
    - **gpg_key**: The name of the GPG key to use to sign packages lists for
      this pocket. The GPG key provided must have a private key associated
      with it.
    - **include_udeb**: Whether the pocket should include selected components
      also for .udeb packages (debian-installer). It's 'false' by default.
    - **mirror_uri**: In 'mirror' mode, the base URI of the repository to
      mirror.
    - **mirror_suite**: In 'mirror' mode, the specific sub-directory under
      dists/ that should be mirrored. If the suite name ends with a '/', the
      remote repository is flat (no dists/ structure, see
      http://wiki.debian.org/RepositoryFormat#Flat_Repository_Format); in this
      case a single value must be passed for the 'components' parameter.
      Packages from the remote repository will be mirrored in the specified
      component. This parameter is optional and defaults to the same name as
      local series and pocket.
    - **mirror_gpg_key**: The name of the GPG key to use to verify the mirrored
      archive signature. If none is given, the stock Ubuntu archive one will be
      used.
    - **pull_pocket**: The name of a pocket in current distribution to sync
      packages from for pockets in 'pull' mode.
    - **pull_series**: The name of the series pull_pocket belongs to. Must be a
      series in the same distribution series belongs to. If not specified, it
      defaults to series.
    - **filter_type**: If specified, the type of the filter of the pocket. Can
      be either 'whitelist' or 'blacklist'.
    - **upload_allow_unsigned**: For pockets in upload mode, a boolean
      indicating whether uploaded packages are required to be signed or not.
      It's 'false' by default.

    For example, the following request creates a pocket named 'updates' in the
    'natty-staging' series of the 'ubuntu' distribution, with packages for
    'main' and 'universe' components for 'i386' architecture. The pocket will
    be in 'pull' mode from the pocket 'updates' of the 'natty' series.
    The latter will be typically configured to fetch packages from an ubuntu
    mirror::

        ?action=CreatePocket&name=updates&series=natty-staging
           &distribution=ubuntu&components.1=main&components.2=universe
           &architectures.1=i386&mode=pull&gpg_key=mykey
           &pull_pocket=updates&pull_series=natty&filter_type=whitelist

    The following errors may be raised:

    - **InvalidPocketMode**: The mode parameter has an invalid value.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **DuplicatePocket**: A pocket with same name already exists in the
      specified series.
    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **GPGKeyHasNoSecret**: The GPG key requested for signing does not have a
      secret key associated.
    - **OperationInProgress**: Another operation is in progress on the
      distribution.
    - **InvalidFilter**: Invalid filter type.
    - **InvalidParameterCombinationError**: The given parameters are not compatible.

    The method returns JSON serialized info of the created pocket::

        {"name": "updates",
         "architectures": ["i386"],
         "creation_time": "2011-07-27T09:07:59Z",
         "components": ["main", "universe"],
         "mode": "pull",
         "gpg_key": "mykey",
         "pull_pocket": "natty/updates"}
    """

    schema = Schema(
        name="CreatePocket",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    "The name of the pocket. It must be unique within "
                    "series, start with an alphanumeric character and only "
                    "contain lowercase letters, numbers and - or + signs."
                ),
            ),
            Unicode("series", doc="The name of the series to create the pocket in."),
            Unicode(
                "distribution",
                doc="The name of the distribution the series belongs to.",
            ),
            List(
                "components",
                Unicode(),
                doc="A list of components the pocket will handle.",
            ),
            List(
                "architectures",
                Unicode(),
                doc="A list of architectures the pocket will handle.",
            ),
            Unicode(
                "mode", doc="The pocket mode. Can be 'pull', 'mirror' and 'upload'."
            ),
            Unicode(
                "gpg_key",
                doc=(
                    "The name of the GPG key to use to sign packages lists "
                    "for this pocket. The GPG key provided must have a "
                    "private key associated with it."
                ),
            ),
            Bool(
                "include_udeb",
                optional=True,
                default=False,
                doc=(
                    "Whether the pocket should include selected components "
                    "also for .udeb packages (debian-installer). "
                    "It's 'false' by default."
                ),
            ),
            Unicode(
                "mirror_uri",
                optional=True,
                doc="The URI to mirror for pockets in 'mirror' mode.",
            ),
            Unicode(
                "mirror_suite",
                optional=True,
                doc=(
                    "The repository entry under dists/ to mirror for pockets "
                    "in 'mirror' mode. This parameter is optional and "
                    "defaults to the same name as local series and pocket. "
                    "If the suite name ends with a '/', the remote "
                    "repository is flat (packages are not grouped in "
                    "components); in this case a single value can be passed "
                    "for the 'components' parameter. Packages from the "
                    "remote repository will be mirrored in the specified "
                    "component."
                ),
            ),
            Unicode(
                "mirror_gpg_key",
                optional=True,
                doc=(
                    "The name of the GPG key to use to verify the mirrored "
                    "archive signature. If none is given, the stock Ubuntu "
                    "archive one will be used."
                ),
            ),
            Unicode(
                "pull_series",
                optional=True,
                doc=(
                    "The name of the series pull_pocket belongs to. Must be "
                    "a series in the same distribution series belongs to. If "
                    "not specified, it defaults to series."
                ),
            ),
            Unicode(
                "pull_pocket",
                optional=True,
                doc=(
                    "The name of a pocket in current distribution to sync "
                    "packages from for pockets in 'pull' mode."
                ),
            ),
            Unicode(
                "filter_type",
                optional=True,
                doc=(
                    "If specified, the type of the filter of the pocket. Can "
                    "be either 'whitelist' or 'blacklist'."
                ),
            ),
            Bool(
                "upload_allow_unsigned",
                optional=True,
                default=False,
                doc=(
                    "For pockets in upload mode, a boolean indicating "
                    "whether uploaded packages are required to be signed or "
                    "not. It's 'false' by default."
                ),
            ),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            InvalidPocketModeError,
            DuplicatePocketError,
            UnknownPocketError,
            UnknownGPGKeyError,
            GPGKeyHasNoSecretError,
            LockError,
            InvalidPocketFilterError,
            MissingParameterError,
            InvalidParameterCombinationError,
        ],
        doc="""Create a pocket associated with a series in the account.""",
    )

    requires = "CreatePocket"

    def query(self, context, call):
        distribution = context  # for readability
        name = call.args.name
        mode = call.args.mode
        filter_type = call.args.filter_type

        if filter_type:
            if mode and mode != "pull":
                raise InvalidPocketModeError(
                    "filter_type cannot be used in 'mirror' or 'upload' mode"
                )
            filter_type = filter_type.lower()
            if filter_type == "whitelist":
                filter_type = FILTER_WHITELIST
            elif filter_type == "blacklist":
                filter_type = FILTER_BLACKLIST
            else:
                raise InvalidPocketFilterError(
                    f"The filter type '{call.args.filter_type}' is invalid"
                )

        [series] = self.get_series_by_names(distribution, names=[call.args.series])

        person = get_person_by_id(call.principal.id)
        [gpg_key] = self.get_gpg_keys_by_names(
            distribution.account, person, names=[call.args.gpg_key]
        )

        if gpg_key.secret_material is None:
            raise GPGKeyHasNoSecretError(
                f"gpg_key '{call.args.gpg_key}' has no secret key"
            )

        pocket = series.get_pockets(names=[name]).one()
        if pocket is not None:
            raise DuplicatePocketError(f"Pocket '{name}' already present")

        self._check_mode_params(call)

        pocket = series.create_pocket(
            name=name,
            gpg_key=gpg_key,
            components=list(call.args.components),
            archs=list(call.args.architectures),
            filter_type=filter_type,
            include_udeb=call.args.include_udeb,
        )

        if mode == "mirror":
            if call.args.mirror_suite and call.args.mirror_suite.endswith("/"):
                # If the suite ends with "/", this is a flat repository.
                if len(call.args.components) > 1:
                    raise InvalidParameterCombinationError(
                        "A single component must be specified for flat repositories"
                    )
            self._set_mirror_mode(pocket, call, distribution.account, person)
        elif mode == "pull":
            self._set_pull_mode(pocket, call, distribution, filter_type)
        else:
            self._set_upload_mode(pocket, call, distribution.account)

        self.lock_distribution(distribution)
        pocket_state = pocket.get_state()
        self.finish = lambda _: pocket_state
        return (
            pocket.get_reprepro(),
            distribution.account_id,
            call.principal.id,
            distribution.name,
        )

    def run_reprepro(self, reprepro):
        """
        Run the reprepro 'export' command to create the distribution tree.
        """
        return reprepro.export()

    def _check_mode_params(self, call):
        """
        Check that the passed parameters are suitable for the selected mode.
        """

        # Map modes to their required parameters and the invalid ones
        params_by_mode = {
            "mirror": (
                ("mirror_uri",),
                ("pull_series", "pull_pocket", "upload_allow_unsigned"),
            ),
            "pull": (
                ("pull_pocket",),
                (
                    "mirror_uri",
                    "mirror_suite",
                    "mirror_gpg_key",
                    "upload_allow_unsigned",
                ),
            ),
            "upload": (
                (),
                (
                    "mirror_uri",
                    "mirror_suite",
                    "mirror_gpg_key",
                    "pull_pocket",
                    "pull_series",
                ),
            ),
        }

        mode = call.args.mode
        if mode not in params_by_mode:
            raise InvalidPocketModeError(f"Pocket mode '{mode}' is invalid")

        required_params, invalid_params = params_by_mode[mode]

        for param in required_params:
            if not getattr(call.args, param):
                raise MissingParameterError(
                    f"The request must contain the parameter {param} for '{mode}' mode"
                )
        for param in invalid_params:
            if getattr(call.args, param):
                raise InvalidParameterCombinationError(
                    f"Can't specify '{param}' with this mode"
                )

    def _set_mirror_mode(self, pocket, call, account, person):
        """Set mirror mode on the created pocket."""
        if call.args.mirror_gpg_key:
            [mirror_gpg_key] = self.get_gpg_keys_by_names(
                account, person, names=[call.args.mirror_gpg_key]
            )
        else:
            mirror_gpg_key = None

        suite = (
            call.args.mirror_suite if call.args.mirror_suite else pocket.get_full_name()
        )
        pocket.set_mirror_mode(call.args.mirror_uri, suite, mirror_gpg_key)

    def _set_pull_mode(self, pocket, call, distribution, filter_type):
        """Set pull mode on the created pocket."""
        if call.args.pull_series:
            [pull_series] = self.get_series_by_names(
                distribution, names=[call.args.pull_series]
            )
        else:
            pull_series = pocket.series
        [pull_pocket] = self.get_pockets_by_names(
            pull_series, names=(call.args.pull_pocket,)
        )
        pocket.set_pull_mode(pull_pocket, filter_type)

    def _set_upload_mode(self, pocket, call, account):
        """Set upload mode on the created pocket."""
        pocket.set_upload_mode(allow_unsigned=call.args.upload_allow_unsigned)


@method
class SyncMirrorPocket(PocketMethod):
    """
    Synchronize a mirror repository pocket.

    This method takes 3 mandatory arguments:

    - **name**: The name of the pocket to synchronize.
    - **series**: The name of the series containing the pocket.
    - **distribution**: The name of the distribution containing the series.

    For example, the following request syncs a pocket named 'updates' in the
    'natty' series of the 'ubuntu' distribution::

        ?action=SyncMirrorPocket&name=updates&series=natty
           &distribution=ubuntu

    The following errors may be raised:

    - **UnknownDistribution**: No distribution with specified name exists.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'mirror' mode.
    """

    requires = "SyncPocket"

    schema = Schema(
        name="SyncMirrorPocket",
        parameters=[
            Unicode("name", doc="The name of the pocket to synchronize."),
            Unicode("series", doc="The name of the series."),
            Unicode("distribution", doc="The name of the distribution."),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            InvalidPocketModeError,
            LockError,
        ],
        doc="""Synchronize a mirror repository pocket.""",
    )

    pocket_modes = ["mirror"]

    def query(self, context, call):
        distribution = context  # for readability
        pocket = self.get_pocket(distribution, call)
        distribution.lock()
        person = get_person_by_id(call.principal.id)
        handler = self.get_action_handler(
            "sync-mirror-pocket", context=distribution.account, person=person
        )
        state = handler.handle({"pocket_id": pocket.id})

        v2 = call.version == FUTURE_VERSION
        if not v2:
            state["children"] = [
                self._get_child_state(distribution.account, state["id"])
            ]

        return state

    def _get_child_state(self, account, parent_id):
        """Return the state of the child activity."""
        parent = get_parent_activity(account, parent_id)
        [child] = parent.children()
        return child.get_state()


@method
class PullPackagesToPocket(SynchronousPocketMethod):
    """
    Import packages to a pocket in pull mode from its parent pocket.

    This method takes 3 mandatory arguments:

    - **name**: The name of the pocket to pull packages to.
    - **series**: The name of the series containing the pocket.
    - **distribution**: The name of the distribution containing the series.

    For example, the following request pulls packages to a pocket named
    'updates' in the 'natty-staging' series of the 'ubuntu' distribution::

        ?action=PullPackagesToPocket&name=updates&series=natty-staging
           &distribution=ubuntu

    The following errors may be raised:

    - **UnknownDistribution**: No distribution with specified name exists.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'pull' mode.
    - **OperationInProgress**: Another operation is in progress on the
      distribution containing the pocket.
    """

    requires = "SyncPocket"

    schema = Schema(
        name="PullPackagesToPocket",
        parameters=[
            Unicode("name", doc="The name of the pocket to pull packages to."),
            Unicode("series", doc="The name of the series."),
            Unicode("distribution", doc="The name of the distribution."),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            InvalidPocketModeError,
            LockError,
        ],
        doc="""
    Import packages to a pocket in pull mode from its parent pocket.
    """,
    )

    pocket_modes = ["pull"]

    def run_reprepro(self, reprepro):
        """Run the reprepro 'pull' command."""
        return reprepro.pull()

    def finish(self, result):
        """Return an empty JSON-valid response."""
        return {}


@method
class RemovePackagesFromPocket(SynchronousPocketMethod):
    """
    Remove packages from pockets in upload mode.

    This method takes 4 mandatory arguments:

    - **distribution**: The name of the distribution containing the series.
    - **series**: The name of the series containing the pocket.
    - **name**: The name of the pocket to remove packages from.
    - **packages.#**: The names of the packages to be removed from the pockets.

    For example, the following request removes packages named haskel
    and golang from a pocket named 'updates' in the 'natty-staging' series of
    the 'ubuntu' distribution::

        ?action=RemovePackagesFromPockets&name=updates&series=natty-staging
           &distribution=ubuntu&packages.1=haskel&packages.2=golang

    The following errors may be raised:

    - **UnknownDistribution**: No distribution with specified name exists.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'upload' mode.
    - **OperationInProgress**: Another operation is in progress on the
      distribution containing the pocket.

    """

    requires = "EditPocket"

    schema = Schema(
        name="RemovePackagesFromPocket",
        parameters=[
            Unicode("name", doc="The name of the pocket to remove packages from."),
            Unicode("series", doc="The name of the series containing the pocket."),
            Unicode(
                "distribution",
                doc="The name of the distribution containing the series.",
            ),
            List(
                "packages",
                Unicode(),
                doc=("A list of names of packages to be removed from the pockets."),
            ),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            InvalidPocketModeError,
            UnknownPocketError,
            LockError,
        ],
        doc="""Remove packages from pockets in upload mode.""",
    )

    pocket_modes = ["upload"]

    def _with_lock(self, repo_data):
        """
        Run reprepro outside of the transaction and attach a callback to unlock
        the distribution upon completion.
        """
        (reprepro, account_id, person_id, distribution_name, packages) = repo_data
        deferred = self.run_reprepro(reprepro, packages)
        deferred.addBoth(
            self._unlock_distribution, account_id, person_id, distribution_name
        )
        deferred.addErrback(self._operation_error)
        deferred.addCallback(self.finish)
        return deferred

    def query(self, context, call):
        distribution = context  # for readability
        packages = call.args.packages
        pocket = self.get_pocket(distribution, call)
        self.lock_distribution(distribution)
        return (
            pocket.get_reprepro(),
            distribution.account_id,
            call.principal.id,
            distribution.name,
            packages,
        )

    def run_reprepro(self, reprepro, packages):
        return reprepro.remove(packages)

    def finish(self, result):
        """Return an empty JSON-valid response."""
        return {}


@method
class DiffPullPocket(SynchronousPocketMethod):
    """
    Return a list of the changes between a pocket configured in pull mode and
    its origin one.

    This method takes 3 mandatory arguments:

    - **name**: The name of the pocket.
    - **series**: The name of the series containing the pocket.
    - **distribution**: The name of the distribution containing the series.

    For example, the following request the diff of a pocket named 'updates' in
    the 'lucid-staging' series of the 'ubuntu' distribution::

        ?action=DiffPullPocket&name=updates&series=lucid-staging
           &distribution=ubuntu

    The following errors may be raised:

    - **UnknownDistribution**: Distribution with specified name doesn't exist.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'pull' mode.
      Can only diff pocket configured in 'pull mode' with their origin
      pocket.
    - **OperationInProgress**: Another operation is in progress on the
      distribution containing the pocket.

    The method returns JSON serialized info of changed packages between
    pockets for each component/architecture pair::

        {"main/i386": {
            "update": [["package1", "1.0.0", "1.0.1"],
                       ["package2", "3.2", "3.3"]],
            "delete": [["package3", "1.1"]]},
         "multiverse/amd64": {
            "update": [["package4", "1.3", "1.4"]],
            "add": [["package5", "1.1"]]}}
    """

    requires = "ViewPocket"

    schema = Schema(
        name="DiffPullPocket",
        parameters=[
            Unicode("name", doc="The name of the pocket."),
            Unicode("series", doc="The name of the series."),
            Unicode("distribution", doc="The name of the distribution."),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            InvalidPocketModeError,
            LockError,
        ],
        doc="""
    Return a list of the changes between a pocket configured in pull mode and
    its origin one.
    """,
    )

    pocket_modes = ["pull"]

    def run_reprepro(self, reprepro):
        """Run the 'dumppull' command."""
        return reprepro.dumppull()


@method
class ListPocket(SynchronousPocketMethod):
    """
    Return a list of the packages in a pocket.

    This method takes 3 mandatory arguments:

    - **name**: The name of the pocket.
    - **series**: The name of the series containing the pocket.
    - **distribution**: The name of the distribution containing the series.

    For example, the following request the list of packages of a pocket named
    'updates' in the 'lucid-staging' series of the 'ubuntu' distribution::

        ?action=ListPocket&name=updates&series=lucid-staging
           &distribution=ubuntu

    The following errors may be raised:

    - **UnknownDistribution**: Distribution with specified name doesn't exist.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **OperationInProgress**: Another operation is in progress on the
      distribution containing the pocket.

    The method returns JSON serialized info of the packages in the pocket, for
    each section and architecture::

        {"main/i386":
            [["package1", "1.0.0"],
             ["package2", "3.2"]],
         "multiverse/amd64":
            [["package4", "1.3"],
             ["package5", "1.1"]]}
    """

    requires = "ViewPocket"

    schema = Schema(
        name="ListPocket",
        parameters=[
            Unicode("name", doc="The name of the pocket."),
            Unicode("series", doc="The name of the series."),
            Unicode("distribution", doc="The name of the distribution."),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            LockError,
        ],
        doc="""Return a list of the packages in a pocket.""",
    )

    def run_reprepro(self, reprepro):
        """Run the 'list' command."""
        return reprepro.list()


@method
class EditPocket(SynchronousPocketMethod):
    """
    Edit configuration for a repository pocket from a series in a distribution.
    Provided details will overwrite the current ones.

    This method takes the following arguments:

    - **name**: The name of the pocket to edit.
    - **series**: The name of the series containing the pocket.
    - **distribution**: The name of the distribution containing the series.
    - **components.#**: An optional list of components the pocket will
      handle. If provided, it will replace the current components.
    - **architectures.#**: An optional list of architectures the pocket will
      handle. If provided, it will replace the current architectures.
    - **gpg_key**: The name of the GPG key to use to sign packages lists for
      this pocket. The GPG key provided must have a private key associated
      with it.
    - **mirror_uri**: Optionally, the URI to mirror for pockets in 'mirror'
      mode.
    - **mirror_suite**: Optionally, the repository entry under dists/ to mirror
      for pockets in 'mirror' mode.
    - **mirror_gpg_key**: Optionally, the name of the GPG key to use to verify
      the mirrored archive signature. If '-' is given, the stock Ubuntu archive
      one will be used.
    - **upload_allow_unsigned**: Optionally, for pockets in upload mode, a
      boolean indicating whether uploaded packages are required to be signed or
      not.
    - **include_udeb**: Optionally, whether the pocket should include selected
      components, also for .udeb packages (debian-installer).

    For example, the following request updates the 'mirror_uri' and
    'mirror-suite' parameters for a pocket named 'updates' from a series named
    'lucid' for the 'ubuntu' distribution::

        ?action=EditPocket&name=updates&series=natty
           &distribution=ubuntu&mirror_uri=http://example.com/mirror/ubuntu
           &mirror_suite=natty-updates

    The following errors may be raised:

    - **UnknownDistribution**: No distribution with specified name exists.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **GPGKeyHasNoSecret**: The GPG key requested for signing does not have a
      secret key associated.
    - **OperationInProgress**: Another operation is in progress on the
      distribution containing the pocket.
    - **InvalidPocketMode**: Provided parameters are not valid for the pocket
      mode.

    The method returns JSON serialized info of the pocket state::

        {"name": "updates",
         "architectures": ["i386"],
         "creation_time": "2011-07-27T09:07:59Z",
         "components": ["main", "universe"],
         "mode": "mirror",
         "mirror_suite": "natty-updates",
         "mirror_uri": "http://example.com/mirror/ubuntu",
         "gpg_key": "mykey"}

    """

    requires = "EditPocket"

    schema = Schema(
        name="EditPocket",
        parameters=[
            Unicode("name", doc="The name of the pocket to edit."),
            Unicode("series", doc="The name of the series containing the pocket."),
            Unicode(
                "distribution",
                doc="The name of the distribution containing the series.",
            ),
            List(
                "components",
                Unicode(),
                optional=True,
                doc="A list of components the pocket will handle.",
            ),
            List(
                "architectures",
                Unicode(),
                optional=True,
                doc="A list of architectures the pocket will handle.",
            ),
            Unicode(
                "gpg_key",
                optional=True,
                doc=(
                    "The name of the GPG key to use to sign packages lists "
                    "for this pocket. The GPG key provided must have a "
                    "private key associated with it."
                ),
            ),
            Unicode(
                "mirror_uri",
                optional=True,
                doc="The URI to mirror for pockets in 'mirror' mode.",
            ),
            Unicode(
                "mirror_suite",
                optional=True,
                doc=(
                    "The repository entry under dists/ to mirror for pockets "
                    "in 'mirror' mode."
                ),
            ),
            Unicode(
                "mirror_gpg_key",
                optional=True,
                doc=(
                    "The name of the GPG key to use to verify the mirrored "
                    "archive signature. If '-' is given, the stock Ubuntu "
                    "archive one will be used."
                ),
            ),
            Bool(
                "upload_allow_unsigned",
                optional=True,
                doc=(
                    "For pockets in upload mode, a boolean indicating "
                    "whether uploaded packages are required to be signed "
                    "or not."
                ),
            ),
            Bool(
                "include_udeb",
                optional=True,
                doc=(
                    "Whether the pocket should include selected components "
                    "also for .udeb packages (debian-installer). "
                ),
            ),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            InvalidPocketModeError,
            LockError,
            GPGKeyHasNoSecretError,
        ],
        doc=(
            "Edit configuration for a repository pocket from a series in "
            "a distribution."
        ),
    )

    @inlineCallbacks
    def run_reprepro(self, reprepro):
        """
        Run the reprepro 'clearvanished' command in case some components or
        architectures are gone, and export indices in case the sign key has
        changed.
        """
        yield reprepro.clearvanished()
        yield reprepro.export()

    def _check_mode_params(self, pocket, call):
        """
        Check that provided parameters are consistent with pocket mode.
        """
        params_by_mode = {
            "mirror": ["mirror_uri", "mirror_suite", "mirror_gpg_key"],
            "upload": ["upload_allow_unsigned"],
        }

        for mode, params in params_by_mode.items():
            for param in params:
                value = getattr(call.args, param)
                right_mode = getattr(pocket, f"is_in_{mode}_mode")()

                if value is not None and not right_mode:
                    raise InvalidPocketModeError(
                        f"Can't set '{param}' for a pocket not in '{mode}' mode."
                    )

    def process_pocket(self, management, pocket, call):
        """
        Update pocket configuration.
        """
        self._check_mode_params(pocket, call)

        params = ("mirror_uri", "mirror_suite", "upload_allow_unsigned")
        for param in params:
            value = getattr(call.args, param)
            if value is not None:
                setattr(pocket, param, value)
        if call.args.architectures:
            pocket.archs = call.args.architectures
        if call.args.components:
            pocket.components = call.args.components
        if call.args.include_udeb is not None:
            pocket.include_udeb = call.args.include_udeb

        if call.args.gpg_key is not None:
            [gpg_key] = self.get_gpg_keys_by_names(
                management.account, management.person, names=[call.args.gpg_key]
            )

            if gpg_key.secret_material is None:
                raise GPGKeyHasNoSecretError(
                    "gpg_key '%s' has no secret key" % call.args.gpg_key
                )

            pocket.gpg_key = gpg_key

        mirror_gpg_key = call.args.mirror_gpg_key
        if mirror_gpg_key == "-":
            pocket.mirror_gpg_key = None
        elif mirror_gpg_key is not None:
            [mirror_key] = self.get_gpg_keys_by_names(
                management.account, management.person, names=[mirror_gpg_key]
            )
            pocket.mirror_gpg_key = mirror_key

        pocket_state = pocket.get_state()
        # Return the updated pocket state.
        self.finish = lambda _: pocket_state


@method
class RemovePocket(SynchronousPocketMethod):
    """
    Remove a repository pocket from a series in a distribution.

    This method takes 3 mandatory arguments:

    - **name**: The name of the pocket to remove.
    - **series**: The name of the series containing the pocket.
    - **distribution**: The name of the distribution containing the series.

    For example, the following request removes a pocket named 'updates' from a
    series named 'lucid' for the 'ubuntu' distribution::

        ?action=RemovePocket&name=updates&series=natty
           &distribution=ubuntu

    The following errors may be raised:

    - **UnknownDistribution**: No distribution with specified name exists.
    - **UnknownSeries**: No series with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists.
    - **OperationInProgress**: Another operation is in progress on the
      distribution containing the pocket.
    """

    requires = "EditPocket"

    schema = Schema(
        name="RemovePocket",
        parameters=[
            Unicode("name", doc="The name of the pocket to remove."),
            Unicode("series", doc="The name of the series containing the pocket."),
            Unicode(
                "distribution",
                doc="The name of the distribution containing the series.",
            ),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            LockError,
        ],
        doc="""Remove a repository pocket from a series in a distribution.""",
    )

    def run_reprepro(self, reprepro):
        """Run the reprepro 'clearvanished' command."""
        return reprepro.clearvanished()

    def process_pocket(self, management, pocket, call):
        """
        Remove pocket and unlink it from associated L{RepositoryProfile}s.
        """
        management.remove_pocket(pocket)

    def finish(self, result):
        """Return an empty JSON-valid response."""
        return {}


class ManageUploaderGPGKeysMethod(PocketMethod):
    """
    Base class for methods which manage GPG keys association/disassociation
    with repository pockets
    """

    requires = "EditPocket"

    schema = Schema(
        name="ManageUploaderGPGKeysMethod",
        parameters=[
            Unicode("name", doc="The name of the pocket on which to associate keys."),
            Unicode("series", doc="The name of the series containing the pocket."),
            Unicode(
                "distribution",
                doc="The name of the distribution containing the series.",
            ),
            List("gpg_keys", Unicode(), doc="A list of GPG keys on which to operate."),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            UnknownPocketError,
            InvalidPocketModeError,
            UnknownGPGKeyError,
            LockError,
            GPGKeyAlreadyAssociatedError,
            GPGKeyNotAssociatedError,
            InvalidPocketModeError,
        ],
        doc="""
    Base class for methods which manage GPG keys association/disassociation
    with repository pockets
    """,
    )

    pocket_modes = ["upload"]

    def query(self, context, call):
        distribution = context  # for readability
        pocket = self.get_pocket(distribution, call)
        person = get_person_by_id(call.principal.id)
        keys = self.get_gpg_keys_by_names(
            distribution.account, person, names=call.args.gpg_keys
        )

        if pocket.upload_allow_unsigned:
            raise InvalidPocketModeError(
                "Pocket is configured to allow unsigned uploads"
            )

        self.manage_uploader_keys(pocket, keys)
        return pocket.get_state()

    def manage_uploader_keys(self, pocket, keys):
        """
        Override this to perform the actual operations on pocket.
        """
        raise NotImplementedError("manage_uploader_keys() must be implemented")


@method
class AddUploaderGPGKeysToPocket(ManageUploaderGPGKeysMethod):
    """
    Add GPG keys to a repository pocket in upload mode to validate uploaded
    packages.

    - **name**: The name of the pocket.
    - **series**: The name of the series the pocket belongs to.
    - **distribution**: The name of the distribution the series belongs to.
    - **gpg_keys.#**: The name(s) of the GPG keys to add.

    Example of a valid request::

        ?action=AddUploaderGPGKeysToPocket&name=devel
           &series=lucid&distribution=ubuntu&gpg_keys.1=mykey

    The following errors may be raised:

    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'upload' mode.
    - **GPGKeyAlreadyAssociated**: GPG key is already added to pocket.

    The method returns JSON serialized info of the pocket::

        {"name": "devel",
         "architectures": ["i386"],
         "creation_time": "2011-07-27T09:07:59Z",
         "components": ["main", "universe"],
         "mode": "upload",
         "upload_gpg_keys": [
            {"id": 11,
            "name": "key",
            "fingerprint": "a404:34a3:e40c:1add:94fa:31b4:30a7:5431:a2eb:521a",
            "has_secret": false}]}
    """

    schema = ManageUploaderGPGKeysMethod.schema.extend(
        name="AddUploaderGPGKeysToPocket",
        doc="""
    Add GPG keys to a repository pocket in upload mode to validate uploaded
    packages.
    """,
    )

    def manage_uploader_keys(self, pocket, keys):
        """Add GPG keys to repository pocket."""
        keys_to_add = set(keys).difference(pocket.upload_gpg_keys)
        pocket.add_upload_gpg_keys(keys_to_add)


@method
class RemoveUploaderGPGKeysFromPocket(ManageUploaderGPGKeysMethod):
    """
    Remove GPG keys for uploaded packages validation from a repository pocket
    in upload mode.

    - **name**: The name of the pocket.
    - **series**: The name of the series the pocket belongs to.
    - **distribution**: The name of the distribution the series belongs to.
    - **gpg_keys.#**: The name(s) of the GPG keys to remove.

    Example of a valid request::

        ?action=RemoveUploaderGPGKeysFromPocket&name=devel
           &series=lucid&distribution=ubuntu&gpg_keys.1=mykey

    The following errors may be raised:

    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'upload' mode.
    - **GPGKeyNotAssociated**: GPG key is not associated with pocket.

    The method returns JSON serialized info of the pocket::

        {"name": "devel",
         "architectures": ["i386"],
         "creation_time": "2011-07-27T09:07:59Z",
         "components": ["main", "universe"],
         "mode": "upload",
         "upload_gpg_keys": []}
    """

    schema = ManageUploaderGPGKeysMethod.schema.extend(
        name="RemoveUploaderGPGKeysFromPocket",
        doc="""
    Remove GPG keys for uploaded packages validation from a repository pocket
    in upload mode.
    """,
    )

    def manage_uploader_keys(self, pocket, keys):
        """Remove GPG keys from repository pocket."""
        keys_to_remove = set(keys).intersection(pocket.upload_gpg_keys)
        pocket.remove_upload_gpg_keys(keys_to_remove)


class ManagePocketPackagesFilter(PocketMethod):
    """
    Base class for methods to add/remove package filters to pockets.
    """

    requires = "EditPocket"

    schema = Schema(
        name="ManagePocketPackagesFilter",
        parameters=[
            Unicode("name", doc="The name of the pocket to operate on."),
            Unicode("series", doc="The name of the series containing the pocket."),
            Unicode(
                "distribution",
                doc="The name of the distribution containing the series.",
            ),
            List(
                "packages",
                Unicode(validator=is_valid_package_name),
                doc=(
                    "A list of names of packages to be added or removed from "
                    "the pocket filter."
                ),
            ),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            InvalidPocketModeError,
            UnknownPocketError,
            NoPocketFilteringError,
            LockError,
        ],
    )

    pocket_modes = ["pull"]

    def query(self, context, call):
        distribution = context  # for readability
        pocket = self.get_pocket(distribution, call)

        if pocket.filter_type is None:
            raise NoPocketFilteringError("Pocket does not support filtering")

        self.manage_pocket_filters(pocket, call.args.packages)
        return pocket.get_state()

    def manage_pocket_filters(self, pocket, packages):
        """
        Override this to perform the actual operations on pocket.
        """
        raise NotImplementedError("manage_packages_filter() must be implemented")


@method
class AddPackageFiltersToPocket(ManagePocketPackagesFilter):
    """
    Add package filters to a repository pocket.  The pocket must be in pull
    mode and support blacklist/whitelist filtering.

    - **name**: The name of the pocket.
    - **series**: The name of the series the pocket belongs to.
    - **distribution**: The name of the distribution the series belongs to.
    - **packages.#**: The name(s) of the packages to add to the filter.

    Example of a valid request::

        ?action=AddPackageFiltersToPocket&name=devel
           &series=lucid&distribution=ubuntu&packages.1=bash&packages.2=python

    The following errors may be raised:

    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'pull' mode.
    - **NoPocketFiltering**: Pocket does not support filtering.

    The method returns JSON serialized info of the pocket::

        {"name": "devel",
         "architectures": ["i386"],
         "creation_time": "2011-07-27T09:07:59Z",
         "components": ["main", "universe"],
         "mode": "pull",
         "filter_type": "whitelist",
         "filters": ["bash", "python"]}
    """

    schema = ManagePocketPackagesFilter.schema.extend(
        name="AddPackageFiltersToPocket",
        doc="""
    Add package filters to a repository pocket.  The pocket must be in pull
    mode and support blacklist/whitelist filtering.
    """,
    )

    def manage_pocket_filters(self, pocket, packages):
        """Add package filters to pocket."""
        pocket.add_filters(packages)


@method
class RemovePackageFiltersFromPocket(ManagePocketPackagesFilter):
    """
    Remove package filters from a repository pocket.  The pocket must be in
    pull mode and support blacklist/whitelist filtering.

    - **name**: The name of the pocket.
    - **series**: The name of the series the pocket belongs to.
    - **distribution**: The name of the distribution the series belongs to.
    - **packages.#**: The name(s) of the packages to remove from the filter.

    Example of a valid request::

        ?action=AddPackageFiltersToPocket&name=devel
           &series=lucid&distribution=ubuntu&packages.1=bash&packages.2=python

    The following errors may be raised:

    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **InvalidPocketMode**: Pocket is not configured in 'pull' mode.
    - **NoPocketFiltering**: Pocket does not support filtering.

    The method returns JSON serialized info of the pocket::

        {"name": "devel",
         "architectures": ["i386"],
         "creation_time": "2011-07-27T09:07:59Z",
         "components": ["main", "universe"],
         "mode": "pull",
         "filter_type": "whitelist",
         "filters": ["cron"]}
    """

    schema = ManagePocketPackagesFilter.schema.extend(
        name="RemovePackageFiltersFromPocket",
        doc="""
    Remove package filters from a repository pocket.  The pocket must be in
    pull mode and support blacklist/whitelist filtering.
    """,
    )

    def manage_pocket_filters(self, pocket, packages):
        """Remove package filters from pocket."""
        pocket.remove_filters(packages)
