# Copyright 2005-20122Canonical Limited. All rights reserved.

"""
Repository Profiles
===================

The methods available here are related to operations on repository profiles.
"""

from zope.security.interfaces import Unauthorized

from canonical.txapi.server.schema import Schema, Unicode, List
from canonical.txapi.server.method import method

from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.errors import UnknownAccessContextError
from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.methods import AccountMethod, TagAssociationMethod
from canonical.landscape.api.error import (
    UnknownPocketError,
    UnknownSeriesError,
    UnknownDistributionError,
    UnknownAPTSourceError as APIUnknownAPTSourceError,
    InvalidParameterCombinationError,
)
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.api.repository.methods import RepositoryMethod
from canonical.landscape.model.repository.aptsource import UnknownAPTSourceError
from canonical.landscape.model.repository.profile import UnknownRepositoryProfileError
from canonical.landscape.model.repository.management import RepositoryManagement


__all__ = [
    "CreateRepositoryProfile",
    "GetRepositoryProfiles",
    "EditRepositoryProfile",
    "RemoveRepositoryProfiles",
    "AssociateRepositoryProfile",
    "DisassociateRepositoryProfile",
    "AddPocketsToRepositoryProfile",
    "RemoveRepositoryProfile",
    "RemovePocketsFromRepositoryProfile",
    "AddAPTSourcesToRepositoryProfile",
    "RemoveAPTSourceFromRepositoryProfile",
    "RemoveAPTSourcesFromRepositoryProfile",
]


@method
class GetRepositoryProfiles(RepositoryMethod):
    """
    Get a list of repository profiles in the account.

    - **names**: A list of repository profile names to get info for.
      If not provided, the call will return all repository profiles in the
      account. (optional)

    Example of a valid request::

        ?action=GetRepositoryProfiles&names.1=lucid-example

    The method returns a JSON serialized list of the repository profiles::

        [{"all_computers": False,
          "description": "Lucid packages for example application",
          "id": 1104,
          "name": "lucid-example",
          "tags": ["example"]}]

    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetRepositoryProfiles",
        doc="Get a list of repository profiles in the account.",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc=(
                    "A list of repository profile names to get info for. "
                    "If this is not provided, the call will return all "
                    "repository profiles for the account."
                ),
            )
        ],
        errors=[],
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """
        Return all repository profiles in the account that match the
        call arguments.
        """
        result = context.get_profiles(names=call.args.names or None)
        return [profile.get_state() for profile in result]


@method
class CreateRepositoryProfile(AccountMethod):
    """
    Create a repository profiles in the account.

    - **title**: The title of the repository profile. It must start with an
      alphanumeric character and only contain lowercase
      letters, numbers and - or + signs.
    - **description**: Description of the repository profile. (optional)
    - **access_group**: The repository profile will be created within this
      access group.  If not specified, the access group the user belongs to
      will be used.

    Example of a valid request::

        ?action=CreateRepositoryProfile&title=Lucid+Example
            &description=Example+Lucid+profile

    The following errors may be raised:

    - **Unauthorised**: The person is not authorised to create repository
      profiles.

    The method returns a JSON representation of the created repository
    profile::

        {"all_computers": False,
          "description": "Example Lucid profile",
          "id": 1104,
          "name": "lucid-example",
          "tags": []}
    """

    requires = "ViewAccount"

    schema = Schema(
        name="CreateRepositoryProfile",
        doc="Create a repository profile.",
        errors=[UnknownAccessContextError, Unauthorized],
        parameters=[
            UnicodeTitle(
                "title",
                doc=(
                    "Title of the repository profile. It must start "
                    "with an alphanumeric character and only "
                    "contain lowercase letters, numbers and - or "
                    "+ signs."
                ),
            ),
            Unicode(
                "description",
                optional=True,
                doc="Description of the repository profile.",
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc=("Optional name of the access group to create " "the profile in."),
            ),
        ],
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """Create a new repository profile."""
        profile = context.create_profile(
            call.args.title, call.args.description, call.args.access_group
        )
        return profile.get_state()


@method
class RemoveRepositoryProfiles(RepositoryMethod):
    """
    DEPRECATED: use RemoveRepositoryProfile instead.

    Remove repository profiles in the account.

    - **names.#**: Name of the repository profile to be removed. Multiple
      names can be supplied.

    """

    requires = "ViewAccount"

    schema = Schema(
        name="RemoveRepositoryProfiles",
        doc="Deprecated: use RemoveRepositoryProfile instead.",
        parameters=[
            List(
                "names",
                Unicode(),
                doc="Names of the repository profiles to be removed.",
            )
        ],
        errors=[UnknownRepositoryProfileError, Unauthorized],
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """
        Remove all repository profiles in the account that match the
        call arguments.
        """
        names = call.args.names
        profiles = self.get_profiles_by_names(context, names=names)
        not_removed = context.remove_profiles(profiles)
        return [profile.get_state() for profile in not_removed]


@method
class RemoveRepositoryProfile(RepositoryMethod):
    """
    Remove a repository profile.

    This method takes one mandatory argument:

    - **name**: Name of the repository profile to be removed.

    Example of a valid request::

        ?action=RemoveRepositoryProfile&name=lucid-example
    """

    requires = "EditRepositoryProfile"

    schema = Schema(
        name="RemoveRepositoryProfile",
        doc="Remove repository profile from the account.",
        parameters=[
            Unicode("name", doc="The name of the repository profile to be removed.")
        ],
        errors=[UnknownRepositoryProfileError, Unauthorized],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """
        Remove a repository profile from the account.
        """
        profile = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        management.remove_profiles([profile])
        return {}


@method
class EditRepositoryProfile(RepositoryMethod):
    """
    Edit a repository profile in the account.

    - **name**: Name of the repository profile to edit.
    - **title**: New title of the profile.
    - **description**: New description of the profile.

    Example of a valid request::

        ?action=EditRepositoryProfile&name=lucid-example
            &description=Example+Lucid+profile

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.

    The method returns a JSON serialized dictionary of the repository
    profile::

        {"all_computers": False,
         "description": "Example Lucid profile",
         "id": 1104,
         "name": "lucid-example",
         "tags": ["example"]}
    """

    requires = "ViewAccount"

    schema = Schema(
        name="EditRepositoryProfile",
        doc="Edit a repository profile.",
        parameters=[
            Unicode("name", doc="Name of the repository profile to edit."),
            UnicodeTitle(
                "title", optional=True, doc="Title of the repository profile."
            ),
            Unicode(
                "description",
                optional=True,
                doc="Description of the repository profile.",
            ),
        ],
        errors=[
            InvalidParameterCombinationError,
            UnknownRepositoryProfileError,
            Unauthorized,
        ],
    )

    route = ManagementRoute(factory=RepositoryManagement)

    def query(self, context, call):
        """
        Look up a repository profile and change its properties.
        """
        if not call.args.title and not call.args.description:
            raise InvalidParameterCombinationError(
                "The request must contain at least one of title and "
                "description parameters."
            )

        profile = context.edit_profile(
            call.args.name, title=call.args.title, description=call.args.description
        )
        return profile.get_state()


@method
class AssociateRepositoryProfile(TagAssociationMethod):
    """
    Associate repository profile to computers with specified tags or to all
    computers.

    - **name**: Name of the repository profile.
    - **tags.#**: List of tag names to associate to the profile.
    - **all_computers**: true if the profile should be associated to all
      computers. This parameter is *optional* and defaults to
      *false*. Individual tags associated to the profile will remain, but they
      will only be effective if the "all_computers" flag is later disabled.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=AssociateRepositoryProfile&name=lucid-example
            &tags.1=server&tags.2=lucid

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.

    The method returns JSON serialized info of the profile status::

        {"name": "lucide-example",
         "id": 178,
         "description": "A profile for ubuntu lucid",
         "tags": ["my-computers", "lucid", "server"],
         "all_computers": False,
         "pockets": []}
    """

    requires = "EditRepositoryProfile"

    schema = TagAssociationMethod.schema.extend(
        name="AssociateRepositoryProfile",
        doc="""Associate repository profile to computers with specified tags
            or to all computers.""",
        errors=[UnknownRepositoryProfileError],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )


@method
class DisassociateRepositoryProfile(TagAssociationMethod):
    """
    Disassociate repository profile from computers with specified tags or from
    all computers.

    If **all_computers=true**, the profile will be unflagged as applying to all
    computers, but will still be enabled for computers which have specific tags
    associated with it.

    - **name**: Name of the repository profile.
    - **tags.#**: List of tag names to disassociate from the profile.
    - **all_computers**: If true, the profile will only remain enabled for
      computers with tags associated to the profile. This parameter is
      *optional* and defaults to *false*.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=DisssociateRepositoryProfile&name=lucid-example
            &tags.1=server&tags.2=natty

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.
    - **UnknownTag**: No tag with the specified name exists.
    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.

    The method returns JSON serialized info of the profile status::

        {"name": "lucide-example",
         "id": 178,
         "description": "A profile for ubuntu lucid",
         "tags": ["desktop", "my-computers"],
         "all_computers": False,
         "pockets": []}
    """

    requires = "EditRepositoryProfile"

    associate = False

    schema = TagAssociationMethod.schema.extend(
        name="DisassociateRepositoryProfile",
        doc="""Disassociate repository profile from computers with specified
            tags or from all computers.""",
        errors=[UnknownRepositoryProfileError],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )


@method
class AddPocketsToRepositoryProfile(RepositoryMethod):
    """
    Add repository pockets to a repository profile.

    An activity will be created to add the given pockets to the APT sources of
    the computers associated with the given profile.

    - **name**: Name of the repository profile.
    - **pockets.#**: The names of the pockets to add.
    - **series**: The name of the series the pockets belongs to.
    - **distribution**: The name of the distribution the series belongs to.

    Example of a valid request::

        ?action=AddPocketsToRepositoryProfile&name=lucid-example
            &pockets.1=staging&pockets.2=release&series=lucid
            &distribution=ubuntu

    The method returns JSON serialized info of the profile status::

        {"name": "lucide-example",
         "id": 178,
         "description": "A profile for ubuntu lucid",
         "tags": ["desktop", "my-computers"],
         "all_computers": False,
         "pockets": [ ... ],
         "apt_sources": []}

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **UnknownDistribution**: No distribution with the specified name exists.
    """

    requires = "EditRepositoryProfile"

    schema = Schema(
        name="AddPocketsToRepositoryProfile",
        doc="""Add repository pockets to a repository profile.
            An activity will be created to add the given pockets to the APT
            sources of the computers associated with the given profile. """,
        errors=[
            UnknownRepositoryProfileError,
            UnknownPocketError,
            UnknownSeriesError,
            UnknownDistributionError,
            Unauthorized,
        ],
        parameters=[
            Unicode("name", doc="Name of the repository profile."),
            List("pockets", Unicode(), doc="The names of the pockets to add."),
            Unicode("series", doc="The name of the series the pockets belongs to."),
            Unicode(
                "distribution",
                doc="The name of the distribution the series belongs to.",
            ),
        ],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """Associate the pockets to the profile."""
        profile = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        [distribution] = self.get_distributions_by_names(
            management, names=[call.args.distribution]
        )
        [series] = self.get_series_by_names(distribution, names=[call.args.series])
        pockets = self.get_pockets_by_names(series, names=call.args.pockets)
        management.edit_profile(profile.name, add_pockets=pockets)
        return profile.get_state()


@method
class RemovePocketsFromRepositoryProfile(RepositoryMethod):
    """
    Remove repository pockets from a repository profile.

    An activity will be created to remove the pockets from the APT sources of
    the computers associated with the given profile.

    - **name**: Name of the repository profile.
    - **pockets.#**: The names of the pockets to remove.
    - **series**: The name of the series the pocket belongs to.
    - **distribution**: The name of the distribution the series belongs to.

    Example of a valid request::

        ?action=RemovePocketsFromRepositoryProfile&name=lucid-example
            &pockets.1=staging&pockets.2=release&series=lucid
            &distribution=ubuntu

    The method returns JSON serialized info of the profile status::

        {"name": "lucide-example",
         "id": 178,
         "description": "A profile for ubuntu lucid",
         "tags": ["desktop", "my-computers"],
         "all_computers": False,
         "pockets": [],
         "apt_sources": []}

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.
    - **UnknownPocket**: No pocket with the specified name exists in the
      distribution.
    - **UnknownSeries**: No series with the specified name exists in the
      distribution.
    - **UnknownDistribution**: No distribution with the specified name exists.
    """

    requires = "EditRepositoryProfile"

    schema = Schema(
        name="RemovePocketsFromRepositoryProfile",
        doc="""Remove repository pockets from a repository profile.
            An activity will be created to remove the pockets from the APT
            sources of the computers associated with the given profile.""",
        errors=[
            UnknownRepositoryProfileError,
            UnknownPocketError,
            UnknownSeriesError,
            UnknownDistributionError,
            Unauthorized,
        ],
        parameters=[
            Unicode("name", doc="Name of the repository profile."),
            List("pockets", Unicode(), doc="The names of the pockets to remove."),
            Unicode("series", doc="The name of the series the pocket belongs to."),
            Unicode(
                "distribution",
                doc="The name of the distribution the series belongs to.",
            ),
        ],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """Disassociate pockets from the profile."""
        profile = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        [distribution] = self.get_distributions_by_names(
            management, names=[call.args.distribution]
        )
        [series] = self.get_series_by_names(distribution, names=[call.args.series])
        pockets = self.get_pockets_by_names(series, names=call.args.pockets)
        management.edit_profile(profile.name, remove_pockets=pockets)
        return profile.get_state()


@method
class AddAPTSourcesToRepositoryProfile(RepositoryMethod):
    """
    Add APT sources to a repository profile.

    An activity will be created to add the given source to the the computers
    associated with the given profile.

    - **name**: Name of the repository profile.
    - **apt_sources.#**: The names of the APT sources to add.

    Example of a valid request::

        ?action=AddAPTSourcesToRepositoryProfile&name=lucid-example
            &apt_sources.1=lucid-mirror&apt_sources.2=lucid-extra

    The method returns JSON serialized info of the profile status::

        {"name": "lucide-example",
         "id": 178,
         "description": "A profile for ubuntu lucid",
         "tags": ["desktop", "my-computers"],
         "all_computers": False,
         "pockets": [],
         "apt_sources": [
             {"line": "deb http://example.com/ubuntu lucid main,
              "gpg_key": None,
              "id": 919,
              "name": "lucid-main"},
             {"line": u'deb http://example.com/ubuntu lucid universe",
              "'gpg_key": None,
              "id": 920,
              "name": "lucid-universe"}]}}

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.
    - **UnknownAPTSource**: No APT source with the specified name exists.
    """

    requires = "EditRepositoryProfile"

    schema = Schema(
        name="AddAPTSourcesToRepositoryProfile",
        doc="""Add APT sources to a repository profile.
            An activity will be created to add the given source to the the
            computers associated with the given profile.""",
        errors=[UnknownRepositoryProfileError, APIUnknownAPTSourceError, Unauthorized],
        parameters=[
            Unicode("name", doc="Name of the repository profile."),
            List("apt_sources", Unicode(), doc="The names of the APT sources to add."),
        ],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """Associate the sources to the profile."""
        profile = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        sources = self.get_apt_sources_by_names(management, names=call.args.apt_sources)
        management.edit_profile(profile.name, add_apt_sources=sources)
        return profile.get_state()


@method
class RemoveAPTSourcesFromRepositoryProfile(RepositoryMethod):
    """
    DEPRECATED: use RemoveAPTSourceFromRepositoryProfile instead.

    Remove APT sources from a repository profile.

    An activity will be created to remove the sources from the computers
    associated with the given profile.

    - **name**: Name of the repository profile.
    - **apt_sources.#**: The names of the APT sources to remove.
    """

    requires = "EditRepositoryProfile"

    schema = Schema(
        name="RemoveAPTSourcesFromRepositoryProfile",
        doc="""
            Deprecated: use RemoveAPTSourceFromRepositoryProfile instead.
        """,
        errors=[UnknownRepositoryProfileError, APIUnknownAPTSourceError, Unauthorized],
        parameters=[
            Unicode("name", doc="Name of the repository profile."),
            List(
                "apt_sources", Unicode(), doc="The names of the APT sources to remove."
            ),
        ],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """Disassociate sources from the profile."""
        profile = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        apt_sources_to_remove = set(call.args.apt_sources).intersection(
            set([x.name for x in profile.apt_sources])
        )
        management.edit_profile(
            profile.name, remove_apt_source_names=apt_sources_to_remove
        )
        return profile.get_state()


@method
class RemoveAPTSourceFromRepositoryProfile(RepositoryMethod):
    """
    Remove APT source from a repository profile.

    An activity will be created to remove the APT source from the computers
    associated with the given repository profile.

    - **name**: Name of the repository profile.
    - **apt_source**: The name of the APT source to remove.

    Example of a valid request::

        ?action=RemoveAPTSourcesFromRepositoryProfile&name=lucid-example
            &apt_source=lucid-mirror

    The method returns JSON serialized info of the profile status::

        {"name": "lucide-example",
         "id": 178,
         "description": "A profile for ubuntu lucid",
         "tags": ["desktop", "my-computers"],
         "all_computers": False,
         "pockets": [],
         "apt_sources": []}

    The following errors may be raised:

    - **UnknownRepositoryProfile**: No profile with the specified name exists.
    - **UnknownAPTSource**: No APT source with the specified name exists.
    """

    requires = "EditRepositoryProfile"

    schema = Schema(
        name="RemoveAPTSourceFromRepositoryProfile",
        doc="""Remove APT source from a repository profile.
            An activity will be created to remove the source from the
            computers associated with the given profile.""",
        errors=[UnknownRepositoryProfileError, UnknownAPTSourceError, Unauthorized],
        parameters=[
            Unicode("name", doc="Name of the repository profile."),
            Unicode("apt_source", doc="The name of the APT source to remove."),
        ],
    )

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_profile_by_name",
        identifier_arg="name",
    )

    def query(self, context, call):
        """Disassociate an APT source from the profile."""
        profile = context  # for readability
        management = self.get_management(RepositoryManagement, call)
        source = call.args.apt_source
        management.edit_profile(profile.name, remove_apt_source_names=[source])
        return profile.get_state()
