# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.txapi.server.schema import (
    Schema,
    Unicode,
    List,
    MissingParameterError,
    Bool,
)
from canonical.txapi.server.method import method

from canonical.lib.text import is_valid_slug
from canonical.landscape.api.routes import SubjectRoute
from canonical.landscape.api.error import (
    DuplicateSeriesError,
    GPGKeyHasNoSecretError,
    InvalidParameterCombinationError,
    UnknownDistributionError,
    UnknownGPGKeyError,
    UnknownSeriesError,
)

from canonical.landscape.api.repository.methods import (
    RepositoryMethod,
    SynchronousRepreproMethod,
)
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.model.repository.series import NoPocketsInSeries
from canonical.landscape.model.repository.pocket import get_suite_name
from canonical.landscape.model.repository.distribution import LockError


@method
class CreateSeries(RepositoryMethod, SynchronousRepreproMethod):
    """
    Create a series associated with a distribution in the account.

    This method takes 2 mandatory arguments:

    - **name**: The name of the series. It must be unique within
      series within the distribution, start with an alphanumeric character and
      only contain lowercase letters, numbers and - or + signs.
    - **distribution**: The name of the distribution to create the series in.
    - **pockets.#**: Pockets that will be created in the series, they will be
      in mirror mode by default. This parameter is **optional**.
    - **components.#**: Components for the created pockets. This parameter is
      **optional** if no pocket is specified.
    - **architectures.#**: List of architectures for the created pockets. This
      parameter is **optional** if no pocket is specified.
    - **gpg_key**: The name of the GPG key to use to sign packages lists of the
      created pockets. This parameter is **optional** if no pocket is
      specified.
    - **mirror_gpg_key**: The name of the GPG key to use to verify the mirrored
      repositories for created pockets. If none is given, the stock Ubuntu
      archive one will be used.
    - **mirror_uri**: The URI to mirror for the created pockets. This parameter
      is **optional** if no pocket is specified.
    - **mirror_series**: Optionally, the remote series to mirror. If not
      specified, it defaults to the name of the series being created. If a
      pockets parameter also passed, each of the created pockets will mirror
      the relevant dists/<mirror_series>-<pocket> repository of the remote
      archive.
    - **include_udeb**: Whether the pocket should include selected components
      also for .udeb packages (debian-installer). It's 'false' by default.

    For example, the following request creates a series named 'lucid' in the
    'ubuntu' distribution, with a 'release' pocket (mirroring the 'lucid'
    suite) and an 'updates' one (mirroring the 'lucid-updates' suite)::

       ?action=CreateSeries&name=lucid&distribution=ubuntu
          &pockets.1=release&pockets.2=updates&components.1=main
          &components.2=universe&components.3=multiverse
          &components.4=restricted&architectures.1=amd64&gpg_key=my-key
          &mirror_uri=http://archive.ubuntu.com/ubuntu

    The following errors may be raised:

    - **InvalidParameterValue**: The Name parameter has an invalid value.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **DuplicateSeries**: A series with same name already exists in the
      specified distribution.
    - **UnknownGPGKey**: No GPG key with the specified name exists.
    - **GPGKeyHasNoSecret**: The GPG key requested for signing does not have a
      secret key associated.
    - **OperationInProgress**: Another operation is in progress on the
      distribution.

    The method returns JSON serialized info of the created series::

        {"name": "lucid",
         "pockets": [],
         "creation_time": "2011-07-27T09:04:03Z"}
    """

    schema = Schema(
        name="CreateSeries",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    "The name of the series. It must be unique within series "
                    "within the distribution, start with an alphanumeric "
                    "character and only contain lowercase letters, numbers "
                    "and - or + signs."
                ),
            ),
            Unicode(
                "distribution",
                doc="The name of the distribution to create the series in.",
            ),
            List(
                "pockets",
                Unicode(validator=is_valid_slug),
                optional=True,
                doc=(
                    "Pockets that will be created in the series, they will "
                    "be in mirror mode by default."
                ),
            ),
            List(
                "components",
                Unicode(),
                optional=True,
                doc=(
                    "List of components for the created pockets. This "
                    "parameter is **optional** if no pocket is specified."
                ),
            ),
            List(
                "architectures",
                Unicode(),
                optional=True,
                doc=(
                    "List of architectures for the created pockets. This "
                    "parameter is **optional** if no pocket is specified"
                ),
            ),
            Unicode(
                "gpg_key",
                optional=True,
                doc=(
                    "The name of the GPG key to use to sign packages lists "
                    "of the created pockets. This parameter is **optional** "
                    "if no pocket is specified."
                ),
            ),
            Unicode(
                "mirror_uri",
                optional=True,
                doc=(
                    "The URI to mirror for the created pockets. This "
                    "parameter is **optional** if no pocket is specified."
                ),
            ),
            Unicode(
                "mirror_series",
                optional=True,
                doc=(
                    "The remote series to mirror. If not specified, it "
                    "defaults to the name of the series being created. If a "
                    "pockets parameter also passed, each of the created "
                    "pockets will mirror the relevant "
                    "dists/<mirror_series>-<pocket> repository of the remote "
                    "archive."
                ),
            ),
            Unicode(
                "mirror_gpg_key",
                optional=True,
                doc=(
                    "The name of the GPG key to use to verify the mirrored "
                    "repositories for created pockets. If none is given, the "
                    "stock Ubuntu archive one will be used."
                ),
            ),
            Bool(
                "include_udeb",
                optional=True,
                default=False,
                doc=(
                    "Whether the pocket should include selected components "
                    "also for .udeb packages (debian-installer). "
                    "It's 'false' by default."
                ),
            ),
        ],
        errors=[
            DuplicateSeriesError,
            GPGKeyHasNoSecretError,
            InvalidParameterCombinationError,
            UnknownDistributionError,
            UnknownGPGKeyError,
            LockError,
            MissingParameterError,
        ],
        doc="""
    Create a series associated with a distribution in the account.
    """,
    )

    requires = "CreateSeries"

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_distribution",
        identifier_arg="distribution",
        unknown_error_factory=UnknownDistributionError,
    )

    def query(self, context, call):
        distribution = context  # for readability
        name = call.args.name
        series = distribution.get_series(names=[name]).one()

        if series is not None:
            raise DuplicateSeriesError("Series '%s' already present" % name)

        if call.args.pockets:
            for param in ("components", "architectures", "mirror_uri", "gpg_key"):
                if not getattr(call.args, param):
                    raise MissingParameterError(param)

            person = get_person_by_id(call.principal.id)
            [gpg_key] = self.get_gpg_keys_by_names(
                distribution.account, person, names=[call.args.gpg_key]
            )

            if gpg_key.secret_material is None:
                raise GPGKeyHasNoSecretError(
                    "gpg_key '%s' has no secret key" % call.args.gpg_key
                )

            if call.args.mirror_gpg_key:
                [mirror_gpg_key] = self.get_gpg_keys_by_names(
                    distribution.account, person, names=[call.args.mirror_gpg_key]
                )
            else:
                mirror_gpg_key = None

        else:
            for param in (
                "components",
                "architectures",
                "gpg_key",
                "mirror_uri",
                "mirror_series",
                "mirror_gpg_key",
            ):
                if getattr(call.args, param):
                    raise InvalidParameterCombinationError(
                        "Can't specify '%s' without 'pockets'" % param
                    )

        self.lock_distribution(distribution)
        series = distribution.create_series(name)
        pockets_full_names = []

        for name in call.args.pockets:
            pocket = series.create_pocket(
                name=name,
                gpg_key=gpg_key,
                components=list(call.args.components),
                archs=list(call.args.architectures),
                include_udeb=call.args.include_udeb,
            )
            if call.args.mirror_series:
                suite = get_suite_name(call.args.mirror_series, name)
            else:
                suite = pocket.get_full_name()
            pocket.set_mirror_mode(call.args.mirror_uri, suite, mirror_gpg_key)
            pockets_full_names.append(pocket.get_full_name())

        series_state = series.get_state()
        self.finish = lambda _: series_state
        return (
            distribution.get_reprepro(),
            distribution.account_id,
            call.principal.id,
            distribution.name,
        )

    def run_reprepro(self, reprepro):
        """
        Run the reprepro 'export' command to create the distribution trees.
        """
        return reprepro.export()


@method
class RemoveSeries(RepositoryMethod, SynchronousRepreproMethod):
    """
    Remove a repository series from a distribution.

    This method takes 2 mandatory arguments:

    - **name**: The name of the series.
    - **distribution**: The name of the distribution.

    For example, the following request removes a series named 'lucid' for the
    'ubuntu' distribution::

       ?action=RemoveSeries&name=lucid&distribution=ubuntu

    The following errors may be raised:

    - **UnknownDistribution**: Distribution with specified name doesn't exist.
    - **UnknownSeries**: No series with the specified name exists.
    """

    schema = Schema(
        name="RemoveSeries",
        parameters=[
            Unicode("name", doc="The name of the series to remove."),
            Unicode("distribution", doc="The name of the distribution."),
        ],
        errors=[UnknownDistributionError, UnknownSeriesError, LockError],
        doc="""
    Remove a repository series from a distribution.
    """,
    )

    requires = "EditSeries"

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_distribution",
        identifier_arg="distribution",
        unknown_error_factory=UnknownDistributionError,
    )

    def query(self, context, call):
        distribution = context  # for readability
        [series] = self.get_series_by_names(distribution, [call.args.name])

        self.lock_distribution(distribution)
        management = self.get_management(RepositoryManagement, call)
        management.remove_series(series)
        self.finish = lambda _: {}
        return (
            distribution.get_reprepro(),
            distribution.account_id,
            call.principal.id,
            distribution.name,
        )

    def run_reprepro(self, reprepro):
        """Remove files for removed pockets."""
        return reprepro.clearvanished()


@method
class DeriveSeries(RepositoryMethod, SynchronousRepreproMethod):
    """
    Derive a series from another one in the same distribution. The derived
    series will have pockets with names corresponding to the origin series,
    each one configured to pull from the pocket in origin series.

    This method takes 3 mandatory arguments:

    - **name**: The name of the derived series. It must be unique within the
      distribution, start with an alphanumeric character and only contain
      lowercase letters, numbers and - or + signs.
    - **origin**: The name of the origin series.
    - **distribution**: The name of the distribution to derive the series in.

    For example, the following request derive a series named 'lucid-staging'
    from the 'lucid' series in the 'ubuntu' distribution::

       ?action=DeriveSeries&name=lucid-staging&origin=lucid
        &distribution=ubuntu

    The following errors may be raised:

    - **InvalidParameterValue**: The Name parameter has an invalid value.
    - **UnknownDistribution**: No distribution with the specified name exists.
    - **UnknownSeries**: No origin series with the specified name exists.
    - **DuplicateSeries**: A series with same name already exists in the
      specified distribution.
    - **EmptySeries**: The series to derive from does not contain pockets.
    - **OperationInProgress**: Another operation is in progress on the
      distribution.

    The method returns JSON serialized info of the derived series::

        {"creation_time": "2011-07-26T15:15:17Z",
         "name": "lucid-staging",
         "pockets": [{"name": "release",
                      "creation_time": "2011-07-26T15:15:17Z",
                      "pull_pocket": "lucid/release",
                      "architectures": ["i386"],
                      "components": ["main", "contrib"],
                      "mode": "pull"},
                     {"name": "updates",
                     "creation_time": "2011-07-26T15:15:17Z",
                     "pull_pocket": "lucid/updates",
                     "architectures": ["i386", "amd64"],
                     "components": ["main"],
                     "mode": "pull"}]}
    """

    schema = Schema(
        name="DeriveSeries",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    "The name of the derived series. It must be unique "
                    "within the distribution, start with an alphanumeric "
                    "character and only contain lowercase letters, numbers "
                    "and - or + signs."
                ),
            ),
            Unicode("origin", doc="The name of the origin series."),
            Unicode(
                "distribution",
                doc="The name of the distribution to derive the series in.",
            ),
        ],
        errors=[
            UnknownDistributionError,
            UnknownSeriesError,
            NoPocketsInSeries,
            DuplicateSeriesError,
            LockError,
        ],
        doc="""
    Derive a series from another one in the same distribution. The derived
    series will have pockets with names corresponding to the origin series,
    each one configured to pull from the pocket in origin series.
    """,
    )

    requires = "CreateSeries"

    route = SubjectRoute(
        factory=RepositoryManagement,
        getter="get_distribution",
        identifier_arg="distribution",
        unknown_error_factory=UnknownDistributionError,
    )

    def query(self, context, call):
        distribution = context  # for readability
        [origin] = self.get_series_by_names(distribution, [call.args.origin])

        series = distribution.get_series(names=[call.args.name]).one()
        if series is not None:
            raise DuplicateSeriesError("Series '%s' already present" % call.args.name)

        self.lock_distribution(distribution)

        try:
            derived = origin.derive_series(call.args.name)
        except NoPocketsInSeries:
            distribution.unlock()
            raise

        series_state = derived.get_state()
        self.finish = lambda _: series_state
        return (
            distribution.get_reprepro(),
            distribution.account_id,
            call.principal.id,
            distribution.name,
        )

    def run_reprepro(self, reprepro):
        """
        Run the reprepro 'export' command to create the distribution trees.
        """
        return reprepro.export()
