# Copyright 2005-2023 Canonical Limited. All rights reserved.

import json
import logging
from base64 import b64encode

from storm.twisted.transact import transact

from canonical.txapi.server.resource import QueryAPI
from canonical.txapi.server.schema import Schema, Unicode, Integer, Date, RawStr

from canonical.landscape.api.error import UnauthorisedError, MultiError
from canonical.landscape.api.principal import Principal, AnonymousCredentials

from canonical.landscape.model.main.person import get_credentials_by_access_key_id


class ServiceAPI(QueryAPI):
    """An API used to configure the Landscape service.

    @param transactor: A L{Transactor} that can be used to execute
        functions that run in a database transaction.
    """

    signature_versions = (2,)
    content_type = "application/json"

    schema = Schema(
        parameters=[
            Unicode("action"),
            RawStr("access_key_id"),
            Date("timestamp", optional=True),
            Date("expires", optional=True),
            Unicode("version", optional=True),
            Unicode("signature_method", optional=True, default="HmacSHA256"),
            Unicode("signature"),
            Integer("signature_version", optional=True, default=2),
        ]
    )

    def __init__(self, transactor, registry=None, path=None):
        QueryAPI.__init__(self, registry=registry, path=path)
        self.transactor = transactor

    def get_call_arguments(self, request):
        """
        Overridden from the base L{QueryAPI.get_call_arguments} because we use
        a different naming scheme for our core API parameters.

        First, parameter names are lowercase and underscore-separated instead
        of CamelCase; second, we want to change the "AWSAccessKeyId" to a
        simple "access_key_id" to avoid Amazon-specific terminology in our API.
        """
        params = {k.decode(): v[-1].decode() for k, v in request.args.items()}
        args, rest = self.schema.extract(params)
        # Get rid of Signature so it doesn't mess with signature verification
        params.pop("signature")
        result = {
            "transport_args": {
                "action": args.action,
                "access_key_id": args.access_key_id,
                "timestamp": args.timestamp,
                "expires": args.expires,
                "version": args.version,
                "signature_method": args.signature_method,
                "signature": args.signature,
                "signature_version": args.signature_version,
            },
            "handler_args": rest,
            "raw_args": params,
        }
        return result

    def get_method(self, call):
        return QueryAPI.get_method(self, call, self.transactor)

    @transact
    def get_principal(self, access_key):
        """
        Retrieve the L{APICredentials} matching the C{access_key}, and return a
        L{Principal} wrapping it.
        """
        if access_key == b"anonymous":
            return Principal(AnonymousCredentials())
        credentials = get_credentials_by_access_key_id(access_key)
        if credentials is None:
            return None
        return Principal(credentials)

    def authorize(self, method, call):
        if not method.authorize(call):
            raise UnauthorisedError("Permission denied.")

    def dump_result(self, result):
        return json.dumps(self._decode_bytes(result))

    def dump_error(self, error, request):
        """
        Convert an exception to a json structure.

        L{APIError}s and L{MultiError}s can be converted.
        """
        if isinstance(error, MultiError):
            reasons = [
                {"message": reason.message, "param": reason.param, "error": reason.code}
                for reason in error.get_reasons()
            ]
            dump = {"error": error.code, "errors": reasons, "message": error.message}
        else:
            dump = {"message": error.message}
            if error.code:
                dump["error"] = error.code
        return json.dumps(dump)

    def _decode_bytes(self, data):
        """
        Recursively decodes bytes to strings to prep for json dumping.
        If UTF-8 decoding doesn't work, b64-encodes first.
        """
        if isinstance(data, bytes):
            logging.warn("API endpoint returns bytes: %s", data)
            try:
                return data.decode()
            except UnicodeDecodeError:
                return b64encode(data).decode()

        if isinstance(data, list):
            return [self._decode_bytes(d) for d in data]

        if isinstance(data, dict):
            return {
                self._decode_bytes(k): self._decode_bytes(v) for k, v in data.items()
            }

        return data
