# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import map
from builtins import object
from canonical.search.engine import SearchCriteriaError

from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.computer.collection import ComputerCollection
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.api.helpers import APIHelper
from canonical.landscape.api.error import (
    InvalidQueryError,
    UnknownComputersError,
    UnknownComputerError,
)
from canonical.landscape.service.security import PersonPrincipal


__all__ = [
    "GlobalRoute",
    "AccountRoute",
    "ManagementRoute",
    "SelectionRoute",
    "SubjectRoute",
    "EntityLookupRoute",
    "CollectionLookupRoute",
]


class GlobalContext(object):
    """A context which is used for methods oriented to AccountManagers."""

    # GlobalContext is passed into the service/security authorise function via
    # GlobalRoute. Authorize will validate
    # principal.has(<permission>, on=<access_context>). When access_context is
    # None, like  GlobalContext, authorize will validate the person has the
    # global permission.
    access_context = None

    def __init__(self, person):
        self.person = person


class GlobalRoute(object):
    """
    A route for methods not bound to a specific account, returning the
    L{GlobalContext}.
    """

    def __call__(self, call):
        return GlobalContext(get_person_by_id(call.principal.id)), None


class AccountRoute(object):
    """Return the L{Account} associated to the caller's credentials."""

    def __call__(self, call):
        account = get_account_by_id(call.principal.account_id)
        return account, account


class ManagementRoute(APIHelper):
    """Return a management object for the credential's L{Account}.

    @param factory: The management class to use.
    """

    def __init__(self, factory):
        self._factory = factory

    def __call__(self, call):
        mgmt = self.get_management(self._factory, call)
        return mgmt, mgmt.account


class SubjectRoute(object):
    """
    Return a model object for the credential's L{Account}.

    @param factory: The management class to use
    @param getter: The getter method to use on the factory.
    @param identifier: The Call field to use as the parameter for the getter
        call.
    @param unknown_error_factory: A factory for "unknown foo" exception types
        which will be invoked if the getter returns None. It will be invoked
        with the same argument that the c{getter} is invoked with.
    """

    def __init__(self, factory, getter, identifier_arg, unknown_error_factory=None):
        self._factory = factory
        self._getter = getter
        self._identifier_arg = identifier_arg
        self._unknown_error_factory = unknown_error_factory

    def __call__(self, call):
        management, account = ManagementRoute(self._factory)(call)

        getter_method = getattr(management, self._getter)
        identifier = getattr(call.args, self._identifier_arg)

        result = getter_method(identifier)
        if result is None and self._unknown_error_factory is not None:
            error = self._unknown_error_factory(identifier)
            raise error
        else:
            return result, account


class EntityLookupRoute(object):
    """Return a model entity using the given EntityLookup.

    @param factory: The EntityLookup class to use to perform the lookup.
    @param key_args: The name of the arguments that compose the entity key.
        They will be combined in a tuple and passed to the lookup. The
        default is to use a single 'name' argument as key.
    """

    def __init__(self, factory, *key_args):
        self._factory = factory
        self._key_args = key_args

    def __call__(self, call):
        lookup = self._factory(PersonPrincipal(call.principal.id))
        account = get_account_by_id(call.principal.account_id)
        key_args = self._key_args if self._key_args else ["name"]
        key = tuple(getattr(call.args, key_arg) for key_arg in key_args)
        if len(key_args) == 1:
            # This is not a compound key, so don't use a tuple.
            key = key[0]
        return lookup(account, key), account


class CollectionLookupRoute(object):
    """Return a model collection using the given CollectionLookup.

    This route will grab filter arguments from the API call and pass them
    to the given CollectionLookup in order to get a Collection object
    configured  to only return the entities matching the given parameters. This
    collection will be the context of the API method and can be used as
    context for CollectionOperation-based service operations.

    @param factory: The CollectionLookup class to use to perform the lookup.
    @param filter_args: The name of the arguments that should be used as filter
        parameters for the collection.
    """

    def __init__(self, factory, *filter_args):
        self._factory = factory
        self._filter_args = filter_args

    def __call__(self, call):
        lookup = self._factory(PersonPrincipal(call.principal.id))
        account = get_account_by_id(call.principal.account_id)
        # Extract filter arguments from the Call object and pass them to the
        # lookup, in order to get a properly filtered collection.
        filters = {arg: getattr(call.args, arg) for arg in self._filter_args}
        return lookup(account, **filters), account


class SelectionRoute(object):
    """Return a L{ComputerSelection} matching the parameters of a C{Call}.

    @param permission: Include in the selection only L{Computer}s that the
        accessing person has this permission on.
    @param error_if_empty: Raise an error if the selection ends up to be empty.
    @param enforce_limits: Apply limits specified by 'limit' and 'offset'
        parameters, if present.
    """

    def __init__(self, permission, error_if_empty=True, enforce_limits=True):
        self.permission = "landscape." + permission
        self.error_if_empty = error_if_empty
        self.enforce_limits = enforce_limits

    def __call__(self, call):
        collection = ComputerCollection(
            account_id=call.principal.account_id,
            person_id=call.principal.id,
            permission=self.permission,
        )

        if "computer_ids" in call.args:
            computer_ids = call.args.computer_ids
            collection = collection.with_computer_ids(call.args.computer_ids)
            error_msg_tmpl = "Invalid computer_ids supplied: {}"
        if "computer_titles" in call.args:
            computer_ids = list(dict(call.args.computer_titles).keys())
            collection = collection.with_computer_ids(computer_ids)
            error_msg_tmpl = "Invalid ids in computer_titles supplied: {}"
        if "query" in call.args:
            try:
                collection = collection.with_search_criteria(call.args.query)
            except SearchCriteriaError as error:
                raise InvalidQueryError("Invalid query '%s'." % str(error))

        result = collection.find()
        if self.enforce_limits and ("limit" in call.args or "offset" in call.args):
            result.order_by(Computer.title)
            result.config(offset=call.args.offset, limit=call.args.limit)
        computers = list(result)

        if "computer_ids" in call.args or "computer_titles" in call.args:
            found_computer_ids = set(computer.id for computer in computers)
            missing_ids = set(computer_ids) - found_computer_ids
            if missing_ids:
                raise UnknownComputerError(
                    error_msg_tmpl.format(",".join(map(str, missing_ids)))
                )

        account = get_account_by_id(call.principal.account_id)
        person = get_person_by_id(call.principal.id)
        selection = ComputerSelection(
            account, person=person, permission=self.permission
        )
        selection.select_computers(computers)

        if self.error_if_empty and not selection.get_selected_count():
            raise UnknownComputersError(
                "No computer(s) matching the query '%s' and "
                "authorised for this operation." % call.args.query
            )

        return selection, account
