# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Saved searches
==============

The methods available here are related to saved searches.
"""
from builtins import str
from canonical.txapi.server.schema import Schema, Integer, Unicode
from canonical.txapi.server.method import method

from canonical.lib.text import is_valid_slug

from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.error import (
    InvalidSavedSearchNameError,
    InvalidSavedSearchCriteriaError,
    UnknownSavedSearchError,
    MissingParameterError,
)
from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.model.account.savedsearch import (
    SavedSearchManagement,
    InvalidSavedSearchName,
    InvalidSavedSearchCriteria,
    DuplicateSavedSearchError,
    EmptyTitleAndSearch,
    UnknownSavedSearch,
)


@method
class GetSavedSearches(AccountMethod):
    """
    Retrieve saved searches associated with the current account.

    It also supports two optional arguments:

    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.

    For example, the following request looks for saved searches associated with
    the account and limits the result to 20 saved searches::

        ?action=GetSavedSearches&limit=20

    The method returns a JSON serialized list of saved searches, like the
    following result::

        [{"name": "first",
          "title": "First Computer",
          "search": "id:1"}]
    """

    requires = "ViewSavedSearch"

    schema = Schema(
        name="GetSavedSearches",
        parameters=[
            Integer(
                "offset",
                default=RESULT_OFFSET,
                optional=True,
                doc="The offset inside the list of results.",
            ),
            Integer(
                "limit",
                default=RESULT_LIMIT,
                optional=True,
                doc=(
                    "The maximum number of results returned by the "
                    "method. It defaults to 1000."
                ),
            ),
        ],
        errors=[],
        doc="""Retrieve saved searches associated with the current account.""",
    )

    route = ManagementRoute(factory=SavedSearchManagement)

    def query(self, context, call):
        """
        Perform a query to get L{SavedSearch}es subject to the limit and
        offset.
        """
        limit = call.args.limit
        offset = call.args.offset
        savedsearches = context.get_saved_searches(limit=limit, offset=offset)
        return [ss.get_state() for ss in savedsearches]


@method
class CreateSavedSearch(AccountMethod):
    """
    Create a new saved search associated with the current account.

    It requires three arguments:

    - **name**: The "slug" name for this saved search. It must consist of only
      lowercase ASCII letters, numbers and hyphens. This is the text which
      must be used when using the "search:name" syntax.  If this parameter is
      not included a name will be generated automatically based on the title.
    - **title**: The display name for the SavedSearch.
    - **search**: The search string to save.

    For example, the following request creates a saved search associated with
    the account::

        ?action=CreateSavedSearch&name=name1&title=Title%201&search=id:1

    The method returns a JSON serialized version of the new saved search, like
    the following result::

        {"name": "first",
         "title": "First Computer",
         "search": "id:1"}

    """

    requires = "CreateSavedSearch"

    schema = Schema(
        name="CreateSavedSearch",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                optional=True,
                doc=(
                    'The "slug" name for this saved search. It must '
                    "consist of only lowercase ASCII letters, numbers "
                    "and hyphens. This is the text which must be used "
                    'when using the "search:name" syntax.  If this '
                    "parameter is not included a name will be generated "
                    "automatically based on the title."
                ),
            ),
            UnicodeTitle("title", doc="The display name for the SavedSearch."),
            Unicode("search", doc="The search string to save."),
        ],
        errors=[
            InvalidSavedSearchNameError,
            InvalidSavedSearchCriteriaError,
            DuplicateSavedSearchError,
        ],
        doc="""Create a new saved search associated with the current account.
        """,
    )

    route = ManagementRoute(factory=SavedSearchManagement)

    def query(self, context, call):
        """
        Perform a query to create a L{SavedSearch} associated with the
        L{Account}.
        """

        try:
            name, title, search = context.validate_create_saved_search(
                call.args.name, call.args.title, call.args.search
            )
            savedsearch = context.create_saved_search(name, title, search)

        except InvalidSavedSearchName as error:
            raise InvalidSavedSearchNameError(str(error))

        except InvalidSavedSearchCriteria as error:
            raise InvalidSavedSearchCriteriaError(str(error))

        return savedsearch.get_state()


@method
class RemoveSavedSearch(AccountMethod):
    """
    Remove a saved search associated with the current account.

    It requires one argument:

    - **name**: The "slug" name for this saved search.

    For example, the following request removes a saved search associated with
    the account::

        ?action=RemoveSavedSearch&name=name1

    """

    requires = "EditSavedSearch"
    schema = Schema(
        name="RemoveSavedSearch",
        parameters=[Unicode("name", doc='The "slug" name for this saved search.')],
        errors=[UnknownSavedSearchError],
        doc="""Remove a saved search associated with the current account.""",
    )

    route = SubjectRoute(
        factory=SavedSearchManagement,
        getter="get_saved_search_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownSavedSearchError,
    )

    def query(self, context, call):
        """
        Remove a L{SavedSearch} associated with the L{Account}
        """
        saved_search = context
        management = self.get_management(SavedSearchManagement, call)
        management.remove_saved_searches([saved_search])
        return {}


@method
class EditSavedSearch(AccountMethod):
    """
    Edit a saved search associated with the current account.

    It takes three arguments:

    - **name**: The "slug" name for this saved search, this is the text which
      must be used when using the "search:name" syntax.  A saved search with
      this name must already exist in the account.
    - **title**: The new display name for the saved search.  If this parameter
      is not included then the title will not be modified.
    - **search**: The search string to save.  If this parameter is not included
      then the search string will not be modified.

    For example, the following request edits the title of a saved search
    associated with the account::

        ?action=EditSavedSearch&name=name1&title=Title%201

    The method returns a JSON serialized version of the new saved search, like
    the following result::

        {"name": "first",
         "title": "First Computer",
         "search": "id:1"}

    """

    requires = "EditSavedSearch"
    schema = Schema(
        name="EditSavedSearch",
        parameters=[
            Unicode(
                "name",
                validator=is_valid_slug,
                doc=(
                    'The "slug" name for this saved search, this is '
                    "the text which must be used when using the "
                    '"search:name" syntax. A saved search with this '
                    "name must already exist in the account."
                ),
            ),
            UnicodeTitle(
                "title",
                optional=True,
                doc=(
                    "The new display name for the saved search. "
                    "If this parameter is not included then the title "
                    "will not be modified."
                ),
            ),
            Unicode(
                "search",
                optional=True,
                doc=(
                    "The search string to save. If this parameter is "
                    "not included then the search string will not be "
                    "modified."
                ),
            ),
        ],
        errors=[
            MissingParameterError,
            UnknownSavedSearchError,
            DuplicateSavedSearchError,
            InvalidSavedSearchCriteriaError,
        ],
        doc="""Edit a saved search associated with the current account.""",
    )

    route = ManagementRoute(factory=SavedSearchManagement)

    def query(self, context, call):
        """
        Edit a L{SavedSearch} associated with the L{Account}
        """
        try:
            context.validate_edit_saved_search(
                call.args.name, call.args.title, call.args.search
            )

        except EmptyTitleAndSearch as error:
            raise MissingParameterError(str(error))

        except UnknownSavedSearch as error:
            raise UnknownSavedSearchError(str(error))

        except InvalidSavedSearchCriteria as error:
            raise InvalidSavedSearchCriteriaError(str(error))

        savedsearch = context.edit_saved_search(
            call.args.name, title=call.args.title, search=call.args.search
        )

        return savedsearch.get_state()
