# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.txapi.server.schema import AbstractParameter as Parameter, Unicode

from canonical.lib.text import is_valid_title


class UnicodeTitle(Unicode):
    """
    A parameter that must be a C{unicode} string, without newlines, and contain
    more than just whitespace.
    """

    kind = "unicode title"

    def parse(self, value):
        value = super(UnicodeTitle, self).parse(value)
        if not is_valid_title(value):
            raise ValueError()
        return value


class Mapping(Parameter):
    """A parameter that maps keys to values."""

    kind = "mapping"
    supports_multiple = True

    def __init__(
        self, name=None, key=None, value=None, optional=False, default=None, doc=None
    ):
        """
        @param key: A Parameter instance which will be used to parse and
            format the keys in the mapping (default: Unicode).
        @param value: A Parameter instance which will be used to parse and
            format the values in the mapping (default: Unicode).
        """
        if key is None:
            key = Unicode()
        elif not hasattr(key, "parse") or not hasattr(key, "format"):
            raise ValueError("expected a Parameter for key")
        if value is None:
            value = Unicode()
        elif not hasattr(value, "parse") or not hasattr(value, "format"):
            raise ValueError("expected a Parameter for value")
        if default is None:
            default = {}
        super(Mapping, self).__init__(name, optional=optional, default=default, doc=doc)
        if key.name is None:
            key.name = name
        self.key = key
        if value.name is None:
            value.name = name
        self.value = value

    def parse(self, arg):
        """Convert the keys and values of the given mapping."""
        try:
            getpairs = arg.items
        except AttributeError:
            items = arg
        else:
            items = getpairs()
        parse_key = self.key.parse
        parse_value = self.value.parse
        result = {}
        for key, value in items:
            key = parse_key(key)
            value = parse_value(value)
            result[key] = value
        # txapi.server.schema.Arguments() does funny stuff when given
        # dicts, so we return the item pairs instead.
        return list(result.items())

    def format(self, arg):
        """Format the keys and values of the given mapping."""
        try:
            getpairs = arg.items
        except AttributeError:
            items = arg
        else:
            items = getpairs()
        format_key = self.key.format
        format_value = self.value.format
        result = {}
        for key, value in items:
            key = format_key(key)
            value = format_value(value)
            result[key] = value
        return result
