# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Utility for converting a L{txapi.server.schema.Schema} to JSON.
"""

# TODO:
# - allow_none? min/max?

from builtins import object
from canonical.landscape.api.errormapping import get_error_code


def serialize_schema(schema):
    """
    Convert a L{txapi.server.schema.Schema} to a format suitable for encoding
    with JSON.
    """
    return SchemaConverter().serialize(schema)


class SchemaConverter(object):
    """
    Helper class for serializeing a L{txapi.server.schema.Schema}.
    """

    def serialize(self, schema):
        """
        Serialize a L{txapi.server.schema.Schema} to a dict.
        """
        result = {
            "name": schema.name,
            "doc": schema.doc,
            "result": {},
            "errors": [],
        }
        result["parameters"] = [
            self._typify(parameter, include_name=True)
            for parameter in schema.get_parameters()
        ]
        if schema.result is not None:
            for field_key, field_type in list(schema.result.items()):
                result["result"][field_key] = self._typify(field_type)
        if schema.errors is not None:
            for error in schema.errors:
                if isinstance(error, str):
                    code = error
                else:
                    code = get_error_code(error)
                if code is None:
                    raise RuntimeError(
                        "Couldn't find an API code for exception %s" % (error,)
                    )
                result["errors"].append({"code": code})
        return result

    def _typify(self, param, include_name=False):
        """
        Return a description of a parameter's type.

        This dispatches to methods beginning with "_typify_" and ending with
        the value of L{Parameter.kind} on the parameter. If such a method
        doesn't exist, it will return a simple description of the type.
        """
        handler = getattr(self, "_typify_" + param.kind, None)
        result = {"type": param.kind}
        if include_name:
            result["name"] = param.name
        if param.optional:
            result["optional"] = param.optional
            result["default"] = param.default
        if param.doc:
            result["doc"] = param.doc
        if handler is not None:
            result.update(handler(param))
        return result

    def _typify_list(self, param):
        """
        _Typify a L{txapi.server.schema.List} while indicating the type of its
        items.
        """
        return {"item": self._typify(param.item)}

    def _typify_mapping(self, param):
        """
        _Typify a c.l.api.schema.Mapping while indicating the type
        of its key and value.
        """
        return {
            "key": self._typify(param.key),
            "value": self._typify(param.value),
        }

    def _typify_structure(self, param):
        """
        _Typify a L{txapi.server.schema.Structure} while indicating the types
        of its fields.
        """
        fields = dict(
            [
                (field_key, self._typify(field_type))
                for field_key, field_type in list(param.fields.items())
            ]
        )
        return {"fields": fields}
