# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Scripts
=======

The methods available here are related to stored scripts.
"""
from canonical.txapi.server.schema import Schema, Integer, Unicode
from canonical.txapi.server.method import method
from zope.security.interfaces import Unauthorized

from canonical.lib.timestamp import parse_datetime

from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.helpers import File, Data
from canonical.landscape.api.error import (
    InvalidTimeError,
    RequiresUserError,
    InvalidQueryError,
    UnknownComputersError,
)
from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.routes import ManagementRoute, SelectionRoute, SubjectRoute

from canonical.landscape.model.script.errors import (
    UnknownScriptError,
    UnknownScriptAttachmentError,
    DuplicateScriptAttachmentError,
    TooManyScriptAttachmentsError,
    ScriptAttachmentSizeError,
    ScriptEncodingError,
)
from canonical.landscape.model.script.interfaces import MIN_TIME_LIMIT, MAX_TIME_LIMIT
from canonical.landscape.model.script.script import (
    ScriptExecution,
    DuplicateScriptError,
    EmptyScriptCodeError,
    EmptyScriptInterpreterError,
)
from canonical.landscape.model.script.management import ScriptManagement
from canonical.landscape.model.main.errors import UnknownAccessContextError


__all__ = [
    "GetScripts",
    "GetScriptCode",
    "ExecuteScript",
    "RemoveScript",
    "CreateScript",
    "EditScript",
    "CopyScript",
    "RemoveScriptAttachment",
    "CreateScriptAttachment",
]


class ScriptBodyConversionMixin(object):
    """
    A mixin to help extracting a script's body (that has to be valid text).
    """

    def get_script_body_from_code(self, code):
        try:
            code = code.decode("utf-8")
        except UnicodeDecodeError:
            raise ScriptEncodingError()
        return code


@method
class GetScripts(AccountMethod):
    """
    Retrieve stored scripts associated with the current account.

    It also supports two optional arguments:

    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.

    For example, the following request looks for stored scripts associated with
    the account and limits the result to 20 scripts::

        ?action=GetScripts&limit=20

    The method returns a JSON serialized list of scripts, like the following
    result::

        [{"creator": {"name": "John Smith", "id": 12345, "email":
                      "john@example.com"}
          "interpreter": "/bin/bash",
          "title": "some script",
          "time_limit": 300,
          "username": "root",
          "id": 12345,
          "access_group": "somegroup",
          "attachments": ["file 1"]}]
    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetScripts",
        parameters=[
            Integer(
                "limit",
                default=RESULT_LIMIT,
                optional=True,
                doc=(
                    "The maximum number of results returned by the "
                    "method. It defaults to 1000."
                ),
            ),
            Integer(
                "offset",
                default=RESULT_OFFSET,
                optional=True,
                doc="The offset inside the list of results.",
            ),
        ],
        errors=[],
        doc="""Retrieve stored scripts associated with the current account.""",
    )

    route = ManagementRoute(factory=ScriptManagement)

    def query(self, context, call):
        """
        Perform a query to get all L{Scripts} associated with the L{Account},
        subject to the limit and offset.
        """
        scripts = context.get_scripts()
        scripts.config(offset=call.args.offset, limit=call.args.limit)
        return [script.get_state() for script in scripts]


@method
class ExecuteScript(AccountMethod):
    """
    Execute script on computers.

    This method requires the following mandatory arguments:

    - **query**: A query string used to select the computers to execute the
      script on. See :ref:`Computer Queries <computer-query>` for details.
    - **script_id**: The id of the script stored in the server.
    - **username**: The username to execute the script as on the client.
      Required if the script has no default username.

    It also supports an optional argument:

    - **deliver_after**: Run the script after the specified time.  The time
      format is YYYY-MM-DDTHH:MM:SSZ.

    For example, the following request will execute the Script with Id 19000 on
    all computers with the server tag::

        ?action=ExecuteScript&query=tag:server&script_id=19000

    The method returns a JSON structure with the activity that was created::

        {'computer_id': None,
         'creation_time': '2012-11-26T23:36:20Z',
         'creator': {'email': 'john@example.com', 'id': 3, 'name': 'John'},
         'id': 155,
         'parent_id': None,
         'summary': 'Run script: foo',
         'type': 'ActivityGroup'}

    The following errors may be raised:

    - **UnknownScript**: No script with the given id could be found.
    - **RequiresUserError**: The script does not have a user to run as, so
      the user must be specified when executing the script.
    """

    schema = Schema(
        name="ExecuteScript",
        parameters=[
            Unicode(
                "query",
                doc=(
                    "A query string used to select the computers to "
                    "execute the script on.   Multiple occurrences will "
                    "be joined with a logical AND."
                ),
            ),
            Integer(
                "script_id", doc="The identity of the script stored in the server."
            ),
            Unicode(
                "username",
                optional=True,
                doc=(
                    "The username to execute the script as on the "
                    "client. Required if the script has no default "
                    "username."
                ),
            ),
            Unicode(
                "deliver_after",
                optional=True,
                doc="A time in the future to deliver the script.",
            ),
        ],
        errors=[
            InvalidTimeError,
            InvalidQueryError,
            UnknownComputersError,
            UnknownScriptError,
            RequiresUserError,
        ],
        doc="""Execute a script on computers.""",
    )

    requires = "ExecuteScript"

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        """
        Find all computers matching query and execute the Script identified by
        script_id on them.
        """
        management = self.get_management(ScriptManagement, call)

        deliver_after = None
        if call.args.deliver_after:
            try:
                deliver_after = parse_datetime(call.args.deliver_after)
            except ValueError:
                raise InvalidTimeError(
                    "Invalid format for deliver_after: '%s'" % call.args.deliver_after
                )

        script = management.get_scripts(ids=[call.args.script_id]).one()
        if script is None:
            raise UnknownScriptError(call.args.script_id)

        if script.username is None and not call.args.username:
            raise RequiresUserError(
                "script_id '%d' requires a username to " "run." % call.args.script_id
            )
        script_user = script.username or call.args.username
        script_execution = ScriptExecution(context)
        activity = script_execution.execute(
            management.person,
            script.title,
            script.interpreter,
            script.code,
            script_user,
            script.time_limit,
            {},
            list(script.attachments),
            deliver_after=deliver_after,
        )

        return activity.get_state()


@method
class RemoveScript(AccountMethod):
    """
    Remove a stored script.

    This method takes one mandatory argument:

    - **script_id**: The id of a stored script identities in the server.

    For example, the following request removes stored script 1::

        ?action=RemoveScript&script_id=1

    The following errors may be raised:

    - **UnknownScript**: No script with the given id could be found.
    - **Unauthorised**: The user doesn't have permission to remove the
      specified script.
    """

    requires = "EditScript"

    schema = Schema(
        name="RemoveScript",
        parameters=[Integer("script_id", doc="The identity of the script to remove.")],
        errors=[UnknownScriptError, Unauthorized],
        doc="""Remove scripts.""",
    )

    route = SubjectRoute(
        factory=ScriptManagement, getter="get_script", identifier_arg="script_id"
    )

    def query(self, context, call):
        management = self.get_management(ScriptManagement, call)
        management.remove_scripts([context])
        return {}


@method
class CreateScript(AccountMethod, ScriptBodyConversionMixin):
    """
    Create a new script.

    - **title**: The title you wish to give the new script.
    - **time_limit**: The time, in seconds, after which the script is
      considered defunct. The process will be killed, and the script execution
      will be marked as failed after this limit expires.
    - **code**: The content of the script, encoded in UTF-8, then encoded in
      base64.
    - **username**: The system username to execute the script as. (optional)
    - **access_group**: The name of the access group to create the new script
      into. (optional)

    For example, the following request creates a new script::

        ?action=CreateScript&title=Example&interepreter=python
            &time_limit=200&code=aGVsbG8=

    The following errors may be raised:

    - **DuplicateScript**: A script with this title already exists in the
      account.
    - **ScriptEncoding**: The code (script body) is not UTF-8 decodeable.
    - **EmptyScriptCode**: The code (script body) is of zero length or
      commented.
    - **EmptyScriptInterpreter**: The code (script body) does not start with
      a script interpreter. (eg: #!/bin/bash)
    - **UnknownAccessContext**: The specified access context could not be found
      for the current script.
    - **Unauthorised**: The user doesn't have permission to create the
      specified script.

    The method returns a JSON serialized description of the script, like the
    following result::

        {"creator": {"name": "John Smith", "id": 12345, "email":
                     "john@example.com"}
         "interpreter": "/bin/bash",
         "title": "some script",
         "time_limit": 300,
         "username": "root",
         "id": 12345,
         "access_group": "somegroup",
         "attachments": ["file 1"]}

    Note: the script content is not returned in this representation, please
    use the GetScriptCode API method to retrieve it instead.
    """

    requires = "ViewAccount"

    schema = Schema(
        name="CreateScript",
        parameters=[
            UnicodeTitle("title", doc="The title of the new script."),
            Integer(
                "time_limit",
                doc="Amount of time to wait for the process to end.",
                min=MIN_TIME_LIMIT,
                max=MAX_TIME_LIMIT,
            ),
            Data("code", doc="The filename holding the script contents."),
            Unicode(
                "username",
                optional=True,
                doc="The user to execute " "the script as.",
            ),
            Unicode(
                "access_group",
                optional=True,
                doc=("The access group for the new script."),
            ),
        ],
        errors=[
            UnknownAccessContextError,
            ScriptEncodingError,
            DuplicateScriptError,
            Unauthorized,
            EmptyScriptCodeError,
            EmptyScriptInterpreterError,
        ],
        doc="""Create a new script.""",
    )

    route = ManagementRoute(factory=ScriptManagement)

    def query(self, context, call):
        script_management = context

        code = self.get_script_body_from_code(call.args.code)

        create_args = {
            "title": call.args.title,
            "code": code,
            "time_limit": call.args.time_limit,
        }
        if call.args.username is not None:
            create_args["username"] = call.args.username
        if call.args.access_group is not None:
            create_args["access_context_name"] = call.args.access_group

        script = script_management.create_script(**create_args)

        return script.get_state()


@method
class EditScript(AccountMethod, ScriptBodyConversionMixin):
    """
    Edit an existing script.

    - **script_id**: The id of the script that will be edited.
    - **title**: The new title you wish to give to the script (optional).
    - **time_limit**: The new time, in seconds, after which the script is
      considered defunct. The process will be killed, and the script execution
      will be marked as failed after this limit expires (optional).
    - **code**: The content of the script, encoded in UTF-8, then encoded in
      base64 (optional).
    - **username**: The new system username to execute the script as
      (optional).

    For example, the following request edits the script with id 2::

        ?action=EditScript&script_id=2&title=NewTitle&interepreter=newpython
            &time_limit=999&username=foo&code=aGVsbG8

    The following errors may be raised:

    - **UnknownScript**: No script with the given id could be found.
    - **DuplicateScript**: A script with this title already exists in the
      account.
    - **ScriptEncoding**: The code (script body) is not UTF-8 decodeable.
    - **EmptyScriptCode**: The code (script body) is of zero length or
      commented.
    - **EmptyScriptInterpreter**: The code (script body) does not start with
      a script interpreter. (eg: #!/bin/bash)
    - **Unauthorised**: The user doesn't have permission to remove the
      specified script.

    The method returns a JSON serialized description of the new script, like
    the following result::

        {"creator": {"name": "John Smith", "id": 12345, "email":
                     "john@example.com"}
         "interpreter": "newpython",
         "title": "NewTitle",
         "time_limit": 999,
         "username": "foo",
         "id": 12345,
         "access_group": "somegroup",
         "attachments": ["file 1"]}

    Note: the script content is not returned in this representation, please
    use the GetScriptCode API method to retrieve it instead.
    """

    requires = "EditScript"

    schema = Schema(
        name="EditScript",
        parameters=[
            Integer(
                "script_id", doc="The identifier of the script you wish " "to edit."
            ),
            UnicodeTitle("title", doc="The new script title.", optional=True),
            Integer(
                "time_limit",
                doc="Amount of time to wait for the process to end.",
                optional=True,
                min=MIN_TIME_LIMIT,
                max=MAX_TIME_LIMIT,
            ),
            Data(
                "code", doc="The filename holding the script contents.", optional=True
            ),
            Unicode(
                "username", doc="The user to execute the script as.", optional=True
            ),
        ],
        errors=[
            UnknownScriptError,
            DuplicateScriptError,
            Unauthorized,
            ScriptEncodingError,
            EmptyScriptCodeError,
            EmptyScriptInterpreterError,
        ],
        doc="""Edit a script.""",
    )

    route = SubjectRoute(
        factory=ScriptManagement, getter="get_script", identifier_arg="script_id"
    )

    def query(self, context, call):
        code = None
        if call.args.code is not None:
            code = self.get_script_body_from_code(call.args.code)

        edit_args = {
            "script_id": call.args.script_id,
            "title": call.args.title,
            "code": code,
            "time_limit": call.args.time_limit,
        }
        if call.args.username is not None:
            edit_args["username"] = call.args.username

        script_management = self.get_management(ScriptManagement, call)
        script = script_management.edit_script(call.args.script_id, edit_args)
        if script == None:
            # No changes were made.
            script = script_management.get_script(call.args.script_id)

        return script.get_state()


@method
class GetScriptCode(AccountMethod):
    """
    Retrieve the code portion of a given script.

    - **script_id**: The id of the script you wish to get the code for.

    The following errors may be raised:

    - **UnknownScript**: No script with the given id could be found.

    For example, the following request will get the code for a given script::

        ?action=GetScriptCode&script_id=1

    The method returns a JSON serialized representation of the new script::

        "import sys\\nprint sys.version"
    """

    requires = "ViewAccount"

    schema = Schema(
        name="GetScriptCode",
        parameters=[
            Integer(
                "script_id",
                doc="The identity of the script you wish to get the code " "for.",
            )
        ],
        errors=[UnknownScriptError],
        doc="""Retrieve the code portion of a given script.""",
    )

    route = SubjectRoute(
        factory=ScriptManagement, getter="get_script", identifier_arg="script_id"
    )

    def query(self, context, call):
        script = context
        return script.get_executable_code()


@method
class CopyScript(AccountMethod):
    """
    Copy an existing script to a script with a new title.

    Arguments:

    - **script_id**: The id of an existing script to copy.
    - **destination_title**: The title of the new script.
    - **access_group**: The access group to place the script in.
      If none is given, the script will be placed in the same access group
      as the existing script. (optional)

    The following errors may be raised:

    - **UnknownScript**: No script with the given id could be found or do not
      have access to the script.
    - **UnknownAccessGroup**: The specified access group could not be found.
    - **DuplicateScript**: The script title is not unique.
    - **Unauthorised**: The user doesn't have permission to create a script
      in the destination access group.

    For example, the following request creates a script from an existing
    script::

        ?action=CopyScript&script_id=1&destination_title=ScriptCopy

    The method returns a JSON serialized representation of the new script::

        {"creator": {"name": "John Smith", "id": 12345, "email":
                     "john@example.com"}
         "interpreter": "/bin/bash",
         "title": "ScriptCopy",
         "time_limit": 300,
         "username": "root",
         "id": 2,
         "access_group": "somegroup",
         "attachments": ["file 1"]}
    """

    requires = "ViewAccount"

    schema = Schema(
        name="CopyScript",
        parameters=[
            Integer("script_id", doc="The identity of the existing script."),
            Unicode("destination_title", doc="The title of the new script."),
            Unicode(
                "access_group",
                optional=True,
                doc=(
                    "The access group for the new script. It defaults "
                    "to the same access group as the existing script."
                ),
            ),
        ],
        errors=[
            UnknownScriptError,
            UnknownAccessContextError,
            DuplicateScriptError,
            Unauthorized,
        ],
        doc="""Copy an existing script to a script with a new name.""",
    )

    route = ManagementRoute(factory=ScriptManagement)

    def query(self, context, call):
        """Copies an existing script."""
        script = context.copy_script(
            call.args.script_id, call.args.destination_title, call.args.access_group
        )
        return script.get_state()


@method
class RemoveScriptAttachment(AccountMethod):
    """
    Remove a script attachment from a given script.

    Arguments:

    - **script_id**: The id of the script the attachment belongs to.
    - **filename**: The filename of the attachment you wish to delete.

    The following errors may be raised:

    - **UnknownScript**: No script with the given id could be found or do not
      have access to the script.
    - **UnknownScriptAttachment**: An attachment with the specified filename
      could not be found in this script.
    - **Unauthorised**: The user doesn't have permission to remove the
      attachment.

    For example, the following request deletes an attachment named "foo.jpg"
    from a script::

        ?action=RemoveScriptAttachment&script_id=1&filename=foo.jpg
    """

    requires = "EditScript"

    schema = Schema(
        name="RemoveScriptAttachment",
        parameters=[
            Integer("script_id", doc="The identity of the script to remove."),
            Unicode("filename", doc="The filename of the attachment to remove."),
        ],
        errors=[UnknownScriptError, UnknownScriptAttachmentError],
        doc="""Remove a script attachment.""",
    )

    route = SubjectRoute(
        factory=ScriptManagement, getter="get_script", identifier_arg="script_id"
    )

    def query(self, context, call):
        script = context
        filename = call.args.filename
        attachment = script.attachments.find(
            filename=filename, account_id=script.account_id
        ).one()
        if attachment is None:
            raise UnknownScriptAttachmentError(str(script.id), filename)
        script.remove_attachments_by_ids([attachment.id])
        self.event_log_from_call(call).create_log_entry(
            script, "Attachment '%s' removed." % filename
        )
        return {}


@method
class CreateScriptAttachment(AccountMethod):
    """
    Creates a script attachment for a given script.

    Arguments:

    - **script_id**: The id of the script the attachment will be attached to.
    - **file**: The file you wish to use as an attachment. The format for this
      parameter is: <filename>$$<base64 encoded file contents>.

    The following errors may be raised:

    - **DuplicateScriptAttachment**: An attachment with the same name already
      exists for the given script.
    - **UnknownScriptAttachment**: An attachment with the specified filename
      could not be found in this script.
    - **TooManyScriptAttachments**: The script you wish the attachment to
      belong to already has the maximum allowed number of scripts.
    - **ScriptAttachmentSize**: The maximum allowed size for attachments is
      already reached, or would be reached should the creation have occured.
    - **Unauthorised**: The user doesn't have permission to create the
      attachment.

    For example, the following request creates an attachment named "foo.txt"
    containing "hello"::

        ?action=CreateScriptAttachment&script_id=1&file=foo.txt$$aGVsbG8=

    The method returns the newly created attachment's filename:

        'foo.txt'
    """

    requires = "EditScript"

    schema = Schema(
        name="CreateScriptAttachment",
        parameters=[
            Integer(
                "script_id",
                doc="The identity of the script to add the attachment " "to.",
            ),
            File("file", doc="The file to attach"),
        ],
        errors=[
            UnknownScriptError,
            DuplicateScriptAttachmentError,
            TooManyScriptAttachmentsError,
            ScriptAttachmentSizeError,
        ],
        doc="""Add a script attachment.""",
    )

    route = SubjectRoute(
        factory=ScriptManagement, getter="get_script", identifier_arg="script_id"
    )

    def query(self, context, call):
        script = context
        (filename, contents) = call.args.file  # contents is binary here.
        attachment = script.add_attachment(filename, contents)
        self.event_log_from_call(call).create_log_entry(
            script, "Attachment '%s' added." % filename
        )
        return attachment.filename
