# Copyright 2005-2022 Canonical Limited. All rights reserved.
import json
import logging

from canonical.landscape.api.routes import GlobalRoute
from canonical.landscape.api.methods import Method
from canonical.txapi.server.schema import Schema, List, Unicode
from canonical.txapi.server.method import method

from zope.component import getUtility
from canonical.landscape.model.configuration import (
    ISystemConfiguration,
    LANDSCAPE_CONFIG_KEYS,
    LANDSCAPE_DEVELOPER_MODE,
    DEVELOPER_ONLY_KEYS,
)
from canonical.landscape.api.error import SettingsValueError


class SettingsMethodMixin(object):
    """
    This mixin is used by both settings methods to hold the "developer mode"
    logic.
    """

    def get_allowed_keys(self, configuration):
        """
        Return a list of keys currently available to the user.

        Different keys are "visible" to the user (reading, but writing too)
        depending on whether developer mode has been enabled.
        """
        developer_mode = configuration.get(LANDSCAPE_DEVELOPER_MODE, False)
        keys = LANDSCAPE_CONFIG_KEYS
        if developer_mode:
            keys += DEVELOPER_ONLY_KEYS
        return set(keys)


@method
class GetSettings(Method, SettingsMethodMixin):
    """Get the list of settings and their associated values.

    This method returns a JSON serialized mapping of configuration keys to
    configuration values.
    """

    requires = "ManageSettings"  # This is only granted on LDS.
    schema = Schema(
        name="GetSettings",
        parameters=[],
        errors=[],
        doc="""
    Get all settings and their value for the current LDS installation.""",
    )

    route = GlobalRoute()

    def query(self, context, call):
        # The key/values actually stored in the table
        configuration = getUtility(ISystemConfiguration)
        visible_keys = self.get_allowed_keys(configuration)
        config = dict.fromkeys(visible_keys, "")
        db_key_values = configuration.get_all()
        config.update(db_key_values)

        return json.dumps(config)


@method
class SetSettings(Method, SettingsMethodMixin):
    """Set a list of key-value configuration entries."""

    requires = "ManageSettings"  # This is only granted on LDS.
    schema = Schema(
        name="SetSettings",
        parameters=[
            List(
                "key_values",
                Unicode(),
                doc=(
                    "Key/value pairs to set, separated by '='. 'true' and "
                    "'false' strings will be interpreted as booleans."
                ),
            )
        ],
        errors=[SettingsValueError],
        doc="""
    Set configuration settings for the current LDS installation.""",
    )

    route = GlobalRoute()

    def query(self, context, call):
        configuration = getUtility(ISystemConfiguration)
        key_values = call.args.key_values

        allowed_keys = self.get_allowed_keys(configuration)
        # Developer mode is always writable.
        allowed_keys.add(LANDSCAPE_DEVELOPER_MODE)

        # Insert the valid key-value pairs in the database.
        for key_value in key_values:
            key, value = self._split_key_value(key_value)

            if key in ("idkfa", LANDSCAPE_DEVELOPER_MODE):
                # Support idkfa as setting for developer mode.
                # In doom, that cheat code gives you access to all weapons.
                # http://knowyourmeme.com/memes/iddqd
                logging.info("DEVELOPER MODE SET TO '{}'".format(value))
                key = LANDSCAPE_DEVELOPER_MODE

            if key not in allowed_keys:
                raise SettingsValueError("Unknown key: {}".format(key))

            # Support booleans
            if value == "true":
                value = True
            elif value == "false":
                value = False
            configuration.set(key, value)

        return {}

    def _split_key_value(self, key_value):
        """
        A method taking a string in "<key>=<value>" format and returning a
        tuple of <key>, <value>. It also serves as validation method.
        """
        if key_value == "=" or key_value.count("=") != 1:
            raise SettingsValueError(
                "Key/value entries must be of the form '<key>=<value>'. "
                "All passed key/values have been ignored."
            )
        key, value = key_value.split("=")
        if not key:
            raise SettingsValueError(
                "The key in a key/value pair cannot be empty. All passed "
                "key/values have been ignored."
            )
        return (key, value)
