"""Application configuration loading functionality for Landscape."""

import os
import socket
import sys
import logging
import warnings

from urllib.parse import quote as urlparse

import transaction
from storm.twisted.transact import Transactor, transact

from zope.app.appsetup.appsetup import config as appconfig
from zope.component import provideUtility, queryUtility, getSiteManager
from zope.event import notify
from zope.session.interfaces import ISessionDataContainer, IClientIdManager

# Ignoring pkg_resources-related warnings from talisker. Some
# ubuntu-packaged python packages have versions, ala 1.2.3-ubuntu1,
# that don't match PEP 440 format.
warnings.filterwarnings("ignore", message=".*invalid version", module="pkg_resources")

import talisker
import talisker.logs

from twisted.application import service, strports
from twisted.application.service import IService
from twisted.cred.portal import Portal
from twisted.internet import reactor, ssl
from twisted.internet.task import LoopingCall
from twisted.protocols.tls import TLSMemoryBIOFactory
from twisted.web import proxy
from twisted.web.server import NOT_DONE_YET
from twisted.web.guard import BasicCredentialFactory, HTTPAuthSessionWrapper

from storm.zope.zstorm import global_zstorm as zstorm

import canonical.logging.handlers
from canonical.oops.interfaces import IOOPSManager
import canonical.talisker.twisted

from canonical.landscape.environment import get_raw_config
from canonical.landscape.setup import read_config, configure_oops
from canonical.landscape.model.main.standalone import setup_lds_script_executions
from canonical.landscape.model.configuration import (
    is_development_mode,
    is_standalone,
    ISystemConfiguration,
    LANDSCAPE_KNOWN_PROXIES,
)
from canonical.landscape.model.events import ProcessStarting
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.model.main.account import (
    get_account_by_name,
    DEFAULT_STANDALONE_ACCOUNT_NAME,
)
from canonical.landscape.model.account.alerts import StandaloneScriptAlert
from canonical.landscape.prometheus.metrics import update_prometheus_saas_metrics

from canonical.landscape.ui.resource import UI
from canonical.landscape.ui.session.manager import LandscapeCookieClientIdManager
from canonical.landscape.ui.session.session import (
    persistent_data_container,
    volatile_data_container,
)
from canonical.landscape.ui.lsopenid import (
    ILandscapeOpenIDFactory,
    LandscapeOpenIDFactory,
)

from canonical.landscape.pppa import (
    ARCHIVE_REALM_NAME,
    ArchiveChecker,
    LicenseOnlyRealm,
)


__all__ = ["setup_logging", "LANDSCAPE_CONFIG", "run_application", "PACKAGE_VERSION"]


LANDSCAPE_CONFIG = "LANDSCAPE_CONFIG"

PACKAGE_VERSION = "23.03+18.2-0landscape0"

DEFAULT_LOGGING_FORMAT = "%(asctime)s.%(msecs)03d %(levelname)s %(name)s %(message)s"


class ZopeService(service.MultiService):
    """A C{MultiService} that fires the C{ProcessStarting} event."""

    def startService(self):
        notify(ProcessStarting())
        transaction.commit()
        for name, store in zstorm.iterstores():
            store.close()
        service.MultiService.startService(self)


class StandaloneScriptAlertChecker(object):
    """Periodically check the state of Standalone cron script alert."""

    interval = 60

    def __init__(self, transactor, clock=reactor):
        self.transactor = transactor
        self._clock = clock

    @transact
    def _check_alert(self):
        account = get_account_by_name(DEFAULT_STANDALONE_ACCOUNT_NAME)
        account.check_rules([StandaloneScriptAlert])

    def start(self):
        call = LoopingCall(self._check_alert)
        call.clock = self._clock
        call.start(self.interval)


class SAASPrometheusMetricUpdater(object):
    """Periodically update prometheus metrics related to hosted accounts."""

    interval = 3600  # Hourly metric updates

    def __init__(self, transactor, clock=reactor):
        self.transactor = transactor
        self._clock = clock

    @transact
    def _update_metrics(self):
        """Refresh values for prometheus metrics into appserver's registry."""
        logging.info("Begin - Updating prometheus SAAS metrics")
        update_prometheus_saas_metrics()
        logging.info("End - Updating prometheus SAAS metrics")

    def start(self):
        call = LoopingCall(self._update_metrics)
        call.clock = self._clock
        call.start(self.interval)


def setup(config, application, instance_type):
    # XXX Set Python's check-interval to a higher value than the default
    # one (100) should supposedly increase performance of threaded processes
    # like Zope. It's not clear what the optimal value should be though,
    # and it likely depends on the specific hardware the code is running on.
    # See also #322536 and these links:
    #
    # https://pypi.python.org/pypi/jarn.checkinterval
    # http://david.wglick.org/2010/on-zope-multiple-cores-and-the-gil/
    #
    # Eventually we should perform some benchmarks to verify what works
    # better for us.
    landscape_config = config.get(instance_type)
    global_config = config.get("global")
    sys.setswitchinterval(landscape_config.get("check-interval", 1200))

    features = _get_zope_features(config, instance_type)
    appconfig(landscape_config["site-definition"], features=features)

    # Setup mailing config here because we need ZCML loaded (for the IMailer
    # utility)
    if "mailer" in landscape_config:
        mailer = landscape_config["mailer"]
        if mailer == "default":
            path = "/tmp/landscape-mail-queue"
        else:
            path = landscape_config["mailer-path"]
            # Setup the zope.sendmail's QueueProcessorThread only
            # for non-default configurations, like staging production
            # and standalone. This is useful mainly if you want to
            # use bin/landscape-shell or bin/py to run some code manually
            # and you don't want the mailer thread to start behind your
            # back (it will stay there after you close the interpreter)
            from zope.sendmail.delivery import QueueProcessorThread
            from zope.sendmail.interfaces import IMailer

            mailerObject = queryUtility(IMailer, "landscape-smtp")
            thread = QueueProcessorThread()
            thread.setMailer(mailerObject)
            thread.setQueuePath(path)
            reactor.callWhenRunning(thread.start)

        # Setup the IMailDelivery utility, which will be used by the
        # canonical.landscape.mail.mail.LandscapeMailer wrapper
        from zope.sendmail.delivery import QueuedMailDelivery

        delivery = QueuedMailDelivery(path)
        provideUtility(delivery, name="landscape-mailer")

    root_service = ZopeService()

    # Give the server it's own thread pool, with the specified
    # number of threads
    threadpool = reactor.getThreadPool()
    threadpool.adjustPoolsize(maxthreads=int(landscape_config["threads"]))

    root_resource_config = {
        "devmode": "devmode" in features,
        "root-url": get_root_url(),
    }
    root_resource = UI(root_resource_config, reactor, threadpool)
    site = canonical.talisker.twisted.TaliskerSite(
        root_resource,
        max_memory_mib=int(global_config.get("max-service-memory", 0)),
    )
    service = strports.service("tcp:%s" % landscape_config["port"], site)
    service.setServiceParent(root_service)

    root_service.setServiceParent(application)

    if is_standalone():
        setup_lds_script_executions()
    else:
        # Setup the license.txt server
        license_realm = LicenseOnlyRealm()
        license_portal = Portal(license_realm, [ArchiveChecker()])
        license_resource = HTTPAuthSessionWrapper(
            license_portal, [BasicCredentialFactory(ARCHIVE_REALM_NAME)]
        )
        root_resource.putChild(b"license.txt", license_resource)


def setup_logging(
    name="root",
    fmt=DEFAULT_LOGGING_FORMAT,
    dateformat="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
    always_log_to_stdout=False,
    stdout=sys.stdout,
):
    """Setup the root logger to log to syslog.

    In addition to syslog, logs to sys.stdout if so configured or if
    we are in the development mode.

    @param name: The name of the logger.
    @param fmt: logging message format template.
    @param dateformat: logging date format template.
    @param level: logging level of messages to include in the output.
    @param always_log_to_stdout: unconditionally log to stdout.
    @param stdout: Standard output stream.
    @return: The log handler setup for logging.
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(level)

    config = get_raw_config()

    if is_development_mode() or always_log_to_stdout:
        canonical.talisker.twisted.configure_logging(name, level, stdout)

    # Get the syslog address from the service config.  If not defined,
    # default to /dev/log.
    if config.has_option("global", "syslog-address"):
        syslog_address = config.get("global", "syslog-address")
        if syslog_address.find(":") != -1:
            host, port = syslog_address.split(":", 1)
            syslog_address = (host, int(port))
    else:
        syslog_address = "/dev/log"

    try:
        # XXX 2015-01-23  Danilo re bug #1412740:
        # Because syslog implementations have limited message size,
        # we split the log message into chunks no more than 8000 bytes.
        # We can drop this and simply use SysLogHandler once we start
        # recording all the relevant debugging data in OOPSes.
        ChunkedSysLogHandler = canonical.logging.handlers.ChunkedSysLogHandler
        syslog_handler = ChunkedSysLogHandler(
            address=syslog_address, facility=logging.handlers.SysLogHandler.LOG_DAEMON
        )
    except socket.error:
        sys.exit("Unable to connect to syslog, exiting.")
    else:
        syslog_fmt = name + ": %(message)s"
        syslog_formatter = talisker.logs.StructuredFormatter(syslog_fmt)
        syslog_handler.setFormatter(syslog_formatter)
        syslog_handler.setLevel(level)
        root_logger.addHandler(syslog_handler)


def run_application(application, registry=None):

    if registry is None:
        registry = getSiteManager()

    instance_type = IService(application).name

    # Create the Chameleon cache directory if it doesn't exist, we need
    # to use get_raw_config() because the read_config below indirectly triggers
    # the import for chameleon.config, and we need to set the CHAMELEON_CACHE
    # environment *before* that import.
    raw_config = get_raw_config()
    if raw_config.has_option("landscape", "mailer-path"):
        mailer_path = raw_config.get("landscape", "mailer-path")
    else:
        mailer_path = "/tmp/mailer"
    parent_path = os.path.dirname(os.path.normpath(mailer_path))

    chameleon_cache = os.path.join(parent_path, "landscape-chameleon-cache")
    os.environ.setdefault("CHAMELEON_CACHE", chameleon_cache)
    if not os.path.exists(os.environ["CHAMELEON_CACHE"]):
        os.makedirs(os.environ["CHAMELEON_CACHE"])

    config = read_config(instance_type)

    landscape_config = config.get(instance_type)
    secret_token = landscape_config.get("secret-token", "No cookie forging please!")
    client_id_manager = LandscapeCookieClientIdManager(secret=secret_token)
    registry.registerUtility(client_id_manager, IClientIdManager)

    openid_provider_url = landscape_config.get("openid-provider-url")
    oidc_issuer = landscape_config.get("oidc-issuer")
    if openid_provider_url:
        openid_logout_url = landscape_config.get("openid-logout-url")
        openid_factory = LandscapeOpenIDFactory(openid_provider_url, openid_logout_url)
        registry.registerUtility(openid_factory, ILandscapeOpenIDFactory)
    elif oidc_issuer:
        # late import, to avoid pulling pyoidc on hosted
        from canonical.landscape.ui.oidc import LandscapeOidcFactory

        oidc_config = {
            k[5:].replace("-", "_"): v
            for k, v in list(landscape_config.items())
            if k.startswith("oidc-")
        }
        oidc_factory = LandscapeOidcFactory(oidc_config=oidc_config)
        registry.registerUtility(oidc_factory, ILandscapeOpenIDFactory)

    if instance_type == "landscape":
        data_container = persistent_data_container
    else:
        data_container = volatile_data_container
    registry.registerUtility(data_container, ISessionDataContainer)

    if "oops-key" in landscape_config:
        oops_manager = configure_oops(instance_type, config)
        registry.registerUtility(oops_manager, IOOPSManager)

    setup(config, application, instance_type)

    if instance_type == "landscape":
        if not is_standalone():
            threadpool = reactor.getThreadPool()
            transactor = Transactor(threadpool)
            SAASPrometheusMetricUpdater(transactor).start()
        known_proxies = landscape_config.get("known-proxies")
        if known_proxies is not None:
            configuration = registry.getUtility(ISystemConfiguration)
            configuration.set(
                LANDSCAPE_KNOWN_PROXIES,
                [proxy.strip() for proxy in known_proxies.split(",")],
            )
            get_main_store().commit()

    if is_standalone():
        threadpool = reactor.getThreadPool()
        threadpool.adjustPoolsize(maxthreads=1)
        transactor = Transactor(threadpool)
        StandaloneScriptAlertChecker(transactor).start()

    return config


def _get_zope_features(config, instance_type):
    """Return a tuple of features to enable in Zope configuration.

    The returned features are based on values in a provided config
    object.
    """
    features = ()
    if config.get(instance_type, "devmode", default="") == "on":
        features += ("devmode",)
        logging.warning(
            "Developer mode is enabled: this is a security risk and "
            "should NOT be enabled on production servers. Developer mode can "
            "be turned off in etc/zope.conf"
        )

    deployment = config.get("global", "deployment-mode", default="standalone")
    if deployment == "standalone":
        features += ("standalone",)

    return features
