# Copyright 2005-2022 Canonical Limited. All rights reserved.

from twisted.internet import reactor
from twisted.internet.task import deferLater
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.web.http import NOT_FOUND, INTERNAL_SERVER_ERROR
from twisted.python import log

from storm.twisted.transact import transact

from canonical.landscape.security.authentication import COMPUTER_ID_HEADER
from canonical.landscape.model.main.computer import get_computer_by_secure_id
from canonical.landscape.model.script.script import (
    AttachmentManagement,
    get_attachment_chunk,
)


class AttachmentResource(Resource):
    """A L{Resource} delivering attachments from the database."""

    CHUNK_SIZE = 8192

    clock = reactor

    def __init__(self, transactor):
        Resource.__init__(self)
        self.transactor = transactor

    def getChild(self, path, request):
        return self

    def render(self, request):
        secure_id = request.getHeader(COMPUTER_ID_HEADER)
        if secure_id is None:
            self.not_found(request)
            return NOT_DONE_YET

        try:
            path = request.prepath[-1]
        except IndexError:
            self.not_found(request)
            return NOT_DONE_YET

        try:
            attachment_id = int(path)
        except ValueError:
            self.not_found(request)
            return NOT_DONE_YET

        deferred = self.get_attachment(secure_id, attachment_id, request)
        deferred.addErrback(self.server_error, request)
        return NOT_DONE_YET

    def not_found(self, request):
        """Set the response code to L{NOT_FOUND} and close the request.

        It has to happen in the main thread.
        """
        request.setResponseCode(NOT_FOUND)
        request.finish()

    def server_error(self, error, request):
        """
        Log the error, set the response code to L{INTERNAL_SERVER_ERROR} and
        close the request.
        """
        log.err(error)
        request.setResponseCode(INTERNAL_SERVER_ERROR)
        request.finish()

    def get_attachment(self, secure_id, attachment_id, request):
        """
        Authenticate the computer with the L{secure_id}, and retrieve the
        attachment C{attachment_id}.
        """
        deferred = self._get_attachment(secure_id, attachment_id)

        def _continue_get_attachment(result):
            if result is None:
                return self.not_found(request)
            filename, size, store_name, data_oid = result
            request.setHeader(
                "Content-Disposition", 'attachment; filename="%s"' % filename
            )
            request.setHeader("Content-Length", str(size))
            return deferLater(
                self.clock,
                0,
                self.get_chunk,
                request,
                store_name,
                data_oid,
                self.CHUNK_SIZE,
                0,
            )

        return deferred.addCallback(_continue_get_attachment)

    @transact
    def _get_attachment(self, secure_id, attachment_id):
        computer = get_computer_by_secure_id(secure_id)
        if computer is None:
            return

        management = AttachmentManagement(computer.account)
        attachment = management.get_attachment(attachment_id)
        if attachment is None:
            return

        filename = attachment.filename.replace('"', '\\"')
        size = attachment.get_size()
        return (filename, size, computer.account.get_store_name(), attachment.data_oid)

    def get_chunk(self, request, store_name, data_oid, size, position):
        """
        Retrieve a chunk of the attachment file from database, and write it on
        the request.
        """
        deferred = self._get_chunk(store_name, data_oid, size, position)

        def _continue_get_chunk(chunk):
            request.write(chunk)
            if len(chunk) < size:
                request.finish()
            else:
                return deferLater(
                    self.clock,
                    0,
                    self.get_chunk,
                    request,
                    store_name,
                    data_oid,
                    size,
                    position + size,
                )

        return deferred.addCallback(_continue_get_chunk)

    @transact
    def _get_chunk(self, store_name, data_oid, size, position):
        return get_attachment_chunk(store_name, data_oid, size, position)
