# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import glob

from configparser import ConfigParser

from canonical.landscape.constants import ROOT_DIR


__all__ = [
    "get_config_name",
    "get_modules",
    "get_hash_id_databases_dir",
    "get_raw_config",
    "get_config_directory",
    "get_config_file_path",
]

LANDSCAPE_CONFIG = "LANDSCAPE_CONFIG"
SERVICE_CONF = "/etc/landscape/service.conf"


def get_config_name():
    # Any value set in the service.conf file will override the
    # environment variable.
    if os.path.exists(SERVICE_CONF):
        config = ConfigParser()
        config.read(SERVICE_CONF)
        if config.has_option("global", "deployment-mode"):
            return config.get("global", "deployment-mode")
    raw = os.environ.get(LANDSCAPE_CONFIG, "default")
    return raw.split("-", 1)[0]


def get_modules(patterns):
    """Get the module or package objects matching the given glob patterns.

    This function will return a list of module objects located under
    the 'canonical.landscape' tree and matching the given patterns. For
    example:

    get_modules('model/*')

    will return all packages or modules located directly under 'model'.
    """
    landscape = os.path.dirname(__file__)
    root = os.path.normpath(os.path.join(landscape, "..", ".."))

    matches = []
    for pattern in patterns:
        matches += glob.glob(os.path.join(landscape, *pattern.split("/")))

    modules = []
    for match in matches:
        path = match[len(root) + 1 :]
        if path.endswith(".py"):
            path = path[: -len(".py")]
        path = ".".join(path.split("/"))
        modules.append(__import__(path, fromlist=path.split(".")[:-1]))
    return modules


def get_config_directory(landscape_root):
    """
    Return the directory containing configuration files based on the
    current environment.
    """
    if os.environ.get("LANDSCAPE_CONFIG_DIR"):
        # XXX This environment variable is used by integration tests in order
        #     to set a configuration directory different than the stock one.
        #     Eventually we should rather support passing this as command line
        #     option.
        return os.environ["LANDSCAPE_CONFIG_DIR"]
    return os.path.join(landscape_root, "configs", get_config_name())


def get_hash_id_databases_dir():
    """
    Read the C{database-dir} config from the C{hash-id-databases.conf} for this
    instance.

    @return: The C{path} to the C{hash-id-databases} directory.
    """
    config_dir = get_config_directory(ROOT_DIR)
    config_filename = os.path.join(config_dir, "hash-id-databases.conf")

    config = ConfigParser()
    config.readfp(open(config_filename))
    return config.get("DEFAULT", "database-dir")


def get_raw_config():
    """Return a C{ConfigParser} instance for the current environment."""
    config_dir = get_config_directory(ROOT_DIR)
    config_file = os.path.join(config_dir, "service.conf")
    cp = ConfigParser()
    cp.read(config_file)
    return cp


def get_config_file_path(name):
    """Return the path to the given config file name."""
    return os.path.join(get_config_directory(ROOT_DIR), name)
