import asyncio
from configparser import ConfigParser
import datetime
import json
import logging
import os
import sys
import urllib.parse

import aio_pika

try:
    import grpc.aio as grpc_aio
except ModuleNotFoundError:
    import grpc.experimental.aio as grpc_aio

from canonical.landscape.environment import get_raw_config
from canonical.lib.text import password_decode

from hostagent_api import hostagent_pb2
from hostagent_api import hostagent_pb2_grpc


def get_amqp_config():
    config_parser = get_raw_config()
    host = config_parser.get("broker", "host", fallback="localhost")
    port = config_parser.get("broker", "port", fallback=5672)
    user = config_parser.get("broker", "user", fallback="guest")
    password = password_decode(
        config_parser.get("broker", "password", fallback="guest")
    )
    vhost = config_parser.get("broker", "hostagent_virtual_host", fallback="/")
    # / is the default vhost for RabbitMQ and needs to be escaped as %2F
    # so we override the safe parameter's default of "/"
    quoted_vhost = urllib.parse.quote(vhost, safe="")
    amqp_url = f"amqp://{user}:{password}@{host}:{port}/{quoted_vhost}"
    task_queue = config_parser.get(
        "broker",
        "hostagent_task_queue",
        fallback="landscape-server-hostagent-task-queue",
    )

    return dict(
        host=host,
        port=port,
        user=user,
        password=password,
        vhost=vhost,
        amqp_url=amqp_url,
        task_queue=task_queue,
    )


def get_grpc_server_config():
    # There are many grpc configuration parameters.
    # See https://grpc.github.io/grpc/core/group__grpc__arg__keys.html
    # They all start with "grpc." and the config file should match.
    # Some values must be sent as integers and not as strings.
    # We just convert all digit string values to integers,
    # rather than use the `getint` method of `ConfigParser`
    # so we don't have to write separate handling rules for every
    # individual grpc configuration parameter.
    config_parser = get_raw_config()
    return [
        (key, int(value)) if value.isdigit() else (key, value)
        for key, value in config_parser.items("grpc")
    ]


class LandscapeHostAgentServicer(hostagent_pb2_grpc.LandscapeHostAgentServicer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.amqp_config = get_amqp_config()
        self.server_options = get_grpc_server_config()

    async def Connect(self, request_iterator, context):
        async for request in request_iterator:
            if not request.uid:
                command = self.assign_uid_command()
                yield command
            else:
                await self.handle_request(request)
                uid = request.uid
                async for c in self.get_commands(uid):
                    yield c

    def assign_uid_command(self):
        import uuid

        uid = uuid.uuid4().hex
        message = json.dumps({"assign_host": uid})
        command = self.get_command(message)
        return command

    async def handle_request(self, request):
        json_serialized_request = self.serialize_request(request)
        connection = await aio_pika.connect_robust(url=self.amqp_config["amqp_url"])
        channel = await connection.channel()
        await channel.declare_queue(self.amqp_config["task_queue"])
        await channel.default_exchange.publish(
            aio_pika.Message(
                body=json_serialized_request,
            ),
            routing_key=self.amqp_config["task_queue"],
        )
        await connection.close()

    def serialize_request(self, request):
        request_as_dict = dict(
            hostname=request.hostname,
            token=request.token,
            uid=request.uid,
            instances=self.serialize_instances(request.instances),
        )
        return json.dumps(request_as_dict).encode()

    def serialize_instances(self, instances):
        return [
            dict(
                id=instance.id,
                instance_state=instance.instance_state,
                name=instance.name,
                version_id=instance.version_id,
            )
            for instance in instances
        ]

    async def get_commands(self, uid):
        connection = await aio_pika.connect_robust(url=self.amqp_config["amqp_url"])

        async with connection:
            channel = await connection.channel()
            queue = await channel.declare_queue(uid, auto_delete=False)

            async with queue.iterator() as queue_iterator:
                async for message in queue_iterator:
                    async with message.process():
                        yield self.get_command(message.body)

    def get_command(self, message_body):
        command_dict = json.loads(message_body)
        command = None

        if "assign_host" in command_dict:
            command = hostagent_pb2.Command(
                assign_host=hostagent_pb2.Command.AssignHost(
                    uid=command_dict["assign_host"],
                )
            )
        elif "install" in command_dict:
            command = hostagent_pb2.Command(
                install=hostagent_pb2.Command.Install(
                    id=command_dict["install"],
                )
            )
        elif "set_default" in command_dict:
            command = hostagent_pb2.Command(
                set_default=hostagent_pb2.Command.SetDefault(
                    id=command_dict["set_default"],
                )
            )
        elif "shutdown_host" in command_dict:
            command = hostagent_pb2.Command(
                shutdown_host=hostagent_pb2.Command.ShutdownHost()
            )
        elif "start" in command_dict:
            command = hostagent_pb2.Command(
                start=hostagent_pb2.Command.Start(
                    id=command_dict["start"],
                )
            )
        elif "stop" in command_dict:
            command = hostagent_pb2.Command(
                stop=hostagent_pb2.Command.Stop(
                    id=command_dict["stop"],
                )
            )
        elif "uninstall" in command_dict:
            command = hostagent_pb2.Command(
                uninstall=hostagent_pb2.Command.Uninstall(
                    id=command_dict["uninstall"],
                )
            )
        else:
            logging.error(f"Unrecognized command: {command_dict}")
        return command

    async def serve(self):
        self._server = grpc_aio.server(options=self.server_options)
        hostagent_pb2_grpc.add_LandscapeHostAgentServicer_to_server(
            LandscapeHostAgentServicer(),
            self._server,
        )
        self._port = self._server.add_insecure_port("[::]:50051")
        await self._server.start()
        return self._server, self._port


async def run_server():
    logging.basicConfig(level=logging.INFO)
    server = LandscapeHostAgentServicer()
    server, port = await server.serve()
    await server.wait_for_termination()


if __name__ == "__main__":
    asyncio.run(run_server())
