// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.activity.detail", function(Y) {

Y.namespace("landscape.activity");


/**
 * Widgets for interacting with activity detail.
 *
 * @module landscape.activity.detail
 */

var CONTENT_BOX = "contentBox",
    GRAPH_WIDTH = "graphWidth",
    ANIM_DURATION = "animDuration",
    WIDTH = "width",
    HEIGHT = "height",
    SPACE = " ",
    DOT = ".",
    NOTE = "-note",
    TABLE_NOTE = "notes",
    GRAPH = ".graph",
    ACTIVITY_STATUS = "activityStatus",
    ACTIVITY_DETAIL = ".activity-detail",
    ACTIVITY_DETAIL_NOTE = ACTIVITY_DETAIL + NOTE,
    ACTIVITY_DETAIL_PROGRESS = ACTIVITY_DETAIL + "-progress",
    ACTIVITY_PROGRESS_GRAPH = ACTIVITY_DETAIL_PROGRESS + SPACE + GRAPH,
    ACTIVITY_PROGRESS_LABEL = ACTIVITY_DETAIL_PROGRESS + SPACE + GRAPH + "-label",
    ACTIVITY_PROGRESS_FILL = GRAPH + "-fill",
    LEVEL_SUCCESS = "success",
    LEVEL_WARNING = "warning",
    LEVEL_PROBLEM = "problem",
    STATUS_CODES = ["good", "warning", "bad"],
    UPDATE_ON_RENDER = "updateOnRender",
    RENDERED = "rendered",
    TABLE_ROW = "table-row",
    TR = "tr",
    DISPLAY = "display",
    TEXT = "text",
    SRC_NODE = "srcNode",
    SELECT_ALL_CONTROLS = "selectAllControls",
    SELECT_NONE_CONTROLS = "selectNoneControls",
    SELECTED_ACTIVITY_IDS = "selectedActivityIds",
    AVAILABLE_ACTIVITY_IDS = "availableActivityIds",

    // Local aliases.
    ActivityStatusPlugin = Y.landscape.activity.ActivityStatusPlugin;

/**
 * A widget for showing a list of root activities.
 *
 * @class ActivityDetail
 * @extends Widget
 * @constructor
 */

var ActivityDetail = function() {
    ActivityDetail.superclass.constructor.apply(this, arguments);
};

Y.mix(ActivityDetail, {
    NAME: "activity_detail",

    ATTRS: {
        /**
         * The source to use when requesting updates about the displayed
         * child activities.
         *
         * @attribute source
         * @type DataSource
         */
        source: {
            value: new Y.DataSource.IO({source: "get_activity_detail"})
        },

        /**
         * The source to use when requesting paginated data.
         *
         * @attribute recordSource
         * @type DataSource
         */
        recordSource: {
            value: new Y.DataSource.IO({source: "available"})
        },

        /**
         * Page size to be used by the pager.
         *
         * @attribute pageSize
         * @type Integer
         */
        pageSize: {
            value: 25
        },

        /**
         * Cache size to be used for the underlying activities
         * DataSource. Effectively, the number of pages to cache.
         *
         * @attribute cacheSize
         * @type Integer
         */
        cacheSize: {
            value: 10
        },

        /**
         * The width of the progress graph
         *
         * @attribute graphWidth
         */
        graphWidth: {
            value: 284
        },

        /**
         * How long should the graph animation run for.
         *
         * @attribute animDuration
         */
        animDuration: {
            value: 0.6
        },

        /**
         * An instance of ActivityStatusPlugin bound to the
         * contextBox.
         *
         * @attribute activityStatus
         * @type ActivityStatusPlugin
         */
        activityStatus: {
            value: null
        },

        /**
         * Should the widget be updated on first render?
         *
         * @attribute updateOnRender
         * @type Boolean
         */
        updateOnRender: {
            value: true
        },

        /**
         * Structure containing information about activities.
         *
         * @attribute detailsData
         * @type Object
         */
        detailsData: {
            value: null
        },

        /**
         * The total number of activities which can be rendered.
         *
         * @attribute totalActivities
         * @type Integer
         */
        totalActivities: {
            value: 0
        },

        /**
         * List of selected activity ids.
         *
         * @attribute selectedActivityIds
         * @type Array
         */
        selectedActivityIds: {
            value: []
        },

        /**
         * List of available activity ids.
         *
         * @attribute availableActivityIds
         * @type Array
         */
        availableActivityIds: {
            value: []
        },

    }
});

Y.extend(ActivityDetail, Y.Widget, {

    SELECT_ALL_TEMPLATE : ["",
"{count}",
" activities selected. ",
"<a href='#' class='select-all-confirm'>Select all ",
"{total}",
" activities</a>"].join(""),

    /**
     * Constructor for the widget. Initializes some instance variables
     * from the provided config.
     */
    initializer: function(config) {
        var srcNode = this.get(SRC_NODE),
            contentBox = this.get("contentBox"),
            recordSource = this.get("recordSource"),
            sourceSchema = {
               metaFields: {"totalResults": "totalRecords"},
               resultListLocator: "records",
               resultFields:[{key: "activity_id"},
                             {key: "status"},
                             {key: "url"},
                             {key: "summary"},
                             {key: "creation_time"},
                             {key: "computer"},
                             {key: "computer_count"},
                             {key: "creator"},
                             {key: "approval_time"},
                             {key: "delivery_time"},
                             {key: "completion_time"},
                             {key: "has_text"},
                             {key: "expandable"}]};

        // If no Notes widget was provided, initialize one ourselves.
        if (!this.get(TABLE_NOTE)) {
            var notes = new Y.landscape.notes.TableNotes({
                contentBox: this.get(CONTENT_BOX).one("#activities-table"),
                animDuration: 0.01});
            this.set(TABLE_NOTE, notes);
        }

        srcNode.plug(ActivityStatusPlugin,
                     {prefixes: ["#activity-row-",
                                 "#activity-detail-"],
                      detailsData: this.get("detailsData")});
        this.set(ACTIVITY_STATUS, srcNode.activity_status);

        if (!Y.one("#activity-selection-form")) {
            this._selectableListing = null;
            return;
        }

        this._selectableListing = new Y.landscape.listing.SelectableListing({
            contentBox: "#activity-selection-form",
            controlsRequiringMarkedRows: [
                "#cancel-activities",
                "#approve-activities",
                "#revert-activities",
                "#reapply-activities"
            ],
            selectAllControls: ["#select-all-activities"],
            selectNoneControls: ["#select-none-activities"]});

        this._expandable = new Y.landscape.listing.ActivitiesListing({
            contentBox: "#activity-selection-form",
            defaultExpanded: false});

        recordSource.plug(
            {fn: Y.Plugin.DataSourceCache, cfg: {max: 10}});
        recordSource.plug(
            {fn: Y.Plugin.DataSourceJSONSchema, cfg: {schema: sourceSchema}});

        var sourceArguments = {detail: true};
        if (!Y.Lang.isUndefined(config.activitySearchCriteria)) {
            sourceArguments.activity_search_criteria = config.activitySearchCriteria;
        }

        // Pagination
        this._pager = new Y.landscape.listing.Pager(
            {contentBox: contentBox.one("#activity-selection-form"),
             renderFirstPage: false,
             rowRenderer: Y.bind(this._renderRecord, this),
             pageSize: this.get("pageSize"),
             recordSource: recordSource,
             recordSourceArguments: sourceArguments});

        this._pager.plug(
            {fn: Y.landscape.listing.PaginationPlugin,
             cfg: {paginationBox: contentBox.one("#pagination-controls")}});
        this._pager.set("totalRecords", this.get("totalActivities"));
        this._pager.set("records", []);
    },

    /**
     * Render the widget.
     */
    renderUI: function() {
        if (!this._selectableListing) {
            return;
        }
        this._selectableListing.renderUI();
        this._expandable.render();
        this._pager.render();
    },

    /*
     * Bind events, animations and whatnot.
     */

    bindUI: function() {
        var contentBox = this.get(CONTENT_BOX),
            graph = contentBox.one(ACTIVITY_PROGRESS_GRAPH),
            formActions = contentBox.one(".form-actions");
        this._anim = {};
        Y.Array.each(STATUS_CODES, function(code) {
            this._anim[code] = new Y.Anim({
                node: graph.one(".bar-body-status ." + code),
                duration: this.get(ANIM_DURATION),
                easing: Y.Easing.easeOut});
        }, this);

        if (formActions) {
            formButtons = formActions.all("input");
            formButtons.on("click", this._onFormButtonClick, this);
        }
        if (!this._selectableListing) {
            return;
        }
        this._selectableListing.bindUI();
        this._selectableListing.on("mark-row", this._onMarkRow, this);
        this._selectableListing.on("unmark-row", this._onUnmarkRow, this);
        Y.each(this._selectableListing.get(SELECT_ALL_CONTROLS), function(node) {
            node.on("click", this._onSelectAll, this);
        }, this);
        Y.each(this._selectableListing.get(SELECT_NONE_CONTROLS), function(node) {
            node.on("click", this._onSelectNone, this);
        }, this);
        this._pager.on("before-page-change", this._collapseExpandedRows, this);
        this._pager.on("after-page-change",
                       this._selectableListing.syncUI, this._selectableListing);

    },

    /*
     * Called when a row is selected, adds the ID to the selected list.
     */
    _onMarkRow: function(evt) {
        var selectedActivityIds = this.get(SELECTED_ACTIVITY_IDS),
            row = evt.details[0].row,
            activityId = parseInt(row.one("input").get("value"), 10);

        if (Y.Array.indexOf(selectedActivityIds, activityId) == -1) {
            selectedActivityIds.push(activityId);
        }
    },

    /*
     * Called when a row is unselected, removes the ID from the selected list.
     */
    _onUnmarkRow: function(evt) {
        var selectedActivityIds = this.get(SELECTED_ACTIVITY_IDS),
            row = evt.details[0].row,
            activityId = parseInt(row.one("input").get("value"), 10),
            index = Y.Array.indexOf(selectedActivityIds, activityId);

        if (index != -1) {
            selectedActivityIds.pop(index);
        }

        var notes = this.get(TABLE_NOTE);
        var previous = Y.one(
            notes.get(CONTENT_BOX)).one("a.select-all-confirm");

        if (previous) { // remove existing note.
            previous.ancestor("tr").remove();
        }
    },

    /*
     * Called when the select-all control is clicked, include all ids.
     */
    _onSelectAll: function(evt) {
        var selectedRows = this._selectableListing.get("selectedRows");
        var selectedActivityIds = [];
        Y.each(selectedRows, function(row) {
            var value = parseInt(row.one("input[type=checkbox]").get("value"));
            selectedActivityIds.push(value);
        });

        this.set(SELECTED_ACTIVITY_IDS, selectedActivityIds.slice());

        if (selectedRows.size() == this._pager.get("totalRecords")) {
            return;
        }

        // Get the table note to render at the top of the table.
        var notes = this.get(TABLE_NOTE);
        var previous = Y.one(
            notes.get(CONTENT_BOX)).one("a.select-all-confirm");

        if (previous) { // remove existing note.
            previous.ancestor("tr").remove();
        }

        notes.addNote("info", Y.substitute(this.SELECT_ALL_TEMPLATE, {
            count: selectedRows.size(),
            total: this._pager.get("totalRecords")
        }));
        // Since td's are used as spacers in this widget, the colspan of the
        // note needs adjustment
        notes.get("contentBox").one("td").setAttribute("colspan", 7);

        // Bind the note link to our confirmation callback.
        var link = Y.one(notes.get(CONTENT_BOX)).one("a.select-all-confirm");
        link.on("click", Y.bind(function(evt) {
            evt.preventDefault();
            this._selectionConfirmAll();
        }, this));
    },

    /**
     * Confirm the selection of all computers, hiding notes if they
     * are visible.
     *
     * @method selectionConfirmAll
     * @public
     */
    _selectionConfirmAll: function() {
        var notes = this.get(TABLE_NOTE);
        notes.hideNotes();

        // Mark all available activities as selected, and no activity
        // ids as unselected.
        this.set(SELECTED_ACTIVITY_IDS,
                 this.get(AVAILABLE_ACTIVITY_IDS),
                 {internal: true});
    },

    /*
     * Called when the select-none control is clicked, clear all selected ids.
     */
    _onSelectNone: function(evt) {
        this.set(SELECTED_ACTIVITY_IDS, []);
        var notes = this.get(TABLE_NOTE);
        notes.hideNotes();
    },

    /*
     * Called when a form button is clicked, updates the form with the selected
     * activity IDs.
     */
    _onFormButtonClick: function(evt) {
        var contentBox = this.get("contentBox"),
            selectionInput = contentBox.one("#selected_ids");
        selectionInput.set("value", "" + this.get(SELECTED_ACTIVITY_IDS));
    },

    /*
     * Collapse expanded rows. This is called before switching page, so we
     * reset to a clear collapsed state.
     */
    _collapseExpandedRows: function() {
        var pagerBox = this._pager.get("contentBox"),
            expandedRows = pagerBox.all(".landscape-expandable-expanded");
        expandedRows.each(function(row, index) {
            row.removeClass("landscape-expandable-expanded");
            row.next().remove();
        });

        var notes = this.get(TABLE_NOTE);
        notes.hideNotes();
    },

    /*
     * Make a request to get the activity details and populate the widget.
     */
    syncUI: function() {
        var rendered = this.get(RENDERED),
            update = this.get(UPDATE_ON_RENDER),
            source = this.get("source"),
            request;

        if (update) {
            // Flush the cache, because the status of previously retrived
            // activities might have changed.
            var recordSource = this.get("recordSource");
            recordSource.cache.flush();

            // Force the page to reload and re-render rows for current page
            this._pager._fetchPageRows();
            // Perform a request and then update the UI.
            request = "?" + this._pager.getRequestQueryString();
            source.sendRequest({request: request, callback: {
                success: Y.bind(this._renderDetail, this)
            }});
        } else {
            if (!rendered) {
                // We were called during render, but requested to not update
                // the widget on render. Now set updateOnRender to 'true' so
                // that subsequent calls to syncUI do update the widget.
                this.set(UPDATE_ON_RENDER, true);
            }

            // Only synchronize the selectableListing, since _renderDetail
            // wasn't called.
            if (this._selectableListing) {
                this._selectableListing.syncUI();
            }
        }
    },

    /*
     * Render result from get_activity_detail.
     */
    _renderDetail: function(response) {
        response = Y.JSON.parse(response.data.responseText);

        var contentBox = this.get(CONTENT_BOX),
            graph = contentBox.one(ACTIVITY_PROGRESS_GRAPH),
            label = contentBox.one(ACTIVITY_PROGRESS_LABEL),
            percentage = response.progress.percentage,
            newOverview = response.progress.overview,
            children = response.children,
            activityStatus = this.get(ACTIVITY_STATUS);

        Y.Array.each(STATUS_CODES, function(code) {
            var barBlock = graph.one(".bar-body-status ." + code),
                oldPercentage = barBlock.getStyle("width"),
                newPercentage = newOverview[code] + "%",
                anim = this._anim[code];
            
            if (oldPercentage == newPercentage) {
                // nothing changed here
                return;
            }

            anim.set("from", {width: oldPercentage});
            anim.set("to", {width: newPercentage});
            anim.run();
        }, this);
        label.set(TEXT, percentage + "%");

        Y.Object.each(response.summary, function(value, key) {
            var node = contentBox.one(
                ACTIVITY_DETAIL_NOTE + SPACE + DOT + key + NOTE);
            if (!Y.Lang.isNull(node)) {
                if (value) {
                    node.set(TEXT, value);
                    node.ancestor(TR).removeClass("hidden").show();
                } else {
                    node.ancestor(TR).hide();
                }
            }
        });


        if (!Y.Lang.isUndefined(children)) {
            Y.Array.each(children, function(child) {
                this.update(child.id, child.status, child.flags);
            }, activityStatus);
            this._selectableListing.syncUI();
        }
    },

    /**
     * Render a single child activity row.
     */
    _renderRecord: function(row, record, index) {
        var detailsData = this.get("detailsData"),
            action = row.one(".activity-action").one("input"),
            expand = row.one(".activity-expand"),
            icon = row.one(".activity-status-icon"),
            label = row.one(".activity-status-label"),
            summary = row.one(".activity-summary"),
            link = row.one(".activity-link"),
            computer = row.one(".activity-computer"),
            selectedActivityIds = this.get(SELECTED_ACTIVITY_IDS),
            sprite;
        if (record.expandable) {
            expand.setHTML(
                "<div class='landscape-expandable-tree-expand no-select' />");
        } else {
            expand.setHTML("");
        }
        row.set("id", "activity-row-" + record.activity_id);
        if (action) {
            action.set("value", record.activity_id);
            if (Y.Array.indexOf(selectedActivityIds, record.activity_id) != -1) {
                action.set("checked", true);
                row.addClass("selected");
            } else {
                action.set("checked", false);
                row.removeClass("selected");
            }
        }
        icon.one("div").remove();
        sprite = new Y.landscape.ui.Sprite({image: record.status.icon});
        sprite.render(icon);
        label.set("text", record.status.title);

        if (record.approval_time || record.has_text || record.delivery_time) {
            label.replaceClass("activity-status-no-result",
                               "activity-status-result");
            label.set("href", record.url + "/result");
        } else {
            label.replaceClass("activity-status-result",
                               "activity-status-no-result");
            label.removeAttribute("href");
        }

        summary.set("text", record.summary);
        link.set("href", record.url);
        if (computer) {
            computer.set("href", record.computer.url);
            computer.set("text", record.computer.title);
        }

        // Update detailsData, so the result overlay can be displayed
        detailsData[record.activity_id] = {
            approval_time: record.approval_time,
            delivery_time: record.delivery_time,
            completion_time: record.completion_time,
            has_text: record.has_text};
    }
});

Y.landscape.activity.ActivityDetail = ActivityDetail;

}, "0.1", {"requires": ["node", "transition", "anim", "json",
                        "datasource-io", "datasource-cache",
                        "datasource-jsonschema",
                        "landscape.ui.sprite",
                        "landscape.notes",
                        "landscape.activity.status",
                        "landscape.listing.selectable",
                        "landscape.listing.activities",
                        "landscape.listing.pager"]});
