// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.activity.schedule", function(Y) {

Y.namespace("landscape.activity");


/**
 * This module provides the ActivitySchedule widget.  It allows users to
 * specify the date and time they want the activity they're creating to be
 * delivered and carried out.
 *
 * @module landscape.activity.schedule
 */


// Delivery states
var DELIVER_NOW = 0,
    DELIVER_AFTER = 1;


/**
 * ActivitySchedule allows users to pick a date/time for an activity.
 *
 * @class ActivitySchedule
 * @extends Widget
 * @constructor
 */
var ActivitySchedule = function() {
    ActivitySchedule.superclass.constructor.apply(this, arguments);
};

ActivitySchedule.NAME = "activity_schedule";

ActivitySchedule.ATTRS = {
    joinedWidgets: { value: null },
    state: { value: DELIVER_NOW },
    deliverAfterTime: { value: null },
    activityButtons: { value: null },
    deliverNowButton: { value: null },
    deliverAfterButton: { value: null },
    deliverAfterTimeInput: { value: null },
    timeZoneOffset: { value: 0 },
    statusMessage: { value: null },

    /**
     * The tab index is null to prevent focusing of the content box when the
     * widget is used.
     *
     * @attribute tabIndex
     * @type Int
     */
    tabIndex: {
        value: null
    }
};

Y.extend(ActivitySchedule, Y.Widget, {
    /**
     * Constructor logic.
     *
     * @method initializer
     * @protected
     */
    initializer: function(config) {
        this.set("joinedWidgets", [this]);

        /**
         * Fires when the widget is in a valid state and changes are safe to
         * apply.
         *
         * @event allow_changes
         */
        this.publish("allow_changes");

        /**
         * Fires when the widget is in an invalid state and changes are not
         * safe to apply.
         *
         * @event deny_changes
         */
        this.publish("deny_changes");
    },

    /**
     * Register event handlers for controls that implement activity scheduling
     * behaviour.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        var deliverNowButton = this.get("deliverNowButton");
        deliverNowButton.on("click", Y.bind(this.setDeliverNow, this));

        var deliverAfterButton = this.get("deliverAfterButton");
        deliverAfterButton.on("click", Y.bind(this.setDeliverAfter, this));

        // FIXME This is untested because we can't simulate the 'change'
        // event.
        var deliverAfterTimeInput = this.get("deliverAfterTimeInput");
        deliverAfterTimeInput.on("valueChange", Y.bind(function(evt) {
            var deliverAfterTimeInput = this.get("deliverAfterTimeInput");
            var date = deliverAfterTimeInput.get("value");
            var widgets = this.get("joinedWidgets");
            for (var i = 0; i < widgets.length; ++i) {
                var widget = widgets[i];
                deliverAfterTimeInput = widget.get("deliverAfterTimeInput");
                deliverAfterTimeInput.set("value", date);
            }
            this.setDeliverAfterTime(date);
        }, this));
    },

    /**
     * Synchronize the user interface with the widget.
     *
     * @method syncUI
     * @protected
     */
    syncUI: function(shouldFocus) {
        var state = this.get("state");
        if (state == DELIVER_NOW) {
            this._syncDeliverNow();
            this.fire("allow_changes");
        } else {
            this._syncDeliverAfter(shouldFocus);
            if (this.get("deliverAfterTime")) {
                this.fire("allow_changes");
            } else {
                this.fire("deny_changes");
            }
        }
    },

    _setStatusMessage: function(message, error) {
        var statusMessage = this.get("statusMessage");
        statusMessage.set("text", message);
        if (error) {
            statusMessage.addClass("error-note");
        } else {
            statusMessage.removeClass("error-note");
        }
    },

    _syncDeliverNow: function() {
        var deliverAfterTimeInput = this.get("deliverAfterTimeInput");

        deliverAfterTimeInput.set("disabled", true);
        deliverAfterTimeInput.set("value", "");
        this._setStatusMessage("");
        this._enableActivityButtons();
    },

    _syncDeliverAfter: function(shouldFocus) {
        var deliverAfterTimeInput = this.get("deliverAfterTimeInput");

        deliverAfterTimeInput.set("disabled", null);
        if (shouldFocus) {
            deliverAfterTimeInput.focus();
        }

        var error = false;
        var statusMessage = "";
        var deliverAfterTime = this.get("deliverAfterTime");
        if (deliverAfterTime) {
            var date = Y.DataType.Date.format(
                deliverAfterTime, {format: "%Y-%m-%d %H:%M"});
            // See _parseDate below for an explanation of this horror...
            var local_now = new Date();
            var local_date = new Date(Date.UTC(
                deliverAfterTime.getFullYear(),
                deliverAfterTime.getMonth(),
                deliverAfterTime.getDate(),
                deliverAfterTime.getHours(),
                deliverAfterTime.getMinutes() - this.get("timeZoneOffset")));
            deliverAfterTimeInput.set("value", date);
            this._enableActivityButtons();
            if (local_date < local_now) {
                statusMessage = "Warning: scheduled time is in the past.";
            }
        } else {
            var userDeliverAfterTime = deliverAfterTimeInput.get("value");
            if (userDeliverAfterTime) {
                error = true;
                statusMessage = "'" + userDeliverAfterTime + "' is invalid.";
            }
            this._disableActivityButtons();
        }
        this._setStatusMessage(statusMessage, error);
    },

    _getPrettyDate: function(date) {
        var months = {0: "Jan", 1: "Feb", 2: "Mar", 3: "Apr",
                      4: "May", 5: "Jun", 6: "Jul", 7: "Aug",
                      8: "Sep", 9: "Oct", 10: "Nov", 11: "Dec"};
        var month = months[date.getMonth()];
        var hours = date.getHours();
        var minutes = date.getMinutes();
        minutes = minutes < 10 ? "0" + minutes : minutes;
        return (month + " " + date.getDate() + ", " + date.getFullYear() +
                " at " + hours + ":" + minutes);
    },

    _enableActivityButtons: function() {
        var activityButtons = this.get("activityButtons");
        if (! activityButtons) {
            return;
        }
        for (var i = 0; i < activityButtons.length; ++i) {
            var activityButton = activityButtons[i];
            activityButton.set("disabled", null);
        }
    },

    _disableActivityButtons: function() {
        var activityButtons = this.get("activityButtons");
        if (! activityButtons) {
            return;
        }
        for (var i = 0; i < activityButtons.length; ++i) {
            var activityButton = activityButtons[i];
            activityButton.set("disabled", true);
        }
    },

    /**
     * Join this widget with another ActivitySchedule, so that they are both in
     * sync with each other and mirror each other's state.
     *
     * @method join
     * @public
     */
    join: function(widget) {
        widget.get("joinedWidgets").push(this);
        this.get("joinedWidgets").push(widget);
    },

    /**
     * Put the widget into 'Deliver Now' mode.  This hides the
     * deliver-after-time input box and enables activity buttons.
     *
     * @method setDeliverNow
     * @public
     */
    setDeliverNow: function() {
        var widgets = this.get("joinedWidgets");
        // Go through widgets backwards so that this one gets handled last.
        // This ensures that an element can be focused without interference
        // from other widgets.
        for (var i = widgets.length; i > 0; --i) {
            var widget = widgets[i-1];
            var deliverAfterTimeInput = widget.get("deliverAfterTimeInput");
            deliverAfterTimeInput.set("disabled", true);
            widget.set("state", DELIVER_NOW);
            widget.syncUI();
        }
    },

    /**
     * Put the widget into 'Deliver After' mode.  This shows the
     * deliver-after-time input box and disables activity buttons until a valid
     * date is provided, or until the widget is put into 'Deliver Now' mode.
     *
     * @method setDeliverAfter
     * @public
     */
    setDeliverAfter: function() {
        var widgets = this.get("joinedWidgets");
        // Go through widgets backwards so that this one gets handled last.
        // This ensures that an element can be focused without interference
        // from other widgets.
        for (var i = widgets.length; i > 0; --i) {
            var index = i - 1;
            var widget = widgets[index];
            widget.set("state", DELIVER_AFTER);
            widget.syncUI(!index);
        }
    },

    /**
     * Set the earliest time at which activity requests should be delivered to
     * clients.  If the date is valid the widget enables activity buttons and
     * set the date as the deliver-after-time.  If the date is invalid the
     * widget disables activity buttons and displays an error message.
     *
     * @method setDeliverAfterTime
     * @public
     */
    setDeliverAfterTime: function(date) {
        date = this._parseDate(date);
        var widgets = this.get("joinedWidgets");
        // Go through widgets backwards so that this one gets handled last.
        // This ensures that an element can be focused without interference
        // from other widgets.
        for (var i = widgets.length; i > 0; --i) {
            var index = i - 1;
            var widget = widgets[index];
            widget.set("deliverAfterTime", date);
            widget.syncUI(!index);
        }
    },

    _parseDate: function(date) {
        var datetime_regex = /^\s*(\d{4})-(\d{2})-(\d{2})\s+(\d{2}):(\d{2})\s*$/;
        var time_regex = /^\s*(\d{2}):(\d{2})\s*$/;
        var match = null;

        match = datetime_regex.exec(date);
        if (match) {
            return Y.landscape.util.createDate(
                parseInt(match[1], 10), // year
                parseInt(match[2], 10), // month
                parseInt(match[3], 10), // day
                parseInt(match[4], 10), // hour
                parseInt(match[5], 10), // minute
                0);
        }
        else {
            // This clunky matching logic is simply to placate JSLint's hatred
            // of assignment expressions...
            match = time_regex.exec(date);
            if (match) {
                // Timezone logic gets a bit mind-bending below (because it's
                // timezones...). "local_now" is the time now in the browser's
                // local TZ. We don't care about the browser's local TZ, but we
                // need a point in time to compare to. "now" is "local_now"
                // shifted into the TZ associated with the user's account.
                var local_now = new Date();
                var now = new Date(
                    local_now.getUTCFullYear(),
                    local_now.getUTCMonth(),
                    local_now.getUTCDate(),
                    local_now.getUTCHours(),
                    local_now.getUTCMinutes() + this.get("timeZoneOffset"));
                // "schedule" is the time, expanded to a full date-time using
                // "now", entered by the user in the form. This is also in the
                // TZ associated with the user's account.
                var schedule = Y.landscape.util.createDate(
                    now.getFullYear(),
                    now.getMonth() + 1,
                    now.getDate(),
                    parseInt(match[1], 10),
                    parseInt(match[2], 10),
                    0);
                if (schedule) {
                    // Finally, "local_schedule" which is "schedule" offset
                    // into the browser's local TZ so we can compare it to
                    // local_now. No, it's not a mistake that we're calling
                    // Date.UTC, and yet not using the getUTC* methods, or that
                    // we're subtracting the offset when we added it above.
                    // Consider "now" and "schedule" to be dumb structs with no
                    // timezone (even though the browser thinks they have one)
                    // and things should become clearer (perhaps after
                    // ingestion of various fermented beverages...)
                    var local_schedule = new Date(Date.UTC(
                        schedule.getFullYear(),
                        schedule.getMonth(),
                        schedule.getDate(),
                        schedule.getHours(),
                        schedule.getMinutes() - this.get("timeZoneOffset")));
                    if (local_schedule < local_now) {
                        schedule = new Date(
                            schedule.getFullYear(),
                            schedule.getMonth(),
                            schedule.getDate() + 1,
                            schedule.getHours(),
                            schedule.getMinutes());
                    }
                    return schedule;
                }
            }
        }
        return null;
    }
});

Y.landscape.activity.ActivitySchedule = ActivitySchedule;


}, "0.1", {"requires": ["event", "node", "widget", "intl", "event-valuechange",
                        "datatype-date-format", "landscape.util.date"]});
