// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.activity.status", function(Y) {

Y.namespace("landscape.activity");

/*
 * A plugin for managing activity status updates given a PageDataSource and a
 * container node.
 *
 * Upon receiving a notification, the plugin will look for the activity related
 * to the notification and update it's status icon and title.
 */

var IMAGE_TEMPLATE = "/static/activity/{name}-14x14.png",
    ACTION_TEMPLATE = [
        "<input name='activity_ids:list:int' ",
        "value='{value}' ",
        "type='checkbox' />"].join(""),
    ACTIVITY_STATUS_RESULT_TEMPLATE = [
        "<a href='{url}' ",
        "title='Activity details'",
        "class='activity-status-result activity-status-label no-select'>",
        "{label}",
        "</a>"].join(""),
    staticFilename = Y.landscape.ui.image.staticFilename,

    ACTIVITY = ".activity",
    ACTIVITY_ACTION = ACTIVITY + "-action",
    ACTIVITY_STATUS = ACTIVITY + "-status",
    ACTIVITY_STATUS_ICON = ACTIVITY_STATUS + "-icon",
    ACTIVITY_STATUS_LABEL = ACTIVITY_STATUS + "-label",
    ACTIVITY_STATUS_RESULT = ACTIVITY_STATUS + "-result",
    CONTENT_UPDATE = "node:contentUpdate",
    CANCELABLE = "cancelable",
    EXPANDABLE = "expandable",
    APPROVABLE = "approvable",
    REVERTABLE = "revertable",
    REAPPLIABLE = "reappliable",
    PREFIXES = "prefixes",
    FLAGS = [CANCELABLE, EXPANDABLE, APPROVABLE, REVERTABLE, REAPPLIABLE];

var ActivityStatusPlugin = function (config) {
    ActivityStatusPlugin.superclass.constructor.apply(this, arguments);
    this._node = config.host;
    this._node.publish(CONTENT_UPDATE);
};

Y.mix(ActivityStatusPlugin, {
      NAME: "activity_status",
      NS: "activity_status",
      ATTRS: {
          /**
           * List of prefixes to use when looking for activities in the page.
           *
           * @attribute prefixes
           * @type List
           */
          prefixes: {
              value: null
          },

          /**
           * Structure containing information about activities.
           *
           * @attribute detailsData
           * @type Object
           */
          detailsData: {
              value: null
          }
      }
});

Y.extend(ActivityStatusPlugin, Y.Plugin.Base, {

    /**
     * Given an activity id, find nodes associated with that activity id by
     * concatenating configured prefixes with the activity id, then update the
     * activity status nodes based on the provided status configuration and
     * flags.
     */
    update: function(activityId, status, flags) {
        var prefixes = this.get(PREFIXES),
            detailsData = this.get("detailsData");
        Y.Array.each(prefixes, function(prefix) {
            var activityNode = this._node.one(prefix + activityId);
            if (!Y.Lang.isNull(activityNode)) {
                var name = status.title.toLowerCase().replace(" ", "-"),
                    image = Y.substitute(IMAGE_TEMPLATE, {name: name}),
                    sprite, label, icon, action, flag, resultNode, anchorNode;

                if (!Y.Lang.isUndefined(flags)) {
                    Y.Array.each(FLAGS, function(flag) {
                        if (!flags[flag]) {
                            this.removeClass(flag);
                        } else {
                            if (!this.hasClass(flag)) {
                                this.addClass(flag);
                            }
                        }
                    }, activityNode);
                    // Check whether the plugin should add a result text icon
                    // when result is true, or remove it when result is false
                    // or not present.
                    if (!Y.Lang.isUndefined(flags.result)) {
                        resultNode = activityNode.one(ACTIVITY_STATUS_RESULT);
                        if (detailsData[activityId]) {
                            detailsData[activityId].has_text = flags.result;
                        }
                        var activityLink = activityNode.one(".activity-link");
                        if (Y.Lang.isNull(resultNode) && !Y.Lang.isNull(activityLink) &&
                            ["undelivered", "canceled"].indexOf(status.name) == -1) {
                            var url = activityLink.get("href") + "/result";
                            var anchor = Y.substitute(
                                ACTIVITY_STATUS_RESULT_TEMPLATE,
                                {url: url, label: status.title});
                            label = activityNode.one(ACTIVITY_STATUS_LABEL);
                            label.replace(anchor);
                        }
                    }
                }

                icon = activityNode.one(ACTIVITY_STATUS_ICON);
                // Activity detail nodes do not include an icon or other
                // details that can be updated.
                if (icon !== null) {
                    icon.one(".icon").remove();
                    sprite = new Y.landscape.ui.Sprite({image: image});
                    sprite.render(icon);

                    label = activityNode.one(ACTIVITY_STATUS_LABEL);
                    label.set("text", status.title);
                    action = activityNode.one(ACTIVITY_ACTION);

                    if (!Y.Lang.isNull(action) && !Y.Lang.isUndefined(flags)) {
                        if (!flags.actionable) {
                            action.all("input").remove();
                            action.set("text", "");
                        } else {
                            if (Y.Lang.isNull(action.one("input"))) {
                                action.set("text", "");
                                action.append(Y.substitute(
                                    ACTION_TEMPLATE, {value: activityId}));
                            }
                        }
                    }
                }
                this._node.fire(CONTENT_UPDATE);
            }
        }, this);
    }
});

Y.landscape.activity.ActivityStatusPlugin = ActivityStatusPlugin;

}, "0.1", {"requires": ["node", "attribute", "plugin",
                        "substitute", "landscape.ui.sprite"]});
