// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.activity.summary", function(Y) {

Y.namespace("landscape.activity");


/**
 * Widgets for interacting with activity summary.
 *
 * @module landscape.activity.summary
 */

/**
 * A widget for showing a list of root activities.
 *
 * @class ActivitySummary
 * @extends Widget
 * @constructor
 */

var ActivitySummary = function() {
    ActivitySummary.superclass.constructor.apply(this, arguments);
};

Y.mix(ActivitySummary, {
    NAME: "activity_summary",

    ATTRS: {
        /**
         * Datasource of activity records.
         *
         * @attribute recordSource
         * @type DataSource
         */
        recordSource: {
            value: null
        },

        /**
         * Cache size to be used for the underlying
         * DataSources. Effectively, the number of pages to cache.
         *
         * @attribute cacheSize
         * @type Integer
         */
        cacheSize: {
            value: 10
        },

        /**
         * Pager instance managed by this widget.
         *
         * @attribute pager
         * @type Pager
         */
        pager: {
            value: null
        },

        /**
         * Page size to be used by the pager.
         *
         * @attribute pageSize
         * @type Integer
         */
        pageSize: {
            value: 25
        },

        /**
         * Node where the pagination box should be rendered.
         *
         * @attribute paginationBox
         * @type Node
         */
        paginationBox: {
            value: null,
            setter: Y.landscape.util.nodeSetter
        },

        /**
         * SelectableListing instance managed by this widget.
         *
         * @attribute selectable
         * @type SelectableListing
         */
        selectableListing: {
            value: null
        },

        /**
         * Should the pager render the first page or progressively
         * enhance what's there?
         *
         * @attribute renderFirstPage
         * @type Boolean
         */
        renderFirstPage: {
            value: true
        },

        /**
         * A list of CSS selectors for controls that, when clicked,
         * should mark all activities.
         *
         * @attribute selectAllControls
         * @type Array
         */
        selectAllControls: {
            value: [],
            setter: Y.landscape.util.nodesSetter
        },

        /**
         * A list of CSS selectors for controls that, when clicked,
         * should unmark all activities.
         *
         * @attribute selectNoneControls
         * @type Array
         */
        selectNoneControls: {
            value: [],
            setter: Y.landscape.util.nodesSetter
        },

        /**
         * A list of CSS selectors for controls that should be enabled only
         * when rows are marked.
         *
         * @attribute controlsRequiringMarkedRows
         * @type Array
         */
        controlsRequiringMarkedRows: {
            value: [],
            setter: Y.landscape.util.nodesSetter
        },

        /**
         * The filter argument to select activities, if any.
         *
         * @attribute filter
         * @type String
         */
        filter: {
            value: null
        },


        /**
         * The total number of activities which can be rendered.
         *
         * @attribute totalActivities
         * @type Integer
         */
        totalActivities: {
            value: 0
        }
    }
});

Y.extend(ActivitySummary, Y.Widget, {

    STATUS_BAR_TEMPLATE: [
        "<div class='bar'>",
        "  <div class='bar-body bar-body-status'>",
        "    <div class='bar-body-block good' style='width: {good}%'></div>",
        "    <div class='bar-body-block warning' style='width: {warning}%'></div>",
        "    <div class='bar-body-block bad' style='width: {bad}%'></div>",
        "  </div>",
        "</div>"].join(""),

    /**
     * Constructor for the widget. Initializes some instance variables
     * from the provided config. Some config options are passed
     * directly to the underlying Pager instance.
     */
    initializer: function(config) {
        var contentBox = this.get("contentBox");
        var recordSource = this.get("recordSource");

        var sourceSchema = {
            metaFields: {"totalResults": "totalRecords"},
            resultListLocator: "records",
            resultFields:[{key: "activity_id"},
                          {key: "overview"},
                          {key: "url"},
                          {key: "summary"},
                          {key: "creation_time"},
                          {key: "computer"},
                          {key: "computer_count"},
                          {key: "creator"}]};

        recordSource.plug(
            {fn: Y.Plugin.DataSourceCache, cfg: {max: 10}});
        recordSource.plug(
            {fn: Y.Plugin.DataSourceJSONSchema, cfg: {schema: sourceSchema}});

        var pageSize = this.get("pageSize");
        var filter = this.get("filter");
        var sourceArguments = {};
        var cacheKey = "?size=" + pageSize;
        if (filter) {
            sourceArguments.filter = filter;
            cacheKey += "&filter=" + filter;
        }

        var renderFirstPage = this.get("renderFirstPage");

        var pagerConfig = {contentBox: contentBox,
                           renderFirstPage: renderFirstPage,
                           rowRenderer: Y.bind(this.rowRender, this),
                           pageSize: pageSize,
                           recordSource: recordSource,
                           recordSourceArguments: sourceArguments};
        var pager = new Y.landscape.listing.Pager(pagerConfig);

        // Pagination
        pager.plug({fn: Y.landscape.listing.PaginationPlugin,
                    cfg: {paginationBox: this.get("paginationBox")}});
        pager.set("totalRecords", this.get("totalActivities"));
        pager.set("records", []);
        this.set("pager", pager);

        if (!renderFirstPage) {
            var firstPageRecords = this._parsePage();
            recordSource.cache.add(
                cacheKey, {meta: {totalResults: this.get("totalActivities")},
                results: firstPageRecords});
        }

        var selectableListing = new Y.landscape.listing.SelectableListing({
            contentBox: contentBox,
            selectAllControls: this.get("selectAllControls"),
            selectNoneControls: this.get("selectNoneControls"),
            controlsRequiringMarkedRows: this.get("controlsRequiringMarkedRows")});
        this.set("selectableListing", selectableListing);
    },

    /**
     * Render the Pager widget and the SelectableListing widget as
     * part of our widget.
     */
    renderUI: function() {
        this.get("pager").render();
        this.get("selectableListing").render();
    },

    /**
     * Helper method to create the HTML template of a link.
     */
    makeLink: function(url, title) {
        var link = Y.Node.create("<a />");
        link.setAttribute("href", url);
        link.set("text", title);
        return link;
    },

    /**
     * Render one single activity row of the summary.
     */
    rowRender: function(row, record, index) {
        var columns = row.all("td");
        var cursor = 0;
        var firstColumn = columns.item(cursor);
        var checkbox = firstColumn.one("input");
        if (checkbox) {
            checkbox.set("value", record.activity_id);
            cursor += 1;
        }
        columns.item(cursor).set("innerHTML", "");
        if (record.overview) {
            columns.item(cursor).set(
                "innerHTML",
                Y.substitute(this.STATUS_BAR_TEMPLATE, record.overview));
        }
        cursor += 1;

        var link = this.makeLink(record.url, record.summary);
        columns.item(cursor).setContent(link);
        cursor += 1;

        columns.item(cursor).set("innerHTML", "");
        if (record.computer) {
            var computerLink;
            computerLink = this.makeLink(
                record.computer.url, record.computer.title);
            columns.item(cursor).setContent(computerLink);
        } else if (record.computer_count) {
            columns.item(cursor).set("text", record.computer_count);
        }
        cursor += 1;

        columns.item(cursor).set("innerHTML", "");
        if (record.creator) {
            var creatorLink = this.makeLink(
                record.creator.url, record.creator.name);
            columns.item(cursor).setContent(creatorLink);
        }
        cursor += 1;

        columns.item(cursor).set("text", record.creation_time);
    },

    /**
     * Helper method to extract the sprite name from a given node.
     */
    extractSprite: function(node) {
        var spriteDiv = node.one("div");
        if (!spriteDiv) {
            return;
        }
        var classes = spriteDiv.get("className").split(" ");
        for (var i = 0; i < classes.length; i++) {
            if (classes[i].slice(0, 6) == "sprite") {
                return classes[i].slice(7);
            }
        }
    },

    /**
     * Parse the current content of the page, returning a record list with the
     * same schema as returned by the server
     *
     * @method _parsePage
     * @private
     */
    _parsePage: function() {
        var records = [];
        var contentBox = this.get("contentBox");
        var record, columns, checkbox, statusText, cursor, bars, barBlocks;
        contentBox.all("tbody tr").each(function(row) {
            record = {};
            columns = row.all("td");
            checkbox = columns.item(0).one("input");
            cursor = 0;
            if (checkbox) {
                record.activity_id = checkbox.get("value");
                cursor += 1;
            }
            
            barBlocks = columns.item(cursor).all(".bar-body-block");
            record.overview = {
                "good": this._parseBarBlock(barBlocks.item(0)),
                "warning": this._parseBarBlock(barBlocks.item(1)),
                "bad": this._parseBarBlock(barBlocks.item(2)),
            };
            cursor += 1;

            var link = columns.item(cursor).one("a");
            record.url = link.get("href");
            record.summary = link.get("text");
            cursor += 1;

            var computerLink = columns.item(cursor).one("a");
            if (computerLink) {
                record.computer = {
                    url: computerLink.get("href"),
                    title: computerLink.get("text")};
            } else {
                record.computer_count = columns.item(cursor).get("text");
            }
            cursor += 1;

            var creatorLink = columns.item(cursor).one("a");
            if (!Y.Lang.isNull(creatorLink)) {
                record.creator = {
                    url: creatorLink.get("href"),
                    name: creatorLink.get("text")};
            }
            cursor += 1;

            record.creation_time = Y.Lang.trim(columns.item(cursor).get("text"));

            records.push(record);
        }, this);
        return records;
    },

    /**
     * Parse an overview bar block returning the associated percentage.
     */
    _parseBarBlock: function(barBlock) {
        return parseInt(barBlock.getStyle("width"));
    },

});

Y.landscape.activity.ActivitySummary = ActivitySummary;

}, "0.1", {"requires": ["node", "io-base", "cache",
                        "datasource-io", "datasource-cache",
                        "datasource-jsonschema",
                        "landscape.listing.selectable",
                        "landscape.ui.sprite", "landscape.util.node",
                        "landscape.listing.pager"]});
