// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.constraint.conditionalinput", function(Y) {

Y.namespace("landscape.constraint");

/**
 * This module provides the ConditionalInput widget, which allows users
 * to click on a checkbox that will enable an input text or disables it.
 *
 * @module landscape.constraint.conditionalinput
 */


/**
 * ConditionalInput allows users to click on a checkbox that will enable
 * an input text or disables it.
 *
 * @class ConditionalInput
 * @extends Widget
 * @constructor
 */

var ConditionalInput = function() {
    ConditionalInput.superclass.constructor.apply(this, arguments);

};


ConditionalInput.NAME = "conditional-input";

ConditionalInput.ATTRS = {
    /**
     * Optionally, the CSS error class to use when our JS widget
     * detects an error in the input designated by the class name
     * ".conditional-input-text".
     *
     * @attribute errorClass
     * @type String
     */
    errorClass: {
        value: "error-note"
    },

    /**
     * Optionally, the error message prefix that will be published at
     * the beginning of all error messages.
     *
     * @attribute errorMessagePrefix
     * @type String
     */
    errorMessagePrefix: {
        value: ""
    }
}; 

Y.extend(ConditionalInput, Y.Widget, {

    initializer: function(config) {
        var contentBox = this.get("contentBox");

        this._errorClass = this.get("errorClass");
        this._errorMessagePrefix = this.get("errorMessagePrefix");
        this._conditionalInputCheckbox = contentBox.one(".conditional-input-checkbox");
        this._conditionalInputText = contentBox.one(".conditional-input-text");
        this._errorMessage = contentBox.one(".conditional-input-error-message");
        /**
         * Fires when the widget is in a valid state and changes are safe to
         * apply.
         *
         * @event allow_changes
         */
        this.publish("allow_changes");

        /**
         * Fires when the widget is in an invalid state and changes are not
         * safe to apply.
         *
         * @event deny_changes
         */
        this.publish("deny_changes");
    },

    bindUI: function() {
        this._conditionalInputCheckbox.on("click", this._onClick, this);
        this._conditionalInputText.on("valuechange", this._onValueChange, this);
    },

    _onClick: function(evt) {
        /**
         * Handle checkbox and text input behavior.
         *
         * @method _onClick
         * @protected
         */
        var checked = this._conditionalInputCheckbox.get("checked"),
            inputValue = this._conditionalInputText.get("value");

        // Clear the error message and sets the input value to zero.
        if (this._errorMessage.hasClass(this._errorClass)) {
            this._conditionalInputText.set("value", "0");
            this._setStatus("", false);
        }

        if (checked) {
            if (inputValue !== ""  && Y.Lang.isNumber(parseInt(inputValue))) {
                // Preserve the value and enable the input.
                this._conditionalInputText.set("disabled", null);
            } else {
                // Reset the value and enable the input.
                this._conditionalInputText.set("disabled", null);
                this._conditionalInputText.set("value", "0");
            }
        } else {
            this._conditionalInputText.set("disabled", true);
        }
        this.fire("allow_changes");
    },

    _onValueChange: function(evt) {
        /**
         * Handle key strokes to allow for numbers only.
         *
         * @method _onValueChange
         * @protected
         */
        var isValidNumber = /^\d+$/.test(evt.newVal);
        var error = false;
        var statusMessage = "";

        if (evt.newVal !== "" && !isValidNumber) {
            // Display an error message if the value is not an empty string
            // and not a valid number.
            error = true;
            statusMessage = "'" + evt.newVal + "' must be a number greater than or equal to zero.";
        } else {
            // Set the text of the input with the new value.
            this._conditionalInputText.set("value", evt.newVal);
        }
        this._setStatus(statusMessage, error);
    },

    _setStatus: function(message, error) {
        /**
         * Sets an error message displaying the message that was passed.
         *
         * @method _onValueChange
         * @protected
         */
        if (error) {
            this._errorMessage.addClass(this._errorClass);
            this.fire("deny_changes");
            message = this._errorMessagePrefix + message;
        } else {
            this._errorMessage.removeClass(this._errorClass);
            this._errorMessage.set("text", "");
            this.fire("allow_changes");
        }
        this._errorMessage.set("text", message);
    }

});

Y.landscape.constraint.ConditionalInput = ConditionalInput;


}, "0.1", {"requires": ["event", "node", "widget"]});
