// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.deferred", function(Y) {

Y.namespace("landscape.deferred");

/**
 * A simplified implementation of Twisted's deferred.
 *
 * @module landscape.deferred
 */

/**
 * A landscape.deferred.Deferred is a promise that a function will at
 * some point have a result. We can attach callback functions to a
 * Deferred, and once it gets a result these callbacks will be
 * called. In addition Deferreds allow the developer to register a
 * callback for an error. The deferred mechanism standardizes the
 * application programmer's interface with all sorts of blocking or
 * delayed operations.
 *
 * @class Deferred
 * @extends Base
 * @constructor
 */

var Deferred = function(config) {
    Deferred.superclass.constructor.apply(this, arguments);
};

Deferred.NAME = "deferred";

Y.extend(Deferred, Y.Base, {

    initializer: function(config) {
        config = config || {};
        this.called = (
            Y.Lang.isUndefined(config.called) ? false : config.called);
        this.result = (
            Y.Lang.isUndefined(config.result) ? null : config.result);
        this.callbacks = (
            Y.Lang.isUndefined(config.callbacks) ? [] : config.callbacks);
    },

    /**
     *  Run all success callbacks that have been added to this
     *  Deferred.
     *
     *  Each callback will have its result passed as the first
     *  argument to the next; this way, the callbacks act as a
     *  'processing chain'. Also, if the success-callback returns an
     *  Error or raises an Exception, processing will continue on the
     *  *error*- callback chain.
     *
     * @method callback
     * @public
     */
    callback: function(result) {
        if (this.called) {
            throw new Error("Already called.");
        }
        this.called = true;
        this.result = result;
        this._runCallbacks();
    },

    /**
     * Run all error callbacks that have been added to this L{Deferred}.
     *
     * Each callback will have its result passed as the first
     * argument to the next; this way, the callbacks act as a
     * 'processing chain'. Also, if the error-callback returns a non-Error
     * or doesn't raise an Exception, processing will continue on the
     * *success*-callback chain.
     *
     * If the argument that's passed to me is not an Error instance,
     * it will be embedded in one.
     *
     * @method errback
     * @public
     */
    errback: function(error) {
        if (this.called) {
            throw new Error("Already called.");
        }
        this.called = true;
        if (!(error instanceof Error)) {
            error = new Error(error);
        }
        this.result = error;
        this._runCallbacks();
    },

    /**
     * Convenience method for adding just a callback.
     *
     * @method addCallback
     * @public
     */
    addCallback: function(callback) {
        return this.addCallbacks(callback, null);
    },

    /**
     * Convenience method for adding just an errback.
     *
     * @method addErrback
     * @public
     */
    addErrback: function(errback) {
        return this.addCallbacks(null, errback);
    },

    /**
     * Convenience method for adding a single callable as both a
     * callback and an errback.
     *
     * @method addBoth
     * @public
     */
    addBoth: function(callback) {
        return this.addCallbacks(callback, callback);
    },

    /**
     * Add a pair of callbacks (success and error) to this
     * Deferred. These will be executed when the 'master' callback is
     * run.
     *
     * @method addCallbacks
     * @public
     */
    addCallbacks: function(callback, errback) {
        this.callbacks.push([callback, errback]);
        if (this.called) {
            this._runCallbacks();
        }
        return this;
    },

    _runCallbacks: function() {
        var result = this.result;
        var callbacks = this.callbacks;
        while (callbacks.length) {
            var callback = callbacks.shift();
            var index = ((result instanceof Error) ? 1 : 0);
            var callable = callback[index];
            if (callable === null) {
                continue;
            }
            try {
                result = callable(result);
            } catch (error) {
                Y.log("Callback error: " + error, "info", "landscape.deferred");
                result = error;
            }
        }
        this.result = result;
    }
});


Y.landscape.deferred.Deferred = Deferred;

/**
 * Sometimes you want to be notified after several different events
 * have all happened, rather than waiting for each one
 * individually. For example, you may want to wait for all the
 * connections in a list to close. landscape.deferred.DeferredList is
 * the way to do this.
 *
 * @class DeferredList
 * @extends Deferred
 * @constructor
 */

var DeferredList = function(config) {
    DeferredList.superclass.constructor.apply(this, arguments);
};

DeferredList.NAME = "deferred_list";

Y.extend(DeferredList, Deferred, {

    initializer: function(config) {
        config = config || {};
        var deferredList = (
            Y.Lang.isUndefined(config.deferredList) ?
                [] : config.deferredList);
        this.fireOnOneErrback = (
            Y.Lang.isUndefined(config.fireOnOneErrback) ?
                false : config.fireOnOneErrback);

        this.resultList = [];

        var deferredCount = deferredList.length;
        this.finishedCount = deferredCount;

        for (var idx=0; idx < deferredCount; idx++) {
            this.resultList.push(undefined);
            deferredList[idx].addCallbacks(
                Y.bind(this._cbDeferred, this, idx, true),
                Y.bind(this._cbDeferred, this, idx, false));
        }
    },

    _cbDeferred: function(index, succeeded, result) {
        this.resultList[index] = [succeeded, result];
        this.finishedCount -= 1;
        if (!this.called) {
            if (!succeeded && this.fireOnOneErrback) {
                this.errback(result);
            } else if (!this.finishedCount) {
                this.callback(this.resultList);
            }
        }
        return result;
    }
});

Y.landscape.deferred.DeferredList = DeferredList;

/**
 * Returns list with result of given L{Deferred}s.
 *
 * This builds on L{DeferredList} but is useful since you don't
 * need to parse the result for success/failure.
 *
 * @function gatherResults
 */
function gatherResults(deferredList) {
    var deferred = new DeferredList({
            "deferredList": deferredList, "fireOnOneErrback": true});
    deferred.addCallback(function(results) {
        var newResults = [];
        Y.Array.each(results, function(result) {
            newResults.push(result[1]);
        });
        return newResults;
    });
    return deferred;
}

Y.landscape.deferred.gatherResults = gatherResults;


/**
 * Return a Deferred that has already had .callback(result) called.
 *
 * This is useful when you're writing synchronous code to an
 * asynchronous interface: i.e., some code is calling you expecting a
 * Deferred result, but you don't actually need to do anything
 * asynchronous. Just return succeed(theResult).
 *
 * See fail for a version of this function that uses a failing
 * Deferred rather than a successful one.
 *
 * @function succeed
 */
function succeed(result) {
    var deferred = new Deferred();
    deferred.callback(result);
    return deferred;
}

Y.landscape.deferred.succeed = succeed;


/**
 * Return a L{Deferred} that has already had C{.errback(result)}
 * called.
 *
 * See L{succeed}'s docstring for rationale.
 *
 * @function succeed
 */
function fail(result) {
    var deferred = new Deferred();
    deferred.errback(result);
    return deferred;
}

Y.landscape.deferred.fail = fail;

}, "0.1", {"requires": ["base"]});
