// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.job.creation", function(Y) {

Y.namespace("landscape.job");

/**
 * Handle form errors returned by a JobPage.
 *
 * It expects a JSON result with 2 fields:
 *  - global_errors, a list of global errors which are going to be displayed as
 *    notes.
 *  - widget_errors, a mapping of widget names to error messages, which are
 *    going to be displayed using the standard form error hint.
 *
 */
function renderFormErrors(result, form) {
    var notes = new Y.landscape.notes.Notes({contentBox: "#notes"}),
        errors;

    try {
        errors = Y.JSON.parse(result);
    } catch (error) {
        notes.addNote("error", result);
        return;
    }

    var global_errors = errors.global_errors;
    var widget_errors = errors.widget_errors;
    if (global_errors) {
        for (var errorIdx in global_errors) {
            if (global_errors.hasOwnProperty(errorIdx)) {
                notes.addNote("error", global_errors[errorIdx]);
            }
        }
    }
    for (var name in widget_errors) {
        if (widget_errors.hasOwnProperty(name)) {
            var html = widget_errors[name];
            var input = form.one("[name='" + name + "']");
            if (!Y.Lang.isNull(input) && !input.hasAttribute("disabled")) {
                var formField = input.ancestor(".form-field");
                var divError = Y.Node.create("<div class='form-error' />");
                divError.appendChild(Y.Node.create(html));
                formField.appendChild(divError);
            }
        }
    }
}

Y.landscape.job.renderFormErrors = renderFormErrors;

/**
 * Plugin for binding a job submission to a Y.landscape.ui.Button widget.
 *
 * @class Binder
 * @extends Plugin
 * @constructor
 */

var Binder = function() {
    Binder.superclass.constructor.apply(this, arguments);
};

Binder.NAME = "binder";

Binder.NS = "binder";

Binder.ATTRS = {

    /**
     * @attribute source
     * @description The page data source used to send the action.
     * @type PageDataSource
     *
     */
    source: {value: null},

    /**
     * @attribute action
     * @description The name of the job to submit.
     * @type String
     *
     */
    action: {value: null},

    /**
     * @attribute data
     * @description Optional object holding extra data to include in the job.
     * @type Object
     *
     */
    data: {value: null},

    /**
     * @attribute base
     * @description Optional base url for sending the action.
     * @type String
     *
     */
    base: {value: null},

    /**
     * @attribute success
     * @description A function to be called when the action succeeds.
     * @type function
     *
     */
    success: {value: null},

    /**
     * @attribute failure
     * @description A function to be called when the action fails. If not
     *     specified the default behavior is to use renderFormErrors to
     *     render the errors.
     * @type function
     *
     */
    failure: {value: null}
};


/**
 * Plugin for binding a job request to a form submission.
 *
 * It should be plugged into a form Y.Node and it provides 'success'
 * and 'failure' events for receiving results and errors from the
 * submitted job.
 *
 * By default errors are rendered (either as global notes for form
 * field errors).
 *
 * @class Binder
 * @extends Plugin
 * @constructor
 */

var FormJob = Y.Base.create("formJob", Y.Plugin.Base, [], {

    initializer: function(config) {
        this.onHostEvent("submit", this._onSubmit);
        this.publish("success");
        this.publish("failure", {defaultFn: this._defFailureFn});
        this.addTarget(this.get("host"));
        this._disabledInputs = [];
        this._notificationHandle = null;
    },

    /**
     * Send a job using the form input.
     */
    _onSubmit: function(e) {
        var io = this.get("io"),
            action = this.get("action"),
            notification = "notification:" + action,
            form = this.get("host");

        e.preventDefault();
        this._uiResetInputs(form);
        this._notificationHandle = Y.once(notification, this._onNotification, this);
        io(action, {
            method: "POST",
            form: {
                id: form,
                useDisabled: true,
            },
            on: {
                failure: Y.bind(this._onRequestFailure, this),
            },
        });
    },

    /**
     * Handle failed form posts by invoking the failure callback.
     */
    _onRequestFailure: function(transaction, request) {
        this._notificationHandle.detach();
        this._uiRestoreInputs();
        this.fire("failure", {data: request.response});
    },

    /**
     * Handle completed jobs. If there are no errors, the success callback
     * gets invoked, otherwise the failure one.
     */
    _onNotification: function(e) {
        var success = this.get("success"),
            failure = this.get("failure");

        this._uiRestoreInputs();

        if (e.error) {
            this.fire("failure", {data: e.error});
        } else {
            this.fire("success", {data: e.result});
        }
    },

    /**
     * Default event handler for the 'failure' event. Render form errors.
     */
    _defFailureFn: function(e) {
        renderFormErrors(e.data, this.get("host"));
    },

    /**
     * Disable all inputs which are not disabled already and reset errors.
     */
    _uiResetInputs: function(form) {
        var inputs = form.all("input"),
            errors = form.all(".form-error");

        this._disabledInputs = inputs.filter(function(input) {
            if (!input.getAttribute("disabled")) {
                input.setAttribute("disabled", "");
                return true;
            }
        });

        errors.remove();
    },

    /**
     * Re-enable all inputs that were disabled by _resetInputs.
     */
    _uiRestoreInputs: function() {
        this._disabledInputs.each(function(input) {
            input.removeAttribute("disabled");
        });
    },

}, {NS: "job",
    EVENT_PREFIX: "job",
    ATTRS: {

    /**
     * @attribute action
     * @description The name of the job to submit. If not given it defaults to
     *    the last URL segment of the 'action' attribute of the host form node.
     * @type String
     *
     */
    action: {
        valueFn: function() {
            var host = this.get("host"),
                action = host.get("action");
            return action.split("/").slice(-1)[0];
        },
    },

    /**
     * @attribute io
     * @description The Y.IO instance used to post the form, for testing.
     * @type Y.IO
     *
     */
    io: {
        value: Y.io,
        cloneDefaultValue: false
    },
}});

Y.landscape.job.FormJob = FormJob;

}, "0.1", {"requires": ["node",
                        "base",
                        "plugin",
                        "io-form",
                        "datasource",
                        "landscape.notes"]});
