// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.legacy.table", function(Y) {

Y.namespace("landscape.legacy");

// A TableListing represents a list of items with selectable rows.
// Selectable rows are TR elements that contain a checkbox; the value
// of the checkbox is used as the ID to get TableRow instances.  The
// selection of a row is toggled when either the TR or the checkbox is
// clicked.

TableListing = function(form_id) {
    this._form_id = form_id;
    this._initialized = false;
    this._rows = {};
};

TableListing.prototype._initialize = function() {
    if (this._initialized) {
        return;
    }

    var form = document.getElementById(this._form_id);
    for (i = 0; i < form.elements.length; ++i) {
        var node = form.elements[i];
        if (node.type == "checkbox" && node.className != "table-unmanaged") {
            var row = this._find_parent(node, "TR");
            if (typeof(row) != "undefined") {
                this._rows[node.value] =
                    new Y.landscape.legacy.TableRow(node.value, row, node);
                // reset the selected state if row already checked
                // Specifically for the fix in bug #75558, if the back
                // button is pressed to go back to the selection
                // screen the checkboxes remain checked but the rows
                // are not highlighted some older Firefox's.
                // get_selection returns the state of the checkbox
                // only, which is then passed into set_selection which
                // then clears the selection and correctly sets the
                // style to highlight the row.
                var state = node.checked;
                this._rows[node.value].set_selection(state);
            }
        }
    }
    this._initialized = true;
};

TableListing.prototype._find_parent = function(node, tag) {
    if (typeof(node) == "undefined") {
        return null;
    } else if (node.parentNode.tagName == tag) {
        return node.parentNode;
    }
    return this._find_parent(node.parentNode, tag);
};

TableListing.prototype.get_rows = function() {
    this._initialize();
    return this._rows;
};

TableListing.prototype.get_row_by_id = function(id) {
    this._initialize();
    return this._rows[id];
};

TableListing.prototype.toggle_selection = function(id) {
    this._initialize();
    row = this.get_row_by_id(id);
    return row.toggle_selection();
};

TableListing.prototype.get_row_count = function() {
    this._initialize();
    var count = 0;
    for (var id in this._rows) {
        count += 1;
    }
    return count;
};

TableListing.prototype.get_selected_row_count = function() {
    this._initialize();
    var count = 0;
    var rows = this._rows;
    for (var id in rows) {
        if (rows.hasOwnProperty(id)) {
            var row = rows[id];
            if (row.get_selection()) {
                count += 1;
            }
        }
    }
    return count;
};

TableListing.prototype.select_all = function() {
    this._set_selection(true);
};

TableListing.prototype.select_none = function() {
    this._set_selection(false);
};

TableListing.prototype._set_selection = function(condition) {
    this._initialize();
    var rows = this._rows;
    for (var id in rows) {
        if (rows.hasOwnProperty(id)) {
            var row = rows[id];
            row.set_selection(condition);
        }
    }
};


Y.landscape.legacy.TableListing = TableListing;


// A TableRow represents a TR and the INPUT[type="checkbox"] it
// contains.  It exposes methods to manipulate the selection state of
// both the row and the checkbox.

TableRow = function(id, row, checkbox) {
    this._id = id;
    this._row = row;
    this._checkbox = checkbox;
};

TableRow.prototype.toggle_selection = function() {
    var condition = ! this.get_selection();
    this.set_selection(condition);
    return condition;
};

TableRow.prototype.get_selection = function() {
    return this._row.className.match("selected");
};

TableRow.prototype.set_selection = function(condition) {
    this._checkbox.checked = condition;
    this.clear_selection();
    if (condition) {
        this._row.className += " selected";
    }
};


TableRow.prototype.clear_selection = function() {
    this._row.className = this._row.className.replace(/\bselected\b/, "");
};

TableRow.prototype.get_cell_by_class = function(cell_class) {
    return this._row.getElementsByClassName(cell_class)[0];
};

Y.landscape.legacy.TableRow = TableRow;


}, "0.1", {"requires": ["node"]});
