// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.legacy.tree", function(Y) {

Y.namespace("landscape.legacy");


// A TreeListing is a specialized form of the standard table listing
// in that it contains parent rows with child rows beneath them.  Each
// parent row represents the all the child rows beneath it.
//
// TreeListing implements the specialized selection logic that makes
// clicking on an aggregate row select (or unselect) all the child
// rows; makes unselecting a child row, when the aggregate row is
// selected, unselect the aggregate row; and selects the aggregate row
// when all child rows become selected.

TreeListing = function(form_id) {
    this._table = new Y.landscape.legacy.TableListing(form_id);
    this._rows = {};
};

TreeListing.prototype.toggle_selection = function(id) {
    var row = this._rows[id];
    if (! row) {
        return this._table.toggle_selection(id);
    }
    return row.toggle_selection();
};

TreeListing.prototype.get_selected_row_count = function() {
    this._table._initialize();
    var count = 0;
    var rows = this._table._rows;
    for (var id in rows) {
        if (rows.hasOwnProperty(id)) {
            var row = rows[id];
            if (row.get_selection()) {
                if (row._row.className.search(/\baggregate-parent-row\b/) == -1) {
                    count += 1;
                }
            }
        }
    }
    return count;
};


TreeListing.prototype.register_branch_row = function(id) {
    var row = this._table.get_row_by_id(id);
    var branch_row = new Y.landscape.legacy.TreeBranchRow(row);
    this._rows[id] = branch_row;
};

TreeListing.prototype.register_child_row = function(id, aggregate_id) {
    var branch_row = this._rows[aggregate_id];
    var row = this._table.get_row_by_id(id);
    var child_row = new Y.landscape.legacy.TreeLeafRow(row, branch_row);
    this._rows[id] = child_row;
    branch_row.add_child(child_row);
};


Y.landscape.legacy.TreeListing = TreeListing;


// A TreeBranchRow is a row that represents multiple child rows.  When
// it's state changes it manipulates the state of it's child rows.  It
// also responds to change notifications received from child rows.

TreeBranchRow = function(row) {
    this._row = row;
    this._child_rows = [];
};

TreeBranchRow.prototype.add_child = function(row) {
    this._child_rows.push(row);
};

TreeBranchRow.prototype.toggle_selection = function() {
    var condition = this._row.toggle_selection();
    for (var i = 0; i < this._child_rows.length; ++i) {
        var child_row = this._child_rows[i];
        child_row.set_selection(condition);
    }
};

TreeBranchRow.prototype.update = function() {
    var condition = true;
    for (var i = 0; i < this._child_rows.length; ++i) {
        var child_row = this._child_rows[i];
        condition = child_row.get_selection();
        if (! condition) {
            break;
        }
    }
    this._row.set_selection(condition);
};


Y.landscape.legacy.TreeBranchRow = TreeBranchRow;


// A TreeLeafRow represents a child row of an TreeBranchRow.  When
// it's selection status changes a notification is raised so that the
// parent TreeBranchRow can update it's state.

TreeLeafRow = function(row, branch_row) {
    this._row = row;
    this._branch_row = branch_row;
};

TreeLeafRow.prototype.toggle_selection = function() {
    this._row.toggle_selection();
    this._branch_row.update();
};

TreeLeafRow.prototype.get_selection = function() {
    return this._row.get_selection();
};

TreeLeafRow.prototype.set_selection = function(condition) {
    this._row.set_selection(condition);
};


Y.landscape.legacy.TreeLeafRow = TreeLeafRow;


}, "0.1", {"requires": ["node", "landscape.legacy.table"]});
