// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.listing.activities", function(Y) {

Y.namespace("landscape.listing");

var staticFilename = Y.landscape.ui.image.staticFilename;

/**
 * Hides or shows a information about activities.
 *
 * @module landscape.listing.activities
 */


/**
 * This class provides the ability turn a basic activity listing into
 * a dynamic one where details are expanded as the user toggles them.
 *
 * @class ActivitiesListing
 * @extends ExpandableListing
 * @constructor
 */
var ACTIVITIES_LISTING = "activities_listing",
    BINDUI = "bindUI",

    // Events
    EXPAND_ROW = "expandable_listing:expand_row",
    COLLAPSE_ROW = "expandable_listing:collapse_row";

var ActivitiesListing = function(config) {
    ActivitiesListing.superclass.constructor.apply(this, arguments);
};

ActivitiesListing.NAME = ACTIVITIES_LISTING;

ActivitiesListing.ATTRS = {
    /**
     * The DataSource class to be customized in tests.
     *
     * @attribute expandSource
     */
    expandSource: {
        value: null
    }
};

Y.extend(ActivitiesListing, Y.landscape.listing.ExpandableListing, {
    /**
     * Constructor logic.
     *
     * @method initializer
     * @protected
     */
    initializer: function(config) {

        Y.after(this._bindUIActivitiesListing, this, BINDUI);
    },

    /**
     * Bind the widget's event handlers.
     *
     * @method _bindUIActivitiesListing
     * @protected
     */
    _bindUIActivitiesListing: function() {
        this.on(EXPAND_ROW, this._expandRow);
        this.on(COLLAPSE_ROW, this._collapseRow);
    },

    /**
     * Expand the details for the current row, by performing a request
     * to the backend and rendering a table with the returned
     * packages.
     *
     * @method _expandRow
     * @protected
     */
    _expandRow: function(event) {
        var row = event.details[0].row;
        row.addClass("landscape-expandable-expanded");
        if (row.hasClass("landscape-expandable-has-content")) {
            row.next().setStyle("display", "table-row");
            return;
        }
        var ancestor = row.ancestor();
        var tr = Y.Node.create("<tr class='landscape-expandable-content'>");
        var contentCell, loadingIcon;
        var loadingIconPath = staticFilename("/static/skin/manx/loading-16x16.gif");
        for (var i=0; i < 6; i++) {
            var td = Y.Node.create("<td>");
            if (i == 4 || i == 5) {
                td.addClass("inner-cell");
            }
            if (i == 4) {
                td.setAttribute("colspan", "2");
                loadingIcon = Y.Node.create("<img src= '" + loadingIconPath + "'/>");
                td.appendChild(loadingIcon);
                contentCell = td;
            }
            tr.appendChild(td);
        }
        ancestor.insertBefore(tr, row.next());
        var link = row.one("a.activity-link");
        var url = link.get("href") + "/expand";
        var source = this.get("expandSource");
        if (!source) {
            source = new Y.DataSource.IO({source: url});
        }
        source.sendRequest(
            {request: null,
             callback:  {
                 success: Y.bind(this._renderPackageChanges, this,
                                 contentCell, loadingIcon)
             }
            });
    },

    /**
     * Hide the details for the current row.
     *
     * @method _collapseRow
     * @protected
     */
    _collapseRow: function(event) {
        var row = event.details[0].row;
        row.replaceClass("landscape-expandable-expanded",
                         "landscape-expandable-has-content");
        row.next().setStyle("display", "none");
    },

    /**
     * Create header text for special changes.
     *
     * @method _createHeader
     * @protected
     */
    _createHeader: function(msg) {
        var header = Y.Node.create("<tr>");
        var headerText = Y.Node.create("<td>");
        headerText.setStyle("border", 0);
        headerText.setAttribute("colspan", 2);
        headerText.set("text", msg);
        headerText.setStyle("fontWeight", "bold");
        header.appendChild(headerText);
        return header;
    },

    /**
     * Render fetched package information into a table, replacing the
     * loading icon.
     *
     * @method _renderPackageChanges
     * @protected
     */
    _renderPackageChanges: function(contentCell, loadingIcon, result) {
        var fragment = Y.config.doc.createDocumentFragment();
        var contentTable = Y.Node.create("<table>");
        fragment.appendChild(Y.Node.getDOMNode(contentTable));
        var contentBody = Y.Node.create("<tbody>");
        contentTable.appendChild(contentBody);
        var changes = Y.JSON.parse(result.data.responseText);
        var isComplemented = false;
        var isOriginal = false;
        Y.Array.each(changes, function(change) {
            if (!isComplemented && change.complemented) {
                isComplemented = true;
                contentBody.appendChild(
                    this._createHeader("Additional required changes"));
            }
            if (!isOriginal && !change.complemented && isComplemented) {
                isOriginal = true;
                contentBody.appendChild(
                    this._createHeader("User requested changes"));
            }
            var tr = Y.Node.create("<tr/>");
            contentBody.appendChild(tr);
            var tdContent = Y.Node.create("<td/>");
            tdContent.setStyle("border", 0);
            tdContent.setStyle("width", "100%");
            var txt, word;

            if (change.packageType === "snap") {
                txt = change.type + " snap " + change["package"].name;
            } else if (change.type == "create-hold") {
                txt = "Hold package '" + change["package"] + " " +
                    change.version + "'";
            } else if (change.type == "remove-hold") {
                txt = "Remove hold on package '" + change["package"] + " " +
                    change.version + "'";
            } else {
                word = change.type.charAt(0).toUpperCase() +
                       change.type.substr(1);
                txt = word + " package " + change["package"] + " ";
                if (change.type == "install" || change.type == "remove") {
                    txt += change.version;
                } else {
                    txt += change.remove_version + " to " +
                           change.install_version;
                }
            }
            tdContent.set("text", txt);
            tr.appendChild(tdContent);
        }, this);
        loadingIcon.remove();
        contentCell.appendChild(fragment);
    }

});

Y.landscape.listing.ActivitiesListing = ActivitiesListing;


}, "0.1", {"requires": ["node", "event", "widget", "base", "json-parse",
                        "datasource-io", "landscape.listing.expandable",
                        "landscape.ui.sprite", "landscape.ui.image"]});
