// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.listing.expandable", function(Y) {

Y.namespace("landscape.listing");

/**
 * Hides or shows a collection of sibling rows.
 *
 * @module landscape.listing.expandable
 */


/**
 * This class provides the ability to turn the rows in a static HTML table
 * into interactive tree-like elements.
 *
 * @class ExpandableListing
 * @extends Widget
 * @constructor
 */
var EXPANDABLE_LISTING = "expandable_listing",

    // Configuration properties
    BOUNDING_BOX = "boundingBox",
    CONTENT_BOX = "contentBox",
    PARENT_ROWS = "parentRows",
    CHILD_ROWS = "childRows",
    BRANCH_TOGGLE_CONTROLS = "branchToggleControls",
    DEFAULT_EXPANDED = "defaultExpanded",
    EXPANDABLE_PARENT = "tbody tr:not(.landscape-expandable-child)",
    EXPANDABLE_CHILD = "tbody tr.landscape-expandable-child",
    EXPAND_CONTROL = ".landscape-expandable-tree-expand",

    // Events
    EXPAND_ROW = "expandable_listing:expand_row",
    COLLAPSE_ROW = "expandable_listing:collapse_row";

var ExpandableListing = function(config) {
    ExpandableListing.superclass.constructor.apply(this, arguments);
};

ExpandableListing.NAME = EXPANDABLE_LISTING;

/**
 * Static object hash used to capture existing markup for progressive
 * enhancement.
 *
 * @property ExpandableListing.HTML_PARSER
 * @type Object
 * @static
 */
ExpandableListing.HTML_PARSER = {
    tableRows: ["tbody tr"],
    parentRows: [EXPANDABLE_PARENT],
    childRows: [EXPANDABLE_CHILD],
    branchToggleControls: [EXPAND_CONTROL]
};

ExpandableListing.ATTRS = {
    /**
     * A list of nodes that should trigger expanding and collapsing
     * the children rows.
     *
     * @attribute branchToggleControls
     * @type Array
     */
    branchToggleControls: {
        value: null
    },

    /**
     * A list of rows that should be considered parent nodes.
     *
     * @attribute parentRows
     * @type Array
     */
    parentRows: {
        value: null
    },

    /**
     * A list of rows that should be considered child nodes.
     *
     * @attribute childRows
     * @type Array
     */
    childRows: {
        value: null
    },

    /**
     * A boolean value indicating if the branches should be expanded
     * by default.
     *
     * @attribute defaultExpanded
     * @type Array
     */
    defaultExpanded: {
        value: null
    },

    /**
     * The tab index is null to prevent focusing of the content box when the
     * widget is used.
     *
     * @attribute tabIndex
     * @type Int
     */
    tabIndex: {
        value: null
    }
};

Y.extend(ExpandableListing, Y.Widget, {
    /**
     * Constructor logic.
     *
     * @method initializer
     * @protected
     */
    initializer: function(config) {
        /**
         * Fires when the user toggles a collapsed branch.
         *
         * @event expand_row
         */
        this.publish(EXPAND_ROW);

        /**
         * Fires when the user toggles an expanded branch.
         *
         * @event collapse_row
         */
        this.publish(COLLAPSE_ROW);
    },

    /**
     * Bind the widget's DOM elements to their event handlers.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        var controls = this.get(BRANCH_TOGGLE_CONTROLS);
        if (controls) {
            controls.addClass("no-select");
        }

        this.get(CONTENT_BOX).delegate("click", function(event) {
            event.halt(true);
            var control = event.currentTarget;
            var row = control.ancestor(EXPANDABLE_PARENT);
            this.toggleRow(row);
            return false;
        }, EXPAND_CONTROL, this);
        
        var default_expanded = this.get(DEFAULT_EXPANDED);
        var parents = this.get(PARENT_ROWS);

        if (parents && default_expanded !== false) {
            parents.each(function(parent) {
                if (default_expanded === true ||
                    (default_expanded === null &&
                     parent.hasClass("landscape-expandable-expanded"))) {
                    this.expandRow(parent);
                } else {
                    this.collapseRow(parent);
                }
            }, this);
        }
    },

    /**
     * Toggle the specified row's marked state.
     *
     * @method toggleRow
     * @private
     */
    toggleRow: function(row) {
        var expand = ! row.hasClass("landscape-expandable-expanded");
        if (expand) {
            this.expandRow(row);
        } else {
            this.collapseRow(row);
        }
    },

    /**
     * Expand the specified row, if not yet expanded.
     *
     * @method expandRow
     * @private
     */
    expandRow: function(row) {
        this._setRowState(row, true);
    },

    /**
     * Collapse the specified row, if not yet collapsed.
     *
     * @method collapseRow
     * @private
     */
    collapseRow: function(row) {
        this._setRowState(row, false);
    },

    /**
     * Set the expanded state of the specified row.
     *
     * @method _setRowState
     * @private
     */
    _setRowState: function(row, expanded) {
        var parents = this.get(PARENT_ROWS);

        if (parents && parents.indexOf(row) < 0){
            return;
        }

        var is_expanded = row.hasClass("landscape-expandable-expanded");
        if (!is_expanded && expanded) {
            row.addClass("landscape-expandable-expanded");
        } else {
            if (is_expanded && !expanded) {
                row.removeClass("landscape-expandable-expanded");
            }
        }

        var children = this.get(CHILD_ROWS);
        if (children) {
            var child = row;
            while ((child = child.next("tr"))) {
                if (!child || children.indexOf(child) < 0) {
                    break;
                }
                if (expanded) {
                    child.addClass("landscape-expandable-expanded");
                    child.setStyle("display", "table-row");
                } else {
                    child.removeClass("landscape-expandable-expanded");
                    child.setStyle("display", "none");
                }
            }
        }

        if (!is_expanded && expanded) {
            this.syncUI();
            this.fire(EXPAND_ROW, {row: row});
        } else if (is_expanded && !expanded) {
                this.syncUI();
                this.fire(COLLAPSE_ROW, {row: row});
        }
    }

});

Y.landscape.listing.ExpandableListing = ExpandableListing;


}, "0.1", {"requires": ["node", "event", "widget", "base"]});
