// Copyright 2005-2013 Canonical Limited.  All rights reserved.

/*

Todo List
---------
 - controls should be disabled if they're not valid
 - autoPageSize option to automatically size the number of rows
   to the available space, (based on dimensions of template row).
 - events should fire off the widget, as private events not the
   global bus.
*/




YUI.add("landscape.listing.pager", function(Y) {

Y.namespace("landscape.listing");

var staticFilename = Y.landscape.ui.image.staticFilename;

/**
 * Present a list of items as a series of pages.
 *
 * @module landscape.listing.pager
 */

var DEFAULT_PAGE_SIZE = 25,
    EVT_BEFORE_PAGE_CHANGE = "before-page-change",
    EVT_BEFORE_PAGE_RENDER = "before-page-render",
    EVT_AFTER_PAGE_CHANGE = "after-page-change",
    PAGINATION_DISABLED = "pagination-disabled",
    PAGINATION_HIDDEN = "pagination-hidden",
    INVALID_DATASOURCE = "Invalid DataSource",
    INVALID_PAGE_SIZE = "Invalid Page Size",
    INVALID_PAGE_NUMBER = "Invalid Page Number",
    PREVIOUS_NODE = "previousNode",
    NEXT_NODE = "nextNode",
    FIRST_NODE = "firstNode",
    LAST_NODE = "lastNode",
    FIRST_ITEM_NODE = "firstItemNode",
    LAST_ITEM_NODE = "lastItemNode",
    NEXT_TEXT = "nextText",
    PREVIOUS_TEXT = "previousText",
    FIRST_TEXT = "firstText",
    LAST_TEXT = "lastText",
    TOTAL_RECORDS_NODE = "totalRecordsNode",
    HOST = "host",
    TEXT = "text",
    CLICK = "click",
    TOTAL_PAGES = "totalPages",
    RECORDS = "records",
    METADATA = "metadata",
    TOTAL_RECORDS = "totalRecords",
    PAGE_SIZE = "pageSize",
    CURRENT_PAGE = "currentPage",
    RENDER = "render",
    ROW_RENDERER = "rowRenderer",
    STATUS_RENDERER = "statusRenderer",
    CHILDREN = "children",
    CONTENT_BOX = "contentBox",
    PAGINATION_BOX = "paginationBox",
    TEMPLATE_ROW = "templateRow",
    REQUEST_CONFIG = "requestConfig";


/**
 *
 * @class Pager
 * @extends Widget
 * @constructor
 */
var Pager = function() {
    Pager.superclass.constructor.apply(this, arguments);
};

Pager.NAME = "pager";

Pager.ATTRS = {
    /**
     * An optional custom function used to update a row node with data.
     * Intended for greater flexibility for accomodating designs when
     * rendering rows. The function is called with the row YUI node,
     * the row data record, and the the row index.
     *
     * @attribute rowRenderer
     * @type Function
     */
    rowRenderer: {value: null},

    /**
     * An optional custom fuction used to update the status message/pager
     * element. The function is called with the widget as its only
     * argument.
     *
     * @attribute statusRenderer
     * @type Function
     */
    statusRenderer: {value: null},

    /**
     * The row to use as a template, when creating new rows.
     * used by the default renderer. Custom renderers may or
     * may not use this. A datasource's record is matched
     * to the columns in a row by name, where the appropriate
     * name is attached as a css class to the column.
     *
     * @attribute templateRow
     * @type Node
     */
    templateRow: {value: null},

    /**
     * A data source to get records from. The recordSource must have a
     * a data schema plugin attached, as a "totalRecords" meta field is
     * *required*.
     *
     * Any additional meta fields are stored on the widget as the
     * "metadata" object attribute. This forms the primary mechanism
     * to flow additional backend data through the widget to its subscribers
     * plugins or custom functions.
     *
     *
     * @attribute recordSource
     * @type DataSource.IO
     */
    recordSource: {value: null,
                   setter: function(value) {
                        // default behavior for yui validators and setters
                        // returning INVALID_ATTRIBUTE is to silently
                        // fail. In the words of Tim Peters,
                        // "Errors should never pass silently.
                        if (!value instanceof Y.DataSource.Local) {
                            throw new Error(INVALID_DATASOURCE);
                        }
                        if (value.schema === undefined) {
                            throw new Error(INVALID_DATASOURCE);
                        }
                        return value;
                    }},

    /**
     * Extra arguments to include when getting data from recordSource.
     *
     * @attribute recordSourceArguments
     * @type Object
     */
    recordSourceArguments: {value: null},

    /**
     * Should the widget attempt to render the first page. In a progressive
     * enhancement scenario, we can set this to false and render the first page
     * directly in html, letting the widget render additional pages.
     *
     * @attribute renderFirstPage
     * @type boolean
     */
    renderFirstPage: {value: true},

    /**
     * All meta values from backend response are stored on this attribute
     *
     * @attribute metadata
     * @type Object
     */
    metadata: {value: null},

    /**
     * Parameter name to append to the url to specify how the sort value
     * will be passed to the DataSource.
     *
     * @attribute sourceSortKeyParameter
     * @type String
     */
    sortKeyParameter: {value: null},

    /**
     * Attribute to hold the current sort key, typically a column name.
     *
     * @attribute sortKey
     * @type String
     */
    sortKey: {value: null},

    /**
     * Parameter name to append to the url when specifing the sort order.
     *
     * @attribute sourceSortDirectionParameter
     * @type String
     */
    sortDirectionParameter: {value: null},

    /**
     * Attribute to hold the current sort direction parameter, typically
     * either ascending or descending values recognized by the backend.
     *
     * @attribute sortValue
     * @type String
     */
    sortDirection: {value: null},

    /**
     * Parameter Name to append to the url to indicate start position
     * in the total record set. The widget automatically sets the value
     * to the currentPage*pageSize.
     *
     * @attribute offsetParameter
     * @type String
     */
    offsetParameter: {value: "offset"},

    /**
     * Parameter Name to append to the url to indicate number of
     * records to return. The widget automatically sets the value
     * to the current pageSize.
     *
     * @attribute pageSizeParameter
     * @type String
     */
    pageSizeParameter: {value: "size"},

    /**
     * An array of objects representing rows for the current page.
     *
     * @attribute records
     * @type Array
     */
    records: {value: null},

    /**
     * The total number of records across all pages. If -1, then the number of
     * total records is unknown, and we will show "many" instead.
     *
     * @attribute totalRecords
     * @type Int
     */
    totalRecords: {value: 0},

    /**
     * The current page.  Default is 1, the first page.
     *
     * @attribute currentPage
     * @type Int
     */
    currentPage: {value: 1},

    /**
     * The maximum number of rows to show at a time.  Defaults to
     * DEFAULT_PAGE_SIZE.
     *
     * @attribute pageSize
     * @type Int
     */
    pageSize: {value: DEFAULT_PAGE_SIZE},

    /**
     * The total number of pages. If unknown, -1 is returned.
     *
     * @attribute totalPages
     * @type Int
     */
    totalPages: {
        getter: function() {
            var totalRecords = this.get(TOTAL_RECORDS);
            if (totalRecords < 0) {
                return -1;
            }
            if (totalRecords) {
                return Math.ceil(totalRecords / this.get(PAGE_SIZE));
            }
            return 0;
        }
    },

    /**
     * Optionally, an element displaying a count of items in the list.
     *
     * @attribute statusMessage
     * @type Node
     */
    statusMessage: {value: null},

    /**
     * Optionally, the control to use as the 'Next' button.
     *
     * @attribute nextControl
     * @type Node
     */
    nextControl: {value: null},

    /**
     * Optionally, the control to use as the 'Previous' button.
     *
     * @attribute previousControl
     * @type Node
     */
    previousControl: {value: null},

    /**
     * The tab index is null by default, to prevent focusing of the content box
     * when the widget is used.
     *
     * @attribute tabIndex
     * @type Int
     */
    tabIndex: {value: null},

    /**
     * Function used to calculate the offset for a page. By default, it simply
     * uses the page size and the number of the page, but it can be customized.
     *
     * @attribute offsetCalculation
     * @type Function
     */
    offsetCalculation: {
        value: function(currentPage, pageSize) {
            return (currentPage - 1) * pageSize;
        }
    },

    /**
     * An object to provide custom configuration to the request to fetch rows
     * data.
     */
    requestConfig: {value: {}}
};

Y.extend(Pager, Y.Widget, {
    /**
     * Initialization Logic. Publishes custom events.
     *
     * @method initializer
     * @protected
     */
    initializer: function() {
      /**
       * Notification event before the widget requests data for a page change.
       * A typical use case is subscribers cleaning up event handlers for
       * rows before the rows are rerendered with new data, or setting a
       * visual loading indicator as new rows are fetched.
       *
       * @event pager:before-page-change
       * @param {EventFacade} e The Event Facade
       */
      this.publish(EVT_BEFORE_PAGE_CHANGE);

      /**
       * Notification event before the widget renders the rows for new
       * page data. A typical use case is using the backend metadata
       * for bookkeeping or prerendering to a given page.
       *
       * @event pager:before-page-render
       * @param {EventFacade} e The Event Facade
       */
      this.publish(EVT_BEFORE_PAGE_RENDER);

      /**
       * Notification event after the page has been rendered with new data.
       * Typically used for attaching behavior via event handlers en masse to
       * the rendereded rows.
       *
       * @event pager:after-page-change
       * @param {EventFacade} e The Event Facade
       */
      this.publish(EVT_AFTER_PAGE_CHANGE,
                   {defaultFn: this.syncUI});
    },

    /**
     * Render the widget.
     *
     * @method renderUI
     * @protected
     */
    renderUI: function() {
        var renderFirstPage = this.get("renderFirstPage");
        var templateRow = this.get(CONTENT_BOX).one("tbody tr");
        this.set(TEMPLATE_ROW, templateRow);
        if (renderFirstPage) {
            templateRow.remove();
            this._fetchPageRows();
        }
    },

    /**
     * Bind event handlers to hook the widget into the DOM.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        var nextControl = this.get("nextControl");
        if (nextControl) {
            nextControl.on(CLICK, function(evt) {
                this.nextPage();
            }, this);
        }

        var previousControl = this.get("previousControl");
        if (previousControl) {
            previousControl.on(CLICK, function(evt) {
                this.previousPage();
            }, this);
        }
    },

    /**
     * Synchronize the widget and the user interface.
     *
     * @method syncUI
     * @protected
     */
    syncUI: function() {
        this._updateStatusMessage();
    },


    /**
     * Advance to the next page.  If this is the last page, this method has no
     * effect.
     *
     * @method nextPage
     * @public
     */
    nextPage: function() {
        var page = this.get(CURRENT_PAGE);
        var totalPages = this.get(TOTAL_PAGES);
        if (totalPages == -1 || page + 1 <= totalPages) {
            this._switchPage(page + 1);
        }
    },

    /**
     * Advance to the previous page.  If this is the first page, this method
     * has no effect.
     *
     * @method previousPage
     * @public
     */
    previousPage: function() {
        var page = this.get(CURRENT_PAGE);
        if (page - 1 >= 1) {
            this._switchPage(page - 1);
        }
    },

    /**
     * Method to switch the current page. Upon changing the current page
     * the UI is automatically re-rendered.
     *
     * @method jumpToPage
     * @public
     */
    jumpToPage: function(pageNumber) {
        var totalPages = this.get(TOTAL_PAGES);
        pageNumber = parseInt(pageNumber, 10);
        if (isNaN(pageNumber)) {
            throw new Error(INVALID_PAGE_NUMBER);
        }
        if ((totalPages != -1 && pageNumber > totalPages) || pageNumber < 1) {
            throw new Error(INVALID_PAGE_NUMBER);
        }
        this._switchPage(pageNumber);
    },

    /**
     * Method to set the page size. Upon changing the page size,
     * the UI is automatically reflowed. We attempt to preserve
     * the contents of the current page (based on the starting offset).
     *
     * @method setPageSize
     * @public
     */
    setPageSize: function(pageSize) {
        pageSize = parseInt(pageSize, 10);
        if (isNaN(pageSize)) {
            throw new Error(INVALID_PAGE_SIZE);
        }
        var currentPage = this.get(CURRENT_PAGE);
        var currentOffset = currentPage * this.get(PAGE_SIZE);
        var nextPage = parseInt(currentOffset / pageSize, 10) + 1;
        this.set(PAGE_SIZE, pageSize);
        this.jumpToPage(nextPage);
    },

    /**
     * Method to specify value for server side sorting of the data source.
     * Calling this method will automatically change the page back to 1.
     *
     *
     * @method setSort
     * @public
     */
    setSort: function(name, direction) {
        if (!this.get("sortKeyParameter")) {
            throw "Sort Key Parameter must be specified.";
        }
        if (direction && !this.get("sortDirectionParameter")) {
            throw "Sort Direction Parameter must be specified.";
        }
        this.set("sortKey", name);
        this.set("sortDirection", direction);
        this._switchPage(1);
    },

    /**
     * Get the number of rows currently rendered.
     */
    getRowCount: function() {
        var tbody = this.get(CONTENT_BOX).one("tbody");
        var rows = tbody.all("tr");
        return rows.size();
    },

    // Returns an object mapping stock query string arguments to values.
    _getBuiltinRecordSourceArguments: function() {
        var args = {};

        var pageSize = this.get(PAGE_SIZE);
        var offsetParameter = this.get("offsetParameter");
        var offsetCalculation = this.get("offsetCalculation");
        var offset = offsetCalculation.apply(this, [this.get(CURRENT_PAGE), pageSize]);
        args[offsetParameter] = offset;

        var pageSizeParameter = this.get("pageSizeParameter");
        args[pageSizeParameter] = pageSize;

        var sortKeyParameter = this.get("sortKeyParameter");
        var sortKey = this.get("sortKey");
        args[sortKeyParameter] = sortKey;

        var sortDirectionParameter = this.get("sortDirectionParameter");
        var sortDirection = this.get("sortDirection");
        args[sortDirectionParameter] = sortDirection;

        return args;
    },

    _buildQueryString: function(args) {
        var queryString = "";
        var firstArgument = true;
        for (var parameter in args) {
            if (args.hasOwnProperty(parameter)) {
                var value = args[parameter];
                if (value) {
                    if (Y.Lang.isArray(value)) {
                        for (var index in value) {
                            if (!firstArgument) {
                                queryString += "&";
                            }
                            queryString += parameter + "=" + encodeURIComponent(value[index]);
                            firstArgument = false;
                        }
                    } else {
                        if (!firstArgument) {
                            queryString += "&";
                        }
                        queryString += parameter + "=" + encodeURIComponent(value);
                    }
                    firstArgument = false;
                }
            }
        }
        return queryString;
    },

    getRequestQueryString: function() {
        var args = Y.merge(this._getBuiltinRecordSourceArguments(),
                           this.get("recordSourceArguments"));
        return this._buildQueryString(args);
    },

    _switchPage: function(page) {
        this.fire(EVT_BEFORE_PAGE_CHANGE, {page: page});
        this.set(CURRENT_PAGE, page);
        this._fetchPageRows();
    },

    _updateStatusMessage: function() {
        // support custom status renderers
        if (this.get(STATUS_RENDERER)) {
            return this.get(STATUS_RENDERER)(this);
        }
        var statusMessage = this.get("statusMessage");
        if (statusMessage) {
            var items = this.get(RECORDS);
            if (items === null) {
                statusMessage.set(TEXT, "Loading...");
            } else {
                var message = "Showing " + this.get(CURRENT_PAGE) + " of " +
                              this.get(TOTAL_PAGES);
                statusMessage.set(TEXT, message);
            }
        }
        return null; // jslint
    },

    /**
     * Create/Remove Table Rows to match the number of rows on the
     * Current Page.
     *
     * @method _renderPageRows
     * @private
     */
    _renderPageRows: function() {
        var i;
        var records = this.get(RECORDS);
        var pageSize = this.get(PAGE_SIZE);
        var tbody = this.get(CONTENT_BOX).one("tbody");

        var rows = tbody.all("tr");
        var existingRowCount = rows.size();
        // how many rows do we need to display the current batch of records
        var neededRowCount = (records.length > pageSize) ? pageSize : records.length;

        // If the current page has too many rows remove the excess
        if (existingRowCount > neededRowCount) {
            for (i = neededRowCount; i < existingRowCount; i++) {
                rows.item(i).remove();
            }
            return;
        }

        // create additional rows as needed
        var templateRow = this.get(TEMPLATE_ROW);

        for (i = existingRowCount; i < neededRowCount; ++i) {
            var row = templateRow.cloneNode(true);
            tbody.appendChild(row);
        }

    },

    _updatePageRows: function() {
        var row, record;
        var records = this.get(RECORDS);
        var rows = this.get(CONTENT_BOX).all("tbody tr");
        var rowCount = rows.size();

        var renderer = this.get(ROW_RENDERER) ?
            this.get(ROW_RENDERER) : this._updateRow;
        for (var i = 0; i < rowCount; i++) {
            row = rows.item(i);
            record = records[i];
            renderer(row, record, i);
        }
    },

    _updateRow: function(row, record, row_index) {
        for (var key in record) {
            if (record.hasOwnProperty(key)) {
                // XXX this might be a bit excessive O(row*col),
                // basically does a query per cell.
                var node = row.one("." + key);
                if (node) {
                    node.set(TEXT, record[key]);
                }
            }
        }
    },

    _fetchPageRows: function() {
        var queryString = this.getRequestQueryString();
        var requestConfig = this.get(REQUEST_CONFIG);

        var config = {request: "?" + queryString,
                      cfg: requestConfig,
                      callback: {
                          success: Y.bind(this._dataRetrievalSuccess, this),
                          failure: Y.bind(this._dataRetrievalFailed, this)
                      }};

        if (! Y.Lang.isUndefined(requestConfig.method) &&
            requestConfig.method.toLowerCase() == "post") {
            config.request = queryString;
        }

        this.get("recordSource").sendRequest(config);
    },

    _dataRetrievalFailed: function(result, errorMessage) {
        var notes = new Y.landscape.notes.Notes({contentBox: "#notes"});
        errorMessage = errorMessage || result.error;
        if (errorMessage === undefined) {
            errorMessage = "List items could not be retrieved.";
        }
        notes.addNote("error", errorMessage);
    },

    _dataRetrievalSuccess: function(result) {
        var records = result.response.results;
        var total_records = result.response.meta.totalResults;

        this.set(TOTAL_RECORDS, total_records);
        this.set(RECORDS, records);
        this.set(METADATA, result.response.meta);

        // fire event after we have data, before we modify rows
        this.fire(EVT_BEFORE_PAGE_RENDER);

        // ensure proper number of rows for this page
        this._renderPageRows();

        // update rows on page
        this._updatePageRows();

        // fire event after render (default handler updates status ui)
        this.fire(EVT_AFTER_PAGE_CHANGE);

        return null; // jslint
    }

});

Y.landscape.listing.Pager = Pager;

var StaticPager = function() {
    StaticPager.superclass.constructor.apply(this, arguments);
};

StaticPager.NAME = "static_pager";

StaticPager.ATTRS = {
    /**
     * Static list containing all the element IDs to paginate.
     */
    availableIds: {
        value: []
    },

    /**
     * Record source retrieving available IDs when sorting changes.
     */
    availableIdsSource: {
        value: null
    }
};

Y.extend(StaticPager, Pager, {

    /**
     * When sorting happens, we retrieve the list of available IDs again.
     *
     * @method setSort
     * @public
     */
    setSort: function(name, direction) {
        if (!this.get("sortKeyParameter")) {
            throw "Sort Key Parameter must be specified.";
        }
        if (direction && !this.get("sortDirectionParameter")) {
            throw "Sort Direction Parameter must be specified.";
        }
        this.set("sortKey", name);
        this.set("sortDirection", direction);
        this._updateAvailableIds();
    },

    /**
     * Override the Pager default arguments building to explicitly pass the IDs
     * we're interesting in.
     *
     * @method _getBuiltinRecordSourceArguments
     * @private
     */
    _getBuiltinRecordSourceArguments: function() {
        var args = {};
        var availableIds = this.get("availableIds");
        var offsetCalculation = this.get("offsetCalculation");
        var pageSize = this.get(PAGE_SIZE);
        var offset = offsetCalculation.apply(
            this, [this.get(CURRENT_PAGE), pageSize]);
        var ids = availableIds.slice(offset, offset + pageSize);

        args.ids = ids.join(",");

        return args;
    },

    /**
     * Specific arguments passed to retrieve the available IDs when sorting
     * changes.
     *
     * @method _getAvailableRecordSourceArguments
     * @private
     */
    _getAvailableRecordSourceArguments: function() {
        var args = {};

        var sortKeyParameter = this.get("sortKeyParameter");
        var sortKey = this.get("sortKey");
        args[sortKeyParameter] = sortKey;

        var sortDirectionParameter = this.get("sortDirectionParameter");
        var sortDirection = this.get("sortDirection");
        args[sortDirectionParameter] = sortDirection;

        return args;
    },

    /**
     * Method that builds the query string to retrive the list of available
     * IDs.
     *
     * @method _getAvailableRequestQueryString
     * @private
     */
    _getAvailableRequestQueryString: function() {
        var args = Y.merge(this._getAvailableRecordSourceArguments(),
                           this.get("recordSourceArguments"));
        return this._buildQueryString(args);
    },

    /**
     * Query the available IDs source to get the sorted list of available
     * elements.
     *
     * @method _updateAvailableIds
     * @private
     */
    _updateAvailableIds: function() {
        var availableIdsSource = this.get("availableIdsSource");
        if (!Y.Lang.isUndefined(availableIdsSource)) {
            var queryString = this._getAvailableRequestQueryString();
            availableIdsSource.sendRequest({
                request: "?" + queryString,
                callback : {
                    success: Y.bind(function(result) {
                        var data = result.response.results[0].responseText;
                        this.set("availableIds", Y.JSON.parse(data));
                        this._switchPage(1);
                    }, this)
                }
            });
        }
    },

});

Y.landscape.listing.StaticPager = StaticPager;


/**
 * Plugin renders pagination and location status elements when the 'render'
 * event on the host widget fires.  The pagination and location status elements
 * are updated when the host's 'after-page-change' event fires.
 *
 * Use it with the Pager widget like so:
 *
 *   var widget = new Y.landscape.listing.Pager({...});
 *   widget.plug(Y.landscape.listing.PaginationPlugin,
 *               {paginationBox: Y.one("#pagination")});
 *
 * The pagination box is the element within which pagination controls are
 * rendered.  It should be a DIV or similar element, the contents of which
 * will be replaced by dynamically generated pagination controls.
 *
 * @class PaginationPlugin
 * @extends Plugin.Base
 * @constructor
 */
var PaginationPlugin = function() {
    PaginationPlugin.superclass.constructor.apply(this, arguments);
};

PaginationPlugin.NAME = "pagination";
PaginationPlugin.NS = "pagination";

PaginationPlugin.ATTRS = {
    /**
     * The element within which to render pagination controls.
     *
     * @attribute paginationBox
     * @type Node
     */
    paginationBox: {value: null},

    /**
     * The text to display as the next-page control, such as 'Older'.  Default
     * is 'Next'.
     *
     * @attribute nextText
     * @type String
     */
    nextText: {value: "Next"},

    /**
     * The text to display as the previous-page control, such as 'Newer'.
     * Default is 'Previous'.
     *
     * @attribute previousText
     * @type String
     */
    previousText: {value: "Previous"},

    /**
     * The text to display as the first-page control, such as 'Newest'.
     * Default is 'First'.
     *
     * @attribute firstText
     * @type String
     */
    firstText: {value: "First"},

    /**
     * The text to display as the last-page control, such as 'Oldest'.
     * Default is 'Last'.
     *
     * @attribute lastText
     * @type String
     */
    lastText: {value: "Last"},

    /**
     * An HTML element that represents the 'Next' control.
     *
     * @attribute nextNode
     * @type Node
     */
    nextNode: {value: null},

    /**
     * An HTML element that represents the 'Previous' control.
     *
     * @attribute previousNode
     * @type Node
     */
    previousNode: {value: null},

    /**
     * An HTML element that represents the 'First' control.
     *
     * @attribute firstNode
     * @type Node
     */
    firstNode: {value: null},

    /**
     * An HTML element that represents the 'Last' control.
     *
     * @attribute lastNode
     * @type Node
     */
    lastNode: {value: null},

    /**
     * An HTML element with the sequence number of the first record on the
     * current page as text content.
     *
     * @attribute firstItemNode
     * @type Node
     */
    firstItemNode: {value: null},

    /**
     * An HTML element with the sequence number of the last record on the
     * current page as text content.
     *
     * @attribute lastItemNode
     * @type Node
     */
    lastItemNode: {value: null},

    /**
     * An HTML element with the total number of records available in the
     * underlying data set as text content.
     *
     * @attribute totalRecordsNode
     * @type Node
     */
    totalRecordsNode: {value: null},

    /**
     * Disable the last element navigation, if it can't be used for the page.
     *
     * @attribute disableLast
     * @type Boolean
     */
    disableLast: {value: false},

    /**
     * Should the "Last" element navigation be hidden?
     *
     * @attribute hideLast
     * @type Boolean
     */
    hideLast: {value: false}
};


Y.extend(PaginationPlugin, Y.Plugin.Base, {
    /**
     * Constructor logic.
     *
     * @method initializer
     * @protected
     */
    initializer: function(config) {
        if (config.nextText) {
            this.set(NEXT_TEXT, config.nextText);
        }
        if (config.previousText) {
            this.set(PREVIOUS_TEXT, config.previousText);
        }

        var host = this.get(HOST);
        if (this.get(PAGINATION_BOX)) {
            host.on(RENDER, Y.bind(this._renderLoadingMessage, this));
            host.after(RENDER, Y.bind(this._renderPaginationControls, this));
        }
    },

    /**
     * Render a loading message in the pagination box.
     *
     * @method _renderLoadingMessage
     * @private
     */
    _renderLoadingMessage: function() {
        var loadingIconPath = staticFilename("/static/skin/manx/loading-16x16.gif");
        var controls = new Y.Node.create(["",
            "<span style='vertical-align: top;'>                    ",
            "  <img src='" + loadingIconPath + "' />   ",
            "  <span style='vertical-align: top;'>Loading...</span> ",
            "</span>                                                ",
            ""].join(""));
        var paginationBox = this.get(PAGINATION_BOX);
        paginationBox.get(CHILDREN).remove();
        paginationBox.appendChild(controls);
    },

    /**
     * Render pagination and location status controls, bind event handlers, and
     * synchronize the plugin with the host widget.
     *
     * @method _renderPaginationControls
     * @private
     */
    _renderPaginationControls: function() {
        this._renderUI();
        this._bindUI();
        this._syncUI();

        var host = this.get(HOST);
        host.after(EVT_AFTER_PAGE_CHANGE, Y.bind(this._syncUI, this));
    },

    /**
     * Render a status message with pagination controls in the pagination
     * box.
     *
     * @method _renderUI
     * @private
     */
    _renderUI: function() {
        var next = this.get(NEXT_TEXT);
        var previous = this.get(PREVIOUS_TEXT);
        var first = this.get(FIRST_TEXT);
        var last = this.get(LAST_TEXT);
        var controls = new Y.Node.create(["",
            "<span class='pagination-controls'>",
            "  <a href='#' class='pagination-first'>" + first + "</a>",
            "  <a href='#' class='pagination-previous'>" + previous + "</a>",
            "  <span class='pagination-label'>",
            "  <span class='pagination-first-item'></span> -",
            "  <span class='pagination-last-item'></span> ",
            "  <span class='pagination-of'>of</span>",
            "  <span class='pagination-total-records'></span>",
            "  </span>",
            "  <a href='#' class='pagination-next'>" + next + "</a>",
            "  <a href='#' class='pagination-last'>" + last + "</a>",
            "</span>",
            ""].join(""));
        var paginationBox = this.get(PAGINATION_BOX);
        paginationBox.get(CHILDREN).remove();
        paginationBox.appendChild(controls);

        this.set(NEXT_NODE,
                 paginationBox.one(".pagination-next"));
        this.set(PREVIOUS_NODE,
                 paginationBox.one(".pagination-previous"));
        this.set(FIRST_NODE,
                 paginationBox.one(".pagination-first"));
        this.set(LAST_NODE,
                 paginationBox.one(".pagination-last"));
        this.set(FIRST_ITEM_NODE,
                 paginationBox.one(".pagination-first-item"));
        this.set(LAST_ITEM_NODE,
                 paginationBox.one(".pagination-last-item"));
        this.set(TOTAL_RECORDS_NODE,
                 paginationBox.one(".pagination-total-records"));
    },

    /**
     * Bind the pagination and location status controls in the pagination box
     * to event handlers.
     *
     * @method _bindUI
     * @private
     */
    _bindUI: function() {
        this.get(NEXT_NODE).on(CLICK, Y.bind(this._nextPage, this));
        this.get(PREVIOUS_NODE).on(CLICK, Y.bind(this._previousPage, this));
        this.get(FIRST_NODE).on(CLICK, Y.bind(this._firstPage, this));
        this.get(LAST_NODE).on(CLICK, Y.bind(this._lastPage, this));
    },

    /**
     * Disable all the controls to stop navigation.
     *
     * @method _disablePagination
     * @private
     */
    _disablePagination: function() {
        this.get(PREVIOUS_NODE).addClass(PAGINATION_DISABLED);
        this.get(NEXT_NODE).addClass(PAGINATION_DISABLED);
        this.get(FIRST_NODE).addClass(PAGINATION_DISABLED);
        this.get(LAST_NODE).addClass(PAGINATION_DISABLED);
    },

    /**
     * Switch to the next page and update pagination and location status
     * controls.
     *
     * @method _nextPage
     * @private
     */
    _nextPage: function(evt) {
        evt.preventDefault();
        if (!evt.target.hasClass(PAGINATION_DISABLED)) {
            this._disablePagination();
            this.get(HOST).nextPage();
        }
    },

    /**
     * Switch to the previous page and update pagination and location status
     * controls.
     *
     * @method _previousPage
     * @private
     */
    _previousPage: function(evt) {
        evt.preventDefault();
        if (!evt.target.hasClass(PAGINATION_DISABLED)) {
            this._disablePagination();
            this.get(HOST).previousPage();
        }
    },

    /**
     * Switch to the first page and update pagination and location status
     * controls.
     *
     * @method _firstPage
     * @private
     */
    _firstPage: function(evt) {
        evt.preventDefault();
        if (!evt.target.hasClass(PAGINATION_DISABLED)) {
            this._disablePagination();
            this.get(HOST).jumpToPage(1);
        }
    },

    /**
     * Switch to the last page and update pagination and location status
     * controls.
     *
     * @method _lastPage
     * @private
     */
    _lastPage: function(evt) {
        evt.preventDefault();
        var lastPage = this.get(HOST).get(TOTAL_PAGES);
        if (!evt.target.hasClass(PAGINATION_DISABLED)) {
            this._disablePagination();
            this.get(HOST).jumpToPage(lastPage);
        }
    },

    /**
     * See if we've pushed a -1 offset onto the listing's offsets
     * list. If so we've reached the end and we should force last page
     * behaviour.
     *
     * @method _forceLastPage
     * @private
     **/
    _forceLastPage: function(host, currentPage, pageSize) {
        var offsetCalculation = host.get("offsetCalculation");
        // Request the offset for the next page, which is already
        // stored on the listing.  If there shouldn't be a next page
        // this offset will be -1.
        var mark = offsetCalculation.apply(this, [currentPage + 1, pageSize]);
        return mark < 0;
    },

    /**
     * Synchronize this plugin's state with the host widget's state.
     *
     * @method _syncUI
     * @private
     */
    _syncUI: function() {
        var host = this.get(HOST);
        var pageSize = host.get(PAGE_SIZE);
        var totalRecordsCount = host.get(TOTAL_RECORDS);
        var currentPage = host.get(CURRENT_PAGE);
        var firstItemCount = (currentPage - 1) * pageSize + 1;
        var existingRowCount = host.getRowCount();
        var lastItemCount = firstItemCount + existingRowCount - 1;
        var onLastPage = false;
        var onlyOnePage = false;

        // If the first page we hit is the only page of results, but
        // we don't explicitly know the totalRecordCount, then
        // totalRecordCount will already have been set to -2 and is
        // made available to use via the meta data on the results (set
        // during initialisation of the record souce), indicating that
        // this is the last page of results. However, when we switch
        // pages the initialisation that sets the meta data on the
        // DataSourceIO is not redone, so the totalRecordsCount value
        // will not be changed from its initial value.  Therefore we
        // have to have a second route to check if we're on the last
        // page of results for a listing with no defined
        // totalRecordsCount.  "Dijkstra would not have liked this".
        if (totalRecordsCount == -1 &&
            this._forceLastPage(host, currentPage, pageSize)) {
            // Set totalRecordsCount to the magic number, -2, to force
            // last page behaviour.
            totalRecordsCount = -2;
        }

        switch (totalRecordsCount) {
        case -2:
            // -2 implies that we do not know the total number of
            // records, but we do know that there are no further
            // records after this batch.
            totalRecordsCount = lastItemCount;
            onLastPage = true;
            onlyOnePage = currentPage <= 1;
            break;
        case -1:
            // -1 implies that we do not know the total number of
            // records, but we do know that there are further results
            // to be returned after this batch.
            onLastPage = existingRowCount < pageSize;
            onlyOnePage = false;
            break;
        default:
            // We know the total number of records, hurrah!
            onlyOnePage = totalRecordsCount <= pageSize;
            onLastPage = totalRecordsCount == lastItemCount;
            break;
        }
        var disableNext = onlyOnePage || onLastPage;
        var disablePrevious = firstItemCount <= 1;
        var disableFirst = currentPage <= 1;
        var disableLast = (
            this.get("disableLast") || onlyOnePage || onLastPage ||
            totalRecordsCount == -1);
        this._updateUI(firstItemCount, lastItemCount, totalRecordsCount,
                       disableNext, disablePrevious, disableFirst, disableLast,
                       this.get("hideLast"));
    },

    /**
     * Update the pagination and location status controls in the pagination box
     * to reflect the specified values.
     *
     * @method _updateUI
     * @private
     */
    _updateUI: function(firstItemCount, lastItemCount, totalRecordsCount,
                        disableNext, disablePrevious, disableFirst, disableLast,
                        hideLast) {
        this.get(FIRST_ITEM_NODE).set(TEXT, firstItemCount);
        this.get(LAST_ITEM_NODE).set(TEXT, lastItemCount);
        if (totalRecordsCount == -1) {
            totalRecordsCount = "many";
        }
        this.get(TOTAL_RECORDS_NODE).set(TEXT, totalRecordsCount);
        if (disablePrevious) {
            this.get(PREVIOUS_NODE).addClass(PAGINATION_DISABLED);
        } else {
            this.get(PREVIOUS_NODE).removeClass(PAGINATION_DISABLED);
        }
        if (disableNext) {
            this.get(NEXT_NODE).addClass(PAGINATION_DISABLED);
        } else {
            this.get(NEXT_NODE).removeClass(PAGINATION_DISABLED);
        }
        if (disableFirst) {
            this.get(FIRST_NODE).addClass(PAGINATION_DISABLED);
        } else {
            this.get(FIRST_NODE).removeClass(PAGINATION_DISABLED);
        }
        if (disableLast) {
            this.get(LAST_NODE).addClass(PAGINATION_DISABLED);
        } else {
            this.get(LAST_NODE).removeClass(PAGINATION_DISABLED);
        }
        if (hideLast) {
            this.get(LAST_NODE).addClass(PAGINATION_HIDDEN);
        }
    }
});

Y.landscape.listing.PaginationPlugin = PaginationPlugin;


}, "0.1", {"requires": ["node", "event", "event-custom", "base",
                        "plugin", "widget", "json", "datasource-io",
                        "landscape.notes", "landscape.ui.image"]});
