// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.listing.selectable", function(Y) {

Y.namespace("landscape.listing");

/**
 * Mark or unmark rows in a table listing.
 *
 * @module landscape.listing.selectable
 */


/**
 * This class provides the ability to turn the rows in a static HTML table
 * into interactive selectable elements.  It can also enable or disable
 * dependant elements, based on the selection state in the table.
 *
 * @class SelectableListing
 * @extends Widget
 * @constructor
 */
var SELECTABLE_LISTING = "selectable_listing",

    // Configuration properties
    BOUNDING_BOX = "boundingBox",
    CONTENT_BOX = "contentBox",
    CONTROLS_REQUIRING_MARKED_ROWS = "controlsRequiringMarkedRows",
    CONTROLS_REQUIRING_ACTIONABLE_ROWS = "controlsRequiringActionableRows",
    SELECT_ALL_CONTROLS = "selectAllControls",
    SELECT_NONE_CONTROLS = "selectNoneControls",
    TOGGLE_ALL_CONTROL = "toggleAllControl",
    SELECTABLE_ROWS = "selectableRows",
    DISABLED_ROWS = "disabledRows",
    SELECTED_ROWS = "selectedRows",
    ACTIONABLE_ROWS = "actionableRows",

    // Events
    MARK_ROW = "selectable_listing:mark-row",
    UNMARK_ROW = "selectable_listing:unmark-row",
    MARK_ROWS = "selectable_listing:mark-rows",
    UNMARK_ROWS = "selectable_listing:unmark-rows",

    // Local bindings
    nodesSetter = Y.landscape.util.nodesSetter,
    nodesMappingSetter = Y.landscape.util.nodesMappingSetter,
    nodeListMappingSetter = Y.landscape.util.nodeListMappingSetter;

var SelectableListing = function() {
    SelectableListing.superclass.constructor.apply(this, arguments);
};

/**
 * Static object hash used to capture existing markup for progressive
 * enhancement.
 *
 * @property SelectableListing.HTML_PARSER
 * @type Object
 * @static
 */
SelectableListing.HTML_PARSER = {
    selectableRows: ["tbody tr.clickable"],
    disabledRows: ["tbody tr.disabled"],
    selectedRows: ["tbody tr.selected"]
};

SelectableListing.NAME = SELECTABLE_LISTING;

SelectableListing.ATTRS = {
    /**
     * A list of table rows that can be selected.
     *
     * @attribute selectableRows
     * @type NodeList
     */
    selectableRows: {
        value: null
    },

    /**
     * A list of CSS selectors for controls that should be disabled unless at
     * least one row is marked.
     *
     * @attribute controlsRequiringMarkedRows
     * @type Array
     */
    controlsRequiringMarkedRows: {
        value: null,
        setter: nodesSetter
    },

    /**
     * A mapping from {action: CSS selector for actionable rows}.
     *
     * @attribute actionableRows
     * @type Object
     */
    actionableRows: {
        value: {},
        setter: nodeListMappingSetter
    },

    /**
     * A mapping from {action: [CSS selector for control]} for controls that
     * should be hidden unless at least one row is actionable.
     *
     * @attribute controlsRequiringActionableRows
     * @type Array
     */
    controlsRequiringActionableRows: {
        value: {},
        setter: nodesMappingSetter
    },

    /**
     * A list of CSS selectors for controls that, when clicked, should mark all
     * rows in the table.
     *
     * @attribute selectAllControls
     * @type Array
     */
    selectAllControls: {
        value: [],
        setter: nodesSetter
    },

    /**
     * A list of CSS selectors for controls that, when clicked, should unmark
     * all rows in the table.
     *
     * @attribute selectNoneControls
     * @type Array
     */
    selectNoneControls: {
        value: [],
        setter: nodesSetter
    },

    /**
     * The name of a checkbox used to toggle between selecting all rows and
     * selecting no rows.
     *
     * @attribute toggleAllControl
     * @type String
     */
    toggleAllControl: {
        value: null,
        setter: Y.one
    },


    /**
     * The tab index is null to prevent focusing of the content box when the
     * widget is used.
     *
     * @attribute tabIndex
     * @type Int
     */
    tabIndex: {
        value: null
    }
};

Y.extend(SelectableListing, Y.Widget, {
    /**
     * Constructor logic.
     *
     * @method initializer
     * @protected
     */
    initializer: function(config) {
        /**
         * Fires when the user marks one or more rows.
         *
         * @event mark-row
         */
        this.publish(MARK_ROW);

        /**
         * Fires when the user unmarks one or more rows.
         *
         * @event unmark-row
         */
        this.publish(UNMARK_ROW);
    },

    /**
     * Bind the widget's DOM elements to their event handlers.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        // Bind handlers for 'Select All' controls, if they're present.
        var selectAll = Y.bind(function(evt) {
            evt.preventDefault();
            this.selectAll();
        }, this);

        Y.Array.each(this.get(SELECT_ALL_CONTROLS), function(control){
            control.on("click", selectAll);
        });

        // Bind handlers for 'Select None' controls, if they're present.
        var selectNone = Y.bind(function(evt) {
            evt.preventDefault();
            this.selectNone();
        }, this);

        Y.Array.each(this.get(SELECT_NONE_CONTROLS), function(control) {
            control.on("click", selectNone);
        });

        // Bind a handler for the toggle all checkbox, if it's present.
        var toggle = Y.bind(function(evt) {
            if (evt.target.get("checked")) {
                this.selectAll();
            } else {
                this.selectNone();
            }
        }, this);

        var toggleAll = this.get(TOGGLE_ALL_CONTROL);
        if (!Y.Lang.isNull(toggleAll)) {
            toggleAll.on("click", toggle);
        }

        var contentBox = this.get(CONTENT_BOX);

        this.set(SELECTABLE_ROWS, contentBox.all("tbody tr.clickable"));
        this.set(DISABLED_ROWS, contentBox.all("tbody tr.disabled"));
        this.set(SELECTED_ROWS, contentBox.all("tbody tr.selected"));

        // Bind handlers to handle row and checkbox click events.
        // Ignore the event if a link was clicked, or if the target
        // element has a 'no-select' CSS class, to prevent the
        // underlying row from being clicked.
        contentBox.delegate("click", function(evt) {
            var target = evt.target;
            if (! (target.hasClass("no-select") ||
                   target.ancestor(".no-select") ||
                   target.get("nodeName").toUpperCase() == "A")) {
                this.toggleRow(evt.currentTarget);
            }
            evt.stopPropagation();
        }, "tbody tr.clickable", this);

        // For disabled rows, prevent the default behaviour of anything inside
        // them.  Also prevent default action on checkbox labels, otherwise two
        // events are fired for each click (one on the label and one on the
        // checkbox). The event is still propagated if the checkbox is inside
        // the label and the event is on the checkbox itself.
        contentBox.delegate("click", function(evt) {
            if (evt.target.get("nodeName").toUpperCase() != "INPUT") {
                evt.stopPropagation();
                evt.preventDefault();
            }
        }, "tbody tr.disabled, tbody label");
    },

    /**
     * Update controls based on the mark state of rows in the listing.
     *
     * @method syncUI
     * @protected
     */
    syncUI: function() {
        // Add clickable CSS class to rows that are selectable and do
        // not have it.
        var contentBox = this.get(CONTENT_BOX);
        var selector = "tbody tr:not(.clickable):not(.disabled) input[type='checkbox']";
        var newlySelectableRows = contentBox.all(selector);

        newlySelectableRows.each(function(checkbox){
            var row = checkbox.ancestor("tr");
            if (!checkbox.get("disabled")) {
                row.addClass("clickable");
            } else {
                row.addClass("disabled");
            }
        });

        // Always refresh the NodeLists, because rows might have been
        // deleted.
        var selectableRows = this.get(SELECTABLE_ROWS);
        var selectedRows = this.get(SELECTED_ROWS);
        var disabledRows = this.get(DISABLED_ROWS);
        selectableRows.refresh();
        selectedRows.refresh();
        disabledRows.refresh();

        var markedRowCount = selectedRows.size();
        var unmarkedRowCount = selectableRows.size() - markedRowCount;
        var disabled = ! markedRowCount;

        Y.Array.each(this.get(CONTROLS_REQUIRING_MARKED_ROWS), function(node) {
            node.set("disabled", disabled);
        });

        Y.Array.each(this.get(SELECT_NONE_CONTROLS), function(control) {
            if (disabled) {
                control.addClass("selected-filter");
            } else {
                control.removeClass("selected-filter");
            }
        });
        Y.Array.each(this.get(SELECT_ALL_CONTROLS), function(control){
            if (!unmarkedRowCount) {
                control.addClass("selected-filter");
            } else {
                control.removeClass("selected-filter");
            }
        });

        var toggleAll = this.get(TOGGLE_ALL_CONTROL);
        if (!Y.Lang.isNull(toggleAll)) {
            if (disabled) {
                toggleAll.set("checked", false);
            } else if (!unmarkedRowCount) {
                toggleAll.set("checked", true);
            }
        }

        var actionableRows = this.get(ACTIONABLE_ROWS),
            rows;
        Y.Object.each(this.get(CONTROLS_REQUIRING_ACTIONABLE_ROWS), function(controls, key) {
            rows = actionableRows[key];

            if (controls.size() && !Y.Lang.isUndefined(rows)) {
                // If filtering down actionable rows to clickable ones results in
                // no actionable rows, hide the associated controls.
                rows.refresh();

                if (!rows.filter(".clickable").size()) {
                    controls.each(function(control) {
                        control.hide();
                    });
                } else {
                    controls.removeClass("hidden");
                    controls.each(function(control) {
                        control.show();
                    });
                }
            }
        });
    },

    /**
     * Toggle the specified row's marked state.
     *
     * @method toggleRow
     * @private
     */
    toggleRow: function(row) {
        if (!row.hasClass("clickable")){
            return false;
        }

        var mark_row = ! row.hasClass("selected");
        var input = row.one("input[type='checkbox']");
        var changed = false;

        if (input) {
            if (input.get("checked") !== mark_row) {
                input.set("checked", mark_row);
            }
        }

        if (mark_row) {
            row.addClass("selected");
        } else {
            row.removeClass("selected");
        }

        this.syncUI();
        if (mark_row) {
            this.fire(MARK_ROW, {row: row});
        } else {
            this.fire(UNMARK_ROW, {row: row});
        }
        return true;
    },

    /**
     * Set the selected rows to the specified list of rows. If a row
     * is currently selected but not contained in this list, unselect
     * it.
     *
     * @method selectRows
     * @private
     */
    selectRows: function(rows) {
        var checkbox, selected, toSelect;
        var selectRows = [];
        var selectBoxes = [];
        var unselectRows = [];
        var unselectBoxes = [];
        var selectedRows = this.get(SELECTED_ROWS);
        var checked = "input[type='checkbox']:checked";
        var unchecked = "input[type='checkbox']:not(:checked)";

        this.get(SELECTABLE_ROWS).each(function(row) {
            selected = (selectedRows.indexOf(row) > -1);
            toSelect = (Y.Array.indexOf(rows, row) > -1);
            if (toSelect) {
                // If the row is in the list of rows to select and is
                // not selected, then select it.
                if (!selected) {
                    selectRows.push(row);
                }
                // If it should be selected but the checkbox is
                // unchecked, we'll need to check the checkbox.
                checkbox = row.one(unchecked);
                if (!Y.Lang.isNull(checkbox)) {
                    selectBoxes.push(checkbox);
                }
            } else {
                // If the row is not in the list of rows to select and is
                // selected, then unselect it.
                if (selected) {
                    unselectRows.push(row);
                }
                // If checkbox is checked, and it should be
                // unselected, we'll need to uncheck the checkbox.
                checkbox = row.one(checked);
                if (!Y.Lang.isNull(checkbox)) {
                    unselectBoxes.push(checkbox);
                }
            }
        });

        selectRows = new Y.NodeList(selectRows);
        selectBoxes = new Y.NodeList(selectBoxes);
        unselectRows = new Y.NodeList(unselectRows);
        unselectBoxes = new Y.NodeList(unselectBoxes);

        if ((selectRows.size() || selectBoxes.size()) ||
            (unselectRows.size() || unselectBoxes.size())) {

            selectRows.addClass("selected");
            selectBoxes.set("checked", true);

            unselectRows.removeClass("selected");
            unselectBoxes.set("checked", false);

            this.syncUI();
            if (selectRows.size()) {
                this.fire(MARK_ROWS, {rows: selectRows});
            }
            if (unselectRows.size()) {
                this.fire(UNMARK_ROWS, {rows: unselectRows});
            }
        }
    },

    /**
     * Mark all rows.
     *
     * @method selectAll
     * @private
     */
    selectAll: function() {
        var contentBox = this.get(CONTENT_BOX);
        var checkbox = "tbody tr.clickable input[type='checkbox']:not(:checked)";
        var notSelectedBoxes = contentBox.all(checkbox);

        if (notSelectedBoxes.size()) {
            var selectableRows = this.get(SELECTABLE_ROWS);
            selectableRows.addClass("selected");
            notSelectedBoxes.set("checked", true);
            this.syncUI();
            this.fire(MARK_ROWS, {rows: selectableRows});
        }
    },

    /**
     * Mark all rows.
     *
     * @method selectNone
     * @private
     */
    selectNone: function() {
        var contentBox = this.get(CONTENT_BOX);
        var checkbox = "tbody tr.clickable input[type='checkbox']:checked";
        var selectedBoxes = contentBox.all(checkbox);

        var count = selectedBoxes.size();
        if (count) {
            var selectedRows = this.get(SELECTED_ROWS);
            selectedRows.removeClass("selected");
            selectedBoxes.set("checked", false);
            this.syncUI();
            this.fire(UNMARK_ROWS, {rows: selectedRows});
        } else {
            this.syncUI();
        }
        return count;
    }
});

Y.landscape.listing.SelectableListing = SelectableListing;


}, "0.1", {"requires": ["node", "selector-css3", "event",
                        "widget", "base", "landscape.util.node"]});
