// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.notes", function(Y) {

Y.namespace("landscape.notes");

/**
 * Manage top-level notes.
 *
 * @module landscape.notes
 */


/**
 * This class allows to add and remove notes from the notes div.
 *
 * @class Notes
 * @extends Widget
 * @constructor
 */
var NOTES = "notes",
    // Configuration properties
    CONTENT_BOX = "contentBox",
    NOTES_CLOSE_EVENT = "notes:close_event",
    ONLY_DYNAMIC = "onlyDynamic";

var Notes = function() {
    Notes.superclass.constructor.apply(this, arguments);
};

Notes.NAME = NOTES;

Notes.ATTRS = {
    hideAnimDuration: { value: 0.7 },
    showAnimDuration: { value: 1.0 },
    /* The default tabIndex is *supposed* to be null, according to docs, but
    it's actually 0, which is inappropriate for non-interactive widgets. */
    tabIndex: { value: null },
    /* If true, assume there are no notes already defined in HTML when
     * calling addNote, in this case we only handle notes added via
     * calls to addNote in javascript.
     */
    onlyDynamic: { value: true }
};

Y.extend(Notes, Y.Widget, {

    NOTE_TEMPLATE : [
        "<div class='note-{category} note'>",
        "  <div class='note-row'>",
        "  <span class='note-category-icon note-category-{category}-icon'></span>",
        "  <span class='note-message'>{message}</span>",
        "  <span class='note-close-icon'></span>",
        "  </div>",
        "</div>"].join(""),

    initializer: function(config) {
        this.publish(NOTES_CLOSE_EVENT);
        this.pending = [];
        this.hideLater = null;
    },

    /**
     * Render notes, displaying it if it has any notes.
     *
     * @method bindUI
     * @protected
     */
    renderUI: function() {
        var container = this.get(CONTENT_BOX);
        if (container.all("#inner-notes div.note").size()) {
            this._showNote();
        }
    },

    /**
     * Bind event handlers.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        Y.one("#notes").on("click", function(evt) {
            this.hideNotes();
        }, this);
    },

    /**
     * Make the note section visible, using an animation.
     *
     * @method _showNote
     * @private
     */
    _showNote: function() {
        var div = this.get(CONTENT_BOX);
        div.setStyle("top", -85);
        div.setStyle("opacity", 0);
        div.setStyle("display", "block");
        var animation = new Y.Anim({
            node: div,
            duration: this.get("showAnimDuration"),
            easing: Y.Easing.easeOut,
            from: {top: -85, opacity: 0},
            to: {top: 45, opacity: 1}
        });

        animation.on("end", function() {
            this.hideLater = Y.later(10000, this, this.hideNotes, this);
        }, this);

        animation.run();
        var note = div.one("#inner-notes div.note");
        var category = "info";
        if (note.hasClass("note-warning")) {
            category = "warning";
        } else if (note.hasClass("note-error")) {
            category = "error";
        }
    },

    /**
     * Add new note, hiding the old ones in the process.
     *
     * @method addNote
     * @public
     */
    addNote: function(category, message) {
        var container = this.get(CONTENT_BOX).one("#inner-notes");
        var onlyDynamic = this.get(ONLY_DYNAMIC);
        if (container.one(".note")) {
            this.pending.unshift([category, message]);

            // If we don't have any statically defined notes then we
            // don't need to hide anything when adding a note.
            if (!onlyDynamic) {
                this.hideNotes();
            }
            this.on(NOTES_CLOSE_EVENT, function() {
                var pending = this.pending.pop();
                if (pending) {
                    this.addNote(pending[0], pending[1]);
                }
            }, this);
            return;
        }

        container.append(Y.substitute(this.NOTE_TEMPLATE, {
            category: category,
            message: message}));

        this._showNote();
    },


    /**
     * Hide the notes.
     *
     * @method hideNotes
     * @public
     */
    hideNotes: function() {
        if (!Y.Lang.isNull(this.hideLater)) {
            this.hideLater.cancel();
        }
        this.hideLater = null;
        var div = this.get(CONTENT_BOX);
        var animation = new Y.Anim({
            node: div,
            duration: this.get("hideAnimDuration"),
            easing: Y.Easing.easeOut,
            from: {opacity: 1},
            to: {opacity: 0}
        });

        animation.on("end", function() {
            div.setStyle("display", "none");
            div.setStyle("opacity", 1);
            // Delete previous note
            div.all("#inner-notes div.note").remove();
            this.fire(NOTES_CLOSE_EVENT);
        }, this);
        animation.run();
    }

});

Y.landscape.notes.Notes = Notes;

/**
 * This class allows to add and remove notes inside a table.
 *
 * @class TableNotes
 * @extends Widget
 * @constructor
 */
var TABLE_NOTES = "table_notes",
    TABLE_NOTES_CLOSE_EVENT = "table_notes:close_event",

    // Local aliases
    getCN = Y.ClassNameManager.getClassName,

    // Local classes
    CLASS_PREFIX = ".",
    TABLE_NOTE_CLASS = getCN(TABLE_NOTES, 'note'),
    CLOSE_BUTTON_CLASS = getCN(TABLE_NOTES, 'close-button');

var TableNotes = function() {
    TableNotes.superclass.constructor.apply(this, arguments);
};

TableNotes.NAME = NOTES;

TableNotes.ATTRS = {
    /* The default tabIndex is *supposed* to be null, according to docs, but
    it's actually 0, which is inappropriate for non-interactive widgets. */
    tabIndex: { value: null }
};

Y.extend(TableNotes, Y.Widget, {

    NOTE_TEMPLATE : [
        "<tr>",
        "  <td colspan='{span}' ",
        "        class='{category}-note center-cell'>",
        "    {message}<a class='{close}' href='#'>Dismiss</a>",
        "  </td>",
        "</tr>"].join(""),
    TABLE_NOTE_CLASS : TABLE_NOTE_CLASS,
    CLOSE_BUTTON_CLASS : CLOSE_BUTTON_CLASS,

    initializer: function(config) {
        this.publish(TABLE_NOTES_CLOSE_EVENT);
    },

    /**
     * Render notes, displaying it if it has any notes.
     *
     * @method bindUI
     * @protected
     */
    renderUI: function() {
        var contentBox = this.get(CONTENT_BOX);
        var notes = contentBox.all(CLASS_PREFIX + TABLE_NOTE_CLASS);
        notes.setStyle("display", "table-row");
    },

    /**
     * Bind event handlers.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        this.get(CONTENT_BOX).all(
            CLASS_PREFIX + CLOSE_BUTTON_CLASS).on("click", function(evt) {
                evt.preventDefault();
                this.hideNotes();
        }, this);
    },

    /**
     * Add new note, hiding the old ones in the process.
     *
     * @method addNote
     * @public
     */
    addNote: function(category, message) {
        var body = this.get(CONTENT_BOX).one("tbody");
        var firstRow = body.one("tr");
        var size = firstRow.all("td").size();

        var messageRow = new Y.Node.create(
            Y.substitute(this.NOTE_TEMPLATE, {
                span: size,
                category: category,
                message: message,
                close: CLOSE_BUTTON_CLASS}));

        messageRow.one(
            CLASS_PREFIX + CLOSE_BUTTON_CLASS).on("click", function(evt) {
            evt.preventDefault();
            this.hideNotes();
        }, this);

        messageRow.addClass(TABLE_NOTE_CLASS);
        body.prepend(messageRow);
    },

    /**
     * Hide the notes.
     *
     * @method hideNotes
     * @public
     */
    hideNotes: function() {
        // Delete previous notes
        var contentBox = this.get(CONTENT_BOX);
        contentBox.all(CLASS_PREFIX + TABLE_NOTE_CLASS).remove();
        this.fire(TABLE_NOTES_CLOSE_EVENT);
    }
});

Y.landscape.notes.TableNotes = TableNotes;

}, "0.1", {"requires": ["substitute", "node", "widget", "anim"]});
