// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.notifications.transport", function(Y) {

Y.namespace("landscape.notifications");

/**
 * DataSource-related utilities that are specific to Landscape.
 *
 * @module landscape.notifications
 */

/**
 * A DataSource that never returns, useful for testing.
 *
 * @class BlackholeDataSource
 * @extends DataSource.Local
 * @constructor
 */

var BlackholeDataSource = function() {
    BlackholeDataSource.superclass.constructor.apply(this, arguments);
};

BlackholeDataSource.NAME = "dataSourceDummy";

Y.extend(BlackholeDataSource, Y.DataSource.Local, {

    _defRequestFn: function(evt) {
    }
});

Y.landscape.notifications.BlackholeDataSource = BlackholeDataSource;

/**
 * A DataSource that always fails, useful for testing.
 *
 * @class DataSourceError
 * @extends DataSource.Local
 * @constructor
 */

var DataSourceError = function() {
    DataSourceError.superclass.constructor.apply(this, arguments);
};

DataSourceError.NAME = "dataSourceError";

DataSourceError.ATTRS = {
    status: { value: undefined }
};

Y.extend(DataSourceError, Y.DataSource.Local, {

    _defRequestFn: function(evt) {
        evt.error = new Error("Expected data failure");
        evt.data = {responseText: this.get("source")};
        var status = this.get("status");
        if (status) {
            evt.data.status = status;
        }
        this.fire("data", evt);
    }
});

Y.landscape.notifications.DataSourceError = DataSourceError;

/**
 * A helper for handling actions and notifications from an
 * ever-polling DataSource backed by Landscape's Ajax framework.
 *
 * @class Transport
 * @extends Base
 * @constructor
 */

var Transport = function() {
    Transport.superclass.constructor.apply(this, arguments);
};

Transport.NAME = "transport";
Transport.EVENT_PREFIX = "notification";

Transport.ATTRS = {

    /**
     * @attribute page_key
     * @description A unique identifier to include in the request to the AJAX
     *     frontend when polling for notifications and action results.
     * @type string
     */
    page_key: {
        value : null
    },

    /**
     * @attribute source
     * @description The DataSource object to use for sending HTTP requests to
     *     the AJAX frontend.
     * @type DataSource
     */
    source: {
        value: new Y.DataSource.IO({source: "/ajax/"})
    },

    /**
     * @attribute later
     * @description The function to use for scheduling actions in the future.
     * @type function
     */
    later: {
        value: Y.later
    },

    /**
     * @attribute io
     * @description The DataSource class to use for creating DataSource objects
     *     when sending actions to the server.
     * @type class
     */
    io: {
        value: Y.DataSource.IO,
        cloneDefaultValue: false
    }
};

Y.extend(Transport, Y.Base, {

    initializer: function() {
        this.page_key = this.get("page_key");
        this._source = this.get("source");
        this._later = this.get("later");
        this._handlers = {};
        this._pendingDeferreds= {};
        this._subscribers = {};
        this._sequence = 0;

        // Delay first poll until the 'onload' event fires such that
        // Webkit-based browsers do not get stuck in the 'loading' state
        // forever.
        Y.on("load", Y.bind(this._pollPage, this));
    },

    /**
     * Schedule a new polling of the DataSource for 'timeout'
     * miliseconds in the future.
     *
     * @method _schedulePollPage
     * @private
     */
    _schedulePollPage: function(timeout) {
        this._later(timeout, this, this._pollPage);
    },

    /**
     * Poll the DataSource for new data, keeping a sequence of how many
     * times the DataSource has been polled.
     *
     * @method _pollPage
     * @private
     */
    _pollPage: function() {
        this._sequence += 1;
        this._source.sendRequest({request: null, callback: {
            success: Y.bind(this._pageSuccess, this),
            failure: Y.bind(this._pageFailure, this)
        }, cfg: {method: "POST", data: "uuid=" + this.page_key + "&sequence=" + this._sequence}});
        this._stopped = false;
    },

    /**
     * Set a handler for a specific action, returning a
     * Deferred. Attach callbacks and errbacks to the returned
     * Deferred to handle results from the DataSource.
     *
     * @method setHandler
     * @public
     */
    setHandler: function(action) {
        if (action in this._pendingDeferreds) {
            var pendingDeferred = this._pendingDeferreds[action];
            delete this._pendingDeferreds[action];
            return pendingDeferred;
        }
        var deferred = new Y.landscape.deferred.Deferred();
        this._handlers[action] = deferred;
        if (this._stopped) {
            this._pollPage();
        }
        return deferred;
    },

    /**
     * Fire a subscriber for a key with a result.
     */
    _fireSubscriber: function(key, result) {
        var subscriber = this._subscribers[key];
        if (subscriber) {
            Y.Array.each(subscriber, function(callback) {
                callback(result);
            });
        }
    },

    /**
     * Handle a success result from the DataSource. If the payload
     * contains an action, call the handler registered for that
     * action. If the payload contains a notification, call the
     * subscriber registered for that notification. After handling the
     * result, reschedule a new poll immediately.
     *
     * @method _pageSuccess
     * @private
     */
    _pageSuccess: function(result) {
        if (!result.data.responseText) {
            // If there is no data, it means that Twisted closed the connection
            // without a response, likely because of a restart. Let's retry
            // later.
            this._schedulePollPage(60000);
            return;
        }
        var data = Y.JSON.parse(result.data.responseText),
            event, details;

        if (data.action) {            
            var deferred = this._handlers[data.action];
            if (deferred) {
                delete this._handlers[data.action];
                deferred.callback(data.result[0]);
            } else {
                this._pendingDeferreds[data.action] = Y.landscape.deferred.succeed(
                    data.result[0]);
            }
            this._fireSubscriber(data.action, data.result[1]);
            event = data.action;
            details =  {result: data.result[0]};
        } else if (data.notification) {
            this._fireSubscriber(data.notification, data.result);
            event = data.notification;
            details = {result: data.result};
        }

        this._fireEvent(event, details);

        this._schedulePollPage(100);
    },

    /**
     * Handle a failure result from the DataSource. If the DataSource
     * failed badly or returned no results, stop polling. If the
     * DataSource timed out, reschedule for a long time in the
     * future. Otherwise, if the DataSource returned a failure action,
     * trigger the handlers for that action and reschedule the transport.
     *
     * @method _pageFailure
     * @private
     */
    _pageFailure: function(result) {
        if (result.data.status >= 500) {
            this._stopped = true;
        } else if (result.data.status === 0 || result.data.status == 408) {
            this._schedulePollPage(60000);
        } else {
            if  (!result.data.responseText) {
                this._stopped = true;
            } else {
                var data = Y.JSON.parse(result.data.responseText);
                var deferred = this._handlers[data.action];
                if (deferred) {
                    delete this._handlers[data.action];
                    deferred.errback(data.error);
                } else {
                    this._pendingDeferreds[data.action] = Y.landscape.deferred.fail(
                        data.error);
                }
                this._fireEvent(data.action, {error: data.error});
                this._schedulePollPage(500);
            }
        }
    },

    /**
     * Fire the given event with the given details, publishing the event first
     * if it's a brand new one.
     */
    _fireEvent: function(event, details) {
        if (!this.getEvent(event)) {
            this.publish(event, {broadcast: 1});
        }
        this.fire(event, details);
    },

    /**
     * Send an action to the server with the given parameters.
     *
     * @method sendAction
     * @public
     * @return A Deferred firing with the result of the action, or failing
     *         with an error if either the action couldn't be sent or it
     *         was sent but returned an error itself (in both cases the
     *         original error is returned).
     */
    sendAction: function(config) {
        config = config || {};
        var action = config.action,
            path = action.replace(/_/g, "-") + "-action",
            fullpath = config.base ? config.base + "/" + path : path,
            url = config.url ? config.url : fullpath,
            IO = this.get("io"),
            source = new IO({source: url}),
            deferred = new Y.landscape.deferred.Deferred(),
            cfg = {method: "POST",
                   data: Y.mix({page_key: this.page_key}, config.data),
                   form: config.form ? {id: config.form, useDisabled: true} : null};
        source.sendRequest({request: null, callback: {
            success: Y.bind(this._sendSuccess, this, action, deferred),
            failure: Y.bind(this._sendFailure, this, action, deferred)
        }, cfg: cfg});

        return deferred;
    },

    /**
     * Handle a successful action submission, subscribing for its result.
     *
     * @method _sendSuccess
     * @private
     */
    _sendSuccess: function(action, deferred, evt) {
        var result;
        if (evt.data.responseText) {
            result = Y.JSON.parse(evt.data.responseText);
            if (!Y.Lang.isNull(result)) {
                if (result.hasOwnProperty("success")) {
                    // For backward compatibility with the old JobPage
                    result = result.success;
                }
                // We already got a response from the server, this means no job
                // was created and we can fire the deferred right away.
                deferred.callback(result);
                return;
            }
        }
        // A job is running, let's add a handler for it.
        this.setHandler(action).addCallbacks(
            function(result) {deferred.callback(result);},
            function(result) {deferred.errback(result);});
    },

    /**
     * Handle a failed action submission, firing back its error.
     *
     * @method _sendSuccess
     * @private
     */
    _sendFailure: function(action, deferred, result) {
        var message;
        if (result.data.status >= 500) {
            message = "Service unavailable.";
        } else{
            message = result.data.responseText;
        }
        deferred.errback(message);
    }
});

Y.landscape.notifications.Transport = Transport;

// Tests don't usually define an actual page key
var page_key = Y.config.landscape ? Y.config.landscape.pageKey : null,
    transport;

if (page_key) {
    transport = new Transport({page_key: page_key});
    // XXX expose the Transport instance for backward compatibility, as
    // some code uses it directly
    Y.landscape.notifications.transport = transport;
}

}, "0.1", {"requires": ["node", "datasource-local", "datasource-io",
                        "landscape.deferred", "json-parse",
                        "querystring-stringify-simple"]});
