// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.packages.listing", function(Y) {

Y.namespace("landscape.packages");

/**
 * This module provides the PackageListing widget.  It shows packages in a list
 * and provide controls to mark them for installation, upgrade, downgrade or
 * removal.
 *
 * The PackageListing widget uses a number of internal components to provide
 * the package listing behaviour.  PackageIndex provides an API for JSON dat
 * generated by the server.  This index is used when building PackageSummaryRow
 * objects, one for each package name in the dataset.  The index is also used
 * when building PackageVersionRow objects representing the relationship of
 * computers to available, installed, upgradable or downgradable packages.  A
 * special PackageVersionFilterRow provides controls that allow the user to
 * change which versions are currently displayed.
 *
 * @module landscape.packages.listing
 */


/**
 * Add a node as the next sibling of an existing node.
 *
 * @function insertNodeAfter
 */
var insertNodeAfter = function(existingNode, newNode) {
    // FIXME Implement this on Y.Node and send a patch upstream.

    // FIXME The logic here will break if you use an existing node with no
    // parent.  This is probably why the provided insertBefore method must be
    // called on the parent of the nodes you wish to manipulate.
    var parentNode = existingNode.ancestor();
    var nextNode = existingNode.next();
    if (nextNode !== null) {
        parentNode.insertBefore(newNode, nextNode);
    } else {
        parentNode.appendChild(newNode);
    }
    return newNode;
};

Y.landscape.packages.insertNodeAfter = insertNodeAfter;


/**
 * Concatenate arguments together using dash to create name suitable for use as
 * a CSS class or HTML ID.  Dot and plus symbols in arguments are replaced with
 * dash.
 *
 * @function getCN
 */
var getCN = function() {
    var sanitizedArguments = [];
    Y.Array.each(arguments, function(argument) {
        argument = argument.replace(/\./g, "-");
        argument = argument.replace(/\+/g, "-");
        sanitizedArguments.push(argument);
    });
    return Y.ClassNameManager.getClassName.apply(this, sanitizedArguments);
};

// Takes a node or a selector and returns a Node, defaulting to
// globally defined selector if none is provided.
var nodeSetter = function(node, name) {
    if (Y.Lang.isUndefined(node) || Y.Lang.isNull(node)){
        node = DEFAULT_NODES[name];
    }
    return Y.bind(Y.landscape.util.nodeSetter, this)(node, name);
};

// Constants used to represent package relationships.  These are
// duplicated from the private constants defined in test_listing.js
// and in canonical.landscape.browser.package.listing.
var AVAILABLE = 1,
    INSTALLED = 2,
    UPGRADE = 3,
    DOWNGRADE = 4,
    PARTIAL = 5,
    LOCKED = 6,
    UNLOCKED = 7;


// Constants are used to identify class and attribute names.  These are used
// primarily as a means of getting as much shrinkage out of the minification
// process as possible.  Another slight benefit of is that they result in
// fewer strings being allocated when the Javascript is processed by the
// browser.
var PACKAGE_LISTING = "package_listing",
    PACKAGE_SUMMARY_BATCH = "package_summary_batch",
    PACKAGE_SUMMARY_ROW = "package_summary_row",
    PACKAGE_SUMMARY_ROW_CLASS = "package-summary-row",
    PACKAGE_SUMMARY_ROW_TOGGLE = "package-summary-row-toggle",
    PACKAGE_VERSION_FILTER_ROW = "package_version_filter_row",
    PACKAGE_VERSION_ROW = "package_version_row",
    ACTION_BUTTON = "action-button",
    DOT = ".",
    CLICKABLE = "clickable",
    CONTENT_BOX = "contentBox",
    PACKAGE_DATA = "packageData",
    COMPUTER_URL = "computerURL",
    PACKAGE_LOADING_STATUS = "packageLoadingStatus",
    SELECT_NONE_CONTROLS = "selectNoneControls",
    SELECT_UPGRADES_CONTROLS = "selectUpgradesControls",
    SUMMARY_ROWS = "summaryRows",
    SUMMARY_ROWS_RENDERED = "summaryRowsRendered",
    RESULT = "result",
    KIND = "kind",
    IMAGE = "image",
    LOCK_STATE = "lockState",
    HAS_AVAILABLE = "hasAvailable",
    HAS_INSTALLED = "hasInstalled",
    HAS_UPGRADES = "hasUpgrades",
    HAS_REMOVABLE = "hasRemovable",
    ATTR_AVAILABLE = "available",
    ATTR_INSTALLED = "installed",
    ATTR_INSTALL = "install",
    ATTR_REMOVE = "remove",
    ATTR_UPGRADE = "upgrade",
    ATTR_DOWNGRADE = "downgrade",
    ATTR_HOLD = "hold",
    ATTR_UNHOLD = "unhold",
    PACKAGE_DATA_ATTRS = [ATTR_AVAILABLE, ATTR_INSTALLED,
                          ATTR_UPGRADE, ATTR_DOWNGRADE],
    PACKAGE_INFO = "packageInfo",
    PACKAGE_ICON = "packageIcon",
    BOUNDING_BOX = "boundingBox",
    WIDGET = "widget",
    VERSION_FILTER_ROW = "versionFilterRow",
    VERSION_ROWS = "versionRows",
    EXPANDABLE = "landscape-expandable-tree-expand",
    EXPANDED = "expanded",
    VERSION_ROWS_RENDERED = "versionRowsRendered",
    EVT_BEFORE_PAGE_CHANGE = "before-page-change",
    COLLAPSE_ICON_PATH = "collapseIconPath",
    EXPAND_ICON_PATH = "expandIconPath",
    MARK_FOR_INSTALL = "ls-pkg-tree-mark-for-install",
    MARK_FOR_INSTALL_ICON_PATH = "markForInstallIconPath",
    MARK_FOR_INSTALL_DISABLED_ICON_PATH = "markForInstallDisabledIconPath",
    MARK_FOR_UPGRADE = "ls-pkg-tree-mark-for-upgrade",
    MARK_FOR_UPGRADE_ICON_PATH = "markForUpgradeIconPath",
    MARK_FOR_UPGRADE_DISABLED_ICON_PATH = "markForUpgradeDisabledIconPath",
    MARK_FOR_DOWNGRADE = "ls-pkg-tree-mark-for-downgrade",
    MARK_FOR_DOWNGRADE_ICON_PATH = "markForDowngradeIconPath",
    MARK_FOR_DOWNGRADE_DISABLED_ICON_PATH = "markForDowngradeDisabledIconPath",
    MARK_FOR_REMOVAL = "ls-pkg-tree-mark-for-removal",
    MARK_FOR_REMOVAL_ICON_PATH = "markForRemovalIconPath",
    MARK_FOR_REMOVAL_DISABLED_ICON_PATH = "markForRemovalDisabledIconPath",
    MARK_FOR_HOLD = "ls-pkg-tree-mark-for-hold",
    MARK_FOR_HOLD_ICON_PATH = "markForHoldIconPath",
    MARK_FOR_HOLD_DISABLED_ICON_PATH = "markForHoldDisabledIconPath",
    MARK_FOR_UNHOLD = "ls-pkg-tree-mark-for-unhold",
    MARK_FOR_UNHOLD_ICON_PATH = "markForUnholdIconPath",
    MARK_FOR_UNHOLD_DISABLED_ICON_PATH = "markForUnholdDisabledIconPath",
    LOCKED_ICON_PATH = "lockedIconPath",
    AVAILABLE_ICON_PATH = "availableIconPath",
    INSTALLED_ICON_PATH = "installedIconPath",
    PARTIAL_ICON_PATH = "partialIconPath",
    INSTALLING_ICON_PATH = "installingIconPath",
    UPGRADING_ICON_PATH = "upgradingIconPath",
    PARTIALING_ICON_PATH = "partialingIconPath",
    DOWNGRADING_ICON_PATH = "downgradingIconPath",
    REMOVING_ICON_PATH = "removingIconPath",
    SUMMARY_ROW = "summaryRow",
    PARENT_NODE = "parentNode",
    CACHE_SIZE = "cacheSize",
    PAGE_SIZE = "pageSize",
    PAGER = "pager",
    PAGINATION_BOX = "paginationBox",
    TOTAL_RECORDS = "totalRecords",
    NODE = "node",
    ID = "id",
    RENDER_LOOP_SIZE = "renderLoopSize",
    RELATION = "relation",
    ROW = "row",
    TOGGLE_ICON = "toggleIcon",
    LATEST_VERSION = "latest",
    SELECTED_INDEX = "selectedIndex",
    TEXT = "text",
    VALUE = "value",
    MARKED = "marked",
    UNMARKED = "unmarked",
    GUTTER_ROW = "gutterRow",
    DEFAULT_NODES = {
        paginationBox: ".package-listing-pagination"
    };

/**
 * @class PackageIndex
 *
 * PackageIndex provides an API on top of JSON data about packages.
 *
 * The JSON data structure provides information about packages and about the
 * relationship of those packages to computers.
 *
 *   var data = {
 *        packages: {
 *            1: {name: "iproute", version: "1.2-1ubuntu3",
 *                summary: "Professional networking tools",
 *                usn: {name: "123-4", summary: "Off-by-one exploit"}},
 *            2: {name: "iproute", version: "1.1-4ubuntu1",
 *                summary: "Professional networking tools"}
 *        },
 *        relations: [{name: iproute, package_ids: [2, 1],
 *                    securityUpgrades: 1, upgrades: 1,
 *                    relations: [{computers: [1], installed: 2, upgrade: 1
 *                                 kind: UPGRADE, latest: true, held: false},
 *                                {computers: [2], available: 2
 *                                 kind: AVAILABLE, latest: true, held: false},
 *                                {computers: [3], installed: 1,
 *                                 kind: INSTALLED, latest: true, held: false},
 *                                {computers: [3], installed: 1, downgrade: 2,
 *                                 kind: DOWNGRADE, latest: true,
 *                                 held: false}]}],
 *        computers: {
 *            1: {title: "kip"},
 *            2: {title: "rex"},
 *            3: {title: "napoleon"}
 *        }
 *   };
 *
 * The 'usn' attribute of the 'data.packages' data objects may
 * be omitted if they don't apply to a particular package.
 *
 */
var PackageIndex = function() {
    this._packageNameIndex = {}; // {packageName: [packageData, ...]}
    this._packageData = [];
};

PackageIndex.prototype = {

    setData: function(data) {
        this._packageNameIndex = {};
        this._packageData = [];
        return this.update(data);
    },

    /**
     * Add 'id' property from object key into value.
     */
    _addIDProperty: function(object) {
        for (var id in object) {
            if (object.hasOwnProperty(id)) {
                object[id].id = id;
            }
        }
    },

    /**
     * Add packages from JSON data to the package index.  Returns the list of
     * new package names added to the index.
     */
    update: function(data) {
        var packageNames = [];
        var allRelations = data.relations;
        var allPackages = data.packages;
        var allComputers = data.computers;

        if (Y.Lang.isUndefined(allRelations) ||
            Y.Lang.isUndefined(allPackages) ||
            Y.Lang.isUndefined(allComputers)) {
            return [];
        }

        // Add the id to the mapping value. It's used in other places and isn't
        // set by the server to avoid duplication.
        this._addIDProperty(allPackages);
        this._addIDProperty(allComputers);

        for (var i in allRelations) {
            var packageData = allRelations[i];
            var packageName = packageData.name;
            var packageIds = packageData.package_ids;
            var packageRelations = packageData.relations;

            // Replace values in packageData with their actual counterparts.
            packageData.packages = [];
            for (var pid in packageData.package_ids) {
                packageData.packages[pid] = allPackages[packageIds[pid]];
            }

            for (var j in packageRelations) {
                var relation = packageRelations[j];

                for (var k in PACKAGE_DATA_ATTRS) {
                    var name = PACKAGE_DATA_ATTRS[k];
                    if (relation.hasOwnProperty(name)) {
                        relation[name] = allPackages[relation[name]];
                    }
                }

                // for a relation with a single computer, we need the full
                // computer object, not just the id
                if (relation.computers.length == 1) {
                    relation.computer = allComputers[relation.computers[0]];
                }
            }
            packageNames.push(packageName);
            this._packageData.push(packageData);
            this._packageNameIndex[packageName] = packageData;
        }
        return packageNames;
    },

    /**
     * Returns an object with information about packages matching the specified
     * name and the relations those packages have.
     * The returned object will match this structure:
     *
     *   var packageInfo = {
     *       name: $name,
     *       packages: [$package1, $package2, ...],
     *       upgrades: 1,
     *       securityUpgrades: 0,
     *       relations: [
     *           {kind: AVAILABLE, computers: [$computer_id]
     *            available: $package, latest: $bool, held: $bool},
     *           {kind: INSTALLED, computers: [$computer_id],
     *            installed: $package, latest: $bool, held: $bool},
     *           {kind: UPGRADE, computer: [$computer_id, $computer_id],
     *            installed: $package, upgrade: $package,
     *            latest: $bool, held: $bool},
     *           {kind: DOWNGRADE, computer: [$computer_id],
     *            installed: $package, downgrade: $package,
     *            latest: $bool, held: $bool}
     *       ]
     *   };
     *
     * 'relations' sorted by package, newest before eldest, and then by
     * computer title.
     */
    getPackageInfo: function(name) {
        if (!this._packageNameIndex.hasOwnProperty(name)) {
            return undefined;
        }
        return this._packageNameIndex[name];
    },

    /**
     * Returns all the package data that we have, ordered by
     * packageName.
     */
    getPackageData: function() {
        return this._packageData;
    }
};

Y.landscape.packages.PackageIndex = PackageIndex;


/**
 * PackageListing displays a list of packages created from a JSON data
 * structure processed by a PackageIndex.  Summary rows, rendered first, have
 * controls to mark selected computers for install, upgrade or removal.  A
 * disclosure triangle can be used to open each row to see details about the
 * computers that either have the package available, installed, upgradable or
 * downgradable.  When opened, version rows are dynamically rendered, and have
 * controls that allows users to mark actions for the package on individual
 * computers.
 *
 * @class PackageListing
 * @extends Widget
 * @constructor
 */
var PackageListing = function() {
    PackageListing.superclass.constructor.apply(this, arguments);
};

PackageListing.NAME = PACKAGE_LISTING;

PackageListing.ATTRS = {
    /**
     * JSON data containing information about packages, computers and the
     * relationships between them.
     *
     * @attribute packageData
     * @type Object
     */
    packageData: {
        value: {}
    },

    /**
     * The base URL to use when building package URLs.
     *
     * @attribute computerURL
     * @type String
     */
    computerURL: {
        value: ""
    },

    /**
     * A list of PackageSummaryRow instances contained in the package listing.
     *
     * @attribute summaryRows
     * @type Array
     */
    summaryRows: {
        valueFn: function() {
            if (this._summaryRows === undefined) {
                this._summaryRows = [];
            }
            return this._summaryRows;
        }
    },

    /**
     * Optionally, the selector for the package loading status element.  This
     * element will be removed when rendering completes.
     *
     * @attribute packageLoadingStatus
     * @type String
     */
    packageLoadingStatus: {
        value: null
    },

    /**
     * A list of selectors for 'Select None' controls.
     *
     * @attribute selectNoneControls
     * @type Array
     */
    selectNoneControls: {
        value: []
    },

    /**
     * A list of selectors for 'Select Upgrades' controls.
     *
     * @attribute selectUpgradesControls
     * @type Array
     */
    selectUpgradesControls: {
        value: []
    },

    /**
     * The path to an icon representing a collapsed disclosure triangle.
     *
     * @attribute collapseIconPath
     * @type String
     */
    collapseIconPath: {
        value: "/static/skin/manx/collapse-tree-14x14.png"
    },

    /**
     * The path to an icon representing an expanded disclosure triangle.
     *
     * @attribute expandIconPath
     * @type String
     */
    expandIconPath: {
        value: "/static/skin/manx/expand-tree-14x14.png"
    },

    /**
     * The path to an icon representing a button used to mark packages for
     * installation.
     *
     * @attribute markForInstallIconPath
     * @type String
     */
    markForInstallIconPath: {
        value: "/static/package/mark-for-install-16x16.png"
    },

    /**
     * The path to an icon representing a disabled mark packages for
     * installation button.
     *
     * @attribute markForInstallDisabledIconPath
     * @type String
     */
    markForInstallDisabledIconPath: {
        value: "/static/package/mark-for-install-disabled-16x16.png"
    },

    /**
     * The path to an icon representing a button used to mark packages for
     * upgrade.
     *
     * @attribute markForUpgradeIconPath
     * @type String
     */
    markForUpgradeIconPath: {
        value: "/static/package/mark-for-upgrade-16x16.png"
    },

    /**
     * The path to an icon representing a disabled mark packages for
     * upgrade button.
     *
     * @attribute markForUpgradeDisabledIconPath
     * @type String
     */
    markForUpgradeDisabledIconPath: {
        value: "/static/package/mark-for-upgrade-disabled-16x16.png"
    },

    /**
     * The path to an icon representing a button used to mark packages for
     * downgrade.
     *
     * @attribute markForDowngradeIconPath
     * @type String
     */
    markForDowngradeIconPath: {
        value: "/static/package/mark-for-downgrade-16x16.png"
    },

    /**
     * The path to an icon representing a disabled mark packages for
     * downgrade button.
     *
     * @attribute markForDowngradeDisabledIconPath
     * @type String
     */
    markForDowngradeDisabledIconPath: {
        value: "/static/package/mark-for-downgrade-disabled-16x16.png"
    },

    /**
     * The path to an icon representing a button used to mark packages for
     * removal.
     *
     * @attribute markForRemovalIconPath
     * @type String
     */
    markForRemovalIconPath: {
        value: "/static/package/mark-for-removal-16x16.png"
    },

    /**
     * The path to an icon representing a disabled mark packages for
     * removal button.
     *
     * @attribute markForRemovalDisabledIconPath
     * @type String
     */
    markForRemovalDisabledIconPath: {
        value: "/static/package/mark-for-removal-disabled-16x16.png"
    },

    /**
     * The path to an icon representing a mark packages for
     * hold button.
     *
     * @attribute markForHoldIconPath
     * @type String
     */
    markForHoldIconPath: {
        value: "/static/package/mark-for-install-16x16.png"
    },

    /**
     * The path to an icon representing a disabled mark packages for
     * hold button.
     *
     * @attribute markForHoldDisabledIconPath
     * @type String
     */
    markForHoldDisabledIconPath: {
        value: "/static/package/mark-for-install-disabled-16x16.png"
    },

    /**
     * The path to an icon representing a mark packages for
     * unhold button.
     *
     * @attribute markForUnholdIconPath
     * @type String
     */
    markForUnholdIconPath: {
        value: "/static/package/mark-for-removal-16x16.png"
    },

    /**
     * The path to an icon representing a disabled mark packages for
     * unhold button.
     *
     * @attribute markForUnholdDisabledIconPath
     * @type String
     */
    markForUnholdDisabledIconPath: {
        value: "/static/package/mark-for-removal-disabled-16x16.png"
    },

    /**
     * The path to an icon overlay representing a lock.
     *
     * @attribute lockedIconPath
     * @type String
     */
    lockedIconPath: {
        value: "/static/locked-16x16.png"
    },

    /**
     * The path to an icon representing a set of packages that are all
     * available, with none actually installed.
     *
     * @attribute availableIconPath
     * @type String
     */
    availableIconPath: {
        value: "/static/package/available-package-16x16.png"
    },

    /**
     * The path to an icon representing a set of packages that are all
     * installed.
     *
     * @attribute installedIconPath
     * @type String
     */
    installedIconPath: {
        value: "/static/package/installed-package-16x16.png"
    },

    /**
     * The path to an icon representing a set of packages that are both
     * available and installed.
     *
     * @attribute installedIconPath
     * @type String
     */
    partialIconPath: {
        value: "/static/package/partial-package-16x16.png"
    },

    /**
     * The path to an icon representing a package marked for install, upgrade
     * or downgrade as well as removal.
     *
     * @attribute partialingIconPath
     * @type String
     */
    partialingIconPath: {
        value: "/static/package/partial-modification-package-16x16.png"
    },

    /**
     * The path to an icon representing a package marked for installation.
     *
     * @attribute installingIconPath
     * @type String
     */
    installingIconPath: {
        value: "/static/package/installing-package-16x16.png"
    },

    /**
     * The path to an icon representing a package marked for upgrade.
     *
     * @attribute upgradingIconPath
     * @type String
     */
    upgradingIconPath: {
        value: "/static/package/upgrading-package-16x16.png"
    },

    /**
     * The path to an icon representing a package marked for downgrade.
     *
     * @attribute downgradingIconPath
     * @type String
     */
    downgradingIconPath: {
        value: "/static/package/downgrading-package-16x16.png"
    },

    /**
     * The path to an icon representing a package marked for removal.
     *
     * @attribute removingIconPath
     * @type String
     */
    removingIconPath: {
        value: "/static/package/removing-package-16x16.png"
    },

    /**
     * The path to an icon representing the existence of package upgrades.
     *
     * @attribute packageUpgradesIconPath
     * @type String
     */
    packageUpgradesIconPath: {
        value: "/static/package/package-upgrades-alert-16x16.png"
    },

    /**
     * The path to an icon representing the existence of security upgrades.
     *
     * @attribute securityUpgradesIconPath
     * @type String
     */
    securityUpgradesIconPath: {
        value: "/static/package/security-upgrades-alert-16x16.png"
    },

    /**
     * The tab index is null to prevent focusing of the content box when the
     * widget is used.
     *
     * @attribute tabIndex
     * @type Int
     */
    tabIndex: {
        value: null
    },

    /**
     * Cache size to be used for the underlying
     * DataSources. Effectively, the number of pages to cache.
     *
     * @attribute cacheSize
     * @type Integer
     */
    cacheSize: {
        value: 10
    },

    /**
     * Pager instance managed by this widget.
     *
     * @attribute pager
     * @type Pager
     */
    pager: {
        value: null
    },

    /**
     * Page size to be used by the pager.
     *
     * @attribute pageSize
     * @type Integer
     */
    pageSize: {
        value: 25
    },

    /**
     * Node where the pagination box should be rendered.
     *
     * @attribute paginationBox
     * @type Node
     */
    paginationBox: {
        value: null,
        setter: nodeSetter
    },

    /**
     * The total number of results available.
     *
     * @attribute totalRecords
     * @type Integer
     */
    totalRecords: {
        value: null
    },

    /**
     * The arguments passed to get results from subsequent pages.
     *
     * @attribute sourceArguments
     * @type Object
     */
    sourceArguments: {
        value: {}
    },

    /**
     * The initial offset of the first page, required to get the second page
     * results.
     *
     * @attribute packageOffset
     * @type Integer
     */
    packageOffset: {
        value: null
    },
};

Y.extend(PackageListing, Y.Widget, {
    SELECT_ALL_LINK : ["",
                       "<a href='#' class='select-all-confirm'>Select all ",
                       "possible upgrades</a>"].join(""),

    /**
     * Constructor logic.
     *
     * @method initializer
     * @protected
     */
    initializer: function(config) {
        config = config || {};

        this._summaryRowsById = {};

        // Store the offsets to navigate the next page
        this.offsets = [this.get("packageOffset")];

        this.notes = new Y.landscape.notes.TableNotes({
            contentBox: this.get(CONTENT_BOX), animDuration: 0.01});

        // Default schema for the Computer Selection
        // DataSource.
        var sourceSchema = {
            metaFields: {"totalResults": TOTAL_RECORDS},
            resultListLocator: "records",
            resultFields:[{key: "name"},
                          {key: "packages"},
                          {key: "upgrades"},
                          {key: "securityUpgrades"},
                          {key: "relations"}]
        };

        var cacheSize = this.get(CACHE_SIZE);

        // If a recordSource wasn't provided, use the default one.
        var recordSource = config.recordSource;
        if (Y.Lang.isUndefined(recordSource)) {
            recordSource = new Y.DataSource.IO({source: ATTR_AVAILABLE});
        }

        // The cached entries are the entire data source response
        // for a page.
        recordSource.plug(
            {fn: Y.Plugin.DataSourceCache, cfg: {max: cacheSize}});
        recordSource.plug(
            {fn: Y.Plugin.DataSourceJSONSchema, cfg: {schema: sourceSchema}});

        recordSource.on("data", function(result) {
            var data = Y.JSON.parse(result.data.responseText);
            this.offsets.push(data.packageOffset);
            this.packageIndex.setData(data.packages);

            var results = this.packageIndex.getPackageData();
            result.data = {
                totalRecords: this.get(TOTAL_RECORDS),
                records: results};
        }, this);

        var sourceArguments = this.get("sourceArguments");
        var cacheKey = "?size=" + this.get(PAGE_SIZE);
        if (sourceArguments.filter) {
            cacheKey += "&filter=" + sourceArguments.filter;
        }
        if (sourceArguments.q) {
            cacheKey += "&q=" + sourceArguments.q;
        }
        if (sourceArguments.usn) {
            cacheKey += "&usn=" + sourceArguments.usn;
        }

        var pagerConfig = {contentBox: this.get(CONTENT_BOX),
                           renderFirstPage: true,
                           rowRenderer: Y.bind(this.rowRender, this),
                           pageSize: this.get(PAGE_SIZE),
                           recordSource: recordSource,
                           offsetCalculation: Y.bind(this.offsetCalculation, this),
                           recordSourceArguments: sourceArguments};

        this.packageIndex = new PackageIndex();
        this.packageIndex.update(this.get(PACKAGE_DATA));

        var results = this.packageIndex.getPackageData();
        recordSource.cache.add(
            cacheKey, {
                meta: {totalResults: this.get(TOTAL_RECORDS)},
                results: results});
        pagerConfig[TOTAL_RECORDS] = this.get(TOTAL_RECORDS);

        // Pager
        var pager = new Y.landscape.listing.Pager(pagerConfig);

        // Pagination
        var paginationBox = this.get(PAGINATION_BOX);
        if (paginationBox) {
            pager.plug(Y.landscape.listing.PaginationPlugin,
                {paginationBox: paginationBox, hideLast: true});
        }

        this.set(PAGER, pager);

        this.selectAll = false;
        this.hideAllSelected = false;

        /**
         * Fired when changes are marked and ready to be applied.
         *
         * @event allow_changes
         */
        this.publish("allow_changes");

        /**
         * Fired when changes aren't marked.
         *
         * @event deny_changes
         */
        this.publish("deny_changes");
    },

    /**
     * Return the offset needed for the given page. It basically returns the
     * offset given by the server so that we have a consistent navigation. This
     * is required because a page with 25 packages can match a query with more
     * than 25 results.
     *
     * @method offsetCalculation
     * @public
     */
    offsetCalculation: function(currentPage, pageSize) {
        if (currentPage == 1) {
            return 0;
        } else {
            return this.offsets[currentPage - 2];
        }
    },

    /**
     * Render rows in the package listing that represent packages and computers
     * with those packages.
     *
     * @method renderUI
     * @protected
     */
    renderUI: function() {
        this.get(PAGER).render();
    },

    /**
     * Register event handlers for controls that implement package selection
     * behaviour.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        var getSummaryRow = Y.bind(function(e) {
            var summaryNode = e.currentTarget.ancestor(
                DOT + PACKAGE_SUMMARY_ROW_CLASS);
            var nodeId = summaryNode.getAttribute(ID);
            return this._summaryRowsById[nodeId];
        }, this);

        this.get(CONTENT_BOX).delegate("click", function(e) {
            getSummaryRow(e).toggle();
        }, DOT + PACKAGE_SUMMARY_ROW_TOGGLE, this);

        this.get(CONTENT_BOX).delegate("click", function(e) {
            getSummaryRow(e).markForInstallation();
            this.deselectAll();
        }, DOT + MARK_FOR_INSTALL, this);

        this.get(CONTENT_BOX).delegate("click", function(e) {
            getSummaryRow(e).markForRemoval();
            this.deselectAll();
        }, DOT + MARK_FOR_REMOVAL, this);

        this.get(CONTENT_BOX).delegate("click", function(e) {
            getSummaryRow(e).markForUpgrade();
            this.deselectAll();
        }, DOT + MARK_FOR_UPGRADE, this);

        // Wire up the 'Select None' buttons.
        var selectNoneControls = this.get(SELECT_NONE_CONTROLS);
        var selectNone = function(evt) {
            evt.preventDefault();

            this._removeAllInputs();

            var summaryRows = this.get(SUMMARY_ROWS);
            Y.Array.each(summaryRows, function(summaryRow) {
                summaryRow.unmark();
            });

            this.deselectAll();
            this.syncUI();
        };

        Y.Array.each(selectNoneControls, function (selectNoneControl) {
            var control = Y.one(selectNoneControl);
            if (control) {
                control.on("click", selectNone, this);
            }
        }, this);

        // Wire up the 'Select Upgrades' buttons.
        var selectUpgradesControls = this.get(SELECT_UPGRADES_CONTROLS);
        var selectUpgrades = function(evt) {
            evt.preventDefault();
            if (this.selectAll) {
                return;
            }

            this._removeAllInputs();

            var count = 0;
            var summaryRows = this.get(SUMMARY_ROWS);
            Y.Array.each(summaryRows, function(summaryRow) {
                if (summaryRow.markUpgrades()) {
                    count += 1;
                }
            });

            var pager = this.get(PAGER);
            var totalPages = pager.get("totalPages");
            var currentPage = pager.get("currentPage");
            var pageFull = pager.getRowCount() == this.get(PAGE_SIZE);
            if (currentPage > 1 || (currentPage == 1 && pageFull && totalPages != 1)) {
                this.notes.hideNotes();
                var message;
                if (!count) {
                    message = "No upgrades on this page. ";
                } else if (count == 1) {
                    message = "1 upgrade selected. ";
                } else {
                    message = count + " upgrades selected. ";
                }
                this.notes.addNote("info", message + this.SELECT_ALL_LINK);

                // Bind the note link to our confirmation callback.
                var link = this.notes.get(CONTENT_BOX)
                    .one("a.select-all-confirm");
                link.on("click", Y.bind(function(evt) {
                    evt.preventDefault();
                    this.selectionConfirmAll();
                }, this));
            }
            this.syncUI();
        };

        Y.Array.each(selectUpgradesControls, function(selectUpgradesControl) {
            var control = Y.one(selectUpgradesControl);
            if (control) {
                control.on("click", selectUpgrades, this);
            }
        }, this);

        var pager = this.get(PAGER);
        // Clean up before the page changes
        pager.on("pager:" + EVT_BEFORE_PAGE_CHANGE, function() {
            // Hide the notes before we switch pages.
            var hideAllSelected = this.hideAllSelected;
            this.notes.hideNotes();
            this.hideAllSelected = hideAllSelected;
            var summaryRows = this.get(SUMMARY_ROWS);
            Y.Array.each(summaryRows, function(summaryRow) {
                if (summaryRow.isMarked()) {
                    summaryRow.unmark();
                }
                if (summaryRow.get(EXPANDED)) {
                    summaryRow.toggle();
                }
            });

            // Removes the extra rows which might have been created
            var contentBox = this.get(CONTENT_BOX);
            contentBox.all("tr.ls-pkg-tree-version-row").remove();
            contentBox.all("tr.version-gutter").remove();
            this.set(SUMMARY_ROWS, []);
        }, this);
        pager.on("pager:after-page-change", function() {
            if (this.selectAll) {
                var summaryRows = this.get(SUMMARY_ROWS);
                Y.Array.each(summaryRows, function(summaryRow) {
                    summaryRow.markUpgrades();
                });
                if (!this.hideAllSelected) {
                    this.notes.addNote("info", "All upgrades selected");
                }
                this.notes.once("table_notes:close_event", function() {
                    this.hideAllSelected = true;
                }, this);
            }
        }, this);
    },

    /**
     * Synchronize the user interface with the widget.
     *
     * @method syncUI
     * @protected
     */
    syncUI: function() {
        this._fireChangesEvent();
        this._removePackageLoadingStatus();
    },

    /**
     * Render a single row on behalf of the pager widget.
     *
     * @method rowRender
     * @protected
     */
    rowRender: function(row, record, index) {
        var summaryRow = new PackageSummaryRow({
            packageInfo: record, widget: this});
        row.all("td").remove();
        summaryRow.renderUI(row);
        this._summaryRowsById[summaryRow.get(ID)] = summaryRow;
        this.get(SUMMARY_ROWS).push(summaryRow);
    },

    /**
     * Confirm the selection of all computers, hiding notes if they
     * are visible.
     *
     * @method selectionConfirmAll
     * @public
     */
    selectionConfirmAll: function() {
        this.notes.hideNotes();
        this.selectAll = true;
        this.hideAllSelected = false;
        var form = this.get(CONTENT_BOX);
        var input = Y.Node.create("<input type='hidden' />");
        input.set("id", "changes-select-all-upgrades");
        input.set("name", "changes-select-all-upgrades");
        var sourceArguments = this.get("sourceArguments");
        inputValue = "filter=" + sourceArguments.filter;
        inputValue += "&q=" + sourceArguments.q;
        inputValue += "&usn=" + sourceArguments.usn;
        input.set("value", inputValue);
        form.appendChild(input);
        this.notes.addNote("info", "All upgrades selected");
        this.notes.once("table_notes:close_event", function() {
            this.hideAllSelected = true;
        }, this);
    },

    /**
     * Unset the selectAll flag.
     *
     * @method deselectAll
     * @public
     */
    deselectAll: function() {
        this.notes.hideNotes();
        if (this.selectAll) {
            this.selectAll = false;
            this.get(CONTENT_BOX).one("#changes-select-all-upgrades").remove();
        }
    },

    /**
     * Remove all inputs for package changes.
     *
     * @method _removeAllInputs
     * @private
     */
    _removeAllInputs: function() {
        this.get(CONTENT_BOX).all("#hidden-inputs > input").remove();
    },

    /**
     * Returns true if at least one row is marked and false if no rows are
     * marked.
     *
     * @method isMarked
     * @public
     */
    isMarked: function() {
        var summaryRows = this.get(SUMMARY_ROWS);
        for (var i in summaryRows) {
            if (summaryRows[i].isMarked()) {
                return true;
            }
        }
        return false;
    },

    /**
     * Fire the 'allow_changes' event if rows are marked, or the 'deny_changes'
     * event if no row are marked.
     *
     * @method _fireChangesEvent
     * @private
     */
    _fireChangesEvent: function() {
        this.fire(this.isMarked() ? "allow_changes" : "deny_changes");
    },

    /**
     * Remove the package loading status element, if its defined in the
     * configuration and the related node is present.
     *
     * @method _removeLoadingStatus
     * @private
     */
    _removePackageLoadingStatus: function() {
        var packageLoadingStatus = this.get(PACKAGE_LOADING_STATUS);
        if (packageLoadingStatus) {
            var node = Y.one(packageLoadingStatus);
            if (node) {
                var parentNode = node.get(PARENT_NODE);
                if (parentNode && parentNode.get("children").size() == 1) {
                    parentNode.remove();
                } else {
                    node.remove();
                }

            }
        }
    },

    getPackageIconPath: function(kind, lockState, marked) {
        if (kind == AVAILABLE) {
           return AVAILABLE_ICON_PATH;
        }
        if (kind == PARTIAL) {
            return PARTIAL_ICON_PATH;
        }
        return INSTALLED_ICON_PATH;
    },

    /**
     * Indicate that the package icon can be used to toggle package holds if
     * the package is installed.
     */
    setPackageIconHoldAction: function(packageIconImage, kind, held) {
        var isInstalled = (kind == INSTALLED ||
                           kind == UPGRADE ||
                           kind == DOWNGRADE);
        if (isInstalled) {
            if (held) {
                packageIconImage.addClass(MARK_FOR_UNHOLD);
                packageIconImage.set("title", "Mark for unhold");
            } else {
                packageIconImage.addClass(MARK_FOR_HOLD);
                packageIconImage.set("title", "Mark for hold");
            }
        }
    },

    /**
     * Update the package icon with the given configuration.
     *
     * @method updatePackageIcon
     * @public
     */
    updatePackageIcon: function(packageIcon, configName, mark, lockState) {
        var iconSource = this.get(configName);
        if (packageIcon.get(IMAGE) != iconSource) {
            packageIcon.set(IMAGE, iconSource);
        }
        var lockIconNodes = packageIcon.get(NODE).all(".ls-pkg-tree-lock-icon");
        if (mark == ATTR_HOLD || lockState == LOCKED || lockState == PARTIAL) {
            if (mark == ATTR_UNHOLD) {
                lockIconNodes.remove();
            } else if (lockIconNodes.size() === 0) {
                var lockedPackageIconImage = new Y.landscape.ui.Sprite({
                    "cssClasses": ["ls-pkg-tree-lock-icon"],
                    "image": this.get(LOCKED_ICON_PATH)});
                lockedPackageIconImage.render(packageIcon.get(NODE));
            }
        } else if (lockIconNodes.size() > 0) {
            lockIconNodes.remove();
        }
    }
});

Y.landscape.packages.PackageListing = PackageListing;

/**
 * A Landscape.PackageSummaryRow represents a top-level row in the package
 * listing.  It displays the name and description of package and can be
 * expanded to show specific data about the package installed on selected
 * computers.
 *
 * @class PackageSummaryRow
 * @extends Base
 * @constructor
 */
var PackageSummaryRow = function(config) {
    PackageSummaryRow.superclass.constructor.apply(this, arguments);
};

PackageSummaryRow.NAME = PACKAGE_SUMMARY_ROW;

PackageSummaryRow.ATTRS = {
    /**
     * An identifier for this row, based on the package name.
     *
     * @attribute id
     * @type Object
     */
    id: {
        value: null
    },

    /**
     * The package info for this summary row.  It includes all versions of the
     * package with the matching name, along with a set of relation objects
     * that represent package version/computer relationships.
     *
     * @attribute packageInfo
     * @type Object
     */
    packageInfo: {
        value: undefined
    },

    /**
     * The package icon for this summary row.
     *
     * @attribute packageIcon
     * @type Image
     */
    packageIcon: {
        value: undefined
    },

    /**
     * The toggle icon for this summary row.
     *
     * @attribute toggleIcon
     * @type Image
     */
    toggleIcon: {
        value: undefined
    },

    /**
     * The package listing widget instance parent of this summary row.
     *
     * @attribute widget
     * @type PackageListing
     */
    widget: {
        value: undefined
    },

    /**
     * The Y.Node representing the TR of this summary row in the package
     * listing.
     *
     * @attribute row
     * @type Node
     */
    row: {
        value: undefined
    },

    /**
     * The Y.Node representing the gutter row for this summary row.  The gutter
     * row is an empty row rendered as the final element, after the version
     * filter and all version rows.  It is only shown when the summary row is
     * expanded and is meant to be a visual indication that there is a
     * separator between the version rows for the expanded summary row and the
     * next summary row in the package listing.  Default is undefined.
     *
     * @attribute gutterRow
     * @type Node
     */
    gutterRow: {
        value: undefined
    },

    /**
     * True if the version filter and version rows are expanded and visible.
     * Default is false.
     *
     * @attribute expanded
     * @type Boolean
     */
    expanded: {
        value: false
    },

    /**
     * Value stores the last mark chosen by the users.  This is used when
     * determining how toggle behaviour should work with the 'Mark for
     * <action>' buttons.  Default is undefined.
     *
     * @attribute marked
     * @type String
     */
    marked: {
        value: UNMARKED
    },

    /**
     * A value that represents the kind of package version/computer relations
     * present for this summary row's package.  If all versions of the package
     * are available, and none installed, on all computers, the value will be
     * AVAILABLE.  If versions of the package are installed on all computers
     * the value will be INSTALLED.  If versions are both installed and
     * available the value will be PARTIAL.
     *
     * @attribute kind
     * @type Int
     */
    kind: {
        value: undefined
    },

    /**
     * A value that represents the lock state of package
     * version/computer relations present for this summary row's
     * package.  If all versions of the package are locked, on all
     * computers, the value will be LOCKED.  If versions of the
     * package are unlocked on all computers the value will be
     * UNLOCKED.  If only some versions are locked in some computers
     * the value will be PARTIAL.
     *
     * @attribute lockState
     * @type Int
     */
    lockState: {
        value: undefined
    },

    /**
     * Flag is true if any of the version row children represent available
     * packages.
     *
     * @attribute hasAvailable
     * @type Boolean
     */
    hasAvailable: {
        value: false
    },

    /**
     * Flag is true if any of the version row children represent upgrade
     * packages.
     *
     * @attribute hasUpgrades
     * @type Boolean
     */
    hasUpgrades: {
        value: false
    },

    /**
     * Flag is true if any of the version row children represent installed
     * packages.
     *
     * @attribute hasInstalled
     * @type Boolean
     */
    hasInstalled: {
        value: false
    },

    /**
     * Flag is true if any of the version row children represent installed
     * packages that can be removed.
     *
     * @attribute hasRemovable
     * @type Boolean
     */
    hasRemovable: {
        value: false
    },

    /**
     * The package version row object, if one is present, otherwise undefined.
     *
     * @attribute versionFilterRow
     * @type PackageVersionFilterRow
     */
    versionFilterRow: {
        valueFn: function() {
            // A version filter row is lazily created, but only if there is
            // more than one version of the package available for any
            // particular computer.
            if (this._versionFilterRow === undefined) {
                var packageInfo = this.get(PACKAGE_INFO);
                var hasMultipleVersions = false;
                var index = {};
                for (var i in packageInfo.relations) {
                    var relation = packageInfo.relations[i];
                    for (var j in relation.computers) {
                        var computer = relation.computers[j];
                        if (! Y.Lang.isUndefined(index[computer])) {
                            hasMultipleVersions = true;
                            break;
                        } else {
                            index[computer] = true;
                        }
                    }
                    if (hasMultipleVersions) {
                        break;
                    }
                }

                if (hasMultipleVersions) {
                    this._versionFilterRow = new PackageVersionFilterRow({
                        summaryRow: this
                    });
                } else {
                    this._versionFilterRow = null;
                }
            }

            if (this._versionFilterRow === null) {
                return undefined;
            } else {
                return this._versionFilterRow;
            }
        }
    },

    /**
     * The list of PackageVersionRow objects managed by this summary row.
     *
     * @attribute versionRows
     * @type Array
     */
    versionRows: {
        valueFn: function() {
            if (this._versionRows === undefined) {
                this._versionRows = [];
            }
            return this._versionRows;
        }
    },

    /**
     * Flag is true is version rows have been rendered.  Default is false.
     *
     * @attribute versionRowsRendered
     * @type Boolean
     */
    versionRowsRendered: {
        value: false
    },
};

Y.extend(PackageSummaryRow, Y.Base, {
    /**
     * Setup the summary row using data provided in the configuration.  As
     * little as possible is done here, preferring lazy evaluation where
     * possible.
     *
     * @method initializer
     */
    initializer: function(config) {
        // Build an index that will allow us to figure out what kind of row
        // this is.
        var packageInfo = this.get(PACKAGE_INFO);
        var index = {all: [], allIndex: {},
                     installed: [], installedIndex: {},
                     held: [], heldIndex: {}};
        var hasAvailable = false;
        var hasInstalled = false;
        var hasUpgrades = false;
        var hasRemovable = false;

        Y.Array.each(packageInfo.relations, function(relation) {
            var kind = relation.kind;
            var held = relation.held;
            var installed = relation.installed;

            Y.Array.each(relation.computers, function(computer) {
                if (!index.allIndex.hasOwnProperty(computer)) {
                    index.all.push(computer);
                    index.allIndex[computer] = true;
                }
                if ((kind == INSTALLED || kind == UPGRADE || kind == DOWNGRADE) &&
                    !index.installedIndex.hasOwnProperty(computer)) {
                    index.installed.push(computer);
                    index.installedIndex[computer] = true;
                }
                if (held && !index.heldIndex.hasOwnProperty(computer)) {
                    index.held.push(computer);
                    index.heldIndex[computer] = true;
                }
            });

            if (kind == AVAILABLE) {
                hasAvailable = true;
            } else if (kind == INSTALLED || kind == DOWNGRADE) {
                hasInstalled = true;
            } else if (kind == UPGRADE && !held) {
                hasUpgrades = true;
            }
            if (installed && !held) {
                hasRemovable = true;
            }
        });

        this.set(ID, getCN("package-summary", packageInfo.name));
        this.set(HAS_AVAILABLE, hasAvailable);
        this.set(HAS_INSTALLED, hasInstalled);
        this.set(HAS_UPGRADES, hasUpgrades);
        this.set(HAS_REMOVABLE, hasRemovable);

        if (index.installed.length == index.all.length) {
            this.set(KIND, INSTALLED);
        } else if (index.installed.length > 0) {
            // Packages are both installed and available.
            this.set(KIND, PARTIAL);
        } else {
            this.set(KIND, AVAILABLE);
        }

        if (index.held.length == index.all.length) {
            this.set(LOCK_STATE, LOCKED);
        } else if (index.held.length > 0) {
            // Packages are both held and unheld.
            this.set(LOCK_STATE, PARTIAL);
        } else {
            this.set(LOCK_STATE, UNLOCKED);
        }

        // Create a package version row for each package relation defined in
        // the package information for this package.
        var versionRows = this.get(VERSION_ROWS);
        Y.Array.each(packageInfo.relations, function(relation) {
            versionRows.push(new PackageVersionRow({
                summaryRow: this, relation: relation}));
        }, this);

        /**
         * An event handler that provides toggling of row marking.
         *
         * @param mark: The marker status to toggle.
         * @param markedCallback: Name of a function taking no arguments that
         *     returns true if the row is marked.
         * @param markCallback: Name of a function taking no arguments that marks
         *     the row.
         * @param allowMarkCallable: A function taking a PackageVersionRow and
         *     returning true if the row should be marked.
         */
        var eventHandler = function(mark, markedCallback,
                                    markCallback, allowMarkCallable) {
            var widget = this.get(WIDGET);
            var alreadyMarked = 0;
            var versionRows = this.get(VERSION_ROWS);
            if (this.isMarked()) {
                Y.Array.each(versionRows, function(versionRow) {
                    if (versionRow[markedCallback]()) {
                        versionRow.unmark();
                        alreadyMarked += 1;
                    }
                });
                this._unmark();
            }
            if (!alreadyMarked) {
                var currentVersion = LATEST_VERSION;
                var versionFilter = this.get(VERSION_FILTER_ROW);
                if (versionFilter !== undefined) {
                    currentVersion = versionFilter.getCurrentVersion();
                    if (currentVersion === undefined) {
                        currentVersion = LATEST_VERSION;
                    }
                }
                Y.Array.each(versionRows, function(versionRow) {
                    if (allowMarkCallable(versionRow, currentVersion)) {
                        versionRow[markCallback]();
                    }
                });
            }
            this.syncUI();
            widget.syncUI();
        };

        // Bind a handler for the 'Mark for installation' button.
        var allowMarkForAvailable = function(versionRow, currentVersion) {
            var relation = versionRow.get(RELATION);
            if (relation.kind == AVAILABLE) {
                if (currentVersion == LATEST_VERSION) {
                    return relation.latest;
                } else {
                    var packageData = currentVersion;
                    return (relation.kind == AVAILABLE &&
                            relation.available.id == packageData.id);
                }
            }
            return null;
        };

        this.markForInstallation = Y.bind(
            eventHandler, this, ATTR_INSTALL,
            "markedForInstallation", "markForInstallation",
            allowMarkForAvailable);

        // Bind a handler for the 'Mark for upgrade' button.
        var allowMarkForUpgrade = function(versionRow, currentVersion) {
            var relation = versionRow.get(RELATION);
            if (relation.kind == UPGRADE) {
                if (currentVersion == LATEST_VERSION) {
                    return relation.latest;
                } else {
                    var packageData = currentVersion;
                    return (relation.installed.id == packageData.id ||
                            relation.upgrade.id == packageData.id);
                }
            }
            return null;
        };

        this.markForUpgrade = Y.bind(
            eventHandler, this, ATTR_UPGRADE,
            "markedForUpgrade", "markForUpgrade",
            allowMarkForUpgrade);

        // Bind a handler for the 'Mark for removal' button.
        var allowMarkForRemoval = function(versionRow, currentVersion) {
            var relation = versionRow.get(RELATION);
            if (relation.kind == INSTALLED ||
                relation.kind == UPGRADE ||
                relation.kind == DOWNGRADE) {
                if (currentVersion == LATEST_VERSION) {
                    return relation.latest;
                } else {
                    var packageData = currentVersion;
                    if (relation.kind == INSTALLED) {
                        return relation.installed.id == packageData.id;
                    } else if (relation.kind == UPGRADE) {
                        return (relation.installed.id == packageData.id ||
                                relation.upgrade.id == packageData.id);
                    } else if (relation.kind == DOWNGRADE) {
                        return (relation.installed.id == packageData.id ||
                                relation.downgrade.id == packageData.id);
                    }
                }
            }
            return null;
        };

        this.markForRemoval = Y.bind(
            eventHandler, this, ATTR_REMOVE,
            "markedForRemoval", "markForRemoval",
            allowMarkForRemoval);
    },

    /**
     * Get the path to the package icon to use for this row.
     **/
    _getPackageIconPath: function() {
        var kind = this.get(KIND);
        var lockState = this.get(LOCK_STATE);
        var marked = this.get(MARKED);

        return this.get(WIDGET).getPackageIconPath(kind, lockState, marked);
    },

    /**
     * Render this summary row and return the resulting node to be added to
     * the package listing.
     *
     * @method renderUI
     */
    renderUI: function(row) {
        var cell;
        var parentWidget = this.get(WIDGET);
        var packageInfo = this.get(PACKAGE_INFO);

        this.set(ROW, row);
        row.set("id", this.get(ID));
        row.addClass(PACKAGE_SUMMARY_ROW_CLASS);

        // Build disclosure triangle column.
        cell = Y.Node.create("<td class='collapse-column-right'>");
        row.appendChild(cell);

        var toggleIcon = new Y.landscape.ui.Sprite({
            "cssClasses": [CLICKABLE, PACKAGE_SUMMARY_ROW_TOGGLE, EXPANDABLE],
            "imageId": getCN("ls-pkg-tree-toggle-version-visibility",
                             packageInfo.name),
            "image": parentWidget.get(COLLAPSE_ICON_PATH)});
        this.set(TOGGLE_ICON, toggleIcon);
        toggleIcon.render(cell);

        // Build package action buttons column.
        cell = Y.Node.create(
            "<td class='expand collapse-column-left collapse-column-right'/>");
        row.appendChild(cell);

        var buttonContainer = Y.Node.create(
            "<div class='package-action-buttons'/>");
        cell.appendChild(buttonContainer);

        var installButton = new Y.landscape.ui.Sprite({
            "cssClasses": [MARK_FOR_INSTALL, ACTION_BUTTON],
            "imageId": getCN(MARK_FOR_INSTALL, packageInfo.name),
            "title": "Mark all installable packages"});

        if (this.get(HAS_AVAILABLE)) {
            installButton.set(
                IMAGE, parentWidget.get(MARK_FOR_INSTALL_ICON_PATH));
        } else {
            installButton.addClass("disabled-action-button");
            installButton.set(
                IMAGE, parentWidget.get(MARK_FOR_INSTALL_DISABLED_ICON_PATH));
        }
        installButton.render(buttonContainer);

        var upgradeButton = new Y.landscape.ui.Sprite({
            "cssClasses": [MARK_FOR_UPGRADE, ACTION_BUTTON],
            "imageId": getCN(MARK_FOR_UPGRADE, packageInfo.name),
            "title": "Mark all upgradable packages"});

        if (this.get(HAS_UPGRADES)) {
            upgradeButton.set(
                IMAGE, parentWidget.get(MARK_FOR_UPGRADE_ICON_PATH));
        } else {
            upgradeButton.addClass("disabled-action-button");
            upgradeButton.set(
                IMAGE, parentWidget.get(MARK_FOR_UPGRADE_DISABLED_ICON_PATH));
        }
        upgradeButton.render(buttonContainer);

        var removeButton = new Y.landscape.ui.Sprite({
            "cssClasses": [MARK_FOR_REMOVAL, ACTION_BUTTON],
            "imageId": getCN(MARK_FOR_REMOVAL, packageInfo.name),
            "title": "Mark all removable packages"});

        if (this.get(HAS_REMOVABLE)) {
            removeButton.set(
                IMAGE, parentWidget.get(MARK_FOR_REMOVAL_ICON_PATH));
        } else {
            removeButton.addClass("disabled-action-button");
            removeButton.set(
                IMAGE, parentWidget.get(MARK_FOR_REMOVAL_DISABLED_ICON_PATH));
        }
        removeButton.render(buttonContainer);

        // Build package name icon column.
        cell = Y.Node.create(
            "<td class='collapse-column-left collapse-column-right'/>");
        row.appendChild(cell);

        var packageIconImage = new Y.landscape.ui.Sprite({
            "cssClasses": ["ls-pkg-tree-action-icon"]});
        this.set(PACKAGE_ICON, packageIconImage);

        var packageIconPath = this._getPackageIconPath();
        packageIconImage.set(
            IMAGE, parentWidget.get(packageIconPath));
        packageIconImage.render(cell);

        if (this.get(LOCK_STATE) == LOCKED ||
            this.get(LOCK_STATE) == PARTIAL
            ) {
            this._setActionIconPath(packageIconPath);
        }

        // Build package name column.
        cell = Y.Node.create("<td class='inner-cell collapse-column-left' />");
        row.appendChild(cell);

        var packageName = Y.Node.create("<nobr />");
        packageName.set(TEXT, packageInfo.name);
        cell.appendChild(packageName);

        // Build package summary column.
        cell = Y.Node.create(
            "<td colspan='2' class='collapse-column-right'/>");
        row.appendChild(cell);

        cell.set(TEXT, packageInfo.packages[0].summary);

        var packageStatusCell = Y.Node.create(
            "<td class='collapse-column-left' />");
        row.appendChild(packageStatusCell);
        if (packageInfo.upgrades) {
            var statusMessage = Y.Node.create("<nobr />");
            packageStatusCell.appendChild(statusMessage);
            var qualifier = "";
            if (packageInfo.securityUpgrades > 0) {
                qualifier = " security";
            }
            if (packageInfo.upgrades > 1) {
                statusMessage.appendChild(
                    Y.Node.create(packageInfo.upgrades + qualifier + " upgrades"));
            } else {
                statusMessage.appendChild(Y.Node.create("1" + qualifier + " upgrade"));
            }
        }

        return row;
    },

    /**
     * Synchronize the user interface state with the widget state.
     *
     * @method syncUI
     */
    syncUI: function() {
        if (this.get(EXPANDED)) {
            this._syncVersionFilterRow();
            this._syncVersionRowVisibility();
        }
        this._syncSummaryRow();
    },

    /**
     * Synchronize the version row filter.
     *
     * @method _syncVersionFilterRow
     * @private
     */
    _syncVersionFilterRow: function() {
        // This method is a no-op when no version filter is present.
        var versionFilterRow = this.get(VERSION_FILTER_ROW);
        if (versionFilterRow === undefined) {
            return;
        }
        versionFilterRow.syncUI();
        versionFilterRow.get(SUMMARY_ROW).get(WIDGET).syncUI();
    },

    /**
     * Synchronize the visibility of version rows based on the state of the
     * version row filter, if one is present.
     *
     * @method _syncVersionRowVisibility
     * @private
     */
    _syncVersionRowVisibility: function() {
        // This method is a no-op when no version filter is present.
        var versionFilterRow = this.get(VERSION_FILTER_ROW);
        if (versionFilterRow === undefined) {
            return;
        }

        // The current version is either the LATEST_VERSION symbol, used to
        // represent the newest version of the package for each computer, or a
        // specific package data object to match version rows with.
        var currentVersion = versionFilterRow.getCurrentVersion();
        var match;
        if (currentVersion == LATEST_VERSION) {
            match = function(relation) {
                return relation.latest;
            };
        } else {
            var packageData = currentVersion;
            match = function(relation) {
                if (relation.kind == UPGRADE) {
                    return (relation.upgrade.id == packageData.id ||
                            relation.installed.id == packageData.id);
                } else if (relation.kind == DOWNGRADE) {
                    return (relation.downgrade.id == packageData.id ||
                            relation.installed.id == packageData.id);
                } else if (relation.kind == INSTALLED) {
                    return relation.installed.id == packageData.id;
                } else {
                    return relation.available.id == packageData.id;
                }
            };
        }

        // A tricky thing with a knife happens here.  Version rows are sorted
        // in groups, newer packages come before older ones, and computers in
        // each group are sorted alphabetically.  An index is used to keep
        // track of the computers that have already been displayed.  Anytime a
        // row has a computer that has already been rendered, we know that we
        // have an older relation for the package, so we suppress it.  This
        // avoids a situation where, when selecting an installed package
        // version with many upgrades or downgrades, computers are displayed
        // more than once.

        var versionRows = this.get(VERSION_ROWS);
        var computerIndex = {};

        Y.Array.each(versionRows, function(versionRow) {
            var relation = versionRow.get(RELATION);

            if (match(relation)) {
                var relationComputers = [];
                Y.Array.each(relation.computers, function(computer) {
                    if (!computerIndex.hasOwnProperty(computer)) {
                        computerIndex[computer] = true;
                        relationComputers.push(computer);
                    }
                });
                if (relationComputers.length > 0) {
                    versionRow.show();
                }
            } else {
                versionRow.hide();
                if (versionRow.get(MARKED) != UNMARKED) {
                    versionRow.unmark();
                }
            }
        });
    },

    /**
     * Synchronize the summary row, updating its mark state based on the state
     * of version rows.
     *
     * @method _syncSummaryRow
     * @private
     */
    _syncSummaryRow: function() {
        var install = 0;
        var upgrade = 0;
        var downgrade = 0;
        var remove = 0;
        var hold = 0;
        var unhold = 0;
        var versionRows = this.get(VERSION_ROWS);

        var i = versionRows.length;
        while (i--) {
            var mark = versionRows[i].get(MARKED);
            if (mark == ATTR_INSTALL) {
                install += 1;
            } else if (mark == ATTR_UPGRADE) {
                upgrade += 1;
            } else if (mark == ATTR_DOWNGRADE) {
                downgrade += 1;
            } else if (mark == ATTR_REMOVE) {
                remove += 1;
            } else if (mark == ATTR_HOLD) {
                hold += 1;
            } else if (mark == ATTR_UNHOLD) {
                unhold += 1;
            }
        }

        this._unmark();

        if ((install || upgrade || downgrade) && remove) {
            this._markForPartial();
        } else if (install) {
            this._markForInstallation();
        } else if (upgrade) {
            this._markForUpgrade();
        } else if (downgrade) {
            this._markForDowngrade();
        } else if (remove) {
            this._markForRemoval();
        } else if (hold) {
            this._markForHold();
        } else if (unhold) {
            this._markForUnhold();
        }
    },

    _getUpgradeCheckbox: function() {
        var row = this.get(ROW);
        return row.one(".package-action-buttons input[type='checkbox']");
    },

    /**
     * Update CSS and icon paths to mark this row for installation.  If the row
     * is already marked for installation, unmark it.
     *
     * @method _markForInstallation
     * @private
     */
    _markForInstallation: function() {
        var parentWidget = this.get(WIDGET);
        var row = this.get(ROW);
        if (!row.hasClass("ls-pkg-tree-install")) {
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-install");
            this._setActionIconPath(INSTALLING_ICON_PATH);
            this.set(MARKED, ATTR_INSTALL);
        }
    },

    /**
     * Update CSS and icon paths to mark this row for upgrade.  If the row is
     * already marked for upgrade, unmark it.
     *
     * @method _markForUpgrade
     * @private
     */
    _markForUpgrade: function() {
        var parentWidget = this.get(WIDGET);
        var row = this.get(ROW);
        if (!row.hasClass("ls-pkg-tree-upgrade")) {
            row.replaceClass("ls-pkg-tree-remove", "ls-pkg-tree-upgrade");
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-upgrade");

            this._setActionIconPath(UPGRADING_ICON_PATH);
            this.set(MARKED, ATTR_UPGRADE);
        }
    },

    /**
     * Update CSS and icon paths to mark this row for downgrade.  If the row is
     * already marked for downgrade, unmark it.
     *
     * @method _markForDowngrade
     * @private
     */
    _markForDowngrade: function() {
        var parentWidget = this.get(WIDGET);
        var row = this.get(ROW);
        if (!row.hasClass("ls-pkg-tree-downgrade")) {
            row.replaceClass("ls-pkg-tree-remove", "ls-pkg-tree-downgrade");
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-downgrade");
            this._setActionIconPath(DOWNGRADING_ICON_PATH);
            this.set(MARKED, ATTR_DOWNGRADE);
        }
    },

    /**
     * Update CSS and icon paths to mark this row for removal.  If the row
     * is already marked for installation, unmark it.
     *
     * @method _markForRemoval
     * @private
     */
    _markForRemoval: function() {
        var parentWidget = this.get(WIDGET);
        var row = this.get(ROW);
        if (!row.hasClass("ls-pkg-tree-remove")) {
            row.replaceClass("ls-pkg-tree-upgrade", "ls-pkg-tree-remove");
            row.replaceClass("ls-pkg-tree-downgrade", "ls-pkg-tree-remove");
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-remove");
            this._setActionIconPath(REMOVING_ICON_PATH);
            this.set(MARKED, ATTR_REMOVE);
        }
    },

    /**
     * Update CSS and icon paths to mark this row as partial, as having version
     * rows marked for installation, upgrade or downgrade as well as removal.
     * If the row is already marked it's unmarked.
     *
     * @method _markForPartial
     * @private
     */
    _markForPartial: function() {
        var parentWidget = this.get(WIDGET);
        var row = this.get(ROW);
        if (!row.hasClass("ls-pkg-tree-partial")) {
            row.replaceClass("ls-pkg-tree-remove", "ls-pkg-tree-partial");
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-partial");
            this._setActionIconPath(PARTIALING_ICON_PATH);
            this.set(MARKED, "partial");
        }
    },

    /**
     * Mark the row to hold the package.
     *
     * @method markForHold
     */
    _markForHold: function() {
        var row = this.get(ROW);
        this.set(MARKED, ATTR_HOLD);
        row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-hold");
        var packageIcon = this.get(PACKAGE_ICON);
        this._setActionIconPath(this._getPackageIconPath());
    },

    /**
     * Mark the row to unhold the package.
     *
     * @method markForUnhold
     */
    _markForUnhold: function() {
        var row = this.get(ROW);
        this.set(MARKED, ATTR_UNHOLD);
        row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-unhold");
        this._setActionIconPath(this._getPackageIconPath());
    },

    /**
     * Update CSS and icon paths to mark this row for removal.  If the row
     * is already marked for installation, unmark it.
     *
     * @method _unmark
     * @private
     */
    _unmark: function() {
        var row = this.get(ROW);
        row.removeClass("ls-pkg-tree-install");
        row.removeClass("ls-pkg-tree-upgrade");
        row.removeClass("ls-pkg-tree-downgrade");
        row.removeClass("ls-pkg-tree-remove");
        row.removeClass("ls-pkg-tree-partial");
        row.removeClass("ls-pkg-tree-hold");
        row.removeClass("ls-pkg-tree-unhold");
        row.addClass("ls-pkg-tree-unmarked");

        this.set(MARKED, UNMARKED);
        this._setActionIconPath(this._getPackageIconPath());
    },

    /**
     * Unmark this row and all of its version rows.
     *
     * @method unmark
     */
    unmark: function() {
        var versionRows = this.get(VERSION_ROWS);
        Y.Array.each(versionRows, function(versionRow) {
            versionRow.unmark();
        });
        this._unmark();
        this.syncUI();
    },

    /**
     * Mark all upgrades.
     *
     * @method markUpgrades
     */
    markUpgrades: function() {
        var currentVersion = LATEST_VERSION;
        var versionFilter = this.get(VERSION_FILTER_ROW);
        if (versionFilter !== undefined) {
            currentVersion = versionFilter.getCurrentVersion();
            if (currentVersion === undefined) {
                currentVersion = LATEST_VERSION;
            }
        }

        var marked = false;
        var versionRows = this.get(VERSION_ROWS);
        for (var i in versionRows) {
            var versionRow = versionRows[i];
            var relation = versionRow.get(RELATION);
            var match = false;
            if (relation.kind == UPGRADE) {
                if (currentVersion == LATEST_VERSION) {
                    match = relation.latest;
                } else {
                    var packageData = currentVersion;
                    match = (relation.installed.id == packageData.id ||
                             relation.upgrade.id == packageData.id);
                }
            }
            if (match) {
                match = versionRow.markForUpgrade();
            }
            marked = marked || match;
        }
        this.syncUI();
        return marked;
    },

    /**
     * Returns true if this row is marked in any way.
     *
     * @method isMarked
     */
    isMarked: function() {
        return this.get(MARKED) != UNMARKED;
    },

    /**
     * Set the path of the action icon using the path retrieved from the
     * widget's configuration using the specified name.
     *
     * @method _setActionIconPath
     * @private
     */
    _setActionIconPath: function(configName) {
        this.get(WIDGET).updatePackageIcon(
            this.get(PACKAGE_ICON), configName, this.get(MARKED),
            this.get(LOCK_STATE));
    },

    /**
     * Toggle version row visibility.
     *
     * @method toggle
     * @public
     */
    toggle: function() {
        if (!this.get(EXPANDED)) {
            this._showVersionFilterAndRows();
        } else {
            this._hideVersionFilterAndRows();
        }
        this.syncUI();
    },

    /**
     * Show the version filter, version rows and the gutter row, rendering them
     * if necessary.
     *
     * @method _showVersionFilterAndRows
     * @private
     */
    _showVersionFilterAndRows: function() {
        var gutterRow;
        var packageInfo = this.get(PACKAGE_INFO);
        var versionFilterRow = this.get(VERSION_FILTER_ROW);
        var versionRows = this.get(VERSION_ROWS);
        if (!this.get(VERSION_ROWS_RENDERED)) {
            // Use a DocumentFragment to speed up rendering of version
            // rows.
            var fragment = Y.config.doc.createDocumentFragment();
            var table = Y.Node.create("<table />");
            var placeholder = Y.Node.create("<tr />");
            table.appendChild(placeholder);
            fragment.appendChild(Y.Node.getDOMNode(table));

            var appendAfterRow = placeholder;

            // Render version filter, if appropriate, and version rows and the
            // gutter row.
            if (versionFilterRow !== undefined) {
                appendAfterRow = versionFilterRow.renderUI(appendAfterRow);
            }
            Y.Array.each(versionRows, function(versionRow) {
                appendAfterRow = versionRow.renderUI(appendAfterRow);
            });

            gutterRow = Y.Node.create(
                "<tr class='version-gutter'>" +
                "<td/><td/><td/><td/><td/><td/><td/>" +
                "</tr>");
            insertNodeAfter(appendAfterRow, gutterRow);
            this.set(GUTTER_ROW, gutterRow);

            // Get rid of the placeholder node.
            placeholder.remove();

            // Create a new fragment from all the other rows that are
            // left on the table and insert it after the summaryRow.
            var rendered = table.all("tr").toFrag();

            // Now that the version rows are rendered, append them
            // after the summary row.
            var summaryRow = Y.one("#" + getCN("package-summary",
                                               packageInfo.name));
            insertNodeAfter(summaryRow, rendered);

            // Finally, bind and sync the UI for the rendered rows.
            if (versionFilterRow !== undefined) {
                versionFilterRow.bindUI();
            }
            Y.Array.each(versionRows, function(versionRow) {
                versionRow.bindUI();
                versionRow.syncUI();
            });
            this._gutterRowDisplayValue = gutterRow.getStyle("display");
            this.set(VERSION_ROWS_RENDERED, true);
        } else {
            // The version filter, if appropriate, version rows and the gutter
            // row are rendered.  They need to made visible.
            if (versionFilterRow !== undefined) {
                versionFilterRow.show();
            }
            Y.Array.each(versionRows, function(versionRow) {
                versionRow.show();
            });
            gutterRow = this.get(GUTTER_ROW);
            gutterRow.setStyle("display", this._gutterRowDisplayValue);
        }

        this.get(ROW).replaceClass("ls-pkg-tree-collapsed",
                                   "ls-pkg-tree-expanded");
        this.set(EXPANDED, true);
        this.get(TOGGLE_ICON).set(IMAGE, this.get(WIDGET).get(EXPAND_ICON_PATH));
    },

    /**
     * Hide the version filter row, if it's present, along with all version
     * rows.
     *
     * @method _hideVersionFilterAndRows
     * @private
     */
    _hideVersionFilterAndRows: function() {
        var versionRows = this.get(VERSION_ROWS);
        Y.Array.each(versionRows, function(versionRow) {
            versionRow.hide();
        });
        var versionFilterRow = this.get(VERSION_FILTER_ROW);
        if (versionFilterRow !== undefined) {
            versionFilterRow.hide();
        }
        var gutterRow = this.get(GUTTER_ROW);
        gutterRow.setStyle("display", "none");

        this.get(ROW).replaceClass("ls-pkg-tree-expanded",
                                   "ls-pkg-tree-collapsed");
        this.set(EXPANDED, false);
        var packageInfo = this.get(PACKAGE_INFO);
        this.get(TOGGLE_ICON).set(IMAGE, this.get(WIDGET).get(COLLAPSE_ICON_PATH));
    }
});

Y.landscape.packages.PackageSummaryRow = PackageSummaryRow;


/**
 * A PackageVersionFilterRow allows a user to change the version of the package
 * they are looking at.  By default, only the latest version of each package on
 * each computer is shown.
 *
 * @class PackageVersionFilterRow
 * @extends Base
 * @constructor
 */
var PackageVersionFilterRow = function(config) {
    PackageVersionFilterRow.superclass.constructor.apply(this, arguments);
};

PackageVersionFilterRow.NAME = PACKAGE_VERSION_FILTER_ROW;

PackageVersionFilterRow.ATTRS = {
    summaryRow: {
        value: undefined
    }
};

Y.extend(PackageVersionFilterRow, Y.Base, {
    /**
     * Render this version filter row and add it to the package listing.  The
     * row is added as a sibling of the specified row.  The method returns the
     * TR Y.Node object created to contain the version filter.
     *
     * @method renderUI
     */
    renderUI: function(appendAfterNode) {
        var summaryRow = this.get(SUMMARY_ROW);
        var parentWidget = summaryRow.get(WIDGET);
        var packageInfo = summaryRow.get(PACKAGE_INFO);
        var row = Y.Node.create(
            "<tr class='ls-pkg-tree-unmarked ls-pkg-tree-version-row' />");
        row.set(
            "id", getCN("ls-pkg-tree-version-filter-row", packageInfo.name));
        this.set(ROW, row);

        // Build disclosure triangle cell.
        var triangleCell = Y.Node.create(
            "<td class='expand inner-cell version-gutter' />");
        row.appendChild(triangleCell);

        // Build package action buttons cell.
        var actionButtonsCell =
            Y.Node.create("<td class='expand inner-cell' />");
        row.appendChild(actionButtonsCell);

        // Build package icon cell.
        var packageIconCell = Y.Node.create("<td/>");
        row.appendChild(packageIconCell);

        // Build version drop-down list in the package name cell.
        var packageNameCell = Y.Node.create("<td class='inner-cell' />");
        row.appendChild(packageNameCell);
        var versionList = Y.Node.create("<select />");
        versionList.set(
            "id", getCN("ls-pkg-tree-version-filter", packageInfo.name));
        versionList.addClass("ls-pkg-tree-version-filter");
        packageNameCell.appendChild(versionList);
        var latestVersionOption = Y.Node.create("<option>Latest</option>");
        latestVersionOption.set(VALUE, LATEST_VERSION);
        versionList.appendChild(latestVersionOption);

        // Find out which package versions should be displayed, by
        // looking at the kind of the relations for this package.
        var relation, packageVersion;
        var versions = {};
        var relations = packageInfo.relations;
        var i = relations.length;

        while (i--) {
            relation = relations[i];
            if (relation.kind !== INSTALLED) {
                if (relation.kind == UPGRADE) {
                    packageVersion = relation.upgrade;
                }
                if (relation.kind == DOWNGRADE) {
                    packageVersion = relation.downgrade;
                }
                if (relation.kind == AVAILABLE) {
                    packageVersion = relation.available;
                }
                if (!versions.hasOwnProperty(packageVersion.id)) {
                    versions[packageVersion.id] = true;
                }
            }
        }

        // Now iterate over all the available package versions in
        // order and if they were selected for display above, append
        // an entry to the selection list for them.
        Y.Array.each(packageInfo.packages, function(packageData) {
            if (versions.hasOwnProperty(packageData.id)) {
                var versionOption = Y.Node.create("<option />");
                versionOption.set(TEXT, packageData.version);
                versionOption.set(VALUE, packageData.id);
                versionList.appendChild(versionOption);
            }
        });

        versionList.set(SELECTED_INDEX, 0);

        // Build summary icon cell.
        //var summaryIconCell = Y.Node.create("<td/>");
      //  row.appendChild(summaryIconCell);

        // Build summary message cell.
        var summaryMessageCell = Y.Node.create(
            "<td colspan='2' class='expand' />");
        summaryMessageCell.set(
            "id", getCN("ls-pkg-tree-version-summary", packageInfo.name));
        row.appendChild(summaryMessageCell);

        // Build status message cell.
        var statusMessageCell = Y.Node.create("<td />");
        row.appendChild(statusMessageCell);

        // Insert row into the DOM last to avoid needless reflow.
        insertNodeAfter(appendAfterNode, row);
        return row;
    },

    /**
     * Register event handlers for elements in this version filter row.
     *
     * @method bindUI
     */
    bindUI: function() {
        var summaryRow = this.get(SUMMARY_ROW);
        var packageInfo = summaryRow.get(PACKAGE_INFO);
        var versionList = Y.one("#" +
            getCN("ls-pkg-tree-version-filter", packageInfo.name));
        versionList.on("change", function(evt) {
            var target = evt.currentTarget;
            var option = target.get("options").item(target.get(SELECTED_INDEX));
            option.setAttribute("selected", "selected");
            summaryRow.syncUI();
            summaryRow.get(WIDGET).syncUI();
        });
    },

    /**
     * Synchronize the user interface state with the widget state.  The message
     * displayed in the summary message cell is updated to match the version
     * selected in the drop-down list.
     *
     * @method syncUI
     */
    syncUI: function() {
        var currentVersion = this.getCurrentVersion();
        if (!currentVersion) {
            return;
        }

        var packageInfo = this.get(SUMMARY_ROW).get(PACKAGE_INFO);
        var message;
        if (currentVersion == LATEST_VERSION) {
            message = ("Showing latest version of " + packageInfo.name +
                       " for all computers");
        } else {
            var packageData = currentVersion;
            message = ("Showing " + packageData.name + " " +
                       packageData.version + " for matching computers");
        }

        var summaryMessageCell = Y.one("#" +
            getCN("ls-pkg-tree-version-summary", packageInfo.name));

        if (summaryMessageCell.get(TEXT) != message) {
            summaryMessageCell.set(TEXT, message);
        }
    },

    /**
     * Get the currently selected version filtering option.
     *
     * If the version list is not present, because it hasn't been rendered or
     * because it's misconfigured, this method returns undefined.  If the
     * 'Latest' entry is selected the special LATEST_VERSION token will be
     * returned.  Otherwise, a package data object for the selected package
     * version will be returned.
     *
     * @method getCurrentVersion
     */
    getCurrentVersion: function() {
        var packageInfo = this.get(SUMMARY_ROW).get(PACKAGE_INFO);
        var versionList = Y.one("#" +
            getCN("ls-pkg-tree-version-filter", packageInfo.name));
        if (versionList) {
            var selectedIndex = versionList.get(SELECTED_INDEX);
            var option = versionList.get("options").item(selectedIndex);
            var currentVersion = option.getAttribute(VALUE);
            if (currentVersion == LATEST_VERSION) {
                return LATEST_VERSION;
            } else {
                // Iterating to find the package we want isn't the
                // awesomest, but most of the time there will be very few
                // packages to iterate over so it's probably okay.
                for (var i in packageInfo.packages) {
                    var packageData = packageInfo.packages[i];
                    if (packageData.id == currentVersion) {
                        return packageData;
                    }
                }
            }
        }
        return undefined;
    },

    /**
     * Adjust the CSS 'display' property to hide the version filter row, used
     * when the summary row is collapsed and version rows are hidden.
     *
     * @method hide
     */
    hide: function() {
        var packageInfo = this.get(SUMMARY_ROW).get(PACKAGE_INFO);
        var row = Y.one("#" +
            getCN("ls-pkg-tree-version-filter-row", packageInfo.name));
        row.setStyle("display", "none");
    },

    /**
     * Adjust the CSS 'display' property to show the version filter row, used
     * when the summary row is expanded and version rows are shown.
     *
     * @method show
     */
    show: function() {
        var packageInfo = this.get(SUMMARY_ROW).get(PACKAGE_INFO);
        var row = Y.one("#" +
            getCN("ls-pkg-tree-version-filter-row", packageInfo.name));
        if (!row.hasClass("ls-pkg-tree-suppress-visibility")) {
            row.setStyle("display", "table-row");
        }
    }
});

Y.landscape.packages.PackageVersionFilterRow = PackageVersionFilterRow;


/**
 * A Landscape.PackageVersionRow represents a relation between versions in a
 * package, along with the computers affected by it.
 *
 * @class PackageVersionRow
 * @extends Base
 * @constructor
 */
var PackageVersionRow = function(config) {
    PackageVersionRow.superclass.constructor.apply(this, arguments);
};

PackageVersionRow.NAME = PACKAGE_VERSION_ROW;

PackageVersionRow.ATTRS = {
    packageIcon: {
        value: undefined
    },

    summaryRow: {
        value: undefined
    },

    relation: {
        value: undefined
    },

    row: {
        value: undefined
    },

    marked: {
        value: UNMARKED
    },
};

Y.extend(PackageVersionRow, Y.Base, {
    /**
     * Render this version row and add it to the package listing.  The row is
     * added as a sibling of the specified row.
     *
     * @method renderUI
     */
    renderUI: function(appendAfterNode) {
        var markForInstallButton,
            markForDowngradeButton,
            markForUpgradeButton,
            markForRemovalButton,
            markedMessageSpan,
            unmarkedMessageSpan,
            message,
            removeMessageSpan,
            downgradeMessageSpan;
        var parentWidget = this.get(SUMMARY_ROW).get(WIDGET);
        var relation = this.get(RELATION);
        var row = Y.Node.create(
            "<tr class='ls-pkg-tree-unmarked ls-pkg-tree-version-row' />");
        insertNodeAfter(appendAfterNode, row);
        this.set(ROW, row);

        // Build the first cell, used to show the disclosure triangle in
        // summary rows.
        row.appendChild(
            Y.Node.create("<td class='version-gutter' />"));

        // Build the action button row.
        var actionButtonCell =
            Y.Node.create("<td class='inner-cell' />");
        row.appendChild(actionButtonCell);
        var actionButtonDiv =
            Y.Node.create("<div class='package-action-buttons' />");
        actionButtonCell.appendChild(actionButtonDiv);

        if (relation.kind == AVAILABLE) {
            markForInstallButton = new Y.landscape.ui.Sprite({
                "cssClasses": [MARK_FOR_INSTALL, ACTION_BUTTON],
                "title": "Mark for installation",
                "imageId": MARK_FOR_INSTALL + "-" + relation.available.id,
                "image": parentWidget.get(MARK_FOR_INSTALL_ICON_PATH)});
            markForInstallButton.render(actionButtonDiv);
        } else if (relation.kind == INSTALLED) {
            markForRemovalButton = new Y.landscape.ui.Sprite({
                "cssClasses": [MARK_FOR_REMOVAL, ACTION_BUTTON],
                "title": "Mark for removal",
                "imageId": MARK_FOR_REMOVAL + "-" + relation.installed.id,
                "image": parentWidget.get(MARK_FOR_REMOVAL_ICON_PATH)});
            // If the relation is not removable, then we add a CSS
            // class to mark the action as disabled and change the
            // icon to match.
            if (relation.held) {
                markForRemovalButton.addClass("disabled-action-button");
                markForRemovalButton.set(IMAGE, parentWidget.get(
                    MARK_FOR_REMOVAL_DISABLED_ICON_PATH));
            }
            markForRemovalButton.render(actionButtonDiv);
        } else if (relation.kind == UPGRADE) {
            markForUpgradeButton = new Y.landscape.ui.Sprite({
                "cssClasses": [MARK_FOR_UPGRADE, ACTION_BUTTON],
                "title": "Mark for upgrade",
                "imageId": MARK_FOR_UPGRADE + "-" + relation.upgrade.id,
                "image": parentWidget.get(MARK_FOR_UPGRADE_ICON_PATH)});
            if (relation.held) {
                markForUpgradeButton.addClass("disabled-action-button");
                markForUpgradeButton.set(IMAGE, parentWidget.get(
                    MARK_FOR_UPGRADE_DISABLED_ICON_PATH));
            }
            markForUpgradeButton.render(actionButtonDiv);

            markForRemovalButton = new Y.landscape.ui.Sprite({
                "cssClasses": [MARK_FOR_REMOVAL, ACTION_BUTTON],
                "title": "Mark for removal",
                "imageId": MARK_FOR_REMOVAL + "-" + relation.upgrade.id,
                "image": parentWidget.get(MARK_FOR_REMOVAL_ICON_PATH)});
            if (relation.held) {
                markForRemovalButton.addClass("disabled-action-button");
                markForRemovalButton.set(IMAGE, parentWidget.get(
                    MARK_FOR_REMOVAL_DISABLED_ICON_PATH));
            }
            markForRemovalButton.render(actionButtonDiv);
        } else {
            markForDowngradeButton = new Y.landscape.ui.Sprite({
                "cssClasses": [MARK_FOR_DOWNGRADE, ACTION_BUTTON],
                "title": "Mark for downgrade",
                "imageId": MARK_FOR_DOWNGRADE + "-" + relation.downgrade.id,
                "image": parentWidget.get(MARK_FOR_DOWNGRADE_ICON_PATH)});
            if (relation.held) {
                markForDowngradeButton.addClass("disabled-action-button");
                markForDowngradeButton.set(IMAGE, parentWidget.get(
                    MARK_FOR_DOWNGRADE_DISABLED_ICON_PATH));
            }
            markForDowngradeButton.render(actionButtonDiv);

            markForRemovalButton = new Y.landscape.ui.Sprite({
                "cssClasses": [MARK_FOR_REMOVAL, ACTION_BUTTON],
                "title": "Mark for removal",
                "imageId": MARK_FOR_REMOVAL + "-" + relation.downgrade.id,
                "image": parentWidget.get(MARK_FOR_REMOVAL_ICON_PATH)});
            if (relation.held) {
                markForRemovalButton.addClass("disabled-action-button");
                markForRemovalButton.set(IMAGE, parentWidget.get(
                    MARK_FOR_REMOVAL_DISABLED_ICON_PATH));
            }
            markForRemovalButton.render(actionButtonDiv);
        }

        // Build computers icon column.
        var computersIconCell = Y.Node.create("<td/>");
        row.appendChild(computersIconCell);

        // Build computers count column.
        var computersCountCell = Y.Node.create("<td class='inner-cell' />");
        row.appendChild(computersCountCell);
        var computersCountLink = Y.Node.create("<a href='#' />");
        computersCountCell.appendChild(computersCountLink);
        var computersText = Y.Node.create("<nobr />");
        var computersLabel = "";
        if (Y.Lang.isUndefined(relation.computer)) {
            // more than one computer in the relation
            computersLabel = relation.computers.length + " computers";

            computersCountLink.on("click", function(e) {
                e.preventDefault();
                this.fire("display-computers", {computers: relation.computers,
                                                original_target: e.target});
            }, parentWidget);
        } else {
            computersLabel = relation.computer.title;
            computersCountLink.setAttribute(
                "href", "/computer/" + relation.computer.id);
        }

        computersText.set(TEXT, computersLabel);
        computersCountLink.appendChild(computersText);

        // Build package icon column.
        var packageIconCell = Y.Node.create("<td/>");
        row.appendChild(packageIconCell);
        var packageIconPath = this._getPackageIconPath();
        var packageIconImage = new Y.landscape.ui.Sprite({
            "cssClasses": ["ls-pkg-tree-action-icon"],
            "image": parentWidget.get(packageIconPath)});
        this.set(PACKAGE_ICON, packageIconImage);
        parentWidget.setPackageIconHoldAction(
            packageIconImage, relation.kind, relation.held);
        // If the relation is held, we need to render a lock overlay
        // on top of the Package Icon.
        this._setPackageIconPath(packageIconPath);
        packageIconImage.render(packageIconCell);

        // Build summary text column.
        var computerURL = parentWidget.get(COMPUTER_URL);
        var createPackageLink = function(packageData) {
            var packageLink = Y.Node.create("<a />");
            var url = computerURL + "/packages/" + packageData.id;
            packageLink.setAttribute("href", url);
            packageLink.set(TEXT, packageData.version);
            return packageLink;
        };

        var packageSummaryCell =
            Y.Node.create("<td class='ls-pkg-tree-summary' />");
        row.appendChild(packageSummaryCell);
        if (relation.kind == AVAILABLE) {
            // Unmarked message.
            unmarkedMessageSpan =
                Y.Node.create("<span class='ls-pkg-tree-unmarked' />");
            packageSummaryCell.appendChild(unmarkedMessageSpan);
            unmarkedMessageSpan.appendChild(
                createPackageLink(relation.available));
            message = "&nbsp;is available";
            if (relation.held) {
                message += " but held";
            }
            unmarkedMessageSpan.appendChild(Y.Node.create(message));

            markedMessageSpan =
                Y.Node.create("<span class='ls-pkg-tree-install' />");
            packageSummaryCell.appendChild(markedMessageSpan);
            markedMessageSpan.appendChild(
                createPackageLink(relation.available));
            markedMessageSpan.appendChild(
                Y.Node.create("&nbsp;is marked to be installed"));
        } else if (relation.kind == INSTALLED) {
            // Unmarked message.
            unmarkedMessageSpan =
                Y.Node.create("<span class='ls-pkg-tree-unmarked' />");
            packageSummaryCell.appendChild(unmarkedMessageSpan);
            unmarkedMessageSpan.appendChild(
                createPackageLink(relation.installed));
            message = "&nbsp;is installed";
            if (relation.held) {
                message += " and held at this version";
            }
            unmarkedMessageSpan.appendChild(
                Y.Node.create(message));

            // Marked for removal message. Only makes sense if the
            // installed package isn't held.
            if (!relation.held) {
                markedMessageSpan =
                    Y.Node.create("<span class='ls-pkg-tree-remove' />");
                packageSummaryCell.appendChild(markedMessageSpan);
                markedMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                markedMessageSpan.appendChild(
                    Y.Node.create("&nbsp;is marked to be removed"));
            }
        } else if (relation.kind == UPGRADE) {
            // Unmarked message.
            unmarkedMessageSpan =
                Y.Node.create("<span class='ls-pkg-tree-unmarked' />");
            packageSummaryCell.appendChild(unmarkedMessageSpan);
            unmarkedMessageSpan.appendChild(
                createPackageLink(relation.installed));

            message = "&nbsp;is installed";
            if (relation.held) {
                message += " and held";
            }
            message += "&nbsp;";
            unmarkedMessageSpan.appendChild(Y.Node.create(message));

            var upgradable = unmarkedMessageSpan;

            // If the relation is held, create a del element
            // around it.
            if (relation.held) {
                upgradable = Y.Node.create("<del />");
                unmarkedMessageSpan.appendChild(upgradable);
            }
            upgradable.appendChild(
                Y.Node.create("and upgradable to&nbsp;"));
            upgradable.appendChild(
                createPackageLink(relation.upgrade));

            // Marked for upgrade message. Should only be created if
            // the relation is not held.
            if (!relation.held) {
                var upgradeMessageSpan = Y.Node.create(
                    "<span class='ls-pkg-tree-upgrade ls-pkg-tree-install' />");
                packageSummaryCell.appendChild(upgradeMessageSpan);
                upgradeMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                upgradeMessageSpan.appendChild(
                    Y.Node.create("&nbsp;marked to be upgraded to&nbsp;"));
                upgradeMessageSpan.appendChild(
                    createPackageLink(relation.upgrade));
            }

            // Marked for removal message. Should only be created if
            // the installed package is removable.
            if (!relation.held) {
                removeMessageSpan =
                    Y.Node.create("<span class='ls-pkg-tree-remove' />");
                packageSummaryCell.appendChild(removeMessageSpan);
                removeMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                removeMessageSpan.appendChild(
                    Y.Node.create("&nbsp;is marked to be removed"));
            }
        } else {
            // Unmarked message.
            unmarkedMessageSpan =
                Y.Node.create("<span class='ls-pkg-tree-unmarked' />");
            packageSummaryCell.appendChild(unmarkedMessageSpan);
            unmarkedMessageSpan.appendChild(
                createPackageLink(relation.installed));

            message = "&nbsp;is installed";
            if (relation.held) {
                message += " and held";
            }
            message += "&nbsp;";
            unmarkedMessageSpan.appendChild(
                Y.Node.create(message));

            var downgradable = unmarkedMessageSpan;

            // If the relation is held, create a del element
            // around it.
            if (relation.held) {
                downgradable = Y.Node.create("<del />");
                unmarkedMessageSpan.appendChild(downgradable);
            }
            downgradable.appendChild(
                Y.Node.create("and downgradable to&nbsp;"));
            downgradable.appendChild(
                createPackageLink(relation.downgrade));

            // Marked for downgrade message. Only makes sense if the
            // relation isn't held.
            if (!relation.held) {
                downgradeMessageSpan = Y.Node.create(
                    "<span class='ls-pkg-tree-downgrade ls-pkg-tree-install' />");
                packageSummaryCell.appendChild(downgradeMessageSpan);
                downgradeMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                downgradeMessageSpan.appendChild(
                    Y.Node.create("&nbsp;marked to be downgraded to&nbsp;"));
                downgradeMessageSpan.appendChild(
                    createPackageLink(relation.downgrade));
            }

            // Marked for removal message. Only makes sense if the
            // installed package is removable.
            if (!relation.held) {
                removeMessageSpan =
                    Y.Node.create("<span class='ls-pkg-tree-remove' />");
                packageSummaryCell.appendChild(removeMessageSpan);
                removeMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                removeMessageSpan.appendChild(
                    Y.Node.create("&nbsp;is marked to be removed"));
            }
        }
        if (relation.installed) {
            if (relation.held) {
                var unheldMessageSpan =
                    Y.Node.create("<span class='ls-pkg-tree-unhold' />");
                packageSummaryCell.appendChild(unheldMessageSpan);
                unheldMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                unheldMessageSpan.appendChild(
                    Y.Node.create("&nbsp;is marked to be unheld"));
            } else {
                var heldMessageSpan =
                    Y.Node.create("<span class='ls-pkg-tree-hold' />");
                packageSummaryCell.appendChild(heldMessageSpan);
                heldMessageSpan.appendChild(
                    createPackageLink(relation.installed));
                heldMessageSpan.appendChild(
                    Y.Node.create("&nbsp;is marked to be held"));
            }
        }

        // Build status text column.
        var statusMessageCell = Y.Node.create("<td />");
        row.appendChild(statusMessageCell);
        if (relation.kind == UPGRADE && relation.upgrade.usn !== undefined) {
            var usnLink = Y.Node.create("<a />");
            usnLink.setAttribute(
                "href",
                "https://www.ubuntu.com/usn/usn-" + relation.upgrade.usn.name);
            usnLink.set(TEXT, "USN " + relation.upgrade.usn.name);
            statusMessageCell.appendChild(usnLink);
            Y.Array.each(relation.upgrade.usn.cves, function(cve) {
              var cveLink = Y.Node.create("<a />");
              cveLink.setAttribute(
                  "href",
                  "https://cve.mitre.org/cgi-bin/cvename.cgi?name=" + cve);
              cveLink.set(TEXT, cve);
              statusMessageCell.appendChild(Y.Node.create("<br />"));
              statusMessageCell.appendChild(cveLink);
            });
        }

        return row;
    },

    /**
     * Register event handlers for elements in this version row.
     *
     * @method bindUI
     */
    bindUI: function() {
        var selector;
        var relation = this.get(RELATION);
        if (relation.kind == AVAILABLE) {
            // Bind a handler for the 'Mark for installation' button.
            selector = "#ls-pkg-tree-mark-for-install-" + relation.available.id;
            this._bindHandler(
                selector, "markForInstallation", "ls-pkg-tree-install");
        } else if (relation.kind == UPGRADE) {
            // Bind a handler for the 'Mark for upgrade' button.
            selector = "#ls-pkg-tree-mark-for-upgrade-" + relation.upgrade.id;
            this._bindHandler(
                selector, "markForUpgrade", "ls-pkg-tree-upgrade");

            // Bind a handler for the 'Mark for removal' button.
            selector = "#ls-pkg-tree-mark-for-removal-" + relation.upgrade.id;
            this._bindHandler(selector, "markForRemoval", "ls-pkg-tree-remove");
        } else if (relation.kind == DOWNGRADE) {
            // Bind a handler for the 'Mark for downgrade' button.
            selector = "#ls-pkg-tree-mark-for-downgrade-" + relation.downgrade.id;
            this._bindHandler(
                selector, "markForDowngrade", "ls-pkg-tree-downgrade");

            // Bind a handler for the 'Mark for removal' button.
            selector = "#ls-pkg-tree-mark-for-removal-" + relation.downgrade.id;
            this._bindHandler(selector, "markForRemoval", "ls-pkg-tree-remove");
        } else if (relation.kind == INSTALLED) {
            // Bind a handler for the 'Mark for removal' button.
            selector = "#ls-pkg-tree-mark-for-removal-" + relation.installed.id;
            this._bindHandler(selector, "markForRemoval", "ls-pkg-tree-remove");
        }
        this._bindHandler(
            "." + MARK_FOR_HOLD, "markForHold", "ls-pkg-tree-hold");
        this._bindHandler(
            "." + MARK_FOR_UNHOLD, "markForUnhold", "ls-pkg-tree-unhold");
    },

    /**
     * Bind an action button that toggles marking and unmarking of the row.
     *
     * @param markCallback: The name of the method to call when this row is
     *     marked.
     * @param cssClass: The name of the CSS class set when the row is marked.
     * @method _bindHandler
     * @private
     */
    _bindHandler: function(selector, markCallback, cssClass) {
        this.get(ROW).all(selector).on("click", function(evt) {
            evt.stopPropagation();
            if (!this.get(ROW).hasClass(cssClass)) {
                this[markCallback]();
            } else {
                this.unmark();
            }
            var summaryRow = this.get(SUMMARY_ROW);
            summaryRow.syncUI();
            var widget = summaryRow.get(WIDGET);
            widget.syncUI();
            widget.deselectAll();
        }, this);
    },

    /**
     * Synchronize the user interface with the widget state.
     *
     * @method syncUI
     */
    syncUI: function() {
        var mark = this.get(MARKED);
        switch (mark) {
        case ATTR_INSTALL:
            this.markForInstallation();
            break;
        case ATTR_UPGRADE:
            this.markForUpgrade();
            break;
        case ATTR_REMOVE:
            this.markForRemoval();
            break;
        case ATTR_HOLD:
            this.markForHold();
            break;
        case UNMARKED:
        case undefined:
            this.unmark();
            break;
        }
    },

    /**
     * Get the path to the package icon to use for this row.
     **/
    _getPackageIconPath: function() {
        var parentWidget = this.get(SUMMARY_ROW).get(WIDGET);
        var relation = this.get(RELATION);
        var lockState = UNLOCKED;
        if (relation.held) {
            lockState = LOCKED;
        }
        return parentWidget.getPackageIconPath(
            relation.kind, lockState, this.get(MARKED));
    },

    // /**
    //  * Returns INPUT elements with values set to mark a package for
    //  * installation or removal.
    //  *
    //  * @param packageProperty: The name of the property on the relation object
    //  *     containing the package data object to mark.
    //  * @param marker: A string, either 'i' or 'r' for installation and removal.
    //  * @method _createInputs
    //  * @private
    //  */
    _createInputs: function(packageProperty, marker) {
        var relation = this.get(RELATION);
        var inputs = [];
        Y.Array.each(relation.computers, function(computerId) {
            var input = Y.Node.create("<input type='hidden' />");
            input.set("id", ("changes-" + computerId + "-" +
                             relation[packageProperty].id));
            input.set("name", "changes-" + computerId + ":list");
            input.set("value", marker + relation[packageProperty].id);
            inputs.push(input);
        });
        return inputs;
    },

    /**
     * Removes INPUT elements from the package listing form, if any exist, for
     * this version row.
     *
     * @param packageProperty: The name of the property on the relation object
     *     containing the package data object to unmark.
     * @method _removeInputs
     * @private
     */
    _removeInputs: function(packageProperty) {
        var form = this.get(SUMMARY_ROW).get(WIDGET).get(CONTENT_BOX);
        var relation = this.get(RELATION);
        var packageId = relation[packageProperty].id;
        // Clear inputs only for the specified package
        Y.Array.each(relation.computers, function(computerId) {
            form.all("#changes-" + computerId + "-" + packageId).remove();
        });
    },

    /**
     * Set the path of the package icon using the path retrieved from the
     * widget's configuration using the specified name.
     *
     * @method _setPackageIconPath
     * @private
     */
    _setPackageIconPath: function(configName) {
        var relation = this.get(RELATION),
            lockState = relation.held ? LOCKED : UNLOCKED,
            widget = this.get(SUMMARY_ROW).get(WIDGET);
        widget.updatePackageIcon(
            this.get(PACKAGE_ICON), configName, this.get(MARKED), lockState);
    },

    /**
     * Adjust CSS styles, form values and icon images to mark this row for
     * installation.
     *
     * @method markForInstallation
     */
    markForInstallation: function() {
        this.unmark();
        var parentWidget = this.get(SUMMARY_ROW).get(WIDGET);
        var row = this.get(ROW);
        if (row) {
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-install");
            this._setPackageIconPath(INSTALLING_ICON_PATH);
        }

        var form = parentWidget.get(CONTENT_BOX);
        var inputs = new Y.NodeList(this._createInputs(ATTR_AVAILABLE, "i"));
        form.one("#hidden-inputs").append(inputs);

        this.set(MARKED, ATTR_INSTALL);
    },

    /**
     * Returns true if this row is marked for installation.
     *
     * @method markedForInstallation
     */
    markedForInstallation: function() {
        return this.get(MARKED) == ATTR_INSTALL;
    },

    /**
     * Adjust CSS to mark this row for upgrade.
     *
     * @method markForUpgrade
     */
    markForUpgrade: function() {
        var relation = this.get(RELATION);
        if (relation.held) {
            return false;
        }
        this.unmark();
        var parentWidget = this.get(SUMMARY_ROW).get(WIDGET);
        var row = this.get(ROW);
        if (row) {
            row.replaceClass("ls-pkg-tree-remove", "ls-pkg-tree-upgrade");
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-upgrade");
            this._setPackageIconPath(UPGRADING_ICON_PATH);
        }

        var form = parentWidget.get(CONTENT_BOX);
        var inputs = new Y.NodeList(
            this._createInputs(ATTR_INSTALLED, "r").concat(
                this._createInputs(ATTR_UPGRADE, "i")));
        form.one("#hidden-inputs").append(inputs);

        this.set(MARKED, ATTR_UPGRADE);
        return true;
    },

    /**
     * Returns true if this row is marked for upgrade.
     *
     * @method markedForUpgrade
     */
    markedForUpgrade: function() {
        return this.get(MARKED) == ATTR_UPGRADE;
    },

    /**
     * Adjust CSS to mark this row for downgrade.
     *
     * @method markForDowngrade
     */
    markForDowngrade: function() {
        var relation = this.get(RELATION);
        if (!relation.held) {
            this.unmark();
            var parentWidget = this.get(SUMMARY_ROW).get(WIDGET);
            var row = this.get(ROW);
            if (row) {
                row.replaceClass("ls-pkg-tree-remove", "ls-pkg-tree-downgrade");
                row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-downgrade");
                this._setPackageIconPath(DOWNGRADING_ICON_PATH);
            }

            var form = parentWidget.get(CONTENT_BOX);
            var inputs = new Y.NodeList(
                this._createInputs(ATTR_INSTALLED, "r").concat(
                    this._createInputs(ATTR_DOWNGRADE, "i")));
            form.one("#hidden-inputs").append(inputs);

            this.set(MARKED, ATTR_DOWNGRADE);
        }
    },

    /**
     * Adjust CSS to mark this row for removal.
     *
     * @method markForRemoval
     */
    markForRemoval: function() {
        var relation = this.get(RELATION);
        if (relation.installed && !relation.held) {
            this.unmark();
            var row = this.get(ROW);
            if (row) {
                row.replaceClass("ls-pkg-tree-upgrade", "ls-pkg-tree-remove");
                row.replaceClass("ls-pkg-tree-downgrade", "ls-pkg-tree-remove");
                row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-remove");
                this._setPackageIconPath(REMOVING_ICON_PATH);
            }

            var form = this.get(SUMMARY_ROW).get(WIDGET).get(CONTENT_BOX);
            var inputs = new Y.NodeList(this._createInputs(ATTR_INSTALLED, "r"));
            form.one("#hidden-inputs").append(inputs);

            this.set(MARKED, ATTR_REMOVE);
        }
    },

    /**
     * Returns true if this row is marked for removal.
     *
     * @method markedForRemoval
     */
    markedForRemoval: function() {
        return this.get(MARKED) == ATTR_REMOVE;
    },

    /**
     * Returns true if this row is marked for hold.
     *
     * @method markedForHold
     */
    markedForHold: function() {
        return this.get(MARKED) == ATTR_HOLD;
    },

    /**
     * Mark this row for hold.
     *
     * @method markForHold
     */
    markForHold: function() {
        var row = this.get(ROW);
        var form = this.get(SUMMARY_ROW).get(WIDGET).get(CONTENT_BOX);

        this.unmark();
        this.set(MARKED, ATTR_HOLD);
        Y.Array.each(this._createInputs(ATTR_INSTALLED, "h"),
                     function(input) {
                         form.appendChild(input);
                     });
        if (row) {
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-hold");
            this._setPackageIconPath(this._getPackageIconPath());
        }
    },

    /**
     * Mark this row for unhold.
     *
     * @method markForUnhold
     */
    markForUnhold: function() {
        var row = this.get(ROW);
        var form = this.get(SUMMARY_ROW).get(WIDGET).get(CONTENT_BOX);

        this.unmark();
        this.set(MARKED, ATTR_UNHOLD);
        Y.Array.each(this._createInputs(ATTR_INSTALLED, "u"),
                     function(input) {
                         form.appendChild(input);
                     });
        if (row) {
            row.replaceClass("ls-pkg-tree-unmarked", "ls-pkg-tree-unhold");
            this._setPackageIconPath(this._getPackageIconPath());
        }
    },

    /**
     * Returns true if this row is marked for unhold.
     *
     * @method markedForUnhold
     */
    markedForUnhold: function() {
        return this.get(MARKED) == ATTR_UNHOLD;
    },

    /**
     * Adjust the CSS to make this row unmarked.
     *
     * @method unmark
     */
    unmark: function() {
        var parentWidget = this.get(SUMMARY_ROW).get(WIDGET);
        var relation = this.get(RELATION);
        var row = this.get(ROW);

        this.set(MARKED, UNMARKED);
        if (row) {
            row.removeClass("ls-pkg-tree-install");
            row.removeClass("ls-pkg-tree-upgrade");
            row.removeClass("ls-pkg-tree-downgrade");
            row.removeClass("ls-pkg-tree-remove");
            row.removeClass("ls-pkg-tree-hold");
            row.removeClass("ls-pkg-tree-unhold");
            row.addClass("ls-pkg-tree-unmarked");

            this._setPackageIconPath(this._getPackageIconPath());
        }

        switch (relation.kind) {
        case AVAILABLE:
            this._removeInputs(ATTR_AVAILABLE);
            break;
        case INSTALLED:
            this._removeInputs(ATTR_INSTALLED);
            break;
        case UPGRADE:
            this._removeInputs(ATTR_INSTALLED);
            this._removeInputs(ATTR_UPGRADE);
            break;
        case DOWNGRADE:
            this._removeInputs(ATTR_INSTALLED);
            this._removeInputs(ATTR_DOWNGRADE);
            break;
        }

    },

    /**
     * Adjust the CSS 'display' property to hide this row.
     *
     * @method hide
     */
    hide: function() {
        this.get(ROW).setStyle("display", "none");
    },

    /**
     * Adjust the CSS 'display' property to show this row.
     *
     * @method show
     */
    show: function() {
        var row = this.get(ROW);
        if (!row.hasClass("ls-pkg-tree-suppress-visibility")) {
            row.setStyle("display", "table-row");
        }
    }
});

Y.landscape.packages.PackageVersionRow = PackageVersionRow;

}, "0.1", {"requires": ["event", "node", "widget", "classnamemanager",
                        "cache", "datasource-io", "datasource-cache",
                        "datasource-jsonschema", "landscape.util.node",
                        "landscape.notes", "landscape.listing.pager",
                        "landscape.ui.image", "landscape.ui.sprite"]});
