// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.selection.computer", function(Y) {

Y.namespace("landscape.selection");

/**
 * Widgets for interacting with computer selection.
 *
 * @module landscape.selection.computer
 */

/**
 * A widget for showing a list of available computers with an optional
 * initial selection. Upon selection change, allows for retrieval of
 * selected computer ids.
 *
 * @class ComputerSelection
 * @extends Widget
 * @constructor
 */

var staticFilename = Y.landscape.ui.image.staticFilename,
    COMPUTER_SELECTION = "computer_selection",
    ACCOUNT_URL = "accountUrl",
    ALL = "all", // A magic marker value denoting ALL computers.
    AVAILABLE_COMPUTER_IDS = "availableComputerIds",
    CACHE_SIZE = "cacheSize",
    CHOOSEONE = "chooseOne",
    COL_PREFIX = "#col_",
    CONTENT_BOX = "contentBox",
    LISTING_CONTROLS = ".listing-controls",
    FILTER_CONTROLS = ".filter-controls",
    NO_COMPUTERS_MESSAGE = "noComputersMessage",
    NOTES = "notes",
    PAGER = "pager",
    PAGE_SIZE = "pageSize",
    PAGINATION_BOX = "paginationBox",
    PAGE_DATASOURCE = "pageDataSource",
    RENDER_FIRST_PAGE = "renderFirstPage",
    RECORDS = "records",
    ROW_PREFIX = "#computer-row-",
    SELECTABLE = "selectable",
    SELECTABLE_LISTING = "selectableListing",
    SELECTABLE_ROWS = "selectableRows",
    SELECTED_ROWS = "selectedRows",
    SELECT_ALL_CONTROLS = "selectAllControls",
    SELECT_NONE_CONTROLS = "selectNoneControls",
    SELECTION_ADD = 1,
    SELECTION_REMOVE = 2,
    SELECTED_COMPUTER_IDS = "selectedComputerIds",
    SELECTION_INFO = "selectionInfo",
    UNSELECTED_COMPUTER_IDS = "unselectedComputerIds",
    REBOOT_REQUIRED_ICON = staticFilename("/static/computer/reboot-required-17x16.png"),
    COL_SORT_UP_ICON =
        Y.UA.ie ? "&nbsp<font face='webdings'>5</font>" : "&nbsp;&#x25B4;",
    COL_SORT_DOWN_ICON =
        Y.UA.ie ? "&nbsp<font face='webdings'>6</font>" : "&nbsp;&#x25BE;",
    COL_SORT_FORWARD_ID = "#sort-fwd-indicator",
    COL_SORT_REVERSE_ID = "#sort-rev-indicator",
    DEFAULT_NODES = {
        paginationBox: ".computer-selection-pagination",
        selectAllControls: [".computer-selection-select-all"],
        selectNoneControls: [".computer-selection-select-none"]
    };


// Takes either an array of Node objects or selectors, or both, and
// returns an array of Node objects.
var nodesSetter = function(nodes, name) {
    if (Y.Lang.isUndefined(nodes) || (Y.Lang.isArray(nodes) && !nodes.length)) {
        nodes = DEFAULT_NODES[name];
    }
    return Y.bind(Y.landscape.util.nodesSetter, this)(nodes, name);
};

// Takes a node or a selector and returns a Node, defaulting to
// globally defined selector if none is provided.
var nodeSetter = function(node, name) {
    if (Y.Lang.isUndefined(node) || Y.Lang.isNull(node)){
        node = DEFAULT_NODES[name];
    }
    return Y.bind(Y.landscape.util.nodeSetter, this)(node, name);
};

var ComputerSelection = function() {
    ComputerSelection.superclass.constructor.apply(this, arguments);
};

Y.mix(ComputerSelection, {
    NAME: COMPUTER_SELECTION,

    ATTRS: {
        /**
         * Base Account URL for the computers used by this widget.
         *
         * @attribute accountUrl
         * @type String
         */
        accountUrl: {
            value: ""
        },

        /**
         * Optional selection informational text to present at the top of the
         * selection table. Used to inform the user in special cases about the
         * use of this particular selection.
         *
         * @attribute headerLabel
         * @type String
         */
        selectionInfo: {
            value: ""
        },

        /**
         * Cache size to be used for the underlying
         * DataSources. Effectively, the number of pages to cache.
         *
         * @attribute cacheSize
         * @type Integer
         */
        cacheSize: {
            value: 10
        },

        /**
         * Pager instance managed by this widget.
         *
         * @attribute pager
         * @type Pager
         */
        pager: {
            value: null
        },

        /**
         * Page size to be used by the pager.
         *
         * @attribute pageSize
         * @type Integer
         */
        pageSize: {
            value: 25
        },

        /**
         * Node where the pagination box should be rendered.
         *
         * @attribute paginationBox
         * @type Node
         */
        paginationBox: {
            value: null,
            setter: nodeSetter
        },

        /**
         * Page datasource for Ajax events.
         *
         * @attribute pageDataSource
         * @type PageDataSource
         */
        pageDataSource: { value: null },

        /**
         * SelectableListing instance managed by this widget.
         *
         * @attribute selectableListing
         * @type SelectableListing
         */
        selectableListing: {
            value: null
        },

        /**
         * Is this widget selectable?
         *
         * @attribute selectable
         * @type boolean
         */
        selectable: {
            value: true
        },

        /**
         * Is this widget's selection limited to choosing only one?
         *
         * @attribute chooseOne
         * @type boolean
         */
        chooseOne: {
            value: false
        },

        /**
         * Should the pager render the first page or progressively
         * enhance what's there?
         *
         * @attribute renderFirstPage
         * @type Boolean
         */
        renderFirstPage: {
            value: true
        },

        /**
         * Is this widget's selection limited to choosing only one?
         *
         * @attribute chooseOne
         * @type String
         */
        noComputersMessage: {
            value: "No computers meet selection criteria."
        },

        /**
         * Notes instance managed by this widget.
         *
         * @attribute notes
         * @type Notes
         */
        notes: {
            value: null
        },

        /**
         * A list of CSS selectors for controls that, when clicked,
         * should mark all computers.
         *
         * @attribute selectAllControls
         * @type Array
         */
        selectAllControls: {
            value: [],
            setter: nodesSetter
        },

        /**
         * A list of CSS selectors for controls that, when clicked,
         * should unmark all computers.
         *
         * @attribute selectNoneControls
         * @type Array
         */
        selectNoneControls: {
            value: [],
            setter: nodesSetter
        },

        /**
         * All of the available computer ids. For a computer to be
         * selectable its id must be part of this list. Selecting all
         * computers means all the availableComputerIds are marked as
         * selected.
         *
         * @attribute availableComputerIds
         * @type Array
         */
        availableComputerIds: {
            value: []
        },

        /**
         * All of the selected computer ids.
         *
         * @attribute selectedComputerIds
         * @type Array
         */
        selectedComputerIds: {
            value: []
        },

        /**
         * All of the unselected computer ids.
         *
         * @attribute unselectedComputerIds
         * @type Array
         */
        unselectedComputerIds: {
            value: []
        }
    }
});


Y.extend(ComputerSelection, Y.Widget, {
    // Externalize some constants for easier testing.
    SORT_FORWARD: COL_SORT_FORWARD_ID,
    SORT_REVERSE: COL_SORT_REVERSE_ID,
    ROW_PREFIX: ROW_PREFIX,
    COL_PREFIX: COL_PREFIX,

    FILTER_TEMPLATE: ["",
"        <div class='filter-controls'>",
"          Select: <a class='computer-selection-select-all' ",
"                     href='#'>All</a> ",
"                  <a class='computer-selection-select-none' href='#'>None</a>",
"        </div>"].join(""),

    // Default CONTENT_BOX content if none is provided.
    CONTENT_TEMPLATE: ["",
"<table class='inline-listing merged-columns'>",
"  <thead>",
"    <tr>",
"      <th class='listing-controls' colspan='7'>",
"        <div class='computer-selection-pagination'></div>",
"      </th>",
"    </tr>",
"    <tr>",
"      <th class='tight-fit column-icon'>&nbsp;</th>",
"      <th id='col_title' class='inner-cell sortable'>Name</th>",
"      <th id='col_hostname' class='inner-cell sortable'>Hostname</th>",
"      <th id='col_last_ping_time'",
"          class='sortable column-text-small'>Last ping time</th>",
"      <th class='tight-fit column-icon'>&nbsp;</th>",
"    </tr>",
"  </thead>",
"  <tr>",
"    <td class='column-box tight-fit'>",
"        <input type='checkbox' name='selection:list' />",
"    </td>",
"    <td class='inner-cell' style='width: 25%;'>",
"      <a href='#'>Computer</computer>",
"    </td>",
"    <td class='inner-cell'>Hostname</td>",
"    <td style='width: 18%;'>Yesterday</td>",
"    <td class='tight-fit'>",
"  </tr>",
"</table>"].join(""),

    SELECT_ALL_TEMPLATE : ["",
"{count}",
" computers selected. ",
"<a href='#' class='select-all-confirm'>Select all ",
"{total}",
" computers</a>"].join(""),

    /**
     * Constructor for the widget. Initializes some instance variables
     * from the provided config. Some config options are passed
     * directly to the underlying Pager instance.
     */
    initializer: function(config) {
        config = config || {};
        var selectable = this.get(SELECTABLE),
            chooseOne = this.get(CHOOSEONE),
            selectionInfo = this.get(SELECTION_INFO),
            contentBox = this.get(CONTENT_BOX),
            listingControls = contentBox.one(LISTING_CONTROLS),
            filterControls = listingControls ? listingControls.one(FILTER_CONTROLS) : null;

        // If chooseOne is specified, we can't have a selectable list
        if (chooseOne) {
            selectable = false;
            this.set(SELECTABLE, selectable);
        }

        var content = ["        <div class='table-header-info'>",
                       selectionInfo,
                       "</div>"].join("");
        listingControls.prepend(Y.Node.create(content));

        if (selectable && Y.Lang.isNull(filterControls)) {
            listingControls.append(Y.Node.create(this.FILTER_TEMPLATE));
        }

        // If no Notes widget was provided, initialize one ourselves.
        if (!this.get(NOTES)) {
            var notes = new Y.landscape.notes.TableNotes({
                contentBox: this.get(CONTENT_BOX), animDuration: 0.01});
            this.set(NOTES, notes);
        }

        // Default schema for the Computer Selection
        // DataSource.
        var sourceSchema = {
            metaFields: {"totalResults": "totalRecords"},
            resultListLocator: "records",
            resultFields:[{key: "title"},
                          {key: "hostname"},
                          {key: "last_activity"},
                          {key: "computer_id"},
                          {key: "selected"},
                          {key: "reboot_required_flag"},
                          {key: "reboot_required_text"}]
        };

        var cacheSize = this.get(CACHE_SIZE);

        // If keywords were provided on construction, initialize the
        // Pager's sourceArguments with them.
        var pageDataSource = this.get(PAGE_DATASOURCE);
        var sourceArguments = {"page_key": pageDataSource.page_key};
        if (!Y.Lang.isUndefined(config.searchCriteria)) {
            sourceArguments.criteria = config.searchCriteria;
        }

        // If a recordSource wasn't provided, use the default one.
        var recordSource = config.recordSource;
        if (Y.Lang.isUndefined(recordSource)) {
            recordSource = new Y.DataSource.IO({source: "available"});
        }

        // The cached entries are the entire data source response
        // for a page.
        recordSource.plug(
            {fn: Y.Plugin.DataSourceCache, cfg: {max: cacheSize}});
        recordSource.plug(
            {fn: Y.Plugin.DataSourceJSONSchema, cfg: {schema: sourceSchema}});

        var pagerConfig = {contentBox: this.get(CONTENT_BOX),
                           sortKeyParameter: "sort",
                           sortDirectionParameter: "sort_dir",
                           renderFirstPage: this.get(RENDER_FIRST_PAGE),
                           rowRenderer: Y.rbind(this.rowRender, this,
                                                selectable, chooseOne),
                           pageSize: this.get(PAGE_SIZE),
                           recordSource: recordSource,
                           recordSourceArguments: sourceArguments};


        // Pager
        var pager = new Y.landscape.listing.StaticPager(pagerConfig);
        pager.set("availableIds", this.get(AVAILABLE_COMPUTER_IDS));
        pager.set("availableIdsSource", config.availableComputerIdsSource);

        this.after(AVAILABLE_COMPUTER_IDS + "Change", function(evt) {
            pager.set("availableIds", evt.newVal);
        });

        // Pagination
        var paginationBox = this.get(PAGINATION_BOX);
        if (paginationBox) {
            pager.plug(Y.landscape.listing.PaginationPlugin,
                {paginationBox: paginationBox});
        }

        this.set(PAGER, pager);

        if (selectable) {
            // Selectable Listing
            var selectableListing = new Y.landscape.listing.SelectableListing(
                this.getAttrs([CONTENT_BOX,
                               SELECT_ALL_CONTROLS,
                               SELECT_NONE_CONTROLS]));
            this.set(SELECTABLE_LISTING, selectableListing);

            // The currently selected computer id
            this.selectedComputerId = null;
        }

        if (!this.get(RENDER_FIRST_PAGE)) {
            // If we don't render the first page, we parse the HTML content to
            // put it in the cache, so that future navigation don't ask for the
            // content already retrieved
            var firstPageRecords = this._parsePage();
            recordSource.cache.add(
                "?" + pager.getRequestQueryString(),
                {meta: {
                    totalResults: this.get(AVAILABLE_COMPUTER_IDS).length},
                results: firstPageRecords});
        }
    },

    /**
     * Renders the Pager widget and the SelectableListing widget as
     * part of our widget.
     */
    renderUI: function() {
        // Render the widget.
        var pager = this.get(PAGER);
        var contentBox = this.get(CONTENT_BOX);
        var notes = this.get(NOTES);

        // On progressive render we have to bootstrap the widget metadata
        // for initialization.
        if (!this.get(RENDER_FIRST_PAGE)) {
            pager.set("totalRecords", this.get(AVAILABLE_COMPUTER_IDS).length);
            pager.set("records", []);
        }

        // After the page changes, sync our UI state to the new page
        // state.
        pager.on("after-page-change", function() {
            var selectComputerIds = [];
            var unselectComputerIds = [];
            Y.each(pager.get(RECORDS), function(record) {
                if (record.selected) {
                    selectComputerIds.push(record.computer_id);
                } else {
                    unselectComputerIds.push(record.computer_id);
                }
            });
            this._changeSelectedComputers(selectComputerIds,
                                          unselectComputerIds, false);
            this.syncUI(true);
        }, this);

        // Clean up before the page changes
        pager.on("before-page-change", function() {
            // Hide the notes before we switch pages.
            if (notes) {
                notes.hideNotes();
            }
        });

        // Render the SelectableListing widget, if one was created.
        var selectableListing = this.get(SELECTABLE_LISTING);
        if (selectableListing) {
            selectableListing.render();
        }

        // And finally render the pager widget.
        pager.render();
    },

    /**
     * Binds event handlers to the underlying Pager widget and to
     * selection buttons.
     */
    bindUI: function() {
        var pager = this.get(PAGER);
        var selectableListing = this.get(SELECTABLE_LISTING);
        var notes = this.get(NOTES);
        var contentBox = this.get(CONTENT_BOX);

        // If the selectedComputerIds are changed externally, sync our UI.
        this.after(SELECTED_COMPUTER_IDS + "Change", function(evt) {
            // Only call syncUI if value was not set internally.
            if (!evt.details[0].internal) {
                var availableComputerIds = this.get(AVAILABLE_COMPUTER_IDS);
                var isSelected = function(item) {
                    return (Y.Array.indexOf(evt.newVal, item) > -1);
                };
                this.set(UNSELECTED_COMPUTER_IDS,
                         Y.Array.reject(availableComputerIds, isSelected));
                this.syncUI();
            }
        });

        var changeSelected = Y.bind(this._changeSelectedComputers, this);

        if (selectableListing) {
            // Whenever a row is clicked, mark the computer represented by
            // that row as selected.
            selectableListing.on("selectable_listing:mark-row", function(evt) {
                var row = evt.details[0].row;
                var computerId = parseInt(row.one("input").get("value"), 10);
                changeSelected([computerId], null, true);
            });

            // When a row is unmarked, remove the computer id represented
            // by that row from the list of selected computer ids.
            selectableListing.on("selectable_listing:unmark-row", function(evt) {
                var row = evt.details[0].row,
                    computerId = parseInt(row.one("input").get("value"), 10);
                changeSelected(null, [computerId], true);

                // Remove the select all notes if present
                var previous = Y.one(notes.get(CONTENT_BOX))
                    .one("a.select-all-confirm");

                if (previous) { // remove existing note.
                    previous.ancestor("tr").remove();
                }
            }, this);

            var selectAll = Y.bind(function(evt) {
                this.selectAll();
            }, this);

            var selectNone = Y.bind(function(evt){
                this.selectNone();
            }, this);

            // Bind page controls to event handlers.
            Y.each(selectableListing.get(SELECT_ALL_CONTROLS), function(node) {
                node.on("click", selectAll);
            }, this);

            Y.each(selectableListing.get(SELECT_NONE_CONTROLS), function(node) {
                node.on("click", selectNone);
            }, this);
        }
        // Make table sortable via binding click handlers to column headers.
        var headers = pager.get(CONTENT_BOX).all("th.sortable");
        headers.on("click", this.sortSelection, this);
    },

    /**
     * Sync UI state with internal state.
     */
    syncUI: function(pagerChange) {
        pagerChange = typeof pagerChange !== 'undefined' ? pagerChange : false;
        var selectableListing = this.get(SELECTABLE_LISTING),
            selectionInfo = this.get(SELECTION_INFO),
            contentBox = this.get(CONTENT_BOX),
            selectionInfoTitle = contentBox.one(".table-header-info");
        var pager = this.get(PAGER);

        if (selectionInfoTitle.getContent() != selectionInfo) {
            // update the selection title
            selectionInfoTitle.setContent(selectionInfo);
        }

        if (pager.get("totalRecords") === 0 && pagerChange) {
                var div = [
                    "<tr id='no-computer-row'><td colspan=6 align='center'>",
                    this.get(NO_COMPUTERS_MESSAGE), "</td></tr>"].join("");
                var overlayContent = Y.Node.create(div);
                Y.one(".inline-listing").one("tbody").appendChild(overlayContent);
        }

        if (selectableListing) {
            // Walk over the selectable rows and mark the ones matching
            // selectedComputerIds as selected.
            var selectRows = [],
                selectedComputerIds = this.get(SELECTED_COMPUTER_IDS);
            selectableListing.syncUI();

            selectableListing.get(SELECTABLE_ROWS).each(function(row) {
                var computerId = parseInt(row.one("input").get("value"), 10);
                if (Y.Array.indexOf(selectedComputerIds, computerId) > -1) {
                    selectRows.push(row);
                }
            });
            selectableListing.selectRows(selectRows);

            Y.each(selectableListing.get(SELECT_NONE_CONTROLS), function(node) {
                if (selectedComputerIds.length) {
                    node.removeClass("selected-filter");
                }
            });
        }
    },

    /**
     * Parse the current content of the page, returning a record list with the
     * same schema as returned by the server
     *
     * @method _parsePage
     * @private
     */
    _parsePage: function() {
        var records = [];
        var contentBox = this.get(CONTENT_BOX);
        var record, columns, checkbox;

        function getImage(parentNode, image) {
            if (!parentNode) {
                return null;
            }
            var node = parentNode.one("*");
            if (!node) {
                return null;
            }
            if (node.get("tagName").toUpperCase() == "IMG") {
                if (node.getAttribute("src") == image) {
                    return node;
                }
            } else {
                if (node.hasClass(
                       Y.landscape.ui.sprite.formatFilename(image))) {
                    return node;
                }
            }
        }

        contentBox.all("tbody tr").each(function(row) {
            columns = row.all("td");
            checkbox = columns.item(0).one("input");
            var rebootIndex = 5;
            if (columns.size() == 5) {
                rebootIndex = 4;
            }
            reboot_required_img = getImage(
                columns.item(rebootIndex), REBOOT_REQUIRED_ICON);
            if (reboot_required_img === null) {
                reboot_required_flag = false;
                reboot_required_text = "";
            } else {
                reboot_required_flag = true;
                reboot_required_text = reboot_required_img.get("title");
            }
            record = {
                selected: checkbox.get("checked"),
                computer_id: parseInt(checkbox.get("value"), 10),
                title: columns.item(1).one("a").get("text"),
                hostname: columns.item(2).get("text"),
                last_activity: columns.item(3).get("text"),
                reboot_required_flag: reboot_required_flag,
                reboot_required_text: reboot_required_text,
            };
            records.push(record);
        });
        return records;
    },

    /**
     * Change the list of currently selected computer ids by adding or
     * removing a row from the selection.
     *
     * @method _changeSelectedComputers
     * @private
     */
    _changeSelectedComputers: function(selectComputerIds, unselectComputerIds,
                                       force) {
        var chooseOne = this.get(CHOOSEONE);
        var selectedComputerIds = this.get(SELECTED_COMPUTER_IDS);
        var unselectedComputerIds = this.get(UNSELECTED_COMPUTER_IDS);

        // Make a copy of the selectedComputerIds to avoid
        // modifying inplace.
        selectedComputerIds = Y.Array(selectedComputerIds, 0, true);
        unselectedComputerIds = Y.Array(unselectedComputerIds, 0, true);

        if (chooseOne) {
            // Now setup the onclick event bindings for our refreshed computer
            // selection content
            var pager = this.get(PAGER);
            var pagerLinks = pager.get("contentBox").all(
                ".computer-selection-select-one");
            var chooseOneComputer = Y.bind(function(evt){
                this.chooseOneComputer(evt);
            }, this);

            Y.each(pagerLinks, function(link) {
                link.on("click", chooseOneComputer);
            }, this);
        }
        Y.each(selectComputerIds,
               Y.bind(this._changeSelectedComputer,
                       this, selectedComputerIds, unselectedComputerIds,
                       force, SELECTION_ADD));
        Y.each(unselectComputerIds,
               Y.bind(this._changeSelectedComputer,
                       this, selectedComputerIds, unselectedComputerIds,
                       force, SELECTION_REMOVE));

        this.set(SELECTED_COMPUTER_IDS,
                 selectedComputerIds, {internal: true});
        this.set(UNSELECTED_COMPUTER_IDS,
                 unselectedComputerIds, {internal: true});
    },

    _changeSelectedComputer: function(selectedComputerIds,
                                      unselectedComputerIds,
                                      force, operation, computerId) {
        var selectionIndex = Y.Array.indexOf(selectedComputerIds, computerId);
        var unselectionIndex = Y.Array.indexOf(unselectedComputerIds, computerId);
        var selected = selectionIndex > -1;
        var unselected = unselectionIndex > -1;

        // If the computer is not selected, only mark it as selected
        // if forcefully asked to do so or if it was not marked as
        // unselected.
        if (!selected && operation == SELECTION_ADD && (force || !unselected)) {
            selectedComputerIds.push(computerId);
            // If the computer was previously unselected and we've
            // been asked to forcefully select it, remove it from
            // unselected.
            if (unselected) {
                unselectedComputerIds.splice(unselectionIndex, 1);
            }
        }

        // If the computer is not unselected, only mark it as
        // unselected if forcefully asked to do so or if it wasn't
        // marked as selected.
        if (!unselected && operation == SELECTION_REMOVE && (force || !selected)) {
            unselectedComputerIds.push(computerId);
            // If the computer was previously selected and we've been
            // forcefully asked to unselect it, remove it from
            // selected.
            if (selected) {
                selectedComputerIds.splice(selectionIndex, 1);
            }
        }
    },

    /**
     * Select all computers in the current page. If there are more
     * available computers which are not in the current page, add a
     * note to that effect, allowing the user to really select all
     * computers.
     *
     * @method selectAll
     * @public
     */
    selectAll: function() {
        var pager = this.get(PAGER);
        var selectedComputerIds = Y.Array(
            this.get(SELECTED_COMPUTER_IDS), 0, true);
        var unselectedComputerIds = Y.Array(
            this.get(UNSELECTED_COMPUTER_IDS), 0, true);

        this.get(SELECTABLE_LISTING).get(SELECTED_ROWS).each(function(row) {
            var checkbox = row.one("input[type='checkbox']:checked");
            if (checkbox) {
                var computerId = parseInt(checkbox.get("value"), 10);
                if (Y.Array.indexOf(selectedComputerIds, computerId) == -1) {
                    selectedComputerIds.push(computerId);
                }
                var unselectedIndex = Y.Array.indexOf(unselectedComputerIds, computerId);
                if (unselectedIndex > -1) {
                    unselectedComputerIds.splice(unselectedIndex, 1);
                }
            }
        });

        // Store new value for selectedComputerIds.
        this.set(SELECTED_COMPUTER_IDS,
                 selectedComputerIds,
                 {internal: true});
        this.set(UNSELECTED_COMPUTER_IDS,
                 unselectedComputerIds,
                 {internal: true});

        if (selectedComputerIds.length == pager.get("totalRecords")) {
            return;
        }

        var notes = this.get(NOTES);
        var previous = Y.one(
            notes.get(CONTENT_BOX)).one("a.select-all-confirm");

        if (previous) { // remove existing note.
            previous.ancestor("tr").remove();
        }

        notes.addNote("info", Y.substitute(this.SELECT_ALL_TEMPLATE, {
            count: selectedComputerIds.length,
            total: pager.get("totalRecords")
        }));

        // Bind the note link to our confirmation callback.
        var link = Y.one(notes.get(CONTENT_BOX)).one("a.select-all-confirm");
        link.on("click", Y.bind(function(evt) {
            evt.preventDefault();
            this.selectionConfirmAll();
        }, this));
    },

    /**
     * Choose one computer in the current page.
     *
     * @method chooseOneComputer
     * @public
     */
    chooseOneComputer: function(evt) {
        var idTitleSplit = evt.currentTarget.get("value").split("__");
        var computerId = idTitleSplit[0];
        var title = idTitleSplit[1];
        this.fire("computerChosen", {computer_id: computerId,
                                      computer_name: title});
    },

    /**
     * Select a single computer in the listing
     *
     * @method selectOneComputer
     * @public
     */
    selectOneComputer: function(computerId) {
        if (this.get(CHOOSEONE)) {
            this.selectedComputerId = computerId;
        }
    },

    /**
     * Confirm the selection of all computers, hiding notes if they
     * are visible.
     *
     * @method selectionConfirmAll
     * @public
     */
    selectionConfirmAll: function() {
        var notes = this.get(NOTES);
        notes.hideNotes();

        // Mark all available computers as selected, and no computer
        // ids as unselected.
        this.set(SELECTED_COMPUTER_IDS,
                 this.get(AVAILABLE_COMPUTER_IDS),
                 {internal: true});
        this.set(UNSELECTED_COMPUTER_IDS,
                 [], {internal: true});
    },

    /**
     * Unselect all computers.
     *
     * @method selectNone
     * @public
     */
    selectNone: function() {
        var notes = this.get(NOTES);
        notes.hideNotes();

        this.set(SELECTED_COMPUTER_IDS, [],
                 {internal: true});
        this.set(UNSELECTED_COMPUTER_IDS,
                 Y.Array(this.get(AVAILABLE_COMPUTER_IDS), 0, true),
                 {internal: true});
    },

    /**
     * Callback for clicking a sortable column header. Adds sort UI indicator
     * to the column header and set's the pager sort column and direction.
     *
     * @method sortSelection
     * @public
     */
    sortSelection: function(evt) {
        var pager = this.get(PAGER);
        var header = evt.currentTarget;
        evt.preventDefault();

        var sortId = null;
        var sortArrow = null;

        // Remove any existing sort indicators on columns.
        var sortedColumns = header.get("parentNode").all("th.sorted");
        var sortDirection = "asc";

        sortedColumns.each(function(col) {
            if (col.get("id") == header.get("id")) {
                return;
            }
            var forwardIndicator = col.one(COL_SORT_FORWARD_ID);
            var reverseIndicator = col.one(COL_SORT_REVERSE_ID);
            if (forwardIndicator) {
                forwardIndicator.remove();
            }
            if (reverseIndicator) {
                reverseIndicator.remove();
            }
            col.removeClass("sorted");
        });

        // UI indicator to denote sort state
        var forwardIndicator = header.one(COL_SORT_FORWARD_ID);
        var reverseIndicator = header.one(COL_SORT_REVERSE_ID);
        if (forwardIndicator) {
            forwardIndicator.remove();
            sortId = COL_SORT_REVERSE_ID.substring(1);
            sortArrow = COL_SORT_UP_ICON;
            sortDirection = "asc";
        } else {
            if (reverseIndicator) {
                reverseIndicator.remove();
            }
            sortId = COL_SORT_FORWARD_ID.substring(1);
            sortArrow = COL_SORT_DOWN_ICON;
            sortDirection = "desc";
        }

        header.addClass("sorted");
        var span = "<span id='" + sortId + "'>" + sortArrow + "</span>";
        header.appendChild(Y.Node.create(span));

        // Refresh the UI
        var prefixLength = COL_PREFIX.substring(1).length;
        var sortColumn = header.get("id").substring(prefixLength);
        pager.setSort(sortColumn, sortDirection);
    },

    /**
     * Pager custom row renderer function for computers, updates row
     * with backend data. When paging between batches, the row
     * renderer function will be updating rows with existing data.
     */
    rowRender: function(row, record, index, selectable, chooseOne) {
        var column = 0;
        row.set("id", ROW_PREFIX.substring(1) + record.computer_id);

        var columns = row.all("td");
        if (selectable || chooseOne) {
            var checkbox = [
                    "<input type='checkbox' name='selection:list' ",
                    "value='", record.computer_id, "' />"].join("");
            if (chooseOne) {
                var checked = "";
                if (record.computer_id == this.selectedComputerId) {
                    checked = "checked='checked' ";
                }
                checkbox = [
                    "<input type='radio' name='selection:list' ",
                    "class='computer-selection-select-one' ",
                    checked,
                    "value='", record.computer_id, "__",
                    record.title.replace("'",""), "' />"].join("");
            }
            columns.item(column++).set("innerHTML", checkbox);
        } else if (columns.item(0).one("input")){
            columns.item(column++).set("innerHTML", "");
        } else {
            column++;
        }

        var computerHref = [this.get(ACCOUNT_URL), "/computer/",
                            record.computer_id, "/info"].join("");

        var computerLink = Y.Node.create("<a />");
        computerLink.setAttribute("href", computerHref);
        computerLink.setAttribute("alt", "View computer");
        computerLink.setAttribute("title", "View computer");
        computerLink.set("text", record.title);
        columns.item(column++).setContent(computerLink);

        columns.item(column++).set("text", record.hostname);
        columns.item(column++).set("text", record.last_activity);

        if (record.reboot_required_flag) {
            var rebootIcon = Y.Node.create("<img />");
            rebootIcon.setAttribute("src", REBOOT_REQUIRED_ICON);
            rebootIcon.setAttribute("class", "icon");
            rebootIcon.setAttribute("title", record.reboot_required_text);
            rebootIcon.setAttribute("alt", "");
            columns.item(column++).setContent(rebootIcon);
        } else { // reset column
            columns.item(column++).set("innerHTML", "");
        }
    }
});

Y.landscape.selection.ComputerSelection = ComputerSelection;

}, "0.1", {"requires": ["array-extras", "node", "event", "io-base", "cache",
                        "substitute", "datasource-io", "datasource-cache",
                        "datasource-function",
                        "datasource-jsonschema", "landscape.notes",
                        "landscape.listing.selectable",
                        "landscape.ui.image",
                        "landscape.ui.sprite",
                        "landscape.util.node",
                        "landscape.listing.pager"]});
