// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.selection.savedsearch", function(Y) {

Y.namespace("landscape.selection");

var stringToSlug = Y.landscape.util.stringToSlug;

/**
 * This widget shows a modal dialog that lets users create a Saved Search
 * As part of init, you can pass an object containing "criteria" and "id"
 * properties, to allow editing of this search.
 *
 * @class SavedSearch
 * @extends Widget
 * @constructor
 */
var SavedSearch = function(config) {
    SavedSearch.superclass.constructor.apply(this, arguments);
};

SavedSearch.NAME = "SavedSearch";
SavedSearch.ATTRS = {
    /**
     * @attribute io
     * @description The DataSource class to use for creating DataSource objects
     *     when sending actions to the server.
     * @type class
     */
    io: {
        value: Y.DataSource.IO,
        cloneDefaultValue: false
    },

    document: {
        value: Y.config.doc,
        cloneDefaultValue: false
    }
};

Y.extend(SavedSearch, Y.Widget, {

    initializer: function(config) {
        var contentBox = this.get("contentBox");

        this._overlayButton = contentBox.all("#save-search-button, " +
                                             ".edit-search-button");
        this._formBox = contentBox.one("#save-search-form");
        this._titleInput = this._formBox.one("input[name='form.title']");
        this._nameInput = this._formBox.one("input[name='form.name']");
        this._searchInput = this._formBox.one("input[name='form.search']");
        this._saveButton = this._formBox.one(
            "input[name='form.actions.save']");

        this._overlay = new Y.landscape.ui.Dialog({
            contentBox: "#save-search-overlay",
            bodyContent: this._formBox,
            visible : false,
            width : "70%",
            align : {
                points : [Y.WidgetPositionAlign.TC,
                          Y.WidgetPositionAlign.TC]
            },
        });

        var button = this._overlayButton.item(0);
        if (button) {
            this._editMode = button.hasClass("edit-search-button");
        } else {
            this._editMode = false;
        }
    },

    renderUI: function() {
        this._formBox.show();
        this._overlay.render();
    },

    syncUI: function() {
        var contentBox = this.get("contentBox"),
            criteriaInput = contentBox.one("input[name='criteria']");
        if (criteriaInput) {
            this._searchInput.set("value", criteriaInput.get("value"));
        }

        // For the editing form, the name field must be disabled because it's
        // not allowed to change the name of a search.
        // Also, disable name suggestion.
        if (this._editMode) {
            this._nameInput.set("disabled", true);
        }
    },

    bindUI: function() {
        this._overlayButton.on("click", this._onOverlayButtonClick, this);
        this._titleInput.on("keyup", this._onTitleKeyup, this);
        this._nameInput.on("focus", this._onNameFocus, this);
        this._saveButton.on("click", this._onSaveClick, this);
    },

    _onOverlayButtonClick: function(evt) {
        evt.preventDefault();
        if (evt.target.hasClass("listing")) {
            var table = evt.target.ancestor("#saved-searches-table");
            this._titleInput.set("value", table.one(".title").get("text"));
            this._nameInput.set("value", evt.target.get("text"));
            this._searchInput.set("value", table.one(".criteria").get("text"));
        }
        this._overlay.show();
    },

    _onTitleKeyup: function(evt) {
        if (this._editMode) {
            // Don't touch the name field in edit mode
            return;
        }
        var charCode = evt.charCode ? evt.charCode : evt.keyCode;

        // Ignore tabs
        if (charCode != 9) {
            this._nameInput.set("value",
                                stringToSlug(this._titleInput.get("value")));
        }
    },

    _onNameFocus: function(evt) {
        // If the name field gains focus, the suggestion from title must
        // be turned off.
        this._titleInput.detach("keyup", this._onTitleKeyup, this);
        this._nameInput.detach("focus", this._onNameFocus, this);
        this._nameInput.on("keypress", this._onNameKeypress, this);
    },

    _onNameKeypress: function(evt) {
        // Everything that is typed in the name field is
        // converted to a valid slug to prevent mistakes.
        var charCode = evt.charCode ? evt.charCode : evt.keyCode,
            charText;

        // ignore backspace and tab.
        if (charCode != 8 && charCode != 9) {
            evt.preventDefault();
            charText = stringToSlug(String.fromCharCode(charCode));
            this._nameInput.set("value",
                                this._nameInput.get("value") + charText);
        }
    },

    _onSaveClick: function(evt) {
        evt.preventDefault();
        var form = this._formBox.one("form"),
            cfg = {
                method: "POST",
                form: {id: form.get("id"), useDisabled: true}
            },
            IO = this.get("io"),
            source = new IO({source: form.get("action")});

        source.sendRequest({on: {success: Y.bind(this._handleSuccess, this),
                                 failure: Y.bind(this._handleFailure, this)},
                            cfg: cfg});
    },

    _handleFailure: function(evt) {
        this._formBox.empty();
        this._formBox.setHTML(evt.data.responseText);
        this._overlay.setStdModContent(Y.WidgetStdMod.BODY,
                                       this._formBox,
                                       Y.WidgetStdMod.REPLACE);

        // XXX we have to install the event listeners again because the nodes
        // have been replaced
        this._titleInput = this._formBox.one("input[name='form.title']");
        this._titleInput.on("keyup", this._onTitleKeyup, this);
        this._nameInput = this._formBox.one("input[name='form.name']");
        this._nameInput.on("keypress", this._onNameKeypress, this);
        if (this._editMode) {
            this._nameInput.set("disabled", true);
        }
        this._searchInput = this._formBox.one("input[name='form.search']");
        this._saveButton = this._formBox.one(
            "input[name='form.actions.save']");
        this._saveButton.on("click", this._onSaveClick, this);
    },

    _handleSuccess: function(evt) {
        var document = this.get("document");
        var nextLocation = document.location.href;
        if (!this._editMode) {
            var name = this._nameInput.get("value");
            if (!name) {
                name = stringToSlug(this._titleInput.get("value"));
            }
            nextLocation += "/search/" + name;
        }
        document.location = nextLocation;
    }
});

Y.landscape.selection.SavedSearch = SavedSearch;

}, "0.1", {"requires" : ["node", "base", "io-form", "datasource-io", "widget",
                         "landscape.ui.dialog", "landscape.util.text",
                         "landscape.notes"]});
