// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.table.formatters", function (Y) {

Y.namespace("landscape.table");

var CHECKBOX_TEMPLATE =
        "<label>" +
        "<input type='checkbox' name='{name}'" +
        " value='{value}'{checked}{disabled}>" +
        "<span>{label}</span>" +
        "</label>";

Y.landscape.table.CHECKBOX_TEMPLATE = CHECKBOX_TEMPLATE;

var STATUS_ICON_TEMPLATE =
        "<span class='activity-status activity-status-{tag}'></span>",
    INFO_ICON_TEMPLATE =
        "<span class='link-button {iconClasses}' data-tooltip='{tooltip}'></span>";

var Formatters = {

    /**
     *  Return a formatter for a column that contains an <input> element
     *  of type checkbox.
     *
     *  The formatter expects the column value (o.value) to be a boolean value
     *  indicating whether the row is currently selected. It will set checkbox
     *  state according to it.
     *
     *  If the column value of is null, then the checkbox will be unchecked
     *  and disabled. Alternatively if the model as a 'disabled' attribute
     *  set to true, the checkbox will be disabled as well (either in checked
     *  or unchecked state, depending on the column value).
     *
     *  If the column has a checkboxLabel string attribute, it's used as label.
     *  If it's a function, it's called with the column and the object and should
     *  return a string, which is used as label.
     *
     *  The value attribute of the <input> element will be set to the ID of
     *  the row model (see Y.Model).
     *
     *  The name attribute of the <input> element can be specified via the
     * 'checkboxName' column config property (it will default to the column key).

     *  The 'selected' model attribute is conventionally used for row selection,
     *  in that case additional CSS classes will be set on the row to allow for
     *  row-level styling.
     */
    checkbox: function (col) {
        col.allowHTML = true;
        var name = col.checkboxName || col.key;
        var checkboxLabel = col.checkboxLabel;
        if (! Y.Lang.isFunction(checkboxLabel)) {
            checkboxLabel = function() {
                return col.checkboxLabel || "";
            };
        }

        return function(o) {
            var disabled = Y.Lang.isNull(o.value) || o.data.disabled;
            var html = Y.Lang.sub(CHECKBOX_TEMPLATE, {
                checked: o.value ? " checked" : "",
                disabled: disabled ? " disabled" : "",
                value: o.data.id || o.record.get("id"),
                name: name,
                label: checkboxLabel(col, o)});

            o.className = "tight";
            o.rowClass = "";
            if (col.key == "selected") {
                if (o.value) {
                    o.rowClass += "yui3-table-row-selected ";
                }
                if (disabled) {
                    o.rowClass += "yui3-table-row-disabled ";
                }
                if (o.data.inUse) {
                    o.rowClass += "yui3-table-row-in-use ";
                }
            }
            return html;
        };
    },

    /**
     * Return a simple formatter to convert bytes into gigabytes
     * (1 GB = 1000^3 bytes) and present in a human friendly form.
     */
    gigabyte: function (col) {
        return function(o) {
            var gigabytes = o.value / Math.pow(1000.0, 3);
            if (gigabytes >= 100) {
                return gigabytes.toFixed();
            } else {
                return gigabytes.toFixed(1);
            }
        };
    },

    /**
     * Convert bytes into a human-readable string.
     *
     * The default is to use the decimal conversion, for example:
     *
     * 1000^3 => '1G'
     *
     * The binary conversion can be used by setting the 'bytesConversion'
     * property to 'binary' in the column object, and that would 
     * result in:
     *
     * 1024^3 => '1GiB'
     */
    bytes: function (col) {
        return function(o) {
            return Y.landscape.util.bytesToString(o.value, col.bytesConversion);
        };
    },

    /**
     * Return a simple formatter to convert bytes into gibibytes
     * (1 GiB = 1024^3 bytes) and present in a human friendly form.
     */
    gibibyte: function (col) {
        return function(o) {
            var gibibytes = o.value / Math.pow(1024.0, 3);
            if (gibibytes >= 100) {
                return gibibytes.toFixed();
            } else {
                return gibibytes.toFixed(1);
            }
        };
    },

    /**
     * Return a formatter for rendering the icon of an activity status.
     */
    statusIcon: function(col) {
        col.allowHTML = true;
        return function(o) {
            var title = o.data.status,
                tag = title.replace(" ", "-").toLowerCase();
            return Y.Lang.sub(STATUS_ICON_TEMPLATE, {tag: tag});
         };
    },

    /**
     * Return a formatter for rendering the text of an activity status.
     */
    statusText: function(col) {
        col.allowHTML = true;
        return function(o) {
            var title = o.data.status;
            if (o.data.result) {
                return "<a href='#'>" + title + "</a>";
            } else {
                return title;
            }
         };
    },

    /**
     * Return a formatter for rendering an information icon.
     *
     * A link-button icon will be rendered if the data includes an
     * 'infoIconClasses' attribute.
     * 
     * Also, if the data has a 'tooltip' attribute, it will be added as
     * a data-tooltip attribute on the tag, for the Tooltip widget to
     * use.
     */
    infoIcon: function(col) {
        col.allowHTML = true;
        return function(o) {
            var result = o.value;
            if (o.data.infoIconClasses) {
                result += Y.Lang.sub(INFO_ICON_TEMPLATE,{
                    tooltip: o.data.tooltip ? o.data.tooltip : "",
                    iconClasses: o.data.infoIconClasses});
            }
            return result;
         };
    },

    /**
     * Return a string with comma-separated elements from a list.
     */ 
    list: function(col) {
        return function(o) {
            return o.value.join(", ");
        };
    }
};

Y.mix(Y.DataTable.BodyView.Formatters, Formatters);

}, "0.1", {"requires": ["datatable-body", "landscape.util.text"]});
