// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.table.selectable", function(Y) {

Y.namespace("landscape.table");

/**
 * DataTable support for (bulk) row or cell selection.
 *
 * This extension synchronizes the checked state of a cell's <input> checkbox
 * with the associated attribute of its model.
 *
 * It also supports bulk cell selection by clicking the checkbox input of the
 * column.
 *
 * The attribute named 'selected' is conventionally used for row-level
 * selection as opposed to cell-level selection. The only difference here
 * is that appropriate CSS classes will be set and unset also at the row (TR)
 * level, rather than only at the cell one (e.g. 'yui3-table-row-selected').
 *
 * @module landscape.table.selectable
 **/

var CELL_CHECKBOX = ".yui3-table-cell input[type='checkbox']:not(:disabled)",
    HEADER_CHECKBOX = ".yui3-table-header input[type='checkbox']";

var Selectable = function() {};

Y.mix(Selectable.prototype, {

    bindUI: function() {
        this.publish("select", {
            emitFacade: true,
            defaultFn: this._defSelectFn,
        });

        // Listen for model changes of each attribute associated with a checkbox
        // column.
        Y.Array.each(this.get("selectableColumns"), function(column) {
            this.after("*:" + column.key + "Change", this._afterModelChange, this);
        }, this);

        this.after("table:renderHeader", this._afterRenderHeader, this);
        this.after("table:renderBody", this._afterRenderBody, this);
        this.delegate("click", this._onCellCheckboxClick, CELL_CHECKBOX, this);
        this.delegate("click", this._onHeaderCheckboxClick, HEADER_CHECKBOX, this);
    },

    /**
     * Set the header checkbox for the given column depending on whether all items
     * are selected.
     *
     * Ignores disabled items.
     */
    _uiSetHeaderCheckbox: function(column) {
        var key = column.key,
            headerSelector = ".yui3-table-header.yui3-table-col-" + key + " input[type='checkbox']",
            headerCheckbox = this.view.tableNode.one(headerSelector),
            headerCell,
            data = this.get("data"),
            // Are there some unselected items which are not disabled? (an item is considered disabled
            // either if it's 'disabled' attribute is true, or the column value is null).
            someUnselected = data.some(function(item) {
                var isSelected = item.get(key) === true,
                    isDisabled = item.get("disabled") === true || Y.Lang.isNull(item.get(key));
                return !(isSelected || isDisabled);
            });

        if (!Y.Lang.isNull(headerCheckbox)) {
            // Re-render the checkbox, because it's not possible to call
            // preventDefault() AND then actually set the checked attribute
            // again.
            headerCell = headerCheckbox.ancestor("th");
            headerCell.setHTML(column.label);
            headerCheckbox = headerCell.one("input");
            headerCheckbox.set("checked", ! someUnselected);
        }
    },

    /**
     * After the header view is rendered, refresh the header checkbox for
     * the selectable columns.
     */
    _afterRenderHeader: function(e) {
        Y.Array.each(this.get("selectableColumns"), function(column) {
            this._uiSetHeaderCheckbox(column);
        }, this);
    },

    /**
     * Handle changes of the selectable attributes in the model items.
     */
    _afterModelChange: function(e) {
        if (e.target.get("disabled") === true) {
            // Disallow selecting a disabled item.
            // XXX: this might be better suited to a Y.Model-based class,
            // since it's pretty generic to any model.
            throw new Error(
                "You can't change 'selected' value on the disabled row.");
        }
        this._uiSetHeaderCheckbox(this.getColumn(e.attrName));
    },

    /**
     * Handler for triggering post-render behavior.
     */
    _afterRenderBody: function(e) {
        Y.Do.after(this._afterRefreshRow, this.view.body, "refreshRow", this);
    },

    /**
     * When a checkbox cell is refreshed after a model change, also refresh the
     * marker class on the parent row.
     */
    _afterRefreshRow: function(row, model, colKeys) {
        if (colKeys.indexOf("selected") >= 0) {
            row.toggleClass("yui3-table-row-selected", model.get("selected"));
        }
    },

    /**
     * Handle mouse clicks on a row's <input> checkbox.
     */
    _onCellCheckboxClick: function(e) {
        var checkbox = e.target,
            checked = checkbox.get("checked"),
            key = this._getColumnKey(checkbox),
            record = this.getRecord(checkbox.get("id"));

        // This is just for sake of precision. It will prevent the browser from
        // toggling the checkbox input, since it will be done anyways by the
        // checkbox row formatter, which will be invoked by DataTable to
        // refresh the row, since the "selected" property of the associated
        // model has changed.
        e.preventDefault();

        this.fire("select", {key: key, models: [record], value: checked});
    },

    /**
     * Handle mouse clicks on a header's <input> checkbox.
     */
    _onHeaderCheckboxClick: function(e) {
        var checkbox = e.target,
            checked = checkbox.get("checked"),
            data = this.get("data"),
            key = this._getColumnKey(checkbox),
            models;

        e.preventDefault();

        // Collect all models that are not disabled.
        models = data.filter(function(model) {
            return !(model.get("disabled") || Y.Lang.isNull(model.get(key)));
        });

        this.fire("select", {key: key, models: models, value: checked});
    },

    // Figure the column key for this checkbox (in case the checkbox name differs
    // from the column name because of the checkboxName knob).
    _getColumnKey: function(checkbox) {
        var columns = this.get("selectableColumns"),
            name = checkbox.get("name"),
            column;

        column = Y.Array.find(columns, function(col) {
            return col.key === name || col.checkboxName === name;
        });

        return column.key;
    },

    /**
     * Default handler for the 'select' event, modifying the model directly
     * by setting the 'selected' attribute as appropriate. It can be
     * prevented in order to implement custom controller logic.
     */
    _defSelectFn: function(e) {
        Y.Array.invoke(e.models, "set", e.key, e.value);
    },

});

Selectable.ATTRS = {

    // Return all columns that are selectable, i.e. that use the checkbox
    // formatter.
    selectableColumns: {
        getter: function() {
            return Y.Array.filter(this.get("columns"), function(column) {
                return column.formatter === "checkbox";
            }, this);
        }
    }
};

Y.landscape.table.Selectable = Selectable;

}, "0.1", {"requires": ["datatable-base", "landscape.table.formatters"]});
