// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.table.total", function(Y) {

Y.namespace("landscape.table");

/**
 * DataTable support for column totals in footer.
 *
 * This extension updates the counts in column footers based on cell values.
 *
 * If a cell in the table footer contains an element with the "counter-number"
 * class, its content will be updated with the sum of values for that column in
 * the table.
 *
 * If a cell in the table footer contains an element with
 * the "counter-selected" class, its content will be updated with the sum
 * values for that column but only for rows which have attribute "selected"
 * set to true.
 *
 * If column configuration has a "fixed_total" property set, the sum of the
 * values in the column will be used as the "selected" count, and the provided
 * "fixed_total" will be used as the "total" count.
 *
 * When column totals are updated (as a consequence of a table model change),
 * the "change" event is fired, with a list of objects with column name,
 * selected and total counts as parameter.
 *
 * A typical footer cell content looks like
 *
 * <div class="total">
 *   <div class="total-selected">0</div>
 *   <div class="total-available">0</div>
 *   <div class="total-caption">total</div>
 * </div>
 *
 * @module landscape.table.total
 **/

var Total = function() {};

var Formatters = Y.DataTable.BodyView.Formatters;

Y.mix(Total.prototype, {

    initializer: function() {
        this.after("dataChange", this._afterModelChanged, this);
        this.after("render", this._afterModelChanged, this);
        this.after("model:change", this._afterModelChanged, this);
        Y.Array.each(["init", "load", "add", "remove", "reset"], function(eventName) {
            this.after("modelList:" + eventName, this._afterModelChanged, this);
        }, this);
    },

    /* Return totals for each data column, and the count of selected rows. */
    getTotals: function() {
        var totals = {},
            countableColumns = this._countableColumns(),
            tableData = this.get("data"),
            selected = tableData.filter(
                function(rowData) {
                    return rowData.get("selected");
                }).length;

        Y.Array.each(this.get("columns"), function(column, index) {
            if (countableColumns.indexOf(index) == -1) {
                // The column is not countable, skip it.
                return;
            }

            var values = tableData.get(column.key),
                columnTotals = {
                    total: Y.Array.reduce(values, 0, function(a, b) {
                        return parseFloat(a) + parseFloat(b); }),
                    selected: 0};

            tableData.each(function(rowData) {
                if (rowData.get("selected")) {
                    columnTotals.selected += parseFloat(rowData.get(column.key));
                }
            });

            if (! Y.Lang.isUndefined(column.fixed_total)) {
                columnTotals.selected = columnTotals.total;
                columnTotals.total = column.fixed_total;
            }

            totals[column.key] = columnTotals;
        }, this);

        return {selected: selected, totals: totals};
    },

    /* Event handler to update column totals when the table model changes. */
    _afterModelChanged: function(event) {
        var tfoot = this.get("contentBox").one("tfoot");
        if (Y.Lang.isNull(tfoot)) {
            return;
        }

        var tableTotals = this.getTotals(),
            tdNodes = tfoot.all("td"),
            fireEvent = false;

        Y.Array.each(this.get("columns"), function(column, index) {
            var totals = tableTotals.totals[column.key],
                formattedTotals = Y.clone(totals);
            if (Y.Lang.isUndefined(totals)) {
                return;
            }

            // Don't try to format columns with fixed footers
            if (Y.Lang.isUndefined(column.footerHtml)) {
                this._formatColumnTotals(column, formattedTotals);
            }

            if (this._updateTotals(tdNodes.item(index), formattedTotals)) {
                fireEvent = true;
            }
        }, this);

        if (fireEvent) {
            this.fire("totals-change", tableTotals);
        }
    },

    /**
     * Format total counters using the column formatter if defined, otherwise
     * leave them unchanged.
     */
    _formatColumnTotals: function(column, totals) {
        var formatterFn = null;
        if (Y.Lang.isFunction(column.formatter)) {
            formatterFn = column.formatter;
        } else if (Y.Lang.isString(column.formatter)) {
            formatterFn = Formatters[column.formatter](column);
        }

        if (! Y.Lang.isNull(formatterFn)) {
            try {
                totals.selected = formatterFn({value: totals.selected});
                totals.total = formatterFn({value: totals.total});
            } catch (e) {
                // If formatting fails for whatever reason
                // (because formatter function uses more
                // than just o.value for example, then do
                // nothing.
            }
        }

        return totals;
    },

    /**
     * Update values for totals in the provided node.
     *
     * @return Whether any counter has been updated.
     */
    _updateTotals: function(node, totals) {
        var selectedNode = node.one(".total-selected"),
            counterNode = node.one(".total-available"),
            updated = false;

        if (! Y.Lang.isNull(counterNode)) {
            counterNode.set("text", totals.total);
            updated = true;
        }
        if (! Y.Lang.isNull(selectedNode)) {
            selectedNode.set("text", totals.selected);
            updated = true;
        }

        return updated;
    },


    /**
     * Return indexes for countable columns.
     */
    _countableColumns: function() {
        var indexes = [];

        var tfoot = this.get("contentBox").one("tfoot");
        if (! Y.Lang.isNull(tfoot)) {
            tfoot.all("td").each(function(node, index) {
                var countable = (
                    node.all(".total-selected, .total-available").size() > 0);
                if (countable) {
                    indexes.push(index);
                }
            });
        }

        return indexes;
    },

});

Y.landscape.table.Total = Total;

}, "0.1", {"requires": ["datatable-base", "datatable-body"]});
