// Copyright 2005-2013 Canonical Limited.  All rights reserved.


YUI.add("landscape.testing", function(Y) {

/**
 * Testing utilities.
 *
 * @module landscape.testing
 * @namespace landscape
 */

Y.namespace("landscape.testing.Runner");

Y.landscape.testing.Runner.add = function(suite) {
    Y.Test.Runner.add(suite);
};

Y.landscape.testing.Runner.run = function(suite) {
    var yconsole = new Y.Console({
        newestOnTop: false,
        useBrowserConsole: true
    });
    yconsole.render("#log");
    Y.Test.Runner.run(suite);
};

}, "0.1", {requires: ["test", "console"]});

YUI.add("landscape.testing.node", function(Y) {

Y.namespace("landscape.testing.node");

/**
 * A helper object to make it easier to make assertions about Y.Node objects.
 *
 * @class NodeAssertion
 * @extends Base
 * @constructor
 */
var NodeAssertion = function(config) {
    NodeAssertion.superclass.constructor.apply(this, arguments);
};

NodeAssertion.NAME = "node_assertion";

NodeAssertion.ATTRS = {
    node: {
        value: undefined
    },

    name: {
        value: undefined
    }
};

Y.extend(NodeAssertion, Y.Base, {
    /**
     * An assertion fails if the Y.Node under test is missing one of the
     * specified CSS classes.
     *
     * @method hasClasses
     */
    hasClasses: function() {
        var node = this.get("node");
        var name = this.get("name");
        for (var i = 0; i < arguments.length; ++i) {
            var className = arguments[i];
            Y.Assert.isTrue(node.hasClass(className),
                            name + " needs " + className + " CSS class.");
        }
        return this;
    },

    /**
     * An assertion fails if the Y.Node under test has one of the specified CSS
     * classes.
     *
     * @method notHasClasses
     */
    notHasClasses: function() {
        var node = this.get("node");
        var name = this.get("name");
        for (var i = 0; i < arguments.length; ++i) {
            var className = arguments[i];
            Y.Assert.isFalse(
                node.hasClass(className),
                name + " shouldn't have " + className + " CSS class.");
        }
        return this;
    },

    /**
     * An assertion fails if the Y.Node under test is missing one of the
     * specified key/value attributes.  It also fails if a matching attribute
     * key is present with an unexpected value.  Y.Node.getAttribute is used to
     * retrieve the value from the node for each specified key.  If it returns
     * null, an attempt is made to get the value with Y.Node.get.
     *
     * @method hasAttributes
     */
    hasAttributes: function(expectedItems) {
        var node = this.get("node");
        var name = this.get("name");
        for (var key in expectedItems) {
            if (expectedItems.hasOwnProperty(key)) {
                var expectedValue = expectedItems[key];
                var actualValue = node.getAttribute(key);
                if (actualValue === null) {
                    actualValue = node.get(key);
                }
                Y.Assert.areNotSame(undefined, actualValue,
                                    key + " attribute is missing for " + name);
                Y.Assert.areEqual(
                    expectedValue, actualValue,
                    key + " attribute has the wrong value for " + name);
            }
        }
        return this;
    },

    /**
     * An assertion fails if the Y.Node under test is missing one of the
     * specified key/value CSS properties.  If also fails if a matching CSS
     * property is present with an unexpected value.
     *
     * @method hasStyles
     */
    hasStyles: function(expectedProperties) {
        var node = this.get("node");
        var name = this.get("name");
        for (var style in expectedProperties) {
            if (expectedProperties.hasOwnProperty(style)) {
                var expectedValue = expectedProperties[style];
                var actualValue = node.getStyle(style);
                Y.Assert.areNotSame(
                    undefined, actualValue,
                    style + " CSS property is missing for " + name);
                Y.Assert.areEqual(
                    expectedValue, actualValue,
                    style + " CSS property has the wrong value for " + name);
            }
        }
        return this;
    },

    /**
     * An assertion fails if the Y.Node under test has HTML content that
     * differs from the expected HTML content.
     *
     * @method hasInnerHTML
     */
    hasInnerHTML: function(expectedHTML) {
        var node = this.get("node");
        var name = this.get("name");
        Y.Assert.areEqual(expectedHTML, node.get("innerHTML"),
                          name + " has unexpected inner HTML.");
        return this;
    },

    /**
     * An assertion fails if the Y.Node under test has HTML content that is the
     * same as the expected HTML content.
     *
     * @method notHasInnerHTML
     */
    notHasInnerHTML: function(unexpectedHTML) {
        var node = this.get("node");
        var name = this.get("name");
        Y.Assert.areNotEqual(unexpectedHTML, node.get("innerHTML"),
                             name + " has unexpected inner HTML.");
        return this;
    },

    /**
     * An assertion fails if the Y.Node under test doesn't have the expected
     * number of child nodes.
     *
     * @method hasChildCount
     */
    hasChildCount: function(expectedCount) {
        var node = this.get("node");
        var name = this.get("name");
        Y.Assert.areEqual(
            expectedCount, node.get("children").size(),
            name + " has unexpected number of child nodes.");
        return this;
    }
});

Y.Assert.checkNode = function(node, name) {
    Y.Assert.areNotSame(null, node, "The " + name + " node is null.");
    return new NodeAssertion({node: node, name: name});
};

// A wrapper for the Y.Event.simulate() function.  The wrapper accepts
// CSS selectors and Node instances instead of requiring raw nodes.
Y.landscape.testing.node.simulate = function(selector, eventType, options) {
    var node = Y.one(selector);
    var rawNode = Y.Node.getDOMNode(node);
    Y.Event.simulate(rawNode, eventType, options);

    // When clicking on an 'option' node, the 'select' node above it
    // doesn't automatically receive the 'change' event, so let's
    // simulate it here.
    if (node.get("tagName").toUpperCase() == "OPTION" && eventType == "click") {
        var select = node.ancestor("select");
        select.set("selectedIndex", node.get("index"));
        Y.Event.simulate(Y.Node.getDOMNode(select), "change");
    }
};

Y.landscape.testing.node.NodeAssertion = NodeAssertion;

}, "0.1", {requires: ["node", "event", "event-simulate"]});


YUI.add("landscape.testing.pager", function(Y) {

Y.namespace("landscape.testing.pager");


// Get the value for 'key' from 'url' or undefined if the key is not present
// in the URL.
function parseQueryString(url, key) {
    var valueBegin = url.indexOf(key);
    if (valueBegin == -1) {
        return undefined;
    }
    valueBegin = valueBegin + key.length + 1;

    var valueEnd = url.substring(valueBegin).indexOf("&");
    valueEnd = (valueEnd == -1) ? url.length : valueBegin + valueEnd;

    return url.substring(valueBegin, valueEnd);
}


// Get an object with 'offset', 'size', 'sortKey' and 'sortType' keys from
// 'url'.
function parseURL(url) {
    return {offset: parseInt(parseQueryString(url, "offset"), 10),
            size: parseInt(parseQueryString(url, "size"), 10),
            sortKey: parseQueryString(url, "sort"),
            sortType: parseQueryString(url, "dir")};
}


// Get a sample data set using data copied from 'templateDataSet' using
// configuration values encoded in 'url'.
function getSampleData(url, templateDataSet) {
    var dataSet = Y.clone(templateDataSet);
    var config = parseURL(url);
    if (config.sortKey !== undefined) {
        dataSet.records.sort(function(x, y) {
            if (x[config.sortKey] < y[config.sortKey]) {
                return config.sortType == "desc" ? 1 : -1;
            } else if (x[config.sortKey] > y[config.sortKey]) {
                return config.sortType == "desc" ? -1 : 1;
            } else {
                return 0;
            }
        });
    }

    if (config.offset) {
        dataSet.records = dataSet.records.slice(config.offset,
                                                config.offset + config.size);
    }
    return dataSet;
}


/* Create a data source that can be used with a Pager in unit tests.  The new
 * data source expects to be used with a URL that includes values for
 * 'offset', 'size' and optionally 'sort' and 'dir' keys in its query string.
 *
 * The result parameter should be the list of items to use when generating
 * results from the data source.  The supported keys are 'sampleName' and
 * 'sampleDescription'.
 * 
 * The totalResults parameter is optional, and should be the number that is
 * returned when the data source is asked how many records it has.
 *
 * @function createDataSource
 */
    function createDataSource(result, totalRecords) {
	var totalResults = result.length;
	if (! Y.Lang.isUndefined(totalRecords)) {
	    totalResults = totalRecords;
	}
        var templateDataSet = {records: result, totalRecords: totalResults};
        var metaFields = {totalResults: "totalRecords"};
        var dataSource = new Y.DataSource.Function({
            source: function(url) {
                var dataSet = getSampleData(url, templateDataSet);
                return {responseText: Y.JSON.stringify(dataSet)};
            }});
        dataSource.plug({
            fn: Y.Plugin.DataSourceJSONSchema,
            cfg: {schema: {metaFields: metaFields,
                           resultListLocator: "records",
                           resultFields: [{key: "sampleName"},
                                          {key: "sampleDescription"}]}}
    });
    return dataSource;
}

Y.landscape.testing.pager.createDataSource = createDataSource;


}, "0.1", {requires: ["datasource-function", "datasource-jsonschema"]});



YUI.add("landscape.testing.datasource", function(Y) {

Y.namespace("landscape.testing.datasource");

var TestablePageDataSource = function() {
    this.init.apply(this, arguments);
};

TestablePageDataSource.prototype = {

    init: function() {
        this.reset();
        this.page_key = "my-page-key";
    },

    /**
     *  Reset all registered results.
     */
    reset: function() {
        this._data = {};
        this._subscribers = {};
        this._actions = {};
        this._deferreds = {};
    },

    /**
     * Set the result of the deferred action.
     *
     * If result is an Error, an errback is added, otherwise a callback.
     *
     * Nothing is done if no deferred is registered for the action, or
     * if the result is undefined.
     */
    _setDeferredResult: function(action, result) {
        var deferred = this._deferreds[action];
        if (Y.Lang.isUndefined(deferred) || Y.Lang.isUndefined(result)) {
            return;
        }
        if (deferred.called) {
            deferred = new Y.landscape.deferred.Deferred();
            this._deferreds[action] = deferred;
        }
        if (result instanceof Error) {
            deferred.errback(result);
        }
        else {
            deferred.callback(result);
        }
    },

    registerResult: function(action, result) {
        this._data[action] = result;
        this._setDeferredResult(action, result);
        var subscriber = this._subscribers[action];
        if (subscriber && !(result instanceof Error)) {
            subscriber(result);
        }
    },

    registerAction: function(action, config, result) {
        this._actions[action] = config;
        this.registerResult(action, result);
    },

    setHandler: function(action) {
        this._deferreds[action] = new Y.landscape.deferred.Deferred();
        var data = this._data[action];
        this._setDeferredResult(action, data);
        return this._deferreds[action];
    },

    sendAction: function(config) {
        var action = config.action;
        if (this._actions[action]) {
            var data = this._actions[action].data;
            var form = this._actions[action].form;
            for (var key in data) {
                if (data.hasOwnProperty(key)) {
                    Y.Assert.areEqual(data[key], config.data[key]);
                }
            }
            Y.Assert.areEqual(form, config.form);
        }
        return this.setHandler(action);
    }
};

Y.landscape.testing.datasource.TestablePageDataSource = TestablePageDataSource;


/**
 * A helper for testing objects that normally use a Y.DataSource.IO.
 *
 * It doesn't try to emulate a real Y.DataSource.IO object. It can send
 * only a single request, emulating either a success or failure.
 *
 * It saves the config its constructor receives.
 */
var TestableIO = function(config) {
    this.config = config;
};

TestableIO.prototype = {

    /**
     * Simulate sending the request to the server.
     *
     * Nothing is done, except for saving the request parameter for later
     * inspection.
     */
    sendRequest: function(request) {
        this.request = request;
    },

    /**
     * Create a request
     */
    _createRequest: function(statusCode, responseText) {
        return {data: {status: statusCode, responseText: responseText}};
    },

    /**
     * Simulate a success by calling the success callback with a 200
     * request.
     */
    simulateSuccess: function(responseText) {
        this.request.callback.success(this._createRequest(200, responseText));
    },

    /**
     * Simulate a failuer by calling the failure callback with a request
     * having the given status code.
     */
    simulateFailure: function(statusCode, responseText) {
        this.request.callback.failure(
            this._createRequest(statusCode, responseText));
    }
};

Y.landscape.testing.datasource.TestableIO = TestableIO;

}, "0.1", {requires: ["landscape.deferred"]});
