// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.button", function(Y) {

Y.namespace("landscape.ui");


/**
 * This module provides the ActiveButton widget.  It manages the
 * enabled/disabled state of a button based on external events.
 *
 * @module landscape.ui.button
 */


/**
 * ActiveButton manages the enabled/disabled state of a button based on
 * external events.
 *
 * @class ActiveButton
 * @extends Widget
 * @constructor
 */
var ActiveButton = function() {
    ActiveButton.superclass.constructor.apply(this, arguments);
};

ActiveButton.NAME = "active_button";

ActiveButton.ATTRS = {
    /**
     * A list of objects with a 'target' key that has the widget to hook on,
     * 'enableEvent' is the name of the event that will fire to indicate that
     * the button should be enabled and 'disableEvent' is the name of the event
     * that will fire to indicate that the button should be disabled.
     *
     * @attribute eventTargets
     * @type Array
     */
    eventTargets: {
        value: null
    },

    /**
     * The button to manage.
     *
     * @attribute button
     * @type Node
     */
    button: {
        value: null
    },

    /**
     * A dict that maps source names to a boolean enabled/disabled status.
     *
     * @attribute statuses
     * @type Object
     */
    statuses: {
        value: null
    },

    /**
     * The tab index is null to prevent focusing of the content box when the
     * widget is used.
     *
     * @attribute tabIndex
     * @type Int
     */
    tabIndex: {
        value: null
    }
};

Y.extend(ActiveButton, Y.Widget, {
    /**
     * Register event handlers for controls that implement active button
     * behaviour.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        var eventTargets = this.get("eventTargets");
        var enableEvent = function(evt) {
            this._getStatuses()[evt.target] = true;
            this.syncUI();
        };
        var disableEvent = function(evt) {
            this._getStatuses()[evt.target] = false;
            this.syncUI();
        };
        for (var i = 0; i < eventTargets.length; ++i) {
            var data = eventTargets[i];
            data.target.on(data.enableEvent, Y.bind(enableEvent, this));
            data.target.on(data.disableEvent, Y.bind(disableEvent, this));
        }
    },

    /**
     * Synchronize the user interface with the widget.
     *
     * @method syncUI
     * @protected
     */
    syncUI: function() {
        var disabled = this.isDisabled();
        this.get("button").set("disabled", disabled);
    },

    /**
     * Returns true if the widget is (or should be) disabled.
     *
     * The button is enabled if all registered targets have an enabled status.
     * If any one of them has a disabled status, the button is disabled.
     *
     * @method isDisabled
     * @public
     */
    isDisabled: function() {
        var statuses = this._getStatuses();
        for (var target in statuses) {
            if (statuses.hasOwnProperty(target) && ! statuses[target]) {
                return true;
            }
        }
        return false;
    },

    _getStatuses: function() {
        if (this.get("statuses") === null) {
            this.set("statuses", {});
        }
        return this.get("statuses");
    }
});

Y.landscape.ui.ActiveButton = ActiveButton;


}, "0.1", {"requires": ["event", "node", "widget"]});
