// Copyright 2005-2014 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.dropdown", function(Y) {

Y.namespace("landscape.ui");

/**
 * z-index to add to the current z-index in order to put the element to the
 * front.
 */
var ZINDEX_OFFSET = 10000;

/**
 * A single item in a ListBox.
 *
 * @class ListBoxItem
 * @extends Widget
 * @constructor
 */
var LISTBOX_ITEM_NAME = "listboxitem";

var LISTBOX_ITEM_ATTRS = {

    /**
     * Description text for for the 'title' property of the listbox item link.
     */
    title: {
        value: null
    },

    /**
     * Text for for the content of the listbox item link.
     */
    value: {
        value: null
    }
};

var LISTBOX_ITEM_PROTOTYPE = {

    BOUNDING_TEMPLATE: "<li></li>",
    CONTENT_TEMPLATE: "<a></a>",

    /*
     * Render the 'li' node.
     */
    renderUI: function() {
        var contentBox = this.get("contentBox"),
            label = contentBox;

        label.setContent(this.get("title"));
    },

    /*
     * Re-sync the UI on attribute change events.
     */
    bindUI: function() {
        this.after("titleChange", this.syncUI);
        this.on("click", this._onClick, this);
    },

    /*
     * Sync the contents of the listbox item.
     */
    syncUI: function() {
        var contentBox = this.get("contentBox"),
            label = contentBox;
        label.setContent(this.get("title"));
    },

    /*
     * Mark the item as selected.
     */
    _onClick: function(e) {
        this.set("selected", 1);
    }

};

var ListBoxItem = Y.Base.create(
    LISTBOX_ITEM_NAME, Y.Widget, [Y.WidgetChild],
    LISTBOX_ITEM_PROTOTYPE, {ATTRS: LISTBOX_ITEM_ATTRS});

Y.landscape.ui.ListBoxItem = ListBoxItem;



/**
 * A widget for generating scrollable selection boxes.
 *
 * @class ListBox
 * @extends Widget
 * @constructor
 */
var LISTBOX_NAME = "listbox";

var LISTBOX_ATTRS = {

    /**
     * Size of the select box.
     */
    size: {
        value: null
    },

    defaultChildType: {
        valueFn: function() {
            return ListBoxItem;
        }
    }
};

var LISTBOX_PROTOTYPE = {

    CONTENT_TEMPLATE: "<ul></ul>",

    initializer: function(config) {
        this.set("showing", true);
    },

    /*
     * Hide the entire ListBox by setting the display CSS attribute to none.
     */
    hide: function() {
        var boundingBox = this.get("boundingBox");
        this.set("showing", false);
        boundingBox.setStyle("display", "none");
    },

    /*
     * Show the entire ListBox by unsetting the display CSS attribute.
     */
    show: function() {
        var boundingBox = this.get("boundingBox");
        this.set("showing", true);
        boundingBox.setStyle("display", "block");
    },

    /*
     * Alternate between shown and hidden list box.
     */
    toggleVisibility: function() {
        if (this.get("showing") === true) {
            this.hide();
        } else {
            this.show();
        }
    }

};

var ListBox = Y.Base.create(
    LISTBOX_NAME, Y.Widget, [Y.WidgetParent],
    LISTBOX_PROTOTYPE, {ATTRS: LISTBOX_ATTRS});

Y.landscape.ui.ListBox = ListBox;


/**
 * Dropdown creates a drop-down list box.
 *
 * @class Dropdown
 * @extends Widget
 * @constructor
 */
var DROPDOWN_NAME = "dropdown";

var DROPDOWN_ATTRS = {

    selected: {
        value: null
    },

    /**
     * If readOnly is set, display the currently selected item and stop
     * the dropdown list from popping down.
     */
    readOnly: {
        value: false
    },

    children: {
        value: null
    },

    // SELECT node that we are enhancing.
    selectNode: {
        value: null
    }

};

var DROPDOWN_PROTOTYPE = {
    INPUT_TEMPLATE: "<div>&nbsp;</div>",

    initializer: function(config) {
        var inputBox = Y.Node.create(this.INPUT_TEMPLATE),
            listBox = new ListBox({
              width: this.get("width"),
              children: this.get("children")
            });
        this._selectbox = inputBox;
        this._listbox = listBox;

        inputBox.setStyle("display", "none");
        inputBox.addClass(this.getClassName("inputbox"));

        if (this.get("selected") !== null &&
            this.get("selected") !== undefined) {
            var selectedItem = this.findItemByValue(this.get("selected"));
            if (selectedItem !== null) {
                selectedItem.set("selected", 1);
            }
        }
    },

    /**
     * Increase/decrease z-index on "inputBox" and "listBox" by ZINDEX_OFFSET.
     *
     * If "revert" is true, decreases the z-index, otherwise increases it.
     */
    _setDivStacking: function(inputBox, listBox, revert) {
        var inputBoxLevel = parseInt(inputBox.getStyle("zIndex")),
            listBoxLevel = parseInt(listBox.getStyle("zIndex")),
            factor = 1;
        if (revert) {
            factor = -1;
        }

        inputBoxLevel = isNaN(inputBoxLevel) ? 0 : inputBoxLevel;
        listBoxLevel = isNaN(listBoxLevel) ? 0 : listBoxLevel;

        inputBox.setStyle("zIndex", factor * ZINDEX_OFFSET + inputBoxLevel);
        listBox.setStyle("zIndex", factor * ZINDEX_OFFSET + listBoxLevel);
    },

    /*
     * Find the ListBoxItem in a contained ListBox with the matching
     * 'value' attribute.
     *
     * If there is no matching ListBoxItem, returns null.
     */
    findItemByValue: function(value) {
        var listBox = this._listbox;
        for (var itemIndex = 0; itemIndex < listBox.size(); itemIndex++) {
            var item = listBox.item(itemIndex);
            if (item.get("value") == value) {
                return item;
            }
        }
        return null;
    },

    /**
     * Return an array of non-hidden ListBoxItems in this Dropdown.
     *
     * If all items are hidden, returns an empty array.
     */
    getVisibleItems: function() {
        var listBox = this._listbox,
            visibleItems = [];

        for (var itemIndex = 0; itemIndex < listBox.size(); itemIndex++) {
            var item = listBox.item(itemIndex);
            if (item.get("visible")) {
                visibleItems.push(item);
            }
        }
        return visibleItems;
    },


    /*
     * Render the input box, the 'ul' node and add the relevant CSS classes.
     */
    renderUI: function() {
        var contentBox = this.get("contentBox"),
            inputBox = this._selectbox,
            listBox = this._listbox,
            selectNode = this.get("selectNode");

        contentBox.append(inputBox);
        listBox.hide();

        var selection = listBox.get("selection");
        if (selection !== undefined) {
            inputBox.setContent(selection.get("title"));
        }

        listBox.render(contentBox);
        inputBox.setStyle("display", "");
        if (selectNode) {
            selectNode.setStyle("display", "none");
        }
    },

    /*
     * Set up event handlers to keep the select box updated with selection
     * changes.
     */
    bindUI: function() {
        var selectBox = this.get("boundingBox"),
            listBox = this._listbox;
        selectBox.on("click", this._openList, this);
        listBox.after("selectionChange", this._afterListboxChange, this);
        this.after("selectedChange", this._afterSelectedChange, this);
        this.after("readOnlyChange", this._afterReadOnlyChange, this);
    },

    /**
     * Update the read-only flag on the dropdown.
     */
    syncUI: function() {
        this._uiSetReadOnly(this.get("readOnly"));
    },

    /*
     * Toggle the display of the dropdown ListBox depending on the
     * status of the "showing" attribute.
     */
    _openList: function(e) {
        if (this.get("readOnly")) {
            return;
        }
        var listBox = this._listbox,
            dropdownBox = this.get("boundingBox");
        if (listBox.get("showing") === true) {
            dropdownBox.removeClass("active");
        } else {
            dropdownBox.addClass("active");
        }
        this._setDivStacking(
            this._selectbox, listBox.get("contentBox"),
            listBox.get("showing"));
        listBox.toggleVisibility();
    },

    /**
     * Event handler to deal with changes to the readOnly attribute.
     * Added just so we keep up with our naming pattern of having event
     * handlers be named "handler-type + event-name".
     */
    _afterReadOnlyChange: function(event) {
        this._uiSetReadOnly(event.newVal);
    },

    /**
     * Adds/removes "readonly" CSS class to the bounding box node depending
     * on the passed-in value for the "readOnly".
     */
    _uiSetReadOnly: function(value) {
        var boundingBox = this.get("boundingBox"),
            readOnlyClass = this.getClassName("readonly");
        // This sets the "readonly" class if readOnly attribute is set,
        // and removes it otherwise.
        boundingBox.toggleClass(readOnlyClass, value);
    },

    /*
     * Event handler for selection change in the contained ListBox.
     */
    _afterListboxChange: function(e) {
        var listBox = this._listbox,
            selection = listBox.get("selection");
        this.set("selected", selection.get("value"), {src: "ui"});
    },

    /*
     * Event handler to deal with changes to the "selected" attribute
     * on the dropdown.
     *
     * Updates the select box text with the description from the current
     * selection and reflects the value change in the selectNode if present.
     */
    _afterSelectedChange: function(e) {
        var listBox = this._listbox,
            dropdownBox = this._selectbox,
            selectNode = this.get("selectNode"),
            selection = this.findItemByValue(this.get("selected"));

        selection.set("selected", 1);
        dropdownBox.setContent(selection.get("title"));
        dropdownBox.removeClass("active");
        if (selectNode) {
            selectNode.set("value", selection.get("value"));
        }
    }

};

var DROPDOWN_HTMLPARSER = {
    selectNode: "select",

    children: function(srcNode) {
        var selectNode = srcNode.one("select");
        if (selectNode) {
            return _getSelectOptionItems(selectNode);
        } else  {
            throw new Error(
                "No SELECT element found inside srcNode.");
        }
    },

    selected: function(srcNode) {
        var selectNode = srcNode.one("select");
        if (selectNode) {
            return selectNode.get("value");
        } else  {
            throw new Error(
                "No SELECT element found inside srcNode.");
        }
    }
};

var Dropdown = Y.Base.create(
    DROPDOWN_NAME, Y.Widget, [],
    DROPDOWN_PROTOTYPE, {ATTRS: DROPDOWN_ATTRS,
                         HTML_PARSER: DROPDOWN_HTMLPARSER });

Y.landscape.ui.Dropdown = Dropdown;


/******************************************************************************
 * Progressive enhancement functions.
 *
 */

/*
 * Return an array of {title, value} pairs matching OPTIONs inside
 * the passed in "selectNode".
 */
function _getSelectOptionItems(selectNode) {
    var children = [],
        selectOptions = selectNode.all("option");
    selectOptions.each(function(itemNode) {
        children.push({
            title: itemNode.getContent(),
            value: itemNode.get("value")
        });
    });
    return children;
}
// Export for testing.
Y.landscape.ui.__getSelectOptionItems = _getSelectOptionItems;


}, "0.1", {"requires": ["plugin", "array-invoke", "node", "base",
                        "widget", "widget-parent", "widget-child",
                        "event-focus"]});
