// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.form", function(Y) {

Y.namespace("landscape.ui");

/*
 * A form field widget with support for input tracking and error rendering.
 *
 * The widget keeps its 'value' attribute in sync with the underlying INPUT
 * element, using the 'change' event. Combined with the ability of rendering
 * errors it provides the basic tooling for implementing inline validation.
 */

var FormField = Y.Base.create("formField", Y.Widget, [], {

    CONTENT_TEMPLATE: null,

    renderUI: function() {
        var contentBox = this.get("contentBox");
        // Assume that progressive enhancement is being used, and the widget's
        // markup is already in place. Eventually we might want to support JS
        // rendering too, but we have no use cases for that yet.
        this._inputNode = contentBox.one("input");
        this._tooltipNode = contentBox.one(".tooltip-error");
    },

    bindUI: function() {
        var inlineValidation = this.get("inlineValidation");
        this._inputNode.on("change", this._onInputChange, this);
        this.after("valueChange", this._afterValueChange, this);
        this.after("readonlyChange", this._afterReadonlyChange, this);
        this.after("errorChange", this._afterErrorChange, this);
    },

    /*
     * Handle changes to the input node's value.
     */
    _onInputChange: function(e) {
        this.set("value", this._inputNode.get("value"), {src: "ui"});
    },

    /*
     * Handle changes to the widget's value attribute.
     */
    _afterValueChange: function(e) {
        if (e.src !== "ui") {
            this._uiSetValue(e.newVal);
        }
    },

    /*
     * Handle changes to the widget's readonly attribute.
     */
    _afterReadonlyChange: function(e) {
        if (e.src !== "ui") {
            this._uiSetReadonly(e.newVal);
        }
    },

    /*
     * Handle changes to the widget's error attribute.
     */
    _afterErrorChange: function(e) {
        this._uiSetError(e.newVal);
    },

    /*
     * Sync the input node with the widget's value.
     */
    _uiSetValue: function(value) {
        this._inputNode.set("value", value);
    },

     /*
     * Sync the input node with the widget's readonly attribute.
     */
    _uiSetReadonly: function(value) {
        if (value) {
            this._inputNode.setAttribute("readonly", "readonly");
        } else {
            this._inputNode.removeAttribute("readonly");
        }
    },

    /*
     * Sync the tooltip error node with the widget's error.
     */
    _uiSetError: function(error) {
       this.get("contentBox").toggleClass("form-field-error", error);
       this._tooltipNode.set("text", error);
    },


}, {

ATTRS: {

    /**
     * The name of the field.
     */
    name: {
        value: null,
    },

    /**
     * Current value of the form field.
     */
    value: {
        value: null,
    },

    /**
     * Whether the field is readonly.
     */
    readonly: {
        value: null,
    },

    /**
     * Form field error. The value should be a string with the error message
     * or null if there's no error.
     */
    error: {
        value: null,
    },

},

HTML_PARSER: {
    name: function(srcNode) {
        var inputNode = srcNode.one("input");
        return inputNode.get("name");
    },

    value: function(srcNode) {
        var inputNode = srcNode.one("input");
        return inputNode.get("value");
    },

    readonly: function(srcNode) {
        var inputNode = srcNode.one("input");
        return Boolean(inputNode.getAttribute("readonly"));
    },

    error: function(srcNode) {
        var errorNode = srcNode.one(".tooltip-error");
        return errorNode.get("text") || null;
    },
}});

Y.landscape.ui.FormField = FormField;

}, "0.1", {"requires": ["base", "event-valuechange", "widget"]});
