// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.image", function(Y) {

Y.namespace("landscape.ui.image");


/**
 * This module provides the Image widget.
 *
 * @module landscape.ui.image
 */


// Configuration properties
var IMAGE = "image",
    ATTR_PREFIX = "_",
    BOUNDING_BOX = "boundingBox",
    CSS_CLASSES = "cssClasses",
    IMAGE_ID = "imageId",
    NODE = "node",
    TITLE = "title";

/**
 * The Image widget is just a helper to ensure rendering images in
 * Javascript is consistent across our UI.
 *
 * @class Image
 * @extends Widget
 * @constructor
 */
var Image = function(config) {
    var attr;

    Image.superclass.constructor.apply(this, arguments);

    for (attr in config) {
        if (config.hasOwnProperty(attr)) {
            this.set(attr, config[attr]);
        }
    }
};

Image.NAME = "Image";

Image.ATTRS = {
    /**
     * One or more CSS classes to add to the generated image.
     *
     * @attribute cssClasses
     * @type Array
     */
    cssClasses: {
        value: null
    },

    /**
     * Path to the image that should be displayed.
     *
     * @attribute image
     * @type String
     */
    image: {
        value: null
    },

    /**
     * Global identifier to assign to the generated image.
     *
     * @attribute imageId
     * @type String
     */
    imageId: {
        value: null
    },

    /**
     * Title to assign to the generated image.
     *
     * @attribute title
     * @type String
     */
    title: {
        value: null
    },

    /**
     * Node representing the image.
     *
     * @attribute node
     * @type Node
     */
    node: {
        valueFn: function() {
            if (this._node === undefined) {
                this._node = Y.Node.create("<img />");
            }
            return this._node;
        }
    },

    boundingBox: {
        valueFn: function() {
            if (this._boundingBox === undefined) {
                this._boundingBox = Y.Node.create("<div />");
            }
            return this._boundingBox;
        }
    }
};

Y.extend(Image, Y.Object, {

    /**
     * A simplified getter implementation to avoid the overhead of
     * using Y.Attribute.
     *
     * @method get
     * @public
     */
    get: function(name) {
        var attrs = this.constructor.ATTRS,
            attr = attrs[name],
            valueFn = attr && attr.valueFn,
            value = this[ATTR_PREFIX + name] || (attr && attr.value);
        if (value === undefined && valueFn) {
            value = valueFn.call(this);
        }
        return value;
    },

    /**
     * A simplified setter implementation to avoid the overhead of
     * Y.Attribute.
     *
     * @method set
     * @public
     */
    set: function(name, value) {
        var changeFn = this[ATTR_PREFIX + name + "Change"],
            oldVal = this.get(name);
        this[ATTR_PREFIX + name] = value;
        if (changeFn) {
            changeFn.call(this, oldVal, value);
        }
    },

    /**
     * Render node to parentNode, if one is specified and different
     * from the boundingBox.
     *
     * @method render
     * @public
     */
    render: function(parentNode) {
        var boundingBox = this.get(BOUNDING_BOX);
        parentNode = parentNode || this.get(BOUNDING_BOX);

        this.renderUI();
        if (!boundingBox.compareTo(parentNode)) {
            parentNode.appendChild(this.get(NODE));
        }
    },

    /**
     * Render stuff that's specific to this widget.
     *
     * @method _renderUI
     * @protected
     */
    _renderUI: function() {
        var node = this.get(NODE);
        node.setAttribute("src", this.get(IMAGE));
        this.get(BOUNDING_BOX).appendChild(node);
    },

    /**
     * Render the contents of the widget.
     *
     * @method renderUI
     * @protected
     */
    renderUI: function() {
        var node = this.get(NODE);
        var id = this.get(IMAGE_ID);
        var title = this.get(TITLE);
        if (! Y.Lang.isNull(id)) {
            node.setAttribute("id", id);
        }
        if (! Y.Lang.isNull(title)) {
            node.setAttribute("title", title);
        }
        this._cssClassesSet(node, null, this.get(CSS_CLASSES));
        this.get(BOUNDING_BOX).addClass("icon");
        this._renderUI();
    },

    /**
     * Update the CSS classes for the managed node. Called from the
     * 'after' callback for the 'cssClasses' property.
     *
     * @method _cssClassesChange
     * @protected
     */
    _cssClassesChange: function(prevVal, newVal) {
        this._cssClassesSet(this.get(NODE), prevVal, newVal);
    },

    /**
     * Given a node and two arrays with previous and current CSS
     * classes, update the node to remove old CSS classe that only
     * existed in previous and add new ones that exist only in
     * current.
     *
     * @method _cssClassesSet
     * @protected
     */
    _cssClassesSet: function(node, previous, current) {
        // This function only works with arrays. If something that is
        // not an array is passed in, it will be ignored.
        if (! Y.Lang.isArray(previous)) {
            previous = new Y.Array();
        }
        if (! Y.Lang.isArray(current)) {
            current = new Y.Array();
        }
        var currentHash = Y.Array.hash(current);
        var previousHash = Y.Array.hash(previous);

        for (var previousClass in previousHash) {
            if (! currentHash[previousClass]) {
                // The current value doesn't contain this class, so
                // remove it from the node already.
                node.removeClass(previousClass);
            } else {
                // Both current and previous values contain this
                // class, so remove it from currentHash since we don't
                // need to look at it again below.
                delete currentHash[previousClass];
            }
        }

        // We already removed the classes that were in previousHash
        // from currentHash above, so we don't need to check for them
        // again. Just go ahead and add the class.
        for (var currentClass in currentHash) {
            if (currentHash.hasOwnProperty(currentClass)) {
                node.addClass(currentClass);
            }
        }
    },

    /**
     * Update the id of the managed node. Called from the 'after'
     * callback for the 'imageId' property.
     *
     * @method _imageIdChange
     * @protected
     */
    _imageIdChange: function (prevVal, newVal) {
        if (newVal) {
            this.get(NODE).setAttribute("id", newVal);
        }
    },

    /**
     * Update the title of the managed node. Called from the 'after'
     * callback for the 'title' property.
     *
     * @method _titleChange
     * @protected
     */
    _titleChange: function (prevVal, newVal) {
        if (newVal) {
            this.get(NODE).setAttribute("title", newVal);
        }
    },

    /**
     * Update the image of the managed node. Called from the 'after'
     * callback for the 'image' property.
     *
     * @method _imageChange
     * @protected
     */
    _imageChange: function(prevVal, newVal) {
        this.get(NODE).setAttribute("src", newVal);
    },

    /**
     * Add a new class to the 'cssClasses' property of this widget,
     * updating the managed node.
     *
     * @method addClass
     * @protected
     */
    addClass: function(className) {
        // Copy the value of cssClasses into a new array, since
        // modifying the array in place then setting the attribute
        // will not trigger an attribute change event.
        var cssClasses = this.get(CSS_CLASSES);
        var newVal = [];
        var i = cssClasses.length;
        while (i--) {
            newVal.push(cssClasses[i]);
        }
        newVal.push(className);
        this.set(CSS_CLASSES, newVal);
    },

    /**
     * Remove a class from the 'cssClasses' property of this widget,
     * updating the managed node.
     *
     * @method addClass
     * @protected
     */
    removeClass: function(className) {
        // Make a new copy of the cssClasses value removing any
        // occurrence of the specified class name, and then set the
        // attribute, triggering an attribute change event.
        var cssClasses = this.get(CSS_CLASSES);
        var newVal = [];
        var i = cssClasses.length;
        while (i--) {
            if (cssClasses[i] !== className) {
                newVal.push(cssClasses[i]);
            }
        }
        this.set(CSS_CLASSES, newVal);
    }
});


Y.landscape.ui.Image = Image;

/**
 * A helper function to generate a CSS class name from a filename
 * according to the conventions in use by Landscape (see
 * c.l.ui.spriteformat).
 */
var staticFilename = function(filename) {
    var config = Y.config.landscape || {base: ""},
        base = config.base || "";
    return base + filename;
};

Y.landscape.ui.image.staticFilename = staticFilename;


}, "0.1", {"requires": ["event", "node", "widget"]});
