// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.menu", function(Y) {

Y.namespace("landscape.ui");

var ITEMS = "items",
    CHANGE = "Change",
    HOST = "host";

// A plugin class that simplifies dynamically adding and removing menu items
// from a NodeMenuNav menu.
function MenuItemsPlugin(config) {
    MenuItemsPlugin.superclass.constructor.apply(this, arguments);
}

MenuItemsPlugin.NAME = "MenuItemsPlugin";
MenuItemsPlugin.NS = ITEMS;

MenuItemsPlugin.ATTRS = {
    /**
     * A list of all the items that can be selected.
     *
     * @attribute items
     */
    items: {
        value: null
    }
};

Y.extend(MenuItemsPlugin, Y.Plugin.Base, {
    MENU_ITEM_TEMPLATE: [
        "<li class='yui3-menuitem'>",
        "<a class='yui3-menuitem-content' ",
        "href='{href}'>{title}</a>",
        "</li>"].join(""),


    /**
     * Initialize the plugin, binding events as needed.
     */
    initializer: function(config) {
        this.after(ITEMS + CHANGE, this._afterItemsChange);
        this._nodes = [];

        // Parse existing markup into internal list of nodes and items.
        this.get(HOST).all(".yui3-menuitem").each(function(node) {
            this._nodes.push(node);
        }, this);

        var items = this.get(ITEMS) || [];
        Y.Array.each(this._nodes, function(node) {
            var item = node.one("a").getAttrs(["text", "href"]);
            item.title = item.text;
            this.push(item);
        }, items);

        // Manually trigger the itemsChange event when the plugin is initialized.
        this.set(ITEMS, items);
    },

    /**
     * Update the underlying DOM structure after a change to the 'items'
     * attribute.
     *
     * @method _afterItemsChange
     */
    _afterItemsChange: function(evt) {
        var oldNodes = this._nodes,
            oldLength = this._nodes.length,
            newItems = Y.Array(evt.newVal),
            newLength = newItems.length,
            host = this.get(HOST),
            container = host.one(".yui3-menu-content ul");

        if (newLength < oldLength) {
            var remove = oldNodes.splice(newLength);
            Y.Array.invoke(remove, "remove");
        }

        if (newLength > 0) {
            this._updateNodesFromItems(oldNodes, newItems, true);
            host.show();
        } else {
            host.hide();
        }

        if (newLength > oldLength) {
            var newNodes = [],
                fragment;
            Y.Array.each(Y.Array(newItems, oldLength),
                Y.rbind(function(item, idx, items, nodes, newNodes) {
                var node = Y.Node.create(
                    Y.Lang.sub(this.MENU_ITEM_TEMPLATE, item));
                nodes.push(node);
                newNodes.push(node);
            }, this, oldNodes, newNodes));
            fragment = new Y.NodeList(newNodes).toFrag();
            container.append(fragment);
        }
    },

    /**
     * Update an array of Node objects with the specified items..
     *
     * @method _updateNodesFromItems
     */
    _updateNodesFromItems: function(nodes, items) {
        Y.Array.each(nodes,
            Y.rbind(function(node, idx, nodes, newItems) {
            var anchor = node.one("a"),
                oldHref = anchor.get("href"),
                oldTitle = anchor.get("text"),
                item = newItems[idx],
                newHref = item.href,
                newTitle = item.title;
            if (!(oldHref == newHref &&
                  oldTitle == newTitle)) {
                anchor.set("href", newHref);
                anchor.set("text", newTitle);
            }
        }, this, items));
    }
});

Y.landscape.ui.MenuItemsPlugin = MenuItemsPlugin;

}, "0.1", {"requires": ["plugin", "array-invoke", "node", "base"]});
