// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.overlay", function(Y) {

Y.namespace("landscape.ui");

var BOUNDING_BOX = "boundingBox",
    CONTENT_BOX = "contentBox",
    OVERLAY = "overlay",
    SHADOW = "shadow",
    ALIGN = "align",
    WIDTH = "width",
    HEIGHT = "height",
    BODY = "body",
    HIDDEN = "hidden",
    FACTORY = "factory",
    WIDGET = "widget",
    RENDERED = "rendered",
    REGION = "region",
    START = "start",
    END = "end",
    VISIBLE = "visible",
    REVERSE = "reverse",
    LOADING_CLASS = "loading-large",
    SLIDING_OVERLAY = "sliding-overlay",
    _getClassName = Y.ClassNameManager.getClassName,
    OVERLAY_CLASS = _getClassName(SLIDING_OVERLAY, OVERLAY),
    SHADOW_CLASS = _getClassName(SLIDING_OVERLAY, SHADOW);

/**
 * Sliding Overlay
 *
 * Actually renders two overlays, one being a 'shadow' on top of a page region,
 * and the other being the container for the specified widget.
 *
 * @class SlidingOverlay
 * @extends Base
 * @constructor
 */
var SlidingOverlay = function(config) {
    SlidingOverlay.superclass.constructor.apply(this, arguments);
};

SlidingOverlay.NAME = "sliding-overlay";

Y.extend(SlidingOverlay, Y.Base, {

    /**
     * The overlay that will contain the actual widget.
     *
     * @attribute overlay
     */
    overlay: null,

    /**
     * The shadow overlay, used to gray out and prevent clicking through the
     * underlying content.
     *
     * @attribute shadow
     */
    shadow: null,

    /**
     * A factory to use for creating the widget that will be contained within
     * the overlay. It will only get called when the SlidingOverlay gets
     * activated.
     *
     * @attribute factory
     */
    factory: null,

    /**
     * The widget to be rendered within the overlay. It will be initialized by
     * using the factory above once the SlidingOverlay gets activated.
     *
     * @attribute widget
     */
    widget: null,

    /**
     * A hidden node used to render the widget.
     *
     * @attribute hidden
     */
    hidden: null,

    /**
     * The node that should be used to get the dimensions for the overlays.
     *
     * @attribute boundingBox
     */
    boundingBox: null,

    /**
     * The animation used to open the overlay.
     *
     * @attribute openShadowAnim
     */
    openShadowAnim: null,

    /**
     * The animation used to show the overlay.
     *
     * @attribute openOverlayAnim
     */
    openOverlayAnim: null,

     /**
      * The animation used when the content widget updates.
      *
      * @attribute contentUpdateAnim
      */
    contentUpdateAnim: null,

    /**
     * Where should both overlays be aligned to.
     *
     * @attribute align
     */
    align: null,

    /**
     * The duration for each animation.
     *
     * @attribute duration
     */
    duration: 0.3,

    /**
     * Initialize attributes from provided config.
     */
    initializer: function(config) {
        Y.mix(this, config, true);
        if (Y.Lang.isString(this[BOUNDING_BOX])) {
            this[BOUNDING_BOX] = Y.one(this[BOUNDING_BOX]);
        }
    },

    /**
     * Render the overlay and shadow widgets and create a hidden widget where
     * the content widget will be rendered to.
     *
     * @method render
     * @public
     */
    render: function(node) {
        node = node || Y.one(BODY);
        var boundingBox = this[BOUNDING_BOX],
            align = this[ALIGN],
            region = boundingBox.get(REGION),
            hidden = Y.Node.create(["",
                "<div style='visibility: hidden; ",
                "z-index: -1; height: 0px;' />"].join("")),
            overlay = new Y.Overlay({
                visible: false,
                width: region.width,
                height: 0,
                zIndex: 2,
                align: align}),
            shadow = new Y.Overlay({
                visible: false,
                width: region.width,
                height: region.height,
                zIndex: 1,
                align: align});
        this[OVERLAY] = overlay;
        this[SHADOW] = shadow;
        this[HIDDEN] = hidden;
        overlay.render(node);
        overlay.get(BOUNDING_BOX).addClass(OVERLAY_CLASS);
        shadow.render(node);
        shadow.get(BOUNDING_BOX).addClass(SHADOW_CLASS);
        node.append(hidden);
        shadow.on("click", function(evt) {
            this.toggle();
        }, this);
        Y.on("windowresize", function(evt) {
            var region = boundingBox.get(REGION);
            shadow.set(WIDTH, region.width);
            shadow.set(HEIGHT, region.height);
            overlay.set(WIDTH, region.width);
        });
    },

    /**
     * Toggle the widget open or closed. Content widget is rendered on the first
     * time the overlay is opened.
     *
     * @method toggle
     * @public
     */
    toggle: function() {
        var widget = this[WIDGET],
            boundingBox = this[BOUNDING_BOX],
            region = boundingBox.get(REGION),
            hidden = this[HIDDEN],
            factory = this[FACTORY],
            overlay = this[OVERLAY],
            overlayBox = overlay.get(BOUNDING_BOX),
            shadow = this[SHADOW],
            shadowBox = shadow.get(BOUNDING_BOX),
            duration = this.duration;

        if (Y.Lang.isNull(widget) && Y.Lang.isFunction(factory)) {
            // Widget doesn't exist yet, so let's initialize it.
            widget = factory();
            this[WIDGET] = widget;
        }

        if (!widget.get(RENDERED)) {
            // Since the widget doesn't exist yet, it means we haven't shown
            // the overlay yet. So animate showing the overlay and the shadow.
            this.openOverlayAnim = new Y.Anim({
                    node: overlayBox,
                    duration: duration,
                    easing: Y.Easing.easeBoth,
                    from: {height: 0},
                    to: {height: 64}
            });

            this.openShadowAnim = new Y.Anim({
                    node: shadowBox,
                    duration: duration,
                    from: {opacity: 0},
                    to: {opacity: 0.8}
            });

            this.openShadowAnim.once(END, function() {
                // When the shadow overlay animation ends, if the widget is
                // not yet visible then remove the loading class from the
                // overlay.
                if (!widget.get(VISIBLE)) {
                    overlayBox.addClass(LOADING_CLASS);
                }
                // Now animate opening the content overlay.
                overlay.show();
                if (this.openOverlayAnim) {
                    this.openOverlayAnim.run();
                }
            }, this);

            // A few helper functions to compute dynamic heights for the
            // animations below.
            var getWidgetHeight = Y.bind(function() {
                return this.get(BOUNDING_BOX).get(REGION).height;
            }, widget);

            var getOverlayHeight = Y.bind(function() {
                return this.get(REGION).height;
            }, overlayBox);

            this.openOverlayAnim.once(END, function() {
                // Setup FX plugins to animate content area from original size
                // to overlay height, and overlay height from zero to widget
                // height.
                overlayBox.plug(Y.Plugin.NodeFX, {
                    from: {height: 0},
                    to: {height: getWidgetHeight},
                    easing: Y.Easing.easeOut,
                    duration: duration,
                    reverse: true
                });

                shadowBox.plug(Y.Plugin.NodeFX, {
                    from: {opacity: 0},
                    to: {opacity: 0.8},
                    duration: duration,
                    reverse: true
                });

                shadowBox.fx.on(START, function() {
                    if (!(shadow.get(VISIBLE) ||
                          shadowBox.fx.get(REVERSE))) {
                        shadow.show();
                    }
                });

                shadowBox.fx.on(END, function() {
                    var reverse = this.get(REVERSE);
                    if (reverse) {
                        shadow.hide();
                    }
                    this.set(REVERSE, !reverse);
                });

                overlayBox.fx.on(START, function() {
                    if (!(overlay.get(VISIBLE) ||
                          overlayBox.fx.get(REVERSE))) {
                        overlay.show();
                    }
                });

                overlayBox.fx.on(END, function() {
                    var reverse = this.get(REVERSE);
                    if (reverse) {
                        overlay.hide();
                    }
                    this.set(REVERSE, !reverse);
                });

            });

            // Animate overlay from it's current height to the
            // current height for the widget.
            this.contentUpdateAnim = new Y.Anim({
                node: overlayBox,
                duration: duration,
                easing: Y.Easing.easeBoth,
                from: {height: getOverlayHeight},
                to: {height: getWidgetHeight}
            });

            widget.on("contentUpdate", function() {
                // Replace the overlay body with the
                // widget and un-hide the widget. It will actually only
                // show up in the page when the containing Overlay
                // widget is visible.
                overlay.set("bodyContent", widget.get(BOUNDING_BOX));
                widget.show();
                overlayBox.removeClass(LOADING_CLASS);
                this.contentUpdateAnim.run();
            }, this);

            shadow.show();
            this.openOverlayAnim.once("start", function() {
                widget.render(hidden);
            });
            this.openShadowAnim.run();

        } else {
            if (overlay.get(VISIBLE)) {
                overlayBox.fx.once(END, function() {
                    shadowBox.fx.run();
                });

                overlayBox.fx.run();
            } else {
                // Ensure shadow and overlay dimensions match the overlay
                // box. It might have changed while the overlay was hidden.
                shadow.set(WIDTH, region.width);
                shadow.set(HEIGHT, region.height);
                overlay.set(WIDTH, region.width);

                shadowBox.fx.once(END, function() {
                    overlayBox.fx.run();
                });

                shadowBox.fx.run();
            }
        }
    }
});

Y.landscape.ui.SlidingOverlay = SlidingOverlay;

}, "0.1", {"requires": ["node", "event", "anim", "overlay", "classnamemanager"]});
