// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.tooltip", function(Y) {

Y.namespace("landscape.ui");

/**
 * This module provides the ToolTip widget. Multiple Tooltips can be used in a
 * single page.
 * It's possible to position the Tooltip at the left, top, right and botoom of
 * the tooltip target, which could be a link for example.
 * It's also possible to specify an element containing the contents of the
 * tooltip, otherwise it will use the tooltip target's data-tooltip attribute
 * for its content.
 *
 * @module landscape.ui.tooltip
 *
 **/

/**
 * ToolTip provides an overlayed tooltip bubble that can be positioned at
 * arbitrary possitions relative to an element containing the 'target'
 * css class.
 *
 * @class ToolTip
 * @extends Widget
 * @constructor
 **/

var ToolTip = function(config) {
    ToolTip.superclass.constructor.apply(this, arguments);
};

ToolTip.NAME = "tooltip";

ToolTip.ATTRS = {
    /**
     * The position of the tooltip.
     * It can be one of the following: left, top, right, bottom.
     *
     * @attribute position
     * @type String
     */
    position: {
        value: "top",
        validator: "_validatePositionFn"
    },

    /**
     * The id of the source of the content for the tooltip. If it's null, the
     * tooltip data- attribute will be used.
     *
     * @attribute bodyContent
     * @type String
     */
    bodyContent: {
        value: null
    },

    /**
     * Tooltip target area: nodes over which mouse causes the tooltip
     * to be shown.
     *
     * @attribute target
     * @type: String selector to pass into delegate calls.
     */
    target: {
        value: ".tooltip-target"
    },

    /**
     * Number of milliseconds to wait before hiding the tooltip.
     *
     * @attribute hideTimeout
     * @type: Integer
     */
    hideTimeout: {
        value: 0
    }
};

Y.extend(ToolTip, Y.Widget, {

    CONTENT_TEMPLATE: null,

    initializer: function (config) {
        this._overlay = new Y.Overlay({visible: false});
        // The current position of the overlay, which is different from the
        // "position" attribute in case auto-positioning is used.
        this._currentAutoPosition = null;
    },

    renderUI: function () {
        // Render the tooltip overlay.
        this._overlay.render(this.get("contentBox"));
        this._overlay.get("contentBox").addClass("tooltip");
        var position = this.get("position");
        // Set default position if one of the 'auto' options is chosen.
        if (position == "auto-h") {
            position = "right";
        } else if (position == "auto-v") {
            position = "bottom";
        }
        this._changePosition(position);
    },

    bindUI: function () {
        // Register mouse hover event handlers for tooltip targets.
        var contentBox = this.get("contentBox");
        contentBox.delegate(
            "mouseenter", this._onMouseEnter, this.get("target"), this);
        contentBox.delegate(
            "mouseleave", this._onMouseLeave, this.get("target"), this);

        this.on("visibleChange", this._onVisibleChange, this);
    },

    /**
     * Event handler for showing and hiding the entire tooltip
     * (eg. with show() and hide() methods).
     */
    _onVisibleChange: function(event) {
        if (event.newVal) {
            this._overlay.show();
        } else {
            this._overlay.hide();
        }
    },

    /**
     * Event handler on all tooltip targets to show and place the tooltip.
     **/
    _onMouseEnter: function(event) {
        var target = event.currentTarget;
        // Whether we are currently in the process of (delayed) hiding
        // of the tooltip.
        this._hidingInProgress = false;
        if (this._timeout) {
            clearTimeout(this._timeout);
        }
        this._setBodyContent(target, this.get("bodyContent"));
        this.show();
    },

    /**
     * Event handler that hides the tooltip when the mouse leaves the
     * tooltip target.
     **/
    _onMouseLeave: function (event) {
        var tooltip = this; overlayNode = this._overlay;
        tooltip._hidingInProgress = true;
        if (this.get("hideTimeout")) {
            this._timeout = setTimeout(function() {
                if (tooltip._hidingInProgress) {
                    tooltip.hide();
                    tooltip._hidingInProgress = false;
                }
            }, this.get("hideTimeout"));
        } else {
            tooltip.hide();
            tooltip._hidingInProgress = false;
        }
    },

    /**
     * Event handler for click events.
     */
    _onClick: function(event) {
        event.preventDefault();
        var target = event.currentTarget;

        this._setBodyContent(target, this.get("bodyContent"));
        this._overlay.get("boundingBox").toggleClass("yui3-overlay-hidden");
    },

    /**
     * Sets the content of the body of the overlay.
     */
    _setBodyContent: function(target, bodyContent) {
        var template = "{tooltip}";
        // If bodyContent is defined, use it as template. Otherwise, the
        // 'data-tooltip' attribute is used by default.
        if (!Y.Lang.isNull(bodyContent)) {
            template = Y.one(bodyContent).getHTML();
        }
        this._overlay.set(
            "bodyContent", Y.Lang.sub(template, this._getTemplateData(target)));
        this._alignOverlay(target);
    },

    /**
     * Return data to populate the tooltip template.
     * Can be overridden in subclasses, by default it returns the target data
     attributes.
     */
    _getTemplateData: function(target) {
        return target.getData();
    },

    /**
     * Make sure it's a supported position.
     **/
    _validatePositionFn: function(value, attribute) {
        var supported = ["left", "top", "right", "bottom", "auto-h", "auto-v"];
        if (supported.indexOf(value) >= 0) {
            return true;
        }
    },

    /*
     * Make sure it's a supported mouse event.
     */
    _validateMouseEventFn: function(value, attribute) {
        var supported = ["click", "mouseenter"];
        if (supported.indexOf(value) >= 0) {
            return true;
        }
    },

    /*
     * Adjust CSS classes to change tooltip position.     
     */
    _changePosition: function(position) {
        var contentBox = this._overlay.get("contentBox");
        if (this._currentAutoPosition) {
            contentBox.removeClass("tooltip-" + this._currentAutoPosition);
        }
        contentBox.addClass("tooltip-" + position);
        this._currentAutoPosition = position;
    },

    /*
     * Align the overlay based on the widget position.  If one of the 'auto'
     * options is chosen, align to the proper one to avoid the tooltip
     * overflowing the 'srcNode'.
     */
    _alignOverlay: function(target) {
        var position = this.get("position"),
            realPosition = position,
            nodeRegion, 
            overlayRegion,
            points;

        if (position == "auto-h" || position == "auto-v") {
            // Try the default position first
            if (position == "auto-h") {
                realPosition = "right";
            } else if (position == "auto-v") {
                realPosition = "bottom";
            }
            points = this._getAlignmentPoints(realPosition);
            this._overlay.align(target, points);
            nodeRegion = this.get("srcNode").get("region");
            overlayRegion = this._overlay.get("srcNode").get("region");

            if (position == "auto-h" && overlayRegion.right > nodeRegion.right) {
                // The overlay overflows on the right, switch position to left.
                realPosition = "left";
            } else if (position == "auto-v" &&
                       overlayRegion.bottom > nodeRegion.bottom) {
                // The overlay overflows on the bottom, switch position to top.
                realPosition = "top";
            }
            this._changePosition(realPosition);
        }

        points = this._getAlignmentPoints(realPosition);
        this._overlay.align(target, points);
    },

    /**
     * Return YUI.Overlay points to align the tooltip.
     *
     * @method _getAlignmentPoints
     * @private
     **/
    _getAlignmentPoints: function(position) {
        switch(position) {
            case "top": // Bottom center with top center
                return ["bc", "tc"];
            case "bottom": // Top center with bottom center
                return ["tc", "bc"];
            case "left": // Right center with left center
                return ["rc", "lc"];
            case "right": // Left center with right center
                return ["lc", "rc"];
        }
    }
});

Y.landscape.ui.ToolTip = ToolTip;


}, "0.1", {"requires": ["node",
                        "base",
                        "widget",
                        "event",
                        "overlay"]});
