// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.ui.url", function(Y) {

Y.namespace("landscape.ui");

/**
 * This widget generates a URL from a name textbox. Optionally, the user can
 * edit the URL directly.
 *
 * @class GeneratedURL
 * @extends Widget
 * @constructor
 */

var GeneratedURL = Y.Base.create("GeneratedURL", Y.Widget, [], {

    /**
     * Bind events to the widget.
     *
     * @method bindUI
     * @protected
     */
    bindUI: function() {
        var nameField = this.get("nameField");
        nameField.after("keyup", this._onNameKeyup, this);

        var urlInput = Y.one("#user-url-text");
        urlInput.on("keydown", this._onURLKeydown, this);

        var editButton = Y.one("#edit-url");
        editButton.on("click", this._onEditClick, this);

        var setButton = Y.one("#set-url");
        setButton.on("click", this._onSetClick, this);

        var cancelButton = Y.one("#cancel");
        cancelButton.on("click", this._onCancelClick, this);

        var baseURL = this.get("baseURL");
        var baseSpan = Y.one("#base-url");
        baseSpan.set("text", baseURL);
    },

    /**
     * Respond to keyup event in name field by generating an appropriate URL
     * unless the user has edited the URL by hand.
     *
     * @method _onNameKeyup
     * @private
     */
    _onNameKeyup: function(e) {
        var edited = this.get("edited");

        if (!edited) {
            var target = e.target;
            var value = target.get("value");
            this._setURL(value);
        }
    },

    /**
     * Returns true if the keycode is in the list of nonprinting keystrokes
     * that a user might use while editing text.
     *
     * @method _isWhitelistedKeystroke
     * @private
     */
    _isWhitelistedKeystroke: function(keyCode) {
        switch(keyCode) {
            case 8:  //backspace
            case 20: //caps lock
            case 35: //end
            case 36: //home
            case 37: //left
            case 38: //up
            case 39: //right
            case 40: //down
            case 45: //insert
            case 46: //delete
                return true;
            default:
                return false;
        }
    },

    /**
     * Respond to keyup event on URL field checking user input and updating the
     * URL.
     *
     * @method _onURLKeydown
     * @private
     */
    _onURLKeydown: function(e) {
        var code = e.keyCode;               
        // XXX: For this to work we need to explicitely allow any characters
        // that a user will want to use like code 8 and 127 for backspace and
        // delete. What other codes do we need here?
        if (this._isWhitelistedKeystroke(code)) {
            return;
        }
        var key = String.fromCharCode(code);
        var cleaned = this._clean(key);
        if (key != cleaned) {
            e.halt();
        }
    },

    /**
     * Set the URL in both the user input box and the URL generation area.
     *
     * @method _setURL
     * @private
     */
    _setURL: function(value) {
        cleaned = this._cleanURL(value);
        var baseURL = this.get("baseURL");

        var url = baseURL.concat("/", cleaned);

        var generatedURL = Y.one("#generated-url-text");
        generatedURL.set("text", url);

        var userURL = Y.one("#user-url-text");
        userURL.set("value",  cleaned);

        var urlField = this.get("urlField");
        urlField.set("value",  url);
    },

    /**
     * Clean text to eliminate characters that would not make a good part of a
     * URL.
     *
     * @method _clean
     * @private
     */
    _clean: function(value) {
        var cleaned = value.replace(/\s/g, "-");
        cleaned = cleaned.replace(/[^\d\w-]/g, "");
        return cleaned;
    },

    /**
     * Clean text so that it makes a good part of a URL.
     *
     * @method _cleanURL
     * @private
     */
    _cleanURL: function(value) {
        cleaned = this._clean(value);
        cleaned = cleaned.toLowerCase();
        return cleaned;
    },

    /**
     * Respond to click event on edit button.
     *
     * @method _onEditClick
     * @private
     */
    _onEditClick: function(e) {
        var generatedURL = Y.one("#generated-url-box");
        generatedURL.setStyle("display", "none");

        var userURL = Y.one("#user-url-box");
        userURL.setStyle("display", "block");

        var userInput = Y.one("#user-url-text");
        var value = userInput.get("value");

        this.set("cancelCache", value);
        this._setURL(value);
    },

    /**
     * Respond to click event on cancel button.
     *
     * @method _onCancelClick
     * @private
     */
    _onCancelClick: function(e) {
        var generatedURL = Y.one("#generated-url-box");
        generatedURL.setStyle("display", "block");

        var userURL = Y.one("#user-url-box");
        userURL.setStyle("display", "none");

        var value = this.get("cancelCache");
        this._setURL(value);
    },

    /**
     * Respond to click event on edit button.
     *
     * @method _onSetClick
     * @private
     */
    _onSetClick: function(e) {
        this.set("edited", true);

        var generatedURL = Y.one("#generated-url-box");
        generatedURL.setStyle("display", "block");

        var userURL = Y.one("#user-url-box");
        userURL.setStyle("display", "none");

        var userInput = Y.one("#user-url-text");
        var value = userInput.get("value");
        this._setURL(value);
    }

}, {
    ATTRS: {
        /**
         * Has the generated URL been edited by the user?
         *
         * @attribute edited
         * @type boolean
         */
         edited: {value: false},

        /**
         * Base URL used at the beginning of the gnerated URL.
         *
         * @attribute baseURL
         * @type string
         */
         baseURL: {value: null},

        /**
         * allows reverting to old value on cancel.
         *
         * @attribute cancelCache
         * @type string
         */
         cancelCache: {value: ""},

        /**
         * Node from which a URL can be generated. Should be a text input field.
         *
         * @attribute nameField
         * @type Node
         */
         nameField: {value: null},

        /**
         * Node to store URL. Should probably be a hidden field.
         *
         * @attribute urlField
         * @type Node
         */
         urlField: {value: null}
     }
});

Y.landscape.ui.GeneratedURL = GeneratedURL;

}, "0.1", {"requires": ["node", "base", "widget"]});
