// Copyright 2005-2013 Canonical Limited.  All rights reserved.

YUI.add("landscape.util.node", function(Y) {

Y.namespace("landscape.util");

var SRC_NODE = "srcNode";

/**
 * Takes either an array of Node objects or selectors, or both, and
 * returns an array of Node objects.
 */
var nodesSetter = function(nodes, name) {
    var items = [];

    var push = Y.bind(function(items, nodeOrSelector) {
        // Can be a proper node or a selector.
        var node = nodeOrSelector;
        if (!(node instanceof Y.Node)) {
            // If it's not a node, create NodeList out of the selector
            // and then recurse.
            nodes = this.get(SRC_NODE).all(nodeOrSelector);
            return nodes.each(push);
        }
        // Otherwise, if not a null node, then push it into the array.
        if (!Y.Lang.isNull(node)){
            items.push(node);
        }
        return null; // jslint
    }, this, items);

    if (nodes instanceof Y.NodeList) {
        nodes.each(push);
    } else {
        Y.Array.each(Y.Array(nodes, 0), push);
    }
    return items;
};

Y.landscape.util.nodesSetter = nodesSetter;

/**
 * Takes a node or a selector and returns a Node.
 */
var nodeSetter = function(node, name) {
    if (node instanceof Y.Node) {
        return node;
    }
    return this.get(SRC_NODE).one(node);
};

Y.landscape.util.nodeSetter = nodeSetter;

/**
 * Takes an object mapping key to an array of selectors and returns a mapping
 * from key to a static Y.NodeList of items matching the given selectors.
 */
var nodesMappingSetter = function(mapping, filter) {
    var setter = Y.bind(nodesSetter, this);
    Y.Object.each(mapping, function(nodesOrSelector, key, mapping) {
        mapping[key] = new Y.NodeList(setter(nodesOrSelector));
    }, this);
    return mapping;
};

Y.landscape.util.nodesMappingSetter = nodesMappingSetter;

/**
 * Takes an object mapping key to a single selector and returns a mapping from
 * key to a live Y.NodeList.
 */
var nodeListMappingSetter = function(mapping, filter) {
    Y.Object.each(mapping, function(selector, key, mapping) {
        if (!(selector instanceof Y.NodeList)) {
            mapping[key] = this.get("srcNode").all(selector);
        }
    }, this);
    return mapping;
};

Y.landscape.util.nodeListMappingSetter = nodeListMappingSetter;

/**
 * Smart scrolling of a node into viewport.
 *
 * The idea is to scroll the least amount needed to show what's important.
 * If the element is below the fold, scroll up and align with the bottom
 * border, if it's above the fold, scroll down and align with the top border.
 *
 * Complications arise when the entire element can't fit into the viewport,
 * so here are all the cases (rotated for terseness).
 *
 * LEGEND:   << above  below >>
 * VIEWPORT:       +----+
 * NODE:      +--+                  a. above the fold => align top
 *                +--+              b. partially above the fold => align top
 *                  +--+            c. fully encompassed => nothing
 *                    +--+          d. partially below the fold => align bottom
 *                        +--+      e. fully below the fold => align bottom
 *
 * VIEWPORT:         +--+
 * NODE:      +----+                f. above the fold => align-top
 *               +----+             g. partially above the fold => nothing
 *                  +----+          h. fully encompassing => nothing
 *                     +----+       i. partially below the fold =>  align top
 *                        +----+    j. fully below the fold => align top
 *
 * VIEWPORT:       +**----+         topOffset represented as '*'
 * NODE:         +--+               k. above the topOffset => align top then
 *                                       scroll by topOffset
 *                  +--+            l. partial above topOffset => align top
 *                                       then scroll by topOffset
 *                    +--+          m. fully encompassed => nothing
 *
 * Some of the choices are controversial.
 *  - for case f), choosing align-bottom might be just as good, but doesn't
 *    matter much since viewport is too big to hold the entire element anyway.
 *  - for case g), we prefer not scrolling because we usually get that when
 *    we expand a smaller element downwards into a bigger element.
 * This is all encompassed in the following logic and suitably tested.
 *
 * @param node: The node to scroll into view
 * @param viewPortNode: Optional scrollable container node in which we check
 *     node's visibility. Defaults to Y.one("body");
 * @param topOffset: Optional additional scroll offset required within
 #     viewPortNode for visibility.
 */
var smartScrollIntoView = function(node, viewPortNode, topOffset) {
    var rect = node._node.getBoundingClientRect(),
        viewPort;
    if (Y.Lang.isUndefined(topOffset)) {
        topOffset = 0;
    }
    if (Y.Lang.isUndefined(viewPortNode)) {
        viewPortNode = Y.one("body");
    }
    viewPort = viewPortNode._node.getBoundingClientRect();
    if (rect.height >= (viewPort.height - topOffset)) {
        if (!(rect.bottom > 0 && rect.top < topOffset)) {
            // Scroll the element so the top of the element is aligned with
            // the top of the viewport plus topOffset.
            node.scrollIntoView(true);
            if (topOffset !== 0) {
                window.scrollBy(0, -topOffset);
            }
        }
    } else {
        if (rect.top < topOffset) {
            node.scrollIntoView(true);
            if (topOffset !== 0) {
                window.scrollBy(0, -topOffset);
            }
        } else if (rect.bottom > viewPort.height) {
            node.scrollIntoView(false);
        }
    }
};
Y.landscape.util.smartScrollIntoView = smartScrollIntoView;

}, "0.1", {"requires": ["node"]});
