import PropTypes from 'prop-types';
import React from 'react';
import { injectStripe } from 'react-stripe-elements';

import { AddressCheckbox, AccountAddress, BillingAddress } from './address';
import { AccountType } from './account-type';
import { getCountryState, getStateState } from './country';
import { isValidVATIN } from './vat';
import { CreditCardForm } from './credit-card-form';
import { getFieldState, ModalInterstitial } from './components';
import { delayDecorator } from './utility';


export class _AddBillingDetailsForm extends React.Component {

    constructor(props) {
        super(props);
        this.state = {
            stripeError: null,
            accountCountry: props.shippingAddress.country,
            accountVAT: props.accountVAT,
            accountType: props.accountType,
            accountName: props.shippingAddress.name,
            accountAddress1: props.shippingAddress.address1,
            accountAddress2: props.shippingAddress.address2,
            accountState: props.shippingAddress.state,
            accountCity: props.shippingAddress.city,
            accountZip: props.shippingAddress.zip,
            billingCountry: "",
            billingName: "",
            billingAddress1: "",
            billingAddress2: "",
            billingState: "",
            billingCity: "",
            billingZip: "",
            sameAddress: false,
            creditCardOK: true,  // Stripe form is initially submittable
        };
        this.handleAddressCheckboxChange = this.handleAddressCheckboxChange.bind(this);
        this.handleBillingAddressChange = this.handleBillingAddressChange.bind(this);
        this.handleAccountAddressChange = this.handleAccountAddressChange.bind(this);
        this.handleCreditCardOKChange = this.handleCreditCardOKChange.bind(this);
        this.handleAccountVATBlur = this.handleAccountVATBlur.bind(this);
        this.handleAccountTypeChange = this.handleAccountTypeChange.bind(this);
        this.refreshFieldValidity = this.refreshFieldValidity.bind(this);
        this.isSubmittable = this.isSubmittable.bind(this);
        this.onCancel = this.onCancel.bind(this);
        this.onSubmit = this.onSubmit.bind(this);
        this.simpleAddressFields = [
            "billingName",
            "billingAddress1",
            "billingCity",
            "billingZip",
            "accountName",
            "accountAddress1",
            "accountCity",
            "accountZip",
        ];
    }

    handleAddressCheckboxChange(sameAddress) {
        this.setState({sameAddress: sameAddress});
    }

    handleAccountAddressChange(items) {
        for (let key in items) {
            // AccountAddress is only allowed to update account address fields
            if (key.startsWith("account")) {
                let stateUpdate = {[key]: items[key]};
                if (this.simpleAddressFields.includes(key)) {
                    stateUpdate[key + "OK"] = getFieldState(items[key]);
                } else if (key === "accountState") {
                    stateUpdate["accountStateOK"] = getStateState(items[key], this.state.accountCountry);
                } else if (key === "accountCountry") {
                    let [isVATOK, errorMessage] = isValidVATIN(this.state.accountVAT, items[key]);
                    stateUpdate["accountCountryOK"] = getCountryState(items[key]);
                    stateUpdate["accountVATOK"] = isVATOK;
                    stateUpdate["accountVATErrorMessage"] = errorMessage;
                    if (this.state.accountStateOK !== undefined) {
                        stateUpdate["accountStateOK"] = getStateState(this.state.accountState, items[key]);
                    }
                }
                this.setState(stateUpdate);
            }
        }
    }

    handleBillingAddressChange(items) {
        for (let key in items) {
            // BillingAddress is only allowed to update billing address fields
            if (key.startsWith("billing")) {
                let stateUpdate = {[key]: items[key]};
                if (this.simpleAddressFields.includes(key)) {
                    stateUpdate[key + "OK"] = getFieldState(items[key]);
                } else if (key === "billingState") {
                    stateUpdate["billingStateOK"] = getStateState(items[key], this.state.billingCountry);
                } else if (key === "billingCountry") {
                    stateUpdate["billingCountryOK"] = getCountryState(items[key]);
                    if (this.state.billingStateOK !== undefined) {
                        stateUpdate["billingStateOK"] = getStateState(this.state.billingState, items[key]);
                    }
                }
                this.setState(stateUpdate);
            }
        }
    }

    handleCreditCardOKChange(isOK) {
        this.setState({creditCardOK: isOK});
    }

    handleAccountVATBlur() {
        const [isVATOK, errorMessage] = isValidVATIN(this.state.accountVAT, this.state.accountCountry);
        this.setState({accountVATOK: isVATOK, accountVATErrorMessage: errorMessage});
    }

    handleAccountTypeChange(value) {
        this.setState({accountType: value});
    }

    onCancel() {
        window.location.replace("./");
    }

    // Click handler when submitting
    refreshFieldValidity() {
        let billingUpdate = {};
        let accountUpdate = {};
        // Coerce empty fields to the empty string
        this.simpleAddressFields.map(key => {
            let map;
            if (key.startsWith("billing")) {
                map = billingUpdate;
            } else if (key.startsWith("account")) {
                map = accountUpdate;
            } else {
                throw "unexpected key";
            }
            map[key] = this.state[key] || "";
        });
        billingUpdate["billingState"] = this.state["billingState"] || "";
        billingUpdate["billingCountry"] = this.state["billingCountry"] || "";
        accountUpdate["accountState"] = this.state["accountState"] || "";
        accountUpdate["accountCountry"] = this.state["accountCountry"] || "";
        this.handleBillingAddressChange(billingUpdate);
        this.handleAccountAddressChange(accountUpdate);
    }

    // Is ths form ready to be submitted?
    isSubmittable() {
        let mustHaveFields = [
            this.state.creditCardOK,
            this.state.accountNameOK,
            this.state.accountAddress1OK,
            this.state.accountCityOK,
            this.state.accountStateOK,
            this.state.accountZipOK,
            this.state.accountCountryOK,
        ];
        if (this.state.accountVATOK !== undefined) {
            // undefined VAT is fine
            mustHaveFields.push(this.state.accountVATOK);
        }
        const billingFields = [
            this.state.billingNameOK,
            this.state.billingAddress1OK,
            this.state.billingCityOK,
            this.state.billingStateOK,
            this.state.billingZipOK,
            this.state.billingCountryOK,
        ];
        if ( !this.state.sameAddress ) {
            mustHaveFields.push(...billingFields);
        }

        return mustHaveFields.every(field => Boolean(field));
    }

    // Return data to be submitted from the form.
    _getSubmitData() {
        let form = new FormData();
        form.set("accountCountry", this.state["accountCountry"]);
        form.set("accountVAT", this.state["accountVAT"]);
        form.set("accountType", this.state["accountType"]);
        form.set("accountName", this.state["accountName"]);
        form.set("accountAddress1", this.state["accountAddress1"]);
        form.set("accountAddress2", this.state["accountAddress2"]);
        form.set("accountState", this.state["accountState"]);
        form.set("accountCity", this.state["accountCity"]);
        form.set("accountZip", this.state["accountZip"]);
        form.set("form-security-token", this.props.formToken);
        return form;
    }

    _getBillingDetails() {
        if (this.state.sameAddress) {
            return {
                "name": this.state.accountName,
                "address": {
                    "city": this.state.accountCity,
                    "country": this.state.accountCountry,
                    "line1": this.state.accountAddress1,
                    "line2": this.state.accountAddress2,
                    "postal_code": this.state.accountZip,
                    "state": this.state.accountState,
                }
            };
        }
        return {
            "name": this.state.billingName,
            "address": {
                "city": this.state.billingCity,
                "country": this.state.billingCountry,
                "line1": this.state.billingAddress1,
                "line2": this.state.billingAddress2,
                "postal_code": this.state.billingZip,
                "state": this.state.billingState,
            }
        };
    }

    onSubmit(event) {
        event.preventDefault();
        if (! this.isSubmittable()) {
            return;
        }
        this.setState({amSubmitting: true});
        let form = this._getSubmitData();
        this.setState({stripeError: null});

        // Bank authentication happens here, if SCA is required.
        this.props.stripe.handleCardSetup(
            this.props.stripeClientSecret,
            {payment_method_data: {billing_details: this._getBillingDetails()}}
        ).then((result) => {
            if (result.error) {
                this.setState({
                    amSubmitting: false,
                    stripeError: result.error.message,
                });
                return;
            }

            form.set("stripeSetupIntent", result.setupIntent.id);
            let fetchConfig = {
                credentials: 'same-origin',
                method: 'POST',
                body: form
            };
            fetch('./create', fetchConfig).then((response) => {
                if (response.error) {
                    // Restarting the SetupIntent flow from scratch as
                    // it is unusable at this point.
                    window.location.reload();
                } else {
                    window.location.replace(response.url);
                }
            }).catch((error) => {
                this.setState({amSubmitting: false});
                throw new Error('Failure to add card: ' + error.message);
            });
        });
    }

    render() {
        return (
            <div className="p-payment">
                <ModalInterstitial isOpen={this.state.amSubmitting} />
                <h2 className="p-payment__title">
                    Payment details &nbsp;&nbsp;
                    <small style={{fontSize: 20}}>
                        <i className="p-icon--padlock p-icon--title"></i> 128-bit SSL encrypted payment
                    </small>
                </h2>
                <form method="POST" onSubmit={this.onSubmit} >
                    <AccountType type={this.state.accountType} onChange={this.handleAccountTypeChange} className='u-clearfix' />
                    <AccountAddress isStripeAddress={this.state.sameAddress}
                        isBusinessAccount={this.state.accountType == "business"}
                        accountName={this.state.accountName}
                        accountNameOK={this.state.accountNameOK}
                        accountCountry={this.state.accountCountry}
                        accountCountryOK={this.state.accountCountryOK}
                        accountVAT={this.state.accountVAT}
                        accountVATOK={this.state.accountVATOK}
                        accountVATErrorMessage={this.state.accountVATErrorMessage}
                        accountAddress1={this.state.accountAddress1}
                        accountAddress1OK={this.state.accountAddress1OK}
                        accountAddress2={this.state.accountAddress2}
                        accountState={this.state.accountState}
                        accountStateOK={this.state.accountStateOK}
                        accountCity={this.state.accountCity}
                        accountCityOK={this.state.accountCityOK}
                        accountZip={this.state.accountZip}
                        accountZipOK={this.state.accountZipOK}
                        onChange={this.handleAccountAddressChange}
                        onVATBlur={delayDecorator(this.handleAccountVATBlur)}
                    />
                    <CreditCardForm
                        onOKChange={this.handleCreditCardOKChange}
                        showWarning={false}/>
                    <AddressCheckbox sameAddress={this.state.sameAddress}
                        onChange={this.handleAddressCheckboxChange} />
                    {this.state.sameAddress || (
                        <BillingAddress billingName={this.state.billingName}
                            billingNameOK={this.state.billingNameOK}
                            billingCountry={this.state.billingCountry}
                            billingCountryOK={this.state.billingCountryOK}
                            billingAddress1={this.state.billingAddress1}
                            billingAddress1OK={this.state.billingAddress1OK}
                            billingAddress2={this.state.billingAddress2}
                            billingState={this.state.billingState}
                            billingStateOK={this.state.billingStateOK}
                            billingCity={this.state.billingCity}
                            billingCityOK={this.state.billingCityOK}
                            billingZip={this.state.billingZip}
                            billingZipOK={this.state.billingZipOK}
                            onChange={this.handleBillingAddressChange}  />)}
                    {this.state.stripeError &&
                        <p className="p-payment__details-error">
                            <span>Error:&nbsp;</span>{this.state.stripeError}
                        </p>
                    }
                    <div className="buttons">
                        <button disabled={this.state.amSubmitting} type="button" onClick={this.onCancel}>Cancel</button>
                        <button disabled={this.state.amSubmitting} className="primary p-payment__button" type="submit" onClick={this.refreshFieldValidity}>Add payment details</button>
                    </div>
                </form>
            </div>);
    }
}

_AddBillingDetailsForm.propTypes = {
    accountVAT: PropTypes.string.isRequired,
    accountType: PropTypes.string.isRequired,
    shippingAddress: PropTypes.shape({
        country: PropTypes.string.isRequired,
        name: PropTypes.string.isRequired,
        address1: PropTypes.string.isRequired,
        address2: PropTypes.string.isRequired,
        city: PropTypes.string.isRequired,
        state: PropTypes.string.isRequired,
        zip: PropTypes.string.isRequired
    }).isRequired,
    formToken: PropTypes.string.isRequired,
    stripeClientSecret: PropTypes.string.isRequired,
    stripe: PropTypes.object,
};

export const AddBillingDetailsForm = injectStripe(_AddBillingDetailsForm);
