import PropTypes from 'prop-types';
import React from 'react';

import { formatMoney } from 'accounting';
import { Receipts, paymentsPropType } from './receipts';
import { CreditCardDetails, CreditCardError, MissingCreditCard, RemoveCardConfirmation, cardPropType } from './credit-card-details';

// Format money from an integer value in 1/100th of cent
function formatCentiCents(centicents) {
    return formatMoney(centicents / 10000);
}


// Format a date in the format YYYY-MM-DD.
function formatDate(datestr) {
    var months = [
        "January", "February", "March", "April", "May", "June", "July", "August",
        "September", "October", "November", "December"];
    var date = new Date(datestr);
    return [
        date.getUTCDate(),
        months[date.getUTCMonth()],
        date.getUTCFullYear()
    ].join(" ");
}


// Format a date period from two date strings in the format YYYY-MM-DD.
function formatDatePeriod(startDate, endDate) {
    var months = [
        "January", "February", "March", "April", "May", "June", "July", "August",
        "September", "October", "November", "December"];
    var start = new Date(startDate);
    var end = new Date(endDate);
    var tokens = [start.getUTCDate(), months[start.getUTCMonth()]];
    if (start.getUTCFullYear() != end.getUTCFullYear()) {
        tokens.push(start.getUTCFullYear());
    }
    tokens = tokens.concat("-", [end.getUTCDate(), months[end.getUTCMonth()], end.getUTCFullYear()]);
    return tokens.join(" ");
}


export class BillingPage extends React.Component {

    constructor(props) {
        super(props);
        this.removeCard = this.removeCard.bind(this);
        this.getSubmitData = this.getSubmitData.bind(this);
        this.state = {
            card: this.props.card,
        };
    }

    getSubmitData() {
        let form = new FormData();
        form.append('form-security-token', this.props.formToken);
        return form;
    }

    removeCard() {
        let fetchConfig = {
            credentials: 'same-origin',
            method: 'POST',
            body: this.getSubmitData(),
        };
        fetch('.', fetchConfig).then((response) => {
            if (!response.ok) {
                throw new Error('Failure to remove card.');
            }
            // If it's OK then the card was successfully removed.
            this.nullifyCard();
        }).catch(function(error) {
            throw new Error('Failure to remove card: ' + error.message);
        });
    }

    nullifyCard() {
        this.setState({card: null});
    }

    _makeCreditCardComponent(stripeError, card, removeCard) {
        if (stripeError) {
            return <CreditCardError />;
        } else {
            if (card !== null) {
                let cardButtons = (
                    <div className="buttons u-clearfix">
                        <RemoveCardConfirmation removeCard={removeCard}/>
                        <a className="button-link" href="./edit">Edit payment details</a>
                    </div>);
                if (!this.props.canEdit) {
                    cardButtons = (<div className="buttons u-clearfix"></div>);
                }
                return <CreditCardDetails card={card} displayAddress={true} buttons={cardButtons}/>;
            }
            return <MissingCreditCard />;
        }
    }

    _makePaymentFailure( failure ) {
        if ( failure ) {
            return (
                <PaymentFailure>
                    <p>We encountered a billing error while processing your outstanding payment.</p>
                    <p>
                    Reason: { failure.message }<br />
                    Due Date: { failure.date }<br />
                    Amount Outstanding: { formatCentiCents(failure.due) }<br />
                    </p>
                    <p>Please correct your card or address details and we will attempt to re-run the payment in about an hour.</p>
                </PaymentFailure>
            );
        }
    }

    _makePaymentAction(action) {
        if (action) {
            return (
                <PaymentFailure>
                    <p>An outstanding payment requires a validation from the card&apos;s owner.</p>
                    <p><a className="button-link" href={action.url}>Validate this payment</a></p>
                    <p>Once validated we will try to process this payment again in about an hour.</p>
                </PaymentFailure>
            );
        }
    }

    render() {
        return (
            <div>
                <h2>Billing</h2>
                { this._makePaymentFailure(this.props.failedPayment) }
                { this._makePaymentAction(this.props.paymentAction) }
                <div id="usage">
                    <h3>Current usage</h3>
                    <div>
                        <CurrentUsage startDate={this.props.usage.period[0]} endDate={this.props.usage.period[1]}/>
                        <Cost cost={this.props.usage.cost} />
                    </div>
                </div>
                <Receipts payments={this.props.payments}/>
                <div id="payment-details">
                    <h3>
                        Payment details &nbsp;&nbsp;
                        <small>
                            <i className="p-icon--padlock p-icon--title"></i> 128-bit SSL encrypted payment
                        </small>
                    </h3>
                    { this._makeCreditCardComponent(this.props.stripeError, this.state.card, this.removeCard) }
                </div>
                <div id="help">
                    <h3>Help</h3>
                    <span>
                        For billing questions, please contact <a href="https://forms.canonical.com/saas/">Canonical Landscape Support</a>.
                    </span>
                </div>
            </div>
        );
    }
}

const creditPropTypes = PropTypes.shape({
    amount: PropTypes.number.isRequired,
    expiry: PropTypes.string.isRequired,
});

const costPropTypes = PropTypes.shape({
    due: PropTypes.number.isRequired,
    usage: PropTypes.number.isRequired,
    vat: PropTypes.number.isRequired,
    vat_rate: PropTypes.number.isRequired,
    used_credit: PropTypes.number.isRequired,
    remaining_credit: PropTypes.number.isRequired,
    remaining_credits: PropTypes.arrayOf(creditPropTypes),
});

BillingPage.propTypes = {
    usage: PropTypes.shape({
        period: PropTypes.arrayOf(PropTypes.string).isRequired,
        usage: PropTypes.array,
        cost: costPropTypes
    }),
    formToken: PropTypes.string.isRequired,
    stripeError: PropTypes.string,
    failedPayment: PropTypes.shape({
        message: PropTypes.string,
        date: PropTypes.string,
        total: PropTypes.integer,
    }),
    paymentAction: PropTypes.shape({
        url: PropTypes.string,
    }),
    canEdit: PropTypes.bool,
    card: cardPropType,
    payments: paymentsPropType,
};
    

class CurrentUsage extends React.Component {
    render() {
        var period = formatDatePeriod(this.props.startDate, this.props.endDate);
        return (
            <div id="current-usage">
                Current billing period: {period}
            </div>);
    }
}

CurrentUsage.propTypes = {
    startDate: PropTypes.string.isRequired,
    endDate: PropTypes.string.isRequired,
};


class Cost extends React.Component {
    render() {
        var netTotal = formatCentiCents(this.props.cost.usage);
        var vat = formatCentiCents(this.props.cost.vat);
        var usedCredit = formatCentiCents(this.props.cost.used_credit);
        var total = formatCentiCents(this.props.cost.due);
        return (
            <div id="current-cost">
                <table id="cost">
                    <tbody>
                        <tr/>
                        <tr>
                            <th>Usage (before tax)</th>
                            <td>{netTotal}</td>
                        </tr>
                        <tr>
                            <th>Used credit</th>
                            <td>-{usedCredit}</td>
                        </tr>
                        <tr id="last-subtotal">
                            <th>VAT</th>
                            <td>{vat}</td>
                        </tr>
                        <tr id="total-row">
                            <th>Amount due</th>
                            <td>{total}</td>
                        </tr>
                        <tr id="remaining-credits-row">
                            <th>Remaining credit</th>
                            <td>
                                <Credits total={this.props.cost.remaining_credit} credits={this.props.cost.remaining_credits} />
                            </td>
                        </tr>
                    </tbody>
                </table>
            </div>
        );
    }
}

Cost.propTypes = {
    cost: costPropTypes,
};


class Credits extends React.Component {
    render() {
        var total = formatCentiCents(this.props.total);

        if (this.props.credits.length == 0) {
            return (
                <span>{total}</span>
            );
        }

        var date;
        var rows = [];
        if (this.props.credits.length == 1 && !this.props.forceList) {
            date = formatDate(this.props.credits[0].expiry);
            rows.push(
                <tr id="credit-expiry-row" key="expiry">
                    <td/>
                    <td>(expires on {date})</td>
                </tr>
            );
        } else {
            rows.push(
                <tr key="header">
                    <th>Amount</th>
                    <th>Expires on</th>
                </tr>
            );
            for (var i = 0; i < this.props.credits.length; i++) {
                var amount = formatCentiCents(this.props.credits[i].amount);
                date = formatDate(this.props.credits[i].expiry);
                rows.push(
                    <tr key={i.toString()}>
                        <td>{amount}</td>
                        <td>{date}</td>
                    </tr>
                );
            }
        }

        return (
            <table id="credits">
                <tbody>
                    <tr id="credits-total-row">
                        <td/>
                        <td>{total}</td>
                    </tr>
                    {rows}
                </tbody>
            </table>
        );
    }
}

Credits.propTypes = {
    total: PropTypes.number.isRequired,
    credits: PropTypes.arrayOf(creditPropTypes),
    forceList: PropTypes.bool,
};


// Render an error message in a div
export function PaymentFailure(props) {
    return (
        <div className="p-error__payment">
            { props.children }
        </div>
    );
}


PaymentFailure.propTypes = {
    children: PropTypes.node
};
