import PropTypes from 'prop-types';
import React from 'react';
import Modal from 'react-modal';

// Compose a label and an input field.
export function LabeledInput(props) {
    // The input is a built in HTML element, therefore we can't modify
    // it.  Thus when we use spread notation ("...") below, we can't
    // just use the props as passed in.  An error will be raised if we
    // attempt to give the input any children, or labelText props.
    //
    // To that end, we clone props to inputProps and kill those two
    // keys.
    let inputProps = Object.assign({}, props);
    delete inputProps.children;
    delete inputProps.labelText;
    return (
        <div>
            <label htmlFor={props.name} className={props.className}>{props.labelText}</label>
            <input {...inputProps} />
            {props.children}
        </div>
    );
}

LabeledInput.propTypes = {
    name: PropTypes.string.isRequired,
    labelText: PropTypes.string.isRequired,
    type: PropTypes.string.isRequired,
    onChange: PropTypes.func,
    placeholder: PropTypes.string,
    value: PropTypes.string,
    className: PropTypes.string,
    children: PropTypes.node
};


// A LabeledInput that can show the result of validation.
//
// This component doesn't perform any validation, but rather indicates
// the positive or negative results of validation (including any
// messages).  This in turn means the component has no opinion on how
// that validation should occur.
export class ValidatedLabeledInput extends React.Component {

    constructor(props) {
        super(props);
        this.onChange = this.onChange.bind(this);
    }

    onChange(event) {
        this.props.onChange(event);
    }

    render() {
        let cls;
        let errorMessage;
        let hasValidation = this.props.isOK !== undefined;

        if (hasValidation) {
            cls = this.props.className || '';
            if (this.props.isOK) {
                cls += ' has-success';
            } else {
                cls += ' has-error';
                errorMessage = this.props.errorMessage;
            }
        }

        // don't pass on the isOK and errorMessage properties
        let inputProps = Object.assign({}, this.props);
        delete inputProps.isOK;
        delete inputProps.errorMessage;

        return (
            <LabeledInput {...inputProps}
                className={cls}
                onChange={this.onChange}>
                {hasValidation && !this.props.isOK &&
                 <div className="p-payment__details-error">
                     <span>Error:&nbsp;</span> 
                     {errorMessage}
                 </div>
                }
            </LabeledInput>
        );
    }
}

ValidatedLabeledInput.propTypes = {
    name: PropTypes.string.isRequired,
    labelText: PropTypes.string.isRequired,
    type: PropTypes.string.isRequired,
    onChange: PropTypes.func,
    placeholder: PropTypes.string,
    value: PropTypes.string,
    isOK: PropTypes.bool,
    errorMessage: PropTypes.string,
    className: PropTypes.string
};


export function getFieldState(value) {
    if (value === undefined) {
        return undefined;
    } else if (value === "") {
        return false;
    } else {
        return true;
    }
}

const modalStyles = {overlay: {cursor: 'wait'}, content: {display: 'none'}};

export function ModalInterstitial(props) {
    return <Modal isOpen={props.isOpen} contentLabel="White noise" style={modalStyles} />;
}

ModalInterstitial.propTypes = {
    isOpen: PropTypes.bool,
};

ModalInterstitial.defaultProps = {
    isOpen: false
};
