import PropTypes from 'prop-types';
import React from 'react';

import { AddressCheckbox, AccountAddress, BillingAddress } from './address';
import { AccountType } from './account-type';
import { getCountryState, getStateState } from './country';
import { isEUCountry } from './eu';
import { isValidVATIN } from './vat';
import { CreditCardForm } from './credit-card-form';
import { CreditCardDetails } from './credit-card-details';
import { getFieldState, ModalInterstitial } from './components';
import { PaymentDetailsHeader } from './payment-details';
import { TermsForm } from './terms-form';
import { delayDecorator } from './utility';
import { injectStripe } from 'react-stripe-elements';

export class _SignupForm extends React.Component {

    constructor(props) {
        super(props);
        this.state = {
            stripeError: null,
            editDetails: true,
            card: null,
            accountType: "personal",
            accountName: "",
            accountCountry: "",
            accountVAT: "",
            accountAddress1: "",
            accountAddress2: "",
            accountState: "",
            accountCity: "",
            accountZip: "",
            sameAddress: false,
            billingName: "",
            billingCountry: "",
            billingAddress1: "",
            billingAddress2: "",
            billingState: "",
            billingCity: "",
            billingZip: "",
            accountUpdates: false,
            paymentMethod: "",
            creditCardOK: true,  // Stripe form is initially submittable
        };
        this.refreshFieldValidity = this.refreshFieldValidity.bind(this);
        this.onSubmit = this.onSubmit.bind(this);
        this.handleAddressCheckboxChange = this.handleAddressCheckboxChange.bind(this);
        this.handleBillingAddressChange = this.handleBillingAddressChange.bind(this);
        this.handleAccountAddressChange = this.handleAccountAddressChange.bind(this);
        this.handleCreditCardOKChange = this.handleCreditCardOKChange.bind(this);
        this.handleAccountVATBlur = this.handleAccountVATBlur.bind(this);
        this.handleAccountTypeChange = this.handleAccountTypeChange.bind(this);
        this.handleUpdatesChange = this.handleUpdatesChange.bind(this);
        this.editCardDetails = this.editCardDetails.bind(this);
        this.finalizeSignup = this.finalizeSignup.bind(this);
        this.simpleAddressFields = [
            "billingName",
            "billingAddress1",
            "billingCity",
            "billingZip",
            "accountName",
            "accountAddress1",
            "accountCity",
            "accountZip",
        ];
    }

    handleAddressCheckboxChange(sameAddress) {
        this.setState({sameAddress: sameAddress});
    }

    handleAccountAddressChange(items) {
        for (let key in items) {
            // AccountAddress is only allowed to update account address fields
            if (key.startsWith("account")) {
                let stateUpdate = {[key]: items[key]};
                if (this.simpleAddressFields.includes(key)) {
                    stateUpdate[key + "OK"] = getFieldState(items[key]);
                } else if (key === "accountState") {
                    stateUpdate["accountStateOK"] = getStateState(items[key], this.state.accountCountry);
                } else if (key === "accountCountry") {
                    let [isVATOK, errorMessage] = isValidVATIN(this.state.accountVAT, items[key]);
                    stateUpdate["accountCountryOK"] = getCountryState(items[key]);
                    stateUpdate["accountVATOK"] = isVATOK;
                    stateUpdate["accountVATErrorMessage"] = errorMessage;
                    if (this.state.accountStateOK !== undefined) {
                        stateUpdate["accountStateOK"] = getStateState(this.state.accountState, items[key]);
                    }
                }
                this.setState(stateUpdate);
            }
        }
    }

    handleBillingAddressChange(items) {
        for (let key in items) {
            // BillingAddress is only allowed to update billing address fields
            if (key.startsWith("billing")) {
                let stateUpdate = {[key]: items[key]};
                if (this.simpleAddressFields.includes(key)) {
                    stateUpdate[key + "OK"] = getFieldState(items[key]);
                } else if (key === "billingState") {
                    stateUpdate["billingStateOK"] = getStateState(items[key], this.state.billingCountry);
                } else if (key === "billingCountry") {
                    stateUpdate["billingCountryOK"] = getCountryState(items[key]);
                    if (this.state.billingStateOK !== undefined) {
                        stateUpdate["billingStateOK"] = getStateState(this.state.billingState, items[key]);
                    }
                }
                this.setState(stateUpdate);
            }
        }
    }

    handleCreditCardOKChange(isOK) {
        this.setState({creditCardOK: isOK});
    }

    handleAccountVATBlur() {
        const [isVATOK, errorMessage] = isValidVATIN(this.state.accountVAT, this.state.accountCountry);
        this.setState({accountVATOK: isVATOK, accountVATErrorMessage: errorMessage});
    }

    handleAccountTypeChange(value) {
        this.setState({accountType: value});
    }

    handleUpdatesChange(value) {
        this.setState({accountUpdates: value});
    }

    editCardDetails() {
        this.setState({editDetails: true});
    }

    // Click handler when submitting
    refreshFieldValidity() {
        let billingUpdate = {};
        let accountUpdate = {};
        // Coerce empty fields to the empty string
        this.simpleAddressFields.map(key => {
            let map;
            if (key.startsWith("billing")) {
                map = billingUpdate;
            } else if (key.startsWith("account")) {
                map = accountUpdate;
            } else {
                throw "unexpected key";
            }
            map[key] = this.state[key] || "";
        });
        billingUpdate["billingState"] = this.state["billingState"] || "";
        billingUpdate["billingCountry"] = this.state["billingCountry"] || "";
        accountUpdate["accountState"] = this.state["accountState"] || "";
        accountUpdate["accountCountry"] = this.state["accountCountry"] || "";
        this.handleBillingAddressChange(billingUpdate);
        this.handleAccountAddressChange(accountUpdate);
    }

    // Is ths form ready to be submitted?
    isSubmittable() {
        let mustHaveFields = [
            this.state.creditCardOK,
            this.state.accountNameOK,
            this.state.accountAddress1OK,
            this.state.accountCityOK,
            this.state.accountStateOK,
            this.state.accountZipOK,
            this.state.accountCountryOK,
        ];
        if (this.state.accountVATOK !== undefined) {
            // undefined VAT is fine
            mustHaveFields.push(this.state.accountVATOK);
        }
        const billingFields = [
            this.state.billingNameOK,
            this.state.billingAddress1OK,
            this.state.billingCityOK,
            this.state.billingStateOK,
            this.state.billingZipOK,
            this.state.billingCountryOK,
        ];
        if ( !this.state.sameAddress ) {
            mustHaveFields.push(...billingFields);
        }

        return mustHaveFields.every(field => Boolean(field));
    }

    _getBillingDetails() {
        if (this.state.sameAddress) {
            return {
                "name": this.state.accountName,
                "address": {
                    "city": this.state.accountCity,
                    "country": this.state.accountCountry,
                    "line1": this.state.accountAddress1,
                    "line2": this.state.accountAddress2,
                    "postal_code": this.state.accountZip,
                    "state": this.state.accountState,
                }
            };
        }
        return {
            "name": this.state.billingName,
            "address": {
                "city": this.state.billingCity,
                "country": this.state.billingCountry,
                "line1": this.state.billingAddress1,
                "line2": this.state.billingAddress2,
                "postal_code": this.state.billingZip,
                "state": this.state.billingState,
            }
        };
    }

    onSubmit(event) {
        event.preventDefault();
        if (! this.isSubmittable()) {
            return;
        }
        this.setState({stripeError: null});
        this.props.stripe.createPaymentMethod(
            'card', {billing_details: this._getBillingDetails()}
        ).then((response) => {
            if (response.error) {
                this.setState({stripeError: response.error});
            } else {
                this.setState({
                    editDetails: false,
                    paymentMethod: response.paymentMethod.id,
                    card: response.paymentMethod.card});
            }
        });

    }

    finalizeSignup(event) {
        event.preventDefault();
        if (this.state.amSubmitting) {
            // don't double-POST in case a form element gets focus despite the modal overlay
            return;
        }
        this.setState({amSubmitting: true});

        // finalize setupIntent before posting, allowing for interactive bank
        // authentication, if SCA is required
        this.props.stripe.handleCardSetup(
            this.props.stripeClientSecret,
            {payment_method: this.state.paymentMethod}
        ).then((result) => {
            if (result.error) {
                this.setState({
                    stripeError: result.error.message,
                    editDetails: true,
                    amSubmitting: false
                });
                return;
            }

            let form = this._getSubmitData();
            form.set("stripeSetupIntent", result.setupIntent.id);
            let fetchConfig = {
                credentials: 'same-origin',
                method: 'POST',
                body: form
            };
            fetch('./signup', fetchConfig).then(() => {
                // Reload whether successful or not, as any change requires
                // restarting the SetupIntent flow from scratch.
                window.location.reload();
            }).catch((error) => {
                this.setState({amSubmitting: false});
                throw new Error('Failure to complete signup: ' + error.message);
            });
        });
    }

    // Return data to be submitted from the form.
    _getSubmitData() {
        let vatNumber = "";
        let isBusiness;
        // Only post the VAT number if it's a business account
        if (this.state.accountType === "business") {
            if (isEUCountry(this.state.accountCountry)) {
                vatNumber = this.state.accountVAT;
            }
            isBusiness = true;
        } else {
            isBusiness = false;
        }
        let form = new FormData();
        form.set("VATcountry", this.state["accountCountry"]);
        form.set("VATnumber", vatNumber);
        form.set("isBusiness", JSON.stringify(isBusiness));
        form.set("accountName", this.state["accountName"]);
        form.set("accountAddress1", this.state["accountAddress1"]);
        form.set("accountAddress2", this.state["accountAddress2"]);
        form.set("accountState", this.state["accountState"]);
        form.set("accountCity", this.state["accountCity"]);
        form.set("accountZip", this.state["accountZip"]);
        form.set("accountUpdates", this.state["accountUpdates"]);
        form.set("form-security-token", this.props.formToken);
        return form;
    }

    componentDidUpdate() {
        let error = document.querySelector(".p-payment__error");
        if (error) {
            error.scrollIntoView(true);
        }
    }

    render() {
        if (this.state.editDetails) {
            let boxClass = "box box-shaped p-box";
            if (!this.props.isLoggedIn) {
                boxClass = boxClass.concat(" is-disabled");
            }
            return (
                <div>
                    <div className={boxClass}>
                        <div className="p-payment">
                            <PaymentDetailsHeader isComplete={false} />
                            <form id="stripeForm" action="" method="POST" onSubmit={this.onSubmit}>
                                <AccountType type={this.state.accountType} onChange={this.handleAccountTypeChange} />
                                <AccountAddress isStripeAddress={this.state.sameAddress}
                                    isBusinessAccount={this.state.accountType == "business"}
                                    accountName={this.state.accountName}
                                    accountNameOK={this.state.accountNameOK}
                                    accountCountry={this.state.accountCountry}
                                    accountCountryOK={this.state.accountCountryOK}
                                    accountVAT={this.state.accountVAT}
                                    accountVATOK={this.state.accountVATOK}
                                    accountVATErrorMessage={this.state.accountVATErrorMessage}
                                    accountAddress1={this.state.accountAddress1}
                                    accountAddress1OK={this.state.accountAddress1OK}
                                    accountAddress2={this.state.accountAddress2}
                                    accountState={this.state.accountState}
                                    accountStateOK={this.state.accountStateOK}
                                    accountCity={this.state.accountCity}
                                    accountCityOK={this.state.accountCityOK}
                                    accountZip={this.state.accountZip}
                                    accountZipOK={this.state.accountZipOK}
                                    onChange={this.handleAccountAddressChange}
                                    onVATBlur={delayDecorator(this.handleAccountVATBlur)}
                                />
                                <CreditCardForm
                                    onOKChange={this.handleCreditCardOKChange}
                                    showWarning={true} />
                                <AddressCheckbox sameAddress={this.state.sameAddress}
                                    onChange={this.handleAddressCheckboxChange} />
                                {this.state.sameAddress || (
                                    <BillingAddress billingName={this.state.billingName}
                                        billingNameOK={this.state.billingNameOK}
                                        billingCountry={this.state.billingCountry}
                                        billingCountryOK={this.state.billingCountryOK}
                                        billingAddress1={this.state.billingAddress1}
                                        billingAddress1OK={this.state.billingAddress1OK}
                                        billingAddress2={this.state.billingAddress2}
                                        billingState={this.state.billingState}
                                        billingStateOK={this.state.billingStateOK}
                                        billingCity={this.state.billingCity}
                                        billingCityOK={this.state.billingCityOK}
                                        billingZip={this.state.billingZip}
                                        billingZipOK={this.state.billingZipOK}
                                        onChange={this.handleBillingAddressChange} />)}
                                {this.state.stripeError &&
                                    <p className="p-payment__error">Something went wrong: {this.state.stripeError}</p>}
                                <button className="p-payment__button" type="submit" value="Submit" onClick={this.refreshFieldValidity}>Add payment details</button>
                            </form>
                        </div>
                    </div>
                    <TermsForm isDisabled={true} onChange={this.handleUpdatesChange} />
                </div>);
        } else {
            let editCardButton = <button className="p-payment__button u-clearfix" onClick={this.editCardDetails}>Edit payment details</button>;
            return (
                <div>
                    <ModalInterstitial isOpen={this.state.amSubmitting} />
                    <div className="box box-shaped p-box">
                        <div className="p-payment">
                            <PaymentDetailsHeader isComplete={true} />
                            <CreditCardDetails card={this.state.card} displayAddress={true} buttons={editCardButton} />
                        </div>
                    </div>
                    <TermsForm isDisabled={false} onChange={this.handleUpdatesChange} onSubmit={this.finalizeSignup} />
                    {this.state.stripeError &&
                        <p className="p-payment__error">Something went wrong: {this.state.stripeError}</p>}
                </div>);
        }
    }

}

_SignupForm.propTypes = {
    formToken: PropTypes.string.isRequired,
    isLoggedIn: PropTypes.bool.isRequired,
    stripeClientSecret: PropTypes.string.isRequired,
    stripe: PropTypes.object,
};

export const SignupForm = injectStripe(_SignupForm);
