# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
from email import charset, encoders
from email.header import Header
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.utils import formataddr, parseaddr

from zope.interface import implementer
from zope.component import getUtility
from zope.sendmail.interfaces import IMailDelivery

from canonical.lib.mail import idna_encode_email_address
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.model.configuration import (
    ISystemConfiguration,
    LANDSCAPE_SYSTEM_EMAIL,
)


# By default the content transfer encoding for utf-8 is base64. We make
# it quoted printable, so that the e-mails can be read without having to
# decode them first.
del charset.CHARSETS["utf-8"]
charset.add_charset("utf-8", charset.SHORTEST, charset.QP, "utf-8")


def format_address(name, email):
    """Format a name and email as an e-mail header string.

    It works like email.Util.formataddr(), except that it handles
    unicode name properly.
    """
    email = idna_encode_email_address(email)
    if not name:
        return email
    name = str(Header(name))
    return formataddr((name, email))


@implementer(ILandscapeMailer)
class LandscapeMailer(object):
    """A convenience wrapper around zope.sendmail.delivery.QueuedMailDelivery.

    This class provides a more friendly interface for sending mails from
    the application code.
    """

    def send(
        self,
        to_addresses,
        subject,
        message_content,
        sender=None,
        bulk=False,
        attachments=None,
    ):
        """Send an email:

        @param to_addresses: a list of addresses to sent the email to
        @param subject: the subject of the email
        @param message_content: the content of the email
        @param sender: optionally the email sender
        @param bulk: if True, set the Precedence header to Bulk
        @param attachments: a 2-tuple of (filename, content-type) of files
            to attach to the email
        """
        mailer = getUtility(IMailDelivery, "landscape-mailer")
        text_message = MIMEText(message_content.encode("utf-8"), "plain", "utf-8")
        if attachments:
            message = MIMEMultipart()
            message.attach(text_message)
        else:
            message = text_message

        configuration = getUtility(ISystemConfiguration)
        if sender is None:
            sender = configuration.get(
                LANDSCAPE_SYSTEM_EMAIL, "landscape-team@canonical.com"
            )
        message["From"] = sender
        message["To"] = ", ".join(to_addresses)
        message["Subject"] = subject
        if bulk:
            message.add_header("Precedence", "Bulk")
        if attachments:
            self._add_attachments(message, attachments)
        mail_addresses = [parseaddr(address)[1] for address in to_addresses]
        mailer.send(sender, mail_addresses, message.as_string())

    def _add_attachments(self, message, attachments):
        """Attach files to the message."""
        for attachment, content_type in attachments:
            part = MIMEBase(*content_type.split("/", 1))
            with open(attachment, "rb") as fd:
                part.set_payload(fd.read())
            encoders.encode_base64(part)
            # Set the filename parameter
            part.add_header(
                "Content-Disposition",
                "attachment",
                filename=os.path.basename(attachment),
            )
            message.attach(part)


landscape_mailer = LandscapeMailer()
