# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
import os
from textwrap import fill

from zope.interface import implementer
from zope.component import provideAdapter, queryMultiAdapter

from canonical.landscape.mail.interfaces import IEmailView


def register_email_adapter(context_class, email_view_class):
    """
    Register a multi-adapter that adapts C{(context_class, Person)} to
    L{IEmailView}.

    @param context_class: The context class to pass to the view.
    @param email_view_class: The view class adapting the context and the
        person.
    """
    from canonical.landscape.model.main.person import Person

    provideAdapter(email_view_class, (context_class, Person), IEmailView)


def get_email_view(context, person):
    """Get the appropriate email view for C{(context, person)}."""
    return queryMultiAdapter((context, person), IEmailView)


def load_template(module_filename, template_filename):
    """Read a template file and return its content.

    @param module_filename: the filename of the Python module in which
        C{template_filename} should be.
    @param template_filename: the filename of the template to load.
    """
    template_directory = os.path.abspath(os.path.dirname(module_filename))
    template_path = os.path.join(template_directory, template_filename)
    return open(template_path).read()


@implementer(IEmailView)
class EmailViewBase(object):
    """Base class for email views in Landscape.

    A a sub-class of L{EmailViewBase} provides the To, Subject and body
    elements of an email to be sent to a L{Person} regarding a context object.

    Sub-classes must implement the L{subject} and L{body} methods and
    optionally the L{to} one (which defaults to L{Person.email}. They
    can make use of the rendering features provided by this base class.

    See the docstring for L{_render} for more details.
    """

    def __init__(self, context, person):
        """
        @param context: The context object the view is based on.
        @param person: An arbitrary object representing the person receiving
            the email.  This can be used by view code to, for example, generate
            localized times and dates.
        """
        self._context = context
        self._person = person

    def _fit(self, text, length, fill=" "):
        """
        Returns C{text} with spaces appended to make it C{length} characters
        long, or truncates C{text} if it is longer than C{length}.
        """
        if len(text) > length:
            return text[:length]
        return text.ljust(length, fill)

    def _bullet(self, text):
        """
        Returns C{text} rendered as a bullet point, with text wrapping at 68
        characters.
        """
        return fill(text, initial_indent=" * ", subsequent_indent="   ", width=68)

    def _truncate(self, text, width):
        """
        Returns C{text} truncated if it exceeds C{width}.  A C{...} suffix is
        added when truncation occurs.
        """
        if len(text) > width:
            return "%s..." % text[: width - 3]
        return text

    def _separator(self, width=67):
        """Returns text to use as a separator."""
        return "-" * width

    def _render(self, template):
        """Render the contents of a template text.

        @param template: A template text in Genshi format. Attributes
            set on the view are directly available in the template text
            as C{$attribute_name}.
        """

        # We import genshi here to avoid a dependency on genshi just to run
        # basic tools like 'schema'.
        from genshi.template.text import TextTemplate

        data = self.__dict__.copy()
        data.update(
            {
                "fit": self._fit,
                "bullet": self._bullet,
                "truncate": self._truncate,
                "separator": self._separator,
                "formatter": self._person.formatter,
            }
        )
        return TextTemplate(template).generate(**data).render("text", encoding=None)

    @property
    def to(self):
        """Return a list of recipients for the To field of the email."""
        return [self._person.email]

    @property
    def subject(self):
        """Return a string to be used in the Subject field of the email."""
        raise NotImplementedError("Subclasses must implement this method!")

    @property
    def body(self):
        """Return the body of the email."""
        raise NotImplementedError("Subclasses must implement this method!")
