# Copyright 2005-2022 Canonical Limited. All rights reserved.

import json
import time
from urllib.parse import urlencode, urljoin

import pycurl
from io import StringIO


class MarketoAPIError(Exception):
    """Generic API error wrapper."""


class MarketoAuthError(MarketoAPIError):
    """Marketo authentication error."""


class MarketoClient(object):
    """API client for marketo requests."""

    def __init__(
        self,
        url,
        client_id,
        client_secret,
        timeout=15,
        _curl=pycurl.Curl,
        _time=time.time,
    ):
        """
        Create a new marketo client.

        @param url: The Marketo api base URL
        @param client_id: the Marketo API client ID
        @param client_secret: the marketo API client secret key
        """
        self.url = url
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout = timeout
        self._access_token = None
        self._auth_expires = None
        self._time = _time
        self._curl = _curl

    def get_leads(self, filter_type, filter_values):
        """Get filtered leads.

        @param filter_type: a field name (e.g. "landscapeEnvironment")
        @param filter_values: a list of field values (e.g. "staging")
        """
        url = self._get_leads_url()
        params = urlencode(
            {"filterType": filter_type, "filterValues": ",".join(filter_values)}
        )
        url = "{}?{}".format(url, params)
        _, result = self._request(url)

        try:
            success = result["success"]
        except KeyError:
            raise MarketoAPIError("Unexpected leads get response", result)
        if not success:
            raise MarketoAPIError("Failed to get leads", result)
        return result["result"]

    def create_lead(self, fields):
        """Create a new lead from fields."""
        self.create_or_update_lead(fields, action="createOnly")

    def create_or_update_lead(self, fields, lookup=None, action="createOrUpdate"):
        """Create or update a lead.

        @param fields: a dict of fields to set on the lead.
        @param lookup: a lookup field used for updates. Defaults to email.
        @param action: one of "createOrUpdate", "createOnly", "updateOnly"
        """
        data = {
            "action": action,
            "input": [fields],
        }
        if lookup is not None:
            data["lookupField"] = lookup
        url = self._get_leads_url()
        _, result = self._request(url, method="POST", data=data)
        try:
            success = result["success"]
        except KeyError:
            raise MarketoAPIError("Unexpected leads post response", result)
        if not success:
            raise MarketoAPIError("Failed to post leads", result)

    def delete_leads(self, lead_ids):
        """Delete a list of leads by id.

        @param lead_ids: a list of marketo leads ids to delete
        """
        url = self._get_leads_url() + "?_method=DELETE"
        data = {"input": [{"id": lead_id} for lead_id in lead_ids]}
        _, result = self._request(url, method="POST", data=data)
        try:
            success = result["success"]
        except KeyError:
            raise MarketoAPIError("Unexpected leads delete response", result)
        if not success:
            raise MarketoAPIError("Failed to delete leads", result)

    def _get_leads_url(self):
        """Returns the lead API URL."""
        return urljoin(self.url, "/rest/v1/leads.json")

    def is_authenticated(self):
        """
        Returns whether this client thinks it has a valid
        access token at this moment.
        """
        return self._access_token and (self._time() < self._auth_expires)

    def _request(self, url, method="GET", data=None, headers=None):
        """Perform an authenticated request to marketo.

        The client will try to request/refresh a valid
        access_token if it doesn't already have one or if the
        server rejects the request because an an expired token.

        @param url: the url of the request
        @param method: HTTP method, either "POST" or "GET"
        @param data: request http post data
        @param headers: a dict of extra HTTP headers to set or override
        @returns a (status, response) tuple
        """
        if not self.is_authenticated():
            self._authenticate()
        request_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer {}".format(self._access_token),
        }
        if headers is not None:
            request_headers.update(headers)
        post_data = json.dumps(data) if data is not None else None

        status_code, response = self._anonymous_request(
            url, method=method, data=post_data, headers=request_headers
        )
        if status_code == 602:
            # Server considers token as expired.
            self._auth_expires = None
            self._access_token = None
            self._authenticate()
        return status_code, response

    def _anonymous_request(self, url, method="GET", data=None, headers=None):
        """Make an unauthenticated request to the marketo service.

        @param url: the request url
        @param method: the request HTTP method, either "GET" or "POST"
        @param data: the request POST body
        @param headers: HTTP headers, as dict
        """
        if headers is None:
            headers = []
        else:
            headers = [": ".join(i) for i in list(headers.items())]

        # disable expect entries, which confuses the service
        headers.append("Expect:")

        response = StringIO()
        curl = self._curl()
        if method == "POST":
            curl.setopt(pycurl.POST, True)
        elif method != "GET":
            raise NotImplementedError

        if data is not None:
            curl.setopt(pycurl.READFUNCTION, StringIO(data).read)
            headers.append("Content-Length: {}".format(len(data)))

        # Suppress signals
        curl.setopt(pycurl.NOSIGNAL, True)
        # Follow redirects
        curl.setopt(pycurl.FOLLOWLOCATION, True)
        curl.setopt(pycurl.MAXREDIRS, 5)
        # Set timeouts
        if self.timeout:
            curl.setopt(pycurl.CONNECTTIMEOUT, self.timeout)
            curl.setopt(pycurl.TIMEOUT, self.timeout)
        curl.setopt(pycurl.HTTPHEADER, headers)
        curl.setopt(pycurl.WRITEFUNCTION, response.write)
        curl.setopt(pycurl.URL, url)
        try:
            curl.perform()
            status = curl.getinfo(pycurl.RESPONSE_CODE)
        except pycurl.error as err:
            raise MarketoAPIError("Unexpected response", err)

        result = response.getvalue()
        try:
            return status, json.loads(result)
        except ValueError as err:
            raise MarketoAPIError("Invalid json response", err)

    def _authenticate(self):
        """Try to authenticate to marketo.

        Upon successful authentication, the access token and its
        expiration are cached on the client for use in requests.
        """

        response = StringIO()
        auth_params = urlencode(
            {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }
        )
        auth_url = urljoin(self.url, "/identity/oauth/token?{}".format(auth_params))
        status, response = self._anonymous_request(url=auth_url)

        if status != 200:
            raise MarketoAuthError(response)

        try:
            self._auth_expires = self._time() + response["expires_in"]
            self._access_token = response["access_token"]
        except KeyError:
            raise MarketoAuthError(
                "Expecting access_token and expires_in in auth response"
            )
