# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
import time
import uuid
from distutils.version import StrictVersion

from hashlib import md5
from datetime import datetime, timedelta

import transaction

from zope.component import getUtility, queryUtility
from zope.security.proxy import removeSecurityProxy

from twisted.web.http import PAYMENT_REQUIRED

from landscape.lib import bpickle

from canonical.message.api import MessageAPI

from canonical.oops.interfaces import IOOPSManager

from canonical.landscape.security.authentication import (
    COMPUTER_ID_HEADER,
    EXCHANGE_TOKEN_HEADER,
)
from canonical.landscape.security.interfaces import IPendingComputerRequest
from canonical.landscape.interfaces import IFileBasedConfiguration

from canonical.landscape.model.main.serveridentity import get_server_uuid
from canonical.landscape.model.main.computer import get_computer_by_secure_id
from canonical.landscape.model.main.license import ProLicenseType
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.message.nextexpected import got_next_expected
from canonical.landscape.message.handlers.ubuntu_pro import update_pro_licensing
from canonical.landscape.message.interfaces import ILoadShaperFactory, ConsistencyError
from canonical.landscape.model.activity.resynchronize import resynchronize_computers
from canonical.landscape.publication.publication import request_should_be_retried


SOFT_LOCK_TIMEOUT = timedelta(minutes=10)

USER_AGENT = "landscape-client/"


class LandscapeMessageAPI(MessageAPI):
    """Mixin for type-based message APIs."""

    _payload = None
    _payload_size = 0

    @property
    def client_api(self):
        """Return the API version reported by the client."""
        payload, _ = self._get_payload(["client-api"])
        return payload["client-api"]

    def _fail_on_big_message(self, snippet):
        """Helper function that throws exception on message being too large"""
        self._logger.error("Ignored big message starting with {}".format(snippet))
        raise AssertionError("Message too big!")

    def _get_content_length(self):
        content_length = self.request.getHeader("content-length")
        try:
            content_length = int(content_length)
            return content_length
        except (ValueError, TypeError):
            return None

    def _read_message(self):
        """
        Returns the bytes from the twisted request. If message is too big,
        throws an AssertionError
        """

        config = getUtility(IFileBasedConfiguration)
        max_msg_size_bytes = config.get(
            "message-server", key="max-msg-size-bytes", default=30 * (1000 * 1000)
        )  # 30MB
        message_snippet_bytes = config.get(
            "message-server", key="message-snippet-bytes", default=1000
        )
        max_msg_size_bytes = int(max_msg_size_bytes)
        message_snippet_bytes = int(message_snippet_bytes)

        content_length = self._get_content_length()
        if content_length:
            if content_length > max_msg_size_bytes:
                self._logger.error(
                    "Failed on {:.2f} MB message!".format(content_length / 1e6)
                )
                snippet = self.request.content.read(message_snippet_bytes)
                self._fail_on_big_message(snippet)

        data = self.request.content.read(max_msg_size_bytes)

        # In case header size is wrong, check if there is a next byte
        if self.request.content.read(1):
            self._logger.error(
                "Failed on {:.2f} MB message bad header len!".format(
                    max_msg_size_bytes / 1e6
                )
            )
            snippet = data[:message_snippet_bytes]
            self._fail_on_big_message(snippet)

        return data

    def _get_payload(self, required_fields):
        """Read and parse the request body.

        @param required_fields: A list of field names that the payload is
            expected to have. If not present, an AssertionError is raised.

        @return. The unmarshalled payload along with it size in kilobytes
        If message size is too big, will raise an assertion error
        """
        if self._payload is None:

            data = self._read_message()

            self._payload = bpickle.loads(data)
            self._payload_size = len(data) / 1024.0
            del data

        for field in required_fields:
            assert field in self._payload, "Invalid payload received"

        return self._payload, self._payload_size

    def _get_server_api(self, request):
        """Return the most recent API we support."""
        available_apis = self.system.get_api_names(self.request)
        highest_api = available_apis[0]
        for available_api in available_apis[1:]:
            if StrictVersion(available_api) >= StrictVersion(highest_api):
                highest_api = available_api
        return highest_api

    def _generate_response_data(self, payload, response_payload):
        """Finish the response payload and return the response bpickle dump.

        This will add the server UUID to the response payload, the highest
        API version this server can handle, and the 'accepted-types' payload
        item (if necessary), and will finally return the bpickle dump of the
        response payload.

        The 'accepted-types' payload item is only added if the client
        is out-of-date with respect to what it thinks the server will
        accept.  In this case, we send it a message updating this
        knowledge.
        """
        config = getUtility(IFileBasedConfiguration)
        try:
            ping_interval = int(config.get("message-server", "ping-interval"))
        except KeyError:
            ping_interval = None
        if ping_interval:
            message = {"type": "set-intervals", "ping": ping_interval}
            response_payload["messages"].append(message)

        if "accepted-types" in payload:
            types = sorted(name for (name, handler) in self.get_handlers())
            digest = md5(";".join(types).encode()).digest()
            if digest != payload["accepted-types"]:
                message = {"type": "accepted-types", "types": types}
                response_payload["messages"].insert(0, message)

        response_payload["server-uuid"] = str(get_server_uuid()).encode()
        response_payload["server-api"] = self._get_server_api(self.request)

        return bpickle.dumps(response_payload)


class ComputerMessageAPI(LandscapeMessageAPI):
    """Process exchanges coming from registered L{Computer}s.

    @ivar max_response_messages: Maximum number of messages to deliver to the
        client in the response body.
    """

    max_response_messages = 100

    def __init__(self, system, request, computer=None, transaction=transaction):
        super(ComputerMessageAPI, self).__init__(system, request)
        self.transaction = transaction

        if computer:
            self.computer = computer
        else:
            self.computer = request.principal.computer

        self._message_store = self.computer.get_message_store()
        self._logger = logging.getLogger("canonical.landscape.message")
        factory = getUtility(ILoadShaperFactory)
        self._load_shaper = factory.get_load_shaper(
            self.computer.get_store_load_average
        )

    def send(self, type, message):
        assert "type" not in message
        message["type"] = type
        self._message_store.add(message)

    def run(self):
        """Handle all messages in the payload.

        If an error occurs during the processing of messages, or
        during parsing the payload itself, the C{consecutive_errors}
        counter on C{self.computer} will be incremented. If no errors
        occur, then the counter will be set back to 0.
        """
        response_payload = {
            "next-expected-sequence": self.computer.next_expected_sequence,
            "messages": [],
        }
        empty_response = self._generate_response_data({}, response_payload)

        if not self._ensure_license():
            client_api = self.client_api
            try:
                client_api = client_api.decode()
            except (AttributeError, UnicodeDecodeError):
                pass
            if client_api >= "3.7":
                self.request.setResponseCode(PAYMENT_REQUIRED)
            return empty_response
        if not self._acquire_soft_lock():
            return empty_response

        try:
            response, errored = self._run()
        except Exception as exception:
            # Abort the transaction, since it might be in a unusable
            # state. This will let the "finally" block still run the
            # needed database changes.
            self.transaction.abort()
            # If we get an exception that will be retried by the transactor,
            # don't consider this failure a final error yet, just re-raise
            # the exception so the run() method will be executed again.
            errored = not request_should_be_retried(self.request, exception)
            raise
        finally:
            if errored:
                self.computer.consecutive_errors += 1
                self._logger.error(
                    "Error processing the payload for "
                    "computer %d, consecutive errors: %d"
                    % (self.computer.id, self.computer.consecutive_errors)
                )
            else:
                self.computer.consecutive_errors = 0
            self.computer.last_exchange_time = datetime.utcnow()
            self.computer.soft_lock_time = None  # Reset the soft lock
            # Commit now to make sure that the changes done in this
            # finally statement really get done. In case of an
            # exception, the raised exception would get to the transactor
            # which would abort the transaction. We're doing our own error
            # handling, aborting the transaction as needed, so we don't need
            # the transactor to abort the transaction for us, and more
            # importantly we want the database changes in this finally clause
            # to be committed no matter what.
            self.transaction.commit()

        return response

    def _ensure_license(self):
        """
        Ensure that the computer has at least one valid license, so the message
        exchange can be allowed.

        @return: a boolean indicating whether the computer has a valid license.
        """
        if self.computer.account.is_disabled:
            return False

        # No license check should be performed for self-service accounts
        if self.computer.account.is_self_service:
            return True

        management = LicenseManagement(self.computer.account)

        current_license = removeSecurityProxy(self.computer.licenses.one())

        if self.computer.has_valid_license():
            if current_license.get_used_overflow_seats():
                non_overflowed = management.auto_assign_license_to_computer(
                    self.computer
                )
                if non_overflowed:
                    logging.info(
                        "Computer %d reassigned to a non-overflowed license"
                        % self.computer.id
                    )
        else:
            logging.warn("Computer %d has no valid licenses." % self.computer.id)

            has_valid_license = management.auto_assign_license_to_computer(
                self.computer
            )
            if has_valid_license:
                logging.info(
                    "Auto-assigned license seat for computer %d." % self.computer.id
                )
            else:
                # If the account isn't expired, then record that an exchange
                # was done so that it's possible to discover which computers
                # are actively trying to connect but which have no valid
                # license. Otherwise do nothing, so we don't get in the way
                # of the expire-accounts job by creating database contention.
                if not self.computer.account.has_expired():
                    self.computer.last_exchange_time = datetime.utcnow()
                return False
        return True

    def _acquire_soft_lock(self):
        """
        Our "soft locking" mechanism prevents the same computer from
        exchanging more than once at the same time.  We do that by
        recording the time at which an exchange starts in the
        soft_lock_time computer attribute.  Then if the same computer
        connects again, we'll notice that this attribute is set, and
        send it away before the first instance finishes running and
        resets it.  If the current soft_lock_time exceeds the preset
        timeout period, we allow the computer to go through.  That means
        it won't deadlock if something bad happens (e.g. server dies).
        """
        if self.computer.soft_lock_time is not None:
            remaining = SOFT_LOCK_TIMEOUT - (
                datetime.utcnow() - self.computer.soft_lock_time
            )
            if remaining > timedelta(0):
                logging.error(
                    "Computer %d is soft-locked (remaining %s)!"
                    % (self.computer.id, remaining)
                )
                return False
            else:
                logging.warning(
                    "Computer %d has an expired soft-lock." % self.computer.id
                )

        self.computer.soft_lock_time = datetime.utcnow()
        return True

    def _run(self):
        required_fields = ("messages", "sequence", "next-expected-sequence")
        payload, payload_size = self._get_payload(required_fields)

        # Clean and realign the message store. Do that before starting
        # to process any received messages, so that a next-expected-sequence
        # that has a gap is handled gracefully.
        store = self._message_store
        got_next_expected(store, payload["next-expected-sequence"])

        if self.computer.next_expected_sequence is None:
            self.computer.next_expected_sequence = payload["sequence"]

        # Get the client message API and version.
        client_api = payload.get("client-api")
        try:
            client_api = client_api.encode()
        except (AttributeError, UnicodeDecodeError):
            pass
        client_version = self.request.getHeader("User-Agent")[len(USER_AGENT) :]
        self.computer.current_api = client_api
        self.computer.current_version = client_version.encode()

        # Clients with API version greater than 3.5 support exchange tokens. We
        # now generate a new token that we will send to the client in the
        # response payload and that the client will have to give us back at the
        # next exchange. This prevents cloning clients, because assuming that
        # two clients with same secure ID have the valid token at a given time,
        # only the client exchanging first will be authenticated and receive a
        # new token for further exchanges.
        if client_api is not None and client_api >= b"3.5":
            self.computer.next_exchange_token = str(uuid.uuid4()).encode()

        if "client-accepted-types" in payload:
            types = payload["client-accepted-types"]
            types = [t.decode() if isinstance(t, bytes) else t for t in types]
            # Checking the hash here isn't strictly necessary, but we
            # want to be sure that clients won't cause extra load by
            # sending the currently accepted types all the time.
            message_types_hash = md5(";".join(sorted(types)).encode()).digest()
            if message_types_hash != self.computer.accepted_message_types_hash:
                self.computer.set_accepted_message_types([x for x in types])

        # Persist this information.  If the message store was created
        # in the constructor, it'll be saved here as well.
        self.transaction.commit()

        errored = self._process_messages(payload, payload_size)

        activities = self.computer.get_deliverable_activities()
        for activity in activities:
            activity.start()

        # Prepare response.
        response_messages = store.get_pending_messages(self.max_response_messages)
        response_payload = {
            "next-expected-sequence": self.computer.next_expected_sequence,
            "next-exchange-token": self.computer.next_exchange_token,
            "messages": removeSecurityProxy(response_messages),
        }
        if self.computer.accepted_message_types_hash is not None:
            response_payload[
                "client-accepted-types-hash"
            ] = self.computer.accepted_message_types_hash

        return self._generate_response_data(payload, response_payload), errored

    def _process_messages(self, payload, payload_size):
        """
        Process each message in the payload.
        """
        sequence = payload["sequence"]
        payload_messages = len(payload["messages"])
        remaining_messages = payload.get("total-messages", payload_messages)
        processed_messages = 0
        ignored_messages = 0
        consistency_error_scopes = set()
        errored = False

        if sequence <= self.computer.next_expected_sequence:

            for message in payload["messages"]:

                if sequence < self.computer.next_expected_sequence:
                    sequence += 1
                    remaining_messages -= 1
                    ignored_messages += 1
                    continue

                if self._load_shaper.get_remaining_time() <= 0:
                    message = (
                        "Refusing further messages for computer %d "
                        "(load=%.2f)" % (self.computer.id, self._load_shaper.get_load())
                    )
                    self._logger.warning(message)
                    break

                remaining_messages -= 1
                processed_messages += 1

                try:
                    try:
                        self.handle(message["type"], message)
                    except ConsistencyError as e:
                        consistency_error_scopes.add(e.scope)
                        logging.exception(
                            "Consistency Error handling "
                            "message %r "
                            "for computer %d: %s"
                            % (message["type"], self.computer.id, repr(message))
                        )
                        self.transaction.abort()
                    sequence += 1
                    self.computer.next_expected_sequence = sequence
                    self.computer.soft_lock_time = datetime.utcnow()
                    self.transaction.commit()
                except Exception as exception:
                    if request_should_be_retried(self.request, exception):
                        raise
                    oops_manager = queryUtility(IOOPSManager)
                    oops_info = ""
                    if oops_manager:
                        oops = oops_manager.get("request")
                        if oops:
                            oops_info = " (OOPS %s)" % oops_manager.dump(oops)
                    self.transaction.abort()
                    self._logger.exception(
                        "Error handling message %r "
                        "for computer %d%s: %s"
                        % (message["type"], self.computer.id, oops_info, repr(message))
                    )
                    errored = True
                    break

        else:
            ignored_messages = payload_messages

        dropped_messages = payload_messages - processed_messages - ignored_messages

        if consistency_error_scopes:
            self._handle_inconsistency(remaining_messages, consistency_error_scopes)

        duration = time.time() - self._load_shaper.get_start_time()
        load = self._load_shaper.get_load() or 0.0
        message = (
            "Processed messages for computer %d (size=%.2fkb, "
            "duration=%.2fs, load=%.2f, ignored=%d, processed=%d, "
            "dropped=%d, remaining=%d, client=%s)"
            % (
                self.computer.id,
                payload_size,
                duration,
                load,
                ignored_messages,
                processed_messages,
                dropped_messages,
                remaining_messages,
                self.computer.current_version,
            )
        )
        self.computer.last_exchange_duration = duration
        self.computer.last_exchange_size = payload_size
        self.computer.ignored_count = ignored_messages
        self.computer.processed_count = processed_messages
        self.computer.dropped_count = dropped_messages
        self.computer.remaining_count = remaining_messages
        self._logger.info(message)
        return errored

    def _handle_inconsistency(self, remaining_messages, scopes):
        """Handle an inconsistency detected while processing a message."""
        if scopes in ([], None):
            scopes = None
        else:
            scopes = list(scopes)
        if not self.computer.is_inconsistent:
            # Flag the computer as being inconsistent and send a
            # resynchronization request.
            self._logger.info(
                "Initiating resynchronization for " "computer %d" % self.computer.id
            )
            self.computer.is_inconsistent = True
            # Scopes needs to be a list so it can be embedded in a message
            # later - sets cannot be marshalled
            account = removeSecurityProxy(self.computer.account)
            resynchronize_computers(account, [self.computer.id], scopes=scopes)
        else:
            # The computer is already flagged as being inconsistent, this means
            # we have already sent it a resynchronization request. However we
            # didn't get any response to it in this batch of messages (see also
            # handlers/resynchronize.py). If there are still messages in the
            # client's outgoing queue, there's hope that one of them is a the
            # resynchronization response we're waiting for, otherwise let's
            # send a new resynchronization request.
            if remaining_messages > 0:
                self._logger.info(
                    "Resynchronization already pending for "
                    "computer %d" % self.computer.id
                )
            else:
                self._logger.warn(
                    "Computer %d lost previous "
                    "resynchronization request. Requesting "
                    "it again." % self.computer.id
                )
                account = removeSecurityProxy(self.computer.account)
                resynchronize_computers(account, [self.computer.id], scopes=scopes)


class AnonymousComputerMessageAPI(LandscapeMessageAPI):
    def __init__(self, system, request):
        super(AnonymousComputerMessageAPI, self).__init__(system, request)

        self._response_messages = []
        self._logger = logging.getLogger("canonical.landscape.message")

    def send(self, type, message):
        assert "type" not in message
        message["type"] = type
        self._response_messages.append(message)

    def run(self):
        required_fields = ("messages",)
        payload, _ = self._get_payload(required_fields)

        messages = payload["messages"]

        is_pending_computer = IPendingComputerRequest.providedBy(self.request)
        secure_id = self.request.getHeader(COMPUTER_ID_HEADER)
        if secure_id and not is_pending_computer:
            # The client is providing a secure ID, but this not associated with
            # a computer or a pending computer. Most probably a computer with
            # this secure ID has been removed, or the client was talking to
            # a different Landscape server before. In either case, let's just
            # discard any messages sent to us and tell the client that we don't
            # know about it, so it can reset itself and ask for a new
            # registration.
            messages = ()
            body = {}
            if self.request.getHeader(EXCHANGE_TOKEN_HEADER) is not None:
                # The client has also provided an invalid exchange token, so
                # it might be a clone. In that case include the title of the
                # original computer in the response so we can show it when
                # the client will try to register again.
                computer = get_computer_by_secure_id(secure_id)
                if computer:
                    self._logger.warn("Detected clone of computer %d" % computer.id)
                    body["clone-of"] = computer.title
            self.send("unknown-id", body)

        for message in messages:
            try:
                self.handle(message["type"], message)
            except Exception as exception:
                transaction.abort()
                if request_should_be_retried(self.request, exception):
                    self._response_messages = []
                    raise
                else:
                    self._logger.exception(
                        "Error handling message %r: %s"
                        % (message["type"], repr(message))
                    )
                    break
            else:
                transaction.commit()

        response_payload = {"messages": self._response_messages}

        return self._generate_response_data(payload, response_payload)


class ComputerMessageAPI_3_0(ComputerMessageAPI):
    """Message API for registered computers, version 3.0."""


class PendingComputerMessageAPI_3_0(AnonymousComputerMessageAPI):
    """Message API for pending computers, version 3.0."""


class AnonymousComputerMessageAPI_3_0(AnonymousComputerMessageAPI):
    """Message API for anonymous computers, version 3.0."""


class ComputerMessageAPI_3_1(ComputerMessageAPI_3_0):
    """Message API for registered computers, version 3.1."""


class PendingComputerMessageAPI_3_1(PendingComputerMessageAPI_3_0):
    """Message API for pending computers, version 3.1."""


class AnonymousComputerMessageAPI_3_1(AnonymousComputerMessageAPI_3_0):
    """Message API for anonymous computers, version 3.1."""


class ComputerMessageAPI_3_2(ComputerMessageAPI_3_1):
    """Message API for registered computers, version 3.2."""


class PendingComputerMessageAPI_3_2(PendingComputerMessageAPI_3_1):
    """Message API for pending computers, version 3.2."""


class AnonymousComputerMessageAPI_3_2(AnonymousComputerMessageAPI_3_1):
    """Message API for anonymous computers, version 3.2."""


class ComputerMessageAPI_3_3(ComputerMessageAPI_3_2):
    """Message API for registered computers, version 3.3."""


class PendingComputerMessageAPI_3_3(PendingComputerMessageAPI_3_2):
    """Message API for pending computers, version 3.3."""


class AnonymousComputerMessageAPI_3_3(AnonymousComputerMessageAPI_3_2):
    """Message API for anonymous computers, version 3.3."""
