from builtins import object
import logging
import os
import random

from twisted.internet import inotify
from twisted.python import filepath

from canonical.landscape.constants import ROOT_DIR


class BackoffStatus(object):
    """
    Figures out if server should backoff a request based on the existence of a
    backoff file. The backoff file can contain a number between 0 and 100
    inclusive which determines what percentage of requests to backoff.
    """

    FILENAME = "backoff.txt"

    def __init__(self, watch_dirpath=None, default_percent=50):

        if not watch_dirpath:
            watch_dirpath = os.path.join(ROOT_DIR, "configs")

        self._backoff_filepath = os.path.join(watch_dirpath, self.FILENAME)
        self._percent_denied = 0
        self._default_percent = default_percent

        self._set_percent()  # Initialize based on current status of file

        # Start watching the file
        self._notifier = inotify.INotify()
        self._notifier.startReading()
        try:
            self._notifier.watch(
                filepath.FilePath(watch_dirpath), callbacks=[self._set_percent_trigger]
            )
        except inotify.INotifyError:
            logging.warning(
                "Invalid directory {} for backoff file!".format(watch_dirpath)
            )

    def _set_percent_trigger(self, ignored, filepath, mask):
        """
        A helper function that calls the update to the backoff percentage only
        when the backoff file is changed via the inotify event
        """
        if filepath.basename().decode("utf8") == self.FILENAME:
            self._set_percent()

    def _set_percent(self):
        """
        Updates the percentage of requests denied (backed off) based on integer
        value in backoff file. Minimum of 0 (no backoff) and max of 100 (always
        backoff). Invalid value in file results to default backoff. And no file
        results to no backoff (0).
        """
        if os.path.isfile(self._backoff_filepath):
            with open(self._backoff_filepath) as fh:
                text = fh.read()
            try:
                self._percent_denied = int(text.strip())
            except ValueError:
                self._percent_denied = self._default_percent
        else:
            self._percent_denied = 0
        self._percent_denied = max(self._percent_denied, 0)
        self._percent_denied = min(self._percent_denied, 100)
        logging.info("Updating percentage backoff to {}%".format(self._percent_denied))

    def should_backoff(self):
        """
        Calculates if server should deny the specific request based on the
        percentage chance from the backoff file. A percent of 0 will never
        trigger, and percent of 100 will always trigger.
        """
        return bool(random.randint(1, 100) <= self._percent_denied)

    def clean_up(self):
        """Prevents unclean reactor error in tests"""
        self._notifier.stopReading()
