# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import logging

from zope.security.proxy import removeSecurityProxy

from landscape.message_schemas import OPERATION_RESULT

from canonical.lib.arguments import coerce_arguments

from canonical.message.handler import FunctionHandler
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.status import (
    ActivityStatusChangeError,
    SUCCEEDED,
    FAILED,
    NullTransitionError,
)


# Legacy activity status values, used in the message protocol by clients
# released before the activity system rework. These values are kept around for
# backward compatibility, in order to properly handle messages from such
# clients.
LEGACY_FAILED_ID = 5
LEGACY_SUCCEEDED_ID = 6


@coerce_arguments(message=OPERATION_RESULT)
def handle_activity_result(message_api, type, message):
    """
    Handle simple messages to update the state of activities.

    If you need more information in the response for your activities,
    create your own message handler.
    """
    computer = message_api.computer

    unproxied_account = removeSecurityProxy(computer.account)
    activity = get_child_activity(unproxied_account, message["operation-id"])

    status = message["status"]
    result_code = message.get("result-code")
    result_text = message.get("result-text")
    if result_text:
        # NUL characters could be returned by scripts. Although technically
        # valid unicode, they will cause DB errors so we replace them from text
        # output by the unicode "replacement character", like for lossy
        # utf8 decoding.
        result_text = result_text.replace("\0", "\ufffd")
    try:
        if status == LEGACY_SUCCEEDED_ID:
            activity.succeed(code=result_code, text=result_text)
        elif status == LEGACY_FAILED_ID:
            activity.fail(code=result_code, text=result_text)
        else:
            raise RuntimeError("How to handle activity status %d?" % status)
    except (ActivityStatusChangeError, NullTransitionError) as e:
        logging.warning(
            "%s while trying to "
            "handle message: %s\n"
            "The error message was: %s\n"
            "The current status of the activity is: %s\n"
            % (e.__class__.__name__, message, e, activity.status.name)
        )
        # Compare against the new status value
        expected = SUCCEEDED if status == LEGACY_SUCCEEDED_ID else FAILED
        if activity.status != expected:
            raise


activity_result_handler = FunctionHandler(handle_activity_result)
