# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from __future__ import absolute_import

from builtins import str
import logging
import time
import transaction
from datetime import datetime
from pprint import pformat

from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy

from landscape.message_schemas import (
    ADD_PACKAGES,
    UNKNOWN_PACKAGE_HASHES,
    PACKAGES,
    CHANGE_PACKAGES_RESULT,
    PACKAGE_REPORTER_RESULT,
)
from landscape.lib.sequenceranges import ranges_to_sequence

from canonical.lib.dpkg import version_compare
from canonical.lib.arguments import coerce_arguments
from canonical.message.handler import FunctionHandler

from canonical.landscape.message.interfaces import ConsistencyError
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.status import (
    ActivityStatusChangeError,
    NullTransitionError,
)
from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.package.package import (
    is_valid_package_name,
    is_valid_package_version,
)
from canonical.landscape.model.package.errors import (
    PackageSearchRequestError,
    PackageSearchConnectionError,
)
from canonical.landscape.model.package.package import (
    Package,
    DEB_PROVIDES,
    DEB_NAME_PROVIDES,
)
from landscape.lib.apt.package.skeleton import PackageSkeleton
from canonical.landscape.model.package.activities import (
    SUCCESS_RESULT,
    ERROR_RESULT,
    DEPENDENCY_ERROR_RESULT,
)
from canonical.landscape.model.package.alerts import (
    PackageUpgradesAlert,
    SecurityUpgradesAlert,
    PackageProfilesAlert,
    PackageReporterAlert,
)
from canonical.landscape.model.package.search import meets_external_threshold
from canonical.landscape.model.resource.package import (
    PackageReporterResultAPI,
    InvalidPackagesDelta,
    InvalidPackageUpgradesRelations,
    update_packages_buffer,
    clear_packages_buffer,
)
from canonical.landscape.model.usn.issue import USNIssueManagement


@coerce_arguments(message=ADD_PACKAGES)
def handle_add_packages(message_api, type, message):
    """Add new packages to the package market."""
    market = PackageMarket()

    hashes = []

    for package in message["packages"]:
        name = package["name"]
        if not is_valid_package_name(name):
            logging.warning(
                "Dropped invalid package '%s' for "
                "computer %d" % (name, message_api.computer.id)
            )
            continue
        version = package["version"]
        if not is_valid_package_version(version):
            logging.warning(
                "Dropped invalid package '%s' with version '%s' "
                "for computer %d" % (name, version, message_api.computer.id)
            )
            continue
        skeleton = PackageSkeleton(package["type"], name, package["version"])
        skeleton.section = package["section"]
        skeleton.summary = package["summary"]
        skeleton.description = package["description"]
        skeleton.size = package["size"]
        skeleton.installed_size = package["installed-size"]
        skeleton.relations = filter_package_relations(message_api.computer, package)
        package_hash = skeleton.get_hash()
        hashes.append(package_hash)
        if not market.get(package_hash):
            market.add(skeleton)

    message_api.send(
        "package-ids",
        {"ids": market.get_ids(hashes), "request-id": message["request-id"]},
    )


def filter_package_relations(computer, package):
    """
    Takes the 'relations' element of a package uploaded by the client and
    removes any invalid relationships.
    """
    valid_relations = []
    for relationship, version in package["relations"]:
        version = version.strip()
        if relationship in (DEB_PROVIDES, DEB_NAME_PROVIDES) and len(
            version.split()
        ) not in (1, 3):
            logging.warning(
                "Dropped invalid provide relation '%s' for "
                "package '%s %s' and computer %d"
                % (version, package["name"], package["version"], computer.id)
            )
        elif version:
            valid_relations.append((relationship, version))
        else:
            logging.warning(
                "Dropped invalid relation '%s' for package "
                "'%s %s' and computer %d"
                % (version, package["name"], package["version"], computer.id)
            )
    return valid_relations


add_packages_handler = FunctionHandler(handle_add_packages)


@coerce_arguments(message=UNKNOWN_PACKAGE_HASHES)
def handle_unknown_package_hashes(message_api, type, message):
    """Requests package ids for given hashes."""
    market = PackageMarket()
    message_api.send(
        "package-ids",
        {"ids": market.get_ids(message["hashes"]), "request-id": message["request-id"]},
    )


unknown_package_hashes_handler = FunctionHandler(handle_unknown_package_hashes)


@coerce_arguments(message=PACKAGES)
def handle_packages(message_api, type, message):
    """Informs the server of changes in the client package state."""
    installed = []
    held = []
    available = []
    upgrades = []
    autoremovable = []
    security = []
    not_installed = []
    not_available = []
    not_upgrades = []
    not_held = []
    not_autoremovable = []
    not_security = []

    ids_by_status = {
        "installed": installed,
        "available": available,
        "available-upgrades": upgrades,
        "locked": held,
        "autoremovable": autoremovable,
        "security": security,
        "not-installed": not_installed,
        "not-available": not_available,
        "not-available-upgrades": not_upgrades,
        "not-locked": not_held,
        "not-autoremovable": not_autoremovable,
        "not-security": not_security,
    }

    package_ids = set()
    for key, ids in ids_by_status.items():
        for item in ranges_to_sequence(message.get(key, ())):
            if not isinstance(item, (int, int)):
                raise ValueError("Invalid package id: %r" % item)
            if len(ids) > 150000:  # Arbitrary protection from DoS.
                raise ValueError("Too many packages.")
            ids.append(item)
        package_ids.update(ids)

    computer = message_api.computer

    # Check that the IDs we got from the client are actually valid package IDs
    # in the package database.
    market = PackageMarket()
    unknown = sorted(package_ids - set(market.get_many(package_ids).values(Package.id)))
    if unknown:
        raise ConsistencyError(
            "Unknown package IDs in message: %s"
            % (", ".join(str(pkg) for pkg in unknown)),
            "package",
        )

    current_version = computer.current_version
    if current_version and version_compare(current_version, "12.04") < 0:
        # This is not an APT client, let's ignore package locks information
        held[:] = []
        not_held[:] = []

    # Merge buffered packages (if any) with new ones. Commit explicitly since
    # the transaction can be later aborted if an inconsistency is detected.
    unproxied_computer = removeSecurityProxy(computer)
    buffer = update_packages_buffer(
        unproxied_computer,
        available,
        upgrades,
        installed,
        held,
        autoremovable,
        security,
        not_available,
        not_upgrades,
        not_installed,
        not_held,
        not_autoremovable,
        not_security,
    )
    transaction.commit()

    try:
        computer_packages = computer.get_packages()
        computer_packages.update(
            available=buffer.available,
            upgrades=buffer.available_upgrades,
            installed=buffer.installed,
            held=buffer.held,
            autoremovable=buffer.autoremovable,
            security=buffer.security,
            not_available=buffer.not_available,
            not_upgrades=buffer.not_available_upgrades,
            not_installed=buffer.not_installed,
            not_held=buffer.not_held,
            not_autoremovable=buffer.not_autoremovable,
            not_security=buffer.not_security,
        )
    except InvalidPackagesDelta as e:
        raise ConsistencyError(str(e), "package")
    except InvalidPackageUpgradesRelations:
        # We don't resync here, since the state might be temporarily
        # inconsistent when package changes are being reported in multiple
        # messages.  The ComputerPackagesBuffer for the computer is not
        # cleared, so the next time a change is reported it's evaluated again.
        #
        # The current transaction is aborted so that changes are not saved to
        # ComputerPackages.
        transaction.abort()
        logging.info(
            "Package info for computer %d is not consistent, "
            "buffering changes." % computer.id
        )
        return

    # State is consistent and ComputerPackages has been updated, so the buffer
    # can be flushed.
    clear_packages_buffer(unproxied_computer)

    if upgrades or not_upgrades or held or not_held:
        computer.check_rules(PackageUpgradesAlert)
        computer.check_rules(SecurityUpgradesAlert)
        usn_management = USNIssueManagement()
        usn_management.check_usns_for_computer(computer)
    if not_installed:
        # If a profile meta-package gets removed and breaks a profile
        # we have to check the alert rules.
        computer.check_rules(PackageProfilesAlert)

    # Send updates about computer packages to the package search service, if
    # available.
    search_client = getUtility(IPackageSearchClient)
    account = computer.account
    if meets_external_threshold(account):
        try:
            start = time.time()
            # The package-search API requires package lists to be sorted.
            search_client.update_computer_packages(
                account_id=computer.account_id,
                computer_id=computer.id,
                available=sorted(buffer.available),
                upgrades=sorted(buffer.available_upgrades),
                installed=sorted(buffer.installed),
                held=sorted(buffer.held),
                not_available=sorted(buffer.not_available),
                not_upgrades=sorted(buffer.not_available_upgrades),
                not_installed=sorted(buffer.not_installed),
                not_held=sorted(buffer.not_held),
            )
            # XXX temporary debug logging
            logging.info(
                "Package search query for %d took %f"
                % (computer.id, time.time() - start)
            )
        except PackageSearchConnectionError:
            # This means that either the package-search service hasn't started
            # yet or that it's down, in either case we just ignore the error,
            # as it's not critical and the package-search should be good when
            # it eventually comes up again
            #
            # XXX temporary debug logging just to track update attempts
            logging.warning(
                "Could not connect to package-search service and send "
                "message %s for computer %d" % (message, computer.id)
            )
        except PackageSearchRequestError as e:
            # This is an unexpected error, we log it but we don't blow up.
            logging.warning(
                "Error updating the package-search service with "
                "message %s for computer %d : %s" % (message, computer.id, e)
            )


packages_handler = FunctionHandler(handle_packages)


@coerce_arguments(message=CHANGE_PACKAGES_RESULT)
def handle_change_packages_result(message_api, _type, message):
    """
    Handle the answer to a change-packages message sent to the client.
    """
    computer = message_api.computer
    activity_id = message["operation-id"]

    change_packages = get_child_activity(
        removeSecurityProxy(computer.account), activity_id
    )

    if change_packages is None:
        # This covers the edge case where a client reports a result for
        # an activity which a server doesn't have.
        # Such a case could occur when a backup of the server DB gets restored
        # to an older state (against current documentation).
        # Rather than leaving this client wedged, we'll drop that result.
        logging.warning("Invalid package-change reported by computer %d." % computer.id)
        return

    result_code = message["result-code"]
    result_text = message.get("result-text")

    must_install = message.get("must-install", ())
    must_remove = message.get("must-remove", ())

    if None in must_install or None in must_remove:
        change_packages.fail(
            code=ERROR_RESULT,
            text="This request could not be completed because package data "
            "is not available yet.  Newly registered computers upload package "
            "information in batches and it can take a couple of hours for "
            "this process to complete.",
        )
        logging.warning("Missing package ids reported by computer %d." % computer.id)
        return

    try:
        if result_code == SUCCESS_RESULT:
            change_packages.info.add_changes(
                install=must_install, remove=must_remove, complemented=True
            )
            change_packages.succeed(code=result_code, text=result_text)
        elif result_code == ERROR_RESULT:
            change_packages.fail(code=result_code, text=result_text)
            # XXX: we only really care if the activity is actually about
            # applying a profile, so we should filter for this case.
            message_api.computer.check_rules(PackageProfilesAlert)
        elif result_code == DEPENDENCY_ERROR_RESULT:
            change_packages.info.complement_changes(
                must_install,
                must_remove,
                result_code=result_code,
                result_text=result_text,
            )
        else:
            raise RuntimeError(
                "Client sent result-code %d in a "
                "package-change-result message. "
                "What do I do with that!?" % result_code
            )
    except (ActivityStatusChangeError, NullTransitionError) as e:
        logging.warning(
            "Error handling package change message %s for "
            "computer %d : %s" % (pformat(message), computer.id, e)
        )


change_packages_result_handler = FunctionHandler(handle_change_packages_result)


@coerce_arguments(message=PACKAGE_REPORTER_RESULT)
def handle_package_reporter_result(message_api, type, message):
    """Handle a C{package_reporter_result} message."""
    computer = message_api.computer
    api = PackageReporterResultAPI(computer.account)
    error = api.get_package_reporter_result(computer)
    timestamp = message.get("report-timestamp")
    if timestamp is not None:
        timestamp = datetime.utcfromtimestamp(timestamp)
    if error is None:
        api.create_package_reporter_result(
            computer, message["code"], message["err"], timestamp=timestamp
        )
    else:
        error.code = message["code"]
        error.err = message["err"]
        error.timestamp = timestamp
    computer.check_rules(PackageReporterAlert)


package_reporter_result_handler = FunctionHandler(handle_package_reporter_result)
