# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging

from landscape.message_schemas import REGISTER, REGISTER_3_3
from landscape.lib.tag import is_valid_tag

from canonical.lib.text import sanitize_hostname
from canonical.lib.arguments import coerce_arguments
from canonical.message.handler import FunctionHandler
from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.model.main.computer import get_computer_by_secure_id
from canonical.landscape.model.configuration import is_standalone, is_development_mode


MAX_PENDING_COMPUTERS = 20


def create_pending_computer(
    account,
    computer_title,
    computer_hostname,
    message_api,
    tags,
    vm_info=b"",
    container_info="",
    access_group=None,
    check_pending=True,
    juju_model_uuid=None,
    juju_machine_id=None,
    clone_id=None,
    ubuntu_pro_info=None,
):
    """
    Create a pending computer (if there is space for one) with the given
    details and send a response to the client with the ID and insecure ID.

    @param account: the account in which the pending computer is created.
    @param computer_title: the title of the pending computer.
    @param computer_hostname: the hostname of the pending computer.
    @param message_api: the transport to use to send messages to the client.
    @params tags: tags list given by the client, comma separated.
    @param vm_info: String with vm_info type from client.
    @param container_info: String with container_info type from client.
    @param access_group: String with access group from client.
    @param check_pending: Boolean indicating if the pending computer alert and
        pending computers limit should be checked or not.
    @param ubuntu_pro_info: String with Ubuntu Pro info from client.
    """
    pending_exceeded = check_pending and (
        account.get_pending_computer_count() >= MAX_PENDING_COMPUTERS
    )
    if pending_exceeded:
        if message_api.client_api < b"3.8":
            error_code = "unknown-account"
        else:
            error_code = "max-pending-computers"
        message_api.send("registration", {"info": error_code})
        return

    if computer_hostname:
        computer_hostname = sanitize_hostname(computer_hostname)
    # Strip linefeeds from the computer title. since the client does not
    # validate them properly.
    computer_title = computer_title.replace("\n", "").replace("\r", "")
    pending_computer = account.create_pending_computer(
        computer_title,
        "Please consider this computer for your account.",
        hostname=computer_hostname,
        tags=tags,
        vm_info=vm_info,
        container_info=container_info,
        access_group=access_group,
        check_pending=check_pending,
        juju_model_uuid=juju_model_uuid,
        juju_machine_id=juju_machine_id,
        clone_id=clone_id,
        ubuntu_pro_info=ubuntu_pro_info,
    )
    message_api.send(
        "set-id", {"id": pending_computer.secure_id, "insecure-id": pending_computer.id}
    )
    return pending_computer


@coerce_arguments(message=REGISTER)
def handle_register(message_api, _type, message):
    _handle_register(message_api, message)


@coerce_arguments(message=REGISTER_3_3)
def handle_register_3_3(message_api, _type, message):
    _handle_register(message_api, message)


def get_clone_id(clone_secure_id):
    """Get clone id from secure id if it exists otherwise return None"""
    if clone_secure_id:
        computer = get_computer_by_secure_id(clone_secure_id)
        if computer:
            return computer.id
    return None


def _handle_register(message_api, message):
    """
    Handle a request for registration, perhaps creating a L{PendingComputer}.

    If the account name and registration password are correct, a
    L{PendingComputer} will be created for the client, and secure and
    insecure IDs will be sent back.
    """
    account_name = message.get("account_name")
    computer_title = message.get("computer_title")
    computer_hostname = message.get("hostname")
    password = message.get("registration_password")
    tags = message.get("tags")
    vm_info = message.get("vm-info", b"").decode("utf-8")
    container_info = message.get("container-info", "")
    access_group = message.get("access_group")
    juju_info = message.get("juju-info", {})
    juju_model_uuid = juju_info.get("environment-uuid")
    juju_machine_id = juju_info.get("machine-id")
    logger = logging.getLogger("canonical.landscape.message")
    clone_secure_id = message.get("clone_secure_id")
    clone_id = get_clone_id(clone_secure_id)
    ubuntu_pro_info = message.get("ubuntu_pro_info")

    # Given that we have client-side tag verification, using is_valid_tag, if
    # we have invalid tags at this point, something odd is going on, ignore the
    # tags.
    if tags is not None:
        invalid_tags = [
            tag
            for tag in [tag.strip() for tag in tags.split(",")]
            if not is_valid_tag(tag)
        ]
        if invalid_tags:
            tags = None

    assert account_name and computer_title, "Invalid 'register' message: %r" % message

    logger.debug(
        "Received registration request "
        "for account %r from computer %r" % (account_name, computer_title)
    )

    account = get_account_by_name(account_name)

    if not (account and account.check_registration_password(password)):
        message_api.send("registration", {"info": "unknown-account"})
        return

    auto_register = False

    if account.has_registration_password() and is_standalone():
        auto_register = True

    if not (is_development_mode() or is_standalone()):
        # Ignore juju model info with non-lds deployment.
        # For SaaS, juju features are disabled, but juju_models would still
        # cause UUID collisions between accounts when models span over
        # multiple tenants, since clients still report that. (LP: #1920089)
        juju_model_uuid = juju_machine_id = None

    pending_computer = create_pending_computer(
        account,
        computer_title,
        computer_hostname,
        message_api,
        tags,
        vm_info=vm_info,
        container_info=container_info,
        access_group=access_group,
        check_pending=not auto_register,
        juju_model_uuid=juju_model_uuid,
        juju_machine_id=juju_machine_id,
        clone_id=clone_id,
        ubuntu_pro_info=ubuntu_pro_info,
    )

    if auto_register:
        pending_computer.accept(person=None)


register_handler = FunctionHandler(handle_register)
register_handler_3_3 = FunctionHandler(handle_register_3_3)
