# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Handlers for message related to resource monitoring data collection.
"""

import logging

from storm.exceptions import IntegrityError, DataError

from zope.security.proxy import removeSecurityProxy

from landscape.message_schemas import (
    COMPUTER_INFO,
    DISTRIBUTION_INFO,
    PROCESSOR_INFO,
    TEMPERATURE,
    MEMORY_INFO,
    ACTIVE_PROCESS_INFO,
    MOUNT_INFO,
    MOUNT_ACTIVITY,
    FREE_SPACE,
    CUSTOM_GRAPH,
    REBOOT_REQUIRED,
    REBOOT_REQUIRED_INFO,
    APT_PREFERENCES,
    NETWORK_DEVICE,
    NETWORK_ACTIVITY,
    HARDWARE_INFO,
    UPDATE_MANAGER_INFO,
    CPU_USAGE,
    CEPH_USAGE,
    SWIFT_USAGE,
    KEYSTONE_TOKEN,
    CLOUD_METADATA,
)
from landscape.constants import APT_PREFERENCES_SIZE_LIMIT

from canonical.lib.hardware import parse_lshw
from canonical.lib.arguments import coerce_arguments

from canonical.message.handler import FunctionHandler
from canonical.landscape.message.interfaces import ConsistencyError
from canonical.landscape.model.main.distribution import (
    create_distribution,
    get_distribution,
)
from canonical.landscape.model.computer.alerts import (
    ComputerRebootAlert,
    ComputerDuplicateAlert,
)
from canonical.landscape.model.package.alerts import EsmDisabledAlert
from canonical.landscape.model.resource.rebootrequiredinfo import RebootRequiredAPI
from canonical.landscape.model.resource.network import NetworkDeviceAPI
from canonical.landscape.model.resource.hardware import store_hardware_for_computer
from canonical.landscape.model.resource.cpuusage import CPUUsageManagement
from canonical.landscape.model.resource.ceph import CephUsageManagement
from canonical.landscape.model.resource.swift import SwiftUsageManagement
from canonical.landscape.model.resource.keystonetoken import KeystoneTokenManagement
from canonical.landscape.model.resource.annotation import AnnotationManagement
from canonical.landscape.model.resource.cloud_instance_metadata import (
    CloudInstanceMetadataManagement,
)


@coerce_arguments(message=CLOUD_METADATA)
def handle_cloud_metadata(message_api, _type, message):
    """
    Process messages containing cloud-specific information for a computer.
    """
    computer = message_api.computer
    management = CloudInstanceMetadataManagement(computer)
    management.set_instance_metadata(
        message["ami-id"], message["instance-id"], message["instance-type"]
    )


cloud_instance_metadata_handler = FunctionHandler(handle_cloud_metadata)


@coerce_arguments(message=COMPUTER_INFO)
def handle_computer_info(message_api, _type, message):
    """Process messages containing general information about a computer."""
    computer = message_api.computer
    if "hostname" in message:
        computer.hostname = message["hostname"]
        computer.check_rules(ComputerDuplicateAlert)
    if "total-memory" in message:
        computer.total_memory = message["total-memory"]
    if "total-swap" in message:
        computer.total_swap = message["total-swap"]

    annotations = message.get("annotations")
    if annotations:
        process_annotations(computer, annotations)


computer_info_handler = FunctionHandler(handle_computer_info)


def process_annotations(computer, data):
    """Process custom annotations and store them."""
    management = AnnotationManagement(computer.account)
    management.add_annotation([computer.id], data)


@coerce_arguments(message=DISTRIBUTION_INFO)
def handle_distribution_info(message_api, _type, message):
    """
    Process messages containing information about the distribution running on
    a computer.
    """
    distributor = message.get("distributor-id")
    description = message.get("description")
    release = message.get("release")
    code_name = message.get("code-name")
    computer = message_api.computer
    distribution = get_distribution(description, distributor, release, code_name)
    if distribution is None:
        distribution = create_distribution(description, distributor, release, code_name)
    computer.distribution = distribution
    # It is necessary to unblock any outstanding
    # ChangeRepositoryProfileRequest here. They are blocked
    # until the computer has its distribution set, and cannot
    # be unblocked by a human using the UI.
    computer.unblock_change_repository_profile_requests()


distribution_info_handler = FunctionHandler(handle_distribution_info)


@coerce_arguments(message=PROCESSOR_INFO)
def handle_processor_info(message_api, _type, message):
    """Process messages produced by the processor-info plugin."""
    computer = message_api.computer

    computer.remove_processors()

    for processor_details in message["processors"]:
        processor_id = processor_details["processor-id"]
        model = processor_details["model"]
        vendor = processor_details.get("vendor", "")
        cache_size = processor_details.get("cache-size", -1)

        computer.create_processor(processor_id, model, vendor, cache_size)


processor_info_handler = FunctionHandler(handle_processor_info)


@coerce_arguments(message=TEMPERATURE)
def handle_temperature(message_api, _type, message):
    """Process a message produced by the temperature plugin."""
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")

    name = message["thermal-zone"]
    thermal_zone = computer.get_thermal_zone_by_name(name)

    if not thermal_zone:
        thermal_zone = computer.create_thermal_zone(name)

    for timestamp, temperature in message["temperatures"]:
        try:
            thermal_zone.add_temperature(timestamp, temperature)
            store.flush()
        except IntegrityError:
            # For each thermal zone, timestamp must be unique, so trying to add
            # a duplicate will cause an IntegrityError. This can happen with
            # cloned computers, we log a message in this case (see LP:1161856).
            logger.error(
                "Duplicated temperature for computer %d, thermal zone '%s'"
                % (computer.id, name)
            )
            store.rollback()
            break


temperature_handler = FunctionHandler(handle_temperature)


def handle_load_average(message_api, _type, message):
    """Process a message produced by the load-average plugin."""
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")
    for timestamp, load_average in message["load-averages"]:
        try:
            computer.create_load_average(timestamp, load_average)
            store.flush()
        except IntegrityError:
            # For each load average value, timestamp must be unique, so trying
            # to add a duplicate will cause an IntegrityError. This can happen
            # with cloned computers, we log a message in this case (see
            # LP:1161856).
            logger.error("Duplicated load average for computer %d" % computer.id)
            store.rollback()
            break


load_average_handler = FunctionHandler(handle_load_average)


@coerce_arguments(message=CPU_USAGE)
def handle_cpu_usage(message_api, _type, message):
    """
    Process a message produced by the cpuusage plugin.
    """
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")
    manager = CPUUsageManagement(computer)
    for timestamp, cpu_usage in message["cpu-usages"]:
        try:
            manager.create_cpu_usage(timestamp, cpu_usage)
            store.flush()
        except IntegrityError:
            # For each cpu usage value, timestamp must be unique, so trying to
            # add a duplicate will cause an IntegrityError. This can happen
            # with cloned computers, we log a message in this case (see
            # LP:1161856).
            logger.error("Duplicated cpu usage for computer %d" % computer.id)
            store.rollback()
            break


cpu_usage_handler = FunctionHandler(handle_cpu_usage)


@coerce_arguments(message=CEPH_USAGE)
def handle_ceph_usage(message_api, _type, message):
    """
    Process a message produced by the ceph plugin.
    """
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")
    manager = CephUsageManagement(computer)
    ring_uuid = message["ring-id"]
    for timestamp, total, available, used in message["data-points"]:
        try:
            manager.create_ceph_usage(timestamp, ring_uuid, total, available, used)
            store.flush()
        except IntegrityError:
            # For each Ceph ring, timestamp must be unique, so trying to add a
            # duplicate will cause an IntegrityError. This can happen with
            # cloned computers, we log a message in this case (see LP:1161856).
            logger.error(
                "Duplicated timestamp for computer %d, Ceph ring '%s'"
                % (computer.id, ring_uuid)
            )
            store.rollback()
            break


ceph_usage_handler = FunctionHandler(handle_ceph_usage)


@coerce_arguments(message=SWIFT_USAGE)
def handle_swift_usage(message_api, _type, message):
    """
    Process a message produced by the swift plugin.
    """
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")
    manager = SwiftUsageManagement(computer)
    for timestamp, device, total, available, used in message["data-points"]:
        try:
            manager.create_swift_usage(timestamp, device, total, available, used)
            store.flush()
        except IntegrityError:
            # For each device in a computer, timestamp must be unique, so
            # trying to add a duplicate will cause an IntegrityError. This can
            # happen with cloned computers, we log a message in this case (see
            # LP:1161856).
            logger.error(
                "Duplicated timestamp for computer %d, Swift device '%s'"
                % (computer.id, device)
            )
            store.rollback()
            break


swift_usage_handler = FunctionHandler(handle_swift_usage)


@coerce_arguments(message=KEYSTONE_TOKEN)
def handle_keystone_token(message_api, _type, message):
    """
    Process a message provided by the keystone-token plugin.
    """
    management = KeystoneTokenManagement(message_api.computer)
    token = message["data"]
    if token is not None:
        management.set_keystone_token(token)


keystone_token_handler = FunctionHandler(handle_keystone_token)


@coerce_arguments(message=MEMORY_INFO)
def handle_memory_info(message_api, _type, message):
    """Process a message produced by the memory-info plugin."""
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")
    for timestamp, free_memory, free_swap in message["memory-info"]:
        try:
            computer.create_memory_info(timestamp, free_memory, free_swap)
            store.flush()
        except IntegrityError:
            # For each memory info, timestamp must be unique, so trying to add
            # a duplicate will cause an IntegrityError. This can happen with
            # cloned computers, we log a message in this case (see LP:1161856).
            logger.error("Duplicated memory info for computer %d" % computer.id)
            store.rollback()
            break
        except DataError:
            # In some cases such as bug 1602374 /proc/meminfo may report
            # free values that are out of range. As those values are
            # wrong and misleading, we drop them and leave a trace in logs.
            logger.warning("Invalid memory-info values for {}".format(computer.id))
            store.rollback()
            break


memory_info_handler = FunctionHandler(handle_memory_info)


@coerce_arguments(message=ACTIVE_PROCESS_INFO)
def handle_active_process_info(message_api, _type, message):
    """Handle messages from the active-process-info plugin."""
    computer = message_api.computer

    if "kill-all-processes" in message:
        computer.remove_active_processes()

    if "kill-processes" in message:
        killed_pids = message["kill-processes"]
        computer.remove_active_processes_with_pids(killed_pids)

    if "update-processes" in message:
        updates = message["update-processes"]
        process_ids = [process["pid"] for process in updates]
        processes = {}
        for process in computer.get_active_processes_by_pids(process_ids):
            processes[process.pid] = process
        updates.sort(key=lambda data: data["pid"])
        for data in updates:
            process = processes.get(data["pid"])
            if process is None:
                msg = (
                    "Couldn't find process with pid %d for computer %d "
                    "to update" % (data["pid"], computer.id)
                )
                raise ConsistencyError(msg, "process")
            else:
                process.uid = data["uid"]
                process.gid = data["gid"]
                process.name = data["name"]
                process.state = data["state"]
                process.start_time = data["start-time"]
                # Kernel threads don't have a VmSize in their
                # /proc/<pid>/status
                process.vm_size = data.get("vm-size")
                process.cpu_utilisation = data.get("percent-cpu")
                if process.cpu_utilisation is not None:
                    process.cpu_utilisation = round(process.cpu_utilisation, 4)

    if "add-processes" in message:
        for process_info in message["add-processes"]:
            pid = process_info["pid"]
            uid = process_info["uid"]
            gid = process_info["gid"]
            name = process_info["name"]
            state = process_info["state"]
            start_time = process_info["start-time"]
            vm_size = process_info.get("vm-size")
            cpu_utilisation = process_info.get("percent-cpu")
            if cpu_utilisation is not None:
                cpu_utilisation = round(cpu_utilisation, 4)
            computer.create_active_process(
                pid, uid, gid, name, state, start_time, vm_size, cpu_utilisation
            )


active_process_info_handler = FunctionHandler(handle_active_process_info)


@coerce_arguments(message=MOUNT_INFO)
def handle_mount_info(message_api, _type, message):
    """Register new mount information for the computer."""
    computer = message_api.computer

    for timestamp, info in message["mount-info"]:
        mount_info = computer.get_mount_info_for_mount_point(info["mount-point"])
        if mount_info:
            mount_info.device = info["device"]
            mount_info.filesystem = info["filesystem"]
            mount_info.total_space = info["total-space"]
        else:
            computer.create_mount_info(
                info["device"],
                info["mount-point"],
                info["filesystem"],
                info["total-space"],
            )


mount_info_handler = FunctionHandler(handle_mount_info)


@coerce_arguments(message=MOUNT_ACTIVITY)
def handle_mount_activity(message_api, _type, message):
    """
    Handle mount_activity messages, as a no-op for backward compatibility.
    """


mount_activity_handler = FunctionHandler(handle_mount_activity)


@coerce_arguments(message=FREE_SPACE)
def handle_free_space(message_api, _type, message):
    """Attach new free space information to a mount info."""
    computer = message_api.computer
    logger = logging.getLogger("canonical.landscape.message")

    got_inconsistency = False
    store = removeSecurityProxy(computer).store
    for timestamp, mount_point, free_space in message["free-space"]:
        mount_info = computer.get_mount_info_for_mount_point(mount_point)
        if mount_info:
            try:
                mount_info.add_free_space(timestamp, free_space)
                store.flush()
            except IntegrityError:
                # For each mount point, timestamp must be unique, so trying
                # to add a duplicate will cause an IntegrityError. This can
                # happen with cloned computers, we log a message in this
                # case (see LP:1161856).
                logger.error(
                    "Duplicated mount info for computer %d, "
                    "mount point '%s'" % (computer.id, mount_point)
                )
                store.rollback()
                break
        else:
            logger.warn(
                "Unknown mount info for computer %d, "
                "mount point '%s'" % (computer.id, mount_point)
            )
            got_inconsistency = True

    if got_inconsistency:
        raise ConsistencyError("Unknown mount info.", "disk")


free_space_handler = FunctionHandler(handle_free_space)


@coerce_arguments(message=HARDWARE_INFO)
def handle_hardware_info(message_api, _type, message):
    """Report computer hardware."""
    data = message["data"]
    if not data:
        return

    logger = logging.getLogger("canonical.landscape.message")
    store_hardware_for_computer(message_api.computer, parse_lshw(data), logger)


hardware_info_handler = FunctionHandler(handle_hardware_info)


@coerce_arguments(message=CUSTOM_GRAPH)
def handle_custom_graph(message_api, _type, message):
    """Report new custom graph data."""
    computer = message_api.computer
    store = removeSecurityProxy(computer).store
    logger = logging.getLogger("canonical.landscape.message")

    graphs = dict(
        (custom_graph.id, custom_graph) for custom_graph in computer.get_custom_graphs()
    )

    new_graph_ids = list(graphs.keys())
    updated_graph_ids = []
    for graph_id, graph_message in message["data"].items():
        graph = graphs.get(graph_id)
        if graph:
            error = graph_message["error"]
            new_graph_ids.remove(graph_id)
            graph_data = computer.get_custom_graph_data_by_id(graph_id)
            try:
                graph_data.append_data_points(graph_message["values"])
                store.flush()
            except IntegrityError:
                # For each graph, timestamp must be unique, so trying to add a
                # duplicate will cause an IntegrityError. This can happen with
                # cloned computers, we log a message in this case (see
                # LP:1161856).
                logger.error(
                    "Duplicated graph data for computer %d, graph %d"
                    % (computer.id, graph_id)
                )
                store.rollback()
                break
            graph_data.error = error
            if graph_message["script-hash"] != graph.script_hash:
                updated_graph_ids.append(graph_id)
        else:
            logger.warn(
                "Unknown graph id %d from computer %d message: %s"
                % (graph_id, computer.id, message)
            )
            message_api.send("custom-graph-remove", {"graph-id": graph_id})

    new_graph_ids.extend(updated_graph_ids)
    for graph_id in new_graph_ids:
        graph = graphs[graph_id]
        message = {
            "interpreter": graph.interpreter,
            "code": graph.code,
            "username": graph.username,
            "graph-id": graph.id,
            "script-hash": graph.script_hash,
        }
        message_api.send("custom-graph-add", message)


custom_graph_handler = FunctionHandler(handle_custom_graph)


def update_reboot_required_flag(computer, message):
    """Update the C{reboot_required_flag} property of the given L{Computer}."""

    if "flag" in message:
        flag = message["flag"]
        if flag != computer.reboot_required_flag:
            computer.reboot_required_flag = flag
            computer.check_rules(ComputerRebootAlert)


@coerce_arguments(message=REBOOT_REQUIRED)
def handle_reboot_required(message_api, _type, message):
    """Handle old reboot-required messages, without the 'packages' field.

    @note: This can be eventually be dropped once we're confident that
        all clients out there have been upgraded.
    """
    computer = message_api.computer
    update_reboot_required_flag(computer, message)


reboot_required_handler = FunctionHandler(handle_reboot_required)


@coerce_arguments(message=REBOOT_REQUIRED_INFO)
def handle_reboot_required_info(message_api, _type, message):
    """Process a message indicating whether a computer needs a reboot."""
    computer = message_api.computer
    update_reboot_required_flag(computer, message)

    # Before processing the 'packages' field we double check that the flag
    # is actually True, though in practice this condition should always hold.
    reboot_api = RebootRequiredAPI(computer)
    if computer.reboot_required_flag:
        if "packages" in message:
            reboot_required_info = reboot_api.get_reboot_required_info()
            if reboot_required_info is None:
                reboot_required_info = reboot_api.create_reboot_required_info()
            packages = message["packages"]
            if reboot_required_info.packages != packages:
                reboot_required_info.packages = packages
    else:
        reboot_api.remove_reboot_required_info()


reboot_required_info_handler = FunctionHandler(handle_reboot_required_info)


@coerce_arguments(message=UPDATE_MANAGER_INFO)
def handle_update_manager_info(message_api, _type, message):
    """
    Process a message indicating a change in the computers update-manager
    preferences.
    """
    computer = message_api.computer
    prompt = message["prompt"]
    if prompt != computer.update_manager_prompt:
        computer.update_manager_prompt = prompt


update_manager_info_handler = FunctionHandler(handle_update_manager_info)


@coerce_arguments(message=APT_PREFERENCES)
def handle_apt_preferences(message_api, _type, message):
    """Process a message reporting a computer's APT pinning settings."""
    computer_packages = message_api.computer.get_packages()
    if "data" in message:
        apt_preferences_data = message["data"]
        data_size = 0
        if apt_preferences_data:
            for filename, contents in apt_preferences_data.items():
                data_size += len(filename) + len(contents)
            if data_size > APT_PREFERENCES_SIZE_LIMIT:
                raise RuntimeError("APT preferences data message is too big!")
        computer_packages.update_apt_preferences_files(apt_preferences_data)
        message_api.computer.check_rules(EsmDisabledAlert)


apt_preferences_handler = FunctionHandler(handle_apt_preferences)


@coerce_arguments(message=NETWORK_DEVICE)
def handle_network_device(message_api, _type, message):
    """
    Handle a message reporting the list of devices on the computer.

    All the devices are reported if it changed: we need to add the ones we
    don't know about, update the existing ones, and remove the ones which are
    not present anymore.
    """
    seen = []
    network_api = NetworkDeviceAPI(message_api.computer)
    db_devices = dict(
        (device.interface, device) for device in network_api.get_devices()
    )
    for device in message["devices"]:
        interface_name = device["interface"]
        db_device = db_devices.get(interface_name, None)
        if db_device is not None:
            db_device.ip_address = device["ip_address"]
            db_device.mac_address = device["mac_address"]
            db_device.broadcast_address = device["broadcast_address"]
            db_device.netmask = device["netmask"]
            db_device.flags = device["flags"]
            seen.append(interface_name)
        else:
            new_device = network_api.create_device(
                device["interface"],
                device["ip_address"],
                device["mac_address"],
                device["broadcast_address"],
                device["netmask"],
                device["flags"],
            )
            db_devices[device["interface"]] = new_device
            seen.append(interface_name)

    for device in message.get("device-speeds", []):
        db_device = db_devices.get(device["interface"], None)
        # Don't create new interfaces if some extra data is in the device-speed
        if db_device is not None:
            db_device.speed = device["speed"]
            db_device.duplex = device["duplex"]

    for device in db_devices.values():
        if device.interface not in seen:
            network_api.remove_device(device)


network_device_handler = FunctionHandler(handle_network_device)


MAX_NETWORK_ITEMS = 250


@coerce_arguments(message=NETWORK_ACTIVITY)
def handle_network_activity(message_api, _type, message):
    """Handle a message reporting the network traffic on a computer."""
    computer = message_api.computer
    network_api = NetworkDeviceAPI(computer)
    logger = logging.getLogger("canonical.landscape.message")

    store = removeSecurityProxy(computer).store
    count = 0
    for interface, activities in message["activities"].items():
        for activity in activities:
            if count >= MAX_NETWORK_ITEMS:
                logger.warn(
                    "More than %d network items from computer %d, ignoring"
                    " the rest of the message." % (MAX_NETWORK_ITEMS, computer.id)
                )
                return
            for value in (activity[1], activity[2]):
                if not 0 <= value < 9223372036854775807:
                    logger.warn(
                        "Ignoring network value %d from computer %d "
                        "message: %s" % (value, computer.id, message)
                    )
                    break
            else:
                try:
                    network_api.create_traffic(
                        interface, activity[0], activity[1], activity[2]
                    )
                    store.flush()
                except IntegrityError:
                    # For each network device, timestamp must be unique, so
                    # trying to add a duplicate will cause an
                    # IntegrityError. This can happen with cloned computers, we
                    # log a message in this case (see LP:1161856).
                    logger.error(
                        "Duplicated network traffic for computer %d, "
                        "interface '%s'" % (computer.id, interface)
                    )
                    store.rollback()
                    break
                count += 1


network_activity_handler = FunctionHandler(handle_network_activity)
