# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from __future__ import absolute_import

import logging
from pprint import pformat

from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy

from landscape.message_schemas import RESYNCHRONIZE

from canonical.lib.arguments import coerce_arguments
from canonical.message.handler import FunctionHandler
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.status import (
    ActivityStatusChangeError,
    NullTransitionError,
)
from canonical.landscape.model.main.computer import InvalidActivityError
from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.package.search import meets_external_threshold
from canonical.landscape.model.package.errors import (
    PackageSearchRequestError,
    PackageSearchConnectionError,
)


@coerce_arguments(message=RESYNCHRONIZE)
def handle_resynchronize(message_api, type, message):
    """
    Emit an event to the server indicating that a client is resynchronizing.
    """
    operation_id = message.get("operation-id")
    try:
        message_api.computer.resynchronize(operation_id)
    except (ActivityStatusChangeError, NullTransitionError) as e:
        logger = logging.getLogger("canonical.landscape.message")
        logger.warning(
            "%s while trying to "
            "handle message: %s for computer %d, this "
            "resynchronisation has already "
            "completed."
            % (e.__class__.__name__, pformat(message), message_api.computer.id)
        )
    except InvalidActivityError as e:
        # This covers the edge case where a client reports a resync which
        # a server doesn't have. Such a case can occur when a backup
        # of the server DB gets restored to an older state.
        #
        # The best we can do to avoid getting the exchange stuck is to log
        # the error. Subsequent messages are likely to trigger another
        # ConsistencyError with a proper scope. That should successfully
        # resynchronize, as the activity will have been re-created.
        logger = logging.getLogger("canonical.landscape.message").warning(
            "%s while trying to handle message: %s for computer %d, this "
            "resynchronisation is unknown.",
            e.__class__.__name__,
            pformat(message),
            message_api.computer.id,
        )
        return

    computer = message_api.computer
    computer.is_inconsistent = False

    # Notify the package search service about this resynchronization, if it's
    # a package-related resync
    search_client = getUtility(IPackageSearchClient)
    account = computer.account
    is_package_resynchronization = False
    if operation_id:
        activity = get_child_activity(
            removeSecurityProxy(computer.account), operation_id
        )
        scopes = activity.info.scopes
        if scopes is None or "package" in scopes:
            is_package_resynchronization = True
    if is_package_resynchronization and meets_external_threshold(account):
        try:
            search_client.resynchronize(account_id=account.id, computer_id=computer.id)
        except PackageSearchConnectionError:
            # This means that either the package-search service hasn't started
            # yet or that it's down, in either case we just ignore the error,
            # as it's not critical and the package-search should be good when
            # it eventually comes up again
            pass
        except PackageSearchRequestError as e:
            # This is an unexpected error, we log it but we don't blow up.
            logging.warning(
                "Error updating the package-search service with "
                "message %s for computer %d : %s" % (message, computer.id, e)
            )


resynchronize_handler = FunctionHandler(handle_resynchronize)
