import logging
from datetime import datetime

from storm.locals import Store

from landscape.message_schemas import SNAPS

from canonical.lib.arguments import coerce_arguments
from canonical.message.handler import FunctionHandler

from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.snap import InstalledSnap, Snap, SnapPublisher

logger = logging.getLogger("canonical.landscape.message")


class CandidateInstalledSnap:
    """
    A Snap that has been reported as installed but is not yet recorded
    in our database as such. Hashes comparably to an InstalledSnap.
    """

    def __init__(
        self,
        snap_id: str,
        version: str,
        revision: str,
        tracking_channel: str,
        held_until: datetime,
        name: str,
        confinement: str,
        summary: str,
        publisher_username: str,
        publisher_validation: str,
    ):
        self.snap_id = snap_id
        self.version = version
        self.revision = revision
        self.tracking_channel = tracking_channel
        self.name = name
        self.confinement = confinement
        self.summary = summary
        self.publisher_username = publisher_username
        self.publisher_validation = publisher_validation

        try:
            self.held_until = held_until and datetime.fromisoformat(held_until)
        except ValueError:
            logger.info(
                f"Unable to parse snap {name} hold date '{held_until}' as datetime."
            )
            self.held_until = None

    def __hash__(self) -> int:
        return hash((self.snap_id, self.revision))

    def __eq__(self, other) -> bool:
        if hasattr(other, "snap_id") and hasattr(other, "revision"):
            return (self.snap_id, self.revision) == (other.snap_id, other.revision)

        return False

    def to_installed_snap(self, computer: Computer, store: Store) -> InstalledSnap:
        """Converts this object into to a snap record in the database, updating corresponding
        records if they exist.

        :param Computer computer: The computer the snap is installed on.
        :param Store store: The database store to interact with.

        :returns InstalledSnap: The storm object, ready to be committed.
        """
        snap = store.find(Snap, id=self.snap_id).one()

        if snap is not None:
            snap_publisher = snap.publisher

            snap.summary = self.summary
        else:
            # Snap record does not exist, need to create it.
            snap_publisher = store.find(
                SnapPublisher, username=self.publisher_username
            ).one()

            if snap_publisher is None:
                # Publisher record does not exist, need to create it
                # before creating Snap record.
                snap_publisher = store.add(
                    SnapPublisher(
                        self.publisher_username,
                        self.publisher_validation,
                    )
                )

            snap = store.add(
                Snap(
                    self.snap_id,
                    self.name,
                    self.summary,
                    snap_publisher,
                )
            )

        prev_installed_snap = store.find(
            InstalledSnap, computer_id=computer.id, snap_id=snap.id
        ).one()

        if prev_installed_snap:
            prev_installed_snap.version = self.version
            prev_installed_snap.revision = self.revision
            prev_installed_snap.tracking_channel = self.tracking_channel
            prev_installed_snap.held_until = self.held_until
            prev_installed_snap.confinement = self.confinement

            return prev_installed_snap

        installed_snap = store.add(
            InstalledSnap(
                computer.id,
                snap.id,
                self.version,
                self.revision,
                self.tracking_channel,
                self.held_until,
                self.confinement,
            )
        )

        return installed_snap


@coerce_arguments(message=SNAPS)
def handle_snaps(message_api, message_type, message):
    """
    The message contains a list of installed snaps. This handler adds
    this info to the snap database.
    """
    computer = message_api.computer
    store = Store.of(computer)

    candidate_snaps = set(
        CandidateInstalledSnap(
            snap["id"],
            snap["version"],
            snap["revision"],
            snap["tracking-channel"],
            snap.get("hold") and snap.get("hold").split(".")[0],
            snap["name"],
            snap["confinement"],
            snap.get("summary", ""),
            snap["publisher"]["username"],
            snap["publisher"]["validation"],
        )
        for snap in message["snaps"]["installed"]
    )

    # Compare and reconcile the list of reported snaps with those we
    # have in the database.
    prev_snaps = set(snap for snap in computer.snaps)
    to_remove = prev_snaps - candidate_snaps

    for installed_snap in to_remove:
        store.remove(installed_snap)

    for candidate_snap in candidate_snaps:
        candidate_snap.to_installed_snap(computer, store)

    store.flush()


snaps_handler = FunctionHandler(handle_snaps)
