from landscape.message_schemas import UBUNTU_PRO_INFO

from canonical.lib.arguments import coerce_arguments
from canonical.message.handler import FunctionHandler

from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.main.license import ProLicenseType


@coerce_arguments(message=UBUNTU_PRO_INFO)
def handle_ubuntu_pro_info(message_api, message_type, message):
    """
    The message contains a JSON dump from calling `ua status`. This
    handler adds this information to the computer.
    """
    ubuntu_pro_info = message.get("ubuntu-pro-info")

    if not ubuntu_pro_info:
        return

    computer = message_api.computer
    computer.ubuntu_pro_info = ubuntu_pro_info
    update_pro_licensing(computer)


def update_pro_licensing(computer):
    """
    Updates a computer's licensing based on incoming new_pro_info.

    This can result in a loss of license if the Pro subscription has
    expired, or a switch of license to the UbuntuProFeatures license
    if a new subscription has become active.
    """
    licenses = computer.licenses
    subscription_type = computer.ubuntu_pro_subscription_type()
    pro_licenses = [lic for lic in licenses if lic.is_ubuntu_pro_license()]

    # Computer has a subscription and a Pro license. Do nothing.
    if subscription_type is ProLicenseType.FULL and pro_licenses:
        return

    manager = LicenseManagement(computer.account)

    # Computer has a subscription but no Pro license. Add one.
    if subscription_type in (ProLicenseType.FULL, ProLicenseType.FREE):
        manager.auto_assign_license_to_computer(computer)
        return

    # Computer has no subscription, but has a Pro license. Remove it.
    if pro_licenses:
        for license in pro_licenses:
            license.remove_computer(computer)

        if not computer.has_valid_license():
            manager.auto_assign_license_to_computer(computer)

        return

    # Computer has no subscription and no Pro license. Do nothing.


ubuntu_pro_info_handler = FunctionHandler(handle_ubuntu_pro_info)
