# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging

from zope.interface import implementer

from landscape.message_schemas import USERS

from canonical.lib.arguments import coerce_arguments
from canonical.message.interfaces import IMessageHandler

from canonical.landscape.message.interfaces import ConsistencyError


@implementer(IMessageHandler)
class UserManagementMessageHandler(object):
    def __init__(self, message_api):
        self._computer = message_api.computer
        self.logger = logging.getLogger("canonical.landscape.message")

    @coerce_arguments(message=USERS)
    def __call__(self, type, message):
        """Handle a C{users} message."""
        self._message = message
        self._create_users()
        self._create_groups()
        self._create_group_members()
        self._update_users()
        self._update_groups()
        self._delete_group_members()
        self._delete_groups()
        self._delete_users()

    def _get_users_by_usernames(self, usernames, field_name, ignore_missing=False):
        users = {}
        for user in self._computer.get_users_by_usernames(usernames):
            users[user.username] = user
        if len(usernames) != len(users):
            missing_names = ",".join(name for name in usernames if name not in users)
            message = "Users specified in the '%s' field don't exist: %s" % (
                field_name,
                missing_names,
            )
            if ignore_missing:
                self.logger.info(message)
            else:
                raise ConsistencyError(message, "users")
        return users

    def _get_groups_by_names(self, names, field_name):
        groups = {}
        for group in self._computer.get_groups_by_names(names):
            groups[group.name] = group
        if len(names) != len(groups):
            missing_names = ",".join(name for name in names if name not in groups)
            raise ConsistencyError(
                "Groups specified in the "
                "'%s' field don't exist: %s" % (field_name, missing_names),
                "users",
            )
        return groups

    def _create_users(self):
        """Process the C{create-users} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        create_users = self._message.get("create-users")
        if not create_users:
            return

        names = [data["username"] for data in create_users]
        users = self._computer.get_users_by_usernames(names)
        if users.count():
            names = ", ".join(user.username for user in users)
            raise ConsistencyError(
                "Users for names specified in the "
                "'create-users' field already exist: %s" % names,
                "users",
            )

        create_users.sort(key=lambda data: data["username"])
        for data in create_users:
            arguments = (
                data["uid"],
                data["username"],
                data["primary-gid"],
                data["name"],
                data["enabled"],
                data["location"],
                data["home-phone"],
                data["work-phone"],
            )
            self._computer.create_user(*arguments)

    def _update_users(self):
        """Process the C{update-users} field if it exists.

        @raises ConsistencyError: Raised if a L{User} can't be
            matched to a UID.
        """
        update_users = self._message.get("update-users")
        if not update_users:
            return

        usernames = [data["username"] for data in update_users]
        users = self._get_users_by_usernames(usernames, "update-users")
        update_users.sort(key=lambda data: data["username"])
        for data in update_users:
            user = users.get(data["username"])
            user.name = data["name"]
            user.location = data["location"]
            user.work_phone = data["work-phone"]
            user.home_phone = data["home-phone"]
            user.primary_gid = data["primary-gid"]
            if user.enabled != data["enabled"]:
                user.enabled = data["enabled"]

    def _delete_users(self):
        """Process the C{delete-users} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        delete_users = self._message.get("delete-users")
        if not delete_users:
            return

        map = self._get_users_by_usernames(delete_users, "delete-users")
        users = sorted(map)
        for i in range(0, len(users), 100):
            self._computer.remove_users_by_names(users[i : i + 100])

    def _create_groups(self):
        """Process the C{create-groups} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        create_groups = self._message.get("create-groups")
        if not create_groups:
            return

        groupnames = [data["name"] for data in create_groups]
        groups = self._computer.get_groups_by_names(groupnames)
        if groups.count():
            names = ", ".join(group.name for group in groups)
            raise ConsistencyError(
                "Groups for group names specified in the "
                "'create-groups' field already exist: %s" % names,
                "users",
            )

        create_groups.sort(key=lambda data: data["name"])
        created_group_names = set()
        for data in create_groups:
            new_group_name = data["name"]
            if new_group_name not in created_group_names:
                self._computer.create_group(data["gid"], new_group_name)
                created_group_names.add(data["name"])
            else:
                self.logger.warn(
                    "create-groups dropping duplicate group %s "
                    "for computer %d" % (new_group_name, self._computer.id)
                )

    def _update_groups(self):
        """Process the C{update-groups} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        update_groups = self._message.get("update-groups")
        if not update_groups:
            return

        names = [data["name"] for data in update_groups]
        groups = self._get_groups_by_names(names, "update-groups")
        update_groups.sort(key=lambda data: data["name"])
        for data in update_groups:
            group = groups.get(data["name"])
            group.gid = data["gid"]

    def _delete_groups(self):
        """Process the C{delete-groups} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        delete_groups = self._message.get("delete-groups")
        if not delete_groups:
            return

        map = self._get_groups_by_names(delete_groups, "delete-groups")
        groups = sorted(map)
        for i in range(0, len(groups), 100):
            self._computer.remove_groups_by_names(groups[i : i + 100])

    def _create_group_members(self):
        """Process the C{create-group-members} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        create_group_members = self._message.get("create-group-members")
        if not create_group_members:
            return
        groups = self._get_groups_by_names(create_group_members, "create-group-members")

        for groupname, usernames in sorted(create_group_members.items()):
            group = groups.get(groupname)
            existing_members = set(group.members)
            new_members = self._get_users_by_usernames(
                usernames, "create-group-members", ignore_missing=True
            )
            new_members = sorted(
                iter(new_members.values()), key=lambda user: user.username
            )
            for user in new_members:
                if user in existing_members:
                    message = (
                        "User %s (%d) is already a member of group %s "
                        "(%d)." % (user.username, user.uid, group.name, group.gid)
                    )
                    raise ConsistencyError(message, "users")
                group.associate_member(user)

    def _delete_group_members(self):
        """Process the C{delete-group-members} field if it exists.

        @raises ConsistencyError: Raised when inconsistencies are
            detected.
        """
        delete_group_members = self._message.get("delete-group-members")
        if not delete_group_members:
            return

        groups = self._get_groups_by_names(delete_group_members, "delete-group-members")

        for groupname, usernames in sorted(delete_group_members.items()):
            group = groups.get(groupname)
            existing_members = set(group.members)
            new_members = self._get_users_by_usernames(
                usernames, "delete-group-members"
            )
            new_members = sorted(
                iter(new_members.values()), key=lambda user: user.username
            )
            for user in new_members:
                if user not in existing_members:
                    message = "User %s (%d) is not a member of group %s (%d)." % (
                        user.username,
                        user.uid,
                        group.name,
                        group.gid,
                    )
                    raise ConsistencyError(message, "users")
                group.dissociate_member(user)
