from __future__ import division

# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from past.utils import old_div
from builtins import object
import time

from zope.interface import implementer
from zope.component import getSiteManager

from canonical.landscape.message.interfaces import ILoadShaperFactory


@implementer(ILoadShaperFactory)
class LoadShaperFactory(object):
    """Create a L{LoadShaper}."""

    def __init__(self, critical_load, good_load, good_duration, recalculate_load=False):
        self.critical_load = critical_load
        self.good_load = good_load
        self.good_duration = good_duration
        self.recalculate_load = recalculate_load

    def get_load_shaper(self, get_load_average):
        """Create a L{LoadShaper}.

        @param get_load_average: A callable taking no arguments that returns
            the current load average for the monitored resource.
        """
        return LoadShaper(
            get_load_average,
            self.critical_load,
            self.good_load,
            self.good_duration,
            self.recalculate_load,
        )


class LoadShaper(object):
    """An algorithm to ensure QoS when a resource is used by many clients.

    This is computed by the following function::

         /    l - g \
        ( 1 - -----  ) * t - (n - i) = r
         \    c - g /

    Where::

        l = Current database load
        g = Good load
        c = Critical load
        t = Time slice given to the computer when it is at the good load
        n = Current time
        i = Initial time
        r = Remaining time

    When the load moves below the good load, the time slice given to the
    computer is linearly increased.  When the load goes beyond the good load,
    the time slice gets linearly decreased up to the critical load.  When it
    reaches the critical load, the time slice is zero.
    """

    def __init__(
        self,
        get_load_average,
        critical_load,
        good_load,
        good_duration,
        recalculate_load=False,
    ):
        """
        @param get_load_average: A callable taking no arguments that returns
            the current load average for the monitored resource.
        @param critical_load_average: A float representing the maximum load
            average we want to put on a resource.
        @param good_load_average: A float representing the load at which the
            time slice is exactly C{good_duration}.
        @param good_duration: A float representing the number of seconds to
            allocate a client use of a resource when the load is
            C{good_load_average}.
        @param recalculate_load: if True, get_load_average will be called every
            time get_remaining_time is called, otherwise average load will be
            cached.
        """
        self._get_load_average = get_load_average
        self._critical_load = critical_load
        self._good_load = good_load
        self._good_duration = good_duration
        self._recalculate_load = recalculate_load
        self._load = None
        self._start_time = time.time()

    def get_remaining_time(self):
        """Get the number of seconds remaining to use the resource."""
        if self._load is None or self._recalculate_load:
            self._load = self._get_load_average()
        current_load = self._load - self._good_load
        if current_load < 0 and not self._good_load:
            current_load = 0
        limit = self._critical_load - self._good_load
        duration = time.time() - self._start_time
        return (1 - old_div(current_load, limit)) * self._good_duration - duration

    def get_load(self):
        """Get the last known load average or None if no value is available."""
        return self._load

    def get_start_time(self):
        """Get the start time."""
        return self._start_time


def setup_load_shaper(config):
    """Set the L{ILoadShaperFactory} utility with values from the given config.

    @param config: A C{dict} with 'critical-load', 'good-load' and
        'good-duration' keys.
    """
    # Defaults taken from config/production/service.conf
    load_shaper = LoadShaperFactory(
        float(config.get("critical-load", 10.0)),
        float(config.get("good-load", 3.0)),
        float(config.get("good-duration", 60.0)),
    )
    sm = getSiteManager()
    sm.registerUtility(load_shaper)
