# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Helpers for reliable persistent message queues.
"""


def got_next_expected(store, next_expected):
    """Our peer has told us what it expects our next message's sequence to be.

    Call this with the message store and sequence number that the peer
    wants next; this will do various things based on what *this* side
    has in its outbound queue store.

    1. The peer expects a sequence greater than what we last
       sent. This is the common case and generally it should be
       expecting last_sent_sequence+len(messages_sent)+1.

    2. The peer expects a sequence number our side has already sent,
       and we no longer have that message. In this case, just send
       *all* messages we have, including the previous generation,
       starting at the sequence number the peer expects (meaning that
       messages have probably been lost).

    3. The peer expects a sequence number we already sent, and we
       still have that message cached. In this case, we send starting
       from that message.
    """
    old_sequence = store.get_sequence()

    if next_expected == old_sequence:
        # Either the first contact, or nothing changed since the last contact.
        return
    elif next_expected > old_sequence:
        store.delete_old_messages()
        # If the result below would be greater than the number of available
        # messages, use the number of available messages. This situation
        # means that for some reason the other side belives to have
        # received additional messages, and if we don't update our own
        # control, the other side will lose messages that were not even
        # created yet, and will get stuck until the counter reaches the
        # expected sequence number.
        pending_offset = min(
            next_expected - old_sequence, store.count_available_messages()
        )
    else:
        # The other side wants something that we've already sent
        # before.  We'll try to send as much as possible.  If the
        # result below would be < 0, it means we don't have everything
        # anymore, so we'll just try to send whatever we have).
        old_pending_offset = store.get_pending_offset()
        pending_offset = max(old_pending_offset + next_expected - old_sequence, 0)

    store.set_pending_offset(pending_offset)
    store.set_sequence(next_expected)
