# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Define a twisted.web Resource that will handle all client HTTP requests to
the message system url (https://<landscape-site>/message-system).
"""

import logging

from zope.security.management import newInteraction, endInteraction

from storm.twisted.transact import transact

from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.web.http import NOT_FOUND, INTERNAL_SERVER_ERROR, OK
from twisted.python.failure import Failure
from twisted.python import log
from zope.component import getUtility

from canonical.message.interfaces import MessageAPINotFound
from canonical.message.system import get_api_version_from_request

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.message.backoff import BackoffStatus
from canonical.landscape.message.system import message_system
from canonical.landscape.security.authentication import ComputerAuthenticationPolicy


class MessageSystemResource(Resource):
    """Handle requests to the message system.

    This class takes care of authenticating a client based on the parameters
    found in the HTTP request headers, and then invoking the appropriate
    message handling code.

    @param transactor: The L{Transactor} to use to run database queries in a
        thread-safe way (see C{storm.twisted.transact})
    @param message_system: The L{LandscapeMessageSystem} instance to use to
        process messages, defaults to C{.system.message_system} but can be
        overridden by tests for mocking.
    """

    def __init__(self, transactor, message_system=message_system):
        Resource.__init__(self)
        self.transactor = transactor
        self.transactor.on_retry = self._on_retry
        self._message_system = message_system
        self._authentication_policy = ComputerAuthenticationPolicy()

        config = getUtility(IFileBasedConfiguration)
        dirpath = config.get("backoff", key="backoff-dirpath", default=None)
        self.backoff_status = BackoffStatus(watch_dirpath=dirpath)

        # Keep track of disconnected requests
        self._failed = []

    def getChild(self, path, request):
        return self

    def render_HEAD(self, request):
        """Return an empty response for HEAD requests.

        This is to support health check pings. Note that we want to have these
        run without using a thread, to not block on the threadpool queue.
        """
        request.setHeader("Content-Type", "text/plain")
        request.setResponseCode(OK)
        return b""

    def render_GET(self, request):
        """Return a brief description of the service."""
        request.setHeader("Content-Type", "text/plain")
        request.setResponseCode(OK)
        return b"Landscape message server"

    def render_POST(self, request):
        """Handle a request.

        @param request: A C{twisted.web.http.Request} instance.
        """

        def aborted(error):
            # This means the request was interrupted by an error (e.g. the
            # closing the connection prematurely). We're still processing
            # the request in another thread see (_handle_request), let's
            # take note of the failure so we won't try to write the result
            # once we finish processing (see _finish)
            self._failed.append(request)

        request.notifyFinish().addErrback(aborted)

        # If some message-handling code fails as a result of a database
        # error, it is possible that this error may be transient. The
        # transactor will attempt to repeat the execution of _handle_request
        # until it either succeeds or we reach the predetermined limit. This
        # behaviour mimicks Zope (see zope.pubisher.publish) and should be
        # completely transparent, however we need to monitor it in order to
        # decide if an oops should be logged. We log only if the final retry
        # fails (see _process_messages in apis.py). In order to achieve this we
        # annotate the request object with the ordinal of the retry, which is
        # used to emulate Request.supportsRetry from Zope and is compatible
        # with c.l.publication.publication.request_should_be_retried.
        #
        # XXX maybe we could find a way to factor code in a way that we don't
        # need to break into this otherwise transparent retry mechanism.
        request.retry = 0

        def supportsRetry():
            return request.retry < self.transactor.retries

        request.supportsRetry = supportsRetry
        deferred = self._handle_request(request)
        deferred.addBoth(self._finish, request)
        return NOT_DONE_YET

    @transact
    def _handle_request(self, request):
        """Dispatch the request to the appropriate L{MessageAPI}."""
        if getattr(request, "principal", None) is None:
            # The request hasn't been authenticated yet, so let's do it. This
            # will be done only once, further retries won't need it
            principal = self._authentication_policy.authenticate(request)
            request.principal = principal
        else:
            request.content.seek(0, 0)
        # Start a new interaction
        request.interaction = None
        newInteraction(request)
        if self.backoff_status.should_backoff():
            logging.warning("Telling client to backoff!")
            request.setResponseCode(429)
            endInteraction()
            return b""
        try:
            result = self._message_system.dispatch(request)
        except MessageAPINotFound:
            api = get_api_version_from_request(request)
            logging.error("Computer %s requested unknown API %s" % (principal.id, api))
            request.setResponseCode(NOT_FOUND)
            result = b""
        finally:
            endInteraction()

        return result

    def _finish(self, result, request):
        """Finish the request.

        @note: This code must be run outside @transact as it's not thread-safe.
        """
        if request in self._failed:
            # The request has terminated while we were processing it, just
            # abort here
            self._failed.remove(request)
            return
        if isinstance(result, Failure):
            log.err(result, "Got error while handling message")
            request.setResponseCode(INTERNAL_SERVER_ERROR)
        else:
            request.setHeader("Content-Type", "application/octet-stream")
            request.write(result)
        request.finish()

    def _on_retry(self, context):
        """Update the Request's retry attribute, so supportsRetry works."""
        request = context.args[1]
        request.retry = context.retry
