# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
A query introspection and debugging middleware for Landscape.

Features
--------

 - Captures total number of queries, total query execution time, and
   total request time.

 - Classification of queries as duplicates or similiar, ie. same statement
   different parameters.

 - Optional capturing of call stacks for each query executed.

Usage
-----

To enable it or other WSGI Middleware set the LANDSCAPE_MIDDLEWARE environment
variable. For example

 export LANDSCAPE_MIDDLEWARE="canonical.landscape.middleware.QueryToolbar"

Stack Traces
------------
Even with output of all the queries executed in the request, it can still be
difficult to determine where in a large application code base queries are
coming from. To assist with this the QueryToolbar middleware has an additional
optional feature to capture application call stacks for each query executed.
This feature can be enabled by setting the environment variable
LANDSCAPE_QUERY_DEBUG Its not enabled by default as the stack extraction is
fairly expensive when aggregated over all the queries in a request, and the
stack extraction time is counted as part of a queries execution time.

Todo
----

 - maybe offer file as urls that open up in editor (requires developer desktop
   customization). ala
   'integrating a new uris scheme handler to gnome and firefox'
   http://people.w3.org/~dom/archives/2005/09/

 - maybe offer an ajax way to get a query execution plan (EXPLAIN OUTPUT)
"""

import os
import sys
import linecache

from html import escape
from time import time
from traceback import print_list
from io import StringIO
from storm.variables import Variable
from zope.component import getUtility

from canonical.oops.interfaces import IOOPSManager
from canonical.oops.stormtracer import OOPSStorm


class MiddlewareResponse(object):
    def __init__(self):
        self.status = None
        self.headers = None

    def replace(self, name, value):
        for index in range(len(self.headers)):
            if not self.headers[index][0] == name:
                continue
            self.headers[index] = (name, value)


class StackInspector(object):
    """
    Stack Frame printing, with application specific behavior to only output
    interesting stacks frames, and to capture additional information in output.
    """

    stack_offset_start = 10

    stack_begin_ignore = set(
        (
            "storm",
            "canonical/oops",
        )
    )

    stack_end_stop = "zope/app/publication"

    stack_ignore_swallow = set(
        (
            ("tal", "talinterpreter.py"),
            ("browserpage", "viewpagetemplatefile.py"),
            (("tales", "expressions.py"), "_eval"),
            ("tales", "tales.py"),
            ("content", "provider.py"),
        )
    )

    stack_special = {
        (("tales", "expressions.py"), "__call__"): "stack_expression",
        (("tales", "pythonexpr.py"), "__call__"): "stack_python_expression",
        (("pagetemplate", "pagetemplate.py"), "pt_render"): "stack_template",
        (("contentprovider", "tales.py"), "__call__"): "stack_render_provider",
        # easier to read on oneline.. the curse of 80 chars.
        (("security", "policy.py"), "checkPermission"): "stack_check_permission",
    }

    def __init__(self):
        self.stack_special = dict(self.stack_special)
        for key in self.stack_special:
            method = getattr(self, self.stack_special[key])
            self.stack_special[key] = method

    def stack_python_expression(self, frame):
        return "Evaluate %s\n    Line %s File %s" % (
            frame.f_locals["self"],
            getattr(frame.f_locals["econtext"], "position", ("unknown",))[0],
            getattr(frame.f_locals["econtext"], "source_file", "unknown"),
        )

    def stack_expression(self, frame):
        return "Evaluate %s\n    Line %s File %s" % (
            frame.f_locals["self"],
            getattr(frame.f_locals["econtext"], "position", ("unknown",))[0],
            getattr(frame.f_locals["econtext"], "source_file", "unknown"),
        )

    def stack_template(self, frame):
        return "Render Template %s" % (frame.f_locals["self"].filename)

    def stack_check_permission(self, frame):
        return "Check Permission %s on %r" % (
            frame.f_locals["permission"],
            escape(repr(frame.f_locals["object"])),
        )

    def stack_render_provider(self, frame):
        return "Render Provider name:%s view:%s context:%s" % (
            frame.f_locals["name"],
            frame.f_locals["view"].__name__,
            escape(repr(frame.f_locals["context"])),
        )

    def extract_stack(self):
        # speed counts, time spent here counts towards sql execution
        # time. even on large stacks this runs in a few milliseconds,
        # but its aggregated over the number of queries in a request.

        # get out of the middleware and oops manager back into storm
        count = self.stack_offset_start  # based on observation
        frame = sys._getframe(self.stack_offset_start)
        start = limit = 0
        output = []

        # find the start and end of the interesting part of the stack
        while not (start and limit) and frame:
            filename = frame.f_code.co_filename
            if not start:
                skip = False
                for path in self.stack_begin_ignore:
                    if path in filename:
                        count += 1
                        frame = frame.f_back
                        skip = True
                        break
                if skip:
                    continue

            if not start:
                start = count

            if self.stack_end_stop in filename:
                limit = count

            func_name = frame.f_code.co_name
            f_key = tuple(filename.rsplit("/", 2)[1:])

            f_special = self.stack_special.get((f_key, func_name))
            ignore = f_key in self.stack_ignore_swallow
            ignore = ignore or (f_key, func_name) in self.stack_ignore_swallow
            if not f_special and ignore:
                count += 1
                frame = frame.f_back
                continue

            lineno = frame.f_lineno
            if f_special:
                line = f_special(frame)
            else:
                line = linecache.getline(filename, lineno, frame.f_globals)
            output.append((filename, lineno, func_name, line))
            count += 1
            frame = frame.f_back

        output.reverse()
        output_buffer = StringIO()
        print_list(output, output_buffer)
        return output_buffer.getvalue()


class QueryToolbar(object):

    # we look for the specified tags sequence at the end of the response
    # to insert the query introspection debug middleware into.
    html_insertion_tags = ["</body>"]

    def __init__(self, application):
        self.application = application
        self.oops_manager = getUtility(IOOPSManager)
        self.capture_callstacks = False
        self.inspector = StackInspector()
        self.setup()

    def setup(self):
        """Setup the application server with instrumentation hooks."""
        self.oops_manager.automatic_removal = False

        # support capturing callstacks for queries
        if os.environ.get("LANDSCAPE_QUERY_DEBUG"):
            self.capture_callstacks = True
            middleware = self

            def connection_raw_execute(
                self, now, connection, raw_cursor, statement, params
            ):
                """Record an executed statement in this OOPS entry."""
                params = tuple(
                    repr(
                        (
                            isinstance(param, Variable)
                            and (param.get(to_db=True),)
                            or (param,)
                        )[0]
                    )
                    for param in params
                )
                stack = middleware.extract_stack()
                self._statements.append([statement, params, now, None, stack])

            OOPSStorm.connection_raw_execute = connection_raw_execute

    def __call__(self, environ, start_response):
        """
        WSGI entry point.
        """
        # Capture the status and headers so we can patch up Content Length
        response = MiddlewareResponse()

        def wrapped_start_response(status, headers, exc_info=None):
            response.status = status
            response.headers = headers

        start_time = time()
        result = self.application(environ, wrapped_start_response)
        request_time = time() - start_time

        # If the result is an iterator just return it (file)
        if not isinstance(result, (tuple, list)) or not len(result) == 1:
            start_response(response.status, response.headers)
            return result

        # If there's no content its probably a redirect
        if not result[0]:
            start_response(response.status, response.headers)
            return result

        # Render the log into the page
        wrapped_result = self.append(
            result[0].decode(), self.render(request_time), SCRIPT
        )

        # Ensure the Content Length header is set correctly, else
        # the browser will truncate the response.
        response.replace("Content-Length", "{}".format(len(wrapped_result)))

        start_response(response.status, response.headers)
        return (wrapped_result.encode(),)

    def append(self, page, content, script=None):
        """
        Attach content to the end of the html page.
        """
        index = len(page)
        for tag in self.html_insertion_tags:
            index = page.lower().rfind(tag, 0, index)
            if index > -1:
                page = page[:index] + content + page[index:]

        if script:
            # Now insert javascript, if any, just before the closing
            # body tag.
            index = page.lower().rfind("</body>")
            if index > -1:
                return page[:index] + script + page[index:]

        return page

    def render(self, request_time):
        """
        Format the storm statements executed during the request for display.
        """
        total_queries, query_time, statements = self.classify(self.extract())
        similiar_queries = duplicate_queries = 0

        html_buffer = list(WIDGET_TEMPLATE)

        for info in statements:
            if info.setdefault("similiar", ""):
                info["similiar"] = info["count"]
            if info.setdefault("duplicate", ""):
                info["duplicate"] = info["count"] - len(info["params"])
                similiar = len(info["params"]) - info["duplicate"]
                info["similiar"] = similiar > 0 and similiar or ""
            duplicate_queries += info["duplicate"] and info["duplicate"] or 0
            similiar_queries += info["similiar"] and info["similiar"] or 0
            info.setdefault("stack_link", "")
            if self.capture_callstacks:
                info["stack_link"] = "&nbsp;<a class='sql-stack' href=''>C</a>"
            html_buffer.append(QUERY_TEMPLATE % info)
            for p in info["params"]:
                html_buffer.append(PARAM_TEMPLATE % str(p))
            if self.capture_callstacks:
                for stack in info["stack"]:
                    html_buffer.append(STACK_TEMPLATE % stack)

        html_buffer.append("</tbody></table></div>")
        html_buffer.append("</div>")

        short_summary = "R:%0.2f SQL#:%d D:%d S:%d T:%0.3f" % (
            request_time,
            total_queries,
            duplicate_queries,
            similiar_queries,
            query_time,
        )
        long_summary = LONG_SUMMARY_TEMPLATE % (
            request_time,
            total_queries,
            duplicate_queries,
            similiar_queries,
            query_time,
        )
        html_buffer.insert(0, SUMMARY_TEMPLATE % (long_summary, short_summary))
        return "\n".join(html_buffer)

    def classify(self, statements):
        """
        Classify sql statements to determine duplicates and similiar
        statements.
        """
        statement_map = {}
        statement_order = []
        total_time = 0
        for statement, params, time_start, time_end, error in statements:
            query_time = time_end - time_start
            if statement not in statement_map:
                statement_order.append(statement)
                statement_map[statement] = d = dict(
                    statement=statement,
                    params=[params],
                    count=1,
                    time=query_time,
                )
                if self.capture_callstacks:
                    d["stack"] = [error]
                total_time += query_time
                continue
            info = statement_map[statement]
            info["time"] += query_time
            total_time += query_time
            info["count"] += 1
            info["similiar"] = True
            if self.capture_callstacks:
                info["stack"].append(error)
            if params not in info["params"]:
                info["params"].append(params)
                continue
            info["duplicate"] = True
        return (
            len(statements),
            total_time,
            [statement_map[statement] for statement in statement_order],
        )

    def extract(self):
        """
        Extract statements from the oops manager.
        """
        oops = self.oops_manager.get("request")
        self.oops_manager.remove("request")
        storm_oops = oops.get("storm")
        statements = storm_oops.get_dump_data()["statements"]
        return statements

    def extract_stack(self):
        return self.inspector.extract_stack()


WIDGET_TEMPLATE = [
    "<div id='wsgi-sql-overlay' style='display: none'>",
    "<div class='sql-widget-bd'><table class='inline-listing'>",
    "<thead><tr>",
    "<th>Time</th>",
    "<th>D</th>",
    "<th>S</th>",
    "<th>P</th>",
    "<th>Statement</th>",
    "</tr></thead>",
    "<tbody>",
]

QUERY_TEMPLATE = """\
<tr><td>%(time)0.6f</td><td>%(duplicate)s</td><td>%(similiar)s</td>
    <td><a class='sql-params' href='#'>P</a>%(stack_link)s</td>
    <td>%(statement)s</td></tr>"""

PARAM_TEMPLATE = """\
<tr class='sql-params sql-hidden'><td colspan='5'>
    <pre>%s</pre></td></tr>"""

STACK_TEMPLATE = """\
<tr class='sql-stack sql-hidden'><td colspan='5'>
    <pre>%s</pre></td></tr>
"""

LONG_SUMMARY_TEMPLATE = (
    "Request Time:%0.2f  Total Queries:%d "
    "Duplicates: %d  Similiar: %d  Query Time:%0.3f"
)

SUMMARY_TEMPLATE = """
<div id='wsgi-debug-toolbar' style='display: none;'>
  <a id='wsgi-sql-summary' title='%s' href='#'>%s</a>
</div>
<style type='text/css'>

#wsgi-sql-overlay {
  color: black;
  background: white;
}

.sql-hidden {
  display: none;
  }

.sql-widget-bd {
  padding: 16px;
 }

.sql-widget-bd tr > td {
  border-right: 1px dotted #ccc;
  padding-right: 8px;
}

</style>
"""

SCRIPT = """
<script type='text/javascript' defer='defer'>
LS.use("landscape.ui.debug");
</script>
"""
