# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.expr import Func

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.access.errors import (
    DuplicateRoleError,
    UnknownRoleError,
    InvalidRoleNameError,
)
from canonical.landscape.model.access.role import (
    StoredRole,
    StoredRolePerson,
    GlobalAdmin,
    is_valid_role_name,
)
from canonical.landscape.model.main.personaccess import PersonAccess
from canonical.landscape.model.access.errors import ReadOnlyRoleError
from canonical.lib.text import describe_list


__all__ = ["AccessControlManagement"]

__clearlist__ = ["AccessControlManagement"]

__requires_prefix__ = "landscape."


@requires_advisor
class AccessControlManagement(ModelManagement):

    requires("ViewAccount", PRIVATE, ["account"])

    _global_admin = None

    @requires("EditRole")
    def create_role(self, name, description=None, permissions=(), access_contexts=()):
        """Create a new custom role in the account.

        @raises DuplicateRoleError: If a role with the given name exists.
        @raises InvalidRoleNameError: If the name is not valid for a role.
        """
        if not is_valid_role_name(name):
            raise InvalidRoleNameError("Role name '%s' is invalid." % name)
        if self.get_roles(names=[name]):
            raise DuplicateRoleError("Role '%s' already exists." % name)

        role = StoredRole(self.account, name, description=description)
        role.add_permissions(permissions)
        role.add_access_contexts(access_contexts)
        self._event_log.create_log_entry(role, "Role created.")
        return role

    @requires("ViewRole")
    def get_roles(self, names=None):
        """
        Get all L{StoredRole}s in the account matching the given criteria.
        """
        roles = []
        global_admin = self.get_global_admin()
        if names is None or global_admin.name in names:
            roles.append(global_admin)
        roles.extend(self._get_roles(names=names))
        return roles

    @requires("ViewRole")
    def get_memberships(self, roles):
        """Return membership information for the given roles.

        The returned structure is a list with the following format:

           [{'person': <person>, 'info': [{'role': <role>, 'flag': <flag>}]}]

        The list has an entry for each person in the account, and each entry
        contains information indicating for each role whether the person has
        has that role in the account. The flags list ordering matches the one
        of the given C{roles} list.
        """
        store = get_main_store()
        roles_by_id = dict(
            (role.id, role) for role in roles if role.name != "GlobalAdmin"
        )
        spec = (Person.id, Func("ARRAY_AGG", StoredRolePerson.role_id))
        result = store.find(
            spec,
            Person.id == StoredRolePerson.person_id,
            StoredRolePerson.role_id.is_in(roles_by_id),
        )
        result.group_by(Person.id)
        # Maps person IDs to the role IDs the person belongs to
        roles_map = dict(result)
        memberships = []
        for person in self.account.people:
            info = []
            for role in roles:
                # XXX This is kind of bad, we have to special case the global
                # role because we don't store associations for it in the
                # role_person table. Eventually we might want to treat it
                # uniformely.
                if role.name == "GlobalAdmin":
                    flag = person.has_role("Owner", self.account.access_context)
                else:
                    flag = role.id in roles_map.get(person.id, [])
                info.append({"role": role, "flag": flag})
            memberships.append({"person": person, "info": info})
        return memberships

    def get_global_admin(self):
        """Return the L{GlobalAdmin} role for the account."""
        if self._global_admin is None:
            self._global_admin = GlobalAdmin(self.account)
        return self._global_admin

    def _get_roles(self, names=None):
        """
        Get all L{StoredRole}s in the account matching the given criteria.
        """
        where = [StoredRole.account_id == self.account.id]
        if names is not None:
            where.append(StoredRole.name.is_in(names))
        store = get_main_store()
        return store.find(StoredRole, *where).order_by(StoredRole.name)

    def get_role(self, name):
        """Get the role with the given name.

        @raises UnknownRoleError: If a role with the given name doesn't exists.
        """
        roles = self.get_roles(names=[name])
        if not roles:
            raise UnknownRoleError("Unknown role name: %s" % name)
        return roles[0]

    def remove_role(self, name):
        """Remove a role from the account.

        @raises UnknownRoleError: If a role with the given name doesn't exists.
        """
        global_admin = self.get_global_admin()
        if name == global_admin.name:
            raise InvalidRoleNameError("Can't remove %s role" % name)

        role = self.get_role(name)
        persons_ids = [person.id for person in role.get_persons()]
        store = get_main_store()
        store.find(
            PersonAccess,
            PersonAccess.person_id.is_in(persons_ids),
            PersonAccess.role_key == role.key,
        ).remove()
        store.remove(role)
        self._event_log.create_log_entry(role, "Role removed.")

    def get_persons_roles(self):
        """
        Return a dict mapping each L{Person} in the L{Account} to his roles.
        """
        persons = {}
        persons_roles = {}
        for person in self.account.people:
            persons[person.id] = person
            persons_roles[person] = []

        result = get_main_store().find(
            (StoredRole, StoredRolePerson.person_id),
            StoredRole.account_id == self.account.id,
            StoredRolePerson.role_id == StoredRole.id,
            StoredRolePerson.person_id.is_in(persons),
        )

        for role, person_id in result:
            persons_roles[persons[person_id]].append(role)

        global_admin_role = self.get_global_admin()
        global_admin_ids = global_admin_role.get_persons().values(Person.id)
        for person_id in global_admin_ids:
            persons_roles[persons[person_id]].append(global_admin_role)

        return persons_roles

    def remove_person_roles(self, person):
        """
        Remove all roles the L{Person} has in the L{Account}.
        """
        stored_role_ids = self._get_roles().values(StoredRole.id)

        get_main_store().find(
            StoredRolePerson,
            StoredRolePerson.role_id.is_in(stored_role_ids),
            StoredRolePerson.person_id == person.id,
        ).remove()

    def copy_role(self, name, new_name, description):
        """Copy a role."""
        role = self.get_role(name)
        access_context_names = [c.name for c in role.get_access_contexts()]
        permissions = role.get_permissions() + role.get_account_permissions()
        emails = [person.email for person in role.get_persons()]
        new_role = self.create_role(
            new_name,
            description,
            permissions=permissions,
            access_contexts=access_context_names,
        )
        new_role.add_persons(emails)
        return new_role

    @requires("EditRole")
    def edit_role(self, name, permissions=None, access_contexts=None):
        """
        Change a role's permissions and/or access contexts.

        Changes are logged to the event log.

        @param name: The current name of the role.
        @param permissions: The new permissions of the role.
        @param access_contexts: The new access groups for the role.
        """
        account_access_contexts = self.account.get_access_contexts()
        access_context_mapping = dict(
            (access_context.name, access_context.description)
            for access_context in account_access_contexts
        )
        role = self.get_role(name)
        if role.name == "GlobalAdmin":
            raise ReadOnlyRoleError(role.name)
        account_permissions = role.get_account_permissions()
        original_permissions = role.get_permissions() + account_permissions
        original_access_contexts = [i.name for i in role.get_access_contexts()]

        # Make the changes down in the specific class
        role.edit(permissions=permissions, access_context_names=access_contexts)

        # Calculate the changes for the activity creation.
        added_permissions = []
        removed_permissions = []
        if permissions is not None:
            added_permissions = sorted(set(permissions) - set(original_permissions))
            removed_permissions = sorted(set(original_permissions) - set(permissions))

        added_access_contexts = []
        removed_access_contexts = []
        if access_contexts is not None:
            added_access_contexts = set(access_contexts) - set(original_access_contexts)
            removed_access_contexts = set(original_access_contexts) - set(
                access_contexts
            )
            added_access_contexts = sorted(
                [
                    "%s (%s)" % (i, access_context_mapping[i])
                    for i in added_access_contexts
                ]
            )
            removed_access_contexts = sorted(
                [
                    "%s (%s)" % (i, access_context_mapping[i])
                    for i in removed_access_contexts
                ]
            )

        if (
            added_permissions
            or removed_permissions
            or added_access_contexts
            or removed_access_contexts
        ):
            self._event_log.create_log_entry(
                role,
                "".join(
                    [
                        describe_list(" Added permission:", added_permissions),
                        describe_list(" Removed permission:", removed_permissions),
                        describe_list(" Added access context:", added_access_contexts),
                        describe_list(
                            " Removed access context:", removed_access_contexts
                        ),
                    ]
                ).strip()
                + ".",
            )

    @requires("EditRole")
    def edit_membership(self, role_name, added=(), removed=()):
        """
        Edit role memberhip.

        @param role_name: The name of the role.
        @param added: A list with the new people being added to the role.
        @param removed: A list with the new people being removed from the role.
        """
        role = self.get_role(role_name)
        original_membership = list(role.get_persons())
        role.add_persons(added)
        role.remove_persons(removed)
        new_membership = list(role.get_persons())

        added_persons = set(new_membership) - set(original_membership)
        removed_persons = set(original_membership) - set(new_membership)

        added_persons = sorted(
            "%s (%s)" % (person.name, person.email) for person in added_persons
        )
        removed_persons = sorted(
            "%s (%s)" % (person.name, person.email) for person in removed_persons
        )

        if added_persons or removed_persons:
            self._event_log.create_log_entry(
                role,
                "".join(
                    [
                        describe_list(" Added person:", added_persons),
                        describe_list(" Removed person:", removed_persons),
                    ]
                ).strip()
                + ".",
            )
