# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from collections import namedtuple, OrderedDict
from canonical.landscape.model.configuration import is_testing_environment


__clearlist__ = ["MACRO_PERMISSIONS", "ACCOUNT_MACRO_PERMISSIONS"]

# NOTICE: If removing or modifying ANY macro permission from this file, we must
# have a DB patch to main to appropriately remove or modify the permission from
# the main.role_permission table

# Default micro-permissions that any stored role gets.
DEFAULT_PERMISSIONS = [
    "ViewActivity",
    "ViewAccessContextManagement",  # This is to let people see the access
    # contexts they have some permission on.
]

# Permissions that are only visible in testing environments (staging, default)
# XXX this list will be empty if there are no "hidden" development permissions
TESTING_PERMISSIONS = set()

MacroPermission = namedtuple(
    "MacroPermission", ["name", "title", "doc", "permissions", "account_permissions"]
)


def get_macro_permissions(access_context=True, account=True, system=True):
    """
    Return all registered macro permissions filtering based on the types
    specified.

    @param access_context: If False, access context permissions (computer
        level) are not included.
    @param account: If False, account-level permissions are not included.
    @param system: If False, system-level permissions are not included.
    """
    return (
        (list(MACRO_PERMISSIONS.keys()) if access_context else [])
        + (list(ACCOUNT_MACRO_PERMISSIONS.keys()) if account else [])
        + (list(SYSTEM_MACRO_PERMISSIONS.keys()) if system else [])
    )


class MacroPermissionRegistry(OrderedDict):
    """A registry for Macro Permissions.

    "Macro Permissions" are user-visible permissions that administrators can
    grant to custom roles.
    """

    def add_macro_permission(
        self, name, title, doc, permissions, extends=None, account_permissions=None
    ):
        """
        Register a macropermission in the map.

        @param name: The name of the permission, a single word.
        @param title: A verbose name of the description.
        @param doc: A help docstring describing the authorised actions of the
            macro-permission. These docstrings are displayed in the
            Roles UI help overlay.
        @param permissions: List of permissions associated with the
            macro-permission.
        @param account_permissions: List of permissions that will bypass the
            access group associations with the role that this macro permission
            is granted to. These permissions are granted to the user directly
            on the account.
        @param extends: optionally, a name of a macro-permission.
            The new permission will include all permissions from this one.
        """
        permissions = set(permissions)
        if not is_testing_environment():
            if name in TESTING_PERMISSIONS:
                # Don't publish this permission on production or LDS
                return None
            # Don't associate any development micro-permissions with this macro
            permissions -= TESTING_PERMISSIONS
            if not permissions:
                # Don't add empty macro-permissions
                return

        if account_permissions is None:
            account_permissions = []
        account_permissions = set(account_permissions)
        if extends is not None:
            permissions.update(self[extends].permissions)
            account_permissions.update(self[extends].account_permissions)
        macro_permission = MacroPermission(
            name, title, doc, sorted(permissions), account_permissions
        )
        self[name] = macro_permission
        return macro_permission


# Global macro-permissions map
MACRO_PERMISSIONS = MacroPermissionRegistry()
ACCOUNT_MACRO_PERMISSIONS = MacroPermissionRegistry()
SYSTEM_MACRO_PERMISSIONS = MacroPermissionRegistry()

# Manage System
SYSTEM_MACRO_PERMISSIONS.add_macro_permission(
    "ManageSystem",
    "Manage System",
    "Manage entire Landscape system",
    [
        "ActivateTrial",
        "CreateAccount",
        "RemoveAccount",
        "CreateLicense",
        "EditKnowledgeBase",
        "EditLicense",
        "EditMaxAttachmentSize",
        "EditMaxPeopleCount",
        "EditSalesforceKey",
        "InviteInitialAdministrator",
        "ManageSupportCases",
        "ReviewKnowledgeBase",
        "ViewAllAccounts",
        "ViewSalesforceKey",
        "ViewSalesforceUser",
        "ViewWedgedComputers",
        "ViewSearch",
        "EditVatRate",
        "AccessUtilityCost",
        "ManageBillingConfig",
        "ViewBilling",
    ],
)


# Manage Account
ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ManageAccount",
    "Manage account",
    "Manage account details",
    [
        "BlockActivity",
        "CancelInvitation",
        "CheckRule",
        "CompleteActivity",
        "CreateAccessContext",
        "CreateComputer",
        "CreateGroup",
        "CreateJujuModel",
        "EditJujuModel",
        "CreateUser",
        "DeleteOldMessages",
        "DeliverMessages",
        "DisableAccountMembership",
        "DisableMembership",
        "DismissActivity",
        "EditAccessContext",
        "EditAccount",
        "EditAlert",
        "EditComputerLicenses",
        "EditGroup",
        "EditNextExpectedSequence",
        "EditPerson",
        "EditPersonRole",
        "EditRole",
        "FetchMessages",
        "GetAlertSubscription",
        "GetResourceStores",
        "InviteAdministrator",
        "ManageAccessGroups",
        "ManageOpenStack",
        "QueryActivity",
        "QueryAlertRule",
        "QueryPackageProfile",
        "RegisterComputer",
        "RejectPendingComputer",
        "RemoveGroup",
        "ResynchronizeComputer",
        "SendMessage",
        "StartActivity",
        "SubscribeAlert",
        "SynchronizeTags",
        "UnblockActivity",
        "ViewAccessContext",
        "ViewAccount",
        "ViewAccountData",
        "ViewAlertRule",
        "ViewOpenStack",
        "ViewComputerSelection",
        "ViewInvitations",
        "ViewJujuModel",
        "ViewKnowledgeBase",
        "ViewLicense",
        "ViewPendingComputer",
        "ViewPermissionInfo",
        "ViewGPGKey",
        "ViewPerson",
        "ViewPersonPassword",
        "ViewPersonResource",
        "ViewRole",
        "ViewEventLog",
        "ViewVotesKnowledgeBase",
        "VoteKnowledgeBase",
    ],
)

# Computer
MACRO_PERMISSIONS.add_macro_permission(
    "ViewComputer",
    "View computers",
    "View computer details, tags, packages and custom graphs",
    [
        "ViewComputer",
        "ViewComputerPackages",
        "ViewCustomGraph",
        "ViewMountInfo",
        "ViewThermalZone",
        "ViewMemoryInfo",
        "ViewLoadAverage",
        "ViewCPUUsage",
        "ViewKeystoneToken",
        "ViewActiveProcess",
        "ViewHistoricProcess",
        "ViewUserManagement",
        "ViewUser",
        "ViewGroup",
        "ViewDevice",
        "ViewHardwareInfo",
        "ViewPingTimes",
        "ViewProcessor",
        "ViewSavedSearch",
    ],
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManageComputer",
    "Manage computers",
    "Manage computers, activities, tags, packages, custom graphs and run " "scripts",
    [
        "ApproveActivity",
        "CancelActivity",
        "CreateActivity",
        "ScheduleActivity",
        "ExecuteScript",
        "ExecuteStoredScript",
        "EditComputerPackages",
        "EditTags",
        "EditCustomGraph",
        "CreateCustomGraph",
        "RemoveComputer",
        "EditGroup",
        "EditUser",
        "ManageActiveProcess",
        "RemoveUser",
        "RequestCreateUser",
        "RevertActivity",
        "RestartComputer",
        "ShutdownComputer",
        "EditComputerDetails",
        "CreateActiveProcess",
        "CreateHistoricProcess",
        "CreateLoadAverage",
        "CreateCPUUsage",
        "CreateMemoryInfo",
        "CreateMountInfo",
        "CreateProcessor",
        "CreateSavedSearch",
        "CreateThermalZone",
        "EditActiveProcess",
        "EditDevice",
        "EditHardwareInfo",
        "EditHistoricProcess",
        "EditLoadAverage",
        "EditCPUUsage",
        "EditKeystoneToken",
        "EditMemoryInfo",
        "EditMountInfo",
        "EditProcessor",
        "EditSavedSearch",
        "EditThermalZone",
        "RemoveActiveProcess",
        "RemoveProcessor",
    ],
    extends="ViewComputer",
)

MACRO_PERMISSIONS.add_macro_permission(
    "AddComputerToAccessGroup",
    "Add computers to an access group",
    "Add computers to an access group",
    ["AddComputerToAccessContext"],
    extends="ViewComputer",
)

MACRO_PERMISSIONS.add_macro_permission(
    "RemoveComputerFromAccessGroup",
    "Remove computers from an access group",
    "Remove computers from an access group",
    ["RemoveComputerFromAccessContext"],
    extends="ViewComputer",
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManagePendingComputer",
    "Manage pending computers",
    "View, accept and reject pending computer registrations",
    ["AcceptPendingComputer"],
    account_permissions=["RejectPendingComputer", "ViewPendingComputer"],
)

# Script
MACRO_PERMISSIONS.add_macro_permission(
    "ViewScript", "View scripts", "View stored scripts", ["ViewScript"]
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManageScript",
    "Manage scripts",
    "Manage stored scripts",
    ["CreateScript", "EditScript"],
    extends="ViewScript",
)

# Manage MAAS
MACRO_PERMISSIONS.add_macro_permission(
    "ManageMAAS",
    "Manage MAAS Servers",
    "Manage MAAS Servers",
    [
        "UseMAAS",
        "CreateMAASServer",
        "EditMAASServer",
        "ViewMAASServer",
        "ViewMAASSystem",
    ],
)

# Repository Management:
# Two separate MACRO sets. 1: Profiles/APTSources 2: distributions/pockets
MACRO_PERMISSIONS.add_macro_permission(
    "ViewRepositoryProfile",
    "View repository profiles",
    "View repository profiles and APT sources",
    ["ViewAPTSource", "ViewRepositoryProfile"],
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManageRepositoryProfile",
    "Manage repository profiles",
    "Manage repository profiles and APT Sources",
    [
        "CreateAPTSource",
        "CreateRepositoryProfile",
        "EditAPTSource",
        "EditRepositoryProfile",
    ],
    extends="ViewRepositoryProfile",
)

# 2: View/ManageRepository for pocket,series distribution creation and sync.
MACRO_PERMISSIONS.add_macro_permission(
    "ViewRepository",
    "View package repositories",
    "View package repositories as distributions, series and pockets",
    ["ViewDistribution", "ViewPocket"],
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManageRepository",
    "Manage package repositories",
    "Manage package repositories as distributions, series and pockets",
    [
        "CreateDistribution",
        "CreatePocket",
        "CreateSeries",
        "EditDistribution",
        "EditPocket",
        "EditSeries",
        "SyncPocket",
    ],
    extends="ViewRepository",
)

# Upgrade Profiles
MACRO_PERMISSIONS.add_macro_permission(
    "ViewUpgradeProfile",
    "View upgrade profiles",
    "View upgrade profiles",
    ["ViewUpgradeProfile"],
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManageUpgradeProfile",
    "Manage upgrade profiles",
    "Manage upgrade profiles",
    ["EditUpgradeProfile"],
    extends="ViewUpgradeProfile",
)

# Package Profiles
MACRO_PERMISSIONS.add_macro_permission(
    "ViewPackageProfile",
    "View package profiles",
    "View package profiles",
    ["ViewPackageProfile"],
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManagePackageProfile",
    "Manage package profiles",
    "Manage package profiles",
    ["CreatePackageProfile", "EditPackageProfile"],
    extends="ViewPackageProfile",
)

# Roles
ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ViewRole", "View Roles", "View Roles", ["ViewRole"]
)

ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ManageRole", "Manage Roles", "Manage Roles", ["EditRole"], extends="ViewRole"
)

ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ViewEventLog",
    "View event log",
    "View actions taken by users and the system itself",
    ["ViewEventLog"],
)

# GPG Keys
# Note: ViewGPGKeys is given to all AccountMembers
ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ManageGPGKey", "Manage GPG keys", "Manage GPG keys", ["CreateGPGKey", "EditGPGKey"]
)

# Upgrade Profiles
MACRO_PERMISSIONS.add_macro_permission(
    "ViewRemovalProfile",
    "View removal profiles",
    "View removal profiles",
    ["ViewRemovalProfile"],
)

MACRO_PERMISSIONS.add_macro_permission(
    "ManageRemovalProfile",
    "Manage removal profiles",
    "Manage removal profiles",
    ["EditRemovalProfile"],
    extends="ViewRemovalProfile",
)

ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ViewBilling", "View Billing", "View Billing", ["ViewBilling"]
)

ACCOUNT_MACRO_PERMISSIONS.add_macro_permission(
    "ManageBilling",
    "Manage Billing",
    "Manage Billing",
    ["ManageBilling"],
    extends="ViewBilling",
)
