# Copyright 2005-2022 Canonical Limited. All rights reserved.

import re

from zope.component import getUtility
from zope.interface import implementer

from storm.locals import Int, Reference, AutoReload, Unicode, Join
from storm.expr import Func

from canonical.security.role import Role
from canonical.security.interfaces import IRoleRegistry
from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.security.roles import STORED_ROLE_KEY_MIN
from canonical.landscape.model.access.permissions import (
    MACRO_PERMISSIONS,
    DEFAULT_PERMISSIONS,
    ACCOUNT_MACRO_PERMISSIONS,
    get_macro_permissions,
)
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.accesscontext import AccessContext
from canonical.landscape.model.main.personaccess import get_persons_with_role
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.alert.account import AccountRuleInfo
from canonical.landscape.model.access.errors import (
    InvalidPermissionsError,
    ReadOnlyRoleError,
)
from canonical.landscape.model.main.personaccess import PersonAccess
from canonical.landscape.model.main.interfaces import (
    IAccountAware,
    IAccessDelegateAware,
)
from canonical.landscape.model.account.invitation import AccountInvitationRole


ROLE_NAME_RE = re.compile(r"^[a-zA-Z][-a-zA-Z0-9+]+$")


__clearlist__ = [
    "get_stored_roles",
    "is_valid_role_name",
    "StoredRole",
    "get_role_registry",
]

__requires_prefix__ = "landscape."


def is_valid_role_name(name):
    """Return true if the provided string is a valid name for a role."""
    return bool(ROLE_NAME_RE.match(name))


@requires_advisor
@implementer(IAccountAware, IAccessDelegateAware)
class StoredRole(object):
    """Represent a role with a custom set of permissions.

    @param name: Short name for the role.
    @param description: Optional description for the role.
    """

    requires("ViewRole", PRIVATE, ["id", "editable", "key"])
    requires("ViewAccount", "EditRole", ["name", "account"])

    __storm_table__ = "role"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    description = Unicode()

    account = Reference(account_id, Account.id)

    editable = True

    def __init__(self, account, name, description=None):
        self.name = name
        self.description = description
        self.account = account

    @property
    def access_delegate(self):
        return self.account

    @property
    def key(self):
        """Return the key of this role."""
        # Note that we just use the database ID plus 100000, so we don't
        # risk collisions with the stock in-memory roles
        return self.id + STORED_ROLE_KEY_MIN

    def _check_permission_validity(self, macro_permissions):
        invalid_macro_permissions = set(macro_permissions) - set(
            get_macro_permissions(system=False)
        )
        if invalid_macro_permissions:
            raise InvalidPermissionsError(invalid_macro_permissions)

    @requires("EditRole")
    def add_permissions(self, macro_permissions):
        """Add permissions to this L{Role}.

        @param macro_permissions : a list of macro permissions that will be
            added to the role.  Macro permissions are expanded to micro
            permissions in L{get_role_registry} below.

        @raises InvalidPermissionsError: If permissions that are not
            listed in the supported ones are passed.

        """
        if not macro_permissions:
            return

        self._check_permission_validity(macro_permissions)
        new_macro_permissions = (
            set(macro_permissions)
            - set(self.get_permissions())
            - set(self.get_account_permissions())
        )
        for macro_permission in new_macro_permissions:
            StoredRolePermission(self, macro_permission)

    @requires("EditRole")
    def remove_permissions(self, macro_permissions):
        """Remove permissions from this L{Role}.

        @param macro_permissions: a list of macro permissions that will be
            removed from the role.

        @raises InvalidPermissionsError: If permissions that are not
            listed in the supported ones are passed.
        """
        self._check_permission_validity(macro_permissions)
        store = get_main_store()
        store.find(
            StoredRolePermission,
            StoredRolePermission.permission.is_in(macro_permissions),
            StoredRolePermission.role_id == self.id,
        ).remove()

    def _get_stored_role_permissions(self):
        """
        Convenience method for retrieving L{StoredRolePermission} objects.
        """
        store = get_main_store()
        result = store.find(
            StoredRolePermission, StoredRolePermission.role_id == self.id
        )
        return list(result.values(StoredRolePermission.permission))

    @requires("ViewRole")
    def get_permissions(self):
        """Return all the macro permissions associated with the role."""
        permissions = set(self._get_stored_role_permissions()).difference(
            list(ACCOUNT_MACRO_PERMISSIONS.keys())
        )
        return list(permissions)

    @requires("ViewRole")
    def get_account_permissions(self):
        """
        Return the account macro permissions filtering out normal macro
        permissions.
        """
        permissions = set(self._get_stored_role_permissions()).difference(
            list(MACRO_PERMISSIONS.keys())
        )
        return list(permissions)

    @requires("EditRole")
    def add_persons(self, emails):
        """Grant this L{Role} to the L{Person}s with the given emails.

        @raises UnknownPersonsErrors: If one or more of the given emails
            are unknown.
        """
        # The following call will raise UnknownPersonsError if some
        # emails are invalid
        persons = self.account.get_persons(emails)
        new_persons = set(persons) - set(self.get_persons())
        for person in new_persons:
            StoredRolePerson(self, person)
            # Update also the low-level person_access table
            for access_context in self.get_access_contexts():
                access_context.add_role(person, self)

    @requires("EditRole")
    def remove_persons(self, emails):
        """Revoke this L{Role} on the L{Person}s with the given emails.

        @raises UnknownPersonsErrors: If one or more of the given emails
            are unknown.
        """
        # The following call will raise UnknownPersonsError if some
        # emails are invalid
        persons = self.account.get_persons(emails)
        person_ids = [person.id for person in persons]
        store = get_main_store()
        store.find(
            StoredRolePerson,
            StoredRolePerson.person_id.is_in(person_ids),
            StoredRolePerson.role_id == self.id,
        ).remove()
        # Update also the low-level person_access table
        for access_context in self.get_access_contexts():
            access_context.remove_persons(self, persons)

    @requires("ViewPerson")
    def get_persons(self):
        """
        Return a result set yielding all L{Person}s that have this role.
        """
        store = get_main_store()
        result = store.find(
            Person,
            Person.id == StoredRolePerson.person_id,
            StoredRolePerson.role_id == self.id,
        )
        return result

    @requires("EditRole")
    def add_access_contexts(self, names):
        """Add the access contexts with the given names to this L{Role}."""
        if not names:
            return

        # The following call will raise UnknownAccessContextsError if some
        # names are invalid
        access_contexts = self.account.get_access_contexts(names=names)
        new_access_contexts = set(access_contexts) - set(self.get_access_contexts())
        for access_context in new_access_contexts:
            StoredRoleAccessContext(self, access_context)
            # Update also the low-level person_access table
            for person in self.get_persons():
                access_context.add_role(person, self)

    @requires("EditRole")
    def remove_access_contexts(self, names):
        """
        Remove the access contexts with the given names from this L{Role}.
        """
        access_contexts = self.account.get_access_contexts(names=names)
        store = get_main_store()
        for access_context in access_contexts:
            for person in self.get_persons():
                access_context.remove_role(person, self)
            result = store.find(
                StoredRoleAccessContext,
                StoredRoleAccessContext.role_id == self.id,
                StoredRoleAccessContext.access_context_id == access_context.id,
            )
            result.remove()

    @requires("ViewAccessContext")
    def get_access_contexts(self):
        """
        Return a result set yielding all L{AccessContext}s that this role has.
        """
        store = get_main_store()
        result = store.find(
            AccessContext,
            AccessContext.id == StoredRoleAccessContext.access_context_id,
            StoredRoleAccessContext.role_id == self.id,
        )
        return result

    @requires("EditRole")
    def edit(self, permissions=None, access_context_names=None):
        """Change the roles parameters in bulk.

        @permissions: The list of new permissions for the role.
        @param access_context_names: The names of the new access contexts.
        """
        if permissions is not None:
            # Permissions
            existing_permissions = set(self._get_stored_role_permissions())
            self.remove_permissions(existing_permissions - set(permissions))
            self.add_permissions(permissions)

        if access_context_names is not None:
            # Access groups
            result = self.get_access_contexts()
            existing_names = set(access_context.name for access_context in result)
            self.remove_access_contexts(existing_names - set(access_context_names))
            self.add_access_contexts(access_context_names)

    @requires("EditRole")
    def remove(self):
        """Remove this role from the account."""
        persons_ids = [person.id for person in self.get_persons()]
        store = get_main_store()
        store.find(
            PersonAccess,
            PersonAccess.person_id.is_in(persons_ids),
            PersonAccess.role_key == self.key,
        ).remove()
        store.find(
            AccountInvitationRole, AccountInvitationRole.role_key == self.key
        ).remove()
        store.remove(self)

    def to_role(self):
        """
        Get a low-level L{Role} suited to be added to the registry.

        Note that permissions are *not* included in the returned Role.
        """
        # Add the account ID as prefix to avoid conflicts between roles
        # with the same name in different accounts.
        name = "%d.%s" % (self.account_id, self.name)
        return Role(name, key=self.key)

    def get_state(self):
        """Return a dict with information on the role."""
        emails = sorted(self.get_persons().values(Person.email))
        names = sorted(self.get_access_contexts().values(AccessContext.name))
        return {
            "name": self.name,
            "description": self.description or "",
            "permissions": self.get_permissions(),
            "global_permissions": self.get_account_permissions(),
            "persons": emails,
            "access_groups": names,
        }


@requires_advisor
@implementer(IAccessDelegateAware)
class GlobalAdmin(object):
    """An in-memory custom role that maps to the internal Owner role."""

    requires("ViewAccount", PRIVATE, ["name"])
    requires("ViewRole", PRIVATE, ["editable"])

    name = "GlobalAdmin"
    description = "Global role with full privileges in the account"

    editable = False

    def __init__(self, account):
        self.account = account

    @property
    def access_delegate(self):
        return self.account

    @property
    def key(self):
        owner = getUtility(IRoleRegistry).get("Owner")
        return owner.key

    @requires("ViewPerson")
    def get_persons(self):
        owner = getUtility(IRoleRegistry).get("Owner")
        return get_persons_with_role(owner, self.account.access_context)

    @requires("EditRole")
    def add_persons(self, emails):
        """
        Grant the GlobalAdmin role to the persons with the given emails.
        """
        # As this an in-memory role, we manipulate directly the
        # person_access table
        owner = getUtility(IRoleRegistry).get("Owner")
        persons = self.account.get_persons(emails)
        new_persons = set(persons) - set(self.get_persons())
        for person in new_persons:
            self.account.access_context.add_role(person, owner)

    @requires("EditRole")
    def remove_persons(self, emails):
        """
        Remove the persons with the given emails from the GlobalAdmin role.
        """
        # As this an in-memory role, we manipulate directly the
        # person_access table
        registry = getUtility(IRoleRegistry)
        owner = registry.get("Owner")
        account_member = registry.get("AccountMember")
        persons = self.account.get_persons(emails)
        main_access_context = self.account.access_context
        main_access_context.remove_persons(owner, persons)

        for person in persons:
            self._unsubscribe_account_rules(person)
            if not person.has_role(account_member, main_access_context):
                main_access_context.add_role(person, account_member)

    def _unsubscribe_account_rules(self, person):
        """
        Unsubscribe the person from account rules.
        """
        for rule in self.account.get_subscribed_rules(person):
            if isinstance(rule.info, AccountRuleInfo):
                rule.unsubscribe(person)

    @requires("ViewAccessContext")
    def get_access_contexts(self):
        """The global admin role is associated with the root access context."""
        return [self.account.access_context]

    def add_access_contexts(self, names):
        raise ReadOnlyRoleError(self.name)

    def remove_access_contexts(self, names):
        raise ReadOnlyRoleError(self.name)

    def get_permissions(self):
        return get_macro_permissions(system=False)

    def add_permissions(self, macro_permissions):
        raise ReadOnlyRoleError(self.name)

    def remove_permissions(self, macro_permissions):
        raise ReadOnlyRoleError(self.name)

    def remove(self):
        raise ReadOnlyRoleError(self.name)

    def edit(self, args, *kwargs):
        raise ReadOnlyRoleError(self.name)

    def get_state(self):
        emails = sorted(self.get_persons().values(Person.email))
        return {
            "name": self.name,
            "description": self.description,
            "permissions": self.get_permissions(),
            "persons": emails,
            "access_groups": [self.account.access_context.name],
        }


class StoredRolePermission(object):
    """Associate a permission to a L{StoredRole}.

    @param role: The L{StoredRole} to add the permission to.
    @param permission: The permission to add.
    """

    __storm_table__ = "role_permission"

    __storm_primary__ = "role_id", "permission"

    role_id = Int(allow_none=False)
    permission = Unicode(allow_none=False)

    role = Reference(role_id, StoredRole.id)

    def __init__(self, role, permission):
        self.role = role
        self.permission = permission


class StoredRoleAccessContext(object):
    """Associate an L{AccessContext} to a L{StoredRole}.

    @param role: The L{StoredRole} to link the L{AccessContext} to.
    @param access_context: The L{AccessContext} to link.
    """

    __storm_table__ = "role_access_context"

    __storm_primary__ = "role_id", "access_context_id"

    role_id = Int(allow_none=False)
    access_context_id = Int(allow_none=False)

    role = Reference(role_id, StoredRole.id)
    access_context = Reference(access_context_id, AccessContext.id)

    def __init__(self, role, access_context):
        self.role = role
        self.access_context = access_context


class StoredRolePerson(object):
    """Grant a L{StoredRole} to a L{Person}.

    @param person: The L{Person} to grant the L{StoredRole} to.
    @param role: The L{StoredRole} to grant.
    """

    __storm_table__ = "role_person"

    __storm_primary__ = "role_id", "person_id"

    role_id = Int(allow_none=False)
    person_id = Int(allow_none=False)

    role = Reference(role_id, StoredRole.id)
    person = Reference(person_id, Person.id)

    def __init__(self, role, person):
        self.role = role
        self.person = person


def get_stored_roles(account_ids):
    """
    Return a list of binary tuples (role, permissions) for all roles available
    in the given C{account_ids}.
    """
    store = get_main_store()
    roles = list(store.find(StoredRole, StoredRole.account_id.is_in(account_ids)))
    spec = (
        StoredRolePermission.role_id,
        Func("ARRAY_AGG", StoredRolePermission.permission),
    )
    join = Join(
        StoredRolePermission, StoredRole, StoredRolePermission.role_id == StoredRole.id
    )
    role_ids = [role.id for role in roles]
    result = store.using(join).find(spec, StoredRole.id.is_in(role_ids))
    result.group_by(StoredRolePermission.role_id)
    permissions_index = dict(result)
    return [(role, permissions_index.get(role.id, [])) for role in roles]


class AccountRole(Role):
    """
    A special type of L{Role} that automatically applies to the account's
    access context.

    When creating the registry of basic L{Role}s based on L{StoredRole}s, with
    L{get_role_registry}, one of these is created for any L{StoredRole} that
    has account permissions of any sort. This is in addition to the "main"
    L{Role}, which will only contain the normal, access-context-bound
    permissions.

    As an example, if a L{StoredRole} contains a macro permission "ManageFoo"
    that contains a regular micro-permission "EditFoo" and an
    account_permission "ViewFoo", then L{get_role_registry} will return two
    roles:

      - a L{Role} named "<account id>.ManageFoo" that has the "EditFoo"
        permission.
      - an L{AccountRole} named "account.<account id>.ManageFoo" that has the
        "ViewFoo" permission.

    No explicit L{PersonAccess} objects will be created with this kind of role.

    This class provides no functionality other than storing the additional data
    that links it with the L{StoredRole}. It's the responsibility of any
    security-checking code to check to see if a role is an L{AccountRole} and
    act accordingly, by applying the L{AccountRoles} to the account's access
    context, and the regular L{Role}s to the associated access contexts.

    See get_access_context_ids_for_person() and Person.get_all_roles().

    @param name: The name of this account role. Should be in the format of
        "account.<account_id>.<name>"
    @param stored_role_id: The ID of the L{StoredRole} that this L{AccountRole}
        is derived from.
    """

    def __init__(self, name, stored_role_id):
        super(AccountRole, self).__init__(name)
        self.stored_role_id = stored_role_id


def get_persons_with_account_permission(account, permission):
    """
    Return Persons in an account that have the specified micro-permission
    on the account access context.
    """
    return [
        person
        for person in account.people
        if person.has_permission(
            "landscape.{}".format(permission), account.access_context
        )
    ]


def get_role_registry(account_ids):
    """
    Return a new L{RoleRegistry} populated with the regular stock roles plus
    all stored roles associated with the given account IDs.
    """
    # Clone the global registry, this will preserve the hard-wired roles and
    # permissions. Note that the existing role instances will be copied by
    # reference so code that uses the global role registry to look up roles
    # will return the very same instaces as if it looked up this cloned
    # registry
    registry = getUtility(IRoleRegistry).clone()

    # Load stored roles and their permissions from the database
    for stored_role, macro_permissions in get_stored_roles(account_ids):
        role = stored_role.to_role()
        registry.add(role)

        # This account role is always instantiated here, but it's only added to
        # the registry in case we actually find any account-permissions in the
        # macro permissions.
        account_role_name = "account.%d.%s" % (stored_role.account_id, stored_role.name)
        account_role = AccountRole(account_role_name, stored_role.id)

        # Map user-level macro permissions to model-level permissions.
        for macro_permission_name in macro_permissions:
            macro_permission = MACRO_PERMISSIONS.get(macro_permission_name)
            if macro_permission is None:
                macro_permission = ACCOUNT_MACRO_PERMISSIONS[macro_permission_name]
                _grant_permissions(registry, account_role, macro_permission.permissions)
            else:
                _grant_permissions(registry, role, macro_permission.permissions)
                _grant_permissions(
                    registry, account_role, macro_permission.account_permissions
                )

        # Default permissions granted to every role
        for permission in DEFAULT_PERMISSIONS:
            registry.grant(role, "landscape." + permission)

    return registry


def _grant_permissions(registry, role, permissions):
    """
    Ensure a role is in the registry, and grant all permissions in the given
    list.
    """
    if role not in registry:
        registry.add(role)
    for permission in permissions:
        registry.grant(role, "landscape." + permission)
