# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import timedelta
from collections import defaultdict

from zope.component import getUtility

from canonical.lib.mail import is_email_address
from canonical.lib.timestamp import utctoday
from canonical.security.interfaces import IRoleRegistry
from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.mail.template import load_template
from canonical.landscape.model.main.account import get_all_accounts
from canonical.landscape.model.configuration import (
    get_root_url,
    ISystemConfiguration,
    LANDSCAPE_SYSTEM_EMAIL,
)
from canonical.landscape.model.alert.account import AccountRuleInfo
from canonical.landscape.model.alert.mail import AlertEmailViewBase


__all__ = [
    "PendingComputersAlert",
    "StandaloneOverflowAlert",
    "LicenseExpiryAlert",
    "StandaloneScriptAlert",
    "StandaloneSystemEmailAlert",
    "check_pending_expired_accounts",
]

__clearlist__ = ["LicenseExpiryAlert", "StandaloneSystemEmailAlert"]

__requires_prefix__ = "landscape."


@requires_advisor
class PendingComputersAlert(AccountRuleInfo):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when pending computers are available"

    notification_delay = timedelta(minutes=15)

    rule_requires = "landscape.ViewPendingComputer"

    def check(self):
        old_state_data = self.get_stored_state_data()
        new_state_data = self.get_state_data()

        if not new_state_data:
            self.state.deactivate()
        elif old_state_data is None or not set(old_state_data).issuperset(
            new_state_data
        ):
            self.state.activate()

    def get_state_data(self):
        """See RuleInfo.get_state_data."""
        from canonical.landscape.model.main.computer import PendingComputer

        result = self.rule.account.get_pending_computers()
        return sorted(result.values(PendingComputer.id))

    def get_summary(self):
        count = self.rule.account.get_pending_computers().count()
        if not count:
            return "No pending computers need authorization"
        elif count > 1:
            return "%s pending computers need authorization" % count
        else:
            return "1 pending computer needs authorization"

    def get_resolution_url(self):
        if self.state.activation_time:
            return "%saccount/%s/pending-computers" % (
                get_root_url(),
                self.rule.account.name,
            )
        return None


class PendingComputersEmailView(AlertEmailViewBase):

    _content_template = load_template(__file__, "pending-computers-alert.msg")

    def __init__(self, context, person):
        super(PendingComputersEmailView, self).__init__(context, person)
        self.pending_computers = []
        for pending_computer in context.rule.account.get_pending_computers():
            self.pending_computers.append(
                (pending_computer.title, pending_computer.hostname)
            )
        self.count = len(self.pending_computers)


@requires_advisor
class LicenseExpiryAlert(AccountRuleInfo):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when license is about to expire"

    notification_delay = timedelta(minutes=15)

    _ALERT_HORIZON = 30

    def check(self):
        """
        If the number of seats available in self._ALERT_HORIZON
        days is less than it was last time we checked AND that lower
        number is less than current seat usage then alert.
        """
        if self.rule.account.is_self_service:
            return
        old_state_data = self.get_stored_state_data()
        new_state_data = self.get_state_data()
        if not isinstance(old_state_data, int):
            previous_missing_seats = 0
        else:
            previous_missing_seats = old_state_data
        missing_seats = new_state_data
        if missing_seats == 0:
            self.state.deactivate()
        elif missing_seats <= previous_missing_seats:
            # Don't reactivate if the usage is steady or going down
            pass
        else:
            self.state.activate()

    def get_state_data(self):
        """See RuleInfo.get_state_data."""
        return self._get_missing_seats_in_days(self._ALERT_HORIZON)

    def get_summary(self):
        future_missing_seats = self.get_state_data()
        if future_missing_seats > 0:
            return "License about to expire"
        else:
            return "No licenses about to expire."

    def get_resolution_url(self):
        if self.state.activation_time:
            return "%saccount/%s/licenses" % (get_root_url(), self.rule.account.name)
        else:
            return None

    def _get_missing_seats_in_days(self, number_of_days):
        from canonical.landscape.model.license.management import (
            LicenseManagement,
            EXPIRABLE_LICENSE_TYPES,
        )

        management = LicenseManagement(self.rule.account)
        date = utctoday() + timedelta(days=number_of_days)

        # Collect per-license-type used and total available seats.
        total_seats = defaultdict(int)
        used_seats = defaultdict(int)
        licenses = management.get_licenses(role_names=EXPIRABLE_LICENSE_TYPES)
        for license in licenses:
            used_seats[license.role_key] += license.get_used_seats()
            is_infinite = license.expiration_date is None
            if is_infinite or license.expiration_date > date:
                total_seats[license.role_key] += license.seats

        # Compute the total number of missing seats, considering which ones can
        # be covered by a more generic type of license.
        registry = getUtility(IRoleRegistry)
        # License roles are sorted so that seats of each role can fallback on
        # the following if they exceed availability, since it's more generic.
        role_keys = [
            registry.get(role).key
            for role in ("ContainerFeatures", "VMFeatures", "BasicFeatures")
        ]
        missing = 0
        for key in role_keys:
            missing = max(0, used_seats[key] - total_seats[key] + missing)
        return missing


class LicenseExpiryEmailView(AlertEmailViewBase):

    _content_template = load_template(__file__, "license-expiry-alert.msg")

    def __init__(self, context, person):
        super(LicenseExpiryEmailView, self).__init__(context, person)
        self.expiration_dates = ", ".join(self._generate_expiring_licenses(context))

    def _generate_expiring_licenses(self, context):
        from canonical.landscape.model.license.management import (
            LicenseManagement,
            EXPIRABLE_LICENSE_TYPES,
        )

        management = LicenseManagement(context.rule.account)
        horizon = utctoday() + timedelta(days=30)

        licenses = management.get_licenses(
            role_names=EXPIRABLE_LICENSE_TYPES, expire_before=horizon
        )
        licenses.sort(key=lambda license: license.expiration_date)
        for license in licenses:
            yield str(license.expiration_date)


def check_pending_expired_accounts():
    """
    Check the LicenseExpiryAlert rule for each account.
    """
    for account in get_all_accounts():
        account.check_rules([LicenseExpiryAlert])


@requires_advisor
class StandaloneOverflowAlert(AccountRuleInfo):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when exceeding licence entitlements"

    notification_delay = timedelta(minutes=15)

    def check(self):
        old_state_data = self.get_stored_state_data()
        new_state_data = self.get_state_data()

        if not new_state_data:
            self.state.deactivate()
        elif old_state_data is None or new_state_data > old_state_data:
            self.state.activate()

    def get_state_data(self):
        """See RuleInfo.get_state_data."""
        overflow_seats = [
            license.get_used_overflow_seats() for license in self.rule.account.licenses
        ]
        if overflow_seats:
            return sum(overflow_seats)

    def get_summary(self):
        used_seats = self.get_state_data()
        if used_seats:
            return "License overflow alert"
        else:
            return "No overflow currently used."

    def get_resolution_url(self):
        if self.state.activation_time:
            return "%saccount/%s/licenses" % (get_root_url(), self.rule.account.name)
        else:
            return None


class StandaloneOverflowEmailView(AlertEmailViewBase):

    _content_template = load_template(__file__, "standalone-overflow-alert.msg")

    def __init__(self, context, person):
        super(StandaloneOverflowEmailView, self).__init__(context, person)
        overflow_seats = [
            license.get_used_overflow_seats()
            for license in context.rule.account.licenses
        ]
        self.overflow_count = sum(overflow_seats)


@requires_advisor
class StandaloneSystemEmailAlert(AccountRuleInfo):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when system email is misconfigured"

    notification_delay = timedelta(minutes=15)

    rule_requires = "landscape.EditAccount"

    def check(self):
        old_state_data = self.get_stored_state_data()
        new_state_data = self.get_state_data()

        if not new_state_data:
            self.state.deactivate()
        elif new_state_data != old_state_data:
            self.state.activate()

    def get_state_data(self):
        """See RuleInfo.get_state_data.

        Return True when system email is misconfigured.
        """
        return not is_email_address(self._get_system_email())

    def get_summary(self):
        system_email = self._get_system_email()
        if is_email_address(system_email):
            return "System email is configured: %s" % system_email
        elif system_email:
            return "System email is misconfigured: %s" % system_email
        else:
            return "System email is unconfigured"

    def get_resolution_url(self):
        if self.state.active:
            account_name = self.rule.account.name
            return "%saccount/%s/standalone-settings" % (get_root_url(), account_name)
        else:
            return None

    def _get_system_email(self):
        """Return LANDSCAPE_SYSTEM_EMAIL from system configuration.

        Return an empty string if unconfigured.
        """
        configuration = getUtility(ISystemConfiguration)
        return configuration.get(LANDSCAPE_SYSTEM_EMAIL, "")


class StandaloneSystemEmailEmailView(AlertEmailViewBase):

    _content_template = load_template(__file__, "standalone-system-email-alert.msg")

    def __init__(self, context, person):
        super(StandaloneSystemEmailEmailView, self).__init__(context, person)
        configuration = getUtility(ISystemConfiguration)
        self.system_email = configuration.get(LANDSCAPE_SYSTEM_EMAIL)


@requires_advisor
class StandaloneScriptAlert(AccountRuleInfo):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when a Landscape cron job fails"

    notification_delay = timedelta(minutes=15)

    def check(self):
        old_state_data = self.get_stored_state_data()
        new_state_data = self.get_state_data()

        if not new_state_data:
            self.state.deactivate()
        elif new_state_data != old_state_data:
            self.state.activate()

    def get_state_data(self):
        from canonical.landscape.model.main.standalone import (
            LDSScriptExecutionManagement,
        )

        management = LDSScriptExecutionManagement()
        return [
            (
                execution.name,
                execution.ran,
                execution.exceeded_interval,
                execution.exceeded_timeout,
                execution.success,
            )
            for execution in management.get_executions(only_errored=True)
        ]

    def get_scripts_by_state(self):
        """
        Return two lists, with names of scripts that failed and did not run.
        """
        failed, not_run, timed_out = [], [], []
        for state_data in self.get_state_data():
            name, ran, exceeded_interval, exceeded_run, success = state_data
            if success is False:  # can be None if the script didn't run
                failed.append(name)
            elif exceeded_run:
                timed_out.append(name)
            elif not ran or exceeded_interval:
                not_run.append(name)
        return failed, not_run, timed_out

    def get_summary(self):
        failed, not_run, timed_out = [
            len(state) for state in self.get_scripts_by_state()
        ]
        total = failed + not_run + timed_out

        if total:
            detail = []
            if not_run:
                detail.append("%d didn't run" % not_run)
            if failed:
                detail.append("%d failed" % failed)
            if timed_out:
                detail.append("%d exceeding run timeout" % timed_out)

            if total == 1:
                message = "%d Landscape script has issues: %s"
            else:
                message = "%d Landscape scripts have issues: %s"
            return message % (total, ", ".join(detail))
        else:
            return "No Landscape script has issues."

    def get_resolution_url(self):
        return "/script-issues" if self.state.active else None


class StandaloneScriptEmailView(AlertEmailViewBase):

    _content_template = load_template(__file__, "standalone-scripts-alert.msg")

    def __init__(self, context, person):
        super(StandaloneScriptEmailView, self).__init__(context, person)
        (
            self.failed,
            self.not_run,
            self.timed_out,
        ) = context.rule.info.get_scripts_by_state()
        self.count = len(self.failed) + len(self.not_run) + len(self.timed_out)
