# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Collection logic to filter and find L{Account}s."""

from storm.locals import Or, Not
from storm.store import EmptyResultSet

from canonical.collection import CollectionBase

from canonical.landscape.model.stormfunc import Today


__all__ = ["AccountCollection"]


class AccountCollection(CollectionBase):
    """A collection of L{Account}s that can be filtered in various ways."""

    def with_account_ids(self, account_ids):
        """Only include L{Account}s with an C{Account.id} in C{account_ids}.

        If C{account_ids} is an empty C{list} the collection will yield an
        C{EmptyResultSet} when its L{find} method is called.

        @param account_ids: A sequence of L{Account.id}s to filter by.
        @return: A new L{AccountCollection} filtered by L{Account.id}s.
        """
        return self._filter_by(account_ids=account_ids)

    def with_account_names(self, account_names):
        """
        Only include L{Account}s with an C{Account.name} in C{account_names}.

        If C{account_names} is an empty C{list} the collection will yield an
        C{EmptyResultSet} when its L{find} method is called.

        @param account_names: A sequence of L{Account.name}s to filter by.
        @return: A new L{AccountCollection} filtered by L{Account.name}s.
        """
        return self._filter_by(account_names=account_names)

    def with_active(self, expiration_date=None):
        """
        Only include L{Account}s with L{License}s that will be active after
        C{expiration_date}.

        @param expiration_date: Optionally, the expiration date to use in the
            check, defaults to now.
        @return: A new L{AccountCollection} filtered by expired L{Account}.
        """
        return self._filter_by(expired=False, expiration_date=expiration_date)

    def with_expired(self, expiration_date=None):
        """Only include L{Account}s that expire before C{expiration_date}.

        @param expiration_date: Optionally, the expiration date to use in the
            check, defaults to now.
        @return: A new L{AccountCollection} filtered by expired L{Account}.
        """
        return self._filter_by(expired=True, expiration_date=expiration_date)

    def with_self_service(self, flag):
        """Only include Accounts with the specified self-service status.

        If None is passed, the self-service status will be ignored.
        """
        return self._filter_by(self_service=flag)

    def _get_result(
        self,
        store,
        where=None,
        account_ids=None,
        account_names=None,
        expired=None,
        expiration_date=None,
        self_service=None,
    ):
        """
        Return a L{ResultSet} configured to yield L{Account}s matching filter
        criteria.
        """
        from canonical.landscape.model.main.account import Account

        if where is None:
            where = []
        if account_ids is not None:
            if not account_ids:
                return EmptyResultSet()
            where.append(Account.id.is_in(account_ids))
        if account_names is not None:
            if not account_names:
                return EmptyResultSet()
            where.append(Account.name.is_in(account_names))
        if expired is not None:
            where.extend(
                self._get_expired_where_clauses(
                    store, expired, expiration_date, account_ids
                )
            )
        if self_service is not None:
            if self_service:
                where.append(Account.stripe_customer_id != None)
            else:
                where.append(Account.stripe_customer_id == None)
        return store.find(Account, *where)

    def _get_expired_where_clauses(self, store, expired, expiration_date, account_ids):
        """Build the where clause for the C{expired} filtering criteria."""
        from canonical.landscape.model.main.account import Account
        from canonical.landscape.model.main.license import License

        if expiration_date is None:
            expiration_date = Today

        where = [
            License.account_id == Account.id,
            License.expiration_date != None,
            License.expiration_date < expiration_date,
        ]
        if account_ids:
            where.append(License.account_id.is_in(account_ids))
        result = store.find(License, *where)
        result.config(distinct=True)
        expired_account_ids = result.get_select_expr(License.account_id)

        where = [
            License.account_id == Account.id,
            License.expiration_date >= expiration_date,
        ]
        if account_ids:
            where.append(License.account_id.is_in(account_ids))
        result = store.find(License, *where)
        result.config(distinct=True)
        active_account_ids = result.get_select_expr(License.account_id)

        where = [License.account_id == Account.id, License.expiration_date == None]
        if account_ids:
            where.append(License.account_id.is_in(account_ids))
        result = store.find(License, *where)
        result.config(distinct=True)
        perpetual_account_ids = result.get_select_expr(License.account_id)

        if expired:
            return [
                Account.id.is_in(expired_account_ids),
                Not(Account.id.is_in(perpetual_account_ids)),
                Not(Account.id.is_in(active_account_ids)),
            ]
        else:
            return [
                Or(
                    Not(Account.id.is_in(expired_account_ids)),
                    Account.id.is_in(perpetual_account_ids),
                )
            ]
