# Copyright 2005-2022 Canonical Limited. All rights reserved.
"""Track which computers are associated with a certain group of tags.

A "computer group" is defined as the set of computers that are associated
with a certain group of tags.

This logic is particularly useful for profiles, that enforce conditions across
all computers with certain tags. Consumers of this machinery use composition,
in that they hold a reference to a computer group ID and are the "owners" of
that computer group.
"""

from storm.locals import Int, Bool, AutoReload, ReferenceSet, Store

from zope.security import proxy

from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.tag import Tag
from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.account.helper import AccountHelper


__all__ = [
    "ComputerGroupItem",
    "ComputerGroupTag",
    "ComputerGroup",
    "AccountComputerGroup",
    "register_profile_info_type",
    "profile_info_types",
]

__requires_prefix__ = "landscape."


profile_info_types = {}
reverse_profile_info_types = {}


def register_profile_info_type(info_type, info_class):
    """Register C{info_class} as a L{ComputerGroup} user.

    The C{info_type} will be stored as C{ComputerGroup.info_type}.

    @param info_type: A unique integer representing C{info_class}.
    @param info_class: A L{ComputerGroupHelper}-based class.
    @raise RuntimeError: Raised if a class is already registered for
        C{info_type}.
    """
    existing_info_class = profile_info_types.get(info_type)
    if existing_info_class is not None:
        raise RuntimeError(
            "%r has the same info_type of %r" % (info_class, existing_info_class)
        )
    profile_info_types[info_type] = info_class
    reverse_profile_info_types[info_class] = info_type
    info_class.group_info_type = info_type


class ComputerGroupItem(ComputerHelper):
    """
    Represent the association of a computer and a computer group.

    This modelizes the 1-N relationship between L{ComputerGroup} and
    L{Computer}, automatically updated with the L{ComputerGroupTag}
    relationship.

    This relation is not strictly needed, as it could be derived from
    the L{ComputerGroupTag} and L{ComputerTag} relations, but
    a denormalization has been done for better performance and
    scalability.
    """

    __storm_table__ = "computer_group_item"
    __storm_primary__ = "computer_id", "computer_group_id"

    computer_group_id = Int(allow_none=False)
    computer_id = Int(allow_none=False)


class ComputerGroupTag(object):
    """
    Represent the assocation of a tag and a computer group.

    This modelizes the 1-N relationship between C{ComputerGroup} and C{Tag}.
    """

    __storm_table__ = "computer_group_tag"
    __storm_primary__ = "tag_id", "computer_group_id"

    computer_group_id = Int(allow_none=False)
    tag_id = Int(allow_none=False)

    def __init__(self, tag=None):
        if tag is not None:
            if proxy.isinstance(tag, (int, int)):
                self.tag_id = tag
            else:
                self.tag_id = tag.id


class ComputerGroup(AccountHelper):
    """
    A group of computers attached to an account.

    This object represents a set of computers using the tags attached to them:
    a computer group is directly linked to a set of tags (or the magic tag "all
    computers"): when a tag is added or removed to a computer, this computer is
    automatically added or removed to the computer group. Similary, if a tag is
    attached or detached from a computer group, all the computers in this tag
    are added or removed from this group (except if they are belonging to
    another tag attached to this group).

    As the logic is quite complicated, and used in several places (for now,
    Rule and CustomGraph), all the dirty work is done here and in
    C{AccountComputerGroup}.
    """

    __storm_table__ = "computer_group"

    id = Int(primary=True, default=AutoReload)
    account_id = Int(allow_none=False)
    all_computers = Bool(allow_none=False, default=False)
    info_type = Int(allow_none=False)
    _info = None

    _computers = ReferenceSet(
        id,
        ComputerGroupItem.computer_group_id,
        ComputerGroupItem.computer_id,
        AccountComputerTwin.id,
    )

    @property
    def computers(self):
        """
        Get the AccountComputerTwin associated with this group.
        Archived/removed computers are ignored.
        """
        return self.account.store.find(
            AccountComputerTwin,
            ComputerGroupItem.computer_id == AccountComputerTwin.id,
            ComputerGroupItem.computer_group_id == self.id,
        )

    tags = ReferenceSet(id, ComputerGroupTag.computer_group_id)

    def __init__(self, account, info):
        super(ComputerGroup, self).__init__(account)
        self.all_computers = False

        self.info_type = info.group_info_type
        self._info = info

    def add_computer(self, twin):
        """
        Add a L{ComputerGroupItem} associating the AccountComputerTwin with
        this group.
        """
        self._computers.add(twin)

    def _get_info(self):
        if self._info is None:
            info_class = profile_info_types[self.info_type]
            self._info = self.account.store.find(
                info_class, info_class.computer_group_id == self.id
            ).one()
        return self._info

    info = property(_get_info)


@requires_advisor
class AccountComputerGroup(AccountHelper):
    """
    Managed computer groups linked to an account when tags are added or
    removed from computers.
    """

    @requires("SynchronizeTags")
    def synchronize_with_added_tags(
        self, tag_ids, computer_ids, group_owner_class, group_owner=None
    ):
        """
        Add missing objects linked to computers when tags are added to
        computers. This method is used by objects having a ComputerGroup to
        manage tags.

        If C{group_owner} is specified, only this object is considered
        instead of all the objects of the account matching the new tags.
        """
        store = self.account.store
        if group_owner is None:
            result = store.find(
                group_owner_class,
                ComputerGroupTag.tag_id.is_in(tag_ids),
                (
                    ComputerGroupTag.computer_group_id
                    == group_owner_class.computer_group_id
                ),
                ComputerGroup.id == group_owner_class.computer_group_id,
                ComputerGroup.all_computers == False,
                group_owner_class.account_id == self.account.id,
            )
            result.config(distinct=True)
            index = dict([(rule.id, rule) for rule in result])
        else:
            if group_owner.all_computers:
                # If all_computers is True, no need to add any objects
                return
            if group_owner.account_id != self.account.id:
                # Sanity check
                return
            index = {group_owner.id: group_owner}

        if index:
            # All (group_owner_id, computer_id) pairings where computer_id
            # is in computer_ids, group_owner_id is in index and computer_id
            # belongs to the ComputerGroup of group_owner_id.
            result = store.find(
                (group_owner_class.id, ComputerGroupItem.computer_id),
                group_owner_class.computer_group_id
                == ComputerGroupItem.computer_group_id,
                group_owner_class.id.is_in(index),
                ComputerGroupItem.computer_id.is_in(computer_ids),
            )

            result.config(distinct=True)
            state_index = set(result)
            for group_owner_id, group_owner in list(index.items()):
                # Add to group_owner all computers associated to the given
                # tag_ids that are not associated to it yet.
                missing_computers_ids = [
                    computer_id
                    for computer_id in computer_ids
                    if not (group_owner_id, computer_id) in state_index
                ]
                # The group_owner's add_computers_with_ids method is supposed
                # to eventually call ComputerGroup.add_computers_with_ids.
                group_owner.add_computers_with_ids(missing_computers_ids)

    @requires("SynchronizeTags")
    def synchronize_with_removed_tags(
        self, tag_ids, computer_ids, group_owner_class, group_owner=None
    ):
        """
        Remove objects linked to computers when tags are removed from
        computers.

        If C{group_owner} is specified, only this object is considered
        instead of all the objects of the account matching the removed tags.
        """
        from canonical.landscape.model.main.tag import ComputerTag

        removed_tag_ids = set(tag_ids)

        store = self.account.store
        if group_owner is None:
            result = store.find(
                group_owner_class,
                group_owner_class.account_id == self.account.id,
                ComputerGroup.id == group_owner_class.computer_group_id,
                ComputerGroup.all_computers == False,
                ComputerGroupTag.computer_group_id
                == group_owner_class.computer_group_id,
                ComputerGroupTag.tag_id.is_in(removed_tag_ids),
            )
            result.config(distinct=True)
            index = dict(
                [(group_owner_item.id, group_owner_item) for group_owner_item in result]
            )
        else:
            if group_owner.all_computers:
                # If all_computers is True, no need to remove any objects
                return
            if group_owner.account_id != self.account.id:
                # Sanity check
                return
            index = {group_owner.id: group_owner}

        tag_index = {}
        all_tag_ids = set()
        result = store.find(
            (group_owner_class.id, ComputerGroupTag.tag_id),
            ComputerGroupTag.computer_group_id == group_owner_class.computer_group_id,
            group_owner_class.id.is_in(index),
        )
        for group_owner_id, tag_id in result:
            if group_owner_id not in tag_index:
                tag_index[group_owner_id] = set([tag_id])
            else:
                tag_index[group_owner_id].add(tag_id)
            all_tag_ids.add(tag_id)

        # Build an index of computers to tags.
        result = Store.of(self.account).find(
            ComputerTag,
            ComputerTag.tag_id.is_in(all_tag_ids),
            ComputerTag.computer_id.is_in(computer_ids),
        )

        computer_tag_index = {}
        for computer_id, tag_id in result.values(
            ComputerTag.computer_id, ComputerTag.tag_id
        ):
            if computer_id not in computer_tag_index:
                computer_tag_index[computer_id] = set([tag_id])
            else:
                computer_tag_index[computer_id].add(tag_id)

        for group_owner_id, tag_ids in list(tag_index.items()):
            removable_computer_ids = set()
            remaining_tag_ids = tag_ids - removed_tag_ids
            for computer_id, computer_tag_ids in list(computer_tag_index.items()):
                if not computer_tag_ids.intersection(remaining_tag_ids):
                    removable_computer_ids.add(computer_id)

            index[group_owner_id].remove_computers_with_ids(removable_computer_ids)

    @requires("ViewAccount")
    def get_owners_by_tags_and_all_computers(self, group_owners, tags):
        """
        Return a mapping between L{ComputerGroup} owner classes and a list
        of instances of that class that have the group enabled for any of
        the specified tags or for all computers.
        """
        tag_ids = get_main_store().find(Tag.id, Tag.name.is_in(tags))

        group_owners_by_class = {}
        for group_owner in group_owners:
            owners_with_all_computers = self.account.store.find(
                group_owner,
                group_owner.account_id == self.account.id,
                ComputerGroup.id == group_owner.computer_group_id,
                ComputerGroup.all_computers == True,
            )
            owners_with_tags = self.account.store.find(
                group_owner,
                group_owner.account_id == self.account.id,
                ComputerGroup.id == group_owner.computer_group_id,
                (ComputerGroupTag.computer_group_id == group_owner.computer_group_id),
                ComputerGroupTag.tag_id.is_in(tag_ids),
                ComputerGroup.all_computers == False,
            )
            result = owners_with_all_computers.union(owners_with_tags)
            group_owners_by_class[group_owner] = list(result)

        return group_owners_by_class
