# Copyright 2005-2022 Canonical Limited. All rights reserved.

from xml.sax.saxutils import escape

from zope.interface import implementer

from storm.locals import Int, Unicode, Reference, AutoReload, RawStr

from landscape.lib.scriptcontent import build_script, generate_script_hash

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.account.interfaces import ICustomGraph
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)
from canonical.landscape.model.script.script import split_code_on_interpreter
from canonical.landscape.model.account.computergroup import (
    ComputerGroup,
    ComputerGroupItem,
)
from canonical.landscape.model.account.helper import ComputerGroupHelper
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)

__all__ = ["CustomGraph", "CustomGraphManagement", "DuplicateCustomGraphError"]

__clearlist__ = ["CustomGraphManagement", "DuplicateCustomGraphError"]

__requires_prefix__ = "landscape."

CUSTOM_GRAPH_INTRODUCED_VERSION = "3.2"

CUSTOM_GRAPH_DISABLED_MESSAGE = """
Script Execution plugin is disabled on this client, please see
<a href="http://help.landscape.canonical.com/RunningScripts">our help site</a>
 for advice on enabling this plugin.
"""


class DuplicateCustomGraphError(DuplicateModelError):
    """
    Raised when a L{CustomGraph} with the same name exists in the account.
    """

    model_name = "Custom graph"


class UnknownCustomGraphError(Exception):
    """
    Raised when a L{CustomGraph} can not be found.
    """

    def __init__(self, graph_id):
        super(UnknownCustomGraphError, self).__init__(
            "graph_id '%d' does not exist." % graph_id
        )


@implementer(ICustomGraph)
@requires_advisor
class CustomGraph(ComputerGroupHelper):

    __storm_table__ = "custom_graph"

    requires(
        "ViewCustomGraph",
        PRIVATE,
        [
            "id",
            "script_hash",
            "all_computers",
            "get_computers",
            "get_computer_count",
            "get_computer_ids",
            "get_tags",
            "access_context",
        ],
    )
    requires(
        "ViewCustomGraph",
        "EditCustomGraph",
        [
            "title",
            "y_axis_title",
            "account",
            "interpreter",
            "code",
            "username",
            "computer_group_id",
        ],
    )
    requires(
        "EditCustomGraph",
        PRIVATE,
        [
            "update_association",
            "update_tags",
            "enable_for_all_computers",
            "disable_for_all_computers",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)

    title = Unicode(allow_none=False)

    y_axis_title = Unicode(allow_none=False)

    _interpreter = Unicode("interpreter", allow_none=False)
    _code = Unicode("code", allow_none=False)
    username = Unicode(allow_none=False)
    script_hash = RawStr(allow_none=False)
    access_context_id = Int(allow_none=False)

    computer_group_id = Int()
    computer_group = Reference(computer_group_id, ComputerGroup.id)

    def __init__(
        self, account, title, y_axis_title, username, code, access_context_id=None
    ):
        super(CustomGraph, self).__init__(account, access_context_id)
        self.title = title
        self.y_axis_title = y_axis_title
        self.username = username
        self.set_code(code)
        self._create_computer_group()

    def _set_code(self, newcode):
        self._code = newcode
        self._update_script_hash()

    def _get_code(self):
        return self._code

    code = property(_get_code, _set_code)

    def _set_interpreter(self, newinterpreter):
        self._interpreter = newinterpreter
        self._update_script_hash()

    def _get_interpreter(self):
        return self._interpreter

    interpreter = property(_get_interpreter, _set_interpreter)

    def _update_script_hash(self):
        hash = generate_script_hash(build_script(self.interpreter, self.code))
        self.script_hash = hash

    @requires("EditCustomGraph")
    def set_code(self, code):
        self.interpreter, self.code = split_code_on_interpreter(code)

    @requires("ViewCustomGraph")
    def get_code(self):
        return "#!" + self.interpreter + "\n" + self.code

    def add_computers_with_ids(self, computer_ids):
        """
        This takes a list of computer IDs and adds them to the custom graph.
        """
        computer_ids = super(CustomGraph, self).add_computers_with_ids(computer_ids)

        if computer_ids:
            from canonical.landscape.model.resource.customgraph import CustomGraphData

            result = self.account.store.find(
                AccountComputerTwin, AccountComputerTwin.id.is_in(computer_ids)
            )
            for twin in result:
                twin.computer.store.add(CustomGraphData(twin.computer, self))

        return computer_ids

    def remove_computers_with_ids(self, computer_ids):
        """
        This takes a list of computer IDs and removes them from the custom
        graph.
        """
        computer_ids = super(CustomGraph, self).remove_computers_with_ids(computer_ids)

        if computer_ids:
            from canonical.landscape.model.resource.customgraph import CustomGraphData

            result = self.account.store.find(
                AccountComputerTwin, AccountComputerTwin.id.is_in(computer_ids)
            )
            for twin in result:
                data_result = twin.computer.store.find(
                    CustomGraphData,
                    CustomGraphData.custom_graph_id == self.id,
                    CustomGraphData.computer_id == twin.id,
                )
                data_result.remove()

        return computer_ids

    @requires("EditCustomGraph")
    def remove(self):
        from canonical.landscape.model.resource.customgraph import CustomGraphData

        computer_ids = [twin.computer.id for twin in self.computer_group.computers]
        for store in get_resource_stores_for_computer_ids(computer_ids):
            store.find(
                CustomGraphData, CustomGraphData.custom_graph_id == self.id
            ).remove()
        super(CustomGraph, self).remove()

    def _get_computers_with_custom_graph_message(self, computer_ids):
        """
        Returns a result set of computers which support custom graphs,
        selecting from the set of computers with ids in computer_ids.
        """
        from canonical.landscape.model.main.computer import (
            Computer,
            MessageType,
            ComputerMessageType,
        )

        store = get_main_store()
        return store.find(
            Computer,
            Computer.id.is_in(computer_ids),
            ComputerMessageType.computer_id == Computer.id,
            MessageType.id == ComputerMessageType.message_type_id,
            MessageType.name == "custom-graph-add",
        )

    @requires("ViewCustomGraph")
    def has_errors(self, selected_ids=None):
        """
        Returns true if a custom graph has any error messages on the associated
        computers.

        @param selected_ids: An optional sequence of L{Computer} ids which
            will be used to restrict the erroring computer ids.
        """
        if selected_ids:
            computers = [
                twin.computer
                for twin in self.computer_group.computers
                if twin.computer.id in selected_ids
            ]
        else:
            computers = [twin.computer for twin in self.computer_group.computers]

        computer_ids = [computer.id for computer in computers]
        for computer in computers:
            if not computer.is_api_compatible(CUSTOM_GRAPH_INTRODUCED_VERSION):
                return True

        result = self._get_computers_with_custom_graph_message(computer_ids)
        if result.count() < len(computer_ids):
            return True

        from canonical.landscape.model.resource.customgraph import CustomGraphData

        stores = get_resource_stores_for_computer_ids(computer_ids)
        for store, computer_ids in list(stores.items()):
            result = store.find(
                CustomGraphData,
                CustomGraphData.custom_graph_id == self.id,
                CustomGraphData.error != "",
                CustomGraphData.computer_id.is_in(computer_ids),
            )
            errors = result.count()
            if errors:
                return True
        return False

    @requires("ViewCustomGraph")
    def get_error_messages(self, person, selected_ids=None):
        """
        Return error messages associated with computers for this custom
        graph.

        @param selected_ids: An optional sequence of L{Computer} ids which
            will be used to restrict the erroring computer ids.
        """
        from canonical.landscape.model.resource.customgraph import CustomGraphData
        from canonical.landscape.model.main.computer import Computer

        access_context_ids = get_access_context_ids_for_person(
            self.account.access_context, "ViewCustomGraph", person
        )
        if selected_ids:
            computers = [
                twin.computer
                for twin in self.computer_group.computers.find(
                    AccountComputerTwin.id.is_in(selected_ids)
                )
            ]
        else:
            computers = [twin.computer for twin in self.computer_group.computers]
        computers = [
            computer
            for computer in computers
            if computer.access_context_id in access_context_ids
        ]
        computer_ids = [computer.id for computer in computers]
        errors = {}
        for computer in computers:
            if not computer.is_api_compatible(CUSTOM_GRAPH_INTRODUCED_VERSION):
                errors[computer] = (
                    "Client version incompatible with custom "
                    "graph feature, please upgrade."
                )

        result = self._get_computers_with_custom_graph_message(computer_ids)
        computer_ids_with_custom_graphs = list(result.values(Computer.id))
        for computer in computers:
            if computer.id not in computer_ids_with_custom_graphs:
                if computer not in errors:
                    errors[computer] = CUSTOM_GRAPH_DISABLED_MESSAGE
        stores = get_resource_stores_for_computer_ids(computer_ids)
        for store, computer_ids in list(stores.items()):
            result = store.find(
                CustomGraphData,
                CustomGraphData.custom_graph_id == self.id,
                CustomGraphData.error != "",
                CustomGraphData.computer_id.is_in(computer_ids),
            )
            for graph_data in result:
                computer = graph_data.computer
                if computer not in errors:
                    errors[computer] = escape(graph_data.error)
        return errors


@requires_advisor
class CustomGraphManagement(ModelManagement):

    requires("ViewAccount", PRIVATE, ["account", "access_context", "person"])

    @requires("ViewAccount")
    def create_graph(
        self,
        title,
        y_axis_title,
        username,
        code,
        access_context_name=DEFAULT_ACCESS_CONTEXT_NAME,
    ):
        """
        Create a custom graph in the account.
        """
        access_context = self.get_access_context(
            access_context_name, "CreateCustomGraph"
        )

        self.check_duplicate(CustomGraph, DuplicateCustomGraphError, title=title)

        graph = CustomGraph(
            self.account,
            title,
            y_axis_title,
            username,
            code,
            access_context_id=access_context.id,
        )
        self.account.store.add(graph)
        return graph

    @requires("ViewAccount")
    def remove_graphs(self, graphs):
        """
        Remove the specified custom graphs from the account.

        @returns: L{CustomGraph}s that didn't get removed because the user
        doesn't have permission on them.
        """
        access_context_ids = self.get_access_context_ids("EditCustomGraph")

        remaining_graphs = []
        for graph in graphs:
            if graph.access_context.id in access_context_ids:
                graph.remove()
            else:
                remaining_graphs.append(graph)
        return remaining_graphs

    @requires("ViewAccount")
    def get_graphs(
        self, ids=None, titles=None, computer_ids=None, permission="ViewCustomGraph"
    ):
        """
        Return LCustomGraph}s associated with this account, by the specified
        criteria.

        @param ids: If provided, only return graphs with specified ids.
        @param titles: If provided, only return graphs with specified titles.
        @param computer_ids: If provided, limit result to graphs applied to
            L{Computer}s by specified Ids.
        @param permission: The permission the L{Person} must have on the
            L{CustomGraph} access context. If C{None}, permissions won't be
            checked.
        """
        where = [
            CustomGraph.account_id == self.account.id,
            CustomGraph.computer_group_id == ComputerGroup.id,
        ]

        if permission is not None:
            access_context_ids = self.get_access_context_ids(permission)
            where.append(CustomGraph.access_context_id.is_in(access_context_ids))
        if ids is not None:
            where.append(CustomGraph.id.is_in(ids))
        if titles is not None:
            where.append(CustomGraph.title.is_in(titles))
        if computer_ids is not None:
            where.extend(
                [
                    (
                        CustomGraph.computer_group_id
                        == ComputerGroupItem.computer_group_id
                    ),
                    ComputerGroupItem.computer_id.is_in(computer_ids),
                ]
            )

        result = self.account.store.find(CustomGraph, *where).config(distinct=True)
        return result.order_by(CustomGraph.title)

    @requires("ViewAccount")
    def get_graph(self, id):
        """
        Return the L{CustomGraph} with the matching id.

        @raises UnknownGraphError: If no L{CustomGraph} with the given id is
            found.
        """
        custom_graph = self.get_graphs(ids=[id]).one()
        if custom_graph is None:
            raise UnknownCustomGraphError(id)
        return custom_graph

    @requires("ViewAccount")
    def edit_graph(self, custom_graph_id, data):
        """
        Edit a stored L{CustomGraph}.

        @returns: True if the L{CustomGraph} was modified, False otherwise.
        """
        db_graph = self.get_graph(custom_graph_id)
        changed = False

        title = data.get("title", None)
        if title is not None and title != db_graph.title:
            self.check_duplicate(CustomGraph, DuplicateCustomGraphError, title=title)
            db_graph.title = title
            changed = True

        code = data.get("code", None)
        existing_code = db_graph.get_code()
        if code is not None and code != existing_code:
            db_graph.set_code(code)
            changed = True

        username = data.get("username", None)
        if username is not None and username != db_graph.username:
            db_graph.username = username
            changed = True

        y_axis_title = data.get("y_axis_title", None)
        if y_axis_title is not None and y_axis_title != db_graph.y_axis_title:
            db_graph.y_axis_title = y_axis_title
            changed = True

        return changed
