from builtins import object
from zope.component import getUtility

from canonical.storm.relation import RelationLink, Relation, RelationResolver
from canonical.security.interfaces import IRoleRegistry
from canonical.security.role import UnknownRoleError
from canonical.lib.credentials import create_access_key, create_secret_key

from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.accesscontext import AccessContext
from canonical.landscape.model.main.account import (
    Account,
    PersonAccount,
    get_account_by_name,
    DEFAULT_ACCESS_CONTEXT_NAME,
    DEFAULT_ACCESS_CONTEXT_TITLE,
    DEFAULT_ACCOUNT_MAX_PEOPLE_COUNT,
)
from canonical.landscape.model.account.store import get_account_store_by_name
from canonical.landscape.model.account.twin import AccountTwin
from canonical.landscape.model.access.role import (
    StoredRole,
    StoredRolePermission,
    StoredRoleAccessContext,
    StoredRolePerson,
)
from canonical.landscape.model.access.stockroles import (
    STOCK_ROLES,
    STOCK_ROLE_DESCRIPTION,
)
from canonical.landscape.model.main.person import Person, APICredentials
from canonical.landscape.model.main.personaccess import PersonAccess
from canonical.landscape.model.main.license import License
from canonical.landscape.security.roles import STORED_ROLE_KEY_MIN

DEFAULT_MAX_PEOPLE = 5
DEFAULT_SEATS = 10

ROOT_ACCESS_CONTEXT_COLUMNS = (
    AccessContext.name,
    AccessContext.description,
)

CHILD_ACCESS_CONTEXT_COLUMNS = (
    AccessContext.parent_id,
    AccessContext.root_id,
    AccessContext.name,
    AccessContext.description,
)

ACCOUNT_COLUMNS = (
    Account.name,
    Account.title,
    Account.registration_password,
    Account.max_people_count,
    Account.access_context_id,
    Account.stripe_customer_id,
    Account.vat_number,
    Account.vat_country,
    Account.marketing_emails,
)

ACCOUNT_TWIN_COLUMNS = (AccountTwin.id,)

STORED_ROLE_COLUMNS = (
    StoredRole.account_id,
    StoredRole.name,
    StoredRole.description,
)

STORED_ROLE_PERMISSION_COLUMNS = (
    StoredRolePermission.role_id,
    StoredRolePermission.permission,
)

STORED_ROLE_ACCESS_CONTEXT_COLUMNS = (
    StoredRoleAccessContext.role_id,
    StoredRoleAccessContext.access_context_id,
)

STORED_ROLE_PERSON_COLUMNS = (
    StoredRolePerson.role_id,
    StoredRolePerson.person_id,
)

LICENSE_COLUMNS = (
    License.account_id,
    License.role_key,
    License.seats,
)

PERSON_ACCOUNT_COLUMNS = (
    PersonAccount.person_id,
    PersonAccount.account_id,
)

API_CREDENTIALS_COLUMNS = (
    APICredentials.access_key_id,
    APICredentials.access_secret_key,
    APICredentials.person_id,
    APICredentials.account_id,
)


PERSON_COLUMNS = (
    Person.name,
    Person.email,
    Person.password,
    Person.identity,
)

PERSON_ACCESS_COLUMNS = (
    PersonAccess.person_id,
    PersonAccess.role_key,
    PersonAccess.access_context_id,
)


class AccountProfile(object):
    """Create one or more accounts, their admins and their roles.

    This class is a helper to create a set of persons and accounts from a given
    profile specifications expressed with dictionaries.

    Example:

    profile = AccountProfile(
        persons={
            "joe": {"name": u"Joe", "email": u"joe@foo.com", "password": u"x"},
        },
        accounts={
            "datatech": {
                "name": u"datatech",
                "title": u"DataTech Ltd.",
                "persons": [{"key": "joe", "roles": ["Owner"]}],
            },
            })
    profile.create()

    will create an account with a single admin. More complex setups are
    possible by passing additional info about access groups and stored roles.

    See unit tests and c/l/sample/profiles.py for more comprehensive examples.
    """

    def __init__(self, persons=None, accounts=None, resolver=None):
        """
        @param persons: A dict describing the person to create. Expected keys
            are "name", "email", and optionally a "password" or "identify"
            entry.
        @param account: A dict specifying the accounts to create, with keys
            being "name", "title", and optionals "max-people" and
            "registration-key".
        @param resolver: The RelationResolver instance to use for managing
            relations. If None, a fresh one will be created.
        """
        self.persons = persons or {}
        self.accounts = accounts or {}
        self.resolver = resolver or RelationResolver()
        self._factory = AccountFactory(self.resolver)

    def create(self):
        """Create the accounts and persons of this profile."""
        self._create_accounts()
        self._create_persons()
        self._create_memberships()
        self._factory.flush()
        self._create_options()

    def _create_accounts(self):
        """Create the accounts defined in the profile."""
        for key, info in self.accounts.items():

            # Create the bare account
            self._factory.create_account(
                info["name"],
                info["title"],
                max_people=info.get("max-people", DEFAULT_MAX_PEOPLE),
                stripe_customer_id=info.get("stripe-customer-id"),
                vat_number=info.get("vat-number"),
                vat_country=info.get("vat-country"),
                registration_key=info.get("registration-key"),
            )

            # Add the license seats
            self._factory.create_license(info["name"], info.get("seats", DEFAULT_SEATS))

            # Add custom access groups
            for access_group in info.get("access-groups", ()):
                self._factory.create_access_group(
                    info["name"], access_group["name"], access_group["description"]
                )

            # Add custom stored roles
            for role in info.get("roles", ()):
                self._factory.create_role(
                    info["name"],
                    role["name"],
                    role["description"],
                    role["access-groups"],
                    role["permissions"],
                )

    def _create_options(self):
        """Create optional account-related entities."""
        for key, info in self.accounts.items():

            # Optionally create stock rules. The flag is true by default, but
            # some tests might want to disable it to avoid triggering rule
            # checks.
            if info.get("stock-rules", True):
                account = get_account_by_name(info["name"])
                account.create_stock_rules()

    def _create_persons(self):
        """Create the persons defined in the profile."""
        for key, info in self.persons.items():
            self._factory.create_person(
                info["name"],
                info["email"],
                password=info.get("password"),
                identity=info.get("identity"),
            )

    def _create_memberships(self):
        """Add persons to accounts, with the desired roles."""
        for key, info in self.accounts.items():
            for link in info["persons"]:
                account_name = info["name"]
                person_email = self.persons[link["key"]]["email"]

                # access_contexts is empty for stock roles as they relate to
                # the root access context. Stored roles will potentially have
                # access contexts to which they relate.
                role_access_contexts = {role: () for role in link["roles"]}
                # Grab stored role access groups
                for role in info.get("roles", ()):
                    name = role["name"]
                    if name not in role_access_contexts:
                        continue  # Skip stock role
                    role_access_contexts[name] = role["access-groups"]
                self._factory.create_membership(
                    account_name, person_email, role_access_contexts
                )


class AccountFactory(object):
    """Helper to create accounts and related objects."""

    def __init__(self, resolver):
        """
        @param resolver: The RelationResolve instance to use to manage
            relations.
        """
        self._resolver = resolver
        self._root_access_context = Relation(ROOT_ACCESS_CONTEXT_COLUMNS)
        self._account = Relation(ACCOUNT_COLUMNS, index=Account.name)
        self._account_twin = Relation(ACCOUNT_TWIN_COLUMNS)
        self._stored_role = Relation(
            STORED_ROLE_COLUMNS, index=(self._account, StoredRole.name)
        )
        self._child_access_context = Relation(
            CHILD_ACCESS_CONTEXT_COLUMNS, index=(self._account, AccessContext.name)
        )
        self._license = Relation(LICENSE_COLUMNS)
        self._stored_role_permission = Relation(STORED_ROLE_PERMISSION_COLUMNS)
        self._stored_role_access_context = Relation(STORED_ROLE_ACCESS_CONTEXT_COLUMNS)
        self._stored_role_person = Relation(STORED_ROLE_PERSON_COLUMNS)
        self._person = Relation(PERSON_COLUMNS, index=Person.email)
        self._person_access = Relation(PERSON_ACCESS_COLUMNS)
        self._person_account = Relation(PERSON_ACCOUNT_COLUMNS)
        self._api_credentials = Relation(API_CREDENTIALS_COLUMNS)

        # Cache for the root URL to avoid unnecessary queries
        self._root_url = None

    def create_account(
        self,
        name,
        title,
        registration_key=None,
        max_people=DEFAULT_ACCOUNT_MAX_PEOPLE_COUNT,
        stripe_customer_id=None,
        vat_number=None,
        vat_country=None,
        marketing_emails=False,
    ):
        """Create an Account with the given parameters."""
        self._root_access_context.add(
            name=DEFAULT_ACCESS_CONTEXT_NAME, description=DEFAULT_ACCESS_CONTEXT_TITLE
        )
        self._account.add(
            name=name,
            title=title,
            registration_password=registration_key,
            max_people_count=max_people,
            stripe_customer_id=stripe_customer_id,
            vat_number=vat_number,
            vat_country=vat_country,
            marketing_emails=marketing_emails,
            access_context_id=RelationLink(self._root_access_context),
        )
        self._account_twin.add(id=RelationLink(self._account))

        # Add stock stored roles
        access_contexts = [DEFAULT_ACCESS_CONTEXT_NAME]
        for role_name, permissions in STOCK_ROLES.items():
            description = STOCK_ROLE_DESCRIPTION.format(role_name)
            self.create_role(name, role_name, description, access_contexts, permissions)

    def create_person(self, name, email, password=None, identity=None, global_roles=()):
        """Create a Person with the given parameters."""
        if identity:
            if self._root_url is None:
                # Cache the root-url, since it doesn't change and we don't
                # want to hit the database unnecessarily
                self._root_url = get_root_url().strip("/")
            identity = identity.format(root_url=self._root_url)
        elif not password:
            # If no password or identity is provided, a default password
            # is set. Mainly useful for tests.
            password = "pwd"

        if password:
            algo = "sha512"
            password = "{}:{}".format(
                algo, Person._encrypt_password(password, algorithm=algo)
            )

        self._person.add(name=name, email=email, password=password, identity=identity)

        for global_role in global_roles:
            self._create_person_access(email, global_role, None)

    def create_license(self, account, seats, kind="BasicFeatures"):
        """Create a license and add seats to it."""
        registry = getUtility(IRoleRegistry)
        self._license.add(
            account_id=RelationLink(self._account, key=account),
            role_key=registry.get(kind).key,
            seats=seats,
        )

    def create_access_group(self, account, name, description):
        """Create an access group in the given account."""
        root_access_context_id_link = RelationLink(
            self._account, key=account, column=Account.access_context_id
        )
        self._child_access_context.add(
            parent_id=root_access_context_id_link,
            root_id=root_access_context_id_link,
            name=name,
            description=description,
        )

    def create_role(self, account, name, description, access_contexts, permissions):
        """Create a stored role.

        @param account: The name of the account to create the role in.
        @param name: The name of the role to create.
        @param description: The role's description.
        @param access_contexts: List of AccessContext names that the role
            will be associated with.
        @param permissions: List of the permission that the role will grant.
        """
        # Add the bare role
        self._stored_role.add(
            account_id=RelationLink(self._account, key=account),
            name=name,
            description=description,
        )

        role_id_link = RelationLink(self._stored_role, key=(account, name))

        # Add the role's access contexts
        root_access_context_id_link = RelationLink(
            self._account, key=account, column=Account.access_context_id
        )
        for access_context in access_contexts:
            if access_context == DEFAULT_ACCESS_CONTEXT_NAME:
                # If the access context is the root one, we get its ID by
                # referencing Account.access_context_id
                access_context_id_link = root_access_context_id_link
            else:
                access_context_id_link = RelationLink(
                    self._child_access_context, key=(account, access_context)
                )
            self._stored_role_access_context.add(
                role_id=role_id_link, access_context_id=access_context_id_link
            )

        # Add the role's permissions
        for permission in permissions:
            self._stored_role_permission.add(
                role_id=role_id_link, permission=permission
            )

    def create_membership(self, account, person, role_access_contexts):
        """Add a person to an account, with given roles and access_contexts.

        @param account: The name of the account to add the membership to.
        @param person: The email of the person the membership is for.
        @param role_access_contexts: Dictionary of role names and
            access context names for the roles the person should be granted.
        """
        root_access_context_id_link = RelationLink(
            self._account, key=account, column=Account.access_context_id
        )
        registry = getUtility(IRoleRegistry)
        for role, access_contexts in role_access_contexts.items():
            # Check if it's a stored role or a stock one
            try:
                registry.get(role)
            except UnknownRoleError:
                # This is a stored role.
                self._create_person_access(
                    person, "AccountMember", root_access_context_id_link
                )
                self._create_stored_role_person(account, role, person)
                self._create_stored_role_person_access(
                    account, person, role, access_contexts
                )
            else:
                self._create_person_access(person, role, root_access_context_id_link)
        person_id_link = RelationLink(self._person, key=person)
        account_id_link = RelationLink(self._account, key=account)
        self._person_account.add(person_id=person_id_link, account_id=account_id_link)
        self._api_credentials.add(
            access_key_id=create_access_key(),
            access_secret_key=create_secret_key(),
            person_id=person_id_link,
            account_id=account_id_link,
        )

    def flush(self):
        """Flush all the data create with this factory to the database."""
        main_store = get_main_store()
        # At the moment we support only one account database
        account_store = get_account_store_by_name("account-1")

        self._resolver.add(self._root_access_context, main_store)
        self._resolver.add(self._account, main_store)
        self._resolver.add(self._account_twin, account_store)
        self._resolver.add(self._child_access_context, main_store)
        self._resolver.add(self._license, main_store)
        self._resolver.add(self._stored_role, main_store)
        self._resolver.add(self._stored_role_permission, main_store)
        self._resolver.add(self._stored_role_access_context, main_store)
        self._resolver.add(self._person, main_store)
        self._resolver.add(self._person_access, main_store)
        self._resolver.add(self._person_account, main_store)
        self._resolver.add(self._api_credentials, main_store)
        self._resolver.add(self._stored_role_person, main_store)

        self._resolver.flush()

    def _create_person_access(self, email, role, access_context):
        """Create a new row in the person_access table."""
        registry = getUtility(IRoleRegistry)
        self._person_access.add(
            person_id=RelationLink(self._person, key=email),
            role_key=registry.get(role).key,
            access_context_id=access_context,
        )

    def _create_stored_role_person_access(self, account, email, role, access_contexts):
        """Create new stored role rows in the person_access table.

        One row will be created for each access_context of the role.
        """

        def stored_role_offset(value):
            """
            StoredRoles are offset by STORED_ROLE_KEY_MIN for
            person_access.role_key.
            """
            return value + STORED_ROLE_KEY_MIN

        # Add a person_access row for each access_context a role applies to
        for access_context in access_contexts:
            access_context_id_link = RelationLink(
                self._child_access_context, key=(account, access_context)
            )
            self._person_access.add(
                person_id=RelationLink(self._person, key=email),
                role_key=RelationLink(
                    self._stored_role,
                    key=(account, role),
                    value_callback=stored_role_offset,
                ),
                access_context_id=access_context_id_link,
            )

    def _create_stored_role_person(self, account, role, email):
        """Create a new row in the role_person table.

        @param account: Name of the account the role belongs to.
        @param role: Name of the stored role to link.
        @param email: Email of the person to link.
        """
        self._stored_role_person.add(
            role_id=RelationLink(self._stored_role, key=(account, role)),
            person_id=RelationLink(self._person, key=email),
        )
