# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from typing import Iterable

from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy
from zope.interface import implementer

from storm.locals import Storm, Delete
from storm.databases.postgres import Returning
from storm.expr import And, In, Not, SQL

from canonical.security.interfaces import IRoleRegistry

from canonical.landscape.model.main.interfaces import IAccountAware, IAccessContextAware
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.tag import get_tags
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.lib.text import describe_list
from canonical.landscape.model.eventlog.management import EventLog


__all__ = ["AccountHelper", "ComputerGroupHelper"]


@implementer(IAccountAware, IAccessContextAware)
class AccountHelper(Storm):
    """Helper for implementing L{IAccountAware} classes."""

    _account = None
    account_id = None

    def __init__(self, account=None):
        if account:
            account = removeSecurityProxy(account)
            self.account = account

            # Emulate the behavior of a "normal" reference to the account
            # (the object would be automatically added).
            if hasattr(self.__class__, "__storm_table__"):
                account.store.add(self)

    def _get_account(self):
        store = get_main_store()
        account_id = self.account_id
        if (
            account_id is not None and not self._account
        ) or self._account.id != account_id:
            from canonical.landscape.model.main.account import Account

            self._account = store.get(Account, account_id)
        return self._account

    def _set_account(self, account):
        self.account_id = account.id
        self._account = account

    account = property(_get_account, _set_account)

    @property
    def access_context(self):
        return self.account.access_context


class ComputerGroupHelper(AccountHelper):
    """
    Base class for model classes that reference a L{ComputerGroup}.
    """

    def __init__(self, account=None, access_context_id=None):
        super(ComputerGroupHelper, self).__init__(account=account)
        self.access_context_id = access_context_id or self.account.access_context.id

    @property
    def access_context(self):
        return self.account.get_access_context(self.access_context_id)

    @property
    def access_children(self):
        return self.get_computers()

    @property
    def access_cutoff_role(self):
        return getUtility(IRoleRegistry).get("AdminViewer")

    def get_computers(self, computer_ids=None):
        """Get the {Computer} objects that are associated with this profile.

        @param computer_ids: Optionally a list of L{Computer.id} to filter by,
            if C{None} all computers associated with the profile will be
            returned.
        @return: A C{ResultSet} yielding L{Computer} objects associated with
            this profile.
        """
        associated_computer_ids = self.get_computer_ids(computer_ids=computer_ids)

        return self.account.get_computers(computer_ids=associated_computer_ids)

    def get_computer_count(self):
        """Return the number of computers with this profile applied."""
        return self.get_computers().count()

    def get_computer_ids(self, computer_ids=None) -> Iterable[int]:
        """Generator yields L{Computer.id}s this profile is associated with."""
        result = self.computer_group.computers
        if computer_ids is not None:
            result = result.find(AccountComputerTwin.id.is_in(computer_ids))

        return result.values(AccountComputerTwin.id)

    def get_child_computer_ids(self, parent_computer_ids) -> Iterable[int]:
        """
        Returns a list of L{Computer.id}s of computers that are children of the
        given ids.
        """
        # SQL query fails if list is empty
        if not parent_computer_ids:
            return []

        from canonical.landscape.model.main.computer import (
            Computer,
            ComputerRelationship,
        )

        # The PostgreSQL ltree type has implicit cast from string but not int
        str_ids = ",".join([f"'{computer_id}'" for computer_id in parent_computer_ids])
        where = And(
            SQL(f"ARRAY[{str_ids}]::LTREE[] @> tree"),
            Not(In(ComputerRelationship.computer_id, parent_computer_ids)),
        )

        store = get_main_store()
        child_ids = store.find(ComputerRelationship.computer_id, where)
        result = self.account.get_computers(
            computer_ids=child_ids,
            access_context_ids=self.access_context.get_all_access_context_ids(),
        )

        return result.values(Computer.id)

    def update_association(
        self, new_tags=None, all_computers=None, changes=None, person=None
    ):
        """
        Consolidate tag and all computer association handling to simplify
        L{Tag} and C{all_computers} association and L{EventLogEntry} tracking
        for L{ComputerGroupHelper} subclasses.
        """
        if changes is None:
            changes = {}

        messages = []
        if all_computers is not None:
            if all_computers != self.computer_group.all_computers:
                if all_computers:
                    self.enable_for_all_computers()
                    changes["associate_all_computers"] = True
                    messages.append("Associated to all computers.")
                else:
                    self.disable_for_all_computers()
                    changes["disassociate_all_computers"] = True
                    messages.append("Disassociated from all computers.")
        if new_tags is not None:
            (tags_added, tags_removed) = self.update_tags(new_tags)
            changes["tags_added"] = [tag.name for tag in tags_added]
            changes["tags_removed"] = [tag.name for tag in tags_removed]
            if changes["tags_added"] or changes["tags_removed"]:
                added_tags = describe_list(
                    "Tags added:", changes["tags_added"], pluralize=False, quote=False
                )
                if added_tags:
                    messages.append("%s." % added_tags)
                removed_tags = describe_list(
                    "Tags removed:",
                    changes["tags_removed"],
                    pluralize=False,
                    quote=False,
                )
                if removed_tags:
                    messages.append("%s." % removed_tags)
        if messages:
            event_log = EventLog(self.account, person)
            event_log.create_log_entry(self, " ".join(messages))
        return changes

    def update_tags(self, new_tags):
        """Replace the set of tags associated to this L{ComputerGroupHelper}.

        @param new_tags: The list of tags to associate.
        """
        new_tags = set(new_tags)
        current_tags = self.get_tags()
        tags_to_be_removed = sorted(
            list(current_tags.difference(new_tags)), key=lambda x: x.name
        )
        tags_to_be_added = sorted(
            list(new_tags.difference(current_tags)), key=lambda x: x.name
        )

        if tags_to_be_added:
            self.add_tags(tags_to_be_added)
        if tags_to_be_removed:
            self.remove_tags(tags_to_be_removed)

        return (tags_to_be_added, tags_to_be_removed)

    def get_tags(self):
        """Return tags associated with this profile."""
        from canonical.landscape.model.account.computergroup import ComputerGroupTag

        result = self.account.store.find(
            ComputerGroupTag.tag_id,
            ComputerGroupTag.computer_group_id == self.computer_group.id,
        )
        return get_tags(result)

    def add_tags(self, tags):
        """Add tags to the ones we're watching.

        Add a new L{ComputerGroupTag} relation for each given tag, and call
        L{AccountComputerGroup.synchronize_with_added_tags}.

        @param tags: The list of tags to be added.
        """
        from canonical.landscape.model.account.computergroup import ComputerGroupTag
        from canonical.landscape.model.main.computer import Computer

        current_tag_ids = [tag.tag_id for tag in self.computer_group.tags]
        new_tags = [tag for tag in tags if tag.id not in current_tag_ids]
        if not new_tags:
            return

        for tag in new_tags:
            self.computer_group.tags.add(ComputerGroupTag(tag))
        result = self.account.get_computers(tags=new_tags)
        computer_ids = set(result.values(Computer.id))
        new_tag_ids = [tag.id for tag in new_tags]
        self.account.computer_group.synchronize_with_added_tags(
            new_tag_ids, computer_ids, self.__class__, self
        )

    def remove_tags(self, tags):
        """Remove tags from the ones we're watching."""
        if not tags:
            return

        from canonical.landscape.model.account.computergroup import ComputerGroupTag
        from canonical.landscape.model.main.computer import Computer

        tag_ids = [tag.id for tag in tags]
        result = self.account.get_computers(tags=tags)
        computer_ids = set(result.values(Computer.id))
        self.account.computer_group.synchronize_with_removed_tags(
            tag_ids, computer_ids, self.__class__, self
        )

        result = self.account.store.find(
            ComputerGroupTag,
            ComputerGroupTag.computer_group_id == self.computer_group.id,
            ComputerGroupTag.tag_id.is_in(tag_ids),
        )
        result.remove()

    @property
    def all_computers(self):
        return self.computer_group.all_computers

    def enable_for_all_computers(self):
        """Enable the C{all_computers} flag for this profile.

        @note: This eventually ends up calling L{add_computers_with_ids}
        """
        if self.computer_group.all_computers:
            return

        from canonical.landscape.model.main.computer import Computer

        self.computer_group.all_computers = True

        result = self.account.get_computers()
        requested_computer_ids = set(result.values(Computer.id))
        existing_computer_ids = set(
            self.computer_group.computers.values(AccountComputerTwin.id)
        )
        missing_computer_ids = requested_computer_ids - existing_computer_ids
        self.add_computers_with_ids(missing_computer_ids)

    def disable_for_all_computers(self):
        """Disable the C{all_computers} flag for this profile."""
        if not self.computer_group.all_computers:
            return

        from canonical.landscape.model.account.computergroup import (
            ComputerGroupTag,
            ComputerGroupItem,
        )
        from canonical.landscape.model.main.computer import Computer

        self.computer_group.all_computers = False
        where = []
        existing_tag_ids = list(
            self.computer_group.tags.values(ComputerGroupTag.tag_id)
        )
        if existing_tag_ids:
            result = self.account.get_computers(tag_ids=existing_tag_ids)
            where.append(
                Not(ComputerGroupItem.computer_id.is_in(result.values(Computer.id)))
            )
        computer_ids = list(
            self.computer_group.computers.find(*where).values(AccountComputerTwin.id)
        )
        self.remove_computers_with_ids(computer_ids)

    def add_computers_with_ids(self, computer_ids):
        """
        Associate computers to the computer group.

        Only ids for computers in the L{AccessContext} of the computer group
        and its descendents will be added.

        @returns: a list of ids for L{Computer}s that were added.
        """
        if not computer_ids:
            return computer_ids

        from canonical.landscape.model.main.computer import Computer

        # Filter the computers which are not in the proper access contexts
        access_context_ids = self.access_context.get_all_access_context_ids()
        computers = self.account.get_computers(
            computer_ids=computer_ids, access_context_ids=access_context_ids
        )
        computer_ids = list(computers.values(Computer.id))

        result = self.account.store.find(
            AccountComputerTwin, AccountComputerTwin.id.is_in(computer_ids)
        )

        for twin in result:
            self.computer_group.add_computer(twin)
        return list(computer_ids)

    def remove_computers_with_ids(self, computer_ids):
        """Disassociate computers from the computer group."""
        if not computer_ids:
            return computer_ids

        from canonical.landscape.model.account.computergroup import ComputerGroupItem

        where = And(
            ComputerGroupItem.computer_group_id == self.computer_group.id,
            ComputerGroupItem.computer_id.is_in(computer_ids),
        )
        result = self.account.store.execute(
            Returning(Delete(where=where), ComputerGroupItem.computer_id)
        )
        return [item[0] for item in result]

    def update_access_context(self, computer_ids, access_context_id):
        """
        Synchronize computers when their access contexts changes to
        C{access_context_id}.
        """
        from canonical.landscape.model.main.computer import Computer

        access_context_ids = self.access_context.get_all_access_context_ids()
        if access_context_id not in access_context_ids:
            self.remove_computers_with_ids(computer_ids)
        else:
            if not self.computer_group.all_computers:
                # Filter the passed computers by our tags
                tags = self.get_tags()
                if not tags:
                    return
                computers = self.account.get_computers(
                    tags=tags, computer_ids=computer_ids
                )
                computer_ids = computers.values(Computer.id)
            existing_computer_ids = set(
                self.computer_group.computers.values(AccountComputerTwin.id)
            )
            missing_computer_ids = set(computer_ids) - existing_computer_ids
            self.add_computers_with_ids(missing_computer_ids)

    def remove(self):
        self.account.store.remove(self)
        self.account.store.remove(self.computer_group)

    def _create_computer_group(self):
        """Create the L{ComputerGroup} associated with this model."""
        from canonical.landscape.model.account.computergroup import ComputerGroup

        self.computer_group = ComputerGroup(self.account, self)
