# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging

from datetime import datetime

from zope.interface import implementer
from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy

from storm.locals import Storm, Unicode, Int, DateTime, Reference, Store, AutoReload

from canonical.lib.text import make_secure_id, describe_list
from canonical.lib.mail import is_email_address
from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE
from canonical.security.interfaces import IRoleRegistry
from canonical.amqp.message.interfaces import IDispatcher

from canonical.landscape.model.main.schema import InvalidEmailAddress
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.person import Person
from canonical.landscape.mail.mail import format_address
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.model.account.errors import (
    MaximumAccountAdministratorsError,
    RepeatedInvitationError,
    IllegalInvitationError,
    PersonAlreadyInAccountError,
    InvalidInvitationError,
)
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.eventlog.management import EventLog


__all__ = [
    "AccountInvitation",
    "get_invitation",
    "get_invitation_by_sfid",
    "INVITATION_EXPIRY_IN_DAYS",
    "create_invitation",
    "InvitationManagement",
    "get_invitation_by_email",
    "AccountInvitationRole",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IAccessDelegateAware)
class AccountInvitation(Storm):

    __storm_table__ = "account_invitation"

    # Note on permissions: AccountInvitations are accessed by
    # arbitrary Persons. So making this object public is OK as long as
    # there are no *factories* that return AccountInvitation objects
    # without verifying a secure ID.

    requires(
        PUBLIC,
        PRIVATE,
        [
            "secure_id",
            "account",
            "name",
            "email",
            "salesforce_key",
            "account_title",
            "creation_time",
            "namespace",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    secure_id = Unicode(allow_none=False)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    email = Unicode(allow_none=False)
    salesforce_key = Unicode()
    creation_time = DateTime(allow_none=False)

    account = Reference(account_id, Account.id)

    def __init__(self, account, name, email, salesforce_key=None):
        if not is_email_address(email):
            raise InvalidEmailAddress(email)

        self.account = account
        # Let's keep the length of the URL under 80 characters so it
        # won't wrap in email clients..
        length = 79 - len("http://landscape.canonical.com/accept-invitation/")
        self.secure_id = make_secure_id(length)
        self.name = name
        self.email = email
        self.salesforce_key = salesforce_key
        self.creation_time = datetime.utcnow()

    @property
    def access_delegate(self):
        return self.account

    @requires(PUBLIC)
    def activate(self, person):
        """Add the given person to the account and delete this invitation.

        @raises: L{MaximumAccountAdministratorsError} if the account is already
            at or over the limit of administrators.
        """
        if self.account in list(person.accounts):
            self.cancel()
            raise PersonAlreadyInAccountError()

        if self.account.people.count() >= self.account.max_people_count:
            self.cancel()
            raise MaximumAccountAdministratorsError()

        store = get_main_store()
        result = store.find(
            AccountInvitationRole, AccountInvitationRole.invitation_id == self.id
        )

        if result.any() is None:
            # No Role associated with this invitation. Remove the invitation
            # and raise an error.
            self.remove()
            raise InvalidInvitationError(
                "The role you were invited to no longer exists."
            )

        # Use the account-specific role registry here to be able to
        # accept invitations for stored roles too
        roles = []
        from canonical.landscape.model.access.role import get_role_registry

        registry = get_role_registry([self.account.id])
        for invitation_role in result:
            role = registry.get(invitation_role.role_key)
            roles.append(role.name)

        self.account.add_person(person, roles=roles)

        if self.salesforce_key:
            self.account.create_salesforce_user(person, self.salesforce_key)

        logging.info(
            "Person %s activated invitation %s for account '%s'"
            % (person.email, self.secure_id, self.account.title)
        )
        event_log = EventLog(self.account, person)
        event_log.create_log_entry(
            self.account,
            "Person %s activated invitation %s." % (person.email, self.secure_id),
        )
        self.remove()

    @requires(PUBLIC)
    def create_job_accept_invitation(self, page_key, person):
        """
        Create an C{accept_invitation} job to accept an invitation.
        """
        dispatcher = getUtility(IDispatcher)
        dispatcher.create_job(
            page_key,
            self.namespace,
            "accept-invitation",
            secure_id=self.secure_id,
            person_id=person.id,
        )

    @requires(PUBLIC)
    def create_job_reject_invitation(self, page_key):
        """
        Create a C{reject_invitation} job to reject an invitation.
        """
        dispatcher = getUtility(IDispatcher)
        dispatcher.create_job(
            page_key, self.namespace, "reject-invitation", secure_id=self.secure_id
        )

    @requires(PUBLIC)
    def cancel(self):
        logging.info(
            "Invitation %s to %s in account '%s' was cancelled"
            % (self.secure_id, self.email, self.account.id)
        )
        self.remove()

    @property
    def account_title(self):
        return removeSecurityProxy(self.account).title

    @property
    def namespace(self):
        """Get an L{Account}-bound namespace for invitation-related pages."""
        return "landscape.account.%d.invitation" % self.account.id

    def remove(self):
        """
        Remove this account invitation.
        """
        Store.of(self).remove(self)

    @requires("InviteAdministrator")
    def send_invitation(invitation):
        """Send an invitation mail to the invited user.

        @param invitation: The L{AccountInvitation} for the user.
        @param root_url: The root URL of the Landscape instance.
        """
        to = format_address(invitation.name, invitation.email)
        account = invitation.account
        subject = INVITATION_SUBJECT % {"account": account.name}
        body = INVITATION_CONTENT % {
            "account_title": account.title,
            "account_name": account.name,
        }
        body += "\n\n" + INVITATION_FOOTER % {
            "root": get_root_url().rstrip("/"),
            "key": invitation.secure_id,
            "days": INVITATION_EXPIRY_IN_DAYS,
        }
        getUtility(ILandscapeMailer).send([to], subject, body)


class AccountInvitationRole(object):
    """Associate a role to a L{AccountInvitation}.

    @param invitation_id: The L{AccountInvitation} which the L{StoredRole} will
        be linked to.
    @param role_key: The L{StoredRole} to link.
    """

    __storm_table__ = "invitation_role"

    __storm_primary__ = "invitation_id", "role_key"

    invitation_id = Int(allow_none=False)
    role_key = Int(allow_none=False)

    invitation = Reference(invitation_id, AccountInvitation.id)

    def __init__(self, invitation_id, role_key):
        self.invitation_id = invitation_id
        self.role_key = role_key


@requires_advisor
class InvitationManagement(ModelManagement):
    @requires("ViewAccount")
    def check_person_can_invite(self):
        """Check that the calling person is allowed to invite to the account.

        This method performs a number of security-related checks to ensure that
        the calling has the right permissions and roles to invite somebody
        to the account. It's needs to be separate from the C{invite} method
        to allow the UI to call it in order to decide whether to render
        the invitation form or not.
        """
        people_count = self.account.people.count()
        invitations_count = self.get_invitations().count()
        # XXX at some point we might want to replace the explicit roles
        # check with some general permission check, so also other type
        # of roles can perform invitations
        is_owner = self.person.has_role("Owner", self.account.access_context)
        can_invite_initial = self.person.has_permission(
            "landscape.InviteInitialAdministrator", None
        )

        if not is_owner and not can_invite_initial:
            raise IllegalInvitationError("You don't have rights to invite")

        if people_count + invitations_count >= self.account.max_people_count:
            raise MaximumAccountAdministratorsError(
                "Can't have more than %d people in the account"
                % self.account.max_people_count
            )

        if people_count == 0 and not can_invite_initial:
            # Only account managers and support providers can invite the
            # first administrator.
            raise IllegalInvitationError(
                "Illegal attempt to invite first administrator"
            )

        if people_count > 0:
            if not is_owner and self.account.has_people_in_role("Owner"):
                # Account managers can only invite further admins as support
                # users
                raise IllegalInvitationError(
                    "Illegal attempt to invite full administrators"
                )

    @requires("InviteAdministrator")
    def invite(self, name, email, salesforce_key=None, role_names=None):
        """Invite an administrator to the L{Account}."""
        self.check_person_can_invite()
        role_registry = getUtility(IRoleRegistry)

        if role_names is None:
            role_names = ["GlobalAdmin"]

        roles = []
        if self.person.has_role("Owner", self.account.access_context):
            # If the person has the Owner role in the account, he can select a
            # specific role in the account. Note that if the person is also
            # an AccountManager this will be behave basically the same, in
            # order to support the corner-case of account managers also having
            # an account (e.g. in hosted)
            from canonical.landscape.model.access.management import (
                AccessControlManagement,
            )

            management = AccessControlManagement(self.account, self.person)
            for role_name in role_names:
                roles.append(management.get_role(role_name))

        elif self.person.has_permission("landscape.InviteInitialAdministrator", None):
            # If the person is a pure AccountManager or SupportProvider,
            # without the Owner role, he can only specify full-access or
            # support-only as roles. This check should never really
            # fail, because both the API and the UI can't be led to such
            # a code path by user's actions, but we assert it here for
            # safety.
            if not set(["Owner", "SupportUser"]).intersection(role_names):
                raise RuntimeError("Illegal role for first administrator")
            for role_name in role_names:
                roles.append(role_registry.get(role_name))
        invitation = create_invitation(self.account, name, email, salesforce_key, roles)

        # Log the invitation to the event log.
        assert len(role_names) != 0, "role_names is unexpectedly empty."
        self._event_log.create_log_entry(
            self.account,
            describe_list(
                "Person %s (%s) invited to be administrator with role:" % (name, email),
                role_names,
            )
            + ".",
        )
        return invitation

    @requires("ViewInvitations")
    def get_invitations(self):
        """Return all available invitations in the account, ordered by name."""
        store = get_main_store()
        result = store.find(AccountInvitation, account_id=self.account.id)
        result.order_by(AccountInvitation.name)
        return result

    @requires("ViewInvitations")
    def get_invitation(self, secure_id):
        """Return the invitation in the account with the given secure ID."""
        store = get_main_store()
        return store.find(
            AccountInvitation, secure_id=secure_id, account_id=self.account.id
        ).one()

    @requires("ViewInvitations")
    def get_invitation_roles(self, invitation_id=None):
        """
        Return all the C{role_key}s associated with this invitation.
        """
        store = get_main_store()
        return store.find(
            AccountInvitationRole.role_key,
            AccountInvitationRole.invitation_id == invitation_id,
        )


def create_invitation(account, name, email, salesforce_key=None, roles=None):
    """Create a new L{AccountInvitation} with the given details.

    @param account: The L{Account} the invitation is for.
    @param name: The name of the invitee.
    @param email: The email of the invitee.
    @param salesforce_key: The Salesforce key the invitee will be linked to.
    @param roles: The L{Role}s the invitee will be granted in the account.
    """
    store = get_main_store()
    result = store.find(
        AccountInvitation,
        AccountInvitation.account_id == account.id,
        AccountInvitation.email == email,
    )
    if result.one():
        raise RepeatedInvitationError(
            "{} is already invited to this account".format(name)
        )

    if account.people.find(Person.email == email).one():
        raise RepeatedInvitationError(
            "{} is already a member of this account".format(name)
        )

    invitation = AccountInvitation(account, name, email, salesforce_key)

    if roles:
        # Link all passed in roles with this invitation.
        for role in roles:
            store.add(AccountInvitationRole(invitation.id, role.key))

    return invitation


def get_invitation(secure_id):
    return get_main_store().find(AccountInvitation, secure_id=secure_id).one()


def get_invitation_by_email(account, email):
    return (
        get_main_store()
        .find(
            AccountInvitation,
            AccountInvitation.account_id == account.id,
            AccountInvitation.email == email,
        )
        .one()
    )


def get_invitation_by_sfid(sfid):
    if sfid is None:
        return None
    return get_main_store().find(AccountInvitation, salesforce_key=sfid).one()


INVITATION_EXPIRY_IN_DAYS = 14

INVITATION_SUBJECT = "You have been invited to the Landscape account %(account)s"

INVITATION_CONTENT = """\
You've been invited to become an administrator in the following Landscape
account:

    %(account_title)s (%(account_name)s)."""

INVITATION_FOOTER = """\
If you were expecting this invitation, please open the following URL to
accept it:

%(root)s/accept-invitation/%(key)s

Please note that this invitation will expire in %(days)s days.

Thanks,

The Canonical Team.
"""
