# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime, timedelta, time
import logging

import transaction

from zope.component import getUtility

from canonical.security.interfaces import IRoleRegistry

from canonical.landscape.constants import UBUNTU_ADVANTAGE_PURCHASE_URL
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.mail.mail import format_address
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.license import License
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store


FOURTEEN_DAY_NOTIFICATION_EMAIL = """
Dear %(name)s,

It's been 14 days since you registered for the Ubuntu Advantage free trial.  We
hope you've found it useful and would like to remind you that your account will
be closing in 16 days.
"""

THREE_DAY_NOTIFICATION_EMAIL = """
Dear %(name)s,

It's been 27 days since you registered for the Ubuntu Advantage free trial.  We
hope you've found it useful and would like to remind you that your account will
be closing in 3 days.
"""

NOTIFICATION_OF_EXPIRY_EMAIL = """
Dear %(name)s,

Your Ubuntu Advantage trial account has now expired. We hope you've found your
experience with it useful!
"""

NOTIFICATION_COMMON_CONTENT = """

Want to experience the full benefit of the Ubuntu Advantage service? You
can still upgrade your account by subscribing to the full package. You'll have
complete access to:

    * Canonical systems management tool Landscape
    * Dedicated support for Ubuntu and Landscape
    * Knowledge Base
    * Comprehensive legal cover

Make your purchase or contact our sales team at {purchase_url}.

Thank you for trying Ubuntu Advantage!


Regards,


The Canonical team
""".format(
    purchase_url=UBUNTU_ADVANTAGE_PURCHASE_URL
)


def get_expiring_trial_licenses(days):
    """
    Returns a list of L{License} objects that are due to expire on a date that
    is "days" from today, that are C{TrialFeatures} licenses.
    """
    start_of_day = datetime.combine(datetime.utcnow() + timedelta(days=days), time())
    end_of_day = start_of_day + timedelta(hours=24)
    store = get_main_store()
    trial_features = getUtility(IRoleRegistry).get("TrialFeatures")

    # We need to find all Accounts which have all Licenses due to expire
    # on the day "days" ago.
    expiring_licenses = store.find(
        License,
        License.expiration_date >= start_of_day,
        License.expiration_date < end_of_day,
        License.account_id == Account.id,
        License.role_key == trial_features.key,
    )
    return expiring_licenses.order_by(License.expiration_date)


def notify_account_admins_of_expiry(account, email_text, subject, days, dry_run=False):
    """
    Sends emails to all admins with L{Owner} role on the account.
    """
    logger = logging.getLogger("trial_account_notifications")
    owner_role = getUtility(IRoleRegistry).get("Owner")
    for admin in account.people.order_by(Person.name):
        if admin.has_role(owner_role, account.access_context):
            recipient = format_address(admin.name, admin.email)
            body = (email_text + NOTIFICATION_COMMON_CONTENT) % {"name": admin.name}
            message = "%s notification to %s (%s) for account '%s' (%s) " "%d days"
            if dry_run:
                message = message % (
                    "Not sending",
                    admin.name,
                    admin.email,
                    account.title,
                    account.name,
                    days,
                )
            else:
                message = message % (
                    "Sending",
                    admin.name,
                    admin.email,
                    account.title,
                    account.name,
                    days,
                )
            logger.info(message)
            if not dry_run:
                mailer = getUtility(ILandscapeMailer)
                mailer.send([recipient], subject, body)
                transaction.commit()


def send_notifications_to_expiring_trials(dry_run=False):
    """
    This finds all expiring C{TrialLicense}s on various day offsets,
    and sends a notification to them.

    @return: the number of licenses detected as expiring.
    """
    expiring_licenses = 0
    try:
        licenses_expiring_in_14_days = get_expiring_trial_licenses(14)
        notification_template = (
            "Ubuntu Advantage free trial notification " "for %s (%s)"
        )
        for license in licenses_expiring_in_14_days:
            expiring_licenses += 1
            notify_account_admins_of_expiry(
                license.account,
                FOURTEEN_DAY_NOTIFICATION_EMAIL,
                notification_template % (license.account.title, license.account.name),
                14,
                dry_run=dry_run,
            )

        licenses_expiring_in_3_days = get_expiring_trial_licenses(3)
        for license in licenses_expiring_in_3_days:
            expiring_licenses += 1
            notify_account_admins_of_expiry(
                license.account,
                THREE_DAY_NOTIFICATION_EMAIL,
                notification_template % (license.account.title, license.account.name),
                3,
                dry_run=dry_run,
            )

        licenses_expiring_today = get_expiring_trial_licenses(0)
        for license in licenses_expiring_today:
            expiring_licenses += 1
            notify_account_admins_of_expiry(
                license.account,
                NOTIFICATION_OF_EXPIRY_EMAIL,
                "Ubuntu Advantage free trial expiration for %s (%s)"
                % (license.account.title, license.account.name),
                0,
                dry_run=dry_run,
            )
    finally:
        transaction.abort()
    return expiring_licenses
