# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, Unicode, AutoReload, Like

from canonical.lib.text import is_valid_slug, slug_from_title, describe_dict

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.selection.selection import ComputerSelection


__all__ = [
    "SavedSearchManagement",
    "SavedSearch",
    "generate_saved_search_name",
    "InvalidSavedSearchName",
    "EmptySavedSearchTitle",
    "EmptySavedSearchCriteria",
    "InvalidSavedSearchCriteria",
    "DuplicateSavedSearchError",
    "EmptySavedSearchNameAndTitle",
    "UnknownSavedSearch",
    "EmptyTitleAndSearch",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


class InvalidSavedSearchName(Exception):
    """Raised when no saved search name was provided."""


class EmptySavedSearchTitle(Exception):
    """Raised when no saved search title was provided."""


class EmptySavedSearchCriteria(Exception):
    """Raised when no search criteria was provided."""


class EmptySavedSearchNameAndTitle(Exception):
    """Raised if neither name or title was provided."""


class InvalidSavedSearchCriteria(Exception):
    """Raised if a saved search criteria contains a search: term."""


class DuplicateSavedSearchError(DuplicateModelError):
    """Raised if a saved search with specified name or title already exists."""

    model_name = "Saved search"


class EmptyTitleAndSearch(Exception):
    """Raised when editing a saved search if title and search are empty."""


class UnknownSavedSearch(Exception):
    """Raised when a saved search could not be found in edit mode."""


@requires_advisor
class SavedSearchManagement(ModelManagement):
    """Management class to access L{SavedSearch}es of an L{Account}."""

    requires("ViewSavedSearch", PRIVATE, ["account"])

    @requires("CreateSavedSearch")
    def create_saved_search(self, name, title, search):
        """
        Creates a L{SavedSearch}.

        @param name: The name of the L{SavedSearch} (must be a valid slug).
        @param title: The title of the L{SavedSearch}.
        @param search: The search criteria of the L{SavedSearch}.
        """
        prefix = "Saved search created with"
        values = {"name": name, "title": title, "search criteria": search}
        message = "%s." % describe_dict(prefix, values, pluralize=False)
        created = SavedSearch(self.account, name, title, search)
        self._event_log.create_log_entry(created, message)
        return created

    @requires("ViewSavedSearch")
    def get_saved_searches(
        self, ids=None, names=None, titles=None, criteria=None, limit=None, offset=None
    ):
        """
        Get L{SavedSearch}es with specified ids, names, or titles.

        @param ids: The ids to retrieve.
        @param names: The SavedSearch names to retrieve.
        @param titles: The SavedSearch titles to retrieve.
        @param criteria: The SavedSearch criteria to retrieve.
        @param limit: Maximum number of results to return.
        @param offset: Offset within results to start returning.
        """
        where = []
        if ids is not None:
            where.append(SavedSearch.id.is_in(ids))
        if names is not None:
            where.append(SavedSearch.name.is_in(names))
        if titles is not None:
            where.append(SavedSearch.title.is_in(titles))
        if criteria is not None:
            where.append(SavedSearch.search.is_in(criteria))
        result_set = self._find(*where).order_by(SavedSearch.name)
        if limit is not None:
            result_set.config(limit=limit)
        if offset is not None:
            result_set.config(offset=offset)
        return result_set

    @requires("ViewSavedSearch")
    def get_saved_search_by_id(self, id):
        """
        Get a L{SavedSearch} with a specific id

        @param id: The SavedSearch id to retrieve.
        """
        return self.get_saved_searches(ids=(id,)).one()

    @requires("ViewSavedSearch")
    def get_saved_search_by_name(self, name):
        """
        Get a L{SavedSearch} with a specific name

        @param name: The SavedSearch name to retrieve.
        """
        return self.get_saved_searches(names=(name,)).one()

    @requires("ViewSavedSearch")
    def get_saved_search_by_title(self, title):
        """
        Get a L{SavedSearch} with a specific title

        @param title: The SavedSearch title to retrieve.
        """
        return self.get_saved_searches(titles=(title,)).one()

    @requires("EditSavedSearch")
    def remove_saved_searches_by_ids(self, ids):
        """
        Remove L{SavedSearch}es from the L{Account}.

        @param ids: The ids of the L{SavedSearch}es to remove.
        """
        result = self.get_saved_searches(ids=ids)
        for saved_search in result:
            self._event_log.create_log_entry(saved_search, "Saved search removed.")
        return result.remove()

    @requires("EditSavedSearch")
    def remove_saved_searches_by_names(self, names):
        """
        Remove L{SavedSearch}es from the L{Account}.

        @param names: The names of the L{SavedSearch}es to remove.
        """
        result = self.get_saved_searches(names=names)
        removed = sorted(result.values(SavedSearch.name))
        for saved_search in result:
            self._event_log.create_log_entry(saved_search, "Saved search removed.")
        result.remove()
        return removed

    @requires("ViewAccount")
    def remove_saved_searches(self, saved_searches):
        """
        Remove L{SavedSearch}es from the L{Account}.

        @param saved_searches: A list of L{SavedSearch}es to be removed.
        """
        for saved_search in saved_searches:
            self._event_log.create_log_entry(saved_search, "Removed")
            self.account.store.remove(saved_search)

    @requires("ViewSavedSearch")
    def get_saved_searches_by_name_prefix(self, prefix):
        """
        Search for L{SavedSearch}es within the account where the name starts
        with C{prefix}.

        @param prefix: The name prefix used for searching.
        """
        return self._find(Like(SavedSearch.name, prefix + "%"))

    def _find(self, *where):
        """Find L{SavedSearch}es in this L{Account}.

        @param where: An optional list for Storm expressions.
        @return: A L{ResultSet} yielding L{SavedSearch}es in the account.
        """
        return self.account.store.find(
            SavedSearch, SavedSearch.account_id == self.account.id, *where
        )

    def validate_create_saved_search(self, name, title, search):
        """
        Validate a L{SavedSearch} at creation time.

        @param name: The name of the L{SavedSearch} (must be a valid slug).
        @param title: The title of the L{SavedSearch}.
        @param search: The search criteria of the L{SavedSearch}.
        """
        if not search:
            raise EmptySavedSearchCriteria("Search must not be blank.")

        if "search:" in search:
            raise InvalidSavedSearchCriteria(
                "search: cannot be used as a " "saved search criteria."
            )

        if not name and not title:
            raise EmptySavedSearchNameAndTitle(
                "A name and a title must be " "provided."
            )

        if not title:
            raise EmptySavedSearchTitle("Title must not be blank.")

        if name:
            if not is_valid_slug(name):
                raise InvalidSavedSearchName(
                    "Not a valid name: '%s'. "
                    "Use only lowercase ASCII and numbers." % name
                )
        else:
            name = generate_saved_search_name(self, title)

        self.check_duplicate(
            SavedSearch, DuplicateSavedSearchError, name=name, title=title
        )

        return (name, title, search)

    def validate_edit_saved_search(self, name, title, search):
        """
        Validate a L{SavedSearch} at editing time.

        @param name: The name of the L{SavedSearch} (must be a valid slug).
        @param title: The title of the L{SavedSearch}.
        @param search: The search criteria of the L{SavedSearch}.
        """
        if not title and not search:
            raise EmptyTitleAndSearch(
                "The request must contain a Title and " "Search parameter."
            )

        savedsearch = self.get_saved_search_by_name(name)
        if not savedsearch:
            raise UnknownSavedSearch("Saved search '%s' not found." % name)

        if title:
            if savedsearch.title != title:
                self.check_duplicate(
                    SavedSearch, DuplicateSavedSearchError, title=title
                )

        if search and "search:" in search:
            raise InvalidSavedSearchCriteria(
                "search: cannot be used as a " "saved search criteria."
            )

    @requires("EditSavedSearch")
    def edit_saved_search(self, name, title=None, search=None):
        """
        Edit a L{SavedSearch}.

        @param name: The name of a existing L{SavedSearch}.
        @param title: The new title of the L{SavedSearch}.
        @param search: The new search criteria of the L{SavedSearch}.
        """
        savedsearch = self.get_saved_search_by_name(name)
        messages = []
        if title and savedsearch.title != title:
            messages.append(
                "Title changed from '%s' to '%s'." % (savedsearch.title, title)
            )
            savedsearch.title = title
        if search and savedsearch.search != search:
            messages.append(
                "Search criteria changed from '%s' to '%s'."
                % (savedsearch.search, search)
            )
            savedsearch.search = search
        if messages:
            self._event_log.create_log_entry(savedsearch, " ".join(messages))
        return savedsearch


def generate_saved_search_name(management, title):
    """
    Generate a unique name for a L{SavedSearch} based on the C{title} which
    can be used in the URL slug when accessing L{SavedSearch}.

    @param management: A L{SavedSearchManagement} object.
    @param title: The title of the new L{SavedSearch}.
    """

    def existing_check(slug):
        return set(
            saved_search.name
            for saved_search in management.get_saved_searches_by_name_prefix(slug)
        )

    return slug_from_title(title, 55, existing_check)


@requires_advisor
class SavedSearch(AccountHelper):

    __storm_table__ = "saved_search"

    requires("ViewSavedSearch", PRIVATE, ["id"])
    requires("ViewSavedSearch", "EditSavedSearch", ["name", "title", "search"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)

    name = Unicode(allow_none=False)
    title = Unicode(allow_none=False)
    search = Unicode(allow_none=False)

    def __init__(self, account, name, title, search):
        super(SavedSearch, self).__init__(account)
        self.name = name
        self.title = title
        self.search = search

    def get_computer_selection_from_search(self):
        """
        Executes the saved search and returns the results as a
        L{ComputerSelection}.
        """
        from canonical.landscape.model.computer.search import ComputerSearch

        search = ComputerSearch(self.account)
        computers = search.find(self.search)
        selection = ComputerSelection(self.account)
        selection.select_computers(list(computers))
        return selection

    @requires("ViewSavedSearch")
    def get_count(self):
        """Return the number of computers matching this search."""
        from canonical.landscape.model.computer.search import ComputerSearch

        search = ComputerSearch(self.account)
        return search.find(self.search).count()

    def get_state(self):
        """
        Return the state of the script in JSON serialisable format.
        """
        return {"name": self.name, "title": self.title, "search": self.search}
