# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from venusian import attach

from zope.interface import implementer
from zope.component import getUtility

from canonical.lib.text import lower_class
from canonical.amqp.message.interfaces import IDispatcher

from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.action.interfaces import IActionHandler


def action(handler_class):
    """Decorator to use to mark an action handler.

    When invoking L{ActionRegistry.scan} the classes marked with this decorator
    will be added to the registry.

    @param handler_class: The L{ActionHandler} class to register.
    """

    def callback(scanner, name, handler_class):
        scanner.registry.add(handler_class)

    attach(handler_class, callback, category="action")
    return handler_class


@implementer(IActionHandler, IAccessDelegateAware)
class ActionHandler(object):
    """Handle an action for performing high-level business logic.

    This class serves as base for classes implementing the business logic
    of actions that are common to both the API and the UI layers.

    Typically such action accepts plain integer IDs or string names as
    input parameters coming from the API or UI layer, and use them to
    look up model objects, performing model lookups or changes and interfacing
    to service gateways such as the job handler, the notification system and
    the mailer.

    @param action: The name of the action to handle.
    @param context: Optionally, a context object for the action.
    @param person: Optionally, the person requesting the action.
    @param uuid: Optionally, the UUID to pass to the job dispatcher when
        creating jobs.

    @ivar namespace: Optional namespace to use when submitting jobs and
        sending notifications.
    @ivar permission: Optional C{SecurityProxy} permission for C{handle}.
    """

    namespace = None
    permission = None

    # By default, the job name is derived from the action handler class
    # name. It can be overridden here.
    job_name = None

    def __init__(self, context=None, person=None, uuid=None):
        self.action = lower_class(self.__class__, suffix="ActionHandler")
        self.context = context
        self.person = person
        self.uuid = uuid

    def handle(self, **params):
        """Handle an action request with the given data.

        @param params: The action parameters.
        @return: The result to be returned to the action caller.

        @raises ActionError: If anything goes wrong.
        """
        raise NotImplementedError()

    def send_job(self, **kwargs):
        """Send a job to the job handler.

        The requested job will have the same name of the action being handled.

        @param kwargs: Keywords arguments that to be passed to the L{JobRunner}
            registered to handle the requested job.
        """
        dispatcher = getUtility(IDispatcher)
        if self.job_name is None:
            job_name = self.action
        else:
            job_name = self.job_name
        dispatcher.create_job(self.uuid, self.namespace, job_name, **kwargs)

    def send_notification(self, content):
        """Send a notification to the notification publisher.

        The notification is sent to the C{ActionHandler.namespace} namespace.

        @param content: A JSON-serializable object with the content of the
            notification to send.
        """
        dispatcher = getUtility(IDispatcher)
        dispatcher.send_notification(content, None, self.action, self.namespace)

    @property
    def access_delegate(self):
        return self.context
