"""Manage the the activity_parent_access cache table.

The activity_parent_access table represents a cache of the access contexts
associated with the children of a parent activity, by entity type.

For example if a parent a has 4 children like:

  - child1 linked to Computer C1 in access context AC1
  - child2 linked to Computer C2 in access context AC1
  - child3 linked to Computer C3 in access context AC2
  - child4 linked to JujuModel J1 in access context AC1

then the cache table will have 3 rows, summarizing the access contexts linked
to the children of a given type:

  - row1: (parent_id, <Computer entity type> , AC1)
  - row2: (parent_id, <Computer entity type> , AC2)
  - row3: (parent_id, <JujuModel entity type> , AC1)

This makes it possible to perform RBAC-related queries against parent
activities without having to look a the children (which is potentially
expensive for parents with lots of children).

See also c.l.model.account.register.register_account_entity_types for the
actual numberic values of the entity types of the various models.
"""
from builtins import object
from storm.locals import Int


DELETE_STATEMENT = """
DELETE FROM activity_parent_access WHERE parent_id IN ?
"""

INSERT_STATEMENT = """
INSERT INTO activity_parent_access
    SELECT DISTINCT parent_id, entity_type, data.access_context_id
        FROM activity JOIN account_entity_data as data
                      ON entity_data_id = data.id
        WHERE parent_id IN ?
"""


class ActivityParentAccess(object):
    """Storm model for the parent activity access cache.

    This class should be used only for generating SELECT queries
    with Store.find. Inserting and deleteting rows is managed by
    the update_activity_parent_access() function.
    """

    __storm_table__ = "activity_parent_access"
    __storm_primary__ = ("parent_id", "entity_type", "access_context_id")

    parent_id = Int(allow_none=False)
    entity_type = Int(allow_none=False)
    access_context_id = Int(allow_none=False)


def update_activity_parent_access(store, parent_ids):
    """Update the activity_parent_access cache for the given parent IDs.

    This function will delete all existing cache rows for the given parent IDs
    and repopulate the cache for those parent IDs, with values from their
    current children.

    It should be called in the following situations:

      - whenever children get added to a parent
      - whenever children get deleted from a parent
      - whenever the account entity linked to a child activity is moved to
        a different access context

    @param store: The account Store the given parent IDs live in.
    @param parent_ids: The ActivityParent IDs to update the cache for.
    """
    if not parent_ids:
        return

    parent_ids = tuple(parent_ids)
    store.execute(DELETE_STATEMENT, (parent_ids,))
    store.execute(INSERT_STATEMENT, (parent_ids,))
