# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer

from storm.locals import Unicode, Int, DateTime

from canonical.lib.timestamp import to_timestamp, format_datetime
from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.activity.interfaces import IReappliableRequest
from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.configuration import get_root_url


__all__ = ["SignalProcessRequest", "ExecuteScriptRequest"]

__clearlist__ = ["SignalProcessRequest", "ExecuteScriptRequest"]

SUCCESS_RESULT = 1
TIMEOUT_RESULT = 102

NOT_SUPPORTED = 500
ATTACHMENTS_NOT_SUPPORTED = 501

__requires_prefix__ = "landscape."


@requires_advisor
class SignalProcessRequest(ComputerRequest):

    __storm_table__ = "signal_process_request"

    signal = Unicode(allow_none=False)
    pid = Int(allow_none=False)
    name = Unicode(allow_none=False)
    start_time = DateTime(allow_none=False)

    def __init__(self, activity, computer, signal, pid, name, start_time):
        self.signal = signal
        self.pid = pid
        self.name = name
        self.start_time = start_time
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(SignalProcessRequest, self).__init__(activity, computer)

    @requires("ViewActivity")
    def get_summary(self):
        return "Signal process %s (PID %s) with %s" % (self.name, self.pid, self.signal)

    def _started(self):
        message_store = self.computer.get_message_store()
        message_store.add(
            {
                "type": "signal-process",
                "operation-id": self.activity.id,
                "pid": self.pid,
                "name": self.name,
                "signal": self.signal,
                "start-time": to_timestamp(self.start_time),
            }
        )


@requires_advisor
@implementer(IReappliableRequest)
class ExecuteScriptRequest(ComputerRequest):
    """
    Execute a script on the client by sending an C{execute-script} message.

    @type interpreter: C{unicode}
    @ivar interpreter: The path to the interpreter that will be put after a #!
    @type code: C{unicode}
    @ivar code: The code that will be executed with the interpreter.
    @type username: C{unicode}
    @ivar username: The user that should execute the script.
    @type time_limit: C{int}
    @ivar time_limit: The number of seconds to wait for the script to finish.
    @type title: C{unicode}
    @ivar title: Title identifier for the script.
    """

    __storm_table__ = "execute_script_request"

    interpreter = Unicode(allow_none=False)
    code = Unicode(allow_none=False)
    username = Unicode(allow_none=False)
    time_limit = Int(allow_none=False)
    title = Unicode()

    reapply_arguments = {
        "interpreter": "interpreter",
        "code": "code",
        "username": "username",
        "time_limit": "time_limit",
        "title": "title",
    }

    def __init__(
        self, activity, computer, interpreter, code, username, time_limit, title
    ):
        self.interpreter = interpreter
        self.code = code
        self.username = username
        self.time_limit = time_limit
        self.title = title

        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(ExecuteScriptRequest, self).__init__(activity, computer)

    @requires("ViewActivity")
    def get_summary(self):
        if self.title is not None:
            return "Run script: %s" % (self.title,)
        else:
            # For old activities
            return "Run %s script" % (self.interpreter,)

    def get_text(self):
        if self.activity.result_code == NOT_SUPPORTED:
            return (
                "To support script execution, the Landscape client on this\n"
                "computer must be upgraded, and the following line must be\n"
                "added to /etc/landscape/client.conf:\n"
                "\n"
                "include_manager_plugins = ScriptExecution\n"
            )
        elif self.activity.result_code == ATTACHMENTS_NOT_SUPPORTED:
            return (
                "To support script attachments, the Landscape client on this"
                "\ncomputer must be upgraded."
            )
        elif self.activity.result_code == TIMEOUT_RESULT:
            return "Script timed out while executing."
        else:
            return super(ExecuteScriptRequest, self).get_text()

    def _started(self):
        """Send an execute-script message."""
        message_store = self.computer.get_message_store()
        client_version = self.computer.current_version

        computer = self.activity.info.computer
        # Supply some useful environment variables to the script.
        env = {
            "LANDSCAPE_ACCOUNT": self.activity.account.name,
            "LANDSCAPE_COMPUTER_ID": str(computer.id),
            "LANDSCAPE_COMPUTER_TAGS": (
                " ".join(tag.name for tag in computer.get_tags())
            ),
            "LANDSCAPE_URL": get_root_url(),
            "LANDSCAPE_ACTIVITY_ID": str(self.activity.id),
            "LANDSCAPE_ACTIVITY_CREATOR_ID": str(self.activity.creator_id),
            "LANDSCAPE_ACTIVITY_CREATION_TIME": format_datetime(
                self.activity.creation_time
            ),
        }

        attachments = {}

        out_of_band_support = (
            client_version.split(b"-")[0] >= b"11.11"
            if client_version is not None
            else False
        )
        if out_of_band_support:
            for attachment in self.activity.parent.attachments:
                attachments[attachment.filename] = attachment.id
        else:
            for attachment in self.activity.parent.attachments:
                attachments[attachment.filename] = attachment.get_data()

        message_store.add(
            {
                "type": "execute-script",
                "interpreter": self.interpreter,
                "code": self.code,
                "username": self.username,
                "time-limit": self.time_limit,
                "operation-id": self.activity.id,
                "attachments": attachments,
                "env": env,
            }
        )
