# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
The activity system tracks events that have occurred and that are in
progress.
"""

from zope.interface import implementer
from zope.security.proxy import removeSecurityProxy

from storm.databases.postgres import JSON
from storm.locals import Enum, DateTime, Int, Reference, Store, Unicode

from canonical.security.protect import requires, PRIVATE, requires_advisor
from canonical.lib.timestamp import format_datetime

from canonical.landscape.model.main.interfaces import (
    IAccessDelegateAware,
    IAccessContextAware,
)
from canonical.landscape.model.activity.status import (
    add_status_transition_event,
    ActivityStatusChangeError,
    WAITING,
    QUEUED,
    INPROGRESS,
    SUCCEEDED,
    FAILED,
    CANCELED,
    SCHEDULED,
    BLOCKED,
    VALID_STATUS_TRANSITIONS,
    STATUS_MAP,
    NullTransitionError,
)
from canonical.landscape.model.activity.interfaces import IActivity, IActivityInfo
from canonical.landscape.model.activity.base import (
    ActivityBase,
    ActivityBaseInfoMeta,
    ActivityBaseInfo,
)
from canonical.landscape.model.activity.timeline import ActivityChildTimeline
from canonical.landscape.model.entity.helper import get_account_entity_class
from canonical.landscape.model.entity.data import AccountEntityData
from future.utils import with_metaclass


__all__ = [
    "AbstractActivityInfo",
    "Activity",
    "ActivityInfo",
    "ActivityInfoMeta",
    "ActivityOrdering",
    "reverse_activity_info_types",
    "activity_info_types",
    "register_activity_info_type",
    "activity_type_search_terms",
    "get_succeeding_activities",
    "get_preceding_activities",
    "get_child_activity",
]


__clearlist__ = ["Activity", "get_child_activity"]


__requires_prefix__ = "landscape."


class ChildrenDisallowedError(Exception):
    """Raised when adding an activity child with an inappropriate info class.

    Info classes must inherit from ActivityParentInfo.
    """


# TODO: rename to ActivityChild when all the parent activity code has
#       been moved out..
@implementer(IActivity, IAccessContextAware)
@requires_advisor
class Activity(ActivityBase):
    """High-level representation of state changes.

    An activity represents a high-level change to the internal state of
    Landscape and possibly to the state of external systems driven by
    Landscape.

    An activity can be created by users via UI or API requests, or by
    Landscape itself.

    The Activity class should not be subclassed. Custom behavior should rather
    be implemented by using the infoheritance pattern and passing an
    ActivityInfo class (see https://storm.canonical.com/Infoheritance).

    Here follows some code exemplifying a typical lifecycle, in which
    an activity is created and then executed::

      activity = parent_activity.add(my_info_class, my_params)
      parent.settle()  # this queues child activities
      activity.start()  # started
      activity.advance(50)  # 50% completed
      activity.succeed()  # completed

    See also canonical.landscape.model.activity.status for more details
    about the activity lifecycle.

    @ivar info_type: The type ID of the info object.
    @ivar info: The object which describes this Activity and extends its
        functionality.
    @ivar creator: Optionally, the Person that created this Activity.
    @ivar creation_time: The time at which this Activity was created.
    @ivar parent: This Activity's parent.
    @ivar computer: Optionally, the computer this activity affects.
    @ivar progress: The overall activity's progress, as a percentage.
    @ivar status: The activity's status, an instance of ActivityStatus
    @ivar result_text: Optionally, a field for capturing the details
        of an operation.
    @ivar entity_data_id: Link to the AccountEntityData of the
        AccountEntity that this activity involves.

    """

    requires(
        "ViewActivity",
        PRIVATE,
        [
            "id",
            "account",
            "access_context",
            "info",
            "creator",
            "creation_time",
            "info_type",
            "account_id",
            "parent_id",
            "status",
            "schedule_after_time",
            "schedule_before_time",
        ],
    )

    requires("ViewActivity", "CreateActivity", ["parent"])

    __storm_table__ = "activity"

    parent_id = Int()

    status = Enum(map=STATUS_MAP, allow_none=True)
    progress = Int(allow_none=False, default=0)
    result_code = Int()
    result_text = Unicode()
    schedule_after_time = DateTime()
    schedule_before_time = DateTime()
    entity_data_id = Int()

    entity_data = Reference(entity_data_id, AccountEntityData.id)

    def __repr__(self):
        return "<Activity %s (%s)>" % (self.info, self.status)

    def __init__(self, info_class, parent, entity=None, **info_class_kwargs):
        """Construct an activity and a correspondent activity info object.

        This is supposed to be used in a way similar to:

            activity = Activity(ActivityInfoSubClass, account=account,
                                parameter_for_info_subclass=value)

        @param info_class: The ActivityInfo subclass to be used as the info
            for this activity.  All keyword parameters besides those
            explicitly specified for this __init__ will be passed to
            the C{info_class} constructor.
        @param parent: The parent activity.
        """
        assert issubclass(
            info_class, ActivityInfo
        ), "info_class not intended for child activities"

        if entity is not None:
            # XXX Eventually the entity parameter will be required, it's
            # currently an optional argument until the transition is completed
            # and all call sites pass it instead of an info-type specific
            # object.

            # XXX When activities are created from the UI, the entity might be
            # security-proxied. Eventually they should be created in the
            # service layer, where there is no security proxy.
            entity = removeSecurityProxy(entity)
            self.entity_data = entity.data

        self.parent = parent

        super(Activity, self).__init__(
            parent.account, info_class, creator=parent.creator, **info_class_kwargs
        )

    def _get_info(self):
        if self._info is None:
            self._info = activity_info_types[self.info_type].load(self)
        return self._info

    def _set_info(self, info):
        self._info = info
        self.info_type = info.info_type

    info = property(_get_info, _set_info)

    @property
    def access_context(self):
        """Return the AccessContext associated to this L{Activity}.

        This is the access context of the related entity.
        """
        return self.entity_data.access_context

    @property
    def entity(self):
        """Return the entity related to the activity."""
        entity_class = get_account_entity_class(self.entity_data.entity_type)
        return self.account.store.find(
            entity_class, entity_class.data_id == self.entity_data.id
        ).one()

    def after(self, activity, minimum_progress=100):
        """Execute this activity after another one.

        This activity won't be executed until the given activity has reached
        the minimum progress.

        @param activity: The L{Activity} we come after.
        @param minimum_progress: The minimum progress the activity we come
            after has to reach before we can start executing.

        If this method is called multiple times with different activities, all
        of them have to reach the given minum progress, before this activity
        can be executed.
        """
        # XXX TODO: check that the given activity has the same parent.
        ordering = ActivityOrdering(self, activity, minimum_progress=minimum_progress)
        self.account.store.add(ordering)

    def wait(self):
        """Prevent activity execution until preceding activities succeed.

        This method will transition the activity to the WAITING state. This
        means the activity will not progress until the activities on which
        this activity is waiting either succeed or reach the specified minimum
        progress (see L{Activity.after}).
        """
        self._transition(WAITING)

    @requires("BlockActivity")
    def block(self):
        """Prevent activity execution until further notice.

        This method will transition the activity to the BLOCKED status.
        The activity will remain blocked until an event external to the
        activity system calls unblocked.

        A typical case is an activity that must wait for user input to
        complete.
        """
        self._transition(BLOCKED)
        self.info._blocked()

    @requires("UnblockActivity")
    def unblock(self):
        """Release a block on an activity allowing it to progress.

        The external condition blocking activity progress has been resolved.
        The activity will transition to its next valid status.

        A typical case is an activity which just received required user input.
        """
        if self.status is not BLOCKED:
            raise ActivityStatusChangeError(
                "Can't unblock activity in {} status".format(self.status.name)
            )

        if self._has_preceding():
            self.wait()
        else:
            self.queue()
        self.info._unblocked()

    @requires("ScheduleActivity")
    def schedule(self, after=None, before=None):
        """Mark the activity as scheduled for later execution.

        This method transition the activity to the SCHEDULED state, which means
        that the activity is scheduled for execution at some point in the
        future.

        @param after: an optional time the activity should be executed after.
        @param before: an optional time the activity should be executed before.
        """
        if not (after or before):
            raise RuntimeError("At least one of 'after' and 'before' must be provided.")

        self._transition(SCHEDULED)

        if after:
            self.schedule_after_time = after
        if before:
            self.schedule_before_time = before

    def queue(self):
        """Queue the activity for execution.

        This method will transition the activity to the QUEUED state, which
        means that the activity is ready to go and should be executed whenever
        possible.

        The L{ActivityInfo} subclass can use the C{_queued} hook to perform
        whatever actions are needed in order to prepare the activity for
        execution. For example, the hook can fire a job that will take care
        of transitioning the activity to the INPROGRESS state and actually
        start its execution.
        """
        self._transition(QUEUED)
        self.info._queued()

    @requires("StartActivity")
    def start(self):
        """Start executing the activity.

        This method will transition the activity to the INPROGRESS state, which
        means that the changes that the activity represents are being
        performed.
        """
        self._transition(INPROGRESS)
        self.progress = 0
        self.info._started()

    def advance(self, progress):
        """Change the amount of progress that the activity has made.

        @param progress: An C{int} ranging between 0 and 100.
        """
        # XXX TODO: check that progress is in the range between 0 and 100,
        #           probably with a database constraint.
        self.progress = progress

        if not self.parent._schedule_update():
            self.parent._update_after_child_progress_change()

    @requires("CompleteActivity")
    def succeed(self, code=None, text=None):
        """Finish the activity successfully.

        This method will transition the activity to the SUCCEEDED state, which
        means that the activity has succeeded and no further change will happen.

        @param code: Optional result code the activity succeeds with, see
            L{canonical.landscape.model.activity.codes} for a list of
            possible values.
        @param text: Optional result text the activity succeeds with.
        """
        self._transition(SUCCEEDED)
        self._finalize(code, text)
        # Automatically advance to 100% progress in case of success.
        self.advance(100)
        self.info._succeeded()

    @requires("CompleteActivity")
    def fail(self, code=None, text=None):
        """Finish the activity unsuccessfully.

        This method will transition the activity to the FAILED state, which
        means that the activity has failed and no further change will happen.

        @param code: Optional result code the activity fails with, see
            L{canonical.landscape.model.activity.codes} for a list of
            possible values.
        @param text: Optional the result text the activity fails with.
        """
        self._transition(FAILED)
        self._finalize(code, text)
        self.info._failed()

    @requires("CancelActivity")
    def cancel(self, code=None, text=None):
        """Cancel the activity.

        This method will transition the activity to the CANCELED state, which
        means that the activity has been canceled and no further change will
        happen.

        @param code: Optional result code the activity is canceled with, see
            L{canonical.landscape.model.activity.codes} for a list of possible
            values.
        @param text: The result text the activity is canceled with, if any.
        """
        self._transition(CANCELED)
        self._finalize(code, text)
        self.info._canceled()

    @requires("ViewActivity")
    def get_state(self, v2=False):
        """
        @return: a dict with L{Activity} information.
        """
        timeline = ActivityChildTimeline(self)

        state = super(Activity, self).get_state(v2=v2)
        state.update(
            {
                "parent_id": self.parent_id,
                "modification_time": format_datetime(timeline.get_modification_time()),
                "completion_time": format_datetime(timeline.get_completion_time()),
                "schedule_before_time": format_datetime(self.schedule_before_time),
                "schedule_after_time": format_datetime(self.schedule_after_time),
                "result_code": self.result_code,
            }
        )

        if self.status is not None:
            if v2:
                state["status"] = self.status.name
            else:
                state["activity_status"] = self.status.name

                # Show the status as "unapproved" in the API for backwards
                # compatibility.
                from canonical.landscape.model.activity.computer import ComputerRequest

                is_blocked_computer_activity = (
                    isinstance(self.info, ComputerRequest) and self.status == BLOCKED
                )
                if is_blocked_computer_activity:
                    state["activity_status"] = "unapproved"

        if not v2:
            # Current API always include a children key for child activities.
            # This is always empty, since we don't support children with
            # children.
            #
            # The key is there because the with_children parameter was True
            # by default, and didn't get passed to children when collecting
            # their states.
            state["children"] = []

        return state

    def _transition(self, new_status):
        """Transition to the given C{status}."""
        self._validate_transition(new_status)

        old_status = self.status
        self.status = new_status
        add_status_transition_event(self, old_status)

        # If the old status is None, it means that we're setting the status on
        # a freshly created activity, otherwise we're updating this activity
        # after it was first created.
        if old_status is None:
            self.parent._increment_status_count(self.status)
        else:
            # Let our parent update itself and possibly unblock other child
            # activities
            if not self.parent._schedule_update():
                self.parent._update_after_child_status_change()

    def _finalize(self, code, text):
        """Update activity attributes for terminal states."""
        self.result_code = code
        self.result_text = text

    def _validate_transition(self, new_status):
        """
        Raise an exception if the transition from the current status to the
        new one is invalid.
        """
        old_name = self.status.name if self.status else None
        new_name = new_status.name if new_status else None

        if old_name == new_name:
            raise NullTransitionError(
                "No transition: {}=>{}".format(old_name, new_name)
            )

        if self.status not in VALID_STATUS_TRANSITIONS[new_status]:
            raise ActivityStatusChangeError(
                "Can't change activity status: {}=>{}".format(old_name, new_name)
            )

    def _has_preceding(self):
        """Whether this activity comes after some other preceding activity."""
        result = self.account.store.find(ActivityOrdering, activity_id=self.id)
        return not result.is_empty()

    ###########################################################################
    # WARNING: all APIs below this marker are deprecated and should eventually
    #          be dropped when activity system rework and consolidation gets
    #          completed, see http://bit.ly/1gRZ0Db.
    ###########################################################################

    @property
    def access_children(self):
        # Child activities don't have children
        return []

    access_cutoff_role = None


activity_info_types = {}
reverse_activity_info_types = {}
stored_activity_info_types = set()
activity_type_search_terms = {}


def register_activity_info_type(info_type, info_class):
    """Register C{info_class} as an L{ActivityInfo} implementation.

    The C{info_type} will be stored as C{Activity.info_type} for any
    activities created with C{info_class}.

    @param info_type: A unique integer representing C{info_class}.
    @param info_class: An L{ActivityInfo}-based class.
    @raise RuntimeError: Raised if a class is already registered for
        C{info_type}.
    """
    existing_info_class = activity_info_types.get(info_type)
    if existing_info_class is not None:
        raise RuntimeError(
            "%r has the same info_type of %r" % (info_class, existing_info_class)
        )
    activity_info_types[info_type] = info_class
    reverse_activity_info_types[info_class] = info_type
    if activity_info_types.get(info_class.__name__) is not None:
        raise RuntimeError(
            "%s is already registered has info search term" % info_class.__name__
        )
    activity_type_search_terms[info_class.__name__] = info_type
    if getattr(info_class, "__storm_table__", None) is not None:
        stored_activity_info_types.add(info_type)
    info_class.info_type = info_type


def unregister_activity_info_type(info_class):
    """Unregister C{info_class} as an L{ActivityInfo} implementation.

    Removes C{info_class} from the registry of known
    C{Activity.info_type}s and cleans up extra attributes.

    @param info_class: An L{ActivityInfo}-based class.
    """
    info_type = info_class.info_type
    if info_type is None:
        # When an info_class is registered, it's info_type attribute
        # is initialized, so if it hasn't been initialized then we've
        # been called to erroneously called to unregister it.
        raise RuntimeError("Info class is not registered: %r" % info_class)
    if info_type in activity_info_types:
        del activity_info_types[info_type]
    if info_class in reverse_activity_info_types:
        del reverse_activity_info_types[info_class]
    if info_class.__name__ in activity_type_search_terms:
        del activity_type_search_terms[info_class.__name__]
    if info_type in stored_activity_info_types:
        stored_activity_info_types.remove(info_type)
    info_class.info_type = None


class ActivityInfoMeta(ActivityBaseInfoMeta):
    """Metaclass for the ActivityInfo class."""

    activity_class = Activity


@implementer(IActivityInfo, IAccessDelegateAware)
@requires_advisor
class ActivityInfo(with_metaclass(ActivityInfoMeta, ActivityBaseInfo)):
    "An object which describes and extends the functionality of an Activity."

    requires(
        "ViewActivity",
        PRIVATE,
        [
            "get_summary",
            "activity",
            "data",
            "get_text",
            "is_approvable",
            "is_cancelable",
            "is_actionable",
            "is_revertable",
            "is_reappliable",
            "is_expandable",
        ],
    )

    @classmethod
    def get_info_type(cls):
        return reverse_activity_info_types[cls]

    def get_state(self):
        """Return a dict with ActivityInfo information."""

        state = super(ActivityInfo, self).get_state()
        state["result_text"] = self.get_text()
        return state

    def get_text(self):
        """Override to provide a custom result text."""
        return self.activity.result_text

    def _set_activity_entity(self, entity):
        """Set the entity associated to an activity."""
        # When activities are created from the UI, the entity might be
        # security-proxy wrapped.
        entity = removeSecurityProxy(entity)
        self.activity.entity_data = entity.data

    def _started(self):
        """Hook called when the activity transitions to the INPROGRESS state."""

    def _queued(self):
        """Hook called when the activity transitions to the QUEUED state."""

    def _canceled(self):
        """Hook called when the activity transitions to the CANCELED state."""

    def _succeeded(self):
        """Hook called when the activity transitions to the SUCCEEDED state."""

    def _failed(self):
        """Hook called when the activity transitions to the FAILED state."""

    def _blocked(self):
        """Hook called when the activity transitions to the BLOCKED state."""

    def _unblocked(self):
        """Hook called when the activity transitions from the BLOCKED state."""


@requires_advisor
class AbstractActivityInfo(ActivityInfo):
    """
    An ActivityInfo type designed to hold generic data.

    NOTE: All new activities (as of 2023-05-26) that require persistent
    data should subclass this class to prevent the proliferation of
    activity-specific database tables.

    Eventually we should collapse existing tables into this model.
    """

    __storm_table__ = "activity_info"

    data = JSON(allow_none=True)

    requires("ViewActivity", PRIVATE, ["data"])

    def __init__(self, data, activity):
        self.data = data

        super().__init__(activity)


class ActivityOrdering:
    """Specifies partial ordering for activities within an activity group.

    In general activities are unordered, but if needed, we can say that
    some activities have to be completed before some other activities
    can be delived.

    The two activities must have the same parent.

    @ivar activity_id: The ID of the L{Activity} that needs to wait for
        some other activity to finish.
    @ivar preceding_activity_id: The ID of the L{Activity} that needs
        to reach the minium progress value before the other L{Activity}
        can be delived.
    @ivar minimum_progress: The progress value the preceding L{Activity}
        has to reach before the other L{Activity} can be delivered.
        Default: 100 (i.e. status of SUCCEEDED)
    """

    __storm_table__ = "activity_ordering"
    __storm_primary__ = "activity_id", "preceding_activity_id"

    activity_id = Int(allow_none=False)
    # XXX: we are fixing the DB column name typo separately.
    preceding_activity_id = Int(allow_none=False, name="preceeding_activity_id")
    minimum_progress = Int(allow_none=False)
    activity = Reference(activity_id, Activity.id)
    preceding_activity = Reference(preceding_activity_id, Activity.id)

    def __init__(self, activity, preceding_activity, minimum_progress=100):
        self.activity = activity
        self.preceding_activity = preceding_activity
        self.minimum_progress = minimum_progress


def get_succeeding_activities(activity, info_classes=None):
    """
    Return a ResultSet of activities that follow the given activity.

    @param info_classes: If given, return only activities whose info class is
        in this list.
    """
    store = Store.of(activity)
    where = [
        ActivityOrdering.activity_id == Activity.id,
        ActivityOrdering.preceding_activity_id == activity.id,
    ]
    if info_classes:
        info_types = [info_class.info_type for info_class in info_classes]
        where.append(Activity.info_type.is_in(info_types))
    return store.find(Activity, *where)


def get_preceding_activities(activity, info_classes=None):
    """Return a ResultSet of activities that precede the given activity.

    @param info_classes: If given, return only activities whose info class is
        in this list.
    """
    store = Store.of(activity)
    where = [
        ActivityOrdering.activity_id == activity.id,
        ActivityOrdering.preceding_activity_id == Activity.id,
    ]
    if info_classes:
        info_types = [info_class.info_type for info_class in info_classes]
        where.append(Activity.info_type.is_in(info_types))
    return store.find(Activity, *where)


def get_child_activity(account, activity_id):
    """Return the Activity with the specified ID in the account."""
    from canonical.landscape.model.activity.collection import ActivityChildCollection

    collection = ActivityChildCollection(
        account.store, account=account, ids=[activity_id]
    )
    result = collection.find()
    return result.one()
