# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import timedelta

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.mail.template import load_template
from canonical.landscape.model.activity.status import BLOCKED
from canonical.landscape.model.alert.computer import (
    ComputerRuleInfo,
    ComputerAlertEmailViewBase,
)
from canonical.landscape.model.computer.twin import AccountComputerTwin


__all__ = ["UnapprovedActivitiesAlert"]


__requires_prefix__ = "landscape."


@requires_advisor
class UnapprovedActivitiesAlert(ComputerRuleInfo):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when activities need approval"

    notification_delay = timedelta(minutes=5)

    def get_summary(self, computer_ids=None):
        from canonical.landscape.model.activity.collection import (
            ActivityChildCollection,
        )

        collection = ActivityChildCollection(
            self.rule.account.store,
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
            statuses=[BLOCKED],
        )
        count = collection.find().count()
        if not count:
            return "No activities need approval"
        elif count > 1:
            return "%s activities need approval" % count
        else:
            return "1 activity needs approval"

    def check(self, computer_ids):
        from canonical.landscape.model.activity.collection import (
            ActivityChildCollection,
        )

        computer_ids = set(computer_ids)
        account = self.rule.account
        collection = ActivityChildCollection(
            account.store,
            account=account,
            statuses=[BLOCKED],
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
        )
        collection = collection.select(AccountComputerTwin.id)
        result = collection.find()
        result.config(distinct=True)

        active_computer_ids = set(result)
        inactive_computer_ids = computer_ids - active_computer_ids
        self.update_states(active_computer_ids, inactive_computer_ids)

    def get_resolution_url(self, computer_ids=None):
        """Get the URL for a page that can be used to resolve this alert.

        If a single L{Activity} requires approval the resolution URL will be
        for that activity.  If many activities require approval the resolution
        URL will be for the page that shows all unapproved activities.

        @param computer_ids: Optionally, a list of L{Computer.id}s to use when
            finding unapproved L{Activity}s.
        @return: A resolution URL or C{None} if no L{Activity}s require
            approval.
        """
        from canonical.landscape.model.activity.collection import (
            ActivityChildCollection,
        )

        result = self.rule.info.get_states(active=True, computer_ids=computer_ids)
        if not result.any():
            return None

        account = self.rule.account
        collection = ActivityChildCollection(
            account.store,
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
            statuses=[BLOCKED],
        )
        result = collection.find()
        if result.count() > 1:
            return "%saccount/%s/activities?filter=unapproved" % (
                get_root_url(),
                account.name,
            )
        else:
            activity = result.one().parent
            return "%saccount/%s/activity/%d" % (
                get_root_url(),
                account.name,
                activity.id,
            )


class UnapprovedActivitiesEmailView(ComputerAlertEmailViewBase):

    _content_template = load_template(__file__, "unapproved-activities-alert.msg")

    def __init__(self, context, person):
        super(UnapprovedActivitiesEmailView, self).__init__(context, person)
        from canonical.landscape.model.computer.activities import (
            get_computer_activities_parents,
        )

        computer_ids = [computer.id for computer in self.affected_computers]
        result = get_computer_activities_parents(
            context.rule.account, computer_ids=computer_ids, statuses=[BLOCKED]
        )
        self.unapproved_activities = [
            activity.info.get_summary() for activity in result
        ]
