# Copyright 2005-2015 Canonical Limited.  All rights reserved.

from builtins import object
from inspect import getargspec

from storm.locals import Int, DateTime, AutoReload, Reference, Store

from canonical.lib.timestamp import format_datetime
from canonical.lib.undef import Undef
from canonical.lib.timestamp import utcnow

from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.activity.interfaces import IExpandableRequest
from future.utils import with_metaclass


class ActivityBase(AccountHelper):
    """High-level representation of state changes.

    An activity represents an high-level change to the internal state of
    Landscape and possibly to the state of external systems driven by
    Landscape.

    An activity can be created by users via UI or API requests, or by
    Landscape itself.

    This serves as a base class of ActivityParent and ActivityChild, holding
    common data and logic.
    """

    id = Int(allow_none=False, primary=True, default=AutoReload)
    info_type = Int(allow_none=False)
    creation_time = DateTime(default_factory=utcnow)
    account_id = Int(allow_none=False)
    creator_id = Int()

    _info = None

    def __init__(self, account, info_class, creator=None, **info_class_kwargs):
        """
        @param account: The Account this activity belongs to.
        @param info_class: The info class for this activity.
        @param creator: The Person creating the activity, or None.
        @param info_class_kwargs: Keyword arguments to pass to the info class.
        """
        super(ActivityBase, self).__init__(account)
        self.creator = creator
        # Note: even though setting self.info below causes self.info_type to be
        # set, we'll first set it explicitly up front because it's a NOT NULL
        # field in the schema, and all of the following assignments (which have
        # lots of behavior in their setters) can easily cause a flush to
        # happen, which would blow up if info_type weren't set.
        self.info_type = info_class.get_info_type()
        self.info = info_class(self, **info_class_kwargs)

    def get_state(self, v2=False):
        """Return a dict with the state of the activity."""
        state = {
            "id": self.id,
            "creation_time": format_datetime(self.creation_time),
            "creator": self.creator.get_state() if self.creator else None,
        }

        if hasattr(self.info, "get_state"):
            has_v2_support = "v2" in getargspec(self.info.get_state).args
            kwargs = {"v2": v2} if has_v2_support else {}
            state.update(self.info.get_state(**kwargs))

        return state

    def _get_creator(self):
        if self.creator_id:
            return get_main_store().get(Person, self.creator_id)
        return None

    def _set_creator(self, person):
        self.creator_id = person.id if person is not None else None

    creator = property(_get_creator, _set_creator)


class ActivityBaseInfoMeta(type):
    """Base Metaclass for the activity info classes.

    It adds references to activity class for subclasses of the info class that
    are stored (they have a database table).
    """

    activity_class = Undef()

    def __new__(cls, name, bases, attrs):
        if "__storm_table__" in attrs:
            activity_id_reference = Int(allow_none=False, primary=True)
            attrs.update(
                {
                    "activity_id": activity_id_reference,
                    "activity": Reference(activity_id_reference, cls.activity_class.id),
                }
            )
        return super(ActivityBaseInfoMeta, cls).__new__(cls, name, bases, attrs)


class ActivityBaseInfo(with_metaclass(ActivityBaseInfoMeta, object)):
    """Describe and extends the functionality of an ActivityBase.

    ActivityBaseInfo should be subclassed to provide a custom database table
    (if necessary) and logic.  All ActivityBaseInfo classes must be registered
    with register_activity_info_type before being used.

    @cvar info_type: Must be specified to a unique integer. See
        register_activity_info_type.
    @cvar _data_class: The class used to instantiate the data attribute.
    @cvar data: An instance of _data_class. It's created with a reference to
        this class and its activity, plus any argument passed to the
        constructor.

    """

    info_type = None

    _data_class = None  # Can be set by subclasses to provide a data attribute
    _data = None

    def __init__(self, activity):
        self.activity = activity

    @property
    def access_delegate(self):
        return self.activity

    @classmethod
    def get_info_type(cls):
        raise NotImplementedError("Subclasses must implement get_info_type")

    @classmethod
    def load(cls, activity):
        if hasattr(cls, "__storm_table__"):
            return Store.of(activity).get(cls, activity.id)
        # Do the same job of our own constructor, but without calling the
        # class constructor, like Storm would do when loading previously
        # created objects.  Override this method to implement custom
        # behavior.
        info = cls.__new__(cls)
        info.activity = activity
        return info

    # This method isn't even present in the base class so that any
    # subclasses that run verifyObject() with the interface will break
    # up on the lack of this method.
    #
    # def get_summary(self):
    #    pass

    def get_state(self):
        """Return a dict with L{ActivityInfo} information."""
        return {"type": self.__class__.__name__, "summary": self.get_summary()}

    def _get_data(self):
        if self._data is None and self._data_class is not None:
            self._data = Store.of(self.activity).get(self._data_class, self.activity.id)
        return self._data

    def _set_data(self, data):
        self._data = data

    data = property(_get_data, _set_data)

    # Those are just placeholders for methods with the same name in
    # ComputerRequest, since only ComputerRequest activities can return a
    # non-False value for them.
    def is_approvable(self):
        """Can this activity be approved?"""
        return False

    def is_cancelable(self):
        """Can this activity be canceled?"""
        return False

    def is_actionable(self):
        """
        Is this item representing an activity which may be acted on?
        """
        return (
            self.is_approvable()
            or self.is_cancelable()
            or self.is_revertable()
            or self.is_reappliable()
        )

    def is_revertable(self):
        """Can this activity be reverted?"""
        return False

    def is_reappliable(self):
        """Can this activity be reapplied?"""
        return False

    def is_expandable(self):
        """Can this activity be expanded to show additional information?"""
        return IExpandableRequest.providedBy(self) and self.has_expanded_content()
