# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Collection logic to filter and find L{Activity}s."""

from storm.locals import Or, And

from canonical.collection import CollectionBase
from canonical.lib.undef import Undef

from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.entity.data import AccountEntityData
from canonical.landscape.model.activity.access import ActivityParentAccess


__all__ = [
    "ActivityChildCollection",
    "ActivityParentCollection",
    "get_child_activities",
]

__clearlist__ = ["ActivityChildCollection"]


class ActivityChildCollection(CollectionBase):
    """A collection of child Activities."""

    model_class = Activity

    def __init__(
        self,
        store=None,
        account=None,
        ids=None,
        parent_id=None,
        access_contexts=None,
        statuses=None,
        info_classes=None,
        entity_class=None,
        entity_ids=None,
    ):
        # Override the init so we can avoid explicitly passing the store
        # if the account is passed.
        if not store:
            if account:
                store = account.store
            else:
                raise RuntimeError("Must specify either the store or the account")

        super(ActivityChildCollection, self).__init__(
            store,
            account=account,
            ids=ids,
            parent_id=parent_id,
            access_contexts=access_contexts,
            statuses=statuses,
            info_classes=info_classes,
            entity_class=entity_class,
            entity_ids=entity_ids,
        )

    @property
    def access_context(self):
        """Delegate security checks on the account root access context."""
        account = self._where["account"]
        return account.access_context

    def with_account(self, account):
        """Filter Activities by account."""
        return self._filter_by(account=account)

    def with_ids(self, ids):
        """Filter Activities by database IDs."""
        return self._filter_by(ids=ids)

    def with_parent_id(self, parent_id):
        """Filter Activities by parent ID."""
        return self._filter_by(parent_id=parent_id)

    def with_access_contexts(self, access_contexts):
        """Filter by entity type and access context IDs.

        @param access_contexts: A dict mapping and entity type numeric
            code to a list of access contexts the user has a certain permission
            on (must be provided by calling code).
        """
        return self._filter_by(access_contexts=access_contexts)

    def with_statuses(self, statuses):
        """Filter Activities by status."""
        return self._filter_by(statuses=statuses)

    def with_info_classes(self, info_classes):
        """Filter Activities by info class."""
        return self._filter_by(info_classes=info_classes)

    def with_entities(self, entity_class, entity_ids=None):
        """Filter Activities by related entity class and optionally IDs.

        @param entity_class: the class of related entity to filter on.
        @param entity_ids: IDs of instances of the specified class.
        """
        return self._filter_by(entity_class=entity_class, entity_ids=entity_ids)

    def _get_result(
        self,
        store,
        where=None,
        account=Undef,
        ids=None,
        parent_id=None,
        access_contexts=None,
        statuses=None,
        info_classes=None,
        entity_class=None,
        entity_ids=None,
    ):
        """Get a ResultSet of Activities matching search criteria."""

        # Make sure the account parameter is set explicit, to minimize the risk
        # of forgetting about it.
        if account is Undef:
            raise RuntimeError(
                "You must explicitly set the 'account' filter to either an "
                "Account object or to None to search across all accounts in "
                "this store"
            )

        if entity_ids is not None and entity_class is None:
            raise RuntimeError(
                "Must specify the entity_class if entity_ids is specified"
            )

        if where is None:
            where = []

        if account is not None:
            where.append(Activity.account_id == account.id)

        if ids is not None:
            where.append(Activity.id.is_in(ids))

        if parent_id is not None:
            where.append(Activity.parent_id == parent_id)

        if access_contexts is not None:
            where.extend(
                [
                    Activity.entity_data_id == AccountEntityData.id,
                    self._get_access_contexts_clause(store, access_contexts),
                ]
            )

        if statuses is not None:
            where.append(Activity.status.is_in(statuses))

        if info_classes is not None:
            info_types = [cls.info_type for cls in info_classes]
            where.append(Activity.info_type.is_in(info_types))

        if entity_class is not None:
            where.append(Activity.entity_data_id == entity_class.data_id)
        if entity_ids is not None:
            where.append(entity_class.id.is_in(entity_ids))

        return store.find(self._select or Activity, *where)

    def _get_access_contexts_clause(self, store, access_contexts):
        """Return the where clause for the access_contexts filter.

        @param access_contexts: Same as with_access_contexts.
        """
        clauses = []
        for entity_type, access_context_ids in access_contexts.items():
            clauses.append(
                And(
                    AccountEntityData.entity_type == entity_type,
                    AccountEntityData.access_context_id.is_in(access_context_ids),
                )
            )

        return Or(*clauses)


class ActivityParentCollection(CollectionBase):
    """A collection of ParentActivities."""

    model_class = ActivityParent

    def __init__(
        self,
        store=None,
        account=None,
        ids=None,
        child_access_contexts=None,
        inprogress=None,
        queued=None,
        waiting=None,
        blocked=None,
        scheduled=None,
        succeeded=None,
        failed=None,
        canceled=None,
        info_classes=None,
    ):
        # Override the init so we can avoid explicitly passing the store
        # if the account is passed.
        if not store:
            if account:
                store = account.store
            else:
                raise RuntimeError("Must specify either the store or the account")
        super(ActivityParentCollection, self).__init__(
            store,
            account=account,
            ids=ids,
            child_access_contexts=child_access_contexts,
            inprogress=inprogress,
            queued=queued,
            waiting=waiting,
            blocked=blocked,
            scheduled=scheduled,
            succeeded=succeeded,
            failed=failed,
            canceled=canceled,
            info_classes=info_classes,
        )

    @property
    def access_context(self):
        """Delegate security checks on the account root access context."""
        account = self._where["account"]
        return account.access_context

    def with_account(self, account):
        """Filter ActivityParent's by account."""
        return self._filter_by(account=account)

    def with_ids(self, ids):
        """Filter ActivityParent's by database IDs."""
        return self._filter_by(ids=ids)

    def with_info_classes(self, info_classes):
        """Filter ActivityParent's info classes."""
        return self._filter_by(info_classes=info_classes)

    def with_child_status(
        self,
        inprogress=None,
        queued=None,
        waiting=None,
        blocked=None,
        scheduled=None,
        succeeded=None,
        failed=None,
        canceled=None,
    ):
        """Filter by child status counts.

        Parameters are 3-value flags (True/False/None) named after
        status names.

        If a flag is True, return only parents that have at least one child
        in the that state.

        If a flag is False return only parents that don't have any child in
        the that state.

        If a flag is None, it's ignored.
        """
        return self._filter_by(
            inprogress=inprogress,
            queued=queued,
            waiting=waiting,
            blocked=blocked,
            scheduled=scheduled,
            succeeded=succeeded,
            failed=failed,
            canceled=canceled,
        )

    def with_child_access_contexts(self, child_access_contexts):
        """Filter by entity type and access context IDs.

        Note that since parent activities are not directly linked to access
        context. The semantics is that a user has access to and ActivityParent
        if this has have one or more child linked to account entities with
        the user has access to.

        @param child_access_contexts: A dict mapping an entity type numeric
            code to a list of access context IDs the user has a certain
            permission on (must be provided by calling code).
        """
        return self._filter_by(child_access_contexts=child_access_contexts)

    def _get_result(
        self,
        store,
        where=None,
        account=Undef,
        ids=None,
        child_access_contexts=None,
        inprogress=None,
        queued=None,
        waiting=None,
        blocked=None,
        scheduled=None,
        succeeded=None,
        failed=None,
        canceled=None,
        info_classes=None,
    ):
        """Get a ResultSet of ActivityParent's matching search criteria."""

        # Make sure the account parameter is set explicit, to minimize the risk
        # of forgetting about it.
        if account is Undef:
            raise RuntimeError(
                "You must explicitly set the 'account' filter to either an "
                "Account object or to None to search across all accounts in "
                "this store"
            )

        if where is None:
            where = []

        if account is not None:
            where.append(ActivityParent.account_id == account.id)

        if ids is not None:
            where.append(ActivityParent.id.is_in(ids))

        if info_classes is not None:
            where.append(
                Or(
                    [
                        ActivityParent.info_type == info_class.get_info_type()
                        for info_class in info_classes
                    ]
                )
            )

        if child_access_contexts is not None:
            where.extend(
                self._get_child_access_contexts_clauses(store, child_access_contexts)
            )

        child_status_clauses = self._get_child_status_clauses(
            dict(
                inprogress=inprogress,
                queued=queued,
                waiting=waiting,
                blocked=blocked,
                scheduled=scheduled,
                succeeded=succeeded,
                failed=failed,
                canceled=canceled,
            )
        )
        if child_status_clauses:
            where.append(Or(child_status_clauses))

        return store.find(self._select or ActivityParent, *where)

    def _get_child_access_contexts_clauses(self, store, child_access_contexts):
        """Return the where clause for the child_access_contexts filter.

        @param child_access_contexts: Same as with_child_access_contexts.
        """
        iterator = iter(child_access_contexts.items())
        access_context_clauses = []
        for entity_type, access_context_ids in iterator:
            access_context_clauses.append(
                And(
                    ActivityParentAccess.entity_type == entity_type,
                    ActivityParentAccess.access_context_id.is_in(access_context_ids),
                )
            )
        return [
            ActivityParentAccess.parent_id == ActivityParent.id,
            Or(*access_context_clauses),
        ]

    def _get_child_status_clauses(self, statuses):
        """Return the where clauses matching the given status filters.

        @param statuses: A map from status name (e.g. "inprogress",
            "blocked") to a boolean, to indicate whether only parent
            activities with a number of children in this status greater
            than 0 should be returned.
        """
        clauses = []
        for status, flag in statuses.items():
            if flag is None:
                continue
            count_column = getattr(ActivityParent, status + "_count")
            if flag:
                clauses.append(count_column > 0)
            else:
                clauses.append(count_column == 0)
        return clauses


def get_child_activities(parent, **kwargs):
    """Return matching children of the specified ActivityParent.

    See ActivityChildCollection for available filters.
    """
    account = parent.account
    collection = ActivityChildCollection(
        account.store, account=account, parent_id=parent.id, **kwargs
    )
    return collection.find()
