# Copyright 2005-2013 Canonical Limited.  All rights reserved.

# HYSTERICAL RAISINS: There is a lot of logic in this module that involves
# children of ComputerRequests, but that logic is obsolete now that
# ComputerRequests can no longer have children. The logic is left only because
# updating all the unit tests to avoid exercising it would take a lot of
# effort.

import transaction

from zope.interface import implementer

from canonical.lib.timestamp import utcnow, format_datetime
from canonical.security.protect import requires, PRIVATE, requires_advisor

from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.computer.interfaces import IComputerAware
from canonical.landscape.model.activity.interfaces import (
    IRevertableRequest,
    IReappliableRequest,
)
from canonical.landscape.model.activity.activity import Activity, ActivityInfo
from canonical.landscape.model.activity.status import (
    QUEUED,
    BLOCKED,
    SUCCEEDED,
    FAILED,
    CANCELED,
    SCHEDULED,
    PENDING_STATUSES,
    ActivityStatusChangeError,
)
from canonical.landscape.model.activity.timeline import ActivityChildTimeline
from canonical.landscape.model.eventlog.management import EventLog


__all__ = ["ComputerRequest", "ComputerRequestMixin", "cancel_expired_activities"]

__clearlist__ = ["ComputerRequest", "ComputerRequestMixin"]

__requires_prefix__ = "landscape."


@implementer(IComputerAware)
@requires_advisor
class ComputerRequestMixin:
    """
    Mixin that provides computer-request related methods and attributes,
    without unnecessarily tying our type to ActivityInfo.

    Only mixes in with ActivityInfo and AbstractActivityInfo classes.
    """

    requires("ViewActivity", PRIVATE, ["computer"])

    @property
    def computer(self):
        if not hasattr(self, "_computer") or self._computer is None:
            from canonical.landscape.model.main.computer import Computer

            store = get_main_store()
            self._computer = store.get(Computer, self.activity.entity.id)

        return self._computer

    @computer.setter
    def computer(self, value):
        self._computer = value

    def get_state(self, v2=False):
        state = super().get_state()
        timeline = ActivityChildTimeline(self.activity)
        state.update(
            {
                "computer_id": self.computer.id,
                "approval_time": format_datetime(timeline.get_approval_time()),
                "delivery_time": format_datetime(timeline.get_delivery_time()),
            }
        )
        if not v2:
            # Legacy delivery timestamps
            state.update(
                {
                    "deliver_after_time": format_datetime(
                        self.activity.schedule_after_time
                    ),
                    "deliver_before_time": format_datetime(
                        self.activity.schedule_before_time
                    ),
                }
            )

        return state

    @requires("RevertActivity")
    def revert(self, parent):
        """
        Check revert preconditions. The logic should be implemented in
        subclasses.
        """
        if not IRevertableRequest.providedBy(self):
            raise ActivityStatusChangeError("Can't revert this activity")

        if self.activity.status is not SUCCEEDED:
            raise ActivityStatusChangeError(
                "Can't revert activity in status %s" % self.activity.status.name
            )

    @requires("CreateActivity")
    def reapply(self, parent):
        """Reapply this activity.

        @return: The newly created activity.
        """
        if not IReappliableRequest.providedBy(self):
            raise ActivityStatusChangeError("Can't reapply this activity")

        if self.activity.status not in (SUCCEEDED, FAILED, CANCELED):
            raise ActivityStatusChangeError(
                "Can't reapply activity in status %s" % self.activity.status.name
            )
        kwargs = {}
        for key, value in self.reapply_arguments.items():
            kwargs[key] = getattr(self, value)

        child = parent.add(self.__class__, computer=self.computer, **kwargs)

        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: %s" % child.info.get_summary()
        )
        return child

    def is_approvable(self):
        """
        Is this item representing a ComputerRequest which may be approved?
        """
        return self.activity.status == BLOCKED

    def is_cancelable(self):
        """Whether the ComputerRequest is cancelable."""
        return self.activity.status in PENDING_STATUSES

    def is_actionable(self):
        """
        Is this item representing a ComputerRequest which may be acted on?
        """
        return (
            self.is_approvable()
            or self.is_cancelable()
            or self.is_revertable()
            or self.is_reappliable()
        )

    def is_revertable(self):
        """
        Is this item representing a ComputerRequest which may be reverted?
        """
        return self.activity.status == SUCCEEDED and IRevertableRequest.providedBy(self)

    def is_reappliable(self):
        """
        Is this item representing a ComputerRequest which may be reapplied?
        """
        return (
            self.activity.status
            in (
                SUCCEEDED,
                FAILED,
                CANCELED,
            )
            and IReappliableRequest.providedBy(self)
        )

    def _canceled(self):
        """Hook called when an activity is CANCELED.

        It check account rules for the computer the activity refers to.
        """
        self._check_rules()

        from canonical.landscape.model.package.activities import (
            ChangePackageProfilesRequest,
        )

        if isinstance(self, ChangePackageProfilesRequest):
            from canonical.landscape.model.package.alerts import PackageProfilesAlert

            self.activity.account.check_rules(
                [PackageProfilesAlert], computer_ids=[self.activity.entity.id]
            )

        create_log_entry(self, "Activity canceled", use_person_from_parent_event=True)

    def _succeeded(self):
        """Hook called when an activity is SUCCEEDED."""
        create_log_entry(self, "Activity succeeded")

    def _failed(self):
        """Hook called when an activity is FAILED."""
        create_log_entry(self, "Activity failed")

    def _unblocked(self):
        """Hook called when the activity is unblocked."""
        create_log_entry(self, "Activity approved", use_person_from_parent_event=True)
        self._check_rules()

    def _blocked(self):
        """Hook called when the activity is BLOCKED."""
        self._check_rules()

    def _check_rules(self):
        """Check rules for the computer associated to the activity.."""
        from canonical.landscape.model.activity.alerts import UnapprovedActivitiesAlert

        result = self.activity.account.get_rules(
            info_classes=[UnapprovedActivitiesAlert]
        )
        for rule in result:
            rule.info.check([self.computer.id])


@requires_advisor
class ComputerRequest(ComputerRequestMixin, ActivityInfo):
    def __init__(self, activity, computer):
        super(ComputerRequest, self).__init__(activity)
        self._set_activity_entity(computer.get_account_twin())


def cancel_expired_activities():
    """
    Cancel all pending activities that are expired.
    """
    now = utcnow()
    for name, store in get_account_stores():
        result = store.find(
            Activity,
            Activity.status.is_in([QUEUED, SCHEDULED, BLOCKED]),
            Activity.schedule_before_time < now,
        )
        for activity in result:
            activity.cancel()
        transaction.commit()


def create_log_entry(info, summary_prefix, use_person_from_parent_event=False):
    """Create a log entry for the info type.

    @param info: a ComputerRequest info type.
    @param summary_prefix: a prefix to be used in the summary message.
    @param use_person_from_parent_event: If True, use the person performing the
      event instead of the activity.creator. Intended for Cancel and Approve
      child activity actions to track the person who performed those actions
      instead of the original activity creator.
    """
    creator = info.activity.creator
    management = EventLog(info.activity.account, info.activity.creator)
    if use_person_from_parent_event and info.activity.parent is not None:
        # Attempt to grab creator from the most recent ActivityGroup event
        parent_entity = info.activity.parent.info
        entry = management.get_entries(about=parent_entity).first()
        if entry and entry.person_id is not None:
            creator = get_person_by_id(entry.person_id)
            management.person = creator

    management.create_log_entry(
        info.computer, "%s: %s" % (summary_prefix, info.get_summary())
    )
