from builtins import object
from datetime import datetime

from canonical.storm.relation import RelationLink, Relation

from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.group import ActivityGroup
from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.entity.data import AccountEntityData
from canonical.landscape.model.account.store import get_account_store_by_name
from canonical.landscape.model.package.activities import ChangePackagesRequest
from canonical.landscape.model.activity.activities import ExecuteScriptRequest


ACTIVITY_PARENT_COLUMNS = (
    ActivityParent.id,
    ActivityParent.account_id,
    ActivityParent.creation_time,
    ActivityParent.info_type,
    ActivityParent.inprogress_count,
    ActivityParent.queued_count,
    ActivityParent.waiting_count,
    ActivityParent.blocked_count,
    ActivityParent.scheduled_count,
    ActivityParent.succeeded_count,
    ActivityParent.failed_count,
    ActivityParent.canceled_count,
)

ACTIVITY_CHILD_COLUMNS = (
    Activity.id,
    Activity.account_id,
    Activity.parent_id,
    Activity.creation_time,
    Activity.info_type,
    Activity.entity_data_id,
    Activity.status,
)

ACTIVITY_GROUP_COLUMNS = (
    ActivityGroup.activity_id,
    ActivityGroup.basic_summary,
    ActivityGroup.summary,
)

CHANGE_PACKAGES_REQUEST_COLUMNS = (
    ChangePackagesRequest.activity_id,
    ChangePackagesRequest.changes,
)

EXECUTE_SCRIPT_REQUEST_COLUMNS = (
    ExecuteScriptRequest.activity_id,
    ExecuteScriptRequest.interpreter,
    ExecuteScriptRequest.code,
    ExecuteScriptRequest.username,
    ExecuteScriptRequest.time_limit,
    ExecuteScriptRequest.title,
)


class ActivityProfile(object):
    """Create activities.

    This class is a helper to create a set of activities from a given
    profile specification expressed with dictionaries.

    Example:

    profile = ActivityProfile(
        resolver,
        activities=[
            {"account": "onward",
             "info_class": StubInfo,
             "children": [
                 {"info_class": StubInfoLeaf},
                 {"info_class": StubInfoLeaf},
             ]},
        ]
    )
    profile.create()

    will create a parent activity with two children. More complex setups are
    possible by passing additional info about the activities.

    See unit tests or code for more comprehensive examples.
    """

    def __init__(self, resolver, activities=None):
        self.resolver = resolver
        self.activities = activities or {}

    def create(self):
        """Create the activities for this profile."""
        factory = ActivityFactory(self.resolver)

        for activity in self.activities:
            parent_kwargs = activity.copy()
            account = parent_kwargs.pop("account")
            children_kwargs = parent_kwargs.pop("children")
            factory.create_activity(account, parent_kwargs, children_kwargs)

        factory.flush()


class ActivityFactory(object):
    """Helper to create activities and related objects."""

    def __init__(self, resolver):
        """
        @param resolver: The RelationResolve instance to use to manage
            relations.
        """
        self._resolver = resolver
        self._account = self._resolver.get(Account)
        self._account_entity_data = self._resolver.get(AccountEntityData)
        self._computer = self._resolver.get(Computer)
        self._activity_parent = Relation(ACTIVITY_PARENT_COLUMNS)
        self._activity_child = Relation(ACTIVITY_CHILD_COLUMNS)
        self._activity_group = Relation(ACTIVITY_GROUP_COLUMNS)
        self._change_packages_request = Relation(CHANGE_PACKAGES_REQUEST_COLUMNS)
        self._execute_script_request = Relation(EXECUTE_SCRIPT_REQUEST_COLUMNS)
        self._id = 0

    def create_activity(self, account, parent_kwargs, children_kwargs):
        """Create an activity with the given details."""
        creation_time = parent_kwargs.pop("creation_time", datetime.utcnow())
        children_counts = parent_kwargs.pop("children_counts", (0,) * 9)
        info_kwargs = parent_kwargs.copy()
        info_class = info_kwargs.pop("info_class")
        account_id_link = RelationLink(self._account, key=account)
        self._id += 1
        parent_id = self._id
        self._activity_parent.add(
            id=parent_id,
            account_id=account_id_link,
            creation_time=creation_time,
            info_type=info_class.get_info_type(),
            inprogress_count=children_counts[0],
            queued_count=children_counts[1],
            waiting_count=children_counts[2],
            blocked_count=children_counts[3],
            scheduled_count=children_counts[4],
            succeeded_count=children_counts[5],
            failed_count=children_counts[6],
            canceled_count=children_counts[7],
        )
        self._create_info(parent_id, info_class, info_kwargs)

        for child_kwargs in children_kwargs:
            self._id += 1
            info_kwargs = child_kwargs.copy()
            info_class = info_kwargs.pop("info_class")
            status = info_kwargs.pop("status", None)

            entity_data_id = None
            if issubclass(info_class, ComputerRequest):
                computer = info_kwargs.pop("computer")
                entity_data_id = RelationLink(self._account_entity_data, key=computer)

            self._activity_child.add(
                id=self._id,
                account_id=account_id_link,
                parent_id=parent_id,
                creation_time=creation_time,
                info_type=info_class.get_info_type(),
                entity_data_id=entity_data_id,
                status=status,
            )
            self._create_info(self._id, info_class, info_kwargs)

    def flush(self):
        """Flush all the data created with this factory to the database."""
        # At the moment we support only one account and resource database
        account_store = get_account_store_by_name("account-1")

        self._resolver.add(self._activity_parent, account_store)
        self._resolver.add(self._activity_child, account_store)
        self._resolver.add(self._activity_group, account_store)
        self._resolver.add(self._change_packages_request, account_store)
        self._resolver.add(self._execute_script_request, account_store)

        self._resolver.flush()

    def _create_info(self, activity_id, info_class, info_kwargs):
        """Create the info class data for the given activity."""
        table = getattr(info_class, "__storm_table__", None)
        if not table:
            return  # This is not a stored info class
        method = getattr(self, "_create_{}".format(table))
        method(activity_id, **info_kwargs)

    def _create_activity_group(self, activity_id, summary):
        """Create an ActivityGroup row."""
        self._activity_group.add(
            activity_id=activity_id, basic_summary=summary.id, summary=summary.summary
        )

    def _create_change_package_request(self, activity_id, **kwargs):
        """Create a ChangePackagesRequest row and associated data."""
        self._change_packages_request.add(activity_id=activity_id, **kwargs)

    def _create_execute_script_request(self, activity_id, **kwargs):
        """Create an ExecuteScriptRequest row and associated data."""
        self._execute_script_request.add(activity_id=activity_id, **kwargs)
