# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Unicode, Enum, Store

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.activity.status import SUCCEEDED, FAILED, CANCELED
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import ActivityParentInfo
from canonical.landscape.model.activity.collection import get_child_activities
from canonical.landscape.model.activity.schedule import (
    InvalidScheduleAfterWindowError,
    MAX_SCHEDULE_AFTER_WINDOW,
)
from canonical.landscape.model.eventlog.management import EventLog


__all__ = [
    "ActivityGroup",
    "HELLO_WORLD_SUMMARY",
    "ADD_USERS_SUMMARY",
    "DELETE_USERS_SUMMARY",
    "EDIT_USERS_SUMMARY",
    "LOCK_USERS_SUMMARY",
    "UNLOCK_USERS_SUMMARY",
    "ADD_TO_GROUPS_SUMMARY",
    "REMOVE_FROM_GROUPS_SUMMARY",
    "SIGNAL_PROCESSES_SUMMARY",
    "CHANGE_PACKAGES_SUMMARY",
    "VERIFY_UPGRADES_SUMMARY",
    "EXECUTE_SCRIPT_SUMMARY",
    "RESTART_SUMMARY",
    "SHUTDOWN_SUMMARY",
    "ACCEPT_COMPUTER_SUMMARY",
    "REJECT_COMPUTER_SUMMARY",
    "START_CLOUD_INSTANCES_SUMMARY",
    "STOP_CLOUD_INSTANCES_SUMMARY",
    "RESTART_CLOUD_INSTANCES_SUMMARY",
    "RELEASE_UPGRADE_SUMMARY",
    "APPLY_PACKAGE_PROFILE_SUMMARY",
    "APPLY_REPOSITORY_PROFILE_SUMMARY",
    "CHANGE_PACKAGE_HOLDS_SUMMARY",
    "TAGS_ADDED_SUMMARY",
    "TAGS_REMOVED_SUMMARY",
    "COMPUTERS_REMOVED_SUMMARY",
    "CHANGE_HA_CLUSTER_STANDBY_SUMMARY",
    "CHANGE_HA_CLUSTER_ONLINE_SUMMARY",
    "OPENSTACK_UPGRADE_SUMMARY",
    "RESYNCHRONIZE_SUMMARY",
    "CHANGE_REPOSITORY_POCKET_SUMMARY",
    "validate_delivery_time",
    "JUJU_DESTROY_ENVIRONMENT_SUMMARY",
    "CLEAN_JUJU_MODEL_SUMMARY",
    "CHANGE_SNAPS_SUMMARY",
    "START_CHILD_SUMMARY",
    "STOP_CHILD_SUMMARY",
    "DELETE_CHILD_SUMMARY",
]

__clearlist__ = [
    "ActivityGroup",
    "SIGNAL_PROCESSES_SUMMARY",
    "validate_delivery_time",
    "RESYNCHRONIZE_SUMMARY",
]

__requires_prefix__ = "landscape."


summaries = {}


class Summary(object):
    """An ActivityGroup summary.

    Maps a database-level ID to a nice English (ha) text string.
    """

    def __init__(self, id, summary):
        self.id = id
        self.summary = summary
        summaries[self] = self.id

    def __repr__(self):
        return f"Summary({self.id}, {self.summary})"


HELLO_WORLD_SUMMARY = Summary(1, "Hello, world!")
ADD_USERS_SUMMARY = Summary(2, "Adding user(s)")
DELETE_USERS_SUMMARY = Summary(3, "Deleting user(s)")
EDIT_USERS_SUMMARY = Summary(4, "Editing user(s)")
LOCK_USERS_SUMMARY = Summary(5, "Locking user(s)")
UNLOCK_USERS_SUMMARY = Summary(6, "Unlocking user(s)")
ADD_TO_GROUPS_SUMMARY = Summary(7, "Adding user(s) to group(s)")
REMOVE_FROM_GROUPS_SUMMARY = Summary(8, "Removing user(s) from group(s)")

SIGNAL_PROCESSES_SUMMARY = Summary(9, "Signalling processes")

CHANGE_PACKAGES_SUMMARY = Summary(10, "Install, hold and/or remove packages")
VERIFY_UPGRADES_SUMMARY = Summary(11, "Verify package upgrades")
EXECUTE_SCRIPT_SUMMARY = Summary(12, "Run script")

RESTART_SUMMARY = Summary(15, "Restart computer(s)")
SHUTDOWN_SUMMARY = Summary(16, "Shutdown computer(s)")

REJECT_COMPUTER_SUMMARY = Summary(17, "Reject computer(s)")
ACCEPT_COMPUTER_SUMMARY = Summary(18, "Accept computer(s)")

START_CLOUD_INSTANCES_SUMMARY = Summary(19, "Start new VM instances")
STOP_CLOUD_INSTANCES_SUMMARY = Summary(20, "Stop VM instances")
RESTART_CLOUD_INSTANCES_SUMMARY = Summary(21, "Restart VM instances")

RELEASE_UPGRADE_SUMMARY = Summary(22, "Perform release upgrade")
APPLY_PACKAGE_PROFILE_SUMMARY = Summary(23, "Apply package profile")
CHANGE_PACKAGE_HOLDS_SUMMARY = Summary(24, "Change package holds")

APPLY_REPOSITORY_PROFILE_SUMMARY = Summary(25, "Apply repository profiles")

CHANGE_PACKAGE_HOLDS_SUMMARY = Summary(26, "Change package holds")

TAGS_ADDED_SUMMARY = Summary(28, "Computer tags added")
TAGS_REMOVED_SUMMARY = Summary(29, "Computer tags removed")

COMPUTERS_REMOVED_SUMMARY = Summary(30, "Computers removed")

CHANGE_HA_CLUSTER_STANDBY_SUMMARY = Summary(
    31, "HA cluster participation set to standby"
)

CHANGE_HA_CLUSTER_ONLINE_SUMMARY = Summary(32, "HA cluster participation set to online")

OPENSTACK_UPGRADE_SUMMARY = Summary(33, "Change packages on OpenStack hosts")

# NOTE: at the moment an ActivityGroup with CHANGE_REPOSITORY_POCKET_SUMMARY
#       will have at most one child (a SyncPocketRequest), so the text of the
#       summary doesn't really matter, as it won't ever be used in practices.
#       See c.l.model.repository.management.RepositoryManagement.
CHANGE_REPOSITORY_POCKET_SUMMARY = Summary(34, "Change a repository pocket")

RESYNCHRONIZE_SUMMARY = Summary(35, "Resynchronize computer(s)")

# NOTE: at the moment an ActivityGroup with JUJU_DESTROY_ENVIRONMENT_SUMMARY
#       will have at most one child, so the text of the summary doesn't
#       really matter, as it won't ever be used in practice.
#       See c.l.service.openstack.regionremove
JUJU_DESTROY_ENVIRONMENT_SUMMARY = Summary(36, "Destroy Juju controller")
CLEAN_JUJU_MODEL_SUMMARY = Summary(
    37, "Release all machines and applications from externally managed juju model"
)

CHANGE_SNAPS_SUMMARY = Summary(38, "Change snaps")

START_CHILD_SUMMARY = Summary(39, "Starting child computer(s)")
STOP_CHILD_SUMMARY = Summary(40, "Stopping child computer(s)")
DELETE_CHILD_SUMMARY = Summary(41, "Deleting child computer(s)")


def validate_delivery_time(run_window_length=None, deliver_after_window=None):
    """
    Validate C{run_window_length} and C{deliver_after_window} for delivery

    Ensure C{run_window_length} expiry hours and C{deliver_after_window} delays
    are integers that will still allow for a valid activity delivery window.
    Raise an InvalidScheduleAfterWindowError exception.
    """
    if deliver_after_window is None:
        return

    try:
        deliver_after_window = int(deliver_after_window)
    except ValueError:
        raise InvalidScheduleAfterWindowError(
            f"The 'deliver delay window' of '{deliver_after_window}' "
            "must be a number greater than or equal to zero"
        )
    if deliver_after_window < 0:
        raise InvalidScheduleAfterWindowError(
            f"The 'deliver delay window' of '{deliver_after_window}' "
            "must be a number greater than or equal to zero"
        )

    if deliver_after_window > MAX_SCHEDULE_AFTER_WINDOW:
        raise InvalidScheduleAfterWindowError(
            f"The 'deliver delay window' of '{deliver_after_window}' minutes "
            "must be a number less than or equal to "
            f"{(MAX_SCHEDULE_AFTER_WINDOW // 24) // 60} days"
        )
    if run_window_length and run_window_length * 60 <= deliver_after_window:
        raise InvalidScheduleAfterWindowError(
            f"The 'deliver delay window' of '{deliver_after_window}' minutes "
            "should be shorter than the 'deliver within' expiration time of "
            f"{run_window_length} {'hours' if run_window_length > 1 else 'hour'}"
        )


@requires_advisor
class ActivityGroup(ActivityParentInfo):
    """A container for related L{Activity}s.

    When many activities are created as a result of a single user action, such
    as when a user requests upgrades on many computers, they are grouped
    together using an C{ActivityGroup}.

    @ivar basic_summary: The integer value of the L{Summary} for this group,
        used as fallback if the children have different summary.
    @ivar summary: The cached text representation of the summary. It's the same
        as all its children if they are identical, otherwise it's the linked
        value of C{basic_summary}
    """

    requires("ViewActivity", PRIVATE, ["summary"])

    __storm_table__ = "activity_group"

    basic_summary = Enum(map=summaries)
    summary = Unicode()

    def __init__(self, activity, summary):
        super(ActivityGroup, self).__init__(activity)
        self.basic_summary = summary
        self.summary = summary.summary

    @property
    def title(self):
        """Get a human-readable title.

        This property is here to make instances compatible with the event log
        machinery.
        """
        return self.get_summary()

    @property
    def id(self):
        """Get a unique ID (with respect to other instances of this type).

        This property is here to make instances compatible with the event log
        machinery.
        """
        return self.activity.id

    def get_summary(self):
        """Get a summary for this L{Activity}.

        @return: A human-readable summary for this L{Activity}.
        """
        return self.summary

    @requires("CreateActivity")
    def create_log_entry(self, prefix=None, message=None, person=None):
        """Create an event log entry with the provided prefix and person.

        @param prefix: Optionally the string to prefix the event log entry
          message with.
        @param message: Optionally, the event log message content. If not
          provided, the value from get_summary will be used.
        @param person: Optionally, the person attributed to the entry created.
          If not provided, the original activity creator will be used.
        """
        if message is None:
            message = self.get_summary()
        # It is policy for event log entries to have a trailing stop
        if message[-1] != ".":
            message = f"{message}."
        if prefix is not None:
            message = f"{prefix}: {message}"
        if person is None:
            person = self.activity.creator
        event_log = EventLog(self.activity.account, person)
        return event_log.create_log_entry(self, message)

    @requires("ViewActivity")
    def get_scheduled_time(self):
        """Return the first scheduled time for this activity."""
        store = Store.of(self)

        result = store.find(
            Activity.schedule_after_time, Activity.parent_id == self.activity.id
        )
        return result.min(Activity.schedule_after_time)

    @requires("CreateActivity")
    def clone(self, creator):
        """Create a new activity group with the same summary as this one.

        @note: This little wrapper around account.create_activity is needed so
            a restricted admin can create 'redo' and 'undo' activities, without
            needing to have the CreateActivity on the account access context.
        """
        return self.activity.account.create_activity(
            ActivityGroup, summary=self.basic_summary, creator=creator
        )

    def _settled(self):
        """Update the activity summary based on children summaries."""
        summary = None
        for child in get_child_activities(self.activity):
            if summary != self.basic_summary.summary:
                if summary is None:
                    summary = child.info.get_summary()
                elif child.info.get_summary() != summary:
                    summary = self.basic_summary.summary

        if summary is None:
            summary = self.basic_summary.summary
        self.summary = summary

    def _finished(self):
        status = self.activity.legacy_status
        if status == SUCCEEDED:
            self.create_log_entry(message="Group activity succeeded")
        elif status == FAILED:
            self.create_log_entry(message="Group activity failed")
        elif status == CANCELED:
            self.create_log_entry(message="Group activity canceled")
