# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.schema import Datetime, Int, Object, TextLine
from zope.interface import Interface, Attribute


__all__ = [
    "IActivity",
    "IActivityBaseInfo",
    "IActivityInfo",
    "IActivityParentInfo",
    "IRevertableRequest",
    "IReappliableRequest",
    "IExpandableRequest",
    "IActivityStatus",
    "IActivityCollection",
]

__clearlist__ = [
    "IRevertableRequest",
    "IReappliableRequest",
    "IExpandableRequest",
    "IActivity",
]


class IActivity(Interface):

    info_type = Int(
        title="Type of associated ActivityInfo",
        description="A unique integer representing the ActivityInfo type.",
        required=True,
        readonly=True,
    )

    creation_time = Datetime(
        title="Creation time",
        description="When this activity was created.",
        required=True,
        readonly=True,
    )

    creator_id = Int(
        title="Creator ID",
        description="ID of the person who created this activity",
        required=False,
        readonly=True,
    )


class IActivityBaseInfo(Interface):

    info_type = Int(
        title="ActivityInfo Type",
        description="A unique integer representing this ActivityInfo type.",
        required=True,
        readonly=True,
    )

    activity = Object(
        title="Activity",
        description="Activity object associated with this ActivityInfo.",
        required=True,
        readonly=True,
        schema=IActivity,
    )

    def get_summary():
        """Returns a human-readable textual summary of the activity."""


class IActivityInfo(IActivityBaseInfo):
    def get_text():
        """Returns textual information for the activity."""


class IActivityParentInfo(IActivityBaseInfo):
    pass


class IActivityStatus(Interface):

    id = Int(
        title="Unique identifier",
        description="A unique identifier for this status.",
        required=True,
        readonly=True,
    )

    name = TextLine(
        title="Name",
        description="A short lower-case-no-spaces name.",
        required=True,
        readonly=True,
    )

    title = TextLine(
        title="Title",
        description="A short sentence case name without punctuation.",
        required=True,
        readonly=True,
    )


class IRevertableRequest(Interface):
    """Represents a L{ComputerRequest} which can be reverted."""

    def revert(parent):
        """Create a new activity to revert the present request.

        @param parent: The parent activity the reverted activity should
            be added to.

        @return: The newly created activity.
        """


class IReappliableRequest(Interface):
    """Represent a L{ComputerRequest} which can be reapplied."""

    reapply_arguments = Attribute(
        "A mapping of instance attributes to __init__ arguments, used to "
        "reapply an activity."
    )

    def reapply(creator):
        """Create a new activity reapplying the present request.

        @param creator: Optionally, the C{Person} asking for the reapplied
            operation.

        @return: The newly created activity.
        """


class IExpandableRequest(Interface):
    """A request with additional content to provide to the view."""

    def has_expanded_content():
        """Return C{True} if the activity has expanded content."""

    def expanded_content():
        """Return the expanded content for that activity."""


class IActivityCollection(Interface):
    """
    A specific collection of activities that can have a summary page rendered
    for it.
    """

    def get_activities(**kwargs):
        """Retrieve activities to be used by the summary page."""
