# Copyright 2005-2014 Canonical Limited.  All rights reserved.

from builtins import str
import logging

from twisted.internet.defer import inlineCallbacks, returnValue

from storm.twisted.transact import transact

from canonical.amqp.job.runner import JobRunner
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.status import INPROGRESS


class ActivityJobRunner(JobRunner):
    """A JobRunner that is connected to an Activity.

    The job accepts the account id and activity id and the activity
    status will be updated automatically. When the job starts, the
    status is set to INPROGRESS. When it's finished successfully it is
    set to SUCCEEDED. If something goes wrong it's set to FAILED.

    The work that the job should do should be put into the _run_activity
    method.

    There's also a method for updating the activity progress,
    self._advance.
    """

    # Whether to automatically mark the activity as succeeded if there are
    # no failures. Sub-classes can set it to False if they don't want this
    # behavior.
    _auto_succeed = True

    @inlineCallbacks
    def run(self, account_id, activity_id):
        """Update the activity status and run the job."""
        yield self._start(account_id, activity_id)
        try:
            yield self._run_activity(account_id, activity_id)
        except Exception as e:
            logging.exception("Failed to execute job: %s" % e)
            yield self._fail(account_id, activity_id, str(e))
        else:
            if self._auto_succeed:
                yield self._succeed(account_id, activity_id)
        returnValue((None, None))

    def _run_activity(self, account_id, activity_id):
        """Run the parts that should be done while INPROGRESS."""
        raise NotImplementedError(self._run_activity)

    def _get_activity(self, account_id, activity_id):
        """Return the L{Activity} with the given details."""
        account = get_account_by_id(account_id)
        return get_child_activity(account, activity_id)

    @transact
    def _start(self, account_id, activity_id):
        """Mark the activity as started."""
        activity = self._get_activity(account_id, activity_id)
        if activity.status != INPROGRESS:
            # The activity could be already INPROGRESS if it wast started by a
            # job that didn't complete because of a crash.
            activity.start()

    @transact
    def _advance(self, account_id, activity_id, progress):
        """Advance the activity to the given progress."""
        activity = self._get_activity(account_id, activity_id)
        activity.advance(progress)

    @transact
    def _fail(self, account_id, activity_id, error_message):
        """Mark the activity as failed."""
        activity = self._get_activity(account_id, activity_id)
        activity.fail(text=error_message)

    @transact
    def _succeed(self, account_id, activity_id):
        """Mark the activity as succeeded."""
        activity = self._get_activity(account_id, activity_id)
        activity.succeed()
