# Copyright 2005-2016 Canonical Limited.  All rights reserved.

"""
The code in this method has been factored out of other parts of activity
modules since it will be eventually dropped, once the activity consolidation
work is done (see http://bit.ly/1gRZ0Db for details) and we update both the
API and UI activity-related functionality to better match the new system (in
particular this means finishing the work around version 2 of the Landscape API
and dropping version 1 after a deprecation period).

This helps cleaning out the core models and logic, and makes it easier to find
call sites that still need to be updated to the new infrastructure.
"""

from zope.security.proxy import removeSecurityProxy

from canonical.search.engine import CriteriaBase

from canonical.landscape.model.entity.data import AccountEntityData
from canonical.landscape.model.entity.helper import get_account_entity_type
from canonical.landscape.model.activity.search import (
    ActivityChildSearchParser,
    ActivityChildSearch,
    ActivityParentSearchParser,
    ActivityParentSearch,
)
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import activity_parent_type_search_terms

__all__ = [
    "MixedActivityParentSearch",
    "MixedActivityChildSearch",
    "get_activity_child_entity_ids",
]

__clearlist__ = ["get_activity_child_entity_ids"]

__requires_prefix__ = "landscape."


class MixedActivityChildSearchParser(ActivityChildSearchParser):
    """A child activity search parser accepting also child-specific terms.

    See MixedActivityChildSearch.
    """

    def _get_type_criteria(self, term):
        # If the given info type is a parent one, we don't match any child.
        if term in activity_parent_type_search_terms:
            return FalseCriteria()
        return super(MixedActivityChildSearchParser, self)._get_type_criteria(term)


class MixedActivityChildSearch(ActivityChildSearch):
    """A child search that also deals with parent-specific terms.

    This search accepts the same terms as ActivityChildSearch, but
    will also handle values that are child-specific.

    It's used to support the GetActivities semantics of the Landscape
    API version 1, which returns a mix of parent and child activities.
    """

    parser_class = MixedActivityChildSearchParser


class MixedActivityParentSearchParser(ActivityParentSearchParser):
    """A parent activity search parser accepting also child-specific terms.

    See MixedActivityParentSearch.
    """

    def _get_parent_id_criteria(self, term):
        # If the parent_id criteria is given we don't want to match any parent,
        # since parent_id is a child-specific term that can match only
        # children.
        return FalseCriteria()

    def _get_computer_criteria(self, term):
        # If the computer criteria is given we don't want to match any parent,
        # since computer is a child-specific term that can match only children.
        return FalseCriteria()

    def _get_type_criteria(self, term):
        # If the given info type is a child one, we don't match any parent.
        if term not in activity_parent_type_search_terms:
            return FalseCriteria()
        return super(MixedActivityParentSearchParser, self)._get_type_criteria(term)


class MixedActivityParentSearch(ActivityParentSearch):
    """A parent search that also deals with child-specific terms.

    This search accepts the same terms as ActivityChildSearch, but
    will also handle the ones that are child-specific.

    It's used to support the GetActivities semantics of the Landscape
    API version 1, which returns a mix of parent and child activities.
    """

    parser_class = MixedActivityParentSearchParser


class FalseCriteria(CriteriaBase):
    """A criteria that always returns a False clause."""

    def get_where_clause(self):
        return False


def get_activity_child_entity_ids(parent, entity_type):
    """Get entity IDs relating to children of the specified parent."""
    entity_type_id = get_account_entity_type(entity_type)

    result = removeSecurityProxy(parent)._get_children(entity_type.id)
    result = result.find(
        Activity.parent_id == parent.id,
        Activity.entity_data_id == AccountEntityData.id,
        AccountEntityData.entity_type == entity_type_id,
        AccountEntityData.id == entity_type.data_id,
    )
    return result.config(distinct=True)
