# Copyright 2005-2022 Canonical Limited. All rights reserved.

from random import randint
from collections import defaultdict
from datetime import timedelta

from zope.interface import implementer
from zope.component import getUtility

from storm.locals import Store, Int, Count, Asc, Desc, ClassAlias, Reference
from storm.expr import Max

from canonical.lib.timestamp import utcnow, format_datetime
from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.amqp.message.interfaces import IDispatcher

from canonical.landscape.model.main.interfaces import (
    IAccessDelegateAware,
    IAccessContextAware,
)
from canonical.landscape.model.activity.interfaces import IActivity, IActivityParentInfo
from canonical.landscape.model.activity.base import (
    ActivityBase,
    ActivityBaseInfoMeta,
    ActivityBaseInfo,
)
from canonical.landscape.model.activity.activity import (
    Activity,
    ActivityOrdering,
    reverse_activity_info_types,
)
from canonical.landscape.model.activity.status import (
    INPROGRESS,
    SUCCEEDED,
    FAILED,
    WAITING,
    CANCELED,
    BLOCKED,
    SCHEDULED,
    QUEUED,
    STATUS_PRIORITY,
    FINAL_STATUSES,
    PENDING_STATUSES,
)
from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError
from canonical.landscape.model.activity.timeline import (
    ActivityChildEvent,
    EVENT_STATUS_TRANSITION,
)
from canonical.landscape.model.activity.access import update_activity_parent_access
from future.utils import with_metaclass


__all__ = [
    "ActivityParent",
    "ActivityParentInfo",
    "activity_parent_info_types",
    "register_activity_parent_info_type",
    "get_parent_ids_for_entities",
    "update_parent_activities_after_changed_children",
    "get_parent_activity",
    "STATUS_COUNTS_COLUMNS",
    "activity_parent_type_search_terms",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


# Map status IDs to the corresponding count column in ActivityParent.
STATUS_COUNTS_COLUMNS = {
    INPROGRESS.id: "inprogress_count",
    QUEUED.id: "queued_count",
    WAITING.id: "waiting_count",
    BLOCKED.id: "blocked_count",
    SCHEDULED.id: "scheduled_count",
    SUCCEEDED.id: "succeeded_count",
    FAILED.id: "failed_count",
    CANCELED.id: "canceled_count",
}


@requires_advisor
@implementer(IActivity, IAccessContextAware)
class ActivityParent(ActivityBase):
    """A parent for activity children.

    It's used to group together related activities.

    Parent activities shouldn't be created using the ParentActivity
    class directly, instead Account.create_activity() should be used.

    All child activities will have the same creator and creation data as
    the parent activity.

    Child activities are added using the add() method:

        child = parent.add(SomeInfoClass, info_class_param_foo="bar")

    When all the children have been added, the parent should be settled
    to allow the children activities to start:

        parent.settle()

    @ivar schedule_after_window: If non-zero, the schedule_after_time parameter
        of child activities will be offset an additional random time of up to
        schedule_after_window minutes in the future, so they won't be
        delivered immediately but rather spread over this time window.
    """

    requires("ViewActivity", "CreateActivity", ["attachments"])
    requires(
        "ViewActivity",
        PRIVATE,
        [
            "id",
            "account_id",
            "account",
            "access_context",
            "info",
            "info_type",
            "creator",
            "creation_time",
            "children",
            "children_count",
            "legacy_status",
            "schedule_after_window",
            "succeeded_count",
            "failed_count",
            "canceled_count",
            "blocked_count",
            "unapproved_count",
        ],
    )

    __storm_table__ = "activity_parent"

    progress = Int(allow_none=False, default=0)

    # Number of children in the various statuses.
    inprogress_count = Int(allow_none=False, default=0)
    queued_count = Int(allow_none=False, default=0)
    waiting_count = Int(allow_none=False, default=0)
    blocked_count = Int(allow_none=False, default=0)
    scheduled_count = Int(allow_none=False, default=0)
    succeeded_count = Int(allow_none=False, default=0)
    failed_count = Int(allow_none=False, default=0)
    canceled_count = Int(allow_none=False, default=0)

    schedule_after_window = Int(default=0)

    _settled = False
    _access_context = None

    # Parent activities with less than this number of children will have their
    # progress/status updated synchronously whenever a child's status/progress
    # changes. Parent activities bigger than this number will be updated with a
    # job, in order to avoid database contention and serialization errors.
    _SYNCHRONOUS_UPDATES_THRESHOLD = 50

    def __init__(
        self,
        account,
        info_class,
        creator=None,
        access_context=None,
        **info_class_kwargs
    ):
        super(ActivityParent, self).__init__(
            account, info_class, creator=creator, **info_class_kwargs
        )
        assert issubclass(
            info_class, ActivityParentInfo
        ), "info_class not intended for parent activities"
        self._access_context = access_context

    def __repr__(self):
        return "<ActivityParent {}>".format(self.info)

    def _get_info(self):
        if self._info is None:
            self._info = activity_parent_info_types[self.info_type].load(self)
        return self._info

    def _set_info(self, info):
        self._info = info
        self.info_type = info.info_type

    info = property(_get_info, _set_info)

    @property
    def access_context(self):
        """Return the AccessContext of the activity.

        This defaults to the access context of the account the activity belongs
        to, or any access context passed during creation.
        """
        return self._access_context or self.account.access_context

    @property
    def namespace(self):
        """Namespace for AMQP notifications related to the ActivityParent."""
        return "{}.activity.{}".format(self.account.namespace, self.id)

    @property
    def completion_time(self):
        """The time at which activity has completed.

        This is the latest completion time of children, if all children have
        completed, None otherwise.

        XXX This is a legacy methods used in the API, it should eventually go
        away. As it is, it might be expensive when a lot of parents are
        involved.
        """
        if self.is_finished():
            children_ids = self._get_children(Activity.id)
            return (
                Store.of(self)
                .find(
                    Max(ActivityChildEvent.timestamp),
                    ActivityChildEvent.activity_id.is_in(children_ids),
                    ActivityChildEvent.event_type == EVENT_STATUS_TRANSITION,
                )
                .one()
            )

    @requires("ViewActivity")
    def get_state(self, v2=False):
        state = super(ActivityParent, self).get_state(v2=v2)
        state["completion_time"] = format_datetime(self.completion_time)

        if v2:
            state["schedule_after_window"] = self.schedule_after_window
        else:
            state.update(
                {
                    "parent_id": None,
                    "deliver_delay_window": self.schedule_after_window,
                    "result_text": None,
                    "result_code": None,
                }
            )

            status = self.legacy_status
            if status is not None:
                state["activity_status"] = status.name

        return state

    @requires("CreateActivity")
    def add(self, info_class, entity=None, **info_class_kwargs):
        """Add a child activity.

        @param info_class: The ActivityInfo subclass of the child.
        @param info_class_kwargs: Keyword arguments to pass to the info
            class.
        @return: The newly added child Activity.
        """
        if self._settled:
            raise RuntimeError("Can't add children, the activity is settled")

        assert not issubclass(
            info_class, ActivityParentInfo
        ), "Info class intended for parents, not for children."
        return Activity(info_class, self, entity=entity, **info_class_kwargs)

    @requires("CreateActivity")
    def settle(self):
        """
        Settle the activity (it no longer accepts adding children), and set
        children statuses.
        """
        if self._settled:
            raise RuntimeError("Already settled")

        for child in self._get_children():

            if child.status is not None:
                # Skip children which already have a status. These are either
                # explicitly set (scheduled and blocked) children, or they have
                # been queued from a previous call to settle()
                continue

            if child._has_preceding():
                child.wait()
            else:
                child.queue()

        self._settled = True
        update_activity_parent_access(Store.of(self), [self.id])
        self.info._settled()

    def alter(self):
        """Allow further addition of children to the parent activity."""
        if not self._settled:
            raise RuntimeError("Not settled")

        self._settled = False

    def abort(self):
        """Cancel all pending children activities."""
        for child in self._get_children():
            if child.status in PENDING_STATUSES:
                child.cancel()

    @requires("ScheduleActivity")
    def schedule(self, after=None, after_window=0, before=None):
        """Schedule child activities based on the specified times.

        @param after: an optional time activities should be executed after.
        @param after_window: if specified, a number of minutes a random delay
            for activities start time will be chosen in.
        @param before: an optional time activities should be executed before.
        """
        if self._settled:
            raise RuntimeError("Can't schedule, the activity is settled")

        if not (after or after_window or before):
            raise RuntimeError(
                "At least one of 'after', 'after_window' and 'before' must be "
                "provided."
            )

        if after_window > 0:
            if after is None:
                after = utcnow()

            max_delta = timedelta(seconds=after_window * 60)

            if before is not None and (after + max_delta) >= before:
                raise InvalidScheduleAfterWindowError(
                    "Activity has a delay window that is later than "
                    "the activity's expiration time."
                )

        for child in self._get_children():
            if child.status is not None:
                continue

            child_after = after
            if after_window > 0:
                # add the random initial delay
                child_after += timedelta(seconds=randint(1, after_window * 60))
            child.schedule(after=child_after, before=before)

        # Set this late so it doesn't get set in case of errors
        self.schedule_after_window = after_window

    @requires("CreateActivity")
    def add_attachment(self, attachment):
        """Add a script attachment."""
        self.attachments.add(attachment)

    def children(
        self,
        offset=0,
        limit=25,
        ascending=True,
        status=None,
        info_class=None,
        status_sort=None,
    ):
        """Return child Activities matching the given criteria.

        @param offset: Return children starting from this point.
        @param limit: Return at most this number of children.
        @param ascending: By default child Activities are sorted by ascending
            status (i.e. most active activities first). Setting this to
            False will reverse the order and sort by descending status.
        @param status: If given, return only children in this status.
        @param info_class: Optionally, only include activities of the specified
            class.
        @param status_sort: An optional custom sort expression for the status.
        """
        # XXX TODO: raise an error if limit exceeds a certain level (e.g. 100).
        if status_sort is None:
            status_sort = Asc if ascending else Desc
        result = self._get_children()

        if status is not None:
            result = result.find(Activity.status == status)
        if info_class is not None:
            info_type = reverse_activity_info_types[info_class]
            result = result.find(Activity.info_type == info_type)

        result.order_by(status_sort(Activity.status), Activity.id)
        result.config(limit=limit, offset=offset)

        return list(result)

    @requires("ViewActivity")
    def is_finished(self):
        """Whether all children are in a final status."""
        statuses = self.children_status()
        finished_count = sum(statuses[status] for status in FINAL_STATUSES)
        return finished_count == self.children_count

    @requires("ViewActivity")
    def has_failed_children(self):
        """Whether the activity has failed children."""
        return self.failed_count > 0

    def has_pending_children(self):
        """Whether the activity has pending children."""
        return self.pending_count > 0

    def has_canceled_children(self):
        """Whether the activity has canceled children."""
        return self.canceled_count > 0

    def has_blocked_children(self):
        """Whether the activity has blocked children."""
        return self.blocked_count > 0

    def has_inprogress_children(self):
        """Whether the activity has inprogess children."""
        return self.inprogress_count > 0

    def has_queued_children(self):
        """Whether the activity has queued children."""
        return self.queued_count > 0

    @property
    def pending_count(self):
        """Return the number of pending children."""
        statuses = self.children_status()
        return sum(statuses[status] for status in PENDING_STATUSES)

    @property
    def children_count(self):
        """Return the total number of children."""
        return sum(self._get_status_count(status) for status in STATUS_PRIORITY)

    def children_status(self):
        """Return the number of children in each possible status."""
        return {status: self._get_status_count(status) for status in STATUS_PRIORITY}

    @property
    def legacy_status(self):
        """Return the legacy overall status of the parent activity.

        This is here to help transition away from ActivityParent.status.  It
        should not be used in new code, and will be eventually dropped as
        parent activities should not have their own status.
        """
        for status in STATUS_PRIORITY:
            if self._get_status_count(status) > 0:
                return status

    def _schedule_update(self):
        """Schedule a job for updating the activity if needed.

        This method checks if the parent is big enough to need to be updated
        asynchronously to avoid database contention. If so fires a job to
        perform the update some time in the future.

        The job will have the 'accumulate' flag on, so it won't be executed
        immediately, but rather wait a little bit in case other children change
        concurrently and queue the same job.

        @return: a boolean value indicating whether a job has been scheduled
        """
        if self.children_count < self._SYNCHRONOUS_UPDATES_THRESHOLD:
            return False

        # The bigger the activity the longer we want to wait before executing
        # the job, since lots of other updates may arrive in the meantime and
        # updating the parent is more costly (more children to process). We
        # put a cap of 1 minute so updates aren't delayed too much though.
        accumulate_delay = min(self.children_count * 0.05, 60)

        dispatcher = getUtility(IDispatcher)
        dispatcher.create_job(
            None,
            self.namespace,
            "activity-update-parent",
            accumulate=True,
            accumulate_delay=accumulate_delay,
            account_id=self.account_id,
            activity_id=self.id,
        )

        return True

    def _increment_status_count(self, status):
        """Increment by one the count attribute for the given status.

        @param status: The Activity.status identifying the counter to
            increment.
        """
        count = self._get_status_count(status) + 1
        self._set_status_count(status, count)

    def _update_status_counts(self):
        """Update the status counts with fresh values from the database."""
        status_counts = self._get_status_counts()
        for status in STATUS_PRIORITY:
            count = status_counts.get(status, 0)
            self._set_status_count(status, count)

    def _get_status_counts(self):
        """Return a dict mapping children statuses to their count.

        @return: A dict of the form Activity.status -> count, which for each
            possible status tells how many children are in that status.
        """
        from canonical.landscape.model.activity.activity import Activity

        find_spec = (Activity.status, Count(Activity.id))
        result = self._get_children(find_spec=find_spec)
        return dict(result.group_by(Activity.status))

    def _get_status_count(self, status):
        """Return the value of the count attribute for the given status."""
        return getattr(self, self._get_status_count_attribute(status))

    def _set_status_count(self, status, count):
        """Set the value of the count attribute for the given status."""
        return setattr(self, self._get_status_count_attribute(status), count)

    def _get_status_count_attribute(self, status):
        """Return the name of the count attribute for the given status."""
        return STATUS_COUNTS_COLUMNS[status.id]

    def _update_after_child_progress_change(self):
        """Updates the progress of a parent activity after a child has changed.

        The progress of this parent activity will be recalculated based on the
        progress of the children.
        """
        progress = self._get_children().avg(Activity.progress)
        self.progress = progress

        # Check if we can unblock some child that was waiting for this one to
        # progress.
        self._check_waiting_children()

        self._notify_changes()

    def _update_after_child_status_change(self):
        """Updates children status counts after child status has changed.

        If the parent activity is finished, the _finished hook is called.
        """
        self._update_status_counts()

        # Allow the info object to react to the change in child status.
        self.info._child_status_changed()

        if self.is_finished():
            self.info._finished()

        self._notify_changes()

    def _notify_changes(self):
        """Send a change notification to this activity's namespace.

        This is used by the UI to know that the progress or status of this
        activity has changed and refresh it via an AJAX call.
        """
        payload = {"id": self.id}
        dispatcher = getUtility(IDispatcher)
        dispatcher.send_notification(payload, None, "activity-changed", self.namespace)

    def _check_waiting_children(self):
        """Check whether some child that is waiting can be queued.

        This should be checked after some of the child activities
        change their progress. If all preceding activities have
        reached the minimum progress, the waiting activity will be queued
        for delivery.
        """
        store = Store.of(self)
        PrecedingActivity = ClassAlias(Activity)
        result = store.find(
            (
                Activity.id,
                PrecedingActivity.progress,
                ActivityOrdering.minimum_progress,
            ),
            Activity.parent_id == self.id,
            Activity.id == ActivityOrdering.activity_id,
            Activity.status == WAITING,
            ActivityOrdering.preceding_activity_id == PrecedingActivity.id,
        )

        preceding_min_progress_state = defaultdict(list)
        for activity_id, preceding_progress, minimum_progress in result:
            preceding_min_progress_state[activity_id].append(
                preceding_progress >= minimum_progress
            )

        queueable_activity_ids = [
            activity_id
            for activity_id in preceding_min_progress_state
            if all(preceding_min_progress_state[activity_id])
        ]

        result = self._get_children()
        for activity in result.find(Activity.id.is_in(queueable_activity_ids)):
            activity.queue()

    def _get_children(self, find_spec=None):
        """Return a L{ResultSet} of the child activities of this activity.

        @param find_spec: The find spec to use for the ResultSet, by default
            it's Activity.
        """
        if find_spec is None:
            find_spec = Activity
        return Store.of(self).find(find_spec, Activity.parent_id == self.id)


Activity.parent = Reference(Activity.parent_id, ActivityParent.id)


class ActivityParentInfoMeta(ActivityBaseInfoMeta):
    """Metaclass for the ActivityParentInfo class."""

    activity_class = ActivityParent


@requires_advisor
@implementer(IActivityParentInfo, IAccessDelegateAware)
class ActivityParentInfo(with_metaclass(ActivityParentInfoMeta, ActivityBaseInfo)):
    """
    An object which describes and extends the functionality of an
    ActivityParent.


    See ActivityInfo for more documentation.
    """

    # We'll set the permission of get_summary here even though we don't define
    # it locally, so that subclasses don't have the burden.
    requires(
        "ViewActivity",
        PRIVATE,
        [
            "get_summary",
            "activity",
            "data",
            "get_state",
            "is_approvable",
            "is_cancelable",
            "is_actionable",
            "is_revertable",
            "is_reappliable",
            "is_expandable",
        ],
    )

    @classmethod
    def get_info_type(cls):
        return reverse_activity_parent_info_types[cls]

    def _settled(self):
        """Hook called after the parent activity is settled."""

    def _child_status_changed(self):
        """Hook called after any child activity's status has changed."""

    def _finished(self):
        """Hook called after the parent activity is finished."""


activity_parent_info_types = {}
reverse_activity_parent_info_types = {}
activity_parent_type_search_terms = {}


def register_activity_parent_info_type(info_type, info_class):
    """Register info_class as an ActivityInfo implementation.

    The info_type will be stored as ActivityParent.info_type for any
    activities created with info_class.

    @param info_type: A unique integer representing info_class.
    @param info_class: An ActivityParentInfo-based class.
    @raise RuntimeError: Raised if a class is already registered for
        info_type.
    """
    existing = activity_parent_info_types.get(info_type)
    if existing is not None:
        raise RuntimeError("%r has the same info_type of %r" % (info_class, existing))
    activity_parent_info_types[info_type] = info_class
    reverse_activity_parent_info_types[info_class] = info_type
    activity_parent_type_search_terms[info_class.__name__] = info_type
    info_class.info_type = info_type


def unregister_activity_parent_info_type(info_class):
    """Unregister info_class as an ActivityParentInfo implementation.

    Removes info_class from the registry of known
    ActivityParent.info_type and cleans up extra attributes.

    @param info_class: An ActivityParentInfo-based class.
    """
    info_type = info_class.info_type
    if info_type is None:
        # When an info_class is registered, it's info_type attribute
        # is initialized, so if it hasn't been initialized then we've
        # been called to erroneously called to unregister it.
        raise RuntimeError("Parent info class is not registered: %r" % info_class)
    if info_type in activity_parent_info_types:
        del activity_parent_info_types[info_type]
        del reverse_activity_parent_info_types[info_class]
        del activity_parent_type_search_terms[info_class.__name__]
    info_class.info_type = None


def get_parent_ids_for_entities(store, entity_class, entity_ids):
    """Return IDs of parent activities related to an entity collection.

    @param store: the store to search entities in.
    @param entity_class: the class of the AccountEntity IDs refer to.
    @param entity_ids: IDs for the entities that the returned parent IDs
        should be associated with. A parent activity is associated with an
        entity if it has at least one child related to it (via
        Activity.entity_data_id).
    """
    from canonical.landscape.model.activity.collection import ActivityChildCollection

    collection = ActivityChildCollection(store)
    collection = collection.with_entities(entity_class, entity_ids)
    result = collection.find()
    result = result.find(Activity.entity_data_id == entity_class.data_id)
    result.config(distinct=True)
    return result.values(Activity.parent_id)


def update_parent_activities_after_changed_children(account, parent_ids):
    """Update parent activities after some children have been changed.

    This should be called when children activities have been removed or changed
    access context.
    This updates counts and progress for parent activities, and removed parents
    that are left with no children.
    """
    parent_ids = set(parent_ids)
    if not parent_ids:
        return

    store = account.store

    result = store.execute(
        """
        UPDATE activity_parent
        SET progress = child.avg_progress
        FROM (
          SELECT parent_id, AVG(progress) AS avg_progress
          FROM activity
          WHERE parent_id IN ?
          GROUP BY parent_id) AS child
        WHERE id = child.parent_id
        RETURNING id
        """,
        [tuple(parent_ids)],
    )

    # Parents that match the query still have children.
    parent_ids_with_children = set(row[0] for row in result)
    if parent_ids_with_children:
        parent_ids_with_children_tuple = tuple(parent_ids_with_children)
        for status_id, column in list(STATUS_COUNTS_COLUMNS.items()):
            result = store.execute(
                """
                UPDATE activity_parent
                SET {} = parent_child_counts.count
                FROM (
                  SELECT activity_parent.id, count(activity.id)
                  FROM activity_parent LEFT OUTER JOIN activity
                    ON activity_parent.id = activity.parent_id
                      AND activity.status = ?
                      AND activity.parent_id IN ?
                  GROUP BY activity_parent.id) AS parent_child_counts
                WHERE activity_parent.id = parent_child_counts.id
                  AND activity_parent.id IN ?
                """.format(
                    column
                ),
                [
                    status_id,
                    parent_ids_with_children_tuple,
                    parent_ids_with_children_tuple,
                ],
            )

    # Also update the access cache since children have been removed.
    update_activity_parent_access(store, parent_ids_with_children)

    parent_ids_to_remove = parent_ids - parent_ids_with_children
    if parent_ids_to_remove:
        store.find(
            ActivityParent, ActivityParent.id.is_in(parent_ids_to_remove)
        ).remove()


def get_parent_activity(account, activity_id):
    """Return the ActivityParent with the specified ID in the account."""
    from canonical.landscape.model.activity.collection import ActivityParentCollection

    collection = ActivityParentCollection(
        account.store, account=account, ids=[activity_id]
    )
    result = collection.find()
    return result.one()
