# Copyright 2019 Canonical Limited. All rights reserved.

from datetime import timedelta
import logging

from storm.expr import Eq
import transaction

from canonical.lib.timestamp import utcnow

from canonical.landscape.model.main.account import get_all_accounts
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import (
    ActivityParent,
    STATUS_COUNTS_COLUMNS,
)
from canonical.landscape.model.activity.status import PENDING_STATUSES


def cleanup_activities(older_than_days, dry_run=False):
    """Remove finished L{ActivityParent} objects with creation date older than
    C{older_than_days} days. Child L{Activity} objects are automatically
    cascade-deleted.

    @param older_than_days: Remove activities older than these many days.
    @param dry_run: True if we should only identify and count the activities
        that are to be removed.
    """
    expiration_time = utcnow() - timedelta(days=older_than_days)
    deleted_parents_count = 0
    deleted_children_count = 0

    # This is equivalent to:
    #  [Eq(ActivityParent.blocked_count, 0),
    #   Eq(ActivityParent.inprogress_count, 0),
    #   Eq(ActivityParent.queued_count, 0),
    #   Eq(ActivityParent.scheduled_count, 0),
    #   Eq(ActivityParent.waiting_count, 0)]
    is_finished_expr = [
        Eq(getattr(ActivityParent, STATUS_COUNTS_COLUMNS[status.id]), 0)
        for status in PENDING_STATUSES
    ]

    for account in get_all_accounts():
        parents = account.store.find(
            ActivityParent,
            ActivityParent.account_id == account.id,
            ActivityParent.creation_time < expiration_time,
            *is_finished_expr
        )
        parents_count = parents.count()
        if parents_count > 0:
            deleted_parents_count += parents_count
            parent_ids = tuple(parents.values(ActivityParent.id))
            children_count = account.store.find(
                Activity, Activity.parent_id.is_in(parent_ids)
            ).count()
            deleted_children_count += children_count
            if not dry_run:
                with transaction.manager as manager:
                    logging.info(
                        "Removing %d parent activities and %d child "
                        "activities for account '%s'.",
                        parents_count,
                        children_count,
                        account.name,
                    )
                    try:
                        parents.remove()
                    except Exception:
                        logging.exception("Error removing activities.")
                        manager.abort()
            else:
                logging.info(
                    "Would remove %d parent activities and %d child "
                    "activities for account '%s'.",
                    parents_count,
                    children_count,
                    account.name,
                )

    if deleted_parents_count == 0:
        logging.info("No activities older than %d days found.", older_than_days)
    elif not dry_run:
        logging.info(
            "Removed %d parent activities and %d child activities "
            "older than %d days.",
            deleted_parents_count,
            deleted_children_count,
            older_than_days,
        )
