# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import List, Unicode

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.activity.group import (
    ActivityGroup,
    RESYNCHRONIZE_SUMMARY,
)


__all__ = ["ResynchronizeRequest", "resynchronize_computers"]

__clearlist__ = ["ResynchronizeRequest"]

__requires_prefix__ = "landscape."


@requires_advisor
class ResynchronizeRequest(ComputerRequest):
    """
    An activity which represents a request from the server to
    resynchronise data between client and server.

    @ivar scopes: a list of scopes the resynchronisation request will pertain
          to on the client.  If no scopes are included then the
          resynchronisation request will be considered to be global.

    """

    __storm_table__ = "resynchronize_request"

    scopes = List(type=Unicode())

    requires("ViewActivity", PRIVATE, ["scopes"])

    def __init__(self, activity, computer, scopes=None, **kwargs):
        self.scopes = scopes

        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(ResynchronizeRequest, self).__init__(activity, computer, **kwargs)

    def _started(self):
        """Queue the resync message when the activity is started."""
        message_store = self.computer.get_message_store()
        message_store.add(
            {
                "type": "resynchronize",
                "operation-id": self.activity.id,
                "scopes": self.scopes,
            }
        )

    def get_summary(self):
        summary = "Resynchronize computer %s" % self.computer.title
        if self.scopes:
            summary += " (scopes: %s)" % ", ".join(self.scopes)
        return summary


def resynchronize_computers(account, computer_ids, scopes=None):
    """Resynchronize the computers with the given IDs."""

    # Note that we don't pass the person on purpose because for now this
    # operation is supposed to be run only by operators and we don't
    # want operator names to show up as creators in the user accounts.
    activity = account.create_activity(ActivityGroup, summary=RESYNCHRONIZE_SUMMARY)

    # XXX This doesn't scale, but for now we don't support
    # resynchronization at scale (as it's a hosted-specific action
    # that can be performed only by Canonical employees with the Operator
    # role).
    for computer in account.get_computers(computer_ids=computer_ids):
        computer.request_resynchronize(activity, scopes=scopes)
    activity.settle()
    return activity
