# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from datetime import datetime

from zope.datetime import parse, SyntaxError, DateError, DateTimeError, TimeError

from storm.expr import Desc, Eq, Ge, Le

from canonical.lib.undef import Undef
from canonical.lib.text import describe_list

from canonical.landscape.model.search import ModelIDCriteria
from canonical.landscape.model.activity.activity import (
    Activity,
    activity_type_search_terms,
)
from canonical.landscape.model.activity.parent import (
    STATUS_COUNTS_COLUMNS,
    ActivityParent,
    activity_parent_type_search_terms,
)
from canonical.search.engine import (
    SearchEngine,
    SearchParser,
    SearchCriteriaError,
    CriteriaBase,
)
from canonical.landscape.model.activity.status import STATUS_MAP
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.person import get_person_by_email
from canonical.landscape.model.main.errors import UnknownPersonsError
from canonical.landscape.model.computer.search import ComputerSearch


__all__ = [
    "ActivityChildSearch",
    "ActivityChildSearchParser",
    "ActivityParentSearchParser",
    "ActivityParentSearch",
    "get_search_help",
]


__clearlist__ = ["get_search_help"]


class ActivityCreatedAfterCriteria(CriteriaBase):
    """Criteria matching activities created after a given timestamp."""

    criteria_prefix = "created-after"
    criteria_attribute = "_created_after"

    _model = Undef()

    _term = Undef()

    def __init__(self, created_after):
        self._created_after = created_after

    @classmethod
    def get_search_help(cls):
        return {
            "terms": ["created-after:<time>"],
            "descriptions": [
                "{} created after the specified timestamp, in "
                "YYYY-MM-DDTHH:MM:SS format.".format(cls._term)
            ],
        }

    def get_where_clause(self):
        """Find the activity created after the given time.

        @return: A Storm expression that will match the specified time.
        """
        return Ge(self._model.creation_time, self._created_after)


class ActivityCreatedBeforeCriteria(CriteriaBase):
    """Criteria matching activities created before a given timestamp."""

    criteria_prefix = "created-before"
    criteria_attribute = "_created_before"

    _model = Undef()

    _term = Undef()

    def __init__(self, created_before):
        self._created_before = created_before

    @classmethod
    def get_search_help(cls):
        return {
            "terms": ["created-before:<time>"],
            "descriptions": [
                "{} created before the specified timestamp, in "
                "YYYY-MM-DDTHH:MM:SS format.".format(cls._term)
            ],
        }

    def get_where_clause(self):
        """Find the activity created before the given time.

        @return: A Storm expression that will match the specified time.
        """
        return Le(self._model.creation_time, self._created_before)


class ActivityTypeCriteria(CriteriaBase):

    criteria_prefix = "type"
    criteria_attribute = "info_class_name"

    _model = Undef()

    _term = Undef()

    # Mapping between info class names and their IDs.
    _info_classes = Undef()

    def __init__(self, info_class_name):
        self._info_class_name = info_class_name

    @classmethod
    def get_search_help(cls):
        return {
            "terms": ["type:<type>"],
            "descriptions": [
                "{} with the specified type. Must be one "
                "of {}".format(
                    cls._term.capitalize(),
                    describe_list(
                        items=list(cls._info_classes().keys()), conjunction="or"
                    ),
                )
            ],
        }

    def get_where_clause(self):
        """Find model instances which are of the specified info_type.

        @return: A Storm expression that will match model instances with the
            specified info type.
        """
        info_type = self._info_classes().get(self._info_class_name)
        if info_type is None:
            raise SearchCriteriaError(
                "'{}' is not a valid type for {}.".format(
                    self._info_class_name, self._term
                )
            )
        return Eq(self._model.info_type, info_type)


class ActivityChildSearchParser(SearchParser):
    """Parser for search critesria for child activities."""

    def __init__(self, query, account):
        super(ActivityChildSearchParser, self).__init__(query)
        self._account = account

    def _get_id_criteria(self, activity_id):
        """Parse the term looking for an activity id.

        @raises SearchCriteriaError: Raised if the term can't be parsed to an
            integer.
        @return: An ActivityChildIDCriteria instance.
        """
        try:
            activity_id = int(activity_id)
        except ValueError:
            raise SearchCriteriaError("'%s' is an invalid activity ID." % activity_id)
        return ActivityChildIDCriteria(activity_id)

    def _get_parent_id_criteria(self, activity_id):
        """Parse the term looking for an activity id.

        @raises SearchCriteriaError: Raised if the term can't be parsed to an
            integer.
        @return: An ActivityChildParentIDCriteria instance.
        """
        try:
            activity_id = int(activity_id)
        except ValueError:
            raise SearchCriteriaError("'%s' is an invalid activity ID." % activity_id)
        return ActivityChildParentIDCriteria(activity_id)

    def _get_status_criteria(self, term):
        """Parse the term looking for an activity status.

        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            known L{ActivityStatus} name.
        @return: An ActivityChildStatusCriteria instance.
        """
        # XXX allow to use the old unapproved status, which is now mapped to
        # blocked.
        if term == "unapproved":
            term = "blocked"

        for status in STATUS_MAP:
            if status.name == term:
                return ActivityChildStatusCriteria(status)
        raise SearchCriteriaError("'%s' is an invalid status value." % term)

    def _parse_datetime_term(self, term):
        """Parse a term in a datetime object."""
        try:
            return datetime(*parse(term)[:6])
        except (SyntaxError, DateError, TimeError, DateTimeError, TypeError):
            raise SearchCriteriaError("'%s' is an invalid creation time value." % term)

    def _get_created_after_criteria(self, term):
        """Parse the term looking for a date.

        @return: An ActivityChildCreatedAfterCriteria instance.
        """
        created_after = self._parse_datetime_term(term)
        return ActivityChildCreatedAfterCriteria(created_after)

    def _get_created_before_criteria(self, term):
        """Parse the term looking for a date.

        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            datetime.
        @return: An ActivityChildCreatedAfterCriteria instance.
        """
        created_before = self._parse_datetime_term(term)
        return ActivityChildCreatedBeforeCriteria(created_before)

    def _get_computer_criteria(self, term):
        """Forward the term to the L{ComputerSearch} mechanism.

        @return: An ActivityChildComputerCriteria instance.
        """
        return ActivityChildComputerCriteria(self._account, term)

    def _get_creator_criteria(self, term):
        """Parse the term looking for a creator email.

        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            known email.
        @return: An ActivityChildCreatorCriteria instance.
        """
        try:
            [person] = self._account.get_persons(emails=[term])
        except UnknownPersonsError:
            raise SearchCriteriaError(
                "'%s' is not a valid email for this account" % term
            )
        return ActivityChildCreatorCriteria(person.id)

    def _get_type_criteria(self, term):
        """Parse the term looking for an activity type name.

        @return: An ActivityChildTypeCriteria instance.
        """
        return ActivityChildTypeCriteria(term)


class ActivityChildIDCriteria(ModelIDCriteria):
    """Criteria that matches child activities by ID."""

    _model = Activity

    _term = "Activity"


class ActivityChildParentIDCriteria(CriteriaBase):

    criteria_prefix = "parent-id"
    criteria_attribute = "_parent_id"

    @staticmethod
    def get_search_help():
        return {
            "terms": ["parent_id:<activity_id>"],
            "descriptions": ["Activities that are children of the specified parent."],
        }

    def __init__(self, parent_id):
        self._parent_id = parent_id

    def get_where_clause(self):
        """Find the activities which are children of the given id.

        @return: A Storm expression that will match the children of the
            specified id.
        """
        return Eq(Activity.parent_id, self._parent_id)


class ActivityChildStatusCriteria(CriteriaBase):
    """A criteria to filter activities by status.

    @param status: an ActivityStatus instance.
    """

    criteria_prefix = "status"
    criteria_attribute = "_status"

    @staticmethod
    def get_search_help():
        return {
            "terms": ["status:<status>"],
            "descriptions": [
                "Activities that have the given status. Must be one of "
                "'unapproved', 'undelivered', 'delivered', 'canceled', "
                "'failed', 'succeeded', 'waiting', 'blocked' or 'scheduled'."
            ],
        }

    def __init__(self, status):
        self._status = status

    def get_where_clause(self):
        """Find the activity with the given status.

        @return: A Storm expression that will match the specified status.
        """
        return Eq(Activity.status, self._status.id)


class ActivityChildCreatedAfterCriteria(ActivityCreatedAfterCriteria):
    """Filter child activities with creation time after the specified one."""

    _model = Activity

    _term = "Activities"


class ActivityChildCreatedBeforeCriteria(ActivityCreatedBeforeCriteria):
    """Filter child activities with creation time before the specified one."""

    _model = Activity

    _term = "Activities"


class ActivityChildCreatorCriteria(CriteriaBase):

    criteria_prefix = "creator"
    criteria_attribute = "_creator_id"

    @staticmethod
    def get_search_help():
        return {
            "terms": ["creator:<email>"],
            "descriptions": [
                "Activities created by the administrator with the specified " "email."
            ],
        }

    def __init__(self, creator_id):
        self._creator_id = creator_id

    def get_where_clause(self):
        """Find the activities which are created by the specified creator_id.

        @return: A Storm expression that will match activities with the
            specified creator id.
        """
        return Eq(Activity.creator_id, self._creator_id)


class ActivityChildComputerCriteria(CriteriaBase):
    """A criteria to search activities by computers.

    @param account_id: The ID of the account in which the search happens,
        forwarded to L{ComputerSearch}.
    @param term: The search term for the computer query.
    """

    criteria_prefix = "computer"
    criteria_attribute = "_term"

    @staticmethod
    def get_search_help():
        return {
            "terms": ["computer:<computer query>"],
            "descriptions": [
                "Activities that are associated with any computers that "
                "match the specified computer query. e.g., "
                "computer:tag:laptop"
            ],
        }

    def __init__(self, account, term):
        self._account = account
        self._term = term

    def get_where_clause(self):
        """Find the activity where the computer matched the associated query.

        @return: A Storm expression that will match the specified computer
            search.
        """
        # Unfortunately, this logic means that we'll do one query in the main
        # database for each terms which mention computers in the activity
        # search. This can't be avoided for now because otherwise we lose
        # correctness between with operator priority. There are tests ensuring
        # that, like test_search_computer_composed_mixed.
        search = ComputerSearch(self._account)
        result = search.find(self._term)
        computer_ids = result.values(Computer.id)
        return (
            AccountComputerTwin.data_id == Activity.entity_data_id,
            AccountComputerTwin.id.is_in(computer_ids),
        )


class ActivityChildTypeCriteria(ActivityTypeCriteria):

    _model = Activity

    _term = "activities"

    @classmethod
    def _info_classes(cls):
        # Allow both parent and child info types in the search
        info_classes = activity_type_search_terms.copy()
        info_classes.update(activity_parent_type_search_terms)
        return info_classes


class ActivityChildSearch(SearchEngine):
    """A search engine for finding child Activities.

    Words provided as search parameters are treated as keywords.  Selector
    prefixes can be used to further customize the search:

     * id - Search for L{Activity}s with the specified id.
     * OR - Search for L{Activity}s matching term A or term B. OR must be in
            all caps.
     * NOT - Search for L{Activity}s not matching the next term. NOT must be in
            all caps.
    """

    model = Activity
    parser_class = ActivityChildSearchParser

    def __init__(self, account, filter_result=True):
        super(ActivityChildSearch, self).__init__(filter_result=filter_result)
        self.account = account

    def get_where_clause(self, search_query):
        where = super(ActivityChildSearch, self).get_where_clause(search_query)
        where.append(Activity.account_id == self.account.id)
        return where

    def _get_search_store(self):
        return self.account.store

    def filter_result(self, result):
        """Sort activities by creation time."""
        result.order_by(Desc(Activity.creation_time))
        result.config(distinct=True)
        return result

    def _build_parser(self, query):
        return self.parser_class(query, self.account)


class ActivityParentSearchParser(SearchParser):
    """Parser for search criteria for parent activities."""

    def _get_id_criteria(self, id):
        """Parse the term looking for a parent activity id.

        @raises SearchCriteriaError: Raised if the term can't be parsed to an
            integer.
        @return: An ActivityParentIDCriteria instance.
        """
        try:
            id = int(id)
        except ValueError:
            raise SearchCriteriaError(
                "'{}' is an invalid parent activity ID.".format(id)
            )
        return ActivityParentIDCriteria(id)

    def _get_created_after_criteria(self, term):
        """Parse the term looking for a date.

        @return: An ActivityParentCreatedAfterCriteria instance.
        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            datetime.
        """
        created_after = self._parse_datetime_term(term)
        return ActivityParentCreatedAfterCriteria(created_after)

    def _get_created_before_criteria(self, term):
        """Parse the term looking for a date.

        @return: An ActivityParenbtCreatedAfterCriteria instance.
        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            datetime.
        """
        created_before = self._parse_datetime_term(term)
        return ActivityParentCreatedBeforeCriteria(created_before)

    def _get_creator_criteria(self, term):
        """Parse the term looking for a creator email.

        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            known email.
        @return: An ActivityParentCreatorCriteria instance.
        """
        return ActivityParentCreatorCriteria(term)

    def _get_type_criteria(self, term):
        """Parse the term looking for an activity parent type name.

        @raises SearchCriteriaError: Raised if the term doesn't match a known
            activity parent type.
        @return: An ActivityParentTypeCriteria instance.
        """
        return ActivityParentTypeCriteria(term)

    def _get_status_criteria(self, term):
        """Parse the term looking for an activity status.

        @raises SearchCriteriaError: Raised if the term can't be parsed to a
            known ActivityStatus name.
        @return: An ActivityParentStatusCriteria instance.
        """
        return ActivityParentStatusCriteria(term)

    def _parse_datetime_term(self, term):
        """Parse a term in a datetime object."""
        try:
            return datetime(*parse(term)[:6])
        except (SyntaxError, DateError, TimeError, DateTimeError, TypeError):
            raise SearchCriteriaError("'%s' is an invalid creation time value." % term)


class ActivityParentSearch(SearchEngine):
    """A search engine for finding parent Activities.

    Words provided as search parameters are treated as keywords.  Selector
    prefixes can be used to further customize the search:

     * id - Search for ActivityParents with the specified id.
     * OR - Search for ActivityParents matching term A or term B. OR must be in
            all caps.
     * NOT - Search for ActivityParents not matching the next term. NOT must be
             in all caps.
    """

    parser_class = ActivityParentSearchParser


class ActivityParentIDCriteria(ModelIDCriteria):
    """Match the ActivityParent with the given id."""

    _model = ActivityParent

    _term = "Activity parent"


class ActivityParentCreatedAfterCriteria(ActivityCreatedAfterCriteria):
    """Filter parent activities with creation time after the specified one."""

    _model = ActivityParent

    _term = "Parent activities"


class ActivityParentCreatedBeforeCriteria(ActivityCreatedBeforeCriteria):
    """Filter parent activities with creation time before the specified one."""

    _model = ActivityParent

    _term = "Parent activities"


class ActivityParentCreatorCriteria(CriteriaBase):
    """Filter parent activities by creator email."""

    criteria_prefix = "creator"
    criteria_attribute = "creator_email"

    def __init__(self, creator_email):
        self.creator_email = creator_email

    @staticmethod
    def get_search_help():
        return {
            "terms": ["creator:<email>"],
            "descriptions": [
                "Parent activities created by the administrator with the "
                "specified email."
            ],
        }

    def get_where_clause(self):
        """Find the activities which are created by the specified creator_id.

        @return: A Storm expression that will match activities with the
            specified creator id.
        """
        person = get_person_by_email(self.creator_email)
        if person is None:
            raise SearchCriteriaError("'%s' is not a valid email" % self.creator_email)
        return Eq(ActivityParent.creator_id, person.id)


class ActivityParentTypeCriteria(ActivityTypeCriteria):
    """Filter parent activities by info type."""

    _model = ActivityParent

    _term = "parent activities"

    @classmethod
    def _info_classes(cls):
        return activity_parent_type_search_terms


class ActivityParentStatusCriteria(CriteriaBase):
    """Filter parent activities by status count greater than zero."""

    criteria_prefix = "status"
    criteria_attribute = "status_name"

    def __init__(self, status_name):
        self.status_name = status_name

    @staticmethod
    def get_search_help():
        return {
            "terms": ["status:<status>"],
            "descriptions": [
                "Parent activities with one or more children in the given "
                "status. Must be one of 'unapproved', 'undelivered', "
                "'delivered', 'canceled', 'failed', 'succeeded', 'waiting', "
                "'blocked' or 'scheduled'."
            ],
        }

    def get_where_clause(self):
        # XXX allow to use the old unapproved status, which is now mapped to
        # blocked.
        status_name = self.status_name
        if status_name == "unapproved":
            status_name = "blocked"

        for status in STATUS_MAP:
            if status.name == status_name:
                column_name = STATUS_COUNTS_COLUMNS[status.id]
                return getattr(ActivityParent, column_name) >= 1
        raise SearchCriteriaError(
            "'{}' is an invalid status value.".format(self.status_name)
        )


def get_search_help():
    """
    Return a list of tuples with available search terms and their
    descriptions for help documentation purposes.
    """
    # The list of criteria_classes must be kept up-to-date with all of the
    # criteria used by ActivityChildSearchParser. Each criteria class needs a
    # search_help attribute.
    criteria_classes = [
        ActivityChildIDCriteria,
        ActivityChildParentIDCriteria,
        ActivityChildStatusCriteria,
        ActivityChildCreatedAfterCriteria,
        ActivityChildCreatedBeforeCriteria,
        ActivityChildCreatorCriteria,
        ActivityChildComputerCriteria,
        ActivityChildTypeCriteria,
    ]
    return [cls.get_search_help() for cls in criteria_classes]
