import json

from zope.interface import implementer

from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.activity.activity import AbstractActivityInfo
from canonical.landscape.model.activity.computer import ComputerRequestMixin
from canonical.landscape.model.activity.interfaces import IExpandableRequest

__requires_prefix__ = "landscape."


@implementer(IExpandableRequest)
@requires_advisor
class SnapActivityBase(ComputerRequestMixin, AbstractActivityInfo):
    """
    A common parent class for all snap-related activities.

    Because snap activities take the same form: an activity type and a
    list of snaps of form
    {"name": "eg", "args": {"channel": "...", "revision": "...", "classic": False}}
    (with optional args), subclasses need only define a class-level `summary`
    and `message_type`.
    """

    def __init__(self, activity, computer, snaps):
        super().__init__(snaps, activity)

        self._set_activity_entity(computer.get_account_twin())

    def __init_subclass__(cls):
        assert hasattr(cls, "summary") and hasattr(
            cls, "message_type"
        ), "SnapActivity classes must define both `summary` and `message_type`"
        assert isinstance(cls.summary, str), "SnapActivity `summary` must be a str"
        assert isinstance(
            cls.message_type, str
        ), "SnapActivity `message_type` must be a str"

        return super().__init_subclass__()

    @requires("ViewActivity")
    def expanded_content(self):
        return [
            {
                "package": s,
                "type": self.message_type.split("-")[0],
                "packageType": "snap",
            }
            for s in self.data
        ]

    @requires("ViewActivity")
    def has_expanded_content(self):
        return True

    def get_summary(self):
        return self.summary

    def _started(self):
        self.computer.get_message_store().add(
            {
                "type": self.message_type,
                "snaps": self.data,
                "operation-id": self.activity.id,
            }
        )


class HoldSnaps(SnapActivityBase):
    """Hold a list of snaps on a computer."""

    summary = "Hold snaps on a computer"
    message_type = "hold-snaps"


class InstallSnaps(SnapActivityBase):
    """Install a list of snaps on a computer."""

    summary = "Install snaps on computer"
    message_type = "install-snaps"


class RefreshSnaps(SnapActivityBase):
    """Refresh a list of snaps on a computer"""

    summary = "Refresh snaps on computer"
    message_type = "refresh-snaps"


class RemoveSnaps(SnapActivityBase):
    """Remove a list of snaps from a computer."""

    summary = "Remove snaps on computer"
    message_type = "remove-snaps"


class UnholdSnaps(SnapActivityBase):
    """Unhold a list of snaps on a computer."""

    summary = "Unhold snaps on a computer"
    message_type = "unhold-snaps"
