import pickle as pickle

from collections import defaultdict

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.accesscontext import AccessContext
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.entity.data import AccountEntityData

PARENT_COLUMNS = (
    ActivityParent.account_id,
    ActivityParent.id,
    ActivityParent.creation_time,
    ActivityParent.inprogress_count,
    ActivityParent.queued_count,
    ActivityParent.waiting_count,
    ActivityParent.blocked_count,
    ActivityParent.scheduled_count,
    ActivityParent.succeeded_count,
    ActivityParent.failed_count,
    ActivityParent.canceled_count,
)

CHILD_COLUMNS = (
    Activity.parent_id,
    AccountEntityData.entity_type,
    AccountEntityData.id,
    AccountEntityData.access_context_id,
)


class ActivityStatistics(object):
    """Collect, dump and load activity statitics.

    This class collects all activities from the database and arrange them in a
    format that can be serialized and deserialized for benchmark purposes (e.g.
    for reproducing the data distribution of a certain deployment and running
    benchmarks against it).

    @ivar activities: A dict mapping account store names to lists of parent
        activities for each account. Each list of parent activity is a dict
        with a 'data' field with information about the parent and a 'children'
        field with information about the children.

        The 'data' field is a tuple with the following values:

          (creation_time, inprogress_count, queued_count, waiting_count, ...)

        see the PARENT_COLUMNS variable for the rest of the counts.

        The 'children' field is a list with a 3-tuple for each child:

          (entity_type, entity_index, access_context_index)

        where entity_type is the entity type of the entity associated with the
        child activity, entity_index is the index indentifying the entity of
        that type in the account (for example if entity_type is 2 entity_index
        is one, then the entity is the second computer in the account), and
        access_context_index is the index of the associated access context (0
        for the root access context, 1 for the first child access context,
        etc).
    """

    def __init__(self):
        # Account activities by account store.
        self.activities = {}

    def collect(self):
        """Collect data from all available activities in the account stores."""
        access_contexts = self._get_access_contexts()
        for name, store in get_account_stores():
            parents = self._get_parents(store)
            children = self._get_children(store)
            self.activities[name] = self._build_activities(
                access_contexts, parents, children
            )

    def dump(self, fd):
        """Write serialized statistics to the given file descriptor."""
        # Create a pickler using the binary format (HIGHEST_PROTOCOL), as
        # by default the text format is used, which is quite expensive in
        # terms of disk size and CPU.
        pickler = pickle.Pickler(fd, protocol=pickle.HIGHEST_PROTOCOL)
        pickler.dump(self.activities)

    def load(self, fd):
        """Read serialized statistics from the given file descriptor."""
        self.activities = pickle.load(fd)

    def _get_access_contexts(self):
        """Get a dict mapping each account ID to its access context IDs."""
        store = get_main_store()
        access_contexts = defaultdict(list)
        result = store.find(
            (Account.id, AccessContext.id),
            AccessContext.root_id == Account.access_context_id,
        )
        for account_id, access_context_id in result.order_by(AccessContext.id):
            access_contexts[account_id].append(access_context_id)
        return access_contexts

    def _get_parents(self, store):
        """Get a dict mapping each account ID to its parent activities."""
        parents = defaultdict(list)
        result = store.find(PARENT_COLUMNS)
        for row in result:
            account_id = row[0]
            parents[account_id].append(row[1:])
        return parents

    def _get_children(self, store):
        """Get a dict mapping each parent ID to its child activities."""
        children = defaultdict(list)
        result = store.find(
            CHILD_COLUMNS, AccountEntityData.id == Activity.entity_data_id
        )
        for row in result:
            parent_id = row[0]
            children[parent_id].append(row[1:])
        return children

    def _build_activities(self, access_contexts, parents, children):
        """Build a serializable data structure with parents and children."""
        activities = []

        while parents:
            account_id, account_parents = parents.popitem()
            account_access_contexts = access_contexts.pop(account_id)
            account_activities = []
            account_entities = {}  # Account entities by type

            for parent in account_parents:
                parent_id = parent[0]
                try:
                    parent_children = children.pop(parent_id)
                except KeyError:
                    # XXX Solve bug of parents without children
                    continue

                for i, child in enumerate(parent_children):
                    # Replace the entity ID and access context ID with a simple
                    # index relative to the number of entities of the given
                    # type and to the access contexts (for example if the
                    # account has 2 access context and this child's access
                    # context ID matches the second access context, then the
                    # index will be 1).
                    entity_type, entity_id, access_context_id = child
                    entities = account_entities.setdefault(entity_type, [])
                    if entity_id not in entities:
                        entities.append(entity_id)  # New entity

                    entity_index = entities.index(entity_id)
                    access_context_index = account_access_contexts.index(
                        access_context_id
                    )

                    parent_children[i] = (
                        entity_type,
                        entity_index,
                        access_context_index,
                    )

                    if sum(parent[2:]) != len(parent_children):
                        # XXX Solve bug of inconsistent children count
                        continue

                account_activities.append(
                    {
                        "data": parent[1:],
                        "children": parent_children,
                    }
                )

            activities.append(account_activities)

        return activities
