# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Activity life-cycle.

An activity follows the following life-cycle, where some of the statuses are
optional (meaning that an activity could skip them)::

  - SCHEDULED: the activity will kick off sometime in the future (optional)
  - BLOCKED: the activity is blocked on some not yet met condition (optional)
  - WAITING: the activity is waiting on preceding activities (optional)
  - QUEUED: the activity is about to be executed (optional)
  - INPROGRESS: the activity is happening but didn't complete yet (optional)
  - SUCCEEDED: the activity has succeeded, no further change will happen
  - FAILED: the activity has failed, no further change will happen
  - CANCELED: the activity has been canceled, no further change will happen

It's possible for an activity to transition back from the INPROGRESS status to
the WAITING status (for example a package activity needing approval) and
re-start from there.
"""

from zope.interface import implementer

from canonical.security.protect import requires, requires_advisor, PRIVATE, PUBLIC

from canonical.landscape.model.activity.interfaces import IActivityStatus
from canonical.landscape.model.activity.timeline import (
    ActivityChildEvent,
    EVENT_STATUS_TRANSITION,
)


__all__ = [
    "ActivityStatus",
    "ActivityStatusChangeError",
    "CANCELED",
    "SUCCEEDED",
    "FAILED",
    "SCHEDULED",
    "WAITING",
    "STATUS_MAP",
    "STATUS_PRIORITY",
    "QUEUED",
    "INPROGRESS",
    "PENDING_STATUSES",
    "FINAL_STATUSES",
    "BLOCKED",
    "VALID_STATUS_TRANSITIONS",
    "add_status_transition_event",
    "NullTransitionError",
]

__clearlist__ = [
    "FAILED",
    "QUEUED",
    "INPROGRESS",
    "CANCELED",
    "SUCCEEDED",
    "SCHEDULED",
    "WAITING",
    "ActivityStatus",
    "BLOCKED",
    "VALID_STATUS_TRANSITIONS",
    "PENDING_STATUSES",
    "ActivityStatusChangeError",
]

__requires_prefix__ = "landscape."


class ActivityStatusChangeError(Exception):
    """Raised when an Activity Status transitions an invalid path."""


class NullTransitionError(Exception):
    """Raised when attempting to transition an activity to its current state.

    This error allows us to see when a transition would be a NoOp.
    """


@requires_advisor
@implementer(IActivityStatus)
class ActivityStatus(object):
    """An activity state."""

    # We'll just PUBLIC this because it doesn't really represent anything
    # sensitive that the referring object doesn't have
    requires(PUBLIC, PRIVATE, ["id", "name", "title"])

    def __init__(self, id=None, name="", title=""):
        self.id = id
        self.name = name
        self.title = title

    def __str__(self):
        return self.title

    def __repr__(self):
        return "ActivityStatus({}, {}, {})".format(self.id, self.name, self.title)

    def copy(self):
        return ActivityStatus(id=self.id, name=self.name, title=self.title)


# IDs for statuses are ordered based on their relevance to the user.
SUCCEEDED = ActivityStatus(1, "succeeded", "Succeeded")
SCHEDULED = ActivityStatus(2, "scheduled", "Scheduled")
WAITING = ActivityStatus(3, "waiting", "Waiting")
QUEUED = ActivityStatus(4, "undelivered", "Queued")
INPROGRESS = ActivityStatus(5, "delivered", "In progress")
BLOCKED = ActivityStatus(6, "blocked", "Blocked")
CANCELED = ActivityStatus(7, "canceled", "Canceled")
FAILED = ActivityStatus(8, "failed", "Failed")

# Define a global ordering for status, to be used by this and other
# modules that require ordering by status.
STATUS_PRIORITY = (
    BLOCKED,
    SCHEDULED,
    QUEUED,
    INPROGRESS,
    FAILED,
    WAITING,
    SUCCEEDED,
    CANCELED,
)

# NOTE: Changing these maps (other than adding new items) will
# probably require a database patch.
STATUS_MAP = {status: status.id for status in STATUS_PRIORITY}

# Statuses in which the activity is considered pending or not
FINAL_STATUSES = (SUCCEEDED, FAILED, CANCELED)
PENDING_STATUSES = tuple(
    status for status in STATUS_MAP if status not in FINAL_STATUSES
)


# Acceptable status transitions. Map valid new statuses to current statuses.
# None values in the old status list mean the new status can be an initial one.
VALID_STATUS_TRANSITIONS = {
    QUEUED: (SCHEDULED, WAITING, BLOCKED, None),
    INPROGRESS: (QUEUED, SCHEDULED),
    FAILED: (INPROGRESS, CANCELED),
    SUCCEEDED: (INPROGRESS, CANCELED),
    CANCELED: PENDING_STATUSES,
    SCHEDULED: (None,),
    WAITING: (BLOCKED, None),
    BLOCKED: (INPROGRESS, None),
}


def add_status_transition_event(activity, old_status):
    """Add ActivityChildEvent for a status transition."""
    new_status = activity.status
    old_status_id = old_status.id if old_status else None
    data = {"old": old_status_id, "new": new_status.id}
    return ActivityChildEvent(activity, event_type=EVENT_STATUS_TRANSITION, data=data)
