# Copyright 2005-2015 Canonical Limited.  All rights reserved.

"""Activity events timeline.

Record events that happen during the lifecycle of activities.
This includes state transitions, as well as events specific to activity types.
"""

from builtins import object
from storm.locals import Store, Storm, AutoReload, Int, DateTime, Reference
from storm.expr import Desc, Cast
from storm.databases.postgres import JSON, JSONTextElement

from canonical.lib.timestamp import utcnow

__all__ = ["ActivityChildTimeline", "ActivityChildEvent", "EVENT_STATUS_TRANSITION"]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


EVENT_STATUS_TRANSITION = 1


class ActivityChildEvent(Storm):
    """An event in a child activity timeline."""

    __storm_table__ = "activity_child_event"

    id = Int(allow_none=False, primary=True, default=AutoReload)
    activity_id = Int(allow_none=False)
    event_type = Int(allow_none=False)
    timestamp = DateTime(allow_none=False, default_factory=utcnow)
    # Arbitrary data specific to the event type
    data = JSON(allow_none=False, default={})

    activity = Reference(activity_id, "Activity.id")

    def __init__(self, activity, event_type, timestamp=None, data=None):
        self.activity = activity
        self.event_type = event_type
        self.timestamp = timestamp or utcnow()
        self.data = data or {}


class ActivityChildTimeline(object):
    """Timeline of events regarding a child activity."""

    def __init__(self, activity):
        self._activity = activity

    def get_events(self, event_type=None):
        """
        Return a ResultSet with ActivityChildEvents for the specified activity,
        possibly filtering by event type, in reverse chronological order.
        """
        where = [ActivityChildEvent.activity_id == self._activity.id]
        if event_type:
            where.append(ActivityChildEvent.event_type == event_type)

        store = Store.of(self._activity)
        result = store.find(ActivityChildEvent, *where)
        result.order_by(Desc(ActivityChildEvent.timestamp))
        return result

    # Methods that follow are used by UI/API and are mostly specific to
    # computer activities. They should be considered legacy and not used in new
    # code.

    def get_modification_time(self):
        """Return the time at which activity status last changed."""
        return self._get_last_status_transition_time()

    def get_completion_time(self):
        """The time at which the activity has completed."""
        from canonical.landscape.model.activity.status import FINAL_STATUSES

        return self._get_last_status_transition_time(FINAL_STATUSES)

    def get_delivery_time(self):
        """Return the time at which the activity was started."""
        from canonical.landscape.model.activity.status import QUEUED, INPROGRESS

        if self._activity.status != QUEUED:
            # If the activity is QUEUED it might have been APPROVED in the
            # past, in which case it has a previous transition to INPROGRESS.
            # In this case, the previous delivery time not be reported.
            return self._get_last_status_transition_time([INPROGRESS])

    def get_approval_time(self):
        """Return the time (if any) of the last approval of the activity."""
        from canonical.landscape.model.activity.status import BLOCKED, QUEUED

        if self._activity.status != BLOCKED:
            # If the activty is BLOCKED, it might have been APPROVED in the
            # past and become BLOCKED again. In this case, the previous
            # approval time should not be reported.
            return self._get_transition_time(BLOCKED, QUEUED)

    def _get_last_status_transition_time(self, statuses=None):
        """Return the time of last status transition of the activity.

        @param statuses: if provided, a list of statuses to limit search to.
            If more than one status is passed, the most recent time is
            returned.
        """
        clauses = []
        if statuses is not None:
            status_ids = [status.id for status in statuses]
            status_expr = Cast(
                JSONTextElement(ActivityChildEvent.data, "new"), "INTEGER"
            )
            clauses.append(status_expr.is_in(status_ids))

        return self._get_last_status_timeline_time(*clauses)

    def _get_transition_time(self, old_status, new_status):
        """
        Return the time (if any) of the last transition from a status to
        another for the activity.
        """
        old_status_expr = Cast(
            JSONTextElement(ActivityChildEvent.data, "old"), "INTEGER"
        )
        new_status_expr = Cast(
            JSONTextElement(ActivityChildEvent.data, "new"), "INTEGER"
        )
        clauses = (old_status_expr == old_status.id, new_status_expr == new_status.id)
        return self._get_last_status_timeline_time(*clauses)

    def _get_last_status_timeline_time(self, *clauses):
        """
        Return the most recent status transition entry for the activity,
        filtering by the specified criteria.
        """
        result = self.get_events(event_type=EVENT_STATUS_TRANSITION)
        result = result.find(*clauses)

        # The timeline is reverse-ordered, so most recent time is first.
        result.config(limit=1)
        event = result.one()
        if event:
            return event.timestamp
