# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime

from storm.locals import Reference

from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.alert.rule import (
    RuleState,
    RuleInfo,
    StoredRuleInfoMixin,
)


__all__ = ["AccountRuleInfo", "StoredAccountRuleInfo"]


__requires_prefix__ = "landscape."


@requires_advisor
class AccountRuleInfo(RuleInfo):
    """
    A RuleInfo that is bound to an account.

    C{rule_requires} should be specified if a particular permission is required
    on the account. If it remains None, then only Owners will be able to view
    the rule.
    """

    rule_requires = None

    def __init__(self, rule):
        super(AccountRuleInfo, self).__init__(rule)
        self.state = RuleState(rule)
        self.check()

    @classmethod
    def is_allowed(class_, person, account):
        """
        Account rules have special permission logic.

        - Owners of the account have access to all rules.
        - restricted admins only gain access to a rule if the info specifies a
          C{rule_requires} other than None, and if the admin has the permission
          specified in C{rule_requires}.
        """
        permission = class_.rule_requires
        if person.has_role("Owner", account.access_context):
            return True
        elif permission is not None and person.has_permission(
            permission, account.access_context
        ):
            return True
        else:
            return False

    @requires("ViewAlertRule")
    def is_active(self):
        return self.state.active

    def _get_state(self):
        if not hasattr(self, "_state"):
            self._state = self.rule.account.store.find(
                RuleState,
                RuleState.computer_id == None,  # Not necessary, but...
                RuleState.rule == self.rule,
            ).one()
        return self._state

    def _set_state(self, state):
        self._state = state

    state = property(_get_state, _set_state)

    def check(self):
        """Check and update the alert rule state.

        Sub-classes should provide an implementation that updates the value of
        the L{RuleState}s C{active} attribute when this method is called.

        @param state: A L{RuleState}.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def _update_notification_time(self):
        """Overridden method updates notification time."""
        self.state.notification.notification_time = datetime.utcnow()

    def get_state_data(self):
        """See L{RuleInfo.get_state_data()}."""
        return None

    def get_stored_state_data(self):
        """See L{RuleInfo.get_stored_state_data()}."""
        return self.state.data


class StoredAccountRuleInfo(AccountRuleInfo, StoredRuleInfoMixin):

    state = Reference(StoredRuleInfoMixin.rule_id, RuleState.rule_id)
