# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Collection logic to filter and find alert L{Rule}s."""

from builtins import map
from storm.expr import Join
from storm.store import EmptyResultSet

from canonical.collection import CollectionBase

from canonical.landscape.model.alert.rule import (
    Rule,
    RuleState,
    RuleSubscription,
    get_rule_info_class,
    get_all_rule_info_classes,
)
from canonical.landscape.model.main.account import get_account_by_id


__all__ = ["RuleCollection"]


class RuleCollection(CollectionBase):
    """A collection of alert L{Rule}s that can be filtered in various ways."""

    def __init__(
        self,
        store,
        account_id=None,
        info_types=None,
        active=None,
        computer_ids=None,
        subscriber_id=None,
        computer_group=None,
        person=None,
    ):
        """
        @param store: the Storm L{Store} to query L{Rule}s against.
        @param person: The L{Person} for which security checking will be
            done. For any account rules that have a C{rule_requires} attribute
            specifying a permission, the person must have that permission on
            the L{Account} in order to view the rule.
        """
        super(RuleCollection, self).__init__(
            store,
            account_id=account_id,
            info_types=info_types,
            active=active,
            computer_ids=computer_ids,
            subscriber_id=subscriber_id,
            computer_group=computer_group,
            person=person,
        )

    def with_account(self, account):
        """Only include L{Rules}s associated to the given L{Account}.

        @return: A new L{RuleCollection} filtered by L{Account}.
        """
        return self._filter_by(account_id=account.id)

    def with_info_classes(self, info_classes):
        """Only include L{Rule}s whose info type is in the given L{RuleInfo}s.

        @param info_classes: a list of L{RuleInfo} sub-classes
        """
        info_types = [info_class.info_type for info_class in info_classes]
        return self._filter_by(info_types=info_types)

    def with_computer_group(self, flag):
        """
        - None: account rules and computer rules will be returned.
        - False: account rules only.
        - True: computer rules only.
        """
        return self._filter_by(computer_group=flag)

    def with_active(self):
        """Only include L{Rule}s with a matching active state."""
        return self._filter_by(active=True)

    def with_inactive(self):
        """Only include L{Rule}s with a matching inactive state."""
        return self._filter_by(active=False)

    def with_computer_ids(self, computer_ids):
        """
        Only include L{Rule}s with a matching C{Computer} in C{computer_ids}.

        @param computer_ids: A sequence of L{Computer.id}s to filter by.
        @return: A new RuleCollection filtered by Computer.ids.
        """
        return self._filter_by(computer_ids=computer_ids)

    def with_computers(self, computers):
        """Only include L{Rule}s applying to the specified computers.

        @param computers: A sequence of L{Computer}s to filter the rule by.
        """
        return self.with_computer_ids([computer.id for computer in computers])

    def with_subscriber(self, person):
        """Only include L{Rule}s the given C{person} is subscribed to.

        @param subscriber: The L{Person} the returned L{Rule}'s should
            be subscribed from.
        """
        return self._filter_by(subscriber_id=person.id)

    def _get_result(
        self,
        store,
        where=None,
        account_id=None,
        info_types=None,
        active=None,
        computer_ids=None,
        subscriber_id=None,
        computer_group=None,
        person=None,
    ):
        """
        Return a L{ResultSet} configured to yield L{Rule}s matching
        filter criteria.
        """
        joins = {}
        if where is None:
            where = []

        if account_id:
            where.append(Rule.account_id == account_id)

        extra_results = None
        if person is not None:
            # 1. Passing a person can affect the other criteria.  First, it
            # will affect which info types will be included in the search. Only
            # account rule info types that the person has access to will be
            # used; any he doesn't have access to will be filtered out.
            if account_id is None:
                raise TypeError("account_id must be passed if person is " "provided")
            # Limit the rules to those that the person has access to.
            account = get_account_by_id(account_id)
            if info_types is None:
                info_classes = get_all_rule_info_classes()
            else:
                info_classes = list(map(get_rule_info_class, info_types))
            # Here we overwrite info_types so we can filter it by those which
            # the person has access to.
            info_types = []
            for info_class in info_classes:
                if info_class.is_allowed(person, account):
                    info_types.append(info_class.info_type)

            # 2. Second, passing a person will also affect which computer IDs
            # will be looked up, if we're searching for active alerts. A Person
            # can only see those alerts on computers that he has ViewComputer
            # on. NOTE: This means that account alerts will not be returned.
            if active and computer_ids is None and computer_group is None:
                computer_ids = person.get_computer_ids(account)
                computer_group = True
                # Since we can't query for alerts-with-specific-computers and
                # alerts-with-no-computers at the same time, we grab another
                # collection for just the account alerts, and later union it
                # with our computer alert query.
                collection = RuleCollection(
                    store,
                    account_id=account.id,
                    active=True,
                    computer_group=False,
                    person=person,
                )
                extra_results = collection.find()

        if computer_ids == []:
            return EmptyResultSet()

        if computer_group is not None:
            if computer_group:
                where.append(Rule.computer_group_id != None)
            else:
                where.append(Rule.computer_group_id == None)

        if info_types is not None:
            where.append(Rule.info_type.is_in(info_types))

        if active is not None:
            if RuleState not in joins:
                joins[RuleState] = Join(Rule, RuleState, Rule.id == RuleState.rule_id)
            if active:
                where.append(RuleState.activation_time != None)
            else:
                where.append(RuleState.activation_time == None)

        if computer_ids is not None:
            if RuleState not in joins:
                joins[RuleState] = Join(Rule, RuleState, Rule.id == RuleState.rule_id)
            where.append(RuleState.computer_id.is_in(computer_ids))

        if subscriber_id is not None:
            joins[RuleSubscription] = Join(
                Rule, RuleSubscription, Rule.id == RuleSubscription.rule_id
            )
            where.append(RuleSubscription.person_id == subscriber_id)

        needs_distinct = False
        if joins:
            store = store.using(*list(joins.values()))
            needs_distinct = True
        result = store.find(Rule, *where)
        result.order_by(Rule.info_type)
        result.config(distinct=needs_distinct)
        if extra_results is not None:
            return result.union(extra_results)
        else:
            return result
