# Copyright 2005-2022 Canonical Limited. All rights reserved.

from collections import defaultdict
from datetime import datetime

import transaction

from storm.locals import Store, Join
from storm.store import EmptyResultSet
from zope.interface import implementer
from zope.event import notify

from canonical.security.protect import requires, requires_advisor
from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.model.events import RuleChangedEvent
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.main.interfaces import IAccessContextComposite
from canonical.landscape.model.alert.rule import (
    RuleStateNotification,
    RuleState,
    RuleInfo,
    StoredRuleInfoMixin,
    Rule,
)
from canonical.landscape.model.alert.mail import AlertEmailViewBase
from canonical.landscape.model.alert.interfaces import IComputerRuleInfo


__all__ = [
    "ComputerRuleInfo",
    "StoredComputerRuleInfo",
    "call_for_enabled_rules_with_transaction",
    "ComputerAlertEmailViewBase",
]

__clearlist__ = ["ComputerRuleInfo"]


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IComputerRuleInfo, IAccessContextComposite)
class ComputerRuleInfo(RuleInfo):
    """The base class for rules pertaining to L{Computer}s.

    Subclasses must implement a C{check} method that takes a sequence of
    L{Computer.id}s as a parameter.  The C{check} method should update the
    L{RuleState}s associated with the specified computer IDs.
    """

    access_cutoff_role = None

    resolution_url = ""

    @requires("ViewAlertRule")
    def is_active(self, computer_ids=None):
        return bool(self.get_states(active=True, computer_ids=computer_ids).any())

    @requires("EditAlert")
    def create_states(self, computer_ids):
        """Create a L{RuleState} for each L{Computer} in C{computers}.

        @param computer_ids: A sequence of L{Computer.id}s.
        @return: A list of L{RuleState}s.
        """
        states = [RuleState(self.rule, id) for id in computer_ids]
        self.check(computer_ids)
        return states

    @requires("ViewAlertRule")
    def get_states(self, active=None, computer_ids=None):
        """Get a L{ResultSet} of L{RuleState}s for this L{Rule}.

        @param active: Optionally, only include L{RuleState}s with a matching
            active state.  Use None for all rule states, True for active ones
            and False for inactive ones.  Default is None.
        @param computer_ids: Optionally, only include L{RuleState}s matching
            L{Computer.id}s in the sequence (if the sequence is empty, an empty
            L{ResultSet} is returned).  Default includes all computers.
        """
        where = [RuleState.rule_id == self.rule.id]
        if active is not None:
            if active:
                where.append(RuleState.activation_time != None)
            else:
                where.append(RuleState.activation_time == None)

        if computer_ids == []:
            return EmptyResultSet()

        if computer_ids is not None:
            where.append(RuleState.computer_id.is_in(computer_ids))
        result = self.rule.account.store.find(RuleState, *where)
        result.order_by(RuleState.computer_id)
        return result

    @requires("ViewAlertRule")
    def get_computers(self, active=None, computer_ids=None):
        """Get a L{ResultSet} of L{Computer}s for this L{Rule}.

        This method has the same signature as L{get_states} but returns a
        L{ResultSet} of L{Computer}s instead of L{RuleState}s.  See
        L{get_states} for information about parameters.
        """
        result = self.get_states(active=active, computer_ids=computer_ids)
        computer_ids = list(result.values(RuleState.computer_id))
        return self.rule.account.get_computers(computer_ids=computer_ids)

    @requires("EditAlert")
    def update_states(self, active_computer_ids=(), inactive_computer_ids=()):
        """Update L{RuleState}s based on specified L{Computer.id}s."""
        # HEADSUP: There is duplication of this logic in the
        #          RuleState.{activate,deactivate}() methods.  Any changes
        #          here should be reflected there.
        activated = deactivated = False
        if active_computer_ids:
            result = self.get_states(computer_ids=active_computer_ids)
            activated = not result.find(RuleState.activation_time == None).is_empty()
            states_data = self.get_state_data(active_computer_ids)
            now = datetime.utcnow()
            for rule_state in result:
                rule_state.activation_time = now
                rule_state.data = states_data.get(rule_state.computer_id)
        if inactive_computer_ids:
            result = self.get_states(computer_ids=inactive_computer_ids).find(
                RuleState.activation_time != None
            )
            deactivated = not result.is_empty()
            result.set(activation_time=None, data=None)
        if activated or deactivated:
            notify(RuleChangedEvent(self.rule))

    @requires("EditAlert")
    def check(self, computer_ids):
        """Check and update L{RuleState}s for C{computer_ids}.

        @param computer_ids: A sequence of L{Computer.id}s.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def _update_notification_time(self):
        """Overridden method updates notification time."""
        result = self.get_states(active=True)
        store = Store.of(self.rule)
        notification = store.find(
            RuleStateNotification,
            RuleStateNotification.rule_state_id.is_in(result.values(RuleState.id)),
        )
        notification.set(notification_time=datetime.utcnow())

    def get_state_data(self, computer_ids):
        """See L{RuleInfo.get_state_data()}."""
        return {}

    def get_stored_state_data(self, computer_ids):
        """See L{RuleInfo.get_state_data()}."""
        if not computer_ids:
            return {}
        store = Store.of(self.rule)
        result = store.find(
            (RuleState.computer_id, RuleState.data),
            RuleState.rule_id == self.rule.id,
            RuleState.computer_id.is_in(computer_ids),
        )
        return dict(result)

    @property
    def access_children(self):
        return self.get_computers()

    def get_resolution_url(self, computer_ids=None):
        """
        Return a URL to resolve the alert, or C{None} if there are no active
        alerts.

        @param computer_ids: a list of computer ids to limit the alert lookup
        """
        states = self.get_states(active=True, computer_ids=computer_ids)
        if states.is_empty():
            return None
        return "%saccount/%s%s" % (
            get_root_url(),
            self.rule.account.name,
            self.resolution_url,
        )


class StoredComputerRuleInfo(ComputerRuleInfo, StoredRuleInfoMixin):
    pass


class ComputerAlertEmailViewBase(AlertEmailViewBase):
    """Base class for computers-related alert email views.

    @ivar affected_computers: a list of L{Computer}s accessible by the
        L{Person} being adapted and affected by the adapted L{RuleInfo}. The
        list is ordered by L{Computer.title} by default, but sub-classes
        can change the sort key by overriding the
        L{_affected_computers_sort_key} method or even completely change
        how the sorting is done by overriding the L{sort_computers} method.
    """

    def __init__(self, context, person):
        super(ComputerAlertEmailViewBase, self).__init__(context, person)
        account = context.rule.account
        computer_ids = person.get_computer_ids(account)
        states = list(context.get_states(active=True, computer_ids=computer_ids))
        self.affected_computers = self.sort_computers(
            [state.computer for state in states]
        )

    def sort_computers(self, computers):
        """
        Returns a list of sorted L{Computer}s, by title, in ascending
        order.
        """
        return sorted(computers, key=self._affected_computers_sort_key)

    def _affected_computers_sort_key(self, computer):
        return computer.title

    @property
    def subject(self):
        """Return a string to be used in the Subject field of this email alert

        The generated subject takes into account only those L{Computer}s
        actually affected by the L{RuleInfo} being adapted.

        If no computers accessible by the L{Person} being adapted are affected
        by the rule, C{None} is returned.
        """
        if self.affected_computers == []:
            return None
        return "[Landscape Alert] %s" % (
            self._info.get_summary(computer_ids=[c.id for c in self.affected_computers])
        )

    @property
    def content(self):
        """Return the content of the alert email body.

        If no computers accessible by the L{Person} being adapted are affected
        by the rule, C{None} is returned.
        """
        if self.affected_computers == []:
            return None
        return super(ComputerAlertEmailViewBase, self).content


def call_for_enabled_rules_with_transaction(rule_info_class, callback):
    """
    Get all the accounts having C{rule_info_class} enabled, and call callback
    for each of these accounts, with the computer ids associated.
    """
    from canonical.landscape.model.account.store import get_account_stores

    for name, store in get_account_stores():
        index = defaultdict(list)
        join = Join(RuleState, Rule, Rule.id == RuleState.rule_id)
        result = store.using(join).find(
            (Rule.account_id, RuleState.computer_id),
            Rule.info_type == rule_info_class.info_type,
        )
        for account_id, computer_id in result:
            index[account_id].append(computer_id)

        # Run rule checks for computers, in a transaction per account.
        for account_id, computer_ids in index.items():
            _call_for_computer_ids_in_account(callback, account_id, computer_ids)


def _call_for_computer_ids_in_account(callback, account_id, computer_ids):
    """Helper function for C{call_for_enabled_rules_with_transaction}.

    Calls the given L{callback} on the L{computer_ids} in the
    L{account} unless the account has been removed or has expired.
    """
    account = get_account_by_id(account_id)
    if account is None or account.has_expired():
        return
    callback(account, computer_ids)
    transaction.commit()
