# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
import transaction

from storm.locals import Or

from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.main.account import get_accounts_by_ids
from canonical.landscape.model.alert.rule import Rule, RuleState, RuleStateNotification
from canonical.landscape.mail.template import EmailViewBase, load_template


__all__ = [
    "AlertEmailViewBase",
    "DefaultAlertEmailView",
    "get_active_notifiable_rules",
    "send_pending_alert_emails",
    "AlertEmailError",
]


class AlertEmailError(Exception):
    """An error occurred sending alert emails."""


def get_active_notifiable_rules(store, account_id=None):
    """Get all L{Rule}s that are active and needed to be notified.

    @param store: An account store.
    @param account_id: Optionally, an L{Account.id} to limit results with.
    """
    where = [
        RuleState.rule_id == Rule.id,
        RuleState.activation_time != None,
        RuleState.id == RuleStateNotification.rule_state_id,
        Or(
            RuleStateNotification.notification_time == None,
            RuleState.activation_time > RuleStateNotification.notification_time,
        ),
    ]
    if account_id:
        where.append(Rule.account_id == account_id)
    result = store.find(Rule, *where)
    result.config(distinct=True)
    return result


def send_pending_alert_emails():
    """Send emails to all accounts with active notifiable alerts."""
    error_occurred = False
    for name, store in get_account_stores():
        result = get_active_notifiable_rules(store)
        account_ids = [rule.account_id for rule in result]
        accounts = get_accounts_by_ids(account_ids)
        active_account_ids = [
            account.id for account in accounts if not account.has_expired()
        ]
        for account_id in active_account_ids:
            for rule in get_active_notifiable_rules(store, account_id):
                try:
                    rule.info.notify()
                except:
                    logging.exception(
                        "Error processing rule notification "
                        "in account %s" % account_id
                    )
                    error_occurred = True
            transaction.commit()
            logging.info("Processed alert notifications for account %s." % account_id)
    if error_occurred:
        raise AlertEmailError()


class AlertEmailViewBase(EmailViewBase):
    """Base class for alert email views.

    This class provides default L{subject} and L{body} methods. The text
    returned by the L{body} method is composed by a header, a content and
    a footer text elements.

    The header and footer elements are supposed to  be common to all alert
    views and are implemented by the L{header} and L{footer} methods, while
    the content element are L{RuleInfo}-specific and must be customised by
    sub-classes through the L{_content_template} variable.
    """

    _header_template = load_template(__file__, "email-header.msg")
    _footer_template = load_template(__file__, "email-footer.msg")

    def __init__(self, context, person):
        super(AlertEmailViewBase, self).__init__(context, person)
        self._info = self._context
        self.account_title = self._info.rule.account.title
        self.resolve_url = "%saccount/%s/alert/%s/resolve" % (
            get_root_url(),
            self._info.rule.account.name,
            self._info.rule.id,
        )
        self.subscription_url = "%saccount/%s/alerts" % (
            get_root_url(),
            self._info.rule.account.name,
        )

    @property
    def subject(self):
        """
        Return the subject of the alert email, using L{RuleInfo.get_summary}.
        """
        return "[Landscape Alert] %s" % (self._info.get_summary())

    @property
    def header(self):
        """Return the header of the alert email body."""
        return self._render(self._header_template)

    @property
    def content(self):
        """Return the content of the alert email body.

        Sub-classes must provide a template through the L{_content_template}
        instance variable.
        """
        return self._render(self._content_template)

    @property
    def footer(self):
        """Return the footer of the alert email body."""
        return self._render(self._footer_template)

    @property
    def body(self):
        """
        Render the body of the email by composing header, content and footer.

        If the content element is C{None}, the body will be C{None} as well.
        """
        if self.content is None:
            return None

        return "%s\n\n%s\n\n%s" % (self.header, self.content, self.footer)


class DefaultAlertEmailView(AlertEmailViewBase):

    _content_template = load_template(__file__, "default-alert.msg")

    def __init__(self, context, person):
        super(DefaultAlertEmailView, self).__init__(context, person)
        self.alert_url = "%saccount/%s/alert/%s" % (
            get_root_url(),
            context.rule.account.name,
            context.rule.id,
        )
