# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
from datetime import timedelta

from zope.event import notify
from zope.component import getUtility
from zope.interface import implementer, implementer_only

from storm.locals import Int, Bool, DateTime, Reference, AutoReload, Pickle

from canonical.lib.timestamp import utcnow

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.mail.template import get_email_view
from canonical.landscape.model.events import RuleChangedEvent
from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.main.interfaces import (
    IAccessContextAware,
    IAccessDelegateAware,
    IAccountAware,
)
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.computer.interfaces import IComputerAware
from canonical.landscape.model.account.helper import ComputerGroupHelper
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.alert.interfaces import IRuleInfo


__all__ = [
    "Rule",
    "RuleStateNotification",
    "RuleState",
    "RuleInfo",
    "RuleSubscription",
    "StoredRuleInfoMixin",
    "register_rule_info_type",
    "unregister_rule_info_types",
    "get_rule_info_class",
    "get_all_rule_info_classes",
    "unregister_rule_info_type",
    "get_rule_info_name",
    "get_all_rule_info_names",
    "get_rule_info_class_by_name",
    "UnknownAlertTypeError",
    "InvalidAlertTypeError",
]

__clearlist__ = ["Rule"]

__requires_prefix__ = "landscape."


class UnknownAlertTypeError(Exception):
    """The specified alert type could not be found."""


class InvalidAlertTypeError(Exception):
    """The specified alert type is not valid for the operation."""


class UnregisteredRuleError(Exception):
    """
    Raised when a RuleInfo type which hasn't been registered with an info_type
    ID is used.
    """


rule_info_types = {}
rule_info_names = {}
reverse_rule_info_types = {}
reverse_rule_info_names = {}


def get_rule_info_class(info_type):
    """
    Get a L{RuleInfo} class by info_type.

    @raises KeyError: If the info_type hasn't been registered.
    """
    return rule_info_types[info_type]


def get_rule_info_class_by_name(name):
    """
    Get a L{RuleInfo} class by name.

    @param name: the name of the RuleInfo subclass we'd like to return.
    """
    try:
        info_type = reverse_rule_info_names[name]
    except KeyError:
        raise UnknownAlertTypeError(
            "No alert type exists matching the name '%s'" % name
        )
    return get_rule_info_class(info_type)


def get_rule_info_name(info_type):
    """
    Get the name associated with a registered L{RuleInfo} class by
    type.

    @param info_type: the info type identifier for which the name
        should be returned.
    """
    return rule_info_names[info_type]


def get_all_rule_info_classes():
    """Get all registered L{RuleInfo} classes."""
    return list(reverse_rule_info_types.keys())


def get_all_rule_info_names():
    """Get the names of all registered L{RuleInfo} subclasses."""
    return list(reverse_rule_info_names.keys())


def register_rule_info_type(info_type, info_class):
    """Register C{info_class} as a L{RuleInfo} implementation.

    The C{info_type} will be stored as C{Rule.info_type} for any rules created
    with C{info_class}.

    @param info_type: A unique integer representing C{info_class}.
    @param info_class: A L{RuleInfo}-based class.
    @raise RuntimeError: Raised if a class is already registered for
        C{info_type}.
    """
    existing_info_class = rule_info_types.get(info_type)
    if existing_info_class is not None:
        raise RuntimeError(
            "%r has the same info_type of %r" % (info_class, existing_info_class)
        )
    name = info_class.get_name()
    rule_info_types[info_type] = info_class
    rule_info_names[info_type] = name
    reverse_rule_info_types[info_class] = info_type
    reverse_rule_info_names[name] = info_type
    info_class.info_type = info_type


def unregister_rule_info_types():
    """Unregister all L{RuleInfo} types, useful for test cleanups."""
    rule_info_types.clear()
    reverse_rule_info_types.clear()


def unregister_rule_info_type(info_type):
    """
    Unregister a single L{RuleInfo} type. Useful for test cleanups.

    Note: This does not clear the C{info_type} attribute that
    L{register_rule_info_type} puts on the L{RuleInfo} class.
    """
    info_class = rule_info_types[info_type]
    del rule_info_types[info_type]
    del reverse_rule_info_types[info_class]


@requires_advisor
@implementer_only(IAccountAware, IAccessDelegateAware)
class Rule(ComputerGroupHelper):
    """
    A rule is the mechanism by which we know when to mark an L{Account} or
    L{Computer} as in an alerted state.

    This class doesn't directly define the logic for deciding whether a
    computer is alerted.  That's the responsibility of the L{RuleInfo}, which
    can be configured with various rule-specific parameters if necessary. The
    C{info_type} defines the L{RuleInfo} type to use for this rule.

    A rule can be enabled for none, some or all L{Computer}s in an L{Account}.
    A flag is used to directly indicate whether or not the rule is enabled for
    all computers.  L{Tag}s can be associated with a rule to enable it for
    computers with matching tags.
    """

    __storm_table__ = "rule"

    id = Int(primary=True, default=AutoReload)
    info_type = Int(allow_none=False)
    account_id = Int(allow_none=False)
    next_notification_time = DateTime()
    access_context_id = Int()
    _info = None

    computer_group_id = Int()
    computer_group = Reference(computer_group_id, ComputerGroup.id)

    requires(
        "ViewAlertRule",
        PRIVATE,
        ["id", "info", "info_type", "account", "all_computers", "get_tags"],
    )
    requires(
        "EditAlert",
        PRIVATE,
        [
            "update_association",
            "update_tags",
            "enable_for_all_computers",
            "disable_for_all_computers",
        ],
    )

    # This is to get rid of AccountHelper's IAccessContextAware interface
    # XXX Will need to be used to track Rule changesets for auditing
    _changes = {}

    # TODO: Add get_affected_computers?
    def __init__(self, account, rule_info_type, **kwds):
        if rule_info_type not in list(rule_info_types.values()):
            raise UnregisteredRuleError(
                "%s is not a registered rule info type." % (rule_info_type,)
            )
        super(Rule, self).__init__(account)
        self.info_type = rule_info_type.get_info_type()
        self._info = rule_info_type(self, **kwds)
        from canonical.landscape.model.alert.computer import ComputerRuleInfo

        if isinstance(self.info, ComputerRuleInfo):
            self._create_computer_group()

    @property
    def info(self):
        """Get this rule's L{RuleInfo} object."""
        if self._info is not None:
            return self._info
        assert self.id is not None
        type = rule_info_types[self.info_type]
        if not hasattr(type, "__storm_table__"):
            info = type.__new__(type)
            info.rule = self
        else:
            info = self.account.store.get(type, self.id)
        self._info = info
        return info

    @property
    def title(self):
        # XXX compatibility code for event logs, see get_log_entity_name,
        # eventually we should find a way to normalize this model.
        return str(self.info.title)

    @requires("GetAlertSubscription")
    def get_subscribers(self):
        """
        Return a L{ResultSet} of the L{Person}s who are subscribed to this
        rule.
        """
        person_ids = self.account.store.find(
            RuleSubscription, RuleSubscription.rule == self
        ).values(RuleSubscription.person_id)
        return get_main_store().find(Person, Person.id.is_in(person_ids))

    @requires("GetAlertSubscription")
    def get_subscription_by_person(self, person):
        """
        Get a L{RuleSubscription} for C{person} or None if one isn't
        available.
        """
        return self._get_subscription_by_person_resultset(person).one()

    def _get_subscription_by_person_resultset(self, person):
        return self.account.store.find(
            RuleSubscription,
            RuleSubscription.rule == self,
            RuleSubscription.person_id == person.id,
        )

    @requires("SubscribeAlert")
    def subscribe(self, person):
        """Subscribe C{person} to this rule.

        No changes will be made if a L{RuleSubscription} already exists for
        C{person}.
        """
        subscription = self.get_subscription_by_person(person)
        if not subscription:
            subscription = RuleSubscription(self, person)
        return subscription

    @requires("SubscribeAlert")
    def unsubscribe(self, person):
        """Unsubscribe C{person} to this rule."""
        self._get_subscription_by_person_resultset(person).remove()

    def add_computers_with_ids(self, computer_ids):
        """
        Create states and computer_group items for the given computer ids.
        """
        if not computer_ids:
            return

        self.info.create_states(computer_ids)
        super(Rule, self).add_computers_with_ids(computer_ids)
        notify(RuleChangedEvent(self))

    def remove_computers_with_ids(self, computer_ids):
        """
        Remove states and computer_group items for the given computer ids.
        """
        if not computer_ids:
            return

        rule_states = self.account.store.find(
            RuleState, RuleState.rule == self, RuleState.computer_id.is_in(computer_ids)
        )
        rule_states.remove()
        super(Rule, self).remove_computers_with_ids(computer_ids)
        notify(RuleChangedEvent(self))

    @property
    def access_delegate(self):
        return self.info

    def _get_rule_status(self, rule, is_computer_rule, computer_ids):
        """Get a textual description of the current rule status."""

        alerted = "Alerted"
        ok = "OK"
        disabled = "Disabled"

        if is_computer_rule:
            is_active = rule.info.is_active(computer_ids=computer_ids)
            if rule.info.get_states(computer_ids=computer_ids).is_empty():
                if not self.get_tags():
                    return disabled
            if is_active:
                return alerted
            else:
                return ok
        else:
            if rule.info.is_active():
                return alerted
            return ok

    def get_state_for_computers(self, computer_ids=None):
        """Get rule state for a set of computer_ids."""
        name = self.info.get_name()

        from canonical.landscape.model.alert.computer import ComputerRuleInfo

        is_computer_rule = isinstance(self.info, ComputerRuleInfo)

        if is_computer_rule:
            all_computers = self.all_computers
            scope = "computer"
            tags = sorted(tag.name for tag in self.get_tags())
        else:
            all_computers = False
            scope = "account"
            tags = []

        status = self._get_rule_status(self, is_computer_rule, computer_ids)

        return {
            "alert_type": name,
            "description": self.title,
            "scope": scope,
            "tags": tags,
            "status": status,
            "all_computers": all_computers,
        }


@implementer(IAccessContextAware, IComputerAware)
class RuleSubscription(object):
    """
    The representation that a L{Person} should get notified when it is
    detected that computers are newly affected by a L{Rule}.
    """

    __storm_table__ = "rule_subscription"
    __storm_primary__ = ("rule_id", "person_id")

    rule_id = Int(allow_none=False)
    person_id = Int(allow_none=False)

    rule = Reference(rule_id, Rule.id)

    requires("ViewAlertRule", PRIVATE, ["rule"])

    def __init__(self, rule, person):
        self.rule = rule
        self.person_id = person.id

    @property
    def person(self):
        return get_main_store().get(self.person_id)

    @property
    def access_context(self):
        return self.rule.account.access_context


class RuleStateNotification(object):

    __storm_table__ = "rule_state_notification"

    rule_state_id = Int(primary=True)
    notification_time = DateTime()

    def __init__(self, rule_state_id, notification_time):
        self.rule_state_id = rule_state_id
        self.notification_time = notification_time


class RuleState(ComputerHelper):
    """
    Represents whether a L{Computer} or an L{Account} is currently 'alerted'
    according to a L{Rule}.

    @ivar active: When some code determines that a L{Rule} applies to a
        particular computer or account, a L{RuleState} should have its
        C{active} attribute set to True. When Landscape detects that the rule
        no longer applies to a computer, the C{active} attribute should be set
        back to False.
    @ivar dismissed: The user can explicitly 'dismiss' an alerted L{Computer}
        or L{Account} as a way of indicating that they no longer want to be
        bothered about this computer or account in this alerted state.
    """

    __storm_table__ = "rule_state"

    requires(
        "ViewAlertRule",
        PRIVATE,
        ["active", "activation_time", "notification_time", "dismissed", "computer_id"],
    )

    id = Int(primary=True, default=AutoReload)
    rule_id = Int(allow_none=False)
    computer_id = Int()
    activation_time = DateTime()
    notification = Reference(id, RuleStateNotification.rule_state_id)
    dismissed = Bool()
    data = Pickle(default=None)

    rule = Reference(rule_id, Rule.id)

    def __init__(self, rule, computer=None):
        super(RuleState, self).__init__(computer)
        self.rule = rule
        self.notification = RuleStateNotification(self.id, None)

    @property
    def notification_time(self):
        return self.notification.notification_time

    @property
    def active(self):
        return bool(self.activation_time)

    def activate(self):
        # HEADSUP: There is duplication of this logic in
        #          ComputerRuleInfo.update_states().  Any changes here
        #          should be reflected there.
        activated = self.activation_time is None
        self.activation_time = utcnow()
        from canonical.landscape.model.alert.computer import ComputerRuleInfo

        if not isinstance(self.rule.info, ComputerRuleInfo):
            self.data = self.rule.info.get_state_data()
        if activated:
            notify(RuleChangedEvent(self.rule))

    def deactivate(self):
        # HEADSUP: There is duplication of this logic in
        #          ComputerRuleInfo.update_states().  Any changes here
        #          should be reflected there.
        deactivated = self.activation_time is not None
        self.activation_time = None
        self.data = None
        if deactivated:
            notify(RuleChangedEvent(self.rule))

    def notify(self):
        if self.activation_time:
            self.notification.notification_time = utcnow()
        else:
            raise RuntimeError(
                "The notify method should not be called with " "an inactive rule state."
            )


@implementer(IRuleInfo, IAccessContextAware, IComputerAware)
class RuleInfo(object):
    """
    The logic and configuration for a rule which decides whether L{Account}s
    or L{Computers} should be in an alerted state.

    C{RuleInfo} should be subclassed to provide a custom database table and
    logic.  All C{RuleInfo} classes must be registered with
    L{register_rule_info_type} before being used; this can also be invoked by
    using a C{<zope:rule>} directive in a configure.zcml file, like so::

      <rule key="999" class="canonical.landscape.foo.bar.RuleInfoSubclass" />

    Subclasses must provide a C{check} method that updates the active flag of
    any related state objects.  Subclasses must provide a C{get_message}
    method that takes a L{ComputerRuleState} and returns a C{(subject,
    message)} 2-tuple, called when email notifications are generated.  A
    C{title} attribute must also be provided.  The title attribute is used by
    presentation code in browser pages and in the subject line of email
    notifications.

    @cvar notification_delay: The amount of time, in seconds, to wait before
        sending notifications.  The default is 0.  Subclasses should provide
        an appropriate value.
    @type notification_delay: timedelta
    """

    requires("ViewAlertRule", PRIVATE, ["rule", "get_name"])

    notification_delay = timedelta(seconds=0)

    def __init__(self, rule):
        self.rule = rule

    @property
    def access_context(self):
        return self.rule.account.access_context

    @classmethod
    def get_info_type(class_):
        return reverse_rule_info_types[class_]

    @classmethod
    def get_name(class_):
        return class_.__name__

    @classmethod
    def is_allowed(class_, person, account):
        """
        Returns True if this L{Rule} type should be visible by the specified
        person in the specified account.
        """
        return True

    def _allow_notifications(self):
        """Returns True if the L{Rule} is allowed to send notifications."""
        now = utcnow()
        if self.rule.next_notification_time is None:
            # If the time since the last notification is unknown, the next
            # notification time is set to the current time plus the
            # notification delay.  The purpose of this delay is to create a
            # window in which data from many computers can be collected, so
            # that fewer emails with more information are sent than reporting
            # as soon as a single computer reports data.
            self.rule.next_notification_time = now + self.notification_delay
        return self.rule.next_notification_time <= now

    def notify(self):
        """Send a status message about this alert to each subscriber."""
        if not self._allow_notifications():
            return

        mailer = getUtility(ILandscapeMailer)
        for subscriber in self.rule.get_subscribers():
            email = get_email_view(self, subscriber)
            if email.body is None:
                continue
            mailer.send(email.to, email.subject, email.body, bulk=True)
            logging.info(
                "Sent alert to %s for account %s about rule: %s"
                % (subscriber.email, self.rule.account.name, self.title)
            )
        self._update_notification_time()
        logging.info("Resetting next notification time for rule %s." % (self.title,))
        self.rule.next_notification_time = None

    def _update_notification_time(self):
        raise NotImplementedError("Subclasses must implement this method!")

    @requires("ViewAlertRule")
    def get_summary(self):
        """Get a summary about the state of the rule.

        The summary should be a statement about the rules status.  For
        example, '2 activities need approval' or '3 computers are offline'.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def check(self, computer_ids=None):
        """Check and activate or deactivate alert rule state, when necessary.

        This method checks conditions and updates rule state.  Computer-based
        rules should check L{RuleState}s for all related L{Computer}s,
        while account-based ones should check just its one state.

        Account-based rule infos will call C{check()} without arguments,
        while computer-based rule infos should pass in the specific
        computer ids which should be checked.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def get_state_data(self, computer_ids=None):
        """Return current state data for given computers or for an account.

        The state.data parameter is a rule-info-defined way to track changes
        in the monitored data while an alert is still active.  An example
        is probably useful to illustrate it: when a security upgrade alert
        is sent when upgrade A is known, only this upgrade is mentioned in
        the email sent.  If a second security upgrade B shows up, the user
        should be alerted about the new fact, even if he didn't yet handle
        the original problem.  The result of this function must be a
        rule-info-specific data value which represents the current state
        for the given alert rule info.

        For ComputerRuleInfos, the first argument must be a list of
        computer_ids, and the function result will be an up-to-date
        map like {computer_id: data} for the given computers.  If a
        computer_id in computer_ids isn't present in the returned map,
        it should be considered that its state data is None.

        For AccountRuleInfos, the first argument should not be provided,
        and the result is the data value itself.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    def get_stored_state_data(self, computer_ids=None):
        """Return the currently stored state data.

        This method behaves similarly to L{get_stored_data()}, except
        that instead of returning *current* values, it will return data
        values currently stored in the database for the given computers
        (for L{ComputerRuleInfo}s) or the associated account
        (for L{AccountRuleInfo}s).

        This is useful for comparisons, and for generation of content
        comparing both values.
        """
        raise NotImplementedError("Subclasses must implement this method!")

    @requires("ViewAlertRule")
    def is_active(self):
        """Returns True if the L{Rule} is currently active."""
        raise NotImplementedError("Subclasses must implement this method!")

    @requires("ViewAlertRule")
    def get_resolution_url(self):
        """
        Return a URL to a component in Landscape that can be used to resolve
        the alerted situation.  If the alert is inactive, or if there is no
        appropriate resolution URL, None should be returned.  Default is None.
        """
        return None

    @requires("ViewAlertRule")
    def get_state(self):
        """
        Return the state of the L{RuleInfo} including its C{alert_type} and
        C{summary}.
        """
        return {
            "alert_type": self.get_name(),
            "summary": self.get_summary(),
            "resolution_url": self.get_resolution_url(),
        }


class StoredRuleInfoMixin(object):

    rule_id = Int(primary=True)
    rule = Reference(rule_id, Rule.id)
