"""Entities modelling the autoscale DSL."""

from builtins import zip
from builtins import object
from canonical.landscape.model.autoscale.errors import InferenceError

# Type codes, used for static type inference
INT = "int"
STRING = "string"
BOOL = "bool"
SERVICE_REF = "service-ref"
METRIC_REF = "metric-ref"
TIME_DELTA = "time-delta"
MACHINE_REF = "machine-ref"
CONTAINER_REF = "container-ref"
UNIT_REF = "unit-ref"


class Program(object):
    """Top-level language entity.

    @ivar statements: The list of L{Statement}s the program is composed of.
    """

    def __init__(self, statements):
        self.statements = statements

    def infer(self):
        """Infer the types of all our statements."""
        stack = Stack()  # Brand new stack
        for statement in self.statements:
            statement.infer(stack)


class Assignment(object):
    """Variable assignment.

    @ivar variable: The L{Variable} being assigned.
    @ivar expression: The L{Expression} to assign.
    """

    def __init__(self, variable, expression):
        self.variable = variable
        self.expression = expression

    def infer(self, stack):
        """Bind the variable's name to the expression's type."""
        stack.add(self.variable.name, self.expression.infer(stack))


class Conditional(object):
    """Conditional statement.

    @ivar guard: The L{Expression} deciding whether to execute the conditional.
    @ivar statements: The statements to execute if the guard is true.
    """

    def __init__(self, guard, statements):
        self.guard = guard
        self.statements = statements

    def infer(self, stack):
        """Check the guard and the statements."""
        if self.guard.infer(stack) != BOOL:
            raise InferenceError("Type error (expected bool)")

        # The statements block gets evaluated in a separate scope
        stack.push()
        for statement in self.statements:
            statement.infer(stack)
        stack.pop()


class Expression(object):
    """Base class for expressions."""


class Integer(Expression):
    """Hold a single integer.

    @ivar value: The integer.
    """

    def __init__(self, value):
        self.value = value

    def infer(self, stack):
        return INT


class Boolean(Expression):
    """Hold a single boolean.

    @ivar value: The boolean.
    """

    def __init__(self, value):
        self.value = value

    def infer(self, stack):
        return BOOL


class String(Expression):
    """Hold a single string.

    @ivar value: The string.
    """

    def __init__(self, value):
        self.value = value

    def infer(self, stack):
        return STRING


class Variable(Expression):
    """Reference to a variable.

    @ivar name: The name of the variable.
    """

    def __init__(self, name):
        self.name = name

    def infer(self, stack):
        """Return the type of the referenced expression."""
        try:
            return stack.get(self.name)
        except KeyError:
            raise InferenceError("Unknown variable")


class Operator(Expression):
    """Expressions with an operator.

    @ivar name: The name of the operator (e.g. "and", "<")
    @ivar operands: A list of operand expressions.
    """

    def __init__(self, name, operands):
        self.name = name
        self.operands = operands

    def infer(self, stack):
        """Infer and check the types of the operands."""
        if self.name in ["<", "<=", "==", ">=", ">"]:
            requires = INT
        elif self.name in ["not", "and", "or"]:
            requires = BOOL
        else:
            raise RuntimeError("Unexpected operator %s" % self.name)
        for operand in self.operands:
            if operand.infer(stack) != requires:
                raise InferenceError()
        return BOOL


class Function(Expression):
    """Function call.

    @ivar name: The name of the function being called.
    @ivar args: A list of L{Expression} arguments.
    @ivar kwargs: A C{dict} of L{Expression} keyword arguments.
    """

    # Map built-in function names to their expected signature, holding
    # positional arguments types, keyword arguments types and return type
    _signatures = {
        "service": ([STRING], {}, SERVICE_REF),
        "average": ([STRING, SERVICE_REF], {}, METRIC_REF),
        "total": ([STRING, SERVICE_REF], {}, METRIC_REF),
        "minutes": ([INT], {}, TIME_DELTA),
        "above": ([INT, METRIC_REF], {"for": TIME_DELTA}, BOOL),
        "below": ([INT, METRIC_REF], {"for": TIME_DELTA}, BOOL),
        "add_machine": ([], {"size": STRING}, MACHINE_REF),
        "add_container": ([MACHINE_REF], {}, CONTAINER_REF),
        "add_unit": ([SERVICE_REF], {"on": CONTAINER_REF}, UNIT_REF),
        "num_units": ([SERVICE_REF], {}, INT),
    }

    def __init__(self, name, args, kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs

    def infer(self, stack):
        """Infer and check the types of the arguments."""
        signature = self._signatures.get(self.name)

        if signature is None:
            raise InferenceError("Unknown function")

        if len(self.args) != len(signature[0]):
            raise InferenceError(
                "Wrong number of arguments (expected %d)" % len(signature[0])
            )

        for arg, requires in zip(self.args, signature[0]):
            if arg.infer(stack) != requires:
                raise InferenceError("Wrong argument type (expected %s)" % requires)

        for name, kwarg in self.kwargs.items():
            requires = signature[1].get(name)

            if requires is None:
                raise InferenceError("Unknown keyword argument %r" % name)

            if kwarg.infer(stack) != requires:
                raise InferenceError("Type error for keyword argument %r" % name)

        return signature[2]


class Stack(object):
    """A dictionary-like class sporting stacked lookups."""

    def __init__(self):
        self._lookups = [{}]

    def push(self):
        """Add a lookup layer, new additions will override existing ones."""
        self._lookups.insert(0, {})

    def pop(self):
        """Remove the topmost lookup layer, restoring former lookups."""
        self._lookups.pop(0)

    def get(self, key):
        """Lookup the value for the given key."""
        for lookup in self._lookups:
            if key in lookup:
                return lookup[key]
        raise KeyError(key)

    def add(self, key, value):
        """Add a value for the given key."""
        self._lookups[0][key] = value
