from builtins import str
from builtins import object
from pyparsing import (
    Word,
    Optional,
    Combine,
    Suppress,
    Group,
    Forward,
    Literal,
    alphas,
    alphanums,
    nums,
    delimitedList,
    lineEnd,
    operatorPrecedence,
    opAssoc,
    oneOf,
    ParseException,
)

from canonical.landscape.model.autoscale.errors import ParseError
from canonical.landscape.model.autoscale.language import (
    Integer,
    Boolean,
    String,
    Variable,
    Function,
    Assignment,
    Program,
    Conditional,
    Operator,
)


class Parser(object):
    """Convert autoscale language strings into abstract syntax trees."""

    def __init__(self):
        self._grammar = self._build_grammar()

    def parse(self, text):
        """Parse the given autoscale language script.

        @param text: The string to parse.
        @return: A list of parsed L{Statement}s.
        """
        try:
            result = self._grammar.parseString(text, parseAll=True)
        except ParseException as error:
            raise ParseError(str(error))
        return Program(list(result))

    def _build_grammar(self):
        """Return a pyparsing grammar modelling the language syntax."""

        # Syntactic symbols and delimiters
        comma = Suppress(",")
        equal = Suppress("=")
        quote = Suppress('"')
        lparen = Suppress("(")
        rparen = Suppress(")")
        lcurly = Suppress("{")
        rcurly = Suppress("}")

        # Valid identifiers
        letter = Word(alphas, exact=1)
        identifier = Combine(letter + Optional(Word(alphanums + "_")))

        # Expressions
        operand = Forward()

        variable = Combine(Suppress("$") + identifier)
        variable.setParseAction(self._parse_variable)

        boolean = Literal("true") | Literal("false")
        boolean.setParseAction(self._parse_boolean)

        string = quote + Word(alphanums + "-") + quote
        string.setParseAction(self._parse_string)

        integer = Word(nums)
        integer.setParseAction(self._parse_integer)

        args = delimitedList(operand)
        kw_assignment = identifier + equal + operand
        kwargs = Optional(comma + delimitedList(Group(kw_assignment)))
        function = identifier + lparen + Group(args) + Group(kwargs) + rparen
        function.setParseAction(self._parse_function)

        operand << (integer | boolean | string | variable | function)

        comparison = oneOf("< > == <= >=")
        expression = operatorPrecedence(
            operand,
            [
                (comparison, 2, opAssoc.LEFT, self._parse_operator),
                ("not", 1, opAssoc.RIGHT, self._parse_operator),
                ("and", 2, opAssoc.LEFT, self._parse_operator),
                ("or", 2, opAssoc.LEFT, self._parse_operator),
            ],
        )

        # Assignments
        assignment = variable + equal + expression
        assignment.setParseAction(self._parse_assignment)

        # Blocks
        block = Forward()

        # Conditionals
        conditional = Suppress("if") + expression + lcurly + block + rcurly
        conditional.setParseAction(self._parse_conditional)

        statement = assignment | expression | conditional
        block << Group(delimitedList(statement, delim=lineEnd))

        # Complete grammar
        grammar = delimitedList(statement, delim=lineEnd)
        return grammar

    def _parse_assignment(self, tokens):
        """Parse an assignment."""
        return Assignment(tokens[0], tokens[1])

    def _parse_conditional(self, tokens):
        """Parse a conditional."""
        return Conditional(tokens[0], tokens[1])

    def _parse_string(self, tokens):
        """Parse a string."""
        return String(tokens[0])

    def _parse_boolean(self, tokens):
        """Parse a boolean."""
        return Boolean(tokens[0] == "true")

    def _parse_integer(self, tokens):
        """Parse an integer."""
        # TODO: handle converion errors
        return Integer(int(tokens[0]))

    def _parse_operator(self, tokens):
        """Parse an operator."""
        token = tokens[0]

        # Unary operator
        if len(token) == 2:
            name = token[0]
            operands = [token[1]]
        else:
            name = token[1]
            # Skip all tokens with odd indexes, as they are all operators,
            # e.g. "3 < 4 < 5" results in the tokens [3, <, 4, <, 5].
            index = 0
            operands = []
            while index < len(token):
                operands.append(token[index])
                index += 2

        return Operator(name, operands)

    def _parse_variable(self, tokens):
        """Parse a variable."""
        return Variable(tokens[0])

    def _parse_function(self, tokens):
        """Parse a function."""
        args = list(tokens[1])
        kwargs = dict((token[0], token[1]) for token in tokens[2])
        return Function(tokens[0], args, kwargs)
