# Copyright 2005-2016 Canonical Limited.  All rights reserved.

"""Track hourly rates applied to managed computers.

Each computer type can have different hourly rates.  Rates apply from the
specified date until another rate with a more recent validity date is added.
"""

from collections import OrderedDict

from storm.locals import Storm, Int, Date, AutoReload, Enum
from storm.expr import Max, Min

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import COMPUTER_TYPE_MAP, COMPUTER_TYPES


class UtilityCost(Storm):
    """Hourly cost for managed computers of a certain type.

    @ivar computer_type: the type of computer the rate applies to.
    @ivar valid_since: the start validity date for the rate.
    @ivar hourly_cost: the management cost per hour, in 100th of cents.
    """

    __storm_table__ = "utility_cost"

    id = Int(allow_none=False, primary=True, default=AutoReload)
    computer_type = Enum(map=COMPUTER_TYPE_MAP)
    valid_since = Date(allow_none=False, default=AutoReload)
    hourly_cost = Int(allow_none=False)

    def __init__(self, computer_type, valid_since, hourly_cost):
        self.computer_type = computer_type
        self.valid_since = valid_since
        if hourly_cost <= 0:
            raise ValueError("Hourly cost must be positive")
        self.hourly_cost = hourly_cost


def set_utility_cost(computer_type, valid_since, hourly_cost):
    """Add a UtilityCost or update an existing one for the same
    date and computer type."""
    store = get_main_store()
    utility_cost = store.find(
        UtilityCost,
        UtilityCost.computer_type == computer_type,
        UtilityCost.valid_since == valid_since,
    ).one()
    if utility_cost is None:
        return store.add(UtilityCost(computer_type, valid_since, hourly_cost))
    else:
        utility_cost.hourly_cost = hourly_cost
        return utility_cost


def get_utility_costs_over_period(start_date, end_date):
    """Return a dict with costs over a time range for each computer type."""
    store = get_main_store()

    start_cost_dates = _get_start_date_for_costs(start_date)
    end_cost_dates = _get_start_date_for_costs(end_date)

    costs_by_computer_type = OrderedDict()
    for computer_type in COMPUTER_TYPES:
        cost = store.find(
            Min(UtilityCost.hourly_cost),
            UtilityCost.computer_type == computer_type,
            UtilityCost.valid_since >= start_cost_dates[computer_type],
            UtilityCost.valid_since <= end_cost_dates[computer_type],
        ).one()
        costs_by_computer_type[computer_type] = cost

    return costs_by_computer_type


def _get_start_date_for_costs(time):
    """
    Return the start time for current cost at the specified time, for each
    computer type.
    """
    store = get_main_store()
    result = store.find(
        (UtilityCost.computer_type, Max(UtilityCost.valid_since)),
        UtilityCost.valid_since <= time,
    )
    return dict(result.group_by(UtilityCost.computer_type))
