# Copyright 2005-2016 Canonical Limited.  All rights reserved.

from collections import namedtuple
from dateutil.relativedelta import relativedelta

from storm.locals import Storm, Reference, Int, Date, AutoReload

from canonical.lib.timestamp import utctoday

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.account import Account


CREDIT_VALIDITY = relativedelta(days=65)


class UtilityCredit(Storm):
    """Utility credit given to an Account."""

    __storm_table__ = "utility_credit"

    id = Int(allow_none=False, primary=True, default=AutoReload)
    account_id = Int(allow_none=False)
    creation_date = Date(allow_none=False, default=AutoReload)
    expiration_date = Date(allow_none=False)
    initial_amount = Int(allow_none=False)
    amount = Int(allow_none=False)

    account = Reference(account_id, Account.id)

    def __init__(self, account, amount):
        if amount <= 0:
            raise ValueError("Credit amount must be positive")
        self.account = account

        self.creation_date = utctoday()
        self.expiration_date = self.creation_date + CREDIT_VALIDITY

        self.initial_amount = amount
        self.amount = amount

    @property
    def snapshot(self):
        return CreditSnapshot(
            self.amount, self.initial_amount, self.creation_date, self.expiration_date
        )


class CreditSnapshot(
    namedtuple("CreditSnapshot", "amount initial creation_date expiry")
):
    """A read-only copy of a credit."""

    @property
    def snapshot(self):
        return self


def subtract_credits(amount, credits):
    """Return a tuple of credits/amount values after subtracting the amount.

    The result is (used credits, remaining credits, remaining amount).
    Both lists of credits are ordered by oldest first.  As well, when
    the original credits are applied to the amount, the oldest credits
    are used up first.  If only part of a credit is used toward the
    original amount then the used portion of it will show up in the
    "used" list and the unused portion in the "remaining" list.
    """
    credits = sorted((c.snapshot for c in credits), key=lambda v: v.creation_date)
    if amount == 0:
        return 0, [], credits
    if not credits:
        return amount, [], []

    used = []
    for credit in credits:
        if amount <= credit.amount:
            used.append(credit._replace(amount=amount))
            break

        # Another credit bites the dust!
        used.append(credit)
        amount -= credit.amount
    else:
        # We used all the credits!
        return amount, used, []

    # We covered the amount with the available credits!
    remaining = credits[len(used) :]
    if amount < credit.amount:
        remaining.insert(0, credit._replace(amount=credit.amount - amount))
    return 0, used, remaining


def remove_expired_credits():
    """Remove expired UtilityCredits across all accounts."""
    cutoff_date = utctoday()
    result = get_main_store().find(
        UtilityCredit, UtilityCredit.expiration_date < cutoff_date
    )
    result.remove()
