# Copyright 2005-2022 Canonical Limited. All rights reserved.

import uuid

from storm.expr import Sum

from canonical.lib.timestamp import utctoday

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import ComputerUtilisation, COMPUTER_TYPES
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.billing.collection import PaymentCollection
from canonical.landscape.model.billing.credit import UtilityCredit, subtract_credits
from canonical.landscape.model.billing.payment import Payment
from canonical.landscape.stripe.charge import MAX_ATTEMPTS


__all__ = ["BillingManagement"]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


class BillingManagement(ModelManagement):
    """Management class for all data related to billing"""

    def add_payment(self, period_start_date, period_end_date, credit_discount, vat):
        """Add a Payment."""
        payment_id = uuid.uuid4().hex
        return self.account.store.add(
            Payment(
                payment_id,
                self.account,
                period_start_date,
                period_end_date,
                credit_discount,
                vat,
            )
        )

    def get_outstanding_payments(self):
        """Return outstanding Payments for the account."""
        collection = PaymentCollection(
            self.account.store,
            account=self.account,
            statuses=["unpaid", "failed", "incomplete", "pending"],
        )
        return collection.find()

    def get_computer_utilisation(self):
        """Return managed hours by computer type.

        @return: a dict mapping computer types to managed machine hours for
            the type.
        """
        result = self._main_store.find(
            (ComputerUtilisation.computer_type, Sum(ComputerUtilisation.managed_hours)),
            ComputerUtilisation.account_id == self.account.id,
        )
        result.group_by(ComputerUtilisation.computer_type)

        utilisation = dict.fromkeys(COMPUTER_TYPES, 0)
        utilisation.update(result)
        return utilisation

    def reset_computer_utilisation(self):
        """Set ComputerUtilisation managed_hours back to 0 for the account."""
        # Remove entries for removed computers.
        result = self._main_store.find(
            ComputerUtilisation,
            ComputerUtilisation.account_id == self.account.id,
            ComputerUtilisation.computer_id == None,
        )
        result.remove()
        # Reset managed hours for remaining computers
        result = self._main_store.find(
            ComputerUtilisation, ComputerUtilisation.account_id == self.account.id
        )
        result.set(managed_hours=0)

    def reset_final_failed_payments(self):
        """Reset failed payments which are at their maximum retry attempt."""
        collection = PaymentCollection(
            self.account.store,
            account=self.account,
            statuses=["failed"],
            num_attempts=MAX_ATTEMPTS,
        )
        for payment in collection.find():
            payment.num_attempts = 0

    def add_credit(self, amount):
        """Add utility credit to the account."""
        return self._main_store.add(UtilityCredit(self.account, amount))

    def subtract_credit(self, amount):
        """Subtract uility credit from the account."""
        credits = list(self.get_credits())
        amount, used, _ = subtract_credits(amount, credits)
        if amount:
            raise RuntimeError("Available credit is less than amount to subtract")

        for i, snapshot in enumerate(used):
            credit = credits[i]
            if snapshot.amount == credit.amount:
                self._main_store.remove(credit)
            else:
                credit.amount -= snapshot.amount

    def get_credit_amount(self):
        """Return the amount of remaining credit."""
        amount = self._get_credits().sum(UtilityCredit.amount)
        if amount is None:  # no credit is present
            amount = 0
        return amount

    def get_credits(self):
        """Return non-expired UtilityCredits order of creation."""
        result = self._get_credits()
        return result.order_by(UtilityCredit.creation_date)

    def _get_credits(self):
        """Return non-expired UtilityCredits."""
        start_date = utctoday()
        return self._main_store.find(
            UtilityCredit,
            UtilityCredit.account_id == self.account.id,
            UtilityCredit.expiration_date >= start_date,
        )

    @property
    def _main_store(self):
        return get_main_store()
