# Copyright 2005-2019 Canonical Limited.  All rights reserved.

"""Incurred charges and payment models.

Accounts based on utility billing are billed monthly based on the managed
machine hours for the previous month.
"""

from datetime import datetime

from storm.locals import (
    Storm,
    Int,
    Date,
    DateTime,
    AutoReload,
    Reference,
    Enum,
    ReferenceSet,
    Unicode,
)

from canonical.landscape.model.entity.helper import AccountEntity
from canonical.landscape.model.main.computer import COMPUTER_TYPE_MAP


PAYMENT_STATUSES = {
    "unpaid": 1,
    "paid": 2,
    "failed": 3,
    "no-charge": 4,
    "pending": 5,
    "incomplete": 6,
}


class Payment(AccountEntity):
    """A payment for incurred costs in an account.

    @ivar payment_id: the unique ID of the payment.
    @ivar period_start_date: start date of the billing period.
    @ivar period_end_date: end date of the billing period.
    @ivar credit_discount: the amount of credit used as discount in 100ths of a
        cent.
    @ivar vat: the total VAT amount for the payment, in 100ths of a cent.
    @ivar status: the status of the payment.
    @ivar payment_date: the date and time charges for this Payment were paid.
    @ivar stripe_charge_id: ID of the Stripe charge associated to the payment,
        if any.
    @ivar stripe_error_code: error code from Stripe for failed Payments.
    @ivar stripe_error_message: error message from Stripe for failed Payments.
    @ivar stripe_charge_idempotency_key: the idempotency key for Stripe API
        payments.
    @ivar stripe_next_action: the URL of the user interaction required to
        have this payment processed.
    @ivar first_failure: the timestamp of the payment's first failure,
        if any.
    @ivar last_failure: the timestamp of the payment's most recent
        failure, if any.
    @ivar num_attempts: the number of times an attempt has been made
        to charge this payment.
    @ivar receipt_uri: the URI of the PDF receipt for this payment in Swift.
    """

    __storm_table__ = "payment"

    creation_time = DateTime(allow_none=False)
    period_start_date = Date(allow_none=False)
    period_end_date = Date(allow_none=False)
    credit_discount = Int(allow_none=False)
    vat = Int(allow_none=False)
    status = Enum(map=PAYMENT_STATUSES, allow_none=False)
    payment_date = DateTime()
    stripe_charge_id = Unicode()
    stripe_error_code = Unicode()
    stripe_error_message = Unicode()
    stripe_charge_idempotency_key = Unicode()
    stripe_next_action = Unicode()
    first_failure = DateTime()
    last_failure = DateTime()
    num_attempts = Int(allow_none=False, default=0)
    receipt_uri = Unicode()

    def __init__(
        self,
        payment_id,
        account,
        period_start_date,
        period_end_date,
        credit_discount,
        vat,
        status="unpaid",
    ):
        self.period_start_date = period_start_date
        self.period_end_date = period_end_date
        self.credit_discount = credit_discount
        self.vat = vat
        self.status = status
        super(Payment, self).__init__(account, payment_id)

    @property
    def payment_id(self):
        return self.data.name

    @property
    def total(self):
        """The total amount before tax."""
        return sum(cost.amount for cost in self.costs)

    @property
    def due(self):
        """The total due amount of the payment in 100th of cents."""
        return self.total - self.credit_discount + self.vat

    @property
    def total_hours(self):
        """The total machine amount of hours spent."""
        return sum(cost.hours for cost in self.costs)

    @property
    def receipt_id(self):
        """
        The ID of the receipt for this Payment, or None if it hasn't been paid.
        """
        if self.payment_date is None:
            return None

        return "{0.name}-{1:%Y}-{1:%m}".format(self.account, self.payment_date)

    def add_cost(
        self, computer_type, hourly_cost, hours, period_start_date, period_end_date
    ):
        """Add a cost to the payment."""
        cost = IncurredCost(
            self, computer_type, hourly_cost, hours, period_start_date, period_end_date
        )
        self.costs.add(cost)
        return cost

    def mark_paid(self, stripe_payment_id, timestamp=None):
        """Record a successful payment."""
        if not stripe_payment_id:
            raise RuntimeError("missing Stripe payment ID")
        if timestamp is None:
            timestamp = datetime.utcnow()
        self._record_result("paid", stripe_payment_id, timestamp=timestamp)

    def mark_pending(self, stripe_payment_id):
        """Record an incomplete payment."""
        if not stripe_payment_id:
            raise RuntimeError("missing Stripe payment ID")
        self._record_result("pending", stripe_payment_id)

    def mark_need_action(self, stripe_payment_id, next_action, error=None):
        """Record an incomplete payment requiring user action."""
        if not stripe_payment_id:
            raise RuntimeError("missing Stripe payment ID")
        self._record_result("incomplete", stripe_payment_id, next_action=next_action)

    def mark_failed(self, error, stripe_payment_id=None, timestamp=None):
        """Record a failed payment."""
        try:
            code, msg = error
        except (ValueError, TypeError):
            raise RuntimeError("expected (code, msg) for error, got {!r}".format(error))
        if not code:
            raise RuntimeError("missing error code")
        self._record_result("failed", stripe_payment_id, code, msg, timestamp)

    def _record_result(
        self,
        status,
        stripe_payment_id,
        errcode=None,
        errmsg=None,
        timestamp=None,
        next_action=None,
    ):
        paidts = errts = timestamp
        if errcode:
            if not errmsg:
                raise RuntimeError("got error code without error message")
            if errts is None:
                errts = datetime.utcnow()
            paidts = None
        elif errmsg:
            raise RuntimeError("got error message without error code")

        if self.status in ("unpaid", "failed"):
            self.num_attempts += 1
        self.status = status

        self.payment_date = paidts
        self.stripe_charge_id = stripe_payment_id
        self.stripe_next_action = next_action

        self.stripe_error_code = errcode
        self.stripe_error_message = errmsg

        if errcode:
            self.stripe_charge_idempotency_key = None
            if self.first_failure is None:
                self.first_failure = errts
            self.last_failure = errts


class IncurredCost(Storm):
    """A cost for managed hours of a computer type in a billing period.

    @ivar payment_id: id of the payment the cost is paid by.
    @ivar computer_type: the type of computer the cost is for.
    @ivar hourly_cost: hourly cost for management of the computer type.
    @ivar hours: number of managed hours.
    @ivar period_start_date: start date of the period the cost is for.
    @ivar period_end_date: end date of the period the cost is for.
    """

    __storm_table__ = "incurred_cost"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    payment_id = Int(allow_none=False)
    computer_type = Enum(map=COMPUTER_TYPE_MAP)
    hourly_cost = Int(allow_none=False)
    hours = Int(allow_none=False)
    period_start_date = Date(allow_none=False)
    period_end_date = Date(allow_none=False)

    payment = Reference(payment_id, Payment.id)

    def __init__(
        self,
        payment,
        computer_type,
        hourly_cost,
        hours,
        period_start_date,
        period_end_date,
    ):
        self.payment = payment
        self.computer_type = computer_type
        self.hourly_cost = hourly_cost
        self.hours = hours
        self.period_start_date = period_start_date
        self.period_end_date = period_end_date

    @property
    def amount(self):
        """The cost amount."""
        return self.hours * self.hourly_cost


Payment.costs = ReferenceSet(
    Payment.id,
    IncurredCost.payment_id,
    order_by=(IncurredCost.period_start_date, IncurredCost.computer_type),
)


def get_incurred_costs(store, payment_ids):
    """Return IncurredCosts.

    @param payment_ids: IDs of Payments the IncurredCosts are associated with
    """
    return store.find(IncurredCost, IncurredCost.payment_id.is_in(payment_ids))


def get_payment_by_id(store, id):
    """Return the payment with the specified ID."""
    return store.find(Payment, Payment.id == id).one()


def get_payments_with_status(store, statuses):
    """Return all payments matching the given payment statuses."""
    return store.find(Payment, Payment.status.is_in(statuses))
