# Copyright 2017 Canonical Limited.  All rights reserved.

from __future__ import division

import datetime

from canonical.landscape.model.billing.payment import Payment
from canonical.lib import random_string
from canonical.lib.timestamp import new_date_period
from canonical.landscape.stripe.testing import (
    new_stripe_charge_id,
    new_stripe_customer_id,
)


def new_payment_id():
    """Return a string that may be used as a payment ID."""
    return random_string(32, upper=False)


def add_cost(payment, total, hours=1, period=None, type="physical"):
    """Add the total cost to given payment.

    If hours is given then the total gets split up accordingly.  If no
    period is provided, the payment's period is used.
    """
    if period is None:
        period = (payment.period_start_date, payment.period_end_date)
    assert hours >= 0
    assert total % hours == 0
    hourly = total // hours
    start, end = period
    return payment.add_cost(type, hourly, hours, start, end)


def new_payment(
    account,
    id=None,
    credit=30,
    vat=10,
    period=None,
    timestamp=None,
    status=None,
    cost=None,
    add=False,
):
    """Return a new Payment on the given account with the provided info.

    The creation_time is set to the provided timestamp (or utcnow()
    if not provided).  The period is set to
    (timestamp - 30 days, timestamp).  If a cost is provided then it
    gets added.  If a status is provided then it gets set along with
    the extra fields that go with the status (e.g. stripe_charge_id).
    """
    if id is None:
        id = new_payment_id()
    if period is None:
        period = new_date_period()
    if timestamp is None:
        timestamp = datetime.datetime.utcnow()
    start, end = period
    payment = Payment(new_payment_id(), account, start, end, 30, 10)
    payment.creation_time = timestamp

    stripe_charge_id = new_stripe_charge_id()
    if status == "paid":
        payment.mark_paid(stripe_charge_id, timestamp)
        payment.stripe_charge_idempotency_key = "<some key>"
    elif status == "pending":
        payment.mark_pending(stripe_charge_id)
        payment.stripe_charge_idempotency_key = "<some key>"
    elif status == "failed":
        error = ("<error code>", "<error message>")
        payment.mark_failed(error, stripe_charge_id, timestamp)
    elif status:
        payment.status = status

    if status in ("paid", "pending"):
        payment.stripe_charge_idempotency_key = "<some key>"

    if account.stripe_customer_id is None:
        account.stripe_customer_id = new_stripe_customer_id()

    if add:
        payment = account.store.add(payment)
    if cost:
        add_cost(payment, cost)
    return payment
