# Copyright 2005-2016 Canonical Limited.  All rights reserved.

"""Functions to compute current utility usage."""

from builtins import object
from collections import namedtuple

from dateutil.relativedelta import relativedelta

from canonical.lib.timestamp import utctoday
from canonical.landscape.model.billing.management import BillingManagement
from canonical.landscape.model.billing.payment import get_incurred_costs
from canonical.landscape.model.billing.cost import get_utility_costs_over_period
from canonical.landscape.model.billing.credit import subtract_credits
from canonical.landscape.model.billing.vat import get_country_vat_rate


UsageSummary = namedtuple("UsageSummary", ["period", "usage", "cost"])


class CostSummary(
    namedtuple("CostSummary", "due usage vat vat_rate used_credit remaining_credits")
):
    """A summary of usage costs."""

    @property
    def remaining_credit(self):
        return sum(credit.amount for credit in self.remaining_credits)


class ComputerTypeUsage(object):
    """Usage and cost information for a computer type."""

    def __init__(self, computer_type, period, hourly_cost, hours):
        self.computer_type = computer_type
        self.period = period
        self.hourly_cost = hourly_cost
        self.hours = hours

    @property
    def cost(self):
        """The total cost for this usage."""
        return self.hourly_cost * self.hours


BILLING_COUNTRY = "GB"  # The country payments are invoiced from

BILLING_DAY = 28  # The day of the month for billing operations


def get_usage_summary(account, include_outstanding=True):
    """Return utilisation and costs for each machine type.

    @param include_outstanding: whether to include costs from outstanding
        Payments.
    """
    management = BillingManagement(account)
    credits = management.get_credits()
    period = _get_current_billing_period()

    discounts = 0
    usage = []
    if include_outstanding:
        payment_ids = []
        for payment in management.get_outstanding_payments():
            payment_ids.append(payment.id)
            # Add back the credit used by outstanding charges
            discounts += payment.credit_discount

        outstanding_costs = get_incurred_costs(account.store, payment_ids)
        usage.extend(_get_costs_usage(outstanding_costs))

    # add current usages last to keep the start_date ordering
    usage.extend(_get_current_usage(management.get_computer_utilisation(), period))
    vat_rate = _get_vat(account)
    cost = _get_cost(credits, discounts, vat_rate, usage)
    return UsageSummary(period, usage, cost)


def charge_account(account, min_charge=None):
    """Create a Payment with IncurredCosts for an account for current usage.

    The managed hours count for computers is reset after the invoice is
    created.

    @param min_charge: the minimum due cost that can be charged, in 100ths of
        cents. No payment is created if the due cost is less than this.
    """
    # XXX This should live up in the service layer as it is a high-level
    # operation over a collection of accounts.
    # This logic can't live in the service layer currently because it is also
    # run by a BatchScript and any service layer operation requires an
    # authorized principal defined with appropriate permissions. Moving this
    # logic back into the service layer would require a new LandscapePrincipal
    # or ScriptPrincipal which can be associated with the cron calls to
    # StripeChargeAccounts.
    # This also affects c.l.m.main.account.get_computers_by_type.
    summary = get_usage_summary(account, include_outstanding=False)
    if summary.cost.usage == 0:
        return

    if min_charge and summary.cost.due and summary.cost.due < min_charge:
        # A Payment is not created if a minimum charge is specified and the due
        # cost is below it, but not zero.  In the latter case, due cost is
        # fully covered by credit and a Payment is still created to document
        # the cost.
        return

    start_date, end_date = summary.period
    management = BillingManagement(account)
    payment = management.add_payment(
        start_date, end_date, summary.cost.used_credit, summary.cost.vat
    )
    management.subtract_credit(summary.cost.used_credit)

    utility_costs = get_utility_costs_over_period(start_date, end_date)
    for usage in summary.usage:
        if usage.cost:
            payment.add_cost(
                usage.computer_type,
                utility_costs[usage.computer_type],
                usage.hours,
                start_date,
                end_date,
            )

    if summary.cost.due == 0:
        # The credit covers the entire costs, mark the payment as not needing
        # charge.
        payment.status = "no-charge"

    management.reset_computer_utilisation()
    return payment


def _get_current_billing_period():
    """Return the current billing period."""
    today = utctoday()
    start = (today - relativedelta(months=1)).replace(day=BILLING_DAY)
    return start, today


def _get_current_usage(utilisation, period):
    """Return a list of ComputerTypeUsages for current costs.

    @param utilisation: a dict mapping computer types to managed machine hours.
    @param period: a tuple with start and end day of the billing period.
    """
    costs = get_utility_costs_over_period(*period)
    return [
        ComputerTypeUsage(computer_type, period, cost, utilisation[computer_type])
        for computer_type, cost in costs.items()
        if utilisation[computer_type] > 0
    ]


def _get_costs_usage(costs):
    """Return usage for outstanding costs by computer type."""
    return [
        ComputerTypeUsage(
            cost.computer_type,
            (cost.period_start_date, cost.period_end_date),
            cost.hourly_cost,
            cost.hours,
        )
        for cost in costs.order_by("period_start_date", "computer_type")
    ]


def _get_cost(credits, discounts, vat_rate, usage):
    """Return cost information."""
    usage_cost = sum(type_usage.cost for type_usage in usage)

    total_cost = usage_cost - discounts
    total_cost, _, remaining_credits = subtract_credits(total_cost, credits)
    used_credit = usage_cost - total_cost

    vat_cost = int(total_cost * vat_rate)
    due_cost = total_cost + vat_cost

    return CostSummary(
        due_cost, usage_cost, vat_cost, vat_rate, used_credit, remaining_credits
    )


def _get_vat(account):
    """Return the VAT percentage to apply for the account.

    The following rules apply for VAT calculation:

    * for private accounts:
      - for EU countries, the country VAT rate is applied
      - for other countries, no VAT is applied

    * for company accounts:
      - if the account country is the UK or they have no VAT
         number, UK VAT rate is applied.
      - otherwise no VAT is applied
    """
    rate = get_country_vat_rate(account.vat_country)
    if rate is None:
        # The account country is not a EU one, since we only track
        # rates for those countries. No VAT is applied in this case.
        return 0.0
    if not account.is_business:
        # Personal accounts in EU always have their country's rate.
        return rate
    else:
        # Company accounts
        if account.vat_country == BILLING_COUNTRY:
            # Don't bother looking up UK rate again.
            return rate
        elif not account.vat_number:
            return get_country_vat_rate(BILLING_COUNTRY)
        else:
            return 0.0
