# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int

from canonical.security.protect import PUBLIC, PRIVATE, requires

from canonical.storm.variables import BytesStr

from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.main.account import get_account_by_id


__all__ = ["CloudRegistrationOTP", "consume_otp_for_instance"]


__clearlist__ = ["consume_otp_for_instance"]


class CloudRegistrationOTP(object):
    """A one-time-password for automatic registration of cloud VMs.

    This is created by the Landscape server whenever it creates a VM on a
    cloud.  The OTP value should be passed to the booting instance through
    some secure out-of-band means, and when that computer tries to register
    with Landscape it will give the OTP back so that we can immediately and
    automatically register it as a Computer without going through the Pending
    stage.

    @param otp: A random C{str} password used as the primary key.
    @param instance: The C{ManagedInstance} attached to the password.
    @param account: Or the C{Account} attached to the password.
    """

    __storm_table__ = "registration_otp"

    otp = BytesStr(primary=True)
    account_id = Int(allow_none=False)

    # Note on permissions: OTPs are accessed by arbitrary clients. So
    # making this object public is OK as long as there are no
    # *factories* that return OTP objects without verifying the
    # one-time-password.
    requires(PUBLIC, PRIVATE, ["instance", "account"])

    def __init__(self, otp, account):
        self.otp = otp
        self.account_id = account.id

    @property
    def account(self):
        return get_account_by_id(self.account_id)


def consume_otp_for_instance(otp, instance_key):
    """
    Get the L{Account} which is associated with the given otp. The OTP is
    deleted so that it can't be used again.
    """
    stores = get_account_stores()
    for name, store in stores:
        otp_instance = store.find(CloudRegistrationOTP, otp=otp).one()
        if otp_instance is not None:
            account = otp_instance.account
            store.remove(otp_instance)
            return account
