# Copyright 2005-2016 Canonical Limited.  All rights reserved.
import json
import logging
import transaction

from pika import ConnectionParameters, PlainCredentials
from pika.adapters import twisted_connection
from pika.exceptions import AMQPConnectionError
from twisted.internet import defer, protocol, reactor
from storm.expr import Desc
from storm.twisted.transact import transact, Transactor

from canonical.landscape.hostagent.config import AMQP_CONFIG
from canonical.landscape.model.account.store import get_account_store_by_name
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.activity.status import BLOCKED
from canonical.landscape.model.activity.collection import (
    ActivityParentCollection,
    ActivityChildCollection,
)
from canonical.security.protect import requires, requires_advisor


__all__ = [
    "DeleteChildComputerActivity",
    "get_computer_activities_parents",
    "get_next_parent_with_blocked_computer_activity",
    "GRPCActivityBase",
    "StartChildComputerActivity",
    "StopChildComputerActivity",
]

__clearlist__ = __all__


def get_computer_activities_parents(
    account, computer_ids=None, statuses=None, info_classes=None, ordered=False
):
    """Return parent activities with children related to specified computers.

    @param account: the Account to search in.
    @param computer_ids: optionally IDs of the computers to search activities
        for. If no IDs are specified, activities for all computers are
        searched.
    @param statuses: optionally limit child activities to specified statuses.
    @param info_classes: optionally limit child activities to the specified
        info classes.
    @param ordered: whether to order parents by creation time and id.
    """
    if not info_classes:
        # Allow empty sequences as parameter
        info_classes = None

    child_collection = ActivityChildCollection(
        account.store,
        account=account,
        statuses=statuses,
        info_classes=info_classes,
        entity_class=AccountComputerTwin,
        entity_ids=computer_ids,
    )
    result = child_collection.find()
    result = result.config(distinct=True)
    parent_ids = result.values(Activity.parent_id)

    # Return parents of the children previously found
    parent_collection = ActivityParentCollection(
        account.store, account=account, ids=parent_ids
    )
    result = parent_collection.find()
    if ordered:
        result = result.order_by(
            Desc(ActivityParent.creation_time), Desc(ActivityParent.id)
        )
    return result.find()


def get_next_parent_with_blocked_computer_activity(activity_parent, computer_ids=None):
    """Return the next parent activity with children in BLOCKED status.

    Parent activities are considered based on their creation time.

    If computer_ids are passed, only parents which children involving specified
    computers are considered.
    """
    account = activity_parent.account

    parents_result = get_computer_activities_parents(
        account, computer_ids=computer_ids, statuses=[BLOCKED]
    )
    parents_result.order_by(ActivityParent.creation_time)
    result = parents_result.find(
        ActivityParent.creation_time > activity_parent.creation_time
    )
    next_blocked = result.first()

    if next_blocked is None:
        # It could be that we got the last activity, try to loop around to the
        # first blocked activity.
        result = parents_result.find(
            ActivityParent.creation_time < activity_parent.creation_time
        )
        next_blocked = result.first()

    return next_blocked


@requires_advisor
class GRPCActivityBase(ComputerRequest):
    """
    Base class for Activities that should put a message on the queue
    for the GRPC server to send along to hostagents.

    Subclasses should implement the `get_message`, `get_summary`,
    `should_cancel_when_queued`, and `cancel_message` methods.
    """

    def __init__(self, activity, computer):
        super().__init__(activity, computer)
        self.message = self.get_message()
        self.queue = None
        self.parent = self.computer.get_parent_computer()
        if self.parent:
            self.queue = self.parent.secure_id.decode()
        self.activity_id = activity.id
        self.account_id = activity.account_id
        self.transactor = Transactor(reactor.getThreadPool())

    def get_message(self):
        """
        Subclasses should construct the message to put on the queue
        for the GRPC message relay server to send to the hostagent.
        """
        raise NotImplementedError()

    def get_summary(self):
        """
        Summary description of the activity.
        """
        raise NotImplementedError()

    def should_cancel_when_queued(self):
        """
        If `True` we cancel the activity immediately when queued.
        If `False` we start the activity.
        """
        raise NotImplementedError()

    def cancel_message(self):
        """
        Message to display when the activity is
        automatically canceled.
        """
        raise NotImplementedError()

    def _queued(self):
        if self.should_cancel_when_queued():
            self.activity.cancel(text=self.cancel_message())
        else:
            self.activity.start()

    def _started(self):
        deferred = defer.ensureDeferred(self.send_grpc_message())
        deferred.addErrback(self._fail)

    @transact
    def _fail(self, failure):
        """Mark the activity as failed."""
        logging.error(failure)
        account = get_account_by_id(self.account_id)
        activity = get_child_activity(account, self.activity_id)
        activity.fail(text=str(failure.value))

    async def send_grpc_message(self):
        config = AMQP_CONFIG
        credentials = PlainCredentials(
            username=config["user"], password=config["password"]
        )
        params = ConnectionParameters(
            host=config["host"],
            port=config["port"],
            virtual_host=config["vhost"],
            credentials=credentials,
        )
        client_creator = protocol.ClientCreator(
            reactor,
            twisted_connection.TwistedProtocolConnection,
            params,
        )
        connection = await client_creator.connectTCP(
            host=config["host"], port=config["port"]
        )
        await connection.ready
        channel = await connection.channel()
        if channel:
            await channel.queue_declare(queue=self.queue)
            await channel.basic_publish(
                exchange="", routing_key=self.queue, body=self.message
            )


@requires_advisor
class DeleteChildComputerActivity(GRPCActivityBase):
    """
    Activity to put a message on the queue for a hostagent to delete a child
    computer instance.
    """

    @requires("ViewActivity")
    def get_summary(self):
        summary = f"Delete instance {self.computer.title}"
        return summary

    def get_message(self):
        return json.dumps({"delete": self.computer.id})

    def should_cancel_when_queued(self):
        return not self.computer.delete_instance_supported()

    def cancel_message(self):
        return f"Delete instance not supported for computer {self.computer.title}"


@requires_advisor
class StartChildComputerActivity(GRPCActivityBase):
    """
    Activity to put a message on the queue for a hostagent to start a child
    computer instance.
    """

    @requires("ViewActivity")
    def get_summary(self):
        summary = f"Start instance {self.computer.title}"
        return summary

    def get_message(self):
        return json.dumps({"start": self.computer.id})

    def should_cancel_when_queued(self):
        return not self.computer.start_instance_supported()

    def cancel_message(self):
        return f"Start instance not supported for computer {self.computer.title}"


@requires_advisor
class StopChildComputerActivity(GRPCActivityBase):
    """
    Activity to put a message on the queue for a hostagent to stop a child
    computer instance.
    """

    @requires("ViewActivity")
    def get_summary(self):
        summary = f"Stop instance {self.computer.title}"
        return summary

    def get_message(self):
        return json.dumps({"stop": self.computer.id})

    def should_cancel_when_queued(self):
        return not self.computer.stop_instance_supported()

    def cancel_message(self):
        return f"Stop instance not supported for computer {self.computer.title}"
