# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime, timedelta

from storm.expr import Func
from storm.locals import Store

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.mail.template import load_template
from canonical.landscape.model.stormfunc import AgeFromNow
from canonical.landscape.model.alert.rule import RuleState
from canonical.landscape.model.alert.computer import (
    ComputerRuleInfo,
    call_for_enabled_rules_with_transaction,
    ComputerAlertEmailViewBase,
)
from canonical.landscape.model.main.computer import (
    PingTime,
    PingServer,
    Computer,
    ComputerStatus,
)
from canonical.landscape.model.resource.rebootrequiredinfo import (
    get_reboot_required_info_for_computers,
)


__all__ = [
    "ComputerOfflineAlert",
    "ComputerRebootAlert",
    "ComputerDuplicateAlert",
    "check_offline_computers",
]


__requires_prefix__ = "landscape."


@requires_advisor
class ComputerDuplicateAlert(ComputerRuleInfo):
    """Alert activates when duplicate hostnames are registered."""

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when a duplicate computer exists"

    notification_delay = timedelta(minutes=5)

    resolution_url = "/computers/criteria/alert:computer-duplicates"

    def get_summary(self, computer_ids=None):
        """Get a summary stating how many duplicates computers are presents.

        @param computer_ids: a list of computers ids to limit the generated
            summary to.
        """
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_id_count = result.count()
        if computer_id_count == 1:
            return "{} computer has duplicates".format(computer_id_count)
        elif computer_id_count > 1:
            return "{} computers have duplicates".format(computer_id_count)
        else:
            return "No computers have duplicates"

    def check(self, computer_ids):
        """Update alert status for computer_ids."""
        if not computer_ids:
            return

        duplicates = set()
        uniques = set()

        store = Store.of(self.rule.account)
        result = store.find(
            (Computer.hostname, Func("array_agg", Computer.id)),
            Computer.account == self.rule.account.id,
        ).group_by(Computer.hostname)

        for hostname, computers in result:
            if len(computers) > 1:
                duplicates.update(computers)
            else:
                uniques.update(computers)
        self.update_states(duplicates, uniques)


@requires_advisor
class ComputerOfflineAlert(ComputerRuleInfo):
    """
    @ivar ping_check_interval: Minutes elapsed since the last ping time to
        consider a L{Computer} offline.
    @type ping_check_interval: C{int}
    @ivar exchange_check_interval: Minutes elapsed since the last exchange
        time to consider a L{Computer} offline.
    @type exchange_check_interval: C{int}
    """

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when computer has not contacted Landscape for some time"

    notification_delay = timedelta(minutes=5)

    ping_check_interval = 5
    exchange_check_interval = 60
    maximum_last_update_delay = 2

    resolution_url = "/computers/criteria/alert:computer-offline"

    def get_summary(self, computer_ids=None):
        """Get a summary stating how many computers are offline.

        @param computer_ids: a list of computers ids to limit the generated
            summary to.
        """
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_id_count = result.count()
        if computer_id_count == 1:
            return (
                "1 computer hasn't contacted Landscape within the "
                "last %d minutes" % self.ping_check_interval
            )
        elif computer_id_count > 1:
            return (
                "%d computers haven't contacted Landscape within the "
                "last %d minutes" % (computer_id_count, self.ping_check_interval)
            )
        else:
            return "No computers offline"

    def check(self, computer_ids):
        """Update online/offline status for C{computer_ids}."""
        if not computer_ids:
            return
        status = ComputerOfflineStatus(
            self,
            computer_ids,
            self.ping_check_interval,
            self.exchange_check_interval,
            self.maximum_last_update_delay,
        )
        offline_computer_ids, online_computer_ids = status.get()
        self.update_states(offline_computer_ids, online_computer_ids)


class ComputerOfflineStatus(object):
    """Algorithm calculates online/offline status for L{Computer}s.

    @param alert: A L{ComputerOfflineAlert}.
    @param computer_ids: A set of ids for L{Computer}s to check.
    @param ping_check_interval: An C{int} representing the longest interval in
        minutes without a ping before a L{Computer} is considered offline.
    @param exchange_check_interval: An C{int} representing the longest
        interval in minutes without a exchange before a L{Computer} is
        considered offline.
    @param maximum_last_update_delay: An C{int} representing the longest time
        in minutes without changes to L{PingServer.last_update_time} before a
        ping server is considered offline.
    """

    def __init__(
        self,
        alert,
        computer_ids,
        ping_check_interval,
        exchange_check_interval,
        maximum_last_update_delay,
    ):
        self.alert = alert
        self.computer_ids = computer_ids
        delta = lambda minutes: timedelta(minutes=minutes)
        self.ping_check_interval = delta(ping_check_interval)
        self.exchange_check_interval = delta(exchange_check_interval)
        self.maximum_last_update_delay = delta(maximum_last_update_delay)
        self.store = Store.of(self.alert.rule.account)

    def get(self):
        """Calculate online/offline status for L{computer_ids}.

        @return: A 2-tuple of C{(offline_computer_ids, online_computer_ids)}.
        """
        self._offline_computer_ids = set()
        self._online_computer_ids = set()
        result = self.alert.get_states(active=True)
        already_offline = list(result.values(RuleState.computer_id))
        ping_check_time = datetime.utcnow() - self.ping_check_interval
        ping_server_up = self._is_ping_server_up()
        ping_computer_ids = self._check_ping_times(ping_check_time, ping_server_up)
        self._check_unknown_ping_times(
            ping_check_time, ping_server_up, ping_computer_ids, already_offline
        )
        exchange_check_time = datetime.utcnow() - self.exchange_check_interval
        self._check_exchange_times(exchange_check_time)
        return self._offline_computer_ids, self._online_computer_ids

    def _is_ping_server_up(self):
        """Check to see if a ping server is up.

        A ping server is considered to be up and running if the
        L{last_update_time} is within the last L{maximum_last_update_delay}
        minutes of the L{ping_check_interval}.

        @return: True if a ping server is up, otherwise False.
        """
        startup_age = AgeFromNow(PingServer.startup_time)
        last_update_age = AgeFromNow(PingServer.last_update_time)
        result = self.store.find(
            PingServer,
            startup_age > self.ping_check_interval,
            last_update_age < self.maximum_last_update_delay,
        )
        return not result.is_empty()

    def _check_ping_times(self, check_time, ping_server_up):
        """Check L{Computer}s with ping times and update result sets.

        L{Computer}s that have a ping time older than C{check_time}, while a
        ping server is up, will be added to the offline set.  The rest will be
        added to the online set.

        @param check_time: A C{datetime}, in UTC, representing the earliest
            L{PingTime.ping_time} considered as being online.
        @param ping_server_up: A flag, True if the a ping server is up, False
            otherwise.
        @return: The set of L{Computer.id}s with ping times.
        """
        result = self.store.find(
            PingTime,
            PingTime.computer_id.is_in(self.computer_ids),
            PingTime.last_ping_time != None,
        )
        for ping_time in result:
            if ping_time.last_ping_time < check_time:
                if ping_server_up:
                    self._offline_computer_ids.add(ping_time.computer_id)
            else:
                self._online_computer_ids.add(ping_time.computer_id)
        return self._offline_computer_ids.union(self._online_computer_ids)

    def _check_unknown_ping_times(
        self, check_time, ping_server_up, all_ping_computer_ids, already_offline
    ):
        """Check L{Computer}s with unknown ping times and update result sets.

        The ping time is unknown if there is no ping time row at all, or if
        there's a ping time without a value.  Computers that are already
        offline are can't come online, and are excluded from checking logic.
        For computers that have been accepted but have not pinged yet, we only
        update their status, ever, if the ping server is up.  Otherwise we
        can't know for sure if they've been online or not.  We only want to
        consider them truly offline if (a) the ping server was up at all and
        (b) their accept time falls within the last L{ping_check_interval}.

        @param check_time: A C{datetime}, in UTC, representing the earliest
            L{Computer.accept_time} considered as being online.
        @param ping_server_up: A flag, True if the a ping server is up, False
            otherwise.
        @param already_offline: A list of L{Computer.id}s already known to be
            offline.
        """
        unknown_computer_ids = set(self.computer_ids) - all_ping_computer_ids
        unknown_computer_ids = unknown_computer_ids.difference(already_offline)
        if ping_server_up and unknown_computer_ids:
            computers_accept_time = self.store.find(
                (Computer.id, Computer.accept_time),
                Computer.id.is_in(unknown_computer_ids),
            )
            for computer_id, accept_time in computers_accept_time:
                if accept_time < check_time:
                    self._offline_computer_ids.add(computer_id)
                else:
                    self._online_computer_ids.add(computer_id)
        self._offline_computer_ids = self._offline_computer_ids.difference(
            already_offline
        )

    def _check_exchange_times(self, check_time):
        """Check L{Computer}s with exchange times and update result sets.

        L{Computer}s that have an exchange time older than C{check_time}, but
        are considered online based on other checks, will be removed from the
        online set and added to the offline set.

        @param check_time: A C{datetime}, in UTC, representing the earliest
            L{Computer.last_exchange_time} considered as being online.
        """
        result = self.store.find(
            (ComputerStatus.computer_id, ComputerStatus.last_exchange_time),
            ComputerStatus.computer_id.is_in(self.computer_ids),
            ComputerStatus.last_exchange_time != None,
            ComputerStatus.soft_lock_time == None,
        )
        for computer_id, exchange_time in result:
            if exchange_time < check_time:
                if computer_id in self._online_computer_ids:
                    self._online_computer_ids.remove(computer_id)
                    self._offline_computer_ids.add(computer_id)


def check_offline_computers():
    """
    Get every L{Account} containing the L{ComputerOfflineAlert} rule, and check
    the rule for all L{Computer}s in each account.
    """

    def callback(account, computer_ids):
        account.check_rules([ComputerOfflineAlert], computer_ids=computer_ids)

    call_for_enabled_rules_with_transaction(ComputerOfflineAlert, callback)


class ComputerOfflineAlertEmailView(ComputerAlertEmailViewBase):

    _content_template = load_template(__file__, "computer-offline-alert.msg")

    def sort_computers(self, computers):
        """
        Returns a list of L{Computer}s, ordered by timestamp, in descending
        order.
        """
        return sorted(computers, key=self._affected_computers_sort_key, reverse=True)

    def _affected_computers_sort_key(self, computer):
        """
        Uses the last_ping_time as sort key and the oldest possible
        timestamp in the case where the computer never pinged us.
        """
        if computer.last_ping_time is None:
            return datetime.utcfromtimestamp(0)
        else:
            return computer.last_ping_time.replace(tzinfo=None)


@requires_advisor
class ComputerRebootAlert(ComputerRuleInfo):
    """Alert activates when computers need to be rebooted."""

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when computer needs to be rebooted"

    notification_delay = timedelta(minutes=5)

    resolution_url = "/computers/criteria/alert:computer-reboot"

    def get_summary(self, computer_ids=None):
        """Get a summary stating how many computers need to be rebooted.

        @param computer_ids: a list of computers ids to limit the generated
            summary to.
        """
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_id_count = result.count()
        if computer_id_count == 1:
            return "1 computer needs to be rebooted"
        elif computer_id_count > 1:
            return "%d computers need to be rebooted" % (computer_id_count,)
        else:
            return "No computers need to be rebooted"

    def check(self, computer_ids):
        if not computer_ids:
            return

        result = self.rule.account.get_computers(
            computer_ids=computer_ids, reboot_required_flag=True
        )
        active_computer_ids = set(result.values(Computer.id))
        inactive_computer_ids = set(computer_ids) - active_computer_ids
        self.update_states(active_computer_ids, inactive_computer_ids)


class ComputerRebootAlertEmailView(ComputerAlertEmailViewBase):

    _content_template = load_template(__file__, "computer-reboot-alert.msg")

    def __init__(self, context, person):
        super(ComputerRebootAlertEmailView, self).__init__(context, person)
        packages = set([])
        for reboot_required_info in get_reboot_required_info_for_computers(
            self.affected_computers
        ):
            if reboot_required_info.packages:
                packages.update(reboot_required_info.packages)
        self.affected_packages = sorted(packages)
