# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Collection logic to filter and find L{Computer}s."""

from storm.expr import LeftJoin
from storm.locals import Like, Or, Select
from storm.store import EmptyResultSet
from storm.variables import UnicodeVariable

from canonical.collection import CollectionBase
from canonical.landscape.model.entity.collection import AccountEntityCollection
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import Computer, ComputerStatus
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.main.account import get_account_by_id


__all__ = ["ComputerCollection", "AccountComputerTwinCollection"]

__clearlist__ = __all__


class ComputerCollection(CollectionBase):
    """A collection of L{Computer}s that can be filtered in various ways."""

    def __init__(
        self,
        store=None,
        computer_ids=None,
        tag_ids=None,
        keywords=None,
        person_id=None,
        permission=None,
        account_id=None,
        reboot_required_flag=None,
        access_context_ids=None,
        with_ping_times=None,
    ):
        if store is None:
            store = get_main_store()
        if (person_id or permission) and not (person_id and permission):
            raise RuntimeError("person_id and permission must be provided together.")
        super(ComputerCollection, self).__init__(
            store,
            computer_ids=computer_ids,
            tag_ids=tag_ids,
            keywords=keywords,
            person_id=person_id,
            permission=permission,
            account_id=account_id,
            reboot_required_flag=reboot_required_flag,
            access_context_ids=access_context_ids,
            with_ping_times=with_ping_times,
        )

    def with_ping_times(self):
        """Make it possible to sort the result by L{PingTime}.

        @return: A new L{ComputerCollection} with L{PingTime}s.
        """
        return self._filter_by(with_ping_times=True)

    def with_computer_ids(self, computer_ids):
        """Only include L{Computer}s with a C{Computer.id} in C{computer_ids}.

        If C{computer_ids} is an empty the list the collection will yield an
        C{EmptyResultSet} when its L{find} method is called.

        @param computer_ids: A sequence of L{Computer.id}s to filter by.
        @return: A new L{ComputerCollection} filtered by L{Computer.id}s.
        """
        return self._filter_by(computer_ids=computer_ids)

    def with_search_criteria(self, search_criteria):
        """Only include L{Computer}s matching the search_criteria.

        @param search_criteria: A search criteria passed to
            L{ComputerSearch.find}.
        @return: A new L{ComputerCollection} filtered by C{search_criteria}
            using the L{ComputerSearch} engine.
        """
        from canonical.landscape.model.computer.search import ComputerSearch

        account = get_account_by_id(self._where["account_id"])
        search = ComputerSearch(account)
        computer_ids = set(search.find(search_criteria).values(Computer.id))
        if self._where["computer_ids"]:
            computer_ids.intersection_update(self._where["computer_ids"])
        return self._filter_by(computer_ids=computer_ids)

    def with_keywords(self, keywords):
        """Only include L{Computer}s that match C{keywords}.

        @param keywords: A sequence of keywords to filter L{Computer}s by.
        @return: A new L{ComputerCollection} filtered by keywords.
        """
        return self._filter_by(keywords=keywords)

    def with_permission(self, person, permission):
        """
        Only include L{Computer}s that are members of an L{AccessContext} for
        which C{person} has a L{Role} granting C{permission}.

        @param person: The L{Person} to check C{permission} for when filtering
            L{Computer}s.
        @param permission: The permission required for L{Computer}s.
        @return: A new L{ComputerCollection} filtered by L{Person} and
            C{permission}.
        """
        return self._filter_by(person_id=person.id, permission=permission)

    def with_account(self, account):
        """Only include L{Computer}s that are members of an L{Account}.

        @param account: An L{Account} to filter L{Computer}s by.
        @return: A new L{ComputerCollection} filtered by L{Account}.
        """
        return self._filter_by(account_id=account.id)

    def with_tags(self, tags):
        """Only include L{Computer}s that have a matching L{Tag}.

        @param tags: A sequence of, optionally mixed, L{Tag}s, L{Tag.id}s or
            strings.
        @return: A new L{ComputerCollection} filtered by L{Tag}s.
        """
        from canonical.landscape.model.main.tag import get_tags

        tag_ids = [tag.id for tag in get_tags(tags)]
        return self._filter_by(tag_ids=tag_ids)

    def with_access_groups(self, access_groups):
        """Only include L{Computer}s that have a matching L{AccessContext}.

        @param access_groups: A sequence of L{AccessContext.name}s
        @return: A new L{ComputerCollection} filtered by L{AccessContext}s.
        """
        account = get_account_by_id(self._where["account_id"])
        access_context_ids = [
            context.id for context in account.get_access_contexts(names=access_groups)
        ]
        return self._filter_by(access_context_ids=access_context_ids)

    def with_reboot_required_flag(self, reboot_required_flag=True):
        """
        Only include L{Computer}s that have a C{reboot_required_flag} attribute
        that matches C{condition}.

        @param condition: Optionally, a C{bool} indicating whether or not to
            include L{Computer}s that require a reboot.  Defaults to C{True}.
        """
        return self._filter_by(reboot_required_flag=reboot_required_flag)

    def with_access_contexts(self, access_contexts):
        """Only include L{Computer}s that are members of C{access_contexts}.

        @param access_contexts: A sequence of L{AccessContext}s.
        @return: A new L{ComputerCollection} filtered by L{AccessContext}.
        """
        access_context_ids = [access_context.id for access_context in access_contexts]
        return self._filter_by(access_context_ids=access_context_ids)

    def _get_result(
        self,
        store,
        where=None,
        computer_ids=None,
        tag_ids=None,
        keywords=None,
        person_id=None,
        permission=None,
        account_id=None,
        reboot_required_flag=None,
        access_context_ids=None,
        with_ping_times=None,
    ):
        """
        Return a L{ResultSet} configured to yield L{Computer}s matching
        filter criteria.
        """
        from canonical.landscape.model.main.tag import ComputerTag
        from canonical.landscape.model.main.computer import Computer, PingTime

        if where is None:
            where = []
        tables = []
        if computer_ids is not None:
            if computer_ids:
                where.append(Computer.id.is_in(computer_ids))
            else:
                return EmptyResultSet()
        if tag_ids:
            tables.append(ComputerTag)
            where.append(ComputerTag.computer_id == Computer.id)
            where.append(ComputerTag.tag_id.is_in(tag_ids))
        if keywords:
            keyword_where = []
            for keyword in keywords:
                keyword = UnicodeVariable("%%%s%%" % (keyword,))
                keyword_where.append(
                    Like(Computer.title, keyword, case_sensitive=False)
                )
                keyword_where.append(
                    Like(Computer.hostname, keyword, case_sensitive=False)
                )
            where.append(Or(*keyword_where))
        if person_id and permission:
            where.extend(
                self._create_with_permission_expressions(
                    store, account_id, person_id, permission
                )
            )
        if account_id:
            where.append(Computer.account_id == account_id)
        if reboot_required_flag is not None:
            tables.append(ComputerStatus)
            where.append(ComputerStatus.computer_id == Computer.id)
            where.append(ComputerStatus.reboot_required_flag == reboot_required_flag)
        if access_context_ids:
            where.append(Computer.access_context_id.is_in(access_context_ids))

        if with_ping_times:
            join = LeftJoin(Computer, PingTime, Computer.id == PingTime.computer_id)
            if tag_ids:
                result = store.find(Computer, *where)
                subselect = result.get_select_expr(Computer.id)
                result = store.using(join).find(Computer, Computer.id.is_in(subselect))
            else:
                result = store.using(tables + [join]).find(Computer, *where)
        elif tables:
            tables.insert(0, Computer)
            result = store.using(tables).find(Computer, *where)
            result.config(distinct=bool(tag_ids))
        else:
            # Let's not use "using" if we can avoid it. That allows us to do
            # more things with the ResultSet, like call the set method.
            result = store.find(Computer, *where)
            result.config(distinct=bool(tag_ids))

        if account_id:
            account = get_account_by_id(account_id)
            # Since get_roles_by_computer() doesn't use a cache for
            # the self-service case, there's no need to preload the
            # cache.
            if not account.is_self_service:
                account.preload_licenses(result.values(Computer.id))
        return result

    def _create_with_permission_expressions(
        self, store, account_id, person_id, permission
    ):
        """Return appropriate Storm expressions for the permission filter."""
        from canonical.landscape.model.main.computer import Computer
        from canonical.landscape.model.main.accesscontext import AccessContext
        from canonical.landscape.model.main.person import Person

        person = store.get(Person, person_id)
        person_roles = person.get_all_roles()

        registry = person.get_role_registry()

        if account_id is not None:
            account = get_account_by_id(account_id)
            if account.access_context_id in person_roles:
                if registry.may(person_roles[account.access_context_id], permission):
                    # The person has a role granting the required permission
                    # on the account main access context. In this case the
                    # account filter is enough, as the person will have that
                    # permission on all computer in the account.
                    return ()

        access_context_ids = []
        for access_context_id in person_roles:
            if registry.may(person_roles[access_context_id], permission):
                if access_context_id is None:
                    # The person has a global role granting the required
                    # permission, there's no need to filter at all.
                    return ()

                access_context_ids.append(access_context_id)

        # A person with a role in a parent access context should have access to
        # computers in its child access contexts.  A subselect is used to get
        # the AccessContext.ids of all access contexts whose parent is one of
        # the access contexts the person has a suitable role on.
        # If Computer.access_context_id matches any of these values, the
        # computer is included in the collection.
        subselect = Select(
            AccessContext.id, where=[AccessContext.parent_id.is_in(access_context_ids)]
        )

        return [
            Or(
                Computer.access_context_id.is_in(access_context_ids),
                Computer.access_context_id.is_in(subselect),
            )
        ]


class AccountComputerTwinCollection(AccountEntityCollection):

    model_class = AccountComputerTwin
