from builtins import str
from builtins import map
from builtins import range
from builtins import object
import random

from datetime import datetime, timedelta

from landscape import SERVER_API, VERSION

from canonical.storm.relation import RelationLink, Relation, RelationResolver

from canonical.landscape.model.main.accesscontext import AccessContext
from canonical.landscape.model.main.tag import Tag, ComputerTag, get_tags
from canonical.landscape.model.main.distribution import Distribution
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.computer import (
    Computer,
    ComputerStatus,
    ComputerUtilisation,
    PingTime,
    get_computer_by_secure_id,
)
from canonical.landscape.model.computer.twin import ComputerTwin, AccountComputerTwin
from canonical.landscape.model.entity.data import AccountEntityData
from canonical.landscape.model.entity.helper import get_account_entity_type
from canonical.landscape.model.resource.rebootrequiredinfo import RebootRequiredInfo
from canonical.landscape.model.account.store import get_account_store_by_name
from canonical.landscape.model.resource.store import get_resource_stores
from canonical.landscape.model.resource.package import ComputerPackages


WSL_HOST_CODE_NAME = "windows"


# Map of possible distributions that computers can be associated with
DISTRIBUTIONS = {
    "dapper": ("Ubuntu 6.06 LTS", "Ubuntu", "6.06"),
    "edgy": ("Ubuntu 6.10", "Ubuntu", "6.10"),
    "feisty": ("Ubuntu 7.04", "Ubuntu", "7.04"),
    "gutsy": ("Ubuntu 7.10", "Ubuntu", "7.10"),
    "hardy": ("Ubuntu 8.04 LTS", "Ubuntu", "8.04"),
    "jaunty": ("Ubuntu 9.04", "Ubuntu", "9.04"),
    "karmic": ("Ubuntu 9.10", "Ubuntu", "9.10"),
    "lucid": ("Ubuntu 10.04 LTS", "Ubuntu", "10.04"),
    "natty": ("Ubuntu 11.04", "Ubuntu", "11.04"),
    "oneiric": ("Ubuntu 11.10", "Ubuntu", "11.10"),
    "precise": ("Ubuntu 12.04 LTS", "Ubuntu", "12.04"),
    "trusty": ("Ubuntu 14.04 LTS", "Ubuntu", "14.04"),
    "xenial": ("Ubuntu 16.04 LTS", "Ubuntu", "16.04"),
    "bionic": ("Ubuntu 18.04 LTS", "Ubuntu", "18.04"),
    WSL_HOST_CODE_NAME: ("Windows 10 / Windows 11", "Microsoft", "10 / 11"),
}

DISTRIBUTION_COLUMNS = (
    Distribution.description,
    Distribution.distributor,
    Distribution.release,
    Distribution.code_name,
)

TAG_COLUMNS = (Tag.name,)

COMPUTER_COLUMNS = (
    Computer.account_id,
    Computer.secure_id,
    Computer.title,
    Computer.access_context_id,
    Computer.resource_database_key,
    Computer.distribution_id,
    Computer.hostname,
    Computer.vm_info,
    Computer.container_info,
    Computer.current_api,
    Computer.current_version,
)

COMPUTER_UTILISATION_COLUMNS = (
    ComputerUtilisation.account_id,
    ComputerUtilisation.computer_id,
    ComputerUtilisation.computer_type,
    ComputerUtilisation.managed_hours,
    ComputerUtilisation.next_managed_hour_time,
)

ACCOUNT_ENTITY_DATA_COLUMNS = (
    AccountEntityData.account_id,
    AccountEntityData.entity_type,
    AccountEntityData.name,
    AccountEntityData.access_context_id,
)

ACCOUNT_COMPUTER_TWIN_COLUMNS = (
    AccountComputerTwin.id,
    AccountComputerTwin.data_id,
)

COMPUTER_TWIN_COLUMNS = (
    ComputerTwin.id,
    ComputerTwin.account_id,
)

COMPUTER_TAG_COLUMNS = (
    ComputerTag.computer_id,
    ComputerTag.tag_id,
)

COMPUTER_STATUS_COLUMNS = (
    ComputerStatus.computer_id,
    ComputerStatus.consecutive_errors,
    ComputerStatus.reboot_required_flag,
)

PING_TIME_COLUMNS = (
    PingTime.computer_id,
    PingTime.last_ping_time,
)

REBOOT_REQUIRED_INFO_COLUMNS = (
    RebootRequiredInfo.computer_id,
    RebootRequiredInfo.packages,
)

COMPUTER_PACKAGES_COLUMNS = (ComputerPackages.computer_id,)


class ComputerProfile(object):
    """Create one or more computers.

    This class is a helper to create a set of computers from a given
    profile specifications expressed with dictionaries.

    Example:

    profile = ComputersProfile(
        computers={
            "computer-1": {
                "account": u"account-1",
                "title": u"Computer 1",
            },
            "computer-2": {
                "account": u"account-1",
                "title": u"Computer 2",
                "hostname": u"some.host",
                "tags: [u"server", u"xenial"],
            },
        }
    )
    profile.create()

    will create a two computers with the given data. More complex setups are
    possible by passing additional info about the the computers.

    See unit tests and c/l/sample/profiles.py for more comprehensive examples.
    """

    def __init__(self, computers=None, resolver=None):
        self.computers = computers or {}
        self.resolver = resolver or RelationResolver()
        self.count = 0  # Keep track of the number of computers created so far

    def create(self, materialize=False):
        """Create the computers of this profile.

        @return: The list of newly created Computers (if materialize is
            set to True).
        """
        factory = ComputerFactory(self.resolver)
        deferred_tags = {}
        secure_ids = []

        for key, info in self.computers.items():

            account = info["account"]
            # Human-readable secure ID, for easier reference
            secure_id = key
            secure_ids.append(secure_id)

            factory.create_computer(
                account,
                secure_id,
                info["title"],
                access_group=info.get("access-group"),
                managed_hours=info.get("managed-hours", 0),
                computer_type=self._get_computer_type(info),
            )

            tags = info.get("tags", ())
            if info.get("rules-states", True):
                # Conditionally create computer rules, since they are expensive
                # and not always needed (for example for tests). In case we do
                # create rule states we need also to defer associating the
                # computer to the tags, since rule creation with tags is not
                # supported at the moment. See the create_rules_for_computer()
                # method of c.l.model.alert.policy.AlertRulePolicy.
                deferred_tags[secure_id] = tags
            else:
                factory.set_tags(secure_id, tags)

            factory.set_attributes(
                secure_id,
                hostname=info.get("hostname"),
                vm_info=info.get("vm-info"),
                container_info=info.get("container-info"),
                current_api=SERVER_API,
                current_version=VERSION,
            )

            factory.set_last_ping_time(
                secure_id, datetime.utcnow() - timedelta(minutes=1 + self.count)
            )

            factory.set_distribution(secure_id, info.get("distribution"))
            factory.set_wedged(secure_id, info.get("wedged", False))
            factory.set_reboot_required(secure_id, info.get("reboot-required", ()))

            self.count += 1

        factory.flush()

        for secure_id, tags in deferred_tags.items():
            computer = get_computer_by_secure_id(secure_id)
            computer.create_rule_states()
            for tag in get_tags(tags):
                computer.add_tag(tag)

        if materialize:
            return list(map(get_computer_by_secure_id, secure_ids))

    def clear(self):
        """Clear the profile, useful dynamic re-use."""
        self.computers.clear()

    def _get_computer_type(self, info):
        """Return the computer type based on specified computer info.

        Note: this implements the same logic as
          c.l.m.main.computer.get_computer_type, but returns integer IDs since
          we're not using the ORM here.
        """
        if info.get("container-info"):
            return 3
        if info.get("vm-info"):
            return 2
        return 1


class ComputerFactory(object):
    """Helper to create computers and related objects."""

    def __init__(self, resolver):
        """
        @param resolver: The RelationResolve instance to use to manage
            relations.
        """
        self._resolver = resolver
        self._account = self._resolver.get(Account)
        _, self._child_access_context = self._resolver.get(AccessContext)
        self._distribution = Relation(
            DISTRIBUTION_COLUMNS, index=Distribution.code_name
        )
        self._tag = Relation(TAG_COLUMNS, index=Tag.name)
        self._computer = Relation(COMPUTER_COLUMNS, index=Computer.secure_id)
        self._account_entity_data = Relation(
            ACCOUNT_ENTITY_DATA_COLUMNS, index=AccountEntityData.name
        )
        self._account_computer_twin = Relation(ACCOUNT_COMPUTER_TWIN_COLUMNS)
        self._computer_twin = Relation(COMPUTER_TWIN_COLUMNS)
        self._computer_packages = Relation(COMPUTER_PACKAGES_COLUMNS)
        self._ping_time = Relation(PING_TIME_COLUMNS)
        self._computer_tag = Relation(COMPUTER_TAG_COLUMNS)
        self._computer_status = Relation(COMPUTER_STATUS_COLUMNS, index=self._computer)
        self._computer_utilisation = Relation(
            COMPUTER_UTILISATION_COLUMNS, index=self._computer
        )
        self._reboot_required_info = Relation(REBOOT_REQUIRED_INFO_COLUMNS)

    def create_computer(
        self,
        account,
        secure_id,
        title,
        access_group=None,
        managed_hours=0,
        computer_type=None,
    ):
        """Create a computer with the given details."""
        if access_group is None or access_group == "global":
            # Use the root access context
            access_context_id_link = RelationLink(
                self._account, key=account, column=Account.access_context_id
            )
        else:
            access_context_id_link = RelationLink(
                self._child_access_context, key=(account, access_group)
            )

        account_id_link = RelationLink(self._account, key=account)
        self._computer.add(
            account_id=account_id_link,
            secure_id=secure_id,
            title=title,
            resource_database_key=Account.default_resource_database_key,
            access_context_id=access_context_id_link,
        )

        computer_id_link = RelationLink(self._computer, key=secure_id)

        self._computer_status.add(
            computer_id=computer_id_link,
            consecutive_errors=0,
            reboot_required_flag=False,
        )

        self._computer_utilisation.add(
            account_id=account_id_link,
            computer_id=computer_id_link,
            computer_type=computer_type,
            managed_hours=managed_hours,
            next_managed_hour_time=datetime.utcnow() + timedelta(hours=1),
        )

        self._account_entity_data.add(
            account_id=account_id_link,
            entity_type=get_account_entity_type(AccountComputerTwin),
            name=str(secure_id),
            access_context_id=access_context_id_link,
        )

        self._account_computer_twin.add(
            id=computer_id_link,
            data_id=RelationLink(self._account_entity_data, key=secure_id),
        )

        self._computer_twin.add(id=computer_id_link, account_id=account_id_link)

        self._computer_packages.add(computer_id=computer_id_link)

    def set_distribution(self, computer, code_name):
        """Associate the computer with the distribution with the given name."""
        if not code_name:
            return
        if not self._distribution.get(code_name):
            description, distributor, release = DISTRIBUTIONS[code_name]
            self._distribution.add(
                code_name=code_name,
                description=description,
                distributor=distributor,
                release=release,
            )

        distribution_id_link = RelationLink(self._distribution, key=code_name)
        self._computer.set(computer, Computer.distribution_id, distribution_id_link)

    def set_last_ping_time(self, computer, last_ping_time):
        """Set the last ping time of the given computer.

        @param computer: The secure_id of the computer to modify.
        @param last_ping_time: The last ping time value to set.
        """
        self._ping_time.add(
            computer_id=RelationLink(self._computer, key=computer),
            last_ping_time=last_ping_time,
        )

    def set_tags(self, computer, tags):
        """Associate the computer with the given tags."""
        computer_id_link = RelationLink(self._computer, key=computer)
        for tag in tags:
            if not self._tag.get(tag):
                self._tag.add(name=tag)
            self._computer_tag.add(
                computer_id=computer_id_link, tag_id=RelationLink(self._tag, key=tag)
            )

    def set_wedged(self, computer, wedged):
        """Set the computer as wedged if 'wedged' is True."""
        if wedged:
            # Set at least 2 consecutive errors
            self.set_status(computer, consecutive_errors=random.choice(range(2, 100)))

    def set_reboot_required(self, computer, packages):
        """If packages is a non-empty list, set the reboot required info."""
        if not packages:
            return
        self.set_status(computer, reboot_required_flag=True)
        self._reboot_required_info.add(
            computer_id=RelationLink(self._computer, key=computer), packages=packages
        )

    def flush(self):
        """Flush all the data created with this factory to the database."""
        main_store = get_main_store()
        # At the moment we support only one account and resource database
        account_store = get_account_store_by_name("account-1")
        _, resource_store = list(get_resource_stores())[0]

        self._resolver.add(self._distribution, main_store)
        self._resolver.add(self._tag, main_store)
        self._resolver.add(self._computer, main_store)
        self._resolver.add(self._account_entity_data, account_store)
        self._resolver.add(self._account_computer_twin, account_store)
        self._resolver.add(self._computer_twin, resource_store)
        self._resolver.add(self._computer_packages, resource_store)
        self._resolver.add(self._ping_time, main_store)
        self._resolver.add(self._computer_tag, main_store)
        self._resolver.add(self._computer_status, main_store)
        self._resolver.add(self._computer_utilisation, main_store)
        self._resolver.add(self._reboot_required_info, resource_store)

        self._resolver.flush()

    def set_status(self, computer, **values):
        """Set a value in the ComputerStatus linked to the given computer."""
        for name, value in values.items():
            column = getattr(ComputerStatus, name)
            self._computer_status.set(computer, column, value)

    def set_attributes(self, computer, **values):
        """Set a attributes in the row associated with the given computer."""
        for name, value in values.items():
            column = getattr(Computer, name)
            self._computer.set(computer, column, value)
