# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.expr import Select

from zope.interface import implementer
from zope.component import adapts, getUtility

from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import Computer, ComputerLicense
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.activity.group import (
    DELETE_CHILD_SUMMARY,
    START_CHILD_SUMMARY,
    STOP_CHILD_SUMMARY,
    RESTART_SUMMARY,
    SHUTDOWN_SUMMARY,
)
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import (
    update_parent_activities_after_changed_children,
)
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.main.interfaces import IPerson, IAccessDelegateAware
from canonical.landscape.model.entity.data import AccountEntityData
from canonical.landscape.model.selection.interfaces import IComputerSelection
from canonical.landscape.model.computer.interfaces import IComputerManagement
from canonical.landscape.model.alert.rule import RuleState
from canonical.landscape.model.package.search import meets_external_threshold
from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.eventlog.management import EventLog


__all__ = ["ComputerManagement"]


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IComputerManagement, IAccessDelegateAware)
class ComputerManagement(ModelManagement):
    """A management API for L{Computer}s in a L{ComputerSelection}."""

    adapts(IComputerSelection, IPerson)

    requires(PUBLIC, PRIVATE, ["selection", "person", "account"])

    def __init__(self, selection, person):
        self.selection = selection
        self.person = person
        super(ComputerManagement, self).__init__(selection.account, person)

    @property
    def access_delegate(self):
        return self.selection

    @requires("ViewComputer")
    def get_computer_ids(self):
        """Return IDs for computers the management acts on."""
        return self.selection.get_selected_computer_ids()

    @requires("RestartComputer")
    def restart(self, deliver_after=None):
        """Restart selected L{Computer}s.

        @param deliver_after: Optionally, a C{datetime} representing the
            earliest time the L{Activity}s created to carry out this request
            should be delivered to selected L{Computer}s.  Defaults to
            immediate delivery.
        """
        return self._reboot("restart", RESTART_SUMMARY, deliver_after=deliver_after)

    @requires("ShutdownComputer")
    def shutdown(self, deliver_after=None):
        """Shutdown selected L{Computer}s.

        @param deliver_after: Optionally, a C{datetime} representing the
            earliest time the L{Activity}s created to carry out this request
            should be delivered to selected L{Computer}s.  Defaults to
            immediate delivery.
        """
        return self._reboot("shutdown", SHUTDOWN_SUMMARY, deliver_after=deliver_after)

    def _reboot(self, reboot_kind, summary, deliver_after=None):
        parent = self.selection.create_activity_group(
            summary=summary, creator=self.person
        )
        for computer in self.selection.get_selected():
            reboot = getattr(computer, reboot_kind)
            reboot(parent, deliver_after=deliver_after)
        parent.info.create_log_entry(
            prefix="Group activity created", person=self.person
        )
        parent.settle()
        return parent

    def remove_computers(self, create_log_entries=True):
        """
        Remove computers with specified ids from main and account stores.
        It also removes all related entities, such as Activities and
        RuleStates.

        @param create_log_entries: Boolean. When False, no event log entries
           are created for the computers removed.
        """
        computer_ids = set(self.selection.get_selected_computer_ids())
        if not computer_ids:
            return

        is_over_threshold = meets_external_threshold(self.account)

        account_store = self.account.store

        # Remove activities for computers
        self._remove_computer_activities(computer_ids)

        # Remove from the account store
        self._remove_computer_account_entity_data(computer_ids)

        if create_log_entries:
            # Create event log entries about the computer removal.
            event_log = EventLog(self.account, self.person)
            event_log.create_log_entries_for_computers(
                store=account_store,
                computer_ids=computer_ids,
                message="Computer removed.",
            )
        self._remove_computers_from_main_store(computer_ids)

        account_store.find(
            RuleState, RuleState.computer_id.is_in(computer_ids)
        ).remove()

        # If, by removing these computers, we cross below the external
        # threshold, then tell the package search service we no longer want it
        # to cache data about this account.
        if is_over_threshold and not (meets_external_threshold(self.account)):
            search_client = getUtility(IPackageSearchClient)
            search_client.remove_account(account_id=self.account.id)

    def _remove_computer_activities(self, computer_ids):
        """Remove activities related to the specified computers."""
        collection = ActivityChildCollection(
            self.account.store,
            account=self.account,
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
        )
        result = collection.find()

        activity_ids = set()
        parent_ids = set()
        pairs = result.values(Activity.id, Activity.parent_id)
        for activity_id, parent_id in pairs:
            activity_ids.add(activity_id)
            parent_ids.add(parent_id)

        account_store = self.account.store
        account_store.find(Activity, Activity.id.is_in(activity_ids)).remove()

        update_parent_activities_after_changed_children(self.account, parent_ids)

    def _remove_computer_account_entity_data(self, computer_ids):
        """
        Remove AccountEntityData (cascading related AccountComputerTwins).
        """
        select = Select(
            AccountComputerTwin.data_id,
            where=[AccountComputerTwin.id.is_in(computer_ids)],
        )
        self.account.store.find(
            AccountEntityData, AccountEntityData.id.is_in(select)
        ).remove()

    def _remove_computers_from_main_store(self, computer_ids):
        """Remove computers from the main store."""
        main_store = get_main_store()
        main_store.find(Computer, Computer.id.is_in(computer_ids)).remove()
        main_store.find(
            ComputerLicense, ComputerLicense.computer_id.is_in(computer_ids)
        ).remove()

    def delete_child_computers(self, person, computer_ids):
        activity_parent = self.selection.create_activity_group(
            summary=DELETE_CHILD_SUMMARY, creator=person
        )

        for computer in self.selection.get_selected():
            if computer.id in computer_ids:
                computer.delete_instance(activity_parent)

        activity_parent.settle()
        activity_parent.info.create_log_entry("Group activity created", person=person)
        return activity_parent

    def start_child_computers(self, person, computer_ids):
        activity_parent = self.selection.create_activity_group(
            summary=START_CHILD_SUMMARY, creator=person
        )

        for computer in self.selection.get_selected():
            if computer.id in computer_ids:
                computer.start_instance(activity_parent)

        activity_parent.settle()
        activity_parent.info.create_log_entry("Group activity created", person=person)
        return activity_parent

    def stop_child_computers(self, person, computer_ids):
        activity_parent = self.selection.create_activity_group(
            summary=STOP_CHILD_SUMMARY, creator=person
        )

        for computer in self.selection.get_selected():
            if computer.id in computer_ids:
                computer.stop_instance(activity_parent)

        activity_parent.settle()
        activity_parent.info.create_log_entry("Group activity created", person=person)
        return activity_parent
