# Copyright 2005-2022 Canonical Limited. All rights reserved.

from typing import Iterable

from zope.interface import implementer
from zope.security.interfaces import Unauthorized

from storm.locals import Bool, Int, AutoReload, Unicode, Reference
from storm.expr import SQL

from canonical.lib.text import slug_from_title, describe_dict
from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.computer import ComputerStatus
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.account.helper import ComputerGroupHelper
from canonical.landscape.model.main.interfaces import IAccessContextComposite
from canonical.landscape.model.computer.interfaces import IRemovalProfile
from canonical.landscape.model.computer.jobs import create_job_for_computers_removal
from canonical.landscape.model.computer.twin import AccountComputerTwin


__all__ = [
    "RemovalProfile",
    "RemovalProfileManagement",
    "UnknownRemovalProfileError",
    "DaysWithoutExchangeValueError",
]


__clearlist__ = ["RemovalProfile", "RemovalProfileManagement"]


__requires_prefix__ = "landscape."


class UnknownRemovalProfileError(Exception):
    """The removal profile couldn't be found."""


class DaysWithoutExchangeValueError(ValueError):
    """The days_without_exchange value was out of range."""


@requires_advisor
@implementer(IRemovalProfile, IAccessContextComposite)
class RemovalProfile(ComputerGroupHelper):
    """
    A removal profile that holds criteria about how to automatically remove
    computers.

    @ivar name: A name for this profile (must be a valid slug)
    @ivar days_without_exchange: The number of days a computer hasn't exchanged
        messages with the server required for this profile to kick in and
        delete the computer.
    @ivar cascade_to_children: If true, removed computers will also include
        child computers (e.g. virtual machines, WSL instances).
    """

    __storm_table__ = "removal_profile"

    requires(
        "ViewRemovalProfile",
        PRIVATE,
        [
            "id",
            "account",
            "all_computers",
            "creation_time",
            "modification_time",
            "access_context",
            "get_computers",
            "get_computer_count",
            "get_computer_ids",
            "get_tags",
            "get_state",
            "name",
        ],
    )
    requires(
        "ViewRemovalProfile",
        "EditRemovalProfile",
        ["title", "days_without_exchange", "cascade_to_children", "computer_group_id"],
    )
    requires(
        "EditRemovalProfile",
        PRIVATE,
        [
            "update_association",
            "remove",
            "update_tags",
            "enable_for_all_computers",
            "disable_for_all_computers",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    title = Unicode(allow_none=False)
    days_without_exchange = Int(allow_none=False)
    cascade_to_children = Bool(allow_none=False, default=False)
    access_context_id = Int(allow_none=False)

    computer_group_id = Int()
    computer_group = Reference(computer_group_id, ComputerGroup.id)

    def __init__(
        self,
        account,
        name,
        title,
        days_without_exchange,
        cascade_to_children=False,
        access_context_id=None,
    ):
        super(RemovalProfile, self).__init__(account, access_context_id)
        self._create_computer_group()
        self.name = name
        self.title = title
        self.days_without_exchange = days_without_exchange
        self.cascade_to_children = cascade_to_children

    def get_computer_ids(self, computer_ids=None) -> Iterable[int]:
        """Returns a list of L{Computer.id}s this profile is associated with."""
        result = list(super().get_computer_ids(computer_ids))

        if self.cascade_to_children:
            result += list(self.get_child_computer_ids(result))

        return result

    def get_state(self):
        """@return: a dict with general state of the removal profile."""
        all_computers = self.computer_group.all_computers
        tags = sorted(tag.name for tag in self.get_tags())
        return {
            "id": self.id,
            "name": self.name,
            "title": self.title,
            "days_without_exchange": self.days_without_exchange,
            "cascade_to_children": self.cascade_to_children,
            "tags": tags,
            "all_computers": all_computers,
            "access_group": self.access_context.name,
        }

    def run(self):
        """
        Remove computers that match the criteria in this removal profile,

        @return: A list of IDs of removed computers.
        """

        store = get_main_store()

        threshold_expr = SQL(
            "now() at time zone 'utc' - interval '%d days'" % self.days_without_exchange
        )
        result = store.find(
            ComputerStatus.computer_id,
            ComputerStatus.computer_id.is_in(
                self.computer_group.computers.values(AccountComputerTwin.id)
            ),
            ComputerStatus.last_exchange_time < threshold_expr,
        )
        computer_ids = list(result)

        if self.cascade_to_children:
            computer_ids += list(self.get_child_computer_ids(computer_ids))

        create_job_for_computers_removal(self.account, computer_ids)

        return computer_ids


@requires_advisor
class RemovalProfileManagement(ModelManagement):

    requires("ViewAccount", PRIVATE, ["account", "person"])

    @requires("ViewAccount")
    def create_removal_profile(
        self,
        title,
        days_without_exchange,
        cascade_to_children=False,
        access_context_name=(DEFAULT_ACCESS_CONTEXT_NAME),
    ):
        """
        Create a L{RemovalProfile}.

        @param title: The title of the removal profile.
        @param days_without_exchange: The number of days the computers haven't
            exchanged messages.
        @param cascade_to_children: If true, removed computers will also include
            child computers (e.g. virtual machines, WSL instances).
        @param access_context_name: The access context of this removal profile.
        """
        # Derive from title.
        def check_existing(slug):
            """
            Retrieve all the removal profile names so it can be used to
            derive a name from the title.
            """
            return self.account.store.find(
                RemovalProfile.name, RemovalProfile.account_id == self.account.id
            )

        name = slug_from_title(title, 50, check_existing)
        if days_without_exchange < 1:
            raise DaysWithoutExchangeValueError(
                "days_without_exchange must be at least 1."
            )

        access_context = self.get_access_context(
            access_context_name, "EditRemovalProfile"
        )

        profile = RemovalProfile(
            self.account,
            name,
            title,
            days_without_exchange,
            cascade_to_children,
            access_context_id=access_context.id,
        )
        prefix = "Removal profile created with"
        values = {
            "name": name,
            "title": title,
            "days without exchange": days_without_exchange,
            "cascade to children": cascade_to_children,
            "access group": access_context.name,
        }
        message = "%s." % describe_dict(prefix, values, pluralize=False)
        self._event_log.create_log_entry(profile, message)

        return profile

    @requires("ViewAccount")
    def remove_removal_profiles(self, profiles):
        """Remove given L{RemovalProfile}s from the account.

        Only profiles that the accessing person has EditRemovalProfile on will
        be removed.

        @return: A C{list} of L{RemovalProfile}s that were not removed.
        """
        access_context_ids = self.get_access_context_ids("EditRemovalProfile")

        remaining_profiles = []
        for profile in profiles:
            if profile.access_context.id in access_context_ids:
                profile.remove()
                self._event_log.create_log_entry(profile, "Removal profile removed.")
            else:
                remaining_profiles.append(profile)

        return remaining_profiles

    @requires("ViewAccount")
    def get_removal_profiles(
        self, ids=None, names=None, titles=None, days_without_exchange=None
    ):
        """
        Returns a list of L{RemovalProfile}s for the current account matching
        the provided parameters.
        """
        where = [RemovalProfile.account_id == self.account.id]

        # XXX don't enforce permission check for now, if the person is not
        # provided to the class (currently the case for RemovalProfileCriteria.
        # This should be required once the person is mandatory for Criteria.
        if self.person is not None:
            access_context_ids = self.get_access_context_ids("ViewRemovalProfile")
            where.extend(
                [
                    RemovalProfile.computer_group_id == ComputerGroup.id,
                    RemovalProfile.access_context_id.is_in(access_context_ids),
                ]
            )
        if ids:
            where.append(RemovalProfile.id.is_in(ids))
        if names:
            where.append(RemovalProfile.name.is_in(names))
        if titles:
            where.append(RemovalProfile.title.is_in(titles))
        if days_without_exchange is not None:
            where.append(RemovalProfile.days_without_exchange == days_without_exchange)
        result = self.account.store.find(RemovalProfile, *where)
        return result.order_by(RemovalProfile.name)

    def get_removal_profile_by_name(self, name):
        """Get a single L{RemovalProfile} by name.

        @raises: UnknownRemovalProfileError if no profile with the given name
            exists in the account.
        """
        profile = self.get_removal_profiles(names=[name]).one()
        if profile is None:
            raise UnknownRemovalProfileError(
                "No removal profile found with name '%s'." % name
            )
        return profile

    @requires("ViewAccount")
    def edit_removal_profile(
        self,
        name,
        title=None,
        days_without_exchange=None,
        cascade_to_children=None,
        tags=None,
    ):
        """
        Edit a C{RemovalProfile} object.

        @param name: The name of the removal profile.
        @param title: The new title of the removal profile.
        @param days_without_exchange: The new number of days without
            exchange for the removal profile.
        @param cascade_to_children: If true, removed computers will also include
            child computers (e.g. virtual machines, WSL instances).
        @param tags: A list of tags to associate with the profile.
        @return The modified profile.
        """
        access_context_ids = self.get_access_context_ids("EditRemovalProfile")
        profile = self.get_removal_profile_by_name(name)
        if profile.access_context.id not in access_context_ids:
            raise Unauthorized(
                "Permission denied. Can't change removal profile '%s'" % name
            )

        profile.update_association(tags, all_computers=None)

        messages = []
        if title is not None and profile.title != title:
            messages.append("Title changed from '%s' to '%s'." % (profile.title, title))
            profile.title = title

        if days_without_exchange is not None:
            if days_without_exchange < 1:
                raise DaysWithoutExchangeValueError(
                    "days_without_exchange must be at least 1."
                )
            if profile.days_without_exchange != days_without_exchange:
                messages.append(
                    "days_without_exchange changed from '%d' to '%d'."
                    % (profile.days_without_exchange, days_without_exchange)
                )
                profile.days_without_exchange = days_without_exchange

        if cascade_to_children and profile.cascade_to_children != cascade_to_children:
            messages.append(
                "cascade_to_children changed from '%s' to '%s'"
                % profile.cascade_to_children,
                cascade_to_children,
            )
            profile.cascade_to_children = cascade_to_children

        if messages:
            self._event_log.create_log_entry(profile, " ".join(messages))
        return profile
