# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Search logic to find L{Computer}s."""

import re
from datetime import datetime, timedelta
from decimal import Decimal

from storm.expr import Or, Like, And, Eq, Column, LeftJoin
from storm.info import ClassAlias

from pyparsing import (
    Word,
    Group,
    alphanums,
    quotedString,
    removeQuotes,
    OneOrMore,
    Literal,
    nestedExpr,
)

from canonical.search.engine import (
    SearchEngine,
    SearchParser,
    SearchCriteriaError,
    CriteriaBase,
)

from canonical.landscape.model.textsearch import (
    get_search_query,
    Match,
    SIMPLE_SEARCH_CONFIG,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.alert.rule import RuleState
from canonical.landscape.model.resource.annotation import Annotation
from canonical.landscape.model.resource.network import AccountNetworkInfo
from canonical.landscape.model.resource.hardware import HardwareInfo, HardwareInfoKey
from canonical.landscape.model.resource.cloud_instance_metadata import (
    CloudInstanceMetadata,
)
from canonical.landscape.model.main.errors import UnknownAccessContextsError
from canonical.landscape.model.main.computer import (
    Computer,
    ComputerStatus,
    ComputerLicense,
    PingTime,
)
from canonical.landscape.model.computer.twin import ComputerTwin
from canonical.landscape.model.configuration import (
    is_testing_environment,
    is_standalone,
)
from canonical.landscape.model.main.license import License
from canonical.landscape.model.entity.data import AccountEntityData


__all__ = [
    "ComputerSearch",
    "AccessGroupCriteria",
    "AlertCriteria",
    "get_alert_names_mapping",
    "UpgradeProfileCriteria",
    "RemovalProfileCriteria",
    "AnnotationCriteria",
    "InstanceIdCriteria",
    "InstanceTypeCriteria",
    "AMIIdCriteria",
    "LicenseIdCriteria",
    "NoLicenseCriteria",
    "LastPingCriteria",
    "get_search_term_help_docs",
]

__clearlist__ = [
    "ComputerSearch",
    "AccessGroupCriteria",
    "AlertCriteria",
    "get_alert_names_mapping",
    "UpgradeProfileCriteria",
    "RemovalProfileCriteria",
    "AnnotationCriteria",
    "InstanceIdCriteria",
    "InstanceTypeCriteria",
    "AMIIdCriteria",
    "LicenseIdCriteria",
    "NoLicenseCriteria",
    "LastPingCriteria",
    "get_search_term_help_docs",
]


# Allow users to only request last-ping within the past year
MAX_LAST_PING_HOURS = 8760

# Define the unicode keyword search outside of _get_grammar(), since
# it's expensive to create.
unicode_printables = "".join(chr(c) for c in range(65536) if not chr(c).isspace())
unicode_word = Word(unicode_printables)


class ComputerSearchParser(SearchParser):
    """Parser converts computer search query strings into criteria objects.

    @param query: A textual representation of a query.
    @param account: The account in which the search happens.
    """

    def __init__(self, query, account):
        super(ComputerSearchParser, self).__init__(query)
        self._account = account
        self._excluded_searches = set()
        if account.is_self_service:
            self._excluded_searches.add(self._get_license_id_criteria)
            self._excluded_searches.add(self._get_no_license_criteria)

    def _get_grammar(self):
        """
        Return a pyparsing grammar supporting computer search specific needs,
        like quotes and ranges.
        """

        def remove_end(string, location, tokens):
            return tokens[0]

        def remove_to(string, location, tokens):
            return [tokens[0], tokens[2]]

        quoted_entry = quotedString.copy().setParseAction(removeQuotes)
        # A search value is a unicode_word or a quoted string.
        entry_value = unicode_word ^ quoted_entry
        # A search term is an alphanumeric, with . and - allowed, ending by :.
        key_value = Word(alphanums + ".-") + Literal(":")
        key_value.setParseAction(remove_end)
        # A range value is defined by 2 alphanumerics separated by TO.
        range_value = Word(alphanums + ".*") + Literal("TO") + Word(alphanums + ".*")
        range_value.setParseAction(remove_to)
        # And a range must be enclosed in brackets.
        bracket_values = nestedExpr("[", "]", range_value)
        # A search entry is a term and a value, a term and a range, OR, NOT, or
        # a unicode keyword search.
        entry = (
            Group(key_value + entry_value)
            ^ Group(key_value + bracket_values)
            ^ Literal("OR")
            ^ Literal("NOT")
            ^ unicode_word
        )
        return OneOrMore(entry)

    def _get_criteria(self, term):
        """Get a criteria object for the current term.

        @raises SearchCriteriaError: Raised if the term doesn't represent a
            valid search criteria.
        @return: A criteria object that can be used to generate where clauses
            suitable for finding matching L{Computer}s.
        """
        if isinstance(term, str):
            selector = "keyword"
            data = term
        elif term[0] == "needs" and term[1] == "reboot":
            selector = "reboot"
            data = ""
        elif term[0] == "needs" and term[1] == "license":
            selector = "no_license"
            data = ""
        else:
            selector, data = term

        selector = selector.replace("-", "_")
        method = getattr(self, "_get_%s_criteria" % (selector,), None)
        if method is None or method in self._excluded_searches:
            return HardwareCriteria(selector, data, self._account)
        elif not isinstance(data, str):
            raise SearchCriteriaError("Ranges not accepted for '%s' search." % selector)
        model = method(data)
        return model

    def _get_keyword_criteria(self, keyword):
        """Get a L{KeywordCriteria} instance for C{keyword}."""
        return KeywordCriteria(keyword)

    def _get_tag_criteria(self, tag_name):
        """Get a L{TagCriteria} for C{tag_name}."""
        from canonical.landscape.model.main.tag import get_tag

        tag = get_tag(tag_name)
        if tag is None:
            raise SearchCriteriaError("'%s' is an unknown tag." % tag_name)
        return TagCriteria(tag.id, self._account)

    def _get_distribution_criteria(self, distribution_name):
        """Get a L{DistributionCriteria} for C{distribution}."""
        return DistributionCriteria(distribution_name)

    def _get_title_criteria(self, title):
        """Get a L{TitleCriteria} for C{title}."""
        return TitleCriteria(title)

    def _get_hostname_criteria(self, hostname):
        """Get a L{HostnameCriteria} for C{hostname}."""
        return HostnameCriteria(hostname)

    def _get_reboot_criteria(self, ignored):
        """Get a L{RebootCriteria} object."""
        return RebootCriteria()

    def _get_alert_criteria(self, criteria_name):
        """Get an L{AlertCriteria} for C{criteria_name}."""
        return AlertCriteria(criteria_name, self._account)

    def _get_ip_criteria(self, ip_address):
        """Get an L{IPAddressCriteria} for C{ip_address}."""
        return IPAddressCriteria(ip_address, self._account)

    def _get_mac_criteria(self, mac_address):
        """Get an L{MacAddressCriteria} for C{mac_address}."""
        return MacAddressCriteria(str(mac_address), self._account)

    def _get_annotation_criteria(self, key_value_term):
        """Get an L{AnnotationCriteria} for C{key_value_term}."""
        return AnnotationCriteria(str(key_value_term), self._account)

    def _get_id_criteria(self, computer_id):
        """Get an L{IDCriteria} for C{computer_id}."""
        try:
            computer_id = int(computer_id)
        except ValueError:
            raise SearchCriteriaError("'%s' is an invalid id" % computer_id)
        return IDCriteria(computer_id)

    def _get_access_group_criteria(self, access_group_name):
        """Get an L{AccessGroupCriteria} for C{access_group_name}."""
        return AccessGroupCriteria(access_group_name, self._account)

    def _get_search_criteria(self, search_name):
        """Get a L{SearchSearchCritera} for a C{search_name}."""
        from canonical.landscape.model.account.search import SavedSearchCriteria

        return SavedSearchCriteria(search_name, self._account)

    def _get_upgrade_profile_criteria(self, upgrade_profile_name):
        """
        Return a L{UpgradeProfileCriteria} for the given
        C{upgrade_profile_name}.
        """
        return UpgradeProfileCriteria(upgrade_profile_name, self._account)

    def _get_removal_profile_criteria(self, removal_profile_name):
        """
        Return a L{RemovalProfileCriteria} for the given
        C{removal_profile_name}.
        """
        return RemovalProfileCriteria(removal_profile_name, self._account)

    def _get_environment_criteria(self, name):
        """Get a L{JujuModelCriteria} for a C{name}."""
        return JujuModelCriteria(name, self._account)

    def _get_service_criteria(self, name):
        """Get a L{ServiceCriteria} for a C{name}."""
        return ServiceCriteria(name, self._account)

    def _get_unit_criteria(self, name):
        """Get a L{UnitCriteria} for a C{name}."""
        return UnitCriteria(name, self._account)

    def _get_instance_id_criteria(self, instance_id):
        """Get a L{InstanceIdCriteria} for a C{instance_id}."""
        return InstanceIdCriteria(instance_id, self._account)

    def _get_instance_type_criteria(self, instance_type):
        """Get a L{InstanceIdCriteria} for a C{instance_type}."""
        return InstanceTypeCriteria(instance_type, self._account)

    def _get_ami_id_criteria(self, ami_id):
        """Get a L{AMIIdCriteria} for a C{ami_id}."""
        return AMIIdCriteria(ami_id, self._account)

    def _get_license_id_criteria(self, license_id):
        """Get a LicenseIdCriteria for a license_id.

        When license_id is none, return a NoLicenseCriteria instance.
        """
        if license_id == "none":
            return NoLicenseCriteria(self._account)
        try:
            license_id = int(license_id)
        except ValueError:
            raise SearchCriteriaError("'%s' is an invalid license id" % license_id)
        return LicenseIdCriteria(license_id, self._account)

    def _get_no_license_criteria(self, _):
        """Get a NoLicenseIdCriteria for an account."""
        return NoLicenseCriteria(self._account)

    def _get_last_ping_criteria(self, number_of_hours):
        "Get a LastPingCriteria for an account"
        try:
            number_of_hours = int(number_of_hours)
        except ValueError:
            raise SearchCriteriaError("'%s' is an invalid integer" % number_of_hours)
        if number_of_hours < 0 or number_of_hours > MAX_LAST_PING_HOURS:
            raise SearchCriteriaError(
                "{} is not a value between 0 and {}".format(
                    number_of_hours, MAX_LAST_PING_HOURS
                )
            )
        return LastPingCriteria(number_of_hours, self._account)


class KeywordCriteria(object):
    """
    Search criteria matches L{Computer}s with L{Computer.title} or
    L{Computer.hostname} fields that match a specified keyword.
    """

    search_help = {
        "terms": ["<keyword>"],
        "descriptions": ["Computers with the title or hostname <keyword>."],
    }

    def __init__(self, keyword):
        self._keyword = keyword

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s matching the specified
        keyword.

        @return: A Storm expression that will match the specified keywords.
        """
        keyword = "%%%s%%" % self._keyword
        # Like() strings get escaped when processing avoiding injection attacks
        title = Like(Computer.title, keyword, case_sensitive=False)
        hostname = Like(Computer.hostname, keyword, case_sensitive=False)
        return Or(title, hostname)

    def to_string(self):
        """Convert this criteria to a string."""
        return "%s" % self._keyword


class AnnotationCriteria(object):

    search_help = {
        "terms": ["annotation:<key>", "annotation:<key>:<string>"],
        "descriptions": [
            "Computers with a custom annotation defining a "
            "specific <key>. <key> can not contain ':'.",
            "Computers with a custom annotation defining a <key> "
            "which has a value containing a specific <string>.",
        ],
    }

    def __init__(self, annotation_term, account):
        self.account = account
        search_terms = annotation_term.split(":", 1)
        self._key = str(search_terms[0])
        self._substring = None
        if len(search_terms) > 1:
            self._substring = str(search_terms[1])

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s having the specified
        matching annotation key or annotation key with a value containing the
        requested string.

        @return: A Storm expression that will match the specified keywords.
        """
        result = set()
        for store in self.account.get_resource_stores():
            where = [
                ComputerTwin.account_id == self.account.id,
                ComputerTwin.id == Annotation.computer_id,
                Annotation.key == self._key,
            ]
            if self._substring:
                # Like() strings get escaped when processing avoiding
                # injection attacks
                substring = "%%%s%%" % self._substring
                where.append(Like(Annotation.value, substring, case_sensitive=True))
            data = store.find(ComputerTwin.id, *where)
            data.config(distinct=True)
            result.update(data)
        return Computer.id.is_in(result)

    def to_string(self):
        """Convert this criteria to a string."""
        annotation_string = "annotation:%s" % self._key
        if self._substring:
            annotation_string += ":%s" % self._substring
        return annotation_string


class CloudInstanceCriteria(object):

    cloud_metadata_attribute = None

    def __init__(self, cloud_metadata, account):
        self._cloud_metadata = cloud_metadata
        self.account = account

    def get_where_clause(self):
        """
        Creates a where clause to find L{Computer}s matching the specified
        C{cloud_metadata}.
        """
        result = set()
        for store in self.account.get_resource_stores():
            where = [
                ComputerTwin.account_id == self.account.id,
                ComputerTwin.id == CloudInstanceMetadata.computer_id,
                Like(
                    self.cloud_metadata_db_column,
                    self._cloud_metadata,
                    case_sensitive=False,
                ),
            ]
            data = store.find(ComputerTwin.id, *where)
            data.config(distinct=True)
            result.update(data)
        return Computer.id.is_in(result)

    def to_string(self):
        """Convert this criteria to a string."""
        return self.cloud_metadata_attribute + ":%s" % self._cloud_metadata


class InstanceIdCriteria(CloudInstanceCriteria):

    cloud_metadata_attribute = "instance-id"
    cloud_metadata_db_column = CloudInstanceMetadata.instance_id

    search_help = {
        "terms": ["instance-id:<id>"],
        "descriptions": ["Cloud instances with a specific instance <id>."],
    }


class InstanceTypeCriteria(CloudInstanceCriteria):

    cloud_metadata_attribute = "instance-type"
    cloud_metadata_db_column = CloudInstanceMetadata.instance_type

    search_help = {
        "terms": ["instance-type:<id>"],
        "descriptions": ["Cloud instances with a specific instance type <id>."],
    }


class AMIIdCriteria(CloudInstanceCriteria):

    cloud_metadata_attribute = "ami-id"
    cloud_metadata_db_column = CloudInstanceMetadata.ami_id

    search_help = {
        "terms": ["ami-id:<id>"],
        "descriptions": ["Cloud instances with a specific AMI <id>."],
    }


class HostnameCriteria(object):

    search_help = {
        "terms": ["hostname:<keyword>"],
        "descriptions": ["Computers with the exact hostname <keyword>."],
    }

    def __init__(self, hostname):
        self._hostname = hostname

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s having the specified
        hostname.

        @return: A Storm expression that will match the specified keywords.
        """
        return Computer.hostname == self._hostname

    def to_string(self):
        """Convert this criteria to a string."""
        return "hostname:%s" % self._hostname


class TitleCriteria(object):

    search_help = {
        "terms": ["title:<keyword>"],
        "descriptions": ["Computers with the exact title <keyword>."],
    }

    def __init__(self, title):
        self._title = title

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s having the specified
        title.

        @return: A Storm expression that will match the specified keywords.
        """
        return Computer.title == self._title

    def to_string(self):
        """Convert this criteria to a string."""
        return "title:%s" % self._title


class DistributionCriteria(object):
    """
    Search criteria matches L{Computer}s running particular L{Distribution}s.
    @name is the search string used to get the distribution ids.
    """

    search_help = {
        "terms": ["distribution:<keyword>"],
        "descriptions": [
            "Computers with the installed distribution named " "<keyword>."
        ],
    }

    def __init__(self, name):
        self._name = name
        self.distribution_ids = self._get_distribution_ids()
        if len(self.distribution_ids) == 0:
            raise SearchCriteriaError("'%s' is an unknown distribution." % self._name)

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s matching the specified
        L{Distribution}.

        @return: A Storm expression that will match the specified
            L{Distribution}.
        """
        return Computer.distribution_id.is_in(self.distribution_ids)

    def _get_distribution_ids(self):
        """
        Get a C{list} of L{Distribution.id} with release versions or code names
        matching C{self._name}.
        """
        from canonical.landscape.model.main.distribution import Distribution

        store = get_main_store()
        result = store.find(
            Distribution.id,
            Or(
                Distribution.release == self._name, Distribution.code_name == self._name
            ),
        )
        return list(result)

    def to_string(self):
        """Convert this criteria to a string."""
        return "distribution:%s" % self._name


class TagCriteria(object):
    """
    Search criteria matches L{Computer}s associated with the specified L{Tag}.

    @param tag_id: The L{Tag.id} to match L{Computer}s against.
    @param account_id: Optionally, an L{Account.id} to use to limit search
        boundaries.  It should be provided when L{Computer}s in a specific
        account are queried to optimize the number of L{Computer.id}s that
        need to be retrieved and compared.
    """

    search_help = {
        "terms": ["tag:<tag_id>"],
        "descriptions": ["Computers associated with the tag <tag_id>."],
    }

    def __init__(self, tag_id, account):
        self._account = account
        self._tag_id = tag_id

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s matching the specified
        L{Tag}.

        @return: A Storm expression that will match the specified L{Tag}.
        """
        from canonical.landscape.model.main.tag import ComputerTag

        store = get_main_store()
        result = store.find(
            ComputerTag.computer_id,
            ComputerTag.tag_id == self._tag_id,
            ComputerTag.computer_id == Computer.id,
            Computer.account_id == self._account.id,
        )
        computer_ids = list(result)
        return Computer.id.is_in(computer_ids)

    def to_string(self):
        """Convert this criteria to a string."""
        from canonical.landscape.model.main.tag import Tag

        store = get_main_store()
        tag = store.get(Tag, self._tag_id)
        return "tag:%s" % tag.name


class RebootCriteria(object):
    """Search criteria matches L{Computer}s that need to be rebooted."""

    search_help = {
        "terms": ["needs:reboot"],
        "descriptions": ["Computers needing a reboot."],
    }

    def get_where_clause(self):
        """Get a where clause to find L{Computer}s that need to be rebooted.

        @return: A Storm expression that will match L{Computer}s that need to
            be rebooted.
        """
        return And(
            ComputerStatus.reboot_required_flag == True,
            ComputerStatus.computer_id == Computer.id,
        )

    def to_string(self):
        """Convert this criteria to a string."""
        return "needs:reboot"


class IPAddressCriteria(CriteriaBase):
    """
    Search criteria matches L{Computer}s that have a matching IP address
    fragment.
    """

    criteria_prefix = "ip"
    criteria_attribute = "_ip_address"
    search_help = {
        "terms": ["ip:<address>"],
        "descriptions": [
            "Computers with IP addresses like <address>.",
            "<address> can be a fragment or partial IP address "
            "representing a subnet such as 91.185.94 to match any "
            "computers in the subnet.",
        ],
    }

    def __init__(self, ip_address, account):
        self._ip_address = ip_address
        self._account = account

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s with a matching IP address.

        @return: A Storm expression that will match L{Computer}s with a network
            device with the IP address.
        """
        if not re.match(r"^[\d\.]+", self._ip_address):
            raise SearchCriteriaError("Invalid IP address %s" % self._ip_address)
        helper = AccountNetworkInfo(self._account)
        result = helper.find_computers_by_ip(self._ip_address)
        return Computer.id.is_in(result)


class MacAddressCriteria(CriteriaBase):
    """
    Search criteria matches L{Computer}s that have a matching MAC address.
    """

    criteria_prefix = "mac"
    criteria_attribute = "_mac_address"
    search_help = {
        "terms": ["mac:<address>"],
        "descriptions": ["Computers with the MAC address <address>."],
    }

    def __init__(self, mac_address, account):
        self._mac_address = mac_address
        self._account = account

    def get_where_clause(self):
        """
        Get a where clause to find L{Computer}s with a matching MAC address.

        @return: A Storm expression that will match L{Computer}s with a network
            device with the MAC address.
        """
        helper = AccountNetworkInfo(self._account)
        result = helper.find_computers_by_mac_address(self._mac_address.lower())
        return Computer.id.is_in(result)


def _get_alert_criteria_tuples():
    """
    Returns a list of ("alert name", <Alert Class>)  for allowed alerts.
    """
    from canonical.landscape.model.package.alerts import (
        PackageUpgradesAlert,
        SecurityUpgradesAlert,
        PackageProfilesAlert,
        PackageReporterAlert,
        EsmDisabledAlert,
    )
    from canonical.landscape.model.computer.alerts import (
        ComputerOfflineAlert,
        ComputerRebootAlert,
        ComputerDuplicateAlert,
    )
    from canonical.landscape.model.activity.alerts import UnapprovedActivitiesAlert

    tuples = [
        ("package-upgrades", PackageUpgradesAlert),
        ("security-upgrades", SecurityUpgradesAlert),
        ("package-profiles", PackageProfilesAlert),
        ("package-reporter", PackageReporterAlert),
        ("esm-disabled", EsmDisabledAlert),
        ("computer-offline", ComputerOfflineAlert),
        ("computer-reboot", ComputerRebootAlert),
        ("computer-duplicates", ComputerDuplicateAlert),
        ("unapproved-activities", UnapprovedActivitiesAlert),
    ]
    return tuples


class AlertCriteria(CriteriaBase):
    """
    Search criteria matches L{Computer}s with the specified alert active.

    @param alert_id: The L{Alert.id} to match L{Computer}s against.
    @param account: Required parameter, used to look for alerts in the account
        being active and the associated computers.
    """

    criteria_prefix = "alert"
    criteria_attribute = "_alert_type"
    search_help = {
        "terms": ["alert:<type>"],
        "descriptions": [
            "Computers with outstanding alerts of the specified " "<type>.",
            "<type> can be %s."
            % ", ".join(term for term, _ in _get_alert_criteria_tuples()),
        ],
    }

    def __init__(self, alert_type, account):
        self._account = account
        self._alert_type = alert_type

    def get_where_clause(self):
        """Get a where clause to find L{Computer}s that have an active alert
        with the specified Alert class.

        @return: A Storm expression that will match L{Computer}s that have the
            alert active.
        """
        computer_ids = set()
        alert_class = get_alert_criteria_mapping().get(self._alert_type)
        if not alert_class:
            raise SearchCriteriaError("'%s' is an unknown alert." % self._alert_type)
        rules = self._account.get_rules(info_classes=[alert_class], active=True)

        for rule in rules:
            computer_ids.update(
                rule.info.get_states(active=True).values(RuleState.computer_id)
            )
        return Computer.id.is_in(computer_ids)


def get_alert_criteria_mapping():
    """
    Returns an {"alert name": <Alert Class>} mapping for allowed alerts.
    """
    return dict(_get_alert_criteria_tuples())


def get_alert_names_mapping():
    """
    Returns an {<Alert Class>: "alert name"} mapping for allowed alerts.
    """
    return dict((cls, name) for name, cls in _get_alert_criteria_tuples())


class IDCriteria(CriteriaBase):

    criteria_prefix = "id"
    criteria_attribute = "_computer_id"
    search_help = {
        "terms": ["id:<computer_id>"],
        "descriptions": ["Computer with the id <computer_id>."],
    }

    def __init__(self, computer_id):
        self._computer_id = computer_id

    def get_where_clause(self):
        """Find the computer with the given id.

        @return: A Storm expression that will match the specified id.
        """
        return Eq(Computer.id, self._computer_id)


class AccessGroupCriteria(CriteriaBase):

    criteria_prefix = "access-group"
    criteria_attribute = "_access_context_name"
    search_help = {
        "terms": ["access-group:<name>"],
        "descriptions": ["Computers associated with the <name> access group."],
    }

    def __init__(self, access_context_name, account):
        self._access_context_name = access_context_name
        self._account = account

    def get_where_clause(self):
        """Find computers in the given access context.

        @return: A Storm expression that will match computers in the specified
        access context.
        """
        try:
            [access_context] = self._account.get_access_contexts(
                names=[self._access_context_name]
            )
        except UnknownAccessContextsError:
            raise SearchCriteriaError(
                "Can't find access group named '%s'" % self._access_context_name
            )

        return Eq(Computer.access_context_id, access_context.id)


class UpgradeProfileCriteria(CriteriaBase):

    criteria_prefix = "upgrade-profile"
    criteria_attribute = "_upgrade_profile_name"
    search_help = {
        "terms": ["upgrade-profile:<name>"],
        "descriptions": ["Computers associated with the <name> upgrade profile."],
    }

    def __init__(self, upgrade_profile_name, account):
        self._upgrade_profile_name = upgrade_profile_name
        self._account = account

    def get_where_clause(self):
        """Find computers with the given upgrade profile.

        @return: A Storm expression that will match the computers associated
            with the profile.
        """
        from canonical.landscape.model.package.autoupgrade import (
            UpgradeProfileManagement,
        )

        management = UpgradeProfileManagement(self._account)
        profile = management.get_upgrade_profiles(
            names=[self._upgrade_profile_name]
        ).one()
        if profile is None:
            raise SearchCriteriaError(
                "Can't find upgrade profile named '%s'" % self._upgrade_profile_name
            )
        return Computer.id.is_in(profile.get_computer_ids())


class RemovalProfileCriteria(CriteriaBase):

    criteria_prefix = "removal-profile"
    criteria_attribute = "_removal_profile_name"
    search_help = {
        "terms": ["removal-profile:<name>"],
        "descriptions": ["Computers associated with the <name> removal profile."],
    }

    def __init__(self, removal_profile_name, account):
        self._removal_profile_name = removal_profile_name
        self._account = account

    def get_where_clause(self):
        """Find computers with the given removal profile.

        @return: A Storm expression that will match the computers associated
            with the profile.
        """
        from canonical.landscape.model.computer.removal import RemovalProfileManagement

        management = RemovalProfileManagement(self._account)
        profile = management.get_removal_profiles(
            names=[self._removal_profile_name]
        ).one()
        if profile is None:
            raise SearchCriteriaError(
                "Can't find removal profile named '%s'" % self._removal_profile_name
            )
        return Computer.id.is_in(profile.get_computer_ids())


class HardwareCriteria(CriteriaBase):

    criteria_attribute = "value_escaped"

    aliases = {"cpu.flags": "cpu.capabilities"}

    # There are too many hardware search terms to extrapolate.
    # Just list the most helpful terms. The rest of the specific search
    # terms will be presented as mouseover info popups in the UI on the
    # computer hardware page.
    search_help = {
        "terms": [
            "vendor:<string>",
            "product:<string>",
            "uuid:<string>",
            "cpu.<attribute>:<string>",
            "disk.<attribute>:<string>",
            "display.<attribute>:<string>",
            "firmware.<attribute>:<string>",
            "memory.<attribute>:<string>",
            "network.capacity:<value_or_range>",
            "network.<attribute>:<string>",
            "network.serial:<mac_address>",
            "serial:<serialnumber>",
        ],
        "descriptions": [
            "Computers containing hardware with the matched properties.",
            "<string> is a partial string to try matching to the search "
            "terms specified.",
            "<attribute> can be: vendor, product, version or size. If size is "
            "searched a <value> or <range> must be provided as the <string>.",
            "<value> is a number followed by a kb, mb, gb, tb, ghz or mhz.",
            "<range> is a set of minimum and maximum <value>s of the form "
            "<term>.<attribute>:[<value> TO <value>].",
            "For example: disk.size:[10Gb TO 2Tb] will match computers "
            "with storage between 10 Gigabytes and 2 Terabytes.",
        ],
    }

    def __init__(self, key, value, account):
        self.key = key
        self.value = value
        self.criteria_prefix = key
        self.account = account

    @property
    def value_escaped(self):
        """
        If the search value contains a space, escape it with quotes for
        C{to_string}.
        """
        if " " in self.value:
            return '"%s"' % self.value
        return self.value

    def _parse_units(self, value, multiple, suffix):
        """Parse a value in the form of $number$unit.

        It understands power from kilo to tera, with an optional C{suffix} to
        indicate the unit.
        """
        parse_re = re.compile("^([\d]+[\.]?[\d]*)([a-zA-Z]+)$")
        match = parse_re.match(value)
        if match is None:
            raise SearchCriteriaError("Invalid unit value '%s'" % value)
        value = Decimal(match.group(1))
        unit = match.group(2).lower()
        mapping = {
            "k": 1,
            "k%s" % suffix: 1,
            "m": 2,
            "m%s" % suffix: 2,
            "g": 3,
            "g%s" % suffix: 3,
            "t": 4,
            "t%s" % suffix: 4,
        }
        factor = mapping.get(unit)
        if factor is not None:
            value *= pow(multiple, factor)
        elif unit != suffix:
            raise SearchCriteriaError("Invalid unit value '%s'" % value)
        return int(value)

    def _parse_integer_criteria(self, multiple, suffix, error):
        """Handle a criteria mapping to an integer in the database.

        Single values are treated with a fuzzing range given by C{error},
        otherwise we manage a range with '*' acting as a wildcard.
        """
        if isinstance(self.value, str):
            value = self._parse_units(self.value, multiple, suffix)
            return [
                HardwareInfo.int_value >= value - value * error,
                HardwareInfo.int_value <= value + value * error,
            ]
        else:
            result = []
            if self.value[0] != "*":
                value = self._parse_units(self.value[0], multiple, suffix)
                result.append(HardwareInfo.int_value >= value)
            if self.value[1] != "*":
                value = self._parse_units(self.value[1], multiple, suffix)
                result.append(HardwareInfo.int_value <= value)
            if not result:
                raise SearchCriteriaError("Invalid range")
            return result

    def _get_memory_size_criteria(self):
        return self._parse_integer_criteria(1024, "b", 0.1)

    def _get_disk_size_criteria(self):
        return self._parse_integer_criteria(1000, "b", 0.05)

    def _get_cpu_size_criteria(self):
        return self._parse_integer_criteria(1000, "hz", 0.05)

    def _get_network_size_criteria(self):
        return self._parse_integer_criteria(1000, "b", 0.05)

    def get_where_clause(self):
        result = set()
        key = self.aliases.get(self.key, self.key)
        for store in self.account.get_resource_stores():
            where = [
                ComputerTwin.account_id == self.account.id,
                ComputerTwin.id == HardwareInfo.computer_id,
            ]
            if key != "hardware":
                for part in key.split("."):
                    part_alias = ClassAlias(HardwareInfoKey)
                    where.extend(
                        [
                            part_alias.hardware_info_id == HardwareInfo.id,
                            part_alias.key == part,
                        ]
                    )
            check_key = key.replace(".", "_")
            method = getattr(self, "_get_%s_criteria" % check_key, None)
            if method is not None:
                where.extend(method())
            elif not isinstance(self.value, str):
                raise SearchCriteriaError(
                    "Range queries not supported for '%s'." % self.key
                )
            else:
                value = get_search_query(self.value, search_config=SIMPLE_SEARCH_CONFIG)
                vector = Column("str_value_vector")
                where.append(Match(value, vector))
            data = store.find(ComputerTwin.id, *where)
            data.config(distinct=True)
            result.update(data)
        return Computer.id.is_in(result)


class JujuModelCriteria(CriteriaBase):

    criteria_prefix = "environment"
    criteria_attribute = "_environment"
    search_help = {
        "terms": ["environment:<uuid>"],
        "descriptions": ["Computers matching the given Juju environment UUID."],
    }

    def __init__(self, juju_model, account):
        self._juju_model = juju_model
        self._account = account

    def get_where_clause(self):
        """Find computers associated with environments with the given name.

        @return: A Storm expression that will match computers associated with
                 environments in the account having the specified name.
        """
        from canonical.landscape.model.juju.machine import Machine
        from canonical.landscape.model.juju.model import JujuModel

        computer_ids = self._account.store.find(
            Machine.computer_id,
            Machine.computer_id != None,
            Machine.juju_model_id == JujuModel.id,
            JujuModel.data_id == AccountEntityData.id,
            AccountEntityData.account_id == self._account.id,
            JujuModel.uuid == self._juju_model,
        )
        return Computer.id.is_in(computer_ids)


class LicenseIdCriteria(CriteriaBase):

    criteria_prefix = "license-id"
    criteria_attribute = "_license_id"
    search_help = {
        "terms": ["license-id:<id>"],
        "descriptions": ["Computers assigned to the license with the given <id>."],
    }

    def __init__(self, license_id, account):
        self._license_id = license_id
        self._account = account

    def get_where_clause(self):
        """Find computers associated with the given C{_license_id}.

        @return: A Storm expression that will match the computers associated
        with the license.
        """
        store = get_main_store()
        result = store.find(
            Computer.id,
            License.account_id == self._account.id,
            ComputerLicense.computer_id == Computer.id,
            ComputerLicense.license_id == License.id,
            License.id == self._license_id,
        )
        return Computer.id.is_in(result)

    def to_string(self):
        """Convert this criteria to a string."""
        return "license-id:%d" % self._license_id


class NoLicenseCriteria(object):
    """Search criteria matches Computers that need a License."""

    search_help = {
        "terms": ["needs:license", "license-id:none"],
        "descriptions": ["Computers without a Landscape license."],
    }

    def __init__(self, account):
        self.account = account

    def get_where_clause(self):
        """Find computers associated with no licenses.

        @return: A Storm expression that will match the computers associated
        with no licenses.
        """
        store = get_main_store()
        using_clause = (
            Computer,
            LeftJoin(ComputerLicense, Computer.id == ComputerLicense.computer_id),
        )
        result = store.using(*using_clause).find(
            Computer.id,
            Computer.account_id == self.account.id,
            ComputerLicense.license_id == None,
        )
        return Computer.id.is_in(result)

    def to_string(self):
        """Convert this criteria to a string."""
        return "license-id:none"


class LastPingCriteria(CriteriaBase):

    criteria_prefix = "last-ping"
    criteria_attribute = "_nr_of_hours"
    search_help = {
        "terms": ["last-ping:<nr-of-hours>"],
        "descriptions": [
            "Computers that were active in the last <nr-of-hours> hours, "
            "where <nr-of-hours> is a value between 0 and {}".format(
                MAX_LAST_PING_HOURS
            )
        ],
    }

    def __init__(self, hours_since_last_ping, account):
        self._nr_of_hours = hours_since_last_ping
        self._account = account

    def get_where_clause(self):
        """Find computers that were active in the last requested number of
        hours.

        @return: A Storm expression that will match the active computers.
        """
        store = get_main_store()
        result = store.find(
            Computer.id,
            Computer.account_id == self._account.id,
            PingTime.computer_id == Computer.id,
            PingTime.last_ping_time
            >= datetime.utcnow() - timedelta(hours=self._nr_of_hours),
        )
        return Computer.id.is_in(result)


class ComputerSearch(SearchEngine):
    """A Search engine for finding L{Computer}s.

    Retrieve computers associated with the current account.

    This method takes one mandatory argument:

     - **Query**: A query string used to filter the returned computers. It can
       be one of the following space-separated criteria:

    Words provided as search parameters are treated as keywords.  Selector
    prefixes can be used to further customize the search:

     * tag - Search for L{Computer}s with the specified tag.
     * distribution - Search for L{Computer}s running a specific Ubuntu
            release (can be code name like 'lucid' or version number like
            '10.04').
     * alert - Search for computers with a named alert being active, alerts can
            be one of "package-upgrades", "security-upgrades",
            "package-profiles", "package-reporter",  "computer-offline",
            "computer-reboot"
     * OR - Search for L{Computer}s matching term A or term B. OR must be in
            all caps.
     * NOT - Search for L{Computer}s not matching the next term. NOT must be in
            all caps

    These prefixes can be mixed and matched with keywords.  For example, the
    following query would match computers with 'appserver' in their title or
    hostname, either with the tag 'server' or running the 'hardy' release of
    Ubuntu::

      appserver tag:server OR distribution:hardy

      appserver alert:package-upgrades

    @param account: The L{Account} used to limit search boundaries.
    """

    model = Computer
    parser_class = ComputerSearchParser

    def __init__(self, account, filter_result=True):
        super(ComputerSearch, self).__init__(filter_result=filter_result)
        self.account = account

    def get_where_clause(self, search_query):
        where = super(ComputerSearch, self).get_where_clause(search_query)
        where.append(Computer.account_id == self.account.id)
        return where

    def _get_search_store(self):
        return get_main_store()

    def filter_result(self, result):
        return result.order_by(self.model.title)

    def _build_parser(self, query):
        return self.parser_class(query, self.account)


def get_search_term_help_docs(account):
    """
    Return a list of dicts with available search terms and their
    descriptions for help documentation purposes.
    """

    # XXX The list of criteria_classes must match all of the criteria used by
    # ComputerSearchParser as this will dynamically generate documentation
    # for our UI help overlays for search help. Each criteria class will need
    # a help method to return a dict of terms and descriptions.
    from canonical.landscape.model.account.search import SavedSearchCriteria

    criteria_classes = [
        KeywordCriteria,
        DistributionCriteria,
        RebootCriteria,
        AlertCriteria,
        IPAddressCriteria,
        MacAddressCriteria,
        IDCriteria,
        AccessGroupCriteria,
        SavedSearchCriteria,
        UpgradeProfileCriteria,
        RemovalProfileCriteria,
        HardwareCriteria,
        TagCriteria,
        HostnameCriteria,
        TitleCriteria,
        AnnotationCriteria,
        InstanceIdCriteria,
        InstanceTypeCriteria,
        AMIIdCriteria,
        LicenseIdCriteria,
        NoLicenseCriteria,
        LastPingCriteria,
    ]
    is_juju_supported_env = is_testing_environment() or is_standalone()
    if is_juju_supported_env:
        criteria_classes.extend([JujuModelCriteria])
    if account.is_self_service:
        criteria_classes.remove(LicenseIdCriteria)
        criteria_classes.remove(NoLicenseCriteria)
    return [cls.search_help for cls in criteria_classes]
