# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from datetime import timedelta

from storm.locals import Store

from canonical.event.subscriber import subscriber
from canonical.landscape.model.main.computer import PingServer
from canonical.landscape.model.events import (
    MaintenanceEvent,
    AddComputerEvent,
    AddTagsEvent,
    RemoveTagsEvent,
)
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.stormfunc import AgeFromNow
from canonical.landscape.model.computer.removal import RemovalProfile
from canonical.landscape.model.account.computergroup import ComputerGroup


@subscriber(IAccount, MaintenanceEvent)
def remove_old_ping_servers(account, unused_event):
    offline_delta = timedelta(days=1)
    store = Store.of(account)

    # Remove ping servers which have not been updated in the last
    # C{offline_delta} period.
    result = store.find(
        PingServer, AgeFromNow(PingServer.last_update_time) > offline_delta
    )
    result.remove()

    # Remove ping servers which have been started before
    # C{offline_delta} period but never performed a single update.
    result = store.find(
        PingServer,
        AgeFromNow(PingServer.startup_time) > offline_delta,
        PingServer.last_update_time == None,
    )
    result.remove()


@subscriber(IAccount, AddComputerEvent, phase=1)
def update_all_computers_removal_profiles(account, event):
    """
    Upon a L{AddComputerEvent} event, link profiles enabled for all computers
    with this newly created computer.
    """
    profiles = account.store.find(
        RemovalProfile,
        RemovalProfile.account_id == account.id,
        RemovalProfile.computer_group_id == ComputerGroup.id,
        ComputerGroup.all_computers == True,
    )
    for profile in profiles:
        profile.add_computers_with_ids([event.computer.id])


@subscriber(IAccount, AddTagsEvent)
def add_tag_to_computers_handler(account, event):
    """Event handler is invoked when L{Tag}s are added to L{Computer}s."""
    account.computer_group.synchronize_with_added_tags(
        event.tag_ids, event.computer_ids, RemovalProfile
    )


@subscriber(IAccount, RemoveTagsEvent)
def remove_tag_from_computers_handler(account, event):
    """Event handler is invoked when L{Tag}s are removed from L{Computer}s."""
    account.computer_group.synchronize_with_removed_tags(
        event.tag_ids, event.computer_ids, RemovalProfile
    )
