# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from storm.locals import Int

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.entity.helper import AccountEntity


__all__ = ["ComputerTwin", "AccountComputerTwin"]

__clearlist__ = ["AccountComputerTwin"]


__requires_prefix__ = "landscape."


class ComputerTwin(ComputerHelper):
    """The representation of a computer in a database other than 'main'.

    The primary purpose of this object is to serve as a point of relational
    stability with all computer-related objects in a remote database, such as
    an account or resource database.
    """

    __storm_table__ = "computer_twin"

    id = Int(primary=True, allow_none=False)
    account_id = Int(allow_none=False)

    def __init__(self, computer):
        super(ComputerTwin, self).__init__(computer)
        self.account_id = computer.account.id

    # This property makes ComputerHelper happy.  When ComputerHelper sets
    # computer_id this code causes the id of this object to be the same as the
    # Computer in the main database represented by this one.
    def _get_computer_id(self):
        return self.id

    def _set_computer_id(self, id):
        self.id = id

    computer_id = property(_get_computer_id, _set_computer_id)

    @property
    def account(self):
        store = get_main_store()
        from canonical.landscape.model.main.account import Account

        return store.get(Account, self.account_id)


class AccountComputerTwin(AccountEntity):
    """The representation of a computer in the 'account' database.

    This class serves both as a point of relational stability with all
    computer-related objects in the account database, and as normalized
    account-related model for enforcing RBAC and security with access
    contexts.
    """

    __storm_table__ = "computer_twin"

    id = Int(primary=True, allow_none=False)

    def __init__(self, computer):
        # Use the secure ID as name, since the computer title isn't guaranteed
        # to be unique.
        name = computer.secure_id.decode()
        super(AccountComputerTwin, self).__init__(
            computer.account, name, access_context=computer.access_context
        )
        self.id = computer.id

    @property
    def computer(self):
        """Return the actual Computer we are the twin of."""
        from canonical.landscape.model.main.computer import get_computer_by_id

        return get_computer_by_id(self.id)
