# Copyright 2005-2022 Canonical Limited. All rights reserved.

from collections import namedtuple
import json
import os
import subprocess
import uuid

from urllib.parse import urlparse, urlunparse

from zope.component import getUtility
from zope.interface import implementer, Interface

from landscape.lib.fs import read_file

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.environment import get_config_name
from canonical.landscape.model.main.store import get_main_store


__clearlist__ = [
    "LANDSCAPE_SYSTEM_EMAIL",
    "LANDSCAPE_ROOT_URL",
    "LANDSCAPE_PING_URL",
    "LANDSCAPE_CONFIG",
    "LANDSCAPE_HTTP_PROXY",
    "LANDSCAPE_HTTPS_PROXY",
    "LANDSCAPE_NO_PROXY",
    "LANDSCAPE_KNOWN_PROXIES",
    "LANDSCAPE_DO_NOT_TRACK_GA",
    "LANDSCAPE_ANONYMOUS_UUID",
    "get_root_url",
    "set_root_url",
    "get_custom_ssl_ca",
    "is_development_mode",
    "is_testing_environment",
    "is_staging_environment",
    "is_standalone",
    "is_selenium_environment",
    "get_archive_url",
    "get_config_name",
    "LANDSCAPE_CUSTOM_SSL_CA",
    "ISystemConfiguration",
    "SystemConfiguration",
    "get_revision",
    "get_static_root_path",
    "LANDSCAPE_HOME",
    "log_revision",
    "get_api_url",
    "get_client_config",
    "is_charm_deployment",
    "get_proxy_configuration",
    "set_proxy_configuration",
    "LANDSCAPE_CONFIG_KEYS",
    "LANDSCAPE_DEVELOPER_MODE",
    "OPENSTACK_UBUNTU_SERIES",
    "OPENSTACK_RELEASE",
    "OPENSTACK_USE_PROPOSED",
    "DEVELOPER_ONLY_KEYS",
    "JUJU_AGENT_STREAM",
    "JUJU_TEST_MODE",
    "STRIPE_PUBLIC_KEY",
    "STRIPE_SECRET_KEY",
    "MARKETO_CREDENTIALS",
    "get_marketo_credentials",
    "set_marketo_credentials",
    "get_stripe_public_key",
    "get_stripe_secret_key",
    "set_stripe_api_keys",
    "get_billing_credentials",
    "set_billing_credentials",
    "BILLING_CREDENTIALS",
    "AUTOPILOT_NO_HA",
    "set_default_system_email",
]

__all__ = __clearlist__

LANDSCAPE_SYSTEM_EMAIL = "landscape.system_email_sender"
LANDSCAPE_ROOT_URL = "landscape.root_url"
LANDSCAPE_PING_URL = "landscape.ping_url"
LANDSCAPE_ANONYMOUS_UUID = "landscape.uuid"

# Used to get an accurate IP for login notes, see #838131
LANDSCAPE_KNOWN_PROXIES = "landscape.known_proxies"

LANDSCAPE_HTTP_PROXY = "landscape.http_proxy"
LANDSCAPE_HTTPS_PROXY = "landscape.https_proxy"
LANDSCAPE_NO_PROXY = "landscape.no_proxy"
LANDSCAPE_DO_NOT_TRACK_GA = "landscape.do_not_track_ga"

LANDSCAPE_DEVELOPER_MODE = "landscape.developer_mode"

JUJU_AGENT_STREAM = "juju.agent-stream"
JUJU_TEST_MODE = "juju.test-mode"
OPENSTACK_UBUNTU_SERIES = "openstack.ubuntu_series"
OPENSTACK_RELEASE = "openstack.openstack_release"
OPENSTACK_USE_PROPOSED = "openstack.use_proposed"
AUTOPILOT_NO_HA = "autopilot.no_ha"

STRIPE_PUBLIC_KEY = "stripe.public_key"
STRIPE_SECRET_KEY = "stripe.secret_key"

BILLING_CREDENTIALS = "billing_credentials"

MARKETO_CREDENTIALS = "marketo"

LANDSCAPE_CONFIG_KEYS = (
    LANDSCAPE_SYSTEM_EMAIL,
    LANDSCAPE_ROOT_URL,
    LANDSCAPE_PING_URL,
    LANDSCAPE_ROOT_URL,
    LANDSCAPE_KNOWN_PROXIES,
    LANDSCAPE_HTTP_PROXY,
    LANDSCAPE_HTTPS_PROXY,
    LANDSCAPE_NO_PROXY,
    LANDSCAPE_DO_NOT_TRACK_GA,
    STRIPE_PUBLIC_KEY,
    STRIPE_SECRET_KEY,
    BILLING_CREDENTIALS,
    MARKETO_CREDENTIALS,
)

DEVELOPER_ONLY_KEYS = (
    OPENSTACK_UBUNTU_SERIES,
    OPENSTACK_RELEASE,
    OPENSTACK_USE_PROPOSED,
    LANDSCAPE_DEVELOPER_MODE,
    JUJU_AGENT_STREAM,
    JUJU_TEST_MODE,
    AUTOPILOT_NO_HA,
)

STRICT_VALIDATE = {JUJU_AGENT_STREAM: ("devel", "proposed", "released")}

LANDSCAPE_CONFIG = "LANDSCAPE_CONFIG"
DEBUG_ENV_VAR = "LANDSCAPE_DEBUG"
LANDSCAPE_CUSTOM_SSL_CA = "/etc/ssl/certs/landscape_server_ca.crt"

LANDSCAPE_HOME = "/opt/canonical/landscape"
LANDSCAPE_RUN = "/var/run/landscape"
LANDSCAPE_DEFAULT = "/etc/default/landscape-server"


class ISystemConfiguration(Interface):
    """
    Provides a global utility for storing changeable configuration values.
    """

    def get(key, default=None):
        """
        Returns the stored configuration value by name, or the default if not
        found.
        """

    def set(key, value):
        """
        Stores the value in the global configuration agains the key name.
        """

    def get_all():
        """
        Retrieve the full mapping of key/values as stored in the database.
        """


@implementer(ISystemConfiguration)
class SystemConfiguration(object):
    """Hold application-wide system configuration details."""

    def _has_key(self, key):
        """Return whether an entry for a particular key exists."""
        store = get_main_store()
        result = store.execute(
            "SELECT COUNT(*) FROM system_configuration " "WHERE key = ? LIMIT 1", (key,)
        ).get_one()
        return result[0] > 0

    def set(self, key, value):
        """
        Insert a system configuration value for the given key. If a value is
        already present for the given key, it will be overwritten.

        @param key: The key for which to store the passed in value.
        @param value: The value to store for the given key. Can be any type, as
            long as its serializable to JSON.
        """
        if isinstance(key, bytes):
            key = key.decode("utf-8")  # "cast" to unicode

        if key in STRICT_VALIDATE:
            if value not in STRICT_VALIDATE[key]:
                raise ValueError(
                    "'%s' must be one of %s" % (value, STRICT_VALIDATE[key])
                )

        if not isinstance(key, str):
            raise TypeError("key must be a string")

        value = json.dumps(value)

        store = get_main_store()

        # If we ever support Postgres 9.5+ only, this is the perfect use case
        # for UPSERT.
        if self._has_key(key):
            store.execute(
                "UPDATE system_configuration SET value = ?" " WHERE key = ?",
                (value, key),
            )
        else:
            store.execute(
                "INSERT INTO system_configuration (key, value)" " VALUES (?, ?)",
                (key, value),
            )

    def get(self, key, default=None):
        """
        Retrieve the value associated with the given system configuration key.

        @param key: The key to retrieve the value for.
        """
        if not isinstance(key, str):
            raise TypeError("key must be a string")

        store = get_main_store()
        # We cast to text here to prevent psycopg2 > 2.5 from doing
        # automatic JSON decoding for us. See
        # http://initd.org/psycopg/docs/extras.html#adapt-json
        result = store.execute(
            "SELECT value::text FROM system_configuration " "WHERE key = ?", (key,)
        ).get_one()
        if result:
            return json.loads(result[0])
        else:
            return default

    def get_all(self):
        """
        Return a mapping of all keys stored in the database and their
        associated value.
        """
        mapping = {}
        store = get_main_store()

        # We cast to text here to prevent psycopg2 > 2.5 from doing
        # automatic JSON decoding for us. See
        # http://initd.org/psycopg/docs/extras.html#adapt-json
        result = store.execute("SELECT key, value::text FROM system_configuration")

        for key, value in result:
            mapping[key] = json.loads(value)

        return mapping


def get_root_url():
    """
    Get the landscape server's root URL path.

    The logic is: First look at environment variables (that overrides anything
    else if set), then look at the database configuration entry.

    If both fail, look at file-based configuration.
    """
    db_config = getUtility(ISystemConfiguration)
    url = os.environ.get("LANDSCAPE_ROOT_URL", None)
    if not url:
        url = db_config.get(LANDSCAPE_ROOT_URL)
    if not url:
        file_config = getUtility(IFileBasedConfiguration)
        url = file_config.get("global", key="root-url", default=None)
        if url is not None:
            db_config.set(LANDSCAPE_ROOT_URL, str(url))
    return url


def set_root_url(url):
    """
    Save the root_url to the database backend.
    """
    if url and not url.endswith("/"):
        raise ValueError("'%s' doesn't end with a '/'" % url)
    getUtility(ISystemConfiguration).set(LANDSCAPE_ROOT_URL, url)


def set_default_system_email(root_url):
    """Set the default system email based on the root url"""
    parsed_url = urlparse(root_url)
    system_email_hostname = parsed_url.hostname
    configuration = getUtility(ISystemConfiguration)
    system_email = "noreply@" + system_email_hostname
    configuration.set(LANDSCAPE_SYSTEM_EMAIL, system_email)


def get_billing_credentials():
    """Retrieve billing openstack credentials."""
    db_config = getUtility(ISystemConfiguration)
    return db_config.get(BILLING_CREDENTIALS, {})


def set_billing_credentials(endpoint, username, password, tenant):
    """Save the billing openstack credentials used to store invoices."""
    credentials = {
        "keystone_url": endpoint,
        "username": username,
        "password": password,
        "tenant": tenant,
    }
    getUtility(ISystemConfiguration).set(BILLING_CREDENTIALS, credentials)


def get_stripe_public_key():
    """Retrieve the public Stripe API key."""
    db_config = getUtility(ISystemConfiguration)
    return db_config.get(STRIPE_PUBLIC_KEY)


def get_stripe_secret_key():
    """Retrieve the secret Stripe API key."""
    db_config = getUtility(ISystemConfiguration)
    return db_config.get(STRIPE_SECRET_KEY)


def set_stripe_api_keys(public_key, secret_key):
    """
    Save the Stripe API keys to the database backend.
    """
    getUtility(ISystemConfiguration).set(STRIPE_PUBLIC_KEY, public_key)
    getUtility(ISystemConfiguration).set(STRIPE_SECRET_KEY, secret_key)


def get_marketo_credentials():
    """Retrieve marketo credentials."""
    db_config = getUtility(ISystemConfiguration)
    return db_config.get(MARKETO_CREDENTIALS, {})


def set_marketo_credentials(url, client_id, client_secret):
    """Save credentials for marketo used to push marketing information."""
    credentials = {
        "url": url,
        "client_id": client_id,
        "client_secret": client_secret,
    }
    getUtility(ISystemConfiguration).set(MARKETO_CREDENTIALS, credentials)


def get_api_url():
    """Return the API endpoint URL."""
    config = getUtility(IFileBasedConfiguration)
    api_url = None
    if config.has_section("api"):
        api_url = config.get("api").get("root-url", None)
    if not api_url:
        api_url = get_root_url()

    if not api_url.endswith("/"):
        api_url += "/"
    if not api_url.endswith("api/"):
        api_url += "api/"

    return _lowercase_hostname(api_url)


def get_archive_url():
    """
    Read archive URL from service.conf/env var or derive it from the root URL.

    When deriving from root URL, the following rules are applied:
     * if the root URL is under landscape.canonical.com, but is neither
       staging nor edge, prepend "archive."
     * if the root URL is landscape.canonical.com or any other subdomain,
       prepend "archive-" (making archive URL eg. "archive-trunk.l.c.c"
       from "trunk.l.c.c").
     * in all other cases, append "/archive" to the URL.
    """
    config = getUtility(IFileBasedConfiguration)
    archive_url = None
    if config.has_section("pppa-proxy"):
        archive_url = config.get("pppa-proxy").get("archive-url", None)

    # One can set archive-url to a special value "RELATIVE", which means
    # that we want to append "archive/" to Landscape server IP for development
    # and testing purposes.  In these cases, root-url is usually unset, so
    # we can let the code below fall-through unchanged.
    if archive_url and archive_url != "RELATIVE":
        return archive_url

    # XXX 2016-11-08 Danilo: The logic below, other than appending "archive/"
    # to the URL for development and test deployments, can go away after we
    # switch to juju-deployed hosted, since the charm will be setting the
    # archive URL in the config directly.
    root_url = get_root_url()
    parsed_url = list(urlparse(root_url))
    if parsed_url[1].endswith("landscape.canonical.com"):
        sub_domain = parsed_url[1].split(".")[0]
        if sub_domain in ["staging", "edge"]:
            # Under staging and edge, we should have 'archive-'
            # instead of 'archive.'
            parsed_url[1] = "archive-" + parsed_url[1]
        else:
            parsed_url[1] = "archive." + parsed_url[1]
    else:
        if not parsed_url[2].endswith("/"):
            parsed_url[2] += "/"
        parsed_url[2] += "archive/"
    return urlunparse(parsed_url)


def get_custom_ssl_ca():
    """
    Returns the contents of the file if LANDSCAPE_CUSTOM_SSL_CA exists,
    otherwise returns C{None}.
    """
    if os.path.exists(LANDSCAPE_CUSTOM_SSL_CA):
        return open(LANDSCAPE_CUSTOM_SSL_CA, "rb").read()
    return None


def is_development_mode():
    """
    Return C{True} if development mode is enabled.
    """
    return get_config_name() == "default"


def is_selenium_environment():
    """
    Return C{True} if we're running the server in a selenium testing
    environment.
    """
    return get_config_name() == "selenium"


def is_testing_environment():
    """
    Return C{True} if the current config name is for a testing environment.
    """
    config = get_config_name()
    testing_env = os.environ.get(DEBUG_ENV_VAR)
    return config in ("default", "staging", "selenium") or testing_env


def is_staging_environment():
    """
    Return C{True} if the current config name is for a staging environment.
    """
    return get_config_name() == "staging"


def is_standalone():
    """
    Return C{True} if the current config name is for a standalone environment.
    """
    return get_config_name() == "standalone"


def get_vcs_revision():
    """Get the short git commit hash for Landscape."""
    try:
        git_process = subprocess.run(
            ["git", "rev-parse", "--short", "HEAD"],
            capture_output=True,
            check=True,
            text=True,
        )
        return git_process.stdout.strip()
    except subprocess.CalledProcessError:
        return None


def get_revision_from_file():
    """
    Get the revision number for Landscape when the revision file is present.
    """
    try:
        content = read_file(os.path.join(LANDSCAPE_HOME, "REVISION"))
    except IOError:
        return None
    else:
        return content.decode().strip()


def get_revision():
    """Exposes revision details about the application."""
    revision = get_revision_from_file()

    if not revision:
        return get_vcs_revision()

    return revision


def log_revision():
    """
    Returns a standard string containing the revision for the current server.
    This is used to log the revision information in a consistent way.
    """
    version = get_revision() or "Unavailable (None)"
    return "At revision: %s" % version


_static_root_path = None
_static_root_name = None


def get_static_root_path():
    """
    The static root path is a unique root-relative path where static
    resources can be found.
    """
    global _static_root_path
    if _static_root_path is None:
        _static_root_path = "/r/" + get_static_root_name()
    return _static_root_path


def get_static_root_name():
    """
    The static root name is the current revision if
    C{is_development_mode} is C{False}, or a L{uuid.uuid4} if
    C{is_development_mode} is C{True}.
    """
    global _static_root_name
    if _static_root_name is None:
        if not is_development_mode():
            _static_root_name = get_revision()
        if _static_root_name is None:
            _static_root_name = str(uuid.uuid4())
        else:
            _static_root_name = str(_static_root_name)
    return _static_root_name


def get_ping_url(root_url):
    """Get the ping URL to use.

    If a LANDSCAPE_PING_URL environment is set, that one is returned.

    If a ping URL is specified in SystemConfiguration, that one is returned.

    By default the http version of the root URL is used.
    """
    configuration = getUtility(ISystemConfiguration)
    ping_url = configuration.get(
        LANDSCAPE_PING_URL, urlunparse(("http",) + urlparse(root_url)[1:]) + "ping"
    )
    return os.environ.get("LANDSCAPE_PING_URL", ping_url)


def get_message_url(root_url):
    """Get the message system URL to use.

    If a LANDSCAPE_MESSAGE_URL environment is set, that one is returned.

    Otherwise the root URL with "/message-system" appended to it is returned.
    """
    return os.environ.get("LANDSCAPE_MESSAGE_URL", root_url + "message-system")


def get_client_config():
    """Configuration for the client to connect to this landscape instance."""

    Config = namedtuple("Config", ["message_url", "ping_url", "ssl_ca_certificate"])
    root_url = get_root_url()
    custom_ssl_ca = get_custom_ssl_ca()
    config = Config(get_message_url(root_url), get_ping_url(root_url), custom_ssl_ca)
    return config


def is_charm_deployment(landscape_default=LANDSCAPE_DEFAULT):
    """Return whether this a charm deployment."""
    if not os.path.exists(landscape_default):
        return False
    data = read_file(landscape_default)
    return b'DEPLOYED_FROM="charm"' in data


def get_proxy_configuration():
    """Return the proxy configuration, if any.

    @return: A dict with 'http_proxy', 'https_proxy' and 'no_proxy' as
        possible keys. The dict will only be populated if there is a proxy
        defined.

    The environment variables http_proxy, https_proxy and no_proxy are
    searched, both the lowercase and uppercase versions.
    """
    proxy_variables = (
        "HTTP_PROXY",
        "HTTPS_PROXY",
        "NO_PROXY",
        "http_proxy",
        "https_proxy",
        "no_proxy",
    )
    proxy_configuration = {}
    for variable in proxy_variables:
        if variable in os.environ:
            proxy_configuration[variable.lower()] = os.environ[variable]
    return proxy_configuration


def set_proxy_configuration():
    """Set the process HTTP proxy environment variables, if configured."""
    system_configuration = getUtility(ISystemConfiguration)

    http_proxy = system_configuration.get(LANDSCAPE_HTTP_PROXY)
    if http_proxy:
        os.environ["http_proxy"] = http_proxy
        os.environ["HTTP_PROXY"] = http_proxy

    https_proxy = system_configuration.get(LANDSCAPE_HTTPS_PROXY)
    if https_proxy:
        os.environ["https_proxy"] = https_proxy
        os.environ["HTTPS_PROXY"] = https_proxy

    no_proxy = system_configuration.get(LANDSCAPE_NO_PROXY)
    if no_proxy:
        os.environ["no_proxy"] = no_proxy
        os.environ["NO_PROXY"] = no_proxy


def _lowercase_hostname(url):
    """Return the provided URL with lowercased hostname."""
    parts = list(urlparse(url))
    parts[1] = parts[1].lower()
    return urlunparse(parts)
