# Copyright 2005-2016 Canonical Limited.  All rights reserved.

from canonical.lib.undef import Undef

from canonical.collection.collection import CollectionBase

from canonical.landscape.model.entity.data import AccountEntityData


class AccountEntityCollection(CollectionBase):
    """Models a set of AccountEntities."""

    @property
    def access_context(self):
        """Return the access context associated to the account, if any."""
        account = self.filters.get("account")
        return account.access_context if account else None

    def with_account(self, account):
        """Filter account entities by account."""
        return self._filter_by(account=account)

    def with_ids(self, ids):
        """Filter account entities by database ID."""
        return self._filter_by(ids=ids)

    def with_names(self, names):
        """Filter account entities by name."""
        return self._filter_by(names=names)

    def with_access_context_ids(self, access_context_ids):
        """Filter account entities by access context IDs."""
        return self._filter_by(access_context_ids=access_context_ids)

    def _get_result(
        self,
        store,
        where=None,
        account=Undef,
        ids=None,
        names=None,
        access_context_ids=None,
    ):
        """Get a ResultSet of account entities matching search criteria."""

        # Make sure the account parameter is set explicit, to minimize the risk
        # of forgetting about it.
        if account is Undef:
            raise RuntimeError(
                "You must explicitly set the 'account' filter to either an "
                "Account object or to None to search across all accounts in "
                "this store"
            )

        if where is None:
            where = []

        # Filter by entity type
        entity_type = self.model_class.get_entity_type()
        where.extend(
            [
                self.model_class.data_id == AccountEntityData.id,
                AccountEntityData.entity_type == entity_type,
            ]
        )

        if account is not None:
            where.append(AccountEntityData.account_id == account.id)

        if ids is not None:
            where.append(self.model_class.id.is_in(ids))

        if names is not None:
            where.append(AccountEntityData.name.is_in(names))

        if access_context_ids is not None:
            where.append(AccountEntityData.access_context_id.is_in(access_context_ids))

        return store.find(self.model_class, *where)
