from builtins import object
from storm.locals import Int, AutoReload, DateTime, Unicode

from canonical.lib.timestamp import format_datetime


class AccountEntityData(object):
    """Hold common data for all entities living in an account.

    Account entities share basic attributes like access contexts and provide
    the foundation for implementing across-the-board behavior like security
    and RBAC.

    @ivar id: The ID of the entity.
    @ivar account_id: The ID of the Account the entity lives in.
    @ivar entity_type: A numeric code identifying the type of the particular
        account entity this row is linked to.
    @ivar name: The name of the entity, it must be unique per account/type.
    @ivar access_context_id: The ID of the AccessContext the entity is linked
        to. The access context is used for enforcing security: in order to
        access or modify an entity a person needs to have a role providing
        the appropriate permission on the access context the entity is linked
        to.
    @ivar creation_time: Time at which the entity was first created.
    """

    __storm_table__ = "account_entity_data"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    entity_type = Int(allow_none=False)
    name = Unicode(allow_none=False)
    access_context_id = Int(allow_none=False)
    creation_time = DateTime(default=AutoReload)

    def __init__(self, account, entity_type, name, access_context=None):
        self.account_id = account.id
        self.entity_type = entity_type
        self.name = name
        if access_context is None:
            access_context = account.access_context
        self.access_context_id = access_context.id

    @property
    def account(self):
        """Return the Account this entity is associated with."""
        from canonical.landscape.model.main.account import get_account_by_id

        return get_account_by_id(self.account_id)

    @property
    def access_context(self):
        """Return the AccessContext this entity is associated with."""
        return self.account.get_access_context(self.access_context_id)

    def get_state(self):
        """Return the state of the AccountEntity.

        Sub-classes of AccountEntity are supposed to implement their own
        get_state() method that invokes this one for getting the basic
        properties common across all account entities.
        """
        return {
            "name": self.name,
            "access_group": self.access_context.name,
            "creation_time": format_datetime(self.creation_time),
        }


def update_entities_access_context(store, entity_class, ids, access_context):
    """Update the access context for entities of a model."""
    entity_table = entity_class.__storm_table__
    query = """
        UPDATE account_entity_data
        SET access_context_id = ?
        FROM {entity_table}
        WHERE account_entity_data.id = {entity_table}.data_id
          AND {entity_table}.id IN ?
        """
    store.execute(
        query.format(entity_table=entity_table), (access_context.id, tuple(ids))
    )


def remove_entity_data_by_ids(store, ids):
    """Remove AccountEntityData matching specified IDs."""
    store.find(AccountEntityData, AccountEntityData.id.is_in(ids)).remove()
