"""Encapsulate data and behavior common to all account-related entities.

The are a number of attributes and associated behavior that are common across
all models that "live" under an account (e.g. computers, regions, repositories,
etc. are all associated to an access context).

This module provides basic abstractions and tooling for re-using such data
definitions and related behavior.

In particular, the AccountEntity and AccountEntityData classes implement the
supertype/subtype relational pattern, where common schema between two models
is kept in a common table (the supertype) which is referenced with a foreign
key in the derived type (the subtype).
"""

from builtins import object
from zope.interface import implementer

from collections import defaultdict

from storm.locals import Int, AutoReload, Reference

from canonical.landscape.model.entity.data import AccountEntityData
from canonical.landscape.model.main.interfaces import IAccessContextAware
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)


@implementer(IAccessContextAware)
class AccountEntity(object):
    """Base class for models of entities living in an account.

    This offers a base class for Storm models that need to map to the
    underlying SQL tables that implement a supertype/subtype relational
    design.
    """

    id = Int(primary=True, allow_none=False, default=AutoReload)
    data_id = Int(allow_none=False)

    data = Reference(data_id, AccountEntityData.id)

    def __init__(self, account, name, access_context=None):
        self.data = AccountEntityData(
            account, self.get_entity_type(), name, access_context=access_context
        )

    @classmethod
    def get_entity_type(cls):
        """Return the numeric code of this account entity subtype."""
        return _account_entity_class_to_type[cls]

    @property
    def account_id(self):
        """Return the ID of the account this entity is associated with."""
        return self.data.account_id

    @property
    def account(self):
        """Return the Account this entity is associated with."""
        return self.data.account

    @property
    def name(self):
        """Return the name of this entity."""
        return self.data.name

    @property
    def access_context(self):
        """Return the access_context of this entity."""
        return self.data.access_context

    @property
    def creation_time(self):
        """Return the creation type of this entity."""
        return self.data.creation_time

    @property
    def store(self):
        """Return the account store."""
        return self.account.store

    def get_state(self):
        """Return a dict serializing the state of this entity."""
        raise NotImplementedError("Must be implemented by sub-classes")


def get_account_entity(account, entity_class, id=None, name=None):
    """Return an AccountEntity-based model by id or by name."""
    if id is not None:
        # We don't need to join the account_entity table here, since we have
        # the subtype database ID.
        entity = account.store.get(entity_class, id)
        # For sanity, let's just assert that it's
        # really linked to the given account.
        if entity:
            assert entity.account is account
    elif name is not None:
        where = _get_data_join_clause(account, entity_class)
        where.append(AccountEntityData.name == name)
        entity = account.store.find(entity_class, *where).one()
    else:
        raise TypeError("Must provide entity id or name")
    return entity


def register_account_entity_type(entity_type, entity_class):
    """Register an new entity type for the given model class."""
    if entity_type in _account_entity_type_to_class:
        raise RuntimeError("Entity type %d is already registered" % entity_type)
    _account_entity_class_to_type[entity_class] = entity_type
    _account_entity_type_to_class[entity_type] = entity_class


def unregister_account_entity_type(entity_class):
    """Unregister an entity type (for testing)."""
    entity_type = _account_entity_class_to_type.pop(entity_class)
    del _account_entity_type_to_class[entity_type]


def register_account_entity_permissions(entity_type, category, permissions):
    """Register permissions required to perform operations on entity types.

    @param entity_type: The numeric code associated with the account entity
    @param category: The category of operations the permission is required
        for (e.g, "view" or "edit")
    @param permissions: A list of permissions to register for the category
    """
    if entity_type not in _account_entity_type_to_class:
        raise RuntimeError("Entity type %d is is not registered" % entity_type)
    categories = _account_entity_type_to_permissions[entity_type]
    categories[category] = permissions


def unregister_account_entity_permissions(entity_type, category):
    """Unregister an entity type permission (for testing)."""
    categories = _account_entity_type_to_permissions[entity_type]
    del categories[category]
    if not categories:
        del _account_entity_type_to_permissions[entity_type]


def get_account_entity_type(entity_class):
    """Return the numeric code of the entity_type of the given entity_class."""
    return _account_entity_class_to_type[entity_class]


def get_account_entity_class(entity_type):
    """Return the entity class for the given entity code."""
    return _account_entity_type_to_class[entity_type]


def get_access_context_ids_by_entity_type(access_context, person, category):
    """Map entity types to access context IDs with the given permission.

    @param access_context: The root of the AccessContext tree to scan for
        roles and permissions.
    @param person: The Person to find roles/permissions for.
    @param category: The category of permissions to look for. Can be "view"
        for getting access contexts IDs on which the person has view
        permissions on, or "edit".

    @return: A dict whose keys are entity type codes and values are all
        access context IDs that the given person has the given category
        of permissions on. For example if the given person has ViewComputer
        on access contexts 3 and 4, and the given category is 'view', then
        the returned dict will look like:

        {2: [3, 4]}

        where 2 is the info type code for computers.

        See also c.l.model.account.register.register_account_entity_types.
    """
    access_context_ids_by_type = {}
    iterator = iter(_account_entity_type_to_permissions.items())
    for entity_type, categories in iterator:
        access_context_ids = set()

        permissions = categories.get(category, [])
        for permission in permissions:
            access_context_ids.update(
                get_access_context_ids_for_person(access_context, permission, person)
            )
        access_context_ids_by_type[entity_type] = access_context_ids

    return access_context_ids_by_type


def _get_data_join_clause(account, entity_class):
    """Get clauses for performing a join between supertype and subtype.

    This function just returns the boilerplate join Storm query clauses
    need to link the AccountEntity and AccountEntityData tables.
    """
    return [
        entity_class.data_id == AccountEntityData.id,
        AccountEntityData.account_id == account.id,
        AccountEntityData.entity_type == entity_class.get_entity_type(),
    ]


# Map a model class to its entity type code
_account_entity_class_to_type = {}

# Map an entity type code to its model class
_account_entity_type_to_class = {}

# Map an entity type code to a map from operation categories to permissions
_account_entity_type_to_permissions = defaultdict(dict)
