# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Enum
from zope.security.proxy import removeSecurityProxy

from canonical.landscape.model.access.role import StoredRole, GlobalAdmin
from canonical.landscape.model.account.customgraph import CustomGraph
from canonical.landscape.model.account.savedsearch import SavedSearch
from canonical.landscape.model.activity.group import ActivityGroup
from canonical.landscape.model.activity.testing import StubInfo
from canonical.landscape.model.alert.rule import Rule
from canonical.landscape.model.computer.removal import RemovalProfile
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.computer import PendingComputer
from canonical.landscape.model.main.tag import Tag
from canonical.landscape.model.package.autoupgrade import UpgradeProfile
from canonical.landscape.model.package.profile import PackageProfile
from canonical.landscape.model.repository.profile import RepositoryProfile
from canonical.landscape.model.script.script import Script


class LogEntityType(object):
    """A logged entity type."""

    def __init__(self, id=None, _class=None, title=""):
        self.id = id
        self._class = _class
        self.title = title

    def __str__(self):
        return self.title

    def __repr__(self):
        class_ = self.__class__
        return "<%s.%s %r>" % (class_.__module__, class_.__name__, self.title)


# For development and testing purposes
class FakeEntity(object):
    id = None
    name = None

    def __init__(self, id, name):
        self.id = id
        self.name = name


class FakeJujuEntity(FakeEntity):
    pass


# List of enumerators for LogEntityTypeEnum which defines constraints on the
# event_log.entity_type column.

# XXX We can add unique values to this list for each new entity type event
# logs will track.  But, changing an existing enum value will require a
# database patch to migrate existing logs of that entity_type to the new index.
ENTITY_TEST = LogEntityType(0, FakeEntity, "Test entity")
ENTITY_COMPUTER = LogEntityType(1, Computer, "Computer")
ENTITY_TAG = LogEntityType(2, Tag, "Tag")
ENTITY_ROLE = LogEntityType(3, StoredRole, "Role")
ENTITY_SCRIPT = LogEntityType(4, Script, "Script")
ENTITY_PACKAGE_PROFILE = LogEntityType(5, PackageProfile, "Package profile")
ENTITY_REMOVAL_PROFILE = LogEntityType(6, RemovalProfile, "Removal profile")
ENTITY_REPOSITORY_PROFILE = LogEntityType(7, RepositoryProfile, "Repository profile")
ENTITY_UPGRADE_PROFILE = LogEntityType(8, UpgradeProfile, "Upgrade profile")
ENTITY_CUSTOM_GRAPH = LogEntityType(9, CustomGraph, "Custom graph")

# XXX Juju is no longer supported - rather than requiring a db migration,
# mark as unused with FakeEntity.
ENTITY_JUJU_ENVIRONMENT = LogEntityType(10, FakeJujuEntity, "Juju environment")

ENTITY_RULE = LogEntityType(11, Rule, "Rule")
ENTITY_GLOBAL_ADMIN = LogEntityType(12, GlobalAdmin, "Global admin")
ENTITY_ACCOUNT = LogEntityType(13, Account, "Account")
ENTITY_SAVED_SEARCH = LogEntityType(15, SavedSearch, "Saved search")
ENTITY_PENDING_COMPUTER = LogEntityType(16, PendingComputer, "Pending computer")
ENTITY_ACTIVITY_GROUP = LogEntityType(17, ActivityGroup, "Activity group")

# This is used for testing only.
ENTITY_STUB_INFO = LogEntityType(9999, StubInfo, "Stub info")

ALL_ENTITY_TYPES = [
    ENTITY_TEST,
    ENTITY_COMPUTER,
    ENTITY_TAG,
    ENTITY_ROLE,
    ENTITY_SCRIPT,
    ENTITY_PACKAGE_PROFILE,
    ENTITY_REMOVAL_PROFILE,
    ENTITY_REPOSITORY_PROFILE,
    ENTITY_UPGRADE_PROFILE,
    ENTITY_CUSTOM_GRAPH,
    ENTITY_JUJU_ENVIRONMENT,
    ENTITY_RULE,
    ENTITY_GLOBAL_ADMIN,
    ENTITY_ACCOUNT,
    ENTITY_SAVED_SEARCH,
    ENTITY_PENDING_COMPUTER,
    ENTITY_ACTIVITY_GROUP,
    ENTITY_STUB_INFO,
]


ENTITY_TYPE_MAP = {}
ENTITY_TYPE_SET_MAP = {}
ENTITY_TYPE_GET_MAP = {}
for entity_type in ALL_ENTITY_TYPES:
    ENTITY_TYPE_MAP[entity_type] = entity_type.id
    ENTITY_TYPE_SET_MAP[entity_type] = entity_type.id
    ENTITY_TYPE_GET_MAP[entity_type._class] = entity_type


class LogEntityTypeEnum(Enum):
    def __init__(self, allow_none=False):
        super(LogEntityTypeEnum, self).__init__(
            map=ENTITY_TYPE_MAP, set_map=ENTITY_TYPE_SET_MAP, allow_none=allow_none
        )


def get_log_entity_type(entity):
    """Determine the event-log-centric entity type of the given object."""
    # Entities can be wrapped in security proxies, and security proxies
    # don't lie about their actual type, so we have to momentarily unwrap
    # the entity in order to identify it.
    return ENTITY_TYPE_GET_MAP[type(removeSecurityProxy(entity))]
