from builtins import str
from storm.locals import Int, Unicode, AutoReload, DateTime

from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.eventlog.entities import LogEntityTypeEnum
from canonical.lib.timestamp import format_datetime
from canonical.landscape.model.eventlog.entities import get_log_entity_type

missing = object()


class EventLogEntry(AccountHelper):
    """An event log entry in an L{Account} of operations causing state changes.

    @ivar account_id: ID of the L{Account} to which this log entry belongs.
    @ivar creation_time: The time at which this entry was created.
    @ivar person_id: ID of the L{Person} performing the operation.
        Defaults to 0 if no person performed the operation.
    @ivar person_name: Name of the L{Person} performing the operation.
        Defaults to "Landscape" if no person performed the operation.
    @ivar entity_type: A string representing the class of entity operated on.
    @ivar entity_id: ID of the entity on which the operation was performed.
    @ivar entity_name: Name of the entity on which the operation was performed.
    @ivar message: The full English text summary of the operation logged and
        all impacts of that operation on the entity.
    """

    __storm_table__ = "event_log_entry"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    creation_time = DateTime(default=AutoReload)
    person_id = Int()
    person_name = Unicode(allow_none=False, default="Landscape")
    entity_type = LogEntityTypeEnum(allow_none=False)
    entity_id = Int()
    entity_name = Unicode(allow_none=False)
    message = Unicode(allow_none=False)

    def __init__(self, account, entity, message, person=None):
        self.entity_type = get_log_entity_type(entity)
        self.entity_name = get_log_entity_name(entity)
        self.entity_id = getattr(entity, "id", None)
        self.message = message
        if person is not None:
            self.person_id = person.id
            self.person_name = person.name
        # Call the base class constructor last to avoid storm flush mid-init
        super(EventLogEntry, self).__init__(account=account)

    def get_state(self):
        """@return: a dict with L{EventLogEntry} information."""
        info = ("id", "person_id", "person_name", "entity_id", "entity_name", "message")
        state = dict((field, getattr(self, field)) for field in info)
        state["creation_time"] = format_datetime(self.creation_time)
        state["entity_type"] = str(self.entity_type)
        return state


def get_log_entity_name(entity):
    """Determine an entity's name via its "name" or "title" attribute."""
    name = getattr(entity, "name", missing)
    if name is missing:
        name = getattr(entity, "title", missing)
    if name is missing:
        raise TypeError(
            "Unable to log entity data. Entity '%s' requires either name "
            "or title attribute." % entity.__class__.__name__
        )
    return name
