# Copyright 2005-2022 Canonical Limited. All rights reserved

from storm.locals import Desc, And, Insert

from canonical.landscape.model.management import ModelManagement


__all__ = ["EventLog"]

__clearlist__ = ["EventLog"]


class EventLog(ModelManagement):
    """Manage Landscape event logs"""

    def clear_log(self):
        """Clear all event log entries from this account.

        Testing purposes only because an C{EventLog} is read-only.
        """
        from canonical.landscape.model.eventlog.entry import EventLogEntry

        result = self.account.store.find(
            EventLogEntry, EventLogEntry.account_id == self.account.id
        )
        result.remove()

    def create_log_entries_for_computers(self, store, computer_ids, message):
        """Add event log entries for a provided list of Computer ids.

        @param computer_ids: A list of ids of computers changed.
        @param message: Prose describing the action performed.
        """
        from canonical.landscape.model.main.computer import Computer
        from canonical.landscape.model.eventlog.entities import ENTITY_COMPUTER
        from canonical.landscape.model.main.store import get_main_store

        main_store = get_main_store()
        result = main_store.find(
            (Computer.id, Computer.title), Computer.id.is_in(computer_ids)
        )

        row_values = []  # The values of all rows that will be inserted
        columns = ["account_id", "entity_type", "entity_id", "entity_name", "message"]
        if self.person:
            columns.extend(["person_id", "person_name"])

        for entity_id, entity_name in result:
            values = [
                self.account.id,
                ENTITY_COMPUTER.id,
                entity_id,
                entity_name,
                message,
            ]
            if self.person:
                values.extend([self.person.id, self.person.name])
            row_values.append(tuple(values))

        if row_values:
            store.execute(Insert(columns, table="event_log_entry", values=row_values))

    def create_log_entry(self, entity, message):
        """Add a log entry.

        @param entity: the thing this log entry is about.
        @param message: prose describing the action performed.

        @returns: the created EventLogEntry.
        """
        from canonical.landscape.model.eventlog.entry import EventLogEntry

        return EventLogEntry(self.account, entity, message, self.person)

    def get_entries_count(self, start_time=None):
        """Return a count of all C{EventLogEntry}s in this account."""
        from canonical.landscape.model.eventlog.entry import EventLogEntry

        where = [EventLogEntry.account_id == self.account.id]
        if start_time:
            where.append(EventLogEntry.creation_time >= start_time)
        return self.account.store.find(EventLogEntry, *where).count()

    def get_entries(
        self, about=None, message=None, start_time=None, offset=0, limit=1000
    ):
        """Return a result set of all the C{EventLogEntry}s in this account.

        @param about: Only return entries about this entity.
        Optionally filter the results to only include logs on or after
        the specified C[start_time}, at the given C{offset} and up to the
        specified C{limit} which defaults to 1000.
        """
        from canonical.landscape.model.eventlog.entry import EventLogEntry
        from canonical.landscape.model.eventlog.entities import get_log_entity_type

        where = [EventLogEntry.account_id == self.account.id]
        if start_time:
            where.append(EventLogEntry.creation_time >= start_time)
        if about is not None:
            where.append(
                And(
                    EventLogEntry.entity_type == get_log_entity_type(about),
                    EventLogEntry.entity_id == about.id,
                )
            )
        if message is not None:
            where.append(EventLogEntry.message == message)
        result = self.account.store.find(EventLogEntry, *where)
        result.order_by(Desc(EventLogEntry.creation_time), EventLogEntry.id)
        return result.config(limit=limit, offset=offset)
