# Copyright 2019 Canonical Limited. All rights reserved.

from datetime import timedelta
import logging

import transaction

from canonical.lib.timestamp import utcnow

from canonical.landscape.model.main.account import get_all_accounts
from canonical.landscape.model.eventlog.entry import EventLogEntry


def cleanup_events(older_than_days, dry_run=False):
    """
    Remove L{EventLogEntry} objects with creation date older than
    C{older_than_days} days.

    @param older_than_days: Remove event logs older than these many days.
    @param dry_run: True if we should only identify and count the event log
        entries that are to be removed.
    """
    expiration_time = utcnow() - timedelta(days=older_than_days)
    deleted_count = 0

    for account in get_all_accounts():
        entries = account.store.find(
            EventLogEntry,
            EventLogEntry.account_id == account.id,
            EventLogEntry.creation_time < expiration_time,
        )
        count = entries.count()
        if count > 0:
            deleted_count += count
            if not dry_run:
                with transaction.manager as manager:
                    logging.info(
                        "Removing %d event log entries for account '%s'.",
                        count,
                        account.name,
                    )
                    try:
                        entries.remove()
                    except Exception:
                        logging.exception("Error removing entries.")
                        manager.abort()
            else:
                logging.info(
                    "Did NOT remove %d event log entries for account '%s'.",
                    count,
                    account.name,
                )

    if deleted_count == 0:
        logging.info("No event log entries older than %d days found.", older_than_days)
    elif not dry_run:
        logging.info(
            "Removed %d event log entries older than %d days.",
            deleted_count,
            older_than_days,
        )
