# Copyright 2005-2015 Canonical Limited.  All rights reserved.

"""API for accessing feature flags.

The public API for querying feature flags is is_feature_enabled().

Tests should use feature_flags as a class/method decorator to override flag
enable/disable state for the duration of a test.
"""

from builtins import str
from builtins import object
from storm.locals import Unicode, Storm

from canonical.landscape.model.main.store import get_main_store


__all__ = [
    "is_feature_enabled",
    "enable_feature",
    "disable_feature",
    "get_feature_flags",
    "FeatureFlags",
    "BUILTIN_FEATURES",
]


__clearlist__ = ["is_feature_enabled", "get_feature_flags", "BUILTIN_FEATURES"]


# A list of built-in features. Once a feature flag gets dropped and the feature
# permanently enabled, it should be added to this list to avoid reuse in the
# future.
#
# This list, plus the enabled feature flags, will be inspected by system
# tests in order to allow system tests to skip absent or disabled features to
# avoid testing features on old releases which may not be supported.
BUILTIN_FEATURES = [
    "osa-nagios",
    "openstack-roles",
    "openstack-image-refresh",
    "maas-2.0",
    "juju-2.0",
    "openstack-lxd",
]

# The set of all known feature flags and the description of those flags
_REGISTERED_FEATURE_FLAGS = {
    "juju-debug": "Set Juju root logger to DEBUG when bootstrapping",
    "juju-gui": "Deploy and link to Juju GUI in the Autopilot",
    "openstack-debug": "Enable debug log in OpenStack services",
    "openstack-maintenance": "Enable support for OpenStack node maintenance",
    "openstack-next-charms": "Use -next charms when deploying OpenStack",
    "self-service": "Enable self-service Landscape",
    "repository-mirror-ui": "Enable the repository mirroring UI",
}


class EnabledFeature(Storm):
    """Presence of a name in this table indicates the feature is enabled.

    An EnabledFeature enables a specific named feature in Landscape that is
    otherwise hidden or unpublished by default. Feature flags are used to
    prevent release of features that are not yet blessed for production
    release.

    @ivar name: An unique name for an enabled feature flag. The name must exist
        in the registered_feature_flags dict.
    @type name: Unicode
    """

    __storm_table__ = "enabled_feature"

    name = Unicode(primary=True, allow_none=False)

    def __init__(self, name):
        self.name = name


def ensure_feature_registered(name):
    """Verify that the named flag is in fact registered."""
    if name not in _feature_flags.registry:
        raise ValueError("unregistered feature name %r" % name)


def _is_feature_enabled_in_database(name):
    """Return True if the flag exists and is enabled."""
    ensure_feature_registered(name)
    store = get_main_store()
    flag = store.find(EnabledFeature, EnabledFeature.name == name).one()
    return flag is not None


def enable_feature(name):
    """Enable the named feature.

    Returns True if any work was done (the feature wasn't already
    enabled), and False if already enabled.  If the feature is not
    allowed to be enabled then None is returned.
    """
    ensure_feature_registered(name)
    if _is_feature_enabled_in_database(name):
        return False
    get_main_store().add(EnabledFeature(name))
    return True


def disable_feature(name):
    """Disable the named feature.

    Returns True if any work was done (the feature wasn't already
    disabled), and False if already disabled.  If the feature is not
    allowed to be disabled then None is returned.
    """
    ensure_feature_registered(name)
    if not _is_feature_enabled_in_database(name):
        return False
    store = get_main_store()
    store.find(EnabledFeature, EnabledFeature.name == name).remove()
    return True


class FeatureFlags(object):
    """A representation of the systems current set of feature flags.

    This class is an implementation detail.  It should only be used directly
    by other feature flag code.
    """

    def __init__(self, registry=_REGISTERED_FEATURE_FLAGS, **flags):
        self.registry = registry
        # All flag names must be unicode.
        flags = {str(name): value for name, value in list(flags.items())}
        # Unspecified flags are disabled.
        self.flag_enabled = {name: False for name in self.registry}
        # Only already-registered names are allowed.
        assert all(name in registry for name in flags), "unrecognized flag"
        self.flag_enabled.update(flags)
        self._overridden_feature_flags = None

    def get_feature_flags(self):
        """Fetch a list of feature flag details (sorted by name)."""
        flags = []
        for name, description in sorted(self.registry.items()):
            flags.append(
                {
                    "name": name,
                    "description": description,
                    "enabled": self.is_feature_enabled(name),
                }
            )
        return flags

    def is_feature_enabled(self, name):
        """Is the named feature enabled?"""
        assert name in self.registry
        return self.flag_enabled[name]

    def install(self):
        """Override the database-provided feature flags with this object."""
        if self._overridden_feature_flags is not None:
            raise RuntimeError("these feature flags are already installed")
        global _feature_flags
        self._overridden_feature_flags = _feature_flags
        _feature_flags = self

    def uninstall(self):
        """Restore the original feature flag functions."""
        if self._overridden_feature_flags is None:
            # already uninstalled
            return
        global _feature_flags
        _feature_flags = self._overridden_feature_flags
        self._overridden_feature_flags = None

    def __enter__(self):
        self.install()
        return self

    def __exit__(self, *ignored):
        self.uninstall()


class DatabaseFeatureFlags(FeatureFlags):
    """Feature flags backed by the database.

    This class is an implementation detail.  It should only be used directly
    by other feature flag code.
    """

    def __init__(self, registry=_REGISTERED_FEATURE_FLAGS):
        super(DatabaseFeatureFlags, self).__init__(registry)

    def is_feature_enabled(self, name):
        """Is the named feature enabled?"""
        return _is_feature_enabled_in_database(name)


# The system-wide feature flags.
_feature_flags = DatabaseFeatureFlags()


def is_feature_enabled(name):
    """Is the named feature enabled?"""
    return _feature_flags.is_feature_enabled(name)


def get_feature_flags():
    """Fetch a list of mappings of feature flag details (sorted by name)."""
    return _feature_flags.get_feature_flags()
