# Copyright 2005-2015 Canonical Limited.  All rights reserved.
from builtins import object
import inspect
from functools import wraps
from unittest import TestCase

from canonical.landscape.model.feature import flag


def feature_flags(name=None, value=True, **flags):
    """Return a decorator that sets the provided feature flags in tests."""
    if name is not None:
        flags[name] = value
    return FeatureFlags(**flags)


class FeatureFlags(object):
    """A class and method decorator for enabling feature flags in tests.

    May also be used as a context manager.
    """

    def __init__(self, **flags):
        self.flags = flags
        # Base the imposed feature flags on the current system-wide feature
        # flags.
        self.base = flag._feature_flags.registry

    def _make_wrapper(self, func):
        func, wrapped_flags = _unwrap(func)
        combined_flags = dict(wrapped_flags, **self.flags)
        flags = flag.FeatureFlags(registry=self.base, **combined_flags)

        @wraps(func)
        def func_wrapper(*args, **kws):
            with flags:
                return func(*args, **kws)

        # These attrs are used by _unwrap() in the nested wrapper case.
        func_wrapper.__wrapped__ = func
        func_wrapper._feature_flags = combined_flags

        return func_wrapper

    def __call__(self, thing):
        if inspect.isclass(thing) and issubclass(thing, TestCase):
            # If this is a test case class, decorate each test method.
            for method_name in dir(thing):
                if method_name.startswith("test_"):
                    method = getattr(thing, method_name)
                    wrapper = self._make_wrapper(method)
                    setattr(thing, method_name, wrapper)
            return thing
        else:
            # Otherwise this is an individual test method, so just wrap it.
            return self._make_wrapper(thing)

    def __enter__(self):
        self.context = flag.FeatureFlags(registry=self.base, **self.flags)
        return self.context.__enter__()

    def __exit__(self, *ignored):
        return self.context.__exit__(*ignored)


def _unwrap(func):
    """Return (func, flags) based on the given func.

    If the provided func looks like it was originally wrapped by a
    FeatureFlags then the returned func will be the originally wrapped
    one and the returned flags will be a copy of the ones with which it
    had been wrapped so far.  Otherwise the provided func is returned
    as-is, along with an empty dict for the flags.
    """
    try:
        flags = func._feature_flags
    except AttributeError:
        return func, {}

    # Get back to the originally wrapped function.
    try:
        func = func.__wrapped__
    except AttributeError:
        pass

    return func, flags
