# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, Unicode, AutoReload, Or

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.account.helper import AccountHelper


__all__ = ["DuplicateGPGKeyError", "GPGKey"]

__clearlist__ = ["DuplicateGPGKeyError", "GPGKey"]

__requires_prefix__ = "landscape."


class DuplicateGPGKeyError(DuplicateModelError):
    """
    Raised when a key with same name or fingerprint is present in the account.
    """

    model_name = "GPG key"


@requires_advisor
class GPGKey(AccountHelper):

    __storm_table__ = "gpg_key"

    requires(
        "ViewGPGKey",
        PRIVATE,
        ["id", "account", "key_id", "get_state", "get_pockets", "get_apt_sources"],
    )
    requires(
        "ViewGPGKey",
        "EditGPGKey",
        ["name", "fingerprint", "material", "secret_material"],
    )
    requires("EditGPGKey", PRIVATE, ["remove"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    fingerprint = Unicode(allow_none=False)
    material = Unicode(allow_none=False)
    secret_material = Unicode()

    def __init__(self, account, name, fingerprint, material, secret_material=None):
        super(GPGKey, self).__init__(account)
        self.name = name
        self.fingerprint = fingerprint
        self.material = material
        self.secret_material = secret_material

    @property
    def key_id(self):
        """Return the key id (the last 16 chars of the fingerprint)."""
        return self.fingerprint.replace(":", "")[-16:].upper()

    def get_state(self):
        """Return a dict with information about the key."""
        return {
            "id": self.id,
            "name": self.name,
            "key_id": self.key_id,
            "fingerprint": self.fingerprint,
            "has_secret": self.secret_material is not None,
        }

    def get_pockets(self):
        """Return the pockets this key is linked to."""
        from canonical.landscape.model.repository.pocket import (
            Pocket,
            PocketUploadGPGKey,
        )

        result = self.account.store.find(
            Pocket,
            Or(Pocket.gpg_key_id == self.id, Pocket.mirror_gpg_key_id == self.id),
        )
        return result.union(
            self.account.store.find(
                Pocket,
                PocketUploadGPGKey.pocket_id == Pocket.id,
                PocketUploadGPGKey.gpg_key_id == self.id,
            )
        ).order_by(Pocket.id)

    def get_apt_sources(self):
        """Return the APT sources this key is linked to."""
        from canonical.landscape.model.repository.aptsource import APTSource

        return self.account.store.find(
            APTSource, APTSource.gpg_key_id == self.id
        ).order_by(APTSource.id)

    @requires("EditGPGKey")
    def remove(self):
        self.account.store.remove(self)
