# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.gpg.key import GPGKey, DuplicateGPGKeyError
from canonical.security.protect import requires, requires_advisor, PRIVATE

__all__ = ["GPGKeyManagement"]

__clearlist__ = ["GPGKeyManagement"]

__requires_prefix__ = "landscape."


@requires_advisor
class GPGKeyManagement(ModelManagement):
    """Manage L{GPGKey}s linked to an L{Account}."""

    requires("ViewGPGKey", PRIVATE, ["account", "person"])

    @requires("CreateGPGKey")
    def create_key(self, name, fingerprint, material, secret_material=None):
        """Create a L{GPGKey}."""
        self.check_duplicate(
            GPGKey, DuplicateGPGKeyError, name=name, fingerprint=fingerprint
        )

        return GPGKey(
            self.account, name, fingerprint, material, secret_material=secret_material
        )

    @requires("ViewGPGKey")
    def get_keys(self, names=None, ids=None, fingerprints=None):
        """Get the account's L{GPGKey}s.

        @param names: Only get L{GPGKey}s with the given names.
        @param ids: Only get L{GPGKey}s with the given ids.
        @param fingerprints: Only get L{GPGKey}s with the given fingerprints.
        """
        where = []
        if names is not None:
            where.append(GPGKey.name.is_in(names))
        if ids is not None:
            where.append(GPGKey.id.is_in(ids))
        if fingerprints is not None:
            where.append(GPGKey.fingerprint.is_in(fingerprints))
        result = self._find(*where)
        return result.order_by(GPGKey.name)

    def get_key(self, name):
        """
        Conveniency method to get a single key.
        """
        return self.get_keys(names=[name]).one()

    @requires("EditGPGKey")
    def remove_keys(self, key_ids):
        """Remove L{GPGKey}s from L{Account}."""
        self._find(GPGKey.id.is_in(key_ids)).remove()

    def _find(self, *where):
        """Find L{GPGKey}s in our L{Account}.

        @param where: An optional list for Storm expressions.
        @return: A L{ResultSet} yielding L{GPGKey}s in the account
            matching the given expressions.
        """
        return self.account.store.find(
            GPGKey, GPGKey.account_id == self.account.id, *where
        )
