# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer
from zope.security import checkPermission

from canonical.lib.timestamp import utcnow

from storm.locals import (
    Unicode,
    DateTime,
    Int,
    Bool,
    ReferenceSet,
    AutoReload,
    Store,
    Reference,
    Count,
)

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.storm.variables import BytesStr

from canonical.landscape.model.textsearch import (
    TSVectorValidator,
    TSVectorPart,
    get_text_from_wiki,
)
from canonical.landscape.model.knowledge.category import Category, ArticleCategoryLink
from canonical.landscape.model.knowledge.store import get_knowledge_store
from canonical.landscape.model.knowledge.interfaces import IArticle
from canonical.landscape.model.knowledge.attachment import Attachment
from canonical.landscape.model.knowledge.vote import ArticleVote


__all__ = [
    "Article",
    "ArticleView",
    "get_article_by_id",
    "get_articles_by_category_id",
    "create_article",
    "may_view_unpublished_articles",
]

__clearlist__ = ["may_view_unpublished_articles"]

__requires_prefix__ = "landscape."


WIKI_VALIDATOR = TSVectorValidator(
    "text_search_vector",
    TSVectorPart("title", weight="A"),
    TSVectorPart("body", weight="B", transform=(get_text_from_wiki,)),
)


@requires_advisor
@implementer(IArticle)
class Article(object):
    """An article in the knowledge base."""

    __storm_table__ = "article"
    requires("ViewKnowledgeBase", PRIVATE, ["id", "creation_time", "categories"])

    requires(
        "ViewKnowledgeBase",
        "EditKnowledgeBase",
        ["title", "body", "published", "modification_time", "attachments"],
    )

    requires("ViewKnowledgeBase", "ViewKnowledgeBase", ["times_viewed"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    title = Unicode(allow_none=False, validator=WIKI_VALIDATOR)
    body = Unicode(allow_none=False, validator=WIKI_VALIDATOR)
    creation_time = DateTime(allow_none=False)
    modification_time = DateTime(allow_none=False)
    published = Bool(allow_none=False)
    times_viewed = Int(allow_none=False, default=AutoReload)
    text_search_vector = Unicode()

    categories = ReferenceSet(
        id,
        ArticleCategoryLink.article_id,
        ArticleCategoryLink.category_id,
        Category.id,
        order_by=Category.title,
    )
    attachments = ReferenceSet(id, Attachment.article_id, order_by=Attachment.filename)
    votes = ReferenceSet(id, ArticleVote.article_id)

    def __init__(self, title, body, published=False):
        now = utcnow()
        self.title = title
        self.body = body
        self.published = published
        self.creation_time = now
        self.modification_time = now

    @requires("EditKnowledgeBase")
    def remove_category_by_ids(self, category_ids):
        """Remove the given categories from this article."""
        if not category_ids:
            return
        store = Store.of(self)
        result = store.find(
            ArticleCategoryLink,
            ArticleCategoryLink.article_id == self.id,
            ArticleCategoryLink.category_id.is_in(category_ids),
        )
        result.remove()

    @requires("EditKnowledgeBase")
    def add_category_by_ids(self, category_ids):
        """Add the given categories to this article."""
        store = Store.of(self)
        for category in category_ids:
            link = ArticleCategoryLink()
            link.category_id = category
            link.article_id = self.id
            store.add(link)

    @requires("EditKnowledgeBase")
    def add_attachment(self, filename, data):
        """
        Attach an attachment to this article.

        @param filename: name of the attachment.
        @type filename: C{unicode}

        @param data: content of the attachment.
        @type data: C{str}

        @return: the attachment instance created.
        @rtype: L{ScriptAttachment}
        """
        attachment = Attachment(filename)
        self.attachments.add(attachment)
        attachment.set_data(data)
        return attachment

    @requires("EditKnowledgeBase")
    def remove_attachments_by_ids(self, ids):
        """
        Remove from the article the attachments matching the given identifiers.
        """
        store = Store.of(self)
        ids = set(ids)
        if not ids:
            return
        result = store.find(Attachment, Attachment.id.is_in(ids))
        result.remove()

    @requires("ViewKnowledgeBase")
    def get_attachment_by_id(self, id):
        """Retrieve one article attachment by its id."""
        return self.attachments.find(Attachment.id == id).one()

    @requires("ViewKnowledgeBase")
    def get_attachment_by_filename(self, filename):
        """Retrieve one article attachment by its filename."""
        return self.attachments.find(Attachment.filename == filename).one()

    @requires("EditKnowledgeBase")
    def remove(self):
        Store.of(self).remove(self)

    @requires("VoteKnowledgeBase")
    def vote(self, person, helpful):
        """
        Make the given person to vote for this article, finding it helpful or
        not.
        """
        current_vote = self.get_vote_by_person(person)
        if current_vote is None:
            self.votes.add(ArticleVote(person, self, helpful))
        else:
            current_vote.helpful = helpful
            current_vote.vote_time = utcnow()

    @requires("ViewKnowledgeBase")
    def get_vote_by_person(self, person):
        """Return the vote made by the given person, if any."""
        return self.votes.find(person_id=person.id).one()

    @requires("ViewVotesKnowledgeBase")
    def get_votes(self):
        """
        Return a dictionary with the numbers of votes finding this article
        helpful and unhelpful.
        """
        result = Store.of(self).find(
            (ArticleVote.helpful, Count(ArticleVote.helpful)),
            ArticleVote.article_id == self.id,
        )
        result.group_by(ArticleVote.helpful)
        votes = {True: 0, False: 0}
        votes.update(result)
        return votes

    @requires("ViewVotesKnowledgeBase")
    def reset_votes(self):
        """Reset all votes on this article."""
        self.votes.find().remove()

    @requires("ViewKnowledgeBase")
    def add_times_viewed(self, client_id):
        """Register a new view of this article, for an user in the specified
        session.

        @param client_id: a C{str} instance identifying the current session of
            the user.
        """
        view = ArticleView(client_id.encode())
        self.views.add(view)

    @requires("ViewKnowledgeBase")
    def get_consolidated_times_viewed(self):
        """Return the number of times this article have been viewed.

        It aggregates the rows in the article_view table for this article,
        removing duplicate views by checking sessions, and add the value to the
        C{times_viewed} attributes which should have been calculated by the
        previous views in maintenance.
        """
        from storm.expr import Select, Alias

        store = Store.of(self)
        subselect = Select(
            ArticleView.client_id,
            ArticleView.article_id == self.id,
            group_by=ArticleView.client_id,
        )
        select = Select(Count(), tables=Alias(subselect, "_tmp"))
        count = store.execute(select).get_all()[0][0]
        return count + self.times_viewed


Attachment.article = Reference(Attachment.article_id, Article.id)


@requires_advisor
class ArticleView(object):
    """A view of an article by an user.

    This object logs when an user views an article in the knowledge base. To
    prevent serialization errors, there is no unique key on article/session, so
    we should be careful to remove potential duplicates when getting info out
    of this table.
    """

    __storm_table__ = "article_view"

    requires("ViewKnowledgeBase", PRIVATE, ["id", "article_id", "article", "client_id"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    article_id = Int(allow_none=False)
    client_id = BytesStr(allow_none=False)
    article = Reference(article_id, Article.id)

    def __init__(self, client_id):
        self.client_id = client_id


Article.views = ReferenceSet(Article.id, ArticleView.article_id)


def get_article_by_id(id):
    """Get an article by its id."""
    return get_knowledge_store().get(Article, id)


def may_view_unpublished_articles():
    """True if the current principal may view unpublished articles."""
    return checkPermission("landscape.EditKnowledgeBase", None) or checkPermission(
        "landscape.ReviewKnowledgeBase", None
    )


def get_articles_by_category_id(category_id):
    """Return the articles in the given category, ordered by title."""
    store = get_knowledge_store()
    where = [
        Article.id == ArticleCategoryLink.article_id,
        ArticleCategoryLink.category_id == category_id,
    ]
    if not may_view_unpublished_articles():
        where.append(Article.published == True)
    result = store.find(Article, *where)
    result.order_by(Article.title)
    return result


def create_article(title, body, published, category_ids):
    """Create a new article in the knowledge base."""
    store = get_knowledge_store()
    article = Article(title, body, published)
    store.add(article)
    article.add_category_by_ids(category_ids)
    return article
