# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from canonical.lib.creoleparser.core import Parser
from canonical.lib.creoleparser.elements import Heading
from canonical.lib.creoleparser.dialects import Creole10


__all__ = ["create_wiki_parser"]
__clearlist__ = __all__


class LandscapeWikiDialect(Creole10):
    """Landscape-specific wiki formatting tweaks.

    Landscape does rounded corner magic on HTML h2 tags. We don't want to have
    article wiki content rendered where one of the article headings is rendered
    above the article title itself (e.g., h1) or styled the same as the article
    title (e.g., h2), we want to force all headings to be h3 or lower. For
    this, we need to define our own dialect.
    """

    def __init__(self, *args, **kwds):
        super(LandscapeWikiDialect, self).__init__(*args, **kwds)
        inline_elements = [
            self.no_wiki,
            self.img,
            self.link,
            self.br,
            self.raw_link,
            self.strong,
            self.em,
        ]
        self.h1 = Heading("h3", "=", inline_elements)
        self.h2 = Heading("h4", "==", inline_elements)
        self.h3 = Heading("h5", "===", inline_elements)
        self.h4 = Heading("h6", "====", inline_elements)
        self.h5 = Heading("h6", "=====", inline_elements)
        self.h6 = Heading("h6", "======", inline_elements)
        headings = [self.h1, self.h2, self.h3, self.h4, self.h5, self.h6]

        if kwds.get("use_additions"):
            block_elements = [
                (self.bodied_block_macro, self.pre, self.block_macro),
                self.blank_line,
                self.table,
                self.dl,
            ]
        else:
            block_elements = [self.pre, self.blank_line, self.table]

        self.block_elements = (
            block_elements
            + headings
            + [self.hr, self.ul, self.ol, self.lone_place_holder, self.p]
        )


def create_wiki_parser(format="xhtml"):
    """
    A convenience function for easily creating a parser from the Landscape
    KnowledgeBase creoleparser dialect.

    @param format: the format the output should be in; creoleparser and genshi
        use name "method" for this, which is somewhat obscure in this context,
        thus the use of the clearer "format" name; legal values are "html",
        "xhtml", "xml", and "text".

    @return: a creoleparser Parser instance.
    """
    return Parser(dialect=LandscapeWikiDialect(use_additions=True), method=format)
