# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from zope.interface import implementer

from storm.locals import Or, Like, Desc

from canonical.landscape.model.textsearch import (
    TSRankCD,
    TSHeadline,
    Match,
    get_search_query,
    AND,
    TEXT_SEARCH_CONFIG,
    RANK_LOG_LENGTH,
)
from canonical.landscape.model.knowledge.interfaces import IKnowledgeBase
from canonical.landscape.model.knowledge.store import get_knowledge_store
from canonical.landscape.model.knowledge.article import (
    Article,
    may_view_unpublished_articles,
)


__all__ = ["KnowledgeBase", "search_articles", "search_articles_simple"]

__clearlist__ = ["KnowledgeBase"]


__requires_prefix__ = "landscape."


@implementer(IKnowledgeBase)
class KnowledgeBase(object):
    """The class associated with the knowledge base search."""


def search_articles_simple(search_string):
    """
    Search for articles those title or body match the search_string provided.
    The search is case insensitive.
    """
    store = get_knowledge_store()
    substring = search_string.replace("%", "\\%")
    substring = "%" + substring + "%"
    where = [
        Or(
            Like(Article.title, substring, case_sensitive=False),
            Like(Article.body, substring, case_sensitive=False),
        )
    ]
    if not may_view_unpublished_articles():
        where.append(Article.published == True)
    result = store.find(Article, *where)
    result.order_by(Article.title)
    return result


def search_articles(search_terms, search_mode=AND, rank_mode=RANK_LOG_LENGTH):
    """
    Perform a full text search against the articles, matching the title or
    body, and returning the article fields as well as search rank and
    excerpt.

    @param search_expression: search_expression can be a simple word or it can
        be a full expression. Multiple search terms are combined with "&", "|",
        and "!" and can be grouped with parentheses to enforce precedence.  In
        the absence of parentheses, "!" binds most tightly, and "&" binds more
        tightly than "|".

    @param search_mode: this gets passed directly to the get_search_query
        method; see that method's docstring for more information.

    @param rank_mode: this is a bit-masked integer with valid values being
        RANK_NO_NORMAL, RANK_LOG_LENGTH, RANK_LENGTH, RANK_HARMONIC_MEAN,
        RANK_UNIQUE_WORDS, RANK_LOG_UNIQUE_WORDS, and RANK_BY_RANK.
        Respectively, these ranking parameters represent ignoring the document
        length, dividing the rank by (1 + logarithm of the document length),
        dividing the rank by the document length, dividing the rank by the mean
        harmonic distance between extents, dividing the rank by the number of
        unique words in the document, dividing the rank by (1 + the log of the
        number of unique words in the document, dividing the rank by (rank +
        1).

    @return: returns a Storm ResultSet, an iterable of (article, rank, excerpt)
        tuples.
    """
    store = get_knowledge_store()
    text_search_query = get_search_query(search_terms, search_mode)
    rank = TSRankCD(Article.text_search_vector, text_search_query, rank_mode)
    excerpt = TSHeadline(TEXT_SEARCH_CONFIG, Article.body, text_search_query)
    where = [Match(text_search_query, Article.text_search_vector)]
    if not may_view_unpublished_articles():
        where.append(Article.published == True)
    columns = (Article, rank, excerpt)
    results = store.find(columns, *where)
    results.order_by(Desc(rank))
    return results
