# Copyright 2005-2022 Canonical Limited. All rights reserved.

from past.builtins import cmp
from functools import cmp_to_key, partial
from collections import defaultdict

from storm.locals import And, Desc, Or

from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy

from canonical.lib.timestamp import utctoday
from canonical.lib.text import make_secure_id
from canonical.security.protect import requires, requires_advisor, PUBLIC
from canonical.security.interfaces import IRoleRegistry

from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.account.alerts import (
    LicenseExpiryAlert,
    StandaloneOverflowAlert,
)
from canonical.landscape.model.main.license import (
    License,
    LicenseSeatError,
    ProLicenseType,
)


__all__ = [
    "ALL_LICENSE_TYPES",
    "COMPUTER_LICENSE_TYPES",
    "VM_LICENSE_TYPE",
    "CONTAINER_LICENSE_TYPE",
    "EXPIRABLE_LICENSE_TYPES",
    "LicenseManagement",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."

_not_set = object()

BASIC_LICENSE_TYPE = "BasicFeatures"
TRIAL_LICENSE_TYPE = "TrialFeatures"
VM_LICENSE_TYPE = "VMFeatures"
CONTAINER_LICENSE_TYPE = "ContainerFeatures"
UBUNTU_PRO_LICENSE_TYPE = "UbuntuProFeatures"
UBUNTU_PRO_FREE_LICENSE_TYPE = "UbuntuProFreeFeatures"

COMPUTER_LICENSE_TYPES = (BASIC_LICENSE_TYPE, TRIAL_LICENSE_TYPE)
ALL_LICENSE_TYPES = COMPUTER_LICENSE_TYPES + (
    VM_LICENSE_TYPE,
    CONTAINER_LICENSE_TYPE,
    UBUNTU_PRO_LICENSE_TYPE,
    UBUNTU_PRO_FREE_LICENSE_TYPE,
)
EXPIRABLE_LICENSE_TYPES = (
    BASIC_LICENSE_TYPE,
    VM_LICENSE_TYPE,
    CONTAINER_LICENSE_TYPE,
)
PRO_LICENSE_TYPES = (UBUNTU_PRO_LICENSE_TYPE, UBUNTU_PRO_FREE_LICENSE_TYPE)


@requires_advisor
class LicenseManagement(ModelManagement):

    # A list of tuples with license type name a list of associated license
    # roles. The list is ordered by decreasing license type.
    LICENSE_TYPES = (
        ("physical", COMPUTER_LICENSE_TYPES),
        ("vm", [VM_LICENSE_TYPE]),
        ("container", [CONTAINER_LICENSE_TYPE]),
    )

    def __init__(self, account, person=None):
        super(LicenseManagement, self).__init__(account, person)
        self._role_registry = getUtility(IRoleRegistry)

    @requires("CreateLicense")
    def create_license(
        self,
        role_name,
        seats,
        expiration_date=None,
        license_key=_not_set,
        salesforce_asset=None,
    ):
        """
        Create a new C{role_name} L{License} with C{seats} seats valid
        until C{expiration_date}.
        """
        role = self._role_registry.get(role_name)
        assert role.key, "License roles must have keys!"
        if license_key is _not_set:
            license_key = make_secure_id()
        license = License(
            self.account,
            role.key,
            seats,
            license_key=license_key,
            salesforce_asset_key=salesforce_asset,
        )
        if expiration_date:
            license.expiration_date = expiration_date

        self.account.check_rules([LicenseExpiryAlert])
        return license

    @requires("CreateLicense")
    def edit_license(
        self,
        license,
        expiration_date,
        salesforce_asset,
    ):
        """
        Update the expiration date and Salesforce asset key of a license.
        """
        license.expiration_date = expiration_date
        license.salesforce_asset_key = salesforce_asset
        return license

    @requires("CreateLicense")
    def find_license(
        self,
        role_name,
        seats,
        salesforce_asset,
    ):
        """
        Find a license for the account matching `role_key`, `seats`,
        and `salesforce_asset` and return it.
        If the account has no license matching all three parameters,
        try to find one that has no `salesforce_asset` and matches both
        `role_key` and `seats` with latest expiration date and return it.
        Failing that, return `None`.
        """
        assert salesforce_asset, "A Salesforce asset key is required!"
        role = self._role_registry.get(role_name)
        assert role.key, "License roles must have keys!"
        # Search by Salesforce asset key and role first
        results = self.account.licenses.find(
            License.role_key == role.key,
            License.seats == seats,
            License.salesforce_asset_key == salesforce_asset,
        )
        license = results.first() if results else None
        if license:
            return license

        # Then search by seats, role, and expiration date for a license,
        # without any Salesforce asset key.
        where = [
            License.role_key == role.key,
            License.seats == seats,
            Or(
                License.salesforce_asset_key == None,
                License.salesforce_asset_key == "",
            ),
        ]

        # None gets sorted last when Ascending. We want it first.
        results = self.account.licenses.find(*where).order_by(Desc("expiration_date"))
        license = results.first() if results else None
        return license

    @requires("ViewAccount")
    def _license_compare(self, license1, license2, role_keys=None):
        """C{cmp} operator to sort licenses.

        Licenses are sorted according to the following criteria:

        - licenses with available seats first
        - if C{role_keys} is provided, licenses are sorted based on their
          C{role_key} according to that order
        - licenses are sorted by descending expiration date
        """
        by_seats = cmp(
            license2.get_available_seats() > 0, license1.get_available_seats() > 0
        )
        if by_seats:
            return by_seats

        if role_keys is not None:
            by_role = cmp(
                role_keys.index(license1.role_key), role_keys.index(license2.role_key)
            )
            if by_role:
                return by_role

        expiration1 = license1.expiration_date
        expiration2 = license2.expiration_date
        if all((expiration1, expiration2)):
            return cmp(expiration2, expiration1)
        if not any((expiration1, expiration2)):
            return 0
        return 1 if expiration1 else -1

    def get_licenses(
        self,
        role_names=None,
        expire_before=None,
        active_only=False,
        available_only=False,
        create_pro_license=False,
    ):
        """Filter account licenses based on the specified criteria.

        @param role_names: A list of names for license roles to include.
        @param expire_before: Only include licenses with expiration date before
            the specified one.
        @param active_only: If C{True}, don't include expired licenses.
        @param available_only: If C{True}, don't include licenses with no
            available seats.
        """
        if role_names and create_pro_license:
            for license_type in PRO_LICENSE_TYPES:
                if license_type in role_names:
                    return self.get_or_create_ubuntu_pro_license(license_type)

        role_keys = self._get_role_keys(role_names) if role_names else None

        where = []
        if role_keys is not None:
            where.append(License.role_key.is_in(role_keys))
        if expire_before is not None:
            where.append(License.expiration_date < expire_before)
        if active_only:
            where.append(
                Or(
                    (License.expiration_date == None),
                    (License.expiration_date >= utctoday()),
                )
            )
        result = self.account.licenses.find(*where)
        if available_only:
            result = [
                license for license in result if license.get_available_seats() > 0
            ]

        key = cmp_to_key(partial(self._license_compare, role_keys=role_keys))
        return sorted(result, key=key)

    def get_or_create_ubuntu_pro_license(self, license_type):
        """
        Returns a license for a computer with an Ubuntu Pro
        subscription, creating one for the account if one does not
        exist.

        @returns: A storm result set containing the Ubuntu Pro license.
        """
        role_key = self._role_registry.get(license_type).key
        result = self.account.licenses.find(License.role_key == role_key)

        if result.is_empty():
            self.create_license(license_type, 1)
            result = self.account.licenses.find(License.role_key == role_key)

        return result

    @requires("ViewAccount")
    def get_license_by_id(self, license_id):
        return self.account.licenses.find(License.id == license_id).one()

    def _get_role_keys(self, role_names):
        """Return keys for specified role names."""
        return [self._role_registry.get(name).key for name in role_names]

    def _add_computer_to_license(self, computer, licenses):
        """
        Attempt to attach a computer to a L{License} from a sequence of
        licenses.

        @param computer: The computer to add to the L{License}
        @param licenses: A sequence of L{License} objects

        @return: True if the computer was added to a License.
        """
        current_licenses = set(computer.licenses)

        for license in licenses:
            if license in current_licenses:
                continue
            try:
                license.add_computer(computer)
            except LicenseSeatError:
                continue
            return True
        return False

    @requires(PUBLIC)
    def auto_assign_license_to_computer(self, computer):
        """
        Attempt to assign a computer to a random valid license:
        - If the computer is normal computer (vm_info not set), assign an
          available normal seat, if any.
        - If the computer is a VM (vm_info is set), this method will first try
          to assign it to an available VM seat, and fallback to a normal seat
          in case no VM seat can be found.
        - N.B. if the computer's Ubuntu Pro info indicates it should
          receive a license, it will, even if there are none available.

        This requires PUBLIC since the computer is used to create a principal
        in the message server (when receiving a new registration request).
        Since the computer doesn't already have a license attached to it, it
        is granted no role (computer roles are granted based on licenses,
        precisely).
        """
        # Remove the security proxy as the computer could be unlicensed and
        # would get Unauthorized on calls to License methods.
        current_license = removeSecurityProxy(computer.licenses.one())
        subscription_type = removeSecurityProxy(computer.ubuntu_pro_subscription_type())

        if subscription_type is ProLicenseType.FULL:
            role_names = (UBUNTU_PRO_LICENSE_TYPE,)
        elif subscription_type is ProLicenseType.FREE:
            role_names = (UBUNTU_PRO_FREE_LICENSE_TYPE,)
        elif computer.container_info:
            role_names = (
                CONTAINER_LICENSE_TYPE,
                VM_LICENSE_TYPE,
            ) + COMPUTER_LICENSE_TYPES
        elif computer.vm_info:
            role_names = (VM_LICENSE_TYPE,) + COMPUTER_LICENSE_TYPES
        else:
            role_names = COMPUTER_LICENSE_TYPES

        available_licenses = self.get_licenses(
            role_names=role_names,
            active_only=True,
            available_only=True,
            create_pro_license=True,
        )
        found_license = self._add_computer_to_license(computer, available_licenses)
        if found_license:
            if current_license is not None:
                current_license.remove_computer(computer)

            rules = [LicenseExpiryAlert]
            if is_standalone():
                rules.append(StandaloneOverflowAlert)
            self.account.check_rules(rules)

        return found_license

    @requires("ViewAccount")
    def get_seats(self):
        """
        Return a dict with available and used seats for each license type, and
        totals.
        """
        license_roles = {
            label: self._get_role_keys(types) for label, types in self.LICENSE_TYPES
        }

        licenses_by_role = defaultdict(list)
        for license in self.get_licenses():
            licenses_by_role[license.role_key].append(license)

        seats = {}
        total_used_count = 0
        total_available_count = 0
        for label, roles in list(license_roles.items()):
            used_count = 0
            available_count = 0
            for role in roles:
                for license in licenses_by_role[role]:
                    used_count += license.get_used_seats()
                    available_count += license.get_available_seats()

            total_used_count += used_count
            total_available_count += available_count
            seats.update(
                {"used-" + label: used_count, "available-" + label: available_count}
            )

        seats.update(
            {"used-total": total_used_count, "available-total": total_available_count}
        )
        return seats

    def check_seats(self, seats_by_type):
        """Check if the specified number of seats is available.

        This takes into account falling back to greater license types
        (if available) when a lesser type doesn't cover requirements.

        @param seats_by_type: a dict with required license seats by kind.
        @return: a dict with missing seats count per kind.
        """
        kinds = [name for name, _ in self.LICENSE_TYPES]
        missing_seats = dict.fromkeys(kinds, 0)
        seats = self.get_seats()

        carry = 0
        for kind in kinds:
            required = seats_by_type.get(kind, 0)
            available = seats["available-" + kind]
            diff = available + carry - required
            if diff > 0:
                carry = diff
            else:
                missing_seats[kind] = -diff

        return missing_seats
