# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer
from zope.cachedescriptors.property import Lazy

from canonical.lib.text import slug_from_title

from storm.locals import Store, Int, Unicode, Reference, ReferenceSet, AutoReload

from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE
from canonical.collection.collection import CollectionBase

from canonical.landscape.model.main.store import get_main_store

from canonical.landscape.model.main.interfaces import (
    IAccessContext,
    IAccountAware,
    ICollectionContext,
)

__all__ = [
    "AccessContext",
    "AccountAccessContexts",
    "create_root_access_context",
    "get_unique_access_context_name",
    "get_access_context",
    "AccessContextCollection",
]

__clearlist__ = ["AccountAccessContexts", "AccessContextCollection"]

__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IAccessContext, ICollectionContext)
class AccessContext(object):
    """Access context for security-related checks.

    Access contexts are organized in hierarchies. Each L{Account} holds
    a reference to a root L{AccessContext}, and all its descendants are
    linked to the root.

    @ivar id: Database identifier.
    @ivar name: Human-readable identifier.
    @ivar description: Optional description of the L{AccessContext}.
    @ivar parent: The parent L{AccessContext} or C{None} for roots.
    @ivar root: The root L{AccessContext} (it points to the access
        context itself in case of roots).
    """

    __storm_table__ = "access_context"

    requires(
        "ViewAccessContext",
        PRIVATE,
        [
            "id",
            "access_context_id",
            "is_root",
            "computers",
            "parent",
            "account",
            "title",
        ],
    )
    requires(
        "ViewAccessContext",
        "EditAccessContext",
        ["name", "description", "children", "parent_id"],
    )

    collection_filter = "root"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    description = Unicode()
    parent_id = Int()
    root_id = Int(allow_none=False, default=AutoReload)

    parent = Reference(parent_id, id)
    root = Reference(root_id, id)
    children = ReferenceSet(id, parent_id, order_by=name)

    def __init__(self, name, description=None, parent=None):
        self.name = name
        self.description = description
        self.parent = parent
        if parent is not None:
            # If there's a parent, then inherit its root. Otherwise default
            # to our own ID as root_id (see the database schema).
            self.root = parent.root

    @property
    def access_context(self):
        """The L{AccessContext} to use for security checks.

        When checking if a person has a certain permission on this
        L{AccessContext}, the L{AccessContext} to check against is ourselves.
        """
        return self

    @property
    def store(self):
        """Return the Storm Store this access context lives in."""
        return Store.of(self)

    @property
    def is_root(self):
        """Whether this L{AccessContext} is a root one."""
        return self.parent is None

    @Lazy  # This isn't a Storm model, so this is OK.  # noqa
    def account(self):
        from canonical.landscape.model.main.account import Account

        result = self.store.find(Account, Account.access_context_id == self.root_id)
        return result.one()

    @requires("EditAccessContext")
    def add_role(self, person, role):
        person.reset_cache()
        from canonical.landscape.model.main.personaccess import PersonAccess

        PersonAccess(person, role, self)

    @requires("ViewAccessContext")
    def get_role_map(self):
        """
        Returns a dictionary mapping {person: set([roles])} for this access
        context, with all persons with a role on this access context.
        """
        # XXX this method is mainly used by the old RBAC UI and should be
        # eventually dropped
        from canonical.landscape.model.main.personaccess import PersonAccess

        store = Store.of(self)
        mapping = {}
        for person_access in store.find(
            PersonAccess, PersonAccess.access_context_id == self.id
        ):
            # use the person-specific role registry here
            registry = person_access.person.get_role_registry()
            role = registry.get(person_access.role_key)
            person_roles = mapping.get(person_access.person)
            if person_roles is None:
                mapping[person_access.person] = set((role,))
            else:
                person_roles.add(role)
        return mapping

    @requires("EditAccessContext")
    def remove_role(self, person, role):
        person.reset_cache()
        from canonical.landscape.model.main.personaccess import PersonAccess

        Store.of(self).find(
            PersonAccess,
            PersonAccess.access_context_id == self.id,
            PersonAccess.person_id == person.id,
            PersonAccess.role_key == role.key,
        ).remove()

    @requires("EditAccessContext")
    def remove_roles(self, person, roles):
        person.reset_cache()
        from canonical.landscape.model.main.personaccess import PersonAccess

        role_keys = [role.key for role in roles if role.key is not None]
        Store.of(self).find(
            PersonAccess,
            PersonAccess.access_context_id == self.id,
            PersonAccess.person_id == person.id,
            PersonAccess.role_key.is_in(role_keys),
        ).remove()

    def remove_persons(self, role, persons):
        """Revoke the given role on the given persons."""
        from canonical.landscape.model.main.personaccess import PersonAccess

        person_ids = [person.id for person in persons]
        Store.of(self).find(
            PersonAccess,
            PersonAccess.access_context_id == self.id,
            PersonAccess.person_id.is_in(person_ids),
            PersonAccess.role_key == role.key,
        ).remove()

    @requires("EditAccessContext")
    def set_roles(self, person, roles):
        """
        Set the C{role}s for C{person} on this access context to be the list in
        roles.
        """
        from canonical.landscape.model.main.personaccess import PersonAccess

        person.reset_cache()
        Store.of(self).find(
            PersonAccess,
            PersonAccess.access_context_id == self.id,
            PersonAccess.person_id == person.id,
        ).remove()
        for role in roles:
            PersonAccess(person, role, self)

    @requires("CreateAccessContext")
    def create_child(self, name, description=None):
        """Create a child L{AccessContext}.

        @param name: Human-readable identifier for the access context.
        @param description: Optional description of the access context.
        @return: The new L{AccessContext}.
        """
        access_context = AccessContext(name, description=description, parent=self)
        store = get_main_store()
        store.add(access_context)
        return access_context

    @requires("ViewAccount")
    def has_children(self):
        """Return a boolean indicating if the access context has children."""
        return self.children.count() > 0

    @requires("ViewAccessContext")
    def get_child_by_id(self, id):
        """Return a child L{AccessContext} with the given id."""
        store = Store.of(self)
        return store.find(
            AccessContext, AccessContext.parent_id == self.id, AccessContext.id == id
        ).one()

    @requires("ViewAccessContext")
    def get_state(self, v2=False):
        """Return a dict with information on the L{AccessContext}.

        @param v2: Whether to use the new version of the L{AccessContext}
            representation.
        """
        if not v2:
            # XXX To be dropped when the API version 2011-08-01 dies
            children = [child.name for child in self.children]
            parent_name = ""
            if self.parent:
                parent_name = self.parent.name
            return {
                "name": self.name,
                "title": self.description,
                "parent": parent_name,
                "children": ",".join(children),
            }

        parent = None
        if self.parent is not None:
            parent = {"id": self.parent.id, "name": self.parent.name}

        children = []
        for child in self.children:
            children.append({"id": child.id, "name": child.name})

        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "parent": parent,
            "children": children,
        }

    @requires("ViewAccount")
    def get_all_access_context_ids(self):
        """
        Helper method returning all the L{AccessContext} IDs related to this
        access context, ie itself and all its children.
        """
        access_context_ids = set([self.id])
        for child in self.children:
            access_context_ids.update(child.get_all_access_context_ids())
        return access_context_ids

    def get_ancestor_access_context_ids(self):
        """
        Return ids for all the ancestor L{AccessContext}s of this
        L{AccessContext}.
        """
        access_context_ids = []
        access_context = self
        while access_context.parent_id is not None:
            access_context_ids.append(access_context.parent_id)
            access_context = access_context.parent
        return access_context_ids

    @requires("ViewAccessContext")
    def get_roles(self):
        """
        Return a list of L{StoredRole}s the L{AccessContext} is linked to,
        either directly or through the hierarchy.
        """
        from canonical.landscape.model.access.role import (
            StoredRole,
            GlobalAdmin,
            StoredRoleAccessContext,
        )

        access_context_ids = self.get_ancestor_access_context_ids()
        access_context_ids.append(self.id)

        result = Store.of(self).find(
            StoredRole,
            StoredRoleAccessContext.role_id == StoredRole.id,
            StoredRoleAccessContext.access_context_id.is_in(access_context_ids),
        )
        roles = list(result.config(distinct=True))
        # Also add the GlobalAdmin role, which is always present.
        roles.append(GlobalAdmin(self.account))
        return roles

    @property
    def title(self):
        # XXX for backward compatibility, will be dropped after the
        # migration to name/description completes
        return self.description


@requires_advisor
@implementer(IAccountAware)
class AccountAccessContexts(object):
    """Routing hook for access context management functionality."""

    requires(PUBLIC, PRIVATE, ["account"])

    def __init__(self, account):
        self.account = account


class AccessContextCollection(CollectionBase):
    """Return result sets of AccessContexts matching given criteria."""

    def __init__(
        self, store=None, root=None, ids=None, names=None, access_context_ids=None
    ):
        if not store:
            store = get_main_store()
        super(AccessContextCollection, self).__init__(
            store,
            root=root,
            ids=ids,
            names=names,
            access_context_ids=access_context_ids,
        )

    @property
    def access_context(self):
        """Delegate security checks on the root access context."""
        return self._where["root"]

    def with_root(self, root):
        """Return AccessContexts that are rooted at this access context."""
        return self._filter_by(root=root)

    def with_ids(self, ids):
        """Return AccessContexts having the given ids."""
        return self._filter_by(ids=ids)

    def with_access_context_ids(self, access_context_ids):
        """Restrict the search to entities linked to these access contexts.

        This method is used to limit the set of returned AccessContext entities
        based on the access contexts a person has access to. However since the
        access context of an AccessContext is the AccessContext itself, this
        is actually just filters previously set IDs furhter, or set new a new
        filter based on the given IDs.

        The use case is a user performing a search for access contexts matching
        given IDs, but we want to return only the ones he has access to.
        """
        return self._filter_by(access_context_ids=access_context_ids)

    def with_names(self, names):
        """Return AccessContexts having the given names."""
        return self._filter_by(names=names)

    def _get_result(
        self,
        store,
        where=None,
        root=None,
        ids=None,
        names=None,
        access_context_ids=None,
    ):
        """Get a ResultSet of AccessContexts matching the search criteria."""
        if where is None:
            where = []

        if root is not None:
            where.append(AccessContext.root_id == root.id)

        filter_ids = None
        if ids and access_context_ids:
            # If we filter both by IDs and access context IDs, take the
            # intersection of the two.
            filter_ids = set(ids) & set(access_context_ids)
        elif ids:
            # Filter only by IDs
            filter_ids = ids
        elif access_context_ids:
            # Filter only by access context IDs
            filter_ids = access_context_ids
        if filter_ids:
            where.append(AccessContext.id.is_in(filter_ids))

        if names:
            where.append(AccessContext.name.is_in(names))

        return store.find(AccessContext, *where)


def create_root_access_context(name, description=None):
    """Create a new L{AccessContext}.

    @param name: The name of the L{AccessContext}.
    @param description: Optional description of the L{AccessContext}.
    """
    access_context = AccessContext(name, description=description)
    get_main_store().add(access_context)
    return access_context


def get_unique_access_context_name(root_access_context, description):
    """
    Generate a unique (within the account) access context name that can be used
    as a slug.
    """
    collection = AccessContextCollection(root=root_access_context)
    result = collection.find()
    values = list(result.values(AccessContext.name, AccessContext.description))
    names, descriptions = list(zip(*values))

    if description in descriptions:
        return None

    name = slug_from_title(description, 55, lambda slug: names)
    return name


def get_access_context(root, key):
    """Return the L{AccessContext} with the given key.

    @param root: The L{AccessContext} under which the lookup is performed.
    @param key: Either an L{AccessContext.id} or L{AccessContext.name}.
    """
    if isinstance(key, int):
        clause = AccessContext.id == key
    else:
        clause = AccessContext.name == key
    store = get_main_store()
    result = store.find(AccessContext, AccessContext.root_id == root.id, clause)
    return result.one()
