# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Core L{Account} object and account management functions.
"""
import shlex
import string
from datetime import datetime, timedelta

from zope.interface import implementer
from zope.component import getUtility
from zope.event import notify
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.security.interfaces import Unauthorized

from storm.locals import (
    Storm,
    Store,
    AutoReload,
    Int,
    Unicode,
    Reference,
    ReferenceSet,
    Select,
    Like,
    And,
    Or,
    DateTime,
    Bool,
    SQL,
    Not,
    In,
)
from storm.zope.interfaces import IZStorm

from canonical.lib import random_string
from canonical.lib.timestamp import utctoday, format_datetime
from canonical.lib.text import is_valid_slug, slug_from_title, make_secure_id
from canonical.security.protect import requires, PUBLIC, PRIVATE, requires_advisor
from canonical.security.interfaces import IRoleRegistry
from canonical.amqp.message.interfaces import IDispatcher

from canonical.landscape.model.account.collection import AccountCollection
from canonical.landscape.model.main.interfaces import (
    IAccount,
    IAccessContextAware,
    IAccountAware,
    ICollectionContext,
)
from canonical.landscape.model.main.errors import (
    UnknownPersonsError,
    UnknownAccessContextsError,
    UnknownAccessContextError,
)
from canonical.landscape.model.main.accesscontext import (
    AccessContext,
    create_root_access_context,
)
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.lowercaseunicode import LowercaseUnicode
from canonical.landscape.model.events import AddComputerEvent
from canonical.landscape.model.store import StoreHolder
from canonical.landscape.security.roles import STORED_ROLE_KEY_MIN
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.activity.group import ActivityGroup
from canonical.landscape.model.configuration import (
    is_staging_environment,
    is_standalone,
)


LENGTH_OF_TRIAL_IN_DAYS = 30

__all__ = [
    "Account",
    "PersonAccount",
    "TrialAccount",
    "get_account_by_name",
    "get_all_accounts",
    "get_accounts_by_ids",
    "get_account_by_id",
    "get_accounts_by_id_range",
    "get_account_by_sfid",
    "get_account_by_salesforce_account_key",
    "DEFAULT_STANDALONE_ACCOUNT_NAME",
    "DEFAULT_STANDALONE_ACCOUNT_MAX_PEOPLE_COUNT",
    "DEFAULT_ACCOUNT_MAX_PEOPLE_COUNT",
    "DEFAULT_ACCESS_CONTEXT_NAME",
    "DEFAULT_ACCESS_CONTEXT_TITLE",
    "DEFAULT_ACCOUNT_MAX_PEOPLE_COUNT",
    "ContextParameterError",
    "generate_account_name",
    "generate_random_account_name",
    "create_account",
    "create_trial_account",
    "get_trial_accounts",
    "is_valid_account_name",
    "RemovedAccountName",
    "LENGTH_OF_TRIAL_IN_DAYS",
    "InvalidAccessContext",
    "PersonNotMemberOfAccountError",
    "get_trial_account_by_secure_id",
    "can_have_trial",
    "remove_account",
    "bootstrap_account",
]

__clearlist__ = [
    "Account",
    "PersonNotMemberOfAccountError",
    "can_have_trial",
    "DEFAULT_ACCESS_CONTEXT_NAME",
    "get_account_by_name",
    "get_account_by_id",
    "create_trial_account",
    "DEFAULT_STANDALONE_ACCOUNT_NAME",
    "generate_account_name",
    "generate_random_account_name",
    "get_account_by_salesforce_account_key",
]


__requires_prefix__ = "landscape."


DEFAULT_ACCESS_CONTEXT_NAME = "global"
DEFAULT_ACCESS_CONTEXT_TITLE = "Global access"
DEFAULT_STANDALONE_ACCOUNT_NAME = "standalone"
DEFAULT_STANDALONE_ACCOUNT_MAX_PEOPLE_COUNT = 1000
DEFAULT_ACCOUNT_MAX_PEOPLE_COUNT = 5


class InvalidAccessContext(Exception):
    """Raised when the access context ID passed to invite is invalid.

    @see: L{Account.invite}
    """


class MissingAccessContextError(Exception):
    """Raised when activating an invitation for a missing access context.

    @see: L{AccountInvitation.activate}
    """


class ContextParameterError(Exception):
    """
    Raised when trying to call a function with an invalid context parameter.

    @see: L{get_account_access_contexts_vocabulary}
    """


class PersonNotMemberOfAccountError(Exception):
    """
    Raised when the person is not a member of the target account.

    @see: L{Account.disable_membership}
    """


class PersonAccount(object):
    """A link object to associate a L{Person} with an L{Account}.

    @cvar __storm_table__: Person/account link objects are stored in the
        C{person_account} table.
    @cvar __storm_primary__: The compund primary key is made up of unique
        (C{person_id}, C{account_id}) tuples.
    @ivar person_id: The L{Person.id<Person>} to link to.
    @type person_id: Int
    @ivar account_id: The C{Account.id} to link to.
    @type account_id: Int
    """

    __storm_table__ = "person_account"
    __storm_primary__ = "person_id", "account_id"

    person_id = Int(allow_none=False)
    account_id = Int(allow_none=False)


@requires_advisor
@implementer(IAccount, IAccessContextAware, ICollectionContext)
class Account(StoreHolder, Storm):
    """An account contains L{Computer}s and the data needed to manage them.

    Account names may not have spaces or punctuation.  They should be made up
    of C{[A-Z,a-z,0-9]} and C{-} (dash).

    @cvar __storm_table__: Account objects are stored in the C{account} table.
    @cvar default_resource_database_key: A key for the default resource
        database to use with new L{Computer}s.
    @type default_resource_database_key: int
    @cvar resource_lifespan: The number of days to keep monitoring data such
        as L{LoadAverage}s, L{Temperature}s, etc.
    @type resource_lifespan: int
    @ivar id: The primary key, automatically generated by the database, for
        this account instance.
    @type id: Int
    @ivar access_context_id: The L{AccessContext.id<AccessContext>} for this
        account.
    @type access_context_id: Int
    @ivar database_key: A key for the account database to use with this
        account.
    @type database_key: Int
    @ivar name: The user-supplied name used to identify this account in URLs.
    @type name: LowercaseUnicode
    @ivar title: The user-supplied title used to represent this account in the
        user interface.
    @type title: Unicode
    @ivar registration_password: The registration password for this account or
        C{None} if there is no password, the default.
    @type registration_password: Unicode
    @ivar max_people_count: The maximum number of L{Person}s to allow in this
        account.  The default is 2.
    @type max_people_count: Int
    @ivar access_context: A C{Reference} to the L{AccessContext} for this
        account.
    @type access_context: AccessContext
    @ivar creation_time: The time when the account has been created.
    @type creation_time: C{datetime}
    @ivar disabled_reason: The reason the account was disabled
        (None if enabled).
    """

    __storm_table__ = "account"

    requires(
        "ViewAccount",
        PRIVATE,
        [
            "id",
            "people",
            "computers",
            "licenses",
            "computer_group",
            "access_context_id",
            "access_context",
            "name",
            "namespace",
            "creation_time",
            "max_attachment_files",
            "managed_hours_start_time",
        ],
    )
    requires(
        "ViewAccount",
        "EditAccount",
        [
            "title",
            "registration_password",
            "archive_token",
            "registration_password_escaped",
        ],
    )
    requires("ViewAccount", "EditMaxPeopleCount", ["max_people_count"])
    requires("ViewAccount", "EditMaxAttachmentSize", ["max_attachment_size"])
    requires("ViewSalesforceKey", "EditSalesforceKey", ["salesforce_account_key"])

    default_resource_database_key = 1
    resource_lifespan = 28
    collection_filter = "account"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    access_context_id = Int(allow_none=False)
    database_key = Int(default=1)
    name = LowercaseUnicode(allow_none=False)
    title = Unicode(allow_none=False)
    registration_password = Unicode()
    max_people_count = Int(allow_none=False)
    max_attachment_size = Int(allow_none=False, default=AutoReload)
    max_attachment_files = 5  # This will possibly be parametrizable later
    creation_time = DateTime(default=AutoReload)
    managed_hours_start_time = DateTime(default=AutoReload)
    disabled_reason = Unicode(allow_none=True)

    salesforce_account_key = Unicode(allow_none=True)
    stripe_customer_id = Unicode(allow_none=True)
    is_business = Bool(allow_none=True)
    vat_number = Unicode(allow_none=True)
    vat_country = Unicode(allow_none=True)
    marketing_emails = Bool(allow_none=False)

    access_context = Reference(access_context_id, AccessContext.id)
    people = ReferenceSet(
        id,
        PersonAccount.account_id,
        PersonAccount.person_id,
        Person.id,
        order_by=Person.name,
    )
    trial_account = Reference(id, "TrialAccount.account_id")

    _computer_group = None
    _computer_roles_cache = None

    def __init__(
        self,
        name,
        title,
        registration_password=None,
        database_key=None,
        max_people_count=None,
        salesforce_account_key=None,
        stripe_customer_id=None,
        is_business=None,
        vat_number=None,
        vat_country=None,
        marketing_emails=False,
    ):
        """
        @param name: Name for this account.
        @type name: str

        @param title: Title for this account.
        @type title: str

        @param registration_password: Registration password for this account.
        @type registration_password: str

        @param database_key: resource database key to use for this account.
        @type database_key: int

        @param salesforce_account_key: Salesforce ID for this account.
        @type salesforce_account_key: str

        @param max_people_count: Maximum number of administrators that
        can be added to this Account
        @type max_people_count: int

        @param stripe_customer_id: Stripe Customer ID for this account.
        @type stripe_customer_id: str

        @param is_business: Is this account for a business or a consumer?
        @type is_business: bool

        @param vat_number: VAT number, in case this is a business account.
        @type vat_number: str

        @param vat_country: VAT country, in case this is a business account.
        @type vat_country: str

        @param marketing_emails: whether the owner of the account
        agreed to receive marketing emails
        @type marketing_emails: bool
        """
        self.access_context = create_root_access_context(
            DEFAULT_ACCESS_CONTEXT_NAME, DEFAULT_ACCESS_CONTEXT_TITLE
        )
        self.name = name
        self.title = title
        self.registration_password = registration_password
        if database_key is not None:
            self.database_key = database_key
        self.max_people_count = max_people_count or DEFAULT_ACCOUNT_MAX_PEOPLE_COUNT
        self.salesforce_account_key = salesforce_account_key
        self.stripe_customer_id = stripe_customer_id
        self.is_business = is_business
        self.vat_number = vat_number
        self.vat_country = vat_country
        self.marketing_emails = marketing_emails

    @property
    def registration_password_escaped(self):
        key = self.registration_password
        if key is not None:
            return shlex.quote(self.registration_password)
        else:
            return ""

    @property
    def namespace(self):
        return "landscape.account.%d" % self.id

    @property
    def computer_group(self):
        """Get the L{AccountComputerGroup} for this account."""
        if self._computer_group is None:
            from canonical.landscape.model.account.computergroup import (
                AccountComputerGroup,
            )

            self._computer_group = AccountComputerGroup(self)
        return self._computer_group

    def reset_cache(self):
        """Reset the cache."""
        self._computer_group = None
        self._computer_roles_cache = None

    __storm_invalidated__ = reset_cache

    @property
    @requires(PUBLIC)
    def is_self_service(self):
        """Is this account self-service?

        Self-service accounts use utility billing, on a pay-as-you-go
        basis. They are only applicable on Landscape SAAS and don't
        use the license machinery.

        These accounts are charged by Landscape via Stripe, and don't
        use TrialAccounts.

        Landscape hosted will have some accounts using the existing UA
        machinery (and integrate tightly with SalesForce) and some
        which are self-service.
        """
        return self.stripe_customer_id is not None

    def get_store_name(self):
        """Get the C{IZStorm} name of the account database for this account."""
        return "account-%d" % self.database_key

    def get_state(self):
        """Get a C{dict} with information about the L{Account}."""
        last_login_time = None
        administrators = []
        trial = self.trial_account

        for person in self.people.order_by(Person.name, Person.email):
            sf_user = self.get_salesforce_user(person)
            administrators.append(
                {
                    "name": person.name,
                    "email": person.email,
                    "openid": person.identity,
                    "salesforce_contact_id": (
                        sf_user.salesforce_key if sf_user else None
                    ),
                }
            )
            if person.last_login_time:
                if not last_login_time:
                    last_login_time = person.last_login_time
                elif person.last_login_time > last_login_time:
                    last_login_time = person.last_login_time

        last_login_time = format_datetime(last_login_time)

        license_data = []
        has_trial_license = False
        role_registry = getUtility(IRoleRegistry)
        trial_role = role_registry.get("TrialFeatures")
        lds_role = role_registry.get("LDSBasicFeatures")
        lds_vm_role = role_registry.get("LDSVMFeatures")
        lds_container_role = role_registry.get("LDSContainerFeatures")
        vm_role = role_registry.get("VMFeatures")
        container_role = role_registry.get("ContainerFeatures")
        basic_role = role_registry.get("BasicFeatures")
        ubuntu_pro_role = role_registry.get("UbuntuProFeatures")

        role_name_mapping = {
            trial_role.key: "Trial",
            lds_role.key: "LDSBasic",
            lds_vm_role.key: "LDSVirtual",
            lds_container_role.key: "LDSContainer",
            vm_role.key: "Virtual",
            container_role.key: "Container",
            basic_role.key: "Basic",
            ubuntu_pro_role.key: "UbuntuPro",
        }

        from canonical.landscape.model.main.license import License

        for license in self.licenses.order_by(License.expiration_date):
            if license.role_key == trial_role.key:
                has_trial_license = True
            license_data.append(
                {
                    "expires": format_datetime(license.expiration_date),
                    "seats": license.seats,
                    "type": role_name_mapping[license.role_key],
                }
            )

        data = {
            "account": self.name,
            "trial_account": len(license_data) == 1 and has_trial_license,
            "creation_time": format_datetime(self.creation_time),
            "creator": None,
            "administrators": administrators,
            "disabled": False,
            "computers": self.get_computers().count(),
            "company": self.title,
            "company_size": None,
            "daytime_phone": None,
            "country": None,
            "region": None,
            "last_login_time": last_login_time,
            "licenses": license_data,
        }
        if self.disabled_reason is not None:
            data["disabled"] = True
            data["disabled_reason"] = self.disabled_reason

        from canonical.landscape.model.main.standalone import AccountLDSLicenseAPI

        api = AccountLDSLicenseAPI(self)
        data["lds_enabled"] = True if api.get_license() else False
        data["creator"] = None if not trial else trial.creator.identity
        data["company_size"] = None if not trial else trial.company_size
        data["daytime_phone"] = None if not trial else trial.daytime_phone
        data["country"] = None if not trial else trial.country
        data["region"] = None if not trial else trial.region
        if trial:
            data["mailing_list_optin"] = trial.mailing_list_optin
        elif self.is_self_service:
            data["mailing_list_optin"] = self.marketing_emails
        else:
            data["mailing_list_optin"] = None
        data["city"] = None if not trial else trial.city
        data["salesforce_account_key"] = self.salesforce_account_key

        if trial:
            data["registration_comments"] = trial.registration_comments

        return data

    @requires(PRIVATE)
    def add_person(self, person, roles, access_context_id=None):
        """Make a L{Person} a member of this account.

        The C{RetiredMember} role, granted to a L{Person} that was a member of
        the account in the past, is removed from the account's access context
        when present.  If a person is added with a custom L{AccessContext}
        they will also be granted the C{AccountMember} role in the account's
        access context.

        @param person: The L{Person} to add to this account.
        @param roles: Optionally, the name of the roles to associate the person
            with.  The default role is C{Owner}. If the role refers to a custom
            L{StoredRole}, then the person will be granted C{AccountMember}
            status on this L{Account}'s global access context and will also be
            added as a member of the specified L{StoredRole}.
        @param access_context_id: Optionally, the
            L{AccessContext.id<AccessContext>} to add the person to.  The
            default access context is the account's. If a different access
            context is specified, the Person will be granted the
            C{AccountMember} role on the account's global access context in
            addition to being granted the specified C{role} on the given access
            context.
        """
        from canonical.landscape.model.access.role import get_role_registry

        registry = get_role_registry([self.id])  # Consider custom roles too
        account_member = registry.get("AccountMember")

        person_added = False
        for role_name in roles:
            role = registry.get(role_name)
            # XXX We should eventually have a better way to detect a stored
            # role, and possibly make this method work with stored roles only.
            if role.key >= STORED_ROLE_KEY_MIN:
                if not person_added:
                    # Add the person because stored_role.add_persons requires
                    # it.
                    self.people.add(person)
                    person_added = True
                    self.access_context.add_role(person, account_member)
                from canonical.landscape.model.access.management import (
                    AccessControlManagement,
                )

                management = AccessControlManagement(self)
                stored_role = management.get_role(role.name.split(".")[1])
                stored_role.add_persons([person.email])
            elif (
                access_context_id is None or access_context_id == self.access_context.id
            ):
                if not person_added:
                    self.people.add(person)
                    self.access_context.add_role(person, role)
                    person_added = True
            else:
                access_context = self.access_context.get_child_by_id(access_context_id)
                if access_context is None:
                    raise MissingAccessContextError()
                if not person_added:
                    access_context.add_role(person, role)
                    self.access_context.add_role(person, account_member)

                    # Add the person now, so in case of
                    # MissingAccessContextError we won't run this at all
                    # (for testing).
                    self.people.add(person)
                    person_added = True

        # If the person was previously in this account, it may have the
        # RetiredMember role.
        retired_member = registry.get("RetiredMember")
        if retired_member in person.get_all_roles()[self.access_context_id]:
            self.access_context.remove_role(person, retired_member)

    def get_persons(self, emails):
        """Return the L{Person}s in the account with the given emails.

        @raises UnknownPersonsError: If one or more emails have no associated
            persons in the account.
        """
        persons = list(self.people.find(Person.email.is_in(emails)))
        known_emails = [person.email for person in persons]
        unknown_emails = set(emails) - set(known_emails)
        if unknown_emails:
            raise UnknownPersonsError(sorted(unknown_emails), self)
        return persons

    def limit_membership(self, person, access_context=None, role="Owner"):
        """
        Limit the membership of the given person to the given role within
        the given access context.

        @note: This method is meant as a testing convenience to setup
            restricted admins, it shouldn't be used for real code.
        """
        if access_context is None:
            access_context = self.access_context

        registry = person.get_role_registry()
        roles = person.get_all_roles()
        self.access_context.remove_roles(person, roles[self.access_context_id])
        self.access_context.add_role(person, registry.get("AccountMember"))
        if role is not None:
            access_context.add_role(person, registry.get(role))

    @requires("DisableMembership")
    def disable_membership(self, person):
        """Remove a L{Person} from this account.

        All roles that the L{Person} has in the account's L{AccessContext}
        will be removed, and a C{RetiredMember} role will be added.  This will
        cause the C{Coworker} role to still be added by the C{PersonByPerson}
        security checker.  L{RuleSubscription}s, if any are present, for
        C{person} will be removed.

        @param person: The L{Person} to remove from this account.
        """
        from canonical.landscape.model.alert.rule import Rule, RuleSubscription

        registry = getUtility(IRoleRegistry)
        retired_member_role = registry.get("RetiredMember")

        all_roles = person.get_all_roles()
        roles = all_roles.get(self.access_context_id, {})

        if person not in self.people and retired_member_role not in roles:
            raise PersonNotMemberOfAccountError(
                "Person '%s' is not a member of account '%s'" % (person.name, self.name)
            )

        self.access_context.remove_roles(person, roles)

        # Remove stored roles for the person in the account
        from canonical.landscape.model.access.management import AccessControlManagement

        management = AccessControlManagement(self)
        management.remove_person_roles(person)

        # Grant the RetiredMember role before doing anything else, this
        # assure that the L{Person} who is calling this code will still
        # have the Coworker role on the L{Person} being deleted from the
        # account. See also the L{PersonByPerson} checker.
        self.access_context.add_role(person, retired_member_role)

        for child_access_context in self.get_access_contexts():
            if (
                child_access_context.id in all_roles
                and child_access_context != self.access_context
            ):
                roles = person.get_all_roles()[child_access_context.id]
                child_access_context.remove_roles(person, roles)

        # Remove the API credential for the L{Person} in this L{Account}.
        from canonical.landscape.model.main.person import APICredentials

        Store.of(self).find(
            APICredentials,
            APICredentials.person == person,
            APICredentials.account == self,
        ).remove()

        # It's important we do this (remove the person from the account) as the
        # last thing; code before this attempts to fetch the roles associated
        # with the Person being deleted (Person.get_all_roles), which might
        # blow up if the association between the Person and Account is broken.
        self.people.remove(person)

        salesforce_user = self.get_salesforce_user(person)
        if salesforce_user is not None:
            Store.of(self).remove(salesforce_user)

        subselect = Select(
            RuleSubscription.rule_id,
            where=[
                And(
                    RuleSubscription.person_id == person.id,
                    RuleSubscription.rule_id == Rule.id,
                    Rule.account_id == self.id,
                )
            ],
        )
        self.store.find(
            RuleSubscription,
            RuleSubscription.rule_id.is_in(subselect),
            RuleSubscription.person_id == person.id,
        ).remove()

    @requires("EditSalesforceKey")
    def create_salesforce_user(self, person, salesforce_key):
        """Create a L{SalesforceUser} for L{Person} using L{salesforce_key}."""
        from canonical.landscape.model.main.salesforceuser import SalesforceUser

        return SalesforceUser(self, person, salesforce_key)

    @requires("ViewAccount")
    def get_salesforce_user(self, person):
        """
        Get the L{SalesforceUser} for C{person}, or None if one isn't
        available.
        """
        from canonical.landscape.model.main.salesforceuser import SalesforceUser

        return (
            Store.of(self)
            .find(
                SalesforceUser,
                SalesforceUser.account_id == self.id,
                SalesforceUser.person_id == person.id,
            )
            .one()
        )

    @requires("ViewAccount")
    def preload_licenses(self, computer_ids):
        """Preload L{License}s for C{computer_ids}.

        The C{_computer_roles_cache} is updated with L{License} data related
        to C{computer_ids}.  This data is used when checkers enforce security
        permissions, when L{Computer} instances are used.  License data is
        only loaded if it isn't already in the cache.
        """
        from canonical.landscape.model.main.license import License
        from canonical.landscape.model.main.computer import ComputerLicense

        if self._computer_roles_cache is not None:
            existing_computer_ids = set(self._computer_roles_cache.keys())
            requested_computer_ids = set(computer_ids)
            computer_ids = requested_computer_ids - existing_computer_ids
        if not computer_ids:
            return

        computer_license_index = {}
        result = Store.of(self).find(
            (ComputerLicense.computer_id, License),
            License.account_id == self.id,
            ComputerLicense.license_id == License.id,
            ComputerLicense.computer_id.is_in(computer_ids),
        )
        for computer_id, license in result:
            if computer_id not in computer_license_index:
                computer_license_index[computer_id] = []
            computer_license_index[computer_id].append(license)

        registry = getUtility(IRoleRegistry)
        for computer_id, licenses in list(computer_license_index.items()):
            self._cache_roles(registry, computer_id, licenses)

    @requires("ViewAccount")
    def reset_licenses(self, computer):
        """Reset information about C{computer} in the internal license cache.

        It will be automatically repopulated when a license requests is next
        made for C{computer} or when L{preload_licenses} is called with
        information about it.  Information about other L{Computer}s is not
        affected.

        @param computer: The L{Computer} to reset.
        """
        if self._computer_roles_cache is not None:
            self._computer_roles_cache.pop(computer.id, None)

    def _cache_roles(self, registry, computer_id, licenses):
        """Cache L{Role}s provided by C{licenses} for C{computer_id}.

        @param registry: A L{RoleRegistry} instance.
        @param computer_id: The L{Computer.id} that C{licenses} are for.
        @param licenses: The L{License}s for C{computer_id}.
        @return: The set of L{Role}s cached for C{computer_id}.
        """
        roles = set()
        for license in licenses:
            if not license.expired():
                role = registry.get(license.role_key)
                if role is not None:
                    roles.add(role)
        if self._computer_roles_cache is None:
            self._computer_roles_cache = {}
        self._computer_roles_cache[computer_id] = roles
        return roles

    @requires(PUBLIC)
    def get_roles_by_computer(self, computer):
        """Get the C{set} of L{Role}s a L{Computer} has in this account."""
        assert computer.account == self
        if self.is_self_service:
            # Self service accounts don't have licenses. The
            # distinction between full machines, VMs and containers is
            # important for pricing but not for c.l.security.
            registry = getUtility(IRoleRegistry)
            return {registry.get("BasicFeatures")}
        if self._computer_roles_cache is None:
            self._computer_roles_cache = {}
        roles = self._computer_roles_cache.get(computer.id)
        if roles is None:
            registry = getUtility(IRoleRegistry)
            roles = self._cache_roles(registry, computer.id, computer.licenses)
        return roles

    @requires("CreateComputer")
    def create_computer(
        self,
        secure_id,
        title,
        access_context=None,
        hostname=None,
        resource_database_key=None,
        insecure_id=None,
        vm_info=None,
        container_info=None,
        juju_model_uuid=None,
        juju_machine_id=None,
        last_exchange_time=None,
        clone_id=None,
        ubuntu_pro_info=None,
    ):
        """Create a computer.

        This will consume one arbitrary license seat from any available
        L{License}.  Resource and account computer twin objects will be
        associated with the new L{Computer}.  All remaining positional and
        keyword arguments will be passed on to the constructor of L{Computer}.

        @param insecure_id: If this keyword argument is passed, it will become
            the computer's insecure_id in both the main and resource stores.
        @param resource_database: If this keyword argument is passed, the
            resource data will be stored in the given resource database.
        @param broadcast: Optionally, a flag that specified whether or not an
            L{AddComputerEvent} is broadcast.  Defaults to True.
        """
        from canonical.landscape.model.main.license import LicenseSeatError
        from canonical.landscape.model.main.computer import (
            Computer,
            ComputerStatus,
            ComputerUtilisation,
            get_computer_type,
        )

        # FIXME Add some nice logic to decide what resource database to use.
        store = Store.of(self)

        if resource_database_key is None:
            resource_database_key = self.default_resource_database_key

        if last_exchange_time is not None:
            last_exchange_time = last_exchange_time.replace(tzinfo=None)

        if access_context is None:
            access_context = self.access_context

        computer = Computer(
            self,
            access_context,
            secure_id,
            title,
            hostname=hostname,
            resource_database_key=resource_database_key,
            insecure_id=insecure_id,
            vm_info=vm_info,
            container_info=container_info,
            juju_model_uuid=juju_model_uuid,
            juju_machine_id=juju_machine_id,
            clone_id=clone_id,
            ubuntu_pro_info=ubuntu_pro_info,
        )

        computer.status = ComputerStatus(computer.id, last_exchange_time)
        computer.utilisation = ComputerUtilisation(
            self.id, computer.id, computer_type=get_computer_type(computer)
        )

        if not self.is_self_service:
            from canonical.landscape.model.license.management import LicenseManagement

            license_management = LicenseManagement(self)
            found_license = license_management.auto_assign_license_to_computer(computer)
            if not found_license:
                store.remove(computer)  # don't leave the Computer around!
                raise LicenseSeatError("No suitable licenses found for computer")

        # Notify creation of new computer
        event = AddComputerEvent(self, computer)
        notify(event)

        return computer

    @requires("CreateAccessContext")
    def create_access_context(self, name, description=None):
        """Create a child L{AccessContext} called C{title}.

        @param name: The human-readable C{unicode} name for the new access
            context.
        @param description: The human-readable C{unicode} description for the
            new access context.
        @return: The new L{AccessContext}.
        """
        return self.access_context.create_child(name, description=description)

    @requires("ViewAccessContext")
    def get_access_context(self, id):
        """Get an L{AccessContext} in this account.

        @param id: The L{AccessContext.id} to match.
        @return: An L{AccessContext} instance, or C{None} if the C{id} doesn't
            match an access context in this account.
        """
        if self.access_context.id == id:
            return self.access_context
        result = get_main_store().find(
            AccessContext,
            AccessContext.id == id,
            AccessContext.root_id == self.access_context.id,
        )
        return result.one()

    @requires("ViewAccount")
    def get_access_context_by_name(self, person, name, permission):
        """
        Get the L{AccessContext} with the given C{name} that the person has
        the C{permission} on.
        """
        access_context_ids = get_access_context_ids_for_person(
            self.access_context, permission, person
        )
        access_context = (
            get_main_store()
            .find(
                AccessContext,
                AccessContext.name == name,
                AccessContext.id.is_in(access_context_ids),
            )
            .one()
        )
        if access_context is None:
            # We do this check last to optimize for success
            if not self.has_access_context(name):
                raise UnknownAccessContextError(name)

            raise Unauthorized(
                "Permission denied. Unauthorised action on access group '%s'." % name
            )
        return access_context

    @requires("ViewAccount")
    def has_access_context(self, name=None, title=None):
        """
        Returns true if this account has an L{AccessContext} at any level with
        the specified C{name} or C{title}.

        @param name: The C{unicode} name to match an L{AccessContext} to.
        @param title: The C{unicode} title to match an L{AccessContext} to.
        """
        clauses = []

        if name is not None:
            if self.access_context.name == name:
                return True
        elif self.access_context.description == title:
            return True

        if name is not None:
            clauses.append(AccessContext.name == name)
        if title is not None:
            clauses.append(AccessContext.description == title)

        result = get_main_store().find(
            AccessContext, AccessContext.root_id == self.access_context.id, *clauses
        )
        return not result.is_empty()

    @requires("ViewAccount")
    def get_access_contexts(self, names=None, titles=None, ids=None):
        """Get the L{AccessContext}s in the account, possibly limiting the
        search to the given names, titles and/or ids.

        @raises UnknownAccessContexts: If one or more names or titles don't
            match any access contexts in the account.
        """
        store = get_main_store()
        root_id = self.access_context.root_id

        clauses = []
        if names:
            clauses.append(AccessContext.name.is_in(names))
        if titles:
            clauses.append(AccessContext.description.is_in(titles))
        if ids is not None:
            clauses.append(AccessContext.id.is_in(ids))

        result = store.find(AccessContext, AccessContext.root_id == root_id, *clauses)
        access_contexts = list(result)
        known_names = []
        known_titles = []
        for access_context in access_contexts:
            known_names.append(access_context.name)
            known_titles.append(access_context.description)

        unknown_names = set(names) - set(known_names) if names else []
        unknown_titles = set(titles) - set(known_titles) if titles else []
        if unknown_names or unknown_titles:
            raise UnknownAccessContextsError(
                sorted(unknown_names), sorted(unknown_titles), self
            )

        return access_contexts

    @requires("ViewAccount")
    def get_tags(self):
        from canonical.landscape.model.main.computer import Computer
        from canonical.landscape.model.main.tag import ComputerTag, Tag
        from canonical.landscape.model.account.computergroup import ComputerGroupTag

        result = self.store.find(
            ComputerGroupTag.tag_id,
            ComputerGroupTag.computer_group_id == ComputerGroup.id,
            ComputerGroup.account_id == self.id,
        )
        result.config(distinct=True)
        tag_ids = list(result)

        subselect = Select(
            ComputerTag.tag_id,
            where=And(
                Computer.id == ComputerTag.computer_id, Computer.account_id == self.id
            ),
        )
        result = Store.of(self).find(
            Tag, Or(Tag.id.is_in(tag_ids), Tag.id.is_in(subselect))
        )
        result.config(distinct=True)
        result.order_by(Tag.name)
        return result

    @requires("ViewAccount")
    def has_tags(self, tags):
        return set(tags) <= set(self.get_tags())

    @requires("ViewAccount")
    def has_people_in_role(self, role):
        """True if there is at least one person with the given role."""
        for person in self.people:
            if person.has_role(role, self.access_context):
                return True
        return False

    @requires("ViewAccount")
    def get_computers(
        self,
        computer_ids=None,
        tags=None,
        tag_ids=None,
        reboot_required_flag=None,
        keywords=None,
        person=None,
        access_context_ids=None,
    ):
        """Get L{Computer}s matching specified criteria.

        @param computer_ids: Optionally, a sequence of L{Computer.id}s to
            match with L{Computer}s.
        @param tags: Optionally, a sequence of L{Tag} objects to match with
            L{Computer}s.  This parameter cannot be provided at the same time
            as C{tag_ids}.
        @param tag_ids: Optionally, a sequence of L{Tag.id}s to match with
            L{Computer}s.  This parameter cannot be provided at the same time
            as C{tags}.
        @param reboot_required_flag: Optionally, a boolean flag describing
            whether or not to include L{Computer}s that either require a
            reboot or don't.  By default, computers are included in the result
            regardless of reboot required status.
        @param keywords: Optionally, a list of keywords to match with
            L{Computer.hostname} and L{Computer.title}.
        @param person: Optionally, the L{Person} to filter L{Computers}s by
            C{ViewComputer} permission for.
        @param access_context_ids: Optionally, a sequence of
            L{AccessContext.id}s to filter L{Computer}s by.
        @raise RuntimeError: Raised if C{tags} and C{tag_ids} are both
            provided.
        @return: A C{ResultSet} with matching L{Computer} instances.
        """
        from canonical.landscape.model.computer.collection import ComputerCollection

        if tags and tag_ids:
            raise RuntimeError("Only one of tags or tag_ids may be provided.")
        if tags:
            tag_ids = [tag.id for tag in tags]
        if person:
            allow_computer_ids = person.get_computer_ids(self)
            if allow_computer_ids is not None:
                if computer_ids is not None:
                    computer_ids = set(computer_ids)
                    computer_ids = computer_ids.intersection(allow_computer_ids)
                else:
                    computer_ids = allow_computer_ids
        collection = ComputerCollection(
            Store.of(self),
            account_id=self.id,
            computer_ids=computer_ids,
            tag_ids=tag_ids,
            keywords=keywords,
            reboot_required_flag=reboot_required_flag,
            access_context_ids=access_context_ids,
        )
        return collection.find()

    @requires("ViewAccount")
    def get_child_computers(self, parent_computer_ids, access_context_ids=None):
        """Get L{Computer}s that are child computers of the given L{Computer}s.

        NOTE: there is a similar method defined in L{ComputerGroupHelper}.

        @param parent_computer_ids: a sequence of parent L{Computer.id}s used to
            match with child L{Computer}s.
        @param access_context_ids: optionally, a sequence of
            L{AccessContext.id}s to filter the child L{Computer}s by.
        @return: a C{ResultSet} with matching L{Computer} instances.
        """
        # SQL query fails if list is empty
        if not parent_computer_ids:
            return []

        from canonical.landscape.model.computer.collection import ComputerCollection
        from canonical.landscape.model.main.computer import (
            Computer,
            ComputerRelationship,
        )

        # The PostgreSQL ltree type has implicit cast from string but not int
        str_ids = ",".join([f"'{computer_id}'" for computer_id in parent_computer_ids])
        where = And(
            SQL(f"ARRAY[{str_ids}]::LTREE[] @> tree"),
            Not(In(ComputerRelationship.computer_id, parent_computer_ids)),
        )

        store = get_main_store()
        child_ids = store.find(ComputerRelationship.computer_id, where)
        collection = ComputerCollection(
            Store.of(self),
            account_id=self.id,
            computer_ids=child_ids,
            access_context_ids=access_context_ids,
        )

        return collection.find()

    @requires("ViewAccount")
    def get_computers_count(self):
        """Get the total number of computers in the account."""
        return self.get_computers().count()

    @requires("ViewAccount")
    def get_computer(self, computer_id):
        """
        Get a computer that's associated with this account.
        """
        from canonical.landscape.model.main.computer import Computer

        computer = Store.of(self).get(Computer, computer_id)
        # Using Store.get and then checking additional things is an
        # optimization to continue using the Store.get cache.
        if computer and computer.account_id == self.id:
            return computer

    @requires("ViewPendingComputer")
    def get_pending_computers(self, person=None):
        """
        Get a result set of all pending computers on the account.

        If person is passed, it will beused to do a security check -- the
        person must have landscape.ViewPendingComputer on some access context
        within the account.
        """
        from canonical.landscape.model.main.computer import PendingComputer

        store = Store.of(self)
        result = store.find(PendingComputer, account_id=self.id)
        result.order_by(PendingComputer.creation_time)
        return result

    @requires("ViewPendingComputer")
    def get_pending_computer_by_id(self, person, id):
        from canonical.landscape.model.main.computer import PendingComputer

        store = Store.of(self)
        return store.find(PendingComputer, account_id=self.id, id=id).one()

    @requires(PUBLIC)
    def get_pending_computer_count(self):
        """Get the number of computers that are pending."""
        from canonical.landscape.model.main.computer import PendingComputer

        store = Store.of(self)
        return store.find(
            PendingComputer, PendingComputer.account_id == self.id
        ).count()

    @requires(PUBLIC)
    def create_pending_computer(self, *args, **kwargs):
        from canonical.landscape.model.main.computer import PendingComputer

        return PendingComputer(self, *args, **kwargs)

    @requires(PUBLIC)
    def check_registration_password(self, password):
        return not self.registration_password or self.registration_password == password

    @requires(PUBLIC)
    def has_registration_password(self):
        """Return true if this account has a registration password set."""
        return bool(self.registration_password)

    @requires("CreateActivity")
    def create_activity(self, info_class, **kwargs):
        from canonical.landscape.model.activity.parent import ActivityParent
        from canonical.landscape.model.eventlog.management import EventLog

        activity = ActivityParent(self, info_class, **kwargs)
        if isinstance(activity.info, ActivityGroup):
            event_log = EventLog(self, kwargs.get("creator"))
            event_log.create_log_entry(activity.info, "Group activity created.")
        return activity

    @requires("GetResourceStores")
    def get_resource_stores(self, computer_ids=None):
        """Fetch resource stores for computers in this account.

        @param computer_ids: Fetch the stores for computers with these IDs.
        """
        from canonical.landscape.model.main.computer import Computer

        zstorm = getUtility(IZStorm)
        where = [Computer.account == self]
        if computer_ids is not None:
            where.append(Computer.id.is_in(computer_ids))
        result = Store.of(self).find(Computer, *where).config(distinct=True)
        resource_store_keys = set(result.values(Computer.resource_database_key))
        stores = set(
            [zstorm.get("resource-%d" % (store_id)) for store_id in resource_store_keys]
        )
        return list(stores)

    @requires("ViewAccount")
    def calculate_cutoff_datetime(self, date=None):
        now = date or datetime.utcnow()
        # one second after midnight today
        today = datetime(now.year, now.month, now.day, 0, 0, 1)
        cutoff_date = today + timedelta(-self.resource_lifespan)
        return cutoff_date

    def remove(self, allow_reuse=False):
        """
        Remove removes this account and all associated computers.

        If allow_reuse is True, do not create a C{RemovedAccountName} to
        prevent re-use of the account's name.
        """
        from canonical.landscape.model.main.computer import Computer
        from canonical.landscape.model.computer.management import ComputerManagement
        from canonical.landscape.model.computer.twin import ComputerTwin
        from canonical.landscape.model.selection.selection import ComputerSelection
        from canonical.landscape.model.resource.store import get_resource_stores

        computer_ids = list(self.computers.values(Computer.id))
        selection = ComputerSelection(self)
        selection.select_computers(computer_ids)
        management = ComputerManagement(selection, None)

        # Don't create log entries which will just be cascade removed when the
        # account is removed.
        management.remove_computers(create_log_entries=False)

        # Remove all ComputerTwins from the resource database
        for _, resource_store in get_resource_stores():
            resource_store.find(
                ComputerTwin, ComputerTwin.id.is_in(computer_ids)
            ).remove()

        twin = self.get_account_twin()
        Store.of(twin).remove(twin)
        name = self.name
        access_context_id = self.access_context_id
        store = Store.of(self)
        store.remove(self)
        store.find(AccessContext, AccessContext.root_id == access_context_id).remove()
        if not allow_reuse:
            store.add(RemovedAccountName(name))

    def get_account_twin(self):
        """
        Return this account's representation in the account database.
        """
        from canonical.landscape.model.account.twin import AccountTwin

        return self.store.find(AccountTwin, AccountTwin.id == self.id).one()

    @property
    @requires(PUBLIC)
    def is_disabled(self):
        """Return True if the account is currently disabled."""
        return self.disabled_reason is not None

    def disable(self, reason):
        """Return True after disabling the account.

        The given string is set as the reason.  If the account
        is already disabled then return False and don't change
        the reason.
        """
        assert reason is not None
        if self.disabled_reason is not None:
            return False
        self.disabled_reason = reason
        return True

    def enable(self):
        """Return True after enabling the account.

        If the account was already enabled, return False.
        """
        if self.disabled_reason is None:
            return False
        self.disabled_reason = None
        return True

    def disable_admins(self):
        """Disable the admins for this account."""
        for person in self.people:
            self.disable_membership(person)

    @requires("EditAlert")
    def create_rule(self, type, **kwargs):
        """Create a L{Rule} associated with this L{Account}.

        Keyword arguments will be passed to the L{Rule} constructor after the
        account.

        @param type: A subclass of L{RuleInfo}.
        """
        from canonical.landscape.model.alert.rule import Rule

        return Rule(self, type, **kwargs)

    @requires("QueryAlertRule")
    def get_rules(
        self,
        info_classes=None,
        active=None,
        computers=None,
        person=None,
        computer_ids=None,
    ):
        """Get a L{ResultSet} of L{Rule}s associated with this account.

        @param info_classes: Optionally, a sequence of L{RuleInfo} class types
            to filter results by.
        @param active: Optionally, only include L{Rule}s with a matching
            active state.  Use None for all rules, True for active and False
            for inactive.  Default is None.
        @param computers: Optionally, a list of computers to filter results by.
        """
        from canonical.landscape.model.alert.collection import RuleCollection

        if info_classes is not None:
            info_types = [info_class.info_type for info_class in info_classes]
        else:
            info_types = None
        if computers is not None:
            rule_computer_ids = [computer.id for computer in computers]
        else:
            rule_computer_ids = None

        if computer_ids is not None:
            if rule_computer_ids is None:
                rule_computer_ids = computer_ids
            else:
                rule_computer_ids.extend(computer_ids)

        collection = RuleCollection(
            self.store,
            account_id=self.id,
            active=active,
            info_types=info_types,
            computer_ids=rule_computer_ids,
            person=person,
        )
        return collection.find()

    def create_stock_roles(self):
        """Create the stock L{StoredRole}s for this account."""
        # Create stock roles
        from canonical.landscape.model.access.stockroles import StockRoleManager

        stock_role_manager = StockRoleManager()
        stock_role_manager.create_stock_roles(self)

    def create_stock_rules(self):
        """Create the stock L{AlertRule}s for this account."""
        # Create stock rules.
        from canonical.landscape.model.alert.policy import AlertRulePolicy

        policy = AlertRulePolicy()
        policy.create_rules_for_account(self)

    @requires("QueryAlertRule")
    def create_job_get_active_rules(self, page_key, person):
        dispatcher = getUtility(IDispatcher)
        return dispatcher.create_job(
            page_key,
            self.namespace,
            "get-active-rules",
            account_id=self.id,
            person_id=person.id,
        )

    @requires("QueryAlertRule")
    def get_rule_by_id(self, id, person):
        from canonical.landscape.model.alert.rule import Rule

        where = [Rule.id == id, Rule.account_id == self.id]
        rule = self.store.find(Rule, *where).one()
        if rule is not None and rule.info.is_allowed(person, self):
            return rule

    @requires("QueryAlertRule")
    def get_subscribed_rules(self, person):
        """Get a L{ResultSet} of L{Rule}s that C{person} is subscribed to."""
        from canonical.landscape.model.alert.collection import RuleCollection

        collection = RuleCollection(
            self.store, account_id=self.id, subscriber_id=person.id
        )
        return collection.find()

    @requires("CheckRules")
    def check_rules(self, info_classes, computer_ids=None):
        """Check L{Rule}s with the specified L{RuleInfo}-based class types.

        @param info_classes: A sequence of L{RuleInfo}-based class types.
        @param computer_ids: A sequence of L{Computer.id}s that must be
            provided if a L{ComputeRuleInfo}-based L{Rule} is in the list of
            C{info_classes}.
        """
        from canonical.landscape.model.alert.account import AccountRuleInfo

        for rule in self.get_rules(info_classes=info_classes):
            if isinstance(rule.info, AccountRuleInfo):
                rule.info.check()
            else:
                if not computer_ids:
                    raise RuntimeError(
                        "Computer.id's must be specified when "
                        "computer-based rules are checked."
                    )
                rule.info.check(computer_ids)

    @requires("RejectPendingComputer")
    def reject_computers(self, person, computers):
        """
        Reject the given computers from this account.

        @param person: the user launching this action.
        @param computers: the list of pending computers to reject.
        """
        for pending_computer in computers:
            pending_computer.reject(person)

    @requires("ViewAccount")
    def accept_computers(self, person, computers):
        """
        Accept the given computers into this account.

        @param person: the user launching this action.
        @param computers: the list of pending computers to accept.

        @return: The computers that were accepted.
        """
        new_computers = []
        for pending in computers:
            computer = pending.accept(person)
            new_computers.append(computer)
        return new_computers

    @requires("ViewComputerSelection")
    def get_computer_statuses(self, computers):
        """Return the L{ComputerStatus} objects for the given computers."""
        from canonical.landscape.model.main.computer import ComputerStatus

        computer_ids = [computer.id for computer in computers]
        store = Store.of(self)
        return store.find(
            ComputerStatus, ComputerStatus.computer_id.is_in(computer_ids)
        )

    @requires("ViewPingTimes")
    def get_ping_times(self, computers):
        """Return the L{PingTime} objects for the given computers."""
        from canonical.landscape.model.main.computer import PingTime

        computer_ids = [computer.id for computer in computers]
        store = Store.of(self)
        return store.find(PingTime, PingTime.computer_id.is_in(computer_ids))

    @requires(PUBLIC)  # Public because unlicensed computers call it in apis.py
    def has_expired(self, now=None):
        """C{True} if all L{License}s in this account are expired.

        @param now: A C{date} object representing the current time.
        """
        collection = AccountCollection(
            get_main_store(), account_ids=[self.id], expired=True, expiration_date=now
        )
        return not collection.find().is_empty()


class TrialAccount(Storm):
    """Capture information about a trial L{Account}.

    @ivar account: The L{Account} to link to.
    @ivar creator: The L{Person} creating this L{Account}.
    @ivar daytime_phone: Daytime phone number for C{creator}.
    @ivar company_size: Number of people in C{creator}'s company.
    @ivar country: ISO 3166 code for the country which the C{creator} comes
        from.
    @ivar region: The administrative region that C{creator} comes from.  This
        may be a 2-letter code for a US state or Canadian province/territory.
    @ivar mailing_list_optin: Boolean flag, True if C{creator} opted in to the
        marketing mailing list.
    """

    __storm_table__ = "trial_account"

    requires(PUBLIC, PRIVATE, ["secure_id"])
    requires("ViewAccount", PRIVATE, ["creator", "account"])
    requires(
        "ViewAccount",
        "EditAccount",
        [
            "daytime_phone",
            "country",
            "company_size",
            "region",
            "mailing_list_optin",
            "name",
            "city",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    secure_id = Unicode()
    account_id = Int(allow_none=True)
    creator_id = Int(allow_none=True)
    company_name = Unicode(allow_none=False)
    daytime_phone = Unicode()
    company_size = Unicode()
    country = Unicode()
    region = Unicode()
    mailing_list_optin = Bool()
    city = Unicode()
    registration_comments = Unicode()

    account = Reference(account_id, Account.id)
    creator = Reference(creator_id, Person.id)

    def __init__(
        self,
        company_name,
        daytime_phone=None,
        company_size=None,
        country=None,
        region=None,
        mailing_list_optin=False,
        city=None,
        registration_comments=None,
    ):
        self.secure_id = make_secure_id()
        self.company_name = company_name
        self.daytime_phone = daytime_phone
        self.company_size = company_size
        self.country = country

        self.region = region
        self.mailing_list_optin = mailing_list_optin
        self.city = city
        self.registration_comments = registration_comments

    @requires("ActivateTrial")
    def activate(self, creator):
        """Activate the trial account.

        This will create an actual L{Account} object using the company name
        provided by the user as title and a trial license.
        """
        title = self.company_name
        name = generate_account_name(title)
        account = bootstrap_account(name, title, person=creator)
        self.account_id = account.id
        self.creator_id = creator.id
        self.secure_id = None  # Prevent the trial from being activated again

        expiration_date = utctoday() + timedelta(days=LENGTH_OF_TRIAL_IN_DAYS)
        from canonical.landscape.model.license.management import LicenseManagement

        LicenseManagement(account).create_license(
            "TrialFeatures", 50, expiration_date=expiration_date
        )
        return account


class RemovedAccountName(object):
    """
    Stores the names of accounts which have been removed from the L{Account}
    table.
    """

    __storm_table__ = "removed_account_name"

    requires("ViewAccount", PRIVATE, ["id", "name"])
    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = LowercaseUnicode(allow_none=False)

    def __init__(self, name):
        self.name = name


def get_account_by_name(name):
    """Get the L{Account} called C{name}, or C{None} if one doesn't exist."""
    collection = AccountCollection(get_main_store(), account_names=[name])
    return collection.find().one()


@requires("ViewAllAccounts")
def get_all_accounts():
    """Get all L{Account}s in the system.

    You probably want to import this from safe.canonical.landscape.main
    """
    collection = AccountCollection(get_main_store())
    return collection.find()


@requires("ViewAllAccounts")
def get_accounts_by_ids(account_ids):
    """Get all L{Account}s matching C{account_ids}."""
    collection = AccountCollection(get_main_store(), account_ids=account_ids)
    return collection.find()


Person.accounts = ReferenceSet(
    Person.id,
    PersonAccount.person_id,
    PersonAccount.account_id,
    Account.id,
    order_by=Account.name,
)


def get_accounts_by_id_range(minimum_id=0, maximum_id=None):
    """Get the L{Account}s with C{id}s within a range.

    Return an ordered resultset of C{id}s greater than or equal to
    C{minimum_id} and less than or equal to C{maximum_id}."""
    where = [Account.id >= minimum_id]
    if maximum_id is not None:
        where += [Account.id <= maximum_id]
    return get_main_store().find(Account, where).order_by(Account.id)


def get_account_by_id(id):
    """Get the L{Account} matching C{id}, or None if one isn't found."""
    return get_main_store().get(Account, id)


def get_account_by_sfid(salesforce_key):
    """
    Returns the account linked to SalesforceUser with this salesforce_key.
    """
    from canonical.landscape.model.main.salesforceuser import SalesforceUser

    sfuser = (
        get_main_store()
        .find(SalesforceUser, SalesforceUser.salesforce_key == salesforce_key)
        .one()
    )
    if sfuser:
        return sfuser.account


def get_account_by_salesforce_account_key(salesforce_account_key):
    """Returns the account linked to the Salesforce account key."""
    return (
        get_main_store()
        .find(Account, Account.salesforce_account_key == salesforce_account_key)
        .one()
    )


def get_account_access_contexts_vocabulary(context):
    """Return a vocabulary for all the access contexts of the given account.

    If the given C{context} doesn't provide L{IAccount}, then
    C{context.account} is used.
    """
    if IAccount.providedBy(context):
        account = context
    elif IAccountAware.providedBy(context):
        account = context.account
    else:
        raise ContextParameterError()

    access_contexts = sorted(
        account.get_access_contexts(),
        key=lambda access_context: access_context.description,
    )
    terms = [
        SimpleTerm(access_context.name, title=access_context.description)
        for access_context in access_contexts
    ]
    return SimpleVocabulary(terms)


def is_valid_account_name(name):
    """
    @return: True if name is a valid Account.name attribute
    """
    # Have we removed an account with this name previously?
    store = get_main_store()
    if store.find(RemovedAccountName.name, RemovedAccountName.name == name).one():
        return False
    return is_valid_slug(name, max_length=60)


def generate_account_name(account_title):
    """
    Generate a unique name for a C{Account} which can be used in the URL slug
    when accessing Accounts.
    """
    return slug_from_title(account_title, 55, _existing_account_check)


def generate_random_account_name():
    """Generate a unique random name for an Account, to be used in URL."""
    # Names are case-insensitive, so only lowercase and letters.
    # This still gives 36^8 possibilities which should be way enough.
    characters = string.ascii_lowercase + string.digits
    name = random_string(8, alphabet=characters)
    while _existing_account_check(name):
        name = random_string(8, alphabet=characters)
    return name


def _existing_account_check(slug):
    """Returns a list of matching account names, if any."""
    store = get_main_store()
    result = store.find(Account, Like(Account.name, slug + "%"))
    matching_account_names = set(result.values(Account.name))
    result2 = store.find(RemovedAccountName, Like(RemovedAccountName.name, slug + "%"))
    matching_account_names.update(result2.values(RemovedAccountName.name))
    return matching_account_names


def create_account(name, title, *args, **kwargs):
    """Create a new L{Account}.

    @param name: Name for this account.
    @type name: str
    @param title: Title for this account.
    @type title: str

    @return: Newly created Account object.
    """
    account = Account(name, title, *args, **kwargs)
    # This will result in an exception being raised if the account
    # name is a duplicate.
    get_main_store().flush()

    from canonical.landscape.model.account.twin import AccountTwin

    account_twin = AccountTwin(account)
    account.store.add(account_twin)
    # We flush the resource store because we want the resource
    # account in place for foreign keys that don't go through the
    # resource account object, and thus have no ordering rules in
    # place.  This behaviour ensures correct ordering of SQL
    # statements emitted by Storm.
    account.store.flush()

    return account


def bootstrap_account(name, title, person=None, registration_key=None, **kwargs):
    """Create an L{Account} with basic setup (initial admin, alerts, roles).

    XXX: This should be actually moved to a CreateAccount service layer
    operation, since this is a high-level operation directly used by the
    API and the UI.
    """
    from canonical.landscape.model.main.license import (
        check_lds_license,
        CHECK_LDS_LICENSE_ERRORS,
    )

    account = create_account(
        name, title, registration_password=registration_key, **kwargs
    )
    account.create_stock_roles()
    account.create_stock_rules()
    if person is not None:
        account.add_person(person, roles=["Owner"])
    if is_standalone():
        try:
            check_lds_license()
        except CHECK_LDS_LICENSE_ERRORS:
            pass
    return account


def remove_account(account):
    """Remove the given L{Account}."""
    from canonical.landscape.model.account.twin import AccountTwin

    twin = account.store.get(AccountTwin, account.id)
    account.store.remove(twin)
    account.store.flush()
    store = get_main_store()
    store.remove(account)
    store.flush()


def can_have_trial(person):
    """Return whether the given person is entitled to have a trial account."""
    return is_staging_environment() or not person.accounts.any()


def create_trial_account(
    company_name,
    daytime_phone=None,
    company_size=None,
    country=None,
    region=None,
    mailing_list_optin=False,
    city=None,
    company_trial=False,
    registration_comments=None,
):
    """Create a new L{Account} with an associated L{TrialAccount}.

    @param company_name: The name of the organisation, it will become the
        account title once the trial is activated.
    @param daytime_phone: The phone number to use for daytime contacts.
    @param company_size: The size of the company using requesting the new
        account.
    @param country: The ISO-3166-1 country code of the company requesting the
        new account.
    @param person_name: The full name of the registrant
    @param mailing_list_optin: True if the registrant has opted into receive
        updates from Canonical.
    @param city: Name of the string the registrant is from
    @param region: The US state or Canadian province of the company requesting
        the new account, if the company is in the US or Canada.
    @param registration_comments: Text comments entered by the user upon
        registration.
    @return: A 2-tuple with the new L{TrialAccount} and L{Account} instances.
    """
    trial = TrialAccount(
        company_name,
        daytime_phone=daytime_phone,
        company_size=company_size,
        country=country,
        region=region,
        mailing_list_optin=mailing_list_optin,
        city=city,
        registration_comments=registration_comments,
    )
    store = get_main_store()
    store.add(trial)
    return trial


def get_trial_accounts(start_time, stop_time=None, person=None):
    """
    Retrieve L{TrialAccount}s and associated L{Account}s created between
    C{start_time} and C{stop_time}.

    @param start_time: Time after which the trials must have been created.
    @type start_time: C{datetime}
    @param stop_time: Optionally, time before which the trials must have been
        created.
    @type stop_time: C{datetime} or C{NoneType}
    @param person: Optionally, restrict trial accounts to ones created by the
        Person.
    @type person: C{Person} or C{NoneType}
    @return: A C{ResultSet} of C{TrialAccount}s and L{Account}s.
    """
    where = [TrialAccount.account_id == Account.id, Account.creation_time >= start_time]
    if stop_time is not None:
        where.append(Account.creation_time < stop_time)
    if person is not None:
        where.append(TrialAccount.creator_id == person.id)
    store = get_main_store()
    result = store.find((TrialAccount, Account), *where)
    result.order_by(Account.creation_time)
    return result


def get_trial_account_by_secure_id(secure_id):
    """Return the L{TrialAccount} with the given C{secure_id}."""
    store = get_main_store()
    return store.find(TrialAccount, TrialAccount.secure_id == secure_id).one()
