# Copyright 2005-2022 Canonical Limited. All rights reserved.
import json
from collections import namedtuple
from datetime import datetime
from hashlib import md5
from json.decoder import JSONDecodeError
from typing import Optional, Union

from pytz import UTC

from zope.interface import implementer
from zope.security.proxy import isinstance, removeSecurityProxy
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.schema.interfaces import IContextSourceBinder
from zope.event import notify
from zope.component import getUtility

from storm.locals import (
    Store,
    Int,
    RawStr,
    Unicode,
    DateTime,
    Reference,
    ReferenceSet,
    AutoReload,
    Bool,
    Storm,
    Float,
    Enum,
)
from storm.expr import (
    Coalesce,
    Update,
    And,
    Cast,
    Not,
    Desc,
    Alias,
    LeftJoin,
    Count,
    Select,
    SQL,
    Join,
)
from storm.store import EmptyResultSet
from storm.databases.postgres import Case

from landscape.lib.tag import is_valid_tag

from canonical.amqp.message.interfaces import IDispatcher

from canonical.lib.timestamp import format_datetime
from canonical.lib.text import make_secure_id, arg_to_bytes

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.storm.variables import BytesStr

from canonical.landscape.model.main.errors import (
    AcceptanceUnauthorized,
    UnknownAccessContextsError,
)
from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.model.computer.activities import (
    DeleteChildComputerActivity,
    StartChildComputerActivity,
    StopChildComputerActivity,
)
from canonical.landscape.model.computer.twin import AccountComputerTwin, ComputerTwin
from canonical.landscape.model.main.interfaces import (
    IComputer,
    IPendingComputer,
    IAccessContextAware,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.tag import (
    get_tags,
    Tag,
    ComputerTag,
    get_tag_names,
    InvalidTagError,
)
from canonical.landscape.model.main.accesscontext import AccessContext
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.license import (
    License,
    ProLicenseType,
    check_lds_license,
)
from canonical.landscape.model.main.distribution import Distribution
from canonical.landscape.model.events import (
    ResynchronizeEvent,
    AddTagsEvent,
    RemoveTagsEvent,
    AccountRulesChangedEvent,
)
from canonical.landscape.model.store import StoreHolder
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.status import (
    QUEUED,
    SCHEDULED,
    PENDING_STATUSES,
    BLOCKED,
)
from canonical.landscape.model.resource.cloud_instance_metadata import (
    CloudInstanceMetadataManagement,
)
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.stormfunc import Now, DatePart, Concat, Floor
from canonical.landscape.model.repository.activities import (
    ChangeRepositoryProfilesRequest,
)
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.store import get_package_store


__all__ = [
    "ComputerLicense",
    "Computer",
    "InvalidActivityError",
    "PendingComputer",
    "PendingComputerToAccept",
    "EmptyTagError",
    "PingTime",
    "PingServer",
    "MessageType",
    "ComputerMessageType",
    "get_computer_by_id",
    "get_computer_by_secure_id",
    "get_computers_by_ids",
    "get_pending_computer_by_id",
    "get_pending_computers_by_juju_model_uuid",
    "get_pending_computers_by_ids",
    "get_pending_computer_by_secure_id",
    "get_wedged_computers",
    "add_tags_to_computers",
    "remove_tags_from_computers",
    "ComputerStatus",
    "ComputerUtilisation",
    "ComputersForPendingComputersVocabularyFactory",
    "cleanup_after_removed_computers",
    "get_computer_by_juju_machine_id",
    "get_computer_ids_by_unknown_juju_machine_ids",
    "update_managed_hours",
    "COMPUTER_TYPES",
    "COMPUTER_TYPE_MAP",
    "get_computer_type_expr",
    "get_computer_type",
    "get_computer_counts_by_type",
    "ComputerRelationship",
]


__clearlist__ = [
    "Computer",
    "PingTime",
    "get_computer_type",
    "PendingComputerToAccept",
    "ComputersForPendingComputersVocabularyFactory",
]

__requires_prefix__ = "landscape."


# Type of computers. The order here is important for the COMPUTER_TYPE_MAP
# below.
COMPUTER_TYPES = ("physical", "vm", "container")

# Map computer types to numeric ids. This is used for database Enums.
COMPUTER_TYPE_MAP = dict((name, idx) for idx, name in enumerate(COMPUTER_TYPES, 1))


class EmptyTagError(Exception):
    pass


class InvalidActivityError(Exception):
    """The provided activity_id does not exist or is invalid."""


class ComputerLicense(object):
    __storm_table__ = "computer_license"
    __storm_primary__ = "computer_id", "license_id"

    computer_id = Int(allow_none=False)
    license_id = Int(allow_none=False)


class MessageType(object):
    __storm_table__ = "message_type"
    __storm_order__ = "name"

    id = Int(allow_none=False, primary=True)
    name = Unicode(allow_none=False)

    def __init__(self, name):
        self.name = name


class ComputerMessageType(Storm):
    __storm_table__ = "computer_message_type"
    __storm_primary__ = ("computer_id", "message_type_id")

    computer_id = Int(allow_none=False)
    message_type_id = Int(allow_none=False)

    computer = Reference(computer_id, "Computer.id")
    message_type = Reference(message_type_id, MessageType.id)

    def __init__(self, computer, message_type):
        self.computer = computer
        self.message_type = message_type


class ComputerStatus(object):
    """ComputerStatus stores the status information about a computer.

    This information is kept in a separate table than the computer one because
    its fields are updated frequently (typically once per exchange) and we want
    to minimize database contention when modifying other tables that are linked
    to the computer table (e.g. computer_tag). See #839510 for a more detailed
    analysis.

    @type computer_id: int
    @param computer_id: The id of the computer that this status information
        belongs to.
    @type last_exchange_time: datetime
    @param last_exchange_time: The time this computer last contacted landscape.

    @type computer_id: int
    @ivar computer_id: The id of the computer that this status information
        belongs to.
    @type next_expected_sequence: int
    @ivar next_expected_sequence: The next sequence id that the client is
        expected to use when contacting the server.
    @type consecutive_errors: int
    @ivar consecutive_errors: The number of errors seen since the last
        successful interaction with the client.
    @type last_exchange_time: datetime
    @ivar last_exchange_time: The time this computer last contacted landscape.
    @type last_exchange_duration: float
    @ivar last_exchange_duration: The length of time for last exchange, in
        seconds.
    @type last_exchange_size: float
    @ivar last_exchange_size: The size of the last exchange, in kilobytes.
    @type ignored_count: int
    @ivar ignored_count: The number of messages ignored during the last
        exchange.
    @type processed_count: int
    @ivar processed_count: The number of messages processed during the last
        exchange.
    @type dropped_count: int
    @ivar dropped_count: The number of messages dropped during the last
        exchange.
    @type remaining_count: int
    @ivar remaining_count: The number of message still waiting to be processed
        on the client after the last exchange.
    @type soft_lock_time: datetime
    @ivar soft_lock_time: The time the last exchange began.  soft_lock_time is
        used to prevent a computer from having multiple simultaneous
        exchanges.
    @type is_inconsistent: bool
    @ivar is_inconsistent: Flag that indicates that client information needs
        to be synchronized with the server.
    @type reboot_required_flag: bool
    @ivar reboot_required_flag: True if the client computer needs to be
        rebooted, false otherwise.
    @type update_manager_prompt: unicode
    @ivar update_manager_prompt: defines when we should indicate that new
        releases are available to upgrade this machine. Valid values are:
        "never"; "lts" (only when LTS releases are available as an upgrade);
        and "normal" (whenever any release is available as an upgrade).
    """

    __storm_table__ = "computer_status"

    computer_id = Int(allow_none=False, primary=True)
    next_expected_sequence = Int()
    next_exchange_token = RawStr(default=None)
    consecutive_errors = Int(default=0)
    last_exchange_time = DateTime()
    last_exchange_duration = Float(default=0)
    last_exchange_size = Float(default=0)
    ignored_count = Int(default=0)
    processed_count = Int(default=0)
    dropped_count = Int(default=0)
    remaining_count = Int(default=0)
    soft_lock_time = DateTime()
    is_inconsistent = Bool(default=False)
    reboot_required_flag = Bool(default=False)
    update_manager_prompt = Unicode(default="normal")

    def __init__(self, computer_id, last_exchange_time):
        self.computer_id = computer_id
        self.last_exchange_time = last_exchange_time


class ComputerUtilisation(object):
    """Store utilisation details about Landscape's management of a computer.

    The pingserver frequently updates these rows for every computer
    which actively pings Landscape. ComputerUtilisation is separated from
    ComputerStatus to reduce the risk of concurrent updates on these two
    tables.

    The account ID and computer type are duplicated here since the computer_id
    is set to None when the related computer is removed, but we still need to
    track usage by type until the cost is charged to the account.

    @type account_id: int
    @param account_id: The ID of the account the computer related to this
        utilisation belongs to.
    @type computer_id: int
    @param computer_id: The id of the computer that managed status information
        belongs to. It is None if the related computer has been removed.
    @type computer_type: Enum
    @param computer_type: The type of computer this utilisation relates to.
    @type managed_hours: int
    @param managed_hours: The count of hours the computer has been managed by
        Landscape. This value is determined by checking whether the machine has
        pinged Landscape within a given hour.
    @type next_managed_hour_time: datetime
    @ivar next_managed_hour_time: The time after which a ping from the computer
        results in an incremented managed_hours count.
    """

    __storm_table__ = "computer_utilisation"

    account_id = Int(allow_none=False)
    # computer_id is not really a primary key, but a foreign key. It's marked
    # primary since storm needs a primary key in the model. This is still a
    # good candidate since it's unique.
    computer_id = Int(primary=True)
    computer_type = Enum(map=COMPUTER_TYPE_MAP)
    managed_hours = Int(allow_none=False, default=0)
    next_managed_hour_time = DateTime(allow_none=False)

    account = Reference(account_id, Account.id)

    def __init__(self, account_id, computer_id, computer_type=None):
        self.account_id = account_id
        self.computer_id = computer_id
        self.computer_type = computer_type


@implementer(IComputer, IAccessContextAware)
@requires_advisor
class Computer(StoreHolder):
    """Our knowledge of a machine that gets managed by Landscape.

    This object acts as a hub for all data represented associated with
    the computer.

    @type secure_id: str
    @ivar secure_id: A large random hash used by clients to identify
        themselves when connecting to the HTTP message server. It's
        "secure" because it should be unguessable and only sent over
        encrypted connections. It will imply access to quite a bit of
        information.
    @type insecure_id: int
    @ivar insecure_id: An ID used by clients connecting to the
        canonical.landscape.pingserver to ask whether this computer has
        messages.  It's 'insecure' because it's sent over the wire without
        encryption, and it should only allow access to very little
        information.
    @type consecutive_errors: int
    @ivar consecutive_errors: The number of consecutive client exchanges
        which have resulted in an exception.
    @type current_api: str
    @ivar current_api: The client API on the most recent message exchange.
    @type current_version: str
    @ivar current_version: The client version on the most recent message
        exchange.
    @ivar last_ping_time: Time that the computer last sent a 'ping' to the
        server.
    @type last_ping_time: datetime
    @ivar reboot_required_flag: Flag indicating if a computer requires a
        reboot.
    @type reboot_required_flag: bool
    @ivar vm_info: String representing the vm (if any) that we detected the
        client running in.
    @type vm_info: str
    @ivar container_info: String representing the container (if any) that we
        detected the client running in.
    @type container_info: str
    @ivar juju_model_uuid: The Juju model UUID this computer was
        deployed in, if any.
    @ivar juju_machine_id: The Juju machine ID this computer was deployed to,
        if any.
    @type update_manager_prompt: unicode
    @ivar update_manager_prompt: defines when we should indicate that new
        releases are available to upgrade this machine. Valid values are:
        "never"; "lts" (only when LTS releases are available as an upgrade);
        and "normal" (whenever any release is available as an upgrade).
    """

    __storm_table__ = "computer"

    requires(
        "ViewComputer",
        PRIVATE,
        [
            "id",
            "resource_database_key",
            "distribution_id",
            "licenses",
            "last_ping_time",
            "accept_time",
            "parsed_ubuntu_pro_info",
            "ubuntu_pro_subscription_type",
        ],
    )
    requires("ViewComputer", "RegisterComputer", ["secure_id"])
    requires("ViewComputer", "EditComputerAccount", ["account_id", "account"])
    requires("ViewComputer", "EditNextExpectedSequence", ["next_expected_sequence"])
    requires("ViewComputer", "EditNextExchangeToken", ["next_exchange_token"])
    requires(
        "ViewComputer",
        "EditComputerDetails",
        [
            "title",
            "hostname",
            "total_memory",
            "total_swap",
            "distribution",
            "comment",
            "vm_info",
            "container_info",
            "juju_model_uuid",
            "juju_machine_id",
            "clone_id",
            "ubuntu_pro_info",
            "snaps",
        ],
    )
    requires(
        "ViewComputer",
        "RemoveComputerFromAccessContext",
        ["access_context_id", "access_context"],
    )
    requires(
        "ViewComputer",
        "DeliverMessages",
        [
            "consecutive_errors",
            "current_api",
            "current_version",
            "last_exchange_time",
            "last_exchange_duration",
            "last_exchange_size",
            "ignored_count",
            "processed_count",
            "dropped_count",
            "remaining_count",
            "soft_lock_time",
            "is_inconsistent",
            "accepted_message_types_hash",
            "reboot_required_flag",
            "update_manager_prompt",
            "status",
        ],
    )
    requires("ViewAccountData", "RegisterComputer", ["insecure_id"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    access_context_id = Int(allow_none=False)
    resource_database_key = Int(allow_none=False)
    distribution_id = Int()
    secure_id = BytesStr(allow_none=False)
    insecure_id = Int(default=AutoReload)
    title = Unicode(allow_none=False)
    comment = Unicode()
    hostname = Unicode()
    total_memory = Int()
    total_swap = Int()
    current_api = BytesStr()
    current_version = BytesStr()
    accepted_message_types_hash = RawStr()
    accept_time = DateTime()
    vm_info = Unicode()
    container_info = Unicode()
    juju_model_uuid = Unicode()
    juju_machine_id = Unicode()
    clone_id = Int(allow_none=True)
    ubuntu_pro_info = Unicode()

    account = Reference(account_id, Account.id)
    access_context = Reference(access_context_id, AccessContext.id)
    licenses = ReferenceSet(
        id, ComputerLicense.computer_id, ComputerLicense.license_id, License.id
    )
    distribution = Reference(distribution_id, Distribution.id)
    accepted_message_types = ReferenceSet(
        id,
        ComputerMessageType.computer_id,
        ComputerMessageType.message_type_id,
        MessageType.id,
    )
    status = Reference(id, ComputerStatus.computer_id)
    utilisation = Reference(id, ComputerUtilisation.computer_id)

    snaps = ReferenceSet(
        id,
        "InstalledSnap.computer_id",
    )

    def __init__(
        self,
        account,
        access_context,
        secure_id,
        title,
        hostname=None,
        id=None,
        resource_database_key=None,
        insecure_id=None,
        vm_info=None,
        container_info=None,
        juju_model_uuid=None,
        juju_machine_id=None,
        clone_id=None,
        ubuntu_pro_info=None,
    ):
        if id is not None:
            self.id = id  # Sample data and tests preset it.
        self.account = account
        self.access_context = access_context
        self.resource_database_key = resource_database_key
        self.secure_id = secure_id
        self.title = title
        self.hostname = hostname
        if insecure_id is not None:
            self.insecure_id = insecure_id
        self.vm_info = vm_info
        self.container_info = container_info
        self.juju_model_uuid = juju_model_uuid
        self.juju_machine_id = juju_machine_id
        self.clone_id = clone_id
        self.ubuntu_pro_info = ubuntu_pro_info
        # Add twin objects to the resource and account stores
        self.store.add(ComputerTwin(self))
        self.account.store.add(AccountComputerTwin(self))

    def __repr__(self):
        return f"<Computer id={self.id} at {hex(id(self))}>"

    def last_ping_time():
        """
        Note: the database representation of last_ping_time is naive with
        regard to timezones.  Because we want to present the timezone to the
        user correctly we insinuate the last_ping_time with UTC when we get it,
        and remove that insinuation when we set it.
        """

        def get(self):
            ping_time = Store.of(self).get(PingTime, self.id)
            if ping_time is not None:
                last_ping_time = ping_time.last_ping_time
                if last_ping_time is not None:
                    last_ping_time = last_ping_time.replace(tzinfo=UTC)
                return last_ping_time

        def set(self, value):
            ping_time = Store.of(self).get(PingTime, self.id)
            if ping_time is None:
                ping_time = PingTime()
                ping_time.computer = self
            if value is not None:
                value = value.replace(tzinfo=None)
            ping_time.last_ping_time = value

        return get, set

    last_ping_time = property(*last_ping_time())

    def parsed_ubuntu_pro_info(self):
        try:
            return json.loads(self.ubuntu_pro_info)
        except (JSONDecodeError, TypeError):
            return self.ubuntu_pro_info

    def ubuntu_pro_subscription_type(self) -> Union[ProLicenseType, None]:
        """
        A computer has a valid Ubuntu Pro subscription iff:
          * Ubuntu Pro info is parseable to a dict
          * there's a contract
          * the contract's effective date is in the past
          * the contract's expiration date is in the future
        """
        parsed_info = self.parsed_ubuntu_pro_info()

        if not isinstance(parsed_info, dict):
            return None

        if "contract" not in parsed_info:
            return None

        contract = parsed_info.get("contract")
        if not isinstance(contract, dict):
            return None

        products = contract.get("products")
        if not isinstance(products, list) or not products:
            return None

        if "free" in products:
            return ProLicenseType.FREE

        now = datetime.now(tz=UTC)

        try:
            effective_date = parsed_info.get("effective")
            effective_date = datetime.fromisoformat(effective_date)

            expiry_date = parsed_info.get("expires")
            expiry_date = datetime.fromisoformat(expiry_date)
        except (TypeError, ValueError):
            return None
        else:
            if effective_date > now:
                return ProLicenseType.INACTIVE

            if expiry_date <= now:
                return ProLicenseType.EXPIRED

        return ProLicenseType.FULL

    def _delegate_to_status(field):
        def get(self):
            return getattr(self.status, field)

        def set(self, value):
            return setattr(self.status, field, value)

        return get, set

    next_expected_sequence = property(*_delegate_to_status("next_expected_sequence"))
    next_exchange_token = property(*_delegate_to_status("next_exchange_token"))
    consecutive_errors = property(*_delegate_to_status("consecutive_errors"))
    last_exchange_time = property(*_delegate_to_status("last_exchange_time"))
    last_exchange_duration = property(*_delegate_to_status("last_exchange_duration"))
    last_exchange_size = property(*_delegate_to_status("last_exchange_size"))
    ignored_count = property(*_delegate_to_status("ignored_count"))
    processed_count = property(*_delegate_to_status("processed_count"))
    dropped_count = property(*_delegate_to_status("dropped_count"))
    remaining_count = property(*_delegate_to_status("remaining_count"))
    soft_lock_time = property(*_delegate_to_status("soft_lock_time"))
    is_inconsistent = property(*_delegate_to_status("is_inconsistent"))
    reboot_required_flag = property(*_delegate_to_status("reboot_required_flag"))
    update_manager_prompt = property(*_delegate_to_status("update_manager_prompt"))

    @requires("DeliverMessages")
    def get_accepted_message_types(self):
        return list(self.accepted_message_types.values(MessageType.name))

    @requires("DeliverMessages")
    def set_accepted_message_types(self, types):
        """
        Given a list of message type names accepted by the client:

        - Ensure that there are MessageType objects in the database for each
          given type,
        - Update the list of message types associated with this computer.
        """
        message_types = Store.of(self).find(MessageType, MessageType.name.is_in(types))
        message_types = list(message_types)
        known_type_names = set([x.name for x in message_types])
        unknown_type_names = set(types) - known_type_names
        for unknown_type in unknown_type_names:
            message_types.append(MessageType(unknown_type))
        self.accepted_message_types.clear()
        for message_type in message_types:
            ComputerMessageType(self, message_type)

        self.accepted_message_types_hash = md5(
            ";".join(sorted(types)).encode()
        ).digest()

    @requires("ViewComputer")
    def has_valid_license(self):
        """Returns C{True} if this computer has a valid license.

        C{False} is returned if all of the licenses associated with this
        computer have expired, or if there are no licenses associated with
        this computer.
        """
        return any(not license.expired() for license in self.licenses)

    def get_store_name(self):
        return "resource-%d" % self.resource_database_key

    @requires("ViewComputer")
    def get_store_load_average(self):
        return self.store.execute("SELECT get_load_average()").get_one()[0]

    def _find(self, cls, *args, **kwargs):
        """Helper ensures all queries are tied to this computer instance."""
        return self.store.find(cls, cls.computer_id == self.id, *args, **kwargs)

    @requires("ViewComputer")
    def get_message_store(self):
        """Return a message store for managing outbound messages."""
        from canonical.landscape.model.resource.message import MessageStore

        message_store = self._find(MessageStore).one()
        if message_store is None:
            return self.store.add(MessageStore(self))
        return message_store

    @requires("CreateLoadAverage")
    def create_load_average(self, timestamp, load_average):
        from canonical.landscape.model.resource.loadaverage import LoadAverage

        load_average = LoadAverage(self, timestamp, load_average)
        self.store.add(load_average)
        return load_average

    @requires("ViewLoadAverage")
    def get_load_averages(self, start=None, stop=None):
        """
        Returns an iterable that yields C{(timestamp, load-average)} 2-tuples.
        """
        from canonical.landscape.model.resource.loadaverage import LoadAverage

        where = []
        if start is not None:
            where.append(LoadAverage.timestamp >= start)
        if stop is not None:
            where.append(LoadAverage.timestamp <= stop)
        results = self._find(LoadAverage, *where)
        results.order_by(LoadAverage.timestamp)
        return results.values(LoadAverage.timestamp, LoadAverage.load_average)

    @requires("CreateMemoryInfo")
    def create_memory_info(self, timestamp, free_memory, free_swap):
        from canonical.landscape.model.resource.memoryinfo import MemoryInfo

        memory_info = MemoryInfo(self, timestamp, free_memory, free_swap)
        self.store.add(memory_info)
        return memory_info

    @requires("ViewMemoryInfo")
    def get_memory_infos(self, start=None, stop=None):
        from canonical.landscape.model.resource.memoryinfo import MemoryInfo

        where = []
        if start is not None:
            where.append(MemoryInfo.timestamp >= start)
        if stop is not None:
            where.append(MemoryInfo.timestamp <= stop)
        result = self._find(MemoryInfo, *where)
        return result.order_by(MemoryInfo.timestamp)

    @requires("ViewMemoryInfo")
    def get_free_memory(self, start=None, stop=None):
        """
        Returns an iterable that yields C{(timestamp, free-memory)} 2-tuples.
        """
        from canonical.landscape.model.resource.memoryinfo import MemoryInfo

        result = self.get_memory_infos(start, stop)
        return result.values(MemoryInfo.timestamp, MemoryInfo.free_memory)

    @requires("ViewMemoryInfo")
    def get_free_swap(self, start=None, stop=None):
        """
        Returns an iterable that yields C{(timestamp, free-swap)} 2-tuples.
        """
        from canonical.landscape.model.resource.memoryinfo import MemoryInfo

        result = self.get_memory_infos(start, stop)
        return result.values(MemoryInfo.timestamp, MemoryInfo.free_swap)

    @requires("CreateProcessor")
    def create_processor(self, processor_id, model, vendor, cache_size):
        from canonical.landscape.model.resource.processor import Processor

        processor = Processor(self, processor_id, model, vendor, cache_size)
        self.store.add(processor)
        return processor

    @requires("ViewProcessor")
    def get_processors(self):
        from canonical.landscape.model.resource.processor import Processor

        results = self._find(Processor)
        return results.order_by(Processor.processor_id)

    @requires("RemoveProcessor")
    def remove_processors(self):
        from canonical.landscape.model.resource.processor import Processor

        self._find(Processor).remove()

    @requires("CreateThermalZone")
    def create_thermal_zone(self, name):
        from canonical.landscape.model.resource.temperature import ThermalZone

        thermal_zone = ThermalZone(self, name)
        self.store.add(thermal_zone)
        return thermal_zone

    @requires("ViewThermalZone")
    def get_thermal_zones(self):
        """Get the thermal zone objects associated with this computer.

        These objects allow you to access information about the
        temperature of the machine.
        """
        from canonical.landscape.model.resource.temperature import ThermalZone

        result = self._find(ThermalZone)
        return result.order_by(ThermalZone.name)

    @requires("ViewThermalZone")
    def get_thermal_zone_by_name(self, name):
        from canonical.landscape.model.resource.temperature import ThermalZone

        return self._find(ThermalZone, name=name).one()

    @requires("CreateActiveProcess")
    def create_active_process(
        self, pid, uid, gid, name, state, start_time, vm_size, sleep_average
    ):
        from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess

        process = ActiveProcess(
            self, pid, uid, gid, name, state, start_time, vm_size, sleep_average
        )
        self.store.add(process)
        return process

    @requires("ViewActiveProcess")
    def get_active_processes(self):
        from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess

        result = self._find(ActiveProcess)
        return result.order_by(ActiveProcess.name)

    @requires("ViewActiveProcess")
    def get_active_processes_by_ids(self, ids):
        from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess

        if not ids:
            return EmptyResultSet()
        result = self._find(ActiveProcess, ActiveProcess.id.is_in(ids))
        return result.order_by(ActiveProcess.pid)

    @requires("ViewActiveProcess")
    def get_active_processes_by_pids(self, pids):
        from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess

        pids = list(pids)
        if not pids:
            return EmptyResultSet()
        result = self._find(ActiveProcess, ActiveProcess.pid.is_in(pids))
        return result.order_by(ActiveProcess.pid)

    @requires("RemoveActiveProcess")
    def remove_active_processes(self):
        from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess

        self._find(ActiveProcess).remove()

    @requires("RemoveActiveProcess")
    def remove_active_processes_with_pids(self, pids):
        from canonical.landscape.model.resource.activeprocessinfo import ActiveProcess

        self._find(ActiveProcess, ActiveProcess.pid.is_in(pids)).remove()

    @requires("CreateHistoricProcess")
    def create_historic_process(
        self, pid, uid, gid, name, exit_code, start_time, stop_time
    ):
        from canonical.landscape.model.resource.historicprocess import HistoricProcess

        process = HistoricProcess(
            self, pid, uid, gid, name, exit_code, start_time, stop_time
        )
        self.store.add(process)
        return process

    @requires("ViewHistoricProcess")
    def get_historic_processes(self):
        from canonical.landscape.model.resource.historicprocess import HistoricProcess

        result = self._find(HistoricProcess)
        return result.order_by(HistoricProcess.start_time)

    @requires("CreateMountInfo")
    def create_mount_info(self, device, mount_point, filesystem, total_space):
        from canonical.landscape.model.resource.mountinfo import MountInfo

        minfo = MountInfo(self, device, mount_point, filesystem, total_space)
        self.store.add(minfo)
        return minfo

    @requires("ViewMountInfo")
    def get_mount_infos(self):
        from canonical.landscape.model.resource.mountinfo import MountInfo

        result = self._find(MountInfo)
        return result.order_by(MountInfo.mount_point)

    @requires("ViewMountInfo")
    def get_mount_info_for_mount_point(self, mount_point):
        from canonical.landscape.model.resource.mountinfo import MountInfo

        return self._find(MountInfo, MountInfo.mount_point == mount_point).one()

    @requires("ViewMountInfo")
    def get_mount_info_for_device(self, device_name):
        from canonical.landscape.model.resource.mountinfo import MountInfo

        return self._find(MountInfo, MountInfo.device == device_name).one()

    @requires("EditTags")
    def add_tag(self, tag, broadcast=True):
        """Associate a L{Tag} with this computer.

        @param tag: Tag object
        @param broadcast: Optionally, a flag that specified whether or not an
            L{AddTagsEvent} is broadcast.  Defaults to True.
        @raises EmptyTagError: Raised when an empty name or None is used.
        """
        # FIXME: This should only accept Tag objects eventually.
        if not tag:
            raise EmptyTagError("Tag must have a name.")

        if not is_valid_tag(tag.name):
            raise InvalidTagError(f"Invalid Tag name. {tag.name}")

        tag_ids = []
        store = Store.of(self)
        if not self.has_tags([tag]):
            store.execute("INSERT INTO computer_tag VALUES (?, ?)", (self.id, tag.id))
        if broadcast:
            tag_ids = [tag.id]
            event = AddTagsEvent(self.account, tag_ids, [self.id])
            notify(event)

    @requires("EditTags")
    def remove_tag(self, tag, broadcast=True):
        """Dissociate a L{Tag} from this computer.

        @param tag: Tag object
        @param broadcast: Optionally, a flag that specified whether or not an
            L{RemoveTagsEvent} is broadcast.  Defaults to True.
        """
        if broadcast:
            tag_ids = [tag.id]
            event = RemoveTagsEvent(self.account, tag_ids, [self.id])
            notify(event)

        self._tags.clear(Tag.id == tag.id)

    @requires("ViewComputer")
    def get_tags(self):
        return list(self._tags.order_by(Tag.name))

    @requires("ViewComputer")
    def has_tags(self, tags):
        return set(tags) <= set(self.get_tags())

    @requires("CreateUser")
    def create_user(
        self,
        uid,
        username,
        primary_gid=None,
        name=None,
        enabled=True,
        location=None,
        home_phone=None,
        work_phone=None,
    ):
        from canonical.landscape.model.resource.user import User

        user = User(
            self,
            uid,
            username,
            primary_gid,
            name,
            enabled,
            location,
            home_phone,
            work_phone,
        )
        self.store.add(user)
        return user

    @requires("RequestCreateUser")
    def request_create_user(
        self,
        parent,
        username,
        name,
        password,
        require_passphrase_reset=False,
        primary_groupname=None,
        location=None,
        home_phone=None,
        work_phone=None,
        deliver_after=None,
    ):
        """Create an L{Activity} for a L{CreateUserRequest}."""
        from canonical.landscape.model.users.activities import CreateUserRequest

        child = parent.add(
            CreateUserRequest,
            computer=self,
            username=username,
            name=name,
            passphrase=password,
            require_passphrase_reset=require_passphrase_reset,
            primary_groupname=primary_groupname,
            location=location,
            home_phone=home_phone,
            work_phone=work_phone,
        )
        if deliver_after:
            child.schedule(after=deliver_after)
        EventLog(self.account, parent.creator).create_log_entry(
            self, f"Activity created: {child.info.get_summary()}"
        )
        return child

    @requires("ViewUser")
    def get_users(self):
        from canonical.landscape.model.resource.user import User

        result = self._find(User)
        result.order_by(User.username)
        return result

    @requires("RemoveUser")
    def remove_user(self, user):
        from canonical.landscape.model.resource.user import User

        assert isinstance(user, User), repr(type(user))
        assert self == user.computer, repr(user.computer)
        self.store.remove(user)

    @requires("RemoveUser")
    def remove_users(self):
        from canonical.landscape.model.resource.user import User

        self._find(User).remove()

    @requires("RemoveUser")
    def remove_users_by_names(self, names):
        from canonical.landscape.model.resource.user import User

        self._find(User, User.username.is_in(names)).remove()

    @requires("CreateGroup")
    def create_group(self, gid, name):
        """
        Create a L{canonical.landscape.model.resource.Group}
        associated with this computer, with the given GID and name.
        """
        from canonical.landscape.model.resource.group import Group

        group = Group(self, gid, name)
        self.store.add(group)
        return group

    @requires("CreateGroup")
    def request_create_group(self, parent, groupname, system=False):
        """Create an L{Activity} for a L{CreateUserRequest}."""
        from canonical.landscape.model.users.activities import CreateGroupRequest

        return parent.add(
            CreateGroupRequest, computer=self, groupname=groupname, system=system
        )

    @requires("ViewGroup")
    def get_groups(self):
        """
        Get all L{canonical.landscape.model.resource.Group} objects
        associated with this computer.
        """
        from canonical.landscape.model.resource.group import Group

        return self._find(Group).order_by(Group.gid)

    @requires("ViewGroup")
    def get_group_by_name(self, name):
        from canonical.landscape.model.resource.group import Group

        return self._find(Group, name=name).one()

    @requires("ViewGroup")
    def get_groups_by_names(self, names):
        from canonical.landscape.model.resource.group import Group

        return self._find(Group, Group.name.is_in(names))

    @requires("ViewGroup")
    def get_group_by_gid(self, gid):
        """Return the first group matching the gid, ordered by name.

        NOTE: This the IEEE Std 1003.1 spec states that getgrgid:
        "shall search the group database for an entry with a matching gid."

        And from testing, it appears that it returns the first one in the file.
        """
        from canonical.landscape.model.resource.group import Group

        return self._find(Group, gid=gid).order_by(Group.name).first()

    @requires("ViewGroup")
    def get_groups_by_gids(self, gids):
        from canonical.landscape.model.resource.group import Group

        return self._find(Group, Group.gid.is_in(gids))

    @requires("ViewUser")
    def get_user_by_username(self, username):
        from canonical.landscape.model.resource.user import User

        return self._find(User, User.username == username).one()

    @requires("ViewUser")
    def get_users_by_usernames(self, usernames):
        from canonical.landscape.model.resource.user import User

        return self._find(User, User.username.is_in(usernames))

    @requires("ViewUser")
    def get_user_by_uid(self, uid):
        """Return the first user matching the uid, ordered by name.

        NOTE: This the IEEE Std 1003.1 spec states that getpwuid:
        "shall search the user database for an entry with a matching uid."

        And from testing, it appears that it returns the first one in the file.
        """
        from canonical.landscape.model.resource.user import User

        return self._find(User, User.uid == uid).order_by(User.username).first()

    @requires("ViewUser")
    def get_users_by_uids(self, uids):
        from canonical.landscape.model.resource.user import User

        return self._find(User, User.uid.is_in(uids))

    @requires("RemoveGroup")
    def remove_group(self, group):
        from canonical.landscape.model.resource.group import Group

        assert isinstance(group, Group), repr(type(group))
        assert self == group.computer, repr(group.computer)
        self.store.remove(group)

    @requires("RemoveGroup")
    def remove_groups(self):
        from canonical.landscape.model.resource.group import Group

        self._find(Group).remove()

    @requires("RemoveGroup")
    def remove_groups_by_names(self, names):
        from canonical.landscape.model.resource.group import Group

        self._find(Group, Group.name.is_in(names)).remove()

    @requires("CreateActivity")
    def create_activity(self, info_class, **kwargs):
        """Create a new activity using the given info_class and kwargs.

        This method is deprecated and is used for testing only.

        It automatically creates a parent if one isn't provided to deal
        with legacy tests that didn't care about activity parents.
        """
        from canonical.landscape.model.activity.testing import StubInfo

        parent = kwargs.pop("parent", None)
        creator = kwargs.pop("creator", None)

        should_create_parent = parent is None
        if should_create_parent:
            parent = self.account.create_activity(StubInfo, creator=creator)
        child = parent.add(info_class, computer=self, **kwargs)
        if should_create_parent:
            parent.settle()
        return child

    @requires("ViewComputerPackages")
    def get_packages(self):
        from canonical.landscape.model.resource.package import ComputerPackages

        return self._find(ComputerPackages).one()

    @requires("ViewComputer")
    def get_activities(self, info_classes=None, statuses=None):
        """Get a ResultSet of ActivityParents for this computer.

        @param info_classes: filter activities by child info classes.
        @param statuses: filter activities by child status.
        """
        from canonical.landscape.model.computer.activities import (
            get_computer_activities_parents,
        )

        result = get_computer_activities_parents(
            self.account, [self.id], statuses=statuses, info_classes=info_classes
        )
        return result.order_by(Desc(ActivityParent.creation_time))

    @requires("ViewActivity")
    def get_pending_activities(self):
        """Return pending activities for the Computer."""
        from canonical.landscape.model.computer.activities import (
            get_computer_activities_parents,
        )

        return get_computer_activities_parents(
            self.account, [self.id], statuses=PENDING_STATUSES
        )

    @requires("UnblockActivity")
    def unblock_change_repository_profile_requests(self):
        collection = ActivityChildCollection(
            self.account.store,
            account=self.account,
            statuses=[BLOCKED],
            info_classes=[ChangeRepositoryProfilesRequest],
            entity_class=AccountComputerTwin,
            entity_ids=[self.id],
        )
        for child in collection.find():
            child.unblock()

    @requires("StartActivity")
    def get_deliverable_activities(self, time=None):
        """
        Get all top-level ComputerRequest Activities which are in the
        QUEUED state, or in the SCHEDULED state, where the time for
        delivery is before now (or the optional time parameter).
        """
        from canonical.landscape.model.activity.activity import Activity
        from canonical.landscape.model.activity.collection import (
            ActivityChildCollection,
        )

        time = time if time is not None else datetime.utcnow()

        collection = ActivityChildCollection(
            self.account.store,
            account=self.account,
            statuses=[QUEUED, SCHEDULED],
            entity_class=AccountComputerTwin,
            entity_ids=[self.id],
        )
        result = collection.find()
        result = result.find(
            Coalesce(Activity.schedule_before_time, time) >= time,
            Coalesce(Activity.schedule_after_time, time) <= time,
        )
        result.order_by(Activity.creation_time)
        return result

    @requires("ResynchronizeComputer")
    def request_resynchronize(self, parent, scopes=None):
        """Tell the client to perform a resynchronization.

        @param parent: The parent activity to use.
        @param scopes: A list of scopes for restricting what data will be
            resynchronized. If C{None} all data will be resynchronized.
        """
        from canonical.landscape.model.activity.resynchronize import (
            ResynchronizeRequest,
        )

        if not self.is_api_compatible("3.6"):
            # Older clients don't support scoped resyncs
            scopes = None
        child = parent.add(ResynchronizeRequest, computer=self, scopes=scopes)
        EventLog(self.account, parent.creator).create_log_entry(
            self, f"Activity created: {child.info.get_summary()}"
        )
        return child

    @requires("ResynchronizeComputer")
    def resynchronize(self, activity_id=None):
        """Handle a request by the client to resynchronize.

        Fire a L{ResynchronizeEvent}, which other code should handle
        mostly by cleaning up state.

        @param activity_id: If specified, the activity ID of an
            existing L{ResynchronizeRequest} activity. The associated
            activity will be marked as having succeeded.
        """
        event_log = EventLog(self.account)
        event_log.create_log_entry(self, "Resynchronize fired on the client.")
        if activity_id:
            activity = get_child_activity(self.account, activity_id)
            if activity is None:
                raise InvalidActivityError
            activity.succeed()
            scopes = activity.info.scopes
        else:
            scopes = None
        notify(ResynchronizeEvent(self, scopes=scopes))

        # Recheck all rules.
        for rule in self.account.get_rules(computers=[self]):
            rule.info.check([self.id])

    @requires("ExecuteScript")
    def request_execute_script(
        self, parent, title, interpreter, code, username, time_limit, deliver_after=None
    ):
        """Request that a script be executed on this computer."""
        from canonical.landscape.model.activity.activities import ExecuteScriptRequest

        child = parent.add(
            ExecuteScriptRequest,
            computer=self,
            title=title,
            interpreter=interpreter,
            code=code,
            username=username,
            time_limit=time_limit,
        )
        if deliver_after:
            child.schedule(after=deliver_after)
        EventLog(self.account, parent.creator).create_log_entry(
            self, f"Activity created: {child.info.get_summary()}"
        )
        return child

    def create_rule_states(self):
        """Create L{RuleState}s for existing computer-based alert rules."""
        from canonical.landscape.model.alert.policy import AlertRulePolicy

        policy = AlertRulePolicy()
        policy.create_rules_for_computer(self)

    @requires("CheckRule")
    def check_rules(self, info_class):
        """
        Given a L{RuleInfo} subclass, check all L{Rule}s which this
        L{Computer} has a L{RuleState} associated with.

        @param info_class: The type of the L{RuleInfo} to check this computer
            for.
        """
        from canonical.landscape.model.alert.rule import Rule, RuleState

        result = self.account.store.find(
            Rule,
            Rule.info_type == info_class.get_info_type(),
            RuleState.rule_id == Rule.id,
            RuleState.computer_id == self.id,
        )
        result.config(distinct=True)
        for rule in result:
            rule.info.check([self.id])

    @requires("RestartComputer")
    def restart(self, parent, deliver_after=None):
        """Restart this computer.

        @param creator: The L{Person} that requested the restart.
        @param parent: Optionally, an L{Activity} to use as the parent for the
            L{RestartRequest} created by this method.
        @param deliver_after: Optionally, a C{datetime} representing the
            earliest time the L{Activity}s created to carry out this request
            should be delivered to this L{Computer}.  Defaults to immediate
            delivery.
        """
        from canonical.landscape.model.activity.shutdown import RestartRequest

        child = parent.add(RestartRequest, computer=self)
        if deliver_after:
            child.schedule(after=deliver_after)
        EventLog(self.account, parent.creator).create_log_entry(
            self, f"Activity created: {child.info.get_summary()}"
        )
        return child

    @requires("ShutdownComputer")
    def shutdown(self, parent, deliver_after=None):
        """Shutdown this computer.

        @param parent: The Activity to use as the parent for the
            L{ShutdownRequest} created by this method.
        @param deliver_after: Optionally, a C{datetime} representing the
            earliest time the L{Activity}s created to carry out this request
            should be delivered to this L{Computer}.  Defaults to immediate
            delivery.
        """
        from canonical.landscape.model.activity.shutdown import ShutdownRequest

        child = parent.add(ShutdownRequest, computer=self)
        if deliver_after:
            child.schedule(after=deliver_after)
        EventLog(self.account, parent.creator).create_log_entry(
            self, f"Activity created: {child.info.get_summary()}"
        )
        return child

    @requires("EditComputerPackages")
    def request_release_upgrade(self, parent):
        """Create an L{Activity} for a L{ReleaseUpgradeRequest}.

        @param creator: The L{Person} creating the activity.
        @param parent: Optionally, the parent of the created activity.
        """
        from canonical.landscape.model.package.activities import ReleaseUpgradeRequest

        meta_release = self.distribution.meta_release
        if not meta_release or not meta_release.upgrade:
            raise RuntimeError("Release upgrade not available")

        child = parent.add(
            ReleaseUpgradeRequest,
            computer=self,
            from_release=meta_release,
            to_release=meta_release.upgrade,
        )
        EventLog(self.account, parent.creator).create_log_entry(
            self, f"Activity created: {child.info.get_summary()}"
        )

        # The release upgrade process will fail for computers with
        # old versions of landscape-client. We cancel any release upgrade
        # activity that involves such a computer and for any computer
        # with an old release.
        should_cancel = False
        message = ""
        if not self.release_upgrade_supported():
            release = self.distribution.release if self.distribution else "Unknown"
            message = (
                "Release upgrade only supported for 16.04 "
                f"and later. This computer has {release}"
            )
            should_cancel = True
        elif self.should_upgrade_client_before_release_upgrade():
            current_version = (
                self.current_version.decode() if self.current_version else "Unknown"
            )
            message = f"Please upgrade landscape-client first. Current version is {current_version}"
            should_cancel = True

        if should_cancel:
            child = removeSecurityProxy(child)
            child.queue()
            child.cancel(text=message)
            EventLog(self.account, parent.creator).create_log_entry(
                self, f"Activity canceled: {child.info.get_summary()}"
            )

        return child

    @requires("ViewCustomGraph")
    def get_custom_graph_data_by_id(self, id):
        """
        Returns a CustomGraphData from the resource database.
        """
        from canonical.landscape.model.resource.customgraph import CustomGraphData

        return self._find(CustomGraphData, CustomGraphData.custom_graph_id == id).one()

    @requires("ViewCustomGraph")
    def get_custom_graph_datas_by_ids(self, ids):
        """
        Returns a CustomGraphData from the resource database.
        """
        from canonical.landscape.model.resource.customgraph import CustomGraphData

        result = self._find(CustomGraphData, CustomGraphData.custom_graph_id.is_in(ids))
        # Arbitrarily order by ID to get a stable order
        return result.order_by(CustomGraphData.id)

    @requires("ViewCustomGraph")
    def get_custom_graph_datas(self):
        """
        Returns all CustomGraphData for this computer from the resource
        database.
        """
        from canonical.landscape.model.resource.customgraph import CustomGraphData

        # Arbitrarily order by ID to get a stable order
        return self._find(CustomGraphData).order_by(CustomGraphData.id)

    @requires("ViewComputer")
    def get_account_twin(self):
        """
        Return the ComputerTwin in the Account database for this computer's
        account.
        """
        return self.account.store.find(
            AccountComputerTwin, AccountComputerTwin.id == self.id
        ).one()

    def get_resource_twin(self):
        """
        Return the ComputerTwin in the Resource database for this computer's
        account.
        """
        return self.store.find(ComputerTwin, ComputerTwin.id == self.id).one()

    @requires("ViewCustomGraph")
    def get_custom_graphs(self):
        """
        Return all custom graphs associated with this Computer.
        """
        from canonical.landscape.model.account.customgraph import CustomGraph
        from canonical.landscape.model.account.computergroup import (
            ComputerGroupItem as GroupItem,
        )

        return self.account.store.find(
            CustomGraph,
            CustomGraph.computer_group_id == GroupItem.computer_group_id,
            GroupItem.computer_id == self.id,
        )

    def is_api_compatible(self, version):
        """Compare the current version of a computer with a version.


        This takes the version and splits it on the decimal character,
        converting both parts to ints, so... "3.2" becomes (3, 2).

        It then does a comparison of the two tuples, and returns true if
        ver2 (the version) is greater than or equal to Computer.current_api.

        We may need to inject 0.0 into the calculation in case the value has
        yet to be populated.
        """
        version = version.encode() if type(version) == str else version
        versions = (self.current_api or b"0.0", version)
        ver1, ver2 = [tuple(map(int, s.split(b"."))) for s in versions]
        return ver1 >= ver2

    def get_state(self):
        """
        @return a dict with L{Computer} information.
        """
        info = (
            "id",
            "title",
            "comment",
            "hostname",
            "total_memory",
            "total_swap",
            "reboot_required_flag",
            "update_manager_prompt",
            "clone_id",
        )

        state = dict((field, getattr(self, field)) for field in info)
        state["last_exchange_time"] = format_datetime(self.last_exchange_time)
        state["last_ping_time"] = format_datetime(self.last_ping_time)
        state["tags"] = [tag.name for tag in self.get_tags()]
        state["access_group"] = self.access_context.name
        if self.distribution is None:
            state["distribution"] = None
        else:
            state["distribution"] = self.distribution.release

        management = CloudInstanceMetadataManagement(self)
        state["cloud_instance_metadata"] = management.get_instance_metadata()
        state["vm_info"] = self.vm_info
        state["container_info"] = self.container_info

        return state

    @requires("ViewComputer")
    def release_upgrade_supported(self):
        """
        Returns True if the release is Ubuntu 16.04 or later, otherwise False.
        """
        if (
            self.distribution
            and self.distribution.distributor == "Ubuntu"
            and self.distribution.release >= "16.04"
        ):
            return True
        return False

    @requires("ViewComputer")
    def should_upgrade_client_before_release_upgrade(self):
        """
        The release upgrade process will fail for computers with
        old versions of landscape-client.

        Returns False only if the Ubuntu release is at least 16.04 and
        an appropriate version of landscape-client is installed,
        otherwise returns True.
        """
        if not self.distribution:
            return True

        if self.distribution.distributor != "Ubuntu":
            return True

        if self.distribution.release < "16.04":
            return True

        # We get the version of landscape-client installed from
        # the database because some versions of landscape-client
        # do not report this to us correctly in the User-Agent
        computer_packages = self.get_packages()
        [installed_package_ids] = computer_packages.get_packages("installed")
        store = get_package_store()
        select = Select(
            Package.version,
            where=And(
                Package.name == "landscape-client",
                Package.id.is_in(installed_package_ids),
            ),
        )
        result = store.execute(select).get_one()
        if not result:
            return True

        current_landscape_client_version = result[0]

        required_landscape_client_version = "18.03"  # xenial
        if self.distribution.release >= "18.04":
            required_landscape_client_version = "23.02"  # bionic and newer

        return current_landscape_client_version < required_landscape_client_version

    @requires("ViewComputer")
    def start_instance_supported(self):
        """
        Whether or not the computer supports the "start" GRPC call.
        """
        parent_computer = self.get_parent_computer()
        if not parent_computer:
            return False
        if parent_computer.distribution.code_name != "windows":
            return False
        return True

    @requires("ViewComputer")
    def stop_instance_supported(self):
        """
        Whether or not the computer supports the "stop" GRPC call.
        """
        return self.start_instance_supported()

    @requires("ViewComputer")
    def delete_instance_supported(self):
        """
        Whether or not the computer supports the "delete" GRPC call.
        """
        return self.start_instance_supported()

    @requires("ViewComputer")
    def get_parent_computer(self):
        origin = [
            Computer,
            Join(ComputerRelationship, ComputerRelationship.parent_id == Computer.id),
        ]
        store = get_main_store()
        parent_computer = (
            store.using(*origin)
            .find(Computer, ComputerRelationship.computer_id == self.id)
            .one()
        )

        return parent_computer

    @requires("RestartComputer")
    def delete_instance(self, activity_parent):
        """Create an L{Activity} for a L{DeleteChildComputerActivity}."""
        activity_child = activity_parent.add(
            DeleteChildComputerActivity,
            computer=self,
        )
        EventLog(self.account, activity_parent.creator).create_log_entry(
            self, f"Activity created: {activity_child.info.get_summary()}"
        )

        return activity_child

    @requires("RestartComputer")
    def start_instance(self, activity_parent):
        """Create an L{Activity} for a L{StartChildComputerActivity}."""
        activity_child = activity_parent.add(
            StartChildComputerActivity,
            computer=self,
        )
        EventLog(self.account, activity_parent.creator).create_log_entry(
            self, f"Activity created: {activity_child.info.get_summary()}"
        )

        return activity_child

    @requires("ShutdownComputer")
    def stop_instance(self, activity_parent):
        """Create an L{Activity} for a L{StartChildComputerActivity}."""
        activity_child = activity_parent.add(
            StopChildComputerActivity,
            computer=self,
        )
        EventLog(self.account, activity_parent.creator).create_log_entry(
            self, f"Activity created: {activity_child.info.get_summary()}"
        )

        return activity_child


class PingTime(object):
    """
    A ping time represents the last time a client has sent the Landscape server
    a 'ping'.
    """

    __storm_table__ = "ping_time"
    computer_id = Int(allow_none=False, primary=True)
    computer = Reference(computer_id, Computer.id)
    last_ping_time = DateTime()

    requires("ViewComputer", PRIVATE, ["computer_id", "last_ping_time"])


class PingServer(object):
    """
    Represents the last time that any ping server was ever seen alive.
    """

    __storm_table__ = "ping_server"
    id = Int(primary=True, allow_none=False, default=AutoReload)
    startup_time = DateTime(allow_none=False, default=AutoReload)
    last_update_time = DateTime()

    requires("ViewComputer", PRIVATE, ["id", "startup_time", "last_update_time"])


class PendingComputerToAccept(
    namedtuple("PendingComputerToAccept", "computer existing access_context")
):
    """Identifies a pending computer to accept."""

    def __new__(cls, computer, existing=None, access_context=None):
        self = super(PendingComputerToAccept, cls).__new__(
            cls, computer, existing, access_context
        )
        return self

    def accept(self, person):
        """Accept this pending computer."""
        return self.computer.accept(person, self.existing, self.access_context)


@requires_advisor
@implementer(IPendingComputer, IAccessContextAware)
class PendingComputer(object):
    """Track an unregistered computer that has contacted Landscape."""

    __storm_table__ = "pending_computer"

    # FIXME Add IP address of where the request came from.

    id = Int(primary=True, allow_none=False, default=AutoReload)
    title = Unicode(allow_none=False)
    description = Unicode()
    hostname = Unicode()
    creation_time = DateTime(allow_none=False)
    secure_id = BytesStr(allow_none=False)
    account_id = Int(allow_none=False)
    client_tags = Unicode()
    vm_info = Unicode()
    container_info = Unicode()
    access_group = Unicode()
    juju_model_uuid = Unicode()
    juju_machine_id = Unicode()
    account = Reference(account_id, Account.id)
    clone_id = Int(allow_none=True)
    ubuntu_pro_info = Unicode()
    computer = None
    _check_pending = True

    requires(
        "ViewPendingComputer",
        PRIVATE,
        [
            "id",
            "secure_id",
            "account_id",
            "account",
            "creation_time",
            "hostname",
            "vm_info",
            "container_info",
            "access_group",
            "juju_model_uuid",
            "juju_machine_id",
            "clone_id",
            "ubuntu_pro_info",
        ],
    )

    # It looks pointless that these attributes get mutated (and thus need
    # protected by Reject), since the whole PendingComputer is
    # immediately thrown out (as it is accepted) as soon as changed.
    # However, accepting is relative to the context in which the new computer
    # will be, and edition happens on global account entity. Thus, we use
    # the global RejectPendingComputer perm for editing.
    requires(
        "ViewPendingComputer",
        "RejectPendingComputer",
        [
            "title",
            "description",
            "computer",
            "hostname",
            "client_tags",
            "access_group",
            "juju_model_uuid",
            "juju_machine_id",
            "clone_id",
        ],
    )

    def __init__(
        self,
        account,
        title,
        description,
        hostname=None,
        tags=None,
        vm_info=None,
        check_pending=True,
        container_info=None,
        access_group=None,
        juju_model_uuid=None,
        juju_machine_id=None,
        clone_id=None,
        ubuntu_pro_info=None,
    ):
        self.account = account
        self.title = title
        self.description = description
        self.hostname = hostname
        self.client_tags = tags
        self.vm_info = vm_info
        self.creation_time = datetime.utcnow()
        self.secure_id = make_secure_id()
        self.container_info = container_info
        self.access_group = access_group
        self.juju_model_uuid = juju_model_uuid
        self.juju_machine_id = juju_machine_id
        self.clone_id = clone_id
        self.ubuntu_pro_info = ubuntu_pro_info
        # This attribute is volatile, only used when we accept the computer
        # right away
        self._check_pending = check_pending
        self._check_rules(self.account)

    def _check_rules(self, account, check_account_limits=False):
        """Update L{PendingComputersAlert}s rule state."""
        from canonical.landscape.model.account.alerts import (
            PendingComputersAlert,
            StandaloneOverflowAlert,
            LicenseExpiryAlert,
        )

        alerts = []
        if self._check_pending:
            alerts.append(PendingComputersAlert)
        if check_account_limits:
            alerts.append(StandaloneOverflowAlert)
            alerts.append(LicenseExpiryAlert)
        if alerts:
            account.check_rules(alerts)

    @requires("ViewAccount")
    def accept(self, person, computer=None, access_context=None):
        """
        Accept this pending computer.

        Create a real Computer object with the same secure ID as this pending
        computer and delete this computer.

        @param computer: If not None, the client that this pending computer
            represents will be associated with that existing computer; its
            title, secure_id, insecure_id will be changed to those of the
            pending computer and its outbound message store will be cleaned up,
            to prevent any old messages from being sent to the new client.
        @param access_context: The access context to place the new computer
            in. If None, it will be associated with the requested access
            context (named in access_group) or the global access context if
            access_group is invalid.
        """
        if is_standalone():
            # Only check the license for LDS deployments.
            check_lds_license()

        if access_context is not None and person is None:
            raise AcceptanceUnauthorized(
                "Must pass a person if passing an access context"
            )
        if person is not None:
            if access_context is None:
                access_context = self.access_context
            if not person.has_permission(
                "landscape.AcceptPendingComputer", access_context
            ):
                raise AcceptanceUnauthorized(
                    f"{person.name} cannot accept pending computers "
                    f"to access group {access_context.name}"
                )

        # If no access context has yet to be chosen and the pending computer
        # requested a particular access context (by giving its name in
        # "access_group" at registration time) then that access context is
        # used if available.
        if access_context is None and self.access_group is not None:
            try:
                [access_context] = self.account.get_access_contexts(
                    names=[self.access_group]
                )
            except UnknownAccessContextsError:
                # The requested access context is not valid.
                pass

        if computer is not None:
            computer.secure_id = self.secure_id
            computer.title = self.title
            computer.hostname = self.hostname
            computer.vm_info = self.vm_info
            computer.container_info = self.container_info
            computer.juju_model_uuid = self.juju_model_uuid
            computer.juju_machine_id = self.juju_machine_id
            computer.insecure_id = self.id
            computer.get_message_store().delete_old_messages()
            computer.next_expected_sequence = None
            computer.access_context = access_context
            computer.clone_id = self.clone_id
            computer.ubuntu_pro_info = self.ubuntu_pro_info
            computer.resynchronize()
        else:
            computer = self.account.create_computer(
                self.secure_id.decode(),
                self.title,
                access_context=access_context,
                insecure_id=self.id,
                hostname=self.hostname,
                last_exchange_time=self.creation_time,
                vm_info=self.vm_info,
                container_info=self.container_info,
                juju_model_uuid=self.juju_model_uuid,
                juju_machine_id=self.juju_machine_id,
                clone_id=self.clone_id,
                ubuntu_pro_info=self.ubuntu_pro_info,
            )
            computer.create_rule_states()
        # If we have valid tags then we create them and add them to the
        # computer
        if self.client_tags:
            tags_to_add = [tag.strip() for tag in self.client_tags.split(",")]
            tags = get_tags(
                [tagname for tagname in tags_to_add if is_valid_tag(tagname)]
            )
            for tag in tags:
                computer.add_tag(tag)

        message_store = computer.get_message_store()
        message_store.add({"type": "registration-complete"})
        event_log = EventLog(self.account, person)
        event_log.create_log_entry(self, "Pending computer accepted.")
        Store.of(self).remove(self)
        self._check_rules(self.account, check_account_limits=True)
        return computer

    @requires("RejectPendingComputer")
    def reject(self, person):
        """
        Reject this pending computer.  Don't create a real Computer
        and delete this computer.
        """
        event_log = EventLog(self.account, person)
        event_log.create_log_entry(self, "Pending computer rejected.")
        # Cache the account to pass to _check_rules.  We can't use
        # self.account because it'll be gone when remove self.
        account = self.account
        Store.of(self).remove(self)
        self._check_rules(account)

    @property
    def access_context(self):
        return self.account.access_context

    def get_state(self):
        """
        Return a dictionary representation of this PendingComputer.
        """
        try:
            tag_names = get_tag_names(self.client_tags)
        except InvalidTagError:
            tag_names = []
        return {
            "id": self.id,
            "title": self.title,
            "hostname": self.hostname,
            "creation_time": format_datetime(self.creation_time),
            "vm_info": self.vm_info,
            "container_info": self.container_info,
            "client_tags": tag_names,
            "access_group": self.access_group,
        }


def get_computer_by_id(id):
    return get_main_store().get(Computer, id)


def get_computers_by_ids(ids):
    return get_main_store().find(Computer, Computer.id.is_in(ids))


@arg_to_bytes
def get_computer_by_secure_id(secure_id):
    return get_main_store().find(Computer, secure_id=secure_id).one()


def get_computer_by_juju_machine_id(juju_model, machine_id):
    """
    Return the L{Computer} with the specified machine ID in the account
    associated to the juju model.
    """
    result = _get_computers_by_juju_model_uuid(juju_model)
    return result.find(Computer.juju_machine_id == machine_id).one()


def get_computer_ids_by_unknown_juju_machine_ids(juju_model, machine_ids):
    """
    Return the L{Computer.id}s with specified machine IDs in the account
    associated to the juju model.
    """
    result = _get_computers_by_juju_model_uuid(juju_model)
    result = result.find(Not(Computer.juju_machine_id.is_in(machine_ids)))
    return list(result.values(Computer.id))


def _get_computers_by_juju_model_uuid(juju_model):
    """
    Return a ResultSet of Computers in the account with the given
    Juju model UUID.
    """
    return get_main_store().find(
        Computer, account_id=juju_model.account_id, juju_model_uuid=juju_model.uuid
    )


def get_wedged_computers():
    """Fetch all computers that have more than one consecutive error."""
    store = get_main_store()
    result = store.find(
        Computer,
        Computer.account_id == Account.id,
        Computer.id == ComputerStatus.computer_id,
        ComputerStatus.consecutive_errors >= 2,
    )
    result.order_by(Account.title, Computer.title)
    return result


@arg_to_bytes
def get_pending_computer_by_secure_id(secure_id):
    return get_main_store().find(PendingComputer, secure_id=secure_id).one()


def get_pending_computer_by_id(id):
    return get_main_store().find(PendingComputer, id=id).one()


def get_pending_computers_by_ids(ids):
    return get_main_store().find(PendingComputer, PendingComputer.id.is_in(ids))


def get_pending_computers_by_juju_model_uuid(account, model_uuid):
    """
    Get all PendingComputers associated to the given Juju model UUID.
    """
    return get_main_store().find(
        PendingComputer, account_id=account.id, juju_model_uuid=model_uuid
    )


def cleanup_after_removed_computers(account, computer_ids):
    """
    Check account rules after computers have been deleted, and create a job to
    remove their resource data through a job.
    """
    from canonical.landscape.model.account.alerts import (
        StandaloneOverflowAlert,
        LicenseExpiryAlert,
    )
    from canonical.landscape.model.computer.alerts import ComputerDuplicateAlert

    account.check_rules([StandaloneOverflowAlert, LicenseExpiryAlert])
    notify(AccountRulesChangedEvent(account))
    account.check_rules([ComputerDuplicateAlert], computer_ids=computer_ids)
    # Create a job for removing resource data for computers.
    dispatcher = getUtility(IDispatcher)
    dispatcher.create_job(
        page_key=None,
        namespace=None,
        action="remove-computer-resources",
        computer_ids=computer_ids,
    )


License.computers = ReferenceSet(
    License.id, ComputerLicense.license_id, ComputerLicense.computer_id, Computer.id
)

Account.computers = ReferenceSet(Account.id, Computer.account_id)

AccessContext.computers = ReferenceSet(AccessContext.id, Computer.access_context_id)

Computer._tags = ReferenceSet(
    Computer.id, ComputerTag.computer_id, ComputerTag.tag_id, Tag.id
)


@implementer(IContextSourceBinder)
class ComputersForPendingComputersVocabularyFactory(object):
    """
    A factory for a vocabulary which includes all computers that the specified
    Person has the C{landscape.AcceptPendingComputer} permission on. This is
    used in the pending computer acceptance page to show the list of computers
    that can be re-associated with a pending computer.

    XXX: This vocabulary can return ALL computers in the account. This should
    be changed so we don't have to load / display all the computers at once; it
    should probably be replaced with a computer selection popup. See bug
    #1037175.
    """

    def __init__(self, account, person):
        self.account = account
        self.person = person

    def __call__(self, context):
        computer_ids = self.person.get_computer_ids(
            self.account, permission="landscape.AcceptPendingComputer"
        )
        if computer_ids is not None:
            computers = list(map(get_computer_by_id, computer_ids))
            computers = sorted(computers, key=lambda c: (c.title, c.id))
        else:
            computers = self.account.get_computers().order_by(
                Computer.title, Computer.id
            )
        terms = [
            SimpleTerm(computer.id, str(computer.id), computer.title)
            for computer in computers
        ]
        return SimpleVocabulary(terms)


def add_tags_to_computers(person, tags, computers):
    """Associate L{Computer}s with tags in C{tags}.

    An L{AddTagsEvent} event is broadcast when this function is called.

    @param tags: A sequence of, optionally mixed, tag names, IDs or L{Tag}
        objects.
    @param computers: A sequence of L{Computer} objects.
    """
    if not computers:
        return
    if not tags:
        return

    account = computers[0].account

    # Collect the changes and if any apply to the computer.
    changes = []
    for computer in computers:
        new_tags = set(tags) - set(computer.get_tags())
        if new_tags:
            tag_names = []
            for tag in new_tags:
                computer.add_tag(tag, broadcast=False)
                tag_names.append(tag.name)
            changes.append({"computer": computer, "new_tags": tag_names})

    # Create the log entries and fire events.
    if changes:
        event_log = EventLog(account, person)

        for change in changes:
            event_log.create_log_entry(
                change["computer"],
                "Tags added: " + ", ".join(sorted(change["new_tags"])) + ".",
            )

        event = AddTagsEvent(
            account,
            [tag.id for tag in get_tags(tags)],
            [computer.id for computer in computers],
        )
        notify(event)


def remove_tags_from_computers(person, tags, computers):
    """Dissociate L{Computer}s from tags in C{tags}.

    A L{RemoveTagsEvent} event is broadcast when this function is called.

    @param tags: A sequence of, optionally mixed, tag names, IDs or L{Tag}
        objects.
    @param computers: A sequence of L{Computer} objects.
    """
    if not computers:
        return
    if not tags:
        return

    account = computers[0].account
    event = RemoveTagsEvent(
        account,
        [tag.id for tag in get_tags(tags)],
        [computer.id for computer in computers],
    )
    notify(event)

    tag_names = set([tag.name for tag in tags])
    event_log = EventLog(account, person)
    for computer in computers:
        existing_tag_names = set([tag.name for tag in computer.get_tags()])
        tags_to_remove = sorted(existing_tag_names.intersection(tag_names))
        for tag in tags:
            computer.remove_tag(tag, broadcast=False)
        if tags_to_remove:
            event_log.create_log_entry(
                computer, "Tags removed: " + ", ".join(sorted(tags_to_remove)) + "."
            )


def update_managed_hours(computer_ids):
    """Update the managed hours count for the given computers.

    The number of hours a computer has been managed by Landscape will be
    updated if the computer's next_managed_hour_time already has
    passed. In that case next_managed_hour_time will be increased by
    1 hour, so that we won't increase the count only once every hour, no
    matter how often this method is called.

    @param computer_ids: IDs of computers to update.
    """
    store = get_main_store()
    managed_hours = ComputerUtilisation.managed_hours + 1
    full_hour_difference = Floor(
        DatePart(Now - ComputerUtilisation.next_managed_hour_time, part="epoch") // 3600
    )
    next_managed_hour_time = ComputerUtilisation.next_managed_hour_time + Cast(
        Concat(full_hour_difference + 1, " hours"), "interval"
    )
    update = Update(
        {
            ComputerUtilisation.managed_hours: managed_hours,
            ComputerUtilisation.next_managed_hour_time: next_managed_hour_time,
        },
        table=ComputerUtilisation,
        where=And(
            ComputerUtilisation.next_managed_hour_time < Now,
            ComputerUtilisation.computer_id.is_in(computer_ids),
        ),
    )
    store.execute(update)
    store.flush()


def get_computer_type_expr():
    """Return a Storm expression returning the computer type as string."""
    case = Case(
        [
            (Coalesce(Computer.container_info, "") != "", "container"),
            (Coalesce(Computer.vm_info, "") != "", "vm"),
        ],
        default="physical",
    )
    return Alias(case)


def get_computer_type(computer):
    """Return the type of a computer."""
    if computer.container_info:
        return "container"
    if computer.vm_info:
        return "vm"
    return "physical"


def get_computer_counts_by_type(account_ids, last_ping_time=None):
    """Return counts by computer type for the account_ids provided.

    Optionally limit results to computers which have pinged later than
    last_ping_time.
    """
    counts = dict.fromkeys(COMPUTER_TYPES, 0)
    machine_type = get_computer_type_expr()
    store = get_main_store()
    where = [Computer.account_id.is_in(account_ids)]
    if last_ping_time:
        where.append(PingTime.last_ping_time >= last_ping_time)
        join = LeftJoin(Computer, PingTime, Computer.id == PingTime.computer_id)
        store = store.using(join)
    result = store.find((machine_type, Count()), *where)
    result.group_by(machine_type)
    counts.update(result)
    return counts


class ComputerRelationship:
    """Our knowledge of the parent-child relationships between machines managed
    by Landscape.

    The `tree` field uses the "ltree" datatype in PostgreSQL::

      https://www.postgresql.org/docs/current/ltree.html

    @ivar tree: the label tree comprised of the computer ids that represents the
    hierarchical position of the computer.
    """

    __storm_table__ = "computer_relationship"

    computer_id = Int(primary=True, allow_none=False)
    parent_id = Int()
    tree = Unicode(allow_none=False)

    computer = Reference(computer_id, Computer.id)
    parent = Reference(parent_id, Computer.id)

    def __init__(self, computer: Computer, parent: Optional[Computer] = None):
        if parent is not None and computer.account_id != parent.account_id:
            raise ValueError("Parent and child computer must be in the same account")

        self.computer = computer
        self.parent = parent
        self.tree = str(computer.id)

        parent_tree = self.find_parent_tree()
        self.update_tree(parent_tree)
        self.update_child_trees()

    def find_parent_tree(self):
        """Find the parent computer's tree if it exists."""

        if self.parent_id is None:
            return None

        store = get_main_store()
        result = store.find(
            ComputerRelationship, ComputerRelationship.computer_id == self.parent_id
        )

        if result.is_empty():
            return self.parent_id

        return result.one().tree

    def update_tree(self, parent_tree: Optional[str] = None):
        """Update this computer's tree."""

        if parent_tree is None:
            self.tree = str(self.computer_id)
        else:
            self.tree = f"{parent_tree}.{self.computer_id}"

    def update_child_trees(self):
        """Update the tree of all child computers."""

        store = get_main_store()
        children = store.find(
            ComputerRelationship,
            ComputerRelationship.parent_id == self.computer_id,
        )

        for child in children:
            child.update_tree(self.tree)
            child.update_child_trees()

    def __repr__(self):
        return (
            f"<Computer id={self.computer_id} with parent computer id={self.parent_id}>"
        )
