# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Core L{DistributedLock} objects.
"""

from builtins import object
import logging
from socket import gethostname

from storm.exceptions import TimeoutError
from storm.locals import Int, Unicode, AutoReload, DateTime

from canonical.lib.timestamp import utcnow
from canonical.landscape.model.main.store import get_main_store


__all__ = [
    "DistributedLock",
    "acquire_distributed_lock",
    "release_distributed_lock",
    "create_distributed_lock",
]

STALE_LOCK_TIMEOUT_IN_SECONDS = 7200  # 2 hours


class DistributedLock(object):
    """Hold locks for synchronizing distributed services

    A distributed lock is a service-supplied lock name used to create a
    database row exclusive lock in order to prevent multiple services
    writing to the database at the same time.

    @ivar id: The primary key, automatically generated by the database, for
        this distributed lock instance.
    @type id: Int
    @ivar name: The service-supplied lock name used to identify this
        lock instance.
    @type name: Unicode
    @ivar creation_time: The time when the lock has been created.
    @type creation_time: C{datetime}
    """

    __storm_table__ = "distributed_lock"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    hostname = Unicode(allow_none=False)
    creation_time = DateTime(default=AutoReload)

    def __init__(self, name, hostname):
        self.name = name
        self.hostname = hostname


def create_distributed_lock(name, hostname):
    """Create a lock in the main database with the hostname which owns it"""
    lock = DistributedLock(name, hostname)
    store = get_main_store()
    store.add(lock)
    return lock


def release_distributed_lock(name, timeout=5400):
    """
    Release a distributed lock held by this script
    """
    store = get_main_store()
    # Set a local timeout on this transaction in case the lock is held.
    store.execute("SET LOCAL statement_timeout = %d" % timeout)
    store.execute("LOCK distributed_lock IN ROW EXCLUSIVE MODE")
    result = store.find(DistributedLock, name=name)
    if not result.is_empty():
        result.remove()
        store.commit()


def acquire_distributed_lock(name, timeout=5400, host_getter=None):
    """
    Grab a lock during this script run or timeout on lock held. Return C{True}
    on success, C{False} if lock is held.
    """
    if not host_getter:  # for testing
        host_getter = gethostname

    store = get_main_store()
    # Set a local timeout on this transaction in case the lock is held.
    store.execute("SET LOCAL statement_timeout = %d" % timeout)
    store.execute("LOCK distributed_lock IN ROW EXCLUSIVE MODE")
    lock = store.find(DistributedLock, name=name).one()
    if lock:
        now = utcnow()
        delta = now - lock.creation_time
        if delta.total_seconds() > STALE_LOCK_TIMEOUT_IN_SECONDS:
            logging.warning(
                "Releasing stale distributed lock %s from host %s."
                % (name, lock.hostname)
            )
            release_distributed_lock(name)
        else:
            logging.info("Existing lock '%s' held by '%s'." % (name, lock.hostname))
            return False
    try:
        lock = create_distributed_lock(name, "%s" % host_getter())
        store.add(lock)
        store.commit()
        return True
    except TimeoutError:
        logging.info("Couldn't obtain '%s' lock. Lock held by another computer." % name)
        store.rollback()
        return False
