# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, Unicode, Reference

from canonical.security.protect import requires, requires_advisor, PRIVATE, PUBLIC
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.metarelease import MetaRelease


__all__ = ["Distribution", "create_distribution", "get_distribution"]


__requires_prefix__ = "landscape."


@requires_advisor
class Distribution(object):

    __storm_table__ = "distribution"

    requires(
        PUBLIC,
        PRIVATE,
        [
            "id",
            "distributor",
            "description",
            "release",
            "code_name",
            "meta_release",
            "release_upgrades",
        ],
    )

    id = Int(primary=True, allow_none=False)
    description = Unicode(allow_none=False)
    distributor = Unicode()
    release = Unicode()
    code_name = Unicode()
    meta_release_id = Int()
    meta_release = Reference(meta_release_id, MetaRelease.id)

    def __init__(self, description, distributor, release, code_name):
        self.description = description
        self.distributor = distributor
        self.release = release
        self.code_name = code_name


def create_distribution(description, distributor, release, code_name):
    """Create a L{Distribution}."""
    from canonical.landscape.model.main.metarelease import get_meta_release

    store = get_main_store()
    distribution = Distribution(description, distributor, release, code_name)
    distribution.meta_release = get_meta_release(code_name)
    return store.add(distribution)


def get_distribution(description=None, distributor=None, release=None, code_name=None):
    """Find the L{Distribution} matching the given criteria.

    @param description: Optionally, the L{Distribution.description} to match.
    @param distributor: Optionally, the L{Distribution.distributor} to match.
    @param release: Optionally, the L{Distribution.release} to match.
    @param code_name: Optionally, the L{Distribution.code_name} to match.
    @raises RuntimeError: Raised if no filtering criteria are provided.
    @return: The matching L{Distribution} or C{None} if one isn't available.
    """
    if not (description or distributor or release or code_name):
        raise RuntimeError("At least one filtering criteria must be provided.")

    where = []
    if description is not None:
        where.append(Distribution.description == description)
    if distributor is not None:
        where.append(Distribution.distributor == distributor)
    if release is not None:
        where.append(Distribution.release == release)
    if code_name is not None:
        where.append(Distribution.code_name == code_name)

    store = get_main_store()
    result = store.find(Distribution, *where)
    return result.one()
