# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import Interface, Attribute
from zope.schema import (
    Datetime,
    Int,
    Password,
    TextLine,
    Choice,
    Object,
    Text,
    Bool,
    List,
)

from canonical.lib.validators import validator
from canonical.lib.text import is_valid_slug
from canonical.landscape.model.main.schema import EmailAddress
from canonical.landscape.ui.format import is_valid_registration_key


__all__ = [
    "IAccessContextAware",
    "IAccessContext",
    "IPendingComputer",
    "IAccount",
    "IComputer",
    "ILicense",
    "ICollectionContext",
    "IPerson",
    "ISalesforceInfo",
    "ITag",
    "IAccountAware",
    "IPersonAware",
    "IAccessContextComposite",
    "IAccessDelegateAware",
    "IMetaRelease",
    "IAccountPerson",
    "ILDSLicenseAuthority",
]

__clearlist__ = __all__


class IAccessContext(Interface):

    name = TextLine(
        title="Name of the access group",
        constraint=validator(is_valid_slug, "Invalid name format."),
        description=("The name can contain alphanumeric characters, '-' and '+'."),
        required=True,
    )

    title = TextLine(
        title="Title",
        description=(
            "A meaningful title for the access group. E.g: " "'West coast servers'."
        ),
        required=True,
    )

    access_context = Choice(
        title="Parent", default="global", vocabulary="access_contexts", required=True
    )


class ICollectionContext(Interface):
    """Mark entities that can be used as collection contexts.

    These entities are containers for other entities and can be used as filters
    for searches using Collection-based classes.
    """

    store = Attribute("The Storm store the contained entities live in.")

    collection_filter = Attribute(
        "The collection filter parameter name to use when searching."
        "for entities contained in this entity."
    )


class IAccessContextAware(Interface):
    access_context = Object(
        title="Access context",
        description="The access context used to secure this object.",
        schema=IAccessContext,
    )


class IAccessContextComposite(IAccessContextAware):
    access_children = List(title="Children", description="Children entities.")

    access_cutoff_role = TextLine(
        title="Cutoff role",
        description="Limit the roles acquired from children entities.",
    )


class IAccessDelegateAware(Interface):
    """Delegate permission checking to another object.

    If an object provides L{IAccessDelegateAware}, the security machinery will
    grant to the principal trying to access the object all permissions the
    principal itself has on the C{access_delegate} attribute object.
    """

    access_delegate = Object(
        title="Access delegate",
        description="The object to delegate permission checking to.",
        schema=IAccessContext,
    )


class IPendingComputer(Interface):

    title = TextLine(
        title="Title", description="A short name for the computer.", required=True
    )

    hostname = TextLine(
        title="Hostname", description="The computer's hostname.", required=False
    )


class IAccount(Interface):

    name = TextLine(
        title="Name",
        description="A short name for the account.",
        min_length=3,
        max_length=50,
        required=True,
        readonly=False,
    )

    title = TextLine(
        title="Title",
        description="A short description of the account.",
        required=True,
        readonly=False,
    )

    registration_password = TextLine(
        title="Registration key",
        constraint=validator(is_valid_registration_key, "Invalid characters."),
        description="The key used by computers when registering "
        "with this account.  An empty key means clients "
        "may register without providing a key. Trailing "
        "spaces or ; or # symbols are not "
        "allowed.",
        min_length=3,
        max_length=50,
        required=False,
        readonly=False,
    )

    creation_time = Datetime(
        title="Created at",
        description="Account creation time",
        required=True,
        readonly=True,
    )

    max_people_count = Int(
        title="Administrators limit", required=True, readonly=False, min=1, max=100
    )

    salesforce_account_key = TextLine(
        title="Salesforce account key",
        description=("The Salesforce identification key for this account."),
        required=False,
    )

    trial_flag = Bool(
        title="Trial account",
        description="Is this account a trial account?",
        default=False,
    )

    max_attachment_size = Int(
        title="Maximum script attachment size",
        description="The maximum size of script attachments, in bytes.",
        required=False,
        readonly=False,
    )


class IComputer(Interface):

    title = TextLine(
        title="Title",
        description="A short title for the computer.",
        min_length=1,
        max_length=50,
        required=True,
        readonly=False,
    )

    hostname = TextLine(
        title="Hostname",
        description="The computer's hostname.",
        min_length=1,
        max_length=100,
        required=True,
        readonly=False,
    )

    total_memory = Int(
        title="Total memory",
        description="The total amount of RAM in the computer (in megabytes).",
        required=False,
        readonly=False,
    )

    total_swap = Int(
        title="Total swap",
        description="The total amount of swap in the computer " "(in megabytes).",
        required=False,
        readonly=False,
    )

    comment = Text(
        title="Comment",
        description="Comments about this computer.",
        required=False,
        readonly=False,
        missing_value="",
    )

    reboot_required_flag = Bool(
        title="Requires reboot",
        description="Flag indicating if this computer requires a reboot.",
        required=True,
        readonly=False,
    )


class IMetaRelease(Interface):

    code_name = TextLine(
        title="Distribution",
        description="The APT archive entry name of this release.",
        min_length=1,
        max_length=50,
        required=True,
    )

    name = TextLine(
        title="Name",
        description="Full name of this release",
        min_length=1,
        max_length=50,
        required=False,
    )

    version = TextLine(
        title="Version",
        description="Version of this release",
        min_length=1,
        max_length=50,
        required=False,
    )

    date = Datetime(
        title="Release date",
        description="The date this release was published.",
        required=False,
    )

    supported = Bool(
        title="Officially supported",
        description="Whether this release is still officially supported.",
        required=True,
    )

    upgrade_tool_tarball_url = TextLine(
        title="Upgrade tool",
        description="URL of the upgrade-tool tarball for this release.",
        required=False,
    )

    upgrade_tool_signature_url = TextLine(
        title="Upgrade tool signature",
        description="URL of the upgrade-tool signature for this release.",
        required=False,
    )

    lts = Bool(
        title="Long term support",
        description="Whether this release is an LTS one.",
        required=True,
    )


class ILicense(Interface):

    expiration_date = Datetime(
        title="Expiry date",
        description="The expiry date of the license.",
        required=False,
        readonly=False,
    )

    seats = Int(
        title="Number of seats",
        description="The number of seats allocated to this license.",
        required=True,
        readonly=False,
    )


class IPerson(Interface):

    name = TextLine(
        title="Name",
        description="First and last name.",
        min_length=1,
        max_length=50,
        required=True,
        readonly=False,
    )

    email = EmailAddress(
        title="E-mail address",
        description="Email address for Landscape notifications.",
        min_length=5,
        max_length=200,
        required=True,
        readonly=False,
    )

    password = Password(
        title="Passphrase",
        description="A secret passphrase.",
        min_length=3,
        max_length=50,
        required=True,
        readonly=False,
    )

    timezone = Choice(
        title="Timezone",
        description="Timezone to use when presenting dates and times.",
        required=True,
        readonly=False,
        vocabulary="timezones",
    )


class ISalesforceInfo(Interface):

    salesforce_key = TextLine(
        title="Salesforce key",
        description=(
            "The contact key of the administrator's Salesforce " "self-service user."
        ),
        required=False,
    )

    support_only = Bool(
        title="Support only user",
        description="The user is only allowed to access support cases.",
        required=False,
    )


class ITag(Interface):

    name = TextLine(
        title="Name",
        description="Tag name",
        min_length=1,
        max_length=50,
        required=True,
        readonly=False,
    )


class IAccountAware(Interface):

    account = Object(
        title="Account",
        description="The account associated with this object.",
        required=True,
        readonly=True,
        schema=IAccount,
    )


class IPersonAware(Interface):

    person = Object(
        title="Person",
        description="The person associated with this object.",
        required=True,
        readonly=True,
        schema=IPerson,
    )


class IAccountPerson(Interface):
    """A marker interface to represent a person navigated inside an account."""


class ILDSLicenseAuthority(Interface):
    """A factory for creating LDS licenses."""

    def create_license_contents(lds_license):
        """Generate signed contents for an LDS license file.

        @param lds_license: L{LDSLicense} to generate the license file from.
        """
