# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime

import json
import logging
import re
import os

from enum import Enum

import transaction

from gnupg import GPG

from zope.component import getUtility
from zope.interface import implementer
from zope.security.proxy import removeSecurityProxy

from storm.locals import Store, Int, Date, Reference, ReferenceSet, AutoReload, Unicode

from canonical.lib.text import describe_list
from canonical.lib.timestamp import utctoday

from canonical.security.protect import requires, requires_advisor, PRIVATE, PUBLIC
from canonical.security.interfaces import IRoleRegistry

from canonical.storm.variables import BytesStr

from canonical.landscape.model.main.interfaces import (
    ILicense,
    IAccountAware,
    IAccessContextAware,
)
from canonical.landscape.model.main.account import Account, get_account_by_name
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.alerts import (
    StandaloneOverflowAlert,
    LicenseExpiryAlert,
)


__all__ = [
    "CHECK_LDS_LICENSE_ERRORS",
    "InvalidLicenseError",
    "InvalidSignatureError",
    "License",
    "LicenseExpiredError",
    "CombinedLicenseFile",
    "LicenseSeatError",
    "MissingLicenseError",
    "ProLicenseType",
    "check_lds_license",
    "create_license_file",
    "get_license_file_paths",
    "get_signing_key_fingerprint",
    "signature_is_valid",
    "get_license_name",
    "LDS_ROLES",
    "InappropriateLicenseError",
    "get_licensee",
    "get_license_by_account_license_key",
    "DEFAULT_LICENSE_PATH",
    "FREE_LICENSE_KEYS",
]

__clearlist__ = [
    "License",
    "LicenseSeatError",
    "InappropriateLicenseError",
    "get_license_name",
    "LDS_ROLES",
    "get_licensee",
    "get_license_file_paths",
    "CHECK_LDS_LICENSE_ERRORS",
    "FREE_LICENSE_KEYS",
]


__requires_prefix__ = "landscape."


LICENSE_KEY_FINGERPRINT = "35338392CC7AF15C581DA3DB72A66665C5C88EE6"
LICENSE_INCLUDE_DIR = "/etc/landscape/license.d"
DEFAULT_LICENSE_PATH = "/etc/landscape/license.txt"

GNUPGHOME = "/var/lib/landscape-server/gnupg"
DATE_FORMAT = "%Y-%m-%d"
LDS_ROLES = ("LDSBasicFeatures", "LDSVMFeatures", "LDSContainerFeatures")
FREE_LICENSE_KEYS = {
    "free_lds_license",
    "free_container_lds_license",
    "free_vm_lds_license",
}

ROLE_NAME_MAPPING = {
    "BasicFeatures": ("Full", "BasicFeatures"),
    "LDSBasicFeatures": ("LDS", "BasicFeatures"),
    "TrialFeatures": ("Trial", "TrialFeatures"),
    "VMFeatures": ("Virtual", "VMFeatures"),
    "LDSVMFeatures": ("LDS Virtual", "VMFeatures"),
    "ContainerFeatures": ("Container", "ContainerFeatures"),
    "LDSContainerFeatures": ("LDS Container", "ContainerFeatures"),
    "UbuntuProFeatures": ("Ubuntu Pro", "UbuntuProFeatures"),
    "UbuntuProFreeFeatures": ("Ubuntu Pro Free", "UbuntuProFreeFeatures"),
}

MAX_FREE_PRO_SEATS = 5


class ProLicenseType(Enum):
    FULL = "full"
    FREE = "free"
    INACTIVE = "inactive"
    EXPIRED = "expired"


class InvalidLicenseError(Exception):
    """Raised when license data is missing or invalid."""


class InvalidSignatureError(Exception):
    """Raised when the GnuPG signature of the license file is invalid."""


class LicenseExpiredError(Exception):
    """Raised when adding a computer to a license which has already expired"""


class LicenseSeatError(Exception):
    """Raised when using more license seats than available"""


class MissingLicenseError(Exception):
    """Raised when the license file could not be found or read."""


class InappropriateLicenseError(Exception):
    """
    Raised when attempting to assign a license to a computer that the
    license is not appropriate for.
    """

    def __init__(self, license_type):
        self.license_type = license_type


@requires_advisor
@implementer(ILicense, IAccountAware, IAccessContextAware)
class License(object):
    """
    A mechanism for granting a role to computers for a limited amount of time.
    """

    __storm_table__ = "license"

    requires(
        "ViewLicense",
        PRIVATE,
        [
            "id",
            "computers",
            "role_key",
            "access_context_id",
            "account",
            "license_key",
            "salesforce_asset_key",
        ],
    )
    requires("ViewLicense", "EditLicense", ["expiration_date", "seats"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    expiration_date = Date()
    role_key = Int(allow_none=False)
    seats = Int(allow_none=False)
    license_key = BytesStr()
    salesforce_asset_key = Unicode()

    account = Reference(account_id, Account.id)

    def __init__(
        self, account, role_key, seats, license_key=None, salesforce_asset_key=None
    ):
        self.account = account
        self.role_key = role_key
        self.seats = seats
        self.license_key = license_key
        self.salesforce_asset_key = salesforce_asset_key

    @property
    def access_context(self):
        return self.account.access_context

    @requires("EditLicense")
    def remove(self):
        Store.of(self).remove(self)

    @requires("ViewLicense")
    def get_state(self):
        role = getUtility(IRoleRegistry).get(self.role_key)
        state = {
            "account": self.account.name,
            "role": role.name,
            "seats": self.seats,
            "license_key": self.license_key.decode(),
        }
        if self.expiration_date is not None:
            state["expiration_date"] = self.expiration_date.strftime(DATE_FORMAT)
        if self.salesforce_asset_key is not None:
            state["salesforce_asset_key"] = self.salesforce_asset_key
        return state

    @requires("ViewLicense")
    def expired(self):
        """C{True} if an expiration date is set and has been reached."""
        now = utctoday()
        return self.expiration_date < now if self.expiration_date is not None else False

    @requires("ViewLicense")
    def is_free(self):
        """Return True if a license is free."""
        if self.license_key.decode() in FREE_LICENSE_KEYS:
            return True

        role_registry = getUtility(IRoleRegistry)
        # TrialFeatures license is considered free as well in Hosted
        trial_features_role = role_registry.get("TrialFeatures")
        return self.role_key == trial_features_role.key

    @requires("EditComputerLicenses")
    def add_computer(self, computer):
        """Add a new license to computer.

        This will check that added computers aren't exceeding the license
        seat limit.

        There's a race condition between the check for license seats and
        the actual insertion in the database. To fix that we lock the
        license in the database with a select for update, and count the
        number of computers really using that license in the association
        table.

        If this license is an UbuntuProFeatures license and the computer
        has an Ubuntu Pro subscription, the number of seats scales to
        accommodate.

        @raises LicenseSeatError: If adding a computer will exceed the
            license seat limit.
        @raises LicenseExpiredError: If this license is no longer
            valid.
        """
        role_registry = getUtility(IRoleRegistry)
        # We allow containers to use a VM license when no container license is
        # available.
        if not computer.vm_info and not computer.container_info:
            vm_features_role = role_registry.get("VMFeatures")
            if self.role_key == vm_features_role.key:
                raise InappropriateLicenseError("Virtual Machine")

        if not computer.container_info:
            container_features_role = role_registry.get("ContainerFeatures")
            if self.role_key == container_features_role.key:
                raise InappropriateLicenseError("Container")

        if self.expired():
            raise LicenseExpiredError(
                f"Computer '{computer.title}' could not be added, license expired."
            )
        store = Store.of(self)
        store.execute(
            "SELECT id FROM license WHERE id=? FOR UPDATE", (self.id,), noresult=True
        )
        # We need a < 1 expression here, because it's technically possible that
        # the number of seats on a license can be set to a value lower than the
        # number of used computers.
        if self.get_unused_seats() < 1:
            if self._can_add_pro_seat(computer):
                self.seats += 1
            else:
                raise LicenseSeatError(
                    f"Computer '{computer.title}' could not be added, no seats left."
                )

        self.computers.add(computer)
        self.account.reset_licenses(computer)

    @requires("ViewLicense")
    def get_used_seats(self):
        """Return the number of currently licensed computers.

        This method returns a count of the number of computers with the same
        license id.
        """
        from canonical.landscape.model.main.computer import ComputerLicense

        store = Store.of(self)
        return store.find(ComputerLicense, license_id=self.id).count()

    @requires("ViewLicense")
    def get_unused_seats(self):
        """Returns the number of unused seats for this license."""
        return self.seats - self.get_used_seats()

    @requires("ViewLicense")
    def get_available_seats(self):
        """Returns the number of unused licenses on this license.

        This method returns 0 if this particular license is expired.
        """
        if self.expired():
            return 0
        else:
            unused_seats = self.get_unused_seats()
            return unused_seats if unused_seats > 0 else 0

    def get_used_overflow_seats(self):
        """Return the number of computers using the overflow capacity."""
        overdrawn = self.get_used_seats() - self.seats
        if overdrawn <= 0:
            return 0
        else:
            return overdrawn

    @requires("EditComputerLicenses")
    def remove_computer(self, computer):
        """Remove a single L{Computer} from this license."""
        self.computers.remove(computer)
        self.account.reset_licenses(computer)
        is_pro_license = (
            self.is_ubuntu_pro_license() or self.is_ubuntu_pro_free_license()
        )

        # If this is an Ubuntu Pro license, reduce the number of seats.
        if is_pro_license and self.seats > 0:
            self.seats -= 1

    @requires("ViewLicense")
    def is_ubuntu_pro_license(self):
        role_registry = getUtility(IRoleRegistry)

        return self.role_key == role_registry.get("UbuntuProFeatures").key

    @requires("ViewLicense")
    def is_ubuntu_pro_free_license(self):
        role_registry = getUtility(IRoleRegistry)

        return self.role_key == role_registry.get("UbuntuProFreeFeatures").key

    def _can_add_pro_seat(self, computer) -> bool:
        # Need to remove the security proxy because the computer might be unlicensed.
        sub_type = removeSecurityProxy(computer.ubuntu_pro_subscription_type())

        if self.is_ubuntu_pro_license():
            return sub_type is ProLicenseType.FULL
        elif self.is_ubuntu_pro_free_license():
            return sub_type is ProLicenseType.FREE and self.seats < MAX_FREE_PRO_SEATS

        return False


Account.licenses = ReferenceSet(
    Account.id, License.account_id, order_by=License.expiration_date
)


class CombinedLicenseFile(object):
    """Representation of all license files for standalone Landscape.

    @ivar licenses: A list of licenses parsed from the file.
    @ivar licensee: The name of the organisation this license is for.
    """

    def __init__(self, licensee, licenses):
        self._licenses = licenses
        self.licensee = licensee

    def get_licenses(self):
        """
        Return a list of dicts with details about each license in the
        CombinedLicenseFile.
        """
        return [
            {
                "role": license["role"],
                "expires": license.get("expires"),
                "seats": license["seats"],
                "license_key": license.get("license_key"),
            }
            for license in self._licenses
        ]


def get_license_file_paths(include_dir=LICENSE_INCLUDE_DIR):
    """Return a list of known paths for licenses.

    @param include_dir: Define a non-standard include_dir for testing

    This separate function allows us to mock and pass arbitrary paths in tests.
    """
    license_paths = [os.environ.get("LANDSCAPE_LICENSE_FILE", DEFAULT_LICENSE_PATH)]
    if os.path.isdir(include_dir):
        files = os.listdir(include_dir)
        if not files:
            # include_dir added to represent expected license paths
            license_paths.append(include_dir)
        for filename in files:
            license_paths.append(os.path.join(include_dir, filename))
    return license_paths


def get_license_by_account_license_key(account, license_key):
    """
    Return the matching L{License} object for this license_key in the specific
    account or None if no matching L{License} object.
    """
    if isinstance(license_key, str):
        license_key = license_key.encode()

    return (
        get_main_store()
        .find(License, license_key=license_key, account_id=account.id)
        .one()
    )


def synchronise_license_file_to_database(license_file, startup=False):
    """
    Process the CombinedLicenseFile and ensure all the licenses in the files
    are correct in the database.
    """
    # Ideally, we'd get the account name from the license_file.licensee
    account = get_account_by_name("standalone")
    if not account:
        return

    try:
        store = get_main_store()
        file_licenses = license_file.get_licenses()
        file_license_keys = set([license["license_key"] for license in file_licenses])

        if startup:
            logging.info("Synchronising license file to database on startup")
            license_keys = [license.license_key for license in account.licenses]
            logging.info(f"Current license keys = {license_keys}")
            logging.info(f"File license keys = {list(file_license_keys)}")

        # Special case upgrades - if there's only one existing license in the
        # database and the license_key is "", then we can set the
        # license_key to be the first license_key in the license file, and it
        # will be configured correctly.
        if account.licenses.count() == 1:
            license = account.licenses.one()
            if not license.license_key:
                license_key = file_licenses[0]["license_key"]

                if license_key is not None:
                    license_key = license_key.encode()

                license.license_key = license_key
                logging.info(
                    "Detected upgrade case - assigned "
                    "key = {}".format(license_key and license_key.decode())
                )

        for license in file_licenses:
            license_key = license["license_key"]
            if license["expires"]:
                license_expiry = license["expires"].strftime(DATE_FORMAT)
            else:
                license_expiry = "Never"

            unknown_role = license["role"] not in ROLE_NAME_MAPPING
            if startup:
                action = "Ignoring" if unknown_role else "Processing"
                logging.info(
                    f"{action} license "
                    f"({license_key}, {license['role']}, {license['seats']}, {license_expiry})"
                )
            if unknown_role:
                # Skip licenses that have unknown role names. This means the
                # license type is newer than the this LDS release, and it's
                # unsupported by it.
                continue
            db_license = get_license_by_account_license_key(account, license_key)
            if db_license:
                license_changed = False
                if db_license.seats != license["seats"]:
                    db_license.seats = license["seats"]
                    license_changed = True

                if db_license.expiration_date != license["expires"]:
                    db_license.expiration_date = license["expires"]
                    license_changed = True

                role = getUtility(IRoleRegistry).get(license["role"])
                if db_license.role_key != role.key:
                    db_license.role_key = role.key
                    license_changed = True

                if license_changed:
                    logging.info(
                        f"Updating license with key {license_key}: seats {license['seats']}, "
                        f"expires {license_expiry} role: {license['role']}"
                    )
            else:
                logging.info(f"Creating new license with key {license_key}")
                from canonical.landscape.model.license.management import (
                    LicenseManagement,
                )

                LicenseManagement(account).create_license(
                    license["role"],
                    license["seats"],
                    license["expires"],
                    license_key=license_key,
                )
        account_license_keys = {
            license.license_key and license.license_key.decode()
            for license in account.licenses
        }
        deletable_keys = account_license_keys.difference(file_license_keys)
        for key in deletable_keys:
            logging.info(f"Deleting unknown license key {key}")
            # Note: we can't do is_in with "None" as a value
            store.find(License, license_key=key.encode()).remove()

        if startup:
            logging.info("Completed license synchronisation on startup.")
            license_keys = [license.license_key for license in account.licenses]
            logging.info(f"Current license keys = {license_keys}")

        account.check_rules([StandaloneOverflowAlert, LicenseExpiryAlert])
    except Exception:
        transaction.abort()
        raise
    else:
        transaction.commit()


# In case of licensing problems one of the following errors is raised by
# check_lds_license().
CHECK_LDS_LICENSE_ERRORS = (
    InvalidLicenseError,
    InvalidSignatureError,
    MissingLicenseError,
)


def check_lds_license(startup=False):
    """
    This function checks the LDS license and is called at system startup,
    during initial user creation and prior to accepting a pending computer.

    This involves
        - making sure the LDS license file exists and is valid
        - inserting or updating the license in the database to reflect the
          data in the license file as necessary

    Please note: this function raises an exception if anything with the
    license is amiss.
    """
    try:
        account = get_account_by_name("standalone")
        # No standalone account found => no license data in the db, we're done.
        if not account:
            logging.info("No standalone account found, license in db not checked.")
            return True

        license_file = create_license_file()
        synchronise_license_file_to_database(license_file, startup=startup)

    except Exception as e:
        logging.exception(e)
        raise
    else:
        return True


def create_license_file(check_signature=True, validate_licenses=None):
    """
    Instantiate a CombinedLicenseFile object if the provided license data is
    correct.

    @param check_signature: if False, doesn't check the signature of the
        file. It's fine after startup when the license has already been
        verified.
    @param validate_licenses: For testing, callable returning a tuple of
        licensee, licenses and errors.

    @raises MissingLicenseError: when no licenses are found at any defined
        license paths.
    @raises InvalidLicenseError: when the only license file found is malformed
        or lacking a required data item.
    @raises InvalidSignatureError: when the GPG signature validation fails
        for the only license file found or when it was signed with the wrong
        key.
    """
    if validate_licenses is None:
        validate_licenses = _validate_all_licenses
    licensee, licenses, errors = validate_licenses(check_signature)
    if not licenses:
        if errors:
            # Raise the first specific error as it is more appropriate than
            # MissingLicense.
            raise errors[0]
        error_message = describe_list(
            "No LDS license file found at",
            get_license_file_paths(),
            conjunction="or",
            pluralize=None,
            quote=False,
        )
        raise MissingLicenseError(error_message)
    return CombinedLicenseFile(licensee, licenses)


def extract_license_from_signing_armour(license_text):
    """
    Extract the actual JSON representing the license from the signed file.
    """
    clearsigned_re = re.compile(
        "-----BEGIN PGP SIGNED MESSAGE-----"
        ".*?(?:\r\n|\n)(?:\r\n|\n)(.*)(?:\r\n|\n)"
        "(-----BEGIN PGP SIGNATURE-----"
        ".*"
        "-----END PGP SIGNATURE-----)",
        re.DOTALL,
    )
    match = clearsigned_re.search(license_text)
    values = match.group(1)
    return values


def get_signing_key_fingerprint():
    """This facilitates the testing with arbitrary signing key fingerprints."""
    return LICENSE_KEY_FINGERPRINT


def signature_is_valid(text_to_check):
    """
    Check that the GnuPG signature is correct and was made using the LDS
    license signing key.

    @param text_to_check: the license file contents.
    @return: None if the signature is valid and was made using the LDS signing
        key, an error message otherwise.
    """
    try:
        gpghome = os.environ.get("LANDSCAPE_GPG_HOME", GNUPGHOME)
        gpg = GPG(gnupghome=gpghome)
        verify = gpg.verify(text_to_check)
    except Exception:
        return "The license signature verification failed"

    if not verify.valid:
        return "The license signature verification failed"
    elif verify.fingerprint != get_signing_key_fingerprint():
        return "The license was not signed with the correct key"


def get_license_name(role_name):
    """
    Return a mapping of role_name to either something suitable for displaying
    for users, or writing to an LDS file.
    """
    return dict(
        list(zip(("display", "lds"), ROLE_NAME_MAPPING.get(role_name, ("", ""))))
    )


def get_licensee():
    """Return the licensee of existing CombinedLicenseFile."""
    license_file = create_license_file(check_signature=False)
    return license_file.licensee


def _parse_license_file(path, check_signature=False):
    """Attempt to read a license at the provided path and extract license info.

    @param path: The path to a license file.
    @param check_signature: If False, don't check the signature of the
        file. It's fine after startup when the license has already been
        verified.
    @returns A tuple containing the licensee and a list of licenses.

    @raises InvalidLicenseError: when the license file is malformed or
        lacking a required data item.
    @raises InvalidSignatureError: when the GPG signature validation fails
        for the license file or when the latter was signed with the wrong key.
    """
    try:
        license_contents = open(path).read()
    except IOError:
        # Couldn't read the file or directory, return no license info
        return None, []

    if check_signature:
        signature_error = signature_is_valid(license_contents)
        if signature_error:
            message = "{}: {}.".format(signature_error, path)
            raise InvalidSignatureError(message)

    license_json = extract_license_from_signing_armour(license_contents)
    try:
        values = json.loads(license_json)
    except ValueError:
        raise InvalidLicenseError(
            "License file {} contains malformed JSON.".format(path)
        )

    if "licensee" not in values:
        raise InvalidLicenseError(
            "License file {} lacks the 'licensee' field.".format(path)
        )

    licensee = values["licensee"]
    if "licenses" not in values:
        raise InvalidLicenseError(
            "License file {} lacks the 'licenses' field.".format(path)
        )

    required_license_keys = set(("role", "seats", "license_key"))
    for license in values["licenses"]:
        if not required_license_keys.issubset(list(license.keys())):
            raise InvalidLicenseError(
                "License file {} contains malformed JSON.".format(path)
            )
        if license.get("expires") is not None:
            license["expires"] = datetime.strptime(
                license["expires"], DATE_FORMAT
            ).date()
        license["license_key"] = str(license["license_key"])
    return licensee, values["licenses"]


def _validate_all_licenses(check_signature=True, get_license_paths=None):
    """Collect and validate all licenses, log and return any validation errors.

    This includes the following checks:
        - is the GPG signature of the license file intact and was the latter
          signed with the LDS license signing key.
        - is the license file well-formed (JSON) and are the required data
          items present.

    @param check_signature: if False, doesn't check the signature of the
        file. It's fine after startup when the license has already been
        verified.
    @param get_license_paths: A function returning a list of license paths
        for testing.

    @returns a tuple of licensee, licenses and exceptions raised by any
        invalid licenses parsed
    """
    valid_licenses = []
    known_keys = set()
    valid_licensee = None
    errors = []
    if not get_license_paths:
        get_license_paths = get_license_file_paths

    for path in get_license_paths():
        try:
            licensee, licenses = _parse_license_file(path, check_signature)
        except (InvalidLicenseError, InvalidSignatureError) as e:
            # Log an error and proceed in hopes of finding a valid license.
            licenses = []
            logging.error(e)
            errors.append(e)

        for license in licenses:
            key = license["license_key"]
            if key in known_keys:
                message = "Ignoring duplicate key {} in license file {}".format(
                    key, path
                )
                logging.warning(message)
            else:
                if not valid_licensee:
                    valid_licensee = licensee  # Even standalone free licensee
                if licensee != "standalone":
                    # Overwrite standalone licensee with paid entitlement
                    valid_licensee = licensee
                known_keys.add(key)
                valid_licenses.append(license)
    return valid_licensee, valid_licenses, errors
