# Copyright 2005-2022 Canonical Limited. All rights reserved.

from email.utils import parsedate
import itertools
from datetime import datetime
from apt_pkg import TagFile
import logging
from zope.interface import implementer
from storm.locals import Int, Unicode, DateTime, Bool, Desc, Reference

from canonical.security.protect import requires, requires_advisor, PRIVATE, PUBLIC
from canonical.landscape.model.main.interfaces import IMetaRelease
from canonical.landscape.model.main.store import get_main_store


__all__ = [
    "MetaRelease",
    "get_meta_releases_by_code_name",
    "get_meta_release",
    "create_meta_release",
    "update_meta_releases_from_tagfiles",
    "update_meta_releases_upgrades",
    "get_meta_releases_after",
]

__clearlist__ = __all__


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IMetaRelease)
class MetaRelease(object):
    """
    Hold meta-information about an Ubuntu release.

    @ivar code_name: The code name of the release.
    @ivar name: The extended name of the release.
    @ivar version: The version of the release.
    @ivar date: The release date.
    @ivar supported: Flag indicating if the release is still supported.
    @ivar upgrade_tool: The URL of the tool to use to upgrade to this release.
    @ivar upgrade_tool_signature: The GPG signature of the upgrade tool.
    @ivar lts: Flag indicating if this release is an LTS one.
    @ivar display_name: The common display name with the version and the code
        name, alternative to C{description}.
    """

    __storm_table__ = "meta_release"

    requires(
        PUBLIC,
        PRIVATE,
        [
            "id",
            "code_name",
            "name",
            "version",
            "date",
            "supported",
            "upgrade_tool_tarball_url",
            "upgrade_tool_signature_url",
            "lts",
            "description",
            "upgrade",
            "display_name",
        ],
    )

    id = Int(primary=True, allow_none=False)
    code_name = Unicode(allow_none=False)
    name = Unicode()
    version = Unicode()
    date = DateTime()
    supported = Bool(default=False)
    upgrade_tool_tarball_url = Unicode()
    upgrade_tool_signature_url = Unicode()
    lts = Bool(default=False)
    upgrade_id = Int(allow_none=True)
    upgrade = Reference(upgrade_id, id)

    def __init__(
        self,
        code_name,
        name=None,
        version=None,
        date=None,
        supported=False,
        upgrade_tool_tarball_url=None,
        upgrade_tool_signature_url=None,
        lts=False,
    ):
        """
        @param code_name: The codename of the meta-release to be created.
        @param name: The extended name of the release.
        @param version: The version of the release.
        @param date: The release date.
        @param supported: Flag indicating if the release is still supported.
        @param upgrade_tool_tarball_url: The URL of the upgrade tool to use.
        @param upgrade_tool_signature_url: GPG signature of the upgrade tool.
        @param lts: Flag indicating if this release is an LTS one.
        """
        self.code_name = code_name
        self.name = name
        self.version = version
        self.date = date
        self.supported = supported
        self.upgrade_tool_tarball_url = upgrade_tool_tarball_url
        self.upgrade_tool_signature_url = upgrade_tool_signature_url
        self.lts = lts

    @property
    def description(self):
        return "Ubuntu %s (%s)" % (self.version, self.code_name)

    @property
    def display_name(self):
        return "%s (%s)" % (self.version, self.name.split()[0])

    def is_valid_upgrade(self):
        """Return C{True} if this meta release is a valid upgrade target."""
        return (
            self.supported
            and self.upgrade_tool_tarball_url
            and self.upgrade_tool_signature_url
        )


def get_meta_releases_by_code_name():
    """Get dictionary mapping L{MetaRelease.code_name}s to L{MetaRelease}s."""
    store = get_main_store()
    result = store.find((MetaRelease.code_name, MetaRelease))
    return dict(result)


def get_meta_releases_after(date):
    """Return L{MetaRelease}s released after the specified date."""
    store = get_main_store()
    result = store.find(MetaRelease, MetaRelease.date > date)
    return result.order_by(Desc(MetaRelease.date))


def get_meta_release(code_name):
    """
    Get the L{MetaRelease} matching C{code_name}, or C{None} if one isn't
    available.
    """
    store = get_main_store()
    result = store.find(MetaRelease, MetaRelease.code_name == code_name)
    return result.one()


def create_meta_release(
    code_name,
    name=None,
    version=None,
    date=None,
    supported=False,
    upgrade_tool_tarball_url=None,
    upgrade_tool_signature_url=None,
    lts=False,
):
    """Create a new L{MetaRelease} adding it to the store.

    @param code_name: The codename of the meta-release to be created.
    @param name: The extended name of the release.
    @param version: The version of the release.
    @param date: The release date.
    @param supported: Flag indicating if the release is still supported.
    @param upgrade_tool_tarball_url: URL of the tarball containing the
        upgrade-tool to use to upgrade to this release.
    @param upgrade_tool_signature_url: The GPG signature of the upgrade tool.
    @param lts: Flag indicating if this release is an LTS one.

    @return: The newly created L{MetaRelease}.
    """
    meta_release = MetaRelease(
        code_name,
        name=name,
        version=version,
        date=date,
        supported=supported,
        upgrade_tool_tarball_url=upgrade_tool_tarball_url,
        upgrade_tool_signature_url=upgrade_tool_signature_url,
        lts=lts,
    )
    store = get_main_store()
    store.add(meta_release)

    # Link distributions with the same code name
    from canonical.landscape.model.main.distribution import Distribution

    for distribution in store.find(Distribution, code_name=code_name):
        distribution.meta_release = meta_release

    return meta_release


def update_meta_releases_upgrades():
    """Update the C{upgrade_id} attribute of all available L{MetaRelease}s.

    This function fetches all available L{MetaRelease}s and check which upgrade
    path is available for each of them, if any.
    """
    store = get_main_store()
    result = store.find(MetaRelease)
    result.order_by(Desc(MetaRelease.date))

    next_meta_release = None
    next_meta_release_lts = None

    for meta_release in result:
        meta_release.upgrade = None

        # For normal releases the only possible upgrade is to the next
        # release, if supported.
        if next_meta_release and next_meta_release.is_valid_upgrade():
            meta_release.upgrade = next_meta_release

        # If we are LTS, check if there is a later supported LTS we can
        # upgrade to.
        if (
            meta_release.lts
            and next_meta_release_lts
            and next_meta_release_lts.is_valid_upgrade()
        ):
            meta_release.upgrade = next_meta_release_lts

        next_meta_release = meta_release
        if meta_release.lts:
            next_meta_release_lts = meta_release


def update_meta_releases_from_tagfiles(meta_release_tagfile, lts_tagfile=""):
    """Update information about MetaReleases reading them from tagfiles.

    @parm meta_release_tagfile: The tagfile containing all known releases.
    @parm lts_tagfile: Optional tagfile containing only LTS releases.

    @note: A sample of meta-release tagfile can be found at:
           http://changelogs.ubuntu.com/meta-release(-lts)
    """
    lts_section_by_code_names = {}
    if lts_tagfile:
        lts_tagfile = TagFile(lts_tagfile)
        lts_section_by_code_names = {
            str(section["Dist"]): section for section in lts_tagfile
        }

    meta_releases = get_meta_releases_by_code_name()
    tag_file = TagFile(meta_release_tagfile)
    for section in itertools.chain(tag_file, lts_tagfile):

        code_name = str(section["Dist"])
        name = str(section["Name"])
        version = str(section["Version"])
        date = datetime(*parsedate(section["Date"])[:6])
        supported = bool(int(section["Supported"]))

        meta_release = meta_releases.get(code_name)
        if not meta_release:
            meta_release = create_meta_release(code_name)

        meta_release.name = name
        meta_release.version = version
        meta_release.date = date
        meta_release.supported = supported

        if "UpgradeTool" in section:
            meta_release.upgrade_tool_tarball_url = str(section["UpgradeTool"])
        if "UpgradeToolSignature" in section:
            meta_release.upgrade_tool_signature_url = str(
                section["UpgradeToolSignature"]
            )
        if code_name in lts_section_by_code_names:
            lts_section = lts_section_by_code_names[code_name]
            for key in list(lts_section.keys()):
                if section[key] != lts_section[key]:
                    logging.warning(
                        "LTS tagfile content for {} doesn't match base"
                        " tagfile.".format(code_name)
                    )
                    break
            meta_release.lts = True
        else:
            meta_release.lts = False
