# Copyright 2005-2022 Canonical Limited. All rights reserved.

import string
import random

from hashlib import sha1, sha512
from datetime import datetime, timedelta

from pytz import timezone, UTC

from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import implementer
from zope.event import notify
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.traversing.interfaces import ITraversable

from storm.locals import Storm, Store, Int, Unicode, DateTime, Reference, AutoReload
from storm.store import EmptyResultSet

from canonical.lib.text import arg_to_bytes, make_secure_id
from canonical.security.interfaces import IRoleRegistry, IRole
from canonical.security.protect import requires, requires_advisor, PRIVATE, PUBLIC

from canonical.storm.variables import BytesStr

from canonical.lib.credentials import create_access_key, create_secret_key
from canonical.landscape.mail.mail import format_address
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.model.main.interfaces import IPerson
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.events import NewUserEvent
from canonical.landscape.model.main.interfaces import IAccessDelegateAware


__all__ = [
    "Person",
    "PasswordReset",
    "Formatter",
    "create_person",
    "create_support_person",
    "get_person_by_email",
    "get_persons_by_email",
    "get_person_by_id",
    "get_persons_by_ids",
    "get_person_by_identity",
    "get_password_reset",
    "get_password_reset_cutoff_time",
    "create_account_manager",
    "APICredentials",
    "get_credentials_by_access_key_id",
    "PersonAccessContextsVocabularyFactory",
]

__clearlist__ = [
    "Person",
    "get_person_by_identity",
    "is_cloud_user",
    "PersonAccessContextsVocabularyFactory",
    "create_person",
    "get_person_by_email",
    "get_persons_by_email",
    "get_person_by_id",
]


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IPerson)
class Person(Storm):
    __storm_table__ = "person"

    requires(PUBLIC, "EditPerson", ["timezone"])
    requires(PUBLIC, PRIVATE, ["formatter", "reset_password"])
    requires("ViewPerson", PRIVATE, ["id", "accounts"])
    requires(
        "ViewPerson",
        "EditPerson",
        ["name", "email", "identity", "last_login_time", "last_login_host"],
    )
    requires("ViewPersonPassword", "EditPerson", ["password"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    email = Unicode(allow_none=False)
    password = BytesStr()
    timezone = Unicode(default="UTC")
    identity = Unicode()
    last_login_time = DateTime()
    last_login_host = Unicode()

    _role_registry = None
    _roles = None
    _computer_id_index = None

    def __init__(self, name, email, password=None, timezone=None, identity=None):
        self.name = name
        self.email = email
        if password is not None:
            self.set_password(password)
        if timezone is not None:
            self.timezone = timezone
        self.identity = identity

    @staticmethod
    def _generate_salt(choices=string.ascii_letters + string.digits):
        system_random = random.SystemRandom()
        return "".join(system_random.choice(choices) for i in range(8))

    @classmethod
    def _encrypt_password(cls, password, salt=None, algorithm="sha512"):
        if salt is None:
            salt = cls._generate_salt()
        encryption_functions = {"sha1": sha1, "sha512": sha512}
        salted = (salt + password).encode()
        return salt + encryption_functions[algorithm](salted).hexdigest()

    @requires("EditPerson")
    def set_password(self, plaintext, salt=None, algorithm="sha512"):
        """
        Set the password, encrypting it before storing it in the
        database.
        """
        self.password = (
            algorithm
            + ":"
            + self._encrypt_password(plaintext, salt=salt, algorithm=algorithm)
        ).encode()

    def reset_password(self, utcnow=None):
        """
        Creates a L{PasswordReset} object that can be used to reset
        this person's passphrase.
        """
        Store.of(self).find(PasswordReset, person_id=self.id).remove()
        return PasswordReset(self, utcnow=utcnow)

    @requires(PUBLIC)
    def check_password(self, password):
        my_pw = self.password.decode()
        split_location = my_pw.find(":")
        algorithm = my_pw[:split_location]
        salt = my_pw[split_location + 1 : split_location + 9]
        old_password = my_pw[split_location + 1 :]
        return old_password == self._encrypt_password(password, salt, algorithm)

    @Lazy  # Since __storm_invalidated__ clears this cache, Lazy is ok. # no qa
    def formatter(self):
        """Get this person's L{Formatter}."""
        return Formatter(self.timezone)

    @requires("CreateAccount")
    def create_account(
        self,
        name,
        title,
        registration_password=None,
        database_key=None,
        salesforce_account_key=None,
        max_people_count=None,
    ):
        """
        Create an account.  The account is not associated with the Person in
        any way.

        (Why is this a method of Person, again?)
        """
        from canonical.landscape.model.main.account import create_account

        return create_account(
            name,
            title,
            registration_password,
            database_key,
            salesforce_account_key,
            max_people_count,
        )

    @requires("ViewAccount")
    def get_default_account(self):
        """Get the default L{Account} for this person.

        The default account changes, depending on the situation:

         * When only a single account is available, it is always the default
           account.
         * If multiple accounts are available, the first non-expired account
           (sorted by L{Account.name}) is the default.
         * If multiple accounts are available and all of them are expired, the
           first account (sorted by L{Account.name}) is the default.
        """
        accounts = list(self.accounts)
        for account in accounts:
            if not account.has_expired():
                return account
        if accounts:
            return accounts[0]
        return None

    @requires("ViewPermissionInfo")
    def get_role_registry(self):
        """
        Get a registry that includes custom roles on the person's accounts.
        """
        if self._role_registry is None:
            from canonical.landscape.model.access.role import get_role_registry
            from canonical.landscape.model.main.account import Account

            account_ids = list(self.accounts.values(Account.id))
            self._role_registry = get_role_registry(account_ids)
        return self._role_registry

    @requires("ViewPermissionInfo")
    def get_all_roles(self):
        """
        Return a mapping of {access_context_id: Role} that apply to
        this person.
        """
        if self._roles is None:
            # See canonical.landscape.security.principals
            from canonical.landscape.model.main.personaccess import PersonAccess
            from canonical.landscape.model.access.role import (
                StoredRole,
                StoredRolePerson,
            )
            from canonical.landscape.model.access.role import AccountRole

            registry = self.get_role_registry()
            store = Store.of(self)
            roles = {}
            for person_access in store.find(PersonAccess, person_id=self.id):
                role = registry.get(person_access.role_key)
                if role is not None:
                    access_context_id = person_access.access_context_id
                    context_roles = roles.get(access_context_id)
                    if context_roles is None:
                        roles[access_context_id] = set((role,))
                    else:
                        context_roles.add(role)

            # There are sometimes magical L{AccountRole}s that contain
            # permissions that must be granted on the toplevel access context
            # in the associated account. The L{AccountRole}s refer to the
            # L{StoredRole}, which allows us to know which account it's for and
            # which people are in it.

            # {stored_role_id: account_role_instance}
            stored_role_ids = dict(
                [
                    (role_.stored_role_id, role_)
                    for role_ in registry.get_all()
                    if isinstance(role_, AccountRole)
                ]
            )

            # stored_roles is all of the L{StoredRole}s that this Person is in
            # and that have L{AccountRole}s associated with them.
            stored_roles = store.find(
                StoredRole,
                StoredRolePerson.role_id == StoredRole.id,
                StoredRolePerson.person_id == self.id,
                StoredRolePerson.role_id.is_in(list(stored_role_ids.keys())),
            )
            for stored_role in stored_roles:
                account_roles = roles.setdefault(
                    stored_role.account.access_context_id, set()
                )
                account_roles.add(stored_role_ids[stored_role.id])

            if not self.accounts.find().is_empty():
                kb_user = registry.get("KnowledgeBaseUser")
                if None in roles:
                    roles[None].add(kb_user)
                else:
                    roles[None] = set((kb_user,))

            global_roles = roles.get(None)
            if global_roles:
                for access_context_id in roles:
                    if access_context_id is not None:
                        roles[access_context_id].update(global_roles)
            self._roles = roles
        return self._roles

    def add_global_role(self, role_name):
        """Grant this person membership of a global role."""
        self._roles = None
        from canonical.landscape.model.main.personaccess import PersonAccess

        role = getUtility(IRoleRegistry).get(role_name)
        PersonAccess(self, role, None)

    def remove_global_role(self, role_name):
        """Revoke this person's membership of a global role."""
        self._roles = None
        from canonical.landscape.model.main.personaccess import PersonAccess

        role = self.get_role_registry().get(role_name)
        Store.of(self).find(
            PersonAccess,
            PersonAccess.person == self,
            PersonAccess.role_key == role.key,
            PersonAccess.access_context_id == None,
        ).remove()

    @requires("ViewPermissionInfo")
    def has_global_role(self, role):
        """Returns a bool indicating whether the person has a role globally.

        @param role: The role to check, either a string or an actual role
            object providing L{IRole}.
        """
        if not IRole.providedBy(role):
            role = getUtility(IRoleRegistry).get(role)
        return role in self.get_all_roles().get(None, ())

    @requires("ViewPerson")
    def has_role(self, role, access_context):
        """Indicates if the person has the given role in an access context.

        This honors access context inheritance; if the person has the role in a
        parent of the passed access context, True will be returned.

        @param role: can be a L{Role} or a string
        """
        if isinstance(role, str):
            registry = self.get_role_registry()
            role = registry.get(role)
        all_roles = self.get_all_roles()

        while access_context is not None:
            roles = all_roles.get(access_context.id)
            if roles is not None and role in roles:
                return True
            access_context = access_context.parent
        return False

    @requires("ViewPerson")
    def has_permission(self, permission, access_context):
        """
        Indicates if the person has the given permission on an access context.

        This honors access context inheritance; if the person has the
        permission in a parent of the passed access context, True will be
        returned.
        """
        roles = self.get_role_registry().get_with_permission(permission)
        for role in roles:
            if self.has_role(role, access_context) or self.has_global_role(role):
                return True
        return False

    @requires("ViewPerson")
    def get_computer_ids(self, account, permission="landscape.ViewComputer"):
        """Get the L{Computer.id}s this person has access to in C{account}.

        @param account: the L{Account} to get L{Computer.id}s from.
        @param permission: the permission needed for the returned computer IDs.

        @return: C{None} if the person has access to all L{Computer}s in
            C{account}, or a C{list} of L{Computer.id}s otherwise.  If the
            person dosn't have access to any L{Computer} in the account, an
            empty C{list} is returned.

        @note: L{Computer.id}s retrieved from this method are cached.  Store
            invalidation resets the cache to ensure transaction safety.
        """
        if self._computer_id_index is None:
            self._computer_id_index = {}
        key = (account.id, permission)
        if key not in self._computer_id_index:
            from canonical.landscape.model.main.computer import Computer
            from canonical.landscape.model.computer.collection import ComputerCollection

            has_full_access = self.has_permission(permission, account.access_context)
            if not has_full_access:
                collection = ComputerCollection(
                    get_main_store(),
                    account_id=account.id,
                    person_id=self.id,
                    permission=permission,
                )
                self._computer_id_index[key] = list(
                    collection.find().values(Computer.id)
                )
            else:
                self._computer_id_index[key] = None
        return self._computer_id_index[key]

    @requires("ViewPerson")
    def reset_cache(self):
        self._computer_id_index = None
        try:
            del self.formatter
        except AttributeError:
            pass
        self._role_registry = None
        self._roles = None

    __storm_invalidated__ = reset_cache

    @requires("EditPerson")
    def update_login_data(self, login_host=""):
        """
        This is a method that needs to be called upon successful login to
        the Landscape application. Currently, it updates the latest login time
        for a user as well as the IP address they were determined to be logging
        in from.
        """
        self.last_login_time = datetime.utcnow()
        if login_host:
            self.last_login_host = login_host

    @requires("EditPerson")
    def create_api_credentials(self, account=None):
        """
        Create an L{APICredentials} for this person in the given account (or
        C{None} for global credentials).
        """
        if account is not None and account not in self.accounts:
            raise ValueError(
                "Can't create API credentials if the person is not a member "
                "of the account."
            )
        if account is None:
            # Check that there is not an existing global credentials
            credentials = self.get_api_credentials()
            if credentials is not None:
                raise ValueError("Can't have more than one global APICredentials")
        access_key_id = create_access_key().encode("utf-8")
        access_secret_key = create_secret_key().encode("utf-8")
        return APICredentials(access_key_id, access_secret_key, self, account)

    @requires("ViewPerson")
    def get_api_credentials(self, account=None):
        """
        Retrieves the associated L{APICredentials} with the optional associated
        C{account}.
        """
        store = Store.of(self)
        return store.find(
            APICredentials,
            APICredentials.person == self,
            APICredentials.account == account,
        ).one()

    @requires("EditPerson")
    def delete_api_credentials(self, account=None):
        """
        Delete L{APICredentials} for this person in the given account (or
        C{None} for global credentials).
        """
        credentials = self.get_api_credentials(account)
        if credentials is None:
            raise ValueError("There are no credentials for this account")
        store = Store.of(credentials)
        store.remove(credentials)

    def get_state(self):
        """
        Returns a representation of a L{Person}.
        """
        return {"name": self.name, "email": self.email, "id": self.id}

    @requires("EditPerson")
    def send_test_email(self, name, email_address):
        """
        Send a test email to the person.

        @param name: The person whom we are addressing.
        @param email_address: The Email address to be tested.
        """
        to = format_address(name, email_address)
        subject = "Landscape Email Test."
        body = (
            """
Dear %s,

This is a test email from the Landscape system.  If you received this
e-mail it means that Landscape is capable of sending system alerts to
you.

If you believe you have received this email by mistake please contact
your system administrator.

Kind regards,
--
Landscape Team,
Canonical Ltd.
"""
            % name
        )
        getUtility(ILandscapeMailer).send([to], subject, body)


@requires_advisor
class PasswordReset(object):
    """An object which can be used to reset a L{Person}'s passphrase.

    This is in support of functionality which wants to allow users who
    have forgetton their passphrase by sending them a one-time key
    over a relatively insecure connection which they can then use to
    reset their passphrase over a secure connection.

    @ivar secure_id: A secure randomly-generated string that must be
    passed to L{reset}.
    """

    # Note on permissions: L{PasswordReset}s are accessed by
    # arbitrary Persons. So making this object public is OK as long as
    # there are no *factories* that return PasswordReset objects
    # without verifying a secure ID.

    __storm_table__ = "password_recovery"

    requires(PUBLIC, PRIVATE, ["secure_id", "creation_time"])  # person?

    secure_id = Unicode(primary=True, allow_none=False)
    person_id = Int(allow_none=False)
    creation_time = DateTime(allow_none=False)

    person = Reference(person_id, Person.id)

    def __init__(self, person, utcnow=None):
        self.person = person
        # Let's keep the length of the URL under 80 characters so it
        # won't wrap in email clients..
        length = 79 - len("http://landscape.canonical.com/reset-passphrase/")
        self.secure_id = make_secure_id(length)
        self.creation_time = utcnow or datetime.utcnow()

    # The General Public can try to reset their password.
    @requires(PUBLIC)
    def reset(self, password):
        """
        Set the associated person's passphrase to the given
        passphrase, and delete this object from the database.
        """
        self.person.set_password(password)
        Store.of(self).remove(self)


@requires_advisor
@implementer(ITraversable)
class Formatter(object):

    requires(PUBLIC, PRIVATE, ["traverse"])

    def __init__(self, local_timezone="UTC"):
        self._tzinfo = timezone(local_timezone)

    @requires(PUBLIC)
    def traverse(self, name, furtherPath):
        return getattr(self, name)

    @requires(PUBLIC)
    def short_timezone(self, time=None, tzname_only=False):
        """Format the time zone.

        @param time: If specified, the datetime to use when querying the offset
            (which may vary due to DST). If omitted, the current datetime is
            used.
        @param tzname_only: If specified, return the time zone only without
            the offset.
        """
        if self._tzinfo is UTC:
            return "UTC"
        else:
            if time is None:
                time = datetime.now(tz=self._tzinfo)
            else:
                time = self._ensure_timezone_aware(time)
                time = time.astimezone(self._tzinfo)
            return time.tzname() if tzname_only else time.strftime("%Z%z")

    @requires(PUBLIC)
    def timezone_offset(self, time=None):
        """
        Return the UTC offset of the timezone as an integer number of minutes.

        @param time: If specified, the datetime to use when querying the offset
            (which may vary due to DST). If omitted, the current datetime is
            used.
        """
        if self._tzinfo is UTC:
            return 0
        else:
            if time is None:
                time = datetime.now()
            return int(self._tzinfo.utcoffset(time).total_seconds()) // 60

    @requires(PUBLIC)
    def format_date(
        self, time, now=None, with_year=False, with_tz=False, localize=True
    ):
        """Format a C{date} or C{datetime} into a long date string.

        If C{time} is a C{datetime} for today or yesterday the string is
        formatted as 'Today at ...' or 'Yesterday at ...', respectively.

        @param time: A C{date} or C{datetime} instance to format.
        @param now: Optionally, a C{datetime} representing the current time.
        @param with_year: Passing C{with_year=True} forces year, month and day
            to be included even if the date is today or yesterday.
        @param with_tz: If C{with_tz} is C{True}, then the short time zone is
            appended to the date string.
        @param localize: Whether to localize and convert to the user timezone
            if the time that was passed is an instance of datetime.
        @return: A formatted long date string.
        """
        now = now if now else datetime.now(self._tzinfo)
        if isinstance(time, datetime):
            formatted = self._format_datetime(time, now, with_year, localize=localize)
        else:
            formatted = self._format_date(time, now, with_year)
        if with_tz:
            formatted += " %s" % self.short_timezone(time, tzname_only=True)

        return formatted

    @requires(PUBLIC)
    def format_short_date(self, date, now=None, with_year=False):
        """
        Returns C{date} formatted as a short date string.  If the date
        is today or yesterday the string is formatted as 'Today ...' or
        'Yesterday ...'.  Details such as the year or month/day are only
        included in the string if they're necessary for disambiguation
        purposes. Passing C{with_year=True} forces year, month and day to be
        included even if the date is today or yesterday.
        """
        if not date:
            return ""
        now = now or datetime.now(self._tzinfo)
        date = self._ensure_timezone_aware(date)
        date = date.astimezone(self._tzinfo)
        if date.date() == now.date() and not with_year:
            return date.strftime("Today %H:%M")
        elif date.year != now.year or with_year:
            return date.strftime("%Y-%m-%d %H:%M")
        else:
            yesterday = now - timedelta(1)
            if date.date() == yesterday.date():
                return date.strftime("Yesterday %H:%M")
        return date.strftime("%b %-d %H:%M")

    size = {
        "TB": 1024 * 1024 * 1024 * 1024,
        "GB": 1024 * 1024 * 1024,
        "MB": 1024 * 1024,
        "KB": 1024,
    }

    @requires(PUBLIC)
    def format_bytes(self, bytes):
        """Convert a byte integer value to a human-readble string."""
        for abbrev, size in self.size.items():
            if bytes >= size:
                return "{:.2f}{}".format(bytes / size, abbrev)

        return "{:d}B".format(bytes)

    @requires(PUBLIC)
    def format_megabytes(self, megabytes):
        """Convert a megabyte integer value to a human-readable string.

        Values below 1024 are represented in MB; values above 1024 are
        represented in GB; and values above 1024*1024 are represented
        in TB.
        """
        if megabytes >= 1048576:
            return "%.2fTB" % (megabytes / 1048576.0)
        elif megabytes >= 1024:
            return "%.2fGB" % (megabytes / 1024.0)
        else:
            return "%dMB" % (megabytes)

    def _format_date(self, time, now, with_year):
        """Format a C{date} instance."""
        if time.year != now.year or with_year:
            return time.strftime("%a %-d %b %Y")
        return time.strftime("%a %-d %b")

    def _format_datetime(self, time, now, with_year, localize=True):
        """Format a C{datetime} instance."""
        if localize:
            time = self._ensure_timezone_aware(time)
            time = time.astimezone(self._tzinfo)
        if not with_year:
            yesterday = now - timedelta(1)
            if (
                time.year == now.year
                and time.month == now.month
                and time.day == now.day
            ):
                return time.strftime("Today at %H:%M")
            elif (
                time.year == yesterday.year
                and time.month == yesterday.month
                and time.day == yesterday.day
            ):
                return time.strftime("Yesterday at %H:%M")
            elif time.year == now.year:
                return time.strftime("%a %-d %b, %H:%M")
        return time.strftime("%a %-d %b %Y, %H:%M")

    def _ensure_timezone_aware(self, date_or_time):
        """Ensure a date or datetime is timezone-aware (not naive).

        If the timezone is not set, it's assumed to be UTC.
        """
        if date_or_time.tzinfo is None:
            return date_or_time.replace(tzinfo=UTC)
        return date_or_time


@requires_advisor
@implementer(IAccessDelegateAware)
class APICredentials(Storm):
    """Hold information about API access for a person.

    @ivar access_key_id: Access key for the API.
    @type access_key_id: C{str}
    @ivar access_secret_key: Secret key for the API, used to sign requests.
    @type access_secret_key: C{str}
    @ivar person_id: Identifier of the person holding those credentials.
    @type person_id: C{int}
    @ivar person: Reference to the person instance.
    @type person: L{Person}
    @ivar account_id: Identifier of the account in which the credentials are
        valid. If C{None}, they are globally valid.
    @type account_id: C{int} or C{NoneType}
    @ivar account: Reference to the account instance.
    @type account: L{Account}
    """

    __storm_table__ = "api_credentials"

    requires("ViewPerson", PRIVATE, ["person", "account"])
    requires("ViewPerson", "EditPerson", ["access_key_id", "access_secret_key"])

    access_key_id = BytesStr(primary=True, allow_none=False)
    access_secret_key = BytesStr(allow_none=False)
    person_id = Int(allow_none=False)
    account_id = Int()

    person = Reference(person_id, Person.id)
    account = Reference(account_id, "Account.id")

    def __init__(self, access_key_id, access_secret_key, person, account):
        self.access_key_id = access_key_id
        self.access_secret_key = access_secret_key
        self.person = person
        self.account = account

    @property
    def access_delegate(self):
        return self.person


def get_person_by_email(email):
    """Return the L{Person} with the given C{email}, or C{None} otherwise.

    @note: The query is case-insensitive.
    """
    return (
        get_main_store()
        .find(Person, Person.email.like(email, case_sensitive=False))
        .one()
    )


def get_persons_by_email(email):
    """
    Returns a ResultSet of L{Person]s with the given C{email}. For SSO, we
    allow duplicate emails. Otherwise, the result should always have zero or
    one items.
    """
    return get_main_store().find(Person, Person.email.like(email, case_sensitive=False))


def get_person_by_id(id):
    return get_main_store().get(Person, id)


def get_person_by_identity(identity):
    return get_main_store().find(Person, identity=identity).one()


def get_persons_by_ids(ids):
    if not ids:
        return EmptyResultSet()
    return get_main_store().find(Person, Person.id.is_in(ids))


@arg_to_bytes
def get_credentials_by_access_key_id(access_key_id):
    """Returns the L{APICredentials} matching the given C{access_key_id}."""
    store = get_main_store()
    return store.get(APICredentials, access_key_id)


def get_password_reset(id):
    """
    This returns the newly reset password with the given secure_id if it hasn't
    expired.
    """
    cutoff_time = get_password_reset_cutoff_time()
    result = get_main_store().find(
        PasswordReset,
        PasswordReset.secure_id == id,
        PasswordReset.creation_time >= cutoff_time,
    )
    return result.one()


def create_person(*args, **kwargs):
    person = Person(*args, **kwargs)
    get_main_store().add(person)
    # Notify creation of new person
    event = NewUserEvent(person)
    notify(event)
    return person


def create_support_person(*args, **kwargs):
    person = create_person(*args, **kwargs)
    person.add_global_role("SupportProvider")
    return person


def create_account_manager(*args, **kwargs):
    person = create_person(*args, **kwargs)
    person.add_global_role("AccountManager")
    return person


def get_password_reset_cutoff_time():
    """
    Returns the cutoff time after which this reset password becomes invalid.
    This is currently 24 hours.
    """
    return datetime.utcnow() - timedelta(hours=24)


@implementer(IContextSourceBinder)
class PersonAccessContextsVocabularyFactory(object):
    """
    L{PersonAccessContextsVocabularyFactory} builds a vocabulary of access
    contexts on which a person has a specific permission.

    @param permission:  the name of the permission you wish to find access
        contexts for.
    @type permission: string
    @param person: optionally, a specific person to check permissions against.
        This should be used when the context does not have an associated
        person.  If is not used then the person associated with the context
        will be used.
    """

    def __init__(self, permission, person=None):
        self.permission = permission
        self.person = person

    def __call__(self, context):
        from canonical.landscape.model.main.personaccess import (
            get_access_context_ids_for_person,
        )

        if not self.person:
            self.person = context.person

        account = context.account
        access_context_ids = get_access_context_ids_for_person(
            account.access_context, self.permission, self.person
        )
        terms = []
        for access_context_id in access_context_ids:
            access_context = context.account.get_access_context(access_context_id)
            terms.append(
                SimpleTerm(access_context.name, title=access_context.description)
            )
        return SimpleVocabulary(terms)
