# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.security.proxy import removeSecurityProxy

from storm.locals import Int, Reference

from canonical.security.protect import requires, requires_advisor
from canonical.landscape.model.main.accesscontext import (
    AccessContext,
    AccessContextCollection,
)
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store


__all__ = ["PersonAccess", "get_persons_with_role", "get_access_context_ids_for_person"]

__clearlist__ = ["get_access_context_ids_for_person"]


__requires_prefix__ = "landscape."


@requires_advisor
class PersonAccess(object):
    __storm_table__ = "person_access"
    __storm_primary__ = "person_id", "role_key", "access_context_id"

    requires(
        "ViewPersonAccess",
        "EditPersonAccess",
        ["person_id", "role_key", "access_context_id", "person", "access_context"],
    )

    person_id = Int(allow_none=False)
    role_key = Int(allow_none=False)
    access_context_id = Int(default=None)

    person = Reference(person_id, Person.id)
    access_context = Reference(access_context_id, AccessContext.id)

    def __init__(self, person, role, access_context):
        self.person = person
        self.role_key = role.key
        self.access_context = access_context
        assert self.role_key is not None


def get_persons_with_role(role, access_context):
    """Return L{Person}s with the given role in the given access context."""
    store = get_main_store()
    result = store.find(
        Person,
        Person.id == PersonAccess.person_id,
        PersonAccess.role_key == role.key,
        PersonAccess.access_context_id == access_context.id,
    )
    return result


def get_access_context_ids_for_person(root_access_context, permission, person):
    """Return access contexts on which the person has the specified permission.

    This function returns the IDs of all AccessContexts under the given root
    access context that the given person has the given permission on, either
    directly or indirectly (through the access context hierarchy).
    """
    if not permission.startswith("landscape."):
        permission = "landscape." + permission

    registry = person.get_role_registry()
    roles = registry.get_with_permission(permission)
    role_keys = [role.key for role in roles if role.key is not None]

    from canonical.landscape.model.access.role import AccountRole

    for role in roles:
        # XXX: Since the registry is proxied, we must remove the proxy so
        # that we can access the role properties in the condition bellow.
        role = removeSecurityProxy(role)
        if (
            isinstance(role, AccountRole)
            and registry.may([role], permission)
            and person.has_role(role, root_access_context)
        ):
            # It's a magic account-level role, which means the permissions
            # within it should apply to all access contexts in the account,
            # so return all access contexts.
            collection = AccessContextCollection(root=root_access_context)
            return list(collection.find().values(AccessContext.id))

    store = get_main_store()

    has_global_permission = not store.find(
        PersonAccess,
        PersonAccess.person_id == person.id,
        PersonAccess.role_key.is_in(role_keys),
        PersonAccess.access_context_id == None,
    ).is_empty()

    access_context_ids = []

    if has_global_permission:
        # The person has a global role with the required permission.
        # This means the person can see *all* access contexts for which
        # she holds that role, relative to this account (she can see all
        # access contexts for all accounts).
        # We can return a list of all access contexts with the given roles.
        access_context_ids = store.find(
            AccessContext.id, AccessContext.root_id == root_access_context.id
        )
        access_context_ids.config(distinct=True)
    else:
        # Get all access context IDs the person has suitable roles on
        context_id = root_access_context.id
        access_context_ids = store.find(
            PersonAccess.access_context_id,
            PersonAccess.person_id == person.id,
            PersonAccess.role_key.is_in(role_keys),
            PersonAccess.access_context_id == AccessContext.id,
            AccessContext.root_id == context_id,
        )
        access_context_ids.config(distinct=True)

        # Get also all access context IDs descending from the above ones
        access_context_ids = set(access_context_ids)
        parent_ids = access_context_ids
        while parent_ids:
            children_ids = store.find(
                AccessContext.id, AccessContext.parent_id.is_in(parent_ids)
            )
            children_ids = children_ids.config(distinct=True)
            children_ids = set(children_ids)
            parent_ids = children_ids - access_context_ids
            access_context_ids.update(children_ids)

    return list(access_context_ids)
