# Copyright 2005-2022 Canonical Limited. All rights reserved.

import time

from zope.interface import implementer
from zope.component import queryUtility

from storm.locals import Storm, Store, Int, Unicode, Reference

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.support.interfaces import ISalesforceConnector
from canonical.landscape.model.main.interfaces import IAccessContextAware
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.store import get_main_store

__all__ = ["SalesforceUser", "get_salesforce_user_by_sfid"]


__clearlist__ = ["get_salesforce_user_by_sfid"]


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IAccessContextAware)
class SalesforceUser(Storm):

    __storm_table__ = "salesforce_user"

    id = Int(primary=True, allow_none=False)
    account_id = Int(allow_none=False)
    person_id = Int(allow_none=False)
    salesforce_key = Unicode(allow_none=False)

    account = Reference(account_id, Account.id)
    person = Reference(person_id, Person.id)

    requires("ViewSalesforceUser", PRIVATE, ["access_context", "account", "person"])
    requires("ViewSalesforceKey", "EditSalesforceKey", ["salesforce_key"])

    def __init__(self, account, person, salesforce_key):
        self.account = account
        self.person = person
        self.salesforce_key = salesforce_key

    @property
    def access_context(self):
        """
        Access to the C{SalesforceUser} is essentially granted by the
        account.  The {Account}'s access context is proxied for this
        reason.  It's critical that the C{ManageSupportCases}
        permission is handled carefully to ensure security integrity.
        """
        return self.account.access_context

    @requires("ManageSupportCases")
    def get_support_cases_url(self, cache=None, create_time=time.time):
        """
        Authenticate the user against the Salesforce support cases
        system and return a session-specific URL.  If authentication
        can't be completed C{None} is returned.

        @param cache: Optionally, a dict-like object to cache data
            used to calculate session timeouts.  This is passed
            through to the L{ISalesforceConnector} utility.
        @param create_time: Optionally, a function taking no arguments
            that returns the current time.
        """
        connector = queryUtility(ISalesforceConnector)
        if connector:
            return connector.get_support_cases_url(
                self.salesforce_key, cache, create_time
            )
        return None

    @requires("EditSalesforceKey")
    def remove(self):
        return Store.of(self).remove(self)


def get_salesforce_user_by_sfid(sfid):
    return get_main_store().find(SalesforceUser, salesforce_key=sfid).one()
