# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Or, Like

from canonical.security.protect import requires

from canonical.landscape.model.main.account import Account
from canonical.landscape.model.account.invitation import AccountInvitation
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.salesforceuser import SalesforceUser
from canonical.landscape.model.main.store import get_main_store


__all__ = ["search"]


@requires("ViewAllAccounts")
def search(search_string):
    """Search for L{Account}s and L{Person}s that match C{substring}.

    Restricts the search results to 50 accounts and 50 people.

    @return: Generator yields matching L{Account}s and L{Person}s.
    """
    store = get_main_store()
    substring = search_string.replace("%", "\\%")
    substring = "%" + substring + "%"

    accounts = store.find(
        Account,
        Or(
            Like(Account.name, substring, case_sensitive=False),
            Like(Account.title, substring, case_sensitive=False),
            Like(Account.salesforce_account_key, substring, case_sensitive=False),
        ),
    )

    people = store.find(
        Person,
        Or(
            Like(Person.name, substring, case_sensitive=False),
            Like(Person.email, substring, case_sensitive=False),
        ),
    )

    invitations = store.find(
        AccountInvitation,
        Or(
            Like(AccountInvitation.name, substring, case_sensitive=False),
            Like(AccountInvitation.email, substring, case_sensitive=False),
            AccountInvitation.salesforce_key == search_string,
        ),
    )

    salesforce_users = store.find(
        Person,
        SalesforceUser.person_id == Person.id,
        SalesforceUser.salesforce_key == search_string,
    )

    people = people.union(salesforce_users)
    people.config(distinct=True)

    return sorted(
        list(accounts) + list(people) + list(invitations), key=lambda x: x.name.lower()
    )
