# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
import uuid

from canonical.landscape.model.main.store import get_main_store


__all__ = ["MissingServerUUIDError", "get_server_uuid", "ensure_server_uuid_exists"]


_cached_server_uuid = None


class MissingServerUUIDError(Exception):
    """Raised when no UUID for the server is found in the database."""


def get_server_uuid(connection=None, store=None):
    """Return the UUID which uniquely identifies this server.

    The server UUID is publicy disclosed, and as such is used only as an
    advisory mechanism to let clients know when they're talking to a
    different server, rather than as some kind of security system.

    @param connection: an optional psycopg2 connection to be used to
        connect to the database.
    @param store: Optionally, a store to execute the statement on.  If not
        provided, uses the main store.
    """
    global _cached_server_uuid
    if _cached_server_uuid is None:
        statement = "SELECT uuid FROM server_identity"
        if connection is None:
            if store is None:
                store = get_main_store()
            result = store.execute(statement).get_one()
        else:
            cursor = connection.cursor()
            cursor.execute(statement)
            result = cursor.fetchone()
        if not result:
            raise MissingServerUUIDError("Server UUID not found")
        _cached_server_uuid = uuid.UUID(result[0])
    return _cached_server_uuid


def ensure_server_uuid_exists(connection=None, store=None):
    """Check if a server UUID already exists, and creates it if not.

    @param connection: Optionally, a psycopg connection to be used in place
        of the defaut IZStorm utility. Note that possible changes to the
        database are not committed.
    @param store: Optionally, a store to execute the statement on.  If not
        provided, uses the main store.
    """
    try:
        get_server_uuid(connection=connection)
    except MissingServerUUIDError:
        statement = "INSERT INTO server_identity VALUES (1, '%s')" % (
            str(uuid.uuid1()),
        )
        if connection is None:
            if store is None:
                store = get_main_store()
            store.execute(statement, noresult=True)
        else:
            cursor = connection.cursor()
            cursor.execute(statement)


def reset_server_uuid_cache():
    global _cached_server_uuid
    _cached_server_uuid = None
