from datetime import datetime

from storm.locals import (
    AutoReload,
    DateTime,
    Int,
    Reference,
    ReferenceSet,
    Storm,
    Unicode,
)

from canonical.security.protect import requires, requires_advisor, PUBLIC


@requires_advisor
class InstalledSnap(Storm):
    """
    Model representing the many-to-many relationship between Computers
    and Snaps. A single revision of a Snap installed on a Computer.
    """

    __storm_table__ = "installed_snap"
    __storm_primary__ = "computer_id", "snap_id"

    computer_id = Int()
    snap_id = Unicode()
    version = Unicode()
    revision = Unicode()
    tracking_channel = Unicode()
    held_until = DateTime()
    confinement = Unicode()

    computer = Reference(computer_id, "Computer.id")
    snap = Reference(snap_id, "Snap.id")

    requires(PUBLIC, names=["snap_id", "revision"])

    def __init__(
        self,
        computer,
        snap,
        version: str,
        revision: str,
        tracking_channel: str,
        held_until: datetime,
        confinement: str = "strict",
    ):
        self.computer = computer
        self.snap = snap
        self.version = version
        self.revision = revision
        self.tracking_channel = tracking_channel
        self.held_until = held_until
        self.confinement = confinement

    def __hash__(self):
        return hash((self.snap_id, self.revision))

    def __eq__(self, other):
        if hasattr(other, "snap_id") and hasattr(other, "revision"):
            return self.snap_id == other.snap_id

        return False


class Snap(Storm):
    """
    Model representing revision-independent data we have about a single
    snap.
    """

    __storm_table__ = "snap"

    id = Unicode(primary=True, allow_none=False, default=AutoReload)
    name = Unicode()
    publisher_id = Int()
    summary = Unicode()

    publisher = Reference(publisher_id, "SnapPublisher.id")
    computers = ReferenceSet(
        id, "ComputerSnap.snap_id", "ComputerSnap.computer_id", "Computer.id"
    )

    def __init__(self, id: str, name: str, summary: str, publisher):
        self.id = id
        self.name = name
        self.summary = summary
        self.publisher = publisher


class SnapPublisher(Storm):

    __storm_table__ = "snap_publisher"

    id = Int(primary=True)
    username = Unicode()
    validation = Unicode()

    def __init__(self, username, validation):
        self.username = username
        self.validation = validation
