# Copyright 2005-2022 Canonical Limited. All rights reserved.

from uuid import uuid4
import json

from gnupg import GPG

from zope.component import getUtility
from zope.interface import implementer

from storm.expr import Func, SQL
from storm.locals import (
    Storm,
    Int,
    Bool,
    Unicode,
    Reference,
    ReferenceSet,
    DateTime,
    AutoReload,
    Or,
    And,
)

from canonical.lib.timestamp import utctoday, utcnow

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.security.interfaces import IRoleRegistry

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.model.main.account import Account, get_account_by_name
from canonical.landscape.model.main.license import License, get_license_name
from canonical.landscape.model.main.interfaces import ILDSLicenseAuthority
from canonical.landscape.model.main.store import get_main_store


__all__ = [
    "LDSLicense",
    "AccountLDSLicenseAPI",
    "LDSLicenseAuthority",
    "check_account_archive_access",
    "generate_license_contents",
    "get_available_releases",
    "synchronize_releases",
    "LDSScriptExecutionManagement",
    "setup_lds_script_executions",
    "LicenseLDSLicense",
]

__clearlist__ = [
    "AccountLDSLicenseAPI",
    "get_available_releases",
    "LDSScriptExecutionManagement",
    "setup_lds_script_executions",
]

__requires_prefix__ = "landscape."


class LicenseLDSLicense(object):
    """A link object to associate an L{LDSLicense} with a L{License}."""

    __storm_table__ = "license_lds_license"
    __storm_primary__ = "lds_license_id", "license_id"

    lds_license_id = Int(allow_none=False)
    license_id = Int(allow_none=False)


class LDSLicense(Storm):
    """A license granting access to Landscape On Premises."""

    __storm_table__ = "lds_license"

    id = Int(primary=True, allow_none=False)
    account_id = Int(allow_none=False)
    apt_key = Unicode(allow_none=False)
    contents = Unicode(allow_none=False)

    account = Reference(account_id, Account.id)

    licenses = ReferenceSet(
        id, LicenseLDSLicense.lds_license_id, LicenseLDSLicense.license_id, License.id
    )

    def __init__(self, account, apt_key):
        self.account = account
        self.apt_key = apt_key
        self.contents = ""

    def add_license(self, license):
        """
        Add a license to the list of licenses associated with this LDSLicense.
        """
        self.licenses.add(license)

    def regenerate_apt_key(self):
        """Regenerate the APT key."""
        self.apt_key = generate_apt_key()

    def regenerate_contents(self):
        """Regenerate the LDS license file."""
        authority = getUtility(ILDSLicenseAuthority)
        self.contents = authority.create_license_contents(self)


def generate_apt_key():
    """Get a randomly generated APT key."""
    return str(uuid4())


class AccountLDSLicenseAPI(object):
    def __init__(self, account):
        self._account = account

    def create_license(self):
        """Create a new L{LDSLicense}.

        This allows LDS role types to be added to the account.
        """
        apt_key = generate_apt_key()
        license = LDSLicense(self._account, apt_key)
        release = self.get_release()
        if not release:
            store = get_main_store()
            releases = get_available_releases()
            if releases:
                release = LDSRelease(self._account.id, str(releases[0]))
                store.add(release)
        return license

    def get_license(self):
        """Get the L{LDSLicense} for the L{Account}.

        @return: An L{LDSLicense} instance or C{None} if one doesn't exist.
        """
        store = get_main_store()
        result = store.find(LDSLicense, LDSLicense.account_id == self._account.id)
        return result.one()

    def remove_license(self):
        """Remove the current L{LDSLicense}."""
        license = self.get_license()
        if license:
            get_main_store().remove(license)

    def get_available_releases(self):
        """
        Proxy L{get_available_releases} for easier access in the UI.
        """
        return get_available_releases()

    def set_release(self, release):
        """The current release for this account."""
        store = get_main_store()
        result = store.find(LDSRelease, account_id=self._account.id)
        result.set(release=release)

    def get_release(self):
        """Return the current release for this account."""
        store = get_main_store()
        release = store.find(LDSRelease, account_id=self._account.id).one()
        if release:
            return release.release

    def has_valid_license(self, apt_key=None):
        """Checks if the L{Account} has a valid, non-expired L{LDSLicense}.

        @param apt_key: If provided, also compare the C{apt_key} when
            considering license validity.
        @return: C{True} if there is a non-expired license, C{False}
            otherwise .
        """
        store = get_main_store()
        where = [
            LDSLicense.account_id == self._account.id,
            LicenseLDSLicense.lds_license_id == LDSLicense.id,
            LicenseLDSLicense.license_id == License.id,
            License.expiration_date >= utctoday(),
        ]
        if apt_key is not None:
            where.append(LDSLicense.apt_key == apt_key)
        result = store.find(LDSLicense.id, *where)
        return bool(result.count())


@implementer(ILDSLicenseAuthority)
class LDSLicenseAuthority(object):
    """A factory to generate content for LDS entitlement files.

    @param gpg_home_path: The path to the GPG private key.
    @param passphrase: The passphrase for the GPG key.
    """

    def __init__(self, gpg_home_path, passphrase):
        self._gpg_home_path = gpg_home_path
        self._passphrase = passphrase

    def create_license_contents(self, lds_license):
        """Generate signed contents for an LDS license file.

        @param lds_license: L{LDSLicense} to generate the license file from.
        """
        contents = generate_license_contents(lds_license)
        return self.sign_contents(contents)

    def sign_contents(self, contents):
        """
        Sign the JSON of the given C{contents}.
        """
        gpg = GPG(gnupghome=self._gpg_home_path)
        json_message = json.dumps(contents, indent=0, separators=(",", ":"))
        return str(gpg.sign(json_message, passphrase=self._passphrase))


def generate_license_contents(lds_license):
    """
    @param lds_license: L{LDSLicense} to generate the license file from.
    """
    # Backwards compatibility - ensure we have the same keys as the original
    # file.
    role_registry = getUtility(IRoleRegistry)
    contents = {"licensee": lds_license.account.name, "licenses": []}
    for license in lds_license.licenses:
        expiration_date = license.expiration_date
        role = role_registry.get(license.role_key)
        license_data = {
            "role": get_license_name(role.name)["lds"],
            "seats": license.seats,
            "license_key": license.license_key.decode("utf-8"),
        }
        if expiration_date:
            license_data["expires"] = expiration_date.isoformat()
        contents["licenses"].append(license_data)
    return contents


def check_account_archive_access(account_name, apt_key):
    """
    Check if access to the Landscape On Premises archive is
    allowed, given an C{Authorization} header.
    """
    account = get_account_by_name(account_name)
    if account is None:
        return False
    elif account.has_expired():
        return False
    api = AccountLDSLicenseAPI(account)
    return api.has_valid_license(apt_key)


def get_available_releases():
    """
    Return the list of available LDS releases, using the PPPA configuration on
    hosted, or the C{LDSAvailableRelease} on standalone.
    """
    if is_standalone():
        store = get_main_store()
        result = store.find(LDSAvailableRelease)
        result.order_by(LDSAvailableRelease.release)
        releases = [release.release for release in result]
    else:
        config = getUtility(IFileBasedConfiguration)
        try:
            section = config.get("pppa-proxy")
        except KeyError:
            return []
        releases = section.get("supported-releases", "").split()
    releases.sort(reverse=True)
    return releases


class LDSRelease(object):
    """Table associating a LDS release and an account."""

    __storm_table__ = "lds_release"
    account_id = Int(allow_none=False, primary=True)
    release = Unicode(allow_none=False)

    def __init__(self, account_id, release):
        self.account_id = account_id
        self.release = release


class LDSAvailableRelease(object):
    """Store the releases published on hosted. Only used in standalone."""

    __storm_table__ = "lds_available_release"
    release = Unicode(allow_none=False, primary=True)

    def __init__(self, release):
        self.release = release


def synchronize_releases(releases):
    """Synchronize the list of available releases."""
    store = get_main_store()
    store.find(LDSAvailableRelease).remove()
    for release in releases:
        store.add(LDSAvailableRelease(str(release)))


DEFAULT_SCRIPT_RUN_TIMEOUT = 5

default = object()


@requires_advisor
class LDSScriptExecutionLog(object):
    """
    Table holding times for the last execution of LDS scripts, such as cron
    scripts.
    """

    __storm_table__ = "lds_script_execution_log"

    requires(
        "ViewScriptIssues",
        PRIVATE,
        [
            "name",
            "start_time",
            "end_time",
            "success",
            "ran",
            "run_interval",
            "run_timeout",
            "exceeded_interval",
            "exceeded_timeout",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    run_interval = Int(allow_none=False)
    run_timeout = Int(allow_none=False)
    start_time = DateTime(allow_none=False)
    end_time = DateTime()
    success = Bool()

    def __init__(
        self,
        name,
        run_interval,
        run_timeout=None,
        start_time=None,
        end_time=None,
        success=None,
    ):
        self.name = name
        self.run_interval = run_interval
        self.run_timeout = run_timeout or DEFAULT_SCRIPT_RUN_TIMEOUT
        self.start_time = start_time or utcnow()
        self.end_time = end_time
        self.success = success

    @property
    def running(self):
        """Return C{True} if the script is currently running."""
        return self.end_time is None

    @property
    def ran(self):
        """Return C{True} if the script has ever run."""
        return not self.running and self.success is not None

    @property
    def exceeded_interval(self):
        """
        Return C{True} if the script hasn't been run for a time longer than its
        C{run_interval}.
        """
        if self.running:
            return False
        interval = (utcnow() - self.end_time).total_seconds() // 60
        return interval > self.run_interval

    @property
    def exceeded_timeout(self):
        """
        Return C{True} if the script is running and exceeded its run timeout.
        """
        if not self.running:
            return False
        interval = (utcnow() - self.start_time).total_seconds() // 60
        return interval > self.run_timeout

    @property
    def errored(self):
        """Return C{True} if the script execution is in any error state."""
        return self.success is False or self.exceeded_interval or self.exceeded_timeout

    def set(self, start_time=default, end_time=default, success=True):
        """
        Set details for the script execution.

        If only C{end_time} is provided, C{start_time} is not changed and
        C{success} is updated.  Otherwise, the C{start_time} is set and
        C{end_time} and L{success} are set to C{None} unless specified.
        """
        if end_time is not default:
            self.end_time = end_time
            self.success = success
            if start_time is not default:
                self.start_time = start_time
        else:
            if start_time is default:
                start_time = utcnow()
            if end_time is default:
                end_time = None
                success = None

            self.start_time = start_time
            self.end_time = end_time
            self.success = success


@requires_advisor
class LDSScriptExecutionManagement(object):
    def create_execution(
        self,
        name,
        run_interval,
        run_timeout=None,
        start_time=None,
        end_time=None,
        success=None,
    ):
        """
        Create a L{LDSScriptExecutionLog} with the specified details.
        """
        execution = LDSScriptExecutionLog(
            name,
            run_interval,
            run_timeout=run_timeout,
            start_time=start_time,
            end_time=end_time,
            success=success,
        )
        get_main_store().add(execution)
        return execution

    def get_execution(self, name):
        """
        Return a L{LDSScriptExecutionLog} with the given name, or C{None} if
        not found.
        """
        return (
            get_main_store()
            .find(LDSScriptExecutionLog, LDSScriptExecutionLog.name == name)
            .one()
        )

    @requires("ViewScriptIssues")
    def get_executions(self, names=None, only_errored=False):
        """
        Return L{LDSScriptExecutionLog}s, possibly filtering only scripts with
        specified names or that didn't run or failed execution.
        """
        where = []
        if names:
            where.append(LDSScriptExecutionLog.name.is_in(names))
        if only_errored:
            now = utcnow()
            exceeded = And(
                Func("age", now, LDSScriptExecutionLog.end_time)
                > LDSScriptExecutionLog.run_interval * SQL("'1 min'::interval")
            )
            timed_out = And(
                LDSScriptExecutionLog.end_time == None,
                Func("age", now, LDSScriptExecutionLog.start_time)
                > LDSScriptExecutionLog.run_timeout * SQL("'1 min'::interval"),
            )
            failed = LDSScriptExecutionLog.success == False

            where.append(Or(exceeded, timed_out, failed))

        result = get_main_store().find(LDSScriptExecutionLog, *where)
        result.order_by(LDSScriptExecutionLog.name)
        return result

    def remove_executions(self, names):
        """
        Remove L{LDSScriptExecutionLog}s with specified names. Unknown names
        are ignored.
        """
        self.get_executions(names=names).remove()

    def mark_start(self, name):
        """
        Set the start time for a script to the current time.
        If a script by the specified name is not found, no action is taken.
        """
        execution = self.get_execution(name)
        if execution:
            execution.set(start_time=utcnow())

    def mark_end(self, name, success):
        """
        Set the end time for a script to the current time and it success state.
        If a script by the specified name is not found, no action is taken.
        """
        execution = self.get_execution(name)
        if execution:
            execution.set(end_time=utcnow(), success=success)


LDS_SCRIPTS = {
    "update_security_db.sh": {"interval": 70, "timeout": 10},
    "maintenance.sh": {"interval": 48 * 60, "timeout": 30},
    "update_alerts.sh": {"interval": 70, "timeout": 10},
    # Other scripts (e.g. releases, metrics) are not function-critical so
    # we will not enforce them here.
}


def setup_lds_script_executions(scripts=LDS_SCRIPTS):
    """
    Sync LDS script entries with the provided dict mapping script names to
    their configuration.
    """
    management = LDSScriptExecutionManagement()
    script_names = set(scripts)
    found_names = set()
    to_remove = set()

    for script in management.get_executions():
        name = script.name
        if name in scripts:
            found_names.add(name)
            config = scripts[name]
            script.run_interval = config["interval"]
            script.run_timeout = config["timeout"]
        else:
            to_remove.add(name)

    for name in script_names - found_names:
        config = scripts[name]
        # New script executions are created with start/end time but with
        # success=None, so that the alert is not raised immediately.
        now = utcnow()
        management.create_execution(
            name,
            config["interval"],
            run_timeout=config["timeout"],
            start_time=now,
            end_time=now,
        )
    if to_remove:
        management.remove_executions(names=to_remove)
