# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from functools import partial
from itertools import chain
import logging
import psycopg2

from datetime import datetime, timedelta

from zope.event import notify
from storm.locals import Store
from storm.exceptions import IntegrityError

from canonical.event.subscriber import subscriber

from canonical.lib.retry import retry, uniform_exponential
from canonical.landscape.model.events import (
    AddComputerEvent,
    MaintenanceEvent,
    NewUserEvent,
    ProcessStarting,
)
from canonical.landscape.model.main.computer import PingTime
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.serveridentity import ensure_server_uuid_exists
from canonical.landscape.model.main.license import (
    CHECK_LDS_LICENSE_ERRORS,
    check_lds_license,
)
from canonical.landscape.model.computer.removal import RemovalProfileManagement
from canonical.landscape.model.main.account import (
    DEFAULT_STANDALONE_ACCOUNT_NAME,
    get_all_accounts,
    DEFAULT_STANDALONE_ACCOUNT_MAX_PEOPLE_COUNT,
    bootstrap_account,
)
from canonical.landscape.model.main.person import (
    get_password_reset_cutoff_time,
    PasswordReset,
    Person,
)
from canonical.landscape.model.account.invitation import (
    AccountInvitation,
    INVITATION_EXPIRY_IN_DAYS,
)
from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.global_maintenance import global_maintenance


@subscriber(IAccount, AddComputerEvent)
def create_ping_time_handler(account, event):
    """Create a L{PingTime} for a computer.

    Called whenever a L{Computer} is added to an L{Account}.
    """
    ping_time = PingTime()
    get_main_store().add(ping_time)
    ping_time.computer = event.computer


@global_maintenance
def maintain_all_accounts(event):
    """Emit MaintenanceEvent for all accounts in the system."""
    logger = logging.getLogger("maintenance-script")
    for account in get_all_accounts():
        logger.info(
            "Started maintenance of account %s at %s" % (account.id, datetime.now())
        )
        maintenance_event = MaintenanceEvent(account)
        try:
            retry(partial(notify, maintenance_event), uniform_exponential(5))
        except:
            logger.exception(
                "Failed to perform maintenance of account {}".format(account.id)
            )
        else:
            logger.info(
                "Ended maintenance of account %s at %s" % (account.id, datetime.now())
            )


@subscriber(IAccount, MaintenanceEvent)
def remove_expired_invitations(account, unused_event):
    logger = logging.getLogger("maintenance-script")
    expiration_date = datetime.utcnow() - timedelta(days=INVITATION_EXPIRY_IN_DAYS)
    result = Store.of(account).find(
        AccountInvitation,
        AccountInvitation.account == account,
        AccountInvitation.creation_time <= expiration_date,
    )
    for invitation in result:
        warning = "Deleting invitation of %r <%s> to account '%s'" % (
            invitation.name,
            invitation.email,
            invitation.account.id,
        )
        if invitation.salesforce_key is not None:
            warning += " with SFID %s" % (invitation.salesforce_key,)
        logger.warning(warning)
        invitation.remove()


@global_maintenance
def maintain_person_data(event):
    """Maintains person data.

    This removes password reset entries, which are older than 24 hours.
    """
    logger = logging.getLogger("maintenance-script")
    store = get_main_store()
    cutoff_time = get_password_reset_cutoff_time()
    values = list(
        store.find(
            (PasswordReset.secure_id, PasswordReset.person_id),
            PasswordReset.creation_time < cutoff_time,
        )
    )
    ids_to_remove = [x[0] for x in values]
    store.find(PasswordReset, PasswordReset.secure_id.is_in(ids_to_remove)).remove()
    for secure_id, person_id in values:
        logger.info(
            "Deleted password reset link %s for user %d" % (secure_id, person_id)
        )


@subscriber(IAccount, MaintenanceEvent)
def remove_old_computers(account, unused_event):
    """
    Remove computers that have not contacted landscape within the limit given
    in L{RemovalProfile}.
    """
    logger = logging.getLogger("maintenance-script")
    management = RemovalProfileManagement(account)
    removed_computer_ids = set(
        chain(*(profile.run() for profile in management.get_removal_profiles()))
    )
    logger.info(
        "Removed %i computer(s) from account '%s'"
        % (len(removed_computer_ids), account.id)
    )


@subscriber(NewUserEvent)
def standalone_create_initial_account_handler(event):
    """Bootstrap a Landscape account upon creation of a user."""
    if not is_standalone():
        return

    person = event.object
    current_account_count = get_all_accounts().count()
    current_person_count = get_main_store().find(Person).count()
    if current_account_count == 0 and current_person_count == 1:
        account = bootstrap_account(
            DEFAULT_STANDALONE_ACCOUNT_NAME, "Organisation", person=person
        )
        account.max_people_count = DEFAULT_STANDALONE_ACCOUNT_MAX_PEOPLE_COUNT
        try:
            # If a LDS license file is present and valid, a matching license
            # will be inserted into the database by check_lds_license().
            check_lds_license()
        except CHECK_LDS_LICENSE_ERRORS:
            pass


@subscriber(ProcessStarting)
def ensure_server_uuid_exists_on_startup(event):
    try:
        store = get_main_store()
        ensure_server_uuid_exists()
    except IntegrityError:
        # This means we've lost a race with another process which
        # managed to get the UUID inserted before us.  Since all
        # we really care about is that there is *a* UUID at all, we
        # just rollback our failed attempt and move on with life.
        store.rollback()
    except psycopg2.Error as e:
        # This catches both psycopg2.ProgrammingError and
        # psycopg2.OperationalError
        # These errors are raised if the schema hasn't been setup yet, or the
        # database server can't be contacted - we log the message, and exit so
        # that Landscape will terminate.
        logging.exception(e)
        raise
    else:
        store.commit()


@subscriber(ProcessStarting)
def check_lds_license_on_startup(event):
    """Check for the LDS license presence upon startup."""
    if not is_standalone():
        return

    try:
        check_lds_license(startup=True)
    except CHECK_LDS_LICENSE_ERRORS:
        pass
