# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Core L{Tag} and L{ComputerTag} objects along with tag retrieval functions.
"""

import re

from zope.security import proxy
from zope.interface import implementer

from storm.locals import Int, Unicode

from landscape.lib.tag import is_valid_tag

from canonical.lib.validators import ValidationError

from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.interfaces import ITag


__all__ = [
    "Tag",
    "ComputerTag",
    "InvalidTagError",
    "get_tags",
    "get_tag_by_name",
    "get_tag",
    "get_tags",
    "get_tag_names",
]

__clearlist__ = ["InvalidTagError", "get_tags", "get_tag_names"]

__requires_prefix__ = "landscape."


class InvalidTagError(ValidationError):
    """Raised a tag name is invalid."""


@implementer(ITag)
@requires_advisor
class Tag(object):
    """A tag is a user-supplied name used to group related L{Computer}s.

    Tag names may not have spaces or punctuation.  They should be made up of
    C{[A-Z,a-z,0-9]} and C{-} (dash).

    @cvar __storm_table__: Tag objects are stored in the C{tag} table.
    @ivar id: The primary key, automatically generated by the database, for
        this tag instance.
    @type id: Int
    @ivar name: The user-supplied name used to identify this tag instance.
    @type name: Unicode
    """

    __storm_table__ = "tag"

    requires(PUBLIC, PRIVATE, ["id", "name"])

    id = Int(primary=True, allow_none=False)
    name = Unicode(allow_none=False)

    def __init__(self, name):
        self.name = name


class ComputerTag(object):
    """A link object to associate a L{Tag} with a L{Computer}.

    @cvar __storm_table__: Computer/tag link objects are stored in the
        C{computer_tag} table.
    @cvar __storm_primary__: The compund primary key is made up of unique
        (C{computer_id}, C{tag_id}) tuples.
    @ivar computer_id: The L{Computer.id<Computer>} to link to.
    @type computer_id: Int
    @ivar tag_id: The L{Tag.id<Tag>} to link to.
    @type tag_id: Int
    """

    __storm_table__ = "computer_tag"
    __storm_primary__ = "computer_id", "tag_id"

    computer_id = Int(allow_none=False)
    tag_id = Int(allow_none=False)


def get_tags(tags):
    """Get a set of L{Tag}s matching search criteria.

    Any tag names not in the database will be automatically added, and the new
    L{Tag} objects will be included in the result.

    @param tags: A sequence of, optionally mixed, tag names, IDs or L{Tag}
        objects to get L{Tag} objects for.
    @return: A C{set} of L{Tag}s matching search criteria.
    """
    store = get_main_store()
    tag_objects = set()
    tag_names = set()
    tag_ids = set()

    # First, separate the sequence we were given into their typed buckets.
    for tag in tags:
        if isinstance(tag, int):
            tag_ids.add(tag)
        elif proxy.isinstance(tag, Tag):
            tag_objects.add(tag)
        else:
            tag_names.add(tag)

    # Then, if one or more tag names were passed in, we try to load all of
    # them in one go.  If one or more aren't loaded in that procedure, it
    # means that we have to create some of them.  We then see which ones
    # are missing, and create them individually.
    if tag_names:
        tag_names_objects = set(store.find(Tag, Tag.name.is_in(tag_names)))
        tag_objects.update(tag_names_objects)
        if len(tag_names_objects) != len(tag_names):
            missing_tags = tag_names - set(tag.name for tag in tag_names_objects)
            for tag in missing_tags:
                tag_id = store.execute("SELECT get_tag(?)", (tag,)).get_one()[0]
                tag_ids.add(tag_id)

    # Finally, if we have any tag_ids passed in, or add to the tag_ids set
    # due to the missing tags above, we update the result set with these.
    if tag_ids:
        tag_objects.update(store.find(Tag, Tag.id.is_in(tag_ids)))

    return tag_objects


def get_tag_by_name(name):
    """Get the L{Tag} for C{name}.

    If the tag name is not in the database, it will be automatically added and
    the new L{Tag} object will be returned.

    @param name: A C{unicode} string with the name of the L{Tag}.
    @return: The L{Tag} object for C{name}.
    """
    return list(get_tags([name]))[0]


def get_tag(name):
    """Get the L{Tag} called C{name} or C{None} if one doesn't exist."""
    store = get_main_store()
    result = store.find(Tag, Tag.name == name)
    return result.one()


def get_tag_names(tags_string):
    """
    Split a string in a list of tag names.

    @param: a C{string} with tags.
    @return: a list of tag names.
    @raises: InvalidTagError: if one or more tags are invalid.
    """
    if not tags_string:
        return []
    tags = [tag for tag in re.split("[, ]+", tags_string) if tag]
    if not all(is_valid_tag(tag) for tag in tags):
        raise InvalidTagError("Invalid tag(s) in string")
    return tags
