# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer
from zope.security.proxy import removeSecurityProxy

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.main.interfaces import IAccessContextAware, IAccountAware
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IAccountAware, IAccessContextAware)
class ModelManagement(object):
    """Base class for management classes.

    @param account: The L{Account} the management object will operate on.
    @param person: The L{Person} requested operations should be run as.
    """

    requires("ViewAccount", PRIVATE, ["account", "access_context"])

    def __init__(self, account, person=None):
        # XXX person should become mandatory once all calling sites pass it
        self.account = removeSecurityProxy(account)
        self.person = person

    @property
    def _event_log(self):
        from canonical.landscape.model.eventlog.management import EventLog

        return EventLog(self.account, self.person)

    @property
    def access_context(self):
        return self.account.access_context

    def get_access_context_ids(self, permission):
        """
        Return access contexts in which the person has the specified
        permission, within the account.
        """
        return get_access_context_ids_for_person(
            self.account.access_context, permission, self.person
        )

    def get_access_context(self, name, permission):
        """
        Get the L{AccessContext} with the given C{name} that the person has
        the C{permission} on.
        """
        return self.account.get_access_context_by_name(self.person, name, permission)

    @requires("ViewAccount")
    def person_can(self, permission):
        """
        Whether the person has the given permission on at least one access
        context in the account.
        """
        return len(self.get_access_context_ids(permission)) > 0

    def check_duplicate(self, model, exception, **fields):
        """
        Raises the provided exception if an instance of the provided model with
        ANY of the specified fields values is already present in the
        account.

        NOTE: this method should eventually go away in favor of the equivalent
              check_duplicate_entity once account models get normalised around
              the AccountEntity base class.

        @param model: the model class to search for duplicates.
        @param exception: an exception class to raise in case of duplicates.
            It should be a subclass of L{DuplicateModelError}.
        @param fields: fields values to check for duplication.
        """
        duplicated_fields = []

        for field, value in list(fields.items()):
            clause = {"account_id": self.account.id, field: value}
            if not self.account.store.find(model, **clause).is_empty():
                duplicated_fields.append(field)

        if duplicated_fields:
            raise exception(duplicated_fields)

    def check_duplicate_entity(self, model, exception, name, **fields):
        """Check for duplicates for an AccountEntity-based model.

        Raises the provided exceptionif an instance of the given
        AccountEntity-based model with the given name or fields is already
        present in the account.

        @param name: The name of the account entity (checked against
            AccountEntityData)
        @param **fields: Additional fields on the specific AccountEntity-based
            model (checked against the given model).
        """
        from canonical.landscape.model.entity.data import AccountEntityData

        duplicated_fields = []
        result = self.account.store.find(
            AccountEntityData, name=name, entity_type=model.get_entity_type()
        )
        if not result.is_empty():
            duplicated_fields.append("name")
        for field, value in list(fields.items()):
            clause = {field: value}
            result = self.account.store.find(
                model,
                model.data_id == AccountEntityData.id,
                AccountEntityData.account_id == self.account.id,
                **clause
            )
            if not result.is_empty():
                duplicated_fields.append(field)

        if duplicated_fields:
            raise exception(duplicated_fields)
