# Copyright 2005-2022 Canonical Limited. All rights reserved.

from decimal import Decimal

from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.main.person import Formatter
from canonical.landscape.model.monitoring.graph import MonitoringGraph, GraphExporter
from canonical.landscape.model.resource.mountinfo import FreeSpace, MountInfo


__all__ = ["UsedSpaceGraph", "UsedSpaceDataExporter"]

__clearlist__ = ["UsedSpaceGraph", "UsedSpaceDataExporter"]


def encode_mount_point(mount_point):
    mount_point = mount_point.replace("/", "_")
    return mount_point.encode("ascii", "replace")


class UsedSpaceGraph(MonitoringGraph):

    label = "Percent"

    monitoring_class = FreeSpace
    monitoring_field = FreeSpace.free_space
    computer_id_field = MountInfo.computer_id

    def __init__(self, selection, mount_point, start, stop, large, tzinfo):
        super(UsedSpaceGraph, self).__init__(selection, start, stop, large, tzinfo)
        self._mount_point = mount_point
        self._total_space = {}
        for computer in self._selection.get_selected():
            mount_info = computer.get_mount_info_for_mount_point(self._mount_point)
            if mount_info is not None:
                total_space = mount_info.total_space
            else:
                total_space = None
            self._total_space[computer.id] = total_space

    def get_filter(self, ids):
        return [
            self.monitoring_class.mount_info_id == MountInfo.id,
            MountInfo.mount_point == self._mount_point,
            MountInfo.computer_id.is_in(ids),
        ]

    def _get_total_space(self):
        total_space = None
        for space in list(self._total_space.values()):
            if total_space is None:
                total_space = space
            elif space != total_space:
                total_space = None
                break
        return total_space

    @property
    def title(self):
        total_space = self._get_total_space()
        if total_space:
            formatter = Formatter()
            text = "Used space for %s of %s" % (
                self._mount_point,
                formatter.format_megabytes(total_space),
            )
        else:
            text = "Used space for %s" % (self._mount_point,)
        return text

    def transform_value(self, computer_id, value, raw=False):
        total_space = self._total_space[computer_id]
        if raw:
            return total_space - value
        else:
            return float((total_space - value) / Decimal(total_space) * 100)


@requires_advisor
class UsedSpaceDataExporter(GraphExporter):

    unit = "MiB"

    @requires("zope.View")
    def get_encoded_mount_point(self):
        return encode_mount_point(self._graph._mount_point)
