# Copyright 2005-2022 Canonical Limited. All rights reserved.

import calendar
import csv

from datetime import datetime
from io import StringIO

from zope.interface import implementer

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.stormfunc import Avg, Mod, utc_timestamp_as_integer
from canonical.landscape.model.monitoring.interfaces import IGraph
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)


__all__ = ["MonitoringGraph", "GraphExporter"]

__clearlist__ = ["MonitoringGraph", "GraphExporter"]


@implementer(IGraph)
class MonitoringGraph(object):
    """
    Base class representing graph data in a generic fashion. Subclasses need
    to provide at least C{monitoring_class} and C{monitoring_field} to query
    the monitoring data. It may also need to customize C{get_filter} and
    C{computer_id_field}.
    """

    transform_value = None

    def __init__(self, selection, start, stop, large, tzinfo):
        self._selection = selection
        self.start = start
        self.stop = stop
        self._large = large
        self.tzinfo = tzinfo
        self.step = self.get_step()

    @property
    def computer_id_field(self):
        return self.monitoring_class.computer_id

    @property
    def legend(self):
        return dict(
            (computer.id, computer.title) for computer in self._selection.get_selected()
        )

    def get_filter(self, ids):
        return [self.monitoring_class.computer_id.is_in(ids)]

    def get_step(self):
        """
        Return a step of 5 minutes which will be used to aggregate data,
        considering a maximum number of data points to return.
        """
        period = self.stop - self.start
        points = period // 300
        max_points = 600 if self._large else 300
        step = None
        if points > max_points:
            step = 300
            while points > max_points:
                points -= 288  # 86400 / 300
                step += 300
        return step

    def get_data(self, raw=False):
        """
        Return a dictionary of computer IDs to a list of (timestamp, value).

        @param raw: If C{True}, don't aggregate any data or transform it before
            returning it.
        """
        computer_ids = self._selection.get_selected_computer_ids()
        store_index = get_resource_stores_for_computer_ids(computer_ids)
        monitoring_class = self.monitoring_class
        result = {}
        computer_id_field = self.computer_id_field
        step = self.step if not raw else None
        for store, ids in list(store_index.items()):
            where = [
                monitoring_class.timestamp >= datetime.utcfromtimestamp(self.start),
                monitoring_class.timestamp <= datetime.utcfromtimestamp(self.stop),
            ]
            where.extend(self.get_filter(ids))
            if step is None:
                values = [
                    computer_id_field,
                    self.monitoring_field,
                    utc_timestamp_as_integer(monitoring_class.timestamp),
                ]
            else:
                values = [
                    computer_id_field,
                    Avg(self.monitoring_field),
                    utc_timestamp_as_integer(monitoring_class.timestamp)
                    - Mod(utc_timestamp_as_integer(monitoring_class.timestamp), step),
                ]
            resultset = store.find(tuple(values), *where)
            resultset.order_by(computer_id_field, values[2])
            if step is not None:
                resultset.group_by(computer_id_field, values[2])
            for computer_id, value, timestamp in resultset:
                if self.transform_value is not None:
                    value = self.transform_value(computer_id, value, raw)
                # We change the timestamp on the server side so it's easier to
                # calculate date and time on the client side.
                localized_timestamp = datetime.fromtimestamp(
                    timestamp, self.tzinfo
                ).timetuple()
                timestamp = calendar.timegm(localized_timestamp)
                result.setdefault(computer_id, []).append((timestamp, value))
        return result


@requires_advisor
@implementer(IAccessDelegateAware)
class GraphExporter(object):

    unit = None

    requires("landscape.ViewPersonResource", PRIVATE, ["selection", "get_csv_data"])

    def __init__(self, graph):
        self._graph = graph

    @property
    def access_delegate(self):
        return self._graph._selection.account

    def get_graph_data(self):
        """get_graph_data returns a tuple of timestamps and graph_data.

        Timestamps are all the timestamps that we have data for.
        The graph_data is a list of lists of data points for each computer,
        with the computer title at the start, any points that are missing for
        an individual computer are None values.
        """
        graph_data = []
        all_data = self._graph.get_data(raw=True)
        for computer_id in all_data:
            all_data[computer_id] = dict(all_data[computer_id])
        timestamps = self._extract_time_stamps(all_data)
        if timestamps:
            sorted_computer_list = sorted(
                self._graph._selection.get_selected(), key=lambda x: x.title
            )
            for computer in sorted_computer_list:
                if self.unit is not None:
                    row_data = ["%s [%s]" % (computer.title, self.unit)]
                else:
                    row_data = [computer.title]
                data = all_data.get(computer.id, {})
                for timestamp in timestamps:
                    row_data.append(data.get(timestamp))
                graph_data.append(row_data)
        return (timestamps, graph_data)

    def get_csv_data(self):
        """
        This outputs the result of get_graph_data in CSV format, using the CSV
        module.
        """
        # First of all, we convert the timestamp values into localtime
        # We don't explicitly add the timezone because this stops OO.o from
        # recognising the column as a time formatted column.
        timestamps, graph_data = self.get_graph_data()
        # format = "%Y-%m-%dT%H:%M:%S %Z"
        tzinfo = self._graph.tzinfo
        timestamps = [datetime.fromtimestamp(x, tzinfo).isoformat() for x in timestamps]

        # We need to put the timestamps into the graph_data before we rotate
        # the data.
        timestamps.insert(0, "Timestamp")
        graph_data.insert(0, timestamps)

        # This rotates the current row-by-row data into columnar format.
        graph_data = list(zip(*graph_data))

        # Now we can write the CSV data out.
        output = StringIO()
        writer = csv.writer(output)
        writer.writerows(graph_data)
        return output.getvalue()

    def _extract_time_stamps(self, graph_data):
        """
        Extract all the unique timestamps from all the data in the selected
        computers.
        """
        timestamps = set()
        for computer, data in list(graph_data.items()):
            timestamps.update(iter(list(data.keys())))
        return sorted(timestamps)
