from builtins import zip
import os
from datetime import datetime, timedelta

import yaml

from canonical.landscape.testing.model import ModelTest
from canonical.landscape.model.main.account import create_account
from canonical.landscape.model.license.management import LicenseManagement


class OKRTest(ModelTest):

    query_name = None  # Subclasses should set this

    def setUp(self):
        super(OKRTest, self).setUp()
        file_path = os.path.join(os.path.dirname(__file__), "config.yaml")
        with open(file_path) as fh:
            config = yaml.load(fh)
        self.query = config["queries"][self.query_name]

    def execute_query(self):
        """Run the query with the specified name."""
        # We currently only have database per query
        [database_name] = self.query["databases"]
        database = self.zstorm.get(database_name)
        result = database.execute(self.query["sql"])
        results_by_column = list(zip(*result.get_all()))
        return dict(list(zip(self.query["metrics"], results_by_column)))


class ComputerWithRemainingMsgTest(OKRTest):

    query_name = "computer_with_remaining_msg"

    with_computer = True

    def setUp(self):
        super(ComputerWithRemainingMsgTest, self).setUp()
        self.account2 = create_account("account2", "Account 2")
        license_management2 = LicenseManagement(self.account2)
        license_management2.create_license("BasicFeatures", 10)
        self.computer2 = self.account2.create_computer("c2", "Computer 2")

    def test_computer_count(self):
        """The query returns number of computers with remaining messages."""
        self.computer.last_exchange_time = datetime.utcnow()
        self.computer.remaining_count = 10
        self.computer2.last_exchange_time = datetime.utcnow()
        self.computer2.remaining_count = 4
        self.assertItemsEqual([2], self.execute_query()["computer_with_remaining_msg"])

    def test_only_active(self):
        """
        The query doesn't include computers that didn't exchange in the last
        hours.
        """
        self.computer.last_exchange_time = datetime.utcnow() - timedelta(hours=2)
        self.computer.remaining_count = 10
        self.computer2.last_exchange_time = datetime.utcnow()
        self.computer2.remaining_count = 4
        self.assertItemsEqual([1], self.execute_query()["computer_with_remaining_msg"])


class MsgCountsTest(OKRTest):

    query_name = "msg_counts"

    with_computer = True

    def setUp(self):
        super(MsgCountsTest, self).setUp()
        self.account2 = create_account("account2", "Account 2")
        license_management2 = LicenseManagement(self.account2)
        license_management2.create_license("BasicFeatures", 10)
        self.computer2 = self.account2.create_computer("c2", "Computer 2")

    def test_message_counts(self):
        """Message counts are returned."""
        self.computer.last_exchange_time = datetime.utcnow()
        self.computer.remaining_count = 10
        self.computer.dropped_count = 4
        self.computer.processed_count = 5
        self.computer.ignored_count = 2
        self.computer2.last_exchange_time = datetime.utcnow()
        self.computer2.remaining_count = 8
        self.computer2.dropped_count = 1
        self.computer2.processed_count = 33
        self.computer2.ignored_count = 21
        result = self.execute_query()
        self.assertItemsEqual([18], result["msg_remaining"])
        self.assertItemsEqual([5], result["msg_dropped"])
        self.assertItemsEqual([38], result["msg_processed"])
        self.assertItemsEqual([23], result["msg_ignored"])

    def test_message_counts_only_active(self):
        """
        Only computers that exchanged in the last hour are considered for
        message counts.
        """
        self.computer.last_exchange_time = datetime.utcnow() - timedelta(hours=2)
        self.computer.remaining_count = 10
        self.computer.dropped_count = 4
        self.computer.processed_count = 5
        self.computer.ignored_count = 2
        self.computer2.last_exchange_time = datetime.utcnow()
        self.computer2.remaining_count = 8
        self.computer2.dropped_count = 1
        self.computer2.processed_count = 33
        self.computer2.ignored_count = 21
        result = self.execute_query()
        self.assertItemsEqual([8], result["msg_remaining"])
        self.assertItemsEqual([1], result["msg_dropped"])
        self.assertItemsEqual([33], result["msg_processed"])
        self.assertItemsEqual([21], result["msg_ignored"])
