# Copyright 2005-2022 Canonical Limited. All rights reserved.

from functools import cmp_to_key
import itertools

from storm.locals import Int, Unicode, Store, List

from zope.interface import implementer

from landscape.constants import POLICY_ALLOW_INSTALLS

from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.activity.interfaces import (
    IRevertableRequest,
    IReappliableRequest,
    IExpandableRequest,
)
from canonical.landscape.model.activity.status import CANCELED
from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.eventlog.management import EventLog
from canonical.lib.text import describe_list
from functools import reduce


__all__ = [
    "ChangePackagesRequest",
    "UpgradeAllPackagesRequest",
    "ChangePackageProfilesRequest",
    "ChangeUpgradeProfilesRequest",
    "ChangePackagesPopulator",
    "ReleaseUpgradeRequest",
    "SUCCESS_RESULT",
    "ERROR_RESULT",
    "DEPENDENCY_ERROR_RESULT",
]

__clearlist__ = [
    "ChangePackagesRequest",
    "UpgradeAllPackagesRequest",
    "ChangePackageProfilesRequest",
    "ChangeUpgradeProfilesRequest",
    "ReleaseUpgradeRequest",
]

__requires_prefix__ = "landscape."


SUCCESS_RESULT = 1
ERROR_RESULT = 100
DEPENDENCY_ERROR_RESULT = 101
INCOMPATIBLE_CLIENT_API = 102

SHOW_PACKAGE_NAME_LIMIT = 3

CHANGE_GENERIC = 0
CHANGE_FEW = 1
CHANGE_REMOVAL = 2
CHANGE_INSTALL = 4
CHANGE_DOWNGRADE = 8
CHANGE_UPGRADE = 16
CHANGE_COMPLEMENTED = 32
CHANGE_HOLD = 64
CHANGE_REMOVE_HOLD = 128

CHANGE_TYPES = [
    CHANGE_REMOVAL,
    CHANGE_INSTALL,
    CHANGE_DOWNGRADE,
    CHANGE_UPGRADE,
    CHANGE_HOLD,
    CHANGE_REMOVE_HOLD,
]


def cmp(a, b):
    return (a > b) - (a < b)


def make_combinations(flags, minimum_combination_length):
    """
    L{make_combinations} generates a L{tuple} of numbers which represent all
    the possible integer values obtainable from a set of flags when they are
    combined with binary or.  Currently we don't try to remove duplicate
    values, because it costs time, but if performance becomes bad on lookups
    this would be an obvious optimisation.  The possible flag values are passed
    in as an iterable type called L{flags}, the L{minimum_combination_length}
    sets the minimum number of flags to be combined, the maximum length is the
    length of the variable L{flags}.

    For example, a simplified UNIX permissions value has 4 flag states:

       - NONE    = 0
       - EXECUTE = 1
       - READ    = 2
       - WRITE   = 4

    Calling L{make_combinations}([NONE, EXECUTE, READ, WRITE], 1) would
    generate the following L{tuple}:

        (0, 1, 2, 4, 1, 2, 4, 3, 5, 6, 3, 5, 6, 7, 7)

    """
    combination_lengths = list(range(minimum_combination_length, len(flags) + 1))
    combinations = (
        itertools.combinations(flags, combination_length)
        for combination_length in combination_lengths
    )
    return tuple(
        reduce(lambda x, y: x | y, combination)
        for combination in itertools.chain(*combinations)
    )


CHANGE_MIXED = make_combinations(CHANGE_TYPES, 2)


class UnknownPackage(object):
    """
    Represent a package unknown on the server, when a client has a wrong cache
    of package IDs.
    """

    name = "unknown"
    version = ""


unknown_package = UnknownPackage()


@requires_advisor
@implementer(IRevertableRequest, IReappliableRequest, IExpandableRequest)
class ChangePackagesRequestBase(ComputerRequest):

    simple_summary_messages = {
        CHANGE_GENERIC: "Install and/or remove packages",
        CHANGE_REMOVAL | CHANGE_FEW: ("Remove %(removal-names)s"),
        CHANGE_INSTALL | CHANGE_FEW: ("Install %(install-names)s"),
        CHANGE_HOLD | CHANGE_FEW: ("Hold %(hold-names)s"),
        CHANGE_REMOVE_HOLD | CHANGE_FEW: ("Remove hold on %(remove-hold-names)s"),
        CHANGE_DOWNGRADE | CHANGE_FEW: ("Downgrade %(downgrade-names)s"),
        CHANGE_UPGRADE | CHANGE_FEW: ("Upgrade %(upgrade-names)s"),
    }

    _packages = None

    reapply_arguments = {}

    changes = List(type=List(type=Int()))

    def __init__(self, activity, computer, **kwargs):
        changes = kwargs.pop("changes", [])
        super(ChangePackagesRequestBase, self).__init__(activity, computer, **kwargs)
        self.changes = changes

    def make_summary_messages(self, flags, info):
        """
        Generate a human readable text summary of the changes based on the
        contents of the L{info} dictionary and the L{flags} passed in.
        """
        simple_message = self.simple_summary_messages.get(flags, None)
        if simple_message:
            return simple_message % info
        message_parts = []
        if flags & CHANGE_REMOVAL:
            message_parts.append("remove %s" % info["removal"])
        if flags & CHANGE_INSTALL:
            message_parts.append("install %s" % info["install"])
        if flags & CHANGE_DOWNGRADE:
            message_parts.append("downgrade %s" % info["downgrade"])
        if flags & CHANGE_UPGRADE:
            message_parts.append("upgrade %s" % info["upgrade"])
        if flags & CHANGE_HOLD:
            message_parts.append("hold %s" % info["hold"])
        if flags & CHANGE_REMOVE_HOLD:
            message_parts.append("remove hold on %s" % info["remove-hold"])

        message = describe_list(items=message_parts, quote=False)
        return message.capitalize()

    def compare_changes(self, first, second):
        """Compare C{first} change to C{second} change.

        Complemented changes are always first, then we sort by flag values
        (remove, install, downgrade, upgrade). Finally we use package names to
        sort changes of the same type.
        """
        first_flag = first[0]
        second_flag = second[0]

        if first_flag != second_flag:
            if first_flag & CHANGE_COMPLEMENTED:
                if second_flag & CHANGE_COMPLEMENTED:
                    return cmp(first_flag, second_flag)
                else:
                    return -1
            if second_flag & CHANGE_COMPLEMENTED:
                return 1
            return cmp(first_flag, second_flag)

        first_pkg = self.activity.info.get_package(first[1])
        second_pkg = self.activity.info.get_package(second[1])

        first_pkg_name = first_pkg.name if first_pkg else None
        second_pkg_name = second_pkg.name if second_pkg else None
        if first_pkg_name != second_pkg_name:
            return cmp(first_pkg_name, second_pkg_name)

        # Fallback on package IDs if we don't know the package names.
        return cmp(first[1], second[1])

    def sort(self):
        """Sort C{self.changes}.

        We use copies to make sure that the changes list is never empty
        (because the list is empty during the sort call).
        """
        changes_copy = self.changes[:]
        changes_copy.sort(key=cmp_to_key(self.compare_changes))
        self.changes = changes_copy

    @requires("CreateActivity")
    def reapply(self, parent):
        """
        Reapply this package activity, creating a new L{ChangePackagesRequest}
        with the same changes.
        """
        activity = super(ChangePackagesRequestBase, self).reapply(parent)

        new_changes = [
            (change[0] - (change[0] & CHANGE_COMPLEMENTED), change[1], change[2])
            for change in self.changes
        ]
        activity.info.changes = new_changes
        return activity

    def get_package(self, package_id, fallback=False):
        """Retrieve the package object for C{package_id}, using a cache.

        @param fallback: if C{True}, returns C{unknown_package} if the package
            is not found (for compatibility purpose).
        """
        # If the package cache has been already initialized and the
        # package being requested is not already in the cache, it
        # means that the activity has been modified (ie, new children
        # have been added). We assume that ``get_package()`` is *only*
        # called from children activities and by no-one else.
        if self._packages is None or package_id not in self._packages:
            all_package_ids = set()
            for change in self.changes:
                flag = change[0]
                all_package_ids.add(change[1])
                if flag & CHANGE_DOWNGRADE or flag & CHANGE_UPGRADE:
                    all_package_ids.add(change[2])

            market = PackageMarket()
            result = market.get_many(all_package_ids)
            self._packages = packages = {}
            for package in result:
                packages[package.id] = package
        if fallback:
            default = unknown_package
        else:
            default = None
        return self._packages.get(package_id, default)

    def get_summary(self):
        """
        Build the summary for the package changes, construction details
        depending on the amount of packages.
        """
        flags = CHANGE_GENERIC
        info = {}

        removals = []
        installs = []
        holds = []
        remove_holds = []
        downgrades = []
        upgrades = []

        for change in self.changes:
            flag = change[0]
            if flag & CHANGE_REMOVAL:
                removals.append(change[1])
            elif flag & CHANGE_INSTALL:
                installs.append(change[1])
            elif flag & CHANGE_HOLD:
                holds.append(change[1])
            elif flag & CHANGE_REMOVE_HOLD:
                remove_holds.append(change[1])
            elif flag & CHANGE_UPGRADE:
                upgrades.append(change[1])
            elif flag & CHANGE_DOWNGRADE:
                downgrades.append(change[1])

        info["removal"] = len(removals)
        info["install"] = len(installs)
        info["hold"] = len(holds)
        info["remove-hold"] = len(remove_holds)
        info["downgrade"] = len(downgrades)
        info["upgrade"] = len(upgrades)

        for action, flag in (
            ("removal", CHANGE_REMOVAL),
            ("install", CHANGE_INSTALL),
            ("hold", CHANGE_HOLD),
            ("remove-hold", CHANGE_REMOVE_HOLD),
            ("downgrade", CHANGE_DOWNGRADE),
            ("upgrade", CHANGE_UPGRADE),
        ):
            if info[action]:
                flags |= flag

        # Optimize so that we only load packages to fetch the
        # package name if we have (a) few changes and (b)
        # non-mixed changes.
        if flags not in CHANGE_MIXED:
            for action, package_ids in (
                ("removal", removals),
                ("install", installs),
                ("hold", holds),
                ("remove-hold", remove_holds),
                ("downgrade", downgrades),
                ("upgrade", upgrades),
            ):
                if 0 < info[action] <= SHOW_PACKAGE_NAME_LIMIT:
                    flags |= CHANGE_FEW
                    info["%s-names" % action] = []
                    for package_id in package_ids:
                        package = self.get_package(package_id, fallback=True)
                        info["%s-names" % action].append(package.name)

        # Post-process ``info`` to compute correct plural form and
        # join package names, if needed.
        for action in (
            "removal",
            "install",
            "hold",
            "remove-hold",
            "downgrade",
            "upgrade",
        ):
            if info[action] == 1:
                info[action] = "%d package" % info[action]
            else:
                info[action] = "%d packages" % info[action]

            action_names = "%s-names" % action
            if action_names in info:
                info[action_names] = describe_list(
                    "package", info[action_names], quote=False
                )

        return self.make_summary_messages(flags, info)

    @requires("EditComputerPackages")
    def add_changes(
        self,
        install=(),
        remove=(),
        hold=(),
        remove_hold=(),
        populator=None,
        complemented=False,
    ):
        """Add package requests to this L{ChangePackagesRequest}.

        This should be called to add install/remove activities to this
        L{ChangePackagesRequest}. It can be called multiple times, and
        it will augment the activities without changing their state.

        @param install: List of package IDs to install.
        @param remove: List of package IDs to remove.
        @param hold: List of package IDs to be held.
        @param complemented: If True, indicates that changes have been
            introduced due to dependency requirements determined by the client.
        """
        run = False
        if populator is None:
            run = True
            populator = ChangePackagesPopulator()
        populator.add_changes(self.activity, install, remove, hold, remove_hold)
        if run:
            populator.run(complemented)

    @requires("EditComputerPackages")
    def complement_changes(
        self, install=(), remove=(), result_code=None, result_text=None
    ):
        """Add more package requests to this L{ChangePackages} activity.

        This should be called when a client comes back indicating that
        e.g. some extra package needs to be installed or removed.

        1. This L{ChangePackages} activity will be marked as unapproved.
        2. All the existing activities will be set back to undelivered.
        3. New activities will be created representing the package changes the
           client requested.  They will be created in the unapproved state.

        @param install: List of package IDs to install.
        @param remove: List of package IDs to remove.
        @param result_code: result code to set on C{change_packages}.
        @param result_text: result text to set on C{change_packages}.
        """
        if self.activity.status == CANCELED:
            # Ignore the changes if the activity is CANCELED. Nothing happened
            # on the client side, and the user asked to ignore this request.
            return

        self.activity.block()
        # XXX Manually update results. Eventually this should be dropped
        # and the info recorded elsewhere (perhaps the timeline), since
        # this is not a final state
        self.activity.result_text = result_text
        self.activity.result_code = result_code

        self.add_changes(install, remove, complemented=True)

        # If we only have complemented upgrades or install, we automatically
        # approve it.
        allowed = CHANGE_INSTALL | CHANGE_UPGRADE
        for change in self.changes:
            flag = change[0]
            if flag & CHANGE_COMPLEMENTED and not flag & allowed:
                break
        else:
            self.activity.unblock()

    def get_message(self, api):
        """
        Get a message to send to the client describing the changes to make.
        """
        removals = []
        installs = []
        holds = []
        hold_removals = []

        for change in self.changes:
            flag = change[0]
            if flag & CHANGE_INSTALL:
                installs.append(change[1])
            elif flag & CHANGE_REMOVAL:
                removals.append(change[1])
            elif flag & CHANGE_HOLD:
                holds.append(change[1])
            elif flag & CHANGE_REMOVE_HOLD:
                hold_removals.append(change[1])
            else:
                installs.append(change[1])
                removals.append(change[2])

        if not (installs or removals or holds or hold_removals):
            # No package changes to perfom
            return

        message = {"type": "change-packages", "operation-id": self.activity.id}
        if installs:
            message["install"] = installs

        if removals:
            message["remove"] = removals

        if holds:
            if not self.computer.is_api_compatible("3.4"):
                self.activity.fail(code=INCOMPATIBLE_CLIENT_API)
                return
            message["hold"] = holds

        if hold_removals:
            if not self.computer.is_api_compatible("3.4"):
                self.activity.fail(code=INCOMPATIBLE_CLIENT_API)
                return
            message["remove-hold"] = hold_removals

        if installs or removals or holds or hold_removals:
            # Add the binaries associated with the packages to install or
            # remove, if any. We include also the remove ones because we
            # might be asking a removal of a package which is not installed
            # anymore, and the client doesn't know its hash.
            market = PackageMarket()
            binaries = list(
                market.get_binaries(installs + removals + holds + hold_removals)
            )
            if binaries:
                if not self.computer.is_api_compatible("3.3"):
                    self.activity.fail(code=INCOMPATIBLE_CLIENT_API)
                    return
                message["binaries"] = binaries

        # If the client supports it, automatically complement the changes
        # when the only involve additional packages to be installed
        if self.computer.is_api_compatible("3.3"):
            message["policy"] = POLICY_ALLOW_INSTALLS

        return message

    def _started(self):
        message = self.get_message(self.computer.current_api)
        if message is not None:
            self.computer.get_message_store().add(message)

    def get_text(self):
        if self.activity.result_code == DEPENDENCY_ERROR_RESULT:
            return (
                "Requested changes require additional modifications "
                "in the package selection."
            )
        elif self.activity.result_code == INCOMPATIBLE_CLIENT_API:
            return (
                "Requested changes require a more recent version of "
                "the landscape-client package."
            )
        return self.activity.result_text

    def _create_activity(self, parent):
        """
        Return a C{ChangePackagesRequest} activity for the same computer.
        """
        return parent.add(self.__class__, computer=self.computer)

    @requires("EditComputerPackages")
    def revert(self, parent):
        """Create a reverted change packages request.

        @return: A new L{ChangePackagesRequest} request for the same computer
            and with the same creator, that will remove all packages installed
            by this activity and re-install all packages it removed.
        """
        super(ChangePackagesRequestBase, self).revert(parent)

        change_packages = self._create_activity(parent)

        new_changes = []

        for change in self.changes:
            flag = change[0]
            if flag & CHANGE_INSTALL:
                new_changes.append((CHANGE_REMOVAL, change[1], None))
            elif flag & CHANGE_REMOVAL:
                new_changes.append((CHANGE_INSTALL, change[1], None))
            elif flag & CHANGE_UPGRADE:
                new_changes.append((CHANGE_DOWNGRADE, change[2], change[1]))
            elif flag & CHANGE_DOWNGRADE:
                new_changes.append((CHANGE_UPGRADE, change[2], change[1]))
            elif flag & CHANGE_HOLD:
                new_changes.append((CHANGE_REMOVE_HOLD, change[1], None))
            elif flag & CHANGE_REMOVE_HOLD:
                new_changes.append((CHANGE_HOLD, change[1], None))
        change_packages.info.changes = new_changes

        EventLog(self.computer.account, parent.creator).create_log_entry(
            self.computer, "Activity created: " + change_packages.info.get_summary()
        )

        return change_packages

    @requires("ViewActivity")
    def has_expanded_content(self):
        """
        Return C{True} all the time, as an empty C{ChangePackagesRequest}
        doesn't make sense.
        """
        return True

    @requires("ViewActivity")
    def expanded_content(self):
        """
        Return a serialized content of attached L{ChangePackage}.
        """
        result = []
        for change in self.changes:
            package = self.get_package(change[1], fallback=True)
            flag = change[0]
            data = {
                "package": package.name,
                "complemented": bool(flag & CHANGE_COMPLEMENTED),
            }
            if flag & CHANGE_INSTALL:
                data["type"] = "install"
                data["version"] = package.version
            elif flag & CHANGE_REMOVAL:
                data["type"] = "remove"
                data["version"] = package.version
            elif flag & CHANGE_UPGRADE:
                data["type"] = "upgrade"
                data["install_version"] = package.version
                removed_package = self.get_package(change[2], fallback=True)
                data["remove_version"] = removed_package.version
            elif flag & CHANGE_DOWNGRADE:
                data["type"] = "downgrade"
                data["install_version"] = package.version
                removed_package = self.get_package(change[2], fallback=True)
                data["remove_version"] = removed_package.version
            elif flag & CHANGE_HOLD:
                data["type"] = "create-hold"
                data["version"] = package.version
            elif flag & CHANGE_REMOVE_HOLD:
                data["type"] = "remove-hold"
                data["version"] = package.version
            result.append(data)
        return result

    @requires("ViewActivity")
    def get_state(self, v2=False):
        """Return a dictionary containing package information."""
        standard_fields = super(ChangePackagesRequestBase, self).get_state(v2=v2)
        all_package_ids = set()
        for change in self.changes:
            flag = change[0]
            all_package_ids.add(change[1])
            if flag & CHANGE_DOWNGRADE or flag & CHANGE_UPGRADE:
                all_package_ids.add(change[2])
        standard_fields["package_ids"] = sorted(all_package_ids)
        standard_fields["changes"] = self.expanded_content()
        return standard_fields


@requires_advisor
class ChangePackagesRequest(ChangePackagesRequestBase):

    __storm_table__ = "change_package_request"

    requires("ViewComputerPackages", "EditComputerPackages", ["changes"])


@requires_advisor
class ChangePackageProfilesRequest(ChangePackagesRequestBase):
    """
    @ivar package_profile: Reference to the package profile which created this
        request. If the profile has been removed since the activity has been
        created, this can be C{None}.
    """

    __storm_table__ = "change_package_profile_request"

    requires("ViewComputerPackages", "EditComputerPackages", ["changes"])

    package_profile_id = Int()
    package_profile_name = Unicode()

    reapply_arguments = {"package_profile": "package_profile"}

    def __init__(self, activity, computer, package_profile):
        super(ChangePackageProfilesRequest, self).__init__(activity, computer)
        if package_profile is not None:
            self.package_profile_id = package_profile.id
            self.package_profile_name = package_profile.name
        else:
            self.package_profile_id = None
            self.package_profile_name = None

    def _create_activity(self, parent):
        """Create a new activity pointing to the same profile as self."""
        return parent.add(
            self.__class__, computer=self.computer, package_profile=self.package_profile
        )

    def get_summary(self):
        """Return information about which package profile is applied."""
        profile_name = self.package_profile_name
        if profile_name is not None:
            return "Apply package profile '%s'" % profile_name
        else:
            return "Profile has been removed"

    def is_actionable(self):
        """
        If no profile is linked, return C{False} so the activity can't be
        retried. Proceed normally otherwise.
        """
        if self.package_profile_id is None:
            return False
        else:
            return super(ChangePackageProfilesRequest, self).is_actionable()


@requires_advisor
class ChangeUpgradeProfilesRequest(ChangePackagesRequestBase):
    """
    @ivar upgrade_profile: Reference to the upgrade profile which created this
        request. If the profile has been removed since the activity has been
        created, this can be C{None}.
    """

    __storm_table__ = "change_upgrade_profile_request"

    requires("ViewComputerPackages", "EditComputerPackages", ["changes"])

    upgrade_profile_id = Int()

    reapply_arguments = {"upgrade_profile": "upgrade_profile"}

    def __init__(self, activity, computer, upgrade_profile):
        super(ChangeUpgradeProfilesRequest, self).__init__(activity, computer)
        if upgrade_profile is not None:
            self.upgrade_profile_id = upgrade_profile.id
        else:
            self.upgrade_profile_id = None

    def _create_activity(self, parent):
        """Create a new activity pointing to the same profile as self."""
        return parent.add(
            self.__class__, computer=self.computer, upgrade_profile=self.upgrade_profile
        )


class ChangePackagesPopulatorItem(object):
    def __init__(self, activity, installs, removals, holds, hold_removals):
        """A modification for a L{ChangePackagesRequest} activity.

        @param activity: Activity to be modified.
        @param install: List of package IDs to install.
        @param remove: List of package IDs to remove.
        @param holds: List of package IDs to hold.
        @param hold_removals: List of package IDs to remove the hold on.
        """
        self.activity = activity
        self.computer = activity.info.computer
        self.installs = installs
        self.removals = removals
        self.holds = holds
        self.hold_removals = hold_removals
        self.existing_installs = set()
        self.existing_removals = set()
        self.existing_holds = set()
        self.existing_hold_removals = set()


class ChangePackagesPopulator(object):
    """
    Populate package changes, creating the necessary L{ChangePackage} rows for
    the given activities.

    This class is used as an optimization, as it's able to handle changes for
    several computers at once, thus getting the packages and calculating the
    relations only once.
    """

    def __init__(self):
        self._changes = []

    def add_changes(self, activity, install, remove, hold, remove_hold):
        """Add a change to handle for the given activity.

        @param install: the list of package IDs which must be installed by the
            activity.
        @type install: C{list}
        @param remove: the list of package IDs which must be removed by the
            activity.
        @type remove: C{list}
        @param hold: the list of package IDs which must be held by the
            activity.
        @type hold: C{list}
        """
        self._changes.append(
            ChangePackagesPopulatorItem(activity, install, remove, hold, remove_hold)
        )

    def run(self, complemented=False):
        """Run the populator, creating the necessary L{ChangePackage}s.

        @param complemented: whether the new changes are client-request
            dependencies or not.
        """
        if not self._changes:
            return

        # Compute the complete set of installs and removals over the
        # requested modifications.
        installs = set()
        removals = set()
        holds = set()
        hold_removals = set()
        change_by_activity_id = {}

        for change in self._changes:
            installs.update(change.installs)
            removals.update(change.removals)
            holds.update(change.holds)
            hold_removals.update(change.hold_removals)
            # Also map parent activity id to change and collect the
            # parent activity ids, so that we can query all the
            # existing installs and removes below.
            change_by_activity_id[change.activity.id] = change

        # Compute existing installs and removes for all requested
        # changes in one go. Also take the opportunity and add those
        # installs and removals to the complete set.
        store = Store.of(self._changes[0].activity)
        change_package_class = self._changes[0].activity.info.__class__

        result = store.find(
            change_package_class,
            change_package_class.activity_id.is_in(change_by_activity_id),
        )
        for change_package in result:
            change = change_by_activity_id[change_package.activity_id]
            for change_detail in change_package.changes:
                flag = change_detail[0]
                if flag & CHANGE_INSTALL:
                    installs.add(change_detail[1])
                    change.existing_installs.add(change_detail[1])
                elif flag & CHANGE_REMOVAL:
                    removals.add(change_detail[1])
                    change.existing_removals.add(change_detail[1])
                elif flag & CHANGE_HOLD:
                    holds.add(change_detail[1])
                    change.existing_holds.add(change_detail[1])
                elif flag & CHANGE_REMOVE_HOLD:
                    hold_removals.add(change_detail[1])
                    change.existing_hold_removals.add(change_detail[1])

        # Fetch all the packages that we might be using here.
        market = PackageMarket()
        all_package_ids = (
            set(installs) | set(removals) | set(holds) | set(hold_removals)
        )
        packages = {}
        for package in market.get_many(all_package_ids):
            packages[package.id] = package

        if not installs and not removals and not holds and not hold_removals:
            return

        # Compute relationship information for the whole set.
        relation_info = market.get_package_relation_info(installs, removals)

        # Process upgrade and downgrades so that we only create them
        # if they are for a package with the same name, regardless of
        # the upgrade/downgrade relationship.
        upgrades = []  # [(install_package_id, remove_package_id), ...]
        downgrades = []  # [(install_package_id, remove_package_id), ...]

        for package_id, upgraded in list(relation_info.upgrades_map.items()):
            install_package = packages[package_id]
            for remove_package_id in upgraded:
                removal_package = packages[remove_package_id]
                if removal_package.name == install_package.name:
                    upgrades.append((package_id, remove_package_id))

        for package_id, downgraded in list(relation_info.downgraded_map.items()):
            install_package = packages[package_id]
            for remove_package_id in downgraded:
                removal_package = packages[remove_package_id]
                if removal_package.name == install_package.name:
                    downgrades.append((package_id, remove_package_id))

        for change in self._changes:
            for source, flag in (
                (upgrades, CHANGE_UPGRADE),
                (downgrades, CHANGE_DOWNGRADE),
            ):
                if complemented:
                    flag |= CHANGE_COMPLEMENTED
                for install_package_id, remove_package_id in source:
                    has_removal = (
                        remove_package_id in change.removals
                        or remove_package_id in change.existing_removals
                    )
                    has_install = (
                        install_package_id in change.installs
                        or install_package_id in change.existing_installs
                    )
                    if has_install and has_removal:
                        # Alright, we detected an upgrade or downgrade here.
                        # Update the 'changes' column of the
                        # ChangePackagesRequest object for this activity
                        # appropriately.
                        change.activity.info.changes.append(
                            [flag, install_package_id, remove_package_id]
                        )

                        # Now walk through existing and requested installs
                        # and removals and cancel them, since we turned
                        # them into an upgrade or downgrade already.
                        if install_package_id in change.installs:
                            change.installs.remove(install_package_id)
                        else:
                            for change_flag in (
                                CHANGE_INSTALL,
                                CHANGE_INSTALL | CHANGE_COMPLEMENTED,
                            ):
                                drop = [change_flag, install_package_id, None]
                                if drop in change.activity.info.changes:
                                    change.activity.info.changes.remove(drop)

                        if remove_package_id in change.removals:
                            change.removals.remove(remove_package_id)
                        else:
                            for change_flag in (
                                CHANGE_REMOVAL,
                                CHANGE_REMOVAL | CHANGE_COMPLEMENTED,
                            ):
                                drop = [change_flag, remove_package_id, None]
                                if drop in change.activity.info.changes:
                                    change.activity.info.changes.remove(drop)

            # Now that upgrades and downgrades have been handled, finally
            # create any leftover installs , removals, holds and hold removals.
            for source, flag in (
                (change.installs, CHANGE_INSTALL),
                (change.removals, CHANGE_REMOVAL),
                (change.holds, CHANGE_HOLD),
                (change.hold_removals, CHANGE_REMOVE_HOLD),
            ):
                if complemented:
                    flag |= CHANGE_COMPLEMENTED
                for package_id in source:
                    change.activity.info.changes.append([flag, package_id, None])
            change.activity.info.sort()


@requires_advisor
class UpgradeAllPackagesRequest(ChangePackagesRequest):
    @requires("ViewActivity")
    def has_expanded_content(self):
        """Return C{True} if this request has any changes attached to it."""
        return bool(self.changes)

    def get_summary(self):
        return "Upgrade all possible packages"

    def _started(self):
        if not self.changes:
            message = {
                "type": "change-packages",
                "upgrade-all": True,
                "operation-id": self.activity.id,
            }
            self.computer.get_message_store().add(message)
        super(UpgradeAllPackagesRequest, self)._started()


@requires_advisor
class ReleaseUpgradeRequest(ComputerRequest):
    """Upgrade a computer to a greater release.

    @ivar from_code_name: The code-name of the release to upgrade from.
    @ivar to_code_name: The code-name of the release to upgrade to.
    @ivar upgrade_tool_tarball_url: The URL of the upgrade-tool to use.
    @ivar upgrade_tool_tarball_signature_url: The upgrade-tool signature.
    """

    __storm_table__ = "release_upgrade_request"

    from_code_name = Unicode(allow_none=False)
    to_code_name = Unicode(allow_none=False)
    upgrade_tool_tarball_url = Unicode(allow_none=False)
    upgrade_tool_signature_url = Unicode(allow_none=False)

    def __init__(self, activity, computer, from_release, to_release):
        """
        @param activity: The L{Activity} this object is the info of.
        @param from_release: The L{MetaRelease} the computer is upgrading from.
        @param to_release: The L{MetaRelease} the computer is upgrading to.
        """
        self.from_code_name = from_release.code_name
        self.to_code_name = to_release.code_name
        self.upgrade_tool_tarball_url = to_release.upgrade_tool_tarball_url
        self.upgrade_tool_signature_url = to_release.upgrade_tool_signature_url
        super(ReleaseUpgradeRequest, self).__init__(activity, computer)

    @requires("ViewActivity")
    def get_summary(self):
        return "Upgrade from %s to %s" % (self.from_code_name, self.to_code_name)

    def _started(self):
        """Send a C{release-upgrade} message."""
        msg = {
            "type": "release-upgrade",
            "code-name": self.to_code_name,
            "upgrade-tool-tarball-url": self.upgrade_tool_tarball_url,
            "upgrade-tool-signature-url": self.upgrade_tool_signature_url,
            "debug": True,  # FIXME: these flags are hard-coded
            "allow-third-party": True,  # and should be exposed (LP:#490748)
            "operation-id": self.activity.id,
        }
        self.computer.get_message_store().add(msg)
        super(ReleaseUpgradeRequest, self)._started()
