# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import timedelta

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.resource.package import (
    get_computer_ids_with_preferences,
    get_computer_ids_with_security_upgrades,
    PackageReporterResultAPI,
)
from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.search import PackageSearch
from canonical.landscape.mail.template import load_template
from canonical.landscape.model.alert.computer import (
    ComputerRuleInfo,
    call_for_enabled_rules_with_transaction,
    ComputerAlertEmailViewBase,
)
from canonical.landscape.model.alert.rule import RuleState
from canonical.landscape.model.package.profile import (
    get_package_profiles_with_computer_ids,
)


__all__ = [
    "EsmDisabledAlert",
    "PackageUpgradesAlert",
    "SecurityUpgradesAlert",
    "PackageProfilesAlert",
    "check_security_upgrades_alerts",
    "PackageReporterAlert",
]


__requires_prefix__ = "landscape."


# Flag objects to get the state of a computers. See _check_computer_state
_UP_TO_DATE = object()
_UPGRADABLE = object()

# Known filename for ESM preferences files. This is created by the ua tool to
# disable EMS repositories using pinned priority.
ESM_UA_NAME_FILTER = "/etc/apt/preferences.d/ubuntu-esm-infra%"


@requires_advisor
class PackageUpgradesAlertBase(ComputerRuleInfo):
    @requires("EditAlert")
    def check(self, computer_ids):
        """
        Calculates the computers that have new upgrades and are up to date
        respectively.
        """
        stored_data = self.get_stored_state_data(computer_ids)
        data = self.get_state_data(computer_ids)
        upgradable_computer_ids = set()
        up_to_date_computer_ids = set()
        for computer_id in computer_ids:
            stored_state = PackageUpgradeRuleStateData.deserialize(
                stored_data.get(computer_id)
            )
            current_state = PackageUpgradeRuleStateData.deserialize(
                data.get(computer_id)
            )
            # Note that the following test works if stored_state
            # is currently None, which may happen with states that were
            # never checked and also at the time the server is updated
            # to support alert diffing (all past states are None).
            check = self._check_computer_state(stored_state, current_state)
            if check is _UP_TO_DATE:
                up_to_date_computer_ids.add(computer_id)
            elif check is _UPGRADABLE:
                upgradable_computer_ids.add(computer_id)

        self.update_states(upgradable_computer_ids, up_to_date_computer_ids)

    def _check_computer_state(self, stored_state, current_state):
        """
        Check the computer state from upgrades, only considering unheld
        upgrades.

        @param current_state: The state of the computer regarding package
            upgrades.
        @type current_state: L{PackageUpgradeRuleStateData}
        @param previous_state: The previous calculated state of the computer.
        @type previous_state: L{PackageUpgradeRuleStateData}
        @return: C{_UP_TO_DATE} or C{_UPGRADABLE} if the state has changed,
            C{None} if the state is the same
        """
        if stored_state.unheld_upgrades != current_state.unheld_upgrades:
            if current_state.unheld_upgrades == []:
                return _UP_TO_DATE
            else:
                # We only have to reactivate the state if the stored data
                # is None (in which case we have no idea about what the
                # previous state was), or if there are new entries.
                if stored_state.was_none or set(current_state.unheld_upgrades) - set(
                    stored_state.unheld_upgrades
                ):
                    return _UPGRADABLE

    def get_state_data(self, computer_ids):
        """See L{RuleInfo.get_state_data()}.

        State data for upgrades is currently stored as a list of package ids
        which are available for upgrade on each computer.
        """
        if not computer_ids:
            return {}

        selection = ComputerSelection(self.rule.account)
        selection.select_computers(computer_ids)

        package_search = PackageSearch(selection)
        upgrades = package_search.get_upgrades_by_computer_id()

        unknown_ids = set(computer_ids).difference(
            selection.get_selected_computer_ids()
        )
        upgrades.update((computer_id, ([], [])) for computer_id in unknown_ids)
        return upgrades


@requires_advisor
class PackageUpgradesAlert(PackageUpgradesAlertBase):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when package upgrades are available"

    notification_delay = timedelta(hours=6)

    resolution_url = (
        "/computers/criteria/alert:package-upgrades" "/packages/list?filter=upgrades"
    )

    @requires("ViewAlertRule")
    def get_summary(self, computer_ids=None):
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_ids = list(result.values(RuleState.computer_id))
        if not computer_ids:
            return "No package upgrades are available"
        elif len(computer_ids) > 1:
            return "%s computers have package upgrades available" % (len(computer_ids),)
        else:
            return "1 computer has package upgrades available"


@requires_advisor
class SecurityUpgradesAlert(PackageUpgradesAlertBase):

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when security upgrades are available"

    notification_delay = timedelta(hours=1)

    resolution_url = (
        "/computers/criteria/alert:security-upgrades" "/packages/list?filter=security"
    )

    def _check_computer_state(self, stored_state, current_state):
        """
        Check the computer state considering security upgrades, taking into
        account both unheld and held upgrades.

        @param current_state: The state of the computer regarding package
            upgrades.
        @type current_state: L{PackageUpgradeRuleStateData}
        @param previous_state: The previous calculated state of the computer.
        @type previous_state: L{PackageUpgradeRuleStateData}
        @return: C{_UP_TO_DATE} or C{_UPGRADABLE} if the state has changed,
            C{None} if the state is the same
        """
        if (
            stored_state.unheld_upgrades != current_state.unheld_upgrades
            or stored_state.held_upgrades != current_state.held_upgrades
        ):
            if (
                current_state.unheld_upgrades == []
                and current_state.held_upgrades == []
            ):
                return _UP_TO_DATE
            else:
                # We only have to reactivate the state if the stored data
                # is None (in which case we have no idea about what the
                # previous state was), or if there are new entries.
                if stored_state.was_none:
                    return _UPGRADABLE

                all_current_upgrades = set(current_state.unheld_upgrades).union(
                    current_state.held_upgrades
                )
                all_stored_upgrades = set(stored_state.unheld_upgrades).union(
                    stored_state.held_upgrades
                )
                if all_current_upgrades - all_stored_upgrades:
                    return _UPGRADABLE

    @requires("ViewAlertRule")
    def get_summary(self, computer_ids=None):
        result = self.rule.info.get_states(active=True, computer_ids=computer_ids)
        computer_ids = list(result.values(RuleState.computer_id))
        if not computer_ids:
            return "No security upgrades are available"
        elif len(computer_ids) > 1:
            return "%s computers have security upgrades available" % (
                len(computer_ids),
            )
        else:
            return "1 computer has security upgrades available"

    def get_state_data(self, computer_ids):
        """See L{RuleInfo.get_state_data()}.

        State data for security upgrades is currently stored as a list of
        package ids which are available for upgrade on each computer.
        """
        # First we get the same basic data considering all upgrades.
        base = super(SecurityUpgradesAlert, self)
        state_data = base.get_state_data(computer_ids)

        # Then we filter that to consider only security upgrades.
        # To do that, we first build a single set of all package ids.
        all_package_ids = set()
        for unheld_upgrades, held_upgrades in list(state_data.values()):
            all_package_ids.update(unheld_upgrades)
            all_package_ids.update(held_upgrades)

        # Now, we query the package database to see which of these
        # packages are known security upgrades.
        market = PackageMarket()
        result = market.get_many(all_package_ids, with_usn=True)
        security_upgrades_ids = set(result.values(Package.id))

        # Finally, we process the data to filter out any package ids
        # not in this set of security_upgrades.
        for unheld_upgrades, held_upgrades in list(state_data.values()):
            unheld_upgrades[:] = [
                id for id in unheld_upgrades if id in security_upgrades_ids
            ]
            held_upgrades[:] = [
                id for id in held_upgrades if id in security_upgrades_ids
            ]

        # PROFIT!
        return state_data


class PackageUpgradesEmailViewBase(ComputerAlertEmailViewBase):
    def __init__(self, context, person):
        super(PackageUpgradesEmailViewBase, self).__init__(context, person)
        self.all_unheld_upgrades = set()

    def get_upgrade_ids(self):
        """Get L{Package.id}s for all alerted computers."""
        upgrade_ids = set()

        # The state data for the package upgrade alert is a mapping of
        # computer ids to available upgrades, so reuse that data since
        # it's expensive to compute.
        stored_data = self._context.get_stored_state_data(
            [computer.id for computer in self.affected_computers]
        )
        for unheld_upgrades, held_upgrades in list(stored_data.values()):
            upgrade_ids.update(set(held_upgrades))
            unheld_upgrades = set(unheld_upgrades)
            upgrade_ids.update(unheld_upgrades)
            self.all_unheld_upgrades.update(unheld_upgrades)
        return upgrade_ids

    def get_package_data(self, with_usn=None):
        """
        Get a L{ResultSet} that yields 3-tuples of
        C{(Package.name, Package.version, Package.usn_name)}
        """
        market = PackageMarket()
        result = market.get_many(self.get_upgrade_ids(), with_usn=with_usn)
        result.order_by(Package.name, Package.version, Package.usn_name)
        return result.values(
            Package.id, Package.name, Package.version, Package.usn_name
        )


class PackageUpgradesEmailView(PackageUpgradesEmailViewBase):

    _content_template = load_template(__file__, "package-upgrades-alert.msg")

    def __init__(self, context, person):
        super(PackageUpgradesEmailView, self).__init__(context, person)
        self.security_upgrades = []
        self.upgrades = []
        for package_id, name, version, usn_name in self.get_package_data():
            # Package is ignored unless it is unheld on at least one
            # computer.
            if package_id not in self.all_unheld_upgrades:
                continue
            if usn_name:
                self.security_upgrades.append((name, version, usn_name))
            else:
                self.upgrades.append((name, version))


class SecurityUpgradesEmailView(PackageUpgradesEmailViewBase):

    _content_template = load_template(__file__, "security-upgrades-alert.msg")

    def __init__(self, context, person):
        super(SecurityUpgradesEmailView, self).__init__(context, person)
        self.security_upgrades = []
        for package_id, name, version, usn_name in self.get_package_data(with_usn=True):
            if package_id in self.all_unheld_upgrades:
                held = ""
            else:
                held = " [HELD]"
            self.security_upgrades.append((name, version, usn_name, held))


@requires_advisor
class PackageProfilesAlert(ComputerRuleInfo):
    """Alert notifying when a L{PackageProfile} has non-compliant computers.

    This alert is triggered whenever a computer associated with a certain
    profile does not have the relevant profile meta-package installed,
    and hence is not compliant with the profile.

    A check for the alert is needed in the following situations:

    1) A computer reports non-installed packages, as one of them could be
       a profile meta-package that got removed.

    2) A computer reports a failed change-packages request, as it might have
       been a request for installing a profile meta-package.

    3) When cancelling a computer L{ChangePackagesRequest} activity, as it
       might have been a request for installing a profile meta-package.

    @see: L{PackageProfile.get_broken_computers}
    """

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when a package profile is not applied"

    notification_delay = timedelta(hours=6)

    resolution_url = "/package-profile?filter=non-compliant"

    @requires("ViewAlertRule")
    def get_summary(self, computer_ids=None):
        """Get a summary stating how many computers have broken profiles.

        @param computer_ids: a list of computers ids to limit the generated
            summary to.
        """
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_ids = list(result.values(RuleState.computer_id))
        if not computer_ids:
            return "All computers are compliant with their profiles"
        elif len(computer_ids) > 1:
            return "%s computers are not compliant with their profiles" % (
                len(computer_ids),
            )
        else:
            return "1 computer is not compliant with its profiles"

    def check(self, computer_ids):
        """Check which computers have a broken profile and which not.

        This method checks all the L{PackageProfile}s associated with the given
        L{Computer.id}s, if a computer is not compliant on any of them, the its
        state will be activated, and if it is compliant it will be deactivated.

        @param computer_ids: Limit the check to the given computers.
        """
        if not computer_ids:
            return

        active_computer_ids = set()

        for profile in get_package_profiles_with_computer_ids(
            self.rule.account, computer_ids
        ):
            broken = profile.get_broken_computer_ids(computer_ids=computer_ids)
            active_computer_ids.update(broken)

        inactive_computer_ids = set(computer_ids) - active_computer_ids

        already_active = self.get_states(active=True, computer_ids=computer_ids).values(
            RuleState.computer_id
        )
        active_computer_ids = active_computer_ids.difference(already_active)

        self.update_states(active_computer_ids, inactive_computer_ids)


class PackageProfilesEmailView(ComputerAlertEmailViewBase):

    _content_template = load_template(__file__, "package-profiles-alert.msg")


def check_security_upgrades_alerts(package_ids):
    """
    Activate L{SecurityUpgradesAlert} rules for all L{Account}s that have
    L{Computer}s with at least one L{Package} from C{package_ids} as an
    available upgrade.  Alert rules that are already active will have their
    activation time updated.  Rules are processed one account at a time.

    @param package_ids: A list of L{Package.id}s affected by L{USN}s to limit
        checks with.
    """
    if not package_ids:
        return

    def callback(account, computer_ids):
        vulnerable_computer_ids = set(
            get_computer_ids_with_security_upgrades(computer_ids, package_ids)
        )
        if vulnerable_computer_ids:
            account.check_rules(
                [SecurityUpgradesAlert], computer_ids=vulnerable_computer_ids
            )

    call_for_enabled_rules_with_transaction(SecurityUpgradesAlert, callback)


class PackageUpgradeRuleStateData(object):
    """
    Access logic for the L{RuleState} data blob for package upgrade alerts.

    @ivar unheld_upgrades: List of unheld upgrade package ids.
    @ivar held_upgrades: List of upgrade package ids where either the
        installed package or the upgrade is held.
    """

    def __init__(self, unheld_upgrades=None, held_upgrades=None):
        if unheld_upgrades:
            self.unheld_upgrades = unheld_upgrades
        else:
            self.unheld_upgrades = []
        if held_upgrades:
            self.held_upgrades = held_upgrades
        else:
            self.held_upgrades = []
        # This is needed in order to capture whether this instance was
        # deserialized from a C{None} stored state.
        self.was_none = None

    def serialize(self):
        """Return data in a format suitable for pickling."""
        return (self.unheld_upgrades, self.held_upgrades)

    @staticmethod
    def deserialize(data):
        """Factory method, creates instances from pickled data.

        The data passed must either be a 2-tuple or None or a list.
        """
        if isinstance(data, tuple) and len(data) == 2:
            data = PackageUpgradeRuleStateData(*data)
            data.was_none = False
        elif data is None:
            data = PackageUpgradeRuleStateData()
            data.was_none = True
            data.unheld_upgrades = None
            data.held_upgrades = None
        else:
            raise ValueError("Invalid 'data' parameter: '%s'" % data)

        return data

    def add_upgrades(self, upgrades, held=False):
        """Add the given C{upgrades} observing the C{held} flag."""
        if not upgrades:
            return
        if held:
            self.held_upgrades = sorted(set(self.held_upgrades) | set(upgrades))
        else:
            self.unheld_upgrades = sorted(set(self.unheld_upgrades) | set(upgrades))

    def remove_upgrades(self, upgrades, held=False):
        """Remove the given C{upgrades} observing the C{held} flag."""
        if not upgrades:
            return
        if held:
            self.held_upgrades = sorted(set(self.held_upgrades) - set(upgrades))
        else:
            self.unheld_upgrades = sorted(set(self.unheld_upgrades) - set(upgrades))


@requires_advisor
class PackageReporterAlert(ComputerRuleInfo):
    """
    This alert is triggered when a computer fails to give a successful result
    from the package reporter.
    """

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when package reporting fails"

    notification_delay = timedelta(hours=6)

    resolution_url = "/computers/criteria/alert:package-reporter" "/packages/reporter"

    @requires("ViewAlertRule")
    def get_summary(self, computer_ids=None):
        """
        Return a summary indicating how many computers fail to report packages,
        if any.
        """
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_ids = list(result.values(RuleState.computer_id))
        if not computer_ids:
            return "All computers are successfully reporting packages"
        elif len(computer_ids) > 1:
            return "%s computers are having issues reporting packages" % (
                len(computer_ids),
            )
        else:
            return "1 computer is having issues reporting packages"

    def check(self, computer_ids):
        """Check if the alert is active for the given computers.

        The alert is set if the exit code if the code is different from 0, and
        different from 1 with an empty error string. The latter is considered
        as false positive when the package reporter fails to run because the
        last run is too recent.
        """
        if not computer_ids:
            return

        api = PackageReporterResultAPI(self.rule.account)
        data = api.get_package_reporter_errors(computer_ids)
        active_computer_ids = set(i[0] for i in data)

        inactive_computer_ids = set(computer_ids) - active_computer_ids

        already_active = self.get_states(active=True, computer_ids=computer_ids).values(
            RuleState.computer_id
        )
        active_computer_ids = active_computer_ids.difference(already_active)

        self.update_states(active_computer_ids, inactive_computer_ids)


class PackageReporterEmailView(ComputerAlertEmailViewBase):

    _content_template = load_template(__file__, "package-reporter-alert.msg")


@requires_advisor
class EsmDisabledAlert(ComputerRuleInfo):
    """
    This alert is triggered when a computer has ESM disabled
    through known apt preferences.
    """

    requires("ViewAlertRule", PRIVATE, ["title"])

    title = "Alert when ESM updates are disabled"

    notification_delay = timedelta(hours=1)

    resolution_url = "/computers/criteria/alert:esm-disabled/packages"

    @requires("ViewAlertRule")
    def get_summary(self, computer_ids=None):
        """
        Return a summary indicating how many computers fail to report packages,
        if any.
        """
        result = self.get_states(active=True, computer_ids=computer_ids)
        computer_ids = list(result.values(RuleState.computer_id))
        if not computer_ids:
            return "No computer has ESM disabled"
        elif len(computer_ids) > 1:
            return "{} computers have ESM disabled".format(
                len(computer_ids),
            )
        else:
            return "1 computer has ESM disabled"

    def check(self, computer_ids):
        """Check if the alert is active for the given computers.

        The alert is set if an "ubuntu-esm-infra*" apt preferences file
        exists for that computer.
        """
        if not computer_ids:
            return

        data = get_computer_ids_with_preferences(computer_ids, ESM_UA_NAME_FILTER)
        active_computer_ids = set(data)

        inactive_computer_ids = set(computer_ids) - active_computer_ids

        already_active = self.get_states(active=True, computer_ids=computer_ids).values(
            RuleState.computer_id
        )
        active_computer_ids = active_computer_ids.difference(already_active)

        self.update_states(active_computer_ids, inactive_computer_ids)


class EsmDisabledEmailView(ComputerAlertEmailViewBase):

    _content_template = load_template(__file__, "esm-disabled-alert.msg")
