# Copyright 2005-2022 Canonical Limited. All rights reserved.

from dateutil.rrule import rrulestr
import pytz

import logging

import transaction

from storm.locals import Bool, Int, Unicode, Reference, AutoReload, DateTime, Enum

from zope.interface import implementer
from zope.security.interfaces import Unauthorized

from canonical.lib.timestamp import utcnow
from datetime import timedelta

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.account.helper import ComputerGroupHelper
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.package.management import PackageManagement
from canonical.landscape.model.package.interfaces import IUpgradeProfile
from canonical.landscape.model.package.activities import ChangeUpgradeProfilesRequest
from canonical.landscape.model.package.errors import UnknownUpgradeProfileError
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.status import QUEUED
from canonical.landscape.model.activity.group import validate_delivery_time
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.main.interfaces import IAccessContextComposite
from canonical.lib.text import slug_from_title, describe_dict


__all__ = [
    "UpgradeProfile",
    "UpgradeProfileManagement",
    "ScheduleFormatException",
    "run_upgrade_profiles",
    "UPGRADE_TYPE_MAP",
    "get_upgrade_type",
    "schedule_to_dict",
    "schedule_to_string",
    "WEEKDAY_MAP",
    "WEEKDAY_DISPLAY_MAP",
]

__clearlist__ = [
    "UpgradeProfileManagement",
    "UpgradeProfile",
    "get_upgrade_type",
    "schedule_to_dict",
    "schedule_to_string",
    "WEEKDAY_MAP",
    "WEEKDAY_DISPLAY_MAP",
]

__requires_prefix__ = "landscape."

UPGRADE_TYPE_MAP = {"all": 0, "security": 1}

WEEKDAY_MAP = {
    "mo": "monday",
    "tu": "tuesday",
    "we": "wednesday",
    "th": "thursday",
    "fr": "friday",
    "sa": "saturday",
    "su": "sunday",
}


WEEKDAY_DISPLAY_MAP = {
    "mo": "Mon",
    "tu": "Tue",
    "we": "Wed",
    "th": "Thu",
    "fr": "Fri",
    "sa": "Sat",
    "su": "Sun",
}


class ScheduleFormatException(Exception):
    """
    This exception is raised when the schedule format is invalid, for instance
    in UpgradeProfile.calculate_next_run_time.
    """

    def __init__(self, message):
        self.message = message


@requires_advisor
@implementer(IUpgradeProfile, IAccessContextComposite)
class UpgradeProfile(ComputerGroupHelper):
    """
    An upgrade profile, holding information about automatic upgrades on set of
    machines.

    @ivar title: A title for this profile.
    @ivar upgrade_type: The type of upgrade this profile will execude. Can be
        "all" or "security" for all updates or security updates only,
        respectively.
    @ivar schedule: The iCal representaion of the recurrence rule for This
        upgrade profile.
    @ivar run_window_length: The length of the run window in hours. This is
        added to the next_run_time to get the datetime after which the
        activities created by this profile should not be delivered anymore.
    @ivar creation_time: The time at which this profile was created.
    @ivar modification_time: The time at which this profile was last modified.
    @ivar next_run_time: The time at which this profile will be run (will
        create activities).
    @ivar name: An optional name for this profile (must be a valid slug)
    @ivar deliver_after_window: If non-zero, the C{deliver_after} parameter of
        child activities in this group will be set to random time values of at
        most C{deliver_after_window} minutes in the future, so they won't be
        delivered immediately but rather spread over this time window.
    @ivar timezone: The timezone the profile is being created for.
    @ivar autoremove: Autoremove dependencies.
    """

    __storm_table__ = "upgrade_profile"

    requires(
        "ViewUpgradeProfile",
        PRIVATE,
        [
            "id",
            "account",
            "all_computers",
            "creation_time",
            "modification_time",
            "next_run_time",
            "access_context",
            "get_computers",
            "get_computer_count",
            "get_computer_ids",
            "get_tags",
            "get_state",
            "deliver_after_window",
        ],
    )
    requires(
        "ViewUpgradeProfile",
        "EditUpgradeProfile",
        [
            "name",
            "title",
            "upgrade_type",
            "schedule",
            "security_upgrade",
            "run_window_length",
            "computer_group_id",
            "timezone",
            "autoremove",
        ],
    )
    requires(
        "EditUpgradeProfile",
        PRIVATE,
        [
            "update_association",
            "update_tags",
            "enable_for_all_computers",
            "disable_for_all_computers",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    title = Unicode(allow_none=False)
    upgrade_type = Enum(map=UPGRADE_TYPE_MAP, allow_none=False)
    schedule = Unicode(allow_none=False)
    run_window_length = Int()
    creation_time = DateTime(default=AutoReload)
    modification_time = DateTime(default=AutoReload)
    next_run_time = DateTime()
    deliver_after_window = Int(default=0)
    access_context_id = Int(allow_none=False)
    timezone = Unicode(allow_none=False)
    autoremove = Bool(allow_none=False, default=False)

    computer_group_id = Int()
    computer_group = Reference(computer_group_id, ComputerGroup.id)

    def __init__(
        self,
        account,
        name,
        title,
        upgrade_type,
        schedule,
        deliver_within=1,
        access_context_id=None,
        deliver_after_window=0,
        timezone="UTC",
        autoremove=False,
    ):
        super(UpgradeProfile, self).__init__(account, access_context_id)
        self._create_computer_group()
        self.title = title
        self.name = name  # Should be a unique slug from management class
        self.upgrade_type = upgrade_type
        self.schedule = schedule
        self.run_window_length = deliver_within
        self.timezone = timezone
        self.calculate_next_run_time()
        self.deliver_after_window = deliver_after_window
        self.autoremove = autoremove

    def _get_security_upgrade(self):
        if self.upgrade_type == "security":
            return True
        return False

    def _set_security_upgrade(self, upgrade):
        self.upgrade_type = get_upgrade_type(upgrade)

    security_upgrade = property(_get_security_upgrade, _set_security_upgrade)

    @requires("EditUpgradeProfile")
    def calculate_next_run_time(self):
        """Compute the value of C{next_run_time} using C{schedule}.

        This uses L{rrulestr} and the format defined in
        U{RFC 2445 <http://www.ietf.org/rfc/rfc2445.txt>}.
        """
        # Do not depend on server time. grab the utc time and convert it to a
        # now in the user timezone.
        tz = pytz.timezone(self.timezone)
        localized_utc_now = pytz.utc.localize(utcnow())
        utc_now_in_local_timezone = localized_utc_now.astimezone(tz)
        from_date = utc_now_in_local_timezone.replace(tzinfo=None)

        try:
            rrule = rrulestr(self.schedule, dtstart=from_date)
        except ValueError:
            raise ScheduleFormatException(
                "The schedule format must comply with RFC 2445"
            )

        self.next_run_time = rrule.after(from_date)

    def run(self):
        """
        Create the activities for applying upgrades, and update the
        C{next_run_time}.

        @return: C{True} if some activities were created, C{False} otherwise.
        """
        self.calculate_next_run_time()

        computers = list(self.get_computers())
        if not computers:
            return False

        selection = ComputerSelection(self.account)
        selection.select_computers(computers)
        management = PackageManagement(selection)
        search = management.get_search()
        changes = search.get_package_upgrades_by_computer_id(
            security_only=self.upgrade_type == "security", autoremove=self.autoremove
        )

        all_computer_ids = [
            computer
            for computer, installs, removals, holds, hold_removals in changes
            if installs or removals
        ]

        if not all_computer_ids:
            return False

        pending_computer_ids = self.get_pending_activities(all_computer_ids).values(
            AccountComputerTwin.id
        )

        # Remove computers which have pending changes created by this profile.
        new_changes = [
            (computer, installs, removals, holds, hold_removals)
            for computer, installs, removals, holds, hold_removals in changes
            if (installs or removals) and computer not in pending_computer_ids
        ]

        if not new_changes:
            return False

        # There is no need to take the display time zone into account for this
        # calculation because the relevant times are stored in UTC internally.
        deliver_before = utcnow() + timedelta(hours=self.run_window_length)

        # Ask the PackageManagement to create activities
        management.change_packages(
            None,
            new_changes,
            deliver_before=deliver_before,
            upgrade_profile=self,
            deliver_after_window=self.deliver_after_window,
        )
        return True

    def get_state(self):
        """Return a dict with information about this profile."""
        all_computers = self.computer_group.all_computers
        tags = sorted(tag.name for tag in self.get_tags())
        result = {
            "id": self.id,
            "title": self.title,
            "name": self.name,
            "access_group": self.access_context.name,
            "upgrade_type": self.upgrade_type,
            "autoremove": self.autoremove,
            "tags": tags,
            "all_computers": all_computers,
            "deliver_within": str(self.run_window_length),
            "deliver_delay_window": str(self.deliver_after_window),
        }
        result.update(schedule_to_dict(self.schedule))
        return result

    @requires("ViewUpgradeProfile")
    def get_pending_activities(self, computer_ids):
        collection = ActivityChildCollection(
            self.account.store,
            account=self.account,
            statuses=[QUEUED],
            info_classes=[ChangeUpgradeProfilesRequest],
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
        )
        result = collection.find()
        return result.find(
            ChangeUpgradeProfilesRequest.activity_id == Activity.id,
            ChangeUpgradeProfilesRequest.upgrade_profile_id == self.id,
        )

    @requires("EditUpgradeProfile")
    def remove(self):
        super(UpgradeProfile, self).remove()


ChangeUpgradeProfilesRequest.upgrade_profile = Reference(
    ChangeUpgradeProfilesRequest.upgrade_profile_id, UpgradeProfile.id
)


@requires_advisor
class UpgradeProfileManagement(ModelManagement):

    requires("ViewAccount", PRIVATE, ["account", "person"])

    @requires("ViewAccount")
    def create_upgrade_profile(
        self,
        title,
        upgrade_type,
        schedule,
        deliver_within=1,
        access_context_name=(DEFAULT_ACCESS_CONTEXT_NAME),
        deliver_after_window=0,
        timezone="UTC",
        autoremove=False,
    ):
        """Create an L{UpgradeProfile} in the database."""

        validate_delivery_time(
            run_window_length=deliver_within, deliver_after_window=deliver_after_window
        )

        def existing_names(slug):
            """Return a set of existing names in this account"""
            return self.account.store.find(
                UpgradeProfile.name, UpgradeProfile.account_id == self.account.id
            )

        name = slug_from_title(title, 50, existing_names)
        access_context = self.get_access_context(
            access_context_name, "EditUpgradeProfile"
        )

        profile = UpgradeProfile(
            self.account,
            name,
            title,
            upgrade_type,
            schedule,
            deliver_within=deliver_within,
            access_context_id=access_context.id,
            deliver_after_window=deliver_after_window,
            timezone=timezone,
            autoremove=autoremove,
        )
        prefix = "Upgrade profile created with"
        values = {
            "name": name,
            "title": title,
            "upgrade type": upgrade_type,
            "autoremove": autoremove,
            "deliver within": deliver_within,
            "schedule:": schedule_to_string(profile.schedule),
            "randomised delay window": deliver_after_window,
            "access group": access_context.name,
        }
        message = "%s." % describe_dict(prefix, values, pluralize=False)
        self._event_log.create_log_entry(profile, message)
        return profile

    @requires("ViewUpgradeProfile")
    def edit(
        self,
        name,
        title=None,
        security_upgrade=None,
        schedule=None,
        deliver_within=None,
        new_tags=None,
        computer_association=None,
        deliver_after_window=None,
        timezone=None,
        autoremove=None,
    ):
        """
        Handle all L{UpgradeProfile} edits within the UpgradeProfileManagement
        for auditing purposes.
        """
        validate_delivery_time(
            run_window_length=deliver_within, deliver_after_window=deliver_after_window
        )
        access_context_ids = self.get_access_context_ids("EditUpgradeProfile")
        profile = self.get_upgrade_profile_by_name(name)
        if profile.access_context.id not in access_context_ids:
            raise Unauthorized(
                "Permission denied. Can't change upgrade profile '%s'" % name
            )

        profile.update_association(new_tags, computer_association)
        update_runtime = False

        messages = []
        if title and title != profile.title:
            messages.append("Title changed from '%s' to '%s'." % (profile.title, title))
            profile.title = title
        if schedule and schedule != profile.schedule:
            messages.append("Schedule changed to '%s'." % schedule_to_string(schedule))
            profile.schedule = schedule
            update_runtime = True
        if autoremove is not None and autoremove != profile.autoremove:
            messages.append("Autoremove changed to '{}'.".format(autoremove))
            profile.autoremove = autoremove
        if security_upgrade is not None:
            if security_upgrade != ("security" == profile.upgrade_type):
                profile.security_upgrade = security_upgrade
                if security_upgrade:
                    messages.append("Activated security only.")
                else:
                    messages.append("Deactivated security only.")
        if deliver_within and deliver_within != profile.run_window_length:
            messages.append(
                "Run window length updated from %d to %d."
                % (profile.run_window_length, deliver_within)
            )
            profile.run_window_length = deliver_within
            update_runtime = True
        if update_runtime:
            profile.calculate_next_run_time()

        if deliver_after_window is not None:
            if profile.deliver_after_window != deliver_after_window:
                messages.append(
                    "Deliver after window updated from %d to %d."
                    % (profile.deliver_after_window, deliver_after_window)
                )
                profile.deliver_after_window = deliver_after_window

        if timezone is not None and timezone != profile.timezone:
            previous = profile.timezone
            profile.timezone = timezone
            messages.append(
                "Time zone updated from %s to %s." % (previous, profile.timezone)
            )

        if messages:
            self._event_log.create_log_entry(profile, " ".join(messages))
        return profile

    @requires("ViewAccount")
    def remove_upgrade_profiles(self, profiles):
        """Remove given L{UpgradeProfile}s from the account.

        Only profiles that the accessing person has EditUpgradeProfile on will
        be removed.

        @return: A C{list} of L{UpgradeProfile}s that were not removed.
        """
        access_context_ids = self.get_access_context_ids("EditUpgradeProfile")

        remaining_profiles = []
        for profile in profiles:
            if profile.access_context.id in access_context_ids:
                self._event_log.create_log_entry(profile, "Upgrade profile removed.")
                profile.remove()
            else:
                remaining_profiles.append(profile)

        return remaining_profiles

    @requires("ViewAccount")
    def get_upgrade_profiles(
        self, ids=None, names=None, titles=None, upgrade_type=None
    ):
        """
        Returns a list of L{UpgradeProfile} for the current account matching
        the provided parameters.
        """
        where = [UpgradeProfile.account_id == self.account.id]
        # XXX don't enforce permission check for now, if the person is not
        # provided to the class. This should be required once the person is
        # mandatory.
        if self.person is not None:
            access_context_ids = self.get_access_context_ids("ViewUpgradeProfile")
            where.extend(
                [
                    UpgradeProfile.computer_group_id == ComputerGroup.id,
                    UpgradeProfile.access_context_id.is_in(access_context_ids),
                ]
            )
        if ids:
            where.append(UpgradeProfile.id.is_in(ids))
        if names:
            where.append(UpgradeProfile.name.is_in(names))
        if titles:
            where.append(UpgradeProfile.title.is_in(titles))
        if upgrade_type is not None:
            where.append(UpgradeProfile.upgrade_type == upgrade_type)
        result = self.account.store.find(UpgradeProfile, *where)
        return result.order_by(UpgradeProfile.name)

    def get_upgrade_profile_by_name(self, name):
        """Get a single L{UpgradeProfile} by name.

        @raises: UnknownUpgradeProfileError if no profile with the given name
            exists in the account.
        """
        profile = self.get_upgrade_profiles(names=[name]).one()
        if not profile:
            raise UnknownUpgradeProfileError(
                "No upgrade profile found with name '%s'." % name
            )
        return profile


def run_upgrade_profiles():
    """Get all L{UpgradeProfile}s that needs to run, and run them."""
    localized_utc_now = pytz.utc.localize(utcnow())
    for name, store in get_account_stores():
        result = store.find(UpgradeProfile)
        result.order_by(UpgradeProfile.next_run_time)
        for profile in result:
            tz = pytz.timezone(profile.timezone)
            local_now = localized_utc_now.astimezone(tz).replace(tzinfo=None)
            if profile.next_run_time < local_now:
                try:
                    if profile.run():
                        logging.info(
                            "Applied upgrade profile '%s' in account '%s'"
                            % (profile.name, profile.account.name)
                        )
                except Exception as e:
                    # Generally we don't want to catch all exceptions in our
                    # model because it might hide unexpected issues in our
                    # code, but run_upgrade_profiles is only called in a cron
                    # job currently and we want that script to attempt running
                    # all profile jobs as well as exit cleanly. So let's log
                    # the exceptions and move on.
                    logging.error(
                        "Unable to apply upgrade profile '%s' in account '%s'."
                        " %s" % (profile.name, profile.account.name, str(e))
                    )
                else:
                    transaction.commit()


def get_upgrade_type(security_upgrade):
    """
    Given whether the upgrade should be a security upgrade, return the correct
    textual representation of it ("all" or "security").
    """
    return "security" if security_upgrade is True else "all"


def schedule_to_dict(schedule):
    """Takes a schedule string and returns a dict of the interesting fields."""
    result = {}
    splits = schedule.split(";")
    for split in splits:
        token, value = split.split("=")
        if token == "BYMINUTE":
            result["at_minute"] = value
        elif token == "BYHOUR":
            result["at_hour"] = value
        elif token == "BYDAY":
            result["on_days"] = value.split(",")
        elif token == "FREQ":
            if value == "WEEKLY":
                result["every"] = "week"
            elif value == "HOURLY":
                result["every"] = "hour"
            else:
                raise ValueError("Unsupported frequency: %r" % value)
        else:
            raise ValueError("Unsupported token: %r" % token)
    return result


def schedule_to_string(schedule):
    """Takes a schedule string and returns a human readable string"""
    schedule_dict = schedule_to_dict(schedule)
    days_string = ",".join(
        WEEKDAY_DISPLAY_MAP[day] for day in schedule_dict.get("on_days", [])
    )
    if schedule_dict.get("every") == "hour":
        result = "Every hour at %s minutes past the hour" % (schedule_dict["at_minute"])
        if days_string:
            result += " on every %s" % days_string
    elif schedule_dict.get("every") == "week":
        result = "Every %s at %02d:%02d" % (
            days_string,
            int(schedule_dict["at_hour"]),
            int(schedule_dict["at_minute"]),
        )
    return result
