# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, RawStr, Reference

from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.package import Package


__all__ = ["PackageBinary", "set_package_binary", "get_package_binaries"]


class PackageBinary(object):
    """Hold the binary data of an actual package.

    @note: In principle this information could be attached as a column to
        to the L{Package} class/table, however to save space and gain some
        performance it is kept separately.
    """

    __storm_table__ = "package_binary"

    package_id = Int(primary=True, allow_none=False)
    data = RawStr(allow_none=False)

    package = Reference(package_id, Package.id)

    def __init__(self, package, data):
        """
        @param package: The L{Package} we're holding the binary of.
        @param data: The package binary data we hold.
        """
        self.package = package
        self.data = data


def set_package_binary(package, data):
    """Set the binary data for the given package.

    If no L{PackageBinary} exists yet for the package, a new one will be
    created, otherwise its data will be replaced.

    @param package: The L{Package} the binary is for.
    @param data: The binary data to store.
    """
    store = get_package_store()
    binary = store.get(PackageBinary, package.id)
    if binary is not None:
        binary.data = data
    else:
        binary = store.add(PackageBinary(package, data))
    return binary


def get_package_binaries(package_ids):
    """Get the package binary data for the given L{Package.id}s.

    @return: A L{ResultSet} yielding binary tuples of package ids and
        package binaries.
    """
    store = get_package_store()
    result = store.find(PackageBinary, PackageBinary.package_id.is_in(package_ids))
    return result.values(PackageBinary.package_id, PackageBinary.data)
