# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import object
from re import findall
from json import dumps, loads

from zope.interface import implementer

from landscape.lib.fetch import fetch, PyCurlError, HTTPCodeError

from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.package.errors import (
    PackageSearchRequestError,
    PackageSearchConnectionError,
)


@implementer(IPackageSearchClient)
class PackageSearchClient(object):
    """Perform synchronous HTTP requests to the package search service."""

    # Because of generic getattr implementation below, for newer zope.interface
    # versions (included in Ubuntu 14.10), we should either fail with an
    # AttributeError for __component_name__ or return an empty unicode string
    # (which it defaults to when __component_name__ is not defined).
    # Alternatively, we can special case it in __getattr__ implementation.
    __component_name__ = ""

    def __init__(self, host, port, fetch=fetch):
        self.host = host
        self.port = port
        self.fetch = fetch

    def __getattr__(self, name):
        method = self._camel(name)

        def query(**kwargs):
            params = dict((self._camel(key), val) for key, val in kwargs.items())
            return self._query(method, params)

        return query

    def _camel(self, term):
        """Convert the given term to CamelCase."""
        return "".join([part.capitalize() for part in term.split("_")])

    def _uncamel(self, term):
        """Convert the given CamelCase term to Python-style."""
        words = findall("[A-Z][a-z]+", term)
        return "_".join(word.lower() for word in words)

    def _query(self, method, params):
        """Perform the HTTP request with the given parameters."""
        url = "http://%s:%d/%s" % (self.host, self.port, method)
        data = dumps(params)
        try:
            response = self.fetch(url, post=True, data=data, total_timeout=60)
        except PyCurlError as error:
            raise PackageSearchConnectionError(str(error))
        except HTTPCodeError as error:
            raise PackageSearchRequestError(loads(error.body)["Error"])
        result = loads(response)
        return dict((self._uncamel(key), val) for key, val in result.items())
