# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Collection logic to filter and find L{Package}s and L{USN}s."""

import operator
import re

from storm.expr import And, Column, Desc, Like, Or
from storm.store import EmptyResultSet

from canonical.collection import CollectionBase

from canonical.landscape.model.textsearch import (
    TSRankCD,
    Match,
    get_search_query,
    AND,
    RANK_LOG_LENGTH,
    Similarity,
)
from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.usn import USN

COMPARISON_OPERATORS_PATTERN = r"[^A-Za-z0-9\s\-\+\.]+"


__all__ = [
    "PackageCollection",
    "USNCollection",
    "parse_package_names",
    "COMPARISON_OPERATORS_PATTERN",
    "get_package_search_rank",
]


class PackageCollection(CollectionBase):
    """A collection of L{Package}s.

    L{Package}s in the collection are ordered by L{Package.name} and
    L{Package.version}, by default.  If text search is used results are order
    by rank first, then name and version.

    @param store: A package C{Store} to search.
    @param package_ids: Optionally, a list of L{Package.id}s to match.
    @param hashes: Optionally, a list of L{Package.hash}es to match.
    @param text: Optionally, text to match in L{Package.name},
        L{Package.summary} and L{Package.description} fields.
    @param usn: Optionally, a boolean indicating whether or not to include
        L{Package}s that are associated with a L{USN}.
    @param usn_name: Optionally, the name of a L{USN} associated with
        L{Package}s to limit search results with.
    """

    def __init__(
        self,
        store,
        package_ids=None,
        hashes=None,
        text=None,
        usn=None,
        usn_name=None,
        package_names=None,
    ):
        super(PackageCollection, self).__init__(
            store,
            package_ids=package_ids,
            hashes=hashes,
            text=text,
            usn=usn,
            usn_name=usn_name,
            package_names=package_names,
        )

    def with_package_ids(self, package_ids):
        """Only include L{Package}s with a C{Package.id} in C{package_ids}.

        @param package_ids: A sequence of L{Package.id}s to filter by.
        @return: A new L{PackageCollection} filtered by L{Package.id}.
        """
        return self._filter_by(package_ids=package_ids)

    def with_package_names(self, package_names):
        """Only include L{Package}s with a C{Package.name} in C{package_names}.

        @param package_names: A sequence of L{Package.name}s to filter by.
        @return: A new L{PackageCollection} filtered by L{Package.name}.
        """
        return self._filter_by(package_names=package_names)

    def with_hashes(self, hashes):
        """Only include L{Package}s with a L{Package.hash} in C{hashes}.

        @param hashes: A sequence of L{Package.hash}es to filter by.
        @return: A new L{PackageCollection} filtered by L{Package.hash}.
        """
        return self._filter_by(hashes=hashes)

    def with_text(self, text=None):
        """
        Only include L{Package}s with name, summary or description matching
        C{text}.

        @param text: A simple word or a full expression.  Multiple search
            terms are combined with C{&}, C{|}, and C{!} and can be grouped
            with parentheses to enforce precedence.  In the absence of
            parentheses, C{!} has the strongest precedence, and C{&} has
            stronger precedence than C{|}.
        @return: A new L{PackageCollection} filtered by keyword.
        """
        return self._filter_by(text=text)

    def with_usn(self, condition=True):
        """Only include L{Package}s that have or don't have a L{USN}.

        @param condition: A boolean indicating whether or not L{Package}s must
            have a related L{USN}.
        @return: A new L{PackageCollection} filtered by L{Package}s that are
            associated with a L{USN}.
        """
        return self._filter_by(usn=condition)

    def with_usn_name(self, name):
        """Only include L{Package}s that have a L{USN} matching C{name}.

        @param name: The L{USN.name} to match.
        @return: A new L{PackageCollection} filtered by L{Package}s that are
            associated with the L{USN} matching C{name}.
        """
        return self._filter_by(usn_name=name)

    def _get_result(
        self,
        store,
        where=None,
        package_ids=None,
        hashes=None,
        text=None,
        usn=None,
        usn_name=None,
        package_names=None,
    ):
        """
        Return a L{ResultSet} configured to yield L{Package}s matching filter
        criteria.
        """
        if where is None:
            where = []

        if package_ids is not None:
            if not package_ids:
                return EmptyResultSet()
            where.append(Package.id.is_in(package_ids))

        if hashes is not None:
            where.append(Package.hash.is_in(hashes))

        if usn is not None:
            if usn:
                where.append(Package.usn_name != None)
            else:
                where.append(Package.usn_name == None)

        if usn_name is not None:
            where.append(Package.usn_name == usn_name)

        if text is not None:
            clause = self._get_text_where_clause(text)
            rank = get_package_search_rank(text)
            where.append(clause)

        if package_names is not None and not text:
            if not package_names:
                return EmptyResultSet()
            clauses = self._get_version_where_clause(package_names)
            where.extend([Or(clauses)])

        result = store.find(self._select or Package, *where)
        if text is not None:
            result.order_by(Desc(rank), Package.name, Desc(Package.version))
        else:
            result.order_by(Package.name, Desc(Package.version))
        return result

    def _get_text_where_clause(self, text):
        """
        Generate a where clause to find L{Package}s matching C{text}.

        @param text: A word or full expression.
        @return: A where clause.
        """
        text_search_query = get_search_query(text, AND)
        clause = Or(
            And(*(Like(Package.name, f"%{word}%") for word in text.split())),
            Match(text_search_query, Column("text_search_vector")),
        )
        return clause

    def _get_version_where_clause(self, packages):
        """
        Generate a where clause for L{Package} name with each specific version.

        @param packages: A list with the names of the packages.
        @return: A list of where clauses with package name and version.
        """
        operators_mapping = {
            ">>": operator.gt,
            "<<": operator.lt,
            ">=": operator.ge,
            "<=": operator.le,
            "=": operator.eq,
        }

        def get_package_parts(package):
            match = re.search(COMPARISON_OPERATORS_PATTERN, package)
            if match:
                operator = match.group()
                package_name, version = package.split(operator)
                return (package_name, operator, version)
            else:
                return None

        where = []
        for package in packages:
            package_parts = get_package_parts(package)
            if package_parts:
                package_name, operator_string, version = [i for i in package_parts]
                try:
                    mapped_operator = operators_mapping[operator_string]
                except:
                    raise ValueError("Operator not supported: %s" % operator_string)
                where += [
                    And(
                        operator.eq(Package.name, package_name),
                        mapped_operator(Package.version, version),
                    )
                ]
            else:
                where += [And(operator.eq(Package.name, package))]
        return where


def get_package_search_rank(text):
    """
    Return the L{Storm} expression to be used for calculating the rank of
    full text searches.
    """
    text_search_query = get_search_query(text, AND)
    rank_mode = RANK_LOG_LENGTH
    return (
        Similarity(Column("name"), text)
        + TSRankCD(Column("text_search_vector"), text_search_query, rank_mode) * 0.1
    )


def parse_package_names(package_names):
    """
    Parse each package name and return a list with names only.

    @return: A list with package names.
    """
    if package_names:
        names = []
        for package in package_names:
            match = re.search(COMPARISON_OPERATORS_PATTERN, package)
            if match:
                names.append(package.split(match.group())[0])
            else:
                names.append(package)
        return package_names if not names else names
    else:
        return []


class USNCollection(CollectionBase):
    """A collection of L{USN}s.

    L{USN}s in the collection are ordered by L{USN.name}, by default.

    @param store: A package C{Store} to search.
    @param package_ids: Optionally, a list of L{Package.id}s to match.
    """

    def __init__(self, store, package_ids=None):
        super(USNCollection, self).__init__(store, package_ids=package_ids)

    def with_package_ids(self, package_ids):
        """
        Only include L{USN}s associated with a C{Package} that has an ID in
        C{package_ids}.

        @param package_ids: A sequence of L{Package.id}s to filter by.
        @return: A new L{USNCollection} filtered by L{Package}s matching a
            L{Package.id}.
        """
        return self._filter_by(package_ids=package_ids)

    def _get_result(self, store, where=None, package_ids=None):
        """
        Return a L{ResultSet} configured to yield L{Package}s matching filter
        criteria.
        """
        if where is None:
            where = []

        if package_ids:
            where.extend([Package.id.is_in(package_ids), USN.name == Package.usn_name])

        result = store.find(USN, *where)
        if package_ids is not None:
            result.config(distinct=True)
        result.order_by(USN.name)
        return result
