# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import re

from zope.schema import Int, TextLine, Text, Object, Datetime, Bool
from zope.interface import Interface

from canonical.landscape.model.selection.interfaces import IComputerSelection


__all__ = [
    "IPackage",
    "IPackageSearchClient",
    "IManagedPackage",
    "IPackageHold",
    "IPackageProfile",
    "VERSION_RE",
    "NAME_RE",
    "IUpgradeProfile",
]

__clearlist__ = __all__


# This is a relaxed version of the Debian policy, but good enough for
# preventing people to shoot themselves in the feet
VERSION_RE = re.compile("[0-9][a-zA-Z0-9.+:~-]*$")
NAME_RE = re.compile("[a-z0-9][a-z0-9+.:-]+$")


class IPackage(Interface):

    id = Int(title="ID", required=True, readonly=True)

    hash = TextLine(
        title="Hash",
        description="Unique hash identifying the package",
        required=True,
        readonly=True,
    )

    type = Int(
        title="Type",
        description="Numeric package type specification",
        required=True,
        readonly=True,
    )

    name = TextLine(
        title="Name",
        description="The name of the package",
        required=True,
        readonly=True,
    )

    version = TextLine(
        title="Version", description="The package version", required=True, readonly=True
    )

    section = TextLine(
        title="Section",
        description="Category this package is part of",
        required=True,
        readonly=True,
    )

    summary = TextLine(
        title="Summary",
        description="One-line description",
        required=True,
        readonly=True,
    )

    description = Text(
        title="Description",
        description="Complete description",
        required=True,
        readonly=True,
    )

    size = Int(
        title="Size",
        description="Size of package files (uninstalled)",
        required=True,
        readonly=True,
    )

    installed_size = Int(
        title="Installed size",
        description="Size occupied by the package when installed",
        required=True,
        readonly=True,
    )

    dpkg_size = Int(
        title="The package size as reported by dpkg (in kB)",
        required=False,
        readonly=True,
    )

    dpkg_installed_size = Int(
        title="The package's installed size, as reported by dpkg (in kB)",
        required=False,
        readonly=True,
    )

    usn = TextLine(
        title="USN",
        description="The USN for this package",
        required=False,
        readonly=True,
    )


class IManagedPackage(Interface):

    computers = Object(
        title="Computer Selection",
        description="Computer selection being managed.",
        required=True,
        readonly=True,
        schema=IComputerSelection,
    )

    package = Object(
        title="Package",
        description="Package being managed.",
        required=True,
        readonly=True,
        schema=IPackage,
    )


class IUSN(Interface):

    name = TextLine(title="Name", description="USN name", required=True, readonly=True)

    summary = TextLine(
        title="Summary", description="USN summary", required=True, readonly=True
    )


class IPackageHold(Interface):

    id = Int(title="ID", required=True, readonly=True)

    name = TextLine(
        title="Name",
        description="Name of the package the hold applies to",
        required=True,
        readonly=True,
        constraint=NAME_RE.match,
    )

    relation = TextLine(
        title="Relation",
        description="Relation of the version condition this hold requires",
        required=True,
        readonly=True,
    )

    version = TextLine(
        title="Version",
        description="Version this hold requires",
        required=True,
        readonly=True,
    )

    def parse():
        """
        Return a tuple with the package, operator and version of this hold.
        """


class IPackageProfile(Interface):

    name = TextLine(
        title="Name",
        description="The name of the package profile",
        constraint=NAME_RE.match,
        required=True,
    )

    title = TextLine(
        title="Title",
        description=(
            "The profile title. The profile title can only contain "
            "alphanumeric characters."
        ),
        required=True,
    )

    description = TextLine(
        title="Description",
        description="Full description of the profile",
        required=True,
    )

    version = Int(
        title="Version",
        description="Autoincremented version of the profile",
        required=True,
        readonly=True,
    )

    creation_time = Datetime(
        title="Creation time",
        description="Time when the profile was created",
        required=True,
        readonly=True,
    )

    modification_time = Datetime(
        title="Modification time", required=True, readonly=True
    )

    package = Object(title="Package", schema=IPackage, required=False, readonly=True)


class IUpgradeProfile(Interface):

    title = TextLine(
        title="Title", description=("Title of the upgrade profile."), required=True
    )

    security_upgrade = Bool(
        title="Only security upgrades",
        description=(
            "Whether this upgrade profile should target only "
            "security upgrades. If unchecked, all upgrades will be "
            "automatically applied."
        ),
        required=True,
        default=False,
    )

    autoremove = Bool(
        title="Autoremove packages",
        description=(
            "Automatically remove packages that were installed to satify "
            "dependencies but are no longer required after the upgrade."
        ),
        required=True,
        default=False,
    )


class IPackageSearchClient(Interface):
    """
    Client for querying the package search service.
    """
