# Copyright 2005-2022 Canonical Limited. All rights reserved.

from collections import defaultdict, namedtuple

from storm.expr import Func

from zope.interface import implementer
from zope.component import adapts, getUtility

from canonical.amqp.message.interfaces import IDispatcher
from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE

from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.distribution import Distribution
from canonical.landscape.model.main.metarelease import MetaRelease
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.main.computer import Computer, ComputerStatus
from canonical.landscape.model.selection.interfaces import IComputerSelection
from canonical.landscape.model.activity.group import (
    CHANGE_PACKAGES_SUMMARY,
    VERIFY_UPGRADES_SUMMARY,
    RELEASE_UPGRADE_SUMMARY,
    APPLY_PACKAGE_PROFILE_SUMMARY,
    validate_delivery_time,
)
from canonical.landscape.model.package.interfaces import IManagedPackage, IPackage
from canonical.landscape.model.package.search import PackageSearch
from canonical.landscape.model.package.activities import (
    ChangePackagesPopulator,
    UpgradeAllPackagesRequest,
    ChangePackageProfilesRequest,
    ChangePackagesRequest,
    ChangeUpgradeProfilesRequest,
)


__all__ = ["PackageManagement", "ManagedPackage"]


__clearlist__ = __all__


__requires_prefix__ = "landscape."


ComputerChangeTuple = namedtuple(
    "ComputerChangeTuple",
    ["computer", "installs", "removals", "holds", "hold_removals"],
)


@requires_advisor
@implementer(IAccessDelegateAware)
class PackageManagement(object):
    """
    API for package operations on selected computers.
    """

    adapts(IComputerSelection)

    requires("ViewComputer", PRIVATE, ["selection"])
    requires(PUBLIC, PRIVATE, ["default_max_results", "external_threshold"])

    # This value should be balanced based on the expense of generating data on
    # the server along with the cost to update client-side structures in
    # Javascript using the generated data.
    default_max_results = 25

    def __init__(self, selection):
        self.selection = selection

    @property
    def access_delegate(self):
        return self.selection

    @requires("EditComputerPackages")
    def change_packages(
        self,
        person,
        computer_changes,
        batch_size=None,
        deliver_after=None,
        deliver_before=None,
        package_profile=None,
        upgrade_profile=None,
        deliver_after_window=0,
    ):
        """Request package changes for selected computers.

        @param person: The L{Person} requesting the change.
        @param computer_changes: A list of C{(Computer.id, installs, removals,
            holds, hold_removals)} 5-tuples where C{installs}, C{removals},
            C{holds} and C{hold_removals} are L{Package.id}s to install,
            remove, hold or release from a hold, on L{Computer}.
        @param computer_changes: A list of C{(Computer.id, installs, removals)}
            3-tuples where C{installs} and C{removals} are L{Package.id}s to
            install or remove on L{Computer}.
        @param package_profile: if specified, indicates that the changes are
            requested by a L{PackageProfile}, so a
            L{ChangePackageProfilesRequest} must be created.
        @param upgrade_profile: if specified, indicates that the changes are
            requested by a L{UpgradeProfile}, so a
            L{ChangeUpgradeProfilesRequest} must be created.
        @param deliver_after: Optionally, a C{datetime} representing the
            earliest time the L{Activity}s created to carry out this request
            should be delivered to selected L{Computer}s.  Defaults to
            immediate delivery.
        @param batch_size: If specified, the maximum number of child activities
            created at once.

        @param deliver_before: If specified, the date until which to deliver
            the created L{Activity} to selected L{Computer}s.
        @param deliver_after_window: If non-zero, the maximum number of minutes
            across which the created L{Activity} delivery times will be
            randomly staggered.

        @return: The parent L{Activity} representing the change request.
        """
        run_window_length = None
        if upgrade_profile:
            run_window_length = upgrade_profile.run_window_length
        validate_delivery_time(
            run_window_length=run_window_length,
            deliver_after_window=deliver_after_window,
        )
        summary = (
            APPLY_PACKAGE_PROFILE_SUMMARY
            if package_profile is not None
            else CHANGE_PACKAGES_SUMMARY
        )
        parent = self.selection.create_activity_group(summary=summary, creator=person)
        self.run_change_packages_batch(
            parent,
            computer_changes,
            batch_size=batch_size,
            deliver_after=deliver_after,
            deliver_after_window=deliver_after_window,
            deliver_before=deliver_before,
            package_profile=package_profile,
            upgrade_profile=upgrade_profile,
        )
        parent.info.create_log_entry("Group activity created", person=person)
        return parent

    def run_change_packages_batch(
        self,
        parent,
        computer_changes,
        batch_size=None,
        deliver_after=None,
        deliver_after_window=0,
        deliver_before=None,
        package_profile=None,
        upgrade_profile=None,
        request_class=None,
        request_kwargs=None,
    ):
        """
        Create the child activities for every computer in the package changes.
        If C{batch_size} is specified and that there are more computers than
        the size, this will create a job to separately handle the additional
        computers.
        """
        # Aggregate all requests for changes into one big task to
        # reduce the amount of queries done for detecting package
        # upgrades to the minimum possible.
        populator = ChangePackagesPopulator()

        batch_changes = computer_changes[:batch_size]
        changes_by_computers = dict(
            (computer_id, (installs, removals, holds, hold_removals))
            for computer_id, installs, removals, holds, hold_removals in batch_changes
        )

        for computer in self.selection.get_selected():
            changes = changes_by_computers.get(computer.id, (None, None, None, None))
            installs, removals, holds, hold_removals = changes

            if installs or removals or holds or hold_removals:
                kwargs = {}
                if request_class is not None and request_kwargs is not None:
                    kwargs.update(request_kwargs)
                elif package_profile is not None:
                    request_class = ChangePackageProfilesRequest
                    kwargs["package_profile"] = package_profile
                elif upgrade_profile is not None:
                    request_class = ChangeUpgradeProfilesRequest
                    kwargs["upgrade_profile"] = upgrade_profile
                else:
                    request_class = ChangePackagesRequest
                activity = parent.add(request_class, computer=computer, **kwargs)
                activity.info.add_changes(
                    installs, removals, holds, hold_removals, populator=populator
                )

        populator.run()

        # Parent methods must be called after the populator has run, since it
        # updates child activities.
        if deliver_after_window or deliver_after or deliver_before:
            parent.schedule(
                after=deliver_after,
                after_window=deliver_after_window,
                before=deliver_before,
            )
        parent.settle()

        # Add the log entries after the populator has run, so that the
        # activities have the right summary.
        for activity in parent.children():
            event_log = EventLog(parent.account, parent.creator)
            event_log.create_log_entry(
                computer, "Activity created: " + activity.info.get_summary()
            )

        if batch_size is not None:
            account = self.selection.account
            remaining_changes = computer_changes[batch_size:]
            if remaining_changes:
                dispatcher = getUtility(IDispatcher)
                if deliver_after is not None:
                    deliver_datetime = deliver_after.strftime("%Y-%m-%dT%H:%M:%SZ")
                else:
                    deliver_datetime = None
                dispatcher.create_job(
                    None,
                    "landscape.account.%d.activity.%d" % (account.id, parent.id),
                    "populate-change-packages-activity",
                    activity_id=parent.id,
                    account_id=account.id,
                    person_id=parent.creator.id,
                    computer_changes=remaining_changes,
                    deliver_after=deliver_datetime,
                    batch_size=batch_size,
                )

    @requires("EditComputerPackages")
    def request_upgrades(self, person, deliver_after_window=0):
        """Request package upgrades for all selected computers.

        @param person: The L{Person} requesting the change.
        @param deliver_after_window: If non-zero, the maximum number of minutes
            across which the created L{Activity} delivery times will be
            randomly staggered.
        @return: The parent L{Activity} representing the upgrade calculation
            request.
        """
        parent = self.selection.create_activity_group(
            summary=VERIFY_UPGRADES_SUMMARY, creator=person
        )
        for computer in self.selection.get_selected():
            child = parent.add(UpgradeAllPackagesRequest, computer=computer)
            event_log = EventLog(parent.account, parent.creator)
            event_log.create_log_entry(
                computer, "Activity created: " + child.info.get_summary()
            )

        if deliver_after_window:
            parent.schedule(after_window=deliver_after_window)
        parent.settle()
        parent.info.create_log_entry("Group activity created", person=person)
        return parent

    @requires("EditComputerPackages")
    def request_release_upgrade(self, person, computer_ids):
        """Request a release upgrade for the given L{Computer.id}s.

        @param person: The L{Person} requesting the release upgrade.
        @param computer_ids: A list of L{Computer}s ids to be
            upgraded. Computers not part of the selection are be ignored.
        @return: The parent L{Activity} representing the change request.
        """
        if not computer_ids:
            raise RuntimeError("No computer ids provided")

        parent = self.selection.create_activity_group(
            summary=RELEASE_UPGRADE_SUMMARY, creator=person
        )

        for computer in self.selection.get_selected():
            if computer.id in computer_ids:
                computer.request_release_upgrade(parent)
        parent.settle()
        parent.info.create_log_entry("Group activity created", person=person)
        return parent

    @requires("ViewComputerPackages")
    def get_computers_with_release_upgrades_count(self):
        """
        Return the count of L{Computer}s with available release upgrades.
        """
        computers_with_upgrades_count = 0
        computer_ids = self.selection.get_selected_computer_ids()
        if len(computer_ids) > 0:
            store = get_main_store()
            # find distributions with a release upgrade
            distribution_upgrades = store.find(
                (Distribution.id, Func("array_agg", MetaRelease.upgrade_id)),
                Distribution.meta_release_id == MetaRelease.id,
                MetaRelease.upgrade != None,
            )
            distribution_upgrades.group_by(Distribution.id)
            upgrade_is_lts = {}
            for distribution_id, upgrade_ids in distribution_upgrades:
                lts_available = store.find(
                    Func("bool_or", MetaRelease.lts), MetaRelease.id.is_in(upgrade_ids)
                ).one()
                upgrade_is_lts[distribution_id] = lts_available
            computers = store.find(
                (
                    Computer.id,
                    ComputerStatus.update_manager_prompt,
                    Computer.distribution_id,
                ),
                Computer.id == ComputerStatus.computer_id,
                Computer.id.is_in(computer_ids),
            )
            for (computer_id, update_manager_prompt, distribution_id) in computers:
                if update_manager_prompt == "never":
                    continue
                if distribution_id in upgrade_is_lts:
                    if (
                        update_manager_prompt == "normal"
                        or upgrade_is_lts[distribution_id] == True
                    ):
                        computers_with_upgrades_count += 1
        return computers_with_upgrades_count

    @requires("ViewComputerPackages")
    def get_release_upgrade_computers_index(self):
        """Return a C{dict} mapping upgradable L{MetaRelease}s to L{Computer}s.

        Each entry in the returned C{dict} maps a 2-tuple of a L{MetaRelease}
        and its upgrade to a C{list} of the associated computer ids in the
        selection.
        """
        store = get_main_store()

        result = store.find(MetaRelease)
        metareleases_by_id = dict(
            (metarelease.id, metarelease) for metarelease in result
        )
        # map distribution ids to metareleases
        distribution_to_metarelease = dict(
            store.find(
                (Distribution.id, Distribution.meta_release_id),
                Distribution.meta_release_id.is_in(metareleases_by_id),
            )
        )

        result = store.find(
            (Computer.distribution_id, Computer),
            Computer.id.is_in(self.selection.get_selected_computer_ids()),
            Computer.distribution_id.is_in(distribution_to_metarelease),
        )
        result.order_by(Computer.distribution_id)

        index = defaultdict(list)
        distribution_id = None
        for distribution_id, computer in result:
            metarelease = metareleases_by_id[
                distribution_to_metarelease[distribution_id]
            ]
            if metarelease.upgrade_id is None:
                continue
            if computer.update_manager_prompt == "never":
                continue
            if computer.update_manager_prompt == "lts":
                upgrade_meta_release = metareleases_by_id[metarelease.upgrade_id]
                if upgrade_meta_release.lts == False:
                    continue
            index[(metarelease, metarelease.upgrade)].append(computer.id)
        return index

    @requires("ViewComputerPackages")
    def get_search(self):
        """Return a L{PackageSearch} to search packages in L{Computer}s."""
        return PackageSearch(self.selection)


@requires_advisor
@implementer(IManagedPackage)
class ManagedPackage(object):

    adapts(IComputerSelection, IPackage)
    requires(PUBLIC, PRIVATE, ["selection"])

    def __init__(self, selection, package):
        self.selection = selection
        self.package = package
