# Copyright 2005-2022 Canonical Limited. All rights reserved.

import base64
import six
from storm.exceptions import IntegrityError
from storm.variables import UnicodeVariable
from storm.store import EmptyResultSet

from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.collection import (
    PackageCollection,
    USNCollection,
)
from canonical.landscape.model.package.binary import (
    set_package_binary,
    get_package_binaries,
)


__all__ = ["PackageMarket"]
__clearlist__ = ["PackageMarket"]


class PackageRelationInfo(object):
    """The L{PackageRelationInfo} categorizes install/removal information.

    Given a set of packages for installation and removal, the full set
    of relationships between them will be computed.

    Mappings from:

      {upgrades_package.id: [upgraded_package_id, ...]}
      {upgraded_package.id: [upgrades_package_id, ...]}
      {downgrades_package.id: [downgraded_package_id, ...]}
      {downgraded_package.id: [downgrades_package_id, ...]}

    ...will be generated.

    In addition to that, the set of installs and removals will be
    cleared from any package that falls into the mappings mentioned
    above.
    """

    def __init__(self, installs, removals, market):
        self.installs = set(installs)
        self.removals = set(removals)
        self.all_packages = self.installs | self.removals
        self._compute(market)

    def _compute(self, market):
        # Since upgrade relationships will not overlap, we pass the
        # whole set on both sides and then we post process the upgrade
        # relationships below.
        upgrade_relations = market.get_upgrade_relations(
            self.all_packages, self.all_packages
        )

        self.upgrades_map = {}
        self.upgraded_map = {}
        self.downgrades_map = {}
        self.downgraded_map = {}

        for upgrades_id, upgraded_id in upgrade_relations:
            if upgrades_id in self.installs:
                self.upgrades_map.setdefault(upgrades_id, []).append(upgraded_id)
                self.upgraded_map.setdefault(upgraded_id, []).append(upgrades_id)
            if upgrades_id in self.removals:
                self.downgrades_map.setdefault(upgrades_id, []).append(upgraded_id)
                self.downgraded_map.setdefault(upgraded_id, []).append(upgrades_id)

        self.installs -= set(self.upgrades_map) | set(self.downgraded_map)
        self.removals -= set(self.upgraded_map) | set(self.downgrades_map)


class PackageMarket(object):
    """The place where all packages are.

    Yes. All packages. All of them in the universe.
    """

    def search(self, query, package_ids=None, hashes=None, with_usn=None):
        """Find packages matching a particular query.

        @type query: unicode
        @param query: The search term which will be used in a substring match
            against package names.
        @param package_ids: A sequence of L{Package.id}s to limit searching
            to.
        @param hashes: A sequence of L{Package.hash}s to limit searching to.
        @param with_usn: A boolean indicating whether to restrict the results
            to those that are security updates.
        @return: A C{ResultSet} containing matching L{Package}s.
        """
        if not package_ids and not hashes:
            return EmptyResultSet()
        collection = PackageCollection(
            get_package_store(),
            package_ids=package_ids,
            hashes=hashes,
            text=query,
            usn=with_usn,
        )
        return collection.find()

    def get_usn_packages(self, package_ids=None):
        """Return a L{ResultSet} that yields L{Package}s with L{USN}s.

        @param package_ids: Optionally, a list of L{Package.id}s to limit
            search results with.
        """
        if package_ids is not None and not package_ids:
            return EmptyResultSet()

        collection = PackageCollection(
            get_package_store(), package_ids=package_ids, usn=True
        )
        return collection.find()

    def get(self, key):
        """Get a L{Package} given its ID or hash."""
        store = get_package_store()
        if isinstance(key, str):
            key = key.encode()

        if isinstance(key, bytes):
            return store.find(Package, hash=key).one()
        else:
            return store.get(Package, key)

    def get_many(self, package_ids, with_usn=None):
        """Get the L{Package}s with the given hashes or IDs.

        @param package_ids: A sequence of L{Package.id}s.
        @param with_usn: A boolean indicating whether to restrict the given
            L{Package}s to those which are security updates.  The default
            includes all matching packages, regardless of L{USN}.
        @return: A C{ResultSet} of matching L{Package}s.
        """
        if not package_ids:
            return EmptyResultSet()

        usn = None
        usn_name = None
        if isinstance(with_usn, six.string_types):
            usn_name = with_usn
        elif with_usn is not None:
            usn = with_usn

        collection = PackageCollection(
            get_package_store(), package_ids=package_ids, usn=usn, usn_name=usn_name
        )
        return collection.find()

    def get_ids(self, hashes):
        """Get the package IDs for the given hashes.

        @param hashes: A sequence of package hashes to fetch the ids for.
        @return: A list of package IDs for packages in the same order
            that the hashes are in. If a hash is unknown, then None
            will be in its place.
        """
        hashes = [h.encode() if isinstance(h, str) else h for h in hashes]
        store = get_package_store()
        result = store.find(Package, Package.hash.is_in(hashes))
        ids_found = dict(result.values(Package.hash, Package.id))

        return list(map(ids_found.get, hashes))

    def get_binaries(self, ids):
        """Get the package binary data for the given L{Package.id}s.

        @return: A generator yielding 3-tuples of the form hash, id,
            binary, wheren binary is base64-encode.
        """
        ids_binaries = dict(get_package_binaries(ids))
        result = self.get_many(list(ids_binaries.keys()))
        ids_hashes = dict(result.values(Package.id, Package.hash))
        for id in ids_binaries:
            yield ids_hashes[id], id, base64.b64encode(ids_binaries[id])

    def _add(self, skeleton):
        """Add a package to the store.

        @param skeleton: A L{PackageSkeleton} holding the package meta-data.
        @return: The package hash.
        """
        package_hash = skeleton.get_hash()
        store = get_package_store()

        params = (
            package_hash,
            skeleton.type,
            UnicodeVariable(skeleton.name),
            UnicodeVariable(skeleton.version),
            UnicodeVariable(skeleton.section),
            UnicodeVariable(skeleton.summary),
            UnicodeVariable(skeleton.description),
            skeleton.size,
            skeleton.installed_size,
        )
        result = store.execute(
            "INSERT INTO package (hash,type,name,version,"
            "section,summary,description,size,installed_size) "
            "VALUES (?,?,?,?,?,?,?,?,?) RETURNING id",
            params,
        )
        package_id = result.get_one()[0]

        return package_id

    def add(self, skeleton, binary=None):
        """Add a package to the store.

        @param skeleton: A L{PackageSkeleton} holding the package meta-data.
        @param binary: Optionally the binary data of the actual package.
        @return: The package hash.
        """
        package_hash = skeleton.get_hash()
        store = get_package_store()

        try:
            self._add(skeleton)
        except IntegrityError:
            # Package is already there, we just ignore the error and move on.
            store.rollback()

        if binary is not None:
            set_package_binary(self.get(package_hash), binary)

        store.commit()
        return package_hash

    def __contains__(self, package_hash):
        store = get_package_store()
        result = store.execute("SELECT id FROM package WHERE hash = ?", (package_hash,))
        return bool(result.get_one())

    def __iter__(self):
        return iter(get_package_store().find(Package).order_by(Package.id))

    def get_upgrade_relations(self, upgrades, upgraded):
        """Return upgrade relations between packages in upgrades and upgraded.

        This method will verify which packages in C{upgrades} actually
        are considered to be an upgrade of the packages in C{upgraded}.
        It will check that the package name is the same and that the
        version in C{upgrades} is greater than the one in C{upgraded}.

        @param upgrades: Sequence of L{Package.id}s that might be upgrading
            packages in C{upgraded}.
        @param upgraded: Sequence of L{Package.id}s that might be upgraded by
            packages in C{upgrades}.
        @return: Iterator generating C{(pkg_id_upgrades, pkg_id_upgraded)}
            2-tuples with the highest version pkg_id_upgrades listed first.
        """
        if not upgrades or not upgraded:
            return  # We can't have pairs that way!
        store = get_package_store()

        result = store.execute(
            """
            SELECT upgrades.id, upgraded.id
            FROM package AS upgrades
            JOIN package AS upgraded
                ON upgraded.name = upgrades.name
            WHERE upgrades.version > upgraded.version
                AND upgrades.id = ANY(?) AND upgraded.id = ANY(?)
            ORDER BY upgrades.version DESC
            """,
            (list(upgrades), list(upgraded)),
        )
        for pair in result:
            yield pair

    def get_usns_for_packages(self, package_ids):
        """Fetch all L{USN} objects that apply to any of the given packages."""
        if not package_ids:
            return EmptyResultSet()

        collection = USNCollection(get_package_store(), package_ids=package_ids)
        return collection.find()

    def get_package_relation_info(self, installs, removals):
        """Return package relations between packages in installs and removals.

        @param installs: Sequence of package ids that are being installed
        @param removals: Sequence of package ids that are being removed
        @param result: A rich L{PackageRelationInfo} object with
                       specific categorization for each of the
                       packages being requested.
        """
        return PackageRelationInfo(installs, removals, self)
