# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer

from storm.locals import Int, RawStr, Unicode, Reference, DateTime
from storm.variables import UnicodeVariable

from canonical.landscape.model.package.usn import USNName, USN
from canonical.landscape.model.package.interfaces import IPackage, NAME_RE, VERSION_RE


__all__ = [
    "PACKAGE",
    "PROVIDES",
    "REQUIRES",
    "UPGRADES",
    "CONFLICTS",
    "DEB_PACKAGE",
    "DEB_PROVIDES",
    "DEB_NAME_PROVIDES",
    "DEB_REQUIRES",
    "DEB_OR_REQUIRES",
    "DEB_UPGRADES",
    "DEB_CONFLICTS",
    "Package",
    "relation_type_to_string",
    "parse_relation_value",
    "build_relation_type",
    "build_relation_value",
    "parse_constraint",
    "InvalidPackageConstraintTypeError",
    "InvalidPackageConstraintError",
    "is_valid_package_name",
    "is_valid_package_version",
]


# First two bytes of relation.type is the kind selector:
PACKAGE = 1 << 0
PROVIDES = 1 << 1
REQUIRES = 1 << 2
UPGRADES = 1 << 3
CONFLICTS = 1 << 4


DEB_PACKAGE = 1 << 16 | PACKAGE
DEB_PROVIDES = 2 << 16 | PROVIDES
DEB_NAME_PROVIDES = 3 << 16 | PROVIDES
DEB_REQUIRES = 4 << 16 | REQUIRES
DEB_OR_REQUIRES = 5 << 16 | REQUIRES
DEB_UPGRADES = 6 << 16 | UPGRADES
DEB_CONFLICTS = 7 << 16 | CONFLICTS

TYPE_TO_STRING = {DEB_REQUIRES: "depends", DEB_CONFLICTS: "conflicts"}

STRING_TO_TYPE = {"depends": DEB_REQUIRES, "conflicts": DEB_CONFLICTS}


class InvalidPackageConstraintTypeError(Exception):
    """
    Raised when a relation type is something other than C{depends} or
    C{conflicts}.
    """


class InvalidPackageConstraintError(Exception):
    """
    Raised when a relation type is something other than C{depends} or
    C{conflicts}.
    """


class DebVersionVariable(UnicodeVariable):
    @staticmethod
    def parse_set(value, from_db):
        if not isinstance(value, str):
            raise TypeError(
                "Expected str or unidcode, found %r: %r" % (type(value), value)
            )
        return str(value)

    @staticmethod
    def parse_get(value, to_db):
        return str(value)


class DebVersion(Unicode):
    """Field for DEBVERSION postgres type."""

    variable_class = DebVersionVariable


class LazyAttribute(object):
    def __init__(self, name, build_method):
        self._name = name
        self._build_method = build_method

    def __get__(self, obj, type):
        self._build_method(obj)
        return getattr(obj, self._name)


class PackageBase(object):

    id = Int(primary=True, allow_none=False)
    hash = RawStr(allow_none=False)
    type = Int(allow_none=False)
    name = Unicode(allow_none=False)
    version = DebVersion(allow_none=False)
    section = Unicode()
    summary = Unicode()
    description = Unicode()
    size = Int()
    installed_size = Int()
    usn_name = USNName(name="usn_id")
    usn = Reference(usn_name, USN.name)

    @property
    def dpkg_size(self):
        """
        Return the package's size as it would be reported by dpkg (in kB,
        instead of bytes).
        """
        # Some packages don't have a size field (most packages in the test
        # fixtures are in this case).
        if self.size is not None:
            return self.size // 1024
        else:
            return 0

    @property
    def dpkg_installed_size(self):
        """
        Return the package's installed_size as it would be reported by dpkg (in
        kB instead of bytes).
        """
        # Some packages don't have an installed_size (most packages in the
        # test fixtures are in this case).
        if self.installed_size is not None:
            return self.installed_size // 1024
        else:
            return 0


@implementer(IPackage)
class Package(PackageBase):
    """A package that is part of the live package map.

    The package database stores a mapping of all relations between stored
    packages.  All packages on all L{Computer}s in Landscape are represented
    in this live map.  This makes it possible to perform the same kinds of
    package calculations that can be performed on a local machine.
    """

    __storm_table__ = "package"

    creation_time = DateTime()

    def __init__(self, hash, type, name, version):
        self.hash = hash
        self.type = type
        self.name = name
        self.version = version

    def __repr__(self):
        return "<Package name=%r version=%r at %s>" % (
            self.name,
            self.version,
            hex(id(self)),
        )

    def get_state(self):
        """Return a serialized state for this package."""
        data = {"name": self.name, "version": self.version, "summary": self.summary}
        if self.usn:
            data["usn"] = {"name": self.usn.name, "summary": self.usn.summary}
        return data


def is_valid_package_name(name):
    """Determine whether or not the given package name is valid."""
    return bool(NAME_RE.match(name))


def is_valid_package_version(name):
    """Determine whether or not the given package version is valid."""
    return bool(VERSION_RE.match(name))


def relation_type_to_string(type):
    """Parse the given relation type code and return a human-readable string."""
    return TYPE_TO_STRING[type]


def parse_relation_value(value):
    """Break down the given relation C{value} into individual clause items.

    @param value: The relation value, e.g. 'pkg1 >= 1.0 | pkg2'.
    @return: A list of items of 3-tuple of the form (package, operator,
          version), e.g. C{[('pkg1', '>=', '1.0'), ('pkg2', '', '')]}.
    """
    items = []
    for clause in value.split("|"):
        tokens = clause.strip().split(" ")
        package = tokens[0]
        if len(tokens) > 1:
            if len(tokens) != 3:
                raise InvalidPackageConstraintError(
                    "Invalid constraint format %r" % clause
                )
            operator = tokens[1]
            version = tokens[2]
        else:
            operator = ""
            version = ""
        items.append((package, operator, version))
    return items


def parse_constraint(constraint):
    """
    Parse a string like "depends python-twisted > 1.0" for constraint data.
    If constraint comes in as a 4-tuple, collapse them non-empty the elements
    of that tuple into a single string before parsing.

    @return a list of tuples of relation type and a relation values.
        For example, given "depends python-twisted > 1.0", the value
            ("depends", "python-twisted", ">", "1.0") will be returned.
    """
    if isinstance(constraint, tuple):
        constraint = " ".join([element for element in constraint if element])

    parts = constraint.split(" ", 1)
    if len(parts) != 2:
        raise InvalidPackageConstraintError(
            "Couldn't parse constraint %r" % (constraint,)
        )
    type, constraint = parts

    # Validate type and constraint format or error
    build_relation_type(type)
    constraints = parse_relation_value(constraint)
    if len(constraints) != 1:
        raise InvalidPackageConstraintError(
            "Couldn't parse constraint %r" % (constraint,)
        )
    return type, constraints[0]


def build_relation_type(kind):
    """
    Return the type code for the given human-readable relation C{kind}.
    """
    type = STRING_TO_TYPE.get(kind)
    if type is None:
        raise InvalidPackageConstraintTypeError("Invalid constraint type %r" % (kind,))
    return type


def build_relation_value(items):
    """Assemble a relation value according to the given clause C{items}.

    @param items: A list of 3-tuple of the form (package, operator, version),
        e.g. C{[('pkg1', '>=', '1.0'), ('pkg2', '', '')]}.
    @return: The relation value for the given items, e.g. 'pkg1 >= 1.0 | pkg2'.
    """
    clauses = []
    for tokens in items:
        stanza = " ".join(tokens).rstrip()
        clauses.append(stanza)
    value = " | ".join(clauses)
    return value
