# Copyright 2005-2022 Canonical Limited. All rights reserved.

import apt_pkg

from collections import defaultdict
from functools import cmp_to_key

from storm.expr import Func, Or, Column, SQLRaw, Max, Desc, Count, LeftJoin

from zope.interface import implementer
from zope.component import getUtility

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.main.computer import Computer, get_computers_by_ids
from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.usn import USN, CVE, parse_usn
from canonical.landscape.model.usn.issue import USNIssue
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)
from canonical.landscape.model.resource.package import ComputerPackages, ArrayOverlap
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.package.collection import (
    PackageCollection,
    get_package_search_rank,
)


__all__ = [
    "AVAILABLE",
    "INSTALLED",
    "UPGRADE",
    "DOWNGRADE",
    "PackageSearch",
    "meets_external_threshold",
    "set_external_threshold",
]

__clearlist__ = __all__


__requires_prefix__ = "landscape."


# Constants used to represent package relationships.  These are
# duplicated from the private constants defined in the
# Y.landscape.packages.listing module and in test_listing.js
AVAILABLE = 1
INSTALLED = 2
UPGRADE = 3
DOWNGRADE = 4

# apt_pkg wants this
apt_pkg.init()


def cmp(a, b):
    """cmp polyfill for python 3."""
    return (a > b) - (a < b)


@requires_advisor
@implementer(IAccessDelegateAware)
class PackageSearch(object):
    """A L{Package} search constrained to a set of selected L{Computer}s.

    @param selection: A L{ComputerSelection} instance to constrain the search
        for L{Package}s to.
    """

    requires("ViewComputerPackages", PRIVATE, ["selection"])

    _STATUSES = ("installed", "available", "available_upgrades", "held")

    def __init__(self, selection):
        self.selection = selection
        self._resource_stores = None
        self._search_client = None
        # Check whether we have enough computers to use the external
        # package-search service.
        if meets_external_threshold(selection.account):
            self._search_client = getUtility(IPackageSearchClient)

    def _get_resource_stores(self):
        if self._resource_stores is None:
            self._resource_stores = get_resource_stores_for_computer_ids(
                self.selection.get_selected_computer_ids()
            )
        return self._resource_stores

    @property
    def access_delegate(self):
        """The L{ComputerSelection} to delegate permission checks to."""
        return self.selection

    def _compare_packages(self, package1, package2):
        """Compare packages by name and version, for sorting packages with
        ascending name and descending versions."""
        return cmp(package1.name, package2.name) or apt_pkg.version_compare(
            package2.version, package1.version
        )

    def _sort_relations(self, relations, packages):
        """
        Return a list of relations sorted by package version and relation kind.

        Relations are sorted by package, newer first, using the appropriate
        field for the relation kind.
        """
        package_map = {
            AVAILABLE: "available",
            INSTALLED: "installed",
            UPGRADE: "upgrade",
            DOWNGRADE: "downgrade",
        }
        relation_package = lambda relation: (relation[package_map[relation["kind"]]])

        def compare(relation1, relation2):
            package1 = packages[relation_package(relation1)]
            package2 = packages[relation_package(relation2)]
            return self._compare_packages(package1, package2) or cmp(
                relation2["kind"], relation1["kind"]
            )

        return sorted(relations, key=cmp_to_key(compare))

    def _process_relations(self, relations, packages):
        """
        Process the provided relations list (in place) as follows:

        - sort packages by name.
        - sort relations for each package by version, newer first.
        - add the C{latest} flag, C{true} for the ones involving the newest
          version for each computer.  This can result in splitting a relation
          in two, with different C{latest} values and computer ids.

        Return a 2-tuple with a list of ids of single computers in a relation
        and the processed relations list.
        """
        single_computer_ids = set()

        for relation in relations:
            # sort package ids
            sorted_package_ids = [
                package.id
                for package in sorted(
                    (packages[package_id] for package_id in relation["package_ids"]),
                    key=cmp_to_key(self._compare_packages),
                )
            ]
            # sort relations
            package_relations = self._sort_relations(relation["relations"], packages)

            upgrades_count = 0
            security_upgrades_count = 0

            with_latest = set()
            new_relations = []

            for package_relation in package_relations:
                package_relation["latest"] = True
                new_relations.append(package_relation)

                computer_ids = set(package_relation["computers"])
                already_with_latest = computer_ids.intersection(with_latest)
                new_with_latest = computer_ids - already_with_latest
                with_latest.update(new_with_latest)

                # get single computers in relations
                if len(already_with_latest) == 1:
                    single_computer_ids.update(already_with_latest)
                if len(new_with_latest) == 1:
                    single_computer_ids.update(new_with_latest)

                # update upgrades and security upgrades count
                if package_relation["kind"] == UPGRADE:
                    new_count = len(new_with_latest)
                    upgrades_count += new_count
                    if packages[package_relation["upgrade"]].usn_name:
                        security_upgrades_count += new_count
                        # If it's not the latest, we want to make sure that at
                        # least one package reports the security upgrade
                        if not security_upgrades_count:
                            security_upgrades_count = 1

                if not new_with_latest:
                    # all computers have a newer version
                    package_relation["latest"] = False
                elif already_with_latest:
                    # some computers have a newer version, create a new
                    # relation for these ones, with latest flag to false
                    new_relation = package_relation.copy()
                    package_relation["computers"] = list(new_with_latest)
                    new_relation["latest"] = False
                    new_relation["computers"] = list(already_with_latest)
                    new_relations.append(new_relation)

            relation.update(
                {
                    "relations": new_relations,
                    "package_ids": sorted_package_ids,
                    "upgrades": upgrades_count,
                    "securityUpgrades": security_upgrades_count,
                }
            )

        return single_computer_ids, relations

    @requires("ViewComputerPackages")
    def has(self, package_id):
        """
        Return C{True} if any computer has the specified package id available
        or installed.
        """
        package_ids = SQLRaw("'{%d}'" % package_id)
        for store, ids in list(self._get_resource_stores().items()):
            result = store.find(
                ComputerPackages.computer_id,
                Or(
                    ArrayOverlap(package_ids, Column("installed")),
                    ArrayOverlap(package_ids, Column("available")),
                ),
                ComputerPackages.computer_id.is_in(ids),
            )
            if not result.is_empty():
                return True
        return False

    @requires("ViewComputerPackages")
    def get_computer_counts(self):
        """
        Get aggregated computer counts for the various package statuses.

        @return: A binary tuple holding the following information:
            - C{dict} mapping package status to number of computers with
              packages in that status in the selection,
            - C{list} of C{dict} mapping L{USN} objects, L{CVE} identifiers,
              and counts of computers affected by those USNs, ordered by
              descending USN number.
        """
        if self._search_client:
            counts = self._get_computer_counts_external()
        else:
            counts = self._get_computer_counts()

        status_counts, usn_counts = counts
        # figure out how many computers are fully up-to-date
        status_counts["up-to-date"] = (
            self.selection.get_selected_count()
            - status_counts["no-packages"]
            - status_counts["upgrade"]
        )

        usn_query = get_package_store().using(USN, LeftJoin(CVE, USN.id == CVE.usn_id))
        usn_query = usn_query.find(
            (USN, Func("ARRAY_REMOVE", Func("ARRAY_AGG", CVE.id), None)),
            USN.id.is_in(usn_counts),
        )
        usn_query = usn_query.group_by(USN.id).order_by(Desc(USN.name))

        usn_counts = [
            {"usn": usn, "cves": cves, "computers_count": usn_counts[usn.id]}
            for usn, cves in usn_query
        ]

        return status_counts, usn_counts

    def _get_computer_counts(self):
        """
        Get aggregated computer counts.

        See C{get_computer_counts()} for details.
        """
        status_counts = dict.fromkeys(
            ("no-packages", "upgrade", "security-upgrade", "held"), 0
        )
        usn_counts = defaultdict(lambda: 0)

        columns = [
            Func("COALESCE", Func("ARRAY_LENGTH", Column(status), 1), 0)
            for status in self._STATUSES
        ]

        for store, ids in list(self._get_resource_stores().items()):
            result = store.find(
                ComputerPackages, ComputerPackages.computer_id.is_in(ids)
            )
            result = store.execute(result.get_select_expr(*columns))
            for counts in result:
                if not any(counts):
                    status_counts["no-packages"] += 1
                    continue
                _, _, upgrade, held = counts
                if upgrade > 0:
                    status_counts["upgrade"] += 1
                if held > 0:
                    status_counts["held"] += 1

            # count computers with security upgrades
            result = store.find(
                Count(USNIssue.computer_id, distinct=True),
                USNIssue.resolve_time == None,
                USNIssue.computer_id.is_in(ids),
            )
            status_counts["security-upgrade"] += result.one()

            # computer counts by USN
            result = store.find(
                (USNIssue.usn_id, Count(USNIssue.computer_id)),
                USNIssue.resolve_time == None,
                USNIssue.computer_id.is_in(ids),
            )
            result.group_by(USNIssue.usn_name)
            for usn_id, count in result:
                usn_counts[usn_id] += count

        return status_counts, usn_counts

    def _get_computer_counts_external(self):
        """
        Get aggregated computer counts querying the external service.

        See C{get_computer_counts()} for details.
        """
        account_id = self.selection.account.id
        computer_ids = list(self.selection.get_selected_computer_ids())
        counts = self._search_client.counts(
            account_id=account_id, computer_ids=computer_ids
        )

        status_counts = {
            "no-packages": len(computer_ids) - counts["any"],
            "upgrade": counts["upgrade"],
            "security-upgrade": counts["security"],
            "held": counts["held"],
        }
        usn_counts = dict(
            (usn_data["Usn"], usn_data["Count"]) for usn_data in counts["usn"]
        )
        return status_counts, usn_counts

    @requires("ViewComputerPackages")
    def get_computer_ids_with_usn_upgrades(self):
        """
        Get aggregated computer ids in the selection which have available USN
        upgrades.

        @return: A C{list} of computer ids with outstanding USN updates
            required.
        """
        computer_ids = []
        for store, ids in list(self._get_resource_stores().items()):
            result = store.find(
                USNIssue.computer_id,
                USNIssue.resolve_time == None,
                USNIssue.computer_id.is_in(ids),
            )
            result.config(distinct=True)
            computer_ids.extend(result)
        return sorted(computer_ids)

    @requires("ViewComputerPackages")
    def get_package_counts_by_computer(self, limit=25, offset=0):
        """
        Get aggregated package counts for the various computers in the
        selection.

        @param limit: Limit the results to this size.
        @param offset: Start the results from this offset.
        @return: a C{list} of C{dict}s holding a L{Computer} and its package
            counts by status, ordered by L{Computer.title}.
        """
        counts = []
        computers = self.selection.get_selected(
            order_by=Computer.title, limit=limit, offset=offset
        )
        computers = dict((computer.id, computer) for computer in computers)
        computer_ids = list(computers.keys())

        columns = [Column("computer_id")]
        columns.extend(
            Func("COALESCE", Func("ARRAY_LENGTH", Column(status), 1), 0)
            for status in self._STATUSES
        )

        resource_stores = get_resource_stores_for_computer_ids(computer_ids)
        for store, ids in list(resource_stores.items()):
            result = store.find(
                ComputerPackages, ComputerPackages.computer_id.is_in(ids)
            )
            result = store.execute(result.get_select_expr(*columns))
            for computer_id, installed, available, upgrade, held in result:
                counts.append(
                    {
                        "computer": computers[computer_id],
                        "available": available,
                        "installed": installed,
                        "held": held,
                        "upgrade": upgrade,
                    }
                )
        return sorted(counts, key=lambda item: item["computer"].title)

    def get_upgrades_by_computer_id(self):
        """
        Return a C{dict} mapping computer ids to a 2-element tuple with
        available upgrades and held upgrades.
        """
        # package_id: [computers_ids that have package_id as available upgrade]
        computers_by_upgrade = {}
        # computer_id: [held_ids]
        held_by_computer = {}
        # all held package ids
        all_held = set()
        # computer_id: ([upgrades], [held_upgrades])
        upgrades_by_computer = defaultdict(lambda: ([], []))

        for store, ids in list(self._get_resource_stores().items()):
            columns = (
                Column("computer_id"),
                Column("available_upgrades"),
                Column("held"),
            )
            result = store.find(
                ComputerPackages, ComputerPackages.computer_id.is_in(ids)
            )
            for row in result.values(*columns):
                computer_id, upgrades, held = row
                # Set the upgrades right away, for the held upgrades
                # we just set a placeholder, we'll populate it below
                upgrades_by_computer[computer_id] = (upgrades, [])
                for upgrade in upgrades:
                    computers = computers_by_upgrade.setdefault(upgrade, [])
                    computers.append(computer_id)
                held_by_computer[computer_id] = set(held)
                all_held.update(held)

        all_upgrades = list(computers_by_upgrade.keys())
        market = PackageMarket()
        relations = market.get_upgrade_relations(all_upgrades, all_held)

        for upgrades_id, upgraded_id in relations:
            for computer_id in computers_by_upgrade[upgrades_id]:
                if upgraded_id in held_by_computer[computer_id]:
                    upgrades_by_computer[computer_id][1].append(upgrades_id)

        # sort held upgrades and filter them out from regular upgrades
        for computer_id in self.selection.get_selected_computer_ids():
            upgrades, held_upgrades = upgrades_by_computer[computer_id]
            upgrades = sorted(set(upgrades).difference(held_upgrades))
            held_upgrades = sorted(held_upgrades)
            upgrades_by_computer[computer_id] = (upgrades, held_upgrades)
        return upgrades_by_computer

    @requires("ViewComputerPackages")
    def get_package_upgrades_by_computer_id(
        self,
        text=None,
        names=None,
        security_only=False,
        autoremove=False,
        usn_name=None,
    ):
        """
        Return a list of tuples (computer_id, installs, removals) representing
        the changes needed to upgrade packages on selected computers.

        @return: A list of tuples (computer_id, installs, removals) to be
        actioned.
        """
        # computer_id -> [IDs of available package upgrades]
        upgrades_by_computer = {}
        # computer_id -> [IDs of non-held installed packages]
        installed_by_computer = {}
        # computer_id -> [IDs of non-held autoremovable packages]
        autoremovable_by_computer = {}
        # IDs of all available package upgrades across the selection
        all_upgrades = set()
        all_security = set()
        # IDs of all non-held installed packages across the selection
        all_installed = set()

        columns = (
            Column("computer_id"),
            Column("installed"),
            Column("held"),
            Column("available_upgrades"),
            Column("autoremovable"),
            Column("security"),
        )

        for store, ids in list(self._get_resource_stores().items()):
            result = store.find(
                ComputerPackages, ComputerPackages.computer_id.is_in(ids)
            )
            for row in result.values(*columns):
                computer_id = row[0]
                # Only consider installed packages that are not held
                installed = set(row[1]).difference(row[2])
                upgrades = set(row[3])
                autoremovable = set(row[4]).difference(row[2])
                security_upgrades = upgrades.intersection(row[5])
                upgrades_by_computer[computer_id] = upgrades
                installed_by_computer[computer_id] = installed
                autoremovable_by_computer[computer_id] = autoremovable
                all_upgrades.update(upgrades)
                all_installed.update(installed)
                all_security.update(security_upgrades)

        # Filter the set of available upgrades with the given criteria
        collection = PackageCollection(
            get_package_store(),
            package_ids=all_upgrades,
            text=text,
            package_names=names,
            usn=security_only or None,
            usn_name=usn_name,
        )
        collection = collection.select((Package.id,))
        all_upgrades = [row[0] for row in collection.find()]

        # Include packages reported as security upgrades
        if security_only:
            all_upgrades = set(all_upgrades).union(all_security)

        market = PackageMarket()
        relations = market.get_upgrade_relations(all_upgrades, all_installed)

        # computer_id -> ([IDs of package upgrades],
        #                 [IDs of upgradable packages])
        computer_changes = defaultdict(lambda: (set(), set()))
        for upgrades_id, upgraded_id in relations:
            for computer_id in upgrades_by_computer:
                if (
                    upgrades_id in upgrades_by_computer[computer_id]
                    and upgraded_id in installed_by_computer[computer_id]
                ):
                    # computers that have the upgraded package installed and
                    # the upgrade available can be upgraded
                    installs, removals = computer_changes[computer_id]
                    # Only add a package upgrade if we are not already
                    # updating this package.  This can happen when a single
                    # package has multiple upgrade candidates.
                    if upgraded_id not in removals:
                        installs.add(upgrades_id)
                        removals.add(upgraded_id)

        if autoremove:
            for computer_id in autoremovable_by_computer:
                _, removals = computer_changes[computer_id]
                removals.update(autoremovable_by_computer[computer_id])

        # return a tuple for each selected computer, also those without
        # install or removals.
        result = []
        for computer_id in self.selection.get_selected_computer_ids():
            installs, removals = computer_changes[computer_id]
            result.append((computer_id, sorted(installs), sorted(removals), [], []))

        return result

    def get_package_states(
        self,
        text=None,
        names=None,
        available=None,
        upgrades=None,
        installed=None,
        held=None,
        limit=25,
        offset=0,
    ):
        """
        Return a list of packages matching criteria, along with lists of
        selected computers IDs grouped by states.

        The list is sorted by package name, with versions newest to oldest.

        For example:

          [(package,
            [(available, installed, held, upgrade, [computer_ids]),
             ...]),
           ...]
        """
        all_states = self._get_computer_package_states()

        # Filter the all_states dictionary in place, using the given status
        # flags
        flags = (available, installed, held, upgrades)
        for package_id in list(all_states.keys()):
            states = []
            for state, computer_ids in list(all_states[package_id].items()):
                for i, flag in enumerate(flags):
                    if flag is not None:
                        # A filtering flag is set, let's make sure that we only
                        # return matching states
                        if flag and not state[i]:
                            break
                        if not flag and state[i]:
                            break
                else:
                    # All filtering flags match, so let's add this state
                    states.append(list(state) + [computer_ids])
            if states:
                all_states[package_id] = states
            else:
                all_states.pop(package_id)

        # Further filter by text and names
        collection = PackageCollection(
            get_package_store(),
            package_ids=list(all_states.keys()),
            text=text,
            package_names=names,
        )
        result = collection.find()
        result.config(limit=limit, offset=offset)
        return [(package, all_states[package.id]) for package in result]

    @requires("ViewComputerPackages")
    def is_using_external(self):
        """Return whether we're using the external package search service."""
        return self._search_client is not None

    @requires("ViewComputerPackages")
    def get_relations_by_package_name(
        self,
        text=None,
        available=None,
        installed=None,
        upgrades=None,
        held=None,
        usn=None,
        usn_name=None,
        limit=25,
        offset=0,
        mark=0,
    ):
        """
        Get relation information about the matching packages across the
        selection.

        Return a structure of the form:

          {"totalPackages": 120,
           "computers": [<Computer object ...>, ...],
           "packages": {package_id:  <Package object pkg-1 0.1-1>, ...},
           "relations": [
               {"name": "pkg-1",
                "relations": [{"kind": UPGRADE,
                               "held": False,
                               "installed": package_id,
                               "upgrade": package_id,
                               "computers": [1, 2, 3, 4]}]},
               ...
               ],
           ...}
        """
        if self._search_client:
            aggregate_search_info = self._get_aggregate_search_info_external(
                text=text,
                available=available,
                installed=installed,
                upgrades=upgrades,
                held=held,
                usn=usn,
                usn_name=usn_name,
                limit=limit,
                mark=mark,
            )
            statuses, names, packages, mark = aggregate_search_info
            count = None
        else:
            # Get aggregated computer status data matching the given criteria
            statuses, count, names, packages = self._get_aggregate_search_info(
                text=text,
                available=available,
                installed=installed,
                upgrades=upgrades,
                held=held,
                usn=usn,
                usn_name=usn_name,
                limit=limit,
                offset=offset,
            )
            mark = None
        # Now figure out what upgrades what among the package IDs that we
        # got above, and arrange this information in relation dicts.
        kinds = [AVAILABLE, INSTALLED, UPGRADE, DOWNGRADE]
        if upgrades:
            kinds = [UPGRADE]
        if installed or held:
            kinds = [INSTALLED, UPGRADE, DOWNGRADE]
        relations = self._get_relations_for_packages(packages, statuses, kinds)
        # Sort relations according to the order of the names list, which is
        # ordered by rank and name.
        relations.sort(key=lambda relation: names.index(relation["name"]))
        single_computer_ids, relations = self._process_relations(relations, packages)
        computers = sorted(
            get_computers_by_ids(single_computer_ids), key=lambda computer: computer.id
        )
        result = {"packages": packages, "relations": relations, "computers": computers}
        if count is not None:
            result["totalPackages"] = count
        else:
            result["mark"] = mark
        return result

    def _get_relations_for_packages(self, packages, statuses, kinds):
        """
        Return relations for the specified packages, grouped by package name.

        @param packages: dict mapping ids to L{Package}s
        @param statuses: a dict with package statuses by package id, with the
            following content:

            {package_id: [
              (available, installed, held, upgrade, [computer_id, ...]),
              ...],
             ...}
        @param kinds: Only return relations whose kind is in this list.
        """
        package_ids = list(statuses.keys())
        market = PackageMarket()
        relations = market.get_upgrade_relations(package_ids, package_ids)

        relations_by_name = defaultdict(
            lambda: {
                "relations": [],
                "package_ids": set(),
                "upgrades": 0,
                "securityUpgrades": 0,
            }
        )
        # map package ids to computer ids that already have a relation
        # involving this package
        computer_ids_in_relation = defaultdict(set)

        for (
            name,
            upgraded_id,
            upgrades_id,
            relation,
        ) in self._get_upgrade_downgrade_relations(
            packages, statuses, relations, kinds
        ):
            relation_data = relations_by_name[name]
            relation_data["relations"].append(relation)
            relation_data["package_ids"].update([upgraded_id, upgrades_id])

            computer_ids = relation["computers"]
            computer_ids_in_relation[upgraded_id].update(computer_ids)
            computer_ids_in_relation[upgrades_id].update(computer_ids)

        for name, package_id, relation in self._get_installed_available_relations(
            packages, statuses, computer_ids_in_relation, kinds
        ):
            relation_data = relations_by_name[name]
            relation_data["relations"].append(relation)
            relation_data["package_ids"].add(package_id)
            computer_ids = relation["computers"]

        for name, relation_data in list(relations_by_name.items()):
            relation_data["name"] = name
        return list(relations_by_name.values())

    def _get_upgrade_downgrade_relations(self, packages, statuses, relations, kinds):
        """
        Generator yielding tuples like
            (package_name, upgraded_id, upgrades_id, relation)
        for UPGRADE and DOWNGRADE kinds.

        @param relations: (upgraded_id, upgrades_id) pairs of package ids.
        """
        # For each upgrades_id/upgraded_id pair we have the following 4
        # possible situations that can occur each on a subset of the
        # selected computers:
        #
        #  0. upgrades_id is available and upgraded_id installed (upgrade)
        #  1. upgrades_id is available and upgraded_id held (held upgrade)
        #  2. upgrades_id is installed and upgraded_id available (downgrade)
        #  3. upgrades_id is held and upgraded_id available (held downgrade)
        #
        #  We yield a relation for each of these situations, if it occurs.

        # List of possible relation types (kind, held)
        relation_types = [
            (UPGRADE, False),
            (UPGRADE, True),
            (DOWNGRADE, False),
            (DOWNGRADE, True),
        ]

        for upgrades_id, upgraded_id in relations:

            package_name = packages[upgrades_id].name

            relation_sets = self._get_upgrade_downgrade_relation_sets(
                upgrades_id, upgraded_id, statuses
            )

            # For each relation type, the associated relation set contains
            # the computer IDs for which upgrades_id/upgraded_id is either
            # an upgrade or downgrade relation (held or not). If the set
            # is non empty yield a new relation of the appropriate type
            for (kind, held), relation_set in zip(relation_types, relation_sets):
                if relation_set and kind in kinds:
                    relation = {
                        "kind": kind,
                        "held": held,
                        "computers": sorted(relation_set),
                    }
                    if kind == UPGRADE:
                        relation.update(
                            {"installed": upgraded_id, "upgrade": upgrades_id}
                        )
                    else:
                        relation.update(
                            {"installed": upgrades_id, "downgrade": upgraded_id}
                        )
                    yield package_name, upgraded_id, upgrades_id, relation

    def _get_upgrade_downgrade_relation_sets(self, upgrades_id, upgraded_id, statuses):
        """
        Given a package upgrade relation (upgrades_id, upgraded_id), return
        a tuple of four computer IDs sets for the various relation types:

          0. upgrades_id is available and upgraded_id installed (upgrade)
          1. upgrades_id is available and upgraded_id held (held upgrade)
          2. upgrades_id is installed and upgraded_id available (downgrade)
          3. upgrades_id is held and upgraded_id available (held downgrade)
        """
        upgrades_computer_ids = set()  # upgrades_id is an upgrade
        upgraded_computer_ids = set()  # upgraded_id is installed
        upgraded_held_computer_ids = set()  # upgraded_id is held
        downgrades_computer_ids = set()  # upgraded_id is available
        downgraded_computer_ids = set()  # upgrades_id is installed
        downgraded_held_computer_ids = set()  # upgrades_id is held

        for (available, installed, held, upgrade, computer_ids) in statuses[
            upgrades_id
        ]:
            if upgrade:
                upgrades_computer_ids.update(computer_ids)
            elif installed:
                if held:
                    downgraded_held_computer_ids.update(computer_ids)
                else:
                    downgraded_computer_ids.update(computer_ids)

        for (available, installed, held, upgrade, computer_ids) in statuses[
            upgraded_id
        ]:
            if installed:
                if held:
                    upgraded_held_computer_ids.update(computer_ids)
                else:
                    upgraded_computer_ids.update(computer_ids)
            elif available:
                downgrades_computer_ids.update(computer_ids)

        return (
            upgrades_computer_ids & upgraded_computer_ids,
            upgrades_computer_ids & upgraded_held_computer_ids,
            downgrades_computer_ids & downgraded_computer_ids,
            downgrades_computer_ids & downgraded_held_computer_ids,
        )

    def _get_installed_available_relations(
        self, packages, statuses, computer_ids_in_relation, kinds
    ):
        """
        Generator yielding tuples with (package_name, package_id, relation) for
        AVAILABLE and INSTALLED kinds.

        @param computer_ids_in_relation: dict mapping package ids to a set of
           computer ids that already have a relation for that package.
        """
        for package_id, package_statuses in list(statuses.items()):
            for (available, installed, held, upgrade, computer_ids) in package_statuses:

                computer_ids = sorted(
                    set(computer_ids) - computer_ids_in_relation[package_id]
                )
                if computer_ids:
                    package = packages[package_id]
                    if installed and INSTALLED in kinds:
                        yield (
                            package.name,
                            package_id,
                            {
                                "kind": INSTALLED,
                                "held": held,
                                "installed": package_id,
                                "computers": computer_ids,
                            },
                        )
                    elif available and AVAILABLE in kinds:
                        yield (
                            package.name,
                            package_id,
                            {
                                "kind": AVAILABLE,
                                "held": False,
                                "available": package_id,
                                "computers": computer_ids,
                            },
                        )

    def _get_aggregate_search_info(
        self,
        text=None,
        names=None,
        available=None,
        upgrades=None,
        installed=None,
        held=None,
        usn=None,
        usn_name=None,
        limit=25,
        offset=0,
    ):
        """
        Return a tuple of four elements providing information about the matches
        of the given search:

          (statuses, count, names, packages)

        where:

          - statuses: C{dict} mapping package IDs to statuses for packages
                matching the given criteria along with the computer IDs they
                refert to, i.e.:

                {package_id:
                    [(available, installed, held, upgrade, [computer_ids]),
                     ...],
                 ...}

          - count: total count of package matches, this doesn't take into
                account the limit/offset parameter, it's a global count

          - names: C{list} of package names matching the criteria, ordered
                by rank (if a text parameter was provided) and name

          - packages: C{dict} mapping package IDs to L{Package} objects, for
                the packages matching the criteria.
        """
        # Get what package IDs are in what state across the selected computers,
        # obtaining a dict of the form:
        #
        # {<package_id>: {<state>: [<computer_ids>]}}
        #
        # where <state> is a combination of four boolean flags (available,
        # installed, held, upgrade). Here we basically need to fetch each row
        # in the computer_packages table for the selected computers.
        all_states = self._get_computer_package_states()

        # Find out which states match the given status criterion. If a status
        # criterion is set to None, then it always matches. This is an
        # application-side filtering that we can't perform with straight SQL
        # because the data in computer_packages is denormalized.
        flags = (available, installed, held, upgrades)
        package_ids = self._get_package_ids_in_states(all_states, flags)

        # Query the packages whose IDs are in the set of matching computer
        # states and that also meet the rest of filtering criteria (i.e. text,
        # names, usn, usn_name, limit, offset).
        collection = PackageCollection(
            get_package_store(),
            package_ids=package_ids,
            text=text,
            package_names=names,
            usn=usn,
            usn_name=usn_name,
        )
        collection = collection.select(Package.name)
        result = collection.find()
        result.group_by(Package.name)
        if text is not None:
            rank = get_package_search_rank(text)
            result.order_by(Desc(Max(rank)), Package.name)
        else:
            result.order_by(Package.name)
        result.config(offset=offset, limit=limit)
        names = list(result)
        # Count the total number of unique package names.
        result = collection.select(Package.name).find()
        count = result.count(Package.name, distinct=True)

        # Now fetch the packages matching the above names, across all states,
        # not only the matching ones. This will include for example packages
        # upgrades.
        collection = PackageCollection(
            get_package_store(),
            package_ids=list(all_states.keys()),
            package_names=names,
        )

        statuses = {}
        packages = {}
        for package in collection.find():
            packages[package.id] = package
            statuses[package.id] = [
                list(state) + [all_states[package.id][state]]
                for state in all_states[package.id]
            ]

        return statuses, count, names, packages

    def _get_aggregate_search_info_external(
        self,
        text=None,
        available=None,
        installed=None,
        upgrades=None,
        held=None,
        usn=None,
        usn_name=None,
        limit=25,
        mark=0,
    ):
        """Using the external packagesearch service, return a tuple of four
        elements providing information about the matches of the given search:

          (statuses, names, packages, mark)

        where:

          - statuses: C{dict} mapping package IDs to statuses for packages
                matching the given criteria along with the computer IDs they
                refert to, i.e.:

                {package_id:
                    [(available, installed, held, upgrade, [computer_ids]),
                     ...],
                 ...}

          - names: C{list} of package names matching the criteria, ordered
                by rank (if a text parameter was provided) and name

          - packages: C{dict} mapping package IDs to L{Package} objects, for
                the packages matching the criteria.

          - mark: C{int} a cursor showing the point at which we hit the
                specified result limit within the search.  Passing this mark to
                a subsequent call will cause it to fetch results beyond this
                mark.

        """
        # Combinations of states are not allowed, even though the method
        # signature implies they are.
        query_state = "any"
        if installed:
            query_state = "installed"
        if upgrades:
            query_state = "upgrade"
        if held:
            query_state = "held"
        if usn:
            query_state = "security"
        account_id = self.selection.account.id
        computer_ids = list(self.selection.get_selected_computer_ids())
        search_criteria = {
            "account_id": account_id,
            "computer_ids": computer_ids,
            "state": query_state,
            "limit": limit,
            "mark": mark,
        }
        if usn_name:
            usn_id = parse_usn(usn_name)
            search_criteria["usn"] = usn_id
        if text:
            search_criteria["text"] = text
        result = self._search_client.search(**search_criteria)
        statuses = {}
        for state in result["states"]:
            package_id = state["Package"]
            for state_dict in state["Groups"]:
                available, upgrade, installed, held = state_dict["Flags"]
                computers = state_dict["Computers"]
                statuses.setdefault(package_id, []).append(
                    [available, installed, held, upgrade, computers]
                )
        names = result["names"]
        market = PackageMarket()
        packages = {}
        for package in market.get_many(list(map(int, list(statuses.keys())))):
            packages[package.id] = package
        mark = result["mark"]
        return statuses, names, packages, mark

    def _get_package_ids_in_states(self, states, flags):
        """
        Return the package IDs present in the states matching the given flags.

        The C{states} parameter is in the form:

            {<package_id>: {<state>: [<computer_ids>]}}

        where <state> is a combination of four boolean flags (available,
        installed, held, upgrade), and the C{flags} parameter is a sequence
        of four boolean values that should be checked against the state. If
        a flag is C{None} it will always match the relevant value in <state>.
        """
        # Find out which states match the given filtering criteria. If a
        # filtering flag is set to None, then it always matches.
        package_ids = []
        for package_id in list(states.keys()):
            for state, computer_ids in list(states[package_id].items()):
                for i, flag in enumerate(flags):
                    if flag is not None:
                        # A filtering flag is set, let's make sure that we only
                        # return matching states
                        if flag and not state[i]:
                            break
                        if not flag and state[i]:
                            break
                else:
                    # All filtering flags match, so let's add this package
                    # to the set of candidate matches
                    package_ids.append(package_id)
        return package_ids

    def _get_computer_package_states(self):
        """Get aggregate computer package states across the selection.

        The returned data structure is of the form:

         {<package_id>: {<state>: [<computer_ids>]}}

        where <state> is a combination of four boolean flags (available,
        installed, held, upgrade).
        """
        all_states = defaultdict(lambda: defaultdict(lambda: []))
        for store, ids in list(self._get_resource_stores().items()):
            columns = (
                Column("computer_id"),
                Column("available"),
                Column("installed"),
                Column("held"),
                Column("available_upgrades"),
            )
            result = store.find(
                ComputerPackages, ComputerPackages.computer_id.is_in(ids)
            )
            for row in result.values(*columns):
                computer_id = row[0]

                # package_id -> (available, installed, held, upgrade)
                states = defaultdict(lambda: (False, False, False, False))
                for i, package_ids in enumerate(row[1:]):
                    for package_id in package_ids:
                        state = list(states[package_id])
                        state[i] = True
                        states[package_id] = tuple(state)

                for package_id, state in list(states.items()):
                    all_states[package_id][state].append(computer_id)
        return all_states


_external_threshold = 50  # High default mainly for tests


def set_external_threshold(threshold):
    """
    Only use the external package search if the account has more than this
    number of computers.
    """
    global _external_threshold
    old_threshold = _external_threshold
    _external_threshold = threshold
    return old_threshold


def meets_external_threshold(account):
    """
    Whether the given account has enough computers to make sure of the external
    package search service.
    """
    return account.get_computers_count() >= _external_threshold
