# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.event.subscriber import subscriber

from canonical.landscape.model.events import AddComputerEvent, ResynchronizeEvent
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.package.profile import (
    PackageProfile,
    get_package_profiles_with_computer_ids,
)
from canonical.landscape.model.package.autoupgrade import UpgradeProfile
from canonical.landscape.model.resource.package import clear_packages_buffer
from canonical.landscape.model.events import AddTagsEvent, RemoveTagsEvent


@subscriber(IAccount, AddComputerEvent, phase=1)
def add_computer_event_package_profile_handler(account, event):
    """
    Event handler is invoked when a L{Computer} is added to an L{Account}.
    """
    profiles = account.store.find(
        PackageProfile,
        PackageProfile.account_id == account.id,
        PackageProfile.computer_group_id == ComputerGroup.id,
        ComputerGroup.all_computers == True,
    )
    for profile in profiles:
        profile.add_computers_with_ids([event.computer.id])

    profiles = account.store.find(
        UpgradeProfile,
        UpgradeProfile.account_id == account.id,
        UpgradeProfile.computer_group_id == ComputerGroup.id,
        ComputerGroup.all_computers == True,
    )
    for profile in profiles:
        profile.add_computers_with_ids([event.computer.id])


@subscriber(ResynchronizeEvent)
def delete_packages(resynchronize):
    """
    Delete all packages in a L{Computer} in response to a
    L{ResynchronizeEvent}.
    """
    # <= 2.2 clients resynchronize but don't know how to do packages,
    # so don't actually resynchronize for those clients.
    if resynchronize.computer.current_api in [b"2.0", b"2.1", b"2.2"]:
        return
    if resynchronize.is_in_scope("package"):
        computer = resynchronize.computer
        computer.get_packages().update(reset=True)
        clear_packages_buffer(computer)


@subscriber(ResynchronizeEvent)
def reapply_profiles(resynchronize):
    """If a computer resynchronizes we need to possibly re-apply its profiles.

    For example when a pending computer is accepted as an existing computer
    a L{ResynchronizeEvent} is triggered, and we catch it in order to apply
    the required profiles on the new computer.
    """
    if not resynchronize.is_in_scope("package"):
        return
    computer = resynchronize.computer
    for profile in get_package_profiles_with_computer_ids(
        computer.account, computer_ids=[computer.id]
    ):
        profile.apply(computer_ids=[computer.id])


@subscriber(IAccount, AddTagsEvent)
def add_tag_to_computers_handler(account, event):
    """Event handler is invoked when L{Tag}s are added to L{Computer}s."""
    account.computer_group.synchronize_with_added_tags(
        event.tag_ids, event.computer_ids, PackageProfile
    )
    account.computer_group.synchronize_with_added_tags(
        event.tag_ids, event.computer_ids, UpgradeProfile
    )


@subscriber(IAccount, RemoveTagsEvent)
def remove_tag_from_computers_handler(account, event):
    """Event handler is invoked when L{Tag}s are removed from L{Computer}s."""
    account.computer_group.synchronize_with_removed_tags(
        event.tag_ids, event.computer_ids, PackageProfile
    )
    account.computer_group.synchronize_with_removed_tags(
        event.tag_ids, event.computer_ids, UpgradeProfile
    )
