# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime
import logging

from zope.component import getUtility

from storm.locals import (
    Int,
    Unicode,
    And,
    Or,
    DateTime,
    AutoReload,
    Reference,
    ReferenceSet,
)
from storm.expr import Eq, Insert
from storm.properties import SimpleProperty
from storm.variables import Variable

from canonical.lib.text import make_secure_id
from canonical.security.protect import PUBLIC, PRIVATE, requires, requires_advisor

from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.errors import (
    PackageSearchRequestError,
    PackageSearchConnectionError,
)


__all__ = ["CVE", "USN", "USNName", "update_from_usn_tool_db", "parse_usn"]

__clearlist__ = ["parse_usn"]


def parse_usn(value):
    major, minor = value.split("-")
    return int(major) * 1000 + int(minor)


def get_usn_binaries(usn_data):
    """Given a USN database entry (a set of nested dictionaries), return a
    set of (package, version) tuples describing the total set of binary
    packages affected by the USN, extracted from the release URLs.
    """
    return {
        (pkg_name, pkg["version"])
        for release in usn_data.get("releases", {}).values()
        for pkg_name, pkg in release.get(
            "allbinaries", release.get("binaries", {})
        ).items()
    }


class USNNameVariable(Variable):
    """Simple bridge to store USN identifiers (123-4) as ints (123004)."""

    def parse_get(self, value, to_db):
        if to_db:
            return value
        return "%d-%d" % (value // 1000, value % 1000)

    def parse_set(self, value, from_db):
        if from_db:
            return value
        try:
            return parse_usn(value)
        except ValueError:
            raise ValueError("Invalid USN name")


class USNName(SimpleProperty):

    variable_class = USNNameVariable


@requires_advisor
class USN(object):
    """Simple representation of a USN.

    @type name: C{str}
    @ivar name: The USN identifier (like "123-4").

    @ivar summary: The summary from the USN.
    """

    __storm_table__ = "usn"

    id = Int(allow_none=False, default=AutoReload)
    name = USNName(name="id", primary=True, allow_none=False)
    summary = Unicode()
    release_time = DateTime()

    def __init__(self, name, summary, release_time):
        self.name = name
        self.summary = summary
        self.release_time = release_time

    @requires(PUBLIC)
    def release_time_date(self):
        return self.release_time.date()


class USNChangeSet(object):
    """
    Changes resulting from a USN database import, with the delta from the
    previous state.
    """

    def __init__(
        self,
        added_usns=None,
        removed_usns=None,
        added_package_usns=None,
        removed_package_usns=None,
    ):
        """
        @param added_usns: a set of ids for USNs added to the database
        @param removed_usns: a set of ids for USNs removed from the database
        @param added_package_usns: a set of (package id, USN id) tuples for
            added package USNs
        @param removed_package_usns: a set of (package id, USN id) tuples for
            removed package USNs
        """
        self.added_usns = added_usns or set()
        self.removed_usns = removed_usns or set()
        self.added_package_usns = added_package_usns or set()
        self.removed_package_usns = removed_package_usns or set()


class CVE(object):
    """
    Simple representation of a CVE, associated with a USN.
    """

    __storm_table__ = "cve"
    __storm_primary__ = "id", "usn_id"

    id = Unicode(allow_none=False)
    usn_id = Int(allow_none=False)
    usn_name = USNName(name="usn_id", allow_none=False)
    usn = Reference(usn_id, USN.id)

    def __init__(self, cve, usn_id):
        self.id = cve
        self.usn_id = usn_id


USN.cves = ReferenceSet(USN.id, CVE.usn_id)


def update_from_usn_tool_db(db):
    """Record USN data and associate with relevant packages.

    A L{USN} object is created for each USN entry in the database. For each
    package which a USN affects, the L{Package.usn} will be set to the USN
    object. If no record exists for a specified package, then it will be
    ignored.

    @param db: A USN database (i.e. a dict) as read from
        https://usn.ubuntu.com/usn-db/database.json.bz2
    @return: A set of L{Package.id}s affected by USNs.
    """
    from canonical.landscape.model.package.package import Package

    def decode(value):
        if isinstance(value, bytes):
            return value.decode("utf-8")
        return value

    store = get_package_store()

    # First, get current packages with USNs, as USNs that are no longer present
    # in the database file are later removed, which causes usn_name to be set
    # to null for packages that had those USNs.
    result = store.find((Package.id, Package.usn_name), Package.usn_name != None)
    current_package_usn = set(
        (pkg_id, parse_usn(usn_name)) for pkg_id, usn_name in result
    )

    existing_usns = set(store.find(USN.id))
    db_usns = set(parse_usn(usn) for usn in db)
    removed_usns = existing_usns - db_usns  # Ids removed from the db

    # Remove USNs that are no longer in the database file; this will also
    # remove CVEs via cascaded foreign-key
    store.find(USN, USN.id.is_in(removed_usns)).remove()

    added_usns = set()  # Ids for new USNs
    db_package_usn = set()  # (pkg_id, usn_id) tuples currently in the db
    for usn_name in db:
        usn_data = db[usn_name]
        usn_release_time = datetime.fromtimestamp(usn_data["timestamp"])
        if parse_usn(usn_name) in existing_usns:
            usn = store.get(USN, usn_name)
            usn.summary = decode(usn_data["summary"])
            usn.release_time = usn_release_time
        else:
            usn = USN(usn_name, decode(usn_data["summary"]), usn_release_time)
            store.add(usn)
            added_usns.add(usn.id)

        where = []
        for pkg_name, pkg_version in get_usn_binaries(usn_data):
            where.append(
                And(
                    Eq(Package.name, decode(pkg_name)),
                    Eq(Package.version, decode(pkg_version)),
                )
            )
        if where:
            result = store.find(Package, Or(*where))
            ids = set(result.values(Package.id))
            db_package_usn.update((package_id, usn.id) for package_id in ids)

    added_package_usns = db_package_usn - current_package_usn
    # Keep the newest USN for each package, since we allow linking only a
    # single USN to a package.
    added_package_usns_map = dict(sorted(added_package_usns))
    removed_package_usns = current_package_usn - db_package_usn
    # This can't have multiple USNs per package, so there's no need to sort.
    removed_package_usns_map = dict(removed_package_usns)

    # Find the set of all CVE,USN relationships currently in the database, and
    # generate the diffs to apply
    existing_cves = set(store.find((CVE.id, CVE.usn_id)))
    db_cves = {
        (decode(cve), parse_usn(usn))
        for usn, usn_data in db.items()
        for cve in usn_data.get("cves", ())
        if cve.startswith("CVE-")
    }
    added_cves = db_cves - existing_cves
    removed_cves = existing_cves - db_cves
    for cve, usn in removed_cves:
        store.find(CVE, CVE.id == cve, CVE.usn_id == usn).remove()
    for cve, usn in added_cves:
        cve = CVE(cve, usn)
        store.add(cve)

    # Set USN to None for packages that are no longer listed in the USN
    # file. The set can contain packages that had their USN updated to a
    # different number, so removals must be processed before additions.
    result = store.find(
        Package, Package.id.is_in(package_id for package_id, _ in removed_package_usns)
    )
    result.set(usn_name=None)

    # Bulk-add USN info for packages.
    if added_package_usns_map:
        temp_table = "new_package_usn_%s" % decode(make_secure_id(length=6))
        store.execute(
            "CREATE TEMPORARY TABLE %s "
            "(pkg_id INTEGER PRIMARY KEY, usn_id INTEGER) ON COMMIT DROP" % temp_table
        )
        store.execute(
            Insert(
                ("pkg_id", "usn_id"),
                table=temp_table,
                values=iter(added_package_usns_map.items()),
            )
        )
        store.execute(
            "UPDATE package SET usn_id = %(temp)s.usn_id FROM %(temp)s "
            "WHERE id = %(temp)s.pkg_id" % {"temp": temp_table}
        )

    # Send updates to the package-search service too.
    if added_package_usns_map or removed_package_usns_map:
        search_client = getUtility(IPackageSearchClient)
        try:
            search_client.update_usns(
                added=added_package_usns_map, removed=removed_package_usns_map
            )
        except PackageSearchConnectionError:
            # This means that either the package-search service hasn't started
            # yet or that it's down, in either case we just ignore the error,
            # since it will have the correct state (from the database) once it
            # comes up.
            logging.warning(
                "Could not connect to package-search service to send USN "
                "package updates."
            )
        except PackageSearchRequestError as e:
            # This is an unexpected error, we log it but we don't blow up.
            logging.warning(
                "Error updating USN information on the package-search "
                "service: %s" % e
            )

    return USNChangeSet(
        added_usns=added_usns,
        removed_usns=removed_usns,
        added_package_usns=added_package_usns,
        removed_package_usns=removed_package_usns,
    )
