# Copyright 2005-2022 Canonical Limited. All rights reserved.

import csv
from io import StringIO
from collections import OrderedDict
from datetime import timedelta

from zope.interface import implementer
from zope.security.proxy import removeSecurityProxy

from storm.expr import Or, Join, Min, Func

from canonical.lib.timestamp import utcnow
from canonical.landscape.model.package.autoupgrade import UpgradeProfile
from canonical.landscape.ui.format import sanitize_csv_value
from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.main.computer import PingTime

from canonical.landscape.model.reporting.interfaces import IReportingManagement
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.computergroup import ComputerGroupItem
from canonical.landscape.model.package.usn import USN, CVE
from canonical.landscape.model.package.search import PackageSearch
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.usn.issue import USNIssueManagement

from canonical.landscape.ui.autoupgrade.schedule import ViewSchedule


__all__ = ["ReportingManagement"]


__clearlist__ = __all__


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IAccessDelegateAware, IReportingManagement)
class ReportingManagement(object):
    """
    API for reporting (aka compliance reports) operations on selected
    computers.
    """

    requires("ViewComputer", PRIVATE, ["selection"])

    def __init__(self, selection):
        self.selection = selection
        self.computer_ids = set(self.selection.get_selected_computer_ids())
        self.main_store = get_main_store()
        self.account_store = removeSecurityProxy(self.selection.account).store
        self._not_covered_ids = None
        self._schedule_to_count = None

    def _add_to_schedule_count(self, profile, schedule_to_count, computer_ids):
        view_schedule = ViewSchedule(profile)
        schedule = profile.schedule
        schedule_to_count[schedule] = {
            "computer_ids": sorted(computer_ids),
            "count": len(computer_ids),
            "display_name": view_schedule.string_display(),
        }

    def _profile_counts(self):
        covered_ids_and_profiles = self.account_store.find(
            (ComputerGroupItem.computer_id, UpgradeProfile),
            UpgradeProfile.account_id == self.selection.account.id,
            UpgradeProfile.computer_group_id == ComputerGroupItem.computer_group_id,
            ComputerGroupItem.computer_id.is_in(self.computer_ids),
        ).order_by(UpgradeProfile.schedule, ComputerGroupItem.computer_id)
        not_covered_ids = set(self.computer_ids)
        schedule_to_count = {}
        last_profile = None
        computer_ids = set()
        for covered_id, profile in covered_ids_and_profiles:
            not_covered_ids.discard(covered_id)
            if not last_profile is None and profile.schedule != last_profile.schedule:
                self._add_to_schedule_count(
                    last_profile, schedule_to_count, computer_ids
                )
                computer_ids = set()
            computer_ids.add(covered_id)
            last_profile = profile
        if last_profile:
            self._add_to_schedule_count(last_profile, schedule_to_count, computer_ids)
        return list(not_covered_ids), list(schedule_to_count.values())

    @property
    def access_delegate(self):
        return self.selection

    @requires("ViewComputerPackages")
    def get_package_search(self):
        """Return a L{PackageSearch} to search packages in L{Computer}s."""
        return PackageSearch(self.selection)

    @requires("ViewComputer")
    def get_not_pinging_computers(self, since_minutes=5):
        """
        Return a count of computers in the current selection which have not
        contacted landscape in the last C{since_minutes}.
        """
        now = utcnow()
        five_minutes = timedelta(minutes=since_minutes)
        five_minutes_ago = now - five_minutes
        return self.main_store.find(
            PingTime.computer_id,
            PingTime.computer_id.is_in(self.computer_ids),
            Or(
                PingTime.last_ping_time < five_minutes_ago,
                PingTime.last_ping_time == None,
            ),
        )

    @requires("ViewUpgradeProfile")
    def get_computers_not_upgraded(self):
        """
        Returns a list of computer ids for the current selection that are not
        covered by an upgrade profile.
        """
        if self._not_covered_ids is None:
            (self._not_covered_ids, self._schedule_to_count) = self._profile_counts()
        return self._not_covered_ids

    @requires("ViewUpgradeProfile")
    def get_upgraded_computers_by_frequency(self):
        """
        Return a dictionary mapping the frequency of an upgrade profile to the
        count of computers that it applies to.

        Since we cannot easily access maps with arbitrary strings from the
        template (zope namespaces forbid spaces for instance), this returns a
        mapping of:

            [{"display_name": "Something with spaces", "count": 5}]
        """
        if self._schedule_to_count is None:
            (self._not_covered_ids, self._schedule_to_count) = self._profile_counts()
        return self._schedule_to_count

    def _initialise_usn_results_structure(self, fixed_in_days):
        """
        Create the empty result structure to be returned by
        C{get_usn_time_to_fix}, with the correct "fixed_in" slots based on the
        C{fixed_in_days} parameter.
        """
        return {
            "pending": set(),
            "fixed_in": OrderedDict(
                list(zip(fixed_in_days, [set() for i in range(len(fixed_in_days))]))
            ),
        }

    def _calculate_usn_constraint_dates(
        self, now, fixed_in_days, in_last, pending_in_days
    ):
        """
        Calculate the three dates required to feed
        C{_get_constrained_usn_data}.
        """
        max_days = max(fixed_in_days)
        min_days = min(fixed_in_days)
        start_date = now - timedelta(days=max_days) - timedelta(days=in_last)
        end_date = now - timedelta(days=min_days)
        pending_cutoff = now - timedelta(days=pending_in_days)
        return start_date, end_date, pending_cutoff

    def _get_constrained_usn_data(self, start_date, pending_cutoff):
        """
        Given a reporting period for USNs (defined by C{start_date} and
        C{end_date}) and a cutoff date for pending USNs, retrieve all the USNs
        required to report on both times to fix USNs and machines with pending
        USNs.
        """
        min_date = min((start_date, pending_cutoff))
        usns = set(get_package_store().find(USN, USN.release_time > min_date))
        release_times = {}
        usn_names = []
        pending_exclusion_names = []
        for usn in usns:
            release_times[usn.name] = usn.release_time
            if usn.release_time > start_date:
                usn_names.append(usn.name)
            if usn.release_time > pending_cutoff:
                pending_exclusion_names.append(usn.name)
        return release_times, usn_names, pending_exclusion_names

    def _update_fixed_in_days(self, result, fixed_in_days, computer_id, elapsed_delta):
        """
        Add a computer ID to the correct "fixed_in" bucket in the results.
        """
        if elapsed_delta is not None:
            for day in fixed_in_days:
                if elapsed_delta < timedelta(days=day):
                    result["fixed_in"][day].add(computer_id)

    @requires("ViewComputer")
    def get_usn_time_to_fix(
        self,
        fixed_in_days=(2, 14, 30),
        pending_in_days=60,
        in_last=30,
        reporting_date=None,
    ):
        """
        Return an aggregated view showing the time it takes to fix USNs for the
        current selection of computers.

        For each C{fixed_in_days}, it returns the set of computer IDs which
        fixed all the USNs released around C{in_last} interval in
        C{fixed_in_days}. For the default values, it means the following
        intervales: -32 to -2, -44 to -14, and -60 to -30.

        It also return the set of computers which have USNs older than
        C{pending_in_days} still not outstanding.

        Optionally passing C{reporting_date} allows us to report on a fixed
        point in time rather than at this exact moment.
        """
        if reporting_date is None:
            now = utcnow()
        else:
            now = reporting_date
        usn_management = USNIssueManagement(self.selection, only_computers=False)
        result = self._initialise_usn_results_structure(fixed_in_days)
        dates = self._calculate_usn_constraint_dates(
            now, fixed_in_days, in_last, pending_in_days
        )
        start_date, end_date, pending_cutoff = dates

        usn_data = self._get_constrained_usn_data(start_date, pending_cutoff)
        release_times, usn_names, pending_exclusion_names = usn_data

        # USN issues are inherently grouped by computer_id (this is important).
        usn_issues = usn_management.get_issues(usn_names)

        last_computer_id = None
        max_elapsed = None
        release_time = None
        not_reported_on_computers = set(self.computer_ids)
        for usn_issue in usn_issues:
            release_time = release_times[usn_issue.usn_name]
            if last_computer_id is not None:
                if usn_issue.computer_id != last_computer_id:
                    self._update_fixed_in_days(
                        result, fixed_in_days, last_computer_id, max_elapsed
                    )
                    max_elapsed = None
            if start_date < release_time < end_date:
                not_reported_on_computers.discard(usn_issue.computer_id)
                if usn_issue.resolve_time is not None:
                    elapsed = usn_issue.resolve_time - release_time
                    if max_elapsed is None or elapsed > max_elapsed:
                        max_elapsed = elapsed
            last_computer_id = usn_issue.computer_id

        if release_time is not None:
            self._update_fixed_in_days(
                result, fixed_in_days, last_computer_id, max_elapsed
            )
        if not_reported_on_computers:
            for day in fixed_in_days:
                result["fixed_in"][day].update(not_reported_on_computers)

        pending_issues = usn_management.get_open_issues_for_computers(self.computer_ids)
        for usn_issue in pending_issues:
            if usn_issue.usn_name not in pending_exclusion_names:
                result["pending"].add(usn_issue.computer_id)
        return result

    def _get_issue_resolutions_by_usn(self, max_days=30):
        """
        Return two dicts which will be used by L{get_tabular_data} below. The
        first dict maps USN names to column headers. The second maps a tuple
        of (computer id, USN name) to the resolution datetime.
        """
        # Get the USNs from the last 30 days
        thirty_one_days_ago = utcnow() - timedelta(days=max_days)
        usns = list(
            get_package_store().find(USN, USN.release_time > thirty_one_days_ago)
        )
        usn_headers = {
            usn.name: "USN {usn.name} "
            "({usn.release_time:%Y-%m-%d %H:%M})".format(usn=usn)
            for usn in usns
        }

        # USNs and their respective patch time
        usn_issue_management = USNIssueManagement(self.selection)
        usn_issues = usn_issue_management.get_issues([usn.name for usn in usns])
        resolutions = {
            (issue.computer_id, issue.usn_name): issue.resolve_time
            for issue in usn_issues
        }

        return usn_headers, resolutions

    def _get_issue_resolutions_by_cve(self, max_days=30):
        """
        Return two dicts which will be used by L{get_tabular_data} below. The
        first dict maps CVE ids to column headers. The second maps a tuple
        of (computer id, CVE id) to the resolution datetime.
        """
        # Get the earliest release time for each CVE in range
        package_store = get_package_store()
        thirty_one_days_ago = utcnow() - timedelta(days=max_days)
        cve_query = (
            package_store.using(USN, Join(CVE, USN.id == CVE.usn_id))
            .find(
                (CVE.id, Min(USN.release_time)), USN.release_time > thirty_one_days_ago
            )
            .group_by(CVE.id)
        )
        cve_headers = {
            cve: "{cve} ({release_time:%Y-%m-%d %H:%M})".format(
                cve=cve, release_time=release_time
            )
            for cve, release_time in cve_query
        }

        # USNs and their respective patch time
        usn_query = (
            package_store.using(USN, Join(CVE, USN.id == CVE.usn_id))
            .find(
                (USN.name, Func("ARRAY_AGG", CVE.id)),
                USN.release_time > thirty_one_days_ago,
            )
            .group_by(USN.name)
        )
        usns = {usn: cves for usn, cves in usn_query}
        usn_issue_management = USNIssueManagement(self.selection)
        usn_issues = usn_issue_management.get_issues(usns)

        resolutions = {}
        for issue in usn_issues:
            for cve in usns.get(issue.usn_name, ()):
                key = (issue.computer_id, cve)
                resolution = resolutions.get(key)
                if resolution is None or issue.resolve_time is None:
                    resolutions[key] = issue.resolve_time
                else:
                    resolutions[key] = min(resolution, issue.resolve_time)
        return cve_headers, resolutions

    def get_tabular_data(self, by_cve=False, max_days=30):
        """
        Get a table of data about computers in the selection.

        The result will contain the following keys:
            - name: The title of the computer.
            - is_pinging: whether the computer has successfully pinged within
              five minutes.
            - schedule: An English description of the schedule.

        In addition, if by_cve is False (the default), a key will be added for
        each USN released in the last max_days. The key name will be the
        identifier of the USN with timestamp, and the value will indicate
        whether the issue is present and if it is, whether it has been
        resolved, and when.

        If by_cve is True, CVEs will be used as the column key instead of USNs,
        but otherwise the logic is identical.
        """
        computer_ids = self.selection.get_selected_computer_ids()
        selected_computers = self.selection.get_selected()

        account_store = removeSecurityProxy(self.selection.account).store
        main_store = get_main_store()

        # Ping times
        five_minutes_ago = utcnow() - timedelta(minutes=5)
        pinging_ids = set(
            main_store.find(
                PingTime.computer_id,
                PingTime.computer_id.is_in(computer_ids),
                PingTime.last_ping_time > five_minutes_ago,
            )
        )

        # Upgrade schedule
        ids_to_profile = account_store.find(
            (ComputerGroupItem.computer_id, UpgradeProfile),
            UpgradeProfile.account_id == self.selection.account.id,
            UpgradeProfile.computer_group_id == ComputerGroupItem.computer_group_id,
            ComputerGroupItem.computer_id.is_in(computer_ids),
        )
        # The ordering of this query is only for testing reliability purposes.
        ids_to_profile = ids_to_profile.order_by(UpgradeProfile.name)

        view_schedules = {}
        computers_to_profiles = {}
        for computer_id, profile in ids_to_profile:
            computers_to_profiles.setdefault(computer_id, [])
            computers_to_profiles[computer_id].append(profile)
            schedule_string = view_schedules.get(profile)
            if not schedule_string:
                view_schedule = ViewSchedule(profile)
                view_schedules[profile] = view_schedule.string_display()

        issue_headers, issue_resolutions = (
            self._get_issue_resolutions_by_cve
            if by_cve
            else self._get_issue_resolutions_by_usn
        )(max_days=max_days)

        result = []
        for computer in selected_computers:
            record = {}
            record["name"] = computer.title
            record["is_pinging"] = computer.id in pinging_ids

            profiles = computers_to_profiles.get(computer.id)
            if profiles is not None:
                schedules = []
                for profile in profiles:
                    schedules.append(view_schedules[profile])
                record["schedule"] = " AND ".join(schedules)
            else:
                record["schedule"] = None

            for issue in issue_headers:
                header = issue_headers[issue]
                try:
                    resolve_time = issue_resolutions[(computer.id, issue)]
                except KeyError:
                    record[header] = False
                else:
                    record[header] = resolve_time
            result.append(record)
        return result

    @requires("ViewComputer")
    def get_csv_data(self, by_cve=False, max_days=30):
        """
        Format the data returned from L{get_tabular_data} as a CSV string.
        """
        output = StringIO()
        writer = csv.writer(output)

        data = self.get_tabular_data(by_cve, max_days=max_days)
        base_keys = ["name", "is_pinging", "schedule"]
        if not data:
            writer.writerow(base_keys)
            return output.getvalue()
        security_keys = sorted(
            key for key in list(data[0].keys()) if key not in base_keys
        )
        header = base_keys + security_keys

        writer.writerow(header)

        for info in data:
            row = []
            for header in base_keys:
                value = info[header]
                if value is not None:
                    value = str(value)
                row.append(value)
            for security_key in security_keys:
                resolution = info[security_key]
                if resolution is False:
                    resolution = "Not present"
                elif resolution is None:
                    resolution = "Unresolved"
                else:
                    resolution = resolution.strftime("%Y-%m-%d %H:%M")
                row.append(resolution)
            row = [sanitize_csv_value(cell) for cell in row]
            writer.writerow(row)
        return output.getvalue()
