# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from storm.locals import Int, Reference

from canonical.lib.text import expand_placeholders
from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.activity.activity import ActivityInfo
from canonical.landscape.model.repository.pocket import Pocket


__all__ = ["ChangeRepositoryProfilesRequest", "SyncPocketRequest"]

__clearlist__ = ["ChangeRepositoryProfilesRequest"]

__requires_prefix__ = "landscape."


class ChangeRepositoryProfilesRequest(ComputerRequest):
    """
    Activity to send a list of package profiles to a L{Computer}.

    The activity creates a message for the client with details for each
    repository profile, if at least one profile is available for that
    L{Computer}.
    """

    def is_approvable(self):
        """
        ChangeRepositoryProfilesRequest never requires human approval.
        """
        return False

    def get_summary(self):
        return "Apply repository profiles"

    def _started(self):
        """
        Send a message to the client, requesting that it update its
        sources.list.
        """
        message = {
            "type": "apt-sources-replace",
            "operation-id": self.activity.id,
            "sources": [],
            "gpg-keys": [],
        }

        from canonical.landscape.model.repository.management import (
            get_profiles_for_computer,
        )

        result = get_profiles_for_computer(self.computer)

        placeholder_values = {}
        distro = self.computer.distribution
        if distro is not None:
            placeholder_values["series"] = distro.code_name

        for profile in result:
            apt_lines = profile.get_apt_source_lines()
            if apt_lines:
                content = expand_placeholders("\n".join(apt_lines), placeholder_values)
            else:
                content = "## No sources associated with this profile\n"

            message["sources"].append({"name": profile.name, "content": content})
            message["gpg-keys"].extend(profile.get_gpg_keys_material())

        self.computer.get_message_store().add(message)


class SyncPocketRequest(ActivityInfo):
    """Activity to sync a L{Pocket} in mirror mode.

    @param activity: The L{Activity} we're attached to.
    @param pocket: The L{Pocket} to be synced.
    """

    __storm_table__ = "sync_pocket_request"

    pocket_id = Int(allow_none=False)

    pocket = Reference(pocket_id, Pocket.id)

    def __init__(self, activity, pocket):
        super(SyncPocketRequest, self).__init__(activity)
        self.pocket = pocket

    def get_summary(self):
        return "Sync pocket '%s' of series '%s' in distribution '%s'" % (
            self.pocket.name,
            self.pocket.series.name,
            self.activity.entity.name,
        )

    def get_state(self):
        state = super(SyncPocketRequest, self).get_state()
        state.update(
            {
                "pocket_id": self.pocket.id,
                "pocket_name": self.pocket.name,
                # XXX the progress field can be dropped once we expose the
                # progress field in ActivityInfo.get_state().
                "progress": self.activity.progress,
            }
        )
        return state
