# Copyright 2005-2022 Canonical Limited. All rights reserved.

import errno
import os

from psycopg2.extensions import TransactionRollbackError

from zope.component import getUtility

from storm.locals import Int, AutoReload, ReferenceSet, Reference, Store

from canonical.reprepro.reprepro import Reprepro

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.configuration import get_proxy_configuration
from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.entity.helper import AccountEntity, get_account_entity
from canonical.landscape.model.repository.pocket import (
    FILTER_WHITELIST,
    FILTER_BLACKLIST,
)
from canonical.landscape.model.repository.series import Series
from canonical.landscape.model.repository.upload import Upload
from canonical.landscape.model.repository.constants import UBUNTU_KEYS

STATUS_UNLOCKED = 0
STATUS_LOCKED = 1


class LockError(Exception):
    """
    Raised when trying to lock a distribution for an operation when it's
    already locked.
    """

    def __init__(self, distribution):
        self.distribution = distribution
        super(LockError, self).__init__(
            f"Another operation is already in progress on distribution '{distribution}'"
        )


class DuplicateDistributionError(DuplicateModelError):
    """Raised when a distribution with same name already exists."""

    model_name = "Distribution"


class Distribution(AccountEntity):
    """A distribution inside the L{Account}'s repository.

    @param account: The L{Account} the distribution is linked to.
    @param name: The name of the distribution.
    """

    __storm_table__ = "distribution"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    status = Int(allow_none=False)

    series = ReferenceSet(id, Series.distribution_id, order_by=Series.name)

    def _get_access_context(self):
        return self.account.get_access_context(self.data.access_context_id)

    def _set_access_context(self, access_context):
        self.data.access_context_id = access_context.id

    access_context = property(_get_access_context, _set_access_context)

    def create_series(self, name):
        """Create a L{Series} in this distribution."""
        return Series(self, name)

    def get_series(self, ids=None, names=None):
        """Get the L{Series} for a L{Distribution}."""
        where = []
        if ids is not None:
            where.append(Series.id.is_in(ids))
        if names is not None:
            where.append(Series.name.is_in(names))
        return self.series.find(*where)

    def get_basedir(self):
        """Returns the base directory for distribution"""
        config = getUtility(IFileBasedConfiguration)
        repodir = config.get("landscape", "repository-path")
        return os.path.join(repodir, self.account.name, self.name)

    def get_reprepro(self):
        """
        Return a L{Reprepro} instance configured with the L{Distribution} data.
        """
        config = getUtility(IFileBasedConfiguration)
        repodir = config.get("landscape", "repository-path")
        accountdir = os.path.join(repodir, self.account.name)
        gpg_home = os.path.join(accountdir, ".gnupg")
        for directory in (repodir, accountdir):
            if os.path.exists(directory):
                if not os.path.isdir(directory):
                    raise RuntimeError(
                        f"Path '{directory}' exists but is" "not a directory"
                    )
            else:
                try:
                    os.mkdir(directory)
                    os.chmod(directory, 0o0755)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise

        # Set the reprepro binary, if different from the default (for tests)
        try:
            binary_path = config.get("landscape", "reprepro-binary")
        except KeyError:
            binary_path = None

        reprepro = Reprepro(self.get_basedir(), binary_path=binary_path)
        reprepro.set_gpg_home(gpg_home)

        for series in self.series:
            for pocket in series.pockets:
                pocket_full_name = pocket.get_full_name()
                udeb_components = pocket.components if pocket.include_udeb else None
                reprepro.add_distribution(
                    pocket_full_name,
                    pocket.components,
                    pocket.archs,
                    mode=pocket.get_mode_name(),
                    sign=pocket.get_sign_with(),
                    udeb_components=udeb_components,
                )

                if pocket.gpg_key is not None:
                    reprepro.add_gpg_key(pocket.gpg_key.secret_material)

                if pocket.is_in_mirror_mode():
                    flat = None
                    if pocket.mirror_suite.endswith("/"):
                        # It's a flat repository, so it doesn't have
                        # components. In this case the pocket should have a
                        # single component which will be the target one for
                        # packages from the flat repository.
                        [flat] = pocket.components
                    config = self._get_proxy_apt_config()
                    reprepro.add_update(
                        pocket_full_name,
                        pocket.mirror_uri,
                        pocket.mirror_suite,
                        pocket.get_verify_release(),
                        flat=flat,
                        config=config,
                    )
                    if pocket.mirror_gpg_key is None:
                        for key in UBUNTU_KEYS:
                            reprepro.add_gpg_key(key.material)
                    else:
                        material = pocket.mirror_gpg_key.material
                        reprepro.add_gpg_key(material)

                if pocket.is_in_pull_mode():
                    filter_action = None
                    if pocket.filter_type == FILTER_BLACKLIST:
                        filter_action = "install"
                        rule_action = "deinstall"
                    if pocket.filter_type == FILTER_WHITELIST:
                        filter_action = "deinstall"
                        rule_action = "install"
                    reprepro.add_pull(
                        pocket_full_name,
                        pocket.pull_pocket.get_full_name(),
                        filter_action,
                    )
                    if filter_action:
                        for name in pocket.filters:
                            reprepro.add_filterlist_rule(
                                pocket_full_name, name, rule_action
                            )

                if pocket.is_in_upload_mode():
                    reprepro.add_incoming(pocket_full_name)
                    if pocket.upload_allow_unsigned:
                        reprepro.add_uploader(pocket_full_name, unsigned=True)
                    else:
                        for key in pocket.upload_gpg_keys:
                            reprepro.add_uploader(pocket_full_name, key.key_id)
                            reprepro.add_gpg_key(key.material)

        return reprepro

    def get_state(self):
        """Return a dict with information on the L{Distribution}."""
        state = self.data.get_state()
        state.update({"series": [series.get_state() for series in self.get_series()]})
        return state

    def lock(self):
        """
        Lock the distribution for operations.

        @raise LockError: if the distribution is already locked.
        """
        self._set_status(STATUS_LOCKED)

    def unlock(self):
        """Unlock the distribution for operations."""
        self._set_status(STATUS_UNLOCKED)

    def _set_status(self, status):
        """Lock/unlock atomically."""
        try:
            current_status = (
                Store.of(self)
                .execute(
                    "SELECT status FROM distribution where id = ? FOR UPDATE",
                    (self.id,),
                )
                .get_one()[0]
            )
        except TransactionRollbackError:
            # another transaction is in progress
            raise LockError(self.name)

        if current_status == STATUS_LOCKED and status == STATUS_LOCKED:
            raise LockError(self.name)

        self.status = status

    def _get_proxy_apt_config(self):
        """Return the APT proxy configuration to be passed to reprepro."""
        config = {}
        proxy_config = get_proxy_configuration()

        http_proxy = proxy_config.get("http_proxy")
        if http_proxy:
            config["Acquire::http::Proxy"] = http_proxy
        https_proxy = proxy_config.get("https_proxy")
        if https_proxy:
            config["Acquire::https::Proxy"] = https_proxy
        no_proxy = proxy_config.get("no_proxy")
        if no_proxy:
            for entry in no_proxy.split(","):
                entry = entry.strip()
                config.update(
                    {
                        "Acquire::http::Proxy::{}".format(entry): "DIRECT",
                        "Acquire::https::Proxy::{}".format(entry): "DIRECT",
                    }
                )
        return config


Series.distribution = Reference(Series.distribution_id, Distribution.id)

Upload.distribution = Reference(Upload.distribution_id, Distribution.id)


def get_distribution(account, id=None, name=None):
    """Return a L{Distribution} by id or name in the L{Account}."""
    return get_account_entity(account, Distribution, id=id, name=name)
