# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
import time

from twisted.python.failure import Failure

from zope.component import getUtility
from zope.security.interfaces import Unauthorized

from twisted.internet.defer import succeed

from storm.twisted.transact import transact

from canonical.reprepro.reprepro import RepreproError
from canonical.amqp.job.runner import JobRunner, job
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.amqp.message.interfaces import IDispatcher
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.activity.activity import get_child_activity


class RepositoryJobRunnerMixin(JobRunner):
    @transact
    def _unlock_distribution(
        self, result, account_id, person_id, distribution_id=None, pocket_id=None
    ):
        """Unlock a distribution given a distribution or pocket ID."""
        account = get_account_by_id(account_id)
        person = get_person_by_id(person_id)
        management = RepositoryManagement(account, person)
        distribution = None
        if distribution_id:
            distribution = management.get_distributions(ids=[distribution_id]).one()
        else:
            pocket = management.get_pockets(ids=[pocket_id]).one()
            if pocket:
                distribution = pocket.series.distribution
        if distribution:
            distribution.unlock()
        if isinstance(result, Failure):
            raise result.value


@job
class SyncMirrorPocketJobRunner(RepositoryJobRunnerMixin):
    """Setup and run reprepro to sync a mirror pocket."""

    _time = time.time  # For testing

    def run(self, account_id, person_id, pocket_id, activity_id):
        deferred = self._start(account_id, activity_id)
        deferred.addCallback(self._get_reprepro, account_id, person_id, pocket_id)
        deferred.addCallback(self._run_reprepro, account_id, activity_id)
        deferred.addCallback(self._succeed, account_id, activity_id)
        deferred.addErrback(self._fail, account_id, activity_id)
        deferred.addBoth(
            self._unlock_distribution, account_id, person_id, pocket_id=pocket_id
        )
        deferred.addCallback(lambda _: (None, None))
        return deferred

    @transact
    def _get_reprepro(self, _, account_id, person_id, pocket_id):
        """Return the config to pass to reprepro."""
        account = get_account_by_id(account_id)
        person = get_person_by_id(person_id)
        management = RepositoryManagement(account, person)
        pocket = management.get_pockets(ids=[pocket_id]).one()
        if pocket is None:
            raise Unauthorized()
        return pocket.get_reprepro()

    @transact
    def _update_activity_progress(self, account_id, activity_id, progress):
        activity = self._get_activity(account_id, activity_id)
        activity.advance(progress)

    def _get_activity(self, account_id, activity_id):
        """Return the activity with the specified id in the account."""
        account = get_account_by_id(account_id)
        return get_child_activity(account, activity_id)

    def _run_reprepro(self, reprepro, account_id, activity_id):
        """Spawn reprepro an track its progress."""
        last_update = [self._time(), None]
        deferred = succeed(None)

        def progress_callback(progress):
            now = self._time()
            # Only update the activity if the last update was performed
            # more then 1 second ago and the progress value has changed
            if now - last_update[0] > 1.0 and progress != last_update[1]:
                last_update[0] = now
                last_update[1] = progress
                deferred.addCallback(
                    lambda _: self._update_activity_progress(
                        account_id, activity_id, progress
                    )
                )

        def wait_for_progress_updates(result):
            # Chain the progress updates deferred, so that we fire
            # only when everything is completed. Then we return the
            # process output.
            return deferred.addCallback(lambda _: result)

        run = reprepro.update(progress_callback=progress_callback)
        run.addCallback(wait_for_progress_updates)
        return run

    @transact
    def _start(self, account_id, activity_id):
        activity = self._get_activity(account_id, activity_id)
        activity.start()

    @transact
    def _succeed(self, _, account_id, activity_id):
        activity = self._get_activity(account_id, activity_id)
        activity.succeed(code=0, text="Repository sync completed")

    @transact
    def _fail(self, failure, account_id, activity_id):
        error = failure.value
        if failure.check(RepreproError):
            code = error.code
            text = str(error.out)
        elif failure.check(Unauthorized):
            # Special-case this exception since it has an empty message
            code = -1
            text = "The user is not authorized to perform this operation"
        else:
            code = -1
            text = str(error)
        activity = self._get_activity(account_id, activity_id)
        activity.fail(code=code, text=text)


@job
class RepreproJobRunner(RepositoryJobRunnerMixin):
    """Perform a reprepro operation for a given pocket or distribution."""

    _time = time.time  # For testing

    def run(
        self,
        account_id,
        person_id,
        reprepro_method,
        distribution_id=None,
        pocket_id=None,
    ):
        """Call a reprepro_method for the given pocket or distribution.

        @param account_id: ID of the Account for the distribution or pocket.
        @param person_id: ID of the person submitting the job.
        @param reprepro_method: The reprepro class method name which will be
            called.
        @param distribution_id: Optionally, the ID of the Distribution for
            which the reprepro operation is performed. If not provided, then
            pocket_id must be set.
        @param pocket_id: Optionally, the ID of the Pocket for
            which the reprepro operation is performed. If not provided, then
            distribution_id must be set.
        """
        assert any(
            [distribution_id, pocket_id]
        ), "Reprepro job requires either distribution_id or pocket_id"
        assert None in (
            distribution_id,
            pocket_id,
        ), "Reprepro job does not accept both distribution_id and pocket_id"

        deferred = self._get_reprepro(account_id, person_id, distribution_id, pocket_id)
        deferred.addCallback(self._run_reprepro, reprepro_method)
        deferred.addBoth(
            self._unlock_distribution,
            account_id,
            person_id,
            distribution_id=distribution_id,
            pocket_id=pocket_id,
        )
        deferred.addCallback(lambda _: (None, None))
        return deferred

    @transact
    def _get_reprepro(self, account_id, person_id, distribution_id, pocket_id):
        """Return the config to pass to reprepro."""
        account = get_account_by_id(account_id)
        person = get_person_by_id(person_id)
        management = RepositoryManagement(account, person)
        if distribution_id:
            distribution = management.get_distributions(ids=[distribution_id]).one()
            if distribution is None:
                raise Unauthorized(
                    "Person is not authorized for this operation on "
                    "distribution id: {}".format(distribution_id)
                )
            return distribution.get_reprepro()
        else:
            pocket = management.get_pockets(ids=[pocket_id]).one()
        if pocket is None:
            raise Unauthorized(
                "Person is not authorized for this operation on "
                "pocket id: {}".format(pocket_id)
            )
        return pocket.get_reprepro()

    def _run_reprepro(self, reprepro, reprepro_method):
        """Call the designated reprepro operation."""
        method = getattr(reprepro, reprepro_method)
        return method()


def create_job_for_reprepro(
    account, person, reprepro_method, distribution=None, pocket=None, page_key=None
):
    """
    Create a job which calls a reprepro method for a pocket or distribution.

    @param account: The Account in which the pocket or distribution lives.
    @param person: The person submitting the job.
    @param reprepro_method: The reprepro method the job will call.
    @param distribution: Optionally, the Distribution on which reprepro will
        operate. If not set, Pocket must be set.
    @param pocket: Optionally, the Pocket on which reprepro will operate. If
        not set, Distribution must be set.
    @param page_key: Optionally, the page key to pass to the created job.
    """
    assert any(
        [distribution, pocket]
    ), "Either distribution or pocket must be specified"
    dispatcher = getUtility(IDispatcher)
    distribution_id = distribution.id if distribution is not None else None
    pocket_id = pocket.id if pocket is not None else None
    dispatcher.create_job(
        page_key=page_key,
        namespace=None,
        action="reprepro",
        account_id=account.id,
        person_id=person.id,
        reprepro_method=reprepro_method,
        distribution_id=distribution_id,
        pocket_id=pocket_id,
    )
