# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.security.interfaces import Unauthorized

from aptsources.sourceslist import SourceEntry

from storm.locals import Select, And

from canonical.lib.text import slug_from_title, describe_dict, describe_list
from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.account.computergroup import ComputerGroupItem
from canonical.landscape.model.repository.distribution import (
    Distribution,
    DuplicateDistributionError,
    get_distribution,
)
from canonical.landscape.model.repository.series import Series
from canonical.landscape.model.repository.pocket import Pocket
from canonical.landscape.model.repository.upload import Upload
from canonical.landscape.model.repository.profile import (
    RepositoryProfile,
    RepositoryProfileAPTSource,
    RepositoryProfilePocket,
    UnknownRepositoryProfileError,
)
from canonical.landscape.model.repository.aptsource import (
    APTSource,
    DuplicateAPTSourceError,
    UnknownAPTSourceError,
)
from canonical.landscape.model.repository.activities import SyncPocketRequest
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.group import (
    ActivityGroup,
    CHANGE_REPOSITORY_POCKET_SUMMARY,
)
from canonical.landscape.model.entity.data import AccountEntityData


__all__ = ["RepositoryManagement", "InvalidAPTLine", "get_profiles_for_computer"]

__clearlist__ = ["RepositoryManagement"]

__requires_prefix__ = "landscape."


class InvalidAPTLine(Exception):
    """Raised when an APT line has invalid format."""


@requires_advisor
class RepositoryManagement(ModelManagement):
    """Manage repository-related objects in an L{Account}."""

    requires("ViewRepositoryProfile", PRIVATE, ["account", "namespace"])

    @property
    def namespace(self):
        """Return the namespace for repository-related notifications."""
        return "landscape.account.%d.repository" % self.account.id

    @requires("CreateRepositoryProfile")
    def create_profile(
        self, title, description=None, access_context_name=DEFAULT_ACCESS_CONTEXT_NAME
    ):
        """Create a L{RepositoryProfile}."""
        access_context = self.get_access_context(
            access_context_name, "CreateRepositoryProfile"
        )

        def check_existing(slug):
            """
            Checks if an existing repository profile exists.
            """
            return self.account.store.find(
                RepositoryProfile.name, RepositoryProfile.account_id == self.account.id
            )

        name = slug_from_title(title, 50, check_existing)
        profile = RepositoryProfile(
            self.account, name, title, description or "", access_context.id
        )
        prefix = "Repository profile created with"
        values = {
            "name": name,
            "title": title,
            "description": description,
            "access group": access_context.name,
        }
        message = "%s." % describe_dict(prefix, values, pluralize=False)
        self._event_log.create_log_entry(profile, message)
        return profile

    @requires("ViewAccount")
    def edit_profile(
        self,
        name,
        title=None,
        description=None,
        add_pockets=(),
        remove_pockets=(),
        add_apt_sources=(),
        remove_apt_source_names=(),
    ):
        """
        Change a L{RepositoryProfile}.

        @param name: The name of the repository profile to be edited.
        @param title: The new title of the repository profile.
        @param description: The new description of the repository profile.
        @param add_pockets: List of L{RepositoryProfilePockets} to link to this
            profile.
        @param remove_pockets: List of L{RepositoryProfilePockets} to unlink
            from this profile.
        @param add_apt_sources: List of L{APTSources} to link to this profile.
        @param remove_apt_source_names: List of names of L{APTSources} to
            unlink from this profile. This is a list of strings instead of
            L{APTSources} because a restricted admin may not have acces to an
            APTSource but can have visibility to the L{APTSource} name.
        """
        access_context_ids = self.get_access_context_ids("EditRepositoryProfile")
        profile = self.get_profile_by_name(name)
        if profile.access_context.id not in access_context_ids:
            raise Unauthorized(
                "Permission denied. Can't change repository profile '%s'" % name
            )

        messages = []

        if title is not None and title != profile.title:
            messages.append("Title changed from '%s' to '%s'." % (profile.title, title))
            profile.title = title

        if description and description != profile.description:
            messages.append("Description changed to '%s'." % description)
            profile.description = description

        added_pockets = set(add_pockets).difference(profile.pockets)
        if added_pockets:
            pocket_names = [
                "%s:%s:%s"
                % (pocket.series.distribution.name, pocket.series.name, pocket.name)
                for pocket in added_pockets
            ]
            messages.append("%s." % describe_list("Added pocket", pocket_names))
            profile.add_pockets(add_pockets)

        removed_pockets = set(remove_pockets).intersection(profile.pockets)
        if removed_pockets:
            pocket_names = [
                "%s:%s:%s"
                % (pocket.series.distribution.name, pocket.series.name, pocket.name)
                for pocket in removed_pockets
            ]
            messages.append("%s." % describe_list("Removed pocket", pocket_names))
            profile.remove_pockets(remove_pockets)

        added_apt_sources = set(add_apt_sources).difference(profile.apt_sources)
        if added_apt_sources:
            apt_source_names = [source.name for source in added_apt_sources]
            messages.append("%s." % describe_list("Added APT source", apt_source_names))
            profile.add_apt_sources(add_apt_sources)

        if remove_apt_source_names:
            messages.append(
                "%s." % describe_list("Removed APT source", remove_apt_source_names)
            )
            for source_name in remove_apt_source_names:
                profile.remove_apt_source(source_name)

        if messages:
            self._event_log.create_log_entry(profile, " ".join(messages))
        return profile

    @requires("ViewAccount")
    def remove_profiles(self, profiles):
        """Remove L{RepositoryProfile}s from L{Account}."""
        access_context_ids = self.get_access_context_ids("EditRepositoryProfile")
        remaining_profiles = []
        # Calling profile.apply() will remove the profile from the
        # linked computers, since the created activity will send all the
        # profiles that the computer are linked to when the message gets
        # delivered, at which time this profile doesn't exist anymore
        # and thus won't be inlcluded in the message.
        for profile in profiles:
            if profile.access_context_id in access_context_ids:
                self._event_log.create_log_entry(profile, "Repository profile removed.")
                profile.apply()
                profile.remove()
            else:
                remaining_profiles.append(profile)
        return remaining_profiles

    @requires("ViewRepositoryProfile")
    def get_profiles(self, ids=None, names=None, apt_sources_ids=None):
        """Get the account's L{RepositoryProfile}s."""
        access_context_ids = self.get_access_context_ids("ViewRepositoryProfile")

        where = [
            RepositoryProfile.account_id == self.account.id,
            RepositoryProfile.access_context_id.is_in(access_context_ids),
        ]
        if ids is not None:
            where.append(RepositoryProfile.id.is_in(ids))
        if names is not None:
            where.append(RepositoryProfile.name.is_in(names))
        if apt_sources_ids is not None:
            where.extend(
                [
                    RepositoryProfileAPTSource.profile_id == RepositoryProfile.id,
                    RepositoryProfileAPTSource.apt_source_id.is_in(apt_sources_ids),
                ]
            )
        result = self.account.store.find(RepositoryProfile, *where)
        return result.config(distinct=True).order_by(RepositoryProfile.name)

    def get_profile_by_name(self, name):
        """Get a single L{RepositoryProfile} by name.

        @raises: UnknownRepositoryProfileError if no profile with the given
            name exists in the account.
        """
        profile = self.get_profiles(names=[name]).one()
        if not profile:
            raise UnknownRepositoryProfileError(
                "No repository profile found with name '%s'." % name
            )
        return profile

    def get_profiles_for_computer(self, computer):
        """Gets a list of RepositoryProfile for a computer."""
        access_context_ids = self.get_access_context_ids("ViewRepositoryProfile")
        result = self.account.store.find(
            RepositoryProfile,
            RepositoryProfile.account_id == computer.account.id,
            RepositoryProfile.access_context_id.is_in(access_context_ids),
            (
                RepositoryProfile.computer_group_id
                == ComputerGroupItem.computer_group_id
            ),
            ComputerGroupItem.computer_id == computer.id,
        )
        return result.config(distinct=True).order_by(RepositoryProfile.name)

    def create_distribution(self, name, access_context=None):
        """Create a new L{Distribution} in the L{Account}.

        @param name: The name of the L{Distribution} to create.
        """
        if access_context is None:
            # We rely on the service layer to check authorization.
            access_context = self.access_context

        if get_distribution(self.account, name=name) is not None:
            raise DuplicateDistributionError(["name"])

        return self.account.store.add(
            Distribution(self.account, name, access_context=access_context)
        )

    def remove_distribution(self, distribution):
        """
        Remove the given L{Distribution}.
        """
        result = self.get_pockets(distribution=distribution)
        self._remove_pocket_references(list(result.values(Pocket.id)))
        # Remove the related entity data
        self.account.store.remove(distribution.data)
        self.account.store.remove(distribution)

    def get_distributions(self, ids=None, names=None):
        """Get the account L{Distribution}s."""
        access_context_ids = self.get_access_context_ids("ViewDistribution")

        where = []
        if ids is not None:
            where.append(Distribution.id.is_in(ids))
        if names is not None:
            where.append(AccountEntityData.name.is_in(names))

        return self.account.store.find(
            Distribution,
            Distribution.data_id == AccountEntityData.id,
            AccountEntityData.account_id == self.account.id,
            AccountEntityData.access_context_id.is_in(access_context_ids),
            *where
        ).order_by(AccountEntityData.name)

    def get_distribution(self, name):
        """
        Conveniency method to get a single distribution, for API
        L{SubjectRoute} use.
        """
        return self.get_distributions(names=[name]).one()

    def get_pockets(self, ids=None, series=None, distribution=None):
        """Return a C{ResultSet} of matching L{Pocket}s.

        @param ids: Get only pockets with the given L{Pocked.id}s.
        @param distribution: Get only pockets in the given L{Distribution}.
        @param series: Get only pockets in the given L{Series}.
        """
        access_context_ids = self.get_access_context_ids("ViewDistribution")

        where = [
            Series.id == Pocket.series_id,
            Distribution.id == Series.distribution_id,
            Distribution.data_id == AccountEntityData.id,
            AccountEntityData.account_id == self.account.id,
            AccountEntityData.access_context_id.is_in(access_context_ids),
        ]
        if ids is not None:
            where.append(Pocket.id.is_in(ids))
        if series is not None:
            where.append(Series.id == series.id)
        if distribution is not None:
            where.append(Distribution.id == distribution.id)
        return self.account.store.find(Pocket, *where)

    def remove_series(self, series):
        """
        Remove specified L{Series}. If unauthorised, return the unaltered
        L{Series}.
        """
        result = self.get_pockets(series=series)
        self._remove_pocket_references(list(result.values(Pocket.id)))
        self.account.store.remove(series)

    def remove_pocket(self, pocket):
        """
        Remove the given L{Pocket}.
        """
        self._remove_pocket_references([pocket.id])
        self.account.store.remove(pocket)

    def _remove_pocket_references(self, ids):
        """Remove all model entities associated to the given L{Pocket.id}s.

        This includes activities and repository profiles.
        """
        store = self.account.store

        # Select all activities of info type SyncPocketRequest associated
        # to the given pocket ids
        subquery = Select(
            Activity.parent_id,
            And(
                SyncPocketRequest.activity_id == Activity.id,
                SyncPocketRequest.pocket_id.is_in(ids),
            ),
            default_tables=[Activity, SyncPocketRequest],
        )

        result = store.find(ActivityParent, ActivityParent.id.is_in(subquery))
        result.remove()

        result = store.find(
            RepositoryProfilePocket, RepositoryProfilePocket.pocket_id.is_in(ids)
        )

        result.remove()

    def get_uploads(self, pocket=None):
        """Return a C{ResultSet} of matching L{Upload}s."""
        where = [
            Distribution.id == Upload.distribution_id,
            Distribution.data_id == AccountEntityData.id,
            AccountEntityData.account_id == self.account.id,
        ]
        if pocket is not None:
            where.append(Upload.pocket_id == pocket.id)
        return self.account.store.find(Upload, *where)

    def create_sync_pocket_activity(self, pocket):
        """Create an L{ActivityGroup} with a L{SyncPocketRequest} child."""
        parent = self.account.create_activity(
            ActivityGroup, summary=CHANGE_REPOSITORY_POCKET_SUMMARY, creator=self.person
        )
        parent.add(SyncPocketRequest, entity=pocket.series.distribution, pocket=pocket)
        parent.settle()
        return parent

    def _validate_apt_line(self, line):
        entry = SourceEntry(line)
        if entry.invalid or entry.type not in ("deb", "deb-src"):
            raise InvalidAPTLine("Invalid APT line format")

    def create_apt_source(
        self, name, line, gpg_key=None, access_context_name=DEFAULT_ACCESS_CONTEXT_NAME
    ):
        """Create an L{APTSource}."""
        access_context = self.get_access_context(access_context_name, "CreateAPTSource")
        self.check_duplicate(APTSource, DuplicateAPTSourceError, name=name)
        self._validate_apt_line(line)
        return APTSource(self.account, name, line, gpg_key, access_context)

    def get_apt_sources(self, names=None, ids=None):
        """
        Get the account's L{APTSource}s.

        Results returned will be filtered to access contexts for which the
        person has C{ViewAPTSource} permission.

        @param names: Get only L{APTSource}s with the given names.
        @param ids: Get only L{APTSource}s with the given ids.
        """
        access_context_ids = self.get_access_context_ids("ViewAPTSource")
        where = [
            APTSource.account_id == self.account.id,
            APTSource.access_context_id.is_in(access_context_ids),
        ]
        if names is not None:
            where.append(APTSource.name.is_in(names))
        if ids is not None:
            where.append(APTSource.id.is_in(ids))
        return self.account.store.find(APTSource, *where).order_by(APTSource.name)

    def get_apt_source_by_name(self, name):
        """Get a single L{APTSource} by name.

        @raises: UnknownAPTSourceError if no profile with the given
            name exists in the account.
        """
        apt_source = self.get_apt_sources(names=[name]).one()
        if not apt_source:
            raise UnknownAPTSourceError("No apt source found with name '%s'." % name)
        return apt_source

    def remove_apt_sources(self, apt_source_ids):
        """Remove L{APTSource}s with given ids."""
        access_context_ids = self.get_access_context_ids("EditAPTSource")

        remaining_apt_sources = []
        for apt_source in self.get_apt_sources(ids=apt_source_ids):
            if apt_source.access_context.id in access_context_ids:
                apt_source.remove()
            else:
                remaining_apt_sources.append(apt_source.get_state())
        return remaining_apt_sources


def get_profiles_for_computer(computer):
    """Return L{RepositoryProfile}s for the L{Computer}."""

    result = computer.account.store.find(
        RepositoryProfile,
        RepositoryProfile.account_id == computer.account.id,
        (RepositoryProfile.computer_group_id == ComputerGroupItem.computer_group_id),
        ComputerGroupItem.computer_id == computer.id,
    )
    return result.config(distinct=True)
