# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from storm.locals import (
    Int,
    Unicode,
    AutoReload,
    DateTime,
    Reference,
    ReferenceSet,
    List,
    Bool,
)

from canonical.lib.timestamp import format_datetime

from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.model.gpg.key import GPGKey
from canonical.landscape.model.repository.upload import Upload
from canonical.landscape.model.repository.constants import DEFAULT_VERIFY_RELEASE

MODE_MIRROR = 0
MODE_PULL = 1
MODE_UPLOAD = 2

FILTER_WHITELIST = 1
FILTER_BLACKLIST = 2


def get_suite_name(series, pocket):
    """
    Return the full pocket name as it will appear in the APT repository,
    (i.e. the name under the dists/ directory).

    @param series: the series name
    @param pocket: the pocket name
    """
    name = series
    if pocket != "release":
        name = "%s-%s" % (name, pocket)
    return name


class PocketUploadGPGKey(object):
    """
    Link pockets in upload mode to GPG keys allowed to sign uploaded packages.
    """

    __storm_table__ = "pocket_upload_gpg_key"
    __storm_primary__ = "pocket_id", "gpg_key_id"

    pocket_id = Int(allow_none=False)
    gpg_key_id = Int(allow_none=False)


class Pocket(object):
    """A pocket part of a L{Series}.

    @param series: The L{Series} we're part of.
    @param name: The name of the pocket.
    @param components: The list of components (e.g. main, universe, ...)
    @param include_udeb: Flag to choose whether the pocket should include
        components for .udeb packages too.
    @param archs: The list of architectures (e.g. i386, amd64, ...)
    @param gpg_key: The GPG key to sign pocket package indices.
    """

    __storm_table__ = "pocket"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    series_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    mode = Int(allow_none=False, default=2)
    gpg_key_id = Int()
    mirror_uri = Unicode()
    mirror_suite = Unicode()
    mirror_gpg_key_id = Int()
    pull_pocket_id = Int()
    upload_allow_unsigned = Bool()
    creation_time = DateTime(default=AutoReload)
    components = List(type=Unicode())
    archs = List(type=Unicode())
    include_udeb = Bool(default=False)
    filter_type = Int()
    filters = List(type=Unicode())

    pull_pocket = Reference(pull_pocket_id, id)
    gpg_key = Reference(gpg_key_id, GPGKey.id)
    mirror_gpg_key = Reference(mirror_gpg_key_id, GPGKey.id)
    upload_gpg_keys = ReferenceSet(
        id,
        PocketUploadGPGKey.pocket_id,
        PocketUploadGPGKey.gpg_key_id,
        GPGKey.id,
        order_by=GPGKey.name,
    )

    def __init__(
        self,
        series,
        name="release",
        components=[],
        archs=[],
        include_udeb=False,
        gpg_key=None,
        filter_type=None,
    ):
        self.series = series
        self.name = name
        self.components = components
        self.archs = archs
        self.include_udeb = include_udeb
        self.gpg_key = gpg_key
        self.filters = []

    def get_full_name(self):
        """
        Return the full pocket name as it will appear in the APT repository.
        """
        return get_suite_name(self.series.name, self.name)

    def get_mode_name(self):
        """
        Return the name of the mode the pocket is in.
        """
        if self.is_in_mirror_mode():
            return "mirror"
        if self.is_in_pull_mode():
            return "pull"
        if self.is_in_upload_mode():
            return "upload"

    def get_filter_type_name(self):
        """
        Return the filter type name for the pocket.
        """
        if self.filter_type == FILTER_BLACKLIST:
            return "blacklist"
        if self.filter_type == FILTER_WHITELIST:
            return "whitelist"

    def get_verify_release(self):
        """
        Get the key to be used in reprepro's VerifyRelease.
        """
        if self.mirror_gpg_key is None:
            return DEFAULT_VERIFY_RELEASE
        else:
            return "{}+".format(self.mirror_gpg_key.key_id)

    def get_sign_with(self):
        """
        Get the key to be used in reprepro's SignWith.
        """
        if self.gpg_key is None:
            return None
        else:
            return self.gpg_key.key_id

    def is_in_mirror_mode(self):
        """Return C{True} if the pocket is in mirror mode, C{False} if not."""
        return self.mode == MODE_MIRROR

    def is_in_pull_mode(self):
        """Return C{True} if the pocket is in pull mode, C{False} if not."""
        return self.mode == MODE_PULL

    def is_in_upload_mode(self):
        """Return C{True} if the pocket is in upload mode, C{False} if not."""
        return self.mode == MODE_UPLOAD

    def _set_mode(
        self,
        mode,
        mirror_uri=None,
        mirror_suite=None,
        mirror_gpg_key=None,
        pull_pocket=None,
        upload_allow_unsigned=None,
        filter_type=None,
    ):
        """Set the details of a mode."""
        self.mode = mode
        self.mirror_uri = mirror_uri
        self.mirror_suite = mirror_suite
        self.mirror_gpg_key = mirror_gpg_key
        self.pull_pocket = pull_pocket
        self.upload_allow_unsigned = upload_allow_unsigned
        self.upload_gpg_keys.clear()
        self.filter_type = filter_type

    def set_mirror_mode(self, mirror_uri, mirror_suite, mirror_gpg_key=None):
        """Set the pocket to mirror mode.

        Packages will be mirrored from an external APT repository.

        @param mirror_uri: The base URI of the APT repository to mirror from.
        @param mirror_suite: The entry under 'dists/' to mirror from.
        @param mirror_gpg_key: The GPG key to use to verify the repository
            signature, if C{None} is given the stock Ubuntu one will be used.
        """
        self._set_mode(
            MODE_MIRROR,
            mirror_uri=mirror_uri,
            mirror_suite=mirror_suite,
            mirror_gpg_key=mirror_gpg_key,
        )

    def set_pull_mode(self, pull_pocket, filter_type=None):
        """Set the pocket to pull mode.

        Packages will be pulled from another pocket.

        @param pull_pocket: The L{Pocket} to pull from.
        """
        self._set_mode(MODE_PULL, pull_pocket=pull_pocket, filter_type=filter_type)

    def set_upload_mode(self, allow_unsigned=True):
        """Set the pocket to upload mode.

        @param allow_unsigned: Whether the uploaded packages should be signed.
        """
        self._set_mode(MODE_UPLOAD, upload_allow_unsigned=allow_unsigned)

    def add_upload_gpg_keys(self, gpg_keys):
        """
        Add the given gpg keys to the set of allowed uploader gpg keys.
        """
        for gpg_key in gpg_keys:
            self.upload_gpg_keys.add(gpg_key)

    def remove_upload_gpg_keys(self, gpg_keys):
        """
        Remove the given gpg keys to the set of allowed uploader gpg keys.
        """
        for gpg_key in gpg_keys:
            self.upload_gpg_keys.remove(gpg_key)

    def add_filters(self, names):
        """Adds specified package names to the filter list."""
        filters = set(self.filters)
        filters.update(names)
        self.filters = sorted(filters)

    def remove_filters(self, names):
        """
        Remove specified package names from the filter list.
        Unknown names will be ignored.
        """
        filters = set(self.filters)
        filters.difference_update(names)
        self.filters = sorted(filters)

    def get_state(self):
        """Return a dict with information on the L{Pocket}."""
        modes = ["mirror", "pull", "upload"]

        gpg_key_state = self.gpg_key.get_state() if self.gpg_key else None

        result = {
            "name": self.name,
            "creation_time": format_datetime(self.creation_time),
            "mode": modes[self.mode],
            "gpg_key": gpg_key_state,
            "components": self.components,
            "architectures": self.archs,
            "include_udeb": self.include_udeb,
            "apt_source_line": self.get_apt_source_line(),
            "series": {
                "name": self.series.name,
                "creation_time": format_datetime(self.series.creation_time),
            },
            "distribution": self.series.distribution.data.get_state(),
        }
        if self.is_in_mirror_mode():
            result["mirror_uri"] = self.mirror_uri
            result["mirror_suite"] = self.mirror_suite
            if self.mirror_gpg_key:
                result["mirror_gpg_key"] = self.mirror_gpg_key.get_state()
        if self.is_in_pull_mode():
            result["pull_pocket"] = "%s/%s" % (
                self.pull_pocket.series.name,
                self.pull_pocket.name,
            )
            result["filter_type"] = self.get_filter_type_name()
            result["filters"] = self.filters
        if self.is_in_upload_mode():
            result["upload_allow_unsigned"] = self.upload_allow_unsigned
            result["upload_gpg_keys"] = [
                key.get_state() for key in self.upload_gpg_keys
            ]
        return result

    def get_reprepro(self):
        """
        Return a L{Reprepro} wrapper configured to operate on this pocket.
        """
        reprepro = self.series.distribution.get_reprepro()
        reprepro.set_distribution(self.get_full_name())
        return reprepro

    def get_apt_source_line(self):
        """Return an Apt source line for this pocket.

        E.g. "deb http://landscape.example.com/ubuntu pocket-name main"
        """
        root_url = get_root_url()
        if root_url is not None:
            if root_url.startswith("https"):
                root_url = root_url.replace("https", "http", 1)

        line = (
            "deb %(root_url)srepository/%(account)s/%(distribution)s"
            " %(pocket)s %(components)s"
            % {
                "root_url": root_url,
                "account": self.series.distribution.account.name,
                "distribution": self.series.distribution.name,
                "pocket": self.get_full_name(),
                "components": " ".join(component for component in self.components),
            }
        )
        return line

    def create_upload(self, changes):
        """Create a new L{Upload} entry in the upload queue.

        @param changes: The filename of the uploaded .changes.
        """
        return Upload(self, changes)


Upload.pocket = Reference(Upload.pocket_id, Pocket.id)


def get_account_pocket_by_id(account_id, pocket_id):
    """
    Return a L{Pocket} in an L{Account} from respective Ids.
    """
    account = get_account_by_id(account_id)
    if account is None:
        return None
    pocket = account.store.get(Pocket, pocket_id)
    # Make sure that the pocket belongs to the given account
    if pocket and pocket.series.distribution.account is not account:
        return None
    return pocket
