# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, Unicode, AutoReload, DateTime, Reference, ReferenceSet

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.account.helper import ComputerGroupHelper
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.activity.group import (
    ActivityGroup,
    APPLY_REPOSITORY_PROFILE_SUMMARY,
)
from canonical.landscape.model.activity.status import (
    QUEUED,
    BLOCKED,
    SCHEDULED,
    PENDING_STATUSES,
)
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.repository.activities import (
    ChangeRepositoryProfilesRequest,
)
from canonical.landscape.model.repository.pocket import Pocket
from canonical.landscape.model.repository.aptsource import (
    APTSource,
    UnknownAPTSourceError,
)


__all__ = [
    "DuplicateRepositoryProfileError",
    "UnknownRepositoryProfileError",
    "RepositoryProfile",
    "RepositoryProfilePocket",
    "RepositoryProfileAPTSource",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


class DuplicateRepositoryProfileError(DuplicateModelError):
    """Raised when a repository profile already exists."""

    model_name = "Repository profile"


class UnknownRepositoryProfileError(Exception):
    """Raised when a repository profile couldn't be found."""


class RepositoryProfilePocket(object):

    __storm_table__ = "repository_profile_pocket"
    __storm_primary__ = "profile_id", "pocket_id"

    profile_id = Int(allow_none=False)
    pocket_id = Int(allow_none=False)


class RepositoryProfileAPTSource(object):

    __storm_table__ = "repository_profile_apt_source"
    __storm_primary__ = "profile_id", "apt_source_id"

    profile_id = Int(allow_none=False)
    apt_source_id = Int(allow_none=False)


@requires_advisor
class RepositoryProfile(ComputerGroupHelper):

    __storm_table__ = "repository_profile"

    requires(
        "ViewRepositoryProfile",
        PRIVATE,
        [
            "id",
            "account",
            "creation_time",
            "get_computers",
            "get_computer_count",
            "get_computer_ids",
            "get_tags",
        ],
    )
    requires("ViewRepositoryProfile", "EditRepositoryProfile", ["name", "description"])
    requires(
        "EditRepositoryProfile",
        PRIVATE,
        [
            "update_tags",
            "add_tags",
            "remove_tags",
            "enable_for_all_computers",
            "disable_for_all_computers",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    title = Unicode(allow_none=False)
    description = Unicode(allow_none=False)
    creation_time = DateTime(default=AutoReload)
    access_context_id = Int(allow_none=False)

    computer_group_id = Int()
    computer_group = Reference(computer_group_id, ComputerGroup.id)

    pockets = ReferenceSet(
        id,
        RepositoryProfilePocket.profile_id,
        RepositoryProfilePocket.pocket_id,
        Pocket.id,
    )

    apt_sources = ReferenceSet(
        id,
        RepositoryProfileAPTSource.profile_id,
        RepositoryProfileAPTSource.apt_source_id,
        APTSource.id,
    )

    _changes = {}

    def __init__(self, account, name, title, description, access_context_id=None):
        super(RepositoryProfile, self).__init__(account, access_context_id)
        self.name = name
        self.title = title
        self.description = description
        self._create_computer_group()

    @requires("ViewRepositoryProfile")
    def get_state(self):
        """Return a dict with information about the profile."""
        all_computers = self.computer_group.all_computers
        tags = sorted(tag.name for tag in self.get_tags())
        pockets = [pocket.get_state() for pocket in self.pockets]
        return {
            "id": self.id,
            "access_group": self.access_context.name,
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "all_computers": all_computers,
            "tags": tags,
            "pockets": pockets,
            "apt_sources": [source.name for source in self.apt_sources],
            "pending_count": self.get_pending_activities().count(),
        }

    def get_apt_source_lines(self):
        """
        Return a list of all deb lines for the profile, including those for
        linked pockets.
        """
        apt_lines = []
        apt_lines.extend(pocket.get_apt_source_line() for pocket in self.pockets)
        apt_lines.extend(source.line for source in self.apt_sources)
        return apt_lines

    def get_gpg_keys_material(self):
        """
        Return a set of material for GPG keys used by profile sources.
        """
        gpg_keys = set()
        gpg_keys.update(
            pocket.gpg_key.material
            for pocket in self.pockets
            if pocket.gpg_key is not None
        )
        gpg_keys.update(
            source.gpg_key.material
            for source in self.apt_sources
            if source.gpg_key is not None
        )
        return gpg_keys

    def add_computers_with_ids(self, computer_ids):
        computer_ids = super(RepositoryProfile, self).add_computers_with_ids(
            computer_ids
        )

        if computer_ids:
            self.apply(computer_ids=computer_ids)
        return computer_ids

    def remove_computers_with_ids(self, computer_ids):
        computer_ids = super(RepositoryProfile, self).remove_computers_with_ids(
            computer_ids
        )

        if computer_ids:
            self.apply(computer_ids=computer_ids)
        return computer_ids

    @requires("EditRepositoryProfile")
    def remove(self):
        super(RepositoryProfile, self).remove()

    @requires("ViewRepositoryProfile")
    def get_pending_activities(self, computer_ids=None, statuses=None, values=Activity):
        """Return the activities matching search criteria.

        @param computer_ids: Return only L{Activity}s involving L{Computer}s
            with these ids.
        @param values: Return only the given storm properties, this is a
            workaround needed because of L{ResultSet.union}.
        @see: L{_pending_clause}
        """
        if computer_ids is None:
            computer_ids = self.computer_group.computers.values(AccountComputerTwin.id)

        if statuses is None:
            statuses = (QUEUED, BLOCKED, SCHEDULED)
        statuses = set(statuses).intersection(PENDING_STATUSES)

        collection = ActivityChildCollection(
            self.account.store,
            account=self.account,
            statuses=statuses,
            info_classes=[ChangeRepositoryProfilesRequest],
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
        )
        result = collection.select(values).find()
        return result.find()

    @requires("EditRepositoryProfile")
    def apply(self, computer_ids=None):
        """Create a group activity for changing package profiles.

        @param computer_ids: Optionally a list of L{Computer.id}s to apply the
            profile on. If C{None} the profile will be installed on all
            associated computers.
        @return: A group L{Activity} having a L{ChangeReposioryProfilesRequest}
            request child for each computer in the profile.
        """
        ids = (
            computer_ids
            if computer_ids
            else [computer.id for computer in self.computer_group.computers]
        )

        computers = self.account.get_computers(computer_ids=ids)

        computers_with_pending = set(
            self.get_pending_activities(computer_ids, values=AccountComputerTwin.id)
        )

        computers = [
            computer
            for computer in computers
            if computer.id not in computers_with_pending
        ]

        if not computers:
            return

        parent = self.account.create_activity(
            ActivityGroup, summary=APPLY_REPOSITORY_PROFILE_SUMMARY, creator=None
        )

        event_log = EventLog(self.account, parent.creator)
        for computer in computers:
            child = parent.add(ChangeRepositoryProfilesRequest, computer=computer)
            if computer.distribution is None:
                child.block()

            event_log.create_log_entry(
                computer, "Activity created: %s" % child.info.get_summary()
            )

        parent.settle()
        return parent

    @requires("EditRepositoryProfile")
    def add_pockets(self, pockets):
        """Link L{Pocket}s to the profile."""
        pockets_to_add = set(pockets).difference(self.pockets)

        if pockets_to_add:
            for pocket in pockets_to_add:
                self.pockets.add(pocket)
            self.apply()

    @requires("EditRepositoryProfile")
    def remove_pockets(self, pockets):
        """Unlink L{Pocket}s from the profile."""
        pockets_to_remove = set(pockets).intersection(self.pockets)

        if pockets_to_remove:
            for pocket in pockets_to_remove:
                self.pockets.remove(pocket)
            self.apply()

    def add_apt_sources(self, sources):
        """Link L{APTSource}s to the profile."""
        sources_to_add = set(sources).difference(self.apt_sources)

        if sources_to_add:
            for source in sources_to_add:
                self.apt_sources.add(source)
            self.apply()

    def remove_apt_sources(self, sources):
        """
        DEPRECATED: use remove_apt_source.
        Unlink L{APTSource}s from the profile.
        """
        sources_to_remove = set(sources).intersection(self.apt_sources)

        if sources_to_remove:
            for source in sources_to_remove:
                self.apt_sources.remove(source)
            self.apply()

    def remove_apt_source(self, apt_source_name):
        """
        Unlink named L{APTSource} from the profile if found, otherwise raise
        an C{UnknownAPTSourceError}.
        """
        apt_source = self.apt_sources.find(name=apt_source_name).one()
        if apt_source is None:
            raise UnknownAPTSourceError(
                "Unknown APT source '%s' in profile '%s'" % (apt_source_name, self.name)
            )
        self.apt_sources.remove(apt_source)
        self.apply()


Pocket.profiles = ReferenceSet(
    Pocket.id,
    RepositoryProfilePocket.pocket_id,
    RepositoryProfilePocket.profile_id,
    RepositoryProfile.id,
)
