# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from storm.locals import Int, Unicode, AutoReload, DateTime, Reference, ReferenceSet

from canonical.landscape.model.repository.pocket import Pocket

from canonical.lib.timestamp import format_datetime


class NoPocketsInSeries(Exception):
    """Exception raised if the series does not contain pockets."""


class Series(object):
    """A series part of a L{Distribution}.

    @param distribution: The L{Distribution} we're part of.
    @param name: The name of the series.
    """

    __storm_table__ = "series"

    id = Int(primary=True, allow_none=False, default=AutoReload)
    name = Unicode(allow_none=False)
    distribution_id = Int(allow_none=False)
    creation_time = DateTime(default=AutoReload)

    pockets = ReferenceSet(id, Pocket.series_id, order_by=Pocket.name)

    def __init__(self, distribution, name):
        self.distribution = distribution
        self.name = name

    def create_pocket(
        self,
        name="release",
        components=[],
        archs=[],
        include_udeb=False,
        gpg_key=None,
        filter_type=None,
    ):
        """Create a L{Pocket} in this series."""
        return Pocket(
            self,
            name,
            components=components,
            archs=archs,
            include_udeb=include_udeb,
            gpg_key=gpg_key,
            filter_type=filter_type,
        )

    def get_pockets(self, ids=None, names=None):
        """Get the L{Pocket}s. for a L{Series}."""
        where = []
        if ids is not None:
            where.append(Pocket.id.is_in(ids))
        if names is not None:
            where.append(Pocket.name.is_in(names))
        return self.pockets.find(*where)

    def derive_series(self, name):
        """
        Derives a L{Series} from this one.

        @param name: the name for the derived series.
        @return: the derived series.

        The derived series will have a pocket for each one in the original
        series, set to pull from the corresponding pocket of the origin series.
        """
        if self.pockets.count() == 0:
            raise NoPocketsInSeries(
                "Series '%s' does not contain pockets." % (self.name,)
            )

        derived = self.distribution.create_series(name)

        for pocket in self.pockets:
            derived_pocket = derived.create_pocket(
                pocket.name,
                components=pocket.components,
                archs=pocket.archs,
                gpg_key=pocket.gpg_key,
                include_udeb=pocket.include_udeb,
            )
            derived_pocket.set_pull_mode(pocket)
        return derived

    def get_state(self):
        """Return a dict with information on the L{Series}."""
        result = {}
        result["name"] = self.name
        result["creation_time"] = format_datetime(self.creation_time)
        result["pockets"] = [pocket.get_state() for pocket in self.get_pockets()]
        return result


Pocket.series = Reference(Pocket.series_id, Series.id)
