# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from canonical.event.subscriber import subscriber
from canonical.landscape.model.account.computergroup import ComputerGroup
from canonical.landscape.model.events import (
    AddTagsEvent,
    RemoveTagsEvent,
    AddComputerEvent,
    ResynchronizeEvent,
)
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.repository.profile import RepositoryProfile


@subscriber(IAccount, AddTagsEvent)
def add_tag_to_computers_handler(account, event):
    """Event handler is invoked when L{Tag}s are added to L{Computer}s."""
    account.computer_group.synchronize_with_added_tags(
        event.tag_ids, event.computer_ids, RepositoryProfile
    )


@subscriber(IAccount, RemoveTagsEvent)
def remove_tag_from_computers_handler(account, event):
    """Event handler is invoked when L{Tag}s are removed from L{Computer}s."""
    account.computer_group.synchronize_with_removed_tags(
        event.tag_ids, event.computer_ids, RepositoryProfile
    )


@subscriber(IAccount, AddComputerEvent)
def add_computer_to_account_handler(account, event):
    """
    Event handler is invoked when a L{Computer} is added to an L{Account}.
    """
    profiles = account.store.find(
        RepositoryProfile,
        RepositoryProfile.account_id == account.id,
        RepositoryProfile.computer_group_id == ComputerGroup.id,
        ComputerGroup.all_computers == True,
    )
    for profile in profiles:
        profile.add_computers_with_ids([event.computer.id])


@subscriber(ResynchronizeEvent)
def reapply_profiles(resynchronize):
    """If a computer resynchronizes we need to possibly re-apply its profiles.

    For example when a pending computer is accepted as an existing computer
    a L{ResynchronizeEvent} is triggered, and we catch it in order to apply
    the required profiles on the new computer.
    """
    if not resynchronize.is_in_scope("package"):
        return

    from canonical.landscape.model.repository.management import (
        get_profiles_for_computer,
    )

    computer = resynchronize.computer
    for profile in get_profiles_for_computer(computer):
        profile.apply(computer_ids=[computer.id])
