# Copyright 2005-2022 Canonical Limited. All rights reserved.

from collections import defaultdict

from storm.locals import Int, Unicode, AutoReload

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.lib.text import describe_list
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.main.computer import get_computers_by_ids
from canonical.landscape.model.computer.helper import ComputerHelper

from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)


__all__ = [
    "Annotation",
    "AnnotationManagement",
    "validated_annotation",
    "InvalidAnnotationKeyError",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


class InvalidAnnotationKeyError(Exception):
    """
    Raised when a annotation key specified contains a colon, C{:}.
    """


def validated_annotation(data):
    """Ensure the keys do not contain invalid character like C{:}.

    Return a C{Dict} of validated annotation with whitespace stripped."""
    invalid_keys = []
    validated_annotation = {}
    for key, value in list(data.items()):
        if key is None or key.find(":") > 0 or len(key.strip()) == 0:
            invalid_keys.append(key)
        else:
            if value:
                validated_annotation[key.strip()] = value.strip()
            else:
                validated_annotation[key.strip()] = value

    if invalid_keys:
        if len(invalid_keys) > 1:
            error_str = "Invalid annotation keys %s" % ", ".join(
                ["'%s'" % key for key in invalid_keys]
            )
        else:
            error_str = "Invalid annotation key '%s'" % invalid_keys[0]
        raise InvalidAnnotationKeyError(
            "%s. Keys can not be empty and can not contain ':'." % error_str
        )
    return validated_annotation


@requires_advisor
class AnnotationManagement(ModelManagement):
    @requires("ViewAccount")
    def get_annotation(self, computer_ids, keys=None):
        """Return a dict of L{Annotation} associated with C{computer_ids}.

        Since API and UI use ComputerSelection routes which already check
        C{ViewComputer} permission, we don't need access context checks within.
        """
        # Are we searching for an invalid key?
        if keys:
            annotation = validated_annotation(dict.fromkeys(keys))
            keys = list(annotation.keys())

        resource_stores = get_resource_stores_for_computer_ids(computer_ids)
        result = defaultdict(dict)
        for store, store_computer_ids in list(resource_stores.items()):
            where = [Annotation.computer_id.is_in(store_computer_ids)]
            if keys:
                where.append(Annotation.key.is_in(keys))
            data = store.find(
                (Annotation.computer_id, Annotation.key, Annotation.value), *where
            )
            for computer_id, key, value in data:
                result[computer_id][key] = value
        return result

    @requires("ViewAccount")
    def add_annotation(self, computer_ids, data):
        """Store a new L{Annotation} for the specified computers.

        Replace any pre-existing key and value with the current provided
        C{data} dictionary. Since API and UI use ComputerSelection routes
        which already check C{EditComputerDetails} permissions. No additional
        permission checks are required within.
        """
        # Are we trying to add an invalid key?
        if data:
            validated_data = validated_annotation(data)

        if not computer_ids:
            return

        computers = get_computers_by_ids(computer_ids)
        validated_keys = sorted(validated_data.keys())
        self.remove_annotation(computer_ids, validated_keys, audit_notice=False)
        for computer in computers:
            for key, value in list(validated_data.items()):
                annotation = Annotation(computer, key, value)
                annotation.store.add(annotation)
            if validated_keys:
                message = "%s." % describe_list("Added annotation key", validated_keys)
                self._event_log.create_log_entry(computer, message)

    @requires("ViewAccount")
    def remove_annotation(self, computer_ids, keys=None, audit_notice=True):
        """Remove L{Annotation} keys from computers matching C{computer_ids}.
        Access context checks for C{EditComputerDetails} permissions are done
        in the API and UI layers."""

        # Are we trying to remove an invalid key?
        if keys:
            annotation = validated_annotation(dict.fromkeys(keys))
            keys = list(annotation.keys())

        resource_stores = get_resource_stores_for_computer_ids(computer_ids)
        for store, store_computer_ids in list(resource_stores.items()):
            computers = get_computers_by_ids(store_computer_ids)
            where = [Annotation.computer_id.is_in(store_computer_ids)]
            if keys:
                where.append(Annotation.key.is_in(keys))
            matched_computers = set(
                store.find(Annotation.computer_id, *where).config(distinct=True)
            )

            store.find(Annotation, *where).remove()

            if audit_notice:
                for computer in computers:
                    if computer.id not in matched_computers:
                        continue
                    # Creates an audit log activity
                    keys = sorted(keys or [])
                    if keys:
                        message = "%s." % describe_list("Removed annotation key", keys)
                        self._event_log.create_log_entry(computer, message)


@requires_advisor
class Annotation(ComputerHelper):
    """Annotation table only tracks user-added computer annotations.

    Annotations can be associated with a computer via an API method call
    AddAnnotationToComputers or through landscape-client annotation messages
    which are sent as part of each computer_info message exchange. On the
    client, these annotations are automatically constructed from any files
    found in the client's C{data_path}/annotations.d directory."""

    __storm_table__ = "annotation"

    requires("ViewComputer", PRIVATE, ["id", "computer"])
    requires("ViewComputer", "EditComputerDetails", ["key", "value"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    computer_id = Int(allow_none=False)
    key = Unicode(allow_none=False)
    value = Unicode()

    def __init__(self, computer, key, value):
        super(Annotation, self).__init__(computer)
        self.key = key
        self.value = value
