# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, Unicode

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.computer.helper import ComputerHelper


__all__ = ["CloudInstanceMetadata", "CloudInstanceMetadataManagement"]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


@requires_advisor
class CloudInstanceMetadata(ComputerHelper):
    """Store Instance Metadata for cloud virtual machine instances.

    Cloud metadata will automatically be added to a computer through
    landscape-client messages. Cloud metadata can be obtained or updated each
    client restart with any discovered EC2 metadata.
    """

    __storm_table__ = "cloud_instance_metadata"

    requires("ViewComputer", PRIVATE, ["computer"])
    requires("ViewComputer", PRIVATE, ["ami_id", "instance_id", "instance_type"])

    computer_id = Int(primary=True, allow_none=False)
    ami_id = Unicode(allow_none=False)
    instance_id = Unicode(allow_none=False)
    instance_type = Unicode(allow_none=False)

    def __init__(self, computer, ami_id, instance_id, instance_type):
        super(CloudInstanceMetadata, self).__init__(computer)
        self.ami_id = ami_id
        self.instance_id = instance_id
        self.instance_type = instance_type


@requires_advisor
class CloudInstanceMetadataManagement(ComputerHelper):
    """
    A helper class to associate a C{CloudInstanceMetadata} object to a
    computer.
    """

    def set_instance_metadata(self, ami_id, instance_id, instance_type):
        """Set the C{key} and C{value} as instance data for a computer.

        If {key} already exists, replace C{value} with the updated {value}.
        """
        result = self.store.find(
            CloudInstanceMetadata, CloudInstanceMetadata.computer_id == self.computer.id
        )
        instance_metadata = result.one()
        if instance_metadata is None:
            instance_metadata = CloudInstanceMetadata(
                self.computer, ami_id, instance_id, instance_type
            )
            self.store.add(instance_metadata)
        else:
            # Change values if the instance metadata already exists
            instance_metadata.ami_id = ami_id
            instance_metadata.instance_id = instance_id
            instance_metadata.instance_type = instance_type
        return instance_metadata

    @requires("ViewComputer")
    def get_instance_metadata(self):
        """
        Returns a C{dict} containing all related L{CloudInstanceMetadata}.
        """
        result = self.store.find(
            CloudInstanceMetadata, CloudInstanceMetadata.computer_id == self.computer.id
        )
        metadata = result.one()
        if metadata:
            return {
                "ami-id": metadata.ami_id,
                "instance-id": metadata.instance_id,
                "instance-type": metadata.instance_type,
            }
        return {}
