# Copyright 2005-2022 Canonical Limited. All rights reserved.

from builtins import object
from zope.interface import implementer

from storm.locals import (
    Int,
    Unicode,
    DateTime,
    ReferenceSet,
    Reference,
    Float,
    AutoReload,
)

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.resource.interfaces import (
    ICustomGraphData,
    ICustomGraphDataPoint,
)


__all__ = ["CustomGraphData", "CustomGraphDataPoint"]


__requires_prefix__ = "landscape."


@implementer(ICustomGraphData)
@requires_advisor
class CustomGraphData(ComputerHelper):
    """Representation of CustomGraphs in the Resource database."""

    __storm_table__ = "custom_graph_data"

    requires("ViewCustomGraph", PRIVATE, ["id"])
    requires(
        "ViewCustomGraph", "EditCustomGraph", ["computer", "error", "custom_graph"]
    )

    _custom_graph = None

    id = Int(primary=True, allow_none=False, default=AutoReload)
    custom_graph_id = Int(allow_none=False)
    computer_id = Int(allow_none=False)
    error = Unicode()

    def __init__(self, computer=None, custom_graph=None):
        super(CustomGraphData, self).__init__(computer)
        if custom_graph:
            self.custom_graph = custom_graph

    def _get_custom_graph(self):
        if (self.custom_graph_id is not None and not self._custom_graph) or (
            self._custom_graph and self._custom_graph.id != self.custom_graph_id
        ):
            from canonical.landscape.model.account import customgraph

            store = self.computer.account.store
            self._custom_graph = store.get(
                customgraph.CustomGraph, self.custom_graph_id
            )
        return self._custom_graph

    def _set_custom_graph(self, custom_graph):
        if custom_graph is None:
            self.custom_graph_id = None
        else:
            self.custom_graph_id = custom_graph.id
        self._custom_graph = custom_graph

    custom_graph = property(_get_custom_graph, _set_custom_graph)

    @requires("ViewCustomGraph")
    def get_data_points(self, start=None, stop=None):
        """
        Returns an iterable that yields C{(timestamp, temperature)} tuples.
        """
        where = []
        if start is not None:
            where.append(CustomGraphDataPoint.timestamp >= start)
        if stop is not None:
            where.append(CustomGraphDataPoint.timestamp <= stop)
        result = self._data_points.find(*where)
        return result.values(CustomGraphDataPoint.timestamp, CustomGraphDataPoint.value)

    @requires("EditCustomGraph")
    def append_data(self, timestamp, value):
        """
        Record a data point for this graph/computer combination.
        """
        self._data_points.add(CustomGraphDataPoint(timestamp, value))

    @requires("EditCustomGraph")
    def append_data_points(self, data_points):
        """
        Record data points for this graph/computer combination.
        """
        for timestamp, value in data_points:
            self._data_points.add(CustomGraphDataPoint(timestamp, value))


@implementer(ICustomGraphDataPoint)
class CustomGraphDataPoint(object):
    """Represents a custom data point over time."""

    __storm_table__ = "custom_graph_data_point"
    __storm_primary__ = "custom_graph_data_id", "timestamp"

    custom_graph_data_id = Int(allow_none=False)
    timestamp = DateTime(allow_none=False)
    value = Float(allow_none=False)

    computer_custom_graph = Reference(custom_graph_data_id, CustomGraphData.id)

    requires("ViewCustomGraphData", PRIVATE, ["custom_graph_data_id"])
    requires(
        "ViewCustomGraphData",
        "EditCustomGraphData",
        ["timestamp", "value", "custom_graph"],
    )

    def __init__(self, timestamp, temperature):
        self.value = temperature
        self.timestamp = timestamp


CustomGraphData._data_points = ReferenceSet(
    CustomGraphData.id,
    CustomGraphDataPoint.custom_graph_data_id,
    order_by=CustomGraphDataPoint.timestamp,
)
