# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer

from storm.locals import Int, Unicode, ReferenceSet

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.main.interfaces import IAccessContextAware
from canonical.landscape.model.resource.interfaces import IGroup
from canonical.landscape.model.resource.user import User
from canonical.landscape.model.eventlog.management import EventLog


__all__ = ["Group", "PendingGroup", "UserGroupLink"]


__requires_prefix__ = "landscape."


class UserGroupLink(object):
    """
    An object for representing the N:M links between users and groups.
    """

    __storm_table__ = "computer_user_group"
    __storm_primary__ = "user_id", "group_id"
    user_id = Int(allow_none=False)
    group_id = Int(allow_none=False)


@requires_advisor
@implementer(IAccessContextAware)
class PendingGroup(object):
    """
    Landscape can create groups on client machines.  Supporting this
    is fairly easy; however, just to throw a wrench into an otherwise
    simple problem, Landscape lets administrators associate Ubuntu
    users with groups that haven't actually been created on client
    machines yet.  C{PendingGroup} exists in order to facilitate this
    behaviour.
    """

    def __init__(self, computer, name):
        super(PendingGroup, self).__init__()
        self.computer = computer
        self.name = name

    @requires("EditGroup")
    def request_add_member(self, parent, username, uid=None):
        from canonical.landscape.model.users.activities import AddGroupMemberRequest

        return parent.add(
            AddGroupMemberRequest,
            computer=self.computer,
            username=username,
            groupname=self.name,
            uid=uid,
        )

    @property
    def access_context(self):
        return self.computer.access_context


@requires_advisor
@implementer(IGroup)
class Group(ComputerHelper):
    """A representation of a UNIX group on a computer.

    @ivar name: The group's name.
    @ivar gid: The Group ID.
    """

    __storm_table__ = "computer_group"

    id = Int(primary=True, allow_none=False)
    computer_id = Int(allow_none=False)
    gid = Int(allow_none=False)
    name = Unicode(allow_none=False)

    members = ReferenceSet(
        id,
        UserGroupLink.group_id,
        UserGroupLink.user_id,
        User.id,
        order_by=User.username,
    )

    requires("ViewGroup", PRIVATE, ["id", "computer"])
    requires("ViewGroup", "EditGroup", ["gid", "name", "members"])

    def __init__(self, computer, gid, name):
        super(Group, self).__init__(computer)
        self.gid = gid
        self.name = name

    @requires("EditGroup")
    def request_add_member(self, parent, username, uid=None):
        from canonical.landscape.model.users.activities import AddGroupMemberRequest

        child = parent.add(
            AddGroupMemberRequest,
            computer=self.computer,
            username=username,
            groupname=self.name,
            uid=uid,
            gid=self.gid,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child

    @requires("EditGroup")
    def request_remove_member(self, parent, uid, username):
        """
        Create an operation which will indicate to the client that a
        member should be removed from this group.
        """
        from canonical.landscape.model.users.activities import RemoveGroupMemberRequest

        child = parent.add(
            RemoveGroupMemberRequest,
            computer=self.computer,
            username=username,
            groupname=self.name,
            uid=uid,
            gid=self.gid,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child

    @requires("EditGroup")
    def associate_member(self, user):
        """Describe a relationship between C{user} and this group.

        See L{request_add_member} to actually add members to groups.
        This method only affects model objects to reflect current
        group/user relations.
        """
        # self.members.add is read-only on proxied objects.
        self.members.add(user)

    @requires("EditGroup")
    def dissociate_member(self, user):
        """Destroy the relationship between C{user} and this group.

        See L{request_remove_member} to actually remove members from
        groups.  This method only affects model objects to reflect
        current group/user relations.
        """
        self.members.remove(user)


User.groups = ReferenceSet(
    User.id,
    UserGroupLink.user_id,
    UserGroupLink.group_id,
    Group.id,
    order_by=Group.name,
)
