# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging

from collections import defaultdict

from storm.locals import Int, Unicode, AutoReload, Like

from zope.security.proxy import removeSecurityProxy

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.computer.helper import ComputerHelper

from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)


__all__ = [
    "HardwareInfoManagement",
    "HardwareInfo",
    "HardwareInfoKey",
    "store_hardware_for_computer",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


def _stringify_value(value, units):
    if units is not None:
        int_value = int(value)
        str_value = "%s %s" % (int_value, units)
    else:
        str_value = value
        int_value = None
    return str_value, int_value


@requires_advisor
class HardwareInfoManagement(AccountHelper):
    @requires("ViewHardwareInfo")
    def get_hardware_info(self, computer):
        """Return L{HardwareInfo}s for the given computer."""
        store = removeSecurityProxy(computer).store
        return store.find(HardwareInfo, HardwareInfo.computer_id == computer.id)

    @requires("ViewHardwareInfo")
    def get_hardware_for_computer_ids(self, computer_ids):
        """
        Return a dict of list of the L{HardwareInfo} keys and values associated
        to the given C{computer_ids}.
        """
        resource_stores = get_resource_stores_for_computer_ids(computer_ids)
        result = defaultdict(list)
        for store, store_computer_ids in resource_stores.items():
            data = store.find(
                (
                    HardwareInfo.computer_id,
                    HardwareInfo.key,
                    HardwareInfo.str_value,
                    HardwareInfo.int_value,
                ),
                HardwareInfo.computer_id.is_in(store_computer_ids),
            )
            for computer_id, key, str_value, int_value in data:
                result[computer_id].append((key, str_value, int_value))
        return result

    def get_hardware_info_for_computer_ids(self, key, computer_ids):
        """
        Return a dict mapping given C{computer_ids} to a list of values of
        the specified C{HardwareInfo} key (as a tuple of string and integer
        values).

        The C{key} parameter can contain "%" to perform "like" matches.
        """
        resource_stores = get_resource_stores_for_computer_ids(computer_ids)
        result = defaultdict(list)

        if "%" in key:
            filter_condition = Like(HardwareInfo.key, key)
        else:
            filter_condition = HardwareInfo.key == key

        for store, store_computer_ids in resource_stores.items():
            data = store.find(
                (
                    HardwareInfo.computer_id,
                    HardwareInfo.str_value,
                    HardwareInfo.int_value,
                ),
                filter_condition,
                HardwareInfo.computer_id.is_in(store_computer_ids),
            )
            for computer_id, str_value, int_value in data:
                result[computer_id].append((str_value, int_value))
        return result

    @requires("EditHardwareInfo")
    def add_hardware_info(self, computer, key, value, units=None):
        """Store a new L{HardwareInfo} for the given computer."""
        str_value, int_value = _stringify_value(value, units)
        info = HardwareInfo(computer, key, str_value, int_value)
        info.store.add(info)
        key = key.replace(":", ".")
        for part in set(key.split(".")):
            info.store.add(HardwareInfoKey(info.id, part))


class HardwareInfoKey(object):
    """Hold parts for L{HardwareInfo.key} to ease search on them.

    The idea is that the query 'disk.size' should match 'usb.storage.disk.size'
    and 'pci.store.disk.size'. Instead of doing a LIKE on the end of the string
    which would perform badly, we do a query on this table on 'disk' and
    'size'.

    @ivar hardware_info_id: Link to L{HardwareInfo.id}.
    @ivar key: A part of the original key.
    """

    __storm_table__ = "hardware_info_key"
    __storm_primary__ = "hardware_info_id", "key"

    hardware_info_id = Int(allow_none=False)
    key = Unicode(allow_none=False)

    def __init__(self, hardware_info_id, key):
        self.hardware_info_id = hardware_info_id
        self.key = key


@requires_advisor
class HardwareInfo(ComputerHelper):
    __storm_table__ = "hardware_info"

    requires("ViewHardwareInfo", PRIVATE, ["id"])
    requires("ViewHardwareInfo", "EditHardwareInfo", ["computer", "key", "value"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    computer_id = Int(allow_none=False)
    key = Unicode(allow_none=False)
    str_value = Unicode(allow_none=False)
    int_value = Int()

    def __init__(self, computer, key, str_value, int_value):
        super(HardwareInfo, self).__init__(computer)
        self.key = key
        self.str_value = str_value
        self.int_value = int_value


def store_hardware_for_computer(computer, data, logger=None):
    """
    Store provided hardware information for the computer.  Previously stored
    info are removed.

    @param computer: the computer to store hardware info for.
    @param data: an iterable returning hardware info tuples.
    @param logger: a logger instance to log invalid values. If not provided,
        the default logger will be used.
    """
    max_int = 9223372036854775807

    if logger is None:
        logger = logging.getLogger()

    management = HardwareInfoManagement(computer.account)

    # Find the existing records to diff against.
    existing_ids = {
        (info.key, info.str_value, info.int_value): info.id
        for info in management.get_hardware_info(computer)
    }

    for key, value, unit in data:
        if unit is not None and int(value) > max_int:
            logger.warn(
                "Ignore too big hardware value '%s' with key "
                "'%s' for computer '%s'" % (value, key, computer.id)
            )
        else:
            # If we already have a matching record, remove it from the existing
            # list so we don't delete it later. Otherwise insert the new row.
            str_value, int_value = _stringify_value(value, unit)
            if not existing_ids.pop((key, str_value, int_value), None):
                management.add_hardware_info(computer, key, value, unit)

    # Remove any existing rows that we didn't find in the new data.
    management.get_hardware_info(computer).find(
        HardwareInfo.id.is_in(list(existing_ids.values()))
    ).remove()
