# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.interface import Interface
from zope.schema import Bool, Datetime, Float, Int, Object, TextLine

from canonical.landscape.model.computer.interfaces import IComputerAware


__all__ = [
    "ILoadAverage",
    "IMemoryInfo",
    "IMountInfo",
    "IFreeSpace",
    "IProcessor",
    "IThermalZone",
    "ITemperature",
    "IUser",
    "IGroup",
    "IActiveProcess",
    "ICustomGraphData",
    "ICustomGraphDataPoint",
    "IComputerPackageHold",
    "IComputerAptPreferencesFiles",
    "ICPUUsage",
]


class ILoadAverage(IComputerAware):

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    load_average = Float(
        title="Load average",
        description="The load average.",
        required=True,
        readonly=False,
    )


class IMemoryInfo(IComputerAware):

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    free_memory = Int(
        title="Free memory",
        description="The amount of free memory (in megabytes).",
        required=True,
        readonly=False,
    )

    free_swap = Int(
        title="Free swap",
        description="The amount of free swap (in megabytes).",
        required=True,
        readonly=False,
    )


class IMountInfo(IComputerAware):

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    device = TextLine(
        title="Device",
        description="The mount point's device.",
        required=True,
        readonly=False,
    )

    mount_point = TextLine(
        title="Mount point",
        description="The mount point's path.",
        required=True,
        readonly=True,
    )

    filesystem = TextLine(
        title="Filesystem",
        description="The mount point's filesystem type.",
        required=True,
        readonly=False,
    )

    total_space = Int(
        title="Total space",
        description="The total amount of space for the mount point.",
        required=True,
        readonly=False,
    )


class IFreeSpace(IComputerAware):

    mount_info = Object(
        title="Mount info object",
        description="The mount info object this free space value is for.",
        required=True,
        readonly=False,
        schema=IMountInfo,
    )

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    free_space = Int(
        title="Free space",
        description="The amount of free space (in megabytes).",
        required=True,
        readonly=False,
    )


class IProcessor(IComputerAware):

    processor_id = Int(
        title="Processor ID",
        description="The processor's ID.",
        required=True,
        readonly=False,
    )

    cache_size = Int(
        title="Cache size",
        description="The size of the processor's cache (in kilobytes).",
        required=True,
        readonly=False,
    )

    model = TextLine(
        title="Model",
        description="The processor's model.",
        required=True,
        readonly=False,
    )

    vendor = TextLine(
        title="Vendor",
        description="The processor's vendor.",
        required=True,
        readonly=False,
    )


class IThermalZone(IComputerAware):

    name = TextLine(
        title="Name",
        description="The thermal zone's name.",
        required=True,
        readonly=False,
    )


class ITemperature(IComputerAware):

    thermal_zone = Object(
        title="Thermal zone",
        description="The thermal zone this temperature is for.",
        required=True,
        readonly=False,
        schema=IThermalZone,
    )

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    temperature = Int(
        title="Temperature",
        description="The temperature.",
        required=True,
        readonly=False,
    )


class IUser(IComputerAware):

    uid = Int(title="UID", description="Linux UID.", required=True, readonly=False)

    enabled = Bool(
        title="Enabled",
        description="Account is enabled?",
        required=True,
        readonly=False,
    )

    name = TextLine(
        title="Name",
        description="The user's full name.",
        required=False,
        readonly=False,
    )

    username = TextLine(
        title="Username",
        description="The user's username.",
        required=True,
        readonly=False,
    )

    location = TextLine(
        title="Location",
        description="The user's location",
        required=False,
        readonly=False,
    )

    home_phone = TextLine(
        title="Home phone number",
        description="The user's home phone number",
        required=False,
        readonly=False,
    )

    work_phone = TextLine(
        title="Work phone number",
        description="The user's work phone number",
        required=False,
        readonly=False,
    )


class IGroup(IComputerAware):
    # Just for a tag to use in ZCML.
    pass


class IActiveProcess(IComputerAware):

    pid = Int(title="PID", description="Linux PID.", required=True, readonly=True)

    uid = Int(title="UID", description="Linux UID.", required=True, readonly=True)

    gid = Int(title="GID", description="Linux GID.", required=True, readonly=True)

    name = TextLine(
        title="Name", description="The process's name.", required=True, readonly=True
    )

    state = TextLine(
        title="State", description="The process's state.", required=True, readonly=True
    )

    start_time = Datetime(
        title="Start time",
        description="The time the process started.",
        required=True,
        readonly=True,
    )

    vm_size = Int(
        title="Virtual memory size",
        description="Virtual memory size of the process.",
        required=True,
        readonly=True,
    )

    start_time = Int(
        title="Sleep average",
        description="The average amount of time the process spends sleeping.",
        required=True,
        readonly=True,
    )

    def get_username():
        """Returns the username of the process uid."""

    def get_groupname():
        """Returns the group of the process gid."""

    def is_running():
        """Returns True if the process state is running."""

    def is_sleeping():
        """Returns True if the process state is sleeping."""

    def is_stopped():
        """Returns True if the process state is stopped."""

    def is_dead():
        """Returns True if the process state is dead."""

    def is_zombie():
        """Returns True if the process state is a zombie."""

    def request_end(command):
        """Queue a command to end this process.

        @param command: The command to add the operation to.
        """

    def request_kill(person):
        """Queue a command to forcefully kill this process.

        @param command: The command to add the operation to.
        """


class ICustomGraphData(IComputerAware):

    error = TextLine(
        title="Error message",
        description="Error message for failing custom graph scripts",
        required=False,
    )


class ICustomGraphDataPoint(IComputerAware):

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    data_value = Float(
        title="Data value",
        description="The value returned by the custom graph on the client.",
        required=True,
        readonly=False,
    )


class IComputerPackageHold(Interface):
    """Associate a package hold with a computer."""


class IComputerAptPreferencesFiles(IComputerAware):
    """Manage the set of APT preferences files associated with a computer."""


class ICPUUsage(IComputerAware):

    timestamp = Datetime(
        title="Time",
        description="The time the data was collected.",
        required=True,
        readonly=False,
    )

    cpu_usage = Float(
        title="CPU Usage", description="The CPU usage", required=True, readonly=False
    )
