# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime

from storm.locals import Store, Int, RawStr, DateTime, Select

from landscape.lib import bpickle

from canonical.security.protect import requires, PRIVATE, requires_advisor
from canonical.landscape.model.computer.helper import ComputerHelper


__all__ = ["MessageStore"]


__requires_prefix__ = "landscape."


@requires_advisor
class MessageStore(ComputerHelper):
    """Contains some meta data about each computer's outbound message queue."""

    __storm_table__ = "message_info"

    requires("FetchMessages", PRIVATE, ["computer"])

    computer_id = Int(primary=True, allow_none=False)
    client_expected_sequence = Int(allow_none=False, default=0)
    client_pending_offset = Int(allow_none=False, default=0)

    @requires("SendMessage")
    def add(self, message):
        m = Message()
        m.computer_id = self.computer_id
        m.data = bpickle.dumps(message)
        m.timestamp = datetime.now()
        Store.of(self).add(m)

    @requires("DeleteOldMessages")
    def delete_old_messages(self):
        if self.client_pending_offset == 0:
            return
        store = Store.of(self)
        # FIXME the list for is_in shouldn't be necessary
        results = store.find(
            Message,
            Message.id.is_in(
                [
                    Select(
                        Message.id,
                        Message.computer_id == self.computer_id,
                        order_by=Message.timestamp,
                        limit=self.client_pending_offset,
                    )
                ]
            ),
        )
        results.remove()

    @requires("FetchMessages")
    def get_pending_messages(self, max=None):
        offset = self.client_pending_offset
        store = Store.of(self)
        results = store.find(Message, computer_id=self.computer_id)
        results = results.order_by(Message.timestamp)
        if max:
            results = results[offset : offset + max]
        else:
            results = results[offset:]
        return [bpickle.loads(m.data) for m in results]

    @requires("FetchMessages")
    def count_available_messages(self):
        store = Store.of(self)
        results = store.find(Message, computer_id=self.computer_id)
        return results.count()

    @requires("FetchMessages")
    def get_sequence(self):
        return self.client_expected_sequence

    @requires("FetchMessages")
    def set_sequence(self, value):
        self.client_expected_sequence = value

    @requires("FetchMessages")
    def get_pending_offset(self):
        return self.client_pending_offset

    @requires("FetchMessages")
    def set_pending_offset(self, value):
        self.client_pending_offset = value


class Message(object):
    """Represents a single message to the client.

    This is an internal object, and shouldn't be used outside of
    this module.  This also means no security or interface
    definitions.
    """

    __storm_table__ = "message"

    id = Int(primary=True, allow_none=False)
    computer_id = Int(allow_none=False)
    data = RawStr(allow_none=False)
    timestamp = DateTime(allow_none=False)
