# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from zope.interface import implementer

from storm.locals import (
    Int,
    Unicode,
    DateTime,
    Reference,
    ReferenceSet,
    Select,
    Not,
    Storm,
)
from storm.expr import Exists

from canonical.security.protect import requires, PRIVATE, requires_advisor

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.resource.interfaces import IMountInfo, IFreeSpace
from canonical.landscape.model.resource.store import get_resource_stores

__all__ = [
    "MountInfo",
    "FreeSpace",
    "AccountMountInfos",
    "remove_unreferenced_mount_infos",
]

__clearlist__ = ["AccountMountInfos"]


__requires_prefix__ = "landscape."


@implementer(IMountInfo)
@requires_advisor
class MountInfo(ComputerHelper, Storm):

    __storm_table__ = "mount_info"

    requires("ViewMountInfo", PRIVATE, ["id", "free_spaces"])
    requires(
        "ViewMountInfo",
        "EditMountInfo",
        ["device", "mount_point", "filesystem", "total_space", "computer"],
    )

    id = Int(primary=True, allow_none=False)
    computer_id = Int(allow_none=False)
    device = Unicode(allow_none=False)
    mount_point = Unicode(allow_none=False)
    filesystem = Unicode(allow_none=False)
    total_space = Int(allow_none=False)

    def __init__(self, computer, device, mount_point, filesystem, total_space):
        super(MountInfo, self).__init__(computer)
        self.device = device
        self.mount_point = mount_point
        self.filesystem = filesystem
        self.total_space = total_space

    @requires("EditMountInfo")
    def add_free_space(self, timestamp, free_space):
        self.free_spaces.add(FreeSpace(timestamp, free_space))

    @requires("ViewMountInfo")
    def get_free_spaces(self, start=None, stop=None):
        """
        Returns an iterable that yields C{(timestamp, free-space)} tuples.
        """
        where = []
        if start is not None:
            where.append(FreeSpace.timestamp >= start)
        if stop is not None:
            where.append(FreeSpace.timestamp <= stop)
        result = self.free_spaces.find(*where)
        return result.values(FreeSpace.timestamp, FreeSpace.free_space)


@implementer(IFreeSpace)
class FreeSpace(object):

    __storm_table__ = "free_space"
    __storm_primary__ = "mount_info_id", "timestamp"

    requires(
        "ViewMountInfo",
        PRIVATE,
        ["mount_info_id", "timestamp", "free_space", "mount_info"],
    )

    mount_info_id = Int(allow_none=False)
    timestamp = DateTime(allow_none=False)
    free_space = Int(allow_none=False)

    mount_info = Reference(mount_info_id, MountInfo.id)

    def __init__(self, timestamp, free_space):
        self.timestamp = timestamp
        self.free_space = free_space

    @property
    def computer(self):
        return self.mount_info.computer


MountInfo.free_spaces = ReferenceSet(
    MountInfo.id, FreeSpace.mount_info_id, order_by=FreeSpace.timestamp
)


class AccountMountInfos(AccountHelper):
    def get_mount_info(self, computer_ids):
        """Return a set of L{MountInfo} for all given computers."""

        from canonical.landscape.model.resource.store import (
            get_resource_stores_for_computer_ids,
        )
        from canonical.landscape.model.resource.mountinfo import MountInfo
        from canonical.landscape.model.computer.twin import ComputerTwin

        per_store_computers = get_resource_stores_for_computer_ids(computer_ids)

        mount_info = set()
        for store, store_computer_ids in per_store_computers.items():
            result = store.find(
                MountInfo,
                MountInfo.computer_id.is_in(store_computer_ids),
                MountInfo.computer_id == ComputerTwin.id,
                ComputerTwin.account_id == self.account_id,
            )
            mount_info.update(result)

        return mount_info

    @requires("ViewMountInfo")
    def get_mount_points(self, computer_ids):
        """
        Return a set with the union of all mount points for the given
        computers.
        """
        return set(
            mount_info.mount_point for mount_info in self.get_mount_info(computer_ids)
        )


def remove_unreferenced_mount_infos():
    """
    This handler removes old L{MountInfo}s for all L{Computer}s in all
    L{Account}s.
    """
    # Remove MountInfo's that don't reference any FreeSpace data points.
    subselect = Select(
        "1",
        tables=[FreeSpace, MountInfo],
        where=[FreeSpace.mount_info_id == MountInfo.id],
    )
    for name, store in get_resource_stores():
        result = store.find(MountInfo, Not(Exists(subselect)))
        result.remove()
