# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Int, DateTime, Like, Bool

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.storm.variables import BytesStr

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.account.helper import AccountHelper


__all__ = [
    "NetworkDevice",
    "NetworkDeviceAPI",
    "NetworkTraffic",
    "AccountNetworkInfo",
    "INBOUND",
    "OUTBOUND",
    "sort_interfaces",
]

__clearlist__ = ["AccountNetworkInfo", "INBOUND", "OUTBOUND", "sort_interfaces"]


__requires_prefix__ = "landscape."


INBOUND = "in"
OUTBOUND = "out"


@requires_advisor
class AccountNetworkInfo(AccountHelper):
    @requires("ViewDevice")
    def get_network_interfaces(self, computer_ids, interface=None):
        """
        Return the union of the interfaces with traffic for the given
        computers.

        @param computer_ids: a list of IDs of computers from which
            interfaces should be found.

        @param interface: Optionally, a specific interface name to
            look for.  If this parameter is None, all interfaces will be
            returned.
        """
        from canonical.landscape.model.resource.store import (
            get_resource_stores_for_computer_ids,
        )
        from canonical.landscape.model.computer.twin import ComputerTwin

        per_store_computers = get_resource_stores_for_computer_ids(computer_ids)
        interfaces = set()
        where = [
            NetworkTraffic.computer_id == ComputerTwin.id,
            ComputerTwin.account_id == self.account_id,
        ]
        if interface:
            where.append(NetworkTraffic.interface == interface.encode())
        for store, store_computer_ids in list(per_store_computers.items()):
            where.append(NetworkTraffic.computer_id.is_in(store_computer_ids))
            result = store.find(NetworkTraffic.interface, *where)
            result.config(distinct=True)
            interfaces.update(r.decode() for r in result)
        return interfaces

    @requires("ViewComputer")
    def find_computers_by_ip(self, ip_address):
        """
        Find all computers with a matching IP address in the associated
        L{Account} and return Computer.id
        """
        from canonical.landscape.model.resource.store import get_resource_stores
        from canonical.landscape.model.computer.twin import ComputerTwin

        computer_ids = set()
        for name, store in get_resource_stores():
            result = store.find(
                ComputerTwin.id,
                Like(NetworkDevice.ip_address, "%s%%" % ip_address),
                NetworkDevice.computer_id == ComputerTwin.id,
                ComputerTwin.account_id == self.account_id,
            )
            result.config(distinct=True)
            computer_ids.update(result)
        return computer_ids

    @requires("ViewDevice")
    def get_devices_for_computer_ids(self, computer_ids):
        """
        Find all devices for each computer_id in the list.
        """
        from canonical.landscape.model.resource.store import get_resource_stores
        from canonical.landscape.model.computer.twin import ComputerTwin

        computers = {}
        for name, store in get_resource_stores():
            result = store.find(
                NetworkDevice,
                NetworkDevice.computer_id.is_in(computer_ids),
                ComputerTwin.account_id == self.account_id,
            )
            result.config(distinct=True)
            for device in result:
                devices = computers.get(device.computer_id, set())
                devices.add(device)
                computers[device.computer_id] = devices
        return computers

    @requires("ViewComputer")
    def find_computers_by_mac_address(self, mac_address):
        """
        Find all computers with a matching MAC address in the associated
        L{Account} and return Computer.id
        """
        from canonical.landscape.model.resource.store import get_resource_stores
        from canonical.landscape.model.computer.twin import ComputerTwin

        computer_ids = set()
        for name, store in get_resource_stores():
            result = store.find(
                ComputerTwin.id,
                Like(NetworkDevice.mac_address, "%%%s%%" % mac_address),
                NetworkDevice.computer_id == ComputerTwin.id,
                ComputerTwin.account_id == self.account_id,
            )
            result.config(distinct=True)
            computer_ids.update(result)
        return computer_ids


@requires_advisor
class NetworkDeviceAPI(ComputerHelper):
    """API to access L{NetworkDevice}s of a L{Computer}."""

    @requires("ViewDevice")
    def get_devices(self):
        """Return all the devices for the computer."""
        return self.store.find(NetworkDevice, computer_id=self.computer.id)

    @requires("ViewDevice")
    def get_device(self, interface):
        """Return the device with the given interface for the computer."""
        return self.store.find(
            NetworkDevice, computer_id=self.computer.id, interface=interface.encode()
        ).one()

    @requires("EditDevice")
    def create_device(
        self,
        interface,
        ip_address,
        mac_address,
        broadcast_address,
        netmask,
        flags,
        speed=None,
        duplex=False,
    ):
        """Create a new L{NetworkDevice} attached to the computer."""
        device = NetworkDevice(
            self.computer,
            interface,
            ip_address,
            mac_address,
            broadcast_address,
            netmask,
            flags,
            speed,
            duplex,
        )
        return self.store.add(device)

    @requires("EditDevice")
    def remove_device(self, device):
        """Remove the given L{NetworkDevice}."""
        self.store.remove(device)

    @requires("EditDevice")
    def create_traffic(self, interface, timestamp, bytes_in, bytes_out):
        """Create a new L{NetworkTraffic} for the given computer."""
        data_point = NetworkTraffic(
            self.computer, interface, timestamp, bytes_in, bytes_out
        )
        return self.store.add(data_point)

    def _get_traffic(self, interface, direction, start, stop):
        """
        Helper method returning a C{ResultSet} of L{NetworkTraffic} for the
        passed parameters.
        """
        where = [
            NetworkTraffic.interface == interface.encode(),
            NetworkTraffic.computer_id == self.computer.id,
        ]
        if start is not None:
            where.append(NetworkTraffic.timestamp >= start)
        if stop is not None:
            where.append(NetworkTraffic.timestamp <= stop)
        return self.store.find(NetworkTraffic, *where)

    @requires("ViewDevice")
    def get_traffic(self, interface, direction, start, stop):
        """
        Retrieve network traffic for the given interface between C{start} and
        C{stop}.

        @param direction: whether C{INBOUND} for inbound traffic or C{OUTBOUND}
            for outbound traffic.
        """
        if direction == INBOUND:
            field = NetworkTraffic.bytes_in
        else:
            field = NetworkTraffic.bytes_out
        result = self._get_traffic(interface, direction, start, stop)
        result.order_by(NetworkTraffic.timestamp)
        return result.values(NetworkTraffic.timestamp, field)

    @requires("ViewDevice")
    def is_interface_active(self, interface, direction, start, stop):
        """Return C{True} if there is traffic for the given interface."""
        return not self._get_traffic(interface, direction, start, stop).is_empty()


@requires_advisor
class NetworkDevice(ComputerHelper):
    """
    Hold information about network devices on a computer.
    """

    __storm_table__ = "network_device"

    id = Int(primary=True, allow_none=False)
    computer_id = Int(allow_none=False)
    interface = BytesStr(allow_none=False)
    ip_address = BytesStr(allow_none=False)
    mac_address = BytesStr(allow_none=False)
    broadcast_address = BytesStr(allow_none=False)
    netmask = BytesStr(allow_none=False)
    flags = Int(allow_none=False)
    speed = Int(allow_none=True)
    duplex = Bool(allow_none=False, default=False)
    creation_time = DateTime()

    requires(
        "ViewDevice",
        "EditDevice",
        [
            "interface",
            "ip_address",
            "mac_address",
            "broadcast_address",
            "netmask",
            "flags",
            "speed",
            "duplex",
        ],
    )

    def __init__(
        self,
        computer,
        interface,
        ip_address,
        mac_address,
        broadcast_address,
        netmask,
        flags,
        speed=None,
        duplex=False,
    ):
        super(NetworkDevice, self).__init__(computer)
        self.interface = interface
        self.ip_address = ip_address
        self.mac_address = mac_address
        self.broadcast_address = broadcast_address
        self.netmask = netmask
        self.flags = flags
        self.speed = speed
        self.duplex = duplex

    def get_state(self):
        """
        Returns a serialisable representation of the C{NetworkDevice}.
        """
        return {
            "interface": self.interface.decode(),
            "ip_address": self.ip_address.decode(),
            "mac_address": self.mac_address.decode(),
            "broadcast_address": self.broadcast_address.decode(),
            "netmask": self.netmask.decode(),
            "speed": self.speed,
            "duplex": self.duplex,
        }


@requires_advisor
class NetworkTraffic(ComputerHelper):
    """Hold information about network traffic on a computer."""

    __storm_table__ = "network_traffic"
    __storm_primary__ = "computer_id", "interface", "timestamp"

    computer_id = Int(allow_none=False)
    interface = BytesStr(allow_none=False)
    timestamp = DateTime(allow_none=False)
    bytes_in = Int(allow_none=False)
    bytes_out = Int(allow_none=False)

    requires("ViewDevice", PRIVATE, ["interface", "timestamp", "bytes_in", "bytes_out"])

    def __init__(self, computer, interface, timestamp, bytes_in, bytes_out):
        super(NetworkTraffic, self).__init__(computer)
        self.interface = interface
        self.timestamp = timestamp
        self.bytes_in = bytes_in
        self.bytes_out = bytes_out


def sort_interfaces(interfaces):
    """Returns a sorted list of interfaces.

    You should use this instead of a plain sorted() call, because
    this will place, for example, eth10 after eth9 instead of
    directly after eth1.
    """

    def split_interface_name(name):
        category = name.rstrip("0123456789")
        if category != name:
            index = int(name[len(category) :])
        else:
            # Some names, like "lo", don't have an index
            index = "-1"
        return (category, index)

    return sorted(interfaces, key=split_interface_name)
