# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import object
from datetime import timedelta
from time import time
import logging
import six

from zope.interface import implementer
from zope.security.proxy import removeSecurityProxy

from storm.locals import Int, Unicode, Store, DateTime, List
from storm.expr import Column, And, Or, BinaryOper, SQLRaw

from canonical.lib.timestamp import utcnow

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import ComputerStatus, get_computer_by_id
from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.computer.twin import ComputerTwin
from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.resource.interfaces import IComputerPackageHold
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
    get_resource_stores,
)
from canonical.landscape.model.resource.aptpreferences import ComputerAptPreferencesFile
from canonical.landscape.model.activity.resynchronize import resynchronize_computers


__all__ = [
    "get_computer_ids_with_security_upgrades",
    "ComputerPackageHold",
    "InvalidPackagesDelta",
    "get_computer_ids_with_installed_package",
    "get_computer_ids_with_preferences",
    "PackageReporterResultAPI",
    "ComputerPackages",
    "ArrayOverlap",
    "get_computer_ids_with_installed",
    "InvalidPackageUpgradesRelations",
    "update_packages_buffer",
    "clear_packages_buffer",
    "purge_stale_packages_buffers",
]


__clearlist__ = [
    "PackageReporterResultAPI",
    "ComputerPackages",
    "InvalidPackagesDelta",
    "InvalidPackageUpgradesRelations",
    "update_packages_buffer",
    "clear_packages_buffer",
]


__requires_prefix__ = "landscape."


# The maximum amount of time the packages buffer should live without having
# exchanges that update it. After this time we just purge it and send a resync
# activity, since we're not likely to get any update at that point (the
# package reporter should have run in those 6 hours and its messages come
# to the server). See purge_stale_packages_buffers below.
PACKAGES_BUFFER_TIMEOUT = timedelta(hours=6)


class InvalidPackagesDelta(Exception):
    """
    The packages delta information is inconsistent with computer packages
    state.
    """


class InvalidPackageUpgradesRelations(Exception):
    """
    The package delta information about upgrade relations between
    installed packages and available upgrades is inconsistent.
    """


@requires_advisor
class ComputerPackages(ComputerHelper):
    """
    Manage packages associated to a L{Computer} and their states.
    """

    __storm_table__ = "computer_packages"

    requires("ViewComputerPackages", PRIVATE, ["computer_id", "computer"])

    computer_id = Int(primary=True, allow_none=False)

    @requires("ViewComputerPackages")
    def get_packages(self, *states):
        """Return a tuple of lists of package ids, one for each state.

        Accepted packagestates are: "avaiable", "available_upgrades",
        "installed" and "held". At least one state must be provided.
        """
        if not states:
            raise ValueError("At least one state must be provided")
        store = Store.of(self)
        return store.execute(
            "SELECT %s FROM computer_packages WHERE computer_id = ?"
            % (", ".join(states)),
            [self.computer_id],
        ).get_one()

    @requires("EditComputerPackages")
    def update(
        self,
        available=(),
        upgrades=(),
        installed=(),
        held=(),
        autoremovable=(),
        security=(),
        not_available=(),
        not_upgrades=(),
        not_installed=(),
        not_held=(),
        not_autoremovable=(),
        not_security=(),
        reset=False,
    ):
        """
        Update packages status for the computer.

        @param available: Sequence of packages and/or package hashes and/or
            package ids of packages available for installation in the given
            system.
        @param upgrades: Sequence of packages and/or package hashes and/or
            package ids of packages available for installation in the given
            system, and that represent upgrades to installed packages.
        @param installed: Sequence of packages and/or package hashes and/or
            package ids of packages currently installed in the given system.
        @param held: Sequence of packages and/or package hashes and/or
            package ids of packages currently held in the given system.
        @param autoremovable: Sequence of packages and/or package hashes
            and/or package ids of packages currently autoremovable in the
            given system.
        @param security: Sequence of packages and/or package hashes
            and/or package ids of packages from the security pocket.
        @param not_available: Sequence of packages and/or package hashes
            and/or package ids of packages that are not available for
            installation in the given system anymore.
        @param not_upgrades: Sequence of packages and/or package hashes and/or
            package ids of packages that are not available for installation in
            the given system anymore or that do not represent an upgrade
            anymore.
        @param not_installed: Sequence of packages and/or package hashes
            and/or package ids of packages that are not installed in the given
            system anymore.
        @param not_held: Sequence of packages and/or package hashes
            and/or package ids of packages that are not held in the given
            system anymore.
        @param not_autoremovable: Sequence of packages and/or package hashes
            and/or package ids of packages that are not autoremovable in the
            given system anymore.
        @param not_security: Sequence of packages and/or package hashes
            and/or package ids of packages that are not from the security
            pocket.
        @param reset: If true, clear the database set of installed, available,
            upgrades and held packages before processing other parameters.
        """
        available = self._parse_packages(available)
        upgrades = self._parse_packages(upgrades)
        installed = self._parse_packages(installed)
        held = self._parse_packages(held)
        autoremovable = self._parse_packages(autoremovable)
        security = self._parse_packages(security)
        not_available = self._parse_packages(not_available)
        not_upgrades = self._parse_packages(not_upgrades)
        not_installed = self._parse_packages(not_installed)
        not_held = self._parse_packages(not_held)
        not_autoremovable = self._parse_packages(not_autoremovable)
        not_security = self._parse_packages(not_security)
        result = Store.of(self).execute(
            "SELECT update_computer_packages(?, ?, ?, ?, ?, ?, ?, ?, ?, ?,"
            " ?, ?, ?, ?)",
            (
                self.computer_id,
                available,
                upgrades,
                installed,
                held,
                autoremovable,
                security,
                not_available,
                not_upgrades,
                not_installed,
                not_held,
                not_autoremovable,
                not_security,
                reset,
            ),
        )
        succeded = result.get_one()[0]
        if not succeded:
            raise InvalidPackagesDelta("Package data are inconsistent")

        # Check that all packages marked as upgrades actually have an upgrade
        # relation with installed packages.  The check must be done after
        # updating package lists, since it needs all upgrades and installed
        # packages.
        all_upgrades, all_installed = self.get_packages(
            "available_upgrades", "installed"
        )
        market = PackageMarket()
        relations = market.get_upgrade_relations(all_upgrades, all_installed)
        relation_upgrades = set(relation[0] for relation in relations)
        if relation_upgrades != set(all_upgrades):
            raise InvalidPackageUpgradesRelations("Invalid/missing upgrades relations")

    @requires("ViewComputerPackages")
    def has(
        self,
        either=(),
        available=(),
        upgrades=(),
        installed=(),
        held=(),
        autoremovable=(),
    ):
        """Check if package is available and/or installed.

        @param either: Method will return True only if all given packages
            are either available or installed.
        @param available: Method will return True only if all given packages
            are available.
        @param upgrades: Method will return True only if all given
            packages are available as upgrades.
        @param installed: Method will return True only if all given packages
            are installed.
        @param held: Method will return True only if all given packages
            are held.
        @param autoremovable: Method will return True only if all
            given packages are autoremovable.
        """
        either = self._parse_packages(either)
        available = self._parse_packages(available)
        available_upgrades = self._parse_packages(upgrades)
        installed = self._parse_packages(installed)
        held = self._parse_packages(held)
        autoremovable = self._parse_packages(autoremovable)
        query = []
        parms = [self.computer_id]

        for id in either:
            query.append(" AND (%s = ANY (available) OR %s = ANY (installed))")
            parms.append(id)
            parms.append(id)

        for id in available:
            query.append(" AND %s = ANY (available)")
            parms.append(id)

        for id in available_upgrades:
            query.append(" AND %s = ANY (available_upgrades)")
            parms.append(id)

        for id in installed:
            query.append(" AND %s = ANY (installed)")
            parms.append(id)

        for id in held:
            query.append(" AND %s = ANY (held)")
            parms.append(id)

        for id in autoremovable:
            query.append(" AND %s = ANY (autoremovable)")
            parms.append(id)

        query = "".join(query)
        parms = tuple(parms)

        store = Store.of(self)
        result = store.execute(
            "SELECT 1 FROM computer_packages WHERE " "computer_id = %s " + query, parms
        )

        return bool(result.get_one())

    @requires("ViewComputer")
    def get_apt_preferences_files(self):
        """Return a C{ResultSet} of all the computer APT preferences files."""
        result = self.computer._find(ComputerAptPreferencesFile)
        return result.order_by(ComputerAptPreferencesFile.filename)

    @requires("EditComputerPackages")
    def update_apt_preferences_files(self, apt_preferences_data):
        """Update the APT preferences files associated with the L{Computer}.

        @param apt_preferences_data: A dictionary mapping APT preferences file
            names with their contents.
        @note: Any APT preferences file previously associated with the
            L{Computer} will be removed.
        """
        apt_preferences_files = self.get_apt_preferences_files()

        if not apt_preferences_data:
            apt_preferences_files.remove()
            return

        existing_filenames = set()
        for apt_preferences_file in apt_preferences_files:
            filename = apt_preferences_file.filename

            existing_filenames.add(filename)
            if filename in apt_preferences_data:
                new_contents = apt_preferences_data[filename]
                old_contents = apt_preferences_file.contents
                if new_contents != old_contents:
                    apt_preferences_file.contents = new_contents
            else:
                self.computer.store.remove(apt_preferences_file)

        new_filenames = set(apt_preferences_data.keys()) - existing_filenames
        for filename in new_filenames:
            contents = apt_preferences_data[filename]
            self.computer.store.add(
                ComputerAptPreferencesFile(self.computer, filename, contents)
            )

    def _parse_packages(self, sequence):
        """
        Return a list of L{Package}s from a sequence of (possibly mixed) ids,
        hashes or L{Package}s.
        """
        package_ids = set()
        hashes = []
        for item in sequence:
            if isinstance(item, Package):
                package_ids.add(item.id)
            elif isinstance(item, six.string_types):
                hashes.append(item)
            else:
                package_ids.add(item)
        if hashes:
            market = PackageMarket()
            package_ids.update(market.get_ids(hashes))
        return list(package_ids)


class ComputerPackagesBuffer(ComputerHelper):
    """Buffer for packages state changes on a Computer.

    When a client reports a message with packages changes that would make the
    package state inconsistent, the changes are added to the buffered state,
    until the combined state of ComputerPackages row and the
    ComputerPackagesBuffer is consistent. At that point the data in the buffer
    is flushed to the ComputerPackages entry.

    This keeps the state in ComputerPackages always consistent, so that the
    package-search service always gets a consistent state.

    See L{canonical.landscape.message.handler.package.handle_packages} for
    details.

    """

    __storm_table__ = "computer_packages_buffer"

    requires("ViewComputerPackages", PRIVATE, ["computer_id", "computer"])

    computer_id = Int(primary=True, allow_none=False)
    last_changed = DateTime(allow_none=False, default_factory=time)
    available = List(type=Int(), allow_none=False)
    not_available = List(type=Int(), allow_none=False)
    available_upgrades = List(type=Int(), allow_none=False)
    autoremovable = List(type=Int(), allow_none=False)
    security = List(type=Int(), allow_none=False)
    not_available_upgrades = List(type=Int(), allow_none=False)
    installed = List(type=Int(), allow_none=False)
    not_installed = List(type=Int(), allow_none=False)
    held = List(type=Int(), allow_none=False)
    not_held = List(type=Int(), allow_none=False)
    not_autoremovable = List(type=Int(), allow_none=False)
    not_security = List(type=Int(), allow_none=False)

    def __init__(
        self,
        computer_id,
        available,
        available_upgrades,
        installed,
        held,
        autoremovable,
        security,
        not_available,
        not_available_upgrades,
        not_installed,
        not_held,
        not_autoremovable,
        not_security,
    ):
        self.computer_id = computer_id
        self.available = available
        self.available_upgrades = available_upgrades
        self.installed = installed
        self.held = held
        self.autoremovable = autoremovable
        self.security = security
        self.not_available = not_available
        self.not_available_upgrades = not_available_upgrades
        self.not_installed = not_installed
        self.not_held = not_held
        self.not_autoremovable = not_autoremovable
        self.not_security = not_security

    def update_available(self, available, not_available):
        """Update the available and not available lists."""
        self._update_state("available", available, not_available)

    def update_installed(self, installed, not_installed):
        """Update the installed and not installed lists."""
        self._update_state("installed", installed, not_installed)

    def update_available_upgrades(self, available_upgrades, not_available_upgrades):
        """Update the available upgrade and not available upgrades lists."""
        self._update_state(
            "available_upgrades", available_upgrades, not_available_upgrades
        )

    def update_held(self, held, not_held):
        """Update the held and not held lists."""
        self._update_state("held", held, not_held)

    def update_autoremovable(self, autoremovable, not_autoremovable):
        """Update the autoremovable and not autoremovable lists."""
        self._update_state("autoremovable", autoremovable, not_autoremovable)

    def update_security(self, security):
        """Update the security and not security lists."""
        self._update_state("security", security, [])

    def _update_state(self, state, add, remove):
        """Update the given state with a new set of added and removed IDs.

        @param state: Either "available", "installed", "available_upgrades",
            or "held".
        @param add: The IDs to add to the state.
        @param remove: The IDs to remove from the state.
        """
        total_add = set(getattr(self, state) + add)
        total_remove = set(getattr(self, "not_" + state) + remove)
        setattr(self, state, total_add - total_remove)
        setattr(self, "not_" + state, total_remove - total_add)


def update_packages_buffer(
    computer,
    available,
    available_upgrades,
    installed,
    held,
    autoremovable,
    security,
    not_available,
    not_available_upgrades,
    not_installed,
    not_held,
    not_autoremovable,
    not_security,
):
    """
    Update the package changes buffer for the specified computer with the
    provided sets of changes.
    """
    packages_buffer = computer.store.find(
        ComputerPackagesBuffer, ComputerPackagesBuffer.computer_id == computer.id
    ).one()

    if packages_buffer is None:
        packages_buffer = ComputerPackagesBuffer(
            computer.id,
            available,
            available_upgrades,
            installed,
            held,
            autoremovable,
            security,
            not_available,
            not_available_upgrades,
            not_installed,
            not_held,
            not_autoremovable,
            not_security,
        )
        computer.store.add(packages_buffer)
    else:
        packages_buffer.update_available(available, not_available)
        packages_buffer.update_installed(installed, not_installed)
        packages_buffer.update_available_upgrades(
            available_upgrades, not_available_upgrades
        )
        packages_buffer.update_held(held, not_held)
        packages_buffer.update_autoremovable(autoremovable, not_autoremovable)
        packages_buffer.update_security(security)
        packages_buffer.last_changed = utcnow()

    return packages_buffer


def clear_packages_buffer(computer):
    """Clear the package changes buffer for the computer."""
    computer.store.find(
        ComputerPackagesBuffer, ComputerPackagesBuffer.computer_id == computer.id
    ).remove()


def _get_last_exchanges(computer_ids):
    """Retrieve the last time a group of computers exchanged."""
    return dict(
        get_main_store().find(
            (ComputerStatus.computer_id, ComputerStatus.last_exchange_time),
            ComputerStatus.computer_id.is_in(computer_ids),
        )
    )


def purge_stale_packages_buffers():
    """Check all L{ComputerPackagesBuffer} rows and remove stale ones.

    A buffer is considered stale if it hasn't been updated for 6 hours or
    more since the last time the computer exchanged. In that case the client
    most probably hasn't any more info to report but yet it's inconsistent,
    so we just try to resynchronize it.
    """
    last_changed_by_id = {}
    for _, store in get_resource_stores():
        result = store.find(
            (ComputerPackagesBuffer.computer_id, ComputerPackagesBuffer.last_changed),
            ComputerTwin.id == ComputerPackagesBuffer.computer_id,
        )
        last_changed_by_id.update(result)

    store = get_main_store()
    result = store.find(
        (ComputerStatus.computer_id, ComputerStatus.last_exchange_time),
        ComputerStatus.computer_id.is_in(last_changed_by_id),
    )
    last_exchange_by_id = dict(result)

    resynchronize_ids = {}
    missing = object()
    for computer_id, last_changed in last_changed_by_id.items():
        last_exchange = last_exchange_by_id.get(computer_id, missing)
        if last_exchange is missing:
            # See bug 1531551 for exegesis.
            logging.warning(
                "Missing last exchange record for computer {}".format(computer_id)
            )
            continue

        if last_changed < last_exchange - PACKAGES_BUFFER_TIMEOUT:
            computer = get_computer_by_id(computer_id)
            account = computer.account
            clear_packages_buffer(computer)
            resynchronize_ids.setdefault(account, []).append(computer.id)

    if resynchronize_ids:
        for account, computer_ids in resynchronize_ids.items():
            resynchronize_computers(account, computer_ids, scopes=["packages"])


@implementer(IComputerPackageHold)
class ComputerPackageHold(object):
    """Associate a L{PackageHold} with a L{Computer}."""

    __storm_table__ = "computer_package_hold"
    __storm_primary__ = "computer_id", "package_hold_id"

    requires("ViewComputerPackages", PRIVATE, ["computer_id", "package_hold_id"])

    computer_id = Int(allow_none=False)
    package_hold_id = Int(allow_none=False)

    def __init__(self, computer_id, package_hold_id):
        """
        @param computer_id: The L{Computer.id} to associate the hold with.
        @param package_hold_id: The L{PackageHold.id} of the hold.
        """
        self.computer_id = computer_id
        self.package_hold_id = package_hold_id


class ArrayNe(BinaryOper):

    __slots__ = ()
    oper = " <> "


class ArrayOverlap(BinaryOper):

    __slots__ = ()
    oper = " && "


def get_computer_ids_with_installed_package(computer_ids, package_id):
    """
    Return a list of L{Computer.id}s that have the given package_id installed.
    """
    store_index = get_resource_stores_for_computer_ids(computer_ids)
    package_ids = SQLRaw("'{%d}'" % package_id)
    matching_computer_ids = []
    for store, ids in store_index.items():
        result = store.find(
            ComputerPackages.computer_id,
            ArrayOverlap(package_ids, Column("installed")),
            ComputerPackages.computer_id.is_in(ids),
        )
        matching_computer_ids.extend(result)
    return matching_computer_ids


def get_computer_ids_with_installed(computer_ids):
    """
    Return a list of L{Computer.id}s that have reported installed packages.
    """
    store_index = get_resource_stores_for_computer_ids(computer_ids)
    matching_computer_ids = []
    for store, ids in store_index.items():
        result = store.find(
            ComputerPackages.computer_id,
            ArrayNe(Column("installed"), SQLRaw("'{}'")),
            ComputerPackages.computer_id.is_in(ids),
        )
        matching_computer_ids.extend(result)
    return matching_computer_ids


def get_computer_ids_with_security_upgrades(computer_ids, package_ids=None):
    """Get L{Computer.id}s that have available security upgrades.

    @param computer_ids: A sequence of L{Computer.id}s to limit checks to.
    @param package_ids: Optionally, a sequence of L{Package.id}s to limit
        checks to.  By default, all L{Package}s with associated L{USN}s are
        included in the search.
    """
    if not computer_ids:
        return []

    from canonical.landscape.model.package.market import PackageMarket
    from canonical.landscape.model.package.package import Package

    market = PackageMarket()
    if package_ids is None:
        # FIXME The number of package IDs returned by this search will grow as
        # USNs are created.  We should implement garbage collection in the
        # package database to ensure that the magnitude of this data doesn't
        # get out of hand.
        package_ids = list(market.get_usn_packages().values(Package.id))
    if not package_ids:
        return []
    package_ids = [str(id) for id in package_ids]
    package_ids = SQLRaw("'{%s}'" % (",".join(package_ids),))
    index = get_resource_stores_for_computer_ids(computer_ids)
    vulnerable_computer_ids = []
    for store, computer_ids in index.items():
        result = store.find(
            ComputerPackages.computer_id,
            ArrayOverlap(package_ids, Column("available_upgrades")),
            ComputerPackages.computer_id.is_in(computer_ids),
        )
        result.config(distinct=True)
        vulnerable_computer_ids.extend(result)
    return vulnerable_computer_ids


def get_computer_ids_with_preferences(computer_ids, like_name=None):
    """
    Get computers with apt preferences files.

    @param computer_ids: A sequence of L{Computer.id}s to limit checks to.
    @param like_name: Optionally, a configuration filename filter to check.
        By default, all preferences files of the computer_ids will be
        included in the search.
    """
    store_index = get_resource_stores_for_computer_ids(computer_ids)
    name_filters = []
    if like_name:
        name_filters.append(ComputerAptPreferencesFile.filename.like(like_name))
    matching_computer_ids = []
    for store, ids in store_index.items():
        # XXX 2022 - This isn't indexed for search.
        # As of this writing, query benchmarks on SaaS don't raise any
        # performance issue due to this table not seeing a lot of data,
        # getting updated infrequently and the alert getting checked on update.
        result = store.find(
            ComputerAptPreferencesFile.computer_id,
            ComputerAptPreferencesFile.computer_id.is_in(ids),
            *name_filters
        )
        matching_computer_ids.extend(result)
    return matching_computer_ids


class PackageReporterResult(ComputerHelper):
    """Store results reported by the package reporter."""

    __storm_table__ = "package_reporter_result"

    requires("ViewComputerPackages", PRIVATE, ["computer_id", "computer"])
    requires(
        "ViewComputerPackages", "EditComputerPackages", ["code", "err", "timestamp"]
    )

    computer_id = Int(primary=True, allow_none=False)
    code = Int(allow_none=False)
    err = Unicode(allow_none=False)
    timestamp = DateTime()

    def __init__(self, computer_id, code, err, timestamp=None):
        self.computer_id = computer_id
        self.code = code
        self.err = err
        self.timestamp = timestamp

    @requires("EditComputerPackages")
    def remove(self):
        self.store.remove(self)


class PackageReporterResultAPI(AccountHelper):
    @requires("ViewComputerPackages")
    def get_package_reporter_result(self, computer):
        """
        Return the L{PackageReporterResult} for the given computer, if it
        exists.
        """
        store = removeSecurityProxy(computer).store
        return store.find(
            PackageReporterResult, PackageReporterResult.computer_id == computer.id
        ).one()

    @requires("EditComputerPackages")
    def create_package_reporter_result(self, computer, code, err, timestamp=None):
        """
        Create a new L{PackageReporterResult} link to this computer.
        """
        store = removeSecurityProxy(computer).store
        result = PackageReporterResult(computer.id, code, err, timestamp=timestamp)
        return store.add(result)

    @requires("ViewComputerPackages")
    def get_package_reporter_errors(self, computer_ids):
        """
        Return a C{set} of (computer_id, code, error, timestamp) for the
        computers having a current package reporting issue.
        """
        data = set()

        if not computer_ids:
            return data

        store_index = get_resource_stores_for_computer_ids(computer_ids)
        for store, store_computer_ids in store_index.items():
            # Reporting is considered failing if the exit code is different
            # from 0, and different from 1 with no error output. The latter is
            # considered as false positive when the package reporter fails to
            # run because the last run is too recent.
            result = store.find(
                (
                    PackageReporterResult.computer_id,
                    PackageReporterResult.code,
                    PackageReporterResult.err,
                    PackageReporterResult.timestamp,
                ),
                PackageReporterResult.computer_id.is_in(store_computer_ids),
                And(
                    PackageReporterResult.code != 0,
                    Or(
                        PackageReporterResult.code != 1, PackageReporterResult.err != ""
                    ),
                ),
            )
            data.update(result)
        return data

    @requires("ViewComputerPackages")
    def get_package_reporter_result_count(self, computer_ids):
        """
        Return the number of computers which reporter any kind of result to the
        server.
        """
        count = 0
        if not computer_ids:
            return count

        store_index = get_resource_stores_for_computer_ids(computer_ids)
        for store, store_computer_ids in store_index.items():
            result = store.find(
                PackageReporterResult,
                PackageReporterResult.computer_id.is_in(store_computer_ids),
            )
            count += result.count()
        return count
