# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from storm.locals import Int, Unicode, List

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)


__all__ = [
    "RebootRequiredInfo",
    "RebootRequiredAPI",
    "get_reboot_required_info_for_computers",
]

__clearlist__ = ["get_reboot_required_info_for_computers"]

__requires_prefix__ = "landscape."


class RebootRequiredInfo(ComputerHelper):
    """Information about a L{Computer} requiring a reboot.

    @param computer: The L{Computer} the information refers to.
    @param packages: The list of packages that requested the reboot.
    """

    __storm_table__ = "reboot_required_info"

    computer_id = Int(primary=True, allow_none=False)
    packages = List(type=Unicode())

    def __init__(self, computer, packages=None):
        super(RebootRequiredInfo, self).__init__(computer)
        self.packages = packages


class RebootRequiredAPI(ComputerHelper):
    def get_reboot_required_info(self):
        """
        Return the L{RebootRequiredInfo} associated with the given L{Computer}.
        """
        return self.store.get(RebootRequiredInfo, self.computer.id)

    def create_reboot_required_info(self, packages=None):
        """Create a L{RebootRequiredInfo} for the given L{Computer}.

        @param packages: A list of package names for the created object.
        """
        reboot_required_info = RebootRequiredInfo(self.computer, packages)
        self.store.add(reboot_required_info)
        return reboot_required_info

    def remove_reboot_required_info(self):
        """Remove the L{RebootRequiredInfo} of the given L{Computer}."""
        reboot_required_info = self.get_reboot_required_info()
        if reboot_required_info:
            self.store.remove(reboot_required_info)


def get_reboot_required_info_for_computers(computers):
    """
    Return the L{RebootRequiredInfo} information for the given L{Computer}s.
    """
    computer_ids = [computer.id for computer in computers]
    computer_ids_by_store = get_resource_stores_for_computer_ids(computer_ids)
    reboot_required_infos = []

    for store, store_computer_ids in computer_ids_by_store.items():
        result = store.find(
            RebootRequiredInfo, RebootRequiredInfo.computer_id.is_in(store_computer_ids)
        )
        reboot_required_infos.extend(list(result))

    return reboot_required_infos
