# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from datetime import datetime

from canonical.event.subscriber import subscriber
from canonical.landscape.environment import get_raw_config
from canonical.landscape.model.events import (
    SchemaMaintenanceEvent,
    ResynchronizeEvent,
    AddComputerEvent,
)
from canonical.landscape.model.resource.loadaverage import LoadAverage
from canonical.landscape.model.resource.memoryinfo import MemoryInfo
from canonical.landscape.model.resource.network import NetworkTraffic
from canonical.landscape.model.resource.temperature import Temperature
from canonical.landscape.model.resource.mountinfo import (
    FreeSpace,
    remove_unreferenced_mount_infos,
)
from canonical.landscape.model.resource.customgraph import CustomGraphDataPoint
from canonical.landscape.model.resource.store import get_resource_stores
from canonical.landscape.schema.partitioner import TablePartitioner
from canonical.landscape.model.resource.network import NetworkDevice
from canonical.landscape.model.resource.package import ComputerPackages
from canonical.landscape.model.resource.cpuusage import CPUUsage
from canonical.landscape.model.resource.ceph import CephUsage
from canonical.landscape.model.resource.swift import SwiftUsage
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.global_maintenance import global_maintenance
from canonical.landscape.model.usn.issue import USNIssue


def maintain_model_table(model):
    """
    maintain_model_table takes care of the maintenance on a table that is
    partitioned by date.  It removes old partition tables and creates new
    partition tables as needed.

    This method maintains 28 days of historical data and creates tables for 7
    days in the future.  To prevent errors, this method should be called more
    often than every 7 days.

    @param model - The model with a partitioned table
    """
    check_field = "timestamp"
    days_retention = 28
    days_preallocated = 7
    date = datetime.utcnow().date()

    raw_config = get_raw_config()
    grant_access_user = raw_config.get("stores", "user")
    for name, store in get_resource_stores():
        partitioner = TablePartitioner(
            model,
            check_field,
            date,
            days_retention,
            days_preallocated,
            store,
            grant_access_user,
        )
        partitioner.update()
        store.commit()


@subscriber(SchemaMaintenanceEvent)
def load_average_maintenance(event):
    """This handler removes old L{LoadAverage}s."""
    maintain_model_table(LoadAverage)


@subscriber(SchemaMaintenanceEvent)
def memory_info_maintenance(event):
    """This handler removes old L{MemoryInfo}s."""
    maintain_model_table(MemoryInfo)


@subscriber(SchemaMaintenanceEvent)
def network_traffic_maintenance(event):
    """Remove old L{NetworkTraffic} data."""
    maintain_model_table(NetworkTraffic)


@subscriber(SchemaMaintenanceEvent)
def temperature_maintenance(event):
    """This handler removes old L{Temperature}s.

    We don't touch L{ThermalZone}s because these are unlikely to
    change often enough to justify removal.
    """
    maintain_model_table(Temperature)


@subscriber(SchemaMaintenanceEvent)
def mount_info_maintenance(event):
    """
    This handler removes old L{FreeSpace} data.
    """
    maintain_model_table(FreeSpace)


@subscriber(SchemaMaintenanceEvent)
def custom_graph_maintenance(event):
    """This handler removes old L{CustomGraphDataPoint}s."""
    maintain_model_table(CustomGraphDataPoint)


@subscriber(SchemaMaintenanceEvent)
def cpu_usage_maintenance(event):
    """This handler removes old L{CPUUsage}s."""
    maintain_model_table(CPUUsage)


@subscriber(SchemaMaintenanceEvent)
def ceph_usage_maintenance(event):
    """This handler removes old L{CephUsage}s."""
    maintain_model_table(CephUsage)


@subscriber(SchemaMaintenanceEvent)
def swift_usage_maintenance(event):
    """This handler removes old L{SwiftUsage}s."""
    maintain_model_table(SwiftUsage)


@global_maintenance
def maintain_unreferenced_mount_infos(event):
    remove_unreferenced_mount_infos()


@subscriber(ResynchronizeEvent)
def delete_groups(resynchronize):
    """
    Delete all groups in a computer in response to a
    L{ResynchronizeEvent}.
    """
    if not resynchronize.is_in_scope("users"):
        return
    resynchronize.computer.remove_groups()


@subscriber(IAccount, AddComputerEvent)
def add_computer_packages(account, event):
    """
    Relate packages to a computer after it is created (in response to
    the L{AddComputerEvent}).
    """
    event.computer.store.add(ComputerPackages(event.computer))


@subscriber(ResynchronizeEvent)
def remove_network_devices(resynchronize):
    """
    Remove all L{NetworkDevice} in response to a L{ResynchronizeEvent}
    .
    """
    if not resynchronize.is_in_scope("network"):
        return
    computer = resynchronize.computer
    result = computer.store.find(
        NetworkDevice, NetworkDevice.computer_id == computer.id
    )
    result.remove()


@subscriber(ResynchronizeEvent)
def delete_users(resynchronize):
    """
    Delete all users in a computer in response to a
    L{ResynchronizeEvent}.
    """
    if not resynchronize.is_in_scope("users"):
        return
    resynchronize.computer.remove_users()


@subscriber(ResynchronizeEvent)
def delete_usn_issues(resynchronize):
    """
    Remove all open L{USNIssue} for the resync computer.
    """
    if not resynchronize.is_in_scope("package"):
        return
    computer = resynchronize.computer
    store = computer.store
    store.find(
        USNIssue, USNIssue.resolve_time == None, USNIssue.computer_id == computer.id
    ).remove()
