# Copyright 2005-2022 Canonical Limited. All rights reserved.

from builtins import object
from zope.interface import implementer

from storm.locals import Int, Unicode, DateTime, Reference, ReferenceSet

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.resource.interfaces import IThermalZone, ITemperature
from canonical.landscape.model.account.helper import AccountHelper


__all__ = ["ThermalZone", "Temperature", "AccountThermalZones"]

__clearlist__ = ["AccountThermalZones"]


__requires_prefix__ = "landscape."


@implementer(IThermalZone)
@requires_advisor
class ThermalZone(ComputerHelper):
    """Thermal zones report temperature data."""

    __storm_table__ = "thermal_zone"

    requires("ViewThermalZone", PRIVATE, ["id"])
    requires("ViewThermalZone", "EditThermalZone", ["computer", "name"])

    id = Int(primary=True, allow_none=False)
    computer_id = Int(allow_none=False)
    name = Unicode(allow_none=False)

    def __init__(self, computer, name):
        super(ThermalZone, self).__init__(computer)
        self.name = name

    @requires("EditThermalZone")
    def add_temperature(self, timestamp, temperature):
        self._temperatures.add(Temperature(timestamp, temperature))

    @requires("ViewThermalZone")
    def get_temperatures(self, start=None, stop=None):
        """
        Returns an iterable that yields C{(timestamp, temperature)} tuples.
        """
        where = []
        if start is not None:
            where.append(Temperature.timestamp >= start)
        if stop is not None:
            where.append(Temperature.timestamp <= stop)
        result = self._temperatures.find(*where)
        return result.values(Temperature.timestamp, Temperature.temperature)


@implementer(ITemperature)
@requires_advisor
class Temperature(object):
    """Represents a temperature over time."""

    __storm_table__ = "temperature"
    __storm_primary__ = "thermal_zone_id", "timestamp"

    thermal_zone_id = Int(allow_none=False)
    timestamp = DateTime(allow_none=False)
    temperature = Int(allow_none=False)

    thermal_zone = Reference(thermal_zone_id, ThermalZone.id)

    requires("ViewThermalZone", PRIVATE, ["thermal_zone_id"])
    requires(
        "ViewThermalZone",
        "EditThermalZone",
        ["timestamp", "temperature", "thermal_zone"],
    )

    def __init__(self, timestamp, temperature):
        self.temperature = temperature
        self.timestamp = timestamp

    @property
    def computer(self):
        return self.thermal_zone.computer


ThermalZone._temperatures = ReferenceSet(
    ThermalZone.id, Temperature.thermal_zone_id, order_by=Temperature.timestamp
)


class AccountThermalZones(AccountHelper):
    @requires("ViewThermalZone")
    def get_thermal_zones(self, computer_ids):
        """
        Return a set with the union of all mount points for the given
        computers.
        """
        from canonical.landscape.model.resource.store import (
            get_resource_stores_for_computer_ids,
        )
        from canonical.landscape.model.resource.temperature import ThermalZone
        from canonical.landscape.model.computer.twin import ComputerTwin

        resource_stores = get_resource_stores_for_computer_ids(computer_ids)
        thermal_zones = set()
        for store, store_computer_ids in resource_stores.items():
            result = store.find(
                ThermalZone,
                ThermalZone.computer_id.is_in(store_computer_ids),
                ThermalZone.computer_id == ComputerTwin.id,
                ComputerTwin.account_id == self.account_id,
            )
            result.config(distinct=True)
            thermal_zones.update(result.values(ThermalZone.name))
        return thermal_zones
