from __future__ import division

# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import zip
from builtins import range
from builtins import object
from past.utils import old_div
from collections import defaultdict

from datetime import timedelta, datetime

from storm.expr import Func, Cast, Alias, Count, Avg, SQLRaw, Select, And, Column

from canonical.lib.timestamp import utctoday
from canonical.lib.mathutils import weighted_average
from canonical.lib.forecast import forecast

from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.resource.ceph import CephUsage
from canonical.landscape.model.resource.swift import SwiftUsage
from canonical.landscape.model.resource.cpuusage import CPUUsage
from canonical.landscape.model.resource.memoryinfo import MemoryInfo


__all__ = ["ComputeResourceUsage", "StorageResourceUsage"]


class ResourceUsage(object):
    """
    Base class for classes performing resource usage computations on a computer
    selection.

    @param selection: The L{ComputerSelection} to compute resource usage for.
    """

    def __init__(self, selection, utctoday=utctoday):
        self.selection = selection
        self.utctoday = utctoday  # overridable for testing purposes

    def _get_value_interval_expr(self, field, intervals):
        """
        Return a Storm expression giving the index of the interval a field
        belongs to, among those provided.

        @param field: the name of the database field for the expression
        @param intervals: a list of 2-tuples with min/max for each interval;
            values should not overlap
        @return: a Storm expression

        """
        expr = ["CASE"]
        for i, (start, stop) in enumerate(intervals):
            expr.append("WHEN %s BETWEEN %s AND %s THEN %s" % (field, start, stop, i))
        expr.append("END")
        return SQLRaw(" ".join(expr))

    def _get_time_interval_index_expr(self, resource_class, name=None):
        """
        Return a Storm expression calculating the number of days back a data
        point row belongs to.

        @param resource_class: The resource Storm model to calculate the
            expression for.
        @param name: the name of the C{Alias} expression, autogenerated if
            C{None}.
        @return: a Storm expression.

        """
        delta = Cast(self.utctoday(), "date") - Cast(resource_class.timestamp, "date")
        expr = Cast(Func("trunc", delta), "integer")
        return Alias(expr) if name is None else Alias(expr, name=name)

    def _get_cutoff_date(self, days):
        """
        Return a date representing the oldest date to filter resource data.

        @param days: The number of days of data to consider.
        @return: The cutoff date.
        """
        today = self.utctoday()
        now = datetime(today.year, today.month, today.day)
        cutoff_date = now - timedelta(days=days - 1)
        return cutoff_date

    def _get_usages(self, value_tuples):
        """
        Return a list of usage in the (0.0, 1.0) range from a list of
        (total, used) tuples.

        @param value_tuples: a list of 2-tuples with (total, used)
        """
        usages = []
        for total, used in value_tuples:
            usages.append(None if total is None else old_div(used, total))
        return usages

    def _get_forecast(self, daily_usages):
        """
        Return a forecast of the number of days in which a resource will max
        out, based on the provided daily usages.

        @param daily_usages: a list of daily usage values (in the 0.0-1.0
            range), newest first
        @return: integer number of days until resource outage, or L{None} if
            the forecast can't be computed.
        """
        # Create a series for the forecast function, as a list of 2-tuples with
        # (timestamp, value), the timestamp is just an integer startging from 0
        # for days, so the forecast unit is number of days.
        #
        # If a day doesn't have a value, it's not included in the series (but
        # timestamp is increased).
        series = [
            value for value in enumerate(reversed(daily_usages)) if value[1] is not None
        ]
        if len(series) < 2:
            # not enough data for forecast
            return None
        days = forecast(series, 1.0)
        if days is not None:
            return int(days)

    def _compute_averages(self, interval_usages):
        """
        Return weighted average for values in each interval.

        @param interval_usages: a list of lists with usages (as a 2-tuple with
            value and weight).
        @return: a list of average values, one for each interval. Average is
            L{None} if the interval has no values.
        """
        averages = []
        for usages in interval_usages:
            if usages:
                averages.append(weighted_average(usages))
            else:
                averages.append(None)
        return averages


class ComputeResourceUsage(ResourceUsage):
    """
    Compute aggregated historical usage and outage forecast for "compute"
    resources (such as RAM and CPU), for a computer selection.
    """

    def get_memory_usage_and_forecast(self, days):
        """
        Return a list of average (total, used) memory on a daily basis for the
        specified number of days, and a forecast of how many days are left
        before resource outage.

        @param days: The number of days of data to consider.
        @return: A 2-tuple with usage and forecast.
        """
        usage = self._get_memory_usage(days)
        return usage, self._get_forecast(self._get_usages(usage))

    def get_cpu_usage_map_and_forecast(self, days, usage_intervals):
        """
        Return a list of lists of counts of computers with CPU usage in each
        interval, on a daily basis for the specified number of days and a
        forecast of how many days are left before resource outage.

        @param days: The number of days of data to consider.
        @param usage_intervals: a list of 2-tuples with min/max values for
            each usage interval. Values must be sorted, in the range 0-100 and
            intervals must not overlap.
        @return: a 2-tuple with a list of lists list of counts in each
            usage/time interval and forecast
        """
        usages = self._get_cpu_usage(days)
        usage_map = self._get_cpu_usage_map(usage_intervals, days)
        return usage_map, self._get_forecast(usages)

    def _get_memory_usage(self, days):
        """
        Return average memory usage in each day, up to the specified number of
        days.

        @param days: The number of days of data to consider.
        @return: a list with average memory usage for each time interval
        """
        selected_computer_ids = self.selection.get_selected_computer_ids()

        result = get_main_store().find(
            (Computer.id, Computer.total_memory),
            Computer.total_memory != None,
            Computer.id.is_in(selected_computer_ids),
        )
        computer_total_memory = dict(result)
        total_memory = sum(computer_total_memory.values())

        # List of total and used memory in each day.
        interval_totals = [None] * days
        interval_used = [None] * days
        if total_memory:
            cutoff_date = self._get_cutoff_date(days)
            # Only include computers that have total_memory set.
            resource_stores = get_resource_stores_for_computer_ids(
                list(computer_total_memory.keys())
            )
            time_interval = self._get_time_interval_index_expr(MemoryInfo)

            for store, computer_ids in resource_stores.items():
                result = store.find(
                    (
                        time_interval,
                        MemoryInfo.computer_id,
                        Cast(Avg(MemoryInfo.free_memory), "float"),
                    ),
                    MemoryInfo.computer_id.is_in(computer_ids),
                    MemoryInfo.timestamp >= cutoff_date,
                )
                result.group_by(MemoryInfo.computer_id, time_interval)

                for interval, computer_id, free in result:
                    total = computer_total_memory[computer_id]
                    used = total - free
                    if interval_totals[interval] is None:
                        interval_totals[interval] = float(total)
                        interval_used[interval] = used
                    else:
                        interval_totals[interval] += total
                        interval_used[interval] += used

        return list(zip(interval_totals, interval_used))

    def _get_cpu_usage(self, days):
        """
        Return average CPU usage in each day, up to the specified number of
        days.

        @param days: The number of days of data to consider.
        @return: a list with average memory usage for each day

        """
        cutoff_date = self._get_cutoff_date(days)
        selected_computer_ids = self.selection.get_selected_computer_ids()
        resource_stores = get_resource_stores_for_computer_ids(selected_computer_ids)
        time_interval = self._get_time_interval_index_expr(CPUUsage)

        interval_usages = [[] for _ in range(days)]
        for store, computer_ids in resource_stores.items():

            result = store.find(
                (time_interval, Avg(CPUUsage.cpu_usage), Count()),
                CPUUsage.computer_id.is_in(computer_ids),
                CPUUsage.timestamp >= cutoff_date,
            )
            result.group_by(time_interval)

            for interval, used, count in result:
                interval_usages[interval].append((used, count))

        return self._compute_averages(interval_usages)

    def _get_cpu_usage_map(self, usage_intervals, days):
        """
        Return counts of computers in each usage interval, in each day, up to
        the specified number of days.

        @param usage_intervals: a list of 2-tuples with min/max values for
            each usage interval. Values must be sorted, in the range 0-100 and
            intervals must not overlap.
        @param days: The number of days of data to consider.
        @return: a list of lists of number of computers in each usage interval,
            for each time interval.
        """
        cutoff_date = self._get_cutoff_date(days)
        selected_computer_ids = self.selection.get_selected_computer_ids()
        resource_stores = get_resource_stores_for_computer_ids(selected_computer_ids)

        interval_usages = [[0] * len(usage_intervals) for _ in range(days)]

        for store, computer_ids in resource_stores.items():
            time_interval = self._get_time_interval_index_expr(
                CPUUsage, name="time_interval"
            )
            average_usage = Alias(
                Func("trunc", Avg(CPUUsage.cpu_usage) * 100), "average_usage"
            )
            usage_interval = Alias(
                self._get_value_interval_expr("average_usage", usage_intervals),
                "usage_interval",
            )

            # Get average usage for each computer in each time interval
            subquery = Select(
                (time_interval, CPUUsage.computer_id, average_usage),
                where=And(
                    CPUUsage.computer_id.is_in(computer_ids),
                    CPUUsage.timestamp >= cutoff_date,
                ),
                group_by=(time_interval, CPUUsage.computer_id),
            )
            # Get computers count for each time and value interval
            query = Select(
                (Column("time_interval"), usage_interval, Count(Column("computer_id"))),
                tables=[Alias(subquery)],
                group_by=(Column("time_interval"), usage_interval),
                order_by=(Column("time_interval"), usage_interval),
            )
            result = store.execute(query)

            for time_interval, usage_interval, count in result:
                interval_usages[time_interval][usage_interval] += count

        return interval_usages


class StorageResourceUsage(ResourceUsage):
    """
    Compute aggregated  historical usage and outage forecast for "storage"
    resources (such as Ceph and Swift), for a computer selection.
    """

    def get_ceph_usage_and_forecast(self, days):
        """
        For each Ceph ring, return a list of average (total, used) space on a
        daily basis, and a forecast of how many days are left before resource
        outage.

        @param days: The number of days of data to consider.
        @return: A 2-tuple with usage and forecast for each ring uuid.

        """
        usage = self._get_ceph_usage(days)
        forecasts = {
            ring_uuid: self._get_forecast(self._get_usages(values))
            for ring_uuid, values in usage.items()
        }
        return usage, forecasts

    def get_swift_usage_and_forecast(self, days):
        """
        Return a list of average (total, used) space for Swift on a daily
        basis, and a forecast of how many days are left before resource
        outage.

        @param days: The number of days of data to consider.
        @return: A 2-tuple with usage and forecast.
        """
        usage = self._get_swift_usage(days)
        return usage, self._get_forecast(self._get_usages(usage))

    def _get_ceph_usage(self, days):
        """Return storage information in the Ceph rings for each day.

        Return a dict mapping Ceph ring UUIDs for selected computers to a list
        of 2-tuples with total and used storage space on each day.

        @param days: The number of days of data to consider.
        @return: a dicts mapping Ceph ring UUIDs to a list of 2-tuples with
            average (total, used) space in bytes.
        """
        cutoff_date = self._get_cutoff_date(days)
        selected_computer_ids = self.selection.get_selected_computer_ids()
        resource_stores = get_resource_stores_for_computer_ids(selected_computer_ids)

        time_interval = self._get_time_interval_index_expr(CephUsage)

        # Map ring uuids to totals and used for each day.  The query groups
        # by (ring_uuid, interval), so it will return a single value for each
        # one, but each store can contribute a (count, total, used) for the
        # (ring_uuid, interval), so we need to keep a list and average on it.
        interval_totals = defaultdict(lambda: [[] for _ in range(days)])
        interval_used = defaultdict(lambda: [[] for _ in range(days)])
        for store, computer_ids in resource_stores.items():
            result = store.find(
                (
                    CephUsage.ring_uuid,
                    time_interval,
                    Count(),
                    Cast(Avg(CephUsage.total), "float"),
                    Cast(Avg(CephUsage.used), "float"),
                ),
                CephUsage.computer_id.is_in(computer_ids),
                CephUsage.timestamp >= cutoff_date,
            )
            result.group_by(CephUsage.ring_uuid, time_interval)

            for ring_uuid, interval, count, total, used in result:
                interval_totals[ring_uuid][interval].append((total, count))
                interval_used[ring_uuid][interval].append((used, count))

        usage = {}
        for ring_uuid in interval_totals:
            totals = self._compute_averages(interval_totals[ring_uuid])
            used = self._compute_averages(interval_used[ring_uuid])
            # List of (total, used) for each day
            usage[ring_uuid] = list(zip(totals, used))

        return usage

    def _get_swift_usage(self, days):
        """Return storage information in the Swift cluster for each day.

        Return list of 2-tuples with total and used storage space on each day.

        @param days: The number of days of data to consider.
        @return: a list  of 2-tuples with average (total, used) space in bytes.
        """
        cutoff_date = self._get_cutoff_date(days)
        selected_computer_ids = self.selection.get_selected_computer_ids()
        resource_stores = get_resource_stores_for_computer_ids(selected_computer_ids)

        time_interval = self._get_time_interval_index_expr(SwiftUsage)

        # List of total and used space for each device in each computer in a
        # time interval.
        interval_totals = [None] * days
        interval_used = [None] * days
        for store, computer_ids in resource_stores.items():
            result = store.find(
                (
                    time_interval,
                    Cast(Avg(SwiftUsage.total), "float"),
                    Cast(Avg(SwiftUsage.used), "float"),
                ),
                SwiftUsage.computer_id.is_in(computer_ids),
                SwiftUsage.timestamp >= cutoff_date,
            )
            result.group_by(SwiftUsage.computer_id, SwiftUsage.device, time_interval)

            # In each interval, sum up total and used space in each device
            for interval, total, used in result:
                if interval_totals[interval] is None:
                    interval_totals[interval] = total
                    interval_used[interval] = used
                else:
                    interval_totals[interval] += total
                    interval_used[interval] += used

        return list(zip(interval_totals, interval_used))
