# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer

from storm.locals import Int, Bool, Unicode

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.computer.helper import ComputerHelper
from canonical.landscape.model.resource.interfaces import IUser
from canonical.landscape.model.eventlog.management import EventLog


__all__ = ["User"]


__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IUser)
class User(ComputerHelper):

    __storm_table__ = "computer_user"

    id = Int(primary=True, allow_none=False)
    computer_id = Int(allow_none=False)
    uid = Int(allow_none=False)
    primary_gid = Int()
    username = Unicode(allow_none=False)
    enabled = Bool(allow_none=False)
    name = Unicode()
    location = Unicode()
    home_phone = Unicode()
    work_phone = Unicode()

    requires("ViewUser", PRIVATE, ["id", "computer", "uid"])
    requires(
        "ViewUser",
        "EditUser",
        [
            "username",
            "name",
            "enabled",
            "location",
            "home_phone",
            "work_phone",
            "primary_gid",
        ],
    )

    def __init__(
        self,
        computer,
        uid,
        username,
        primary_gid,
        name,
        enabled,
        location=None,
        home_phone=None,
        work_phone=None,
    ):
        super(User, self).__init__(computer)
        self.uid = uid
        self.username = username
        self.primary_gid = primary_gid
        self.name = name
        self.enabled = enabled
        self.location = location
        self.home_phone = home_phone
        self.work_phone = work_phone

    @requires("ViewUser")
    def get_primary_group(self):
        return self.computer.get_group_by_gid(self.primary_gid)

    @requires("EditUser")
    def request_lock_user(self, parent):
        from canonical.landscape.model.users.activities import LockUserRequest

        child = parent.add(
            LockUserRequest,
            computer=self.computer,
            uid=self.uid,
            username=self.username,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child

    @requires("EditUser")
    def request_unlock_user(self, parent):
        from canonical.landscape.model.users.activities import UnlockUserRequest

        child = parent.add(
            UnlockUserRequest,
            computer=self.computer,
            uid=self.uid,
            username=self.username,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child

    @requires("RemoveUser")
    def request_delete_user(self, parent, delete_home):
        from canonical.landscape.model.users.activities import DeleteUserRequest

        child = parent.add(
            DeleteUserRequest,
            computer=self.computer,
            uid=self.uid,
            username=self.username,
            delete_home=delete_home,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child

    @requires("EditUser")
    def request_edit_user(
        self,
        parent,
        name=None,
        primary_groupname=None,
        passphrase=None,
        location=None,
        home_phone=None,
        work_phone=None,
    ):
        from canonical.landscape.model.users.activities import EditUserRequest

        child = parent.add(
            EditUserRequest,
            computer=self.computer,
            username=self.username,
            name=name,
            uid=self.uid,
            primary_groupname=primary_groupname,
            passphrase=passphrase,
            location=location,
            home_phone=home_phone,
            work_phone=work_phone,
        )
        event_log = EventLog(parent.account, parent.creator)
        event_log.create_log_entry(
            self.computer, "Activity created: " + child.info.get_summary()
        )
        return child
