import ast
import logging

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.internet.task import deferLater

from storm.twisted.transact import transact

from canonical.amqp.job.runner import JobRunner, job

from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.model.activity.activity import get_child_activity


@job
class SandboxHelloJobRunner(JobRunner):
    """Hello world job that just returns a string."""

    def run(self, delay=0):
        """
        @param delay: How many seconds the job should take to complete.
        """
        deferred = Deferred()

        def finish():
            logging.info("Sandbox hello job")
            deferred.callback(("Hello :)", None))

        deferLater(reactor, int(delay), finish)
        return deferred


@job
class SandboxBoomJobRunner(JobRunner):
    """Failing job that raises an exception."""

    def run(self):
        logging.info("Sandbox boom job")
        raise Exception("Boom!")


@job
class SandboxExpressionJobRunner(JobRunner):
    """Validate a Python expression."""

    def run(self, expression):
        logging.info("Sandbox expression job for %s" % expression)
        try:
            ast.parse(expression).compile()
        except SyntaxError:
            error = "Invalid syntax"
        else:
            error = None
        return ({"error": error}, None)


@job
class SandboxSingleActivityJobRunner(JobRunner):
    """Simulate the execution of an activity."""

    def run(self, activity_id):
        logging.info("Sandbox activity %d" % activity_id)
        simulator = _ActivitySimulator(self.transactor, activity_id)
        simulator.execute()
        return (None, None)


class _ActivitySimulator(object):
    """Simulate the execution of an activity.

    The activity will transition from QUEUED to SUCCEEDED and its progress
    will advance linearly.
    """

    def __init__(self, transactor, activity_id):
        self.transactor = transactor
        self.activity_id = activity_id

    @inlineCallbacks
    def execute(self, duration=1, delay=1):
        """Execute the activity.

        @param duration: Number of seconds the execution should last.
        @param delay: Delay the execution by this number of seconds.
        """
        yield self._defer(delay, self._start)
        interval = float(duration) / 10
        for i in range(0, 10):
            yield self._defer(interval, self._advance, (i + 1) * 10)
        yield self._succeed()

    def _defer(self, seconds, function, *args):
        """Execute a function in the future."""
        return deferLater(reactor, seconds, function, *args)

    @transact
    def _start(self):
        activity = self._get_activity()
        activity.start()

    @transact
    def _advance(self, progress):
        activity = self._get_activity()
        activity.advance(progress)

    @transact
    def _succeed(self):
        activity = self._get_activity()
        activity.succeed()

    def _get_activity(self):
        account = get_account_by_name("sandbox")
        return get_child_activity(account, int(self.activity_id))
