# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.lib.text import describe_dict
from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.script.errors import UnknownScriptError
from canonical.landscape.model.script.script import Script, DuplicateScriptError


__all__ = ["ScriptManagement"]

__clearlist__ = __all__


__requires_prefix__ = "landscape."


@requires_advisor
class ScriptManagement(ModelManagement):

    requires("ViewAccount", PRIVATE, ["account", "person"])

    @requires("ViewAccount")
    def create_script(
        self,
        title,
        code,
        time_limit,
        username=None,
        attachments=None,
        access_context_name=DEFAULT_ACCESS_CONTEXT_NAME,
    ):
        """
        Create a stored script in the account.

        @param attachments: if specified, dictionary of attachments in the form
            filename: content.
        @type attachments: C{dict}
        """
        access_context = self.get_access_context(access_context_name, "CreateScript")

        self.check_duplicate(Script, DuplicateScriptError, title=title)

        script = Script(
            self.account,
            self.person,
            code,
            title,
            time_limit,
            username,
            access_context=access_context,
        )
        if attachments is not None:
            for filename, data in attachments.items():
                script.add_attachment(filename, data)
                self._event_log.create_log_entry(
                    script, "Attachment '%s' added." % filename
                )

        details = {
            "creator": self.person,
            "interpreter": script.interpreter,
            "title": title,
            "time_limit": time_limit,
            "username": username,
            "creator": self.person.name,
        }
        self._event_log.create_log_entry(script, describe_dict(items=details))
        return script

    @requires("ViewAccount")
    def edit_script(self, script_id, data):
        """
        Edit a stored script in the account.

        @type attachments: C{dict}
        """
        db_script = self.get_script(script_id)
        changes = {}

        title = data.get("title", None)
        if title is not None and title != db_script.title:
            self.check_duplicate(Script, DuplicateScriptError, title=title)
            db_script.title = title
            changes.update({"title": title})

        code = data.get("code", None)
        existing_code = "#!" + db_script.interpreter + "\n" + db_script.code
        if code is not None and code != existing_code:
            db_script.set_code(code)
            changes.update({"code_changed": True})
            # Do not store the diff in the database (it's big)

        time_limit = data.get("time_limit", None)
        if time_limit is not None and time_limit != db_script.time_limit:
            db_script.time_limit = time_limit
            changes.update({"time_limit": time_limit})

        username = data.get("username", None)
        if username is not None and username != db_script.username:
            db_script.username = username
            changes.update({"username": username})

        if not changes:
            return None
        self._event_log.create_log_entry(db_script, describe_dict("Change:", changes))
        return db_script

    @requires("ViewAccount")
    def get_scripts(self, ids=None, titles=None, permission="ViewScript"):
        """
        Get all scripts in the account that the person has access to.

        If C{permission} is C{None}, the resultset won't be filtered by access
        contexts accessible by the user.
        """
        where = [Script.account_id == self.account.id]
        if permission is not None:
            # XXX this is needed to check for duplicated names in the UI, as
            # the user might not see all scripts with the default permission.
            # Eventually, we should move the whole validation to the model.
            access_context_ids = self.get_access_context_ids(permission)
            where.append(Script.access_context_id.is_in(access_context_ids))
        if ids is not None:
            where.append(Script.id.is_in(ids))
        if titles is not None:
            where.append(Script.title.is_in(titles))

        result = self.account.store.find(Script, *where)
        result.order_by(Script.title)
        return result

    @requires("ViewAccount")
    def get_script(self, id):
        """Return the L{Script} with the matching ID.

        @raises UnknownScriptError: If no L{Script{ with the given ID is found.
        """
        script = self.get_scripts(ids=[id]).one()
        if script is None:
            raise UnknownScriptError(id)
        return script

    @requires("ViewAccount")
    def remove_scripts(self, scripts):
        """Remove the given scripts.

        @returns: L{Script}s that didn't get removed because the user doesn't
        have permission on them.
        """
        access_context_ids = self.get_access_context_ids("EditScript")

        remaining_scripts = []
        for script in scripts:
            if script.access_context_id in access_context_ids:
                self.account.store.remove(script)
                self._event_log.create_log_entry(script, "Script removed.")
            else:
                remaining_scripts.append(script)
        return remaining_scripts

    @requires("ViewAccount")
    def copy_script(self, script_id, title, access_context_name):
        """Copy a script."""
        script = self.get_script(script_id)
        if access_context_name is None:
            access_context_name = script.access_context.name
        attachments = dict(
            [
                (attachment.filename, attachment.get_data())
                for attachment in script.attachments
            ]
        )
        code = "#!" + script.interpreter + "\n" + script.code
        new_script = self.create_script(
            title,
            code,
            script.time_limit,
            script.username,
            attachments,
            access_context_name,
        )
        return new_script
