# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging

from zope.component import getUtility
from zope.interface import implementer

from storm.locals import Int, Unicode, AutoReload, Reference, ReferenceSet, Store
from storm.expr import NamedFunc
from storm.zope.interfaces import IZStorm

from canonical.security.protect import requires, requires_advisor, PRIVATE

from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.activity.group import EXECUTE_SCRIPT_SUMMARY
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.main.person import Person
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.helper import AccountHelper
from canonical.landscape.model.script.interfaces import IScript
from canonical.landscape.model.script.errors import (
    DuplicateScriptAttachmentError,
    TooManyScriptAttachmentsError,
    ScriptAttachmentSizeError,
)


__all__ = [
    "Script",
    "ScriptExecution",
    "ScriptAttachment",
    "Attachment",
    "ScriptAttachmentLink",
    "Length",
    "AttachmentManagement",
    "DuplicateScriptError",
    "get_attachment_chunk",
    "get_script_by_id",
    "EmptyScriptCodeError",
    "EmptyScriptInterpreterError",
    "split_code_on_interpreter",
]

__clearlist__ = [
    "ScriptExecution",
    "DuplicateScriptError",
    "EmptyScriptCodeError",
    "EmptyScriptInterpreterError",
    "split_code_on_interpreter",
]

__requires_prefix__ = "landscape."


class DuplicateScriptError(DuplicateModelError):
    """Raised when a script with the same title already exists."""

    model_name = "Script"


class EmptyScriptCodeError(Exception):
    """Raised when a script code body containts no viable content."""


class EmptyScriptInterpreterError(Exception):
    """Raised when a script has no interpreter."""


@requires_advisor
class ScriptExecution(object):
    """
    Allow execution of scripts over a selection of computers.
    """

    def __init__(self, selection):
        self.selection = selection

    @requires("CreateActivity")
    def execute(
        self,
        person,
        title,
        interpreter,
        code,
        username,
        time_limit,
        attachments,
        existing_attachments,
        deliver_after=None,
        batch_size=None,
    ):
        """
        @param person: the executor of the script.
        @type person: L{canonical.landscape.model.main.person.Person}

        @param title: title identifier for the script.
        @type title: C{unicode}

        @param interpreter: interpreter to be used to launch the script on
            client side.
        @type interpreter: C{unicode}

        @param code: code body of the script to be launched.
        @type code: C{unicode}

        @param username: the script will be run as this user.
        @type username: C{unicode}

        @param time_limit: amount of time to wait for the process to end before
            the client kills and returns a timeout error.
        @type time_limit: C{int}

        @param attachments: attachments for this specific script execution
            request, to be stored.
        @type attachments: C{dict} of (filename, data)

        @param existing_attachments: attachments already stored in the
            database.
        @type existing_attachments: C{list} of C{Attachment} instance

        @param deliver_after: Optionally, a C{datetime} representing the
            earliest time the L{Activity}s created to carry out this request
            should be delivered to selected L{Computer}s.  Defaults to
            immediate delivery.

        @param batch_size: If specified, the maximum number of activities
            created in each run.

        @return: The parent L{Activity} representing the script execution.
        """
        parent = self.selection.create_activity_group(
            summary=EXECUTE_SCRIPT_SUMMARY, creator=person
        )
        for filename, data in list(attachments.items()):
            attachment = Attachment(filename, self.selection.account)
            attachment.set_data(data)
            parent.add_attachment(attachment)
        for attachment in existing_attachments:
            parent.add_attachment(attachment)

        self.run_execute_batch(
            person,
            title,
            interpreter,
            code,
            username,
            time_limit,
            deliver_after,
            batch_size,
            parent,
        )
        parent.info.create_log_entry("Group activity created", person=person)
        return parent

    def run_execute_batch(
        self,
        person,
        title,
        interpreter,
        code,
        username,
        time_limit,
        deliver_after,
        batch_size,
        parent,
    ):
        """
        Create the activities for every computer, possibly splitting the work
        into separate jobs if batch size is specified and there are more
        computers to handle.
        """
        if batch_size is not None:
            logging.warning("Batch usage is currently ignored")
        selection = self.selection
        for computer in selection.get_selected():
            computer.request_execute_script(
                parent,
                title,
                interpreter,
                code,
                username,
                time_limit,
                deliver_after=deliver_after,
            )
        parent.settle()


@requires_advisor
@implementer(IScript)
class Script(AccountHelper):
    """
    Represent a script which is stored to be run several times.

    @ivar id: numeric identifier.
    @type id: C{int}

    @ivar account_id: identifier of the account this script is linked to.
    @type account_id: C{int}

    @param code: code body of the script.
    @type code: C{unicode}

    @ivar title: title of the script.
    @type title: C{unicode}

    @ivar time_limit: default amount of time to wait for the process to end.
    @type time_limit: C{int}

    @ivar username: default user to execute the script as (optional).
    @type username: C{unicode}

    @ivar creator_id: identifier of the person who created this script.
    @type creator_id: C{int}
    """

    __storm_table__ = "script"
    id = Int(primary=True, default=AutoReload)
    account_id = Int(allow_none=False)
    interpreter = Unicode(allow_none=False)
    code = Unicode(allow_none=False)
    title = Unicode(allow_none=False)
    time_limit = Int(allow_none=False)
    username = Unicode()
    creator_id = Int(allow_none=False)
    access_context_id = Int(allow_none=False)

    requires(
        "ViewScript",
        "EditScript",
        [
            "id",
            "interpreter",
            "code",
            "title",
            "account",
            "time_limit",
            "username",
            "creator",
            "attachments",
        ],
    )
    requires("ViewScript", PRIVATE, ["access_context", "access_context_id"])

    def __init__(
        self,
        account,
        creator,
        code,
        title,
        time_limit,
        username=None,
        access_context=None,
    ):
        self.interpreter, self.code = split_code_on_interpreter(code)
        super(Script, self).__init__(account)
        self.creator = creator
        self.title = title
        self.time_limit = time_limit
        self.username = username
        if access_context is None:
            access_context = self.account.access_context
        self.access_context = access_context

    def _get_access_context(self):
        return self.account.get_access_context(self.access_context_id)

    def _set_access_context(self, access_context):
        self.access_context_id = access_context.id

    access_context = property(_get_access_context, _set_access_context)

    def _get_creator(self):
        if self.creator_id:
            return get_main_store().get(Person, self.creator_id)
        return None

    def _set_creator(self, person):
        if person is not None:
            self.creator_id = person.id
        else:
            self.creator_id = None

    creator = property(_get_creator, _set_creator)

    @requires("EditScript")
    def set_code(self, code):
        self.interpreter, self.code = split_code_on_interpreter(code)

    @requires("ViewScript")
    def get_code(self):
        return "#!" + self.interpreter + "\n" + self.code

    @requires("EditScript")
    def add_attachment(self, filename, data):
        """
        Attach an attachment  to this script.

        @param filename: name of the attachment.
        @type filename: C{unicode}

        @param data: content of the attachment.
        @type data: C{str}

        @return: the attachment instance created.
        @rtype: L{ScriptAttachment}
        """
        attachments = list(self.attachments)
        max_size = self.account.max_attachment_size
        max_files = self.account.max_attachment_files

        if len(data) > max_size:
            raise ScriptAttachmentSizeError(self.id, max_size)

        if len(attachments) >= max_files:
            raise TooManyScriptAttachmentsError(self.id, max_files)

        filenames = [att.filename for att in attachments]
        if filename in filenames:
            raise DuplicateScriptAttachmentError(self.id, filename)

        new_attachment = ScriptAttachment(filename, self.account)
        new_attachment.set_data(data)
        self.attachments.add(new_attachment)
        return new_attachment

    @requires("EditScript")
    def remove_attachments_by_ids(self, ids):
        """
        Remove attachments of this script matching the given identifiers.
        """
        ids = set(ids)
        if not ids:
            return

        # We do it in two phases, because we want to exactly know the
        # attachments linked to this script
        result = self.attachments.find(ScriptAttachment.id.is_in(ids))
        ids = set(result.values(ScriptAttachment.id))
        if not ids:
            return

        self.attachments.clear(ScriptAttachment.id.is_in(ids))
        # Now let's delete actual attachments, but only if there is no
        # script execution using it
        store = Store.of(self)
        attachments = store.find(
            ScriptExecutionAttachment,
            ScriptExecutionAttachment.attachment_id.is_in(ids),
        )
        attachments_kept = set(
            script_execution.attachment_id for script_execution in attachments
        )
        ids_to_delete = ids - attachments_kept
        if ids_to_delete:
            store.find(Attachment, Attachment.id.is_in(ids_to_delete)).remove()

    @requires("ViewScript")
    def get_attachment_by_id(self, id):
        """
        Retrieve one script attachment by its id.
        """
        return self.attachments.find(ScriptAttachment.id == id).one()

    @requires("ViewScripts")
    def get_state(self):
        """Return the state of the script in JSON serialisable format."""
        return {
            "id": self.id,
            "access_group": self.access_context.name,
            "creator": self.creator.get_state() if self.creator else None,
            "title": self.title,
            "time_limit": self.time_limit,
            "username": self.username if self.username else "",
            "attachments": [
                attachment.filename for attachment in list(self.attachments)
            ],
        }

    @requires("ViewScripts")
    def get_executable_code(self):
        """Return script code in an executable format."""
        if self.interpreter:
            first_line = "#!" + self.interpreter
            script = "\n".join((first_line, self.code))
        else:
            script = self.code
        return script


class Length(NamedFunc):
    name = "LENGTH"


@requires_advisor
@implementer(IAccessDelegateAware)
class Attachment(AccountHelper):
    """
    Stored attachments for an account.
    """

    __storm_table__ = "attachment"

    id = Int(primary=True, default=AutoReload)
    filename = Unicode(allow_none=False)
    account_id = Int(allow_none=False)
    data_oid = Int(allow_none=False)

    requires("ViewScript", "EditScript", ["id", "filename"])

    def __init__(self, filename, account):
        super(Attachment, self).__init__(account)
        self.filename = filename
        self.data_oid = self._create()

    def _get_activity(self):
        # Attachments are either for a script or an activity parent. By
        # default, check for an activity parent. ScriptAttachment will
        # check if it's attached to a script.
        store = Store.of(self)
        result = store.find(
            ActivityParent,
            ActivityParent.id == ScriptExecutionAttachment.activity_id,
            ScriptExecutionAttachment.attachment_id == self.id,
        )
        return result.one()

    _activity = property(_get_activity)

    @property
    def access_delegate(self):
        return self._activity

    @requires("ViewScript")
    def get_data(self):
        fd = self._open()
        data = self._read(fd)
        self._close(fd)
        return data

    @requires("EditScript")
    def set_data(self, data):
        fd = self._open()
        self._write(fd, data)
        self._close(fd)

    @requires("ViewScript")
    def get_size(self):
        fd = self._open()
        size = self._size(fd)
        self._close(fd)
        return size

    def _open(self):
        # Note: 393216 = CAST('20000' | x'40000' AS integer)
        result = self.account.store.execute(
            "SELECT lo_open(?, 393216)", (self.data_oid,)
        )
        return result.get_one()[0]

    def _close(self, fd):
        self.account.store.execute("SELECT lo_close(?)", (fd,))

    def _size(self, fd):
        self.account.store.execute("SELECT lo_lseek(?, 0, 2)", (fd,))
        result = self.account.store.execute("SELECT lo_tell(?)", (fd,))
        self.account.store.execute("SELECT lo_lseek(?, 0, 0)", (fd,))
        return result.get_one()[0]

    def _create(self):
        store = self.account.store
        # Prevent Storm from flushing to early (self.data_oid is still unset)
        store.block_implicit_flushes()
        server_version = int(store.execute("SHOW server_version_num").get_one()[0])
        result = store.execute("SELECT lo_creat(-1)")
        data_oid = result.get_one()[0]
        if server_version >= 90000:
            # Enforce ownership to landscape when > 9.0
            store.execute("ALTER LARGE OBJECT ? OWNER TO landscape", (data_oid,))
        store.unblock_implicit_flushes()
        return data_oid

    def _read(self, fd):
        size = self._size(fd)
        result = self.account.store.execute("SELECT loread(?, ?)", (fd, size))
        return result.get_one()[0].tobytes().decode()

    def _write(self, fd, data):
        self.account.store.execute("SELECT lowrite(?, ?)", (fd, data))


@requires_advisor
class ScriptAttachment(Attachment):
    """
    Attachment specific to a stored script.
    """

    requires("ViewScript", "EditScript", ["script"])

    def _get_script(self):
        store = Store.of(self)
        result = store.find(
            ScriptAttachmentLink, ScriptAttachmentLink.attachment_id == self.id
        ).one()
        return result.script

    script = property(_get_script)

    @property
    def access_delegate(self):
        return self.script


class ScriptAttachmentLink(object):
    """
    Link between Script and ScriptAttachment.
    """

    __storm_table__ = "script_attachment"
    __storm_primary__ = "script_id", "attachment_id"
    script_id = Int(allow_none=False)
    attachment_id = Int(allow_none=False)

    script = Reference(script_id, Script.id)


Script.attachments = ReferenceSet(
    Script.id,
    ScriptAttachmentLink.script_id,
    ScriptAttachmentLink.attachment_id,
    ScriptAttachment.id,
    order_by=ScriptAttachment.filename,
)


class ScriptExecutionAttachment(object):

    __storm_table__ = "execute_script_attachment"
    __storm_primary__ = "activity_id", "attachment_id"
    activity_id = Int(allow_none=False)
    attachment_id = Int(allow_none=False)


ActivityParent.attachments = ReferenceSet(
    ActivityParent.id,
    ScriptExecutionAttachment.activity_id,
    ScriptExecutionAttachment.attachment_id,
    Attachment.id,
    order_by=Attachment.filename,
)


class AttachmentManagement(AccountHelper):
    def get_attachment(self, attachment_id):
        """Retrieve an attachment given its ID."""
        return self.account.store.find(
            Attachment,
            Attachment.id == attachment_id,
            Attachment.account_id == self.account_id,
        ).one()


def get_attachment_chunk(store_name, data_oid, size, position):
    """Return a chunk of a stored attachment."""
    zstorm = getUtility(IZStorm)
    store = zstorm.get(store_name)
    result = store.execute("SELECT lo_open(?, 393216)", (data_oid,))
    fd = result.get_one()[0]
    store.execute("SELECT lo_lseek(?, ?, 0)", (fd, position))
    result = store.execute("SELECT loread(?, ?)", (fd, size))
    chunk = result.get_one()[0].tobytes()
    store.execute("SELECT lo_close(?)", (fd,))
    return chunk


def get_script_by_id(account, script_id):
    result = account.store.find(
        Script, Script.account_id == account.id, Script.id == script_id
    )
    return result.one()


def _validate_script_content(content):
    """
    Return C{True} if Script content is viable and uncommented code.
    """
    if content is None or len(content.strip()) <= 0:
        return False

    for line in content.split("\n"):
        stripped_line = line.strip()
        if stripped_line and not stripped_line.startswith("#"):
            return True
    return False


def split_code_on_interpreter(code):
    """
    Split up a script into two parts, the interpreter (without the
    shebang characters) and the rest of the code. Raise an
    C{EmptyScriptCodeError} because we don't allow empty code content.
    """
    if code is None:
        interpreter = None
        rest = None
    elif code.startswith("#!"):
        result = code.split("\n", 1)
        if len(result) == 1:
            interpreter = code[2:]
            rest = None
        else:
            interpreter = result[0][2:]
            rest = result[1]
    else:
        interpreter = ""
        rest = code

    if not _validate_script_content(rest):
        raise EmptyScriptCodeError(
            "Script code content appears empty or only contains "
            "whitespace. Ensure code is not commented out and newlines "
            "are preserved."
        )
    if not interpreter:
        raise EmptyScriptInterpreterError(
            "Script code content does not start with an interpreter "
            "line. eg: #!/bin/bash"
        )
    return interpreter, rest
