# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from functools import reduce
from datetime import timedelta

from storm.expr import Desc, Count, Or
from storm.variables import TimeDeltaVariable

from canonical.landscape.model.main.computer import (
    Computer,
    PingTime,
    ComputerLicense,
    ComputerStatus,
)
from canonical.landscape.model.main.tag import Tag, ComputerTag
from canonical.landscape.model.main.license import License
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
)
from canonical.landscape.model.resource.user import User
from canonical.landscape.model.resource.processor import Processor
from canonical.landscape.model.main.store import get_main_store

__all__ = ["ComputerSelectionInfo"]

__clearlist__ = __all__


class ComputerSelectionInfo(object):
    """
    Helper class to retrive aggregated info about computers in a
    L{ComputerSelection}.

    @param selection: the computer selection
    """

    def __init__(self, selection):
        self.selection = selection
        self._main_store = get_main_store()

    def get_last_ping_time_computer(self):
        """
        Return the computer in the selection with the most recent ping time.
        """
        computer_ids = self.selection.get_selected_computer_ids()
        result = self._main_store.find(
            PingTime,
            PingTime.last_ping_time != None,
            PingTime.computer_id.is_in(computer_ids),
        )
        ping_time = result.order_by(Desc(PingTime.last_ping_time)).first()
        if ping_time is not None:
            return ping_time.computer
        else:
            return None

    def get_last_accepted_computer(self):
        """
        Return the computer most recently accepted computer in the selection.
        """
        result = self.selection.get_selected(
            order_by=Desc(Computer.accept_time), limit=1
        )
        return result[0] if result else None

    def get_computers_processor_info(self, fields):
        """
        Return a set of processor info for selected computers.

        @param fields: a list of names of of L{Process} model fields to include
        """
        info = set()
        stores = get_resource_stores_for_computer_ids(
            self.selection.get_selected_computer_ids()
        )

        for store, computer_ids in stores.items():
            result = store.find(
                tuple(getattr(Processor, field) for field in fields),
                Processor.computer_id.is_in(computer_ids),
            )
            result.config(distinct=True)
            info.update(result)

        return info

    def get_tags_count(self):
        """
        Return a dict with tag names for selected computers and their count.
        """
        computer_ids = self.selection.get_selected_computer_ids()
        result = self._main_store.find(
            (Tag.name, Count()),
            Tag.id == ComputerTag.tag_id,
            ComputerTag.computer_id.is_in(computer_ids),
        )
        result.group_by(Tag.name)
        return dict(result)

    def get_ping_exchange_diff_count(self):
        """
        Return a count of the number of machines in the selection that have
        a significant difference between their last ping time and their
        last exchange time.
        """
        computer_ids = self.selection.get_selected_computer_ids()
        return self._main_store.find(
            Count(),
            ComputerStatus.computer_id.is_in(computer_ids),
            ComputerStatus.computer_id == PingTime.computer_id,
            Or(
                ComputerStatus.last_exchange_time - PingTime.last_ping_time
                > TimeDeltaVariable(timedelta(hours=1)),
                PingTime.last_ping_time - ComputerStatus.last_exchange_time
                > TimeDeltaVariable(timedelta(hours=1)),
            ),
        ).one()

    def get_licenses(self):
        """Return the licenses for selected computers."""
        computer_ids = self.selection.get_selected_computer_ids()
        result = self._main_store.find(
            License,
            ComputerLicense.license_id == License.id,
            ComputerLicense.computer_id == Computer.id,
            Computer.id.is_in(computer_ids),
        )
        return result.config(distinct=True)

    def get_users_info(self, fields):
        """
        Return a set of user info for selected computers.  Only users common to
        all computers (based on the provided fields) are returned.

        @param fields: a list of names of of L{Process} model fields to include
        """
        info = []
        stores = get_resource_stores_for_computer_ids(
            self.selection.get_selected_computer_ids()
        )
        fields_expr = tuple(getattr(User, field) for field in fields)

        for store, computer_ids in stores.items():
            result = store.find(fields_expr, User.computer_id.is_in(computer_ids))
            result.group_by(*fields_expr)
            result.having(Count() == len(computer_ids))
            info.append(set(result))

        return reduce(set.intersection, info) if info else set()
