# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer
from zope.security import proxy

from storm.store import EmptyResultSet
from storm.expr import Desc, Asc
from storm.locals import Int, List

from canonical.search.engine import SearchCriteriaError

from canonical.landscape.model.main.interfaces import IAccountAware, IPersonAware
from canonical.landscape.model.main.computer import Computer, PingTime
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.selection.interfaces import IComputerSelection
from canonical.landscape.model.computer.collection import ComputerCollection

__all__ = ["ComputerSelection", "SessionComputerSelection", "ComputerSelectionBase"]

__clearlist__ = __all__


def reset_cache(f):
    def _f(self, *args, **kwargs):
        self._reset_cache()
        return f(self, *args, **kwargs)

    return _f


NO_ORDER = object()


@implementer(IAccountAware, IPersonAware)
class ComputerSelectionBase(object):

    access_cutoff_role = None

    _selection_cache = None
    _available_cache = None
    _available_count = None
    _available_ids = None
    _access_contexts = None

    def __init__(self, account, person=None, permission=None):
        self.account = account
        self.person = person
        self.permission = permission
        self._included_computers = set()
        self._included_tags = set()
        self._included_access_groups = set()
        self._included_search_criteria = None
        self._included_search_name = None
        self._main_store = get_main_store()

    def _reset_cache(self):
        self._selection_cache = None
        self._available_cache = None
        self._available_count = None
        self._available_ids = None
        self._access_contexts = None

    def get_available_count(self):
        if self._available_count is not None:
            return self._available_count
        try:
            computers = self._find()
        except SearchCriteriaError:
            self._available_count = 0
        else:
            self._available_count = computers.count()
        return self._available_count

    def get_available_ids(self, order_by=NO_ORDER):
        """
        Return the list of available computer IDs, ordered by C{order_by} if
        specified.
        """
        if self._available_ids is None:
            self._available_ids = {}
        available_ids = self._available_ids
        if order_by not in available_ids:
            try:
                computers = self._find(order_by=order_by)
            except SearchCriteriaError:
                available_ids[order_by] = []
            else:
                if self._included_tags and order_by is not NO_ORDER:
                    # We need to include the additional column in the select
                    expr = order_by
                    if isinstance(expr, (Asc, Desc)):
                        expr = expr.expr
                    if expr is None:
                        expr = Computer.title
                    result = computers.values(Computer.id, expr)
                    available_ids[order_by] = [row[0] for row in result]
                else:
                    available_ids[order_by] = list(computers.values(Computer.id))
        return available_ids[order_by]

    @reset_cache
    def clear(self):
        self.clear_included_computers()
        self.clear_included_tags()
        self.clear_included_access_groups()
        self.clear_selected_computers()
        self.clear_search_criteria()

    def get_available(self, order_by=None, offset=None, limit=None):
        """
        Return a list of the available computers, using a cache to prevent
        useless queries.
        """
        cache_key = (order_by, offset, limit)
        if self._available_cache is None:
            self._available_cache = {}
        if cache_key not in self._available_cache:
            available = self._find(order_by=order_by, offset=offset, limit=limit)
            self._available_cache[cache_key] = list(available)
        return self._available_cache[cache_key]

    def get_selected_count(self):
        """Get the count of selected computer ids."""
        return len(self.get_selected_computer_ids())

    def get_selected(self, order_by=None, offset=None, limit=None):
        """
        Return a list of the selected computers, using a cache to prevent
        useless queries.
        """
        cache_key = (order_by, offset, limit)
        if self._selection_cache is None:
            self._selection_cache = {}
        if cache_key not in self._selection_cache:
            try:
                selected = self._find(
                    selected=True, order_by=order_by, offset=offset, limit=limit
                )
            except SearchCriteriaError:
                selected = []
            self._selection_cache[cache_key] = list(selected)
        return self._selection_cache[cache_key]

    def is_selected(self, computer):
        """Is the given computer selected?"""
        id = computer
        if proxy.isinstance(computer, Computer):
            id = computer.id
        return id in self.get_selected_computer_ids()

    def get_access_contexts(self):
        """Return the L{AcccessContext}s linked to the selected computers."""
        if self._access_contexts is None:
            result = self._find(selected=True, order_by=NO_ORDER)
            result.config(distinct=True)
            ids = list(result.values(Computer.access_context_id))
            access_contexts = self.account.get_access_contexts(ids=ids)
            self._access_contexts = access_contexts
        return self._access_contexts

    def __iter__(self):
        """Iterating over the selection yields selected computers."""
        return iter(self.get_selected())

    def get_computers_by_ids(self, computer_ids):
        """
        Return some specific computers given their IDs.
        """
        collection = ComputerCollection(self._main_store, account_id=self.account.id)
        if self.person is not None:
            permission = self.permission
            if permission is None:
                permission = "landscape.ViewComputer"
            collection = collection.with_permission(self.person, permission)
        collection = collection.with_computer_ids(computer_ids)
        return collection.find()

    def _find(self, selected=False, order_by=None, limit=None, offset=None):
        collection = ComputerCollection(self._main_store, account_id=self.account.id)
        # Most of the ordering in the UI just uses attributes.  We also have a
        # case where we want to order by the last time a machine was accessed,
        # which requires an outer join.  We explicitly hardcode support for
        # that here.
        if order_by and order_by is not NO_ORDER:
            expr = order_by
            if isinstance(order_by, Desc):
                expr = order_by.expr
            if expr.cls is PingTime:
                collection = collection.with_ping_times()

        if self.person is not None:
            permission = self.permission
            if permission is None:
                permission = "landscape.ViewComputer"
            collection = collection.with_permission(self.person, permission)

        if selected:
            selected_computer_ids = self.get_selected_computer_ids(available_only=False)
            if not selected_computer_ids:
                return EmptyResultSet()
            else:
                collection = collection.with_computer_ids(selected_computer_ids)

        included_tag_ids = [tag.id for tag in self.get_included_tags()]
        if included_tag_ids:
            collection = collection.with_tags(included_tag_ids)

        included_group_names = [
            group.name for group in self.get_included_access_groups()
        ]
        if included_group_names:
            collection = collection.with_access_groups(included_group_names)

        included_computer_ids = self.get_included_computer_ids()
        if included_computer_ids:
            collection = collection.with_computer_ids(included_computer_ids)

        included_search_criteria = self.get_search_criteria()
        if included_search_criteria:
            collection = collection.with_search_criteria(included_search_criteria)

        result = collection.find()
        if order_by is None:
            result.order_by(Computer.title)
        elif order_by is not NO_ORDER:
            result.order_by(order_by)
        if limit or offset:
            result.config(offset=offset, limit=limit)
        return result

    def _ids_for_computers(self, computers):
        ids = set()
        for computer in computers:
            if proxy.isinstance(computer, Computer):
                ids.add(computer.id)
            elif isinstance(computer, int):
                ids.add(computer)
            else:
                raise TypeError("%r isn't a computer" % (computer,))
        return ids

    @reset_cache
    def clear_included_computers(self):
        self._included_computers.clear()

    def get_included_computer_ids(self):
        return self._included_computers

    @reset_cache
    def include_computers(self, computers):
        """
        Define a set of computers and limit the results of subsequent selection
        operations such as set_search_criteria. Multiple calls to
        include_computers will add any new computers to the set. If
        L{ComputerSelection.include_computers} has never been called for this
        selection, default behavior is to match any computer in the account.
        """
        computers = self._ids_for_computers(computers)
        self._included_computers.update(computers)

    @reset_cache
    def discard_included_computers(self, computers):
        computers = self._ids_for_computers(computers)
        self._included_computers.difference_update(computers)

    @reset_cache
    def clear_included_tags(self):
        self._included_tags.clear()

    @reset_cache
    def clear_included_access_groups(self):
        self._included_access_groups.clear()

    def get_included_tags(self):
        return self._included_tags

    def get_included_access_groups(self):
        return self._included_access_groups

    def get_search_criteria(self):
        return self._included_search_criteria

    def get_search_name(self):
        return self._included_search_name

    @reset_cache
    def clear_search_criteria(self):
        self._included_search_criteria = None
        self.clear_search_name()

    def clear_search_name(self):
        self._included_search_name = None

    @reset_cache
    def include_tags(self, tags):
        self._included_tags.update(tags)

    @reset_cache
    def include_access_groups(self, access_groups):
        self._included_access_groups.update(access_groups)

    @reset_cache
    def set_search_criteria(self, search_criteria):
        self._included_search_criteria = search_criteria

    @reset_cache
    def set_saved_search_criteria(self, search_name):
        from canonical.landscape.model.account.savedsearch import SavedSearchManagement

        management = SavedSearchManagement(self.account)
        saved_search = management.get_saved_search_by_name(search_name)
        if saved_search:
            self._included_search_criteria = saved_search.search
            self.set_saved_search_name(search_name)
        else:
            self._included_search_criteria = None

    def set_saved_search_name(self, search_name):
        self._included_search_name = search_name

    @reset_cache
    def discard_included_tags(self, tags):
        self._included_tags = self._included_tags - set(tags)

    @reset_cache
    def discard_included_access_groups(self, access_groups):
        self._access_groups = self._access_groups - set(access_groups)

    def create_activity_group(self, **kwargs):
        """Create an ActivityGroup activity."""
        from canonical.landscape.model.activity.parent import ActivityParent
        from canonical.landscape.model.activity.group import ActivityGroup

        # Create the parent in the same access context as any of the selected
        # computers, as the user may not have permissions on the global
        # context, but should have it on the selection on which the parent
        # acts.
        computers = self.get_selected(limit=1)
        try:
            (selected,) = computers
        except ValueError:
            context = None
        else:
            context = selected.access_context
        return ActivityParent(
            self.account, ActivityGroup, access_context=context, **kwargs
        )

    @property
    def access_context(self):
        return self.account.access_context


@implementer(IComputerSelection)
class ComputerSelection(ComputerSelectionBase):
    def __init__(self, account, person=None, permission=None):
        super(ComputerSelection, self).__init__(
            account, person=person, permission=permission
        )
        self._selected_computer_ids = set()

    def get_selected_computer_ids(self, available_only=True):
        """Get the set of selected computer ids.

        @param available_only: If C{False}, all the selected
            computer ids are returned. Otherwise, only selected computer
            ids which are also available computer ids are returned.
        """
        if available_only:
            available = set(self.get_available_ids())
            return set(self._selected_computer_ids).intersection(available)
        return self._selected_computer_ids

    @reset_cache
    def select_computers(self, computers):
        ids = self._ids_for_computers(computers)
        self._selected_computer_ids.update(ids)

    @reset_cache
    def clear_selected_computers(self):
        self._selected_computer_ids.clear()

    @reset_cache
    def unselect_computers(self, computers):
        ids = self._ids_for_computers(computers)
        self._selected_computer_ids.difference_update(ids)


@implementer(IComputerSelection)
class SessionComputerSelection(ComputerSelectionBase):
    """Computer selection persists selection state changes to the session."""

    _selected_computer_ids = None

    def _reset_cache(self):
        super(SessionComputerSelection, self)._reset_cache()
        self._selected_computer_ids = None

    def _get_data(self):
        result = self._main_store.find(
            ComputerSelectionData, person_id=self.person.id, account_id=self.account.id
        )
        data = result.one()
        if data is None:
            data = ComputerSelectionData(self.person, self.account)
            self._main_store.add(data)
        return data

    def _get(self):
        data = self._get_data()
        return set(data.selected_computer_ids)

    def _add(self, items):
        data = self._get_data()
        current = set(data.selected_computer_ids)
        updated = set(current)
        updated.update(items)
        if current != updated:
            data.selected_computer_ids = list(updated)

    def _subtract(self, items):
        data = self._get_data()
        current = set(data.selected_computer_ids)
        updated = current - set(items)
        if current != updated:
            data.selected_computer_ids = list(updated)

    def _clear(self):
        data = self._get_data()
        data.selected_computer_ids = []

    def get_selected_computer_ids(self, available_only=True):
        """Get the set of selected computer ids.

        @param available_only: If C{False}, all the selected
            computer ids are returned. Otherwise, only selected computer
            ids which are also available computer ids are returned.
        """
        if self._selected_computer_ids is None:
            self._selected_computer_ids = self._get()
        if available_only:
            available = set(self.get_available_ids())
            return set(self._selected_computer_ids).intersection(available)
        return self._selected_computer_ids

    @reset_cache
    def select_computers(self, computers):
        ids = self._ids_for_computers(computers)
        self._add(ids)

    @reset_cache
    def unselect_computers(self, computers):
        ids = self._ids_for_computers(computers)
        self._subtract(ids)

    @reset_cache
    def clear_selected_computers(self):
        self._clear()


class ComputerSelectionData(object):
    """Store data about the current computer selection for a L{Person}.

    @param person: The L{Person} the selection is associated with.
    """

    __storm_table__ = "computer_selection_data"
    __storm_primary__ = "person_id", "account_id"

    person_id = Int(allow_none=False)
    account_id = Int(allow_none=False)
    selected_computer_ids = List(type=Int())

    def __init__(self, person, account):
        self.person_id = person.id
        self.account_id = account.id
        self.selected_computer_ids = []
