from collections.abc import Iterator
from collections import namedtuple
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Mapping

import storm.expr
from storm.expr import And
from storm.expr import Or
from storm.expr import Count
from storm.expr import Join
from storm.expr import LeftJoin
from storm.expr import Max
from storm.expr import Select

from zope.component import adapts
from zope.interface import implementer
from zope.security.checker import defineChecker
from zope.security.proxy import removeSecurityProxy

from canonical.lib.dpkg import version_compare

from canonical.security.protect import PRIVATE, PUBLIC
from canonical.security.protect import requires
from canonical.security.protect import requires_advisor

from canonical.landscape.model.activity.group import (
    CHANGE_SNAPS_SUMMARY,
    validate_delivery_time,
)
from canonical.landscape.model.activity.snap import (
    HoldSnaps,
    InstallSnaps,
    RefreshSnaps,
    RemoveSnaps,
    UnholdSnaps,
)
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.snap import InstalledSnap, Snap
from canonical.landscape.model.selection.interfaces import IComputerSelection


__all__ = ["SNAP_ACTIONS", "SnapManagement"]
__clearlist__ = __all__
__requires_prefix__ = "landscape."


SNAP_ACTIONS = {
    "install": InstallSnaps,
    "refresh": RefreshSnaps,
    "remove": RemoveSnaps,
    "hold": HoldSnaps,
    "unhold": UnholdSnaps,
}

SUPPORTED_CLIENT_VERSION = "23.02+git6282"


@requires_advisor
@dataclass
class SnapComputerInfo:

    requires(PUBLIC, PUBLIC)

    version: str
    revision: str
    tracking_channel: str
    held_until: datetime
    id: int
    title: str


@requires_advisor
@dataclass
class ComputerGroupedSnap:

    requires(PUBLIC, PUBLIC)

    name: str
    summary: str
    computers: List[SnapComputerInfo]

    @property
    def all_held(self):
        """Snap is held for all computers in the group."""
        return all(computer.held_until for computer in self.computers)


@requires_advisor
class ComputerGroupedSnaps(Iterator):
    """
    A class representing a iterator of installed snaps, grouped by snap
    name and containing a seqeunce of all the computers with any rev of
    the snap installed.
    """

    def __init__(self, **search_kwargs):
        self._result = None
        self._curr_snap_i = 0
        self.store = get_main_store()

        self._select = Select(
            (
                Snap.name,
                Snap.summary,
                InstalledSnap.version,
                InstalledSnap.revision,
                InstalledSnap.tracking_channel,
                InstalledSnap.held_until,
                InstalledSnap.computer_id,
                Computer.title,
            ),
            tables=(
                InstalledSnap,
                Join(Snap, InstalledSnap.snap_id == Snap.id),
                Join(Computer, InstalledSnap.computer_id == Computer.id),
            ),
            order_by=Snap.name,
            **search_kwargs,
        )

    def __next__(self) -> ComputerGroupedSnap:
        """
        Iterates over the result set, combining like-named snaps into
        single records.
        """
        if not self._result:
            # Lazily populate the result set.
            self._result = list(self.store.execute(self._select))

        try:
            (
                name,
                summary,
                version,
                revision,
                tracking_channel,
                held_until,
                id,
                title,
            ) = self._result[self._curr_snap_i]
            curr_snap = ComputerGroupedSnap(
                name,
                summary,
                [
                    SnapComputerInfo(
                        version, revision, tracking_channel, held_until, id, title
                    )
                ],
            )
            self._curr_snap_i += 1
        except IndexError:
            raise StopIteration

        # For each record after the current one, combine it with the
        # current record until we see one with a different name.
        # This works because the records are ordered by name.
        for snap in self._result[self._curr_snap_i :]:
            (
                name,
                summary,
                version,
                revision,
                tracking_channel,
                held_until,
                id,
                title,
            ) = snap
            next_snap = ComputerGroupedSnap(
                name,
                summary,
                [
                    SnapComputerInfo(
                        version, revision, tracking_channel, held_until, id, title
                    )
                ],
            )

            if curr_snap.name != next_snap.name:
                break

            self._curr_snap_i += 1
            curr_snap.computers += next_snap.computers

        return curr_snap


@requires_advisor
@implementer(IAccessDelegateAware)
class SnapManagement:
    """API for snap operations on selected computers."""

    adapts(IComputerSelection)

    requires("ViewComputer", PRIVATE, ["selection"])

    def __init__(self, selection):
        self.selection = selection
        self._has_supported_computers = None
        self._has_unsupported_computers = None

    @property
    def access_delegate(self):
        return self.selection

    @requires("ViewComputerPackages")
    def get_snap_counts_by_computer(self, offset=0, limit=25):
        """
        Produces a list of dicts, pairing each computer object in the selection
        with a count of the number of snaps it has in each state (installed, currently).
        """
        store = get_main_store()

        join = [
            Computer,
            LeftJoin(InstalledSnap, InstalledSnap.computer_id == Computer.id),
        ]
        computers_with_snaps = (
            store.using(*join)
            .find(
                (Computer, Count(InstalledSnap.snap_id)),
                Computer.id.is_in(c.id for c in self.supported_computers()),
            )
            .group_by(Computer.id)
            .order_by(Computer.id)
            .config(offset=offset, limit=limit)
        )

        return [{"computer": c, "installed": i} for c, i in computers_with_snaps]

    @requires("ViewComputerPackages")
    def search_snaps(self, status="installed", offset=0, limit=25, text=None):
        """
        Produces a list of snaps that match the provided criteria and contain
        the text string, relative to the current selection.
        """
        # XXX: only listing installed and held snaps until we support reaching out to snap store.
        if status not in ("installed", "held"):
            return []

        store = get_main_store()
        computer_ids = self.selection.get_selected_computer_ids()

        expr = InstalledSnap.computer_id.is_in(computer_ids)

        if text:
            expr = And(
                expr,
                Or(
                    InstalledSnap.snap_id.like(f"%{text}%"),
                    Snap.name.like(f"%{text}%"),
                ),
            )

        if status == "held":
            expr = And(
                expr,
                InstalledSnap.held_until != None,
            )

        cgs = list(ComputerGroupedSnaps(where=expr, offset=offset, limit=limit))
        return cgs

    @requires("EditComputerPackages")
    def change_snaps(
        self, person, changes={}, deliver_after=None, deliver_after_window=0
    ):
        """
        Requests snaps be removed for the selected computers.

        `changes` is assumed to only contain change types from
        `SNAP_ACTIONS`.

        Returns the parent activity and whether some computers could not
        receive the activity because their client version is not new
        enough.
        """
        validate_delivery_time(deliver_after_window=deliver_after_window)

        activity = self.selection.create_activity_group(
            summary=CHANGE_SNAPS_SUMMARY, creator=person
        )

        for computer in self.supported_computers():
            for action, snaps in changes.items():
                if action != "install":
                    affected_snaps = self._computer_installed_snaps(computer, snaps)
                else:
                    affected_snaps = snaps

                if not affected_snaps:
                    continue

                child = activity.add(
                    SNAP_ACTIONS[action], computer=computer, snaps=affected_snaps
                )
                EventLog(activity.account, activity.creator).create_log_entry(
                    computer,
                    "Activity created: " + child.info.get_summary(),
                )

        if deliver_after or deliver_after_window:
            activity.schedule(after=deliver_after, after_window=deliver_after_window)

        activity.settle()
        activity.info.create_log_entry("Group activity created", person=person)
        return activity

    @property
    @requires("ViewComputerPackages")
    def has_supported_computers(self):
        if self._has_supported_computers is None:
            self._has_supported_computers = any(self.supported_computers())

        return self._has_supported_computers

    @property
    @requires("ViewComputerPackages")
    def has_unsupported_computers(self):
        if self._has_unsupported_computers is None:
            self._has_unsupported_computers = any(self.unsupported_computers())

        return self._has_unsupported_computers

    def supported_computers(self):
        return (
            c
            for c in self.selection.get_selected()
            if c.current_version
            and version_compare(c.current_version, SUPPORTED_CLIENT_VERSION) >= 0
        )

    def unsupported_computers(self):
        return (
            c
            for c in self.selection.get_selected()
            if c.current_version
            and version_compare(c.current_version, SUPPORTED_CLIENT_VERSION) < 0
        )

    def _computer_installed_snaps(
        self, computer: Computer, snaps: List[Mapping[str, str]]
    ):
        store = get_main_store()
        snap_names = [s["name"] for s in snaps]
        result = store.execute(
            Select(
                Snap.name,
                tables=(Snap, Join(InstalledSnap, Snap.id == InstalledSnap.snap_id)),
                where=And(
                    InstalledSnap.computer_id == computer.id,
                    Snap.name.is_in(snap_names),
                ),
            )
        )

        return [s for s in snaps if (s["name"],) in list(result)]
