# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
Landscape embeds the Salesforce support cases system in an IFRAME.  It doesn't
connect directly to Salesforce for authentication.  Instead, it connects to the
self-service component (SSP) of the Salesforce integration proxy (SFI proxy)
and requests an authenticated URL using a stored self-service contact ID.
Access to this proxy in Landscape is provided by the ISalesforceConnector
utility.
"""

from future import standard_library

standard_library.install_aliases()
from builtins import str
from builtins import object
import logging
import time
import pycurl
import urllib.request
import urllib.parse
import urllib.error
import os
from io import StringIO
from xmlrpc.client import ServerProxy, Fault, Transport, ProtocolError

from zope.interface import implementer

from canonical.landscape.model.support.interfaces import ISalesforceConnector


__all__ = [
    "NotEntitledError",
    "SalesforceTimeoutError",
    "get_sanitized_url",
    "SalesforceConnector",
]


__clearlist__ = ["NotEntitledError", "SalesforceTimeoutError", "get_sanitized_url"]


class NotEntitledError(Exception):
    """
    Exception raised when a user doesn't have access to support cases.
    """


class SalesforceTimeoutError(Exception):
    """
    Exception raised when we get a timeout error while contacting
    the SFI proxy.
    """


@implementer(ISalesforceConnector)
class SalesforceConnector(object):
    def __init__(self, url, session_timeout, server_factory=None):
        """
        @param url: The Saleforce integration (SFI) proxy URL.
        @param session_timeout: Timeout is used to determine when to
            get a fresh support cases URL from the SFI proxy.
        @param server_factory: Optionally, a callable that takes the
            SFI proxy URL argument and returns an
            C{xmlrpclib.ServerProxy}-like object.
        """
        server_factory = server_factory or ServerProxy
        self._server = server_factory(url, PyCURLTransport(url))
        self.url = url
        self.session_timeout = session_timeout

    def get_support_cases_url(self, salesforce_key, cache=None, create_time=time.time):
        """
        Authenticate C{salesforce_key} with the self-service proxy
        (SSP) system.

        @param salesforce_key: The contact ID to get a URL for.
        @param cache: Optionally, a dict-like object for cached data.
           This is used to keep track of the last request time and
           support cases URL.  Timing data here is used to determine
           whether or not to request a URL from the SFI proxy.
        @param create_time: Optionally, a function taking no paramters
           that returns the current time.
        @return: A session-specific URL for access to the support
            cases system or C{None} if a session can't be created.
        """
        if cache is not None:
            now = create_time()
            data = cache.get(str(salesforce_key))
            if data:
                url = data["cases_url"]
                refresh_time = data["cases_refresh_time"]
                if now < (refresh_time + self.session_timeout):
                    return url

        try:
            result = self._server.getSelfServiceSession(salesforce_key)
        except Fault as f:
            if f.faultCode == "NotFound":
                logging.warning(
                    "Self-service user not found for SFID %s" % (salesforce_key,)
                )
                raise NotEntitledError(f.faultString)
            elif f.faultCode == "NotEntitled":
                logging.warning(
                    "Self-service user not entitled to support "
                    "for SFID %s" % (salesforce_key,)
                )
                raise NotEntitledError(f.faultString)
            else:
                logging.getLogger("SiteError").exception(f)
                return None
        except ProtocolError as e:
            if e.errcode == 28:
                # "Operation Timeout", according to curl's documentation
                logging.warning(
                    "Timeout when retrieving self-service URL for "
                    "SFID %s from %s" % (salesforce_key, e.url)
                )
                raise SalesforceTimeoutError()
            else:
                logging.getLogger("SiteError").exception(e)
                return None
        except Exception as e:
            logging.getLogger("SiteError").exception(e)
            return None

        if result:
            url = result["url"]
            sanitized_url = None
            if url:
                sanitized_url = get_sanitized_url(url)
            logging.info(
                "Received self-service URL %s for SFID %s"
                % (sanitized_url, salesforce_key)
            )
            if cache is not None:
                cache[str(salesforce_key)] = {
                    "cases_url": url,
                    "cases_refresh_time": now,
                }
            return url
        logging.warning("Could not get a self-service URL for SFID %s" % salesforce_key)
        return None


class PyCURLTransport(Transport):
    """Handles a cURL HTTP transaction to an XML-RPC server."""

    def __init__(self, url, timeout=15):
        Transport.__init__(self)

        self._timeout = timeout
        # Flag needed for xmlrpclib
        self.verbose = False
        self.url = url

    def request(self, host, handler, request_body, verbose=False):
        """Performs actual request"""
        data = StringIO()

        curl = pycurl.Curl()
        # Suppress signals
        curl.setopt(pycurl.NOSIGNAL, True)
        # Follow redirects
        curl.setopt(pycurl.FOLLOWLOCATION, True)
        curl.setopt(pycurl.MAXREDIRS, 5)

        url_type = urllib.parse.splittype(self.url)[0]
        # Set timeouts
        if self._timeout:
            curl.setopt(pycurl.CONNECTTIMEOUT, self._timeout)
            curl.setopt(pycurl.TIMEOUT, self._timeout)
        curl.setopt(pycurl.URL, str("%s://%s%s" % (url_type, host, handler)))
        # Unfortunately the proxy certificate is self-signed
        curl.setopt(pycurl.SSL_VERIFYPEER, False)
        curl.setopt(pycurl.POST, True)
        curl.setopt(pycurl.POSTFIELDS, request_body)
        curl.setopt(pycurl.WRITEFUNCTION, data.write)
        curl.setopt(pycurl.VERBOSE, verbose)
        self.verbose = verbose
        try:
            curl.perform()
            status = curl.getinfo(pycurl.RESPONSE_CODE)
        except pycurl.error as err:
            raise ProtocolError(host + handler, err[0], err[1], None)

        if status != 200:
            raise ProtocolError(host + handler, status, data.getvalue(), None)

        data.seek(0)
        return self.parse_response(data)


def get_sanitized_url(url):
    """
    We only want to log the real salesforce URL when debugging a problem.
    In the default case, with debugging disabled, we have to sanitize
    the URL because it allows temporary access to the customer's support
    cases.
    """
    if "SALESFORCE_DEBUG_URL" in os.environ:
        return url
    return "<sanitized>"
