# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
This module defines classes used by storm to represent PostgreSQL operators and
functions that are used by the knowledge base code.
"""
from builtins import range
from builtins import object
import re

from storm.expr import CompoundOper, NamedFunc, Coalesce

from canonical.landscape.model.knowledge.parser import create_wiki_parser


__all__ = [
    "Match",
    "TSRankCD",
    "TSHeadline",
    "ToTSQuery",
    "PlainToTSQuery",
    "ToTSVector",
    "Replace",
    "Text",
    "Strip",
    "get_search_query",
    "AND",
    "OR",
    "RAW",
    "TEXT_SEARCH_CONFIG",
    "RANK_NO_NORMAL",
    "RANK_LOG_LENGTH",
    "RANK_LENGTH",
    "RANK_HARMONIC_MEAN",
    "RANK_UNIQUE_WORDS",
    "RANK_LOG_UNIQUE_WORDS",
    "RANK_BY_RANK",
    "SIMPLE_SEARCH_CONFIG",
    "Similarity",
    "TSVectorValidator",
    "TSVectorPart",
    "get_text_from_wiki",
]


AND = "AND"
OR = "OR"
RAW = "RAW"

# PostgreSQL has a text search ranking function that takes an additional
# parameter which determines the type of normalization to perform on the rank.
# The parameters are integers and defined here for clarity and for importing to
# other Python code.
#
# See /docs/8.3/static/textsearch-controls.html#TEXTSEARCH-RANKING from
# http://www.postgresql.org
RANK_NO_NORMAL = 0
RANK_LOG_LENGTH = 1
RANK_LENGTH = 2
RANK_HARMONIC_MEAN = 4
RANK_UNIQUE_WORDS = 8
RANK_LOG_UNIQUE_WORDS = 16
RANK_BY_RANK = 32

# This is the knowledge-base-specific language configuration for full text
# search.
TEXT_SEARCH_CONFIG = "public.landscape_english"

# This is used for e.g. hardware search, where stemming etc. probably aren't
# desirable.
SIMPLE_SEARCH_CONFIG = "simple"


class Match(CompoundOper):
    """
    The full text match operator is used to compare a compiled search
    (tsquery) expression to a text search vector (tsvector). In PostgreSQL, the
    operator returns a "true" value if the tsvector matches the tsquery.
    """

    __slots__ = ()
    oper = "@@"


class TSRankCD(NamedFunc):
    """
    This full text search rank is a PostgreSQL function that computes the cover
    density of the search terms in the given text search vector.

    SQL usage is the following:
        ts_rank_cd(
            [ weights float4[], ]
            vector tsvector,
            query tsquery [,
            normalization integer ])

    This function returns float4.
    """

    __slots__ = ()
    name = "TS_RANK_CD"


class TSHeadline(NamedFunc):
    """
    This full text search function shows fragments of the document with marked
    search terms.

    SQL usage is the following:
        ts_headline(
            [ config regconfig, ]
            document text,
            query tsquery
            [, options text ])

    This function returns text with highlighted words wrapped in <b></b> by
    default.
    """

    __slots__ = ()
    name = "TS_HEADLINE"


class ToTSQuery(NamedFunc):
    """
    This function converts user-written search expressions into a proper
    PostgreSQL full text search query, doing such things as normalizing words
    in the search terms. The correct format produced by this function is needed
    in order to match a text search query against a text search vector.
    """

    __slots__ = ()
    name = "TO_TSQUERY"


class PlainToTSQuery(NamedFunc):
    """
    This function transforms unformatted text to a tsquery. The text is
    parsed and normalized much as for to_tsvector, then the AND operator "&"
    is inserted between surviving words.
    """

    __slots__ = ()
    name = "PLAINTO_TSQUERY"


class TSQueryAND(CompoundOper):
    """Combine two or more TSQuery objects with an AND operator"""

    __slots__ = ()
    oper = "&&"


class TSQueryOR(CompoundOper):
    """Combine two or more TSQuery objects with an OR operator"""

    __slots__ = ()
    oper = "||"


class ToTSVector(NamedFunc):
    """
    Parses and normalizes a document string, converting text into a proper
    PostgreSQL tsquery data type.
    """

    __slots__ = ()
    name = "TO_TSVECTOR"


class SetWeight(NamedFunc):
    """
    Sets the weight of a TSVector.
    """

    __slots__ = ()
    name = "SETWEIGHT"


class Replace(NamedFunc):
    """String replacement."""

    __slots__ = ()
    name = "REPLACE"


class Text(NamedFunc):
    """Cast a given type as text."""

    __slots__ = ()
    name = "TEXT"


class Strip(NamedFunc):
    """Strip a string."""

    __slots__ = ()
    name = "STRIP"


class Concat(CompoundOper):
    """Concatenate multiple strings"""

    __slots__ = ()
    oper = "||"


class Similarity(NamedFunc):
    """Calculate trigram similarity"""

    __slots__ = ()
    name = "similarity"


class TSVectorPart(object):
    """Describes how a model attribute is converted to a document part.

    Each document part can have different weight, and optionally a
    list of callables that should be applied to prepare the attribute
    value for use in a L{TSVector}.

    Valid weight labels are "A", "B", "C", "D", and defaults to "D"
    if none provided.

    Default weights for each one are "1.0", "0.5", "0.2", "0.1", in
    that order.
    """

    def __init__(self, name, weight="D", transform=None, config=TEXT_SEARCH_CONFIG):
        self.name = name
        self._config = config
        self._weight = weight
        self._transform = transform

    def __call__(self, value=None):
        """
        Transform the value provided, returning a TSVector with the
        specified weight.
        """
        if self._transform is not None:
            for transform in self._transform:
                value = transform(value)
        return SetWeight(ToTSVector(self._config, Coalesce(value, "")), self._weight)


class TSVectorValidator(object):
    """
    This validator object is a trick; we are not checking values nor
    changing the resultant value of the passed attribute. Instead, we
    are using the validation mechanism to set custom values in the
    database, in particular, values needed in order to perform full
    text search. This obviates the need for a database trigger, and
    ultimately, allows us to do more in Python than the database
    trigger could do easily.
    """

    def __init__(self, name, *parts):
        self._name = name
        self._parts = parts or None

    def __call__(self, model_object, attr=None, value=None):
        vector = self.get(model_object, attr=attr, value=value)
        setattr(model_object, self._name, vector)
        return value

    def get(self, model_object, attr=None, value=None):
        vectors = []

        if self._parts is not None:
            for part in self._parts:
                if part.name == attr:
                    part_value = value
                else:
                    part_value = getattr(model_object, part.name)
                if part_value is not None:
                    vectors.append(part(part_value))

        if vectors:
            return Concat(*vectors)
        return ToTSVector("")


def get_search_query(search_terms, search_mode=AND, search_config=TEXT_SEARCH_CONFIG):
    """
    This utility function separates the logic for parsing user input. For now,
    this is fairly simple, providing three basic uses, as defined below.

    @param search_terms: this can either be a list of individual words which
        need to be joined to form a search expression or can be a search
        expression itself, this being determined by the search_mode parameter.

    @param search_mode: search_mode can be one of three things:
        OR - all terms are joined with an OR operator "|" to form a search
        expression suitable for ts_query.

        AND - all terms are passed to the PostgreSQL plainto_tsquery function,
        which normalizes the search terms, ANDing the results.

        RAW - the search_terms variable is treated like a search expression,
        ready to be passed to ts_query. In such an expression, multiple search
        terms are combined with "&", "|", and "!" and can be grouped with
        parentheses to enforce precedence. In the absence of parentheses, "!"
        binds most tightly, and "&" binds more tightly than "|".
    """
    if search_mode == RAW:
        query_function = ToTSQuery
        search_expression = search_terms
    elif search_mode == OR:
        query_function = ToTSQuery
        # Convert the vector to text.
        text = Text(Strip(ToTSVector(search_terms)))
        # Remove the single quotes that the vector put around terms.
        text = Replace(text, "'", "")
        # Insert the OR operator between surviving search terms.
        search_expression = Replace(text, " ", " | ")
    elif search_mode == AND:
        query_function = PlainToTSQuery
        search_expression = search_terms
    return query_function(search_config, search_expression)


def get_text_from_wiki(text, format="text"):
    """Cleanup text from wiki formatting."""
    parser = create_wiki_parser(format=format)
    return parser(text).strip().decode("utf-8")
