# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import Unicode, Bool, Int

from canonical.landscape.model.activity.computer import ComputerRequest


__all__ = [
    "CreateUserRequest",
    "EditUserRequest",
    "DeleteUserRequest",
    "LockUserRequest",
    "UnlockUserRequest",
    "CreateGroupRequest",
    "AddGroupMemberRequest",
    "RemoveGroupMemberRequest",
]


__clearlist__ = ["CreateUserRequest", "CreateGroupRequest", "AddGroupMemberRequest"]


class CreateUserRequest(ComputerRequest):

    __storm_table__ = "create_user_request"

    username = Unicode(allow_none=False)
    name = Unicode()
    passphrase = Unicode()
    require_passphrase_reset = Bool()
    primary_groupname = Unicode()
    location = Unicode()
    home_phone = Unicode()
    work_phone = Unicode()

    def __init__(
        self,
        activity,
        computer,
        username,
        name=None,
        passphrase=None,
        require_passphrase_reset=False,
        primary_groupname=None,
        location=None,
        home_phone=None,
        work_phone=None,
    ):
        self.username = username
        self.name = name
        self.passphrase = passphrase
        self.require_passphrase_reset = require_passphrase_reset
        self.primary_groupname = primary_groupname
        self.location = location
        self.home_phone = home_phone
        self.work_phone = work_phone
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(CreateUserRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Create user %s" % self.username

    def _started(self):
        message = {
            "type": "add-user",
            "operation-id": self.activity.id,
            "username": self.username,
            "name": self.name,
            "password": self.passphrase,
            "require-password-reset": self.require_passphrase_reset,
            "primary-group-name": self.primary_groupname,
            "location": self.location,
            "home-number": self.home_phone,
            "work-number": self.work_phone,
        }
        self.computer.get_message_store().add(message)


class EditUserRequest(ComputerRequest):

    __storm_table__ = "edit_user_request"

    uid = Int(allow_none=False)
    username = Unicode(allow_none=False)
    primary_groupname = Unicode()
    name = Unicode()
    passphrase = Unicode()
    location = Unicode()
    work_phone = Unicode()
    home_phone = Unicode()

    def __init__(
        self,
        activity,
        computer,
        uid,
        username,
        name=None,
        primary_groupname=None,
        passphrase=None,
        location=None,
        home_phone=None,
        work_phone=None,
    ):
        self.uid = uid
        self.username = username
        self.name = name
        self.primary_groupname = primary_groupname
        self.passphrase = passphrase
        self.location = location
        self.work_phone = work_phone
        self.home_phone = home_phone
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(EditUserRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Edit user %s (UID %d)" % (self.username, self.uid)

    def get_text(self):
        if self.activity.result_code == 500:
            return "Landscape client must be upgraded to " "support this operation."
        return super(EditUserRequest, self).get_text()

    def _started(self):
        message_store = self.computer.get_message_store()
        api = self.computer.current_version
        api = api and api.decode()
        # If the client won't understand a primary group edit, then fail
        # this operation.
        if api in ["2.0", "2.1"] and self.primary_groupname is not None:
            self.activity.fail(code=500)
        elif api == "2.0":
            message_store.add(
                {
                    "type": "set-user-details",
                    "username": self.username,
                    "name": self.name,
                    "password": self.passphrase,
                    "location": self.location,
                    "work-number": self.work_phone,
                    "home-number": self.home_phone,
                    "operation-id": self.activity.id,
                }
            )
        elif api == "2.1":
            message_store.add(
                {
                    "type": "edit-user",
                    "uid": self.uid,
                    "name": self.name,
                    "password": self.passphrase,
                    "location": self.location,
                    "work-number": self.work_phone,
                    "home-number": self.home_phone,
                    "operation-id": self.activity.id,
                }
            )
        else:
            message_store.add(
                {
                    "type": "edit-user",
                    "uid": self.uid,
                    "username": self.username,
                    "name": self.name,
                    "password": self.passphrase,
                    "location": self.location,
                    "work-number": self.work_phone,
                    "home-number": self.home_phone,
                    "operation-id": self.activity.id,
                    "primary-group-name": self.primary_groupname,
                }
            )


class DeleteUserRequest(ComputerRequest):

    __storm_table__ = "delete_user_request"

    uid = Int()
    username = Unicode(allow_none=False)
    delete_home = Bool(allow_none=False)

    def __init__(self, activity, computer, uid, username, delete_home):
        self.uid = uid
        self.username = username
        self.delete_home = delete_home
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(DeleteUserRequest, self).__init__(activity, computer)

    def get_summary(self):
        if self.uid is not None:
            return "Delete user %s (UID %d)" % (self.username, self.uid)
        return "Delete user %s" % self.username

    def _started(self):
        message_store = self.computer.get_message_store()
        message_store.add(
            {
                "type": "remove-user",
                "username": self.username,
                "uid": self.uid,
                "delete-home": self.delete_home,
                "operation-id": self.activity.id,
            }
        )


class LockUserRequest(ComputerRequest):

    __storm_table__ = "lock_user_request"

    uid = Int(allow_none=False)
    username = Unicode(allow_none=False)

    def __init__(self, activity, computer, uid, username):
        self.uid = uid
        self.username = username
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(LockUserRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Lock out user %s (UID %d)" % (self.username, self.uid)

    def _started(self):
        message_store = self.computer.get_message_store()
        message_store.add(
            {
                "type": "lock-user",
                "username": self.username,
                "uid": self.uid,
                "operation-id": self.activity.id,
            }
        )


class UnlockUserRequest(ComputerRequest):

    __storm_table__ = "unlock_user_request"

    uid = Int(allow_none=False)
    username = Unicode(allow_none=False)

    def __init__(self, activity, computer, uid, username):
        self.uid = uid
        self.username = username
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(UnlockUserRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Unlock user %s (UID %d)" % (self.username, self.uid)

    def _started(self):
        message_store = self.computer.get_message_store()
        message_store.add(
            {
                "type": "unlock-user",
                "username": self.username,
                "uid": self.uid,
                "operation-id": self.activity.id,
            }
        )


class CreateGroupRequest(ComputerRequest):

    __storm_table__ = "create_group_request"

    groupname = Unicode(allow_none=False)
    system = Bool()

    def __init__(self, activity, computer, groupname, system=False):
        self.groupname = groupname
        self.system = system
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(CreateGroupRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Create group %s" % self.groupname

    def _started(self):
        message_store = self.computer.get_message_store()
        api = self.computer.current_version
        api = api and api.decode()
        if api == "2.0":
            message_store.add(
                {
                    "type": "create-group",
                    "name": self.groupname,
                    "system-group": self.system,
                    "operation-id": self.activity.id,
                }
            )
        elif api == "2.1":
            message_store.add(
                {
                    "type": "add-group",
                    "name": self.groupname,
                    "system-group": self.system,
                    "operation-id": self.activity.id,
                }
            )
        else:
            message_store.add(
                {
                    "type": "add-group",
                    "groupname": self.groupname,
                    "system-group": self.system,
                    "operation-id": self.activity.id,
                }
            )


class AddGroupMemberRequest(ComputerRequest):

    __storm_table__ = "add_group_member_request"

    username = Unicode(allow_none=False)
    groupname = Unicode(allow_none=False)
    uid = Int()
    gid = Int()

    def __init__(self, activity, computer, username, groupname, uid=None, gid=None):
        self.username = username
        self.groupname = groupname
        self.uid = uid
        self.gid = gid
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(AddGroupMemberRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Add user %s to group %s" % (self.username, self.groupname)

    def _started(self):
        api = self.computer.current_version
        api = api and api.decode()
        if api == "2.0":
            message = {
                "type": "add-group-member",
                "member": self.username,
                "group": self.groupname,
                "operation-id": self.activity.id,
            }
        elif api == "2.1":
            message = {
                "type": "add-group-member",
                "username": self.username,
                "groupname": self.groupname,
                "uid": self.uid,
                "gid": self.gid,
                "operation-id": self.activity.id,
            }
        else:
            message = {
                "type": "add-group-member",
                "username": self.username,
                "groupname": self.groupname,
                "operation-id": self.activity.id,
            }
        self.computer.get_message_store().add(message)


class RemoveGroupMemberRequest(ComputerRequest):

    __storm_table__ = "remove_group_member_request"

    username = Unicode(allow_none=False)
    groupname = Unicode(allow_none=False)
    uid = Int(allow_none=False)
    gid = Int(allow_none=False)

    def __init__(self, activity, computer, username, groupname, uid, gid):
        self.username = username
        self.groupname = groupname
        self.uid = uid
        self.gid = gid
        # Call the base class constructor last, so that our own custom
        # attributes are fully initialized, because Storm could end up
        # causing a flush mid-initialization.
        super(RemoveGroupMemberRequest, self).__init__(activity, computer)

    def get_summary(self):
        return "Remove user %s from group %s" % (self.username, self.groupname)

    def _started(self):
        api = self.computer.current_version
        api = api and api.decode()
        message_store = self.computer.get_message_store()
        if api == "2.0":
            message_store.add(
                {
                    "type": "remove-group-member",
                    "member": self.username,
                    "group": self.groupname,
                    "operation-id": self.activity.id,
                }
            )
        elif api == "2.1":
            message_store.add(
                {
                    "type": "remove-group-member",
                    "username": self.username,
                    "groupname": self.groupname,
                    "uid": self.uid,
                    "gid": self.gid,
                    "operation-id": self.activity.id,
                }
            )
        else:
            message_store.add(
                {
                    "type": "remove-group-member",
                    "username": self.username,
                    "groupname": self.groupname,
                    "operation-id": self.activity.id,
                }
            )
