# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer
from zope.security.proxy import removeSecurityProxy

from canonical.security.protect import requires, requires_advisor, PRIVATE
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.resource.user import User
from canonical.landscape.model.resource.group import Group, UserGroupLink, PendingGroup
from canonical.landscape.model.activity.group import (
    ADD_USERS_SUMMARY,
    DELETE_USERS_SUMMARY,
    EDIT_USERS_SUMMARY,
    LOCK_USERS_SUMMARY,
    UNLOCK_USERS_SUMMARY,
    REMOVE_FROM_GROUPS_SUMMARY,
    ADD_TO_GROUPS_SUMMARY,
)


__all__ = ["UserSelection", "UserManagement"]

__clearlist__ = ["UserSelection", "UserManagement"]

__requires_prefix__ = "landscape."


@requires_advisor
@implementer(IAccessDelegateAware)
class UserSelection(object):
    # XXX UserSelection does not provide methods for selecting users, but
    # requires them to be passed in to the constructor. Perhaps it should
    # be changed to be more similar to ComputerSelection.
    """
    A representation of multiple users across multiple machines which
    provides operations for managing those users by sending messages
    to the machines.

    The methods that send messages to clients to perform user
    management operations will create L{Command} objencts. They create
    one Command per *username* when performing a given operation.
    """

    requires("ViewUserManagement", PRIVATE, ["users", "management", "account"])

    def __init__(self, management, users=None):
        if users is None:
            users = []
        self.users = sorted(users, key=lambda u: (u.computer.title, u.username))
        self.management = management
        self.account = self.management.selection.account

    @property
    def access_delegate(self):
        return self.management.selection

    @requires("RemoveUser")
    def delete_users(self, delete_home):
        person = self.management.person
        parent = self.management.selection.create_activity_group(
            creator=person, summary=DELETE_USERS_SUMMARY
        )
        for user in self.users:
            user.request_delete_user(parent, delete_home)
        parent.info.create_log_entry("Group activity created", person=person)
        parent.settle()

    @requires("EditUser")
    def lock_users(self):
        person = self.management.person
        parent = self.management.selection.create_activity_group(
            creator=person, summary=LOCK_USERS_SUMMARY
        )
        for user in self.users:
            user.request_lock_user(parent)
        parent.settle()
        parent.info.create_log_entry("Group activity created", person=person)

    @requires("EditUser")
    def unlock_users(self):
        person = self.management.person
        parent = self.management.selection.create_activity_group(
            creator=person, summary=UNLOCK_USERS_SUMMARY
        )
        for user in self.users:
            user.request_unlock_user(parent)
        parent.settle()
        parent.info.create_log_entry("Group activity created", person=person)

    @requires("EditUser")
    def edit_users(
        self,
        primary_groupname=None,
        name=None,
        passphrase=None,
        location=None,
        work_phone=None,
        home_phone=None,
    ):
        person = self.management.person
        parent = self.management.selection.create_activity_group(
            creator=person, summary=EDIT_USERS_SUMMARY
        )
        for user in self.users:
            if primary_groupname:
                group = user.computer.get_group_by_name(primary_groupname)
                if not group:
                    user.computer.request_create_group(parent, primary_groupname)
            user.request_edit_user(
                parent,
                name=name,
                primary_groupname=primary_groupname,
                passphrase=passphrase,
                location=location,
                work_phone=work_phone,
                home_phone=home_phone,
            )
        parent.settle()
        parent.info.create_log_entry("Group activity created", person=person)
        return parent

    @requires("EditUser")
    def add_to_groups(self, groupnames):
        """Request to add the selected users to the named groups.

        If a group does not exist on any given computer that a user is
        being added to, the group will be implicitly created.
        """
        parent = None
        person = self.management.person
        for user in self.users:
            computer = user.computer
            for groupname in sorted(groupnames):
                group = computer.get_group_by_name(groupname)
                if not group or user not in group.members:
                    if parent is None:
                        selection = self.management.selection
                        parent = selection.create_activity_group(
                            creator=person, summary=ADD_TO_GROUPS_SUMMARY
                        )
                    if not group:
                        group = PendingGroup(computer, groupname)
                        computer.request_create_group(parent, groupname)
                    group.request_add_member(parent, user.username, uid=user.uid)
        if parent is not None:
            parent.settle()
            parent.info.create_log_entry("Group activity created", person=person)

    @requires("EditUser")
    def remove_from_groups(self, groupnames):
        """Request to remove the selected users from the named groups."""
        parent = None
        person = self.management.person
        for user in self.users:
            computer = user.computer
            for groupname in sorted(groupnames):
                group = computer.get_group_by_name(groupname)
                if group and user in group.members:
                    if parent is None:
                        selection = self.management.selection
                        parent = selection.create_activity_group(
                            creator=person, summary=REMOVE_FROM_GROUPS_SUMMARY
                        )
                    group.request_remove_member(parent, user.uid, user.username)
        if parent is not None:
            parent.settle()
            parent.info.create_log_entry("Group activity created", person=person)

    def _get_resource_store_data(self, index=None):
        # FIXME This code is duplicated in temperature.py and probably
        # elsewhere.  It ought to be centralized.
        resource_stores = {}
        computer_twins = {}

        # Get all resource database stores, and which computers are in them.
        if index is not None:
            offset = index - 1
            limit = 1
        else:
            offset = None
            limit = None
        selection = self.management.selection.get_selected(offset=offset, limit=limit)
        for computer in selection:
            rdb_id = computer.resource_database_key
            if rdb_id not in resource_stores:
                unsafe_computer = removeSecurityProxy(computer)
                resource_stores[rdb_id] = unsafe_computer.store
                computer_twins[rdb_id] = []
            computer_twins[rdb_id].append(computer.id)
        return resource_stores, computer_twins

    @requires("ViewGroup")
    def all_groups(self, index=None):
        """Get the union of groups available on all selected computers.

        @param index: If not C{None} return only data about the selected
            computer at the given index.
        @return: The set of all groups, as strings, that are available for all
                 selected computers from whatever resource databases are in
                 use by those computers.
        """
        stores, computers = self._get_resource_store_data(index=index)
        groups = set()
        for rdb_id in stores:
            store = stores[rdb_id]
            computer_ids = computers[rdb_id]
            result = store.find(Group, Group.computer_id.is_in(computer_ids))
            result.config(distinct=True)
            groups.update(result.values(Group.name))
        return groups

    @requires("ViewUserManagement")
    def get_computer_group_user_tuples(self, index=None):
        """
        Get 3-tuples of the form C{(Computer, Group.name, User.username)} for
        the current selection.

        @param index: If not C{None} return only data about the selected
            computer at the given index.
        """
        usernames = [user.username for user in self.users]
        stores, computers = self._get_resource_store_data(index=index)
        for rdb_id in stores:
            store = stores[rdb_id]
            computer_ids = computers[rdb_id]
            result = store.find(
                (Group, User),
                UserGroupLink.group_id == Group.id,
                UserGroupLink.user_id == User.id,
                Group.computer_id.is_in(computer_ids),
                User.username.is_in(usernames),
            )
            for group, user in result:
                assert group and user
                yield group.computer, group.name, user.username

    @requires("ViewUserManagement")
    def get_computers(self):
        """
        Returns a set of computers.  Computers are only returned if
        they match a user in the selection.  This may be different
        from the computers in the computer selection.
        """
        computers = set()
        for user in self.users:
            computers.add(user.computer)
        return computers


@requires_advisor
@implementer(IAccessDelegateAware)
class UserManagement(object):
    """Represents the user management features of Landscape."""

    requires("ViewUserManagement", PRIVATE, ["selection", "person"])

    def __init__(self, selection, person):
        self.selection = selection
        self.person = person

    @property
    def access_delegate(self):
        return self.selection

    @requires("RequestCreateUser")
    def create_user(
        self,
        username,
        name,
        password,
        require_passphrase_reset=False,
        primary_groupname=None,
        location=None,
        home_phone=None,
        work_phone=None,
        offset=None,
        limit=None,
    ):
        """Create a user on each of the selected computers.

        @return: The parent L{Activity} created to contain individual child
            activities created to create users on selected computers.
        """
        parent = self.selection.create_activity_group(
            creator=self.person, summary=ADD_USERS_SUMMARY
        )
        for computer in self.selection.get_selected(offset=offset, limit=limit):

            computer.request_create_user(
                parent,
                username,
                name,
                password,
                require_passphrase_reset,
                primary_groupname=primary_groupname,
                location=location,
                home_phone=home_phone,
                work_phone=work_phone,
            )
        parent.settle()
        parent.info.create_log_entry("Group activity created", person=self.person)
        return parent
