# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
import time
from datetime import datetime
from operator import attrgetter

import transaction

from zope.security.proxy import removeSecurityProxy

from storm.locals import DateTime, Int, Not
from storm.expr import Column, SQLRaw

from canonical.lib.retry import retry, uniform_exponential
from canonical.landscape.model.package.package import Package
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.usn import USN, USNName
from canonical.landscape.model.resource.package import ComputerPackages, ArrayOverlap
from canonical.landscape.model.resource.store import (
    get_resource_stores_for_computer_ids,
    get_resource_stores,
)

__all__ = ["USNIssue", "USNIssueManagement", "update_usn_issues"]


class USNIssue(object):
    """Represents a USN affecting a computer.

    @type name: C{str}
    @ivar name: The USN identifier (like "123-4").

    @ivar summary: The summary from the USN.
    """

    __storm_table__ = "usn_issue"
    __storm_primary__ = "usn_name", "computer_id"

    usn_id = Int(allow_none=False)
    usn_name = USNName(name="usn_id")
    computer_id = Int(allow_none=False)
    resolve_time = DateTime()

    def __init__(self, usn_name, computer_id):
        self.usn_name = usn_name
        self.computer_id = computer_id

    def __lt__(self, other):
        if not isinstance(other, USNIssue):
            return True

        return self.usn_id < other.usn_id


class USNIssueManagement(object):
    """Manage USNIssues for a all computers in a resource store."""

    def __init__(self, selection=None, only_computers=False):
        """
        @param only_computers: If specified, the methods C{get_open_issues},
            C{get_resolved_issues} and C{get_issues} will only return computer
            IDs instead of C{USNIssue}s.
        """
        self.selection = selection
        self._only_computers = only_computers

    def _get_resources_stores(self):
        """Get a list of resource stores and computer ids that are in the store

        A list of tuples (store, computer_ids) are returned. If
        computer_ids is None, it means that all computers should be
        considered in the store, i.e. that there is no selection.
        """
        if self.selection is not None:
            computer_ids = self.selection.get_selected_computer_ids()
            return list(get_resource_stores_for_computer_ids(computer_ids).items())
        else:
            return [(store, None) for _, store in get_resource_stores()]

    def _find(self, *extra_where):
        """Find C{USNIssue} records for the selected computers.

        A list of records are returned, since multiple stores may have
        to be queried.

        @param *extra_where: List of conditions to be passed to C{store.find()}
        """
        result = []
        for store, computer_ids in self._get_resources_stores():
            where = list(extra_where)
            if computer_ids is not None:
                where.append(USNIssue.computer_id.is_in(computer_ids))
            resultset = store.find(USNIssue, *where)
            if self._only_computers:
                resultset = resultset.values(USNIssue.computer_id)
            result.extend(list(resultset))

        return sorted(result)

    def get_open_issues(self, usn_name):
        """Return open C{USNIssue}s for the given usn name."""
        return self._find(USNIssue.usn_name == usn_name, USNIssue.resolve_time == None)

    def get_open_issues_for_computers(self, computer_ids):
        """Return open C{USNIssue}s for computer ids."""
        return self._find(
            USNIssue.computer_id.is_in(computer_ids), USNIssue.resolve_time == None
        )

    def get_resolved_issues(self, usn_names, resolved_before=None, resolved_after=None):
        """
        Retrieve L{USNIssue}s of a list of C{usn_names} for the computers in
        the selection, with by C{resolve_time} between C{resolved_before} and
        C{resolved_after} if specified.
        """
        where = [USNIssue.usn_name.is_in(usn_names)]
        if resolved_before is not None:
            where.append(USNIssue.resolve_time < resolved_before)
        if resolved_after is not None:
            where.append(USNIssue.resolve_time > resolved_after)
        if resolved_after is None and resolved_before is None:
            where.append(USNIssue.resolve_time != None)
        return self._find(*where)

    def get_issues(self, usn_names):
        """Get all C{USNIssue}s for the given USNs."""
        return self._find(USNIssue.usn_name.is_in(usn_names))

    def set_affected_computers(self, usn_name, computer_ids):
        """Indicate which computers are affected by a USN.

        If a computer id is in the store but is not given here, it
        will be considered as not being affected. Thus if there is an
        open C{USNIssue} it will be marked as resolved.

        If there isn't a C{USNIssue} for a given computer, one will be
        created. If a closed C{USNIssue} exists, it will be reopened.
        """
        open_issues = self.get_open_issues(usn_name)
        for issue in open_issues:
            if issue.computer_id not in computer_ids:
                issue.resolve_time = datetime.utcnow()
        computer_stores = get_resource_stores_for_computer_ids(computer_ids)
        for store, computer_ids in list(computer_stores.items()):
            issues = store.find(
                USNIssue,
                USNIssue.usn_name == usn_name,
                USNIssue.computer_id.is_in(computer_ids),
            )
            existing_issues = dict((issue.computer_id, issue) for issue in issues)
            for computer_id in computer_ids:
                existing_issue = existing_issues.get(computer_id)
                if existing_issue is None:
                    new_issue = USNIssue(usn_name, computer_id)
                    store.add(new_issue)
                else:
                    existing_issue.resolve_time = None

    def check_usn(self, usn_name):
        """Check which computers are affected by the given USN name.

        C{USNIssue}s will be created for affected computers that don't
        have a record yet, and unresolved C{USNIssue}s will be marked as
        resolved for computers that aren't affected by the USN anymore.
        """
        package_store = get_package_store()
        package_ids = set(package_store.find(Package.id, Package.usn_name == usn_name))
        sql_package_ids = SQLRaw(
            "'{%s}'" % (",".join([str(package_id) for package_id in package_ids]),)
        )
        base_where = [ArrayOverlap(sql_package_ids, Column("available_upgrades"))]
        affected_computers = set()
        for store, computer_ids in self._get_resources_stores():
            where = list(base_where)
            if computer_ids is not None:
                where.append(ComputerPackages.computer_id.is_in(computer_ids))
            result = store.find(ComputerPackages.computer_id, *where)
            for computer_id in result:
                affected_computers.add(computer_id)
        self.set_affected_computers(usn_name, affected_computers)

    def check_usns_for_computer(self, computer):
        """
        Check the state of the USNs applicable to the given computer

        This is made for the single computer case, so it probably doesn't scale
        well. Look at C{check_usn} if you need to check a lot of computers.
        """
        package_store = get_package_store()
        now = datetime.utcnow()
        store = removeSecurityProxy(computer).store

        packages = removeSecurityProxy(computer.get_packages())

        upgrades = set(packages.get_packages("available_upgrades")[0])

        if upgrades:
            # Retrieve all USNs associated with the available upgrades.
            upgrade_usns = set(
                package_store.find(
                    Package.usn_name,
                    Package.usn_name != None,
                    Package.id.is_in(upgrades),
                )
            )

            # First reset resolve_time for existing USNIssues with packages in
            # upgrades.
            reset_issues = store.find(
                USNIssue,
                USNIssue.usn_name.is_in(upgrade_usns),
                USNIssue.computer_id == computer.id,
            )
            reset_issues.set(resolve_time=None)

            existing_usns = set(reset_issues.values(USNIssue.usn_name))
            # Then create new USNIssue for the remaining list
            for usn in upgrade_usns - existing_usns:
                new_issue = USNIssue(usn, computer.id)
                store.add(new_issue)
        else:
            # There is no available upgrades, resolve all opened USNs.
            upgrade_usns = set()

        # Resolved all opened USNs which don't have any associated upgrades.
        solved_issues = store.find(
            USNIssue,
            USNIssue.resolve_time == None,
            USNIssue.computer_id == computer.id,
            Not(USNIssue.usn_name.is_in(upgrade_usns)),
        )
        solved_issues.set(resolve_time=now)


def update_usn_issues(txn=transaction, batch_size=50, retries=3):
    """Update the USN issues for all computers and all USNs."""
    total_start = time.time()
    logger = logging.getLogger("usn-script")
    package_store = get_package_store()
    usn_names = list(package_store.find(USN).values(USN.name))
    usn_count = len(usn_names)
    issue_management = USNIssueManagement()
    limit = batch_size
    offset = 0

    def update():
        for usn_name in usn_names[offset : offset + limit]:
            start = time.time()
            issue_management.check_usn(usn_name)
            end = time.time()
            logger.debug(
                "Updated USN patch tracking for USN %s in %.2f seconds"
                % (usn_name, end - start)
            )

    while offset < usn_count:
        retry(update, uniform_exponential(retries), transaction=txn)
        offset += limit

    total_end = time.time()
    logger.info(
        "Updated USN patch tracking for %d USNs in %.2f seconds"
        % (usn_count, total_end - total_start)
    )
