# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
import apt_pkg
import os

from twisted.internet.defer import succeed, Deferred

from storm.twisted.transact import transact

from zope.component import getUtility

from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.repository.distribution import (
    get_distribution,
    LockError,
)
from canonical.landscape.model.repository.upload import get_uploads, remove_upload


class Upload(object):
    """
    A serialized version of the RepositoryUpload class with no reference to the
    database model.

    @param upload: the RepositoryUpload class
    """

    def __init__(self, upload):
        self.id = upload.id
        self.account_id = upload.account.id
        self.distribution_id = upload.distribution.id
        self.pocket_id = upload.pocket.id
        self.pocket_state = upload.pocket.get_state()
        self.changes = upload.changes
        self.changes_path = upload.get_changes_path()
        self.reprepro = upload.pocket.get_reprepro()

    def process(self):
        return self.reprepro.processincoming(self.changes)


class IncomingQueue(object):
    """Process uploads in the incoming queue."""

    def __init__(self, transactor, reactor):
        self.transactor = transactor
        self._processing = False
        self._process_again = False
        self._reactor = reactor

    def notify(self):
        """Called by the upload service when a new upload is available."""
        if not self._processing:
            self.process()
        else:
            self._process_again = True

    def process(self):
        """Process all incoming uploads."""
        self._process_again = False
        self._processing = True
        deferred = self._get_uploads()
        deferred.addCallback(self._process_uploads)
        deferred.addCallback(self._check_pending)
        return deferred

    @transact
    def _get_uploads(self):
        """Get all the uploads in the queue."""
        return [Upload(upload) for upload in get_uploads()]

    def _process_uploads(self, uploads):
        """Process all available the package uploads in the queue."""
        deferred = succeed(None)
        for upload in uploads:
            deferred.addCallback(lambda _: self._process_upload(upload=upload))
        return deferred

    def _check_pending(self, _):
        """Re-schedule process if there are still pending uploads."""
        if self._process_again:
            deferred = Deferred()
            deferred.addCallback(lambda _: self.process())
            self._reactor.callLater(10, deferred.callback, None)
            return deferred
        else:
            self._processing = False

    def _process_upload(self, upload):
        """Process a single package upload importing it in the pocket."""

        def handle_success(errors):
            deferred = succeed(upload)
            if errors:
                deferred.addCallback(self._report_errors, errors)
            return deferred.addCallback(self._finish_process_upload, remove=True)

        def handle_failure(failure):
            # We get here only if acquiring the distribution lock fails
            # or if the reprepro process ended with a signal
            self._process_again = True
            if not failure.check(LockError):
                # Unlock the distribution but keep the upload for a later run
                return self._finish_process_upload(upload, remove=False)

        deferred = self._acquire_distribution_lock(upload)
        deferred.addCallback(lambda _: upload.process())
        deferred.addCallback(handle_success)
        deferred.addErrback(handle_failure)
        return deferred

    @transact
    def _finish_process_upload(self, upload, remove=False):
        """Release the distribution lock and possibly delete the upload."""
        account = get_account_by_id(upload.account_id)
        if remove:
            remove_upload(account, upload.id)

        distribution = get_distribution(account, id=upload.distribution_id)
        distribution.unlock()

    @transact
    def _acquire_distribution_lock(self, upload):
        """
        Acquire the lock in distribution associated with the given upload.
        """
        account = get_account_by_id(upload.account_id)
        distribution = get_distribution(account, id=upload.distribution_id)
        distribution.lock()

    # This is run in a transaction as the mailer requires it.  Without it,
    # mails are left in the queue.  Note that this is not unit-tested as there
    # is no easy way to do it.
    @transact
    def _report_errors(self, upload, errors):
        """
        Report errors for failed upload imports

        @param upload: the failed Upload tuple
        @param errors: a dict with import errors
        """
        for section in self._generate_sections(upload.changes_path):
            changed_by = section.get("Changed-By")
            if changed_by is not None:
                mailer = getUtility(ILandscapeMailer)
                subject = "Package import failed for '%s'" % upload.changes

                message = ["The following error(s) occured in package import:\n"]
                if "no-distro" in errors["errors"]:
                    message.append(
                        "* Couldn't find a destination pocket " "for package."
                    )
                if "no-sign" in errors["errors"]:
                    message.append(
                        "* Uploaded data is not signed, but the "
                        "destination pocket requires it"
                    )
                if errors["unverified_files"]:
                    message.append(
                        "* The signature on following file(s) could not "
                        "be verified.  Please make sure to import the GPG "
                        "key(s) used to sign the files or use a different "
                        "key to sign them."
                    )
                    for unverified in errors["unverified_files"]:
                        message.append("  - %s (key id %s)" % unverified)
                if errors["invalid_files"]:
                    message.append(
                        "* The following file(s) are invalid (e.g. "
                        "corrupted, incorrectly signed):"
                    )
                    for invalid in errors["invalid_files"]:
                        message.append("  - %s" % invalid)
                if errors["missing_files"]:
                    message.append(
                        "* The following file(s) listed in the .changes "
                        "file could't be found:"
                    )
                    for missing in errors["missing_files"]:
                        message.append("  - %s" % missing)
                if errors["unknown_arch_files"]:
                    message.append(
                        "* The architecture for the following file(s) "
                        "doesn't match those of the pocket:"
                    )
                    for unknown_arch in errors["unknown_arch_files"]:
                        message.append("  - %s" % unknown_arch)
                if errors["udeb_files"]:
                    message.append(
                        "* The following file(s) couldn't be imported "
                        "as the pocket doesn't support .udeb packages:"
                    )
                    for udeb in errors["udeb_files"]:
                        message.append("  - %s" % udeb)

                message = "\n".join(message) + "\n"
                mailer.send([changed_by], subject, message)

            if "Files" in section:
                upload_path = os.path.dirname(upload.changes_path)
                for line in section["Files"].splitlines():
                    path = os.path.join(upload_path, line.split()[-1])
                    if os.path.exists(path):
                        os.remove(path)

        os.remove(upload.changes_path)
        return upload

    def _generate_sections(self, changes_file):
        """
        Generate sections from a .changes file.

        If apt_pkg blows up trying to read a section, terminate the generator.
        """
        fd = apt_pkg.open_maybe_clear_signed_file(changes_file)
        try:
            for section in apt_pkg.TagFile(fd):
                yield section
        finally:
            os.close(fd)
