# Copyright 2005-2022 Canonical Limited. All rights reserved.

import errno
import re
import os

from twisted.python import log
from zope.component import getUtility

from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

from storm.twisted.transact import transact

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.main.account import (
    get_account_by_name,
    get_account_by_id,
)
from canonical.landscape.model.repository.distribution import get_distribution


class UploadServiceError(Exception):
    """Raised when an error occurs while uploading a package."""

    def __init__(self, code, response):
        self.code = code
        self.response = response


class ResourceNotFound(UploadServiceError):
    """Raised when the path for the resource is invalid."""

    def __init__(self, path):
        super(ResourceNotFound, self).__init__(
            code=404, response="Resource not found: %s" % path
        )


class ForbiddenResource(UploadServiceError):
    """
    Raised when the requested resource for PUT already exists but can't be
    overwritten (it could be a directory or a pocket not in upload mode).
    """

    def __init__(self, path):
        super(ForbiddenResource, self).__init__(
            code=403, response="Forbidden: %s" % path
        )


class UploadService(Resource):
    """
    An HTTP service to handle deb packages upload to pockets.

    @param upload_path: the local base path for the uploads
    @param path: the HTTP base path for the service
    """

    _url_path_re = re.compile(
        r"^/+(?P<account>[^/]+)/+(?P<distribution>[^/]+)/+(?P<series>[^/]+)"
        r"/+(?P<pocket>[^/]+)/+(?P<filename>[^/]*)$"
    )

    def __init__(self, transactor, incoming_queue, path=""):
        Resource.__init__(self)
        self.transactor = transactor
        self.path = path
        self.incoming_queue = incoming_queue

    def isLeaf(self):
        return True

    def _handle_file_upload(self, resources, request):
        """Write the uploaded data to disk."""
        config = getUtility(IFileBasedConfiguration)
        repodir = config.get("landscape", "repository-path")
        accountdir = os.path.join(repodir, resources["account"]["name"])
        distrodir = os.path.join(accountdir, resources["distribution"]["name"])
        upload_path = os.path.join(distrodir, "incoming")

        if not resources["filename"]:
            # the path in the request is actually a directory
            raise ForbiddenResource(request.path)

        for directory in (repodir, accountdir, distrodir, upload_path):
            try:
                os.mkdir(directory)
                os.chmod(directory, 0o755)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

        upload_filename = os.path.join(upload_path, resources["filename"])
        reuploaded = os.path.exists(upload_filename)

        with open(upload_filename, "wb") as uploaded_file:
            while True:
                content = request.content.read()
                if not content:
                    break
                uploaded_file.write(content)

        code = 200 if reuploaded else 201
        if upload_filename.endswith(".changes") and not reuploaded:
            deferred = self._queue_upload(resources)
            deferred.addCallback(lambda _: self.incoming_queue.notify())
            return deferred.addCallback(lambda _: code)
        else:
            return code

    @transact
    def _queue_upload(self, resources):
        """Add an entry to the database queue disk about the new upload."""
        account = get_account_by_id(resources["account"]["id"])
        distribution = get_distribution(account, name=resources["distribution"]["name"])
        series = distribution.get_series(names=[resources["series"]["name"]]).one()
        pocket = series.get_pockets(names=[resources["pocket"]["name"]]).one()
        pocket.create_upload(resources["filename"])

    def handle(self, request):
        """Handle the HTTP request."""
        deferred = self._get_resources(request)
        deferred.addCallback(self._handle_file_upload, request)

        def write_response(code):
            request.setResponseCode(code)
            request.setHeader("Content-Type", "text/plain")
            request.finish()

        def write_error(failure):
            if failure.check(UploadServiceError):
                code = failure.value.code
                response = failure.value.response.encode()
            else:
                log.err(failure)
                code = 500
                response = b"System error"
            request.setResponseCode(code)
            request.write(response)
            request.finish()

        deferred.addCallback(write_response)
        deferred.addErrback(write_error)
        return deferred

    @transact
    def _get_resources(self, request):
        """
        Get L{Account} and L{Distribution} based on the request path.

        @return: A dict with account, distribution, series and pocket names and
           ids, and the name of the file being uploaded
        """
        path = request.path
        try:  # This is str in tests but bytes in general
            path = path.decode()
        except AttributeError:
            pass

        match = self._url_path_re.match(path)
        if match is None:
            raise ResourceNotFound(path)
        parts = dict((key, value) for key, value in list(match.groupdict().items()))

        account = get_account_by_name(parts["account"])
        if not account:
            raise ResourceNotFound(path)
        distribution = get_distribution(account, name=parts["distribution"])
        if not distribution:
            raise ResourceNotFound(path)
        series = distribution.get_series(names=[parts["series"]]).one()
        if not series:
            raise ResourceNotFound(path)
        pocket = series.get_pockets(names=[parts["pocket"]]).one()
        if not pocket:
            raise ResourceNotFound(path)
        if not pocket.is_in_upload_mode():
            raise ForbiddenResource(path)
        return {
            "account": {"id": account.id, "name": account.name},
            "distribution": {"id": distribution.id, "name": distribution.name},
            "series": {"id": series.id, "name": series.name},
            "pocket": {"id": pocket.id, "name": pocket.name},
            "filename": parts["filename"],
        }

    def render_PUT(self, request):
        self.handle(request)
        return NOT_DONE_YET

    def render_GET(self, request):
        request.setHeader("Content-Type", "text/plain")
        return b"Landscape package upload service"
