# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Implementation of the server side of lightweight exchange-triggering system.
"""

from builtins import object
from datetime import datetime, timedelta

from twisted.internet import reactor
from twisted.web.resource import Resource

from storm.expr import Select, Coalesce
from storm.twisted.transact import transact

from landscape.lib.bpickle import dumps

from canonical.landscape.model.stormfunc import Now
from canonical.landscape.model.computer.alerts import ComputerOfflineAlert
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.resource.store import get_resource_stores
from canonical.landscape.model.computer.collection import ComputerCollection
from canonical.landscape.model.main.account import Account, get_accounts_by_ids
from canonical.landscape.model.main.computer import (
    Computer,
    ComputerStatus,
    PingTime,
    PingServer,
    update_managed_hours,
)
from canonical.landscape.model.activity.status import QUEUED, SCHEDULED
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.pingserver.pingtracker import PingTracker
from canonical.landscape.security.authentication import get_host_from_request


class WebPing(Resource):
    """A web resource which answers: Are there messages for computer X?

    @ivar ping: A callable that accepts an C{insecure_id} and a C{network_key}
        as arguments and returns True for those which represent computers
        which currently have pending messages and are allowed to connect to
        exchange messages.
    """

    def __init__(self, ping):
        Resource.__init__(self)
        self.ping = ping

    def render_HEAD(self, request):
        return b""

    def render_POST(self, request):
        insecure_id = request.args.get(b"insecure_id", [None])[0]
        if insecure_id is None:
            return dumps({"error": "provide insecure_id"})
        # Don't check for known proxies, since we don't have access to
        # the configuration, where the known proxies are stored.
        computer_has_messages = self.ping(
            int(insecure_id), get_host_from_request(request, check_known_proxies=False)
        )
        return dumps({"messages": computer_has_messages})

    render_GET = render_POST


# BIG O, SHOOOW TIIIIME
# <http://en.wikipedia.org/wiki/The_Big_O>

# - Create a mapping of computer IDs to the total number of messages currently
#   in that computer called "message_count".
# - Join it with message_info when rows have matching computer_ids.
# - Out of that joined expression, get all computer.id on computers whose
#   message-count is greater than the pending-offset (i.e. those with active
#   messages).
PENDING_MESSAGE_QUERY = """
SELECT message_info.computer_id FROM
    ((SELECT computer_id, COUNT(computer_id) AS total
      FROM message GROUP BY computer_id) AS message_count
     NATURAL JOIN message_info)
WHERE message_count.total > message_info.client_pending_offset
"""


class ComputersNeedingNotification(object):
    """Schedules checks for computers that have messages.

    @ivar tracker: A reference to a C{PingTracker}, which encapsulates
    all the logic needed for tracking which computers pinged us.

    @ivar server_id: The id of the ping server we're connected to.
    """

    def __init__(self, transactor):
        """
        Account and resource stores B{really ought} to have had
        C{SET TRANSACTION ISOLATION LEVEL READ COMMITTED} executed in them.

        @param main_store: The main L{Store}
        @param account_stores: A list of account L{Store}s.
        @param resource_stores: A list of resource L{Store}s.
        """
        self.tracker = PingTracker()
        self.server_id = None
        self.server_startup_time = None
        self.transactor = transactor

    @transact
    def check_for_messages(self, time=None):
        """
        Go over all account and resource databases and accumulate insecure-ids
        for computers which have pending messages and undelivered activities.

        @return: A L{set} of computer IDs which have have pending messages.
        """
        computer_ids = set()
        time = time if time is not None else datetime.utcnow()
        # Get IDs for computers with pending activities.
        for name, account_store in get_account_stores():

            collection = ActivityChildCollection(
                store=account_store,
                account=None,
                statuses=[QUEUED, SCHEDULED],
                entity_class=AccountComputerTwin,
            )
            collection = collection.select(AccountComputerTwin.id)
            result = collection.find()
            result = result.find(
                Coalesce(Activity.schedule_before_time, time) >= time,
                Coalesce(Activity.schedule_after_time, time) <= time,
            )
            computer_ids.update(result)

        # Get IDs for computers with pending messages.
        for name, resource_store in get_resource_stores():
            result = resource_store.execute(PENDING_MESSAGE_QUERY)
            computer_ids.update(row[0] for row in result.get_all())

        # get insecure IDs for computers with messasges
        collection = ComputerCollection()
        collection = collection.with_computer_ids(computer_ids)
        result = collection.find()
        return set(result.values(Computer.insecure_id))

    def _check(self):
        """
        Check for pending messages. This must be run in order for
        L{self.tracker.computer_has_messages} to have an up-to-date value.
        This should be done by calling L{start} instead of directly
        calling L{_check}.
        """
        deferred = self.check_for_messages()
        deferred.addCallback(self.tracker.set_computers_with_messages)
        return deferred

    @transact
    def _write_ping_times(self, insecure_ids):
        """
        Set the C{last_ping_time} to C{Now} for each of the computers with the
        given C{insecure_ids}, then run a check for the L{ComputerOfflineAlert}
        for each all the accounts related to the provided C{computer_ids}.
        """
        store = get_main_store()

        # Record the last time ping times were written out to the
        # database.
        if self.server_id is None:
            server_info = PingServer()
            server_info.last_update_time = Now
            server_info.startup_time = self.server_startup_time
            store.add(server_info)
            self.server_id = server_info.id
        else:
            store.find(PingServer, PingServer.id == self.server_id).set(
                last_update_time=Now
            )

        computer_ids = Select(Computer.id, Computer.insecure_id.is_in(insecure_ids))
        result = store.find(PingTime, PingTime.computer_id.is_in(computer_ids))
        result.set(last_ping_time=Now)

        update_managed_hours(computer_ids)

        exchange_time = datetime.utcnow() - timedelta(
            minutes=ComputerOfflineAlert.exchange_check_interval
        )

        # Now find the accounts related to the computer_ids that we've changed.
        result = store.find(
            (Account.id, Computer.id),
            Account.id == Computer.account_id,
            Computer.id.is_in(computer_ids),
            ComputerStatus.computer_id == Computer.id,
            (
                (ComputerStatus.last_exchange_time == None)
                or (ComputerStatus.last_exchange_time >= exchange_time)
            ),
        )

        computers_by_account = {}
        for account_id, computer_id in result:
            if account_id not in computers_by_account:
                computers_by_account[account_id] = set()
            computers_by_account[account_id].add(computer_id)

        # And update the states for the ComputerOfflineAlert check directly
        # with the computers known to be online, instead of running a full
        # rule check which would check both online and offline computers.
        accounts = get_accounts_by_ids(list(computers_by_account.keys()))
        for account in accounts:
            computer_ids = computers_by_account[account.id]
            for rule in account.get_rules(info_classes=[ComputerOfflineAlert]):
                rule.info.update_states(inactive_computer_ids=computer_ids)

    def start(self, check_interval, write_interval, schedule=reactor.callLater):
        """
        Occasionally check for new messages for computers, ensuring
        L{self.tracker.computer_has_messages} will return reasonably
        up-to-date values.

        @param check_interval: The interval at which to update knowledge of
            computers with outstanding messages from the database.
        @param write_interval: The interval at which to write the last ping
            time values to the database, based on computers that have recently
            pinged.
        @param schedule: A callable similar to
            L{reactor.callLater}. Default is L{reactor.callLater}.
        """

        def reschedule_check(ignored):
            self._check_call = schedule(check_interval, check_loop)
            return ignored

        def check_loop():
            deferred = self._check()
            deferred.addBoth(reschedule_check)
            self._check_deferred = deferred

        def reschedule_write(ignored):
            self._write_call = schedule(write_interval, write_loop)
            return ignored

        def write_loop():
            insecure_ids = self.tracker.get_pings().copy()

            def write_failed(failure):
                self.tracker.get_pings().update(insecure_ids)
                return failure

            self.tracker.clear_pings()
            deferred = self._write_ping_times(insecure_ids)
            deferred.addErrback(write_failed)
            deferred.addBoth(reschedule_write)
            self._write_deferred = deferred

        self.server_startup_time = datetime.utcnow()
        self._check_deferred = None
        self._write_deferred = None
        self._check_call = schedule(check_interval, check_loop)
        self._write_call = schedule(write_interval, write_loop)

    def stop(self):
        """Stop periodically updating."""
        self._check_call.cancel()
        self._write_call.cancel()
        self.server_id = None
