# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from collections import deque
import time


class PingTracker(object):
    """Predictive load distribution engine for pings.

    The primary goal of the ping tracker is to distribute pings in a way
    that prevents the message server from having to handle a very large
    number of connections at once.  It does that by keeping a history of
    of the recent pings that have been received and whether these pings
    have been answered positively or not.  It then uses that information
    to decide whether a computer will receive a positive answer or not.

    When a computer is flagged as having available messages to be
    collected, two main parameters are used to decide whether a positive
    answer will really be provided in response to a ping:

    - Window duration: it's the amount of time we'll look back in time
      to consider how many pings happened in the recent history.

    - Ping limit: it's the maximum number of pings which may receive a
      positive answer within the time set by the window duration.

    In addition to these, the ping tracker may also decide to answer
    negatively to a given ping if the same computer connected within
    a certain amount of time.  The maximum penalty applied is the value
    set as the penalty window duration.  That value will be scaled
    down based on the current fill of the ping window.
    """

    window_duration = 5
    window_ping_limit = 10
    penalty_window_duration = 120

    def __init__(self, get_current_time=time.time):
        self._get_current_time = get_current_time
        self._insecure_ids_with_messages = set()
        self._pings = set()  # set([insecure_id, ...])
        self._window = deque()  # [ping_time, ...]
        self._penalty_window = deque()  # [(insecure_id, ping_time), ...]
        self._positive_pings = {}  # {insecure_id: last_ping_time, ...}

    def set_window_duration(self, seconds):
        """Set the ping window duration.

        Within the amount of time set as the window duration, the ping
        server will answer positively to at most C{window_ping_limit}
        pings.
        """
        self.window_duration = seconds

    def set_penalty_window_duration(self, seconds):
        """Set the maximum penalty inflicted on reconnecting computers.

        Applying a penalty on a computer in this context means returning
        a negative answer for the ping when in fact we knew it had a
        message awaiting on the server.

        The penalty window is the maximum amount of time since the previous
        ping which the computer may be denied.  The actual time will be
        determined by multiplying the penalty window by the current fill
        percentage of the ping window.

        For example, if the ping limit within the window is 10 and we
        received 5 pings within the last 10 seconds, it means that the
        window right now is 50% full, and thus the actual penalty is half
        of the total penalty window.
        """
        self.penalty_window_duration = seconds

    def set_window_ping_limit(self, number_of_pings):
        """
        Set the maximum number of positive pings within the window duration.
        """
        self.window_ping_limit = number_of_pings

    def set_computers_with_messages(self, insecure_ids):
        """Replace the known list of computers which have awaiting messages."""
        self._insecure_ids_with_messages.clear()
        self._insecure_ids_with_messages.update(insecure_ids)

    def computer_has_messages(self, insecure_id):
        """Return True if messages await the computer with C{insecure_id}."""
        return insecure_id in self._insecure_ids_with_messages

    def get_pings(self):
        """Get a set of insecure_ids from computers which pinged.

        The returned set will contain all insecure_ids which pinged since
        the last time they were explicitly cleared with L{clear_pings()}.
        """
        return self._pings

    def clear_pings(self):
        """Reset the set of pings returned by L{get_pings()}."""
        self._pings.clear()

    def get_current_penalty(self):
        """Get current penalty based on recent pings.

        The current penalty is based on how full the ping window is at
        the present moment.  Computers will only receive a true ping
        if their last ping was made before the current penalty window
        started.
        """
        window_fill = len(self._window) / float(self.window_ping_limit)
        return window_fill * self.penalty_window_duration

    def get_last_positive_ping(self, insecure_id, network_key=None):
        """Get timestamp for the computer's last positive ping answer.

        @param insecure_id: Insecure ID for the computer.
        @param network_key: Parameter to reduce the chances of a spoofing
            attack on the ping server from working.  Computers with
            different network keys will not share the last ping time even
            if the C{insecure_id} is the same.  As a consequence, they
            won't share penalties.
        """
        return self._positive_pings.get((insecure_id, network_key))

    def ping(self, insecure_id, network_key=None):
        """Track a ping from a computer and tell whether it should connect.

        Log messages are generated to track performance of the ping tracker
        algorithm.  Due to the volume of data generated, log messages use a
        terse format with symbolic keys::

          A - Represents whether or not the computer is authorized to connect.
              1 is used for true and 0 is used for false.
          M - Represents whether or not pending messages are present when the
              call is made.  1 is used for true and 0 is used for false.
          C - The insecure ID of the computer pinging the server.

        The log messages generated use these symbols to capture information
        about the outcome of a ping.  For example, C{R:0 PM:1 ID:42} indicates
        that the computer with insecure ID 42 was denied access and it has
        pending messages.

        @param insecure_id: Insecure ID for the computer.
        @param network_key: Parameter to reduce the chances of a spoofing
            attack on the ping server from working.  Computers with
            different network keys will not share the last ping time even
            if the C{insecure_id} is the same.  As a consequence, they
            won't share penalties.
        """
        # Track this ping at all times.
        self._pings.add(insecure_id)

        # Remove old pings whose times were dropped off the window.
        now = self._get_current_time()
        window_start = now - self.window_duration
        while self._window and self._window[0] <= window_start:
            self._window.popleft()

        # The penalty window deque is used to clean up the positive pings that
        # happened before the current penalty_window_duration.
        penalty_window_start = now - self.penalty_window_duration
        while (
            self._penalty_window and self._penalty_window[0][-1] <= penalty_window_start
        ):
            ping_key, ping_time = self._penalty_window.popleft()

            # We compare to the stored positive ping time to ensure that if
            # the last positive ping is found more than once in the list (ie,
            # if there are two positive pings for the same computer within the
            # penalty window) we don't drop it off positive pings.
            if self._positive_pings[ping_key] == ping_time:
                del self._positive_pings[ping_key]

        # Check if the machine has messages to exchange.
        messages_waiting = self.computer_has_messages(insecure_id)
        if len(self._window) < self.window_ping_limit and messages_waiting:
            last_ping = self._positive_pings.get((insecure_id, network_key))
            current_penality = self.get_current_penalty()
            if last_ping is None or now - last_ping >= current_penality:
                ping_key = insecure_id, network_key
                self._positive_pings[ping_key] = now
                self._window.append(now)
                self._penalty_window.append((ping_key, now))
                return True

        return False
