# Copyright 2005-2022 Canonical Limited. All rights reserved.

from urllib.parse import urlparse
from urllib.parse import quote as urlquote

from zope.interface import implementer

from twisted.cred.credentials import IAnonymous, IUsernamePassword
from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.error import UnauthorizedLogin
from twisted.cred.portal import IRealm
from twisted.internet.defer import succeed, fail
from twisted.web.resource import IResource, Resource
from twisted.web.http import UNAUTHORIZED

from canonical.twisted.proxy import (
    SecureReverseProxyResource,
    ReverseProxyResourceWithFinish,
)

from canonical.landscape.model.main.standalone import (
    AccountLDSLicenseAPI,
    check_account_archive_access,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.account import get_account_by_name


__all__ = ["ArchiveChecker", "ArchiveRealm", "LicenseOnlyRealm", "ARCHIVE_REALM_NAME"]


ARCHIVE_REALM_NAME = "Landscape on Premises Archive"


class LicenseResource(Resource):
    """A resource rendering the current license file for the customer."""

    def render(self, request):
        account_name = request.getUser().decode()
        try:
            account = get_account_by_name(account_name)
            api = AccountLDSLicenseAPI(account)
            contents = api.get_license().contents
        finally:
            get_main_store().rollback()

        request.setHeader("Content-Type", "text-plain")
        request.setHeader("Content-Disposition", "attachment; filename=license.txt")
        return contents.encode()


class HealthResource(Resource):
    """Health check resource responding to HAProxy HEAD or GET pings to /."""

    isLeaf = True

    def render_GET(self, request):
        return ARCHIVE_REALM_NAME

    def render_HEAD(self, request):
        return b""


class UnauthorizedResource(Resource):
    """Return a 401 Unauthorized error."""

    isLeaf = True

    def render_GET(self, request):
        request.setResponseCode(UNAUTHORIZED)
        request.setHeader(
            "WWW-Authenticate", "basic realm={}".format(ARCHIVE_REALM_NAME)
        )
        return b""


class UnauthenticatedResource(Resource):
    """
    A resource raising unauthorized for any request with path different from
    "/".
    """

    def __init__(self):
        Resource.__init__(self)
        self.putChild(b"", HealthResource())

    def getChild(self, name, request):
        return UnauthorizedResource()


@implementer(ICredentialsChecker)
class ArchiveChecker(object):
    """
    A credentials checker using C{check_account_archive_access} for the access.
    """

    credentialInterfaces = (IAnonymous, IUsernamePassword)

    def requestAvatarId(self, credentials):
        if IAnonymous.providedBy(credentials):
            return succeed(None)
        account_name = credentials.username.decode()
        token = credentials.password.decode()
        try:
            if check_account_archive_access(account_name, token):
                return succeed(credentials.username)
        finally:
            get_main_store().rollback()
        return fail(UnauthorizedLogin())


@implementer(IRealm)
class ArchiveRealm(object):
    """
    A authentication realm retrieving the default release for the logged
    account, and returning a L{SecureReverseProxyResourceForPPPA} pointing to
    it.
    """

    def __init__(self, config, certificates, verifyHostname):
        self.config = config
        self.certificates = certificates
        self.verifyHostname = verifyHostname

    def requestAvatar(self, avatarId, mind, *interfaces):
        if IResource in interfaces:

            if avatarId is None:  # This means no Authorization was provided
                return (IResource, UnauthenticatedResource(), lambda: None)

            account_name = avatarId.decode()
            try:
                account = get_account_by_name(account_name)
                api = AccountLDSLicenseAPI(account)
                release = api.get_release()
            finally:
                get_main_store().rollback()
            if not release:
                raise NotImplementedError()
            pppa_url = self.config.get("%s-url" % release)

            netloc, path = urlparse(pppa_url)[1:3]
            if "@" in netloc:
                # This is a private PPA url with HTTP auth info, let's get them
                # and create a resource with a secure reverse proxy (SSL).
                auth, host = netloc.split("@")
                archive_resource = SecureReverseProxyResourceForPPPA(
                    host,
                    443,
                    path,
                    auth,
                    True,
                    self.certificates,
                    self.verifyHostname,
                    self.config,
                )
            else:
                # This is a public PPA url with no auth info, let's just create
                # a plain reverse proxy (no SSL)
                auth = None
                host = netloc
                archive_resource = ReverseProxyResourceWithFinish(host, 80, path)

            archive_resource.putChild(b"license.txt", LicenseResource())
            return (IResource, archive_resource, lambda: None)
        raise NotImplementedError()


@implementer(IRealm)
class LicenseOnlyRealm:
    """
    Authentication realm retrieving only the license resource for the authenticated user.
    """

    def requestAvatar(self, avatarId, mind, *interfaces):
        if IResource in interfaces:

            if avatarId is None:
                return (IResource, UnauthorizedResource(), lambda: None)

            return (IResource, LicenseResource(), lambda: None)

        raise NotImplementedError()


class SecureReverseProxyResourceForPPPA(SecureReverseProxyResource):
    def __init__(
        self,
        host,
        port,
        path,
        auth,
        verify=False,
        caCerts=None,
        verifyCallback=None,
        url_map=None,
    ):
        SecureReverseProxyResource.__init__(
            self, host, port, path, auth, verify, caCerts, verifyCallback
        )
        if url_map is None:
            url_map = {}
        self.url_map = url_map

    def getChild(self, path, request):
        """
        Override C{getChild} to manipulate the auth and path in order to
        proxy to different PPAs.
        """
        base_path, new_path, auth = self.get_path_and_auth(path, request)

        return SecureReverseProxyResource(
            self.host,
            self.port,
            base_path + "/" + urlquote(new_path, safe=""),
            auth,
            self.verify,
            self.caCerts,
            self.verifyCallback,
            self.reactor,
        )

    def get_path_and_auth(self, request_path, request):
        """
        If the path contains a key that maps to a url in url_map, generate the
        path and auth from the mapped url.
        """
        url = self.url_map.get("%s-url" % request_path)
        if url is None:
            path = self.path
            auth = self.auth
            request.postpath[0:0] = request.prepath
            del request.prepath[:]
        else:
            netloc, path = urlparse(url)[1:3]
            auth = netloc.split("@")[0]

        children = path.split("/")
        base_path = "/".join(children[:-1])
        new_path = children[-1]

        return base_path, new_path, auth
