# Copyright 2005-2017 Canonical Limited.  All rights reserved.

"""
Functions related to creating, providing or parsing and Prometheus data.
"""
from __future__ import division

from builtins import str
from past.utils import old_div
from builtins import object
from datetime import timedelta
from collections import defaultdict
from uuid import uuid4
from prometheus_client import Counter, Gauge, Histogram, REGISTRY

from zope.component import getUtility
from zope.interface import implementer

from canonical.landscape.model.account.collection import AccountCollection
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.computer import get_computer_counts_by_type
from canonical.landscape.model.billing.usage import get_usage_summary
from canonical.landscape.model.billing.payment import get_payments_with_status
from canonical.landscape.model.configuration import (
    ISystemConfiguration,
    LANDSCAPE_ANONYMOUS_UUID,
    LANDSCAPE_DO_NOT_TRACK_GA,
    get_revision,
)
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.main.account import (
    DEFAULT_STANDALONE_ACCOUNT_NAME,
    get_account_by_name,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.prometheus.interfaces import IPrometheusMetrics
from canonical.lib.timestamp import format_datetime, utcnow

__all__ = [
    "get_standalone_account_metrics",
    "PrometheusMetrics",
    "update_prometheus_saas_metrics",
]

__clearlist__ = __all__

__requires_prefix__ = "landscape."


@implementer(IPrometheusMetrics)
class PrometheusMetrics(object):
    """Collection of metrics in the provided registry."""

    default_registry = REGISTRY
    registry = None

    def __init__(self, registry=None):
        self.registry = registry or self.default_registry

        # Initialize metrics for the configured registry.
        self._metrics = {
            "landscape_http_requests": Counter(
                "landscape_http_requests",
                "Count of HTTP requests",
                ["revision", "code", "method", "path"],
                registry=self.registry,
            ),
            "landscape_cloud_regions": Counter(
                "landscape_cloud_regions",
                "Count of OpenStack regions",
                ["revision", "computers", "compute", "object", "block", "zones"],
                registry=self.registry,
            ),
            "landscape_maas_servers": Counter(
                "landscape_maas_servers",
                "Count of configured MAAS servers",
                ["revision"],
                registry=self.registry,
            ),
            "landscape_on_prem_computers": Counter(
                "landscape_on_prem_computers",
                "Count of Landscape On-Premises computers by type",
                ["revision", "type"],
                registry=self.registry,
            ),
            "landscape_on_prem_licenses": Counter(
                "landscape_on_prem_licenses",
                "Count of Landscape On-Premises computers by type",
                ["revision", "type"],
                registry=self.registry,
            ),
            "landscape_self_service_accounts": Gauge(
                "landscape_self_service_accounts",
                "Count of Landscape Self-Service accounts",
                registry=self.registry,
            ),
            "landscape_self_service_computers": Gauge(
                "landscape_self_service_computers",
                "Count of self-serivce computers active within the last hour",
                ["type"],
                registry=self.registry,
            ),
            "landscape_self_service_failed_payments": Gauge(
                "landscape_self_service_failed_payments",
                "Count of self service failed payments",
                registry=self.registry,
            ),
            "landscape_self_service_payments": Gauge(
                "landscape_self_service_payments",
                "Sum of self-service payments across all accounts in cents",
                registry=self.registry,
            ),
            "landscape_self_service_usage": Histogram(
                "landscape_self_service_usage",
                (
                    "Per-account self-service billable machine-hours for the "
                    "current period, in dollars"
                ),
                buckets=(10, 100, 500, 1000, float("inf")),
                registry=self.registry,
            ),
            "landscape_self_service_free_usage": Histogram(
                "landscape_self_service_free_usage",
                (
                    "Per Account self-service machine-hours for the current "
                    "period for which free credits apply, in dollars"
                ),
                buckets=(1, 10, 30, 50, 100, float("inf")),
                registry=self.registry,
            ),
            "landscape_accounts": Gauge(
                "landscape_accounts",
                "Count of Landscape active accounts",
                registry=self.registry,
            ),
            "landscape_computers": Gauge(
                "landscape_computers",
                "Count of computers active within the last hour",
                ["type"],
                registry=self.registry,
            ),
        }

    def get(self, name):
        """Get a registered metric by name."""
        return self._metrics.get(name)


def get_standalone_account_metrics():
    """Return a dictionary describing standalone configuration data.

    Data included: maas, cloud, number of registered computers and licenses.
    """
    configuration = getUtility(ISystemConfiguration)
    if configuration.get(LANDSCAPE_DO_NOT_TRACK_GA):
        return {}
    metrics = {"revision": get_revision()}
    account = get_account_by_name(DEFAULT_STANDALONE_ACCOUNT_NAME)
    if account:
        uuid = configuration.get(LANDSCAPE_ANONYMOUS_UUID)
        if not uuid:
            uuid = str(uuid4())
            # XXX We need to handle transaction commit boundaries here
            # in order to persist this change from BatchScript callers
            # otherwise the uuid generated never gets committed to the db.
            configuration.set(LANDSCAPE_ANONYMOUS_UUID, uuid)
        metrics["uuid"] = uuid
        metrics["created"] = format_datetime(account.creation_time)
        metrics["license"] = _get_license_metrics(account)
        metrics["maas"] = 0  # Maas has been removed from codebase
        metrics["computers"] = get_computer_counts_by_type([account.id])
    return metrics


def _update_self_service_metrics():
    """Calculate and store self-service related metrics in Prometheus."""
    metrics = getUtility(IPrometheusMetrics)
    collection = AccountCollection(get_main_store()).with_self_service(True)
    _update_computer_type_metrics("landscape_self_service_computers", collection)
    accounts = collection.find()
    metrics.get("landscape_self_service_accounts").set(accounts.count())
    failed_payments = 0
    total_paid = 0
    for _, store in get_account_stores():
        failed_payments += get_payments_with_status(store, ["failed"]).count()
        for payment in get_payments_with_status(store, ["paid"]):
            total_paid += payment.due
    metrics.get("landscape_self_service_payments").set(old_div(total_paid, 100))
    metrics.get("landscape_self_service_failed_payments").set(failed_payments)
    _update_self_service_usage_metrics(metrics, accounts)


def _update_self_service_usage_metrics(metrics, accounts):
    """Calculate and store credit usage on self-service accounts.

    Usage is accumulated for histogram, so for a period, the ratios in buckets
    should match the distribution, but queries for actual numbers
    (e.g. _sum, _count) have to be queried using delta() or increase().
    Also, since buckets are fixed, that means the <10 will be full
    at the start of the period, but should have less toward the end.
    """
    usage = metrics.get("landscape_self_service_usage")
    free_usage = metrics.get("landscape_self_service_free_usage")
    for account in accounts:
        account_summary = get_usage_summary(account, include_outstanding=False)
        usage.observe(account_summary.cost.due / 10000.0)
        free_credits_used = account_summary.cost.used_credit
        if free_credits_used:
            free_usage.observe(free_credits_used / 10000.0)


def update_prometheus_saas_metrics():
    """Calculate and refresh SAAS account metrics into Prometheus."""
    metrics = getUtility(IPrometheusMetrics)

    # Self-service account metrics
    _update_self_service_metrics()

    # SAAS account metrics
    collection = AccountCollection(get_main_store()).with_active()
    _update_computer_type_metrics("landscape_computers", collection)
    saas_accounts = collection.find()
    metrics.get("landscape_accounts").set(saas_accounts.count())


def _update_computer_type_metrics(metric_name, collection):
    """Calculate and store computer type counts in the named metric."""
    metrics = getUtility(IPrometheusMetrics)
    computer_metric = metrics.get(metric_name)
    one_hour_ago = utcnow() - timedelta(hours=1)
    account_ids = list(collection.find().values(Account.id))
    saas_computer_counts = get_computer_counts_by_type(
        account_ids, last_ping_time=one_hour_ago
    )
    for _type, count in saas_computer_counts.items():
        computer_metric.labels(type=_type).set(count)


def _get_license_metrics(account):
    """Return license details for the account."""
    license_seats = defaultdict(int)
    license_management = LicenseManagement(account)
    for license in license_management.get_licenses(active_only=True):
        license_seats[license.role_key] += license.seats
    return license_seats


def _get_cloud_metrics(account):
    """Return details of any cloud deployed as a dict."""
    cloud_management = OpenStackManagement(account)
    clouds = cloud_management.get_clouds()
    clouds_data = []
    for cloud in clouds:
        cloud_data = {"creation": format_datetime(cloud.creation_time), "regions": []}
        for region in cloud_management.get_regions(cloud):
            cloud_data["regions"].append(_get_region_metrics(region))
        clouds_data.append(cloud_data)
    return clouds_data


def _get_region_metrics(region):
    """Return details of a region deployed as a dict."""
    region_data = {}
    region_data["compute"] = ",".join(
        sorted([hypervisor.name for hypervisor in region.hypervisors])
    )
    region_data["object"] = region.object_storage
    region_data["block"] = region.block_storage
    service_config = get_region_service_config(region)
    region_config = service_config.region_config
    region_data["zones"] = [len(zone.machines) for zone in region_config.zone_configs]
    return region_data
