# Copyright 2005-2022 Canonical Limited. All rights reserved.

import random
import time

from zope.app.publication.browser import BrowserPublication
from zope.component import queryUtility, getUtility, adapter
from zope.security.interfaces import Unauthorized
from zope.publisher.interfaces import (
    NotFound as ZopeNotFound,
    Redirect as ZopeRedirect,
    Retry as ZopeRetry,
)
from zope.publisher.interfaces.browser import IBrowserSkinType
from zope.publisher.skinnable import applySkin
from zope.traversing.interfaces import IBeforeTraverseEvent
from zope.pagetemplate.engine import zopeTraverser

import transaction

from storm.zope.interfaces import IZStorm
from storm.twisted.transact import RETRIABLE_ERRORS
from storm.exceptions import DisconnectionError

from canonical.oops.interfaces import IOOPSManager
from canonical.routes import Retry
from canonical.routes.publisher import proxify

from canonical.landscape.model.root.interfaces import ILandscapeRoot


def request_should_be_retried(request, exception):
    """Return whether the request should be retried in case of an exception.

    In order for the request to be retried, the exception has to be of a
    retriable kind (i.e a database error that might not happen the next
    time), and the request must be retriable.

    If the request isn't retriable, it usually means that the request
    has already been retried a few times, and it's time to give up.

    @param request: A C{BrowserRequest} of which C{supportsRetry} will be
        consulted.
    @param exception: The C{Exception} that might cause a retry to happen.

    """
    return request.supportsRetry() and isinstance(exception, RETRIABLE_ERRORS)


class LandscapeBrowserPublication(BrowserPublication):

    root_object_interface = ILandscapeRoot

    def __init__(self, db):
        super(LandscapeBrowserPublication, self).__init__(db)
        self.application = db

    def proxy(self, ob):
        """Possibly wrap the given object with a security proxy.

        This method will be invoked by Zope's machinery and it's customized
        to return no security proxy in case the given object has a __no_proxy__
        marker, which means that relevant security was already checked in
        the service layer.
        """
        return proxify(ob)

    def beforeTraversal(self, request):
        zstorm = getUtility(IZStorm)
        for name, store in zstorm.iterstores():
            store.rollback()

        request.response.setHeader("X-Frame-Options", "SAMEORIGIN")
        request.response.setHeader("X-Content-Type-Options", "nosniff")
        request.response.setHeader("X-XSS-Protection", "1; mode=block")
        request.response.setHeader("Strict-Transport-Security", "max-age=2592000")
        super(LandscapeBrowserPublication, self).beforeTraversal(request)

    def getApplication(self, request):
        return self.application

    def handleException(self, object, request, exc_info, retry_allowed=True):
        # We can't use retry_allowed, as we would go one step further than
        # we want, and thus the oops manager would lose track of the
        # request.
        if request_should_be_retried(request, exc_info[1]):
            if isinstance(exc_info[1], DisconnectionError):
                # If we got a disconnection, calling rollback may not be enough
                # because psycopg2 doesn't necessarily use the connection, so
                # we call a dummy query to be sure that all the stores are
                # correct.
                zstorm = getUtility(IZStorm)
                for name, store in zstorm.iterstores():
                    try:
                        store.execute("SELECT 1")
                    except DisconnectionError:
                        pass
            transaction.abort()
            count = getattr(request, "_retry_count", 0) + 1
            time.sleep(random.uniform(1, 2 ** count))
            raise Retry(exc_info)
        # An error reached the publication. Flag the OOPS as wanted,
        # unless the OOPS is an Unauthorized error created by an anonymous.
        oops_manager = queryUtility(IOOPSManager)
        if (
            oops_manager
            and (
                exc_info[0] != Unauthorized
                or request.principal.id != "landscape.anonymous"
            )
            and not isinstance(exc_info[1], (ZopeNotFound, ZopeRedirect, ZopeRetry))
        ):
            oops = oops_manager.get("request")
            if oops is not None:
                oops.wanted = True

        base = super(LandscapeBrowserPublication, self)
        base.handleException(object, request, exc_info, retry_allowed)


@adapter(IBeforeTraverseEvent)
def skin_switch(event):
    """Switches current request's skin if a cookie is present."""
    request = event.request

    # Some tests cause this event to fire, but don't provide a request at all.
    if request is None:
        return

    if request.getHeader("X-Requested-With", "").lower() == "xmlhttprequest":
        name = "Ajax"
    elif request.getHeader("Accept") == "application/json":
        name = "JSON"
    else:
        name = request.cookies.get("skin")

    if name is not None:
        skin = queryUtility(IBrowserSkinType, name)
        if skin is not None:
            applySkin(request, skin)


# Set the proxy factory of the default Zope traverser, so objects returned
# by TALES expressions will be proxified only if they didn't come from the
# service layer, which sets __no_proxy__.
zopeTraverser.proxify = proxify
