# Copyright 2005-2022 Canonical Limited. All rights reserved.
"""Build a sample APT repository that can evolve along different 'stages'."""

from builtins import str
import os

from twisted.internet import reactor
from twisted.internet.defer import succeed

from landscape.lib.fs import create_file, read_file

from canonical.lib.archive import DebFile, FileInfo
from canonical.reprepro.reprepro import Reprepro

from canonical.testing.gpg import SAMPLE_SECRET_KEY_MATERIAL

from canonical.landscape.constants import ROOT_DIR

BASE_DIR = os.path.join(ROOT_DIR, "debarchive")
STAGE_FILENAME = os.path.join(BASE_DIR, "stage")
DISTRIBUTIONS = [
    [
        "lucid",
        ["main"],
        ["i386"],
        "upload",
        "test@example.com",
        SAMPLE_SECRET_KEY_MATERIAL,
    ]
]


def build_repo():
    """Build an APT repository under the C{BASE_DIR} directory.

    Subsequent invocations of this function (also from different processes)
    will 'advance' the repository between stages. Currently only 2 stages
    are available, see also L{_step}.
    """
    if not os.path.exists(BASE_DIR):
        os.mkdir(BASE_DIR)
    if not os.path.exists(STAGE_FILENAME):
        stage = 0
    else:
        stage = int(read_file(STAGE_FILENAME)) + 1
    reactor.callWhenRunning(_step, stage)
    reactor.run()


def _step(stage):
    """
    Make the repository advance through a series of stages, one step at a time.
    """
    deferred = succeed(None)

    if stage == 0:
        del_packages = []
        add_packages = [
            ("python-mode", "5.1.0-1", "all"),
            ("python-psycopg2", "2.4-1", "i386"),
        ]
    elif stage == 1:
        del_packages = ["python-psycopg2"]
        add_packages = [
            ("python-mode", "5.1.0-2", "all"),
            ("fetchmail", "6.3.19-1ubuntu1", "i386"),
        ]
    else:
        del_packages = []
        add_packages = []

    for package in del_packages:

        def create_del_deb_callback(package):
            return lambda _: _del_deb("lucid", package)

        deferred.addCallback(create_del_deb_callback(package))

    for package in add_packages:

        def create_add_deb_callback(package):
            return lambda _: _add_deb("lucid", package)

        deferred.addCallback(create_add_deb_callback(package))

    def finish(result):
        create_file(STAGE_FILENAME, content=b"%d" % stage)
        reactor.stop()

    deferred.addCallback(finish)
    return deferred


def _build_deb(deb_details):
    """Build a .deb file with the given details.

    @return: The filename of the created .deb
    """
    (name, version, arch) = deb_details
    deb = DebFile()
    deb.add(FileInfo("./", mode=0o755))
    deb.add(FileInfo("./some", b"Some data\n", mode=0o600))
    deb.field("Package", name)
    deb.field("Version", version)
    deb.field("Priority", "optional")
    deb.field("Maintainer", "landscape@canonical.com")
    deb.field("Architecture", "all")
    deb.field("Section", "python")
    deb.field("Installed-Size", str(deb.size()))
    deb.field("Description", "A sample package\n" " This is a nice sample package.\n")
    filename = "%s_%s_%s.deb" % (name, version, arch)
    create_file(os.path.join(BASE_DIR, filename), deb.pack())
    return filename


def _get_reprepro():
    """Create a properly configured L{Reprepro} instance."""
    reprepro = Reprepro(BASE_DIR)
    reprepro.set_gpg_home(os.path.join(BASE_DIR, ".gnupg"))
    for name, components, archs, mode, sign, key in DISTRIBUTIONS:
        reprepro.add_distribution(name, components, archs, mode=mode, sign=sign)
        reprepro.add_gpg_key(key)
    return reprepro


def _add_deb(distribution, package):
    """Delete a .deb from a distribution."""
    reprepro = _get_reprepro()
    filename = _build_deb(package)
    reprepro.set_distribution(distribution)
    deferred = reprepro.includedeb(filename)

    def cleanup(_):
        os.unlink(os.path.join(BASE_DIR, filename))

    deferred.addCallback(cleanup)
    return deferred


def _del_deb(distribution, package):
    """Add a .deb to a distribution."""
    reprepro = _get_reprepro()
    reprepro.set_distribution(distribution)
    return reprepro.remove(package)
