from builtins import range
from copy import deepcopy
from random import choice

from faker import Faker

from canonical.lib.text import make_secure_id
from canonical.landscape.model.account.factory import AccountProfile
from canonical.landscape.model.computer.factory import ComputerProfile

# Persons common to both devel and standalone sample data
PERSONS_COMMON = {
    "john": {
        "name": "John Smith",
        "email": "john@example.com",
        "identity": "{root_url}/id/john",
    },
    "jane": {
        "name": "Jane Robertson",
        "email": "jane@example.com",
        "password": "pwd",
    },
    "bob": {
        "name": "Bob Mellow",
        "email": "bob@example.com",
        "identity": "{root_url}/id/bob",
    },
    "noel": {
        "name": "Noel Coward",
        "email": "noel@example.com",
        "identity": "{root_url}/id/noel",
    },
    "oscar": {
        "name": "Oscar Blake",
        "email": "oscar@example.com",
        "identity": "{root_url}/id/oscar",
    },
    "robert": {
        "name": "Robert Frost",
        "email": "robert@example.com",
        "identity": "{root_url}/id/robert",
    },
}

# Persons specific to devel sample data
PERSONS_DEVEL = deepcopy(PERSONS_COMMON)
PERSONS_DEVEL.update(
    {
        "stan": {
            "name": "Stan Peters",
            "email": "stan@example.com",
            "identity": "{root_url}/id/stan",
            "global-roles": ["SupportProvider"],
        },
        "marge": {
            "name": "Marge Peters",
            "email": "marge@example.com",
            "identity": "{root_url}/id/marge",
            "global-roles": ["AccountManager", "Operator"],
        },
        "bill": {
            "name": "Bill Peters",
            "email": "bill@example.com",
            "identity": "{root_url}/id/bill",
        },
    }
)

# Persons specific to standalone sample data
PERSONS_STANDALONE = deepcopy(PERSONS_COMMON)
PERSONS_STANDALONE.update(
    {
        "stan": {
            "name": "Stan Peters",
            "email": "stan@example.com",
            "password": "pwd",
        },
        "marge": {
            "name": "Marge Peters",
            "email": "marge@example.com",
            "password": "pwd",
        },
    }
)

ACCOUNT_ACCESS_GROUPS_COMMON = [
    {"name": "server", "description": "Server machines"},
    {"name": "desktop", "description": "Desktop machines"},
]

ACCOUNT_ROLES_COMMON = [
    {
        "name": "ServerAdmin",
        "description": "Servers administrators",
        "access-groups": ["server"],
        "permissions": [
            "ManageComputer",
            "ViewScript",
            "ManagePackageProfile",
            "ManageUpgradeProfile",
        ],
    },
    {
        "name": "ServerAuditor",
        "description": "Servers auditors",
        "access-groups": ["server"],
        "permissions": [
            "ViewComputer",
            "ViewScript",
            "ViewPackageProfile",
            "ViewUpgradeProfile",
        ],
    },
    {
        "name": "DesktopAdmin",
        "description": "Desktops administrators",
        "access-groups": ["desktop"],
        "permissions": [
            "ManageComputer",
            "ViewScript",
            "ManagePackageProfile",
        ],
    },
]

# Persons belonging to both onward and standalone accounts.
ACCOUNT_PERSONS_COMMON = [
    {"key": "john", "roles": ["Owner"]},
    {"key": "jane", "roles": ["Owner"]},
    {"key": "noel", "roles": ["ServerAdmin"]},
    {"key": "oscar", "roles": ["DesktopAdmin"]},
    {"key": "robert", "roles": ["ServerAuditor"]},
]

# Accounts specific to devel sample data
ACCOUNTS_DEVEL = {
    "onward": {
        "name": "onward",
        "title": "Onward, Inc.",
        "max-people": 20,
        "seats": 25,
        "access-groups": deepcopy(ACCOUNT_ACCESS_GROUPS_COMMON),
        "roles": deepcopy(ACCOUNT_ROLES_COMMON),
        "persons": deepcopy(ACCOUNT_PERSONS_COMMON)
        + [
            {"key": "bob", "roles": ["SupportUser"]},
        ],
    },
    "upside": {
        "name": "upside",
        "title": "Upside Software, Ltd.",
        "max-people": 5,
        "persons": [
            {"key": "john", "roles": ["Owner"]},
            {"key": "bob", "roles": ["Owner"]},
        ],
    },
    "utility": {
        "name": "utility",
        "title": "Utility, Inc.",
        "max-people": 20,
        "stripe-customer-id": "cus_9KivsKAQIaQ6ey",
        "vat-number": "999999",
        "vat-country": "NL",
        "persons": [
            {"key": "bill", "roles": ["Owner"]},
        ],
    },
    "no-admins": {
        "name": "no-admins",
        "max-people": 20,
        "title": "No admins",
        "persons": [],
    },
}

# Accounts specific to standalone sample data
ACCOUNTS_STANDALONE = {
    "standalone": {
        "name": "standalone",
        "title": "Organization",
        "max-people": 20,
        "seats": 60,
        "access-groups": deepcopy(ACCOUNT_ACCESS_GROUPS_COMMON),
        "roles": deepcopy(ACCOUNT_ROLES_COMMON),
        "persons": deepcopy(ACCOUNT_PERSONS_COMMON)
        + [
            {"key": "stan", "roles": ["Owner"]},
            {"key": "marge", "roles": ["Owner"]},
        ],
    },
}

# Computer common to both onward and standalone sample data
COMPUTERS_COMMON = {
    "computer-1": {
        "title": "Application Server 1",
        "hostname": "appserv1",
        "tags": ["webfarm", "server", "lucid"],
        "distribution": "lucid",
        "access-group": "server",
    },
    "computer-2": {
        "title": "Application Server 2",
        "hostname": "appserv2",
        "tags": ["webfarm", "server", "precise"],
        "distribution": "precise",
        "wedged": True,
        "access-group": "server",
    },
    "computer-3": {
        "title": "Database Server 1",
        "hostname": "dbserv1",
        "tags": ["webfarm", "server", "lucid"],
        "distribution": "lucid",
        "access-group": "server",
    },
    "computer-4": {
        "title": "John's Laptop",
        "hostname": "john",
        "tags": ["laptop", "natty"],
        "distribution": "natty",
        "access-group": "desktop",
    },
    "computer-5": {
        "title": "Jane's Laptop",
        "hostname": "jane",
        "tags": ["laptop", "lucid"],
        "distribution": "lucid",
        "reboot-required": ["linux-image"],
    },
    "computer-10": {
        "title": "Noel's Windows Laptop",
        "hostname": "noel",
        "tags": ["laptop", "windows"],
        "distribution": "windows",
        "access-group": "server",
    },
}

# Extra computers common to both onward and standalone sample data
COMPUTERS_COMMON_EXTRA = {
    "computer-6": {
        "title": "Custom Web Server",
        "hostname": "customwebserver",
        "tags": ["server", "precise"],
        "distribution": "precise",
        "access_context": "servers",
    },
    "computer-7": {
        "title": "HP Z400 Desktop",
        "hostname": "hp1",
        "tags": ["server", "oneiric"],
        "distribution": "oneiric",
    },
    "computer-8": {
        "title": "KVM Guest",
        "hostname": "guest1",
        "tags": ["server", "precise"],
        "distribution": "precise",
    },
    "computer-8": {
        "title": "KVM Guest",
        "hostname": "guest1",
        "tags": ["server", "precise"],
        "distribution": "precise",
    },
    "computer-9": {
        "title": "LXC Container",
        "hostname": "lxc1",
        "tags": ["server", "precise"],
        "distribution": "precise",
        "container-info": "lxc",
    },
    "computer-11": {
        "title": "Bionic WSL",
        "hostname": "bionic-wsl",
        "tags": ["bionic", "wsl"],
        "distribution": "bionic",
    },
}

# Computers for the Upside sample account
COMPUTERS_UPSIDE = {
    "upside-computer-1": {
        "account": "upside",
        "title": "John's Workstation",
        "hostname": "john",
        "tags": ["gutsy"],
        "distribution": "gutsy",
    },
    "upside-computer-2": {
        "account": "upside",
        "title": "Dapper Testing Box",
        "hostname": "dapper",
        "tags": ["dapper", "testing"],
        "distribution": "dapper",
    },
    "upside-computer-3": {
        "account": "upside",
        "title": "Edgy Testing Box",
        "hostname": "edgy",
        "tags": ["edgy", "testing"],
        "distribution": "edgy",
        "wedged": True,
    },
    "upside-computer-4": {
        "account": "upside",
        "title": "Feisty Testing Box",
        "hostname": "feisty",
        "tags": ["feisty", "testing"],
        "distribution": "feisty",
    },
    "upside-computer-5": {
        "account": "upside",
        "title": "Feisty Virtual Box",
        "hostname": "feisty-vm",
        "tags": ["feisty", "testing"],
        "distribution": "feisty",
        "vm_info": "kvm",
    },
}

# Computers for the Utility sample account
COMPUTERS_UTILITY = {
    "utility-computer-1": {
        "account": "utility",
        "title": "Computer 1",
        "hostname": "one",
        "distribution": "xenial",
        "managed-hours": 100,
    },
    "utility-computer-2": {
        "account": "utility",
        "title": "Computer 2",
        "hostname": "two",
        "distribution": "xenial",
        "vm-info": "kvm",
        "managed-hours": 500,
    },
    "utility-computer-3": {
        "account": "utility",
        "title": "Computer 3",
        "hostname": "three",
        "distribution": "xenial",
        "container-info": "lxc",
        "managed-hours": 2000,
    },
}


class SampleAccountProfileDevel(AccountProfile):
    """Account profile for devel sample data."""

    def __init__(self):
        super(SampleAccountProfileDevel, self).__init__()
        self.persons.update(deepcopy(PERSONS_DEVEL))
        self.accounts.update(deepcopy(ACCOUNTS_DEVEL))


class SampleAccountProfileStandalone(AccountProfile):
    """Account profile for standalone sample data."""

    def __init__(self):
        super(SampleAccountProfileStandalone, self).__init__()
        self.persons.update(deepcopy(PERSONS_STANDALONE))
        self.accounts.update(deepcopy(ACCOUNTS_STANDALONE))


class SampleComputerProfileCommon(ComputerProfile):
    """Computer profile for the onward/standalone accounts."""

    def __init__(self, account):
        super(SampleComputerProfileCommon, self).__init__()
        self._account = account
        computers = deepcopy(COMPUTERS_COMMON)
        self.computers.update(_get_with_account(account, computers))

    def add_common_extra(self):
        """Add extra common computers (used for non-OpenStack sample data)."""
        computers = deepcopy(COMPUTERS_COMMON_EXTRA)
        self.computers.update(_get_with_account(self._account, computers))

    def add_extra(self, count):
        """Add the given number of extra computers."""
        distributions = ["hardy", "jaunty", "karmic"]
        start = len(self.computers) + 1
        computers = {}
        for i in range(start, start + count):
            distribution = choice(distributions)
            computers["computer-{}".format(i)] = {
                "title": "Desktop {}".format(i),
                "hostname": "desktop-{}".format(i),
                "tags": ["desktop", distribution],
                "distribution": distribution,
                "access-group": "desktop",
            }
        self.computers.update(_get_with_account(self._account, computers))


class SampleComputerProfileUpside(ComputerProfile):
    """Computer profile for the upside account."""

    def __init__(self):
        super(SampleComputerProfileUpside, self).__init__()
        self.computers.update(deepcopy(COMPUTERS_UPSIDE))


class SampleComputerProfileUtility(ComputerProfile):
    """Computer profile for the Utility account."""

    def __init__(self):
        super(SampleComputerProfileUtility, self).__init__()
        self.computers.update(deepcopy(COMPUTERS_UTILITY))


class SampleComputerProfileJuju(ComputerProfile):
    """Computer profile for creating juju computers."""

    def __init__(self, account):
        super(SampleComputerProfileJuju, self).__init__()
        self._account = account

    def add_unit_computers(self, service, count):
        """Add the given count of unit-associated computers for a service."""
        # XXX Set container_info to lxc in order to use a container
        # license if available, so we can build sample data in a
        # regular landscape-server deployment with the free license,
        # except for a few services that will get full licenses (the
        # free license has 10 basic seats and 50 container seats)
        computers = {}
        if service in ("nova-compute", "ceph-osd", "keystone"):
            container_info = None
        else:
            container_info = "lxc"
        for i in range(count):
            if container_info:
                title = "juju-machine-{}-lxc-{}".format(self.count, i)
                hostname = title
            else:
                title = _faker.domain_word()
                hostname = "{}.{}".format(title, _faker.domain_name().split(".")[0])
            computers[make_secure_id()] = {
                "title": hostname,
                "hostname": hostname,
                "container-info": container_info,
                "tags": ["juju"],
            }
        self.computers.update(_get_with_account(self._account, computers))


def _get_with_account(account, computers):
    """Set the account name in the given computers."""
    # Prefix the secure ID with account name (which will be either
    # 'onward' or 'standalone').
    for secure_id in list(computers.keys()):
        # Prefix the secure ID with account name (which will be either
        # 'onward' or 'standalone').
        computer = computers.pop(secure_id)
        computer["account"] = account
        computers["{}-{}".format(account, secure_id)] = computer
    return computers


_faker = Faker()
_faker.seed(4321)  # Make test data repeatable
