import transaction

from collections import defaultdict

from storm.uri import URI

from canonical.landscape.schema.utils import (
    user_exists,
    create_user,
    set_password,
    acquire_lock,
    create_database,
    database_exists,
)


def schema_bootstrap(zstorm, password, user="landscape"):
    """Ensure that the Landscape PostgreSQL user and databases exist.

    This function will ensure that a 'landscape' user with the given password
    exists across all PosgreSQL instances we're configured with (stores can
    live in different PostgreSQL instances for more performant sharding) and
    that the required databases exist on those PostgreSQL instances.

    @param zstorm: The C{ZStorm} instance holding configuration for the various
       stores, including the postgres one, see L{canonical.landscape.setup}.
    """
    # Map Store instances for the various PostgreSQL hosts we're configured
    # with to the list of database names that should live on that host.
    databases_by_postgres_store = defaultdict(list)

    first = True
    for name, default_uri in sorted(zstorm.get_default_uris().items()):
        if name.startswith("postgres-"):
            if first:
                postgres_store = zstorm.get(name)
                # Let's try to get a mutex on the first postgres store we
                # find. Since every other schema bootstrap process racing with
                # us will loop over the default URIs in the same order (see the
                # sorted() call above), if we gain this mutex on the first
                # postgres database, then we're good to go on all of them (any
                # concurrent process will block here until we're done).
                acquire_lock(postgres_store)
            continue

        # Add this non-postgres database to the list of the databases of the
        # associated postgres store.
        uri = URI(default_uri)
        postgres_store = zstorm.get("postgres-%s" % (uri.host or "localhost"))
        databases = databases_by_postgres_store[postgres_store]
        databases.append(uri.database)

    for postgres_store, databases in databases_by_postgres_store.items():
        _ensure_user(postgres_store, user, password)
        _ensure_databases(postgres_store, databases)

    transaction.commit()


def _ensure_user(postgres_store, user, password):
    """Ensure that a PostgreSQL instance has the given user/password.

    @param postgres_store: A L{Store}s connected the the PostgreSQL instance
        we want to check.
    """
    if not user_exists(postgres_store, user):
        create_user(postgres_store, user, password)
    else:
        set_password(postgres_store, user, password)


def _ensure_databases(postgres_store, databases):
    """Ensure that the given databases exist on the given PostgreSQL instance.

    @param postgres_store: A L{Store}s connected the the PostgreSQL instance
        we want to check.
    @param databases: A list of database names that should exist on this
        instance.
    """
    for database in databases:
        if not database_exists(postgres_store, database):
            create_database(postgres_store, database)
