# Copyright 2005-2015 Canonical Limited.  All rights reserved.

import transaction

from prettytable import from_db_cursor


class MoreThanOneConnectionException(Exception):
    def __init__(self, database_name, row_count, table):
        super(MoreThanOneConnectionException, self).__init__(
            "Found %d connections to the database '%s', bailing out:\n%s"
            % (row_count, database_name, table)
        )
        self.database_name = database_name
        self.connections = row_count


def ensure_no_other_connections(zstorm):
    """
    Ensure that we're the only ones connected the Landscape stores in the
    given ZStorm.
    """
    # Filter out stores connecting as superuser (the one whose name starts
    # with 'postgres-'). Also, Sort the stores by name to ensure reliable
    # tests.
    name_stores = []
    for name, store in zstorm.iterstores():
        if not name.startswith("postgres-"):
            name_stores.append((name, store))
    name_stores.sort(key=lambda x: x[0])

    # First block further connections across all our databases
    for name, store in name_stores:
        store.execute(
            "UPDATE pg_database SET datallowconn = false"
            " WHERE datname = current_database()"
        )
    transaction.commit()

    # Secondly ensure that no other process except us is connected.
    for name, store in name_stores:
        activity_result = store.execute(
            "SELECT * FROM pg_stat_activity"
            " WHERE datname = current_database()"
            "  AND usename != 'postgres'"
        )  # Don't consider postgres activity
        if activity_result.rowcount > 1:
            # The DB name is used just to help out the admins figure out what
            # is going on.
            result = store.execute("SELECT current_database()")
            database_name = result.get_one()[0]  # It's a one-tuple
            # Don't leave the databases locked!
            allow_further_connections(zstorm)
            # bail out with an error
            table = from_db_cursor(activity_result._raw_cursor)
            raise MoreThanOneConnectionException(
                database_name, activity_result.rowcount, table
            )


def allow_further_connections(zstorm):
    """Set allowed connections back to True."""
    # Sort the stores by name to ensure reliable tests.
    name_stores = list(zstorm.iterstores())
    name_stores.sort(key=lambda x: x[0])

    for name, store in name_stores:
        if name.startswith("postgres-"):
            continue
        store.execute(
            "UPDATE pg_database SET datallowconn = true"
            " WHERE datname = current_database()"
        )
    transaction.commit()
