# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import os.path
import subprocess
import transaction

from glob import glob

from canonical.landscape.model.main.serveridentity import (
    ensure_server_uuid_exists,
    get_server_uuid,
)
from canonical.lib.text import password_decode
from canonical.landscape.schema.utils import (
    get_database_name,
    get_tables,
    is_table_empty,
)

PG_RESTORE_BINARY = "/usr/bin/pg_restore"
HASHID_STORES_PATH = "/var/lib/landscape/hash-id-databases"
PACKAGE_STOCK_DATA_FILENAME = os.path.join(
    HASHID_STORES_PATH, "landscape-standalone-package.sql"
)


def load_hash_id_databases(
    zstorm,
    config,
    stores_path=HASHID_STORES_PATH,
    dump_filename=PACKAGE_STOCK_DATA_FILENAME,
):
    """
    Loads package hash IDs database dumps and caches from disk.

    Attempts to load a pg_dump from PACKAGE_STOCK_DATA_FILENAME.
    If there is already some data in the package database, it exits.

    Then it renames the pre-generated "uuid_" sqlite downloadable caches
    in HASHID_STORES_PATH to match the currently running server UUID.
    If there is no UUID for the current server, it is generated before
    proceeding.
    """

    package_store = zstorm.get("package")
    _load_hash_id_dump(package_store, config, dump_filename)

    main_store = zstorm.get("main")
    ensure_server_uuid_exists(store=main_store)
    server_uuid = str(get_server_uuid(store=main_store))
    result = _rename_stock_store_filenames(stores_path, server_uuid)
    if result > 0:
        print(
            ("Renamed %d stock hash-id stores against uuid %s." % (result, server_uuid))
        )
    else:
        print("No stock hash-id files to rename.")


def _load_hash_id_dump(package_store, config, dump_filename):
    """
    Loads pg_dump from `dump_filename` into "package" store.

    Runs only if all tables in the "package" database are empty.
    Calls pg_restore on the passed-in `dump_filename`.
    """
    all_tables = get_tables(package_store, exclude=["patch"])
    is_empty = all([is_table_empty(package_store, table) for table in all_tables])

    if not is_empty:
        print("Package database already initialized.")
        return

    if not os.path.exists(dump_filename):
        print("No package data to load.")
        return

    # Ensure all the database operations are done before we attempt
    # a pg_restore run: pg_restore's attempt to disable triggers blocks
    # on all queries, including the simplest read-only ones like
    # "select current_database()" from get_database_name() call.
    transaction.commit()

    database_name = get_database_name(package_store)
    _run_pg_restore(config, database_name, dump_filename)


def _rename_stock_store_filenames(stores_path, server_uuid):
    """
    Rename the stock hash-id store files to match the given `server_uuid`.

    Return number of renamed files.
    """
    if not os.path.exists(stores_path):
        return 0
    stock_store_filenames = glob(os.path.join(stores_path, "uuid_*"))
    for stock_store_filename in stock_store_filenames:
        new_store_filename = stock_store_filename.replace("uuid_", "%s_" % server_uuid)
        os.rename(stock_store_filename, new_store_filename)
    return len(stock_store_filenames)


def _run_pg_restore(configuration, database, dump_filename):
    """
    Loads data-only from `dump_filename` into `database` with pg_restore.

    Database connection parameters are extracted from the `configuration`
    object:
      * host/port from 'host' option in [stores] section
      * username from 'store_user' option in [schema] section
      * password from 'store_password' option in [schema] section.

    Raises PgRestoreError on failure.
    """
    env = os.environ.copy()
    env["PGPASSWORD"] = password_decode(configuration.get("schema", "store_password"))
    server = configuration.get("stores", "host")
    host_parts = server.split(":", 1)
    host = host_parts[0]
    args = [
        PG_RESTORE_BINARY,
        "--disable-triggers",
        "--no-owner",
        "--data-only",
        "--host=%s" % host,
        "--username=%s" % configuration.get("schema", "store_user"),
        "--dbname=%s" % database,
        dump_filename,
    ]
    if len(host_parts) > 1:
        args.insert(5, "--port=%s" % host_parts[1])
    pipe = subprocess.Popen(args, env=env)
    exit_code = pipe.wait()
    if exit_code != 0:
        raise PgRestoreError(" ".join(args))
    else:
        # Remove the dump filename on successful restoration.
        os.remove(dump_filename)


class PgRestoreError(Exception):
    """Raised upon execution failure of the pg_restore program."""
