# Copyright 2005-2013 Canonical Limited.  All rights reserved.


from builtins import str
from builtins import object
from datetime import timedelta

from storm.exceptions import TimeoutError
from storm.info import get_cls_info
from storm.tracer import remove_all_tracers, get_tracers, install_tracer
from twisted.python import log

from canonical.landscape.schema.utils import entity_exists, get_table_oid, rename_table


STATEMENT_TIMEOUT = 30 * 60 * 1000  # 30 minutes


class PartitionError(Exception):
    """Raised when the table cannot be partitioned correctly."""


class TablePartitionInfo(object):
    """
    TablePartitionInfo contains information used to create a single table
    in a partitioned table.

    @param master_table - The original or main table in the partitioned
        table set.
    @param table - The name of an individual table in the partitioned table
        set.
    @param check_field - The name of the table field that the table is
        partitioned on.
    @param from_date - The earliest date that will be inserted in this
        table.
    @param to_date - The latest date that will be inserted in this table
        (non-inclusive).
    @param table_exists - True if the tables exists in the database, False
        otherwise.
    """

    def __init__(self):
        self.master_table = None
        self.table = None
        self.check_field = None
        self.from_date = None
        self.to_date = None
        self.table_exists = False


class TablePartitioner(object):
    """
    TablePartitioner creates a partitioned table from a
    non-partitioned table.  Data from the partitioned table is
    migrated to the new format.  The table is partitioned by one table
    per day.

    Once the initial partitioned setup is complete, the update()
    method should continue to be called on a periodic basis to create
    new tables for upcoming days and drop old tables.

    @param master_table_name - The original or main table in the partitioned
        table set.
    @param check_field - The name of the table field that the table is
        partitioned on.
    @param date - The "current" date that days_retention and days_preallocated
        is centered around.
    @param days_retention - The number of days (tables) that are retained in
        the past.  Tables older than this are dropped.
    @param days_preallocated - The number of days in the future to create
        tables for.
    @param store - The Storm database store used for database access.
    @param grant_access_user - The database user that should have access to
        newly created tables.
    """

    def __init__(
        self,
        model,
        check_field,
        date,
        days_retention,
        days_preallocated,
        store,
        grant_access_user,
    ):
        self.model = model
        self.master_table_name = str(model.__storm_table__)
        self.insert_trigger_name = self.master_table_name + "_insert_trigger"
        self.check_field = check_field
        self.date = date
        self.days_retention = days_retention
        self.days_preallocated = days_preallocated
        self.store = store

    def formatted(self, name, day):
        """
        Given a table name and a date, return the name of the corresponding
        partition table.

        @param name - The base name of the table
        @param day - The date

        Eg: A name of "log" and a date of 1975-01-28 would result in
        "log_19750128".
        """
        return name + day.strftime("_%Y%m%d")

    def _create_child_table(
        self, table_name, check_field, from_date, to_date, master_table_name
    ):
        """
        Create a partition table that inherits from the main table with the
        correct table constraints.
        """
        template = """CREATE TABLE %s (
            LIKE %s INCLUDING DEFAULTS INCLUDING CONSTRAINTS INCLUDING INDEXES,
            CHECK ( %s >= DATE ? AND %s < DATE ? )
            ) INHERITS (%s)
            """
        create = template % (
            table_name,
            master_table_name,
            check_field,
            check_field,
            master_table_name,
        )
        self.store.execute(create, (from_date, to_date))
        self._copy_foreign_keys(master_table_name, table_name)
        self._grant_access(table_name)

    def _copy_foreign_keys(self, from_table, to_table):
        """
        CREATE TABLE LIKE loses foreign keys.  Explicity copy them from
        from_table to to_table.
        """
        from_table_oid = get_table_oid(self.store, from_table)
        key_query = """SELECT
            conname, pg_catalog.pg_get_constraintdef(r.oid, true) AS condef
            FROM pg_catalog.pg_constraint r
            WHERE r.conrelid = ? AND r.contype = 'f'
        """
        result = self.store.execute(key_query, [from_table_oid])
        for row in result.get_all():
            # Use the new table name as fkey name.
            new_fkey_name = row[0].replace(from_table, to_table, 1)
            key_defn = row[1]
            query = """ALTER TABLE %s ADD CONSTRAINT "%s" %s"""
            self.store.execute(query % (to_table, new_fkey_name, key_defn))

    def _grant_access(self, table):
        template = "GRANT SELECT, INSERT, DELETE ON %s TO %s"
        self.store.execute(template % (table, "landscape"))

    def _drop_old_tables(self):
        """
        Delete tables older than C{retention_days} back from the specified
        date.
        """
        oldest_date = self.date - timedelta(days=self.days_retention)
        oldest_table = self.formatted(self.master_table_name, oldest_date)

        result = self.store.execute(
            "SELECT tablename FROM pg_tables WHERE "
            "schemaname='public' AND tablename LIKE ? "
            "AND tablename < ?",
            (self.master_table_name + "_%", oldest_table),
        )

        # Do separate transactions, as table dropping aggressively locks
        # tables and we want gracefully recover from timeout errors.
        self.store.commit()

        try:
            for table in result:
                self.store.execute("DROP TABLE %s" % table)
                self.store.commit()
        except TimeoutError as e:
            log.err("Timeout dropping table partition {}: {}".format(table[0], e))
            self.store.rollback()

    def _performance_ordering(self, infos):
        """Return a new list of infos ordered to give the best performance.

        The optimal ordering is:
        - Today
        - Tomorrow
        - The next day
        - ...
        - Yesterday
        - The day before
        - ...

        Today will not appear in the list if the table is newly migrated.  In
        this case the list will remain unchanged.
        """
        infos = sorted(infos, key=lambda info: info.from_date)
        if not infos:
            return infos

        days = [info.from_date for info in infos]
        if self.date < days[0]:
            return infos
        if self.date > days[-1]:
            return list(reversed(infos))

        if self.date in days:
            today_idx = days.index(self.date)
        else:
            today_idx = 0
        past_days = infos[:today_idx]
        past_days.reverse()
        future_days = infos[today_idx:]

        return future_days + past_days

    def _create_insert_function(self, insert_trigger_name, infos):
        """
        Create the function that inserts data into the correct partition
        table.  A insert trigger should also be created that references
        this function.
        """
        if not infos:
            return

        reordered_infos = self._performance_ordering(infos)

        lines = []
        first = True
        for info in reordered_infos:
            if first:
                keyword = "IF"
                first = False
            else:
                keyword = "ELSIF"

            template = """%s ( NEW.%s >= DATE '%s' AND
                NEW.%s < DATE '%s' ) THEN
                    INSERT INTO %s VALUES (NEW.*);
                """
            lines.append(
                template
                % (
                    keyword,
                    info.check_field,
                    info.from_date,
                    info.check_field,
                    info.to_date,
                    info.table,
                )
            )
        prefix = (
            """CREATE OR REPLACE FUNCTION %s()
            RETURNS TRIGGER AS $$
            BEGIN
            """
            % insert_trigger_name
        )
        suffix = (
            """ELSE
            RAISE NOTICE 'Date out of range.  Discarded by %s.';
            END IF;
            RETURN NEW;
            END;
            $$
            LANGUAGE plpgsql;
            """
            % insert_trigger_name
        )
        lines.insert(0, prefix)
        lines.append(suffix)
        statement = "\n".join(lines)
        self.store.execute(statement)

        # Delete row from the main table, as it has been added to one of the
        # child tables. We can't RETURN NULL from the BEFORE trigger as it
        # would break Storm which uses RETURNING on INSERTs.
        primary_key = (column.name for column in get_cls_info(self.model).primary_key)
        conditions = " AND ".join("%s = NEW.%s" % (key, key) for key in primary_key)

        statement = """CREATE OR REPLACE FUNCTION %(trigger)s_after()
            RETURNS TRIGGER AS $$
            BEGIN
              DELETE FROM ONLY %(table)s WHERE %(conditions)s;
              RETURN NULL;
            END;
            $$
            LANGUAGE plpgsql;
            """ % {
            "trigger": insert_trigger_name,
            "table": info.master_table,
            "conditions": conditions,
        }
        self.store.execute(statement)

    def _insert_trigger_exists(self, name):
        """Return C{True} if a trigger with the specified name exist."""
        result = self.store.execute("SELECT 1 FROM pg_trigger WHERE tgname=?", [name])
        return result.get_one() is not None

    def _create_insert_trigger(self, master_table_name, insert_trigger_name):
        """
        Create an insert trigger that redirects writes to the main table to
        be processed by the database function defined elsewhere.
        """
        create = """CREATE TRIGGER %s
            BEFORE INSERT ON %s
            FOR EACH ROW EXECUTE PROCEDURE %s();
            """ % (
            insert_trigger_name,
            master_table_name,
            insert_trigger_name,
        )
        self.store.execute(create)
        create = """CREATE TRIGGER %s_after
            AFTER INSERT ON %s
            FOR EACH ROW EXECUTE PROCEDURE %s_after();
            """ % (
            insert_trigger_name,
            master_table_name,
            insert_trigger_name,
        )
        self.store.execute(create)

    def migration_needed(self):
        """
        Return C{True} if no partition tables are present, C{False} otherwise.
        """
        result = self.store.execute(
            "SELECT 1 FROM pg_tables WHERE " "schemaname='public' AND tablename LIKE ?",
            (self.master_table_name + "_%",),
        )
        return result.get_one() is None

    def migrate_existing_data(self, date, check_field):
        """
        Handle migrating from a non-parititioned schema to a partitioned
        one.  There may be existing data in the main table, so rename it to
        the partition table for the current day and recreate the main table.

        Note that the table will have a mixture of data, so performance
        will likely be negatively impacted until this table ages and is
        deleted.

        @param date - The date to use as a suffix on the migrated table.
        @param check_field - The name of the field to add date range checking
            to.
        """
        # Production uses a TimeoutTracer which interferes with the ability
        # to run long queries.  Preserve CaptureTracer because the tests use
        # it.
        old_tracers = get_tracers()
        # Use the CaptureTracer class name as a string.  We do not want to
        # have production code depend on python-fixtures, a test library.
        new_tracers = [
            t for t in old_tracers if t.__class__.__name__ == "CaptureTracer"
        ]
        remove_all_tracers()
        for tracer in new_tracers:
            install_tracer(tracer)

        try:
            today_table = self.formatted(self.master_table_name, date)
            self.store.execute("SET statement_timeout=%i" % STATEMENT_TIMEOUT)
            self._delete_old_data(self.master_table_name, check_field, date)
            # Create a new master table, and make the previous table inherit
            # from it. We use a temporary name for the master so everything
            # related to the table evenutally gets properly renamed.
            temp_name = self.master_table_name + "_master"
            self._swap_master_table(
                temp_name, self.master_table_name, check_field, date
            )
            rename_table(self.store, self.master_table_name, today_table)
            rename_table(self.store, temp_name, self.master_table_name)

            # If the table has a sequence for the id column, the owner of the
            # sequence must be updated. This will eventually go away, once all
            # partitioned resource tables will have the id column dropped.
            index_exists = entity_exists(
                self.store, "%s_id_seq" % self.master_table_name, kind="S"
            )
            if index_exists:
                self.store.execute(
                    "ALTER SEQUENCE %(table)s_id_seq OWNED BY %(table)s.id"
                    % {"table": self.master_table_name}
                )

            self._copy_foreign_keys(today_table, self.master_table_name)
            self._grant_access(self.master_table_name)
            self.store.execute("RESET statement_timeout")
        finally:
            remove_all_tracers()
            for tracer in old_tracers:
                install_tracer(tracer)

    def _generate_info_list(self):
        """
        _generate_info_list is a helper function that sets up several variables
        for each table in the partition.
        """
        info_list = []

        date = self.date - timedelta(days=self.days_retention)
        date_stop = self.date + timedelta(days=self.days_preallocated)
        oldest_table_found = False
        while date <= date_stop:
            info = TablePartitionInfo()
            info.master_table = self.master_table_name
            info.check_field = self.check_field
            info.table = self.formatted(self.master_table_name, date)
            info.from_date = date
            info.to_date = date + timedelta(days=1)
            info.table_exists = entity_exists(self.store, info.table)

            if info.table_exists and not oldest_table_found:
                # Handle the newly migrated table case, where the complete
                # list of historical tables may not yet exist.
                info.from_date = self.date - timedelta(days=self.days_retention)
                oldest_table_found = True

            if oldest_table_found:
                info_list.append(info)

            date += timedelta(days=1)

        return info_list

    def _delete_old_data(self, table, check_field, date):
        """Delete data older than retention date."""
        query = "DELETE FROM %s WHERE %s > DATE ?" % (
            self.master_table_name,
            check_field,
        )
        self.store.execute(query, [date])

    def _swap_master_table(self, name, origin, check_field, date):
        """
        Create a new table with the specified name based on the origin one, and
        make it the master table (updating the origin to inherit from it.
        """
        create = """
            CREATE TABLE %s (LIKE %s
              INCLUDING DEFAULTS INCLUDING CONSTRAINTS INCLUDING INDEXES)
        """
        self.store.execute(create % (name, origin))
        self.store.execute("ALTER TABLE %s INHERIT %s" % (origin, name))
        add_constraint = """
            ALTER TABLE %s ADD CONSTRAINT %s_timestamp_check
            CHECK ( %s < DATE ? )"""
        self.store.execute(
            add_constraint % (origin, origin, check_field), [date + timedelta(1)]
        )

    def update(self):
        """
        Perform a set of database operations that migrate from a
        non-partitioned table layout to a partitioned layout, and then
        maintain the partition tables as needed thereafter.
        """
        if self.migration_needed():
            if self.days_preallocated < 1:
                raise PartitionError(
                    "days_preallocated must be at least 1 "
                    "to migrate to partition schema."
                )
            self.migrate_existing_data(self.date + timedelta(1), self.check_field)

        info_list = self._generate_info_list()
        for info in info_list:
            if not info.table_exists:
                self._create_child_table(
                    info.table,
                    info.check_field,
                    info.from_date,
                    info.to_date,
                    info.master_table,
                )

        self._create_insert_function(self.insert_trigger_name, info_list)
        if not self._insert_trigger_exists(self.insert_trigger_name):
            self._create_insert_trigger(
                self.master_table_name, self.insert_trigger_name
            )

        self._drop_old_tables()
