# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Add stocked procedure to insert/update session_data"""
from __future__ import print_function

statements = [
    """
DROP TABLE IF EXISTS session_pkg_data
""",
    """
DROP TABLE IF EXISTS session_data
""",
    """
CREATE TABLE session_data (
    client_id BYTEA PRIMARY KEY,
    generation INTEGER NOT NULL DEFAULT 0,
    access_time TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE 'UTC'))
""",
    """
GRANT SELECT, INSERT, UPDATE, DELETE
    ON TABLE session_data TO landscape, landscape_maintenance
""",
    """
CREATE TABLE session_pkg_data (
    client_id BYTEA NOT NULL,
    generation INTEGER NOT NULL,
    product_id BYTEA NOT NULL,
    key BYTEA NOT NULL,
    data BYTEA NOT NULL,
    PRIMARY KEY (client_id, generation, product_id, key))
""",
    """
GRANT SELECT, INSERT, UPDATE, DELETE
    ON TABLE session_pkg_data TO landscape, landscape_maintenance
""",
    """
CREATE OR REPLACE LANGUAGE plpgsql
""",
    """
CREATE FUNCTION get_generation(provided_client_id BYTEA,
                               update_access_time INTERVAL,
                               update_generation INTERVAL) RETURNS INTEGER AS
$$
DECLARE
    session_record RECORD;
BEGIN
    -- Do we have a fresh generation which has a recent access time?
    SELECT INTO session_record generation FROM session_data
        WHERE session_data.client_id = provided_client_id AND
              session_data.access_time >
              now() AT TIME ZONE 'UTC' - update_access_time;

    IF FOUND THEN
        -- Fantastic!  That's the cheapest as no updating is necessary.
        RETURN session_record.generation;
    END IF;

    -- No luck.  Either we'll have to update an older one, or create it.
    -- For doing that we lock the row to prevent race conditions.
    SELECT INTO session_record generation, access_time FROM session_data
        WHERE session_data.client_id = provided_client_id FOR UPDATE;

    IF FOUND THEN

        -- Ok, there is an existing row.  Now we have to figure if the
        -- row has expired, or if it just needs to get a new access time.
        -- Note that for this test to work correctly, and thus the whole
        -- system, it's important that the isolation level be set to
        -- READ COMMITTED (or AUTOCOMMIT).  Otherwise, even though we've
        -- locked the row above, we're looking at old data, and given
        -- the right conditions, this could allow for two different
        -- generations to be used in parallel in a short interval of time.
        IF session_record.access_time <
           now() AT TIME ZONE 'UTC' - update_generation THEN

            -- Yeah, it's pretty old.  Let's bump the generation.
            UPDATE session_data SET generation=generation+1,
                                    access_time=now() AT TIME ZONE 'UTC'
                WHERE session_data.client_id = provided_client_id;

            RETURN session_record.generation+1;

        ELSE

            -- It's recent enough.  Just update the access time.
            UPDATE session_data SET access_time=now() AT TIME ZONE 'UTC'
                WHERE session_data.client_id = provided_client_id;

            RETURN session_record.generation;

        END IF;
    END IF;

    -- There's no row whatsoever.  We'll have to create it.
    BEGIN
        INSERT INTO session_data (client_id) VALUES (provided_client_id);
    EXCEPTION
        WHEN unique_violation THEN
            NULL; -- Someone else won the race.  We don't care.
    END;
    -- Whatever happened, it's the first generation.
    RETURN 0;
END;
$$
LANGUAGE plpgsql
""",
    """
GRANT EXECUTE ON FUNCTION get_generation(BYTEA, INTERVAL, INTERVAL)
    TO landscape, landscape_maintenance
""",
]


def apply(store):
    print(__doc__)
    for statement in statements:
        store.execute(statement)
