# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Replace free_space.id with a composite key."""
from __future__ import print_function


STATEMENTS = [
    """
    ALTER TABLE %(table)s DROP COLUMN id CASCADE
    """,
    """
    ALTER TABLE %(table)s
        ADD CONSTRAINT %(table)s_mount_info_id_timestamp_key
        PRIMARY KEY (mount_info_id, timestamp)
    """,
    """
    DROP INDEX IF EXISTS %(table)s_mount_info_id_idx
    """,
    """
    DROP INDEX IF EXISTS %(table)s_timestamp_idx
    """,
]


def apply(store):
    print(__doc__)

    # We need to change the schema both on the parent table and on the
    # children ones
    tables = ["free_space"]
    result = store.execute(
        "SELECT tablename FROM pg_tables "
        "WHERE schemaname='public' AND tablename LIKE 'free_space_%'"
    )
    tables.extend(row[0] for row in result.get_all())
    partitioned = len(tables) > 1

    for table in tables:
        # If the table has not been partitioned, discard duplicates in it,
        # otherwise act on each child table and exclude the parent, as it
        # doesn't have own data.
        if not partitioned or (partitioned and table != "free_space"):
            _drop_duplicates(store, table)

        _patch_table(store, table)


def _drop_duplicates(store, table):
    """
    Keep only one row for each (computer_id, interface, timestamp), since it
    will become a primary key.
    """
    result = store.execute(
        "SELECT array_agg(id) FROM %s GROUP BY mount_info_id, "
        "timestamp HAVING count(id) > 1" % table
    )
    ids = []
    for row in result:
        ids.extend(sorted(row[0])[:-1])  # Keep the last inserted row
    n = len(ids)
    if n > 0:
        markers = ",".join(["?"] * n)
        store.execute("DELETE FROM %s WHERE id IN (%s)" % (table, markers), ids)


def _patch_table(store, table):
    """Run schema changes on table"""
    for statement in STATEMENTS:
        store.execute(statement % dict(table=table))
