# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Replace network_traffic.id with a composite key."""
from __future__ import print_function

STATEMENTS = [
    """
    ALTER TABLE %(table)s DROP COLUMN id CASCADE
    """,
    """
    ALTER TABLE %(table)s
        ADD CONSTRAINT %(table)s_network_traffic_pkey
        PRIMARY KEY (computer_id, interface, timestamp)
    """,
]


def apply(store):
    print(__doc__)

    # We need to change the schema both on the parent table and children ones
    tables = ["network_traffic"]
    result = store.execute(
        "SELECT tablename FROM pg_tables "
        "WHERE schemaname='public' AND tablename LIKE 'network_traffic_%'"
    )
    tables.extend(row[0] for row in result.get_all())
    partitioned = len(tables) > 1

    for table in tables:
        # If the table has not been partitioned, discard duplicates in it,
        # otherwise act on each child table and exclude the parent, as it
        # doesn't have own data.
        if not partitioned or (partitioned and table != "network_traffic"):
            _drop_duplicates(store, table)

        _patch_table(store, table)

    # Replace the trigger on the network_traffic table if present
    trigger_name = "network_traffic_insert_trigger_after"
    result = store.execute("SELECT 1 FROM pg_trigger WHERE tgname = ?", [trigger_name])
    if result.get_one() is not None:
        store.execute(
            """
         CREATE OR REPLACE FUNCTION %s()
           RETURNS TRIGGER AS $$
             BEGIN
               DELETE FROM ONLY network_traffic
                 WHERE computer_id = NEW.computer_id
                 AND interface = NEW.interface
                 AND timestamp = NEW.timestamp;
               RETURN NULL;
             END;
           $$
           LANGUAGE plpgsql"""
            % trigger_name
        )


def _drop_duplicates(store, table):
    """
    Keep only one row for each (computer_id, interface, timestamp), since it
    will become a primary key.
    """
    result = store.execute(
        "SELECT array_agg(id) FROM %s GROUP BY computer_id, interface, "
        "timestamp HAVING count(id) > 1" % table
    )
    ids = []
    for row in result:
        ids.extend(sorted(row[0])[:-1])  # Keep the last inserted row
    n = len(ids)
    if n > 0:
        markers = ",".join(["?"] * n)
        store.execute("DELETE FROM %s WHERE id IN (%s)" % (table, markers), ids)


def _patch_table(store, table):
    """Run schema changes on table"""
    for statement in STATEMENTS:
        store.execute(statement % dict(table=table))
