# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Replace id column on resource tables with a composite key."""
from __future__ import print_function


STATEMENTS = [
    """
    ALTER TABLE %(table)s DROP COLUMN id CASCADE
    """,
    """
    DROP INDEX IF EXISTS %(table)s_computer_id_idx
    """,
    """
    ALTER TABLE %(table)s
        ADD CONSTRAINT %(table)s_pkey
        PRIMARY KEY (%(primary_key)s)
    """,
]

TABLES = {
    ("computer_id", "timestamp"): [
        "ceph_usage",
        "cpu_usage",
        "load_average",
        "memory_info",
    ],
    ("thermal_zone_id", "timestamp"): ["temperature"],
    ("custom_graph_data_id", "timestamp"): ["custom_graph_data_point"],
}


def apply(store):
    print(__doc__)

    for primary_key, tables in TABLES.items():
        for table in tables:
            _apply(store, table, primary_key)

    store.execute(
        "DROP INDEX IF EXISTS " "custom_graph_data_point_custom_graph_data_id_idx"
    )


def _apply(store, table_name, primary_key):
    """
    Apply the patch to the specified table and its children (if they exist).

    @param store: the store to execute queries on
    @param table_name: name of the parent table
    @param primary_key: a tuple with composite primary key columns
    """

    # We need to change the schema both on the parent table and children ones
    tables = [table_name]
    result = store.execute(
        "SELECT tablename FROM pg_tables "
        "WHERE schemaname='public' AND tablename LIKE '%s_%%'" % table_name
    )
    tables.extend(row[0] for row in result.get_all())
    partitioned = len(tables) > 1

    _drop_old_primary_key(store, table_name, tables)

    for table in tables:
        # If the table has not been partitioned, discard duplicates in it,
        # otherwise act on each child table and exclude the parent, as it
        # doesn't have own data.
        if not partitioned or (partitioned and table != table_name):
            _drop_duplicates(store, table, primary_key)

        _patch_table(store, table, primary_key)

    # Replace the trigger on the table if present
    trigger_name = "%s_insert_trigger_after" % table_name
    result = store.execute("SELECT 1 FROM pg_trigger WHERE tgname = ?", [trigger_name])

    where = " AND ".join("%s = NEW.%s" % (key, key) for key in primary_key)
    if result.get_one() is not None:
        store.execute(
            """
         CREATE OR REPLACE FUNCTION %(trigger)s()
           RETURNS TRIGGER AS $$
             BEGIN
               DELETE FROM ONLY %(table)s
                 WHERE %(where)s;
               RETURN NULL;
             END;
           $$
           LANGUAGE plpgsql"""
            % dict(trigger=trigger_name, table=table_name, where=where)
        )


def _drop_old_primary_key(store, table_name, tables):
    """
    Drop primary keys based on the table 'id' column, so the name can be reused
    for the composite key.

    @param store: the store to execute queries on
    @param table_name: name of the parent table
    @param tables: names of the tables to look for the constraint to drop
    """
    for table in tables:
        store.execute(
            "ALTER TABLE %(table)s DROP CONSTRAINT IF EXISTS %(name)s_pkey"
            % dict(table=table, name=table_name)
        )


def _drop_duplicates(store, table, primary_key):
    """
    Keep only one row for each row with same value for columns in the the
    composite primary key.
    """
    group_by = ",".join(primary_key)
    result = store.execute(
        "SELECT array_agg(id) FROM %(table)s GROUP BY %(group_by)s "
        "HAVING count(id) > 1" % dict(table=table, group_by=group_by)
    )
    ids = []
    for row in result:
        ids.extend(sorted(row[0])[:-1])  # Keep the last inserted row
    n = len(ids)
    if n > 0:
        markers = ",".join(["?"] * n)
        store.execute("DELETE FROM %s WHERE id IN (%s)" % (table, markers), ids)


def _patch_table(store, table, primary_key):
    """Run schema changes on table"""
    primary_key = ",".join(primary_key)
    for statement in STATEMENTS:
        store.execute(statement % dict(table=table, primary_key=primary_key))
