# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Delete related provisioning activities in synchronize_activities_for_computers.
"""
from __future__ import print_function


STATEMENTS = [
    """
CREATE OR
REPLACE FUNCTION synchronize_activities_for_computers(computer_ids INTEGER[])
    RETURNS VOID AS
$$

# Detect which version of postgres is calling the function
postgres_lessthan_9 = False
if isinstance(computer_ids, str):
    postgres_lessthan_9 = True


# Compatibility function to handle differences between postgres 8 and 9
def querylist_to_pythonlist(input_list):
    if postgres_lessthan_9:
        return list(map(int, input_list[1:-1].split(",")))
    else:
        return list(input_list)


# Compatibility function to handle differences between postgres 8 and 9
def pythonlist_to_querylist(input_list):
    if postgres_lessthan_9:
        return "{%s}" % (",".join(str(element) for element in input_list))
    else:
        return list(input_list)


# Get highest priority status from a set of statuses.
STATUS_PRIORITY = [1, 7, 2, 3, 5, 6, 4]

def get_highest_priority_status(statuses):
    if len(statuses) == 1:
        return statuses.pop()
    else:
        for status in STATUS_PRIORITY:
            if status in statuses:
                return status
    return None

# Get activity ids of parent activities that have children with specified
# computer_ids
select_group_activities_plan = plpy.prepare('''
    SELECT activity.id, activity_group.children_data FROM activity,
    activity_group WHERE activity.id = activity_group.activity_id AND
    activity.info_type = 6 AND activity_group.children_data != '{}' AND
    EXISTS (SELECT 1 FROM activity AS child_activity
    WHERE child_activity.parent_id = activity.id AND
    child_activity.computer_id = ANY($1))
''', ["int4[]"])

select_status_plan = plpy.prepare('''
    SELECT computer_request_data.status FROM computer_request_data, activity
    WHERE computer_request_data.activity_id = activity.id AND
    activity.parent_id = $1
''', ["int4"])

delete_child_activities_plan = plpy.prepare('''
    DELETE FROM activity WHERE computer_id = ANY($1) AND parent_id = $2
''', ["int4[]", "int4"])

delete_activity_plan = plpy.prepare('''
    DELETE FROM activity WHERE id = $1
''', ["int4"])

update_activity_plan = plpy.prepare('''
    UPDATE activity_group SET children_data = $2, status = $3
    WHERE activity_id = $1
''', ["int4", "int4[]", "int4"])

select_remaining_activities = plpy.prepare('''
    SELECT id, parent_id FROM activity WHERE computer_id = ANY($1)
    AND parent_id is not NULL
''', ["int4[]"])

delete_empty_activities_notice = plpy.prepare('''
    DELETE FROM activity WHERE activity.info_type = 6 AND activity.id = ANY($1)
    AND NOT EXISTS (SELECT 1 FROM activity AS child_activity WHERE
    child_activity.parent_id = activity.id)
''', ["int4[]"])

select_provisioned_system_ids = plpy.prepare('''
    SELECT id FROM provisioned_system WHERE computer_id = ANY($1)
''', ["int4[]"])

delete_provisioning_activities = plpy.prepare('''
    DELETE FROM activity WHERE id IN (
        SELECT activity_id FROM provision_system_request
        WHERE provisioned_system_id = ANY($1) UNION
        SELECT activity_id FROM terminate_system_request
        WHERE provisioned_system_id = ANY($1))
''', ["int4[]"])


computer_ids_list = querylist_to_pythonlist(computer_ids)

# First, delete computer request activities
for activity in plpy.execute(select_group_activities_plan, [computer_ids]):
    activity_id = activity["id"]
    children_data_list = querylist_to_pythonlist(activity["children_data"])
    new_children_data = set(children_data_list)
    for computer_id in children_data_list:
        if computer_id in computer_ids_list:
            new_children_data.remove(computer_id)

    plpy.execute(delete_child_activities_plan, [computer_ids, activity_id])

    # If no children are remaining, delete the parent activity
    if not new_children_data:
        plpy.execute(delete_activity_plan, [activity_id])
    else:
        # Update the group with the highest priority status and new
        # children_data
        statuses = list(plpy.execute(select_status_plan, [activity_id]))
        status = get_highest_priority_status([s["status"] for s in statuses])
        new_children_data = pythonlist_to_querylist(new_children_data)
        plpy.execute(update_activity_plan,
                     [activity_id, new_children_data, status])

# Let's take care of notice activities
parent_ids = set()
for activity in plpy.execute(select_remaining_activities, [computer_ids]):
    activity_id = activity["id"]
    parent_id = activity["parent_id"]
    parent_ids.add(parent_id)
    plpy.execute(delete_activity_plan, [activity_id])

parent_ids = pythonlist_to_querylist(parent_ids)
plpy.execute(delete_empty_activities_notice, [parent_ids])

# Remove provisioning activities
provisioned_system_ids = plpy.execute(
    select_provisioned_system_ids, [computer_ids])
provisioned_system_ids = pythonlist_to_querylist(
    [id["id"] for id in provisioned_system_ids])
plpy.execute(delete_provisioning_activities, [provisioned_system_ids])

$$ LANGUAGE plpythonu
"""
]


def apply(store):
    print(__doc__)
    for statement in STATEMENTS:
        store.execute(statement)
