# Copyright 2005-2014 Canonical Limited.  All rights reserved.

"""
Clear inconsistent package information for computers and request resync.
WARNING: This could take several minutes.
"""
from __future__ import print_function

from builtins import map
from builtins import str
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.account.store import get_account_stores


def apply(store):
    print(__doc__)

    # Map computers and their account ids
    computer_account_ids = dict(
        store.execute(
            """SELECT id, account_id FROM computer_twin
            WHERE NOT removed"""
        ).get_all()
    )

    inconsistent_computer_ids = find_inconsistent_computer_ids(
        store, list(computer_account_ids.keys())
    )
    if inconsistent_computer_ids:
        process_inconsistent_computers(
            store, computer_account_ids, inconsistent_computer_ids
        )


def find_inconsistent_computer_ids(store, computer_ids):
    """
    Return a set with ids for computers with inconsistent package information.
    """
    computer_ids = frozenset(computer_ids)
    inconsistent_computer_ids = set()

    # Find computers with inconsistent package information, skipping removed
    # ones.
    result = store.execute(
        """\
        SELECT computer_id FROM (
          SELECT computer_id,
            bool(icount(held - installed) +
                 icount(available_upgrades - available) +
                 icount(available_upgrades & installed))
            AS inconsistent
          FROM computer_packages JOIN computer_twin
            ON computer_packages.computer_id = computer_twin.id
          WHERE NOT removed
        ) AS sub
        WHERE inconsistent"""
    )
    inconsistent_computer_ids.update(row[0] for row in result)

    package_store = get_package_store()
    # Find computers with inconsistent upgraded/upgrades relations.  Skip
    # computers that are already marked as inconsistent.
    for computer_id in computer_ids - inconsistent_computer_ids:
        result = store.execute(
            """SELECT available, available_upgrades, installed
            FROM computer_packages WHERE computer_id = ?""",
            [computer_id],
        ).get_one()
        upgrades_ids = frozenset(result[1])
        installed_ids = frozenset(result[2])

        if not upgrades_ids:
            continue

        if installed_ids:
            result = package_store.execute(
                """\
                SELECT array_agg(upgrades.id)
                FROM package AS upgrades
                JOIN package AS upgraded
                  ON upgraded.name = upgrades.name
                WHERE upgrades.version > upgraded.version
                  AND upgrades.id IN ? AND upgraded.id IN ?
                """,
                (tuple(upgrades_ids), tuple(installed_ids)),
            ).get_one()
            actual_upgrades = result[0]
            if actual_upgrades is None:
                actual_upgrades = []
        else:
            actual_upgrades = []

        if upgrades_ids != frozenset(actual_upgrades):
            inconsistent_computer_ids.add(computer_id)

    return inconsistent_computer_ids


def process_inconsistent_computers(
    store, computer_account_ids, inconsistent_computer_ids
):
    """
    Clear package information for inconsistent computers and create activities
    to ask them to resync.
    """
    inconsistent_computer_ids = tuple(inconsistent_computer_ids)

    account_ids = tuple(account_id for _, account_id in computer_account_ids.items())

    main_store = get_main_store()
    account_ids_to_stores = get_account_ids_to_stores_map(main_store, account_ids)

    # Map computer ids to their current api version.
    computer_current_apis = dict(
        main_store.execute(
            "SELECT id, current_api FROM computer WHERE id IN ?",
            [inconsistent_computer_ids],
        )
    )

    # Create resync requests for inconsistent computers
    for computer_id in inconsistent_computer_ids:
        account_id = computer_account_ids[computer_id]
        account_store = account_ids_to_stores[account_id]
        # Create activity and its info entry
        activity_id = account_store.execute(
            """INSERT INTO activity
            (info_type, account_id, computer_id, pending, status)
            VALUES (4, ?, ?, TRUE, 2)
            RETURNING id""",
            (account_id, computer_id),
        ).get_one()[0]
        account_store.execute(
            """INSERT INTO computer_request_data
            (activity_id, modification_time, computer_id)
            VALUES (?, now(), ?)""",
            [activity_id, computer_id],
        )

        # Older clients don't support scoped resyncs.
        computer_api = computer_current_apis[computer_id]
        supports_scoped_resync = (
            computer_api is not None
            and is_computer_api_compatible(str(computer_api), "3.6")
        )
        if supports_scoped_resync:
            scopes = ["package"]
        else:
            scopes = None

        account_store.execute(
            """INSERT INTO resynchronize_request (activity_id, scopes)
            VALUES (?, ?)""",
            [activity_id, scopes],
        )

    # Mark computers as inconsistent
    main_store.execute(
        """UPDATE computer_status SET is_inconsistent = TRUE
        WHERE computer_id in ?""",
        [inconsistent_computer_ids],
    )

    # Clear computer packages, opened USN issues and buffer for inconsistent
    # computers.
    store.execute(
        """\
        UPDATE computer_packages
          SET available = '{}', available_upgrades = '{}',
            installed = '{}', held = '{}'
        WHERE computer_id IN ?""",
        [inconsistent_computer_ids],
    )
    store.execute(
        """DELETE FROM usn_issue
        WHERE computer_id IN ? AND resolve_time IS NULL""",
        [inconsistent_computer_ids],
    )
    store.execute(
        "DELETE FROM computer_packages_buffer WHERE computer_id IN ?",
        [inconsistent_computer_ids],
    )


def get_account_ids_to_stores_map(main_store, account_ids):
    """Return a dict mapping account ids to their stores."""

    # Map store id to store
    account_stores = {
        int(key.strip("account-")): store for key, store in get_account_stores()
    }

    # Map account id to store id
    account_store_ids = dict(
        main_store.execute(
            "SELECT id, database_key FROM account WHERE id IN ?", [tuple(account_ids)]
        ).get_all()
    )

    return {
        account_id: account_stores[account_store_ids[account_id]]
        for account_id in account_ids
    }


def is_computer_api_compatible(computer_version, version):
    """Compare the current version of a computer API with a version.

    Note: this is a copy of Computer.is_api_compatible().
    """
    ver1, ver2 = [
        tuple(map(int, s.split("."))) for s in (computer_version or "0.0", version)
    ]
    return ver1 >= ver2
