# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Use apt instead of smart in process_relation() trigger"""
from __future__ import print_function

from canonical.landscape.schema.package import SchemaTemplate


STATEMENTS = [
    """
CREATE OR REPLACE FUNCTION process_relation() RETURNS TRIGGER AS
$$
func = SD.get("func")
if func is not None:
    return func(TD["new"])

import apt_pkg

apt_pkg.init()

def parse_value(value):
    # Parses a relation.value into (name, comparison, version).
    # Comparison and version may be None.
    tokens = value.split()
    if len(tokens) == 1:
        return tokens[0], None, None
    return tokens

def deb_matches(provides_name, provides_version,
                relation_name, relation_cmp, relation_version):
    # Checks whether the given 'provides' name and version satisfies the given
    # relation requirements (name, comparison, and version).
    return bool(relation_name == provides_name and
                (not relation_version or
                 provides_version and
                 apt_pkg.check_dep(provides_version,
                                   relation_cmp, relation_version)))

# The query for finding what relation might match the named provides.
select_relations_plan = plpy.prepare('''
    SELECT * FROM relation
    WHERE relation.type IN (@@DEB_REQUIRES, @@DEB_OR_REQUIRES,
                            @@DEB_UPGRADES, @@DEB_CONFLICTS)
      AND relation.id IN (SELECT relation_name.relation_id
                          FROM relation_name, name WHERE
                          relation_name.name_id = name.id AND
                          name.name = $1)
''', ["text"])

# ... for finding what provides might match the named relation.
select_provides_plan = plpy.prepare('''
    SELECT * FROM relation
    WHERE relation.type IN (@@DEB_PROVIDES, @@DEB_NAME_PROVIDES)
      AND relation.id IN (SELECT relation_name.relation_id
                          FROM relation_name, name WHERE
                          relation_name.name_id = name.id AND
                          name.name = $1)
''', ["text"])

# ... for finding what virtual provides might match the named relation.
select_simple_provides_plan = plpy.prepare('''
    SELECT * FROM relation
    WHERE relation.type = @@DEB_PROVIDES
      AND relation.id IN (SELECT relation_name.relation_id
                          FROM relation_name, name WHERE
                          relation_name.name_id = name.id AND
                          name.name = $1)
''', ["text"])

# ... for finding what named provides might match the named relation.
select_name_provides_plan = plpy.prepare('''
    SELECT * FROM relation
    WHERE relation.type = @@DEB_NAME_PROVIDES
      AND relation.id IN (SELECT relation_name.relation_id
                          FROM relation_name, name WHERE
                          relation_name.name_id = name.id AND
                          name.name = $1)
''', ["text"])

insert_relation_plan = plpy.prepare('''
    INSERT INTO provides_relation VALUES ($1, $2)
''', ["int4", "int4"])

insert_relation_name_plan = plpy.prepare('''
    INSERT INTO relation_name VALUES (get_name_id($1), $2)
''', ["text", "int4"])

def func(new):
    # A new relation has been inserted. Find everything that it "relates" to
    # :-) and insert it into the graph at all relevant places.
    type = new["type"]

    # The hard one first: provides are relevant to many kinds of relations.
    if type == @@DEB_PROVIDES or type == @@DEB_NAME_PROVIDES:
        name, cmp, version = parse_value(new["value"])
        for relation in plpy.execute(select_relations_plan, [name]):
            if relation["type"] == @@DEB_OR_REQUIRES:
                for sub_value in relation["value"].split("|"):
                    if deb_matches(name, version, *parse_value(sub_value)):
                        plpy.execute(insert_relation_plan,
                                     [new["id"], relation["id"]])
                        break
            elif deb_matches(name, version, *parse_value(relation["value"])):
                # upgrades don't care about virtual provides.
                if (relation["type"] != @@DEB_UPGRADES
                    or type == @@DEB_NAME_PROVIDES):
                    plpy.execute(insert_relation_plan,
                                 [new["id"], relation["id"]])
        plpy.execute(insert_relation_name_plan, [name, new["id"]])

    elif type == @@DEB_REQUIRES or type == @@DEB_CONFLICTS:
        value = parse_value(new["value"])
        for provides in plpy.execute(select_provides_plan, [value[0]]):
            name, equals, version = parse_value(provides["value"])
            if deb_matches(name, version, *value):
                plpy.execute(insert_relation_plan, [provides["id"], new["id"]])
        plpy.execute(insert_relation_name_plan, [value[0], new["id"]])

    elif type == @@DEB_UPGRADES:
        # upgrades only care about package name provides.
        value = parse_value(new["value"])
        for provides in plpy.execute(select_name_provides_plan, [value[0]]):
            name, equals, version = parse_value(provides["value"])
            if deb_matches(name, version, *value):
                plpy.execute(insert_relation_plan, [provides["id"], new["id"]])
        plpy.execute(insert_relation_name_plan, [value[0], new["id"]])

    elif type == @@DEB_OR_REQUIRES:
        added = set() # IDs of provides already linked.
        names = set() # names of requires to be registered.
        for sub_value in new["value"].split("|"):
            value = parse_value(sub_value)
            name = value[0]
            names.add(name)
            for provides in plpy.execute(select_provides_plan, [name]):
                # don't duplicate links with the same relation.
                if provides["id"] not in added:
                    name, equals, version = parse_value(provides["value"])
                    if deb_matches(name, version, *value):
                        plpy.execute(insert_relation_plan,
                                     [provides["id"], new["id"]])
                        added.add(provides["id"])
        for name in names:
            plpy.execute(insert_relation_name_plan, [name, new["id"]])

SD["func"] = func
return func(TD["new"])

$$
LANGUAGE plpythonu
"""
]


def apply(store):
    print(__doc__)

    for statement in STATEMENTS:
        store.execute(SchemaTemplate(statement).substitute())
