# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Add an index on operation (root_parent_id) and improve oper trigger."""
from __future__ import print_function


def apply(store):
    print(__doc__)
    for statement in STATEMENTS:
        store.execute(statement)


STATEMENTS = [
    """
CREATE INDEX operation_root_parent_id_idx ON operation (root_parent_id)
""",
    """
DROP TRIGGER operation_modification_time_changed ON operation
""",
    """
DROP FUNCTION update_operation_order_time()
""",
    """
CREATE FUNCTION update_operation_order_time_on_insert()
  RETURNS trigger AS $PROC$
DECLARE
  root_parent record;
BEGIN
    IF new.modification_time IS NULL THEN
        RAISE EXCEPTION 'NULL operation.modification_time not accepted';
    END IF;

    new.order_time = new.modification_time;

    IF new.root_parent_id IS NOT NULL THEN
        SELECT INTO root_parent order_time FROM operation
               WHERE operation.id = new.root_parent_id;
        IF root_parent.order_time > new.order_time THEN
            new.order_time = root_parent.order_time;
        ELSE
            UPDATE operation SET order_time = new.order_time
                   WHERE (operation.root_parent_id = new.root_parent_id OR
                          operation.id = new.root_parent_id) AND
                         operation.order_time < new.order_time;
        END IF;
    END IF;
    RETURN new;
END
$PROC$ LANGUAGE plpgsql;
""",
    """
CREATE FUNCTION update_operation_order_time_on_update()
  RETURNS trigger AS $PROC$
DECLARE
  root_parent_id integer;
  root_parent record;
BEGIN
    -- If someone is updating our order_time externally,
    -- everything is up to whoever is doing it.
    IF old.order_time != new.order_time THEN
        RETURN new;
    END IF;

    IF new.modification_time IS NULL THEN
        RAISE EXCEPTION 'NULL operation.modification_time not accepted';
    END IF;
    IF new.order_time IS NULL THEN
        RAISE EXCEPTION 'NULL operation.order_time not accepted';
    END IF;

    -- If this op is being placed into a new/different graph, it
    -- may need to have its order_time updated because it may be at an
    -- older order_time but relating to a graph with a newer order_time.
    IF (new.root_parent_id IS NOT NULL AND
          (old.root_parent_id IS NULL OR
           old.root_parent_id != new.root_parent_id)) THEN
        SELECT INTO root_parent order_time FROM operation
               WHERE operation.id = new.root_parent_id;
        IF root_parent.order_time > new.order_time THEN
            new.order_time = root_parent.order_time;
        END IF;
    END IF;

    -- Now, if the modification_time is newer than the order_time,
    -- update the order_time for the whole tree.
    IF new.order_time < new.modification_time THEN
        new.order_time = new.modification_time;
        IF new.root_parent_id IS NULL THEN
            UPDATE operation SET order_time = new.order_time
                   WHERE operation.root_parent_id = new.id;
        ELSE
            UPDATE operation SET order_time = new.order_time
                   WHERE (operation.root_parent_id = new.root_parent_id OR
                          operation.id = new.root_parent_id)
                          AND operation.id != new.id -- avoid recursion
                          AND operation.order_time < new.modification_time;
        END IF;
    END IF;

  RETURN new;
END
$PROC$ LANGUAGE plpgsql;
""",
    """
CREATE TRIGGER update_order_time_on_insert
BEFORE INSERT ON operation
  FOR EACH ROW EXECUTE PROCEDURE update_operation_order_time_on_insert();
""",
    """
CREATE TRIGGER update_order_time_on_update
BEFORE UPDATE ON operation
  FOR EACH ROW EXECUTE PROCEDURE update_operation_order_time_on_update();
""",
]
