# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Convert Juju environment names to slug fields."""
from __future__ import print_function


from builtins import str
import re
from unidecode import unidecode


def apply(store):
    print(__doc__)

    existing_names = [
        result[0] for result in store.execute("SELECT name FROM juju_environment")
    ]
    names = set(existing_names)

    def existing_check(slug):
        return (name for name in names if name.startswith(slug))

    for name in existing_names:
        new_name = slug_from_title(name, 20, existing_check)
        names.add(new_name)
        store.execute(
            "UPDATE juju_environment SET name = ? WHERE name = ?", (new_name, name)
        )


# for future-proofing slug_from_title in case signature changes
def slug_from_title(title, max_root_length, existing_check):
    """
    Generate a name suitable for a URL from a free-form title provided by a
    user.

    @param title: The unicode title.
    @param max_root_length: The maximum length of the root to use before adding
        digits for unicity.
    @param existing_check: A callable taking the root used, which returns an
        iterable of names which are already used.
    """
    ascii_title = ""
    for character in title:
        ascii_chars = unidecode(character)
        # In order to avoid the need for a default slug value, we need to
        # ensure everything is encoded. If we encounter an undecodeable
        # character we fall back to IDNA
        if ascii_chars == "":
            ascii_title = title.encode("idna")
            break
        ascii_title += ascii_chars
    # There are a special set of characters that won't be encoded in the IDNA
    # format because they are in the ASCII set, but nevertheless cannot appear
    # in a slug.
    slug = str(re.sub("[^\w\s-]", "", ascii_title).strip())
    slug = re.sub("[-\s]+", "-", slug.lower())
    if slug == "":
        slug = "item"
    root = slug = slug[:max_root_length]
    counter = 0
    existing = existing_check(slug)
    while slug in existing:
        counter += 1
        slug = "%s%d" % (root, counter)
    return slug
