# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""More sensible indexes for the resource database."""
from __future__ import print_function

from storm.locals import Delete, In
from storm.expr import Column


def remove_device_duplicates(store):
    duplicated = store.execute(
        """
SELECT device.computer_id, device.id, duplicates.count, device.udi FROM device,
   (SELECT computer_id, udi, count
    FROM (SELECT computer_id,  udi, COUNT(udi)
          FROM device GROUP BY computer_id, udi)
         AS sub WHERE count > 1) as duplicates
 WHERE duplicates.udi = device.udi
       AND device.computer_id = duplicates.computer_id;
"""
    )
    computer_id_and_udi_to_deleted_count = {}
    deletions = []
    for computer_id, device_id, count, udi in duplicated:
        deleted_so_far = computer_id_and_udi_to_deleted_count.setdefault(
            (computer_id, udi), 0
        )
        if deleted_so_far < count - 1:
            deletions.append(device_id)
            computer_id_and_udi_to_deleted_count[computer_id, udi] += 1
    if deletions:
        store.execute(Delete(In(Column("id"), deletions), "device"))


def apply(store):
    print(__doc__)

    remove_device_duplicates(store)

    for statement in SCHEMA_CHANGES:
        store.execute(statement)


SCHEMA_CHANGES = [
    """
ALTER TABLE device
    ADD CONSTRAINT device_computer_id_udi_key UNIQUE (computer_id, udi)
""",
    """
CREATE INDEX message_computer_id_idx ON message (computer_id)
""",
    """
CREATE INDEX message_info_computer_id_idx ON message_info (computer_id)
""",
    """
CREATE INDEX memory_info_computer_id_idx ON memory_info (computer_id)
""",
    """
CREATE INDEX active_process_computer_id_idx ON active_process (computer_id)
""",
    """
CREATE INDEX historic_process_computer_id_idx ON historic_process (computer_id)
""",
    """
CREATE INDEX load_average_computer_id_idx ON load_average (computer_id)
""",
    """
CREATE INDEX processor_computer_id_idx ON processor (computer_id)
""",
    """
CREATE INDEX thermal_zone_computer_id_idx ON thermal_zone (computer_id)
""",
    """
CREATE INDEX client_uptime_computer_id_idx ON client_uptime (computer_id)
""",
    """
CREATE INDEX mount_info_computer_id_idx ON mount_info (computer_id)
""",
    """
CREATE INDEX computer_uptime_computer_id_idx ON computer_uptime (computer_id)
""",
    """
CREATE INDEX computer_user_computer_id_idx ON computer_user (computer_id)
""",
    """
CREATE INDEX computer_group_computer_id_idx ON computer_group (computer_id)
""",
]
