# Copyright 2005-2015 Canonical Ltd. All rights reserved.

"""
Delete removed computers, related activities and drop computer_twin.removed
from account store.
WARNING: This could take several minutes.
"""
from __future__ import print_function


def apply(store):
    print(__doc__)

    _delete_activities_for_removed_computers(store)
    _delete_removed_computers(store)
    _drop_removed_column(store)


def _delete_removed_computers(store):
    """Delete computers marked as removed."""
    store.execute("DELETE FROM computer_twin WHERE removed")


def _drop_removed_column(store):
    """Drop the computer_twin.removed column."""
    store.execute("ALTER TABLE computer_twin DROP COLUMN removed")


def _delete_activities_for_removed_computers(store):
    """Delete activities related to computers marked as removed.

    This also removed parent activities of the removed ones that are left with
    no children.
    """
    parent_activity_ids = _delete_removed_computer_activities(store)
    if parent_activity_ids:
        _delete_empty_parent_activities(store, parent_activity_ids)


def _delete_removed_computer_activities(store):
    """
    Remove activities for removed computers, and return IDs of their parent
    IDs.
    """
    return tuple(
        set(
            row[0]
            for row in store.execute(
                "DELETE FROM activity"
                " USING computer_request_data, computer_twin"
                " WHERE computer_request_data.computer_id = computer_twin.id AND"
                "       computer_request_data.activity_id = activity.id AND"
                "       computer_twin.removed"
                " RETURNING parent_id"
            )
        )
    )


def _delete_empty_parent_activities(store, parent_activity_ids):
    """Remove specified parent activities if they have no children."""

    # Find which of the parent activities no longer have children.
    parent_ids_with_children = tuple(
        row[0]
        for row in store.execute(
            "SELECT parent_id FROM activity WHERE parent_id IN ?", [parent_activity_ids]
        )
    )
    parent_ids_without_children = tuple(
        set(parent_activity_ids).difference(parent_ids_with_children)
    )

    if parent_ids_without_children:
        store.execute(
            "DELETE FROM activity WHERE id IN ?", [parent_ids_without_children]
        )
