# Copyright 2005-2016 Canonical Limited.  All rights reserved.

"""Add computer-duplicate alert to all accounts."""
from __future__ import print_function

from builtins import zip
from canonical.landscape.model.main.store import get_main_store

# Alert info type code from c.l.model.alert.policy.ComputerDuplicateAlert
ALERT_INFO_TYPE = 402
# Profile info type code from c.l.model.account.register
RULE_INFO_TYPE = 1

SELECT_ACCOUNT_WITH_RULE = """
SELECT DISTINCT account_id FROM rule WHERE info_type = ?
"""

SELECT_ACCESS_CONTEXTS = """
SELECT id, access_context_id FROM account
"""

SELECT_ACCOUNT_TWINS = """
SELECT id FROM account_twin
"""

INSERT_COMPUTER_GROUP = """
INSERT INTO computer_group (account_id, info_type, all_computers)
VALUES (?, ?, ?)
RETURNING id
"""

INSERT_ALERT_RULE = """
INSERT INTO rule (account_id, access_context_id, computer_group_id, info_type)
VALUES (?, ?, ?, ?)
RETURNING id
"""

INSERT_COMPUTER_GROUP_ITEM = """
INSERT INTO computer_group_item (computer_group_id, computer_id)
SELECT ?, computer_twin.id
  FROM computer_twin JOIN account_entity_data
    ON computer_twin.data_id = account_entity_data.id
  WHERE account_entity_data.account_id = ?
"""

INSERT_RULE_STATE = """
INSERT INTO rule_state (rule_id, computer_id)
SELECT ?, computer_twin.id
  FROM computer_twin JOIN account_entity_data
    ON computer_twin.data_id = account_entity_data.id
  WHERE account_entity_data.account_id = ?
"""

INSERT_RULE_STATE_NOTIFICATION = """
INSERT INTO rule_state_notification (rule_state_id)
SELECT id
  FROM rule_state
 WHERE rule_id = ?
"""


def apply(store):
    print(__doc__)

    # Find all the accounts that already have the rule.
    accounts_with_rules = set(
        store.execute(SELECT_ACCOUNT_WITH_RULE, (ALERT_INFO_TYPE,)).get_all()
    )

    # Build a mapping from each account ID to it's global access context ID.
    main_store = get_main_store()
    account_to_access_context = dict(
        main_store.execute(SELECT_ACCESS_CONTEXTS).get_all()
    )

    # Find all the accounts which don't have pre-existing rules.
    accounts = set(account_to_access_context)
    accounts_without_rules = accounts - accounts_with_rules

    # We need to skip any accounts with missing twins.  We don't know why this
    # happens, but this isn't the place to fix it.
    twins = set(*list(zip(*store.execute(SELECT_ACCOUNT_TWINS).get_all())))
    missing_twins = accounts - twins

    # Now add a rule to each account
    for account_id in accounts_without_rules:
        # If there is no account twin, adding the rule will violate a foreign
        # key constraint, so skip this account.
        if account_id in missing_twins:
            continue

        # Make a new computer group containing all computers
        group_id = store.execute(
            INSERT_COMPUTER_GROUP, (account_id, RULE_INFO_TYPE, True)
        ).get_one()[0]

        # Actually add all computers to the computer group
        store.execute(INSERT_COMPUTER_GROUP_ITEM, (group_id, account_id))

        access_context = account_to_access_context[account_id]

        # Make a rule for the ComputerDuplicateAlert
        result = store.execute(
            INSERT_ALERT_RULE, (account_id, access_context, group_id, ALERT_INFO_TYPE)
        )
        rule_id = result.get_one()[0]

        # Add activation records for all computers
        store.execute(INSERT_RULE_STATE, (rule_id, account_id))

        # Add notification records for the rule states)
        store.execute(INSERT_RULE_STATE_NOTIFICATION, (rule_id,))
