from __future__ import print_function

# Copyright 2005-2016 Canonical Ltd. All rights reserved.

from builtins import str
import json

from canonical.landscape.schema.patch_395 import __doc__


EVENT_STATUS_TRANSITION = 1

SUCCEEDED = 1
QUEUED = 4
INPROGRESS = 5
BLOCKED = 6
CANCELED = 7
FAILED = 8
UNAPPROVED = 1001

FINAL_STATUSES = (SUCCEEDED, FAILED, CANCELED)


def apply(store):
    print(__doc__)

    # Use a prepared statement to speed up inserts, since the same query is
    # repeated many times.
    store.execute(
        """
        PREPARE insert_event AS
          INSERT INTO activity_child_event (
              activity_id, timestamp, event_type, data)
            VALUES ($1, $2, $3, $4)
        """
    )

    result = store.execute(
        """
        SELECT id, status, completion_time, approval_time, delivery_time,
               modification_time
        FROM activity LEFT JOIN computer_request_data
          ON activity.id = computer_request_data.activity_id
        """
    )
    for row in result:
        (
            activity_id,
            status,
            completion_time,
            approval_time,
            delivery_time,
            modification_time,
        ) = row

        # The UNAPPROVED status is really a case of BLOCKED. Eventually the
        # former will be dropped.
        if status == UNAPPROVED:
            status = BLOCKED

        # Only computer activities have modification time
        is_computer_activity = modification_time is not None

        has_entries = False

        if completion_time and status in FINAL_STATUSES:
            old_status = None

            if status == CANCELED:
                # Computer activities can only be canceled if UNAPPROVED or
                # QUEUED.
                if delivery_time:
                    # The activity has been started and became BLOCKED
                    # (UNAPPROVED).
                    old_status = BLOCKED
                elif is_computer_activity:
                    old_status = QUEUED
            else:
                old_status = INPROGRESS

            insert_event(
                store, activity_id, completion_time, status, old_status=old_status
            )
            has_entries = True
        else:
            if approval_time:
                # The approval time is the time when the activity transitioned
                # from BLOCKED (previously UNAPPROVED) to INPROGRESS.
                insert_event(
                    store, activity_id, approval_time, INPROGRESS, old_status=BLOCKED
                )
                has_entries = True

            if delivery_time and delivery_time != approval_time:
                # A computer activity is INPROGRESS when it gets delivered.  If
                # there is an approval time, the activity was BLOCKED
                # (UNAPPROVED) before.
                insert_event(
                    store,
                    activity_id,
                    delivery_time,
                    INPROGRESS,
                    old_status=BLOCKED if approval_time else None,
                )
                has_entries = True

        if modification_time and not has_entries:
            # The modification time is meaningful only if there's no other
            # entry for the current status.
            insert_event(store, activity_id, modification_time, status)


def insert_event(store, activity_id, timestamp, new_status, old_status=None):
    """Insert an entry in activity_child_event with specified details."""
    data = str(json.dumps({"old": old_status, "new": new_status}))
    store.execute(
        "EXECUTE insert_event(?, ?, ?, ?)",
        [activity_id, timestamp, EVENT_STATUS_TRANSITION, data],
    )
