# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Create triggers to keep package tsvector columns up-to-date."""
from __future__ import print_function


def apply(store):
    print(__doc__)
    for statement in STATEMENTS:
        store.execute(statement)


STATEMENTS = [
    """
GRANT SELECT, INSERT, UPDATE
    ON TABLE package TO landscape
""",
    """
CREATE FUNCTION text_as_grams(i_txt TEXT) RETURNS TEXT AS $$

import re

WORD_SPLIT_PATTERN = re.compile(r"\s+", re.UNICODE)

def get_word_grams(text, unique=True):
    '''
    This function operates on a single word, returning trigrams for that
    word.

    Text grams are non-semantic, non-phonological word components determined
    solely by the positions of letters in a word. The most common form used is
    word "trigrams", where a word of three letters or more can generate a word
    part for every sequential block of three letters.

    This is intended to be used with PostgreSQL's full text search in order to
    provide substring search support. As such, keep in mind that PostgreSQL
    will process all n-grams with its to_tsvector function. This could include
    changes from stemming or exclusion due to matching a stop word.

    @param unique: a boolean that is used to determine if the calculated
        trigrams should include dupicates or not. Defaults to removing the
        duplicates.
    '''
    gram_length = 3
    text_length = len(text)
    if 0 < text_length <= gram_length:
        return [text]

    grams = []
    offset = gram_length - 1
    for index in xrange(text_length - offset):
        gram = text[index:index+gram_length]
        if unique and gram in grams:
            continue
        grams.append(gram)
    return grams

def get_text_grams_as_text(text, unique=True, separator=u" ",
                           include_original=False):
    '''Return text grams as text, instead of as a list.

    This function operates on multiple words and can be used against multi-word
    text in, returning all the trigrams for the words present.

    For information on the parameters, see the docstring for C{get_word_grams}.
    '''
    # NOTE: Remember to update the C{text_as_grams} database function
    # in the package schema whenever you change this function.
    grams = []
    words = WORD_SPLIT_PATTERN.split(text)
    for word in words:
        grams.extend(get_word_grams(word, unique=unique))
    if include_original:
        grams.append(text)
    return separator.join(grams)

return get_text_grams_as_text(i_txt)

$$ LANGUAGE plpythonu
""",
    """
GRANT EXECUTE
    ON FUNCTION text_as_grams(i_txt TEXT) TO landscape
""",
    """
CREATE FUNCTION update_package_tsvector()
  RETURNS trigger AS $PROC$
BEGIN
    IF TG_OP = 'INSERT' THEN
        IF (NEW.name_search_vector IS NULL) THEN
            NEW.name_search_vector :=
                SETWEIGHT(TO_TSVECTOR('simple',
                                      COALESCE(text_as_grams(NEW.name), '')),
                          'A');
        END IF;
        IF (NEW.text_search_vector IS NULL) THEN
            NEW.text_search_vector :=
                SETWEIGHT(TO_TSVECTOR('public.landscape_english',
                                      COALESCE(NEW.summary, '')), 'B') ||
                SETWEIGHT(TO_TSVECTOR('public.landscape_english',
                                      COALESCE(NEW.description, '')), 'C');
        END IF;
    END IF;

    IF TG_OP = 'UPDATE' THEN
        IF (NEW.name_search_vector = OLD.name_search_vector AND
            NOT NEW.name = OLD.name) THEN
            NEW.name_search_vector :=
                SETWEIGHT(TO_TSVECTOR(
                    'simple', COALESCE(text_as_grams(NEW.name), '')), 'A');
        END IF;
        IF (NEW.text_search_vector = OLD.text_search_vector AND
            (NOT (NEW.summary = OLD.summary AND
                  NEW.description = OLD.description))) THEN
            NEW.text_search_vector :=
                SETWEIGHT(TO_TSVECTOR('public.landscape_english',
                                      COALESCE(NEW.summary, '')), 'B') ||
                SETWEIGHT(TO_TSVECTOR('public.landscape_english',
                                      COALESCE(NEW.description, '')), 'C');
        END IF;
    END IF;
    RETURN NEW;
END
$PROC$ LANGUAGE plpgsql;
""",
    """
GRANT EXECUTE
    ON FUNCTION update_package_tsvector() TO landscape
""",
    """
CREATE TRIGGER package_inserted
BEFORE INSERT OR UPDATE ON package
    FOR EACH ROW EXECUTE PROCEDURE update_package_tsvector();
""",
]
