from __future__ import print_function

# Copyright 2005-2016 Canonical Ltd. All rights reserved.

from . import __doc__

STATEMENTS = [
    # Create the new table and copy existing parent activities.
    #
    # Note that the table already existed but was unused. It's easier to drop and
    # re-create it than to change the schema.
    """
DROP TABLE activity_parent
""",
    """
CREATE TABLE activity_parent (
  id INTEGER PRIMARY KEY DEFAULT NEXTVAL('activity_id_seq'::regclass),
  info_type INTEGER NOT NULL,
  creation_time TIMESTAMP WITHOUT TIME ZONE DEFAULT NOW() NOT NULL,
  account_id INTEGER NOT NULL,
  creator_id INTEGER,
  progress integer DEFAULT 0,
  inprogress_count integer DEFAULT 0 NOT NULL,
  queued_count integer DEFAULT 0 NOT NULL,
  waiting_count integer DEFAULT 0 NOT NULL,
  blocked_count integer DEFAULT 0 NOT NULL,
  scheduled_count integer DEFAULT 0 NOT NULL,
  succeeded_count integer DEFAULT 0 NOT NULL,
  failed_count integer DEFAULT 0 NOT NULL,
  canceled_count integer DEFAULT 0 NOT NULL,
  schedule_after_window integer DEFAULT 0,
  CONSTRAINT percentage_progress CHECK ((progress >= 0) AND (progress <= 100))
)
""",
    """
GRANT SELECT, INSERT, DELETE, UPDATE ON TABLE activity_parent TO landscape
""",
    """
CREATE INDEX activity_parent_account_id_creation_time_blocked_idx
ON activity_parent (account_id, creation_time)
WHERE blocked_count > 0
""",
    """
CREATE INDEX activity_parent_account_id_creation_time_pending_idx
ON activity_parent (account_id, creation_time)
WHERE (scheduled_count > 0) OR (waiting_count > 0)
   OR (queued_count > 0) OR (inprogress_count > 0) OR (blocked_count > 0)
""",
    """
CREATE INDEX activity_parent_account_id_idx ON activity_parent (account_id)
""",
    """
CREATE INDEX activity_parent_creation_time_idx
ON activity_parent (creation_time)
""",
    """
CREATE INDEX activity_parent_creator_id_idx
ON activity_parent USING btree(creator_id)
""",
    """
INSERT INTO activity_parent (
    id, info_type, creation_time, account_id, creator_id, progress,
    inprogress_count, queued_count, waiting_count, blocked_count,
    scheduled_count, succeeded_count, failed_count, canceled_count,
    schedule_after_window)
  (SELECT id, info_type, creation_time, account_id, creator_id, progress,
    inprogress_count, queued_count, waiting_count, blocked_count,
    scheduled_count, succeeded_count, failed_count, canceled_count,
    schedule_after_window
   FROM activity
   WHERE parent_id IS NULL)
""",
    # Update the parent reference first otherwise deleting old parent entries
    # would cause a cascade and delete all children too.
    """
ALTER TABLE activity
DROP CONSTRAINT activity_parent_id_fkey,
ADD CONSTRAINT activity_parent_id_fkey FOREIGN KEY (parent_id)
  REFERENCES activity_parent(id)
  ON DELETE CASCADE
""",
    # Remove the primary key constraint before dropping parent activities from the
    # old table, otherwise entries here get cascaded too.
    """
ALTER TABLE activity_parent_access
DROP CONSTRAINT activity_parent_access_parent_id_fkey
""",
    # Delete parent activities from the children table.
    """
DELETE FROM activity WHERE parent_id IS NULL
""",
    # Change the children table to only support children.
    """
ALTER TABLE activity
ALTER COLUMN parent_id SET NOT NULL,
DROP COLUMN inprogress_count,
DROP COLUMN queued_count,
DROP COLUMN waiting_count,
DROP COLUMN blocked_count,
DROP COLUMN scheduled_count,
DROP COLUMN succeeded_count,
DROP COLUMN failed_count,
DROP COLUMN canceled_count,
DROP COLUMN schedule_after_window
""",
    # Recreate the index without the child filter.
    """
DROP INDEX activity_account_id_creation_time_idx
""",
    """
CREATE INDEX activity_account_id_creation_time_idx
ON activity (account_id, creation_time)
""",
    # Create the index on the parent table as well
    """
CREATE INDEX activity_parent_account_id_creation_time_idx
ON activity_parent (account_id, creation_time)
""",
    # Re-add the primary key constraint referencing the parent table.
    """
ALTER TABLE activity_parent_access
ADD CONSTRAINT activity_parent_access_parent_id_fkey
  FOREIGN KEY (parent_id) REFERENCES activity_parent(id)
  ON DELETE CASCADE
""",
]

# Tables that have a foreign key to the activity id for parent activities.
# These now need to reference the activity_parent.id column.
REFERENCING_TABLES = (
    "execute_script_attachment",
    "activity_group",
    "region_request_data",
    "openstack_region_request",
    "node_maintenance_request",
)


def apply(store):
    print(__doc__)
    for table in REFERENCING_TABLES:
        store.execute(
            """
            ALTER TABLE {table}
              DROP CONSTRAINT {table}_activity_id_fkey
            """.format(
                table=table
            )
        )

    for statement in STATEMENTS:
        store.execute(statement)

    for table in REFERENCING_TABLES:
        store.execute(
            """
            ALTER TABLE {table}
              ADD CONSTRAINT {table}_activity_id_fkey
                FOREIGN KEY (activity_id) REFERENCES activity_parent(id)
                ON DELETE CASCADE
            """.format(
                table=table
            )
        )

        # activity_parent_access
