from __future__ import print_function

# Copyright 2005-2020 Canonical Ltd. All rights reserved.

from . import __doc__

STATEMENTS = [
    """
    CREATE OR REPLACE PROCEDURAL LANGUAGE plpython3u
    """,
    r"""
CREATE OR REPLACE FUNCTION public.text_as_grams(i_txt text) RETURNS text
    LANGUAGE plpython3u
    AS $$

import re

WORD_SPLIT_PATTERN = re.compile(r"\s+", re.UNICODE)

def get_word_grams(text, unique=True):
    '''
    This function operates on a single word, returning trigrams for that
    word.

    Text grams are non-semantic, non-phonological word components determined
    solely by the positions of letters in a word. The most common form used is
    word "trigrams", where a word of three letters or more can generate a word
    part for every sequential block of three letters.

    This is intended to be used with PostgreSQL's full text search in order to
    provide substring search support. As such, keep in mind that PostgreSQL
    will process all n-grams with its to_tsvector function. This could include
    changes from stemming or exclusion due to matching a stop word.

    @param unique: a boolean that is used to determine if the calculated
        trigrams should include dupicates or not. Defaults to removing the
        duplicates.
    '''
    gram_length = 3
    text_length = len(text)
    if 0 < text_length <= gram_length:
        return [text]

    grams = []
    offset = gram_length - 1
    for index in range(text_length - offset):
        gram = text[index:index+gram_length]
        if unique and gram in grams:
            continue
        grams.append(gram)
    return grams

def get_text_grams(text, unique=True):
    '''
    This function operates on multiple words and can be used against multi-word
    text in, returning all the trigrams for the words present.

    For information on the parameters, see the docstring for C{get_word_grams}.
    '''
    grams = []
    words = WORD_SPLIT_PATTERN.split(text)
    for word in words:
        grams.extend(get_word_grams(word, unique=unique))
    return grams


def get_text_grams_as_text(text, unique=True, separator=u" ",
                           include_original=False):
    '''Return text grams as text, instead of as a list.
    '''
    grams = get_text_grams(text, unique=unique)
    if include_original:
        grams.append(text)
    return separator.join(grams)

return get_text_grams_as_text(i_txt)

$$;
    """,
    """DROP PROCEDURAL LANGUAGE IF EXISTS plpythonu""",
]


def apply(store):
    print(__doc__)
    for statement in STATEMENTS:
        store.execute(statement)
