from __future__ import print_function

# Copyright 2005-2020 Canonical Ltd. All rights reserved.

from . import __doc__

STATEMENTS = [
    """
CREATE OR REPLACE PROCEDURAL LANGUAGE plpython3u
""",
    """
CREATE OR REPLACE FUNCTION public.get_load_average() RETURNS double precision
    LANGUAGE plpython3u
    AS $$
    import os
    return os.getloadavg()[0]
$$;
""",
    """
CREATE OR REPLACE FUNCTION public.update_computer_packages(
  computer_id integer, available integer[], available_upgrades integer[],
  installed integer[], held integer[], autoremovable integer[],
  security integer[], not_available integer[],
  not_available_upgrades integer[], not_installed integer[],
  not_held integer[], not_autoremovable integer[], not_security integer[],
  reset boolean) RETURNS boolean
    LANGUAGE plpython3u
    AS $_$
  def update_set(s, add, remove):
    '''Adds/remove elements to a set and return a list'''
    s.difference_update(remove)
    s.update(add)
    return list(s)

  def validate_deltas(s, add, remove):
    '''
    Check that all elements in the add sequence are not in s, and all in the
    remove one are.
    '''
    return not (set(remove).difference(s) or s.intersection(add))

  fields_to_deltas = (
    ("available", available, not_available),
    ("available_upgrades", available_upgrades, not_available_upgrades),
    ("installed", installed, not_installed),
    ("held", held, not_held),
    ("autoremovable", autoremovable, not_autoremovable),
    ("security", security, not_security))

  if reset:
    packages = {
      "available": set(),
      "available_upgrades": set(),
      "installed": set(),
      "held": set(),
      "autoremovable": set(),
      "security": set()}
  else:
    # get current packages
    packages_plan = plpy.prepare('''
      SELECT available, available_upgrades, installed, held, autoremovable,
        security
      FROM computer_packages
      WHERE computer_id = $1''', ["int"])
    packages = plpy.execute(packages_plan, [computer_id])[0]
    packages = dict((key, set(value)) for key, value in packages.items())

  # Validate packages to add/remove, checking for package state that all new
  # packages are not already present, and all removed packages are currently
  # present.
  for field, add, remove in fields_to_deltas:
     if not validate_deltas(packages[field], add, remove):
       plpy.warning("Invalid deltas for '%s'" % field)
       return False

  params = [
    update_set(packages[field], add, remove)
    for field, add, remove in fields_to_deltas]
  params.append(computer_id)

  # Cross-state validation, after package sets have been updated with deltas.
  # This ensures that the overall state for the computer is consistent.
  if packages["held"] - packages["installed"]:
    plpy.warning("Packages in held must be installed")
    return False
  if packages["available_upgrades"] - packages["available"]:
    plpy.warning("Packages in available_upgrades must be available")
    return False
  if packages["available_upgrades"] & packages["installed"]:
    plpy.warning("Packages in available_upgrades must not be installed")
    return False
  if packages["autoremovable"] - packages["installed"]:
    plpy.warning("Packages in autoremovable must be installed")
    return False

  update_packages_plan = plpy.prepare('''
      UPDATE computer_packages
      SET available = $1, available_upgrades = $2, installed = $3, held = $4,
          autoremovable = $5, security = $6
      WHERE computer_id = $7''',
     ["int[]", "int[]", "int[]", "int[]", "int[]", "int[]", "int"])
  plpy.execute(update_packages_plan, params)

  return True
$_$;
""",
    """
DROP PROCEDURAL LANGUAGE IF EXISTS plpythonu
""",
]


def apply(store):
    print(__doc__)
    for statement in STATEMENTS:
        store.execute(statement)
