# Copyright 2005-2023 Canonical Ltd. All rights reserved.

from . import __doc__

STATEMENTS = [
    """
CREATE OR REPLACE FUNCTION public.apply_set_delta(
        base integer[], add integer[], remove integer[], reset boolean)
    RETURNS integer[] LANGUAGE plpgsql AS $$
DECLARE
    _base integer[];
    result integer[];
BEGIN
    IF NOT reset THEN
        _base := base;
    ELSE
        _base = ARRAY[]::integer[];
    END IF;

    -- Check that the delta is consistent with the base.
    IF _base && add OR NOT _base @> remove THEN
        RETURN NULL;
    END IF;

    SELECT INTO STRICT result
        COALESCE(array_agg(e ORDER BY e), ARRAY[]::integer[])
    FROM (
        SELECT unnest(_base) e
        UNION SELECT unnest(add)
        EXCEPT SELECT unnest(remove)
    ) r;
    RETURN result;
END;
$$;

CREATE OR REPLACE FUNCTION public.update_computer_packages(
  computer_id integer, available integer[], available_upgrades integer[],
  installed integer[], held integer[], autoremovable integer[],
  security integer[], not_available integer[],
  not_available_upgrades integer[], not_installed integer[],
  not_held integer[], not_autoremovable integer[], not_security integer[],
  reset boolean) RETURNS boolean
    LANGUAGE plpgsql
    AS $$
DECLARE
  old computer_packages%ROWTYPE;
  new_available integer[];
  new_available_upgrades integer[];
  new_installed integer[];
  new_held integer[];
  new_autoremovable integer[];
  new_security integer[];
BEGIN
    SELECT * INTO STRICT old FROM computer_packages
    WHERE computer_packages.computer_id = update_computer_packages.computer_id;

    new_available := apply_set_delta(
        old.available, available, not_available, reset);
    new_available_upgrades := apply_set_delta(
        old.available_upgrades, available_upgrades, not_available_upgrades,
        reset);
    new_installed := apply_set_delta(
        old.installed, installed, not_installed, reset);
    new_held := apply_set_delta(old.held, held, not_held, reset);
    new_autoremovable := apply_set_delta(
        old.autoremovable, autoremovable, not_autoremovable, reset);
    new_security := apply_set_delta(
        old.security, security, not_security, reset);

    -- If any of the new values are NULL, a delta was invalid. Return an error.
    IF new_available IS NULL OR new_available_upgrades IS NULL
        OR new_installed IS NULL OR new_held IS NULL
        OR new_autoremovable IS NULL OR new_security IS NULL THEN
        RETURN false;
    END IF;

    -- Verify the invariant relations between the new sets.
    IF NOT new_held <@ new_installed
        OR NOT new_available_upgrades <@ new_available
        OR new_available_upgrades && new_installed
        OR NOT new_autoremovable <@ new_installed THEN
        RETURN false;
    END IF;

    UPDATE computer_packages
    SET
        available = new_available,
        available_upgrades = new_available_upgrades,
        installed = new_installed,
        held = new_held,
        autoremovable = new_autoremovable,
        security = new_security
    WHERE computer_packages.computer_id = update_computer_packages.computer_id;
    RETURN true;
END;
$$;
"""
]


def apply(store):
    print(__doc__)
    for statement in STATEMENTS:
        store.execute(statement)
