# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Fill the new full text search column for existing packages.
"""
from __future__ import print_function
from __future__ import division

from past.utils import old_div
import time
from datetime import datetime, timedelta
from storm.expr import Column, NamedFunc
from storm.locals import Or, And, Update

from canonical.landscape.model.package import Package
from canonical.landscape.model.textsearch import (
    Concat,
    ToTSVector,
    SetWeight,
    SIMPLE_SEARCH_CONFIG,
    TEXT_SEARCH_CONFIG,
)


BATCH_COUNT = 1000


class TextAsGrams(NamedFunc):
    __slots__ = ()
    name = "text_as_grams"


def apply(store):
    print(__doc__)

    name_search_vector_column = Column("name_search_vector")
    text_search_vector_column = Column("text_search_vector")

    result = store.find(
        Package.id,
        Or(name_search_vector_column == None, text_search_vector_column == None),
    )
    total = result.count()
    print("Found %d packages that need to have their columns filled." % total)

    name_search_vector = SetWeight(
        ToTSVector(SIMPLE_SEARCH_CONFIG, TextAsGrams(Package.name)), "A"
    )

    text_search_vector = Concat(
        SetWeight(ToTSVector(TEXT_SEARCH_CONFIG, Package.summary), "B"),
        SetWeight(ToTSVector(TEXT_SEARCH_CONFIG, Package.description), "C"),
    )

    # Update name_search_vector and text_search_vector column in
    # batches, by performing an UPDATE database query for the whole
    # batch. We commit after each batch in order to keep the
    # transaction short.
    while True:
        start_time = time.time()
        result = store.find(
            Package.id,
            Or(name_search_vector_column == None, text_search_vector_column == None),
        )
        pending = result.count()
        if not pending:
            # Everything is up-to-date? Bail out!
            break

        # Limit this operation to a reasonable batch size.
        result.order_by(Package.id)
        start = result.first()

        if pending <= BATCH_COUNT:
            end = result.last()
        else:
            result.config(offset=BATCH_COUNT)
            end = result.first()

        print("Filling tsvector for packages %d to %d." % (start, end), end=" ")
        where = And(
            Or(name_search_vector_column == None, text_search_vector_column == None),
            Package.id >= start,
            Package.id <= end,
        )
        store.execute(
            Update(
                {
                    text_search_vector_column: text_search_vector,
                    name_search_vector_column: name_search_vector,
                },
                where,
                table=Package,
            )
        )
        store.commit()

        duration = time.time() - start_time
        remaining = (old_div(pending, BATCH_COUNT)) * duration
        completion = datetime.utcnow() + timedelta(seconds=remaining)
        print("[%.2fs r:%.2fs c:%s]" % (duration, remaining, completion.isoformat()))
