# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Add cache to the activity_group table."""
from __future__ import print_function

import time
import transaction

from canonical.landscape.model.activity.group import ActivityGroup


def apply(store):
    print(__doc__)
    store.execute("ALTER TABLE activity_group RENAME COLUMN summary to basic_summary")
    store.execute("ALTER TABLE activity_group ADD COLUMN summary TEXT")
    store.execute("ALTER TABLE activity_group ADD COLUMN children_info_type INTEGER")
    store.execute("ALTER TABLE activity_group RENAME COLUMN data to children_statuses")

    last_time = time.time()
    result = store.find(ActivityGroup)
    info_count = result.count()
    i = 0
    for i, info in enumerate(result.find()):
        info.summary = info.basic_summary.summary
        info.recompute_cache()
        if i and not i % 50:
            transaction.commit()
            last_time = print_progress(i, info_count, last_time)

    transaction.commit()
    if i:
        print_progress(i, info_count, last_time)

    store.execute("ALTER TABLE activity_group ALTER COLUMN summary SET NOT NULL")

    store.execute(
        """
        DROP FUNCTION
        synchronize_activities_for_computers(computer_ids INTEGER[])
        """
    )

    store.execute(
        """
CREATE FUNCTION synchronize_activities_for_computers(computer_ids INTEGER[])
    RETURNS VOID AS
$$

def splice(sequence, splice_size):
    for i in xrange(0, len(sequence), splice_size):
        yield sequence[i:i+splice_size]

select_group_activities_plan = plpy.prepare('''
    SELECT activity.id, activity_group.children_statuses FROM activity,
    activity_group WHERE activity.id = activity_group.activity_id AND
    activity.info_type = 6 AND activity_group.children_statuses != '{}' AND
    EXISTS (SELECT 1 FROM activity AS child_activity
    WHERE child_activity.root_parent_id = activity.id AND
    child_activity.computer_id = ANY($1))
''', ["int4[]"])

delete_child_activities_plan = plpy.prepare('''
    DELETE FROM activity WHERE computer_id = ANY($1) AND (root_parent_id = $2
    OR parent_id = $2)
''', ["int4[]", "int4"])

delete_activity_plan = plpy.prepare('''
    DELETE FROM activity WHERE id = $1
''', ["int4"])

update_activity_plan = plpy.prepare('''
    UPDATE activity_group SET children_statuses = $2 WHERE activity_id = $1
''', ["int4", "int4[]"])

select_remaining_activities = plpy.prepare('''
    SELECT id, root_parent_id FROM activity WHERE computer_id = ANY($1)
    AND root_parent_id is not NULL
''', ["int4[]"])

delete_empty_activities_notice = plpy.prepare('''
    DELETE FROM activity WHERE activity.info_type = 6 AND activity.id = ANY($1)
    AND NOT EXISTS (SELECT 1 FROM activity AS child_activity WHERE
    child_activity.root_parent_id = activity.id)
''', ["int4[]"])

# First, delete computer request activities
computer_ids_set = set(map(int, computer_ids[1:-1].split(",")))
for activity in plpy.execute(select_group_activities_plan, [computer_ids]):
    activity_id = activity["id"]
    children_statuses = activity["children_statuses"]
    real_children_statuses = [
        (int(i[1:]), int(j[:-1]))
        for i, j in splice(children_statuses[1:-1].split(","), 2)]
    for computer_id, status in real_children_statuses[:]:
        if computer_id in computer_ids_set:
            real_children_statuses.remove((computer_id, status))

    plpy.execute(delete_child_activities_plan, [computer_ids, activity_id])

    if not real_children_statuses:
        plpy.execute(delete_activity_plan, [activity_id])
    else:
        new_children_statuses = ",".join(
            "{%d,%d}" % (computer_id, status_id)
            for computer_id, status_id in real_children_statuses)
        new_children_statuses = "{%s}" % new_children_statuses
        plpy.execute(
            update_activity_plan, [activity_id, new_children_statuses])

# Let's take care of notice activities
parent_ids = set()
for activity in plpy.execute(select_remaining_activities, [computer_ids]):
    activity_id = activity["id"]
    parent_id = activity["root_parent_id"]
    parent_ids.add(parent_id)
    plpy.execute(delete_activity_plan, [activity_id])

parent_ids = "{%s}" % (",".join(map(str, parent_ids)),)
plpy.execute(delete_empty_activities_notice, [parent_ids])

$$ LANGUAGE plpythonu
"""
    )


def print_progress(current, total, last_time):
    now = time.time()
    print("%s of %s took %.02fs" % (current, total, now - last_time))
    return now
