# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""RESOURCE DATABASE: OPERATIONS / PACKAGE MANAGEMENT MAJOR UPGRADE"""
from __future__ import print_function

from storm.expr import Select, In, Column

from canonical.landscape.model.main import get_main_store


creations = [
    """
CREATE TABLE operation (
    id SERIAL PRIMARY KEY,
    type INTEGER NOT NULL,
    command_id INTEGER,
    parent_id INTEGER REFERENCES operation ON DELETE CASCADE,
    root_parent_id INTEGER REFERENCES operation ON DELETE CASCADE,
    computer_id INTEGER NOT NULL REFERENCES computer ON DELETE CASCADE,
    child_count INTEGER NOT NULL DEFAULT 0,
    child_index INTEGER NOT NULL DEFAULT 0,
    status INTEGER NOT NULL,
    creation_time TIMESTAMP NOT NULL,
    approval_time TIMESTAMP,
    delivery_time TIMESTAMP,
    completion_time TIMESTAMP,
    modification_time TIMESTAMP NOT NULL,
    order_time TIMESTAMP,
    result_code INTEGER,
    result_text TEXT)
""",
    """
CREATE TABLE generic_user_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    username TEXT NOT NULL)
""",
    """
CREATE TABLE delete_user_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    username TEXT NOT NULL,
    delete_home BOOLEAN NOT NULL)
""",
    """
CREATE TABLE edit_user_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    username TEXT NOT NULL,
    name TEXT,
    passphrase TEXT,
    location TEXT,
    work_phone TEXT,
    home_phone TEXT)
""",
    """
CREATE TABLE add_user_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    username TEXT NOT NULL,
    name TEXT NOT NULL,
    passphrase TEXT NOT NULL,
    require_password_reset BOOLEAN NOT NULL)
""",
    """
CREATE TABLE add_group_member_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    username TEXT NOT NULL,
    groupname TEXT NOT NULL)
""",
    """
CREATE TABLE remove_group_member_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    username TEXT NOT NULL,
    groupname TEXT NOT NULL)
""",
    """
CREATE TABLE create_group_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    groupname TEXT NOT NULL)
""",
    """
CREATE TABLE signal_process_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    signal TEXT NOT NULL,
    pid INTEGER NOT NULL,
    name TEXT NOT NULL)
""",
    """
CREATE TABLE change_package_opinfo (
    operation_id INTEGER PRIMARY KEY REFERENCES operation ON DELETE CASCADE,
    package_id INTEGER NOT NULL)
""",
    """
CREATE LANGUAGE plpythonu
""",
    """
CREATE LANGUAGE plpgsql
""",
    """
CREATE FUNCTION change_array(source_array INT[],
                             remove_array INT[],
                             append_array INT[]) RETURNS INT[] AS
$$
    if args[0] != "{}":
        result = set(args[0][1:-1].split(','))
    else:
        result = set()
    if args[1] != "{}":
        result.difference_update(args[1][1:-1].split(','))
    if args[2] != "{}":
        result.update(args[2][1:-1].split(','))
    return '{%s}' % ','.join(result)
$$
LANGUAGE plpythonu
""",
    """
CREATE TABLE computer_packages (
    computer_id SERIAL NOT NULL PRIMARY KEY REFERENCES computer
        ON DELETE CASCADE,
    available INT[] NOT NULL DEFAULT '{}',
    installed INT[] NOT NULL DEFAULT '{}')
""",
    """
CREATE FUNCTION update_operation_order_time() RETURNS trigger AS $PROC$
DECLARE
  root_parent integer;
  parent record;
BEGIN
  IF tg_op = 'INSERT' THEN
    new.order_time := new.modification_time;
    IF new.root_parent_id IS NOT NULL THEN
      UPDATE operation SET order_time = new.modification_time
             WHERE (operation.root_parent_id = new.root_parent_id OR
                    operation.id = new.root_parent_id) AND
                   operation.id != new.id;
    END IF;
  ELSE
    IF (old.modification_time != new.modification_time) OR
       (old.root_parent_id IS NULL != new.root_parent_id IS NULL OR
        old.root_parent_id != new.root_parent_id) THEN
      IF new.order_time IS NULL OR new.modification_time > new.order_time THEN
        new.order_time = new.modification_time;
      END IF;
      IF new.root_parent_id IS NULL THEN
        root_parent := new.id;
      ELSE
        root_parent := new.root_parent_id;
      END IF;
      UPDATE operation SET order_time = new.modification_time
             WHERE (operation.root_parent_id = root_parent OR
                    operation.id = root_parent)
                    AND operation.id != new.id -- avoid recursion
                    AND operation.order_time < new.modification_time;
    END IF;
    IF (old.root_parent_id IS NULL != new.root_parent_id IS NULL OR
        old.root_parent_id != new.root_parent_id) THEN
      -- If this op is being placed into a new/different graph, it
      -- needs to have its order_time updated because it may be at an
      -- older order_time but relating to a graph with a newer
      -- order_time.
      SELECT INTO parent order_time FROM operation
             WHERE operation.id = new.parent_id;
      new.order_time = parent.order_time;
    END IF;
  END IF;
  RETURN new;
END
$PROC$ LANGUAGE plpgsql;
""",
    """
CREATE TRIGGER operation_modification_time_changed
BEFORE UPDATE OR INSERT ON operation
  FOR EACH ROW EXECUTE PROCEDURE update_operation_order_time();
""",
]


def apply(store):
    print("start of patch 10: %s" % __doc__)
    for creation in creations:
        store.execute(creation)
    print("associating ComputerTwin with Account ids")

    main_store = get_main_store()
    store.execute("ALTER TABLE computer ADD COLUMN account_id INTEGER")
    rcomputer_ids = [x[0] for x in store.execute("SELECT id from computer").get_all()]
    if rcomputer_ids:
        # Use Storm syntax for its In() support :-)
        query = main_store.execute(
            Select(
                (Column("id"), Column("account_id")),
                In(Column("id"), rcomputer_ids),
                "computer",
            )
        )

        for computer_id, account_id in query:
            store.execute(
                "UPDATE computer SET account_id = ? WHERE id = ?",
                [account_id, computer_id],
            )

        print("successfully associated all ComputerTwins")
    store.execute("ALTER TABLE computer ALTER COLUMN account_id SET NOT NULL")
