"""Utility functions for database schema inspection and manipulation."""

# The id of the postgres advisory lock we use to lock concurrent processes
# trying to perform schema operations.
from builtins import str

SCHEMA_ADVISORY_LOCK = 1000


def get_postgres_version(store):
    """Return the PostgreSQL server version, as integer (e.g. 90306)."""
    result = store.execute("SHOW server_version_num")
    return int(result.get_one()[0])


def entity_exists(store, name, kind="r"):
    """Return whether the specified name exists in postgres pg_class table.

    @param name: name of the table-like resource
    @param kind: a letter indicating the class type, as follows (from
       Postgres documentation):

         r = ordinary table
         i = index
         S = sequence
         v = view
         c = composite type
         s = special
         t = TOAST table
    """
    result = store.execute(
        "SELECT 1 FROM pg_catalog.pg_class WHERE relname = ? AND relkind = ?",
        (str(name), str(kind)),
    )
    return result.get_one() is not None


def user_exists(store, name):
    """Return whether a user with the specified name exists in PostgreSQL.

    @param name: The name of the user to check.
    """
    result = store.execute(
        "SELECT 1 FROM pg_catalog.pg_user WHERE usename = ?", (str(name),)
    )
    return result.get_one() is not None


def create_user(store, name, password):
    """Create the PostgreSQL user with the given name.

    @param name: The name of the user to create.
    @param password: The password to use.
    """
    store.execute("CREATE USER \"%s\"WITH PASSWORD '%s'" % (str(name), str(password)))


def set_password(store, name, password):
    """Set the password for the given PostgreSQL user.

    @param name: The name of the user to set the password for.
    @param password: The password to use.
    """
    store.execute("ALTER USER \"%s\" WITH PASSWORD '%s'" % (str(name), str(password)))


def database_exists(store, name):
    """Return whether a database with the specified name exists in PostgreSQL.

    @param name: The name of the database to check.
    """
    result = store.execute(
        "SELECT 1 FROM pg_catalog.pg_database WHERE datname = ?", (str(name),)
    )
    return result.get_one() is not None


def create_database(store, name):
    """Create a new database with the given name.

    @param name: The name of the database to check.
    """
    store.execute('CREATE DATABASE "%s" OWNER postgres' % (str(name),))


def acquire_lock(store):
    """Acquire a lock on the store.

    Concurrent connections racing for performing operations on a store can
    use this function in order to acquire an exclusive lock and hence force
    their serialization (the connection losing the race will have to wait
    till the winner releases the lock or closes the connection).

    Timeouts are applied using the regular process-level tracer in the schema
    script (see the C{set_allowed_request_time} call in the schema script).
    """
    store.execute("SELECT pg_advisory_lock(%d)" % SCHEMA_ADVISORY_LOCK)


def release_lock(store):
    """Release a lock previously acquired with C{acquire_lock}."""
    store.execute("SELECT pg_advisory_unlock(%d)" % SCHEMA_ADVISORY_LOCK)


def get_table_oid(store, table):
    """Return the table OID based from the name."""
    query = "SELECT oid FROM pg_catalog.pg_class WHERE relname = ?"
    result = store.execute(query, [str(table)])
    return result.get_one()[0]


def find_table_constraints(store, table):
    """Return a list of constraints for the table.

    The returned list contains tuples with constraint name and definition.

    Primary keys and unique constraints are excluded here since they are
    reported as indexes.
    """
    table_oid = get_table_oid(store, table)
    query = """
        SELECT conname, pg_catalog.pg_get_constraintdef(oid, true) AS condef
        FROM pg_catalog.pg_constraint
        WHERE conrelid = ? AND contype NOT IN ('p', 'u')
    """
    return store.execute(query, [table_oid]).get_all()


def find_table_indexes(store, table):
    """Return a list of indexes for the table."""
    query = """
        SELECT
            pg_class.relname AS name
        FROM pg_class
            LEFT JOIN pg_namespace
                ON pg_namespace.oid = pg_class.relnamespace
            LEFT JOIN pg_index
                ON pg_index.indexrelid = pg_class.oid
            LEFT JOIN pg_class pg_class_table
                ON pg_index.indrelid = pg_class_table.oid
        WHERE pg_class.relkind = 'i'
            AND pg_namespace.nspname <> 'pg_catalog'
            AND pg_namespace.nspname <> 'information_schema'
            AND pg_namespace.nspname !~ '^pg_toast'
            AND pg_table_is_visible(pg_class.oid)
            AND pg_class_table.relname = ?
    """
    result = store.execute(query, [table]).get_all()
    return [row[0] for row in result]


def rename_table(store, from_table, to_table):
    """
    Rename table along with its constraints, such as foreign keys, check
    constraints, indexes.
    """
    # Rename constraints
    constraints = find_table_constraints(store, from_table)
    for constraint_name, constraint_def in constraints:
        new_name = constraint_name.replace(from_table, to_table, 1)
        store.execute(
            "ALTER TABLE %s DROP CONSTRAINT %s, ADD CONSTRAINT %s %s"
            % (from_table, constraint_name, new_name, constraint_def)
        )

    # Rename indexes
    for index_name in find_table_indexes(store, from_table):
        new_name = index_name.replace(from_table, to_table, 1)
        store.execute("ALTER INDEX %s RENAME TO %s" % (index_name, new_name))

    # Finally rename the table itself
    store.execute("ALTER TABLE %s RENAME TO %s" % (from_table, to_table))


def get_tables(store, exclude=()):
    """
    Query PostgreSQL for the list of tables in the connected database store.

    @param store: Storm database store.
    @param exclude: optionally a list of table names to exclude.
    @return a list of strings of table names that are already in the
        connected database.
    """
    cursor = store.execute(
        """
        SELECT c.relname
        FROM pg_catalog.pg_class c
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
        WHERE c.relkind IN ('r','')
            AND n.nspname NOT IN ('pg_catalog', 'pg_toast')
            AND pg_catalog.pg_table_is_visible(c.oid)
        """
    )
    tables = [x[0] for x in cursor.get_all()]
    if exclude:
        tables = list(set(tables).difference(set(exclude)))
    return tables


def is_table_empty(store, table):
    """Return True if `table` has no row, False otherwise."""
    cursor = store.execute("SELECT * FROM %s LIMIT 1" % table)
    return cursor.get_one() is None


def get_database_name(store):
    """Return the database name for the given PostgreSQL store."""
    query = store.execute("SELECT current_database()")
    result = query.get_one()
    if result:
        return result[0]
    else:
        return None
