# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Patch verification functionality for the Landscape server."""
import sys
import logging
import psycopg2
from traceback import format_exception
import transaction

from zope.component import getUtility

from storm.zope.interfaces import IZStorm
from storm.schema.patch import PatchApplier, PatchSet

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.resource.store import get_resource_stores
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.session.store import get_session_store
from canonical.landscape.model.knowledge.store import get_knowledge_store
from canonical.landscape.schema import main as main_patches
from canonical.landscape.schema import account as account_patches
from canonical.landscape.schema import resource as resource_patches
from canonical.landscape.schema import package as package_patches
from canonical.landscape.schema import session as session_patches
from canonical.landscape.schema import knowledge as knowledge_patches


def check_pending_patches(
    event=None,
    log=logging.error,
    patch_applier=PatchApplier,
    main_store=None,
    resource_stores=None,
    account_stores=None,
    package_store=None,
    session_store=None,
    knowledge_store=None,
    clean_stores=True,
):
    """Shutdown the server if unapplied or unknown patches exist

    Function checks the main database, package database, and all related
    account and resource databases for unapplied or unknown patches.  If a
    database is missing patches or has unknown patches an exception is raised
    and informative error messages are written to the log.
    """
    # Start from fresh, so that the stores don't contain any pending queries
    transaction.abort()
    messages = []
    error_messages = []
    zstorm = getUtility(IZStorm)

    def check_store(name, store, patch_package):
        if clean_stores:
            try:
                # Remove the store before it registers with the transaction
                # manager
                zstorm.remove(store)
            except KeyError:
                pass
        sub_level = name.split("-")[0]
        patch_set = PatchSet(patch_package, sub_level=sub_level)
        applier = patch_applier(store, patch_set)
        versions = applier.get_unknown_patch_versions()
        if versions:
            messages.append("%s has unknown patches." % name)
        if applier.has_pending_patches():
            messages.append("%s has unapplied patches." % name)
        store.rollback()
        if clean_stores:
            store.close()

    try:
        main = main_store or get_main_store()
        check_store("main", main, main_patches)
    except psycopg2.Error:
        # This catches both psycopg2.ProgrammingError and
        # psycopg2.OperationalError
        exc_type, exc_value, exc_tb = sys.exc_info()
        error = "".join(format_exception(exc_type, exc_value, exc_tb))
        error_messages.append(error)

    try:
        for name, store in resource_stores or get_resource_stores():
            check_store(name, store, resource_patches)
    except psycopg2.Error:
        exc_type, exc_value, exc_tb = sys.exc_info()
        error = "".join(format_exception(exc_type, exc_value, exc_tb))
        error_messages.append(error)

    try:
        for name, store in account_stores or get_account_stores():
            check_store(name, store, account_patches)
    except psycopg2.Error:
        exc_type, exc_value, exc_tb = sys.exc_info()
        error = "".join(format_exception(exc_type, exc_value, exc_tb))
        error_messages.append(error)

    try:
        package = package_store or get_package_store()
        check_store("package", package, package_patches)
    except psycopg2.Error:
        exc_type, exc_value, exc_tb = sys.exc_info()
        error = "".join(format_exception(exc_type, exc_value, exc_tb))
        error_messages.append(error)

    if "session" in zstorm.get_default_uris():
        try:
            session = session_store or get_session_store()
            check_store("session", session, session_patches)
        except psycopg2.Error:
            exc_type, exc_value, exc_tb = sys.exc_info()
            error = "".join(format_exception(exc_type, exc_value, exc_tb))
            error_messages.append(error)

    if "knowledge" in zstorm.get_default_uris():
        try:
            knowledge = knowledge_store or get_knowledge_store()
            check_store("knowledge", knowledge, knowledge_patches)
        except psycopg2.Error:
            exc_type, exc_value, exc_tb = sys.exc_info()
            error = "".join(format_exception(exc_type, exc_value, exc_tb))
            error_messages.append(error)

    if error_messages:
        for error_message in error_messages:
            log(error_message)
        log("Shutting down the server!")
        raise SystemExit(2)

    if messages:
        for message in messages:
            log(message)
        log("Shutting down the server!")
        raise RuntimeError("Databases need schema changes")
