# Copyright 2005-2022 Canonical Limited. All rights reserved.

import argparse
import sys

import transaction

from canonical.landscape.model.main.account import (
    get_account_by_sfid,
    get_account_by_name,
    get_account_by_salesforce_account_key,
)

from canonical.landscape.setup import load_config
from canonical.landscape.scripts.utils import manual_confirm
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.resource.store import get_resource_stores


OPS = [
    "remove",
    "reset-admins",
    "disable",
    "enable",
]

ENCODING = getattr(sys.stdin, "encoding", None) or "utf-8"

ID_KINDS = {
    "name": get_account_by_name,
    "sfid": get_account_by_sfid,
    "sfkey": get_account_by_salesforce_account_key,
}


def _parse_id(rawid, defaultkind="name"):
    kind, sep, rawid = rawid.partition(":")
    if not sep or not rawid:
        rawid = kind + sep
        kind = defaultkind
    elif kind not in ID_KINDS:
        rawid = kind + sep + rawid
        kind = defaultkind
    return kind, rawid


def _get_account(key, kind):
    get_account = ID_KINDS[kind]
    return get_account(key)


def _iter_accounts(ops):
    for op, idkind, rawid in ops:
        # The default isolation level that Storm uses for PostgreSQL is
        # repeatable read which is excessive for many operations and causes
        # serialization issues when account removal conflicts with a ping time
        # update.  Read committed is sufficient for all ops in this case.
        # Transaction isolation can only be set at transaction start, and each
        # specified op begins its transaction here (this is the first query),
        # hence the slightly odd positioning.
        sql = "SET TRANSACTION ISOLATION LEVEL READ COMMITTED"
        get_main_store().execute(sql)
        if op == "remove":
            for name, store in get_resource_stores():
                store.execute(sql)
            for name, store in get_account_stores():
                store.execute(sql)

        account = _get_account(rawid, idkind)
        yield op, rawid, account


class ValidationError(RuntimeError):
    """Validation failed."""


def _validate(account, op):
    if not account:
        raise ValidationError("Account %s not found.")
    if op == "remove" and not account.is_disabled:
        raise ValidationError("Account %s must be disabled before removal.")
    if op == "reset-admins" and list(account.people) == []:
        raise ValidationError("Account %s admins are already reset.")
    if op == "disable" and account.is_disabled:
        raise ValidationError("Account %s is already disabled.")
    if op == "enable" and not account.is_disabled:
        raise ValidationError("Account %s is already enabled.")


def _handle_op(op, rawid, account, confirm, do_op):
    if op not in OPS:
        raise RuntimeError("unsupported op %r" % op)
    try:
        _validate(account, op)
    except ValidationError as e:
        return str(e) % rawid

    if confirm and not confirm(account.name, op):
        return "skipped"
    do_op(account, op)
    return "success"


def _do_op(account, op):
    if op == "enable":
        account.enable()
    elif op == "disable":
        account.disable("requested via CLI")
    elif op == "reset-admins":
        account.disable_admins()
    elif op == "remove":
        account.remove()


#######################################
# the script


class Ops(argparse.Action):
    def __init__(self, *args, **kwargs):
        op = kwargs.pop("op", None)
        kind = kwargs.pop("kind", None)

        kwargs.setdefault("nargs", "+")
        if "metavar" not in kwargs:
            if not kind:
                metavar = "ANY_ID"
                # metavar = "NAME|sfid:SF_ID|sfkey:SF_KEY"
            elif kind == "sfid":
                metavar = "SF_ID"
            elif kind == "sfkey":
                metavar = "SF_KEY"
            kwargs["metavar"] = metavar
        super(Ops, self).__init__(*args, **kwargs)

        if not op:
            for opt in self.option_strings:
                maybe = opt.strip("-").split("-")[0]
                if maybe in OPS:
                    op = maybe
                    break
            else:
                raise RuntimeError("missing op")
        if op not in OPS:
            raise RuntimeError("unsupported operation %r" % op)

        if kind is not None and kind not in ID_KINDS:
            raise RuntimeError("unsupported ID kind %r" % kind)

        self.op = op
        self.legacy_kind = kind

    def __call__(self, parser, namespace, values, option_string=None):
        for rawid in values:
            kind, rawid = _parse_id(rawid, self.legacy_kind or "name")
            if self.legacy_kind and kind != self.legacy_kind:
                parser.error(
                    "kind prefix %r did not match %r" % (kind, self.legacy_kind)
                )
            vars(namespace).setdefault(self.dest, []).append((self.op, kind, rawid))


def parse_args(argv=None):
    parser = argparse.ArgumentParser(
        prog=argv[0] if argv else None,
        description="""
                An ID may be an account name, a Salesforce ID, or a
                Salesforce account key.  All IDs except account names
                must have the appropriate prefix.  Use an "sfid:" prefix
                for Salesforce IDs and an "sfkey:" prefix for Salesforce
                account keys.""",
        epilog="Note: Removal of an account with live admins will fail.",
    )

    parser.add_argument(
        "--yes",
        dest="confirm",
        action="store_false",
        default=True,
        help="Don't ask for confirmation of any request",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Don't actually do anything, just confirm the " "account.",
    )

    # ops
    parser.set_defaults(ops=[])
    add_ops_arg = lambda opt, **kwargs: parser.add_argument(
        opt, dest="ops", action=Ops, **kwargs
    )

    add_ops_arg("--remove", help="Remove account(s)")
    add_ops_arg("--reset-admins", help="Reset account(s) admins", op="reset-admins")
    add_ops_arg("--disable", help="Disable account(s)")
    add_ops_arg("--enable", help="Re-enable account(s)")

    # Do not include the legacy options in the usage text.
    parser.usage = parser.format_usage().strip("usage: ")

    # legacy aliases
    legacy = parser.add_argument_group("legacy options")
    add_legacy_arg = lambda opt, **kwargs: legacy.add_argument(
        opt, dest="ops", action=Ops, **kwargs
    )

    add_legacy_arg(
        "--remove-by-user-sfid", kind="sfid", help="Remove account(s) by Salesforce ID"
    )
    add_legacy_arg(
        "--remove-by-account-sfid",
        kind="sfkey",
        help="Remove account(s) by Salesforce key",
    )
    add_legacy_arg(
        "--disable-by-user-sfid",
        kind="sfid",
        help="Disable account(s) by Salesforce ID",
    )
    add_legacy_arg(
        "--disable-by-account-sfid",
        kind="sfkey",
        help="Disable account(s) by Salesforce key",
    )

    # Parse the args.
    args = parser.parse_args(argv[1:] if argv else None)

    if not args.ops:
        parser.error("at least one operation required")

    if args.dry_run:
        args.confirm = False

    return args


def main(ops, confirm=None, do_op=None):
    if do_op is None:
        do_op = _do_op

    def log(msg):
        sys.stdout.write(msg + "\n")

    total_counts = {op: 0 for op in OPS}
    success_counts = {op: 0 for op in OPS}
    failure_counts = {op: 0 for op in OPS}

    # Handle each op.
    for op, rawid, account in _iter_accounts(ops):
        total_counts[op] += 1
        result = _handle_op(op, rawid, account, confirm, do_op)
        if result == "success":
            if op == "reset-admins":
                log("Reset admins on account %s." % (account.name,))
            else:
                log("%sd account %s." % (op.capitalize(), account.name))
            success_counts[op] += 1
        elif result == "skipped":
            if op == "reset-admins":
                log("Admins on account %s not reset." % (account.name,))
            else:
                log("Account %s not %sd." % (account.name, op))
            # not counted as success or failure
        else:  # failure
            log(result)
            failure_counts[op] += 1
        transaction.commit()

    # Summarize the results.
    for op in OPS:
        if not total_counts[op]:
            continue
        if op == "reset-admins":
            log(
                "Reset admins on %d accounts, had %d failures."
                % (success_counts[op], failure_counts[op])
            )
        else:
            log(
                "%sd %d accounts, had %d failures."
                % (op.capitalize(), success_counts[op], failure_counts[op])
            )
    return (sum(success_counts.values()), sum(failure_counts.values()))


def account_tool(argv=None):
    """Handles command line arguments for the landscape-accounts tool.

    @return: A C{(total_successes, total_failures)} 2-tuple.
    """
    if argv:
        argv = [sys.argv[0]] + argv
    args = parse_args(argv)

    confirm = None
    if args.confirm:
        actions = {
            "remove": "removal",
            "reset-admins": "resetting admins",
            "disable": "disabling",
            "enable": "enabling",
        }

        def confirm(name, op):
            action = actions[op]
            prompt = "Please confirm %s of %s, type YES to confirm: "
            return manual_confirm(prompt % (action, name))

    do_op = None if not args.dry_run else (lambda *args: None)

    return main(args.ops, confirm, do_op)


def run():
    print("Loading site configuration...")
    load_config("maintenance")

    successes, failures = account_tool()

    if failures:
        sys.exit(1)
