# Copyright 2005-2022 Canonical Limited. All rights reserved.

from collections import deque
import glob
import logging
import os
import sys
import time

from zope.component import getUtility

from storm.zope.interfaces import IZStorm

from canonical.lib.pidfile import PidFile, PidFileTakenError

from canonical.landscape.setup import load_config, run_amqp_client_in_thread
from canonical.landscape.schema.verify import check_pending_patches
from canonical.landscape.application import setup_logging
from canonical.landscape.model.configuration import is_charm_deployment
from canonical.landscape.model.configuration import set_proxy_configuration

MAINTENANCE_FLAG = "/opt/canonical/landscape/maintenance.txt"
PID_DIR = "/var/run/landscape"


class BatchScript(object):
    """Base class for batch and maintenance scripts.

    @ivar _name: The name of the script, it will also be the name of the
        logger to use.
    @ivar _config: Section of services.conf use as script configuration.
    @ivar _setup_amqp: Whether to setup a connection with the AMQP broker (for
        scripts that interact with the job-system or async-frontend for
        notifications)
    @ivar _log_level: Logging level to set.
    @ivar _log_message_format: Logging message format.
    @ivar _log_date_format: Logging date format.
    """

    _name = property(lambda _: old_div(1, 0))  # Must be set by subclasses
    _config = "maintenance"
    _setup_amqp = False

    _log_level = logging.INFO
    _log_message_format = "%(asctime)s %(levelname)s %(name)s %(message)s"
    _log_date_format = "%Y-%m-%d %H:%M:%S"

    # For testing purposes
    run_amqp_client_in_thread = staticmethod(run_amqp_client_in_thread)

    def __init__(
        self,
        args=sys.argv,
        stdin=sys.stdin,
        stdout=sys.stdout,
        pid_dir=PID_DIR,
        maintenance_flag=MAINTENANCE_FLAG,
        ignore_maintenance_flag=False,
    ):
        """
        @param args: Command line arguments.
        @param stdin: Standard input stream, for testing.
        @param stdout: Standard output stream.
        @param pid_dir: The directory to save the PID file in.
        @param maintenance_flag: Path to the maintenance flag file, if None,
                                 Default is the system default location
                                defined in MAINTENANCE_FLAG.
        @param ignore_maintenance_flag: If True, don't check the maintenance
                                        flag. Default is False.
        """
        self._args = args
        self._stdin = stdin
        self._stdout = stdout
        self._stop_amqp_client = None
        self._pid_dir = pid_dir
        self._maintenance_flag = maintenance_flag
        self._ignore_maintenance_flag = ignore_maintenance_flag
        self._cleanup_callbacks = deque()

    def __call__(self):
        """Invoke the script."""
        try:
            self.setup()
            code = self.run()
        finally:
            self._cleanup()
        if code is None:
            code = 0
        sys.exit(code)

    def _set_up_logging(self):
        setup_logging(
            name=self._name,
            fmt=self._log_message_format,
            dateformat=self._log_date_format,
            level=self._log_level,
            stdout=self._stdout,
        )

    def setup(self):
        """Boiler plate set-up tasks that all scripts need."""

        # Set-up logging
        self._set_up_logging()

        # Honor maintenance flag
        if not self._ignore_maintenance_flag and os.path.exists(self._maintenance_flag):
            logging.info("Maintenance file found, skipping.")
            sys.exit(0)

        # XXX For now we only want to write PID files for charm-deployed
        # scripts.
        if is_charm_deployment():
            # Write the PID file or exit if a valid one exists.
            pid_file = self._get_pid_file()
            try:
                pid_file.write()
            except PidFileTakenError:
                logging.info("Another instance of this script is already running")
                sys.exit(0)
            else:
                self._add_cleanup_callback(pid_file.remove)

        # Load configuration
        config = load_config(self._config)

        # Setup AMQP
        if self._setup_amqp:
            self._add_cleanup_callback(self.run_amqp_client_in_thread(config))

        # Check database patch level
        check_pending_patches(clean_stores=False)

        # Setup HTTP proxy environment variables
        set_proxy_configuration()

        self._add_cleanup_callback(self._clean_stores)

    def run(self):
        """Actual logic of the script, should be implemented by sub-classes.

        @return: The exit code of the script. If None is returned, the script
            will exit with 0.
        """
        raise NotImplementedError()

    def _add_cleanup_callback(self, callback):
        """Add a cleanup function."""
        self._cleanup_callbacks.appendleft(callback)

    def _cleanup(self):
        """Exit sequence."""
        for cleanup_callback in self._cleanup_callbacks:
            cleanup_callback()

    def _clean_stores(self):
        """Clean and remove store storm stores."""
        # Clean stores from scripts main thread to avoid them going down in a
        # less ideal threaded context. This is due to storm's use of __del__,
        # which may or may not get triggered from anywhere, including after
        # SystemExit. (LP: #1764456)
        zstorm = getUtility(IZStorm)
        for name in sorted(zstorm.get_default_uris()):
            store = zstorm.get(name)
            zstorm.remove(store)
            store.close()

    def _get_pid_file(self):
        """Return a PidFile pointing to our PID file."""
        # We prefix the file name with 'batch-' in order to be able to locate
        # PID files of batch scripts in wait_batch_scripts.
        path = os.path.join(self._pid_dir, "batch-%s.pid" % self._name)
        return PidFile(path)


def wait_batch_scripts(timeout=60 * 15, pid_dir=PID_DIR, pause=1):
    """Wait for all running batch scripts to terminate."""
    start = time.time()
    while True:
        paths = glob.glob(os.path.join(pid_dir, "batch-*.pid"))
        if not paths:
            return
        for path in paths:
            pid_file = PidFile(path)
            while True:
                pid = pid_file.check()
                if pid is None:
                    break
                if time.time() - start > timeout:
                    os.kill(pid, 9)
                time.sleep(pause)
