# Copyright 2023 Canonical Limited. All rights reserved.

import argparse
import logging
import sys
import transaction

from canonical.landscape.api.account import bootstrap_account_helper
from canonical.landscape.application import setup_logging
from canonical.landscape.setup import load_config


def get_option_parser():
    """
    Create and populate the options for the OptionParser.
    """
    parser = argparse.ArgumentParser(description="Removes old event log entries.")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--admin_password",
        help="The password of the account administrator, provided if not using openid",
    )
    group.add_argument(
        "--identity",
        help="The identity used for openid login",
    )

    parser.add_argument(
        "--admin_email",
        required=True,
        help="The email address of the account administrator.",
    )
    parser.add_argument(
        "--admin_name",
        required=True,
        help="The full name of the account administrator.",
    )
    parser.add_argument(
        "--system_email",
        help="The email address that will appear on emails sent by Landscape.",
    )
    parser.add_argument("--registration_key", help="The account registration key.")
    parser.add_argument(
        "--root_url",
        required=True,
        help="The URL used to access Landscape's UI. In charmed deployment,"
        "this takes the form https://<apache_unit_ip>/.",
    )

    return parser


def run_bootstrap_account(args):
    """
    When run, this creates the "standalone" account with the specified
    admin. Will fail if the account already exists.
    """
    parser = get_option_parser()
    args = parser.parse_args(args)
    with transaction.manager:
        bootstrap_account_helper(
            admin_email=args.admin_email,
            admin_name=args.admin_name,
            admin_password=args.admin_password,
            system_email=args.system_email,
            registration_key=args.registration_key,
            root_url=args.root_url,
            identity=args.identity,
        )
    logging.info("Bootstrap account successful!")


def run():
    setup_logging("bootstrap-account", level=logging.INFO)
    load_config("maintenance")
    run_bootstrap_account(sys.argv[1:])


if __name__ == "__main__":
    run()
