# Copyright 2005-2022 Canonical Limited. All rights reserved.

import sys
import argparse
import transaction

from canonical.landscape.setup import load_config
from canonical.landscape.model.main.person import get_person_by_identity
from canonical.landscape.model.main.account import PersonAccount, Account
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.resource.store import get_resource_stores
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.scripts.utils import manual_confirm


class CanonicalLeavers(object):
    """
    Script class to remove Landscape privileges from Canonical employees
    leaving the company.

    This class is meant to be run interactively, as it will prompt the user for
    confirmation before any action is taken.
    """

    def __init__(self):
        self.main_store = get_main_store()

    def _get_arguments_parser(self):
        parser = argparse.ArgumentParser(
            description="""\
Remove special privileges from users who are not Canonical employees anymore.

This means, specifically:
1. Removing all global privileges (support provider status etc...)
2. Deleting accounts for which this user is the sole user
3. Removing the user from accounts that have other admins (shared accounts).
            """,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        parser.add_argument(
            "person",
            type=self._openid_to_person,
            metavar="OpenID-URL",
            help="Full or partial OpenID URL identifying the user to remove",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="show what would be done but don't change anything",
        )
        return parser

    def _remove_global_roles(self, person, global_roles):
        """
        This method removes all global roles for the given person.
        """
        for role_name in global_roles:
            person.remove_global_role(role_name)

    def _get_accounts_for_person_id(self, person_id):
        """
        Given a L{Person} id, return a list of L{Account}s.
        """
        matches = self.main_store.find(
            Account,
            PersonAccount.person_id == person_id,
            Account.id == PersonAccount.account_id,
        )
        return list(matches)

    def _filter_accounts_to_remove(self, accounts):
        """
        Split a list of accounts into two lists, one of accounts that are
        eligible to be deleted, and one of accounts that are not.
        """
        to_delete = []
        to_keep = []
        for account in accounts:
            if account.people.count() == 1:
                to_delete.append(account)
            else:
                to_keep.append(account)
        return to_delete, to_keep

    def _delete_accounts(self, accounts):
        for account in accounts:
            account.remove()

    def _remove_person_from_accounts(self, person, accounts):
        """
        Given a L{Person} and a list of L{Account}s, remove the person from the
        corresponding account.
        """
        for account in accounts:
            account.disable_membership(person)

    def _get_person_from_short_openid(self, openid):
        """
        Given a short openID (only the last part of the URL), return a matching
        Person object, or C{None}.
        """
        # Try with the launchpad URL first
        prefixes = [
            "https://login.launchpad.net/+id/%s",
            "https://login.ubuntu.com/+id/%s",
        ]
        person = None
        for prefix in prefixes:
            person = get_person_by_identity(prefix % openid)
            if person is not None:
                break
        return person

    def _openid_to_person(self, openid):
        """
        Given an OpenID URL, return a matching L{Person}.
        """
        if not (openid.startswith("http://") or openid.startswith("https://")):
            person = self._get_person_from_short_openid(openid)
        else:
            person = get_person_by_identity(openid)

        if person is None:
            msg = "Cannot find user matching OpenID URL: %s" % openid
            raise argparse.ArgumentTypeError(msg)
        else:
            return person

    def _confirm_actions(
        self, dry_run, person, global_roles, to_delete, to_keep, stdout
    ):
        """
        This method prints out a summary of changes that will happen, and
        prompts the user for confirmation before continuing.
        Should the user refuse, this method will cause a sys.exit().
        """
        stdout.write("%s (%s)" % (person.name, person.email))

        if len(global_roles):
            stdout.write("\n\n")
            stdout.write(
                "The following global roles will be removed from the " "user:\n"
            )
            for role_name in global_roles:
                stdout.write("- %s \n" % (role_name))

        if len(to_delete):
            stdout.write("\n\n")
            stdout.write("The following accounts will be deleted:\n")
            for account in to_delete:
                stdout.write("- %s (%s)\n" % (account.title, account.name))

        if len(to_keep):
            stdout.write("\n\n")
            stdout.write("User will be removed from:\n")
            for account in to_keep:
                # Print the account name and the license information
                stdout.write("- %s (%s)\n" % (account.title, account.name))
                licenses = LicenseManagement(account).get_licenses(active_only=True)
                for license in licenses:
                    print_date = "NONE!"
                    expiration_date = license.expiration_date
                    if expiration_date is not None:
                        print_date = expiration_date.strftime("%Y-%m-%d")
                    output = " * Id: %s, Seats: %s, Expiration date: %s\n"
                    stdout.write(output % (license.id, license.seats, print_date))

        stdout.write("\n\n")
        if dry_run:
            return False
        return manual_confirm("Type 'YES' to proceed, anything else to " "quit: ")

    def run(self, args, stdout=sys.stdout):
        # The default isolation level that Storm uses for PostgreSQL is
        # repeatable read which is excessive for many operations and causes
        # serialization issues when account removal conflicts with a ping time
        # update.  Read committed is sufficient for removal, but isolation can
        # only be set at transaction start and we don't know exactly which
        # stores we're going to hit. Hence, the brute force and ignorance
        # solution of just dropping *all* stores we *might* touch to read
        # committed before doing anything (even parse_args uses queries)
        sql = "SET TRANSACTION ISOLATION LEVEL READ COMMITTED"
        self.main_store.execute(sql)
        for name, store in get_resource_stores():
            store.execute(sql)
        for name, store in get_account_stores():
            store.execute(sql)

        parser = self._get_arguments_parser()
        args = parser.parse_args(args)

        person = args.person
        accounts = self._get_accounts_for_person_id(person.id)

        to_delete, to_keep = self._filter_accounts_to_remove(accounts)
        global_roles = [role.name for role in person.get_all_roles().get(None, [])]

        if not any([to_delete, to_keep, global_roles, accounts]):
            stdout.write("User already removed.\n")
            return
        proceed = self._confirm_actions(
            args.dry_run, person, global_roles, to_delete, to_keep, stdout
        )
        if not proceed:
            transaction.abort()
            return "Cancelled."

        stdout.write("Removing global roles... ")
        self._remove_global_roles(person, global_roles)
        stdout.write("DONE\nRemoving user from accounts... ")
        self._remove_person_from_accounts(person, accounts)
        stdout.write("DONE\nDeleting accounts... ")
        self._delete_accounts(to_delete)
        stdout.write("DONE\nCommitting transaction... ")
        transaction.commit()
        stdout.write("DONE\n")


def run():
    load_config("maintenance")

    leavers = CanonicalLeavers()
    sys.exit(leavers.run(sys.argv[1:]))
