from __future__ import division

# Copyright 2017 Canonical Limited.  All rights reserved.

from past.utils import old_div
import datetime
from decimal import Decimal
import logging
import textwrap

from zope.component import getUtility

from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.mail.mail import format_address
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.model.billing.collection import PaymentCollection
from canonical.landscape.model.billing.payment import Payment
from canonical.landscape.scripts.batch import BatchScript


class CheckFailedPayments(BatchScript):

    _name = "check-failed-payments"

    INTERVAL_HOURS = 6  # The cron should run every 4 hours so we go bigger.
    NOTIFY = "landscape-billing@lists.canonical.com"

    def run(self, start=None):
        if start is None:
            interval = datetime.timedelta(hours=self.INTERVAL_HOURS)
            start = datetime.datetime.utcnow() - interval
        failed = list(self._iter_failed_payments(start))
        logger = logging.getLogger("check_failed_payments")

        if failed:
            self._send_notification(failed, logger)
        elif start.time() < datetime.time(0o4, 00, 00):  # once per day
            # Essentially, send a keep-alive.
            self._send_notification(failed, logger)
        else:
            logger.info("no payment failures so not sending notification")

    def _iter_failed_payments(self, start):
        for store_name, store in get_account_stores():
            payments = (
                PaymentCollection(
                    store,
                    account=None,
                    statuses=["failed"],
                )
                .find()
                .find(
                    Payment.first_failure >= start,
                )
            )
            for payment in payments:
                if payment.account.is_self_service:
                    yield payment

    def _send_notification(self, payments, logger):
        recipient, subject, body = format_notification(
            self.NOTIFY, payments, self.INTERVAL_HOURS
        )

        logger.info(
            ("sending notification to {} for {} failed payments").format(
                self.NOTIFY, len(payments)
            )
        )

        mailer = getUtility(ILandscapeMailer)
        mailer.send([recipient], subject, body)


def format_notification(addr, payments, interval_hours):
    payments = sorted(payments, key=lambda v: v.first_failure)
    numfailed = len(payments)
    recipient = format_address("", addr)
    subject = "{} Failed Landscape Self-Service Payments".format(numfailed)

    if not payments:
        body = ("No failed payments within the last {} hours.").format(interval_hours)
        return recipient, subject, body

    failed = [
        (" {:>3}   {!s:19}   {!s:10} - {!s:10}   {:>8}").format(
            i,
            p.first_failure.replace(microsecond=0),
            p.period_start_date,
            p.period_end_date,
            "${:.2f}".format(old_div(Decimal(p.due), 10000)),
        )
        for i, p in enumerate(payments, 1)
    ]
    reasons = [
        (" {:>3}   {:20}   {}").format(i, p.stripe_error_code, p.stripe_error_message)
        for i, p in enumerate(payments, 1)
    ]
    lsinfo = [
        (" {:>3}   {:20}   {:32}").format(i, p.account.name, p.payment_id)
        for i, p in enumerate(payments, 1)
    ]
    stripeinfo = [
        (" {:>3}   {:18}   {:27}").format(
            i, p.account.stripe_customer_id, p.stripe_charge_id
        )
        for i, p in enumerate(payments, 1)
    ]
    body = textwrap.dedent(
        """\
        The following payments failed within the last {} hours:

          #         timestamp                period             amount
        ----- --------------------- ------------------------- ----------
        {}
        ----- --------------------- ------------------------- ----------
        total: {}

        reasons:
          #            code              reason
        ----- ---------------------- ---------------
        {}
        ----- ---------------------- ---------------

        Landscape info:
          #          customer                   payment id
        ----- ---------------------- ----------------------------------
        {}
        ----- ---------------------- ----------------------------------

        Stripe info:
          #       customer id               charge id
        ----- -------------------- -----------------------------
        {}
        ----- -------------------- -----------------------------
        """
    ).format(
        interval_hours,
        "\n".join(failed),
        numfailed,
        "\n".join(reasons),
        "\n".join(lsinfo),
        "\n".join(stripeinfo),
    )

    return recipient, subject, body


run = CheckFailedPayments()
