# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Feature flag listing, enabling, and disabling."""

import argparse
import sys

import transaction
from prettytable import PrettyTable

from canonical.landscape.model.feature.flag import (
    enable_feature,
    disable_feature,
    get_feature_flags,
)
from canonical.landscape.setup import load_config


def action_list(args):
    """List the registered feature flags and their status."""
    table = PrettyTable(["Name", "Status", "Description"])
    table.align = "l"
    for flag in get_feature_flags():
        table.add_row(
            [
                flag["name"],
                "Enabled" if flag["enabled"] else "Disabled",
                flag["description"],
            ]
        )
    print(table)


def action_enable(args):
    """Enable specified feature flags."""
    act_on_flags(args.flag, enable_feature, "enabled")


def action_disable(args):
    """Disable specified feature flags."""
    act_on_flags(args.flag, disable_feature, "disabled")


def act_on_flags(flags, method, status):
    for flag in flags:
        result = method(str(flag))
        if result:
            message = status
        elif result is None:
            message = "not {}: change not supported".format(status)
        else:
            message = "already " + status
        print(("{} {}".format(flag, message)))


def main(*args):
    """Validate the command name and call the appropriate handler."""
    parser = argparse.ArgumentParser(prog="featureflags", description=__doc__)
    subparsers = parser.add_subparsers(
        metavar="ACTION", dest="action", help="The command to run"
    )
    subparsers.required = True
    subparsers.add_parser("list", help="List available flags")
    enable_flag = subparsers.add_parser("enable", help="Enable flags")
    enable_flag.add_argument(
        "flag",
        choices=[flag["name"] for flag in get_feature_flags()],
        nargs="+",
        help="Space-separated feature flag names to enable.",
    )
    disable_flags = subparsers.add_parser("disable", help="Disable flags")
    disable_flags.add_argument(
        "flag",
        choices=[flag["name"] for flag in get_feature_flags()],
        nargs="+",
        help="Space-separated feature flag names to disable.",
    )

    args = parser.parse_args(args)

    action = globals()["action_{}".format(args.action)]
    action(args)


def run(argv=sys.argv, main=main, load_config=load_config):
    load_config("scripts")
    with transaction.manager:
        return main(*argv[1:])
