#!/usr/bin/env python3

# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
A script which generates a JSON-formatted description of our API for use in the
API client library and command line program.
"""
import sys
from json import dumps
import argparse

from canonical.txapi.server.registry import Registry

from canonical.landscape.environment import get_modules
from canonical.landscape.api.schema2json import serialize_schema


def get_arguments_parser():
    parser = argparse.ArgumentParser(
        description="Generate API schema for client library."
    )
    parser.add_argument(
        "file", type=argparse.FileType("w"), help="File to save output to."
    )
    parser.add_argument(
        "--hidden",
        default=False,
        action="store_true",
        help="Also generate schema for hidden API methods.",
    )
    return parser


def run():
    args = get_arguments_parser().parse_args(sys.argv[1:])
    registry = Registry()
    for module in get_modules(["api/*.py", "api/repository"]):
        registry.scan(module, ignore=[".tests"])
    result = {}
    for action_name in registry.get_actions():
        for version in registry.get_versions(action_name):
            handler = registry.get(action_name, version)
            if not getattr(handler, "hidden", False) or args.hidden:
                result.setdefault(action_name, {})
                try:
                    schema = handler.schemas[version]
                except AttributeError:
                    schema = handler.schema
                result[action_name][version] = serialize_schema(schema)
                if handler.cli_name is not None:
                    result[action_name][version]["cli_name"] = handler.cli_name
                schema_name = result[action_name][version]["name"]
                if schema_name != action_name:
                    print(
                        "WARNING: action name != schema name (%s != %s)"
                        % (action_name, schema_name)
                    )
    args.file.write(dumps(result, sort_keys=True, indent=4))
    args.file.close()
