from __future__ import division

# Copyright 2005-2016 Canonical Limited.  All rights reserved.

from future import standard_library

standard_library.install_aliases()
from builtins import str
from past.utils import old_div
import argparse
import decimal
import locale
import logging
import os.path
import pkg_resources
from shutil import rmtree
import sys
from tempfile import mkdtemp
from urllib.parse import urljoin

from chameleon import PageTemplateLoader

import stripe
import transaction

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.error import ConnectionRefusedError, ConnectionLost, TimeoutError
from twisted.internet.task import react

from canonical.landscape.model.account.store import (
    get_account_stores,
    get_account_store_by_name,
)
from canonical.landscape.model.billing.collection import PaymentCollection
from canonical.landscape.model.billing.payment import Payment, get_payment_by_id
from canonical.landscape.model.billing.receipt import email_pdf_receipt
from canonical.landscape.model.configuration import (
    get_stripe_secret_key,
    get_billing_credentials,
)
from canonical.landscape.stripe.charge import get_charge
from canonical.landscape.stripe.payment import get_payment_intent
from canonical.landscape.stripe.customer import (
    get_customer,
    get_shipping,
    _get_default_payment_method,
)
from canonical.landscape.ui.constants import STATIC_DIR
from canonical.landscape.scripts.batch import BatchScript
from canonical.openstack.api import HTTPError
from canonical.openstack.clientshelper import OpenStackClientsHelper, ServiceNotFound
from canonical.pdf.pdf import html_to_pdf, pdf_to_pdfa


IMAGES = ["landscape-logo-black-text.png", "credit-card.png"]

RECEIPT_TEMPLATE = (
    "Landscape_receipt_{0.creation_time:%Y}-{0.creation_time:%m}-"
    "{0.creation_time:%d}_ID{0.id}.pdf"
)


class GeneratePaymentReceipts(BatchScript):
    """Generate receipt PDFs for completed payments."""

    _name = "generate-payment-receipts"

    openstack_helper = OpenStackClientsHelper()

    def run(self, argv=sys.argv[1:]):
        self._args = self._parse_args(argv)
        react(self._run)

    @inlineCallbacks
    def _run(self, reactor):
        for account_store_name, payment_id in self._iter_payment_ids():
            try:
                yield self._generate_pdf_email_and_upload(
                    account_store_name, payment_id
                )
            except Exception:
                logging.exception(
                    "Failed to generate PDF for payment ID {}".format(payment_id)
                )

    def _parse_args(self, argv):
        """Parse command line arguments."""
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "--keep-interim",
            action="store_true",
            help="Keep interim HTML files around.",
        )
        return parser.parse_args(argv)

    def _iter_payment_ids(self):
        """Return an iterator yielding successful payments."""
        for store_name, store in get_account_stores():
            collection = PaymentCollection(
                store,
                account=None,
                statuses=["paid", "no-charge"],
                with_receipt_uri=False,
            )
            for payment_id in collection.find().values(Payment.id):
                yield store_name, payment_id

    @inlineCallbacks
    def _generate_pdf_email_and_upload(self, account_store_name, payment_id):
        """Create a receipt PDF."""
        with transaction.manager:
            store = get_account_store_by_name(account_store_name)
            payment = get_payment_by_id(store, payment_id)
            logging.info(
                "Creating a receipt for payment ID {}, account ID {}".format(
                    payment.payment_id, payment.account.id
                )
            )
            html = _generate_html(payment)
            directory, pdfa_path = _convert_to_pdf(payment_id, html)
            payment.receipt_uri = yield _upload_receipt(
                payment, pdfa_path, self.openstack_helper
            )
            if payment.receipt_uri:
                email_pdf_receipt(payment, pdfa_path)
            if self._args.keep_interim:
                logging.info(
                    "Generated receipt for payment {} in {}".format(
                        payment_id, directory
                    )
                )
            else:
                rmtree(directory)


@inlineCallbacks
def _upload_receipt(payment, pdf_path, openstack_helper):
    """Upload the PDF to Swift and return its URI."""
    credentials = get_billing_credentials()
    assert credentials, "Empty credentials"
    try:
        swift_client = yield openstack_helper.get_client("object-store", **credentials)
        container_name = str(payment.account.id)
        yield swift_client.create_container(container_name)
        receipt_filename = RECEIPT_TEMPLATE.format(payment)
        with open(pdf_path, "rb") as fh:
            pdf_data = fh.read()
        yield swift_client.create_object(container_name, receipt_filename, pdf_data)
    except (ServiceNotFound, HTTPError) as e:
        logging.exception(
            "Couldn't upload the receipt for payment ID "
            "{}: {}".format(payment.payment_id, type(e))
        )
        returnValue(None)
    except (ConnectionRefusedError, ConnectionLost, TimeoutError) as e:
        logging.warning(
            "Couldn't upload the receipt for payment ID {}: {}".format(
                payment.payment_id, type(e)
            )
        )
        returnValue(None)
    receipt_uri = u"{}/{}".format(container_name, receipt_filename)
    returnValue(receipt_uri)


def _convert_to_pdf(payment_id, html):
    """Convert HTML for the given payment to PDF/A in the directory."""
    directory = mkdtemp(prefix="receipt-html")
    html_path = os.path.join(directory, "{}.html".format(payment_id))
    pdf_path = os.path.join(directory, "{}.pdf.tmp".format(payment_id))
    pdfa_path = os.path.join(directory, "{}.pdf".format(payment_id))
    with open(html_path, "wb") as fh:
        fh.write(html.encode("utf-8"))
    html_to_pdf(html_path, pdf_path, IMAGES)
    pdf_to_pdfa(pdf_path, pdfa_path)
    return directory, pdfa_path


def _generate_html(payment):
    """Make the HTML for the receipt page."""
    context = _get_context(payment)
    # We use an empty string here because we want the directory
    path_to_templates = os.path.abspath(
        pkg_resources.resource_filename("canonical.landscape.ui.billing", "")
    )
    templates = PageTemplateLoader(path_to_templates)
    template = templates["receipt.pt"]
    return template.render(encoding=None, translate=None, **context)


def _get_context(payment):
    """Collect the context for rendering the template."""
    api_key = get_stripe_secret_key()
    customer = get_customer(api_key, payment.account.stripe_customer_id)
    no_charge = payment.status == "no-charge"
    shipping = get_shipping(customer)
    shipping_name = shipping["name"]
    address = shipping["address"]
    shipping_address = _format_address(
        address["line1"],
        address["line2"],
        address["city"],
        address["state"],
        address["postal_code"],
        address["country"],
    )
    context = {
        "account": payment.account.name,
        "shipping_name": shipping_name,
        "shipping_address": shipping_address,
        "vat_number": payment.account.vat_number,
        "date": str(payment.creation_time.date()),
        "receipt_id": payment.payment_id,
        "currency": "USD",
        "total_before_tax": _format_centicents_total(payment.total),
        "vat": _format_centicents_total(payment.vat),
        "credit": _format_centicents_total(payment.credit_discount),
        "paid": _format_centicents_total(payment.due),
    }
    logo_path = os.path.join(
        STATIC_DIR, "skin", "hokan", "landscape-logo-black-min.svg"
    )
    context["logo_url"] = urljoin("file://", logo_path)
    credit_card_image_path = os.path.join(STATIC_DIR, "billing", "credit-card.png")
    card_holder_details = _get_card_holder_details(
        payment, customer, no_charge, api_key
    )
    context.update(
        {
            "cardholder_name": card_holder_details["name"],
            "cardholder_address": card_holder_details["address"],
            "billing_summary": _format_billing_summary(
                payment.period_start_date, payment.period_end_date
            ),
            "billing_range": _format_billing_range(
                payment.period_start_date, payment.period_end_date
            ),
            "costs": [
                {
                    "unit_price": _format_centicents_rate(cost.hourly_cost),
                    "computer_type": _format_computer_type(cost.computer_type),
                    "total": _format_centicents_total(cost.amount),
                    "hours": _format_hours(cost.hours),
                }
                for cost in payment.costs
            ],
        }
    )
    if not no_charge:
        context.update(
            {
                "credit_card_url": urljoin("file://", credit_card_image_path),
                "card_type": card_holder_details["type"],
                "card_expiration": card_holder_details["expiration"],
                "card_last_four": card_holder_details["last_four"],
            }
        )

    return context


def _get_card_holder_details(payment, customer, no_charge, api_key):
    """Get the card-holder's details."""
    if no_charge:
        # Use the default payment source to get the customer address
        payment_method = _get_default_payment_method(customer)
        if not payment_method:
            # No charge was done (e.g. under free credit) and no default
            # billing source is defined, so fill with placeholder
            # for the void receipt generation.
            return {"name": "n/a", "address": "n/a"}
    else:
        payment_intent_id = payment.stripe_charge_id
        if payment.stripe_charge_id.startswith("ch_"):
            # XXX legacy charges in case any are still unprocessed
            payment_method = get_charge(api_key, payment_intent_id)
        else:
            intent = get_payment_intent(api_key, payment_intent_id)
            payment_method = stripe.PaymentMethod.retrieve(
                intent.payment_method, api_key=api_key
            )

    address = _format_address(
        payment_method.billing_details.address.line1,
        payment_method.billing_details.address.line2,
        payment_method.billing_details.address.city,
        payment_method.billing_details.address.state,
        payment_method.billing_details.address.postal_code,
        payment_method.billing_details.address.country,
    )
    details = {"name": payment_method.billing_details.name, "address": address}
    if not no_charge:
        try:
            card = payment_method.card
        except AttributeError:
            # XXX legacy charges in case any are still unprocessed
            card = payment_method.source
        # Only include credit card info if there's a charge
        details.update(
            {
                "type": card.brand,
                "expiration": u"{}/{}".format(card.exp_month, card.exp_year),
                "last_four": card.last4,
            }
        )
    return details


def _format_address(line1, line2, city, state, zip, country):
    """Return a list with formatted address details."""
    address = [line1]
    if line2:
        address.append(line2)
    address.extend([u"{}, {} {}".format(city, state, zip), country])
    return address


def _format_computer_type(computer_type):
    """Format the computer type to a human readable label."""
    return {
        "physical": "physical machine",
        "vm": "virtual machine",
        "container": "container",
    }[computer_type]


def _format_centicents_total(centicents):
    """Format the centicents total in dollars.

    e.g. "$1,234.56"
    """
    TWO_PLACES = decimal.Decimal("0.01")
    total = decimal.Decimal(centicents)
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
    dollars = (old_div(total, 10000)).quantize(TWO_PLACES)
    return locale.currency(dollars, grouping=True)


def _format_centicents_rate(centicents):
    """Format the hourly centicents rate in dollars.

    e.g. "$0.0001"
    """
    FOUR_PLACES = decimal.Decimal("0.0001")
    rate = decimal.Decimal(centicents)
    dollar_rate = str((old_div(rate, 10000)).quantize(FOUR_PLACES))
    naive = "${}".format(dollar_rate.rstrip("0"))
    unit, sep, fraction = naive.partition(".")
    if len(fraction) < 2:
        fraction += "0"
    return unit + sep + fraction


def _format_hours(hours):
    """Format the number of hours.

    e.g. "12,345"
    """
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
    return locale.format("%d", hours, grouping=True)


def _format_billing_summary(start, end):
    """Format the summary of the billing.

    e.g. "June 2016"
    """
    if start.month == end.month:
        return "{0:%B} {0:%Y}".format(start)
    elif start.year == end.year:
        return "{0:%B} - {1:%B} {1:%Y}".format(start, end)
    else:
        return "{0:%B} {0:%Y} - {1:%B} {1:%Y}".format(start, end)


def _format_billing_range(start, end):
    """Format the range of the billing.

    e.g. "1 June - 30 June 2016"
    """
    if start.year == end.year:
        return "{0.day} {0:%B} - {1.day} {1:%B} {1:%Y}".format(start, end)
    else:
        return "{0.day} {0:%B} {0:%Y} - {1.day} {1:%B} {1:%Y}".format(start, end)


run = GeneratePaymentReceipts()
