#!/usr/bin/python3

"""Query Charm Store API for latest revisions of charms used in LSCI."""
from __future__ import print_function

import argparse
import json

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.task import react
from treq import get, json_content

from canonical.landscape.model.openstack.config import DEFAULT_UBUNTU_SERIES
from canonical.landscape.model.openstack.services import (
    get_charm_configs,
    get_charm_configs_path,
)


def get_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--next",
        action="store_true",
        default=False,
        help="use next charms configuration instead of stable ones.",
    )
    parser.add_argument(
        "--update",
        action="store_true",
        default=False,
        help="update services.py with the latest revisions.",
    )
    parser.add_argument(
        "--ubuntu-series",
        type=str,
        default=DEFAULT_UBUNTU_SERIES,
        metavar="SERIES",
        help="The Ubuntu series the charms are for (default %(default)s).",
    )
    return parser


@inlineCallbacks
def get_latest(charm):
    charm_id = charm["url"].lstrip("cs:").rsplit("-", 1)[0]
    url = "https://api.jujucharms.com/charmstore/v4/%s/meta/any" % charm_id
    response = yield get(url.encode("utf-8"))
    charm_info = yield json_content(response)
    try:
        returnValue(charm_info["Id"])
    except KeyError:
        returnValue("N/A")


@inlineCallbacks
def print_revisions(reactor):
    parser = get_parser()
    options = parser.parse_args()

    charm_series = "next" if options.__next__ else "stable"
    charm_configs = get_charm_configs(
        options.ubuntu_series, charm_series, with_comments=True
    )
    if not charm_configs:
        return

    max_name_len = max(len(name) for name in charm_configs)
    max_url_len = max(len(charm["url"]) for charm in list(charm_configs.values()))
    for name, config in sorted(charm_configs.items()):
        charm_url = config["url"]
        if not options.update:
            print(name.ljust(max_name_len), end=" ")
            print(charm_url.ljust(max_url_len), end=" ")
        latest_url = yield get_latest(config)
        if options.update:
            if charm_url != latest_url:
                config["url"] = latest_url
        elif (
            charm_url.startswith("cs:" + options.ubuntu_series)
            or charm_url.startswith("cs:~openstack-charmers-next/")
        ) and charm_url == latest_url:
            print("✓")
        elif charm_url == latest_url:
            print("⚠")
        else:
            print("✗\t", latest_url)
    if options.update:
        configs_path = get_charm_configs_path(options.ubuntu_series, charm_series)
        with open(configs_path, "w") as configs_file:
            json.dump(charm_configs, configs_file, indent=2, sort_keys=True)
            # Write a newline, since some editors add a newline at the
            # end automatically if one isn't there.
            configs_file.write("\n")


def main():
    react(print_revisions)


if __name__ == "__main__":
    main()
