# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Support for building pre-canned hash=>id databases."""

import os
import sys
import shutil
import tempfile
import optparse
import locale

import transaction

from configparser import ConfigParser, NoOptionError
from landscape.lib.apt.package.store import HashIdStore
from landscape.lib.apt.package.facade import AptFacade, ChannelError
from landscape.lib.apt.package.skeleton import build_skeleton_apt

from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.main.serveridentity import get_server_uuid
from canonical.landscape.environment import get_config_directory
from canonical.landscape.scripts import get_top_level_directory
from canonical.landscape.scripts.batch import BatchScript


class HashIdTool(object):
    """
    Tool to create L{HashIdStore}s using package information from APT channels.

    Optionally this class can populate the L{PackageMarket} with the packages
    associated with the given APT channels which are not yet contained in the
    market.
    """

    def __init__(self, facade, market):
        """
        @param facade: The L{AptFacade} to use for fetching APT channels.
        @param market: The L{PackageMarket} to get/add packages from/to.
        """
        self._facade = facade
        self._market = market

    def load_apt_deb_channels(self, channels, arch):
        """
        Load the given apt-deb C{channels} for the given C{arch}.

        @param channels: A C{list} of C{(url, distribution, components)} tuples
        @param arch: The DPKG architecture to use.
        """
        self._facade.refetch_package_index = True
        self._facade.set_arch(arch)
        self._facade.reset_channels()
        for channel in channels:
            # Trust local for bionic+ to avoid signature validation.
            self._facade.add_channel_apt_deb(*channel, trusted="yes")
        self._facade.reload_channels()

    def populate_market(self):
        """
        Add to the L{PackageMarket} the packages coming from loaded channels.

        If a package available in the load APT channels is already present
        in the L{PackageMarket} it will be ignored.

        @return: The number of new packages that were added.
        """
        hashes = self._facade.get_package_hashes()
        ids = self._market.get_ids(hashes)
        count = 0
        for hash, id in zip(hashes, ids):
            if id is None:
                package = self._facade.get_package_by_hash(hash)
                skeleton = build_skeleton_apt(
                    package, with_unicode=True, with_info=True
                )
                self._market.add(skeleton)
                count += 1
                if count % 500 == 0:
                    transaction.commit()
        transaction.commit()
        return count

    def create_hash_id_store(self, store_filename):
        """
        Create an hash->ids store for the packages coming from loaded channels.

        If a package is not present in the L{PackageMarket}, its hash->id
        mapping will not be included in the generated store.
        """
        hashes = self._facade.get_package_hashes()
        ids = self._market.get_ids(hashes)
        hash_ids = {}
        for hash, id in zip(hashes, ids):
            if id is not None:
                hash_ids[hash] = id
        temp_store_filename = tempfile.mktemp(dir=os.path.dirname(store_filename))
        store = HashIdStore(temp_store_filename)
        store.set_hash_ids(hash_ids)
        os.rename(temp_store_filename, store_filename)

    def parse(self, args):
        """
        Parse the provided command line options.

        @return: a C{ConfigParser} object set according to the provided
            options.
        """
        options = {
            "config": "The configuration file to use.",
            "database-dir": "Directory to save the hash-id stores to",
            "prefix": "The prefix for the stores filenames (default "
            "is the server uuid)",
            "populate": "Add initial packages to the database (which " "must be empty)",
            "releases": "A list of releases to process (e.g. 'hardy')",
            "url": "The APT archive to fetch packages lists from",
            "variants": "Additional release variants (e.g. 'updates')",
            "components": "List of components to include (e.g. 'main')",
            "archs": "List of DPKG archs to process (e.g. 'i386')",
        }
        parser = optparse.OptionParser()
        for option in options:
            parser.add_option(
                "--%s" % option, type="string", action="store", help=options[option]
            )
        options, args = parser.parse_args(args)

        config = ConfigParser()
        if options.config:
            # We use readfp to get nice error messages if the file can't be
            # read
            with open(options.config) as fp:
                config.readfp(fp)
        if options.database_dir:
            config.set("DEFAULT", "database-dir", options.database_dir)
        if options.prefix:
            config.set("DEFAULT", "prefix", options.prefix)
        if options.populate:
            config.set("DEFAULT", "populate", options.populate)
        if options.releases:
            for release in options.releases.split():
                config.add_section(release)
        for section in config.sections():
            for option in ["url", "variants", "components", "archs"]:
                if getattr(options, option) is not None:
                    config.set(section, option, getattr(options, option))

        return config

    def run(self, args):
        """
        Populate the L{PackageMarket} using APT channels defined in a file.

        The format of the file is the same as RFC 822, with one section for
        each APT suite of channels, for example:
        """
        config = self.parse(args)
        store_directory = config.get("DEFAULT", "database-dir")
        if not os.path.exists(store_directory):
            os.mkdir(store_directory)
        stage_directory = tempfile.mkdtemp(prefix=".stage.", dir=store_directory)
        trash_directory = tempfile.mkdtemp(prefix=".trash.", dir=store_directory)
        try:
            self._run_into_stage(config, stage_directory)
            for filename in os.listdir(stage_directory):
                source = os.path.join(stage_directory, filename)
                target = os.path.join(store_directory, filename)
                if os.path.exists(target):
                    shutil.move(target, trash_directory)
                shutil.move(source, store_directory)
        finally:
            shutil.rmtree(stage_directory)
            shutil.rmtree(trash_directory)

    def _run_into_stage(self, config, stage_directory):
        """
        Populate the L{PackageMarket} using APT channels given.

        @param config: A L{ConfigParser} instance.
        @param stage_directory: Where to put the hash id stores.
        """
        try:
            prefix = config.get("DEFAULT", "prefix")
        except NoOptionError:
            prefix = get_server_uuid()
        try:
            populate = config.getboolean("DEFAULT", "populate")
        except NoOptionError:
            populate = False
        if populate:
            store = get_package_store()
            result = store.execute("SELECT id FROM package")
            if result.get_one():
                sys.stderr.write("ERROR: Package database not empty.\n")
                sys.exit(1)
        for section in config.sections():
            url = config.get(section, "url")
            variants = config.get(section, "variants").split()
            # We derive the distro name from the section. "normal" sections
            # are just the distro name, but in some cases we need to have a
            # <distro>-<some identifier> format.
            # One such case is for the ARM archives, where the distro name is
            # the same but the url is different.
            name = section.split("-", 1)[0]
            codenames = [name] + ["%s-%s" % (name, v) for v in variants]
            components = config.get(section, "components").split()
            archs = config.get(section, "archs").split()
            for arch in archs:
                channels = [(url, codename, components) for codename in codenames]
                try:
                    self.load_apt_deb_channels(channels, arch)
                except (ChannelError, IOError) as e:
                    # The exceptions that apt throws when updating all
                    # inherit from IOError.
                    sys.stdout.write("%s.\n" % e)
                    continue
                if populate:
                    count = self.populate_market()
                    sys.stdout.write(
                        "Added %d new packages from %s-%s.\n" % (count, name, arch)
                    )
                self.create_hash_id_store(
                    os.path.join(stage_directory, "%s_%s_%s" % (prefix, name, arch))
                )


class HashIdDatabases(BatchScript):
    """Drive the HashIdTool to update or build hash-id databases."""

    _name = "hash-ids"
    _config = "scripts"

    def run(self):
        # Set the locale to get the right i18n package descriptions
        locale.setlocale(locale.LC_ALL, "en_US.UTF-8")

        apt_dir = tempfile.mkdtemp()
        # XXX make AptFacade properly set keyring when using root dir.
        os.makedirs(os.path.join(apt_dir, "etc/apt"))
        with open(os.path.join(apt_dir, "etc/apt/apt.conf"), "w") as apt_conf:
            apt_conf.write('Acquire::AllowInsecureRepositories "1";')

        facade = AptFacade(root=apt_dir)
        market = PackageMarket()
        tool = HashIdTool(facade, market)
        args = self._args[1:]
        if "--config" not in args:
            root_directory = get_top_level_directory()
            hash_id_databases_conf = os.path.join(
                get_config_directory(root_directory), "hash-id-databases.conf"
            )
            args.extend(["--config", hash_id_databases_conf])

        # Force globally-readable umask for hash files, as they need to be
        # readable by apache.
        os.umask(0o022)
        tool.run(args)
        shutil.rmtree(apt_dir)


run = HashIdDatabases()
run_ignore_maintenance = HashIdDatabases(ignore_maintenance_flag=True)
