# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Support for building pre-canned hash=>id databases."""

import os
import sys
import tempfile
import shutil

import transaction

from landscape.lib.fetch import fetch, FetchError

from canonical.landscape.model.configuration import is_testing_environment
from canonical.landscape.model.main.metarelease import (
    update_meta_releases_from_tagfiles,
    update_meta_releases_upgrades,
)
from canonical.landscape.scripts.batch import BatchScript


CHANGELOGS_URL = "http://changelogs.ubuntu.com/"
META_RELEASE = "meta-release"
META_RELEASE_LTS = "meta-release-lts"


class MetaReleases(BatchScript):
    """Update the database with fresh meta-release data."""

    _name = "meta-releases"
    _config = "scripts"

    def run(self):
        code = update_meta_releases(with_development=is_testing_environment())
        return code


def update_meta_releases(with_development=False, err_pipe=sys.stdout):
    """Fetch all needed meta-release files and update the database.

    @param with_development: A flag indicating whether the current development
        release should be considered or not.
    """
    directory = tempfile.mkdtemp()

    for meta_release in [META_RELEASE, META_RELEASE_LTS]:

        url = CHANGELOGS_URL + meta_release
        if with_development:
            url += "-development"
        try:
            data = fetch(url)
        except FetchError as error:
            err_pipe.write("Failed to fetch %s: '%s'\n" % (url, str(error)))
            shutil.rmtree(directory)
            return 1

        fd = open(os.path.join(directory, meta_release), "wb")
        fd.write(data)
        fd.close()

    update_meta_releases_from_tagfiles(
        os.path.join(directory, META_RELEASE),
        lts_tagfile=os.path.join(directory, META_RELEASE_LTS),
    )
    update_meta_releases_upgrades()
    transaction.commit()
    shutil.rmtree(directory)

    return 0


run = MetaReleases()
