# Copyright 2005-2022 Canonical Limited. All rights reserved.

import optparse
import sys
import time
import json
import os
from subprocess import check_output, check_call

import transaction


from storm.exceptions import TimeoutError
from zope.component import getUtility

from storm.zope.interfaces import IZStorm
from storm.schema.sharding import Sharding
from storm.schema.schema import UnappliedPatchesError

from canonical.timer.stormtimeout import set_allowed_request_time

from canonical.landscape.constants import ROOT_DIR
from canonical.landscape.model.configuration import (
    get_root_url,
    is_charm_deployment,
    ISystemConfiguration,
    set_root_url,
    is_standalone,
    LANDSCAPE_HTTP_PROXY,
    LANDSCAPE_HTTPS_PROXY,
    LANDSCAPE_NO_PROXY,
    LANDSCAPE_SYSTEM_EMAIL,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.license import (
    DEFAULT_LICENSE_PATH,
    FREE_LICENSE_KEYS,
)
from canonical.landscape.model.account.store import get_account_stores
from canonical.landscape.schema import (
    main,
    account,
    resource,
    package,
    session,
    knowledge,
)
from canonical.landscape.schema.connections import (
    ensure_no_other_connections,
    allow_further_connections,
    MoreThanOneConnectionException,
)
from canonical.landscape.schema.bootstrap import schema_bootstrap
from canonical.landscape.schema.hashids import load_hash_id_databases
from canonical.landscape.sample.sample import create_stock_sample_data
from canonical.landscape.setup import load_config
from canonical.landscape.interfaces import IFileBasedConfiguration


# Patch version of the account schema just before the point were we switched
# to a single patch version across all database schemas (see #1407606).
LEGACY_VERSIONS_MARKER = 300

MAX_PREPARED_TRANSACTIONS_QUERY = (
    "SELECT setting FROM pg_settings WHERE name = 'max_prepared_transactions'"
)


# Datatypes to build when the --demo-data option is passed.
DEMO_DATA_DATATYPES = {
    "computers": True,
    "free-disk-space": True,
    "free-memory-and-swap": True,
    "load-averages": True,
    "temperatures": True,
    "network-traffic": True,
    "active-processes": True,
    "hardware": True,
    "packages": True,
    "package-activities": True,
    "script-activities": True,
    "users-and-groups": True,
    "cpu-usage": True,
    "ceph-usage": True,
    "compute-usage": True,
    "swift-usage": True,
    "user-and-group-activities": True,
    "custom-graph": True,
    "scripts": True,
    "openstack": True,
    "package-upgrades": 120,
    "multiple-package-upgrades": 20,
    "package-usns": True,
    "package-holds": 40,
    "system-email": "stan@example.com",
    "extra-juju-environment": False,
    "utilisation": True,
}


# Command template for dumping the schema of a database and updating
# its latest.sql file
LATEST_SQL_UPDATE_COMMAND = """
    pg_dump -T patch --schema-only --no-owner landscape-test-%(database)s \
      | grep -v -e "^\(GRANT\|REVOKE\) ALL" > \
    %(root)s/canonical/landscape/schema/%(schema)s.sql
"""


def print_api_credentials(credentials, stream=sys.stdout):
    """Print out an L{APICredentials} object.

    @note: This output is parsed by robot tests.
    """
    key = credentials.access_key_id
    secret = credentials.access_secret_key
    email = credentials.person.email
    if credentials.account:
        account_name = credentials.account.name
    else:
        account_name = "None"
    stream.write("API credentials: %s %s %s %s\n" % (email, account_name, key, secret))


def get_datatypes(options):
    """Return a dictionary of enabled sample data types.

    @param options: The options object containing the values for all command
        line options.
    """
    datatypes = {}
    datatypes_options = [
        ("computers", options.with_computers),
        ("free-disk-space", options.with_free_disk_space),
        ("free-memory-and-swap", options.with_free_memory_and_swap),
        ("load-averages", options.with_load_averages),
        ("temperatures", options.with_temperatures),
        ("active-processes", options.with_active_processes),
        ("hardware", options.with_hardware),
        ("packages", options.with_packages),
        ("package-activities", options.with_package_activities),
        ("script-activities", options.with_script_activities),
        ("users-and-groups", options.with_users_and_groups),
        ("cpu-usage", options.with_cpu_usage),
        ("ceph-usage", options.with_ceph_usage),
        ("compute-usage", options.with_ceph_usage),
        ("swift-usage", options.with_swift_usage),
        ("user-and-group-activities", options.with_user_and_group_activities),
        ("salesforce-users", options.with_salesforce_users),
        ("custom-graph", options.with_custom_graph),
        ("scripts", options.with_scripts),
        ("network-traffic", options.with_network_traffic),
        ("maas2", options.with_maas2),
        ("openstack", options.with_openstack),
        ("external-nagios", options.with_external_nagios),
        ("create-lds-account-only", options.create_lds_account_only),
        ("utilisation", options.with_utilisation),
    ]

    for datatype, enabled in datatypes_options:
        if enabled or options.with_all_sample_data:
            datatypes[datatype] = True

    if options.with_extra_computers > 0:
        datatypes["extra-computers"] = (
            options.with_extra_computers,
            options.with_extra_computers_data,
        )

    if options.with_extra_interfaces > 0:
        datatypes["extra-interfaces"] = options.with_extra_interfaces

    if options.with_package_test_computers > 0:
        datatypes["package-test-computers"] = options.with_package_test_computers
    if options.with_endpoint_url:
        datatypes["endpoint-url"] = options.with_endpoint_url

    if options.with_root_url:
        datatypes["root-url"] = options.with_root_url

    if options.with_system_email:
        datatypes["system-email"] = options.with_system_email

    if options.with_account_password:
        datatypes["account_password"] = options.with_account_password

    if options.admin_name:
        datatypes["admin-name"] = options.admin_name

    if options.admin_email:
        datatypes["admin-email"] = options.admin_email

    if options.admin_password:
        datatypes["admin-password"] = options.admin_password

    if datatypes.get("packages"):

        if options.with_package_upgrades:
            datatypes["package-upgrades"] = options.with_package_upgrades

        if options.with_multiple_package_upgrades:
            datatypes[
                "multiple-package-upgrades"
            ] = options.with_multiple_package_upgrades

        if options.with_package_usns:
            datatypes["package-usns"] = options.with_package_usns

        if options.with_package_holds:
            datatypes["package-holds"] = options.with_package_holds

    if options.with_utilisation:
        datatypes["utilisation"] = options.with_utilisation

    return datatypes


def handle_existing_licenses(
    license_file=DEFAULT_LICENSE_PATH, check_output=check_output
):
    """
    Remove existing licenses that contain only free licenses that were
    previously installed.
    """
    if os.path.exists(license_file):
        output = check_output(["gpg", "--skip-verify", "--output", "-", license_file])
        contents = json.loads(output)
        collected_license_keys = set(
            license["license_key"] for license in contents["licenses"]
        )
        if not collected_license_keys.difference(FREE_LICENSE_KEYS):
            os.remove(license_file)


def run(args=None, check_output=check_output, license_file=DEFAULT_LICENSE_PATH):

    if args is None:
        args = sys.argv[1:]
    parser = optparse.OptionParser()
    parser.add_option("--with-all-sample-data", action="store_true")
    parser.add_option("--with-computers", action="store_true")
    parser.add_option("--with-package-test-computers", type=int, default=0)
    parser.add_option("--with-extra-computers", type=int, default=0)
    parser.add_option("--with-extra-computers-data", action="store_true")
    parser.add_option("--with-extra-interfaces", type=int, default=0)
    parser.add_option("--with-free-disk-space", action="store_true")
    parser.add_option("--with-free-memory-and-swap", action="store_true")
    parser.add_option("--with-load-averages", action="store_true")
    parser.add_option("--with-temperatures", action="store_true")
    parser.add_option("--with-active-processes", action="store_true")
    parser.add_option("--with-hardware", action="store_true")
    parser.add_option("--with-packages", action="store_true")
    parser.add_option("--with-package-activities", action="store_true")
    parser.add_option("--with-script-activities", action="store_true")
    parser.add_option("--with-users-and-groups", action="store_true")
    parser.add_option("--with-cpu-usage", action="store_true")
    parser.add_option("--with-ceph-usage", action="store_true")
    parser.add_option("--with-compute-usage", action="store_true")
    parser.add_option("--with-swift-usage", action="store_true")
    parser.add_option("--with-user-and-group-activities", action="store_true")
    parser.add_option("--with-salesforce-users", action="store_true")
    parser.add_option("--with-custom-graph", action="store_true")
    parser.add_option("--with-scripts", action="store_true")
    parser.add_option("--with-endpoint-url")
    parser.add_option("--with-network-traffic", action="store_true")
    parser.add_option("--with-root-url")
    parser.add_option("--with-http-proxy")
    parser.add_option("--with-https-proxy")
    parser.add_option("--with-no-proxy")
    parser.add_option("--with-system-email")
    parser.add_option("--with-account-password")
    parser.add_option("--with-package-upgrades", type=int, default=5)
    parser.add_option("--with-multiple-package-upgrades", type=int, default=2)
    parser.add_option("--with-package-usns", type=int, default=1)
    parser.add_option("--with-package-holds", type=int, default=2)
    parser.add_option("--with-maas", action="store_true")
    parser.add_option("--with-maas2", action="store_true")
    parser.add_option("--with-openstack", action="store_true")
    parser.add_option("--with-external-nagios", action="store_true")
    parser.add_option("--with-utilisation", action="store_true")
    parser.add_option("--admin-name")
    parser.add_option("--admin-email")
    parser.add_option("--admin-password")
    parser.add_option("--create-lds-account-only", action="store_true")
    parser.add_option("--configure-lds-only", action="store_true")
    parser.add_option("--timeout", type=int, default=86400)
    parser.add_option("--bootstrap", action="store_true")
    parser.add_option("--demo-data", action="store_true")
    parser.add_option("--update-latest", action="store_true")

    options, args = parser.parse_args(args)
    datatypes = get_datatypes(options)

    print("Loading site configuration...")
    load_config("schema")

    handle_existing_licenses(check_output=check_output, license_file=license_file)

    zstorm = getUtility(IZStorm)
    _set_two_phase_commit_if_available(zstorm)

    timeout = 86400
    if options.timeout:
        timeout = options.timeout
    set_allowed_request_time(timeout)

    if options.bootstrap:
        config = getUtility(IFileBasedConfiguration)
        password = config.get("stores").get("password")
        schema_bootstrap(zstorm, password)

    print(("Setting up database schemas (will timeout after %d seconds) ..." % timeout))

    schemas = _build_schemas()
    _ensure_parallel_patch_versioning(schemas)
    sharding = _build_sharding(schemas)

    try:
        schema_created = create_schema_for_empty_databases(schemas)
        if not schema_created and not options.bootstrap:
            for name, schema, store in schemas:
                if name == "main":
                    main_schema = schema
                    main_store = store
                    break
            try:
                main_schema.check(main_store)
            except UnappliedPatchesError:
                # Ensure we're the only connected user to the database,
                # otherwise bail out.
                if is_charm_deployment():
                    try:
                        ensure_no_other_connections(zstorm)
                    except MoreThanOneConnectionException as e:
                        print((str(e)))
                        sys.exit(-1)
                # Perform the schema upgrade
                try:
                    sharding.upgrade()
                except Exception:
                    transaction.abort()
                    raise
                finally:
                    if is_charm_deployment():
                        allow_further_connections(zstorm)
    except TimeoutError as e:
        print("TimeoutError occured during schema upgrade.")
        print("You can increase the timeout using the --timeout option.")
        transaction.abort()
        # Stop the schema timer, and reinstall the old tracers.
        print((str(e)))
        sys.exit(-1)

    # XXX Disable two phase commit after schema upgrade, since the
    #     sample data code doesn't work with it (due to the direct store
    #     commit in the package market).
    transaction.commit()
    _unset_two_phase_commit(zstorm)

    if options.configure_lds_only:
        _configure_lds(options)
        transaction.commit()
        return

    if options.bootstrap:
        # We can load hash ID databases only after schemas have been set up.
        print("Loading dumps of package hash ID databases...")
        load_hash_id_databases(zstorm, config)
        set_initial_proxy_config(
            options.with_http_proxy, options.with_https_proxy, options.with_no_proxy
        )

    if options.demo_data:
        sys.stdout.write(
            "WARNING: All Landscape data will be erased. Continue? (YES/NO) "
        )
        answer = sys.stdin.readline()
        if answer.strip() != "YES":
            sys.stdout.write("Exiting.\n")
            return
        sharding.delete()
        datatypes.update(DEMO_DATA_DATATYPES)

    if datatypes:
        print("Setting up sample data...")
        # FIXME The start/stop times we use to generate time
        # range-based data should be a command-line parameter, instead
        # of being hard-coded to 6h here.
        stop_time = time.time()
        start_time = stop_time - 6 * 60 * 60
        # Align start-time to the step interval.
        start_time = (start_time // 300) * 300
        snapshots = int((stop_time - start_time) / 300)
        print(
            (
                "Generating %d snapshots from %d to %d."
                % (snapshots, start_time, stop_time)
            )
        )
        invitations, credentials = create_stock_sample_data(
            start_time, stop_time, datatypes
        )
        for invitation in invitations:
            invitation_url = "%saccept-invitation/%s" % (
                get_root_url(),
                invitation.secure_id,
            )
            print(("Invitation at %s" % invitation_url))

        for credential in credentials:
            print_api_credentials(credential)

    if options.update_latest:
        _update_latest_sql(schemas)

    print(
        (
            "Schema patch version: {}".format(
                _get_current_patch_version(get_main_store())
            )
        )
    )
    # Commit everything at the end, this will use 2-phase commit.
    transaction.commit()


def _build_schemas():
    """Build the Storm C{Schema}s to apply to the stores, with their names.

    @return: A list of 2-tuples of the form (<schema name>, <schema object>).
    """
    schemas = []
    zstorm = getUtility(IZStorm)

    for module in [main, account, resource, package, session, knowledge]:
        name = module.__name__.split(".")[-1]
        schema = module.create_schema()
        schema.autocommit(False)
        for store_name in zstorm.get_default_uris():
            if store_name.startswith(name):
                store = zstorm.get(store_name)
        schemas.append((name, schema, store))

    return schemas


def _build_sharding(schemas):
    """Build the Storm C{Sharding} instance used to apply schemas."""
    sharding = Sharding()

    for schema_name, schema, store in schemas:
        sharding.add(store, schema)

    return sharding


def set_initial_proxy_config(http_proxy, https_proxy, no_proxy):
    """Set the initial configuration for HTTP proxing."""
    system_configuration = getUtility(ISystemConfiguration)
    if http_proxy:
        system_configuration.set(LANDSCAPE_HTTP_PROXY, http_proxy)
    if https_proxy:
        system_configuration.set(LANDSCAPE_HTTPS_PROXY, https_proxy)
    if no_proxy:
        system_configuration.set(LANDSCAPE_NO_PROXY, no_proxy)


def create_schema_for_empty_databases(schemas, sharding=None, is_empty=None):
    """Create schema for each database that has no schema.

    @param schemas: A sequence of 3-tuples with name, schema, store for any
        schema module.
    """
    if sharding is None:
        sharding = Sharding()
    if is_empty is None:
        is_empty = _is_store_empty
    needs_schema = False
    for name, schema, store in schemas:
        if is_empty(store):
            needs_schema = True
            sharding.add(store, schema)
    if needs_schema:
        sharding.create()
    return needs_schema


def _ensure_parallel_patch_versioning(schemas):
    """Migration code for transitioning to parallel patch versioning.

    Once we're sure that no deployment is left that is still using the
    non-parallel patch numbering, this code can be dropped.
    """
    # Our target patch version is whatever the current patch version of
    # the account database is, since it's the one with highest number
    # at the time parallel patching has been introduced.
    _, account_store = list(get_account_stores())[0]
    target_version = _get_current_patch_version(account_store)

    # If there's no patch at all, we're creating the schema from scratch, so
    # there's nothing to do
    if target_version is None:
        return

    # If the patch version of the main database is the same as the one of
    # the account database, it means that migration was already performed
    if _get_current_patch_version(get_main_store()) == target_version:
        return

    zstorm = getUtility(IZStorm)
    for schema_name, schema, store in schemas:
        for store_name in zstorm.get_default_uris():
            _apply_legacy_patches(store, schema)


def _apply_legacy_patches(store, schema):
    """Apply pre-parallel patches to the given store.

    This function will apply all pending patches up to the point where parallel
    patching was introduced, so we can then go from here.
    """
    # Get all already-applied versions
    result = store.execute("SELECT version FROM patch")
    versions = [row[0] for row in result.get_all()]

    for version in range(1, LEGACY_VERSIONS_MARKER + 1):
        if version not in versions:
            # This version hasn't been applied yet, let's do it now.
            # XXX the method name "advance" is misleading since all
            # the it does it's applying the patch with given version,
            # and not really any iterim patches. It should be renamed
            # in Storm.
            schema.advance(store, version)


def _is_store_empty(store):
    """Return True if store has no patch table defined."""
    # Check if the patch table is there, if not it means there's no applied
    # patch at all.
    result = store.execute(
        "SELECT * FROM information_schema.tables "
        "WHERE table_schema = 'public' AND table_name = 'patch'"
    )
    return result.rowcount == 0


def _get_current_patch_version(store):
    """Get the current patch version of the given store."""
    if _is_store_empty(store):
        return None

    result = store.execute("SELECT max(version) FROM patch")
    return result.get_one()[0]


def _set_two_phase_commit_if_available(zstorm):
    """Set two-phase commit if PostgreSQL allows prepared transactions.

    # XXX eventually we should enable two-phase commit everywhere if
    # available, not only in the schema script.
    """
    tpc = False
    for name in zstorm.get_default_uris():
        if not name.startswith("postgres-"):
            continue
        store = zstorm.get(name)
        result = store.execute(MAX_PREPARED_TRANSACTIONS_QUERY)
        if int(result.get_one()[0]) > 0:
            tpc = True
        break
    if not tpc:
        print(
            "WARNING: PostgreSQL has max_prepared_transactions set to 0, not "
            "using two-phase commit."
        )
        return
    for name in zstorm.get_default_uris():
        if name.startswith("postgres-"):
            continue
        zstorm.set_default_tpc(name, True)


def _unset_two_phase_commit(zstorm):
    """Unset two-phase commit."""
    for name, store in zstorm.iterstores():
        if name.startswith("postgres-"):
            continue
        zstorm.set_default_tpc(name, False)
        store._tpc = False


def _update_latest_sql(schemas):
    zstorm = getUtility(IZStorm)

    for schema_name, _, _ in schemas:
        for store_name in zstorm.get_default_uris():
            if store_name.startswith(schema_name):
                command = LATEST_SQL_UPDATE_COMMAND % {
                    "database": store_name,
                    "schema": schema_name,
                    "root": ROOT_DIR,
                }
                check_call([command], shell=True)
                break


def _configure_lds(options):
    """Set LDS configuration."""
    if not is_standalone():
        return
    config = getUtility(ISystemConfiguration)
    if options.with_http_proxy is not None:
        config.set(LANDSCAPE_HTTP_PROXY, options.with_http_proxy)
    if options.with_https_proxy is not None:
        config.set(LANDSCAPE_HTTPS_PROXY, options.with_https_proxy)
    if options.with_no_proxy is not None:
        config.set(LANDSCAPE_NO_PROXY, options.with_no_proxy)
    if options.with_root_url is not None:
        set_root_url(options.with_root_url)
    if options.with_system_email is not None:
        config.set(LANDSCAPE_SYSTEM_EMAIL, options.with_system_email)
