# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import re
import sys
import fnmatch
import optparse

from PIL import Image

from canonical.landscape.ui.spriteformat import class_name_from_filename


IMAGE_EXTS = (".png", ".gif")
IMAGE_RE = re.compile(
    "[\"']((.*)\.(%s))[\"']" % ("|".join([ext[1:] for ext in IMAGE_EXTS]))
)

CSS_ICON_DEFAULT = """
.icon {
  display: -moz-inline-stack;
  display: inline-block;
  zoom: 1;
  *display: inline;
}

.smaller-icon {
  width: 14px;
  height: 14px;
}

.small-icon {
  width: 16px;
  height: 16px;
}

.large-icon {
  width: 32px;
  height: 32px;
}
"""

CSS_ICON_TEMPLATE = """
.%(class-name)s {
  background-repeat: %(repeat)s;
  background-image: url("%(image)s"); %(sprite-ref)s
}
"""

CSS_TEMPLATE = """
.%(class-name)s {
  background-image: url("%(image)s"); %(sprite-ref)s
  background-repeat: %(repeat)s;
  width: %(width)spx;
  height: %(height)spx;
}
"""

SPRITE_TEMPLATE = """
/** \
sprite: %(sprite-id)s; \
sprite-image: url('sprites/%(sprite-id)s.png'); \
sprite-layout: %(sprite-layout)s; \
*/
"""

SPRITE_REFERENCE = """\
/** \
sprite-ref: %(sprite-id)s; \
sprite-alignment: %(sprite-alignment)s; \
*/
"""


def normalize(path):
    return os.path.abspath(os.path.normpath(os.path.expanduser(path)))


def relative_path(from_file, to_file):
    """Return the relative path between from_file and to_file."""
    # Create a list of directories to each file.
    abs_from_path = os.path.split(os.path.abspath(from_file))[0].split(os.path.sep)
    abs_to_path = os.path.split(os.path.abspath(to_file))[0].split(os.path.sep)
    # Remove all the directories they have in common.
    while abs_from_path and abs_to_path and abs_from_path[0] == abs_to_path[0]:
        del abs_from_path[0]
        del abs_to_path[0]

    # For each directory still in the from_path, we need to go up a level.
    rel_path = [os.path.pardir] * len(abs_from_path)
    rel_path.extend(abs_to_path)
    rel_path.append(os.path.basename(to_file))

    return os.path.join(*rel_path)


def get_style(size):
    """Get a custom style based on image dimensions."""
    width, height = size
    style = ""
    repeat = "no-repeat"

    if width == 1:
        repeat = "repeat-x"
        style = "width: 10px"
    elif height == 1:
        repeat = "repeat-y"
        style = "height: 10px"
    return style, repeat


def classify(size):
    """Classify an image into a specific group, based on image dimensions."""
    width, height = size
    group_name = "general"
    if width == height:
        if width == 1:
            group_name = "bg"
        if width == 14:
            group_name = "smaller-icon"
        elif width == 16:
            group_name = "small-icon"
        elif width == 32:
            group_name = "large-icon"
    elif width == 1 or height == 1:
        group_name = "bg"
    elif (10 <= width <= 32) and (10 <= height <= 32):
        group_name = "other-icon"
    return group_name


def images_from_zcml(filename):
    """Given a ZCML file, parse it and return a list of image filenames."""
    images = set()
    zcml = open(filename, "r")
    try:
        for line in zcml:
            m = IMAGE_RE.search(line)
            if m:
                images.add(m.group(1))
        return images
    finally:
        zcml.close()


def find_images(root, exclude=None, group_by_section=None):
    """
    Find images from root directory, filtering out based on the
    exclusion patterns specified.
    """
    if exclude is None:
        exclude = ()
    if group_by_section is None:
        group_by_section = {}

    images = {}

    for base, dirs, files in os.walk(root):
        for rule in exclude:
            for remove in fnmatch.filter(dirs, rule):
                dirs.remove(remove)

        for rule in exclude:
            files = set(files) - set(fnmatch.filter(files, rule))

        for fname in files:
            if os.path.splitext(fname)[-1] not in IMAGE_EXTS:
                continue

            image = os.path.join(base, fname)
            section_name = None

            for name, candidates in list(group_by_section.items()):
                for candidate in candidates:
                    if image.endswith(candidate):
                        section_name = name
                        break
                if section_name is not None:
                    break

            if section_name is None:
                section_name = os.path.basename(base)

            section = images.setdefault((section_name, Image.open(image).size), [])
            section.append(image)

    return images


def generate_sprites(images_by_section, target):
    data = []
    seen = {}
    for (section, size), images in list(images_by_section.items()):
        group_name = classify(size)
        style, repeat = get_style(size)
        for image in sorted(images):
            sprite_group_name = group_name
            # For images classified as 'smaller', map them to the
            # 'small' sprite, since they fit into that space. Same
            # applies for 'other' and 'large', respectively.
            if group_name == "smaller-icon":
                sprite_group_name = "small-icon"
            elif group_name == "other-icon":
                sprite_group_name = "large-icon"
            layout = "vertical"
            if repeat == "repeat-y":
                layout = "horizontal"
            sprite_id = "%s-%s-%s" % (section, sprite_group_name, layout)
            if sprite_id not in seen:
                seen[sprite_id] = True
                data.append(
                    (SPRITE_TEMPLATE, {"sprite-id": sprite_id, "sprite-layout": layout})
                )

            class_name = class_name_from_filename(image)

            if group_name in ("smaller-icon", "small-icon", "large-icon"):
                template = CSS_ICON_TEMPLATE
            else:
                template = CSS_TEMPLATE

            alignment = "repeat"
            if repeat == "no-repeat":
                alignment = "left"

            width, height = size
            data.append(
                (
                    template,
                    {
                        "class-name": class_name,
                        "width": width,
                        "height": height,
                        "repeat": repeat,
                        "image": relative_path(target, image),
                        "sprite-id": sprite_id,
                        "sprite-alignment": alignment,
                        "sprite-ref": SPRITE_REFERENCE
                        % {"sprite-id": sprite_id, "sprite-alignment": alignment},
                    },
                )
            )
    return data


def minimize_sprites(images_by_section):
    for (section, size), images in list(images_by_section.items()):
        if len(images) < 5:
            images_by_section.setdefault(("general", size), []).extend(images)
            del images_by_section[(section, size)]
    return images_by_section


def main(args):
    parser = optparse.OptionParser()
    parser.add_option(
        "-x",
        "--exclude",
        action="append",
        help="Globbing expression to match for exclusions.",
    )
    parser.add_option("-t", "--target", help="Target directory for generated content.")
    parser.add_option("-s", "--source", help="Source directory for images.")
    parser.add_option(
        "-f",
        "--from-zcml",
        help=(
            "Parse a ZCML file and group sprites found. "
            "Use '<groupname>=<zcml-file>' to pick the group "
            "name for the images found in the ZCML file. "
            "Those images will end up in a sprite named after "
            "<groupname>, and will be removed from their "
            "original group."
        ),
        action="append",
    )
    opts, args = parser.parse_args(args)

    if not opts.target:
        raise ValueError("Target is required")
    if not opts.source:
        raise ValueError("Source is required")

    opts.target = normalize(opts.target)
    opts.source = normalize(opts.source)

    group_by_section = {}
    if opts.from_zcml is not None:
        for group in opts.from_zcml:
            section, fname = group.split("=")
            group_by_section[section] = images_from_zcml(fname)
    images_by_section = find_images(opts.source, opts.exclude, group_by_section)
    images_by_section = minimize_sprites(images_by_section)

    css_path = os.path.join(opts.target, "sprite.css")
    css = open(css_path, "w")
    css.write(CSS_ICON_DEFAULT)

    for template, args in generate_sprites(images_by_section, css_path):
        css.write(template % args)

    css.close()


def run():
    main(sys.argv[1:])
