# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import re
import subprocess
from platform import node
from OpenSSL.crypto import load_certificate, FILETYPE_PEM


from canonical.landscape.scripts.standalone.base import SetupScriptError
from canonical.landscape.scripts.standalone.plugins.base import TemplatePlugin
from canonical.landscape.model.configuration import (
    LANDSCAPE_CUSTOM_SSL_CA,
    LANDSCAPE_HOME,
)


class UnknownSSLHostnameError(Exception):
    """
    Raised when we are unable to determine the hostname for an SSL
    certificate file.
    """


LANDSCAPE_CERTIFICATE_FILE = "/etc/ssl/certs/landscape_server.pem"
LANDSCAPE_CERTIFICATE_KEY_FILE = "/etc/ssl/private/landscape_server.key"
AVAILABLE_SITES_DIR = "/etc/apache2/sites-available/"


def is_valid_hostname(hostname):
    """
    @returns True unless hostname
             * is None or 0 in length
             * is > 255 characters in length
             * has a segment > 63 characters in length
             * has double dots.
             * a segment begins or ends with -
             * contains characters other than alphanumeric and - or .
    """
    if hostname is None or len(hostname) > 255:
        return False
    allowed = re.compile("[A-Z\d-]{1,63}$", re.IGNORECASE)
    for segment in hostname.split("."):
        if segment.startswith("-") or segment.endswith("-"):
            return False
        if not allowed.match(segment):
            return False
    return True


def check_common_name(common_name, logger):
    """
    Check that common_name is a valid hostname. If common_name starts with '*'
    replace the '*' with the first segment returned by platform#node.

    @param common_name: the Common Name to check.
    @param logger: a C{Logger} used to report any errors.
    """
    if common_name is not None and common_name.startswith("*"):
        hostname = node().split(".")[0]
        logger.warn(
            "Replacing * in commonName %s with local hostname %s"
            % (common_name, hostname)
        )
        common_name = "%s%s" % (hostname, common_name[1:])

    if not is_valid_hostname(common_name):
        logger.error("%s is not a valid hostname." % common_name)
        raise UnknownSSLHostnameError()
    return common_name


def get_common_name_from_certificate_file(filename, logger):
    """
    Use OpenSSL to parse an X.509 certificate and extract the common name
    from the certificate.


    @param filename: Filename to extract the common name from.
    @param logger: a C{Logger} used to report any errors to.
    @return: The Common Name extracted from the certificate file.
    @raises: UnknownSSLHostnameError if none could be extracted.
    """
    data = open(filename).read()

    x509 = load_certificate(FILETYPE_PEM, data)
    common_name = x509.get_subject().commonName
    return check_common_name(common_name, logger)


class ApacheSSLPlugin(TemplatePlugin):
    """
    This plugin updates the landscape virtual machine Apache configuration.
    """

    DEFAULT_TEMPLATE_FILE = os.path.join(
        LANDSCAPE_HOME, "standalone/templates/apache_default.tmpl"
    )

    landscape_certificate_file = LANDSCAPE_CERTIFICATE_FILE
    landscape_certificate_key_file = LANDSCAPE_CERTIFICATE_KEY_FILE
    available_sites_dir = AVAILABLE_SITES_DIR

    log_message = "Setting up Apache SSL configuration ..."

    def main(self):
        if not self.has_ssl_certificate_and_key_file():
            raise SetupScriptError("SSL Certificate and key file not found.")
        super(ApacheSSLPlugin, self).main()
        self.clean_previous_vhosts()
        self.enable_apache_vhost(os.path.basename(self.get_default_output_file()))
        self.reload_apache()

    def enable_apache_vhost(self, vhost):
        """
        Runs a2ensite to enable a Vhost for Apache.

        @param vhost: vhost to enable
        """
        command_line = ["/usr/sbin/a2ensite", vhost]
        process = subprocess.Popen(command_line, stdout=subprocess.PIPE)
        output, errors = process.communicate()
        return output

    def disable_apache_vhost(self, vhost):
        """
        Runs a2dissite to disable the specified Vhost for Apache.

        @param vhost: vhost to disable.
        """
        command_line = ["/usr/sbin/a2dissite", vhost]
        process = subprocess.Popen(command_line, stdout=subprocess.PIPE)
        output, errors = process.communicate()
        return output

    def reload_apache(self):
        """
        Run the command-line to ask Apache to reload its configuration.
        """
        command_line = ["/etc/init.d/apache2", "reload"]
        process = subprocess.Popen(command_line, stdout=subprocess.PIPE)
        output, errors = process.communicate()
        return output

    def get_substitution_variables(self):
        """
        Return the variables available to the Genshi template.
        """
        # NOTE: This is not a standard option, and should only be used by
        # people who understand the implications!
        if os.path.exists(LANDSCAPE_CUSTOM_SSL_CA):
            ssl_certificate_chain_crt = LANDSCAPE_CUSTOM_SSL_CA
        else:
            ssl_certificate_chain_crt = None

        hostname = get_common_name_from_certificate_file(
            self.landscape_certificate_file, self.logger
        )

        data = {
            "ssl_certificate_crt": self.landscape_certificate_file,
            "ssl_certificate_key": self.landscape_certificate_key_file,
            "ssl_certificate_chain_crt": ssl_certificate_chain_crt,
            "hostname": hostname,
        }
        return data

    def clean_previous_vhosts(self):
        """
        Disable old-format site config files and remove them from apache's
        sites-available folder.

        Removes any files in sites-available/sites-enabled that do not have a
        ".conf" extension.  We (quickstart script) placed files like this in
        previous versions of landscape.  In apache version 2.4 (trusty), apache
        now requires ".conf" as the extension for all site config files.
        """
        old_vhost_file = os.path.splitext(self.get_default_output_file())[0]
        if os.path.exists(old_vhost_file):
            old_vhost = os.path.basename(old_vhost_file)
            self.disable_apache_vhost(old_vhost)
            os.unlink(old_vhost_file)

    def get_default_output_file(self):
        """
        Override the default get_default_output_file to return a sensible name
        based upon the common_name appened with .conf.
        """
        return os.path.join(
            self.available_sites_dir,
            "%s.conf"
            % get_common_name_from_certificate_file(
                self.landscape_certificate_file, self.logger
            ),
        )

    def has_ssl_certificate_and_key_file(self):
        """
        Returns True if we have both certificate files needed for the Apache
        SSL setup.
        """
        return os.path.exists(self.landscape_certificate_file) and os.path.exists(
            self.landscape_certificate_key_file
        )
