# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
from base64 import encodebytes
from configparser import RawConfigParser

from canonical.landscape.scripts.standalone.base import GUEST_BASE
from canonical.landscape.scripts.standalone.plugins.base import Plugin, TemplatePlugin
from canonical.landscape.scripts.standalone.plugins.apache import (
    get_common_name_from_certificate_file,
    LANDSCAPE_CERTIFICATE_FILE,
)
from canonical.landscape.model.configuration import get_custom_ssl_ca


class LandscapeConfigBasePlugin(TemplatePlugin):
    """
    This plugin base class is used to replace various conf and ZCML files that
    are used implicitly by developers in a development environment with ones
    that substitute the defined value for the base install directory used by
    the landscape appliance.

    In addition to convenience for developers (e.g., not having to create
    /srv/landscape and any necessary subdirectories), this plugin allows for
    maintaining a single definition of the base directory, freeing developers
    from maintaining it in a collection of other files as well as in the
    scripts.stanalone.host module.
    """

    def get_substitution_variables(self):
        """
        Obtain the Landscape virtual appliance install base directory and
        return them in a dictionary for use by the template engine.
        """
        return {
            "base_dir": GUEST_BASE,
        }


class ClientBootstrapPlugin(Plugin):
    """
    This plugin generates a landscape-client.conf file for bootstrapping
    clients.
    """

    bootstrap_conf_file = os.path.join(GUEST_BASE, "apacheroot/bootstrap.conf")

    log_message = "Setting up Landscape client configuration import file ..."

    def main(self):
        super(ClientBootstrapPlugin, self).main()
        hostname = get_common_name_from_certificate_file(
            LANDSCAPE_CERTIFICATE_FILE, self.logger
        )
        config = RawConfigParser()
        config.add_section("client")
        config.set("client", "account_name", "standalone")
        config.set("client", "url", "https://%s/message-system" % hostname)
        config.set("client", "ping_url", "http://%s/ping" % hostname)
        ssl_ca = get_custom_ssl_ca()
        if ssl_ca:
            ssl_ca = ssl_ca.encode()

            encoded_value = "base64:" + "\n".join(
                [" " + l for l in encodebytes(ssl_ca).decode().splitlines()]
            )
            config.set("client", "ssl_public_key", encoded_value)
        config.write(open(self.bootstrap_conf_file, "w"))
