# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
from xml.sax import saxutils
from configparser import RawConfigParser
import os

from genshi.template.text import TextTemplate

from canonical.landscape.scripts.standalone.base import SetupScriptError

from landscape.client.manager.scriptexecution import get_user_info


class Plugin(object):
    """
    The plugin class needs two things from the setup script that runs is:
         * the configuration information, and
         * the logger
    The setup script is the parent object, and the configuration and logger are
    both obtained from accessing the plugin's parent attribute.

    In addition, if we want the plugin to log a start up message, then the
    log_message attribute needs to be set to something appropriate.
    """

    MESSAGE_PREFIX = "Landscape virtual appliance - "
    service_configuration_file = "/etc/landscape/service.conf"

    log_message = ""

    def __init__(self, logger):
        self.logger = logger

    def log(self, *args, **kwargs):
        """Simple logging.log wrapper. Will skip logging if not set up."""
        if self.logger:
            self.logger.log(*args, **kwargs)

    def main(self):
        """Define all the methods to be called when the plugin is run.

        This method needs to be super'ed by all subclasses, and then have the
        appropriate super class' plugin methods called.
        """
        if self.log_message:
            self.log(logging.INFO, self.log_message)

    def run(self):
        """Execute the plugin.

        This method is responsible for wrapping all the calls made in any
        "main" method. This method should not be overridden; instead, any
        special execution patterns should be implemented in other methods and
        placed in a superclassed "main" method.
        """
        try:
            self.main()
        except Exception as error:
            self.log(logging.ERROR, str(error))
            raise SetupScriptError(error)


class TemplatePlugin(Plugin):
    """This is a base class for plugins that need Genshi templating."""

    # These need to be defined in the subclass. The DEFAULT_OUTPUT_FILE
    # constant is the path to the configuration file that will be generated by
    # the TemplatePlugin subclass.
    DEFAULT_OUTPUT_FILE = ""
    # The DEFAULT_TEMPLATE_FILE constant is the path to the template file that
    # will be used to generate the final file, after values have been
    # substituted in.
    DEFAULT_TEMPLATE_FILE = ""

    def __init__(self, script, template_file=None, output_file=None):
        super(TemplatePlugin, self).__init__(script)
        self._output_file = output_file
        self.template_file = template_file or self.DEFAULT_TEMPLATE_FILE

    @property
    def output_file(self):
        return self._output_file or self.get_default_output_file()

    def get_default_output_file(self):
        """
        Returns the default output file for a TemplatePlugin subclass.
        """
        return self.DEFAULT_OUTPUT_FILE

    def get_substitution_variables(self):
        """
        Return a dictionary of named variables and their values that will be
        used by the templating engine to substitute values in the plugin's
        template file for final output.
        """
        raise NotImplementedError()

    def main(self):
        """
        This runs the template with the variables in the config and
        generates the working file.
        """
        super(TemplatePlugin, self).main()
        input_handle = open(self.template_file)
        template = TextTemplate(input_handle)
        input_handle.close()
        variables = self.get_substitution_variables()
        output_handle = open(self.output_file, "w")
        text = template.generate(**variables).render("text")
        output_handle.write(text)
        output_handle.close()

    def xml_escape(self, string):
        return saxutils.escape(string)


def write_service_configuration(
    sections, filename, user="landscape", overwrite=True, new_filename=None
):
    """
    Update or create a new configuration file, with the named sections, and
    chown and chmod it to be readable by the named user.


    @param sections: Dictionary of sections to be written to the
        service.conf in the form:
          "section_name": {"key_name": key_value}
        If the value for a key is C{None}, the key is removed from the config.
    @param filename: the filename to be written
    @param user: the username the file should be chown'ed to.
    @param overwrite: True if any previous values should be overwritten.
    @return True if the configuration was changed.
    """
    config = RawConfigParser()
    if not config.read([filename]):
        new_file = True
    else:
        new_file = False

    changed = False
    for section, items in list(sections.items()):
        if not config.has_section(section):
            config.add_section(section)
            changed = True

        for key, value in list(items.items()):
            has_option = config.has_option(section, key)

            if has_option and not overwrite:
                continue

            if value is None:
                existed = config.remove_option(section, key)
                if existed:
                    changed = True
            elif not has_option or config.get(section, key) != value:
                config.set(section, key, value)
                changed = True

    output_filename = new_filename or filename
    with open(output_filename, "w+") as config_file:
        config.write(config_file)

    if new_file or (output_filename != filename):
        chmod_and_chown_file(output_filename, user)

    return changed


def delete_from_service_configuration(sections, filename):
    """
    Deletes the named sections from a configuration file. If filename does not
    point to a file, nothing is done and False is returned.

    @param sections: list of sections to be deleted.
    @param filename: the filename to be written if sections are deleted.
    @return True if the configuration was changed.
    """
    changed_configuration = False
    config = RawConfigParser()
    if config.read([filename]):
        for section in sections:
            if config.has_section(section):
                config.remove_section(section)
                changed_configuration = True

    if changed_configuration:
        with open(filename, "w+") as config_file:
            config.write(config_file)

    return changed_configuration


def chmod_and_chown_file(filename, user):
    """Chown and chmod filename to be rw-r----- and uid:0."""
    uid, _, _ = get_user_info(user)
    os.chmod(filename, 0o640)
    os.chown(filename, uid, 0)
