# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import sys
import subprocess
from configparser import RawConfigParser
import pwd

from io import StringIO
from logging import INFO, ERROR, WARN

import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

from canonical.lib.text import password_decode, password_encode, generate_password

from canonical.landscape.application import LANDSCAPE_CONFIG
from canonical.landscape.scripts import schema
from canonical.landscape.scripts.standalone.base import SetupScriptError
from canonical.landscape.scripts.standalone.plugins.base import Plugin
from canonical.landscape.model.main.serveridentity import (
    ensure_server_uuid_exists,
    get_server_uuid,
)

STANDALONE_SCHEMA = "standalone-schema"
STANDALONE_DATABASES = [
    "landscape-standalone-main",
    "landscape-standalone-account-1",
    "landscape-standalone-resource-1",
    "landscape-standalone-package",
    "landscape-standalone-session",
    "landscape-standalone-knowledge",
]

HASHID_STORES_PATH = "/var/lib/landscape/hash-id-databases"
PACKAGE_STOCK_DATA_FILENAME = os.path.join(
    HASHID_STORES_PATH, "landscape-standalone-package.sql"
)

AUTOMATICALLY_GENERATED_PASSWORD_LENGTH = 16


class PasswordGenerationMixin(object):
    def generate_password(self):
        """Return a randomly chosen password string."""
        return generate_password(AUTOMATICALLY_GENERATED_PASSWORD_LENGTH)


class ServiceConfigurationMixin(object):
    """
    Used as a base class for classes that have to read or write to the
    /etc/landscape/service.conf file.
    """

    def write_service_configuration(self, configuration):
        """
        Write out a service.conf format file using the configuration
        dictionary.

        The keys in the dictionary are:
            host: The hostname to connect to.
            password: The password for the landscape user.
            superuser: The landscape schema user.
            superuser_password: The password for the Landscape schema user.
        """
        configuration_file = RawConfigParser()
        # the .read() method returns the list of files it read
        if not configuration_file.read([self.service_configuration_file]):
            new_file = True
        else:
            new_file = False

        if not configuration_file.has_section("stores"):
            configuration_file.add_section("stores")
        configuration_file.set("stores", "user", "landscape")

        if configuration.get("host"):
            configuration_file.set("stores", "host", configuration["host"])

        if configuration.get("port"):
            configuration_file.set("stores", "port", configuration["port"])

        if configuration.get("password"):
            configuration_file.set("stores", "password", configuration["password"])

        if not configuration_file.has_section("schema"):
            configuration_file.add_section("schema")

        if configuration.get("superuser"):
            configuration_file.set("schema", "store_user", configuration["superuser"])

        if configuration.get("superuser_password"):
            configuration_file.set(
                "schema", "store_password", configuration["superuser_password"]
            )

        with open(self.service_configuration_file, "w+") as output_file:
            configuration_file.write(output_file)

        if new_file:
            landscape = pwd.getpwnam("landscape")
            os.chmod(self.service_configuration_file, 0o640)
            os.chown(self.service_configuration_file, 0, landscape[3])

    def read_service_configuration(self):
        """
        Read /etc/landscape/service.conf and return the store information for
        connections to the database.

        The keys in the dictionary are:
            host: The hostname to connect to.
            password: The password for the landscape user.
            superuser: The landscape schema user.
            superuser_password: The password for the Landscape schema user.
        """
        configuration_file = RawConfigParser()
        if not (
            configuration_file.read([self.service_configuration_file])
            == [self.service_configuration_file]
        ):
            raise SetupScriptError("Missing service configuration file.")
        configuration = {}
        port = None
        if configuration_file.has_option("stores", "host"):
            host_config = configuration_file.get("stores", "host")
            if ":" in host_config:
                # host and port
                host, port = host_config.split(":")
                configuration["host"] = host
                configuration["port"] = port
            else:
                # host only
                configuration["host"] = host_config

        if port is None and configuration_file.has_option("stores", "port"):
            configuration["port"] = configuration_file.get("stores", "port")

        if configuration_file.has_option("stores", "password"):
            configuration["password"] = password_decode(
                configuration_file.get("stores", "password")
            )

        if configuration_file.has_option("schema", "store_user"):
            configuration["superuser"] = configuration_file.get("schema", "store_user")

        if configuration_file.has_option("schema", "store_password"):
            configuration["superuser_password"] = password_decode(
                configuration_file.get("schema", "store_password")
            )

        return configuration


class DatabaseConnectionBase(ServiceConfigurationMixin):
    """
    This base class should be used by plugins that require a connection to
    the databases specified in /etc/landscape/service.conf.
    """

    def get_dsn(self, dbname):
        """
        @return DSN for connecting to the database, using the main class's
        get_config() method to get configuration details.
        """
        configuration = self.read_service_configuration()
        dsn = "dbname=%s user=%s host=%s" % (
            dbname,
            configuration["superuser"],
            configuration.get("host", "localhost"),
        )

        if configuration.get("port"):
            dsn = "%s port=%s" % (dsn, configuration.get("port"))
        if configuration.get("superuser_password"):
            dsn = "%s password=%s" % (
                dsn,
                password_decode(configuration.get("superuser_password")),
            )
        return dsn

    def get_connection(self, dbname):
        connection = psycopg2.connect(self.get_dsn(dbname))
        connection.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        return connection


class LandscapeBootstrapPlugin(Plugin, PasswordGenerationMixin, DatabaseConnectionBase):
    """
    This plugin creates the required users for Landscape if they don't
    already exist.
    """

    log_message = "Bootstrapping from service.conf file ..."

    def get_existing_users(self, connection):
        """Query Postgresql for the list of users that it has currently.

        @return a list of strings of usernames that are already in
                Postgresql.
        """
        cursor = connection.cursor()
        cursor.execute("SELECT usename FROM pg_user WHERE usename " "LIKE 'landscape%'")
        return [x[0] for x in cursor.fetchall()]

    def main(self, connection=None):
        """Entry point for all Plugins."""
        super(LandscapeBootstrapPlugin, self).main()
        connection = connection or self.get_connection("postgres")
        existing_users = self.get_existing_users(connection)
        cursor = connection.cursor()
        missing_users = [
            (user, key)
            for user, key in [("landscape", "password")]
            if user not in existing_users
        ]
        new_configuration = {}
        for user, key in missing_users:
            try:
                password = self.generate_password()
                cursor.execute(
                    "CREATE USER %s WITH PASSWORD '%s' NOCREATEDB "
                    "NOSUPERUSER " % (user, password)
                )
                new_configuration[key] = password_encode(password)
            except psycopg2.ProgrammingError as e:
                self.log(INFO, "Error creating user '%s'." % user)
                message = "Encountered database error:\n\n" + str(e)
                self.log(ERROR, message)
            else:
                self.log(INFO, "Created user '%s'." % user)
        connection.commit()
        self.write_service_configuration(new_configuration)


class ServerConfToServiceConfPlugin(Plugin, ServiceConfigurationMixin):
    """
    This Plugin migrates an existing /etc/landscape/server.conf file to
    /etc/landscape/service.conf
    """

    server_configuration_file = "/etc/landscape/server.conf"

    log_message = "Setting up database configuration ..."

    def has_existing_configuration(self):
        """
        Returns True if we have a pre-existing configuration to migrate.
        """
        return os.path.exists(self.server_configuration_file)

    def get_existing_configuration(self, config):
        """
        Parse the old server.conf format, and return the values in a
        dictionary.

        @param config: RawConfigParser that has read a server.conf file.

        The keys in the dictionary are:
            host: The host to connect to.
            password: The password for the landscape user.
            superuser: The landscape schema user.
            superuser_password: The password for the Landscape schema user.
        """
        return {
            "host": config.get("database", "host"),
            "password": password_encode(config.get("database", "password")),
            "superuser": config.get("database", "superuser"),
            "superuser_password": password_encode(
                config.get("database", "superuser-password")
            ),
        }

    def main(self):
        """
        Check to see if there is an existing configuration file, if not, skip
        the migration, otherwise, get the existing configuration from the
        server configuration file, and write it to the new service.conf file.

        Rename the old server.conf file to server.conf.old.
        """
        if not self.has_existing_configuration():
            self.log(INFO, "Skipping configuration migration ...")
        else:
            config = RawConfigParser()
            config.read([self.server_configuration_file])
            if not config.has_section("database"):
                self.log(INFO, "Skipping configuration migration ...")
            else:
                self.log(INFO, "Migrating configuration from VM setup ...")
                existing_configuration = self.get_existing_configuration(config)
                self.write_service_configuration(existing_configuration)
                os.rename(
                    self.server_configuration_file,
                    self.server_configuration_file + ".old",
                )


class DatabaseCreationPlugin(Plugin, DatabaseConnectionBase):
    """This plugin creates the Landscape databases if they are not created."""

    log_message = "Checking Landscape databases ..."

    def get_existing_databases(self, connection):
        """
        Query Postgresql for the list of databases that it has currently.

        @return a list of strings of database names that are already in
                Postgresql.
        """
        cursor = connection.cursor()
        cursor.execute(
            "SELECT datname, datctype FROM pg_database "
            "WHERE datname LIKE 'landscape%'"
        )
        names = []
        for name, ctype in cursor.fetchall():
            names.append(name)
            if not ctype.endswith("UTF-8"):
                self.log(
                    WARN,
                    "Database %(name)r encoding is not UTF-8: %(ctype)r",
                    {"name": name, "ctype": ctype},
                )
        return names

    def main(self, connection=None):
        """Entry point for Plugin."""
        super(DatabaseCreationPlugin, self).main()
        connection = connection or self.get_connection("postgres")
        existing_databases = self.get_existing_databases(connection)
        cursor = connection.cursor()
        for database in [
            database
            for database in STANDALONE_DATABASES
            if database not in existing_databases
        ]:
            try:
                # Use template0 to have the cleanest new database, and to
                # make sure we get an UTF-8 database regardless of the
                # environment on cluster creation.
                cursor.execute(
                    'CREATE DATABASE "%s" OWNER postgres'
                    ' TEMPLATE "template0" ENCODING UTF8'
                    ' LC_CTYPE="C.UTF-8" LC_COLLATE="C.UTF-8"' % database
                )
            except psycopg2.ProgrammingError as e:
                self.log(INFO, "Error creating database '%s'." % database)
                message = "Encountered database error:\n\n" + str(e)
                self.log(ERROR, message)
            else:
                self.log(INFO, "Created database '%s'." % database)
        connection.commit()


class PgRestoreError(Exception):
    """Raised upon execution failure of the pg_restore program."""


class DatabasePackageLoadPlugin(Plugin, DatabaseConnectionBase):
    """This plugin loads the stock package data if the package db is empty."""

    log_message = "Checking package database initial data ..."

    pg_restore_binary = "/usr/bin/pg_restore"
    package_database = "landscape-standalone-package"
    package_stock_data_filename = PACKAGE_STOCK_DATA_FILENAME

    def get_tables(self, connection, exclude=[]):
        """
        Query Postgresql for the list of tables of the connected database.

        @param exclude: optionally a list of table names to exclude.
        @return a list of strings of table names that are already in the
            connected database.
        """
        cursor = connection.cursor()
        cursor.execute(
            """
            SELECT c.relname
            FROM pg_catalog.pg_class c
                LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE c.relkind IN ('r','')
                AND n.nspname NOT IN ('pg_catalog', 'pg_toast')
                AND pg_catalog.pg_table_is_visible(c.oid)
            """
        )
        tables = [x[0] for x in cursor.fetchall()]
        if exclude:
            tables = list(set(tables).difference(set(exclude)))
        return tables

    def is_empty_table(self, connection, table):
        """Return C{True} if C{table} has no row, C{False} otherwise."""

        cursor = connection.cursor()
        cursor.execute("SELECT * FROM %s LIMIT 1" % table)
        return cursor.fetchone() is None

    def run_pg_restore(self):
        """Load the stock package data to the package database."""
        env = os.environ.copy()
        configuration = self.read_service_configuration()
        env["PGPASSWORD"] = password_decode(configuration.get("superuser_password"))
        args = [
            self.pg_restore_binary,
            "--disable-triggers",
            "--no-owner",
            "--data-only",
            "--host=%s" % configuration.get("host"),
            "--username=%s" % configuration.get("superuser"),
            "--dbname=%s" % self.package_database,
            self.package_stock_data_filename,
        ]
        pipe = subprocess.Popen(args, env=env)
        exit_code = os.waitpid(pipe.pid, 0)[1]
        if exit_code != 0:
            raise PgRestoreError(" ".join(args))

    def has_already_package_data(self, connection):
        """Indicate if the package database already contains data."""
        # On fresh deployments the patch table is the only non-empty one,
        # as it got populated by the patch plugin.
        tables = self.get_tables(connection, exclude=["patch"])
        for table in tables:
            if not self.is_empty_table(connection, table):
                return True
        return False

    def main(self):
        super(DatabasePackageLoadPlugin, self).main()
        connection = self.get_connection(self.package_database)
        if self.has_already_package_data(connection):
            self.log(INFO, "Package database already initialized.")
            return
        if not os.path.exists(self.package_stock_data_filename):
            self.log(INFO, "No package data to load.")
            return
        try:
            self.log(INFO, "Loading stock package database ...")
            self.run_pg_restore()
        except PgRestoreError as error:
            self.log(INFO, "Error loading stock package database.")
            message = "Load command failed:\n\n" + str(error)
            self.log(ERROR, message)
        else:
            self.log(INFO, "Stock package database loaded successfully.")
            os.remove(self.package_stock_data_filename)


class DatabasePatchPlugin(Plugin):
    """This plugin patches the Landscape databases."""

    log_message = "Checking database schema ..."

    def set_environment(self):
        """
        This method sets the LANDSCAPE_CONFIG environment variable to the value
        needed for running the schema patching code. It returns a tuple of the
        original value of the environment variable and a boolean that
        represents whether or not the environment variable should be deleted
        when done. The latter is determined by checking if the variable existed
        in the environment originally or not.
        """
        do_delete = False
        if LANDSCAPE_CONFIG not in os.environ:
            do_delete = True
        original_value = os.getenv(LANDSCAPE_CONFIG)
        os.environ[LANDSCAPE_CONFIG] = STANDALONE_SCHEMA
        return (original_value, do_delete)

    def unset_environment(self, original_value, do_delete):
        """
        This method does one of two things: returns the LANDSCAPE_CONFIG
        environment variable back to the provided original value, or it deletes
        it entirely if the do_delete boolean parameter is set to True.
        """
        if do_delete:
            del os.environ[LANDSCAPE_CONFIG]
        else:
            os.environ[LANDSCAPE_CONFIG] = original_value

    def update_schema(self):
        """
        Calls down to the underlying Landscape schema tool, to setup the
        Schema.
        """
        args = []
        capture = StringIO()
        capture.write("Schema configuration output:\n\n")
        old_stdout = sys.stdout
        sys.stdout = capture
        try:
            schema.run(args=args)
        except Exception:
            self.log(ERROR, "Encountered database error", exc_info=True)
        finally:
            self.log(INFO, capture.getvalue())
            sys.stdout = old_stdout

    def main(self):
        """Entry point for Plugin."""
        super(DatabasePatchPlugin, self).main()
        original_value, do_delete = self.set_environment()
        self.update_schema()
        self.unset_environment(original_value, do_delete)


class HashIdStoresPlugin(Plugin, DatabaseConnectionBase):
    """This plugin renames the stock hash-id stores against the server UUID."""

    log_message = "Renaming stock hash-id stores ..."
    stores_path = HASHID_STORES_PATH
    package_stock_data_filename = PACKAGE_STOCK_DATA_FILENAME

    def rename_stock_store_filenames(self, server_uuid):
        """
        Rename the stock hash-id store files to match the given C{server_uuid}.
        """
        stock_store_basenames = []
        if not os.path.exists(self.stores_path):
            return 0
        for basename in os.listdir(self.stores_path):
            if basename.startswith("uuid_"):
                stock_store_basenames.append(basename)
        for stock_store_basename in stock_store_basenames:
            new_store_basename = stock_store_basename.replace(
                "uuid_", "%s_" % server_uuid
            )
            os.rename(
                os.path.join(self.stores_path, stock_store_basename),
                os.path.join(self.stores_path, new_store_basename),
            )
        return len(stock_store_basenames)

    def main(self):
        """
        Ensure the server UUID exists and possibly rename stock hash-id stores.
        """
        super(HashIdStoresPlugin, self).main()
        if os.path.exists(self.package_stock_data_filename):
            self.log(
                INFO,
                "Stock package database not loaded, " "ignoring stock hash-id stores.",
            )
            return
        connection = self.get_connection("landscape-standalone-main")
        ensure_server_uuid_exists(connection=connection)
        server_uuid = str(get_server_uuid(connection=connection))
        connection.close()
        result = self.rename_stock_store_filenames(server_uuid)
        if result > 0:
            self.log(
                INFO,
                "Renamed %d stock hash-id stores against uuid %s"
                % (result, server_uuid),
            )
        else:
            self.log(INFO, "No stock file to rename")
