# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import subprocess

from logging import INFO, WARN

from landscape.client.manager.scriptexecution import get_user_info

from canonical.lib.text import password_encode

from canonical.landscape.scripts.standalone.plugins.base import Plugin
from canonical.landscape.scripts.standalone.plugins.database import (
    PasswordGenerationMixin,
    STANDALONE_DATABASES,
    ServiceConfigurationMixin,
)


LANDSCAPE_CONFIG_HEADER = "# Inserted by Landscape PostgreSQL configuration\n"
LANDSCAPE_CONFIG_LINE = "local\tall\tlandscape\tmd5\n"


class PsqlPlugin(Plugin):
    """Base class for talking direct to the psql binary."""

    def _communicate_with_psql(self, query):
        """
        Open a connection as the postgres user and execute query, returning the
        result for further processing.
        """
        process = subprocess.run(
            [
                "sudo",
                "-u",
                "postgres",
                "-s",
                "/usr/bin/psql",
                "--tuples-only",
                "--no-align",
                "--field-separator=','",
            ],
            input=query,
            capture_output=True,
            text=True,
        )
        output = process.stdout
        errors = process.stderr
        return output.strip()

    def get_existing_users(self, prefix):
        """Query Postgresql for the list of users that it has currently.

        @return a Set of strings of usernames that are already in
                Postgresql.
        """
        output = self._communicate_with_psql(
            "SELECT usename FROM pg_user WHERE usename LIKE '%s%%'" % prefix
        )
        return set([username for username in output.split("\n") if username])


class LocalLandscapeBootstrapPlugin(
    PsqlPlugin, PasswordGenerationMixin, ServiceConfigurationMixin
):
    """
    This plugin creates the required users for Landscape if they don't
    already exist.
    """

    log_message = "Bootstrapping service.conf using psql ..."

    def main(self):
        """Entry point for all Plugins."""
        super(LocalLandscapeBootstrapPlugin, self).main()
        service_conf_data = {}
        existing_users = self.get_existing_users("landscape")
        required_users = set(["landscape_superuser", "landscape"])
        for user in sorted(required_users.difference(existing_users)):
            password = self.generate_password()
            hidden_password = password_encode(password)
            if user == "landscape_superuser":
                self._communicate_with_psql(
                    "CREATE ROLE landscape_superuser"
                    " PASSWORD '%s'"
                    " SUPERUSER CREATEDB CREATEROLE INHERIT LOGIN" % password
                )
                service_conf_data["superuser"] = "landscape_superuser"
                service_conf_data["superuser_password"] = hidden_password
            else:
                self._communicate_with_psql(
                    "CREATE ROLE %s"
                    " PASSWORD '%s'"
                    " NOSUPERUSER NOCREATEDB NOCREATEROLE INHERIT LOGIN"
                    % (user, password)
                )
                if user == "landscape":
                    service_conf_data["password"] = hidden_password
            self.log(INFO, "Created user '%s'." % user)
        service_conf_data["host"] = "localhost"
        self.write_service_configuration(service_conf_data)


class LocalDatabaseCreationPlugin(PsqlPlugin):
    """This plugin creates the Landscape databases if they are not created."""

    log_message = "Checking Landscape databases ..."

    def get_existing_databases(self):
        """
        Query Postgresql for the list of databases that it has currently.

        @return a set of strings of database names that are already in
                Postgresql.
        """
        output = self._communicate_with_psql(
            "SELECT datname, datctype FROM pg_database"
            " WHERE datname LIKE 'landscape%'"
        )

        names = []
        for row in output.split("\n"):
            if not row:
                continue
            name, ctype = row.split("','")
            names.append(name)
            if not ctype.endswith("UTF-8"):
                self.log(
                    WARN,
                    "Database '%(name)s' encoding is not UTF-8: '%(ctype)s'",
                    {"name": name, "ctype": ctype},
                )
        return set(names)

    def main(self, connection=None):
        """Entry point for Plugin."""
        super(LocalDatabaseCreationPlugin, self).main()
        existing_databases = self.get_existing_databases()

        missing_databases = sorted(
            set(STANDALONE_DATABASES).difference(existing_databases)
        )
        psql_query = ""
        for database in missing_databases:
            psql_query += (
                'CREATE DATABASE "%s" OWNER postgres'
                ' TEMPLATE "template0" ENCODING UTF8'
                ' LC_CTYPE="C.UTF-8" LC_COLLATE="C.UTF-8";\n' % database
            )
        if psql_query:
            self._communicate_with_psql(psql_query)


class PostgreSQLPlugin(Plugin):
    """
    Plugin for configuring PostgreSQL according to the LDS Quickstart
    instructions.
    """

    log_message = "Configuring PostgreSQL for Landscape"

    psql12_config = "/etc/postgresql/12/main/pg_hba.conf"
    psql13_config = "/etc/postgresql/13/main/pg_hba.conf"
    psql14_config = "/etc/postgresql/14/main/pg_hba.conf"
    psql_init_script = "/etc/init.d/postgresql"

    def main(self, connection=None):
        """Entry point for all Plugins."""
        super(PostgreSQLPlugin, self).main()
        if not self.has_landscape_configuration():
            self.write_landscape_configuration()
            self.reload_postgresql()

    def reload_postgresql(self):
        """
        Run the command-line to ask PostgreSQL to reload its
        configuration.
        """
        command_line = [self.get_init_script(), "reload"]
        process = subprocess.run(command_line, capture_output=True, text=True)
        output = process.stdout
        errors = process.stderr
        return output

    def has_landscape_configuration(self):
        """
        Return True if this pg_hba.conf file already has already been
        configured for Landscape.
        """
        with open(self.get_config()) as f:
            for line in f.readlines():
                if not line.startswith("#"):
                    components = line.split()
                    if len(components) < 3:
                        continue
                    conn_type, db, users = components[0:3]
                    users = users.split(",")
                    if conn_type == "local" and db == "all" and "landscape" in users:
                        return True
        return False

    def write_landscape_configuration(self):
        """
        This inserts the Landscape configuration into the correct place in the
        PostgreSQL configuration file.
        """
        hba_config = self.get_config()
        temporary_filename = hba_config + ".new"
        new_file = open(temporary_filename, "w")
        landscape_config_written = False
        with open(hba_config) as f:
            for line in f.readlines():
                # Comment lines, we can just write out...
                if line.startswith("#"):
                    new_file.write(line)
                else:
                    # Identify this line, is it the local/all/postgres/ident
                    # line?
                    components = line.split()
                    if components[0:4] == ["local", "all", "postgres", "ident"]:
                        new_file.write(line)
                        new_file.write(LANDSCAPE_CONFIG_HEADER)
                        new_file.write(LANDSCAPE_CONFIG_LINE)
                        landscape_config_written = True
                    else:
                        new_file.write(line)
        # If for some unknown reason, we didn't find the line we were looking
        # for, let's write out the Landscape configuration anyway.
        if not landscape_config_written:
            new_file.write(LANDSCAPE_CONFIG_HEADER)
            new_file.write(LANDSCAPE_CONFIG_LINE)
        new_file.close()
        # Rename the existing file to .old, and the new file to the existing
        # filename
        os.rename(hba_config, hba_config + ".old")
        os.rename(temporary_filename, hba_config)
        os.chmod(hba_config, 0o640)
        # We should check for the existence of the postgres user, but given
        # that we have a dependency on postgresql, this should never happen.
        uid, gid, home_dir = get_user_info("postgres")
        os.chown(hba_config, 0, gid)

    def get_init_script(self):
        """Get the psql init script if it exists."""
        if os.path.exists(self.psql_init_script):
            return self.psql_init_script
        else:
            raise RuntimeError("Postgresql's initscript not found.")

    def get_config(self):
        """Get the psql config file, favor 14 if conflicts found."""
        if os.path.exists(self.psql14_config):
            return self.psql14_config
        elif os.path.exists(self.psql13_config):
            return self.psql13_config
        elif os.path.exists(self.psql12_config):
            return self.psql12_config
        else:
            raise RuntimeError("Postgresql's pg_hba.conf not found.")
