# Copyright 2005-2022 Canonical Limited. All rights reserved.

import subprocess
import os
import re
from configparser import RawConfigParser
from logging import INFO

from zope.component import queryUtility, provideUtility

from twisted.internet import reactor
from twisted.internet.threads import blockingCallFromThread

from canonical.lib.text import password_encode, password_decode

from canonical.landscape.setup import run_amqp_client_in_thread
from canonical.landscape.scripts.standalone.plugins.base import (
    Plugin,
    write_service_configuration,
)
from canonical.landscape.scripts.standalone.base import SetupScriptError
from canonical.landscape.scripts.standalone.plugins.database import (
    PasswordGenerationMixin,
)
from canonical.amqp.message.dispatcher import global_dispatcher
from canonical.amqp.message.publisher import global_publisher
from canonical.amqp.message.interfaces import IDispatcher, IPublisher


def run_rabbitmqctl(option, *args):
    """
    Runs rabbitmqctl and processes the stdout, validates that we get the
    expected output, and returns output, without the header or footer.
    """
    command_line = ["/usr/sbin/rabbitmqctl", "-q", option, *args]
    environment = os.environ.copy()
    environment["LC_ALL"] = "C"
    process = subprocess.run(
        command_line, capture_output=True, env=environment, text=True
    )
    output = process.stdout
    errors = process.stderr

    if errors:
        raise SetupScriptError("Error setting up rabbitmqctl: {}".format(errors))

    return output.strip().split("\n")


class RabbitMQPluginBase(Plugin):
    """Base Functionality for getting information from RabbitMQ."""

    def get_existing_users(self):
        """
        Spawn rabbitmqctl and parse the output, to get the list of existing
        users.
        """
        # list_users has a header\nuser1\nuser2\nfooter\n layout.
        # the header is "Listing users ..."
        output = run_rabbitmqctl("list_users")
        return sorted(line.split()[0] for line in output)

    def get_existing_vhosts(self):
        """
        Spawn rabbitmqctl and parse the output, to get the list of existing
        vhosts.
        """
        # list_vhosts has a header\nuser1\nuser2\nfooter\n layout.
        # the header is "Listing vhosts ..."
        return sorted(run_rabbitmqctl("list_vhosts"))

    def get_existing_permissions(self, vhost):
        """
        Spawn rabbitmqctl and parse the output, to get the list of existing
        permissions for the named vhost.
        """
        result = run_rabbitmqctl("list_permissions", "-p", vhost)
        permissions = {}
        for line in result:
            if not line:
                continue

            user, configure, read, write = tuple(line.split("\t"))
            permissions[user] = (configure, read, write)
        return permissions


class RabbitMQPlugin(RabbitMQPluginBase, PasswordGenerationMixin):
    """
    Configure RabbitMQ for Landscape on the local machine.
    """

    log_message = "Configuring RabbitMQ for Landscape ..."

    def main(self):
        """Entry point for all Plugins."""
        super(RabbitMQPlugin, self).main()
        vhosts = self.get_existing_vhosts()
        if "landscape" not in vhosts:
            self.create_landscape_vhost()

        users = self.get_existing_users()
        if "landscape" not in users:
            password = self.create_landscape_user()
            self.update_service_configuration(password)

        existing_permissions = self.get_existing_permissions("landscape")
        if not existing_permissions.get("landscape") == ("", ".*", ".*"):
            self.setup_landscape_user_permissions()

    def create_landscape_vhost(self):
        """Creates a landscape vhost in RabbitMQ."""
        run_rabbitmqctl("add_vhost", "landscape")

    def create_landscape_user(self):
        """Creates a landscape user with a generated password in RabbitMQ."""
        password = self.generate_password()
        while password.startswith("-"):
            # Passwords starting with `-` are interpreted by
            # rabbitmqctl as options and it will error.
            password = self.generate_password()

        run_rabbitmqctl("add_user", "landscape", password)
        return password

    def setup_landscape_user_permissions(self):
        """
        Sets up the required permissions for the landscape user in the
        landscape vhost.
        """
        run_rabbitmqctl(
            "set_permissions", "-p", "landscape", "landscape", ".*", ".*", ".*"
        )

    def update_service_configuration(self, password):
        """
        Update service.conf file to store a base64 encoded RabbitMQ password.
        """
        config = {
            "port": 5672,
            "host": "localhost",
            "user": "landscape",
            "vhost": "landscape",
            "password": password_encode(password),
        }
        write_service_configuration(
            {"broker": config}, self.service_configuration_file, "landscape"
        )


class RabbitMQUpgradePlugin(RabbitMQPluginBase, PasswordGenerationMixin):
    """
    Upgrade local RabbitMQ settings.
    """

    log_message = "Checking local RabbitMQ settings."

    def main(self):
        """Entry point for all Plugins."""
        super(RabbitMQUpgradePlugin, self).main()
        current_configuration = RawConfigParser()
        current_configuration.read([self.service_configuration_file])
        if current_configuration.has_section("broker"):
            config = dict(current_configuration.items("broker"))
            password = config.get("password")
            if (password == "landscape") and ("landscape" in self.get_existing_users()):
                password = self.generate_password()
                config = {
                    "port": current_configuration.get("broker", "port"),
                    "host": current_configuration.get("broker", "host"),
                    "vhost": current_configuration.get("broker", "vhost"),
                    "user": current_configuration.get("broker", "user"),
                    "password": password_encode(password),
                }
                self._change_password(password)
                write_service_configuration(
                    {"broker": config}, self.service_configuration_file, "landscape"
                )
                self.log(INFO, "RabbitMQ password updated.")
            self._delete_non_durable_queues(config)

    def _change_password(self, password):
        """Set RabbitMQ's password for user 'landscape' to the given value."""
        run_rabbitmqctl("change_password", "landscape", password)

    def _delete_non_durable_queues(self, config):
        """Delete non-durable RabbitMQ queues, if any."""
        queues = self._get_non_durable_queues()
        if not queues:
            return  # Nothing to do

        dispatcher = queryUtility(IDispatcher)
        if dispatcher is None:
            provideUtility(global_dispatcher, IDispatcher)
            provideUtility(global_publisher, IPublisher)
            dispatcher = global_dispatcher

        config["password"] = password_decode(config["password"])
        stop_amqp_client = run_amqp_client_in_thread({"broker": config})
        channel = dispatcher._publisher._channel
        for queue in queues:
            blockingCallFromThread(reactor, channel.queue_delete, queue=queue)

        stop_amqp_client()
        self.log(INFO, "RabbitMQ non-durable queues updated.")

    def _get_non_durable_queues(self):
        """Get a all queues that were not created with the durable flag."""
        queues = []
        lines = run_rabbitmqctl("list_queues", "-p", "landscape", "name", "durable")
        for line in lines:
            if not line:
                continue

            name, durable = line.split()
            if durable == "false":
                queues.append(name)
        return queues
