# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import shutil
import glob

from logging import INFO
from random import SystemRandom
from string import ascii_letters, digits
from configparser import RawConfigParser

from canonical.landscape.scripts.standalone.plugins.base import (
    Plugin,
    write_service_configuration,
    delete_from_service_configuration,
    chmod_and_chown_file,
)


def generate_secret_token():
    """
    Generate a string with 1024 bits of entropy.

    Using 172 * the 62 possible alphanumerics [a-z][A-Z][0-9] provides:
    math.log(62, 2) * 172 ~= 1024 bits of entropy.
    """
    random_chooser = SystemRandom()
    alphanumerics = ascii_letters[0:52] + digits
    return "".join(random_chooser.choice(alphanumerics) for _ in range(172))


class QuickStartUpgradePlugin(Plugin):
    """
    Upgrade the Landscape service.conf file with missing items.
    """

    log_message = "Upgrading service.conf ..."

    mailer_path = "/var/lib/landscape/landscape-mail-queue"

    def main(self):
        """Entry point for all Plugins."""
        super(QuickStartUpgradePlugin, self).main()
        new_filename = self.service_configuration_file + ".new"
        # Make a copy of the service configuration file.
        if os.path.exists(self.service_configuration_file):
            shutil.copy(self.service_configuration_file, new_filename)
            chmod_and_chown_file(new_filename, "landscape")

            self.update_job_handler_section(new_filename)
            self.update_api_section(new_filename)
            self.update_maintenance_section(new_filename)
            self.update_scripts_section(new_filename)
            self.update_pingserver_section(new_filename)
            self.update_message_server_section(new_filename)
            self.update_landscape_section(new_filename)
            self.update_package_upload_section(new_filename)
            self.update_load_shaper_section(new_filename)
            self.update_package_search_section(new_filename)
            self.delete_statsd_section(new_filename)
            self.delete_clouddeck_section(new_filename)
            self.delete_combo_loader_section(new_filename)
            if os.path.exists(self.service_configuration_file):
                os.rename(
                    self.service_configuration_file,
                    self.service_configuration_file + ".old",
                )
            os.rename(new_filename, self.service_configuration_file)
            self.migrate_maildirs()

    def update_global_section(self, filename):
        """
        Update the global section of service.conf to remove the now unused
        log-path variable.
        """
        current_configuration = RawConfigParser()
        current_configuration.read([filename])
        if current_configuration.has_option("global", "log-path"):
            config = {"log-path": None}

        updated = write_service_configuration(
            {"global": config}, filename, overwrite=True
        )
        if updated:
            self.log(INFO, "Updated global section.")

    def update_job_handler_section(self, filename):
        """
        Upgrade the job-handler section to be able to send out mail.
        """
        config = {
            "mailer": "queue",
            "mailer-path": "/var/lib/landscape/landscape-mail-queue",
        }
        if write_service_configuration(
            {"job-handler": config}, filename, overwrite=True
        ):
            self.log(INFO, "Updated job-handler section.")

    def update_api_section(self, filename):
        """
        Update the api server configuration.
        """
        # FIXME: We should record that we changed the port once, and not do it
        # again on subsequent runs.
        base_port = "9080"
        current_configuration = RawConfigParser()
        current_configuration.read([filename])
        if current_configuration.has_option("api", "base-port"):
            current_port = current_configuration.get("api", "base-port")
            if current_port != "9091":
                base_port = current_port
        config = {
            "mailer": "queue",
            "mailer-path": "/var/lib/landscape/landscape-mail-queue",
        }
        updated = write_service_configuration({"api": config}, filename, overwrite=True)
        updated |= write_service_configuration(
            {"api": {"base-port": base_port}}, filename, overwrite=True
        )
        if updated:
            self.log(INFO, "Updated api section.")

    def update_maintenance_section(self, filename):
        """
        Update the mailer-path in the maintenance section.
        """
        config = {
            "mailer-path": "/var/lib/landscape/landscape-mail-queue",
            "store_user": None,
            "store_password": None,
        }
        if write_service_configuration(
            {"maintenance": config}, filename, overwrite=True
        ):
            self.log(INFO, "Updated maintenance section.")

    def update_scripts_section(self, filename):
        """
        Update the mailer-path in the scripts section.
        """
        config = {"mailer-path": "/var/lib/landscape/landscape-mail-queue"}
        if write_service_configuration({"scripts": config}, filename, overwrite=True):
            self.log(INFO, "Updated scripts section.")

    def update_pingserver_section(self, filename):
        """
        Update the pingserver configuration.
        """
        base_port = "8070"
        current_configuration = RawConfigParser()
        current_configuration.read([filename])
        if current_configuration.has_option("pingserver", "base-port"):
            current_port = current_configuration.get("pingserver", "base-port")
            if current_port != "8082":
                base_port = current_port
        config = {"base-port": base_port}
        if write_service_configuration(
            {"pingserver": config}, filename, overwrite=True
        ):
            self.log(INFO, "Updated pingserver section.")

    def update_message_server_section(self, filename):
        """
        Update the message-server configuration.
        """
        base_port = "8090"
        current_configuration = RawConfigParser()
        current_configuration.read([filename])
        if current_configuration.has_option("message-server", "base-port"):
            current_port = current_configuration.get("message-server", "base-port")
            if current_port != "8081":
                base_port = current_port
        config = {
            "base-port": base_port,
            "oops-key": "DM%(instance-letter)s",
            "access-log": "message-server_access-%(instance)s.log",
        }
        if write_service_configuration(
            {"message-server": config}, filename, overwrite=True
        ):
            self.log(INFO, "Updated message-server section.")

    def update_landscape_section(self, filename):
        """
        Update the landscape configuration.
        """
        config = {
            "access-log": "appserver_access-%(instance)s.log",
            "oops-key": "DF%(instance-letter)s",
            "mailer-path": "/var/lib/landscape/landscape-mail-queue",
            "reprepro-binary": ("/opt/canonical/landscape/scripts/reprepro-wrapper.sh"),
        }

        config_no_overwrite = {
            "repository-path": "/var/lib/landscape/landscape-repository",
            "secret-token": generate_secret_token(),
        }

        if any(
            (
                write_service_configuration(
                    {"landscape": config}, filename, overwrite=True
                ),
                write_service_configuration(
                    {"landscape": config_no_overwrite}, filename, overwrite=False
                ),
            )
        ):
            self.log(INFO, "Updated landscape section.")

    def update_package_upload_section(self, filename):
        """
        Update the package-upload service configuration.
        """
        config = {
            "stores": "main account-1",
            "threads": 10,
            "port": 9100,
            "root-url": "http://localhost:9100",
            "mailer": "queue",
        }

        if write_service_configuration(
            {"package-upload": config}, filename, overwrite=False
        ):
            self.log(INFO, "Updated package-upload section.")

        config = {"mailer-path": "/var/lib/landscape/landscape-mail-queue"}
        if write_service_configuration(
            {"package-upload": config}, filename, overwrite=True
        ):
            self.log(INFO, "Updated package-upload mail directory.")

    def update_load_shaper_section(self, filename):
        """
        Update the load-shaper service configuration.
        """
        config = {"critical-load": 10.0, "good-load": 3.0, "good-duration": 60.0}

        if write_service_configuration(
            {"load-shaper": config}, filename, overwrite=False
        ):
            self.log(INFO, "Updated load-shaper section.")

    def delete_statsd_section(self, filename):
        """
        Remove the statsd section.
        """
        if delete_from_service_configuration(["statsd", "db-stats"], filename):
            self.log(INFO, "Removed statsd section.")

    def delete_clouddeck_section(self, filename):
        """
        Remove the clouddeck section.
        """
        if delete_from_service_configuration(["clouddeck"], filename):
            self.log(INFO, "Removed clouddeck section.")

    def delete_combo_loader_section(self, filename):
        """
        Remove the combo loader section.
        """
        if delete_from_service_configuration(["combo-loader"], filename):
            self.log(INFO, "Removed combo-loader section.")

    def update_package_search_section(self, filename):
        """
        Update the package-search service configuration.
        """
        config = {
            "port": "9099",
            "stores": "main package resource-1",
            "account-threshold": "0",
            "pid-path": "/var/run/landscape/landscape-package-search-1.pid",
        }

        if write_service_configuration(
            {"package-search": config}, filename, overwrite=False
        ):
            # There was no package-search section at all
            self.log(INFO, "Updated package-search section.")
        else:
            # There's a package-search section, let's see if need to update
            # the threshold.
            file_config = RawConfigParser()
            file_config.read([filename])
            if file_config.has_option("package-search", "account-threshold"):
                threshold = file_config.get("package-search", "account-threshold")
                if threshold == "800":
                    # The threshold hasn't been changed from the previous
                    # default, update it to always use package-search.
                    config = {"account-threshold": "0"}
                    write_service_configuration(
                        {"package-search": config}, filename, overwrite=True
                    )
                    self.log(INFO, "Updated package-search threshold.")

    def migrate_maildirs(self):
        """
        Migrate multiple maildirs to a single one.
        """
        legacy_mailer_paths = glob.glob("%s-*" % self.mailer_path)

        if not legacy_mailer_paths:
            # Nothing to migrate
            return

        subdirs = ["cur", "new", "tmp"]

        if not os.path.exists(self.mailer_path):
            os.rename(legacy_mailer_paths.pop(), self.mailer_path)

        for legacy_mailer_path in legacy_mailer_paths:
            for subdir in subdirs:
                legacy_subdir = os.path.join(legacy_mailer_path, subdir)
                current_subdir = os.path.join(self.mailer_path, subdir)
                messages = [
                    filename
                    for filename in os.listdir(legacy_subdir)
                    if not filename.startswith(".")
                ]
                for message in messages:
                    os.rename(
                        os.path.join(legacy_subdir, message),
                        os.path.join(current_subdir, message),
                    )
            shutil.rmtree(legacy_mailer_path)
