# Copyright 2005-2022 Canonical Limited. All rights reserved.

import sys
import logging

from canonical.landscape.scripts.standalone.base import (
    StandaloneScript,
    SetupScriptError,
)
from canonical.landscape.scripts.standalone.plugins.database import (
    ServerConfToServiceConfPlugin,
    DatabasePatchPlugin,
    DatabasePackageLoadPlugin,
    HashIdStoresPlugin,
)
from canonical.landscape.scripts.standalone.plugins.local_database import (
    LocalDatabaseCreationPlugin,
    LocalLandscapeBootstrapPlugin,
    PostgreSQLPlugin,
)
from canonical.landscape.scripts.standalone.plugins.rabbitmq import (
    RabbitMQPlugin,
    RabbitMQUpgradePlugin,
)
from canonical.landscape.scripts.standalone.plugins.apache import ApacheSSLPlugin
from canonical.landscape.scripts.standalone.plugins.upgrades import (
    QuickStartUpgradePlugin,
)
from canonical.landscape.scripts import get_top_level_directory
from canonical.landscape.scripts.standalone.helpers import has_root_privileges
from canonical.landscape.application import setup_logging


class StandaloneQuickstartScript(StandaloneScript):
    """
    This is a StandaloneScript subclass that runs and does everything necessary
    to create and setup Landscape databases and schemas, plus generating
    configuration files by instantiating templates with user-provided values.
    """

    MESSAGE_PREFIX = "Landscape server setup - "

    def __init__(self, top_level_directory="", logger=None):
        super(StandaloneQuickstartScript, self).__init__(
            logger,
            plugins=[
                # Patch service.conf file if necessary.
                QuickStartUpgradePlugin,
                # Configure RabbitMQ.
                RabbitMQUpgradePlugin,
                RabbitMQPlugin,
                # Configure Apache for SSL.
                ApacheSSLPlugin,
                # Update the pg_hba.conf file.
                PostgreSQLPlugin,
                # Configuration migration.
                ServerConfToServiceConfPlugin,
                # Bootstrap the landscape superuser and landscape user.
                LocalLandscapeBootstrapPlugin,
                # Database creation.
                LocalDatabaseCreationPlugin,
                # Database patching.
                DatabasePatchPlugin,
                # Package database dump loading.
                DatabasePackageLoadPlugin,
                # Hash-id stores renaming.
                HashIdStoresPlugin,
            ],
        )

        self.top_level_directory = top_level_directory


def main(top_level_directory="", logger=None):
    """
    This function instantiates and runs the setup class. It is called from a
    script, which is in turn called remotely (via ssh login) from the virtual
    appliance host machine.
    """
    script = StandaloneQuickstartScript(
        top_level_directory=top_level_directory, logger=logger
    )
    try:
        script.run()
    except SetupScriptError:
        # The error has already been logged at this point, so we just want to
        # silently exit the script.
        sys.exit(1)


def run():
    if not has_root_privileges():
        logging.error("%s must run with root privileges" % sys.argv[0])
        raise SystemExit(-1)
    setup_logging("landscape-quickstart", always_log_to_stdout=True)
    logger = logging.getLogger("landscape-quickstart")
    main(get_top_level_directory(), logger=logger)
