# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import sys
import logging

from canonical.landscape.scripts.standalone.base import (
    StandaloneScript,
    SetupScriptError,
)
from canonical.landscape.scripts.standalone.plugins.database import (
    ServerConfToServiceConfPlugin,
    DatabaseCreationPlugin,
    DatabasePatchPlugin,
    DatabasePackageLoadPlugin,
    HashIdStoresPlugin,
    LandscapeBootstrapPlugin,
)
from canonical.landscape.scripts import get_top_level_directory
from canonical.landscape.scripts.standalone.helpers import has_root_privileges
from canonical.landscape.application import setup_logging


class StandaloneSetupScript(StandaloneScript):
    """
    This is a StandaloneScript subclass that runs and does everything necessary
    to create and setup Landscape databases and schemas, plus generating
    configuration files by instantiating templates with user-provided values.
    """

    MESSAGE_PREFIX = "Landscape server setup - "

    def __init__(self, top_level_directory="", logger=None):
        super(StandaloneSetupScript, self).__init__(
            logger,
            plugins=[
                # Database configuration.
                LandscapeBootstrapPlugin,
                ServerConfToServiceConfPlugin,
                # Database creation.
                DatabaseCreationPlugin,
                # Database patching.
                DatabasePatchPlugin,
                # Package database dump loading.
                DatabasePackageLoadPlugin,
                # Hash-id stores renaming.
                HashIdStoresPlugin,
            ],
        )

        self.top_level_directory = top_level_directory


def main(top_level_directory="", logger=None):
    """
    This function instantiates and runs the setup class. It is called from a
    script, which is in turn called remotely (via ssh login) from the virtual
    appliance host machine.
    """
    script = StandaloneSetupScript(
        top_level_directory=top_level_directory, logger=logger
    )
    try:
        script.run()
    except SetupScriptError:
        # The error has already been logged at this point, so we just want to
        # silently exit the script.
        sys.exit(1)


def run():
    if not has_root_privileges():
        logging.error("%s must run with root privileges" % sys.argv[0])
        raise SystemExit(-1)
    setup_logging("landscape-setup", always_log_to_stdout=True)
    logger = logging.getLogger("landscape-setup")
    main(get_top_level_directory(), logger)
