# Copyright 2005-2022 Canonical Limited. All rights reserved.

import argparse
from datetime import datetime
import logging
import sys

import transaction

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.main.account import Account, get_account_by_id
from canonical.landscape.model.account.collection import AccountCollection
from canonical.landscape.model.billing.management import BillingManagement
from canonical.landscape.model.billing.alerts import FailedBillingAlert
from canonical.landscape.model.configuration import get_stripe_secret_key
from canonical.landscape.stripe.payment import MAX_ATTEMPTS, try_charge_payment
from canonical.landscape.scripts.batch import BatchScript


def _isatty(stdout):
    try:
        isatty = stdout.isatty
    except AttributeError:
        return False
    return isatty() if callable(isatty) else bool(isatty)


class StripeChargeAccounts(BatchScript):
    """Charge self-service accounts through stripe."""

    _name = "stripe-charge-accounts"
    _setup_amqp = True

    @classmethod
    def _parse_args(cls, argv=sys.argv):
        parser = argparse.ArgumentParser()
        parser.add_argument("--force", action="store_true")
        parser.add_argument("account_ids", nargs="*")
        args = parser.parse_args(argv[1:])
        return args

    def __init__(self, *args, **kwargs):
        BatchScript.__init__(self, *args, **kwargs)
        if _isatty(self._stdout):
            self._log_level = logging.DEBUG

    def run(self):
        args = self._parse_args(self._args)
        self._account_ids = args.account_ids
        self._force = args.force

        stripe_key = get_stripe_secret_key()
        for account_id in self._get_account_ids():
            try:
                self._charge_account(stripe_key, account_id)
            except Exception:
                logging.exception(
                    "Charging failed for account ID {}".format(account_id)
                )
            else:
                logging.info("Charging completed for account ID {}".format(account_id))

    def _get_account_ids(self):
        """Return ids for self-service accounts."""
        if self._account_ids:
            return [int(acctid) for acctid in self._account_ids]
        collection = AccountCollection(get_main_store(), self_service=True)
        return collection.find().values(Account.id)

    def _charge_account(self, stripe_key, account_id):
        """Create Stripe charges for outstanding payments in an account."""
        with transaction.manager:
            account = get_account_by_id(account_id)
            any_success = False
            for payment in self._iter_payments(account):
                try_charge_payment(stripe_key, payment)
                if payment.status == "paid":
                    any_success = True

                # Commit between charges, so that the charge status
                # gets persisted if exception occurs in other payments.
                # We don't want to charge the same charge because of a
                # rollback.
                transaction.commit()

            if any_success:
                # reset_blocked payments on any successful payment
                BillingManagement(account).reset_final_failed_payments()

            # Set or clear an alert on the account.
            account.check_rules([FailedBillingAlert])

    def _iter_payments(self, account):
        today = datetime.utcnow().date()
        management = BillingManagement(account)
        for payment in management.get_outstanding_payments():
            if not self._force:
                if payment.num_attempts >= MAX_ATTEMPTS:
                    logging.debug(
                        ("not trying payment {} (hit max attempts)").format(payment.id)
                    )
                    continue
                # Do not retry on the same day.
                if payment.last_failure is not None:
                    if payment.last_failure.date() >= today:
                        logging.debug(
                            ("not trying payment {} (already tried today)").format(
                                payment.id
                            )
                        )
                        continue
            yield payment


run = StripeChargeAccounts()
