# Copyright 2005-2022 Canonical Limited. All rights reserved.
"""Unblock repository activities that are stuck after killing reprepro."""

import os
import sys
import logging
import subprocess
from argparse import ArgumentParser

from canonical.landscape.scripts.batch import BatchScript
from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.model.account.store import get_account_store_by_name
from canonical.landscape.model.repository.activities import SyncPocketRequest
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.repository.distribution import (
    Distribution,
    STATUS_UNLOCKED,
)
from canonical.landscape.model.activity.status import INPROGRESS


class UnblockRepoActivities(BatchScript):
    """Unblock repository activities that are stuck after killing reprepro."""

    _name = "unblock-repo-activities"
    _config = "scripts"
    _reprepro = "reprepro"

    def run(self):
        parser = ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "--account",
            action="store",
            default="standalone",
            help="The account against which activities are stuck "
            "(default: %(default)s)",
        )
        args = parser.parse_args(self._args[1:])

        try:
            self._reprepro_not_running()
            self._unblock_stores(self._get_stores(args.account))
        except (RuntimeError, ValueError) as exc:
            logging.error(str(exc))
            return 1
        else:
            return 0

    def _reprepro_not_running(self):
        """
        Ensure reprepro is not currently running; this is crudely done with
        pgrep against the specified *reprepro* binary.
        """
        result = subprocess.call(
            ["pgrep", self._reprepro],
            stdout=open(os.devnull, "wb"),
            stderr=subprocess.STDOUT,
        )
        if result == 0:
            raise RuntimeError("Existing %s process found; aborting" % self._reprepro)
        elif result != 1:
            raise RuntimeError("Error running pgrep")

    def _get_stores(self, account):
        """
        Return a mapping of names to account stores that need entries in their
        distribution tables unlocking.
        """
        accounts = [get_account_by_name(account)]
        if accounts[0] is None:
            raise ValueError("Account %s not found" % account)

        # Get the (unique) set of all account stores we need to deal with
        stores = {
            name: get_account_store_by_name(name)
            for name in {
                account.get_store_name()
                for account in accounts
                if account.database_key is not None
            }
        }
        return stores

    def _unblock_stores(self, stores):
        """
        Given a mapping of names to account stores, unlock all distributions
        in the store, and cancel any pending pocket mirror sync requests.
        """
        for name, store in list(stores.items()):
            logging.info("Processing account store %s" % name)
            for dist in store.find(
                Distribution, Distribution.status != STATUS_UNLOCKED
            ):
                logging.info("  Unlocking distribution %s" % dist.data.name)
                dist.unlock()
                lockfile = os.path.join(dist.get_basedir(), "locks", "lock")
                dblockfile = os.path.join(dist.get_basedir(), "db", "lockfile")
                if os.path.exists(lockfile):
                    logging.info("  Removing stale reprepro lock")
                    os.remove(lockfile)
                if os.path.exists(dblockfile):
                    logging.info("  Removing stale reprepro db lock")
                    os.remove(dblockfile)
            for sync_request in store.find(
                SyncPocketRequest,
                Activity.id == SyncPocketRequest.activity_id,
                Activity.status == INPROGRESS,
            ):
                logging.info(
                    "  Cancelling sync request for %s"
                    % sync_request.pocket.get_full_name()
                )
                sync_request.activity.cancel()
            store.commit()


run = UnblockRepoActivities()
