# Copyright 2005-2022 Canonical Limited. All rights reserved.

import urllib.request
import urllib.parse
import urllib.error

import six
import transaction

from storm.exceptions import NotOneError

from zope.interface import implementer
from zope.component import getUtility, getMultiAdapter, hooks
from zope.event import notify
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.publisher.interfaces.http import IHTTPRequest
from zope.pluggableauth import PluggableAuthentication
from zope.pluggableauth.interfaces import (
    IFoundPrincipalFactory,
    IAuthenticatedPrincipalFactory,
    IAuthenticatorPlugin,
    ICredentialsPlugin,
    IPrincipalInfo,
)
from zope.session.interfaces import ISession, IClientIdManager

from canonical.notes.session import SessionNotes
from canonical.pam import pam

from canonical.landscape.model.main.computer import (
    get_computer_by_secure_id,
    get_pending_computer_by_secure_id,
)
from canonical.landscape.model.main.person import (
    get_person_by_email,
    get_person_by_id,
    get_person_by_identity,
    Person,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.security.principals import (
    ComputerPrincipal,
    PendingComputerPrincipal,
    UnknownComputerPrincipal,
    PersonPrincipal,
    AnonymousPrincipal,
)
from canonical.landscape.ui.session.session import ILandscapeSession
from canonical.landscape.model.configuration import (
    ISystemConfiguration,
    LANDSCAPE_KNOWN_PROXIES,
)
from canonical.landscape.security.interfaces import PrincipalLoginEvent


COMPUTER_ID_HEADER = "X-Computer-ID"
EXCHANGE_TOKEN_HEADER = "X-Exchange-Token"
UNKNOWN_HOST = "unknown"


def normalize_header_host(host):
    """
    This utility function cleans up the host or IP string obtained from a
    header and normalizes null values to "unknown".

    See http://wiki.squid-cache.org/SquidFaq/ConfiguringSquid
    under X-Forwarded section for more details.
    """
    host = str(host).strip().lower()
    if host == "none" or host == "":
        host = UNKNOWN_HOST
    return host


def get_host_from_request(request, check_known_proxies=True):
    """
    This utility function attempts to obtain a hostname or IP from the
    'X-Forwarded-For' header, or, failing that, from the 'Remote-Addr' header.

    The 'X-Forwarded-For' header can potentially contain multiple client/proxy
    hosts. We are interested in the originating one, which by convention, is
    placed first in the list.

    See http://en.wikipedia.org/wiki/X-Forwarded-For#Format for more details.

    To prevent spoofing of the host, a list of known proxies that the
    request is passing through can be set in the configuration. If this
    is set, the host left  of the last known proxy will be used, since
    the other values can't be trusted.

    @param check_known_proxies: Whether to check for known proxies. Can
        be disabled, if for example the ISystemConfiguration utility isn't
        available.
    """
    forwarded = normalize_header_host(request.getHeader("X-Forwarded-For"))
    remote = normalize_header_host(request.getHeader("Remote-Addr"))
    host = UNKNOWN_HOST
    if forwarded == UNKNOWN_HOST and remote and remote != UNKNOWN_HOST:
        host = remote
    else:
        known_proxies = None
        if check_known_proxies:
            configuration = getUtility(ISystemConfiguration)
            known_proxies = configuration.get(LANDSCAPE_KNOWN_PROXIES)
        for address in reversed(forwarded.split(",")):
            host = address
            if known_proxies is not None and address not in known_proxies:
                break
    return normalize_header_host(host)


def write_last_login_notes(person_id, notes):
    """
    Used after password and openid authentication, this function updates the
    session notes with a users previous login time and host.
    """
    message = None
    person = get_person_by_id(person_id)
    if person.last_login_time:
        login_time = person.formatter.format_short_date(person.last_login_time)
        message = "Last login was %s" % (login_time,)
    if message:
        if person.last_login_host:
            message = "%s from host %s" % (message, person.last_login_host)
        notes.info("%s." % (message,))


def principal_login_subscriber(event):
    """Update last login information and session notes on user login"""
    person = event.principal.unproxied_person
    request = event.request

    # Regenerate a new cookie
    client_id_manager = getUtility(IClientIdManager)
    client_id_manager.setRequestId(request, client_id_manager.generateUniqueId())

    session_data = ISession(request)["landscape.authentication"]
    session_data["person_id"] = person.id

    if event.update:
        notes = SessionNotes(request)
        write_last_login_notes(person.id, notes)
        person.update_login_data(get_host_from_request(event.request))


@implementer(IPrincipalInfo)
class LandscapePrincipalInfo(object):
    """A wrapper for the Landscape-specific representation of a principal.

    For compatiblity with previous auth implementation, we have the auth
    implementation return principal objects wrapped in an info, as different
    auth forms entail different types of principals and typically already
    have in hand an underlying rdb representation of a principal.
    """

    title = ""

    def __init__(self, principal):
        self.principal = principal
        self.id = principal.id


class LandscapeAuthentication(PluggableAuthentication):
    """
    Landscape specific subclass of pau utility that duplicates the same
    functionality there plus.
      - removes id prefix mangling
      - performs request decoration by principal objects ( non anonymous )
      - on invalid login attempts stores session
      - instead of returning none, returns an AnonymousPrincipal
    """

    # tests instantiate auth utility directly
    authenticatorPlugins = ("computer", "database", "session", "pam")
    credentialsPlugins = ("default",)

    def authenticate(self, request):
        authenticator_plugins = [
            plugin for name, plugin in self.getAuthenticatorPlugins() if plugin
        ]

        info = None
        for name, credential_plugin in self.getCredentialsPlugins():
            credentials = credential_plugin.extractCredentials(request)
            for authentication_plugin in authenticator_plugins:
                info = authentication_plugin.authenticateCredentials(credentials)
                if info is None:
                    continue
                info.credentialsPlugin = credential_plugin
                info.authenticatorPlugin = authentication_plugin
                principal = getMultiAdapter(
                    (info, request), IAuthenticatedPrincipalFactory
                )(self)
                principal.decorate_request(request)
                return principal
        # only display notes on invalid login attempt, do this in the auth
        # utility to avoid putting this responsibility on the auth plugins.
        if info is None and request.get("login.password"):
            self.logout(request)
            # XXX We shouldn't create a session just to store this
            # note. See Bug #605917.
            # Use a specific session key for authentication errors.
            message = "Invalid login credentials. Please try again."
            session = ISession(request)
            session["landscape.authentication.error"].update(
                {"authentication_error": message}
            )

            transaction.commit()
        return AnonymousPrincipal()

    def unauthorized(self, id, request):
        if id != "landscape.anonymous" or not IHTTPRequest.providedBy(request):
            return
        return super(LandscapeAuthentication, self).unauthorized(id, request)


@implementer(IAuthenticatedPrincipalFactory)
class LandscapePrincipalFactory(object):
    """PAU principal factory for landscape, for authenticated principals"""

    def __init__(self, info, request):
        self.info = info
        self.request = request

    def __call__(self, authentication):
        return self.info.principal


@implementer(IFoundPrincipalFactory)
class LandscapeFoundPrincipalFactory(object):
    """PAU principal factory for landscape, for introspected principals"""

    def __init__(self, info):
        self.info = info

    def __call__(self, authentication):
        return self.info.principal


@implementer(ICredentialsPlugin)
class LandscapeCredentialExtraction(object):
    """Extracts Principal Credentials from Request"""

    def extractCredentials(self, request):
        """Extract credentials from request"""
        # XXX We need to import form_token inline to avoid triggering chameleon
        # imports before the CHAMELEON_CACHE environment variable gets set.
        from canonical.content.form import form_token

        credentials = {"request": request}
        if not form_token.validate(request):
            return credentials
        login = request.get("login.password")
        if login and isinstance(login, six.string_types):
            credentials["login.password"] = login
        email = request.get("login.email")
        if email and isinstance(email, six.string_types):
            credentials["login.email"] = email
        return credentials

    def challenge(self, request):
        """Redirect user to a login form"""
        if not IHTTPRequest.providedBy(request):
            return False

        site = hooks.getSite()

        # We build the onward URL like this because unauthorized can be
        # called at any point in the traversal, using request.getURL() uses
        # the traversed objects to build the URL, we may not have seen all
        # the items in the path at this point.
        stack = request.getTraversalStack()
        stack.reverse()
        query = request.get("QUERY_STRING", "")
        # Encode url path to avoid redirecting to an unescaped utf8 URL.
        if stack:
            encoded_stack = urllib.parse.quote("/".join(stack).encode("utf-8"))
            next_url = "/".join([request.getURL(), encoded_stack])
        else:
            next_url = request.getURL()

        if query:
            next_url += "?" + query
        next_url = next_url.encode("utf-8")

        url = "%s%s?%s" % (
            absoluteURL(site, request),
            "/login/authenticate",
            urllib.parse.urlencode({"next_url": next_url}),
        )
        request.response.redirect(url)
        return True

    def logout(self, request):
        """Expire a user's credentials"""
        client_id_manager = getUtility(IClientIdManager)
        client_id = client_id_manager.getRequestId(request)
        if client_id is None:
            # We don't have a session (no client id in the request),
            # so don't bother with expiring.
            return
        # Make sure we make the session before expiring the cookie,
        # otherwise we'll end up expiring a newly made session and
        # leaving the original session open
        session = ILandscapeSession(request)
        request.response.expireCookie(client_id_manager.namespace)
        session.expire()
        transaction.commit()


@implementer(IAuthenticatorPlugin)
class LandscapeDatabaseAuthentication(object):
    """Database authentication for Users"""

    def authenticateCredentials(self, credentials):
        if not ("login.email" in credentials and "login.password" in credentials):
            return
        person = get_person_by_email(credentials["login.email"])
        if not person or person.identity:
            return
        if not person.check_password(credentials["login.password"]):
            return
        principal = PersonPrincipal(person)
        notify(PrincipalLoginEvent(principal, credentials["request"]))
        return LandscapePrincipalInfo(principal)

    def principalInfo(self, id):
        if not isinstance(id, six.string_types) or not id.isdigit():
            return
        person = get_person_by_id(int(id))
        if not person:
            return
        return LandscapePrincipalInfo(PersonPrincipal(person))


class LandscapePAMAuthentication(LandscapeDatabaseAuthentication):
    """Authenticate credentials against underlying PAM implementation."""

    def authenticateCredentials(self, credentials):
        # If we can't authenticate the user, exit immediately.
        if not ("login.email" in credentials and "login.password" in credentials):
            return

        person, identity = self._get_person_identity(credentials["login.email"])

        # This calls down to PAM and authenticates the credentials against the
        # "landscape" service, if the authentication fails, then we exit.
        if not pam.authenticate(identity, credentials["login.password"], "landscape"):
            return

        # If we don't have the person, then we can conclude that the person is
        # a valid user, but that there's no person object stored for them
        # Return an AnonymousPrincipal, this indicates that this IAuthenticator
        # plugin succeeded (it returned a LandscapePrincipalInfo).
        if not person:
            principal = AnonymousPrincipal()
        else:
            # Otherwise we have a fully valid user authenticated via PAM
            principal = PersonPrincipal(person)
            notify(PrincipalLoginEvent(principal, credentials["request"]))
        return LandscapePrincipalInfo(principal)

    def _get_person_identity(self, provided_identity):
        """
        PAM identities can be configured not to use the FQDN. We check what's
        been provided and handle non-FQDNs by looking up and returning the
        appropriate Person record.

        Returns the Person record and the corresponding identity to auth.
        """
        person = get_person_by_identity(provided_identity)

        if person:
            # Nice, they exist with the identity as-is.
            return person, provided_identity

        if "@" in provided_identity:
            # FQDN identity, but we don't have a record of them, split off the
            # domain and try again.
            identity = provided_identity.split("@")[0]
            return get_person_by_identity(identity), identity

        # Last-ditch effort: try to get person whose identity starts with the
        # provided non-FQDN identity.
        try:
            person = (
                get_main_store()
                .find(
                    Person,
                    Person.identity.like(
                        provided_identity + "@%", case_sensitive=False
                    ),
                )
                .one()
            )

            if person:
                # Even though a person exists, we can't be certain it's the
                # right one until we auth against their full identity.
                return person, person.identity
        except NotOneError:
            # Multiple records match, we can't afford to provide the wrong
            # record so fall through.
            pass

        return None, provided_identity


@implementer(IAuthenticatorPlugin)
class LandscapeSessionAuthentication(object):
    """Database Session authentication for Users"""

    def authenticateCredentials(self, credentials):
        if "bypass_session" in credentials:
            return

        # Check if there's a client id set for the session. If no
        # client id exists, don't bother checking for credentials in
        # the session, to avoid creating a session needlessly.
        client_id_manager = getUtility(IClientIdManager)
        client_id = client_id_manager.getRequestId(credentials["request"])
        if client_id is None:
            # There's no client id set, which means the user hasn't
            # been authenticated yet.
            return

        session = ISession(credentials["request"])
        session_data = session["landscape.authentication"]
        person_id = session_data.get("person_id")
        if person_id is None:
            return
        person = get_person_by_id(person_id)
        if not person:
            return
        return LandscapePrincipalInfo(PersonPrincipal(person))

    def principalInfo(self, id):
        return  # Let database principal auth do this


class ComputerAuthenticationPolicy(object):
    def authenticate(self, request):
        """
        Lookup a L{Computer} or a L{PendingComputer} with the given secure ID.
        """
        principal = UnknownComputerPrincipal()

        secure_id = request.getHeader(COMPUTER_ID_HEADER)
        if secure_id is not None:
            computer = get_computer_by_secure_id(secure_id)
            if computer is not None:
                # Check that the provided exchange token matches the one we
                # had saved during the last exchange. If there's no token it
                # means either that we're talking to an older client or that
                # the client itself can't provide the token (e.g. because the
                # last exchange failed due to network errors), in any case we
                # just authenticate the request.
                token = request.getHeader(EXCHANGE_TOKEN_HEADER)
                try:
                    token = token.encode()
                except AttributeError:
                    pass
                if token == computer.next_exchange_token or token is None:
                    principal = ComputerPrincipal(computer)
            else:
                pending_computer = get_pending_computer_by_secure_id(secure_id)
                if pending_computer is not None:
                    principal = PendingComputerPrincipal(pending_computer)

        principal.decorate_request(request)
        return principal
