# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
The Landscape security system delegates the decision of whether to grant
somebody a certain permission on a certain object to a set of permission
checkers.

Each checker is role-based, that means it first determines which roles the
accessing entity has on the accessed object, and then outputs a positive
decision if and only if the entity has at least one role granting the required
permission.
"""

from zope.component import adapts, adapter, queryMultiAdapter
from zope.interface import implementer, Interface
from zope.security.proxy import removeSecurityProxy

from canonical.security.policy import RolePermissionChecker
from canonical.security.interfaces import IPermissionChecker

from canonical.landscape.security.interfaces import (
    IPersonPrincipal,
    IComputerPrincipal,
    IPendingComputerPrincipal,
)
from canonical.landscape.security.roles import SUPPORT_ROLES
from canonical.landscape.model.main.interfaces import (
    IAccessContextAware,
    IAccessContext,
    IPerson,
    IComputer,
    IPendingComputer,
    IAccount,
    IAccountAware,
    IAccessContextComposite,
    IAccessDelegateAware,
)
from canonical.landscape.model.computer.interfaces import IComputerAware
from canonical.landscape.model.selection.interfaces import IComputerSelection
from canonical.landscape.model.license.management import ALL_LICENSE_TYPES


class AnythingByPerson(RolePermissionChecker):

    adapts(Interface, IPersonPrincipal)

    def get_role_registry(self):
        """Return the person's role registry, which includes custom roles."""
        return self.principal.unproxied_person.get_role_registry()

    def get_person_roles(self):
        """
        Returns the "Person" role and all global roles the L{Person} has.
        """
        person = self.registry.get("Person")
        assert person
        roles = set()
        # get global roles
        roles.update(self.principal.get_roles(None))
        roles.add(person)
        return roles

    get_roles = get_person_roles


class AccessContextAwareByPerson(AnythingByPerson):

    adapts(IAccessContextAware, IPersonPrincipal)

    def get_roles(self):
        """Return the set of roles this person has in the adapted object.

        Beside the direct access context associated with the adapted object,
        this method also takes its possible ancestor access contexts into
        account (parent, grand-parent, etc.).

        So the returned set contains all roles the person has on the direct
        access context plus any role he/she has in the ancestor ones.

        This ensures that a person with a given role in a higher-level access
        context, will be granted that role on all its descendants as well, even
        if there's no explicit row in the person_access table for it.
        """
        roles = self.get_person_roles()
        roles.update(self.principal.get_roles(self.context.access_context))

        access_context = self.context.access_context
        while access_context.parent_id is not None:
            roles.update(self.principal.get_roles(access_context.parent))
            access_context = access_context.parent

        return roles


class AccessContextCompositeByPerson(AccessContextAwareByPerson):

    adapts(IAccessContextComposite, IPersonPrincipal)

    def __call__(self, permission):
        """
        Before performing the standard checks, check the AccountManager role.

        In order to decide whether to display Salesforce-related menu items,
        the navigation bar checks if the accessing L{Person} has some
        permissions which are only granted by the AccountManager and the
        SupportProvider roles. If the context object being accessed is a
        composite one with a lots of children, the composite checker
        ends up loading all the children of the object trying to check if
        any is associated with an access  context providing the required
        permission. This check is unnecessary and kills performance, so this
         workaround is needed.
        """
        global_roles = set(self.registry.get(role) for role in SUPPORT_ROLES)
        granting_roles = set(self.registry.get_with_permission(permission))
        person_roles = set(self.get_person_roles())
        if granting_roles.issubset(global_roles):
            return len(person_roles.intersection(granting_roles)) > 0
        return super(AccessContextCompositeByPerson, self).__call__(permission)

    def get_roles(self):
        """Return the union of the roles on the context and on its children.

        Baside returning the roles the accessing L{Person} has on the access
        context associated to the L{IAccessContextComposite} being accessed,
        this method also returns the roles the accessing L{Person} has on the
        access contexts associated to the children of the context object
        being accessed. This is done recursively for all descendants of the
        children as well.

        The L{IAccessContextAwareByPerson.get_roles} method is used to
        determine the roles on the single nodes of the composite tree, so
        parent access context are taken into account.

        A cutoff role can be used to limit the roles that can be acquired from
        the children. In particular if the C{access_cutoff_role} attribute of
        the context object is not C{None}, then all roles that the person has
        on the context's children and that descend from C{access_cutoff_role}
        will be replaced by C{access_cutoff_role} itself.

        A use case of the cutoff role feature is the L{CustomGraph} composite.
        In this case the children of the composite are the computers the custom
        graph is associated with. A L{Person} with the Owner role on the access
        context associated to one these child computers, should be able to view
        the custom graph, but to edit it. That's why L{CustomGraph} sets the
        cutoff role to AdminViewer, which grants C{ViewCustomGraph} but not
        C{EditCustomGraph}.

        @return: a generator yielding the L{Role}s the accessing L{Person} has
            on the L{AccessContext} associated with the context and also on the
            L{AccessContext}s associated with its children, recursively.
        """
        self._yielded_roles = []
        roles = super(AccessContextCompositeByPerson, self).get_roles()
        for role in roles:
            yield role
            self._yielded_roles.append(role)
        for role in self._get_children_roles(self.context.access_children):
            # This role has been yielded by self._get_children_roles, and it
            # was already appended to self._yielded_roles, so we don't need to
            # re-append it.
            yield role

    def _get_children_roles(self, children):
        """Return the roles the person has on the given C{children} objects."""

        for child in children:

            child_roles = AccessContextAwareByPerson(child, self.principal).get_roles()

            if self.context.access_cutoff_role:
                roles_to_cutoff = []
                for role in child_roles:
                    if role.is_descendant_of(self.context.access_cutoff_role):
                        roles_to_cutoff.append(role)

                if roles_to_cutoff:
                    child_roles.difference_update(roles_to_cutoff)
                    child_roles.add(self.context.access_cutoff_role)

            for role in child_roles:
                if role not in self._yielded_roles:
                    yield role
                    self._yielded_roles.append(role)

            if IAccessContextComposite.providedBy(child):
                for role in self._get_children_roles(child.access_children):
                    # This role has been yielded by self._get_children_roles,
                    # and it was already appended to self._yielded_roles, so
                    # we don't need to re-append it.
                    yield role


class ComputerSelectionByPerson(AccessContextAwareByPerson):

    adapts(IComputerSelection, IPersonPrincipal)

    def __call__(self, permission):
        """
        The given C{permission} is granted if at least one of following
        statements holds:

        -  the accessing person has the given permission on the root access
           context.

        -  the accessing person has the given permission on all the access
           contexts linked to selected computers.

        In all other cases the permission is not granted.
        """
        if super(ComputerSelectionByPerson, self).__call__(permission):
            return True
        if not self.context.get_selected_count():
            # If the selection is empty we just deny access
            return False
        person = self.principal.unproxied_person
        for access_context in self.context.get_access_contexts():
            if not person.has_permission(permission, access_context):
                return False
        return True


class AccessContextByPerson(AnythingByPerson):

    adapts(IAccessContext, IPersonPrincipal)

    def get_roles(self):
        roles = self.get_person_roles()
        access_context = self.context
        while access_context is not None:
            roles.update(self.principal.get_roles(access_context))
            access_context = access_context.parent
        return roles


class PersonByPerson(AnythingByPerson):
    """
    Grant C{Owner} if the object-person in question is the same as the context-
    person, and grant C{Coworker} if the context-person shares any access
    contexts with the object-person, other than RetiredMember.
    """

    adapts(IPerson, IPersonPrincipal)

    def get_roles(self):
        roles = self.get_person_roles()

        person = self.principal.unproxied_person
        if self.context == person:
            owner = self.registry.get("Owner")
            assert owner
            roles.add(owner)
        # Does this person have roles related to accounts that are
        # common with the principal?
        else:
            # FIXME We should probably be gathering this data with a
            # single query instead of mashing it together by hand
            # here.
            person_roles = person.get_all_roles()
            retired_member = self.registry.get("RetiredMember")
            context_roles = self.context.get_all_roles()
            for key in person_roles:
                if key in context_roles and key is not None:
                    if retired_member in person_roles[key]:
                        continue
                    coworker = self.registry.get("Coworker")
                    assert coworker
                    roles.add(coworker)
                    break

        return roles


class PendingComputerByPendingComputer(RolePermissionChecker):

    adapts(IPendingComputer, IPendingComputerPrincipal)

    def get_roles(self):
        if self.context is self.principal.unproxied_pending_computer:
            pending = self.registry.get("PendingComputer")
            assert pending
            return set((pending,))

        return None


class ComputerByPerson(AccessContextAwareByPerson):

    adapts(IComputer, IPersonPrincipal)

    def get_roles(self):
        roles = super(ComputerByPerson, self).get_roles()
        account = self.context.account
        roles.update(account.get_roles_by_computer(self.context))
        return roles


class ComputerByComputer(RolePermissionChecker):

    adapts(IComputer, IComputerPrincipal)

    def get_roles(self):
        if self.context is self.principal.unproxied_computer:
            computer = self.registry.get("Computer")
            assert computer

            account = self.context.account
            roles = account.get_roles_by_computer(self.context).copy()
            roles.add(computer)
            return roles

        return None


class AccountByComputer(RolePermissionChecker):
    adapts(IAccount, IComputerPrincipal)

    def _get_account(self):
        return self.context

    def get_roles(self):
        account = self.principal.unproxied_computer.account
        if self._get_account() is not account:
            return None
        computer = self.principal.unproxied_computer
        license_roles = set(
            self.registry.get(role_name) for role_name in ALL_LICENSE_TYPES
        )
        roles = computer.account.get_roles_by_computer(computer).copy()
        if roles & license_roles:
            roles.add(self.registry.get("Computer"))
        return roles


@implementer(IPermissionChecker)
@adapter(IAccessDelegateAware, IPersonPrincipal)
def AccessDelegateAwareByPerson(context, principal):
    """
    Pass on the security check to the checker for the access delegate object.

    Grants exactly the same priviliges the accessing L{Person} would get
    when trying to access the C{access_delegate} attribute of the context
    object being accessed.
    """
    # Removing the security proxy is needed in case the context object
    # was actually unproxied while its access_delegate attribute is so.
    # This situation can happen expecially in tests.
    delegate = removeSecurityProxy(context.access_delegate)
    checker = queryMultiAdapter((delegate, principal), IPermissionChecker)
    return checker


@implementer(IPermissionChecker)
@adapter(IComputerAware, IPersonPrincipal)
def ComputerAwareByPerson(context, principal):
    return ComputerByPerson(context.computer, principal)


@implementer(IPermissionChecker)
@adapter(IComputerAware, IComputerPrincipal)
def ComputerAwareByComputer(context, principal):
    return ComputerByComputer(context.computer, principal)


class AccountAwareByComputer(AccountByComputer):
    adapts(IAccountAware, IComputerPrincipal)

    def _get_account(self):
        return self.context.account
