# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.component import getUtility

from canonical.security.interfaces import IRoleRegistry

from canonical.landscape.environment import get_config_name


def setup_grants(registry=None):
    """Setup the given C{registry} with the Landscape application permissions.

    These roles are generally "global" roles (they are not specific to one
    account), and are therefore deployment-wide.

    The "Owner" role is an exception: it defines the set of permissions granted
    to the user-visible per-account "GlobalAdmin" role. The permissions it
    defines therefore will only apply to the account for which the user is
    a GlobalAdmin.

    The AccountMember role is another exception, where permissions defined are
    granted only for the account which the user is a member of.

    @param registry: The L{RoleRegistry} to setup, if not given the global one
        will be used.
    """
    from canonical.landscape.model.configuration import (
        is_standalone,
        is_testing_environment,
    )

    on_testing = is_testing_environment()
    on_lds = is_standalone()
    on_hosted = not on_lds

    lds = object()
    hosted = object()
    testing = object()

    if registry is None:
        registry = getUtility(IRoleRegistry)

    def grant(role, permission, on=None):
        if on is not None:
            if on is lds and not on_lds:
                return
            if on is hosted and not on_hosted:
                return
            if on is testing and not on_testing:
                return
        registry.grant(registry.get(role), permission)

    # For Zope 3-native compatibility. Don't use zope.View.
    grant("Anonymous", "zope.View")

    # A Person is a logged-in user.
    grant("Person", "landscape.ViewPersonResource")
    grant("Person", "landscape.ViewSearch", on=hosted)
    grant("Person", "landscape.ActivateTrial", on=hosted)
    grant("Person", "landscape.ManageSettings", on=lds)
    grant("Person", "landscape.ViewScriptIssues", on=lds)

    grant("Coworker", "landscape.ViewPerson")
    grant("Coworker", "landscape.ViewPermissionInfo")

    grant("SupportUser", "landscape.ViewAccount")
    grant("SupportUser", "landscape.ManageSupportCases")
    grant("SupportUser", "landscape.ViewAccessContext")
    grant("SupportUser", "landscape.ViewSalesforceKey", on=hosted)

    # The Owner role is the set of permissions granted to the user-visible
    # "GlobalAdmin" role (as seen in the UI).
    grant("Owner", "landscape.CreateComputer")
    grant("Owner", "landscape.AcceptPendingComputer")
    grant("Owner", "landscape.RejectPendingComputer")
    grant("Owner", "landscape.AddComputerToAccessContext")
    grant("Owner", "landscape.RemoveComputerFromAccessContext")
    grant("Owner", "landscape.ResynchronizeComputer")
    grant("Owner", "landscape.DeleteOldMessages")
    grant("Owner", "landscape.ApproveActivity")
    grant("Owner", "landscape.CancelActivity")
    grant("Owner", "landscape.DismissActivity")
    grant("Owner", "landscape.RevertActivity")
    grant("Owner", "landscape.ScheduleActivity")
    grant("Owner", "landscape.RequestCreateUser")
    grant("Owner", "landscape.EditUser")
    grant("Owner", "landscape.RemoveUser")
    grant("Owner", "landscape.EditAccessContext")
    grant("Owner", "landscape.EditAccount")
    grant("Owner", "landscape.SendMessage")
    grant("Owner", "landscape.EditComputerDetails")
    grant("Owner", "landscape.EditTags")
    grant("Owner", "landscape.RemoveComputer")
    grant("Owner", "landscape.EditComputerLicenses")
    grant("Owner", "landscape.EditPerson")
    grant("Owner", "landscape.ManageActiveProcess")
    grant("Owner", "landscape.CreateGroup")
    grant("Owner", "landscape.EditGroup")
    grant("Owner", "landscape.EditComputerPackages")
    grant("Owner", "landscape.RegisterComputer")
    grant("Owner", "landscape.EditNextExpectedSequence")
    grant("Owner", "landscape.ExecuteScript")
    grant("Owner", "landscape.ExecuteStoredScript")
    grant("Owner", "landscape.EditScript")
    grant("Owner", "landscape.CreateScript")
    grant("Owner", "landscape.ViewAccountData")
    grant("Owner", "landscape.RestartComputer")
    grant("Owner", "landscape.ShutdownComputer")
    grant("Owner", "landscape.CreateCustomGraph")
    grant("Owner", "landscape.EditCustomGraph")
    grant("Owner", "landscape.EditAlert")
    grant("Owner", "landscape.InviteAdministrator")
    grant("Owner", "landscape.CancelInvitation")
    grant("Owner", "landscape.DisableMembership")
    grant("Owner", "landscape.CreateAccessContext")
    grant("Owner", "landscape.ViewAccessContextManagement")
    grant("Owner", "landscape.CreatePackageProfile")
    grant("Owner", "landscape.EditPackageProfile")
    grant("Owner", "landscape.CreateRepositoryProfile")
    grant("Owner", "landscape.EditRepositoryProfile")
    grant("Owner", "landscape.ViewAPTSource")
    grant("Owner", "landscape.CreateAPTSource")
    grant("Owner", "landscape.EditAPTSource")
    grant("Owner", "landscape.CreateDistribution")
    grant("Owner", "landscape.ViewDistribution")
    grant("Owner", "landscape.EditDistribution")
    grant("Owner", "landscape.CreateSeries")
    grant("Owner", "landscape.EditSeries")
    grant("Owner", "landscape.CreatePocket")
    grant("Owner", "landscape.EditPocket")
    grant("Owner", "landscape.SyncPocket")
    grant("Owner", "landscape.ViewGPGKey")
    grant("Owner", "landscape.CreateGPGKey")
    grant("Owner", "landscape.EditGPGKey")
    grant("Owner", "landscape.CreateMAASServer")
    grant("Owner", "landscape.EditMAASServer")
    grant("Owner", "landscape.EditUpgradeProfile")
    grant("Owner", "landscape.ViewRole")
    grant("Owner", "landscape.EditRole")
    grant("Owner", "landscape.ViewEventLog")
    grant("Owner", "landscape.ManageAccessGroups")
    grant("Owner", "landscape.CreateJujuModel")
    grant("Owner", "landscape.EditJujuModel")
    grant("Owner", "landscape.EditRemovalProfile")
    grant("Owner", "landscape.ViewOpenStack")
    grant("Owner", "landscape.ManageOpenStack")
    grant("Owner", "landscape.ViewBilling")
    grant("Owner", "landscape.ManageBilling")

    # secret?
    grant("Owner", "landscape.ViewPermissionInfo")
    grant("Owner", "landscape.ViewPersonPassword")

    # Testing
    if get_config_name() == "default":
        # Permissions for stub account entities
        grant("Owner", "landscape.ViewStub")
        grant("Owner", "landscape.EditStub")

    # XXX AdminViewer is used only in tests.
    grant("AdminViewer", "landscape.ViewPendingComputer")
    grant("AdminViewer", "landscape.ViewActiveProcess")
    grant("AdminViewer", "landscape.ViewComputer")
    grant("AdminViewer", "landscape.ViewHistoricProcess")
    grant("AdminViewer", "landscape.ViewInvitations")
    grant("AdminViewer", "landscape.ViewLicense")
    grant("AdminViewer", "landscape.ViewLoadAverage")
    grant("AdminViewer", "landscape.ViewCPUUsage")
    grant("AdminViewer", "landscape.ViewKeystoneToken")
    grant("AdminViewer", "landscape.ViewMemoryInfo")
    grant("AdminViewer", "landscape.ViewMountInfo")
    grant("AdminViewer", "landscape.ViewProcessor")
    grant("AdminViewer", "landscape.ViewThermalZone")
    grant("AdminViewer", "landscape.ViewUser")
    grant("AdminViewer", "landscape.ViewDevice")
    grant("AdminViewer", "landscape.ViewUserManagement")
    grant("AdminViewer", "landscape.ViewGroup")
    grant("AdminViewer", "landscape.ViewComputerPackages")
    grant("AdminViewer", "landscape.GetResourceStores")
    grant("AdminViewer", "landscape.ViewSalesforceUser", on=hosted)
    grant("AdminViewer", "landscape.ViewScript")
    grant("AdminViewer", "landscape.ViewMAASServer")
    grant("AdminViewer", "landscape.ViewMAASSystem")
    grant("AdminViewer", "landscape.UseMAAS")
    grant("AdminViewer", "landscape.ViewCustomGraph")
    grant("AdminViewer", "landscape.ViewPackageProfile")
    grant("AdminViewer", "landscape.ViewRepositoryProfile")
    grant("AdminViewer", "landscape.ViewAPTSource")
    grant("AdminViewer", "landscape.ViewDistribution")
    grant("AdminViewer", "landscape.ViewPocket")
    grant("AdminViewer", "landscape.ViewHardwareInfo")
    grant("AdminViewer", "landscape.ViewUpgradeProfile")
    grant("AdminViewer", "landscape.ViewJujuModel")
    grant("AdminViewer", "landscape.ViewRemovalProfile")

    # The Computer role is the role computers are given when interacting with
    # landscape over the client-server api.
    grant("Computer", "landscape.CheckRule")
    grant("Computer", "landscape.CreateActivity")
    grant("Computer", "landscape.BlockActivity")
    grant("Computer", "landscape.UnblockActivity")
    grant("Computer", "landscape.ResynchronizeComputer")
    grant("Computer", "landscape.ViewActivity")
    grant("Computer", "landscape.QueryActivity")
    grant("Computer", "landscape.StartActivity")
    grant("Computer", "landscape.CompleteActivity")
    grant("Computer", "landscape.CreateActiveProcess")
    grant("Computer", "landscape.CreateHistoricProcess")
    grant("Computer", "landscape.CreateLoadAverage")
    grant("Computer", "landscape.CreateCPUUsage")
    grant("Computer", "landscape.CreateProcessor")
    grant("Computer", "landscape.CreateMountInfo")
    grant("Computer", "landscape.CreateMemoryInfo")
    grant("Computer", "landscape.CreateThermalZone")
    grant("Computer", "landscape.CreateUser")
    grant("Computer", "landscape.EditUser")
    grant("Computer", "landscape.SendMessage")
    grant("Computer", "landscape.EditComputerDetails")
    grant("Computer", "landscape.EditTags")
    grant("Computer", "landscape.EditActiveProcess")
    grant("Computer", "landscape.EditHistoricProcess")
    grant("Computer", "landscape.EditLoadAverage")
    grant("Computer", "landscape.EditCPUUsage")
    grant("Computer", "landscape.EditKeystoneToken")
    grant("Computer", "landscape.EditMemoryInfo")
    grant("Computer", "landscape.EditMountInfo")
    grant("Computer", "landscape.EditComputerPackages")
    grant("Computer", "landscape.EditProcessor")
    grant("Computer", "landscape.EditThermalZone")
    grant("Computer", "landscape.FetchMessages")
    grant("Computer", "landscape.DeliverMessages")
    grant("Computer", "landscape.DeleteOldMessages")
    grant("Computer", "landscape.EditNextExpectedSequence")
    grant("Computer", "landscape.EditNextExchangeToken")
    grant("Computer", "landscape.RemoveActiveProcess")
    grant("Computer", "landscape.RemoveProcessor")
    grant("Computer", "landscape.RemoveUser")
    grant("Computer", "landscape.ViewGroup")
    grant("Computer", "landscape.EditGroup")
    grant("Computer", "landscape.CreateGroup")
    grant("Computer", "landscape.RemoveGroup")
    grant("Computer", "landscape.ViewAccount")
    grant("Computer", "landscape.ViewActiveProcess")
    grant("Computer", "landscape.ViewComputer")
    grant("Computer", "landscape.ViewHistoricProcess")
    grant("Computer", "landscape.ViewLicense")
    grant("Computer", "landscape.ViewLoadAverage")
    grant("Computer", "landscape.ViewCPUUsage")
    grant("Computer", "landscape.ViewKeystoneToken")
    grant("Computer", "landscape.ViewMemoryInfo")
    grant("Computer", "landscape.ViewMountInfo")
    grant("Computer", "landscape.ViewProcessor")
    grant("Computer", "landscape.ViewComputerPackages")
    grant("Computer", "landscape.ViewThermalZone")
    grant("Computer", "landscape.ViewUser")
    grant("Computer", "landscape.ViewDevice")
    grant("Computer", "landscape.EditDevice")
    grant("Computer", "landscape.ViewCustomGraph")
    grant("Computer", "landscape.EditCustomGraph")
    grant("Computer", "landscape.ViewPackageProfile")
    grant("Computer", "landscape.ViewHardwareInfo")
    grant("Computer", "landscape.EditHardwareInfo")
    grant("Computer", "landscape.EditComputerLicenses")

    grant("UbuntuProFeatures", "landscape.ViewLicense")
    grant("UbuntuProFeatures", "landscape.EditComputerLicenses")
    grant("UbuntuProFreeFeatures", "landscape.ViewLicense")
    grant("UbuntuProFreeFeatures", "landscape.EditComputerLicenses")

    # SupportProvider is a global role!!! Be careful.
    # SupportProviders are generally Canonical employees. They have read-only
    # access to all accounts in order to help solve customer problems.
    grant("SupportProvider", "landscape.ViewRole")
    grant("SupportProvider", "landscape.ViewAccount")
    grant("SupportProvider", "landscape.ViewEventLog")
    grant("SupportProvider", "landscape.ViewInvitations")
    grant("SupportProvider", "landscape.ViewGPGKey")
    grant("SupportProvider", "landscape.ViewPerson")
    grant("SupportProvider", "landscape.ViewPermissionInfo")
    grant("SupportProvider", "landscape.ViewCustomGraph")
    grant("SupportProvider", "landscape.ViewSavedSearch")
    grant("SupportProvider", "landscape.ViewAccessContext")
    grant("SupportProvider", "landscape.ViewAccessContextManagement")
    grant("SupportProvider", "landscape.InviteAdministrator")
    grant("SupportProvider", "landscape.InviteInitialAdministrator")
    grant("SupportProvider", "landscape.CancelInvitation")

    grant("SupportProvider", "landscape.ViewAllAccounts", on=hosted)
    grant("SupportProvider", "landscape.ViewWedgedComputers", on=hosted)
    grant("SupportProvider", "landscape.ViewKnowledgeBase", on=hosted)
    grant("SupportProvider", "landscape.EditKnowledgeBase", on=hosted)
    grant("SupportProvider", "landscape.ViewVotesKnowledgeBase", on=hosted)

    grant("KnowledgeBaseUser", "landscape.ViewKnowledgeBase", on=hosted)
    grant("KnowledgeBaseUser", "landscape.VoteKnowledgeBase", on=hosted)
    grant("KnowledgeBaseReviewer", "landscape.ViewKnowledgeBase", on=hosted)
    grant("KnowledgeBaseReviewer", "landscape.ReviewKnowledgeBase", on=hosted)

    # Operator is a global role
    grant("Operator", "landscape.ResynchronizeComputer")

    # AccountManager is a global role that allows creating new accounts.
    # Hosted only!
    grant("AccountManager", "landscape.EditMaxPeopleCount", on=hosted)
    grant("AccountManager", "landscape.EditMaxAttachmentSize", on=hosted)
    grant("AccountManager", "landscape.EditPersonRole", on=hosted)
    grant("AccountManager", "landscape.CreateAccount", on=hosted)
    grant("AccountManager", "landscape.RemoveAccount", on=hosted)
    grant("AccountManager", "landscape.CreateLicense", on=hosted)
    grant("AccountManager", "landscape.InviteAdministrator", on=hosted)
    grant("AccountManager", "landscape.CancelInvitation", on=hosted)
    grant("AccountManager", "landscape.EditLicense", on=hosted)
    grant("AccountManager", "landscape.EditSalesforceKey", on=hosted)
    grant("AccountManager", "landscape.EditVatRate", on=hosted)
    grant("AccountManager", "landscape.AccessUtilityCost", on=hosted)
    grant("AccountManager", "landscape.ManageBillingConfig", on=hosted)
    grant("AccountManager", "landscape.ViewBilling", on=hosted)

    # AccountMember is granted to all users in an account.
    # These permissions are account-specific, like for "Owner".
    grant("AccountMember", "landscape.CreateActivity")
    grant("AccountMember", "landscape.ViewActivity")
    grant("AccountMember", "landscape.ViewGPGKey")
    grant("AccountMember", "landscape.ViewPerson")
    grant("AccountMember", "landscape.ViewPingTimes")
    grant("AccountMember", "landscape.ViewComputerSelection")
    grant("AccountMember", "landscape.QueryActivity")
    grant("AccountMember", "landscape.QueryAlertRule")
    grant("AccountMember", "landscape.SynchronizeTags")
    grant("AccountMember", "landscape.QueryPackageProfile")
    grant("AccountMember", "landscape.ViewLicense")
    grant("AccountMember", "landscape.DisableAccountMembership")
    grant("AccountMember", "landscape.SubscribeAlert")
    grant("AccountMember", "landscape.ViewAlertRule")
    grant("AccountMember", "landscape.GetAlertSubscription")
    grant("AccountMember", "landscape.ViewSavedSearch")
    grant("AccountMember", "landscape.CreateSavedSearch")
    grant("AccountMember", "landscape.EditSavedSearch")
