# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Landscape's import policy is defined here.

Landscape uses security infrastructure that relies on the proper use
of L{SafeImporter}-based imports.  A Landscape-specific import policy
that enforces the correct use of C{safe} amongst other things is
defined here.  The policies currently in use:

 * Ensure that C{safe} is always used when it should be.
 * Ensure that interfaces and other unprotected objects can be
   imported from sensitive modules without using C{safe}.
 * Ensure that only whitelisted names can be safely imported from
   sensitive modules into untrusted modules.
 * Ensure that imported names are in the C{__all__} of the modules
   they're being imported from.
 * Ensure that C{import *} can only be used with modules that define
   C{__all__}.
"""

from canonical.security.importguardian import ImportGuardian
from canonical.security.importpolicy import (
    FromStarNoModuleAllPolicy,
    NotInModuleAllPolicy,
    ImportIntoPolicy,
    ImportFromPolicy,
    SafeImportPolicy,
)


def create_import_policy():
    importer = ImportGuardian()

    # Don't allow C{from foo import *} for modules that don't define
    # C{__all__}.
    importer.add_policy(FromStarNoModuleAllPolicy())
    # Don't allow C{from foo import bar} if C{bar} isn't defined in
    # C{__all__}.
    policy = NotInModuleAllPolicy()
    policy.include_regexp(r"^canonical.*$")
    importer.add_policy(policy)

    # Allow test modules to be imported without invoking policy checks.
    policy = ImportIntoPolicy()
    policy.exclude_regexp(r"^.*tests.*$")
    policy.exclude_regexp(r"^zope\.testing\.doctest$")

    # Same for all schema and schema patches.
    policy.exclude_regexp(r"^canonical\.landscape\.schema.*$")
    importer.add_policy(policy)

    # Make sure policy checks are run for all Landscape modules.
    policy = ImportFromPolicy()
    policy.include_regexp(r"^canonical\.landscape.*$")
    importer.add_policy(policy)

    # Make sure safe is used appropriately.
    policy = SafeImportPolicy()
    import_from = [r"^canonical.landscape.model.*$"]
    import_into = [r"^canonical.landscape.ui.*$"]
    whitelist = [
        "get_account_by_name",
        "get_command_computer_tuples",
        "get_computer_by_id",
        "get_computers_by_ids",
        "get_pending_computers_by_ids",
        "get_password_reset",
        "get_person_by_email",
        "get_person_by_id",
        "get_package_relations_for_computer_ids",
    ]
    clearlist = ()
    policy.require_safe(import_from, import_into, whitelist, clearlist)
    importer.add_policy(policy)
    return importer


import_guardian = create_import_policy()
