# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from zope.interface import directlyProvides, implementer

from zope.security.checker import ProxyFactory

from canonical.landscape.security.interfaces import (
    IComputerPrincipal,
    IPendingComputerPrincipal,
    IPersonPrincipal,
    IAnonymousPrincipal,
    IComputerRequest,
    IPendingComputerRequest,
    IUnknownComputerRequest,
)
from canonical.landscape.service.security import PersonPrincipal as ServicePrincipal


_empty_set = frozenset()


@implementer(IPersonPrincipal)
class PersonPrincipal(object):
    def __init__(self, person):

        self.id = "landscape.person.%d" % person.id
        self.title = person.name
        self.description = ""

        self.person = ProxyFactory(person)
        self.unproxied_person = person

    def decorate_request(self, request):
        pass

    def get_roles(self, access_context=None):
        """Get this person's roles that apply to the given access context.

        If the given access context is None, only global roles will
        be returned.
        """
        all_roles = self.unproxied_person.get_all_roles()
        if access_context is not None:
            access_context_id = access_context.id
        else:
            access_context_id = None
        roles = all_roles.get(access_context_id)
        if roles is None:
            return all_roles.get(None, _empty_set)
        return roles

    def has(self, permission, on=None):
        """Check if the accessing person has the given permission.

        See also L{canonical.landscape.service.security.PersonPrincipal.has}.
        """
        principal = ServicePrincipal(self.unproxied_person.id)
        return principal.has(permission, on=on)


@implementer(IComputerPrincipal)
class ComputerPrincipal(object):
    def __init__(self, computer):
        self._computer = computer

        self.id = "landscape.computer.%d" % computer.id
        self.title = computer.title
        self.description = ""

        self.computer = ProxyFactory(computer)
        self.unproxied_computer = computer

    def decorate_request(self, request):
        directlyProvides(request, IComputerRequest)


@implementer(IPendingComputerPrincipal)
class PendingComputerPrincipal(object):
    def __init__(self, pending_computer):
        self.id = "landscape.pending_computer.%d" % pending_computer.id
        self.title = pending_computer.title
        self.description = "pending computer"

        self.pending_computer = ProxyFactory(pending_computer)
        self.unproxied_pending_computer = pending_computer

    def decorate_request(self, request):
        directlyProvides(request, IPendingComputerRequest)


class UnknownComputerPrincipal(object):
    """Principal for requests coming from unknown computer.

    A client request will be associated with this principal if it doesn't
    contain a secure ID, or if the secure ID it provides is unknown.
    """

    def __init__(self):
        self.id = "landscape.unknown_computer"
        self.title = "unknown computer"
        self.description = "unknown computer"

    def decorate_request(self, request):
        directlyProvides(request, IUnknownComputerRequest)


@implementer(IAnonymousPrincipal)
class AnonymousPrincipal(object):
    def __new__(cls):
        singleton = cls.__dict__.get("__singleton")
        if not singleton:
            singleton = object.__new__(cls)
            setattr(cls, "__singleton", singleton)
        return singleton

    def __init__(self):
        self.id = "landscape.anonymous"
        self.title = "Anonymous"
        self.description = "Anonymous"

    def decorate_request(self, request):
        pass
