# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.component import getUtility
from zope.security.interfaces import NoInteraction
from zope.security.management import getInteraction

from canonical.security.interfaces import IRoleRegistry
from canonical.security.role import Role


__all__ = ["STORED_ROLE_KEY_MIN", "setup_roles", "is_support", "SUPPORT_ROLES"]


STORED_ROLE_KEY_MIN = 100000


SUPPORT_ROLES = [
    "SupportProvider",
    "AccountManager",
]


def _is_logged_in():
    try:
        getInteraction()
    except NoInteraction:
        return False
    else:
        return True


def is_support(request):
    """Return True if the request's person has a global support role.

    If noone is logged in then return False.  If the request doesn't
    have a person then return False.
    """
    if not _is_logged_in():
        return None

    try:
        person = request.principal.person
    except AttributeError:
        return False

    for role in SUPPORT_ROLES:
        if person.has_global_role(role):
            return True
    return False


def setup_roles(registry=None):
    """Setup the given C{registry} with the Landscape application roles.

    @param registry: The L{RoleRegistry} to setup, if not given the global one
        will be used.
    """
    if registry is None:
        registry = getUtility(IRoleRegistry)

    def role(name, default=False, key=None, inherits=None):
        if inherits is not None:
            inherits = [registry.get(parent) for parent in inherits]
        registry.add(Role(name, key=key, inherits=inherits), default=default)

    # Special roles

    role("Anonymous", default=True)
    role("Computer")
    role("Person")
    role("PendingComputer")
    role("SupportUser", key=5)
    role("AccountMember", key=8, inherits=["SupportUser"])
    role("AdminViewer", key=9, inherits=["AccountMember"])
    role("KnowledgeBaseUser")
    role("KnowledgeBaseReviewer", key=10)

    # Person-specific roles

    role("Owner", key=1, inherits=["AdminViewer"])
    role("Coworker", key=2)
    role("SupportProvider", key=3, inherits=["AdminViewer"])
    role("RetiredMember", key=4)
    role("AccountManager", key=6, inherits=["SupportProvider"])
    role("Operator", key=7)
    role("AllFeatures", key=10000)
    role("BasicFeatures", key=10001)
    role("LDSBasicFeatures", key=10002)
    role("LDSVMFeatures", key=10005)
    role("LDSContainerFeatures", key=10007)
    role("TrialFeatures", key=10003, inherits=["BasicFeatures"])
    role("VMFeatures", key=10004, inherits=["BasicFeatures"])
    role("ContainerFeatures", key=10006, inherits=["BasicFeatures"])
    role("UbuntuProFeatures", key=10008, inherits=["BasicFeatures"])
    role("UbuntuProFreeFeatures", key=10009, inherits=["BasicFeatures"])

    # NEVER create roles with a key at or above STORED_ROLE_KEY_MIN
    # (100k). Stored role keys start at 100000.
