# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from canonical.landscape.service.operation import ServiceOperation
from canonical.landscape.service.accesscontext.errors import InvalidAccessGroupError
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.entity.helper import AccountEntityData
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.script.script import Script
from canonical.landscape.model.repository.aptsource import APTSource
from canonical.landscape.model.account.customgraph import CustomGraph
from canonical.landscape.model.alert.rule import Rule
from canonical.landscape.model.package.profile import PackageProfile
from canonical.landscape.model.package.autoupgrade import UpgradeProfile
from canonical.landscape.model.computer.removal import RemovalProfile
from canonical.landscape.model.repository.profile import RepositoryProfile
from canonical.landscape.model.activity.access import (
    ActivityParentAccess,
    update_activity_parent_access,
)


class RemoveAccessContext(ServiceOperation):
    """Remove an L{AccessContext}.

    @ivar context: The L{AccessContext} to remove.
    """

    _requires = "EditAccessContext"

    @classmethod
    def errors(cls):
        errors = super(RemoveAccessContext, cls).errors()
        return errors | set([InvalidAccessGroupError])

    def _perform(self):
        """Remove the access context.

        @return: The parent of the removed L{AccessContext} (all entities
            linked to the removed access context are migrated to the
            parent access context).
        """
        access_context = self.context

        if access_context.is_root:
            message = "Can't remove the default access group"
            raise InvalidAccessGroupError(message)

        # Change all model entities linked to this access context or to its
        # descendants to point to the parent of this access context
        main_store = get_main_store()
        account = access_context.account
        account_store = account.store
        ids = access_context.get_all_access_context_ids()
        # New models are being consolidated with AccountEntity, which is what
        # holds the access context information. Legacy models hold access
        # contexts directly.
        stores = {
            main_store: [Computer],
            account_store: [
                AccountEntityData,
                Script,
                APTSource,
                CustomGraph,
                Rule,
                PackageProfile,
                UpgradeProfile,
                RemovalProfile,
                RepositoryProfile,
            ],
        }
        for store, model_classes in list(stores.items()):
            for model_class in model_classes:
                self._set_access_context_to_parent(store, model_class, ids)

        # removing the access context will remove the whole tree under it
        main_store.remove(access_context)

        self._update_entity_activity_parents_access(account_store, ids)

        return access_context.parent

    def _set_access_context_to_parent(self, store, model_class, ids):
        """
        Update the access context of the models matching the given access
        context IDs. The new access context will be our access context's
        parent.

        @param store: The store the models live in.
        @param model_class: The class of the models to update.
        @param ids:
        """
        result = store.find(
            model_class,
            model_class.account_id == self.context.account.id,
            model_class.access_context_id.is_in(ids),
        )
        result.set(access_context_id=self.context.parent_id)

    def _update_entity_activity_parents_access(self, store, ids):
        """Update access cache for parent activities.

        Parents that have child activities in access groups with specified IDs
        are updated.
        """
        result = store.find(
            ActivityParentAccess.parent_id,
            ActivityParentAccess.access_context_id.is_in(ids),
        )
        parent_ids = list(result)
        if parent_ids:
            update_activity_parent_access(store, parent_ids)
