# Copyright 2005-2014 Canonical Limited.  All rights reserved.

from collections import Counter

from canonical.landscape.service.errors import ServiceError

from canonical.landscape.model.activity.status import (
    ActivityStatusChangeError,
    NullTransitionError,
)


class InvalidActivityStatusChangeError(ServiceError):
    """Raised when performing an invalid operation on an L{Activity}."""

    def __init__(
        self, summary, aggregate_errors, successful_activities, old_status_counts
    ):
        super(InvalidActivityStatusChangeError, self).__init__(summary)
        self.aggregate_errors = aggregate_errors
        self.successful_activities = successful_activities
        self.old_status_counts = old_status_counts

    def summary_by_message(self, verb, noun):
        """Return errors messages, grouped and counted."""
        counts = Counter(err.__class__ for err in list(self.aggregate_errors.values()))
        summary = []
        for error_class, count in counts.most_common():
            summary.append(_summary_for_type(count, verb, noun, error_class))
        return summary


class ActivityExcludedError(Exception):
    """
    Exception to raise when an activity does not require the action
    that is being attempted.
    """


def _summary_for_type(count, verb, noun, error_class):
    """Return a description of a homogenous group of exceptions."""
    if count == 1:
        entity = "activity"
        passive_verb = "is"
        auxillary = "does"
    else:
        entity = "activities"
        passive_verb = "are"
        auxillary = "do"
    if error_class == ActivityExcludedError:
        return "{} {} {} not require {}.".format(count, entity, auxillary, noun)
    elif error_class == NullTransitionError:
        return "{} {} {} already {}.".format(count, entity, passive_verb, verb)
    elif error_class == ActivityStatusChangeError:
        return "{} {} {} in a state that cannot be {}.".format(
            count, entity, passive_verb, verb
        )
    raise ValueError("Should not be handling {}.".format(error_class))
