# Copyright 2005-2016 Canonical Limited.  All rights reserved.

from builtins import str
from collections import namedtuple

from storm.locals import Desc
from storm.expr import Func
from storm.store import EmptyResultSet

from canonical.collection.empty import EmptyCollection

from canonical.landscape.service.operation import ServiceOperation
from canonical.landscape.service.collection.errors import InvalidQueryError
from canonical.landscape.service.collection.search import get_search_clause
from canonical.landscape.model.entity.helper import (
    get_access_context_ids_by_entity_type,
)
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.status import (
    BLOCKED,
    FAILED,
    PENDING_STATUSES,
    STATUS_PRIORITY,
)
from canonical.landscape.model.activity.search import ActivityChildSearch
from canonical.landscape.model.activity.legacy import (
    MixedActivityParentSearch,
    MixedActivityChildSearch,
)
from canonical.landscape.model.activity.collection import (
    ActivityParentCollection,
    ActivityChildCollection,
)


# Special status identifying scheduled, waiting, queued or in-progress child
# activities
PENDING = object()

GetActivityParentsResult = namedtuple(
    "GetActivityParentsResult",
    [
        "total",  # Total number of matching parent activities
        "items",  # Materialized ActivityParent objects, according to the limit
    ],
)


class GetActivityParents(ServiceOperation):
    """Get ActivityParents in the account.

    @ivar context: The ActivityParentCollection returning matching entities.
    """

    _requires = "ViewAccount"

    @classmethod
    def errors(cls):
        errors = super(GetActivityParents, cls).errors()
        return errors | set([InvalidQueryError])

    def _perform(self, limit=25, offset=0, status=None, info_classes=None, query=None):
        """
        @param limit: Maximum number of parent activities to return.
        @param offset: Fetch only activities from this offset on.
        @param status: Return only parents that have at least one child in
            the given status. Can be BLOCKED, FAILED or the special status
            PENDING (scheduled, waiting, queued or in-progress).
        @param info_classes: Return only parens that have at least one child
            with the given info clasees.
        @param query: If not None, a search query string to filter the returned
            parent activities. A the moment it supports a mix of parent and
            child search terms, since that's the current use case of the UI.
            See also the mixed search parsers in c.l.model.activity.legacy.
        @return: A GetActivityParentsResult object whose 'count' attribute is
            set to the total number of parent activities matching the given
            parameters, and whose 'items' attribute is a sub-set of matching
            parent activities according to the given limit and offset, ordered
            by creation time.
        """
        if status not in (PENDING, BLOCKED, FAILED, None):
            # This is not strictly needed, but since service operations
            # are use-case oriented and we for now the UI only these three
            # use cases separate, we add this safe guard for sanity.
            raise RuntimeError("Can't filter by status '{}'".format(status))

        result = self._get_activities_resultset(status, info_classes, query)

        # Since we're joining the activity_parent_access table, we might
        # get duplicate rows, so before counting and materializing the
        # result set, let's order it and use DISTINCT ON to remove the dupes.
        result.order_by(Desc(ActivityParent.id), Desc(ActivityParent.creation_time))
        result.config(distinct=(ActivityParent.id,))

        count = result.count()

        # Fetch the first N ones, newest first
        result.config(limit=limit, offset=offset)

        return GetActivityParentsResult(total=count, items=list(result))

    def _get_activities_resultset(self, status, info_classes, query):
        """Return a ResultSet of matching activities."""
        collection = self.context

        # For each entity type that might be involved in a child activity
        # figure out which are the access contexts the accessing principal
        # has required view permission on. We only want to return parents
        # that have at least one child that the principal has access to.
        child_access_contexts = get_access_context_ids_by_entity_type(
            collection.access_context, self.principal.person, "view"
        )

        if status is PENDING:
            collection = collection.with_child_status(
                scheduled=True, waiting=True, queued=True, inprogress=True
            )
        elif status is BLOCKED:
            collection = collection.with_child_status(blocked=True)
        elif status is FAILED:
            collection = collection.with_child_status(failed=True)

        if query or info_classes:
            collection = self._apply_children_filters(
                collection, child_access_contexts, info_classes, query, status=status
            )
        else:
            # When using a query filter we need to perform a double look up,
            # and the collection we get back is already security-filtered. See
            # _apply_children_filters.
            collection = collection.with_child_access_contexts(child_access_contexts)

        # Figure the total number of matching activities
        return collection.find()

    def _apply_children_filters(
        self, collection, child_access_contexts, info_classes, query, status=None
    ):
        """Apply filters for getting parents matching the given filters.

        @param collection: The ActivityParentCollection to apply filters to.
        @param info_clases: Info classes to filter children, or None.
        @param query: A query string with mixed parent/child criteria.
        @param status: An optional status to filter children by.
        @return: A new ActivityParentCollection with the filters applied.
        """
        account = collection.filters["account"]

        # Here we perform a UNION of two separate searches, one against parents
        # and one against children to get all matches. The query will return
        # the ActivityParent.id column for parents and the Activity.parent_id
        # column for children, so we get a uniform result set of matching
        # parent IDs.
        if info_classes:
            result = EmptyResultSet()
        else:
            result = _get_parents_for_query(
                account, child_access_contexts, query, ActivityParent.id
            )

        result = result.union(
            _get_children_for_query(
                account,
                child_access_contexts,
                query,
                Activity.parent_id,
                info_classes=info_classes,
                status=status,
            )
        )

        # Since the parent IDs we get across the UNION of the two queries
        # above might be duplicate, let's apply DISTINCT as well.
        ids = list(result.config(distinct=True))

        if not ids:
            # If there are no matches we want to return an empty result
            # set.
            collection = EmptyCollection(account.store)
        else:
            collection = collection.with_ids(ids)

        return collection

    def _get_child_access_contexts(self, root_access_context):
        """Get a map from entity types to eligible access context IDs on them.

        For each entity type that might be involved in a child activity, this
        will allow figure out which are the access contexts the principal
        has required entity view permission on.
        """
        return get_access_context_ids_by_entity_type(
            root_access_context, self.principal.person, "view"
        )


class GetActivityChildren(ServiceOperation):
    """Get children activities in the account.

    @ivar context: The ActivityChildCollection returning matching entities.
    """

    _requires = "ViewAccount"

    @classmethod
    def errors(cls):
        errors = super(GetActivityChildren, cls).errors()
        return errors | set([InvalidQueryError])

    def _perform(self, limit=25, offset=0, query=None):
        result = self._get_activities_resultset(query)
        # Order results since limit/offset are applied.
        result.order_by(self._get_order_clause())
        result.config(limit=limit, offset=offset)
        return result

    def _get_activities_resultset(self, query):
        """Return a ResultSet of matching activities."""
        collection = self.context

        access_contexts = get_access_context_ids_by_entity_type(
            collection.access_context, self.principal.person, "view"
        )
        collection = collection.with_access_contexts(access_contexts)

        result = collection.find()

        if query:
            account = collection.filters["account"]
            clause = get_search_clause(ActivityChildSearch(account), query)
            result = result.find(*clause)

        return result

    def _get_order_clause(self):
        """Return the ordering clause for the Activities ResultSet."""
        status_ids_array = "{{{}}}".format(
            ",".join(str(status.id) for status in STATUS_PRIORITY)
        )
        return Func("idx", status_ids_array, Activity.status), Activity.id


def _get_parents_for_query(account, child_access_contexts, query, column):
    """Return the activity parents matching the given search query.

    @param account: The Account to search parent activities in.
    @param child_access_context: A map from all entity types that children
        can be associated with to access context IDs the person has that
        entity type's view permission on.
    @param query: The query string for the search.
    @param column: The Storm column that the ResultSet should yield.
    """
    clause = get_search_clause(MixedActivityParentSearch(account), query)
    collection = ActivityParentCollection(
        account=account, child_access_contexts=child_access_contexts
    )
    collection = collection.select(column)
    result = collection.find()
    return result.find(*clause)


def _get_children_for_query(
    account, child_access_contexts, query, column, info_classes=None, status=None
):
    """Return the activity children matching the given search query.

    @param account: The Account to search child activities in.
    @param child_access_context: A map from all entity types that children
        can be associated with to access context IDs the person has that
        entity type's view permission on.
    @param query: The query string for the search.
    @param column: The Storm column that the ResultSet should yield.
    @param info_classes: An optional list of info classes to filter children by
    @param status: An optional activity status to filter by.
    """
    clause = get_search_clause(MixedActivityChildSearch(account), query)
    collection = ActivityChildCollection(
        account=account,
        access_contexts=child_access_contexts,
        info_classes=info_classes,
    )
    collection = collection.select(column)
    result = collection.find()

    if status is PENDING:
        result = result.find(Activity.status in PENDING_STATUSES)
    elif status is not None:
        result = result.find(Activity.status == status)

    return result.find(*clause)
