# Copyright 2005-2016 Canonical Limited.  All rights reserved.
from operator import attrgetter

from storm.locals import Desc

from canonical.landscape.service.operation import ServiceOperation
from canonical.landscape.service.collection.errors import InvalidQueryError
from canonical.landscape.service.activity.get import (
    _get_parents_for_query,
    _get_children_for_query,
)
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.collection import (
    ActivityParentCollection,
    ActivityChildCollection,
)
from canonical.landscape.model.entity.helper import (
    get_access_context_ids_by_entity_type,
)


class GetActivityMixed(ServiceOperation):
    """Get activity parents or children matching the given criteria.

    This service operation is meant to be used only by the GetActivities
    method of version 1 of the API, whose behavior is a bit fuzzy in that
    it can return both activity parents and activity children (something
    that doesn't really match the activity system model anymore). Eventually
    we'll want to introduce a new version 2 API which will have two separate
    methods, one for parents and one for children.

    See c.l.api.activity.GetActivities.

    @ivar context: The Account to get activities from.
    """

    _requires = "ViewAccount"

    @classmethod
    def errors(cls):
        errors = super(GetActivityMixed, cls).errors()
        return errors | set([InvalidQueryError])

    def _perform(self, query="", limit=1000, offset=0):
        """
        @param query: A search string used to filter the returned activities.
            See c.l.model.activity.search for supported criteria.
        @param limit: Maximum number of activities to return. The default is
            1000, the same as the GetActivities API method.
        @param offset: Return activities starting from this offset.
        """
        account = self.context
        access_contexts = get_access_context_ids_by_entity_type(
            account.access_context, self.principal.person, "view"
        )

        # Build a result set that performs a UNION of the parents and children
        # result sets, fetching only the ID column and applying limit/offset.
        result = _get_parents_for_query(
            account, access_contexts, query, ActivityParent.id
        )
        result = result.union(
            _get_children_for_query(account, access_contexts, query, Activity.id)
        )
        result.order_by(Desc(Activity.id))
        result.config(limit=limit, offset=offset)

        # Fetch all parents and children with the matching IDs.
        ids = list(result)
        parents = list(ActivityParentCollection(account=account, ids=ids).find())
        children = list(ActivityChildCollection(account=account, ids=ids).find())

        # Materialize the objects and sort them by ID.
        activities = sorted(parents + children, key=attrgetter("id"), reverse=True)
        return activities
