# Copyright 2005-2014 Canonical Limited.  All rights reserved.

from builtins import str
from collections import namedtuple

from canonical.lib.undef import Undef

from canonical.landscape.model.activity.status import (
    STATUS_PRIORITY,
    ActivityStatusChangeError,
    NullTransitionError,
)
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.search import ActivityChildSearch
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.entity.helper import (
    get_access_context_ids_by_entity_type,
)
from canonical.landscape.model.activity.collection import ActivityChildCollection

from canonical.landscape.service.collection.search import get_search_clause
from canonical.landscape.service.collection.errors import InvalidQueryError
from canonical.landscape.service.activity.errors import (
    InvalidActivityStatusChangeError,
    ActivityExcludedError,
)
from canonical.landscape.service.operation import ServiceOperation


ActivitiesChildrenOperationResult = namedtuple(
    "ActivitiesChildrenOperationResult",
    [
        # List of activities the operation has acted upon.
        "activities",
        # Dict with counts for each activity status *before*
        # performing the operation.
        "old_status_counts",
    ],
)


class ActivityChildrenOperation(ServiceOperation):
    """
    Base class for operating on child activities.

    @ivar context: The Account to operate on.
    """

    _requires = "ViewAccount"

    # Name of the method to update the activity status.
    _status_update_method = Undef()
    _exclusion_method = None
    _noun = ""

    @classmethod
    def errors(cls):
        errors = super(ActivityChildrenOperation, cls).errors()
        return errors | {InvalidQueryError, InvalidActivityStatusChangeError}

    def _perform(self, ids=None, statuses=None, query=""):
        """Perform the operation on matching activities.

        @param ids: An optional list of ids of children activities to match.
        @param statuses: An optional list of statuses to filter activities.
        @param query: A search string used to filter the returned activities.
            See c.l.model.activity.search for supported criteria.
        """
        collection = self._get_collection(ids, statuses, query)

        errors = {}
        successful_activities = []
        parent_ids = []
        account = collection.filters["account"]
        event_log = EventLog(account, self.principal.person)

        counts_by_old_status = dict.fromkeys(STATUS_PRIORITY, 0)

        if ids:
            excluded = set(ids)
        else:
            excluded = set()

        for activity in collection.find():
            excluded.discard(activity.id)
            counts_by_old_status[activity.status] += 1

            if self._exclusion_method:
                include = getattr(activity, self._exclusion_method, lambda: True)
                if not include():
                    # If the exclusion method returns False we skip it.
                    errors[activity.id] = ActivityExcludedError(
                        "Excluded as it does not require {}.".format(self._noun)
                    )
                    continue

            if activity.parent.id not in parent_ids:
                parent_ids.append(activity.parent.id)
                event_log.create_log_entry(
                    activity.parent.info, self._get_operation_summary()
                )

            try:
                # Change the activity status
                getattr(activity, self._status_update_method)()
            except (ActivityStatusChangeError, NullTransitionError) as error:
                errors[activity.id] = error
            else:
                successful_activities.append(activity)

        for activity_id in excluded:
            errors[activity_id] = ActivityExcludedError(
                "Excluded as it did not meet the criteria for your {}".format(
                    self._noun
                )
            )
        if errors:
            message = "\n".join(
                [
                    "{} => {}".format(id, str(message))
                    for id, message in list(errors.items())
                ]
            )
            raise InvalidActivityStatusChangeError(
                message, errors, successful_activities, counts_by_old_status
            )

        return ActivitiesChildrenOperationResult(
            successful_activities, counts_by_old_status
        )

    def _get_operation_summary(self):
        """
        Return a unicode summary describing the operation performed on the
        collection.
        """
        raise NotImplementedError()

    def _get_collection(self, ids, statuses, query):
        # Find access context the principal has permission on
        access_contexts = get_access_context_ids_by_entity_type(
            self.context.access_context, self.principal.person, "edit"
        )

        collection = ActivityChildCollection(
            account=self.context,
            ids=ids,
            statuses=statuses,
            access_contexts=access_contexts,
        )

        collection = collection.select(Activity.id)
        result = collection.find()

        search = ActivityChildSearch(self.context)
        clause = get_search_clause(search, query)
        activity_ids = result.find(*clause)

        # return a new collection with matching activities
        return ActivityChildCollection(account=self.context, ids=activity_ids)
