# Copyright 2005-2015 Canonical Limited.  All rights reserved.

from canonical.lib.undef import Undef

from canonical.landscape.service.security import authorize
from canonical.landscape.service.collection.operation import CollectionOperation
from canonical.landscape.service.accesscontext.lookup import AccessContextLookup
from canonical.landscape.model.activity.parent import (
    get_parent_ids_for_entities,
    update_parent_activities_after_changed_children,
)
from canonical.landscape.model.entity.data import update_entities_access_context


class CollectionAccessContextChange(CollectionOperation):
    """Change access context for entities in the collection.

    @ivar _target_requires: The permission the accessing principal must have
        on the target access context.
    """

    _target_requires = Undef()

    @classmethod
    def errors(cls):
        return (
            super(CollectionAccessContextChange, cls).errors()
            | AccessContextLookup.errors()
        )

    def _perform(self, collection, access_context_name):
        account = collection.filters["account"]

        lookup = AccessContextLookup(self.principal)
        access_context = lookup(account, access_context_name)
        authorize(self.principal, access_context, self._target_requires)
        access_context = lookup(account, access_context_name)
        entity_ids = self._get_collection_entity_ids(collection)

        if entity_ids:
            self._update_access_context(collection, entity_ids, access_context)
            self._after(account, access_context, entity_ids)
        return entity_ids

    def _update_access_context(self, collection, entity_ids, access_context):
        """Update access context in entities and related activities.

        Update the access context in entities in the collection and the access
        cache for parent activities that have children related to these
        entities.
        """
        account = collection.filters["account"]
        update_entities_access_context(
            account.store, collection.model_class, entity_ids, access_context
        )
        parent_ids = get_parent_ids_for_entities(
            account.store, collection.model_class, entity_ids
        )
        if parent_ids:
            update_parent_activities_after_changed_children(account, parent_ids)

    def _after(self, account, access_context, entity_ids):
        """Hook for subclasses to perform additional operations.

        This is called after access contexts have been updated.
        """
