# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from canonical.landscape.service.operation import ServiceOperation
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)
from canonical.landscape.service.collection.errors import UnknownEntityIDsError


class CollectionOperation(ServiceOperation):
    """Base class for operations acting on a collection of model entities.

    @ivar context: The Collection object that the model entities to operate
        on should belong to. This collection must come from a CollectionLookup
        and hence already have basic filters already applied. The operation
        will be performed on the matching entities that the user has permission
        on (see _requires below).

    @ivar _requires: The permission the accessing principal must have on each
        individual model entity belonging to the context collection. The
        semantics of this attribute differ slightly from the regular
        Operation._requires one in that the latter ensures that the principal
        has a certain permission on the Collection object itself, not on its
        individual items.
    """

    @classmethod
    def errors(cls):
        base_errors = super(CollectionOperation, cls).errors()
        return base_errors | {UnknownEntityIDsError}

    def __call__(self, *args, **kwargs):
        """Perform the operation on the entities the user has permission on.

        This method will figure out which of the entities in the context
        collection the user can act on and which he can't. It will then let
        the subclass implementation perform the actual operation business
        logic on the eligible entities.
        """
        collection = self.context
        access_context_ids = get_access_context_ids_for_person(
            collection.access_context, self._requires, self.principal.person
        )
        collection = collection.with_access_context_ids(access_context_ids)
        return self._perform(collection, *args, **kwargs)

    def _perform(self, collection, *args, **kwargs):
        """Perform the core logic of the operation.

        @param collection: A collection which has the same filters as the
            context collection, plus one to further filter the entities
            by access contexts, matching only the ones the principal has
            the required permission on (see _requires).
        """
        raise NotImplementedError("Must be implemented by sub-classes")

    def _get_entity_ids(self, collection):
        """Return IDs for entities in the collection.

        It raises an error if the IDs returned by the collection passed as
        argument are not the same returned by the context collection. This
        means the person doesn't have the required permissions on some of the
        involved entities.

        @param collection: a collection which has the same filters as the
            context collection, plus one to further filter the entities by
            access contexts, matching only the ones the principal has the
            required permission on (see _requires).

        """
        allowed_ids = self._get_collection_entity_ids(collection)

        # If IDs have been explicitly requested, consider missing also those
        # that are not there at all (which wouldn't be included in all_ids
        # below).
        requested_ids = set(self.context.filters.get("ids", ()))
        if requested_ids:
            missing_ids = requested_ids - allowed_ids
        else:
            all_ids = self._get_collection_entity_ids(self.context)
            missing_ids = all_ids - allowed_ids

        if missing_ids:
            raise UnknownEntityIDsError(missing_ids)
        return allowed_ids

    def _get_collection_entity_ids(self, collection):
        """Return a set of entity IDs from the collection."""
        return set(collection.find().values(collection.model_class.id))
