# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from canonical.landscape.service.operation import ServiceOperation
from canonical.landscape.service.entity.errors import DuplicateEntityError


class EditEntity(ServiceOperation):
    """Base class for operations editing model entities.

    @cvar _duplicate_error: The L{ServiceError} subclass to use to raise an
        error when the user tries to modify an entity and another entity with
        the same name already exists.

    @ivar context: The model entity to be edited.
    """

    _duplicate_error = DuplicateEntityError  # To be overridden by sub-classes

    @classmethod
    def errors(cls):
        return super(EditEntity, cls).errors() | set([cls._duplicate_error])

    def _perform(self, name=None, *args, **kwargs):
        """
        @param name: The modified name, or C{None} if unchanged.
        """
        if name is not None and name != self.context.name:
            entity = self._get_entity(name)
            if entity is not None:
                raise self._duplicate_error(name)
            self.context.name = name
        return self._edit(name, *args, **kwargs)

    def _edit(self, name, *args, **kwargs):
        """
        Subclasses should override this method with the actual edition logic.
        """
        raise NotImplementedError()

    def _get_entity(self, name):
        """Return any existing entity matching the given C{name}."""
        raise NotImplementedError("Must be defined by sub-classes")
