# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from canonical.security.role import UnknownRoleError
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.main.license import (
    get_license_by_account_license_key,
    LDS_ROLES,
)
from canonical.landscape.model.main.standalone import AccountLDSLicenseAPI
from canonical.landscape.service.entity.create import CreateEntity
from canonical.landscape.service.license.errors import (
    DuplicateLicenseError,
    InvalidLicenseError,
)
from canonical.lib.text import make_secure_id


class CreateLicense(CreateEntity):
    """Create a L{License}.

    @ivar context: C{None} because this is a GlobalContext operation.
    """

    _requires = "CreateLicense"

    _duplicate_error = DuplicateLicenseError

    @classmethod
    def errors(cls):
        errors = super(CreateLicense, cls).errors()
        return errors | set([InvalidLicenseError])

    def _create(self, license_key, account, role_name, seats, expiration_date=None):
        """Create a license and update the LDS license if the role requires it.

        @param license_key: The license key for the L{License}, can be C{None}.
            If C{None}, license_key will be automatically generated.
        @param account: The L{Account} in which to create the L{License}.
        @param role_name: The role name of the L{License}.
        @param seats: The number of seats included in this L{License}.
        @param expiration_date: Optional date after which the L{License}
            expires.
        @return: The newly created L{License}.
        """

        # Check for license duplicates within this account
        license = get_license_by_account_license_key(account, license_key)
        if license is not None:
            raise self._duplicate_error(license_key)

        if license_key is None:
            license_key = make_secure_id()

        if role_name in LDS_ROLES:
            if expiration_date is None:
                # LDS releases 12.09 & 13.09 cannot parse licenses without an
                # expiration date. The next public release of LDS will support
                # parsing licenses without expiration dates. When we decide to
                # drop support for 12.09 and 13.09, we can drop this check.
                raise InvalidLicenseError(
                    "Cannot create license. All LDS license roles require an "
                    "expiration date"
                )

        try:
            license = LicenseManagement(account).create_license(
                role_name=role_name,
                seats=seats,
                expiration_date=expiration_date,
                license_key=license_key,
            )
        except UnknownRoleError:
            message = "Cannot create license, invalid role specified %s" % role_name
            raise InvalidLicenseError(message)

        if role_name in LDS_ROLES:
            # Add this license to the account's LDSLicense contents
            api = AccountLDSLicenseAPI(account)
            lds_license = api.get_license()
            if lds_license is None:
                lds_license = api.create_license()
            lds_license.add_license(license)
            lds_license.regenerate_contents()

        return license

    def _get_entity(self, key):
        """
        XXX Perform license duplicate checks in _create. License_key is
        unique scoped at the account level. So, we can only check for
        duplicates within the account in which we will create. _get_entity
        currently limits us to only checking for a global unique entity
        key which is insufficient for our duplicate checks unless we encoded
        account_name:license_key into the get_entity request.
        """
        return None  # No duplicate found
