# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from canonical.lib.undef import Undef

from canonical.landscape.service.security import authorize
from canonical.landscape.service.errors import UnauthorisedError


class ServiceLookup(object):
    """Lookup model objects to be used as contexts for L{ServiceOperation}s.

    This class serves as a base class for classes looking up context objects,
    on which L{ServiceOperation}s will be performed.

    Sub-classes of this class are used by API routes to mimic the way that
    traversal works in the UI, finding a context object that API methods
    can pass to the L{ServiceOperation}s they invoke.

    @ivar _requires: The permission the accessing principal must have on the
       resource being looked up. It can be set to C{None} if not particular
       permission is required.
    """

    _requires = Undef()  # This *must* be defined by subclasses

    def __init__(self, principal):
        self.principal = principal

    def __call__(self, *args, **kwargs):
        """Perform the lookup.

        Sub-classes can add specific lookup parameters.
        """
        context = self._perform(*args, **kwargs)
        authorize(self.principal, context, self._requires)
        context.__no_proxy__ = True
        return context

    def _perform(self, *args, **kwargs):
        raise NotImplementedError("Must be implemented by sub-classes.")

    @classmethod
    def errors(cls):
        """Set of all L{ServiceError}s that this lookup might raise."""
        return set([UnauthorisedError])
