# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from zope.security.proxy import removeSecurityProxy

from canonical.lib.undef import Undef

from canonical.landscape.service.security import PersonPrincipal, authorize
from canonical.landscape.service.errors import UnauthorisedError


class ServiceOperation(object):
    """Handle an operation for performing high-level business logic.

    This class serves as base for classes implementing the business logic
    for specific use cases needed by the UI or by the API.

    Typically such operations accepts plain integer IDs or string names has
    input parameters, and use them for performing model changes and interfacing
    with external components as the job handler, the notification system and
    the mailer.

    A L{ServiceOperation} is also in charge of applying any relevant security
    checks or policies.

    @ivar context: The operation context, typically this is an instance of
        a model object which the operation will access or modify.
    @ivar principal: Who or what is performing the operation, typically an
        instance of L{PersonPrincipal}.

    @ivar _requires: The permission the accessing principal must have on the
       context in order to perform the operation. This serves as a first coarse
       grained "gateway check" protecting the operation. It should be set to
       C{None} by operations that enforce security entirely within the their
       logic.
    """

    _requires = Undef()  # This *must* be defined by subclasses

    def __init__(self, context, principal):
        # XXX Remove the security proxy around the context. This is a temporary
        # measure until we completely transition the UI layer to build on top
        # of the service layer, at that point security proxies can be dropped
        # altogether because security will be enforced here.
        self.context = removeSecurityProxy(context)

        # XXX Replace the UI-level principal with our service-level one, as for
        # the security proxy this is a temporary measure while we transition.
        if not isinstance(principal, PersonPrincipal):
            principal = PersonPrincipal(principal.unproxied_person.id)
        self.principal = principal

    def __call__(self, *args, **kwargs):
        """Perform the given operation.

        This base implementation just supports coarse-grained security checks
        against the C{_requires} attribute.

        Sub-classes should implement the service operation's logic in a
        C{_perform} method.

        @return: The result of the operation, usually a model object, a
            collection of them, or some other data structure.
        @raises L{ServiceError}: If the operation could not be performed.
        """
        # TODO (lp-1673839): Respect provided access groups.
        # Some service operations support requesting a different access
        # group than the default ("global").  However, the authorize
        # call here explicitly uses the access group of the operation's
        # context.  In many places where service operations are used
        # (e.g. several entity-related API methods), the operation's
        # context is ignorant of the requested access group found in the
        # call args.  Thus the below authorize() call will check the
        # wrong access group.  If the principal (i.e. user) is not
        # permitted the service operation's required access, as is the
        # case for restricted admins, then the authorize() call will
        # fail erroneously.  Consequently the context passed in to
        # authorize() must be one that reflects the requested access
        # group.
        authorize(self.principal, self.context, self._requires)
        return self._perform(*args, **kwargs)

    def _perform(self, *args, **kwargs):
        """Perform subclass-specific logic after coarse-grained authorization.

        Subclasses must implement this method with the actual service logic and
        are free to change the method signature. If the C{_requires} attribute
        is defined, subclasses can assume that coarse-grained security was
        already checked (but they might still want to perform ad-hoc security
        checks/filtering).
        """
        raise NotImplementedError()

    @classmethod
    def errors(cls):
        """A set of all L{ServiceError}s that this operation might raise."""
        errors = set()
        if cls._requires is not None:
            errors.add(UnauthorisedError)
        return errors
