# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from zope.cachedescriptors.property import Lazy

from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.service.errors import UnauthorisedError


class PersonPrincipal(object):
    """A principal representing a human user accessing the system.

    It is responsible for answering questions about what the user may or
    may not do within a given context.
    """

    def __init__(self, person_id):
        """
        @param person_id: The L{Person.id} of the accessing person.
        """
        self._person_id = person_id

    def has(self, permission, on=None):
        """Check whether the person has the given permission.

        @param permission: The permission to check.
        @param on: Optionally, an L{AccessContext} to check the permission
            on. If not given the permission will be checked against the global
            context.
        """
        # XXX eventually we'll want to move rbac-related logic out of Person
        # and have it here.
        legacy_permission = "landscape.%s" % permission
        return self.person.has_permission(legacy_permission, on)

    @Lazy
    def person(self):
        """Return the L{Person} linked to our person ID."""
        return get_person_by_id(self._person_id)

    def __eq__(self, other):
        if self is other:
            return True
        try:
            if self._person_id == other._person_id:
                return True
        except AttributeError:
            pass
        return False


def authorize(principal, context, permission):
    """
    Ensure that the given principal has the given permission on the context.

    @raises UnauthorisedError: If the security check fails.
    """
    if permission is not None:
        access_context = context.access_context
        if not principal.has(permission, on=access_context):
            raise UnauthorisedError()
