# Copyright 2005-2022 Canonical Limited. All rights reserved.

from functools import partial
import logging
import os
from threading import Thread

import amqp

from six.moves.urllib.parse import quote as url_quote

from zope.component import provideUtility, getUtility, getSiteManager
from zope.configuration.config import ConfigurationMachine
from zope.configuration.xmlconfig import file as zcmlFile, registerCommonDirectives

from twisted.application.internet import TCPClient
from twisted.python import log
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.threads import blockingCallFromThread

from storm.zope.interfaces import IZStorm
from storm.zope.zstorm import global_zstorm

from landscape.lib.fs import read_file

from canonical.lib.text import password_decode
from canonical.amqp.connection import BrokerConnections
from canonical.amqp.message.interfaces import IPublisher
from canonical.oops.utility import create_standard_oops_manager

from canonical.landscape.constants import ROOT_DIR
from canonical.landscape.environment import get_config_directory, get_modules
from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.setup import setup
from canonical.landscape.model.main.standalone import LDSLicenseAuthority
from canonical.landscape.model.package.client import PackageSearchClient
from canonical.landscape.model.package.search import set_external_threshold
from canonical.landscape.model.action.registry import ActionRegistry
from canonical.landscape.model.action.interfaces import IActionRegistry

from canonical.txlongpoll.client import AMQFactory


def setup_broker(config, connected, disconnected, application, reactor=None):
    """Setup a L{AMQFactory} with the given configuration."""
    broker_connections = BrokerConnections(connected, disconnected)
    broker_config = config.get("broker")
    broker_port = int(broker_config["port"])
    broker_hosts = [host.strip() for host in broker_config["host"].split(",")]
    broker_user = broker_config["user"]
    broker_password = password_decode(broker_config["password"])
    broker_vhost = broker_config["vhost"]
    for broker_host in broker_hosts:
        factory = AMQFactory(
            broker_user,
            broker_password,
            broker_vhost,
            partial(broker_connections.connected, broker_host),
            partial(broker_connections.disconnected, broker_host),
            lambda connector_reason: log.err(connector_reason[1], "Connection failed"),
        )
        client = TCPClient(broker_host, broker_port, factory, reactor=reactor)
        client.setServiceParent(application)


def decode_passwords_in_config(config):
    """
    Loop through the known passwords in the service.conf file, and replace them
    if necessary with the decoded version.
    """
    for section, key in [
        ("stores", "password"),
        ("maintenance", "store_password"),
        ("broker", "password"),
        ("schema", "store_password"),
    ]:
        if config.has_section(section):
            section_data = config.get(section)
            if section_data.get(key):
                section_data[key] = password_decode(section_data[key])
                config.set(section, section_data)
    return config


def get_instance_id():
    """Compute an instance id from the running configuration."""
    config_name = os.environ.get("LANDSCAPE_CONFIG", "default")
    try:
        instance = int(config_name.split("-")[1])
    except (ValueError, IndexError):
        instance = 1
    return instance


def get_instance_name(instance_type):
    """
    Compute a instance-specific name from the running configuration, given an
    C{instance_type}.
    """
    return "%s-%d" % (instance_type, get_instance_id())


def read_config(instance_type, registry=None):
    """
    Read configuration for the given instance type, and do some basic setup,
    like creating the default stores.

    @param instance_type: The type of service or script being configured,
       should have a matching entry in service.conf (e.g. "message-server").
    @param registry: The Zope registry to use, for testing.

    @return: The L{Config} object created.
    """
    if registry is None:
        registry = getSiteManager()

    setup()

    config_dir = get_config_directory(ROOT_DIR)
    service_content = read_file(os.path.join(config_dir, "service.conf"))
    service_content = service_content.decode("utf-8")
    instance_name = get_instance_name(instance_type)
    instance = int(instance_name.split("-")[-1])

    config = registry.getUtility(IFileBasedConfiguration)
    config.load(
        service_content % {"instance": instance, "instance-letter": chr(64 + instance)}
    )
    config = decode_passwords_in_config(config)

    instance_config = config.get(instance_type)
    if "base-port" in instance_config:
        instance_config["port"] = int(instance_config["base-port"]) + instance - 1

    site_definition = os.path.join(config_dir, "%s.zcml" % instance_type)
    if not os.path.exists(site_definition):
        site_definition = os.path.join(config_dir, "basic.zcml")
    instance_config["site-definition"] = site_definition

    if "gpg-home-path" in instance_config:
        passphrase_path = instance_config["gpg-passphrase-path"]
        passphrase = read_file(passphrase_path).strip().decode()
        authority = LDSLicenseAuthority(instance_config["gpg-home-path"], passphrase)
        registry.registerUtility(authority)

    # Provide a client to query the package search service.
    if config.has_section("package-search"):
        package_search_config = config.get("package-search")
        hostname = package_search_config.get("host", "localhost")
        port = int(package_search_config.get("port"))
        # Set the computers count threshold for using the package-search
        # service. If None, the default threshold will be used (50). See
        # also c.l.model.package.search.
        threshold = package_search_config.get("account-threshold")
        if threshold is not None:
            set_external_threshold(int(threshold))
        registry.registerUtility(PackageSearchClient(hostname, port))

    # Copy the section, to avoid mutating the global config object.
    main_store_options = config.get("stores").copy()
    host = main_store_options["host"]

    # We need to extract any port definition from the hostname definition, as
    # this part (without the hostname itself) is used when running ./bin/schema
    # with the super_user
    #
    # XXX: this will fail once we support IPv6 (and assuming we accept IP
    # addresses as hostnames.  PostgreSQL accepts an IPv6 address surrounded by
    # square brackets - we could do the same.
    port_index = host.find(":")
    host_port = ""
    if port_index > -1:
        host_port = "//%%2Fvar%%2Frun%%2Fpostgresql%s/" % host[port_index:]
    main_store_options["host_port"] = host_port

    uri_template = "postgres://%(user)s:%(password)s@%(host)s/%(database)s"

    if "store_user" in instance_config:
        user = instance_config["store_user"]
        if user == "superuser":
            uri_template = "postgres:%(host_port)s%(database)s"
        else:
            main_store_options["user"] = url_quote(user, safe="")

        if "store_password" in instance_config:
            main_store_options["password"] = url_quote(
                instance_config["store_password"], safe=""
            )

    zstorm = registry.queryUtility(IZStorm)
    if zstorm is None:
        zstorm = global_zstorm

    for store in instance_config.get("stores", "").split():
        store_options = main_store_options.copy()
        host_option = "%s-host" % store
        if host_option in store_options:
            store_options["host"] = store_options[host_option]

        # config can include options. Don't quote.
        store_options["database"] = store_options[store]

        zstorm.set_default_uri(store, uri_template % store_options)
        if instance_type == "schema":
            postgres = "postgres-%s" % store_options["host"].split(":")[0]
            if postgres not in zstorm.get_default_uris():
                postgres_options = store_options.copy()
                # The autocommit option is needed because CREATE DATABASE does
                # not work inside transactions.
                postgres_options["database"] = "postgres?isolation=autocommit"
                postgres_uri = uri_template % postgres_options
                zstorm.set_default_uri(postgres, postgres_uri)

    modules = get_modules(["model/*/actions", "model/*/actions.py"])
    action_registry = ActionRegistry()
    action_registry.scan(modules)
    registry.registerUtility(action_registry, IActionRegistry)

    return config


def load_config(instance_type, setup_config=None, provide_utility=None):
    """Load the Landscape application server config.

    This is a useful function for scripts and non-webapp code to call
    in order to do things based on the appserver config.
    """
    if setup_config is None:
        setup_config = read_config
    if provide_utility is None:
        provide_utility = provideUtility

    config = setup_config(instance_type)
    instance_config = config.get(instance_type)

    if "mailer" in instance_config:
        from zope.sendmail.delivery import QueuedMailDelivery

        path = instance_config["mailer-path"]
        delivery = QueuedMailDelivery(path)
        provide_utility(delivery, name="landscape-mailer")
        # XXX: note that we're not setting zope.sendmail's QueueProcessorThread
        # here, because we just rely on the app-server to run one for us.

    if "threads" in instance_config:
        threadpool = reactor.getThreadPool()
        threadpool.adjustPoolsize(maxthreads=int(instance_config["threads"]))

    # "devmode" must be set before loading the ZCML.
    if instance_config.get("devmode") == "on":
        configuration_context = ConfigurationMachine()
        registerCommonDirectives(configuration_context)
        configuration_context.provideFeature("devmode")
    else:
        configuration_context = None

    context = zcmlFile(
        instance_config["site-definition"], context=configuration_context
    )

    provide_utility(context)

    return config


def run_amqp_client_in_thread(
    config,
    reactor=reactor,
    publisher=None,
    Thread=Thread,
    blockingCallFromThread=blockingCallFromThread,
):
    """Run an L{AMQFactory} client and the associated reactor in a thread.

    This is useful when we need to send messages from a script which doesn't
    use the reactor as main loop. A new thread will be spawned where the
    reactor will be run and used to connect to the AMQP broker. This function
    will return once the connection is ready.

    Note that this function will raise an error in case it's not possible to
    connect to the AMQP broker.

    @return: A callable that can be used by the calling thread to stop the
        reactor and terminate its thread.
    """
    if publisher is None:
        publisher = getUtility(IPublisher)
    started = Deferred()

    def connected(broker_host, passthrough):
        logging.info("Connected to AMQP broker: {}".format(broker_host))
        publisher.connected(passthrough).chainDeferred(started)

    def disconnected(broker_host):
        logging.info("Disconnected from AMQP broker: {}".format(broker_host))
        publisher.disconnected()

    def failed(broker_host, result):
        connector, reason = result
        logging.info(
            "Connection to AMQP broker failed: {}: {}".format(broker_host, reason)
        )
        if factories[broker_host].retries > factories[broker_host].maxRetries:
            del factories[broker_host]
            if not factories:
                # Let's give up. This will trigger the except block
                # below
                started.errback(reason)
            else:
                broker_host = list(factories.keys())[0]
                logging.info("Connecting to AMQP broker: {}".format(broker_host))
                reactor.connectTCP(
                    broker_host,
                    int(broker_configuration.get("port", 5672)),
                    factories[broker_host],
                )

    def run():
        broker_host = sorted(factories.keys())[0]
        logging.info("Connecting to AMQP broker: {}".format(broker_host))
        reactor.connectTCP(
            broker_host,
            int(broker_configuration.get("port", 5672)),
            factories[broker_host],
        )
        reactor.run(installSignalHandlers=False)

    broker_configuration = config.get("broker")
    broker_hosts = [host.strip() for host in broker_configuration["host"].split(",")]
    factories = {}
    for broker_host in sorted(broker_hosts):
        factory = AMQFactory(
            broker_configuration["user"],
            broker_configuration["password"],
            broker_configuration["vhost"],
            partial(connected, broker_host),
            partial(disconnected, broker_host),
            partial(failed, broker_host),
        )
        factory.maxRetries = 2  # Try to connect at most 3 times then giveup
        factories[broker_host] = factory

    thread = Thread(target=run)
    thread.start()

    def stop_amqp_client():
        blockingCallFromThread(reactor, reactor.stop)
        thread.join()

    try:
        # Pass the started deferred to the blockingCallFromThread function
        # and shutdown the reactor if it raises an exception (that means
        # that the factory's maxRetries was reached and we're giving up)
        blockingCallFromThread(reactor, lambda: started)
    except:
        # Stop the reactor and re-raise the exception so scripts using this
        # function never execute further code.
        stop_amqp_client()
        raise

    return stop_amqp_client


# XXX: Currently OOPSes are only hooked up to appserver and
# message-server, we should configure them for every service.
def configure_oops(instance_type, config):
    """Create an OOPS manager for this service.

    @param instance_type: The type of service or script being configured,
       should have a matching entry in service.conf (e.g. "message-server").
    @param: config: The Config object returned by read_config()

    @return: An OOPSManager
    """
    oops_config = config.get("oops") if config.has_section("oops") else {}
    global_config = config.get("global")
    landscape_config = config.get(instance_type)
    oops_key = landscape_config.get("oops-key")
    soft_timeout = landscape_config.get("soft-timeout")
    path = oops_config.get("path", global_config.get("oops-path"))
    amqp_exchange = oops_config.get("amqp-exchange", "")
    amqp_key = oops_config.get("amqp-key")

    oops_manager = create_standard_oops_manager(oops_key, soft_timeout)
    if path:
        oops_manager.add_local_publishing(path)
    if amqp_key:
        broker_config = config.get("broker")
        rabbit_factory = partial(
            amqp.Connection,
            host=broker_config["host"],
            userid=broker_config["user"],
            password=broker_config["password"],
            virtual_host=broker_config["vhost"],
        )
        oops_manager.add_remote_publishing(rabbit_factory, amqp_exchange, amqp_key)
    return oops_manager
