// Copyright 2005-2013 Canonical Limited.  All rights reserved.

/*
 * Setup Pagination widget over the computers in a profile.
 *
 * Expects to have the following predefined variables on the page.
 *
 *   - computer_ids: the full set of computers ids across pagination.
 *     This is used when the user does a select all.
 *
 *   - account_url:  url for the landscape account, used as a prefix when
 *     generating individual computer urls.
 * *
 *   - broken_only: show only non compliant computers, and add additional
 *     controls for reapplying the profile.
 *
 */
LS.use("node", "event", "io-base", "datasource-io", "datasource-cache",
       "datasource-jsonschema", "landscape.listing.pager", function(Y) {

var selectedComputers = [];
var selectionPager = null;
ROW_PREFIX = "row";

function rowRenderer(row, record, index){
    var checkbox = "";

    if (broken_only) {
        row.set("id", ROW_PREFIX+record.computer_id);
        var isSelected = (selectedComputers.indexOf(record.computer_id) != -1) ? true : false;
        checkbox = "<input type='checkbox' name='selection:list'";
        checkbox += " value='" + record.computer_id + "' ";
        if (isSelected) {
            checkbox += (" checked" + "='checked'>");
            row.addClass("selected");
        } else {
            checkbox += ">";
        }

        // wrap the checkbox for styling
        checkbox = "<label>" + checkbox + "<span class='checkbox'></span></label>";
    }

    var columns = row.all("td");
    columns.item(0).set("innerHTML", checkbox);

    var computerHref = account_url + "/computer/";
    computerHref += record.computer_id + "/info";

    var computerLink = "<a href='" + computerHref + "'>" +record.title;
    computerLink += "</a>";

    columns.item(1).set("innerHTML", computerLink);
    columns.item(2).set("text", record.hostname);
    columns.item(3).set("text", record.last_activity);
}

function selectionUpdateStatus(){
    var selectedCount = selectedComputers.length;
    var rowCount = selectionPager.get("totalRecords");
    var text = "No computers selected";
    if (rowCount && selectedCount) {
        text = (rowCount > 1) ? "computers" : "computer";
        text = (selectedCount + " of " + rowCount + " " +
                text + " selected");
    }
    if (rowCount && (rowCount == selectedCount)) {
        Y.one("#select-all").addClass("selected-filter");
    } else {
        Y.one("#select-all").removeClass("selected-filter");
    }
    Y.one("#selection-status-message").set("text", text);
    if (selectedCount) {
        Y.one("#apply-button").set("disabled", false);
        Y.one("#select-none").removeClass("selected-filter");
    } else {
        Y.one("#apply-button").set("disabled", true);
        Y.one("#select-none").addClass("selected-filter");
    }
}

function selectionSelectAll(evt){
    var boxes = selectionPager.get("contentBox").all("tbody input");
    boxes.each(function(box, index, box_list) {
        box.set("checked", true);
        box.get("parentNode").get("parentNode").addClass("selected");
    });
    selectedComputers = availableComputerIds;
    selectionUpdateStatus();
    evt.preventDefault();
}

function selectionSelectNone(evt){
    var boxes = selectionPager.get("contentBox").all("tbody input");
    boxes.each( function( box, idx, box_list) {
        box.set("checked", false);
        box.get("parentNode").get("parentNode").removeClass("selected");
    });
    selectedComputers = [];
    selectionUpdateStatus();
    evt.preventDefault();
}

function onReapply(evt){
    evt.preventDefault();
    var input = Y.one("#selection_json");
    input.set("value", "["+selectedComputers+"]");
    Y.one("#computer-selection-form").submit();
}


Y.on("domready", function() {
    if (! Y.one("#profile-computers")){
        return;
    }

    var sourceSchema = {
        metaFields: {"totalResults": "totalRecords"},
        resultListLocator: "records",
        resultFields:[{key: "title"},
                      {key: "hostname"},
                      {key: "last_activity"},
                      {key: "computer_id"}]
    };

    var sourceArguments = null;
    if (broken_only) {
        sourceArguments = {filter:"non-compliant"};
    }

    var profileComputers = new Y.DataSource.IO({source: "json"});

    profileComputers.plug(
        {fn: Y.Plugin.DataSourceJSONSchema, cfg: {schema: sourceSchema }});

    // The cache entries are the entire data source response for a page. So
    // for example the default value here, caches 10 pages.
    profileComputers.plug({fn: Y.Plugin.DataSourceCache, cfg: {max: 10}});
         var data_table = Y.one("#profile-computers");
         var config = {contentBox: data_table,
                  sortKeyParameter: "sort",
                  sortDirectionParameter: "sort_dir",
                  recordSourceArguments: sourceArguments,
                  pageSize: PAGE_SIZE,
                  rowRenderer: rowRenderer,
                  recordSource: profileComputers};

    // Pager
    selectionPager = new Y.landscape.listing.Pager(config);

    // Pagination Controls
    selectionPager.plug(Y.landscape.listing.PaginationPlugin,
                        {paginationBox: Y.one("#pagination")});
    selectionPager.render();

    // Bind page controls to event handlers.
    if (broken_only) {
        Y.one("#select-all").on("click", selectionSelectAll);
        Y.one("#select-none").on("click", selectionSelectNone);
    }

    // Bind Row Event Handlers to Page In / Page Out Events
    selectionPager.on("pager:after-page-change", function() {
        var rows = selectionPager.get("contentBox").all("tbody tr");
        rows.on("computer-row|click", function(evt, boxes) {
            var row = evt.currentTarget;
            // (un)highlight row.
            row.toggleClass("selected");
            // only perform row selection on broken computers.
            if (!broken_only){
                return null;
            }
            // Add or Remove row from selected computers.
            var computerId = parseInt(row.one("input").get("value"), 10);
            var selectionIndex = selectedComputers.indexOf(computerId);
            if (selectionIndex > -1) {
                selectedComputers.splice(selectionIndex, 1);
                row.all("input[type=checkbox]").set("checked", false);
            } else {
                selectedComputers.push(computerId);
                row.all("input[type=checkbox]").set("checked", true);
            }
            selectionUpdateStatus();
            return null;
        });
    });

    selectionPager.on("pager:before-page-change", function() {
        // Unbind event handlers behavior on paged in rows.
        Y.detach("computer-row|*");
        // Reset 'Selected' CSS"
        var tbody = selectionPager.get("contentBox").one("tbody");
        tbody.all("tr").removeClass("selected");
    });

   // Bind the Action handler to serialize the selected computers.
   var applyButton = Y.one("#apply-button");
   if (applyButton) {
       applyButton.on("click", onReapply);
   }

   // Update Status
   if (broken_only){
       selectionUpdateStatus();
   }

});
});
