from past.utils import old_div
import logging
import uuid

import stripe


"""
From Nell at stripe.com via email (14Nov2017):

Typically, once a card has been retried three times and has not
succeeded, it's unlikely that the card will succeed. Our analytics show
that only 0.1% of all charges get successfully authorized between the
5th and 12th retry.

Given the current state and increased monitoring here, adjustments to
your retry logic will be needed. To fix the issue, you can set a cap on
the number of times you retry a charge on a particular card. While it's
been somewhat relaxed historically, card network rules limit the number
of times a card can be tried in a single day and can assess fines
against a merchant for exceeding these limits.

We would recommend setting the max limit at 8 without charging
a particular card more than once in a single day.
"""
MAX_ATTEMPTS = 4


def charge_payment(stripe_secret_key, payment):
    """Charge an account with the due amount in a Payment.

    @param stripe_secret_key: The secret key for the Stripe API
    @param payment: a Payment with the details of the charge.
    """
    assert payment.status in {
        "unpaid",
        "failed",
    }, "Payment {0.id} is {0.status!r} not 'unpaid' or 'failed'".format(payment)
    if payment.num_attempts >= MAX_ATTEMPTS:
        logging.warn("attempting payment after max attempts reached")

    account = payment.account
    assert account.stripe_customer_id, "Can't charge a non-stripe account: {}".format(
        account.name
    )
    description = "Canonical Landscape Payment: {} for period ending: {}".format(
        payment.payment_id, payment.period_end_date
    )
    if payment.stripe_charge_idempotency_key:
        idempotency_key = payment.stripe_charge_idempotency_key
    else:
        idempotency_key = uuid.uuid4().hex
        payment.stripe_charge_idempotency_key = idempotency_key
    # stripe requires an int amount of cents
    total_due = old_div(payment.due, 100)
    metadata = {
        "vat-country": account.vat_country,
        "vat-number": account.vat_number,
        "vat-amount": payment.vat,
        "ls-acct": account.name,
    }
    try:
        charge = stripe.Charge.create(
            api_key=stripe_secret_key,
            idempotency_key=idempotency_key,
            customer=account.stripe_customer_id,
            description=description,
            amount=total_due,
            currency="usd",
            metadata=metadata,
        )
    except stripe.error.CardError as exception:
        # Only mark the payment as failed for this kind of error, since other
        # errors are not user-visible.
        error = exception.json_body["error"]
        payment.mark_failed((error["code"], error["message"]))
        return

    if charge.status == "succeeded":
        payment.mark_paid(charge.id)
    elif charge.status == "pending":
        payment.mark_pending(charge.id)
    elif charge.status == "failed":
        error = (charge["failure_code"], charge["failure_message"])
        payment.mark_failed(error, charge.id)
    else:
        assert False, "Unknown charge status '{}'".format(charge.status)

    return charge


def get_charge(stripe_secret_key, charge_id):
    """Get the Stripe Charge with the given id."""
    return stripe.Charge.retrieve(charge_id, api_key=stripe_secret_key)
