# Copyright 2005-2019 Canonical Limited.  All rights reserved.

import logging

import stripe


class InvalidSetupError(Exception):
    """The SetupIntent provided can't be used to process payments."""


def create_customer(stripe_secret_key, intent_id, description, shipping_address):
    """Create a Stripe customer and return it.

    @param stripe_secret_key: the secret Stripe API key
    @param intent_id: the stripe setup intent which has a payment method
    @param description: the description of the new customer,
                        currently the account name
    """
    payment_method = None
    if intent_id:
        # Make sure the SCA flow succeded.
        intent = stripe.SetupIntent.retrieve(intent_id, api_key=stripe_secret_key)
        if intent.status != "succeeded" or not intent.payment_method:
            raise InvalidSetupError
        payment_method = intent.payment_method

    name = shipping_address.pop("name")
    shipping = {"name": name, "address": shipping_address}
    # XXX we need to handle all stripe connection, api, auth errors.
    customer = stripe.Customer.create(
        payment_method=payment_method,
        invoice_settings={"default_payment_method": payment_method},
        description=description,
        shipping=shipping,
        api_key=stripe_secret_key,
    )
    return customer


def get_customer(stripe_secret_key, customer_id):
    """Retrieve a Stripe customer by its ID.

    @param stripe_secret_key: the secret Stripe API key
    @param customer_id: The Stripe customer ID
    """
    return stripe.Customer.retrieve(customer_id, api_key=stripe_secret_key)


def get_shipping(customer):
    """Return the shipping details for a given customer."""
    address = customer.shipping.address
    return {
        "name": customer.shipping.name or "",
        "address": {
            "line1": address.line1 or "",
            "line2": address.line2 or "",
            "city": address.city or "",
            "state": address.state or "",
            "postal_code": address.postal_code or "",
            "country": address.country or "",
        },
    }


def _get_default_payment_method(customer):
    method_id = get_default_payment_method_id(customer)
    if not method_id:
        return None
    return stripe.PaymentMethod.retrieve(method_id, api_key=customer.api_key)


def get_default_payment_method_id(customer):
    """Return the id of the default PaymentMethod of the customer or None."""
    method_id = customer.invoice_settings.default_payment_method
    if not method_id:
        # Try with legacy source, which should be usable with PaymentMethod API
        if customer.default_source:
            method_id = customer.default_source
        else:
            return None
    return method_id


def get_card_details(customer):
    """Get card details from Stripe for the given customer.

    @param customer: The Stripe customer
    """
    payment_method = _get_default_payment_method(customer)
    if not payment_method:
        return None
    details = payment_method.billing_details
    card = payment_method.card
    return dict(
        name=details.name or "",
        address_line1=details.address.line1 or "",
        address_line2=details.address.line2 or "",
        address_city=details.address.city or "",
        address_state=details.address.state or "",
        address_zip=details.address.postal_code or "",
        address_country=details.address.country or "",
        exp_month=card.exp_month,
        exp_year=card.exp_year,
        last4=card.last4 or "",
        brand=card.brand or "",
    )


def update_credit_card(customer, card_details):
    """Update the details of a customer's credit card on Stripe."""
    payment_method = _get_default_payment_method(customer)
    if payment_method:
        return stripe.PaymentMethod.modify(
            payment_method.id,
            api_key=customer.api_key,
            billing_details={
                "name": card_details["name"] or None,
                "address": {
                    "line1": card_details["address_line1"] or None,
                    "line2": card_details["address_line2"] or None,
                    "city": card_details["address_city"] or None,
                    "state": card_details["address_state"] or None,
                    "country": card_details["address_country"] or None,
                    "postal_code": card_details["address_zip"] or None,
                },
            },
        )


def delete_credit_card(customer):
    """Remove the customer's credit card on Stripe."""
    card = _get_default_payment_method(customer)
    if card:
        card.detach()


def add_credit_card(customer, intent_id):
    """Associate a new credit card to the Stripe customer.

    @param customer: the stripe customer object
    @param intent_id: the stripe setup intent which has a payment method
    """
    payment_method = _get_default_payment_method(customer)
    if payment_method:
        # Customer already has a default payment method
        logging.warning(
            "Trying to add a new credit card to a customer "
            "who already has one. Aborting."
        )
        return None

    # Make sure the SCA flow succeded.
    intent = stripe.SetupIntent.retrieve(intent_id, api_key=customer.api_key)
    if intent.status != "succeeded":
        logging.warning(
            "Trying to add a payment method with setup intent %s", intent.status
        )
        return None
    elif not intent.payment_method:
        logging.warning("Trying to add a from SetupIntent without payment method.")
        return None

    payment_method = stripe.PaymentMethod.attach(
        intent.payment_method, customer=customer.id, api_key=customer.api_key
    )
    customer.invoice_settings.default_payment_method = intent.payment_method
    customer.save()
    return payment_method


def set_customer_metadata(customer, acctname):
    """Update the customer's metadata."""
    metadata = {
        "ls-acct": acctname,
    }
    customer.metadata.update(metadata)
    return customer.save()


def update_customer(customer, shipping):
    """Update the customer's details on Stripe."""
    customer.shipping["name"] = shipping["name"]
    shipping_address = shipping["address"]
    customer.shipping["address"] = {
        "city": shipping_address["city"],
        "country": shipping_address["country"],
        "line1": shipping_address["line1"],
        "line2": shipping_address["line2"],
        "postal_code": shipping_address["postal_code"],
        "state": shipping_address["state"],
    }
    return customer.save()


def setup_intent(stripe_secret_key):
    """
    Create a setup intent, which handles customer strong authentication
    for future off_session payments.
    """
    return stripe.SetupIntent.create(usage="off_session", api_key=stripe_secret_key)
