# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
import uuid

import stripe

from canonical.landscape.model.configuration import get_root_url
from canonical.landscape.stripe import customer as stripe_customer


"""
From Nell at stripe.com via email (14Nov2017):

Typically, once a card has been retried three times and has not
succeeded, it's unlikely that the card will succeed. Our analytics show
that only 0.1% of all charges get successfully authorized between the
5th and 12th retry.

Given the current state and increased monitoring here, adjustments to
your retry logic will be needed. To fix the issue, you can set a cap on
the number of times you retry a charge on a particular card. While it's
been somewhat relaxed historically, card network rules limit the number
of times a card can be tried in a single day and can assess fines
against a merchant for exceeding these limits.

We would recommend setting the max limit at 8 without charging
a particular card more than once in a single day.
"""
MAX_ATTEMPTS = 4


# generic messages for each status
PAYMENT_INTENTS_STATUSES = {
    "requires_payment_method": "A payment method is required.",
    "requires_confirmation": "The payment is pending confirmation.",
    "requires_action": "A user action is required to complete the payment.",
    "processing": "The payment is being processed.",
    "canceled": "The payment has been canceled.",
    "succeeded": "The payment was successful.",
}

CHARGEABLE_STATUSES = ("unpaid", "failed", "incomplete", "pending")


def try_charge_payment(stripe_secret_key, payment):
    """Try to advance a payment intent in order to have it charged.

    @param stripe_secret_key: the secret key for the Stripe API
    @param payment: a Payment with the details of the charge
    @return the stripe payment intent, or None if failed
    """
    assert (
        payment.status in CHARGEABLE_STATUSES
    ), "Payment {0.id} is {0.status!r} not in {1!r}".format(
        payment, CHARGEABLE_STATUSES
    )

    if payment.num_attempts >= MAX_ATTEMPTS:
        logging.warning("attempting payment after max attempts reached")

    account = payment.account
    assert account.stripe_customer_id, "Can't charge a non-stripe account: {}".format(
        account.name
    )
    customer = stripe_customer.get_customer(
        stripe_secret_key, account.stripe_customer_id
    )

    payment_method_id = stripe_customer.get_default_payment_method_id(customer)
    if not payment_method_id:
        payment.mark_failed(
            ("requires_payment_method", "A payment method is required"),
            payment.stripe_payment_id,
        )
        return

    try:
        intent = _get_or_create_intent(stripe_secret_key, payment, payment_method_id)
    except stripe.error.CardError as exception:
        # Only mark the payment as failed for this kind of error, since other
        # errors are not user-visible.
        error = exception.json_body["error"]
        payment.mark_failed(
            (error["code"], error["message"]), error.get("payment_intent", {}).get("id")
        )
        payment.stripe_charge_idempotency_key = None
        return

    # Are we already done?
    if intent.status == "succeeded":
        payment.mark_paid(intent.id)
        return intent
    if intent.status in "canceled":
        error = ("canceled", intent.cancellation_reason)
        payment.mark_failed(error, intent.id)
        return intent

    if intent.status in {"requires_confirmation", "requires_payment_method"} or (
        intent.status == "requires_action"
        and intent.payment_method != payment_method_id
    ):
        try:
            intent.confirm(payment_method=payment_method_id, return_url=get_root_url())
        except stripe.error.CardError:
            # The error is stored on the intent when the payment is attempted,
            # but an exception can be raised when the charge is instantly
            # rejected (e.g. expired card). So we just refresh the intent
            # data for the following logic to evaluate against the latest
            # payment status.
            intent.refresh()

    if intent.status == "succeeded":
        payment.mark_paid(intent.id)
    elif intent.status in "canceled":
        error = ("canceled", intent.cancellation_reason)
        payment.mark_failed(_get_status_from_intent(intent), intent.id)
    elif intent.status == "processing":
        payment.mark_pending(intent.id)
    elif intent.status == "requires_payment_method":
        # Treat as error. This status occurs if we try to charge a customer
        # and get a failure or if there are no default payment method.
        payment.mark_failed(_get_status_from_intent(intent), intent.id)
    elif intent.status == "requires_confirmation":
        # The payment is ready to be processed and should be confirmed
        # on the next run. Unless there is a race, getting here is unlikely.
        payment.mark_pending(intent.id)
    elif intent.status == "requires_action":
        payment.mark_need_action(intent.id, intent.next_action.redirect_to_url.url)
    else:
        assert False, "Unknown payment status '{}'".format(intent.status)

    return intent


def _get_status_from_intent(intent):
    """Get a (code, message) tuple from the intent."""
    if intent.last_payment_error:
        return (intent.last_payment_error.code, intent.last_payment_error.message)

    return (intent.status, PAYMENT_INTENTS_STATUSES[intent.status])


def _get_or_create_intent(stripe_secret_key, payment, payment_method_id):
    """Get an actionable stripe payment intent for a payment."""
    intent_id = payment.stripe_charge_id
    # Ignore (deprecated) charges, if there is one as those are not actionable.
    if intent_id and not intent_id.startswith("ch_"):
        intent = get_payment_intent(stripe_secret_key, intent_id)

        # Ignore canceled intents, as they can't be re-used.
        if intent.status != "canceled":
            return intent

    # We need a new payment intent.
    description = "Canonical Landscape Payment: {} for period ending: {}".format(
        payment.payment_id, payment.period_end_date
    )
    if payment.stripe_charge_idempotency_key:
        idempotency_key = payment.stripe_charge_idempotency_key
    else:
        idempotency_key = uuid.uuid4().hex
        payment.stripe_charge_idempotency_key = idempotency_key
    # stripe requires an int amount of cents
    account = payment.account
    total_due = payment.due // 100
    metadata = {
        "vat-country": account.vat_country,
        "vat-number": account.vat_number,
        "vat-amount": payment.vat,
        "ls-acct": account.name,
    }

    # Create a payment intent if there aren't any or if (legacy)
    # charge_id exist. PaymentIntent are state machines ending
    # either by payment or cancellation.
    has_payment_intent = intent_id and not intent_id.startswith("ch_")

    if has_payment_intent:
        intent = get_payment_intent(stripe_secret_key, intent_id)
    else:
        intent = stripe.PaymentIntent.create(
            api_key=stripe_secret_key,
            idempotency_key=idempotency_key,
            customer=account.stripe_customer_id,
            description=description,
            amount=total_due,
            currency="usd",
            metadata=metadata,
            payment_method=payment_method_id,
            confirmation_method="manual",
        )
    return intent


def get_payment_intent(stripe_secret_key, intent_id):
    """Get the Stripe Payment Intent with the given id."""
    return stripe.PaymentIntent.retrieve(intent_id, api_key=stripe_secret_key)
