# Copyright 2005-2016 Canonical Limited.  All rights reserved.
from builtins import object
import uuid
import collections

from canonical.lib import random_string


def new_stripe_customer_id():
    """Return a str that may be used as a Stripe customer ID."""
    return "cus_{}".format(random_string(14))


def new_stripe_charge_id():
    """Return a str that may be used as a Stripe charge ID."""
    return "ch_{}".format(random_string(24))


class PaymentMethod(object):
    """A Stripe payment method."""

    BillingDetails = collections.namedtuple("BillingDetails", "address name")
    BillingAddress = collections.namedtuple(
        "BillingAddress", "line1 line2 city state country postal_code"
    )
    Card = collections.namedtuple("Card", "brand exp_month exp_year last4")

    def __init__(
        self,
        address_line1="",
        address_line2="",
        address_city="",
        address_state="",
        address_country="",
        address_zip="",
        name="",
        brand="",
        exp_month="",
        exp_year="",
        last4="",
    ):
        self.billing_details = self.BillingDetails(
            self.BillingAddress(
                address_line1,
                address_line2,
                address_city,
                address_state,
                address_country,
                address_zip,
            ),
            name,
        )
        self.card = self.Card(
            brand,
            exp_month,
            exp_year,
            last4,
        )
        self.id = "pm_{}".format(uuid.uuid4().hex)


class FakePaymentIntent(object):
    """A fake payment intent."""

    PaymentError = collections.namedtuple("PaymentError", "code message")
    NextAction = collections.namedtuple("NextAction", "redirect_to_url")
    Redirect = collections.namedtuple("Redirect", "url")

    def __init__(
        self,
        id,
        status="succeeded",
        payment_method=None,
        last_error=None,
        next_action_url=None,
    ):
        self.id = id
        self.status = status
        self.payment_method = payment_method
        self.last_payment_error = self.PaymentError(*last_error) if last_error else None
        self.return_url = None
        self.next_action = (
            self.NextAction(self.Redirect(next_action_url)) if next_action_url else None
        )
        self.confirm_calls = 0

    def confirm(self, payment_method=None, return_url=None):
        self.confirm_calls += 1
        self.payment_method = payment_method
        self.return_url = return_url
        return self

    def refresh(self):
        pass


class FakeCharge(object):
    """A fake stripe Charge."""

    # Pseudo-memoisation of FakeCharge.retrieve, allowing tests to
    # mutate the charge that will be returned.
    _instance_registry = {}

    def __init__(self, id, status="succeeded", details=None, source=None):
        self.id = id
        self.status = status
        self.details = details if details else {}
        self.source = source if source else PaymentMethod()

    def __getitem__(self, item):
        return self.details[item]

    @classmethod
    def retrieve(cls, charge_id, api_key=None, **kwargs):
        return cls._instance_registry.setdefault(charge_id, cls(charge_id, **kwargs))


class FakeAddress(object):
    def __init__(
        self, line1="", line2="", city="", state="", postal_code="", country=""
    ):
        self.line1 = line1
        self.line2 = line2
        self.city = city
        self.state = state
        self.postal_code = postal_code
        self.country = country

    def __setitem__(self, key, item):
        setattr(self, key, item)

    def __getitem__(self, key):
        return getattr(self, key)


class FakeShippingAddress(object):
    """A fake shipping address"""

    def __init__(
        self, name="", line1="", line2="", city="", state="", postal_code="", country=""
    ):
        self.name = name
        self.address = FakeAddress(line1, line2, city, state, postal_code, country)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)


class InvoiceSettings(object):
    def __init__(self, payment_method=None):
        self.default_payment_method = payment_method


class FakeCustomer(object):
    """A fake stripe Customer object."""

    api_key = "some_key"

    def __init__(self, payment_method=None, shipping_address=None, metadata=None):
        self._save_count = 0
        self.shipping = shipping_address
        self.metadata = metadata or {}
        self.invoice_settings = InvoiceSettings(payment_method)
        self.default_source = None
        self.id = "cus_{}".format(uuid.uuid4().hex)

    def save(self):
        self._save_count += 1

    def assert_was_saved(self):
        assert self._save_count > 0


class FakeSetupIntent(object):
    def __init__(self, status="succeeded", payment_method=None):
        self.id = "seti_{}".format(uuid.uuid4().hex)
        self.status = status
        self.payment_method = payment_method
