# Copyright 2005-2013 Canonical Limited.  All rights reserved.
from __future__ import print_function

from future import standard_library

standard_library.install_aliases()
from builtins import range
import logging
import os
import sys
import time
import socket
import tempfile
import subprocess
import xmlrpc.client
import urllib.request
import urllib.error
import urllib.parse
import signal

import transaction

from selenium.webdriver import Chrome
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait

from testresources import TestResourceManager

from canonical.lib.config import Config

from functools import wraps
from datetime import datetime

from zope.session.interfaces import IClientIdManager

from canonical.url import get_url
from canonical.amqp.message.dispatcher import global_dispatcher
from canonical.amqp.message.interfaces import IDispatcher

from canonical.landscape.application import setup_logging
from canonical.landscape.ui.session.manager import LandscapeCookieClientIdManager
from canonical.landscape.constants import ROOT_DIR
from canonical.landscape.setup import run_amqp_client_in_thread
from canonical.landscape.environment import get_config_directory
from canonical.landscape.model.configuration import set_root_url
from canonical.landscape.testing.model import ModelTest
from canonical.landscape.testing.ui import UITest

from canonical.testing.fixtures import UtilityFixture
from canonical.testing.resources import SystemZcmlResourceManager


def _screenshot_decorator(test):
    """
    Method decorator to take a screen shot if the wrapped method
    (selnium based) raises an exception.  This decorator assumes the self
    reference has a driver member that is the selenium webdriver.  It ignores
    any errors with the screenshot process (including decorating things that
    don't act like selenium objects)
    """

    @wraps(test)
    def inner(*args, **kwargs):
        try:
            test(*args, **kwargs)
        except:
            try:
                if hasattr(test, "__self__"):
                    testObject = test.__self__
                    timestamp = datetime.now().isoformat().replace(":", "")
                    filename = testObject.id() + "_" + timestamp + ".png"
                    testObject.driver.get_screenshot_as_file(filename)
                    print("Took Screenshot: %s" % filename)
            except:
                print("Failed screenshot process: %s" % sys.exc_info()[0])
                pass
            raise

    return inner


def start_selenium_supervisor(tries=60, wait=0.5):
    supervisor = os.path.join(ROOT_DIR, "bin/selenium-supervisord")
    cmd = supervisor.split() + ["-n", "-edebug"]

    # Redirect stderr through a temporary file, so that it doesn't
    # block and we don't get an IOError on readline(), apparently
    # caused by an unhandled SIGINT (Google for it. :)
    fd, name = tempfile.mkstemp()
    stdout = open(name)
    rc = None
    try:
        try:
            proc = subprocess.Popen(
                cmd,
                shell=False,
                stdin=subprocess.PIPE,
                stdout=fd,
                stderr=subprocess.STDOUT,
                close_fds=True,
            )

            while True:
                rc = proc.poll()
                if rc is not None:
                    break
                line = stdout.readline()
                if "supervisord started" in line:
                    # process is running.
                    break
                if not tries:
                    break
                time.sleep(wait)
                tries -= 1
        finally:
            stdout.seek(0)

        if rc is None:
            rc = proc.poll()
        if rc is not None:
            raise ValueError(
                "Failed to execute supervisord:"
                "\nError: (%s) %s" % (rc, stdout.read())
            )
    finally:
        os.close(fd)
        stdout.close()
        os.remove(name)
    return proc


def start_supervisor_processes(supervisor_url, proxy=None, processes=None):
    from supervisor.xmlrpc import SupervisorTransport

    if proxy is None:
        proxy = xmlrpc.client.ServerProxy(
            supervisor_url,
            SupervisorTransport("landscape", "landscape", supervisor_url),
        )

    proxy.supervisor.clearAllProcessLogs()
    if processes == ALL_SELENIUM_PROCESSES:
        proxy.supervisor.startAllProcesses(False)
    else:
        for name in processes:
            proxy.supervisor.startProcess(name, False)
    return proxy


def check_supervisor_processes(processes, proxy, tries=30, wait=1):
    processes = set(processes)
    while True:
        running = set()
        failed = set()
        for info in proxy.supervisor.getAllProcessInfo():
            if info["name"] in processes:
                if info["statename"] == "RUNNING":
                    running.add(info["name"])
                if info["statename"] in ("FATAL", "EXITED"):
                    failed.add(info["name"])
        failed.update(processes - running)
        if not failed:
            break
        if not tries:
            messages = []
            for process in failed:
                messages.append("Failed to start %s:" % process)
                messages.append(proxy.supervisor.readProcessStdoutLog(process, 0, 0))
            raise ValueError("\n".join(messages))
        # Not all processes have started yet. Give them a little
        # longer to finish starting.
        time.sleep(wait)
        tries -= 1


def start_selenium(host="localhost", port=4445, browser="firefox", wait=1):

    tries = 60

    while True:
        try:
            log_path = os.path.join(os.getcwd(), "selenium.log")
            if os.path.exists(log_path):
                os.remove(log_path)
            driver = Chrome("chromedriver", service_log_path=log_path)
            return driver, driver.quit
        except (socket.error, IOError, urllib.error.URLError):
            if not tries:
                raise
            # The Selenium Server might not have started yet. Sleep
            # and try again.
            time.sleep(wait)
            tries -= 1


ALL_SELENIUM_PROCESSES = ("landscape", "job-handler", "async-frontend")


def start_selenium_processes(browser, processes):
    """Start all the processes managed by Selenium.

    @param browser: The browser to be used. Can be overriden by setting
        the SELENIUM_BROWSER environment variable.
    @param processes: The list of process names to be started. If
        C{None}, all processes will be started.
    """
    proc = start_selenium_supervisor()
    add_helper_process(proc.pid)

    supervisor_url = "http://127.0.0.1:9002"
    proxy = start_supervisor_processes(supervisor_url, processes=processes)

    driver, quit = start_selenium(browser=browser)

    check_supervisor_processes(processes, proxy)
    return driver, quit


class HeadlessDisplayResourceManager(TestResourceManager):
    """
    Resource manager for running test in headless mode.

    It basically does what xvfb-run does, but only what is needed for starting
    a browser.
    """

    _start_display_number = 99
    _lock_tmp_dir = "/tmp"

    _xvfb_proc = None

    def _find_free_display_number(self):
        """Return the next free display number.

        It looks through the X lock files to find out which display
        numbers are in use currently.
        """
        display_number = self._start_display_number
        while os.path.exists(
            os.path.join(self._lock_tmp_dir, ".X%s-lock" % display_number)
        ):
            display_number += 1
        return display_number

    def make(self, dependencies):
        disabled = "DISABLE_HEADLESS_DISPLAY" in os.environ
        if self._xvfb_proc is None and not disabled:
            display = ":%d" % self._find_free_display_number()
            os.environ["DISPLAY"] = display
            self._xvfb_proc = subprocess.Popen(
                ["Xvfb", display],
                shell=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                close_fds=True,
            )
            add_helper_process(self._xvfb_proc.pid)
        return self


class SeleniumResourceManager(TestResourceManager):
    """Resource Manager for tests using Selenium."""

    resources = [
        ("system", SystemZcmlResourceManager()),
        ("headless", HeadlessDisplayResourceManager()),
    ]

    browser = "firefox"
    processes = ALL_SELENIUM_PROCESSES

    driver = None

    def make(self, dependencies):
        if self.driver is None:
            # Setup a working amqp client so that tests can send out
            # notifications, to make it easier to trigger notifications from
            # selenium tests.
            os.environ["LANDSCAPE_CONFIG"] = "selenium"
            os.environ["LANDSCAPE_ROOT_URL"] = "http://localhost:9980/"
            config_dir = get_config_directory(ROOT_DIR)
            config_file = os.path.join(config_dir, "service.conf")
            config = Config()
            config.load(open(config_file).read())
            setup_logging("selenium-browser", level=logging.INFO)
            self.stop_amqp_client = run_amqp_client_in_thread(config)

            self.base_url = "http://localhost:9980"
            set_root_url(self.base_url + "/")
            transaction.commit()
            self.driver, self.quit = start_selenium_processes(
                self.browser, self.processes
            )
            self.driver.get(self.base_url)
        return self


# XXX workaround the fact that it'is not possible tear down resources after all
# tests are run, this would need cooperation from the test runner.
_tests_count = 0


class SeleniumTest(UITest):
    """Test-case for Selenium tests."""

    with_person = True
    with_license = True
    with_wsgi = True
    with_zcml = True

    selenium_manager = SeleniumResourceManager()

    def __init__(self, methodName="runTest"):
        super(SeleniumTest, self).__init__(methodName)
        global _tests_count
        _tests_count += 1

    def setUp(self):
        """Setup some default data used by many tests."""
        super(SeleniumTest, self).setUp()
        self.base_url = self.selenium.base_url
        self.browser = self.selenium.browser
        self.driver = self.selenium.driver
        self.driver.implicitly_wait(10)
        self.person.identity = "%s/id/person" % self.base_url
        # Re-register dispatcher utilities, since some base classes replace
        # them by fake ones.
        self.useFixture(UtilityFixture(global_dispatcher, IDispatcher))
        self.manager = LandscapeCookieClientIdManager()
        self.useFixture(UtilityFixture(self.manager, IClientIdManager))
        # Allow for subclasses to do extra setup before committing.
        self.seleniumSetUp()
        transaction.commit()

    def tearDown(self):
        # Override superclass method since we don't want to automatically call
        # logout() here.
        super(ModelTest, self).tearDown()
        global _tests_count
        _tests_count -= 1
        if _tests_count == 0:
            # We're done with all tests, kill everything
            terminate_helper_processes()
            self.selenium.quit()
            self.selenium.stop_amqp_client()

    def get_resources(self):
        resources = super(SeleniumTest, self).get_resources()
        return resources + [("selenium", self.selenium_manager)]

    def login(self, driver=None):
        """Generic login method."""
        if driver is None:
            driver = self.driver
        driver.get(self.base_url + "/")
        self.assertBrowserTitle("Landscape | Home", driver=driver)
        login = driver.find_element_by_css_selector("header.banner .login a")
        login.click()
        self.assertBrowserTitle("Fake OpenID provider - Decide", driver=driver)
        identity = driver.find_element_by_id("identity")
        identity.clear()
        identity.send_keys(self.person.identity)
        succeed = driver.find_element_by_id("success-button")
        succeed.click()
        # We should be either at the dashboard, or the account page,
        # depending on how many accounts the person is in.
        self.spin_assert(
            "Title didn't match Account or Dashboard.",
            lambda: (
                driver.title.endswith("Account - Landscape")
                or driver.title.endswith("Dashboard - Landscape")
            ),
        )

    def logout(self, driver=None):
        """Generic logout method."""
        if driver is None:
            driver = self.driver
        driver.get(self.base_url + "/dashboard")
        logout_element = driver.find_element_by_class_name("logout-info")
        logout_element.find_element_by_tag_name("a").click()
        self.assertBrowserTitle("Welcome! - Landscape")

    def get_url(self, context, request=None):
        if request is None:
            request = self.get_request()
        return get_url(context, request)

    def commit(self):
        transaction.commit()

    def abort(self):
        transaction.abort()

    def seleniumSetUp(self):
        pass

    def assertBrowserTitle(self, title, driver=None):
        if driver is None:
            driver = self.driver
        try:
            WebDriverWait(driver, 10).until(expected_conditions.title_is(title))
        except Exception:
            self.fail("Title is %s, not %s" % (driver.title, title))

    def spin_assert(self, msg, assertion, loops=60):
        """
        Loop over a defined amount of times trying to assert if some action
        took place before failing.
        """
        for i in range(loops):
            try:
                self.assertTrue(assertion())
                return
            except Exception:
                pass
        self.fail(msg)

    def assert_note_contains(self, text, driver):
        """
        This will check the note-message at the top of the page that pops
        up, it will spin for effectively 10 seconds waiting for text to
        appear.

        @param text: text to assert that the message contains
        @param driver: web driver to use
        """
        for i in range(10):
            note = self.driver.find_element_by_css_selector("div.note-message")
            if text in note.text:
                return
            else:
                msg = "'%s' not contained in '%s'" % (text, note.text)
                time.sleep(1)
        self.fail(msg)

    if not os.environ.get("LANDSCAPE_TEST_SELENIUM"):

        def run(self, result):
            result.startTest(self)
            result.addSkip(self, "Skipping Selenium tests")
            result.stopTest(self)


# List of pids for processes spwaned by test resources. They are terminated at
# the end of the test run.
_helper_processes = []


def add_helper_process(pid):
    _helper_processes.append(pid)


def terminate_helper_processes():
    for pid in _helper_processes:
        pid = int(pid)
        os.kill(pid, signal.SIGTERM)
        os.waitpid(pid, 0)
