# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import object
import uuid
import json

from zope.interface import implementer

from email import message_from_string
from zope.sendmail.interfaces import IMailDelivery

from twisted.internet.defer import maybeDeferred, succeed

from storm.twisted.testing import FakeTransactor

from landscape import CLIENT_API

from canonical.amqp.message.interfaces import IDispatcher
from canonical.lib.text import lower_class

from canonical.landscape.testing.request import TestRequest

from canonical.landscape.model.main.interfaces import ILDSLicenseAuthority
from canonical.landscape.model.main.standalone import generate_license_contents


@implementer(IDispatcher)
class FakeDispatcher(object):
    """
    A fake C{Dispatcher} implementation, which only logs the calls.
    """

    def __init__(self, jobs=None):
        self.calls = []
        self.actions = []
        self.errors = []
        self.notifications = []
        self.jobs = jobs

    def create_job_queue(self, namespaces):
        return str(uuid.uuid4())

    def create_job(self, page_key, namespace, action, **kwargs):
        self.calls.append((page_key, namespace, action, kwargs))

    def run_job(self, page_key):
        job_runners = {}
        for job in self.jobs:
            name = lower_class(job.__class__, suffix="JobRunner")
            job_runners[name] = job
        for call in self.calls:
            uuid, namespace, action_name, arguments = call
            arguments.pop("accumulate", None)
            runner = job_runners[action_name]
            runner.transactor = FakeTransactor()
            if uuid == page_key:
                return maybeDeferred(runner.run, **arguments)

    def send_action(self, result, uuid, action):
        self.actions.append((result, uuid, action))

    def send_error(self, result, uuid, action):
        self.errors.append((result, uuid, action))

    def send_notification(self, result, uuid, action, namespace):
        self.notifications.append((result, uuid, action, namespace))


@implementer(IMailDelivery)
class FakeMailer(object):
    def __init__(self):
        self.messages = []

    def send(self, sender, to, message):
        self.messages.append(
            {"sender": sender, "to": to, "message": message_from_string(message)}
        )

    def __getattr__(self, attr):
        # Accessing message attributes returns values from the last message
        if attr not in ("sender", "to", "message"):
            raise AttributeError("'FakeMailer' object has no attribute '%s'" % attr)

        if not self.messages:
            return None
        else:
            return self.messages[-1][attr]


class FakeMessageStore(object):
    def __init__(self):
        self.messages = []

    def add(self, message):
        self.messages.append(message)


class FakeMessageAPI(object):
    """A fake API for messages."""

    def __init__(self, computer):
        self.computer = computer
        self.messages = []
        self.request = TestRequest()
        self.client_api = CLIENT_API

    def send(self, type, message):
        message["type"] = type
        self.messages.append(message)


@implementer(ILDSLicenseAuthority)
class FakeLicenseAuthority(object):
    def create_license_contents(self, lds_license):
        """Return JSON encoded license details."""
        contents = generate_license_contents(lds_license)
        return self.sign_contents(contents)

    def sign_contents(self, contents):
        return json.dumps(contents)


class FakeFetch(object):
    """
    A fake landscape.lib.fetch.fetch() implementation for mocking transports
    in clients interacting with JSON APIs.
    """

    def __init__(self):
        self.response = {}
        self.status = 200
        self.error = None
        self.called = False

    def __call__(self, url, post=False, data="", total_timeout=600):
        self.called = True
        self.url = url
        self.post = post
        self.data = data
        self.total_timeout = total_timeout
        if self.status != 200:
            from landscape.lib.fetch import HTTPCodeError

            raise HTTPCodeError(self.status, json.dumps(self.response))
        if self.error is not None:
            from landscape.lib.fetch import PyCurlError

            raise PyCurlError(500, self.error)
        return json.dumps(self.response)


class FakeReactor(object):
    """
    A fake Twisted reactor.
    """

    def __init__(self):
        self.calls = []
        self.tcp_connections = []

    def run(self, *args, **kwargs):
        pass

    def stop(self):
        pass

    def callFromThread(self, function, *args, **kwargs):
        self.calls.append((function, args, kwargs))
        return succeed(function(*args, **kwargs))

    def callLater(self, delay, func):
        pass

    def connectTCP(self, hostname, port, client):
        client.clock = self
        self.tcp_connections.append((hostname, port, client))
