# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from future import standard_library

standard_library.install_aliases()
import os
from datetime import datetime
from time import time
from urllib.parse import urlencode

from fixtures import Fixture, TempDir

from zope.component import getUtility

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.ui.lsopenid import set_consumer_factory, reset_consumer_factory

from landscape.lib.apt.package.skeleton import PackageSkeleton
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.usn import USN, CVE, parse_usn
from canonical.landscape.model.package.package import (
    DEB_PACKAGE,
    DEB_NAME_PROVIDES,
    DEB_UPGRADES,
)
from canonical.landscape.testing.lsopenid import FakeConsumerFactory


class PackageFixture(Fixture):
    def setUp(self):
        super(PackageFixture, self).setUp()
        self.market = PackageMarket()

    def get_usn(self, usn_name, release_time=None, cves=None):
        """Return a L{USN} for C{usn_name}, creating one if necessary."""

        if release_time is None:
            release_time = datetime.utcnow()
        store = get_package_store()
        usn = store.get(USN, usn_name)
        if release_time is None:
            release_time = datetime.utcnow()
        if not usn:
            usn = USN(usn_name, "Vulnerability of USN-" + usn_name, release_time)
            store.add(usn)
        if cves is not None:
            usn_id = parse_usn(usn_name)
            for cve_id in cves:
                cve = store.get(CVE, (cve_id, usn_id))
                if not cve:
                    cve = CVE(cve_id, usn_id)
                    store.add(cve)
        return usn

    def create_package(
        self,
        name,
        version,
        relations=(),
        provides=None,
        upgrades=None,
        usn_name=None,
        summary=None,
        description=None,
        cves=None,
        security_pocket=None,
    ):
        """Create a L{Package} for use in tests.

        @param name: The name of the L{Package}.
        @param version: The version of the L{Package}.
        @param relations: Optionally, a list of relations, such as
            C{DEB_UPGRADES} or C{DEB_NAME_PROVIDES}, to add to the package.
        @param usn_name: Optionally, the name of the USN to associate with the
            package.
        @param security_pocket: Whether that package comes from security
        @return: A new L{Package} instance.
        """
        skeleton = PackageSkeleton(DEB_PACKAGE, name, version)
        if summary is None:
            summary = "A summary for %s %s." % (name, version)
        skeleton.summary = summary
        if description is not None:
            skeleton.description = description
        if provides:
            for provided in provides:
                skeleton.add_relation(DEB_NAME_PROVIDES, provided)
        if upgrades:
            skeleton.add_relation(DEB_UPGRADES, upgrades)
        for type, value in relations:
            skeleton.add_relation(type, value)
        hash = self.market.add(skeleton)
        package = self.market.get(hash)
        if usn_name is not None:
            package.usn = self.get_usn(usn_name, cves=cves)
        if security_pocket is not None:
            package.security_pocket = security_pocket

        return package


class RepositoryFixture(TempDir):
    def setUp(self):
        super(RepositoryFixture, self).setUp()
        self.config = getUtility(IFileBasedConfiguration)
        self.config.set("landscape", "repository-path", self.path)
        self.config.set("landscape", "reprepro-binary", "/usr/bin/reprepro")


class TimeFreezeFixture(Fixture):
    """
    Freeze C{time.time} so that it always returns the same time.

    @ivar now: a C{datetime} representing the frozen value for now().
    """

    def __init__(self, mocker):
        self.mocker = mocker

    def setUp(self):
        super(TimeFreezeFixture, self).setUp()
        frozen_time = int(time())
        self.now = datetime.utcfromtimestamp(frozen_time)
        time_mock = self.mocker.replace(time)
        time_mock()
        self.mocker.result(frozen_time)
        self.mocker.count(0, None)
        self.mocker.replay()


class FakeOpenIDConsumerFixture(Fixture):
    def __init__(self, result, return_to_query=None):
        self.result = result
        self.return_to_query = return_to_query

    def setUp(self):
        super(FakeOpenIDConsumerFixture, self).setUp()
        consumer_factory = FakeConsumerFactory()
        auth_request = consumer_factory.add_begin_response("https://login.ubuntu.com/")
        auth_request.add_redirectURL_response(
            "http://localhost/", result="https://provider.example.com/"
        )
        return_to = "http://localhost/login/handle-openid"
        if self.return_to_query:
            return_to += "?" + urlencode(self.return_to_query)
        consumer_factory.add_complete_result(return_to, self.result)
        set_consumer_factory(consumer_factory)
        self.addCleanup(reset_consumer_factory)


class CleanEnvironmentFixture(Fixture):
    """A fixture the cleans os.environ so that it's empty.

    This is used to ensure that environment variables don't affect the
    test run by accident.

    The original environment is restored when the fixture is torn down.
    """

    original_environ = None

    def setUp(self):
        super(CleanEnvironmentFixture, self).setUp()
        self.original_environ = os.environ.copy()
        os.environ.clear()
        self.addCleanup(os.environ.clear)
        self.addCleanup(os.environ.update, self.original_environ)
