# Copyright 2005-2015 Canonical Limited.  All rights reserved.

import os
import inspect

from bs4 import BeautifulSoup

from canonical.landscape.constants import ROOT_DIR, LANDSCAPE_DIR
from canonical.landscape.testing.ui import UITest

_REQUIRES_MARKER = '{"requires":'
_USES_MARKER = "YUI().use("


class JavascriptMarkup(UITest):
    """Generate sample HTML markup for use in Javascript tests.

    Sub-classes must implement a method with a name that starts with the
    prefix "render_", which should setup model data and then invoke render(),
    passing the URL they want to generate markup for.

    @cvar modules: Optional list that sub-classes can fill with the paths
        of the Javascript modules they want to include as dependencies,
        relative to the c/l/javascript directory. If set to None (the default)
        the dependencies will be set automatically by calculating the
        dependency closure of the test module.
    """

    modules = None

    def setUp(self):
        super(JavascriptMarkup, self).setUp()
        # Derive name of HTML from the Python file that contains the
        # rendering class.
        self.base_path = inspect.getfile(self.__class__).split(".")[0]
        self.browser = self.get_browser(url=None, headers=self.get_headers())
        for key, value in self.get_headers().items():
            self.browser.addHeader(key, value)

    def get_features(self):
        """Add the 'javascript-testing' ZCML feature.

        Sample HTML pages need some specific tweaks (like disabling regular
        scripts and making assets paths relative to the sample page), which
        are triggered via ZCML (see c/l/javascript/testing/configure.zcml).
        """
        features = super(JavascriptMarkup, self).get_features()
        features += ("javascript-testing",)
        return features

    def get_headers(self):
        """Return additional request headers for the test browser.

        The additional headers are used to pass to the templating engine
        information like the test file being rendered, and its YUI modules.
        """
        modules = self.modules
        if modules is None:
            modules = _get_dependency_closure(self.base_path + ".js")
        return {
            "X-Javascript-Testing-Base-Path": self.base_path,
            "X-Javascript-Testing-Modules": ",".join(modules),
        }

    def render(self, url):
        """Render the sample HTML file, along with test javascript code."""
        self.browser.open(url)

        soup = BeautifulSoup(self.browser.contents, "html5lib")

        # Base only exists when the page contains relative links.
        base = soup.find("base")
        if base is not None:
            base.replace_with("")

        body = soup.find("body")
        if body.attrs.setdefault("class", ""):
            body.attrs["class"] += " "
        body.attrs["class"] += "yui3-skin-sam"
        path = "{}-auto.html".format(self.base_path)
        with open(os.path.join(ROOT_DIR, path), "wb") as fd:
            fd.write(soup.prettify().encode("utf-8"))


def _get_dependency_closure(test_module):
    """Return the dependency closure of the test module to run.

    This function will look at the YUI.use() declaration in the test module
    and return all direct and indirect module dependencies.
    """
    queue = _get_uses(test_module)
    modules = set()
    while queue:
        module = queue.pop()
        if module not in modules:
            modules.add(module)
            dependencies = _get_requires(module)
            queue |= dependencies - modules
    return modules


def _get_uses(test_module):
    """Return the set of dependencies as found in the YUI.use() call."""
    with open(os.path.join(ROOT_DIR, test_module)) as fd:
        text = fd.read()
    marker_index = text.find(_USES_MARKER)
    if marker_index == -1:
        raise RuntimeError(
            "Test module %s has no 'YUI.use()' declaration" % test_module
        )

    start_index = text.find("(", marker_index) + 1
    end_index = text.find("function(Y)", start_index)

    dependencies = _parse_dependencies(text[start_index:end_index])
    dependencies.remove("testing.js")
    return dependencies


def _get_requires(module):
    """Return the set of dependencies as found in the 'requires' property."""
    with open(os.path.join(LANDSCAPE_DIR, "javascript", module)) as fd:
        text = fd.read()
    marker_index = text.rfind(_REQUIRES_MARKER)
    if marker_index == -1:
        raise RuntimeError("Module %s has no 'requires' declaration" % module)

    start_index = text.find("[", marker_index) + 1
    end_index = text.find("]", start_index)

    return _parse_dependencies(text[start_index:end_index])


def _parse_dependencies(text):
    """Convert a string of dependencies into a list."""
    modules = set()
    for entry in text.split(","):
        entry = entry.strip('" \n')
        if entry.startswith("landscape."):
            modules.add(entry.replace("landscape.", "").replace(".", "/") + ".js")
    return modules
