# Copyright 2005-2022 Canonical Limited. All rights reserved.

import json
import os
import shutil
import sys
import tempfile
import time
import re
from datetime import datetime, timedelta
from unittest import mock
from unittest.case import _AssertRaisesContext

import transaction

from fixtures import EnvironmentVariableFixture

from gnupg import GPG

from twisted.python.failure import Failure

from zope.security.proxy import removeSecurityProxy

from canonical.lib.timestamp import utctoday
from canonical.lib.text import make_secure_id


class StoresTestMixin(object):

    alt_resource_database_key = 2

    def commit(self):
        transaction.commit()

    def abort(self):
        transaction.abort()

    def get_main_store(self):
        return self.zstorm.get("main")

    def get_account_store(self):
        return self.zstorm.get("account-1")

    def get_alt_account_store(self):
        return self.zstorm.get("account-2")

    def get_resource_store(self):
        return self.zstorm.get("resource-1")

    def get_alt_resource_store(self):
        return self.zstorm.get("resource-2")

    def get_package_store(self):
        return self.zstorm.get("package")

    def get_session_store(self):
        return self.zstorm.get("session")

    def get_session_autocommit_store(self):
        return self.zstorm.get("session-autocommit")

    def get_knowledge_store(self):
        return self.zstorm.get("knowledge")


class AssertionsTestMixin(object):
    def assertRaises(self, excClass, callableObj=None, *args, **kwargs):
        # Emulate the standard library's assertRaises which can return a
        # context manager.
        if callableObj is None:
            return _AssertRaisesContext(excClass, self)

        try:
            result = callableObj(*args, **kwargs)
        except excClass as exception:
            # Unlike the stdlib assertRaises, this version returns the
            # exception for (optional) inspection by the test.
            return exception
        except:
            self.fail(
                "%s raised instead of %s:\n %s"
                % (sys.exc_info()[0], excClass.__name__, Failure().getTraceback())
            )
        else:
            self.fail("%s not raised (%r returned)" % (excClass.__name__, result))

    # FIXME This should be removed and replaced with code that is
    # deterministic.  Code relying on this assertion may fail if, for
    # example, more than 1 second elapses between the time a value is
    # set and the time this assertion is called.
    def assertAlmostNow(self, value, tolerance=timedelta(seconds=1), msg=None):
        """Ensure that a L{datetime} is very close to the current UTC time.

        The assertion ensures that C{value} is within L{tolerance}
        seconds of C{datetime.utcnow}.  If provided, C{msg} is used
        when the assertion fails.
        """
        self.assertApproximates(value, datetime.utcnow(), tolerance, msg)

    def assertElementsAlmostEqual(self, first_list, second_list):
        """
        Use C{assertAlmostEqual} on each corresponding couple of element of
        two lists, if they are float.
        """
        self.assertEqual(len(first_list), len(second_list))
        for first, second in zip(first_list, second_list):
            if isinstance(first, float) and isinstance(second, float):
                self.assertAlmostEqual(first, second)
            else:
                self.assertEqual(first, second)

    def assertSQLIn(self, expected_statement, all_statements):
        """Ensure that C{expected_statement} is in C{all_statements}.

        Whitespace is completely collapsed before statements are compared.
        """
        self._assertSQL(self.assertIn, expected_statement, all_statements)

    def assertSQLNotIn(self, expected_statement, all_statements):
        """Ensure that C{expected_statement} is not in C{all_statements}.

        Whitespace is completely collapsed before statements are compared.
        """
        self._assertSQL(self.assertNotIn, expected_statement, all_statements)

    def _assertSQL(self, func, expected_statement, all_statements):
        regexp = re.compile("\s")

        def crush(statement):
            statement = regexp.sub("", statement)
            return statement

        func(
            crush(expected_statement),
            [crush(statement) for statement in all_statements],
            "didn't pass: %s(%s, %s)" % (func, expected_statement, all_statements),
        )


class LicenseTestMixin(object):
    def setup_gnupg_environment(self, use_trustdb=False):
        """Set up the gnupg environment so that the test key rings are used.

        @return: the GNUPGHOME path

        """
        from canonical.landscape.constants import ROOT_DIR

        tempdir = tempfile.TemporaryDirectory()
        self.addCleanup(tempdir.cleanup)
        gpg_path = tempdir.name
        test_path = os.path.join(ROOT_DIR, "canonical", "landscape", "testing")
        gpg_files = ["pubring.gpg", "secring.gpg"]
        if use_trustdb:
            gpg_files.append("trustdb.gpg")
        for gpg_file in gpg_files:
            shutil.copy(
                os.path.join(test_path, gpg_file), os.path.join(gpg_path, gpg_file)
            )

        # Use the key rings set up for testing.
        self.useFixture(EnvironmentVariableFixture("LANDSCAPE_GPG_HOME", gpg_path))
        return gpg_path

    def sign_data(self, content, key="43FF61AC", use_trustdb=False):
        """Sign the given content."""
        gpg_path = self.setup_gnupg_environment(use_trustdb=use_trustdb)
        gpg = GPG(gnupghome=gpg_path)
        if gpg.version[0] == 1:
            return str(gpg.sign(content, keyid=key))

        # don't use keyid, as --default-key does not fail on gpg2+
        return str(gpg.sign(content, extra_args=["--local-user", key]))

    def shift_todays_date(self, offset_in_days=365):
        """Return today's date shifted by the given offset.

        @param offset_in_days: The number of days the result should differ
            from today's date.  Use negative offsets for dates in the past.
        @return: A C{date} instance with only the date parts set.
        """
        return utctoday() + timedelta(days=offset_in_days)

    def prepare_license_file(
        self,
        licensee="onward",
        role="BasicFeatures",
        seats=789,
        expires=None,
        json_text=None,
        key="43FF61AC",
        use_trustdb=False,
        license_key=None,
    ):
        """Return the signed license file contents with the data given.

        @param licensee: The name of the licensee in the LDS license file.
        @param role: The name of the role in the LDS license file.
        @param seats: The number of seats in the LDS license file.
        @param expires: The expiration date in the LDS license file.
        """
        if json_text:
            signed_json = self.sign_data(json_text, use_trustdb=use_trustdb)
        else:
            if not expires:
                expires = self.shift_todays_date()
            keys = ("licensee", "role", "seats", "expires", "licenses")
            license_key_string = license_key or make_secure_id()
            license_key_string = license_key_string
            values = (
                licensee,
                role,
                seats,
                expires.isoformat(),
                [
                    {
                        "role": role,
                        "seats": seats,
                        "expires": expires.isoformat(),
                        "license_key": license_key_string,
                    }
                ],
            )
            signed_json = self.sign_data(
                json.dumps(dict(list(zip(keys, values)))), key, use_trustdb=use_trustdb
            )
        self.mock_license_path_and_key(signed_json)
        return signed_json

    def mock_signing_key(self):
        fingerprint = "E7654CDF9E45F30423FE7A3075B0EB6143FF61AC"

        patch = mock.patch(
            "canonical.landscape.model.main.license." + "get_signing_key_fingerprint",
            new=mock.Mock(return_value=fingerprint),
        )

        patch.start()
        self.addCleanup(patch.stop)

    def mock_license_path_and_key(self, signed_json):
        """
        Prepare the ground so that C{create_license_file} can be invoked
        with success.

        This involves mocking the C{get_license_file_paths} and the
        C{get_signing_key_fingerprint} functions.
        """
        from canonical.landscape.model.main.license import get_license_file_paths

        self.default_license_path = self.makeFile(content=signed_json)
        license_path_mock = self.mocker.replace(get_license_file_paths)
        self.expect(license_path_mock()).result([self.default_license_path]).count(
            0, None
        )
        self.mock_signing_key()
        self.mocker.replay()

    def initialize_license(
        self,
        licensee="onward",
        role="BasicFeatures",
        seats=789,
        expires=None,
        license_key=None,
    ):
        """Prepare a LDS license observing the data passed.

        @param licensee: The name of the licensee in the LDS license file.
        @param role: The name of the role in the LDS license file.
        @param seats: The number of seats in the LDS license file.
        @param expires: The expiration date in the LDS license file.
        @return: the L{LicenseFile} instance with the matching data.
        """
        from canonical.landscape.model.main.license import create_license_file

        self.prepare_license_file(
            licensee, role, seats, expires, license_key=license_key
        )
        return create_license_file()

    def setup_missing_license(self, include_dir=None):
        """
        Make sure the LDS license is not found even though the
        /etc/landscape/license.txt file exists in the file system.
        """
        from canonical.landscape.model.main.license import get_license_file_paths

        self.setup_gnupg_environment()
        self.default_license_path = "/this-path-should-not-exist-%s" % str(time.time())
        paths = [self.default_license_path]
        if include_dir:
            paths.append(include_dir)
        license_path_mock = self.mocker.replace(get_license_file_paths)
        self.expect(license_path_mock()).result(paths).count(0, None)
        self.mocker.replay()

    def expire_all_licenses(self, computer):
        """Invalidate all licenses for L{computer}."""
        yesterday = datetime.utcnow() - timedelta(days=1)
        insecure_computer = removeSecurityProxy(computer)
        for insecure_license in insecure_computer.licenses:
            insecure_license.expiration_date = yesterday
