# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
from datetime import timedelta

from testresources import ResourcedTestCase
from fixtures import EnvironmentVariableFixture

from zope.component import getUtility
from zope.security.proxy import ProxyFactory

from canonical.testing.fixtures import UtilityFixture
from canonical.testing.unit import UnitTest
from canonical.testing.runner import get_worker
from canonical.testing.resources import (
    ConfigurationContextResourceManager,
    CoreZcmlResourceManager,
)
from canonical.testing.twist import TwistedTestCase
from canonical.security.interfaces import IRoleRegistry
from canonical.security.role import Role, RoleRegistry
from canonical.lib.timestamp import utctoday

from canonical.landscape.security.roles import setup_roles
from canonical.landscape.security.grants import setup_grants
from canonical.landscape.mail.mail import landscape_mailer
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.testing.fakes import FakeMailer, FakeMessageAPI, FakeFetch
from canonical.landscape.testing.resources import (
    ZStormResourceManager,
    DispatcherResourceManager,
    SetupResourceManager,
    ModelZcmlResourceManager,
)
from canonical.landscape.testing.fixture import PackageFixture, TimeFreezeFixture
from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.main.account import create_account
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.main.person import create_person, create_support_person
from canonical.landscape.model.configuration import (
    LANDSCAPE_CONFIG,
    get_root_url,
    set_root_url,
)
from canonical.landscape.model.activity.testing import StubInfo, StubRequest
from canonical.landscape.model.package.interfaces import IPackageSearchClient
from canonical.landscape.model.package.client import PackageSearchClient
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.testing.login import login, logout
from canonical.landscape.testing.mixins import (
    StoresTestMixin,
    AssertionsTestMixin,
    LicenseTestMixin,
)


class ModelTest(AssertionsTestMixin, StoresTestMixin, LicenseTestMixin, UnitTest):

    with_account = False
    with_person = False
    with_license = False  # implies with_account
    with_computer = False  # implies with_license
    with_stub_request = False  # implies with_computer

    with_zcml = False
    with_standalone = False  # implies with_zcml
    with_root_url = False
    with_mailer = False
    with_alert_rule_policy = False
    with_market = False
    with_message_api = False  # implies with_computer

    dispatcher_manager = DispatcherResourceManager()
    setup_manager = SetupResourceManager()
    configuration_context_manager = ConfigurationContextResourceManager()
    core_zcml_manager = CoreZcmlResourceManager()
    model_zcml_manager = ModelZcmlResourceManager()

    account = None
    computer = None
    license = None
    person = None

    person_password = "passwd"

    _zstorm_managers = {}  # Class attribute, common to all instances

    # Hack to not clean up resources in tearDown(). The tearDown() isn't
    # called when setUp() fails (e.g. when creating test data), which
    # can cause all following tests to fail with a db error. Instead of
    # cleaning up the resources in tearDown() we add a cleanup function
    # after setting up the resources, which will be run also if setUp
    # fails. This can be removed when bug 1010379 is fixed.
    realTearDownResources = ResourcedTestCase.tearDownResources

    def tearDownResources(self):
        pass

    def setUp(self):
        self.resources = self.get_resources()
        super(ModelTest, self).setUp()
        self.addCleanup(self.realTearDownResources)

        if self.with_standalone:
            self.setup_standalone()
        if self.with_root_url:
            self.setup_root_url()
        if self.with_mailer:
            self.setup_mailer()
        if self.with_alert_rule_policy:
            self.setup_alert_rule_policy()
        if self.with_account:
            self.create_account()
        if self.with_license:
            self.create_license()
        if self.with_computer:
            self.create_computer()
        if self.with_stub_request:
            self.create_stub_request()
        if self.with_message_api:
            self.create_message_api()
        if self.with_person:
            self.create_person()
        if self.with_market:
            self.setup_market()

    def tearDown(self):
        super(ModelTest, self).tearDown()
        self.account = None
        self.license = None
        self.computer = None

    def get_resources(self):
        """Get the needed resources according to the requested features."""
        resources = []

        if self.with_zcml:
            resources += [
                ("context", self.configuration_context_manager),
                ("core", self.core_zcml_manager),
                ("model", self.model_zcml_manager),
            ]

        resources += [
            ("setup", self.setup_manager),
            ("dispatcher", self.dispatcher_manager),
            ("zstorm", self.get_zstorm_manager()),
        ]
        return resources

    def get_zstorm_manager(self):
        """Return the L{ZStormResourceManager} instance for this test."""
        zstorm_user = "landscape"
        group = self._get_group()
        zstorm_manager_key = "%s-%s" % (zstorm_user, group)
        if zstorm_manager_key not in self._zstorm_managers:
            self._zstorm_managers[zstorm_manager_key] = ZStormResourceManager(
                user=zstorm_user, group=group
            )
        return self._zstorm_managers[zstorm_manager_key]

    @property
    def file_based_config(self):
        return getUtility(IFileBasedConfiguration)

    @property
    def role_registry(self):
        return getUtility(IRoleRegistry)

    def setup_role(
        self, role_name, context_name, permissions, person_email, computer_name=None
    ):
        """
        Quickly set up a new in-memory role, with:
        - a new access context
        - the specified (micro) permissions
        - a new person with the given email
        - optionally, a new computer associated with it.

        @return: A tuple of (L{Role}, L{AccessContext}, L{Person},
            L{Computer}).
        """
        # completely arbitrary role key. TODO: Make it dynamic, and allow
        # multiple setup_role() calls by incrementing the key.
        role = Role(role_name, key=90000)
        self.addCleanup(self.role_registry.remove, role)
        self.role_registry.add(role)
        context = self.account.create_access_context(context_name, context_name)
        if computer_name:
            computer = self.account.create_computer(
                computer_name, computer_name, access_context=context
            )
        else:
            computer = None
        person = create_person("RoleUser", person_email, "secret")
        for permission in permissions:
            self.role_registry.grant(role, "landscape." + permission)
        self.account.add_person(person, roles=[role_name], access_context_id=context.id)
        return (role, context, person, computer)

    def create_account(self):
        """Create a sample account."""
        account_name = "standalone" if self.with_standalone else "account"
        self.account = create_account(account_name, "Account")
        self.license_management = LicenseManagement(self.account)

    def create_license(self):
        """Create a sample license."""
        if self.account is None:
            self.create_account()
        self.license = self.license_management.create_license(
            "BasicFeatures", 10, expiration_date=utctoday() + timedelta(days=365)
        )

    def create_computer(self):
        """Create a sample computer."""
        if self.license is None:
            self.create_license()
        self.computer = self.account.create_computer(
            "sid", "Computer", hostname="testhostname"
        )

    def create_stub_request(self):
        """Create a sample computer request."""
        if self.computer is None:
            self.create_computer()
        self.parent = self.account.create_activity(StubInfo)
        self.activity = self.parent.add(StubRequest, computer=self.computer)
        self.parent.settle()
        self.info = self.activity.info

    def create_message_api(self):
        """Create a fake MessageAPI for C{self.computer}."""
        if self.computer is None:
            self.create_computer()
        self.message_api = FakeMessageAPI(ProxyFactory(self.computer))

    def create_person(self):
        """Create a sample person, adding it to the account if available."""
        self.person = create_person(
            "Person", "person@example.com", self.person_password
        )
        if self.account is not None:
            self.account.add_person(self.person, roles=["Owner"])
            # This is created in the account; some methods of LicenseManagement
            # need the person to be set.
            self.license_management.person = self.person

    def create_access_context_owner(self):
        """
        Return a binary tuple with a new L{Person} and a new L{AccessContext}.

        The returned person is a member of C{self.account} and is granted the
        Owner role on the returned access context.
        """
        person = create_person("OtherPerson", "person@support.com", "pwd")
        access_context = self.account.create_access_context(
            "some-context", "Some context"
        )
        self.account.add_person(
            person, roles=["Owner"], access_context_id=access_context.id
        )
        return person, access_context

    def create_support_only_person(self):
        """Create a sample person with the SupportUser role in the account."""
        support_only_person = create_person(
            "OtherPerson", "person@support.com", self.person_password
        )
        self.account.add_person(support_only_person, roles=["SupportUser"])
        return support_only_person

    def create_support_provider(self):
        """Create a sample support provider person."""
        support_provider = create_support_person(
            "SupportPerson", "person@canonical.com", self.person_password
        )
        return support_provider

    def create_kb_reviewer(self):
        """Create a sample knowledge base reviewer person."""
        kb_reviewer = create_person(
            "Kb Reviewer", "kb_reviewer@xample.com", self.person_password
        )
        kb_reviewer.add_global_role("KnowledgeBaseReviewer")
        return kb_reviewer

    def setup_standalone(self):
        """Setup global state to be LDS-like."""
        fixture = EnvironmentVariableFixture(LANDSCAPE_CONFIG, "standalone")
        self.useFixture(fixture)
        # Replace the regular registry with one having LDS-specific settings
        registry = RoleRegistry()
        setup_roles(registry=registry)
        setup_grants(registry=registry)
        self.useFixture(UtilityFixture(registry))

    def setup_root_url(self):
        """Setup the root url to 'http://localhost/'.

        It can be enabled automatically by setting with_root_url to True.
        """
        self.useFixture(EnvironmentVariableFixture("LANDSCAPE_ROOT_URL"))
        orig_root_url = get_root_url()
        set_root_url("http://localhost/")
        self.addCleanup(set_root_url, orig_root_url)

    def setup_mailer(self):
        """Setup a mock mailer."""
        self.mailer = FakeMailer()
        self.useFixture(UtilityFixture(self.mailer, name="landscape-mailer"))
        self.useFixture(UtilityFixture(landscape_mailer, ILandscapeMailer))

    def setup_market(self):
        """Setup package market."""
        self.package_fixture = self.useFixture(PackageFixture())
        self.market = self.package_fixture.market
        self.package_search_client = PackageSearchClient(
            "localhost", 9099, fetch=FakeFetch()
        )
        self.useFixture(
            UtilityFixture(self.package_search_client, provided=IPackageSearchClient)
        )

    def freeze_time(self):
        """Always return the same value for L{time.time}."""
        return self.useFixture(TimeFreezeFixture(self.mocker)).now

    def login(self, whom=None):
        self.addCleanup(self.logout)
        return login(whom)

    def logout(self):
        logout()

    def _get_group(self):
        """Return the database suffix this model test should use."""
        if os.environ.get("LANDSCAPE_TEST_SELENIUM"):
            return "selenium"
        worker = get_worker()
        if worker is None:
            worker = 0
        return "group-layer-%d" % worker

    def remove_account_computer_ids(self, account, computer_ids):
        """Remove computers from the specified account."""
        selection = ComputerSelection(self.account)
        selection.select_computers(computer_ids)
        management = ComputerManagement(selection, self.person)
        management.remove_computers()


class TwistedModelTest(ModelTest, TwistedTestCase):
    """
    A L{ModelTest} supporting C{Deferred}s.
    """

    _parents = []

    def run(self, result=None):
        TwistedTestCase.run(self, result=result)

    def setUp(self):
        TwistedTestCase.setUp(self)
        ModelTest.setUp(self)

    def tearDown(self):
        ModelTest.tearDown(self)
        TwistedTestCase.tearDown(self)
