# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import random
import shutil
import tempfile
import time
import transaction

from testresources import TestResourceManager

from zope.component import provideUtility, getUtility
from zope.component.hooks import setHooks, resetHooks, setSite
from zope.configuration.xmlconfig import file

from storm.locals import create_database, Store
from storm.zope.testing import ZStormResourceManager as _ZStormResourceManager
from storm.twisted.transact import RETRIABLE_ERRORS

from canonical.testing.wsgiopenid import OpenIdWSGIFetcher
from canonical.testing.resources import ZcmlResourceManager, zcml_extension

from canonical.landscape.schema import (
    main,
    account,
    resource,
    package,
    session,
    knowledge,
)
from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.testing.fakes import FakeDispatcher
from canonical.landscape.environment import get_config_file_path
from canonical.landscape.model.setup import setup
from canonical.landscape.model.root.landscape import create_root


class ZStormResourceManager(_ZStormResourceManager):
    """Provide a L{ZStorm} resource configured with the Landscape databases.

    @param user: The database user to connect as.
    @param group: The database group to use, a different set of databases is
        used for each layer.
    """

    vertical_patching = False
    clean_attempts = 1

    def __init__(self, user="landscape", group="test"):
        self._group = group
        self.schema_stamp_dir = "/tmp/landscape-zstorm-cache-%s" % group
        super(ZStormResourceManager, self).__init__(self._get_databases(user))

    def clean(self, resource):
        """Retry serialization errors if clean_attempts is greater than 1.

        XXX This feature should be probably pushed upstream.
        """
        attempt = 1
        while True:
            try:
                super(ZStormResourceManager, self).clean(resource)
            except RETRIABLE_ERRORS:
                transaction.abort()
                if attempt < self.clean_attempts:
                    # We haven't exhausted the attempts, let's try harder
                    time.sleep(random.uniform(1, 2 ** attempt))
                    attempt += 1
                else:
                    raise
            else:
                break

    def _get_databases(self, user):
        """Get the database data needed by the L{ZStormResourceManager}.

        @return: A C{dict} mapping L{ZStorm} database names to their URI and
        L{Schema} object.
        """
        schemas = [
            ("main", main),
            ("account-1", account),
            ("account-2", account),
            ("resource-1", resource),
            ("resource-2", resource),
            ("package", package),
            ("session", session),
            ("session-autocommit", None),  # Don't create schema twice
            ("knowledge", knowledge),
        ]

        uri_template = "postgres://%(user)s:landscape@localhost/%(database)s"
        schema_uri_template = "postgres:%(database)s"

        postgres = Store(create_database("postgres:postgres?" "isolation=autocommit"))
        existing = [
            row[0] for row in postgres.execute("SELECT datname FROM pg_database")
        ]
        databases = []

        for name, module in schemas:
            options = ""
            database = "landscape-%s-%s" % (self._group, name)
            if database.endswith("-autocommit"):
                database = database[:-11]
                options = "?isolation=autocommit"
            uri = uri_template % {"database": database, "user": user} + options
            if module is None:
                schema = None
                schema_uri = None
            else:
                schema_uri = schema_uri_template % {"database": database}
                schema = module.create_schema()
            databases.append(
                {"name": name, "uri": uri, "schema": schema, "schema-uri": schema_uri}
            )
            if database not in existing:
                postgres.execute('CREATE DATABASE "%s"' % database)
                existing.append(database)
                # Reset the schema stamp dir, so we start fresh.
                if os.path.exists(self.schema_stamp_dir):
                    shutil.rmtree(self.schema_stamp_dir)

        postgres.close()
        return databases


class DispatcherResourceManager(TestResourceManager):
    """Provide a L{SystemConfiguration} resource for test cases."""

    def make(self, dependencies):
        dispatcher = FakeDispatcher()
        provideUtility(dispatcher)
        return dispatcher


class SetupResourceManager(TestResourceManager):

    _setup = False

    def make(self, dependencies):
        if not self._setup:
            setup()
            self._config = getUtility(IFileBasedConfiguration)
            self._setup = True
        self._config.set("landscape", "repository-path", tempfile.mkdtemp())

    def clean(self, resource):
        if os.path.isdir(self._config.get("landscape", "repository-path")):
            shutil.rmtree(self._config.get("landscape", "repository-path"))
        self._config.clear()


class ModelZcmlResourceManager(ZcmlResourceManager):

    _zcml = "ftesting-model.zcml"


class WSGICommonResourceManager(ZcmlResourceManager):
    """Load the Zope app and Canonical extensions zcml needed by the UI.

    They are common between standalone and non-standalone setups.
    """

    _zcml = "ftesting-ui-common.zcml"
    _init = False

    def make(self, dependencies):
        super(WSGICommonResourceManager, self).make(dependencies)
        if not self._init:
            from zope.session.interfaces import IClientIdManager
            from canonical.landscape.ui.session.manager import (
                LandscapeCookieClientIdManager,
            )

            # Register additional global utilities
            provideUtility(LandscapeCookieClientIdManager(), IClientIdManager)

            # Chameleon cache
            chameleon_cache = "/tmp/landscape-chameleon-cache"
            os.environ["CHAMELEON_CACHE"] = chameleon_cache
            if not os.path.exists(chameleon_cache):
                os.makedirs(chameleon_cache)

            self._init = True

        # Call setHooks() so that the ZCML loaded by WSGIResourceManager
        # instances with different features will populate their root
        # object's registry instead of the global one.
        setHooks()

    def clean(self, resource):
        resetHooks()


class WSGIResourceManager(TestResourceManager):
    """
    Load the UI zcml and build the WSGI Zope app, for use in the test browser.

    @param features: A list of ZCML features that will change the site
        configuration, e.g. 'standalone'.
    """

    _wsgi = None
    _root = None
    _zcml = "ftesting-ui.zcml"

    def __init__(self, features=()):
        super(WSGIResourceManager, self).__init__()
        self._features = features

    def make(self, dependencies):
        if not self._root:
            self._root = create_root()

        # Together with the setHooks call later on, this will make us use
        # the root object's registry instead of the global one.
        setSite(self._root)

        if not self._wsgi:
            from canonical.landscape.ui.wsgi import create_zope_app

            # Load UI's ZCML, re-using the existing configuration context
            path = get_config_file_path(self._zcml)
            context = zcml_extension(features=self._features)
            file(path, context=context)

            config = {
                "devmode": True,
                "root": self._root,
                "root-url": "http://localhost/",
            }
            self._wsgi = self._get_wsgi_wrapper(create_zope_app(config))

        return self._wsgi

    def clean(self, resource):
        setSite(None)

    def _get_wsgi_wrapper(self, wsgi):
        def wsgi_wrapper(environ, start_response):
            from openid import oidutil
            from openid.fetchers import setDefaultFetcher, getDefaultFetcher

            orig_openid_log = oidutil.log
            orig_openid_fetcher = getDefaultFetcher()
            # Quiet OpenID logging
            oidutil.log = lambda message, level=0: None
            # Set a WSGI-based fetcher for open ID, to be able to use a test
            # OpenID provider without performing real HTTP requests
            setDefaultFetcher(OpenIdWSGIFetcher(wsgi))

            # Persist changes performed by test code before the WSGI request
            # is executed.
            transaction.commit()
            result = wsgi(environ, start_response)

            oidutil.log = orig_openid_log
            setDefaultFetcher(orig_openid_fetcher)

            # Restore the site
            setSite(self._root)
            return result

        return wsgi_wrapper
