# Copyright 2005-2022 Canonical Limited. All rights reserved.

from email.message import Message
from http.cookiejar import CookieJar, Cookie
from io import StringIO
from urllib.response import addinfourl

from zope.event import notify
from zope.component import getUtility, queryUtility, provideUtility
from zope.publisher.skinnable import setDefaultSkin
from zope.session.interfaces import IClientIdManager
from zope.testbrowser.browser import HostNotAllowed

from z3c.etestbrowser.testing import ExtendedTestBrowser as Browser

from canonical.testing.resources import (
    SystemZcmlResourceManager,
    PublicationZcmlResourceManager,
)
from canonical.content.form import set_form_token

from canonical.landscape.ui.lsopenid import (
    ILandscapeOpenIDFactory,
    LandscapeOpenIDFactory,
)
from canonical.landscape.testing.model import ModelTest
from canonical.landscape.testing.resources import (
    WSGICommonResourceManager,
    WSGIResourceManager,
)
from canonical.landscape.security.principals import PersonPrincipal, AnonymousPrincipal
from canonical.landscape.security.interfaces import PrincipalLoginEvent
from canonical.landscape.testing.request import TestRequest

_wsgi_managers = {}


class AccountContext(object):
    """A convenience class for testing pages that require a context object
    that has an account attribute.
    """

    def __init__(self, account):
        self.account = account


class UITest(ModelTest):
    """Base class for UI tests."""

    with_wsgi = False  # Implies with_zcml, with_root_url
    with_openid = False

    system_zcml_manager = SystemZcmlResourceManager()
    publication_zcml_manager = PublicationZcmlResourceManager()
    wsgi_common_manager = WSGICommonResourceManager()

    _default = object()  # Marker for method parameters

    def setUp(self):
        super(UITest, self).setUp()
        if self.with_wsgi and not self.with_root_url:
            self.setup_root_url()
        if self.with_openid:
            self.setup_openid()

    def setup_openid(self):
        """
        Setup a Landscape OpenID utility that uses test WSGI OpenID provider.
        """
        openid_login = "http://localhost/openid"
        openid_logout = "http://localhost/openidlogout"
        openid_factory = LandscapeOpenIDFactory(openid_login, openid_logout)
        original = queryUtility(ILandscapeOpenIDFactory)
        self.addCleanup(provideUtility, original, ILandscapeOpenIDFactory)
        provideUtility(openid_factory, ILandscapeOpenIDFactory)

    def get_resources(self):
        resources = []
        if self.with_wsgi:
            self.with_zcml = True
            resources.append(("system", self.system_zcml_manager))
            resources.append(("publication", self.publication_zcml_manager))
            resources.append(("wsgi_common", self.wsgi_common_manager))
            features = self.get_features()
            if features not in _wsgi_managers:
                _wsgi_managers[features] = WSGIResourceManager(features)
            resources.append(("wsgi", _wsgi_managers[features]))
        # XXX this is not nice but we need to have the dispatcher resource
        # run after the system one, because otherwise the real IDispatcher
        # declared in the zcml will override the fake one
        resources = super(UITest, self).get_resources() + resources
        resources.remove(("dispatcher", self.dispatcher_manager))
        resources.append(("dispatcher", self.dispatcher_manager))
        return resources

    def get_features(self):
        """Return the ZCML features to enable for the test.

        Sub-classes can add specific features to tweak the ZCML configuration.
        """
        features = ()
        if self.with_standalone:
            features += ("standalone",)
        return features

    def get_browser(
        self, person=_default, url=_default, handleErrors=False, headers={}
    ):
        """Return a test L{Browser} instance.

        @param person: A person the browser will be pre-authenticated as. To
            get an unauthenticated browser, pass C{None}.
        @param url: The URL the browser should automatically navigate to. The
            default is to navigate to the home page, pass None to not open any
            page.
        @param handleErrors: Whether the browser should handle errors raised
            inside the WSGI application.
        """
        browser = Browser(wsgi_app=self.wsgi)

        for key, value in headers.items():
            browser.addHeader(key, value)

        # errors even if they're happening in the initial page.
        browser.handleErrors = handleErrors
        if person is self._default:
            person = self.person
        if person:
            authenticate_browser(browser, person)
        if url is self._default:
            url = "/"
        if url:
            if url[0] == "/":
                url = "http://localhost" + url
            browser.open(str(url))
        return browser

    def get_request(
        self, environ=None, person=_default, form=None, method=None, body_instream=None
    ):
        """Get a TestRequest

        @param body_instream: an input stream for the request body.
        @param environ: A C{dict} to update the request's environment with.
        @param person: A L{Person} to set the request's principal to. The
            default is C{self.person}.
        @param form: A C{dict} to update the request's form values with.
        @param method: The HTTP method the request uses.
        """
        request = TestRequest(body_instream=body_instream, environ=environ, form=form)
        if person is self._default:
            person = self.person
        # Build a principal with an unproxied person, for testing
        principal = PersonPrincipal(person) if person else AnonymousPrincipal()
        request.setPrincipal(principal)
        if method:
            request.method = method
            if method == "POST":
                set_form_token(request)
        setDefaultSkin(request)
        return request


def authenticate_browser(browser, person):
    """Setup C{person} as authenticated for the given C{browser}.

    It creates a session on the server side, and makes sure the browser has a
    cookie pointing to that session.
    """
    principal = PersonPrincipal(person)
    request = TestRequest()
    notify(PrincipalLoginEvent(principal, request, False))
    manager = getUtility(IClientIdManager)
    sid = manager.getRequestId(request)

    try:
        browser.open("http://localhost")
    except HostNotAllowed:
        browser.open("http://localhost/login")

    browser.cookies["landscape_session"] = sid
