# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from future import standard_library

standard_library.install_aliases()
import urllib.parse

from zope.security.interfaces import Unauthorized
from zope.cachedescriptors.property import Lazy

from canonical.routes import Redirect
from canonical.security.status import is_disabled
from canonical.url import get_url

from canonical.landscape.application import PACKAGE_VERSION
from canonical.landscape.model.main.license import get_licensee
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.main.standalone import AccountLDSLicenseAPI
from canonical.landscape.model.activity.status import BLOCKED
from canonical.landscape.model.package.activities import (
    ChangePackagesRequest,
    ChangePackageProfilesRequest,
    ChangeUpgradeProfilesRequest,
    UpgradeAllPackagesRequest,
    ReleaseUpgradeRequest,
)
from canonical.landscape.ui.computer.info import CloneNameMixin
from canonical.landscape.ui.page.page import ActionPage
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.sidebar.page import SidebarPage
from canonical.landscape.ui.person.person import BasePersonPageMixin
from canonical.landscape.model.configuration import get_root_url, get_archive_url
from canonical.landscape.service.activity.lookup import ActivityParentCollectionLookup
from canonical.landscape.service.activity.get import PENDING, GetActivityParents


class AccountDashboard(AccountPage, CloneNameMixin):

    title = "Account"
    maximum_pending_computers = 5
    maximum_pending_activities = 5
    maximum_unapproved_activities = 5

    def __init__(self, context, request):
        super(AccountDashboard, self).__init__(context, request)
        if self.is_support_only_user(context):
            raise Redirect(get_url(self.context, self.request) + "/kb")

    @Lazy
    def seats(self):
        return LicenseManagement(self.context).get_seats()

    def get_active_rules(self):
        """Get a list of active L{Rule}s for the current L{Account}."""
        try:
            return list(
                self.context.get_rules(
                    active=True, person=self.request.principal.person
                )
            )
        except Unauthorized:
            return []

    def _get_pending_computers_resultset(self):
        """
        Return a C{storm.store.ResultSet} of computers pending approval
        in this current account.
        """
        account = self.context
        person = self.request.principal.person
        return account.get_pending_computers(person)

    def get_pending_computers(self):
        """Get a list of pending computers for the current L{Account}.

        Respects C{maximum_pending_computers} and returns at most that
        many L{PendingComputer} records.
        """
        try:
            pending_computers = self._get_pending_computers_resultset()
            pending_computers.config(limit=self.maximum_pending_computers)
            return list(pending_computers)
        except Unauthorized:
            return []

    def get_pending_computers_count(self):
        """Get a total list of pending computers for the current L{Account}.

        Ignores C{maximum_pending_computers} and returns a full count.
        """
        try:
            pending_computers = self._get_pending_computers_resultset()
            return pending_computers.count()
        except Unauthorized:
            return 0

    def get_pending_activities(self):
        """
        Get a list of pending activities for the 'pending-activities'
        provider.

        Respects C{maximum_pending_activities}.
        """
        return self._pending_activities.items

    def get_pending_activities_count(self):
        """
        Get a total count of pending activities for the 'pending-activities'
        provider.

        Ignores C{maximum_pending_activities}.
        """
        return self._pending_activities.total

    def get_unapproved_activities(self):
        """
        Get a list of unapproved activities for the 'unapproved-activities'
        provider.

        Respects C{maximum_unapproved_activities}.
        """
        return self._unapproved_activities.items

    def get_unapproved_activities_count(self):
        """
        Get a total count of unapproved activities for the
        'unapproved-activities' provider.

        Ignores C{maximum_unapproved_activities}.
        """
        return self._unapproved_activities.total

    @Lazy
    def all_releases(self):
        """
        Iterable yielding known on prem releases, i.e. supported ones and the
        current one (even if unsupported), sorted by newest first.
        """
        releases = set(self.lds_releases)
        releases.add(self.current_release)
        return sorted(releases, reverse=True)

    @Lazy
    def lds_releases(self):
        """Return a tuple with supported on prem releases."""
        api = AccountLDSLicenseAPI(self.context)
        return tuple(api.get_available_releases())

    @Lazy
    def current_release(self):
        """Return the current on prem release based on the package version."""
        # PACKAGE_VERSION is always of the format YY-MM-<minor-revision>
        # To get on prem major release number we need only the first 5 chars
        return PACKAGE_VERSION[:5]

    @Lazy
    def hosted_account_name(self):
        """Return the account name in hosted, using the licensee."""
        return get_licensee()

    @Lazy
    def newer_release_available(self):
        """Return C{True} if the current release is not the latest one."""
        if not self.current_release or not self.lds_releases:
            return False
        return self.lds_releases[0] > self.current_release

    @Lazy
    def unsupported_release(self):
        return self.lds_releases and self.current_release not in self.lds_releases

    @Lazy
    def _pending_activities(self):
        """Get pending parent activities for the account, honoring RBAC."""
        account = self.context
        collection = self._service_lookup(
            ActivityParentCollectionLookup, context=account
        )
        operation = GetActivityParents(collection, self.request.principal)
        return operation(limit=self.maximum_pending_activities, status=PENDING)

    @Lazy
    def _unapproved_activities(self):
        """Get unapproved parent activities for the account, honoring RBAC."""
        collection = self._service_lookup(
            ActivityParentCollectionLookup, context=self.context
        )
        operation = GetActivityParents(collection, self.request.principal)
        approvable_info_classes = [
            ChangePackagesRequest,
            ChangePackageProfilesRequest,
            ChangeUpgradeProfilesRequest,
            UpgradeAllPackagesRequest,
            ReleaseUpgradeRequest,
        ]
        return operation(
            limit=self.maximum_unapproved_activities,
            status=BLOCKED,
            info_classes=approvable_info_classes,
        )


class HowToRegisterPage(AccountPage):

    title = "How to register a Landscape client"

    def get_root_url(self):
        """
        Return the system configured "Root URL" for this installation.
        """
        return get_root_url()

    def get_http_root_url(self, url=""):
        """
        Return the system configured "Root URL" changed to use "http" rather
        than "https".

        If the root_url is currently None, return ""
        """
        root_url = url or self.get_root_url()
        if root_url is not None:
            parsed_url = list(urllib.parse.urlparse(root_url))
            # We always get a 6 item tuple back from urlparse
            parsed_url[0] = "http"
            return urllib.parse.urlunparse(parsed_url)
        else:
            return ""


class HowToRegisterOlderPage(HowToRegisterPage):

    title = "How to register an older Ubuntu version with Landscape"


class HowToGetDedicatedPage(AccountPage):

    title = "How to get Landscape On Premises"

    def __init__(self, context, request):
        super(HowToGetDedicatedPage, self).__init__(context, request)
        self.api = AccountLDSLicenseAPI(self.context)
        self.license = self.api.get_license()
        if not self.license:
            raise Unauthorized("No Landscape On Premises access")

    def regenerate_apt_key(self):
        """Regenerate APT key credentials."""
        self.license.regenerate_apt_key()
        self.notes.info("A new APT key has been generated.")
        url = "%s/self-hosted" % get_url(self.context, self.request)
        self.request.response.redirect(url)

    def update_release(self):
        """Endpoint updating the current default release for the account."""
        release = self.request.form["release"]
        self.api.set_release(release)
        self.notes.info("The target release has been updated.")
        url = "%s/self-hosted" % get_url(self.context, self.request)
        self.request.response.redirect(url)

    def get_release(self):
        """Retrieve the current default on prem release."""
        return self.api.get_release()

    def get_available_releases(self):
        """Retrieve the list of supported on prem releases."""
        return self.api.get_available_releases()

    def get_archive_url(self):
        """
        Return the archive URL for this L{Account}, including the private APT
        key.
        """
        archive_url = get_archive_url()
        parsed_url = list(urllib.parse.urlparse(archive_url))
        parsed_url[1] = "%s:%s@%s" % (
            self.context.name,
            self.license.apt_key,
            parsed_url[1],
        )
        return urllib.parse.urlunparse(parsed_url)

    def get_license_url(self):
        """
        Return the L{License} URL for this L{Account}, including the private
        APT key.
        """
        root_url = get_root_url()
        parsed_url = list(urllib.parse.urlparse(root_url))
        parsed_url[1] = "%s:%s@%s" % (
            self.context.name,
            self.license.apt_key,
            parsed_url[1],
        )
        if not parsed_url[2].endswith("/"):
            parsed_url[2] += "/"
        parsed_url[2] += "license.txt"
        return urllib.parse.urlunparse(parsed_url)


class AccountExpiredPage(SidebarPage):

    title = "Account Expired"

    def __init__(self, context, request):
        super(AccountExpiredPage, self).__init__(context, request)
        if not self.context.has_expired():
            raise Redirect(get_url(self.context, self.request))


class AccountDisabledPage(AccountPage):

    title = "Account Disabled"

    def __init__(self, context, request):
        super(AccountDisabledPage, self).__init__(context, request)
        if not is_disabled(context):
            raise Redirect(get_url(context, self.request))


class AccountPersonPage(AccountPage, BasePersonPageMixin):

    title = "Settings"


class AccountEditPersonPage(AccountPage, BasePersonPageMixin):

    title = "Edit settings"


class GetActiveRulesActionPage(ActionPage):
    """Spawn a job returning which alerts are currently active.

    This action page is hit by the javascript code that refreshes the alert
    menu whenever it reveices an AJAX notification from the server that alert
    rules are changed.

    Note that we need to decouple the notification about the change from the
    fetching of what actually change because alerts change according to the
    priviliges of the person looking at the page.
    """
