# Copyright 2005-2022 Canonical Limited. All rights reserved.

from math import floor

from json import dumps

import transaction

from zope.publisher.browser import BrowserPage
from zope.security.proxy import removeSecurityProxy

from canonical.routes import Redirect
from canonical.url import get_url_object, get_url
from canonical.content.decorator import json

from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.computer.activities import (
    get_next_parent_with_blocked_computer_activity,
)
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.collection import ActivityChildCollection
from canonical.landscape.model.activity.status import (
    ActivityStatusChangeError,
    BLOCKED,
    QUEUED,
    INPROGRESS,
    CANCELED,
    FAILED,
    SUCCEEDED,
    SCHEDULED,
    WAITING,
)
from canonical.landscape.service.collection.errors import InvalidQueryError
from canonical.landscape.service.activity.lookup import ActivityChildCollectionLookup
from canonical.landscape.service.activity.errors import InvalidActivityStatusChangeError
from canonical.landscape.service.activity.get import GetActivityChildren
from canonical.landscape.service.activity.cancel import CancelChildActivities
from canonical.landscape.service.activity.unblock import UnblockChildActivities
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.activity.view import get_activity_view
from canonical.landscape.ui.activity.interfaces import IActivityView
from canonical.landscape.ui.activity.icon import ActivityStatusIcon
from canonical.lib.text import describe_list

PROBLEM = "problem"
WARNING = "warning"
SUCCESS = "success"

APPROVED_QUEUED = object()
APPROVED_INPROGRESS = object()
APPROVED_CANCELED = object()
APPROVED_SUCCEEDED = object()
APPROVED_FAILED = object()

APPROVED_MAP = {
    QUEUED: APPROVED_QUEUED,
    INPROGRESS: APPROVED_INPROGRESS,
    CANCELED: APPROVED_CANCELED,
    SUCCEEDED: APPROVED_SUCCEEDED,
    FAILED: APPROVED_FAILED,
}

LEVEL_MAP = {
    PROBLEM: (FAILED, APPROVED_FAILED),
    WARNING: (BLOCKED,),
    SUCCESS: (
        QUEUED,
        WAITING,
        SCHEDULED,
        CANCELED,
        INPROGRESS,
        SUCCEEDED,
        APPROVED_QUEUED,
        APPROVED_CANCELED,
        APPROVED_INPROGRESS,
        APPROVED_SUCCEEDED,
    ),
}

LEVEL_ORDER = (PROBLEM, WARNING, SUCCESS)

STATUS_WEIGHT = {
    QUEUED: 0,
    SCHEDULED: 0,
    WAITING: 0,
    INPROGRESS: 1,
    APPROVED_QUEUED: 2,
    BLOCKED: 2,
    APPROVED_INPROGRESS: 3,
    CANCELED: 4,
    APPROVED_CANCELED: 4,
    FAILED: 4,
    APPROVED_FAILED: 4,
    SUCCEEDED: 4,
    APPROVED_SUCCEEDED: 4,
}

MAX_WEIGHT = max(STATUS_WEIGHT.values())


class ActivityPageBase(object):
    """A page mixin which helps for rendering information about activities.

    This class can be used to help render and manipulate one or more
    activities.
    """

    title = "Activity"
    interface = IActivityView
    select_all = False

    summary_messages = {
        BLOCKED: "%s pending approval",
        SCHEDULED: "%s scheduled for delivery",
        QUEUED: "%s awaiting delivery",
        INPROGRESS: "%s awaiting results",
        CANCELED: "%s canceled",
        FAILED: "%s failed",
        SUCCEEDED: "%s finished successfully",
        WAITING: "%s waiting for dependent activities",
    }

    had_unapproved_activities = False

    def __init__(self, context, request):
        super(ActivityPageBase, self).__init__(context, request)

        for key in ("cancel", "approve", "revert", "reapply"):
            if key in self.request:
                self.update(key)
                break

        self._generate_flat_tree()

        # This logic used to be in the template, making it overly
        # complex. If the number of columns gets modified in the
        # template, this code will need to be updated as well.
        if self.tree_has_computer:
            self.description_span = self.max_depth + (1 if self.tree_has_status else 2)
        else:
            self.description_span = self.max_depth + (3 if self.tree_has_status else 4)
        self.full_span = (
            4
            + self.description_span
            + (1 if self.tree_has_computer else 0)
            + (1 if self.tree_has_status else 0)
        )

        if "select-all-activities" in request:
            self.select_all = True

    def get_activity_view(self, activity):
        return get_activity_view(activity, self.request, self.interface)

    def get_activities_for_tree(self):
        """Activities returned by this method are used to generate the tree."""
        return []

    def get_status_summary(self, levels=None):
        """A verbose status summary with a count of activities for each status."""
        summary_messages = []
        if levels is None:
            levels = LEVEL_ORDER

        # Remap status_summary to combine approved activities into their
        # non-approved counterparts.
        status_summary = self.status_summary.copy()
        for status, approved_status in list(APPROVED_MAP.items()):
            status_summary[status] += status_summary[approved_status]
            status_summary[approved_status] = 0

        # First entry in gets a 'qualifier', but only the first, so we
        # don't repeat it over and over for each state.
        qualifier_added = False
        for level in levels:
            for status in LEVEL_MAP[level]:
                count = status_summary[status]
                if not count:
                    continue
                if not qualifier_added:
                    count = "%s %s" % (count, count > 1 and "activities" or "activity")
                    qualifier_added = True
                summary_messages.append(self.summary_messages[status] % count)
        return describe_list(items=summary_messages, sort=False, quote=False)

    def get_progress_and_overview(self, status_summary=None, total=None):
        """Compute graph size and status to be displayed visually"""
        if status_summary is None:
            status_summary = self.status_summary
        if total is None:
            total = self.activity_count

        # If there are no activities at all, why bother?
        if not total:
            return 0, {"good": 0, "warning": 0, "bad": 0}

        # Current Level should always be the highest level. We walk
        # the levels in reverse order to ensure that.
        count = 0
        for level in reversed(LEVEL_ORDER):
            for status in LEVEL_MAP[level]:
                if status_summary[status]:
                    count += status_summary[status] * STATUS_WEIGHT[status]

        # We use a weighted percentage fill. Depending on which state
        # a certain computer is, it will account for a bigger chunk of
        # the progress.
        percentage = (count / MAX_WEIGHT) / total

        overview = self.activity_view.get_children_overview()

        return int(floor(percentage * 100)), overview

    def _generate_flat_tree(self):
        self.tree_has_approvable = False
        self.tree_has_cancelable = False
        self.tree_has_revertable = False
        self.tree_has_reappliable = False
        self.tree_has_text = False
        self.tree_has_computer = False
        self.tree_has_computer_request = False
        self.tree_has_status = False
        self.max_depth = 0

        self.flat_tree = [
            self.get_activity_view(activity)
            for activity in self.get_activities_for_tree()
        ]
        details_data = {}
        format_date = self.request.principal.person.formatter.format_date

        i = 0
        while i != len(self.flat_tree):
            view = self.flat_tree[i]
            i += 1
            self.flat_tree[i:i] = view.get_children()
            self.max_depth = max(self.max_depth, view.get_depth())
            if view.is_approvable():
                self.tree_has_approvable = True
            if view.is_cancelable():
                self.tree_has_cancelable = True
            if view.is_revertable():
                self.tree_has_revertable = True
            if view.is_reappliable():
                self.tree_has_reappliable = True
            if view.get_text():
                self.tree_has_text = True
            if view.get_computer():
                self.tree_has_computer = True
            if not self.tree_has_computer_request and view.is_computer_request():
                self.tree_has_computer_request = True
            status = view.get_status()
            if status:
                self.tree_has_status = True
            approval_time = view.get_approval_time()
            if approval_time is not None:
                approval_time = format_date(approval_time, with_tz=True)
            delivery_time = view.get_delivery_time()
            if delivery_time is not None:
                delivery_time = format_date(delivery_time, with_tz=True)
            completion_time = view.get_completion_time()
            if completion_time is not None:
                completion_time = format_date(completion_time, with_tz=True)
            details = {
                "approval_time": approval_time,
                "delivery_time": delivery_time,
                "completion_time": completion_time,
                "has_text": bool(view.get_text()),
            }
            details_data[view.get_activity_id()] = details

        self.details_data = dumps(details_data)

        self.tree_has_actionable = (
            self.tree_has_approvable
            or self.tree_has_cancelable
            or self.tree_has_revertable
            or self.tree_has_reappliable
        )

    def _render_operation_notes(self, verb, success_count, notes):
        """Render page notes about successes and failures verb."""
        for note in notes:
            self.notes.warning(note)

        if success_count > 1:
            self.notes.info("%d activities were %s." % (success_count, verb))
        elif success_count == 1:
            self.notes.info("1 activity was %s." % verb)

    def _update_cancel(self, activity_ids, account):
        """
        Cancel the given activity_ids, catching cancellation errors and
        printing corresponding notes.
        """
        # Limit the collection to activities which are authorized & cancelable
        operation = CancelChildActivities(account, self.request.principal)

        all_activities_count = len(activity_ids)
        try:
            result = operation(ids=activity_ids)
        except InvalidActivityStatusChangeError as err:
            notes = err.summary_by_message("canceled", "cancelation")
            old_status_counts = err.old_status_counts
            succeeded_count = len(err.successful_activities)
        else:
            old_status_counts = result.old_status_counts
            succeeded_count = len(result.activities)
            notes = []

        had_inprogress = old_status_counts[INPROGRESS] > 0
        if had_inprogress:
            notes.append(
                "You canceled an activity that was already in progress. "
                "If the affected computer(s) perform the operation, the "
                "status of the activity will be updated accordingly."
            )

        self.had_unapproved_activities = old_status_counts[BLOCKED] > 0

        self._render_operation_notes("canceled", succeeded_count, notes)

        if succeeded_count == all_activities_count:
            transaction.commit()
            redirect_target = result.activities[0].parent
            raise Redirect(get_url(redirect_target, self.request))

    def _update_approve(self, activity_ids, account):
        """
        Approve the given activities.
        """
        operation = UnblockChildActivities(account, self.request.principal)
        try:
            result = operation(ids=activity_ids)
        except InvalidActivityStatusChangeError as err:
            notes = err.summary_by_message("approved", "approval")
            succeeded_count = len(err.successful_activities)
        else:
            notes = []
            succeeded_count = len(result.activities)

        self._render_operation_notes("approved", succeeded_count, notes)
        self.had_unapproved_activities = True

    def _update_revert(self, activity_ids, account):
        """
        Revert the given activities, aborting on the first error encountered.
        """
        collection = ActivityChildCollection(
            account.store, account=account, ids=activity_ids
        )
        activities = list(collection.find())
        creator = self.request.principal.person
        previous_parent = activities[0].parent
        parent = previous_parent.info.clone(creator)
        try:
            for activity in activities:
                child = activity.info.revert(parent)
        except ActivityStatusChangeError:
            transaction.abort()
            self.notes.error("Some selected activities can't be undone.")
        else:
            parent.settle()
            count = len(activities)
            if count == 1:
                msg = "1 activity was created"
            else:
                msg = "%d activities were created" % (count,)
            self.notes.info("%s to undo the changes." % msg)
            transaction.commit()
            redirect_target = parent
            if redirect_target is None:
                redirect_target = child
            raise Redirect(get_url(redirect_target, self.request))

    def _update_reapply(self, activity_ids, account):
        """
        Reapply the given activities, relinking previous attachments and
        aborting on the first error encountered.
        """
        collection = ActivityChildCollection(
            account.store, account=account, ids=activity_ids
        )
        activities = list(collection.find())
        creator = self.request.principal.person
        previous_parent = activities[0].parent
        if not previous_parent:
            parent = None
        else:
            parent = previous_parent.info.clone(creator)
            for attachment in previous_parent.attachments:
                parent.add_attachment(attachment)
        try:
            for activity in activities:
                child = activity.info.reapply(parent)
        except ActivityStatusChangeError:
            transaction.abort()
            self.notes.error("Some selected activities can't be redone.")
        else:
            count = len(activities)
            if count == 1:
                msg = "1 activity was created"
            else:
                msg = "%d activities were created" % (count,)
            self.notes.info("%s to redo the changes." % msg)

            if parent.schedule_after_window:
                parent.schedule(after_window=parent.schedule_after_window)
            parent.settle()
            transaction.commit()

            redirect_target = parent
            if redirect_target is None:
                redirect_target = child
            raise Redirect(get_url(redirect_target, self.request))

    def update(self, action):
        """Update activity statuses using data submitted in a request.

        Activities can be cancelled, or approved. The "activity_ids"
        request argument should hold a list of activity IDs, and the action to
        be taken will be decided by the presence of "cancel", or "approve"
        arguments in the request.
        """
        selected_ids = self.request.get("selected_ids", None)
        if selected_ids:
            activity_ids = [int(value) for value in selected_ids.split(",")]
        elif self.request.get("activity_id", None):
            # Non-group activity
            activity_ids = [self.request.get("activity_id")]
        else:
            activity_ids = ()
        account = get_url_object(self.context, self.request, IAccount)
        if not activity_ids:
            self.notes.error("You must select activities before making " "changes.")
        else:
            account = removeSecurityProxy(account)
            getattr(self, "_update_%s" % (action,))(activity_ids, account)

    @json
    def available_activities_json(self):
        """
        Produce a json list of the available activities, to be displayed by the
        pagination mechanism.
        """
        records = []
        formatter = self.request.principal.person.formatter
        for activity in self.flat_tree:
            record = {
                "activity_id": activity.get_activity_id(),
                "url": activity.get_url(),
                "summary": activity.get_summary(),
                "creation_time": formatter.format_short_date(
                    activity.get_creation_time()
                ),
            }
            status = activity.get_status()
            if activity.is_parent():
                record["overview"] = activity.get_children_overview()
            if status is not None:
                status_icon = ActivityStatusIcon(status, self.request)
                record["status"] = {
                    "icon": status_icon.small(),
                    "title": status_icon.get_path_and_title()[1],
                }
                if self.request.get("detail") and activity.is_computer_request():
                    approval_time = activity.get_approval_time()
                    if approval_time is not None:
                        approval_time = formatter.format_date(
                            approval_time, with_tz=True
                        )
                    delivery_time = activity.get_delivery_time()
                    if delivery_time is not None:
                        delivery_time = formatter.format_date(
                            delivery_time, with_tz=True
                        )
                    completion_time = activity.get_completion_time()
                    if completion_time is not None:
                        completion_time = formatter.format_date(
                            completion_time, with_tz=True
                        )
                    record.update(
                        {
                            "approval_time": approval_time,
                            "delivery_time": delivery_time,
                            "completion_time": completion_time,
                            "expandable": activity.is_expandable(),
                            "has_text": bool(activity.get_text()),
                        }
                    )

            computer = activity.get_computer()
            if computer is not None:
                record["computer"] = {
                    "title": computer.title,
                    "url": get_url(computer, self.request),
                }
            else:
                record["computer_count"] = activity.get_computer_count()
            creator = activity.get_creator()
            if creator is not None:
                record["creator"] = {
                    "name": creator.name,
                    "url": get_url(creator, self.request),
                }
            records.append(record)

        self.request.response.setHeader("Content-Type", "application/json")

        return {"totalRecords": self.total_activities, "records": records}


class ActivityPageMixin(ActivityPageBase):
    """
    A mixin for rendering a page about one activity.

    @ivar next_unapproved_activity: The next activity that is
        unapproved. None if there are no other unapproved activities.
    """

    max_text_size = 4096
    next_unapproved_activity = None
    activity_search_criteria = ""

    def __init__(self, context, request):
        super(ActivityPageMixin, self).__init__(context, request)
        # The context is actually a CopyCat, so it's not possible to check if
        # it's a parent with isinstance.
        parent = getattr(self.context, "parent", self.context)
        self.namespaces.append(
            "landscape.account.%d.activity.%d" % (parent.account.id, parent.id)
        )
        self.activity_view = self.get_activity_view(context)
        status = self.activity_view.get_status()
        # XXX Compare by status id since get_status() might return a copy of
        # BLOCKED (with same id but different title).
        show_unapproved_navigation = self.had_unapproved_activities or (
            status is not None and status.id == BLOCKED.id
        )
        if show_unapproved_navigation:
            self.next_unapproved_activity = self.get_next_unapproved_activity()

        # XXX we probably don't need a flattened tree anymore, because we
        # support only one-level depth activity groups, for now this will
        # just remove the parent from the list, because we don't want to
        # show it in the paginated table of the children.
        self.flat_tree_with_root = self.flat_tree
        self.flat_tree = self.flat_tree_with_root[1:]

    def get_activities_for_tree(self):
        # The context is actually a CopyCat, so it's not possible to check if
        # it's a parent with isinstance.
        activity = getattr(self.context, "parent", self.context)
        person = self.request.principal.person
        computer_ids = person.get_computer_ids(activity.account)

        # Update the status summary used by the detail view and
        # the total_activities attribute which is used by the
        # available_activities_json method for AJAX requests
        self.available_ids = []
        self.activity_count = 0
        self.status_summary = {}
        for level in LEVEL_ORDER:
            for status in LEVEL_MAP[level]:
                self.status_summary[status] = 0
        self.activity_search_criteria = self.request.get("activity_search_criteria")

        collection = self._service_lookup(
            ActivityChildCollectionLookup, activity.account
        )
        collection = collection.with_parent_id(activity.id)
        if computer_ids is not None:
            collection = collection.with_entities(
                AccountComputerTwin, entity_ids=computer_ids
            )
        operation = GetActivityChildren(collection, self.request.principal)

        try:
            result = operation(limit=None, query=self.activity_search_criteria)
        except InvalidQueryError:
            self.notes.error(
                "Search criteria %s is invalid." % (self.activity_search_criteria,)
            )
            self.total_activities = 0
            return []
        for activity_id, status in result.values(Activity.id, Activity.status):
            self.available_ids.append(activity_id)
            self.activity_count += 1
            self.status_summary[status] += 1
        self.total_activities = len(self.available_ids)
        if not self.total_activities and self.activity_search_criteria:
            self.notes.warning(
                "No activities meeting the given criteria could be found."
            )
        return [activity]

    def get_next_unapproved_activity(self):
        """Return the next unapproved activity.

        See get_next_parent_with_blocked_computer_activity for more
        information. This method exists so that subclasses of
        this class can override it, to provide a more filtered view.
        """
        person = self.request.principal.person
        computer_ids = person.get_computer_ids(self.context.account)
        return get_next_parent_with_blocked_computer_activity(
            removeSecurityProxy(self.context), computer_ids=computer_ids
        )

    def view_safe_activity_search_criteria(self):
        """
        Escape C{activity_search_criteria} so that it's safe for Javascript.
        """
        return self.view_safe_js_value(self.activity_search_criteria)


class AccountActivityPage(ActivityPageMixin, AccountPage):
    """Page which renders an activity without the toolbar."""


class ExpandActivityPage(BrowserPage):
    """Publish the C{expanded_content} of L{IExpandableRequest} activities."""

    def __call__(self):
        self.activity_view = get_activity_view(
            self.context, self.request, IActivityView
        )
        if self.activity_view.is_expandable():
            return self.activity_view.expanded_content()


class ActivityDetailPage(AccountActivityPage):
    """Publish the activity detail info as JSON."""

    @json
    def __call__(self):
        progress, overview = self.get_progress_and_overview()

        children = []
        for view in self.flat_tree_with_root:
            child = {
                "id": view.get_activity_id(),
                "flags": {
                    "approvable": view.is_approvable(),
                    "actionable": view.is_actionable(),
                    "cancelable": view.is_cancelable(),
                    "revertable": view.is_revertable(),
                    "expandable": view.is_expandable(),
                    "reappliable": view.is_reappliable(),
                    "result": bool(view.get_text()),
                },
            }
            status = view.get_status()
            if status is not None:
                child["status"] = {"name": status.name, "title": status.title}
            children.append(child)

        return {
            "summary": {
                "problem": self.get_status_summary(levels=("problem",)),
                "warning": self.get_status_summary(levels=("warning",)),
                "success": self.get_status_summary(levels=("success",)),
            },
            "progress": {"percentage": progress, "overview": overview},
            "children": children,
        }
