# Copyright 2005-2022 Canonical Limited. All rights reserved.

import json

from canonical.url import get_url

from canonical.landscape.model.configuration import is_selenium_environment
from canonical.landscape.model.activity.status import BLOCKED, FAILED
from canonical.landscape.ui.page.page import Page
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.activity.activity import ActivityPageBase
from canonical.landscape.ui.activity.interfaces import ISummaryActivityView
from canonical.content.download import set_csv_download
from canonical.landscape.model.activity.activitycsv import get_csv
from canonical.landscape.service.collection.errors import InvalidQueryError
from canonical.landscape.service.activity.lookup import ActivityParentCollectionLookup
from canonical.landscape.service.activity.get import PENDING, GetActivityParents


class ActivitiesPageMixin(ActivityPageBase):
    """A page which renders multiple top-level activities.

    It is sometimes subclassed to provide custom behavior, but doesn't need to
    be. The context object must provide L{IActivityCollection}.

    This renders information about many activities, and provides a way to get
    to detail pages for individual L{Activity} instances.

    @cvar title: The page title.
    @cvar maximum_activities: The maximum number of top-level L{Activity}
        objects to list on the page.
    @ivar maximum_activity_count: The number of activities after which we
        will display "many" as a total instead of the actual total, for
        performance reasons.
    """

    title = "Activities"
    maximum_activities = 25
    maximum_activity_count = 5000  # see is_selenium_environment check below
    maximum_download = 1000
    interface = ISummaryActivityView
    filters = ["", "pending", "unapproved", "failed"]
    filter = ""
    safe_filter = ""  # This is safe to pass back to the UI's JS
    activity_search_criteria = ""

    def __init__(self, context, request):
        if is_selenium_environment():
            # Reduce the maximum maximum activities to 50 in testing
            # environments. If it weren't for our selenium tests relying on
            # this, I would just monkey-patch it in the tests, but that's not
            # possible because the selenium tests spawn a separate landscape
            # server.
            self.maximum_activity_count = 50

        self.filter = request.get("filter", "")
        self.safe_filter = json.dumps(self.filter)
        super(ActivitiesPageMixin, self).__init__(context, request)

    def download(self):
        """
        Handle a request to download activities as JSON or CSV.
        """
        self.maximum_activities = self.maximum_download
        result = self._get_matching_activities(
            self.request.get("download_filter"),
            self.request.get("download_activity_search_criteria"),
        )
        activities = result.items
        filename = "activities"
        result = []
        for activity in activities:
            result.append(activity)
            if len(activities) < self.maximum_download:
                limit = self.maximum_download - len(activities)
                result.extend(activity.children(limit=limit))
        activities = result
        result = get_csv(activities)
        set_csv_download(self.request.response, filename)
        return result

    def view_safe_activity_search_criteria(self):
        """
        Escape C{activity_search_criteria} so that it's safe for Javascript.
        """
        return self.view_safe_js_value(self.activity_search_criteria)

    def get_filters(self):
        """Retrieve the list of activity filters to display."""
        base_url = get_url(self.context, self.request)
        for filter_type in self.filters:
            data = {
                "href": "",
                "css_class": "",
                "title": filter_type.capitalize() or "All",
                "filter_type": filter_type,
            }
            if filter_type:
                data["href"] = "%s?filter=%s" % (base_url, filter_type)
            if filter_type == self.filter:
                data["css_class"] = "selected-filter"
            yield data

    def _get_matching_activities(self, filter, search):
        """
        Get a L{ResultSet} of activities matching the given filter and search
        criteria.
        """
        self.activity_search_criteria = search
        if filter == "pending":
            result = self.get_pending_activities()
        elif filter == "unapproved":
            result = self.get_unapproved_activities()
        elif filter == "failed":
            result = self.get_failed_activities()
        else:
            result = self.get_activities()
        return result

    def get_activities_for_tree(self, limit=None):
        search = self.request.get("activity_search_criteria")
        try:
            result = self._get_matching_activities(self.filter, search)
        except InvalidQueryError:
            self.notes.error(
                "Search criteria %s is invalid." % (self.activity_search_criteria,)
            )
            self.total_activities = -1
            return []

        total_activities = result.total

        if total_activities > self.maximum_activity_count:
            self.total_activities = -1
        else:
            self.total_activities = total_activities

        return result.items

    def get_activities(self):
        return self._get_activities()

    def get_pending_activities(self):
        return self._get_activities(status=PENDING)

    def get_unapproved_activities(self):
        return self._get_activities(status=BLOCKED)

    def get_failed_activities(self):
        return self._get_activities(status=FAILED)

    def _get_activities(self, **kwargs):
        account = self.context.account
        filters = self._get_collection_filters()
        collection = self._service_lookup(
            ActivityParentCollectionLookup, context=account, **filters
        )
        operation = GetActivityParents(collection, self.request.principal)

        kwargs["limit"] = int(self.request.get("size", self.maximum_activities))
        kwargs["offset"] = int(self.request.get("offset", 0))
        kwargs["query"] = self.activity_search_criteria

        return operation(**kwargs)

    def _get_collection_filters(self):
        """Add additional filters to the activity parents collection."""
        return {}


class DownloadActivitiesPage(ActivitiesPageMixin, Page):
    """Download activities in CSV format."""


class AccountActivitiesPage(ActivitiesPageMixin, AccountPage):
    """
    Page displaying activities in an account.
    """
