# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Activity views control the way activities are rendered in the user
interface.

Different views of L{Activity} graphs need to show differing amounts of
detail.  For example, a summary view may only show root activities, whereas a
detailed view may include all activities.  The adapters in this module wrap
L{IActivity} objects and provide an interface for traversing to their
children.  Some of the adapters allow traversal, other do not.  Adapters can
be registered for specific L{ActivityInfo} to customize the way the tree is
traversed based on the types of nodes encountered during traversal.
"""

from builtins import object
from zope.traversing.interfaces import ITraversable
from zope.interface import implementer
from zope.component import getMultiAdapter
from zope.security import proxy
from zope.publisher.interfaces import NotFound as ZopeNotFound

from canonical.url import get_url
from canonical.content.decorator import json

from canonical.landscape.model.activity.status import BLOCKED
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.activity.computer import ComputerRequest
from canonical.landscape.model.activity.group import ActivityGroup
from canonical.landscape.model.activity.timeline import ActivityChildTimeline
from canonical.landscape.model.activity.legacy import get_activity_child_entity_ids
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.service.activity.lookup import ActivityChildCollectionLookup
from canonical.landscape.service.activity.get import GetActivityChildren
from canonical.landscape.ui.activity.interfaces import IActivityView
from canonical.landscape.ui.service import service_lookup

MIN_OVERVIEW_PERCENTAGE = 5


def get_activity_view(activity, request, interface=IActivityView, parent=None):
    """Return an activity view object providing the given interface."""
    return getMultiAdapter((activity.info, request, interface, parent), interface)


def cached(method):
    """Very simple and fast method result caching mechanism.

    It works only on methods that take no arguments, and is used
    to greatly reduce the costs of calling the same method several
    times; something that is done very frequently on page templates.
    """
    cached_result_attr = "_cached_result_" + method.__name__

    def inner(self):
        try:
            return getattr(self, cached_result_attr)
        except AttributeError:
            result = method(self)
            setattr(self, cached_result_attr, result)
            return result

    return inner


@implementer(IActivityView, ITraversable)
class ActivityViewBase(object):
    """Base class used for custom implementations of an ActivityView"""

    def __init__(self, request):
        self._request = request
        self._depth = 0

    def traverse(self, name, further_path):
        # FIXME Why is this needed?  The traversal test seems to pass
        #       fine without this, but the template was breaking. See
        #       Bug #388788.
        return getattr(self, name)()

    def is_visible(self):
        """Should this item be displayed in the tree?

        If this method returns False the activity should not be included in
        rendered output.  The L{get_children} method should always be called
        to determine whether or not children are rendered, regardless of
        whether this activity is displayed or not.
        """
        return True

    def get_depth(self):
        """Return the number of parents above this view.

        Note that depth is relative to the first view queried. It
        doesn't necessarily reflect the activity hierarchy stored in
        the database.
        """
        return self._depth

    def get_activity_id(self):
        """Return the id of the activity associated with this tree item.

        Non-activity tree items may return None, so the template
        should deal with this case too.
        """
        return None

    def get_creation_time(self):
        """Return the creation time or None if one isn't available."""
        return None

    def get_approval_time(self):
        """Return the approval time or None if one isn't available."""
        return None

    def get_delivery_time(self):
        """Return the delivery time or None if one isn't available."""
        return None

    def get_completion_time(self):
        """Return the completion time or None if one isn't available."""
        return None

    def is_approvable(self):
        """
        Is this item representing a ComputerRequest which may be approved?
        """
        return False

    def is_cancelable(self):
        """
        Is this item representing a ComputerRequest which may be canceled?
        """
        return False

    def is_actionable(self):
        """
        Is this item representing a ComputerRequest which may be acted on?
        """
        return False

    def is_revertable(self):
        """
        Is this item representing a ComputerRequest which may be reverted?
        """
        return False

    def is_reappliable(self):
        """
        Is this item representing a ComputerRequest which may be reapplied?
        """
        return False

    def get_text(self):
        """Return the result text for the underlying activity."""
        return ""

    def get_children(self):
        """
        Return activity views to be represented as subelements of this view.
        """
        return []

    def get_url(self):
        """Return the URL to a detail page for the underlying activity."""
        return None

    def get_creator(self):
        """Return the creator or None if one isn't available."""
        return None

    def get_computer(self):
        """Return the computer or None if one isn't available."""
        return None

    def get_computer_count(self):
        """Return a string representing the attached computers."""
        return None

    def is_scheduled(self):
        """Is this item scheduled for delivery?"""
        return False

    def get_scheduled_time(self):
        """
        Get the time this item is scheduled for, or None if it's not
        scheduled.
        """
        return None

    # Chameleon doesn't supoort traversing function calls like get_xx()/++yy++,
    # so we provide these properties in order to write xx/++yy++ instead.

    @property
    def creation_time(self):
        return self.get_creation_time()

    @property
    def creator(self):
        return self.get_creator()

    @property
    def computer(self):
        return self.get_computer()

    @property
    def scheduled_time(self):
        return self.get_scheduled_time()


class ActivityView(ActivityViewBase):
    """Facade used to encapsulate the properties of an Activity for rendering.

    A class offering an IActivityView interface is obtained through the
    get_activity_view() function.

    This view mechanism makes it possible to customize the way in which an
    activity is rendered in a tree or in an activity page by providing an
    adapter that requires the following interfaces:

      (IMyActivityInfo, IMyBrowserRequest, Interface)

    The interface that must be provided is dependent on where the activity
    is being rendered (e.g. ISummaryActivityView is used in the summary
    page).
    """

    max_children = 25

    def __init__(self, info, request, interface, parent=None):
        super(ActivityView, self).__init__(request)
        # XXX Please do not expose activity or its info directly.  Exposing
        #     them means that the template will depend on database-specific
        #     objects, and we do want to offer non-database views in
        #     the future.  For instance, think about having a single
        #     PackageChanges activity which knows about all the packages in
        #     a single activity (a single database row), but which still
        #     renders to the user as if all these changes were subelements
        #     of the real activity.
        self._info = info
        self._interface = interface
        self._activity = info.activity
        self._timeline = ActivityChildTimeline(self._activity)

        self._parent = parent

        if parent is not None:
            self._depth = parent.get_depth() + 1
        else:
            self._depth = 0

    @cached
    def is_approvable(self):
        """
        Is this item representing a ComputerRequest which may be approved?
        """
        return self._info.is_approvable()

    @cached
    def is_cancelable(self):
        """
        Is this item representing a ComputerRequest which may be canceled?
        """
        if self.is_parent():
            return False
        else:
            return self._activity.info.is_cancelable()

    def is_actionable(self):
        """
        Is this item representing a ComputerRequest which may be acted on?
        """
        return (
            self.is_approvable()
            or self.is_cancelable()
            or self.is_revertable()
            or self.is_reappliable()
        )

    @cached
    def is_revertable(self):
        """
        Is this item representing a ComputerRequest which may be reverted?
        """
        return self._info.is_revertable()

    @cached
    def is_reappliable(self):
        """
        Is this item representing a ComputerRequest which may be reapplied?
        """
        return self._info.is_reappliable()

    @cached
    def is_expandable(self):
        """Can this activity be expanded to show additional information?"""
        return self._info.is_expandable()

    @json
    def expanded_content(self):
        """Retrieve the C{expanded_content} of L{IExpandableRequest}."""
        return proxy.removeSecurityProxy(self._info.expanded_content())

    def get_activity_id(self):
        """Return the id of the activity associated with this tree item."""
        return self._activity.id

    def get_status(self):
        """Return the activity status."""
        # XXX the parent status is still used in the JSON page reporting
        # available activities. Use the legacy status for now.
        if self.is_parent():
            status = self._activity.legacy_status
        else:
            status = self._activity.status

        # XXX If an activity is blocked and it's computer activities (or
        # activity group, which contains computer activities), return a BLOCKED
        # status with the title changed, so that the UI displays still "Need
        # approval".
        is_unapproved_computer_request = (
            proxy.isinstance(self._info, (ComputerRequest, ActivityGroup))
            and status == BLOCKED
        )

        if is_unapproved_computer_request:
            if self.is_approvable():
                status = proxy.removeSecurityProxy(status).copy()
                status.title = "Needs approval"
            else:
                status = proxy.removeSecurityProxy(status).copy()
                status.title = "Blocked"
        return status

    status = property(get_status)

    @cached
    def get_creation_time(self):
        """Return the creation time or None if one isn't available."""
        return self._activity.creation_time

    @cached
    def get_approval_time(self):
        """Return the approval time or None if one isn't available."""
        if self.is_computer_request():
            return self._timeline.get_approval_time()

    @cached
    def get_delivery_time(self):
        """Return the delivery time or None if one isn't available."""
        if self.is_computer_request():
            return self._timeline.get_delivery_time()

    @cached
    def get_completion_time(self):
        """Return the completion time."""
        return self._timeline.get_completion_time()

    @cached
    def get_creator(self):
        """Return the creator or None if one isn't available."""
        return self._activity.creator

    @cached
    def get_computer(self):
        """Return the computer or None if one isn't available."""
        if self.is_computer_request():
            return self._info.computer

    @cached
    def is_computer_request(self):
        return proxy.isinstance(self._info, ComputerRequest) or hasattr(
            self._info, "computer"
        )

    @cached
    def get_summary(self):
        """Return a summary of the underlying activity or activities."""
        return self._info.get_summary()

    @cached
    def get_text(self):
        """Return the result text for the underlying activity."""
        if self.is_parent():
            return ""
        return self._info.get_text() or ""

    @cached
    def get_url(self):
        """Return the URL to a detail page for the underlying activity."""
        return get_url(self._activity, self._request)

    @cached
    def _get_computer_ids(self):
        """
        Helper method returing the computers the current person is allowed to
        see.
        """
        person = self._request.principal.person
        return person.get_computer_ids(self._activity.account)

    @cached
    def get_children(self):
        """Return L{ActivityView}s to be represented as children of this view.

        If the L{Person} accessing the view is not an owner in the L{Account},
        only child L{Activity}s for which the person has C{ViewComputer}
        permission on the activity's L{Computer} are included in the result.
        """
        if not self.is_parent():
            return []

        limit = self.max_children
        offset = int(self._request.get("offset", 0))
        computer_ids = self._get_computer_ids()
        search = self._request.get("activity_search_criteria")

        collection = self._service_lookup(
            ActivityChildCollectionLookup, self._activity.account
        )
        collection = collection.with_parent_id(self._activity.id)
        if computer_ids is not None:
            collection = collection.with_entities(
                AccountComputerTwin, entity_ids=computer_ids
            )
        operation = GetActivityChildren(collection, self._request.principal)
        result = operation(limit=limit, offset=offset, query=search)
        return [
            get_activity_view(child, self._request, self._interface, self)
            for child in result
        ]

    @cached
    def is_scheduled(self):
        """Is this item scheduled for delivery?"""
        if self.is_parent():
            return False

        return bool(self._info.activity.schedule_after_time)

    @cached
    def get_scheduled_time(self):
        """
        Get the time this item is scheduled for, or None if it's not
        scheduled.
        """
        if not self.is_parent():
            return self._info.activity.schedule_after_time

    def is_parent(self):
        """
        Is this item representing a ComputerRequest which may be approved?
        """
        return proxy.isinstance(self._activity, ActivityParent)

    @cached
    def get_children_overview(self):
        """Get an overview of the status of the various children.

        @return: a dict with three keys 'good', 'warning' and 'bad', each one
            mapping to the percentage of children with a status in that
            category.
        """
        if not self.is_parent():
            return None
        activity = self._activity
        total = activity.children_count
        good = activity.succeeded_count
        bad = activity.failed_count + activity.canceled_count
        warning = activity.blocked_count
        status = {
            "good": good,
            "warning": warning,
            "bad": bad,
        }
        for key, value in status.items():
            if value > 0:
                percentage = float(value) / total * 100
                status[key] = max(percentage, MIN_OVERVIEW_PERCENTAGE)
        return status

    def _service_lookup(self, lookup_class, *args, **kwargs):
        """Convenience for performing service layer lookups.

        XXX This duplicates Page._service_lookup. It should be eventually
        cleaned up by moving the logic into a page.

        @see canonical.landscape.ui.service.lookup_entity
        """
        try:
            return service_lookup(self._request, lookup_class, *args, **kwargs)
        except ZopeNotFound as error:
            # Set the context object of the NotFound error to this view.
            error.ob = self
            raise error


class NoChildrenActivityView(ActivityView):
    def get_children(self):
        return []


class ActivityGroupViewMixin(object):
    @cached
    def is_scheduled(self):
        """
        Return C{True} if at least one children is scheduled for this group of
        activities.
        """
        return bool(self.get_scheduled_time())

    @cached
    def get_scheduled_time(self):
        """
        Get the time this item is scheduled for, by accumulating all the times
        and finding the earliest (min) value.
        """
        # self._info is an ActivityGroup instance
        scheduled_time = self._info.get_scheduled_time()
        return scheduled_time

    @cached
    def get_computer_count(self):
        """
        Return a description of the number of computers if there is more than
        one available.
        """
        matched_computer_ids = self._get_children_computer_ids()
        count = len(matched_computer_ids)
        if count > 1:
            return "%d computers" % count

    @cached
    def get_computer(self):
        """
        Return a computer if this group's children all have the same computer.
        """
        matched_computer_ids = self._get_children_computer_ids()
        if len(matched_computer_ids) == 1:
            return self._activity.account.get_computer(matched_computer_ids.pop())

    def _get_children_computer_ids(self):
        """
        Helper to return IDs for computers linked to children of the activity
        that the person has access to.
        """
        computer_ids = self._get_computer_ids()
        activity_computer_ids = set(
            get_activity_child_entity_ids(self._activity, AccountComputerTwin)
        )
        if computer_ids is None:
            # The person has access to al computers.
            return activity_computer_ids

        return activity_computer_ids.intersection(computer_ids)


class ActivityGroupNoChildrenView(ActivityGroupViewMixin, NoChildrenActivityView):
    """
    Specialized L{ActivityView} for L{ActivityGroup}s rendered on
    L{ISummaryActivityView} views.
    """


class ActivityGroupView(ActivityGroupViewMixin, ActivityView):
    """
    Specialized L{ActivityView} for L{ActivityGroup}s rendered on
    L{IActivityView} views.
    """
