# Copyright 2005-2022 Canonical Limited. All rights reserved.

import transaction
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.schema import Choice, TextLine, Bool, Set
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.session.interfaces import ISession

from canonical.content.form import Fields, action
from canonical.routes import Redirect
from canonical.url.url import get_url

from canonical.landscape.ui.form import LandscapeFormProvider, MultiCheckBoxWidget
from canonical.landscape.ui.page.page import LandscapePage
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.salesforce.key import validate_salesforce_key
from canonical.landscape.ui.administrator.interfaces import (
    INewOpenIDPerson,
    INewLocalPerson,
    INewPAMPerson,
)
from canonical.landscape.model.main.interfaces import IPerson, ISalesforceInfo
from safe.canonical.landscape.model.main.person import (
    get_person_by_email,
    get_persons_by_ids,
    create_person,
    get_person_by_identity,
)
from canonical.landscape.model.account.management import AccountManagement
from canonical.landscape.model.eventlog.management import EventLog
from safe.canonical.landscape.model.account.invitation import get_invitation
from canonical.landscape.model.account.errors import (
    MaximumAccountAdministratorsError,
    RepeatedInvitationError,
    IllegalInvitationError,
)
from canonical.landscape.model.access.role import get_role_registry
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.pam import pam
from canonical.landscape.model.configuration import (
    set_default_system_email,
    set_root_url,
    get_root_url,
)
from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.main.account import (
    get_account_by_name,
    DEFAULT_STANDALONE_ACCOUNT_NAME,
)
from canonical.landscape.model.configuration import is_standalone
from landscape.lib.fs import read_file


def get_system_timezone(path="/etc/timezone"):
    """Get system timezone."""
    tzname = read_file(path).decode().strip()
    return "UTC" if "UTC" in tzname.split("/") else str(tzname)


class AdministratorPage(AccountPage):

    title = "Administrators"

    @Lazy
    def invitations(self):
        return list(self.context.get_invitations())

    @Lazy
    def account_title(self):
        return self.context.account.title

    @Lazy
    def people_count(self):
        return self.context.account.people.count()

    @Lazy
    def persons(self):
        return list(self.context.account.people)

    @Lazy
    def remaining_persons(self):
        """The number of remaining seats for account administrators."""
        remaining_persons = self.max_people_count - self.people_count
        if remaining_persons > 0:
            return remaining_persons
        return 0

    @Lazy
    def max_people_count(self):
        return self.context.account.max_people_count

    @Lazy
    def can_view_salesforce_keys(self):
        """Returns true if a user should be able to see salesforce keys."""
        principal = self.request.principal
        return checkPermission("landscape.ViewSalesforceKey", principal)

    @Lazy
    def can_invite_administrator(self):
        """Returns C{True} if the user can invite a new administrator."""
        try:
            self.context.check_person_can_invite()
        except IllegalInvitationError:
            return False
        return True

    def needs_administrator_limit_notice(self):
        """Indicate whether to show the administrator limit reached notice."""
        return not self.can_invite_administrator and checkPermission(
            "landscape.InviteAdministrator", self.context
        )

    def update_invitations(self):
        """Cancel or resend invitations."""

        def cancel(invitation):
            invitation.cancel()

        def resend(invitation):
            invitation.send_invitation()

        action_name = None
        url = "/dashboard"

        if "cancel_invitations" in self.request:
            action_name = "cancel_invitations"
            action_function = cancel
            action_participle = "cancelled"

        if "resend_invitations" in self.request:
            action_name = "resend_invitations"
            action_function = resend
            action_participle = "resent"

        event_log = EventLog(self.account, self.request.principal.person)
        if action_name in self.request:
            secure_ids = self.request.get("invitations", ())
            if secure_ids:
                failed_count = 0
                for secure_id in secure_ids:
                    invitation = self.context.get_invitation(secure_id)
                    if invitation:
                        action_function(invitation)
                        message = "%s invitation of %s." % (
                            action_participle.capitalize(),
                            invitation.email,
                        )
                        self.notes.info(message)
                        event_log.create_log_entry(self.account, message)
                    else:
                        failed_count += 1
                if failed_count > 1:
                    self.notes.warning(
                        "%d invitations could not be %s."
                        % (failed_count, action_participle)
                    )
                elif failed_count == 1:
                    self.notes.warning(
                        "1 invitation could not be %s." % action_participle
                    )
            else:
                self.notes.error(
                    "Invitations must be selected before they "
                    "can be %s." % action_participle
                )
            url = self.request.getURL(1)
        self.request.response.redirect(url)

    def disable_administrators(self):
        person_ids = self.request.form.get("disabled_administrators", ())
        management = AccountManagement(
            self.context.account, self.request.principal.person
        )

        for person in get_persons_by_ids([int(id) for id in person_ids]):
            if person.id == self.request.principal.person.id:
                self.notes.warning(
                    "%s is still an administrator of %s "
                    "(you can't disable yourself)." % (person.name, self.account_title)
                )
                continue
            management.disable_membership(person)
            self.notes.info(
                "%s is no longer an administrator of %s."
                % (person.name, self.account_title)
            )
        self.request.response.redirect(self.request.getURL(1))

    @Lazy
    def _role_registry(self):
        """Return the account's role registry, caching it for performance."""
        return get_role_registry([self.context.account.id])


class InviteAdministratorForm(LandscapeFormProvider):

    dropdown_css_class = "dropdown-spacer-left"

    @Lazy
    def fields(self):
        """
        Build form fields dynamically according to the person's role and
        the account's state.
        """
        fields = Fields(IPerson).select("name", "email")

        person = self.request.principal.person

        if person.has_role("Owner", self.context.access_context):
            # Note that if a person is also an AccountManager, he will be
            # shown the exactly same form as ordinary Owners. This case occurs
            # only for some Canonical employees that are account managers
            # and also have personal accounts.

            management = AccessControlManagement(self.context.account, person)
            roles = management.get_roles()
            terms = [SimpleTerm(role.name, title=role.name) for role in roles]
            vocabulary = SimpleVocabulary(terms)
            fields += Fields(
                Set(
                    __name__="role",
                    title="Role",
                    description=("The roles to be granted " "to the invited person."),
                    required=True,
                    default=set(["GlobalAdmin"]),
                    value_type=Choice(vocabulary=vocabulary),
                )
            )
            fields["role"].custom_widget = MultiCheckBoxWidget

        elif person.has_permission("landscape.InviteInitialAdministrator", None):
            # The person is a "pure" AccountManager or SupportProvider
            # without any other role in the account.
            fields += Fields(ISalesforceInfo).select("salesforce_key")

            if self.view.people_count == 0:
                # It's possible to chose the role of the invitee (either full
                # owner or support-only user) only if the account is empty,
                # otherwise it will be forced to support-only.
                fields += Fields(
                    Bool(
                        __name__="support_only",
                        title="Support only user",
                        description=(
                            "The user is only allowed to access " "support cases."
                        ),
                        required=False,
                    )
                )
        return fields

    @action("Invite", primary=True)
    def add_administrator(self, action, data):
        name, email = data["name"], data["email"]
        person = self.request.principal.person
        salesforce_key = None  # only relevant for pure AccountManagers

        if person.has_role("Owner", self.context.access_context):
            # If the person has the Owner role in the account, he can select a
            # specific role in the account. Note that if the person is also
            # an AccountManager he will be basically offered the same form as
            # an Owner.
            roles = data["role"]
            if not roles:
                self.add_error("At least one role must be selected.", "role")
                return

        elif person.has_permission("landscape.InviteInitialAdministrator", None):
            # The person is a pure AccountManager or SupportProvider,
            # without the Owner role
            salesforce_key = data["salesforce_key"]
            if self.view.people_count == 0 and not data["support_only"]:
                # If the account is empty and the user has not checked the
                # support-only box, then invite as Owner
                roles = ["Owner"]
            else:
                # In all other cases invite as SupportOnly
                roles = ["SupportUser"]

        try:
            invitation = self.context.invite(
                name, email, salesforce_key=salesforce_key, role_names=roles
            )

            invitation.send_invitation()
            self.view.notes.info("%s has been invited" % (name,))

        except MaximumAccountAdministratorsError:
            self.view.notes.error(
                "This account already has the maximum " "number of administrators."
            )
        except RepeatedInvitationError as error:
            self.view.notes.error(str(error))
        self.request.response.redirect(get_url(self.context, self.request))

    def validate(self, action, data):
        person = self.request.principal.person

        if person.has_permission(
            "landscape.InviteInitialAdministrator", None
        ) and not person.has_role("Owner", self.context.access_context):

            is_futher_invitation = self.view.people_count > 0
            is_support_only = data.get("support_only") or is_futher_invitation
            salesforce_key = data.get("salesforce_key")

            if is_support_only and not salesforce_key:
                self.add_error(
                    "Support-only users must have a Salesforce" " contact ID",
                    "salesforce_key",
                )

            if salesforce_key:
                validate_salesforce_key(self, "salesforce_key", salesforce_key)


class NewPerson(LandscapePage):
    """
    A general "new person" page that will be used for both password-based sign
    ups (e.g., standalone server) as well as OpenID-based sign ups.
    """

    title = "New user"
    layout = "fixed"
    invitation = None

    def __init__(self, context, request):
        super(NewPerson, self).__init__(context, request)
        invitation_id = request.get("invitation")
        if invitation_id:
            self.invitation = get_invitation(invitation_id)  # may be None


class StartNewOpenIDPerson(NewPerson):
    """
    The starting page for the two-phase OpenID-based sign up process.
    """

    def __init__(self, context, request):
        super(StartNewOpenIDPerson, self).__init__(context, request)
        self.login_url = "/login/authenticate"
        if self.invitation:
            self.login_url += "?invitation=%s" % self.invitation.secure_id


class NewPAMPerson(NewPerson):
    """
    Used as a subclass to hang the slightly different NewPAMPerson form off.
    """


class CompleteNewOpenIDPerson(LandscapePage):

    title = "New user"
    layout = "fixed"


class NewPersonForm(LandscapeFormProvider):
    """
    A general page to be used as a base class for local password-authenticated
    has been established.
    """

    fields = None
    invitation = None
    summary = "Welcome to Landscape"

    def __init__(self, context, request, view):
        super(NewPersonForm, self).__init__(context, request, view)
        invitation_id = request.get("invitation")
        if invitation_id:
            self.invitation = get_invitation(invitation_id)
        self.system_timezone = get_system_timezone()

    def get_initial_data(self):
        if self.invitation:
            return {
                "name": self.invitation.name,
                "email": self.invitation.email,
                "invitation": self.invitation.secure_id,
            }
        session = ISession(self.request)
        new_person = session["landscape.authentication"]
        return {"name": new_person.get("name"), "email": new_person.get("email")}

    def create_person(self, data):
        """
        Subclasses need to implement the means by which a user is created, as
        this will change in different cases.
        """
        raise NotImplementedError

    @action("Sign up", primary=True)
    def new_user(self, action, data):

        no_account = get_account_by_name(DEFAULT_STANDALONE_ACCOUNT_NAME) is None
        if is_standalone() and no_account:
            # XXX Legacy package-based LDS deployments were do not have a
            # root-url service.conf config key, like the charm-based
            # deployments, so we set the initial value here.
            config = getUtility(IFileBasedConfiguration)
            if not config.get("global", key="root-url", default=None):
                root_url = self.request.URL[0]
                if not root_url.endswith("/"):
                    root_url += "/"
                    set_root_url(root_url)
            else:
                root_url = get_root_url()

            set_default_system_email(root_url)

        # Unwrap the person so that we can access ID and, in
        # Invitation.activate called below, the email address.
        person = removeSecurityProxy(self.create_person(data))

        # Log the user in before redirecting.
        session_data = ISession(self.request)["landscape.authentication"]
        session_data["person_id"] = person.id
        transaction.commit()

        # If the user was invited, send him back to the invitation URL with
        # the auto-accept parameter set, otherwise send him to his dashboard.
        if self.request.get("form.invitation"):
            invitation = get_invitation(self.request.get("form.invitation"))
            url = get_url(invitation, self.request) + "?auto-accept=true"
        else:
            url = self.request.form.get("form.next_url", "/dashboard")
        self.request.response.redirect(url)

    def validate(self, action, data):
        if "form.email" not in self.widget_errors:
            if get_person_by_email(data["email"]) is not None:
                self.add_error("Person with that email already exists.", "email")

        if data.get("password") != data.get("verified_password"):
            self.add_error("Password don't match", "verified_password")


class NewLocalPersonForm(NewPersonForm):
    """
    A user registration form for Landscape instances that are using password
    authentication and not OpenID, such as when the server is running in
    stand-alone mode.
    """

    fields = Fields(INewLocalPerson).hide("invitation")

    def create_person(self, data):
        return create_person(
            data["name"], data["email"], data["password"], self.system_timezone
        )


class NewPAMPersonForm(NewPersonForm):
    """
    A user registration form for Landscape instances that are using password
    authentication and not OpenID, such as when the server is running in
    stand-alone mode.
    """

    fields = Fields(INewPAMPerson).hide("invitation")

    def create_person(self, data):
        person = create_person(
            data["name"],
            data["email"],
            None,
            timezone=self.system_timezone,
            identity=data.get("identity"),
        )
        return person

    def validate(self, action, data):
        if "form.identity" not in self.widget_errors:
            if get_person_by_identity(data.get("identity")):
                self.add_error("Person with that identity already exists.", "identity")
        if not pam.authenticate(
            data.get("identity"), data.get("password"), "landscape"
        ):
            self.add_error("Unable to validate credentials.", "identity")
        if "form.email" not in self.widget_errors:
            if get_person_by_email(data["email"]) is not None:
                self.add_error("Person with that email already exists.", "email")


class CompleteNewOpenIDPersonForm(NewPersonForm):
    """
    A page with a form for filling out user details after Ubuntu SSO identity
    has been established.

    session["landscape.authentication"]["identity"] should be set before
    redirecting a user to this page.
    """

    fields = Fields(INewOpenIDPerson).hide("invitation")

    def __init__(self, context, request, view):
        super(CompleteNewOpenIDPersonForm, self).__init__(context, request, view)
        session = ISession(request)
        self.identity = session["landscape.authentication"].get("identity")
        if self.identity is None:
            raise Redirect("/")
        next_url = self.request.get("next_url")
        if next_url:
            self.fields += Fields(TextLine(__name__="next_url", default=next_url))
            # It's unfortunate that we have to hide the invitation field again,
            # but our extended Fields API doesn't play nice with formlib's
            # Fields.__add__, which returns a new instance that ignores our
            # extended class attributes.
            self.fields = self.fields.hide("next_url", "invitation")

    def create_person(self, data):
        return create_person(
            data["name"],
            data["email"],
            timezone=self.system_timezone,
            identity=self.identity,
        )
