# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.cachedescriptors.property import Lazy
from zope.session.interfaces import ISession

from canonical.content.form import action
from canonical.url.url import get_url

from canonical.landscape.ui.page.page import LandscapePage
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.model.account.errors import (
    MaximumAccountAdministratorsError,
    PersonAlreadyInAccountError,
    InvalidInvitationError,
)


class AcceptInvitationPage(LandscapePage):

    title = "Accept invitation"

    def __call__(self):
        if not self._is_intended_recipient():
            self.notes.error(
                "This invitation is intended for a different " "recipient."
            )
            self.request.response.redirect("/")
            return None

        return super(AcceptInvitationPage, self).__call__()

    @Lazy
    def account_title(self):
        return self.context.account_title

    def _is_intended_recipient(self):
        """
        Returns True if the request is from an OpenID user and the invitation
        was sent to one of their emails.
        """
        auth_data = ISession(self.request)["landscape.authentication"]
        emails = auth_data.get("extra_emails", [])

        email = auth_data.get("email")
        if email:
            emails = emails + [email]
        emails = [e.lower() for e in emails]

        # If there are no emails, then this is not an OpenID user - assume
        # they are the intended recipient.
        return not emails or self.context.email.lower() in emails


class AcceptInvitationForm(LandscapeFormProvider):
    """Form for accepting or rejecting invitations."""

    def update(self):
        if self.request.get("auto-accept") == "true":
            # Automatically activate the invitation
            self._activate()
        super(AcceptInvitationForm, self).update()

    @action("Accept", primary=True)
    def accept(self, action, data):
        self._activate()

    def _activate(self):
        """Activate the invitation for the accessing person."""
        account = self.context.account
        url = "/"
        try:
            self.context.activate(self.request.principal.person)
        except MaximumAccountAdministratorsError:
            self.add_error(
                "The '%s' account has reached its "
                "administrator limit." % account.title
            )
        except PersonAlreadyInAccountError:
            self.add_error(
                "You are already member of the '%s' account. "
                "Invitation cancelled." % account.title
            )
        except InvalidInvitationError:
            self.add_error("The role you were invited to no longer exists.")
        else:
            # If everything went well, redirect to the account
            url = get_url(account, self.request)
            self.view.notes.info(
                "You are now a member of the '%s' account." % account.title
            )
        self.request.response.redirect(url)

    @action("Reject", primary=False)
    def reject(self, action, data):
        self.context.cancel()
        self.request.response.redirect("/")
