# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.security import proxy

from canonical.url import get_url, quote_url

from canonical.landscape.model.alert.computer import ComputerRuleInfo
from canonical.landscape.model.computer.search import (
    AlertCriteria,
    get_alert_names_mapping,
)
from canonical.landscape.ui.form import TagAssociationHelper
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.alert.view import get_alert_view


class AlertPage(AccountPage):
    """View the current status of and details about a particular L{Rule}."""

    title = "Alert"

    def __init__(self, context, request):
        super(AlertPage, self).__init__(context, request)
        if self.is_computer_rule():
            self.tags = sorted([tag.name for tag in context.get_tags()])
        self.alert = get_alert_view(context, request)

    def is_subscribed(self):
        person = self.request.principal.person
        return bool(self.context.get_subscription_by_person(person))

    def subscribe(self):
        person = self.request.principal.person
        if self.is_subscribed():
            self.notes.info("You are already subscribed to this alert.")
        else:
            self.context.subscribe(person)
            self.notes.info("You are subscribed to this alert.")
        self.request.response.redirect(get_url(self.context, self.request))

    def unsubscribe(self):
        person = self.request.principal.person
        self.context.unsubscribe(person)
        self.notes.info("You are not subscribed to this alert.")
        self.request.response.redirect(get_url(self.context, self.request))

    def is_computer_rule(self):
        return proxy.isinstance(self.context.info, ComputerRuleInfo)

    @property
    def computer_search_link(self):
        mapping = get_alert_names_mapping()
        alert_name = mapping.get(self.context.info.__class__)

        url = get_url(self.context.account, self.request)
        criteria = AlertCriteria(alert_name, self.context.account)
        criteria_url = quote_url(criteria.to_string())
        return "%s/computers/criteria/%s" % (url, criteria_url)


class EditAlertPage(AccountPage, TagAssociationHelper):
    """Update choices that affect which L{Computer}s a L{Rule} is enabled for.

    The page shows a checkbox for 'All computers' which, when checked, enables
    the L{Rule} for all L{Computer}s in the L{Account}.  It also shows a
    checkbox for each L{Tag} in the account.  A tag checkbox is checked if the
    rule is associated with it.

    @ivar all_tags: A list of L{Tag} names, sorted alphabetically.
    @ivar tags: A list of L{Tag} names, sorted alphabetically, associated with
        the L{Rule} used as the context of the page.
    """

    title = "Edit alert"

    update_requires = "landscape.EditAlert"

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        TagAssociationHelper.__init__(self, context, request)
        if self.request.method == "POST":
            self.update("alert")
