# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.cachedescriptors.property import Lazy
from zope.interface import implementer

from canonical.notes.interfaces import INFO, WARNING
from canonical.url import get_url

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.alert.interfaces import IAlertRules
from canonical.landscape.service.alert.get import GetAlertsOperation


def find_subscription_changes(subscribing, selected, rules_by_id, subscriptions):
    """Determine which alert subscriptions the user wants to change.

    @param subscribing: A boolean indicating whether the user is subscribing
        (True) or unsubscribing (False).
    @param selected: A sequence of rule IDs corresponding to the alert rules
        the user checked on the submitted form.
    @param rules_by_id: A mapping of rule ID to rules.  This method does not
        depend on the internals of rules (type, attributes, methods, etc.).
    @param subscriptions: The users current subscriptions (IDs of rules) as a
        collection that supports the "in" operator.

    @returns: A list of the rules which should have their subscriptions
        changed.
    """
    changes = []
    for rule_id, rule in list(rules_by_id.items()):
        if rule_id in selected:
            subscribed = rule_id in subscriptions
            # Do we need to change this subscription?
            if subscribing ^ subscribed:
                changes.append(rule)
    return changes


def make_notification(subscribing, selected, changes):
    """Construct a message (and severity) describing the changes.

    @param subscribing: A boolean indicating whether the user is  subscribing
        (True) or unsubscribing (False).
    @param selected: A sequence of rule IDs corresponding to the alert rules
        the user checked on the submitted form.
    @param changes: A list of the rules which should have their subscriptions
        changed.

    @returns: A tuple of severity and message.
    """
    if subscribing:
        verb = "subscribed to"
    else:
        verb = "unsubscribed from"

    if not selected:
        return (WARNING, "Please select alerts to be %s." % verb)

    if len(changes) > 1:
        return (INFO, "You have %s %d alerts." % (verb, len(changes)))
    elif len(changes) == 1:
        change = changes[0]
        rule = change["rule"]
        return (INFO, "You have %s %r." % (verb, rule.info.title))
    else:
        return (WARNING, "No changes.")


@implementer(IAlertRules)
class AlertRules(object):
    """Context object for the alert rules listing page."""

    def __init__(self, account):
        self.account = account


def rule_sort_key(rule):
    return rule.info_type, rule.info.title


class AlertRulesPage(AccountPage):

    title = "Alerts"

    @Lazy
    def rules(self):
        """A summary of the rules to which this person could subscribe."""
        operation = GetAlertsOperation(self.context.account, self.request.principal)
        rules = operation()
        for rule in rules:
            rule["status_type"] = self._get_status_class(rule["status"])
            rule["enabled_for"] = self._get_enabled_for_text(rule)
        return rules

    def update(self):
        """Apply the user's requested alert subscription changes."""
        url = get_url(self.context, self.request)
        selected = [int(id) for id in self.request.get("rules", ())]
        rules_by_id = dict((rule["rule"].id, rule) for rule in self.rules)
        # The user is either subscribing or unsubscribing.
        subscribing = "subscribe" in self.request
        changes = find_subscription_changes(
            subscribing, selected, rules_by_id, self._get_subscriptions()
        )
        for summary in changes:
            rule = summary["rule"]
            if subscribing:
                rule.subscribe(self.request.principal.person)
            else:
                rule.unsubscribe(self.request.principal.person)

        self.notes.add(*make_notification(subscribing, selected, changes))
        self.request.response.redirect(url)

    def _get_status_class(self, status):
        """Return the CSS class associated with a given alert status."""
        if status == "Alerted":
            return "bad"
        return "good"

    def _get_subscriptions(self):
        """This user's current set of subscriptions.

        @returns: A set of rule IDs.
        """
        rules = self.context.account.get_subscribed_rules(self.request.principal.person)
        return set(rule.id for rule in rules)

    def _get_enabled_for_text(self, rule):
        """Return a text summary of the rules 'scope', 'all_computers' and
        'tags' fields.
        """
        if rule["scope"] == "account":
            return "Account"
        else:
            if rule["all_computers"] == True:
                return "All computers"
            else:
                tags = rule["tags"]
                if len(tags) == 0:
                    return "No computers"
                else:
                    return "Computers tagged with " + ", ".join(tags)
