# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer
from zope.component import getMultiAdapter

from canonical.landscape.model.main.computer import Computer
from canonical.landscape.ui.alert.interfaces import IAlertView


def get_alert_view(rule, request):
    """Return a rule view adapting the given objects."""
    return getMultiAdapter((rule.info, request), IAlertView)


@implementer(IAlertView)
class AlertView(object):
    """View provides details about alert objects."""

    resolution_text = "View more details about this alert"

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @property
    def summary(self):
        """Get a summary about the state of the rule."""
        return self.context.get_summary()

    @property
    def resolution_url(self):
        """Return a URL that can be used to resolve the alerted situation."""
        return self.context.get_resolution_url()

    @property
    def active(self):
        """Return C{True} if the rule is active."""
        return self.context.is_active()

    @property
    def name(self):
        """Return the user readable name of the underlying RuleInfo class."""
        return self.context.get_name()


class PendingComputersAlertView(AlertView):

    resolution_text = "View pending computers"


class ComputerAlertView(AlertView):
    def _get_computer_ids(self):
        return self.request.principal.person.get_computer_ids(self.context.rule.account)

    @property
    def summary(self):
        """Get a summary about the state of the rule.

        The generated summary will take into account only those computers
        the L{Person} associated to the request has access to.
        """
        return self.context.get_summary(computer_ids=self._get_computer_ids())

    @property
    def resolution_url(self):
        """Return a URL that can be used to resolve the alerted situation.

        The generated URL will take into account only those computers
        the L{Person} associated to the request has access to.
        """
        return self.context.get_resolution_url(computer_ids=self._get_computer_ids())

    @property
    def active(self):
        """Return C{True} if the rule is active.

        The returned value takes into account only those computers
        the L{Person} associated to the request has access to.
        """
        return self.context.is_active(computer_ids=self._get_computer_ids())

    @property
    def computers(self):
        """Get all accessible L{Computer}s affected by the rule."""
        result = self.context.get_computers(
            active=True, computer_ids=self._get_computer_ids()
        )
        result.order_by(Computer.title)
        return list(result)


class ComputerOfflineAlertView(ComputerAlertView):

    resolution_text = "View offline computers"


class UnapprovedActivitiesAlertView(ComputerAlertView):

    resolution_text = "View unapproved activities"


class PackageUpgradesAlertView(ComputerAlertView):

    resolution_text = "View package upgrades"


class SecurityUpgradesAlertView(ComputerAlertView):

    resolution_text = "View security upgrades"


class PackageProfilesAlertView(ComputerAlertView):

    resolution_text = "View package profiles"


class EsmDisabledAlertView(ComputerAlertView):

    resolution_text = (
        "Attach a UA subscription on eligible computers to"
        " ensure ESM updates are applied"
    )


class StandaloneOverflowAlertView(AlertView):

    resolution_text = "Please log a sales issue support case online."


class StandaloneSystemEmailAlertView(AlertView):

    resolution_text = "Edit system email address"


class LicenseExpiryAlertView(AlertView):

    resolution_text = "Please log a sales issue support case online."


class StandaloneScriptAlertView(AlertView):

    resolution_text = "View Landscape scripts that have issues"


class PackageReporterAlertView(ComputerAlertView):

    resolution_text = "View package reporting issues"


class FailedBillingAlertView(AlertView):

    resolution_text = "Please verify your billing details"
