# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.browserpage import ViewPageTemplateFile
from zope.schema import Choice
from zope.security.interfaces import Unauthorized

from canonical.url import get_url

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.autoupgrade.schedule import (
    validate_custom_form,
    form_to_ical,
    ViewSchedule,
)

from canonical.landscape.model.package.interfaces import IUpgradeProfile
from canonical.content.form import Fields, action
from canonical.landscape.model.package.autoupgrade import get_upgrade_type
from canonical.landscape.model.main.person import PersonAccessContextsVocabularyFactory
from canonical.landscape.model.activity.group import validate_delivery_time
from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError
from canonical.landscape.ui.timezones import Timezone


class CreateUpgradeProfilePage(AccountPage):

    title = "Create an upgrade profile"

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        if not self.context.person_can("EditUpgradeProfile"):
            raise Unauthorized()


def get_deliver_delay_window(context, request):
    """
    Utility method to load the C{deliver_delay_window} either from the request
    or from the context, in that order.
    """
    deliver_delay_window = request.get("form.deliver_delay_window", None)
    if deliver_delay_window is not None:
        return deliver_delay_window
    if hasattr(context, "deliver_after_window"):
        return context.deliver_after_window
    else:
        return 0


def pre_validate_deliver_after_window(deliver_delay_window, form_errors):
    """
    Checks if the value that was passed is a valid digit or if it's a digit,
    it's not negative.
    """
    if (
        deliver_delay_window is not None
        and not deliver_delay_window.isdigit()
        or isinstance(deliver_delay_window, int)
        and deliver_delay_window < 0
    ):
        form_errors["deliver_delay_window"] = (
            "The 'deliver delay window' of '%s' must be a number greater than"
            " or equal to zero." % deliver_delay_window
        )


class CreateUpgradeProfileForm(LandscapeFormProvider):

    summary = "Create an upgrade profile"

    fields = Fields(IUpgradeProfile)

    additional_content = ViewPageTemplateFile("schedule-fields.pt")

    dropdown_css_class = "dropdown-spacer-upgrade"

    def __init__(self, context, request, view):
        super(CreateUpgradeProfileForm, self).__init__(context, request, view)
        timezone = request.principal.person.timezone
        self.schedule = ViewSchedule(None, timezone, request.form)
        self.custom_form_errors = {}
        access_context = Choice(
            __name__="access_context",
            title="Access group",
            source=PersonAccessContextsVocabularyFactory("EditUpgradeProfile"),
            default="Global",
            description="The access group the profile should belong to.",
            required=True,
        )
        user_tz = self.request.principal.person.timezone
        utc_offset = Timezone(user_tz).offset
        offset_to_display = "(GMT %s) %s" % (
            utc_offset[:-2] + ":" + utc_offset[-2:],
            user_tz,
        )
        self.description = "All times in your currently selected timezone" ": %s" % (
            offset_to_display
        )
        self.fields += Fields(access_context)
        self.deliver_delay_window = get_deliver_delay_window(self.context, self.request)

    def _add_custom_form_error(self, message, key):
        """
        This method adds an error to the custom form error dict, that is then
        accessible form the template.
        """
        self.custom_form_errors[key] = message

    def get_initial_data(self):
        return {"access_context": self.context.account.access_context.name}

    def validate(self, action, data):
        validate_custom_form(
            self.request.form,
            self.request.principal.person.timezone,
            data,
            self._add_custom_form_error,
        )

        # Custom validation for checking the combination of expiry time and
        # delivery delay window.
        deliver_delay_window = self.request.form.get("form.deliver_delay_window", None)
        pre_validate_deliver_after_window(deliver_delay_window, self.custom_form_errors)

        try:
            deliver_delay_window = self.request.form.get("form.deliver_delay_window", 0)
            expire_time = data.get("deliver_within", None)
            validate_delivery_time(
                run_window_length=expire_time, deliver_after_window=deliver_delay_window
            )
        except InvalidScheduleAfterWindowError as e:
            self._add_custom_form_error(str(e), "deliver_delay_window")

    def has_errors(self):
        """
        Overrides the superclass's behavior to account for the custom form
        validation.
        """
        normal_errors = super(CreateUpgradeProfileForm, self).has_errors()
        return normal_errors or self.custom_form_errors

    @action("Save", primary=True)
    def save(self, action, data):
        title = data["title"]
        security_upgrade = data["security_upgrade"]
        autoremove = data.get("autoremove", False)

        schedule = form_to_ical(data)
        deliver_within = data.get("deliver_within") or 1

        deliver_delay_window = int(self.request.get("form.deliver_delay_window", 0))

        timezone = self.request.principal.person.timezone

        profile = self.context.create_upgrade_profile(
            title,
            get_upgrade_type(security_upgrade),
            schedule,
            deliver_within,
            access_context_name=data["access_context"],
            deliver_after_window=deliver_delay_window,
            timezone=timezone,
            autoremove=autoremove,
        )
        self.request.response.redirect(get_url(profile, self.request))
