# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from zope.browserpage import ViewPageTemplateFile
from zope.schema import TextLine
from zope.cachedescriptors.property import Lazy

from canonical.url import get_url

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.autoupgrade.create import (
    get_deliver_delay_window,
    pre_validate_deliver_after_window,
)

from canonical.landscape.model.package.interfaces import IUpgradeProfile
from canonical.landscape.model.package.autoupgrade import UpgradeProfileManagement
from canonical.content.form import Fields, action
from canonical.landscape.ui.autoupgrade.schedule import (
    validate_custom_form,
    form_to_ical,
    ViewSchedule,
)
from canonical.landscape.model.activity.group import validate_delivery_time
from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError
from canonical.landscape.ui.timezones import Timezone


class EditUpgradeProfilePage(AccountPage):

    title = "Edit an upgrade profile"

    requires = "landscape.EditUpgradeProfile"


class EditUpgradeProfileForm(LandscapeFormProvider):

    summary = "Edit an upgrade profile"

    additional_content = ViewPageTemplateFile("schedule-fields.pt")

    def __init__(self, context, request, view):
        super(EditUpgradeProfileForm, self).__init__(context, request, view)
        self.manager = UpgradeProfileManagement(
            context.account, self.request.principal.person
        )
        timezone = request.principal.person.timezone
        self.schedule = ViewSchedule(context, timezone, request.form)
        self.custom_form_errors = {}
        user_tz = self.request.principal.person.timezone
        utc_offset = Timezone(user_tz).offset
        offset_to_display = "(GMT %s) %s" % (
            utc_offset[:-2] + ":" + utc_offset[-2:],
            user_tz,
        )
        self.description = "All times in your currently selected timezone" ": %s" % (
            offset_to_display
        )
        self.deliver_delay_window = get_deliver_delay_window(self.context, self.request)

    @Lazy
    def fields(self):
        fields = Fields(IUpgradeProfile, render_context=True)
        fields += Fields(
            TextLine(
                __name__="access_context",
                title="Access group",
                readonly=True,
                default=self.context.access_context.title,
            )
        )
        return fields

    def _add_custom_form_error(self, message, key):
        """
        This method adds an error to the custom form error dict, that is then
        accessible from the template.
        """
        self.custom_form_errors[key] = message

    def validate(self, action, data):
        validate_custom_form(
            self.request.form,
            self.request.principal.person.timezone,
            data,
            self._add_custom_form_error,
        )

        # Custom validation for checking the combination of expiry time and
        # delivery delay window.
        deliver_delay_window = self.request.form.get("form.deliver_delay_window", None)

        pre_validate_deliver_after_window(deliver_delay_window, self.custom_form_errors)

        try:
            validate_delivery_time(
                run_window_length=data.get("deliver_within"),
                deliver_after_window=deliver_delay_window,
            )
        except InvalidScheduleAfterWindowError as e:
            self._add_custom_form_error(str(e), "deliver_delay_window")

    def has_errors(self):
        """
        Overrides the superclass's behavior to account for the custom form
        validation.
        """
        normal_errors = super(EditUpgradeProfileForm, self).has_errors()
        return normal_errors or self.custom_form_errors

    @action("Save", primary=True)
    def save(self, action, data):
        title = data.get("title", None)
        schedule = form_to_ical(data)
        security_upgrade = data.get("security_upgrade", None)
        autoremove = data.get("autoremove", False)
        deliver_within = data.get("deliver_within")

        deliver_delay_window = int(self.request.get("form.deliver_delay_window", 0))
        timezone = self.request.principal.person.timezone

        self.manager.edit(
            self.context.name,
            title=title,
            security_upgrade=security_upgrade,
            deliver_within=deliver_within,
            schedule=schedule,
            deliver_after_window=deliver_delay_window,
            timezone=timezone,
            autoremove=autoremove,
        )
        self.request.response.redirect(get_url(self.context, self.request))
