# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import object
from datetime import datetime
import pytz
import six

from canonical.lib.timestamp import utcnow
from canonical.landscape.model.package.autoupgrade import (
    schedule_to_dict,
    WEEKDAY_MAP,
    WEEKDAY_DISPLAY_MAP,
)


def form_to_ical(fields):
    """
    Converts the form values for the custom upgrade schedule widget to
    an iCal-valid schedule string. This does only the conversion, and does not
    check for the semantics.
    """
    ical = ""

    frequency = fields.get("frequency")
    if frequency:
        ical += "FREQ=%s" % frequency
    else:
        return ""  # There is no point in returning anything invalid

    days_string = fields.get("weekdays")
    if days_string:
        if not isinstance(days_string, six.string_types):
            days_string = ",".join(days_string)
        ical += ";BYDAY=%s" % days_string

    if frequency == "WEEKLY":
        hour = fields.get("at_hour")
        if hour is not None:
            ical += ";BYHOUR=%s" % hour

    minute = fields.get("at_minute")
    if minute is not None:
        ical += ";BYMINUTE=%s" % minute

    return ical


def validate_custom_form(request_form, timezone_name, data, error_callback):
    """
    This validates the fields from the custom schedule "widget".
    """
    frequency = request_form.get("form.frequency")
    data["frequency"] = frequency  # Save it on data
    data["weekdays"] = request_form.get("form.weekdays")

    if frequency == "WEEKLY":
        # There should be at least one day selected
        if not data["weekdays"]:
            error_callback("Please select at least one day of the week.", "weekdays")

        time = request_form.get("form.at_time")
        try:
            hour, minute = time.split(":")
            hour = int(hour)
            minute = int(minute)
        except ValueError:
            error_callback("Time should be in the HH:MM format.", "at_time")
        else:

            if not (0 <= hour <= 23) or not (0 <= minute <= 59):
                error_callback("Please enter a valid time.", "at_time")
            else:
                # In order to make sure we don't depend on server date/time,
                # use utcnow() and convert it to the user tz and replace it
                # with the hour and minutes passed in.
                tz = pytz.timezone(timezone_name)
                utc_now = pytz.utc.localize(datetime.utcnow())
                localized_now = utc_now.astimezone(tz)
                localized_now = localized_now.replace(hour=hour, minute=minute)

                data["at_hour"] = localized_now.hour
                data["at_minute"] = localized_now.minute

        expire_after = request_form.get("form.expire_time")
        try:
            expire_after = int(expire_after)
        except (TypeError, ValueError):
            error_callback(
                "Expiry time should be a valid number of hours.", "expire_time"
            )
        else:
            if not 0 < expire_after:
                error_callback("Expiry time should be positive.", "expire_time")
            data["deliver_within"] = expire_after

    else:
        # Hourly schedule
        minute = request_form.get("form.at_minute")
        try:
            minute = int(minute)
            if not (0 <= minute <= 59):
                raise ValueError
        except (TypeError, ValueError):
            error_callback("Minute should be a number between 0 and 59.", "at_minute")
        else:
            tz = pytz.timezone(timezone_name)
            local_now = datetime.now(tz)
            local_target = local_now.replace(minute=minute)

            data["at_minute"] = local_target.minute
            data["deliver_within"] = 1  # hourly should assert 1 hour expiry


def two_digits_minutes(minute):
    """
    Always returns a two digits string representation of the passed-in minutes
    (for nicer display).
    """
    if minute is None or int(minute) < 0 or int(minute) > 59:
        return None
    return "%02d" % int(minute)


class ViewSchedule(object):
    """
    A class to represent the schedule when passed to templates. This is just
    a simple class that holds the various fields exposed to the template, used
    to create create/edit forms and the detail page (dashboard).
    """

    is_hourly_frequency = None
    deliver_within = None
    at_minute = None
    at_time = None
    on_day = None
    monday = None
    tuesday = None
    wednesday = None
    thursday = None
    friday = None
    saturday = None
    sunday = None
    display_weekdays = ""

    def __init__(self, upgrade_profile, tzinfo="UTC", posted_form=None):
        tz = pytz.timezone(tzinfo)
        utc_now = pytz.utc.localize(utcnow())
        now = utc_now.astimezone(tz)

        if posted_form is None:
            posted_form = {}

        if upgrade_profile is None:
            schedule_dict = {}
        else:
            schedule_dict = schedule_to_dict(upgrade_profile.schedule)

        self.is_hourly_frequency = schedule_dict.get("every") == "hour"

        if not schedule_dict and posted_form:
            self.is_hourly_frequency = posted_form.get("form.frequency") == "HOURLY"

        if upgrade_profile is not None:
            self.deliver_within = str(upgrade_profile.run_window_length)
        else:
            self.deliver_within = "1"

        if posted_form.get("form.expire_time"):
            self.deliver_within = posted_form.get("form.expire_time")

        self.at_minute = two_digits_minutes(schedule_dict.get("at_minute"))
        if self.at_minute is None:
            # There is no at_minute defined - default to now (in current
            # timezone)
            if posted_form.get("form.at_minute"):
                self.at_minute = posted_form.get("form.at_minute")
            else:
                self.at_minute = two_digits_minutes(now.minute)

        hour = schedule_dict.get("at_hour")
        if hour is not None:
            target_datetime = now.replace(hour=int(hour), minute=int(self.at_minute))

            self.at_time = "%02d:%02d" % (target_datetime.hour, target_datetime.minute)

        else:
            # Set the current time (for the given timezone) as default
            minute = int(self.at_minute)
            hour = now.hour
            self.at_time = "%02d:%02d" % (hour, minute)
            if posted_form.get("form.at_time"):
                self.at_time = posted_form.get("form.at_time")

        days = schedule_dict.get("on_days")

        if days is not None and not posted_form:

            # Days is returned as a JSON list - it is wrapped in []
            weekdays = days
            self.display_weekdays = [WEEKDAY_DISPLAY_MAP[day] for day in weekdays]
            self.display_weekdays = ", ".join(self.display_weekdays)

            self.on_day = days
            for short, longer in WEEKDAY_MAP.items():
                if short in days:
                    setattr(self, longer, True)

        elif not self.is_hourly_frequency:
            # Default to saturday and sunday
            self.on_day = "sa,su"
            self.display_weekdays = "Sat, Sun"
            self.saturday = True
            self.sunday = True

        if posted_form:
            posted_days = posted_form.get("form.weekdays")
            # Revert the potentially set defaults
            self.on_day = ""
            self.saturday = None
            self.sunday = None
            if posted_days:
                self.on_day = ",".join(posted_days)
                for short, longer in WEEKDAY_MAP.items():
                    if short in posted_days:
                        setattr(self, longer, True)

    def string_display(self):
        """
        Returns a nice string representation to be used in views.
        """
        if self.is_hourly_frequency:
            display = "Every hour at %s minutes past the hour" % self.at_minute
            if self.display_weekdays:
                display += " on every %s" % self.display_weekdays
            return display
        else:
            return "Every %s at %s" % (self.display_weekdays, self.at_time)
