# Copyright 2005-2017 Canonical Limited.  All rights reserved.

from builtins import str
from zope.cachedescriptors.property import Lazy
from zope.security import checkPermission
from zope.security.interfaces import Unauthorized

import stripe

from canonical.lib.timestamp import format_date
from canonical.lib.serialize import namedtuple_to_dict
from canonical.content.decorator import json
from canonical.url import get_url

from canonical.landscape.stripe.customer import (
    get_card_details,
    get_customer,
    delete_credit_card,
)
from canonical.landscape.model.configuration import get_stripe_secret_key
from canonical.landscape.service.billing.summary import GetUsageSummary
from canonical.landscape.service.billing.payment import PaymentCollectionLookup
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.model.eventlog.management import EventLog


class BillingPage(AccountPage):

    title = "Billing"

    def __call__(self):
        result = super(BillingPage, self).__call__()
        if self.request.method == "POST":
            return self._handle_post()
        return result

    def _handle_post(self):
        """A POST here deletes the card."""
        if not checkPermission("landscape.ManageBilling", self.context.account):
            raise Unauthorized()
        api_key = get_stripe_secret_key()
        customer_id = self.context.account.stripe_customer_id
        try:
            customer = get_customer(api_key, customer_id)
        except stripe.error.StripeError:
            return None
        else:
            delete_credit_card(customer)
            event_log = EventLog(self.context.account, self.context.person)
            event_log.create_log_entry(
                self.context.account, "Removed payment method from account."
            )

        self.request.response.redirect(get_url(self.context, self.request))

    @Lazy
    def billing_js_url(self):
        return self.static_root_path() + "/javascript/bundles/billing-min.js"

    @Lazy
    def react_js_url(self):
        return self.static_root_path() + "/javascript/bundles/react-min.js"

    @Lazy
    @json
    def usage_json(self):
        """Return usages and costs for all computer types."""
        get_usage_summary = GetUsageSummary(
            self.context.account, self.request.principal
        )
        usage = get_usage_summary()
        remaining_credit = usage.cost.remaining_credit
        usage = namedtuple_to_dict(usage)
        usage["cost"]["remaining_credit"] = remaining_credit

        usage["period"] = self._format_period(usage["period"])

        # Format the remaining credits.
        cost = usage["cost"]
        for i, credit in enumerate(cost["remaining_credits"]):
            credit = namedtuple_to_dict(credit)
            credit["expiry"] = format_date(credit["expiry"])
            del credit["creation_date"]  # not needed
            cost["remaining_credits"][i] = credit

        # Remove per-machine-type usage, since it's not needed
        del usage["usage"]

        return usage

    @Lazy
    @json
    def payments_json(self):
        """Return a list of available payments"""
        account = self.context.account
        payments = self._get_payments(with_receipt_uri=True)
        return [
            {
                "end_date": payment.period_end_date.strftime("%B %Y"),
                "amount": payment.due,
                "usage": payment.total_hours,
                "url": self._get_receipt_url(account.id, payment.id),
            }
            for payment in payments
        ]

    @Lazy
    @json
    def credit_card_json(self):
        """Return card information and errors from Stripe"""
        api_key = get_stripe_secret_key()
        customer_id = self.context.account.stripe_customer_id
        try:
            customer = get_customer(api_key, customer_id)
        except stripe.error.StripeError as e:
            return None, str(e)
        else:
            return get_card_details(customer), None

    @Lazy
    @json
    def can_edit(self):
        """
        Return whether the user is authorized to edit billing information.
        """
        return checkPermission("landscape.ManageBilling", self.context.account)

    @Lazy
    @json
    def failed_payment_json(self):
        """
        Return the error message associated with a failed charge
        against the customers credit card.
        """
        if self._failed_payment:
            return {
                "message": self._failed_payment.stripe_error_message,
                "date": format_date(self._failed_payment.period_end_date),
                "due": self._failed_payment.due,
            }
        return None

    @Lazy
    @json
    def payment_action_json(self):
        """
        Return any pending user action on a payment.
        """
        payments_lookup = PaymentCollectionLookup(self.request.principal)
        account = self.context.account
        result = (
            payments_lookup(account, with_next_action=True, statuses=["incomplete"])
            .find()
            .last()
        )
        if result:
            return {
                "url": result.stripe_next_action,
            }
        return None

    def _get_receipt_url(self, account_id, payment_id):
        return "/download/receipt/{}-{}".format(account_id, payment_id)

    def _format_period(self, period):
        return tuple(format_date(date) for date in period)

    @Lazy
    def _failed_payment(self):
        """
        Return any outstanding, failed payment associated with this
        account.
        """
        return self._get_payments(statuses=["failed"]).any()

    def _get_payments(self, with_receipt_uri=None, statuses=None):
        """
        Return a collection of Payment objects related to to this
        account.
        """
        # Note, this method is not Lazy as it has more than one call
        # site, each expecting contradictory results.  Those call
        # site, however, should be Lazy.
        payments_lookup = PaymentCollectionLookup(self.request.principal)
        account = self.context.account
        result = payments_lookup(
            account, with_receipt_uri=with_receipt_uri, statuses=statuses
        )
        return result.find()
