# Copyright 2005-2017 Canonical Limited.  All rights reserved.

from zope.cachedescriptors.property import Lazy

import stripe
import transaction

from canonical.content.decorator import json
from canonical.landscape.model.billing.management import BillingManagement
from canonical.landscape.model.configuration import (
    get_stripe_secret_key,
    get_stripe_public_key,
)
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.stripe.customer import (
    get_customer,
    update_customer,
    add_credit_card,
    get_card_details,
    setup_intent,
)
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.page.csp import (
    make_csp_header,
    CSP_KIND_SCRIPT,
    CSP_KIND_IMG,
)
from canonical.routes import Redirect
from canonical.url import get_url


class CreateBillingPage(AccountPage):

    title = "Add billing details"

    csp = make_csp_header(
        {
            CSP_KIND_SCRIPT: ("https://js.stripe.com/",),
            CSP_KIND_IMG: ("assets.ubuntu.com",),
        }
    )

    requires = "landscape.ManageBilling"

    def __call__(self):
        card = get_card_details(self._customer)
        if card:
            # User already has a card associated
            raise Redirect(get_url(self.context, self.request))
        result = super(CreateBillingPage, self).__call__()
        if self.request.method == "POST":
            return self._handle_post()
        return result

    def _handle_post(self):
        """Create a new credit card object and update the customer."""
        shipping = self._get_shipping_details(self.request.form)
        update_customer(self._customer, shipping)

        intent_id = self.request.form["stripeSetupIntent"]
        try:
            add_credit_card(self._customer, intent_id)
        except stripe.error.CardError as e:
            self.request.response.setStatus(403)
            return e.code

        event_log = EventLog(self.context.account, self.context.person)
        event_log.create_log_entry(
            self.context.account, "Added payment method to account."
        )

        self.context.account.vat_number = self.request.form["accountVAT"]
        self.context.account.vat_country = self.request.form["accountCountry"]
        self.context.account.is_business = (
            self.request.form["accountType"] == "business"
        )
        reset_payment_retries(self.context.account)
        transaction.commit()

        self.request.response.redirect(get_url(self.context, self.request))

    @Lazy
    def create_js_url(self):
        return self.static_root_path() + "/javascript/bundles/billing-create-min.js"

    @Lazy
    def react_js_url(self):
        return self.static_root_path() + "/javascript/bundles/react-min.js"

    @Lazy
    def stripe_publishable_key(self):
        return get_stripe_public_key()

    @Lazy
    def stripe_client_secret(self):
        """Get a SetupIntent client secret for card validation flow."""
        setup = setup_intent(get_stripe_secret_key())
        return setup.client_secret

    @Lazy
    def _customer(self):
        api_key = get_stripe_secret_key()
        customer_id = self.context.account.stripe_customer_id
        try:
            return get_customer(api_key, customer_id)
        except stripe.error.StripeError:
            return None

    @Lazy
    @json
    def shipping_address_json(self):
        """Get the shipping address from Stripe."""
        if self._customer:
            shipping_address = self._customer.shipping.address
            return {
                "name": self._customer.shipping.name or "",
                "address1": shipping_address.line1 or "",
                "address2": shipping_address.line2 or "",
                "city": shipping_address.city or "",
                "state": shipping_address.state or "",
                "zip": shipping_address.postal_code or "",
                "country": shipping_address.country or "",
            }

    @Lazy
    @json
    def account_vat_json(self):
        """Get the VAT number of this account."""
        return self.context.account.vat_number

    @Lazy
    @json
    def account_type_json(self):
        """Get the VAT number of this account."""
        return "business" if self.context.account.is_business else "personal"

    def _get_shipping_details(self, form):
        """Return a dict with shipping data extracted from form data."""
        return {
            "name": form.get("accountName"),
            "address": {
                "line1": form.get("accountAddress1"),
                "line2": form.get("accountAddress2"),
                "city": form.get("accountCity"),
                "country": form.get("accountCountry"),
                "postal_code": form.get("accountZip"),
                "state": form.get("accountState"),
            },
        }


def reset_payment_retries(account):
    """Mark all outstanding payments for retry."""
    billing_management = BillingManagement(account)
    for payment in billing_management.get_outstanding_payments():
        payment.num_attempts = 0
        payment.last_failure = None
