# Copyright 2005-2017 Canonical Limited.  All rights reserved.

from future import standard_library

standard_library.install_aliases()
from urllib.parse import urlencode

from storm.twisted.transact import transact, Transactor
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.web.resource import Resource, NoResource
from twisted.web.util import DeferredResource
from twisted.web.server import NOT_DONE_YET

from zope.component import getUtility
from zope.session.interfaces import ISessionDataContainer

from canonical.landscape.model.configuration import is_development_mode
from canonical.landscape.service.errors import UnauthorisedError
from canonical.landscape.ui.billing.resource import (
    BillingResource,
    NoPaymentError,
    NoReceiptError,
)


class ReceiptParentResource(Resource):

    transactor = Transactor(reactor.getThreadPool())  # for testing

    @transact
    def _get_person_id(self, request):
        """Return the Person.id associated with this session."""
        client_id = request.getCookie(b"landscape_session")
        sdc = getUtility(ISessionDataContainer)
        session_data = None
        try:
            client_id_str = client_id.decode() if client_id else None
            session_data = sdc[client_id_str]["landscape.authentication"]
            person_id = session_data.get("person_id")
        finally:
            # The SessionDataContainer is actually backed by Storm and
            # does queries against PostgreSQL, using the
            # session-autocommit store.
            #
            # In normal Zope land (i.e. the rest of Landscape UI) we
            # use store.rollback() to close any left over transactions
            # at the start of the request parsing (see
            # c.l.publication.LandscapeBrowserPublication.beforeTraversal).
            #
            # Since we don't want to re-auth the session, or otherwise
            # cause any changes to the db, purely by reading
            # something, we need to rollback.
            if session_data is not None:
                session_data.store.rollback()
        return person_id

    @inlineCallbacks
    def _getChild(self, path, request):
        """Parse the requested URL to idenfity requested receipt.

        Returns NoResource() if there were problems with the URL.

        Returns ForbiddenResource() if we were unable to identify the
        Person associated with the session.
        """
        person_id = yield self._get_person_id(request)
        if person_id is None:
            returnValue(UnauthenticatedAccess())
        try:
            account_id, payment_id = path.split(b"-", 1)
            receipt = Receipt(int(account_id), int(payment_id), person_id)
        except Exception:
            returnValue(NoResource())
        else:
            returnValue(receipt)

    def getChild(self, path, request):
        """Deferred wrapper to the transacted _getChild."""
        return DeferredResource(self._getChild(path, request))


class UnauthenticatedAccess(Resource):

    """Twisted web resource for login redirection."""

    isLeaf = True

    def render_GET(self, request):
        """Redirect to authentication page, pointing back at the Receipt."""
        url = "/login/authenticate?{}".format(urlencode({"next_url": request.uri}))
        request.redirect(url)
        return b""


class Receipt(Resource):

    """Twisted web resource for the receipt."""

    isLeaf = True

    def __init__(self, account_id, payment_id, person_id):
        self._model = BillingResource(account_id, payment_id, person_id)

    def _send_bytes(self, receipt, request):
        """Write a receipt out to the request.

        The receipt is a tuple containing the receipt contents as
        bytes, and filename.
        """
        pdf_bytes, filename = receipt
        request.setHeader("Content-Type", "application/pdf")
        request.setHeader(
            "Content-Disposition", 'attachment; filename="{}"'.format(filename)
        )
        request.write(pdf_bytes)
        request.finish()

    def _403_unauthorised_access(self, failure, request):
        """If service layer forbids access, we 403."""
        failure.trap(UnauthorisedError)
        request.setResponseCode(403)
        request.finish()

    def _404_missing_payments_and_receipts(self, failure, request):
        """Missing things are 404."""
        failure.trap(NoPaymentError, NoReceiptError)
        request.setResponseCode(404)
        # TODO: Add fancier 404
        request.finish()

    def _500_unhandled(self, failure, request):
        """Unhandled errors are 500."""
        failure.printTraceback()
        request.setResponseCode(500)
        if is_development_mode():
            request.setHeader("Content-Type", "text/plain")
            failure.printDetailedTraceback(request)
        request.finish()

    def render_GET(self, request):
        """Download the receipt and send it on."""
        deferred = self._model.download_receipt()
        deferred.addCallback(self._send_bytes, request)
        deferred.addErrback(self._403_unauthorised_access, request)
        deferred.addErrback(self._404_missing_payments_and_receipts, request)
        deferred.addErrback(self._500_unhandled, request)

        return NOT_DONE_YET
