# Copyright 2005-2017 Canonical Limited.  All rights reserved.

from zope.cachedescriptors.property import Lazy

import stripe

import transaction

from canonical.content.decorator import json

from canonical.landscape.stripe.customer import (
    get_card_details,
    get_customer,
    update_customer,
    update_credit_card,
)
from canonical.landscape.model.configuration import get_stripe_secret_key
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.billing.create import reset_payment_retries
from canonical.routes import Redirect
from canonical.url import get_url


class EditBillingPage(AccountPage):

    title = "Edit billing details"

    requires = "landscape.ManageBilling"

    def __call__(self):
        card = get_card_details(self._customer)
        if card is None:
            # There is nothing to edit
            raise Redirect(get_url(self.context, self.request))
        result = super(EditBillingPage, self).__call__()
        if self.request.method == "POST":
            self._handle_post()
        return result

    @Lazy
    def _customer(self):
        api_key = get_stripe_secret_key()
        customer_id = self.context.account.stripe_customer_id
        try:
            return get_customer(api_key, customer_id)
        except stripe.error.StripeError:
            return None

    @Lazy
    @json
    def credit_card_json(self):
        """Return card information from Stripe"""
        if self._customer:
            return get_card_details(self._customer)

    @Lazy
    @json
    def shipping_address_json(self):
        """Get the shipping address from Stripe."""
        if self._customer:
            shipping_address = self._customer.shipping.address
            return {
                "name": self._customer.shipping.name or "",
                "address1": shipping_address.line1 or "",
                "address2": shipping_address.line2 or "",
                "city": shipping_address.city or "",
                "state": shipping_address.state or "",
                "zip": shipping_address.postal_code or "",
                "country": shipping_address.country or "",
            }

    @Lazy
    @json
    def account_vat_json(self):
        """Get the VAT number of this account."""
        return self.context.account.vat_number

    @Lazy
    @json
    def account_type_json(self):
        """Get the VAT number of this account."""
        return "business" if self.context.account.is_business else "personal"

    @Lazy
    def edit_js_url(self):
        return self.static_root_path() + "/javascript/bundles/billing-edit-min.js"

    @Lazy
    def react_js_url(self):
        return self.static_root_path() + "/javascript/bundles/react-min.js"

    def _handle_post(self):
        """Update stripe and database with new billing details."""
        shipping = self._get_shipping_details(self.request.form)
        credit_card_details = self._get_credit_card_details(self.request.form)
        update_customer(self._customer, shipping)
        update_credit_card(self._customer, credit_card_details)
        event_log = EventLog(self.context.account, self.context.person)
        event_log.create_log_entry(
            self.context.account, "Updated payment method on account."
        )
        self.context.account.vat_number = self.request.form["accountVAT"]
        self.context.account.vat_country = self.request.form["accountCountry"]
        self.context.account.is_business = (
            self.request.form["accountType"] == "business"
        )
        reset_payment_retries(self.context.account)
        transaction.commit()
        self.request.response.redirect("/~/billing")

    def _get_shipping_details(self, form):
        """Return a dict with shipping data extracted from form data."""
        return {
            "name": form.get("accountName"),
            "address": {
                "line1": form.get("accountAddress1"),
                "line2": form.get("accountAddress2"),
                "city": form.get("accountCity"),
                "country": form.get("accountCountry"),
                "postal_code": form.get("accountZip"),
                "state": form.get("accountState"),
            },
        }

    def _get_credit_card_details(self, form):
        """Return a dict with the credit card detailis from form data."""
        field_map = (
            ("name", "billingName"),
            ("address_line1", "billingAddress1"),
            ("address_line2", "billingAddress2"),
            ("address_city", "billingCity"),
            ("address_country", "billingCountry"),
            ("address_zip", "billingZip"),
            ("address_state", "billingState"),
        )
        return {stripe_key: form.get(form_key) for stripe_key, form_key in field_map}
