# Copyright 2005-2017 Canonical Limited.  All rights reserved.

from builtins import object
from storm.twisted.transact import transact, Transactor

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue

from canonical.openstack.clientshelper import OpenStackClientsHelper
from canonical.landscape.model.configuration import get_billing_credentials
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.service.security import PersonPrincipal
from canonical.landscape.service.billing.payment import PaymentCollectionLookup


__clearlist__ = ["BillingResource", "NoReceiptError", "NoPaymentError"]


class NoPaymentError(Exception):

    """Failure to find given Payment."""

    def __init__(self, payment_id):
        super(NoPaymentError, self).__init__(
            "Could not find paid payment with id {}".format(payment_id)
        )


class NoReceiptError(Exception):

    """Failure to download receipt from Swift."""

    def __init__(self, receipt_uri):
        super(NoReceiptError, self).__init__(
            "Failed to download {} from Swift".format(receipt_uri)
        )


class BillingResource(object):

    openstack_helper = OpenStackClientsHelper()  # for testing

    transactor = Transactor(reactor.getThreadPool())  # for testing

    def __init__(self, account_id, payment_id, person_id):
        self._account_id = account_id
        self._payment_id = payment_id
        self._person_id = person_id

    def _get_receipt_filename(self, year, month, day):
        """Return the user-friendly receipt filename used for the download."""
        # NOTE, this filename is different from the one used within
        # Swift, this is by design, apparently it's better UX if we
        # don't include the receipt ID.
        return "Landscape_receipt_{0}-{1:02}-{2:02}.pdf".format(year, month, day)

    @transact
    def _get_payment_receipt(self):
        """Get the receipt URI and receipt filename for a given payment"""
        principal = PersonPrincipal(self._person_id)
        lookup = PaymentCollectionLookup(principal)
        account = get_account_by_id(self._account_id)
        collection = lookup(account)

        filtered = collection.with_ids([self._payment_id]).with_receipt_uri(True)
        payment = filtered.find().one()
        if payment is None:
            raise NoPaymentError(self._payment_id)
        filename = self._get_receipt_filename(
            payment.creation_time.year,
            payment.creation_time.month,
            payment.creation_time.day,
        )
        credentials = get_billing_credentials()
        return (payment.receipt_uri, filename, credentials)

    @inlineCallbacks
    def download_receipt(self):
        """Download the receipt, and return the bytes"""
        receipt_uri, filename, credentials = yield self._get_payment_receipt()

        # XXX TODO: figure out a better way to deal with "Empty credentials"
        assert credentials, "Empty credentials"
        container, path = receipt_uri.split("/", 1)
        swift_client = yield self.openstack_helper.get_client(
            "object-store", **credentials
        )
        try:
            contents = yield swift_client.get_object(container, path)
        except:
            raise NoReceiptError(receipt_uri)
        returnValue((contents, filename))
