# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime

from zope.schema import Choice
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.component import getUtility

from storm.locals import Store

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.security.interfaces import IRoleRegistry
from canonical.lib.timestamp import utctoday

from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.toolbar.page import ToolbarPage
from canonical.landscape.ui.license.license import NEVER_EXPIRY_DATE
from canonical.landscape.model.main.interfaces import IComputer
from canonical.landscape.model.license.management import (
    LicenseManagement,
    ALL_LICENSE_TYPES,
)
from canonical.landscape.model.main.license import License, InappropriateLicenseError


LICENSE_BASE_NAME = "Landscape"


def get_licenses_of_interest(account, current_license):
    """
    Return a list of licenses of interest.

    The licenses of interest are:
        - active licenses for the computer's account
        - the license the computer is linked to (if any)

    A license is considered active if it has not expired yet and
    if it still has seats free.
    """
    licenses = LicenseManagement(account).get_licenses(
        role_names=ALL_LICENSE_TYPES, active_only=True
    )
    licenses = [license for license in licenses if license.get_available_seats() > 0]

    # The current license - if given - should be the first element of the
    # resulting list. We thus don't want it sorted along with the other
    # active licenses.
    if current_license and current_license in licenses:
        licenses.remove(current_license)

    first = datetime.utcfromtimestamp(0).date()
    # Sort the active licenses by expiration date, latest date first.
    licenses = sorted(
        licenses,
        key=lambda license: license.expiration_date
        if license.expiration_date is not None
        else first,
        reverse=True,
    )

    if current_license:
        # Put the current license in front.
        licenses = [current_license] + licenses

    return licenses


def get_license_vocabulary(licenses):
    """
    Return a C{SimpleVocabulary} for the given L{License}s.

    The value of the vocabulary terms will be the C{License.id}. The title of
    the vocabulary terms will be a string combining the license name, the days
    left and the number of free seats.
    One term is always returned: the "No license" term with a value of "-1".
    """
    terms = [SimpleTerm("-1", title="No license")]
    title = ""

    registry = getUtility(IRoleRegistry)
    types_to_names = {
        "VMFeatures": "{} for virtual machines",
        "ContainerFeatures": "{} for containers",
        "TrialFeatures": "{} trial",
        "UbuntuProFeatures": "Ubuntu Pro",
        "UbuntuProFreeFeatures": "Ubuntu Pro Free",
    }
    keys_to_names = {
        registry.get(name).key: message for name, message in types_to_names.items()
    }

    for license in licenses:
        license_type = LICENSE_BASE_NAME
        if license.role_key in keys_to_names:
            license_type = keys_to_names[license.role_key].format(LICENSE_BASE_NAME)
        seats_free = license.get_available_seats()
        seat_unit = "seats" if seats_free > 1 else "seat"
        if license.expiration_date:
            time_left = license.expiration_date - utctoday()
            if time_left.days <= 0:
                title = "%s, expired since %s" % (
                    license_type,
                    license.expiration_date.strftime("%b %e, %Y"),
                )
            elif time_left.days > 1:
                if license.expiration_date is NEVER_EXPIRY_DATE:
                    # Special case for licenses that never expire.
                    title = "%s, never expires, %s %s free" % (
                        license_type,
                        seats_free,
                        seat_unit,
                    )
                else:
                    title = "%s, %s days left, %s %s free" % (
                        license_type,
                        time_left.days,
                        seats_free,
                        seat_unit,
                    )
            else:
                title = "%s, 1 day left, %s %s free" % (
                    license_type,
                    seats_free,
                    seat_unit,
                )
        else:
            title = "%s, never expires, %s %s free"
            title %= (license_type, seats_free, seat_unit)
        terms.append(SimpleTerm(str(license.id), title=title))
    return SimpleVocabulary(terms)


def get_current_license(computer):
    """Return the current license for the given L{Computer} or C{None}.

    We currently assume that each computer is linked to one license at most.
    This will need to be reworked once we adopt more sophisticated licensing
    models where a computer may be linked to multiple licenses of different
    types.
    """
    licenses = list(computer.licenses)
    assert len(licenses) < 2, "Computer linked to more than one license"
    return licenses[0] if licenses else None


def update_license(computer, old_id, new_id):
    """
    Update the license for the given L{Computer} according to the old and new
    C{License.id}.
    """
    if old_id == new_id:
        return None
    if new_id > 0:
        # Computer needs to be linked to a new license, add it.
        Store.of(computer).get(License, new_id).add_computer(computer)
    if old_id > 0:
        # Computer needs to be removed from a previous license, do it.
        Store.of(computer).get(License, old_id).remove_computer(computer)

    message = ""
    if old_id > 0:
        if new_id > 0:
            message = "The selected license has been applied to the computer."
        else:
            message = "Computer was removed from license."
    else:
        message = "Computer was added to license."

    return message


class EditComputerPage(ToolbarPage):

    title = "Edit computer"


class EditComputerForm(LandscapeFormProvider):

    summary = "Edit computer"
    description = "Please provide the following information."

    fields = Fields(IComputer, render_context=True)
    fields = fields.select("title", "comment")

    def __init__(self, *args, **kwds):
        super(EditComputerForm, self).__init__(*args, **kwds)
        current_license = get_current_license(self.context)
        licenses = get_licenses_of_interest(self.context.account, current_license)
        vocabulary = get_license_vocabulary(licenses)
        # Record the original license ID for this computer.
        if current_license:
            self.original_license_id = current_license.id
        else:
            self.original_license_id = -1
        description = "The license for this computer."
        self.fields += Fields(
            Choice(
                __name__="filtered_licenses",
                title="License",
                description=description,
                required=True,
                default=str(self.original_license_id),
                vocabulary=vocabulary,
            )
        )

    @action("Save", primary=True)
    def handle_edit_action(self, action, data):
        licensing_change = None
        new_license_id = int(data["filtered_licenses"])

        try:
            licensing_change = update_license(
                self.context, self.original_license_id, new_license_id
            )
        except InappropriateLicenseError as e:
            if e.license_type == "Container":
                self.view.notes.error(
                    "Computer '%s' cannot be associated with a "
                    "container license." % self.context.title
                )
            else:
                self.view.notes.error(
                    "Computer '%s' cannot be associated with a "
                    "virtual machine license." % self.context.title
                )

        # Remove the license from the form data so that the subsequent
        # computer property updates can go ahead without errors.
        data.pop("filtered_licenses")

        if licensing_change:
            self.view.notes.info(licensing_change)

        if self.apply_changes(self.context, data):
            self.view.notes.info("Changes saved.")
        elif not licensing_change:
            self.view.notes.info("No changes made.")
        self.request.response.redirect(get_url(self.context, self.request))
