# Copyright 2005-2022 Canonical Limited. All rights reserved.

from json import dumps

from zope.schema import Choice
from zope.security.interfaces import Unauthorized
from zope.security.management import checkPermission
from zope.security.proxy import removeSecurityProxy

import transaction

from canonical.content.form import Fields, action
from canonical.lib.text import describe_list
from canonical.routes import Redirect
from canonical.url import get_url

from canonical.landscape.model.computer.activities import (
    get_computer_activities_parents,
)
from canonical.landscape.model.activity.status import BLOCKED, PENDING_STATUSES
from canonical.landscape.ui.toolbar.page import ToolbarPage
from canonical.landscape.ui.reporting.dashboard import ComputerLinkView, ComputerLink
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.activity.schedule import get_deliver_after_time
from canonical.landscape.model.selection.info import ComputerSelectionInfo
from canonical.landscape.model.main.tag import InvalidTagError, get_tags, get_tag_names
from safe.canonical.landscape.model.main.computer import (
    add_tags_to_computers,
    remove_tags_from_computers,
    get_computer_by_id,
)
from canonical.landscape.model.main.license import (
    InappropriateLicenseError,
    LicenseSeatError,
)
from canonical.landscape.model.account.alerts import LicenseExpiryAlert
from canonical.landscape.ui.computer.computer import (
    get_license_vocabulary,
    get_licenses_of_interest,
    update_license,
)
from canonical.landscape.ui.hardware.hardware import (
    ComputerHardwareOverviewPage,
    NOT_AVAILABLE,
)
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.model.resource.hardware import HardwareInfoManagement
from canonical.landscape.model.resource.annotation import AnnotationManagement
from canonical.landscape.model.resource.cloud_instance_metadata import (
    CloudInstanceMetadataManagement,
)
from canonical.landscape.model.main.person import PersonAccessContextsVocabularyFactory
from canonical.landscape.service.computer.lookup import (
    AccountComputerTwinCollectionLookup,
)
from canonical.landscape.service.computer.accesscontextchange import (
    ComputerAccessContextChange,
)


def format_title_if_clone(computer):
    """Get the computer title with a clone suffix if it's cloned"""
    computer_name = computer.title
    clone_name = get_title_from_clone_id(computer.clone_id)
    if clone_name:
        if clone_name == computer_name:
            title = "{} (clone)".format(computer_name)
        else:
            title = "{} (clone of {})".format(computer_name, clone_name)
    else:
        title = computer_name
    return title


def get_title_from_clone_id(clone_id):
    """Get the computer title given a clone ID only if user has permissions"""
    if clone_id:
        cloned_computer = get_computer_by_id(clone_id)
        if cloned_computer:
            try:
                return cloned_computer.title
            except Unauthorized:
                return None
    return None


class CloneNameMixin(object):
    def format_computer_title(self, computer):
        """A helper function for formatting the computer title when cloned"""
        return format_title_if_clone(computer)


class ComputerInfoPage(ToolbarPage):
    """Display information for selected computers.

    @ivar pending_activities: list of pending activities for selected
        computers, limited to C{maximum_pending_activities}.
    @type pending_activities: C{list}

    @ivar unapproved_activities: list of unapproved activities for selected
        computers, limit to C{maximum_unapproved_activities}.
    @type unapproved_activities: C{list}

    @ivar pending_activities_count: total number of pending activities for
        selected computers.
    @type pending_activities_count: C{int}

    @ivar unapproved_activities_count: total number of unapproved activities
        for selected computers.
    @type unapproved_activities_count: C{int}
    """

    title = "Information"
    maximum_pending_activities = 15
    maximum_unapproved_activities = 15

    def __init__(self, context, request):
        super(ComputerInfoPage, self).__init__(context, request)
        self.computers = self.context.selection.get_selected()
        self.selection_info = ComputerSelectionInfo(self.context.selection)
        self._computer_count = self.context.selection.get_selected_count()
        self.is_many_computers = self._computer_count > 1
        if self.is_many_computers:
            self.computer_clause = "the selected computers"
        else:
            self.computer_clause = "this computer"

        if self._computer_count == 0:
            message = (
                "Computers must be selected before information " "can be displayed."
            )
            self.notes.error(message)
            transaction.commit()
            next_url = get_url(self.context.selection, self.request)
            raise Redirect(next_url.encode("utf-8"))

        # Used to disable/enable the Update Access Group Button
        self.disable_access_groups = "true"
        computers_by_access_group = {}
        for computer in self.computers:
            group = computers_by_access_group.setdefault(computer.access_context, [])
            group.append(computer.id)
            can_add_computer_to_group = checkPermission(
                "landscape.AddComputerToAccessContext", computer
            )
            can_remove_computer_from_group = checkPermission(
                "landscape.RemoveComputerFromAccessContext", computer
            )
            if can_add_computer_to_group and can_remove_computer_from_group:
                self.disable_access_groups = "false"

        self.access_group_links = {}
        data = {}
        for access_context, computer_ids in list(computers_by_access_group.items()):
            data["%s" % access_context.name] = sorted(computer_ids)
            data["title-%s" % access_context.name] = (
                "Computers in access group: %s" % access_context.name
            )
            url = '<a href="%s">%s</a>' % (
                get_url(access_context, self.request),
                access_context.title,
            )
            self.access_group_links[access_context.id] = ComputerLinkView(
                ComputerLink(
                    "%s" % access_context.name,
                    len(computer_ids),
                    "",
                    "are part of the %s access group" % url,
                    "is part of the %s access group" % url,
                ),
                self.request,
            )()

        self.computer_data = dumps(data).replace("</script", "<\\/script")

        # XXX: Consuming hardware information directly from the
        # ComputerHardwareOverviewPage is not advised.
        # The logic in there must be factored out into an API.
        self.hardware_info = ComputerHardwareOverviewPage(
            context, self.request
        ).hardware
        self.hardware_management = HardwareInfoManagement(self.account)

        self.annotation_management = AnnotationManagement(
            self.account, self.request.principal.person
        )

        self.alternate_annotation_content = None
        self._initialize_computer_details()

        self.has_operator_role = self.request.principal.person.has_global_role(
            "Operator"
        )

        # activities info
        account = removeSecurityProxy(self.context.selection.account)
        computer_ids = self.context.get_computer_ids()
        pending_activities = get_computer_activities_parents(
            account, computer_ids=computer_ids, statuses=PENDING_STATUSES, ordered=True
        )
        self.pending_activities_count = pending_activities.count()
        self.pending_activities = list(
            pending_activities.config(limit=self.maximum_pending_activities)
        )

        unapproved_activities = get_computer_activities_parents(
            account, computer_ids, statuses=[BLOCKED]
        )
        self.unapproved_activities_count = unapproved_activities.count()
        self.unapproved_activities = list(
            unapproved_activities.config(limit=self.maximum_unapproved_activities)
        )

    def get_unapproved_activities(self):
        """
        Get a list of unapproved activities for the 'unapproved-activities'
        provider.
        """
        return self.unapproved_activities

    def get_pending_activities(self):
        """
        Get a list of pending activities for the 'pending-activities' provider.
        """
        return self.pending_activities

    def _initialize_computer_details(self):
        """
        Initialize the variables displayed in the template, managing the cases
        where several computers are selected, and when the information is not
        available.
        """
        self._tag_counts = self.selection_info.get_tags_count()

        self.ping_exchange_diff_count = (
            self.selection_info.get_ping_exchange_diff_count()
        )

        if self.is_many_computers:
            self.computer_title = "%d computers selected" % self._computer_count
            self.computer_hostname = None
            self.computer_processor = None
            self.computer_memory = None
            self.computer_network = None
            self.computer_storage = None
            self.computer_video = None
            self.clone_id = None
            self.hypervisor = None
            self.container = None
            self.distribution = None
            access_contexts = set()
            distributions = set()
            clone_ids = set()
            self.annotations = None
            self.alternate_annotation_content = "Various"

            # if no computer has a reported ping time, we display the last in
            # the selection
            self.ping_time_computer = (
                self.selection_info.get_last_ping_time_computer() or self.computers[-1]
            )

            self.accept_time_computer = self.selection_info.get_last_accepted_computer()

            self.instance_metadata = None
            self.repository_profiles = None

            for computer in self.computers:
                distributions.add(computer.distribution)
                access_contexts.add(computer.access_context)
                clone_ids.add(computer.clone_id)

            if None in distributions:
                distributions.remove(None)
            if len(distributions) == 1:
                distribution = distributions.pop()
                self.distribution = "%s (%s)" % (
                    distribution.description,
                    distribution.code_name,
                )
            elif len(distributions) > 1:
                self.distribution = "Various"
            else:
                self.distribution = "Unknown"

            if len(access_contexts) == 1:
                self.access_context = access_contexts.pop()
            else:
                self.access_context = None

            if len(clone_ids) == 1:
                self.clone_id = clone_ids.pop()
                self.clone_name = get_title_from_clone_id(self.clone_id)
            else:
                self.clone_name = "Various"

        else:
            # We only have one computer selected.
            computer = self.computers[0]
            self.computer_title = computer.title

            self.computer_hostname = computer.hostname
            self.ping_time_computer = computer
            self.accept_time_computer = computer
            self.computer_comment = computer.comment
            if computer.distribution:
                description = computer.distribution.description
                code_name = computer.distribution.code_name
                self.distribution = "%s (%s)" % (description, code_name)
            else:
                self.distribution = "Unknown"

            self.computer_memory = self.hardware_info["memory"]["size"]

            # Check if model is a valid one otherwise try vendor.
            processor = self.hardware_info["cpu"]["model"]
            if processor == NOT_AVAILABLE:
                processor = self.hardware_info["cpu"]["vendor"]
            self.computer_processor = processor

            self.computer_storage = self.hardware_info["storage"]

            self.computer_network = self.hardware_info["network"]

            self.computer_video = self.hardware_info["display"]

            self.hypervisor = computer.vm_info

            self.container = computer.container_info

            self.clone_id = computer.clone_id
            self.clone_name = get_title_from_clone_id(self.clone_id)

            self.product = self.hardware_info["system"]["model"]
            self.serial = self.hardware_info["system"]["serial"]

            management = CloudInstanceMetadataManagement(computer)
            self.instance_metadata = management.get_instance_metadata()

            self.access_context = computer.access_context
            result = self.annotation_management.get_annotation([computer.id])
            if result:
                self.annotations = [
                    {"key": key, "value": value}
                    for key, value in sorted(result[computer.id].items())
                ]
            else:
                self.annotations = None
                self.alternate_annotation_content = "Not defined"

            repo_management = RepositoryManagement(
                self.context.account, person=self.context.person
            )
            profiles = repo_management.get_profiles_for_computer(computer)
            self.repository_profiles = [profile.name for profile in profiles]

    def get_tag_count(self, tag):
        return self._tag_counts.get(tag, 0)

    def tags(self):
        return sorted(self._tag_counts.keys())

    def reboot(self):
        """
        A web exposed method to restart or shutdown the selected
        computers.
        """
        self._check_request_csrf_token(require_post=True)
        form_keys = list(self.request.form.keys())
        if "restart" not in form_keys and "shutdown" not in form_keys:
            return self.request.response.redirect(get_url(self.context, self.request))

        deliver_after_time = get_deliver_after_time(self.request)
        if "restart" in self.request:
            parent = self.context.restart(deliver_after=deliver_after_time)
            self.notes.info("A restart is being requested.")
        elif "shutdown" in self.request:
            parent = self.context.shutdown(deliver_after=deliver_after_time)
            self.notes.info("A shutdown is being requested.")
        self.request.response.redirect(get_url(parent, self.request))

    def edit_comment(self):
        """
        A web exposed method to set the comment for the selected
        computers.
        """
        self._check_request_csrf_token(require_post=True)
        for computer in self.computers:
            computer.comment = self.request.get("form.comment", "")

        self.notes.info("Comment updated in selected computers.")
        url = get_url(self.context, self.request)
        self.request.response.redirect(url)

    def update_tags(self):
        """
        A web exposed method to update the tags of the selected
        computers.
        """
        self._check_request_csrf_token(require_post=True)
        if "add_tags" in self.request or "remove_tags" in self.request:
            person = self.request.principal.person
            add_tags = "add_tags" in self.request
            tags = self.request.get("tags", "").strip()
            if not tags and add_tags:
                self.notes.error("Tags must be provided before they can " "be added.")
            elif not tags and not add_tags:
                self.notes.error("Tags must be provided before they can be " "removed.")
            else:
                try:
                    tags_to_apply = get_tag_names(tags)
                except InvalidTagError:
                    self.notes.error(
                        "Tag names may only contain alphanumeric " "characters."
                    )
                    url = get_url(self.context, self.request)
                    self.request.response.redirect(url)
                    return

                tags = get_tags(tags_to_apply)
                if add_tags:
                    add_tags_to_computers(person, tags, self.computers)
                    self.notes.info(
                        "%s added to selected computers." % self._get_tag_text(tags)
                    )
                else:
                    remove_tags_from_computers(person, tags, self.computers)
                    included_tags = self.context.selection.get_included_tags()
                    for tag in tags:
                        if tag not in included_tags:
                            self.context.selection.discard_included_tags([tag])

                    self.notes.info(
                        "%s removed from selected computers." % self._get_tag_text(tags)
                    )
                removed = tags.intersection(self.context.selection.get_included_tags())
                if removed and "remove_tags" in self.request:
                    self.context.selection.discard_included_tags(removed)

        url = get_url(self.context, self.request)
        self.request.response.redirect(url)

    def _get_tag_text(self, tags):
        text = ", ".join([tag.name for tag in tags])
        if len(tags) > 1:
            return "%s tags" % text
        return "%s tag" % text


class EditComputerLicenseForm(LandscapeFormProvider):
    def __init__(self, context, request, view):
        super(EditComputerLicenseForm, self).__init__(context, request, view)
        self.computers = view.computers
        licenses = list(view.selection_info.get_licenses())
        default = licenses[0] if len(licenses) == 1 else None

        if view.is_many_computers:
            short_description = (
                "The license the selected computers " "are associated with."
            )
        else:
            short_description = "The license this computer is associated with."
        self.description = (
            "You can change here the license used by %s to contact "
            "Landscape, in particular when one license is about to "
            "expire." % (view.computer_clause,)
        )
        licenses = get_licenses_of_interest(self.context.account, default)
        vocabulary = get_license_vocabulary(licenses)
        if default is not None:
            default = str(default.id)
        self.fields = Fields(
            Choice(
                __name__="license_id",
                title="License",
                description=short_description,
                default=default,
                required=True,
                vocabulary=vocabulary,
            )
        )

    @action("Update license", primary=True)
    def save(self, action, data):
        """Save changes to the license of selected computers."""
        license_id = int(data.get("license_id", -1))

        updated = False
        inappropriate_computers = []
        unlicensed_computers = []
        for computer in self.computers:
            old_license = computer.licenses.one()
            old_id = old_license.id if old_license else -1
            try:
                if update_license(computer, old_id, license_id) is not None:
                    updated = True
            except InappropriateLicenseError:
                inappropriate_computers.append(computer.title)
            except LicenseSeatError:
                unlicensed_computers.append(computer.title)

        if updated:
            # Unproxy the account since persons don't have CheckRule
            # permission.
            unproxied_account = removeSecurityProxy(self.context.account)
            unproxied_account.check_rules([LicenseExpiryAlert])
            if self.view.is_many_computers:
                self.view.notes.info("License updated for the selected " "computers.")
            else:
                self.view.notes.info("License updated for this computer.")

        if inappropriate_computers:
            computers = describe_list("Computer", inappropriate_computers)
            self.view.notes.error(
                "%s cannot be associated with a " "virtual machine license." % computers
            )

        if unlicensed_computers:
            computers = describe_list("Computer", unlicensed_computers)
            self.view.notes.error(
                "%s could not be associated with the "
                "license, not enough seats remaining." % computers
            )

        self.request.response.redirect(get_url(self.context, self.request))


class EditComputerAccessContextForm(LandscapeFormProvider):
    """Form for editing the access context of the selected computer.

    This provider adapts the L{ComputerInfoPage} view.
    """

    def __init__(self, context, request, view):
        super(EditComputerAccessContextForm, self).__init__(context, request, view)
        self.person = request.principal.person
        if view.access_context is None:
            default = None
        elif view.access_context.id == context.account.access_context_id:
            default = context.account.access_context.name
        else:
            default = view.access_context.name

        readonly = view.disable_access_groups == "true"
        if readonly:
            # List Computers and their particular access group in info.pt
            self.fields = Fields()
        else:
            self.fields = Fields(
                Choice(
                    __name__="access_context",
                    title="Access group",
                    source=PersonAccessContextsVocabularyFactory(
                        "AddComputerToAccessContext", person=self.person
                    ),
                    default=default,
                    readonly=readonly,
                    description="The access group the computers should belong to.",
                    required=True,
                )
            )

    @action("Update access group", primary=True, name="update-access-group")
    def save(self, action, data):
        """Save changes to the access context of selected computers."""
        access_context_name = data.get("access_context")
        if access_context_name is None:
            # Use the global access context
            access_context_name = self.context.account.access_context_name

        lookup = AccountComputerTwinCollectionLookup(self.request.principal)
        collection = lookup(self.context.account, ids=self.context.get_computer_ids())
        operation = ComputerAccessContextChange(collection, self.request.principal)
        updated_ids = operation(access_context_name)
        updated_groups = {}
        readonly_groups = {}
        # Count number of computers updated/not-updated
        for computer in self.view.computers:
            if computer.id in updated_ids:
                group = updated_groups.setdefault(computer.access_context.title, 0)
                group += 1
            else:
                group = readonly_groups.setdefault(computer.access_context.title, 0)
                group += 1

        if self.view.is_many_computers:
            if readonly_groups:
                if updated_groups:
                    self.view.notes.warning(
                        "Access group only updated for computers in access "
                        "groups: %s." % ", ".join(list(updated_groups.keys()))
                    )
                self.view.notes.warning(
                    "Unable to update computers in access groups: %s. "
                    "Unauthorized" % ", ".join(list(readonly_groups.keys()))
                )
            else:
                self.view.notes.info(
                    "Access group updated for all selected " "computers."
                )
        else:
            if readonly_groups:
                self.view.notes.warning(
                    "Unable to update access group for this computer. " "Unauthorized"
                )
            else:
                self.view.notes.info("Access group updated for this computer.")
        self.request.response.redirect(get_url(self.context, self.request))
