# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from zope.interface import implementer
from zope.schema import Choice, Text
from zope.security.interfaces import Unauthorized

import transaction

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.model.main.interfaces import IPendingComputer
from canonical.landscape.model.main.license import (
    LicenseSeatError,
    CHECK_LDS_LICENSE_ERRORS,
)
from canonical.landscape.model.main.computer import (
    PendingComputerToAccept,
    ComputersForPendingComputersVocabularyFactory,
)
from canonical.landscape.model.main.person import PersonAccessContextsVocabularyFactory
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.errors import (
    UnknownAccessContextsError,
    UnknownAccessContextError,
)
from canonical.landscape.ui.form import LandscapeFormProvider, TagWidgetMixin
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.computer.info import CloneNameMixin
from canonical.landscape.ui.computer.interfaces import IPendingComputers
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)
from safe.canonical.landscape.model.main.computer import (
    get_computer_by_id,
    get_pending_computers_by_ids,
)


@implementer(IPendingComputers)
class PendingComputers(object):
    """Context object for the pending computers listing page."""

    def __init__(self, account, select_all=False):
        self.account = account
        self.checked = "checked" if select_all else None


class PendingComputersPage(AccountPage, TagWidgetMixin, CloneNameMixin):

    title = "Pending computers"

    def __init__(self, context, request):
        super(PendingComputersPage, self).__init__(context, request)
        account = self.context.account
        person = request.principal.person
        self.computers = list(account.get_pending_computers(person))

        self.have_client_access_groups = any(
            computer.access_group for computer in self.computers
        )
        # Basically duplicating the functionality of
        # PersonAccessContextsVocabularyFactory here, but since the pending
        # computers listing doesn't use a form we can't directly use a
        # vocabulary.
        context_ids = get_access_context_ids_for_person(
            account.access_context, "landscape.AcceptPendingComputer", person
        )
        self.access_groups = []
        for context_id in context_ids:
            context = context.account.get_access_context(context_id)
            item = {"name": context.name, "title": context.title}
            if (
                context.name == DEFAULT_ACCESS_CONTEXT_NAME
                and not self.have_client_access_groups
            ):
                item["selected"] = "selected"
            else:
                item["selected"] = None
            self.access_groups.append(item)
        if not self.access_groups:
            self.access_groups.append({"name": "", "title": "(None)"})

    def client_access_group(self, client_access_group):
        """Get a URL or None for a given client access group."""
        if not client_access_group:
            return
        try:
            [access_context] = self.context.account.get_access_contexts(
                names=[client_access_group]
            )
        except UnknownAccessContextsError:
            return
        else:
            return get_url(access_context, self.request)

    def update(self):
        """
        Handle actions on C{pending-computers/update}: accept selected pending
        computers, reject them, or select all elements (when javascript is
        disabled).
        """
        self._check_request_csrf_token()

        url = get_url(self.context, self.request)
        ids = [int(id) for id in self.request.get("computers", ())]
        if ids:
            pending_computers = list(get_pending_computers_by_ids(ids))
            if "accept" in self.request:
                self._accept_pending_computers(pending_computers)
            elif "reject" in self.request:
                self._reject_pending_computers(pending_computers)
        elif "accept" in self.request or "reject" in self.request:
            self.notes.error(
                "Pending computers must be selected before you "
                "can perform operations on them."
            )
        elif "select_all" in self.request:
            url = "%s?select_all" % (url,)

        self.request.response.redirect(url)

    def summary(self):
        count = len(self.computers)
        if not count:
            return "No pending computers"
        elif count == 1:
            return "1 pending computer"
        return "%d pending computers" % count

    def _filter_selected_computers(self, pending_computers):
        """
        Filter the selected pending computers to remove computers which
        refer to either invalid access groups or access groups which
        the current user doesn't have access to.
        """
        account = self.context.account
        person = self.request.principal.person
        for pending_computer in pending_computers:
            if not pending_computer.access_group:
                continue
            try:
                access_context = account.get_access_context_by_name(
                    person,
                    pending_computer.access_group,
                    "landscape.AcceptPendingComputer",
                )
            except UnknownAccessContextError:
                continue
            except Unauthorized:
                continue
            else:
                yield pending_computer, access_context

    def _accept_pending_computers(self, selected_computers):
        account = self.context.account
        account_title = account.title
        person = self.request.principal.person
        if not self.tag_validate(self.notes.error):
            return
        context_name = self.request.get("access_context")
        if not context_name:
            raise Unauthorized()
        if context_name == "*client*":
            pending = self._filter_selected_computers(selected_computers)
        else:
            access_context = account.get_access_context_by_name(
                person, context_name, "landscape.AcceptPendingComputer"
            )
            pending = ((c, access_context) for c in selected_computers)
        to_accept = []
        for pending_computer, access_context in pending:
            to_accept.append(
                PendingComputerToAccept(pending_computer, None, access_context)
            )
        try:
            computers = account.accept_computers(person, to_accept)
        except CHECK_LDS_LICENSE_ERRORS:
            transaction.abort()
            self.notes.error(
                "You don't have a valid license file. Please check"
                " /etc/landscape/license.txt."
            )
        except LicenseSeatError:
            # Even if we check for seats before, it can fail
            # because of concurrent accesses
            transaction.abort()
            self.notes.error(
                "There are not enough licenses available for %s." % (account_title,)
            )
        else:
            self.apply_tags(person, computers)
            n_pending = len(to_accept)
            if n_pending == 1:
                self.notes.info(
                    "%s accepted into %s."
                    % (to_accept[0].computer.title, account_title)
                )
            elif n_pending > 1:
                self.notes.info(
                    "%d computers accepted into %s." % (n_pending, account_title)
                )
            n_invalid_access_groups = len(selected_computers) - n_pending
            if n_invalid_access_groups == 1:
                self.notes.warning(
                    "Unable to accept 1 computer into its "
                    "client provided access group."
                )
            elif n_invalid_access_groups > 1:
                self.notes.warning(
                    "Unable to accept %d computers into their "
                    "client provided access groups." % n_invalid_access_groups
                )

    def _reject_pending_computers(self, pending_computers):
        account = self.context.account
        account_title = account.title
        person = self.request.principal.person
        n_pending = len(pending_computers)
        if n_pending == 1:
            self.notes.info(
                "%s rejected from %s." % (pending_computers[0].title, account_title)
            )
        elif n_pending > 1:
            self.notes.info(
                "%d computers rejected from %s." % (n_pending, account_title)
            )
        account.reject_computers(person, pending_computers)


class PendingComputerPage(AccountPage, TagWidgetMixin):

    title = "Pending computer"


class ActivationForm(LandscapeFormProvider, TagWidgetMixin):

    fields = Fields(IPendingComputer, render_context=True)

    def __init__(self, context, request, view):
        super(ActivationForm, self).__init__(context, request, view)
        computer_field = Choice(
            __name__="computer",
            title="Computer",
            description="""
                If an existing computer lost its registration and is now
                trying to re-register itself select the existing computer
                here.  This will prevent a duplicate computer from being
                created and make sure its data doesn't get fragmented.  If
                an existing computer isn't selected, this pending computer
                will become a new computer in the system once accepted.""",
            required=False,
            source=ComputersForPendingComputersVocabularyFactory(
                context.account, request.principal.person
            ),
        )
        self.fields += Fields(computer_field)
        default_access_group = DEFAULT_ACCESS_CONTEXT_NAME
        if context.access_group:
            try:
                context.account.get_access_context_by_name(
                    request.principal.person,
                    context.access_group,
                    "landscape.AcceptPendingComputer",
                )
            except (UnknownAccessContextError, Unauthorized):
                client_provided_access_group_field = Text(
                    __name__="access_group",
                    title="Client provided access group",
                    description="""
                    The access group the computer was configured in the client
                    to belong to.""",
                    readonly=True,
                )
                self.fields += Fields(
                    client_provided_access_group_field, render_context=True
                )
            else:
                default_access_group = context.access_group
        access_context_field = Choice(
            __name__="access_context",
            title="Access group",
            source=PersonAccessContextsVocabularyFactory(
                "AcceptPendingComputer", person=request.principal.person
            ),
            default=default_access_group,
            description="The access group the computer should belong to.",
            required=True,
        )
        self.fields += Fields(access_context_field)

    def validate(self, action, data):
        self.tag_validate(self.add_error)

    @action("Accept", primary=True)
    def handle_accept(self, action, data):
        access_context_name = data.pop("access_context")
        self.apply_changes(self.context, data)
        computer_id = data.get("computer")
        account = self.context.account
        person = self.request.principal.person

        computer = None
        if computer_id:
            computer = get_computer_by_id(int(computer_id))

        access_context = account.get_access_context_by_name(
            person, access_context_name, "landscape.AcceptPendingComputer"
        )

        url = "%s/pending-computers" % get_url(account, self.request)
        try:
            new_computer = self.context.accept(
                person, computer, access_context=access_context
            )
        except CHECK_LDS_LICENSE_ERRORS:
            self.view.notes.error(
                "You don't have a valid license file. Please check "
                "/etc/landscape/license.txt."
            )
        except LicenseSeatError:
            # Even if we check for seats before, it can fail because of
            # concurrent accesses
            self.view.notes.error(
                "There are no licenses available for %s. "
                "%s was not accepted." % (account.title, self.context.title)
            )
        else:
            self.apply_tags(person, [new_computer])
            url = "%s/info" % get_url(new_computer, self.request)
            self.view.notes.info(
                "%s has been accepted into %s." % (new_computer.title, account.title)
            )
        self.request.response.redirect(url)

    @action("Reject")
    def handle_reject(self, action, data):
        self.context.reject(self.request.principal.person)
        account = self.context.account
        self.view.notes.info(
            "%s has been rejected from %s."
            % (self.context.title, self.context.account.title)
        )
        url = get_url(account, self.request)
        self.request.response.redirect("%s/pending-computers" % (url,))
