# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from zope.schema import Choice, TextLine
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.security.management import checkPermission
from zope.security.interfaces import Unauthorized

from canonical.content.form import Fields, action

from canonical.url import get_url

from canonical.landscape.model.account.interfaces import ICustomGraph
from canonical.landscape.model.main.person import PersonAccessContextsVocabularyFactory
from canonical.landscape.model.account.customgraph import (
    CustomGraphManagement,
    DuplicateCustomGraphError,
)
from canonical.landscape.model.script.script import (
    split_code_on_interpreter,
    EmptyScriptInterpreterError,
    EmptyScriptCodeError,
)
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import (
    LandscapeFormProvider,
    LongTextWidget,
    MultiCheckBoxWidget,
    TagAssociationHelper,
)
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)


class CustomGraphsPage(AccountPage):

    title = "Graphs"

    def __init__(self, context, request):
        super(CustomGraphsPage, self).__init__(context, request)
        self.custom_graphs = list(self.context.get_graphs())
        self.can_create = self.context.person_can("CreateCustomGraph")
        remove_access_contexts = get_access_context_ids_for_person(
            self.context.account.access_context,
            "EditCustomGraph",
            request.principal.person,
        )
        self.can_remove = "false"
        for custom_graph in self.custom_graphs:
            if custom_graph.access_context.id in remove_access_contexts:
                self.can_remove = "true"

    def do_removals(self, selected_graph_ids):
        graphs = [
            graph for graph in self.custom_graphs if graph.id in selected_graph_ids
        ]
        remaining = self.context.remove_graphs(graphs)
        if not remaining:
            return
        if len(remaining) == 1:
            message = "Unable to remove graph: %s. Unauthorised."
        else:
            message = "Unable to remove graphs: %s. Unauthorised."
        return message % ", ".join(str(graph.title) for graph in remaining)

    def update(self):
        """
        Method is called by Zope's traversal machinery when an action is called
        on the graph selection page.  This method should only be called
        when form data has been POST'd to update the selection; it always
        performs a redirect.
        """
        self._check_request_csrf_token(require_post=True)

        if "remove" in self.request:
            selected_graph_ids = self.request.form.get("graph_id", ())
            if selected_graph_ids:
                selected_graph_ids = [int(id) for id in selected_graph_ids]
                message = self.do_removals(selected_graph_ids)
                if message:
                    self.notes.warning(message)
                self.request.response.redirect(get_url(self.context, self.request))
        elif "create" in self.request:
            self.request.response.redirect(
                get_url(self.context, self.request) + "/create"
            )
        elif "select_all" in self.request:
            # Only used when javascript is disabled
            self.request.response.redirect(
                get_url(self.context, self.request) + "?select_all=1"
            )
        else:
            self.request.response.redirect(get_url(self.context, self.request))

    def get_selection_message(self, graph):
        """
        Generate a message reporting the tags this graph is associated with,
        and hence the computers it is enabled for.
        """
        access_group_message = " in the <a href='%s'>%s</a> access group" % (
            get_url(graph.access_context, self.request),
            graph.access_context.title,
        )

        if graph.all_computers:
            message = "All computers"
        else:
            tags = sorted([tag.name for tag in graph.get_tags()])
            if not tags:
                message = "No computers"
            else:
                message = "Computers with %s %s" % (
                    ", ".join(tags),
                    "tag" if len(tags) == 1 else "tags",
                )
        return message + access_group_message


class CreateCustomGraphPage(AccountPage, TagAssociationHelper):
    """Page used to create graphs.

    This page is similar to the script creation page.

    @ivar title: The page title.
    """

    title = "Create graph"

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        TagAssociationHelper.__init__(self, context, request)
        if not context.person_can("CreateCustomGraph"):
            raise Unauthorized()
        self.disable_save = "false"


class CreateCustomGraphForm(LandscapeFormProvider):

    summary = "Create graph"

    fields = Fields(ICustomGraph)

    fields["title"].custom_widget = LongTextWidget
    fields["tags"].custom_widget = MultiCheckBoxWidget

    dropdown_css_class = "dropdown-spacer-customchart"

    def __init__(self, context, request, view):
        super(CreateCustomGraphForm, self).__init__(context, request, view)
        self.access_context_factory = PersonAccessContextsVocabularyFactory(
            "CreateCustomGraph", person=self.request.principal.person
        )
        access_context_choice = Choice(
            __name__="access_context",
            title="Access group",
            source=self.access_context_factory,
            default="Global",
            description="The access group the graph should belong to.",
            required=True,
        )
        self.fields += Fields(access_context_choice)

    def get_initial_data(self):
        return {"access_context": self.context.account.access_context_id}

    def validate(self, action, data):
        super(CreateCustomGraphForm, self).validate(action, data)
        try:
            split_code_on_interpreter(data.get("code"))
        except EmptyScriptCodeError:
            self.add_error("Script does nothing", "code")
        except EmptyScriptInterpreterError:
            self.add_error("Script needs an interpreter", "code")

    @action("Save", primary=True)
    def save(self, action, data):
        try:
            graph = self.context.create_graph(
                data["title"],
                data["y_axis_title"],
                data["username"],
                data["code"],
                data.get("access_context"),
            )
        except DuplicateCustomGraphError:
            self.add_error("This title is already used in another graph.")
            return

        self.view.notes.info("'%s' has been created." % graph.title)
        url = get_url(self.context.account, self.request) + "/custom-graphs"
        return self.view.update("graph", True, url, data, graph)


class EditCustomGraphPage(AccountPage, TagAssociationHelper):

    title = "Edit graph"

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        TagAssociationHelper.__init__(self, context, request)
        self.disable_save = "true"
        if checkPermission("landscape.EditCustomGraph", context):
            self.disable_save = "false"


class EditCustomGraphForm(LandscapeFormProvider):

    summary = "Edit graph"

    fields = Fields(ICustomGraph, render_context=True)

    fields["title"].custom_widget = LongTextWidget
    fields["tags"].custom_widget = MultiCheckBoxWidget

    def __init__(self, context, request, view):
        super(EditCustomGraphForm, self).__init__(context, request, view)
        if view.disable_save == "true":
            self.summary = "View graph"
        access_context = TextLine(
            __name__="access_context",
            title="Access group",
            default=context.access_context.title,
            readonly=True,
        )
        self.fields += Fields(access_context)
        self.management = CustomGraphManagement(
            self.context.account, self.request.principal.person
        )

    def get_initial_data(self):
        data = {
            "title": self.context.title,
            "code": self.context.get_code(),
            "username": self.context.username,
            "y_axis_title": self.context.y_axis_title,
            "tags": self.view.tag_names,
        }
        return data

    def validate(self, action, data):
        """
        If the title changed, check that it isn't already used.
        """
        title = data.get("title")
        if title:
            graph = self.management.get_graphs(titles=[title], permission=None).one()
            if graph and graph.id != self.context.id:
                self.add_error("This title is already used in another graph.")
        try:
            split_code_on_interpreter(data.get("code"))
        except EmptyScriptCodeError:
            self.add_error("Script does nothing", "code")
        except EmptyScriptInterpreterError:
            self.add_error("Script needs an interpreter", "code")

    @action("Save", primary=True)
    def save(self, action, data):
        tag_data = {}
        tag_data["tags"] = data.pop("tags", [])
        tag_data["all_computers"] = data.pop("all_computers", False)

        changed = self.management.edit_graph(self.context.id, data)
        if changed:
            self.view.notes.info("'%s' has been updated." % self.context.title)

        url = get_url(self.management, self.request)
        return self.view.update("graph", changed, url, tag_data)


class CustomGraphErrorsPage(AccountPage):
    """
    This outputs computers and their current errors for a given graph.
    """

    title = "Custom errors"

    def get_errors(self):
        error_messages = self.context.get_error_messages(self.request.principal.person)
        error_computers = sorted(iter(error_messages.keys()), key=lambda x: x.title)
        for computer in error_computers:
            error = error_messages[computer]
            yield {"computer": computer, "message": error}


def get_account_tags(custom_graph):
    """
    This is a simple vocabulary which provides the list of tags for a custom
    graph for editing on the graph editor.
    """
    tag_names = [tag.name for tag in custom_graph.account.get_tags()]
    vocabulary_terms = []
    for tag_name in tag_names:
        term = SimpleTerm(
            tag_name, tag_name.encode("ascii", "xmlcharrefreplace"), title=tag_name
        )
        vocabulary_terms.append(term)
    return SimpleVocabulary(vocabulary_terms)
