# Copyright 2005-2022 Canonical Limited. All rights reserved.

import logging
import sys
import traceback

from zope.browser.interfaces import ISystemErrorView
from zope.authentication.interfaces import IAuthentication
from zope.component import queryUtility, getUtility
from zope.interface import implementer
from zope.cachedescriptors.property import Lazy

from canonical.landscape.constants import UBUNTU_ADVANTAGE_PURCHASE_URL
from canonical.landscape.security.authentication import LandscapeAuthentication
from canonical.landscape.ui.page.page import LandscapePage
from canonical.oops.interfaces import IOOPSManager


@implementer(ISystemErrorView)
class ExceptionPage(LandscapePage):
    """Base page for all exceptions."""

    title = "Error"
    ubuntu_advantage_url = UBUNTU_ADVANTAGE_PURCHASE_URL

    @Lazy
    def exception(self):
        exc_type, exc, tb = sys.exc_info()
        return {
            "name": exc_type.__name__,
            "message": str(exc),
            "traceback": "\n".join(line.strip() for line in traceback.format_tb(tb)),
        }

    def isSystemError(self):
        return False


class SystemExceptionPage(ExceptionPage):
    def __init__(self, context, request):
        super(SystemExceptionPage, self).__init__(context, request)
        oops_manager = queryUtility(IOOPSManager)
        self.oops_id = None
        if oops_manager:
            oops = oops_manager.get("request")
            if oops is not None and oops.wanted:
                self.oops_id = oops_manager.dump(oops)
                logging.error("Dumping %s.", self.oops_id)
                oops.get("metadata")["id"] = self.oops_id

    def isSystemError(self):
        return True


@implementer(ISystemErrorView)
class ClientExceptionView(object):
    """A view for exceptions which occur during client processing."""

    def __init__(self, context, request):
        pass

    def isSystemError(self):
        """Errors of this type should be logged, so return True."""
        return True

    def __call__(self, REQUEST):
        # Note: The parameter must be called REQUEST because there is
        # a (undocumented? unsupported?) special case in
        # zope.publisher.publish.mapply which passes the request
        # object to functions with REQUEST as a parameter.

        # XXX At some point we may want to return rich data, in
        # bpickle format? Currently the client simply ignores content
        # for status 500 messages.
        REQUEST.response.setStatus(500)
        return "Error handling client request"


class UnauthorizedPage(SystemExceptionPage):
    """Page for unauthorized exceptions."""

    def __call__(self):
        authentication = LandscapeAuthentication()
        authentication.unauthorized(self.request.principal.id, self.request)
        return super(UnauthorizedPage, self).__call__()


def NoPage(context, request):
    return None
