# Copyright 2005-2022 Canonical Limited. All rights reserved.

from xml.sax.saxutils import escape

from zope.browserpage import ViewPageTemplateFile
from zope.formlib.interfaces import InputErrors
from zope.formlib.widget import renderElement
from zope.formlib.widgets import ListSequenceWidget
from zope.security.proxy import removeSecurityProxy
from zope.security.interfaces import Unauthorized
from zope.security.management import checkPermission
from zope.formlib.widget import SimpleInputWidget
from zope.formlib.textwidgets import TextWidget, BytesWidget
from zope.formlib.widgets import (
    CheckBoxWidget as _CheckBoxWidget,
    MultiCheckBoxWidget as _MultiCheckBoxWidget,
    DropdownWidget as _DropdownWidget,
    SelectWidget,
)
from zope.schema import TextLine

from landscape.lib.tag import is_valid_tag

from canonical.content.form import TemplateFormContentProvider
from canonical.url import get_url

from canonical.landscape.model.main.tag import get_tags, get_tag_names
from canonical.landscape.model.main.tag import InvalidTagError
from safe.canonical.landscape.model.main.computer import add_tags_to_computers
from canonical.lib.secrets import SecretUnicode
from canonical.lib.text import describe_list


SelectWidget._messageNoValue = "Please select"


class LandscapeFormProvider(TemplateFormContentProvider):
    """
    Specialized template form content provider writes global errors
    using C{view.notes.error}.
    """

    render_global_errors = False

    def add_error(self, error, field_name=None):
        """Add an error message to the form.

        Global errors apply to the entire form and are emitted as
        error L{notes<canonical.notes.interfaces.INoteContainer>}.  If
        C{field_name} is provided it will be used with the standard
        error logic for field errors.
        """
        super(LandscapeFormProvider, self).add_error(error, field_name)
        error = str(error)
        if not field_name:
            self.view.notes.error(error)


class InputWrapperWidget(SimpleInputWidget):
    """An input widget that wraps the input value.

    Subclasses should set wrap_value to a function that wraps the
    given value.
    """

    wrap_input = None

    @classmethod
    def subclass(cls, base, wrap_input):
        """Return a subclass of this class and the given base class."""
        if not issubclass(base, SimpleInputWidget):
            raise NotImplementedError
        sub = type(cls.__name__, (cls, base), {})
        sub.wrap_input = wrap_input
        return sub

    def getInputValue(self):
        if self.wrap_input is None:
            raise TypeError("wrap_input must be a function")
        value = super(InputWrapperWidget, self).getInputValue()
        return self.wrap_input(value)


class LongTextWidget(TextWidget):
    """
    A text widget with a longer size, used for descriptions.
    """

    displayWidth = 50


class LongBytesWidget(BytesWidget):

    displayWidth = 50


class CheckBoxWidget(_CheckBoxWidget):
    """This is a wrapper for the CheckBoxWidget in Zope3.

    It wraps the checkbox element in a label, adding an extra empty span for
    styling.
    """

    def __call__(self):
        markup = super(CheckBoxWidget, self).__call__()
        # The enclosing <label> element needs a "for" attribute for Firefox.
        return "<label for='%s'>%s<span class='checkbox'></span></label>" % (
            self.name,
            markup,
        )


class MultiCheckBoxWidget(_MultiCheckBoxWidget):
    """This is a wrapper for the MultiCheckBoxWidget in Zope3.

    It automatically provides a vocabulary for the selection and adds markup
    for stying checkboxes.
    """

    def __init__(self, field, request):
        vocabulary = field.value_type.vocabulary
        super(MultiCheckBoxWidget, self).__init__(field, vocabulary, request)

    def _renderItem(self, index, text, value, name, cssClass, checked=False):
        # This duplicates the parent class method, adding an extra span.
        kw = {}
        if checked:
            kw["checked"] = "checked"
        id = "%s.%s" % (name, index)
        elem = renderElement(
            "input",
            type="checkbox",
            cssClass=cssClass,
            name=name,
            id=id,
            value=value,
            **kw
        )
        contents = elem + renderElement(
            "span", cssClass="checkbox", contents=escape(text)
        )
        return renderElement("label", contents=contents, **{"for": id})


def DropdownWidget(field, request):
    """This is a wrapper for the DropdownWidget in Zope3."""
    vocabulary = field.vocabulary
    widget = _DropdownWidget(field, vocabulary, request)
    widget.explicit_empty_selection = False
    return widget


class TagWidgetMixin(object):
    """
    Mixin for adding a tag selection/edit widget to your page or form.

    It adds facility for validating and applying tags, and expects a context
    with a C{account} attribute.
    """

    additional_content = ViewPageTemplateFile("tag-widget.pt")

    def tags(self):
        """
        Return the list of tags that will be displayed in the selection widget.
        """
        return self.context.account.get_tags()

    def tag_validate(self, add_error):
        """
        Check that all the posted tags only contains alphanumeric characters.

        @param add_error: function called with the error message as parameter
            when a tag doesn't match the specification. Typically it will be
            C{self.add_error} for a L{LandscapeFormProvider}, and
            C{self.notes.error} for a page. It will be called only once even if
            several tags doesn't match.

        @return: C{True} if all the tags are OK, or C{False if at least one tag
            doesn't match.
        """
        tags = set(self.request.form.get("tags", "").split())
        for tag in tags:
            if not is_valid_tag(tag):
                add_error("Tag names may only contain alphanumeric characters.")
                return False
        return True

    def apply_tags(self, person, computers):
        """
        Add the posted tags to the given list of computers. Tags must have been
        validated using C{tag_validate} before.
        """
        tags = set(self.request.form.get("tags", "").split())
        tags = get_tags(tags)
        add_tags_to_computers(person, tags, computers)


class TagAssociationHelper(object):
    """
    Helper class to associate an object supporting C{ComputerGroup} with a tag.

    @cvar tag_association_form: template form allowing to choose tags, to be
        include in another template.
    @type tag_association_form: C{ViewPageTemplateFile}
    """

    tag_association_form = ViewPageTemplateFile("tag-association.pt")

    update_requires = None

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.all_tags = list(context.account.get_tags())
        if getattr(context, "get_tags", None) is not None:
            self.tags = removeSecurityProxy(context.get_tags())
            self.tag_names = sorted([tag.name for tag in self.tags])
        else:
            self.tags = set()
            self.tag_names = []

        if self.update_requires is not None:
            self.can_update = checkPermission(self.update_requires, self.context)
        else:
            self.can_update = True

    def update(self, item, changed=False, url=None, data=None, context=None):
        """Method to call to update the tag association of an object.

        @param item: string representation of the object manipulated.
        @type item: C{str}
        @param changed: if C{True}, don't print "No changes" notice even if no
            tags changed.
        @type changed: C{bool}
        @param url: if specified, the URL where the request is going to be
            redirected, instead of the context URL.
        @type url: C{str}
        @param data: object holding form data, default to the request.
        @type data: C{dict}-like
        @param context: the context object, default to the current context.
        @type context: C{object}
        """
        if data is None:
            data = self.request
        if context is None:
            context = self.context
        if not url:
            url = get_url(context, self.request)

        if self.update_requires is not None:
            if not checkPermission(self.update_requires, context):
                raise Unauthorized()

        all_computers = bool(data.get("all_computers"))
        selected_tag_names = data.get("tags", [])
        try:
            new_tag_names = get_tag_names(data.get("new_tag_name", ""))
        except InvalidTagError:
            self.notes.error("Tag names may only contain alphanumeric characters.")
            return
        else:
            selected_tag_names.extend(new_tag_names)
        selected_tags = get_tags(selected_tag_names)

        if not (all_computers == context.all_computers and selected_tags == self.tags):
            if hasattr(context, "update_association"):
                if all_computers != context.all_computers:
                    if all_computers:
                        selected_tags = None
                        self.notes.info("This %s is enabled for all computers." % item)
                    else:
                        if selected_tags:
                            self.notes.info(
                                "This %s is enabled for computers with "
                                "selected tags." % item
                            )
                changes = context.update_association(
                    new_tags=selected_tags,
                    all_computers=all_computers,
                    person=self.request.principal.person,
                )
                if not all_computers:
                    if changes.get("tags_added", []):
                        self.notes.info(
                            "Added tag %s to %s."
                            % (describe_list(items=changes["tags_added"]), item)
                        )
                    if changes.get("tags_removed", []):
                        self.notes.info(
                            "Removed tag %s from %s."
                            % (describe_list(items=changes["tags_removed"]), item)
                        )
                if not all_computers and not selected_tags:
                    self.notes.warning("This %s is disabled for all computers." % item)
        elif not changed:
            self.notes.info("No changes.")

        self.request.response.redirect(url)


class SimpleListSequenceWidget(ListSequenceWidget):
    """
    A simplified list sequence widget, that doesn't require particular markup.
    """

    def hasInput(self):
        return self.name in self.request.form

    def _generateSequence(self):
        sequence = []
        for i, value in enumerate(self.request.form[self.name]):
            widget = self._getWidget(i)
            self.request.form[widget.name] = value
            try:
                sequence.append(widget.getInputValue())
            except InputErrors:
                self._error = widget._error
                raise self._error
        return sequence


class SecretTextLine(TextLine):
    """A zope.schema.TextLine with a secret value."""

    custom_widget = InputWrapperWidget.subclass(TextWidget, SecretUnicode)

    def fromUnicode(self, str):
        str = super(SecretTextLine, self).fromUnicode()
        if not isinstance(str, str):
            return str
        return SecretUnicode(str)

    def get(self, object):
        value = super(SecretTextLine, self).get(object)
        if not isinstance(value, str):
            return value
        return SecretUnicode(value)

    def query(self, object, default=None):
        value = super(SecretTextLine, self).query(object, default=default)
        if not isinstance(value, str):
            return value
        return SecretUnicode(value)
