# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import range

__all__ = ["split", "is_valid_registration_key", "sanitize_csv_value"]


def split(items, columns):
    """
    Split a list of items into a list of lists, with each inner list
    containing items to display in a particular column.

    When the number of items isn't a perfect factor of C{columns}, the
    overflow is evenly distributed in the right-most columns.  For example,
    given 8 items with 3 columns requested, the following arrangement of
    columns will be produced::

       1 4 7
       2 5 8
       3 6

    @param items: The list of items to split into columns.  The elements in
        this list will not be sorted automatically.
    @param columns: The number of columns to split items into.
    """
    if columns < 1:
        raise RuntimeError("columns must be greater than 0.")

    result = []
    # If we have equal-to or fewer items than columns requested, separate the
    # items into single-item columns and pad the result with empty columns, up
    # to the number columns requested.
    if len(items) <= columns:
        for item in items:
            result.append([item])
        result.extend([[] for i in range(columns - len(result))])
    # If we have more items than columns, and more than one column requested,
    # we'll split them in a way that leaves underflowed columns on the
    # right-hand side.
    else:
        items_per_column = len(items) // columns
        overflow = len(items) % columns
        end = 0
        for i in range(columns):
            begin = end
            end = begin + items_per_column
            if overflow:
                end = end + 1
                overflow = overflow - 1
            result.append(items[begin:end])
    return result


def is_valid_registration_key(text):
    """
    Landscape client files cannot contain certain characters
    """
    if "#" in text or ";" in text:
        return False
    if text != text.strip():  # Trailing spaces
        return False
    return True


def sanitize_csv_value(value):
    """
    Protects against spreadsheet formula injection based attacks
    """
    if value:
        if value.startswith("="):
            value = "'" + value
    return value
