# Copyright 2005-2022 Canonical Limited. All rights reserved.

import csv
import itertools
import logging
import sys
import re
from collections import defaultdict

from io import StringIO

from canonical.lib.hardware import HardwareFormatter

from canonical.content.download import set_csv_download
from canonical.landscape.ui.toolbar.page import ComputerPaginatorPage
from canonical.landscape.model.resource.hardware import HardwareInfoManagement

NOT_AVAILABLE = "Not available"


class ComputerHardwareOverviewPage(ComputerPaginatorPage):

    title = "Hardware"

    # For each hardware class, specify a list of regexpes to filter out or
    # match keys.
    _hardware_classes = {
        "system": {"filters": [], "matches": ["core.", "^configuration"]},
        "cpu": {"filters": [], "matches": ["cpu"]},
        "memory": {"filters": [], "matches": ["^memory"]},
        "network": {"filters": [], "matches": ["^pci.*network"]},
        "storage": {"filters": [], "matches": ["disk"]},
        "display": {"filters": [], "matches": ["display"]},
        "multimedia": {"filters": [], "matches": ["multimedia"]},
        "pci": {"filters": ["multimedia", "usb"], "matches": ["pci"]},
        "usb": {"filters": [], "matches": ["usb"]},
        "scsi": {"filters": [], "matches": ["scsi"]},
    }

    def __init__(self, context, request):
        super(ComputerHardwareOverviewPage, self).__init__(context, request)

        self.management = HardwareInfoManagement(self.account)
        hardware_info = self.management.get_hardware_for_computer_ids(
            [self.computer.id]
        )[self.computer.id]
        self.hardware = self._get_hardware(hardware_info)
        self._validate_hardware(hardware_info)

    def _validate_hardware(self, hardware_info):
        """
        Quick health check to ensure we have extracted some valid hardware
        data appropriately. If we have significant gaps in data, we will
        log some warnings in landscape logs.
        """
        if len(hardware_info) == 0:
            return

        # Collect counts of potential key matches from lshw.
        # We have a problem if we have a count of potential key matches but
        # no hardware data found for that class.
        counts = dict.fromkeys(self._hardware_classes, 0)

        all_hardware = [key for key, _, _ in hardware_info]

        # Count up potential hardware class matches from all_hardware
        for hardware_name in all_hardware:
            for hardware_class, config in self._hardware_classes.items():
                # Check match patterns for a potential match for each hardware
                # class. Tally each match and move to the next hardware_name.

                filters = config["filters"]
                if any(re.search(regex, hardware_name) for regex in filters):
                    continue

                matches = config["matches"]
                if any(re.search(regex, hardware_name) for regex in matches):
                    counts[hardware_class] += 1
                    break

        warn_hardware_class_data = []
        for hardware_class, devices in list(self.hardware.items()):
            valid_data = False
            if isinstance(devices, dict):
                for attribute, value in list(devices.items()):
                    if value != NOT_AVAILABLE:
                        valid_data = True
            elif isinstance(devices, list):
                if len(devices) > 0:
                    valid_data = True
            else:
                if devices != NOT_AVAILABLE:
                    valid_data = True

            # Only append warning if we don't see any valid_data, but
            # we have potential matches in hardware_class_data
            if not valid_data:
                if counts[hardware_class] > 0:
                    warn_hardware_class_data.append(hardware_class)

        if len(warn_hardware_class_data) > 0:
            logging.warn(
                "Hardware Data: Could not parse %s data for computer %s. "
                "There are %s attributes seen from lshw which could be "
                "related to this data."
                % (
                    ", ".join(warn_hardware_class_data),
                    self.computer.id,
                    ", ".join(
                        "%d %s" % (counts[key], key) for key in warn_hardware_class_data
                    ),
                )
            )

    def _flatten_info(self, info, prefix=""):
        """
        Generator yielding (key, value) pairs with flattened hardware info.
        """
        if isinstance(info, dict):
            for key, value in info.items():
                _prefix = "%s.%s" % (prefix, key) if prefix else key
                for result in self._flatten_info(value, prefix=_prefix):
                    yield result
        elif isinstance(info, list):
            for index, value in enumerate(info):
                _prefix = "%s:%s" % (prefix, index) if prefix else index
                for result in self._flatten_info(value, prefix=_prefix):
                    yield result
        else:
            yield (prefix, info)

    def download_csv(self):
        """
        Return hardware information in CSV format for the computer selection.
        """
        set_csv_download(self.request.response, "hardware")
        computers_hardware = self.management.get_hardware_for_computer_ids(
            self.context.selection.get_selected_computer_ids()
        )

        column_headers = ("Computer Id", "Hardware Key", "Value")
        csv_data = [column_headers]

        for computer_id, hardware_info in computers_hardware.items():
            csv_data.extend(
                (computer_id, key, value)
                for key, value in self._flatten_info(self._get_hardware(hardware_info))
                if value != NOT_AVAILABLE
            )

        output = StringIO()
        writer = csv.writer(output)
        writer.writerows(csv_data)
        return output.getvalue()

    def _get_hardware(self, hardware_info):
        """
        Collect and return all parsed hardware information. If colletion of
        a class of hardware fails, log the error and move on and fill in the
        other class data.
        """
        parsed_hardware = {}
        for hardware_class in self._hardware_classes:
            try:
                get_hardware = getattr(self, "_get_%s_info" % hardware_class)
                parsed_hardware[hardware_class] = get_hardware(hardware_info)
            except Exception as error:
                # Catch any exceptions which could range from invalid index
                # errors to divsion by zero
                logging.error(
                    "Hardware Data: Failure parsing %s data on computer %d: %s"
                    " (line:%d)."
                    % (
                        hardware_class,
                        self.computer.id,
                        str(error),
                        sys.exc_info()[-1].tb_next.tb_lineno,
                    )
                )
                parsed_hardware[hardware_class] = []
        return parsed_hardware

    def _get_system_info(self, hardware_info):
        """Return system information."""
        vendor = get_value(hardware_info, "core.vendor")
        serial = get_value(hardware_info, "serial")
        model = get_value(hardware_info, "product")
        chassis = get_value(hardware_info, "configuration.chassis")

        # BIOS info is pulled from firmware
        bios_info = group_by_key(hardware_info, "firmware", start=True)
        bios_vendor = get_value(bios_info, "firmware.vendor")
        bios_version = get_value(bios_info, "firmware.version")
        bios_date = get_value(bios_info, "firmware.date")

        bios_capabilities = []
        capabilities = group_by_key(bios_info, "capabilities")
        for key, str_value, int_value in capabilities:
            flag = str_value.split(" ")
            if len(flag) > 1:
                # we have flag code and description
                bios_capabilities.append({"code": flag[0], "title": " ".join(flag[1:])})
            else:
                # we don't have description. Use code as title.
                bios_capabilities.append({"code": flag[0], "title": flag[0]})

        system_info = {
            "vendor": vendor,
            "serial": serial,
            "model": model,
            "chassis": chassis,
            "bios_vendor": bios_vendor,
            "bios_version": bios_version,
            "bios_date": bios_date,
            "bios_capabilities": bios_capabilities,
        }
        return system_info

    def _get_cpu_info(self, hardware_info):
        """Return CPU information."""
        vendor = NOT_AVAILABLE
        model = NOT_AVAILABLE
        flags = []
        architecture = NOT_AVAILABLE
        clock_speed = NOT_AVAILABLE
        cache_info = {}

        cpus = group_by_resource_key(hardware_info, "cpu", ignore=["logicalcpu"])
        for cpu in cpus:
            if vendor == NOT_AVAILABLE:
                vendor = get_value(cpu, "vendor")
            if model == NOT_AVAILABLE:
                model = get_value(cpu, "product")

            if flags == []:
                flag_key = group_by_key(cpu, "capabilities", ignore="cache")
                for key, str_value, int_value in flag_key:
                    flag = str_value.split(" ")
                    if len(flag) > 1:
                        # we have flag code and description
                        flags.append({"code": flag[0], "title": " ".join(flag[1:])})
                    else:
                        # we don't have description. Use code as title.
                        flags.append({"code": flag[0], "title": flag[0]})

            if architecture == NOT_AVAILABLE:
                architecture = get_value(cpu, "width")

            for key in ("capacity", "clock", "size"):
                if clock_speed == NOT_AVAILABLE:
                    clock_speed = group_by_key(cpu, key, ignore="cache")
                    clock_speed = clock_speed[0][1] if clock_speed else NOT_AVAILABLE
                    clock_speed = HardwareFormatter("clockspeed", clock_speed).format()

            if cache_info == {}:
                cache_levels = []
                cache_key = group_by_key(cpu, "cache")
                if not cache_levels:
                    for level_key in ("slot", "description"):
                        # Do the best to find cache level information using
                        # some preferred keys
                        cache_levels = group_by_key(cache_key, level_key)
                        if cache_levels:
                            cache_sizes = group_by_key(cache_key, "size")
                            if cache_sizes:
                                cache_levels = [i[1] for i in cache_levels]
                                cache_sizes = [i[1] for i in cache_sizes]
                                cache_sizes = [
                                    HardwareFormatter("cpucache", x).format()
                                    for x in cache_sizes
                                ]
                                cache_info = dict(list(zip(cache_levels, cache_sizes)))

        cpu_info = {
            "vendor": vendor,
            "model": model,
            "flags": flags,
            "architecture": architecture,
            "clock_speed": clock_speed,
            "cache": cache_info,
        }
        return cpu_info

    @staticmethod
    def _get_memory_info(hardware_info):
        """
        Get memory information.

        Size information can be retrieved from many different keys, depending
        on the stored information, such as:

        memory.size
        memory.bank.size
        memory:0.size
        memory:0.bank:0.size

        The algorithm first looks for a summary record and if not found it sums
        up each bank of RAM to find the total.
        """
        computer_memory = group_by_key(hardware_info, "memory", start=True)
        # We are going to aggregate over all devices, we don't care what
        # number they are and the numbers just get in the way.
        computer_memory = list(map(deindex_hardware, computer_memory))

        memory_groups = group_by_resource_key(hardware_info, "memory")

        sizes = []
        # Try to get the size attribute from "memory" entries.
        for group in memory_groups:
            size = get_value(group, "size", int_value=True, default=None)
            if size is not None:
                sizes.append(size)
            else:
                # There's no "size" attribute, use "memory.bank:X.size"
                # instead.
                bank_groups = group_by_resource_key(group, "bank")
                for group in bank_groups:
                    bank_size = get_value(group, "size", int_value=True, default=None)
                    if bank_size is not None:
                        sizes.append(bank_size)

        total = sum(sizes)
        if total > 0:
            size = HardwareFormatter("memory", total).format() + " RAM"
        else:
            size = NOT_AVAILABLE
        return {"size": size}

    def _get_network_info(self, hardware_info):
        """Return network information"""
        network_info = []
        for group in group_by_resource_key(hardware_info, "network"):
            vendor = get_value(group, "vendor")
            product = get_value(group, "product")
            description = get_value(group, "description")
            mac = get_value(group, "serial", start=False)
            ip = get_value(group, "configuration.ip", start=False)

            data = {
                "vendor": vendor,
                "product": product,
                "description": description,
                "mac": mac,
                "ip": ip,
            }
            network_info.append(data)
        return network_info or NOT_AVAILABLE

    def _get_storage_info(self, hardware_info):
        """Return storage information"""
        storage_info = []

        for group in itertools.chain(
            group_by_resource_key(hardware_info, "disk"),
            group_by_resource_key(hardware_info, r"scsi(:\d+)?.virtio\d*"),
            group_by_resource_key(hardware_info, r"nvme\d*"),
        ):
            vendor = get_value(group, "vendor")
            product = get_value(group, "product")
            description = get_value(group, "description")
            size = get_value(group, "size", default=None) or get_value(
                group, "namespace.size"
            )
            size = HardwareFormatter("disk", size).format()

            volume_group = group_by_resource_key(group, "volume")
            partitions = []
            for volume in volume_group:
                partition_description = get_value(volume, "description")
                partition_size = get_value(volume, "size")
                partition_size = HardwareFormatter("disk", partition_size).format()
                partition_filesystem = get_value(volume, "configuration.filesystem")
                data = {
                    "description": partition_description,
                    "size": partition_size,
                    "filesystem": partition_filesystem,
                }
                partitions.append(data)

            data = {
                "vendor": vendor,
                "product": product,
                "description": description,
                "size": size,
                "partitions": partitions,
            }
            storage_info.append(data)
        return storage_info or NOT_AVAILABLE

    def _get_display_info(self, hardware_info):
        """Return display information"""
        display = group_by_key(hardware_info, "display")

        vendor = get_value(display, "vendor", start=False)
        product = get_value(display, "product", start=False)

        display_info = {"vendor": vendor, "model": product}
        return display_info

    def _get_multimedia_info(self, hardware_info):
        """Return multimedia information."""
        multimedia = group_by_key(hardware_info, "multimedia")
        vendor = get_value(multimedia, "vendor", start=False)
        model = get_value(multimedia, "product", start=False)

        data = {"vendor": vendor, "model": model}
        return data

    def _get_pci_info(self, hardware_info):
        """Return PCI information."""
        pci_info = []
        for group in group_by_resource_key(hardware_info, "pci.pci"):
            attached_devices = []
            bus_info = {
                "vendor": get_value(group, "vendor"),
                "model": get_value(group, "product"),
                "description": get_value(group, "description"),
                "attached_devices": attached_devices,
            }
            device_keys = set()
            for info in group:
                key = info[0]
                split = key.split(".", 1)
                if len(split) == 2:
                    device_key = split[0]
                    # Ingnore bus information sections
                    if device_key not in ("resources", "configuration"):
                        device_keys.add(device_key)

            for device_key in device_keys:
                [device_group] = group_by_resource_key(group, device_key, start=True)
                attached_devices.append(
                    {
                        "vendor": get_value(device_group, "vendor"),
                        "model": get_value(device_group, "product"),
                        "description": get_value(device_group, "description"),
                    }
                )
            pci_info.append(bus_info)

        return pci_info

    def _get_usb_info(self, hardware_info):
        """Return USB info."""
        return self._get_devices_info(hardware_info, "pci.usb")

    def _get_scsi_info(self, hardware_info):
        """Return SCSI disk info."""
        return self._get_devices_info(hardware_info, "scsi.disk")

    def _get_devices_info(self, hardware_info, device_key):
        """Return info for devices with the specified key."""
        usb_info = []
        for device in group_by_resource_key(hardware_info, device_key):
            vendor = get_value(device, "vendor")
            model = get_value(device, "product")
            description = get_value(device, "description")
            if not (
                vendor == NOT_AVAILABLE
                and model == NOT_AVAILABLE
                and description == NOT_AVAILABLE
            ):
                data = {"vendor": vendor, "model": model, "description": description}
                usb_info.append(data)
        return usb_info


def get_value(
    hardware_info, key, default=NOT_AVAILABLE, start=True, ignore=None, int_value=False
):
    """
    Extracts hardware information using the provided key or return a default.

    @param hardware_info: A list containing hardware information.
    @param key: The key to be used to retrieve the information.
    @param default: A default value to be returned if no key is found.
    @param start: If the match must happen at the beginning of the string.
    @param ignore: The keys to ignore.
    @param int_value: Use the int_value or not.
    """
    key = group_by_key(hardware_info, key, start=start, ignore=ignore)
    if not key:
        return default
    key = key[0]  # multiple keys may match, use the first one
    return key[2] if int_value else key[1]


def group_by_key(hardware_info, key, start=False, ignore=None):
    """
    Loops over the hardware_info list and collect keys containing the passed in
    key.

    @param hardware_info: A list containing hardware information in tuples.
    @param key: The key to be used as a filter.
    @param start: Perform or not an exact match at the beginning of the key.
    @param ignore: A list containing keys to be excluded from the final result.
    """
    split_key = lambda key: key.replace(":", ".").split(".")

    def _remove_keys(keys, ignore):
        # It looks tricky but it's not.
        # It takes the key, replaces occurrances of ":" by ".", breaks it by
        # ".", creates a set and check the intersection with the ignore key
        # list and returns it.
        return [key for key in keys if not set(split_key(key[0])).intersection(ignore)]

    if ignore is not None and not isinstance(ignore, list):
        ignore_keys = [ignore]
    else:
        ignore_keys = ignore
    if start:
        # Checks for matches at the beginning of string and then remove the
        # ones that should be ignored.
        keys = [info for info in hardware_info if info[0].startswith(key)]
    else:
        # Checks for matches in the key string and remove the ones to be
        # ignored.
        keys = [info for info in hardware_info if key in info[0]]
    return _remove_keys(keys, ignore_keys) if ignore_keys else keys


def group_by_resource_key(hardware_info, resource_key, ignore=None, start=False):
    """Group hardware info based on the provided resource key.

    Return a list of lists of tuples from C{hardware_info}, grouped by key.

    Each group contains hardware info whose keys are equal up tp the provided
    C{resource_key}, taking into account also indices that follow it.

    For example, if resource_key="bar", info with keys "foo.bar:0.baz" and
    "foo.bar:0.bza" will go in the same, while "foo.bar:0.baz" and
    "foo.bar:1.bza" will go in different groups because of the index.

    Note that keys prefixes up to the C{resource_key} (and index if present)
    are stripped from the result, so "foo.bar:0.baz" -> "baz".

    @param ignore: If specified, a list of key tokens that will be ignored.
    @param start: If true, match the C{resource_key} at the beginning of the
        info keys.
    """
    # Escape dots since the key is used in a regexp
    resource_key = resource_key.replace(".", r"\.")
    prefix_exp = "" if start else r".*\b"
    regexp = re.compile(r"(%s%s(:[0-9]+)?\.)" % (prefix_exp, resource_key))

    if ignore is not None:
        ignore = set(ignore)

    split_key = lambda key: key.replace(":", ".").split(".")

    resource_info = defaultdict(list)
    for info in hardware_info:
        key = info[0]

        if ignore is not None:
            if ignore.intersection(split_key(key)):
                # The key contains at least one element to ignore, skip it
                continue

        match = regexp.match(key)
        if match:
            prefix = match.groups()[0]
            # Strip the common prefix
            stripped_info = (key[len(prefix) :], info[1], info[2])
            resource_info[prefix].append(stripped_info)
    # Sort items based on the original prefix, in case indices were present.
    return [info_list for _, info_list in sorted(resource_info.items())]


HARDWARE_DEINDEXER_REGEX = re.compile("(:\d+)(?=\.)")


def deindex_hardware(items):
    """Remove any device indices from a single record of hardware info."""
    key = items[0]
    rest = items[1:]
    deindexed_key = HARDWARE_DEINDEXER_REGEX.sub("", key)
    return (deindexed_key,) + rest
