# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.security.interfaces import Unauthorized
from zope.security import checkPermission
from zope.component import getUtility
from zope.session.interfaces import IClientIdManager

from canonical.url import get_url

from canonical.landscape.ui.knowledge.page import KnowledgeBaseFilteredSidebarPage
from canonical.landscape.model.knowledge.parser import create_wiki_parser
from canonical.landscape.model.knowledge.article import may_view_unpublished_articles


class ArticlePage(KnowledgeBaseFilteredSidebarPage):
    """Display an article of the knowledge base."""

    title = "Knowledge base article"

    def __init__(self, context, request):
        super(ArticlePage, self).__init__(context, request)
        if not context.published:
            if not may_view_unpublished_articles():
                raise Unauthorized("Can't access this article")
        self.search_terms = self.request.get("search-terms")
        self.vote = self.context.get_vote_by_person(self.request.principal.person)
        self.helpful_votes = self.total_votes = 0
        if checkPermission("landscape.ViewVotesKnowledgeBase", None):
            votes = self.context.get_votes()
            self.helpful_votes = votes[True]
            self.total_votes = votes[True] + votes[False]
        person = self.request.principal.person
        # We only count views from customer, not support people
        if not person.has_global_role("SupportProvider"):
            client_id_manager = getUtility(IClientIdManager)
            client_id = client_id_manager.getClientId(request)
            self.context.add_times_viewed(client_id)

    def may_view_status(self):
        """
        Helper method used to determine whether a user may view the article
        status.
        """
        return may_view_unpublished_articles()

    def get_rendered_body(self):
        """Render the wiki markup body data as XHTML."""
        parser = create_wiki_parser(format="xhtml")
        return parser(self.context.body).decode("utf-8")

    def remove_article(self):
        self.notes.info("Removed article '%s'." % self.context.title)
        self.context.remove()
        self.request.response.redirect("/kb")

    def vote_article(self):
        if "helpful" in self.request.form:
            self.context.vote(self.request.principal.person, True)
        elif "unhelpful" in self.request.form:
            self.context.vote(self.request.principal.person, False)
        self.notes.info("Thank you for your vote.")
        self.request.response.redirect(get_url(self.context, self.request))

    def reset_votes(self):
        self.context.reset_votes()
        self.notes.info("Votes reset.")
        self.request.response.redirect(get_url(self.context, self.request))
