# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import mimetypes
from datetime import datetime

from zope.browserpage import ViewPageTemplateFile

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.ui.knowledge.page import KnowledgeBaseSidebarPage
from canonical.landscape.ui.form import (
    LandscapeFormProvider,
    LongTextWidget,
    MultiCheckBoxWidget,
)
from canonical.landscape.ui.script.attachment import is_image_type
from canonical.landscape.model.knowledge.interfaces import IArticle
from canonical.landscape.model.knowledge.category import Category


class EditArticlePage(KnowledgeBaseSidebarPage):

    title = "Edit article"


class EditArticleForm(LandscapeFormProvider):

    summary = "Edit article"

    fields = Fields(IArticle, render_context=True)

    fields["title"].custom_widget = LongTextWidget
    fields["categories"].custom_widget = MultiCheckBoxWidget

    additional_content = ViewPageTemplateFile("edit-upload-fields.pt")

    attachments_error = ""

    def __init__(self, context, request, view):
        super(EditArticleForm, self).__init__(context, request, view)
        self.original_category_ids = set(self.context.categories.values(Category.id))

    def format_size(self, size):
        return self.request.principal.person.formatter.format_bytes(size)

    def get_initial_data(self):
        data = {
            "title": self.context.title,
            "body": self.context.body,
            "published": self.context.published,
            "categories": [category.id for category in self.context.categories],
        }
        return data

    def is_image(self, filename):
        mimetype, encoding = mimetypes.guess_type(filename)
        return is_image_type(mimetype)

    @action("Save", primary=True)
    def save(self, action, data):
        # First remove selected attachments
        to_remove = []
        for attachment in self.context.attachments:
            input_name = "form.remove_attachments.%d" % (attachment.id,)
            if input_name in self.request.form:
                to_remove.append(attachment.id)
        if to_remove:
            self.context.remove_attachments_by_ids(to_remove)
            if len(to_remove) == 1:
                self.view.notes.info("Attachment removed.")
            else:
                self.view.notes.info("Attachments removed.")

        existing_filenames = [
            attachment.filename for attachment in self.context.attachments
        ]

        attachment = self.request.form.get("form.new_attachment")
        if attachment:
            # Check for unique filenames
            if attachment.filename in existing_filenames:
                self.attachments_error = "Attachments must have unique file names."
                return
            attachment_data = attachment.read()
            self.context.add_attachment(attachment.filename, attachment_data)
            self.view.notes.info("Attachment added.")

        categories = data.pop("categories", ())
        categories = set(categories)
        if categories != self.original_category_ids:
            self.context.remove_category_by_ids(self.original_category_ids - categories)
            self.context.add_category_by_ids(categories - self.original_category_ids)
            self.view.notes.info("Categories of the article changed.")

        if self.apply_changes(self.context, data):
            self.view.notes.info("Article changed.")
            self.context.modification_time = datetime.utcnow()
        elif (
            categories == self.original_category_ids
            and not attachment
            and not to_remove
        ):
            self.view.notes.info("No changes made.")
        else:
            self.context.modification_time = datetime.utcnow()

        if not attachment:
            self.request.response.redirect(get_url(self.context, self.request))
