# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import date
import logging

from zope.browserpage import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility

from canonical.content.form import Fields, action
from canonical.security.interfaces import IRoleRegistry
from canonical.url import get_url


from canonical.landscape.constants import UBUNTU_ADVANTAGE_PURCHASE_URL
from canonical.landscape.model.computer.search import (
    LicenseIdCriteria,
    NoLicenseCriteria,
)
from canonical.landscape.model.main.interfaces import ILicense
from canonical.landscape.model.main.standalone import AccountLDSLicenseAPI
from canonical.landscape.model.main.license import get_license_name, LDS_ROLES
from canonical.landscape.model.license.management import LicenseManagement

from canonical.landscape.security.roles import is_support
from canonical.landscape.service.license.edit import EditLicense

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider


# Licenses which expire on this date will be labelled as never expiring.
# This date matches the one for the 'free_lds_license' in:
#
#   tree/opt/canonical/landscape/free-license.txt
#
# and they must be kept in sync.
NEVER_EXPIRY_DATE = date(3000, 1, 1)


class LicensePageMixin(object):
    """
    Mixin class with utility methods for license-related pages.
    """

    def role_name(self, role_key):
        """Return a display name for the given role key."""
        role = getUtility(IRoleRegistry).get(role_key)
        return get_license_name(role.name)["display"]

    def format_date(self, expiration_date):
        """Return a formatted date string representing the expiration_date.

        Return "Never" if the expiration date does not exist or is far in the
        future.
        """
        if expiration_date in (None, NEVER_EXPIRY_DATE):
            return "Never"

        return self.request.principal.person.formatter.format_date(
            expiration_date, with_year=True
        )


class LicenseOverviewPage(AccountPage, LicensePageMixin):

    title = "Licenses"
    add_license_url = UBUNTU_ADVANTAGE_PURCHASE_URL

    @Lazy
    def licenses(self):
        """Return the list of licenses in the account."""
        return list(self.context.licenses)

    def licensed_computers_url(self, license_id):
        """
        Return the computer search url for computers matching this license.
        """
        criteria = LicenseIdCriteria(license_id, self.context)
        license_search = "/computers/?criteria={}".format(criteria.to_string())
        return get_url(self.context, self.request) + license_search

    @Lazy
    def unlicensed_computers_url(self):
        """
        Return the computer search url for computers matching this license.
        """
        criteria = NoLicenseCriteria(self.context)
        license_search = "/computers/?criteria={}".format(criteria.to_string())
        return get_url(self.context, self.request) + license_search

    @Lazy
    def is_support(self):
        """Return True if the current user has support privileges."""
        return is_support(self.request)


class LicensePage(AccountPage):

    title = "License"

    @Lazy
    def is_lds(self):
        """Return C{True} if the license is for Landscape On Prem."""
        role_registry = getUtility(IRoleRegistry)
        lds_role_keys = [role_registry.get(role).key for role in LDS_ROLES]
        return self.context.role_key in lds_role_keys


class EditLicenseForm(LandscapeFormProvider):

    summary = "Edit license"

    def __init__(self, context, request, view):
        super(EditLicenseForm, self).__init__(context, request, view)
        self.management = LicenseManagement(self.context.account)

    @Lazy
    def fields(self):
        fields = Fields(ILicense, render_context=True)
        fields["expiration_date"].field.required = self.view.is_lds
        return fields

    @action("Save", primary=True)
    def save(self, action, data):
        old_seats = self.context.seats
        old_expiration_date = self.context.expiration_date
        new_seats = data["seats"]
        new_expiration_date = data["expiration_date"]

        operation = EditLicense(self.context, self.request.principal)
        changed = operation(expiration_date=new_expiration_date, seats=new_seats)
        if changed:
            self.view.notes.info("The license has been updated.")
            logging.info(
                "License %d in Account %s updated by Person %d: "
                "seats %d to %d expiry %s to %s"
                % (
                    self.context.id,
                    self.context.account.name,
                    self.request.principal.person.id,
                    old_seats,
                    new_seats,
                    old_expiration_date.isoformat() if old_expiration_date else "None",
                    new_expiration_date.isoformat() if new_expiration_date else "None",
                )
            )

        if self.view.is_lds:
            license_api = AccountLDSLicenseAPI(self.context.account)
            lds_license = license_api.get_license()
            lds_license.regenerate_contents()

        self.request.response.redirect(
            get_url(self.context.account, self.request) + "/licenses"
        )


class RemoveLicensePage(LicensePage):

    title = "Remove license"


class RemoveLicenseForm(LandscapeFormProvider, LicensePageMixin):

    summary = "Remove license"

    additional_content = ViewPageTemplateFile("remove-license.pt")

    def __init__(self, context, request, view):
        super(RemoveLicenseForm, self).__init__(context, request, view)
        self.management = LicenseManagement(self.context.account)
        self.redirect_url = get_url(self.context.account, self.request) + "/licenses"

    @action("Confirm")
    def remove(self, action, data):
        computers = [computer for computer in self.context.computers]
        self.context.remove()

        # Auto-update computers
        untransferred_computers = 0
        transferred_computers = 0
        for computer in computers:
            # auto_assign... adds computer to a new appropriate license type
            found_license = self.management.auto_assign_license_to_computer(computer)
            if found_license:
                transferred_computers += 1
            else:
                untransferred_computers += 1

        if self.view.is_lds:
            license_api = AccountLDSLicenseAPI(self.context.account)
            lds_license = license_api.get_license()
            lds_license.regenerate_contents()

        if untransferred_computers > 0:
            remain_text = "computer remains"
            if untransferred_computers > 1:
                remain_text = "computers remain"
            self.view.notes.warning(
                "The selected license has been removed. %d %s "
                "without an active license." % (untransferred_computers, remain_text)
            )
        else:
            additional_text = ""
            if transferred_computers > 0:
                additional_text = " License transferred for %d computer%s." % (
                    transferred_computers,
                    "s" if transferred_computers > 1 else "",
                )

            self.view.notes.info(
                "The selected license has been removed.%s" % additional_text
            )
        self.request.response.redirect(self.redirect_url)

    @action("Cancel", primary=True)
    def cancel(self, action, data):
        self.request.response.redirect(self.redirect_url)
