# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
OpenID integration for Landscape.
"""
import urllib.request
import urllib.parse
import urllib.error

from zope.interface import Interface
from zope.session.interfaces import ISession

from openid.consumer.consumer import Consumer as DefaultConsumer, SUCCESS
from openid.consumer.discover import DiscoveryFailure
from openid.extensions import pape
from openid.extensions.sreg import SRegRequest, SRegResponse
from openid.extensions.ax import AttrInfo, FetchRequest, FetchResponse

from canonical.landscape.model.configuration import get_root_url

from canonical.landscape.model.openid.store import get_openid_store


def set_consumer_factory(factory):
    LandscapeOpenID.default_consumer_factory = factory


def reset_consumer_factory():
    LandscapeOpenID.default_consumer_factory = DefaultConsumer


class SSODownError(Exception):
    """The SSO server can't be contacted."""


class OpenIDAuth(object):
    """Hold information about a successful OpenID authentication."""

    AX_EXTRA_EMAILS_URI = "http://ns.login.ubuntu.com/2022/emails"

    def __init__(
        self, identity_url, name, email, timezone, nickname="", extra_emails=[]
    ):
        self.identity_url = identity_url
        self.name = name
        self.email = email
        self.timezone = timezone
        self.nickname = nickname  # Launchpad username
        self.extra_emails = extra_emails


class LandscapeOpenID(object):
    """
    An object that knows about Landscape's policy for doing OpenID
    authentication.  Most importantly, the URL of Landscape and the OpenID
    provider: Ubuntu.
    """

    default_consumer_factory = DefaultConsumer

    def __init__(self, request, provider_url, consumer_factory=None):
        if consumer_factory is None:
            consumer_factory = self.default_consumer_factory
        self.openid_provider_url = provider_url
        store = get_openid_store()
        session_data = ISession(request)["landscape.openid"]
        self.request = request
        self.consumer = consumer_factory(session_data, store)

    def begin(self, return_to_suffix):
        """
        Redirect the request to the OpenID provider so that he can
        authenticate.
        """
        try:
            auth = self.consumer.begin(self.openid_provider_url)
        except DiscoveryFailure:
            raise SSODownError()
        pape_request = pape.Request(max_auth_age=30 * 60)
        auth.addExtension(pape_request)
        sreg_request = SRegRequest(
            required=["email", "fullname"], optional=["timezone", "nickname"]
        )
        auth.addExtension(sreg_request)

        ax_request = FetchRequest()  # For additional verified emails
        ax_request.add(AttrInfo(OpenIDAuth.AX_EXTRA_EMAILS_URI, required=True))
        auth.addExtension(ax_request)

        root_url = get_root_url()
        if not root_url:
            root_url = self.request.getApplicationURL() + "/"
        url = auth.redirectURL(root_url, root_url + return_to_suffix)
        self.request.response.redirect(url, trusted=True)

    def complete(self):
        """
        Given a request that contains a response from an OpenID provider (via
        the user agent), check if OpenID authentication succeeded.  If it has,
        an authentication object will be returned.
        """
        url = self.request.getURL()
        if self.request.get("QUERY_STRING"):
            url += "?%s" % self.request.get("QUERY_STRING")
        status = self.consumer.complete(self.request.form, url)
        if status.status == SUCCESS:
            identity_url = str(status.identity_url)
            sreg_response = SRegResponse.fromSuccessResponse(status)
            if sreg_response is not None:
                email_address = str(sreg_response.get("email"))
                name = str(sreg_response.get("fullname"))
                timezone = str(sreg_response.get("timezone", ""))
                nickname = str(sreg_response.get("nickname", ""))
            else:
                email_address = name = timezone = nickname = ""

            ax_response = FetchResponse.fromSuccessResponse(status)
            extra_emails = []
            if ax_response is not None:
                emails_field = ax_response.getSingle(OpenIDAuth.AX_EXTRA_EMAILS_URI)
                if emails_field:
                    extra_emails = str(emails_field).split(",")
                    extra_emails = [e.strip() for e in extra_emails]
                    extra_emails = [e for e in extra_emails if e]  # Remove ''

            return OpenIDAuth(
                identity_url, name, email_address, timezone, nickname, extra_emails
            )
        else:
            return None


class ILandscapeOpenIDFactory(Interface):
    """A factory for Landscape/OpenID policies."""

    def get_logout_url(return_to):
        """The URL to redirect to when logging out."""

    def get_landscape_openid(request, consumer_factory=None):
        """Return an object like L{LandscapeOpenID}."""


class LandscapeOpenIDFactory(object):
    """A Landscape/OpenID policy which uses the real Ubuntu OpenID provider."""

    def __init__(self, provider_url, logout_url):
        self.logout_url = logout_url
        self.provider_url = provider_url

    def get_logout_url(self, return_to):
        return "{}?{}".format(
            self.logout_url, urllib.parse.urlencode({"return_to": return_to})
        )

    def get_landscape_openid(self, request, consumer_factory=None):
        return LandscapeOpenID(request, self.provider_url, consumer_factory)
