# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.security import checkPermission

from canonical.url import get_url, get_url_object

from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.model.feature.flag import is_feature_enabled
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.ui.navigation import (
    LandscapeNavigation,
    LandscapeNavigationItem,
)


MAY_EDIT_ACCOUNT = [
    "landscape.EditAccount",
    "landscape.EditMaxPeopleCount",
]


def _check_permissions(context, permissions):
    """Return True if the context has the required permission(s).

    "permissions" may be a string (single permission), in which case
    a simple check is applied.  Otherwise each string in "permissions"
    is checked and the whole set passes if any of them pass (i.e. an
    OR operation).
    """
    if isinstance(permissions, str):
        return checkPermission(permissions, context)

    return any(checkPermission(perm, context) for perm in permissions)


class AccountNavigationItem(LandscapeNavigationItem):

    PERMISSIONS = None

    def _get_permission_context(self):
        try:
            return self.account
        except AttributeError:
            return self.parent.account

    # used by NavigationItem.is_available()
    def is_permitted(self):
        if self.PERMISSIONS:
            context = self._get_permission_context()
            if not _check_permissions(context, self.PERMISSIONS):
                return False
        return super(AccountNavigationItem, self).is_permitted()


class SettingsNavigation(LandscapeNavigation):
    """A parent navigation for account-level settings."""

    name = "account-settings"
    title = "Settings"

    def __init__(self, context, request, parent=None):
        super(SettingsNavigation, self).__init__(context, request)
        self.account = get_url_object(self.context, request, IAccount)


class AdministratorsItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewPerson"

    title = "Administrators"

    @property
    def description(self):
        return "Manage administrative access to %s" % self.parent.account.title

    @property
    def action(self):
        return "%s/administrators" % get_url(self.parent.account, self.request)


class ScriptsItem(AccountNavigationItem):

    title = "Scripts"
    description = "Manage stored scripts"

    @property
    def action(self):
        return "%s/scripts" % get_url(self.parent.account, self.request)

    def is_selected(self):
        selected = super(ScriptsItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return self.parent.account.name + "/script/" in url


class CustomGraphsItem(AccountNavigationItem):

    title = "Graphs"
    description = "Manage custom graphs"

    @property
    def action(self):
        return "%s/custom-graphs" % get_url(self.parent.account, self.request)

    def is_selected(self):
        selected = super(CustomGraphsItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return self.parent.account.name + "/custom-graph/" in url


class ManageAccessContextsItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewAccessContextManagement"

    title = "Access groups"
    description = "Manage access groups"

    @property
    def action(self):
        return "%s/access-groups" % get_url(self.parent.account, self.request)

    def is_selected(self):
        selected = super(ManageAccessContextsItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return self.parent.account.name + "/access-group/" in url


class RolesItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewRole"

    title = "Roles"
    description = "Manage roles"

    @property
    def action(self):
        return "%s/roles" % get_url(self.parent.account, self.request)

    def is_selected(self):
        selected = super(RolesItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return self.parent.account.name + "/roles/" in url


class AlertsItem(AccountNavigationItem):

    PERMISSIONS = "landscape.QueryAlertRule"

    title = "Alerts"
    description = "Manage alerts"

    @property
    def action(self):
        return "%s/alerts" % get_url(self.parent.account, self.request)

    def is_selected(self):
        selected = super(AlertsItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return self.parent.account.name + "/alert/" in url


class ActivitiesItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewActivity"

    title = "Activities"

    @property
    def description(self):
        return "See activities for computers in %s" % self.parent.account.title

    @property
    def action(self):
        return "%s/activities" % get_url(self.parent.account, self.request)

    def _can_access_computers(self):
        # computer_ids is [] if the person doesn't have access to any
        # computers in the account.
        person = self.request.principal.person
        computer_ids = person.get_computer_ids(self.parent.account)
        return computer_ids is None or len(computer_ids) > 0

    def is_available(self):
        return (
            self._can_access_computers() or super(ActivitiesItem, self).is_available()
        )

    def is_selected(self):
        selected = super(ActivitiesItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return self.parent.account.name + "/activity/" in url


class ConsolidatedProfilesItem(AccountNavigationItem):

    title = "Profiles"
    description = "Manage computer profiles"

    @property
    def action(self):
        return "%s/profiles" % get_url(self.parent.account, self.request)

    def is_selected(self):
        selected = super(ConsolidatedProfilesItem, self).is_selected()
        if selected:
            return selected
        url = self.request.getURL()
        return "profile" in url


class SavedSearchesItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewSavedSearch"

    title = "Searches"
    description = "Manage saved searches"

    @property
    def action(self):
        return "%s/saved-searches" % get_url(self.parent.account, self.request)


class EventLogItem(AccountNavigationItem):

    title = "Events"
    description = "All event log entries from the last 90 days"

    @property
    def action(self):
        return "%s/event-log" % get_url(self.parent.account, self.request)


class AccountItem(AccountNavigationItem):

    PERMISSIONS = MAY_EDIT_ACCOUNT

    title = "Account"
    description = "Edit account details"

    @property
    def action(self):
        return "%s/edit" % get_url(self.parent.account, self.request)


class LDSSettingsItem(AccountNavigationItem):

    PERMISSIONS = MAY_EDIT_ACCOUNT

    title = "Settings"
    description = "Settings"

    @property
    def action(self):
        return "%s/standalone-settings" % get_url(self.parent.account, self.request)

    def is_available(self):
        return is_standalone() and super(LDSSettingsItem, self).is_available()


class BillingItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewBilling"

    title = "Billing"
    description = "View Landscape billing information"

    @property
    def action(self):
        return "%s/billing" % get_url(self.parent.account, self.request)

    def is_available(self):
        return (
            is_feature_enabled("self-service")
            and self.parent.account.is_self_service
            and super(BillingItem, self).is_available()
        )


class LicensesItem(AccountNavigationItem):

    title = "Licenses"
    description = "View Landscape licenses"

    @property
    def action(self):
        return "%s/licenses" % get_url(self.parent.account, self.request)

    def is_available(self):
        return (
            not self.parent.account.is_self_service
            and super(LicensesItem, self).is_available()
        )
