# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.url import get_url
from canonical.routes.publisher import get_route_object_by_interface

from canonical.landscape.security.interfaces import IPersonPrincipal
from canonical.landscape.model.selection.interfaces import IComputerSelection
from canonical.landscape.ui.account.routes import get_current_account
from canonical.landscape.ui.menu.account import AccountNavigationItem
from canonical.landscape.ui.navigation import LandscapeNavigation


class AccountsNavigation(LandscapeNavigation):
    """
    The accounts navigation is organized in a structure similar
    to this one::

      AccountsNavigation
      |
      |\_ AccountItem
      |   |
      |   |\_ AccountItem (zero or more)
      |
      |\_ SystemsManagementIcon
    """

    name = "accounts"
    split = True

    def get_items(self):
        """Build items dynamically depending on the person's accounts."""
        items = []

        dashboard_item = DashboardItem(self.context, self.request, self)

        current_account = get_current_account(self.request)
        if IPersonPrincipal.providedBy(self.request.principal):
            accounts = list(self.request.principal.person.accounts)
        else:
            accounts = []

        if current_account is None:
            # We're not inside an account, just show the dashboard tab
            # and the global support one
            items = [
                dashboard_item,
                GlobalSupportItem(self.context, self.request, self),
            ]
        else:
            # We're inside an account, show all account related tabs
            current_account_item = AccountItem(
                self.context, self.request, self, current_account
            )

            items = [
                current_account_item,
                SystemsManagementItem(self.context, self.request, current_account_item),
                SupportItem(self.context, self.request, current_account_item),
            ]

            # It can be possible for persons not belonging to an account to
            # visit the pages of a certain account (e.g. support providers)
            # so we need perform the check below befor trying to remove the
            # current account from the accounts list.
            if current_account in accounts:
                accounts.remove(current_account)

            if accounts:
                # The dashboard item is shown only if the person has more
                # than one account
                current_account_item.items.append(dashboard_item)

        # Add available accounts to the first item's sub-menu
        first_item = items[0]
        for account in accounts:
            first_item.items.append(
                AccountItem(self.context, self.request, first_item, account)
            )

        return items


class DashboardItem(AccountNavigationItem):

    PERMISSIONS = "landscape.ViewPersonResource"

    name = "account-dashboard"
    title = "Dashboard"
    action = "/dashboard"
    order = 10

    def _get_permission_context(self):
        return self.context


class AccountItem(AccountNavigationItem):

    order = 12

    def __init__(self, context, request, parent, account):
        super(AccountItem, self).__init__(context, request, parent)
        self.account = account

    @property
    def name(self):
        return "account-%s-item" % self.account.name

    @property
    def title(self):
        return self.account.title

    @property
    def action(self):
        return get_url(self.account, self.request)

    def is_selected(self):
        """
        L{AccountItem} is selected when current url is below the account.
        """
        if self.parent.parent:
            # We're a leaf item of the account sub-menu
            return False

        # The account tab is selected only if no other tab is selected
        for item in self.parent.get_items()[1:]:
            if item.is_selected():
                return False
        return True


class SystemsManagementItem(AccountNavigationItem):
    """A navigation item for systems management."""

    PERMISSIONS = "landscape.ViewComputerSelection"

    name = "account-systems-management"
    title = "Computers"
    order = 13
    account = None

    def __init__(self, context, request, parent=None):
        super(SystemsManagementItem, self).__init__(context, request, parent)
        if parent is not None:
            self.account = parent.account

    @property
    def description(self):
        return "Manage computer systems for %s" % self.account.title

    @property
    def action(self):
        return "%s/computers" % get_url(self.account, self.request)

    def is_selected(self):
        selection = get_route_object_by_interface(IComputerSelection, self.request)
        if selection is not None:
            return selection.account == self.account
        return False


class SupportItem(AccountNavigationItem):
    """A navigation item for support."""

    PERMISSIONS = "landscape.ViewKnowledgeBase"

    name = "account-support"
    title = "Support"
    order = 17
    account = None
    action = "http://support.canonical.com"

    def __init__(self, context, request, parent=None):
        super(SupportItem, self).__init__(context, request, parent)
        if parent is not None:
            self.account = parent.account

    @property
    def description(self):
        return "Get support for %s" % self.account.title

    def is_selected(self):
        url = self.request.getURL(path_only=True)
        kb_action = "%s/kb" % get_url(self.account, self.request)
        return url.startswith(kb_action)

    def is_available(self):
        return (
            not self.account.is_self_service and super(SupportItem, self).is_available()
        )


class GlobalSupportItem(AccountNavigationItem):
    """A navigation item for support."""

    PERMISSIONS = "landscape.EditKnowledgeBase"

    name = "support"
    action = "/kb"
    title = "Support"
    order = 17
    account = None

    @property
    def description(self):
        return "Access support"
