# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.navigation import Navigation, NavigationItem
from canonical.url import get_url, get_url_object

from canonical.landscape.model.selection.interfaces import IComputerSelection


class SystemsManagementNavigation(Navigation):
    """Second-level navigation menu for the systems management section."""

    name = "systems-management"


class SelectionNavigationItem(NavigationItem):
    """A navigation item relative to the current computer selection."""

    path_name = None
    alternative_path_name = None

    @property
    def action(self):
        """Return an action URL relative to the current computer selection."""
        if self.path_name is None:
            raise ValueError("The 'path_name' attribute must be set.")
        return self._get_action_url(self.path_name)

    def _get_action_url(self, path_name):
        """
        Given a C{path_name}, compute an action url relative to the
        computer selection.
        """
        selection = get_url_object(self.context, self.request, IComputerSelection)
        return get_url(selection, self.request) + "/" + path_name

    def is_selected(self):
        """
        In addition to any checks the base class does, also compare to
        C{alternative_path_name} if present.
        """
        selected = super(SelectionNavigationItem, self).is_selected()
        if selected:
            return True
        if self.alternative_path_name is None:
            return False
        alternative_action = self._get_action_url(self.alternative_path_name)
        return self.request.getURL(path_only=True).startswith(alternative_action)


class SystemsInfo(SelectionNavigationItem):
    """Overview of currently selected computers."""

    path_name = "info"
    title = "Info"
    description = "See an overview for the selected computers"
    order = 10


class SystemsActivities(SelectionNavigationItem):
    """Activities for the selected computers."""

    path_name = "activities"
    alternative_path_name = "activity"
    title = "Activities"
    description = "See activities for the selected computers"
    order = 20


class SystemsHardware(SelectionNavigationItem):
    """Hardware overview for the selected computers."""

    path_name = "hardware"
    title = "Hardware"
    description = "See a hardware overview for the selected computers"
    order = 30


class SystemsUbuntuProInfo(SelectionNavigationItem):
    """Ubuntu Pro Info overview for the selected computers."""

    path_name = "ubuntu_pro_info"
    title = "Ubuntu Pro"
    description = "See the Ubuntu Pro information for the selected computers"
    order = 35


class SystemsMonitoring(SelectionNavigationItem):
    """Monitoring for the selected computers."""

    path_name = "monitoring"
    title = "Monitoring"
    description = "See monitored activity for the selected computers"
    order = 40


class SystemsScripts(SelectionNavigationItem):
    """Run scripts on selected computers."""

    path_name = "scripts"
    title = "Scripts"
    description = "Run scripts on selected computers"
    order = 50


class SystemsProcesses(SelectionNavigationItem):
    """View active processes on selected computers."""

    path_name = "processes"
    title = "Processes"
    description = "View active processes"
    order = 60


class SystemsPackages(SelectionNavigationItem):
    """Manage packages on the selected computers."""

    path_name = "packages"
    title = "Packages"
    description = "Manage software packages on the selected computers"
    order = 70


class SystemsSnaps(SelectionNavigationItem):
    """Manage snaps on the selected computers."""

    path_name = "snaps"
    title = "Snaps"
    description = "Manage snaps on the selected computers"
    order = 75


class SystemsUsers(SelectionNavigationItem):
    """Manage users on the selected computers."""

    path_name = "users"
    title = "Users"
    description = "Manage users on the selected computers"
    order = 80


class SystemsReporting(SelectionNavigationItem):
    """Display reporting information on the selected computers."""

    path_name = "reports"
    title = "Reports"
    description = "Display reporting information on the selected computers"
    order = 90
