# Copyright 2005-2022 Canonical Limited. All rights reserved.

import json

from decimal import Decimal
from urllib.parse import urlencode

from zope.cachedescriptors.property import Lazy

from zope.publisher.interfaces import NotFound
from zope.publisher.browser import BrowserPage
from zope.security.proxy import removeSecurityProxy

from canonical.url import get_url

from canonical.landscape.ui.toolbar.page import ToolbarPage, ComputerPaginatorPage
from canonical.content.download import set_csv_download
from canonical.landscape.model.account.customgraph import CustomGraphManagement
from canonical.landscape.model.resource.mountinfo import AccountMountInfos
from canonical.landscape.model.resource.temperature import AccountThermalZones
from canonical.landscape.model.resource.network import (
    AccountNetworkInfo,
    sort_interfaces,
)


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)


class ComputerMonitoringPage(ComputerPaginatorPage, ToolbarPage):
    """
    Base class for all monitoring pages provides filter and is the
    base for the monitoring-content-area provider.
    """

    max_per_page = 5

    def __init__(self, context, request):
        super(ComputerMonitoringPage, self).__init__(context, request)
        self.account = context.selection.account
        self.days = self._get_days()
        self.timezone = self.request.principal.person.timezone

    def _get_days(self):
        days = self.request.get("days", "1")
        try:
            days = int(days)
        except ValueError:
            return "1"
        return str(days)

    def query_string(self, extra_params=None):
        """
        Return a query string for the page, with optional additional
        parameters.
        """
        pieces = [("days", self.days)]
        if extra_params:
            pieces.extend(iter(list(extra_params.items())))

        for param in ("start", "index"):
            value = self.request.get(param)
            if value:
                pieces.append((param, value))

        pieces = [(str(key), str(value_)) for key, value_ in pieces]
        return "?" + urlencode(sorted(pieces))

    def graph_image_url(self, name, extra_params=None):
        """
        Return the link for a graph image, based on the graph name.

        The link points to a graph which displays data for computers in the
        current page.
        """
        return "%s/computer/%s/monitoring/graphs/%s%s" % (
            get_url(self.account, self.request),
            "+".join(str(id) for id in self.computer_ids),
            name,
            self.query_string(extra_params),
        )

    def graph_csv_url(self, name, extra_params=None):
        """
        Return the link for graph CSV data, based on the graph name.

        The linked CSV document contains data for all computers in the
        selection.
        """
        return "%s/graphs/%s_as_csv%s" % (
            get_url(self.context, self.request),
            name,
            self.query_string(extra_params),
        )

    def graph_detail_page_url(self, name, extra_params=None):
        """
        Return the link for graph detail page, based on the graph name.
        """
        return "%s/%s%s" % (
            get_url(self.context, self.request),
            name,
            self.query_string(extra_params),
        )


class CustomGraphsMixin(object):
    @Lazy
    def custom_graphs(self):
        """
        Return a list of custom graphs for computers displayed in the page.
        """
        management = CustomGraphManagement(self.account, self.request.principal.person)
        return list(management.get_graphs(computer_ids=self.computer_ids))


class ComputerMonitoringOverviewPage(ComputerMonitoringPage, CustomGraphsMixin):

    title = "Computer monitoring"

    context_url = ""

    # How many interfaces should we provide graphs for? If the page
    # has more NICs than this links to per-interface graph pages will
    # be provided instead.
    MAXIMUM_INTERFACE_GRAPHS = 3

    def __init__(self, context, request):
        super(ComputerMonitoringOverviewPage, self).__init__(context, request)
        account_infos = AccountMountInfos(self.account)
        self.mount_points = sorted(account_infos.get_mount_points(self.computer_ids))
        account_zones = AccountThermalZones(self.account)
        self.thermal_zones = sorted(account_zones.get_thermal_zones(self.computer_ids))
        account_infos = AccountNetworkInfo(self.account)
        self.network_interfaces = sort_interfaces(
            account_infos.get_network_interfaces(self.computer_ids)
        )
        if "lo" in self.network_interfaces:
            self.network_interfaces.remove("lo")
        self.should_graph_interfaces = (
            len(self.network_interfaces) <= self.MAXIMUM_INTERFACE_GRAPHS
        )


class ComputerDetailMonitoringPage(ComputerMonitoringPage):
    """Base class for per-resource detailed graph pages."""

    def query_string(self, extra_params=None):
        params = {"large": 1}
        if extra_params:
            params.update(extra_params)
        return super(ComputerDetailMonitoringPage, self).query_string(params)


class ComputerDiskMonitoringPage(ComputerDetailMonitoringPage):

    title = "Computer disk monitoring"

    context_url = "/disk"

    def __init__(self, context, request):
        super(ComputerDiskMonitoringPage, self).__init__(context, request)
        account_infos = AccountMountInfos(self.account)
        self.mount_points = sorted(account_infos.get_mount_points(self.computer_ids))


class ComputerLoadMonitoringPage(ComputerDetailMonitoringPage):

    title = "Computer load monitoring"

    context_url = "/load"


class ComputerMemoryMonitoringPage(ComputerDetailMonitoringPage):

    title = "Computer memory monitoring"

    context_url = "/memory"


class ComputerTemperatureMonitoringPage(ComputerDetailMonitoringPage):

    title = "Computer temperature monitoring"

    context_url = "/temperature"

    def __init__(self, context, request):
        super(ComputerTemperatureMonitoringPage, self).__init__(context, request)
        account_zones = AccountThermalZones(self.account)
        self.thermal_zones = sorted(account_zones.get_thermal_zones(self.computer_ids))


class CustomGraphsPage(ComputerDetailMonitoringPage, CustomGraphsMixin):

    title = "Custom graphs"

    context_url = "/custom"


class ComputerNetworkMonitoringPage(ComputerDetailMonitoringPage):

    title = "Computer network monitoring"

    context_url = "/network"

    def __init__(self, context, request):
        super(ComputerNetworkMonitoringPage, self).__init__(context, request)
        interface = request.get("interface", None)
        account_infos = AccountNetworkInfo(self.account)
        self.network_interfaces = sort_interfaces(
            account_infos.get_network_interfaces(self.computer_ids, interface=interface)
        )
        if "lo" in self.network_interfaces:
            self.network_interfaces.remove("lo")


class BaseDataAsCSVPage(BrowserPage):
    def __call__(self, **kwargs):
        set_csv_download(self.request.response, self.filename)
        return self.context.get_csv_data(**kwargs)


class CustomGraphAsCSVPage(BaseDataAsCSVPage):
    def __call__(self):
        # Remove the security proxy in order to bypass the access delegate
        # permission checker which relies right on the C{graph} attribute.
        if not removeSecurityProxy(self.context).graph:
            raise NotFound(self, "Graph not found")
        return super(CustomGraphAsCSVPage, self).__call__()

    @property
    def filename(self):
        return "%s" % self.context.graph.title


class LoadDataAsCSVPage(BaseDataAsCSVPage):

    filename = "load_data"


class SwapDataAsCSVPage(BaseDataAsCSVPage):

    filename = "swap_data"


class MemoryDataAsCSVPage(BaseDataAsCSVPage):

    filename = "memory_data"


class TemperatureDataAsCSVPage(BaseDataAsCSVPage):
    @property
    def filename(self):
        return "temperature_%s" % self.context.thermal_zone


class UsedSpaceDataAsCSVPage(BaseDataAsCSVPage):
    @property
    def filename(self):
        return "mount_{}".format(self.context.get_encoded_mount_point().decode())


class NetworkDataAsCSVPage(BaseDataAsCSVPage):
    @property
    def filename(self):
        return "network_%s" % self.context.get_network_interface()


class GraphView(BrowserPage):
    def __call__(self):
        data = removeSecurityProxy(self.context.get_data())
        step = self.context.step
        start = self.context.start
        stop = self.context.stop
        if step:
            start -= start % step
            stop -= stop % step
        else:
            step = 300
        result = {
            "step": step,
            "title": self.context.title,
            "start": start,
            "stop": stop,
            "legend": removeSecurityProxy(self.context.legend),
            "label": self.context.label,
            "data": data,
        }
        self.request.response.setHeader("Content-Type", "application/json")
        return json.dumps(result, cls=DecimalEncoder)
