# Copyright 2005-2022 Canonical Limited. All rights reserved.

from canonical.landscape.ui.toolbar.page import ComputerPaginatorPage
from zope.security.management import checkPermission


class ComputerProcessesPage(ComputerPaginatorPage):

    title = "Processes"

    _process_state_labels = {
        "D": "Uninterruptible",
        "R": "Running",
        "S": "Sleeping",
        "T": "Stopped",
        "W": "Paging",
        "X": "Dead",
        "Z": "Zombie",
    }

    def __init__(self, context, request):
        super(ComputerProcessesPage, self).__init__(context, request)
        # processes is a list that contains elements which are single
        # processes or lists of processes.
        self.processes = list(self.computer.get_active_processes())
        self.date_sort_keys = self._get_custom_date_sort_keys(self.processes)
        self._row_css_class = ""
        self.info_age = self.computer.last_exchange_time
        formatter = self.request.principal.person.formatter
        self._cached_users = {}
        self._cached_groups = {}
        updated_at = formatter.format_short_date(self.info_age)
        if updated_at:
            self.notes.info("Last updated at %s." % updated_at)
        self.disable_buttons = "true"
        if checkPermission("landscape.ManageActiveProcess", self.computer):
            self.disable_buttons = "false"

    def _get_custom_date_sort_keys(self, processes):
        flat_processes = []
        for process in processes:
            if isinstance(process, list):
                flat_processes += process
            else:
                flat_processes.append(process)

        start_keys = list(
            zip(
                [process.start_time for process in flat_processes],
                list(range(len(flat_processes))),
            )
        )

        # now sort by the date, the list is now [(date, i), ...]
        start_keys.sort(key=lambda i: i[0])
        # next, get a list of 2-tuple integers. the first integer
        # is the position of the row in date sorted order, the second integer
        # is the original position of the row
        int_keys = list(zip(list(range(len(start_keys))), [k[1] for k in start_keys]))
        # now sort by the original position.  the date sorted
        # position can now be used as the custom key
        int_keys.sort(key=lambda i: i[1])
        items = []
        for i in range(len(int_keys)):
            process = flat_processes[i]
            date_position = int_keys[i][0]
            items.append((process, date_position))
        return dict(items)

    def state_label(self, process):
        return self._process_state_labels.get(process.state.decode())

    def get_username_or_uid(self, process):
        cache_key = (process.computer.id, process.uid)
        if cache_key not in self._cached_users:
            username = process.get_username()
            if username:
                self._cached_users[cache_key] = username
            else:
                self._cached_users[cache_key] = str(process.uid)
        return self._cached_users[cache_key]

    def get_groupname_or_gid(self, process):
        cache_key = (process.computer.id, process.gid)
        if cache_key not in self._cached_groups:
            groupname = process.get_groupname()
            if groupname:
                self._cached_groups[cache_key] = groupname
            else:
                self._cached_groups[cache_key] = str(process.gid)
        return self._cached_groups[cache_key]
