# Copyright 2005-2022 Canonical Limited. All rights reserved.

import datetime
import time

from pytz import timezone

from canonical.routes import route, NotFound, Redirect
from canonical.url import get_url

from canonical.landscape.model.account.customgraph import CustomGraphManagement
from canonical.landscape.model.monitoring.disk import (
    UsedSpaceGraph,
    UsedSpaceDataExporter,
)
from canonical.landscape.model.monitoring.load import LoadGraph, LoadDataExporter
from canonical.landscape.model.monitoring.memory import (
    UsedMemoryGraph,
    UsedMemoryDataExporter,
)
from canonical.landscape.model.monitoring.swap import (
    UsedSwapGraph,
    UsedSwapDataExporter,
)
from canonical.landscape.model.monitoring.temperature import (
    TemperatureGraph,
    TemperatureDataExporter,
)
from canonical.landscape.model.monitoring.custom import (
    CustomGraphGraph,
    CustomGraphExporter,
)
from canonical.landscape.model.monitoring.network import (
    NetworkTrafficGraph,
    NetworkTrafficExporter,
)
from canonical.landscape.ui.routes import LandscapeRoutes
from canonical.landscape.model.resource.network import INBOUND, OUTBOUND
from canonical.landscape.model.activity.group import SIGNAL_PROCESSES_SUMMARY
from canonical.landscape.ui.page.page import check_request_csrf_token


EPOCH = datetime.datetime(1970, 1, 1)

ONE_DAY = 60 * 60 * 24

STEP_SIZE = 300

MAX_DAYS = 3000  # Around 10 years


def extract_start_stop(request):
    """Extract arguments from a request to build a time interval.

    It extracts if present:
        * start, in the form of YYYYMMDD date
        * days, as an integer

    If not present, or if their values are invalid, C{days} default to 1 and
    C{start} default to now minus the period calculated with C{days}.
    """
    try:
        num_days = abs(int(request.get("days", 1)))
    except ValueError:
        num_days = 1
    num_days = min(num_days, MAX_DAYS)
    period = num_days * ONE_DAY

    start = request.get("start")
    if start:
        try:
            timestamp = datetime.datetime(
                int(start[:4]), int(start[4:6]), int(start[6:])
            )
        except ValueError:
            start = None
        except OverflowError:
            start = None
        else:
            delta = timestamp - EPOCH
            start = delta.days * ONE_DAY + delta.seconds
            stop = start + period

    if start is None:
        stop = int(time.time())
        stop -= stop % STEP_SIZE
        start = stop - period

    return start, stop


class GraphRoutes(LandscapeRoutes):
    def get_start_stop(self):
        return extract_start_stop(self.request)

    def wants_large(self):
        return self.request.get("large") == "1"

    def build(self, cls, *args, **kwargs):
        start, stop = self.get_start_stop()
        if "start" not in kwargs:
            kwargs["start"] = start
        if "stop" not in kwargs:
            kwargs["stop"] = stop
        if "large" not in kwargs:
            kwargs["large"] = self.wants_large()
        person = getattr(self.request.principal, "person", None)
        if person is not None:
            kwargs["tzinfo"] = timezone(person.timezone)
        return cls(self.context.selection, *args, **kwargs)

    def build_csv(self, cls, graph_cls, *args, **kwargs):
        return cls(self.build(graph_cls, *args, **kwargs))

    @route("disk")
    def disk(self):
        mount_point = self.request.get("mount_point")
        if mount_point:
            return self.build(UsedSpaceGraph, mount_point)

    @route("disk_as_csv")
    def disk_as_csv(self):
        mount_point = self.request.get("mount_point")
        if mount_point:
            return self.build_csv(
                UsedSpaceDataExporter, UsedSpaceGraph, mount_point=mount_point
            )

    @route("load")
    def load(self):
        return self.build(LoadGraph)

    @route("load_as_csv")
    def load_as_csv(self):
        return self.build_csv(LoadDataExporter, LoadGraph)

    @route("memory")
    def memory(self):
        return self.build(UsedMemoryGraph)

    @route("memory_as_csv")
    def memory_as_csv(self):
        return self.build_csv(UsedMemoryDataExporter, UsedMemoryGraph)

    @route("swap")
    def swap(self):
        return self.build(UsedSwapGraph)

    @route("swap_as_csv")
    def swap_as_csv(self):
        return self.build_csv(UsedSwapDataExporter, UsedSwapGraph)

    @route("temperature")
    def temperature(self):
        thermal_zone = self.request.get("thermal_zone", None)
        if thermal_zone is None:
            raise NotFound(self, "temperature")
        return self.build(TemperatureGraph, thermal_zone)

    @route("temperature_as_csv")
    def temperature_as_csv(self):
        thermal_zone = self.request.get("thermal_zone", None)
        if thermal_zone is None:
            raise NotFound(self, "temperature")
        return self.build_csv(
            TemperatureDataExporter, TemperatureGraph, thermal_zone=thermal_zone
        )

    @route("custom")
    def custom(self):
        account = self.context.selection.account
        graph_id = self.request.get("custom_graph")
        if graph_id:
            management = CustomGraphManagement(account, self.request.principal.person)
            graph = management.get_graphs(ids=[int(graph_id)]).one()
            if graph is not None:
                return self.build(CustomGraphGraph, graph)

    @route("custom_graph_as_csv")
    def custom_graph_as_csv(self):
        account = self.context.selection.account
        graph_id = self.request.get("custom_graph")
        if graph_id:
            management = CustomGraphManagement(account, self.request.principal.person)
            graph = management.get_graphs(ids=[int(graph_id)]).one()
            if graph is not None:
                return self.build_csv(
                    CustomGraphExporter, CustomGraphGraph, custom_graph=graph
                )

    @route("network")
    def network(self):
        """
        Route network graph requests, extracting the C{interface} and
        C{direction} arguments.
        """
        interface = self.request.get("interface")
        direction = self.request.get("direction")
        if interface and direction in (INBOUND, OUTBOUND):
            return self.build(
                NetworkTrafficGraph, interface=interface, direction=direction
            )

    @route("network_as_csv")
    def network_as_csv(self):
        """
        Route network CSV requests, extracting the C{interface} and
        C{direction} arguments.
        """
        interface = self.request.get("interface")
        direction = self.request.get("direction")
        if interface and direction in (INBOUND, OUTBOUND):
            return self.build_csv(
                NetworkTrafficExporter,
                NetworkTrafficGraph,
                interface=interface,
                direction=direction,
            )


class ProcessRoutes(LandscapeRoutes):
    def get_processes(self):
        """
        There's a potential race condition here that this code does
        nothing about: that is, in the time between a user seeing a
        page with a list of processes and selecting some to end or
        kill, processes in the selection could have been reported by a
        client as killed.  In such cases, this code will end up
        searching for a process ID which no longer exists in the
        database.  If all the selected processes have been reported as
        killed the user will get a seemingly strange error like, 'You
        need to select processes before any can be killed'.
        """
        check_request_csrf_token(self.request, require_post=True)

        if "processes" not in self.request or "computer" not in self.request:
            return []

        computer_id = int(self.request["computer"])
        process_ids = (int(process) for process in self.request["processes"])

        account = self.context.selection.account
        computer = account.get_computers(computer_ids=[computer_id]).one()
        processes = computer.get_active_processes_by_ids(process_ids)
        return list(processes)

    @route("action")
    def process_action(self):
        person = self.request.principal.person
        processes = self.get_processes()
        action = self.request.get("action")

        if action in ("end", "kill"):
            if not processes:
                self.notes.error(
                    "Processes must be selected before they can " "be %sed." % action
                )
            else:
                group = self.context.selection.create_activity_group(
                    creator=person, summary=SIGNAL_PROCESSES_SUMMARY
                )
                for process in processes:
                    if action == "end":
                        process.request_end(group)
                    else:
                        process.request_kill(group)

                group.settle()
                group.info.create_log_entry("Group activity created", person=person)
                if len(processes) == 1:
                    self.notes.info("Queued command to %s 1 process." % action)
                else:
                    self.notes.info(
                        "Queued commands to %s %d processes." % (action, len(processes))
                    )

        url = get_url(self.context.selection, self.request)
        raise Redirect("%s/processes" % (url,))
