import base64

from zope.cachedescriptors.property import Lazy

from canonical.landscape.ui.urlformat import read_asset
from canonical.landscape.ui.page.page import Page


IMAGE_ASSETS = {
    "logo-landscape": "skin/hokan/logo-landscape.png",
    "logo-canonical": "skin/hokan/logo-canonical.png",
}

CSS_ASSETS = [
    "yui/cssreset/cssreset-min.css",
    "yui/cssgrids/cssgrids-min.css",
    "skin/hokan/base-min.css",
    "skin/hokan/layout-min.css",
    "skin/hokan/widgets-core-min.css",
]

CSS_OVERRIDES = """
.link-button-image-landscape {
    background-image: url(data:image/png;base64,%(logo-landscape)s);
}
.link-button-image-canonical {
    background-image: url(data:image/png;base64,%(logo-canonical)s);
}

"""


class OfflinePage(Page):
    """Base class for static offline pages to be used by HAProxy.

    From the HAProxy documentation:

    "The files should not exceed the configured buffer size (BUFSIZE), which
     generally is 8 or 16 kB, otherwise they will be truncated. It is also wise
     not to put any reference to local contents (eg: images) in order to avoid
     loops between the client and HAProxy when all servers are down, causing an
     error to be returned instead of an image. For better HTTP compliance, it
     is recommended that all header lines end with CR-LF and not LF alone."

    See also:

    http://cbonte.github.io/haproxy-dconv/configuration-1.5.html#4-errorfile

    and

    http://cbonte.github.io/haproxy-dconv/configuration-1.5.html#1.3.1
    """

    title = "System error"

    @Lazy
    def css(self):
        """Return all CSS rules needed by offline pages.

        This data will be used by the page template to inline all CSS rules
        in a <style> element, rather then having <link> elements that reference
        external resources. This the offline page content served HAProxy will
        keep working even if the application can't be reached and serve static
        content.
        """
        css = ""
        for asset in CSS_ASSETS:
            css += read_asset(asset)
        css += CSS_OVERRIDES % self._images
        return css

    @Lazy
    def _images(self):
        """Map image names to their base64-encoded versions."""
        images = {}
        for name, asset in IMAGE_ASSETS.items():
            data = read_asset(asset)
            images[name] = base64.b64encode(data)
        return images

    def __call__(self):
        result = super(OfflinePage, self).__call__()
        self.request.response.setStatus(self.http_status)
        return result
