# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
OpenID Connect (oidc) integration for Landscape.
"""
import logging
from six import text_type as str
from six.moves.urllib.parse import urlencode
from urllib.parse import parse_qs, urlencode, urlparse

from zope.session.interfaces import ISession

from oic.oic import PyoidcError
from oic.oic.consumer import Consumer, ErrorResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD

from canonical.landscape.ui.lsopenid import OpenIDAuth
from canonical.landscape.model.configuration import get_root_url


class OidcConsumer(Consumer):
    def dictionary(self):
        """Returns a dict of consumer state."""
        # default consumer tries to store way to many unserializable fields
        # on the session, most of which never get read again.
        return {
            "seed": self.seed,
            "nonce": self.nonce,
            "redirect_uris": self.redirect_uris,
        }


class LandscapeOidc(object):
    """
    An object that knows about Landscape's policy for doing OpenID
    authentication.  Most importantly, the URL of Landscape and the
    oidc endpoints.
    """

    default_consumer_factory = OidcConsumer

    def __init__(self, request, oidc_config, consumer_factory=None):
        consumer_factory = consumer_factory or self.default_consumer_factory

        session_data = ISession(request)["landscape.oidc"]
        self.request = request
        self.consumer_config = {
            "response_type": "code",
            "scope": "openid profile email",
            "authz_page": "",
        }
        client_config = {
            "client_id": oidc_config.get("client_id"),
            "client_authn_method": CLIENT_AUTHN_METHOD,
        }
        self.consumer = consumer_factory(
            session_data,
            self.consumer_config,
            server_info=oidc_config,
            client_config=client_config,
        )
        self.consumer.client_secret = oidc_config.get("client_secret")
        self.consumer.provider_config(oidc_config["issuer"])

    def begin(self, return_to_suffix):
        """
        Redirect the request to the OpenID provider so that he can
        authenticate.
        """
        return_to_suffix = self.extract_query_params(return_to_suffix)
        self.consumer_config["authz_page"] = return_to_suffix
        path = get_root_url() or self.request.getApplicationURL()
        path = path.rstrip("/")

        sid, location = self.consumer.begin(path=path)
        self.request.response.redirect(location, trusted=True)

    def extract_query_params(self, url):
        """
        We might have extra query params. OIDC providers may not
        recognize the resulting url as a valid configured
        redirect uri, so we extract them and set them as cookies.
        """
        parsed_url = urlparse(url)
        query = parse_qs(parsed_url.query, keep_blank_values=True)
        for key, value in query.items():
            self.request.response.setCookie(key, value[0] if value else "")
        return parsed_url.path

    def complete(self):
        """
        Given a request that contains a response from an OpenID provider (via
        the user agent), check if OpenID authentication succeeded. If it has,
        an authentication object will be returned.
        """
        query_string = self.request.get("QUERY_STRING")
        try:
            # In authorization flow, we only get a code/state at this step
            auth_res, access, idtoken = self.consumer.parse_authz(query_string)
            if isinstance(auth_res, ErrorResponse):
                logging.error("oidc access error: %r", auth_res)
                return None
            state = auth_res.get("state", "")

            access = self.consumer.complete(state)
            if isinstance(access, ErrorResponse):
                logging.error("oidc token error: %r", access)
                return None
            # optionally update the state, if provided
            state = access.get("state", state)
            idtoken = access["id_token"]

            userinfo = self.consumer.do_user_info_request(state=state, schema="openid")
            if isinstance(userinfo, ErrorResponse):
                # We don't error here because the needed info can be in the token
                logging.error("oidc userinfo error: {}".format(userinfo))
            else:  # Some providers (e.g. MS) return sub-only userinfo
                idtoken.update(userinfo)

        except PyoidcError as err:
            logging.error("OIDC flow returned an error: %s", err)
            return None

        # Complete the id_token claims
        identity_url = idtoken["sub"]
        email_address = idtoken.get("email", "").strip()

        if not email_address or not identity_url:
            return None

        name = idtoken.get("name", "")
        if not name:
            given_name = idtoken.get("given_name", "")
            family_name = idtoken.get("family_name", "")
            name = "{} {}".format(given_name, family_name)
            name = name.strip()  # Strip in case ' ' so it becomes ''
        timezone = idtoken.get("zoneinfo", "")
        return OpenIDAuth(identity_url, name, email_address, timezone)


class LandscapeOidcFactory(object):
    """A Landscape/OpenID policy which uses OpenID-connect."""

    def __init__(self, oidc_config=None):
        self.config = oidc_config

    def get_logout_url(self, return_to):
        """Discover the end_session_endpoint, if there is one."""

        try:
            # Allow explicit override in server config, since most OP
            # don't expose one, and not having any yields a poor experience.
            return self.config["logout_url"]
        except KeyError:
            pass

        discovered = OidcConsumer(None, None).provider_config(self.config["issuer"])
        logout_url = discovered.get("end_session_endpoint")

        if not logout_url:
            # According to the spec, session management endpoints are optional.
            # Google doesn't return any, ADFS may (KB4038801), and Azure does.
            # If we don't have a return url, just return directly.
            return return_to

        # XXX Some OP use hints about the identity signing out. I have yet to
        # see one which needs this, so that's not done here.
        return "{}?{}".format(
            logout_url, urlencode({"post_logout_redirect_uri": return_to})
        )

    def get_landscape_openid(self, request, consumer_factory=None):
        oidc = LandscapeOidc(request, self.config, consumer_factory=consumer_factory)
        return oidc
