# Copyright 2005-2023 Canonical Limited. All rights reserved.

from urllib.parse import parse_qsl

from zope.security.proxy import removeSecurityProxy
from zope.cachedescriptors.property import Lazy

from canonical.url import get_url, get_context_url

from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.model.main.standalone import LDSScriptExecutionManagement
from safe.canonical.landscape.model.package.alerts import EsmDisabledAlert
from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError

from canonical.landscape.ui.activity.schedule import (
    get_deliver_after_time,
    get_deliver_delay_window,
)
from canonical.landscape.ui.action import ActionItem
from canonical.landscape.ui.toolbar.page import ToolbarPage


NO_PACKAGE_INFORMATION_AVAILABLE = (
    "No package information is available for the selected computers."
)

FILTER_URL = "list?filter="

SUMMARIES = {
    "security-upgrade": ActionItem(
        icon="/static/package/security-upgrades-alert-16x16.png",
        summary="%(count)s computers have %(action)s",
        summary_single="%(count)s computer has %(action)s",
        action="security upgrades to install",
        action_url=FILTER_URL + "security",
    ),
    "upgrade": ActionItem(
        icon="/static/package/package-upgrades-alert-16x16.png",
        summary="%(count)s computers have %(action)s",
        summary_single="%(count)s computer has %(action)s",
        action="upgrades to install",
        action_url=FILTER_URL + "upgrades",
    ),
    "held": ActionItem(
        icon="/static/package/package-held-16x16.png",
        summary="%(count)s computers have %(action)s",
        summary_single="%(count)s computer has %(action)s",
        action="held packages",
        action_url=FILTER_URL + "held",
    ),
    "up-to-date": ActionItem(
        icon="/static/activity/succeeded-16x16.png",
        summary="%(count)s computers are %(action)s",
        summary_single="%(count)s computer is %(action)s",
        action="fully up-to-date",
        action_url="",
    ),
    "no-packages": ActionItem(
        icon="/static/alerts/alert-16x16.png",
        summary="%(count)s computers have not %(action)s",
        summary_single="%(count)s computer has not %(action)s",
        action="reported package information",
        action_url="",
    ),
}


class PackageDashboardPage(ToolbarPage):

    title = "Package management"
    paginate = False

    def __init__(self, context, request):
        super(PackageDashboardPage, self).__init__(context, request)
        self._search = self.context.get_search()
        # Is *any* package available at all for the given selection of
        # computers?
        if self._has_no_packages:
            self.notes.warning(NO_PACKAGE_INFORMATION_AVAILABLE)
        else:
            # If any packages are available, when were they last updated?
            # Strictly, this should take the minimum of the selection but
            # that's a potentially lengthy operation so as in the process
            # page we arbitrarily pick the first machine's update time
            for computer in self.context.selection:
                if computer.last_exchange_time is not None:
                    formatter = self.request.principal.person.formatter
                    self.notes.info(
                        "Last updated at %s."
                        % formatter.format_short_date(computer.last_exchange_time)
                    )
                    break

        if is_standalone():
            management = LDSScriptExecutionManagement()
            execution = management.get_execution("update_security_db.sh")
            if execution and execution.errored:
                self.notes.error(
                    "USN update cron script is not running, "
                    "USN information may be out of date."
                )

    @Lazy
    def has_disabled_esm(self):
        """Returns if selected computers have the ESM disabled alert active."""
        alert = removeSecurityProxy(EsmDisabledAlert)
        return not self.context.selection.account.get_rules(
            [alert], active=True, computers=self.context.selection
        ).is_empty()

    @Lazy
    def release_upgrades_count(self):
        """Return how many computers have an available release upgrade."""
        return self.context.get_computers_with_release_upgrades_count()

    @Lazy
    def counts_by_computer(self):
        """For the first 5 selected computer, count packages in each state."""
        return self._search.get_package_counts_by_computer(offset=0, limit=5)

    @Lazy
    def _counts(self):
        return self._search.get_computer_counts()

    @Lazy
    def _computer_counts(self):
        return self._counts[0]

    @Lazy
    def usn_computers(self):
        return self._counts[1]

    @Lazy
    def _has_no_packages(self):
        """Whether none of the selected computers has reported packages."""
        no_packages_count = self._computer_counts["no-packages"]
        return no_packages_count == self.context.selection.get_selected_count()

    @Lazy
    def controls_disabled(self):
        """Whether to disable form controls in the template."""
        return "disabled" if self._has_no_packages else ""

    def request_upgrades(self):
        """
        Request all possible package upgrades on computers in the selection.
        """
        self._check_request_csrf_token(require_post=True)
        deliver_delay_window = get_deliver_delay_window(self.request)

        activity = self.context.request_upgrades(
            self.request.principal.person, deliver_after_window=deliver_delay_window
        )
        self.notes.info("Upgrade verification queued")
        self.request.response.redirect(get_url(activity, self.request))

    def apply_changes(self):
        self._check_request_csrf_token()
        account = self.context.selection.account
        person = self.request.principal.person

        if not any(name.startswith("changes-") for name in self.request.form):
            self.notes.error("No changes selected")
            self.request.response.redirect(get_url(self.context, self.request))
            return

        deliver_after = get_deliver_after_time(self.request)
        # Random delay window in minutes
        deliver_delay_window = get_deliver_delay_window(self.request)

        try:
            activity = self.context.change_packages(
                person,
                self._get_computer_changes(),
                deliver_after=deliver_after,
                deliver_after_window=deliver_delay_window,
            )
        except InvalidScheduleAfterWindowError as error:
            self.notes.error(str(error))
            self.request.response.redirect(get_url(self.context, self.request))
            return

        self.notes.info("Package changes queued")
        account_url = get_url(account, self.request)
        activity_path = get_context_url(activity, self.request).path
        self.request.response.redirect("%s/%s" % (account_url, activity_path))

    @Lazy
    def package_status_summaries(self):
        """
        Return summaries about the status of the packages across the selection.
        """
        return self._compute_summaries(
            "security-upgrade", "upgrade", "held", "up-to-date"
        )

    @Lazy
    def package_data_summaries(self):
        """
        Return a summary about which computers have not reported package data.
        """
        return self._compute_summaries("no-packages")

    def _compute_summaries(self, *kinds):
        """
        Compute summaries for the actions of the given kind to be shown in the
        Summary section of the dashboard.
        """
        account_name = self.context.selection.account.name
        summaries = []
        for kind in kinds:
            computers_count = self._computer_counts.get(kind)
            if computers_count:
                item = SUMMARIES[kind]
                summary = item.get_summary(account_name, computers_count)
                summaries.append(summary)
        return summaries

    def _get_computer_changes(self):
        """
        Build the list of computer changes sent by the browser. If the all
        upgrades flag is specified, it will recalculate the upgrades to do,
        otherwise it will extract the packages from the form.

        @return: A list of (computer, list of package IDs to install, list of
            package IDs to remove).
        """
        computer_changes = []
        all_upgrades = self.request.get("changes-select-all-upgrades")
        if all_upgrades:
            if isinstance(all_upgrades, list):
                # If several values are passed, only consider the first
                all_upgrades = all_upgrades[0]
            arguments = dict(parse_qsl(all_upgrades))
            query = arguments.get("q")
            package_filter = arguments.get("filter")
            usn = arguments.get("usn")

            if package_filter in ("held", "installed"):
                raise RuntimeError("Can't upgrade held or installed.")
            computer_changes = removeSecurityProxy(
                self._search.get_package_upgrades_by_computer_id(
                    text=query,
                    usn_name=usn,
                    security_only=(package_filter == "security"),
                )
            )

        else:
            for computer_id in self.context.selection.get_selected_computer_ids():
                installs = []
                removals = []
                holds = []
                remove_holds = []
                for value in self.request.get("changes-%d" % computer_id, ()):
                    if not value:
                        continue
                    if value[0] == "i":
                        installs.append(int(value[1:]))
                    elif value[0] == "r":
                        removals.append(int(value[1:]))
                    elif value[0] == "h":
                        holds.append(int(value[1:]))
                    elif value[0] == "u":
                        remove_holds.append(int(value[1:]))
                    else:
                        raise RuntimeError("Improper value: %s" % repr(value))
                computer_changes.append(
                    (computer_id, installs, removals, holds, remove_holds)
                )
        return computer_changes
