# Copyright 2005-2022 Canonical Limited. All rights reserved.

from pytz import timezone, UTC
from json import dumps

from zope.security.proxy import removeSecurityProxy

from canonical.lib.timestamp import utcnow
from canonical.url import get_url
from canonical.content.decorator import json
from canonical.landscape.ui.toolbar.page import ToolbarPage


class PackageListingPage(ToolbarPage):

    title = "Package Listing"
    total_packages = 0
    package_offset = 0
    packages = None

    def __init__(self, context, request):
        super(PackageListingPage, self).__init__(context, request)
        self.package_filter = self.request.get("filter")
        self.text = self.request.get("q", "").strip() or None
        self.usn_name = self.request.get("usn") or None
        self.package_data = {}
        self.package_offset = 0
        self.total_packages = 0
        self.has_packages = False
        self.size = int(self.request.get("size", "25"))
        search = context.get_search()

        tz = timezone(request.principal.person.timezone)
        utc_now = utcnow()
        # We need to localize it to make it timezone and DST aware
        utc_target = UTC.localize(utc_now)
        # Translate the date into the user's local timezone
        local_target = utc_target.astimezone(tz)
        self.at_time = local_target.strftime("%Y-%m-%d %H:%M")

        if self.size <= 0:
            self.notes.error("Invalid listing size: %d." % self.size)
            return

        if not (self.text or self.usn_name or self.package_filter):
            self.notes.error("No search terms provided.")
            return

        try:
            offset = int(self.request.get("offset", 0))
            filter_options = self._get_filter_options(self.package_filter)
            filter_options["limit"] = self.size
            filter_options["text"] = self.text
            filter_options["usn_name"] = self.usn_name
            if search.is_using_external():
                filter_options["mark"] = offset
            else:
                filter_options["offset"] = offset
            search_result = search.get_relations_by_package_name(**filter_options)
        except ValueError as e:
            self.notes.error("%s." % e)
        else:
            self.package_data = self._get_package_info(
                removeSecurityProxy(search_result)
            )
            if search.is_using_external():
                self.package_offset = search_result["mark"]
                if self.package_offset == -1:
                    # Because of the extreme difficulty of getting a value into
                    # the c.l.javascript.listing.pager.PaginationPlugin, and
                    # have it updated with each new page of results, we're now
                    # building on top of the existing hack and setting the
                    # total_packages value to -2 to indicate that total
                    # packages is unknown, but we know there are no more
                    # results after this batch.
                    self.total_packages = -2
                else:
                    self.total_packages = -1
                self.has_packages = len(search_result["packages"]) > 0
            else:
                self.package_offset = offset + len(search_result["relations"])
                self.total_packages = search_result.get("totalPackages", 0)
                self.has_packages = self.total_packages > 0

            if not self.has_packages:
                self.notes.error("No packages matching your search were found.")

    def _get_filter_options(self, package_filter):
        filter_options = {}
        if package_filter in ("installed", "held", "upgrades"):
            filter_options[package_filter] = True
        elif package_filter == "security":
            filter_options["usn"] = True
            filter_options["upgrades"] = True
        return filter_options

    def packages_script(self):
        javascript = (
            "var computerURL = %s;\n"
            "var pageSize = %d;\n"
            "var packageData = %s;\n"
            "var totalPackages = %d;\n"
            "var packageOffset = %d;\n"
            "var sourceArguments = {filter: %s, q: %s, usn: %s};\n"
        ) % (
            self.view_safe_js_value(get_url(self.context.selection, self.request)),
            self.size,
            dumps(self.package_data),
            self.total_packages,
            self.package_offset,
            self.view_safe_js_value(self.request.get("filter", "")),
            self.view_safe_js_value(self.request.get("q", "")),
            self.view_safe_js_value(self.request.get("usn", "")),
        )
        return javascript.replace("</script", "<\\/script")

    def _get_package_info(self, search_result):
        """Get a C{dict} with L{Package}s search results from the results
        returned by L{PackageSearch.search.get_relations_by_package_name()}.

        The C{dict} has information about L{Computer}s, L{Package}s and the
        relationships between each L{Computer} and each L{Package}.  It is in
        a format that can be converted to JSON for use by Javascript code that
        renders the package listing.  It has the following structure::

          {'computers': {id: title, ...},
          'packages': {id: {'name': name, 'version': version,
                             'summary': summary,
                             'usn': {'name': name, 'summary': summary}}, ...},
           'relations': [{'name': package_name,
                          'package_ids': [id1, id2, ...],
                          'upgrades': upgrade_count,
                          'securityUpgrades': upgrade_count,
                          'relations': [computer_relation1, computer_relation2,
                                        ...]},
                         ...]
          }
        """
        relations = search_result["relations"]

        if not relations:
            return {}

        packages = self._get_packages_data(search_result["packages"])
        computers = self._get_computers_data(search_result["computers"])

        return {"computers": computers, "packages": packages, "relations": relations}

    def _get_computers_data(self, computers):
        """
        Returns a C{dict} with computers information in the following format:
          {id: {'title': title}, ... }
        """
        return dict((computer.id, {"title": computer.title}) for computer in computers)

    def _get_packages_data(self, packages_by_id):
        """
        Returns a C{dict} with the packages found and in the following format:
           {id: {'name': name, 'version': version, 'summary': summary,
                 'usn': {'name': name, 'summary': summary}}, ...},
        """

        def get_data(package):
            data = {
                "name": package.name,
                "version": package.version,
                "summary": package.summary,
            }
            if package.usn:
                data["usn"] = {
                    "name": package.usn.name,
                    "summary": package.usn.summary,
                    "cves": [cve.id for cve in package.usn.cves],
                }
            return data

        return dict(
            (package_id, get_data(package))
            for package_id, package in packages_by_id.items()
        )

    @json
    def available_package_json(self):
        """
        Returns a structure containing all the package information needed for
        paginate the package listing.
        """
        self.request.response.setHeader("Content-Type", "application/json")
        return {
            "totalPackages": self.total_packages,
            "packageOffset": self.package_offset,
            "packages": self.package_data,
        }
