# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime

from pytz import UTC, timezone

from zope.contentprovider.interfaces import IContentProvider
from zope.security.interfaces import Unauthorized
from zope.security.management import checkPermission
from zope.component import getMultiAdapter, getUtility
from zope.interface import implementer, directlyProvides
from zope.authentication.interfaces import IUnauthenticatedPrincipal
from zope.cachedescriptors.property import Lazy
from zope.publisher.browser import BrowserPage
from zope.publisher.interfaces import NotFound as ZopeNotFound

from zope.formlib.form import setUpWidgets, getWidgetsData
from zope.formlib.interfaces import IWidgetInputError

from canonical.lib.serialize import json_dumps
from canonical.lib.text import lower_class
from canonical.amqp.message.interfaces import IDispatcher
from canonical.content.form import (
    Fields,
    render_form_token,
    check_invariants,
    form_token,
)
from canonical.content.decorator import json
from canonical.notes.session import SessionNotes
from canonical.routes import BadRequest
from canonical.routes.publisher import get_route_object_by_interface
from canonical.security.interfaces import IRoleRegistry
from canonical.statuscode.interfaces import IStatusCode

from canonical.landscape.application import PACKAGE_VERSION
from safe.canonical.landscape.model.action.registry import get_action_handler
from canonical.landscape.model.action.error import ActionError
from canonical.landscape.model.configuration import (
    is_testing_environment,
    is_staging_environment,
    is_standalone,
    get_root_url,
    get_static_root_path,
    is_development_mode,
)
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.ui.service import service_lookup
from canonical.landscape.ui.page.ajax import AjaxPage
from canonical.landscape.ui.page.csp import make_csp_header, DEFAULT_CSP
from canonical.landscape.ui.interfaces import ILandscapePage, IActionRequest
from canonical.landscape.interfaces import IFileBasedConfiguration


# A list of URLs on which we should NOT check CSRF for POSTs. This is for when
# a post to a page invalidates the session (it's the case when logging in for
# example). Since the CSRF token is stored in the session, the check otherwise
# always fails.
CSRF_WHITELIST = frozenset(["/login/authenticate", "/payment-notification"])


@implementer(ILandscapePage)
class Page(BrowserPage):
    """Base class for Zope views returning HTML content.

    It requires an L{IContentProvider} named 'page' adapting the view, the
    request and the context.

    Normally, the content provider is a template registered for the request's
    skin interface. The template is populated by additional content providers
    for the given context and view.

    @ivar title: Used by the skin template for the C{<title>} header tag.

    @ivar layout: Either "fluid", "fixed" or "wide". This only affects
        the central part of the page (i.e. the "page" and "toolbar_"
        providers).  See also the page template at
        C{canonical/landscape/ui/skin/hokan/page.pt}.

    @ivar account: For account pages (i.e. all pages handling URLs under the
        C{/account/<account-name/} tree), the L{Account} they are bound to.

    @ivar namespaces: A list of notification namespaces the browser page will
        be bound to. The JavaScript browser code will receive all notifications
        that the server generates against such namespaces. All account pages
        are automatically bound to the L{Account.namespace}, and will receive
        account-level notifications such as C{account_rules_changed} (see
        L{c.l.model.alert.subscribers}). Unless there are specific visibility
        requirements (e.g. you only want to send a certain notifications to
        certain admins of in a certain account) the account namespace should be
        enough, because the JavaScript code of a page will set subscribers only
        for notifications it cares about and ignore the rest.

    @ivar page_key: A UUID identifying the page. It will be included as a data
        attribute in the C{<body>} element of the template and the JavaScript
        code must present it back to the server when polling for notifications,
        so the notification can "authenticate" it and know what notifications
        to dispatch. See also C{c/l/javascript/notifications/transport.js}.
    """

    __page_attribute__ = None

    title = "Landscape"
    tight = ""  # CSS class for the page's tight-content div
    page_key = None
    layout = "fluid"
    csp = make_csp_header(DEFAULT_CSP, inherit=None)

    requires = None

    def __init__(self, context, request):
        super(Page, self).__init__(context, request)
        self.year = datetime.today().year
        self.version = PACKAGE_VERSION.split("-")[0]
        self.namespaces = []
        self.account = None
        if request is not None:
            if "page_key" in request:
                # When JavaScript client code performs HTTP requests that end
                # up invoking job-firing Zope views, it should include the page
                # key in the request, so those jobs will be fired re-using that
                # key and the client code itself will receive a notification
                # when the job is done.
                self.page_key = request["page_key"]
            self.account = get_route_object_by_interface(IAccount, request)
            if self.account is not None:
                self.namespaces.append(self.account.namespace)

    @Lazy
    def permission_context(self):
        """
        The context L{LandscapePage.requires} permission is checked on.

        Defaults to the page context, can be overriden by subclasses.
        """
        return self.context

    @Lazy
    def notes(self):
        """
        An L{INoteContainer} instance that can be used to generate user
        messages.
        """
        return SessionNotes(self.request)

    @Lazy
    def js_page_key(self):
        """
        Safely escapes the page key so that it can rendered in JS scripts.
        """
        return self.view_safe_js_value(self.page_key)

    def root_url(self):
        """Return the root url."""
        return get_root_url()

    def static_root_path(self):
        """Return a path prefix to be used for static resources."""
        return get_static_root_path()

    def is_authenticated(self):
        """Return C{True} if the request principal is authenticated."""
        # XXX This should be an IPathAdapter for principals.
        return (
            self.request.principal is not None
            and not IUnauthenticatedPrincipal.providedBy(self.request.principal)
        )

    def is_staging_environment(self):
        """Return C{True} if we're running in a staging environment."""
        return is_staging_environment()

    def is_testing_environment(self):
        """Return C{True} if we're running in a testing environment."""
        return is_testing_environment()

    def is_development_mode(self):
        """
        Return C{True} if we're running in a dev environment, which does not
        include staging.
        """
        return is_development_mode()

    def is_selenium_environment(self):
        """Return C{True} if we're running in a selenium environment."""
        config = getUtility(IFileBasedConfiguration)
        is_selenium = config.get("global", key="selenium", default=None)
        return bool(is_selenium)

    def is_standalone(self):
        """Return C{True} if we're running in a standalone environment."""
        return is_standalone()

    def is_support_only_user(self, account):
        """
        Returns C{True} if the current logged person is a support-only user in
        the given C{account}.
        """
        index = self.request.principal.person.get_all_roles()
        registry = getUtility(IRoleRegistry)
        owner_role = registry.get("Owner")
        support_only_role = registry.get("SupportUser")
        for access_context_id, roles in list(index.items()):
            if (
                access_context_id == account.access_context_id
                and support_only_role in roles
                and owner_role not in roles
            ):
                return True
        return False

    def is_free_license(self):
        """Return True if we're running with only free licenses."""
        if self.account is None:
            return True
        return all(license.is_free() for license in self.account.licenses)

    def set_default_status_code(self):
        """Set status code using IStatusCode if it wasn't previously set."""
        if self.request.response.getStatus() == 599:
            status_code = IStatusCode(self.context, None)
            if status_code is not None:
                self.request.response.setStatus(status_code)

    def browserDefault(self, request):
        """Get the default object.

        @param request: The request object used during traversal.
        @return: Returns an object and a sequence of names.  If the sequence
            of names is not empty, then a traversal step is made for each
            name.  After the publisher gets to the end of the sequence, it
            will call browserDefault on the last traversed object.
        @seealso: C{zope.publisher.interfaces.browser.IBrowserPublisher} for
            more information about this method.
        """
        if self.__page_attribute__ is not None:
            return getattr(self, self.__page_attribute__), ()
        return self, ()

    def _initialize_job_queue(self):
        """
        Initialize the job queue for this page, setting the page_key attribute,
        for the given namespaces.
        """
        if self.page_key is None:
            dispatcher = getUtility(IDispatcher)
            self.page_key = dispatcher.create_job_queue(self.namespaces)

        # Fire a job to fetch active alerts (for account pages)
        self._create_alerts_job()

    def _create_alerts_job(self):
        """Fire a job to fetch active account alerts."""
        if self.account is not None and checkPermission(
            "landscape.QueryAlertRule", self.account
        ):
            self.account.create_job_get_active_rules(
                self.page_key, self.request.principal.person
            )

    def _before_render(self):
        """Hook for sub-classes to run code before the provider is rendered.

        Sub-classes can use this hook to perform tasks like firing jobs or
        initialize state that will be used by the content provider.
        """

    def localize_time(
        self, time: datetime, time_format: str = "%Y-%m-%d %H:%M:%S"
    ) -> str:
        utc_time = UTC.localize(time)

        if not self.is_authenticated():
            return utc_time.strftime(time_format)

        tz = timezone(self.request.principal.person.timezone)
        local_time = utc_time.astimezone(tz)
        return local_time.strftime(time_format)

    def render_form_token(self):
        """Return the HTML input for the form token used for CSRF protection.

        C{form_token.validate} needs to be called at submit time to check it.
        """
        return render_form_token(self.request)

    def get_form_token(self):
        """Return the value of the form token used for CSRF protection.

        C{form_token.validate} needs to be called at submit time to check it.
        """
        return form_token.get(self.request)

    def view_safe_js_value(self, value):
        """
        Escape value to make it safe to include in script content.
        """
        value = json_dumps(value)
        return value.replace("</script", "<\\/script")

    def _check_request_csrf_token(self, require_post=False):
        return check_request_csrf_token(self.request, require_post=require_post)

    def _service_lookup(self, lookup_class, *args, **kwargs):
        """Convenience for performing service layer lookups.

        @see canonical.landscape.ui.service.lookup_entity
        """
        try:
            return service_lookup(self.request, lookup_class, *args, **kwargs)
        except ZopeNotFound as error:
            # Set the context object of the NotFound error to this view.
            error.ob = self
            raise error

    def __call__(self):
        """Render the page."""
        if self.requires is not None:
            if not checkPermission(self.requires, self.permission_context):
                raise Unauthorized()
        # HEAD requests should bail out early to avoid creating
        # sessions from HAProxy pings and the like.
        if self.request.method == "HEAD":
            return None
        if self.request.get("PATH_INFO") not in CSRF_WHITELIST:
            self._check_request_csrf_token()

        # Initialize the job queue
        self._initialize_job_queue()

        # Let sub-classes execute custom code.
        self._before_render()

        # Asking for a content provider here rather than providing the
        # template in this view by itself allows different base pages
        # for different skins.
        provider = getMultiAdapter(
            (self.context, self.request, self), IContentProvider, "page"
        )

        provider.update()
        result = provider.render()
        self.request.response.addHeader("Content-Security-Policy", self.csp)

        self.set_default_status_code()
        if self.request.response.getStatus() in (302, 303):
            return None
        return result

    def redirect(self, redirect_url, trusted=None, status=None):
        """
        A wrapper around Zope's redirect to protect against reponse splitting.

        To do so newlines and carriage-returns are trimmed from redirect urls.
        This is similar what's implemented in canonical.routes.publisher, but
        for callsites which are outside of publisher wrapping.
        """
        sanititized_url = redirect_url.replace("\r", "").replace("\n", "")

        if status is not None or trusted is not None:
            # Duplicated to avoid possibility of changing defaults
            return self.request.response.redirect(
                sanititized_url, trusted=trusted, status=status
            )

        return self.request.response.redirect(sanititized_url)


class LandscapePage(Page):
    """Base class for page views using legacy design patterns."""


class ActionPage(AjaxPage):
    """Base page for handling AJAX requests from the browser.

    By default the L{ActionHandler} matching the name of the sub-class will be
    invoked an passed the keyword arguments returned by the L{get_params}
    method. For example, given these action page and handler:

        class SampleActionPage(ActionPage):

            def get_params(self, data):
                return {'handler_parameter': data['some.form.field']}

        class SampleActionHandler(ActionHandler): # This would be in the model

            def handle(self, handler_parameter):
                # Do business logic
                return {'hello': 'world'}

    the dictionary returned by SampleActionPage.get_params will be passed
    as keyword arguments to SampleActionHandler.handle. In many cases you
    don't even need to customize get_params, if the form field names match
    exactly the parameter names.

    If the action handler's handle() method executes without errors, the page
    returns to the browser a JSON-encoded dictionary containing a 'success' key
    mapping to the action handler's result.

    In case of errors, the dictionary will contain a 'widget_errors' key whose
    value is a dictionary mapping form field names to the associated errors,
    and a 'global_errors' key pointing to a list of global errors.

    @ivar fields: Optional Zope L{Fields} instance that the will be used for
        validating and converting the input before it gets passed on to the
        action handler.
    @ivar prefix: Optional name prefix when validating from parameters.
    """

    fields = Fields()
    prefix = ""

    @json
    def __call__(self):
        action = lower_class(self.__class__, suffix="ActionPage")
        handler = get_action_handler(
            action,
            uuid=self.request.form.pop("page_key", self.page_key),
            context=self.context,
            person=self.request.principal.person,
        )
        status = 200
        try:
            data = self.get_params(self._validate())
            result = handler.handle(**data)
        except ActionError as error:
            widget_errors = {}
            global_errors = []
            for reason in error.get_reasons():
                if reason.param is not None:
                    widget_errors[reason.param] = reason.message
                else:
                    global_errors.append(reason.message)
            result = {"widget_errors": widget_errors, "global_errors": global_errors}
            status = 400
        except Unauthorized:
            widget_errors = {}
            global_errors = ["Unauthorized"]
            result = {"widget_errors": widget_errors, "global_errors": global_errors}
            # This is not 401 since that would cause the browser to use
            # a default authorization prompt.
            status = 403
        self.request.response.setStatus(status)
        return result

    def get_params(self, data):
        """Get a C{dict} of keyword params to be passed to the action handler.

        By default the validated form data are passed.

        Sub-classes can override this to tweak the names and the values of
        the parameters being passed.
        """
        return data

    def _validate(self):
        """Validate the form data.

        If the sub-class defines a C{fields} attribute, it will be used to
        validate and convert the form data.
        """
        if self.fields is None:
            # Just pass the raw parameters over to the action handler
            data = self.request.form.copy()
        else:
            # Validate the raw parameters against the given schema
            action_error = ActionError()
            has_widget_errors = False

            # Remove the security proxy so on the schema so we can access
            # its state.
            directlyProvides(self.request, IActionRequest)
            widgets = setUpWidgets(
                self.fields, self.prefix, request=self.request, context=self.context
            )
            data = {}
            errors = getWidgetsData(widgets, self.prefix, data)
            for error in errors:
                if IWidgetInputError.providedBy(error):
                    has_widget_errors = True
                else:
                    action_error.add_reason(message=error)
            if has_widget_errors:
                for widget in widgets:
                    message = widget.error()
                    if message:
                        action_error.add_reason(message=message, param=widget.name)
            for invalid in check_invariants(self.fields, data):
                action_error.add_reason(message=str(invalid))

            if action_error.get_reasons():
                raise action_error

        return data


def check_request_csrf_token(request, require_post=False):
    """Check for a valid CSRF protection token in POST requests.

    If a POST request without a valid CSRF token is performed,
    L{Unauthorized} is raised.

    For GET requests, this is a no-op.

    @param request: The request to validate.
    @param require_post: If C{True}, require a POST method, raise
        L{BadRequest} otherwise.
    """
    if request.method == "POST":
        if not form_token.validate(request):
            raise Unauthorized("Missing/invalid CSRF token")
    elif require_post:
        raise BadRequest("POST is required")
    return True
