# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import object
from storm.locals import Desc


class PaginationMixin(object):
    """A View Mixin for pagination request variable extraction."""

    # subclasses should define this value as a dictionary of
    # column name -> column.
    _sortable_columns = None

    def get_sort_column(self, sort_column, sort_order=None):
        """
        Given a sort column name and value for direction, return a storm
        column suitable for ordering, via looking up the sort column
        name through the subclass defined _sortable_columns class attribute
        dictionary.

        The default sorting order is ascending, to get a descending sort
        order the string value 'desc' must be passed as an argument.
        """
        if not self._sortable_columns:
            raise NotImplementedError("Subclass must define _sortable_columns")

        if sort_column not in self._sortable_columns:
            return None

        sort_column = self._sortable_columns[sort_column]

        if sort_order == "desc":
            sort_column = Desc(sort_column)

        return sort_column

    def get_sort_variables(self, default_limit=20):
        """Extract from the request pagination values.

        @return: A dictionary with the following keys: C{limit}, C{offset},
            C{sort} and C{sort_order}.  C{limit} and C{offset} are integers.
        """
        limit = self.request.get("size")
        offset = self.request.get("offset")
        if limit and limit.isdigit():
            limit = int(limit)
        else:
            limit = default_limit

        if offset and offset.isdigit():
            offset = int(offset)
        else:
            offset = 0

        sort_column = self.request.get("sort")
        sort_order = self.request.get("sort_dir")

        return {
            "limit": limit,
            "offset": offset,
            "sort": sort_column,
            "sort_order": sort_order,
        }
