# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.schema import TextLine, Password, Choice
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from zope.security import checkPermission
from zope.cachedescriptors.property import Lazy
from zope.event import notify
from zope.session.interfaces import ISession

from canonical.content.form import Fields, action
from canonical.pam import pam
from canonical.url import get_url

from canonical.landscape.model.configuration import get_api_url
from canonical.landscape.model.main.interfaces import IPerson
from canonical.landscape.model.main.schema import EmailAddress
from canonical.landscape.security.interfaces import PrincipalLoginEvent
from canonical.landscape.ui.form import LandscapeFormProvider, DropdownWidget
from canonical.landscape.ui.page.page import LandscapePage
from canonical.landscape.ui.sidebar.page import SidebarPage
from canonical.landscape.ui.salesforce.key import validate_salesforce_key
from canonical.landscape.model.main.person import get_person_by_email
from canonical.landscape.model.access.management import AccessControlManagement


class BasePersonPageMixin(object):
    @Lazy
    def person_is_current_user(self):
        """Return C{True} if the context is the person logged in."""
        return self.context == self.request.principal.person

    @Lazy
    def edit_allowed(self):
        """Return C{True} if the person is allowed to edit the person page."""
        return self.person_is_current_user or checkPermission(
            "landscape.EditSalesforceKey", self.context
        )

    @Lazy
    def accounts(self):
        """Return the list of accounts associated with the viewed person."""
        return list(self.context.accounts)

    @Lazy
    def api_endpoint(self):
        """Return the API endpoint."""
        return get_api_url()

    @Lazy
    def show_api_credentials(self):
        """Returns C{True} if API credentials should be be displayed in page.

        We display API credentials if the user is looking at his own person
        page and if he is member of at least one account.
        """
        return self.person_is_current_user and len(self.accounts) > 0

    @Lazy
    def api_credentials(self):
        """Return API credentials for current Person."""
        person = self.context
        credentials = {}
        for account in self.accounts:
            credentials[account] = person.get_api_credentials(account)
        return credentials

    def generate_api_credentials(self):
        """Generate new API credentials for current Person."""
        self._check_request_csrf_token()

        if self.request.method == "POST":
            person = self.context
            account = self.account
            if person.get_api_credentials(account) is not None:
                person.delete_api_credentials(account)
            person.create_api_credentials(account)

        url = get_url(self.context, self.request)
        self.request.response.redirect(url)

    @Lazy
    def roles_by_account(self):
        """
        Return a dictionary, with accounts as its keys and lists or roles as
        its values.
        """
        result = {}
        for account in self.accounts:
            management = AccessControlManagement(account)
            roles = management.get_persons_roles().get(self.context, [])
            result[account] = sorted(role.name for role in roles)
        return result


class PersonPage(SidebarPage, BasePersonPageMixin):

    title = "Settings"


class EditPersonPage(SidebarPage, BasePersonPageMixin):

    title = "Edit settings"


class EditPersonForm(LandscapeFormProvider):

    summary = "Personal details"

    description = "Please provide the following information."

    dropdown_css_class = "dropdown-spacer-left"

    def __init__(self, *args, **kwargs):
        super(EditPersonForm, self).__init__(*args, **kwargs)
        self.has_pam = pam.has_service("landscape")
        # We only include password-related fields if the person is still using
        # a password instead of an identity URL.
        selected = ["name", "timezone", "email"]
        fields = Fields(IPerson, render_context=True).select(
            "name", "timezone", "password"
        )
        fields["timezone"].custom_widget = DropdownWidget
        fields += Fields(
            EmailAddress(
                __name__="confirm_email",
                title="Confirm email",
                min_length=5,
                max_length=200,
                required=False,
            ),
            render_context=False,
        )

        # Only activate drop down email field for ubuntu sso login
        if self.is_ubuntu_sso():

            # Create email drop down for Ubuntu SSO login
            description = (
                "Email address for Landscape notifications. "
                "To add an email, it must be in your Ubuntu SSO. "
                "Click on the link below the Save button. "
                "You may need to login again to see changes."
            )

            emails = self.get_sso_emails()
            terms = [
                SimpleTerm(emails[i], token=emails[i], title=emails[i])
                for i in range(len(emails))
            ]
            fields += Fields(
                Choice(
                    __name__="email",
                    title="Email address",
                    description=description,
                    source=SimpleVocabulary(terms),
                )
            )

        else:
            selected.append("confirm_email")  # Confirm email when not sso
            fields += Fields(IPerson, render_context=True).select("email")

        if self.context.identity is not None:
            fields.omit("password")
        else:
            if self.has_pam:
                fields += Fields(
                    TextLine(
                        __name__="identity",
                        title="Identity",
                        description="PAM authentication " "identity.",
                        required=True,
                    ),
                    render_context=False,
                )
                selected += ["identity"]
            selected += ["password"]
            fields["password"].render_context = False
            fields["password"].field.required = False
            if not self.has_pam:
                fields += Fields(
                    Password(
                        __name__="confirm_password",
                        title="Confirm passphrase",
                        min_length=3,
                        max_length=50,
                        required=False,
                    ),
                    render_context=False,
                )
                selected += ["confirm_password"]
        self.fields = fields.select(*selected)

    def is_ubuntu_sso(self):
        """Returns True if ubuntu login is used on saas"""
        return self.context.identity and not self.view.is_standalone()

    def get_sso_emails(self):
        """
        Get a list of verified emails and the currently set email which may
        or may not be verified
        """
        session_data = ISession(self.request)["landscape.authentication"]
        extra_emails = session_data.get("extra_emails", [])
        sso_email = session_data.get("email")
        emails = [self.context.email, sso_email] + extra_emails
        emails = [e for e in emails if e]  # Remove Nones or empty values
        emails = sorted(set(emails))
        return emails

    def get_initial_data(self):
        """
        Prepare the form fields prior to display.
        """
        data = {
            "name": self.context.name,
            "email": self.context.email,
            "timezone": self.context.timezone,
            "confirm_email": self.context.email,
        }
        return data

    def validate(self, action, data):

        # Pass the email confirmation step since the field is not included
        if self.is_ubuntu_sso():
            data["confirm_email"] = data["email"]

        # If we've not caught any previous errors with our form data
        if (
            "form.email" not in self.widget_errors
            and "form.confirm_email" not in self.widget_errors
        ):
            if data["email"] != data["confirm_email"]:
                self.add_error(
                    "Email address and confirmation must match", "confirm_email"
                )
            if "form.confirm_email" not in self.widget_errors:
                person = get_person_by_email(data["email"])
                if person and (person != self.context):
                    self.add_error("Email address already in use", "email")

        if self.context.identity is None:
            if self.has_pam:
                if not pam.authenticate(
                    data["identity"], data["password"], "landscape"
                ):
                    self.add_error("Unable to validate credentials.", "identity")
            else:
                if (
                    "form.password" not in self.widget_errors
                    and "form.confirm_password" not in self.widget_errors
                ):
                    if data["password"] != Password.UNCHANGED_PASSWORD:
                        if data.get("password") != data.get("confirm_password"):
                            self.add_error("Passwords must match", "confirm_password")

    @action("Save", primary=True)
    def handle_edit_action(self, action, data):
        # Cleanup our data, we know that validate will have stopped any bad
        # data.
        del data["confirm_email"]
        if self.context.identity is None:
            if "confirm_password" in data:
                del data["confirm_password"]
            password = data["password"]
            del data["password"]

            if not self.has_pam:
                if password != Password.UNCHANGED_PASSWORD:
                    self.context.set_password(password)

                    event = PrincipalLoginEvent(
                        self.request.principal, self.request, False
                    )
                    notify(event)
            else:
                self.context.password = None

        else:
            password = Password.UNCHANGED_PASSWORD

        if (
            self.apply_changes(self.context, data)
            or password != Password.UNCHANGED_PASSWORD
        ):
            self.view.notes.info("Settings changed")
        else:
            self.view.notes.info("No changes made")

        self.request.response.redirect(get_url(self.context, self.request))

    @action("Test EMail", name="test-email")
    def handle_test_email(self, action, data):
        email = data["email"]
        name = data["name"]
        self.context.send_test_email(name, email)
        self.view.notes.info("Test email sent to %s at the address: %s" % (name, email))


class EditSalesforceUserForm(LandscapeFormProvider):

    prefix = "salesforce_form"

    summary = "Salesforce information"

    description = "Salesforce keys for the person."

    def __init__(self, context, request, view):
        super(EditSalesforceUserForm, self).__init__(context, request, view)
        fields = []
        self._defaults = {}
        self.accounts = list(self.context.accounts)
        for account in self.accounts:
            salesforce_user = account.get_salesforce_user(self.context)
            field = TextLine(
                __name__="salesforce_user_%d" % account.id,
                title="Salesforce key for %s" % account.title,
                required=False,
            )
            if salesforce_user is not None:
                self._defaults[field.__name__] = salesforce_user.salesforce_key
                field.default = salesforce_user.salesforce_key
            fields.append(field)

        self.fields = Fields(*fields)

    def validate(self, action, data):
        for name, salesforce_key in list(data.items()):
            if salesforce_key:
                if self._defaults.get(name) != salesforce_key:
                    if salesforce_key in list(self._defaults.values()):
                        for name, default_key in list(self._defaults.items()):
                            if default_key == salesforce_key and data[name] is None:
                                validate_salesforce_key(
                                    self, name, salesforce_key, check_user=False
                                )
                                break
                        else:
                            validate_salesforce_key(self, name, salesforce_key)
                    else:
                        validate_salesforce_key(self, name, salesforce_key)
                if list(data.values()).count(salesforce_key) > 1:
                    self.add_error("Salesforce keys must be unique", name)

    @action("Save", primary=True)
    def save(self, action, data):
        modified = False
        for account in self.accounts:
            salesforce_user = account.get_salesforce_user(self.context)
            key = data["salesforce_user_%d" % account.id]
            if salesforce_user is None:
                if key is not None:
                    # Adding a key
                    account.create_salesforce_user(self.context, key)
                    self.view.notes.info("Key added for %s." % account.name)
                    modified = True
            elif key is None:
                # Deleting a key
                salesforce_user.remove()
                self.view.notes.info("Key deleted for %s." % account.name)
                modified = True
            elif salesforce_user.salesforce_key != key:
                # Changing a key
                salesforce_user.salesforce_key = key
                self.view.notes.info("Key modified for %s." % account.name)
                modified = True
        if not modified:
            self.view.notes.info("No changes made.")
        self.request.response.redirect(get_url(self.context, self.request))


class OpenIDMigration(LandscapePage):

    title = "Authentication migration"
