# Copyright 2005-2022 Canonical Limited. All rights reserved.

from json import loads

from canonical.content.decorator import json

from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.pager import PaginationMixin


class PackageProfileComputerPage(AccountPage, PaginationMixin):

    page_size = 25

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        self.title = "Package profile '%s' Computers" % context.name
        self._computer_ids = None
        self.broken_only = False
        if self.request.form.get("filter", "") == "non-compliant":
            self.broken_only = True
        # display only computers the person has access to.
        selection = ComputerSelection(
            self.context.account, self.request.principal.person
        )
        self._accessible_computer_ids = selection.get_available_ids()

    def show_broken_computers_only(self):
        """
        Return the boolean attribute flag on whether only broken computers
        are to be shown in form suitable for a javascript variable.
        """
        return str(self.broken_only).lower()

    def update(self):
        """
        (Re) Apply the package profile to the specified computers.

        This method is URL Traversable.
        """
        self._check_request_csrf_token(require_post=True)
        computer_ids = self.request.form.get("selection_json")
        if not computer_ids:
            return self.request.response.redirect(".")
        computer_ids = loads(computer_ids)
        self.context.apply(computer_ids)
        # setup a status message for the user
        note = "Profile reapplied to %d " % (len(computer_ids))
        if len(computer_ids) > 1:
            note += "computers"
        else:
            note += "computer"
        self.notes.info(note)
        self.request.response.redirect(".")

    def get_broken_computer_ids(self):
        """
        Returns a list of the available computer ids. Only returns value if
        the view is showing only non-compliant computers.
        """
        if not self._computer_ids:
            if self.broken_only:
                self._computer_ids = self.context.get_broken_computer_ids(
                    computer_ids=self._accessible_computer_ids
                )
            else:
                self._computer_ids = []
        return list(self._computer_ids)

    @json
    def json(self):
        """
        Returns paginated computers batches as JSON.
        """
        results = []
        # date formatter
        format_date = self.request.principal.person.formatter.format_date
        info = self.get_sort_variables()

        if self.broken_only:
            result = self.context.get_broken_computers(
                computer_ids=self._accessible_computer_ids
            )
        else:
            result = self.context.get_computers(
                computer_ids=self._accessible_computer_ids
            )

        result.order_by(Computer.title)
        result.config(limit=self.page_size, offset=info["offset"])
        for computer in result:
            if computer.last_ping_time:
                last_activity = format_date(computer.last_ping_time)
            else:
                last_activity = ""
            info = {
                "title": computer.title,
                "computer_id": computer.id,
                "last_activity": last_activity,
                "hostname": computer.hostname,
            }
            results.append(info)
        if self.broken_only:
            total = len(self.get_broken_computer_ids())
        else:
            computer_ids = set(self.context.get_computer_ids())
            computer_ids.intersection_update(self._accessible_computer_ids)
            total = len(computer_ids)
        self.request.response.setHeader("Content-Type", "application/json")
        return {
            "totalRecords": total,
            "records": results,
            "broken_only": self.broken_only,
        }
