# Copyright 2005-2023 Canonical Limited. All rights reserved.

import re

from zope.browserpage import ViewPageTemplateFile
from zope.schema import ValidationError, Choice
from zope.security.interfaces import Unauthorized
from zope.cachedescriptors.property import Lazy

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.model.package.interfaces import IPackageProfile, VERSION_RE

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider
from safe.canonical.landscape.model.main.computer import get_computer_by_id
from canonical.landscape.model.package.profile import PackageProfileError
from canonical.landscape.model.main.person import PersonAccessContextsVocabularyFactory


def form_constraint_helper(form):
    """
    Intermediate step of converting constraints from form into a list of dicts
    See more info about the 'record:' format here https://zope.readthedocs.io/en/latest/zdgbook/ObjectPublishing.html
    """
    constraints = []
    constraint_num = 0
    for key, value in form.items():
        if key.startswith("constraint"):
            for item in value:
                item.idx = str(constraint_num)  # Constraints are numbered in the HTML
                constraint_num += 1
            constraints.extend(value)  # Value is a list with one element
    return constraints


def extract_constraints(form):
    """
    Extract constraints from a form.

    @return: a tuple of (package_constraints, package_error).
    """
    package_constraints = set()
    package_error = ""
    name_field = IPackageProfile.get("name")

    constraints = form_constraint_helper(form)

    for constraint in constraints:
        if not constraint.package:
            if not constraint.operator:
                continue
            package_error = "Package name can't be empty. "
        else:
            try:
                name_field.validate(constraint.package)
            except ValidationError:
                package_error = "Invalid format for package name. "

        if constraint.operator and not constraint.version:
            package_error += "If an operator is selected a version must be provided"
        elif not constraint.operator:
            # Force empty string for sanity, especially with tests
            constraint.version = ""

        if constraint.version and not VERSION_RE.match(constraint.version):
            package_error += "Invalid format for version."

        if package_error:
            break

        package_constraints.add(
            (
                constraint.relation,
                constraint.package,
                constraint.operator,
                constraint.version,
            )
        )

    if not package_error and not package_constraints:
        # This is untested, but this is a safety check
        package_error = "At least one package must be specified."
    return package_constraints, package_error


class CreatePackageProfilePage(AccountPage):

    title = "Create package profile"

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        self._computer_ids = None
        if not self.context.person_can("CreatePackageProfile"):
            raise Unauthorized()

    def get_computer_ids_with_installed_packages(self):
        """
        Returns a list of the available computer ids that we can use to create
        a package profile. Only returns ids of computers that have installed
        packages listed.
        """
        if not self._computer_ids:
            self._computer_ids = self.context.get_computer_ids_with_packages()
        return list(self._computer_ids)


class CreatePackageProfileForm(LandscapeFormProvider):

    additional_content = ViewPageTemplateFile("create-fields.pt")
    package_error = ""

    dropdown_css_class = "dropdown-spacer-package-profile"

    max_files = 1
    attachment_help_text = (
        "The attachment provided should be formatted as either a package "
        'profile CSV file or the output of running "dpkg --get-selections" '
        "on the computer with desired packages installed. That attachment "
        "will be used to fill in the package contraints for this created "
        "profile."
    )

    empty_constraint = {
        "relation": "depends",
        "package": "",
        "operator": "",
        "version": "",
        "idx": "0",  # This is used to number the constraints in the html
    }

    def __init__(self, context, request, view):
        super(CreatePackageProfileForm, self).__init__(context, request, view)

        self.is_create = True  # report to zope we are the create action
        self.computer = None
        self.file_name = None
        self.file_data = None
        self.constraints = [self.empty_constraint]

        if request.method == "POST":
            computer_id = request.form.get("computer-id")
            self.file_name = request.form.get("file-name")
            self.file_data = self.request.form.get("file-data")
            file_name, file_data = self._get_attachment()
            if computer_id:
                self.computer = get_computer_by_id(int(computer_id))
                if not self.computer:
                    self.add_error(
                        "Could not retrieve computer by id: %s" % computer_id
                    )
                else:
                    self.profile_description = (
                        "Profile created from packages installed on the "
                        "computer %s." % self.computer.title
                    )

            elif file_name:
                self.file_name, self.file_data = file_name, file_data
                if self.file_data is not None:
                    # protect newlines in case the form is resubmitted
                    self.file_data = self.file_data.replace("\n", "\\n").replace(
                        "\r", "\\r"
                    )
            else:
                constraints = form_constraint_helper(request.form)
                if constraints:
                    self.constraints = constraints

    @Lazy
    def fields(self):
        access_context = Choice(
            __name__="access_context",
            title="Access group",
            source=PersonAccessContextsVocabularyFactory("CreatePackageProfile"),
            default="Global",
            description="The access group the profile should belong to.",
            required=True,
        )
        fields = Fields(IPackageProfile).omit("access_context", "name")
        fields += Fields(access_context)
        return fields.hide("creation_time", "modification_time", "version", "package")

    @Lazy
    def summary(self):
        if self.computer is None:
            return "Create package profile"

        return "Create package profile from computer: %s" % self.computer.title

    @Lazy
    def name(self):
        if self.computer is None:
            return None

        name = self.computer.title.replace(" ", "-").lower()
        name = re.sub("[^a-z0-9-]", "", name)
        match = re.search("^([a-z][\w-]+)", name)
        if match:
            name = "computer-%s-profile" % match.group()
        else:
            name = "computer-%s-profile" % self.computer.id
        return name

    @Lazy
    def max_attachment_size(self):
        format_bytes = self.request.principal.person.formatter.format_bytes
        return format_bytes(self.context.account.max_attachment_size)

    def _get_attachment(self):
        """
        Return a tuple with file name and UTF-8 content for the attachment in
        the form.  Both will be C{None} if an attachment is not present.  File
        content can be C{None} if it can't be decoded as UTF-8.
        """
        max_attachment_size = self.context.account.max_attachment_size
        self.max_files = self.context.account.max_attachment_files

        attachment = self.request.form.get("form.attachment")
        if not attachment:
            return None, None

        # Let's try not to read data uselessly if the attachment is too big
        attachment.seek(0, 2)
        total_length = attachment.tell()

        if total_length > max_attachment_size:
            format_bytes = self.request.principal.person.formatter.format_bytes
            self.package_error = "Attachment size is too large: %s (maximum is %s)" % (
                format_bytes(total_length),
                format_bytes(max_attachment_size),
            )
            return None, None

        attachment.seek(0)

        try:
            attachment_data = attachment.read().decode()
        except UnicodeDecodeError:
            attachment_data = None

        return attachment.filename, attachment_data

    def get_initial_data(self):
        initial_data = {"access_context": self.context.account.access_context.name}
        # Did we already choose to create from computer?
        computer_id = None
        computer_name = None
        if self.computer:
            computer_id = self.computer.id
            computer_name = self.computer.title
        if self.name:
            initial_data.update(
                {
                    "name": self.name,
                    "description": self.profile_description,
                    "computer-id": computer_id,
                    "computer-name": computer_name,
                }
            )
        return initial_data

    def validate(self, action, data):
        if self.file_name and not self.file_data:
            self.add_error("Attachment is not dpkg --get-selections or CSV data.")

    def form_save_helper(self, action, data):
        """Used for testing because we cannot use decorator in test case"""

        # computer-id is set the computer.id if we have selected to create our
        # package profile from an existing computer
        computer_id = self.request.form.get("computer-id")
        computer = None
        package_constraints = []

        if computer_id:
            computer = get_computer_by_id(int(computer_id))
            if not computer:
                self.package_error(
                    "Could not retrieve computer by id: %s."
                    "<br />Please select another computer." % computer_id
                )
        elif not self.file_data:
            package_constraints, self.package_error = extract_constraints(
                self.request.form
            )

        # Exit if we errored above
        if self.package_error:
            return

        if self.file_data:
            # Revert any form data modification done
            self.file_data = self.file_data.replace("\\n", "\n").replace("\\r", "\r")

        try:
            profile = self.context.create_package_profile(
                data["title"],
                data["description"],
                material=self.file_data,
                computer=computer,
                access_context_name=data["access_context"],
            )
        except PackageProfileError as error:
            self.package_error = "%s" % str(error)
            return

        profile.modify_package_constraints(add_constraints=package_constraints)
        profile.on_create(self.context.person)
        self.request.response.redirect(get_url(profile, self.request))

    @action("Save", primary=True)
    def save(self, action, data):
        self.form_save_helper(action, data)
