# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.security.management import checkPermission

from canonical.landscape.model.activity.status import (
    BLOCKED,
    QUEUED,
    SCHEDULED,
    INPROGRESS,
)
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import TagAssociationHelper
from canonical.landscape.ui.action import ActionItem
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)


OPERATOR = {
    "<": "<",
    "<=": "\N{LESS-THAN OR EQUAL TO}",
    "=": "=",
    ">=": "\N{GREATER-THAN OR EQUAL TO}",
    ">": ">",
}


PENDING_ITEM = ActionItem(
    icon="/static/activity/undelivered-16x16.png",
    summary="%(count)s computers are %(action)s.",
    summary_single="%(count)s computer is %(action)s.",
    action="applying the profile",
    action_url="activities",
)

NO_PENDING_ITEM = ActionItem(
    icon="/static/activity/succeeded-16x16.png",
    summary="The profile is not pending on any computer.",
    summary_single="",
    action="",
    action_url=None,
)

BLOCKED_ITEM = ActionItem(
    icon="/static/activity/blocked-16x16.png",
    summary="%(count)s computers are %(action)s.",
    summary_single="%(count)s computer is %(action)s.",
    action="waiting for profile activity approval",
    action_url="activities?filter=unapproved",
)

NO_BLOCKED_ITEM = ActionItem(
    icon="/static/activity/succeeded-16x16.png",
    summary="The profile has no unapproved activity on any computer.",
    summary_single="",
    action="",
    action_url=None,
)

ASSOCIATED_COMPUTER_ITEM = ActionItem(
    icon="/static/computer/pending-computers-16x16.png",
    summary="%(count)s computers are %(action)s.",
    summary_single="%(count)s computer is %(action)s.",
    action="associated with this profile",
    action_url="computers",
)

NO_ASSOCIATED_COMPUTER_ITEM = ActionItem(
    icon="/static/computer/computer-16x16.png",
    summary="The profile is not associated to any computer.",
    summary_single="",
    action="",
    action_url=None,
)

BROKEN_COMPUTER_ITEM = ActionItem(
    icon="/static/computer/pending-computers-alert-active-16x16.png",
    summary="%(count)s computers are %(action)s.",
    summary_single="1 computer is %(action)s.",
    action="not compliant with the profile",
    action_url="computers?filter=non-compliant",
)

NO_BROKEN_COMPUTER_ITEM = ActionItem(
    icon="/static/computer/computers-16x16.png",
    summary=("All associated computers are in " "compliance or applying the profile."),
    summary_single="The associated computer is in compliance",
    action="computers",
    action_url=None,
)


class PackageProfileDashboardPage(AccountPage, TagAssociationHelper):

    update_requires = "landscape.EditPackageProfile"

    def __init__(self, context, request):
        AccountPage.__init__(self, context, request)
        TagAssociationHelper.__init__(self, context, request)
        self.title = "Package profile '%s'" % context.name
        self.depends = []
        self.conflicts = []
        self.can_edit = checkPermission("landscape.EditPackageProfile", context)
        # We should be allowed to copy a View-only profile if we have Create
        # permissions in any of our access groups/roles.
        copy_access_ids = get_access_context_ids_for_person(
            self.context.account.access_context,
            "CreatePackageProfile",
            self.request.principal.person,
        )
        self.can_copy = len(copy_access_ids) > 0

        for constraint in context.constraints:
            kind = constraint.parse()[0]
            if kind == "conflicts":
                self.conflicts.append(constraint)
            else:
                self.depends.append(constraint)
        self.access_group_text = self.context.access_context.title
        self._initialize_summary()
        if self.request.method == "POST":
            self.update("profile")

    def _initialize_summary(self):
        # XXX in the UI "pending" currently means all non-final statuses,
        # *except* BLOCKED (which is presented with the legacy "unapproved"
        # term).
        result = self.context.get_pending_activities(
            statuses=[QUEUED, SCHEDULED, INPROGRESS]
        )
        pending_computers_count = result.count()
        if pending_computers_count:
            pending_item = PENDING_ITEM
        else:
            pending_item = NO_PENDING_ITEM
        self.pending_summary = pending_item.get_summary(None, pending_computers_count)
        self.pending_icon = pending_item.icon

        result = self.context.get_pending_activities(statuses=[BLOCKED])
        blocked_computers_count = result.count()
        if blocked_computers_count:
            blocked_item = BLOCKED_ITEM
        else:
            blocked_item = NO_BLOCKED_ITEM
        self.blocked_summary = blocked_item.get_summary(None, blocked_computers_count)
        self.blocked_icon = blocked_item.icon

        broken_computer_ids = self.context.get_broken_computer_ids()
        if broken_computer_ids:
            broken_item = BROKEN_COMPUTER_ITEM
        else:
            broken_item = NO_BROKEN_COMPUTER_ITEM
        self.broken_summary = broken_item.get_summary(None, len(broken_computer_ids))
        self.broken_icon = broken_item.icon

        computer_count = self.context.get_computer_count()
        if computer_count:
            associated_item = ASSOCIATED_COMPUTER_ITEM
        else:
            associated_item = NO_ASSOCIATED_COMPUTER_ITEM
        self.associated_icon = associated_item.icon
        self.associated_summary = associated_item.get_summary(None, computer_count)

    def get_constraint_data(self, constraint):
        """Return a C{dict} holding information about the given C{constraint}.

        The avaialable keys are 'name' (the package name) and
        'version' (a string representation of the operator/version
        pair).
        """
        _, package, operator, version = constraint.parse()

        data = {"name": package}
        if operator:
            data["version"] = "%s %s" % (OPERATOR[operator], version)
        else:
            data["version"] = "any"
        return data
