# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.cachedescriptors.property import Lazy

from canonical.url import get_url
from canonical.content.decorator import json

from canonical.landscape.ui.account.page import AccountPage
from canonical.content.download import set_csv_download


class PackageProfileListingPage(AccountPage):

    title = "Package profiles"

    def __init__(self, context, request):
        super(PackageProfileListingPage, self).__init__(context, request)
        computer_filter = request.get("filter")
        self.no_filter = computer_filter is None
        self.only_non_compliant = computer_filter == "non-compliant"
        self._non_compliant_counts_by_profile = {}
        self.package_profiles = []
        self.can_create = self.context.person_can("CreatePackageProfile")
        self.can_remove = self.context.person_can("EditPackageProfile")

        for profile in context.get_package_profiles():
            count = len(profile.get_broken_computer_ids())
            # Filter out profiles without non-compliant computers
            if count == 0 and self.only_non_compliant:
                continue
            self._non_compliant_counts_by_profile[profile] = count
            self.package_profiles.append(profile)

    @Lazy
    @json
    def buttons(self):
        """Return a list of actionable form buttons."""
        buttons = ["#download_csv"]
        if self.can_remove:
            buttons.append("#remove_button")
        return buttons

    def update(self):
        """
        Method is called by Zope's traversal machinery when a profile removal
        or CSV download is requested.  This method should only be called when
        form data has been POST'd to update the selection.  It always performs
        a redirect.
        """
        self._check_request_csrf_token(require_post=True)
        selected_profiles = [
            profile
            for profile in self.package_profiles
            if profile.id in self.request.form.get("profile_id", ())
        ]

        if selected_profiles:
            if "download_csv" in self.request:
                csv_data = []
                for profile in selected_profiles:
                    if not csv_data:
                        basename = "packageprofile_%s" % profile.name
                        csv_data.append(profile.get_csv_data())
                    else:
                        # Delete header row from subsequent CSV content
                        csv_data.append(profile.get_csv_data(include_header=False))
                        basename = "packageprofiles"

                set_csv_download(self.request.response, basename)

                return "".join(csv_data)

            if "remove" in self.request:
                not_removed = self.context.remove_package_profiles(selected_profiles)
                if not_removed:
                    names = ", ".join(profile.name for profile in not_removed)
                    self.notes.warning(
                        "Unable to remove package profiles: %s. Unauthorized" % names
                    )

        self.request.response.redirect(get_url(self.context, self.request))

    def get_enabled_for_text(self, profile):
        """
        Generate a message reporting the tags this profile is associated with,
        and hence the computers it is enabled for.
        """
        access_group_message = " in the <a href='%s'>%s</a> access group" % (
            get_url(profile.access_context, self.request),
            profile.access_context.title,
        )

        if profile.all_computers:
            message = "All computers"
        else:
            tags = sorted([tag.name for tag in profile.get_tags()])
            if not tags:
                message = "No computers"
            else:
                message = "Computers with %s %s" % (
                    ", ".join(tags),
                    "tag" if len(tags) == 1 else "tags",
                )
        return message + access_group_message

    def _get_computer_count_text(self, count):
        """Generate a message according to the given computer C{count}."""
        if count == 0:
            return None
        elif count == 1:
            return "1 computer"
        else:
            return "%d computers" % count

    def get_pending_on_text(self, profile):
        """
        Generate a message reporting how many computers are pending for
        this profile.
        """
        # The same query is internally performed by the get_broken_computer_ids
        # method call in the constructor. We could avoid it, see Bug #535597.
        count = profile.get_pending_activities().count()
        return self._get_computer_count_text(count)

    def get_non_compliant_on_text(self, profile):
        """
        Generate a message reporting how many computers are not compliant with
        this profile.
        """
        count = self._non_compliant_counts_by_profile[profile]
        return self._get_computer_count_text(count)
