# Copyright 2005-2022 Canonical Limited. All rights reserved.

import json
from datetime import datetime

from pytz import timezone

from zope.security.proxy import removeSecurityProxy
from zope.publisher.browser import BrowserView
from zope.cachedescriptors.property import Lazy

from canonical.lib.text import format_float

from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.model.main.standalone import LDSScriptExecutionManagement
from canonical.landscape.ui.toolbar.page import ToolbarPage


class ComputerLinkView(BrowserView):
    def __call__(self):
        if not self.context.count:
            return "No computers %s." % self.context.plural_text
        text = self.context.plural_text
        computer = "computers"
        if self.context.count == 1:
            text = self.context.single_text
            computer = "computer"
        if text:
            text = " %s." % text
        link_markup = '<a href="#" id="select-computers-%s">%d %s</a> ' % (
            self.context.name,
            self.context.count,
            computer,
        )
        if self.context.percent:
            link_markup += "(%s)" % self.context.percent
        link_markup += "%s" % text
        return link_markup


class ComputerLink(object):
    def __init__(self, name, count, percent, plural_text, single_text):
        self.name = name
        self.count = count
        self.percent = percent
        self.plural_text = plural_text
        self.single_text = single_text


class ReportingDashboardPage(ToolbarPage):

    title = "Reports"

    def __init__(self, context, request):
        super(ReportingDashboardPage, self).__init__(context, request)

        # The search object to act on the computer selection.
        self.search = self.context.get_package_search()

        # The time boundaries to report on
        self.fixed_in_days = [2, 14, 30]
        self.pending_in_days = 60

        # Counts
        self.computer_ids = self.context.selection.get_selected_computer_ids()
        self.computers_count = len(self.computer_ids)
        self.security_upgrades = list(self.search.get_computer_ids_with_usn_upgrades())
        self.security_upgrades_count = len(self.security_upgrades)
        self.no_security_upgrades_count = (
            self.computers_count - self.security_upgrades_count
        )
        self.not_reporting = list(self.context.get_not_pinging_computers())
        self.not_reporting_count = len(self.not_reporting)
        self.reporting_count = self.computers_count - self.not_reporting_count

        # Upgrade profiles
        self.no_upgrade = self.context.get_computers_not_upgraded()
        self.upgrade = list(set(self.computer_ids) - set(self.no_upgrade))
        self.no_upgrade_count = len(self.no_upgrade)
        self.upgrade_count = self.computers_count - self.no_upgrade_count
        data = {
            "computers-count": self.computers_count,
            "title-security-upgrades": "Computers with available security upgrades",
            "security-upgrades": self.security_upgrades,
            "title-no-security-upgrades": "Computers securely patched",
            "no-security-upgrades": list(
                set(self.computer_ids) - set(self.security_upgrades)
            ),
            "title-not-reporting": "Computers which have not reported to the server within the "
            "last 5 minutes",
            "not-reporting": self.not_reporting,
            "title-reporting": "Computers which have contacted the server within the "
            "last 5 minutes",
            "reporting": list(set(self.computer_ids) - set(self.not_reporting)),
            "title-no-upgrade": "Computers which are not covered by upgrade profiles",
            "no-upgrade": list(self.no_upgrade),
            "title-upgrade": "Computers covered by upgrade profiles",
            "upgrade": list(set(self.computer_ids) - set(self.no_upgrade)),
            "usn-fixed-in": self.fixed_in_days,
            "usn-pending-in": self.pending_in_days,
        }

        self.upgrades_frequency = self.context.get_upgraded_computers_by_frequency()

        # USN breakdown
        self.usn_breakdown = removeSecurityProxy(
            self.context.get_usn_time_to_fix(
                fixed_in_days=self.fixed_in_days, pending_in_days=self.pending_in_days
            )
        )

        for fixed_in, computer_ids in list(self.usn_breakdown["fixed_in"].items()):

            data["usn-fixed-in-%d" % fixed_in] = list(computer_ids)
            data["usn-not-fixed-in-%d" % fixed_in] = list(
                set(self.computer_ids) - computer_ids
            )
            data["title-usn-fixed-in-%d" % fixed_in] = (
                "Computers which have applied USNs in %d days or less" % fixed_in
            )
            data["title-usn-not-fixed-in-%d" % fixed_in] = (
                "Computers which have not applied USNs in %d days or less" % fixed_in
            )
            self.usn_breakdown["usn-link-fixed-in-%d" % fixed_in] = ComputerLinkView(
                ComputerLink(
                    "usn-fixed-in-%d" % fixed_in,
                    len(computer_ids),
                    self.display_percent(len(computer_ids)),
                    "have applied USNs in %d days or less" % fixed_in,
                    "has applied USNs in %d days or less" % fixed_in,
                ),
                self.request,
            )()
            not_fixed = self.computers_count - len(computer_ids)
            self.usn_breakdown[
                "usn-link-not-fixed-in-%d" % fixed_in
            ] = ComputerLinkView(
                ComputerLink(
                    "usn-not-fixed-in-%d" % fixed_in,
                    not_fixed,
                    self.display_percent(not_fixed),
                    "have not applied USNs in %d days or less" % fixed_in,
                    "has not applied USNs in %d days or less" % fixed_in,
                ),
                self.request,
            )()

        data["usn-pending"] = list(self.usn_breakdown["pending"])
        data["title-usn-pending"] = (
            "Computers with outstanding USNs older than %d days" % self.pending_in_days
        )
        data["usn-not-pending"] = list(
            set(self.computer_ids) - self.usn_breakdown["pending"]
        )
        data["title-usn-not-pending"] = (
            "Computers with no outstanding USNs older than %d days"
            % self.pending_in_days
        )
        self.usn_breakdown["pending-link"] = ComputerLinkView(
            ComputerLink(
                "usn-pending",
                len(self.usn_breakdown["pending"]),
                self.display_percent(len(self.usn_breakdown["pending"])),
                "have outstanding USNs older than %d days" % self.pending_in_days,
                "has outstanding USNs older than %d days" % self.pending_in_days,
            ),
            self.request,
        )()
        not_pending = self.computers_count - len(self.usn_breakdown["pending"])
        self.usn_breakdown["not-pending-link"] = ComputerLinkView(
            ComputerLink(
                "usn-not-pending",
                not_pending,
                self.display_percent(not_pending),
                "have patched all USNs older than %d days" % self.pending_in_days,
                "has patched all USNs older than %d days" % self.pending_in_days,
            ),
            self.request,
        )()

        self.computer_data = json.dumps(data).replace("</script", "<\\/script")

        self.security_upgrades_link = ComputerLink(
            "security-upgrades",
            self.security_upgrades_count,
            self.display_percent(self.security_upgrades_count),
            "are not yet patched",
            "is not yet patched",
        )

        self.no_security_upgrades_link = ComputerLink(
            "no-security-upgrades",
            self.no_security_upgrades_count,
            self.display_percent(self.no_security_upgrades_count),
            "are securely patched",
            "is securely patched",
        )

        self.reporting_link = ComputerLink(
            "reporting",
            self.reporting_count,
            self.display_percent(self.reporting_count),
            "have contacted the server within the last 5 minutes",
            "has contacted the server within the last 5 minutes",
        )

        self.not_reporting_link = ComputerLink(
            "not-reporting",
            self.not_reporting_count,
            self.display_percent(self.not_reporting_count),
            "have not contacted the server within the last 5 minutes",
            "has not contacted the server within the last 5 minutes",
        )

        self.upgrade_link = ComputerLink(
            "upgrade",
            self.upgrade_count,
            self.display_percent(self.upgrade_count),
            "are covered by upgrade profiles",
            "is covered by upgrade profiles",
        )

        self.no_upgrade_link = ComputerLink(
            "no-upgrade",
            self.no_upgrade_count,
            self.display_percent(self.no_upgrade_count),
            "are not covered by upgrade profiles",
            "is not covered by upgrade profiles",
        )

        if is_standalone():
            management = LDSScriptExecutionManagement()
            execution = management.get_execution("update_security_db.sh")
            if execution and execution.errored:
                self.notes.error(
                    "USN update cron script is not running, "
                    "USN information may be out of date."
                )

    def display_percent(self, part, total=None):
        """
        A small utility to format percentages for the view.
        """
        if total is None:
            total = len(self.computer_ids)
        result = 100 * (part / total)
        return "%s%%" % format_float(result)

    @Lazy
    def date_string(self):
        """
        Returns information about the time the report was generated. Includes
        time and user time zone.
        """
        tzinfo = timezone(self.request.principal.person.timezone)
        return datetime.now(tz=tzinfo).strftime("Generated at %H:%M on %d %B %Y %Z")
