# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os

from twisted.web.proxy import ReverseProxyResource
from twisted.web.resource import Resource, NoResource
from twisted.web.static import File
from twisted.web.util import Redirect, redirectTo

from canonical.talisker.twisted import TaliskerWSGIResource
from canonical.twisted.proxy import ReverseProxyResourceWithFinish
from canonical.twisted.http import set_caching_headers

from canonical.landscape.ui.constants import (
    STATIC_DIR,
    JAVASCRIPT_DIR,
    FAKE_CLOUD_IMAGES,
)
from canonical.landscape.ui.download import DownloadResource
from canonical.landscape.ui.wsgi import create_zope_app, create_combo_app
from canonical.landscape.environment import get_hash_id_databases_dir, get_raw_config


class CustomRedirect(Redirect):
    """This is so the http status code can be set to 301 on a redirect"""

    def render(self, request):
        redirect = redirectTo(self.url, request)
        request.setResponseCode(301)
        return redirect


class UI(Resource):
    """Root C{Resource} for the web interface.

    This resource forwards missing children to the actual C{WSGIResource},
    allowing to reverse-proxy some services and assets when running in
    development mode.
    """

    def __init__(self, config=None, reactor=None, threadpool=None):
        Resource.__init__(self)
        if config is None:
            config = {}
        self._create_zope_resource(config, reactor, threadpool)
        self._create_assets_resource(reactor, threadpool)
        self._create_proxy_resource("async-frontend", "ajax")
        self._create_repository_resource()
        if config.get("devmode"):
            self._create_devmode_resources()

    def getChild(self, name, request):
        """If it's not a child registered with C{putChild} then WSGI it is."""
        request.postpath[0:0] = request.prepath
        del request.prepath[:]
        return self._zope_resource

    def _create_zope_resource(self, config, reactor, threadpool):
        """Create the Zope WSGI resource."""
        self._zope_application = create_zope_app(config)
        self._zope_resource = TaliskerWSGIResource(
            reactor, threadpool, self._zope_application
        )

    def _create_assets_resource(self, reactor, threadpool):
        """Create the combo loader resource."""
        static_resource = File(STATIC_DIR)
        javascript_resource = File(JAVASCRIPT_DIR)
        combo_resource = TaliskerWSGIResource(reactor, threadpool, create_combo_app())
        hash_id_databases_resource = File(get_hash_id_databases_dir())
        download_resource = DownloadResource()

        class Assets(Resource):
            """A C{Resource} that returns the same resource for any path."""

            assets = Resource()
            assets.putChild(b"combo", combo_resource)
            assets.putChild(b"static", static_resource)
            assets.putChild(b"javascript", javascript_resource)

            def getChild(self, path, request):
                """Always return the C{assets}."""
                # Set caching for a very long time (10 years), because we want
                # assets to never exipire (we just change their URL with new
                # releases).
                set_caching_headers(request, days=365 * 10)
                return self.assets

        class StaticAsset(File):
            def getChild(self, path, request):
                url = request.uri.decode().strip("/")
                if url.startswith("static/doc/user-guide") or url.startswith(
                    "static/doc/api"
                ):
                    return CustomRedirect(b"https://docs.ubuntu.com/landscape/en")
                request.setHeader("X-Frame-Options", "SAMEORIGIN")
                return super(StaticAsset, self).getChild(path, request)

        # Allow only access to the API and help guide. Everything else
        # in the static directory should be accessed through the /r/
        # resource to allow for cache invalidation.
        docs_resource = Resource()
        doc_path = os.path.join(STATIC_DIR.encode(), b"doc")
        docs_resource.putChild(b"doc", StaticAsset(doc_path))

        self.putChild(b"r", Assets())
        self.putChild(b"combo", combo_resource)
        self.putChild(b"hash-id-databases", hash_id_databases_resource)
        self.putChild(b"static", docs_resource)
        self.putChild(b"download", download_resource)

    def _create_repository_resource(self):
        """Create a static resource for serving Landscape-managed repositories.

        Additional security will ensure that only URLs under dists/ or pool/
        will be allowed.
        """
        config = get_raw_config()
        if not config.has_option("landscape", "repository-path"):
            return
        repository_path = config.get("landscape", "repository-path").encode()

        class Repository(Resource):
            """A C{Resource} that returns the same resource for any path."""

            def getChild(self, path, request):
                """Always return the C{assets}."""
                if len(request.prepath) <= 3 and len(request.postpath) > 0:
                    return self
                if request.prepath[-1] in (b"dists", b"pool"):
                    path = os.path.join(repository_path, *request.prepath[1:])
                    return File(path)
                return NoResource("No such resource.")

        repository_resource = Repository()
        self.putChild(b"repository", repository_resource)

    def _create_proxy_resource(self, service, endpoint):
        """Create a proxy resource for the given service at the given endpoint.

        @param service: Name of the service to proxy, e.g. 'async-frontend'.
        @param endpoint: Endpoint the service will be exposed at (e.g. 'ajax').
        """
        config = get_raw_config()
        port = int(config.get(service, "base-port"))
        resource = ReverseProxyResource("localhost", port, b"")
        self.putChild(endpoint.encode(), resource)

    def _create_devmode_resources(self):

        # Register a proxy for the pppa-proxy (oh, my!)
        pppa_proxy_port = int(os.environ.get("LANDSCAPE_PPPA_PROXY_PORT", 9093))
        pppa_proxy_resource = ReverseProxyResourceWithFinish(
            "localhost", pppa_proxy_port, b""
        )

        # Register a proxy for the message-server
        message_server_proxy_port = 8089
        message_server_proxy_resource = ReverseProxyResourceWithFinish(
            "localhost", message_server_proxy_port, b"/"
        )
        attachment_proxy_resource = ReverseProxyResourceWithFinish(
            "localhost", message_server_proxy_port, b"/attachment"
        )

        cloud_images_resource = File(FAKE_CLOUD_IMAGES)

        self.putChild(b"archive", pppa_proxy_resource)
        self.putChild(b"message-system", message_server_proxy_resource)
        self.putChild(b"attachment", attachment_proxy_resource)
        self.putChild(b"cloud-images", cloud_images_resource)
