# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.cachedescriptors.property import Lazy
from zope.browserpage import ViewPageTemplateFile
from zope.security.checker import ProxyFactory

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider, MultiCheckBoxWidget
from canonical.landscape.ui.role.create import RolePage
from canonical.landscape.ui.role.interfaces import IRoleForm
from canonical.landscape.model.access.errors import DuplicateRoleError
from canonical.landscape.ui.accesscontext.summary import build_structure
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.landscape.model.access.permissions import ACCOUNT_MACRO_PERMISSIONS


class EditRolePage(RolePage):

    can_copy = True

    title = "Edit role"

    @Lazy
    def is_stored(self):
        permissions = {permission: False for permission in ACCOUNT_MACRO_PERMISSIONS}
        for permission in self.context.get_account_permissions():
            permissions[permission] = True
        return permissions


class EditRoleForm(LandscapeFormProvider):

    prefix = "role"

    render_groups = ViewPageTemplateFile("render-groups.pt")

    @Lazy
    def summary(self):
        return "Edit role '%s'" % self.context.name

    @Lazy
    def access_groups(self):
        return build_structure(self.context)

    @Lazy
    def role_access_groups(self):
        return [
            access_context.name for access_context in self.context.get_access_contexts()
        ]

    @Lazy
    def fields(self):
        fields = Fields(IRoleForm)
        fields["permissions"].custom_widget = MultiCheckBoxWidget
        return fields

    def get_initial_data(self):
        return {
            "name": self.context.name,
            "permissions": self.context.get_permissions(),
            "account_permissions": self.context.get_account_permissions(),
        }

    @action("Save", primary=True)
    def save(self, action, data):
        try:
            access_contexts = data.get("access_contexts", [])
            context_permissions = data.get("permissions", set())
            account_permissions = data.get("account_permissions", set())
            permissions = context_permissions.union(account_permissions)

            # Wrap AccessControlManagement explicitely, to get permission
            # check, as it's in the clearlist but we still want security here.
            management = ProxyFactory(
                AccessControlManagement(
                    self.context.account, self.request.principal.person
                )
            )
            management.edit_role(
                self.context.name,
                permissions=permissions,
                access_contexts=access_contexts,
            )
        except DuplicateRoleError:
            self.add_error("Role name already taken.", "name")
            return

        message = "Role '%s' successfully saved." % self.context.name
        url = "%s/roles" % get_url(self.context.account, self.request)
        self.view.notes.info(message)
        self.request.response.redirect(url)

    @action("Delete")
    def remove(self, action, data):
        """
        Redirect to a confirmation form.
        """
        url = get_url(self.context, self.request) + "/confirm-remove"
        self.request.response.redirect(url)


class RemoveRolePage(AccountPage):
    """Confirmation page for removing stored roles."""

    title = "Remove role?"

    requires = "landscape.EditRole"

    @Lazy
    def permission_context(self):
        # The Person must be a full administrator.
        return self.context.account


class RemoveRoleForm(LandscapeFormProvider):

    summary = "Remove role?"

    @Lazy
    def description(self):
        return "Are you sure you want to remove the '%s' role?" % self.context.name

    @action("Confirm")
    def remove(self, action, data):
        """Remove role."""
        management = AccessControlManagement(
            self.context.account, self.request.principal.person
        )
        management.remove_role(self.context.name)
        self.view.notes.info("Role '%s' has been removed." % self.context.name)
        url = "%s/roles" % get_url(self.context.account, self.request)
        self.request.response.redirect(url)

    @action("Cancel", primary=True)
    def cancel(self, action, data):
        """Redirect to the role page."""
        url = "%s/edit" % get_url(self.context, self.request)
        self.request.response.redirect(url)
