# Copyright 2005-2022 Canonical Limited. All rights reserved.

from json import load as json_load
from prometheus_client import (
    CONTENT_TYPE_LATEST,
    REGISTRY,
    generate_latest,
)

from zope.component import getUtility
from zope.publisher.browser import BrowserView

from canonical.content.decorator import json

from canonical.landscape.prometheus.decorator import http_metrics
from canonical.landscape.prometheus.interfaces import IPrometheusMetrics
from canonical.landscape.model.main.standalone import get_available_releases
from canonical.landscape.ui.page.page import LandscapePage
from canonical.security.interfaces import IRoleRegistry


class LDSMetricsPage(BrowserView):
    def _parse_computer_metrics(self, lds_metrics, prometheus_metrics):
        """Parse computer-related metrics and update prometheus."""
        computers = lds_metrics.get("computers", {})
        prometheus_metric = prometheus_metrics.get("landscape_on_prem_computers")
        for computer_type in ("physical", "vm", "container"):
            if computer_type in computers:
                prometheus_metric.labels(
                    revision=self.http_metrics_revision,
                    type=computer_type,
                ).inc(computers.get(computer_type))

    def _parse_license_metrics(self, lds_metrics, prometheus_metrics):
        """Parse license-related metrics and update prometheus."""
        prometheus_metric = prometheus_metrics.get("landscape_on_prem_licenses")
        registry = getUtility(IRoleRegistry)
        # The anonymous metrics report contains license roles as strings
        # For example {"10001": 100} for a license w/ 100 physical seats.
        vm_key = "{}".format(registry.get("VMFeatures").key)
        container_key = "{}".format(registry.get("ContainerFeatures").key)
        physical_key = "{}".format(registry.get("BasicFeatures").key)
        license = lds_metrics["license"]
        computers = {
            "physical": license.get(physical_key, 0),
            "vm": license.get(vm_key, 0),
            "container": license.get(container_key, 0),
        }
        for computer_type in computers:
            if computer_type in computers:
                prometheus_metric.labels(
                    revision=self.http_metrics_revision,
                    type=computer_type,
                ).inc(computers.get(computer_type))

    def _parse_metrics(self, lds_metrics):
        """Parse Landscape metrics and update prometheus counters."""
        # http_metrics_revision is used by http_metrics decorator
        self.http_metrics_revision = lds_metrics.get("revision", "")
        metrics = getUtility(IPrometheusMetrics)
        self._parse_computer_metrics(lds_metrics, metrics)
        self._parse_license_metrics(lds_metrics, metrics)
        if lds_metrics.get("maas", 0) > 0:
            prometheus_metric = metrics.get("landscape_maas_servers")
            prometheus_metric.labels(revision=self.http_metrics_revision).inc()
        for cloud in lds_metrics.get("clouds", ()):
            for region in cloud["regions"]:
                prometheus_metric = metrics.get("landscape_cloud_regions")
                prometheus_metric.labels(
                    revision=self.http_metrics_revision,
                    computers=sum(region["zones"]),
                    compute=region["compute"],
                    object=region["object"],
                    block=region["block"],
                    zones=len(region["zones"]),
                ).inc()

    @json
    @http_metrics
    def __call__(self):
        """Track anonymous metrics posted by OnPrem installs."""
        self.request.response.setHeader("Content-Type", "application/json")
        status = 200
        response = {}
        if self.request.method != "POST":
            status = 405
            response = {"error": "POST request expected"}
        else:
            body = self.request.bodyStream
            try:
                self._parse_metrics(json_load(body))
            except ValueError as e:
                status = 400
                response = {"error": "Invalid request format, expected json."}
            except KeyError as e:
                status = 400
                response = {"error": "Invalid request: Missing key {}".format(e)}
        self.request.response.setStatus(status)
        return response


class LDSReleasesPage(BrowserView):
    @json
    @http_metrics
    def __call__(self):
        """Exposes all LDS releases in a JSON document."""
        self.request.response.setHeader("Content-Type", "application/json")
        self.request.response.setStatus(200)
        return get_available_releases()


class LDSPrometheusMetricsPage(BrowserView):
    def __call__(self):
        """Handler for publishing prometheus metrics."""
        self.request.response.setHeader("Content-Type", CONTENT_TYPE_LATEST)
        return generate_latest(REGISTRY)


class LDSNoLicensePage(LandscapePage):

    title = "No license available"
