# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import textwrap
from urllib.parse import urlencode

from apt.cache import Cache as APTCache

from zope.component import getUtility, queryUtility
from zope.security.proxy import removeSecurityProxy
from zope.session.interfaces import ISession
from zope.publisher.browser import BrowserPage
from zope.publisher.interfaces import NotFound

from canonical.content.form import Fields
from canonical.content.decorator import json
from canonical.pam import pam
from canonical.url import get_url

from canonical.landscape.application import PACKAGE_VERSION
from canonical.landscape.model.root.interfaces import ILogin
from canonical.landscape.security.interfaces import IAnonymousPrincipal
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.page.page import Page, LandscapePage
from canonical.landscape.ui.page.csp import make_csp_header, CSP_KIND_SCRIPT
from canonical.landscape.ui.lsopenid import ILandscapeOpenIDFactory, SSODownError
from canonical.landscape.model.configuration import (
    is_standalone,
    get_root_url,
    get_revision,
)
from canonical.landscape.model.main.license import get_license_file_paths
from canonical.landscape.model.feature.flag import BUILTIN_FEATURES, get_feature_flags
from canonical.landscape.model.main.account import (
    get_account_by_name,
    DEFAULT_STANDALONE_ACCOUNT_NAME,
)

from safe.canonical.landscape.model.account.invitation import get_invitation


STANDALONE_SETTINGS = "/account/standalone/standalone-settings"


class LoginPage(LandscapePage):

    title = "Login"


class AboutPage(LandscapePage):

    title = "About"

    version = PACKAGE_VERSION

    def revision(self):
        return get_revision()


class JSONAboutPage(BrowserPage):
    """Return information about Landscape in JSON format."""

    _packages = ("juju",)

    @json
    def __call__(self):
        packages = self._get_package_versions()
        # The landscape-server package version is built-in
        packages["landscape-server"] = PACKAGE_VERSION
        return {
            "revision": get_revision(),
            "packages": packages,
            "builtin-features": sorted(BUILTIN_FEATURES),
            "feature-flags": get_feature_flags(),
            "is-bootstrapped": self._is_bootstrapped(),
        }

    def _get_package_versions(self):
        """Return the versions of important packages."""
        with APTCache() as cache:
            return {
                package: cache[package].installed.version for package in self._packages
            }

    def _is_bootstrapped(self):
        """Return whether this is a bootstrapped OPL instance."""
        return bool(
            is_standalone() and get_account_by_name(DEFAULT_STANDALONE_ACCOUNT_NAME)
        )


class WellKnownSecurityPage(BrowserPage):
    def __call__(self):
        if is_standalone():
            raise NotFound(self, "Page not found")

        return textwrap.dedent(
            """\
            Contact: mailto:security@canonical.com.
            Expires: 2024-01-01T02:59:00.000Z
            Hiring: https://canonical.com/careers/all?search=Security
            """
        )


class AuthenticatePage(LandscapePage):

    title = "Enter passphrase"

    def __init__(self, context, request):
        super(AuthenticatePage, self).__init__(context, request)

        if "login.email" not in request:
            self._begin_openid(request)
        else:
            if IAnonymousPrincipal.providedBy(self.request.principal):
                self.request.response.redirect("/login")

    def _begin_openid(self, request):
        """
        Handle either openid or openid-connect, depending on the registered
        factory.
        """
        return_to = "login/handle-openid"
        params = {}
        next_url = self.request.get("next_url", "")
        invitation = self.request.get("invitation")
        create = self.request.get("create")

        # Ignore next_url if it's not a relative path to the application,
        # to prevent XSS attacks.
        next_url_is_local = next_url.startswith(get_root_url()) or next_url.startswith(
            "/"
        )
        if next_url_is_local:
            params["next_url"] = next_url.encode("utf-8")
        if invitation:
            params["invitation"] = invitation
        if create:
            params["create"] = create
        if params:
            # The urllib quoting isn't technically necessary, since it
            # seems python-openid will deal with improperly escaped URLs.
            # I'm just doing it to maintain my mental well-being.
            return_to += "?" + urlencode(params)
        factory = getUtility(ILandscapeOpenIDFactory)
        openid = factory.get_landscape_openid(request)
        try:
            openid.begin(return_to)
        except SSODownError:
            self.request.response.redirect("/")
            self.notes.error(
                "Couldn't reach authentication servers, please try again later"
            )


class LoginFormMixin(object):

    invitation = None

    def __init__(self, context, request):
        invitation_id = request.get("invitation")
        if invitation_id:
            self.invitation = get_invitation(invitation_id)

        self.is_standalone = is_standalone()
        self.has_pam_authentication = pam.has_service("landscape")
        self.standalone_login_prompt = (
            self.has_pam_authentication and "Login:" or "E-mail address:"
        )
        if self.is_standalone:
            self.use_openid = queryUtility(ILandscapeOpenIDFactory) is not None
            self.login_form_action = request.get("next_url", "/dashboard")
            if self.has_pam_authentication:
                self.new_user_link = "/new-pam-user"
            else:
                self.new_user_link = "/new-standalone-user"
        else:
            self.login_form_action = "/login/authenticate"

    def next_url(self):
        """
        Provide the next_url to go to after logging in.
        """
        next_url = self.request.get("next_url")

        if self.invitation:
            next_url = "/accept-invitation/%s" % (self.invitation.secure_id,)

        if self.is_standalone:
            if not get_root_url():
                next_url = STANDALONE_SETTINGS
        return next_url

    def invitation_email(self):
        if self.invitation:
            return self.invitation.email
        return None


class FrontPage(LoginFormMixin, Page):
    """This is the front page that users see when they access Landscape."""

    title = "Welcome!"
    layout = "fixed"
    csp = make_csp_header(
        {
            CSP_KIND_SCRIPT: (
                "https://ajax.googleapis.com/ajax/libs/",
                "assets.ubuntu.com",
            ),
        }
    )

    def __init__(self, context, request):
        LoginFormMixin.__init__(self, context, request)
        Page.__init__(self, context, request)

    def __call__(self):
        if self.request.method == "HEAD":
            return None
        if self.is_standalone:
            account = get_account_by_name("standalone")
            if account:
                count = removeSecurityProxy(account).people.count()
                self.lds_has_administrators = count > 0
            else:
                # If not, check if a license file is present
                self.lds_has_administrators = False
                license_paths = get_license_file_paths()
                if not any(os.path.isfile(path) for path in license_paths):
                    redirect_url = "/lds-no-license"
                elif self.has_pam_authentication:
                    redirect_url = "/new-pam-user"
                elif self.use_openid:
                    redirect_url = "/new-user"
                else:
                    redirect_url = "/new-standalone-user"
                self.redirect(redirect_url)

        if self.is_authenticated():
            default_url = "/dashboard"
            account = self.request.principal.person.get_default_account()
            if account is not None:
                default_url = get_url(account, self.request)
            next_url = self.request.get("next_url", default_url)
            self.redirect(next_url)
        else:
            self.errors = None
            session = ISession(self.request)
            errors = session["landscape.authentication.error"]
            if "authentication_error" in errors:
                self.errors = errors["authentication_error"]
                errors.clear()

            if not self.is_standalone:  # Only redirect on saas or selenium
                if self.is_selenium_environment() or not self.is_development_mode():
                    self.redirect("/login/authenticate")

            return super(FrontPage, self).__call__()


class FeaturesPage(FrontPage):

    title = "Landscape Features"


class TryLandscapePage(FrontPage):

    title = "Try Landscape"


class SetUpOnPremPage(FrontPage):

    title = "Set up on premises"


class LoginFormProvider(LoginFormMixin, LandscapeFormProvider):

    template_filename = "login-form.pt"
    prefix = "login"
    fields = Fields(ILogin)

    def __init__(self, context, request, view):
        LoginFormMixin.__init__(self, context, request)
        LandscapeFormProvider.__init__(self, context, request, view)


class FeatureFlagsPage(LandscapePage):

    title = "Features"

    def get_feature_flags(self):
        return get_feature_flags()
