# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Passphrase reset."""

from zope.traversing.browser.absoluteurl import absoluteURL
from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy

from canonical.url import get_url, get_url_object
from canonical.content.form import Fields, action

from canonical.landscape.model.main.schema import EmailAddress
from canonical.landscape.model.root.interfaces import ILandscapeRoot
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.page.page import LandscapePage
from canonical.landscape.ui.root.interfaces import IResetPassphrase
from safe.canonical.landscape.model.main.person import get_person_by_email


RESET_SUBJECT = "New passphrase for your Landscape account"
RESET_CONTENT = """\
In order to regain access to your Landscape account,
%(email)s
you must visit the following URL in your web browser to reset the
passphrase:

%(root)s/reset-passphrase/%(key)s

Please note this link is only valid for 24 hours.
"""


class ResetPassphrase(LandscapePage):

    title = "Reset your passphrase"
    layout = "fixed"


class ResetPassphraseForm(LandscapeFormProvider):

    summary = "Reset your passphrase"
    description = "Enter your e-mail address"
    fields = Fields(
        EmailAddress(
            __name__="email",
            title="E-mail address",
            description="Your previously-registered e-mail address",
            min_length=5,
            max_length=200,
            required=True,
        )
    )

    @action("Reset passphrase")
    def reset_passphrase(self, action, data):
        """Reset the passphrase associated with the provided email address.

        An email containing a link to reset the passphrase will be sent to the
        provided email address. In case the provided email address differs by
        case from the one stored in the database, the latter will be used.

        If the person associated with the provided email address uses
        Ubuntu SSO authentication, the client's browser will be redirected
        to L{ResetPassphraseOnUbuntuSSO).
        """
        email = data["email"]
        root = get_url_object(self.context, self.request, ILandscapeRoot)

        # UNPROXIED person object; anyone can try to reset a passphrase.
        person = get_person_by_email(email)
        if person and removeSecurityProxy(person).identity is not None:
            self.request.response.redirect(
                get_url(self.context, self.request) + "reset-passphrase-on-ubuntu-sso"
            )
            return

        self.view.notes.info(
            "An email with passphrase reset instructions "
            "has been sent to %s." % (email,)
        )
        if person is not None:
            reset = person.reset_password()
            # Use the email address stored in the database
            email = removeSecurityProxy(person).email
            body = RESET_CONTENT % {
                "email": email,
                "root": absoluteURL(root, self.request),
                "key": reset.secure_id,
            }
            getUtility(ILandscapeMailer).send([email], RESET_SUBJECT, body)
            self.request.response.redirect(get_url(self.context, self.request))
        else:
            # redirect to self
            self.request.response.redirect(
                get_url(self.context, self.request) + "forgot-my-passphrase"
            )


class ResetPassphraseOnUbuntuSSO(LandscapePage):
    """Redirect to Ubuntu SSO's forgotten password page."""

    title = "Reset your passphrase on Ubuntu Single Sign On"
    redirect_url = "https://login.ubuntu.com/+forgot_password"
    redirect_delay = 1

    def __call__(self):
        self.request.response.setHeader(
            "Refresh", "{}; url={}".format(self.redirect_delay, self.redirect_url)
        )
        self.request.response.setHeader("Location", self.redirect_url)
        return super(ResetPassphraseOnUbuntuSSO, self).__call__()


class ChangePassphrase(LandscapePage):

    title = "Change your passphrase"


class ChangePassphraseForm(LandscapeFormProvider):

    fields = Fields(IResetPassphrase)

    @action("Change passphrase")
    def change_passphrase(self, action, data):
        self.view.notes.info("Your passphrase has been changed.")
        self.context.reset(data["passphrase"])
        self.request.response.redirect("/")
