# Copyright 2005-2022 Canonical Limited. All rights reserved.

import urllib.request
import urllib.parse
import urllib.error
from urllib.parse import urlsplit, urlunsplit, parse_qsl

from zope.security.proxy import removeSecurityProxy
from zope.security.interfaces import Unauthorized
from zope.component import getUtility, getAdapter, queryUtility
from zope.session.interfaces import ISession, IClientIdManager, ISessionDataContainer

from canonical.routes import route, Redirect
from canonical.url import get_url

from canonical.landscape.ui.lsopenid import ILandscapeOpenIDFactory
from canonical.landscape.ui.routes import LandscapeRoutes, check_disabled
from canonical.landscape.model.configuration import is_standalone, get_root_url
from canonical.landscape.model.dashboard.dashboard import Dashboard
from canonical.landscape.model.knowledge.search import KnowledgeBase
from canonical.landscape.model.root.login import Login
from canonical.landscape.model.main.standalone import LDSScriptExecutionManagement
from canonical.landscape.security.authentication import (
    LandscapeAuthentication,
    get_host_from_request,
    write_last_login_notes,
)
from canonical.landscape.security.interfaces import IAnonymousPrincipal
from canonical.landscape.security.roles import is_support
from canonical.pam import pam

from safe.canonical.landscape.model.main.account import get_account_by_name
from safe.canonical.landscape.model.account.invitation import get_invitation
from safe.canonical.landscape.model.main.person import (
    get_person_by_id,
    get_password_reset,
    get_person_by_email,
)
from safe.canonical.landscape.model.main.computer import get_computer_by_id
from canonical.landscape.model.main.person import get_person_by_identity, create_person
from canonical.landscape.model.feature.flag import is_feature_enabled
from canonical.landscape.model.sandbox.sandbox import Sandbox
from canonical.landscape.ui.skin.ajax.interfaces import IAjaxSkin


class LandscapeLoginRoutes(LandscapeRoutes):
    def _update_login_data(self, auth):
        """
        If we have a record of the person associated with the supplied
        identity URL, we need to update their session data and continue. If
        not, they are a new user and we need them to provide us with more data.
        """
        redirect_target = None
        session = ISession(self.request)
        person_id = self.context.get_person_id_by_identity(auth.identity_url)
        if person_id is None and self.request.get("create") == "yes":
            # If we were passed the 'create' flag, and there's no person with
            # the given email yet, then automatically create the new user
            if not get_person_by_email(auth.email):
                person = create_person(
                    auth.name, auth.email, identity=auth.identity_url
                )
                person_id = person.id

        auth_data = session["landscape.authentication"]
        auth_data["identity"] = auth.identity_url
        auth_data["name"] = auth.name
        auth_data["email"] = auth.email
        auth_data["timezone"] = auth.timezone
        auth_data["nickname"] = auth.nickname
        auth_data["extra_emails"] = auth.extra_emails

        if person_id is not None:
            auth_data["person_id"] = person_id
            write_last_login_notes(person_id, self.notes)
            client_host = get_host_from_request(self.request)
            self.context.update_login_data(auth.identity_url, client_host)
        else:
            # New user, let them fill out their details
            redirect_target = "/complete-new-user"
            invitation_id = self.request.get("invitation")
            if invitation_id:
                redirect_target += "?invitation=%s" % invitation_id
            elif is_feature_enabled("self-service"):
                # If there is no invitation, then the user is signing up for a
                # new account, so redirect to the signup page.
                redirect_target = "/signup"

        return redirect_target

    def _get_default_start_page(self, identity):
        """Get the default start page for the L{Person} with C{identity}.

        The default start page is different, depending on a variety of
        circumstances:

         * If a C{next_url} field is defined in the request, the specified URL
           is returned.
         * If no L{Person} matches C{identity}, the dashboard URL is returned.
         * If the L{Person} is not a member of an L{Account}, the dashboard
           URL is returned.
         * If the L{Person} is a member of one or more L{Account}s the default
           account URL is returned.
        """
        if "next_url" in self.request:
            return self.request["next_url"]

        person = get_person_by_identity(identity)
        if not person:
            return "/dashboard"

        target_account = person.get_default_account()
        if not target_account:
            return "/dashboard"

        return get_url(target_account, self.request)

    def _get_redirect_target(self, auth):
        """
        For a given identity, determine the appropriate redirection
        location
        """
        invitation_id = self.request.get("invitation")
        invitation = None
        if invitation_id is not None:
            invitation = get_invitation(invitation_id)
        if invitation is not None:
            redirect_target = "/accept-invitation/" + invitation.secure_id
        elif self.context.accounts_are_expired(auth.identity_url):
            redirect_target = "/account/%s/expired" % (
                self.context.get_default_account_name(auth.identity_url)
            )
        else:
            redirect_target = self._get_default_start_page(auth.identity_url)

        new_person_redirect = self._update_login_data(auth)
        if new_person_redirect:
            next_url = self.request.get("next_url")
            if next_url:
                new_person_redirect = self._add_request_parameter(
                    new_person_redirect, "next_url", next_url
                )
            redirect_target = new_person_redirect
        return redirect_target.encode("utf-8")

    def _add_request_parameter(self, url, key, value):
        """
        Given a URL, this will break apart the query string (if any) and add a
        given key and value to it.

        @param url: URL to modify
        @type url: C{String}
        @param key: Parameter name to add to the query string.
        @type key: C{String}
        @param value: Value to be appended, this will be urlencoded
        @type value: C{String}
        @return: New URL with appended key/value pair
        """
        scheme, location, path, query, fragment = urlsplit(url)
        query_params = parse_qsl(query)
        query_params.append(
            (
                key,
                value,
            )
        )
        return urlunsplit(
            (scheme, location, path, urllib.parse.urlencode(query_params), fragment)
        )

    @route("handle-openid")
    def handle_openid(self):
        # This does cryptographic verification that the authentication data is
        # real, based on earlier exchange with the provider during begin().
        factory = getUtility(ILandscapeOpenIDFactory)
        openid = factory.get_landscape_openid(self.request)
        auth = openid.complete()
        if auth is not None:
            self._reset_client_id(self.request)
            redirect_target = self._get_redirect_target(auth)
        else:
            self.notes.error("Ubuntu Single Sign On authentication failed.")
            redirect_target = "/" if is_standalone() else "/login"
            invitation_id = self.request.get("invitation")
            if invitation_id:
                redirect_target += "?invitation=%s" % invitation_id
        raise Redirect(redirect_target)

    def _reset_client_id(self, request):
        """
        Replace the existing client identifier with a newly generated one.

        Because insecure sessions can transition to a secure
        state via an OpenID authentication, it is important that the publicly
        communicated identifier for the session is replaced when we enter that
        secure context, thus avoiding session fixation attacks.
        """
        session = ISession(self.request)
        old_client_id = session.client_id
        client_id_manager = getUtility(IClientIdManager)
        new_client_id = client_id_manager.generateUniqueId()
        client_id_manager.setRequestId(request, new_client_id)
        sdc = getUtility(ISessionDataContainer)
        sdc.reset_client_id(old_client_id, new_client_id)
        self.client_id = new_client_id


class LandscapeRootRoutes(LandscapeRoutes):

    logout_url = None
    features_url = "https://www.ubuntu.com/landscape/features"
    pricing_url = "https://www.ubuntu.com/landscape/pricing"
    install_url = "https://www.ubuntu.com/landscape/install"

    def __init__(self, context, request):
        super(LandscapeRootRoutes, self).__init__(context, request)
        if self.request.method == "HEAD":
            return None
        self.openid_factory = queryUtility(ILandscapeOpenIDFactory)
        if self.openid_factory is not None:
            self.logout_url = self.openid_factory.get_logout_url(get_root_url())
            self.trusted_redirects = [
                self.logout_url,
                self.features_url,
                self.pricing_url,
                self.install_url,
            ]

    def logged_in(self):
        return not IAnonymousPrincipal.providedBy(self.request.principal)

    @route("settings", permission="landscape.ViewPersonResource")
    def settings(self):
        return self.request.principal.person

    @route("person/<id>", permission="landscape.ViewPersonResource")
    def person(self, id):
        try:
            person_id = int(id)
        except ValueError:
            return None
        return get_person_by_id(person_id)

    @route("account/<name>")
    def account(self, name):
        if not self.logged_in():
            raise Unauthorized()

        account = get_account_by_name(name)
        if account is None:
            return None

        is_expiration = self.request["PATH_INFO"].endswith("/expired")
        account = removeSecurityProxy(account)

        if not is_support(self.request):
            if account not in self.request.principal.person.accounts:
                # Don't raise Unauthorized to avoid disclosing account names
                return None
            if account.has_expired() and not is_expiration:
                raise Redirect("/account/%s/expired" % account.name)

            check_disabled(LandscapeRoutes(account, self.request))

        return account

    @route("computer/<id>", permission="landscape.ViewPersonResource")
    def computer(self, id):
        computer = get_computer_by_id(int(id))
        if computer is None:
            # No computer matches the ID, return None to let the route system
            # raises NotFound.
            return None
        url = get_url(computer, self.request)
        raise Redirect(url)

    @route("dashboard", permission="landscape.ViewPersonResource")
    def dashboard(self):
        return Dashboard(self.request.principal.person)

    @route("kb", permission="landscape.ViewKnowledgeBase")
    def kb_base(self):
        return KnowledgeBase()

    @route("javascript")
    def javascript(self):
        return getAdapter(self.request, name="javascript")

    @route("static")
    def static(self):
        return getAdapter(self.request, name="static")

    @route("login")
    def login(self):
        if is_standalone() and not self.logout_url:
            next_url = self.request.get("next_url")
            if next_url:
                next_url = next_url.encode("utf-8")
                new_url = "/?%s" % urllib.parse.urlencode({"next_url": next_url})
            else:
                new_url = "/"
            raise Redirect(new_url)
        return Login()

    @route("redirect")
    def redirect(self):
        person = self.request.principal.person if self.logged_in() else None
        account = person.accounts.first() if person else None

        destination = self.request.get("next_url")
        # If we don't get a next_url, then we look to see if the person is
        # logged in and has an account, if they have, we send them to the
        # first account, otherwise, they're not associated with an account,
        # so we send them to the dashboard, where they should receive (in
        # the LDS case), the information on how to contact their LDS admin.
        if not destination:
            destination = "/dashboard"
            if account:
                destination = get_url(account, self.request)

        if "notes.info" in self.request:
            self.notes.info(self.request["notes.info"])

        if "notes.error" in self.request:
            self.notes.error(self.request["notes.error"])

        if isinstance(destination, str):
            destination = destination.encode("utf-8")
        raise Redirect(destination)

    @route("logout")
    def logout(self):
        """Logout the currently logged in person.

        If the person is using OpenID, we redirect him to the Ubuntu SSO
        logout. Otherwise, we redirect to the frontpage with a message. In both
        cases we clean up the local session.
        """
        identity = None
        if self.openid_factory is not None and self.logged_in():
            identity = self.request.principal.person.identity
        auth = LandscapeAuthentication()
        auth.logout(self.request)
        if identity and self.logout_url:
            url = self.logout_url
        else:
            url = get_url(self.context, self.request)

        raise Redirect(url)

    @route("accept-invitation/<id>")
    def accept_invitation(self, id):
        invitation = get_invitation(id)
        if not invitation:
            # No invitation with the given ID. Returning none raises NotFound.
            return None

        if not self.logged_in():
            # if they are not logged in yet, it's likely they are a new
            # user that has just been invited.  To stem the tide of
            # confusion, take them directly to the new user page.
            # from there, they will get a chance to login as an
            # existing user or create a new one.
            if is_standalone():
                if pam.has_service("landscape"):
                    redirect_url = "/new-pam-user"
                elif self.openid_factory is not None:
                    redirect_url = "/new-user"
                else:
                    redirect_url = "/new-standalone-user"
            else:
                redirect_url = "/new-user"
            redirect_url += "?invitation=%s" % invitation.secure_id
            raise Redirect(redirect_url)
        else:
            # otherwise the user is currently logged in.  So, we should
            # take them right to the invitation.
            return invitation

    @route("reset-passphrase/<id>")
    def reset_passphrase(self, id):
        reset = get_password_reset(id)
        if not reset:
            self.notes.info("This password reset link is no longer valid.")
            raise Redirect("/dashboard")
        return reset

    @route("script-issues")
    def script_issues(self):
        return LDSScriptExecutionManagement()

    @route("sandbox")
    def sandbox(self):
        sandbox = Sandbox()

        # Start fresh unless this is an AJAX request
        if not IAjaxSkin.providedBy(self.request):
            sandbox.reset()

        sandbox.__no_proxy__ = True  # Security is in the service layer
        return sandbox

    @route("~/<account_page>")
    def my_account_pages(self, account_page):
        if not self.logged_in():
            url = "/login/authenticate?next_url=/~/{}".format(account_page)
            raise Redirect(url)
        person = self.request.principal.person
        account = person.get_default_account()
        if account is None:
            return None  # This makes the routes system return a 404.

        account_url = get_url(account, self.request)
        raise Redirect("{}/{}".format(account_url, account_page))

    @route("landscape-features")
    def landscape_features_old_link(self):
        raise Redirect(self.features_url, status=301)

    @route("try-landscape")
    def try_landscape_old_link(self):
        raise Redirect(self.pricing_url, status=301)

    @route("set-up-on-prem")
    def set_up_landscape_old_link(self):
        raise Redirect(self.install_url, status=301)
