# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.cachedescriptors.property import Lazy
from zope.publisher.interfaces import NotFound
from zope.component import getUtility
from zope.session.interfaces import ISession

from canonical.url import get_url
from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.configuration import (
    get_stripe_public_key,
    get_stripe_secret_key,
)
from canonical.landscape.model.feature.flag import is_feature_enabled
from canonical.landscape.ui.page.page import Page
from canonical.landscape.ui.page.csp import (
    make_csp_header,
    CSP_KIND_SCRIPT,
    CSP_KIND_IMG,
)
from canonical.landscape.stripe.customer import setup_intent
from canonical.landscape.marketo.client import MarketoClient


class SignupPage(Page):

    title = "Signup"
    csp = make_csp_header(
        {
            CSP_KIND_SCRIPT: ("https://js.stripe.com/",),
            CSP_KIND_IMG: ("assets.ubuntu.com",),
        }
    )

    def __call__(self):
        result = super(SignupPage, self).__call__()
        if not is_feature_enabled("self-service"):
            raise NotFound(self, "Page not found")

        if self.person:
            account = self.person.get_default_account()
            if account:
                # If the person is part of an account, redirect to the
                # dashboard.
                self.request.response.redirect(get_url(account, self.request))
                return

        return result

    @Lazy
    def person(self):
        """Return the logged in person, or None if not logged in."""
        return getattr(self.request.principal, "person", None)

    @Lazy
    def openid_details(self):
        """Return a dict with the user details from OpenID."""
        openid_details = ISession(self.request)["landscape.authentication"]
        return {k.decode(): v for k, v in openid_details.items()}

    @Lazy
    def is_logged_in(self):
        """Whether the user is logged in."""
        return bool(self.openid_details)

    @Lazy
    def stripe_client_secret(self):
        """Return a SetupIntent client_secret used to authenticate cards."""
        intent = setup_intent(get_stripe_secret_key())
        return intent.client_secret

    @Lazy
    def signup_js_url(self):
        return self.static_root_path() + "/javascript/bundles/signup-min.js"

    @Lazy
    def react_js_url(self):
        return self.static_root_path() + "/javascript/bundles/react-min.js"

    @Lazy
    def stripe_publishable_key(self):
        return get_stripe_public_key()

    @Lazy
    def openid_url(self):
        config = getUtility(IFileBasedConfiguration)
        return config.get("landscape", "openid-provider-url")
