# Copyright 2005-2014 Canonical Ltd. All rights reserved.

from builtins import str
from canonical.landscape.model.main.account import get_account_by_salesforce_account_key
from canonical.landscape.model.account.invitation import get_invitation_by_sfid
from canonical.landscape.model.main.salesforceuser import get_salesforce_user_by_sfid


__all__ = [
    "check_salesforce_key",
    "check_salesforce_account_key",
    "validate_salesforce_key",
    "validate_salesforce_account_key",
]


__clearlist__ = [
    "check_salesforce_key",
    "check_salesforce_account_key",
    "validate_salesforce_key",
    "validate_salesforce_account_key",
]


# A special Salesforce code, specified at the beginning of an object
# ID, means the object is a contact.
CONTACT_OBJECT_SIGNATURE = "003"

# The same applies to Salesforce account key, which is also preceded
# by a version identifier.
ACCOUNT_OBJECT_SIGNATURE = "1-001"


def check_salesforce_key(key):
    """
    Check if the given is a correct Salesforce key.
    """
    return (
        key.startswith(CONTACT_OBJECT_SIGNATURE)
        and len(key) in (15, 18)
        and str(key).isalnum()
    )


def check_salesforce_account_key(key):
    """Check if the given is a correct Salesforce account key.

    The key is preceded by a version identifier prefix.
    """
    return (
        key.startswith(ACCOUNT_OBJECT_SIGNATURE)
        and len(key) in (17, 20)
        and str(key[2:]).isalnum()
    )


SALESFORCE_KEY_ERROR = (
    "Salesforce contact IDs must be 15 or 18 characters, "
    "start with '%s' and may only contain alphanumeric characters"
    % (CONTACT_OBJECT_SIGNATURE,)
)


SALESFORCE_ACCOUNT_KEY_ERROR = (
    "Salesforce account keys must be 17 or 20 "
    "characters, start with '%s' and may only contain alphanumeric "
    "characters after the version prefix" % (ACCOUNT_OBJECT_SIGNATURE,)
)


def validate_salesforce_key(form, form_field, salesforce_key, check_user=True):
    """Check that a Salesforce key is valid.

    The key must
    1) comply with the format required of SF keys,
    2) not be used by any current account invitations and
    3) not be used by any other people.

    @param form: The form object (with C{add_error} method) to add errors to.
    @param form_field: The name of the form field to pass to C{form.add_error}.
    @param salesforce_key: The salesforce key to check.
    @param check_user: whehter we should check for users with existing key or
        not.
    """
    if salesforce_key and not check_salesforce_key(salesforce_key):
        form.add_error(SALESFORCE_KEY_ERROR, form_field)

    if check_user:
        sfuser = get_salesforce_user_by_sfid(salesforce_key)
        if sfuser is not None:
            form.add_error(
                "Salesforce key is already used by %s (%s) in account %s"
                % (sfuser.person.name, sfuser.person.email, sfuser.account.name),
                form_field,
            )

    invitation = get_invitation_by_sfid(salesforce_key)
    if invitation is not None:
        form.add_error(
            "Salesforce key is already used by an invitation to %s (%s) "
            "in account %s"
            % (invitation.name, invitation.email, invitation.account.name),
            form_field,
        )


def validate_salesforce_account_key(form, form_field, salesforce_account_key):
    """Check that a Salesforce account key is valid.

    The key must
    1) comply with the format required of SF keys,
    2) not be used by any current account and

    @param form: The form object (with C{add_error} method) to add errors to.
    @param form_field: The name of the form field to pass to C{form.add_error}.
    @param salesforce_account_key: The Salesforce account key to check.
    """
    if salesforce_account_key and not check_salesforce_account_key(
        salesforce_account_key
    ):
        form.add_error(SALESFORCE_ACCOUNT_KEY_ERROR, form_field)

    account = get_account_by_salesforce_account_key(salesforce_account_key)
    if account is not None:
        form.add_error(
            "Salesforce account key is already used by account %s (%s) "
            % (account.title, account.name),
            form_field,
        )
