# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import range
from builtins import object
from zope.browserpage import ViewPageTemplateFile
from zope.schema import Choice
from zope.security.interfaces import Unauthorized

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider, LongTextWidget
from canonical.landscape.model.script.interfaces import IScript
from canonical.landscape.model.script.script import (
    DuplicateScriptError,
    EmptyScriptCodeError,
    EmptyScriptInterpreterError,
    split_code_on_interpreter,
)
from canonical.landscape.model.main.person import PersonAccessContextsVocabularyFactory
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME


class CreateScriptPage(AccountPage):
    """Page used to create scripts.

    The page has controls to specify source code, default script timeout and
    script name. It ensures that new scripts have all required fields.

    @ivar title: The page title.
    """

    title = "Create script"


class ScriptAttachmentsFormMixin(object):
    """
    Mixin used to define some common behavior for managing script attachments.
    """

    attachments_error = ""
    attachment_help_text = (
        "On the client, the attachments will be placed in the directory "
        "whose name is accessible through the environment variable "
        "LANDSCAPE_ATTACHMENTS. They'll be deleted once the script has been "
        "run."
    )

    def __init__(self, *args, **kwargs):
        super(ScriptAttachmentsFormMixin, self).__init__(*args, **kwargs)
        self.max_files = self.context.account.max_attachment_files

    def max_attachment_size(self):
        format_bytes = self.request.principal.person.formatter.format_bytes
        return format_bytes(self.context.account.max_attachment_size)

    def _check_attachments(self, total_length=0, existing_filenames=()):
        """
        @param total_length: total size of existing attachments
        @type total_length: C{int}

        @param existing_filenames:  list of existing attachments filenames
        @type existing_filenames: C{list} of C{unicode}
        """
        max_attachment_size = self.context.account.max_attachment_size
        attachments = {}
        for i in range(self.max_files):
            attachment = self.request.form.get("form.attachments.%d" % (i,))
            if not attachment:
                continue
            # Check for unique filenames
            if (
                attachment.filename in attachments
                or attachment.filename in existing_filenames
            ):
                self.attachments_error = "Attachments must have unique file names."
                return
            # Let's try not to read data uselessly if the attachment is too big
            attachment.seek(0, 2)
            total_length += attachment.tell()
            if total_length > max_attachment_size:
                break
            attachment.seek(0)
            attachment_data = attachment.read()
            attachments[attachment.filename] = attachment_data
        if total_length > max_attachment_size:
            format_bytes = self.request.principal.person.formatter.format_bytes
            self.attachments_error = (
                "Attachment size is too large: %s (maximum is %s)"
                % (format_bytes(total_length), format_bytes(max_attachment_size))
            )
            return
        return attachments


class CreateScriptForm(ScriptAttachmentsFormMixin, LandscapeFormProvider):

    summary = "Create script"

    fields = Fields(IScript)

    fields["title"].custom_widget = LongTextWidget

    additional_content = ViewPageTemplateFile("create-upload-fields.pt")

    dropdown_css_class = "dropdown-spacer-script"

    def __init__(self, context, request, view):
        super(CreateScriptForm, self).__init__(context, request, view)
        access_context = Choice(
            __name__="access_context",
            title="Access group",
            source=PersonAccessContextsVocabularyFactory("EditScript"),
            default=DEFAULT_ACCESS_CONTEXT_NAME,
            description="The access group the script should belong to.",
            required=True,
        )
        self.fields += Fields(access_context)
        if not self.context.person_can("EditScript"):
            raise Unauthorized()

    def validate(self, action, data):
        super(CreateScriptForm, self).validate(action, data)
        try:
            split_code_on_interpreter(data.get("code"))
        except EmptyScriptCodeError:
            self.add_error("Script does nothing", "code")
        except EmptyScriptInterpreterError:
            self.add_error("Script needs an interpreter", "code")

    def get_initial_data(self):
        return {"access_context": self.context.account.access_context.name}

    @action("Save", primary=True)
    def save(self, action, data):
        """Create a new script.

        If attachments have been added to the script, but are not referenced
        by it, a warning will be displayed and a new script will not be
        created.
        """
        attachments = self._check_attachments()
        if attachments is None:
            return

        try:
            self.context.create_script(
                data["title"],
                data["code"],
                data["time_limit"],
                data["username"],
                attachments,
                access_context_name=data["access_context"],
            )
        except DuplicateScriptError:
            self.add_error(
                "This title already exists. Please use a different one.", "title"
            )
            return

        if attachments and "LANDSCAPE_ATTACHMENTS" not in data["code"]:
            self.view.notes.warning(
                "This script contains one or more attachments, but you don't "
                "seem to refer to them using the LANDSCAPE_ATTACHMENTS "
                "variable in your code. Please check that your code is "
                "correct."
            )
        self.request.response.redirect(get_url(self.context, self.request))
