# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.browserpage import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.schema import TextLine
from zope.security.management import checkPermission

from canonical.content.form import Fields, action
from canonical.lib.text import describe_list
from canonical.url import get_url

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider, LongTextWidget
from canonical.landscape.model.eventlog.management import EventLog
from canonical.landscape.model.script.interfaces import IScript
from canonical.landscape.model.script.management import ScriptManagement
from canonical.landscape.ui.script.create import ScriptAttachmentsFormMixin
from canonical.landscape.model.script.script import (
    split_code_on_interpreter,
    EmptyScriptCodeError,
    EmptyScriptInterpreterError,
)


class EditScriptPage(AccountPage):

    title = "Edit script"

    def __init__(self, context, request):
        super(EditScriptPage, self).__init__(context, request)
        self.disable_save = "true"
        if checkPermission("landscape.EditScript", context):
            self.disable_save = "false"


class EditScriptForm(ScriptAttachmentsFormMixin, LandscapeFormProvider):

    summary = "Edit script"

    additional_content = ViewPageTemplateFile("edit-upload-fields.pt")

    def __init__(self, context, request, view):
        super(EditScriptForm, self).__init__(context, request, view)
        if view.disable_save == "true":
            self.summary = "View script"
        self.management = ScriptManagement(
            self.context.account, self.request.principal.person
        )

    @Lazy
    def fields(self):
        fields = Fields(IScript)
        fields["title"].custom_widget = LongTextWidget
        fields += Fields(
            TextLine(
                __name__="access_context",
                title="Access group",
                readonly=True,
                default=self.context.access_context.title,
            )
        )
        return fields

    def get_initial_data(self):
        return {
            "title": self.context.title,
            "username": self.context.username,
            "time_limit": self.context.time_limit,
            "code": self.context.get_code(),
        }

    def format_size(self, size):
        return self.request.principal.person.formatter.format_bytes(size)

    def validate(self, action, data):
        super(EditScriptForm, self).validate(action, data)
        title = data.get("title")
        if title is not None:
            existing_script = self.management.get_scripts(
                titles=[title], permission=None
            ).one()
            if existing_script and existing_script.id != self.context.id:
                self.add_error(
                    "This title already exists. Please use a different one.", "title"
                )
        try:
            split_code_on_interpreter(data.get("code"))
        except EmptyScriptCodeError:
            self.add_error("Script does nothing", "code")
        except EmptyScriptInterpreterError:
            self.add_error("Script needs an interpreter", "code")

    @action("Save", primary=True)
    def save(self, action, data):
        # First remove selected attachments
        event_log = EventLog(self.context.account, self.request.principal.person)
        ids_to_remove = []
        filenames_to_remove = []
        for attachment in self.context.attachments:
            input_name = "form.remove_attachments.%d" % (attachment.id,)
            if input_name in self.request.form:
                ids_to_remove.append(attachment.id)
                filenames_to_remove.append(attachment.filename)
        if ids_to_remove:
            self.context.remove_attachments_by_ids(ids_to_remove)
            if len(ids_to_remove) == 1:
                self.view.notes.info("Attachment removed.")
            else:
                self.view.notes.info("Attachments removed.")
            event_log.create_log_entry(
                self.context, describe_list("Removed attachment:", filenames_to_remove)
            )

        # Then get new ones
        total_length = 0
        existing_filenames = []
        for attachment in self.context.attachments:
            total_length += attachment.get_size()
            existing_filenames.append(attachment.filename)

        attachments = self._check_attachments(total_length, existing_filenames)

        if attachments is None:
            return

        if self.management.edit_script(self.context.id, data):
            self.view.notes.info("Script changed.")
        elif not attachments and not ids_to_remove:
            self.view.notes.info("No changes made.")

        if attachments:
            for filename, data in attachments.items():
                self.context.add_attachment(filename, data)
                event_log.create_log_entry(
                    self.context, "Attachment '%s' added." % filename
                )
            if len(attachments) == 1:
                self.view.notes.info("Attachment added.")
            else:
                self.view.notes.info("Attachments added.")

        if (
            self.context.attachments.any()
            and "LANDSCAPE_ATTACHMENTS" not in self.context.code
        ):
            self.view.notes.warning(
                "This script contains one or more attachments, but you don't "
                "seem to refer to them using the LANDSCAPE_ATTACHMENTS "
                "variable in your code. Please check that your code is "
                "correct."
            )

        self.request.response.redirect(get_url(self.management, self.request))
