# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.security import checkPermission
from zope.browserpage import ViewPageTemplateFile

from canonical.content.form import Fields, action
from canonical.content.decorator import json
from canonical.url import get_url

from canonical.landscape.model.selection.info import ComputerSelectionInfo
from canonical.landscape.ui.toolbar.page import ToolbarPage
from canonical.landscape.ui.form import LandscapeFormProvider, LongTextWidget
from canonical.landscape.ui.activity.schedule import get_deliver_after_time
from canonical.landscape.model.script.management import ScriptManagement
from canonical.landscape.model.script.script import (
    ScriptExecution,
    DuplicateScriptError,
    split_code_on_interpreter,
    EmptyScriptCodeError,
    EmptyScriptInterpreterError,
)
from canonical.landscape.model.script.interfaces import IScript, IScriptExecution
from canonical.landscape.ui.script.create import ScriptAttachmentsFormMixin


NO_COMPUTER_SELECTED = "Computers must be selected before scripts can be run."


class ExecuteScriptWizardPage(ToolbarPage):
    """
    Navigation page to choose to either load a script or start with an
    empty one for the run script page.

    @ivar title: The page title.
    """

    title = "Select script to run"

    def __init__(self, context, request):
        super(ExecuteScriptWizardPage, self).__init__(context, request)
        self.scripts = []
        if not self.context.get_selected_count():
            self.notes.error(NO_COMPUTER_SELECTED)
            self.request.response.redirect(get_url(self.context, self.request))
        else:
            script_management = ScriptManagement(
                self.context.account, self.request.principal.person
            )
            self.scripts = list(script_management.get_scripts())
            if not self.scripts:
                self.request.response.redirect(
                    get_url(self.context, self.request) + "/scripts/run"
                )

    def toolbar_class(self):
        return ""


class ExecuteScriptPage(ToolbarPage):
    """
    Form page to run a script on the selected computers. It can preload the
    execution form with a script, or start with default values.

    @ivar title: The page title.
    """

    title = "Run a script"

    def __init__(self, context, request):
        super(ExecuteScriptPage, self).__init__(context, request)
        if not self.context.get_selected_count():
            self.notes.error(NO_COMPUTER_SELECTED)
            self.request.response.redirect(get_url(self.context, self.request))

    def toolbar_class(self):
        return ""

    @json
    def get_execute_script_users(self):
        """
        Return a list of common users in the current selection of computers,
        encoded as JSON.
        """
        selection_info = ComputerSelectionInfo(self.context)
        return sorted(
            (info[0] for info in selection_info.get_users_info(["username"])),
            key=lambda username: (username != "nobody", username),
        )


class ExecuteScriptForm(ScriptAttachmentsFormMixin, LandscapeFormProvider):

    summary = "Run a script"

    fields = Fields(IScript).select("title", "code", "time_limit") + Fields(
        IScriptExecution
    )

    fields["title"].custom_widget = LongTextWidget

    additional_content = ViewPageTemplateFile("run-upload-fields.pt")
    script_id = None

    def __init__(self, context, request, view):
        super(ExecuteScriptForm, self).__init__(context, request, view)
        self.script_management = ScriptManagement(
            self.context.account, self.request.principal.person
        )

    def get_initial_data(self):
        self.attachments = []
        script_id = self.request.get("load_script", "")
        if script_id:
            script = self.script_management.get_scripts(ids=[int(script_id)]).one()
            if script is None:
                self.view.notes.warning("Unknown script.")
                return {}
            self.script_id = script.id
            self.attachments = script.attachments
            result = {
                "code": script.get_code(),
                "title": script.title,
                "time_limit": script.time_limit,
                "username": script.username,
            }
            return result
        else:
            return {}

    def validate(self, action, data):
        super(ExecuteScriptForm, self).validate(action, data)
        try:
            split_code_on_interpreter(data.get("code"))
        except EmptyScriptCodeError:
            self.add_error("Script does nothing", "code")
        except EmptyScriptInterpreterError:
            self.add_error("Script needs an interpreter", "code")

    def update(self):
        """Select form fields on-the-fly, basing on the user's privileges."""
        field_names = ["title", "code", "time_limit", "username"]
        if checkPermission(
            "landscape.EditAccount", self.context.account
        ) and not self.request.get("load_script"):
            field_names.append("save_script")
        self.fields = self.fields.select(*field_names)
        super(ExecuteScriptForm, self).update()

    def _run_and_redirect(self, data, attachments, existing_attachments):
        person = self.request.principal.person
        deliver_after = get_deliver_after_time(self.request)
        execution = ScriptExecution(self.context)

        interpreter, code = split_code_on_interpreter(data["code"])
        activity = execution.execute(
            person,
            data["title"],
            interpreter,
            code,
            data["username"],
            data["time_limit"],
            attachments,
            existing_attachments,
            deliver_after=deliver_after,
        )

        self.request.response.redirect(get_url(activity, self.request))

    @action("Run", primary=True)
    def run(self, action, data):
        total_length = 0
        existing_attachments = ()
        existing_filenames = ()
        script_id = self.request.get("load_script")
        if script_id:
            script = self.script_management.get_scripts(ids=[int(script_id)]).one()
            if script is not None:
                existing_attachments = list(script.attachments)
                total_length = sum(
                    [attachment.get_size() for attachment in existing_attachments]
                )
                existing_filenames = [
                    attachment.filename for attachment in existing_attachments
                ]

        attachments = self._check_attachments(total_length, existing_filenames)
        if attachments is None:
            return

        if data.get("save_script"):
            try:
                script = self.script_management.create_script(
                    data["title"],
                    data["code"],
                    data["time_limit"],
                    data["username"],
                    attachments,
                )
            except DuplicateScriptError:
                self.add_error(
                    "This title already exists. Please use a different one.", "title"
                )
                return
            else:
                self.view.notes.info("'%s' has been saved." % (data["title"],))

        self._run_and_redirect(data, attachments, existing_attachments)
