# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.component import getMultiAdapter
from zope.security.proxy import removeSecurityProxy
from zope.interface import directlyProvides

from canonical.lib.copycat import Copycat
from canonical.routes import route, Redirect
from canonical.url import get_url

from canonical.landscape.model.activity.parent import get_parent_activity
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.routes import ComputerSelectionActivities
from canonical.landscape.ui.selection.activity import IComputerSelectionActivity
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.users.users import UserManagement
from canonical.landscape.model.hardware.hardware import HardwareInventory
from canonical.landscape.model.monitoring.monitoring import Monitoring
from canonical.landscape.model.monitoring.processes import Processes
from canonical.landscape.model.package.management import PackageManagement
from canonical.landscape.model.snap.management import SnapManagement
from canonical.landscape.model.computer.interfaces import IComputerManagement
from canonical.landscape.model.reporting.management import ReportingManagement
from canonical.landscape.model.ubuntu_pro.ubuntu_pro import UbuntuProInfo
from canonical.landscape.ui.routes import LandscapeRoutes
from canonical.landscape.model.main.errors import UnknownAccessContextsError
from safe.canonical.landscape.model.main.tag import get_tags
from canonical.landscape.service.computer.resynchronize import ResynchronizeComputers


class ComputerSelectionRoutes(LandscapeRoutes):
    def _redirect(self, message):
        self.notes.error(message)
        raise Redirect(get_url(self.context, self.request))

    @route("info")
    def info(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before " "information can be viewed."
            )
        else:
            person = self.request.principal.person
            return getMultiAdapter((self.context, person), IComputerManagement)

    @route("edit")
    def edit(self):
        count = self.context.get_selected_count()
        if count == 1:
            computer = self.context.get_selected(limit=1)[0]
            url = get_url(computer, self.request)
            raise Redirect("%s/edit" % (url,))
            return
        if not count:
            message = "A computer must be selected before it can be edited."
        else:
            message = "Only a single computer can be edited at a time."
        self._redirect(message)

    @route("tags/<tag>")
    def selection(self, tag):
        tags = tag.split("+")
        self.context.clear_included_tags()
        self.context.include_tags(get_tags(tags))
        return self.context

    @route("accessgroups/<accessgroup>")
    def group_selection(self, accessgroup):
        accessgroups = accessgroup.split("+")
        self.context.clear_included_access_groups()
        try:
            self.context.include_access_groups(
                self.context.account.get_access_contexts(accessgroups)
            )
        except UnknownAccessContextsError:
            if len(accessgroups) > 1:
                message = (
                    "One of the following access groups specified is unknown: "
                    "%s. Please specify existing access group names."
                    % ", ".join(accessgroups)
                )
            else:
                message = (
                    "The access group %s is unknown. Please specify an "
                    "existing access group name." % ",".join(accessgroups)
                )
            self.notes.error(message)
        return self.context

    @route("search/<searchname>")
    def search(self, searchname):
        self.context.clear_search_criteria()
        self.context.set_saved_search_criteria(searchname)
        self.context.select_computers(self.context.get_available_ids())
        return self.context

    @route("criteria/<searchcriteria>")
    def criteria(self, searchcriteria):
        if not self.request.getTraversalStack():
            # If the stack is empty, the computer listing page has been
            # requested.  In this case we don't select computers, but redirect
            # to the listing page with the session selection and criteria set
            # so that the user can select computers from the search result.
            url = get_url(self.context, self.request)
            raise Redirect("%s?criteria=%s" % (url, searchcriteria))

        # Remove the double encoding of '/'
        searchcriteria = searchcriteria.replace("%2F", "/")

        # this does not alter the SessionComputerSelection, rather returns a
        # new one
        selection = ComputerSelection(
            self.context.account, self.context.person, self.context.permission
        )
        selection.clear_search_criteria()
        selection.set_search_criteria(searchcriteria)
        selected_ids = set(selection.get_available_ids())
        if "computer_ids" in self.request:
            computer_ids = set(
                int(id) for id in self.request["computer_ids"].split(",")
            )
            selected_ids = selected_ids.intersection(computer_ids)
        selection.select_computers(selected_ids)
        return selection

    @route("activities")
    def activities(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before " "activities can be viewed."
            )
        else:
            return ComputerSelectionActivities(self.context)

    @route("activity/<id>", id_pattern="\d+")
    def activity(self, id):
        account = removeSecurityProxy(self.context.account)
        activity_id = int(id)
        # The route currently supports both parent and child activities, so try
        # both. There's no risk of clashing since IDs are unique across parents
        # and children, and we plan to maintain uniqueness.
        activity = get_parent_activity(account, activity_id)
        if activity is None:
            activity = get_child_activity(account, activity_id)

        selected_computer_ids = self.context.get_selected_computer_ids()
        if not activity:
            self._redirect("Activity not found.")
        elif not selected_computer_ids:
            self._redirect(
                "Computers must be selected before " "activities can be viewed."
            )
        else:
            # Use a copycat of the activity decorated with an interface
            # signaling it's being viewed inside a ComputerSelection.  We
            # can't modify the Activity instance directly as it might get
            # cached and reused in other requests.
            #
            # It's safe to unwrap it here, because the publisher will wrap
            # the Copycat object again, and it will be protected with the
            # same checker of the activity itself (it's a class attribute).
            copycat = Copycat(removeSecurityProxy(activity))
            directlyProvides(copycat, IComputerSelectionActivity)
            Copycat.activate(copycat)
            return copycat

    @route("users")
    def users(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before " "user actions can be performed."
            )
        else:
            return UserManagement(self.context, self.request.principal.person)

    @route("monitoring")
    def monitoring(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before "
                "resource information can be displayed."
            )
        else:
            return Monitoring(self.context)

    @route("processes")
    def processes(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before "
                "active processes can be displayed."
            )
        else:
            return Processes(self.context)

    @route("packages")
    def packages(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before package "
                "management actions can be performed."
            )
        else:
            return PackageManagement(self.context)

    @route("snaps")
    def snaps(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before snap management actions can be performed."
            )
        else:
            return SnapManagement(self.context)

    @route("hardware")
    def hardware(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before hardware "
                "inventory information can be displayed."
            )
        else:
            return HardwareInventory(self.context)

    @route("ubuntu_pro_info")
    def ubuntu_pro_info(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before Ubuntu Pro info " "can be displayed"
            )
        else:
            return UbuntuProInfo(self.context)

    @route("resynchronize")
    def resynchronize(self):
        person = self.request.principal.person
        if person.has_global_role("Operator"):
            account = self.context.account
            operation = ResynchronizeComputers(account, self.request.principal)
            operation(self.context.get_selected_computer_ids())
            if self.context.get_selected_count() > 1:
                self.notes.info("Data for these computers will be resynchronized.")
            else:
                self.notes.info("Data for this computer will be resynchronized.")
        info_url = get_url(self.context, self.request) + "/info"
        raise Redirect(info_url)

    @route("reports")
    def reports(self):
        if not self.context.get_selected_count():
            self._redirect(
                "Computers must be selected before reporting "
                "information can be displayed."
            )
        else:
            return ReportingManagement(self.context)

    # May be interesting in the future:
    #
    # def default_traverse(self, stack):
    #    computers = self.context.get_selected_count()
    #    if len(computers) == 1:
    #        return computers[0]
    #    return None
