# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from builtins import str
from zope.cachedescriptors.property import Lazy

from canonical.content.form import Fields, action
from canonical.url import get_url_object

from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.model.account.interfaces import ISavedSearch
from canonical.landscape.model.account.savedsearch import (
    SavedSearchManagement,
    generate_saved_search_name,
    InvalidSavedSearchName,
    InvalidSavedSearchCriteria,
    DuplicateSavedSearchError,
    EmptySavedSearchTitle,
    EmptySavedSearchCriteria,
    EmptySavedSearchNameAndTitle,
)


class SavedSearchForm(LandscapeFormProvider):

    description = "Please provide the following information."
    fields = Fields(ISavedSearch, render_context=False)

    def __init__(self, context, request, view):
        super(SavedSearchForm, self).__init__(context, request, view)
        self.account = get_url_object(self.context, request, IAccount)
        self.management = SavedSearchManagement(self.account, request.principal.person)

    @Lazy
    def edit_saved_search_name(self):
        return self.get_edit_saved_search_name()

    @Lazy
    def summary(self):
        if self.edit_saved_search_name:
            return "Edit search"
        else:
            return "Save search"

    def get_edit_saved_search_name(self):
        return self.context.get_search_name()

    def get_initial_data(self):
        """
        When editing a saved search, the form must be rendered with the
        existing values.
        """
        if self.edit_saved_search_name:
            saved_search = self.management.get_saved_search_by_name(
                self.edit_saved_search_name
            )
            saved_search_dictionary = {}
            for field in ("name", "title", "search"):
                saved_search_dictionary[field] = getattr(saved_search, field, "")
            return saved_search_dictionary
        else:
            return {}

    def add_error(self, error, field_name=None):
        super(SavedSearchForm, self).add_error(error, field_name)
        self.request.response.setStatus(400)

    def validate(self, action, data):
        """
        Special validation made for embedded form providers.

        Since this form is hidden inside the selection view and its submission
        is made by xhr, in order for the javascript error handling to work, the
        response status must be set with an error code.
        Error 500 will suffice.
        """
        name = data.get("name", "")
        title = data.get("title", "")
        search = data.get("search", "")

        try:
            name, title, search = self.management.validate_create_saved_search(
                name, title, search
            )

        except InvalidSavedSearchCriteria as error:
            self.add_error(str(error), "search")

        except EmptySavedSearchNameAndTitle:
            self.add_error("A name must be provided.", "name")
            self.add_error("A title must be provided.", "title")

        except DuplicateSavedSearchError as error:
            if not self.edit_saved_search_name:
                for field in error.fields:
                    self.add_error(
                        "Saved search with this %s already exists." % field, field
                    )

        except EmptySavedSearchTitle as error:
            if not self.edit_saved_search_name:
                self.add_error(str(error), "title")

        except EmptySavedSearchCriteria as error:
            self.add_error(str(error), "search")

        except InvalidSavedSearchName as error:
            if not self.edit_saved_search_name:
                self.add_error(str(error), "name")

    @action("Save")
    def handle_save(self, action, data):
        """
        Handle ajax form submission and create or edit the saved search.
        """
        name = data.get("name", None)
        if name is None or name == "":
            name = generate_saved_search_name(self.management, data.get("title"))
            data["name"] = name
        name, title, search = (data.get("name"), data.get("title"), data.get("search"))
        if not self.edit_saved_search_name:
            # A new saved search is being created.
            self.management.create_saved_search(name, title, search)
        else:
            # The saved search is being edited.
            self.management.edit_saved_search(
                self.edit_saved_search_name, title=title, search=search
            )
        self.view.notes.info("Search successfully saved.")
        self.view.notes.remove_on_commit = False


class DeleteSavedSearchForm(LandscapeFormProvider):

    summary = "Delete saved search"

    description = "Are you sure you want to delete the following saved search?"

    fields = Fields(ISavedSearch, for_display=True).omit("search").hide("name")

    def __init__(self, context, request, view):
        super(DeleteSavedSearchForm, self).__init__(context, request, view)
        self.account = get_url_object(self.context, request, IAccount)
        self.management = SavedSearchManagement(self.account)

    def get_initial_data(self):
        """Pre-load the title for display"""
        saved_search_name = self.context.get_search_name()
        saved_search = self.management.get_saved_search_by_name(saved_search_name)
        if saved_search:
            return {"title": saved_search.title, "name": saved_search.name}
        else:
            return {}

    @action("Delete")
    def handle_delete(self, action, data):
        """
        A dummy action that is used just to render the form button.
        """
        return ""
