# Copyright 2005-2022 Canonical Limited. All rights reserved.

from json import loads
import re

from canonical.url import get_url
from canonical.content.decorator import json
from canonical.search.engine import SearchCriteriaError

from canonical.content.download import set_csv_download
from canonical.landscape.ui.computer.info import format_title_if_clone
from canonical.landscape.ui.toolbar.page import ToolbarPage
from canonical.landscape.ui.pager import PaginationMixin
from canonical.landscape.ui.dateformat import format
from canonical.landscape.model.main.computer import Computer, PingTime
from canonical.landscape.model.resource.rebootrequiredinfo import (
    get_reboot_required_info_for_computers,
)
from canonical.landscape.model.selection.csvselection import SelectionAsCSV

DEFAULT_PAGE_SIZE = 25

needs_reboot_regex = re.compile(r"\b(alert:computer-reboot|needs:reboot)\b")


def is_reboot_search(search_string):
    """Does the given search include a computer-needs-reboot criterion?"""
    if search_string is None:
        return False
    else:
        return bool(needs_reboot_regex.search(search_string))


class ComputerSelectionPage(ToolbarPage, PaginationMixin):

    title = "Select computers"

    hide_picker = True

    _sortable_columns = {
        "last_ping_time": PingTime.last_ping_time,
        "hostname": Computer.hostname,
        "title": Computer.title,
    }

    def __init__(self, context, request):
        super(ComputerSelectionPage, self).__init__(context, request)
        self.tag_url = None
        self.group_url = None
        tags = self.context.get_included_tags()
        if tags:
            url = get_url(self.context.account, self.request)
            tag_names = [tag.name for tag in tags]
            self.tag_url = "%s/tags/%s" % (url, "+".join(tag_names))
        groups = self.context.get_included_access_groups()
        if groups:
            url = get_url(self.context.account, self.request)
            group_names = [group.name for group in groups]
            self.group_url = "%s/accessgroups/%s" % (url, "+".join(group_names))
        self._computer_rows = None
        self._available_count = None
        self._computer_ids_selected = None
        self._computer_ids_available = None

        criteria = request.get("criteria")
        self.search_criteria = self.context.get_search_criteria()
        if criteria and not self.search_criteria:
            self.context.set_search_criteria(criteria)
            self.search_criteria = criteria
        self.no_computers = self.context.account.get_computers().count() == 0
        self.page_size = DEFAULT_PAGE_SIZE
        self.search_error = False

    def view_safe_search_criteria(self):
        """Escape C{search_criteria} so that it's safe for Javascript."""
        return self.view_safe_js_value(self.search_criteria)

    def update(self):
        """
        Method is called by Zope's traversal machinery when a computer
        selection update page is selected.  This method should only be
        called when form data has been POST'd to update the selection;
        it always performs a redirect back to the non-update page to
        get redirect-after-POST behaviour.
        """
        if "select_all" not in self.request and "select_none" not in self.request:
            selected_computer_ids = self.request.form.get("selection", ())
            selection_json = self.request.form.get("selection_json", "")

            if selection_json:
                selected_computer_ids = loads("[%s]" % selection_json)

            selected_computer_ids = [int(id) for id in selected_computer_ids]

            self.context.clear_selected_computers()
            self.context.select_computers(selected_computer_ids)

        if "download_csv" in self.request:
            set_csv_download(self.request.response, "selection")
            return SelectionAsCSV(self.context).get_csv_data()

        if "notes.info" in self.request:
            self.notes.info(self.request["notes.info"])
        if "notes.error" in self.request:
            self.notes.error(self.request["notes.error"])

        next_url = self.request.get("next_url")
        if next_url is None:
            next_url = get_url(self.context, self.request)

        self.redirect(next_url)

    def noop_ajax(self):
        """No Op (Do Nothing) URL Endpoint for ajax updates.

        Method is exposed to direct traversal. Called by ajax when updating
        the selection with either 'select_none' or 'select_all'. Those form
        values are directly intercepted by the account routes to update the
        selection. This method merely serves as an URL endpoint for the ajax
        updates.
        """
        return ""

    def available_computers_json(self):
        """
        This method is exposed to direct traversal. The browser uses
        this method to fetch batches of available computers for
        pagination in the ui.
        """
        if "all_computer_ids" in self.request:
            # We use the all_computer_ids parameter to avoid leakage from
            # existing computer selections in the computer_ids parameter.
            # Canonically when we create a selectable list of computers with
            # packages from which to generate a package profile.  See Bug:
            # #1004208 for the issue.
            computer_ids = set(
                int(id) for id in self.request["all_computer_ids"].split(",")
            )
            self.context.include_computers(computer_ids)
        return self._computers_json(selected=False)

    def selected_computers_json(self):
        """
        This method is exposed to direct traversal. The browser uses this
        method to fetch batches of selected computers for pagination in the ui.
        """
        return self._computers_json(selected=True)

    @json
    def _computers_json(self, selected=False):
        """
        Fetch batches of computers for pagination in the ui.
        """
        if "ids" in self.request:
            computer_ids = [
                int(computer_id) for computer_id in self.request["ids"].split(",")
            ]
            computers = sorted(
                self.context.get_computers_by_ids(computer_ids),
                key=lambda computer: computer_ids.index(computer.id),
            )
            records = self._format_computers(computers)
        else:
            info = self.get_sort_variables()
            sort_column = self.get_sort_column(info["sort"], info["sort_order"])
            records = self.computers(
                selected=selected,
                limit=info["limit"],
                offset=info["offset"],
                order_by=sort_column,
            )
        if selected:
            data = {
                "totalRecords": self.context.get_selected_count(),
                "records": records,
            }
        else:
            data = {
                "totalRecords": self.context.get_available_count(),
                "records": records,
            }

        self.request.response.setHeader("Content-Type", "application/json")
        return data

    def selected_ids(self):
        """
        Return the ids of all selected computers applicable to the current
        context.

        This fetches all the selected ids within the current context. For
        example we might have a 100 computers in our selection but be browsing
        a tag with only 10 computers, so we perform an intersection here to
        realize how many computers are selected in this context.
        """
        if self._computer_ids_selected is not None:
            return self._computer_ids_selected
        try:
            self._computer_ids_selected = list(self.context.get_selected_computer_ids())
        except SearchCriteriaError:
            pass
        return self._computer_ids_selected

    @json
    def selected_ids_json(self):
        """
        Return the selected ids as a JSON array. Used for rendering
        the ComputerSelection widget from other pages.
        """
        self.request.response.setHeader("Content-Type", "application/json")
        return self.selected_ids()

    def available_ids(self):
        """
        Returns the ids of all available computers that can be selected.
        """
        if not self._computer_ids_available:
            info = self.get_sort_variables()
            sort_column = self.get_sort_column(info["sort"], info["sort_order"])
            try:
                self._computer_ids_available = list(
                    self.context.get_available_ids(order_by=sort_column)
                )
            except SearchCriteriaError:
                pass
        return self._computer_ids_available

    @json
    def available_ids_json(self):
        """
        Return the available ids as a JSON array. Used for rendering
        the ComputerSelection widget from other pages.
        """
        self.request.response.setHeader("Content-Type", "application/json")
        return self.available_ids()

    def computers(self, selected=False, offset=None, limit=None, order_by=None):
        """Fetches computer objects.

        @param offset: The starting point/index into the total set of results
            that the batch begins with.
        @param limit: How many computers to return.  Default is
            L{DEFAULT_PAGE_SIZE}.
        @param order_by: Attribute to order selection by see
            L{SORTABLE_COLUMNS}
        """
        if limit is None:
            limit = self.page_size
        if self._computer_rows:
            return self._computer_rows
        elif not offset and limit:
            # The view renders in paginated batches. If we had previously
            # fetched all computers, then we need to invalidate the cache,
            # else we'll get all the cached computers. more details in
            # more details in bug:78685 (comments 14,15). the cache isn't
            # directly exposed via api. so we invoke a cache reseting
            # method with empty args.
            self.context.unselect_computers([])
        try:
            # Fetch the batch of computers
            if selected:
                computers = self.context.get_selected(
                    offset=offset, limit=limit, order_by=order_by
                )
            else:
                computers = self.context.get_available(
                    offset=offset, limit=limit, order_by=order_by
                )
        except SearchCriteriaError:
            if not self.search_error:
                message = "Invalid search criteria '%s'" % self.request.get("criteria")
                self.notes.error(message)
                self.search_error = True
            return []

        # Store ping times and computer statuses to cache the objects
        self._ping_times = list(self.context.account.get_ping_times(computers))
        self._computer_status = list(
            self.context.account.get_computer_statuses(computers)
        )

        # Render the batch
        self._computer_rows = self._format_computers(computers)
        return self._computer_rows

    def _format_computers(self, computers):
        computer_rows = []
        for computer in computers:
            is_selected = self.context.is_selected(computer.id)

            if computer.last_ping_time:
                last_activity = format(computer.last_ping_time, self.request)
            else:
                last_activity = ""
            packages = set()
            for reboot_required_info in get_reboot_required_info_for_computers(
                [computer]
            ):
                if reboot_required_info.packages:
                    packages.update(reboot_required_info.packages)
            if computer.reboot_required_flag:
                reboot_text = "This computer needs to be restarted"
                if packages:
                    reboot_text += (
                        " due to changes in the following packages:\n\n"
                        + "\n".join(sorted(packages))
                    )
            else:
                reboot_text = ""
            row = {
                "title": format_title_if_clone(computer),
                "computer_id": computer.id,
                "reboot_required_flag": computer.reboot_required_flag,
                "reboot_required_text": reboot_text,
                "last_activity": last_activity,
                "selected": is_selected,
                "has_cloud_instance": False,
                "hostname": computer.hostname,
            }
            computer_rows.append(row)
        return computer_rows

    def status(self):
        """Status of the computer selection is rendered in javascript."""
        selected = len(self.selected_ids())
        available = len(self.available_ids())
        if not available:
            return "No computers"
        elif not selected and available == 1:
            text = "computer"
        elif selected == 1:
            text = "computer"
        else:
            text = "computers"
        if selected:
            return "%d of %d %s selected" % (selected, available, text)
        return "%d %s" % (available, text)

    def render_save_search_form(self):
        """
        The search form must be rendered conditionally only when necessary.

        The form will be rendered only when a non-empty search was performed or
        when the save form was submitted.
        """
        return bool("form.actions.save" in self.request.form or self.search_criteria)

    def is_saved_search(self):
        """
        If the context is a saved search, the return C{True} so the "Edit this
        search" and "Delete this search"  button is displayed in the user
        interface.
        """
        return self.context.get_search_name() is not None

    def is_reboot_search(self):
        """
        If the context is a search for computers needing to reboot, return
        C{True} so the corresponding button is displayed in the user
        interface.
        """
        return is_reboot_search(self.search_criteria)
