"""Utilities for dealing with the service layer."""

from zope.security.interfaces import Unauthorized
from zope.publisher.interfaces import NotFound

from canonical.landscape.service.errors import UnauthorisedError
from canonical.landscape.service.entity.errors import UnknownEntityError
from canonical.landscape.service.security import PersonPrincipal


def service_lookup(request, lookup_class, *args, **kwargs):
    """Convenience for performing service layer entity lookups.

    This helper takes care of instantiating the lookup and invoke it with
    the given arguments. Unknown entity and authorisation errors are
    handled appropriately and translated to their Zope equivalent. When
    successful, entity lookups return a single model entity object,
    or an entity collection in case of collection lookups.

    NOTE: you generally want to call either LandscapeRoutes._lookup
          or Page._lookup, which specialise the behavior a little
          bit depending if the calling context is a route or a view.

    @param lookup_class: The ServiceLookup sub-class to use to perform
        the lookup.
    @param args: Positional arguments to pass to the lookup.
    @param kwargs: Keyword arguments to pass to the lookup.
    """
    # XXX Replace the UI-level principal with our service-level one. This
    # is a temporary measure while we transition, eventually the request
    # object will directly hold a reference to the service layer principal.
    principal = PersonPrincipal(request.principal.unproxied_person.id)
    lookup = lookup_class(principal)
    try:
        result = lookup(*args, **kwargs)
    except UnknownEntityError as error:
        # This can be raised only by an EntityLookup, not by a
        # CollectionLookup.
        raise NotFound(None, str(error))
    except UnauthorisedError:
        raise Unauthorized(None, "", request)
    return result
