# Copyright 2005-2022 Canonical Limited. All rights reserved.

import collections
from datetime import timedelta

from zope.component import getUtility
from zope.interface import implementer, Interface
from zope.schema import TextLine
from zope.session.session import Session, RAMSessionDataContainer
from zope.session.interfaces import (
    ISessionDataContainer,
    ISessionData,
    ISessionPkgData,
    ISession,
)

from storm.zope.interfaces import IZStorm

from canonical.landscape.model.session.data import SessionPkgData

# The generation idea works like this: for each client id, we have a row
# in the session_data table, containing the generation (a sequential
# integer), and the last access time.  Whenever we have a request, we
# ask the database for the current generation for this client id using
# the get_generation() function.  Internally, this function will perform
# one of the following actions, in a fast and race-proof way (assuming
# that the transaction is in AUTOCOMMIT or READ COMMITTED mode):
#
# a) Create a new row with the current access time and a generation of 0
#    and return it, if the client id was never seen;
# b) Return the current generation immediately, if the last access time
#    was updated recently enough;
# c) Return the current generation after updating the access time, if the
#    access time isn't old enough to have expired;
# d) Increment the current generation and update the access time, and
#    return the updated generation, if the access time is old enough to
#    have expired.
#
# Then, the retrieved generation is simply used as part of the key in the
# session_pkg_data table.  Whenever the generation is bumped due to
# expiration of the access time, old rows are transparently forgotten,
# and may be collected at a later time with no trouble.


@implementer(ISessionDataContainer)
class SessionDataContainer(object):
    """
    An L{ISessionDataContainer} that retrieves sessions from the database.

    @ivar timeout: time in seconds after which a session will be deleted if
        not accessed.
    @type timeout: C{int}

    @ivar resolution: time in seconds separating sweep and last accessed
        update.
    @type resolution: C{int}
    """

    resolution = timedelta(seconds=10 * 60)
    timeout = timedelta(seconds=3600 * 24)

    def __getitem__(self, client_id):
        """
        Return a SessionData for the given C{client_id}. If it already exists,
        it returns it from the database, if not, it creates a new one.
        """
        client_id = str(client_id)
        store = getUtility(IZStorm).get("session-autocommit")
        # XXX In the future, we'll most probably want to cache the generation
        #     locally for all threads (protected by a lock), but this function
        #     is very fast and this is far from being an issue for a long time,
        #     so we can push the first version to production without any
        #     caching.
        result = store.execute(
            "SELECT get_generation(?, INTERVAL ?, INTERVAL ?)",
            (client_id, str(self.resolution), str(self.timeout)),
        )
        return SessionData(client_id, result.get_one()[0])

    def reset_client_id(self, old_client_id, new_client_id):
        """
        When transitioning a session from an insecure to a secure context, for
        example when a session is authenticated with an external OpenID
        service, it is important to change the identifier communicated to the
        users browser in order to prevent session fixation attacks. When this
        happens L{reset_client_id} should be called to associate the session
        data with the new identifier, and disassociate it from the old
        identifier.
        """
        old_client_id = str(old_client_id)
        new_client_id = str(new_client_id)
        store = getUtility(IZStorm).get("session")
        store.execute(
            "UPDATE session_data SET client_id = ? WHERE client_id = ?",
            (new_client_id, old_client_id),
        )
        store.execute(
            "UPDATE session_pkg_data SET client_id = ? WHERE client_id = ?",
            (new_client_id, old_client_id),
        )
        store.commit()


class ILandscapeSession(ISession):
    """Extends ISessionData with a Landscape-specific method."""

    def expire():
        """Expire this session from the database."""


@implementer(ILandscapeSession)
class LandscapeSession(Session):
    def expire(self):
        session_data_container = self._sdc("")
        session_data = session_data_container[self.client_id]
        session_data.expire()


@implementer(ISessionData)
class SessionData:
    def __init__(self, client_id, generation):
        self.client_id = client_id
        self.generation = generation

    def get(self, product_id, default):
        return self[product_id]

    def __getitem__(self, product_id):
        """
        Return a SessionPkgData.
        """
        return SessionPkgDataProxy(self.client_id, self.generation, product_id)

    def expire(self):
        """This removes all session keys for this session."""
        store = getUtility(IZStorm).get("session")
        store.find(
            SessionPkgData,
            client_id=self.client_id.encode(),
            generation=self.generation,
        ).remove()


@implementer(ISessionPkgData)
class SessionPkgDataProxy(collections.abc.MutableMapping, object):
    def __init__(self, client_id, generation, product_id):
        self.client_id = client_id.encode()
        self.generation = generation
        self.product_id = product_id.encode()
        self.store = getUtility(IZStorm).get("session")

        self._query_kwargs = {
            "client_id": self.client_id,
            "generation": self.generation,
            "product_id": self.product_id,
        }

    def __getitem__(self, key):
        if isinstance(key, str):
            key = key.encode()

        obj = self.store.find(SessionPkgData, key=key, **self._query_kwargs)

        if obj.is_empty():
            raise KeyError(key)

        return obj.one().data

    def __setitem__(self, key, value):
        if isinstance(key, str):
            key = key.encode()

        obj = self.store.find(SessionPkgData, key=key, **self._query_kwargs)

        if not obj.is_empty():
            obj = obj.one()
            obj.data = value
        else:
            obj = SessionPkgData(
                self.client_id, self.generation, self.product_id, key, value
            )
            self.store.add(obj)

    def __delitem__(self, key):
        if isinstance(key, str):
            key = key.encode()

        obj = self.store.find(SessionPkgData, key=key, **self._query_kwargs)

        if obj.is_empty():
            raise KeyError(key)

        self.store.remove(obj.one())

    def __iter__(self):
        return self.keys()

    def __len__(self):
        return self.store.find(SessionPkgData, **self._query_kwargs).count()

    def keys(self):
        return self.store.find(SessionPkgData, **self._query_kwargs).values(
            SessionPkgData.key
        )


class ISessionTimeoutDirective(Interface):

    hours = TextLine(
        title="Hours",
        description="Number of hours to timeout sessions after",
        required=True,
    )


def session_timeout_directive(_context, hours):
    """
    Implementation of the <session_timeout> directive.
    """
    new_value = int(hours)
    SessionDataContainer.timeout = timedelta(seconds=3600 * new_value)


persistent_data_container = SessionDataContainer()

volatile_data_container = RAMSessionDataContainer()
