# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.cachedescriptors.property import Lazy

from canonical.url import get_url, get_url_object
from canonical.navigation import NavigationItem

from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.selection.interfaces import IComputerSelection
from canonical.landscape.model.account.savedsearch import SavedSearchManagement


class SearchNavigation(NavigationItem):

    name = "systems-management-searches"
    title = "Saved searches"
    description = "Saved searches in this account"
    expanded = True

    def __init__(self, context, request, parent=None):
        super(SearchNavigation, self).__init__(context, request, parent)
        self.account = get_url_object(self.context, request, IAccount)

    @Lazy
    def searches(self):
        """
        Fetch a list of searches associated with the current account.
        """
        saved_search_management = SavedSearchManagement(self.account)
        return list(saved_search_management.get_saved_searches())

    def get_items(self):
        """
        Dynamically generate L{SearchNavigationItem} items for each search.
        """
        return [
            SearchNavigationItem(self.context, self.request, self, search)
            for search in self.searches
        ]

    def is_available(self):
        """
        L{SearchNavigation} is only available if the user can see computers.
        """
        person = self.request.principal.person
        computer_ids = person.get_computer_ids(self.account)
        if computer_ids is None or computer_ids:
            return len(self.searches) != 0
        else:
            # computer_ids is [] if the person doesn't have access to any
            # computers in the account.  In this case we hide the 'Search'
            # item.
            return False


class SearchNavigationItem(NavigationItem):

    order = 1000

    def __init__(self, context, request, parent, search):
        super(SearchNavigationItem, self).__init__(context, request, parent)
        self.saved_search = search

    @property
    def title(self):
        return "%s" % self.saved_search.title

    @property
    def description(self):
        if not self.is_selected():
            return "Find %s computers matching search '%s'" % (
                self.parent.account.title,
                self.saved_search.search,
            )
        return "See all computers in %s" % self.parent.account.title

    @property
    def action(self):
        if not self.is_selected():
            return "%s/computers/search/%s" % (
                get_url(self.parent.account, self.request),
                self.saved_search.name,
            )
        return "%s/computers" % get_url(self.parent.account, self.request)

    def is_selected(self):
        """Selected if the search is included in the computer selection."""
        selection = get_url_object(self.context, self.request, IComputerSelection)
        if selection is not None:
            return selection.get_search_name() == self.saved_search.name
        return False
