# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.cachedescriptors.property import Lazy

from canonical.content.provider import TemplateContentProvider
from canonical.navigation import Navigation, NavigationItem
from canonical.url import get_url, get_url_object, quote_url

from canonical.landscape.model.account.management import AccountManagement
from canonical.landscape.model.main.interfaces import IAccount
from canonical.landscape.model.selection.interfaces import IComputerSelection


class SystemsManagementSidebar(TemplateContentProvider):
    """Content provider for rendering the computers sidebar."""

    template_filename = "systems.pt"

    @Lazy
    def available_count(self):
        """Return the count of available computers."""
        selection = get_url_object(self.context, self.request, "selection")
        return selection.get_available_count()


class SystemsManagementNavigation(Navigation):

    name = "systems-management-sidebar"
    horizontal = False


class AccessGroupsNavigation(NavigationItem):

    name = "systems-management-access-groups"
    title = "Access Groups"
    description = "Access groups associated with this account"
    expanded = True

    def __init__(self, context, request, parent=None):
        super(AccessGroupsNavigation, self).__init__(context, request, parent)
        self.person = self.request.principal.person
        self.account = get_url_object(self.context, request, IAccount)
        self.management = AccountManagement(self.account, self.person)

    @Lazy
    def _access_groups(self):
        """
        Fetch a list of (access_group, count) tuples visible by the current
        person.
        """
        return self.management.get_access_contexts_with_counts()

    def get_items(self):
        """
        Dynamically generate L{AccessGroupNavigationItem} items for each
        L{AccessContext}.
        """
        return [
            AccessGroupNavigationItem(
                self.context,
                self.request,
                self,
                group["access_context"],
                group["computers_count"],
            )
            for group in self._access_groups
        ]

    def is_available(self):
        """
        L{AccessGroupsNavigation} is only available if the user can see
        access groups.
        """
        return len(self._access_groups) != 0


class AccessGroupNavigationItem(NavigationItem):

    order = 1000

    def __init__(self, context, request, parent, access_group, count):
        super(AccessGroupNavigationItem, self).__init__(context, request, parent)
        self.access_group = access_group
        self.count = count

    @property
    def title(self):
        return "%s (%s)" % (self.access_group.name, self.count)

    @property
    def description(self):
        if not self.is_selected():
            return "Find %s computers in the '%s' access group" % (
                self.parent.account.title,
                self.access_group.name,
            )
        return "See all computers in %s" % self.parent.account.title

    @property
    def action(self):
        if not self.is_selected():
            return "%s/computers/accessgroups/%s" % (
                get_url(self.parent.account, self.request),
                quote_url(self.access_group.name),
            )
        return "%s/computers" % get_url(self.parent.account, self.request)

    def is_selected(self):
        """
        Selected if the access_group is included in the computer selection.
        """
        selection = get_url_object(self.context, self.request, IComputerSelection)
        if selection is not None:
            return (
                self.access_group in selection.get_included_access_groups()
                and selection.account == self.parent.account
            )
        return False


class TagsNavigation(NavigationItem):

    name = "systems-management-tags"
    title = "Tags"
    description = "Tags associated with this account"
    expanded = True

    def __init__(self, context, request, parent=None):
        super(TagsNavigation, self).__init__(context, request, parent)
        self.person = self.request.principal.person
        self.account = get_url_object(self.context, request, IAccount)
        self.management = AccountManagement(self.account, self.person)

    @Lazy
    def _tags(self):
        """
        Fetch a list of (tag, count) tuples visible by the current person.
        """
        return self.management.get_tags_with_counts()

    def get_items(self):
        """Dynamically generate L{TagNavigationItem} items for each tag."""
        return [
            TagNavigationItem(
                self.context, self.request, self, row["tag"], row["computers_count"]
            )
            for row in self._tags
        ]

    def is_available(self):
        """
        L{TagsNavigation} is only available if the user can see tags.
        """
        return len(self._tags) != 0


class TagNavigationItem(NavigationItem):

    order = 1000

    def __init__(self, context, request, parent, tag, count):
        super(TagNavigationItem, self).__init__(context, request, parent)
        self.tag = tag
        self.count = count

    @property
    def title(self):
        return "%s (%s)" % (self.tag.name, self.count)

    @property
    def description(self):
        if not self.is_selected():
            return "Find %s computers with the '%s' tag" % (
                self.parent.account.title,
                self.tag.name,
            )
        return "See all computers in %s" % self.parent.account.title

    @property
    def action(self):
        if not self.is_selected():
            return "%s/computers/tags/%s" % (
                get_url(self.parent.account, self.request),
                quote_url(self.tag.name),
            )
        return "%s/computers" % get_url(self.parent.account, self.request)

    def is_selected(self):
        """Selected if the tag is included in the computer selection."""
        selection = get_url_object(self.context, self.request, IComputerSelection)
        if selection is not None:
            return (
                self.tag in selection.get_included_tags()
                and selection.account == self.parent.account
            )
        return False
