from zope.cachedescriptors.property import Lazy

from canonical.url import get_context_url, get_url

from canonical.landscape.ui.activity.schedule import (
    get_deliver_after_time,
    get_deliver_delay_window,
)
from canonical.landscape.ui.toolbar.page import ToolbarPage
from canonical.landscape.model.activity.schedule import InvalidScheduleAfterWindowError
from canonical.landscape.model.snap.management import SNAP_ACTIONS


class SnapDashboardPage(ToolbarPage):

    title = "Snap management"
    paginate = False

    @Lazy
    def counts_by_computer(self):
        """For the first 5 selected computers, count snaps in each state."""
        if self.context.has_unsupported_computers:
            self.notes.warning(
                "Some of the selected machines are incompatible with Landscape Snap Monitoring and"
                " will not appear below. Please upgrade Landscape Client on these machines."
            )

        return self.context.get_snap_counts_by_computer(offset=0, limit=5)

    def apply_changes(self):
        """
        Bundles up the POSTed snaps by the action type provided and
        creates a group activity for the selected computers.
        """
        if self.request.method != "POST":
            return self()

        self._check_request_csrf_token()
        error_redirect_url = get_url(self.context, self.request)

        # If none of the selected computers supports snap management,
        # there's no point in continuing.
        if not self.context.has_supported_computers:
            self.notes.error(
                "None of the selected machines is compatible with Landscape Snap Management. Please"
                " upgrade Landscape Client on these machines."
            )
            self.request.response.redirect(error_redirect_url)
            return

        account = self.context.selection.account
        person = self.request.principal.person

        deliver_after = get_deliver_after_time(self.request)
        deliver_delay_window = get_deliver_delay_window(self.request)
        changes = self._convert_form_to_snap_changes()

        if not changes:
            # No supported SNAP_ACTIONS had snaps provided.
            self.notes.error("No changes selected")
            self.request.response.redirect(error_redirect_url)
            return

        try:
            activity = self.context.change_snaps(
                person,
                changes,
                deliver_after=deliver_after,
                deliver_after_window=deliver_delay_window,
            )
        except InvalidScheduleAfterWindowError as e:
            self.notes.error(str(e))
            self.request.response.redirect(error_redirect_url)
            return

        if self.context.has_unsupported_computers:
            self.notes.warning(
                "Some of the selected clients are incompatible with Landscape Snap Management and "
                "have not had changes queued. Please upgrade Landscape Client on these machines."
            )

        self.notes.info("Snap changes queued")
        account_url = get_url(account, self.request)
        activity_path = get_context_url(activity, self.request).path
        self.request.response.redirect(f"{account_url}/{activity_path}")

    def _convert_form_to_snap_changes(self):
        """
        Converts form arguments to the format consumed by a snap
        Activity. That is, converts a dict of form:
        {
            "snapname1": "refresh",
            "snapname1-channel": "latest/beta",
            ...
            "snapname2": "remove",
            ...
        }
        to one of form:
        {
            "refresh": [
                {
                    "name": "snapname1",
                    "args": {
                        "channel": "latest/beta",
                        ...
                    }
                },
                ...
            ],
            "remove": [
                {"name": "snapname2"},
                ...
            ],
            ...
        }
        Probably took more lines of comment than there will be of code.
        """
        changes = {}
        snaps = {}

        for k, v in self.request.form.items():
            fields = k.split("_")
            snap_name = fields[0]
            data_type = fields[1] if len(fields) > 1 else "action"

            if snap_name in snaps:
                snaps[snap_name][data_type] = v
            else:
                snaps[snap_name] = {data_type: v}

        for snap, snap_args in snaps.items():
            if "action" not in snap_args:
                continue

            action = snap_args.pop("action")

            if action not in SNAP_ACTIONS:
                continue

            snap_args.update({"name": snap})

            if action in changes:
                changes[action].append(snap_args)
            else:
                changes[action] = [snap_args]

        return changes
