from datetime import datetime

from pytz import UTC, timezone

from canonical.landscape.ui.snap.dashboard import SnapDashboardPage


class SnapInstallPage(SnapDashboardPage):
    """
    A specialization of SnapDashboardPage that uses a different form
    structure specific to install actions.
    """

    title = "Install Snaps"

    def __init__(self, context, request):
        super().__init__(context, request)

        tz = timezone(request.principal.person.timezone)
        utc_now = datetime.now()
        utc_target = UTC.localize(utc_now)
        local_target = utc_target.astimezone(tz)
        self.at_time = local_target.strftime("%Y-%m-%d %H:%M")

    def _convert_form_to_snap_changes(self):
        """
        This form comes in essentially as:
        {
            "name_0": "snapname",
            "channel_0": "latest/stable",
            "revision_0": "123",
            ...
        }
        through to name_n, channel_n, revision_n.
        We want it as:
        {
            "install": [
                {
                    "name": "snapname",
                    "args": {
                        "channel": "latest/stable",
                        "revision": "123",
                    }
                },
                ...
            ]
        }
        We filter out empty values because they might have slipped by
        front-end validation and for 'args' they are not required.
        """
        snaps = {}

        for k, v in self.request.form.items():
            fields = k.split("_")

            if len(fields) < 2:
                continue

            field = fields[0]
            snap_number = fields[1]

            if field == "classic":
                v = v == "on"

            if snap_number not in snaps:
                snaps[snap_number] = {}

            snap = snaps[snap_number]

            if field == "name" and v:
                snap["name"] = v
            elif v and "args" in snap:
                snap["args"][field] = v
            elif v:
                snap["args"] = {field: v}

        if not snaps:
            return {}

        return {"install": list(snaps.values())}
