from datetime import datetime

from zope.cachedescriptors.property import Lazy

from canonical.landscape.model.snap.management import SNAP_ACTIONS
from canonical.landscape.ui.toolbar.page import ToolbarPage

DEFAULT_PAGE_SIZE = 25
DEFAULT_OFFSET = 0


class SnapListingPage(ToolbarPage):
    title = "Snap Listing"

    def __init__(self, context, request):
        super().__init__(context, request)

        self.total_snaps = 0
        self.snap_offset = 0
        self.snap_data = []
        self.has_snaps = False
        self.snap_filter = self.request.get("filter")
        self.page_size = self._get_int_param("size", DEFAULT_PAGE_SIZE)
        self.offset = self._get_int_param("offset", DEFAULT_OFFSET)
        self.text = self._get_text()

        if self.page_size <= 0:
            self.notes.error(f"Invalid listing size: {self.size}")
            return

        if not any([self.text, self.snap_filter]):
            self.notes.error("No search terms provided.")
            return

        self.snap_data = self._search_results()

        if self.snap_data:
            self.has_snaps = True

        self.at_time = self.localize_time(datetime.utcnow())

    @Lazy
    def snap_actions(self):
        return SNAP_ACTIONS

    @Lazy
    def snap_store_url(self):
        return "https://snapstore.io"

    def _get_int_param(self, param, default):
        """Parses param with name `param` as an int, returning `default` if it's unparseable."""
        result = self.request.get(param)

        if not result:
            return default

        try:
            return int(result)
        except ValueError:
            return default

    def _get_text(self):
        """Parses the text queryparameter, returning None if it's empty."""
        text = self.request.get("q", "").strip()

        if text:
            return text

        return None

    def _search_results(self):
        """Queries the local snap database for snaps matching the filter criteria."""
        return self.context.search_snaps(
            status=self.snap_filter,
            offset=self.offset,
            limit=self.page_size,
            text=self.text,
        )
