# Copyright 2005-2022 Canonical Limited. All rights reserved.

from urllib.parse import urlencode

from zope.cachedescriptors.property import Lazy

from canonical.landscape.ui.sidebar.page import SidebarPage
from canonical.landscape.model.selection.interfaces import IComputerSelection


class ToolbarPage(SidebarPage):
    """A page with the legacy toolbar area."""

    status = ""

    def toolbar_class(self):
        if IComputerSelection.providedBy(self.context):
            return "toolbar-item-disabled"


class ComputerPaginatorPage(ToolbarPage):
    """
    A toolbar page which allows referencing a single computer in the current
    selection through its index.

    This allows to create detail pages for a computer in the selection context
    with pagination links to move through computers in the selection.
    """

    index_param = "index"
    max_per_page = 1

    @Lazy
    def max_index(self):
        """The maximum index of the selection."""
        return self.context.selection.get_selected_count()

    @Lazy
    def is_single_computer(self):
        """Whether the selection includes a single computer."""
        return self.max_index == 1

    @Lazy
    def current_index(self):
        """The current index in the pagination."""
        try:
            index = int(self.request.get(self.index_param))
        except (ValueError, TypeError):
            index = 1

        if index > self.max_index:
            index = self.max_index
        if index < 1:
            index = 1
        return index

    @Lazy
    def max_page_index(self):
        """The maximum selection index in the page."""
        max_displayable_index = self.current_index + self.max_per_page - 1
        return min(max_displayable_index, self.max_index)

    @Lazy
    def display_single(self):
        """Whether the page displays a single computer."""
        return self.current_index == self.max_page_index

    @Lazy
    def computers(self):
        """Return the computers displayed in the page."""
        return self.context.selection.get_selected(
            offset=self.current_index - 1, limit=self.max_per_page
        )

    @Lazy
    def computer(self):
        """Return the first computer in the selection slice."""
        return self.computers[0] if self.computers else None

    @Lazy
    def computer_ids(self):
        """Return Ids for L{Computer}s in the current selection slice."""
        return [computer.id for computer in self.computers]

    def _update_query_string(self, params):
        """Return an updated query string with the specified parameters."""
        query = dict(self.request.form) if self.request.method == "GET" else {}
        for key, value in list(params.items()):
            # Remove null values
            if value is None:
                query.pop(key, None)
            else:
                query[key] = value
        # Encode all params as UTF-8
        for key, value in list(query.items()):
            if isinstance(value, list):
                value = [str(elem).encode("utf8") for elem in value]
            else:
                value = str(value).encode("utf8")
            query[key] = value
        return urlencode(sorted(query.items()), doseq=True)

    def _get_index_url(self, index):
        query_string = self._update_query_string({self.index_param: index})
        return "?" + query_string if query_string else ""

    def get_next_url(self):
        index = min(self.current_index + self.max_per_page, self.max_index)
        return self._get_index_url(index if index > self.max_page_index else None)

    def get_prev_url(self):
        return self._get_index_url(
            max(self.current_index - self.max_per_page, 1)
            if self.current_index > 1
            else None
        )

    def get_first_url(self):
        return self._get_index_url(1 if self.current_index > 1 else None)

    def get_last_url(self):
        next_index = self.current_index + self.max_per_page
        remainder = self.max_index % self.max_per_page
        last_slice_index = (
            (self.max_index - remainder + 1)
            if remainder
            else (self.max_index - self.max_per_page + 1)
        )
        return self._get_index_url(
            last_slice_index if last_slice_index >= next_index else None
        )
