# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os

from landscape.lib.fs import read_file

from canonical.landscape.model.configuration import (
    get_static_root_path,
    is_development_mode,
)
from canonical.landscape.ui.constants import STATIC_DIR


__all__ = ["static_url", "get_debug", "read_asset"]

_DEBUG = {
    key: int(os.getenv("DEBUG_" + key.upper(), 0))
    for key in ("css", "js", "yui", "log")
}

_COMBO_PREFIX = "/combo?"


def static_url(context, request):
    """Format a path as relative to the static root."""
    url = context.replace(" ", "").replace("\n", "")  # For multi-line values
    if url.startswith(_COMBO_PREFIX):
        _check_debug_flags(request)

        parts = url[len(_COMBO_PREFIX) :].split("&")
        for i, part in enumerate(parts):
            for key, debug in list(_DEBUG.items()):
                ext = "." + key
                if part.endswith(ext) and not debug:
                    parts[i] = part[: -len(ext)] + "-min" + ext

        url = _COMBO_PREFIX + "&".join(parts)

    return get_static_root_path() + url


def _check_debug_flags(request):
    """Check if the request contains debug flags.

    This function looks for a C{debug_<type>} values in the request for
    supported types and enable/disable debugging accordingly.

    Use C{debug_<type>=0} to disable specific debugging and C{debug_<type>=1}
    to enable it.  The default setting for this parameter can be defined using
    the C{DEBUG_<TYPE>} environment variable.

    """
    # Allow a request variable to toggle global state if we are using the
    # "default" Landscape configuration (i.e., we are in development mode).
    if is_development_mode():
        for key in _DEBUG:
            debug = request.get("debug_" + key)
            if debug is not None:
                _DEBUG[key] = int(debug)


def get_debug(key):
    """Return whether debug is enabled for the specified asset key."""
    return bool(_DEBUG[key])


def reset_debug():
    """Reset debug for all keys turning it off."""
    for key in _DEBUG:
        _DEBUG[key] = 0


def read_asset(asset):
    """Return the content of a static asset file."""
    return read_file(os.path.join(STATIC_DIR, asset))
