# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import base64

from zope.cachedescriptors.property import Lazy
from zope.formlib.form import FormField
from zope.schema import List
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.toolbar.page import ToolbarPage
from canonical.landscape.ui.users.interfaces import IEditUser
from canonical.landscape.ui.users.optional import OptionalWidget, OptionalDropdownWidget
from canonical.landscape.ui.users.passthrough import HiddenPassThroughList


class EditUserPage(ToolbarPage):

    requires = "landscape.EditUser"

    def __init__(self, context, request):
        super(EditUserPage, self).__init__(context, request)
        self._groups = None

    @Lazy
    def title(self):
        if len(self.context.users) > 1:
            return "Edit Users"
        return "Edit User"

    def update_groups(self):
        """Handle a form submission for addition or removal of groups."""
        self._check_request_csrf_token()

        groups = self.request["group.groups"]
        if not groups:
            self.notes.error(
                "At least one group must be selected before "
                "group operations can be performed."
            )
        else:
            groups = groups.split()
            if "group.add" in self.request:
                self.context.add_to_groups(groups)
                self.notes.info(
                    "Group %s will be updated." % ", ".join(sorted(set(groups)))
                )
            elif "group.remove" in self.request:
                self.context.remove_from_groups(groups)
                self.notes.info(
                    "Group %s will be updated." % ", ".join(sorted(set(groups)))
                )

        url = "%s/edit" % get_url(self.context, self.request)
        self.request.response.redirect(url)

    def groups(self):
        """
        Generates a list of dicts, one per group.  The data is used by
        the template to generate the inline listing that displays
        groups.  The format of each dict is:

        {'name': <group name>,
         'membership': <'Full'|'Partial'|'None'>
         'membership_tooltip': <detailed membership description>}

        Membership is 'None' when no selected users are part of the
        group on selected computers (which is sometimes true because
        the group doesn't exist on computer(s)).  It's 'Partial' when
        selected users are on members of the group on some machines
        (which, again, can involve a group not being on a computer).

        The algorithm implemented here iterates over 3-tuples of
        (computer, group name, username) and generates a weight for
        each group.  The weight is compared with the weight for a
        perfect match to determine membership status.
        """
        if self._groups is not None:
            return self._groups

        if self.request and "index" in self.request:
            index = int(self.request.get("index"))
            groupnames = self.context.all_groups(index)
            group_data = self.context.get_computer_group_user_tuples(index)
        else:
            # For backward compatibility with tests that monkey-patch these
            # methods
            groupnames = self.context.all_groups()
            group_data = self.context.get_computer_group_user_tuples()

        index = {}
        for groupname in groupnames:
            index[groupname] = {"weight": 0, "tooltips": []}
        for computer, groupname, username in group_data:
            data = index[groupname]
            data["weight"] += 1
            data["tooltips"].append("%s on %s" % (username, computer.title))

        computer_count = self.context.management.selection.get_selected_count()
        user_count = len(set([user.username for user in self.context.users]))
        perfect_weight = computer_count * user_count

        self._groups = []
        for groupname in sorted(index.keys()):
            data = index[groupname]
            weight = data["weight"]
            status = self._calculate_membership(perfect_weight, weight)
            data["tooltips"].sort()
            tooltip = ", ".join(data["tooltips"])
            membership = {
                "name": groupname,
                "membership": status,
                "membership_tooltip": tooltip,
            }
            self._groups.append(membership)
        self._groups.sort(key=self._sort_key)
        return self._groups

    def _sort_key(self, item):
        sort_order = ["Full", "Partial", "None"]
        membership = sort_order.index(item["membership"])
        return (membership, item["name"])

    def _calculate_membership(self, perfect_weight, weight):
        """Returns a 'Full', 'Partial' or 'None' string indicating membership.

        @param perfect_weight: The score that represents a complete
            match.
        @param weight: The actual score.
        """
        assert perfect_weight >= weight
        if perfect_weight == weight:
            return "Full"
        elif weight:
            return "Partial"
        return "None"

    def _get_computer_text(self, computers):
        if len(computers) > 1:
            computer_count = "%d computers" % len(computers)
        else:
            computer_count = "1 computer"
        computer_tooltip = ", ".join(sorted([c.title for c in computers]))
        return computer_count, computer_tooltip


def user_passthrough(field, request):
    return HiddenPassThroughList(field, request, "users", "users")


class EditUserForm(LandscapeFormProvider):

    prefix = "edit-user"

    summary = "User details"
    fields = Fields(
        FormField(
            List(__name__="users", title="users"),
            custom_widget=user_passthrough,
            name="users",
        ),
        IEditUser,
    ).hide("users")

    for field_name in ("name", "passphrase", "location", "work_phone", "home_phone"):
        fields[field_name].custom_widget = OptionalWidget
    fields["primary_group"].custom_widget = OptionalDropdownWidget

    def get_initial_data(self):
        """
        Get the initial data for the widgets in this form.

        For any fields which are the same across all users in the
        current user selection, their value will be the value of that
        field. For any fields which have different values, there will
        be no default value.
        """
        user_group_names = get_unique_primary_group_names(self.context)
        # If the selected users have more than one primary group between them,
        # then we use "(Various)", otherwise, we only have one groupname, so we
        # can use that.
        if len(user_group_names) > 1:
            primary_group = "(Various)"
        elif len(user_group_names) == 0:
            primary_group = ""
        else:
            primary_group = list(user_group_names)[0]

        user = self.context.users[0]
        data = {
            "name": user.name,
            "location": user.location,
            "home_phone": user.home_phone,
            "work_phone": user.work_phone,
            "primary_group": primary_group,
        }
        for user in self.context.users:
            for attr in ("name", "location", "home_phone", "work_phone"):
                if getattr(user, attr) != data[attr]:
                    data[attr] = ""
        return data

    def validate(self, action, data):
        del data["users"]
        if data["primary_group"] == "(Various)":
            self.add_error("Please select a valid group.", field_name="primary_group")
        if not [_f for _f in list(data.values()) if _f]:
            self.add_error("No changes!")

    @action("Save", primary=True)
    def handle_save_action(self, action, data):
        self.context.edit_users(
            primary_groupname=data["primary_group"],
            name=data["name"],
            passphrase=data["passphrase"],
            location=data["location"],
            work_phone=data["work_phone"],
            home_phone=data["home_phone"],
        )
        usernames = sorted(set(u.username for u in self.context.users))
        self.view.notes.info("User %s will be edited." % ", ".join(usernames))
        url = get_url(self.context.management, self.request)
        self.request.response.redirect(url)


def get_unique_primary_group_names(user_selection):
    """
    Return a set of primary group names for all users in
    C{user_selection}.
    """
    # We need to get the unique list of all groups that users are in.
    selected_user_groups = [
        user.computer.get_group_by_gid(user.primary_gid)
        for user in user_selection.users
        if user.primary_gid is not None
    ]
    return set([group.name for group in selected_user_groups if group])


def get_groupnames(user_selection):
    # Get all groups on all computers, for all selected users.
    groupnames = [group for group in user_selection.all_groups()]
    groupnames.sort()
    groups = [
        (
            SimpleTerm(
                groupname, base64.b64encode(groupname.encode("utf-8")), title=groupname
            )
        )
        for groupname in groupnames
    ]

    user_group_names = get_unique_primary_group_names(user_selection)

    # If the users are members of more than one primary group, then we put
    # (Various) as the top item.
    if len(user_group_names) > 1:
        groups.insert(0, SimpleTerm("(Various)"))

    return SimpleVocabulary(groups)
