# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.session.interfaces import ISession

from canonical.routes import route, Redirect
from canonical.url import get_url, get_url_object

from canonical.landscape.model.users.users import UserSelection
from safe.canonical.landscape.model.main.computer import get_computer_by_id
from canonical.landscape.ui.page.page import check_request_csrf_token
from canonical.landscape.ui.routes import LandscapeRoutes
from canonical.landscape.ui.users.users import get_user_from_descriptor


LOCATIONS = {
    "add": "add",
    "delete": "selection/delete",
    "edit": "selection/edit",
    "update-groups": "selection/update-groups",
    "lock": "selection/lock",
    "unlock": "selection/unlock",
}


class UserRoutes(LandscapeRoutes):
    def update_session(self):
        """
        Processes the HTTP arguments specifying users to perform
        actions on and stores them in the session.
        """
        # NOTE: There is a tricky situation here involving the edge
        # case of no users selected.

        # When no users are selected, self.request["users"] is simply
        # not set, so this method will NOT clear existing
        # user-selection data from the session, causing strange
        # behavior where clicking on a button with no users visibly
        # selected will actually perform an operation on some
        # previously-selected users. Since this method can't really do
        # anything about that, the User Management page, upon
        # rendering, clears the user data from the session: see
        # page.UsersPage._clear_session_data.

        # In the special case that an aggregate computer row is
        # selected, the user descriptor will be encoded with the names
        # of all of the users that are effected by the aggregate
        # operation.  The format of the encoding is
        # 'NAME:ID1:ID2:...' where ID is the computer id and NAME
        # is the username.

        if "users" in self.request:
            if not isinstance(self.request["users"], list):
                pre_users = [self.request["users"]]
            else:
                pre_users = self.request["users"]

            users = set()
            for user_descriptor in pre_users:
                if user_descriptor != "":
                    if ":" in user_descriptor:
                        all_descriptors = user_descriptor.split(":")
                        user_name = all_descriptors.pop(0)
                        for computer_id in all_descriptors:
                            users.add("%s,%s" % (computer_id, user_name))
                    else:
                        users.add(user_descriptor)
            users = sorted(users)

            session = ISession(self.request)["users"]
            session["users"] = users

    @route("action")
    def user_action(self):
        """
        Route for handling the user listing screen actions: processes
        the form and adds selected users to the session (with
        L{update_session}) and redirects to the appropriate location
        based on which button the user pressed.
        """
        # This is very sad, but the only thing we can get at is the
        # value of the submit button, so we unfortunately have to use
        # this view-level value as a key to find out what the user
        # wanted to do.
        check_request_csrf_token(self.request, require_post=True)

        action_title = self.request.get("action")
        if action_title is None:
            raise Redirect(self.request.getURL(1))
        action = action_title.rsplit(None, 1)[0].replace(" ", "_").lower()
        action = LOCATIONS[action]
        self.update_session()
        index = self.request.get("index")
        raise Redirect("%s/%s?index=%d" % (self.request.getURL(1), action, index))

    @route("selection")
    def selection(self):
        """
        Get a UserSelection based on session data (see L{user_action})
        or request args. This has to handle both cases because the
        both the forms and the form handlers for user management
        actions work on the UserSelection; when rendering the form,
        the data from the session will be used, and when handling the
        action, user data will be in the request.
        """
        check_request_csrf_token(self.request)

        if "users" in self.request:
            user_descriptors = self.request["users"]
            if not isinstance(user_descriptors, list):
                user_descriptors = [user_descriptors]
        else:
            session = ISession(self.request)["users"]
            user_descriptors = session.get("users", ())

        computer_selection = get_url_object(self.context, self.request, "selection")
        url = get_url(computer_selection, self.request)
        if not user_descriptors:
            self.notes.error(
                "Users must be selected before user actions " "can be performed."
            )
            raise Redirect(url + "/users")
        users = list(map(get_user_from_descriptor, user_descriptors))
        missing_users = users.count(None)
        if missing_users:
            if missing_users == len(user_descriptors):
                if len(user_descriptors) == 1:
                    computer_id, username = user_descriptors[0].split(",", 1)
                    computer = get_computer_by_id(int(computer_id))
                    msg = "User '%s' doesn't exist anymore on computer " "'%s'." % (
                        username,
                        computer.title,
                    )
                else:
                    msg = "Selected users don't exist anymore."
                self.notes.error(msg)
                raise Redirect(url + "/users")
            else:
                if missing_users == 1:
                    index = users.index(None)
                    computer_id, username = user_descriptors[index].split(",", 1)
                    computer = get_computer_by_id(int(computer_id))
                    msg = "User '%s' doesn't exist anymore on computer " "'%s'." % (
                        username,
                        computer.title,
                    )
                else:
                    msg = "%d selected users don't exist anymore." % (missing_users,)
                self.notes.warning(msg)
                users = set(users)
                users.remove(None)
        users = sorted(set(users), key=lambda u: (u.username, u.computer.title))
        return UserSelection(self.context, users)
