# Copyright 2005-2014 Canonical Limited.  All rights reserved.

import logging
import os
import tempfile
from email.encoders import encode_base64
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart

from treq import request, content

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.error import (
    ConnectError,
    ConnectingCancelledError,
    DNSLookupError,
)

from canonical.launchpad.errors import UnableToUploadBlob, UnexpectedResponse


@inlineCallbacks
def upload_directory_to_blob_store(blob_store, path, tags=None, bug_title=None):
    """Upload the files in a directory as a Launchpad blob.

    @param blob_store: A BlobStoreAPI implementation.
    @param path: The path to the directory to upload.
    @param tags: Any tags the bug report should get, when filed using
        the returned token
    @param bug_title: The title the bug report should get, when filed
        using the returned token.

    A mime/multipart message will be created containing all the files in
    the directory and uploaded to Launchpad. All files are expected to
    be text files.

    The returned token should be appended to the +filebug path, in
    order for the files to be attached automatically. I.e.:

        https://launchpad.net/landscape/+filebug/$token
    """
    fd, temp_filename = tempfile.mkstemp()
    # XXX: Using the email package to construct the mime/multipart
    #     message isn't a good idea for large files. To avoid memory spikes,
    #     we should stream the files ourselves the temporary file.
    msg = MIMEMultipart()
    if tags is not None:
        msg["Tags"] = ",".join(tags)
    if bug_title is not None:
        msg["Subject"] = bug_title
    for filename in sorted(os.listdir(path)):
        with open(os.path.join(path, filename), "r") as directory_file:
            contents = directory_file.read()
        if len(contents) == 0:
            # Launchpad breaks if empty files are included.
            continue
        attachment = MIMEBase("text", "plain")
        attachment.add_header("Content-Disposition", "attachment", filename=filename)
        with open(os.path.join(path, filename), "r") as directory_file:
            attachment.set_payload(contents)
        encode_base64(attachment)
        msg.attach(attachment)
    with os.fdopen(fd, "w") as log_files:
        log_files.write(msg.as_string())
        log_files.write("\n")
    try:
        token = yield blob_store.upload_file(temp_filename)
    except (ConnectError, ConnectingCancelledError):
        logging.warning("Couldn't connect to launchpad.net")
        raise UnableToUploadBlob()
    except DNSLookupError:
        logging.warning("DNS lookup failed for launchpad.net")
        raise UnableToUploadBlob()
    except UnexpectedResponse:
        # We log the unexpected response in upload_file().
        raise UnableToUploadBlob()
    except:
        # Catch all exceptions, since we don't want unexpected exceptions
        # break error reporting. For expected errors (i.e. where we
        # can't access launchpad.net, a warning is logged. For
        # unexpected errors, the exception is logged, so that we'll
        # (hopefully) notice it and fix it..
        logging.exception(
            "Unexpected error when POSTing to " "https://launchpad.net/+storeblob"
        )
        raise UnableToUploadBlob()
    finally:
        os.remove(temp_filename)
    returnValue(token)


class BlobStoreAPI(object):
    """An API class for talking to the Launchpad blob store.

    The blob store is used to upload files that can later be
    automatically attached to a bug report.

    @ivar request: A function for issuing a request. Defaults to
        treq.request, but can be overridden in tests.
    """

    def __init__(self, request=request):
        self.request = request

    @inlineCallbacks
    def upload_file(self, filepath):
        """Upload the file and return the token for accessing it later.

        @param filepath: The path to a multipart/mime file that contains
            the files to be attached to a bug report.

        The returned token should be appended to the +filebug path, in
        order for the files to be attached automatically. I.e.:

            https://launchpad.net/landscape/+filebug/$token
        """
        files = {
            "field.blob": (
                "landscape-logs",
                "application/octet-stream",
                open(filepath, "r"),
            )
        }
        # The 10 second timeout is arbitrary. We could raise or lower
        # it as needed. It's there to avoid this call from hanging
        # forever if there's a firewall in place dropping the request.
        response = yield self.request(
            method="POST",
            url="https://launchpad.net/+storeblob",
            data={"FORM_SUBMIT": "1"},
            files=files,
            timeout=10,
        )

        if response.code != 200:
            error_message = yield content(response)
            logging.warning(
                "Unexpected response from +storeblob: %s: %s"
                % (response.code, error_message)
            )
            raise UnexpectedResponse(response)

        if not response.headers.getRawHeaders("X-Launchpad-Blob-Token"):
            logging.warning("No token found in response from +storeblob")
            raise UnexpectedResponse(response)

        returnValue(response.headers.getRawHeaders("X-Launchpad-Blob-Token")[0])
