"""Create ar archives."""

import gzip
import time
import hashlib
import os

from datetime import datetime

from io import BytesIO, StringIO
from subprocess import Popen, PIPE

from gnupg import GPG

import six
import tarfile


class FileInfo(object):
    """Abstract a file-system node, either file or directory."""

    def __init__(self, path, data=None, mode=None, uid=0, gid=0, mtime=None):
        """
        @param path: The filesystem path. A path ending with a trailing C{'/'}
            is considered to be a directory, otherwise a file.
        @param data: The data contained in the item.  It can be used only
            if the item is a file.
        @param mode: Unix permissions.
        @param uid: Unix user ID.
        @param gid: Unix group ID.
        @param mtime: Last modification time.
        """
        self.path = path

        if self.is_dir() and data is not None:
            raise RuntimeError("Directories can't have data")
        if not self.is_dir():
            if data is None:
                raise RuntimeError("Files must have data")
            if type(data) is not bytes:
                raise RuntimeError("File data must be bytes, not {}".format(type(data)))

        self.data = data

        self.mode = mode
        if self.mode is None:
            if self.is_dir():
                self.mode = 0o755
            else:
                self.mode = 0o644
        self.uid = uid
        self.gid = gid
        self.mtime = mtime or time.time()
        self.user = "root"
        self.group = "root"

    def is_dir(self):
        """Indicate whether the item is a directory."""
        return self.path.endswith("/")

    def size(self):
        """Return the size of the member data."""
        if self.data is None:
            return 0
        return len(self.data)

    def md5(self):
        """Return the m5sum of the item, if a file."""
        if self.is_dir():
            raise RuntimeError("Directory doesn's have md5")
        md5 = hashlib.md5()
        md5.update(self.data)
        return md5.hexdigest().encode("utf-8")


class ArMember(object):
    """Represent one member in an L{ArFile}.

    Each member of an ar archive will be extracted into a file matching
    the given criteria.
    """

    def __init__(self, file):
        """
        @param file: The L{FileInfo} object this member will archive.
        """
        self._file = file

    def header(self):
        """Return the header for the member as documented in AR(5)."""
        data = []
        data.append(self._pad(self._file.path, 16))
        data.append(self._pad(str(self._file.mtime), 12))
        data.append(self._pad(str(self._file.uid), 6))
        data.append(self._pad(str(self._file.gid), 6))
        data.append(self._pad("%o" % self._file.mode, 8))
        data.append(self._pad(str(self._file.size()), 10))
        data.append("`\n")

        return "".join(data)

    def _pad(self, string, length):
        """Pad a C{string} with trailing spaces up to C{length}."""
        return (string + length * " ")[:length]

    def data(self):
        """Return the data of the member."""
        data = self._file.data

        # Possibly add a trailing newline, because the ar format specifies
        # that the members must have an even number of bytes.
        if (self._file.size() % 2) == 1:
            data += b"\n"

        return data


class ArFile(object):
    """Represent a complete ar archive."""

    def __init__(self, files=[]):
        """
        @param files: Optional a list of L{FileInfo}s objects to be included.
        """
        self._members = [ArMember(file) for file in files]

    def add(self, file):
        """Add a file to the archive"""
        self._members.append(ArMember(file))

    def pack(self):
        """Pack all members together and return an in-memory ar archive."""
        data = []
        data.append(b"!<arch>\n")

        for member in self._members:
            data.append(member.header().encode("utf-8") + member.data())

        return b"".join(data)


class TarMember(object):
    """Represent one member in an L{TarFile}."""

    def __init__(self, item):
        """
        @param file: The L{FileInfo} object this member will archive,
            either a file or a directory.
        """
        self._item = item

    def info(self):
        """Return the C{TarInfo} object for archiving the item."""
        info = tarfile.TarInfo(self._item.path)
        if self._item.is_dir():
            info.type = tarfile.DIRTYPE
        info.mode = self._item.mode
        info.mtime = self._item.mtime
        info.uid = self._item.uid
        info.gid = self._item.gid
        info.uname = self._item.user
        info.gname = self._item.group
        info.size = self._item.size()
        return info

    def data(self):
        """Return the C{BytesIO} object for archiving the item."""
        if self._item.is_dir():
            return None
        else:
            return BytesIO(self._item.data)


class TarFile(object):
    """Represent a complete gzip'ed tar archive."""

    def __init__(self, files=[]):
        self._items = [TarMember(file) for file in files]

    def add(self, item):
        """Add an item to the archive, either file or directory."""
        self._items.append(TarMember(item))

    def pack(self):
        """Return an in-memory gzip'ed tar archive for the given items."""
        buf = BytesIO()
        tar = tarfile.open(fileobj=buf, mode="w", format=tarfile.GNU_FORMAT)
        for item in self._items:
            tar.addfile(item.info(), item.data())
        tar.close()
        out = BytesIO()
        gz = gzip.GzipFile(fileobj=out, mode="wb")
        gz.write(buf.getvalue())
        gz.close()

        return out.getvalue()


class DebFile(object):
    """A Debian package archive."""

    def __init__(self):
        self._format = "2.0"
        self._data = []
        self._fields = []

    def add(self, item):
        """Add an item to the package, either file or directory.

        @param item: A L{FileInfo} object.
        """
        self._data.append(item)

    def size(self):
        """Total size of the items in this L{DebFile}."""
        size = 0
        for item in self._data:
            size += item.size()
        return size

    def field(self, name, value):
        """Set a field in the control file of this L{DebFile}.

        @param name: The field name.
        @param value: The field value.
        """
        self._fields.append((name, value))

    def control(self):
        """
        Return a L{FileInfo} object for the control file of this L{DebFile}.
        """
        entries = []
        for name, value in self._fields:
            entries.append("%s: %s" % (name, value))
        data = "\n".join(entries) + "\n"
        return FileInfo("./control", data=data.encode("utf-8"))

    def md5sums(self):
        """
        Return a L{FileInfo} with md5sums of each file of this L{DebFile}.
        """
        entries = []
        for item in self._data:
            if not item.is_dir():
                entries.append(item.md5() + b"  " + item.path.encode("utf-8"))
        data = b"\n".join(entries) + b"\n"
        return FileInfo("./md5sums", data=data)

    def pack(self):
        """Return the Debian package with the loaded data and control."""

        control = TarFile([FileInfo("./"), self.control(), self.md5sums()])
        data = TarFile(self._data)

        ar = ArFile(
            [
                FileInfo("debian-binary", data=(self._format + "\n").encode("utf-8")),
                FileInfo("control.tar.gz", data=control.pack()),
                FileInfo("data.tar.gz", data=data.pack()),
            ]
        )

        return ar.pack()


def dpkg_info(path):
    """Run C{dpkg -I} and C{dpkg -c} against the given path.

    @return: The concatenated output of the two commands.
    """
    env = {"LC_ALL": "C"}  # The output of dpkg is locale-dependent
    process1 = Popen(["dpkg", "-I", path], stdout=PIPE, env=env)
    process1.wait()
    lines = process1.stdout.readlines()
    # Skipe the size line, because it can vary at every gzip run.
    lines.pop(1)
    output1 = b"".join(lines)

    process2 = Popen(["dpkg", "-c", path], stdout=PIPE)
    process2.wait()
    output2 = process2.stdout.read()
    return output1 + output2


class ControlFile(object):
    """A file with Debian control files syntax."""

    def __init__(self):
        self._current_section = []
        self._sections = [self._current_section]

    def new_section(self):
        """Begins a new section."""
        self._current_section = []
        self._sections.append(self._current_section)

    def add_field(self, name, value):
        """
        Add a field to the current section.

        @param name: field name
        @param value: field name, can be a list for multiline fields
        """
        self._current_section.append((name, value))

    def _format_section(self, section):
        result = []
        for name, value in section:
            if isinstance(value, six.string_types):
                content = value
            else:
                content = "\n ".join(value)
            result.append("%s: %s" % (name, content))
        return "\n".join(result)

    def content(self, signed_with=None, gpg_options={}):
        """
        Return the content of the file.

        @param signed_with: a gpg key id to sign the content.
        @param gpg_options: a dict with gpg configuration options
        """
        if self._sections[0]:
            contents = "\n\n".join(
                self._format_section(section) for section in self._sections if section
            )
            contents += "\n"
        else:
            contents = ""

        if signed_with is not None:
            sign = GPG(**gpg_options).sign(contents, keyid=signed_with)
            contents = str(sign)
        return contents


class ChangesFile(ControlFile):
    """
    A .changes file. for a .deb file

    @param name: package name
    @param version: package version
    @param arch: architecture for package
    @param distro: destination distribution
    @param maintainer: name and email of the package maintainer
    @param section: package section
    @param priority: package priority
    @param source: source package name (defaults to package name)
    @param description: one line description for package
    @param changed_by: name and email of the package creator
    @param changes: package changelog
    """

    def __init__(
        self,
        name,
        version,
        arch,
        distro,
        maintainer,
        section,
        priority,
        source=None,
        description=".",
        changes=".",
        changed_by=None,
        date=None,
    ):
        super(ChangesFile, self).__init__()
        self.name = name
        self.version = version
        self.arch = arch
        self._section = section
        self._priority = priority

        self.add_field("Format", "1.8")
        if date is None:
            date = datetime.now().strftime("%a, %d %b %Y %T %z")
        self.add_field("Date", date)
        self.add_field("Source", source if source is not None else name)
        self.add_field("Binary", name)
        self.add_field("Architecture", arch)
        self.add_field("Version", version)
        self.add_field("Distribution", distro)
        self.add_field("Maintainer", maintainer)
        if changed_by is not None:
            self.add_field("Changed-By", changed_by)
        self.add_field("Changes", ["", changes])
        self.add_field("Description", ["", description])
        self._files = [""]
        self.add_field("Files", self._files)

    def add_file(self, filename, size=None, md5=None):
        """Add a file to the .changes list."""
        if not size:
            size = os.stat(filename).st_size
        if md5 is None:
            with open(filename, "rb") as file:
                md5 = hashlib.md5(file.read()).hexdigest()
        self._files.append(
            "%s %d %s %s %s"
            % (md5, size, self._section, self._priority, os.path.basename(filename))
        )

    @property
    def default_name(self):
        """Default .changes filename."""
        return "%s_%s_%s.changes" % (self.name, self.version, self.arch)
