from inspect import getargspec
import types

from canonical.lib.decorator import merge_function_metadata


class ArgumentReplacer(object):
    """A decorator helper that filters arguments to be passed to a function.

    Create one with the original function and a function to replace
    the arguments, and then call replace_arguments for each call to
    the function.
    """

    def __init__(self, original_function, argument_replacer):
        """
        @param original_function: The function which will be called
            soon. The function must *not* define any * or ** parameters.
        @param argument_replacer: A function which will be called for
            every argument. It will be passed a parameter name and the
            associated argument. It should return the new value.
        """
        self.original_function = original_function
        self.argument_replacer = argument_replacer

        spec = getargspec(original_function)
        self.all_arguments = spec[0]

        if hasattr(original_function, "__self__"):
            # If it's bound, ignore the bound arguments.
            self.all_arguments = self.all_arguments[1:]

    def replace_arguments(self, args, kwargs):
        """Filter some arguments destined to be passed to a function.

        @param args: Original positional arguments.
        @param kwargs: Original keyword arguments.

        @return: new arguments and kwarguments.
        """
        args = list(args)
        kwargs = kwargs.copy()

        for name_index, name in enumerate(self.all_arguments):
            # Ok, we've got the name of the argument. Let's find the value
            # of the argument in our original arguments and replace it
            # whether it's a positional or a keyword argument.
            if name_index < len(args):
                # Must be a positional argument
                value = args[name_index]
                args[name_index] = self.argument_replacer(name, value)
            else:
                # Must be a keyword argument
                if name not in kwargs:
                    # Oh, but it wasn't passed in. Ignore it.
                    continue
                value = kwargs[name]
                kwargs[name] = self.argument_replacer(name, value)
        return args, kwargs


def coerce_arguments(**schemas):
    """A decorator factory which returns a decorator which coerces arguments.

    @param schemas: keyword arguments. The keys should be parameters
        of the original function. The values should be schemas
        (objects with C{coerce} methods that take and return a single
        value) that the arguments for those parameters will be coerced
        with.
    """

    def replacer(name, value):
        if name in schemas:
            return schemas[name].coerce(value)
        return value

    def decorator(original):
        argument_replacer = ArgumentReplacer(original, replacer)

        def replacement(*args, **kwargs):
            new_args, new_kwargs = argument_replacer.replace_arguments(args, kwargs)
            return original(*new_args, **new_kwargs)

        merge_function_metadata(original, replacement)
        return replacement

    return decorator
