import os
import configparser
from io import StringIO


_UNSET = object()  # Marker to indicate that no value was set


class Config(object):
    """Hold configuration parameters."""

    def __init__(self, path=None):
        self._sections = {}
        if path:
            self.load_from_file(path)

    def has_section(self, section):
        """Return true if the configuration has the named section."""
        return self._sections.get(section) is not None

    def get(self, section, key=None, default=_UNSET):
        """Get a configuration value.

        @param section: The section to get the value from.
        @param key: The key in the section to get the value of. If C{None}
            a C{dict} of the all the section's key->value pairs is returned.
        @param default: An optional default value to return if either the
             section or key are missing.
        """
        try:
            if key is None:
                return self._sections[section]
            return self._sections[section][key]
        except KeyError:
            # If we've passed a default then we can return that here.
            # Note, we don't use None as that can be a valid default
            # value.
            if default is _UNSET:
                raise
            return default

    def set(self, section, key, value=None):
        """Set a configuration value.

        @param section: The section to set the key in.
        @param key: The key to set.
        @param value: The value to set, if C{None} then C{key} will be
            used as bulk data for the whole section.
        """
        if value is not None:
            self._sections.setdefault(section, {})[key] = value
        else:
            self._sections[section] = key

    def load(self, data):
        """Load parameters from a configuration content.

        @param data: The content of a configuration file.
        """
        parser = configparser.RawConfigParser()
        parser.readfp(StringIO(data))
        for section in parser.sections():
            self._sections[section] = dict(parser.items(section))

    def load_from_file(self, path):
        """Load parameters from a configuration file.

        @parm path: The path to the configuration file to load.

        @note: The configuration file can also contain an 'include' directive
           in its first line for importing parameters from another file.
        """
        with open(path) as fp:
            lines = fp.readlines()

        first_line = lines[0]
        first_entry = first_line.split()
        if first_entry[0] == "include":
            include_path = os.path.join(os.path.dirname(path), first_entry[2])
            self.load_from_file(include_path)
            lines = lines[1:]

        data = "\n".join(lines)
        self.load(data)

    def clear(self):
        """Clear configuration data by dropping all sections."""
        self._sections.clear()
