from zope.interface.declarations import getObjectSpecification


class Copycat(object):
    """Wrapper that forwards attribute getting and setting to another object.

    A copycat object will forward any attributes accessed on it to the
    original object it's wrapping, unless this attribute was set on the
    copycat itself before activation.

    After the copycat object is activated, setting an attribute on it
    will also be forwarded to the original object.  If this attribute
    existed on the copycat itself, it is ignored from then on.

    The Copycat works well when zope interfaces are directlyProvide()d
    on it before activation.  Only the copycat will have the given
    interface, and it will still inherit the other interfaces from the
    original object.

    Here is an example of attribute behavior:

        >>> original = type("Original", (object,), {})()
        >>> copycat = Copycat(original)

    Attribute access is forwarded to the original:

        >>> original.attr = "original value"
        >>> copycat.attr
        'original value'

    Unless the copycat defines its own attribute:

        >>> copycat.attr = "copycat value"
        >>> original.attr
        'original value'
        >>> copycat.attr
        'copycat value'

    After activation, the access behavior is still the same:

        >>> Copycat.activate(copycat)
        >>> original.attr
        'original value'
        >>> copycat.attr
        'copycat value'

    But changing an attribute affects the real object:

        >>> copycat.attr = "modified value"

        >>> original.attr
        'modified value'
        >>> copycat.attr
        'modified value'
    """

    def __init__(self, original):
        self.__copycat_original__ = original
        self.__copycat_activated__ = False

    def activate(self):
        """Make copycat ready for passing on to code unaware of the wrapper.

        This method can't be called directly in the instance, since there
        might be an "activate" attribute in the original object that would
        conflict with it.  Instead, use it as:

            Copycat.activate(copycat)

        "Activating" a copycat means making it ready for passing it on
        to code that isn't aware that it is a wrapper.  Right now the
        only practical effect is that setting attributes on the copycat
        will modify the original object in addition to the copycat.
        """
        self.__copycat_activated__ = True

    def __getattribute__(self, name):
        copycat_dict = object.__getattribute__(self, "__dict__")
        if name in copycat_dict:
            return copycat_dict[name]
        if name == "__providedBy__":
            return getObjectSpecification(self)
        return object.__getattribute__(copycat_dict["__copycat_original__"], name)

    def __setattr__(self, name, value):
        copycat_dict = object.__getattribute__(self, "__dict__")
        if name.startswith("__copycat_") or not copycat_dict["__copycat_activated__"]:
            return object.__setattr__(self, name, value)
        copycat_dict.pop(name, None)
        setattr(self.__copycat_original__, name, value)

    def __eq__(self, other):
        if self.__copycat_activated__:
            return self.__copycat_original__ == other
        else:
            return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        if self.__copycat_activated__:
            return hash(self.__copycat_original__)
        else:
            return id(self)
