# Copyright 2005-2012 Canonical Limited. All rights reserved.

from functools import wraps
import re

from lxml import etree


CPU_ID_REGEX = re.compile(r"^cpu(:\d+)?$")


def coerce_value(f):
    @wraps(f)
    def wrapper(self, value):
        if isinstance(value, str):
            to_format = value.split(" ")
            if to_format[0].isdigit():
                to_format = int(to_format[0])
            else:
                return value
        else:
            to_format = value
        return f(self, to_format)

    return wrapper


class HardwareFormatter(object):
    """
    Nicely format a numeric value to string based on its type, adding the
    proper unit and handling multiples.

    @param resource: the name of the resource type (e.g.: disk, memory,
        clockspeed)
    @param value: the value to format
    """

    def __init__(self, resource, value):
        self.resource = resource
        self.value = value

    def format(self):
        try:
            method = getattr(self, "_get_%s" % self.resource, None)
            return method(self.value)
        except:
            raise AttributeError(
                "Could not format %s using %s formatter." % (self.value, self.resource)
            )

    @coerce_value
    def _get_clockspeed(self, value):
        """Format clock speed."""
        if value > 1000000000:
            # GHz
            formatted = "%0.2f GHz" % (value / 1000000000.0)
            return formatted
        if value == 1000000000:
            # Exact 1 GHz
            return "1 GHz"
        if value < 1000000000:
            # MHz
            formatted = value // 1000000
            return str(formatted) + " MHz"

    @coerce_value
    def _get_memory(self, value):
        """Format RAM memory."""
        if value >= ((1024 ** 4) * 999.5):
            # Report 1024^5 in TiB, %g will shift into %e, so use %f
            return "%0.1f TiB" % (value / float(1024 ** 4))
        if value >= ((1024 ** 3) * 999.5):
            # TiB
            return "%0.3g TiB" % (value / float(1024 ** 4))
        if value >= ((1024 ** 2) * 999.5):
            # GiB
            return "%0.3g GiB" % (value / float(1024 ** 3))
        if value < ((1024 ** 2) * 999.5):
            # Display n MiB, use %f so we don't shift into %e
            return "%0.1f MiB" % (value / float(1024 ** 2))

    @coerce_value
    def _get_disk(self, value):
        """Format disk size."""
        if value >= 1000 ** 4:
            # TB
            return "%s TB" % (value // 1000 ** 4)
        if value >= 1000 ** 3:
            # GB
            return "%s GB" % (value // 1000 ** 3)
        if value >= 1000 ** 2:
            # MB
            return "%s MB" % (value // 1000 ** 2)

    @coerce_value
    def _get_cpucache(self, value):
        """Format cpu cache."""
        if value >= 1048576:
            # MiB
            return "%s MiB" % (value // 1048576)
        if value >= 1024:
            # KiB
            return "%s KiB" % (value // 1024)


def parse_lshw(data):
    """Parse XML output of the 'lshw' command.

    Yields tuples in the form (key, value, unit) for each node, where the key
    is a flattened path for the node in the XML tree.
    """
    parser = etree.XMLParser(recover=True)  # recover makes it tolerant
    doc = etree.XML(data, parser)
    if doc is None:
        return iter([])

    if doc.tag == "list":
        for node in doc:
            return _parse_lshw_node(node, "")
    else:
        return _parse_lshw_node(doc, "")


def _parse_lshw_node(node, prefix):
    """Parse a single node from the XML tree, recursively parsing its children.

    Yields a (key, value, unit) tuple for each node.
    """
    for child in node:
        if child.tag != "node":
            text = child.text.strip() if child.text is not None else ""
            if not text:
                for grandchild in child:
                    key = grandchild.get("id")
                    if not key:
                        key = grandchild.get("type")
                        if not key:
                            continue
                    value = grandchild.get("value")
                    if value:
                        key = "%s%s.%s" % (prefix, child.tag, key)
                        yield key, value, None
                    else:
                        child_text = grandchild.text
                        if child_text:
                            key = "%s [%s]" % (key, child_text.strip())
                        yield "%s%s" % (prefix, child.tag), key, None
            else:
                if text == "Not Available":
                    continue
                units = child.get("units")
                value = text
                key = "%s%s" % (prefix, child.tag)
                yield key, value, units
        else:
            if prefix not in ("core.", ""):
                new_prefix = "%s%s." % (prefix, child.get("id"))
            else:
                new_prefix = "%s." % child.get("id")
            for result in _parse_lshw_node(child, new_prefix):
                yield result
            # If the current node is a CPU that is disabled, yield a tuple
            # describing it as such.
            is_cpu_node = CPU_ID_REGEX.match(child.get("id"))
            if is_cpu_node and child.get("disabled") == "true":
                # new_prefix contains the current node's ID and a trailing dot.
                yield ("%sdisabled" % new_prefix, "true", None)
