# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import re
from canonical.lib.text import is_ip_address


def idna_encode_email_address(email_address):
    """
    Convert unicode domain in email address to punycoded ASCII, in compilance
    with RFC 3490, and return the complete email address in this form.
    """
    local_part, domain = email_address.split("@")
    ascii_domain = domain.encode("idna").decode("utf-8")
    return local_part + "@" + ascii_domain


def is_email_address(address_string):
    """
    Return True if the address string is compliant with RFC 2822 after
    internationalised domain names are converted to ASCII punycode in
    compliance with RFC 3490 (IDNA).
    """
    # RFC 2822 also defines optional display-name which precedes the <mailbox>
    mailbox_match = re.search("^(.*)[^<]*<(.*)>$", address_string)
    if mailbox_match:
        display_name, mailbox = mailbox_match.groups()
        if not _is_email_display_name(display_name):
            return False
    else:
        mailbox = address_string
    try:
        local_part, domain = mailbox.split("@")
    except ValueError:
        return False

    # RFC 3490 toASCII conversion of unicode to punycode for domains.
    ascii_domain = str(domain).encode("idna")

    return _is_valid_domain(ascii_domain) and _is_valid_email_address_atom(
        local_part.encode("utf-8")
    )


def _is_email_display_name(display_name):
    """Return True if display_name meets RFC2822 definition."""
    display_name.strip()
    for word in display_name.split():
        if not _is_valid_email_address_atom(word.encode("utf-8")):
            return False
    return True


def _is_valid_domain(domain):
    """
    Return True if domain neither begins or ends with a dot, and is otherwise a
    valid email address atom.
    """
    return (
        not domain.startswith(b".")
        and not domain.endswith(b".")
        and _is_valid_email_address_atom(domain)
        and not is_ip_address(domain.decode("utf-8"))
    )


def _is_valid_email_address_atom(atom):
    """
    Return True if the atom contains only valid characters for an email atom.
    """
    # RFC6530 provides for Unicode characters to be included in the atoms of
    # an email address.  This RFC does not yet have wide adoption, so, for
    # the moment, Unicode characters will be treated as invalid in our
    # atoms.

    # Internationalisation of the domain name system (IDNA) as per RFC 3490 is
    # already in place and should be supported within email addresses, however
    # this is achieved by using a standardised punycode whilst the mail is in
    # transport.  It is the job of the mail client to encode and decode the
    # punycode form, this allows the SMTP server to work in ASCII only mode,
    # whilst the user inputs and sees their unicode characters. It is assumed
    # that this method will be run on atoms that have already been converted to
    # punycode using the RFC 3490 toASCII algorithm.

    # A valid email address, after escape sequences and quoted section
    # have been removed, may not contain any of the following characters:
    #
    #   0-32: ASCII Control characters and white space characters.
    #     34: "
    #     40: (
    #     41: )
    #     44: ,
    #     58: :
    #     59: ;
    #     60: <
    #     62: >
    #     64: @
    #     91: [
    #     92: \
    #     93: ]
    #    123: {
    #    125: }
    #   >127: Non ASCII characters
    excluded = [34, 40, 41, 44, 58, 59, 60, 62, 64, 91, 92, 93, 123, 125]

    if len(atom) == 0:
        return False

    for character in atom:
        if character < 33 or character > 127 or character in excluded:
            return False
    return True
