# Copyright 2005-2015 Canonical Limited.  All rights reserved.

import os
import errno
import logging


class PidFile(object):
    def __init__(self, path):
        self._path = path

    def write(self):
        """Write our PID to the file."""
        if self.check() is not None:
            # There's already a running process owning this PID file
            raise PidFileTakenError()

        # Here we could race with another process trying to acquire the lock
        # file, so we create and open the file atomically. This ensures that
        # only one of the two processes wins the race.
        try:
            fh = os.open(self._path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        except OSError as error:
            if error.errno == errno.EEXIST:
                raise PidFileTakenError()
            else:
                # This is unexpected, let's just propagate the exception.
                raise

        with os.fdopen(fh, "w") as fd:
            fd.write("%d" % os.getpid())

    def read(self):
        """Read the PID in the file."""
        if not os.path.exists(self._path):
            return None
        with open(self._path) as fd:
            pid = int(fd.read())
        return pid

    def check(self):
        """Check whether the PID file is valid.

        @return: The PID of the running process if the PID file exists and its
            content matches the PID of a running process, and None otherwise.
        """
        pid = self.read()
        if pid is None:
            return None

        try:
            os.kill(pid, 0)
        except OSError:
            logging.info("Removing stale PID file %s" % self._path)
            self.remove()
            return None

        return pid

    def remove(self):
        """Remove our PID file."""
        os.remove(self._path)


class PidFileTakenError(Exception):
    """Raised when another running process is owning the PID file."""
