# Copyright 2016 Canonical Limited.  All rights reserved.

from builtins import str
from builtins import zip
from builtins import range
import six


OMITTED = "<omitted>"
SECRET = "__secret__"


def safe_repr(value, replacement=OMITTED):
    """Return an obscured repr for a sensitive value."""
    return repr(replacement)


def _iteritems(container):
    try:
        return list(container.items())
    except (AttributeError, TypeError):
        if hasattr(container, "__iter__"):
            return enumerate(container)
        else:
            return list(zip(list(range(len(container))), container))


def serialize(value, marker=SECRET):
    """Return the value with all secrets marked.

    Every secret value is replaced with this:

      {"__secret__": value}
    """
    if hasattr(value, "__setitem__"):
        container = value
        # Recurse through the container.
        try:
            items = _iteritems(container)
        except TypeError:
            # Unsupported container type.
            # TODO: raise TypeError?
            pass
        else:
            for key, val in items:
                if getattr(key, "__secret__", False):
                    # TODO: Allow this?
                    raise TypeError("secret keys not supported")
                container[key] = serialize(val, marker)

    if getattr(value, "__secret__", False):
        value = {marker: value}
    return value


def deserialize(value, marker=SECRET):
    """Return the value with all secrets marked with the __secret__ attribute.

    The following is collapsed into the __secret__ attribute.

      {"__secret__": value}

    If __secret__ cannot be set on the inner value then the inner value
    is wrapped appropriately.
    """
    if isinstance(value, six.string_types):
        return value
    try:
        size = len(value)
    except TypeError:
        # Not a container and not marked as secret.
        return value

    # Handle a secret marker.
    if size == 1:
        try:
            value = value[marker]
        except (TypeError, KeyError):
            # Not marked as secret
            pass
        else:
            value = deserialize(value)
            return wrap_secret(value)

    # Handle containers.
    container = value
    if not hasattr(container, "__setitem__"):
        # Not a container we need to deal with.
        # TODO: Fail?
        return container

    # Recurse through the container.
    try:
        items = _iteritems(container)
    except TypeError:
        # Unsupported container type.
        # TODO: raise TypeError?
        pass
    else:
        for key, val in items:
            container[key] = deserialize(val, marker)

    return container


def wrap_secret(value):
    """Return a copy/proxy of the value that has __secret__ set to True."""
    # TODO: Take a more generic approach (e.g. dynamically generate a
    # wrapper type.
    if type(value) is str:
        return SecretUnicode(value)

    raise TypeError("unsupported type for value {!r}".format(value))


def is_secret(value):
    """Return True if the value is secret.

    At present this means any SecretUnicode instance.
    """
    if isinstance(value, SecretUnicode):
        return True

    return False


class SecretUnicode(str):
    """A string wrapper that protects sensitive information."""

    __secret__ = True

    def __repr__(self):
        return safe_repr(str(self))
