"""
This module contains functions useful for serializing data.
"""
import json

from zope.security.proxy import removeSecurityProxy

from . import secrets


def serialize(object, allowed_attributes=[]):
    """
    Prepare an object for safe serialization by converting allowed attributes
    to key/value pairs.

    @param object:              an object that will be serialized
    @param allowed_attributes:  a list of attribute names
    @return:                    a dictionary of allowed key/value pairs
    """
    return dict(
        [(attribute, getattr(object, attribute)) for attribute in allowed_attributes]
    )


def json_dumps(obj):
    obj = secrets.serialize(obj)
    return json.dumps(obj, default=removeSecurityProxy)


def json_loads(encoded):
    payload = json.loads(encoded)
    payload = secrets.deserialize(payload)
    return payload


def namedtuple_to_dict(obj):
    """Convert a namedtuple to dict with tuple fields as keys."""
    to_dict = obj._asdict()
    _scan_dict(to_dict)
    return to_dict


def _scan_dict(d):
    """Scan a dict and convert namedtuple values to dicts inplace."""
    for key, value in d.items():
        if isinstance(value, tuple) and hasattr(value, "_asdict"):
            d[key] = namedtuple_to_dict(value)
        elif isinstance(value, dict):
            _scan_dict(value)
