# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
This module contains functions for manipulating text.
"""

import re
import string
import random
from base64 import b64decode, b64encode
from netaddr import valid_ipv4, valid_ipv6

import six
from unidecode import unidecode

from six.moves import range as xrange

ENCODING_PREFIX = "b64"


def arg_to_bytes(func):
    """Encodes str to bytes if not bytes"""

    def wrapper(bytes_or_str):
        if type(bytes_or_str) == str:
            bytes_or_str = bytes_or_str.encode("utf8")
        return func(bytes_or_str)

    return wrapper


def expand_placeholders(text, values):
    """
    Expand placeholders in text (in the form of ${name}) with actual values.
    Unknown names are not replaced.

    @param text: the text to expand.
    @param value: a dict mapping placeholders to actual values.
    @return the text with values for placeholders.
    """
    result = text
    for placeholder, value in list(values.items()):
        result = result.replace("${%s}" % placeholder, value)
    return result


def sanitize_hostname(hostname):
    """Sanitize a hostname.

    If the hostname contains multiple lines, only the first one is kept.
    Characters that are not allowed in a hostname are stripped.

    Note that this doesn't validate hostnames strictly, it's just meant
    to sanitize hostname provided by clients.

    """
    hostname = hostname.splitlines()[0]
    # strip all invalid chars.
    return re.sub(r"[^a-zA-Z0-9.-]", "", hostname)


LOWER_ALPHANUMERIC_RE = re.compile(r"^[a-z0-9_]*$")


def is_valid_lower_alphanumeric(name):
    """Check if name contains only lowercase alphanumerics and underscores."""
    if not name:
        return False
    return bool(LOWER_ALPHANUMERIC_RE.match(name))


ONLY_SPACES_RE = re.compile(r"\s+$", re.UNICODE)


def is_valid_title(title):
    """
    Check if the provided string is valid as a title. It must neither contain
    newlines, nor consist entirely of white-space characters.

    @param title: the title to check.
    @return: bool indicating whether it is a valid title
    """
    if not title:
        return False

    only_spaces = bool(ONLY_SPACES_RE.match(title))
    invalid = only_spaces or "\n" in title
    return not invalid


SLUG_NAME_RE = re.compile(r"^[a-z0-9][-a-z0-9+]*$")


def is_valid_slug(name, min_length=1, max_length=None):
    """
    Check if name is a valid "slug".

    @param name: the name to check.
    @param min_length: optional minimum length.
    @param max_length: optional maximum length.
    @return: True if the name is valid
    """
    length = len(name)

    if min_length < 1:
        raise ValueError("min_length must be at least 1")

    if max_length is not None:
        if max_length < min_length:
            raise ValueError("max_length must be greater than min_length")
        if length > max_length:
            return False

    if length < min_length:
        return False
    return bool(SLUG_NAME_RE.match(name))


def slug_from_title(title, max_root_length, existing_check):
    """
    Generate a name suitable for a URL from a free-form title provided by a
    user.

    @param title: The unicode title.
    @param max_root_length: The maximum length of the root to use before adding
        digits for unicity.
    @param existing_check: A callable taking the root used, which returns an
        iterable of names which are already used.
    """
    ascii_title = ""
    for character in title:
        ascii_chars = unidecode(character)
        # In order to avoid the need for a default slug value, we need to
        # ensure everything is encoded. If we encounter an undecodeable
        # character we fall back to IDNA
        if ascii_chars == "":
            ascii_title = title.encode("idna")
            break
        ascii_title += ascii_chars
    # There are a special set of characters that won't be encoded in the IDNA
    # format because they are in the ASCII set, but nevertheless cannot appear
    # in a slug.

    if type(ascii_title) == bytes:
        ascii_title = ascii_title.decode("utf-8")

    slug = re.sub("[^\w\s-]", "", ascii_title).strip()
    slug = re.sub("[-\s]+", "-", slug.lower())
    if slug == "":
        slug = "item"
    root = slug = slug[:max_root_length]
    counter = 0
    existing = existing_check(slug)
    while slug in existing:
        counter += 1
        slug = "%s%d" % (root, counter)
    return slug


def lower_class(cls, suffix=None):
    """Split a class name in lowercase words and join them with a separator.

    @param suffix: If not C{None} a class name suffix that will be stripped.
    @returns the processed class name as a string.
    """
    name = cls.__name__
    if suffix is not None and name.endswith(suffix):
        name = name[: -len(suffix)]
    words = re.findall("[A-Z][a-z]+", name)
    return "-".join(word.lower() for word in words)


def capitalize_words(string):
    """Capitalize each word in a string."""
    return " ".join([word.capitalize() for word in string.split(" ")])


def password_decode(password):
    """
    Returns b64decoded version of password, or password if it can't be
    b64decoded.
    """
    if password.startswith("{}:".format(ENCODING_PREFIX)):
        return b64decode(password[len(ENCODING_PREFIX) :]).decode()
    else:
        return password


@arg_to_bytes
def password_encode(password):
    """
    Obfuscate a password to prevent casual reading.
    """
    return "{}:{}".format(ENCODING_PREFIX, b64encode(password).decode())


def make_secure_id(length=64):
    system_random = random.SystemRandom()
    choices = string.digits + string.ascii_letters
    return "".join(system_random.choices(choices, k=length))


def describe_list(
    prefix="", items=None, conjunction="and", pluralize="s", sort=True, quote=True
):
    """Convert a sequence of items into an English list.

    @param prefix: optional prefix to place at the beginning of the returned
        English string.  If the prefix ends in a colon, it is preserved even
        after pluralization.
    @param items: list of items to concatenate with commas or conjunction.
    @param conjunction: string to use as a separator between last two items.
    @param pluralize: The string to use when pluralizing the prefix when the
        number of items to listify is greater than one or False to disable
        pluralization.  The default behavior is to pluralize with an 's'.
    @param sort: If True then format the items in sorted order.
    @param quote: If True then quote all of the items with a single quote.
    @return string of concatenated items.
    """
    # An easy mistake to make is to provide the list of items as the first
    # argument instead of the prefix.
    if not isinstance(prefix, six.string_types):
        raise TypeError("Prefix must be a string.")
    # No items means a returned empty string, ignoring any prefix.
    if not items:
        return ""
    if sort:
        items = sorted(items)
    if quote:
        items = ["'%s'" % item for item in items]

    # Remove any trailing colon so pluralization will work.
    colonize = prefix.endswith(":")
    if colonize:
        prefix = prefix[:-1]

    # Adjust the prefix, if provided.
    if prefix:
        # We only pluralize if it is reqested and there are multiple items.
        if pluralize and len(items) > 1:
            prefix += pluralize
        # If the prefix originally had a colon, put it back.
        if colonize:
            prefix += ":"
        prefix += " "

    if len(items) == 1:
        return "".join([prefix, items[0]])
    else:
        return " ".join([prefix + ", ".join(items[:-1]), conjunction, items[-1]])


def describe_dict(
    prefix="", items=None, conjunction="and", pluralize="s", sort=True, quote=True
):
    """Convert a mapping of items to English.

    @param prefix: optional prefix to place at the beginning of the returned
        English string.  If the prefix ends in a colon, it is preserved even
        after pluralization.
    @param items: mapping of items to concatenate with commas or conjunction.
    @param conjunction: string to use as a separator between last two items.
    @param pluralize: The string to use when pluralizing the prefix when the
        number of items to listify is greater than one or False to disable
        pluralization.  The default behavior is to pluralize with an 's'.
    @param sort: If True then format any values that are lists in sorted
        order.  This parameter does not affect the order the keys are
        formatted, which is always in sorted order.
    @param quote: If True then quote the values.  This parameter does not
        affect the quoting of keys/value pairs of the mapping itself, which
        are always unquoted.
    @return: The English text describing the dictionary of labels and values.
        For example:
            describe_dict(
                "Box created with", {'name": "Bob", "title": "MrMan"})
            returns "Box created with name 'Bob' and title 'MrMan'"

    Values of None and empty sequence will be omitted from the result.
    """
    described = []
    for key, value in sorted(items.items()):
        if value is None:
            continue

        if not isinstance(value, list):
            value = [value]

        if len(value) == 0:
            continue

        described.append(
            describe_list(
                key,
                value,
                conjunction=conjunction,
                pluralize=False,
                sort=sort,
                quote=quote,
            )
        )

    return describe_list(prefix, described, conjunction, pluralize, quote=False)


def format_float(float_value, decimals=2):
    """
    Return a string from a float value, without decimals if the value is
    integer.
    """
    formatting = "%d" if float_value.is_integer() else "%%.%df" % decimals
    return formatting % float_value


def generate_password(length, include_specials=True):
    """Return a randomly chosen password string."""
    rand = random.SystemRandom()
    allowed_characters = "ABCDEFGHIJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz0123456789"
    if include_specials:
        allowed_characters += "^_-."
    return "".join([rand.choice(allowed_characters) for i in range(length)])


def trim_text(content, width=120, length=10):
    """Trim the given text in width and length.

    @param content: The text to trim.
    @param width: The maximum number of characters that each line of the
        returned text should have.
    @param length: The maximum number of lines that the returned text
        should have.

    @return: The trimmed text.
    """
    lines = content.split("\n")
    return "\n".join(line[:width] for line in lines[:length])


def is_ip_address(address):
    """Return True if address is either an IPV4 or IPV6 compliant address."""
    return address and (valid_ipv4(address) or valid_ipv6(address))


def fqdn_to_hostname(hostname):
    """Convert a FQDN to a hostname."""
    return hostname.split(".")[0]


def len_in_bytes(text):
    """Returns length of the string in bytes.

    If text is a unicode string, encodes it to UTF-8 first,
    otherwise returns the length of the passed in bytestring.
    """
    if isinstance(text, str):
        return len(text.encode("utf-8"))
    else:
        return len(text)
