from __future__ import division

import os

from datetime import datetime, date, timedelta

import six


__all__ = [
    "to_timestamp",
    "utctoday",
    "date_from_epoch",
    "format_datetime",
    "new_date_period",
]


TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

DATE_FORMAT = "%Y-%m-%d"


def to_timestamp(date, epoch=datetime.utcfromtimestamp(0), subsecond_resolution=False):
    delta = date - epoch
    timestamp = (delta.days * 60 * 60 * 24) + delta.seconds
    if subsecond_resolution:
        timestamp += delta.microseconds / 1000000
    return timestamp


def utctoday():
    return datetime.utcnow().date()


def date_from_epoch(epoch):
    """Return the date from a timestamp"""
    return datetime.fromtimestamp(epoch).date()


def utcnow():
    """Wrap C{datetime.utcnow} for testing purpose."""
    return datetime.utcnow()


def format_datetime(timestamp):
    """Format timestamp from a C{date} or C{datetime} object.

    @param timestamp: a C{date} or C{datetime} object
    @return: a C{str}, or C{None} if parameter is C{None}."""
    if timestamp is None:
        return None
    if isinstance(timestamp, date):
        tt = timestamp.timetuple()
    else:
        # Convert timestamp to UTC.
        tt = timestamp.utctimetuple()
    utc_timestamp = datetime(
        tt.tm_year, tt.tm_mon, tt.tm_mday, tt.tm_hour, tt.tm_min, tt.tm_sec
    )
    return utc_timestamp.strftime(TIMESTAMP_FORMAT)


def format_date(date):
    """Format date to string."""
    if date is None:
        return
    return date.strftime(DATE_FORMAT)


def parse_datetime(time_string):
    """
    Parse a time string into a C{datetime}. The format is the same as The
    one used in L{format_datetime), making it the reverse function.
    """
    return datetime.strptime(time_string, TIMESTAMP_FORMAT)


def compare_modification_times(path1, path2):
    """Return True if path1's modification time is more recent than path2's."""
    stat1 = os.stat(path1)
    stat2 = os.stat(path2)
    return stat1.st_mtime > stat2.st_mtime


def new_date_period(start=None, delta=None):
    """Return (start, end) for the given start date and delta (days).

    If no start is given then today is used.  If start is a string then
    it is parsed as YYYY-MM-DD.  If no delta is given then a delta of
    30 days is used.
    """
    if delta is None:
        delta = timedelta(days=30)
    elif isinstance(delta, int):
        delta = timedelta(days=delta)
    elif not isinstance(delta, timedelta):
        delta = timedelta(**delta)

    if start is None:
        end = date.today()
        start = end - delta
    else:
        if isinstance(start, six.string_types):
            start = datetime.strptime(start, "%Y-%m-%d").date()
        elif isinstance(start, datetime):
            start = start.date()
        elif not isinstance(start, date):
            start = date(*start)
        end = start + delta
    return (start, end)
