# Copyright 2005-2012 Canonical Limited. All rights reserved.

"""
Miscellaneous functions for validators.
"""

from zope.schema import ValidationError as ZopeValidationError


class ValidationError(Exception):
    """
    Base class for validation error that should be trapped by the C{validator}
    wrapper.
    """


def validator(validator, message):
    """
    Wrap a validation function for use as a validator in a zope.schema.Field
    constraint.

    Exceptions from the validator method are trapped if they inherit from
    C{ValidationError}.

    To return the appropriate error message, the validator must raise a
    ValidationError with the docstring and the message set to the provided
    error message.
    """

    class FieldValidationError(ZopeValidationError):
        __doc__ = message

    def validate(value):
        try:
            valid = bool(validator(value))
        except ValidationError:
            valid = False

        if not valid:
            raise FieldValidationError(message)
        return True

    return validate


def _check_address_tokens(tokens, base, count):
    if not len(tokens) == count:
        return False
    try:
        for token in tokens:
            value = int(token, base)
            if not 0 <= value <= 255:
                return False
    except ValueError:
        return False

    return True


def is_valid_ip_address(text):
    """Check that the string is a valid IP address"""
    tokens = text.split(".")
    return _check_address_tokens(tokens, 10, 4)


def is_valid_mac_address(text):
    """
    Check that the string is a valid MAC address. See C{_tokenize_mac_address}
    for valid separators.
    """
    tokens = _tokenize_mac_address(text)
    return all(len(token) == 2 for token in tokens) and _check_address_tokens(
        tokens, 16, 6
    )


def _tokenize_mac_address(text):
    """
    Split the provided string considered as a MAC address.

    Elements in the string can be separated by ':', '-' or without any
    separator.
    """
    text = text.lower()
    if ":" in text:
        separator = ":"
    elif "-" in text:
        separator = "-"
    else:
        separator = None

    if separator:
        tokens = text.split(separator)
    else:
        tokens = [text[i : i + 2] for i in range(0, len(text), 2)]
    return tokens


def normalize_mac_address(text):
    """
    Return a colon-separated string for the MAC address, or None if the string
    is not a valid address.
    """
    if is_valid_mac_address(text):
        return ":".join(_tokenize_mac_address(text))
