import time
import transaction

from datetime import datetime


class TimeoutError(Exception):
    """Raised when the condition doesn't become true."""

    def __init__(self):
        message = "Timeout at {}".format(datetime.utcnow())
        super(TimeoutError, self).__init__(message)


def wait_until(condition, polling=1, timeout=180, sleep=time.sleep):
    """Wait until the given condition becomes true.

    This function will "poll" the given condition callable every second
    for by default, until it returns True or until 180 seconds have passed.

    @param condition: A callable returning True or False.
    @param polling: Tweak the interval between two subsequent polls.
    @param timeout: Try for maximum this amount of time.
    @param sleep: The sleep function to use (for testing).
    """
    tries = int(timeout / polling)
    while tries > 0:
        # Commit the transaction in case the condition we're checking depends
        # on database state (so we can see updates from other transactions
        # that got committed in the meantime).
        transaction.commit()
        if condition():
            return
        tries -= 1
        sleep(polling)
    raise TimeoutError()
