# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Landscape logging handlers."""

import logging
import logging.handlers

from canonical.lib.text import len_in_bytes
from canonical.landscape.model.configuration import is_development_mode


__all__ = ["ChunkedSysLogHandler"]


class ChunkedSysLogHandler(logging.handlers.SysLogHandler):
    """Splits long log messages into several `message_size`-sized chunks.

    Each message chunk is individually logged with an addition of an extra
    continuation message in non-final chunks indicating that the log message
    has not yet been fully emitted.
    """

    CONTINUATION_MESSAGE = "\n(log message truncated, continued below)"

    def __init__(
        self,
        address=("localhost", logging.handlers.SYSLOG_UDP_PORT),
        facility=logging.handlers.SysLogHandler.LOG_USER,
        socktype=None,
        max_message_size=8000,
    ):
        """Adds `message_size` argument to SysLogHandler's constructor."""
        super(ChunkedSysLogHandler, self).__init__(
            address=address, facility=facility, socktype=socktype
        )

        # In development mode we'd like to see any new errors that
        # occur, but in production we want it to just keep plugging
        # away even if something happens to rsyslog.
        #
        # If raiseExceptions is false then the handleError method of
        # the handler will never inspect sys.exc_info, which saves us
        # leaking a bit of memory every time.  Again, this is
        # desirable in production.
        logging.raiseExceptions = is_development_mode()

        self.max_message_size = max_message_size

    @staticmethod
    def _split_message(message, max_message_size, continuation_message="\n(cont)"):
        """Split 'message' into 'max_message_size'd chunks."""
        assert (
            len(continuation_message) < max_message_size
        ), "Continuation message is longer (%d) than max_message_size %d." % (
            len(continuation_message),
            max_message_size,
        )
        chunks = []
        while len_in_bytes(message) > max_message_size:
            max_size = max_message_size - len(continuation_message)
            part_msg = message[0:max_size]
            # Shorten so that UTF-8 encoded representation fits into the
            # alloted space.
            while len_in_bytes(part_msg) > max_size:
                part_msg = part_msg[:-1]
            message = message[len(part_msg) :]
            part_msg += continuation_message
            chunks.append(part_msg)
        chunks.append(message)
        return chunks

    def emit(self, record):
        """Split long log messages into `max_message_size`-sized chunks.

        Shorter message are emitted in full, and longer ones as several
        syslog messages."""
        original_args = record.args
        original_message = record.msg
        chunks = self._split_message(
            record.getMessage(), self.max_message_size, self.CONTINUATION_MESSAGE
        )
        for chunk in chunks:
            # The record has already been formatted by virtue of
            # calling record.getMessage()
            record.args = None
            record.msg = chunk
            super(ChunkedSysLogHandler, self).emit(record)

        record.args = original_args
        record.msg = original_message
