# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""
A MessageAPI provides a way to dispatch messages from a request to the proper
registered handler.

An API class can extend an existing one, inheriting all the registered
handlers, and can possibly add new ones, remove handlers or provide a different
handler for an existing message.
"""

from zope.component import queryAdapter, getAdapters
from zope.interface import implementer

from canonical.message.interfaces import (
    IMessageAPI,
    IMessageHandler,
    MessageHandlerNotFound,
)


_marker = object()


@implementer(IMessageAPI)
class MessageAPI(object):
    """
    A convenience class, meant for subclassing, which provides some
    utility methods that makes it easier to implement L{IMessageAPI}.

    Notably, this implementation will cache handler objects for the
    life of MessageAPI instances.
    """

    def __init__(self, system, request):
        self.system = system
        self.request = request
        self._handlers = {}

    def handle(self, type, body):
        """Find and call the message handler appropriate for the given type.

        The handler will be looked up as an adapter from this
        MessageAPI instance to L{IMessageHandler}, with C{type} as its
        name. The handler will then be called with the C{type} and
        C{body}.

        If the handler has already been looked up, then this will use
        the same one instead of querying for adapters again. This
        allows handler objects to store instance state.

        @param type: The message type. The IMessageHandler adapter
            must be registered with its name as this type.
        @param body: Will be passed as the second argument to the
            handler.

        @return: The result of the handler call.

        @raises MessageHandlerNotFound: if no message handler is found
        """
        handler = self._handlers.get(type)
        if handler is None:
            if type is None:
                type = ""
            try:  # Unlike Zope in Python 2, cannot be bytes or None
                type = type.decode()
            except (AttributeError, UnicodeDecodeError):
                pass
            handler = queryAdapter(self, IMessageHandler, name=type, default=_marker)
            if handler is _marker:
                raise MessageHandlerNotFound(self, type)
            self._handlers[type] = handler
        return handler(type, body)

    def get_handlers(self):
        """
        Return all IMessageHandler adapters for this MessageAPI
        instance.
        """
        return getAdapters((self,), IMessageHandler)
