# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
The message mechanism is formed by three main components: a message
system, one or more message APIs, and one or more message handlers.

The message system is responsible for investigating a request and
dispatching it to the right message API, while the message API is
responsible for investigating the request and processing it to
dispatch received messages to message handlers registered for the
given message API and message type.

Registration of the different components happens through adapters: APIs are
registered for MessageSystems, and handlers are registered for APIs.

It's possible to check which handlers are available for a given API by querying
the message system for a message API.
"""

from zope.component import queryMultiAdapter, getAdapters
from zope.interface import implementer

from canonical.message.interfaces import IMessageSystem, IMessageAPI, MessageAPINotFound


MESSAGE_API_HEADER = "X-Message-API"


_marker = object()


def get_api_version_from_request(request):
    return request.getHeader(MESSAGE_API_HEADER)


@implementer(IMessageSystem)
class MessageSystem(object):
    """An IMessageSystem implementor that can deal with Zope requests.

    To register message APIs with this message system, provide an
    adapter from (IMessageSystem, IRequest) to IMessageAPI with a name
    matching the version number the message API will respond to (that
    is, specified in the X-Message-API header of HTTP requests).
    """

    def get_message_api(self, request):
        """Get the appropriate IMessageAPI provider.

        The one with the name matching the X-Message-API header in the
        request.
        """
        api = get_api_version_from_request(request)
        if api is None:
            api = ""
        try:  # Unlike Zope in Python 2, cannot be bytes or None
            api = api.decode()
        except (AttributeError, UnicodeDecodeError):
            pass
        message_api = queryMultiAdapter(
            (self, request), IMessageAPI, name=api, default=_marker
        )
        if message_api is _marker:
            raise MessageAPINotFound(self, request, api)
        return message_api

    def get_api_names(self, request):
        """Return the names of the available L{IMessageAPI}s for a request."""
        names = [name for name, _ in getAdapters((self, request), IMessageAPI)]
        return sorted(names)

    def dispatch(self, request):
        message_api = self.get_message_api(request)
        return message_api.run()
