# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.component import getAdapters
from zope.interface import implementer
from zope.traversing.adapters import DefaultTraversable
from zope.location.interfaces import LocationError

from zope.security.management import checkPermission

from zope.pagetemplate.engine import Engine

from canonical.navigation.interfaces import INavigation, INavigationItem
from canonical.security.status import is_context_disabled


__all__ = ["Navigation", "NavigationItem", "TALESNavigationItem"]


@implementer(INavigation)
class Navigation(DefaultTraversable):

    IS_AVAILABLE = True  # default; subclasses must opt out
    IS_PERMITTED = True  # default; subclasses must opt out
    IS_SUPPORT = False  # default

    def __init__(self, context, request, parent=None):
        super(Navigation, self).__init__(self)
        self.context = context
        self.request = request
        self.parent = parent
        self.items = []

    def is_permitted(self):
        return self.IS_PERMITTED

    def is_support(self):
        return self.IS_SUPPORT

    def is_available(self):
        if not self.is_permitted():
            return False

        if not self.is_support():
            if is_context_disabled(self):
                return False

        return self.IS_AVAILABLE

    def get_items(self):
        items = self.items[:]
        for name, adapter in getAdapters(
            (self.context, self.request, self), INavigation
        ):
            if INavigationItem.providedBy(adapter):
                items.append(adapter)
            else:
                items.extend(adapter.get_items())
        return items

    def __iter__(self):
        items = self.get_items()
        items.sort()
        return iter(items)


@implementer(INavigationItem)
class NavigationItem(Navigation):

    name = ""
    title = ""
    description = ""
    action = None
    order = 0
    icon = None
    extra_class = ""

    def is_selected(self):
        action = self.action
        if action is not None:
            return self.request.getURL(path_only=True).startswith(action)
        return False

    def is_external(self):
        action = self.action
        if action is not None:
            if not action.startswith("/"):
                return True
        return False

    def __lt__(self, other):
        return (self.order, self.title) < (other.order, other.title)


class TALESNavigationItemMeta(type):
    """Dynamically compile expressions at class creation."""

    def __new__(cls, name, bases, dict):
        cls = type.__new__(cls, name, bases, dict)

        for attr in cls.__expression_attr__:
            expr = getattr(cls, "{}_expr".format(attr), None)

            if expr is not None:
                setattr(cls, "_compiled_{}_expr".format(attr), Engine.compile(expr))

        return cls


class TALESNavigationItem(NavigationItem, metaclass=TALESNavigationItemMeta):

    __expression_attr__ = [
        "title",
        "description",
        "action",
        "icon",
        "available",
        "selected",
    ]

    permission = None

    def __init__(self, context, request, parent=None):
        super(TALESNavigationItem, self).__init__(context, request, parent)

        tales_context = TALESContext(
            item=self,
            context=context,
            request=request,
            parent=parent,
            nothing=None,
            true=True,
            false=False,
        )

        for attr in self.__expression_attr__:
            if not getattr(self, attr, None):
                compiled_expr = getattr(self, "_compiled_%s_expr" % attr, None)
                if compiled_expr is not None:
                    setattr(self, attr, compiled_expr(tales_context))

    def is_available(self):
        if self.permission is not None:
            if not checkPermission(self.permission, self.context):
                return False
        marker = self
        available = getattr(self, "available", marker)
        if available is not marker:
            return bool(available)
        return super(TALESNavigationItem, self).is_available()

    def is_selected(self):
        marker = self
        selected = getattr(self, "selected", marker)
        if selected is not marker:
            return bool(selected)
        return super(TALESNavigationItem, self).is_selected()


class TALESContext(dict):

    _engine = Engine

    @property
    def request(self):
        return self["request"]

    @property
    def vars(self):
        return self
