# Copyright 2005-2022 Canonical Limited. All rights reserved.

import time
import json

import transaction

from zope.interface import Interface, implementer
from zope.component import adapter, getUtility
from zope.session.interfaces import ISession, IClientIdManager


from canonical.notes.base import NoteContainerBase
from canonical.notes.interfaces import INoteContainer, INFO, WARNING, ERROR


class SessionNotes(NoteContainerBase):

    session_key = "canonical.notes"
    time = time.time  # For testing

    _session = None

    # XXX for ajax pages that want to render the note on the next page load
    remove_on_commit = True

    def __init__(self, request):
        self._request = request
        session = self.get_session()
        if session is not None:
            for kind, text in sorted(session.values()):
                self._kind = max(self._kind, kind)
                self._text += "%s " % text

    def _schedule_note_removal(self):
        """Remove all notes if the user has a chance to see them.

        We assume the user could see the notes if the request ends in a
        non-error state and is not a redirect.
        """
        session = self.get_session()
        if session is not None and self.remove_on_commit:
            trans = transaction.get()

            def remove_note():
                if self._request.response.getStatus() not in (302, 303):
                    for note_key in session.keys():
                        del session[note_key]

            trans.addBeforeCommitHook(remove_note)

    def as_dict(self):
        """Get the current notes, grouped by type."""
        self._schedule_note_removal()
        session = self.get_session()
        notes = {"info": [], "warning": [], "error": []}
        if session is not None:
            for kind, note in session.values():
                if kind == INFO:
                    notes["info"].append(note)
                if kind == WARNING:
                    notes["warning"].append(note)
                if kind == ERROR:
                    notes["error"].append(note)
        return notes

    def as_json(self):
        """Get the current notes, grouped by type and JSON-encoded."""
        return json.dumps(self.as_dict())

    def content(self):
        # We only want to display notes once.
        self._schedule_note_removal()
        return super(SessionNotes, self).content()

    def get_session(self, create=False):
        if self._session is None:
            # Check if there's a client id set for the session. If no
            # client id exists, don't lookup an ISession adapter for the
            # request just yet, to avoid creating a session needlessly.
            client_id_manager = getUtility(IClientIdManager)
            client_id = client_id_manager.getRequestId(self._request)
            if client_id is not None or create:
                # Either there's a client id set, which means a
                # session exists already, or we've been asked to
                # create a session anyway.
                self._session = ISession(self._request)[self.session_key]
        return self._session

    def add(self, kind, text):
        session = self.get_session(create=True)
        note_key = "note-%f" % self.time()
        while note_key in session:
            # Make the key unique, in case two notes are added within
            # the same time.time() frame.
            note_key += "-1"
        session[note_key] = (kind, text)
        self._kind = max(self._kind, kind)
        self._text += "%s " % text

    def extract(self):
        """
        Retrieve the list of notes in the session, mainly to be compared during
        tests.
        """
        session = self.get_session()
        if session is not None:
            return sorted(session.values())
        return []


@adapter(Interface, Interface)
@implementer(INoteContainer)
def default_session_notes(context, request):
    return SessionNotes(request)
