# Copyright 2005-2022 Canonical Limited. All rights reserved.

import time
import cgitb
import collections
import sys
import textwrap

from six.moves.urllib.parse import parse_qsl, urlencode, urljoin

from oic.oic.provider import Provider
from oic.utils.sdb import create_session_db, SessionDB
from oic.utils.authn.authn_context import AuthnBroker
from oic.utils.authz import AuthzHandling
from oic.utils.authn.client import verify_client
from oic.utils.userinfo import UserInfo


class BogusAuthn:
    """Mock authenticator, which accepts any identity you ask for."""

    def create_cookie(self, *args, **kwargs):
        pass

    def authenticated_as(self, cookie, authorization, max_age):
        return {"uid": authorization}, time.time()


class OpenIdConnectProviderMiddleware(object):
    """
    This middleware is used to create a simple OpenID-Connect provider.
    OIDC is based on oauth2 and is not to be confused with OpenID.
    (e.g. MS IDP provides oidc, Ubuntu SSO provides openid)
    """

    def __init__(self, app, base_url):
        self.app = app
        self.base_url = base_url + "oidc"
        self.sessions = {}

        client_id = "develclient"
        client_secret = "develsecret"
        # list of accepted clients
        self.clients = {
            "develclient": {
                "client_id": client_id,
                "client_secret": client_secret,
                # allow all response types on this OP
                "response_types": [
                    "code",
                    "id_token",
                    "token",
                    "access_token",
                ],
                # allowed redirect URIs
                "redirect_uris": [
                    (urljoin(self.base_url, "/login/handle-openid"), ""),
                ],
                "registration_access_token": "",
                # avoid setting encryption keys for this service
                "id_token_signed_response_alg": "none",
                "client_salt": "4",
            },
        }
        self.userinfo = collections.defaultdict(
            lambda: {
                "email": "kurt@example.com",
                "name": "Kurt Lee",
                "zoneinfo": "Europe/Paris",
            }
        )

        patch_sdb()
        self.sdb = create_session_db(self.base_url, "secret", "password", self.sessions)

        broker = AuthnBroker()
        broker.add("bogus", BogusAuthn())
        authz = AuthzHandling()
        userinfo = UserInfo(self.userinfo)
        self.provider = Provider(
            self.base_url.rstrip("/"),
            self.sdb,
            self.clients,
            broker,
            userinfo,
            authz,
            verify_client,
            baseurl=self.base_url,
            symkey="notmuchsecret",
        )
        self.provider.keyjar.add_symmetric(client_id, client_secret)

    def __call__(self, environ, start_response):
        path = environ["PATH_INFO"]
        if path.startswith("/oidc"):
            handler = OpenIdConnectRequestHandler(environ, start_response, self)
            return (handler(),)
        else:
            return self.app(environ, start_response)


class OpenIdConnectRequestHandler(object):
    """
    Handles a single request to the OpenIdConnectProviderMiddleware
    """

    def __init__(self, environ, start_response, server):
        self.environ = environ
        self.start_response = start_response
        self.post_data = None
        self.query = None
        self.status = "200 OK"
        self.headers = []
        self.result = b""
        self.server = server

    def __call__(self):
        self.query = dict(parse_qsl(self.environ.get("QUERY_STRING", "")))

        if self.environ.get("REQUEST_METHOD", "GET") == "POST":
            self.do_POST()
        else:
            self.do_GET()
        self.start_response(self.status, self.headers)
        return self.result

    def do_GET(self):
        try:
            path = self.environ["PATH_INFO"]
            if path.strip("/") == "oidc":
                self.show_main_page()
            elif path == "/oidc/.well-known/openid-configuration":
                # technically optional endpoint
                self.configuration()
            elif path == "/oidc/authorization":
                # A good OP would verify the request here.
                # In order to keep BogusAuthn simple, we ask for the identity
                # beforehand. A side effect is an invalid client_id
                # will raise errors after showing the login form
                # instead of before. Another side effect is you can
                # easily tamper the request to test validation errors.
                self.show_login_page()
            elif path == "/oidc/end_session":
                self.end_session()
            elif path.startswith("/oidc/userinfo"):
                self._handle_endpoint(self.server.provider.userinfo_endpoint)
            else:
                self.set_status(404)
        except Exception:
            self.set_status(500)
            self.set_header("Content-type", "text/html")
            self.result = cgitb.html(sys.exc_info(), context=10)

    def do_POST(self):
        try:
            content_length = int(self.environ["CONTENT_LENGTH"])
            self.post_data = self.environ["wsgi.input"].read(content_length).decode()
            self.query.update(parse_qsl(self.post_data))

            path = self.environ["PATH_INFO"]
            if path == "/oidc/authorization":
                if self.query["action"] == "Fail":
                    self.reject_auth()
                elif self.query["action"] == "Succeed":
                    self.accept_auth(self.query["subject"])
            elif path.startswith("/oidc/token"):
                self._handle_endpoint(self.server.provider.token_endpoint)
            elif path.startswith("/oidc/userinfo"):
                self._handle_endpoint(self.server.provider.userinfo_endpoint)
            else:
                self.set_status(404)
        except Exception:
            self.set_status(500)
            self.set_header("Content-type", "text/html")

    def configuration(self):
        """Render the openid discovery configuration."""
        response = self.server.provider.create_providerinfo()
        self.headers = [("Content-type", "application/json")]
        self.result = response.to_json().encode()

    def end_session(self):
        """Just redirect. There are no global sessions here."""
        self.set_status(303)
        redirect = self.query.get("post_logout_redirect_uri", "/")
        self.set_header("Location", redirect)

    def accept_auth(self, subject):
        """Return a successful authz data to the RP."""
        # Pass desired subject to authn for Bogus
        response = self.server.provider.authorization_endpoint(
            self.query, authn=subject
        )
        self._serve_response(response)

    def reject_auth(self):
        """Return an error response to the RP."""
        params = {
            "error": "access_denied",
            "error_description": "User pressed the Fail button",
        }
        if "state" in self.query:
            params["state"] = self.query["state"]

        location = "{}?{}".format(self.query["redirect_uri"], urlencode(params))
        self.set_status(303)
        self.set_header("Location", location)

    def _handle_endpoint(self, endpoint):
        """Pass request data to endpoints and serve the response."""
        request = self.post_data or self.environ.get("QUERY_STRING")
        response = endpoint(
            request=request, authn=self.environ.get("HTTP_AUTHORIZATION")
        )
        self._serve_response(response)

    def _serve_response(self, response):
        """Serve an OIDC Reponse."""
        self.set_status(response.status_code)
        for header in response.headers:
            self.set_header(*header)
        if response.status_code >= 300 and response.status_code < 400:
            self.set_header("Location", response.message)
        else:
            self.result = response.message

    def set_status(self, code):
        """Set http status code."""
        from twisted.web.http import RESPONSES

        self.status = "%d %s" % (code, RESPONSES[code])

    def set_header(self, header, value):
        """Add an http header."""
        self.headers.append((header, value))

    def show_main_page(self):
        """Render a description page."""
        endpoints = [
            "<li>%s/%s</li>" % (self.server.base_url, ep.url)
            for ep in self.server.provider.endp
        ]
        conf_url = self.server.base_url + "/.well-known/openid-configuration"
        content = """
            <p>
            This is a fake authorization provider powered by python-oidc.
            </p>
            Here are the interesting endpoints:
            <ul>
            <li>{conf_url}</li>
            {ep}
            </ul>""".format(
            conf_url=conf_url, ep="".join(endpoints)
        )
        self.show_page(200, "Main oidc page", body=content)

    def show_login_page(self):
        """Render a simple login form."""
        content = """
            <form method="POST">
                <label>Identity (subject):
                    <input id="subject" name="subject" value="{}"/>
                </label>
                <p>Do you want this authentication to succeed or fail?</p>
                <input type="submit" name="action" value="Succeed"/>
                <input type="submit" name="action" value="Fail"/>
                <script>
                    var el=document.getElementById("subject");
                    var v = el.value;
                    el.value = "";
                    el.focus();
                    el.value = v;
                </script>
            </form>
        """.format(
            urljoin(self.server.base_url, "/id/")
        )
        self.show_page(200, "OpenIdConnect Login", body=content)

    def show_page(self, code, title, body=""):
        """Renders a simple page."""
        self.set_status(code)
        self.headers = [("Content-type", "text/html")]
        self.result = (
            textwrap.dedent(
                """
            <!DOCTYPE html>
            <html>
              <head>
                <meta charset="utf-8"/>
                <title>{title}</title>
              </head>
              <body>
                <h1>Fake OpenID-Connect provider</h1>
                <div style="border: 1px #777 solid; padding: 2em;
                            background: #ddc">{body}</div>
              </body>
            </html>
            """
            )
            .format(title=title, body=body)
            .encode()
        )


def patch_sdb():
    """Monkey-patch a session database to return uid as subjects (sub).

    Normally, sub are obfuscated and possibly distinct between client_id.
    To allow matching our sample data identities, we patch to return the uid.
    """

    def _do_sub(self, sid, client_salt, sector_id="", subject_type="public"):
        authn_event = self.get_authentication_event(sid)
        sub = authn_event.uid
        self.update(sid, "sub", sub)
        return sub

    SessionDB.do_sub = _do_sub
