# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import sys
import zipfile
import os.path
from datetime import datetime

from landscape.lib.bpickle import loads


SEPARATOR = "------\n"


def format_frames(frames, file):
    """Format and write frames.

    @param frames: is a list of frames as used by Failure.frames, with
        each frame being a list of
        (funcName, fileName, lineNumber, line, locals.items(), globals.items())
    @type frames: list
    """
    for filename, lineno, method, line, localVars, globalVars in frames:
        file.write("%s:%s: %s(...)\n" % (filename, lineno, method))
        file.write(" [ Locals ]\n")
        for name, val in localVars.items():
            file.write("  %s : %s\n" % (name, repr(val)))
        file.write(" ( Globals )\n")
        for name, val in globalVars.items():
            file.write("  %s : %s\n" % (name, repr(val)))


def format_statement(log, file):
    """
    This writes oops tracing statements to the log file.

    ['threadName', 'name', 'thread', 'created', 'process', 'args', 'module',
    'filename', 'levelno', 'msecs', 'pathname', 'lineno', 'asctime', 'msg',
    'message', 'relativeCreated', 'levelname']
    """
    file.write("%(pathname)s:%(lineno)s: %(name)s(...)\n" % log)
    file.write("%(msg)s\n" % log)


def format_storm_statements(statements, file):
    """
    This writes storm tracing statements to the log file.
    """
    for statement, params, start_time, end_time, error in statements:
        if end_time:
            time_taken = "%0.4fs" % (end_time - start_time)
        else:
            time_taken = "Unknown"
        start_time = datetime.utcfromtimestamp(start_time)
        file.write("%s %s\n" % (start_time.strftime("%X"), time_taken))
        file.write(statement % params + "\n")
        file.write("%s" % error + "\n")
        file.write(SEPARATOR)


def extract_oops_directory(zipfile):
    """
    Returns the directory name of the OOPS in a zipfile.
    """
    # Extract the directory in the OOPS file.
    return os.path.dirname(zipfile.filelist[0].filename)


def dump_oops_records(log_entries, file):
    """
    Write statement tracking lines to the log file.
    """
    for log_entry in log_entries:
        if "oops_traceback" in log_entry:
            if "exception" in log_entry["oops_traceback"]:
                file.write(log_entry["oops_traceback"]["exception"] + "\n")
            format_frames(log_entry["oops_traceback"]["frames"], file)
        else:
            format_statement(log_entry, file)
        file.write(SEPARATOR)


def output_traceback(log_entries, file):
    """
    Write statement tracking lines to the log file in the form of a
    traceback.
    """
    for log_entry in log_entries:
        if "oops_traceback" in log_entry:
            traceback = log_entry["oops_traceback"]
            if "exception" in traceback:
                frames = sorted(traceback["frames"], reverse=True)
                for filename, lineno, method, line, localVars, globalVars in frames:
                    file.write(
                        'File "%s", line %s, in %s\n' % (filename, lineno, method)
                    )
                    file.write("  %s\n" % line)
                file.write(traceback["exception"] + "\n")


def dump_oops_log(oops_file, dir=None):
    """
    Trigger the dumping of each of the components of an oops file, oops_file is
    assumed to be a ZipFile object.
    """
    oops_id = extract_oops_directory(oops_file)

    log_path = os.path.join((dir or "."), oops_id)
    if not os.path.isdir(log_path):
        os.makedirs(log_path)

    log_entries = loads(oops_file.read("%s/log" % oops_id))["records"]

    traceback_file_name = os.path.join(log_path, "traceback.txt")
    traceback_file = open(traceback_file_name, "w")
    output_traceback(log_entries, traceback_file)
    traceback_file.close()

    log_file_name = os.path.join(log_path, "log.txt")
    log_file = open(log_file_name, "w")
    dump_oops_records(log_entries, log_file)
    log_file.close()

    statement_file_name = os.path.join(log_path, "statements.txt")
    statement_file = open(statement_file_name, "w")
    storm = loads(oops_file.read("%s/storm" % oops_id))
    format_storm_statements(storm["statements"], statement_file)
    statement_file.close()

    metadata_file_name = os.path.join(log_path, "metadata.txt")
    metadata_file = open(metadata_file_name, "w")
    metadata = loads(oops_file.read("%s/metadata" % oops_id))
    metadata_file.write(str(metadata))
    metadata_file.close()


def dump_oops_file(oops_filename, dir=None):
    """
    This kicks off the dumping of a given filename assumed to be the filename
    of an OOPS dump file.
    """
    oops_file = zipfile.ZipFile(oops_filename)
    dump_oops_log(oops_file, dir)


def run():
    for filename in sys.argv[1:]:
        dump_oops_file(filename)
