# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from traceback import format_exception, format_exception_only
import threading
import linecache
import logging
import weakref
import sys
import io


class OOPSEntryError(Exception):
    """Raised when the entry_name added already exists in the OOPS."""


class OOPS(object):
    """Storage for a sequence of events through composition of OOPS entries.

    An OOPS stores a collection of details for the timespan of a given
    event (e.g. a request), so that understanding what happened is easier.
    It does by storing individual OOPS entries which represents individual
    aspects of the application (logging, database interaction, etc).
    """

    version = 1

    def __init__(self):
        """Construct the OOPS instance."""
        self._entries = {}
        self.wanted = False

    def keep(self):
        """Keeping means flagging the OOPS as wanted."""
        self.wanted = True

    def add(self, entry):
        """Add an entry to the OOPS.  It'll named with entry.entry_name.

        The reason why the name isn't passed in through an argument is to
        discourage the creation of non-documented entries.
        """
        entry_name = getattr(entry, "entry_name", None)
        if not entry_name:
            raise TypeError("%r has no entry_name attribute" % entry)
        if entry_name in self._entries:
            raise OOPSEntryError(
                "%r has entry_name %r, " "which was already added" % (entry, entry_name)
            )
        self._entries[entry.entry_name] = entry

    def get(self, name):
        """Return the entry with the given name."""
        return self._entries.get(name)

    def get_all(self):
        """Return all entries."""
        return list(self._entries.values())

    def get_dump_data(self):
        """Return a dictionary with {entry_name: entry_dump_data}."""
        return dict(
            (entry_name, entry.get_dump_data())
            for entry_name, entry in list(self._entries.items())
        )


class OOPSMetaData(dict):
    """Generic OOPS entry for storing key=>value fields."""

    entry_name = "metadata"
    entry_version = 1

    def get_dump_data(self):
        return self.copy()


def safe_repr(o):
    """
    safe_repr(anything) -> string

    Returns a string representation of an object, or a string containing a
    traceback, if that object's __repr__ raised an exception.
    """

    def _determine_class_name(x):
        try:
            c = x.__class__
        except:
            c = type(x)
        try:
            return "%s.%s" % (c.__module__, c.__name__)
        except:
            try:
                return str(c)
            except:
                return "BROKEN CLASS"

    try:
        return repr(o)
    except:
        import traceback

        str_io = io.StringIO()
        traceback.print_exc(file=str_io)
        klass = _determine_class_name(o)
        tb_value = str_io.getvalue()
        return "<%s instance at 0x%x with repr error %s>" % (klass, id(o), tb_value)


class OOPSTraceback(object):
    """OOPS entry that stores a detailed data dump of an exception.

    The data dump includes information that enables rebuilding the
    traceback, plus other details such as local variables of each
    stack frame.
    """

    entry_name = "traceback"
    entry_version = 1

    def __init__(self, type=None, value=None, traceback=None):
        if type is not None and value is not None:
            exception = "".join(format_exception_only(type, value)).strip()
            self._exception = exception
        else:
            self._exception = None

        self._frames = []
        self._traceback = None

        if traceback is None:
            type, value, traceback = sys.exc_info()

        if traceback is not None:
            self._traceback = "".join(format_exception(type, value, traceback))

            # Below we're extracting information about the *full* stack,
            # so we first walk the traceback up to frame where the
            # exception was raised, and then walk downwards on the frame
            # stack.
            #
            # One issue we have to keep in mind is that lines may have
            # changed on *frames* which are located between the point
            # where the exception was raised and the point where it was
            # caught.  So for these frames we use the lines from the
            # traceback object.

            traceback_lines = {}
            while True:
                traceback_lines[traceback.tb_frame] = traceback.tb_lineno
                if not traceback.tb_next:
                    break
                traceback = traceback.tb_next

            frame = traceback.tb_frame
            while frame:
                lineno = traceback_lines.get(frame, frame.f_lineno)
                code = frame.f_code
                name = code.co_name
                filename = code.co_filename
                line = linecache.getline(filename, lineno)

                locals = dict(
                    (key, safe_repr(value))
                    for (key, value) in list(frame.f_locals.items())
                )
                globals = dict(
                    (key, safe_repr(value))
                    for (key, value) in list(frame.f_globals.items())
                    if key != "__builtins__"
                )

                self._frames.append(
                    (filename, lineno, name, line.strip(), locals, globals)
                )

                frame = frame.f_back

    def get_dump_data(self):
        return {
            "frames": self._frames,
            "exception": self._exception,
            "traceback": self._traceback,
        }


class OOPSLoggingHandler(logging.Handler):
    """Logging handler that forwards logged records to all existent OOPSLogs."""

    def emit(self, record):
        logs = OOPSLog.get_for_logging()
        for log in logs:
            log.add_record(record)


class OOPSLog(object):
    """OOPS entry that stores all logged records created while it's alive.

    Each instance may specify if it should keep track only of records
    logged in the current thread, or globally.

    All fields existent in a log record are saved, so it provides much more
    information than a usual textual representation would (e.g. the file
    and line in which an entry was logged).
    """

    entry_name = "log"
    entry_version = 1

    _global = weakref.WeakKeyDictionary()
    _local = threading.local()

    def __init__(self, thread_local=False):
        self._records = []

        if thread_local:
            try:
                logs = self._local.logs
            except AttributeError:
                logs = self._local.logs = weakref.WeakKeyDictionary()
            logs[self] = True
        else:
            self._global[self] = True

    @classmethod
    def get_for_logging(self):
        """Return all OOPSLogs registered for the current thread plus globals."""
        oops_logs = list(self._global.keys())
        try:
            oops_logs.extend(list(self._local.logs.keys()))
        except AttributeError:
            pass
        return oops_logs

    def add_record(self, record):
        """Log a record in this specific OOPSLog instance."""
        record = record.__dict__.copy()
        record.pop("exc_text")
        exc_info = record.pop("exc_info")
        if exc_info:
            record["oops_traceback"] = OOPSTraceback(*exc_info).get_dump_data()
        self._records.append(record)

    def get_dump_data(self):
        return {"records": self._records}
