# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
import threading

import oops
import oops_amqp
from oops_datedir_repo import DateDirRepo
from zope.interface import implementer

from canonical.oops.interfaces import IOOPSManager
from canonical.oops.lsoops import OOPS

from canonical.landscape.model.configuration import get_revision


@implementer(IOOPSManager)
class OOPSManager(object):
    """
    The OOPS manager is an utility which keeps track of all the OOPSes
    in progress in the system by well known names, so that any piece of
    code is able to retrieve a given OOPS and enrich it with details.

    @ivar _oops_config: An instance of oops.Config
    """

    def __init__(self, key, soft_timeout=None, automatic_removal=True):
        self.key = key
        self._oops = {}
        self._local = threading.local()
        self.soft_timeout = soft_timeout
        if self.soft_timeout is not None:
            self.soft_timeout = int(self.soft_timeout)
        self.automatic_removal = automatic_removal
        self._oops_config = self.configure()

    def add(self, name, thread_local=False):
        """Create an OOPS with the given name.

        @param thread_local: If True, the OOPS is available globally, and
            get() requests will be able to obtain it from any thread.
        """
        oops = OOPS()
        if thread_local:
            setattr(self._local, name, oops)
        else:
            self._oops[name] = oops
        return oops

    def get(self, name):
        """Get the OOPS instance registered with the given name."""
        oops = getattr(self._local, name, None)
        if oops is None:
            oops = self._oops.get(name)
        return oops

    def remove(self, name):
        """Remove the OOPS instance registered with the given name."""
        if name in self._oops:
            del self._oops[name]
        if hasattr(self._local, name):
            delattr(self._local, name)

    def configure(self, config_factory=oops.Config, get_revision=get_revision):
        """Configure the OOPS manager using the named section from the config.

        @param config_factory: Optional factory to make an oops.Config.
        """
        oops_config = config_factory()

        # What do we want in our reports?
        # Constants:
        oops_config.template["revno"] = get_revision()
        oops_config.template["reporter"] = self.key
        return oops_config

    def _add_publisher(self, publisher):
        """Add a publisher for OOPSes."""
        self._oops_config.publishers.append(publisher)

    def add_local_publishing(self, path):
        """Publish OOPSes to local path.

        @param path: The local path to dump dated OOPSes to.
        """
        if not os.path.isdir(path):
            os.makedirs(path)
        oops_datedir_repo = DateDirRepo(path)
        self._add_publisher(oops_datedir_repo.publish)

    def add_remote_publishing(self, rabbit_factory, exchange, routing_key):
        """Publish OOPSes to AMQP.

        @param rabbit_factory: A factory for creating a connection to RabbitMQ.
        @param exchange: The exchange name to submit the OOPSes to.
        @param routing_key: The routing key to submit the OOPses to.
        """
        amqp_publisher = oops_amqp.Publisher(rabbit_factory, exchange, routing_key)
        self._add_publisher(amqp_publisher)

    def dump(self, oops):
        """Publish the given OOPS instance."""
        pyoops = self._oops_config.create()
        pyoops.update(munge_oops(oops.get_dump_data()))
        self._oops_config.publish(pyoops)
        return pyoops["id"]


def munge_oops(lsoops_data):
    """Munge the data from Landscape to Launchpad-like OOPS format."""
    storm = lsoops_data.pop("storm", {})
    lsoops_data["timeline"] = list(
        _generate_munged_timeline(storm.get("statements", []))
    )
    metadata = lsoops_data.pop("metadata", {})
    lsoops_data["url"] = metadata.get("url")
    lsoops_data["duration"] = metadata.get("request-duration")
    for record in lsoops_data.get("log", {}).get("records", []):
        if "oops_traceback" in record:
            oops_traceback = record.pop("oops_traceback")
            lsoops_data["tb_text"] = oops_traceback["traceback"]
            break
    return lsoops_data


def _generate_munged_timeline(storm_statements):
    """Generate tuples suitable for timeline from Storm tracer statements."""
    # OOPSes can store an identification for the database each query was
    # executed on: since we do not have that in Landscape, use None everywhere.
    database_id = None
    base_start_time = None
    for statement, params, start_time, end_time, error in storm_statements:
        if base_start_time is None:
            base_start_time = start_time
        start_time_ms = (start_time - base_start_time) * 1000
        end_time_ms = (end_time - base_start_time) * 1000
        yield (start_time_ms, end_time_ms, database_id, statement % params, error)
