# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import threading
import weakref
import time

from storm.variables import Variable


class OOPSStormTracer(object):
    """Tracer registered globally with Storm to log executed statements.

    This tracer will forward all executed statements to each of the
    instantiated OOPSStorm entries.
    """

    # For testing purposes allow this to be overriden.
    now = time.time

    def connection_raw_execute(self, connection, raw_cursor, statement, params):

        now = self.now()
        for entry in OOPSStorm.get_for_tracing():
            entry.connection_raw_execute(now, connection, raw_cursor, statement, params)

    def connection_raw_execute_error(
        self, connection, raw_cursor, statement, params, error
    ):
        now = self.now()
        for entry in OOPSStorm.get_for_tracing():
            entry.connection_raw_execute_error(
                now, connection, raw_cursor, statement, params, error
            )

    def connection_raw_execute_success(self, connection, raw_cursor, statement, params):
        now = self.now()
        for entry in OOPSStorm.get_for_tracing():
            entry.connection_raw_execute_success(
                now, connection, raw_cursor, statement, params
            )


class OOPSStorm(object):
    """OOPS entry which keeps track of executed statements.

    It may be registered globally or in the current thread only.  In the
    later case, only statements executed in the current thread will be
    recorded.
    """

    entry_name = "storm"
    entry_version = 1

    _global = weakref.WeakKeyDictionary()
    _local = threading.local()

    def __init__(self, thread_local=False):
        # Each entry look like [statement, params, start time, end time, error]
        self._statements = []

        if thread_local:
            try:
                logs = self._local.logs
            except AttributeError:
                logs = self._local.logs = weakref.WeakKeyDictionary()
            logs[self] = True
        else:
            self._global[self] = True

    @classmethod
    def get_for_tracing(cls):
        """
        Return all OOPSStorms registered for the current thread plus globals.
        """
        for entry in cls._global:
            yield entry
        try:
            logs = cls._local.logs
        except AttributeError:
            pass
        else:
            for entry in logs:
                yield entry

    def connection_raw_execute(self, now, connection, raw_cursor, statement, params):
        """Record an executed statement in this OOPS entry."""
        # This isn't beautiful, but it's fast. Sorry! :-(
        params = tuple(
            repr(
                (isinstance(param, Variable) and (param.get(to_db=True),) or (param,))[
                    0
                ]
            )
            for param in params
        )
        self._statements.append([statement, params, now, None, None])

    def connection_raw_execute_error(
        self, now, connection, raw_cursor, statement, params, error
    ):
        """Record the error raised when the given statement was executed."""
        last = self._statements[-1]
        last[-2] = now
        last[-1] = str(error)

    def connection_raw_execute_success(
        self, now, connection, raw_cursor, statement, params
    ):
        """
        Record the timing of the successful execution of the given statement.
        """
        last = self._statements[-1]
        last[-2] = now

    def get_dump_data(self):
        statements = []
        for item in self._statements:
            statements.append(tuple(item))
        return {"statements": statements}
