# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import logging
import time

from zope.component import queryUtility
from zope.location.interfaces import ISite
from zope.publisher.interfaces.http import IHTTPApplicationRequest

from storm.tracer import install_tracer, remove_tracer_type

from canonical.oops.interfaces import IOOPSManager
from canonical.oops.manager import OOPSManager
from canonical.oops.stormtracer import OOPSStorm, OOPSStormTracer
from canonical.oops.lsoops import OOPSLog, OOPSMetaData, OOPSLoggingHandler


def start_request(event):
    """Register the "request" OOPS with the OOPSManager utility."""
    if ISite.providedBy(event.object):
        manager = queryUtility(IOOPSManager)
        if manager:
            oops = manager.add("request", thread_local=True)
            oops.add(OOPSStorm(thread_local=True))
            oops.add(OOPSLog(thread_local=True))
            metadata = OOPSMetaData()
            metadata["request-start"] = time.time()
            oops.add(metadata)


def end_request(event):
    """
    Remove the "request" OOPS from the OOPSManager utility, maybe dumping it.
    """
    manager = queryUtility(IOOPSManager)
    if manager:
        oops = manager.get("request")
        if oops and oops.get("metadata").get("id") is None:
            metadata = oops.get("metadata")
            request_end = time.time()
            metadata["request-end"] = request_end
            if IHTTPApplicationRequest.providedBy(event.request):
                metadata["url"] = event.request.getURL()
            request_start = metadata["request-start"]
            request_duration = request_end - request_start
            metadata["request-duration"] = request_duration
            dump_oops = False
            soft_oops = False
            if event.request is not None:
                dump_oops = event.request.get("dump_oops") == "1"
            if not dump_oops and not oops.wanted and manager.soft_timeout is not None:
                soft_oops = request_duration > manager.soft_timeout
            if dump_oops or oops.wanted or soft_oops:
                oops_id = ", ".join(manager.dump(oops))
                if dump_oops:
                    logging.info("Dump of %s explicitly requested.", oops_id)
                elif soft_oops:
                    logging.error("Dumping soft %s.", oops_id)
                else:
                    logging.error("Dumping %s.", oops_id)
            if manager.automatic_removal:
                manager.remove("request")


def create_standard_oops_manager(key, soft_timeout):
    """Create a standard L{OOPSManager}.

    @param key: The C{string} key used to categorise the OOPS.
    @param soft_timeout: The C{int} value of the soft timeout, or C{None}.
    """
    remove_tracer_type(OOPSStormTracer)
    install_tracer(OOPSStormTracer())
    root_logger = logging.getLogger()
    for handler in root_logger.handlers:
        if isinstance(handler, OOPSLoggingHandler):
            break
    else:
        root_logger.addHandler(OOPSLoggingHandler())
    manager = OOPSManager(key, soft_timeout)
    return manager
