# Copyright 2005-2022 Canonical Limited. All rights reserved.

import time
import cgitb
import html
import sys
import urllib.parse

from openid.extensions import ax, sreg
from openid.server import server
from openid.store.memstore import MemoryStore
from openid.consumer import discover

from canonical.landscape.ui.lsopenid import OpenIDAuth


class OpenIdProviderMiddleware(object):
    """
    This middleware is used to create simple OpenID provider.
    """

    def __init__(self, app, base_url):
        self.app = app
        self.base_url = base_url
        store = MemoryStore()
        self.openid = server.Server(store, self.base_url + "openidserver")
        self.approved = {}
        self.lastCheckIDRequest = {}

    def __call__(self, environ, start_response):
        path = environ["PATH_INFO"]
        if path.startswith(("/openid", "/yadis/", "/id/")):
            handler = OpenIdRequestHandler(environ, start_response, self)
            return (handler(),)
        else:
            return self.app(environ, start_response)


class OpenIdRequestHandler(object):
    """
    Handles a single request to the OpenIdServerMiddleware
    """

    def __init__(self, environ, start_response, server):
        self.environ = environ
        self.start_response = start_response
        self.status = "200 OK"
        self.headers = []
        self.result = ""
        self.server = server
        self.user = None
        self.fullname = "Kurt Lee"
        self.timezone = "Europe/Paris"
        self.nickname = "klee123"
        self.ax_attrs = {}
        self.default_emails = "klee123@yahoo.com, klee123@hotmail.com"

    def __call__(self):
        if self.environ.get("REQUEST_METHOD", "GET") == "POST":
            self.do_POST()
        else:
            self.do_GET()
        self.start_response(self.status, self.headers)
        return self.result.encode()

    def set_emails_ax(self, emails):
        """Helper function to set the ax field for verified emails"""
        if emails:
            self.ax_attrs = {OpenIDAuth.AX_EXTRA_EMAILS_URI: emails}

    def do_GET(self):
        try:
            self.query = dict(
                urllib.parse.parse_qsl(self.environ.get("QUERY_STRING", ""))
            )

            path = self.environ["PATH_INFO"]
            if path == "/openid":
                self.showMainPage()
            elif path == "/openidserver":
                self.serverEndPoint(self.query)
            elif path == "/openidlogout":
                self.doLogin()
            elif path.startswith("/id/"):
                self.showIdPage(path)
            elif path.startswith("/yadis/"):
                self.showYadis(path[7:])
            elif path == "/openidserveryadis":
                self.showServerYadis()
            else:
                self.set_status(404)
        except:
            self.set_status(500)
            self.set_header("Content-type", "text/html")
            self.result = cgitb.html(sys.exc_info(), context=10)

    def do_POST(self):
        try:
            content_length = int(self.environ["CONTENT_LENGTH"])
            post_data = self.environ["wsgi.input"].read(content_length).decode()

            self.query = {}
            for k, v in urllib.parse.parse_qsl(post_data):
                self.query[k] = v

            path = self.environ["PATH_INFO"]
            if path == "/openidserver":
                self.serverEndPoint(self.query)

            elif path == "/openidallow":
                self.handleAllow(self.query)
            else:
                self.set_status(404)
        except:
            self.set_status(500)
            self.set_header("Content-type", "text/html")
            self.result = cgitb.html(sys.exc_info(), context=10)

    def handleAllow(self, query):
        # pretend this next bit is keying off the user's session or something,
        # right?
        request = self.server.lastCheckIDRequest.get(self.user)

        if "yes" in query:
            if "login_as" in query:
                self.user = self.query["login_as"]

            if request.idSelect():
                identity = query["identity"]
            else:
                identity = request.identity

            self.set_emails_ax(query.get("emails"))

            trust_root = request.trust_root
            if self.query.get("remember", "no") == "yes":
                self.server.approved[(identity, trust_root)] = "always"

            # Set a hopefully unique email address using the username
            # provided in the identity
            id_url_base = self.server.base_url + "id/"
            self.email = identity[len(id_url_base) :] + "@example.com"

            response = self.approved(request, identity)

        elif "no" in query:
            response = request.answer(False)

        else:
            assert False, "strange allow post.  %r" % (query,)

        self.displayResponse(response)

    def isAuthorized(self, identity_url, trust_root):
        if self.user is None:
            return False

        if identity_url != self.server.base_url + "id/" + self.user:
            return False

        key = (identity_url, trust_root)
        return self.server.approved.get(key) is not None

    def serverEndPoint(self, query):
        try:
            request = self.server.openid.decodeRequest(query)
        except server.ProtocolError as why:
            self.displayResponse(why)
            return

        if request is None:
            # Display text indicating that this is an endpoint.
            self.showAboutPage()
            return

        if request.mode in ["checkid_immediate", "checkid_setup"]:
            self.handleCheckIDRequest(request)
        else:
            response = self.server.openid.handleRequest(request)
            self.displayResponse(response)

    def addSRegResponse(self, request, response):
        sreg_req = sreg.SRegRequest.fromOpenIDRequest(request)

        sreg_data = {}
        for field in sreg_req.allRequestedFields():
            # None is returned to prevent exception when sreg field missing
            value = getattr(self, field, None)
            if value:
                sreg_data[field] = value

        sreg_resp = sreg.SRegResponse.extractResponse(sreg_req, sreg_data)
        response.addExtension(sreg_resp)

    def addAxResponse(self, request, response):
        ax_req = ax.FetchRequest.fromOpenIDRequest(request)

        ax_resp = ax.FetchResponse()
        for item in ax_req.iterAttrs():
            key = item.type_uri
            if key in self.ax_attrs:
                ax_resp.addValue(key, self.ax_attrs[key])

        response.addExtension(ax_resp)

    def approved(self, request, identifier=None):
        response = request.answer(True, identity=identifier)
        self.addSRegResponse(request, response)
        self.addAxResponse(request, response)
        return response

    def handleCheckIDRequest(self, request):
        is_authorized = self.isAuthorized(request.identity, request.trust_root)
        if is_authorized:
            response = self.approved(request)
            self.displayResponse(response)
        elif request.immediate:
            response = request.answer(False)
            self.displayResponse(response)
        else:
            self.server.lastCheckIDRequest[self.user] = request
            self.showDecidePage(request)

    def set_status(self, code):
        from twisted.web.http import RESPONSES

        self.status = "%d %s" % (code, RESPONSES[code])

    def set_header(self, header, value):
        self.headers.append((header, value))

    def displayResponse(self, response):
        try:
            webresponse = self.server.openid.encodeResponse(response)
        except server.EncodingError as why:
            text = why.response.encodeToKVForm()
            self.showErrorPage("<pre>%s</pre>" % html.escape(text))
            return

        self.set_status(webresponse.code)
        for header, value in list(webresponse.headers.items()):
            self.set_header(header, value)
        self.set_header("Content-type", "text/html")
        self.writeUserHeader()

        if webresponse.body:
            self.result = webresponse.body

    def doLogin(self):
        if "submit" in self.query:
            if "user" in self.query:
                self.user = self.query["user"]
            else:
                self.user = None
            self.redirect(self.query["success_to"])
        elif "cancel" in self.query:
            self.redirect(self.query["fail_to"])
        elif "return_to" in self.query:
            self.redirect(self.query["return_to"])
        else:
            assert 0, "strange login %r" % (self.query,)

    def redirect(self, url):
        self.set_status(302)
        self.set_header("Content-Type", "text/html")
        self.set_header("Location", url)
        self.writeUserHeader()

    def writeUserHeader(self):
        if self.user is None:
            t1970 = time.gmtime(0)
            expires = time.strftime("Expires=%a, %d-%b-%y %H:%M:%S GMT", t1970)
            self.set_header("Set-Cookie", "user=;%s" % expires)
        else:
            self.set_header("Set-Cookie", "user=%s" % self.user)

    def quoteattr(self, s):
        qs = html.escape(s, 1)
        return '"%s"' % (qs,)

    def showAboutPage(self):
        endpoint_url = self.server.base_url + "openidserver"

        def link(url):
            url_attr = self.quoteattr(url)
            url_text = html.escape(url)
            return "<a href=%s><code>%s</code></a>" % (url_attr, url_text)

        def term(url, text):
            return "<dt>%s</dt><dd>%s</dd>" % (link(url), text)

        resources = [
            (self.server.base_url, "This example server's home page"),
            (
                "http://www.openidenabled.com/",
                "An OpenID community Web site, home of this library",
            ),
            ("http://www.openid.net/", "the official OpenID Web site"),
        ]

        resource_markup = "".join([term(url, text) for url, text in resources])

        self.showPage(
            200,
            "This is an OpenID server",
            msg="""\
        <p>%s is an OpenID server endpoint.<p>
        <p>For more information about OpenID, see:</p>
        <dl>
        %s
        </dl>
        """
            % (
                link(endpoint_url),
                resource_markup,
            ),
        )

    def showErrorPage(self, error_message):
        self.showPage(
            400,
            "Error Processing Request",
            err="""\
        <p>%s</p>
        """
            % error_message,
        )

    def showDecidePage(self, request):
        id_url_base = self.server.base_url + "id/"
        # XXX: This may break if there are any synonyms for id_url_base,
        # such as referring to it by IP address or a CNAME.
        assert request.identity.startswith(id_url_base) or request.idSelect(), repr(
            (request.identity, id_url_base)
        )
        expected_user = request.identity[len(id_url_base) :]

        if request.idSelect():  # We are being asked to select an ID
            fdata = {
                "id_url_base": id_url_base,
                "trust_root": request.trust_root,
                "emails": str(self.default_emails),
            }
            form = (
                """\
            <form method="POST" action="/openidallow">
            <table>
              <tr>
                <td>Identity:</td>
                <td>
                  <input id='identity' type='text' name='identity' size='30'>
                </td>
              </tr>
              <tr>
                <td>Additional Emails:</td>
                <td>
                    <input id='emails' type='text' name='emails' size='30'
                     value='%(emails)s'>
                </td>
              </tr>
            </table>
            <p>Do you want this authentication to succeed or fail?</p>
            <input type="hidden" id="remember" name="remember" value="yes"
                />
            <input type="submit" id="success-button"
                   name="yes" value="Succeed" />
            </form>
            <form method="POST" action="/openidallow">
            <input type="hidden" name="fake-openid-failure" value="yes" />
            <input type="submit" name="no" value="Fail" />
            </form>
            <script type="text/javascript" defer="defer">
              var identityInput = document.getElementById("identity");
              identityInput.value = '%(id_url_base)s';
              identityInput.focus();
            </script>
            """
                % fdata
            )
        elif expected_user == self.user:
            fdata = {"identity": request.identity, "trust_root": request.trust_root}
            form = (
                """\
            <table>
              <tr><td>Identity:</td><td>%(identity)s</td></tr>
              <tr><td>Trust Root:</td><td>%(trust_root)s</td></tr>
            </table>
            <p>Allow this authentication to proceed?</p>
            <form method="POST" action="/openidallow">
              <input type="checkbox" id="remember" name="remember" value="yes"
                  /><label for="remember">Remember this
                  decision</label><br />
              <input type="submit" name="yes" value="yes" />
              <input type="submit" name="no" value="no" />
            </form>"""
                % fdata
            )
        else:
            fdata = {
                "identity": request.identity,
                "trust_root": request.trust_root,
                "expected_user": expected_user,
            }
            form = (
                """\
            <table>
              <tr><td>Identity:</td><td>%(identity)s</td></tr>
            </table>
            <p>Allow this authentication to proceed?</p>
            <form method="POST" action="/openidallow">
              <input type="checkbox" id="remember" name="remember" value="yes"
                  /><label for="remember">Remember this
                  decision</label><br />
              <input type="hidden" name="login_as" value="%(expected_user)s"/>
              <input type="submit" name="yes" value="yes" />
              <input type="submit" name="no" value="no" />
            </form>"""
                % fdata
            )

        self.showPage(200, "Decide", form=form)

    def showIdPage(self, path):
        link_tag = '<link rel="openid.server" href="%s/openidserver">' % (
            self.server.base_url
        )
        yadis_loc_tag = '<meta http-equiv="x-xrds-location" content="%s">' % (
            self.server.base_url + "yadis/" + path[4:]
        )
        disco_tags = link_tag + yadis_loc_tag
        ident = self.server.base_url + path[1:]

        approved_trust_roots = []
        for (aident, trust_root) in list(self.server.approved.keys()):
            if aident == ident:
                trs = "<li><tt>%s</tt></li>\n" % html.escape(trust_root)
                approved_trust_roots.append(trs)

        if approved_trust_roots:
            prepend = "<p>Approved trust roots:</p>\n<ul>\n"
            approved_trust_roots.insert(0, prepend)
            approved_trust_roots.append("</ul>\n")
            msg = "".join(approved_trust_roots)
        else:
            msg = ""

        self.showPage(
            200,
            "An Identity Page",
            head_extras=disco_tags,
            msg="""\
        <p>This is an identity page for %s.</p>
        %s
        """
            % (ident, msg),
        )

    def showYadis(self, user):
        self.set_status(200)
        self.set_header("Content-type", "application/xrds+xml")

        endpoint_url = self.server.base_url + "openidserver"
        user_url = self.server.base_url + "id/" + user
        self.result = """\
<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS
    xmlns:xrds="xri://$xrds"
    xmlns="xri://$xrd*($v*2.0)">
  <XRD>

    <Service priority="0">
      <Type>%s</Type>
      <Type>%s</Type>
      <URI>%s</URI>
      <LocalID>%s</LocalID>
    </Service>

  </XRD>
</xrds:XRDS>
""" % (
            discover.OPENID_2_0_TYPE,
            discover.OPENID_1_0_TYPE,
            endpoint_url,
            user_url,
        )

    def showServerYadis(self):
        self.headers = [("Content-type", "application/xrds+xml")]

        endpoint_url = self.server.base_url + "openidserver"
        self.result = """\
<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS
    xmlns:xrds="xri://$xrds"
    xmlns="xri://$xrd*($v*2.0)">
  <XRD>

    <Service priority="0">
      <Type>%s</Type>
      <URI>%s</URI>
    </Service>

  </XRD>
</xrds:XRDS>
""" % (
            discover.OPENID_IDP_2_0_TYPE,
            endpoint_url,
        )

    def showMainPage(self):
        yadis_tag = '<meta http-equiv="x-xrds-location" content="%s">' % (
            self.server.base_url + "openidserveryadis"
        )
        if self.user:
            openid_url = self.server.base_url + "id/" + self.user
            user_message = """\
            <p>You are logged in as %s. Your OpenID identity URL is
            <tt><a href=%s>%s</a></tt>. Enter that URL at an OpenID
            consumer to test this server.</p>
            """ % (
                self.user,
                self.quoteattr(openid_url),
                openid_url,
            )
        else:
            user_message = """\
            <p>This server uses a cookie to remember who you are in
            order to simulate a standard Web user experience. You are
            not <a href='/login'>logged in</a>.</p>"""

        self.showPage(
            200,
            "Main Page",
            head_extras=yadis_tag,
            msg="""\
        <p>This is a simple OpenID server implemented using the <a
        href="http://openid.schtuff.com/">Python OpenID
        library</a>.</p>

        %s

        <p>To use this server with a consumer, the consumer must be
        able to fetch HTTP pages from this web server. If this
        computer is behind a firewall, you will not be able to use
        OpenID consumers outside of the firewall with it.</p>

        <p>The URL for this server is <a href=%s><tt>%s</tt></a>.</p>
        """
            % (
                user_message,
                self.quoteattr(self.server.base_url),
                self.server.base_url,
            ),
        )

    def showPage(
        self, response_code, title, head_extras="", msg=None, err=None, form=None
    ):

        body = ""

        if err is not None:
            body += (
                """\
            <div class="error">
              %s
            </div>
            """
                % err
            )

        if msg is not None:
            body += (
                """\
            <div class="message">
              %s
            </div>
            """
                % msg
            )

        if form is not None:
            body += (
                """\
            <div class="form">
              %s
            </div>
            """
                % form
            )

        contents = {
            "title": "Fake OpenID provider - " + title,
            "head_extras": head_extras,
            "body": body,
        }

        self.headers = [("Content-type", "text/html")]
        self.result = (
            """<html>
  <head>
    <title>%(title)s</title>
    %(head_extras)s
  </head>
  <style type="text/css">
      h1 a:link {
          color: black;
          text-decoration: none;
      }
      h1 a:visited {
          color: black;
          text-decoration: none;
      }
      h1 a:hover {
          text-decoration: underline;
      }
      body {
        font-family: verdana,sans-serif;
        width: 50em;
        margin: 1em;
      }
      div {
        padding: .5em;
      }
      table {
        margin: none;
        padding: none;
      }
      .banner {
        padding: none 1em 1em 1em;
        width: 100%%;
      }
      .leftbanner {
        text-align: left;
      }
      .rightbanner {
        text-align: right;
        font-size: smaller;
      }
      .error {
        border: 1px solid #ff0000;
        background: #ffaaaa;
        margin: .5em;
      }
      .message {
        border: 1px solid #2233ff;
        background: #eeeeff;
        margin: .5em;
      }
      .form {
        border: 1px solid #777777;
        background: #ddddcc;
        padding-bottom: 0em;
      }
      dd {
        margin-bottom: 0.5em;
      }
  </style>
  <body>
    <table class="banner">
      <tr>
        <td class="leftbanner">
          <h1><a href="/">Fake OpenID provider</a></h1>
        </td>
      </tr>
    </table>
%(body)s
  </body>
</html>
"""
            % contents
        )
