# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import json
from urllib.parse import urlencode

from treq import request, content

from twisted.python.failure import Failure
from twisted.python import log
from twisted.internet import reactor
from twisted.internet.task import deferLater


class HTTPError(Exception):
    """A non-200 HTTP code was returned.

    @param code: The HTTP status code.
    @param phrase: The HTTP response phrase (the "OK" in "200 OK").
    @param content: The response body, as a C{str}.
    @param url: The URL that was originally requested.
    """

    def __init__(self, method, code, content, url, phrase):
        self.method = method
        self.code = code
        self.phrase = phrase
        self.content = content
        self.url = url

    def __str__(self):
        return "%s request for %s failed with code %s %s: %r" % (
            self.method,
            self.url,
            self.code,
            self.phrase,
            self.content,
        )


class _Request(object):
    """
    A completely private encapsulation of state and logic associated with
    performing a "single" request (with retries).  This only exists as a
    factoring convenience, to make it easy to store all of the ephemeral data
    associated with a request instead of passing it around between several
    private methods.

    This class is implementation-bound to L{JSONClient}, and uses its private
    attributes and methods.
    """

    def __init__(self, client, method, url, headers, body, max_blind_retries):
        self.client = client
        self.method = method
        self.url = url
        self.headers = headers
        self.body = body
        self.auth_retry_attempts = 0
        self.blind_retry_attempts = 0
        self.max_blind_retries = max_blind_retries

    def try_request(self, auth_retry=False, blind_retry=False):
        """Attempt to do a request.

        If a 401 is received, reauth and retry.  If any other error is
        received, retry.

        @param auth_retry: If True, treat this request as an auth retry.
        @param blind_retry: If True, treat this request as a blind retry.
        """
        if auth_retry:
            self.auth_retry_attempts += 1
        if blind_retry:
            self.blind_retry_attempts += 1
        log.msg(
            "%s %s headers=%s auth_retry_attempts=%s "
            "blind_retry_attempts=%s"
            % (
                self.method,
                self.url,
                self.headers,
                self.auth_retry_attempts,
                self.blind_retry_attempts,
            ),
            system="OSAPI",
        )
        request_headers = self.client.default_headers.copy()
        if self.headers is not None:
            request_headers.update(self.headers)
        req = self.client._request(
            self.method,
            self.url,
            headers=request_headers,
            data=self.body,
            persistent=False,
        )  # Apparently swift doesn't like persistent
        # connections, see #1494710
        result = req.addCallback(self._got_response)
        result.addErrback(self._got_error)
        return result

    def _extract_content(self, response):
        """
        Extract the content from the HTTP response, either as a Python object
        loaded from JSON, a string, or None.
        """

        def _decode(result):
            if result:
                return json.loads(result)

        result = content(response)

        is_json = False
        for content_type in response.headers.getRawHeaders("content-type", []):
            if ";" in content_type:
                content_type, _ = content_type.split(";", 1)
            if content_type == "application/json":
                is_json = True
        if is_json:
            result = result.addCallback(_decode)
        return result

    def _got_response(self, response):
        if response.code == 401:
            if (
                self.client._reauth_function is not None
                and self.auth_retry_attempts < self.client._auth_retries
            ):
                result = self.client._reauth_function(self.client)
                result.addCallback(lambda ignored: self.try_request(auth_retry=True))
                return result
        if response.code < 200 or response.code >= 300:
            text_content = content(response)
            text_content.addCallback(
                lambda content: Failure(
                    HTTPError(
                        self.method, response.code, content, self.url, response.phrase
                    )
                )
            )
            return text_content
        else:
            return self._extract_content(response)

    def _got_error(self, failure):
        if self.blind_retry_attempts < self.max_blind_retries:
            log.err(
                failure,
                "Retrying after failure: %s %s "
                "headers=%s retry_attempts=%s"
                % (self.method, self.url, self.headers, self.blind_retry_attempts),
                system="OSAPI",
            )
            delay = deferLater(
                self.client._clock,
                self.client._blind_retry_delay,
                self.try_request,
                blind_retry=True,
            )
            return delay
        else:
            return failure


class OpenStackClient(object):
    """
    An HTTP client that has a convenience for setting the x-auth-token
    header, which is specific to OpenStack and retrying requests.
    """

    _default_url_encoding = "utf-8"
    _reauth_function = None

    def __init__(
        self,
        endpoint,
        auth_token=None,
        _request=request,
        blind_retries=0,
        blind_retry_delay=0,
        clock=reactor,
        auth_retries=1,
    ):
        """
        @param endpoint: A base URL to use for requests.
        @param auth_token: If specified, it will be set as the default value of
            the x-auth-token header, and set as the C{auth_token} attribute.
        @param blind_retries: the maximum number of retries to perform in the
            event of *any* failure. This is an *evil hack*, because we don't
            really know if requests performed are idempotent, or whether an
            error return value means that they have had no effect.
        @param blind_retry_delay: The delay between blind retries.
        """
        self.endpoint = endpoint
        self.default_headers = {
            "user-agent": "Landscape OpenStack API client",
        }
        if auth_token is not None:
            self.set_auth_token(auth_token)
        self._request = _request
        self._blind_retries = blind_retries
        self._blind_retry_delay = blind_retry_delay
        self._clock = clock
        self._auth_retries = auth_retries

    def set_requester(self, request):
        """
        For testing purposes. Specify the method to perform HTTP requests.
        """
        self._request = request

    def _queryize(self, parameters):
        """
        Convert a dict like::

            {"a": "b",
             "multi": ["foo", "bar"]}

        to a list of two-tuples like::

            [("a", "b"),
             ("multi", "foo"),
             ("multi", "bar")]
        """
        result = []
        for parameter, value in list(parameters.items()):
            if isinstance(value, list):
                for element in value:
                    result.append((parameter, element))
            else:
                result.append((parameter, value))
        return result

    def _concatenate_url(self, base, suffix, parameters):
        """
        Concatenate two segments of a URL, taking care not to duplicate "/"
        characters, and adding any parameters to the end.
        """
        base = base.rstrip("/")
        result = base + "/" + suffix
        if parameters is not None:
            result += "?" + urlencode(self._queryize(parameters))
        return result

    def set_auth_token(self, auth_token):
        """Set the X-Auth-Token header for future requests with this client."""
        if isinstance(auth_token, str):
            # This should never be non-ascii. I'm doing it because the token is
            # usually retrieved from a json document which means it's unicode.
            auth_token = auth_token.encode("ascii")
        self.default_headers["x-auth-token"] = auth_token

    def request(
        self,
        method,
        resource,
        headers=None,
        body=None,
        parameters=None,
        blind_retries=None,
    ):
        """Perform an HTTP request."""
        url = self._concatenate_url(self.endpoint, resource, parameters)
        max_blind_retries = self._blind_retries
        if blind_retries is not None:
            max_blind_retries = blind_retries
        request = _Request(self, method, url, headers, body, max_blind_retries)
        return request.try_request()

    def set_reauth_function(self, function):
        """
        Set a function to be invoked to try reauthenticating in the event of an
        HTTP 401 response.

        OpenStack tokens expire occasionally.  A request using an expired token
        will return a 401. The passed in will be invoked in this event, to
        provide an opportunity to reauthenticate and allow seemlessly retrying
        the original request that caused the original 401.
        """
        self._reauth_function = function


class JSONClient(OpenStackClient):
    """An HTTP client which defaults to sending and accepting JSON."""

    def __init__(self, *args, **kwargs):
        super(JSONClient, self).__init__(*args, **kwargs)
        self.default_headers["content-type"] = "application/json"
        self.default_headers["accept"] = "application/json"

    def json_request(
        self,
        method,
        resource,
        headers=None,
        body=None,
        parameters=None,
        blind_retries=None,
    ):
        """
        Perform an HTTP request with a JSON request body and a JSON response
        body.
        """
        if body is not None:
            body = json.dumps(body).encode()
        return self.request(method, resource, headers, body, parameters, blind_retries)
