# Copyright 2005-2014 Canonical Limited.  All rights reserved.

from functools import partial

from twisted.internet.defer import inlineCallbacks, returnValue

from canonical.openstack.keystone import KeystoneClient
from canonical.openstack.neutron import NeutronClient
from canonical.openstack.nova import NovaClient
from canonical.openstack.glance import GlanceClient
from canonical.openstack.swift import SwiftClient


class ServiceNotFound(Exception):
    """A service couldn't be found in the keystone service catalog."""

    def __init__(self, service):
        self.service = service

    def __str__(self):
        return (
            "The %s service couldn't be found in the Keystone service "
            "catalog." % (self.service)
        )


class OpenStackClientsHelper(object):
    """A helper for getting OpenStack clients.

    It will authenticate with keystone, requesting a token and endpoint,
    before returning the requested client.
    """

    def __init__(self):
        self.clients = {
            "image": GlanceClient,
            "identity": KeystoneClient,
            "network": NeutronClient,
            "object-store": SwiftClient,
            "compute": NovaClient,
        }

    @inlineCallbacks
    def get_client(self, service, keystone_url, username, password, tenant):
        """Get an OpenStack client for a given service type.

        @param service: The type of the service as specified in the
            keystone catalog, e.g. "image", "network", "object-store".
        @param keystone_url: The URL to the keystone endpoint.
        @param username: The username to authenticate with.
        @param password: The password to authenticate with.
        @param tenant: The tenant to authenticate with.

        @raises ServiceNotFound: If the service isn't present in the
            keystone catalog.

        @returns: A deferred with the requested OpenStack client.
        """
        endpoint, token = yield self._get_authenticated_endpoint(
            service, keystone_url, username, password, tenant
        )
        client = self.clients[service](endpoint, token)
        client.set_reauth_function(
            partial(self._reauth, service, keystone_url, username, password, tenant)
        )
        returnValue(client)

    @inlineCallbacks
    def _get_authenticated_endpoint(
        self, service_type, keystone_url, username, password, tenant
    ):
        """Authenticate with keystone for the given service.

        @returns: A deferred with the endpoint and authentication token
            for the requested service.
        """
        client = self.clients["identity"](keystone_url)
        auth_info = yield client.authenticate(
            username=username, password=password, tenant=tenant
        )
        token = auth_info["access"]["token"]["id"]
        services = auth_info["access"]["serviceCatalog"]
        endpoint = None
        for service in services:
            if service["type"] == service_type:
                endpoint = service["endpoints"][0]["publicURL"]
                break
        else:
            raise ServiceNotFound(service_type)
        returnValue((endpoint, token))

    @inlineCallbacks
    def _reauth(self, service_type, keystone_url, username, password, tenant, client):
        """Reauthenticate the client.

        A new token will be gotten from keystone, and the token gets
        passed on to the client, which will use it for future requests.
        """
        _, token = yield self._get_authenticated_endpoint(
            service_type=service_type,
            keystone_url=keystone_url,
            username=username,
            password=password,
            tenant=tenant,
        )
        client.set_auth_token(token)
        returnValue(token)
