# Copyright 2005-2015 Canonical Limited.  All rights reserved.

import json

from twisted.web.error import Error
from twisted.web.resource import Resource
from twisted.web.http import BAD_REQUEST

KEYSTONE_PORT = 35357
GLANCE_PORT = 20000
NOVA_PORT = 20001
NEUTRON_PORT = 20002
SWIFT_PORT = 20003

AUTH_INFO = {
    "access": {
        "token": {"id": "x:y:z"},
        "serviceCatalog": [
            {
                "type": "identity",
                "endpoints": [{"publicURL": "http://localhost:%d" % KEYSTONE_PORT}],
            },
            {
                "type": "image",
                "endpoints": [{"publicURL": "http://localhost:%d" % GLANCE_PORT}],
            },
            {
                "type": "compute",
                "endpoints": [{"publicURL": "http://localhost:%d" % NOVA_PORT}],
            },
            {
                "type": "network",
                "endpoints": [{"publicURL": "http://localhost:%d" % NEUTRON_PORT}],
            },
            {
                "type": "object-store",
                "endpoints": [{"publicURL": "http://localhost:%d" % SWIFT_PORT}],
            },
        ],
    }
}


class FakeKeystone(Resource):

    isLeaf = True

    def render(self, request):
        method = request.method
        entity = request.path.split("/")[2]
        handler = getattr(self, "_%s_%s" % (method, entity))
        request.setHeader("Content-Type", "application/json")
        return json.dumps(handler(request))

    def _GET_projects(self, request):
        return {
            "projects": [
                {"name": "admin", "id": "xyz"},
                {"name": "landscape", "id": "xyz"},
                {"name": "person_project", "id": "xyz"},
            ]
        }

    def _GET_users(self, request):
        return {
            "users": [
                {"name": "landscape", "id": "xyz"},
                {"name": "person@example.com", "id": "xyz"},
            ]
        }

    def _GET_roles(self, request):
        return {
            "roles": [{"name": "Admin", "id": "xyz"}, {"name": "Member", "id": "xyz"}]
        }

    def _POST_tokens(self, request):
        return AUTH_INFO

    def _POST_services(self, request):
        return {"service": {"id": "xyz"}}

    def _POST_endpoints(self, request):
        return {}

    def _POST_projects(self, request):
        return {}

    def _POST_users(self, request):
        return {}

    def _PUT_projects(self, request):
        return {}


class FakeGlance(Resource):

    isLeaf = True

    def render_GET(self, request):
        # For now the only GET action we support is getting image details, and
        # we hard-wire two active images (one for KVM and one for LXD) so that
        # WaitForAutoSyncImageJobRunner always passes.
        request.setHeader("Content-Type", "application/json")
        return json.dumps(
            {
                "images": [
                    {"status": "active", "properties": {"hypervisor_type": "qemu"}},
                    {"status": "active", "properties": {"hypervisor_type": "lxc"}},
                ]
            }
        )

    def render_POST(self, request):
        # For now the only POST action we support is uploading images, which
        # doesn't need any return value
        request.setHeader("Content-Type", "application/json")
        return json.dumps({})


class FakeNova(Resource):

    isLeaf = True

    def render_GET(self, request):
        request.setHeader("Content-Type", "application/json")
        endpoint = request.postpath[0]
        if endpoint == "os-aggregates":
            response = {"aggregates": []}
        else:
            response = {"flavors": []}
        return json.dumps(response)

    def render_POST(self, request):
        request.setHeader("Content-Type", "application/json")
        return json.dumps({"aggregate": {"id": "fake-id"}})

    def render_PUT(self, request):
        request.setHeader("Content-Type", "application/json")
        return json.dumps({})


class FakeNeutron(Resource):

    # This list needs to match agents list as defined in
    # c.l.m.openstack.jobs.network.NetworkSetupHelper.
    # Look for _check_neutron_agents_up.
    NEUTRON_AGENTS_BINARIES = (
        "neutron-dhcp-agent",
        "neutron-l3-agent",
        "neutron-metadata-agent",
        "neutron-openvswitch-agent",
    )

    # XXX We'll report running neutron agents on all machines in the 'london'
    #     physical zone of fake-maas, which is the one used by integration
    #     tests. In reality there are only 3 neutron units in the deployment
    #     built by the test, however it's not possible to now upfront which
    #     of machines they'll use.
    NEUTRON_AGENTS_HOSTS = ("machine_a", "machine_b", "machine_c", "machine_e")

    isLeaf = True

    def __init__(self):
        self.networks = {}

    def render(self, request):
        if not request.path.startswith("/v2.0/"):
            raise Error(BAD_REQUEST)
        method = request.method
        entity = request.path.split("/")[2].replace("-", "_")
        handler = getattr(self, "_%s_%s" % (method, entity))
        request.setHeader("Content-Type", "application/json")
        return json.dumps(handler(request))

    def _GET_agents(self, request):
        agents = []
        for host in self.NEUTRON_AGENTS_HOSTS:
            for binary in self.NEUTRON_AGENTS_BINARIES:
                agents.append(
                    {
                        "host": host,
                        "binary": binary,
                        "alive": True,
                        "admin_state_up": True,
                    }
                )
        return {"agents": agents}

    def _GET_networks(self, request):
        return {"networks": list(self.networks.values())}

    def _GET_subnets(self, request):
        return {"subnets": []}

    def _GET_routers(self, request):
        return {"routers": []}

    def _GET_security_groups(self, request):
        return {"security_groups": []}

    def _POST_networks(self, request):
        network = json.loads(request.content.read())
        network_id = "net-id-%d" % len(self.networks)
        network["id"] = network_id
        self.networks[network_id] = network
        return {"network": {"id": network_id}}

    def _DELETE_networks(self, request):
        self.networks.pop(request.postpath[-1])
        return {}

    def _POST_subnets(self, request):
        return {"subnet": {"id": "xyz"}}

    def _POST_routers(self, request):
        return {"router": {"id": "xyz"}}

    def _PUT_routers(self, request):
        return {}

    def _PUT_quotas(self, request):
        return {}


class FakeSwift(Resource):

    isLeaf = True

    def render_PUT(self, request):
        return json.dumps({})
