# Copyright 2005-2015 Canonical Limited.  All rights reserved.

from canonical.openstack.api import JSONClient


class KeystoneClient(JSONClient):
    """
    A client for the Keystone authentication API.

    Most of the methods on this class work only with the Keystone "admin" API
    (port 35357), and not the regular user API (port 5000).  The exceptions are
    L{authenticate} and L{get_projects}.  The v3 API theoretically unifies all
    of this, so it should eventually become moot.
    """

    def __init__(self, endpoint, *args, **kwargs):
        # Strip the version from the endpoint, since KeystoneClient uses
        # a mix of v2.0 and v3 API calls. We still allow a versioned
        # endpoint to be passed in, since that's what all the other
        # OpenStack client expect.
        # XXX: Ideally we should require a versioned endoint and use
        #      only that version. If we have a v2.0 endpoint, there's no
        #      guarantee that the v3 API is available.
        if endpoint.endswith("/v2.0") or endpoint.endswith("/v3"):
            parts = endpoint.split("/")
            endpoint = "/".join(parts[:-1]) + "/"
        super(KeystoneClient, self).__init__(endpoint, *args, **kwargs)

    def authenticate(self, username, password, tenant):
        """Authenticate with the Keystone service.

        Returns a token, some metadata, and information about other
        OpenStack endpoints.
        """
        # We're not using v3 to authenticate because it *seems* that there's
        # no way to auth without knowing what domain you want to authenticate
        # to.
        data = {
            "auth": {
                "tenantName": tenant,
                "passwordCredentials": {"username": username, "password": password},
            }
        }
        return self.json_request("POST", "v2.0/tokens", body=data)

    def invalidate_token(self, token):
        """
        Invalidate a token.
        """
        # v3 version is "DELETE v3/auth/token" with x-subject-token header set.
        # I got a 404 when I tried it, so we're using v2.0.
        # This may only be possible when we switch to v3 for authenticating.
        return self.json_request("DELETE", "v2.0/tokens/%s" % (token,))

    def delete_project(self, project_id):
        """Delete the project given its ID."""
        return self.json_request("DELETE", "v3/projects/%s" % (project_id,))

    def add_project(self, name, description, domain_id="default"):
        """Add a project."""
        data = {
            "project": {
                "name": name,
                "description": description,
                "domain_id": domain_id,
                "enabled": True,
            }
        }
        return self.json_request("POST", "v3/projects", body=data)

    def get_projects(self):
        """Get the projects that a token has access to.

        Requires an auth token to have been set.
        """
        return self.json_request("GET", "v3/projects")

    def _filter(self, result, root_key, comparison):
        for item in result[root_key]:
            if comparison(item):
                return item

    def get_project_named(self, name):
        """Get a project with a specified name, case-insensitive.

        If it can't be found, None will be returned.
        """
        name = name.lower()
        return self.get_projects().addCallback(
            self._filter, "projects", lambda project: project["name"].lower() == name
        )

    def get_role_named(self, name):
        """Get a role with the specified name, case-insensitive.

        If it can't be found, None will be returned.
        """
        name = name.lower()
        return self.get_roles().addCallback(
            self._filter, "roles", lambda role: role["name"].lower() == name
        )

    def get_user_named(self, name):
        """Get a user with the specified name.

        If it can't be found, None will be returned.
        """
        return self.get_users().addCallback(
            self._filter, "users", lambda user: user["name"] == name
        )

    def get_token(self, token):
        """Return data about an already-authenticated token.

        This returns the same data as L{authenticate}, minus the service
        catalog.
        """
        # When upgrading to v3, we need to set X-Subject-Token and not pass it
        # in the URL.
        return self.json_request("GET", "v2.0/tokens/%s" % (token,))

    def get_endpoints(self, token):
        """Return the endpoints that an authenticated token can access.

        This returns the same data as the "serviceCatalog" returned from
        L{authenticate}.
        """
        # When upgrading to v3, we need to set X-Subject-Token and not pass it
        # in the URL.
        return self.json_request("GET", "v2.0/tokens/%s/endpoints" % (token,))

    def add_user(self, name, email, password, project_id=None):
        """Add a user."""
        data = {
            "user": {
                "name": name,
                "password": password,
                "email": email,
                "enabled": True,
            }
        }
        if project_id is not None:
            data["user"]["default_project_id"] = project_id

        return self.json_request("POST", "v3/users", body=data)

    def get_users(self):
        """Get a list of all users."""
        return self.json_request("GET", "v3/users")

    def delete_user(self, user_id):
        """Delete a user, given its ID."""
        return self.json_request("DELETE", "v3/users/%s" % (user_id,))

    def get_user_roles(self, user_id):
        """Get the roles of the user, given its ID."""
        return self.json_request("GET", "v3/users/%s/roles" % (user_id,))

    def get_user_roles_in_project(self, user_id, tenant_id):
        """
        Get the roles of a user in a project, given it's user ID and tenant ID.
        """
        return self.json_request(
            "GET", "v2.0/tenants/%s/users/%s/roles" % (tenant_id, user_id)
        )

    def get_user_projects(self, user_id):
        """Get the projects of the user, given its ID."""
        return self.json_request("GET", "v3/users/%s/projects" % (user_id,))

    def get_roles(self):
        """Get information about all roles in the system."""
        return self.json_request("GET", "v3/roles")

    def grant_project_role(self, user_id, project_id, role_id):
        """Grant C{role_id} to C{user_id} on C{project_id}"""
        url = "v3/projects/%s/users/%s/roles/%s" % (project_id, user_id, role_id)
        return self.json_request("PUT", url)

    def create_service(self, name, service_type, description):
        """Create a new service."""
        data = {
            "service": {"type": service_type, "name": name, "description": description}
        }
        deferred = self.json_request("POST", "v3/services", body=data)
        return deferred.addCallback(lambda result: result["service"])

    def create_endpoint(self, service_id, region, url, interface):
        """Create a new endpoint for a service."""
        data = {
            "endpoint": {
                "service_id": service_id,
                "region": region,
                "interface": interface,
                "url": url,
            }
        }
        return self.json_request("POST", "v3/endpoints", body=data)
