# Copyright 2005-2015 Canonical Limited.  All rights reserved.

from canonical.openstack.api import JSONClient


class NeutronClient(JSONClient):
    """A client for the Neutron API.

    The methods in here reflect the ones in python-neutronclient, which
    can provide more information on what should be passed to the method.
    The REST API documentation for Neutron also helps, since the methods
    in here are a light wrapper around it.
    """

    def list_agents(self):
        """List agents"""
        return self.json_request("GET", "v2.0/agents")

    def create_network(self, body):
        """Create a network."""
        return self.json_request("POST", "v2.0/networks", body=body)

    def list_networks(self, name=None):
        """List networks (optionally named C{name})."""
        parameters = None
        if name is not None:
            parameters = {"name": name}
        return self.json_request("GET", "v2.0/networks", parameters=parameters)

    def delete_network(self, network_id):
        """Delete a network."""
        return self.json_request("DELETE", "v2.0/networks/%s" % network_id)

    def create_subnet(self, body):
        """Create a subnet."""
        return self.json_request("POST", "v2.0/subnets", body=body)

    def list_subnets(self, name=None):
        """List subnets (optionally named C{name})."""
        parameters = None
        if name is not None:
            parameters = {"name": name}
        return self.json_request("GET", "v2.0/subnets", parameters=parameters)

    def create_router(self, body):
        """Create a router."""
        return self.json_request("POST", "v2.0/routers", body=body)

    def list_routers(self, name=None):
        """List routers (optionally named C{name})."""
        parameters = None
        if name is not None:
            parameters = {"name": name}
        return self.json_request("GET", "v2.0/routers", parameters=parameters)

    def delete_router(self, router_id):
        """Delete a router."""
        return self.json_request("DELETE", "v2.0/routers/%s" % router_id)

    def list_router_l3agents(self, router_id):
        """List the L3 agents for a router."""
        return self.json_request("GET", "v2.0/routers/%s/l3-agents" % router_id)

    def add_router_l3agent(self, agent_id, router_id):
        """Add an L3 agent for a router."""
        body = {"router_id": router_id}
        return self.json_request(
            "POST", "v2.0/agents/%s/l3-routers" % agent_id, body=body
        )

    def remove_router_l3agent(self, agent_id, router_id):
        """Remove an L3 agent for a router."""
        return self.json_request(
            "DELETE", "v2.0/agents/%s/l3-routers/%s" % (agent_id, router_id)
        )

    def add_router_interface(self, router_id, body):
        """Add an interface to a router."""
        return self.json_request(
            "PUT", "v2.0/routers/%s/add_router_interface" % router_id, body=body
        )

    def remove_router_interface(self, router_id, body):
        """Remove an interface from a router."""
        return self.json_request(
            "PUT", "v2.0/routers/%s/remove_router_interface" % router_id, body=body
        )

    def set_router_gateway(self, router_id, network_id):
        """Configure a gateway for a router."""
        body = {"router": {"external_gateway_info": network_id}}
        return self.json_request("PUT", "v2.0/routers/%s" % router_id, body=body)

    def list_security_groups(self, name=None):
        """List security groups."""
        parameters = None
        if name is not None:
            parameters = {"name": name}
        return self.json_request("GET", "v2.0/security-groups", parameters=parameters)

    def add_security_group_rule(self, body):
        """Add a security group rule."""
        return self.json_request("POST", "v2.0/security-group-rules", body=body)

    def list_security_group_rules(self, security_group_id=None):
        """List security group rules."""
        parameters = None
        if security_group_id is not None:
            parameters = {"security_group_id": security_group_id}
        return self.json_request(
            "GET", "v2.0/security-group-rules", parameters=parameters
        )

    def delete_security_group_rule(self, security_group_rule_id):
        """Delete a security group rule."""
        return self.json_request(
            "DELETE", "v2.0/security-group-rules/%s" % security_group_rule_id
        )

    def update_quotas(
        self,
        tenant_id,
        floatingip=None,
        port=None,
        security_group=None,
        security_group_rule=None,
    ):
        """Update the quota for a tentant."""
        quotas = {
            "floatingip": floatingip,
            "port": port,
            "security_group": security_group,
            "security_group_rule": security_group_rule,
        }
        quotas = {
            key: value for key, value in list(quotas.items()) if value is not None
        }
        data = {"quota": quotas}
        return self.json_request("PUT", "v2.0/quotas/%s" % tenant_id, body=data)
