# Copyright 2005-2015 Canonical Limited.  All rights reserved.

from canonical.openstack.api import JSONClient


class NovaClient(JSONClient):
    """A client for the Nova API."""

    def create_aggregate(self, aggregate_name, availability_zone_name):
        """Create a host aggregate."""
        data = {
            "aggregate": {
                "name": aggregate_name,
                "availability_zone": availability_zone_name,
            }
        }
        return self.json_request("POST", "os-aggregates", body=data)

    def list_aggregates(self):
        """Get a list of host aggregates."""
        return self.json_request("GET", "os-aggregates")

    def delete_aggregate(self, aggregate_id):
        """Delete named host aggregate(s)."""
        return self.request("DELETE", "os-aggregates/%s" % aggregate_id)

    def add_host_to_aggregate(self, aggregate_id, host_name):
        """Add a host to a host aggregate."""
        data = {"add_host": {"host": host_name}}
        return self.json_request(
            "POST", "os-aggregates/%s/action" % aggregate_id, body=data
        )

    def remove_host_from_aggregate(self, aggregate_id, host_name):
        """Remove a host from a host aggregate."""
        data = {"remove_host": {"host": host_name}}
        return self.json_request(
            "POST", "os-aggregates/%s/action" % aggregate_id, body=data
        )

    def list_flavors(self):
        """Get the list of current flavors."""
        return self.json_request("GET", "flavors")

    def create_flavor(self, name, vcpus, ram, disk, swap):
        """Create a compute flavor."""
        data = {
            "flavor": {
                "name": name,
                "vcpus": vcpus,
                "ram": ram,
                "disk": disk,
                "os-flavor-access:is_public": True,
                "rxtx_factor": 1.0,
                "OS-FLV-EXT-DATA:ephemeral": 0,
                "swap": swap,
            }
        }
        return self.json_request("POST", "flavors", body=data)

    def update_quotas(
        self,
        tenant_id,
        instances=None,
        cores=None,
        ram=None,
        floating_ips=None,
        security_groups=None,
        security_group_rules=None,
    ):
        """Update the quota for a tentant."""
        quotas = {
            "instances": instances,
            "cores": cores,
            "ram": ram,
            "floating_ips": floating_ips,
            "security_groups": security_groups,
            "security_group_rules": security_group_rules,
        }
        quotas = {
            key: value for key, value in list(quotas.items()) if value is not None
        }
        data = {"quota_set": quotas}
        return self.json_request("PUT", "os-quota-sets/%s" % tenant_id, body=data)

    def list_servers_details(self, host=None, status=None, all_tenants=False):
        """List details for servers.

        Optionally, the result can be filtered by server status or host.
        """
        uri = "servers/detail"
        if all_tenants:
            uri += "?all_tenants=1"

        data = {}
        if host:
            data["host"] = host
        if status:
            data["status"] = status
        return self.json_request("GET", uri, body=data)

    def migrate_live_server(
        self, server_id, target=None, block_migration=False, disk_over_commit=False
    ):
        """Live-migrate a server to a new host."""
        data = {
            "os-migrateLive": {
                "host": target,
                "block_migration": block_migration,
                "disk_over_commit": disk_over_commit,
            }
        }
        return self.json_request(
            "POST", "servers/{}/action".format(server_id), body=data
        )

    def enable_service(self, host, service):
        """Enable a service on a host."""
        data = {"host": host, "binary": service}
        return self.json_request("PUT", "os-services/enable", body=data)

    def disable_service(self, host, service):
        """Disable a service on a host."""
        data = {"host": host, "binary": service}
        return self.json_request("PUT", "os-services/disable", body=data)

    def disable_service_log_reason(self, host, service, reason):
        """Log the reason a service on a host has been disabled."""
        data = {"host": host, "binary": service, "disabled_reason": reason}
        return self.json_request("PUT", "os-services/disable-log-reason", body=data)
