# Copyright 2005-2013 Canonical Limited.  All rights reserved.
from uuid import uuid1, uuid4

from twisted.internet.defer import succeed

from canonical.openstack.keystone import KeystoneClient


class FakeResponseKeystoneClient(KeystoneClient):
    """
    A fake Keystone client that accepts custom response for each client method.

    XXX: This fake client is deprecated. Use and enhance
    FakeKeystoneClient instead, since it provides a better API, in that
    it keeps track of the state internally.
    """

    def __init__(
        self,
        get_users_response,
        add_user_reponse,
        get_projects_response,
        get_roles_response,
        grant_project_role_response,
        authenticate_reponse,
        get_user_projects_response,
        get_user_roles_in_project,
    ):
        self.get_users_response = get_users_response
        self.add_user_response = add_user_reponse
        self.get_projects_response = get_projects_response
        self.get_roles_response = get_roles_response
        self.grant_project_role_response = grant_project_role_response
        self.authenticate_response = authenticate_reponse
        self.get_user_projects_response = get_user_projects_response
        self.get_user_roles_in_project_response = get_user_roles_in_project
        self.called = {}
        super(FakeResponseKeystoneClient, self).__init__("http://example.com/")

    def _get_response(self, response_name):
        """
        Return specific response for each C{response_name} passed.
        """
        response = getattr(self, response_name)
        return succeed(response if response is not None else {})

    def get_users(self):
        """
        Emulate the C{get_users} method of L{KeystoneClient}.
        """
        return self._get_response("get_users_response")

    def add_user(self, *args, **kwargs):
        """
        Emulate the C{add_users} method of L{KeystoneClient}.
        """
        return self._get_response("add_user_response")

    def get_projects(self):
        """
        Emulate the C{get_projects} method of L{KeystoneClient}.
        """
        return self._get_response("get_projects_response")

    def get_roles(self):
        """
        Emulate the C{get_roles} method of L{KeystoneClient}.
        """
        return self._get_response("get_roles_response")

    def grant_project_role(self, *args, **kwargs):
        """
        Emulate the C{grant_project_role} method of L{KeystoneClient}.
        """
        self.record_values("grant_project_role", args, kwargs)
        return None

    def authenticate(self, *args, **kwargs):
        """
        Emulate the C{authenticate} method of L{KeystoneClient}.
        """
        return self._get_response("authenticate_response")

    def get_user_projects(self, *args, **kwargs):
        """
        Emulate the C{get_user_projects} method of L{KeystoneClient}.
        """
        return self._get_response("get_user_projects_response")

    def get_user_roles_in_project(self, *args, **kwargs):
        """
        Emulate the C{get_user_roles_in_project} method of L{KeystoneClient}.
        """
        return self._get_response("get_user_roles_in_project_response")

    def record_values(self, method_reponse, args, kwargs):
        """
        Update the C{called} dictionary with the values that were passed to the
        method.
        """
        self.called[method_reponse] = {"args": args, "kwargs": kwargs}

    def json_request(self, *args, **kwargs):
        raise RuntimeError("This shouldn't be called in tests!")


class FakeOpenStackClient(object):
    """Base class for all fake OpenStack clients.

    This provides all the necessary methods they clients need to provide
    in order to be used in tests.
    """

    _reauth_function = None
    auth_token = None

    def set_reauth_function(self, function):
        self._reauth_function = function

    def set_auth_token(self, auth_token):
        self.auth_token = auth_token


class FakeKeystoneClient(FakeOpenStackClient):
    """A fake Keystone client that can be used in tests.

    Instead of talking over the network, it keeps track of the state
    internally.

    @ivar authentications: A list of all authentication attempts made.
    @ivar services: The service catalog available in keystone.
    @ivar endpoint: The endpoint the client was created with.
    """

    def __init__(self):
        super(FakeKeystoneClient, self).__init__()
        self.client_creations = []
        self.authentications = []
        self.tokens = []
        self.services = [
            {
                "type": "network",
                "id": str(uuid1()),
                "endpoints": [{"publicURL": "http://network.example.com/"}],
            },
            {
                "type": "object-store",
                "id": str(uuid1()),
                "endpoints": [{"publicURL": "http://object.example.com/v1"}],
            },
            {
                "type": "identity",
                "id": str(uuid1()),
                "endpoints": [{"publicURL": "http://keystone.example.com/"}],
            },
            {
                "type": "image",
                "id": str(uuid1()),
                "endpoints": [{"publicURL": "http://glance"}],
            },
            {
                "type": "compute",
                "id": str(uuid1()),
                "endpoints": [{"publicURL": "http://nova"}],
            },
        ]

    def authenticate(self, username, password, tenant):
        token = str(uuid1())
        self.authentications.append((username, password, tenant))
        self.tokens.append(token)
        return succeed(
            {"access": {"token": {"id": token}, "serviceCatalog": self.services}}
        )

    def create_service(self, name, service_type, description):
        """Create a service."""
        service = {
            "type": service_type,
            "name": name,
            "description": description,
            "id": str(uuid1()),
            "endpoints": [],
        }
        self.services.append(service)
        return succeed(service)

    def create_endpoint(self, service_id, region, url, interface):
        """Create a new endpoint for a service."""
        for service in self.services:
            if service["id"] == service_id:
                break
        else:
            raise AssertionError(
                "Tried creating an endpoint for a service that doesn't exist."
            )
        for endpoint in service["endpoints"]:
            if endpoint["region"] == region:
                break
        else:
            endpoint = {"region": region}
            service["endpoints"].append(endpoint)
        endpoint[interface + "URL"] = url
        return succeed(None)

    def get_project_named(self, name):
        """Get details about the given project/tenant."""
        return succeed({"name": name, "id": "%s-id" % name})

    def add_project(self, name, description, domain_id="default"):
        """Create a new project."""
        return succeed(None)

    def add_user(self, username, email, password, project_id=None):
        """Create an user."""
        return succeed(None)

    def get_user_named(self, username):
        """Retrieve user information."""
        user_info = {"id": username + "-id"}
        return succeed(user_info)

    def get_role_named(self, role_name):
        """Retrieve role information."""
        role_info = {"id": role_name + "-id"}
        return succeed(role_info)

    def grant_project_role(self, user_id, project_id, role_id):
        """Grant a role to a user."""
        return succeed(None)

    def __call__(self, endpoint=None, auth_token=None):
        self.client_creations.append((endpoint, auth_token))
        self.endpoint = endpoint
        self.auth_token = auth_token
        return self


class FakeNovaClient(FakeOpenStackClient):
    """A fake Nova client that can be used in tests.

    Instead of talking over the network, it keeps track of the state
    internally.

    @ivar host_aggregates: The host aggregates made using the client.
    """

    def __init__(self, *args, **kwargs):
        super(FakeNovaClient, self).__init__()
        self._host_aggregates = {}
        self._flavors = []
        self.quotas = {}
        self.services = {}

    @property
    def host_aggregates(self):
        return list(self._host_aggregates.values())

    def add_host_aggregate(self, aggregate_info):
        self._host_aggregates[aggregate_info["id"]] = aggregate_info

    def list_aggregates(self):
        return succeed({"aggregates": self.host_aggregates})

    def delete_aggregate(self, aggregate_id):
        if self._host_aggregates[aggregate_id]["hosts"]:
            raise AssertionError("Can't delete non-empty host aggregates.")
        del self._host_aggregates[aggregate_id]
        return succeed({})

    def create_aggregate(self, aggregate_name, availability_zone_name):
        aggregate_id = str(uuid1())
        host_aggregate = {
            "name": aggregate_name,
            "id": aggregate_id,
            "availability_zone": availability_zone_name,
            "hosts": [],
        }
        self._host_aggregates[aggregate_id] = host_aggregate
        # The response doesn't contain the "hosts" key.
        aggregate = host_aggregate.copy()
        del aggregate["hosts"]
        return succeed({"aggregate": aggregate})

    def list_flavors(self):
        # The GET /flavors API only return basic information about flavors
        return succeed(
            {
                "flavors": [
                    {
                        "id": flavor["id"],
                        "name": flavor["name"],
                        "links": flavor["links"],
                    }
                    for flavor in self._flavors
                ]
            }
        )

    def create_flavor(self, name, vcpus, ram, disk, swap):
        uuid = str(uuid4())
        flavor = {
            "id": uuid,
            "name": name,
            "vcpus": 1,
            "ram": ram,
            "disk": disk,
            "swap": swap,
            "os-flavor-access:is_public": True,
            "id": "1",
            "links": [
                {"href": "http://nova/flavors/{}".format(uuid), "rel": "self"},
                {"href": "http://nova/flavors/{}".format(uuid), "rel": "bookmark"},
            ],
        }
        self._flavors.append(flavor)
        return succeed({"flavor": flavor})

    def add_host_to_aggregate(self, aggregate_id, host_name):
        self._host_aggregates[aggregate_id]["hosts"].append(host_name)
        return succeed({})

    def remove_host_from_aggregate(self, aggregate_id, host_name):
        self._host_aggregates[aggregate_id]["hosts"].remove(host_name)
        return succeed({})

    def update_quotas(
        self,
        tenant_id,
        instances=None,
        cores=None,
        ram=None,
        floating_ips=None,
        security_groups=None,
        security_group_rules=None,
    ):
        """Update the quota for a tentant."""
        new_quotas = {
            "floating_ips": floating_ips,
            "instances": instances,
            "cores": cores,
            "ram": ram,
            "security_groups": security_groups,
            "security_group_rules": security_group_rules,
        }
        new_quotas = {
            key: value for key, value in list(new_quotas.items()) if value is not None
        }
        self.quotas.setdefault(tenant_id, {}).update(new_quotas)
        return succeed(None)

    def enable_service(self, host, service):
        self.services[host, service] = "enabled"
        return {"service": {"host": host, "binary": service, "status": "enabled"}}

    def disable_service(self, host, service):
        self.services[host, service] = ("disabled", None)
        return {"service": {"host": host, "binary": service, "status": "disabled"}}

    def disable_service_log_reason(self, host, service, reason):
        # The service must be disabled to set the reason
        assert self.services[host, service][0] == "disabled"
        self.services[host, service] = ("disabled", reason)
        return {
            "service": {
                "host": host,
                "binary": service,
                "disabled_reason": reason,
                "status": "disabled",
            }
        }

    def __call__(self, endpoint, auth_token):
        self.endpoint = endpoint
        self.auth_token = auth_token
        return self


class FakeGlanceClient(FakeOpenStackClient):
    """A fake Glance client that can be used in tests.

    Instead of talking over the network, it keeps track of the state
    internally.

    @ivar uploads: The uploads made using the client.
    @ivar images: The details of the images in Glance.
    """

    def __init__(self, *args):
        super(FakeGlanceClient, self).__init__()
        self.uploads = []
        self.deletes = []
        self.images = []

    def delete_image(self, image_id):
        """Delete an image."""
        self.deletes.append(image_id)
        return succeed(None)

    def images_detail(self):
        """Return detailed information about all images."""
        return succeed({"images": self.images})

    def __call__(self, endpoint, auth_token):
        self.endpoint = endpoint
        self.auth_token = auth_token
        return self


class FakeNeutronClient(FakeOpenStackClient):
    """A Neutron client that can be used in tests.

    Instead of talking over the network, it keeps track of the created
    networks and routers internally.

    @ivar agents: A list of agents.
    @ivar networks: A list of all networks that were created.
    @ivar subnets: A list of all subnets that were created.
    @ivar routers: A list of all routers that were created.
    @ivar security_groups: A list of all security groups that were created.
    @ivar security_group_rules: A list of all security group rules
             that were created.
    @ivar quotas: A dictionary of quotas for tenants.
    """

    def __init__(self):
        super(FakeNeutronClient, self).__init__()
        self.agents = []
        self.networks = []
        self.subnets = []
        self.routers = []
        self.security_groups = []
        self.security_group_rules = []
        # Used to track adds and deletes across multiple neutron clients
        self.added_security_group_rules = []
        self.deleted_security_group_rules = []

        self.quotas = {}

    def list_agents(self):
        """List agents."""
        return succeed({"agents": self.agents})

    def create_network(self, options):
        """Create a network."""
        created_network = {"network": options["network"].copy()}
        created_network["network"]["id"] = str(uuid1())
        self.networks.append(created_network["network"])
        return succeed(created_network)

    def list_networks(self, name=None):
        """List networks, optionally matching C{name}."""
        networks = [
            network
            for network in self.networks
            if name is None or network["name"] == name
        ]
        return succeed({"networks": networks})

    def delete_network(self, network_id):
        """Delete a network."""
        [network] = [
            network for network in self.networks if network["id"] == network_id
        ]
        self.networks.remove(network)
        return succeed(None)

    def create_subnet(self, options):
        """Create a subnet."""
        created_network = {"subnet": options["subnet"].copy()}
        created_network["subnet"]["id"] = str(uuid1())
        self.subnets.append(created_network["subnet"])
        return succeed(created_network)

    def list_subnets(self, name=None):
        """List subnets, optionally matching C{name}."""
        subnets = [
            subnet for subnet in self.subnets if name is None or subnet["name"] == name
        ]
        return succeed({"subnets": subnets})

    def create_router(self, options):
        """Create a router."""
        created_router = {"router": options["router"].copy()}
        created_router["router"]["id"] = str(uuid1())
        self.routers.append(created_router["router"])
        return succeed(created_router)

    def list_routers(self, name=None):
        """List routers, optionally matching C{name}."""
        routers = [
            router for router in self.routers if name is None or router["name"] == name
        ]
        return succeed({"routers": routers})

    def delete_router(self, router_id):
        """Delete a router."""
        [router] = [router for router in self.routers if router["id"] == router_id]
        assert (
            "subnet_id" not in router
        ), "Cannot remove a router with an active interface"
        self.routers.remove(router)
        return succeed(None)

    def add_router_interface(self, router_id, options):
        """Add an interface to a router."""
        self._update_router(router_id, options)
        return succeed(None)

    def remove_router_interface(self, router_id, options):
        """Remove an interface from a router."""
        [router] = [router for router in self.routers if router["id"] == router_id]
        del router["subnet_id"]
        return succeed(None)

    def set_router_gateway(self, router_id, options):
        """Set a default gateway for a router."""
        self._update_router(router_id, options)
        return succeed(None)

    def list_security_groups(self, name=None):
        """List security groups."""
        groups = [
            group
            for group in self.security_groups
            if name is None or group["name"] == name
        ]
        return succeed({"security_groups": groups})

    def add_security_group_rule(self, body):
        """Add a security group rule."""
        rule = body["security_group_rule"]
        self.added_security_group_rules.append(sorted(rule.items()))
        rule["id"] = rule.get("id", str(uuid1()))
        self.security_group_rules.append(rule)
        return succeed(None)

    def list_security_group_rules(self, security_group_id=None):
        """List security group rules."""
        rules = [
            rule
            for rule in self.security_group_rules
            if security_group_id is None
            or rule["security_group_id"] == security_group_id
        ]
        return succeed({"security_group_rules": rules})

    def delete_security_group_rule(self, security_group_rule_id):
        """Delete a security group rule."""
        [rule] = [
            rule
            for rule in self.security_group_rules
            if rule["id"] == security_group_rule_id
        ]
        self.deleted_security_group_rules.append(sorted(rule.items()))
        self.security_group_rules.remove(rule)
        return succeed(None)

    def update_quotas(
        self,
        tenant_id,
        floatingip=None,
        port=None,
        security_group=None,
        security_group_rule=None,
    ):
        """Update the quota for a tentant."""
        new_quotas = {
            "floatingip": floatingip,
            "port": port,
            "security_group": security_group,
            "security_group_rule": security_group_rule,
        }
        new_quotas = {
            key: value for key, value in list(new_quotas.items()) if value is not None
        }
        self.quotas.setdefault(tenant_id, {}).update(new_quotas)
        return succeed(None)

    def _update_router(self, router_id, options):
        """Update a router with the given options.

        This an internal function and isn't part of the real
        NeutronClient API.
        """
        [router] = [router for router in self.routers if router["id"] == router_id]
        router.update(options)
        return succeed(None)

    def __call__(self, endpoint, auth_token):
        self.endpoint = endpoint
        self.auth_token = auth_token
        return self


class FakeSwiftClient(FakeOpenStackClient):
    """A fake Swift client that can be used in tests.

    Instead of talking over the network, it keeps track of the state
    internally.

    @ivar containers: The containers that were created, with the objects
        inside them as a dict (path -> data).
    """

    def __init__(self):
        super(FakeSwiftClient, self).__init__()
        self.containers = {}

    def create_container(self, name, public=False):
        if name not in self.containers:
            self.containers[name] = {"objects": {}}
        self.containers[name]["public"] = public
        return succeed(None)

    def create_object(self, container, name, data):
        self.containers[container]["objects"][name] = data
        return succeed(None)

    def get_object(self, container, name):
        return succeed(self.containers[container]["objects"][name])

    def __call__(self, endpoint, auth_token):
        self.endpoint = endpoint
        self.auth_token = auth_token
        return self
