# Copyright 2005-2016 Canonical Limited.  All rights reserved.

import os
from subprocess import check_call


GLOBAL_OPTIONS = ["-O", "Portrait", "-s", "A4", "-q", "--enable-local-file-access"]


PAGE_OPTIONS = ["--disable-javascript"]


PATH_TO_ICC = "/usr/share/color/icc/compatibleWithAdobeRGB1998.icc"


class HTMLToPDFError(Exception):
    """Exception to raise when something is wrong with HTMLToPDFConversion."""


def html_to_pdf(html_path, pdf_path, images, stdout=None):
    """
    For an HTML file at a given path, generate a PDF representation at
    a second path.

    @param images: list of images to allow access to
    """
    _validate_output_path(pdf_path)
    wkhtmltopdf = ["wkhtmltopdf"] + GLOBAL_OPTIONS + [html_path]
    allow_options = []
    for image in images:
        allow_options += ["--allow", image]
    wkhtmltopdf += PAGE_OPTIONS + allow_options + [pdf_path]
    command = ["xvfb-run", "-a"] + wkhtmltopdf
    check_call(command, stdout=stdout)
    _validate_output(pdf_path)


def pdf_to_pdfa(pdf_path, pdfa_path, stdout=None, stderr=None):
    """
    Given a path to a valid PDF file, create the PDF/A version of the
    same file at a provided path.

    PDF/A is a restriction of the PDF 1.4 format to ensure that the
    file is archivable.  In short this means ensuring that all images
    and fonts are embedded, and setting metadata tags to indicate it's
    compliance level.

    PDF/A is the baseline for most international standards, and
    regulations concerning the archival of PDF files.
    """
    command = [
        "gs",
        "-q",
        "-dPDFA=2",
        "-dBATCH",
        "-dNOPAUSE",
        "-dNOOUTERSAVE",
        "-dSAFER",
        "-sDEVICE=pdfwrite",
        "-sOutputICCProfile={}".format(PATH_TO_ICC),
        "-dPDFACompatibilityPolicy=1",
        "-sColorConversionStrategy=RGB",
    ]
    command.append("-sOutputFile=" + pdfa_path)
    command.append(pdf_path)
    check_call(command, stdout=stdout, stderr=stderr)


def _validate_output_path(output_path):
    """Raise an exception if the provided output path is not writeable.

    This is necessary because wkhtmltopdf (as packaged in Ubuntu),
    will fail silently if it can't write to the target file.
    """
    output_dir = os.path.dirname(os.path.abspath(output_path))

    if not os.path.exists(output_dir):
        raise HTMLToPDFError("Output path {} doesn't exist.".format(output_dir))
    if not os.path.isdir(output_dir):
        raise HTMLToPDFError("Output path '{}', is not a directory.".format(output_dir))
    if not os.access(output_dir, os.W_OK):
        raise HTMLToPDFError("Output path '{}', is not writable.".format(output_dir))


def _validate_output(output_path):
    """Raise an exception if no output exists.

    As per _validate_output_path, this exists because wkhtmltopdf
    isn't reliable when it comes to it's return code when it fails to
    output a file.  The upstream version has improved, so at some
    point in the future we should be able to remove these checks and
    rely on the CalledProcessError instead.
    """
    if (not os.path.exists(output_path)) or os.path.isdir(output_path):
        raise HTMLToPDFError("Failed to create PDF file {}.".format(output_path))
