# Copyright 2005-2022 Canonical Limited. All rights reserved.

from landscape.lib.fs import create_file

from canonical.lib.archive import ControlFile


class MissingStanzaKeyField(Exception):
    """Raised when stanza info lack the key field."""


class DuplicatedStanza(Exception):
    """Raised when trying to add a stanza with a name that already exists."""


class RepreproConfigFile(ControlFile):
    """
    This class handles creation of configuration files for reprepro.

    A config file contains a series of stanzas separated by an empty line.

    A stanza is identified by the value of the property referenced as
    C{key_field}. This value must be unique in the file.

    Stanzas may be added at any time, the current stanzas list is written each
    time C{write()} is called.

    @param filename: the name of the config file to write.
    @param key_field: the name of the field used as stanza identifier.
    """

    def __init__(self, filename, key_field="Name"):
        super(RepreproConfigFile, self).__init__()
        self._filename = filename
        self._key_field = key_field
        self._key_names = []

    def add_stanza(self, info):
        """
        Add a configuration stanza with specified info.
        Info must contain the L{key_field}.

        If a stanza with the same name is already present, an exception is
        raised.

        @param info: a dict with stanza parameters and values
        """
        name = info.get(self._key_field)

        if not name:
            raise MissingStanzaKeyField("Missing %s field." % self._key_field)

        if name in self._key_names:
            raise DuplicatedStanza(
                "Stanza with %s '%s' already exists." % (self._key_field, name)
            )

        self._key_names.append(name)

        for key, value in list(info.items()):
            self.add_field(key, value)
        self.new_section()

    def get_stanzas(self):
        """Return the configured stanzas."""
        return [dict(section) for section in self._sections if section]

    def write(self):
        """Write configuration to file."""
        create_file(self._filename, self.content().encode("utf-8"))


class RepreproUploadersFile(object):
    """
    This class handles creation of an uploaders configuration file for
    reprepro.

    An uploaders file contains a series of allow statements.

    Allows may be added at any time, the current list is written each time
    C{write()} is called.

    @param filename: the name of the config file to write.
    """

    def __init__(self, filename):
        self._filename = filename
        self._stanzas = []

    def allow(self, key=None, unsigned=False):
        """
        Allow the specified key, or any key or unsigned.

        @param key: a GPG key id or fingerprint, if None, allows any key
        @param unsigned: if True also allows unsigned content
        """

        if unsigned:
            who = "anybody"
        else:
            who = "key %s" % key if key is not None else "any key"

        self._stanzas.append("allow * by %s" % who)

    def get_stanzas(self):
        """Return the configured stanzas."""
        return list(self._stanzas)

    def write(self):
        """Write configuration to file."""
        create_file(self._filename, ("\n".join(self._stanzas) + "\n").encode("utf-8"))


class RepreproFilterListFile(object):
    """
    Handle a configuration file for reprepro's FilterList option.

    A filter file contains a list of packages and actions for them.

    @param filename: the name of the config file to write.
    """

    def __init__(self, filename):
        self._filename = filename
        self._rules = []

    def add_rule(self, package, action):
        """
        Add a rule for a package.

        @param package: The package to add a rule for.
        @param action: The action to perform on the package.
        """
        self._rules.append((package, action))

    def get_rules(self):
        """
        Return the list of package rules.
        """
        return list(self._rules)

    def write(self):
        """Write configuration to file."""
        content = "\n".join("{} {}".format(*rule) for rule in self._rules)
        content += "\n"
        create_file(self._filename, content.encode())
