# Copyright 2005-2022 Canonical Limited. All rights reserved.

import errno
import os
import re
import shutil

from twisted.python import log
from twisted.internet.defer import DeferredFilesystemLock, succeed

from landscape.lib.twisted_util import spawn_process
from canonical.reprepro.config import (
    RepreproConfigFile,
    RepreproUploadersFile,
    RepreproFilterListFile,
)
from canonical.gnupg.gnupg import GnuPG, GnuPGError


class RepreproError(Exception):
    """Raised when reprepro fails."""

    def __init__(self, out, err, code=None, signal=None):
        self.out = out
        self.err = err
        self.code = code
        self.signal = signal

        if code is not None:
            reason = f"exit code {code}"
        if signal is not None:
            reason = f"signal {signal}"

        super(RepreproError, self).__init__(
            f"reprepro ended with {reason} (out='{out}', err='{err}')"
        )


class Reprepro(object):
    """
    Handle execution of reprepro process for a specific installation

    @param binary_path: the path to the reprepro binary.
    @param basedir: the base directory for the installation
    """

    _GPG_OPTIONS = {"personal-digest-preferences": "SHA512 SHA384 SHA256 SHA224"}

    _progress_re = re.compile(r"Got ([0-9]+)%")
    _quoted_names_re = re.compile(r"'(.*?)'")
    _list_re = re.compile(
        r"(?P<udeb>.+\|)?(?P<distro>.+)\|(?P<component>.+)\|(?P<arch>.+): "
        r"(?P<package>.+) (?P<version>.+)"
    )
    _dump_re = re.compile(r"[^ ']+")

    def __init__(self, basedir, binary_path=None):
        self._gpg_home = None
        self._gpg_keys = set()
        self.binary_path = binary_path or "/usr/bin/reprepro"
        self._basedir = basedir
        self._tempdir = os.path.join(self._basedir, "temp")
        self._incomingdir = os.path.join(self._basedir, "incoming")
        self._confdir = os.path.join(self._basedir, "conf")
        self._lockdir = os.path.join(self._basedir, "locks")
        self._lockfile = DeferredFilesystemLock(os.path.join(self._lockdir, "lock"))
        self.reset()

    def reset(self):
        """Reset the configuration."""
        self._distributions = RepreproConfigFile(
            os.path.join(self._confdir, "distributions"), key_field="Codename"
        )
        self._pulls = RepreproConfigFile(os.path.join(self._confdir, "pulls"))
        self._updates = RepreproConfigFile(os.path.join(self._confdir, "updates"))
        self._incomings = RepreproConfigFile(os.path.join(self._confdir, "incoming"))
        self._uploaders = {}
        self._filterlists = {}
        self._distribution = None

    def add_distribution(
        self,
        name,
        components,
        archs,
        mode=None,
        sign=None,
        udeb_components=None,
        suite=None,
    ):
        """
        Add a distribution to reprepro installation.

        @param name: The name of the distribution to add.
        @param components: The components of the distribution.
        @param archs: The architectures of the distribution.
        @param sign: The GPG key used to sign the distribution.
        @param udeb_components: The debian-installer components of the
            distribution, if any.
        @param suite: The suite (stable, testing). Defaults to the codename.
        """
        stanza = {
            "Codename": name,
            # choose-mirror on ISO fails on missing suite
            "Suite": suite if suite is not None else name,
            "Components": " ".join(components),
            "Architectures": " ".join(archs),
        }
        if udeb_components:
            # Must come after Components in the config stanza, or reprepro will
            # complain.
            stanza["UDebComponents"] = " ".join(udeb_components)

        if mode == "mirror":
            stanza["Update"] = f"- update-{name}"
        if mode == "pull":
            stanza["Pull"] = f"pull-{name}"
        if mode == "upload":
            uploader_name = f"uploader-{name}"
            stanza["Uploaders"] = uploader_name
            self._uploaders[uploader_name] = RepreproUploadersFile(
                os.path.join(self._confdir, uploader_name)
            )

        if sign is not None:
            stanza["SignWith"] = sign

        self._distributions.add_stanza(stanza)

    def add_pull(self, distribution, parent, filter_action=None):
        """
        Add a pull rule to the reprepro installation.

        @param distribution: The distribution the rule refers to.
        @param parent: The distribution to pull from.
        @param filter_action: The default filter action for packages not listed
            in a filter rule. If None, no filtering will be enabled.
        """
        stanza = {"Name": f"pull-{distribution}", "From": parent}
        if filter_action is not None:
            filter_name = f"filterlist-{distribution}"
            stanza["FilterList"] = f"{filter_action} {filter_name}"
            self._filterlists[filter_name] = RepreproFilterListFile(
                os.path.join(self._confdir, filter_name)
            )

        self._pulls.add_stanza(stanza)

    def add_update(self, distribution, uri, suite, sign, flat=None, config=None):
        """
        Add an update rule to the reprepro installation.

        @param distribution: The distribution the rule refers to.
        @param uri: The uri of the remote repository.
        @param sign: The GPG key to use to verify the remote repository.
        @param suite: The name of the remote suite to pull.
        @param flat: If provided, the target component for packages from a flat
            repostiory.
        @param config: If provided, it should be  a dict with key/value pairs
            for extra APT config options.
        """
        stanza = {
            "Name": f"update-{distribution}",
            "Method": uri,
            "VerifyRelease": sign,
        }
        if suite:
            stanza["Suite"] = suite
        if flat:
            stanza["Flat"] = flat
        if config:
            stanza["Config"] = [f"{k}={v}" for k, v in config.items()]
        self._updates.add_stanza(stanza)

    def add_incoming(self, distribution):
        """
        Add an incoming config file to reprepro installation.

        @param distribution: The distribution name to accept uploads for.
        """
        stanza = {
            "Name": f"incoming-{distribution}",
            "Allow": distribution,
            "Permit": "unused_files",
            "Cleanup": "unused_files",
            "TempDir": self._tempdir,
            "IncomingDir": self._incomingdir,
        }
        self._incomings.add_stanza(stanza)

    def add_uploader(self, distribution, key=None, unsigned=False):
        """
        Add an uploader rule for a distribution

        @param distribution: The distribution to add the permission to.
        @param key: GPG key id or fingerprint the incoming .changes
            should be signed with, if None any key is valid.
        @param unsigned: if True unsigned .changes are accepted.
        """
        uploader_file = self._uploaders[f"uploader-{distribution}"]
        uploader_file.allow(key=key, unsigned=unsigned)

    def add_filterlist_rule(self, distribution, package, action):
        """
        Add a filter rule on a package for a distribution.

        @param distribution: The distribution to add the filter rule to.
        @param package: The name of package in the rule.
        @param action: The action to perform on the package.
        """
        filterlist = self._filterlists[f"filterlist-{distribution}"]
        filterlist.add_rule(package, action)

    def set_gpg_home(self, gpg_home):
        """Set the reprepro command to run."""
        self._gpg_home = gpg_home

    def add_gpg_key(self, gpg_key):
        """Add a GPG key to the keyring used for signatures."""
        self._gpg_keys.add(gpg_key)

    def get_gpg_keys(self):
        """Get the configured GPG keys."""
        return list(self._gpg_keys)

    def get_basedir(self):
        """Return the configured base directory."""
        return self._basedir

    def get_dbdir(self):
        """Return reprepro db directory."""
        return os.path.join(self._basedir, "db")

    def get_gpg_home(self):
        """Return the configured GnuPG home directory."""
        return self._gpg_home

    def get_distsdir(self, dist=None):
        """Return configured dists directory.

        @param dist: If not C{None} it returns the path to that specific dist.
        """
        result = os.path.join(self._basedir, "dists")
        if dist is not None:
            result = os.path.join(result, dist)
        return result

    def get_distributions(self):
        """Return the configured distributions."""
        return self._distributions.get_stanzas()

    def get_pulls(self):
        """Return the configured pulls."""
        return self._pulls.get_stanzas()

    def get_updates(self):
        """Return the configured updates."""
        return self._updates.get_stanzas()

    def get_incomings(self):
        """Return the configured incomings."""
        return self._incomings.get_stanzas()

    def get_uploaders(self):
        """Return the configured uploaders."""
        return dict(
            (key, value.get_stanzas()) for key, value in self._uploaders.items()
        )

    def get_filterlist_rules(self):
        """Return the configured filterlist rules."""
        return dict(
            (key, value.get_rules()) for key, value in self._filterlists.items()
        )

    def _run(self, args=(), progress_callback=None):
        """
        Run a reprepro command with the specified argument list.

        @param progress_callback: optional callback for operation progress,
             called with an integer indicating percentage. Not all operations
             have progress report.
        @return: a deferred returning stdout, stderr and exit value
        """
        default_args = ["--ignore=wrongarchitecture", "--ignore=undefinedtarget"]
        if self._gpg_home is not None:
            default_args += ["--gnupghome", self._gpg_home]
        args = [str(arg) for arg in default_args + args]
        if progress_callback:

            def line_received(line):
                line = line.decode()
                match = self._progress_re.match(line)
                if match:
                    progress_callback(int(match.groups()[0]))

        else:
            line_received = None

        def with_lock(result):
            deferred = succeed(None)

            def import_keys(_):
                def log_error(path, error):
                    log.msg(f"Ignoring failure while removing {path}: {error}")

                if os.path.exists(self._gpg_home):
                    for filename in os.listdir(self._gpg_home):
                        path = os.path.join(self._gpg_home, filename)
                        if os.path.isdir(path):
                            try:
                                shutil.rmtree(path)
                            except Exception as exc:
                                log_error(path, exc)
                        else:
                            try:
                                os.remove(path)
                            except OSError as error:
                                log_error(path, error)
                gnupg = GnuPG(self._gpg_home, options=self._GPG_OPTIONS)
                return gnupg.import_keys(self._gpg_keys)

            def spawn_reprepro(_):
                self._distributions.write()
                self._pulls.write()
                self._updates.write()
                self._incomings.write()
                for uploader in list(self._uploaders.values()):
                    uploader.write()
                for filterlist in list(self._filterlists.values()):
                    filterlist.write()
                # The process must be created using a PTY to get line-buffered
                # output, otherwise the output gets processed at the end
                # because pipe buffers are much bigger.
                return spawn_process(
                    self.binary_path,
                    args=args,
                    path=self._basedir,
                    usePTY=True,
                    line_received=line_received,
                )

            def unlock(result):
                self._lockfile.unlock()
                return result

            if self._gpg_keys:
                deferred.addCallback(import_keys)

            deferred.addCallback(spawn_reprepro)
            deferred.addBoth(unlock)
            deferred.addErrback(self._handle_error)
            deferred.addCallback(self._check_exit_code)
            return deferred

        for directory in (
            self._basedir,
            self._confdir,
            self._lockdir,
            self._tempdir,
            self._incomingdir,
        ):
            if os.path.exists(directory):
                if not os.path.isdir(directory):
                    raise RuntimeError(
                        f"Path '{directory}' exists but is not a directory"
                    )
            else:
                try:
                    os.mkdir(directory)
                    os.chmod(directory, 0o755)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise

        lock = self._lockfile.deferUntilLocked()
        return lock.addCallback(with_lock)

    def _check_exit_code(self, exit_info):
        """Raise a RepreproError in case of non-zero exit code."""
        (out, err, code) = exit_info
        if code != 0:
            error = RepreproError(out, err, code=code)
            log.err(error)
            raise error
        return out, err

    def _handle_error(self, failure):
        """Raise a RuntimeError in case of a non-zero exit code."""
        if failure.check(GnuPGError):
            error = failure.value
        else:
            out, err, signal = failure.value.args
            error = RepreproError(out, err, signal=signal)
        log.err(error)
        raise error

    def set_distribution(self, distribution):
        """
        Set the distribution to work on, for the commands that support it.
        """
        self._distribution = distribution

    def processincoming(self, changes):
        """
        Run the 'processincoming' reprepro command with the given .changes

        @param changes: Path to the .changes file to import.
        @return: A deferred firing to a C{dict} reporting any error that
            occurred, if there was no error the C{dict} will be empty.
        """
        deferred = self._run(
            ["-VV", "processincoming", f"incoming-{self._distribution}", changes]
        )
        deferred.addCallback(lambda _: {})
        deferred.addErrback(self._parse_processincoming)
        return deferred

    def _parse_processincoming(self, failure):
        """
        Parse lines from 'processincoming' output and return possible errors in
        package import.

        @return: a dict with errors, or None if no error happened.
        """
        failure.trap(RepreproError)

        lines = failure.value.out.decode().splitlines()
        if not lines:
            return failure

        result = {
            "missing_files": [],
            "invalid_files": [],
            "unknown_arch_files": [],
            "unverified_files": [],
            "udeb_files": [],
            "errors": [],
        }

        for line in lines:
            names = self._quoted_names_re.findall(line)
            if "not found in the incoming dir" in line:
                result["missing_files"].append(names[-1])
            if "does not match expectations" in line:
                result["invalid_files"].append(names[0])
            if "public key missing" in line:
                # list of tuples with filenames and id of the unknown GPG key
                # used to sign it
                result["unverified_files"].append((names[1], names[0]))
            if "No distribution accepting" in line:
                result["errors"].append("no-sign")
            if "No distribution found" in line:
                result["errors"].append("no-distro")
            if "valid architecture" in line:
                result["unknown_arch_files"].append(names[1])
            if "Cannot put file" in line:
                result["udeb_files"].append(names[0])
        return result

    def export(self):
        """
        Run the 'export' reprepro command.

        If no distribution has been set, export all changed distributions.
        """
        command = ["export"]
        if self._distribution is not None:
            command.append(self._distribution)

        if self.get_distributions():
            deferred = self._run(command)
            return deferred.addCallback(self._after_export)
        else:
            return succeed(True)

    def _after_export(self, xxx_todo_changeme1):
        (out, err) = xxx_todo_changeme1
        return True

    def update(self, progress_callback=None):
        """
        Run the 'update' reprepro command for the selected distribution.

        @param progress_callback: optional callback for operation progress,
             called with an integer indicating percentage. Not all operations
             have progress report.
        """
        deferred = self._run(
            [
                "--noskipold",
                "--export=lookedat",
                "--show-percent",
                "update",
                self._distribution,
            ],
            progress_callback=progress_callback,
        )

        return deferred.addCallback(lambda _: True)

    def pull(self):
        """
        Run the 'pull' reprepro command for the selected distribution.
        """
        deferred = self._run(
            ["--noskipold", "--export=lookedat", "-VV", "pull", self._distribution]
        )
        return deferred.addCallback(lambda _: True)

    def dumppull(self):
        """
        Run the 'dumppull' reprepro command for the selected distribution.
        """
        deferred = self._run(["dumppull", self._distribution])
        return deferred.addCallback(self._parse_dump)

    def dumpupdate(self):
        """
        Run the 'dumpupdate' reprepro command for the selected distribution.
        """
        deferred = self._run(["dumpupdate", self._distribution])
        return deferred.addCallback(self._parse_dump)

    def _parse_dump(self, pipes):
        """
        Parse lines from 'dumppull' or 'dumpupdate' commands, returning a dict
        as follows:

          {'<component>/<arch>':
             {'update': [(package, old_version, new_version), ...],
              'add': [(package, version), ...],
              'keep': [(package, version), ...],
              'delete': [(package, version) ...]},
           '<component>/<arch>/udeb':
             {...},
           ....}
        """
        (out, err) = pipes
        result = {}
        for line in out.splitlines():
            line = line.decode()

            if line == "":
                continue

            if line.startswith("Updates needed for"):
                [names] = self._quoted_names_re.findall(line)
                tokens = names.split("|")
                section = tokens[-2:]
                # If there are 4 elements, the first one is 'u' for udeb
                # packages.
                if len(tokens) == 4:
                    section.append("udeb")
            else:
                info = self._dump_re.findall(line)
                action = info.pop(0)

                if action == "add":
                    del info[3]
                    del info[1]
                elif action == "update":
                    del info[3]
                elif action == "keep" and info[2] == "unavailable":
                    # In this case the version is the "best candidate".  If the
                    # value is not "unavailable" the package hasn't changed, so
                    # it won't be included in the diff; otherwise it has been
                    # kept because of a filter or because removed from the
                    # source.
                    del info[2]
                elif action != "delete":
                    continue

                component_arch = result.setdefault("/".join(section), {})
                infos = component_arch.setdefault(action, [])
                infos.append(tuple(info))
        return result

    def list(self):
        """
        Run the 'list' reprepro command for the selected distribution.
        """
        deferred = self._run(["list", self._distribution])
        return deferred.addCallback(self._parse_list)

    def _parse_list(self, pipes):
        """
        Parse lines from 'list', returning a dict as follows:

          {'<component>/<arch>': [(package, version), ...],
           '<component>/<arch>/udeb': [(package, version), ...],
          ...}
        """
        (out, err) = pipes
        result = {}
        for line in out.splitlines():
            line = line.decode()

            if line == "":
                continue

            match = self._list_re.match(line)
            if match:
                details = match.groupdict()
                if details["udeb"]:
                    section = "%(component)s/%(arch)s/udeb" % details
                else:
                    section = "%(component)s/%(arch)s" % details
                component_arch = result.setdefault(section, [])
                component_arch.append((details["package"], details["version"]))
        return result

    def clearvanished(self):
        """
        Run the 'clearvanished' reprepro command.
        """
        if not self.get_distributions():
            # Clean everything if there is no distribution left.
            if os.path.exists(self._basedir):
                shutil.rmtree(self._basedir)
            return succeed(True)

        deferred = self._run(["--delete", "clearvanished"])
        deferred.addCallback(self._after_clearvanished)
        return deferred

    def _after_clearvanished(self, pipes):
        """
        Remove dists file for removed distributions, cleans everything if there
        is no distribution left.
        """
        (out, err) = pipes
        out = out.decode()
        for line in out.splitlines():
            if line.startswith("Deleting vanished identifier"):
                [names] = self._quoted_names_re.findall(line)
                distribution = names.split("|")[0]
                shutil.rmtree(
                    os.path.join(self.get_distsdir(distribution)), ignore_errors=True
                )
        return True

    def includedeb(self, deb):
        """
        Run the 'includedeb' reprepro command to import the given deb.
        """
        return self._run(["includedeb", self._distribution, deb])

    def remove(self, packages):
        """
        Run the 'remove' reprepro command to remove the given packages.
        """
        command = ["remove", self._distribution]
        command.extend(packages)
        deferred = self._run(command)
        return deferred.addCallback(lambda _: True)
