# Copyright 2005-2022 Canonical Limited. All rights reserved.

import sys

from zope.interface import implementer
from zope.component import queryMultiAdapter
from zope.security.proxy import ProxyFactory, removeSecurityProxy, isinstance
from zope.security.management import checkPermission
from zope.security.interfaces import Unauthorized
from zope.publisher.interfaces import (
    NotFound,
    BadRequest,
    Redirect as ZopeRedirect,
    Retry,
)
from zope.publisher.browser import BrowserPage
from zope.publisher.publish import mapply
from zope.publisher.defaultview import getDefaultViewName

from canonical.routes.interfaces import IRoutesPublisher
from canonical.routes.routes import Routes


__all__ = [
    "RoutesPublisher",
    "route",
    "NonePage",
    "get_route_objects",
    "get_route_object",
    "get_route_object_by_interface",
    # From Zope:
    "NotFound",
    "BadRequest",
    "Redirect",
    "Retry",
]


EXPR_ATTR = "__route_expressions__"
OBJECTS_ANNOTATION_KEY = "canonical.routes.objects"


class Redirect(ZopeRedirect):
    """A wrapper around Zope's Redirect.

    The key feature here is that this one protects against
    HTTP Response Splitting (CWE-113) (lp:1695010)
    and open redirects in urlparse (python 35748) (lp:1929620).
    """

    def __init__(self, location, status=None):

        self.status = status

        # Accoring to https://tools.ietf.org/html/rfc2616#section-4.2,
        # "\r\n" isn't allowed in HTTP headers.  Furthermore,
        # section-5.1 in the same RFC is even more restrictive.  It
        # indicates that the request line from the client can't have
        # either \r or \n in it.  Failure to comply on either count
        # opens the door to vulnerabilities like "HTTP Response
        # Splitting" (CWE-113).  Since the stdlib, twisted, etc. aren't
        # helping us out, we take matters into our own hands here.
        try:
            location = location.decode()
        except (AttributeError, UnicodeDecodeError):
            pass
        location = location.replace("\r", "")
        location = location.replace("\n", "")

        # Unescaped backslashes aren't standard, and browsers are
        # reinterpreting them in our back.
        # See https://bugs.python.org/issue35748
        location = location.replace("\\", "/")
        super(Redirect, self).__init__(location)


def route(route_expr, **kwargs):
    frame = sys._getframe(1)
    locals = frame.f_locals
    route_expressions = locals.get(EXPR_ATTR)
    if not route_expressions:
        locals[EXPR_ATTR] = route_expressions = []

    def decorator(handler):
        kwargs["__handler"] = handler
        route_expressions.append((route_expr, kwargs))
        return handler

    return decorator


def get_route_objects(request, default=None):
    return request.annotations.get(OBJECTS_ANNOTATION_KEY, default)


def get_route_object(context_type, request, default=None):
    for object in get_route_objects(request, ()):
        # We may want to extend this to understand interfaces in the future.
        if isinstance(object, context_type):
            return object
    return default


def get_route_object_by_interface(context_interface, request):
    """Retrieve an element in the navigation tree by its interface."""
    for context in get_route_objects(request, ()):
        if context_interface.providedBy(context):
            return context


@implementer(IRoutesPublisher)
class RoutesPublisher(object):

    default_view = None

    trusted_redirects = []

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.__add_route_object(context)

    def __add_route_object(self, object):
        object = proxify(object)
        objects = self.request.annotations.get(OBJECTS_ANNOTATION_KEY)
        if objects is None:
            self.request.annotations[OBJECTS_ANNOTATION_KEY] = [object]
        elif removeSecurityProxy(objects[0]) is not removeSecurityProxy(object):
            objects.insert(0, object)

    def browserDefault(self, request):
        if self.default_view is not None:
            return self.default_view, ()
        ob = self.context
        view_name = getDefaultViewName(ob, request)
        return ob, (view_name,)

    def publishTraverse(self, request, name):
        stack = request.getTraversalStack()
        stack.append(name)
        stack.reverse()
        path = "/".join(stack)

        cls = type(self)

        try:
            routes = cls._routes
        except AttributeError:
            cls._routes = routes = Routes()
            for mro_cls in cls.__mro__:
                for route_expr, kwargs in mro_cls.__dict__.get(EXPR_ATTR, ()):
                    routes.add(route_expr + "[/...]", **kwargs)

        match = routes.match(path)
        if match is None:
            names = stack[:]
            while names:
                name = "/".join(names)
                view = queryMultiAdapter((self.context, request), name=name)
                if view is not None:
                    routes = queryMultiAdapter(
                        (self.context, request), IRoutesPublisher, name=name
                    )
                    if routes:
                        routes.default_view = view
                        view = routes
                    new_stack = stack[len(names) :]
                    break
                names.pop(-1)
            else:
                new_stack = stack[:]
                view = self.default_traverse(new_stack)
                if view is None:
                    raise NotFound(self.context, name)
                self.__add_route_object(view)
        else:
            etc = match.get("etc")
            new_stack = etc and etc.split("/") or []

        offset = len(stack) - len(new_stack)

        assert new_stack == stack[offset:], (
            "The new stack should be the original stack, maybe with "
            "initial items popped."
        )

        # pop() and readd if needed. This makes it possible to have a
        # new_stack which is the same as the stack, and thus not get
        # any names consumed or traversed in this interaction.
        request._traversed_names.pop(-1)
        request._traversed_names.extend(stack[:offset])

        new_stack.reverse()
        request.setTraversalStack(new_stack)

        if match is None:
            return view
        else:
            permission = match.get("permission")
            if permission and not checkPermission(permission, self.context):
                raise Unauthorized(self, match["__handler"].__name__, permission)
            try:
                result = mapply(match["__handler"], (self,), match)
            except ZopeRedirect as e:
                try:  # Sometimes the base exception is called with no status
                    status = e.status
                except AttributeError:
                    status = None
                location = e.getLocation()
                trusted = location in self.trusted_redirects
                request.response.redirect(location, trusted=trusted, status=status)
                return NonePage
            if result is None:
                raise NotFound(self.context, name)
            self.__add_route_object(result)
            return result

    def default_traverse(self, stack):
        return None


class NonePageType(BrowserPage):
    def __call__(self):
        return None

    def publishTraverse(self, request, name):
        return self


NonePage = NonePageType(None, None)


def proxify(ob):
    """
    Wrap the given object with a security proxy, unless __no_proxy__ is set.
    """
    if getattr(ob, "__no_proxy__", None):
        return ob
    return ProxyFactory(ob)
