# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import re

import six


VARIABLE_NAME_RE = re.compile("^[a-zA-Z_][a-zA-Z_0-9]*$")


class RouteError(Exception):
    pass


class RouteNotFoundError(RouteError):
    pass


class Routes(object):
    def __init__(self):
        self._routes = []
        self._names = {}

    def add(self, route_expr, route_name=None, **kwargs):
        if not isinstance(route_expr, str):
            raise RouteError(
                "Route expression must be string, not %s" % type(route_expr)
            )
        optional_stack = []
        optional = None
        components = []
        variables = set()
        defaults = {}
        l = len(route_expr)
        i = 0
        while i < l:
            c = route_expr[i]
            if c == "[":
                optional_stack.append((optional, components))
                optional = Optional()
                components.append(optional)
                components = optional.components
                i += 1
            elif c == "]":
                if not optional:
                    raise RouteError("Unopened optional block")
                if not components:
                    raise RouteError("Empty optional block")
                optional, components = optional_stack.pop()
                i += 1
            elif c == ">":
                raise RouteError("Closing unopened variable")
            elif c == "<" or c == "." and route_expr[i : i + 3] == "...":
                if c == ".":
                    variable = "etc"
                    pattern = ".+?"
                    i += 3
                else:
                    i += 1
                    j = route_expr.find(">", i)
                    if j == -1:
                        raise RouteError("Unclosed variable")
                    variable = route_expr[i:j]
                    pattern = None
                    i = j + 1
                if variable in kwargs:
                    raise RouteError(
                        "Keyword and variable with the same name:"
                        "%s (missing _default suffix?)" % variable
                    )
                if not VARIABLE_NAME_RE.match(variable):
                    raise RouteError("Invalid variable name: %r" % variable)
                if variable in variables:
                    raise RouteError(
                        "Two or more variables with same name: %r" % variable
                    )
                variables.add(variable)
                pattern_keyword = variable + "_pattern"
                if pattern_keyword in kwargs:
                    pattern = kwargs.pop(pattern_keyword)
                default_keyword = variable + "_default"
                if default_keyword in kwargs:
                    value = kwargs.pop(default_keyword)
                    if value is not None and not isinstance(value, six.string_types):
                        raise RouteError(
                            "Invalid type for default: %s (%s)"
                            % (type(value), default_keyword)
                        )
                    defaults[variable] = value
                elif optional:
                    defaults[variable] = None
                components.append(Variable(variable, pattern))
            else:
                value = ""
                while i < l:
                    c = route_expr[i]
                    if c in "<>[]":
                        break
                    elif c == "." and route_expr[i : i + 3] == "...":
                        break
                    elif c == "\\":
                        i += 1
                        if i == l:
                            break
                        c = route_expr[i]
                    value += c
                    i += 1
                if value:
                    components.append(Literal(value))

        if optional:
            raise RouteError("Unclosed optional block")

        pattern_keywords = [x for x in kwargs if x.endswith("_pattern")]
        default_keywords = [x for x in kwargs if x.endswith("_default")]
        if pattern_keywords or default_keywords:
            invalid_keywords = default_keywords + pattern_keywords
            raise RouteError(
                "Missing variables or reserved keywords: " + ", ".join(invalid_keywords)
            )

        route = Route(components, kwargs, defaults)
        self._routes.append(route)
        self._names[route_name] = route

    def match(self, path):
        for route in self._routes:
            match = route.pattern_match(path)
            if match:
                result = route.match_model.copy()
                for key, value in list(match.groupdict().items()):
                    if value is not None:
                        result[key] = value
                return result
        return None

    def generate(self, name, **values):
        route = self._names.get(name)
        if not route:
            raise RouteNotFoundError(name)
        return route.generate(values)


class Route(object):
    def __init__(self, components, values, defaults):
        self.components = components
        self.defaults = defaults
        self.match_model = values.copy()
        self.match_model.update(defaults)
        self.pattern_match = re.compile(self.get_pattern()).match

    def get_pattern(self):
        return "^" + "".join([x.get_pattern() for x in self.components]) + "$"

    def generate(self, values):
        path = ""
        defaults = self.defaults
        for component in self.components:
            path += component.generate(values, defaults)
        return path


# Components


class Variable(object):
    def __init__(self, name, pattern=None):
        self._name = name
        if pattern is None:
            self._pattern = "(?P<%s>[^/]+?)" % name
        else:
            self._pattern = "(?P<%s>%s)" % (name, pattern)

    def get_pattern(self):
        return self._pattern

    def generate(self, values, defaults):
        name = self._name
        value = values.get(name)
        if value is None:
            value = defaults.get(name)
            if value is None:
                raise RouteError("Missing value for variable %r" % name)
        return value

    def changed(self, values):
        return values.get(self._name) is not None


class Literal(object):
    def __init__(self, value):
        self._value = value
        self._pattern = re.escape(value)

    def get_pattern(self):
        return self._pattern

    def generate(self, values, defaults):
        return self._value

    def changed(self, values):
        return False


class Optional(object):
    def __init__(self):
        self.components = []

    def get_pattern(self):
        return "(?:%s)?" % "".join([x.get_pattern() for x in self.components])

    def generate(self, values, defaults):
        path = ""
        for component in self.components:
            if component.changed(values):
                for component in self.components:
                    path += component.generate(values, defaults)
                break
        return path

    def changed(self, values):
        for component in self.components:
            if component.changed(values):
                return True
        return False
