# Copyright 2005-2022 Canonical Limited. All rights reserved.

from storm.locals import And, Or, Not


__all__ = ["SearchEngine", "SearchCriteriaError", "SearchParser", "CriteriaBase"]

__clearlist__ = ["SearchCriteriaError"]


class SearchCriteriaError(Exception):
    """Raised if an invalid search term is specified in a query."""


class SearchEngine(object):
    """
    A GMail-inspired search engine for finding objects for a specific model.

    @cvar model: the Storm model the engine will search.
    @cvar parser_class: a L{SearchParser} subclass, to parse search
       criteria.
    """

    model = None
    parser_class = None

    def __init__(self, filter_result=True):
        self._filter_result = filter_result

    def _get_search_store(self):
        """Return the store for the search."""
        raise NotImplementedError("_get_search_store() must be implemented")

    def get_search_criteria(self, query):
        """
        Parse search string and returns search criteria using the parser class.
        """
        parser = self._build_parser(query)
        return parser.parse()

    def get_where_clause(self, search_query):
        """
        Get a Storm expression that can be used to match all objects that meet
        the search criteria.

        @param query: A search query string that can be parsed by a
            search parser.
        @return: A storm expression that will match the instances of
            C{self.model} according to the query.
        """
        if not search_query:
            return []
        criteria = self.get_search_criteria(search_query)
        return criteria.get_where_clause()

    def find(self, query):
        """Find instances of C{self.model} matching C{query}.

        @param query: A search query string that can be parsed by a
            search parser.
        @return: A L{ResultSet} that yields instances of C{self.model}.
        """
        store = self._get_search_store()
        clause = self.get_where_clause(query)
        result = store.find(self.model, And(clause))
        if self._filter_result:
            return self.filter_result(result)
        else:
            return result

    def filter_result(self, result):
        """This can be overridden to filter/sort result set."""
        return result

    def _build_parser(self, query):
        """Create the parser instance to use.

        Sub-classes can override this method if they need to pass additional
        arguments to the parser class.
        """
        return self.parser_class(query)


class CriteriaBase(object):
    """
    Base class for criteria objects.
    """

    criteria_prefix = None
    criteria_attribute = None

    def get_where_clause(self):
        """
        Creates a C{Storm} where expression and return it.
        """
        raise NotImplementedError("Must be overrided by subclasses.")

    def to_string(self):
        """
        Returns a string representation of the criteria.
        """
        if self.criteria_prefix is None:
            raise AttributeError("A criteria_prefix attribute must be" " defined.")
        if self.criteria_attribute is None:
            raise AttributeError("A criteria_attribute attribute must be" " defined.")
        criteria_value = getattr(self, self.criteria_attribute, "")
        if criteria_value == "":
            raise AttributeError(
                "The %s attribute cannot return an empty"
                " value." % self.criteria_attribute
            )
        return "%s:%s" % (self.criteria_prefix, criteria_value)


class NotCriteria(object):
    """
    A criteria that negates another one.
    """

    def __init__(self, criteria):
        self.criteria = criteria

    def get_where_clause(self):
        return Not(self.criteria.get_where_clause())

    def to_string(self):
        return "NOT %s" % (self.criteria.to_string())


class GroupBase(object):
    def __init__(self, criteria):
        self.criteria = criteria

    def add(self, criteria):
        """Add a criteria object to this group.

        @param criteria: A criteria object, such as a L{KeywordCriteria},
            L{TagCriteria}, etc.
        """
        self.criteria.append(criteria)


class AndGroup(GroupBase):
    """
    A group of criteria that must all be matched for objects to be included in
    a search result.
    """

    def get_where_clause(self):
        """
        Get a where clause to find objects matching all specified sub-criteria.

        @return: A Storm expression that will match all sub-criteria.
        """
        return [criteria.get_where_clause() for criteria in self.criteria]

    def to_string(self):
        """Convert this criteria to a string."""
        return " ".join([criteria.to_string() for criteria in self.criteria])


class OrGroup(GroupBase):
    """
    A group of criteria for which only one must match for objects to be
    included in a search result.
    """

    def get_where_clause(self):
        """
        Get a where clause to find objects matching at least one specified
        sub-criteria.

        @return: A Storm expression that will match at least one sub-criteria.
        """
        where = []
        for criteria in self.criteria:
            where.append(criteria.get_where_clause())
        return Or(where)

    def to_string(self):
        """Convert this criteria to a string."""
        return " OR ".join([criteria.to_string() for criteria in self.criteria])


class SearchParser(object):
    """Convert query strings to criteria objects.

    Sub-classes are required to implement a _get_<name>_criteria method for
    each criteria they support, returning the criteria object for the <name>
    keyword in the query string.
    """

    def __init__(self, query):
        self._query = query.strip()
        self._current = None
        self._terms = None
        self._criteria = []
        self._or_criteria = None
        self._grammar = self._get_grammar()

    def parse(self):
        """
        Parse the query string and convert it into a criteria object that can
        be used to generate an SQL where clause suitable for use with Storm.

        @raises SearchCriteriaError: Raised if the query string contains an
            invalid search.
        @return: An L{AndGroup} criteria object containing sub-criteria
            specified in the query string.
        """
        if not self._query:
            return

        from pyparsing import ParseException

        try:
            self._terms = self._grammar.parseString(self._query)
        except ParseException:
            raise SearchCriteriaError("Couldn't parse search query '%s'." % self._query)

        self._advance()
        while self._current is not None:
            if self._current == "OR":
                if not self._criteria:
                    raise SearchCriteriaError("OR can't be the first term in a query.")
                if self._or_criteria is None:
                    last_criteria = self._criteria.pop()
                    self._or_criteria = OrGroup([last_criteria])
                    self._criteria.append(self._or_criteria)
                self._advance()
                if self._current in ("OR", None):
                    raise SearchCriteriaError("OR is missing subsequent term.")
                if self._current == "NOT":
                    self._advance()
                    if self._current in ("OR", "NOT", None):
                        raise SearchCriteriaError("NOT is missing a subsequent term.")
                    criteria = NotCriteria(self._get_criteria(self._current))
                else:
                    criteria = self._get_criteria(self._current)
                self._or_criteria.add(criteria)
            elif self._current == "NOT":
                self._advance()
                if self._current in ("NOT", "OR", None):
                    raise SearchCriteriaError("NOT is missing a subsequent term.")
                criteria = self._get_criteria(self._current)
                self._criteria.append(NotCriteria(criteria))
            else:
                criteria = self._get_criteria(self._current)
                self._criteria.append(criteria)
                self._or_criteria = None
            self._advance()

        return AndGroup(self._criteria)

    def _get_grammar(self):
        """
        Return the basic grammar supporting list of key/values separated by a
        colon.
        """
        from pyparsing import Word, Group, alphanums, OneOrMore, Literal, printables

        def remove_end(string, location, tokens):
            return tokens[0]

        # A search value is a printable
        entry_value = Word(printables)
        # A search term is an alphanumeric, with . and - allowed, ending by :.
        key_value = Word(alphanums + ".-") + Literal(":")
        key_value.setParseAction(remove_end)
        # A search entry is a term and a value, or OR, or NOT
        entry = Group(key_value + entry_value) ^ Literal("OR") ^ Literal("NOT")
        return OneOrMore(entry)

    def _advance(self):
        """Advance to the next term and set it as the current one."""
        if self._terms:
            self._current = self._terms.pop(0)
        else:
            self._current = None

    def _get_criteria(self, term):
        """Get a criteria object for the current term.

        @raises SearchCriteriaError: Raised if the term doesn't represent a
            valid search criteria.
        @return: A criteria object that can be used to generate where clauses
            suitable for finding matching objects.
        """
        if isinstance(term, str):
            raise SearchCriteriaError("'%s' is an invalid term." % term)

        selector, data = term

        selector = selector.replace("-", "_")
        method = getattr(self, "_get_%s_criteria" % (selector,), None)
        if method is None:
            raise SearchCriteriaError("'%s' is an unknown selector." % selector)
        return method(data)
