# Copyright 2005-2022 Canonical Limited.  All rights reserved.

"""This module contains the L{ImportGuardian}."""

from six.moves import builtins


class ImportGuardian(object):
    """Policy-based importer is a replacement for C{builtins.__import__}.

    Various kinds of import situations may be undesirable in an application.
    There may be coding conventions that should be adhered to or security
    infrastructure that must be used.  It may be undesirable to allow
    C{import *} from a module that doesn't define C{__all__} or there may be
    classes which much always be imported with L{SafeImporter}, for example.
    A set of L{stock import policies<canonical.security.importpolicy>} are
    available to tailor the behaviour of the importer.
    """

    def __init__(self):
        self._original_importer = None
        self._policies = []

    def install(self):
        """Replace the current importer with this guardian.

        If this importer is the installed one, nothing will be changed.

        @raise RuntimeError: Raised if another L{ImportGuardian} is installed.
        """
        if builtins.__import__ is self:
            return
        if isinstance(builtins.__import__, ImportGuardian):
            raise RuntimeError(
                "An ImportGuardian is already installed.  Only "
                "one importer may be installed at a time."
            )
        self._original_importer = builtins.__import__
        builtins.__import__ = self

    def uninstall(self):
        """Restore the importer that was installed previously."""
        if self._original_importer:
            builtins.__import__ = self._original_importer
            self._original_importer = None

    def add_policy(self, policy):
        """Add C{policy} to the importer."""
        self._policies.append(policy)

    def __call__(self, import_from, globals={}, locals={}, fromlist=[], level=0):
        """Perform an import.

        When a module is successfully imported policies are consulted to
        determine whether or not the import is authorized.  Checking occurs in
        a two step process:

         1. The C{override_checks} method is called on each policy.  If any
            policy return False checking is disabled and the import is
            allowed.

         2. If all policies require checking, the C{check} method is called on
            each policy.  The check logic raises an exception if the import
            should be denied.

        @raise ImportError: Raised if the import can be performed.
        @raise ImportPolicyError: Raised by policies when constraints
            aren't satisfied.
        """
        module = self._original_importer(import_from, globals, locals, fromlist, level)
        if globals:
            context = ImportContext(
                module, import_from, globals, locals, fromlist, level
            )
            if context.import_into:
                for policy in self._policies:
                    if policy.override_checks(context):
                        break
                else:
                    for policy in self._policies:
                        policy.check(context)

        return module


class ImportContext(object):
    def __init__(self, module, import_from, globals, locals, fromlist, level):
        self.module = module
        self.import_into = globals.get("__name__")
        self.import_from = import_from
        self.globals = globals
        self.locals = locals
        self.from_list = None
        self.level = level
        if fromlist:
            self.from_list = list(fromlist)
