# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""Stock policies for use with the import guardian."""

import re


__all__ = [
    "FromStarNoModuleAllPolicy",
    "NotInModuleAllPolicy",
    "ImportFromPolicy",
    "ImportIntoPolicy",
    "SafeImportPolicy",
    "FromStarNoModuleAllPolicyError",
    "NotInModuleAllPolicyError",
    "SafeImportPolicyError",
    "ImportPolicyError",
    "WhitelistPolicyError",
]


class ImportPolicyError(ImportError):
    """Raised when an import policy denies an import."""

    def __init__(self, import_into, import_from, *args):
        ImportError.__init__(self, import_into, import_from, *args)
        self.import_into = import_into
        self.import_from = import_from

    def format_message(self):
        return "Generic ImportPolicyError: '%s' imported into '%s'." % (
            self.import_from,
            self.import_into,
        )

    def __str__(self):
        return self.format_message()


class FromStarNoModuleAllPolicyError(ImportPolicyError):
    """Raised when C{import *} is used with a module that has no C{__all__}."""

    def format_message(self):
        return (
            "You should not 'from %s import *' because it has no __all__ "
            "(called in '%s')." % (self.import_from, self.import_into)
        )


class NotInModuleAllPolicyError(ImportPolicyError):
    """
    Raised when an imported name does not appear in a module's C{__all__}.
    """

    def __init__(self, import_into, import_from, name):
        ImportPolicyError.__init__(self, import_into, import_from, name)
        self.name = name

    def format_message(self):
        return (
            "You should not 'from %s import %s' because '%s' is "
            "not in __all__ (called in '%s')."
            % (self.import_from, self.name, self.name, self.import_into)
        )


class SafeImportPolicyError(ImportPolicyError):
    """Raised when import needs to use a safe import."""

    def __init__(self, import_into, import_from, prefix):
        ImportPolicyError.__init__(self, import_into, import_from, prefix)
        self.prefix = prefix

    def format_message(self):
        return (
            "You cannot import %s directly from %s.  Use the safe prefix "
            "'%s' to import security-sensitive code from %s."
            % (self.import_from, self.import_into, self.prefix, self.import_from)
        )


class WhitelistPolicyError(ImportPolicyError):
    """Raised when a name used with safe import isn't in the whitelist."""

    def __init__(self, import_into, import_from, name):
        ImportPolicyError.__init__(self, import_into, import_from, name)
        self.name = name

    def format_message(self):
        return (
            "You cannot import %s directly into %s from %s.  "
            "It is not in the whitelist."
            % (self.name, self.import_into, self.import_from)
        )


class ImportPolicyBase(object):
    def override_checks(self, guardian):
        """All policy checks are disabled if this method returns C{True}."""
        return False

    def check(self, guardian):
        """Raise an exception if policy checks deny the import."""
        pass


class FromStarNoModuleAllPolicy(ImportPolicyBase):
    """Only allow C{import *} for modules that define C{__all__}."""

    def check(self, context):
        """
        @raise FromStarNoModuleAllPolicyError: Raised if C{import *}
            is used to import a module that doesn't define C{__all__}.
        """
        if context.from_list == ["*"] and not hasattr(context.module, "__all__"):
            raise FromStarNoModuleAllPolicyError(
                context.import_into, context.import_from
            )


class RegexpFilterPolicy(ImportPolicyBase):
    """
    Filtering policy limits the scope of code that is subject to
    import policing.
    """

    def __init__(self):
        self.include_patterns = set()
        self._include_regexps = set()
        self.exclude_patterns = set()
        self._exclude_regexps = set()

    def include_regexp(self, pattern):
        """
        Imports from modules with names that match C{pattern} will be
        subject to import policies.
        """
        self.include_patterns.add(pattern)
        self._include_regexps.add(re.compile(pattern))

    def exclude_regexp(self, pattern):
        """
        Imports from modules with names that match C{pattern} will not
        be subject to import policies.
        """
        self.exclude_patterns.add(pattern)
        self._exclude_regexps.add(re.compile(pattern))

    def override_checks(self, context):
        """
        Exclude patterns have higher precedence than include patterns.
        """
        value = self.get_value(context)
        if not value:
            return False
        if self._exclude_regexps:
            for regexp in self._exclude_regexps:
                if regexp.match(value):
                    return True

        if not self._include_regexps:
            return False

        for regexp in self._include_regexps:
            if regexp.match(value):
                return False
        return True


class ImportFromPolicy(RegexpFilterPolicy):
    """Filtering policy limits the scope of possible imports from a module."""

    def get_value(self, context):
        return context.import_from


class ImportIntoPolicy(RegexpFilterPolicy):
    """Filtering policy limits the scope of possible imports into a module."""

    def get_value(self, context):
        return context.import_into


class NotInModuleAllPolicy(ImportIntoPolicy):
    """Only allow names to be imported that are in a module's C{__all__}."""

    def check(self, context):
        """
        @raise NotInModuleAllPolicyError: Raised if a name is imported
            from a module that isn't in C{__all__}.
        """
        if (
            context.from_list == ["*"]
            or not context.from_list
            or not hasattr(context.module, "__all__")
        ):
            return

        whitelist = set(context.module.__all__)
        for name in context.from_list:
            if name not in whitelist:
                raise NotInModuleAllPolicyError(
                    context.import_into, context.import_from, name
                )


class SafeImportPolicy(ImportPolicyBase):
    """Policy ensures that security sensitive code is imported safely."""

    def __init__(self, prefix="safe"):
        self.exclude_patterns = set()
        self._exclude_regexps = set()
        self._prefix = prefix
        self._rules = []

    def require_safe(self, import_from=(), import_into=(), whitelist=(), clearlist=()):
        """Require safe imports for a security sensitive module.

        Policy requires safe imports when names from C{import_from}
        are imported into modules listed in C{import_into}.  Safe
        imports are required from all modules if C{import_into} is
        empty.  Only names from C{whitelist} can be imported by
        safe-requiring modules.  All names can be imported by
        safe-requiring modules if C{whitelist} is empty.  Names in
        C{clearlist} may be imported without safe by any module.

        @param import_from: A list of regexp patterns to match against
            modules names used to determine if imports from them must
            be made safely.
        @param import_into: A list of regexp patterns to match against
            module names.  The import must be made safely if a module
            name requesting an import matches patterns in this list.
        @param whitelist: A list of names that can be imported from
            the sensitive module.  If C{whitelist} is empty all names
            can be imported.
        @param clearlist: A list of names that can be imported
            without using a safe import.
        """
        import_from_regexp = [re.compile(pattern) for pattern in import_from]
        import_into_regexp = [re.compile(pattern) for pattern in import_into]
        rule = SafetyRule(
            set(import_from_regexp),
            set(import_into_regexp),
            set(whitelist),
            set(clearlist),
            self._prefix,
        )
        self._rules.append(rule)

    def check(self, context):
        """
        @raise SafeImportPolicyError: Raised if the import needs to be safe.
        @raise WhitelistPolicyError: Raised if a safe import tries to
            import a protected name.
        """
        for rule in self._rules:
            rule.check(context)


class SafetyRule(object):
    """Provide policy checking functionality for safe import rules."""

    def __init__(self, import_from, import_into, whitelist, clearlist, prefix):
        self.import_from = import_from
        self.import_into = import_into
        self.whitelist = whitelist
        self.clearlist = clearlist
        self.prefix = prefix

    def check(self, context):
        """Check the rule against C{context}.

        @raise SafeImportPolicyError: Raised if the import needs to be safe.
        @raise WhitelistPolicyError: Raised if a safe import tries to
            import a protected name.
        """
        if self.guard_import_from(context):
            if self.untrusted_import_into(context):
                if not self.check_clearlist(context):
                    self.check_safe(context)
                    self.check_whitelist(context)

    def guard_import_from(self, context):
        """Returns C{True} if the module being imported is safe sensitive."""
        for regexp in self.import_from:
            unsafe_name = context.import_from
            if unsafe_name.startswith(self.prefix + "."):
                unsafe_name = unsafe_name[len(self.prefix) + 1 :]
            if regexp.match(unsafe_name):
                return True
        return False

    def untrusted_import_into(self, context):
        """Returns C{True} if the module doing the import is untrusted."""
        if not self.import_into:
            return True
        for regexp in self.import_into:
            if regexp.match(context.import_into):
                return True
        return False

    def check_clearlist(self, context):
        """Return C{True} if the names in the from list are clearlisted."""
        clearlist = set()
        clearlist.update(self.clearlist)
        if hasattr(context.module, "__clearlist__"):
            clearlist.update(context.module.__clearlist__)
        if context.from_list:
            from_list = set(context.from_list)
            return from_list.issubset(clearlist)
        return False

    def check_whitelist(self, context):
        """Check that the names being imported are in the whitelist."""
        if self.whitelist:
            if context.from_list:
                from_list = set(context.from_list)
                diff = from_list - self.whitelist
                if diff and not diff.issubset(self.clearlist):
                    raise WhitelistPolicyError(
                        context.import_into, context.import_from, diff.pop()
                    )

    def check_safe(self, context):
        """Check that the import is safe."""
        if not context.import_from.startswith(self.prefix + "."):
            raise SafeImportPolicyError(
                context.import_into, context.import_from, self.prefix
            )
