# Copyright 2005-2013 Canonical Limited.  All rights reserved.

from zope.security.simplepolicies import ParanoidSecurityPolicy
from zope.security.management import system_user
from zope.security.checker import CheckerPublic
from zope.security.proxy import removeSecurityProxy
from zope.component import queryMultiAdapter, getUtility
from zope.interface import implementer

from canonical.security.interfaces import IPermissionChecker, IRoleRegistry


class SecurityPolicy(ParanoidSecurityPolicy):

    role_registry_name = ""

    def __init__(self, *participations):
        super(SecurityPolicy, self).__init__(*participations)
        self.role_registry = getUtility(IRoleRegistry, self.role_registry_name)

    def checkPermission(self, permission, object):
        if permission is CheckerPublic:
            return True

        object = removeSecurityProxy(object)

        for participation in self.participations:
            principal = participation.principal
            if principal is system_user:
                return True
            checker = queryMultiAdapter((object, principal), IPermissionChecker)
            if checker:
                decision = checker(permission)
                if decision is not None:
                    return decision

        # Check default roles.
        decision = self.role_registry.may(None, permission)
        if decision is not None:
            return decision

        return False


@implementer(IPermissionChecker)
class RolePermissionChecker(object):
    """Base class for implementing role-based permission checkers.

    A L{RolePermissionChecker} grants permissions based the roles that the
    adapted principal has on the adapted object.

    Subclasses must implement the L[get_roles} method, which is used to
    determine the roles the accessing principal has on the accessed context,
    and then an L{IRoleRegistry} utility is used to determine if any of
    these roles grant the required permission.

    @ivar registry: The L{IRoleRegistry} named utility used to check
       if a role grants a given permission, the global one by default.
    @ivar context: The accessed object.
    @ivar principal: The accessing object.
    """

    registry_name = ""

    def __init__(self, context, principal):
        self.context = context
        self.principal = principal
        self.registry = self.get_role_registry()
        self._roles = self  # self is marker, None means no roles.

    def get_role_registry(self):
        """By default the global role registry is returned."""
        return getUtility(IRoleRegistry, self.registry_name)

    def get_roles(self):
        raise NotImplementedError("get_roles() must be implemented")

    def __call__(self, permission):
        """Indicates if the principal has the given permission on the context.

        Determine if the roles returned by L{get_roles} grant the requred
        permission.
        """
        if self._roles is self:
            self._roles = self.get_roles()
        if self._roles:
            decision = self.registry.may(self._roles, permission)
            if decision is None:
                decision = self.registry.may(None, permission)
        else:
            decision = self.registry.may(None, permission)
        return decision
