# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import sys
import logging

import six

from zope.interface.interface import Method
from zope.interface.advice import addClassAdvisor
from zope.security.checker import Checker, CheckerPublic
from zope.security.checker import defineChecker, getCheckerForInstancesOf


__all__ = [
    "requires",
    "PUBLIC",
    "PRIVATE",
    "UnknownPermissionError",
    "protect_names",
    "protect_class",
    "protect_iface",
]


class UnknownPermissionError(Exception):
    """Raised when a permission isn't registered as a utility."""


PUBLIC = CheckerPublic
PRIVATE = object()


def requires(get_permission, set_permission=None, names=None):
    """Apply get and/or set permissions class-wide, or for a named item.

    This function is useful in three situations:

    1. To define class-wide security requirements:
       requires("PermissionName", "PermissionName")

    2. To define security requirements for a particular class item:
       requires("PermissionName", "PermissionName", ["method_name"])

    3. To define security requirements using decorator syntax:
       @requires("PermissionName", "PermissionName")
       def method_name(self):
           pass

    It accomplishes these goals by setting up temporary advice data,
    which is used during class creation to apply security permissions.
    When security permissions have been applied, the advice data is
    destroyed.
    """
    frame = sys._getframe(1)
    locals = frame.f_locals
    globals = frame.f_globals
    prefix = locals.get("__requires_prefix__")
    if prefix is None:
        prefix = globals.get("__requires_prefix__")
    if prefix is not None:
        if isinstance(get_permission, six.string_types):
            get_permission = prefix + get_permission
        if isinstance(set_permission, six.string_types):
            set_permission = prefix + set_permission
    if locals is not globals:
        # Class-level definition.
        advice_data = locals.get("__requires_advice_data__")
        if not advice_data:
            locals["__requires_advice_data__"] = advice_data = []
            # addClassAdvisor(_requires_advisor)  XXX impossible in py3
        advice_data.append((get_permission, set_permission, names))

        def class_decorator(func):
            assert names is None, "'requires' decorator has only two arguments"
            advice_data.pop()  # Undo non-decorator logic.
            advice_data.append((get_permission, set_permission, [func.__name__]))
            return func

        return class_decorator
    else:
        # Module-level decorator definition
        def module_decorator(func):
            assert names is None, "'requires' decorator has only two arguments"
            assert func.__module__, "Can't tell what module the function is in"
            module = __import__(func.__module__, None, None, [""])
            checker = getCheckerForInstancesOf(module)
            if checker is None:
                checker = Checker({}, {})
                defineChecker(module, checker)
            checker.get_permissions[func.__name__] = get_permission
            if set_permission is not None:
                checker.set_permissions[func.__name__] = set_permission
            return func

        return module_decorator


def requires_advisor(cls):
    for base in cls.__mro__:
        advice_data = getattr(base, "__requires_advice_data__", None)
        if advice_data:
            for get_permission, set_permission, names in advice_data:
                if names is None:
                    protect_class(cls, get_permission, set_permission, override=False)
                else:
                    protect_names(cls, names, get_permission, set_permission)
    cls.__Security_checker__ = getCheckerForInstancesOf(cls)
    return cls


def protect_names(cls, names, get_permission, set_permission=None, override=True):
    checker = getCheckerForInstancesOf(cls)
    if not checker:
        checker = Checker({}, {})
        defineChecker(cls, checker)
    for name in names:
        if get_permission is PRIVATE:
            if override or name not in checker.get_permissions:
                checker.get_permissions[name] = None
        elif get_permission is not None and (
            override or name not in checker.set_permissions
        ):
            checker.get_permissions[name] = get_permission
        if set_permission is PRIVATE:
            if override or name not in checker.set_permissions:
                checker.set_permissions[name] = None
        elif set_permission is not None and (
            override or name not in checker.set_permissions
        ):
            checker.set_permissions[name] = set_permission


def protect_class(cls, get_permission, set_permission=None, override=True):
    attributes = []
    methods = []
    names = dir(cls)

    if hasattr(cls, "__dataclass_fields__"):
        # Dataclass fields don't exist yet. This lets us recognize them.
        names += list(cls.__dataclass_fields__.keys())

    for name in names:
        if name[0] == "_" and name not in ("__iter__", "__getitem__", "__getslice__"):
            pass
        elif callable(getattr(cls, name, None)):
            methods.append(name)
        else:
            attributes.append(name)

    protect_names(cls, attributes, get_permission, set_permission, override)
    protect_names(cls, methods, get_permission, None, override)


def protect_iface(cls, iface, get_permission, set_permission=None, override=True):
    attributes = []
    methods = []
    for name, descr in iface.namesAndDescriptions(all=True):
        if isinstance(descr, Method):
            methods.append(name)
        else:
            attributes.append(name)
    protect_names(cls, attributes, get_permission, set_permission, override)
    protect_names(cls, methods, get_permission, None, override)
