# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.interface import implementer

from canonical.security.interfaces import IRoleRegistry, IRole
from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE


class DuplicateRoleError(Exception):
    """Raised when a duplicate role is registered."""


class UnknownRoleError(Exception):
    """Raised when an unknown role is used."""


@requires_advisor
@implementer(IRoleRegistry)
class RoleRegistry(object):
    """Registry for role => permission information."""

    requires(PUBLIC, PRIVATE, ["get", "get_with_permission", "may"])

    def __init__(self):
        self._roles = {}  # {role.key|role.name: role}
        self._perms = {}  # {role: {perm: True|False}}
        self._default = set()  # set([role1, role2])

        # Built just-in-time including inheritance information.
        self._full_perms = None

    def add(self, role, default=False):
        """Add given role to registry.

        @param role: Role being added to the registry.
        @param default: If true, this role will be added to the set of
            default roles.  Default roles are used when may() is
            given None as the C{roles} argument.
        """
        if role.key is not None and role.key in self._roles:
            raise DuplicateRoleError(
                "Role with key %r is already registered" % role.key
            )
        if role.name in self._roles:
            raise DuplicateRoleError(
                "Role with name %r is already registered" % role.name
            )
        if role.key is not None:
            self._roles[role.key] = role
        self._roles[role.name] = role
        self._perms[role] = {}
        self._full_perms = None

        if default:
            self._default.add(role)

    def get(self, key):
        """Return a role with the given name or key.

        @param key: Role name or key.
        @return: The role.
        """
        role = self._roles.get(key)
        if role is None:
            raise UnknownRoleError("Role %r isn't registered" % key)
        return role

    def get_all(self):
        """Return a list of all roles."""
        return set(self._perms.keys())

    def get_all_permissions(self, prefix_filter=None):
        """
        Return a set of all permissions in the registry.

        @param prefix_filter: An optional filter to limit the returned
            permissions to one prefix. For example to get all permissions that
            start with "landscape." use a prefix_filter of "landscape". If a
            prefix_filter is supplied the prefix is stripped from the returned
            permissions.
        """
        perms = set()
        for role in list(self._perms.keys()):
            if prefix_filter is None:
                # No prefix filter, just add all of them
                perms += set(self._perms[role].keys())
            else:
                for permission in list(self._perms[role].keys()):
                    prefix, permission_name = permission.split(".")
                    if prefix == prefix_filter:
                        perms.add(permission_name)
        return perms

    def get_with_permission(self, permission):
        """Returns a list of roles that provide the specified permission."""
        if self._full_perms is None:
            self._build_full_perms()

        roles = []
        for role, index in list(self._full_perms.items()):
            if index.get(permission):
                roles.append(role)
        return roles

    def __contains__(self, key):
        """Return whether a role is present in the registry."""
        if isinstance(key, Role):
            key = key.name
        return key in self._roles

    def remove(self, role):
        """Remove the given role from the registry.

        @param role: Role to be removed from the registry.
        """
        keyed_role = self._roles.get(role.key)
        named_role = self._roles.get(role.name)
        if keyed_role is None and named_role is None:
            raise UnknownRoleError("Role %r isn't registered" % role.name)
        if keyed_role is not role and keyed_role is not None:
            raise UnknownRoleError(
                "Role with key %r doesn't match what's " "registered" % role.key
            )
        if named_role is not role and named_role is not None:
            raise UnknownRoleError(
                "Role with name %r doesn't match what's " "registered" % role.name
            )
        del self._perms[role]
        del self._roles[role.name]
        if role.key is not None:
            del self._roles[role.key]
        self._default.discard(role)
        self._full_perms = None

    def grant(self, role, perm):
        """Grant a permission to a role.

        @param role: Role to be granted the given permission.
        @param perm: Permission to be granted.
        """
        try:
            self._perms[role][perm] = True
        except KeyError:
            raise UnknownRoleError("Role %r isn't registered" % role.name)
        self._full_perms = None

    def forbid(self, role, perm):
        """Deny access to a given permission explicitly.

        @param role: Role to be forbidden from having the given permission.
        @param perm: Permission to be forbidden.

        If another role in the list given to C{may()} has the permission,
        it will be allowed even if it was forbidden for this one role.
        This forbid logic is most useful with three-state checks (allow,
        deny, unknown).
        """
        try:
            self._perms[role][perm] = False
        except KeyError:
            raise UnknownRoleError("Role %r isn't registered" % role.name)
        self._full_perms = None

    def revoke(self, role, perm):
        """Revoke a grant previously made.

        @param role: Role to have the grant revoked.
        @param perm: Permission to be revoked.
        """
        try:
            perms = self._perms[role]
        except KeyError:
            raise UnknownRoleError("Role %r isn't registered" % role.name)
        if perm in perms:
            del perms[perm]
        self._full_perms = None

    def _build_full_perms(self):
        self._full_perms = full_perms = {}

        # Initialize consolidated state from raw settings.
        for role in self._perms:
            full_perms[role] = self._perms[role].copy()

        # While any change was made to the consolidated state, ...
        updated = True
        while updated:
            updated = False

            # ... for each role known, ...
            for role in self._perms:
                role_perms = full_perms[role]

                # ... for each role inherited by said known role, ...
                for irole in role.inherits:

                    # ... for each permission of the inherited role, ...
                    irole_perms = full_perms.get(irole, ())
                    for perm in irole_perms:

                        # ... copy the permission from the inherited role
                        # to the known role, but only if it is a change,
                        # and if it's not doing a True => False change
                        # (grants have a higher precedence than forbids).
                        if (
                            not role_perms.get(perm)
                            and role_perms.get(perm) != irole_perms[perm]
                        ):
                            role_perms[perm] = irole_perms[perm]
                            updated = True

    def may(self, roles, perm):
        """Check if the given permission is allowed to given roles.

        @param roles: Sequence of roles, or None for checking default roles.
        @param perm: Permission to be checked for.
        @return: True if any of the roles were granted the permission, False
            if any of them were forbidden, or None otherwise.
        """
        if self._full_perms is None:
            self._build_full_perms()

        if roles is None:
            roles = self._default

        full_perms = self._full_perms
        result = None
        for role in roles:
            try:
                setting = full_perms[role].get(perm)
            except KeyError:
                raise UnknownRoleError("Role %r isn't registered" % role.name)
            if setting is True:
                return True
            if setting is False:
                result = False
        return result

    def clone(self):
        """
        Return a copy of the registry, keeping references to registered roles.
        """
        registry = RoleRegistry()
        registry._roles = self._roles.copy()
        registry._perms = self._perms.copy()
        registry._default = self._default.copy()
        return registry


@requires_advisor
@implementer(IRole)
class Role(object):

    requires(PUBLIC, PRIVATE, ["name", "title", "description", "key"])

    key = None
    inherits = ()

    def __init__(self, name, title=None, description=None, key=None, inherits=None):
        self.name = name
        self.title = title
        self.description = description
        if key is not None:
            if not isinstance(key, int):
                raise TypeError("Role keys must be numeric")
            self.key = key
        if inherits is not None:
            self.inherits = tuple(inherits)

    def __repr__(self):
        return "<Role %r>" % (self.name,)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def is_descendant_of(self, role):
        """Return a boolean indicating if we're descendant of C{role}."""
        for ancestor in self.inherits:
            if ancestor is role or ancestor.is_descendant_of(role):
                return True

        return False


global_role_registry = RoleRegistry()
