# Copyright 2005-2013 Canonical Limited.  All rights reserved.

import sys
import types

import importlib

from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec

from zope.proxy import ProxyBase, getProxiedObject
from zope.security.checker import (
    Checker,
    Proxy,
    NoProxy,
    selectChecker,
    getCheckerForInstancesOf,
)


__all__ = ["install", "uninstall", "uninstall_all"]


def install(prefix="safe"):
    """Install the safe importer."""
    for importer in sys.meta_path:
        if isinstance(importer, SafeImporter) and importer.prefix == prefix:
            return

    importer = SafeImporter(prefix)
    sys.meta_path.append(importer)


def uninstall(prefix="safe"):
    """Uninstall the safe importer."""
    for i in range(len(sys.meta_path) - 1, -1, -1):
        importer = sys.meta_path[i]
        if isinstance(importer, SafeImporter) and importer.prefix == prefix:
            del sys.meta_path[i]


def uninstall_all():
    for i in range(len(sys.meta_path) - 1, -1, -1):
        importer = sys.meta_path[i]
        if isinstance(importer, SafeImporter):
            del sys.meta_path[i]


closed_checker = Checker({}, {})


def getChecker(object):
    checker = getattr(object, "__Security_checker__", None)
    if checker is None:
        checker = getCheckerForInstancesOf(type(object))
        while not (checker in (None, NoProxy) or isinstance(checker, Checker)):
            checker = checker(object)
    return checker


class SecuredModule(ProxyBase):
    __slots__ = ("__path__", "__submodules__")

    def __init__(self, module):
        super(SecuredModule, self).__init__(module)
        super(SecuredModule, self).__setattr__("__path__", None)
        super(SecuredModule, self).__setattr__("__submodules__", {})

    def __setattr__(self, name, value):
        if isinstance(value, SecuredModule):
            submodules = ProxyBase.__getattribute__(self, "__submodules__")
            submodules[name] = value

    def __getattribute__(self, name):
        # The import hook system depends on __path__ and __spec__.
        if name == "__spec__":
            return ProxyBase.__getattribute__(self, "__spec__")

        if name == "__path__":
            return ProxyBase.__getattribute__(self, "__path__")

        submodules = ProxyBase.__getattribute__(self, "__submodules__")
        if name in submodules:
            return submodules[name]

        # Get the real module and the object being accessed, and
        # test to see if we don't have to proxy it, or if it's
        # doing module traversal.
        module = getProxiedObject(self)
        object = getattr(module, name)
        if type(object) in (Proxy, SecuredModule):
            return object
        if isinstance(object, types.ModuleType):
            return SecuredModule(object)

        # Check if the module was tagged as not wanting a proxy,
        # or if there's no checker registered at all. In the later
        # case we define a checker so that upcoming security
        # definitions will use the same checker.
        module_checker = getCheckerForInstancesOf(module)
        if module_checker is NoProxy:
            return getattr(module, name)

        def precondition():
            selectChecker(module).check(module, name)

        checker = getChecker(object)
        if checker is None:
            return Proxy(object, closed_checker)
        if checker is NoProxy:
            return Proxy(object, PreconditionChecker(precondition))
        return Proxy(object, PreconditionProxyChecker(precondition, checker))


class PreconditionChecker(object):
    def __init__(self, precondition):
        self._precondition = precondition

    def check_setattr(self, object, name):
        self._precondition()

    def check_getattr(self, object, name):
        self._precondition()

    check = check_getattr


class PreconditionProxyChecker(object):
    def __init__(self, precondition, checker):
        self._precondition = precondition
        self._check_getattr = checker.check_getattr
        self._check_setattr = checker.check_setattr
        self.proxy = checker.proxy

    def check_setattr(self, object, name):
        self._precondition()
        self._check_setattr(object, name)

    def check_getattr(self, object, name):
        self._precondition()
        self._check_getattr(object, name)

    check = check_getattr


class SafeImporter(MetaPathFinder, Loader):
    def __init__(self, prefix="safe"):
        self.prefix = prefix
        self._safe_module_spec = ModuleSpec(prefix, self)
        self._safe_module = types.ModuleType(prefix.split(".")[-1])
        self._safe_module.__spec__ = self._safe_module_spec
        self._safe_module.__path__ = [prefix.replace(".", "/")]

    def __repr__(self):
        return "SafeImporter"

    def find_spec(self, fullname, path, target=None):
        if not fullname.startswith(self.prefix):
            return None

        if fullname == self.prefix:
            # Top-level import.
            return self._safe_module_spec

        return ModuleSpec(fullname, self)

    def create_module(self, spec):
        if spec.name == self.prefix:
            return self._safe_module

        name = spec.name[len(self.prefix) + 1 :]
        module = importlib.import_module(name)

        if isinstance(module, SecuredModule):
            return module

        module.__spec__ = spec
        secured_module = SecuredModule(module)

        return secured_module

    def exec_module(self, module):
        # Nothing to do. Module will either be executed within
        # self.create_module or it doesn't need to be executed.
        return
