# Copyright 2017 Canonical Limited.  All rights reserved.

from canonical.lib.graph import iter_lineage


__all__ = [
    "UnknownStatusError",
    "is_enabled",
    "is_disabled",
    "is_context_disabled",
]


NOT_SET = object()


class UnknownStatusError(RuntimeError):
    """Indicates that an object's status could not be determined."""


def is_enabled(obj, default=NOT_SET):
    """Return True if obj is enabled.

    Return the default if the status of the object is unknown.  If no
    default is provided, raise UnknownStatusError.

    See is_disabled().
    """
    try:
        return not is_disabled(obj)
    except UnknownStatusError:
        if default is NOT_SET:
            raise
        return default


def is_disabled(obj, default=NOT_SET):
    """Return True if obj is disabled.

    Return the default if the status of the object is unknown.  If no
    default is provided, raise UnknownStatusError.

    The following are checked:
      - obj.is_disabled
    """
    try:
        return obj.is_disabled
    except AttributeError:
        if default is NOT_SET:
            raise UnknownStatusError
        return default


def is_context_disabled(obj, default=False):
    """Return True if obj.context is disabled.

    If the object is not disabled or doesn't have a context then return
    the default.

    If obj has parents then check their contexts too.
    """
    for obj in iter_lineage(obj):
        try:
            ctx = obj.context
        except AttributeError:
            continue
        if is_disabled(ctx, default):
            return True
    else:
        return default
