from canonical.storm.compile import compile_columns, compile_rows


STATEMENT_TEMPLATE = "INSERT INTO {} ({}) VALUES {}"


class BulkInsert(object):
    """Utility for performing fast bulk insertion of data.

    This class will leverage some Storm object-relational mapping features, but
    use low-level drivers to get the job done fast, without the overhead
    involved with Storm objects.
    """

    def __init__(self, store, columns):
        """
        @param store: The store to execute the bulk insert in.
        @param columns: The Storm model class columns to include in the INSERT.
        """
        self._store = store
        self._rows = []  # List of tuples with INSERT values
        self._columns = columns

    def add(self, rows):
        """Add rows to the bulk insertion.

        @param rows: A list of tuples, each of one matching the columns
            passed to the constructor.
        """
        self._rows.extend(rows)

    def execute(self, returning=None):
        """Execute a bulk insertion of all rows added with add().

        @param returning: Optional list of columns to include as RETURNING
            clause.
        @return: A result set yielding the values of created primary keys,
            if any.
        """
        columns = compile_columns(self._columns)
        rows = compile_rows(self._rows)
        model_class = self._columns[0].cls
        table = model_class.__storm_table__

        statement = STATEMENT_TEMPLATE.format(table, columns, rows)

        if returning:
            returning_columns = compile_columns(returning)
            statement += " RETURNING {}".format(returning_columns)

        # Use the Storm connection wrapper around the underlying
        # database driver.
        result = self._store._connection.execute(statement, self._rows)

        if returning:
            return result.get_all()
