from storm.exceptions import CompileError
from storm.info import get_cls_info

from canonical.storm.compile import compile_columns
from canonical.storm.insert import BulkInsert


class RelationLink(object):
    """Placeholder for row values that will be resolved at flush time.

    Instances of this class can be used as placeholders for row values
    that are not known at the time the row is inserted or updated by the
    application code, typically because the values are actually foreign keys
    referencing surrogate keys (e.g. serial IDs) whose actual value is
    assigned by the database (e.g. with a sequence) and not known yet.
    """

    __slots__ = ("relation", "key", "column", "value_callback", "_column_index")

    def __init__(self, relation, key=None, column=None, value_callback=None):
        """
        @param relation: The foreign Relation this link is referencing.
        @param key: If passed, then the relation should have an index, and
            this placeholder will be resolved with the ID matching the
            provided key.
        @param column: If passed, then the placeholder will be resolved
            using value of this column from the given relation.
        @param value_callback: Optional callback which accepts the value
            obtained via the relation key or column lookup and returns a
            modified value.
        """
        self.relation = relation
        self.key = key
        self.column = column
        self.value_callback = value_callback
        self._check()

    def _check(self):
        """Check that the RelationLink points to a valid relation."""

        # If the link has a key, make sure that the foreign relation has a
        # matching index.
        if self.key:
            key = self.key
            if not isinstance(key, tuple):
                key = (key,)
            if not self.relation.index:
                raise CompileError(
                    "Can't create a keyed link to relation without " "index"
                )
            if len(key) != len(self.relation.index):
                raise CompileError("Link key and relation index don't match")

        # If the link points to a column in the foreign relation, makes sure
        # that the column exists.
        if self.column:
            for i, column in enumerate(self.relation.columns):
                if column is self.column:
                    break
            else:
                column_names = compile_columns(self.relation.columns)
                raise CompileError(
                    "Link column '{}' doesn't match any relation column. "
                    "Valid columns are: {}".format(self.column.name, column_names)
                )
            self._column_index = i


class Relation(object):
    """A database table relation into which new rows can be inserted."""

    def __init__(self, columns, index=None):
        """
        @param columns: A list of Storm column objects defining the table
            columns this Relation will deal with. They must all belong to the
            same Storm model class.
        @param index: A single column or a tuple of (optional) relations and
            columns defining an index for this relation. Relation components
            of the index must come first.
        """
        # This will raise an error if columns don't all belong to the
        # same model.
        compile_columns(columns)
        self.columns = columns
        self._primary = None
        self._set_index(index)
        self._rows = []
        self._dirty = False  # Indicate whether the relation needs flushing

        self._resolved_primary = None  # Track resolved primary keys
        self._resolved_index = None  # Track resolved relation indexes
        self._resolved_rows = None  # Track resolved relation rows

    def link(self, key=None, column=None, value_callback=None):
        """Create a RelationLink object pointing to a row of this relation.

        @param key: Optional index key identifying the foreign relation row
            the link will resolve to.
        @param column: Optional column identifying the row value of foreign
            relation row that this link will resolve to. The default is to
            resolve to the primary key.
        @param value_callback: Optional callback which accepts the value
            obtained via the relation key or column lookup and returns a
            modified value.
        """
        return RelationLink(self, key=key, column=column, value_callback=value_callback)

    def get(self, key):
        """Return the row to be inserted, with the given key.

        Will raise an error if no index has been defined for this relation.
        """
        if not self.index:
            raise CompileError("Can't get rows by key without an index")
        i = self._index.get(key)
        if i is None:
            return
        return self._rows[i]

    def set(self, key, column, value):
        """Update the row to be inserted with the given column/value.

        Will raise an error if no index has been defined for this relation.
        """
        if not self.index:
            raise CompileError("Can't get rows by key without an index")
        i = self._index.get(key)
        if i is None:
            raise CompileError("No row matching key '{}'".format(key))
        row = list(self._rows[i])
        for j in range(len(self.columns)):
            if self.columns[j] is column:
                row[j] = value
                self._rows[i] = tuple(row)
                break
        else:
            raise CompileError("No matching column found")

    def add(self, **kwargs):
        """Add a new row to the relation.

        @param kwargs: A dict mapping column names to the values the new row
            should have.
        """
        values = kwargs.copy()
        row = ()

        # Make sure there are no unknown column names
        for column in self.columns:
            row += (values.pop(column.name, None),)
        if values:
            column_names = ", ".join(sorted(values.keys()))
            raise CompileError("Unknown columns: {}".format(column_names))

        # Make sure all index values are set
        index_relations = list(self._index_relations)
        key = [None] * len(self.index or [])
        for i, value in enumerate(row):
            if isinstance(value, RelationLink):
                link = value
                link.relation._ensure_primary()

                # If the link participates to the index, check that it has a
                # key.
                if link.relation in index_relations:
                    if not link.key:
                        raise CompileError("Foreign index without foreign key")
                    key_item = index_relations.index(link.relation)
                    key[key_item] = link.key
                    index_relations.remove(link.relation)
            else:
                if i in self._index_columns:
                    if value is None:
                        raise CompileError("Index value can't be None")
                    key_item = len(self._index_relations) + self._index_columns.index(i)
                    key[key_item] = value

        # Check that all relations participating to the index were assigned
        # some link value.
        if index_relations:
            raise CompileError("Index value is missing foreign keys")

        self._rows.append(row)

        # Update the index on the rows to be inserted.
        if self.index:
            key = tuple(key)
            if len(key) == 1:
                key = key[0]
            self._index[key] = len(self._rows) - 1

        self._dirty = True

    def _set_index(self, index):
        """Set the given index on this Relation object.

        This method will parse the index and separate the foreign relations
        from the columns components.
        """
        self.index = index
        self._index_relations = []
        self._index_columns = []
        self._index = {}

        if self.index is None:
            return

        self._ensure_primary()

        if not isinstance(self.index, tuple):
            self.index = (self.index,)

        # Break up the index into foreign relations and own columns, and make
        # sure foreign relations come first.
        for item in self.index:

            for i, column in enumerate(self.columns):
                # We need to perform the comparison using 'is' and not '==',
                # since the latter would return an Eq expression, not a bool
                # value indicating whether the two objects are the same. This
                # also means that we can't compact this loop with something
                # like 'if column in self._columns[name]'.
                if isinstance(item, Relation):
                    relation = item
                    if self._index_columns:
                        raise CompileError("Foreign relations must come first")
                    if not relation._index_columns:
                        raise CompileError("Foreign relation has no index")
                    self._index_relations.append(relation)
                    break
                else:
                    if item is column:
                        self._index_columns.append(i)
                        break
            else:
                raise CompileError("Index column doesn't match any relation column")

    def _ensure_primary(self):
        """Ensure that our _primary attribute is set.

        The presence of the _primary attribute will make the flushing code use
        RETURNING when inserting the rows of this relation. This is necessary
        when the relation is used as RelationLink by another relation.
        """
        if self._primary:
            return
        info = get_cls_info(self.columns[0].cls)
        if len(info.primary_key) > 1:
            raise CompileError("Compound primary keys are not supported")
        self._primary = info.primary_key


class RelationResolver(object):
    """Flush Relation rows resolving RelationLink placeholders.

    This class will perform actual inserts for Relation rows and will use the
    RETURNING clause to get back the primary keys needed to resolve placeholder
    row values which were set with RelationLinks.
    """

    def __init__(self):
        self._relations = []
        self._store_by_relation = {}  # Map Relation objects to their stores

    def get(self, model_class):
        """Return the relation(s) for the given Storm model.

        @return: A single Relation object, if there's only one Relation with
            columns belonging to the given model object, or a list of Relation
            objects if there are more than one. If there's no match, None is
            returned.
        """
        matching = [
            relation
            for relation in self._relations
            if relation.columns[0].cls is model_class
        ]

        if len(matching) == 0:
            return None
        elif len(matching) == 1:
            return matching[0]
        else:
            return matching

    def add(self, relation, store):
        """Add a new Relation to the resolver.

        @param relation: The Relation to add. If the rows of the Relation
            reference another Relation via RelationLink, the other Relation
            must be added first.
        @param store: The Store that the Relation's table lives in.
        """
        self._relations.append(relation)
        self._store_by_relation[relation] = store

    def flush(self):
        """Flush all rows of all relations, resolving references."""
        for relation in self._relations:
            if relation._dirty:
                self._flush_relation(relation)
                relation._dirty = False

    def _flush_relation(self, relation):
        """Flush the rows of a single relation.

        @param relation: The Relation to flush.
        """
        store = self._store_by_relation[relation]

        # Resolve all ReferenceLink instances and replace them with actual
        # values
        rows = self._resolve_rows(relation)

        # Actually insert the resolved rows into the database.
        insert = BulkInsert(store, relation.columns)
        insert.add(rows)
        result = insert.execute(returning=relation._primary)

        # Track primary keys if the relation is referenced by other relations.
        if relation._primary:
            relation._resolved_primary = [row[0] for row in result]

        # Build an index mapping relation index keys to the primary key.
        if relation.index:
            relation._resolved_index = self._resolve_index(relation)
            relation._resolved_rows = dict(zip(relation._resolved_primary, rows))

    def _resolve_rows(self, relation):
        """Resolve all ReferenceLink instances of the given relation."""
        rows = []
        for i, row in enumerate(relation._rows):
            rows.append(self._resolve_row(relation, i, row))
        return rows

    def _resolve_row(self, relation, i, row):
        """Resolve a single row of a relation.

        @param relation: The Relation the row belongs to.
        @param i: The index of the row being resolved (from relation._rows).
        @param row: The relation row being resolved.
        """
        resolved = []
        for value in row:

            # If the value is a RelationLink we need to resolve it.
            if isinstance(value, RelationLink):
                link = value
                if link.key is None:
                    # If there's no key, we resolve to the primary ID of the
                    # foreign relation row matching our row index.
                    value = link.relation._resolved_primary[i]
                else:
                    # If there's a key we resolve to the primary ID of the
                    # foreign relation row matching the key.
                    key = link.key
                    if not isinstance(key, tuple):
                        key = (key,)
                    # Get the primary key from the resolved index of the
                    # foreign relation.
                    value = link.relation._resolved_index[key]
                    # If a column is specified, resolve to the value of the
                    # given column in the linked row.
                    if link.column:
                        link_row = link.relation._resolved_rows[value]
                        value = link_row[link._column_index]
                if link.value_callback:
                    value = link.value_callback(value)

            resolved.append(value)

        return tuple(resolved)

    def _resolve_index(self, relation):
        """Resolve the index of the given relation.

        This method will create the given relation index mapping index keys
        to the actual primary keys fetched at INSERT time.
        """
        resolved_index = {}
        index_relations = relation._index_relations
        index_columns = relation._index_columns
        for i, row in enumerate(relation._rows):
            key = [None] * (len(index_relations) + len(index_columns))
            for item, value in enumerate(row):
                if item in index_columns:
                    # This component of the index is a plain value from the
                    # inserted row.
                    key_item = len(index_relations) + index_columns.index(item)
                    key[key_item] = value
                elif isinstance(value, RelationLink):
                    link = value
                    if link.relation not in index_relations:
                        # This linked relation doesn't partecipate in the index
                        continue
                    key_item = index_relations.index(link.relation)
                    key[key_item] = link.key

            key = tuple(key)
            if key in resolved_index:
                raise CompileError("Duplicate index key '{}'".format(key))
            resolved_index[key] = relation._resolved_primary[i]

        return resolved_index
