# Copyright 2005-2022 Canonical Limited. All rights reserved.

__metaclass__ = type

import logging
import os
import sys
from collections import OrderedDict

import talisker.logs

logging.setLoggerClass(talisker.logs.StructuredLogger)

from twisted.web.wsgi import WSGIResource

import talisker
import talisker.logs
import talisker.sentry
import talisker.wsgi

import werkzeug.local

from canonical.twisted.site import MaxMemorySite

logger = logging.getLogger("talisker.twisted")


class TaliskerWSGIResource(WSGIResource):
    """A WSGIResource that wraps the WSGI app with Talisker middleware.

    It also marks the twisted request object as having been handled by WSGI.
    This means it will not be logged by the TaliskerSite logging.
    """

    def __init__(self, reactor, threadpool, application):
        application = talisker.wsgi.wrap(application)
        WSGIResource.__init__(self, reactor, threadpool, application)

    def render(self, request):
        request._wsgi = True
        return WSGIResource.render(self, request)


class TaliskerSite(MaxMemorySite):
    """A Site that logs requests using Talisker's logging."""

    def log(self, request):
        if getattr(request, "_wsgi", False):
            # WSGI already logged
            return

        try:
            log_twisted_request(request)
        except Exception:
            logger.exception("failed to write access log")

        # this is a handy place to clear things
        talisker.clear_contexts()


def get_metadata(
    environ,
    status=None,
    headers=None,
    duration=None,
    length=None,
    exc_info=None,
    filepath=None,
):
    """Return an ordered dictionary of request metadata for logging."""
    if headers is None:
        headers = {}
    else:
        headers = dict((k.lower(), v) for k, v in headers)
    extra = OrderedDict()
    extra["method"] = environ.get("REQUEST_METHOD")
    script = environ.get("SCRIPT_NAME", "")
    path = environ.get("PATH_INFO", "").decode()
    extra["path"] = script + "/" + path.lstrip("/")
    qs = environ.get("QUERY_STRING")
    if qs:
        extra["qs"] = environ.get("QUERY_STRING")
    if status:
        extra["status"] = status
    if "VIEW_NAME" in environ:
        extra["view"] = environ["VIEW_NAME"]
    elif "x-view-name" in headers:
        extra["view"] = headers["x-view-name"]
    if duration:
        extra["duration_ms"] = round(duration * 1000, 3)
    extra["ip"] = environ.get("REMOTE_ADDR", None)
    extra["proto"] = environ.get("SERVER_PROTOCOL")
    if length:
        extra["length"] = length
    if filepath is not None:
        extra["filepath"] = filepath
    request_length = environ.get("CONTENT_LENGTH")
    if request_length:
        try:
            extra["request_length"] = int(request_length)
        except ValueError:
            pass
    content_type = environ.get("CONTENT_TYPE")
    if content_type:
        extra["request_type"] = content_type
    referrer = environ.get("HTTP_REFERER", None)
    if referrer is not None:
        extra["referrer"] = environ.get("HTTP_REFERER", None)
    if "HTTP_X_FORWARDED_FOR" in environ:
        extra["forwarded"] = environ["HTTP_X_FORWARDED_FOR"]
    if "HTTP_USER_AGENT" in environ:
        extra["ua"] = environ["HTTP_USER_AGENT"]

    if exc_info and exc_info[0]:
        extra["exc_type"] = str(exc_info[0].__name__)
        extra["trailer"] = "".join(traceback.format_exception(*exc_info))

    tracking = getattr(werkzeug.local.Local(), "request_tracking", {})
    for name, tracker in tracking.items():
        extra[name + "_count"] = tracker.count
        extra[name + "_time_ms"] = tracker.time

    msg = "{method} {path}{0}".format("?" if "qs" in extra else "", **extra)
    return msg, extra


def log_twisted_request(request):
    """This logs twisted.web requests Talisker-style."""
    # mimic a WSGI environ, since we know how to extra metadata from them
    uri, _, qs = request.uri.partition(b"?")

    if hasattr(request, "getClientAddress"):
        # support newer twisted
        ip = request.getClientAddress().host
    else:
        ip = request.getClientIP()

    environ = {
        "REQUEST_METHOD": request.method.decode(),
        "REMOTE_ADDR": ip,
        "PATH_INFO": uri,
        "QUERY_STRING": qs,
        "CONTENT_TYPE": request.getHeader(b"content-type") or "",
        "CONTENT_LENGTH": request.getHeader(b"content-length") or "",
        "SERVER_NAME": request.getRequestHostname(),
        "SERVER_PORT": str(request.getHost().port),
        "SERVER_PROTOCOL": request.clientproto,
    }

    for name, values in request.requestHeaders.getAllRawHeaders():
        name = (b"HTTP_" + name.upper().replace(b"-", b"_")).decode()
        environ[name] = (b",".join(values).replace(b"\n", b" ")).decode()

    msg, extra = get_metadata(
        environ,
        request.code,
        list(request.responseHeaders.getAllRawHeaders()),
    )
    logger.info(msg, extra=extra)


def initialise(env=os.environ):
    """Initialise Talisker for landscape.

    Configures logging, and logs Talisker configuration.

    Note: we can not use talisker.initialise() as usual, as landscape's
    expectations are different.
    """
    logging.setLoggerClass(talisker.logs.StructuredLogger)
    config = talisker.get_config(env)

    logger = logging.getLogger(__name__)
    config_extra = {k: v.value for k, v in list(config.metadata().items()) if v.raw}
    if config_extra:
        logger.info("talisker configured", extra=config_extra)
    if config.ERRORS:
        errors = {name: str(err) for name, err in list(config.ERRORS.items())}
        logger.error("configuration errors", extra=errors)

    talisker.clear_contexts()
    return config


def configure_logging(name, level, stream):
    """Configure Talisker logging for landscape.

    This hardcodes the logger name in the log format, which is what landscape
    expects. It can also point logs at a specific output stream, which is used
    in testing.
    """
    config = talisker.get_config()

    if config.colour:
        formatter = talisker.logs.ColouredFormatter(style=config.colour)
        formatter._fmt = formatter._fmt.replace("%(name)s", name)
    else:
        named_format = talisker.logs.StructuredFormatter.FORMAT.replace(
            "%(name)s", name
        )
        formatter = talisker.logs.StructuredFormatter(named_format)

    handler = talisker.logs.get_talisker_handler()
    handler.stream = stream
    talisker.logs.add_talisker_handler(level, handler, formatter)
