# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os

from zope.component import provideUtility, getUtility, queryUtility
from zope.configuration.interfaces import IConfigurationContext
from zope.configuration.config import ConfigurationMachine
from zope.configuration.xmlconfig import (
    registerCommonDirectives,
    file as xml_file,
    string as xml_string,
)

from canonical.testing.fixtures import ComponentsFixture

from testresources import TestResourceManager

from .fixtures import Fixture


class ConfigurationContextResourceManager(TestResourceManager):
    """Setup the ZCML configuration context."""

    def make(self, dependencies):
        context = queryUtility(IConfigurationContext)
        if not context:
            context = ConfigurationMachine()
            registerCommonDirectives(context)
            provideUtility(context)

        return context


class ZcmlResourceManager(TestResourceManager):
    """Load a file with ZCML directives."""

    _seen = set()  # Class attribute, common to all instances

    def make(self, dependencies):
        if self._zcml not in self._seen:
            path = _get_path(self._zcml)
            context = getUtility(IConfigurationContext)
            xml_file(path, context=context)
            self._seen.add(self._zcml)


class CoreZcmlResourceManager(ZcmlResourceManager):

    _zcml = "ftesting-core.zcml"


class SystemZcmlResourceManager(ZcmlResourceManager):

    _zcml = "ftesting-system.zcml"


class PublicationZcmlResourceManager(ZcmlResourceManager):

    _zcml = "ftesting-publication.zcml"


class ZcmlComponentsFixture(Fixture):
    """Load a ZCML config file in an overlay Zope registry."""

    def __init__(self, zcml):
        super(ZcmlComponentsFixture, self).__init__()
        self._zcml = zcml

    def setUp(self):
        super(ZcmlComponentsFixture, self).setUp()
        self.useFixture(ComponentsFixture())
        context = getUtility(IConfigurationContext)
        xml_file(_get_path(self._zcml), context=context)


def zcml(s, packages=(), context=None):
    """
    Execute a ZCML string after loading 'meta.zcml' for specified packages.
    """
    for package in packages:
        context = xml_file("meta.zcml", context=context, package=package)
    return xml_string(s, context=context)


def zcml_extension(features=()):
    """
    Return a configuration context that can be used to extend an existing
    context, for example for having zcml directives register components
    against an overridden registry.
    """
    context = ConfigurationMachine()
    registerCommonDirectives(context)
    # The zope.configuration package doesn't offer any API to reuse,
    # clone or extend existing ConfigurationMachine objects, so in
    # order to re-use the directives already registered before, we
    # need to copy them in our brand new machine.
    context._registry.update(getUtility(IConfigurationContext)._registry)
    for feature in features:
        context.provideFeature(feature)
    return context


def _get_path(zcml):
    # XXX checking LANDSCAPE_CONFIG is probably not needed and should
    # be arguably done differently, however we're keeping it like this
    # in case one wants to use this test resource with another config
    name = os.environ.get("LANDSCAPE_CONFIG", "default").split("-", 1)[0]
    root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    return os.path.join(root, "configs", name, zcml)
