# Copyright 2005-2022 Canonical Limited. All rights reserved.

import signal
from io import BytesIO

from zope.interface import implementer

from twisted.trial.unittest import TestCase
from twisted.web.iweb import IRequest
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.python import log


class FakeThreadPool(object):
    """
    A fake L{twisted.python.threadpool.ThreadPool}, running function inside the
    main thread instead for easing tests.
    """

    def callInThreadWithCallback(self, onResult, func, *args, **kw):
        success = True
        try:
            result = func(*args, **kw)
        except:
            result = Failure()
            success = False

        onResult(success, result)


class TwistedTestCase(TestCase):
    def setUp(self):
        if log.defaultObserver is not None:
            try:
                log.defaultObserver.stop()
            except ValueError:
                pass

    def tearDown(self):
        if log.defaultObserver is not None:
            log.defaultObserver.start()
        TestCase.tearDown(self)
        # Trial should restore the handler itself, but doesn't.
        # See bug #3888 in Twisted tracker.
        signal.signal(signal.SIGINT, signal.default_int_handler)


@implementer(IRequest)
class FakeRequest(object):
    """
    Fake HTTP request object

    @param method: HTTP method.
    @param path: The request path.
    @param host: The hostname (and optionally port)?
    @param content: A string with request content or a file-like object to read
      content from.
    @param args: The request arguments.
    """

    def __init__(
        self,
        method="GET",
        path="",
        host="example.com",
        content=b"",
        args=None,
        headers=None,
    ):
        self.method = method
        self.path = path
        self.host = host
        self.args = args
        if isinstance(content, bytes):
            self.content = BytesIO(content)
        elif isinstance(content, str):
            self.content = BytesIO(content.encode())
        else:
            self.content = content
        self.written = BytesIO()
        self.finished = False
        self.code = None
        self.headers = {"Host": self.host}
        if headers:
            self.headers.update(headers)
        self.notifications = []
        self.postpath = path.split("/")[1:]
        self.prepath = []

    def write(self, content):
        assert type(content) is bytes, "Only bytes should be written"
        self.written.write(content)

    def finish(self):
        if self.code is None:
            self.code = 200
        self.finished = True
        for deferred in self.notifications:
            deferred.callback(None)

    def connectionLost(self, reason):
        for d in self.notifications:
            d.errback(reason)

    def setResponseCode(self, code):
        self.code = code

    def notifyFinish(self):
        deferred = Deferred()
        self.notifications.append(deferred)
        return deferred

    def setHeader(self, key, value):
        self.headers[key] = value

    def getHeader(self, key):
        return self.headers.get(key)

    @property
    def response(self):
        return self.written.getvalue()
