# Copyright 2005-2014 Canonical Limited.  All rights reserved.

"""Workarounds and improvements for Twisted HTTP facilities."""

from wsgiref.handlers import format_date_time
from time import time


def _escape(s):
    """Copy of C{twisted.web.http._escape}.

    Since this is a private function in Twisted 14.0 we copy it over to avoid
    future breakages in case Twisted removes it.

    Return a string like python repr, but always escaped as if surrounding
    quotes were double quotes.

    @param s: The string to escape.
    @type s: L{bytes} or L{unicode}

    @return: An escaped string.
    @rtype: L{unicode}
    """
    if not isinstance(s, bytes):
        s = s.encode("ascii")

    r = repr(s)
    if not isinstance(r, str):
        r = r.decode("ascii")
    if r.startswith("b"):
        r = r[1:]
    if r.startswith("'"):
        return r[1:-1].replace('"', '\\"').replace("\\'", "'")
    return r[1:-1]


def combinedLogFormatterWithUser(timestamp, request):
    """Custom version of C{twisted.web.http.combinedLogFormatter}.

    Custom version of the stock http log formatter of Twisted 14.0,
    adding support for logging the HTTP request user name (as found
    in the Authentication header).

    This code could be dropped if Twisted eventually supports logging the HTTP
    request user name (https://twistedmatrix.com/trac/ticket/772).

    See also:

    http://httpd.apache.org/docs/1.3/logs.html

    for more information about the combined log format.
    """
    user = _escape(request.getUser() or b"-")
    referrer = _escape(request.getHeader(b"referer") or b"-")
    agent = _escape(request.getHeader(b"user-agent") or b"-")
    line = (
        "%(ip)s - %(user)s %(timestamp)s "
        '"%(method)s %(uri)s %(protocol)s" '
        '%(code)d %(length)s "%(referrer)s" "%(agent)s"\n'
        % dict(
            ip=_escape(request.getClientIP() or b"-"),
            user=user,
            timestamp=timestamp,
            method=_escape(request.method),
            uri=_escape(request.uri),
            protocol=_escape(request.clientproto),
            code=request.code,
            length=request.sentLength or "-",
            referrer=referrer,
            agent=agent,
        )
    )
    return line


def set_caching_headers(request, days=1, now=time):
    """Set caching headers on the given request.

    @param request: The twisted.web.http.Request object to set the response
        headers on.
    @param days: Number of days in the future to set the expiration at.
    @param now: The function to use to obtain the current time.
    """
    delta = 60 * 60 * 24 * days
    request.setHeader("expires", format_date_time(now() + delta))
    cache_control = "max-age=%s, public" % (60 * 60 * 24 * days)
    request.setHeader("cache-control", cache_control)
