# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Workarounds and improvements for C{twisted.web} facilities."""

from urllib.parse import urlparse

from html import escape
from urllib.parse import quote as urlquote

from OpenSSL import SSL

from twisted.internet import reactor, ssl
from twisted.web.server import NOT_DONE_YET
from twisted.web.proxy import ReverseProxyResource, ProxyClient, ProxyClientFactory


class ProxyClientWithFinish(ProxyClient):
    """
    Customize C{ProxyClient} to workaround bug #4480 in Twisted.
    """

    def __init__(self, command, rest, version, headers, data, father):
        ProxyClient.__init__(self, command, rest, version, headers, data, father)
        self.father.notifyFinish().addBoth(self._fatherFinished)

    def _fatherFinished(self, reason):
        """
        If the main connection died, close the connection to the child, and set
        the finished flag so that we don't try to close several times.
        """
        self._finished = True
        self.transport.loseConnection()

    def connectionLost(self, reason):
        if reason.check(SSL.Error):
            # Validating the SSL context has failed. Return an
            # appropriate error.
            self.father.setResponseCode(501, "Gateway error")
            self.father.responseHeaders.setRawHeaders("Content-Type", ["text/html"])
            self.father.write("<H1>Could not connect</H1>")
            self.father.write("<P>%s</P>" % escape(str(reason.value)))
        return ProxyClient.connectionLost(self, reason)


class ProxyClientFactoryWithFinish(ProxyClientFactory):
    """
    A custom C{ProxyClientFactory} using L{ProxyClientWithFinish}.
    """

    protocol = ProxyClientWithFinish


class ReverseProxyResourceWithFinish(ReverseProxyResource):
    """
    A custom C{ReverseProxyResource} using L{ProxyClientFactoryWithFinish}.
    """

    proxyClientFactoryClass = ProxyClientFactoryWithFinish

    def getChild(self, path, request):
        """
        Override C{getChild} to return an instance of the same class instead of
        C{ReverseProxyResource} instance, thus keeping the correct value for
        C{proxyClientFactoryClass}.
        """
        self_path = self.path.encode() if isinstance(self.path, str) else self.path

        return ReverseProxyResourceWithFinish(
            self.host,
            self.port,
            self_path + b"/" + urlquote(path, safe="").encode(),
            self.reactor,
        )


class VerifyingContextFactory(ssl.CertificateOptions):
    def __init__(self, verify, caCerts, verifyCallback=None):
        ssl.CertificateOptions.__init__(self, verify=verify, caCerts=caCerts)
        self.verifyCallback = verifyCallback

    def _makeContext(self):
        context = ssl.CertificateOptions._makeContext(self)
        if self.verifyCallback is not None:
            context.set_verify(
                SSL.VERIFY_PEER | SSL.VERIFY_FAIL_IF_NO_PEER_CERT, self.verifyCallback
            )
        return context


class SecureReverseProxyResource(ReverseProxyResourceWithFinish):
    def __init__(
        self,
        host,
        port,
        path,
        auth,
        verify=False,
        caCerts=None,
        verifyCallback=None,
        reactor=reactor,
    ):
        ReverseProxyResourceWithFinish.__init__(self, host, port, path, reactor)
        self.auth = auth
        self.verify = verify
        self.caCerts = caCerts
        self.verifyCallback = verifyCallback

    def getChild(self, path, request):
        """
        Override C{getChild} to return an instance of the same class instead of
        C{ReverseProxyResource} instance, thus keeping the correct value for
        C{proxyClientFactoryClass}.
        """
        return SecureReverseProxyResource(
            self.host,
            self.port,
            self.path + "/" + urlquote(path, safe=""),
            self.auth,
            self.verify,
            self.caCerts,
            self.verifyCallback,
            self.reactor,
        )

    def render(self, request):
        """
        Render a request by forwarding it to the proxied server.
        """
        # This sets the user in the Request for subsequent logging
        request.getUser()
        # RFC 2616 tells us that we can omit the port if it's the default port,
        # but we have to provide it otherwise
        if self.port == 443:
            host = self.host
        else:
            host = "%s:%d" % (self.host, self.port)
        request.requestHeaders.setRawHeaders("host", [host])
        request.requestHeaders.setRawHeaders(
            "authorization", ["Basic %s" % self.auth.encode("base64")[:-1]]
        )
        request.content.seek(0, 0)
        qs = urlparse(request.uri)[4]
        if qs:
            rest = self.path + "?" + qs
        else:
            rest = self.path

        clientFactory = self.proxyClientFactoryClass(
            request.method,
            rest,
            request.clientproto,
            request.getAllHeaders(),
            request.content.read(),
            request,
        )
        sslContext = VerifyingContextFactory(
            self.verify, self.caCerts, self.verifyCallback
        )
        self.reactor.connectSSL(self.host, self.port, clientFactory, sslContext)
        return NOT_DONE_YET
