# Copyright 2005-2014 Canonical Limited.  All rights reserved.

from functools import partial
import logging

from twisted.internet import reactor
from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet.task import deferLater


class Retryer(object):
    """An object for retrying calls using a single deferred.

    Usage:

        retry = Retryer(retry_delay=10, max_retries=30)
        deferred = retry(func, arg, kwarg=value)

    If the func() call fails, it will be retried until the maximum
    numbers of retries is reach, when the deferred's errback will fire
    with the last error.

    @param retry_delay: The number of seconds to wait between each retry.
    @param max_retries: The number of times to retry calling the function.
    """

    def __init__(self, retry_delay=30, max_retries=120, reactor=reactor):
        self.retry_delay = retry_delay
        self.max_retries = max_retries
        self.reactor = reactor

    def __call__(self, func, *args, **kwargs):
        """Call the specified function, retrying it if it fails.

        The deferred fires its callback with the result from the
        function to call when it finishes successfully. If it doesn't
        succeeds before the maximum numbers of retries is reach, the
        deferred's errback fires with the last error.
        """
        # If keyword argument 'retryer_ignore_errors' is passed,
        # consider it a function to decide if a particular failure
        # should be ignored and turned into a success.
        self._ignore_errors_func = kwargs.pop("retryer_ignore_errors", None)

        call = RetryCall(func, *args, **kwargs)
        self._schedule_call(call)
        return call.deferred

    def _schedule_call(self, call, delay=0):
        """Schedule the function call be to run."""
        if delay == 0:
            retry_call = maybeDeferred(call.func)
        else:
            retry_call = deferLater(self.reactor, delay, call.func)
        retry_call.addCallback(call.deferred.callback)
        retry_call.addErrback(self._handle_error, call)

    def _handle_error(self, failure, call):
        """Handle errors calling the function.

        If the number of retries have been reached, fire the errback. If
        not, schedule a new call.
        """
        call.failures += 1

        if self.max_retries is not None and call.failures >= self.max_retries:
            logging.error("RetryingCall for '%s' failed last retry." % call.name)
            call.deferred.errback(failure)
        else:
            if self._ignore_errors_func and self._ignore_errors_func(failure):
                logging.info(
                    "Call for '%s' failed with '%s' but ignored."
                    % (call.name, failure.getErrorMessage())
                )
                call.deferred.callback("success")
            else:
                if self.max_retries is not None:
                    logging.info(
                        "RetryingCall for '%s' failed, "
                        "trying %d more time(s):"
                        % (call.name, self.max_retries - call.failures)
                    )
                else:
                    logging.info(
                        "RetryingCall for '%s' failed, will try again." % (call.name)
                    )
                logging.info(failure.getBriefTraceback())
                self._schedule_call(call, delay=self.retry_delay)


class RetryCall(object):
    """Hold info for a function call that must be retried."""

    def __init__(self, func, *args, **kwargs):
        self.name = func.__name__
        # Partially-apply the function, to avoid carrying around arguments
        self.func = partial(func, *args, **kwargs)
        self.deferred = Deferred()
        self.failures = 0
