import logging
import os
import sys

from twisted.web.server import Site
from twisted.internet import reactor

PROC_FILE = "/proc/self/status"

logger = logging.getLogger("canonical.twisted.site")


class MaxMemorySite(Site):
    """
    A Site that restarts the service when it hits its configurable
    `max_requests` number of served requests.
    """

    def __init__(self, resource, *, max_memory_mib: int = 0, **kwargs):
        super().__init__(resource, **kwargs)

        self._served_requests = 0
        self._request_check_rate = 500  # Check every 500 requests.

        self._max_memory_mib = max_memory_mib  # If 0, there is no limit.

    def getResourceFor(self, request):
        """
        We hook in here, as this method is guaranteed to be called once
        for every request.
        """
        if self._should_restart():
            self._restart_process()

        return super().getResourceFor(request)

    def _should_restart(self):
        """
        A restart is required if we have served
        `self._request_check_rate` requests and
        `ru_maxrss` > `self._max_memory_mib`.
        """
        if not self._max_memory_mib:
            return False

        self._served_requests += 1
        logger.debug(f"Requests: {self._served_requests}")

        if self._served_requests < self._request_check_rate:
            return False

        self._served_requests = 0

        with open(PROC_FILE) as proc_status:
            vm_rss_kib = 0
            for line in proc_status:
                if "VmRSS:" in line:
                    vm_rss_kib = int(line.split()[1])

        logging.debug(f"Current process vm RSS: {vm_rss_kib} kiB")
        vm_rss_mib = vm_rss_kib // 1024

        return vm_rss_mib >= self._max_memory_mib

    def _restart_process(self):
        """
        Restart the current process at the next opportunity, using a
        deferred so we don't interrupt the current execution.
        """
        logger.info("Restarting due to max memory rule")

        def restart_process():
            for arg in sys.argv:
                if "pidfile" in arg:
                    pidfile = arg.split("=")[1]
                    os.remove(pidfile)
                    break

            os.execv(sys.argv[0], sys.argv)

        reactor.callLater(2, restart_process)
