# Copyright 2005-2022 Canonical Limited. All rights reserved.

from io import StringIO

from twisted.internet.address import IPv4Address
from twisted.internet.defer import succeed
from twisted.web.client import ResponseDone
from twisted.web.http_headers import Headers
from twisted.python.failure import Failure


class FakeRequest(object):

    clientproto = "HTTP/1.1"
    code = 200
    sentLength = 50
    user = "testing"

    def __init__(
        self, method="GET", uri="/url/", args=None, headers=None, cookies=None
    ):
        try:
            uri = uri.encode()
        except AttributeError:
            pass
        self._closed = False
        self.method = method
        self.uri = uri
        self.postpath = uri.split(b"/")[1:]
        self.prepath = []
        self.args = args
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.response_headers = {}
        self.redirected_url = None
        self.body = StringIO()

    def getClientIP(self):
        return "1.1.1.1"

    def getClientAddress(self):
        return IPv4Address("TCP", self.getClientIP(), 9999)

    def getUser(self):
        return self.user

    def getHeader(self, header):
        return self.headers.get(header)

    def setHeader(self, key, value):
        self.response_headers[key] = value

    def getCookie(self, key):
        return self.cookies.get(key)

    def setResponseCode(self, code):
        self.code = code

    def redirect(self, url):
        self.redirected_url = url

    def write(self, _bytes):
        if self._closed:
            raise RuntimeError("Unable to write to a closed request")
        return self.body.write(_bytes)

    def finish(self):
        self._closed = True


class FakeHTTPResponse(object):
    """
    A very simple fake for twisted.web's HTTPResponse objects. It basically
    only does as much as treq requires to extract the content.
    """

    def __init__(self, content, code, content_type="application/json"):
        try:
            content = content.encode()
        except AttributeError:
            pass
        self.content = content
        self.code = code
        self.length = len(content)
        self.phrase = "OK"
        self.headers = Headers()
        if content_type is not None:
            self.headers.addRawHeader("content-type", content_type)

    def deliverBody(self, protocol):
        protocol.dataReceived(self.content)
        protocol.connectionLost(Failure(ResponseDone()))


class Requester(object):
    """
    A fake for a single invocation of L{treq.request}.
    """

    def __init__(
        self,
        response_content,
        response_code=200,
        response_headers=None,
        content_type="application/json",
    ):
        self.response_content = response_content.encode()
        self.response_code = response_code
        self.content_type = content_type
        if response_headers is None:
            response_headers = {}
        self.response_headers = response_headers

    def fake_request(
        self,
        method,
        url,
        headers=None,
        data=None,
        files=None,
        timeout=None,
        persistent=True,
    ):
        """
        Record the arguments that would be passed to treq.request.
        """
        self.method = method
        self.url = url
        self.headers = headers
        self.data = data
        self.files = files
        self.timeout = timeout
        self.persistent = persistent
        response = FakeHTTPResponse(
            self.response_content, self.response_code, self.content_type
        )
        for k, v in list(self.response_headers.items()):
            response.headers.setRawHeaders(k, [v])
        return succeed(response)


class SequenceRequester(object):
    def __init__(self, responses):
        self.responses = responses
        self.responses_iter = iter(responses)
        self.requests = []

    def fake_request(self, method, url, headers, data, persistent=True):
        self.requests.append(
            {
                "method": method,
                "url": url,
                "headers": headers,
                "data": data,
                "persistent": persistent,
            }
        )
        return succeed(next(self.responses_iter))
