# Copyright (C) 2009 Robert Collins <robertc@robertcollins.net>
# Licensed under the txaws license available at /LICENSE in the txaws source.

import configparser
import os

import attr

from .util import hmac_sha1, hmac_sha256


ENV_ACCESS_KEY = "ACCESS_KEY_ID"
ENV_PROFILE = "TXAPI_PROFILE"
ENV_SECRET_KEY = "SECRET_ACCESS_KEY"
ENV_SHARED_CREDENTIALS_FILE = "SHARED_CREDENTIALS_FILE"


class CredentialsNotFoundError(Exception):
    """
    txapi failed to locate access credentials, and none were explicitly given.
    """


@attr.s(init=False)
class Credentials:
    """Represents an access key and secret access key pair for authenticating
    with an API.

    @param access_key: The access key to use. If None, the env var
        ACCESS_KEY_ID is used.
    @param secret_key: The secret key to use. If None, the env var
        SECRET_ACCESS_KEY is used.
    @param environ: The environment. If unspecified, L{os.environ} is used.
    @raise CredentialsNotFoundError: No access key or secret key was provided,
        nor could they be found in the environment or filesystem.
    """

    access_key = attr.ib()
    secret_key = attr.ib(repr=False)

    def __init__(self, access_key="", secret_key="", environ=os.environ):
        if not access_key:
            access_key = environ.get(ENV_ACCESS_KEY)
            if not access_key:
                access_key, _ = self._load_shared_credentials(environ=environ)
        if not secret_key:
            secret_key = environ.get(ENV_SECRET_KEY)
            if not secret_key:
                _, secret_key = self._load_shared_credentials(environ=environ)

        self.access_key = access_key
        self.secret_key = secret_key

    @staticmethod
    def _load_shared_credentials(environ, profile=None):
        if profile is None:
            profile = environ.get(ENV_PROFILE, "default")

        credential_path = environ.get(
            ENV_SHARED_CREDENTIALS_FILE, os.path.expanduser("~/.txapi/credentials")
        )
        config = configparser.SafeConfigParser()

        if not config.read([credential_path]):
            raise CredentialsNotFoundError(
                "Could not find credentials in the environment or filesystem"
            )

        if not config.has_section(profile):
            raise CredentialsNotFoundError("No such profile {!r}".format(profile))

        try:
            return (
                config.get(profile, "access_key_id"),
                config.get(profile, "secret_access_key"),
            )
        except configparser.NoOptionError as error:
            raise CredentialsNotFoundError(
                "Profile {0.section!r} has no {0.option!r}".format(error)
            )

    def sign(self, bytestring, hash_type="HmacSHA256"):
        """Uses these credentials to sign `bytestring`."""
        if hash_type == "HmacSHA256":
            return hmac_sha256(self.secret_key, bytestring).decode()

        if hash_type == "HmacSHA1":
            return hmac_sha1(self.secret_key, bytestring).decode()

        raise RuntimeError("Unsupported hash type: '{}'".format(hash_type))
