from uuid import uuid4

from .exception import APIError


class Call:
    """Holds info about a single API call initiated by an HTTP request.

    @param raw_params: The raw parameters for the action to be executed, a
        dict mapping parameter names to parameter values.
    @param principal: The principal issuing this API L{Call}
    @param action: The action to be performed.

    @ivar id: A unique identifier for the API call.
    @ivar principal: The principal performing the call.
    @ivar args: An L{Arguments} instance holding parameters extracted from the
        raw parameters according to the L{Schema}. Becames available after
        calling C{parse}.
    @ivar rest: Extra parameters not included in the schema.
    @ivar version: The version of the API call. Defaults to "2012-08-15".
    """

    def __init__(
        self,
        raw_params=None,
        principal=None,
        action=None,
        version="2012-08-15",
        id=str(uuid4()),
    ):
        self.id = id
        self.action = action
        self.version = version
        self.principal = principal

        self._raw_params = {}
        if raw_params is not None:
            self._raw_params.update(raw_params)

    def parse(self, schema, strict=True):
        """Updates C{args} and C{rest}, parsing the raw request arguments.

        @param schema: The L{Schema} the parameters must be extracted with.
        @param strict: If C{True}, an error is raised if parameters not in the
            schema are found. Otherwise, these are saved in C{rest}.
        """
        self.args, self.rest = schema.extract(self._raw_params)

        if strict and self.rest:
            raise APIError(
                400,
                "UnknownParameter",
                "The parameter {} is not recognized".format(list(self.rest.keys())[0]),
            )

    def get_raw_params(self):
        """Return a C{dict} holding the raw API call parameters.

        The format of the dictionary is C({"ParamName": param_value}).
        """
        return self._raw_params.copy()
