from itertools import product

from venusian import attach


def method(method_class):
    """Class decorator for an API method.

    When invoking L{Registry.scan} the classes decorated with this will be
    added to the registry.

    @param cls: The L{Method} class to register.
    """

    def callback(scanner, name, method_class):
        if method_class.actions is not None:
            actions = method_class.actions
        else:
            actions = [name]

        if method_class.versions is not None:
            versions = method_class.versions
        else:
            versions = [None]

        for action, version in product(actions, versions):
            scanner.registry.add(method_class, action=action, version=version)

    attach(method_class, callback, category="method")

    return method_class


class Method:
    """Handler of a single HTTP request to an API resource.

    @cvar actions: List of actions that the Method can handle, if C{None} the
        class name will be used as the only supported action.
    @cvar versions: List of versions that the Method can handle, if C{None} all
        versions will be supported.
    """

    actions = None
    versions = None

    def invoke(self, call):
        """Invokes this method for executing the given C{call}."""
        raise NotImplementedError("Sub-classes have to implement the invoke method.")

    def is_available(self):
        """Returns True if this method is available.

        Override this to dynamically decide whether specific methods are
        available.
        """
        return True
