from venusian import Scanner

from .exception import APIError


class Registry:
    """Register API L{Method}s to handle specific actions and versions."""

    def __init__(self):
        self._by_action = {}

    def add(self, klass, action, version=None):
        """Add a class to the registry.

        @param klass: The class to add
        @param action: The action that the class can handle
        @param version: The version that the class can handle
        """
        by_version = self._by_action.setdefault(action, {})

        if version in by_version:
            raise RuntimeError(
                "A method was already registered for action {} "
                "in version {}".format(action, version)
            )

        by_version[version] = klass

    def check(self, action, version=None):
        """Check if the given action is supported in the given version.

        @raises APIError: If there's no method class registered for handling
                          the given action or version.
        """
        by_version = self._by_action.get(action)

        if not by_version:
            raise APIError(
                400,
                "InvalidAction",
                "The action {} is not valid " "for this web service.".format(action),
            )

        if None not in by_version:
            # There's no catch-all method. Let's try the version-specific one.
            if version not in by_version:
                raise APIError(400, "InvalidVersion", "Invalid API version.")

    def get(self, action, version=None):
        """Get the method class, handing the given action and version."""
        by_version = self._by_action[action]

        return by_version.get(version, by_version.get(None))

    def scan(self, module, onerror=None, ignore=None, decorator="method"):
        """Scans `module` for objects with `decorator` and registers them."""
        scanner = Scanner(registry=self)
        kwargs = {"onerror": onerror, "categories": [decorator]}

        if ignore is not None:
            # Only pass it if specified, for backward compatibility.
            kwargs["ignore"] = ignore

        scanner.scan(module, **kwargs)

    def get_actions(self):
        """Get a list of all action names."""
        return list(self._by_action.keys())

    def get_versions(self, action_name):
        """Get a list of versions supported by the given action name."""
        return list(self._by_action[action_name].keys())
