# Copyright (C) 2009 Duncan McGreggor <duncan@canonical.com>
# Copyright (C) 2009 Robert Collins <robertc@robertcollins.net>
# Licenced under the txaws licence available at /LICENSE in the txaws source.

import warnings

from .util import parse


class ServiceEndpoint:
    """
    @param uri: The URI of the service.
    @param method: The HTTP method used when accessing the service.
    @param ssl_hostname_verification: If True, SSL hostname verification is
        performed when connecting to the endpoint.
    """

    def __init__(self, uri="", method="GET", ssl_hostname_verification=True):
        self.host = ""
        self.port = None
        self.path = "/"
        self.method = method
        self.ssl_hostname_verification = ssl_hostname_verification

        if not self.ssl_hostname_verification:
            warnings.warn(
                "Operating with certificate verification disabled!", stacklevel=2
            )

        self._parse_uri(uri)

        if not self.scheme:
            self.scheme = "http"

    def get_canonical_host(self):
        """Returns the canonical host as per the HTTP header spec."""
        host = self.host.lower()

        if self.port is not None:
            host = "{}:{}".format(host, self.port)

        return host

    def set_canonical_host(self, canonical_host):
        """
        Sets host and port from a canonical host string per the Host HTTP
        header spec.
        """
        parts = canonical_host.lower().split(":")
        self.host = parts[0]

        if len(parts) > 1 and parts[1]:
            self.port = int(parts[1])
        else:
            self.port = None

    def get_uri(self):
        """Get a URI representation of the service."""
        return "{}://{}{}".format(self.scheme, self.get_canonical_host(), self.path)

    def _parse_uri(self, uri):
        scheme, host, port, path = parse(uri, defaultPort=False)

        self.scheme = scheme
        self.host = host
        self.port = port
        self.path = path
