#!/bin/sh

ROOTDIR=/opt/canonical
export PYTHONPATH
export STORM_CEXTENSIONS=1
export LANDSCAPE_CONFIG=standalone
export LANG=C
export TZ=UTC
PATH=/usr/sbin:/usr/bin:/sbin:/bin:$PATH
PYTHON=/usr/bin/python3
LSROOT=${ROOTDIR}/landscape
LSUSER=landscape
DESC="Landscape $LANDSCAPE_CONFIG"
DEFAULTS_FILE=/etc/default/landscape-server
ENVIRONMENT_FILE=/etc/environment

if [ -f "$ENVIRONMENT_FILE" ] ; then
    . $ENVIRONMENT_FILE
fi
export http_proxy https_proxy no_proxy
export HTTP_PROXY HTTPS_PROXY NO_PROXY

# Parse syslog address from Landscape service configuration.
# Defaults to local syslog.
SERVICE_CONF_FILE=/etc/landscape/service.conf
if [ -r $SERVICE_CONF_FILE ]; then
    SYSLOG_ADDRESS=$(awk -F '[\t ]*=[\t ]*' '/^syslog-address/ { print $2; };' $SERVICE_CONF_FILE)
else
    SYSLOG_ADDRESS=""
fi

PIDDIR=/var/run/landscape
LOCKDIR=/var/lock
MAINTENANCE_FILE=/opt/canonical/landscape/maintenance.txt
background=""

get_logger_arguments() {
    if echo $SYSLOG_ADDRESS | grep -q ':'; then
        host=`echo $SYSLOG_ADDRESS | cut -d: -f1`
        port=`echo $SYSLOG_ADDRESS | cut -d: -f2`
        echo "--server $host --port $port"
    else
        if [ -n "$SYSLOG_ADDRESS" ] && [ "$SYSLOG_ADDRESS" != "/dev/log" ]
        then
            echo "--socket $SYSLOG_ADDRESS"
        else
            echo ""
        fi
    fi
}

# Log to syslog with a particular tag using the INFO priority.
#  $1: tag to log with (eg. a script name).
#  $2...: message to log.
log_to_syslog() {
    tag=$1
    shift
    args=`get_logger_arguments`
    logger $args -p info -t $tag -- "$@"
}

# Logs to both syslog and stdout.
#  $1: Script/logger name: log tag to log with.
#  $2: Message to log.
log_to_syslog_and_stdout() {
    tag=$1
    shift
    echo "$@"
    log_to_syslog "$tag" "$@"
}

# Pipe command output to syslog and stderr with a particular tag using
# the ERROR priority.
#  $1: tag to log with (eg. a script name).
pipe_to_syslog() {
    tag=$1
    args=`get_logger_arguments`
    logger $args -s -p user.error -t $tag
}

# useful for cron style scripts to mark start and stop in log
#  $1: Script/logger name: log tag to log with.
#  $2: Message ("Message" at Mon Aug 29 19:15:03 UTC 2011)
mark_log() {
    log_to_syslog "$1" "--- $2 at" $(date -u) "---"
}

# useful for cron style scripts to alert admin on stdout
#  $1: Script name that alert is for.
alert_admin() {
    echo "Error running $0: $?"
    echo "Check out the syslog output for script $1."
}

# Include landscape-server defaults if available
if [ -f "$DEFAULTS_FILE" ] ; then
    . $DEFAULTS_FILE
fi

. /lib/lsb/init-functions

schema_check() {
    local schema_tool="$LSROOT/landscape-schema-check"
    [ "$VERBOSE" != no ] && log_daemon_msg "Checking database schema"
    # see http://mywiki.wooledge.org/BashPitfalls#local_varname.3D.24.28command.29
    local output
    output=$($schema_tool 2>&1)
    local result=$?
    log_end_msg $result
    if [ "$result" -ne "0" ]; then
        # 2nd line says "shutting down", which we are
        # not really doing
        echo "$output" | grep ERROR | head -n 1
        case "$result" in
            1) echo "Please upgrade the schema" ;;
            2) echo "Error verifying schema" ;;
            *) echo "Unknown exit code from landscape-schema-check $result" ;;
        esac
    fi
    return $result
}

check_privsep_dir() {
    # Create the PrivSep empty dir if necessary
    if [ ! -d $PIDDIR ]; then
        mkdir $PIDDIR
        chmod 0755 $PIDDIR
        chown landscape:landscape $PIDDIR
    fi
}

get_distributed_lock() {
    local command="$ROOTDIR/landscape/get-distributed-lock"

    $command "$@"
}

# Acquire the distributed lock for the script provided as argument. This
# prevents multiple computers from running a script at a given time.
#
# The script name can be either the full path of the script or any unique
# string used to prevent shared access.
acquire_lock() {
    get_distributed_lock $1
}

# Release the distributed lock for the script provided as argument.
#
# The script name can be either the full path of the script or any unique
# string used to prevent shared access.
release_lock() {
    get_distributed_lock $1 --release
}

# Fetch the proxy configuration from the database and set it on the env.
load_proxy_config() {
    eval `$ROOTDIR/landscape/get-proxy-config`
}

# Exit with success if cron scripts should not be run on the machine.
exit_if_not_running_cron() {
    if [ "$RUN_ALL" = "no" -a "$RUN_CRON" = "no" ]; then
        exit 0
    fi
}

# $1: the name of the script to run
# $2: optional lifetime limit for the lock file
#     defaults to 1.5h (5400s)
run_script()
{
    local script_name=$1

    exit_if_not_running_cron
    acquire_lock $script_name || exit 1

    if [ "$?" -eq 0 ]; then
        mark_log "$script_name" "Start"
        cd ${ROOTDIR}/landscape && ./$script_name
        if [ "$?" -ne 0 ]; then
            alert_admin $script_name
        fi
        mark_log "$script_name" "End"
    fi

    release_lock $script_name
}

# $1: instance id, a number between 1 and 9
#     if ommitted, 1 will be used
do_start()
{
    local instance=1
    [ "$1" ] && instance="$1"
    local tries=10
    local pidfile=$PIDDIR/landscape-$NAME-$instance.pid

    LANDSCAPE_CONFIG="$LANDSCAPE_CONFIG-$instance" \
        start-stop-daemon --start --quiet --chuid $LSUSER \
        --exec $LSROOT/$SCRIPT $background -- \
        --rundir=$LSROOT --pidfile=$pidfile \
        || return 2

    # in this case, start-stop-daemon will not check for successful start of
    # a process.  So, we should do that ourselves.
    if [ -n "$background" ]; then
        local start_loop=0
        while /bin/true; do
            start_loop=`expr $start_loop + 1`
            if [ $start_loop -ge $tries ]; then
                return 2
            fi
            if [ -r $pidfile ]; then
                status_of_proc -p \
                    $pidfile $NAME "$NAME $instance" > /dev/null 2>&1
                if [ $? -eq 0 ]; then
                    return
                fi
            fi
            sleep 1
        done
    fi
}

# $1: instance id, a number between 1 and 9
#     if ommitted, 1 will be used
do_stop()
{
    local instance=1
    [ "$1" ] && instance="$1"
    local pidfile=$PIDDIR/landscape-$NAME-$instance.pid
    start-stop-daemon --stop --quiet --pidfile $pidfile \
       --retry=TERM/30/KILL/5 --user $LSUSER
    local retval=$?
    [ $retval = 2 ] && return 2
    # Many daemons don't delete their pidfiles when they exit.
    rm -f $pidfile
    return $retval
}

# $1: instance id, a number between 1 and 9
#     if ommitted, 1 will be used
do_reload()
{
    local instance=1
    [ "$1" ] && instance="$1"
    local pidfile=$PIDDIR/landscape-$NAME-$instance.pid

    if [ -f "$pidfile" ] ; then
        local pid=$(head -1 $pidfile)
        kill -HUP $pid
    else
        return 2
    fi
}

# $1: instance id, a number between 1 and 9
#     if ommitted, 1 will be used
do_status()
{
    local instance=1
    [ "$1" ] && instance="$1"
    local pidfile=$PIDDIR/landscape-$NAME-$instance.pid

    # status_of_proc($0) = akin to "ps -ef |grep", not really useful to us.
    # status_of_proc($1) = name of the daemon in the message
    status_of_proc -p $pidfile $NAME "$NAME $instance"
}

# $1: the command used, like start or stop.
# $2: the specific RUN value for the service, like "no", "0", or the number of
#     instances.
# $3  onwards: optionally, one of the following strings:
#     no-reload: to indicate that reload is not supported.
#     background: to indicate that the service doesn't daemonize itself and that
#                 the initscript should do it
run()
{
    local action="$1"
    local run="$2"
    shift 2
    # should we even run?
    if [ "$RUN_ALL" = "no" -o "$RUN_ALL" = "0" ] && \
       [ "$run" = "no" -o "$run" = "0" ]; then
        exit 0
    fi

    local server_count=1

    # "yes" means one
    if [ "$RUN_ALL" = "yes" -o "$run" = "yes" ]; then
        server_count=1
    fi

    # it could also have a number
    if echo "$RUN_ALL" | grep -qE "^[1-9]$"; then
        server_count="$RUN_ALL"
    fi

    # but if there is a number in the more specific
    # env var, we take that one
    if echo "$run" | grep -qE "^[1-9]$"; then
        server_count="$run"
    fi

    # does this service support multiple copies?
    if [ "$server_count" -gt 1 ] && \
       [ "$NAME" != "api" -a \
         "$NAME" != "appserver" -a \
         "$NAME" != "message-server" -a \
         "$NAME" != "pingserver" ]; then
        echo "Service $NAME does not support multiple copies, acting on only one." >&2
        echo "Please fix /etc/default/landscape-server to avoid this warning." >&2
        server_count=1
    fi

    local usage="Usage: $0 {start|stop|restart|force-reload|reload|status}"

    while [ -n "$1" ]; do
        case "$1" in
            no-reload)
                usage="Usage: $0 {start|stop|restart|force-reload|status}"
                if [ "$action" = "reload" ]; then
                    echo $usage >&2
                    exit 3
                fi
                shift
                ;;
            background)
                local background="-b"
                shift
                ;;
            *)
                echo "Unknown extra option \"$1\"" >&2
                exit 3
                ;;
        esac
    done

    check_privsep_dir

    case "$action" in
        start)
            schema_check || exit 0
            for i in `seq $server_count`; do
                # Do not start while in maintenance.
                if [ -e "$MAINTENANCE_FILE" ]; then
                    [ "$VERBOSE" != no ] && log_daemon_msg "Maintenance file exists.  Not starting."
                    exit 0
                fi
                # if it's package-search, and there is no [package-search] section
                # in service.conf, skip it
                if [ "$NAME" = "package-search" ]; then
                    if ! grep -qE "^\[package-search\]" /etc/landscape/service.conf; then
                        [ "$VERBOSE" != no ] && log_daemon_msg "Skipping $DESC" "$NAME" "not configured yet"
                        [ "$VERBOSE" != no ] && log_end_msg 0
                        break
                    fi
                fi
                [ "$VERBOSE" != no ] && log_daemon_msg "Starting $DESC" "$NAME" "$i"
                do_start $i
                case "$?" in
                    0|1) [ "$VERBOSE" != no ] && log_end_msg 0 ;;
                    2) [ "$VERBOSE" != no ] && log_end_msg 1 ;;
                esac
            done
            ;;

        stop)
            for i in `seq $server_count`; do
                [ "$VERBOSE" != no ] && log_daemon_msg "Stopping $DESC" "$NAME" "$i"
                do_stop $i
                case "$?" in
                    0|1) [ "$VERBOSE" != no ] && log_end_msg 0 ;;
                    2) [ "$VERBOSE" != no ] && log_end_msg 1 ;;
                esac
            done
            ;;

        reload)
            for i in `seq $server_count`; do
                [ "$VERBOSE" != no ] && log_daemon_msg "Reloading $DESC" "$NAME" "$i"
                do_reload $i
                case "$?" in
                    0) [ "$VERBOSE" != no ] && log_end_msg 0 ;;
                    2) ;; # Don't log anything if the pid file doesn't exist
                    *) [ "$VERBOSE" != no ] && log_end_msg 1 ;;
                esac
            done
            ;;

        restart|force-reload)
            schema_check || exec "$0" stop
            for i in `seq $server_count`; do
                [ "$VERBOSE" != no ] && log_daemon_msg "Restarting $DESC" "$NAME" "$i"
                do_stop $i
                case "$?" in
                  0|1)
                    if [ -e "$MAINTENANCE_FILE" ]; then
                        [ "$VERBOSE" != no ] && log_daemon_msg "Maintenance file exists.  Not restarting."
                        exit 0
                    fi
                    do_start $i
                    case "$?" in
                        0) [ "$VERBOSE" != no ] && log_end_msg 0 ;;
                        1) [ "$VERBOSE" != no ] && log_end_msg 1 ;; # Old process is still running
                        *) [ "$VERBOSE" != no ] && log_end_msg 1 ;; # Failed to start
                    esac
                    ;;
                  *)
                    # Failed to stop
                    [ "$VERBOSE" != no ] && log_end_msg 1
                    ;;
                esac
            done
            ;;
        status)
            local status_exit=0
            for i in `seq $server_count`; do
                do_status $i || status_exit=$?
            done
            exit $status_exit
            ;;


        *)
            echo $usage >&2
            exit 3
            ;;
    esac
}

# $1: the name of the logger to use
exit_if_maintenance_mode()
{
    if [ -f /opt/canonical/landscape/maintenance.txt ] ; then
        log_to_syslog_and_stdout "$1" "Maintenance mode detected, exiting."
        exit 0
    fi
}
