from functools import wraps
import glob
import os
import shutil
import unittest

from testtools.content import text_content

from canonical.landscape.constants import ROOT_DIR
from canonical.landscape.model.feature.flag import enable_feature
from canonical.landscape.testing.model import ModelTest

from tests.helpers.resources import ResourcedTest, get_root_url


ARTIFACTS_DIR = os.path.join(ROOT_DIR, "tests/artifacts")


class IntegrationTest(ResourcedTest, ModelTest):
    def setUp(self):
        super(IntegrationTest, self).setUp()
        # We want to retry database DELETE statements because we might be
        # racing with other transactions running in the service processes.
        self.addOnException(self._add_logs)
        zstorm_manager = self.get_zstorm_manager()
        zstorm_manager.clean_attempts = 3
        self.setUpDatabase()
        # We want to commit any database fixture data, so the services can
        # see it.
        self.commit()

    def setUpDatabase(self):
        """Hook for sub-classes wanting to setup database data.

        All the setup will be automatically committed.
        """
        pass

    def enable_feature_flags(self, *features):
        """Enable specified feature flags."""
        changed_features = []
        for feature in features:
            if enable_feature(feature):
                changed_features.append(feature)
        self.commit()

    def create_person(self):
        super(IntegrationTest, self).create_person()
        # Also set the person OpenId identity.
        self.person.identity = "%s/id/person" % get_root_url()

    def _add_logs(self, exc_info):
        """Save logs of the started services."""
        for log in glob.glob(os.path.join(self.working_dir, "*.log")):
            self._add_log(log, "landscape")

    def _add_log(self, log, group):
        detail = os.path.basename(log)
        artifact = self._get_artifact_path("_{}-{}".format(group, detail))
        shutil.copyfile(log, artifact)
        self.addDetail(detail, text_content(artifact))

    def _get_artifact_path(self, suffix):
        """Return the full path to an artifact, given the file suffix."""
        return os.path.join(ARTIFACTS_DIR, self.shortDescription() + suffix)


def skip_version(version):
    """Skip the decorated test method if the Juju version matches."""
    if version.endswith((".x", ".X")):
        version = version[:-1]

    def decorator(f):
        @wraps(f)
        def wrapper(self, version=version):
            if version.endswith("."):
                if not self.JUJU_VERSION.startswith(version):
                    return f(self)
                version += "x"
            elif not self.JUJU_VERSION == version:
                return f(self)
            unittest.skip("Juju version {} not supported".format(version))

        return wrapper

    return decorator
