from collections import namedtuple

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.expected_conditions import (
    invisibility_of_element_located,
)

from tests.helpers.pages.account.base import AccountBase

RECEIPTS_LIST = "#receipt-list"
RECEIPTS_TABLE = "#receipts"
RECEIPTS_ROWS = "#receipts tbody tr"
EDIT_BUTTON = ".button-link"
ADD_BUTTON = "button[type='submit']"

BILLING_ERROR_DIV = ".p-error__payment"
CREDIT_CARD_NAME_P = ".p-payment__card-name"
CREDIT_CARD_PAYMENT_DETAILS_DIV = ".p-payment__details"
CREDIT_CARD_PAYMENT_DETAILS_CARD_DIV = ".p-payment__details-card"
CREDIT_CARD_PAYMENT_DETAILS_EXPIRY_DIV = ".p-payment__details-exp"
CREDIT_CARD_PAYMENT_DETAILS_ADDRESS = ".p-payment__details-address"
CREDIT_CARD_REMOVE_LINK = ".p-payment__remove-link"


Receipt = namedtuple("Receipt", ["name", "usage", "amount"])

Card = namedtuple("Card", ["name", "last4", "brand", "expiry_date", "address"])


class Billing(AccountBase):
    """Account billing page."""

    _title = "Billing - Landscape"

    def receipts_list(self):
        """Return a list of Receipts.

        Each Receipt has name, usage and amount properties.
        """
        self.log("Listing receipts")
        self._wait_presence(RECEIPTS_LIST)
        rows = []
        try:
            self._one(RECEIPTS_TABLE)
        except NoSuchElementException:
            # In the no-receipts case
            pass
        else:
            for row in self._all(RECEIPTS_ROWS):
                cells = self._all("td", row)
                name = cells[0].text
                usage = cells[1].text
                amount = cells[2].text
                rows.append(Receipt(name, usage, amount))
        return rows

    def have_credit_card(self):
        self.log("Checking if we have a credit card")
        try:
            self._one(CREDIT_CARD_NAME_P)
        except NoSuchElementException:
            return False
        else:
            return True

    def get_credit_card(self):
        """Get the details of the credit card.

        Card has last4, name, brand, expiry_date and address properties.
        """
        self.log("Getting credit card")
        self._wait_visible(CREDIT_CARD_NAME_P)
        name = self._one(CREDIT_CARD_NAME_P).text
        self._wait_visible(CREDIT_CARD_PAYMENT_DETAILS_DIV)
        card_summary = self._one(CREDIT_CARD_PAYMENT_DETAILS_CARD_DIV)
        brand, last4 = card_summary.text.split(" - ")
        expiry_text = self._one(CREDIT_CARD_PAYMENT_DETAILS_EXPIRY_DIV).text
        expiry_date = expiry_text.split(": ", 1)[1]
        address_element = self._one(CREDIT_CARD_PAYMENT_DETAILS_ADDRESS)
        address = "\n".join(
            span.text for span in self._all("span", ancestor=address_element)
        )
        return Card(name, last4, brand, expiry_date, address)

    def edit_billing_details(self):
        self.log("Choose edit billing details")
        self._wait_visible(EDIT_BUTTON)
        self._click(EDIT_BUTTON)
        self.log("Navigating to the Edit page")
        return EditBilling(self._driver)

    def remove_credit_card(self):
        """Remove the credit card."""
        self._wait_visible(CREDIT_CARD_REMOVE_LINK)
        self._click(CREDIT_CARD_REMOVE_LINK)
        return BillingRemoveCard(self._driver)

    def add_billing_details(self):
        """
        Click the add payment method button. This should only be visible
        if the user doesn't have a card associated.
        """
        self.log("Choose to add billing details")
        self._wait_visible(ADD_BUTTON)
        self._one(ADD_BUTTON).click()
        self.log("Navigating to the Add page")
        return CreateBilling(self._driver)

    def get_billing_error(self):
        self.log("Getting billing error")
        self._wait_visible(BILLING_ERROR_DIV)
        div = self._one(BILLING_ERROR_DIV)
        return div.text


# This is the name of the class that react-modal gives the content
REMOVE_CARD_MODAL_DIALOG = ".ReactModal__Content"
REMOVE_CARD_MODAL_DIALOG_BUTTONS_DIV = REMOVE_CARD_MODAL_DIALOG + " .buttons"


class BillingRemoveCard(AccountBase):

    """Modal confirmation dialog for removing credit card."""

    _title = "Billing - Landscape"

    def _page_wait_ready(self):
        self._wait_presence(REMOVE_CARD_MODAL_DIALOG)

    def confirm_remove(self):
        """Confirm the card removal."""
        self._wait_visible(REMOVE_CARD_MODAL_DIALOG_BUTTONS_DIV)
        buttons_div = self._one(REMOVE_CARD_MODAL_DIALOG_BUTTONS_DIV)
        buttons = self._all("button", ancestor=buttons_div)
        cancel, remove = buttons
        assert remove.text == "Remove"
        remove.click()
        condition = invisibility_of_element_located(
            (By.CSS_SELECTOR, REMOVE_CARD_MODAL_DIALOG)
        )
        self._wait(condition)
        return Billing(self._driver)

    def cancel(self):
        """Cancel the card removal."""
        buttons_div = self._one(REMOVE_CARD_MODAL_DIALOG_BUTTONS_DIV)
        buttons = self._all("button", ancestor=buttons_div)
        cancel, remove = buttons
        assert cancel.text == "Cancel"
        cancel.click()
        condition = invisibility_of_element_located(
            (By.CSS_SELECTOR, REMOVE_CARD_MODAL_DIALOG)
        )
        self._wait(condition)
        return Billing(self._driver)


from tests.helpers.pages.account.edit_billing import EditBilling
from tests.helpers.pages.account.create_billing import CreateBilling
