from collections import namedtuple

from selenium.webdriver.common.action_chains import ActionChains
from selenium.common.exceptions import NoSuchElementException

from tests.helpers.pages.account.base import AccountBase

ACCOUNT_TYPE_LABEL = "label[class='p-payment__use-label']"

ACCOUNT_COUNTRY = "select[name='account-country']"
ACCOUNT_VAT = "input[name='VATNumber']"
ACCOUNT_NAME = "input[name='account-name']"
ACCOUNT_ADDRESS_1 = "input[name='account-address1']"
ACCOUNT_ADDRESS_2 = "input[name='account-address2']"
ACCOUNT_STATE = "input[name='account-state']"
ACCOUNT_CITY = "input[name='account-city']"
ACCOUNT_ZIP = "input[name='account-zip']"

BILLING_COUNTRY = "select[name='billing-country']"
BILLING_NAME = "input[name='billing-name']"
BILLING_ADDRESS_1 = "input[name='billing-address1']"
BILLING_ADDRESS_2 = "input[name='billing-address2']"
BILLING_STATE = "input[name='billing-state']"
BILLING_CITY = "input[name='billing-city']"
BILLING_ZIP = "input[name='billing-zip']"

Address = namedtuple(
    "Address",
    ["country", "name", "address_line_1", "address_line_2", "state", "city", "zipcode"],
)

CREDIT_CARD_NUMBER = "#card-number-label"
CREDIT_CARD_EXP = "#card-expiry-label"
CREDIT_CARD_CVC = "#card-cvc-label"

PAYMENT_ERROR = ".p-payment__details-error"

SUBMIT_BUTTON = "button[type='submit']"


class CreateBilling(AccountBase):
    """Account add new billing details page."""

    _title = "Add billing details - Landscape"

    def get_shipping_address(self):
        """Get the details of the shipping address."""
        self.log("Getting shipping address")
        self._wait_visible(ACCOUNT_COUNTRY)
        country = self._value_of(ACCOUNT_COUNTRY)
        name = self._value_of(ACCOUNT_NAME)
        address1 = self._value_of(ACCOUNT_ADDRESS_1)
        address2 = self._value_of(ACCOUNT_ADDRESS_2)
        state = self._value_of(ACCOUNT_STATE)
        city = self._value_of(ACCOUNT_CITY)
        zipcode = self._value_of(ACCOUNT_ZIP)
        return Address(country, name, address1, address2, state, city, zipcode)

    def get_vat_number(self):
        """Get the VAT number value, if it's present."""
        try:
            return self._value_of(ACCOUNT_VAT)
        except NoSuchElementException:
            return None

    def get_account_type(self):
        """Get the type of account (business or personal)"""
        for account_option in self._all(ACCOUNT_TYPE_LABEL):
            radio = self._one("input", ancestor=account_option)
            if radio.get_property("checked"):
                account_type = account_option.text
                self.log("Account type {} is selected".format(account_type))
                return account_type
        else:
            raise ValueError("All account types are unchecked?")

    def get_error(self):
        self._wait_visible(PAYMENT_ERROR)
        return self._one(PAYMENT_ERROR).text

    def fill_in_billing_address(self, address):
        self._fill(BILLING_NAME, address["name"])
        self._fill(BILLING_ADDRESS_1, address["line1"])
        self._fill(BILLING_STATE, address["state"])
        self._fill(BILLING_CITY, address["city"])
        self._fill(BILLING_ZIP, address["postal_code"])
        self._select(BILLING_COUNTRY, address["country"])

    def fill_in_credit_card_details(self, number, exp_month, exp_year, cvc):
        self.fill_in_frame(CREDIT_CARD_NUMBER, number)
        self.fill_in_frame(CREDIT_CARD_EXP, "{} / {}".format(exp_month, exp_year))
        self.fill_in_frame(CREDIT_CARD_CVC, cvc)

    def submit(self):
        self.log("Getting submit button")
        self._wait_visible(SUBMIT_BUTTON)
        self._click(SUBMIT_BUTTON)
        self._wait_redirect(Billing)
        return Billing(self._driver)

    def submit_with_error(self):
        """In case of error, submit doesn't redirect to the Billing page"""
        self.log("Getting submit button")
        self._wait_visible(SUBMIT_BUTTON)
        self._click(SUBMIT_BUTTON)

    def fill_in_frame(self, label_selector, value):
        """Fill a field in an iframe."""
        chain = ActionChains(self._driver)
        chain.click(self._driver.find_element_by_css_selector(label_selector))
        # validation is wonky if typing too fast
        for key in value:
            chain.send_keys(key)
        chain.perform()


from tests.helpers.pages.account.billing import Billing
