from collections import namedtuple

from selenium.common.exceptions import NoSuchElementException

from tests.helpers.pages.account.base import AccountBase

CREDIT_CARD_NAME_P = ".p-payment__card-name"
CREDIT_CARD_PAYMENT_DETAILS_DIV = ".p-payment__details"
CREDIT_CARD_PAYMENT_DETAILS_CARD_DIV = ".p-payment__details-card"
CREDIT_CARD_PAYMENT_DETAILS_EXPIRY_DIV = ".p-payment__details-exp"

Card = namedtuple("Card", ["name", "last4", "brand", "expiry_date"])

ACCOUNT_TYPE_LABEL = "label[class='p-payment__use-label']"

ACCOUNT_COUNTRY = "select[name='account-country']"
ACCOUNT_VAT = "input[name='VATNumber']"
ACCOUNT_NAME = "input[name='account-name']"
ACCOUNT_ADDRESS_1 = "input[name='account-address1']"
ACCOUNT_ADDRESS_2 = "input[name='account-address2']"
ACCOUNT_STATE = "input[name='account-state']"
ACCOUNT_CITY = "input[name='account-city']"
ACCOUNT_ZIP = "input[name='account-zip']"

BILLING_COUNTRY = "select[name='billing-country']"
BILLING_NAME = "input[name='billing-name']"
BILLING_ADDRESS_1 = "input[name='billing-address1']"
BILLING_ADDRESS_2 = "input[name='billing-address2']"
BILLING_STATE = "input[name='billing-state']"
BILLING_CITY = "input[name='billing-city']"
BILLING_ZIP = "input[name='billing-zip']"

Address = namedtuple(
    "Address",
    ["country", "name", "address_line_1", "address_line_2", "state", "city", "zipcode"],
)

SUBMIT_BUTTON = "button[type='submit']"


class EditBilling(AccountBase):
    """Account edit billing page."""

    _title = "Edit billing details - Landscape"

    def get_credit_card(self):
        """Get the details of the credit card.

        Card has last4, name, brand, expiry_date and address properties.
        """
        self.log("Getting credit card")
        self._wait_visible(CREDIT_CARD_NAME_P)
        name = self._one(CREDIT_CARD_NAME_P).text
        self._wait_visible(CREDIT_CARD_PAYMENT_DETAILS_DIV)
        card_summary = self._one(CREDIT_CARD_PAYMENT_DETAILS_CARD_DIV)
        brand, last4 = card_summary.text.split(" - ")
        expiry_text = self._one(CREDIT_CARD_PAYMENT_DETAILS_EXPIRY_DIV).text
        expiry_date = expiry_text.split(": ", 1)[1]
        return Card(name, last4, brand, expiry_date)

    def get_shipping_address(self):
        """Get the details of the shipping address."""
        self.log("Getting shipping address")
        self._wait_visible(ACCOUNT_COUNTRY)
        country = self._value_of(ACCOUNT_COUNTRY)
        name = self._value_of(ACCOUNT_NAME)
        address1 = self._value_of(ACCOUNT_ADDRESS_1)
        address2 = self._value_of(ACCOUNT_ADDRESS_2)
        state = self._value_of(ACCOUNT_STATE)
        city = self._value_of(ACCOUNT_CITY)
        zipcode = self._value_of(ACCOUNT_ZIP)
        return Address(country, name, address1, address2, state, city, zipcode)

    def get_billing_address(self):
        """Get the details of the billing address."""
        self.log("Getting billing address")
        self._wait_visible(BILLING_COUNTRY)
        country = self._value_of(BILLING_COUNTRY)
        name = self._value_of(BILLING_NAME)
        address1 = self._value_of(BILLING_ADDRESS_1)
        address2 = self._value_of(BILLING_ADDRESS_2)
        state = self._value_of(BILLING_STATE)
        city = self._value_of(BILLING_CITY)
        zipcode = self._value_of(BILLING_ZIP)
        return Address(country, name, address1, address2, state, city, zipcode)

    def get_vat_number(self):
        """Get the VAT number value, if it's present."""
        try:
            return self._value_of(ACCOUNT_VAT)
        except NoSuchElementException:
            return None

    def get_account_type(self):
        """Get the type of account (business or personal)"""
        for account_option in self._all(ACCOUNT_TYPE_LABEL):
            radio = self._one("input", ancestor=account_option)
            if radio.get_property("checked"):
                account_type = account_option.text
                self.log("Account type {} is selected".format(account_type))
                return account_type
        else:
            raise ValueError("All account types are unchecked?")

    def fill_in_billing(self, name, address, city, zipcode, country):
        self._fill(BILLING_NAME, name)
        self._fill(BILLING_ADDRESS_1, address)
        self._fill(BILLING_CITY, city)
        self._fill(BILLING_ZIP, zipcode)
        self._select(BILLING_COUNTRY, country)

    def submit(self):
        self.log("Getting submit button")
        self._wait_visible(SUBMIT_BUTTON)
        self._click(SUBMIT_BUTTON)
        self._wait_redirect(Billing)
        return Billing(self._driver)


from tests.helpers.pages.account.billing import Billing
