import time

from zope.cachedescriptors.property import Lazy

from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.expected_conditions import (
    title_contains,
    presence_of_element_located,
    visibility_of_element_located,
    element_to_be_clickable,
)
from selenium.webdriver.support.ui import WebDriverWait


TIMEOUT = 60  # Default timeout for waiting elements


class PageObject(object):
    """High-level API for exposing and driving the functionality of a page.

    @cvar logger: Optional logger object which is expected to have a
        single "log" method and be configured for logging. This object is a
        class-level instance, shared across all actual page object instances.
    """

    logger = None

    _title = None  # Text that MUST be in the title (defined by subclasses)

    def __init__(self, driver):
        """
        @param driver: An instance of a Selenium driver.
        """
        self.log("Switch to page {}".format(self.__class__.__name__))
        self._driver = driver
        assert self._title is not None, "Missing title for %s" % self.__class__
        assert self.title_matches(driver), "Title {} doesn't match {}".format(
            self.title, self._title
        )
        self._page_wait_yui()
        self._page_wait_ready()

    @classmethod
    def title_matches(cls, driver):
        """Wait for the page title to match the expected one."""
        wait = WebDriverWait(driver, TIMEOUT)
        condition = title_contains(cls._title)
        try:
            wait.until(condition)
        except TimeoutException:
            return False
        else:
            return True

    @property
    def title(self):
        """Expose page title to system tests."""
        return self._driver.title

    @Lazy
    def feature_flags(self):
        """Returns a dict mapping feature flag names to their status."""
        return self._get_js_var("LANDSCAPE_FEATURE_FLAGS")

    def log(self, message):
        """Forward the message to the configured logger, if any."""
        if self.logger is not None:
            self.logger.info(message)

    def _page_wait_yui(self):
        """Wait for YUI javascript to kick in.

        This is needed because many tests rely on YUI to be loaded and event
        listeners to be registered.
        """
        # XXX Wait for *some* YUI to be loaded, but since there's no generic
        # way to wait for *all* YUI code to be loaded, we also sleep a bit and
        # hope for the best.
        self._wait_presence("html.yui3-js-enabled")
        time.sleep(2)

    def _page_wait_ready(self):
        """Page-specific code that will wait for the page to be ready.

        Typically this is used by sub-classes when there's dynamic content
        being generated at page load time via AJAX calls, and the page won't
        be fully functional until such content is there.
        """

    # Conveniences around the Selenium driver

    def _one(self, selector, ancestor=None):
        """Find the first element matching the given selector.

        @param ancestor: Optional ancestor to limit the search tree.
        """
        root = ancestor or self._driver
        return root.find_element_by_css_selector(selector)

    def _all(self, selector, ancestor=None):
        """Find all elements matching the given selector."""
        root = ancestor if ancestor else self._driver
        return root.find_elements_by_css_selector(selector)

    def _fill(self, selector, text):
        """Fill with the given text the input matching the given selector."""
        field = self._one(selector)
        field.clear()
        field.send_keys(text)

    def _select(self, selector, value):
        """Select the option with the given value for the given selector"""
        option = self._one(selector + " option[value={}]".format(value))
        option.click()

    def _click(self, selector):
        """Click the element matching the given selector."""
        element = self._one(selector)
        # Scroll the window to show the element being clicked
        self._driver.execute_script("arguments[0].scrollIntoView(true);", element)
        self._wait(element_to_be_clickable((By.CSS_SELECTOR, selector)))
        element.click()

    def _click_link(self, text):
        """Click the hyperlink matching the given text."""
        element = self._driver.find_element_by_link_text(text)
        element.click()

    def _value_of(self, selector):
        """Get the value of the element matching the given selector."""
        element = self._one(selector)
        return element.get_property("value")

    def _wait(self, condition, timeout=TIMEOUT):
        """Wait for the given condition to realize."""
        wait = WebDriverWait(self._driver, timeout)
        return wait.until(condition)

    def _wait_presence(self, selector, timeout=TIMEOUT):
        """Wait for the presence of an element matching the given selector."""
        condition = presence_of_element_located((By.CSS_SELECTOR, selector))
        return self._wait(condition, timeout=timeout)

    def _wait_visible(self, selector, timeout=TIMEOUT):
        """Wait for the element matching the given selector to be visible."""
        condition = visibility_of_element_located((By.CSS_SELECTOR, selector))
        return self._wait(condition, timeout=timeout)

    def _wait_redirect(self, page_class, timeout=TIMEOUT):
        """Wait for an HTTP redirect to a new page and return a page instance.

        @param page_class : The class for the new page.
        """
        self._wait(page_class.title_matches, timeout=timeout)
        return page_class(self._driver)

    def _get_js_var(self, variable):
        """Return the value of a (global) javascript variable in the page."""
        return self._driver.execute_script("return {};".format(variable))
