import os
import os.path
import tempfile
import shutil
import subprocess
import signal
import sys
import time

from jinja2 import FileSystemLoader, Environment
from selenium import webdriver
from testresources import TestResourceManager
from testtools import TestCase
from zope.component import provideUtility

from canonical.lib.undef import Undef
from canonical.lib.config import Config
from canonical.amqp.message.dispatcher import global_dispatcher
from canonical.amqp.message.publisher import global_publisher
from canonical.amqp.message.interfaces import IDispatcher, IPublisher
from canonical.testing.runner import get_worker

from canonical.landscape.constants import ROOT_DIR
from canonical.landscape.setup import run_amqp_client_in_thread


JUJU_VERSION = None


class ServiceResourceManager(TestResourceManager):
    """Manages a single service process.

    This resource will take care of spawning a specific Landscape service.
    """

    _executable = Undef()  # Name of the service script to spawn
    _stdout_start_marker = Undef()  # Wait for this text to appear in the log

    def __init__(self, juju_version=None):
        super(ServiceResourceManager, self).__init__()
        self.juju_version = juju_version or JUJU_VERSION
        self._started = False

    def make(self, dependencies):
        if self._started:
            return self

        working_dir = ensure_working_dir(self.juju_version)
        stdout = self._get_stdout(working_dir, "w")
        command_line = self._get_command_line(working_dir)
        env = self._get_env(working_dir)
        process = subprocess.Popen(command_line, env=env, stdout=stdout)
        self._wait_start(process, self._get_stdout(working_dir, "r"))
        self._started = True
        return self

    def clean(self, resource):
        # We do nothing here.  Instead this is handled in ResourcedTest below.
        pass

    def _get_command_line(self, working_dir):
        """Return the command line arguments to use."""
        interpreter = sys.executable
        script = os.path.join(ROOT_DIR, self._executable)
        pidfile = os.path.join(working_dir, self._executable + ".pid")
        return [interpreter, script, "-n", "--pidfile=%s" % pidfile]

    def _get_stdout(self, working_dir, mode):
        """Return a file descriptor for the standard output."""
        logfile = os.path.join(working_dir, self._executable + ".log")
        return open(logfile, mode)

    def _get_env(self, working_dir):
        """Return the environment variables to use."""
        env = os.environ.copy()
        env["LANDSCAPE_CONFIG_DIR"] = _get_config_dir(working_dir)
        env["LANDSCAPE_ROOT_URL"] = get_root_url() + "/"
        return env

    def _wait_start(self, process, stdout, tries=50, wait=0.5):
        """Wait for the service to start, checking the start marker text."""
        returncode = None
        try:
            while True:
                returncode = process.poll()
                if returncode is not None:
                    break
                out = stdout.read()
                if self._stdout_start_marker in out:
                    # process is running.
                    break
                if not tries:
                    break
                time.sleep(wait)
                tries -= 1
        finally:
            stdout.seek(0)

        if returncode is not None:
            raise RuntimeError(
                "Failed to execute %s:\nError: (%s) %s"
                % (self._executable, returncode, stdout.read())
            )

        add_service(process.pid)


class AppserverResourceManager(ServiceResourceManager):
    """Manages the Landscape appserver process."""

    _executable = "landscape"
    _stdout_start_marker = "Notifications publisher connected to AMQP broker"

    def _get_env(self, working_dir):
        env = super(AppserverResourceManager, self)._get_env(working_dir)
        env["DEBUG_LOG"] = "1"
        return env


class AsyncFrontendResourceManager(ServiceResourceManager):
    """Manages the Landscape async-frontend process."""

    _executable = "async-frontend"
    _stdout_start_marker = "Async frontend connected to AMQP broker"


class JobHandlerResourceManager(ServiceResourceManager):
    """Manages the Landscape async-frontend process."""

    _executable = "job-handler"
    _stdout_start_marker = "Notifications publisher connected to AMQP broker"


class AMQPClientResourceManager(TestResourceManager):
    """Setup a working amqp client so that tests can send out notifications."""

    _stop_amqp_client = None

    def __init__(self, juju_version=None):
        super(AMQPClientResourceManager, self).__init__()
        self.juju_version = juju_version or JUJU_VERSION

    def make(self, dependencies):
        provideUtility(global_dispatcher, IDispatcher)
        provideUtility(global_publisher, IPublisher)
        if self._stop_amqp_client is None:
            _working_dir = _working_dirs[self.juju_version]
            service_conf = _get_service_conf(_working_dir, self.juju_version)
            config = Config()
            config.load(service_conf)
            self._stop_amqp_client = run_amqp_client_in_thread(config)
        return self

    def destroy(self):
        if self._stop_amqp_client:
            self._stop_amqp_client()


# XXX workaround the fact that it is not possible tear down resources after all
# tests are run, this would need cooperation from the test runner, so it's a
# limitation of the unittest module.
_integration_tests_count = 0


class ResourcedTest(TestCase):
    """A TestCase class that manages integration test resources.

    This class is meant for subclassing, mixing in with
    canonical.landscape.testing.model.ModelTest.  ModelTest makes use
    of the get_resources() method.  ResourcedTest should come before
    ModelTest in the list of base classes.  Note that the resources
    defined here will combined with the resources defined in ModelTest.
    The primary intended subclass is
    tests.helpers.integration.IntegrationTest.

    The most important thing to notice is that the resources here are
    not cleaned up after each test method, unlike normal resources.
    See ResourcedTest.run() for more info.

    """

    JUJU_VERSION = JUJU_VERSION

    # Set with_XXX to True in a subclass to activiate the XXX resource.
    RESOURCES = {
        "appserver": AppserverResourceManager,
        "async_frontend": AsyncFrontendResourceManager,
        "job_handler": JobHandlerResourceManager,
        "amqp_client": AMQPClientResourceManager,
    }

    _resource_cache = {}

    @classmethod
    def _resource_from_cache(cls, name):
        try:
            resource_cache = cls._resource_cache[cls.JUJU_VERSION]
        except KeyError:
            resource_cache = cls._resource_cache[cls.JUJU_VERSION] = {}
        try:
            return resource_cache[name]
        except KeyError:
            newres = cls.RESOURCES[name]
            resource_cache[name] = newres(cls.JUJU_VERSION)
            return resource_cache[name]

    def __init__(self, methodName="runTest"):
        super(ResourcedTest, self).__init__(methodName)
        global _integration_tests_count
        _integration_tests_count += 1

    def run(self, result=None):
        """Run normally and then clean up once all tests have run.

        Normally resources are cleaned up after each test method.
        However, for the integration tests we are going to share the
        resources across all the tests.  The run() method facilitates
        this by only cleaning up resources once *all* the tests have
        completed.

        """
        try:
            super(ResourcedTest, self).run(result=result)
        finally:
            global _integration_tests_count
            _integration_tests_count -= 1
            if _integration_tests_count == 0 or result.shouldStop:
                self._destroy_resources()

    def setUp(self):
        super(ResourcedTest, self).setUp()
        self.working_dir = ensure_working_dir(self.JUJU_VERSION)

    def get_resources(self):
        resources = super(ResourcedTest, self).get_resources()
        for name in self.RESOURCES:
            if getattr(self, "with_" + name, False):
                resource = self._resource_from_cache(name)
                resources.append((name, resource))
        return resources

    def open_artifact(self, filename):
        """Return an open file from the working dir."""
        return open(os.path.join(self.working_dir, filename))

    def read_artifact(self, filename):
        """Return the content of the file from the working dir."""
        with self.open_artifact(filename) as f:
            for line in f:
                yield line

    def _destroy_resources(self):
        """Destroy all long-running resources."""
        (self._resource_from_cache("amqp_client")).destroy()
        terminate_selenium_driver()
        terminate_services()
        cleanup_working_dir()


def get_root_url():
    """Return the root URL for the appserver."""
    return "http://localhost:%d" % _get_port()


def add_service(pid):
    """Add process ID to the list of running services."""
    _services.append(pid)


def terminate_services():
    """Terminate all running services."""
    for pid in _services:
        pid = int(pid)
        os.kill(pid, signal.SIGTERM)
        os.waitpid(pid, 0)


def ensure_working_dir(juju_version):
    """Ensure that the services working directory is ready.

    The working directory holds the configuration files and log files for all
    running services and the script wrapping the fake-juju binary.
    """
    try:
        _working_dir = _working_dirs[juju_version]
    except KeyError:
        _working_dir = _working_dirs[juju_version] = tempfile.mkdtemp(
            prefix="landscape-integration-test-"
        )
        config_dir = _get_config_dir(_working_dir)
        stock_config_dir = os.path.join(ROOT_DIR, "configs", "default")
        os.mkdir(config_dir)

        # Symlink all ZCML files
        for filename in os.listdir(stock_config_dir):
            if filename == "service.conf":
                continue
            source = os.path.join(stock_config_dir, filename)
            link_name = os.path.join(config_dir, filename)
            os.symlink(source, link_name)

        # Generate the service.conf
        with open(os.path.join(config_dir, "service.conf"), "w") as fd:
            fd.write(_get_service_conf(_working_dir, juju_version))

    return _working_dir


def cleanup_working_dir():
    """Remove the working directory."""
    for juju_version in list(_working_dirs):
        _working_dir = _working_dirs.pop(juju_version)
        shutil.rmtree(_working_dir)


def get_selenium_driver():
    """Return the Selenium Chrome driver for Chromium."""
    global _selenium_driver
    if _selenium_driver is None:
        options = webdriver.ChromeOptions()
        options.add_argument("--headless")
        options.add_argument("--disable-gpu")
        options.add_experimental_option("w3c", False)
        # Our CSS rules don't play well with small screens, so some
        # elements might get considered as non-displayed.
        options.add_argument("--window-size=1600,1200")
        _selenium_driver = webdriver.Chrome("chromedriver", chrome_options=options)
    return _selenium_driver


def terminate_selenium_driver():
    """Stops the Chromium process of the selenium driver."""
    global _selenium_driver
    if _selenium_driver is not None:
        _selenium_driver.quit()
        _selenium_driver = None


def _get_config_dir(working_dir):
    """Return the path to the config directory."""
    return os.path.join(working_dir, "config")


def _get_service_conf(working_dir, juju_version):
    """Return the content of the service.conf file to write."""
    templates_dir = os.path.dirname(__file__)
    loader = Environment(loader=FileSystemLoader(templates_dir))
    template = loader.get_template("service.conf")
    worker = get_worker() or 0
    context = {
        "worker": worker,
        "port": _get_port(worker),
        "juju": "fake-juju-{}".format(juju_version),
        "working_dir": working_dir,
    }
    return template.render(context)


def _get_port(worker=None):
    """Return the base port for the services."""
    if worker is None:
        worker = get_worker() or 0
    return 10000 + 100 * worker


_working_dirs = {}
_services = []
_selenium_driver = None
