# -*- coding: utf-8 -*-
import socket

from datetime import timedelta, date
from canonical.lib.timestamp import utctoday

from tests.helpers.ui import UITest
from tests.helpers.pages.account.billing import Billing, BillingRemoveCard

from canonical.landscape.model.configuration import set_root_url, set_stripe_api_keys
from canonical.landscape.model.billing.cost import set_utility_cost
from canonical.landscape.sample.sample import (
    SAMPLE_STRIPE_PUBLIC_KEY,
    SAMPLE_STRIPE_SECRET_KEY,
)
from canonical.landscape.stripe.customer import create_customer, get_customer
from canonical.landscape.stripe.payment import try_charge_payment
from canonical.landscape.model.billing.management import BillingManagement

from stripe import PaymentMethod
from selenium.common.exceptions import TimeoutException


STRIPE_CUSTOMER_ID = "cus_9Te3N4PCSHJoaz"


class BillingTest(UITest):
    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        self.account.stripe_customer_id = STRIPE_CUSTOMER_ID
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)

    def test_billing(self):
        """
        Navigating to the billing page returns an instance of the Billing
        page object without error.
        """
        dashboard = self.login()
        billing_page = dashboard.navigation_go_to_billing()
        self.assertIsInstance(billing_page, Billing)

    def test_empty_list_receipts(self):
        """Receipts can be listed."""
        page = self.login().navigation_go_to_billing()
        receipts = page.receipts_list()
        self.assertEqual([], receipts)

    def test_credit_card_details(self):
        """Credit card details."""
        page = self.login().navigation_go_to_billing()
        card = page.get_credit_card()
        self.assertEqual("Foo Bar", card.name)
        self.assertEqual("4242", card.last4)
        self.assertEqual("visa", card.brand)
        self.assertEqual("11/2018", card.expiry_date)
        self.assertEqual("Subačiaus. 17-25\nVilnius 01300\nLT", card.address)


class BillingWithReceiptsTest(UITest):
    def setUp(self):
        self.start_date = date(2016, 2, 1)
        self.end_date = date(2016, 2, 29)
        super(BillingWithReceiptsTest, self).setUp()

    def _add_payment(self):
        management = BillingManagement(self.account)
        payment = management.add_payment(self.start_date, self.end_date, 0, 0)
        payment.add_cost("physical", 2000, 1000, self.start_date, self.end_date)
        payment.add_cost("vm", 1000, 10000, self.start_date, self.end_date)
        payment.add_cost("container", 500, 1000, self.start_date, self.end_date)
        payment.stripe_charge_id = "ch_18vPV2FtAHAMnKQ7LepEGO8L"
        payment.status = "paid"
        payment.receipt_uri = "foo/bar"

    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        self.account.stripe_customer_id = STRIPE_CUSTOMER_ID
        set_utility_cost("physical", self.start_date, 100)
        set_utility_cost("vm", self.start_date, 10)
        set_utility_cost("container", self.start_date, 1)
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)
        self._add_payment()

    def test_list_receipts(self):
        """Receipts can be listed."""
        page = self.login().navigation_go_to_billing()
        [receipt] = page.receipts_list()
        self.assertEqual("February 2016", receipt.name)
        self.assertEqual("12,000", receipt.usage)
        self.assertEqual("$1,250.00", receipt.amount)


class FullFlowBillingTest(UITest):
    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)

        shipping_address = {
            "name": "New Customer",
            "line1": "AddressLine1",
            "line2": "AddressLine2",
            "city": "City",
            "state": "State",
            "postal_code": "1234",
            "country": "FR",
        }
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)
        # Create a new stripe customer without a card
        customer = create_customer(
            SAMPLE_STRIPE_SECRET_KEY, None, "test_customer", shipping_address
        )
        self.customer_id = customer.id
        self.account.stripe_customer_id = self.customer_id

    def test_billing_flow(self):
        page = self.login().navigation_go_to_billing()

        # Add a card to the new customer
        add_page = page.add_billing_details()
        add_page.fill_in_credit_card_details("4242424242424242", "12", "40", "123")
        billing_address = {
            "name": "New Customer",
            "line1": "AddressLine1",
            "line2": "AddressLine2",
            "city": "City",
            "state": "State",
            "postal_code": "1234",
            "country": "FR",
        }
        add_page.fill_in_billing_address(billing_address)
        page = add_page.submit()
        card = page.get_credit_card()
        self.assertEqual("New Customer", card.name)
        self.assertEqual("4242", card.last4)
        self.assertEqual("visa", card.brand)
        self.assertEqual("12/2040", card.expiry_date)
        self.assertEqual("AddressLine1\nCity 1234\nFR", card.address)

        # Remove the previously added card
        confirmation_page = page.remove_credit_card()
        self.assertIsInstance(confirmation_page, BillingRemoveCard)
        page = confirmation_page.confirm_remove()
        self.assertIsInstance(page, Billing)
        self.assertFalse(page.have_credit_card())

    def tearDown(self):
        super(FullFlowBillingTest, self).tearDown()
        customer = get_customer(SAMPLE_STRIPE_SECRET_KEY, self.customer_id)
        customer.delete()


class FailedChargeTest(UITest):
    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)
        shipping_address = {
            "name": "New Customer",
            "line1": "AddressLine1",
            "line2": "AddressLine2",
            "city": "City",
            "state": "State",
            "postal_code": "1234",
            "country": "FR",
        }
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)
        # Create a new stripe customer without a card
        customer = create_customer(
            SAMPLE_STRIPE_SECRET_KEY, None, "test_customer", shipping_address
        )
        self.account.stripe_customer_id = self.customer_id = customer.id

        # Special test card which does not fail attaching, but fails billing.
        # Other test cards fail too early to properly exercise billing failure.
        pm_card = "pm_card_chargeCustomerFail"
        method = PaymentMethod.attach(
            pm_card, customer=self.customer_id, api_key=SAMPLE_STRIPE_SECRET_KEY
        )
        customer = get_customer(SAMPLE_STRIPE_SECRET_KEY, self.customer_id)
        customer.invoice_settings.default_payment_method = method.id
        customer.save()

        management = BillingManagement(self.account)
        start_date = date(2020, 1, 1)
        end_date = date(2020, 2, 1)
        payment = management.add_payment(start_date, end_date, 0, 0)
        payment.add_cost("physical", 2000, 1000, start_date, end_date)
        try_charge_payment(SAMPLE_STRIPE_SECRET_KEY, payment)

    def tearDown(self):
        super(FailedChargeTest, self).tearDown()
        customer = get_customer(SAMPLE_STRIPE_SECRET_KEY, self.customer_id)
        customer.delete()

    def test_failed_charges(self):
        """
        Failed charge are counted and show in the UI.
        """
        page = self.login().navigation_go_to_billing()
        err = page.get_billing_error()
        expected = (
            "We encountered a billing error while processing your outstanding"
            " payment.\n"
            "Reason: Your card was declined.\n"
            "Due Date: 2020-02-01\n"
            "Amount Outstanding: $200.00\n"
            "Please correct your card or address details and we will attempt"
            " to re-run the payment in about an hour."
        )
        self.assertEqual(expected, err)


class ChargeWithManualConfirmation(UITest):

    PAYMENT_ERROR = (
        "An outstanding payment requires a validation from the"
        " card's owner.\n"
        "Validate this payment\n"
        "Once validated we will try to process this payment again in"
        " about an hour."
    )

    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        # confirmation requires a valid (non-localhost) hostname
        set_root_url("http://{}:8080/".format(socket.getfqdn()))
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)
        shipping_address = {
            "name": "New Customer",
            "line1": "AddressLine1",
            "line2": "AddressLine2",
            "city": "City",
            "state": "State",
            "postal_code": "1234",
            "country": "FR",
        }
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)
        # Create a new stripe customer without a card
        customer = create_customer(
            SAMPLE_STRIPE_SECRET_KEY, None, "test_customer", shipping_address
        )
        self.account.stripe_customer_id = self.customer_id = customer.id

        # Special test card which does not fail attaching, but fails billing.
        # Other test cards fail too early to properly exercise billing failure.
        pm_card = "pm_card_threeDSecure2Required"
        method = PaymentMethod.attach(
            pm_card, customer=self.customer_id, api_key=SAMPLE_STRIPE_SECRET_KEY
        )
        customer = get_customer(SAMPLE_STRIPE_SECRET_KEY, self.customer_id)
        customer.invoice_settings.default_payment_method = method.id
        customer.save()

        self.management = BillingManagement(self.account)
        start_date = date(2020, 1, 1)
        end_date = date(2020, 2, 1)
        self.payment = self.management.add_payment(start_date, end_date, 0, 0)
        self.payment.add_cost("physical", 2000, 1000, start_date, end_date)
        try_charge_payment(SAMPLE_STRIPE_SECRET_KEY, self.payment)

    def tearDown(self):
        super(ChargeWithManualConfirmation, self).tearDown()
        customer = get_customer(SAMPLE_STRIPE_SECRET_KEY, self.customer_id)
        customer.delete()

    def test_need_manual_confirm(self):
        """
        Manual charge confirmation link is displayed in ui when applicable.
        """
        page = self.login().navigation_go_to_billing()
        err = page.get_billing_error()
        self.assertEqual(self.PAYMENT_ERROR, err)

    def test_multiple_incomplete_payments(self):
        """
        When multiple incomplete payments are present, no exceptions should be
        raised (LP #1973248). When resolving, the next one should show up,
        until there's no more billing error.
        """

        # Add an extra payment that doesn't go through
        start_date = date(2020, 1, 1)
        end_date = date(2020, 2, 1)
        payment = self.management.add_payment(start_date, end_date, 0, 0)
        payment.add_cost("physical", 3000, 1000, start_date, end_date)
        try_charge_payment(SAMPLE_STRIPE_SECRET_KEY, payment)
        self.commit()  # This is called automatically in the setup function

        page = self.login().navigation_go_to_billing()
        err = page.get_billing_error()
        self.assertEqual(self.PAYMENT_ERROR, err)

        # Resolve the first one
        payment.mark_paid(payment.stripe_charge_id)
        self.commit()

        page.navigation_go_to_billing()
        err = page.get_billing_error()
        self.assertEqual(self.PAYMENT_ERROR, err)

        # Resolve second one
        self.payment.mark_paid(self.payment.stripe_charge_id)
        self.commit()

        # Make sure billing error is not present
        page.navigation_go_to_billing()
        self.assertRaises(TimeoutException, page.get_billing_error)
