from datetime import timedelta
from canonical.lib.timestamp import utctoday

from tests.helpers.ui import UITest
from tests.helpers.pages.account.create_billing import CreateBilling

from canonical.landscape.model.configuration import set_stripe_api_keys
from canonical.landscape.model.billing.cost import set_utility_cost
from canonical.landscape.stripe.customer import create_customer, get_customer
from canonical.landscape.sample.sample import (
    SAMPLE_STRIPE_PUBLIC_KEY,
    SAMPLE_STRIPE_SECRET_KEY,
)


# Customer with no card associated
STRIPE_CUSTOMER_ID = "cus_AN00tRe9WUWRLD"


class CreateBillingTest(UITest):
    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        self.account.stripe_customer_id = STRIPE_CUSTOMER_ID
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)

        # Setup the billing page, since we need it for navigation
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)

    def test_create_billing(self):
        """
        Navigating from the billing page returns an instance of the
        CreateBilling page object without error.
        """
        billing_page = self.login().navigation_go_to_billing()
        create_billing_page = billing_page.add_billing_details()
        self.assertIsInstance(create_billing_page, CreateBilling)

    def test_has_shipping_address(self):
        """
        Previous shipping address should be present on the page.
        """
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.add_billing_details()
        address = page.get_shipping_address()
        self.assertEqual("John Smith", address.name)
        self.assertEqual("GB", address.country)
        self.assertEqual("82 Shawdene Rd", address.address_line_1)
        self.assertEqual("", address.address_line_2)
        self.assertEqual("Manchester", address.state)
        self.assertEqual("Wythenshawe", address.city)
        self.assertEqual("M22 4AH", address.zipcode)

    def test_vat_for_business_account(self):
        """
        The page includes a VAT field if the user has
        a business account and is in the EU.
        """
        self.account.vat_number = "GB12345"
        self.account.is_business = True
        self.commit()
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.add_billing_details()
        vat = page.get_vat_number()
        self.assertEqual("GB12345", vat)

    def test_vat_for_personal_account(self):
        """
        The page doesn't include VAT if the user has
        a personal account.
        """
        self.account.is_business = False
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.add_billing_details()
        vat = page.get_vat_number()
        self.assertIsNone(vat)

    def test_account_type(self):
        """The page correctly displays the account type."""
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.add_billing_details()
        account_type = page.get_account_type()
        self.assertEqual("Personal use", account_type)


class CreateBillingStripeTest(UITest):
    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)

        shipping_address = {
            "name": "New Customer",
            "line1": "AddressLine1",
            "line2": "AddressLine2",
            "city": "City",
            "state": "State",
            "postal_code": "1234",
            "country": "FR",
        }
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)
        # Create a new stripe customer without a card
        customer = create_customer(
            SAMPLE_STRIPE_SECRET_KEY, None, "test_customer", shipping_address
        )
        self.customer_id = customer.id
        self.account.stripe_customer_id = self.customer_id

    def test_card_declined_error(self):
        """The page will display an error, if a declined card is submitted"""
        page = self.login().navigation_go_to_billing()

        add_page = page.add_billing_details()
        add_page.fill_in_credit_card_details("4000000000000002", "12", "40", "123")
        billing_address = {
            "name": "New Customer",
            "line1": "AddressLine1",
            "line2": "AddressLine2",
            "city": "City",
            "state": "State",
            "postal_code": "1234",
            "country": "FR",
        }
        add_page.fill_in_billing_address(billing_address)
        add_page.submit_with_error()
        error = add_page.get_error()
        self.assertEqual("Error: Your card has been declined.", error)

    def tearDown(self):
        super(CreateBillingStripeTest, self).tearDown()
        customer = get_customer(SAMPLE_STRIPE_SECRET_KEY, self.customer_id)
        customer.delete()
