from datetime import timedelta
from canonical.lib.timestamp import utctoday

from tests.helpers.ui import UITest
from tests.helpers.pages.account.edit_billing import EditBilling

from canonical.landscape.model.configuration import set_stripe_api_keys
from canonical.landscape.model.billing.cost import set_utility_cost
from canonical.landscape.sample.sample import (
    SAMPLE_STRIPE_PUBLIC_KEY,
    SAMPLE_STRIPE_SECRET_KEY,
)


STRIPE_CUSTOMER_ID = "cus_AHoUZzrrXJ7U79"


class EditBillingTest(UITest):
    def setUpDatabase(self):
        self.enable_feature_flags("self-service")
        self.account.stripe_customer_id = STRIPE_CUSTOMER_ID
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)

        # Setup the billing page, since we need it for navigation
        valid_since = utctoday() - timedelta(days=100)
        set_utility_cost("physical", valid_since, 100)
        set_utility_cost("vm", valid_since, 10)
        set_utility_cost("container", valid_since, 1)

    def test_edit_billing(self):
        """
        Navigating from the billing page returns an instance of the EditBilling
        page object without error.
        """
        billing_page = self.login().navigation_go_to_billing()
        edit_billing_page = billing_page.edit_billing_details()
        self.assertIsInstance(edit_billing_page, EditBilling)

    def test_credit_card_details(self):
        """
        Credit card details should be present on the page.
        """
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.edit_billing_details()
        card = page.get_credit_card()
        self.assertEqual("John Smith", card.name)
        self.assertEqual("4242", card.last4)
        self.assertEqual("visa", card.brand)
        self.assertEqual("12/2040", card.expiry_date)

    def test_has_billing_address(self):
        """
        Previous billing address should be present on the page.
        """
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.edit_billing_details()
        address = page.get_billing_address()
        self.assertEqual("John Smith", address.name)
        self.assertEqual("US", address.country)
        self.assertEqual("8785 Windfall St.", address.address_line_1)
        self.assertEqual("", address.address_line_2)
        self.assertEqual("PA", address.state)
        self.assertEqual("Whitehall", address.city)
        self.assertEqual("18052", address.zipcode)

    def test_has_shipping_address(self):
        """
        Previous shipping address should be present on the page.
        """
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.edit_billing_details()
        address = page.get_shipping_address()
        self.assertEqual("John Smith", address.name)
        self.assertEqual("GB", address.country)
        self.assertEqual("82 Shawdene Rd", address.address_line_1)
        self.assertEqual("", address.address_line_2)
        self.assertEqual("Manchester", address.state)
        self.assertEqual("Wythenshawe", address.city)
        self.assertEqual("M22 4AH", address.zipcode)

    def test_vat_for_business_account(self):
        """
        The page includes a VAT field if the user has
        a business account and is in the EU.
        """
        self.account.vat_number = "GB12345"
        self.account.is_business = True
        self.commit()
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.edit_billing_details()
        vat = page.get_vat_number()
        self.assertEqual("GB12345", vat)

    def test_vat_for_personal_account(self):
        """
        The page doesn't include VAT if the user has
        a personal account.
        """
        self.account.is_business = False
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.edit_billing_details()
        vat = page.get_vat_number()
        self.assertIsNone(vat)

    def test_account_type(self):
        """The page correctly displays the account type."""
        billing_page = self.login().navigation_go_to_billing()
        page = billing_page.edit_billing_details()
        account_type = page.get_account_type()
        self.assertEqual("Personal use", account_type)

    def test_submission(self):
        """
        The information on the page gets saved upon submission.
        """
        # Since the info actually gets saved, we need a new stripe customer
        self.account.stripe_customer_id = "cus_AMfJRa7tVVw1So"
        self.commit()
        billing_page = self.login().navigation_go_to_billing()
        edit_page = billing_page.edit_billing_details()
        edit_page.fill_in_billing("John Smith", "Some street", "Paris", "1111", "FR")
        billing_page = edit_page.submit()
        card = billing_page.get_credit_card()
        self.assertEqual("John Smith", card.name)
        self.assertEqual("Some street\nParis 1111\nFR", card.address)
