# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Package Profiles
================

A package profile is a set of package constraints that can be applied to
managed computers.

The methods available are related to package profile management. The user can
search, create, copy, read, update and delete profiles associated with the
account.
"""

from copy import deepcopy

from canonical.txapi.server.schema import Schema, Unicode, Integer, List
from canonical.txapi.server.method import method
from zope.security.interfaces import Unauthorized

from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.main.errors import UnknownAccessContextError
from canonical.landscape.api.error import (
    UnknownComputerError,
    InvalidParameterCombinationError,
    MissingParameterError,
)

from canonical.landscape.api.schema import UnicodeTitle
from canonical.landscape.api.methods import AccountMethod, TagAssociationMethod
from canonical.landscape.api.routes import ManagementRoute, SubjectRoute
from canonical.landscape.model.package.profile import (
    PackageProfilesManagement,
    UnknownPackageProfileNameError,
    DuplicatePackageProfileError,
    NoFoundPackagesError,
    InvalidPackageProfileMaterialError,
    EmptyPackageProfileError,
    InvalidPackageProfileNameError,
)
from canonical.landscape.model.package.package import (
    InvalidPackageConstraintError,
    InvalidPackageConstraintTypeError,
)

from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.computer.collection import ComputerCollection


__all__ = [
    "GetPackageProfiles",
    "CopyPackageProfile",
    "CreatePackageProfile",
    "RemovePackageProfile",
    "AssociatePackageProfile",
    "DisassociatePackageProfile",
    "EditPackageProfile",
]


@method
class GetPackageProfiles(AccountMethod):
    """
    Get the details of all Package Profiles defined in the account.

    - **names**: Specific profile names to limit results. (optional)

    For example, the following command limits the results to return the single
    package profile named "mysqlprofile" in this account::

        ?action=GetPackageProfiles&names.1=mysqlprofile

    The method returns a JSON representation of the package profiles. It
    includes a list of computers which are constrained by this profile as well
    as a list of constrained computers which are non-compliant::

        [{"id": 2,
          "creation_time": "2012-03-06T18:52:09Z",
          "computers": {"constrained": [4, 13, 79, 945],
                        "non-compliant": [13]},
          "modification_time": "2012-03-06T18:52:09Z",
          "description": "Package profile to lock down mysql to >= 5.1.0",
          "version": "2",
          "tags": ["server", "desktop"],
          "constraints": [("depends", "mysql >= 5.1.0"),
                          ("conflicts", "python-mysqldb < 1.2.2")],
          "name": "mysqlprofile",
          "title": "MySQL Profile"}]
    """

    requires = "ViewAccount"
    schema = Schema(
        name="GetPackageProfiles",
        parameters=[
            List(
                "names",
                Unicode(),
                optional=True,
                doc="A list of package profile names to limit the result.",
            )
        ],
        errors=[],
        doc="""Get the details of all Package Profiles defined in the account.
            """,
    )

    route = ManagementRoute(factory=PackageProfilesManagement)

    def query(self, context, call):
        """
        Return a list of C{PackageProfile}s
        """
        result = context.get_package_profiles(names=call.args.names or None)
        return [profile.get_state() for profile in result]


@method
class CopyPackageProfile(AccountMethod):
    """
    Copy an existing package profile to a package profile with a new name.

    - **name**: A name of the existing package profile to copy.
    - **destination_name**: The profile name of the copied package profile.
      (optional)
    - **title**: The title of the copied package profile.  By default the
      title of the source profile is used. (optional)
    - **description**: The descrption of the copied package profile.  By
      default the descrption of the source profile is used. (optional)
    - **access_group**: An optional name of the access group to copy the
      profile to. Defaults to the origin's access group.

    For example, the following command copies the package profile
    'mysqlprofile' to 'copiedprofilename' within registered with the account::

        ?action=CopyPackageProfile&name=mysqlprofile
            &destination_name=copiedprofilename

    The method returns a JSON representation of the copied package profile::

        [{"id": 3,
          "creation_time": "2012-03-06T18:52:09Z",
          "computers": {"constrained": [4, 13, 79, 945],
                        "non-compliant": [13]},
          "modification_time": "2012-03-06T18:52:09Z",
          "description": "Package profile to lock down mysql to >= 5.1.0",
          "version": "2",
          "constraints": [("depends", "mysql >= 5.1.0"),
                          ("conflicts", "python-mysqldb < 1.2.2")],
          "name": "copiedprofilename"}]

    The following errors may be raised:

    - **UnknownPackageProfileName**: source package profile name does not
      exist.
    - **DuplicatePackageProfile**: destination package profile name already
      exists.
    - **UnknownAccessGroup**: The given access group is not known.
    - **Unauthorised**: The person is not authorised to copy package
      profiles in the given access group.
    """

    requires = "ViewAccount"
    schema = Schema(
        name="CopyPackageProfile",
        parameters=[
            Unicode("name", doc="A name of the existing package profile to copy."),
            Unicode(
                "destination_name",
                optional=True,
                doc="The profile name of the copied package profile.",
            ),
            Unicode(
                "title",
                optional=True,
                doc=(
                    "A title for the new profile. If not specified, the"
                    " title of the source profile is used."
                ),
            ),
            Unicode(
                "description",
                optional=True,
                doc=(
                    "A description for the new profile. If not specified,"
                    " the title of the source profile is used."
                ),
            ),
            Unicode(
                "access_group",
                optional=True,
                doc=(
                    "Name of the access group to copy the "
                    "profile to. Defaults to the origin's access group."
                ),
            ),
        ],
        errors=[
            UnknownPackageProfileNameError,
            DuplicatePackageProfileError,
            UnknownAccessContextError,
            Unauthorized,
        ],
        doc="""
    Copy an existing package profile to a package profile with a new name and
    optionally a different title and description.
    """,
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )

    def query(self, context, call):
        """
        Copy a C{PackageProfile} and return the copied C{PackageProfile} state.
        """
        # We don't technically need the source profile object to copy it, since
        # the management method takes the source profile's name. We make the
        # profile the context only for improved permission checking.
        management = self.get_management(PackageProfilesManagement, call)
        copy = management.copy_package_profile(
            call.args.name,
            name=call.args.destination_name,
            title=call.args.title,
            description=call.args.description,
            access_context_name=call.args.access_group,
        )
        return copy.get_state()


@method
class CreatePackageProfile(AccountMethod):
    """
    Create a package profile.

    - **title**: The title of the package profile to create.
    - **description**: The description of the new profile.
    - **source_computer_id**: A computer ID to find a computer which will be
      used as the basis of the package profile. (optional)
    - **material**: Package data in the format of 'dpkg --get-selections' or
      CSV (as exported by Landscape). (optional)
    - **constraints**: A list of constraints to populate the package
      profile. (optional)
    - **access_group**: An optional name of the access group to create the
      profile into.

    source_computer_id and material are mutually exclusive.

    For example, the following command creates a package profile 'newprofile'
    based on the packages installed on computer 500::

        ?action=CreatePackageProfile&title=New%20profile
            &description=New%20profile
            &source_computer_id=500

    The method returns a JSON representation of the created package profile::

        [{"id": 3,
          "creation_time": "2012-03-06T18:52:09Z",
          "computers": {"constrained": [4, 13, 79, 945],
                        "non-compliant": [13]},
          "modification_time": "2012-03-06T18:52:09Z",
          "description": "New profile",
          "version": "1",
          "constraints": [("depends", "mysql >= 5.1.0"),
                          ("conflicts", "python-mysqldb < 1.2.2")],
          "name": "new-profile",
          "title": "New profile"}]

    The following errors may be raised:

    - **InvalidPackageProfileName**: The name specified is not valid.
    - **InvalidParameterCombination**: source_computer_id and material were
      both passed.
    - **UnknownComputer**: An invalid source_computer_id was passed.
    - **InvalidPackageProfileMaterial**: The material passed could not be
      parsed.
    - **NoFoundPackages**: The specified computer did not have any package data
    - **UnknownAccessGroup**: The given access group is not known.
    - **Unauthorised**: The person is not authorised to create package
      profiles in the given access group.  profile with no constraints. At
      least one valid constraint must be provided.

    """

    schema = Schema(
        name="CreatePackageProfile",
        parameters=[
            UnicodeTitle("title", doc="The title of the package profile to create."),
            Unicode("description", doc="The description of the new profile."),
            Integer(
                "source_computer_id",
                optional=True,
                doc=(
                    "A computer ID to find a computer which will be used "
                    "as the basis of the package profile."
                ),
            ),
            Unicode(
                "material",
                optional=True,
                doc=(
                    "Package data in the format of "
                    "'dpkg --get-selections' or CSV (as exported by "
                    "Landscape)."
                ),
            ),
            List(
                "constraints",
                Unicode(),
                optional=True,
                doc=(
                    "Alternative to material, constraint specifications "
                    'in the form of "depends packagename" or '
                    '"conflicts packagename < 1.0".'
                ),
            ),
            Unicode(
                "access_group",
                optional=True,
                default=DEFAULT_ACCESS_CONTEXT_NAME,
                doc=("Optional name of the access group to create " "the profile into"),
            ),
        ],
        errors=[
            InvalidPackageProfileNameError,
            InvalidParameterCombinationError,
            EmptyPackageProfileError,
            UnknownComputerError,
            InvalidPackageProfileMaterialError,
            NoFoundPackagesError,
            UnknownAccessContextError,
            InvalidPackageConstraintError,
            Unauthorized,
            InvalidPackageConstraintTypeError,
            DuplicatePackageProfileError,
        ],
        doc="""
    Create a package profile.

    source_computer_id and material are mutually exclusive.
    """,
    )

    requires = "ViewAccount"

    route = ManagementRoute(factory=PackageProfilesManagement)

    def query(self, context, call):
        """
        Create a package profile and return its state.
        """
        if call.args.source_computer_id:
            collection = ComputerCollection(
                get_main_store(),
                account_id=call.principal.account_id,
                person_id=call.principal.id,
                permission="landscape.ViewComputerPackages",
            )
            collection = collection.with_computer_ids([call.args.source_computer_id])
            computer = collection.find().one()
            if computer is None:
                raise UnknownComputerError(
                    "Unknown computer %s" % (call.args.source_computer_id,)
                )
        else:
            computer = None

        profile = context.create_package_profile(
            call.args.title,
            call.args.description,
            material=call.args.material,
            computer=computer,
            access_context_name=call.args.access_group,
        )

        person = get_person_by_id(call.principal.id)
        profile.modify_package_constraints(add_constraints=call.args.constraints)

        profile.on_create(person)
        return profile.get_state()


@method
class RemovePackageProfile(AccountMethod):
    """
    Remove a package profile, given its name.

    Example of a valid request that removes profile 'foobar'::

        ?action=RemovePackageProfile&name=foobar

    The following errors may be raised:

    - **UnknownPackageProfileName**: No package profile with the specified name
      exists.
    """

    requires = "EditPackageProfile"

    schema = Schema(
        name="RemovePackageProfile",
        parameters=[Unicode("name", doc="The name of the package profile to remove.")],
        errors=[UnknownPackageProfileNameError],
        doc="""Remove a package profile, given its name.""",
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )

    def query(self, context, call):
        """Remove a package profile."""
        profile = context  # For readability
        # Since the delete logic is already on the management class, let's
        # reuse it.
        management = self.get_management(PackageProfilesManagement, call)
        management.remove_package_profiles([profile])
        return {}


@method
class AssociatePackageProfile(TagAssociationMethod):
    """
    Associate a package profile to computers with specific tags or to all
    computers.

    - **name**: Name of the package profile.
    - **tags.#**: List of tag names to associate to the profile.
    - **all_computers**: true if the profile should be associated to all
      computers. This parameter is *optional* and defaults to
      *false*. Individual tags associated to the profile will remain, but they
      will only be effective if the "all_computers" flag is later disabled.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request that associates the profile 'foobar' with the
    tags 'desktop' and 'laptop'::

        ?action=AssociatePackageProfile&name=foobar
            &tags.1=desktop
            &tags.2=laptop
            &all_computers=false

    The following errors may be raised:

    - **UnknownPackageProfileName**: No package profile with the specified name
      exists.

    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.

    The state of the profile will be returned.
    """

    requires = "EditPackageProfile"

    schema = TagAssociationMethod.schema.extend(
        name="AssociatePackageProfile",
        errors=[UnknownPackageProfileNameError],
        doc="""
    Associate a package profile to computers with specific tags or to all
    computers. If a tag with a given name doesn't exist, it will be
    automatically created.

    An 'all_computers' value of 'true' and a list of tags are mutually
    exclusive. Only one or the other may be passed.""",
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )


@method
class DisassociatePackageProfile(TagAssociationMethod):
    """
    Disassociate a package profile from computers with specified tags or from
    all computers.

    If **all_computers=true**, the profile will be unflagged as applying to all
    computers, but will still be enabled for computers which have specific tags
    associated with it.

    - **name**: Name of the package profile.
    - **tags.#**: List of tag names to disassociate from the profile.
    - **all_computers**: If true, the profile will only remain enabled for
      computers with tags associated to the profile. This parameter is
      *optional* and defaults to *false*.

    **tags.#** and **all_computers=true** are mutually exclusive.

    Example of a valid request::

        ?action=DisassociatePackageProfile&name=foobar
            &tags.1=desktop&tags.2=laptop

    The following errors may be raised:

    - **UnknownPackageProfileName**: No package profile with the specified name
      exists.
    - **UnknownTag**: No tag with the specified name exists.
    - **InvalidParameterCombination**: Some tags were specified and a 'true'
      value for 'all_computers' was passed.

    The state of the profile will be returned.
    """

    requires = "EditPackageProfile"

    associate = False

    schema = TagAssociationMethod.schema.extend(
        name="DisassociatePackageProfile",
        errors=[UnknownPackageProfileNameError],
        doc="""Disassociate package profile from computers with specified
            tags or from all computers.""",
    )

    route = SubjectRoute(
        factory=PackageProfilesManagement,
        getter="get_package_profile_by_name",
        identifier_arg="name",
        unknown_error_factory=UnknownPackageProfileNameError,
    )


@method
class EditPackageProfile(AccountMethod):
    """
    Add or remove constraints related to a package profile. Constraints can be
    dependencies or conflicts.

    - **name**: Name of the package profile.
    - **title**: The new title of the package profile.
    - **add_constraints.#**: Constraint specifications to add in the form of
      "depends packagename" or "conflicts packagename < 1.0".
    - **remove_constraints.#**: Constraint specifications to remove in the form
      of "depends packagename" or "conflicts packagename < 1.0".

    Example of a valid request::

        ?action=EditPackageProfile&name=foobar
            &add_constraints.1=depends%20packagename
            &add_constraints.2=conflicts%20python%20%3C%202.5
            &remove_constraints.1=depends%20postgres

    The following errors may be raised:

    - **InvalidPackageConstraint**: The specified package constraint could not
      be parsed.
    - **InvalidConstraintType**: A constraint type other than 'depends' or
      'conflicts' was provided.
    - **EmptyPackageProfile**: Cannot remove constraints as it leaves a
      profile with no constraints. An empty package profile should be removed
      instead.

    The state of the profile will be returned.
    """

    # Note: This method is not split in "add_contraints" and
    # "remove_contraints" like most of the rest of the API since every change
    # to constraints regenerates a deb package and all associated activities
    # and is therefore quite expensive.
    requires = "ViewAccount"

    schema = Schema(
        name="EditPackageProfile",
        parameters=[
            Unicode("name", doc="The name of the package profile."),
            UnicodeTitle(
                "title", doc="The new title of the package profile.", optional=True
            ),
            List(
                "add_constraints",
                Unicode(),
                optional=True,
                doc=(
                    "List of constraints specifications to add in the form "
                    'of "depends packagename" or '
                    '"conflicts packagename < 1.0".'
                ),
            ),
            List(
                "remove_constraints",
                Unicode(),
                optional=True,
                doc=(
                    "List of constraints specifications to remove in the "
                    'form of "depends packagename" or '
                    '"conflicts packagename < 1.0".'
                ),
            ),
        ],
        errors=[
            UnknownPackageProfileNameError,
            InvalidPackageConstraintTypeError,
            InvalidPackageConstraintError,
            EmptyPackageProfileError,
            MissingParameterError,
            Unauthorized,
        ],
        doc="""
    Add or remove constraints related to a package profile. Constraints can
    be dependencies or conflicts.""",
    )

    route = ManagementRoute(factory=PackageProfilesManagement)

    def query(self, context, call):
        """
        Edit package constraints related to a package profile.

        @param context: The PackageProfile being edited.
        """
        profile = context.get_package_profile_by_name(call.args.name)
        if profile is None:
            raise UnknownPackageProfileNameError(call.args.name)

        if not any(
            (call.args.remove_constraints, call.args.add_constraints, call.args.title)
        ):
            raise MissingParameterError(
                "The request must contain at least one of the parameters "
                "add_constraints, remove_constraints or title."
            )

        context.edit(
            profile.name,
            title=call.args.title,
            add_constraints=call.args.add_constraints,
            remove_constraints=call.args.remove_constraints,
        )

        return profile.get_state()


@method
class ModifyPackageProfile(EditPackageProfile):
    """
    L{ModifyPackageProfile} is deprecated and now exists only as an alias for
    L{EditPackageProfile} to avoid breaking the API for old scripts.
    """

    schema = deepcopy(EditPackageProfile.schema)
    schema.name = "ModifyPackageProfile"
    schema.doc = (
        "THIS FUNCTION IS DEPRECATED. please use "
        + "edit-package-profile instead. "
        + schema.doc
    )
