# Copyright 2005-2022 Canonical Limited. All rights reserved.

from datetime import datetime
from canonical.lib.timestamp import format_datetime
from io import StringIO
import csv
import re

from zope.interface import implementer
from zope.security.interfaces import Unauthorized

from storm.locals import Int, Unicode, Reference, ReferenceSet, AutoReload, DateTime

from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE

from canonical.landscape.model.error import DuplicateModelError
from canonical.landscape.model.main.account import DEFAULT_ACCESS_CONTEXT_NAME
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.tag import get_tag_by_name
from canonical.landscape.model.main.interfaces import IAccessDelegateAware
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.account.helper import ComputerGroupHelper
from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.account.computergroup import (
    ComputerGroup,
    ComputerGroupItem,
)
from canonical.landscape.model.package.market import PackageMarket
from canonical.landscape.model.package.interfaces import IPackageProfile
from canonical.landscape.model.package.package import is_valid_package_name
from landscape.lib.apt.package.skeleton import PackageSkeleton
from canonical.landscape.model.package.search import PackageSearch
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.package.management import PackageManagement
from canonical.landscape.model.package.package import (
    Package,
    DEB_PACKAGE,
    DEB_REQUIRES,
    DEB_CONFLICTS,
    DEB_NAME_PROVIDES,
    DEB_UPGRADES,
    parse_relation_value,
    relation_type_to_string,
    parse_constraint,
    build_relation_value,
    build_relation_type,
    InvalidPackageConstraintError,
)
from canonical.lib.archive import DebFile, FileInfo
from canonical.lib.text import slug_from_title, describe_dict
from canonical.landscape.model.activity.activity import Activity
from canonical.landscape.model.activity.interfaces import IActivityCollection
from canonical.landscape.model.activity.status import (
    QUEUED,
    BLOCKED,
    SCHEDULED,
    INPROGRESS,
    PENDING_STATUSES,
)
from canonical.landscape.model.activity.collection import (
    ActivityChildCollection,
    ActivityParentCollection,
)
from canonical.landscape.model.package.activities import ChangePackageProfilesRequest
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.resource.package import (
    get_computer_ids_with_installed_package,
    get_computer_ids_with_installed,
)


__all__ = [
    "PackageProfilesManagement",
    "PackageProfile",
    "PackageProfileConstraint",
    "PackageProfileActivities",
    "UnknownPackageProfileNameError",
    "DuplicatePackageProfileError",
    "InvalidPackageProfileNameError",
    "InvalidPackageProfileMaterialError",
    "PackageProfileError",
    "EmptyPackageProfileError",
    "NoFoundPackagesError",
    "InvalidParameterCombinationError",
    "get_package_profiles_with_computer_ids",
]

__clearlist__ = [
    "PackageProfilesManagement",
    "PackageProfile",
    "PackageProfileActivities",
    "UnknownPackageProfileNameError",
    "DuplicatePackageProfileError",
    "InvalidPackageProfileNameError",
    "EmptyPackageProfileError",
    "InvalidPackageProfileMaterialError",
    "PackageProfileError",
]

__requires_prefix__ = "landscape."


# Two supported formats for package profile csv.
HEADER_V1 = "Id,Name,Data Type,Value,Constraint Type,Constraint Value"
HEADER_V1_COLUMNS = 6

HEADER_V2 = "Id,Name,Title,Data Type,Value,Constraint Type,Constraint Value"
HEADER_V2_COLUMNS = 7


class PackageProfileError(Exception):
    """
    Generic base class for package profile related errors.
    """


class NoFoundPackagesError(PackageProfileError):
    """
    Raised when creating a new L{PackageProfile} if computer provided does
    not report any installed packages. Without this information the
    L{PackageProfile} can not be created.
    """


class InvalidParameterCombinationError(PackageProfileError):
    """
    Raised when creating a new L{PackageProfile} if both parameters material
    and computer are passed. These parameters are mutually exclusive.
    """


class DuplicatePackageProfileError(DuplicateModelError, PackageProfileError):
    """
    Raised when creating or copying a L{PackageProfile} if
    the C{PackageProfile.name} already exists.
    """

    model_name = "Package profile"


class UnknownPackageProfileNameError(PackageProfileError):
    """
    Raised when unable to find a L{PackageProfile} by the specified C{name}.
    """


class InvalidPackageProfileMaterialError(PackageProfileError):
    """
    Raised if package profile material is either too long or if the format is
    neither CSV nor dpkg --get-selections output.
    """


class InvalidPackageProfileNameError(PackageProfileError):
    """
    Raised when the name of a package profile doesn't meet its constraints.
    """


class EmptyPackageProfileError(PackageProfileError):
    """
    Raised if a L{PackageProfile} is created without any valid constraints.
    """


def remove_trailing_empty_line(lines):
    """Removes an empty sting from a the last element in a C{List}"""
    if not len(lines[-1].strip()):
        lines = lines[:-1]
    return lines


@requires_advisor
@implementer(IActivityCollection)
class PackageProfilesManagement(ModelManagement):
    """
    Provide methods to manage and retrieve L{PackageProfile}s linked to an
    account.
    """

    # Desktop systems can easily grow over 3000 packages when running
    # dpkg --get-selections, so double that amount should be enough for most
    # practical cases.
    max_import_lines = 6000

    requires("ViewAccount", PRIVATE, ["account", "access_context", "person"])

    def _parse_get_selections(self, name, create_profile, material):
        """
        Attempt to parse material as dpkg --get-selections content to create
        a L{PackageProfile}. Since we would like to avoid big DOS attacks, set
        a somewhat appropriate max for the number of lines we expect to parse
        to C{PackageProfilesManagement.max_import_lines}.

        @param name: the name of the package profile.
        @param create_profile: a callable creating and returning a
            L{PackageProfile} to populate.
        @param material: CSV or 'dpkg --get-selections' material that can be
            used to create the package profile

        @raises InvalidPackageProfileMaterialError: if errors occur parsing
            the dpkg --get-selections content
        @raises InvalidPackageProfileMaterialError: if material is invalid
            dpkg --get-selections format

        @returns C{False}: if material format isn't does not appear to be
            dpkg --get-selections content.
        @returns C{True}: on successful parse.
        """
        packages = []

        # Using [A-Za-z0-9] instead of \w because we don't want to match
        # non-ascii chars.
        dpkg_regex = (
            r"[A-Za-z0-9]+([-+:.A-Za-z0-9]+)?\s+" r"(install|deinstall|purge|hold)\s*"
        )
        if not re.match(dpkg_regex, material):
            # We don't look like --get-selections output
            return None

        lines = material.split("\n")
        lines = remove_trailing_empty_line(lines)

        for line_count, line in enumerate(lines):
            if not re.match(dpkg_regex + r"$", line):
                message = (
                    "Can not import package profile %s.\n"
                    "Invalid format of dpkg --get-selections material "
                    "on line %d." % (name, line_count + 1)
                )
                raise InvalidPackageProfileMaterialError(message)
            [package, state] = re.split("\s+", line.strip())

            package = package.split(":")[0]  # Drop platform-aliases ':i386'

            if state in ["install", "hold"]:
                # We handle all dpkg get-selections format requirements in the
                # re.match above. Don't retest unicode format or name.
                packages.append(str(package))

        profile = create_profile()
        for package in packages:
            profile._add_package_constraint("depends", package)
        return profile

    def _parse_csv_row(self, name, row, line_count, column_count, copy_name):
        """
        Parse a CSV row and raise any errors for invalid row content.
        Return a C{tuple} containing C{tags}, C{constraint} and an
        C{all_computers} C{bool}.
        """
        valid_data_types = [
            "DESCRIPTION",
            "TAGS",
            "ALLCOMPUTERS",
            "DEPENDS",
            "CONFLICTS",
            "VERSION",
            "CREATIONTIME",
            "MODIFICATIONTIME",
        ]
        tags = []
        constraint = None
        all_computers = None
        if len(row) != column_count:
            message = (
                "Can not import package profile %s from CSV.\n"
                "Line %d has %d columns instead of %d."
                % (name, line_count + 1, len(row), column_count)
            )
            raise InvalidPackageProfileMaterialError(message)

        if copy_name:
            # Backward-compatibility with HEADER_V1
            (id, profile_name, data_type, value, constraint_type, version) = row
            title = profile_name
        else:
            (id, profile_name, title, data_type, value, constraint_type, version) = row

        data_type = data_type.upper()
        if line_count > 0 and data_type not in valid_data_types:
            message = (
                "Can not import package profile %s.\n"
                'CSV contains invalid Data Type "%s" on line %d.'
                % (name, data_type, line_count + 1)
            )
            raise InvalidPackageProfileMaterialError(message)

        if data_type == "TAGS":
            tag_names = value.split(" ")
            for tag in tag_names:
                tags.append(get_tag_by_name(str(tag)))

        if data_type == "ALLCOMPUTERS" and value.upper() == "TRUE":
            all_computers = True
        if data_type in ["DEPENDS", "CONFLICTS"]:
            if constraint_type.upper() == "ANY":
                constraint_type = ""
                version = ""
            try:
                # Validate unicode and name regex format
                package = str(value)
                constraint = (data_type.lower(), package, constraint_type, version)
                if not is_valid_package_name(package):
                    raise InvalidPackageConstraintError()
            except (UnicodeDecodeError, InvalidPackageConstraintError):
                message = (
                    "Can not import package profile %s.\n"
                    "Invalid package constraint format "
                    "on line %d." % (name, line_count + 1)
                )
                raise InvalidPackageConstraintError(message)
        return (tags, constraint, all_computers)

    def _parse_csv_material(self, name, create_profile, material):
        """
        Attempt to parse material as CSV content to create a L{PackageProfile}.
        Parse CSV data_type keys: C{DEPENDS}, C{CONFLICTS}, C{ALLCOMPUTERS},
        C{TAGS}. If material does not appear to be CSV content, return C{None}.

        @param name: the name of the package profile.
        @param create_profile: a callable creating and returning a
            L{PackageProfile} to populate.
        @param material: CSV or 'dpkg --get-selections' material that can be
            used to create the package profile

        @raises InvalidPackageProfileMaterialError: if material is invalid
            CSV format

        @returns C{False}: if material format isn't CSV and C{True} on success.
        """

        tags = []
        constraints = []
        all_computers = False

        # Are we Package Profile CSV format?
        valid_header = False
        copy_name = False
        if material[0 : len(HEADER_V1)] == HEADER_V1:
            valid_header = True
            column_count = HEADER_V1_COLUMNS
            copy_name = True
        elif material[0 : len(HEADER_V2)] == HEADER_V2:
            valid_header = True
            column_count = HEADER_V2_COLUMNS
        if not valid_header:
            return None

        # prune any carriage returns so we can split on line-feed
        material = material.replace("\r", "")
        try:
            # Don't trust csv.Sniffer at all. It incorrectly discovers
            # dialect.delimiter and dialect.lineterminator in CSV material
            lines = material.split("\n")
            lines = remove_trailing_empty_line(lines)
            # Don't trust dialect.delimiter on poorly formatted csv
            # Instead use delimiter="," and we will parse files appropriately
            csv_reader = csv.reader(lines, delimiter=",")
        except csv.Error:
            return None

        for line_count, row in enumerate(csv_reader):
            (new_tags, constraint, new_all_computers) = self._parse_csv_row(
                name, row, line_count, column_count, copy_name
            )
            if new_tags:
                tags.extend(new_tags)
            if constraint:
                constraints.append(constraint)
            if new_all_computers:
                all_computers = new_all_computers

        profile = create_profile()
        for (data_type, package, constraint_type, version) in constraints:
            profile._add_package_constraint(
                data_type, package, constraint_type, version
            )
        if tags or all_computers:
            # Need to call on_create first to generate a package & deps for
            # the profile constraints before adding tags or computer
            # association through update_association
            profile.on_create(self.person)
        profile.update_association(new_tags=tags, all_computers=all_computers)
        return profile

    @requires("ViewAccount")
    def get_computer_ids_with_packages(self):
        """
        Get L{Computer.id}s that can be used as a basis for creating a
        package profile. Only computers which have reported installed packages
        will be returned.
        """
        access_context_ids = self.get_access_context_ids("ViewComputer")
        result = self.account.get_computers()
        result = result.find(Computer.access_context_id.is_in(access_context_ids))
        account_computer_ids = result.values(Computer.id)
        return get_computer_ids_with_installed(account_computer_ids)

    @requires("ViewAccount")
    def create_package_profile(
        self,
        title,
        description,
        material=None,
        computer=None,
        constraint_operator="",
        access_context_name=DEFAULT_ACCESS_CONTEXT_NAME,
        max_material_lines=max_import_lines,
    ):
        """
        Create a new L{PackageProfile} associated with this account.
        If material is specified, parse and import all L{PackageProfile}
        characteristics based on either CSV or dpkg --get-selections material.

        @param title: the title of the package profile.
        @param description: the description of the package profile.
        @param material: CSV or 'dpkg --get-selections' material that can be
            used to create the package profile (optional)
        @param computer: the existing computer from which to create a new
            profile (optional)
        @param constraint_operator: an operator string such as ">", ">=", "="
            can be passed when creating a profile from an existing computer. If
            set, all package constraints for this profile will be set using the
            operator compared with the current package.versions installed on
            the computer (optional).
        @param max_material_lines: A limit to the number of lines of imported
            material allowed when creating a profile. Default is
            C{PackageProfilesManagement.max_import_lines}.

        @raises InvalidPackageProfileMaterialError: if material provided is
            greater than allowed C{max_material_lines} or invalid CSV and dpkg
            --get-selections format
        """
        # Derive from title.
        def check_existing(slug):
            """
            Retrieve all the package profile names so it can be used to
            derive a name from the title.
            """
            return self.account.store.find(
                PackageProfile.name, PackageProfile.account_id == self.account.id
            )

        name = slug_from_title(title, 50, check_existing)

        def create_profile():
            access_context = self.get_access_context(
                access_context_name, "CreatePackageProfile"
            )

            return PackageProfile(
                self.account, name, title, description, access_context.id
            )

        if material and computer:
            message = (
                "Invalid parameter combination. Specify only one "
                "of the parameters material and computer."
            )
            raise InvalidParameterCombinationError(message)

        if material:
            lines = material.split("\n")
            lines = remove_trailing_empty_line(lines)

            if len(lines) > max_material_lines:
                message = (
                    "Package profile content length of %d lines is "
                    "greater than allowed maximum of %d."
                    % (len(lines), max_material_lines)
                )
                raise InvalidPackageProfileMaterialError(message)
            profile = self._parse_csv_material(name, create_profile, material)
            if not profile:
                profile = self._parse_get_selections(name, create_profile, material)
                if not profile:
                    message = (
                        "Package profile material is not valid CSV "
                        "or dpkg --get-selections format."
                    )
                    raise InvalidPackageProfileMaterialError(message)
        elif computer:
            selection = ComputerSelection(self.account)
            selection.select_computers([computer])
            search = PackageSearch(selection)
            packages = [
                package
                for package, states in search.get_package_states(
                    installed=True, limit=None
                )
            ]
            if len(packages) == 0:
                raise NoFoundPackagesError("Computer didn't report package information")

            profile = create_profile()
            for package in packages:
                version = package.version if constraint_operator else ""
                profile._add_package_constraint(
                    "depends", str(package.name), constraint_operator, version
                )
        else:  # Create generic profile
            profile = create_profile()
        message = get_event_log_message_from_profile(profile)
        self._event_log.create_log_entry(profile, message)
        return profile

    @requires("CreatePackageProfile")
    def copy_package_profile(
        self,
        source_name,
        name=None,
        title=None,
        description=None,
        access_context_name=None,
    ):
        """
        Copy an existing PackageProfile to a new PackageProfile named
        "<source_name>-copy" by default, or with the specified name.

        @param source_name: the name of the source package profile to copy.
        @param name: the new name of the copied package profile (optional).
            If destination_name is None, the new name will be based on the
            source "<source_name>-copy".
        @param title: optionally, a title for the new profile. If not
            specified, the same title of the source profile is used.
        @param description: optionally, a description for the new profile. If
            not specified, the same description as the source profile is used.
        @param access_context_name: The name of the access context to copy the
            package profile into. If None, the new profile will be copied to
            the same access context as the source.

        @raises UnknownPackageProfileNameError if a source package profile with
            parameter name does not exist.
        @raises DuplicatePackageProfileError if a package profile already
            exists with the name specified by the paremeter destination_name.
        """
        source = self.get_package_profile_by_name(source_name)
        if not source:
            raise UnknownPackageProfileNameError(
                "Package profile named {} not found.".format(source_name)
            )
        if not name:
            index = 1
            name = "{}-copy-{}".format(source_name, index)
            while self.get_package_profile_by_name(name):
                index += 1
                name = "{}-copy-{}".format(source_name, index)
        else:
            self.check_duplicate(
                PackageProfile, DuplicatePackageProfileError, name=name
            )
        if access_context_name is not None:
            access_context = self.get_access_context(
                access_context_name, "CreatePackageProfile"
            )
        else:
            access_context = self.get_access_context(
                source.access_context.name, "CreatePackageProfile"
            )
        if title is None:
            title = source.title
        if description is None:
            description = source.description

        destination = PackageProfile(
            self.account, name, title, description, access_context.id
        )

        # Copy constraints to destination
        for constraint in source.constraints:
            (kind, package, operator, version) = constraint.parse()
            destination._add_package_constraint(kind, str(package), operator, version)

        destination.on_create(self.person)

        message = get_event_log_message_from_profile(destination)
        self._event_log.create_log_entry(destination, message)
        return destination

    @requires("ViewAccount")
    def get_package_profiles(self, names=None):
        """
        Return all the L{PackageProfile}s associated with this account, sorted
        by name.

        @param names: If given, return only package profiles with the given
            names.
        """
        access_context_ids = self.get_access_context_ids("ViewPackageProfile")
        where = [
            PackageProfile.account_id == self.account.id,
            PackageProfile.computer_group_id == ComputerGroup.id,
            PackageProfile.access_context_id.is_in(access_context_ids),
        ]
        if names is not None:
            where.append(PackageProfile.name.is_in(names))
        result = self.account.store.find(PackageProfile, *where)
        return result.order_by(PackageProfile.name)

    @requires("ViewAccount")
    def remove_package_profiles(self, profiles):
        """Remove the given L{PackageProfile}s.

        @param profiles: C{list} of C{PackageProfile}s.

        @return: L{PackageProfile}s that didn't get removed because the user
            doesn't have permission on them.
        """
        access_context_ids = self.get_access_context_ids("EditPackageProfile")

        profile_ids = [profile.id for profile in profiles]
        computer_ids = set(
            self.account.store.find(
                ComputerGroupItem.computer_id,
                ComputerGroupItem.computer_group_id == PackageProfile.computer_group_id,
                PackageProfile.computer_group_id == ComputerGroup.id,
                PackageProfile.access_context_id.is_in(access_context_ids),
                PackageProfile.id.is_in(profile_ids),
                PackageProfile.account_id == self.account.id,
            )
        )

        remaining_profiles = []
        for profile in profiles:
            if profile.access_context.id in access_context_ids:
                self._event_log.create_log_entry(profile, "Package profile removed.")
                profile.remove()
            else:
                remaining_profiles.append(profile)

        if computer_ids:
            from canonical.landscape.model.package.alerts import PackageProfilesAlert

            self.account.check_rules([PackageProfilesAlert], computer_ids)

        return remaining_profiles

    @requires("QueryPackageProfile")
    def get_package_profile_by_id(self, profile_id):
        """Return a L{PackageProfile} by its ID.

        @param id: C{PackageProfile.id}.
        """
        profile = self.account.store.get(PackageProfile, profile_id)
        if profile is not None and profile.account_id != self.account.id:
            return
        return profile

    @requires("QueryPackageProfile")
    def get_package_profile_by_name(self, name):
        """Return a L{PackageProfile} by its name.

        @param name: C{PackageProfile.name}.
        """
        return self.account.store.find(
            PackageProfile,
            PackageProfile.name == name,
            PackageProfile.account_id == self.account.id,
        ).one()

    @requires("ViewAccount")
    def edit(
        self,
        name,
        title=None,
        description=None,
        add_constraints=[],
        remove_constraints=[],
        new_tags=None,
        computer_association=None,
    ):
        """
        Change a package profile.

        @param name: The name of the package profile to be edited.
        @param title: The new title of the package profile.
        @param description: The description of the package profile.
        @param package_constraints: The package constraints for this package
            profile.
        @param new_tags: The tags this profile will be associated with.
        @param computer_association: If the profile should be associated to all
            computers.
        """
        access_context_ids = self.get_access_context_ids("EditPackageProfile")
        profile = self.get_package_profile_by_name(name)
        if profile.access_context.id not in access_context_ids:
            raise Unauthorized(
                "Permission denied. Can't change package profile '%s'" % name
            )

        profile.update_association(new_tags, computer_association)

        messages = []
        if title is not None and title != profile.title:
            messages.append("Title changed from '%s' to '%s'." % (profile.title, title))
            profile.title = title

        if description and description != profile.description:
            messages.append("Description changed to '%s'." % description)
            profile.description = description

        detailed_changes = profile.modify_package_constraints(
            add_constraints, remove_constraints
        )
        prefix = "Updated package constraints with"
        constraint_message = describe_dict(prefix, detailed_changes, pluralize=False)
        if constraint_message:
            messages.append("%s." % constraint_message)
        if messages:
            profile.on_change()
            self._event_log.create_log_entry(profile, " ".join(messages))
        return profile


@requires_advisor
@implementer(IAccessDelegateAware)
class PackageProfileConstraint(object):
    """Represent one package constraint inside a L{PackageProfile}.

    @ivar id: The primary key of the constraint.
    @ivar package_profile_id: The id of the package profile we're linked to.
    @ivar type: The type can be either L{DEB_REQUIRES} or L{DEB_CONFLICTS}.
    @ivar value: The value of the constraint, for example 'foo-pkg >= 1.0'.
    """

    __storm_table__ = "package_profile_constraint"

    requires("ViewPackageProfile", PRIVATE, ["id"])
    requires("ViewPackageProfile", "EditPackageProfile", ["type", "value"])

    id = Int(primary=True, allow_none=False, default=AutoReload)
    package_profile_id = Int(allow_none=False)
    type = Int(allow_none=False)
    value = Unicode(allow_none=False)

    def __init__(self, type, value):
        self.type = type
        self.value = value

    @property
    def access_delegate(self):
        return self.package_profile

    @requires("ViewPackageProfile")
    def parse(self):
        """Parse the constraint type and value.

        @return: A 4-tuple of the form (kind, package, operator, version).
        """
        kind = relation_type_to_string(self.type)
        items = parse_relation_value(self.value)
        # For now we don't support multiple items in the relation value,
        # like for example 'pkg1 | pkg2'.
        assert len(items) == 1
        package, operator, version = items[0]
        return kind, package, operator, version


@requires_advisor
@implementer(IPackageProfile)
class PackageProfile(ComputerGroupHelper):
    """
    A package profile, modeling a set of package constraints which can be
    turned into a meta-package.
    """

    __storm_table__ = "package_profile"

    requires(
        "ViewPackageProfile",
        PRIVATE,
        [
            "id",
            "account",
            "all_computers",
            "creation_time",
            "modification_time",
            "version",
            "constraints",
            "package",
        ],
    )

    # Define security on methods inherited from ComputerGroupHelper
    requires(
        "ViewPackageProfile",
        PRIVATE,
        [
            "get_computers",
            "get_computer_count",
            "get_computer_ids",
            "get_tags",
            "access_context",
        ],
    )

    requires(
        "ViewPackageProfile",
        "EditPackageProfile",
        ["name", "title", "description", "computer_group_id"],
    )

    requires(
        "EditPackageProfile",
        PRIVATE,
        [
            "update_association",
            "update_tags",
            "enable_for_all_computers",
            "disable_for_all_computers",
        ],
    )

    id = Int(primary=True, allow_none=False, default=AutoReload)
    account_id = Int(allow_none=False)
    name = Unicode(allow_none=False)
    title = Unicode(allow_none=False)
    description = Unicode(allow_none=False)
    version = Int(allow_none=False)
    creation_time = DateTime(default=AutoReload)
    modification_time = DateTime(default=AutoReload)
    access_context_id = Int(allow_none=False)

    computer_group_id = Int()
    computer_group = Reference(computer_group_id, ComputerGroup.id)

    constraints = ReferenceSet(
        id,
        PackageProfileConstraint.package_profile_id,
        order_by=PackageProfileConstraint.value,
    )

    package_id = Int()
    _package = None

    def __init__(self, account, name, title, description, access_context_id=None):
        super(PackageProfile, self).__init__(account, access_context_id)
        self._create_computer_group()
        self.name = name
        self.title = title
        self.description = description
        self.version = 1

    @requires("ViewPackageProfile")
    def get_broken_computers(self, computer_ids=None):
        """Return the L{Computer}s that aren't compliant with this profile.

        A L{Computer} is not compliant with a certain L{PackageProfile} if
        all of the following conditions are true:

        1) It is associated to the profile.
        2) It doesn't have the profile meta-package installed.
        3) There is no pending activity trying to install or upgrade the
           profile meta-package on the computer.

        @param computer_ids: Limit the list of the returned L{Computer}s
            to the ones with matching L{Computer.id}s.
        @return: A C{ResultSet} yielding L{Computer} objects associated with
            this profile that are not in compliance.
        """
        computer_ids = self.get_broken_computer_ids(computer_ids=computer_ids)
        return self.account.get_computers(computer_ids=computer_ids)

    @requires("ViewPackageProfile")
    def get_broken_computer_ids(self, computer_ids=None):
        """Get L{Computer.id}s this profile is not compliant with."""
        pending_computer_ids = set(
            self.get_pending_activities(
                computer_ids=computer_ids, values=AccountComputerTwin.id
            )
        )

        ids = set(self.get_computer_ids())
        if computer_ids is not None:
            ids.intersection_update(computer_ids)

        # Filter out the computers which have an on-going activity for
        # installing the profile, as they are not broken by definition.
        not_pending = ids - pending_computer_ids
        if not_pending:
            # Check if some of the computers that are not pending have
            # the profile already installed, in that case we're good
            installed = set(
                get_computer_ids_with_installed_package(not_pending, self.package.id)
            )
            not_pending -= installed

        return not_pending

    def _add_package_constraint(self, kind, package, operator="", version=""):
        """Add a L{PackageProfileConstraint} to this profile.

        @param kind: A string indicating the constraint type, e.g. 'depends'.
        @param package: The package name.
        @param operator: The comparison operator if any, e.g. '<'.
        @param version: The version constraint if any, e.g. '1.2-3'.
        @raises InvalidPackageConstraintError if the package name
            doesn't meet package name format constraints.
        """
        operators = ["", "<", "<=", ">=", ">", "="]
        if not is_valid_package_name(package):
            raise InvalidPackageConstraintError(
                "Package constraint has an invalid package name format."
            )
        if operator not in operators:
            raise InvalidPackageConstraintError(
                "Package constraint has an invalid package operator %s." % operator
            )

        type = build_relation_type(kind)
        value = build_relation_value([(package, operator, version)])
        constraint = self.constraints.find(
            PackageProfileConstraint.package_profile_id == self.id,
            PackageProfileConstraint.type == type,
            PackageProfileConstraint.value == value,
        ).one()
        if not constraint:
            constraint = PackageProfileConstraint(type, value)
            self.constraints.add(constraint)
        return constraint

    def _remove_package_constraint(self, kind, package, operator, version):
        """
        Remove a L{PackageProfileConstraint} based in its value.
        """
        value = build_relation_value([(package, operator, version)])
        type = build_relation_type(kind)
        constraint = self.constraints.find(
            PackageProfileConstraint.package_profile_id == self.id,
            PackageProfileConstraint.type == type,
            PackageProfileConstraint.value == value,
        )
        constraint.remove()

    @requires("EditPackageProfile")
    def modify_package_constraints(self, add_constraints=[], remove_constraints=[]):
        """
        Modify constraints of an existing L{PackageProfileConstraint} based
        on the set of add_constraints and remove_constraints. Raise an
        C{EmptyPackageProfileError} before taking any action on the profile
        if the end result would be a L{PackageProfile} without constraints.
        """
        constraints = set(
            (relation_type_to_string(constraint.type), constraint.value)
            for constraint in self.constraints
        )

        additions = [parse_constraint(to_add) for to_add in add_constraints]
        removals = [parse_constraint(to_remove) for to_remove in remove_constraints]

        # Used to seed detailed event log change messages
        detailed_changes = {
            "added constraint depends:": [],
            "added constraint conflicts:": [],
            "removed constraint depends:": [],
            "removed constraint conflicts:": [],
        }

        # Add any constraints to reference set
        for (type, items) in additions:
            value = build_relation_value([items])
            new_constraint = set([(type, value)])
            if not constraints.intersection(new_constraint):
                details = detailed_changes["added constraint %s:" % type]
                details.append(value)
            constraints |= new_constraint

        # Remove any constraints from reference set
        for (type, items) in removals:
            value = build_relation_value([items])
            removed_constraint = set([(type, value)])
            if constraints.intersection(removed_constraint):
                details = detailed_changes["removed constraint %s:" % type]
                details.append(value)
            constraints -= removed_constraint
        if not constraints:
            raise EmptyPackageProfileError(
                "Package profiles must have at least one package constraint."
            )

        for type, (package, operator, version) in additions:
            self._add_package_constraint(type, package, operator, version)
        for type, (package, operator, version) in removals:
            self._remove_package_constraint(type, package, operator, version)
        return detailed_changes

    @requires("EditPackageProfile")
    def on_create(self, creator):
        """Called when the profile is created.."""
        self.package = self.to_package()

    @requires("EditPackageProfile")
    def on_change(self):
        """Called when the profile is changed."""
        self.version += 1
        self.modification_time = datetime.utcnow()
        for activity in self.get_pending_activities(
            statuses=[BLOCKED, QUEUED, SCHEDULED]
        ):
            activity.cancel()
        self.package = self.to_package()
        self.apply()

    def to_package(self):
        """
        Convert a package profile into a package.

        Will add a package to the global market if its not already
        present.
        """
        market = PackageMarket()
        package_hash = market.add(self.to_skeleton(), binary=self.to_binary())
        return market.get(package_hash)

    def add_computers_with_ids(self, computer_ids):
        computer_ids = super(PackageProfile, self).add_computers_with_ids(computer_ids)

        if computer_ids:
            self.apply(computer_ids=computer_ids)
        return computer_ids

    def remove_computers_with_ids(self, computer_ids):
        computer_ids = super(PackageProfile, self).remove_computers_with_ids(
            computer_ids
        )

        if computer_ids:
            from canonical.landscape.model.package.alerts import PackageProfilesAlert

            self.account.check_rules([PackageProfilesAlert], computer_ids=computer_ids)

        return computer_ids

    @requires("EditPackageProfile")
    def remove(self):
        super(PackageProfile, self).remove()

    @requires("ViewPackageProfile")
    def get_csv_data(self, include_header=True):
        """
        Return data from the current L{PackageProfile}.
        @param include_header: If C{False} remove header row from CSV output.
        """
        csv_data = []

        csv_data.append(
            (self.id, self.name, self.title, "DESCRIPTION", self.description, "", "")
        )
        csv_data.append(
            (self.id, self.name, self.title, "CREATIONTIME", self.creation_time, "", "")
        )
        csv_data.append(
            (
                self.id,
                self.name,
                self.title,
                "MODIFICATIONTIME",
                self.modification_time,
                "",
                "",
            )
        )
        csv_data.append(
            (self.id, self.name, self.title, "VERSION", self.version, "", "")
        )

        if self.all_computers:
            csv_data.append(
                (self.id, self.name, self.title, "ALLCOMPUTERS", "TRUE", "", "")
            )
        else:
            csv_data.append(
                (self.id, self.name, self.title, "ALLCOMPUTERS", "FALSE", "", "")
            )

        tags = " ".join(sorted(tag.name for tag in self.get_tags()))
        if tags:
            csv_data.append((self.id, self.name, self.title, "TAGS", tags, "", ""))

        for constraint in self.constraints:
            (kind, package, operator, version) = constraint.parse()

            csv_data.append(
                (
                    self.id,
                    self.name,
                    self.title,
                    kind.upper(),
                    package,
                    operator,
                    version,
                )
            )

        output = StringIO()
        writer = csv.writer(output)
        csv_data = sorted(csv_data)
        if include_header:
            csv_data.insert(0, HEADER_V2.split(","))
        writer.writerows(csv_data)
        return output.getvalue()

    def get_state(self):
        """@return a dict with general state of the package profile."""
        info = ("id", "name", "title", "description")

        state = dict((field, getattr(self, field)) for field in info)
        state["creation_time"] = format_datetime(self.creation_time)
        state["modification_time"] = format_datetime(self.modification_time)
        state["version"] = str(self.version)
        state["constraints"] = list(
            (relation_type_to_string(constraint.type), constraint.value)
            for constraint in self.constraints
        )
        state["computers"] = {}
        state["computers"]["constrained"] = sorted(
            list(computer.id for computer in self.access_children)
        )
        state["computers"]["non-compliant"] = sorted(
            list(self.get_broken_computer_ids())
        )
        state["tags"] = list(tag.name for tag in self.get_tags())
        state["all_computers"] = self.all_computers
        state["access_group"] = self.access_context.name

        return state

    def to_skeleton(self):
        """Convert a package profile into a package skeleton.

        @return: A L{PackageSkeleton} object matching the meta-data and
            constraints of the profile.
        """
        name = "landscape-profile-%s-%s" % (
            self.account.name.replace("_", "-"),
            self.name,
        )
        version = str(self.version)
        skeleton = PackageSkeleton(DEB_PACKAGE, name, version)
        skeleton.summary = "Landscape meta-package for profile %s" % self.name
        skeleton.description = self.description
        skeleton.size = 32768
        skeleton.installed_size = 64
        skeleton.section = "metapackages"

        skeleton.add_relation(DEB_NAME_PROVIDES, "%s = %s" % (name, version))
        skeleton.add_relation(DEB_UPGRADES, "%s < %s" % (name, version))

        for constraint in self.constraints:
            skeleton.add_relation(constraint.type, constraint.value)

        return skeleton

    def _get_package(self):
        if self.package_id is None:
            return None
        if not self._package:
            store = get_package_store()
            self._package = store.get(Package, self.package_id)
        return self._package

    def _set_package(self, package):
        self.package_id = package.id
        self._package = package

    def _reset_profile(self):
        """Reset our L{Package} so it's reloaded at each transaction."""
        self._package = None

    __storm_invalidated__ = _reset_profile

    package = property(_get_package, _set_package)

    def to_binary(self):
        """Return a binary Debian package built against the profile."""
        skeleton = self.to_skeleton()
        deb = DebFile()
        deb.field("Package", str(skeleton.name))
        deb.field("Version", str(skeleton.version))
        deb.field("Architecture", "all")
        deb.field("Maintainer", "Landscape Team " "<landscape-team@canonical.com>")
        deb.field("Installed-Size", str(skeleton.installed_size))

        constraints = {DEB_REQUIRES: [], DEB_CONFLICTS: []}
        for type, value in skeleton.relations:
            entry = str(value)
            # Operators > and < are replaced by >> and << in control files.
            entry = entry.replace(" > ", " >> ")
            entry = entry.replace(" < ", " << ")
            # If we have a versioned constraint, wrap the version between
            # parenthesis, for example "foo <= 1.0" becomes "foo (<= 1.0)"
            if " " in entry:
                entry = entry.replace(" ", " (", 1) + ")"
            if type in constraints:
                constraints[type].append(entry)

        if constraints[DEB_REQUIRES]:
            deb.field("Depends", ", ".join(constraints[DEB_REQUIRES]))

        if constraints[DEB_CONFLICTS]:
            deb.field("Conflicts", ", ".join(constraints[DEB_CONFLICTS]))

        deb.add(FileInfo("./"))
        deb.add(FileInfo("./usr/"))
        deb.add(FileInfo("./usr/share/"))
        deb.add(FileInfo("./usr/share/doc/"))
        deb.add(FileInfo("./usr/share/doc/%s/" % str(skeleton.name)))
        deb.add(
            FileInfo(
                "./usr/share/doc/%s/README" % str(skeleton.name),
                b"This is a meta-package generated by Landscape, "
                b"please use the Landscape web UI to remove this "
                b"package\n.",
            )
        )

        deb.field("Section", str(skeleton.section))
        deb.field("Priority", "optional")
        description = "%s\n%s" % (
            skeleton.summary,
            self._format_description(skeleton.description),
        )
        deb.field("Description", description)

        return deb.pack()

    def _format_description(self, description):
        """Format a description for inclusion in a Debian control file.

        The description in a Debian control file must be indented by one
        space and blank lines must be represented as a dot.

        @return: A formatted description suitable for use with a L{DebFile}.
        """
        lines = []
        for line in description.split("\n"):
            if not line:
                line = "."
            lines.append(" " + line)
        return "\n".join(lines)

    @requires("EditPackageProfile")
    def apply(self, computer_ids=None):
        """Create a group activity for installing the profile meta-package.

        @param computer_ids: Optionally a list of L{Computer.id}s to apply the
            profile on. If C{None} the profile will be installed on all
            associated computers.
        @return: A group L{Activity} having a L{ChangePackageProfilesRequest}
            request child for each computer in the profile that doesn't have
            the meta-package installed.
        """
        computers_by_id = dict(
            (computer.id, computer) for computer in self.get_computers(computer_ids)
        )
        computer_ids = set(computers_by_id)
        computer_ids_with_installed = set(
            get_computer_ids_with_installed_package(computer_ids, self.package.id)
        )
        # exclude computers with packages already installed
        computer_ids -= computer_ids_with_installed
        if not computer_ids:
            return None

        # exclude computers with a pending activity for applying the profile
        computer_ids -= set(
            self.get_pending_activities(
                computer_ids=computer_ids, values=AccountComputerTwin.id
            )
        )
        if not computer_ids:
            return None

        selection = ComputerSelection(self.account)
        selection.select_computers(computer_ids)
        package_management = PackageManagement(selection)

        package_changes = []
        market = PackageMarket()
        for computer_id in computer_ids:
            computer_packages = computers_by_id[computer_id].get_packages()

            install = [self.package.id]
            # If previous versions of the profile are installed, ask for them
            # to be removed
            [installed] = computer_packages.get_packages("installed")
            remove = [
                upgraded
                for upgrades, upgraded in market.get_upgrade_relations(
                    install, installed
                )
            ]

            package_changes.append((computer_id, install, remove, (), ()))

        activity = package_management.change_packages(
            None, package_changes, package_profile=self
        )

        from canonical.landscape.model.package.alerts import PackageProfilesAlert

        self.account.check_rules([PackageProfilesAlert], computer_ids=computer_ids)

        return activity

    @requires("ViewPackageProfile")
    def get_pending_root_activities(self, statuses=None):
        """
        Return all the root activities which have at least one child currently
        applying the profile.

        @param statuses: if provided, a list of statuses to filter the pending
            activities.
        @type statuses: C{list}
        """
        result = self.get_pending_activities(
            statuses=statuses, values=Activity.parent_id
        )
        parent_ids = result.config(distinct=True)
        collection = ActivityParentCollection(self.account.store, ids=parent_ids)
        return collection.find()

    @requires("ViewPackageProfile")
    def get_pending_activities(self, computer_ids=None, statuses=None, values=Activity):
        """Return the activities installing the current version of the profile.

        @param computer_ids: Return only L{Activity}s involving L{Computer}s
            with these ids.
        @param values: Return only the given storm properties, this is a
            workaround needed because of L{ResultSet.union}.
        """
        if computer_ids is None:
            computer_ids = self.computer_group.computers.values(AccountComputerTwin.id)

        if statuses is None:
            statuses = (QUEUED, BLOCKED, SCHEDULED, INPROGRESS)
        statuses = set(statuses).intersection(PENDING_STATUSES)

        collection = ActivityChildCollection(
            self.account.store,
            account=self.account,
            statuses=statuses,
            info_classes=[ChangePackageProfilesRequest],
            entity_class=AccountComputerTwin,
            entity_ids=computer_ids,
        )
        collection = collection.select(values)
        result = collection.find()
        return result.find(
            ChangePackageProfilesRequest.activity_id == Activity.id,
            ChangePackageProfilesRequest.package_profile_id == self.id,
        )


PackageProfileConstraint.package_profile = Reference(
    PackageProfileConstraint.package_profile_id, PackageProfile.id
)


ChangePackageProfilesRequest.package_profile = Reference(
    ChangePackageProfilesRequest.package_profile_id, PackageProfile.id
)


@requires_advisor
@implementer(IActivityCollection)
class PackageProfileActivities(object):

    requires(PUBLIC, PRIVATE, ["profile", "account"])

    def __init__(self, profile):
        self.profile = profile
        self.account = profile.account

    @requires(PUBLIC)
    def get_activities(self, pending=True, statuses=None, **kwargs):
        """
        Pass-through to get_pending_root_activities.
        """
        return self.profile.get_pending_root_activities(statuses=statuses)


def get_event_log_message_from_profile(profile):
    """Return a message detailing all attributes of the created profile."""
    prefix = "Package profile created with"
    state = profile.get_state()
    depends = []
    conflicts = []
    for constraint_type, value in state["constraints"]:
        if constraint_type == "depends":
            depends.append(value)
        else:
            conflicts.append(value)
    values = {
        "name": profile.name,
        "title": profile.title,
        "constraint dependencies:": depends,
        "constraint conflicts:": conflicts,
        "description": state["description"],
        "access group": state["access_group"],
    }
    return "%s." % describe_dict(prefix, values, pluralize=False)


def get_package_profiles_with_computer_ids(account, computer_ids):
    """
    Return a result set yielding L{PackageProfile}s in the given account and
    linked to the given computer IDs.
    """
    result = account.store.find(PackageProfile, PackageProfile.account_id == account.id)
    return result.config(distinct=True).find(
        PackageProfile.computer_group_id == ComputerGroupItem.computer_group_id,
        ComputerGroupItem.computer_id.is_in(computer_ids),
    )
