# Copyright 2005-2022 Canonical Limited. All rights reserved.

from builtins import next
from builtins import str
from builtins import range
from past.utils import old_div
from builtins import object
import os
import fnmatch
from itertools import count
from random import randint, random, choice, sample, uniform
from string import ascii_lowercase
from datetime import datetime, timedelta
from collections import defaultdict
from uuid import uuid1
import textwrap
import yaml

from storm.expr import Func

import apt_pkg

from zope.component import getUtility, provideUtility

import transaction

from landscape.lib import bpickle
from landscape.lib.fs import read_file

from canonical.lib.timestamp import utctoday, utcnow
from canonical.lib.text import generate_password

from canonical.landscape.constants import ROOT_DIR
from canonical.landscape.model.configuration import (
    is_standalone,
    set_root_url,
    ISystemConfiguration,
    LANDSCAPE_SYSTEM_EMAIL,
    set_stripe_api_keys,
)
from canonical.landscape.model.feature.flag import is_feature_enabled
from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.model.main.interfaces import ILDSLicenseAuthority
from canonical.landscape.model.main.person import create_person, get_person_by_email
from canonical.landscape.model.main.tag import get_tag_by_name, get_tags
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.package.store import get_package_store
from canonical.landscape.model.main.standalone import (
    AccountLDSLicenseAPI,
    LDSScriptExecutionManagement,
)
from canonical.landscape.model.main.computer import PingServer, ComputerRelationship
from canonical.landscape.model.main.metarelease import (
    create_meta_release,
    update_meta_releases_upgrades,
)
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.package.market import PackageMarket
from landscape.lib.apt.package.skeleton import PackageSkeleton
from canonical.landscape.model.package.usn import USN
from canonical.landscape.model.usn.issue import USNIssue, USNIssueManagement
from canonical.landscape.model.package.management import PackageManagement
from canonical.landscape.model.package.package import (
    DEB_PACKAGE,
    DEB_UPGRADES,
    DEB_NAME_PROVIDES,
    Package,
)
from canonical.landscape.model.resource.cpuusage import CPUUsageManagement
from canonical.landscape.model.resource.ceph import CephUsageManagement
from canonical.landscape.model.resource.swift import SwiftUsageManagement
from canonical.landscape.model.resource.annotation import AnnotationManagement
from canonical.landscape.model.resource.network import NetworkDeviceAPI
from canonical.landscape.model.resource.package import PackageReporterResultAPI
from canonical.landscape.model.resource.rebootrequiredinfo import RebootRequiredInfo
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.users.users import UserSelection, UserManagement
from canonical.landscape.model.script.management import ScriptManagement
from canonical.landscape.model.activity.collection import get_child_activities
from canonical.landscape.model.activity.group import (
    ActivityGroup,
    ADD_TO_GROUPS_SUMMARY,
    CHANGE_PACKAGES_SUMMARY,
    EXECUTE_SCRIPT_SUMMARY,
    RESTART_SUMMARY,
)
from canonical.landscape.model.package.alerts import (
    EsmDisabledAlert,
    PackageUpgradesAlert,
    SecurityUpgradesAlert,
    PackageReporterAlert,
)
from canonical.landscape.model.computer.alerts import (
    ComputerOfflineAlert,
    ComputerRebootAlert,
)
from canonical.landscape.model.package.activities import ChangePackagesRequest
from canonical.landscape.model.package.profile import PackageProfilesManagement
from canonical.landscape.model.computer.removal import RemovalProfileManagement
from canonical.landscape.model.package.autoupgrade import UpgradeProfileManagement
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.testing.fakes import FakeLicenseAuthority
from canonical.landscape.model.knowledge.article import Article
from canonical.landscape.model.knowledge.category import Category
from canonical.landscape.model.knowledge.store import get_knowledge_store
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.landscape.model.access.role import get_role_registry
from canonical.landscape.sample import kb_articles
from canonical.landscape.model.gpg.management import GPGKeyManagement
from canonical.landscape.model.account.customgraph import CustomGraphManagement
from canonical.landscape.model.account.savedsearch import SavedSearchManagement
from canonical.landscape.model.account.invitation import create_invitation
from canonical.landscape.model.account.alerts import StandaloneScriptAlert
from canonical.landscape.model.billing.vat import add_vat_rate
from canonical.landscape.model.billing.alerts import FailedBillingAlert
from canonical.landscape.model.resource.hardware import (
    store_hardware_for_computer,
    HardwareInfoManagement,
)
from canonical.landscape.model.resource.cloud_instance_metadata import (
    CloudInstanceMetadataManagement,
)
from canonical.testing.gpg import (
    SAMPLE_KEY_FINGERPRINT,
    SAMPLE_SECRET_KEY_FINGERPRINT,
    SAMPLE_KEY_MATERIAL,
    SAMPLE_SECRET_KEY_MATERIAL,
    SAMPLE_SECRET_KEY_PUBLIC_MATERIAL,
)
from canonical.landscape.sample.profiles import (
    SampleAccountProfileDevel,
    SampleAccountProfileStandalone,
    SampleComputerProfileCommon,
    SampleComputerProfileUpside,
    SampleComputerProfileUtility,
    SampleComputerProfileJuju,
)
from canonical.landscape.model.main.accesscontext import get_access_context
from canonical.landscape.model.billing.cost import set_utility_cost
from canonical.landscape.model.billing.management import BillingManagement

apt_pkg.init()

DEFAULT_ACCEPTED_MESSAGE_TYPES = [
    "add-group",
    "add-group-member",
    "add-user",
    "change-packages",
    "custom-graph-add",
    "custom-graph-remove",
    "edit-group",
    "edit-user",
    "execute-script",
    "hold-user",
    "package-ids",
    "remove-group",
    "remove-group-member",
    "remove-user",
    "shutdown",
    "signal-process",
    "unhold-user",
    "accepted-types",
    "registration",
    "resynchronize",
    "set-id",
    "set-intervals",
    "unknown-id",
]


SAMPLE_FILES_PATH = os.path.join(ROOT_DIR, "canonical", "landscape", "sample")

# These belong to an actual Stripe account used for testing purposes
SAMPLE_STRIPE_PUBLIC_KEY = "pk_test_sAspqwMrOOW5mMpkoPGXTolW00oIgvDpR2"
SAMPLE_STRIPE_SECRET_KEY = "sk_test_Bagidoxx6MqNbNnF8PV4veEU00NiYcz6Zn"

GB = 1024 ** 3


def postcommit(func):
    def inner(*args, **kwargs):
        result = func(*args, **kwargs)
        transaction.commit()
        return result

    return inner


class SampleData(object):
    @postcommit
    def build_knowledge_base(self):
        """Create sample data for the knowledge base."""
        store = get_knowledge_store()
        store.add(Category("Hardware"))
        feature = store.add(Category("Feature"))
        store.add(Category("Debug"))
        ubuntu = store.add(Category("Ubuntu"))
        landscape = store.add(Category("Landscape"))

        article1 = Article("Article 1", kb_articles.article1, True)
        store.add(article1)
        article1.categories.add(feature)
        article1.categories.add(landscape)
        article1.categories.add(ubuntu)

        article2 = Article("Article 2", kb_articles.article2, True)
        store.add(article2)
        article2.categories.add(feature)

        article3 = Article("Article 3", kb_articles.article3)
        store.add(article3)
        article3.categories.add(feature)

    @postcommit
    def build_ping_server(self):
        store = get_main_store()
        ping_server = PingServer()
        ping_server.startup_time = datetime.utcnow() - timedelta(minutes=30)
        ping_server.last_update_time = datetime.utcnow()
        store.add(ping_server)

    @postcommit
    def build_utility_pricing(self):
        """Add utility prices for each computer type."""
        today = utctoday()
        before = today - timedelta(days=60)
        set_utility_cost("physical", before, 200)
        set_utility_cost("vm", before, 20)
        set_utility_cost("container", before, 2)
        set_utility_cost("physical", today, 100)
        set_utility_cost("vm", today, 10)
        set_utility_cost("container", today, 1)

    @postcommit
    def add_vat_rates(self):
        """Add VAT rates for a few EU countries"""
        add_vat_rate("GB", rate=0.2)
        add_vat_rate("IT", rate=0.22)
        add_vat_rate("NL", rate=0.21)

    @postcommit
    def set_sample_stripe_api_keys(self):
        """Set Stripe API keys"""
        set_stripe_api_keys(SAMPLE_STRIPE_PUBLIC_KEY, SAMPLE_STRIPE_SECRET_KEY)


class SampleAccount(object):
    def __init__(self, account):
        self.account = account

    @postcommit
    def build_invitation(self, name, email, salesforce_key=None, role="Owner"):
        role_registry = get_role_registry([self.account.id])
        return create_invitation(
            self.account,
            name,
            email,
            salesforce_key=salesforce_key,
            roles=[role_registry.get(role)],
        )

    @postcommit
    def build_pending_computer(self, *args, **kwargs):
        return self.account.create_pending_computer(*args, **kwargs)


class SampleComputer(object):

    # Seconds between sample data points
    interval = 300
    distribution = None

    def __init__(self, computer):
        self.computer = computer
        if self.computer.distribution:
            self.distribution = self.computer.distribution.code_name

    @postcommit
    def build_free_disk_space(
        self, start_time, stop_time, mount_point, device, filesystem, total
    ):
        mount_info = self.computer.create_mount_info(
            device, mount_point, filesystem, total
        )
        time = start_time
        min_value = delta = old_div(total, 10)
        max_value = total - delta
        free_space = self._random_int(
            delta * 2, min_range=min_value, max_range=max_value
        )
        while time <= stop_time:
            mount_info.add_free_space(time, free_space)
            free_space = self._random_int(
                delta, previous=free_space, min_range=min_value, max_range=max_value
            )
            time += self.interval

    total_memory = 1024
    total_swap = 1024

    @postcommit
    def build_free_memory_and_swap(self, start_time, stop_time):
        self.computer.total_memory = self.total_memory
        self.computer.total_swap = self.total_swap
        time = start_time
        min_value = delta = 256
        max_memory = self.total_memory - delta
        max_swap = self.total_swap - delta

        # Initial value starts with a wider range
        free_memory = self._random_int(
            delta * 2, min_range=min_value, max_range=max_memory
        )
        free_swap = self._random_int(delta * 2, min_range=min_value, max_range=max_swap)

        while time <= stop_time:
            self.computer.create_memory_info(time, free_memory, free_swap)
            free_memory = self._random_int(
                delta, previous=free_memory, min_range=min_value, max_range=max_memory
            )
            free_swap = self._random_int(
                delta, previous=free_swap, min_range=min_value, max_range=max_memory
            )
            time += self.interval

    @postcommit
    def build_network_traffic(
        self, start_time, stop_time, interface, interface_speed=None
    ):
        api = NetworkDeviceAPI(self.computer)
        ip = choice(list(range(60, 65)))
        api.create_device(
            interface,
            b"192.168.0.%d" % ip,
            b"00:1e:c9:6c:d7:ac",
            b"192.168.0.255",
            b"255.255.255.0",
            1,
            speed=interface_speed,
        )
        time = start_time
        min_value = 100000
        max_value = 1000000
        delta = 15000

        value_in = self._random_int(
            delta=delta, min_range=min_value, max_range=max_value
        )
        value_out = self._random_int(
            delta=delta, min_range=min_value, max_range=max_value
        )
        while time <= stop_time:
            api.create_traffic(interface, time, value_in, value_out)
            value_in = self._random_int(
                delta=delta, previous=value_in, min_range=min_value, max_range=max_value
            )
            value_out = self._random_int(
                delta=delta,
                previous=value_out,
                min_range=min_value,
                max_range=max_value,
            )
            time += self.interval

    @postcommit
    def build_load_averages(self, start_time, stop_time):
        time = start_time
        load = randint(0, 450)
        while time <= stop_time:
            self.computer.create_load_average(time, load / 100.0)
            load = self._random_int(10, previous=load, min_range=0, max_range=450)
            time += self.interval

    @postcommit
    def build_cpu_usage(self, start_time, stop_time):
        time = start_time
        manager = CPUUsageManagement(self.computer)
        usage = randint(0, 100)
        while time <= stop_time:
            manager.create_cpu_usage(time, usage / 100.0)
            usage = self._random_int(10, previous=usage)
            time += self.interval

    @postcommit
    def build_ceph_usage(
        self, start_time, stop_time, uuid="ecbb8960-0e21-11e2-b495-83a88f44db01"
    ):
        """
        Build some Ceph usage info for the computer.
        """
        a_week = timedelta(days=7).total_seconds()
        time = start_time
        manager = CephUsageManagement(self.computer)

        total = 10240 * GB

        multipliers = [float(x) / 10 for x in sample(list(range(1, 11)), 5)]
        while time <= stop_time:
            week = int(old_div((time - start_time), a_week))
            if old_div(week, 2) == 1:
                total += 2048 * GB
            avail, used = self._get_avail_and_used(multipliers[week], total)
            manager.create_ceph_usage(time, uuid, total, avail, used)
            time += self.interval

    @postcommit
    def build_swift_usage(self, start_time, stop_time):
        """
        Build some Swift usage info for the computer.
        """
        a_week = timedelta(days=7).total_seconds()
        time = start_time
        manager = SwiftUsageManagement(self.computer)

        total = 1024 * GB
        num_devices = 1

        multipliers = [float(x) / 10 for x in sample(list(range(1, 11)), 5)]
        while time <= stop_time:
            week = int(old_div((time - start_time), a_week))
            if old_div(week, 2) == 1:
                num_devices += 1
            devices = ["sd%s" % x for x in ascii_lowercase[1 : num_devices + 1]]
            for device in devices:
                avail, used = self._get_avail_and_used(multipliers[week], total)
                manager.create_swift_usage(time, device, total, avail, used)
            time += self.interval

    @postcommit
    def build_temperatures(self, start_time, stop_time, name, passive=43, critical=87):
        thermal_zone = self.computer.create_thermal_zone(name)
        time = start_time
        min_temp = 50
        max_temp = critical + 2
        delta = 5
        temp = min(100, randint(0, critical + 2))
        temp = self._random_int(delta=delta, min_range=min_temp, max_range=max_temp)

        while time <= stop_time:
            thermal_zone.add_temperature(time, temp)
            temp = self._random_int(
                delta=delta, previous=temp, min_range=min_temp, max_range=max_temp
            )
            time += self.interval

    @postcommit
    def build_users_and_groups(self, users=None, stock_users=None, stock_groups=None):
        """Create stock groups and, optionally, users.

        L{AddUser} and L{CreateGroup} operations are created, the same
        as they would if this data had been reported by a client.

        @param users: Optionally, a list of C{(uid, username, name,
            enabled, location, home, work)} tuples.  Any users
            provided in this list are created in addition to stock
            users and groups.
        @param stock_users: Optionally, a list of C{(uid, username,
            name, enabled, location, home, work)} tuples representing
            stock system users.  If this value is None a set of stock
            system users similar to the set of system users for a
            standard Ubuntu installation will be created.
        @param stock_groups: Optionally, a list of C{(gid, groupname)}
            tuples representing stock groups to create.  If this value
            is None a set of stock groups similar to the set of groups
            for a standard Ubuntu installation will be created.
        """
        new_users = []
        if users:
            # We add users to a group with the same gid as uid
            for uid, username, name, enabled, location, home, work in users:
                user = self.computer.create_user(
                    uid, username, uid, name, enabled, location, home, work
                )
                group = self.computer.create_group(uid, username)
                group.associate_member(user)
                new_users.append(user)

        memberships = set(
            [
                "adm",
                "dialout",
                "cdrom",
                "floppy",
                "audio",
                "dip",
                "video",
                "plugdev",
                "netdev",
                "scanner",
                "lpadmin",
                "powerdev",
                "admin",
            ]
        )
        groups = {}
        if stock_groups is None:
            stock_groups = [
                (0, "root"),
                (1, "daemon"),
                (2, "bin"),
                (3, "sys"),
                (4, "adm"),
                (5, "tty"),
                (6, "disk"),
                (7, "lp"),
                (8, "mail"),
                (9, "news"),
                (10, "uucp"),
                (12, "man"),
                (13, "proxy"),
                (15, "kmem"),
                (20, "dialout"),
                (21, "fax"),
                (22, "voice"),
                (24, "cdrom"),
                (25, "floppy"),
                (26, "tape"),
                (27, "sudo"),
                (29, "audio"),
                (30, "dip"),
                (33, "www-data"),
                (34, "backup"),
                (37, "operator"),
                (38, "list"),
                (39, "irc"),
                (40, "src"),
                (41, "gnats"),
                (42, "shadow"),
                (43, "utmp"),
                (44, "video"),
                (45, "sasl"),
                (46, "plugdev"),
                (50, "staff"),
                (60, "games"),
                (100, "users"),
                (101, "dhcp"),
                (102, "netdev"),
                (103, "syslog"),
                (104, "klog"),
                (105, "scanner"),
                (106, "nvram"),
                (107, "messagebus"),
                (108, "ssl-cert"),
                (109, "lpadmin"),
                (110, "crontab"),
                (111, "ssh"),
                (112, "avahi-autoipd"),
                (113, "avahi"),
                (114, "haldaemon"),
                (115, "powerdev"),
                (116, "gdm"),
                (117, "admin"),
                (118, "slocate"),
                (119, "fuse"),
                (120, "postgres"),
            ]

        for gid, groupname in stock_groups:
            groups[groupname] = group = self.computer.create_group(gid, groupname)
            if groupname in memberships:
                for user in new_users:
                    group.associate_member(user)

        stock_system_groups = None
        if stock_users is None:
            stock_users = [
                (0, "root", 0, "root", False, None, None, None),
                (1, "daemon", 1, "daemon", True, None, None, None),
                (2, "bin", 2, "bin", True, None, None, None),
                (3, "sys", 3, "sys", True, None, None, None),
                (4, "sync", 65534, "sync", True, None, None, None),
                (5, "games", 60, "games", True, None, None, None),
                (6, "man", 12, "man", True, None, None, None),
                (7, "lp", 7, "lp", True, None, None, None),
                (8, "mail", 8, "mail", True, None, None, None),
                (9, "news", 9, "news", True, None, None, None),
                (10, "uucp", 10, "uucp", True, None, None, None),
                (13, "proxy", 13, "proxy", True, None, None, None),
                (33, "www-data", 33, "www-data", True, None, None, None),
                (34, "backup", 34, "backup", True, None, None, None),
                (38, "list", 38, "Mailing List Manager", True, None, None, None),
                (39, "irc", 39, "irc", True, None, None, None),
                (
                    41,
                    "gnats",
                    41,
                    "Gnats Bug-Reporting System (admin)",
                    True,
                    None,
                    None,
                    None,
                ),
                (65534, "nobody", 65534, "nobody", True, None, None, None),
                (100, "dhcp", 100, "", True, None, None, None),
                (101, "syslog", 101, "", True, None, None, None),
                (102, "klog", 102, "", True, None, None, None),
                (103, "messagebus", 103, "", True, None, None, None),
                (
                    104,
                    "avahi-autoipd",
                    111,
                    "Avahi autoip daemon",
                    True,
                    None,
                    None,
                    None,
                ),
                (105, "avahi", 105, "Avahi mDNS daemon", True, None, None, None),
                (106, "cupsys", 109, "", True, None, None, None),
                (
                    107,
                    "haldaemon",
                    107,
                    "Hardware abstraction layer",
                    True,
                    None,
                    None,
                    None,
                ),
                (108, "hplip", 7, "HPLIP system user", True, None, None, None),
                (109, "gdm", 116, "Gnome Display Manager", True, None, None, None),
            ]

            stock_system_groups = {
                "cupsys": ["lp", "dialout", "scanner", "ssl-cert"],
                "haldaemon": ["cdrom", "floppy", "plugdev", "powerdev"],
                "hplip": ["scanner"],
            }

        # We use the uid as the gid for the primary_group
        for uid, username, gid, name, enabled, location, home, work in stock_users:
            user = self.computer.create_user(
                uid, username, gid, name, enabled, location, home, work
            )
            groupnames = stock_system_groups.get(username)
            if groupnames:
                group = groups[groupname]
                group.associate_member(user)

    @postcommit
    def build_hardware(self, hardware):
        filename = os.path.join(SAMPLE_FILES_PATH, "hardware-%d.bpickle" % hardware)
        data = read_file(filename)
        store_hardware_for_computer(self.computer, bpickle.loads(data))

    @postcommit
    def build_annotation(self, account, person, data):
        management = AnnotationManagement(account, person)
        management.add_annotation([self.computer.id], data)

    @postcommit
    def build_cloud_instance_metadata(self, ami_id, instance_type, instance_id):
        metadata_manager = CloudInstanceMetadataManagement(self.computer)
        metadata_manager.set_instance_metadata(
            ami_id=ami_id, instance_id=instance_id, instance_type=instance_type
        )

    def build_packages(self, available_ids, installed_ids, upgrade_ids, held_ids):
        self.computer.get_packages().update(
            available=available_ids,
            installed=installed_ids,
            upgrades=upgrade_ids,
            held=held_ids,
        )

    def build_script_activities(self, parent):
        self.computer.request_execute_script(
            parent, "Shutdown", "sh", "shutdown -r +5", "john", 300
        )

    @postcommit
    def build_active_processes(self, start_time, processes=None):
        data = processes
        if data is None:
            data = [
                (1, 0, 0, "init", 0.0, 2948, b"S"),
                (2, 0, 0, "kthreadd", 0.0, 0, b"S"),
                (3, 0, 0, "migration/0", 0.0, 0, b"S"),
                (4, 0, 0, "ksoftirqd/0", 0.0, 0, b"S"),
                (5, 0, 0, "watchdog/0", 0.0, 0, b"S"),
                (6, 0, 0, "migration/1", 0.0, 0, b"S"),
                (7, 0, 0, "ksoftirqd/1", 0.0, 0, b"R"),
                (8, 0, 0, "watchdog/1", 0.0, 0, b"S"),
                (9, 0, 0, "events/0", 0.0, 0, b"S"),
                (10, 0, 0, "events/1", 0.0, 0, b"S"),
                (11, 0, 0, "khelper", 0.0, 0, b"S"),
                (5133, 0, 0, "X", 3.1, 156816, b"R"),
                (2071, 0, 0, "ksuspend_usbd", 0.0, 0, b"S"),
                (2072, 0, 0, "khubd", 0.0, 0, b"S"),
                (6171, 1000, 1000, "firefox-bin", 8.2, 234732, b"S"),
                (12828, 1000, 1000, "python", 2.1, 17396, b"R"),
                (6173, 1000, 1000, "gnome-terminal", 0.0, 63828, b"S"),
                (31, 0, 0, "kblockd/0", 0.0, 0, b"S"),
                (32, 0, 0, "kblockd/1", 0.0, 0, b"S"),
                (33, 0, 0, "kacpid", 0.0, 0, b"S"),
                (34, 0, 0, "kacpi_notify", 0.0, 0, b"S"),
                (6181, 1000, 1000, "sh", 0.0, 1752, b"S"),
                (177, 0, 0, "pdflush", 0.0, 0, b"S"),
                (5127, 0, 0, "gdm", 0.0, 13228, b"S"),
                (178, 0, 0, "pdflush", 0.0, 0, b"S"),
                (5168, 0, 0, "sshd", 0.1, 5280, b"S"),
                (6197, 1000, 1000, "thunderbird-bin", 2.2, 201140, b"S"),
                (6199, 1000, 1000, "emacs-snapshot-gtk", 2.1, 38596, b"S"),
                (6200, 1000, 1000, "gnome-pty-helper", 2.1, 2664, b"S"),
                (6201, 1000, 1000, "bash", 0.5, 6220, b"S"),
                (5130, 0, 1000, "gdm", 0.2, 13672, b"S"),
                (6206, 1000, 1000, "python", 0.2, 40796, b"S"),
                (7691, 1000, 1000, "aspell", 0.8, 8628, b"S"),
                (6240, 1000, 1000, "bash", 0.14, 6220, b"S"),
                (6155, 1000, 1000, "sh", 0.0, 1756, b"S"),
                (5197, 0, 0, "cupsd", 0.1, 5800, b"S"),
                (5731, 1000, 1000, "gnome-keyring-daemon", 0.1, 13308, b"S"),
                (5716, 0, 0, "khidpd_05ac022c", 0.0, 0, b"S"),
                (5728, 107, 114, "event4", 0.0, 2160, b"S"),
                (2145, 0, 0, "ata/0", 0.0, 0, b"S"),
                (2146, 0, 0, "ata/1", 0.0, 0, b"S"),
                (2147, 0, 0, "ata_aux", 0.0, 0, b"S"),
                (5734, 1000, 1000, "x-session-manager", 0.0, 27952, b"S"),
                (4714, 0, 0, "kondemand/0", 0.0, 0, b"S"),
                (4715, 0, 0, "kondemand/1", 0.0, 0, b"S"),
                (2161, 0, 0, "khpsbpkt", 0.0, 0, b"S"),
                (5244, 111, 120, "postgres", 0.0, 38000, b"S"),
                (5249, 111, 120, "postgres: writer process", 0.0, 38000, b"S"),
                (5250, 111, 120, "postgres: stats collector process", 0.0, 9860, b"S"),
                (2667, 0, 0, "udevd", 0.0, 3040, b"S"),
                (6167, 1000, 1000, "sh", 0.0, 1756, b"S"),
                (5776, 1000, 1000, "gpg-agent", 0.0, 4008, b"S"),
                (5778, 1000, 1000, "ssh-agent", 0.0, 4432, b"S"),
                (5780, 1000, 1000, "gconfd-2", 0.0, 7088, b"S"),
                (149, 0, 0, "kseriod", 0.2, 0, b"S"),
                (5786, 1000, 1000, "seahorse-agent", 0.1, 30012, b"S"),
                (5792, 1000, 1000, "dbus-daemon", 0.0, 2908, b"S"),
                (5794, 1000, 1000, "gnome-settings-daemon", 0.1, 38868, b"S"),
                (11940, 1000, 1000, "bash", 0.2, 6224, b"S"),
                (5830, 1000, 1000, "bonobo-activation-server", 0.1, 41024, b"S"),
                (5917, 1000, 1000, "evolution-data-server-1.12", 0.1, 67168, b"S"),
                (2225, 0, 0, "scsi_eh_0", 0.0, 0, b"S"),
                (2226, 0, 0, "scsi_eh_1", 0.0, 0, b"S"),
                (179, 0, 0, "kswapd0", 0.0, 0, b"S"),
                (3764, 0, 0, "scsi_eh_4", 0.0, 0, b"S"),
                (3765, 0, 0, "usb-storage", 0.0, 0, b"S"),
                (5817, 1000, 1000, "gnome-screensaver", 0.0, 16784, b"S"),
                (5819, 1000, 1000, "gnome-panel", 0.8, 47232, b"S"),
                (5821, 1000, 1000, "nautilus", 0.0, 87200, b"S"),
                (5825, 1000, 1000, "gnome-volume-manager", 0.0, 19676, b"S"),
                (4806, 101, 102, "syslogd", 0.1, 1916, b"S"),
                (5833, 1000, 1000, "gnome-vfs-daemon", 0.9, 8748, b"S"),
                (5834, 1000, 1000, "vino-session", 0.0, 19556, b"S"),
                (5837, 1000, 1000, "bluetooth-applet", 0.0, 26364, b"S"),
                (5838, 1000, 1000, "gaim", 0.0, 68480, b"S"),
                (5841, 1000, 1000, "update-notifier", 0.0, 35176, b"S"),
                (5848, 1000, 1000, "evolution-alarm-notify", 0.1, 69192, b"S"),
                (2265, 0, 0, "scsi_eh_2", 0.2, 0, b"S"),
                (2266, 0, 0, "scsi_eh_3", 0.2, 0, b"S"),
                (5071, 0, 0, "hald-addon-cpufreq", 0.3, 3156, b"S"),
                (5854, 1000, 1000, "nm-applet", 0.0, 44948, b"S"),
                (5856, 1000, 1000, "python", 0.0, 23008, b"S"),
                (5857, 1000, 1000, "gnome-cups-icon", 0.0, 58748, b"S"),
                (6885, 1000, 1000, "notification-daemon", 0.0, 31860, b"S"),
                (230, 0, 0, "aio/0", 0.0, 0, b"S"),
                (231, 0, 0, "aio/1", 0.0, 0, b"S"),
                (5866, 1000, 1000, "gnome-power-manager", 0.1, 22732, b"S"),
                (2293, 0, 0, "knodemgrd_0", 0.2, 0, b"S"),
                (5366, 0, 0, "console-kit-daemon", 0.0, 7456, b"S"),
                (4859, 0, 0, "dd", 0.0, 1840, b"S"),
                (4861, 102, 103, "klogd", 0.0, 2496, b"S"),
                (6146, 1000, 1000, "xchat-gnome", 0.0, 45060, b"S"),
                (5900, 1000, 1000, "evolution-exchange-storage", 0.0, 36668, b"S"),
                (5907, 1000, 1000, "mapping-daemon", 0.0, 2664, b"S"),
                (4889, 103, 106, "dbus-daemon", 0.1, 3028, b"S"),
                (5850, 1000, 1000, "trackerd", 0.1, 47884, b"R"),
                (6192, 1000, 1000, "sh", 0.0, 1756, b"S"),
                (5928, 1000, 1000, "trashapplet", 0.0, 0, b"S"),
                (4905, 0, 0, "NetworkManager", 0.2, 29980, b"S"),
                (4919, 0, 0, "NetworkManagerDispatcher", 0.0, 3280, b"S"),
                (5515, 0, 0, "krfcommd", 0.0, 0, b"S"),
                (4932, 0, 0, "system-tools-backends", 0.0, 3084, b"S"),
                (4933, 0, 0, "dbus-daemon", 0.7, 2776, b"S"),
                (
                    5453,
                    105,
                    111,
                    "avahi-daemon: registering [rex.local]",
                    0.0,
                    2736,
                    b"S",
                ),
                (5454, 105, 111, "avahi-daemon: chroot helper", 0.0, 2736, b"S"),
                (4952, 107, 114, "hald", 0.0, 5316, b"S"),
                (4953, 0, 0, "hald-runner", 0.0, 3092, b"S"),
                (5978, 0, 0, "wpa_supplicant", 0.0, 3852, b"S"),
                (5468, 0, 0, "dhcdbd", 0.0, 1992, b"S"),
                (5863, 1000, 1000, "seahorse-agent", 0.0, 33952, b"S"),
                (5996, 1000, 1000, "python", 4.1, 65716, b"S"),
                (5488, 0, 0, "hcid", 0.0, 2924, b"S"),
                (5809, 1000, 1000, "sh", 0.0, 1756, b"S"),
                (6015, 1000, 1000, "gweather-applet-2", 0.0, 46548, b"S"),
                (6016, 100, 101, "dhclient", 0.0, 2420, b"S"),
                (5505, 0, 0, "bluetoothd-service-input", 0.0, 2772, b"S"),
                (5506, 0, 0, "hidd", 0.0, 1928, b"S"),
                (5508, 0, 0, "bluetoothd-service-audio", 0.0, 2844, b"S"),
                (4491, 0, 0, "getty", 0.0, 1696, b"S"),
                (4492, 0, 0, "getty", 0.0, 1696, b"S"),
                (4494, 0, 0, "getty", 0.0, 1692, b"S"),
                (4496, 0, 0, "getty", 0.0, 1692, b"S"),
                (4498, 0, 0, "getty", 0.0, 1696, b"S"),
                (4499, 0, 0, "getty", 0.0, 1692, b"S"),
                (8084, 1000, 1000, "gtk-window-decorator", 0.0, 17360, b"S"),
                (5810, 1000, 1000, "esd", 0.0, 3384, b"S"),
                (2461, 0, 0, "kjournald", 0.0, 0, b"S"),
                (6047, 1000, 1000, "cpufreq-applet", 0.0, 20488, b"S"),
                (6049, 1000, 1000, "mixer_applet2", 0.0, 45304, b"S"),
                (6129, 0, 0, "khidpd_05ac030c", 0.0, 0, b"S"),
                (5548, 0, 0, "atd", 0.0, 1964, b"S"),
                (6045, 1000, 1000, "mono", 0.0, 63200, b"S"),
                (8120, 1000, 1000, "metacity", 0.0, 18900, b"S"),
                (5567, 0, 0, "cron", 0.0, 2336, b"S"),
                (5062, 107, 114, "event1", 0.0, 2160, b"S"),
                (5068, 107, 114, "event2", 0.0, 2164, b"S"),
                (5069, 107, 114, "event3", 0.0, 2164, b"S"),
                (4685, 0, 0, "acpid", 0.0, 2436, b"S"),
                (5072, 107, 114, "acpid.socket", 0.0, 2164, b"S"),
                (5098, 107, 114, "scd0 (every 2 sec)", 0.0, 3264, b"S"),
                (5101, 107, 114, "sdd (every 2 sec)", 0.0, 3264, b"S"),
                (5103, 107, 114, "sdb (every 2 sec)", 0.0, 3264, b"S"),
                (5105, 107, 114, "sde (every 2 sec)", 0.0, 3260, b"S"),
                (5107, 107, 114, "sdc (every 2 sec)", 0.0, 3264, b"S"),
            ]

        for pid, uid, gid, name, cpu_utilisation, vm_size, state in data:
            self.computer.create_active_process(
                pid, uid, gid, name, state, start_time, vm_size, cpu_utilisation
            )

    @postcommit
    def build_reboot_required_info(self, packages):
        reboot_required_info = RebootRequiredInfo(self.computer, packages=packages)
        self.computer.store.add(reboot_required_info)

    @postcommit
    def build_computer_utilisation(self, usage):
        """Set managed hours count for the computer."""
        self.computer.utilisation.managed_hours = usage

    def _get_avail_and_used(self, base_usage, total):
        """
        Return new availalable/used values based on the total and a base
        usage.
        """
        usage = base_usage + uniform(-0.001, 0.001)
        usage = min(1.0, max(0.0, usage))
        used = int(total * usage)
        available = total - used
        return available, used

    def _random_int(self, delta, previous=None, min_range=0, max_range=100):
        """
        Return a random integer value within the provided delta above or below
        the previous value, up to the provided min or max range.

        If the previous value is C{None}, a random value in the range is used.
        """
        if previous is None:
            previous = randint(min_range, max_range)
        delta_value = randint(-delta, delta)
        return max(min_range, min(max_range, previous + delta_value))


def create_stock_sample_data(start_time, stop_time, datatypes):
    """Create sample data suitable for demo and testing purposes.

    The purpose of this recipe is to create sample data that is,
    generally speaking, pretty realistic.  Time range-based data is
    created within the bounds of C{start_time} and C{stop_time}.

    @param start_time: The start timestamp for time range-based sample
        data.  This value must be aligned to the step interval.
    @param stop_time: The stop timestamp for time range-based sample
        data.
    @param datatypes: A dictionary whose keys are strings, each referring
        to a sample datatype to generate data for. The value of the keys
        can be used to pass parameters for the datatype to generate, e.g.
        the number of extra computers.
    """
    data = SampleData()

    data.build_ping_server()
    if not is_standalone():
        data.build_utility_pricing()
        data.add_vat_rates()
        data.set_sample_stripe_api_keys()

    credentials = []
    invitations = []

    # NOTE: the code path below is only taken if --create-lds-account-only
    #       is passed to the schema script (/usr/sbin/landscape-schema) and its
    #       only use case is the 'bootstrap' action of the landscape-server
    #       charm, in order to create the first admin of a freshlyd deployed
    #       OPL. It's not really related to sample or demo data.
    if is_standalone() and datatypes.get("create-lds-account-only", False):
        # It's safe to assume that these fields are there, since the bootstrap
        # action requires them, and the charm code will set them in the command
        # line. See lib/bootstrap.py in the landscape-server charm.
        name = datatypes["admin-name"]
        email = datatypes["admin-email"]
        password = datatypes["admin-password"]
        existing_account = get_account_by_name("standalone")
        if existing_account:
            # It seems we are already bootstrapped, in this case just no-op
            # and return the existing credentials if available.
            person = get_person_by_email(email)
            if person is not None:
                credentials.append(person.get_api_credentials(existing_account))
            return invitations, credentials

        # This call to create_person also creates the standalone account
        # as side effect
        admin = create_person(name, email, password=password)

        standalone_account = get_account_by_name("standalone")
        credentials.append(admin.create_api_credentials(standalone_account))
        return invitations, credentials

    extra_computers_count, extra_computers_data = datatypes.get(
        "extra-computers", (0, False)
    )
    package_test_computers_count = datatypes.get("package-test-computers", 0)
    extra_seats = extra_computers_count + package_test_computers_count
    if "openstack" in datatypes:
        extra_seats += 25  # additional computers for clouds are created

    if is_standalone():
        profile = SampleAccountProfileStandalone()
        if "root-url" in datatypes:
            set_root_url(datatypes["root-url"])
        if "system-email" in datatypes:
            configuration = getUtility(ISystemConfiguration)
            configuration.set(LANDSCAPE_SYSTEM_EMAIL, datatypes["system-email"])
        profile.accounts["standalone"]["registration-key"] = datatypes.get(
            "account_password"
        )
    else:
        profile = SampleAccountProfileDevel()
        profile.accounts["onward"]["seats"] += extra_seats
        profile.accounts["onward"]["registration-key"] = datatypes.get(
            "account_password"
        )

    resolver = profile.resolver  # We'll use the same resolver everywhere
    profile.create()

    # Create some knowledge base data
    data.build_knowledge_base()

    # John and Jane are partners in Onward, Inc., a successful e-commerce
    # business selling top-of-the-line motivation books!  Every day in this
    # business is a new exciting opportunity to learn and grow!  It's
    # fantastic!
    marge = get_person_by_email("marge@example.com")
    john = get_person_by_email("john@example.com")
    jane = get_person_by_email("jane@example.com")
    bob = get_person_by_email("bob@example.com")
    noel = get_person_by_email("noel@example.com")
    oscar = get_person_by_email("oscar@example.com")
    robert = get_person_by_email("robert@example.com")

    if is_standalone():
        account = get_account_by_name("standalone")
        # Marge is a regular admin in standalone
        credentials.append(marge.get_api_credentials(account))
    else:
        account = get_account_by_name("onward")
        # Marge is an AccountManager in dev mode
        marge.add_global_role("AccountManager")
        marge.create_api_credentials()
        credentials.append(marge.get_api_credentials())

    credentials.append(john.get_api_credentials(account))
    credentials.append(jane.get_api_credentials(account))
    credentials.append(noel.get_api_credentials(account))
    credentials.append(oscar.get_api_credentials(account))
    credentials.append(robert.get_api_credentials(account))

    saved_search_management = SavedSearchManagement(account, john)
    saved_search_management.create_saved_search(
        "package-upgrades", "Computers with upgrades", "alert:package-upgrades"
    )
    saved_search_management.create_saved_search(
        "security-upgrades",
        "Computers with security upgrades",
        "alert:security-upgrades",
    )
    # This search should have no matching results...
    saved_search_management.create_saved_search(
        "local-subnet", "Local subnet", "ip:192.168.11"
    )

    authority = FakeLicenseAuthority()
    provideUtility(authority, ILDSLicenseAuthority)

    if not is_standalone():
        api = AccountLDSLicenseAPI(account)
        expiration_date = utctoday() + timedelta(days=365)
        license = api.create_license()
        license_management = LicenseManagement(account)
        license.add_license(
            license_management.create_license("LDSBasicFeatures", 5, expiration_date)
        )
        license.add_license(
            license_management.create_license("ContainerFeatures", 10, expiration_date)
        )
        license.apt_key = "7846c5ce-235c-11e0-970c-60fb42f8d69c"
        license.regenerate_contents()

    new_account = SampleAccount(account)
    new_account.build_pending_computer("Precise Testing Box", None, "precise")
    new_account.build_pending_computer("Other Box 1", None, "precise")
    new_account.build_pending_computer(
        "Other Box 2", None, "precise", access_group="webservers"
    )
    servers = get_access_context(account.access_context, "server")
    desktops = get_access_context(account.access_context, "desktop")

    profile = SampleComputerProfileCommon(account.name)
    profile.resolver = resolver

    # Build extra computers only if we're not adding openstack data, since
    # we want openstack data to fit in the free license.
    if not datatypes.get("openstack", False):
        profile.add_common_extra()

    if extra_computers_count > 0:
        profile.add_extra(extra_computers_count)

    # Return the index of the computer by looking at the secure ID
    def computer_index(computer):
        return int(computer.secure_id.split(b"-")[-1])

    computers = [
        SampleComputer(computer)
        for computer in sorted(profile.create(materialize=True), key=computer_index)
    ]
    license_management = LicenseManagement(account)
    for computer in computers:
        license_management.auto_assign_license_to_computer(computer.computer)

    if not datatypes.get("openstack", False):
        # Skip resource data for the last extra common computer
        computers.pop(8)

    if extra_computers_count and not extra_computers_data:
        # Skip resource data for the extra computers
        computers = computers[:-extra_computers_count]

    computer1, computer2, computer3, computer4, computer5 = computers[:5]
    computer4.build_annotation(
        new_account.account, john, {"Location": "Bldg 3, Floor 2, Cube:104"}
    )
    if len(computers) > 6:
        computers[7].build_cloud_instance_metadata("a-1234", "i-5678", "m1.tiny")

    # add disabled ESM-infra marker to one machine
    computer1.computer.get_packages().update_apt_preferences_files(
        {
            "/etc/apt/preferences.d/ubuntu-esm-infra": textwrap.dedent(
                """
            # Written by ubuntu-advantage-tools
            Package: *
            Pin: release o=UbuntuESM, n=trusty
            Pin-Priority: never"""
            )
        }
    )

    # assume we have a hardware-#.bpickle of sample data for each computer
    sample_count = len(
        fnmatch.filter(os.listdir(SAMPLE_FILES_PATH), "hardware-*.bpickle")
    )

    scheduled_time = datetime.utcnow() + timedelta(hours=3)
    parent = new_account.account.create_activity(
        ActivityGroup, summary=RESTART_SUMMARY, creator=john
    )
    computer5.computer.restart(parent, deliver_after=scheduled_time)
    parent.settle()

    # Add a couple removal profiles
    removal_profiles = RemovalProfileManagement(new_account.account, john)
    removal_profiles.create_removal_profile(
        "Weekday online profile", 3, access_context_name=desktops.name
    )
    removal_profiles.create_removal_profile("Monthly online profile", 30)

    # build extra computers with sample package data for scaling tests
    if package_test_computers_count > 0:
        build_package_test_computers_package_data(
            new_account, package_test_computers_count
        )

    users = [
        (1000, "john", "John Smith", True, None, None, None),
        (
            1001,
            "jane",
            "Jane Robertson",
            True,
            "Vancouver",
            "604-873-1352",
            "778-873-1352",
        ),
    ]
    generate_sample_data(
        start_time, stop_time, sample_count, computers, datatypes, users
    )

    if "user-and-group-activities" in datatypes:
        build_user_and_group_activities(new_account, john, jane, computers[:3])

    if "script-activities" in datatypes:
        build_script_activities(john, computers[1:4])

    computers_with_packages = set(computers[:5])

    juju_computers = ()

    if "packages" in datatypes:
        available = build_packages(
            "gutsy-packages.bpickle",
            computers_with_packages,
            datatypes["package-upgrades"],
            datatypes["multiple-package-upgrades"],
            datatypes["package-usns"],
            datatypes["package-holds"],
        )
        management = PackageProfilesManagement(new_account.account, john)
        profile = management.create_package_profile(
            "web-profile", "This is a profile for Web Servers"
        )
        profile.modify_package_constraints(
            [
                "depends apache2",
                "depends python-twisted > 8.0",
                "depends postgresql-8.3",
                "conflicts mysql-server",
            ]
        )
        profile.on_create(john)
        available.add(profile.package.hash)
        build_repository_profiles(new_account.account, jane)
        build_repository_pockets(new_account.account, jane)

        upgrade_profiles = UpgradeProfileManagement(new_account.account, jane)
        upgrade_profile = upgrade_profiles.create_upgrade_profile(
            "web-daily", "all", "FREQ=HOURLY;BYMINUTE=30"
        )
        upgrade_profile.update_tags([get_tag_by_name("server")])
        if "package-activities" in datatypes:
            build_package_activities(john, computers_with_packages, available)

    if "custom-graph" in datatypes:
        build_custom_graph(new_account, john, computers[:3], start_time, stop_time)

    if "scripts" in datatypes:
        build_scripts(new_account, john)
        script_management = ScriptManagement(new_account.account, john)
        scripts = list(script_management.get_scripts())
        scripts[0].access_context = servers
        scripts[3].access_context = servers

    if "salesforce-users" in datatypes:
        build_salesforce_users(new_account, john, jane, bob)

    for computer in computers[2:5]:
        computer.computer.check_rules(PackageUpgradesAlert)
        computer.computer.check_rules(SecurityUpgradesAlert)

    for computer_wrapper in computers:
        computer = computer_wrapper.computer
        computer.check_rules(ComputerOfflineAlert)
        computer.check_rules(ComputerRebootAlert)
        computer.check_rules(PackageReporterAlert)
        computer.check_rules(EsmDisabledAlert)

    empty = get_tag_by_name("empty")
    for rule in new_account.account.get_rules():
        if rule.info.info_type == PackageUpgradesAlert.info_type:
            rule.update_tags([empty])
        rule.subscribe(john)

    # Add relationship between Windows machine and WSL instance
    windows_computers = []
    wsl_computers = []

    for computer in computers:
        if computer.computer.distribution.code_name == "windows":
            windows_computers.append(computer.computer)

        if "WSL" in computer.computer.title:
            wsl_computers.append(computer.computer)

    relationship = ComputerRelationship(
        computer=wsl_computers[0],
        parent=windows_computers[0],
    )

    # John runs a small side business developing and selling motivationware.
    # He has several test machines he uses to ensure his software runs
    # correctly on all the great versions of Ubuntu Canonical keeps releasing!
    # Bob helps him out from time to time.
    if not is_standalone():
        upside_account = get_account_by_name("upside")
        LicenseManagement(upside_account).create_license("VMFeatures", 5)
        profile = SampleComputerProfileUpside()
        profile.resolver = resolver
        upside_computers = [
            SampleComputer(comp)
            for comp in sorted(profile.create(materialize=True), key=computer_index)
        ]

        license_management = LicenseManagement(account)
        for unlicensed_computer in upside_computers:
            license_management.auto_assign_license_to_computer(
                unlicensed_computer.computer
            )
        users = [
            (1000, "john", "John Smith", True, None, None, None),
            (
                1001,
                "prova\N{LATIN SMALL LETTER J WITH CIRCUMFLEX}o",
                "Juan Prova\N{LATIN SMALL LETTER J WITH CIRCUMFLEX}o",
                True,
                None,
                None,
                None,
            ),
        ]
        generate_sample_data(
            start_time, stop_time, sample_count, upside_computers, datatypes, users
        )

        # Add computers for the utility billing based account.
        utility_account = get_account_by_name("utility")
        utility_account.create_rule(FailedBillingAlert)
        LicenseManagement(upside_account).create_license("BasicFeatures", 5)
        profile = SampleComputerProfileUtility()
        profile.resolver = resolver
        profile.create(materialize=True)
        management = BillingManagement(utility_account)
        management.add_credit(1000000)
        start = utctoday() - timedelta(days=80)
        end = utctoday() - timedelta(days=50)
        payment1 = management.add_payment(start, end, 2000, 0)
        payment1.add_cost("vm", 20, 100, start, end)
        payment1.status = "no-charge"
        start = utctoday() - timedelta(days=50)
        end = utctoday() - timedelta(days=20)
        payment2 = management.add_payment(start, end, 0, 200)
        payment2.add_cost("physical", 200, 100, start, end)
        payment2.status = "paid"
        payment2.stripe_charge_id = "ch_1924oiFtAHAMnKQ7FRYDzUql"
        payment2.payment_date = utcnow() - timedelta(days=10)

    # See: http://changelogs.ubuntu.com/meta-release
    create_meta_release(
        "hardy",
        "Hardy Heron",
        "8.04",
        date=datetime(2008, 5, 24),
        supported=True,
        lts=True,
    )
    create_meta_release(
        "lucid",
        "Lucid Lynx",
        "10.04 LTS",
        supported=True,
        lts=True,
        date=datetime(2010, 4, 29),
    )
    create_meta_release(
        "maverick", "Maverick Meerkat", "10.10", date=datetime(2010, 10, 10)
    )
    create_meta_release(
        "natty", "Natty Narwhal", "11.04", supported=True, date=datetime(2011, 4, 28)
    )
    create_meta_release(
        "oneiric",
        "Oneiric Ocelot",
        "11.10",
        supported=True,
        date=datetime(2011, 10, 13),
    )
    create_meta_release(
        "precise",
        "Precise Pangolin",
        "12.04",
        supported=True,
        lts=True,
        date=datetime(2012, 4, 26),
        upgrade_tool_signature_url="http://signature",
        upgrade_tool_tarball_url="http://tarball",
    )
    # quantal, raring, saucy
    create_meta_release(
        "trusty",
        "Trusty Tahr",
        "14.04",
        supported=True,
        lts=True,
        date=datetime(2014, 4, 17),
        upgrade_tool_signature_url="http://signature",
        upgrade_tool_tarball_url="http://tarball",
    )
    # utopic, vivid, wily
    create_meta_release(
        "xenial",
        "Xenial Xerus",
        "16.04",
        supported=True,
        lts=True,
        date=datetime(2016, 4, 21),
        upgrade_tool_signature_url="http://signature",
        upgrade_tool_tarball_url="http://tarball",
    )
    create_meta_release(
        "yakkety",
        "Yakkety Yak",
        "16.10",
        supported=True,
        date=datetime(2016, 10, 13),
        upgrade_tool_signature_url="http://signature",
        upgrade_tool_tarball_url="http://tarball",
    )
    create_meta_release(
        "zesty",
        "Zesty Zapus",
        "17.04",
        supported=True,
        date=datetime(2017, 4, 13),
        upgrade_tool_signature_url="http://signature",
        upgrade_tool_tarball_url="http://tarball",
    )
    update_meta_releases_upgrades()

    invitations.append(new_account.build_invitation("Joe Wright", "joe@example.com"))

    if not is_standalone():
        sf_key = "00300000000TEDS"
        invitations.append(
            new_account.build_invitation(
                "Ted Support",
                "ted@example.com",
                salesforce_key=sf_key,
                role="SupportUser",
            )
        )
    access_control_management = AccessControlManagement(account)
    [desktop_admin] = access_control_management.get_roles(["DesktopAdmin"])
    invitations.append(
        new_account.build_invitation(
            "Ben Junior", "ben@example.com", role=desktop_admin.key
        )
    )

    # Populate USNIssue table for created computers.
    build_usn_issues()

    # Create LDS cron scripts
    if is_standalone():
        management = LDSScriptExecutionManagement()
        now = datetime.now()
        start_time = now - timedelta(minutes=40)
        end_time = now - timedelta(minutes=35)
        execution = management.get_execution("update_security_db.sh")
        if execution is None:
            management.create_execution(
                "update_security_db.sh",
                30,
                start_time=start_time,
                end_time=end_time,
                success=False,
            )
        else:
            execution.set(start_time=start_time, end_time=end_time, success=False)
        standalone_account = get_account_by_name("standalone")
        standalone_account.check_rules(info_classes=[StandaloneScriptAlert])

    # The invitation is returned here so that calling code can print or log a
    # working URL for it.  We want to avoid producing output here since this
    # code is exercised by automated tests.
    return invitations, credentials


def generate_sample_data(
    start_time,
    stop_time,
    sample_count,
    computers,
    datatypes,
    users,
    fixed_packages=False,
):
    """Generates sample data.

    @param start_time: For time-based data, start data-point.
    @param stop_time: For time-based data, end data-point.
    @param sample_count: Number of sample computer hardware bpickles we
           have.
    @param computers: List of L{Computer} objects to create the data for.
    @param datatypes: Sequence of strings indicating which data to generate.
    @param users: List of tuples of user data in the form (uid, username,
           name, enabled, location, home, work)
    @param fixed_packages: if False, the packages configured for the
    computer will be dependent on the value of the computer id, True
    indicates that the packages configured should always be the same.

    """
    for i, computer in enumerate(computers):
        if "free-disk-space" in datatypes:
            if i % 2:
                computer.build_free_disk_space(
                    start_time, stop_time, "/", "/dev/hda1", "ext3", 150 * 1024
                )
            else:
                computer.build_free_disk_space(
                    start_time, stop_time, "/", "/dev/hda1", "ext3", 78 * 1024
                )
        if "free-memory-and-swap" in datatypes:
            computer.build_free_memory_and_swap(start_time, stop_time)
        if "load-averages" in datatypes:
            computer.build_load_averages(start_time, stop_time)
        if "temperatures" in datatypes:
            computer.build_temperatures(start_time, stop_time, "THM0")
            if i % 2:
                computer.build_temperatures(
                    start_time, stop_time, "THM1", passive=45, critical=94
                )
        if "cpu-usage" in datatypes:
            computer.build_cpu_usage(start_time, stop_time)
        if "users-and-groups" in datatypes:
            computer.build_users_and_groups(users=users)
        if "hardware" in datatypes:
            computer.build_hardware(i % sample_count)
        if "active-processes" in datatypes:
            computer.build_active_processes(start_time)
        if "network-traffic" in datatypes:
            computer.build_network_traffic(start_time, stop_time, b"eth0")
            interface_count = datatypes.get("extra-interfaces", 0)
            if interface_count:
                print("Making {} extra interfaces".format(interface_count))
                for interface_id in range(1, interface_count + 1):
                    computer.build_network_traffic(
                        start_time, stop_time, "eth{}".format(interface_id)
                    )
        if "utilisation" in datatypes:
            computer.build_computer_utilisation(randint(1, 100))
        computer.computer.store.commit()


def build_salesforce_users(onward, john, jane, bob):
    """Create sample Salesforce users."""
    onward.account.create_salesforce_user(john, "00300000000JOHN")
    onward.account.create_salesforce_user(jane, "00300000000JANE")
    onward.account.create_salesforce_user(bob, "003000000000BOBB")


def build_user_and_group_activities(onward, john, jane, computers):
    """Create sample user and group requests.

    @param onward: The L{SampleAccount} for Onward, Inc.
    @param john: The L{Person} for John Smith.
    @param jane: The L{Person} for Jane Robertson.
    @param computers: A list of L{SampleComputer} objects.
    """
    first_computer = computers[0].computer
    parent = onward.account.create_activity(
        ActivityGroup, creator=john, summary=ADD_TO_GROUPS_SUMMARY
    )
    activity = first_computer.request_create_group(parent, "bizdev")
    parent.settle()
    activity.start()
    selection = ComputerSelection(onward.account)
    selection.select_computers([computer.computer for computer in computers])
    management = UserManagement(selection, john)
    parent = management.create_user("joe", "Joe Wright", "pwd")
    for i, activity in enumerate(get_child_activities(parent)):
        if i < 2:
            activity.start()
            if i == 1:
                activity.fail(
                    text="""\
adduser: The user `joe' already exists.
"""
                )
            elif i == 0:
                activity.succeed(
                    text="""\
Adding user `joe' (UID 1001) ...
Done.
"""
                )

    parent = onward.account.create_activity(
        ActivityGroup, creator=jane, summary=ADD_TO_GROUPS_SUMMARY
    )

    create_group_activities = [
        computer.request_create_group(parent, "sales")
        for computer in [c.computer for c in computers]
    ]
    parent.settle()
    for activity in create_group_activities:
        computer = activity.info.computer
        activity.start()
        activity.succeed(
            text="""\
Adding group `sales' (GID 1003) ...
Done.
"""
        )

    # Create sample data for the various kinds of user activities we have.
    management = UserManagement(selection, jane)
    users = []
    for computer in selection:
        users.extend(computer.get_users_by_usernames(["john", "jane"]))
    selection = UserSelection(management, users)
    selection.lock_users()
    selection.unlock_users()
    parent = selection.edit_users(location="321 Forward Blvd.")
    for activity in get_child_activities(parent):
        activity.start()
        activity.fail(text="Forward Blvd. is lame.")

    selection.add_to_groups(["bizdev", "sales"])
    selection.remove_from_groups(["admin"])


def new_package_version(package, version):
    """Build an upgrade version of C{package}."""
    upgrade = PackageSkeleton(package.type, package.name, version)
    upgrade.relations = list(package.relations)
    for type, value in upgrade.relations:
        if type == DEB_UPGRADES or type == DEB_NAME_PROVIDES:
            upgrade.relations.remove((type, value))
            break
    upgrade.add_relation(DEB_UPGRADES, "%s < %s" % (package.name, version))
    upgrade.add_relation(DEB_NAME_PROVIDES, "%s = %s" % (package.name, version))
    upgrade.summary = package.summary
    upgrade.description = package.description
    return upgrade


def build_packages(
    package_data_filename, computers, upgrades, multiple_upgrades, usn_count, holds
):
    """Create sample packages.

    The sample data in C{gutsy-packages.bpickle} includes all installed
    packages from a machine running Ubuntu Server 7.04 (gutsy).  All packages
    that start with 'python' are also included.
    """
    computers = list(computers)  # needs to be a list
    market = PackageMarket()
    available = set()
    installed = set()
    held = set()
    available_upgrade = set()
    package_data_path = os.path.join(os.path.dirname(__file__), package_data_filename)
    package_data = open(package_data_path, "rb").read()
    transaction.commit()
    usn_release_base = datetime(year=2012, month=6, day=1)
    for data in bpickle.loads(package_data):
        skeleton = PackageSkeleton(
            DEB_PACKAGE, data["name"].decode("utf-8"), data["version"].decode("utf-8")
        )
        skeleton.summary = data["summary"]
        skeleton.description = ""
        for relation in data["relations"]:
            relation = (relation[0], relation[1].decode("utf-8"))
            skeleton.add_relation(*relation)
        hash = market.add(skeleton)
        available.add(hash)
        if data["installed"]:
            installed.add(hash)
            if holds:
                holds -= 1
                held.add(hash)
            if upgrades:
                upgrades -= 1
                upgrade = new_package_version(skeleton, skeleton.version + "-1")
                hash = market.add(upgrade)
                available.add(hash)
                available_upgrade.add(hash)
            if multiple_upgrades:
                multiple_upgrades -= 1
                upgrade = new_package_version(skeleton, skeleton.version + "-2")
                hash = market.add(upgrade)
                available.add(hash)
                available_upgrade.add(hash)
            elif usn_count:
                package = market.get(hash)
                usn = USN(
                    "%03d-1" % usn_count,
                    "Security issue %d" % usn_count,
                    usn_release_base + timedelta(days=usn_count),
                )
                package.usn = usn
                usn_count -= 1

    available_ids = market.get_ids(available)
    installed_ids = market.get_ids(installed)
    upgrades_ids = market.get_ids(available_upgrade)
    held_ids = market.get_ids(held)
    for computer in computers:
        computer.build_packages(available_ids, installed_ids, upgrades_ids, held_ids)

    reporter_api = PackageReporterResultAPI(computers[0].computer.account)

    for computer in computers[:2]:
        reporter_api.create_package_reporter_result(
            computer.computer, 2, "Some error happened.", timestamp=datetime.utcnow()
        )

    # Create a USN from "29 days ago", and mark one computer as having
    # been patched 28 days ago, one as being patch 13 days ago and one as
    # not yet patched for the reporting page.
    now = datetime.utcnow()
    twenty_nine_days_ago = now - timedelta(days=29)
    twenty_eight_days_ago = now - timedelta(days=28)
    thirteen_days_ago = now - timedelta(days=13)

    package_store = get_package_store()
    usn = USN("1234-1", "A test USN.", release_time=twenty_nine_days_ago)
    package_store.add(usn)

    issue1 = USNIssue(usn.name, computers[0].computer.id)
    issue1.resolve_time = twenty_eight_days_ago
    computers[0].computer.store.add(issue1)

    issue2 = USNIssue(usn.name, computers[1].computer.id)
    issue2.resolve_time = thirteen_days_ago
    computers[1].computer.store.add(issue2)

    return available


def build_usn_issues():
    """Populate USNIssue table based on security upgrades."""
    management = USNIssueManagement()
    store = get_package_store()

    result = store.find(USN.name)

    for usn_name in result:
        management.check_usn(usn_name)


def build_repository_profiles(account, person):
    """Create some repository profiles for the account."""
    management = RepositoryManagement(account, person)
    server_tag = get_tag_by_name("server")
    precise_tag = get_tag_by_name("precise")
    management.create_profile("testing", "Testing versions")
    profile2 = management.create_profile("development", "Development versions")
    profile2.update_tags([server_tag, precise_tag])
    profile3 = management.create_profile("production", "Production versions")

    computer = account.get_computers(tags=(server_tag,))[0]
    activity = profile3.get_pending_activities((computer.id,))[0]
    activity.start()
    output = read_file(os.path.join(SAMPLE_FILES_PATH, "package-install.out")).decode(
        "UTF-8"
    )
    activity.succeed(text=output)


def build_repository_pockets(account, person):
    """Create some repository pockets for the account."""
    gpg_key_management = GPGKeyManagement(account, person)
    # Create a sample GPG key (not used).
    gpg_key_management.create_key("sample", SAMPLE_KEY_FINGERPRINT, SAMPLE_KEY_MATERIAL)
    # Create a GPG key used to sign repositories.  It is used also as key for
    # the mirror_gpg_key since the local aptserver also signs lists with this
    # key.
    sign_key = gpg_key_management.create_key(
        "sign-key",
        SAMPLE_SECRET_KEY_FINGERPRINT,
        SAMPLE_SECRET_KEY_PUBLIC_MATERIAL,
        SAMPLE_SECRET_KEY_MATERIAL,
    )
    management = RepositoryManagement(account, person)
    distribution = management.create_distribution("ubuntu")
    series = distribution.create_series("lucid")
    pocket1 = series.create_pocket(
        components=["main"], archs=["i386"], gpg_key=sign_key, include_udeb=False
    )
    pocket1.set_mirror_mode("http://localhost:8085/", "lucid", sign_key)
    pocket2 = series.create_pocket(
        "python2.7", components=["main"], archs=["i386", "source"]
    )
    pocket2.set_upload_mode()

    if not is_standalone():
        # The local APT server is available only in devmode
        os.system("./bin/aptserver -n --step")

    activity = management.create_sync_pocket_activity(pocket=pocket1)
    [child] = activity.children()
    child.start()
    child.advance(35)

    activity = management.create_sync_pocket_activity(pocket=pocket2)
    [child] = activity.children()
    child.start()
    child.succeed()


def build_custom_graph(account, person, computers, start_time, stop_time):
    management = CustomGraphManagement(account.account, person)
    custom_graph1 = management.create_graph(
        "Web requests", "Requests per second", "user", "#!/bin/sh\necho 1.0"
    )
    custom_graph2 = management.create_graph(
        "Database requests", "Requests per second", "user", "#!/bin/sh\necho 1.0"
    )

    server_tag = get_tag_by_name("server")
    custom_graph1.update_tags([server_tag])
    custom_graph2.update_tags([server_tag])
    for computer in computers:
        graph1 = computer.computer.get_custom_graph_data_by_id(custom_graph1.id)
        graph2 = computer.computer.get_custom_graph_data_by_id(custom_graph2.id)
        time = start_time
        data1 = random() * computer.computer.id
        data2 = random() * (computer.computer.id * 10)
        while time <= stop_time:
            graph1.append_data(time, data1)
            graph2.append_data(time, data2)
            data1 = abs(data1 + random() * randint(-2, 2))
            data2 = abs(data2 + random() * randint(-2, 2))
            time += computer.interval
        if graph1.id % 3:
            graph1.error = "ValueError: invalid literal for float(): error"
        if graph2.id % 2:
            graph2.error = "ValueError: invalid literal for float(): error"


def build_scripts(account, person):
    """Create sample scripts in the script library for the given account."""
    landscape_vars_script = """import json
from os import environ
lsvars = {}
for key, value in environ.items():
    if key.startswith("LANDSCAPE_"):
        lsvars[key] = value
print(json.dumps(lsvars))
"""
    management = ScriptManagement(account.account, person)
    management.create_script(
        "List temporary files", "#!/bin/bash\nls /tmp", 10, "nobody", {}
    )
    code = "#!/bin/bash\npython3 $LANDSCAPE_ATTACHMENTS/run.py"
    management.create_script(
        "Execute python attachment", code, 20, "nobody", {"run.py": "print('OK!')"}
    )
    code = "#!/usr/bin/python3\n" + landscape_vars_script
    management.create_script(
        "Show Landscape environment variables",
        code,
        20,
        "nobody",
        {"hello.txt": "Hello!"},
    )
    management.create_script(
        "delete me",
        "#!/bin/bash\nls -la $LANDSCAPE_ATTACHMENTS",
        10,
        "nobody",
        {"otherhello.txt": "hey!"},
    )


def build_package_activities(person, computers, available_hashes):
    """Create sample package activities.

    @param available_hashes: A list of available package hashes.
    @param person: A L{Person} to use as the creator of package activities.
    @param computers: A list of L{SampleComputer} instances.
    """
    computers = list(computers)
    account = computers[0].computer.account

    selection = ComputerSelection(account)
    computer_ids = [sample_computer.computer.id for sample_computer in computers]
    selection.select_computers(computer_ids)
    package_management = PackageManagement(selection)
    package_management.request_upgrades(person)

    market = PackageMarket()
    find_package = lambda name: market.search(name, hashes=available_hashes)[0]
    installing = (
        find_package("python-twisted-snmp").id,
        find_package("python-twisted-web2").id,
    )
    removing = (find_package("python-bittorrent").id,)
    changes = [
        (computer_id, installing, removing, (), ()) for computer_id in computer_ids
    ]
    package_management.change_packages(person, changes)
    installing = (
        find_package("python-twisted-mail").id,
        find_package("python-twisted-conch").id,
    )
    changes = [(computer_id, installing, (), (), ()) for computer_id in computer_ids]
    activity_group = package_management.change_packages(person, changes)
    output = read_file(os.path.join(SAMPLE_FILES_PATH, "package-install.out")).decode(
        "UTF-8"
    )
    for activity in get_child_activities(activity_group):
        activity.start()
        activity.succeed(text=output)

    installing = (
        find_package("python-nevow").id,
        find_package("python-xml").id,
    )
    changes = [(computer_id, installing, (), (), ()) for computer_id in computer_ids]
    activity_group = package_management.change_packages(person, changes)
    for activity in get_child_activities(activity_group):
        activity.start()
        activity.fail(text="Don't do that please!")

    parent = account.create_activity(
        ActivityGroup, creator=person, summary=CHANGE_PACKAGES_SUMMARY
    )
    activity = parent.add(ChangePackagesRequest, computer=computers[0].computer)
    parent.settle()
    activity.info.add_changes([find_package("python-pcapy").id])
    activity.start()
    activity.block()

    parent = account.create_activity(
        ActivityGroup, creator=person, summary=CHANGE_PACKAGES_SUMMARY
    )
    activity = parent.add(ChangePackagesRequest, computer=computers[3].computer)
    parent.settle()
    activity.info.add_changes([find_package("python-adns").id])
    activity.start()
    activity.block()

    # Create a group activity for applying a profile on two computers
    account_profiles = PackageProfilesManagement(account, person)
    profile = account_profiles.get_package_profiles().one()
    profile.on_create(person)
    profile.update_tags([get_tag_by_name("laptop")])
    group = profile.get_pending_root_activities().one()
    [child1, child2, child3] = get_child_activities(group)
    child1.info.add_changes([find_package("apache2").id])
    child1.start()
    child1.block()
    child2.info.add_changes([find_package("postgresql-8.3").id])
    child2.start()
    child2.succeed()


def build_script_activities(person, computers):
    """Create sample execute-script activities.

    @param person: A L{Person} to use as the creator of activities.
    @param computers: A list of L{SampleComputer} instances.
    """
    account = computers[0].computer.account
    parent = account.create_activity(
        ActivityGroup, creator=person, summary=EXECUTE_SCRIPT_SUMMARY
    )
    for index, computer in enumerate(computers):
        computer.build_script_activities(parent)
    # Create a single activity in blocked status for viewing.
    # Typically this will be a CreateProjectRequest activity within an
    # OpenstackRegionDeploy, but we want simple sample data for viewing.
    activity = parent.children()[0]
    activity.block()
    parent.settle()


def _get_packages_and_upgrades():
    """
    Return a 2-tuple with the following:

    - a dict mapping package names to a list of package ids for the name. The
      list can contain 1 or 2 elements; in the latter case they are sorted by
      ascending version.
    - a list of names for packages with upgrades.
    """
    names_to_ids = {}
    names_with_upgrades = []

    package_store = get_package_store()
    result = package_store.find(
        (
            Package.name,
            Func("ARRAY_AGG", Package.id),
            Func("ARRAY_AGG", Package.version),
        )
    )
    result.group_by(Package.name)

    for name, ids, versions in result:
        if len(ids) > 1:
            names_with_upgrades.append(name)
            # the test database has exactly 2 versions for each package with
            # upgrades. Sort the ids as (upgraded, upgrades)
            if apt_pkg.version_compare(versions[0], versions[1]) > 0:
                ids[0], ids[1] = ids[1], ids[0]
        names_to_ids[name] = ids[:2]

    return names_to_ids, names_with_upgrades


def _build_computer_packages(
    names_to_ids,
    names_with_upgrades,
    available_count,
    upgrades_count,
    installed_count,
    held_count,
):
    """
    Build sample package data from the provided package info and counts.

    @returns: a tuple of lists with available, upgrades, installed and held
        package ids.
    """
    # get some package names which have more than one version.
    upgrades = sample(names_with_upgrades, upgrades_count)
    # get some other packages that are available but don't have upgrades.
    available = sample(
        set(names_to_ids) - set(upgrades), available_count - upgrades_count
    )

    upgrade_ids = []
    installed_ids = []

    for name in upgrades:
        upgraded_id, upgrades_id = names_to_ids[name]
        installed_ids.append(upgraded_id)
        upgrade_ids.append(upgrades_id)

    # add to installed some packages that don't have upgrades.
    installed_not_upgrades = sample(
        set(names_to_ids) - set(upgrades) - set(available),
        installed_count - upgrades_count,
    )
    installed_ids.extend(names_to_ids[name][0] for name in installed_not_upgrades)

    # available packages also includes upgrades.
    available_ids = [names_to_ids[package][0] for package in available]
    available_ids.extend(upgrade_ids)

    # mark some installed packages as held.
    held_ids = sample(installed_ids, held_count)

    return available_ids, upgrade_ids, installed_ids, held_ids


def build_package_test_computers_package_data(account, computers_count):
    """
    Build sample package data for computers.
    """
    # number of packages in each state for each computer.
    AVAILABLE_PER_COMPUTER = 38000
    UPGRADES_PER_COMPUTER = 500
    INSTALLED_PER_COMPUTER = 2000
    HELD_PER_COMPUTER = 10
    # comptuers for each group will have the same package info.
    COMPUTERS_PER_GROUP = 500

    [tag] = get_tags(["package-test"])

    names_to_ids, names_with_upgrades = _get_packages_and_upgrades()

    iterations = lambda total, by: old_div(total, by) + int(bool(total % by))

    # for each computers group, build the same package data.
    for i in range(0, iterations(computers_count, COMPUTERS_PER_GROUP)):
        start = i * COMPUTERS_PER_GROUP + 1
        end = min((i + 1) * COMPUTERS_PER_GROUP, computers_count)
        print("Building package data from computers %d to %d..." % (start, end))

        available_ids, upgrade_ids, installed_ids, held_ids = _build_computer_packages(
            names_to_ids,
            names_with_upgrades,
            AVAILABLE_PER_COMPUTER,
            UPGRADES_PER_COMPUTER,
            INSTALLED_PER_COMPUTER,
            HELD_PER_COMPUTER,
        )

        computer_ids = []
        for n in range(start, end + 1):
            computer = account.account.create_computer(
                "on-secure-pkg%d" % n, "Package test %d" % n, "package-testhost-%d" % n
            )
            computer.add_tag(tag)
            computer_ids.append(computer.id)

        # this assumes all computers are in the same store.
        resource_store = computer.store
        resource_store.execute(
            "UPDATE computer_packages "
            "SET available = ?, available_upgrades = ?, "
            "    installed = ?, held = ? "
            "WHERE computer_id IN ?",
            (available_ids, upgrade_ids, installed_ids, held_ids, tuple(computer_ids)),
        )
