# Copyright 2005-2022 Canonical Limited. All rights reserved.

from zope.security.interfaces import Unauthorized
from zope.security.proxy import removeSecurityProxy
from zope.interface import directlyProvides

from canonical.lib.copycat import Copycat
from canonical.routes import route, Redirect, BadRequest
from canonical.routes.publisher import get_route_object_by_interface
from canonical.url import get_url

from canonical.landscape.security.interfaces import IPersonPrincipal
from canonical.landscape.model.activity.parent import get_parent_activity
from canonical.landscape.model.activity.activity import get_child_activity
from canonical.landscape.model.activity.routes import AccountActivities
from canonical.landscape.model.script.management import ScriptManagement
from canonical.landscape.model.account.customgraph import CustomGraphManagement
from canonical.landscape.model.license.management import LicenseManagement
from canonical.landscape.model.selection.selection import (
    SessionComputerSelection,
    ComputerSelection,
)
from canonical.landscape.model.configuration import (
    is_testing_environment,
    is_standalone,
)
from canonical.landscape.model.main.accesscontext import AccountAccessContexts
from canonical.landscape.model.main.interfaces import IAccountPerson, IAccount
from canonical.landscape.model.package.profile import PackageProfilesManagement
from canonical.landscape.model.package.autoupgrade import UpgradeProfileManagement
from canonical.landscape.model.computer.removal import RemovalProfileManagement
from canonical.landscape.model.account.savedsearch import SavedSearchManagement
from canonical.landscape.ui.computer.pending import PendingComputers
from canonical.landscape.ui.knowledge.routes import AccountKnowledgeBase
from canonical.landscape.ui.alert.rule import AlertRules
from canonical.landscape.ui.alert.view import get_alert_view
from canonical.landscape.ui.routes import LandscapeRoutes
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.account.invitation import InvitationManagement
from canonical.landscape.model.access.management import AccessControlManagement
from canonical.landscape.model.access.errors import UnknownRoleError
from canonical.landscape.service.accesscontext.lookup import AccessContextLookup
from canonical.landscape.service.errors import UnauthorisedError
from canonical.landscape.service.billing.lookup import BillingServiceLookup
from canonical.landscape.model.eventlog.management import EventLog


class AccountRoutes(LandscapeRoutes):

    CHECK_ANONYMOUS = True

    @route("activities")
    def activities(self):
        return AccountActivities(self.context)

    @route("activity/<id>", id_pattern="\d+")
    def activity(self, id):
        account = removeSecurityProxy(self.context)
        activity_id = int(id)
        # The route currently supports both parent and child activities, so try
        # both. There's no risk of clashing since IDs are unique across parents
        # and children, and we plan to maintain uniqueness.
        activity = get_parent_activity(account, activity_id)
        if activity is None:
            activity = get_child_activity(account, activity_id)
        return activity

    @route("alert/<id>/resolve", id_pattern="\d+")
    def resolve_alert(self, id):
        person = self.request.principal.person
        rule = self.context.get_rule_by_id(int(id), person=person)
        if not rule:
            return None
        if rule.info.is_active():
            url = get_alert_view(rule, self.request).resolution_url
            if not url:
                url = get_url(rule, self.request)
            raise Redirect(url)
        else:
            self.notes.warning("This alert has already been resolved.")
            raise Redirect(get_url(rule, self.request))

    @route("alert/<id>", id_pattern="\d+")
    def alert(self, id):
        person = self.request.principal.person
        return self.context.get_rule_by_id(int(id), person=person)

    @route("alerts")
    def alerts(self):
        return AlertRules(self.context)

    @route("administrators")
    def administrators(self):
        person = self.request.principal.person
        return InvitationManagement(self.context, person)

    @route("computers")
    def computers(self):
        selection = SessionComputerSelection(
            self.context, self.request.principal.person
        )
        if "select_all" in self.request:
            selection.select_computers(selection.get_available_ids())
        elif "select_none" in self.request:
            selection.clear_selected_computers()
        elif "computer_ids" in self.request:
            try:
                computer_ids = set(
                    int(id) for id in self.request["computer_ids"].split(",")
                )
            except ValueError:
                raise BadRequest("Invalid computer ids in request")

            selected_computer_ids = set(selection.get_selected_computer_ids())
            selection = ComputerSelection(self.context)
            filtered_ids = computer_ids.intersection(selected_computer_ids)
            selection.include_computers(filtered_ids)
            selection.select_computers(filtered_ids)

        return selection

    @route("computer/<id>")
    def computer(self, id):
        try:
            if "+" in id:
                ids = [int(x) for x in id.split("+")]
                selection = ComputerSelection(self.context)
                selection.include_computers(ids)
                selection.select_computers(ids)
                return selection
            else:
                return self.context.get_computer(int(id))
        except ValueError:
            return None

    @route("pending-computers")
    def pending_computers(self):
        return PendingComputers(self.context, select_all="select_all" in self.request)

    @route("pending-computer/<id>", id_pattern="\d+")
    def pending_computer(self, id):
        return self.context.get_pending_computer_by_id(
            self.request.principal.person, int(id)
        )

    @route("kb")
    def kb(self):
        return AccountKnowledgeBase(self.context)

    @route("scripts")
    def scripts(self):
        return ScriptManagement(self.context, self.request.principal.person)

    @route("script/<id>", id_pattern="\d+")
    def script(self, id):
        management = ScriptManagement(self.context, self.request.principal.person)
        return management.get_scripts(ids=[int(id)]).one()

    @route("custom-graphs")
    def custom_graphs(self):
        return CustomGraphManagement(self.context, self.request.principal.person)

    @route("custom-graph/<id>", id_pattern="\d+")
    def custom_graph(self, id):
        management = CustomGraphManagement(self.context, self.request.principal.person)
        return management.get_graphs(ids=[int(id)]).one()

    @route("access-group/<id>", id_pattern="\d+")
    def access_context(self, id):
        return self._service_lookup(AccessContextLookup, self.context, int(id))

    @route("access-groups", permission="landscape.ViewAccessContextManagement")
    def access_contexts(self):
        return AccountAccessContexts(self.context)

    @route("package-profile")
    def package_profile(self):
        return PackageProfilesManagement(self.context, self.request.principal.person)

    @route("removal-profile")
    def removal_profile(self):
        return RemovalProfileManagement(self.context, self.request.principal.person)

    @route("saved-searches")
    def saved_searches(self):
        return SavedSearchManagement(self.context, self.request.principal.person)

    @route("upgrade-profile")
    def upgrade_profile(self):
        return UpgradeProfileManagement(self.context, self.request.principal.person)

    @route("settings")
    def settings(self):
        copycat = Copycat(removeSecurityProxy(self.request.principal.person))
        directlyProvides(copycat, IAccountPerson)
        Copycat.activate(copycat)
        return copycat

    @route("person/<id>", permission="landscape.ViewPersonResource")
    def person(self, id):
        try:
            person_id = int(id)
        except ValueError:
            return None
        person = get_person_by_id(person_id)
        if person is not None:
            copycat = Copycat(person)
            directlyProvides(copycat, IAccountPerson)
            Copycat.activate(copycat)
            return copycat

    @route("license/<id>")
    def license(self, id):
        license_management = LicenseManagement(self.context)
        return license_management.get_license_by_id(int(id))

    @route("roles")
    def roles(self):
        return AccessControlManagement(self.context, self.request.principal.person)

    @route("role/<name>")
    def role(self, name):
        management = AccessControlManagement(
            self.context, self.request.principal.person
        )
        try:
            return management.get_role(name)
        except UnknownRoleError:
            return None

    @route("event-log", permission="landscape.ViewEventLog")
    def eventlog(self):
        return EventLog(self.context, self.request.principal.person)

    @route("billing")
    def billing(self):
        """Return a BillingManagement object or raise Unauthorized.

        User permissions are queried to determine the return value.
        """
        lookup = BillingServiceLookup(self.request.principal)
        try:
            context = lookup(self.context)
        except UnauthorisedError:
            raise Unauthorized()
        return context


def get_current_account(request):
    """Get the current account the request is for, if any."""
    if IPersonPrincipal.providedBy(request.principal):
        return get_route_object_by_interface(IAccount, request)
