# Copyright 2005-2023 Canonical Limited.  All rights reserved.

from zope.browserpage import ViewPageTemplateFile
from zope.schema import TextLine
from zope.security.proxy import removeSecurityProxy
from zope.cachedescriptors.property import Lazy

from canonical.content.form import Fields, action
from canonical.url import get_url

from canonical.landscape.model.package.interfaces import IPackageProfile
from canonical.landscape.model.package.profile import PackageProfilesManagement

from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.form import LandscapeFormProvider
from canonical.landscape.ui.profile.create import (
    extract_constraints,
    form_constraint_helper,
)


class EditPackageProfilePage(AccountPage):

    title = "Edit package profile"

    def get_computer_ids_with_installed_packages(self):
        """
        We do not allow PackageProfile edit page to specify a computer.
        Return an empty list so as not to allow choosing a computer.
        """
        return list()


class EditPackageProfileForm(LandscapeFormProvider):

    additional_content = ViewPageTemplateFile("create-fields.pt")
    package_error = ""

    def __init__(self, context, request, view):
        super(EditPackageProfileForm, self).__init__(context, request, view)
        self.is_create = False  # report to zope we are not the create action
        self.summary = "Edit package profile '%s'" % context.title
        self.raw_constraints = list(context.constraints)

        if request.method == "POST":
            self.constraints = form_constraint_helper(request.form)
        else:
            self.constraints = []
            for idx, raw_constraint in enumerate(self.raw_constraints):
                kind, package, operator, version = raw_constraint.parse()
                self.constraints.append(
                    {
                        "relation": kind,
                        "package": package,
                        "operator": operator,
                        "version": version,
                        "idx": str(idx),  # Constraints are numbered in the HTML
                    }
                )

    @Lazy
    def fields(self):
        fields = Fields(IPackageProfile, render_context=True)
        fields = fields.select("title", "description")
        fields += Fields(
            TextLine(
                __name__="access_context",
                title="Access group",
                readonly=True,
                default=self.context.access_context.title,
            )
        )
        return fields

    @action("Save", primary=True)
    def save(self, action, data):
        package_constraints, self.package_error = extract_constraints(self.request.form)
        if self.package_error:
            return

        management = PackageProfilesManagement(
            self.context.account, self.request.principal.person
        )

        add_constraints = []
        remove_constraints = []
        original_constraints = dict(
            (removeSecurityProxy(constraint.parse()), constraint.id)
            for constraint in self.raw_constraints
        )
        original_constraints_set = set(original_constraints)
        if original_constraints_set != package_constraints:
            add_constraints = package_constraints - original_constraints_set
            remove_constraints = original_constraints_set - package_constraints

        management.edit(
            self.context.name,
            title=data.get("title"),
            description=data.get("description"),
            add_constraints=add_constraints,
            remove_constraints=remove_constraints,
        )

        self.request.response.redirect(get_url(self.context, self.request))
